/* 
**
**  AuditSystem.h - a set of routines to implement
**	an event log feature in the IOC.
**
**  int AuditInitLog (char *file,	- audit file (Audit.log is default)
**		      int maxrecords,	- the number of records
**		      int varSize);	- varSize portion in each record
**
**  int AuditAddEvent ( int log,	- log identifier from AuditInitLog
**			int Class,	- report group (see existing definitions
**			int Level,	- error, warning, info, auditSuccess, auditFail
**			char *IDString, - authority id (CN, ...)
**			char *TaskName, - audit source code
**			char *Vartext, int VarSize); - free text area for addition info
**
**  int AuditSearchEvents (int log,	- log identifier from AuditInitLog
**			int Class, int Level, char *IDString, char *TaskName);
**
**  AUDIT_REC *AuditNextEvent (int log);- log identifier from AuditInitLog
**
**  void AuditClearEvents (char *auditfile,- file name
**			int CutOffTime);
**
**  void AuditClearAll (int log);	- log identifier from AuditInitLog
**
**  void AuditCloseLog (int log);	- log identifier from AuditInitLog
**
**  int AuditReportFile (int log,	- log identifier from AuditInitLog
**			 char *FileName,- filename of output
**			 int CutOffTime);-stop reporting at this time
**
**  void AuditPrintFile (char *auditfile,-audit file to print
**			 int Class, int Level, char *IDString, char *TaskName)
**
**  void AuditTail (	int numRecs,	- how many records to show
**			char *auditfile,-audit file to print
**			int Class, int Level, char *IDString, char *TaskName);
**
*/


#include "sslShared.h"	/* get maximum DN length */

/* Audit class definitions (must be power of 2 values to allow OR searching) */
typedef enum {
    AUDIT_FILE = 1,
    AUDIT_LOGON = 2,
    AUDIT_LOGOFF = 4,
    AUDIT_READ = 8,
    AUDIT_WRITE = 16,
    AUDIT_CREATE = 32,
    AUDIT_SESSION = 64
} AUDIT_CLASS;

/* Audit level definitions (must be power of t2 values to allow OR searching */
typedef enum {
    AUDIT_ERROR = 1,
    AUDIT_WARNING = 2,
    AUDIT_INFO = 4,
    AUDIT_SUCCESS = 8,
    AUDIT_FAIL = 16
} AUDIT_LEVEL;

typedef char	AUDIT_IDENT [X509_MAX_DN];
typedef char	AUDIT_TASK [32];

typedef struct {
    char	fileName[32];	    /* name of audit file */
    int		maxRecords;	    /* max number of records */
    int		varSize;	    /* size of variable part of record */
    int		nextRec;	    /* next record location */
    int		firstRec;	    /* index of first record */
    int		written;	    /* is log written to yet? */
} AUDIT_HEADER;

typedef struct audit_log {
    struct audit_log *nextLog;	    /* chain of active logs */
    AUDIT_HEADER header;	    /* header (as in file */
    int		fd;		    /* file descriptor */
    int		usecnt;		    /* user count */
#ifdef vxWorks
    SEM_ID	mxSem;		    /* mutual exclusion semaphore */
#endif vxWorks
} AUDIT_LOG, *AUDIT_ID;

typedef struct {
    time_t	auditTime;	    /* time of record */
    int		Class;		    /* audit class */
    int		Level;		    /* audit level */
    AUDIT_IDENT	IDString;	    /* ID */
    AUDIT_TASK	TaskName;	    /* task generating audit */
} AUDIT_REC;

typedef struct audit_search {
    struct audit_search *next;	    /* chain of active search controls */
    int		firstRec;	    /* id of first record at start of search */
    int		nextRec;	    /* id of next record to return */
    int		taskID;		    /* owner task */
    AUDIT_LOG	*log;		    /* log identifier */
    int		Class;		    /* select audit class - OR of class bits */
    int		Level;		    /* select aidit level - OR of level bits */
    AUDIT_IDENT	IDString;	    /* ID search */
    AUDIT_TASK	TaskName;	    /* task generating audit */
    AUDIT_REC	*retRec;	    /* search record data */
} AUDIT_SEARCH;

/* Published data */
extern AUDIT_ID auditLog;
extern char systemLogFile [];

/* define the audit routines */
AUDIT_ID InitSystemLog (char *logfile, int maxRecs);

AUDIT_ID AuditInitLog (char *file, int maxrecords, int varSize);

int AuditAddEvent (AUDIT_ID log, int Class, int Level, AUDIT_IDENT IDString, char *TaskName,
		    char *Vartext);

int AuditSearchEvents (AUDIT_ID log, int Class, int Level, AUDIT_IDENT IDString, AUDIT_TASK TaskName);

AUDIT_REC *AuditNextEvent (AUDIT_ID log);

void AuditClearEvents (char *auditfile, int CutOffTime);

void AuditClearAll (AUDIT_ID log);

void AuditCloseLog (AUDIT_ID log);

int AuditReportFile (char *auditFile, char *FileName, int CutOffTime);

void AuditPrintFile (char *auditfile,
		     int Class, int Level, char *IDString, char *TaskName);

void AuditTail (int numRecs, char *auditfile,
		int Class, int Level, char *IDString, char *TaskName);

void AuditPrint (AUDIT_REC *rec, int varSize, FILE *file);

int AuditVarSize (AUDIT_ID log);
