/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	   CHANNEL ACCESS DEFINITIONS FOR APPLICATIONS		*/
/*								*/
/*	This file defines the basic control blocks and macros	*/
/*	that are required for an application to use the new	*/
/*	implementation of channel access. We have gone to	*/
/*	great pains to keep the two user interface structures	*/
/*	with the same size and layout so that applications	*/
/*	won't have to be recompiled, just relinked with our	*/
/*	new libraries. That is our goal. If we can't achieve	*/
/*	that, then the user will have to recompile. That isn't	*/
/*	too bad because it only takes a few minutes. However,	*/
/*	100% source compatability with the previous implemen-	*/
/*	tation of channel access is a mandatory requirement.	*/
/*								*/
/****************************************************************/
/*	DANGER!	DANGER!	WARNING! WARNING! DANGER! DANGER!	*/
/****************************************************************/
/*								*/
/*	This header file has the same name as the one in the	*/
/*	original implementation of channel access. I had to	*/
/*	do this in order to get 100% source compatibility.	*/
/*	Be careful not to mix these versions together. You	*/
/*	will not like the results.				*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#ifndef INCLcadef
#define INCLcadef

#include "dbDefs.h"
#include "db_access.h"
#include "dataTypes.h"
#include "agentErrs.h"
#include "tisAlarm.h"

#ifdef AXP
#ifndef POINTER64
#define POINTER64
#endif
#ifndef Pointer64fromToken
#define Pointer64fromToken(tok)	TokenPointers[tok-500000]
extern  int   TokenFromPointer64 (void *);
extern	void *TokenPointers[];
#endif
#endif

/******
	The user's interface to a channel is the channel identifier,
	really a pointer to the following structure. The fields that
	are in the original structure but not used in this implemen-
	tation will have their original names and a note stating
	that we're not using them.

	In the original implementation, the channel name followed
	the structure in memory. I included them in the structure
	so that the code will not be as messy looking. This makes
	the structure a bit bigger, but that won't matter because
	the new stuff has a higher offset than the original stuff.
******/

enum chState
    {
    neverOpened,
    prevOpened,
    currOpen,
    closed	/* I don't need this one */
    };

typedef struct chanBlock
    {
    struct chanBlock	*pNextChan;
    struct chanBlock	*pPrevChan;
    short		nativeType;
    short		nativeCount;
    short		iocix;		/* I don't need this */
    void		*paddr;		/* Used  for taskID in vxWorks! */
    void		*puser;
    enum chState	chanState;
    short		buildType;
    short		buildCount;
    short		nameLength;
    void		*buildValue;
    void		(*connFunc)();
    struct evntBlock	*pFirstEvnt;
    struct evntBlock	*pPrevEvnt;	/* I don't need this */
    int			evntCount;	/* I don't need this */
    int			handle;		/* This is new */
    void		(*callBack)();
    void		*callBackArg;
    char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
#ifdef vxWorks
    void		*pDbAddr;
#endif vxWorks
#ifdef POINTER64
    int			chidToken;
#endif
    } *chid;

/******
	The user's interface to an event is the event identifier,
	really a pointer to the following structure.
******/

typedef struct evntBlock
    {
    struct evntBlock	*pNextEvnt;
    struct evntBlock	*pPrevEvnt;
    void		(*userFunc)();
    void		*userArg;
    chid		pChan;
    int			type;		/* I don't need this (yet?) */
    int			count;		/* I don't need this (yet?) */
    float		float1;		/* Reserved for expansion */
    float		float2;		/* Reserved for expansion */
    float		float3;		/* Reserved for expansion */
    unsigned short	mask;
    } *evid;

/****************************************************************/
/*								*/
/*	TYPEDEFS AND OTHER MACROS FOR CHANNEL ACCESS USERS	*/
/*								*/
/****************************************************************/

typedef int	chtype;

/****************************************************************/
/*								*/
/*	DEFINITIONS REQUIRED FOR SOURCE COMPATIBILITY		*/
/*								*/
/*	The following definitions are equivalent to the ones	*/
/*	in the original <cadef.h>. We have to reimplement	*/
/*	them here so that users will get equivalent structure	*/
/*	and symbol definitions when they include this file	*/
/*	instead of the original one. Once again, 100 percent	*/
/*	source compatibility is a mandatory requirement.	*/
/*								*/
/****************************************************************/

#define	CA_OP_CONN_UP	6
#define CA_OP_CONN_DOWN	7
#ifdef __cplusplus
// some c++ compilers are unhappy with variable "chid" and typedef "chid"
struct connection_handler_args
    {
    chid	chanid;
    long	op;
    };

struct event_handler_args
    {
    void	*usr;
    chid	chanid;
    long	type;
    long	count;
    void	*dbr;
    };

struct	exception_handler_args
    {
    void	*usr;
    chid	chanid;
    long 	type;
    long 	count;
    void 	*addr;
    long	stat;
    long	op;
    char	*ctx;
    };
#else
struct connection_handler_args
    {
    chid	chid;
    long	op;
    };

struct event_handler_args
    {
    void	*usr;
    chid	chid;
    long	type;
    long	count;
    void	*dbr;
    };

struct	exception_handler_args
    {
    void	*usr;
    chid	chid;
    long 	type;
    long 	count;
    void 	*addr;
    long	stat;
    long	op;
    char	*ctx;
    };
#endif

#define TYPENOTCONN	(-1)

#ifndef	DBE_VALUE
#define	DBE_VALUE	AGNT_VALUE
#define DBE_LOG		AGNT_LOG
#define	DBE_ALARM	AGNT_ALARM
#endif	!DBEVALUE

#define	ECA_ADDFAIL	AGNT_ADDFAIL
#define ECA_ALLOCMEM	AGNT_ALLOCMEM
#define ECA_BADCHID	AGNT_BADCHID
#define ECA_BADCOUNT	AGNT_BADCOUNT
#define ECA_BADTYPE	AGNT_BADTYPE
#define ECA_DISCONN	AGNT_DISCONN
#define ECA_EVDISALLOW	AGNT_EVDISALLOW	
#define ECA_GETFAIL	AGNT_GETFAIL
#define ECA_NOCACTX	AGNT_NOCACTX
#define ECA_NORMAL	AGNT_NORMAL
#define ECA_PUTFAIL	AGNT_PUTFAIL
#define ECA_STRTOBIG	AGNT_STRTOBIG
#define ECA_TIMEOUT	AGNT_TIMEOUT

enum channel_state
    {
    cs_never_conn, 
    cs_prev_conn, 
    cs_conn, 
    cs_closed
    };

/****************************************************************/
/*								*/
/*	I HATE THINGS THAT LOOK LIKE FUNCTIONS THAT ARE		*/
/*	IMPLEMENTED AS MACROS WITH A STEADFAST PASSION!		*/
/*								*/
/*	The following functions were #defines in the original	*/
/*	implementation of ca. I made them real functions in	*/
/*	aliasDefs.c. That way, you might be able to debug	*/
/*	something some day.					*/
/*								*/
/****************************************************************/

char	*ca_name(chid);
short	ca_field_type(chid);
short	ca_element_count(chid);
char	*ca_message(int);
void	*ca_puser(chid);
enum channel_state ca_state(chid);

/****************************************************************/
/*								*/
/*		FUNCTION PROTOTYPES FOR BUFFER MANAGMENT	*/
/*								*/
/*	The following routines are used to allocate and		*/
/*	deallocate chid and evid structures. These routines	*/
/*	are internal to the channel access library and should	*/
/*	never be called by the end user.			*/
/*								*/
/****************************************************************/

chid	newCHID(void);
void	deleteCHID(chid chid);
evid	newEVID(void);
void	deleteEVID(evid);

/****************************************************************/
/*								*/
/*		TRANSMIT BUFFER SIZE				*/
/*								*/
/*	This is the size of the buffer that stacks up msgs	*/
/*	from the application to Channel Acquisition and		*/
/*	Control. The buffer actually lives in globalData.c.	*/
/*								*/
/****************************************************************/

#define	CA_TX_BUFF_SIZE		1024

/****************************************************************/
/*								*/
/*		RECEIVE BUFFER SIZE				*/
/*								*/
/*	This is the size of the buffer that is read from	*/
/*	Channel Acquisition and Control. Note that it is	*/
/*	bigger because, in general, channel access apps		*/
/*	receive much more data than they send.			*/
/*								*/
/****************************************************************/

#define CA_RX_BUFF_SIZE		5120


/****************************************************************/
/*								*/
/* Function prototypes						*/
/*								*/
/****************************************************************/

int ca_task_initialize (void);
int ca_add_fd_registration(void (*func)(), void *arg);
int ca_add_exception_event (void (*pfunc)(), void *arg);
int ca_register_as_archiver(void (*pDataFunc)(), void (*pConnFunc)());
int ca_register_as_alarm_handler(void (*pDataFunc)(tisAlarm *), 
				 void (*pConnFunc)(char *, short));
int ca_get_params (
    char	*chanName,	/* Channel Name */
    short	type,		/* DBR_xxx_PARAM (see paramStructs.h) */
    void	(*paramFunc)(),	/* Callback routine */
    void	*pArg		/* Argument for the callback */
);
int ca_pend_event (float timeOut);
int ca_pend_io (float timeOut);
int ca_clear_channel (chid pChid);
int ca_clear_event (evid pEvnt);
int ca_add_event (int type, void *chanPtr,
	      void (*funcPtr)(struct event_handler_args), void *argPtr, void *eventPtr);
int ca_add_array_event(int type, int count, void *chanPtr,
		       void (*funcPtr)(struct event_handler_args), void *argPtr,
		       float float1, float float2, float float3,
		       void *eventPtr);
int ca_add_masked_array_event(int type, int count,
			      chid chan,
			      void (*pFunc)(struct event_handler_args), void *pArg,
			      float f1, float f2, float f3,
			      evid *pEvid, int mask);
int ca_put(int type, void *chanPtr, void *valPtr);
int ca_flush_io (void);
int ca_build_and_connect(char *name, int type, int count,
			 chid *pChid, void *pValue,
			 void (*pFunc)(struct connection_handler_args), void *pUser);
int ca_search (char *name, void *chanPtr);
int ca_array_get (int type, int count, chid pChid, void *pValue);
int ca_array_get_callback(int type, int count, chid pChid,
			  void (*pFunc)(struct event_handler_args), void *parg);
int ca_array_put(int type, int count, chid pChid, void *pValue);
int ca_build_read_and_monitor ( char	*name,
				void	(*connFunc)(struct connection_handler_args),
				void	*pUser,
				void	(*readFunc)(struct event_handler_args),
				void	*readArg,
				void	(*monFunc)(struct event_handler_args),
				void	*monArg,
				chid	*pChid,
				evid	*pEvid);
int ca_complete_bld_read_mon (chid pChid);

int caMsgProc(float, int);
int caMsgHandler(short, char *, char *);
int getMore(void);
int sendCAQ (short , char *, int);
int recvCAQ (char *, int);
/* because the cacontext.h file includes cadef.h, prevent this define until CATX is defined */
#ifdef CATXDEFINED
int caValidChid(chid pChid, CATX *ctx);
#endif CATXDEFINED
int caConvType(char *, short, char *, short, int);
int caConvParamData(char *, short, char *);

int caConvAlarmData (char *, tisAlarm *);
int ca_ack_alarm (char *pvName, TS_STAMP *stamp);
int ca_recv_alarm (int serial, int stateLess, char *pvName, TS_STAMP *pStamp);
int ca_enable_alarm (char *pvName);
int ca_disable_alarm (char *pvName);

#ifdef vxWorks
void caClientInit(void);
int ca_import (int tid);
void ca_task_exit(int taskId);
#else
void ca_task_exit (void);
#endif vxWorks

#endif INCLcadef
