/* db_access.h */
/* share/epicsH @(#)db_access.h	1.6     7/17/91 */

/*
 *      Author:          Bob Dalesio
 *      Date:            4-4-88
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 *
 *	Modification Log:
 *	-----------------
 * .01	05-02-88	lrd	add control enumeration field, structure & size
 * .02	05-12-88	lrd	add array structures
 *				change structure names dbf_ to dbr_
 *				add VALID_FIELD_TYPE & INVALID_FIELD_TYPE macros
 *				add db_access_val union of all access strucures
 *				add MAX_STRING_SIZE 40 define
 * .03	08-08-88	lrd	add no_elements to db_addr
 *				removed OLD_DBF_ARRAY and all OLD_DBR_ARRAY variations
 *				made value the last field of each structure
 *				db_access_val union modified for changes
 *				dbr_size array modified for changes
 * .04	11-29-88	lrd	fixed declaration of union db_access_val
 * .05	12-17-88	lrd	added extra processing field
 * .06	01-13-89	lrd	added EP_CALC for modifying calculations on-line
 * .07	03-12-90	joh	added INCLdb_accessh for auto include 
 * .08	04-06-90	lrd	add OLD_DBF_CHAR, OLD_DBF_LONG, OLD_DBF_DOUBLE
 *				and OLD_DBR_TIME
 * .09	04-06-90	joh	added dbr_size_n(type, count)
 * .10	04-06-90	joh	changed all dbr_size stuff to unsigned
 * .11	04-27-90	rac	added date to OLD_DBR_TIME structures
 * .12	05-02-90	rac	changed back to old way of doing sts and gr;
 *				added dbr and dbf keywords and macros
 * .13	08-07-90	rac	re-enter various changes since .12; change
 *				some char structure members to short; change
 *				time structures once more.
 * .14	10-29-90	joh	forced size and value size to allocate in
 *				one module alone. 
 * .15	10-29-90	joh	added the readonly keyword for VAX C after 
 *				failing to find any way to get the CA VMS share 
 *				image to work without polluting the source.
 * .16  01-07-91        rac     fix a problem with dbf_text_to_type()
 * .17	06-03-91	joh	SPARC padding. 
 * .18	06-04-91	joh	added VALID_DB_FIELD(x) & INVALID_DB_FIELD(x)
 *				macros.
 * .19  03-01-91        mda     integration of ANL/LANL code
 * .20	06-25-91	joh	added #include db_addr.h
 * .21  07-06-93        ems     dbr_gr_enum, inc struct from 24 to 45 strs
 *                              dbr_ctrl_enum inc struct from 24 to 45 strs
 *
 * NOTES:
 * .01	06-04-91	joh	the value must allways be the last item 
 *				in the structure so arrays work.
 * .02	06-05-91	joh	structures are aligned to their largest memeber
 *				for the SPARC. CA takes care of alignment
 *				between structures as part of the CA protocol.
 */

#ifndef INCLdb_accessh
#define INCLdb_accessh

#ifndef INC_tsDefs_h
#   include <tsDefs.h>
#endif


/* 
 *database access address structure 
 */
#ifndef INC_db_addrh
#include	<db_addr.h>
#endif

/* database field types */

#define OLD_DBF_STRING	0
#define	OLD_DBF_INT		1
#define	OLD_DBF_SHORT	1
#define	OLD_DBF_FLOAT	2
#define	OLD_DBF_ENUM	3
#define	OLD_DBF_CHAR	4
#define	OLD_DBF_LONG	5
#define	OLD_DBF_DOUBLE	6
#define OLD_DBF_NO_ACCESS	7
#define	OLD_LAST_TYPE	OLD_DBF_DOUBLE
#define	OLD_VALID_DB_FIELD(x)	((x >= 0) && (x <= OLD_LAST_TYPE))
#define	OLD_INVALID_DB_FIELD(x)	((x < 0) || (x > OLD_LAST_TYPE))

/* data request buffer types */
#define OLD_DBR_STRING	OLD_DBF_STRING	
#define	OLD_DBR_INT		OLD_DBF_INT		
#define	OLD_DBR_SHORT	OLD_DBF_INT		
#define	OLD_DBR_FLOAT	OLD_DBF_FLOAT	
#define	OLD_DBR_ENUM	OLD_DBF_ENUM
#define	OLD_DBR_CHAR	OLD_DBF_CHAR
#define	OLD_DBR_LONG	OLD_DBF_LONG
#define	OLD_DBR_DOUBLE	OLD_DBF_DOUBLE
#define OLD_DBR_STS_STRING	7
#define	OLD_DBR_STS_INT	8
#define	OLD_DBR_STS_SHORT	8
#define	OLD_DBR_STS_FLOAT	9
#define	OLD_DBR_STS_ENUM	10
#define	OLD_DBR_STS_CHAR	11
#define	OLD_DBR_STS_LONG	12
#define	OLD_DBR_STS_DOUBLE	13
#define	OLD_DBR_TIME_STRING	14
#define	OLD_DBR_TIME_INT	15
#define	OLD_DBR_TIME_SHORT	15
#define	OLD_DBR_TIME_FLOAT	16
#define	OLD_DBR_TIME_ENUM	17
#define	OLD_DBR_TIME_CHAR	18
#define	OLD_DBR_TIME_LONG	19
#define	OLD_DBR_TIME_DOUBLE	20
#define	OLD_DBR_GR_STRING	21
#define	OLD_DBR_GR_INT	22
#define	OLD_DBR_GR_SHORT	22
#define	OLD_DBR_GR_FLOAT	23
#define	OLD_DBR_GR_ENUM	24
#define	OLD_DBR_GR_CHAR	25
#define	OLD_DBR_GR_LONG	26
#define	OLD_DBR_GR_DOUBLE	27
#define	OLD_DBR_CTRL_STRING	28
#define OLD_DBR_CTRL_INT	29
#define OLD_DBR_CTRL_SHORT	29
#define	OLD_DBR_CTRL_FLOAT	30
#define OLD_DBR_CTRL_ENUM	31
#define	OLD_DBR_CTRL_CHAR	32
#define	OLD_DBR_CTRL_LONG	33
#define	OLD_DBR_CTRL_DOUBLE	34
#define	OLD_LAST_BUFFER_TYPE	OLD_DBR_CTRL_DOUBLE

/* WARNING WARNING WARNING WARNING!
	If you change LAST_BUFFER_TYPE, you must also make a
	change to paramStructs.h. See that file for details.
   WARNING WARNING WARNING WARNING! */

#define	OLD_VALID_DB_REQ(x)	((x >= 0) && (x <= OLD_LAST_BUFFER_TYPE))
#define	OLD_INVALID_DB_REQ(x)	((x < 0) || (x > OLD_LAST_BUFFER_TYPE))

/* extra processing flags */
#define	EP_SCAN	0x01
#define	EP_INIT	0x02
#define	EP_CALC	0x03

/* values returned for each field type
 * 	OLD_DBR_STRING	returns a NULL terminated string
 *	OLD_DBR_SHORT	returns an unsigned short
 *	OLD_DBR_INT		returns an unsigned short
 *	OLD_DBR_FLOAT	returns an IEEE floating point value
 *	OLD_DBR_ENUM	returns an unsigned short which is the enum item
 *	OLD_DBR_CHAR	returns an unsigned char
 *	OLD_DBR_LONG	returns an unsigned long
 *	OLD_DBR_DOUBLE	returns a double precision floating point number
 *	OLD_DBR_STS_STRING	returns a string status structure (dbr_sts_string)
 *	OLD_DBR_STS_SHORT	returns a short status structure (dbr_sts_short)
 *	OLD_DBR_STS_INT	returns a short status structure (dbr_sts_int)
 *	OLD_DBR_STS_FLOAT	returns a float status structure (dbr_sts_float)
 *	OLD_DBR_STS_ENUM	returns an enum status structure (dbr_sts_enum)
 *	OLD_DBR_STS_CHAR	returns a char status structure (dbr_sts_char)
 *	OLD_DBR_STS_LONG	returns a long status structure (dbr_sts_long)
 *	OLD_DBR_STS_DOUBLE	returns a double status structure (dbr_sts_double)
 *	OLD_DBR_TIME_STRING	returns a string time structure (dbr_time_string)
 *	OLD_DBR_TIME_SHORT	returns a short time structure (dbr_time_short)
 *	OLD_DBR_TIME_INT	returns a short time structure (dbr_time_short)
 *	OLD_DBR_TIME_FLOAT	returns a float time structure (dbr_time_float)
 *	OLD_DBR_TIME_ENUM	returns an enum time structure (dbr_time_enum)
 *	OLD_DBR_TIME_CHAR	returns a char time structure (dbr_time_char)
 *	OLD_DBR_TIME_LONG	returns a long time structure (dbr_time_long)
 *	OLD_DBR_TIME_DOUBLE	returns a double time structure (dbr_time_double)
 *	OLD_DBR_GR_STRING	returns a graphic string structure (dbr_gr_string)
 *	OLD_DBR_GR_SHORT	returns a graphic short structure (dbr_gr_short)
 *	OLD_DBR_GR_INT	returns a graphic short structure (dbr_gr_int)
 *	OLD_DBR_GR_FLOAT	returns a graphic float structure (dbr_gr_float)
 *	OLD_DBR_GR_ENUM	returns a graphic enum structure (dbr_gr_enum)
 *	OLD_DBR_GR_CHAR	returns a graphic char structure (dbr_gr_char)
 *	OLD_DBR_GR_LONG	returns a graphic long structure (dbr_gr_long)
 *	OLD_DBR_GR_DOUBLE	returns a graphic double structure (dbr_gr_double)
 *	OLD_DBR_CTRL_STRING	returns a control string structure (dbr_ctrl_int)
 *	OLD_DBR_CTRL_SHORT	returns a control short structure (dbr_ctrl_short)
 *	OLD_DBR_CTRL_INT	returns a control short structure (dbr_ctrl_int)
 *	OLD_DBR_CTRL_FLOAT	returns a control float structure (dbr_ctrl_float)
 *	OLD_DBR_CTRL_ENUM	returns a control enum structure (dbr_ctrl_enum)
 *	OLD_DBR_CTRL_CHAR	returns a control char structure (dbr_ctrl_char)
 *	OLD_DBR_CTRL_LONG	returns a control long structure (dbr_ctrl_long)
 *	OLD_DBR_CTRL_DOUBLE	returns a control double structure (dbr_ctrl_double)
 */
#define MAX_STRING_SIZE	80

/* VALUES WITH STATUS STRUCTURES */

/* structure for a  string status field */
struct dbr_sts_string{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	char	value[MAX_STRING_SIZE];	/* current value */
	char	aenb[4];		/* "YES" or "NO" */
	char	ack[4];			/* "YES" or "NO" */
	char	desc[28];
	char	egu[8];
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	char	value[MAX_STRING_SIZE] __attribute__ ((packed)) ;	/* current value */
	char	aenb[4] __attribute__ ((packed)) ;		/* "YES" or "NO" */
	char	ack[4] __attribute__ ((packed)) ;			/* "YES" or "NO" */
	char	desc[28] __attribute__ ((packed)) ;
	char	egu[8] __attribute__ ((packed)) ;
#endif /* AXP */
};

/* structure for an short status field */
struct dbr_sts_int{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	short	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	short	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};
struct dbr_sts_short{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	short	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	short	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a  float status field */
struct dbr_sts_float{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	float	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	float	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a  enum status field */
struct dbr_sts_enum{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	short	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	short	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a char status field */
struct dbr_sts_char{
#ifndef AXP
	short		status;	 	/* status of value */
	short		severity;	/* severity of alarm */
	char		RISC_pad;	/* RISC alignment */
	unsigned char	value;		/* current value */
#else /* AXP */
	short		status __attribute__ ((packed)) ;	 	/* status of value */
	short		severity __attribute__ ((packed)) ;	/* severity of alarm */
	char		RISC_pad __attribute__ ((packed)) ;	/* RISC alignment */
	unsigned char	value __attribute__ ((packed)) ;		/* current value */
#endif /* AXP */
};

/* structure for a long status field */
struct dbr_sts_long{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	long	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	long	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a double status field */
struct dbr_sts_double{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	long	RISC_pad;		/* RISC alignment */
	double	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	long	RISC_pad __attribute__ ((packed)) ;		/* RISC alignment */
	double	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* VALUES WITH STATUS AND TIME STRUCTURES */

/* structure for a  string time field */
struct dbr_time_string{
#ifndef AXP
	short		status;	 		/* status of value */
	short		severity;		/* severity of alarm */
	TS_STAMP	stamp;			/* time stamp */
	char		value[MAX_STRING_SIZE];	/* current value */
#else /* AXP */
	short		status __attribute__ ((packed)) ;	 		/* status of value */
	short		severity __attribute__ ((packed)) ;		/* severity of alarm */
	TS_STAMP	stamp __attribute__ ((packed)) ;			/* time stamp */
	char		value[MAX_STRING_SIZE] __attribute__ ((packed)) ;	/* current value */
#endif /* AXP */
};

/* structure for an short time field */
struct dbr_time_short{
#ifndef AXP
	short		status;	 		/* status of value */
	short		severity;		/* severity of alarm */
	TS_STAMP	stamp;			/* time stamp */
	short		RISC_pad;		/* RISC alignment */
	short		value;			/* current value */
#else /* AXP */
	short		status __attribute__ ((packed)) ;	 		/* status of value */
	short		severity __attribute__ ((packed)) ;		/* severity of alarm */
	TS_STAMP	stamp __attribute__ ((packed)) ;			/* time stamp */
	short		RISC_pad __attribute__ ((packed)) ;		/* RISC alignment */
	short		value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a  float time field */
struct dbr_time_float{
#ifndef AXP
	short		status;	 		/* status of value */
	short		severity;		/* severity of alarm */
	TS_STAMP	stamp;			/* time stamp */
	float		value;			/* current value */
#else /* AXP */
	short		status __attribute__ ((packed)) ;	 		/* status of value */
	short		severity __attribute__ ((packed)) ;		/* severity of alarm */
	TS_STAMP	stamp __attribute__ ((packed)) ;			/* time stamp */
	float		value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a  enum time field */
struct dbr_time_enum{
#ifndef AXP
	short		status;	 		/* status of value */
	short		severity;		/* severity of alarm */
	TS_STAMP	stamp;			/* time stamp */
	short		RISC_pad;		/* RISC alignment */
	short		value;			/* current value */
#else /* AXP */
	short		status __attribute__ ((packed)) ;	 		/* status of value */
	short		severity __attribute__ ((packed)) ;		/* severity of alarm */
	TS_STAMP	stamp __attribute__ ((packed)) ;			/* time stamp */
	short		RISC_pad __attribute__ ((packed)) ;		/* RISC alignment */
	short		value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a char time field */
struct dbr_time_char{
#ifndef AXP
	short			status;	 		/* status of value */
	short			severity;		/* severity of alarm */
	TS_STAMP		stamp;			/* time stamp */
	short			RISC_pad0;		/* RISC alignment */
	char			RISC_pad1;		/* RISC alignment */
	unsigned char		value;			/* current value */
#else /* AXP */
	short			status __attribute__ ((packed)) ;	 		/* status of value */
	short			severity __attribute__ ((packed)) ;		/* severity of alarm */
	TS_STAMP		stamp __attribute__ ((packed)) ;			/* time stamp */
	short			RISC_pad0 __attribute__ ((packed)) ;		/* RISC alignment */
	char			RISC_pad1 __attribute__ ((packed)) ;		/* RISC alignment */
	unsigned char		value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a long time field */
struct dbr_time_long{
#ifndef AXP
	short		status;	 		/* status of value */
	short		severity;		/* severity of alarm */
	TS_STAMP	stamp;			/* time stamp */
	long		value;			/* current value */
#else /* AXP */
	short		status __attribute__ ((packed)) ;	 		/* status of value */
	short		severity __attribute__ ((packed)) ;		/* severity of alarm */
	TS_STAMP	stamp __attribute__ ((packed)) ;			/* time stamp */
	long		value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a double time field */
struct dbr_time_double{
#ifndef AXP
	short		status;	 		/* status of value */
	short		severity;		/* severity of alarm */
	TS_STAMP	stamp;			/* time stamp */
	long		RISC_pad;		/* RISC alignment */
	double		value;			/* current value */
#else /* AXP */
	short		status __attribute__ ((packed)) ;	 		/* status of value */
	short		severity __attribute__ ((packed)) ;		/* severity of alarm */
	TS_STAMP	stamp __attribute__ ((packed)) ;			/* time stamp */
	long		RISC_pad __attribute__ ((packed)) ;		/* RISC alignment */
	double		value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* VALUES WITH STATUS AND GRAPHIC STRUCTURES */

/* structure for a graphic string */
	/* not implemented; use struct_dbr_sts_string */

/* structure for a graphic short field */
struct dbr_gr_int{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	char	units[8];		/* units of value */
	short	upper_disp_limit;	/* upper limit of graph */
	short	lower_disp_limit;	/* lower limit of graph */
	short	upper_alarm_limit;	
	short	upper_warning_limit;
	short	lower_warning_limit;
	short	lower_alarm_limit;
	short	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	char	units[8] __attribute__ ((packed)) ;		/* units of value */
	short	upper_disp_limit __attribute__ ((packed)) ;	/* upper limit of graph */
	short	lower_disp_limit __attribute__ ((packed)) ;	/* lower limit of graph */
	short	upper_alarm_limit __attribute__ ((packed)) ;	
	short	upper_warning_limit __attribute__ ((packed)) ;
	short	lower_warning_limit __attribute__ ((packed)) ;
	short	lower_alarm_limit __attribute__ ((packed)) ;
	short	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};
struct dbr_gr_short{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	char	units[8];		/* units of value */
	short	upper_disp_limit;	/* upper limit of graph */
	short	lower_disp_limit;	/* lower limit of graph */
	short	upper_alarm_limit;	
	short	upper_warning_limit;
	short	lower_warning_limit;
	short	lower_alarm_limit;
	short	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	char	units[8] __attribute__ ((packed)) ;		/* units of value */
	short	upper_disp_limit __attribute__ ((packed)) ;	/* upper limit of graph */
	short	lower_disp_limit __attribute__ ((packed)) ;	/* lower limit of graph */
	short	upper_alarm_limit __attribute__ ((packed)) ;	
	short	upper_warning_limit __attribute__ ((packed)) ;
	short	lower_warning_limit __attribute__ ((packed)) ;
	short	lower_alarm_limit __attribute__ ((packed)) ;
	short	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a graphic floating point field */
struct dbr_gr_float{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	short	precision;		/* number of decimal places */
	short	RISC_pad0;		/* RISC alignment */
	char	units[8];		/* units of value */
	float	upper_disp_limit;	/* upper limit of graph */
	float	lower_disp_limit;	/* lower limit of graph */
	float	upper_alarm_limit;	
	float	upper_warning_limit;
	float	lower_warning_limit;
	float	lower_alarm_limit;
	float	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	short	precision __attribute__ ((packed)) ;		/* number of decimal places */
	short	RISC_pad0 __attribute__ ((packed)) ;		/* RISC alignment */
	char	units[8] __attribute__ ((packed)) ;		/* units of value */
	float	upper_disp_limit __attribute__ ((packed)) ;	/* upper limit of graph */
	float	lower_disp_limit __attribute__ ((packed)) ;	/* lower limit of graph */
	float	upper_alarm_limit __attribute__ ((packed)) ;	
	float	upper_warning_limit __attribute__ ((packed)) ;
	float	lower_warning_limit __attribute__ ((packed)) ;
	float	lower_alarm_limit __attribute__ ((packed)) ;
	float	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a graphic enumeration field */
struct dbr_gr_enum{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	short	no_str;			/* number of strings */
	char	strs[64][26];		/* state strings */
	short	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	short	no_str __attribute__ ((packed)) ;			/* number of strings */
	char	strs[64][26] __attribute__ ((packed)) ;		/* state strings */
	short	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a graphic char field */
struct dbr_gr_char{
#ifndef AXP
	short			status;	 		/* status of value */
	short			severity;		/* severity of alarm */
	char			units[8];		/* units of value */
	unsigned char		upper_disp_limit;	/* upper limit of graph */
	unsigned char		lower_disp_limit;	/* lower limit of graph */
	unsigned char		upper_alarm_limit;	
	unsigned char		upper_warning_limit;
	unsigned char		lower_warning_limit;
	unsigned char		lower_alarm_limit;
	char			RISC_pad;		/* RISC alignment */
	unsigned char		value;			/* current value */
#else /* AXP */
	short			status __attribute__ ((packed)) ;	 		/* status of value */
	short			severity __attribute__ ((packed)) ;		/* severity of alarm */
	char			units[8] __attribute__ ((packed)) ;		/* units of value */
	unsigned char		upper_disp_limit __attribute__ ((packed)) ;	/* upper limit of graph */
	unsigned char		lower_disp_limit __attribute__ ((packed)) ;	/* lower limit of graph */
	unsigned char		upper_alarm_limit __attribute__ ((packed)) ;	
	unsigned char		upper_warning_limit __attribute__ ((packed)) ;
	unsigned char		lower_warning_limit __attribute__ ((packed)) ;
	unsigned char		lower_alarm_limit __attribute__ ((packed)) ;
	char			RISC_pad __attribute__ ((packed)) ;		/* RISC alignment */
	unsigned char		value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a graphic long field */
struct dbr_gr_long{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	char	units[8];		/* units of value */
	long	upper_disp_limit;	/* upper limit of graph */
	long	lower_disp_limit;	/* lower limit of graph */
	long	upper_alarm_limit;	
	long	upper_warning_limit;
	long	lower_warning_limit;
	long	lower_alarm_limit;
	long	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	char	units[8] __attribute__ ((packed)) ;		/* units of value */
	long	upper_disp_limit __attribute__ ((packed)) ;	/* upper limit of graph */
	long	lower_disp_limit __attribute__ ((packed)) ;	/* lower limit of graph */
	long	upper_alarm_limit __attribute__ ((packed)) ;	
	long	upper_warning_limit __attribute__ ((packed)) ;
	long	lower_warning_limit __attribute__ ((packed)) ;
	long	lower_alarm_limit __attribute__ ((packed)) ;
	long	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a graphic double field */
struct dbr_gr_double{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	short	precision;		/* number of decimal places */
	short	RISC_pad0;		/* RISC alignment */
	char	units[8];		/* units of value */
	double	upper_disp_limit;	/* upper limit of graph */
	double	lower_disp_limit;	/* lower limit of graph */
	double	upper_alarm_limit;	
	double	upper_warning_limit;
	double	lower_warning_limit;
	double	lower_alarm_limit;
	double	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	short	precision __attribute__ ((packed)) ;		/* number of decimal places */
	short	RISC_pad0 __attribute__ ((packed)) ;		/* RISC alignment */
	char	units[8] __attribute__ ((packed)) ;		/* units of value */
	double	upper_disp_limit __attribute__ ((packed)) ;	/* upper limit of graph */
	double	lower_disp_limit __attribute__ ((packed)) ;	/* lower limit of graph */
	double	upper_alarm_limit __attribute__ ((packed)) ;	
	double	upper_warning_limit __attribute__ ((packed)) ;
	double	lower_warning_limit __attribute__ ((packed)) ;
	double	lower_alarm_limit __attribute__ ((packed)) ;
	double	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* VALUES WITH STATUS, GRAPHIC and CONTROL STRUCTURES */

/* structure for a control string */
	/* not implemented; use struct_dbr_sts_string */

/* structure for a control integer */
struct dbr_ctrl_int{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	char	units[8];		/* units of value */
	short	upper_disp_limit;	/* upper limit of graph */
	short	lower_disp_limit;	/* lower limit of graph */
	short	upper_alarm_limit;	
	short	upper_warning_limit;
	short	lower_warning_limit;
	short	lower_alarm_limit;
	short	upper_ctrl_limit;	/* upper control limit */
	short	lower_ctrl_limit;	/* lower control limit */
	short	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	char	units[8] __attribute__ ((packed)) ;		/* units of value */
	short	upper_disp_limit __attribute__ ((packed)) ;	/* upper limit of graph */
	short	lower_disp_limit __attribute__ ((packed)) ;	/* lower limit of graph */
	short	upper_alarm_limit __attribute__ ((packed)) ;	
	short	upper_warning_limit __attribute__ ((packed)) ;
	short	lower_warning_limit __attribute__ ((packed)) ;
	short	lower_alarm_limit __attribute__ ((packed)) ;
	short	upper_ctrl_limit __attribute__ ((packed)) ;	/* upper control limit */
	short	lower_ctrl_limit __attribute__ ((packed)) ;	/* lower control limit */
	short	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};
struct dbr_ctrl_short{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	char	units[8];		/* units of value */
	short	upper_disp_limit;	/* upper limit of graph */
	short	lower_disp_limit;	/* lower limit of graph */
	short	upper_alarm_limit;	
	short	upper_warning_limit;
	short	lower_warning_limit;
	short	lower_alarm_limit;
	short	upper_ctrl_limit;	/* upper control limit */
	short	lower_ctrl_limit;	/* lower control limit */
	short	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	char	units[8] __attribute__ ((packed)) ;		/* units of value */
	short	upper_disp_limit __attribute__ ((packed)) ;	/* upper limit of graph */
	short	lower_disp_limit __attribute__ ((packed)) ;	/* lower limit of graph */
	short	upper_alarm_limit __attribute__ ((packed)) ;	
	short	upper_warning_limit __attribute__ ((packed)) ;
	short	lower_warning_limit __attribute__ ((packed)) ;
	short	lower_alarm_limit __attribute__ ((packed)) ;
	short	upper_ctrl_limit __attribute__ ((packed)) ;	/* upper control limit */
	short	lower_ctrl_limit __attribute__ ((packed)) ;	/* lower control limit */
	short	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a control floating point field */
struct dbr_ctrl_float{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	short	precision;		/* number of decimal places */
	short	RISC_pad;		/* RISC alignment */
	char	units[8];		/* units of value */
	float	upper_disp_limit;	/* upper limit of graph */
	float	lower_disp_limit;	/* lower limit of graph */
	float	upper_alarm_limit;	
	float	upper_warning_limit;
	float	lower_warning_limit;
	float	lower_alarm_limit;
 	float	upper_ctrl_limit;	/* upper control limit */
	float	lower_ctrl_limit;	/* lower control limit */
	float	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	short	precision __attribute__ ((packed)) ;		/* number of decimal places */
	short	RISC_pad __attribute__ ((packed)) ;		/* RISC alignment */
	char	units[8] __attribute__ ((packed)) ;		/* units of value */
	float	upper_disp_limit __attribute__ ((packed)) ;	/* upper limit of graph */
	float	lower_disp_limit __attribute__ ((packed)) ;	/* lower limit of graph */
	float	upper_alarm_limit __attribute__ ((packed)) ;	
	float	upper_warning_limit __attribute__ ((packed)) ;
	float	lower_warning_limit __attribute__ ((packed)) ;
	float	lower_alarm_limit __attribute__ ((packed)) ;
 	float	upper_ctrl_limit __attribute__ ((packed)) ;	/* upper control limit */
	float	lower_ctrl_limit __attribute__ ((packed)) ;	/* lower control limit */
	float	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a control enumeration field */
struct dbr_ctrl_enum{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	short	no_str;			/* number of strings */
	char	strs[64][26];		/* state strings */
	short	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	short	no_str __attribute__ ((packed)) ;			/* number of strings */
	char	strs[64][26] __attribute__ ((packed)) ;		/* state strings */
	short	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a control char field */
struct dbr_ctrl_char{
#ifndef AXP
	short		status;	 		/* status of value */
	short		severity;		/* severity of alarm */
	char		units[8];		/* units of value */
	unsigned char	upper_disp_limit;	/* upper limit of graph */
	unsigned char	lower_disp_limit;	/* lower limit of graph */
	unsigned char	upper_alarm_limit;	
	unsigned char	upper_warning_limit;
	unsigned char	lower_warning_limit;
	unsigned char	lower_alarm_limit;
	unsigned char	upper_ctrl_limit;	/* upper control limit */
	unsigned char	lower_ctrl_limit;	/* lower control limit */
	char		RISC_pad;		/* RISC alignment */
	unsigned char	value;			/* current value */
#else /* AXP */
	short		status __attribute__ ((packed)) ;	 		/* status of value */
	short		severity __attribute__ ((packed)) ;		/* severity of alarm */
	char		units[8] __attribute__ ((packed)) ;		/* units of value */
	unsigned char	upper_disp_limit __attribute__ ((packed)) ;	/* upper limit of graph */
	unsigned char	lower_disp_limit __attribute__ ((packed)) ;	/* lower limit of graph */
	unsigned char	upper_alarm_limit __attribute__ ((packed)) ;	
	unsigned char	upper_warning_limit __attribute__ ((packed)) ;
	unsigned char	lower_warning_limit __attribute__ ((packed)) ;
	unsigned char	lower_alarm_limit __attribute__ ((packed)) ;
	unsigned char	upper_ctrl_limit __attribute__ ((packed)) ;	/* upper control limit */
	unsigned char	lower_ctrl_limit __attribute__ ((packed)) ;	/* lower control limit */
	char		RISC_pad __attribute__ ((packed)) ;		/* RISC alignment */
	unsigned char	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a control long field */
struct dbr_ctrl_long{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	char	units[8];		/* units of value */
	long	upper_disp_limit;	/* upper limit of graph */
	long	lower_disp_limit;	/* lower limit of graph */
	long	upper_alarm_limit;	
	long	upper_warning_limit;
	long	lower_warning_limit;
	long	lower_alarm_limit;
	long	upper_ctrl_limit;	/* upper control limit */
	long	lower_ctrl_limit;	/* lower control limit */
	long	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	char	units[8] __attribute__ ((packed)) ;		/* units of value */
	long	upper_disp_limit __attribute__ ((packed)) ;	/* upper limit of graph */
	long	lower_disp_limit __attribute__ ((packed)) ;	/* lower limit of graph */
	long	upper_alarm_limit __attribute__ ((packed)) ;	
	long	upper_warning_limit __attribute__ ((packed)) ;
	long	lower_warning_limit __attribute__ ((packed)) ;
	long	lower_alarm_limit __attribute__ ((packed)) ;
	long	upper_ctrl_limit __attribute__ ((packed)) ;	/* upper control limit */
	long	lower_ctrl_limit __attribute__ ((packed)) ;	/* lower control limit */
	long	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

/* structure for a control double field */
struct dbr_ctrl_double{
#ifndef AXP
	short	status;	 		/* status of value */
	short	severity;		/* severity of alarm */
	short	precision;		/* number of decimal places */
	short	RISC_pad0;		/* RISC alignment */
	char	units[8];		/* units of value */
	double	upper_disp_limit;	/* upper limit of graph */
	double	lower_disp_limit;	/* lower limit of graph */
	double	upper_alarm_limit;	
	double	upper_warning_limit;
	double	lower_warning_limit;
	double	lower_alarm_limit;
	double	upper_ctrl_limit;	/* upper control limit */
	double	lower_ctrl_limit;	/* lower control limit */
	double	value;			/* current value */
#else /* AXP */
	short	status __attribute__ ((packed)) ;	 		/* status of value */
	short	severity __attribute__ ((packed)) ;		/* severity of alarm */
	short	precision __attribute__ ((packed)) ;		/* number of decimal places */
	short	RISC_pad0 __attribute__ ((packed)) ;		/* RISC alignment */
	char	units[8] __attribute__ ((packed)) ;		/* units of value */
	double	upper_disp_limit __attribute__ ((packed)) ;	/* upper limit of graph */
	double	lower_disp_limit __attribute__ ((packed)) ;	/* lower limit of graph */
	double	upper_alarm_limit __attribute__ ((packed)) ;	
	double	upper_warning_limit __attribute__ ((packed)) ;
	double	lower_warning_limit __attribute__ ((packed)) ;
	double	lower_alarm_limit __attribute__ ((packed)) ;
	double	upper_ctrl_limit __attribute__ ((packed)) ;	/* upper control limit */
	double	lower_ctrl_limit __attribute__ ((packed)) ;	/* lower control limit */
	double	value __attribute__ ((packed)) ;			/* current value */
#endif /* AXP */
};

#define dbr_size_n(TYPE,COUNT)\
((unsigned)(COUNT==1?dbr_size[TYPE]:dbr_size[TYPE]+((COUNT)-1)*dbr_value_size[TYPE]))
/* sizes for each field */

#ifdef VAXC
readonly
#endif
#ifndef DB_TEXT_GLBLSOURCE
extern unsigned short dbr_size[];
#else
unsigned short dbr_size[] = {
	MAX_STRING_SIZE,		/* string max size		*/
	sizeof(short),			/* short			*/
	sizeof(float),			/* IEEE Float			*/
	sizeof(short),			/* item number			*/
	sizeof(char),			/* character			*/
	sizeof(long),			/* long				*/
	sizeof(double),			/* double			*/
	sizeof(struct dbr_sts_string),	/* string field	with status	*/
	sizeof(struct dbr_sts_short),	/* short field with status	*/
	sizeof(struct dbr_sts_float),	/* float field with status	*/
	sizeof(struct dbr_sts_enum),	/* item number with status	*/
	sizeof(struct dbr_sts_char),	/* char field with status	*/
	sizeof(struct dbr_sts_long),	/* long field with status	*/
	sizeof(struct dbr_sts_double),	/* double field with time	*/
	sizeof(struct dbr_time_string),	/* string field	with time	*/
	sizeof(struct dbr_time_short),	/* short field with time	*/
	sizeof(struct dbr_time_float),	/* float field with time	*/
	sizeof(struct dbr_time_enum),	/* item number with time	*/
	sizeof(struct dbr_time_char),	/* char field with time		*/
	sizeof(struct dbr_time_long),	/* long field with time		*/
	sizeof(struct dbr_time_double),	/* double field with time	*/
	sizeof(struct dbr_sts_string),	/* graphic string info		*/
	sizeof(struct dbr_gr_short),	/* graphic short info		*/
	sizeof(struct dbr_gr_float),	/* graphic float info		*/
	sizeof(struct dbr_gr_enum),	/* graphic item info		*/
	sizeof(struct dbr_gr_char),	/* graphic char info		*/
	sizeof(struct dbr_gr_long),	/* graphic long info		*/
	sizeof(struct dbr_gr_double),	/* graphic double info		*/
	sizeof(struct dbr_sts_string),	/* control string info		*/
	sizeof(struct dbr_ctrl_short),	/* control short info		*/
	sizeof(struct dbr_ctrl_float),	/* control float info		*/
	sizeof(struct dbr_ctrl_enum),	/* control item info		*/
	sizeof(struct dbr_ctrl_char),	/* control char info		*/
	sizeof(struct dbr_ctrl_long),	/* control long info		*/
	sizeof(struct dbr_ctrl_double),	/* control double info		*/
};
#endif

/* sizes for each field's value */
#ifdef VAXC
readonly
#endif
#ifndef DB_TEXT_GLBLSOURCE
extern unsigned short dbr_value_size[];
#else
unsigned short dbr_value_size[] = {
	MAX_STRING_SIZE,	/* string max size		*/
	sizeof(short),		/* short			*/
	sizeof(float),		/* IEEE Float			*/
	sizeof(short),		/* item number			*/
	sizeof(char),		/* character			*/
	sizeof(long),		/* long				*/
	sizeof(double),		/* double			*/
	MAX_STRING_SIZE,	/* string max size		*/
	sizeof(short),		/* short			*/
	sizeof(float),		/* IEEE Float			*/
	sizeof(short),		/* item number			*/
	sizeof(char),		/* character			*/
	sizeof(long),		/* long				*/
	sizeof(double),		/* double			*/
	MAX_STRING_SIZE,	/* string max size		*/
	sizeof(short),		/* short			*/
	sizeof(float),		/* IEEE Float			*/
	sizeof(short),		/* item number			*/
	sizeof(char),		/* character			*/
	sizeof(long),		/* long				*/
	sizeof(double),		/* double			*/
	MAX_STRING_SIZE,	/* string max size		*/
	sizeof(short),		/* short			*/
	sizeof(float),		/* IEEE Float			*/
	sizeof(short),		/* item number			*/
	sizeof(char),		/* character			*/
	sizeof(long),		/* long				*/
	sizeof(double),		/* double			*/
	MAX_STRING_SIZE,	/* string max size		*/
	sizeof(short),		/* short			*/
	sizeof(float),		/* IEEE Float			*/
	sizeof(short),		/* item number			*/
	sizeof(char),		/* character			*/
	sizeof(long),		/* long				*/
	sizeof(double),		/* double			*/
};
#endif

/* 
 * ptr to value given a pointer to the structure and the DBR type
 */
#define dbr_value_ptr(PDBR, OLD_DBR_TYPE) \
((void *)(((char *)PDBR)+dbr_value_offset[OLD_DBR_TYPE]))

/* 
 * byte offset of member in structure
 */
#define BYTE_OS(STRUCTURE, MEMBER) \
((int) & (((STRUCTURE *)0) -> MEMBER))

/* 
 * ptr to value given a pointer to the structure and the structure declaration
 */
#define dbr_value_ptr_from_structure(PDBR, STRUCTURE)\
((void *)(((char *)PDBR)+BYTE_OS(STRUCTURE, value)))

#ifdef VAXC
readonly
#endif
#ifndef DB_TEXT_GLBLSOURCE
extern unsigned short dbr_value_offset[];
#else
unsigned short dbr_value_offset[] = {
	0,					/* string			*/
	0,					/* short			*/
	0,					/* IEEE Float			*/
	0,					/* item number			*/
	0,					/* character			*/
	0,					/* long				*/
	0,					/* IEEE double			*/
	BYTE_OS(struct dbr_sts_string,value[0]),/* string field	with status	*/
	BYTE_OS(struct dbr_sts_short,value),	/* short field with status	*/
	BYTE_OS(struct dbr_sts_float,value),	/* float field with status	*/
	BYTE_OS(struct dbr_sts_enum,value),	/* item number with status	*/
	BYTE_OS(struct dbr_sts_char,value),	/* char field with status	*/
	BYTE_OS(struct dbr_sts_long,value),	/* long field with status	*/
	BYTE_OS(struct dbr_sts_double,value),	/* double field with time	*/
	BYTE_OS(struct dbr_time_string,value[0]),/* string field with time	*/
	BYTE_OS(struct dbr_time_short,value),	/* short field with time	*/
	BYTE_OS(struct dbr_time_float,value),	/* float field with time	*/
	BYTE_OS(struct dbr_time_enum,value),	/* item number with time	*/
	BYTE_OS(struct dbr_time_char,value),	/* char field with time		*/
	BYTE_OS(struct dbr_time_long,value),	/* long field with time		*/
	BYTE_OS(struct dbr_time_double,value),	/* double field with time	*/
	BYTE_OS(struct dbr_sts_string,value[0]),/* graphic string info		*/
	BYTE_OS(struct dbr_gr_short,value),	/* graphic short info		*/
	BYTE_OS(struct dbr_gr_float,value),	/* graphic float info		*/
	BYTE_OS(struct dbr_gr_enum,value),	/* graphic item info		*/
	BYTE_OS(struct dbr_gr_char,value),	/* graphic char info		*/
	BYTE_OS(struct dbr_gr_long,value),	/* graphic long info		*/
	BYTE_OS(struct dbr_gr_double,value),	/* graphic double info		*/
	BYTE_OS(struct dbr_sts_string,value[0]),/* control string info		*/
	BYTE_OS(struct dbr_ctrl_short,value),	/* control short info		*/
	BYTE_OS(struct dbr_ctrl_float,value),	/* control float info		*/
	BYTE_OS(struct dbr_ctrl_enum,value),	/* control item info		*/
	BYTE_OS(struct dbr_ctrl_char,value),	/* control char info		*/
	BYTE_OS(struct dbr_ctrl_long,value),	/* control long info		*/
	BYTE_OS(struct dbr_ctrl_double,value),	/* control double info		*/
};
#endif

/* union for each fetch buffers */
union db_access_val{
	char	strval[MAX_STRING_SIZE];	/* string max size	      */
	short			shrtval;	/* short		      */
	short			intval;		/* short		      */
	float			fltval;		/* IEEE Float		      */
	short			enmval;		/* item number		      */
	unsigned char		charval;	/* character		      */
	long			longval;	/* long			      */
	double			doubleval;	/* double		      */
	struct dbr_sts_string	sstrval;	/* string field	with status   */
	struct dbr_sts_short	sshrtval;	/* short field with status    */
	struct dbr_sts_float	sfltval;	/* float field with status    */
	struct dbr_sts_enum	senmval;	/* item number with status    */
	struct dbr_sts_char	schrval;	/* char field with status     */
	struct dbr_sts_long	slngval;	/* long field with status     */
	struct dbr_sts_double	sdblval;	/* double field with time     */
	struct dbr_time_string	tstrval;	/* string field	with time     */
	struct dbr_time_short	tshrtval;	/* short field with time      */
	struct dbr_time_float	tfltval;	/* float field with time      */
	struct dbr_time_enum	tenmval;	/* item number with time      */
	struct dbr_time_char	tchrval;	/* char field with time	      */
	struct dbr_time_long	tlngval;	/* long field with time	      */
	struct dbr_time_double	tdblval;	/* double field with time     */
	struct dbr_sts_string	gstrval;	/* graphic string info	      */
	struct dbr_gr_short	gshrtval;	/* graphic short info	      */
	struct dbr_gr_float	gfltval;	/* graphic float info	      */
	struct dbr_gr_enum	genmval;	/* graphic item info	      */
	struct dbr_gr_char	gchrval;	/* graphic char info	      */
	struct dbr_gr_long	glngval;	/* graphic long info	      */
	struct dbr_gr_double	gdblval;	/* graphic double info	      */
	struct dbr_sts_string	cstrval;	/* control string info	      */
	struct dbr_ctrl_short	cshrtval;	/* control short info	      */
	struct dbr_ctrl_float	cfltval;	/* control float info	      */
	struct dbr_ctrl_enum	cenmval;	/* control item info	      */
	struct dbr_ctrl_char	cchrval;	/* control char info	      */
	struct dbr_ctrl_long	clngval;	/* control long info	      */
	struct dbr_ctrl_double	cdblval;	/* control double info	      */
};


/*----------------------------------------------------------------------------
* repository for some useful PV database constants and utilities
*
* item dimensions
*    db_strval_dim		dimension for string values
*    db_units_dim		dimension for record units text
*    db_desc_dim		dimension for record description text
*    db_name_dim		dimension for channel names (record.field\0)
*    db_state_dim		number of states possible in a state table
*    db_state_text_dim		dimension for a state text string
*		usage: char state_table[db_state_dim][db_state_text_dim]
*
* type checking macros -- return non-zero if condition is true, zero otherwise
*
*      int dbf_type_is_valid(type)	type is a valid OLD_DBF_xxx
*      int dbr_type_is_valid(type)	type is a valid OLD_DBR_xxx
*      int dbr_type_is_plain(type)	type is a valid plain OLD_DBR_xxx
*      int dbr_type_is_STS(type)	type is a valid OLD_DBR_STS_xxx
*      int dbr_type_is_TIME(type)	type is a valid OLD_DBR_TIME_xxx
*      int dbr_type_is_GR(type)		type is a valid OLD_DBR_GR_xxx
*      int dbr_type_is_CTRL(type)	type is a valid OLD_DBR_CTRL_xxx
*      int dbr_type_is_STRING(type)	type is a valid OLD_DBR_STRING_xxx
*      int dbr_type_is_SHORT(type)	type is a valid OLD_DBR_SHORT_xxx
*      int dbr_type_is_FLOAT(type)	type is a valid OLD_DBR_FLOAT_xxx
*      int dbr_type_is_ENUM(type)	type is a valid OLD_DBR_ENUM_xxx
*      int dbr_type_is_CHAR(type)	type is a valid OLD_DBR_CHAR_xxx
*      int dbr_type_is_LONG(type)	type is a valid OLD_DBR_LONG_xxx
*      int dbr_type_is_DOUBLE(type)	type is a valid OLD_DBR_DOUBLE_xxx
*
* type conversion macros
*
*    char *dbf_type_to_text(type)	returns text matching OLD_DBF_xxx
*     void dbf_text_to_type(text, type) finds OLD_DBF_xxx matching text
*      int dbf_type_to_DBR(type)	returns OLD_DBR_xxx matching OLD_DBF_xxx
*      int dbf_type_to_OLD_DBR_TIME(type)	returns OLD_DBR_TIME_xxx matching OLD_DBF_xxx
*      int dbf_type_to_OLD_DBR_GR(type)	returns OLD_DBR_GR_xxx matching OLD_DBF_xxx
*      int dbf_type_to_OLD_DBR_CTRL(type)	returns OLD_DBR_CTRL_xxx matching OLD_DBF_xxx
*    char *dbr_type_to_text(type)	returns text matching OLD_DBR_xxx
*     void dbr_text_to_type(text, type) finds OLD_DBR_xxx matching text
*---------------------------------------------------------------------------*/
#define db_strval_dim MAX_STRING_SIZE
#define db_units_dim	8
#define db_desc_dim	24
#define db_name_dim	36
#define db_state_dim	16
#define db_state_text_dim 26

#define dbf_type_is_valid(type)   ((type) >= 0 && (type) <= OLD_LAST_TYPE)
#define dbr_type_is_valid(type)   ((type) >= 0 && (type) <= OLD_LAST_BUFFER_TYPE)
#define dbr_type_is_plain(type)   \
		((type) >= OLD_DBR_STRING && (type) <= OLD_DBR_DOUBLE)
#define dbr_type_is_STS(type)   \
		((type) >= OLD_DBR_STS_STRING && (type) <= OLD_DBR_STS_DOUBLE)
#define dbr_type_is_TIME(type)   \
		((type) >= OLD_DBR_TIME_STRING && (type) <= OLD_DBR_TIME_DOUBLE)
#define dbr_type_is_GR(type)   \
		((type) >= OLD_DBR_GR_STRING && (type) <= OLD_DBR_GR_DOUBLE)
#define dbr_type_is_CTRL(type)   \
		((type) >= OLD_DBR_CTRL_STRING && (type) <= OLD_DBR_CTRL_DOUBLE)
#define dbr_type_is_STRING(type)   \
		((type) >= 0 && (type) <= OLD_LAST_BUFFER_TYPE && \
		 (type)%(OLD_LAST_TYPE+1) == OLD_DBR_STRING)
#define dbr_type_is_SHORT(type)   \
		((type) >= 0 && (type) <= OLD_LAST_BUFFER_TYPE && \
		 (type)%(OLD_LAST_TYPE+1) == OLD_DBR_SHORT)
#define dbr_type_is_FLOAT(type)   \
		((type) >= 0 && (type) <= OLD_LAST_BUFFER_TYPE && \
		 (type)%(OLD_LAST_TYPE+1) == OLD_DBR_FLOAT)
#define dbr_type_is_ENUM(type)   \
		((type) >= 0 && (type) <= OLD_LAST_BUFFER_TYPE && \
		 (type)%(OLD_LAST_TYPE+1) == OLD_DBR_ENUM)
#define dbr_type_is_CHAR(type)   \
		((type) >= 0 && (type) <= OLD_LAST_BUFFER_TYPE && \
		 (type)%(OLD_LAST_TYPE+1) == OLD_DBR_CHAR)
#define dbr_type_is_LONG(type)   \
		((type) >= 0 && (type) <= OLD_LAST_BUFFER_TYPE && \
		 (type)%(OLD_LAST_TYPE+1) == OLD_DBR_LONG)
#define dbr_type_is_DOUBLE(type)   \
		((type) >= 0 && (type) <= OLD_LAST_BUFFER_TYPE && \
		 (type)%(OLD_LAST_TYPE+1) == OLD_DBR_DOUBLE)

#define dbf_type_to_text(type)   \
    (  ((type) >= -1 && (type) <= dbf_text_dim-2) ? \
        dbf_text[(type+1)] : dbf_text_invalid  )

#define dbf_text_to_type(text, type)   \
    for (type=dbf_text_dim-3; type>=0; type--) { \
        if (strcmp(text, dbf_text[type+1]) == 0) \
        break; \
    }

#define dbr_type_to_text(type)   \
    (  ((type) >= 0 && (type) < dbr_text_dim) ? \
        dbr_text[(type)] : dbr_text_invalid  )

#define dbr_text_to_type(text, type)   \
    for (type=dbr_text_dim-1; type>=0; type--) { \
        if (strcmp(text, dbr_text[type]) == 0) \
        break; \
    }

#define dbf_type_to_DBR(type)   \
    (((type) >= 0 && (type) <= dbf_text_dim-3) ? \
        (type)   :  -1  )

#define dbf_type_to_DBR_STS(type)  \
    (((type) >= 0 && (type) <= dbf_text_dim-3) ? \
        (type) + (dbf_text_dim-2)   :  -1  )

#define dbf_type_to_DBR_TIME(type)  \
    (((type) >= 0 && (type) <= dbf_text_dim-3) ? \
        (type) + 2*(dbf_text_dim-2)   :  -1  )

#define dbf_type_to_DBR_GR(type)  \
    (((type) >= 0 && (type) <= dbf_text_dim-3) ? \
        (type) + 3*(dbf_text_dim-2)   :  -1  )

#define dbf_type_to_DBR_CTRL(type)  \
    (((type) >= 0 && (type) <= dbf_text_dim-3) ? \
        (type) + 4*(dbf_text_dim-2)   :  -1  )



#ifndef DB_TEXT_GLBLSOURCE
#   ifdef VAXC
    readonly
#   endif
    extern char		*dbf_text[], *dbf_text_invalid;

#   ifdef VAXC
    readonly
#   endif
    extern short	dbf_text_dim;
#else
#   ifdef VAXC
    readonly
#   endif
    char *dbf_text[] = {
	"TYPENOTCONN",
	"OLD_DBF_STRING",
	"OLD_DBF_SHORT",
	"OLD_DBF_FLOAT",
	"OLD_DBF_ENUM",
	"OLD_DBF_CHAR",
	"OLD_DBF_LONG",
	"OLD_DBF_DOUBLE",
	"OLD_DBF_NO_ACCESS"
    };
#   ifdef VAXC
    readonly
#   endif
    char    *dbf_text_invalid = "OLD_DBF_invalid";
#   ifdef VAXC
    readonly
#   endif
    short   dbf_text_dim = (sizeof dbf_text)/(sizeof (char *));
#endif

#ifndef DB_TEXT_GLBLSOURCE
#   ifdef VAXC
    readonly
#   endif
    extern char		*dbr_text[], *dbr_text_invalid;
#   ifdef VAXC
    readonly
#   endif
    extern short	dbr_text_dim;
#else
#   ifdef VAXC
    readonly
#   endif
    char *dbr_text[] = {
	"OLD_DBR_STRING",
	"OLD_DBR_SHORT",
	"OLD_DBR_FLOAT",
	"OLD_DBR_ENUM",
	"OLD_DBR_CHAR",
	"OLD_DBR_LONG",
	"OLD_DBR_DOUBLE",
	"OLD_DBR_STS_STRING",
	"OLD_DBR_STS_SHORT",
	"OLD_DBR_STS_FLOAT",
	"OLD_DBR_STS_ENUM",
	"OLD_DBR_STS_CHAR",
	"OLD_DBR_STS_LONG",
	"OLD_DBR_STS_DOUBLE",
	"OLD_DBR_TIME_STRING",
	"OLD_DBR_TIME_SHORT",
	"OLD_DBR_TIME_FLOAT",
	"OLD_DBR_TIME_ENUM",
	"OLD_DBR_TIME_CHAR",
	"OLD_DBR_TIME_LONG",
	"OLD_DBR_TIME_DOUBLE",
	"OLD_DBR_GR_STRING",
	"OLD_DBR_GR_SHORT",
	"OLD_DBR_GR_FLOAT",
	"OLD_DBR_GR_ENUM",
	"OLD_DBR_GR_CHAR",
	"OLD_DBR_GR_LONG",
	"OLD_DBR_GR_DOUBLE",
	"OLD_DBR_CTRL_STRING",
	"OLD_DBR_CTRL_SHORT",
	"OLD_DBR_CTRL_FLOAT",
	"OLD_DBR_CTRL_ENUM",
	"OLD_DBR_CTRL_CHAR",
	"OLD_DBR_CTRL_LONG",
	"OLD_DBR_CTRL_DOUBLE"
    };
#   ifdef VAXC
    readonly
#   endif
    char    *dbr_text_invalid = "OLD_DBR_invalid";
#   ifdef VAXC
    readonly
#   endif
    short   dbr_text_dim = (sizeof dbr_text)/(sizeof (char *));
#endif

#endif /* INCLdb_accessh */
