/*	@(#)envDefs.h	1.4 8/18/91
 *	Author:	Roger A. Cole
 *	Date:	07-20-91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	07-20-91	rac	initial version
 * .02  08-07-91	joh	added ioc log env
 *
 * make options
 *	-DvxWorks	makes a version for VxWorks
 *	-DNDEBUG	don't compile assert() checking
 *      -DDEBUG         compile various debug code, including checks on
 *                      malloc'd memory
 */
/*+/mod***********************************************************************
* TITLE	envDefs.h - definitions for environment get/set routines
*
* DESCRIPTION
*	This file defines the environment parameters for EPICS.  These
*	ENV_PARAM's are initialized by $epics/share/bin/envSetupParams for
*	use by EPICS programs running under UNIX and VxWorks.
*
*	User programs can define their own environment parameters for their
*	own use--the only caveat is that such parameters aren't automatically
*	setup by EPICS.
*
* SEE ALSO
*	$epics/share/bin/envSetupParams, envSubr.c
*
*-***************************************************************************/

typedef struct envParam {
    char	*name;		/* text name of the parameter */
    char	dflt[80];	/* default value for the parameter */
} ENV_PARAM;

#ifndef ENV_PRIVATE_DATA
    extern ENV_PARAM EPICS_TS_MIN_WEST;
    extern ENV_PARAM EPICS_SYSCLK_INET;
    extern ENV_PARAM EPICS_IOCMCLK_INET;
    extern ENV_PARAM EPICS_SYSCLK_PORT;
    extern ENV_PARAM EPICS_IOCMCLK_PORT;
    extern ENV_PARAM EPICS_CMD_PROTO_PORT;
    extern ENV_PARAM EPICS_AR_PORT;
    extern ENV_PARAM EPICS_IOC_LOG_PORT;
    extern ENV_PARAM EPICS_IOC_LOG_INET;
    extern ENV_PARAM EPICS_IOC_LOG_FILE_LIMIT;
    extern ENV_PARAM EPICS_IOC_LOG_FILE_NAME;
#else
    ENV_PARAM EPICS_TS_MIN_WEST={"EPICS_TS_MIN_WEST",""};
    ENV_PARAM EPICS_SYSCLK_INET={"EPICS_SYSCLK_INET",""};
    ENV_PARAM EPICS_IOCMCLK_INET={"EPICS_IOCMCLK_INET",""};
    ENV_PARAM EPICS_SYSCLK_PORT={"EPICS_SYSCLK_PORT",""};
    ENV_PARAM EPICS_IOCMCLK_PORT={"EPICS_IOCMCLK_PORT",""};
    ENV_PARAM EPICS_CMD_PROTO_PORT={"EPICS_CMD_PROTO_PORT",""};
    ENV_PARAM EPICS_AR_PORT={"EPICS_AR_PORT",""};
    ENV_PARAM EPICS_IOC_LOG_PORT={"EPICS_IOC_LOG_PORT","6500"};
    ENV_PARAM EPICS_IOC_LOG_INET={"EPICS_IOC_LOG_INET",""};
    ENV_PARAM EPICS_IOC_LOG_FILE_LIMIT={"EPICS_IOC_LOG_FILE_LIMIT","1000000"};
    ENV_PARAM EPICS_IOC_LOG_FILE_NAME=
		{"EPICS_IOC_LOG_FILE_NAME","./iocLog.text"};
#endif

char *envGetConfigParam(ENV_PARAM *pParam, int bufDim, char *pBuf);
long envPrtConfigParam(ENV_PARAM *pParam);
long envSetConfigParam(ENV_PARAM *pParam, char *value);
long envGetInetAddrConfigParam(ENV_PARAM *pParam, struct in_addr *pAddr);
long envGetDoubleConfigParam(ENV_PARAM *pParam, double *pDouble);
long envGetLongConfigParam(ENV_PARAM *pParam, long *pLong);

