/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include "sdrLoad.h"
#include "dbAddr.h"
 
extern struct recType		*dbRecType;
extern struct arrChoiceSet	*choiceGbl;
extern struct choiceRec		*choiceRec;
extern struct devChoiceRec	*choiceDev;
extern struct recDevSup		*devSup;
extern struct drvSup		*drvSup;
extern struct arrBrkTable	*cvtTable;
extern struct choiceSet		*choiceCvt;
extern struct recDes		*dbRecDes;
extern struct recSup		*recSup;
extern struct pvd		*dbPvd;
extern struct recHeader		*dbRecords;

/* define structures */

typedef struct pointdata
{
  char 	pvname[PVNAME_SZ + 1];
  unsigned short  recType;		/* index into *dbRecType */
  unsigned short  crc;			/* checksum as computed by dbClient */
  unsigned int	timestamp;		/* modification time */
  short numTableSize;			/* size of array from Table */
  short numStrings;			/* number of enum strings */
  double  *numberTable;			/* pointer to array of Disp & Cntrl values */
  char 	*enumTable;			/* pointer to array of enum strings */
  void	*currValue;			/* pointer to current value */
} PTDATA, *PTDATA_PTR;


typedef struct sdrfield
{
    char	sdrField[FLDNAME_SZ];
    char	operandId[21];
    DATA4	*dbFile;
    FIELD4	*dbFld;
    struct sdrfield	*nextField;
} SDRFIELD, *SDRFIELD_PTR;

typedef struct udbpvs
{
    char	pvname[PVNAME_SZ + 1];
    int		usagecnt;
    struct udbpvs *nextUdb;
} UDBPVS, *UDBPVS_PTR;

typedef struct _dbfiles
{
  char  dbName[80];		/* name from PDBXREF */
  DATA4 *dbFile;		/* zero means file not open yet */
  struct _dbfiles *nextFile;	/* next in list or zero if end */
} DBFILES, *DBFILES_PTR;
extern DBFILES_PTR *dbFileArray;
/* database stuff */

#ifdef WIN32
#define FA_DATABASEPATH "\\tis4000db\\"
#else
#define FA_DATABASEPATH "/tis4000db/"
#endif WIN32
#define	DBTREENAME	"dbTree.txt"
#define FA_STRING_SIZE	32	/* this is the same as ATTR_STR_LEN in chanAcq.c */

CODE4	FA_CodeBase;
DATA4	*FA_Pvmap;
DATA4	*FA_XrefFile;
FIELD4	*FA_PvmapNAME;
FIELD4	*FA_PvmapTYPE;
FIELD4	*FA_PvmapDTYP;
FIELD4	*FA_PvmapCRC;
FIELD4	*FA_PvmapTSTP;
FIELD4 *dbaseField;
FIELD4 *dbaseFile;
FIELD4 *operandID;
FIELD4 *deleteFlag;

/* specials stuff */

typedef enum {SPECL_DISPLOW,SPECL_DISPHIGH,SPECL_CTRLOW,SPECL_CTRHIGH,SPECL_PREC} SPECL_CONTROL;

typedef enum {PASS_1, PASS_2} SPECL_PASS;

/*
  Special dispatch table
*/
extern double (*dbSpecials[])(char *, short, char *, SPECL_CONTROL);

typeDesc **bldDesDataTable ();
void	FA_convert_dbAddr (struct dbAddr *, char *);
int	FA_dbNameToAddr (char *, int, struct dbAddr *);
int	FA_InitAttributes (void);
int	FA_InitDatabase (void);
int	FA_InitAvlTree (typeDesc **descTable);
void	FA_removeTrailingSpaces (char *);
int	FA_NameCompare(void *, void *);
PTDATA_PTR FA_FindPoint(char *name);
double	FA_GetLimitValue(char *pv,short index,limitStruct *limit);
double	FA_GetDBValue(char *pv,short index,char *sdrName);
char	*FA_GetEnumValue(char *pv,short index,enumField *eptr,int *tSize);
DATA4	*getFile(char *fname,short index);
PTDATA_PTR FA_AddNode(PTDATA_PTR ptPtr);
PTDATA_PTR FA_AddLeaf(char *pvname,typeDesc **descTable);
void	FA_AssignValues(PTDATA_PTR pData,typeDesc **descTable);
void	FA_AssignUDBValues(PTDATA_PTR pData,typeDesc **descTable);
DATA4 *getFile(char *fname, short index);
int getAttributes (char *pvname, int pointType, int	maxString,
		   char stringList[][32], int longestString,
		   unsigned short crc,unsigned timestamp,
		   short *nativeType, short *nativeCount,
		   double *hopr, double *lopr, double *drvh,
		   double *drvl, short *prec,short *numStrings);
int FA_FetchData(typeDesc **descTable,char *pv,short index,
		 unsigned short crc,unsigned int tstamp,
		 double *dl,double *dh,double *cl,double *ch,double *pr,
		 char **eStrings,int *numStrs);
int FA_FetchSpecialData(typeDesc **descTable,char *pv,short index,
			double *dl,double *dh,double *cl,double *ch);

void FA_WriteNode(FILE *fp,PTDATA_PTR node);	/* called during traversal */
int FA_WriteTree(char *filename);
int FA_ReadTree(char *filename);

void FA_SetValuePtr(char *name, void *ptr);
void *FA_GetValuePtr(char *name);

