/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		LOCAL DB FOR IOC NAMES AND PRIV			*/
/*								*/
/*	Definitions for local ioc TIS database support.		*/
/*								*/
/*	Author: Grant Irani					*/
/*								*/
/****************************************************************/

#ifndef IOC_TISDB_H
#define IOC_TISDB_H

/* Load system size definitions */
#include "dbDefs.h"

typedef struct
{
    char		iocName[IOCNAME_SZ+1];
    struct in_addr	inetAddr;
} IOCCACHE;

typedef struct
{
    char		resName[RESOURCE_SZ+1];
    IOCCACHE		*iocPtr1;
    IOCCACHE		*iocPtr2;
} RESCACHE;

typedef struct
{
    char		pointName[PVNAME_SZ+1];
    short		pointType;
    RESCACHE		*resPtr;
    char		privgroup[PROTECTIONGRP_SZ+1];
} POINTNAME;

typedef struct
{
    short		pointType;
    char		fieldname[FLDNAME_SZ+1];
    short		level;
} FLDPROT;

typedef struct
{
    char		userName[OPERATOR_SZ+1];
    char		group[OPERATOR_SZ+1];
} SYSUSER;

typedef struct
{
    char		group[OPERATOR_SZ+1];
    char		privGroup[PROTECTIONGRP_SZ+1];
    short		level;
} PRIVTAB;

/******************************************************************************
**
**  Published data
**
*******************************************************************************/

extern IOCCACHE    *TISDBiocAddrs;
extern int	    TISDBnumIocs;

extern RESCACHE    *TISDBresources;
extern int	    TISDBnumRes;

extern POINTNAME   *TISDBpointNames;
extern int	    TISDBnumNames;


/******************************************************************************
**
**  Access routines
**
*******************************************************************************/

/* Initialize the data */
void			InitTisDB (char *dataFile);

/* Print a point entry, or print all if NULL passed */
POINTNAME		*iocNSPoints (char *name);

/* Find a point entry */
POINTNAME		*findPoint (char *name);

/* Print a IOC entry, or print all if NULL passed */
IOCCACHE		*iocNSIocs (char *name);

/* Find a IOC entry */
IOCCACHE		*findIOC (char *name);

/* Print a resource entry, or print all if NULL passed */
RESCACHE		*iocNSResources (char *name);

/* Find a resource entry */
RESCACHE		*findRes (char *name);

/* Print a fieldProt entry, or print all if NULL passed */
FLDPROT			*iocFldProt (char *fldname, int type);

/* Find a fieldProt entry */
FLDPROT			*findFldProt (char *fldname, int type);

/* Print a sysUser entry, or print all if NULL passed */
SYSUSER			*iocSysUser (char *name);

/* Find a sysUser entry */
SYSUSER			*findSysUser (char *name);

/* Print a privTab entry, or print all if NULL passed */
PRIVTAB			*iocPrivTab (char *usergroup, char *pgroup);

/* Find a privTab entry */
PRIVTAB			*findPrivTab (char *usergroup, char *pgroup);

#endif IOC_TISDB_H
