/* link.h */
/* share/epicsH @(#)link.h	1.7     6/7/91 */

/*
 *      Original Author: Bob Dalesio
 *      Current Author:  Marty Kraimer
 *      Date:            6-1-90
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  mm-dd-yy        iii     Comment
 * .02  05-10-94        wep     added definition for Valmet
 * .03  08-25-94        wep     added definition for Trane
 * .04	11-01-94	ems	added HYUNDAI_IO
 * .05  01-09-95	whs	added RTP_IO
 * .06  01-27-95        wep     added definition for Ge
 * .07  01-27-95        wep     added definition for Hart
 * .08  01-27-95        wep     added definition for Sv85
 * .09  04-10-95        wep     added definition for Bailey
 * .10	07-19-95	ems	added SECURITY_IO
 * .11  07-27-95        wep     added definition for Netpac
 * .12	09-14-95	ems	added EAB_IO
 * .13	09-18-95	ems	added VAB_IO
 * .14	09-19-95	ems	added NUM_LINK_TYPES
 * .15  10-20-95        wep     added definitions for Tosg2 and Tosg3
 * .16	11-15-95	ems	added CMPMTR_IO
 * .17  11-29-95	whs	added ANALOGIC_IO
 * .18  01-17-96	wep	added METONE_IO
 * .19  09-18-96	wep	added MR350_IO
 * .20  02-12-97	wep	added LIMITORQUE_IO, MULTILIN_IO, and WPCS_IO
 * .21  07-28-97	wep	added TANKWAY_IO and NWMOD_IO
 * .22  09-15-97	wep	removed SECURITY_IO and added INF7_IO
 * .23  02-03-98	wep	added BENTLY_IO
 * .24  02-12-98	wep	replaced DANIELS_IO with DANIEL_IO
 * .25  06-01-98	wep	added PABPLC_IO
 * .26	04-15-99	ems	chg HART to QUAD4, chg GE to SEC add OMNIFLOW
 *				and MOMENTUM
 * .27	04-16-99	ems	added DNP
*/


/* link types */
#ifndef INClinkLT
#define INClinkLT
#define CONSTANT	0
#define PV_LINK		1
#define VME_IO		2
#define CAMAC_IO	3
#define	AB_IO		4
#define GPIB_IO		5
#define BITBUS_IO	6
#define DB_LINK		10
#define CA_LINK		11
#define INST_IO		12		/* instrument */
#define	BBGPIB_IO	13		/* bitbus -> gpib */

/****************************************/
/*					*/
/*		WARNING!		*/
/*					*/
/* It is assumed that all link types	*/
/* >= SAB_IO use the tate_io struct	*/
/* for their parameters. If you change	*/
/* this, also modify redund/localEndian	*/
/*					*/
/****************************************/

#define	SAB_IO		14		/* Serial AB */
#define	SMOD_IO		15		/* Serial Modbus */
#define	SMIC_IO		16		/* Serial Microsol */
#define	SABRTU_IO	17		/* Serial AB RTU */
#define	STCS_IO		18		/* Serial TCS */
#define	SAD_IO		19		/* Serial AD */
#define	RES_IO		20		/* Resource Control / Status */
#define SBDX_IO		21		/* Bendix LMP */
#define SBRKS_IO	22		/* Brooks Meter */
#define TI_IO		23		/* TI PLC */
#define V984_IO		25		/* VME984 Modicon */
#define TIWAY_IO	26		/* TIWAY dev */
#define ARCOM_IO	27		/* Arcom dev */
#define MDS_IO		28		/* Mds dev */
#define ENRAF_IO	29		/* Enraf dev */
#define DANIEL_IO	30		/* Daniel dev */
#define QEI_IO		31		/* Qei dev */
#define MAPBD_IO	32		/* Mapbd dev */
#define CSI_IO		33		/* Csi dev */
#define RTU_IO		34		/* Rtu dev */
#define VALMET_IO       35              /* Valmet dev */
#define TRANE_IO        36              /* Trane dev */
#define	HYUNDAI_IO	37		/* Hyundai dev */
#define RTP_IO		38		/* RTP / DAP dev */
#define SEC_IO		39		/* SEC UPS dev */
#define QUAD4_IO	40		/* Quad4 meter dev */
#define SV85_IO		41		/* Sv85 dev */
#define BAILEY_IO	42		/* Bailey dev */
#define	INF7_IO		43		/* Inf7 dev */
#define NETPAC_IO	44		/* Netpac dev */
#define	EAB_IO		45		/* Ethernet Allen Bradley */
#define	VAB_IO		46		/* VME Allen Bradley Scanner */
#define	TOSG2_IO	47		/* Toshiba G2 Inverter */
#define	TOSG3_IO	48		/* Toshiba G3 Inverter */
#define	CMPMTR_IO	49		/* Compumotor */
#define ANALOGIC_IO	50		/* Analogic */
#define METONE_IO	51		/* Metone: Met One PCX */
#define MR350_IO	52		/* Mr350: MR-350 DCT */
#define LIMITORQUE_IO	53		/* Limitorque: L120 Valve Control */
#define	MULTILIN_IO	54		/* GE Multilin: SR469 & SR750 */	
#define	WPCS_IO		55		/* IBR Water Particle Counter System */
#define	TANKWAY_IO	56		/* L & J Engineering Tankway Protocol */
#define	NWMOD_IO	57		/* North Wales: Modbus Special */
#define	BENTLY_IO	58		/* Bently Nevada 3300/03 SDI Modbus Interface */
#define	PABPLC_IO	59		/* Peer-To-Peer Allen-Bradley Driver */
#define	OMNIFLOW_IO	60		/* Omniflow meter */
#define	MOMENTUM_IO	61		/* Momentum driver */
#define	SMODSLV_IO	62		/* Modbus Slave Driver */
#define	DNP_IO		63		/* DNP driver */
#define	CB_IO		64		/* CB driver */


/************************************************************************
*  
* NOTE:
*	if you add to link list update NUM_LINK_TYPES because it is used
*	in .../share/src/db/dbtest.c
*	ALSO:
*		add name of IO string to the boards[array] in db/dbTest.c 
*		for dbpr to print out the name of the hardware, see
*		db/dbTest.c::dbprReportLink()
*
*************************************************************************/
#define	NUM_LINK_TYPES	(CB_IO + 1)

/****************************************/
/*					*/
/*		WARNING!		*/
/*					*/
/* It is assumed that all link types	*/
/* >= SAB_IO use the tate_io struct	*/
/* for their parameters. If you change	*/
/* this, also modify redund/localEndian	*/
/*					*/
/****************************************/

#endif  INClinkLT


#ifndef INClinkh
#define INClinkh 1
#define	INSTIO_FLD_SZ	36	/* field size for instio */
#define	LINK_PARAM_SZ	32

/*????? what about vxi*/

/* structure of a link to a process variable*/
struct pv_link{
	short process_passive; /* should in/out passive link be processed*/
	short maximize_sevr;   /* maximize sevr of link. stat=LINK_ALARM*/
	char pvname[PVNAME_SZ];
	char fldname[FLDNAME_SZ];
};
/* structure of a link to a database link */
struct db_link{
	short process_passive; /* should in/out passive link be processed*/
	short maximize_sevr;   /* maximize sevr of link. stat=LINK_ALARM*/
	void *pdbAddr;
	char pvname[PVNAME_SZ];	/* Copied so we can synchronize */
	char fldname[FLDNAME_SZ];	/* Copied so we can synchronize */
};
/* structure of a link to a channel access link */
struct ca_link{
	short process_passive; /* should in/out passive link be processed*/
	short maximize_sevr;   /* maximize sevr of link. stat=LINK_ALARM*/
	void *pca;
	char pvname[PVNAME_SZ];	/* Copied so we can synchronize */
	char fldname[FLDNAME_SZ];	/* Copied so we can synchronize */
};


/* structure of a VME io channel */
struct vmeio{
	short	card;
	short	signal;
};

/* structure of a CAMAC io channel */
struct camacio{
	short	branch;
	short	crate;
	short	slot;
	short	channel;
};

/* structure of a Allen-Bradley io channel */
struct abio{
	short	link;
	short	adapter;
	short	card;
	short	signal;
	short	plc_flag;
};

/* structure of a gpib io channel */
struct gpibio{
	short	link;
	short	addr;		/* device address */
	char	parm[LINK_PARAM_SZ];
};

/* structure of a bitbus io channel */
struct	bitbusio{
	unsigned char	link;
	unsigned char	node;
	unsigned char	port;
	unsigned char	signal;
	char	parm[LINK_PARAM_SZ];
};

/* structure of a bitbus to gpib io channel */
struct	bbgpibio{
	unsigned char	link;
	unsigned char	bbaddr;
	unsigned char	gpibaddr;
	unsigned char	pad;
	char	parm[LINK_PARAM_SZ];
};

/* structure of an instrument io link */
struct	instio {
	char	string[INSTIO_FLD_SZ];		/* the cat of location.
							signal.parameter  */
};
/* structure of a Serial Modbus io channel */
struct tateio{
	short	link;
	short	adapter;
	short	card;
	short	signal;
};

/* union of possible address structures */
union value{
	float		value;		/* constant */
	struct pv_link	pv_link;	/* link to process variable*/
	struct vmeio	vmeio;		/* vme io point */
	struct camacio	camacio;	/* camac io point */
	struct abio	abio;		/* allen-bradley io point */
	struct gpibio	gpibio;
	struct bitbusio	bitbusio;
	struct db_link	db_link;	/* Data base link	*/
	struct ca_link	ca_link;	/* Channel Access link	*/
	struct instio	instio;		/* instrument io link */
	struct bbgpibio bbgpibio;	/* bitbus to gpib io link */
	struct tateio	tateio;		/* Serial Modbus io point */
};

struct link{
#ifndef AXP
	long		type;
	union value	value;
#else
	int		type;
	union value	value __attribute__ ((packed));
#endif
};
#endif
