/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		TIS 4000 CENTRAL MESSAGE FACILITY		*/
/*								*/
/*	This file defines the TIS 4000 message facility. There	*/
/*	are four routines that are as similar as possible	*/
/*	to the Unix system message logging routines.		*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#ifndef	tismsgmonH
#define	tismsgmonH

/****************************************************************/
/*								*/
/*			MESSAGE ROUTINES			*/
/*								*/
/****************************************************************/

	/*
	Similar to openlog():

		tisOpenLog(char *ident);

			ident:		points to a character string that identifies
					the source of the message.

	*/
int tisOpenLog(char *ident, int ceiling, int crossOver, void (*sigUsr1)(), void (*sigUsr2)());


	/*
	Similar to syslog():

		tisSysLog(int debugLevel, int maxFreq, char *message, parameters...);

			debugLevel:	Debug level of the current message. Messages are only
					printed when debug level of the message is less than
					or equal to the current debug level. The default debug
					level is zero.

			maxFreq:	Maximum frequency of this message in seconds.

			message:	Same as corresponding argument to syslog.
					See Unix man pages.

			parameters:	Same as corresponding argument to syslog.
					See Unix man pages.
	*/
/* the parameters are determined from the format */
int tisSysLog(int debugLevel, int maxFreq, char *message, ... );


int tisGetLogLevel(void);


	/*
	YOU DON'T HAVE TO CALL THIS ROUTINE IF YOU ARE USING THE DEFAULT SIGNAL HANDLERS.
	YOU MUST CALL THIS ROUTINE IF YOU ARE USING YOUR OWN SIGNAL HANDLERS.

	Similar to setlogmask():

		tisBumpLogLevel(int incr)

			incr:		The amount to increase or decrease the current debug level.
					This argument is usually +1 or -1. Decrements below level
					zero or above the ceiling are silently ignored. The routine 
					returns the current debug level (i.e, after the 
					increment/decrement). Changes in debug level are always logged.
	*/
int tisBumpLogLevel(int incr);


	/*
	Same parameters as closelog():

		tisCloseLog(void);
	*/
void tisCloseLog(void);

#endif	tismsgmonH

