/* pvd.h */

/* dbPvd.h
 *
 * record name and field name convert subroutines
 *
*/

/*
 * Subroutines
 *
 * get_recfld	convert a record type and field name to offset, size, field type
 *      args
 *		type		record type
 *		pfldname	pointer to the field name
 *		ptype		pointer to the field type
 *		poffset		pointer to the field offset
 *		psize		pointer to the field size
 *	returns
 *		0		found
 *		-1		not found
 *
 * pvdHash		get a collision chain index for a PV name
 *	args
 *		str		pointer to the Process Variable name
 *	returns
 *		hval		the collision chain index
 * pvdPvSearch 	find the specified PV name in the PVD
 *	args
 *		name		pointer to the Process Variable name
 *		hashinx		the collision chain index
 *	returns
 *		>= 0		PVD entry index of the specified PV
 *		-1		not found
 * pvdGetFld	fetches the field type etc.
 *      args
 *		type		record type
 *		pfldname	pointer to field name
 *	returns
 *		NULL		not found
 *		struct fldDes*	If found
 * pvdGetRec	fetches the TYPE and RECORD_NO of a Process Variable
 *	arguments
 *		ppv_name	pointer to the name of the Process Variable
 *		type		pointer to the record type of the PV
 *		record_no	record number of this PV
 *	outputs
 *		-1		record not defined
 *		0		record found
 */


#ifndef EPICS_V2
int get_recfld (short	type,		/*record type*/
		char	*pfldname,	/*field name */
		short	*ptype,		/*field type */
		short	*poffset,	/*field offset*/
		short	*psize);	/*field size */
#endif /* EPICS_V2 */

int		add_pv(char *input,short type,short record_no);
short		pvdHash (char *str);
short		pvdPvSearch (char *pname, short hashinx);
struct fldDes	*pvdGetFld (short type,	/* record type		*/
			    char *pfldname);/* field name		*/
int		pvdGetRec (char *ppv_name,
			   short *ptype,
			   short *precord_no);
int		pvdGetAddr (char *ppv_name, char **record_address);
int		pvdSetAddr (char *ppv_name, char *addr);
