/* recSup.h	Record Support		*/
/* share/epicsH @(#)recSup.h	1.5     6/7/91 */

/*
 *      Author:          Marty Kraimer
 *      Date:            6-1-90
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  mm-dd-yy        iii     Comment
 * .02  09-13-93	ems	added S_rec_Alloc
 * .03  10-18-93	ems	added S_rec_dbName
 */
#ifndef INCrecSuph
#define INCrecSuph 1

typedef long (*RECSUPFUN) ();	/* ptr to record support function*/

struct rset {	/* record support entry table */
	long		number;		/*number of support routines	*/
	RECSUPFUN	report;		/*print report			*/
	RECSUPFUN	init;		/*init support			*/
	RECSUPFUN	init_record;	/*init record			*/
	RECSUPFUN	process;	/*process record		*/
	RECSUPFUN	special;	/*special processing		*/
	RECSUPFUN	get_value;	/*get value field		*/
	RECSUPFUN	cvt_dbaddr;	/*cvt  dbAddr			*/
	RECSUPFUN	get_array_info;
	RECSUPFUN	put_array_info;
	RECSUPFUN	get_units;
	RECSUPFUN	get_precision;
	RECSUPFUN	get_enum_str;	/*get string from enum item*/
	RECSUPFUN	get_enum_strs;	/*get all enum strings		*/
	RECSUPFUN	put_enum_str;	/*put string from enum item*/
	RECSUPFUN	get_graphic_double;
	RECSUPFUN	get_control_double;
	RECSUPFUN	get_alarm_double;
	};
struct recSup {
	long		number;		/*number of record types	*/
	struct rset	**papRset;	/*ptr to arr of ptr to rset	*/
	};
#define RSETNUMBER ( (sizeof(struct rset) - sizeof(long))/sizeof(RECSUPFUN) )
extern struct recSup *recSup;

#define S_rec_noRSET     (M_recSup| 1) /*Missing record support entry table*/
#define S_rec_Alloc      (M_recSup| 3) /* failed m,c alloc */
#define S_rec_dbName     (M_recSup| 5) /* dbNameToAddr failed */

/*************************************************************************
 *
 * The following is a valid reference
 *
 * recSup->papRset[1]->report(<args>)
 *************************************************************************/


/* Definition os structure for routine get_value */

struct valueDes {
	long	field_type;
	long	no_elements;
	void	*pvalue;
};

/**************************************************************************
 * The following macro returns either the addr of rset or NULL
 * A typical usage is:
 *
 *	struct rset *prset;
 *	if(!(prset=GET_PRSET(type))) {<null}
 ***********************************************************************/
#define GET_PRSET(REC_TYPE)\
(\
    recSup\
    ?(\
	(((REC_TYPE) < 1) || ((REC_TYPE) >= recSup->number))\
	?   NULL\
	:   (recSup->papRset[(REC_TYPE)])\
    )\
    : NULL\
)



/************************************************************************
 * report(FILE fp,struct db_addr *paddr);
 * init();
 * init_record(precord);
 * process(struct db_addr *paddr);
 * special(struct db_addr *paddr, after);
 * get_value(precord,struct valueDes *p);
 * cvt_dbaddr(struct db_addr *paddr);
 * get_array_info(paddr,long *no_elements,long *offset);
 * put_array_info(paddr,nNew);
 * get_units(paddr,char units[8]);
 * get_precision(struct db_addr *paddr,long *precision);
 * get_enum_str(paddr,pbuffer);
 * get_enum_strs(paddr,struct dbr_enumStrs *p);
 * put_enum_str(paddr,pbuffer);
 * get_graphic_double(paddr,struct dbr_grDouble *p);
 * get_control_double(paddr,struct dbr_ctrlDouble *p);
 * get_alarm_double(paddr,struct dbr_ctrlDouble *p);
 ***********************************************************************/
#endif
