/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/* serLib.h - definitions for Serial Interface Library header */

#ifndef __SERLIB__
#define __SERLIB__
#include <vxWorks.h>
#include <semLib.h>
#include "serhwLib.h"

/* error codes to place in errno */

#define modSER (0x00700000)
#define S_ser_UNINITIALIZED          (modSER|1)
#define S_ser_ALREADY_RUNNING        (modSER|2)
#define S_ser_NO_MEMORY              (modSER|3)
#define S_ser_LINEID_OUT_OF_RANGE    (modSER|4)
#define S_ser_LINEID_ALREADY_USED    (modSER|5)
#define S_ser_LINEID_UNINITIALIZED   (modSER|6)
#define S_ser_UNSUPPORTED_IOCTL      (modSER|7)
#define S_ser_PARAMETER_OUT_OF_RANGE (modSER|8)

/*
  control structure for mapping a given logical channel to
  the underlying characteristic module
*/

struct _line
{
/*interface to hardware*/
  struct _hwifc *hwifc;		/*std ifc dispatch blk*/
  USHORT hwcfg;			/*packed serial characteristics*/
  USHORT modem;			/*current modem control/status standardized*/
/*interface to user task*/
  BOOL rxRaw;			/*line is not being error filtered*/
  VOIDFUNCPTR rxHook;		/*user rxISR routine*/
  void *rxParm;			/*user rxISR parameter (p0)*/
  VOIDFUNCPTR abHook;		/*user rxISR(ERROR) routine (!rxRaw)*/
  void *abParm;			/*user rxISR(ERROR) parameter (p0)*/
  VOIDFUNCPTR txHook;		/*user txISR routine*/
  void *txParm;			/*user txISR parameter (p0)*/
  VOIDFUNCPTR stHook;		/*user statchangeISR routine*/
  void *stParm;			/*user statchangeISR parameter (p0)*/
/*housekeeping*/
  ULONG numrxints;
  ULONG numtxints;
  ULONG numstints;
  ULONG numrxerrs;
  ULONG numparity;
  ULONG numframing;
  ULONG numoverrun;
  ULONG numunderrun;
  ULONG numcrcerrors;
  SEM_ID mx;			/*mutual exclusion channel protection*/
};

/*
  enumerated constants for use by upper level applications to select
  particular IOCTL functions supported by the Serial Interface Library -
  additional required parameters are indicated by the associated comments
*/

enum _serioctl
{
  serIOCTLCODES = 0x1000,	/*try not to compete with std ioctls*/
  serGETSTATS,     /* */
  serGETDATARATE,  /* ushort* */
  serGETMODE,      /* char* */
  serGETBAUD,      /* ushort* */
  serGETPARITY,    /* ushort* */
  serGETFRAMING,   /* ushort*, ushort* */
  serGETMDMSTAT,   /* ushort* */
  serSETMODE,      /* char* */
  serSETBAUD,      /* enum bX */
  serSETPARITY,    /* enum parX */
  serSETFRAMING,   /* enum bitsX, enum stopX */
  serSETMDMCTL,    /* ushort newval, ushort chgmask */
  serSETRXHOOK,    /* void (*)() [, parm] */
  serSETABHOOK,    /* void (*)() [, parm] */
  serSETTXHOOK,    /* void (*)() [, parm] */
  serSETSTHOOK,    /* void (*)() [, parm] */
  serSETRX_RAW,    /* */
  serSETRX_FILTER, /* */
  serRESETDEV,     /* */
  serTXUNDRUN,     /* Controls underrun action in sync mode (CRC/ABORT) */
  serMARKFLAG,	   /* Control Idle condition in sync mode (MARK/FLAG) */
  serSETADDR,      /* Set physical address in syn mode */
  serINT = 0x8000, /*flag for interrupt level*/
  ser_= -1
};

/* prototypes for published routines */

STATUS serLibInit( int );
STATUS serOpen( int );
STATUS serClose( int );
STATUS serWrite( int, UCHAR );
STATUS serWriteBlk( int, UCHAR*, int );
STATUS serRead( int );
STATUS serReadBlk( int, UCHAR*, int, UCHAR );
STATUS serIoctl( int, int, ... );
void serMapShow( void );

#endif /*__SERLIB__*/
