/*sockTune.h,v 1.1.1.1 1999/08/03 14:41:29 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		SOCKET I/O TUNING PARAMETERS			*/
/*								*/
/*	This structure contains the paramters that control	*/
/*	the usage of a TCP/IP socket connection. There is	*/
/*	one instance this structure for each IOC. Another	*/
/*	process can request these structures from Socket	*/
/*	Creation and Maintenance and can also send them		*/
/*	back so that the system can be dynamically tuned.	*/
/*								*/
/*	Author:	Gordon Uchenick					*/
/*								*/
/****************************************************************/

typedef struct
    {
    struct in_addr stInet;		/* Inet Address */
    int		stMaxIdle;	/* Transmit idle time before boomerang */
    int		stMaxQuiet;	/* # Boomerangs missing to declare node dead */
    int		stMaxBlock;	/* Maximum # bytes in a single transmit */
    int		stInqRate;	/* Interval between connect attempts */
    } sockTune;

/******
	The tuning parameters in this structure have the following
	meaning:

	stInet:
		Not really a tuning parameter, a zero indicates that
		this entry is a default to be used for connections
		that have not yet been explicitily tuned.

	stMaxIdle:
		Socket Creation and Maintenance keeps track of the
		time of day for each transmission. When the time of
		the last transmission was more than stMaxIdle seconds
		ago, any outstanding buffer will be flushed. If the
		buffer is empty, the boomerang will be sent.

	stMaxQuiet:
		stMaxQuiet is the number of boomerangs that have to be
		missing in order to declare the IOC connection dead.
		Nominally, stMaxQuiet would be three to five.  There are 
		three reasons for implementing our own keep alive instead 
		of using TCP/IP facilities. First, we can change our tuning
		parameters without having to remake the kernel. Secondly, we 
		can change these parameters while the system is running, 
		making it much easier to tune the system while it is running.
		Third, we can use a set of parameters for each socket
		which is desirable when some nodes are connected via LAN
		and others by WAN.

	stMaxBlock:
		The max number of bytes that we transmit in a single call
		to send(). TCP/IP will package bytes into transmission
		envelopes on its own, but this does give the user a little
		bit of control on the size of outgoing packets.

	stInqRate:
		When an IOC is not connected, Socket Creation and Maintenance
		sends UDP messages to it. Each message is an invitation for
		the IOC to create a connection to the workstation. There will
		be a delay of stInqRate seconds between these invitations.
******/


/****************************************************************/
/*								*/
/*		DEFAULT TUNING PARAMETERS			*/
/*								*/
/*	The following values are to be used when no defaults	*/
/*	have been set up by the user. These numbers are based	*/
/*	on a system with a mix of local and remote IOCs.	*/
/*								*/
/****************************************************************/

sockTune defaultTuning =
    {
    {0},			/* Inet Address */
    6,			/* 6 seconds idle before boomerang */
    3,			/* 3 boomerangs missing to determine death */
    1024,		/* 1K byte messages */
    60,			/* 1 minute reconnection attempts */
    };

#define MAX_XMIT_BUFFER	4096	/* Biggest msg we'll ever send */

/****************************************************************/
/*								*/
/*		TUNING PARAMETER FILE				*/
/*								*/
/*	There is no point in tuning the system if the user	*/
/*	will have to do it every time that it is started.	*/
/*	We will keep all tuning parameters in a file.		*/
/*	This file will be named /tis4000db/socketTuning.	*/
/*	Socket Creation and Maintenance will be the only	*/
/*	program that reads and writes this file. Any tuning	*/
/*	utility is to do its job by requesting the current	*/
/*	tuning parameters, modifying them, and then sending	*/
/*	the modified parameters back.				*/
/*								*/
/****************************************************************/

#define TUNING_FILE_NAME	"/tis4000db/socketTuning"
