/* special.h */
/* share/epicsH @(#)special.h	1.6     6/7/91 */

/*
 *      Author:          Marty Kraimer
 *      Date:            6-1-90
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  06-07-91        mrk     Cleaned up
 * .02  04-05-93        ems     added SPC_STRAP_TBL
 * .03  04-08-93        ems     added SPC_API_TBL
 * .04  09-13-93	ems	added SPC_DT_VAL,SPC_DT_NUMP
 * .05  10-08-93	ems	added SPC_DM_INAM, SPC_DM_ONAM
 * .06  06-02-94	ems	added SPC_PID_CT,SPC_LL_TIM0,SPC_RAMP_LTIM
 * .07	06-06-94	ems	added SPC_TD_CTCK and SPC_TD_CBMD
 * .08	09-12-94	ems	added SPC_ACC_RUN, SPC_ACC_LTCK
 * .09	10-04-94	ems	chg SPC_ACC_RUN to SPC_ACC_RTOT etc
 * .10	10-10-94	ems	added SPC_ACC_NFTM
 * .11	10-13-94	ems	added SPC_ACC_FSEC and SPC_ACC_RSEC
 * .12	01-18-95	ems	added SPC_UDB_CTCK
 * .13	05-15-95	ems	added SPC_PTO_RVAL
 * .14	06-20-95	ems	added SPC_PAT_CT, SPC_PAT_RVAL
 * .15	01-11-96	ems	added SPC_UDB_ZCMF
 * .16  01-31-96	ems	added SPC_AO_VAL, SPC_BO_VAL, SPC_MBBO_VAL, SPC_SO_VAL
 * .17	02-05-96	ems	added SPC_BO_CBFL
 * .18	02-09-96	ems	added SPC_PUMP_TICKS
 * .19	04-16-96	ems	added SPC_DDC_VAL, SPC_DDC_OMSL
 * .20	04-30-96	ems	deleted SPC_DDC_OMSL added SPC_DDC_LTCK and SPC_DDC_TSXX
 * .21	05-03-96	ems	added SPC_PAT_HAND
 * .22	05-22-96	ems	added SPC_LDFW_CTCK
 */

#ifndef INCspecialh
#define INCspecialh 1
/* types 1-99 are global. Record specific must start with 100 */
#define SPC_NOMOD	1	/* Field must not be modified */
#define SPC_DBADDR	2	/* db_name_to_addr must call cvt_dbaddr	*/
#define SPC_SCAN	3	/* A scan related field is being changed */

/* useful when record support must be notified of a field changing value */
#define SPC_MOD		100

/* used by all records that support a reset field */
#define SPC_RESET	101	/* The res field is being modified */

/* Specific to conversion (Currently only ai */
#define SPC_LINCONV	102	/* A linear conversion field is being changed */

/* Specific to calculation records */
#define SPC_CALC	103	/* The CALC field is being changed */

/* Specific to records that use strapping tables, ex pipe, tank */
#define SPC_STRAP_TBL	104	/* the strapping table name is being changed */
#define SPC_API_TBL     105	/* the API table number is being changed */

/* Specific to deadtime records */
#define SPC_DT_VAL 	106	/* if VAL chg's chg STAT and SEVR when in MANUAL */
#define SPC_DT_NUMP     107	/* deadtime, num to push onto que */
#define SPC_DT_DETS     108     /* when DETS chgs update SDET */

/* Specific to datamove records */
#define SPC_DM_INAM     109     /* when input channel name changes */
#define SPC_DM_ONAM     110     /* when output channel name changes */

/* Specific to those records with a CNST field */
#define SPC_CNST	111	/* field CNST has been modified */

/* Specific for PID records */
#define SPC_PID_SPAN	112	/* if EGUF or EGUL changes */
#define SPC_PID_GAP	113	/* if setpoint gap  changes */
#define SPC_PID_SPTC	114	/* if the setpoitncrossing changes */
#define SPC_PID_CT      115	/* current clock ticks */
#define SPC_LL_TIM0	116	/* leadlag clock ticks */

/* special on ramp records */
#define SPC_RAMP_LTIM	117	/* ramp clock ticks */

/* speicals on time delay record */
#define SPC_TD_CTCK	118	/* time delay clock ticks */
#define SPC_TD_CBMD	119	/* time delay callback mode */

/* specials on accumulator fields */
#define SPC_ACC_RTOT	120	/* running totals */
#define SPC_ACC_RAVG	121	/* running average */
#define SPC_ACC_RMIN	122	/* running minimum */
#define SPC_ACC_RMAX	123	/* running maximum */
#define SPC_ACC_RSMP	124	/* running samples */
#define SPC_ACC_LTCK	125	/* last tick count field */
#define SPC_ACC_NFTM	126	/* special to mod next freeze time */
#define	SPC_ACC_FSEC	127	/* special on freeze seconds */
#define	SPC_ACC_RSEC	128	/* last reset seconds */

/* specials on udb (user defined block) fields */
#define	SPC_UDB_CTCK	129	/* current clock ticks */
#define	SPC_UDB_ZCMF	130	/* zero cmd flds in slave */

/* special for the PID record */
#define	SPC_PID_ASTP	131	/* setpoint field */

/* special for the PTO record */
#define	SPC_PTO_RVAL	132	/* RVAL field */

/* specials for te PAT record */
#define	SPC_PAT_CT	133	/* clock tick field */
#define	SPC_PAT_RVAL	134	/* on/off time field */

/* specials for ao, bo, mbbo, stringout */
#define	SPC_AO_VAL	135	/* if VAL chg's will set ao.lovl */
#define SPC_BO_VAL	136	/* if VAL chg's will set bo.lovl */
#define	SPC_MBBO_VAL	137	/* if VAL chg's will set mbbo.lovl */
#define SPC_SO_VAL	138	/* if VAL chg's will set so.lovl */
#define SPC_BO_CBFL	139	/* call back flag fo BO */

/* special for a pump */
#define	SPC_PUMP_TICKS	140	/* if TTIM or ACRT chg, set TCKS */

/* specials for the Digital Drum Controller block */
#define	SPC_DDC_VAL	141	/* special on the VAL field */
#define	SPC_DDC_LTCK	142	/* special on the LTCK field */
#define	SPC_DDC_TSXX	143	/* special on the TSxx time string fields */

/* special for PAT block */
#define	SPC_PAT_HAND	144

/* special for Lead Follow blocks */
#define	SPC_LDFW_CTCK	145	/* for all fields that deal with clock ticks */
#endif
