/* task_params.h  */
/* share/epicsH @(#)task_params.h	1.8     10/24/91 */

/*  Parameters for tasks on IOC  */
/*
 *      Authors:         Andy Kozubal, Jeff Hill, and Bob Dalesio
 *      Date:            2-24-89
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	07-23-91	ges	Add time-stamp task.
 * .02	09-12-91	joh	stack sizes increased for v5 vxWorks.
 * .03	10-24-91	lrd	Increased stack sizes for scan tasks
 * .04  12-12-91        joh     Increased stack size for the
 *                              wfDoneTask 
 * .05	05-22-92	lrd	added the allen-bradley binary input Change of State scanner
 * .06	04-27-94	ems	chg DATAMOVE stack from 2048 to 4096
 * .07	09-20-94	ems	chg periodic scan stack from 9000 to 32768
 * .08	10-11-94	ems	increase EVENT SCAN stack from 9000 to 16000
 * .09	02-01-95	ems	added DBOINIT_STACK
 * .10	02-23-95	ems	added TCNT_RESET values
 */

#ifndef INCtaskLibh 		
#include <taskLib.h>		
#endif

#define VXTASKIDSELF 0

/* Task Names */
#define IOEVENTSCAN_NAME	"ioEventScanner"
#define EVENTSCAN_NAME		"eventScanner"
#define CALLBACK_NAME          "callbackTask"
#define SMCMD_NAME		"smCommandTask"
#define SMRESP_NAME		"smResponseTask"
#define ABDONE_NAME		"abDoneTask"
#define ABSCAN_NAME		"abScanTask"
#define ABCOS_NAME		"abBiCosScanner"
#define PERIODSCAN_NAME		"periodicScanner"
#define MOMENTARY_NAME		"momentaryTask"
#define WFDONE_NAME		"wfDoneTask"
#define	SEQUENCER_NAME		"sequencer"
#define	MONITOR_NAME		"monitor"
#define	SCANNER_NAME		"scanner"	
#define CA_SRVR_NAME		"CA Server"
#define SOCKMNT_NAME		"sockMnt"
#define CHANACQ_NAME		"chanAcq"
#define WDSCAN_NAME		"wdScanTask"
#define SMIOTEST_NAME		"smInoutTest"
#define SMROTTEST_NAME		"smRotateTest"
#define EVENT_PEND_NAME		"event task"
#define	TIMESTAMP_NAME		"timeStamp"
#define	DBSERVER_NAME		"dbServer"
#define	RESMGR_NAME		"resMgr"
#define DBOINIT_NAME		"initUpdate"
#define DATAMOVE_NAME		"datamove"
#define	SNLMON_NAME		"snlMon"
#define PROCESS_STATS_NAME	"print_process_stats"
#define	DBEMULATOR_NAME		"dbEmulator"
#define TRAVERSE_NAME		"traverseDb"
#define	TCNT_RESET_NAME		"tcnt_reset"
#define	RBTASK_NAME		"rbTask"

/* Task ID's */
/*
These will go away
*/
#define	RESERVED_TID	1	/* used by vxWorks */
#define	MONITOR_TID	2
#define	SCANNER_TID	3	/* Also uses 11-15 */
#define	SEQUENCER_TID	7

/* Task priorities */
/* Gordon Uchenick:
	Change all tasks that were at 100 or below to be above 100.
	The reason for this change was to be able to get into the
	system when a task goes into a compute loop. If a task
	has a higher priority than tNetTask (and perhaps tPortmap?)
	then you can't get in with vxgdb and see what's going on.
	I kept the relationships between the same. The old
	task priorities are commented out.
*/
#define RESMGR_PRI	99	/* resMgr higher than channel access! */
#define	TIMESTAMP_PRI	100	/*42*/	    /* Time-stamp task - interrupt */
#define	DBSERVER_PRI	101	/*39*/	    /* dbServer - just lower than all scan tasks */
#define	CALLBACK_PRI	101	/*39*/      /* callback task - generall callback task */
#define IOEVENTSCAN_PRI 102	/*40*/      /* I/O Event Scanner - Runs on I/O interrupt */
#define EVENTSCAN_PRI   103	/*41*/      /* Event Scanner - Runs on a global event */
#define SMCMD_PRI 	104	/*42*/      /* Stepper Motor Command Task - Waits for cmds */
#define SMRESP_PRI      105	/*43*/      /* Stepper Motor Resp Task - Waits for resps */
#define ABCOS_PRI       105	/*43*/      /* Allen-Bradley Binary Input COS io_event wakeup */
#define ABDONE_PRI      106	/*44*/      /* Allen-Bradley Resp Task - Interrupt Driven */
#define ABSCAN_PRI      107	/*45*/      /* Allen-Bradley Scan Task - Base Rate .1 secs */
#define MOMENTARY_PRI   108	/*48*/      /* Momentary output - posted from watchdog */
#define PERIODSCAN_PRI  109	/*49*/      /* Periodic Scanner - Base rate of .1 seconds */
#define WFDONE_PRI      109	/*49*/      /* Waveform Task - Base Rate of .1 second */
#define	SEQUENCER_PRI	110	/*50*/
#define	MONITOR_PRI	111	/*100*/	
#define DATAMOVE_PRI	112     /* datamove task priority, waits fo CA event */
#define	SCANNER_PRI	150	
#define	DBEMULATOR_PRI	150		/* dbClient emulator task for MICROSOL or 68000 IOCs */
#define CA_SRVR_PRI	181	/* Channel Access Server */
#define SOCKMNT_PRI	181	/* Socket Creation and Maintenance */
#define CHANACQ_PRI	181	/* Channel Acquisition and Control */
#define DBOINIT_PRI	200	/* Persistent init/ periodic slave update */
#define WDSCAN_PRI      200     /* Watchdog Scan Task - runs every 6 seconds */
#define	RBTASK_PRI	202	/* priority of readback task rbTask */
#define SMIOTEST_PRI    205     /* Stepper Mtr in/out test - runs every .1 sec */
#define SMROTTEST_PRI   205     /* Stepper Mtr rotate test - runs every .1 sec */
#define	SNLMON_PRI	210	/* SNL Monitor task */
#define PROCESS_STATS_PRI 215	/* print_process_stats */
#define TRAVERSE_PRI	215	/* Database traversal task */
#define	TCNT_RESET_PRI	215	/* reset bi/tcnt value	*/

/* Task delay times (tics) - 60 tics per second */
#define ABSCAN_RATE     10
#define	SEQUENCER_DELAY	12
#define	MONITOR_DELAY	15
#define	SCANNER_DELAY	15
#define WDSCAN_DELAY    360
#define CA_ONLINE_DELAY	(60*5) 	/* units are sec independent of tick rate - joh */

/* Task creation options */
#define IOEVENTSCAN_OPT	VX_FP_TASK
#define EVENTSCAN_OPT	VX_FP_TASK
#define CALLBACK_OPT	VX_FP_TASK
#define SMCMD_OPT	VX_FP_TASK
#define SMRESP_OPT	VX_FP_TASK
#define ABDONE_OPT	VX_FP_TASK
#define ABCOS_OPT	VX_FP_TASK
#define ABSCAN_OPT	VX_FP_TASK
#define MOMENTARY_OPT	VX_FP_TASK
#define PERIODSCAN_OPT  VX_FP_TASK
#define WFDONE_OPT      VX_FP_TASK
#define SEQUENCER_OPT	VX_FP_TASK | VX_STDIO
#define MONITOR_OPT	VX_FP_TASK
#define SCANNER_OPT	VX_FP_TASK
#define CA_SRVR_OPT	VX_FP_TASK
#define SOCKMNT_OPT	VX_FP_TASK
#define CHANACQ_OPT	VX_FP_TASK
#define WDSCAN_OPT	VX_FP_TASK
#define SMIOTEST_OPT	VX_FP_TASK
#define SMROTTEST_OPT	VX_FP_TASK
#define EVENT_PEND_OPT	VX_FP_TASK
#define	TIMESTAMP_OPT	VX_FP_TASK
#define DATAMOVE_OPT	VX_FP_TASK
#define	SNLMON_OPT	0
#define PROCESS_STATS_OPT VX_FP_TASK | VX_STDIO
#define TRAVERSE_OPT	VX_FP_TASK
#define	TCNT_RESET_OPT	VX_FP_TASK | VX_STDIO
#define	RBTASK_OPT	VX_FP_TASK	/* task options for the readback task rbTask */

/* Task stack sizes */
#define IOEVENTSCAN_STACK	9000
#define EVENTSCAN_STACK		32768
#define CALLBACK_STACK		6000
#define SMCMD_STACK		2000
#define SMRESP_STACK		2000
#define ABCOS_STACK		2000
#define ABDONE_STACK		2000
#define ABSCAN_STACK		2000
#define MOMENTARY_STACK		4000	/* Moved up from 1000 for Epics vs Gtacs */
#define PERIODSCAN_STACK	32768
#define WFDONE_STACK		9000
#define SEQUENCER_STACK		4096
#define MONITOR_STACK		4096
#define SCANNER_STACK		2048
#define RSP_SRVR_STACK		4096
#define RESMGR_STACK		4096
#define CA_SRVR_STACK		32768
#define SOCKMNT_STACK		8192
#define CHANACQ_STACK		20480
#define WDSCAN_STACK		2000	/* moved up by 1000 for v5 vxWorks */
#define SMIOTEST_STACK    	1000
#define SMROTTEST_STACK   	1000
#define EVENT_PEND_STACK   	(4096+2048)	/* moved up by 2K for alarmvu */
#define	TIMESTAMP_STACK		3000
#define DATAMOVE_STACK		4096
#define	SNLMON_STACK		4000
#define PROCESS_STATS_STACK	2048
#define	DBEMULATOR_STACK	10000	/* stack size for dbEmulator */
#define TRAVERSE_STACK		2048
#define	DBOINIT_STACK		8192	/* stack size for dbo_init_task */
#define	TCNT_RESET_STACK	2048	/* stack size for tcnt_reset task */
#define	RBTASK_STACK		32768	/* stack size for the readback task rbTask */

