/*drvCompumotor.h,v 1.1.1.1 1999/08/03 14:42:09 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/*
**	File:	drvCompumotor.h    	
**	Auth:	William H. Szostak   	       	
**
**	This file contains the data structures used by the Modicon Modbus   	
**	I/O driver.   	
**   	
*/

#define	TRUE	1
#define FALSE	0
#define NULLP	0
#define CR	0x0d
#define LF	0x0a

/*	number of lines of compumotor 	*/

#define	CMPMTRLINES	4

/*	number of compumotor controllers per line */

#define	NUMCMPMTRS	16

#define CMPMTRDOS	0		/* Card number for digital outputs */
#define CMPMTRDIS	1		/* Card number for digital inputs */
#define CMPMTRAIS	2		/* Card number for analog inputs */
#define CMPMTRAOS	3		/* Card number for registers */
#define	CMPMTRSIN	2		/* Card number for string inputs */
#define	CMPMTRSOT	3		/* Card number for string outputs */
#define PSEUDO		33		/* Card number for pseudo i/o */

/* Number of Groups of items */

#define NUMCMPMTRDOS	1		/* 1 group of 32 digital outputs */
#define NUMCMPMTRDIS	8		/* 8 groups of 32 digital inputs */
#define NUMCMPMTRAIS	6		/* 6 analog inputs */
#define NUMCMPMTRAOS	5		/* 5 analog output */
#define	NUMCMPMTRSIN	1		/* 1 passthru response, string inputs */
#define	NUMCMPMTRSOT	1		/* 1 string output for passthru cmds */
#define BITBLOCK        32		/* size of digital groups */

/* Number of items */

#define MINDONO		0		/* min number of digital outputs */
#define MINDINO		0		/* min number of digital inputs */
#define MINAINO		0		/* min number of analogs inputs */
#define MINAONO		0		/* min number of analogs outputs */
#define	MINSINO		0		/* min number of string inputs */
#define	MINSONO		0		/* min number of string outputs */

/* Number of items */

#define MAXDONO		NUMCMPMTRDOS * BITBLOCK	/* max number of digital outputs */
#define MAXDINO		NUMCMPMTRDIS * BITBLOCK	/* max number of digital inputs	*/
#define MAXAINO		NUMCMPMTRAIS		/* max number of analog inputs */
#define MAXAONO		NUMCMPMTRAOS		/* max number of analog outputs */
#define	MAXSINO		NUMCMPMTRSIN		/* max number of string inputs */
#define	MAXSONO		NUMCMPMTRSOT		/* max number of string outputs */


/*
**	Conversion types for analog inputs	*/
**
*/
#define	AICNVINTEGER	10
#define	AICNVDOUBLEINT	11
#define	AICNVFLOAT	12
#define	AICNVDOUBLEFLT	13
#define	AICNVASCIIFLT	14
#define	AICNVSTATUS	15
#define	AICNVTIME	16
#define	AICNVASCII_2	17
#define	AICNVASCII_4	18
#define	AICNVASCII_12	19
#define	AICNVASCII_20	20
#define	MAXCONV		20

/* The following must match the definition in recChoice.ascii */

#define NOCONV 0
#define LINEAR 1
#define ASCFLT 2


#define    WATCHDOGTIMEOUT       10 * 60	/* watchdog monitor timeout */


/*
**	Communications I/O Structures
**
*/

#define		RXSEM		0
#define		TXSEM		1
#define		MAXRXMSG	256
#define		MAXTXMSG	256


typedef	struct
{
	unsigned short		rxstate;
	unsigned short		fixedlen;
	unsigned short		rxmsglenexpected;
        unsigned short          rxmsglen;
       	unsigned short		rxix;
	SEMAPHORE		*semid;
	unsigned short		*status;
        unsigned char           rxprompt;
        unsigned char           txfunction;
       	unsigned char		rxbuf [MAXRXMSG];
} RXLINE;


typedef	struct
{
	unsigned short		protocol;
	unsigned short		txdisable;
	unsigned short		txtocnt;
	unsigned short		txhwerrcnt;
	unsigned short		txmsglen;
	unsigned short		txix;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		txbuf [MAXTXMSG];
} TXLINE;

/*
**	Communications I/O Constants
**
*/

#define	RXGOOD			1
#define	RXBAD			2
#define	RXTIMEOUT		60	/* 16 ms ticks */
#define	ACTIVITYTIME		300	/* must see activity in 5 seconds */

#define	NORESP			FALSE
#define	FIXEDRESP		TRUE
#define	ONERETRY		0x8000	/* OR'ed into line number for xfermsg */
#define	RXIGNORE		0
#define	RXADRST			1
#define	RXFNCST			2
#define	RXLENST			3
#define	RXDATAST		4

#define	TXGOOD			1
#define	TXBAD			2
#define	IO_TIMEOUT		3	/* Code returned by timer */


#define	TXTIMEOUT		60	/* 16 ms ticks: one second */

#define	MAXRETRYS		3	/* retransmission trys */

#define	MAXCFGCK		8	/* Try to Bring online each cycle */
#define	BRINGBACKCNT		4	/* Poll loops per cycle */

#define	XFERGOOD		1
#define	XFERBAD			2

#define	RESNAMELEN		16	/* length of a resource name */
#define	SLWWINDOW		900	/* # of ticks to allow slave to try */

/*
** Output Queue Management Structures for Modbus Protocol
**
*/

#define AOUTTYPE	1
#define BOUTTYPE	2
#define SOUTTYPE	3

#define OUTQSIZE	128
#define	STRINGMSGLEN	80

typedef struct
{
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  aosignal;
	short           aovalue;
	double		dblaovalue;
	unsigned short	conversion;
} AOUT, *AOUTPTR;

typedef struct
{
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  bosignal;
	long            bomask;
	long            bovalue;
} BOUT, *BOUTPTR;

typedef struct
{
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  sosignal;
	long            solen;
	unsigned char   sovalue [STRINGMSGLEN];
} SOUT, *SOUTPTR;

typedef struct
{
	unsigned short  otype;
	union
	{
		AOUT    aonode;
		BOUT    bonode;
		SOUT	sonode;
	} out;
} OUTQENT;

typedef struct
{
		unsigned short	ooutix;
		unsigned short	oinix;
		OUTQENT         outque [OUTQSIZE];
} OUTPUTQUE;

/*
**	Internal IOC Structure for Compumotor controller
**
*/
typedef	struct
{
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned long		last_read;
} CMPMTRDO;

typedef	struct
{
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned long		last_read;
} CMPMTRDI;

typedef struct
{
	unsigned short		was_requested;
	unsigned short		data_valid;
	double			ioregs;
} CMPMTRAO;

typedef struct
{
	unsigned short		was_requested;
	unsigned short		data_valid;
	double			aivals;
} CMPMTRAI;

typedef struct
{
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned char		last_response [STRINGMSGLEN];
} CMPMTRSI;

typedef struct
{
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned char		last_command [STRINGMSGLEN];
}CMPMTRSO;

typedef	struct
{
	unsigned short  was_requested;	/* 0  The scan task has accesses this RTU */
	unsigned short  is_responding;	/* 1  The RTU is responding */
	unsigned short  holdflag;	/*    The RTU data is in hold */
	unsigned short  no_dataout_msgs;/* 5  Count Output Msgs */
	unsigned short  no_datain_msgs;	/* 6  Count Data Requests */
	unsigned short  no_rxnoresp;	/* 7  Count No Responses */
	unsigned short  no_txnoresp;	/* 8  Count No Responses */
	unsigned short  no_retries;	/* 10 Count Retries */
	unsigned short  no_rxhwerrs;	/* 11 Count Comm Hardware Errors */
	unsigned short  no_txhwerrs;	/* 12 Count Comm Hardware Errors */
	unsigned short  no_offlines;	/* 14 Count Off Line Transitions */
	unsigned short  cmd_reset_stat;	/* 15 Comand to reset stats */
	unsigned short  cmd_force_offline;/* 16 Command to Force Offline */
	unsigned short  badoutcnt;	/* 17 Count Failed output commands */
        unsigned short  tscanpass;	/* Total number of scanpasses. */
        unsigned short  msgs;		/* Count of msgs which got good responses */
					/*  inc on each exit from 'good' msgxfer*/
        unsigned long   responsetime;	/* Total time (ticks) waiting for good */
					/* responses.*/
					/* accumu (exit-entry) time for 'good' */
					/* msgxfers */
	CMPMTRDO	dos [NUMCMPMTRDOS];
	CMPMTRDI	dis [NUMCMPMTRDIS];
	CMPMTRAI	ais [NUMCMPMTRAIS];
	CMPMTRAO	aos [NUMCMPMTRAOS];
	CMPMTRSI	sin [NUMCMPMTRSIN];
	CMPMTRSO	sot [NUMCMPMTRSOT];
} CMPMTR;


/* Command Results */

#define CMDDONE                 0x00     /* the command was sent and the response was error free.   */
#define CMDFAIL                 0x01     /* there were errors in sending or receiving the command.  */
#define CMDNACK                 0x02     /* response was error free, but command was not completed. */
#define CMDREJECT               0x03     /* driver rejected command due to unsupported data type.   */
#define CMDBUSY			0x05	 /* The command was sent, but the device is busy */

/* task definitions */

#define CMPMTR_STACK		8000   /* Stack size              */
#define	CMPMTR_PRIORITY		152    /* default priority        */
#define	CMPMTR_TASK_OPTIONS	8      /* Floating point          */




/*
** Mutual-Exclusion Control for diagnostic/data radio channel sharing */
**
*/

#define		MUXTIMEOUT	60 * 30	/* 16.67 ms ticks (60Hz): 30 seconds 	*/


typedef struct
{
        unsigned long   TXintcount;
        unsigned long   RXintcount;
        unsigned long   msgxfercount;
        unsigned long   startingtick;
} LINESTATS;

/* Trace Bit Definitions */

#define ALLTRAFFICTRACE      0x0001
#define MODBUSNACKTRACE      0x0002        /* MODBUS NACK errors.                               */
#define CONFIGURATIONTRACE   0x0010        /* trace configuration errors in driver calls.       */
#define RUNTIMEERRORTRACE    0x0020        /* trace run time errors in driver calls.            */
#define CALLREJECTTRACE      0x0040        /* trace non-error based driver call rejections.     */


/*	End of File:	drvCompumotor.h */
