/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/



/****************************************************************************
 *
 *  MODULE NAME            : devHyundai.c
 *  AUTHOR                 : emsteel
 *  DATE                   : 10 Nov 94
 *
 *
 * DESCRIPTION: device support for Hyundai
 *
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : 
 *  REVISION DATE          :
 *  REVISION:
 *
****************************************************************************/

/****************************************************************************
 *
 * Programmer's Notes:
 *
****************************************************************************/


#include	<vxWorks.h>
#include	<types.h>
#include  	<stdioLib.h>

#include	<alarm.h>
#include	<cvtTable.h>
#include	<dbAccess.h>
#include	<dbDefs.h>
#include	<devSup.h>
#include	<link.h>
#include	<module_types.h>
#include	<special.h>
#include	<cvnTypes.h>   /* defines for conversion types */
#include	<aiRecord.h>
#include	<aoRecord.h>
#include	<biRecord.h>
#include	<boRecord.h>
#include	<mbbiRecord.h>
#include	<mbboRecord.h>
#include	<ptoRecord.h>
#include	<tdRecord.h>

#define  IO_TYPE     HYUNDAI_IO               /* NOTE: change here */
#define  AIDRIVER    hyundai_aidriver         /* NOTE: change here */
#define  AODRIVER    hyundai_aodriver         /* NOTE: change here */
#define  AOREAD      hyundai_aoread           /* NOTE: change here */
#define  BIDRIVER    hyundai_bidriver         /* NOTE: change here */
#define  BODRIVER    hyundai_bodriver         /* NOTE: change here */
#define  BOREAD      hyundai_boread           /* NOTE: change here */

static char *io_name = "Hyundai";             /* NOTE: change here */


/* Create the dset for devAiHyundai */

static long ai_init_record();
static long read_ai();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   read_ai;
   DEVSUPFUN   special_linconv;
} devAiHyundai = {                        /* NOTE: change here */
   6,
   NULL,
   NULL,
   ai_init_record,
   NULL,
   read_ai,
   NULL};

/* Create the dset for devAoHyundai */

static long ao_init_record();
static int read_ao ();
static long write_ao();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   write_ao;
   DEVSUPFUN   special_linconv;
} devAoHyundai = {                        /* NOTE: change here */
   6,
   NULL,
   NULL,
   ao_init_record,
   NULL,
   write_ao,
   NULL};

/* Create the dset for devBiHyundai */

static long bi_init_record ();
static long read_bi ();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   read_bi;
} devBiHyundai = {                        /* NOTE: change here */
   5,
   NULL,
   NULL,
   bi_init_record,
   NULL,
   read_bi};

/* Create the dset for devBoHyundai */

static long bo_init_record ();
static long write_bo ();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   write_bo;
} devBoHyundai = {                        /* NOTE: change here */
   5,
   NULL,
   NULL,
   bo_init_record,
   NULL,
   write_bo};

/* Create the dset for devMbbiHyundai */

static long mbbi_init_record ();
static long read_mbbi ();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   read_mbbi;
} devMbbiHyundai = {                      /* NOTE: change here */
   5,
   NULL,
   NULL,
   mbbi_init_record,
   NULL,
   read_mbbi};

/* Create the dset for devMbboHyundai */

static long mbbo_init_record ();
static long write_mbbo ();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   write_mbbo;
} devMbboHyundai = {                      /* NOTE: change here */
   5,
   NULL,
   NULL,
   mbbo_init_record,
   NULL,
   write_mbbo};

/* Create the dset for devPtoHyundai */

static long pto_init_record ();
static long write_pto ();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   write_pto;
} devPtoHyundai = {                       /* NOTE: change here */
   5,
   NULL,
   NULL,
   pto_init_record,
   NULL,
   write_pto};

/* Create the dset for devTdHyundai */

static long td_init_record ();
static long write_td ();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   write_td;
} devTdHyundai = {                        /* NOTE: change here */
   5,
   NULL,
   NULL,
   td_init_record,
   NULL,
   write_td};


static long masks [] = {
   0x00000001, 0x00000002, 0x00000004, 0x00000008,
   0x00000010, 0x00000020, 0x00000040, 0x00000080,
   0x00000100, 0x00000200, 0x00000400, 0x00000800,
   0x00001000, 0x00002000, 0x00004000, 0x00008000,
   0x00010000, 0x00020000, 0x00040000, 0x00080000,
   0x00100000, 0x00200000, 0x00400000, 0x00800000,
   0x01000000, 0x02000000, 0x04000000, 0x08000000,
   0x10000000, 0x20000000, 0x40000000, 0x80000000,
};


/********************************** AI ************************************/

static long ai_init_record (pai)
   struct   aiRecord *pai;
{
   struct   tateio   *ptateio;
   char              message[128];
   unsigned short    value;
   int               status;
   union u_value {
      short    ivalue;
      long     lvalue;
      float    fvalue;
      double   dvalue;
      char     afvalue [100];
  }  uval;

	switch (pai->inp.type)
   	{

      	case (IO_TYPE):
            break;

      	default:
            strcpy (message, pai->name);
            strcat (message, ": devAi");
            strcat (message, io_name);
            strcat (message, " (init_record) Illegal INP field");
            errMessage (S_db_badField, message);
            return (S_db_badField);
   	}

      	/* call driver so that it configures card */

   	ptateio = (struct tateio *) &(pai->inp.value);
   	status = AIDRIVER (pai->name, ptateio->link, ptateio->adapter,
			ptateio->card, ptateio->signal, &uval,
			pai->linr, pai->resn, pai->dtyp);

   	if (status == 0)
   	{
      	    switch (pai->linr)
      	    {
            case (CNVINTEGER):
        	pai->val = uval.ivalue;
		break;

            case (CNVDOUBLEINT):
        	pai->val = uval.lvalue;
        	break;

	    case (CNVFLOAT):
            	pai->val = uval.fvalue;
            	break;

  	    case (CNVDOUBLEFLT):
            	pai->val = uval.dvalue;
            	break;

	    case (CNVASCIIFLT):
		status = -1;

	    case NOCONV:
	    case LINEAR:
	    case CNVSQRROOT:
	    default:
            	pai->rval = uval.ivalue;
            	break;
	    }
	}
	return (status);
}

static long read_ai (pai)
   struct aiRecord   *pai;
{
   struct   tateio   *ptateio;
   int               status;
   unsigned short    value;

   union u_value {
      short    ivalue;
      long     lvalue;
      float    fvalue;
      double   dvalue;
      char     afvalue [100];
  }  uval;

	ptateio = (struct tateio *) &(pai->inp.value);
	status = AIDRIVER (pai->name, ptateio->link, ptateio->adapter,
			ptateio->card, ptateio->signal, &uval,
			pai->linr, 0, pai->dtyp);

	/* setting status to 2 tells ai's process() not to convert value */

	if (status == 0)
	{
	    switch(pai->linr)
	    {
	    case (CNVINTEGER):
	    	pai->val = uval.ivalue;
		status = 2;
		break;

	    case (CNVDOUBLEINT):
		pai->val = uval.lvalue;
		status = 2;
		break;

	    case (CNVFLOAT):
		pai->val = uval.fvalue;
		status = 2;
		break;

	    case (CNVDOUBLEFLT):
		pai->val = uval.dvalue;
		status = 2;
		break;

	    case (CNVASCIIFLT):
		status = -1;

	    case NOCONV:
	    case LINEAR:
	    case CNVSQRROOT:
	    default:
	    	pai->rval = uval.ivalue;
		break;
	    }
	}

	if (status == -1)
	{
	    if ((pai->aenb != ALARM_DISABLED) && (pai->nsev < VALID_ALARM))
	    {
		pai->nsta = READ_ALARM;
		pai->nsev = VALID_ALARM;
	    }
	    status = 2;         /*dont convert*/
	}

	if (status == -3)
	{
	    if ((pai->aenb != ALARM_DISABLED) && (pai->nsev < VALID_ALARM))
	    {
		pai->nsta = SIGNAL_ALARM;
		pai->nsev = VALID_ALARM;
	    }
	    status = 2;         /*dont convert*/
	}
	return (status);
}


/*********************************** AO **********************************/

static long ao_init_record (pao)
   struct   aoRecord *pao;
{
   struct   tateio   *ptateio;
   char              message [128];
   short             value;

	switch (pao->out.type)
   	{

      	case (IO_TYPE):
            break;

	default:
            strcpy (message, pao->name);
            strcat (message, ": devAo");
            strcat (message, io_name);
            strcat (message, " (init_record) Illegal OUT field");
            errMessage (S_db_badField,message);
            return (S_db_badField);
	}

      	/* call driver so that it configures card */

   	return (read_ao (pao));
}

static long write_ao (pao)
   struct aoRecord   *pao;
{
   struct   tateio   *ptateio;
   int               status;
   int               value;
   double            dvalue;


	ptateio = (struct tateio *) &(pao->out.value);
   	dvalue = pao->oval;
   	value = pao->rval;
   	status = AODRIVER (pao->name,ptateio->link, ptateio->adapter,
			ptateio->card, ptateio->signal,dvalue, value,
			pao->linr, 0, pao->dtyp);
   	if(status == -1)
   	{
      	    if((pao->aenb != ALARM_DISABLED) && (pao->nsev < VALID_ALARM))
      	    {
         	pao->nsta = WRITE_ALARM;
         	pao->nsev = VALID_ALARM;
      	    }
   	}
   	else if (status == -2)
   	{
      	    status = 0;
      	    if (pao->nsev < VALID_ALARM )
      	    {
         	pao->nsta = HW_LIMIT_ALARM;
         	pao->nsev = VALID_ALARM;
      	    }
   	}

	if (status == -3)
	{
	    if ((pao->aenb != ALARM_DISABLED) && (pao->nsev < VALID_ALARM))
	    {
		pao->nsta = SIGNAL_ALARM;
		pao->nsev = VALID_ALARM;
	    }
	    status = 2;         /*dont convert*/
	}
	return (status);
}


static int read_ao (pao)
   struct   aoRecord *pao;
{
   struct   tateio   *ptateio = &pao->out.value.tateio;
   int               status;

   union u_value {
      short    ivalue;
      long     lvalue;
      float    fvalue;
      double   dvalue;
      char     afvalue [100];
   }  uval;

	/* get the value from the ao driver */

   	status = AOREAD (pao->name,ptateio->link, ptateio->adapter,
			ptateio->card, ptateio->signal, &uval,
			pao->linr, &pao->resn [0], pao->dtyp);

   	if (status == 0)
   	{
      	    switch (pao->linr)
            {
            case ASCFLT:   /* ascii float conversion */
            	pao->val = uval.fvalue;
            	break;

            case NOCONV:   /* no conversion */
            case LINEAR:   /* linear conversion */
            default:
            	pao->rval = uval.ivalue;
            	break;
      	    }
   	}
	return (status);
}

/*********************************** BI ***********************************/

static long bi_init_record (pbi)
   struct   biRecord *pbi;
{
   struct   tateio   *ptateio;
   char              message [128];
   int               status = 0;
   unsigned long     value;

	switch (pbi->inp.type)
	{
	case (IO_TYPE) :
            ptateio = (struct tateio *) &(pbi->inp.value);
            pbi->mask = masks [((ptateio->signal) % 32)];
            status = BIDRIVER (pbi->name,ptateio->link, ptateio->adapter, 
                            ptateio->card, ptateio->signal, pbi->mask, &value,
                            &pbi->resn [0], pbi->dtyp);
	    if (!status)
	    {
		pbi->rval = value;
	    }
	    break;

	default :
            strcpy (message, pbi->name);
            strcat (message, ": devBi");
            strcat (message, io_name);
            strcat (message, " (init_record) Illegal INP field");
            errMessage (S_db_badField, message);
            status = S_db_badField;
	    break;
	}
	return (status);
}

static long read_bi (pbi)
   struct   biRecord *pbi;
{
   struct   tateio   *ptateio;
   int               status;
   unsigned long     value;

	ptateio = (struct tateio *) &(pbi->inp.value);
	status = BIDRIVER (pbi->name, ptateio->link, ptateio->adapter,
			ptateio->card, ptateio->signal, pbi->mask,
			&value, 0, pbi->dtyp);
	if (status == 0)
	{
	    pbi->rval = value;
	    return (status);
	}
	if (status == -1)
	{
	    if ((pbi->aenb != ALARM_DISABLED) && (pbi->nsev < VALID_ALARM))
      	    {
        	pbi->nsta = READ_ALARM;
        	pbi->nsev = VALID_ALARM;
      	    }
      	    status = 2;
	}
	if (status == -3)
	{
	    if ((pbi->aenb != ALARM_DISABLED) && (pbi->nsev < VALID_ALARM))
	    {
		pbi->nsta = SIGNAL_ALARM;
		pbi->nsev = VALID_ALARM;
	    }
	    status = 2;         /*dont convert*/
	}
	return (status);
}


/*********************************** BO ***********************************/

static long bo_init_record (pbo)
   struct   boRecord *pbo;
{
   struct   tateio   *ptateio;
   char              message [128];
   unsigned long     value;
   int               status = 0;

	switch (pbo->out.type)
	{
      	case (IO_TYPE) :
            ptateio = (struct tateio *) &(pbo->out.value);

            /* read the value via the driver */

            pbo->mask = masks [((ptateio->signal) % 32)];
            status = BOREAD (pbo->name,ptateio->link, ptateio->adapter, 
                          ptateio->card, ptateio->signal, &value, pbo->mask, 
                          &pbo->resn [0], pbo->dtyp);
	    if (status == 0)
            {
            	pbo->rval = value;
            }
            break;

      	default :
            strcpy (message, pbo->name);
            strcat (message, ": devBo");
            strcat (message, io_name);
            strcat (message, " (init_record) Illegal OUT field");
            errMessage (S_db_badField, message);
            status = S_db_badField;
	    break;
   	}
	return (status);
}

static long write_bo (pbo)
   struct   boRecord *pbo;
{
   struct   tateio   *ptateio;
   int               status;

   	ptateio = (struct tateio *) &(pbo->out.value);
   	status = BODRIVER (pbo->name, ptateio->link, ptateio->adapter,
			ptateio->card, ptateio->signal, pbo->rval,
			pbo->mask, 0, pbo->dtyp);
   	if (status != 0)
   	{
      	    if ((pbo->aenb != ALARM_DISABLED) && (pbo->nsev < VALID_ALARM))
      	    {
         	pbo->nsta = WRITE_ALARM;
         	pbo->nsev = VALID_ALARM;
      	    }
	}
   	return (status);
}


/*********************************** MBBI **********************************/

static long mbbi_init_record (pmbbi)
   struct   mbbiRecord  *pmbbi;
{
   struct   tateio   *ptateio;
   char              message [128];
   int               status = 0;
   unsigned long     value;

   	switch (pmbbi->inp.type)
   	{
      	case (IO_TYPE) :
            ptateio = (struct tateio *) &(pmbbi->inp.value);
            pmbbi->shft = ((pmbbi->inp.value.tateio.signal) % 32);
            pmbbi->mask <<= pmbbi->shft;
            status = BIDRIVER (pmbbi->name,ptateio->link, ptateio->adapter, 
                            ptateio->card, ptateio->signal, pmbbi->mask, &value,
                            &pmbbi->resn [0], pmbbi->dtyp);
	    if (!status)
            {
	    	pmbbi->rval = value;
            }
            break;

      	default :
            strcpy (message, pmbbi->name);
            strcat (message, ": devMbbi");
            strcat (message, io_name);
            strcat (message, " (init_record) Illegal INP field");
            errMessage (S_db_badField, message);
            status = S_db_badField;
	    break;
   	}
	return (status);
}

static   long   read_mbbi (pmbbi)
   struct   mbbiRecord  *pmbbi;
{
   struct   tateio   *ptateio;
   int               status = 0;
   unsigned long     value;

   	ptateio = (struct tateio *) &(pmbbi->inp.value);
   	status = BIDRIVER (pmbbi->name,ptateio->link, ptateio->adapter, 
		      ptateio->card, ptateio->signal, pmbbi->mask,
		      &value, 0, pmbbi->dtyp);
   	if (status == 0)
   	{
      	    pmbbi->rval = value;
   	}
	if (status == -1)
	{
   	    if ((pmbbi->aenb != ALARM_DISABLED) && (pmbbi->nsev < VALID_ALARM))
      	    {
         	pmbbi->nsta = READ_ALARM;
         	pmbbi->nsev = VALID_ALARM;
      	    }
	}
	if (status == -3)
	{
	    if ((pmbbi->aenb != ALARM_DISABLED) && (pmbbi->nsev < VALID_ALARM))
	    {
		pmbbi->nsta = SIGNAL_ALARM;
		pmbbi->nsev = VALID_ALARM;
	    }
	    status = -1;         /*dont convert*/
	}
	return (status);
}


/********************************** MBBO *********************************/

static long mbbo_init_record   (pmbbo)
   struct   mbboRecord  *pmbbo;
{
   char                 message [128];
   unsigned   long      value;
   struct   tateio      *ptateio;
   int                  status = 0;

   	switch (pmbbo->out.type)
   	{
      	case (IO_TYPE) :
            ptateio = &(pmbbo->out.value.tateio);
            pmbbo->shft = ((ptateio->signal) % 32);
            pmbbo->mask <<= pmbbo->shft;
            status = BOREAD (pmbbo->name, ptateio->link, ptateio->adapter, 
                          ptateio->card, ptateio->signal, &value, pmbbo->mask,
                          &pmbbo->resn [0], pmbbo->dtyp);
            if (status == 0)
	    {
            	pmbbo->rval = value;
	    }
            break;

      	default :
            strcpy (message, pmbbo->name);
            strcat (message, ": devMbbo");
            strcat (message, io_name);
            strcat (message, " (init_record) Illegal OUT field");
            errMessage (S_db_badField, message);
            status = S_db_badField;
	    break;
   	}
   	return (status);
}

static long write_mbbo (pmbbo)
   struct   mbboRecord  *pmbbo;
{
   struct   tateio   *ptateio;
   int               status;
   unsigned long     value;

   	ptateio = &(pmbbo->out.value.tateio);
   	status = BODRIVER (pmbbo->name,ptateio->link, ptateio->adapter, 
                      ptateio->card, ptateio->signal, pmbbo->rval, 
                      pmbbo->mask, 0, pmbbo->dtyp);
   	if(status != 0)
   	{
      	    if ((pmbbo->aenb != ALARM_DISABLED) && (pmbbo->nsev < VALID_ALARM))
      	    {
         	pmbbo->nsta = WRITE_ALARM;
         	pmbbo->nsev = VALID_ALARM;
      	    }
   	}
	return (status);
}


/************************************ PTO *********************************/

static long pto_init_record (ppto, bit_no)
   struct   ptoRecord   *ppto;
   unsigned int   bit_no;
{
   struct   tateio   *ptateio;
   char              message [128];
   unsigned long     value;
   int               status = 0;

	switch (bit_no)
	{

      	case 0:
            switch (ppto->pout.type)
            {
            case (IO_TYPE) :
                ptateio = (struct tateio *) &(ppto->pout.value);

                /* read the value via the driver */

               	ppto->pmsk = masks [((ptateio->signal) % 32)];
               	status = BOREAD (ppto->name,ptateio->link, ptateio->adapter, 
                                ptateio->card, ptateio->signal, &value, 
                                ppto->pmsk, &ppto->resn [0], ppto->dtyp);
	       if (!status)
		   ppto -> rbv = value;
               	break;

               	default :
                    strcpy (message, ppto->name);
                    strcat (message, ": devPto");
               	    strcat (message, io_name);
                    strcat (message, " (init_record) Illegal POS OUT field");
                    errMessage (S_db_badField, message);
                    status = S_db_badField;
            }
	    break;

	case 1:
            switch (ppto->nout.type)
            {
	    case (IO_TYPE) :
             	ptateio = (struct tateio *) &(ppto->nout.value);

             	/* read the value via the driver */

             	ppto->nmsk = masks [((ptateio->signal) % 32)];
             	status = BOREAD (ppto->name,ptateio->link, ptateio->adapter, 
                                ptateio->card, ptateio->signal, &value, 
                                ppto->nmsk, &ppto->resn [0], ppto->dtyp);
	       if (!status)
		   ppto -> rbv = value;
             	break;

	     default :
		strcpy (message, ppto->name);
               	strcat (message, ": devPto");
               	strcat (message, io_name);
               	strcat (message, " (init_record) Illegal NEG OUT field");
               	errMessage (S_db_badField, message);
               	status = S_db_badField;
             }
	     break;

	default:
            strcpy (message, ppto->name);
            strcat (message, ": devPto");
            strcat (message, io_name);
            strcat (message, " (init_record) Illegal bit no.");
            errMessage (S_db_badField, message);
            status = S_db_badField;

   	} /* end switch bit_no */

	return (status);
}

static long write_pto (ppto, bit_no)
   struct   ptoRecord   *ppto;
   unsigned int         bit_no;
{
   struct   tateio   *ptateio;
   int               status;
   char              message [128];

   	switch (bit_no)
   	{
	case 0:
            ptateio = (struct tateio *) &(ppto->pout.value);
            status = BODRIVER (ppto->name,ptateio->link, ptateio->adapter, 
                            ptateio->card, ptateio->signal, ppto->rbv, 
                            ppto->pmsk, 0, ppto->dtyp);
            break;

      	case 1:
            ptateio = (struct tateio *) &(ppto->nout.value);
            status = BODRIVER (ppto->name,ptateio->link, ptateio->adapter, 
                            ptateio->card, ptateio->signal, ppto->rbv, 
                            ppto->nmsk, 0);
            break;

      	default:
            strcpy (message, ppto->name);
            strcat (message, ": devPto");
            strcat (message, io_name);
            strcat (message, " (write_pto) Illegal bit no.");
            errMessage (S_db_badField, message);
            status = S_db_badField;
   	}

   	if (status != 0)
   	{
      	    if ((ppto->aenb != ALARM_DISABLED) && (ppto->nsev < VALID_ALARM))
      	    {
         	ppto->nsta = WRITE_ALARM;
         	ppto->nsev = VALID_ALARM;
      	    }
   	}
	return (status);
}


/************************************ TD **********************************/

static long td_init_record (ptd)
   struct   tdRecord *ptd;
{
   struct   tateio   *ptateio;
   char              message[128];
   unsigned long     value;
   int               status = 0;

   	switch (ptd->out.type)
   	{
      	case (IO_TYPE) :
            ptateio = (struct tateio *) &(ptd->out.value);

            /* read the value via the driver */

            ptd->mask = masks [((ptateio->signal - 1) % 32)];
            status = BOREAD (ptd->name,ptateio->link, ptateio->adapter, 
                          ptateio->card, ptateio->signal, &value, ptd->mask, 
                          &ptd->resn [0], ptd->dtyp);
            if(status == 0)
            {
            	ptd->rval = value;
            }
            break;

      	default :
            strcpy (message, ptd->name);
            strcat (message, ": devTd");
            strcat (message, io_name);
            strcat (message, " (init_record) Illegal OUT field");
            errMessage (S_db_badField, message);
            status = S_db_badField;
	    break;
   	}
	return (status);
}

static long write_td (ptd)
   struct   tdRecord *ptd;
{
   struct   tateio   *ptateio;
   int               status;

   	ptateio = (struct tateio *) &(ptd->out.value);
   	status = BODRIVER (ptd->name,ptateio->link, ptateio->adapter, ptateio->card,
                      ptateio->signal, ptd->rval, ptd->mask, 0, ptd->dtyp);
   	if (status != 0)
   	{
      	    if ((ptd->aenb != ALARM_DISABLED) && (ptd -> nsev < VALID_ALARM))
      	    {
         	ptd->nsta = WRITE_ALARM;
         	ptd->nsev = VALID_ALARM;
      	    }
   	}
	return (status);
}
