/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/**********
*
*	File:	drvHyundai.h							       	*
*	This file contains the data structures used by the Hyundai
*	I/O driver.                                                            	       	*                                                                             	      	**********/
#ifndef	INCDRVHYUNDAI
#define	INCDRVHYUNDAI

#define	HHI_PORT		5800

#define TRUE	1
#define FALSE	0
#define NULLP	0
#define CR	0x0d
typedef unsigned char	UBYTE;
typedef unsigned short	USINT;


/**********
*                                                                          		*	The IOC (PCU) driver will maintain a data structure which will be updated
*	every 250 ms from the SHARED MEMORY REGION after the ICM has completed
*	a READ cycle. The driver will then update the SHARED MEMORY REGION with
*	the output values and initiate a WRITE cycle.
*
**********/

/*	Define the card types for the HHI MPM	*/

#define	HHI_EMPTY	0x00		/* Empty slot */
#define	HHI_DIM32	0x01		/* 32 point Digital Input Module */
#define	HHI_DIM16	0x02		/* 16 point Digital Input Module */
#define	HHI_DIM8	0x03		/* 8 point Digital Input Module */
#define	HHI_PIM2	0x04		/* 2 point Pulse Input Module */
#define	HHI_AIM16	0x08		/* 16 point Analog Input Module */
#define	HHI_DOM16	0x10		/* 16 point Analog Output Module */
#define	HHI_AOM8	0x18		/* 8 point Analog Ouptut Module */
#define	HHI_AOM4	0x19		/* 4 point Analog Output Module */
#define	PSEUDO		99		/* Pseudo points for stats & controls */

/*	Define the TIS card types for the HHI MPM	*/

#define	TIS_AOM8	21
#define	TIS_AOM4	22
#define	TIS_AIM16	24
#define	TIS_DOM16	32
#define	TIS_DIM32	36
#define	TIS_DIM16	37
#define	TIS_DIM8	38

#define	NUM_ICM		1		/* Number of ICM's in the backplane */
#define	NUM_MPM		4		/* Number of MPM's attached to an ICM */
#define	NUM_MPM_SLOTS	16		/* Number of slots in each ICM */
#define	NUM_ICM_SLOTS	(NUM_MPM * NUM_MPM_SLOTS)
#define	NUM_PTS_SLOT	32		/* Maximum number of points per slot */
#define	NUM_MPM_PTS	(NUM_MPM_SLOTS * NUM_PTS_SLOT)
#define	NUM_ICM_PTS	(NUM_MPM * NUM_MPM_PTS)

/*	Define the card status	*/

#define	SENSOR_GOOD	0x00
#define	SENSOR_BAD	0x01
#define	SENSOR_FAULT	0x02
#define	CARD_FAULT	0x03
#define	COMM_ERROR	0x04
#define	VME_TIMEOUT	0xff

/*	Define the Shared Memory communication constants	*/

#define	MPM_INPUT	0x01
#define	MPM_OUTPUT	0x02
#define	MPM_INPUT_R	0x81		/* Success response */
#define	MPM_OUTPUT_R	0x82		/* Success response */

/*	Define redundancy constants	*/

#define	MAXCFGCK		8	/* Try to Bring online each cycle */
#define	BRINGBACKCNT		4	/* Poll loops per cycle */

#define	ACTIVITYTIME		300	/* must see activity in 5 seconds */	
#define	RESNAMELEN		16	/* length of a resource name */
#define	SLWWINDOW		900	/* # of ticks to allow slave to try */

/*	Define the VME specific interrupt vectors for the FORCE 40	*/

#define VME_INT_GEN       ((char *)0xff803e00)
#define VME_INT_VEC_REGS  ((char *)0xff803e01)
#define FGA_ICRVME2      ((char *)0xffd00208)
#define FGA_VIRQGEN2      0x04
#define INT_VEC_FGA      0x40
#define VINT_VEC_NO      (INT_VEC_FGA + 0x38)
#define INT_ENABLE       0x08


/**********
*
*	Define message types for downloading
*
**********/

#define MPM_LOAD	0x20
#define TNO_LOAD	0x21
#define DIP_LOAD	0x22
#define DOP_LOAD	0x23
#define AIP_LOAD	0x24
#define AOP_LOAD	0x25

extern short hyundaiCards[];
extern short hyundaiCounts[];

/**********
*
*	Define the common structures shared by the IOC (PCU) and the ICM.
*
**********/

typedef struct
    {
	UBYTE		revision_no;
	UBYTE		pcm_dnld_flag;  
	UBYTE		mpm_dnld_flag [NUM_MPM];
	UBYTE		total_mpm_no;
	UBYTE		operation_flag;
	UBYTE		operation_status;
	UBYTE		backup_period;
	UBYTE		trend_period;
	UBYTE		min_except_time;
	UBYTE		max_except_time;
	UBYTE		dev_limit;
	UBYTE		main_reserve_flag;
	UBYTE		time_set;
	UBYTE		external_reset;
    } PCM_EXEC_STRUCT; 

/*  THE STRUCTURE FOR CONFIGURATION */

typedef struct
    {
	UBYTE		type;
	UBYTE		addr;
    } TYPE_CFG_STRUCT;

typedef struct
    {      
	TYPE_CFG_STRUCT	slot [NUM_MPM_SLOTS];
    } MPM_CFG_STRUCT;


typedef struct
    {
	USINT		unit_no;
	UBYTE		slot_no;
	UBYTE		point_no;
	USINT		block_no;
	UBYTE		alm_type;
	UBYTE		contact_type;
    } DI_STRUCT;

typedef struct
    {
	USINT		unit_no;
	UBYTE		slot_no;
	UBYTE		point_no;
	USINT		block_no;
    } DO_STRUCT;

typedef struct
    {
	USINT		unit_no;
	UBYTE		slot_no;
	UBYTE		point_no;
	USINT		block_no;
	UBYTE		alm_type;
	UBYTE		EU_id;
	double		zero;
	double		span;
	double		hl_deadband;
	double		h_limit;
	double		l_limit;
    } AI_STRUCT;

typedef struct
    {
	USINT		unit_no;
	UBYTE		slot_no;
	UBYTE		point_no;
	USINT		block_no;
	USINT		dummy;
	double		zero;
	double		span;   
    } AO_STRUCT;

typedef struct
    {
	int		di_tno;
	int		do_tno;
	int		ai_tno;
	int		ao_tno;
	int		pi_tno;
	int		po_pno;		
    } TNO_STRUCT;

/*  THE STRUCTURE FOR BLOCK DATA */

typedef struct
    {
	UBYTE     	type;
	UBYTE		quality;
	UBYTE		exp_count;
	UBYTE		alarm;
	USINT		pntr;
	double   	value;
	double		old_value;
} BLK_STRUCT;

typedef struct
    {
	BLK_STRUCT	b_data [NUM_ICM_PTS];
    } PCS_BLK_STRUCT;

typedef struct
    {
	USINT		block_no;
	double		value;
	UBYTE		alarm;
    } ALM_STRUCT;

typedef struct
    {
	UBYTE		func [NUM_ICM_PTS]; /* fault */
	USINT		data [NUM_ICM_PTS];
    } SLOT_DATA_STRUCT;

typedef struct
    {
	SLOT_DATA_STRUCT	slot [NUM_ICM_SLOTS];
    } MPM_DATA_STRUCT; 

typedef struct
    {
	int			access;
	UBYTE		PI_sync;
    } SM_STRUCT;

/**********
*
*	Internal IOC Structures for Hyundai IO
*
**********/

typedef	struct
    {
	unsigned char	was_requested;
	unsigned char	output_pending;
	unsigned char	redundant_update;
	unsigned char	status;
	unsigned short	value;
    } POINT_IMAGE;

typedef struct
    {
	unsigned char	card_type;		/* Hyundai code extracted from hyundaiCards[] above */
	unsigned char	card_status;
	unsigned long	dig_inputs;
    } MPM_CONFIG;
	
typedef	struct
    {
	unsigned short  was_requested;		/* 0  The scan task has accesses this RTU 	*/
	unsigned short  is_responding;		/* 1  The RTU is responding */
	unsigned short  holdflag;		/*    The RTU data is in hold */
	unsigned short  no_dataerr;		/* 9  Count Message Echo Errors */
	unsigned short  no_offlines;		/* 14 Count Off Line Transitions */
	unsigned short  cmd_reset_stat;		/* 15 Comand to reset stats */
	unsigned short  cmd_force_offline;	/* 16 Command to Force Offline 			*/
	unsigned char	mpm_status [NUM_MPM];	/* 100 MBBI */
	unsigned char 	card_0_stat [NUM_MPM];	/* 101 BI */
	unsigned char	card_1_stat [NUM_MPM];	/* 102 BI */
	unsigned char	card_2_stat [NUM_MPM];	/* 103 BI */
	unsigned char	card_3_stat [NUM_MPM];	/* 104 BI */
	unsigned char	card_4_stat [NUM_MPM];	/* 105 BI */
	unsigned char	card_5_stat [NUM_MPM];	/* 106 BI */
	unsigned char	card_6_stat [NUM_MPM];	/* 107 BI */
	unsigned char	card_7_stat [NUM_MPM];	/* 108 BI */
	unsigned char	card_8_stat [NUM_MPM];	/* 109 BI */
	unsigned char	card_9_stat [NUM_MPM];	/* 110 BI */
	unsigned char	card_10_stat [NUM_MPM];	/* 111 BI */
	unsigned char	card_11_stat [NUM_MPM];	/* 112 BI */
	unsigned char	card_12_stat [NUM_MPM]; /* 113 BI */
	unsigned char	card_13_stat [NUM_MPM]; /* 114 BI */
	unsigned char	card_14_stat [NUM_MPM]; /* 115 BI */
	unsigned char	card_15_stat [NUM_MPM]; /* 116 BI */
	unsigned char	icm_status;		/* 117 BI */
	unsigned char	partner_pcu_stat;	/* 120 MBBI */
	int		udpFd;			/* fd for UDP sockert */
	int		tcpFd;			/* fd where slave connects */
	int		redFd;			/* fd for redundant update */
	int		redTxBuffsize;		/* Current Tx buffer size for redFd */
	int		redRxBuffsize;		/* Current Rx buffer size for redFd */
	int		redRxBuffmax;		/* Has Rx buffer topped out ? */
	char		sockBuff [4096];	/* For tx/rx on UDP, tx only on TCP */
	int		sockBytes;		/* # bytes in sockBuff */
	char		tcpRxBuff [4096];	/* For rx only on TCP */
	int		tcpRxBuffBytes;		/* # bytes in tcpRxBuff */
	int		tcpRxBuffIdx;		/* Next byte to be output */
	long		lastSeqTx;		/* Last sequence number sent 118 AI*/
	long		lastSeqRx;		/* Last sequence number ack'd 119 AI*/
	long		lastSeqRxTime;		/* From TickGet () */
	char		lastStrobeRx;		/* null, 'M' or 'S' */
	long		lastStrobeRxTime;	/* From TickGet () */
	long		partnerInet;		/* 0 = unknown */
	long		icmInet;		/* Always known */
	SEM_ID		redundSem;		/* Wakes up redundancy task */
	RES_TABLE	*radr;			/* Our entry in resource table */
	unsigned char	config_changed;
	MPM_CONFIG	config [NUM_ICM_SLOTS];
	POINT_IMAGE	image [NUM_ICM_PTS];
	SM_STRUCT	sm;
	PCM_EXEC_STRUCT	exec;
	PCS_BLK_STRUCT	blk;
    }	HHI_IO;

/* task definitions */

#define HHI_STACK		8000   /* Stack size              */
#define	HHI_PRIORITY		152    /* default priority        */
#define	HHI_TASK_OPTIONS	8      /* Floating point          */


/* Message definitions for redundant updates */

#define	HHI_ACK		1
#define HHI_REDUPDT	2

typedef struct
    {
    unsigned int	ackSeqNo;
    } HHI_ACK_MSG;

typedef struct
    {
    unsigned int	ackSeqNo;
    short		blockNo;
    unsigned short	value;
    } HHI_UPDT_MSG;

#endif
/*	End of File:	drvHyundai.h */
