/*drvMapbd.h,v 1.1.1.1 1999/08/03 14:43:18 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
**	File:	drvMapbd.h							       	
**	Auth:	Walter E. Palmer                                        	       	
**	Edit:	04/20/94        WEP: based on drvModicon.h                              
**              01/17/95        WEP: add more trace bit definitions                     
**
*/

#define TRUE	1
#define FALSE	0
#define NULLP	0
#define CR	0x0d

/*	number of lines connected to individual map boards   */
#define		MAPLINES	20

/*
**	The Map board is only written to, not read from. Therefore there is no
**      data polling by the driver. However, the TIS 4000 system does read
**      communication statistics and write channel control information through
**      the use of pseudo points. Therefore the appropriate data types for both
**      the pseudo points and the output points are supported in this driver.
*/


#define IDLETASKINTERVAL 4*60                   /* perform idle task once every 4 seconds. */
                                                /* ACTIVITYTIME = 5 seconds (see below */


#define LAMPS         0                         /* Class number for normal lamp output  */
#define NUMLAMPS        512                     /* 512 map board lamps.                 */
/*#define NUMLAMPS        25                       ** 512 map board lamps.                 */
#define MINLAMPNUM      1
#define MAXLAMPNUM      (NUMLAMPS*2)-1
#define LOOPBACKLAMPNUMBER MAXLAMPNUM           /* 'spare' address to do loopback test  */

#define ALLTEST       1                         /* Class number for ALL lamp testing    */
#define NUMALLTEST      5                       /* 5 specials for testing ALL lamps     */
#define MINALLTESTNUM   0
#define MAXALLTESTNUM   4

#define ONETEST	      2                         /* Class number for single lamp testing */
#define NUMONETEST      5                       /* 5 specials for testing single lamps  */
#define MINONETESTNUM   5
#define MAXONETESTNUM   9

#define PSEUDO       33                         /* Class number for statistics i/o      */



#define MAPBOARDMSGLEN  8                       /* fixed command length and echo length */

#define WATCHDOGTIMEOUT 10*60                   /* watchdog monitor timeout.            */

/*
**	Communications I/O Structures
**
*/

#define		RXSEM		0
#define		TXSEM		1
#define		MAXRXMSG	256
#define		MAXTXMSG	256


typedef	struct {
	unsigned short		protocol;
	unsigned short		rxstate;
	unsigned short		rxtocnt;
	unsigned short		rxhwerrcnt;
	unsigned short		fixedlen;
	unsigned short		rxmsglen;
	unsigned short		rxix;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		rxbuf[MAXRXMSG];
		} 	RXLINE;


typedef	struct {
	unsigned short		protocol;
	unsigned short		txdisable;
	unsigned short		txtocnt;
	unsigned short		txhwerrcnt;
	unsigned short		txmsglen;
	unsigned short		txix;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		txbuf[MAXTXMSG];
		} 	TXLINE;
/*
**	Communications I/O Constants
**
*/

#define		RXGOOD			1
#define		RXBAD			2
#define		RXTIMEOUT		60	/* 16 ms ticks 				*/
#define		ACTIVITYTIME		300	/* must see activity in 5 seconds 	*/

/*#define		VARRESP			FALSE
#define		FIXEDRESP		TRUE
*/
#define		ONERETRY		0x8000	/* OR'ed into line number for xfermsg 	*/
#define		RXIGNORE		0
#define		RXGETDATA		1

#define		TXGOOD			1
#define		TXBAD			2

/*#define		TXTIMEOUT		1255	** 16 ms ticks: just under 21 seconds 	*/
#define		TXTIMEOUT		60	/* 16 ms ticks: 1 second 	*/

#define		IO_TIMEOUT		3	/* Code returned by timer 		*/

#define		MAXRETRYS		3	/* retransmission trys 			*/

#define		MAXCFGCK		8	/* Try to Bring online each cycle 	*/
#define		BRINGBACKCNT		4	/* Poll loops per cycle 		*/

#define		XFERGOOD		1
#define		XFERBAD			2

#define		RESNAMELEN		16	/* length of a resource name 		*/
#define		SLWWINDOW		900	/* # of ticks to allow slave to try 	*/


/*
** Output Queue Management Structures for Modbus Protocol
**
*/

#define         AOUTTYPE        1
#define         BOUTTYPE        2
#define         OUTQSIZE        128

typedef struct {
	unsigned short  line;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  aosignal;
	short           aovalue;
	double		dblaovalue;
	unsigned short	conversion;
		} AOUT, *AOUTPTR;

typedef struct {
	unsigned short  line;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  bosignal;
	long            bomask;
	long            bovalue;
		} BOUT, *BOUTPTR;

typedef struct {
		unsigned short  otype;
		union   {
			AOUT    aonode;
			BOUT    bonode;
			}out;
	       } OUTQENT;

typedef struct {
		unsigned short	ooutix;
		unsigned short	oinix;
		OUTQENT         outque[OUTQSIZE];
		} OUTPUTQUE;
/*
**	Map Board image and control structure
**
*/

typedef	struct	{
	unsigned short  was_requested;		/* 0  The scan task has accesses this RTU 	*/
	unsigned short  is_responding;		/* 1  The RTU is responding 			*/
	unsigned short  holdflag;		/*    The RTU data is in hold 			*/
                                                /* 2  Program load Status 			*/
                                                /* 3  Program Load Block Number 		*/
                                                /* 4  Program Number to Load 			*/
	unsigned short  no_dataout_msgs;	/* 5  Count Output Msgs 			*/
	unsigned short  no_datain_msgs;		/* 6  Count Data Requests 			*/
	unsigned short  no_rxnoresp;		/* 7  Count No Responses 			*/
	unsigned short  no_txnoresp;		/* 8  Count No Responses 			*/
	unsigned short  no_dataerr;		/* 9  Count Message Echo Errors 		*/
	unsigned short  no_retries;		/* 10 Count Retries 				*/
	unsigned short  no_rxhwerrs;		/* 11 Count Comm Hardware Errors 		*/
	unsigned short  no_txhwerrs;		/* 12 Count Comm Hardware Errors 		*/
/*	unsigned short  no_crcerrs;		** 13 Count CRC Errors 				*/
	unsigned short  no_offlines;		/* 14 Count Off Line Transitions		*/
	unsigned short  cmd_reset_stat;		/* 15 Comand to reset stats			*/
	unsigned short  cmd_force_offline;	/* 16 Command to Force Offline 			*/
	unsigned short  badoutcnt;		/* 17 Count Failed output commands 		*/
						/* 18 Back up Mode 1 = Auto 0 = Man		*/
						/* 19 Active Channel 1= Priamry 0 = Backup 	*/
	                                        /* 20 Program Load Error Code 			*/
	                                        /* 21 Command the PLC to Start 			*/
                                                /* 22 Command the PLC to Stop 			*/
	char            lamp[NUMLAMPS];
	}	MAPBOARD;

/* task definitions */

#define MAPBD_STACK		8000   /* Stack size              */
#define	MAPBD_PRIORITY		152    /* default priority        */
#define	MAPBD_TASK_OPTIONS	8      /* Floating point          */

/*
** Mutual-Exclusion Control for diagnostic/data radio channel sharing */
/*
*/
#define		MUXTIMEOUT	60*30	/* 16.67 ms ticks (60Hz): 30 seconds 	*/


/* Trace Bit Definitions */

#define ALLTRAFFICTRACE      0x0001
#define CONFIGURATIONTRACE   0x0010        /* trace configuration errors in driver calls */
#define RUNTIMEERRORTRACE    0x0020        /* trace run time errors in driver calls */
#define CALLREJECTTRACE      0x0040        /* trace non-error based driver call rejections */

/*	End of File:	drvMapbd.h */
