/****************************************************************/
/*                                                              */
/*  This software is the proprietary property of Tate           */
/*  Integrated Systems, L.P.  Patents and copyrights are        */
/*  pending.                                                    */
/*                                                              */
/*  This software is furnished under a license and may          */
/*  be used and copied only in accordance with the terms        */
/*  of such license and with the inclusion of the above         */
/*  notice. This software or any other copies thereof           */
/*  may not be provided or otherwise made available to          */
/*  any other person. No title to and ownership of the          */
/*  software is hereby transferred.                             */
/*                                                              */
/*  The information in this software is subject to              */
/*  change without notice and should not be construed           */
/*  as a commitment by Tate Integrated Systems L.P. or          */
/*  its third party suppliers.                                  */
/*                                                              */
/*  Tate Integrated Systems, L.P. (TIS) and its third           */
/*  party suppliers assume no responsibility for the            */
/*  use or inability to use any of its software. TIS            */
/*  software is provided "as is" without warranty of            */
/*  any kind and TIS expressly disclaims all implied            */
/*  warranties, including but not limited to the implied        */
/*  warranties of merchantability and fitness for a             */
/*  particular purpose.	                                        */
/*                                                              */
/*  Notice: Notwithstanding any other lease or license          */
/*  that may pertain to or accompany the delivery of this       */
/*  computer software, the rights of the Government             */
/*  regarding its use, reproduction, and disclosure are         */
/*  as set forth in Section 52.227-19 of the FARS               */
/*  Computer Software-Restricted Rights clause.                 */
/*                                                              */
/****************************************************************/
/*
/*  File: drvMds.c						                         
/*  Edit:  03/07/94     WEP: created based on drvmodbus.c                                
/*      :  05/05/94     WEP: change all failed msgxfer holdflag=FALSE to holdflag=TRUE.  
/*      :  06/14/94     WEP: modified register_resource to print pointname causing error.
/*      :  06/22/94     WEP: corrected performance and use of the cmd_force_offline flag.
/*      :  10/27/94     WEP: modified MAIN to pre-null plcs data structure.              
/*                         : modified allocated_plc to wait to write pointer to the plcs 
/*                         : pointer structure until the allocated memory has been       
/*                         : completely initialized.                                     
*/

#define  MAIN        mds_driver           /* NOTE: change here */
#define  AIDRIVER    mds_aidriver         /* NOTE: change here */
#define  AODRIVER    mds_aodriver         /* NOTE: change here */
#define  AOREAD      mds_aoread           /* NOTE: change here */
#define  BIDRIVER    mds_bidriver         /* NOTE: change here */
#define  BODRIVER    mds_bodriver         /* NOTE: change here */
#define  BOREAD      mds_boread           /* NOTE: change here */

#define  TRACEINFO   mdstraceinfo         /* NOTE: change here */
#define  TRACEMASK   mdstracemask         /* NOTE: change here */
#define  TRACEADD    mdstraceadd          /* NOTE: change here */
#define  TRACESUB    mdstracesub          /* NOTE: change here */

#define  STOPDRIVER  stopmds              /* NOTE: change here */
#define  STARTDRIVER startmds             /* NOTE: change here */
#define  SETDELAYS   mdssetdelays         /* NOTE: change here */
#define  TABLEINFO   mdsradio             /* NOTE: change here */
#define  LINEINFO    mdsline              /* NOTE: change here */
#define  SHOWMAP     mdsmap               /* NOTE: change here */

#include <ctype.h>
#include <vxWorks.h>
#include <sysLib.h>
#include <semLib.h>
#include <wdLib.h>
#include <rngLib.h>
#include <task_params.h>
#include <stdioLib.h>
#include <dbDefs.h>
#include <drvSup.h>
#include <module_types.h>
#include <drvMds.h>		          /* NOTE: change here */
#include <drvMds_proto.h> 		  /* NOTE: change here */
#include <restable.h>


struct	{
		long		number;
		DRVSUPFUN	report;
		DRVSUPFUN	init;
	}
	drvMds={ 2, NULL, init};

/*
/*      External Line Level Data Structures (declared in drvcommon.c)
/*
*/
extern	unsigned char	resnames[MDSLINES][RESNAMELEN];	/* array of resource name strings.		*/

extern	RES_TABLE	*prestable[MDSLINES];		/* array of pointers to resource tables 	*/

extern	unsigned short	resresponse[MDSLINES];		/* array of resource response flags. 		*/
							/* False=not active resource(slave),		*/
							/* True= active resource(master) 		*/

extern	unsigned short	line_activity[MDSLINES];	/* true/false line active flag. calculated	*/
							/* using activity_time below. 			*/
extern	unsigned short	lastactivity[MDSLINES];		/* assigned value of line_activity prior to	*/
							/* it's recalculation.   			*/

extern	unsigned short	modtrace[MDSLINES];		/* controls output of communication error log	*/
							/* messages (true/false) 			*/
extern	unsigned short	mltrace[MDSLINES];		/* controls output of certain resource related	*/
							/* log messages (level) 			*/
extern	unsigned short	modrestrace[MDSLINES];		/* controls output of MASTER/SLAVE arbitration	*/
							/* log messages (binary) 			*/

extern	SEM_ID		semids[MDSLINES][2];		/* 'new' semaphore logic. the old SEMAPHORE is	*/
							/* still used. 					*/

extern	short		linetimers[MDSLINES];		/* except on init (init_line_tabs), all		*/
							/* references are commented out		 	*/

/*extern SEM_ID		linesems[MDSLINES];		/* not used. no references in the code at all	*/
                                                	
extern	unsigned short	intermsgdelay[MDSLINES];	/* holds value passed in task spawn. used for	*/
							/* each msgxfer attempt 			*/
extern	unsigned short	pretxdelay[MDSLINES];		/* holds post RTS delay before TX. 		*/
extern	unsigned short	posttxdelay[MDSLINES];		/* holds post TX delay before dropping RTS. 	*/

extern	long		activity_time[MDSLINES];	/* holds timetick of last activity. used to	*/
							/* set line_activity flag  			*/

extern  SEM_ID		linemux[MDSLINES];		/* 1/03/94: mutual-exclusion semaphore for	*/
							/* shared resource of the data link 		*/
							/* controls access to common path shared by	*/
							/* data and diagnostic channels. 		*/
							/* originally designed for use with MDS remote	*/
							/* radio diagnostics. 				*/

extern  unsigned short  halfduplex[MDSLINES];           /* store half/full duplex line status           */

extern	unsigned short	prime_state[MDSLINES];

extern	unsigned long	release_time[IOCLINES];		/* holds tick value needed to release the	*/
							/* semaphore to the primary.			*/

extern	unsigned long	master_time[IOCLINES];		/* holds tick value needed to begin next master	*/
							/* radio poll.					*/

extern	unsigned long	remote_time[IOCLINES];		/* holds tick value needed to begin next remote	*/
							/* radio poll.					*/

extern  unsigned long	offline_time[IOCLINES];		/* holds tick value needed for next attempt to  */
							/* bring offline radios back on line.           */

extern  unsigned short  nextradio[MDSLINES];            /* holds number of next radio to be time polled */

extern  unsigned short  nextoffline[MDSLINES];          /* holds number of next radio to be checked for */
                                                        /* offline status and attempted reconnection.   */

extern	unsigned short	prime_line[MDSLINES];           /* store the primeline value from the MAIN call */

/*
/*
/*      Static MDS Radio Structures
/*
/*
*/

static	RXLINE          *rxlinetab[MDSLINES] =
			{NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,};/* array of pointers to dynamically		*/
			 				/* allocated comm controls/buffers 		*/

static	TXLINE          *txlinetab[MDSLINES] =
			{NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,};/* array of pointers to dynamically		*/
			 				/* allocated comm controls/buffers 		*/

static	OUTPUTQUE	*oq[MDSLINES] =
			{NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,
			 NULL, NULL, NULL, NULL, NULL,};/* array of pointers to dynamically		*/
			 				/* allocated  output queues and	                */
							/* their control structures 			*/



static  int		taskId [MDSLINES];

static	MDSRADIO *radios[MDSLINES][NUMMDSRADIOS+1];
					/* two dimensional array of pointers to the fixed length   	*/
					/* dynamically allocated Radio control and data structures.  	*/
					/* extra pointer is for Query structure used independently 	*/
					/* of the others. it also allows for the array index to    	*/
					/* match the mod value.                                    	*/

static	char	user[40],password[40];	/* holds id and password of the user that spawned the driver	*/
					/* task	required by host to check file read privileges on Radio	*/
					/* load files 							*/

static char    *bools[] = {
	"TRUE ",
	"FALSE"
    };

static char *tasknames[] = {
				"MDS Ch0","MDS Ch1","MDS Ch2","MDS Ch3","MDS Ch4","MDS Ch5",
				"MDS Ch6","MDS Ch7","MDS Ch8","MDS Ch9","MDS Ch10","MDS Ch11",
				"MDS Ch12","MDS Ch13","MDS Ch14","MDS Ch15","MDS Ch16","MDS Ch17",
				"MDS Ch18","MDS Ch19","MDS Ch20"
			   };


static LOADMAP *maptable[MDSLINES]; /* holds control structure for address translation map file. */


static char oneshot;/* temp variable */


static unsigned short   masterinterval[MDSLINES];       /* store master radio scanning interval (secs). */
static unsigned short   remoteinterval[MDSLINES];       /* store remote radio scanning interval (secs). */

static	unsigned short	trace[MDSLINES];		/* controls output of system error messages:	*/
							/* bit mapped controls                          */

/**********
*
*	This routine is required by iocinit in the EPICS system.
*
**********/
static long init ()
    {
	return (0);
    }
/********
/*
/*	This routine causes a one-time poll of a specific remote radio - for testing only.
/*
********/
void one ()
    {
	oneshot = TRUE;
    }

/********
/*
/*	This routine sets the poll interval values for the master & slave radios.
/*
********/
void mdstime (unsigned short line, unsigned short master, unsigned short remote)
    {
	masterinterval[line] = master;
	remoteinterval[line] = remote;
    }

/********
/*
/*	This routine enables logging of comm trace messages
/*
********/
void TRACEINFO (unsigned short line)
    {
        /* note: this information must match definitions in drvMds.h */
        printf ("\n Debug Tracing Bit Definitions for mdstraceon():-\n\n");

        printf ( "MASTERCOMMTRACE =    %d :", ((trace[line] & MASTERCOMMTRACE)?1:0));
        printf ( "(0x0001)  Communications:    Master Radio Only related messages    \n");
        printf ( "REMOTECOMMTRACE =    %d :", ((trace[line] & REMOTECOMMTRACE)?1:0));
        printf ( "(0x0002)  Communications:    Remote Radio related messages         \n");
        printf ( "GENERALCOMMTRACE =   %d :", ((trace[line] & GENERALCOMMTRACE)?1:0));
        printf ( "(0x0004)  Communications:    All communiations errors              \n");
        printf ( "MASTERPOLLTRACE =    %d :", ((trace[line] & MASTERPOLLTRACE)?1:0));
        printf ( "(0x0010)  Communications:    Beginning of master radio poll        \n");
        printf ( "REMOTEPOLLTRACE =    %d :", ((trace[line] & REMOTEPOLLTRACE)?1:0));
        printf ( "(0x0020)  Communications:    Beginning of remote radio poll        \n");
        printf ( "DEMANDACTIONTRACE =  %d :", ((trace[line] & DEMANDACTIONTRACE)?1:0));
        printf ( "(0x0040)  Communications:    All outputs and forced loopback tests \n");
        printf ( "MUTEXTRACE =         %d :", ((trace[line] & MUTEXTRACE)?1:0));
        printf ( "(0x0100)  Mutual Exclusion:  Semaphore activity                    \n");
        printf ( "RESOURCESTATETRACE = %d :", ((trace[line] & RESOURCESTATETRACE)?1:0));
        printf ( "(0x0200)  Resource State:    Master/Slave activity                 \n");
        printf ( "RESOURCEERRORTRACE = %d :", ((trace[line] & RESOURCEERRORTRACE)?1:0));
        printf ( "(0x0400)  Resource Error:    Resource not allocated                \n");
        printf ( "RADIOMAPTRACE =      %d :", ((trace[line] & RADIOMAPTRACE)?1:0));
        printf ( "(0x0800)  Radio Map:         Initialization and running details    \n");
        printf ( "OUTPURQUEUETRACE =   %d :", ((trace[line] & OUTPUTQUEUETRACE)?1:0));
        printf ( "(0x1000)  Output Queue:      Full indication                       \n");
        printf ( "CALLOCTRACE =        %d :", ((trace[line] & CALLOCTRACE)?1:0));
        printf ( "(0x2000)  Memory Allocation: Noted only (errors ALWAYS reported)   \n");
    }

/********
/*
/*	This routine writes the supplied mask to the trace mask 
/*
********/
void TRACEMASK (unsigned short line, unsigned short mask)
    {
	trace[line] = mask;
        TRACEINFO(line);
    }

/********
/*
/*	This routine ORs in the supplied mask to the trace mask
/*
********/
void TRACEADD (unsigned short line, unsigned short mask)
    {
	trace[line] = trace[line] | mask;
        TRACEINFO(line);
    }

/********
/*
/*	This routine ANDs out the supplied mask from the trace mask
/*
********/
void TRACESUB (unsigned short line, unsigned short mask)
    {
	trace[line] = trace[line] & ~mask;
        TRACEINFO(line);
    }


int STOPDRIVER (short ln)
{
	if (comm_disconnect (ln) != -1)
	{
	    taskDelete (taskId [ln]);
	    return (1);
	}
	else
	{
	    return (-1);
	}
}

int STARTDRIVER (unsigned short primeline,  /* line number of the primary/cooperating device driver.*/
                 unsigned short line,       /* the line number of this driver.                      */
                 char *mode,                /* UART initializations string (comma delimited ascii). */
                 unsigned short delay,      /* intermessage delay value, expressed in ticks.        */
                 unsigned short hdflag,     /* half duplex flag. 1=half duplex mode, 1!=full duplex */
                 unsigned short delay1,     /* half duplex post RTS assert, pre TX delay.           */
                 unsigned short delay2,     /* half duplex post TX, pre RTS drop delay.             */
                 unsigned short timer1,     /* master radio polling interval, expressed in seconds. */
                 unsigned short timer2)     /* remote radio polling interval, expressed in seconds. */
{
	if ((line < IOCLINES) && (primeline < MDSLINES) )
	{
	    taskId [line] = taskSpawn (tasknames[line],MDS_PRIORITY,MDS_TASK_OPTIONS,MDS_STACK,
			  MAIN, primeline, line, mode, delay, hdflag, delay1, delay2, timer1, timer2);
	    logMsg ("\nStarted MDS driver on line %d.", line);
	    printf (" It shares a common data path with driver on line %d.\n", primeline);
	    remCurIdGet (user,password);
	    return (1);
	}
	else
	{
	    logMsg ("\nFailed MDS driver installation. Invalid line number %d\n", line);
	    return (-1);
	}
}

/********
/*
/* Set communication logic time delays to handle half/full duplex, radio delays, etc */ 
/*
********/
void SETDELAYS (unsigned short line,
		unsigned short tickcount1,
		unsigned short tickcount2,
		unsigned short tickcount3)
    {
	intermsgdelay[line] = tickcount1;  	/* at beginning new message/response sequence.		*/
	pretxdelay[line]    = tickcount2;  	/* after RTS assert but before beginning transmission.	*/
	posttxdelay[line]   = tickcount3;  	/* after sending last character of transmission before	*/
						/* dropping RTS  */
    }


/*******										
/*      MAIN	                                                                      	
/*                                                                                     	
/*      The purpose of this task is to send diagnostic commands both periodically and   
/*      on demand to the MDS radio diagnostic channel. Information from both the        
/*      Master radio and the Remote radios will be acquired and the most recent values  
/*      will be stored. Values output to the master & remote radios cannot be read      
/*      back, as the MDS units do not preserve most recent output values.               
/*
*******/                                                                                     	
/* not called from within this driver 						      */

void MAIN (unsigned short primeline, unsigned short line, unsigned char *mode,
	   unsigned short delay, unsigned short hdflag, unsigned short delay1, unsigned short delay2,
           unsigned short timer1, unsigned short timer2)
    {
	unsigned short  mod, cfgchk, cnt, modspolled;
	unsigned short  activity, pstate, lstate, execute, command,i;
	unsigned short  prime_pstate, prime_lstate;
	long	        this_time, last_time;
	MDSRADIO        *radio;
	RXLINE          *rxp;
	TXLINE          *txp;
        unsigned short  x;
        int             temp;

/* Initialize the data structures */

        for (i = 0; i < NUMMDSRADIOS+1; i++)
        {
            radios[line][i] = NULLP;
        }

        prime_line[line] = primeline;   /* this hold the line number of the Primary/Companion driver */

	intermsgdelay[line] = delay;

	halfduplex[line]  = hdflag;
        pretxdelay[line]  = delay1;
	posttxdelay[line] = delay2;

        masterinterval[line] = timer1;
        remoteinterval[line] = timer2;

        resresponse[line] = FALSE;
        resnames[line][0] = 0;
        prestable[line] = NULL;

        trace[line] = (
/*                       MASTERCOMMTRACE     | /* Communications: Master Radio Only related messages    */
/*                       REMOTECOMMTRACE     | /* Communications: Remote Radio related messages         */
                       GENERALCOMMTRACE    | /* Communications: All communication errors              */
/*                       MASTERPOLLTRACE     | /* Communications: Beginning of master radio poll        */
                       REMOTEPOLLTRACE     | /* Communications: Beginning of remote radio poll        */
                       DEMANDACTIONTRACE   | /* Communications: All outputs and forced loopback tests */
/*                       MUTEXTRACE          | /* Mutual Exclusion: semaphore activity                  */
                       RESOURCESTATETRACE  | /* Resource State: Master/Slave activity                 */
                       RESOURCEERRORTRACE  | /* Resource State: error details                         */
/*                       RADIOMAPTRACE       | /* Radio Map: initialization and running details         */
                       OUTPUTQUEUETRACE    | /* Output Queue: Full indication                         */
                       CALLOCTRACE         | /* Memory Allocation: noted only; errors ALWAYS reported */
                       NULL);                /* (here for format/style purposes only)                 */

        activity_time[line] = tickGet ();
        lastactivity[line] = line_activity[line] = FALSE;

        nextradio[line] = FIRSTRADIOADDRESS;   /* start at beginning of poll sequence. */

/*oneshot = FALSE;
/*while (!oneshot)
/*  taskDelay(1*TICKSPERSEC);   /* wait till engineer types 'one' from terminal */
/*oneshot = FALSE;
*/
        /* output queue initialization */
        oq[line] = (OUTPUTQUE *)calloc(1,(sizeof(OUTPUTQUE)));
        if (!oq[line])
        {
                logMsg ("Not enough memory for line %d data structure (data output queue)\n", line);
                return;
        }
        outqueinit(line);


        /* communication initialization */
        rxlinetab[line] = (RXLINE *)calloc(1,(sizeof(RXLINE)));
        if (!rxlinetab[line])
        {
                logMsg ("Not enough memory for line %d data structure (comm: rx)\n",line);
                return;
        }
        txlinetab[line] = (TXLINE *)calloc(1,(sizeof(TXLINE)));
        if (!txlinetab[line])
        {
                logMsg ("Not enough memory for line %d data structure (comm: tx)\n",line);
                return;
        }
        taskLock();  /* make terminal display easier to read - avoids interleaving of other task data */
        comm_init (line, mode, &tx_int_proc, &rx_int_proc);
        clear_rts (line);
        taskUnlock();

        /* Initialize the Rx to ignore interrupts, prior to enabling */
        rxp=rxlinetab[line];
        rxp->rxstate = RXIGNORE;
        enable_rx_ints(line);

        /* give initial default values to communication timeout parameters */
        rxp->rxtimeout = RXTIMEOUT;
        txp=txlinetab[line];
        txp->txtimeout = TXTIMEOUT;



        /* make sure the Primary driver has been loaded by checking for an uninitialized semaphore. */
        if (!linemux[prime_line[line]])
        {
                logMsg("Terminate: MDS driver for line");
                printf(" %d must be loaded after Primary driver for line %d\n", line, prime_line[line]);
                return;
        }

        /* allocate maptable structure */
        maptable[line] = (LOADMAP *)calloc(1,(sizeof(LOADMAP)));
        if (!txlinetab[line])
        {
                logMsg ("Not enough memory for line %d data structure (maptable)\n",line);
                return;
        }
        /* open the address map file and build translation table */
        if (!loadmapfile(line))
        {
                logMsg("Terminate: Error on loading Address Map file for line %d MDS driver.\n", line);
                return;
        }
        /* allocate Master Radio data structure */
        if (!allocate_master(line))
        {
                logMsg("Terminate: Cannot allocate MRDS for line %d MDS driver.\n", line);
                return;				/* if not allocated, try. if can't allocate, get out. */
        }

        /* Wait for prime resource to be defined via first call to */
        /*  its 'register_resource' via DRIVER/READ call.          */

        x = 59*4; /* wait 59 sec before printing message, print there after every 60 seconds */
        while ( (resnames[prime_line[line]][0] == 0) || (prestable[prime_line[line]] == NULL) )
        {
                if ((trace[line] & RESOURCEERRORTRACE) && (x++ > (60*4)))
                {
                   logMsg ( "No Resource Defined for Primeline - %d \n", prime_line[line]);
                   x = 0;
                   if (oneshot)
                   {
                           oneshot = FALSE;
                           calloneshot();
                   }
                }
                taskDelay (TICKSPERSEC/4);
        }

        /* Wait for resource to be defined via first call to */
        /*   our 'register_resource' via DRIVER/READ call.   */

        x = 59*4; /* wait 1 sec before printing message, print there after every 60 seconds */

        while ( (resnames[line][0] == 0) || (prestable[line] == NULL) )
        {
                if ((trace[line] & RESOURCEERRORTRACE) && (x++ > (60*4)))
                {
                   logMsg ( "No Resource Defined for line - %d \n", line);
                   x = 0;
                }
                taskDelay (TICKSPERSEC/4);
        }

        /* Start out as slave */
        resClearCaAnswer(prestable[line]);		/* start quiet */
        resPutLocalState(prestable[line],RES_SLV);
        if (trace[line] & RESOURCESTATETRACE)
           logMsg ("Initialize to RES_SLV state\n");

/*        taskDelay(300); */

        activity_time[line] = tickGet();
        lastactivity[line] = line_activity[line] = FALSE;
        last_time = tickGet(); 				/* Initialize the time  */

	master_time[line] = tickGet();
	remote_time[line] = tickGet() + 1*TICKSPERSEC;

        while (TRUE)
        {
                monitor_state(line);            /* maintain state and perform master radio updates.     */
                check_demand_list(line);
                check_remote_radios(line, remoteinterval[line]);
                check_demand_list(line);
                check_offline_radios(line, remoteinterval[line]*4); /* for now, take 25% more bandwidth */
                if (oneshot)
                {
                        oneshot = FALSE;
                        calloneshot();
                }
		taskDelay(12);
        }
    } /* end of MAIN procedure */

void calloneshot()
{

        LINEINFO();
/*        AODRIVER(0,2,0,33,21,0,1,0,0); */

}
/*
/*	monitor_state:                                                                  
/*                                                                                      
/*      	This procedure returns to the caller after checking and updating the	
/*		local state of this driver. It checks the state change of the		
/*		primary driver and changes the local state appropriately. If the	
/*		current state is master, it checks to see if the Master radio needs     
/*              to be scanned.           						
/*
*******/
/* this routine only called from MAIN */

void monitor_state(unsigned short line)
{
	unsigned short  activity, pstate, lstate /*, execute, command,i */;
	unsigned short  prime_pstate, prime_lstate;
	int temp;
	int error;
        unsigned int entrytime, exittime;
        int tailflag;

	resGetStates(prestable[line] ,&lstate, &pstate);
	resGetStates(prestable[prime_line[line]], &prime_lstate, &prime_pstate);

	if (prime_lstate == RES_MSTR)
	{
		if (lstate != RES_MSTR)		/* if we were not MASTER already,	*/
						/* become MASTER now.			*/
                {
			while (pstate == RES_MSTR)
                        {
                                taskDelay(TICKSPERSEC);	/* wait 1 sec to avoid stepping on partner. */
				resGetStates(prestable[line] ,&lstate, &pstate);
			        if (trace[line] & RESOURCESTATETRACE)
				    logMsg ("Waiting for partner to leave RES_MSTR  state...\n");
                        }
                        lstate = RES_MSTR;
			resPutLocalState(prestable[line], RES_MSTR);
			if (trace[line] & RESOURCESTATETRACE)
				logMsg ("Became RES_MSTR  P= %x A= %d \n", pstate, activity);
			resSetCaAnswer(prestable[line]);/* Assume control by taking	*/
							/* the CA responsibilities.	*/
			prepresponses(line);
                        /* Make sure driver responding. driver won't issue commands if not set. */
               		resresponse[line] = TRUE;
                       }
                }
	if (prime_lstate == RES_SLV)
        {
		if (lstate != RES_SLV)		/* if we were not SLAVE already,	        */
						/* become SLAVE now.			        */
		{
                        /* Make sure driver is not responding.*/
               		resresponse[line] = FALSE;
			lstate = RES_SLV;
			resPutLocalState(prestable[line],RES_SLV);
			if (trace[line] & RESOURCESTATETRACE)
				logMsg ("Became RES_SLV  P= %x A= %d \n", pstate, activity);
			resClearCaAnswer(prestable[line]);
		}
	}
	if (prime_lstate == RES_SLVWANT)
        {
		if (lstate != RES_SLVWANT)	/* if we were not SLAVE WANT already,	        */
						/* become SLAVE WANT now.			*/
		{
                        /* don't disable the driver - let it keep responding */
                        lstate = RES_SLVWANT;
			resPutLocalState(prestable[line],RES_SLVWANT);
			if (trace[line] & RESOURCESTATETRACE)
				logMsg ("Became RES_SLVWANT  P= %x A= %d \n", pstate, activity);
			resClearCaAnswer(prestable[line]);
		}
	}
	if (prime_lstate == RES_SLVWANTCA)
        {
		if (lstate != RES_SLVWANTCA)	/* if we were not SLAVE WANT CA already,	*/
						/* become SLAVE WANT CA now.			*/
		{
                        /* don't disable the driver - let it keep responding */
			lstate = RES_SLVWANTCA;
			resPutLocalState(prestable[line],RES_SLVWANTCA);
			if (trace[line] & RESOURCESTATETRACE)
				logMsg ("Became RES_SLVWANTCA  P= %x A= %d \n", pstate, activity);
			resClearCaAnswer(prestable[line]);
		}
	}
        if (lstate == RES_MSTR)
        {
                check_master_radio(line, masterinterval[line]);
        }
} /* end of monitor_state */
/*
/*	get_the_semaphore:                                                              
/*                                                                                      
/*      	This procedure returns to the caller when the semaphore has been	
/*		acquired. This procedure also detects the state change of the		
/*		primary driver and changes the local state appropriately. If the	
/*		semaphore is not available, this procedure waits forever till it	
/*		is ready. During long waits, an error message is periodically		
/*		displayed on the IOC console.						
/*
*******/
/* this routine only called from MAIN */

void get_the_semaphore(unsigned short line)
{
	unsigned short  activity, pstate, lstate /*, execute, command,i */;
	unsigned short  prime_pstate, prime_lstate;
	int temp;
	int error;
        unsigned int entrytime, exittime;
        int tailflag;

        entrytime = tickGet();
        error = 0;
        tailflag = FALSE;
       	temp = semTake(linemux[prime_line[line]], 60);	/* try to get the semaphore, short wait */
	if (temp == ERROR)
	{
                tailflag = TRUE;
		if (trace[line] & MUTEXTRACE)
	            logMsg ("Begin Wait for Mutual-Exclusion Semaphore\n");
	}
	else
	{
		if (trace[line] & MUTEXTRACE)
	            logMsg ("Acquired Mutual-Exclusion Semaphore\n");
	}
        while(TRUE)
	{
		resGetStates(prestable[line] ,&lstate, &pstate);
		resGetStates(prestable[prime_line[line]], &prime_lstate, &prime_pstate);

                /* this is a place where we could do Master Radio poll. whether we do           */
                /* or do not have the semaphore, we are in master state. Thus, we can poll      */
                /* for master radio things. in theory, we could do the master radio output      */
                /* things here too, but as they will be fairly infrequent, why bother?          */ 
                if (lstate == RES_MSTR)
                {
                        check_master_radio(line, masterinterval[line]);
                }
		if (prime_lstate == RES_MSTR)
		{
			if (lstate != RES_MSTR)		/* if we were not MASTER already,	*/
							/* become MASTER now.			*/
                        {
				while (pstate == RES_MSTR)
                                {
                                        taskDelay(TICKSPERSEC);	/* avoid stepping on partner. */
					resGetStates(prestable[line] ,&lstate, &pstate);
				        if (trace[line] & RESOURCESTATETRACE)
					    logMsg ("Waiting for partner to leave RES_MSTR  state...\n");
                                }
                                lstate = RES_MSTR;
				resPutLocalState(prestable[line], RES_MSTR);
				if (trace[line] & RESOURCESTATETRACE)
					logMsg ("Became RES_MSTR  P= %x A= %d \n", pstate, activity);
				resSetCaAnswer(prestable[line]);/* Assume control by taking	*/
								/* the CA responsibilities.	*/
				prepresponses(line);
                                /* Make sure driver responding. driver won't issue commands if not set. */
                		resresponse[line] = TRUE;
                        }
                }
		if (prime_lstate == RES_SLV)
                {
			if (lstate != RES_SLV)		/* if we were not SLAVE already,	        */
							/* become SLAVE now.			        */
			{
                                /* Make sure driver is not responding.*/
                		resresponse[line] = FALSE;
				lstate = RES_SLV;
				resPutLocalState(prestable[line],RES_SLV);
				if (trace[line] & RESOURCESTATETRACE)
					logMsg ("Became RES_SLV  P= %x A= %d \n", pstate, activity);
				resClearCaAnswer(prestable[line]);
			}
		}
		if (prime_lstate == RES_SLVWANT)
                {
			if (lstate != RES_SLVWANT)	/* if we were not SLAVE WANT already,	        */
							/* become SLAVE WANT now.			*/
			{
                                /* don't disable the driver - let it keep responding */
                                lstate = RES_SLVWANT;
				resPutLocalState(prestable[line],RES_SLVWANT);
				if (trace[line] & RESOURCESTATETRACE)
					logMsg ("Became RES_SLVWANT  P= %x A= %d \n", pstate, activity);
				resClearCaAnswer(prestable[line]);
			}
		}
		if (prime_lstate == RES_SLVWANTCA)
                {
			if (lstate != RES_SLVWANTCA)	/* if we were not SLAVE WANT CA already,	*/
							/* become SLAVE WANT CA now.			*/
			{
                                /* don't disable the driver - let it keep responding */
				lstate = RES_SLVWANTCA;
				resPutLocalState(prestable[line],RES_SLVWANTCA);
				if (trace[line] & RESOURCESTATETRACE)
					logMsg ("Became RES_SLVWANTCA  P= %x A= %d \n", pstate, activity);
				resClearCaAnswer(prestable[line]);
			}
		}
		if (temp != ERROR)
		{
                        exittime = tickGet();
                        if (tailflag == TRUE)
                        {
		                if (trace[line] & MUTEXTRACE)
                                {
                                    logMsg("Total Wait Time for Mutual-Exclusion Semaphore");
                                    printf(" was %d ticks\n", (exittime-entrytime));
                                }
                        }
                        release_time[line] = exittime;	/* this so that we can Give it back	*/
							/* immediately if required. 		*/
                	break;				/* on leaving here, we are MASTER 	*/
		}
		temp = semTake(linemux[prime_line[line]], 2);	/* try to get the semaphore, short wait */
	}
} /* end of get_the_semaphore procedure */
/*
/*	give_the_semaphore:	This procedure immediately releases the semaphore for the	*/
/*				primary driver to use. The releasetime value is set when an 	*/
/*				actual diagnostic action takes place. if a master radio action,	*/
/*				no delay is needed. if remote radio, then it's 12 sec.		*/
/*
*******/
/* this routine only called from MAIN */

void give_the_semaphore(unsigned short line)
{
        unsigned int temp;
        int      delta;

        delta = release_time[line] - tickGet();
        if (delta > 0)
        {
                if (trace[line] & MUTEXTRACE)
                {
        	        logMsg ("Waiting for Deadtime countdown for next %d ticks ... ", delta);
                }
                while (release_time[line] > tickGet())	/* wait till 'radio channel release'    */
						        /* time out has expired.                */
	        {
		        taskDelay(10);			/* reduced inquiry rate won't affect    */
						        /* performance significantly            */
	        }
		if (trace[line] & MUTEXTRACE)
                {
	                printf (" Wait completed.\n");
                }
	}
        temp = semGive(linemux[prime_line[line]]); /* give up the semaphore 			*/

	if (temp == ERROR)
	{
	        if (trace[line] & MUTEXTRACE)
                {
		        logMsg ("Radio Diagnostics: ");
		        printf ("Mutual-Exclusion Semaphore Value Error on line %d\n",line);
                }
	}
	else
	{
		if (trace[line] & MUTEXTRACE)
	            logMsg ("Released Mutual-Exclusion Semaphore\n");
	}
} /* end of give_the_semaphore procedure */




/*
/*	check_demand_list:										
/*      	checks output que and returns if and when empty.					
/*      	if there is something on the queue it attempts to process EVERY                         
/*      	pending output, doing immediate retries. If attempts fail the request is dropped.	
/*                                                                                              	
*******/
/* this routine only called from MAIN */

static void check_demand_list(unsigned short line)
    {
	OUTQENT *optr;
	unsigned short  link, adapter, card, signal, value;
	MDSRADIO  *radio;

/* work through the entire list before exiting */

	if ((optr=ogetnextp(line)) != NULLP)
        {
                get_the_semaphore(line);        /* if can't get it right away,  will poll master radio */
        }
        else
        {
                return;
        }
        while ( optr != NULLP )
	{
	    if ((optr->otype) == AOUTTYPE)
	    {
		link = optr->out.aonode.link;
		adapter = optr->out.aonode.adapter;
		radio = radios[link][adapter];
		if (radio == NULL)
		{
			if (!allocate_radio(link, adapter))
				return; 	/* if not allocated, try. if can't allocate, get out. */
                        radio = radios[link][adapter];
		}

		signal = optr->out.aonode.aosignal;
		value = optr->out.aonode.aovalue;
		if ((!radio->cmd_force_offline) && (resresponse[link]))
		{
		        if (trace[line] & DEMANDACTIONTRACE)
			   logMsg ( "Sending Analog Output Command: Radio=%d\n", adapter);
                        if (!send_ao_commands(line, adapter, card, signal, value))
                        {
                                radio->badoutcnt++;
                        }
                }
            }
      	    if ((optr->otype) == BOUTTYPE)
	    {
		link = optr->out.bonode.link;
		adapter = optr->out.bonode.adapter;
		radio = radios[link][adapter];
		if (radio == NULL)
		{
			if (!allocate_radio(link, adapter))
				return; 	/* if not allocated, try. if can't allocate, get out. */
                        radio = radios[link][adapter];
		}

		signal = optr->out.bonode.bosignal;
		value = optr->out.bonode.bovalue;
		if ((!radio->cmd_force_offline) && (resresponse[link]))
		{
		        if (trace[line] & DEMANDACTIONTRACE)
			   logMsg ( "Sending Binary Output Command: Radio=%d\n", adapter);
                        if (!send_bo_commands(line, adapter, card, signal, value))
		        {
			        radio->badoutcnt++;
		        }
                }
	    }
      	    if ((optr->otype) == FORCEREMOTELOOPBACKTYPE)
	    {
		link = optr->out.aonode.link;
		adapter = optr->out.aonode.adapter;
		radio = radios[link][adapter];
		if (radio == NULL)
		{
			if (!allocate_radio(link, adapter))
				return; 	/* if not allocated, try. if can't allocate, get out. */
                        radio = radios[link][adapter];
		}
		if ((!radio->cmd_force_offline) && (resresponse[link]))
		{
		        if (trace[line] & DEMANDACTIONTRACE)
                        {
			        logMsg ( "Forcing Remote Diagnostic Loopback Test:");
                                printf ( " Remote Radio=%d\n", adapter);
                        }
		        process_remote_radio(line, adapter);
			if (!radio->is_responding)
			{
                                if (trace[line] & (REMOTEPOLLTRACE | DEMANDACTIONTRACE))
                                   logMsg("Remote Radio %d (%d) not responding.\n",
                                     adapter,
                                     maptable[line]->address[(int)adapter].hardware );
			        mark_not_valid(radio);
                                if (radio->retry_count++ >= FAILRETRYCOUNT)
                                {
                                     if (trace[line] & REMOTEPOLLTRACE)
                                         logMsg("Taking Remote Radio %d(%d) Offline.\n",
                                                adapter,
                                                maptable[line]->address[(int)adapter].hardware );
                                     radio->offline = TRUE;
                                     radio->retry_count = 0; /* use to count offline */
                                                             /* retry failures.      */
                                }
		        }
                }
	    }
	    oreleasefull(line);     /* release the output que entry */
            optr = ogetnextp(line); /* get value for next 'while' test */
	}
        give_the_semaphore(line);

  } /* end of check_demand_list */


/*
/*	check_master_radio:							
/*          checks to see if time to acquire new data from the master radio.    
/*          if it is time, it generates the time for the next process and then  
/*          proceeds with this process.                                          
*/
/* this routine only called from MAIN */

static unsigned int check_master_radio(unsigned short line, unsigned short timer1)
{
        MDSRADIO *master;
	long	current_time;

        master = radios[line][MASTERADDRESS];
	current_time = tickGet();
        /* note: if time came to poll, and was_requested was not true, we don't poll,   */
        /*       but, as soon as it becomes true, we will poll the master. (no waiting) */
	if ((current_time >= master_time[line]) &&
            (master->was_requested == TRUE) &&
            (!master->cmd_force_offline))   /* added 6-23-94 */
        {
		master_time[line] = current_time + ( timer1 * TICKSPERSEC );
		if (trace[line] & MASTERPOLLTRACE)
			logMsg ( "Performing Timed Scan of Master Radio\n");

		process_master_radio(line);
		/* Check for master not responding. if detected, no action to take - just say it. */
		if (!master->is_responding)
		{
		        if (trace[line] & (MASTERCOMMTRACE |
                                           REMOTECOMMTRACE |
                                           GENERALCOMMTRACE |
                                           MASTERPOLLTRACE ))
                        {
			        logMsg ("The Master Radio is not responding to diagnostic commands\n");
                        }
		}
        }
	return(0);

} /* end of check_master_radio */


/*
/*	process_master_radio:						                
/*              builds and sends command to the master radio to obtain master radio     
/*              level status information.                                               
/*
*/
/* this routine only called from check_master_radio */

static	unsigned short process_master_radio (unsigned short line)
    {
	LOADMAP         *mapinfo;
        MDSRADIO        *master;
        TXLINE          *txp;
        RXLINE          *rxp;
        unsigned short  i;
        unsigned short  masterhwaddress;

        master = radios[line][MASTERADDRESS];

	mapinfo = maptable[line];	     /* establish the pointer.   */
        masterhwaddress = mapinfo->address[MASTERADDRESS].hardware;

        /* R1->T1 */
        /* Set up pointers for comm buffers */
        txp = txlinetab[line & (~ONERETRY)];
	rxp = rxlinetab[line & (~ONERETRY)];

	/* Count number of data request messages */
	master->no_datain_msgs++;

        /* R1 Message: Request Alarm Status Information */
	/* Format the output message */
	txp->txbuf[0] =  MDSMSGSTARTBYTE;             /* MDS Message Start Byte    */
	txp->txbuf[1] =  masterhwaddress;             /* MDS Master Radio Address  */  
	txp->txbuf[2] =  REQUESTALARMSTATUS;          /* MDS Message               */
	txp->txbuf[3] =  MDSMSGENDBYTE;               /* MDS Message End Byte      */
	txp->txmsglen =  REQUESTALARMSTATUSLEN;       /* MDS Message Length        */

        /* set command-specific communication timeout values */
        txp->txtimeout = R1TIMEOUT;
        rxp->rxtimeout = T1TIMEOUT;

        /* Send out the message */
        if (mdsmsgxfer(line, master, T1LEN, NOCHECKSUM, MASTERHOLDOFFTIME, MASTERCOMMTRACE) == XFERBAD)
	{
	    master->no_offlines++;
	    master->holdflag = TRUE;
	    master->is_responding = FALSE;
	    return(-1);
	}

	/* Store the returned data */
        /* store Master Radio AI values */
        master->aismds[AN1A].last_read = (unsigned short) rxp->rxbuf[3];
        master->aismds[AN1B].last_read = (unsigned short) rxp->rxbuf[4];
        master->aismds[AN1A].data_valid = TRUE;
        master->aismds[AN1B].data_valid = TRUE;

        /* store Master Radio BI values */
	master->bismds[DIG1].last_read = (unsigned long) rxp->rxbuf[1];
	master->bismds[DIG2].last_read = (unsigned long) rxp->rxbuf[2];
	master->bismds[DIG3].last_read = (unsigned long) rxp->rxbuf[5];
	master->bismds[DIG1].data_valid = TRUE;
	master->bismds[DIG2].data_valid = TRUE;
	master->bismds[DIG3].data_valid = TRUE;


        /* R3->T2 */
	/* Reset up pointers for comm buffers */
	txp = txlinetab[line & (~ONERETRY)];
	rxp = rxlinetab[line & (~ONERETRY)];

        /* R3 Message: Request Transmitter Information */
	/* Format the output message */
	txp->txbuf[0] =  MDSMSGSTARTBYTE;       /* MDS Message Start Byte    */
	txp->txbuf[1] =  masterhwaddress;       /* MDS Master Radio Address  */  
	txp->txbuf[2] =  REQUESTTXDATA;         /* MDS Message               */
	txp->txbuf[3] =  MDSMSGENDBYTE;         /* MDS Message End Byte      */
	txp->txmsglen = REQUESTTXDATALEN;       /* MDS Message Length        */

        /* set command-specific communication timeout values */
        txp->txtimeout = R3TIMEOUT;
        rxp->rxtimeout = T2TIMEOUT;

	/* Send out the message */
        if (mdsmsgxfer(line, master, T2LEN, NOCHECKSUM, MASTERHOLDOFFTIME, MASTERCOMMTRACE) == XFERBAD)
	{
	    master->no_offlines++;
	    master->holdflag = TRUE;
	    master->is_responding = FALSE;
	    return(-1);
	}

	/* Store the returned data                      */
        /* store Master Radio AI values */
        master->aismds[AN2A].last_read = (unsigned short) rxp->rxbuf[3];
        master->aismds[AN2B].last_read = (unsigned short) rxp->rxbuf[4];
        master->aismds[AN3A].last_read = (unsigned short) rxp->rxbuf[5];
        master->aismds[AN3B].last_read = (unsigned short) rxp->rxbuf[6];
        master->aismds[AN4A].last_read = (unsigned short) rxp->rxbuf[7];
        master->aismds[AN4B].last_read = (unsigned short) rxp->rxbuf[8];
        master->aismds[AN2A].data_valid = TRUE;
        master->aismds[AN2B].data_valid = TRUE;
        master->aismds[AN3A].data_valid = TRUE;
        master->aismds[AN4B].data_valid = TRUE;
        master->aismds[AN4A].data_valid = TRUE;
        master->aismds[AN4B].data_valid = TRUE;

        /* store Master Radio BI values */
	master->bismds[DIG1].last_read = (unsigned long) rxp->rxbuf[1];
	master->bismds[DIG2].last_read = (unsigned long) rxp->rxbuf[2];
	master->bismds[DIG1].data_valid = TRUE;
	master->bismds[DIG2].data_valid = TRUE;

        /* R4->T3 */
	/* Reset the  pointers for comm buffers */
	txp = txlinetab[line & (~ONERETRY)];
	rxp = rxlinetab[line & (~ONERETRY)];

	/* Count number of data request messages */
	master->no_datain_msgs++;

        /* R4 Message: Request Receiver Information */
	/* Format the output message */
	txp->txbuf[0] =  MDSMSGSTARTBYTE;       /* MDS Message Start Byte    */
	txp->txbuf[1] =  masterhwaddress;       /* MDS Master Radio Address  */  
	txp->txbuf[2] =  REQUESTRXDATA;         /* MDS Message               */
	txp->txbuf[3] =  MDSMSGENDBYTE;         /* MDS Message End Byte      */
	txp->txmsglen = REQUESTRXDATALEN;       /* MDS Message Length        */

        /* set command-specific communication timeout values */
        txp->txtimeout = R4TIMEOUT;
        rxp->rxtimeout = T3TIMEOUT;

        /* Send out the message */
        if (mdsmsgxfer(line, master, T3LEN, NOCHECKSUM, MASTERHOLDOFFTIME, MASTERCOMMTRACE) == XFERBAD)
	{
	    master->no_offlines++;
	    master->holdflag = TRUE;
	    master->is_responding = FALSE;
	    return(-1);
	}

	/* Store the returned data                      */
        /* store Master Radio AI values */
        master->aismds[AN5A].last_read = (unsigned short) rxp->rxbuf[3];
        master->aismds[AN5B].last_read = (unsigned short) rxp->rxbuf[4];
        master->aismds[AN6A].last_read = (unsigned short) rxp->rxbuf[5];
        master->aismds[AN6B].last_read = (unsigned short) rxp->rxbuf[6];
        master->aismds[AN7A].last_read = (unsigned short) rxp->rxbuf[7];
        master->aismds[AN7B].last_read = (unsigned short) rxp->rxbuf[8];
        master->aismds[AN5A].data_valid = TRUE;
        master->aismds[AN5B].data_valid = TRUE;
        master->aismds[AN6A].data_valid = TRUE;
        master->aismds[AN6B].data_valid = TRUE;
        master->aismds[AN7A].data_valid = TRUE;
        master->aismds[AN7B].data_valid = TRUE;

        /* store Master Radio BI values */
	master->bismds[DIG1].last_read = (unsigned long) rxp->rxbuf[1];
	master->bismds[DIG2].last_read = (unsigned long) rxp->rxbuf[2];
        master->bismds[DIG1].data_valid = TRUE;
        master->bismds[DIG2].data_valid = TRUE;

        /* R9->T7 */
	/* Reset the pointers for comm buffers */
	txp = txlinetab[line & (~ONERETRY)];
	rxp = rxlinetab[line & (~ONERETRY)];

        /* R9 Message: Request Master Status Information */
	/* Format the output message */
	txp->txbuf[0] =  MDSMSGSTARTBYTE;       /* MDS Message Start Byte    */
	txp->txbuf[1] =  masterhwaddress;       /* MDS Master Radio Address  */  
	txp->txbuf[2] =  MASTERSTATUS;          /* MDS Message               */
	txp->txbuf[3] =  MDSMSGENDBYTE;         /* MDS Message End Byte      */
	txp->txmsglen =  MASTERSTATUSLEN;       /* MDS Message Length        */

        /* set command-specific communication timeout values */
        txp->txtimeout = R9TIMEOUT;
        rxp->rxtimeout = T7TIMEOUT;

	/* Send out the message */
        if (mdsmsgxfer(line, master, T7LEN, NOCHECKSUM, MASTERHOLDOFFTIME, MASTERCOMMTRACE) == XFERBAD)
	{
	    master->no_offlines++;
	    master->holdflag = TRUE;
	    master->is_responding = FALSE;
	    return(-1);
	}

	/* Store the returned data                      */
        /* store Master Radio AI values */
        master->aismds[AN2A].last_read = (unsigned short) rxp->rxbuf[3];
        master->aismds[AN2B].last_read = (unsigned short) rxp->rxbuf[4];
        master->aismds[AN1A].last_read = (unsigned short) rxp->rxbuf[5];
        master->aismds[AN1B].last_read = (unsigned short) rxp->rxbuf[6];
        master->aismds[AN15].last_read = (unsigned short) rxp->rxbuf[7];
        master->aismds[AN16].last_read = (unsigned short) rxp->rxbuf[8];
        master->aismds[AN17].last_read = (unsigned short) rxp->rxbuf[9];
        master->aismds[REV1].last_read = (unsigned short) rxp->rxbuf[13];
        master->aismds[AN2A].data_valid = TRUE;
        master->aismds[AN2B].data_valid = TRUE;
        master->aismds[AN1A].data_valid = TRUE;
        master->aismds[AN1B].data_valid = TRUE;
        master->aismds[AN15].data_valid = TRUE;
        master->aismds[AN16].data_valid = TRUE;
        master->aismds[AN17].data_valid = TRUE;
        master->aismds[REV1].data_valid = TRUE;

        /* store Master Radio BI values */
	master->bismds[DIG1].last_read = (unsigned long) rxp->rxbuf[1];
	master->bismds[DIG2].last_read = (unsigned long) rxp->rxbuf[2];
	master->bismds[DIG3].last_read = (unsigned long) rxp->rxbuf[10];
	master->bismds[DIG7].last_read = (unsigned long) rxp->rxbuf[12];
	master->bismds[DIG8].last_read = (unsigned long) rxp->rxbuf[11];
        master->bismds[DIG1].data_valid = TRUE;
        master->bismds[DIG2].data_valid = TRUE;
        master->bismds[DIG3].data_valid = TRUE;
        master->bismds[DIG7].data_valid = TRUE;
        master->bismds[DIG8].data_valid = TRUE;

        /* take radio off of hold, if put on hold after becoming master */
        master->holdflag = FALSE;

        /* when all have been done successfully. get out. */
	master->is_responding = TRUE;
/*        master->data_valid = TRUE; */
	return(0);
    }
/* end of process_master_radio */
/*
/*	check_remote_radios:										
/*      	first checks to see if it is time to perform a scheduled remote advanced loopback       
/*              diagnostic test on another remote radio. if it is,                                      
/*                         it checks all possible remote radios in the configuration. if it finds one   
/*                         'eligible' for testing, it tries to perform the test. when the test is
/*                         completed, it marks the radio status appropriately and gets out. Only one    
/*                         'eligible' radio is tested per scheduled time slot.                          
*******/
/* this routine only called from MAIN */

static void check_remote_radios(unsigned short line, unsigned short timer2)
{
	long		current_time;
	unsigned short	status;
	unsigned short	radioschecked;
	unsigned short	radiospolled;
        unsigned short  radionum;
        MDSRADIO *radio;

	current_time = tickGet();
	if (current_time >= remote_time[line])
        {
                get_the_semaphore(line);        /* if can't get it right away,  will poll master radio */

		remote_time[line] = current_time + ( timer2 * TICKSPERSEC );
		status = FALSE;
		radiospolled = 0;
		radioschecked = 0;
                /* check up to maximum possible, but once one is done, get out. */
                while ((radiospolled == 0) && (radioschecked < NUMMDSRADIOS))
		{
                        radionum = nextradio[line];
                        radio = radios[line][radionum]; 	/* shorthand pointer 			*/
		    	if (radio)				/* verify pointer validity prior to use */
		    	{
				if (radio->cmd_force_offline)
				{ 
					/* if user has turned this radio offline via pseudo commands,	*/
					/* mark it as not responding 					*/
					radio->is_responding = FALSE;
					mark_not_valid(radio);
				}
                                /* note: we still poll 'not_responding' radios.                         */
				else if (radio->was_requested) /* note: as always, only poll if asked   */
				{
					radiospolled++;
					/* later, we may check to see if radio was 'recently' polled, 	*/
					/* and if so, go poll another for which the data is older.	*/

                                        if (trace[line] & REMOTEPOLLTRACE)
                                           logMsg("Performing Timed Scan of Remote Radio %d (%d)\n",
                                                   radionum,
                                                   maptable[line]->address[(int)radionum].hardware );

					process_remote_radio(line, radionum); /* I/O done here! 	*/

					if (!radio->is_responding)
					{
                                                if (trace[line] & REMOTEPOLLTRACE)
                                                   logMsg("Remote Radio %d (%d) not responding.\n",
                                                     radionum,
                                                     maptable[line]->address[(int)radionum].hardware );
						mark_not_valid(radio);
                                                if (radio->retry_count++ >= FAILRETRYCOUNT)
                                                {
                                                   if (trace[line] & REMOTEPOLLTRACE)
                                                      logMsg("Taking Remote Radio %d(%d) Offline.\n",
                                                        radionum,
                                                        maptable[line]->address[(int)radionum].hardware );
                                                   radio->offline = TRUE;
                                                   radio->retry_count = 0; /* use to count offline */
                                                                           /* retry failures.      */
                                                }
		                       	}
				}
		        }
			radioschecked++; /* we only check one 'active' radio per time slot. */
			if (nextradio[line]++ == LASTRADIOADDRESS+1)
                           nextradio[line] = FIRSTRADIOADDRESS;
		}
                give_the_semaphore(line);

		/* Check for no one responding. if detected, no action to take - just say it. */
		if (!any_responding(line))
		{
		        if (trace[line] & (REMOTECOMMTRACE | GENERALCOMMTRACE))
                        {
			        logMsg ("None of the Remote Radios ");
			        printf ("is responding to diagnostic commands\n");
                        }
		}
/*		if (radiospolled == 0)  */
/*			taskDelay(60);  */
/*		else taskDelay(10);     */
	}
} /* end of check_remote_radio */

/*
/*	check_offline_radios:										
/*      	first checks to see if it is time to attempt to bring back on line an offline remote    
/*              radio. if it is,                                                                        
/*                         it checks all possible remote radios in the configuration. if it finds one   
/*                         'eligible' for testing (offline but not forced offline), it tries to perform 
/*                         the test. when the test is completed, it marks the radio status appropriately
/*                         and gets out. Only one 'eligible' radio is tested per scheduled time slot.   
/*   
*******/
/* this routine only called from MAIN */

static void check_offline_radios(unsigned short line, unsigned short timer3)
{
	long		current_time;
	unsigned short	status;
	unsigned short	radioschecked;
	unsigned short	radiospolled;
        MDSRADIO *radio;

	current_time = tickGet();
	if (current_time >= remote_time[line])
        {
                get_the_semaphore(line);        /* if can't get it right away,  will poll master radio */

		remote_time[line] = current_time + timer3;
		status = FALSE;
		radiospolled = 0;
		radioschecked = 0;
                /* check up to maximum possible, but once one is done, get out. */
                while ((radiospolled == 0) && (radioschecked < NUMMDSRADIOS))
		{
                        radio = radios[line][nextoffline[line]]; /* shorthand pointer                   */
		    	if (radio)				 /* verify validity prior to use.       */
		    	{
				if ((!radio->cmd_force_offline) &&
                                     (radio->was_requested) &&
                                     (radio->offline))
				{
					radiospolled++;          /* poll one 'offline' radio/timeslot.  */
                                        on_line (line, nextoffline[line]); /* I/O done here!    */
				}
		        }
			radioschecked++;
			if (nextoffline[line]++ == LASTRADIOADDRESS+1)
                           nextoffline[line] = FIRSTRADIOADDRESS;
		}
                give_the_semaphore(line);

/*		if (radiospolled == 0) */
/*			taskDelay(60); */
/*		else taskDelay(10);    */
	}
} /* end of check_offline_radios */

/*
/* process_remote_radio                                                	                        
/*           build message to perform remote diagnostics. send message. validate that message   
/*           was properly received and acted on by the remote radio. if it was, store the       
/*           returned data in the appropriate location.                                         
y/*
*******/
/* this routine called from check_remote_radio and process_demand_list  */

static	void process_remote_radio (unsigned short line, unsigned short radionum)
    {
	LOADMAP         *mapinfo;
	MDSRADIO        *radio;
	MDSRADIO        *master;
        TXLINE          *txp;
	RXLINE          *rxp;
        unsigned short   radiohwaddress;
        unsigned short   masterhwaddress;
	unsigned short   i;

	/* Translate radio's logical to hardware address */
	mapinfo = maptable[line];	/* establish the pointer.   */
        if (mapinfo->address[radionum].valid_flag != TRUE)
	{
	    return;
	}
        masterhwaddress = mapinfo->address[MASTERADDRESS].hardware;
        radiohwaddress = mapinfo->address[radionum].hardware;

        /* Setup pointers to radio data structures */
	master = radios[line][MASTERADDRESS];
	radio =  radios[line][radionum];

	/* Set up pointers for comm buffers */
	txp = txlinetab[line & (~ONERETRY)];
	rxp = rxlinetab[line & (~ONERETRY)];

	/* Count number of data request messages */
	radio->no_datain_msgs++;

        /* R6 Message: Request Remote Diagnostics Loopback Test    */

	/* Format the output message */
	txp->txbuf[0] =  MDSMSGSTARTBYTE;                  /* MDS Message Start Byte    */
	txp->txbuf[1] =  masterhwaddress;                  /* MDS Master Radio Address  */  
	txp->txbuf[2] =  REMOTEDIAGNOSTICS;
	makedtmf(radiohwaddress, &(txp->txbuf[3]));        /* MDS Special 2 DTMF Digits */
	txp->txbuf[5] =  MDSMSGSPARE;
	txp->txbuf[6] =  MDSMSGSPARE;
	txp->txbuf[7] =  MDSMSGENDBYTE;                    /* MDS Message End Byte      */

	txp->txmsglen = REMOTEDIAGNOSTICSLEN;

        /* set command-specific communication timeout values */
        txp->txtimeout = R6TIMEOUT;
        rxp->rxtimeout = T4TIMEOUT;

	/* Send out the message */
	if (mdsmsgxfer(line, radio, T4LEN, NOCHECKSUM, REMOTEHOLDOFFTIME, REMOTECOMMTRACE) == XFERBAD)
	{
	    master->no_offlines++;
	    master->holdflag = TRUE;
	    master->is_responding = FALSE;
	    radio->no_offlines++;
	    radio->holdflag = TRUE;
	    radio->is_responding = FALSE;
	    return;
	}

	/* Store the returned data                      */

        /* store Master Radio AI values */
        master->aismds[AN5A].last_read = (unsigned short) rxp->rxbuf[3];
        master->aismds[AN5B].last_read = (unsigned short) rxp->rxbuf[4];
        master->aismds[AN6A].last_read = (unsigned short) rxp->rxbuf[5];
        master->aismds[AN6B].last_read = (unsigned short) rxp->rxbuf[6];
        master->aismds[AN7A].last_read = (unsigned short) rxp->rxbuf[7];
        master->aismds[AN7B].last_read = (unsigned short) rxp->rxbuf[8];
        master->aismds[AN5A].data_valid = TRUE;
        master->aismds[AN5B].data_valid = TRUE;
        master->aismds[AN6A].data_valid = TRUE;
        master->aismds[AN6B].data_valid = TRUE;
        master->aismds[AN7A].data_valid = TRUE;
        master->aismds[AN7B].data_valid = TRUE;

        /* store Master Radio BI values */
	master->bismds[DIG1].last_read = (unsigned long) rxp->rxbuf[1];
	master->bismds[DIG2].last_read = (unsigned long) rxp->rxbuf[2];
	master->bismds[DIG5].last_read = (unsigned long) rxp->rxbuf[18];
	master->bismds[DIG6].last_read = (unsigned long) rxp->rxbuf[17];
        master->bismds[DIG1].data_valid = TRUE;
        master->bismds[DIG2].data_valid = TRUE;
        master->bismds[DIG5].data_valid = TRUE;
        master->bismds[DIG6].data_valid = TRUE;


	/* check status bits for master to remote communication errors                                 */
        /* have put this here for now so that we reference the data structure by proper name reference */
        /* this also lets us acquire the master data, which should be valid.                           */
	if (master->bismds[DIG6].last_read & 0x0F)
        {
	    radio->no_offlines++;
	    radio->holdflag = TRUE;
	    radio->is_responding = FALSE;
	    return;
	}

        /* store Remote Radio AI values */
        radio->aismds[AN8].last_read  = (unsigned short) rxp->rxbuf[9];
        radio->aismds[AN9].last_read  = (unsigned short) rxp->rxbuf[10];
        radio->aismds[AN10].last_read = (unsigned short) rxp->rxbuf[11];
        radio->aismds[AN11].last_read = (unsigned short) rxp->rxbuf[12];
        radio->aismds[AN12].last_read = (unsigned short) rxp->rxbuf[13];
        radio->aismds[AN13].last_read = (unsigned short) rxp->rxbuf[14];
        radio->aismds[AN14].last_read = (unsigned short) rxp->rxbuf[15];
        radio->aismds[AN8].data_valid = TRUE;
        radio->aismds[AN9].data_valid = TRUE;
        radio->aismds[AN10].data_valid = TRUE;
        radio->aismds[AN11].data_valid = TRUE;
        radio->aismds[AN12].data_valid = TRUE;
        radio->aismds[AN13].data_valid = TRUE;
        radio->aismds[AN14].data_valid = TRUE;

        /* store Remote Radio BI values */
	radio->bismds[DIG4].last_read = (unsigned long) rxp->rxbuf[16];
        radio->bismds[DIG4].data_valid = TRUE;

        /* take radio off of hold, if put on hold after becoming master */
        radio->holdflag = FALSE;

        /* each time that we we do the loop back, we must also get EEPOT values.          */

        /* enable the Master Diagnostics: this is a special no-response preamble header   */
        /* R10 Message: Request Remote Maintenance Diagnostics Activation                 */
        if (send_masterdiagnostics(line, ENABLEDIAGS ) != 0)
           return;             /* if command fails, report the error and get out          */
        /* R11 Message: Request Remote Maintenance Diagnostics Activation                 */
        /* R11->T9: setup for access to the remote maintenance channel                    */
        if (send_remoteactivate(line, radionum, radio) != 0)
           return;             /* if command fails, report the error and get out          */
        /* remote diagnostic channel is OPEN                                              */

        /* enable the Master Diagnostics: this is a special no-response preamble header   */
        /* R10 Message: Request Remote Maintenance Diagnostics Activation                 */
        if (send_masterdiagnostics(line, ENABLEDIAGS ) != 0)
           return;             /* if command fails, report the error and get out          */
        /* R15 Message: Request Remote Maintenance Diagnostics Operating Parameters Control */
        /* R15->T9, close the access to the remote maintenance channel                    */
        if (send_remotecontrol(line, radionum, radio, 0, CLOSECHANNEL) != 0)
           return;             /* if command fails, report the error and get out          */
        /* remote maintenance channel is CLOSED                                           */


        /* when all have been done successfully. get out. */
	radio->is_responding = TRUE;
/*        radio->data_valid = TRUE; */
	return;
    }
 /* end of process_remote_radio */

/*
/* 	send_bo_commands                                                        	        
/*           this routine converts the output signal number to the appropriate                  
/*           MDS radio message/command, builds the message, sends it, and stores the results.   
/*
*******/
/* this routine only called from process_demand_list  */

static	unsigned short send_bo_commands (unsigned short line, unsigned short radionum,
					 unsigned short card, unsigned short signal,
					 unsigned short value)
    {
	MDSRADIO *master;
	MDSRADIO *radio;
	TXLINE  *txp;
	RXLINE  *rxp;
	unsigned short  i, ix, lsignal;
	unsigned long lmask, lval, tmask ,tval;
	unsigned short mdscmd;


	/* Set up pointers for comm buffers */
	txp = txlinetab[line & (~ONERETRY)];
	rxp = rxlinetab[line & (~ONERETRY)];

        /* Setup pointer to master radio data structure */
	master =  radios[line][MASTERADDRESS];

        /* Setup pointer to radio data structure */
	radio =  radios[line][radionum];

        /* translate the signal number into a specific Radio command */
	switch (signal)
	{
	case 0:		/* Switch TX to Side A */
		mdscmd = SWITCHTXA;
		/* finish building message & send it */
		break;
	case 1:		/* Switch TX to Side B */
		mdscmd = SWITCHTXB;
		/* finish building message & send it */
		break;
	case 2:		/* Switch TX to Auto   */
		mdscmd = SWITCHTXAUTO;
		/* finish building message & send it */
		break;
	case 3:		/* Switch RX to Side A */
		mdscmd = SWITCHRXA;
		/* finish building message & send it */
		break;
	case 4:		/* Switch RX to Side B */
		mdscmd = SWITCHRXB;
		/* finish building message & send it */
		break;
	case 5:		/* Switch RX to Auto   */
		mdscmd = SWITCHRXAUTO;
		/* finish building message & send it */
		break;


	case 7:		/* Enable RX AFC*/
                value = 0;
		mdscmd = ENABLERXAFC;
                	/* finish building message & send it */
		break;
	case 8:		/* Enable TX AFC */
                value = 0;
		mdscmd = ENABLETXAFC;
		/* finish building message & send it */
		break;
	case 9:		/* Disable both TX & RX AFC(Automatic Frequency Control) */
                value = 0;
		mdscmd = DISABLEAFC;
		/* finish building message & send it */
		break;
	default:
		return (-1);
	}

	/* Set up pointers for comm buffers     */
	txp = txlinetab[line & (~ONERETRY)];
	rxp = rxlinetab[line & (~ONERETRY)];

	/* Count number of data request messages */
	radio->no_dataout_msgs++;

        if (radionum == MASTERADDRESS)
        {
                /* R2 Message: Request Remote Maintenance Diagnostics Activation    */
                /* R11->T9: setup for access to the remote maintenance channel */
                if (send_remotecontrolmaster(line, radionum, card, signal, mdscmd ) != 0)
                   return(-1);             /* if command fails, report the error and get out */
                /* when command has been sent successfully. get out. */
                radio->is_responding = TRUE;
                return(0);
        }
        else
        {
              /* enable the Master Diagnostics: this is a special no-response preamble header   */
              /* R10 Message: Request Remote Maintenance Diagnostics Activation                 */
              if (send_masterdiagnostics(line, ENABLEDIAGS ) != 0)
                 return(-1);             /* if command fails, report the error and get out      */
              /* R11 Message: Request Remote Maintenance Diagnostics Activation                 */
              /* R11->T9: setup for access to the remote maintenance channel                    */
              if (send_remoteactivate(line, radionum, radio) != 0)
                 return(-1);             /* if command fails, report the error and get out      */
              /* remote diagnostic channel is OPEN                                              */

              /* enable the Master Diagnostics: this is a special no-response preamble header   */
              /* R10 Message: Request Remote Maintenance Diagnostics Activation                 */
              if (send_masterdiagnostics(line, ENABLEDIAGS ) != 0)
                 return(-1);             /* if command fails, report the error and get out      */
              /* R15 Message: Request Remote Maintenance Diagnostics Operating Parameters Control */
              /* R15->T9, perform requested BO write to remote radio                            */
              if (send_remotecontrol(line, radionum, radio, value, mdscmd) != 0)
                 return(-1);             /* if command fails, report the error and get out      */
              /* the remote action has been accomplished                                        */

	      /* Store EEPot Settings*/
              value = 0;
	      mdscmd = STOREEEPOTS;
              /* enable the Master Diagnostics: this is a special no-response preamble header   */
              /* R10 Message: Request Remote Maintenance Diagnostics Activation                 */
              if (send_masterdiagnostics(line, ENABLEDIAGS ) != 0)
                 return(-1);             /* if command fails, report the error and get out      */
              /* R15 Message: Request Remote Maintenance Diagnostics Operating Parameters Control */
              /* R15->T9, perform requested BO write to remote radio                            */
              if (send_remotecontrol(line, radionum, radio, value, mdscmd) != 0)
                 return(-1);             /* if command fails, report the error and get out      */
              /* the remote action has been accomplished                                        */
              /* enable the Master Diagnostics: this is a special no-response preamble header   */
              /* R10 Message: Request Remote Maintenance Diagnostics Activation                 */
              if (send_masterdiagnostics(line, ENABLEDIAGS ) != 0)
                 return(-1);             /* if command fails, report the error and get out      */
              /* R15 Message: Request Remote Maintenance Diagnostics Operating Parameters Control */
              /* R15->T9, close the access to the remote maintenance channel                    */
              if (send_remotecontrol(line, radionum, radio, value, CLOSECHANNEL) != 0)
                 return(-1);             /* if command fails, report the error and get out      */
              /* remote maintenance channel is CLOSED                                           */
              /* when command has been sent successfully. get out. */
              radio->is_responding = TRUE;
              master->is_responding = TRUE;
              return(0);
        }
    }
/* end of send_bo_commands */

/******
/*                                                                                      
/*  send_remotecontrolmaster                                                            
/*          builds and sends the self-named message/command to the master radio,        
/*          checks the validity and stores the results.                                 
/******
*/
/* this routine only called from send_bo_commands  */

static unsigned short send_remotecontrolmaster(unsigned short line,
                                               unsigned short radionum,
                                               unsigned short card,
                                               unsigned short signal,
                                               unsigned char mdscmd)
{
	LOADMAP         *mapinfo;
	MDSRADIO        *master;
	TXLINE          *txp;
	RXLINE          *rxp;
        unsigned short  masterhwaddress;

        /* Setup pointer to radio data structure */
	master = radios[line][MASTERADDRESS];

	mapinfo = maptable[line];	/* establish the pointer.   */
        masterhwaddress = mapinfo->address[MASTERADDRESS].hardware;

	/* Set up pointers for comm buffers     */
	txp = txlinetab[line & (~ONERETRY)];
	rxp = rxlinetab[line & (~ONERETRY)];

        /* R2 Message: Remote Control of Master         */
	/* Format the output message */
	txp->txbuf[0] =  MDSMSGSTARTBYTE;                  /* MDS Message Start Byte    */
	txp->txbuf[1] =  masterhwaddress;                  /* MDS Master Radio Address  */
	txp->txbuf[2] =  REMOTECONTROLMASTER;
	txp->txbuf[3] =  mdscmd;
	txp->txbuf[4] =  masterhwaddress;                  /* MDS Master Radio Address  */
	txp->txbuf[5] =  REMOTECONTROLMASTER;
	txp->txbuf[6] =  mdscmd;
	txp->txbuf[7] =  masterhwaddress;                  /* MDS Master Radio Address  */
	txp->txbuf[8] =  REMOTECONTROLMASTER;
	txp->txbuf[9] =  mdscmd;
	txp->txbuf[10] = MDSMSGENDBYTE;                    /* MDS Message End Byte      */

	txp->txmsglen =  REMOTECONTROLMASTERLEN;

        /* set command-specific communication timeout values */
        txp->txtimeout = R2TIMEOUT;
        rxp->rxtimeout = T1TIMEOUT;

	/* Send out the message */
        if (mdsmsgxfer(line, master, T1LEN, NOCHECKSUM, MASTERHOLDOFFTIME, REMOTECOMMTRACE) == XFERBAD)
	{
	    master->no_offlines++;
	    master->holdflag = TRUE;
	    master->is_responding = FALSE;
	    return(-1);
	}

	/* Store the returned data                      */
        /* store Master Radio AI values */
        master->aismds[AN1A].last_read = (unsigned short) rxp->rxbuf[3];
        master->aismds[AN1B].last_read = (unsigned short) rxp->rxbuf[4];
        master->aismds[AN1A].data_valid = TRUE;
        master->aismds[AN1B].data_valid = TRUE;

        /* store Master Radio BI values */
	master->bismds[DIG1].last_read = (unsigned long) rxp->rxbuf[1];
	master->bismds[DIG2].last_read = (unsigned long) rxp->rxbuf[2];
	master->bismds[DIG3].last_read = (unsigned long) rxp->rxbuf[5];
        master->bismds[DIG1].data_valid = TRUE;
        master->bismds[DIG2].data_valid = TRUE;
        master->bismds[DIG3].data_valid = TRUE;

        /* remote diagnostic action is completed. */
        return(0);
}
/* end of send_remotecontrolmaster */


/*                                                                	                        
/* send_ao_commands                                                                             
/*           this routine converts the output signal number and value to the appropriate        
/*           MDS radio message/command, builds the message, sends it, and stores the results.   
/*                                                                                              
*/
/* this routine only called from process_demand_list  */

static	unsigned short send_ao_commands (unsigned short line, unsigned short radionum,
                                                unsigned short card, unsigned short signal,
                                                unsigned char value)
/* only called from process_demand_list  */
{
	LOADMAP         *mapinfo;
	MDSRADIO        *radio;
	MDSRADIO        *master;
	TXLINE 		*txp;
	RXLINE 		*rxp;
        unsigned short   masterhwaddress;
        unsigned char    mdscmd;


        /* Setup pointer to radio data structure */
	radio = radios[line][radionum];
	master = radios[line][MASTERADDRESS];

	mapinfo = maptable[line];	/* establish the pointer.   */
        masterhwaddress = mapinfo->address[MASTERADDRESS].hardware;

	/* Count number of data request messages */
	radio->no_dataout_msgs++;

        /* first translate the signal number into a specific Radio command and verify its validity */
        switch (signal)
	{
	        case 0:		/* Increment RF TX Power by Value 	*/
                     mdscmd = INCRFTXPOWER;
		     break;
	        case 1:		/* Decrement RF TX Power by Value 	*/
		     mdscmd = DECRFTXPOWER;
		     break;
	        case 2:		/* Increment RF Frequency by Value	*/
		     mdscmd = INCRFFREQUENCY;
		     break;
                case 3:		/* Decrement RF Frequency by Value	*/
		     mdscmd = DECRFFREQUENCY;
		     break;
                case 4:		/* Increment RF Deviation by Value	*/
		     mdscmd = INCRFDEVIATION;
		     break;
                case 5:		/* Decrement RF Deviation by Value	*/
		     mdscmd = DECRFDEVIATION;
		     break;
                default:        /* if not valid command, get out & report error */
		     return (-1);
	}
	/* Count number of data request messages */
	master->no_dataout_msgs++;


        /* enable the Master Diagnostics: this is a special no-response preamble header */
        /* R10 Message: Request Remote Maintenance Diagnostics Activation               */
        if (send_masterdiagnostics(line, ENABLEDIAGS ) != 0)
           return(-1);             /* if command fails, report the error and get out    */
        /* R11 Message: Request Remote Maintenance Diagnostics Activation               */
        /* R11->T9: setup for access to the remote maintenance channel                  */
        if (send_remoteactivate(line, radionum, radio) != 0)
           return(-1);             /* if command fails, report the error and get out    */
        /* remote diagnostic channel is OPEN                                            */

        /* enable the Master Diagnostics: this is a special no-response preamble header */
        /* R10 Message: Request Remote Maintenance Diagnostics Activation               */
        if (send_masterdiagnostics(line, ENABLEDIAGS ) != 0)
           return(-1);             /* if command fails, report the error and get out    */
        /* R15 Message: Request Remote Maintenance Diagnostics Operating Parameters Control */
        /* R15->T9, perform requested AO write to remote radio */
        if (send_remotecontrol(line, radionum, radio, value, mdscmd) != 0)
           return(-1);             /* if command fails, report the error and get out    */
        /* the remote action has been accomplished                                      */

        /* Store EEPot Settings*/
        value = 0;
        mdscmd = STOREEEPOTS;
        /* enable the Master Diagnostics: this is a special no-response preamble header */
        /* R10 Message: Request Remote Maintenance Diagnostics Activation               */
        if (send_masterdiagnostics(line, ENABLEDIAGS ) != 0)
           return(-1);             /* if command fails, report the error and get out    */
        /* R15 Message: Request Remote Maintenance Diagnostics Operating Parameters Control */
        /* R15->T9, perform requested BO write to remote radio                          */
        if (send_remotecontrol(line, radionum, radio, value, mdscmd) != 0)
           return(-1);             /* if command fails, report the error and get out    */
        /* the remote action has been accomplished                                      */

        /* enable the Master Diagnostics: this is a special no-response preamble header */
        /* R10 Message: Request Remote Maintenance Diagnostics Activation               */
        if (send_masterdiagnostics(line, ENABLEDIAGS ) != 0)
           return(-1);             /* if command fails, report the error and get out    */
        /* R15 Message: Request Remote Maintenance Diagnostics Operating Parameters Control */
        /* R15->T9, close the access to the remote maintenance channel                  */
        if (send_remotecontrol(line, radionum, radio, value, CLOSECHANNEL) != 0)
           return(-1);             /* if command fails, report the error and get out    */
        /* remote maintenance channel is CLOSED                                         */

        /* when command has been sent successfully. get out. */
        master->is_responding = TRUE;
        radio->is_responding = TRUE;
	return(0);
    }
/* end of send_ao_commands */


/******
/*                                                                                      
/*  send_masterdiagnostics                                                                  
/*          builds and sends the self-named message/command to the master radio.        
/*          as there is no response, validity cannot be checked, only a good send.      
/******
*/
/* this routine only called from send_ao_commands  */

static unsigned short send_masterdiagnostics(unsigned short line, unsigned char mdscmd)
{
	LOADMAP         *mapinfo;
	MDSRADIO        *master;
	TXLINE          *txp;
	RXLINE          *rxp;
        unsigned short  masterhwaddress;

        /* Setup pointer to master radio data structure */
	master = radios[line][MASTERADDRESS];

	mapinfo = maptable[line];	/* establish the pointer.   */
        masterhwaddress = mapinfo->address[MASTERADDRESS].hardware;

	/* Set up pointers for comm buffer     */
	txp = txlinetab[line & (~ONERETRY)];

        /* R10 Message: Master Control: Enable Master Diagnostics         */
	/* Format the output message */
	txp->txbuf[0] =  MDSMSGSTARTBYTE;                  /* MDS Message Start Byte    */
	txp->txbuf[1] =  masterhwaddress;                  /* MDS Master Radio Address  */
	txp->txbuf[2] =  MASTERCONTROL;
	txp->txbuf[3] =  mdscmd;
	txp->txbuf[4] =  masterhwaddress;                  /* MDS Master Radio Address  */
	txp->txbuf[5] =  MASTERCONTROL;
	txp->txbuf[6] =  mdscmd;
	txp->txbuf[7] =  masterhwaddress;                  /* MDS Master Radio Address  */
	txp->txbuf[8] =  MASTERCONTROL;
	txp->txbuf[9] =  mdscmd;
	txp->txbuf[10] = MDSMSGENDBYTE;                    /* MDS Message End Byte      */

	txp->txmsglen =  MASTERCONTROLLEN;

        /* set command-specific communication timeout value */
        txp->txtimeout = R10TIMEOUT;

	/* Send out the message */
        if (mdsmsgout(line, master, 0, 0, MASTERHOLDOFFTIME, REMOTECOMMTRACE) == XFERBAD)
	{
	    master->no_offlines++;
	    master->holdflag = TRUE;
	    master->is_responding = FALSE;
	    return(-1);
	}
        /* remote diagnostic action is completed. */
        return(0);
}
/* end of send_masterdiagnostics */


/*
/* send_remoteactivate                                                                        
/*           this routine builds the self-named message, sends it, and stores the results.      
/*                                                                                              
*/
/* this routine only called from send_ao_commands  */

static int send_remoteactivate(unsigned short line, unsigned short radionum, MDSRADIO *radio)
{
	LOADMAP          *mapinfo;
        MDSRADIO         *master;
        TXLINE 		 *txp;
        RXLINE 		 *rxp;
        unsigned short    radiohwaddress;
        unsigned short    masterhwaddress;

	/* Translate radio's logical to hardware address */
	mapinfo = maptable[line];	/* establish the pointer.   */
        if (mapinfo->address[radionum].valid_flag != TRUE)
	{
	    return(-1);
	}
        masterhwaddress = mapinfo->address[MASTERADDRESS].hardware;
        radiohwaddress = mapinfo->address[radionum].hardware;

        /* R11->T9: setup for access to the remote maintenance channel */
        /* Set up pointers for comm buffers     */

        txp = txlinetab[line & (~ONERETRY)];
        rxp = rxlinetab[line & (~ONERETRY)];

        /* Setup pointer to radio data structure */
        master = radios[line][MASTERADDRESS];

        /* Count number of data request messages for remote radio */
        radio->no_datain_msgs++;

        /* R11 Message: Request Remote Maintenance Diagnostics Activation    */

        /* Format the output message */
        txp->txbuf[0] =  MDSMSGSTARTBYTE;                  /* MDS Message Start Byte    */
        txp->txbuf[1] =  masterhwaddress;                  /* MDS Master Radio Address  */  
        txp->txbuf[2] =  REMOTEACTIVATE;
        txp->txbuf[3] =  MDSMSGSPARE;
        txp->txbuf[4] =  MDSMSGSPARE;
        makedtmf(radiohwaddress, &(txp->txbuf[5]));        /* MDS Special 2 DTMF Digits */
        txp->txbuf[7] =  MDSMSGSPARE;
        txp->txbuf[8] =  MDSMSGSPARE;
        makecrc(&(txp->txbuf[1]), 8, &(txp->txbuf[9]));
        txp->txbuf[11] =  MDSMSGENDBYTE;                   /* MDS Message End Byte      */

        txp->txmsglen = REMOTEACTIVATELEN;

        /* set command-specific communication timeout values */
        txp->txtimeout = R11TIMEOUT;
        rxp->rxtimeout = T9TIMEOUT;

        /* Send out the message */
        if (mdsmsgxfer(line, radio, T9LEN, NOCHECKSUM, MAINTENANCETIME, REMOTECOMMTRACE) == XFERBAD)
        {
	    master->no_offlines++;
	    master->holdflag = TRUE;
	    master->is_responding = FALSE;
	    radio->no_offlines++;
	    radio->holdflag = TRUE;
	    radio->is_responding = FALSE;
	    return(-1);
	}

	/* Store the returned data                      */
        /* store Master Radio AI values */
        master->aismds[AN5A].last_read = (unsigned short) rxp->rxbuf[3];
        master->aismds[AN5B].last_read = (unsigned short) rxp->rxbuf[4];
        master->aismds[AN6A].last_read = (unsigned short) rxp->rxbuf[5];
        master->aismds[AN6B].last_read = (unsigned short) rxp->rxbuf[6];
        master->aismds[AN7A].last_read = (unsigned short) rxp->rxbuf[7];
        master->aismds[AN7B].last_read = (unsigned short) rxp->rxbuf[8];
        master->aismds[AN5A].data_valid = TRUE;
        master->aismds[AN5B].data_valid = TRUE;
        master->aismds[AN6A].data_valid = TRUE;
        master->aismds[AN6B].data_valid = TRUE;
        master->aismds[AN7A].data_valid = TRUE;
        master->aismds[AN7B].data_valid = TRUE;

        /* store Master Radio BI values */
	master->bismds[DIG1].last_read = (unsigned long) rxp->rxbuf[1];
	master->bismds[DIG2].last_read = (unsigned long) rxp->rxbuf[2];
	master->bismds[DIG6].last_read = (unsigned long) rxp->rxbuf[17];
        master->bismds[DIG1].data_valid = TRUE;
        master->bismds[DIG2].data_valid = TRUE;
        master->bismds[DIG6].data_valid = TRUE;

	/* check status bits for master to remote communication errors                                 */
        /* have put this here for now so that we reference the data structure by proper name reference */
        /* this also lets us acquire the master data, which should be valid.                           */
	if (master->bismds[DIG6].last_read & 0x0F)
        {
	    radio->no_offlines++;
	    radio->holdflag = TRUE;
	    radio->is_responding = FALSE;
	    return(-1);
	}
        /* remote diagnostic channel is OPEN */

        /* since radio status is good, store the returned data */

        /* store Remote Radio AI values */
        radio->aismds[AN9].last_read  = (unsigned short) rxp->rxbuf[11];
        radio->aismds[AN10].last_read = (unsigned short) rxp->rxbuf[10];
        radio->aismds[AN13].last_read = (unsigned short) rxp->rxbuf[9];
        radio->aismds[AN14].last_read = (unsigned short) rxp->rxbuf[15];
        radio->aismds[EEP1].last_read = (unsigned short) rxp->rxbuf[12];
        radio->aismds[EEP2].last_read = (unsigned short) rxp->rxbuf[13];
        radio->aismds[EEP3].last_read = (unsigned short) rxp->rxbuf[14];
        radio->aismds[SPR1].last_read = (unsigned short) rxp->rxbuf[18];
        radio->aismds[SPR2].last_read = (unsigned short) rxp->rxbuf[19];
        radio->aismds[AN9].data_valid  = TRUE;
        radio->aismds[AN10].data_valid = TRUE;
        radio->aismds[AN13].data_valid = TRUE;
        radio->aismds[AN14].data_valid = TRUE;
        radio->aismds[EEP1].data_valid = TRUE;
        radio->aismds[EEP2].data_valid = TRUE;
        radio->aismds[EEP3].data_valid = TRUE;
        radio->aismds[SPR1].data_valid = TRUE;
        radio->aismds[SPR2].data_valid = TRUE;

        /* store Remote Radio BI values */
	radio->bismds[DIG4].last_read = (unsigned long) rxp->rxbuf[16];
        radio->bismds[DIG4].data_valid = TRUE;

        /* when all have been done successfully. get out. */
	radio->is_responding = TRUE;
/*	radio->data_valid = TRUE; */
        return(0);
}
/* end of send_remoteactivate */

/*
/* send_remotecontrol                                                                        
/*           this routine builds the self-named message, sends it, and stores the results.      
/*                                                                                              
*/
/* this routine only called from send_ao_commands  */

static int send_remotecontrol(unsigned short line, unsigned short radionum, MDSRADIO *radio,
                         unsigned short value, unsigned char mdscmd)
{
	LOADMAP          *mapinfo;
	MDSRADIO         *master;
	TXLINE 		 *txp;
	RXLINE 		 *rxp;
        unsigned short    radiohwaddress;
        unsigned short    masterhwaddress;

	/* Translate radio's logical to hardware address */
	mapinfo = maptable[line];	/* establish the pointer.   */
        if (mapinfo->address[radionum].valid_flag != TRUE)
	{
	    return(-1);
	}
        radiohwaddress = mapinfo->address[radionum].hardware;
        masterhwaddress = mapinfo->address[MASTERADDRESS].hardware;

	/* Set up pointers for comm buffers     */
	txp = txlinetab[line & (~ONERETRY)];
	rxp = rxlinetab[line & (~ONERETRY)];

        /* Setup pointer to radio data structure */
	master = radios[line][MASTERADDRESS];

	/* Count number of data request messages */
	radio->no_datain_msgs++;

        /* R15 Message: Request Remote Maintenance Diagnostics Operating Parameters Control      */
	/* Format the output message */
	txp->txbuf[0] =  MDSMSGSTARTBYTE;                  /* MDS Message Start Byte     */
	txp->txbuf[1] =  masterhwaddress;                  /* MDS Master Radio Address   */
	txp->txbuf[2] =  REMOTECONTROL;
	txp->txbuf[3] =  mdscmd;
	txp->txbuf[4] =  value;
	makedtmf(radiohwaddress, &(txp->txbuf[5]));         /* MDS Special 2 DTMF Digits */
	txp->txbuf[7] =  MDSMSGSPARE;
	txp->txbuf[8] =  MDSMSGSPARE;
	makecrc(&(txp->txbuf[1]), 8, &(txp->txbuf[9]));
	txp->txbuf[11] =  MDSMSGENDBYTE;                    /* MDS Message End Byte      */

	txp->txmsglen = REMOTECONTROLLEN;

        /* set command-specific communication timeout values */
        txp->txtimeout = R15TIMEOUT;
        rxp->rxtimeout = T9TIMEOUT;

	/* Send out the message */
        if (mdsmsgxfer(line, radio, T9LEN, NOCHECKSUM, MAINTENANCETIME, REMOTECOMMTRACE) == XFERBAD)
	{
	    master->no_offlines++;
	    master->holdflag = TRUE;
	    master->is_responding = FALSE;
	    radio->no_offlines++;
	    radio->holdflag = TRUE;
	    radio->is_responding = FALSE;
	    return(-1);
	}

	/* Store the returned data                      */
        /* store Master Radio AI values */
        master->aismds[AN5A].last_read = (unsigned short) rxp->rxbuf[3];
        master->aismds[AN5B].last_read = (unsigned short) rxp->rxbuf[4];
        master->aismds[AN6A].last_read = (unsigned short) rxp->rxbuf[5];
        master->aismds[AN6B].last_read = (unsigned short) rxp->rxbuf[6];
        master->aismds[AN7A].last_read = (unsigned short) rxp->rxbuf[7];
        master->aismds[AN7B].last_read = (unsigned short) rxp->rxbuf[8];
        master->aismds[AN5A].data_valid = TRUE;
        master->aismds[AN5B].data_valid = TRUE;
        master->aismds[AN6A].data_valid = TRUE;
        master->aismds[AN6B].data_valid = TRUE;
        master->aismds[AN7A].data_valid = TRUE;
        master->aismds[AN7B].data_valid = TRUE;

        /* store Master Radio BI values */
	master->bismds[DIG1].last_read = (unsigned long) rxp->rxbuf[1];
	master->bismds[DIG2].last_read = (unsigned long) rxp->rxbuf[2];
	master->bismds[DIG6].last_read = (unsigned long) rxp->rxbuf[17];
        master->bismds[DIG1].data_valid = TRUE;
        master->bismds[DIG2].data_valid = TRUE;
        master->bismds[DIG6].data_valid = TRUE;

	/* check status bits for master to remote communication errors                                 */
        /* have put this here for now so that we reference the data structure by proper name reference */
        /* this also lets us acquire the master data, which should be valid.                           */
	if (master->bismds[DIG6].last_read & 0x0F)
        {
	    radio->no_offlines++;
	    radio->holdflag = TRUE;
	    radio->is_responding = FALSE;
	    return(-1);
	}
        if (mdscmd != CLOSECHANNEL)
        {
        	/* Store the returned data                      */
                /* store Remote Radio AI values */
                radio->aismds[AN9].last_read  = (unsigned short) rxp->rxbuf[11];
                radio->aismds[AN10].last_read = (unsigned short) rxp->rxbuf[10];
                radio->aismds[AN13].last_read = (unsigned short) rxp->rxbuf[9];
                radio->aismds[AN14].last_read = (unsigned short) rxp->rxbuf[15];
                radio->aismds[EEP1].last_read = (unsigned short) rxp->rxbuf[12];
                radio->aismds[EEP2].last_read = (unsigned short) rxp->rxbuf[13];
                radio->aismds[EEP3].last_read = (unsigned short) rxp->rxbuf[14];
                radio->aismds[SPR1].last_read = (unsigned short) rxp->rxbuf[18];
                radio->aismds[SPR2].last_read = (unsigned short) rxp->rxbuf[19];
                radio->aismds[AN9].data_valid  = TRUE;
                radio->aismds[AN10].data_valid = TRUE;
                radio->aismds[AN13].data_valid = TRUE;
                radio->aismds[AN14].data_valid = TRUE;
                radio->aismds[EEP1].data_valid = TRUE;
                radio->aismds[EEP2].data_valid = TRUE;
                radio->aismds[EEP3].data_valid = TRUE;
                radio->aismds[SPR1].data_valid = TRUE;
                radio->aismds[SPR2].data_valid = TRUE;

                /* store Remote Radio BI values */
                radio->bismds[DIG4].last_read = (unsigned long) rxp->rxbuf[16];
                radio->bismds[DIG4].data_valid = TRUE;

                /* when all have been done successfully. get out. */
        	radio->is_responding = TRUE;
/*	        radio->data_valid = TRUE; */
                return(0);
        }
/* remote diagnostic action is completed. */
   return(0);
}
/* end of send_remotecontrol */
/*
/* makedtmf 									
/*                                                                        	
*/
/* called from                                                                  */

static	void makedtmf (unsigned short radionum, unsigned char *retbuf)
    {
	unsigned char   x, x1, x2, y, y1, y2;

        x  = radionum/100;
        if((x1 = x/10) == 0)
               x1 = 0x0A;
        if((x2 = x%10) == 0)
               x2 = 0x0A;
	retbuf[0] = ((x1 << 4) | x2);
        y  = radionum%100;
        if((y1 = y/10) == 0)
               y1 = 0x0A;
        if((y2 = y%10) == 0)
               y2 = 0x0A;
	retbuf[1] = ((y1 << 4) | y2);
}
/* end of makedtmf */


/*                                                                                       
/* makecrc 									         
/*         this routine builds and stores the 16 bit crc value for the supplied string.  
/*                                                                                       
*/
/* this routine only called from send_remoteactivate. */

static	void makecrc (unsigned char *chkbuf, int len, char unsigned *retbuf)

    {
	unsigned short checksum;
	unsigned int     i, j;
	unsigned short   flag;
	unsigned char   tempch;

/*	checksum = 0x0ffff; */
	checksum = 0x00000;
	for (j = 0; j < len; j++)
	{
	    tempch = chkbuf [j];
	    checksum ^= tempch;
	    for (i = 0; i < 8; i++)
	    {
		flag = checksum & 1;
		checksum >>= 1;
		if (flag)
		{
		    checksum ^= (unsigned short) 0x0a001;
		}
	    }
	}
	retbuf[1] = ((checksum >> 8) & 0xff);
	retbuf[0] = (checksum & 0xff);
}
/* end of makecrc */

/*
/* makechecksum 									         
/*         this routine builds and stores the simple 8 bit checksum for the supplied string.     
/*                                                                                               
*/
/* this routine only called from mdsmsgxfer. */

static	void makechecksum (unsigned char *chkbuf, int len, char unsigned *retbuf)
    {
	unsigned short checksum;
	unsigned int     i, j;
	unsigned short   flag;
	unsigned char   tempch;

	checksum = 0x00000;
	for (j = 0; j < len; j++)
	{
	    checksum += chkbuf [j];
	}
	retbuf[1] = ((checksum >> 8) & 0xff);
	retbuf[0] = (checksum & 0xff);
        return;
    }
/* end of makechecksum */

/*                                                                                      
/* on_line                                                                              
/*         try to bring a Mds Remote Radio on line.                                     
/*                                                                                      
/*
*******/
/* this routine only called from MAIN */

static void on_line (unsigned short line, unsigned short radionum)
    {
	MDSRADIO  *radio;

	radio = radios[line][radionum];
	if (radio)			/* verify pointer prior to use */
	{
	        if (trace[line] & REMOTEPOLLTRACE)
		{
			logMsg ( "Attempt to bring Online remote radio # %d \n",
				radionum );
	        }
		/* Use the loop back function to see if the unit is there and responding */
		process_remote_radio(line, nextoffline[line]); /* I/O done here!    */
		if (radio->is_responding)
		{
                      radio->offline = FALSE;
                      radio->retry_count = 0;
                      mark_not_valid(radio);
	              if (trace[line] & REMOTEPOLLTRACE)
		      {
			logMsg ( "Remote radio # %d successfully brought Online \n",
				radionum );
	              }
                }
        }
    }
/* end of on_line */


/*                                                              
/*	mark_not_valid						
/*								
/* 	Mark the MDS Remote Radio data elements as not valid	
/*                                             			
*******/                                             	
/* called from prepresponses      		*/    	

static void mark_not_valid (MDSRADIO  *radio)
    {
	unsigned short i;

	for (i = 0 ; (i<NUMMDSBISXXX) ; i++)
	{
	    radio->bismds[i].data_valid = FALSE;
	}

	for (i = 0 ; (i<NUMMDSAISXXX) ; i++)
	{
	    radio->aismds[i].data_valid = FALSE;
	}
    }
/* end of mark_not_valid */
/*                                                                                     
/*      AIDRIVER									
/*                                                                              	
/*      This routines main purpose is to return the analog input value for the         	
/*      specified radio data. For a specific radio data value a (-1) will be		
/*	returned if there is no valid data ready.                       		
/*                                                                                     	
*/
/* this routine not called from within this driver */

int AIDRIVER (char *pointname, unsigned short link, unsigned short adapter, unsigned short card,
		unsigned short signal, short *pvalue, unsigned short conversion,
		unsigned char *resname)
    {
	MDSRADIO  *radio;
/*         LOADMAP   *mapinfo; */
        unsigned short  blkix;
	unsigned short	*pdata;
        unsigned short zero;

        zero = 0;

	if (link > MDSLINES)
		return (-1);
	if (adapter > NUMMDSRADIOS)
		return (-1);
	if ((card != MAINS) && (card != MPSEUDO))
		return (-1);
	if (conversion != conversion)
		return (-1);                /* No warnings */

	if ((card == MAINS) && (signal > MAXAINO))
		return (-1);

	radio = radios[link][adapter];
	if (radio == NULL)
	{
		if (!allocate_radio(link, adapter))
			return(-1);         /* if not allocated, try. if can't allocate, get out. */
                radio = radios[link][adapter];
	}
	radio->was_requested = TRUE;

        if ((maptable[link]->address[(int)adapter].valid_flag != TRUE) && (trace[link] & RADIOMAPTRACE))
        {
                logMsg ("Hardware address not defined for logical radio #%d \n", adapter);
                return(-1);
        }

	if (card == MPSEUDO)
	{
	    switch (signal)
	    {
	    case 0: /* Were the radio data values requested */
		*pvalue = radio->was_requested;
		break;
	    case 1: /* Is the radio responding */
		*pvalue = radio->is_responding;
		break;

	    case 5: /* Output Messages */
		*pvalue = radio->no_dataout_msgs & 0x7fff;
		break;
	    case 6: /* Input Messages */
		*pvalue = radio->no_datain_msgs & 0x7fff;
		break;
	    case 7: /* No Receiver Responses */
		*pvalue = radio->no_rxnoresp & 0x7fff;
		break;
	    case 8: /* No Transmitter Response */
		*pvalue = radio->no_txnoresp & 0x7fff;
		break;
	    case 9: /* Data Echo Errors */
		*pvalue = radio->no_dataerrs & 0x7fff;
		break;
	    case 10: /* Message Retries */
		*pvalue = radio->no_retries & 0x7fff;
		break;
	    case 11: /* RX Hardware errors */
		*pvalue = radio->no_rxhwerrs & 0x7fff;
		break;
	    case 12: /* TX Hardware errors */
		*pvalue = radio->no_txhwerrs & 0x7fff;
		break;
	    case 13: /* CRC errors */
		*pvalue = radio->no_crcerrs & 0x7fff;
		break;
	    case 14: /* Offline transitions */
		*pvalue = radio->no_offlines & 0x7fff;
		break;
	    case 15: /* reset statistics */
		*pvalue = radio->cmd_reset_stat;
		break;
	    case 16: /* Force the radio offline */
		*pvalue = radio->cmd_force_offline;
		break;
	    case 17: /* Failed output commands */
		*pvalue = radio->badoutcnt & 0x7fff;
		break;

	    case 24: /* Fail/Retry Counter */
		*pvalue = radio->retry_count & 0x7fff;
		break;
	    case 25: /* Radio Offline - out of polling loop */
		*pvalue = radio->offline;
		break;
	    case 26: /* Is the radio on hold */
		*pvalue = radio->holdflag;
		break;
	    case 27: /* Message Length errors */
		*pvalue = radio->no_rxlenerrs & 0x7fff;
		break;
	    case 28: /* Checksum errors */
		*pvalue = radio->no_chksumerrs & 0x7fff;
		break;
            case 29: /* read master radio polling time: only valid for master radio */
                if (adapter != MASTERADDRESS)
		    return (-1);
                *pvalue = masterinterval[link];
		break;
            case 30: /* read remote radio polling time: only valid for master radio */
                if (adapter != MASTERADDRESS)
		    return (-1);
                *pvalue = remoteinterval[link];
		break;
            case 31: /* hardware/DTMF address of this master/remote radio */
                *pvalue = maptable[link]->address[(int)adapter].hardware;
                break;

	    default:
		return (-1);
	    }
	    return (0);
	}

	if (resname != 0)
	{ 
	    register_resource(link, radio, resname, pointname);
	}

	if (card == MAINS)
	{
	    radio->aismds [signal].was_requested = TRUE;
	    if (!resresponse[link])
	    	return (-1);
	    if (radio->holdflag)
	    	return (-2);
	    if (!radio->is_responding)
	    	return (-1);
	    if (!radio->aismds[signal].data_valid)
	    {
		return (-1);
	    }
	    pdata = &(radio->aismds[signal].last_read);
	    *pvalue = *pdata;
	}
	return(0);
} /* end of AIDRIVER */

/*  										
/*      AODRIVER                                                                
/*                                                                            	
/*      This routines main purpose is to set requested analog outputs to the  	
/*      specified module.                                                     	
*/                                                                            	
/* this routine not called from within this driver */

int AODRIVER (char *pointname, unsigned short link, unsigned short adapter, unsigned short card,
	    unsigned short signal, double dvalue, unsigned short value,
	    unsigned short conversion, unsigned char *resname)
    {
	MDSRADIO  *radio;
	unsigned short  blkix;
	OUTQENT *oqp;

	if (link > MDSLINES)
		return (-1);
	if (adapter > NUMMDSRADIOS)
		return (-1);
	if ((card != MREGS) && (card != MPSEUDO))
		return (-1);

	radio = radios[link][adapter];
	if (radio == NULL)
	{
		if (!allocate_radio(link, adapter))
			return(-1); /* if not allocated, try. if can't allocate, get out. */
                radio = radios[link][adapter];
	}

        if ((maptable[link]->address[(int)adapter].valid_flag != TRUE) && (trace[link] & RADIOMAPTRACE))
        {
                logMsg ("Hardware address not defined for logical radio #%d\n", adapter);
                return(-1);
        }

/* printf("\n%s<[%d,%d,%d,%d] Value=%d: Resreponse=%x:Is_responding=%x)>",
                                            pointname,
                                            link, adapter, card, signal,
                                            value,
                                            resresponse[link],
                                            radio->is_responding,
                                            blkix);
*/
	if (card == MPSEUDO)
	{
	    switch (signal)
	    {
	    case 15: /* Reset the accumulated statistics */
		if (value != 0)
			init_stats(radio);
		break;
	    case 16: /* Force the Radio online/offline */
		radio->cmd_force_offline = (value != 0);
                if (value != 0 )
                    radio->is_responding = FALSE;/* only required when asserting cmd_force_offline */
		break;

            case 29: /* set master radio polling time: only valid for master radio */
                if (adapter != MASTERADDRESS)
		    return (-1);
                masterinterval[link] = value;
		break;

            case 30: /* set remote radio polling time: only valid for master radio */
                if (adapter != MASTERADDRESS)
		    return (-1);
                remoteinterval[link] = value;
		break;

            case 32: /* force a diagnostic loopback on a remote radio: not valid for master radio */
                if (adapter == MASTERADDRESS)
		   return (-1);
                radio->was_requested = TRUE;

                oqp = ogetemptyp (link);
                if (oqp == NULLP)
                {
                        return (-1);		/* output que is full */
                }
                oqp->otype = FORCEREMOTELOOPBACKTYPE;
                oqp->out.aonode.link = link;
                oqp->out.aonode.adapter = adapter;/* the command is sent to each remote radio   */
                                                  /* individually.                              */
                oqp->out.aonode.card = 0;
                oqp->out.aonode.aosignal = 0;
                oqp->out.aonode.aovalue = 0;
                oqp->out.aonode.spare = 0;
                omarkfull (link);		/* Mark this entry as ready for output.         */
		break;
            default:
		return (-1);
            }
	    return (0);
	}
	if (resname != 0)
	{ 
	    register_resource(link, radio, resname, pointname);
	}
	if (signal > MAXAONO)
	{
	    return (-1);
	}
	radio->was_requested = TRUE;

	if (!resresponse[link])
		return (-1);

	if ((!radio->is_responding) || (radio->cmd_force_offline))
	{
	    return (-1);
	}

       /* there are no 'real' AO points defined for the Master Radio */
        if (adapter == MASTERADDRESS)
		return (-1);

        /* store the output request in the output queue for demand list processing */
	oqp = ogetemptyp (link);
	if (oqp == NULLP)
	{
	    return (-1);		/* output que is full */
	}
	oqp->otype = AOUTTYPE;
	oqp->out.aonode.link = link;
	oqp->out.aonode.adapter = adapter;
	oqp->out.aonode.card = card;
	oqp->out.aonode.aosignal = signal;

	oqp->out.aonode.aovalue = value;

	omarkfull (link);		/* Mark this entry as ready for output */

	return (0);
    }

/* end of AODRIVER */

/*                                                                                     
/*      BIDRIVER                                                                        
/*                                                                               	
/*      The main purpose of this routine is to return the request digital input  	
/*      value(s) from the requested card. This routine will return a (-1) if the  	
/*      requested module is the wrong type or the data is not yet valid.          	
/*                                                                                	
*/
/* this routine not called from within this driver */

int BIDRIVER (char *pointname, unsigned short link, unsigned short adapter, unsigned short card,
		unsigned short signal, unsigned long mask, unsigned long *pvalue,
		unsigned char *resname)
{
	MDSRADIO  *radio;
	unsigned short blkix;
	unsigned long tvalue;

	if (link > MDSLINES)
		return (-1);
	if (adapter > NUMMDSRADIOS)
		return (-1);
	if ((card != MDINS) && (card != MPSEUDO))
		return (-1);

	radio = radios[link][adapter];
	if (radio == NULL)
	{
                if (!allocate_radio(link, adapter))
                   return; /* if not allocated, try. if can't allocate, get out. */
                radio = radios[link][adapter];
	}

	radio->was_requested = TRUE;

        if ((maptable[link]->address[(int)adapter].valid_flag != TRUE) && (trace[link] & RADIOMAPTRACE))
        {
                logMsg ("Hardware address not defined for logical radio #%d\n", adapter);
                return;
        }

	if (card == MPSEUDO)
	{
	    switch (signal)
	    {
	        case 0: /* Were the radio data values requested */
		     *pvalue = radio->was_requested;
		     break;
	        case 1: /* Is the radio responding */
		     *pvalue = radio->is_responding;
		     break;

	        case 15: /* reset statistics */
		     *pvalue = radio->cmd_reset_stat;
		     break;
	        case 16: /* Force the radio offline */
		     *pvalue = radio->cmd_force_offline;
		     break;

	        case 25: /* Radio Offline - out of polling loop */
		     *pvalue = radio->offline;
		     break;
	        case 26: /* Is the radio on hold */
		     *pvalue = radio->holdflag;
		     break;

		default:
		     return (-1);
	    }
	    return (0);
	}

        /* 'real' BI data points only valid for master radio in specified range.      */
        if ((adapter == MASTERADDRESS ) && (( signal >= 24 ) && (signal <= 31 )))
           return (-1);

        /* 'real' BI data points only valid for remote radio in specified range.      */
        if ((adapter != MASTERADDRESS ) && (( signal < 24 ) || (signal > 31 )))
           return (-1);

	if (resname != 0)
	{ 
	    register_resource(link, radio, resname, pointname);
	}

	blkix = ((signal / 8) & 0xFFFC);	/* 8 bits per digital BLOCK */
						/* mask aligned to 32 bits  */
	tvalue = 0;

	if (card == MDINS)
	{
	    if (signal > MAXBINO)
	    {
		return (-1);
	    }

	    radio->was_requested = TRUE;
	    if (!resresponse[link])
			return (-1);
	    if (radio->holdflag)
			return (-2);
	    if (!radio->is_responding)
			return (-1);

	    if ( (mask & 0x000000FF) != 0)
	    {
	        radio->bismds [blkix].was_requested = TRUE;
		if (!resresponse[link])
			return (-1);
		if (radio->holdflag)
			return (-2);
		if (!radio->is_responding)
			return (-1);
		if (!radio->bismds [blkix+0].data_valid)
		{
		    return (-1);
		}
	        tvalue = radio->bismds [blkix+0].last_read & mask;                      /* byte 1 */
	    }
	    if ( (mask & 0x0000FF00) != 0)
	    {
	        radio->bismds [blkix+1].was_requested = TRUE;
		if (!resresponse[link])
			return (-1);
		if (radio->holdflag)
			return (-2);
		if (!radio->is_responding)
			return (-1);
		if (!radio->bismds [blkix+1].data_valid)
		{
		    return (-1);
		}
	        tvalue += (unsigned long) (radio->bismds [blkix+1].last_read) << 8;     /* byte 2 */
	    }
	    if ( (mask & 0x00FF0000) != 0)
	    {
	        radio->bismds [blkix+2].was_requested = TRUE;
		if (!resresponse[link])
			return (-1);
		if (radio->holdflag)
			return (-2);
		if (!radio->is_responding)
			return (-1);
		if (!radio->bismds [blkix+2].data_valid)
		{
		    return (-1);
                }
	        tvalue += (unsigned long) (radio->bismds [blkix+2].last_read) << 16;    /* byte 3 */
	    }
	    if ( (mask & 0xFF000000) != 0)
	    {
	        radio->bismds [blkix+3].was_requested = TRUE;
		if (!resresponse[link])
			return (-1);
		if (radio->holdflag)
			return (-2);
		if (!radio->is_responding)
			return (-1);
		if (!radio->bismds [blkix+3].data_valid)
		{
		    return (-1);
		}
	        tvalue += (unsigned long) (radio->bismds [blkix+3].last_read) << 24;    /* byte 4 */
	    }
            *pvalue = tvalue & mask;
	    return (0);
	}
*pvalue = tvalue & mask;
printf("<BI-1>");
	return (-1);
    }
/* end of BIDRIVER */

/*                                                                                	
/*      BODRIVER                                                                        
/*                                                                                 	
/*      The main purpose of this routine is to set the request digital output      	
/*      value(s) on the requested card. This routine will return a (-1) if the     	
/*      requested module is the wrong type or not configured yet.                  	
/*                                                                                 	
*/
/* this routine not called from within this driver */


int BODRIVER (char *pointname, unsigned short link, unsigned short adapter, unsigned short card,
		unsigned short signal, unsigned long value, unsigned long mask,
		unsigned char *resname)
    {
	MDSRADIO  *radio;
	MDSRADIO  *master;
	OUTQENT   *oqp;
	unsigned short  blkix;

	if (link > MDSLINES)
		return (-1);
	if (adapter > NUMMDSRADIOS)
		return (-1);
	if ((card != MCOILS) && (card != MPSEUDO))
		return (-1);
        if ((maptable[link]->address[(int)adapter].valid_flag != TRUE) && (trace[link] & RADIOMAPTRACE))
        {
                logMsg ("Hardware address not defined for logical radio #%d\n", adapter);
                return;
        }

	radio = radios[link][adapter];
	if (radio == NULL)
	{
                if (!allocate_radio(link, adapter))
                   return; /* if not allocated, try. if can't allocate, get out. */
                radio = radios[link][adapter];
	}
	radio->was_requested = TRUE;

	if (card == MPSEUDO)
	{
	    switch (signal & 0x1f)
	    {
	    case 15: /* Reset the accumulated statistics */
		if (value != 0) init_stats(radio);
		break;

	    case 16: /* Force the radio offline */
		radio->cmd_force_offline = (value != 0);
                if (value != 0 )
                    radio->is_responding = FALSE;/* only required when asserting cmd_force_offline */
		break;

            case 32: /* force a diagnostic loopback on a remote radio: not valid for master radio */
                if (adapter == MASTERADDRESS)
		   return (-1);
                oqp = ogetemptyp (link);
                if (oqp == NULLP)
                {
                        return (-1);		/* output que is full */
                }
                oqp->otype = FORCEREMOTELOOPBACKTYPE;
                oqp->out.aonode.link = link;
                oqp->out.aonode.adapter = adapter;/* the command is sent to each remote radio   */
                                                  /* individually.                              */
                oqp->out.aonode.card = 0;
                oqp->out.aonode.aosignal = 0;
                oqp->out.aonode.aovalue = 0;
                oqp->out.aonode.spare = 0;
                omarkfull (link);		/* Mark this entry as ready for output.         */
		break;
	    default:
		return (-1);
	    }
	    return (0);
	}

	if (signal > MAXBONO)
	{
	    return (-1);
	}
	/* 'real' BO data points are only valid for: master radio: signals 0 - 5, remote radios: signal 7 - 9. */
		if ((adapter == MASTERADDRESS) && (signal > 5))
		return (-1);
		if ((adapter != MASTERADDRESS) && ((signal < 7) || (signal >9)) )
		return (-1);

	master = radios[link][adapter];
	master->was_requested = TRUE;

	if (resname != 0)
	{ 
	    register_resource(link, master, resname, pointname);
	}
	if (!resresponse[link])
		return (-1);

	if ((!master->is_responding) || (master->cmd_force_offline))
	{
	    return (-1);
	}

	oqp = ogetemptyp (link);
	if (oqp == NULLP)
	{
	    return (-1);		/* output que is full */
	}
	oqp->otype = BOUTTYPE;
	oqp->out.bonode.link = link;
	oqp->out.bonode.adapter = adapter;
	oqp->out.bonode.card = card;
	oqp->out.bonode.bosignal = signal;
	oqp->out.bonode.bomask = mask;
	oqp->out.bonode.bovalue = value;

	omarkfull (link);		/* Mark this entry as ready for output */

	return (0);
    }
/* end of BODRIVER */

/*                                                                             
/*      AOREAD                                                                  
/*                                                                          	
/*      This routine is stubbed out to satisfy dev level logic. the MDS radio	
/* 	does not hold the ouptut values, so we cannot retrieve them. we just	
/*	stub this routine out and return a zero value and no error.		
*/
/* this routine not called from within this driver */


int AOREAD (char *pointname, unsigned short link, unsigned short adapter, unsigned short card,
	  unsigned short signal, unsigned short *pvalue,
	  unsigned short conv, unsigned char *resname)
    {
	MDSRADIO  *radio;
        unsigned short  blkix;
	unsigned short	*pdata;
        unsigned short zero;

        zero = 0;

	if (link > MDSLINES)
		return (-1);
	if (adapter > NUMMDSRADIOS)
		return (-1);
	if ((card == MREGS) && (signal <= MAXAONO))   /* if 'normal' data, then just return zero.  */
	{	*pvalue = 0;
	        return (0);
        }
	if ((card == MREGS) && (signal > MAXAONO))
		return (-1);
	if ((card != MPSEUDO))                        /* AOREAD only valid for some pseudo commands */
		return (-1);

	radio = radios[link][adapter];
	if (radio == NULL)
	{
		if (!allocate_radio(link, adapter))
			return(-1);         /* if not allocated, try. if can't allocate, get out. */
                radio = radios[link][adapter];
	}
	radio->was_requested = TRUE;

        if ((maptable[link]->address[(int)adapter].valid_flag != TRUE) && (trace[link] & RADIOMAPTRACE))
        {
                logMsg ("Hardware address not defined for logical radio #%d \n", adapter);
                return(-1);
        }

	if (card == MPSEUDO)
	{
	    switch (signal)
	    {
	    case 15: /* reset statistics */
		*pvalue = radio->cmd_reset_stat;
		break;
	    case 16: /* Force the radio offline */
		*pvalue = radio->cmd_force_offline;
		break;

            case 29: /* read master radio polling time: only valid for master radio */
                if (adapter != MASTERADDRESS)
		    return (-1);
                *pvalue = masterinterval[link];
		break;
            case 30: /* read remote radio polling time: only valid for master radio */
                if (adapter != MASTERADDRESS)
		    return (-1);
                *pvalue = remoteinterval[link];
		break;

	    default:
		return (-1);
	    }
	    return (0);
	}
	return (-1);
    }
/* end of AOREAD */







/*                                                                              
/*      BOREAD                                                                  
/*                                                                           	
/*      This routine is stubbed out to satisfy dev level logic. the MDS radio   
/* 	does not hold the ouptut values, so we cannot retrieve them. we just	
/*	stub this routine out and return a zero value and no error.		
*/
/* this routine not called from within this driver */


int BOREAD (char *pointname, unsigned short link, unsigned short adapter, unsigned short card,
		unsigned short signal, unsigned long *pvalue, unsigned long mask,
		unsigned char *resname)

    {
	MDSRADIO  *radio;
        unsigned short  blkix;
	unsigned short	*pdata;
        unsigned short zero;

        zero = 0;

	if (link > MDSLINES)
		return (-1);
	if (adapter > NUMMDSRADIOS)
		return (-1);
	if ((card == MCOILS) && (signal <= MAXAONO))   /* if 'normal' data, then just return zero.  */
	{	*pvalue = 0;
	        return (0);
        }
	if ((card == MCOILS) && (signal > MAXAONO))
		return (-1);
	if ((card != MPSEUDO))                        /* AOREAD only valid for some pseudo commands */
		return (-1);

	radio = radios[link][adapter];
	if (radio == NULL)
	{
		if (!allocate_radio(link, adapter))
			return(-1);         /* if not allocated, try. if can't allocate, get out. */
                radio = radios[link][adapter];
	}
	radio->was_requested = TRUE;

        if ((maptable[link]->address[(int)adapter].valid_flag != TRUE) && (trace[link] & RADIOMAPTRACE))
        {
                logMsg ("Hardware address not defined for logical radio #%d \n", adapter);
                return(-1);
        }

	if (card == MPSEUDO)
	{
	    switch (signal)
	    {
	    case 15: /* reset statistics */
		*pvalue = radio->cmd_reset_stat;
		break;
	    case 16: /* Force the radio offline */
		*pvalue = radio->cmd_force_offline;
		break;

	    default:
		return (-1);
	    }
	    return (0);
	}
	return (-1);
    }
/* end of BOREAD */

/*                                                                                     
/* register_resource                                                              	
/*                                                                                	
/*	This routine is used to initialize the resource pointers
/*  If this is the first time the resource for this line is being                  	
/*  referenced the pointer to the resource table entry will be set                  	
/*  and the resource name will be saved. If the resource name has                    	
/*  already been set, this routine will verify that all other points                 	
/*  for this driver are using the same resource name.                                	
/*                                                                                   	
*******/                                                                                    	
/* this routine called from AIDRIVER, AODRIVER, BIDRIVER, and BODRIVER */                               	

static void  register_resource (unsigned short link, MDSRADIO *pmod,
                                unsigned char *resname, char *pointname)
    {
	if (resnames[link][0] == 0) /* This is the first reference */
	{
	    if (resGetByName( resname, &prestable[link]) != RES_OK)
	    {
		logMsg ("Resource not found in table.(Point Name = %s, Resource Name = %s)\n",
                                pointname,
				resname );
	    }
	    else strcpy(resnames[link], resname);
	}
	else
	{
	    if (strcmp(resnames[link], resname) != 0)
	    {
		logMsg ("Resource name[%s] conflict.(Point Name = %s, Resource Name = %s)\n",
				resnames[link],
                                pointname,
				resname );
	    }
	}
    }
/* end of register_resource */

/*                                                                  		
/*  outqueinit
/*         Initialize the Queue                                      		
/*                                                                              
*******/                                                                        	
/* called from MAIN */

static	void outqueinit (unsigned short line)
    {
	oq[line]->ooutix = oq[line]->oinix = 0;
    }
/* end of outqueinit */


/*                                                                     		
/*  ogetempty                                                                   
/*       get the next available slot on the output list         		
/*                                                                              
*******/
/*    called from AODRIVER and BODRIVER   */

static	OUTQENT *ogetemptyp (unsigned short line)
    {
	if (((oq[line]->oinix + 1) % OUTQSIZE) != oq[line]->ooutix)
	{
	    return (&oq[line]->outque[oq[line]->oinix]);
	}
	if (trace[line] & OUTPUTQUEUETRACE)
		logMsg ( "Output Que Full line - %d\n",line);
	return (NULLP);
    }
/* end of ogetemptyp */


/*                                                                    		
/* omarkfull                                                          		
/*      mark the next block as full                                     	
/*                                                                              
*******/
/*  called from AODRIVER and BODRIVER   */

static	void omarkfull (unsigned short line)
	{
	    oq[line]->oinix = ((oq[line]->oinix + 1) % OUTQSIZE);
	}
/* end of omarkfull */


/*                                                                      	
/*  ogetnextp - get the next full slot on the output list               	
/*
*******/                                                                      	
/* only called from process_demand_list */

static	OUTQENT *ogetnextp (unsigned short line)
    {
	if (oq[line]->ooutix != oq[line]->oinix)
	{
	    return (&oq[line]->outque[oq[line]->ooutix]);
	}
	else return (NULLP);
    }
/* end of ogetnextp */



/*                                                       	               	
/* oreleasefull - release the current full output block    	              	
/*
*******/                                                                            	
/* only called from process_demand_list */

static	void oreleasefull (unsigned short line)
    {
	if (oq[line]->ooutix == oq[line]->oinix)
		return;
	oq[line]->ooutix = ((oq[line]->ooutix + 1) % OUTQSIZE);
    }
/* end of oreleasefull */


/*                                                          			
/*      link_status                                                           	
/*                                                                            	
/*      The purpose of this routine is to print out the state of each of the  	
/*      Mds radios and the statistics gathered about them.                      
/*                                                                           	
*******/
/* not called from within this driver */

static	void link_status (unsigned short line)
    {
    if (line > MDSLINES-1) return;
/* do nothing for now */
    }
/* end of link_status */


/*                                                                              
/*	allocate_radio
/*                                                                              
/*	The purpose of this routine is to allocate the memory for a single      
/*	radio data structure. If not successful, a console error message is  	
/*	printed and an error flag is returned to the caller.                    
/*	If successful, the newly allocated data structures are initialized.     
/*                                                                             	
*******/
/* this procedure called from check_demand_list, AIDRIVER, AODRIVER, and BIDRIVER */

static int allocate_radio ( unsigned short line, unsigned short radio )
   {
        MDSRADIO   *pradio;

	pradio = (MDSRADIO *)calloc(1,(sizeof(MDSRADIO)));
	if (!pradio)
	{
		logMsg ("Unable to allocate memory for ");
		printf ("MDSRADIO data structure for radio address %d\n", radio);
		 return(FALSE);
	}
	else
	{
	        if (trace[line] & CALLOCTRACE)
                {
		        logMsg ("Allocated memory for ");
                        printf ("MDSRADIO data structure for radio address %d\n", radio);
                }
		init_line_tab(line, pradio);	     /*	initialize the data structures. */
                radios[line][radio] = pradio;        /* fill table AFTER structure init */
	}

	return(TRUE);
   }
/* end of allocate_radio */

/*
/*                                                                              
/*	allocate_master                                                         
/*                                                                              
/*	The purpose of this routine is to allocate the memory for the           
/*	master radio data structure. If not successful, a console error         
/*      message is printed and an error flag is returned to the caller.         
/*	If successful, the newly allocated data structures are initialized.     
/*                                                                             	
*******/
/* this procedure only called from MAIN */

static int allocate_master (unsigned short line)
   {
        MDSRADIO   *pmaster;

	pmaster = (MDSRADIO *)calloc(1,(sizeof(MDSRADIO)));
	if (!pmaster)
	{
		logMsg ("Unable to allocate memory for MDS Master Radio\n");
		 return(FALSE);
	}
	else
	{
	        if (trace[line] & CALLOCTRACE)
                {
		        logMsg ("Allocated memory for MDS Master Radio\n");
                }
		init_line_tab(line, pmaster);	     /*	initialize the data structures. */
                radios[line][MASTERADDRESS] = pmaster;
	}

	return(TRUE);
   }
/* end of allocate_master */


/*
/*      start_rx                                                        	
/*                                                                     	 	
/*      The purpose of this routine is to initialize a line table        	
/*      for receive and to enable the receiver.                           	
/*                                                                        	
*/
/* only called from mdsmsgxfer */

static	void start_rx ( unsigned short line,SEMAPHORE *semid,unsigned short *pstatus )

    {
	RXLINE                  *rxp;

	rxp = rxlinetab[line];
	rxp->rxix = 0;
	rxp->semid = semid;
	rxp->status = pstatus;
	rxp->rxstate = RXADRST;

    }
/* end of start_rx */





/*                                                                           	
/*      stop_rx                                                              	
/*                                                                           	
/*      The purpose of this routine is to initialize a line table            	
/*      for receive and to disable the receiver.                             	
/*                                                                           	
*/
/* only called from msgxfer */


static	void stop_rx ( unsigned short line )

    {
	RXLINE                  *rxp;

	rxp = rxlinetab[line];
	rxp->rxstate = RXIGNORE;
    }
/* end of stop_rx */


/*                                                                        	
/*      rx_int_proc                                                        	
/*                                                                         	
/*      The purpose of this routine is to process a receive character interrupt	
/*      for a given line.                                                     	
*/
/* referenced in system call in MAIN */

static	void rx_int_proc ( unsigned short line )
{
	unsigned char   rxchar;
	unsigned short  rxstat;
	unsigned short  rx_err_mask;
	RXLINE          *rxp;

	read_rx_hw(line,&rxchar,&rxstat,&rx_err_mask);

	rxp = rxlinetab[line];

	/* Something is sending in data */
	line_activity[line] = TRUE;

	if (rxp->rxstate == RXIGNORE)
		return;

        /* if status and error info is non-zero, check index */
        if ((rxstat & rx_err_mask) != 0)
	{
            /* if this is prior to the first good character, ignore errors - probably line noise */
	    if (rxp->rxix == 0)
	    	return;
	    /* if communications error, flag the error, store relevant info and get out */
            *(rxp->status) = RXBAD;
	    rxp->rxstate = RXIGNORE;
	    rxp->rxbuf[100] = rxstat;
	    rxp->rxbuf[101] = rxchar;
	    return;
	}

        /* if this is first character, make sure it is the message header byte. if not, ignore. */
	if ((rxp->rxix == 0) && (rxchar != MDSMSGSTARTBYTE))
	     return;

       /* store the data in the receive buffer                      */
        rxp->rxbuf[rxp->rxix++] = rxchar;
        rxp->rxmsglen--;
        if (rxp->rxmsglen <= 0)
        {
                *(rxp->status) = RXGOOD;
                rxp->rxstate=RXIGNORE;
                semGive(rxp->semid);
        }
        return;
}
/* end of rx_int_proc */

/*                                                                           	
/*      check_line_activity                                                     
/*                                                                             	
/*      The purpose of this routine is to process a receive character interrupt	
/*      for a given line.                                                      	
/*                                                                             	
*/
/* was previously called from MAIN, not currently called.  */


static	int check_line_activity (unsigned short line, unsigned short predelay )

    {
	RXLINE          *rxp;
	long		thistime;
	short		status;

	if (predelay != 0)
		taskDelay(predelay);
	thistime = tickGet();
	if ((thistime-activity_time[line]) < ACTIVITYTIME)
		return (lastactivity[line] || (line_activity[line]));
	lastactivity[line] = status = line_activity[line];
	line_activity[line] = FALSE;
	activity_time[line] = thistime;
	rxp = rxlinetab[line];
	rxp->rxstate = RXIGNORE;

	return (status);
    }
/* end of check_line_activity */


/*                                                                         	
/*      start_tx                                                            	
/*                                                                          	
/*      The purpose of this routine is to initialize a line table           	
/*      for send and to enable the transmitter.                              	
/*                                                                            	
*/
/* only called from msgxfer  */


static	void start_tx (unsigned short line,SEMAPHORE *semid,unsigned short *pstatus )

    {

	TXLINE                  *txp;

	disable_tx_ints(line);
	txp = txlinetab[line];
	txp->txix = 0;
	txp->semid = semid;
	txp->status = pstatus;
	txp->txdisable = FALSE;

	/* Might have to output the first character here */


	enable_tx_ints(line);

    }
/* end of start_tx */

/*
/*      stop_tx                                                                	
/*                                                                             	
/*      The purpose of this routine is to initialize a line table              	
/*      for receive and to disable the receiver.                               	
/*                                                                             	
*/
/* only called from msgxfer  */


static	void stop_tx (unsigned short line )

    {
	TXLINE			*txp;

	disable_tx_ints(line);
	txp = txlinetab[line];
	txp->txdisable=TRUE;
    }
/* end of stop_tx */


/*                                                                           	
/*      tx_int_proc                                                           	
/*                                                                            	
/*      The purpose of this routine is to process a transmit character interrupt
/*      for a given line.                                                  	
/*                                                                         	
*/
/* only called from MAIN  */


static	void tx_int_proc (unsigned short line )

    {
	char                    txchar;
	unsigned short          txstat;
	unsigned short          tx_err_mask;
	TXLINE                  *txp;

	txp = txlinetab[line];
	if (txp->txdisable)
		return;
	if (txp->txix == txp->txmsglen)
	{
	    txp->txdisable=TRUE;
	    disable_tx_ints (line);
	    *(txp->status) = TXGOOD;
	    semGive(txp->semid);
	    return;
	}
	else
	{
	    txchar = txp->txbuf[txp->txix++];
	    write_tx_hw ( line,txchar,&txstat,&tx_err_mask );
	    if ( (txstat & tx_err_mask) != 0 )
	    {
	 	txp->txdisable=TRUE;
	        disable_tx_ints (line);
	        *(txp->status) = TXBAD;
	        semGive(txp->semid);
	        return;
	    }

	}

	if (txp->txix >= MAXTXMSG)
	{
	    txp->txdisable=TRUE;
	    disable_tx_ints (line);
	    *(txp->status) = TXBAD;
	    semGive(txp->semid);
	    return;
	}

    }
/* end of tx_int_proc */


/*
/*      mdsmsgout                                                             	
/*                                                                           	
/*      The purpose of this routine is to transmit a buffer to the MDS master   
/*      radio. It Does NOT await a response. The transmit and receive           
/*      buffers are in their respective line tables. This routine will retry 	
/*      the message up to a maximum number of times or until a valid       	
/*      transmission has occurred.                                              
/*              returns         XFERGOOD if transmit is OK.               	
/*                              XFERBAD  if transmit not OK.
/*                                                                              
*/
/* called from 						*/

static	unsigned short  mdsmsgout (unsigned short fline, 
				 MDSRADIO *radio, unsigned short rxlength,
                                 unsigned short checksumflag, unsigned long time, unsigned short type)
    {
	unsigned short  retry, txstatus, rxstatus, txlen, rxlen, line, ix;
	RXLINE          *rxp;
	TXLINE          *txp;
	unsigned char   csums[2];
        unsigned short  checksumpass;
        unsigned int    tick1, tick2, tick3, tick4;
        unsigned int    delta12, delta23, delta34;

        tick1 = tick2 = tick3 = tick4 = 0;     /* pre-zero all time keep registers */
        delta12 = delta23 = delta34 = 0;

	line = (fline & (~ONERETRY));          /* strip off the no retry flag*/

	rxp = rxlinetab[line];
	txp = txlinetab[line];

	for (retry=((fline == line)?0:MAXRETRYS-1); (retry<MAXRETRYS); retry++ )
	{
	    if (intermsgdelay[line] != 0)
	    {
		taskDelay (intermsgdelay[line]);
	    }

	    set_rts(line);

	    if (pretxdelay[line] != 0)
	    {
		taskDelay (pretxdelay[line]);
	    }
	    txstatus = IO_TIMEOUT;
	    rxp->rxmsglen = rxlength;
	    rxstatus = IO_TIMEOUT;
	    semTake(semids[line][TXSEM], NO_WAIT);  /* take it 'on behalf' of tx interrupt */

            set_release_time (line, time);          /* adjust 'release channel' semaphore 'oneshot' timer.*/

	    txlen = txp->txmsglen;

            if ((trace[line]) & (type))
            {
               printf("TX:");
               for (ix = 0; ix < (12 + 24) ; ix++) /* 12 is maximum TX message length, pad past RX */
               {
                    if (ix < txlen)
/*                       printf("_%2x", (char *) txp->txbuf[ix]); */
                       printf("_%2x", txp->txbuf[ix]);
                    else
                       printf("...");
               } 
               printf(".... ");
            }

            tick1 = tickGet();
	    start_tx(line, semids[line][TXSEM], &txstatus); /* this fires off the transmission. */

	    /* 'wait' here for tx_int_proc to complete and release the semaphore. */
            if ((semTake(semids[line][TXSEM], txp->txtimeout)) == ERROR)
		txstatus = IO_TIMEOUT;

            tick2 = tickGet();
            delta12 = (tick2-tick1)%10000;

            stop_tx(line);

	    if (posttxdelay[line] != 0)
	    {
		taskDelay (posttxdelay[line]);
	    }

	    if (halfduplex[line] = TRUE)
               clear_rts(line);    /* half duplex rts line control */

            tick3 = tickGet();
            delta23 = (tick3-tick2)%10000;


	    if (txstatus == TXGOOD)
	    {
                if ((trace[line]) & (type))
                        printf("%3d %3d\n", delta12, delta23);
		return (XFERGOOD);
	    } /* end of (txstatus == TXGOOD) 'if' block */

	    else if
	    (txstatus == IO_TIMEOUT)
	    {
		if (trace[line] & (GENERALCOMMTRACE | MASTERCOMMTRACE | REMOTECOMMTRACE))
			logMsg ("Transmit Timeout\n");
		radio->no_txnoresp++;
	    }
	    else
	    {
		if (trace[line] & (GENERALCOMMTRACE | MASTERCOMMTRACE | REMOTECOMMTRACE))
			logMsg ("Transmit Error\n");
		radio->no_txhwerrs++;
	    }
	if (retry != MAXRETRYS)
		radio->no_retries++;
	}
        /* after all retries have been exhausted, we give up and exit with error status */
        if ((trace[line]) & (type))
                printf("%3d %3d\n", delta12, delta23);
	if (trace[line] & (GENERALCOMMTRACE | MASTERCOMMTRACE | REMOTECOMMTRACE))
		logMsg ("Xfer Bad\n");
	return (XFERBAD);

    }
/* end of mdsmsgout */


/*
/*      mdsmsgxfer                                                             	
/*                                                                           	
/*      The purpose of this routine is to transmit a buffer to the         	
/*      analog devices I/O and to await a response. The transmit and receive	
/*      buffers are in their respective line tables. This routine will retry 	
/*      the message up to a maximum number of times or until a response      	
/*      is received with a proper checksum.                                  	
/*              returns         XFERGOOD if response received                	
/*                              XFERBAD  if no good response is received     	
/*                                                                           	
*/
/* called from 						*/

static	unsigned short  mdsmsgxfer (unsigned short fline, 
				 MDSRADIO *radio, unsigned short rxlength,
                                 unsigned short checksumflag, unsigned long time, unsigned short type)
    {
	unsigned short  retry, txstatus, rxstatus, txlen, rxlen, line, ix;
	RXLINE          *rxp;
	TXLINE          *txp;
	unsigned char   csums[2];
        unsigned short  checksumpass;
        unsigned int    tick1, tick2, tick3, tick4;
        unsigned int    delta12, delta23, delta34;

        tick1 = tick2 = tick3 = tick4 = 0;     /* pre-zero all time keep registers */
        delta12 = delta23 = delta34 = 0;

	line = (fline & (~ONERETRY));          /* strip off the no retry flag*/

	rxp = rxlinetab[line];
	txp = txlinetab[line];

	for (retry=((fline == line)?0:MAXRETRYS-1); (retry<MAXRETRYS); retry++ )
	{
	    if (intermsgdelay[line] != 0)
	    {
		taskDelay (intermsgdelay[line]);
	    }
	    set_rts(line);

	    if (pretxdelay[line] != 0)
	    {
		taskDelay (pretxdelay[line]);
	    }
	    txstatus = IO_TIMEOUT;
	    rxp->rxmsglen = rxlength;
	    rxstatus = IO_TIMEOUT;
	    semTake(semids[line][TXSEM], NO_WAIT);  /* take it 'on behalf' of tx interrupt */
	    semTake(semids[line][RXSEM], NO_WAIT);  /* take it 'on behalf' of rx interrupt */

            set_release_time (line, time);          /* adjust 'release channel' semaphore 'oneshot' timer.*/

	    txlen = txp->txmsglen;

            if ((trace[line]) & (type))
            {
               printf("TX:");
               for (ix = 0; ix < 12 ; ix++) /* 12 is maximum TX message length */
               {
                    if (ix < txlen)
/*                       printf("_%2x", (char *) txp->txbuf[ix]); */
                       printf("_%2x", txp->txbuf[ix]);
                    else
                       printf("...");
               } 
               printf(" ");
            }

            tick1 = tickGet();
            start_rx(line, semids[line][RXSEM], &rxstatus); /* be ready to receive immediately. */
	    start_tx(line, semids[line][TXSEM], &txstatus); /* this fires off the transmission. */

	    /* 'wait' here for tx_int_proc to complete and release the semaphore. */
            if ((semTake(semids[line][TXSEM], txp->txtimeout)) == ERROR)
		txstatus = IO_TIMEOUT;

            tick2 = tickGet();
            delta12 = (tick2-tick1)%10000;

            stop_tx(line);

	    if (posttxdelay[line] != 0)
	    {
		taskDelay (posttxdelay[line]);
	    }
	    if (halfduplex[line] = TRUE)
	       clear_rts(line);    /* half duplex rts line control */

            tick3 = tickGet();
            delta23 = (tick3-tick2)%10000;


	    if (txstatus == TXGOOD)
	    {
                /* 'wait' here for rx_int_proc to complete and release the semaphore.  */
		if ((semTake(semids[line][RXSEM], rxp->rxtimeout)) == ERROR)
		    rxstatus = IO_TIMEOUT;

		stop_rx(line);
                tick4 = tickGet();
                delta34 = (tick4-tick3)%10000;

		rxlen = rxp->rxix;

		if (rxstatus == RXGOOD)
		{
		    checksumpass = FALSE;

                    if ((trace[line]) & (type))
                    {
                       printf("RX:");
                       for (ix = 0; ix < 24 ; ix++) /* 24 is maximum RX message length */
                       {
                            if (ix < rxlen)
/*                                printf("_%2x", (char *) rxp->rxbuf[ix]); */
                                printf("_%2x", rxp->rxbuf[ix]);
                            else
                                printf("...");
                       } 
	               printf(" ");
 	            }

                    if (rxlen != rxlength) 
		    {
		        if (trace[line] & (GENERALCOMMTRACE | MASTERCOMMTRACE | REMOTECOMMTRACE))
                        {
                                /* note the error, but fall through to retry logic loop */
                                logMsg("Invalid rxlen...    ");
                                printf ("Expected = %d, Actual = %d\n", rxlength, rxlen);
			        radio->no_rxlenerrs++;
                        }
		    }
		    else
		    {
                        if (checksumflag)
                        {
                                /* MDS has a one byte checksum */
				makechecksum(&rxp->rxbuf [0], rxlen-2, csums);
                                if (csums[0] != rxp->rxbuf[rxp->rxix-2])
                                {
                                        checksumpass = FALSE;
                                        radio->no_crcerrs++;
		                        if (trace[line] & (GENERALCOMMTRACE |
                                                            MASTERCOMMTRACE |
                                                            REMOTECOMMTRACE))
                                        {
                                           logMsg ("Invalid Checksum ");
                                           printf ("Received = %x, Calculated = %x\n",
					          rxp->rxbuf[rxp->rxix-2],
					          csums [0]);
                                        }
                                }
                        }
                        else
                        {
                                checksumpass = TRUE;
                        }
		    }
                    /* if checksum, if present, is good, and the start & stop bytes are */
                    /* the fixed values they should be, then that's the best we can do. */
		    if ( (checksumpass) &&
                         (rxp->rxbuf[0]  == MDSMSGSTARTBYTE) &&
                         (rxp->rxbuf[rxp->rxix-1] == MDSMSGENDBYTE) )
		    {
                        if ((trace[line]) & (type))
                                printf("%3d %3d %3d\n", delta12, delta23, delta34);
			return (XFERGOOD);
		    }
		    else
		    {
			radio->no_dataerrs++;
		        if (trace[line] & (GENERALCOMMTRACE | MASTERCOMMTRACE | REMOTECOMMTRACE))
                        {
				logMsg ("Bad Message Response. Length = %d\n", rxlen);
				printf ("Message: ");
                                for (ix = 0; ix < rxlen; ix++)
                                {
/*                                        printf("[%x]", (char *) rxp->rxbuf[ix]); */
                                        printf("[%x]", rxp->rxbuf[ix]);
			        }
	                        printf("\n");
		        }
                    } /* end of (rxstatus == RXGOOD) 'if' block */
                }
		else if
		(rxstatus == IO_TIMEOUT)
		{
		    if (trace[line] & (GENERALCOMMTRACE | MASTERCOMMTRACE | REMOTECOMMTRACE))
		    	logMsg ("Receive Timeout... Length: %d\n", rxlen);
		    radio->no_rxnoresp++;
		}
		else
		{
		    if (trace[line] & (GENERALCOMMTRACE | MASTERCOMMTRACE | REMOTECOMMTRACE))
                    {
		    	logMsg ("Receive Error\n");
		    	logMsg ( "RXHWERR: %2x, RXCHR: %2x, RXCNT: %2x\n",
				 rxp->rxbuf[100],
				 rxp->rxbuf[101],
				 rxp->rxix);
			radio->no_rxhwerrs++;
                    }
		}
	    } /* end of (txstatus == TXGOOD) 'if' block */
	    else if
	    (txstatus == IO_TIMEOUT)
	    {
		if (trace[line] & (GENERALCOMMTRACE | MASTERCOMMTRACE | REMOTECOMMTRACE))
			logMsg ("Transmit Timeout\n");
		radio->no_txnoresp++;
	    }
	    else
	    {
		if (trace[line] & (GENERALCOMMTRACE | MASTERCOMMTRACE | REMOTECOMMTRACE))
			logMsg ("Transmit Error\n");
		radio->no_txhwerrs++;
	    }
	if (retry != MAXRETRYS)
		radio->no_retries++;
	}
        /* after all retries have been exhausted, we give up and exit with error status */
        if ((trace[line]) & (type))
                printf("%3d %3d %3d\n", delta12, delta23, delta34);

	if (trace[line] & (GENERALCOMMTRACE | MASTERCOMMTRACE | REMOTECOMMTRACE))
		logMsg ("Xfer Bad\n");
	return (XFERBAD);

    }
/* end of mdsmsgxfer */


/*
/*      set_release_time                                                        
/*                                                                              
/*      The purpose of this routine is to set the release_time[] variable for   
/*       proper give_the_semaphore oneshot behavior.                            
/*                                                                              
*******/
/* called from mdsmsgout and mdsmsgxfer  */


static	void set_release_time (unsigned short line, unsigned long delay)
{
    unsigned long time;

    time =  tickGet() + delay;	/* mirror the Master radio delay	*/
    if (release_time[line] <= time)
       release_time[line] = time;
}
/* end of set_release_time */




/*                                                                                      
/*      init_line_tab                                                                   
/*                                                                                      
/*      The purpose of this routine is to initialize the line tables for the            
/*      individual radio on the line.                                                   
/*                                                                                      
*******/
/* only called from allocate_radio  */

static	void init_line_tab (unsigned short line, MDSRADIO *radio)

    {
	radio->is_responding = FALSE;
	radio->holdflag = FALSE;
	radio->was_requested = FALSE;
	radio->retry_count = 0;
	radio->offline = FALSE;
	radio->cmd_force_offline = FALSE;

	init_radio(radio);
	init_stats(radio);

	linetimers[line] = gettimer();
	if (linetimers[line] == -1)
	{
		/*  put an error message in here */
	}
	/* Initialize this tasks semaphore */
	semids[line][0]=semBCreate(SEM_Q_PRIORITY, SEM_EMPTY);
	semids[line][1]=semBCreate(SEM_Q_PRIORITY, SEM_EMPTY);
    }
/* end of int_line_tab */




/*											
/* Cold Init an Mds Radio								
/*											
*******/
/* only called from init_line_tab  */


static void init_radio (MDSRADIO  *radio)
    {
       	radio->was_requested = FALSE;
        mark_not_valid (radio);

    }
/* end of init_radio */



/*                                                            		
/*                                                              	
/* initialize the statistics for an Mds Radio	                  	
/*                                                              	
/*
*******/                                                               	
/* only called from init_line_tabs  */


static	void init_stats (MDSRADIO *radio)
    {
	radio->no_dataout_msgs = 0;  /* 5  Count Output Msgs 		*/
	radio->no_datain_msgs = 0;   /* 6  Count Data Requests 		*/
	radio->no_rxnoresp = 0;      /* 7  Count No Responses 		*/
	radio->no_txnoresp = 0;      /* 8  Count No Responses 		*/
	radio->no_dataerrs = 0;      /* 9  Count Message Echo Errors 	*/
	radio->no_retries = 0;       /* 10 Count Retries 		*/
	radio->no_rxhwerrs = 0;      /* 11 Count Comm Hardware Errors	*/
	radio->no_txhwerrs = 0;      /* 12 Count Comm Hardware Errors 	*/
	radio->no_crcerrs = 0;       /* 13 Count CRC Errors 		*/
	radio->no_chksumerrs = 0;    /* 13 Count Checksum Errors 	*/
	radio->no_offlines = 0;      /* 14 Count Off Line Transitions 	*/
	radio->badoutcnt = 0;	     /* 17 Count failed output commands	*/
        radio->no_rxlenerrs = 0;     /* 27 Count Response Length Errors */ 
        radio->no_chksumerrs = 0;    /* 28 Count Checksum errors        */
    }
/* end of init_stats */


/*                                                                           	
/*     prepresponses
/*                                                                             	
/*      The purpose of this routine is to mark the Radios as on hold and the   	
/*	data as not valid until read again from the Radio                      	
/*                                                                           	
/*
*******/
/* only called from MAIN */

static	void prepresponses (unsigned short line)
    {
	unsigned short  radionum;
	MDSRADIO  *radio;

	for (radionum = 1; (radionum <= NUMMDSRADIOS); radionum++)
	{
	    radio = radios[line][radionum];
	    if (radio)				/* verify pointer validity prior to use */
	    {
		radio->holdflag = TRUE;
		mark_not_valid(radio);
            }
	}
    }
/* end of prepresponses */



/*                                                                  	
/*      any_responding                                               	
/*                                                                   	
/*      The purpose of this routine is to see if any units are       	
/*	responding on this line                                      	
/*                                                                   	
/*
*******/
/* only called from MAIN */

static	int any_responding (unsigned short line)

    {
	unsigned short  radionum;
	MDSRADIO  *radio;
        unsigned short requested;

        requested = 0;
	for (radionum = 1; (radionum <= NUMMDSRADIOS); radionum++)
	{
	    radio = radios[line][radionum];
	    if (radio)			/* verify pointer validity prior to use */
	    {
		if (radio->was_requested)
			 requested++;;
                /* check cmd_force_offline flag because it is set in an asynchronous task */
                if ((radio->is_responding) && (!radio->cmd_force_offline))
			 return(TRUE);
	    }
	}
	if (requested == 0)
        {
           return(TRUE); /* don't report an error if none are requested */
        }
        else return(FALSE);
    }
/* end of any_responding */

/*
/*	Load MDS Address Translation Mapping File                                
/*                                                                               
*/                                                                 	                
/* only called from MAIN */

static int loadmapfile (unsigned short line)
{
        FILE                *handle;
        LOADMAP             *mapinfo;
        char		    filename[30];
        char		    addr[8];
        int 		    status;
/*	long		    pos; */
        char                nextchar, *pnextchar;
        unsigned short      i, state;
        unsigned int        hardware_address, logical_address;
        unsigned short      entry_count;

        pnextchar = &nextchar;

	mapinfo = maptable[line];	/* establish the pointer.   */
	/* If already one being loaded, something is wrong.         */
	if (mapinfo->loading)
		return(FALSE);
	mapinfo->line = line;	        /* easy way to find it      */
	mapinfo->loading = TRUE;

    /* first, initialize the array to all null/invalid values */
    for (i=0; (i<=NUMMDSRADIOS) ;i++)
    {
        mapinfo->address[i].hardware = NULL;
        mapinfo->address[i].valid_flag = FALSE;
    }

	/* Make File Name	*/
	strcpy(mapinfo->filename,"MdsLine");
	addr[0] = '0'+(line/10);	/* append Line Number       */
	addr[1] = '0' + (line % 10);
	addr[2] = 0;
	strcat(mapinfo->filename,addr);
	strcat(mapinfo->filename,".map");


	/* Open File existing file for Read Only */
	iam(user,password);
	if ((mapinfo->mapfile = fopen(&mapinfo->filename[0],"r")) == NULL)
	{
	    terminateload(mapinfo, line ,FALSE, CANNOTOPENFILE);
	    return(FALSE);
	}

        entry_count = 0;

        handle = mapinfo->mapfile;
        state = START;
        nextchar = getc(handle);    /* get first character in file */

        while (TRUE)
        {
/*       printf ("\nState = %x, Line = %x, Character = %c  ", state, entry_count, nextchar); */
            switch (state)
            {
               case START:
                    /* ignore spaces & tabs at beginning of line */
                    eatspaces(pnextchar, mapinfo->mapfile);
                    switch (nextchar)
                    {
                         case '\n':
                         case '#':
                              eatline(pnextchar, mapinfo->mapfile);
                              break;
                         case '(':
                              nextchar = getc(handle); /* eat the left paren */
                              eatspaces(pnextchar, mapinfo->mapfile); /* eat any spaces inside paren */
                              state = FIRSTPARAM;
                              break;
                         case EOF:
                              state = WRAPUP;
                              break;
                         default:
                              terminateload(mapinfo, line, FALSE, SYNTAXERROR);
                              return(FALSE);     /* syntax error */
                              break;
                    }
                    break;

               case FIRSTPARAM:
                    hardware_address = 0;
                    if (!((nextchar >= '0') && (nextchar <= '9')))
                    {
                         terminateload(mapinfo, line, FALSE, SYNTAXERROR);
                         return(FALSE);         /* syntax error */
                    }
                    while ((nextchar >= '0') && (nextchar <= '9'))
                    {
                        hardware_address = hardware_address*10 + (nextchar-'0');
                        if (hardware_address > 9999)
                        {
                           terminateload(mapinfo, line, FALSE, SYNTAXERROR);
                           return(FALSE);         /* syntax error */
                        }
                        nextchar = getc(handle);  /* eat that digit, get the next */
                    }
                    eatspaces(pnextchar, mapinfo->mapfile);
                    if (nextchar != ',')
                    {
                        terminateload(mapinfo,line,FALSE,SYNTAXERROR);
                        return(FALSE);         /* syntax error */
                    }
                    nextchar = getc(handle);  /* eat the comma */
                    eatspaces(pnextchar, mapinfo->mapfile);
                    state = SECONDPARAM;
                    break;

               case SECONDPARAM:
                    logical_address = 0;
                    if (!((nextchar >= '0') && (nextchar <= '9')))
                    {
                         terminateload(mapinfo,line,FALSE,SYNTAXERROR);
                         return(FALSE);         /* syntax error */
                    }
                    while ((nextchar >= '0') && (nextchar <= '9'))
                    {
                        logical_address = logical_address*10 + (nextchar-'0');
                        if (logical_address > NUMMDSRADIOS)
                        {
                           terminateload(mapinfo,line,FALSE,BADRADIOADDRESS);
                           return(FALSE);         /* radio address exceeded max value */
                        }
                        nextchar = getc(handle);  /* eat that digit, get the next */
                    }
                    eatspaces(pnextchar, mapinfo->mapfile);
                    if (nextchar != ')')
                    {
                         terminateload(mapinfo,line,FALSE,SYNTAXERROR);
                         return(FALSE);         /* syntax error */
                    }
                    eatline(pnextchar, mapinfo->mapfile); /* eat up the rest of the line */
                    state = STOREVALUE;
                    break;

               case STOREVALUE:
                    entry_count++;
                    mapinfo->address[logical_address].hardware = hardware_address;
                    mapinfo->address[logical_address].valid_flag = TRUE;
                    state = START;   /* process next line of input, if there is one. */
                    break;

               case WRAPUP:
                    if (trace[line] & RADIOMAPTRACE)
	            {
	                logMsg("Number of syntactically correct entries = %d\n", entry_count);
	                return;
	            }
                    if (mapinfo->address[MASTERADDRESS].valid_flag == TRUE)
                    {
                         terminateload(mapinfo,line,TRUE,LOADOK);
                         return(TRUE);
                    }
                    else
                    {
                         terminateload(mapinfo,line,FALSE,NOMASTERADDRESS);
                         return(FALSE);               /* failure to define master radio address */
                    }
                    break;
               default:
                         terminateload(mapinfo,line,FALSE,SYNTAXERROR);
                         return(FALSE);                /* syntax error */
                       break;
           }
        }
}
/* end of loadmapfile */


/*                                                                          
/*      terminateload                                                       
/*	Abort the Address Translation Map Load and Report Errors                
/*                                                                          
*/                                                                                              
/* called from loadmapfile */

static void terminateload (LOADMAP *mapinfo,unsigned short line,
	       unsigned short success, unsigned short code)
{
    /* Close The File		*/
    fclose(mapinfo->mapfile);
    mapinfo->loading = FALSE;
    mapinfo->map_load_status = success;
    mapinfo->map_load_code = code;

    return;
}
/* end of terminateload */



/*                                                                          
/* eatspaces                                                                
/*        entered with ch containing first character under consideration    
/*        exit with ch pointing to next character to be considered.         
/*                                                                          
 */
/* only called from readmapfile */

static void eatspaces(char *ch, FILE *handle)
{
    while ((*ch == ' ') || (*ch == '\t'))
    {
        *ch = getc(handle);
    }
    return;  /* only leave when non-space character encountered */
}



/*                                                                          
/* eatline                                                                  
/*        entered with ch containing first character under consideration    
/*        exit with ch pointing to next character to be considered.         
/*                                                                          
 */
/* only called from readmapfile */

static void eatline(char *ch, FILE *handle)
{
    while ((*ch != '\n') && (*ch != EOF))
    {
        *ch = getc(handle);
    }
    while (*ch == '\n') 
    {
        *ch = getc(handle);
    }
    return; /* only leave when non-end-of-line or end-of-file encountered */
}
/*
/*      TABLEINFO                                                       	
/*                                                                           	
/*      The purpose of this routine is to dump the radio state description     	
/*      portion of the MDS driver tables for the selected line.                 
/*      It will print out two lines for each radio which is responding.         
/*      The first line will give the general radio information, the second      
/*      line will give radio-specific information.                              
*/
/* not called from within this driver */

void TABLEINFO (unsigned short line, unsigned short radioaddress)
    {
	MDSRADIO        *radio;
	short           ix;
	unsigned	short	i, n, j; 

	XXXBIS		*biptr;
	XXXAIS		*aiptr;
	unsigned	short   address, linecnt;
	unsigned	char	buffer[200];

	radio = radios[line][radioaddress];
	if (!radio)
	{
	    logMsg("The Radio Address for this Line is Not Defined \n");
	    return;
	}

        printf ("    Was requested       =%6d\t",                              radio->was_requested);
	printf               ("    *Retry Count        =%6d\t",                radio->retry_count);
	printf                              ("    Reset Statistics    =%6d\n", radio->cmd_reset_stat);

        printf ("    Is responding       =%6d\t",                              radio->is_responding);
	printf               ("    *Offline flag       =%6d\t",                radio->offline);
	printf                              ("    Force Offline       =%6d\n", radio->cmd_force_offline);

        printf ("    Resource Enabled    =%6d\t",                              resresponse[line]);
	printf               ("    *Hold Flag          =%6d\t",                radio->holdflag);
	printf                              ("    # Off Lines         =%6d\n", radio->no_offlines);
	printf ("\n");

	printf ("    # Data out messages =%6d\t",                              radio->no_dataout_msgs);
	printf               ("    # No Tx Responses   =%6d\t",                radio->no_txnoresp);
	printf                              ("    # No Rx Response    =%6d\n", radio->no_rxnoresp);

	printf ("    # Data in messages  =%6d\t",                              radio->no_datain_msgs);
	printf               ("    # Tx Hw Errors      =%6d\t",                radio->no_txhwerrs);
	printf                              ("    # Rx Hw Errors      =%6d\n", radio->no_rxhwerrs);

	printf ("    # Retries           =%6d\t",                              radio->no_retries);
	printf               ("    # Failed Outputs    =%6d\t",                radio->badoutcnt);
	printf                              ("    *# RX Length Errors =%6d\n", radio->no_rxlenerrs);

	printf ("                                 \t"                                            );
	printf               ("                                  \t"                             );
	printf                              ("    # Crc Errors        =%6d\n", radio->no_crcerrs);

	printf ("                                 \t"                                            );
	printf               ("                                 \t"                              );
	printf                              ("    # Total Data Errors =%6d\n", radio->no_dataerrs);


	printf ("\n");
/*	printf ("\n"); */
/*	printf ("    *Data valid         = %d \n", radio->data_valid); */

	linecnt = 1;
	printbihdr(buffer, linecnt);
	address = 0;
	for (i=0; i < (NUMMDSBISXXX); i += 4)
	{
		printf("   %4d:       ", address);
		for (n = 0; n < 4; n++)
		{		
                        if ( (i+n)  >=  NUMMDSBISXXX )
                           break;
               		biptr = &radio->bismds[i+n];
		        printf("%1x  %1x   %2x       ",
		                biptr->was_requested, biptr->data_valid, biptr->last_read);
                }
		printf("\n");
		linecnt += 1;
		if ( (linecnt >= 17) && ((i + 1)  < (NUMMDSBISXXX)) )
		{
			linecnt = 1;
			printbihdr(buffer, linecnt);
		}
		address += 4;
	}
	linecnt = 1;
	printaihdr(buffer, linecnt);
	address = 0;
	for (i=0; i < (NUMMDSAISXXX); i += 4)
	{
		printf("   %4d:       ", address);
		for (n = 0; n < 4; n++)
		{		
                        if ( (i+n)  >=  NUMMDSAISXXX )
                           break;
                        aiptr = &radio->aismds[i+n];
                        printf("%1x  %1x   %2x       ",
		                aiptr->was_requested, aiptr->data_valid, aiptr->last_read);
                }
		printf("\n");
		linecnt += 1;
		if ( (linecnt >= 17) && ((i + 1)  < (NUMMDSAISXXX/8)) )
		{
			linecnt = 1;
			printaihdr(buffer, linecnt);
		}
		address += 4;
	}

    }

/*
/* only called from TABLEINFO 
*/
static void printbihdr (unsigned char *buffer, unsigned short cnt)
{	
	taskDelay(180);
/*	printf("\n"); */
	sprintf(buffer,
                  "----------------------- MDS Radio Report - Digital Inputs ----------------------");
	printf( "%s\n", buffer);
	cnt += 1;
        sprintf(buffer,
                "BI Address:    R  V  Value     R  V  Value     R  V  Value     R  V  Value");
 	printf( "%s\n", buffer);
 	cnt += 1;
}


/*
/* only called from TABLEINFO 
*/
static void printaihdr (unsigned char *buffer, unsigned short cnt)
{
	taskDelay(180);
/*	printf("\n"); */
	sprintf(buffer,
                  "----------------------- MDS Radio Report - Analog Inputs -----------------------");
	printf( "%s\n", buffer);
	cnt += 1;
        sprintf(buffer,
                "AI Address:    R  V  Value     R  V  Value     R  V  Value     R  V  Value");
 	printf( "%s\n", buffer);
 	cnt += 1;
}
/*
/*      SHOWMAP                                                                 
/*                                                                           	
/*      The purpose of this routine is to dump the logical address to          	
/*      hardware address translation map for the master and remote radios.      
/*      It will print out an entry for the master radio listing the address 	
/*      of the master radio on the comm line. The table then lists all 100      
/*      remote radio logical addresses, its status(valid), and the DTMF value.  
*/
/* not called from within this driver */

void SHOWMAP (unsigned short line)
    {
	LOADMAP         *mapinfo;
	unsigned	short	i, n; 
	unsigned	short   address, linecnt;
	unsigned	char	buffer[200];

	mapinfo = maptable[line];	/* establish the pointer.   */

        if (!mapinfo)
	{
	    logMsg("Address Map Table not allocated. Pointer not valid. \n");
	    return;
	}
        if (mapinfo->address[0].valid_flag != TRUE)
	{
	    logMsg("No valid address map table entry for Master Radio. \n");
	    return;
	}
	printf ("\n");
	printf ("Master Radio Hardware Address = %d \n", mapinfo->address[0].hardware);
	printf ("\n");

	linecnt = 1;
	printmaphdr(buffer, linecnt);
	for (i = 1; i < (NUMMDSRADIOS); i += 10)
	{
		for (n = 0; n < 10; n++)
		{		
                        if ( (i+n)  >  NUMMDSRADIOS )
                           break;
                        if (mapinfo->address[i+n].valid_flag == TRUE)
                        {
                                printf("  %3d:%04d", (i+n), mapinfo->address[i+n].hardware);
                        }                
                        else
                        {
		                printf("  %3d:----", (i+n), mapinfo->address[i+n].hardware);
                        }
                }
		printf("\n");
		linecnt += 1;
		if ( (linecnt >= 17) && ((i + 1)  < (NUMMDSRADIOS)) )
		{
			linecnt = 1;
                        taskDelay(4*TICKSPERSEC);
			printmaphdr(buffer, linecnt);
		}
	}

    }

/*
/* only called from SHOWMAP 
*/
static void printmaphdr (unsigned char *buffer, unsigned short cnt)
{
	printf("\n");
	sprintf(buffer,
"----------------------- MDS Radio Report - Logical:Hardware Address Mapping ------------------------");
	printf( "%s\n", buffer);
	cnt += 1;
        sprintf(buffer,
"  Log:Hrdw  Log:Hrdw  Log:Hrdw  Log:Hrdw  Log:Hrdw  Log:Hrdw  Log:Hrdw  Log:Hrdw  Log:Hrdw  Log:Hrdw");
 	printf( "%s\n", buffer);
 	cnt += 1;
	printf ("\n");
	cnt += 1;
}

/*
**                                                                	 	
**	The purpose of this routine is to display the contents            	
**	of the static variables contained within this driver.             	
**                                                                        	
*/
/* this routine not called from within this driver */

void LINEINFO ()
{
	short	i;
	char	buffer[300];
	char	*resrsp, reschar[5];
	char	*lineact, linechar[5];

        resrsp = &reschar[0];
        lineact = &linechar[0];

printf(
"---------------------------------- Mds Radio Line Information Table --------------------------------");
printf("\n");
printf(
"Line  Resource   Resource   Line    Delay Times: Prime Prime LineMux     Next   Next   TRACE");
printf("\n");
printf(
"Num     Name     Response Activity  IM  pTX TXp  Line  State Semaphore   Radio  Offln  BITS    Qin  Qout");
printf("\n");
	for (i = 0 ; i < MDSLINES; i++)
	{
		if(prestable[i] == NULL)          /* check for an initialized resource table */
		{
			continue;
		}
		if (resresponse[i] == 0)
		{
			resrsp = "FALSE";
		}
		else
		{
			resrsp = "TRUE ";
		}
		if (line_activity[i] == 0)
		{
			lineact = "FALSE";
		}
		else
		{
			lineact = "TRUE ";
		}
		sprintf(buffer, 
/*                "%2d %10s  %5s     %5s   %2d  %2d  %2d    %4d   %4d %6d %8d  %2d %2d %2d %2d", */
                " %2d   %10s  %5s    %5s    %2d  %2d  %2d  %4d  %4d    %8d %6d  %3d      %3x    %3x   %3x",
			i,
                        &resnames[i],
			&resrsp[0],
			&lineact[0],
			intermsgdelay[i],
			pretxdelay[i],
			posttxdelay[i],
			prime_line[i],
                        prime_state[i],
                        linemux[i],
                        nextradio[i],
                        nextoffline[i],
                        trace[i],
                        oq[i]->oinix,
                        oq[i]->ooutix);
		printf( "%s\n", buffer); 
	}
	return;
}



/*	End of File: drvMds.c */

