/*drvModiconslv.h,v 1.1.1.1 1999/08/03 14:43:22 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
**	File: drvModiconslv.h							       	
**
**	This file contains the data structures used by the Modicon Modbus	       	
**	I/O driver.                                                            	       	
**                                                                             	      	
*/

#define TRUE	1
#define FALSE	0
#define NULLP	0
#define CR	0x0d

/*	number of lines of modbus 	*/
#define		MODLINES	32
/*	number of modbus plcs per line */
#define		NUMMODPLCS	10

/*      number of groups which can be defined in the PCU/LGU */
#define         MAXGROUPNUM     16

#define         RES_LOCKED       0

/*                                                                           		
*	The IOC is emulating a Modicon PLC
*		The IOC will write to all 4 memory types AIN,REG,DIN,Coil
*		It will serve this data to the master and accept write commands
*		to the REG and COIL address space. The size of the address space
*		supported and its starting memory address is specified below 
*/

/* Number of Groups of items */

#define NUMCOILS		2048			/* 2048 coils				*/
#define NUMDINS		2048			/* 2048 dins				*/
#define NUMAINS		1024			/* 1024 AINs				*/
#define NUMREGS		1024			/* 1024 Regs				*/
#define COILSTART		0			/* Start Address of Coil Space	*/
#define DINSTART		0			/* Start Address of Din Space		*/
#define AINSTART		0			/* Start Address of Ain Space		*/
#define REGSTART		0			/* Start Address of Reg Space		*/



/* The card field of the database blocks will contain the following constants
*  to addrewss the various typres of memory
*/

#define MCOILS		0			/* Card number for coils 		*/
#define MDINS		1			/* Card number for digital inputs 	*/
#define MAINS		3			/* Card number for analog inputs 	*/
#define MREGS		4			/* Card number for registers 		*/
#define MPSEUDO		33			/* Card number for pseudo i/o 		*/


						/*  to request in one packet        	*/

/* The following must match the definition in recChoice.ascii */

/*
**	Conversion types for analog inputs
**
*/
#define		AICNVINTEGER	10
#define		AICNVDOUBLEINT	11
#define		AICNVFLOAT		12
#define		AICNVDOUBLEFLT	13
#define		AICNVASCIIFLT	14
#define		AICNVSTATUS		15
#define		AICNVTIME		16
#define		AICNVASCII_2	17
#define		AICNVASCII_4	18
#define		AICNVASCII_12	19
#define		AICNVASCII_20	20
#define		MAXCONV		20

#define NOCONV 0
#define LINEAR 1
#define ASCFLT 2


#define    WATCHDOGTIMEOUT       10*60 /* watchdog monitor timeout.  */


/*
**	Communications I/O Structures
**
*/

#define		RXSEM		0
#define		TXSEM		1
#define		MAXRXMSG	256
#define		MAXTXMSG	256


typedef	struct {
	unsigned short		rxstate;
	unsigned short		fixedlen;
	unsigned short		rxmsglenexpected;
      unsigned short          rxmsglen;
      unsigned short		rxix;
	SEMAPHORE			*semid;
	unsigned short		*status;
      unsigned char           txaddress;
      unsigned char           txfunction;
      unsigned char		rxbuf[MAXRXMSG];
      unsigned short          rxwinix;
      unsigned char           rxwindow[MAXRXMSG*2];
		} 	RXLINE;


typedef	struct {
	unsigned short		protocol;
	unsigned short		txdisable;
	unsigned short		txtocnt;
	unsigned short		txhwerrcnt;
	unsigned short		txmsglen;
	unsigned short		txix;
	SEMAPHORE			*semid;
	unsigned short		*status;
	unsigned char		txbuf[MAXTXMSG];
		} 	TXLINE;
/*
**	Communications I/O Constants
**
*/

#define		RXGOOD		1
#define		RXBAD			2
#define		RXTIMEOUT		60		/* 16 ms ticks 				*/
#define		ACTIVITYTIME	300		/* must see activity in 5 seconds 	*/

#define		VARRESP		FALSE
#define		FIXEDRESP		TRUE
#define		ONERETRY		0x8000	/* OR'ed into line number for xfermsg 	*/


/* Interrupt State Machine Constants */


#define		RXIGNORE          0
#define		RXADRST           1
#define		RXFNCST           2
#define		RXSTARTADDRHI     3
#define		RXSTARTADDRLO     4
#define		RXLENHI           5
#define		RXLENLO           6
#define		RXDATAST		7

#define		TXGOOD		1
#define		TXBAD			2
#define		IO_TIMEOUT		3	/* Code returned by timer 		*/


#define		TXTIMEOUT		60	/* 16 ms ticks: one second              */

#define		MAXRETRYS		3	/* retransmission trys 			*/

#define		XFERGOOD		1
#define		XFERBAD		2

#define		RESNAMELEN		16	/* length of a resource name 		*/

/*
** Output Queue Management Structures for Modbus Protocol
**
*/

#define         AOUTTYPE        1
#define         BOUTTYPE        2
#define         SOUTTYPE        3

#define         OUTQSIZE        128
#define	    STRINGMSGLEN	  80

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  aosignal;
	short           aovalue;
	double	    dblaovalue;
	unsigned short  conversion;
		} AOUT, *AOUTPTR;

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  bosignal;
	long            bomask;
	long            bovalue;
		} BOUT, *BOUTPTR;

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  sosignal;
	long            solen;
	unsigned char   sovalue[STRINGMSGLEN];
		} SOUT, *SOUTPTR;

typedef struct {
		unsigned short  otype;
		union   {
			AOUT    aonode;
			BOUT    bonode;
			SOUT	sonode;
			}out;
	       } OUTQENT;

typedef struct {
		unsigned short	ooutix;
		unsigned short	oinix;
		OUTQENT         outque[OUTQSIZE];
		} OUTPUTQUE;
/*
**	Internal IOC Structure for MODICON PLC
**
*/

typedef	struct	{
	unsigned short  was_requested;		/* 0  The scan task has accesses this RTU 	*/
	unsigned short  is_responding;		/* 1  The RTU is responding 			*/
	unsigned short  holdflag;		/*    The RTU data is in hold 			*/
	unsigned short  no_dataout_msgs;	/* 5  Count Output Msgs 			*/
	unsigned short  no_datain_msgs;		/* 6  Count Data Requests 			*/
	unsigned short  no_rxnoresp;		/* 7  Count No Responses 			*/
	unsigned short  no_txnoresp;		/* 8  Count No Responses 			*/
	unsigned short  no_dataerr;		/* 9  Count Message Echo Errors 		*/
	unsigned short  no_retries;		/* 10 Count Retries 				*/
	unsigned short  no_txhwerrs;		/* 12 Count Comm Hardware Errors 		*/
	unsigned short  no_crcerrs;		/* 13 Count CRC Errors 				*/
	unsigned short  cmd_reset_stat;		/* 15 Comand to reset stats			*/
	unsigned short  cmd_force_offline;	/* 16 Command to Force Offline 			*/
	unsigned char	coils[NUMCOILS/8];
	unsigned short	coilsdef[NUMCOILS/8];
	unsigned short	coilsvalid[NUMCOILS/8];
	unsigned char	dis[NUMDINS/8];
	unsigned short	dinsdef[NUMDINS/8];
	unsigned short	dinsvalid[NUMDINS/8];
	unsigned short	ais[NUMAINS];
	unsigned short	ainsdef[NUMAINS];
	unsigned short	ainsvalid[NUMAINS];
	unsigned short	regs[NUMREGS];
	unsigned short	regsdef[NUMREGS];
	unsigned short	regsvalid[NUMREGS];
	}	MODPLC;
/*
** MODBUS Protocol Message Idecies
*/
#define MSGADDR		0
#define MSGFUNC		1
#define MSGSTARTHI	2
#define MSGSTARTLO	3
#define MSGLENHI		4
#define MSGLENLO		5
#define MSGDIAGFUNCHI   2
#define MSGDIAGFUNCLO   3
#define MSGDIAGDATAHI   4
#define MSGDIAGDATALO   5
/*
** MODBUS Protocol function codes
*/
#define MODCOILREAD		0x01  /* READ COIL STATUS: (ON/OFF) read a group of logic coils.     */
#define MODDIREAD		0x02  /* READ INPUT STATUS:(ON/OFF) of a group of discrete inputs.   */
#define MODREGREAD		0x03  /* READ HOLDING REGISTERS:    get current binary value in one  */
                                      /*                            or more holding registers.       */
#define MODAIREAD		0x04  /* READ INPUT REGISTERS:      get current binary value in one  */
                                      /*                            or more input registers.         */

#define MODCOILWRITE		0x05  /* FORCE SINGLE COIL:         force logic coil to ON/OFF state.*/
#define MODREGWRITE		0x06  /* PRESET SINGLE REGISTER:    place a specific binary value    */
                                      /*                            into a holding register.         */

#define MODLOOPBACK		0x08  /* LOOPBACK DIAGNOSTIC TEST                                    */
#define MLOOPDATAQUERYFUNC    0x00  /* Loop back subfunction data query code */
#define MODMCOILWRITE		0x0F  /* FORCE MULTIPLE COILS:      force series of consecutive      */
                                      /*                            logic coils to ON/OFF states.    */
#define MODMREGWRITE		0x10  /* PRESET MULTIPLE REGISTERS: put specific binary values into  */
                                      /*                            consecutive series of holding    */
                                      /*                            registers.                       */

#define FNCERRBIT		0x80  /* Modbus error function code */


/* Various message lengths used in message transmission and response validation */
#define MODVARMGSOVHD		0x07
#define MODCRCLEN			0x02
#define MODWRITEMREGHLEN	0x07
#define MODWRITEMCOILHLEN	0x07
#define MODWRITEHDRLEN		0x04	/* wep: not currently used */
#define MODREADMSGLEN		0x06
#define MODSETCOILLEN		0x06
#define MODSETREGLEN		0x06
#define MODSET1REGLEN		0x09
#define MODSET2REGLEN		0x0b
#define MODSET3REGLEN		0x0d
#define MODSET4REGLEN		0x0f
#define MODMSETCOILLEN		0x07	/* wep: not currently used */
#define MODMSETCOILRESPLEN	0x06
#define MODSETCOILRESPLEN	0x06
#define MODCRCLEN		0x02
#define MODLOOPBACKLEN		0x06
#define MODLOOPBACKRESPLEN	0x08	/* Including CRC */
#define MODWRITERESPLEN		0x08	/* Including CRC */
#define MODSETMREGRESPLEN	0x06
#define MODSETREGRESPLEN	0x06
#define NAKMSGLEN		0x03	/* bytes left get */

/*
** MODBUS Exception Codes
*/

#define MODEXCEPTFNC		0x01
#define MODEXCEPTADR		0x02
#define MODEXCEPTDATA		0x03
#define MODEXCEPSLAVE		0x04
#define MODEXCEPTSACK		0x05
#define MODEXCEPTBUSY		0x06

/* Command Results */

#define CMDDONE                 0x00     /* the command was sent and the response was error free.   */
#define CMDFAIL                 0x01     /* there were errors in sending or receiving the command.  */
#define CMDNACK                 0x02     /* response was error free, but command was not completed. */
#define CMDREJECT               0x03     /* driver rejected command due to unsupported data type.   */
#define CMDBUSY			0x05	 /* The command was sent, but the device is busy */

/* task definitions */

#define MODSLV_STACK		8000   /* Stack size              */
#define	MODSLV_PRIORITY		152    /* default priority        */
#define	MODSLV_TASK_OPTIONS	8      /* Floating point          */

typedef struct {
        unsigned long   TXintcount;
        unsigned long   RXintcount;
        unsigned long   RXbadcount;
        unsigned long   msgxfercount;
        unsigned long   startingtick;
               }        LINESTATS;

/* Trace Bit Definitions */

#define ALLTRAFFICTRACE      0x0001
#define MODBUSNACKTRACE      0x0002        /* MODBUS NACK errors.                               */
#define CONFIGURATIONTRACE   0x0010        /* trace configuration errors in driver calls.       */
#define RUNTIMEERRORTRACE    0x0020        /* trace run time errors in driver calls.            */
#define CALLREJECTTRACE      0x0040        /* trace non-error based driver call rejections.     */


/*	End of File:	drvModiconslv.h */
