/*drvMr350.h,v 1.1.1.1 1999/08/03 14:43:23 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
**  File: drvMr350.h                                                         
**  Edit: 03/03/97      WEP: created based on drvModicon.h                   
**        09/04/97	WEP: added support for Resource Group Lock function  
**        10/28/97	WEP: added 0xff tx post-transmission logic.          
**        10/29/97	WEP: added 0xff tx pre-transmission logic.           
**
**	This file contains the data structures used by                       
**       the MR-350 Data Collection Terminal (DCT) I/O driver.               
**                                                                           
*/

#define TRUE             1
#define FALSE	         0
#define NULLP            0
#define CR            0x0d
#define TICKSPERSEC     60

/*	number of lines of MR-350 Drivers  */
#define         MR350LINES  32

/*	number of Data Collection Terminals (CDT) per line */
#define		NUMDCTS   3

/*      number of groups which can be defined in the PCU/LGU */
#define         MAXGROUPNUM     16

#define         RES_LOCKED       0

#define DINS            0               /* Card number for digital inputs      */
#define DOUTS	        1               /* Card number for digital outputs     */
#define LEDOUTS         2               /* Card number for LED outputs         */
#define LCDOUTS         3               /* Card number for LCD string outputs  */
#define STRINGINS       4               /* Card number for string inputs       */
#define BUZZEROUTS      5               /* Card number for buzzer outputs      */
#define SITEINS         6               /* Card number for site status         */
#define USERIDS         7               /* Card number for User ID             */
#define PERIMETERS      8               /* Card number for Perimeter BI/BO     */
#define NEWSTUFF        9               /* Card number for Miscellaneous Stuff */
#define NEWFILES        9               /* Card number for Miscellaneous Stuff */
#define DELAYS          9               /* Card number for Miscellaneous Stuff */
#define TIMEDATE        9               /* Card number for Miscellaneous Stuff */
#define USERSIN        10               /* Card number for logged in users     */
#define DETAILS        11               /* Card number for other details       */

#define MPSEUDO	       33               /* Card number for pseudo i/o          */
                                        
/* Max Number of Groups of items */
#define NUMDINS         4               /* Max number of digital inputs        */
#define NUMDOUTS        2               /* Max number of digital outputs       */
#define NUMLEDOUTS      4               /* Max number of LED outputs           */
#define NUMLCDOUTS      2               /* Max number of LCD string outputs    */
#define NUMSTRINGINS    2               /* Max number of string inputs         */
#define NUMBUZZEROUTS   4               /* Max number of buzzer outputs        */
#define NUMSITEINS      1               /* Max number of site status regs      */
#define NUMUSERIDS      1               /* Max number of User ID regs          */
#define NUMPERIMETERS   1               /* Max number of Perimeter BI/BOs      */
#define NUMNEWSTUFF     8               /* Max number of Miscellaneous Stuff   */
#define NUMUSERSIN     20               /* Max number of logged in users       */
#define NUMDETAILS      2               /* Max number of other details         */

#define NUMNEWFILES     1               /* Max number of Load New File BO      */
#define NUMTIMEDATE     1               /* Max number of Download Time/Date    */
#define NUMDELAYS       8               /* Max number of access delay values   */

/*
**	Conversion types for analog inputs	*/
**
*/
#define		AICNVINTEGER	10
#define		AICNVDOUBLEINT	11
#define		AICNVFLOAT	12
#define		AICNVDOUBLEFLT	13
#define		AICNVASCIIFLT	14
#define		AICNVSTATUS	15
#define		AICNVTIME	16
#define		AICNVASCII_2	17
#define		AICNVASCII_4	18
#define		AICNVASCII_12	19
#define		AICNVASCII_20	20
#define		MAXCONV		20

/* The following must match the definition in recChoice.ascii */
#define NOCONV 0
#define LINEAR 1
#define ASCFLT 2

#define    WATCHDOGTIMEOUT   10*60 /* watchdog monitor timeout.  */

/*
**	Communications I/O Structures
**
*/

#define		RXSEM		0
#define		TXSEM		1
#define		MAXRXMSG	128
#define		MAXTXMSG	128

typedef	struct {
	unsigned short		rxstate;
        unsigned short          rxspecial;
        unsigned short          rxmsglen;
       	unsigned short		rxix;
	unsigned short		*status;
        unsigned short          dummyshort;
        unsigned long           rawchksum;
       	unsigned char		rxbuf[MAXRXMSG];
	SEMAPHORE		*semid;
		} 	RXLINE;


typedef	struct {
	unsigned short		txff;
	unsigned short		txffcount;
	unsigned short		txstate;
        unsigned short          txspecial;
	unsigned short		txdisable;
	unsigned short		txmsglen;
	unsigned short		txix;
	unsigned short		*status;
        unsigned char           dummychar;
        unsigned char           chksum1;
        unsigned char           chksum2;
	unsigned char		address;
        unsigned long           rawchksum;
	unsigned char		txbuf[MAXTXMSG];
	SEMAPHORE		*semid;
		} 	TXLINE;
/*
**	Communications I/O Constants
*/

/* RX State Names */
#define  RXIGNORE    0
#define  RXSTX       1
#define  RXDATA	     2

/* RX Status Names */
#define	 RXGOOD	     1
#define	 RXBAD       2
#define	 RXACK       3
#define	 RXNAK       4
#define	 RXEOT       5


/* TX State Names */
#define  TXIGNORE    0
#define  PRETXSTX    1
#define  PRETXACK    2
#define  PRETXNAK    3
#define  TXSTX       4
#define  TXACK       5
#define  TXNAK       6
#define  TXDATA      7
#define  TXCHKSUM1   8
#define  TXCHKSUM2   9
#define  TXADDRESS  10
#define  TXFF       11


/* TX Status Names */
#define  IO_TIMEOUT  0
#define	 TXGOOD      1
#define	 TXBAD       2
#define	 TXBAD3      3
#define	 TXBAD4      4
#define	 TXBAD5      5
#define	 TXBAD6      6
#define	 TXBAD7      7
#define	 TXBAD8      8
#define	 TXBAD9      9
#define	 TXBAD10    10
#define	 TXBAD11    11
#define	 TXBAD12    12

/* TX FF State Names */
#define  PRETXACKFLAG    0x01
#define  PRETXNAKFLAG    0x02
#define  PRETXSTXFLAG    0x04
#define  POSTTXACKFLAG   0x08
#define  POSTTXNAKFLAG   0x10
#define  POSTTXADDFLAG   0x20




#define	    RXTIMEOUT          15   /* 16 ms ticks = 250 msec              */
#define     ACTIVITYTIME      300   /* must see activity in 5 seconds      */
#define     ONERETRY       0x8000   /* OR'ed into line number for xfermsg  */
#define     TXTIMEOUT          30   /* 16 ms ticks: 1/2 second * 2 = 1 sec */
#define     MAXRETRYS           3   /* retransmission trys                 */
#define     MAXCFGCK            8   /* Try to Bring online each cycle      */
#define     BRINGBACKCNT        4   /* Poll loops per cycle                */
#define     XFERGOOD            1
#define     XFERBAD             2
#define     RESNAMELEN         16   /* length of a resource name           */
#define     SLWWINDOW         900   /* # of ticks to allow slave to try    */


/*
** Output Queue Management Structures for MR-350 Unit Protocol
**
*/
#define         AOUTTYPE        1
#define         BOUTTYPE        2
#define         SOUTTYPE        3

#define         OUTQSIZE        128
#define		STRINGMSGLEN	80

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  aosignal;
	short           aovalue;
	double		dblaovalue;
	unsigned short	conversion;
		} AOUT, *AOUTPTR;

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  bosignal;
	long            bomask;
	long            bovalue;
		} BOUT, *BOUTPTR;

typedef struct {
    unsigned short  link;
    unsigned short  adapter;
    unsigned short  card;
    unsigned short  sosignal;
    long            solen;
    unsigned char   sovalue[STRINGMSGLEN];
               } SOUT, *SOUTPTR;

typedef struct {
     unsigned short  otype;
     union   {
                AOUT    aonode;
                BOUT    bonode;
                SOUT	sonode;
             }out;
	       } OUTQENT;

typedef struct {
    unsigned short	ooutix;
    unsigned short	oinix;
    OUTQENT             outque[OUTQSIZE];
               } OUTPUTQUE;





/*
**	Internal IOC Structure for MR350 DCT
**
*/
typedef	struct {
    unsigned short        was_requested;
    unsigned short        data_valid;
    unsigned short        new_flag;
    unsigned char         last_read;
               } DCTFIN;

typedef	struct {
    unsigned short        was_requested;
    unsigned short        data_valid;
    unsigned short        new_flag;
    unsigned char         last_read;
               } DCTDINS;

typedef	struct {
    unsigned short        was_requested;
    unsigned short        data_valid;
    unsigned char         last_write;
               } DCTDOUTS;

typedef	struct {
    unsigned short        was_requested;
    unsigned short        data_valid;
    unsigned char         last_write;
               } DCTLEDOUTS;

typedef	struct {
    unsigned short        was_requested;
    unsigned short        data_valid;
    unsigned char         last_write[81];
               } DCTLCDOUTS;

typedef	struct {
    unsigned short        was_requested;
    unsigned short        data_valid;
    unsigned short        new_flag;
    unsigned char         last_read[81];
               } DCTSTRINGINS;

typedef	struct {
    unsigned short        was_requested;
    unsigned short        data_valid;
    unsigned short        last_write;
               } DCTBUZZEROUTS;



typedef	struct {
    unsigned short        was_requested;
    unsigned short        data_valid;
    unsigned short        new_flag;
    unsigned char         last_read;
               } XSITEINS;

typedef	struct {
    unsigned short        was_requested;
    unsigned short        data_valid;
    unsigned short        new_flag;
    unsigned char         last_read[81];
               } XUSERIDS;

typedef	struct {
    unsigned short        was_requested;
    unsigned short        data_valid;
    unsigned short        new_flag;
    unsigned char         last_write;
               } XPERIMETERS;


typedef	struct {
    unsigned short        was_requested;
    unsigned short        data_valid;
    unsigned short        new_flag;
    unsigned char         last_write;
               } XNEWFILES;

typedef	struct {
    unsigned short        was_requested;
    unsigned short        data_valid;
    unsigned short        new_flag;
    unsigned char         last_read[81];
               } XTIMEDATE;


typedef	struct {
    unsigned short        was_requested;
/*
    unsigned short        data_valid;
    unsigned short        new_flag;
*/
    unsigned short        last_write;
               } XDELAYS;


#define XUSERLEN 15      /* this is equivelant to SLM_NAME_SIZE */
 
typedef	struct {
    unsigned short        was_requested;
    unsigned short        data_valid;
    unsigned short        new_flag;
    unsigned char         last_read[XUSERLEN+1];
               } XUSERSIN;


/* Message definitions for redundant updates */
typedef struct
{
             short           line;
             short           usersincount;
             long            statetimeout;
             XUSERSIN        usersin [NUMUSERSIN];
} RSAREDUPD;





typedef	struct	{
    unsigned short  was_requested;	/* 0  The scan task has accesses this RTU 	*/
    unsigned short  is_responding;	/* 1  The RTU is responding 			*/
    unsigned short  holdflag;		/*    The RTU data is in hold 			*/
    unsigned short  newflag;		/*    The RTU data is in hold 			*/

    unsigned short  no_dataout_msgs;	/* 5  Count Output Msgs 			*/
    unsigned short  no_datain_msgs;	/* 6  Count Data Requests 			*/
    unsigned short  no_rxnoresp;	/* 7  Count No Responses 			*/
    unsigned short  no_txnoresp;	/* 8  Count No Responses 			*/
    unsigned short  no_dataerr;		/* 9  Count Message Echo Errors 		*/
    unsigned short  no_retries;		/* 10 Count Retries 				*/
    unsigned short  no_rxhwerrs;	/* 11 Count Comm Hardware Errors 		*/
    unsigned short  no_txhwerrs;	/* 12 Count Comm Hardware Errors 		*/
    unsigned short  no_cserrs;		/* 13 Count Checksum Errors 			*/
    unsigned short  no_offlines;	/* 14 Count Off Line Transitions		*/
    unsigned short  cmd_reset_stat;	/* 15 Comand to reset stats			*/
    unsigned short  cmd_force_offline;	/* 16 Command to Force Offline 			*/
    unsigned short  badoutcnt;		/* 17 Count Failed output commands 		*/
    unsigned short  no_mr350_msgs;	/* 18 Count MR-350 response messages  		*/

    unsigned short  tscanpass;		/* Total number of scanpasses. incremented at   */
                                        /*  the start of each dct scan.                 */
    unsigned short  msgs;		/* Count of messages which got good responses.  */
                                        /*  incremented on each exit from 'good' msgxfer*/
    unsigned long   responsetime;	/* Total time(ticks) waiting for good responses.*/
                                        /* accumu (exit-entry) time for 'good' msgxfers */
    unsigned short  mr350sendto;

    DCTFIN         fin;
    DCTDINS        dins[NUMDINS];
    DCTDOUTS       douts[NUMDOUTS];
    DCTLEDOUTS     ledouts[NUMLEDOUTS];
    DCTLCDOUTS     lcdouts[NUMLCDOUTS];
    DCTSTRINGINS   stringins[NUMSTRINGINS];
    DCTBUZZEROUTS  buzzerouts[NUMBUZZEROUTS];
}
    MR350;


typedef	struct
{
    XSITEINS        siteins[NUMSITEINS];
    XUSERIDS        userids[NUMUSERIDS];
    XDELAYS         delays[NUMDELAYS];
    XPERIMETERS     perimeters[NUMPERIMETERS];
    XNEWFILES       newfiles[NUMNEWFILES];
    XTIMEDATE       timedate[NUMTIMEDATE];
    XUSERSIN        usersin[NUMUSERSIN];
    short           usersincount;
    short           timeremaining;
}   ACCESS;


/* ASCII Character Definitions */
#define STX 0x02
#define ETX 0x03
#define EOT 0x04
#define ACK 0x06
#define NAK 0x15
#define ESC 0x1B


/*
** User Defined function Codes
*/
#define MINIMUMCMD         '0'
#define REQDINS            '0'     /* Request Digital Inputs command.  */
#define SETDOUTS           '1'     /* Set Digital Outputs command.     */
#define SETLEDOUTS         '2'     /* Set LED Outputs command.         */
#define SETLCDOUTS         '3'     /* Set LCD String Outputs command.  */
#define SETBUZZEROUTS      '4'     /* Set Buzzer Outputs command.      */
#define SETKLAXON          '5'     /* Set Klaxon Parameters.           */
#define SETDEBUGMODE       '6'     /* Set Debug Mode.                  */
#define MAXIMUMCMD         '6'



/*
** User Defined Message Codes 
*/
#define BADGETYPE      'B'      /* a badge has been scanned in the cardreader. */
#define KEYPADTYPE     'K'      /* an id has been entered from the keypad.     */
#define LOGOUTTYPE     'L'      /* the logout function key has been pressed.   */
#define PHOTOTYPE      'P'      /* a digital input has changed value.          */



/* Various message lengths used in message transmission and response validation */
#define LOOPBACKLEN         10     /* wep: not currently used.         */
#define CHKSUMLEN           02
#define REQDINSLEN          03
#define SETDOUTSLEN         05
#define SETLEDOUTSLEN       07
#define SETLCDOUTSLEN       40
#define SETBUZZEROUTSLEN     8


/* Command Results */

#define CMDDONE             00     /* the command was sent and the response was error free.   */
#define CMDFAIL             01     /* there were errors in sending or receiving the command.  */
#define CMDNACK             02     /* response was error free, but command was not completed. */
#define CMDREJECT           03     /* driver rejected command due to unsupported data type.   */
#define CMDBUSY             05     /* The command was sent, but the device is busy            */



/* Access Control Task State Definitions */
#define UNKNOWN              0

#define TESTSTATE14         14
#define TESTSTATE15         15




/* task definitions */

#define DCT_STACK            8000      /* Stack size              */
#define	DCT_PRIORITY          152      /* default priority        */
#define	DCT_TASK_OPTIONS        8      /* Floating point          */


/*
** Mutual-Exclusion Control for diagnostic/data radio channel sharing */
/*
*/

#define		MUXTIMEOUT	60*30	/* 16.67 ms ticks (60Hz): 30 seconds 	*/


typedef struct {
        unsigned long   TXintcount;
        unsigned long   RXintcount;
        unsigned long   RXbadcount;
        unsigned long   msgxfercount;
        unsigned long   startingtick;
               }        LINESTATS;

/* Trace Bit Definitions */

#define ALLTRAFFICTRACE      0x0001
/* #define DCTNACKTRACE         0x0002        ** MR-350 DCT NACK errors.                           */
#define CONFIGURATIONTRACE   0x0010        /* trace configuration errors in driver calls.       */
#define RUNTIMEERRORTRACE    0x0020        /* trace run time errors in driver calls.            */
#define CALLREJECTTRACE      0x0040        /* trace non-error based driver call rejections.     */
#define RSADEBUGENABLE       0x0080        /* enable rsa_debug tracing.                         */



/* ------------------------------------------------------------ */
/* ------------------------------------------------------------ */
/* ----    Remote Site Access Control Task Declarations    ---- */
/* ------------------------------------------------------------ */
/* ------------------------------------------------------------ */


/*
** Remote Site Access Task Queue Management Structures for MR-350 Unit Protocol
**
*/
#define         KLAXONCMD        1
#define         LEDCMD           2
#define         LOGCMD           3
#define         LCDCMD           4

#define         TASKQSIZE      128

typedef struct {
	unsigned short  link;
	unsigned short  unit;
	unsigned short  command;
	unsigned char   parameter[80];
               } TASKOUT;

typedef struct {
    unsigned short	ooutix;
    unsigned short	oinix;
    TASKOUT             taskque[TASKQSIZE];
               } TASKQUE;



/*************************/
/* Constant Declarations */
/*************************/

#define MAXBADRETRY 1

#define	ROOT_POINTNAME	"_rsa_state"

/* actions taken on accessing the current state */
#define GETIT          1
#define SETIT          0


/* user list management outcome states */
#define LISTFULL       0
#define INSERTED       1
#define DELETED        2


/* right-end truncation length used when displaying user name on MR-350 Line 2 */
#define  MAXIDLENGTH   8


#define UNDEFINED      0
#define SECURE         1
#define ALERT          2
#define OCCUPIED       3
#define SECURING       4
#define ALARM          5
/* #define WEIRD         99 */


#define VALIDLEDNUM    0
#define INVALIDLEDNUM  1
#define LEDOFF         0
#define LEDON          1
#define LOGINNUM       0
#define LOGOUTNUM      1
#define LOGOFF         0
#define LOGON          1


#define IDLE           0        /* line 2 message state */
#define VALIDLOGIN     1        /* line 2 message state */
#define VALIDLOGOUT    2        /* line 2 message state */
#define DISPLAYCOUNT   3        /* line 2 message state */
#define INVALIDID      4        /* line 2 message state */


#define LOGIN          0
#define LOGOUT         1


/* system default values */
#define ALERTTIMEOUT        30   /*     30 secs  */
#define SECURINGTIMEOUT     30   /*     30 secs  */
#define WARNINGTIMEOUT       5   /*      5 secs  */
#define MESSAGETIMEOUT       2   /*      2 secs  */
#define ALARMTIMEOUT      2*60   /*   2 minutes  */
#define RETRYCOUNT           2

#define REFRESHINTERVAL   5*60   /* every 5 mniutes */

#define MAX_RSA_ALERT     5*60   /* max timeout (05 minutes) */
#define MAX_RSA_SECURING  5*60   /* max timeout (05 minutes) */
#define MAX_RSA_WARNING     30   /* max timeout (30 seconds) */
#define MAX_RSA_MESSAGE     30   /* max timeout (30 seconds) */
#define MAX_RSA_ALARM    10*60   /* max timeout (10 minutes) */
#define MAX_RSA_RETRY        9

#define RSA_ALERT       2  /*   Signal Number for Login Timer Delay (seconds).              */
                           /*    Intruder Alert: time window for known user to login.       */

#define RSA_RETRY       3  /*   Signal Number for Login Retry Counts.                       */

#define RSA_SECURING    4  /*   Signal Number for Logout Timer Delay (seconds).             */
                           /*    Securing Site: time window for securing the site.          */

#define RSA_WARNING     5  /*   Signal Number for Warning Buzzer Duration (seconds).        */
                           /*    Alarm Klaxon is sounded for this length of time when:      */
                           /*     Site Secure, and a known but unexpected login occurs.     */
                           /*     Site Occupied, and LOGOUT pressed while perimeter open.   */
                           /*     Securing Site, and known user logs in.                    */

#define RSA_MESSAGE     6  /*** Signal Number for Login Message Delay (seconds).            */
                           /*    Message is display on LCD this length of time when:        */
                           /*     Site Secure, and known user logs in. "UUUUUUU LOGIN".     */
                           /*     Site Secure, after "UUUUUUU LOGIN". "01 logged in".       */
                           /*     Site Secure, and unknown user logs in. "INVALID ID".      */
                           /*     Intruder Alert, and known user logs in. "UUUUUUU LOGIN".  */
                           /*     Intruder Alert, after "UUUUUUU LOGIN". "01 logged in".    */
                           /*     Intruder Alert, and unknown user logs in. "INVALID ID".   */
                           /*     Intruder Alarm, and known user logs in. "UUUUUUU LOGIN".  */
                           /*     Intruder Alarm, after "UUUUUUU LOGIN". "01 logged in".    */
                           /*     Intruder Alarm, and unknown user logs in. "INVALID ID".   */
                           /*     Site Occupied, and known user logs in. "UUUUUUU LOGIN".   */
                           /*     Site Occupied, after "UUUUUUU LOGIN". "NN logged in".     */
                           /*     Site Occupied, and listed user logs in. "UUUUUUU LOGOUT". */
                           /*     Site Occupied, and unknown user logs in. "INVALID ID".    */
                           /*     Securing Site, and known user logs in. "UUUUUUU LOGIN".   */
                           /*     Securing Site, after "UUUUUUU LOGIN". "01 logged in".     */
                           /*     Securing Site, and unknown user logs in. "INVALID ID".    */

#define RSA_ALARM       7  /*** Signal Number for Warning Buzzer Duration (seconds)             */
                           /*    Goto Intruder alarm and sound Alarm Klaxon for this time when: */
                           /*     Site Secure, and unknown user logs in.                        */
                           /*     Intruder Alert, and unknown user logs in, and max retry.      */
                           /*     Intruder Alarm, and unknown user logs in, and Klaxon not on.  */

/* if more delays are added, make sure NUMDELAYS is modified accordingly */





#define RSA_STACK            8000      /* Stack size              */
#define	RSA_PRIORITY          153      /* default priority        */
#define	RSA_TASK_OPTIONS        8      /* Floating point          */




/*	End of File:	drvMr350.h */
