/*drvNetpac.h,v 1.1.1.1 1999/08/03 14:43:25 tis4000 Exp */

/****************************************************************/
/*                                                              */
/*    This software is the proprietary property of Tate         */
/*    Integrated Systems, L.P.  Patents and copyrights are      */
/*    pending.                                                  */
/*                                                              */
/*    This software is furnished under a license and may        */
/*    be used and copied only in accordance with the terms      */
/*    of such license and with the inclusion of the above       */
/*    notice. This software or any other copies thereof         */
/*    may not be provided or otherwise made available to        */
/*    any other person. No title to and ownership of the        */
/*    software is hereby transferred.                           */
/*                                                              */
/*    The information in this software is subject to            */
/*    change without notice and should not be construed         */
/*    as a commitment by Tate Integrated Systems L.P. or        */
/*    its third party suppliers.                                */
/*                                                              */
/*    Tate Integrated Systems, L.P. (TIS) and its third         */
/*    party suppliers assume no responsibility for the          */
/*    use or inability to use any of its software. TIS          */
/*    software is provided "as is" without warranty of          */
/*    any kind and TIS expressly disclaims all implied          */
/*    warranties, including but not limited to the implied      */
/*    warranties of merchantability and fitness for a           */
/*    particular purpose.                                       */
/*                                                              */
/*    Notice: Notwithstanding any other lease or license        */
/*    that may pertain to or accompany the delivery of this     */
/*    computer software, the rights of the Government           */
/*    regarding its use, reproduction, and disclosure are       */
/*    as set forth in Section 52.227-19 of the FARS             */
/*    Computer Software-Restricted Rights clause.               */
/*                                                              */
/****************************************************************/
/*
/*  File: drvNetpac.h
/*  Edit:  7/25/95     wep : created                              
/*                         :
/*
*/                                                                                                                      

#define TRUE    1
#define FALSE   0
#define NULLP   0
#define CR      0x0d

/*  number of lines of Netpac              */
#define     NETPACLINES    20

/*  number of netpac modules per line      */
#define     NUMMODS        16

/*  maximum number of channels per module  */
#define     NUMCHANNELS    20

#define     MINAINO         0
#define     MAXAINO        19

#define     MINDINO         1
#define     MAXDINO         1

/*                                                                           		
/* The IOC will assume that the data to be read from each Netpac module will            
/*   will come from a fixed number of channels, not exceeding 20 per module.            
/*   It will read all of the channels in one module in one read request to              
/*   minimize the amount of communications overhead.                                    
/*                                                                                      
*/

#define MAINS           0               /* Card number for analog inputs              */
#define MODULE       1000               /* Card number for module status              */
#define MPSEUDO      9999               /* Card number for pseudo i/o                 */


/*
/*	Conversion types for analog inputs	*/
/*
*/
#define		AICNVINTEGER    10
#define		AICNVDOUBLEINT  11
#define		AICNVFLOAT      12
#define		AICNVDOUBLEFLT  13
#define		AICNVASCIIFLT   14

#define		AICNVSQRROOT    15

#define		AICNVSTATUS     15
#define		AICNVTIME       16
#define		AICNVASCII_2    17
#define		AICNVASCII_4    18
#define		AICNVASCII_12   19
#define		AICNVASCII_20   20
#define		MAXCONV         20

/* The following must match the definition in recChoice.ascii */
#define NOCONV 0
#define LINEAR 1
#define ASCFLT 2


#define    WATCHDOGTIMEOUT       10*60 /* watchdog monitor timeout.  */


/*
/*	Communications I/O Structures
/*
*/

#define		RXSEM		0
#define		TXSEM		1
#define		MAXRXMSG	256
#define		MAXTXMSG	256


typedef	struct {
	unsigned short		rxstate;
	unsigned short		fixedlen;
	unsigned short		rxmsglenexpected;
        unsigned short          rxmsglen;
       	unsigned short		rxix;
	SEMAPHORE		*semid;
	unsigned short		*status;
        unsigned char           txaddress;
        unsigned char           txfunction;
       	unsigned char		rxbuf[MAXRXMSG];
		} 	RXLINE;


typedef	struct {
	unsigned short		protocol;
	unsigned short		txdisable;
	unsigned short		txtocnt;
	unsigned short		txhwerrcnt;
	unsigned short		txmsglen;
	unsigned short		txix;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		txbuf[MAXTXMSG];
		} 	TXLINE;
/*
/*	Communications I/O Constants
/*
*/

#define		RXGOOD			1
#define		RXBAD			2
#define		RXTIMEOUT		60	/* 16 ms ticks 				*/
#define		ACTIVITYTIME		300	/* must see activity in 5 seconds 	*/

#define		VARRESP			FALSE
#define		FIXEDRESP		TRUE
#define		ONERETRY		0x8000	/* OR'ed into line number for xfermsg 	*/

#define		RXIGNORE                0
#define		RXCOLON                 1
#define		RXAT                    2
#define		RXRESPONSE              3

#define		TXGOOD			1
#define		TXBAD			2
#define		IO_TIMEOUT		3	/* Code returned by timer 		*/


#define		TXTIMEOUT		30	/* 16 ms ticks: one second/2            */

#define		MAXRETRYS		3	/* retransmission trys 			*/

#define		MAXCFGCK		8	/* Try to Bring online each cycle 	*/
#define		BRINGBACKCNT		4	/* Poll loops per cycle 		*/

#define		XFERGOOD		1
#define		XFERBAD			2

#define		RESNAMELEN		16	/* length of a resource name 		*/
#define		SLWWINDOW		900	/* # of ticks to allow slave to try 	*/

/*
/*	Internal IOC Structure for Netpac Modules
/*
*/
typedef	struct {
    unsigned short  was_requested;
    unsigned short  device_type;
    unsigned short  initialized;

    unsigned short  data_valid;
    unsigned char   raw_read[8];
    float           last_read;
    unsigned short  data_status;
               } INPUTS;


typedef struct {
    unsigned short  was_requested;      /* 0  The scan task has accesses this RTU       */
    unsigned short  is_responding;      /* 1  The RTU is responding                     */
    unsigned short  holdflag;           /*    The RTU data is in hold                   */
    unsigned short  no_dataout_msgs;    /* 5  Count Output Msgs                         */
    unsigned short  no_datain_msgs;     /* 6  Count Data Requests                       */
    unsigned short  no_rxnoresp;        /* 7  Count No Responses                        */
    unsigned short  no_txnoresp;        /* 8  Count No Responses                        */
    unsigned short  no_dataerr;         /* 9  Count Message Echo Errors                 */
    unsigned short  no_retries;         /* 10 Count Retries                             */
    unsigned short  no_rxhwerrs;        /* 11 Count Comm Hardware Errors                */
    unsigned short  no_txhwerrs;        /* 12 Count Comm Hardware Errors                */
    unsigned short  no_chksumerrs;      /* 13 Count Checksum Errors                     */
    unsigned short  no_offlines;        /* 14 Count Off Line Transitions                */
    unsigned short  cmd_reset_stat;     /* 15 Comand to reset stats                     */
    unsigned short  cmd_force_offline;  /* 16 Command to Force Offline                  */
    unsigned short  badoutcnt;          /* 17 Count Failed output commands              */
    unsigned short  no_frameerrs;       /* 18 Count Framing Errors                      */
    unsigned short  tscanpass;          /* Total number of scanpasses. incremented at   */
                                        /*  the start of each plc scan.                 */
    unsigned short  msgs;               /* Count of messages which got good responses.  */
                                        /*  incremented on each exit from 'good' msgxfer*/
    unsigned long   responsetime;       /* Total time(ticks) waiting for good responses.*/
                                        /* accumu (exit-entry) time for 'good' msgxfers */
    unsigned short  status;
    unsigned short  statusclass;
    unsigned short  status_valid;
    unsigned short  modulestatus[2];

    unsigned short  noerror;
    unsigned short  progerror;
    unsigned short  commerror;
    unsigned short  fatalerror;
    unsigned short  configerror;
    unsigned short  unknownerror;

    unsigned short  error00;
    unsigned short  error01;
    unsigned short  error02;
    unsigned short  error03;
    unsigned short  error04;
    unsigned short  error10;
    unsigned short  error11;
    unsigned short  error12;
    unsigned short  error40;
    unsigned short  error41;
    unsigned short  error42;
    unsigned short  error43;
    unsigned short  error44;
    unsigned short  error45;
    unsigned short  error50;

    INPUTS  inputs[NUMCHANNELS];
   } NETPAC;
/*
/* Netpac Protocol function codes
*/

#define LOOPBACK        'T'   /* MODULE LOOPBACK                                        */
#define READSTATUS      'A'   /* READ MODULE STATUS                                     */
#define READANALOGDATA  'D'   /* READ ANALOG DATA                                       */
#define ENGINEERINGUNIT 'E'   /* SET ENGINEERING UNITS                                  */
#define TEMPERATUREUNIT 'F'   /* SET TEMPERATURE UNITS                                  */
#define GOSCAN          'G'   /* GO SCAN THE MODULE                                     */
#define DATAFORMAT      'H'   /* SET DATA FORMAT                                        */
#define TALKMODE        'T'   /* SET TALK MODE ACTIVE                                   */

#define FNCERRBIT       0x80  /* ------- error function code                            */



/* Various message lengths used in message transmission validation                      */
#define LOOPBACKLEN        4  /* MODULE LOOPBACK LENGTH                                 */
#define READSTATUSLEN      4  /* MODULE STATUS LENGTH                                   */
#define READANALOGDATALEN  4  /* ANALOG DATA LENGTH                                     */
#define ENGINEERINGUNITLEN 8  /* SET ENGINEERING UNITS REPLY LENGTH                     */
#define TEMPERATUREUNITLEN 5  /* SET TEMPERATURE UNITS REPLY LENGTH                     */
#define GOSCANLEN          4  /* GO SCAN THE MODULE REPLY LENGTH                        */
#define DATAFORMATLEN      5  /* SET DATA FORMAT                                        */
#define TALKMODELEN        4  /* SET TALK MODE ACTIVE                                   */

#define CHKSUMLEN          2

/* Various message lengths used in message response validation                          */
#define STATUSREPLYLEN     5  /* Status Reply Length                                    */
#define DATAREPLYLEN    2+(20*8)  /* Read Analog Data Reply Length                          */

/*
/* Netpac Command Reply Error Codes
*/

#define NOERRORS           00    /* No errors, no new commands received    */
#define COMMANDOK          01    /* Command received, no errors to report  */
#define PROGRAMMINGERROR   02    /* Programming error                      */
#define POWEUPFLAGERROR    03    /* Power up flag not set                  */
#define FRAMINGERROR       04    /* Serial framing error                   */

#define PROMERROR          10    /* PROM check error                       */
#define RAMERROR           11    /* RAM check error                        */
#define VCOERROR           12    /* VCO check error                        */

#define CHANNELRANGEERROR  40    /* Channel number out of range            */
#define CARDNUMBERERROR    41    /* Card number not installed              */
#define EUERROR            42    /* EU is not 40 for <V> (Value)           */
#define VALUERANGEERROR    43    /* Value out of range                     */
#define OVERRANGEERROR     44    /* Overrange                              */
#define POWERFAILURE       45    /* Power Failure                          */

#define CHECKSUMERROR      50    /* Checksum error for any module          */
#define CHECKSUMERROR00    50    /* Checksum error for module 00           */
#define CHECKSUMERROR15    65    /* Checksum error for module 15           */


#define NOERROR            0
#define COMMERROR          1
#define PROGERROR          2
#define FATALERROR         3
#define CONFIGERROR        4
#define UNKNOWNERROR       5



/*
/* Netpac Channel Error Codes
*/

#define GOODDATA           00    /*                                        */
#define SKIPCHANNEL        01    /*                                        */
#define OVERRANGE          02    /*                                        */
#define OPENTHERMOCOUPLE   03    /*                                        */
#define PARITYERROR        04    /*                                        */
#define COMMUNICATIONERROR 05    /*                                        */
#define MATHERROR          06    /*                                        */






/* Command Results */
#define CMDDONE          0x00     /* the command was sent and the response was error free.   */
#define CMDFAIL          0x01     /* there were errors in sending or receiving the command.  */
#define CMDNACK          0x02     /* response was error free, but command was not completed. */
#define CMDREJECT        0x03     /* driver rejected command due to unsupported data type.   */
#define CMDBUSY	         0x05     /* The command was sent, but the device is busy.           */ 


/* task definitions */
#define NETPAC_STACK        8000  /* Stack size              */
#define	NETPAC_PRIORITY     152   /* default priority        */
#define	NETPAC_TASK_OPTIONS 8     /* Floating point          */


/*
/* Mutual-Exclusion Control for diagnostic/data radio channel sharing */
/*
*/

#define		MUXTIMEOUT	60*30	/* 16.67 ms ticks (60Hz): 30 seconds 	*/

typedef struct {
        unsigned long   TXintcount;
        unsigned long   RXintcount;
        unsigned long   msgxfercount;
        unsigned long   startingtick;
               }        LINESTATS;

/* Trace Bit Definitions */

#define ALLTRAFFICTRACE      0x0001
#define MODBUSNACKTRACE      0x0002        /* MODBUS NACK errors.                               */
#define CONFIGURATIONTRACE   0x0010        /* trace configuration errors in driver calls.       */
#define RUNTIMEERRORTRACE    0x0020        /* trace run time errors in driver calls.            */
#define CALLREJECTTRACE      0x0040        /* trace non-error based driver call rejections.     */

#define OKTRACE              0x0100
#define COMMTRACE            0x0200
#define PROGTRACE            0x0400
#define FATALTRACE           0x0800
#define CONFIGTRACE          0x1000
#define UNKNOWNTRACE         0x2000

#define ASCIITRAFFICTRACE    0x4000        /* when combined with ALLTRAFFICTRACE,               */
                                           /*  prints comm traffic in ascii.                    */
/*	End of File:	drvNetpac.h */
