/*drvQei.h,v 1.1.1.1 1999/08/03 14:44:15 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
/*	File:	drvQei.h							       	
/*	Auth:	John L. Dalesio                                        	       	       	
/*	Edit:	05/06/94        WEP: based on smod_driver                               
/*              06/25/96        WEP: added DATATRAFFICTRACE flag                        
/*                                                                                      
/*	This file contains the data structures used by the QEI I/O driver.              
/*                                                                                      
*/

#define TRUE	1
#define FALSE	0
#define NULLP	0
#define CR	0x0d

/*	number of lines of modbus 	*/
#define		QEILINES	20
/*	number of modbus plcs per line */
#define		NUMQEIRTUS	256

/*                                                                           		
/*	The IOC will assume that the data to be exchanged with each RTU will           	
/*	come from a fixed number of dis,acc,and ais.               	
/*	It will read these initially in polls for all data, and the in
/*	requests for upsets.
*/

#define QEIAIS		0			/* Card number for analog inputs 	*/
#define QEIACCS		1			/* Card number for accumulators 	*/
#define QEIDIS		3			/* Card number for digital inputs 	*/
#define QEIDOS		4			/* Card number for digital outputs 	*/
#define QEIAOS		5			/* Card number for analog outputs 	*/
#define PSEUDO		33			/* Card number for pseudo i/o 		*/
#define POLLCONTROL	22			/* Card Number for polling control      */

#define MAXAI		128			/* max number of analog inputs  	*/
#define MAXACC		8			/* max number of accumulators 		*/
#define MAXDI		256			/* max number of digital inputs 	*/
#define MAXDIBLK	(MAXDI/16)		/* max number of digital input cards 	*/
#define MAXDO		128			/* max number of digital outputs 	*/
#define MAXAO		128			/* max number of analog outputs 	*/


/*
/*	Conversion types for analog inputs	*/
/*
*/
#define		AICNVINTEGER	10
#define		AICNVDOUBLEINT	11
#define		AICNVFLOAT	12
#define		AICNVDOUBLEFLT	13
#define		AICNVASCIIFLT	14
#define		AICNVSTATUS	15
#define		AICNVTIME	16
#define		AICNVASCII_2	17
#define		AICNVASCII_4	18
#define		AICNVASCII_12	19
#define		AICNVASCII_20	20
#define		MAXCONV		20

/* The following must match the definition in recChoice.ascii */
#define NOCONV 0
#define LINEAR 1
#define ASCFLT 2


#define    WATCHDOGTIMEOUT       10*60 /* watchdog monitor timeout.  */


/*
/*	Communications I/O Structures
/*
*/

#define		RXSEM		0
#define		TXSEM		1
#define		MAXRXMSG	(256*4)+10
#define		MAXTXMSG	256

typedef	struct {
	unsigned short		protocol;
	unsigned short		rxstate;
	unsigned short		rxtocnt;
	unsigned short		rxhwerrcnt;
	unsigned short		fixedlen;
	unsigned short		rxmsglen;
	unsigned short		rxix;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		rxbuf[MAXRXMSG];
		} 	RXLINE;


typedef	struct {
	unsigned short		protocol;
	unsigned short		txdisable;
	unsigned short		txtocnt;
	unsigned short		txhwerrcnt;
	unsigned short		txmsglen;
	unsigned short		txix;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		txbuf[MAXTXMSG];
		} 	TXLINE;
/*
/*	Communications I/O Constants
/*
*/
#define		RXGOOD			1
#define		RXBAD			2
#define		RXTIMEOUT		180	/* 16 ms ticks 				*/
#define		ACTIVITYTIME		300	/* must see activity in 5 seconds 	*/

#define		VARRESP			FALSE
#define		FIXEDRESP		TRUE
#define		ONERETRY		0x8000	/* OR'ed into line number for xfermsg 	*/
#define		RXIGNORE		0
#define		RXBLKCNT		1
#define		RXDATA			2

#define		TXGOOD			1
#define		TXBAD			2

#define		TXTIMEOUT              60	/* 16 ms ticks: one second              */

#define		IO_TIMEOUT		3	/* Code returned by timer 		*/

#define		MAXRETRYS		5	/* retransmission trys 			*/

#define		MAXCFGCK		8	/* Try to Bring online each cycle 	*/
#define		BRINGBACKCNT		4	/* Poll loops per cycle 		*/

#define		XFERGOOD		1
#define		XFERBAD			2
#define		XFERGOODCIU		3

#define		RESNAMELEN		16	/* length of a resource name 		*/
#define		SLWWINDOW		900	/* # of ticks to allow slave to try 	*/


/*
/* Output Queue Management Structures for Modbus Protocol
/*
*/

#define         AOUTTYPE        1
#define         BOUTTYPE        2
#define         OUTQSIZE        128

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  aosignal;
	short           aovalue;
	double		dblaovalue;
	unsigned short	conversion;
		} AOUT, *AOUTPTR;

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  bosignal;
	long            bomask;
	long            bovalue;
		} BOUT, *BOUTPTR;

typedef struct {
		unsigned short  otype;
		union   {
			AOUT    aonode;
			BOUT    bonode;
			}out;
	       } OUTQENT;

typedef struct {
		unsigned short	ooutix;
		unsigned short	oinix;
		OUTQENT         outque[OUTQSIZE];
		} OUTPUTQUE;


typedef	struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned long		last_read;
		}	QEIDI;

typedef struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	short			last_read;
		}	QEIAI;

typedef struct {
	unsigned short		was_requested;
	unsigned short		data_valid;
	unsigned short		last_read;
		}	QEIACC;

typedef struct {
	unsigned char		station[8];
		}	ID;

typedef	struct	{
	unsigned short  was_requested;		/* 0  The scan task has accesses this RTU 	*/
	unsigned short  is_responding;		/* 1  The RTU is responding 			*/
	unsigned short  holdflag;		/*    The RTU data is in hold 			*/
	unsigned short  no_dataout_msgs;	/* 5  Count Output Msgs 			*/
	unsigned short  no_datain_msgs;		/* 6  Count Data Requests 			*/
	unsigned short  no_rxnoresp;		/* 7  Count No Responses 			*/
	unsigned short  no_txnoresp;		/* 8  Count No Responses 			*/
	unsigned short  no_dataerr;		/* 9  Count Message Echo Errors 		*/
	unsigned short  no_retries;		/* 10 Count Retries 				*/
	unsigned short  no_rxhwerrs;		/* 11 Count Comm Hardware Errors 		*/
	unsigned short  no_txhwerrs;		/* 12 Count Comm Hardware Errors 		*/
	unsigned short  no_crcerrs;		/* 13 Count CRC Errors 				*/
	unsigned short  no_offlines;		/* 14 Count Off Line Transitions		*/
	unsigned short  cmd_reset_stat;		/* 15 Comand to reset stats			*/
	unsigned short  cmd_force_offline;	/* 16 Command to Force Offline 			*/
	unsigned short  badoutcnt;		/* 17 Count Failed output commands 		*/
	unsigned short  no_zeroresponses;       /*    Count Zero Byte Message Responses         */

	/* These command may be implemented as broadcast pseudo control                         */
	unsigned short	accreset;		/* 30 Broadcast Acc Reset                       */
	unsigned short	accfreeze;		/* 31 Broadcast Acc Freeze                      */
	unsigned short  xferreset;		/* 32 Broadcast Acc Transfer and reset          */

	short		setpoints[MAXAO];	/* Analog outputs                               */
	unsigned short	dos[MAXDO];		/* Digital outputs                              */
	QEIACC		accs[MAXACC];		/* Accumulators                                 */
	QEIAI		ais[MAXAI];		/* Analog Inputs                                */
	QEIDI		dis[MAXDIBLK];		/* Digital Inputs                               */
	unsigned char	stationname[8];		/* Name of station                              */

	/* Polling Control values */
	unsigned short	forcescan;		/* pseudo control for scan now                  */
	unsigned short	fast_poll;		/* this unit is on fast poll                    */
	unsigned short	scan_freq;		/* pseudo control for scan frequency            */
	unsigned long	last_scan_no;		/* cycle counter of last scan                   */
	unsigned short	poll_due;		/* this unit is due for a poll                  */
	unsigned short	poll_all_due;		/* this unit is due for a poll for all data     */
	unsigned short	poll_acc_due;		/* this unit is due for a poll for acc          */


        unsigned short  tscanpass;		/* Total number of scanpasses. incremented at   */
                                                /*  the start of each plc scan.                 */
        unsigned short  msgs;			/* Count of messages which got good responses.  */
                                                /*  incremented on each exit from 'good' msgxfer*/
        unsigned long   responsetime;		/* Total time(ticks) waiting for good responses.*/
                                                /* accumu (exit-entry) time for 'good' msgxfers */

	}	QEIRTU;

/*
/* Poll Control for QEI Line
*/
typedef	struct {
	unsigned long	scan_number;
	unsigned short	lastscanned[10]; 	/* Addresses of last 10 units scanned           */
	unsigned short	lastscannedstat[10];	/* On Line status of last 10 units scanned      */
	unsigned short	pollreason[10];		/* Type of poll list last 10 were taken from    */
	ID		id[10];			/* Station Name of last 10 stations polled      */
	unsigned short	nosincenormal;		/* Number of polls since last normal scan       */
	unsigned short	nosincefast;		/* Number of polls since last fast scan         */
	unsigned short	nosinceoffline;		/* Number of polls since last offline scan      */
	unsigned short	noperoffline;		/* Number of polls between offline scans        */
	unsigned short	noperfast;		/* Number of polls between fast scans           */
	unsigned short  minsperacc;		/* Minutes between accumulator polls            */
	unsigned short	minsperall;		/* Minutes between polls for all data           */
	unsigned short	minstillacc;		/* Minutes till next poll for accumulators      */
	unsigned short	minstillall;		/* Minutes till next poll for all data          */
	unsigned long	lastalltime;		/* last till all data was fetched               */
	unsigned long	lastacctime;		/* last till all accs were fetched              */
	unsigned short  lastsetfreq;		/* Last unit polled for set freq scan           */
	unsigned short  lastforced;		/* Last unit polled for forced poll             */
	unsigned short  lastfast;		/* Last unit polled for fast poll               */
	unsigned short  lastoffline;		/* Last unit polled for offline scan            */
	unsigned short  lastnormal;		/* Last unit polled for normal scan             */

    } POLLCNTRL;


/* QEI Communications constants */

/* TTC Codes */
#define	M2RPRESING	(4*2)
#define	M2RPREMORE	(0*2)
#define	M2RDATAMORE	(2*2)
#define	M2RDATALAST	(6*2)
#define	R2MPRESING	(5*2)
#define	R2MPREMORE	(1*2)
#define	R2MDATAMORE	(3*2)
#define	R2MDATALAST	(7*2)

/* Opcodes */
#define	RESTART		(0)
#define	POLLRESET	(1)
#define	POLLNORESET	(2)
#define	POLLALL		(9)
#define	SELECT_DO	(10)
#define	EXECON		(11)
#define	EXECOFF		(12)
#define	POLLACC		(13)
#define	RESETACC	(14)
#define	FREEZEACC	(15)
#define	SELECT_AO	(16)
#define	EXEC_AO		(17)
#define	XFERRESETACC	(20)

#define	STOPBIT		0x80

#define SINGLELEN	4
#define SETPOINTLEN	8

#define DIGDATAFLAG	0x10
#define ACCFLAG		0x01000000


/* task definitions */

#define QEI_STACK		8000            /* Stack size              */
#define	QEI_PRIORITY		152             /* default priority        */
#define	QEI_TASK_OPTIONS	8               /* Floating point          */




/*
/* Mutual-Exclusion Control for diagnostic/data radio channel sharing */
/*
*/

#define		MUXTIMEOUT	60*30	        /* 16.67 ms ticks (60Hz): 30 seconds 	*/


/*
/* simulation data structure */
/*
*/
typedef	struct	{

	unsigned short	dosel;
	unsigned short	aosel;
	unsigned short	aoval;
	unsigned short  outputstation;		

	unsigned short	setpoints[MAXAO];	/* Analog outputs       */
	unsigned short	dos[MAXDO];		/* Digital outputs      */
	unsigned short	accs[MAXACC];		/* Accumulators         */
	unsigned short	ais[MAXAI];		/* Analog Inputs        */
	unsigned short	dis[MAXDIBLK];		/* Digital Inputs       */
	unsigned short	cfaccs[MAXACC];		/* Accumulators         */
	unsigned short	cfais[MAXAI];		/* Analog Inputs        */
	unsigned short	cfdis[MAXDIBLK];	/* Digital Inputs       */
	unsigned short	ocfaccs[MAXACC];	/* Accumulators         */
	unsigned short	ocfais[MAXAI];		/* Analog Inputs        */
	unsigned short	ocfdis[MAXDIBLK];	/* Digital Inputs       */

	}	QEISIM;


/* Scan Types for RTUS */

#define NORMAL		0
#define OFFLINESCAN	1
#define FASTSCAN	2
#define SETFREQSCAN	3
#define FORCEDSCAN	4

#define DEFAULTPOLLSPEROFFLINE 256
#define DEFAULTPOLLSPERFAST	5


typedef struct {
        unsigned long   TXintcount;
        unsigned long   RXintcount;
        unsigned long   msgxfercount;
        unsigned long   startingtick;
               }        LINESTATS;


/* Trace Bit Definitions */

#define ALLTRAFFICTRACE      0x0001
#define MODBUSNACKTRACE      0x0002        /* Communication NACK errors.                        */
#define DATATRAFFICTRACE     0x0004        /* While tracing traffic, display the data, too.     */
#define CONFIGURATIONTRACE   0x0010        /* trace configuration errors in driver calls.       */
#define RUNTIMEERRORTRACE    0x0020        /* trace run time errors in driver calls.            */
#define CALLREJECTTRACE      0x0040        /* trace non-error based driver call rejections.     */

/*	End of File:	drvQei.h */
