/****
*
*	File name :		vxquad4.c
*	Author :		Steve Johnson
*	Creation date :	September 14, 1998
*
*	Copyright (c) 1998,1999 Sigma Six Corporation
*
*	Provides the following entry points:
*
*		Quad4Start
*		Quad4Stop
*
****/

#include <sscommon.h>
#include <ssnames.h>
#include <sstrace.h>
#include <sslib.h>
#include "q4task.h"

static void	Quad4Task (char *);

static int		task_pid;
static char		*ini_name;

/****
*
*	Call this to start the Quad4 Meter Master Task
*
****/

int Quad4Start (char *ini_file, int priority)
{
	ini_name = (char *) malloc (MAX_STRING);
	if (!ini_name) {
		printf ("Quad4Start: malloc failed %d bytes\n", MAX_STRING);
		return -1;
	}
	if (!priority) {
		priority = 200;
	}
	strcpy (ini_name, ini_file);
	task_pid = taskSpawn ("tQuad4", priority, 0, 0x8000,
						  (FUNCPTR) Quad4Task,
						  (int) ini_name, 0, 0, 0, 0, 0, 0, 0, 0, 0);
	if (task_pid == ERROR) {
		printf ("Quad4Start: taskSpawn failed, errno = %d\n", errno);
		free (ini_name);
		return -2;
	}
	return 0;
}

int Quad4AddMeter (char *name)
{
	if (!quad4_add_meter (name)) {
		printf ("Quad 4: quad4_add_meter (%s) failed\n", name);
		return -1;
	}
	return 0;
}

/****
*
*	Call this if you need to stop the Quad4 Meter task
*	Undoes what Quad4Start does and stops the task
*
****/

int Quad4Stop (void)
{
	quad4_stop ();
	return 0;
}

UCHAR *Quad4GetTablePtr (USHORT port, USHORT table_number)
{
	return quad4_table_addr (port, table_number);
}

int Quad4DebugOn (void)
{
	quad4_debug_on ();
	return 0;
}

int Quad4DebugOff (void)
{
	quad4_debug_off ();
	return 0;
}

/****
*
*	This is the Task Entry point used in the spawn
*
****/

static void Quad4Task (char *ini_file)
{
	int		status;

	printf ("Quad4Task: Task Started\n");
	status = sslib_add (ini_file, QUAD4_MAIN_GROUP);
	if (status) {
		printf ("Quad4Task, sslib_add failed = %d\n", status);
		free (ini_file);
		return;
	}
	if (!quad4_init ()) {
		printf ("Quad4Task, quad4_init failed\n");
		sslib_remove ();
		free (ini_file);
		return;
	}
	quad4_loop ();
	sslib_remove ();
	printf ("Quad4Task: task terminated\n");
	free (ini_file);
}

