/****************************************************************/
/*                                                              */
/*    This software is the proprietary property of Tate         */
/*    Integrated Systems, L.P.  Patents and copyrights are      */
/*    pending.                                                  */
/*                                                              */
/*    This software is furnished under a license and may        */
/*    be used and copied only in accordance with the terms      */
/*    of such license and with the inclusion of the above       */
/*    notice. This software or any other copies thereof         */
/*    may not be provided or otherwise made available to        */
/*    any other person. No title to and ownership of the        */
/*    software is hereby transferred.                           */
/*                                                              */
/*    The information in this software is subject to            */
/*    change without notice and should not be construed         */
/*    as a commitment by Tate Integrated Systems L.P. or        */
/*    its third party suppliers.                                */
/*                                                              */
/*    Tate Integrated Systems, L.P. (TIS) and its third         */
/*    party suppliers assume no responsibility for the          */
/*    use or inability to use any of its software. TIS          */
/*    software is provided "as is" without warranty of          */
/*    any kind and TIS expressly disclaims all implied          */
/*    warranties, including but not limited to the implied      */
/*    warranties of merchantability and fitness for a           */
/*    particular purpose.                                       */
/*                                                              */
/*    Notice: Notwithstanding any other lease or license        */
/*    that may pertain to or accompany the delivery of this     */
/*    computer software, the rights of the Government           */
/*    regarding its use, reproduction, and disclosure are       */
/*    as set forth in Section 52.227-19 of the FARS             */
/*    Computer Software-Restricted Rights clause.               */
/*                                                              */
/****************************************************************/
/*
**  File: drvQuad4.c                                                          
**  Edit:  1/21/99  wep : created.
**         5/14/99  wep : added support for 12 bit signed and unsigned AI data
**                                                                           
*/                                                                                                                      

#define  VERSIONNUM     2                   /* NOTE: change here */

#include <vxWorks.h>
#include <sysLib.h>
#include <semLib.h>
#include <wdLib.h>
#include <rngLib.h>
#include <logLib.h>
#include <tickLib.h>
#include <remLib.h>
#include <stdioLib.h>
#include <string.h>
#include <tisByteOrder.h>
#include <task_params.h>
#include <dbDefs.h>
#include <drvSup.h>
#include <restable.h>
#include <iocsiodvr_bridge.h>
#include <serialFile.h>
#include <resgrouptask.h>
#include <cvnTypes.h>   /* defines for conversion types */

#include "drvQuad4.h"                          /* NOTE: change here */
#include "drvapi.h"                    /* NOTE: change here */

/*
#include <resGroupTask.h>
*/

/***********************************/
/***********************************/
/******** Begin Prototypes *********/
/***********************************/
/***********************************/

/**********
*
*    This routine is required by iocinit in the EPICS system.
*
**********/
static  long init ();

        void quad4help (void);

        void quad4trlv (unsigned short level);

        void quad4contraceon (void);

        void quad4contraceoff (void);

        void quad4info (void);

        void quad4table (unsigned short link, unsigned short adapter, 
                         unsigned short card, unsigned short signal);

        unsigned char* Quad4GetTablePtr (unsigned short port, unsigned short tablenumber);

/*********************************/
/*********************************/
/******** End Prototypes *********/
/*********************************/
/*********************************/

struct {
            long        number;
            DRVSUPFUN   report;
            DRVSUPFUN   init;
       }
        drvQuad4={ 2, NULL, init};


static  unsigned short xmltrace;

static	unsigned short xtrace;


/**********
*
*    This routine is required by iocinit in the EPICS system.
*
**********/
static long init ()
{
    return (0);
}

/********
**
**    These routine call common help routine
**
********/
void quad4Help (void)
{
    quad4help();
}
void Quad4help (void)
{
    quad4help();
}
void Quad4_help (void)
{
    quad4help();
}
void quad4_help (void)
{
    quad4help();
}
/********
**
**    This routine outputs information to help with debugging.
**
********/
void quad4help (void)
{
    printf (" There is no Quad4 Driver Task to start. \n");
    printf ("                                         \n");
    printf ("  (v%d)                                  \n", VERSIONNUM);
    printf ("\n");
    printf ("\n");
    printf ("Quad4 'Driver' Display Functions (* = default)\n");
    printf ("\n");
    printf (" quad4contraceon(N)     Enable display of configuration errors in periodicScanTask calls.  \n");
    printf ("*quad4contraceoff(N)    Disable display of configuration errors in periodicScanTask calls. \n");
    printf ("\n");
    printf (" quad4trlv(N,X)         Enable the following message for line N:                 \n");
    printf ("                         X=0,  display none of the messages below. (default)     \n");
    printf ("                         X=1,  AIDRIVER: normal call return.                     \n");
    printf ("                         X=2,  BIDRIVER: normal call return.                     \n");
    printf ("                         X=3 to 254,  reserved.                                  \n");
    printf ("\n");
    printf (" quad4table(N,A,C,S)  Interrogate System Memory:                                 \n");
    printf ("                         N:  Port Number                                         \n");
    printf ("                         A:  (Page Number * 100) + Table Number                  \n"); 
    printf ("                         C:  Byte Offset                                         \n");
    printf ("                         S:  Bit Offset                                          \n");
    printf ("\n");
    printf (" quad4help            Display this information.                                  \n");
    printf ("\n");
}


/********
**
**    This routine controls logging of trace messages
**
********/
void quad4trlv (unsigned short level)
{
    xmltrace = level;
}

/********
**
**    These routines enable and disable tracing of database configuration errors in driver calls
**
********/
void quad4contraceon (void)
{
    xtrace = xtrace | CONFIGURATIONTRACE;
}

void quad4contraceoff (void)
{
    xtrace = xtrace & ~CONFIGURATIONTRACE;
}


/********
**
** Show Driver Info and Current State Information  
**
********/
void quad4info (void)
{
    printf ("trace        = %d\n", xtrace);
    printf ("xmltrace     = %d\n", xmltrace);
}


/*                                                                                     
**      quad4aidriver                                                                  
**                                                                                     
**      This routines quad4driver purpose is to return the analog input value for      
**      the specified module. For a module value                                       
**      a (-1) will be returned if there is no valid data ready.                       
**                                                                                     
*/
/* this routine not called from within this driver */

int quad4aidriver (char *pointname, unsigned short link, unsigned short adapter, 
                   unsigned short card, unsigned short signal, short *pvalue, 
                   unsigned short conversion, unsigned char *resname)
{
    unsigned short  port, tablenumber;
    unsigned char  *datapointer;
    unsigned long   tableoffset, pagenumber;
    char            buffer [200];
    unsigned short  signalvalid,temp;

    port = link;
    tablenumber = adapter%100;
    pagenumber = adapter/100;

    signalvalid = TRUE;

    if (signal)
    {
        if (signal != 4)
        {
            signalvalid = FALSE;
        }
        else
        {
            if ((conversion != CNVSIGNED12BIT) &&
                (conversion != CNVUNSIGNED12BIT))
            {
                signalvalid = FALSE;
            }
        }
    }

    if (!signalvalid)
    {
        if (xtrace & CONFIGURATIONTRACE)
            logMsg("AIDRIVER: Illegal Signal Value:[%s] L=%d, A=%d, C=%d, S=%d\n",
                    *pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    datapointer = Quad4GetTablePtr (port, tablenumber);
    if (!datapointer)
    {
        if (xtrace & CONFIGURATIONTRACE)
            logMsg("AIDRIVER: GetTablePtr Returned Invalid Pointer. Illegal Port or Table Number:[%s] L=%d, A=%d, C=%d, S=%d\n",
                    *pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    tableoffset = ((pagenumber * 0x10000) + card);

    datapointer = datapointer + tableoffset;

    switch (conversion)
    {
        case (CNVDOUBLEINT):
            *(long*) (pvalue) = ntohl(*(long*)(datapointer));
            break;

        case (CNVFLOAT):
             ntohf( (*(float*) (datapointer)), (*(float*) (pvalue)) );
            break;

        case (CNVDOUBLEFLT):
             ntohd( (*(double*) (datapointer)), (*(double*) (pvalue)) );
            break;

        case (CNVSIGNED12BIT):
            temp = ntohs( (*(short*) (datapointer)) );
            if (signal == 0)
                temp =  (temp >> 4) & 0xfff;       /* right shift data. */
            else if (signal == 4)
                temp = temp & 0xfff;
            else temp = 0;
            if (temp & 0x800)
		temp = temp | 0xf000;
            *(short*) (pvalue) = temp;
            break;

        case (CNVUNSIGNED12BIT):
            temp = ntohs( (*(short*) (datapointer)) );
            if (signal == 0)
                temp =  (temp >> 4) & 0xfff;       /* right shift data. */
            else if (signal == 4)
                temp = temp & 0xfff;
            else temp = 0;
            *(short*) (pvalue) = temp;
            break;

        case (CNVINTEGER):
        default:
            *(short*) (pvalue) = *(short*) (datapointer);
            break;
    }

    if (xmltrace == 1)
    {
        switch (conversion)
        {
            case (CNVDOUBLEINT):
               sprintf (buffer,"AIDRIVER: Data Returned:[%s] L=%d, A=%d, C=%d, S=%d, datapointer = 0x%x, datavalue = 0x%lx [%ld]\n",
                        pointname, link, adapter, card, signal, (unsigned int) datapointer,  *(long*) pvalue,  *(long*) pvalue);
               break;

            case (CNVFLOAT):
               sprintf (buffer,"AIDRIVER: Data Returned:[%s] L=%d, A=%d, C=%d, S=%d, datapointer = 0x%x, datavalue = 0x%lx [%f]\n",
                        pointname, link, adapter, card, signal, (unsigned int) datapointer,  *(long*) pvalue,  *(float*) pvalue);
               break;

            case (CNVDOUBLEFLT):
               sprintf (buffer,"AIDRIVER: Data Returned:[%s] L=%d, A=%d, C=%d, S=%d, datapointer = 0x%x, datavalue = 0x%lx [%f]\n",
                        pointname, link, adapter, card, signal, (unsigned int) datapointer,  *(long*) pvalue,  *(double*) pvalue);
               break;

            case (CNVINTEGER):
            case (CNVSIGNED12BIT):
            case (CNVUNSIGNED12BIT):
            default:
               sprintf (buffer,"AIDRIVER: Data Returned:[%s] L=%d, A=%d, C=%d, S=%d, datapointer = 0x%x, datavalue = 0x%x [%d]\n",
                        pointname, link, adapter, card, signal, (unsigned int) datapointer,  *(short*) pvalue,  *(short*) pvalue);
               break;
        }
        logMsg ( "%s\n", (int) buffer, 0, 0, 0, 0, 0);
    }
    return (0);
}


/*                                                                                
**      quad4aodriver                                                             
**                                                                                
**      This routines quad4driver purpose is to set requested analog outputs to   
**      the specified module.                                                     
**                                                                                
*/
/* this routine not called from within this driver */

int quad4aodriver (char *pointname, unsigned short link, unsigned short adapter, 
                   unsigned short card, unsigned short signal, double dvalue, 
                   unsigned short value, unsigned short conversion, unsigned char *resname)
{
    if (xtrace & CONFIGURATIONTRACE)
        logMsg("AODRIVER: Illegal Point Type Quad4 Output Not Supported:[%s]\n", *pointname, 0, 0, 0, 0, 0);
    return (-1);
}


/*                                                                               
**      quad4aoread                                                              
**                                                                               
**      This routines quad4driver purpose is to return the value last readback   
**      from the analog output hardware. This is used to initialize analog       
**      outputs.                                                                 
**                                                                               
*/
/* this routine not called from within this driver */

int quad4aoread (char *pointname, unsigned short link, unsigned short adapter, 
                 unsigned short card, unsigned short signal, unsigned short *pvalue,
                 unsigned short conversion, unsigned char *resname)

{
    if (xtrace & CONFIGURATIONTRACE)
        logMsg("AOREAD: Illegal Point Type Hart/Quad4 Output Not Supported:[%s]\n", *pointname, 0, 0, 0, 0, 0);
    return (-1);
}


/*                                                                                    
**      quad4bidriver                                                                 
**                                                                                    
**      The quad4driver purpose of this routine is to return the request digital      
**      input value(s) from the requested card. This routine will return a (-1) if    
**      the requested module is the wrong type or the data is not yet valid.          
**                                                                                    
*/
/* this routine not called from within this driver */

int quad4bidriver (char *pointname, unsigned short link, unsigned short adapter, 
                   unsigned short card, unsigned short signal, unsigned long mask, 
                   unsigned long *pvalue, unsigned char *resname)
{
    unsigned short  port, tablenumber;
    unsigned char  *datapointer;
    unsigned long   tableoffset, pagenumber;
    char            buffer [200];

    port = link;
    tablenumber = adapter%100;
    pagenumber = adapter/100;

    datapointer = Quad4GetTablePtr (port, tablenumber);
    if (!datapointer)
    {
        if (xtrace & CONFIGURATIONTRACE)
            logMsg("BIDRIVER: GetTablePtr Returned Invalid Pointer. Illegal Port or Table Number:[%s] L=%d, A=%d, C=%d, S=%d\n",
                    *pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    tableoffset = (pagenumber * 0x10000) + card;

    datapointer = datapointer + tableoffset;

    *(pvalue+0) = *(short*) (datapointer+0);
    *(pvalue+2) = *(short*) (datapointer+2);

    if (xmltrace == 2)
    {
        sprintf (buffer,"BIDRIVER: Data Returned:[%s] L=%d, A=%d, C=%d, S=%d, datapointer = 0x%x, datavalue =  0x%x [%d]\n",
                 pointname, link, adapter, card, signal, (unsigned int) datapointer, 
                 (unsigned int) *pvalue, (unsigned int) *pvalue);
        logMsg ( "%s\n", (int) buffer, 0, 0, 0, 0, 0);
    }
    return (0);
}


/*                                                                                     
**      quad4bodriver                                                                  
**                                                                                     
**      The quad4driver purpose of this routine is to set the request digital          
**      output value(s) on the requested card. This routine will return a (-1) if      
**      the requested module is the wrong type or not configured yet.                  
**                                                                                     
*/
/* this routine not called from within this driver */

int quad4bodriver (char *pointname, unsigned short link, unsigned short adapter, 
                   unsigned short card, unsigned short signal, unsigned long value, 
                   unsigned long mask, unsigned char *resname)
{
    if (xtrace & CONFIGURATIONTRACE)
        logMsg("BODRIVER: Illegal Point Type Hart/Quad4 Output Not Supported:[%s]\n", *pointname, 0, 0, 0, 0, 0);
    return (-1);
}


/*                                                                              
**      quad4boread                                                             
**                                                                              
**      This routines quad4driver purpose is to return the value last read      
**      back from the digital output hardware. This is used to initialize       
**      digital outputs. This will return a (-1) if the module is the wrong     
**      type or the data is not yet valid                                       
**                                                                              
*/
/* this routine not called from within this driver */

int quad4boread (char *pointname, unsigned short link, unsigned short adapter, unsigned short card,
        unsigned short signal, unsigned long *pvalue, unsigned long mask,
        unsigned char *resname)

{
    if (xtrace & CONFIGURATIONTRACE)
        logMsg("BOREAD: Illegal Point Type Hart/Quad4 Output Not Supported:[%s]\n", *pointname, 0, 0, 0, 0, 0);
    return (-1);
}


/*                                                                                     
**      quad4sidriver                                                                   
**                                                                              	
**      This routines main purpose is to return the analog input value for the         	
**      specified module. For a module value                                           	
**      a (-1) will be returned if there is no valid data ready.                       	
**                                                                                     	
*/
/* this routine not called from within this driver */

int quad4sidriver (char *pointname, unsigned short link, unsigned short adapter, unsigned short card,
                   unsigned short signal, short *pvalue, unsigned short len, unsigned char *resname)
{
    unsigned short  port, tablenumber;
    unsigned char  *datapointer;
    unsigned long   tableoffset, pagenumber;
    char            buffer1 [200];
    char            buffer2 [200];

    port = link;
    tablenumber = adapter%100;
    pagenumber = adapter/100;

    datapointer = Quad4GetTablePtr (port, tablenumber);
    if (!datapointer)
    {
        if (xtrace & CONFIGURATIONTRACE)
            logMsg("SIDRIVER: GetTablePtr Returned Invalid Pointer. Illegal Port or Table Number:[%s] L=%d, A=%d, C=%d, S=%d\n",
                    *pointname, link, adapter, card, signal, 0);
        return (-1);
    }

    tableoffset = (pagenumber * 0x10000) + card;

    datapointer = datapointer + tableoffset;

    strncpy ((unsigned char *) pvalue, datapointer, len); 
    pvalue[len+1] = 0;

    strncpy (&buffer1[0], datapointer, len); 
    buffer1[len+1] = 0;

    if (xmltrace == 3)
    {
        logMsg ( "buffer1 = '%s'\n", (int) buffer1, 0, 0, 0, 0, 0);
        sprintf (buffer2,"SIDRIVER: Data Returned:[%s] L=%d, A=%d, C=%d, S=%d, datapointer = 0x%x, datastring = '%s'\n",
                 pointname, link, adapter, card, signal, (unsigned int) datapointer, &buffer1[0]);
        logMsg ( "%s\n", (int) buffer2, 0, 0, 0, 0, 0);
    }
    return (0);
}


/*                                                                      
**      quad4tableinfo                                                  
**                                                                      
**      The purpose of this routine is twofold:                         
**         1) dump the module level statistics for the modbus driver.   
**         2) dump the data for the specified Quad4 module.             
**                                                                      
*/
/* not called from within this driver */

void quad4table (unsigned short link, unsigned short adapter, 
                 unsigned short card, unsigned short signal)
{
    unsigned short  port, tablenumber;
    unsigned char  *datapointer;
    unsigned long   tableoffset, pagenumber;
    unsigned char   datavaluec;
    unsigned short  datavalues;
    unsigned long   datavaluel; 
    float           datavaluef;
    double          datavalued;

    char            buffer1 [200];

    port = link;
    tablenumber = adapter%100;
    pagenumber = adapter/100;

    printf("quad4table: Port = %d [0x%x], Page Number = %d [0x%x],  Table Number = %d [0x%x]\n",
                        port,  port, (int) pagenumber, (int) pagenumber, (int) tablenumber, (int) tablenumber);

    datapointer = Quad4GetTablePtr (port, tablenumber);
    if (!datapointer)
    {
        printf("quad4table: GetTablePtr Returned Invalid Pointer. Illegal Port or Table Number: L=%d, A=%d, C=%d, S=%d\n",
                link, adapter, card, signal);
        return;
    }
    printf("quad4table: GetTablePtr Returned Valid Pointer: %d [0x%x]\n", (unsigned int) datapointer, (unsigned int) datapointer);

    tableoffset = ((pagenumber * 0x10000) + card);
    printf("quad4table: Calculated Table Offset = %d [0x%x]\n", (unsigned int) tableoffset, (unsigned int) tableoffset);

    datapointer = datapointer + tableoffset;

    datavaluec = (unsigned char)  *(unsigned char*)  datapointer;
    datavalues = (unsigned short) *(unsigned short*) datapointer;
    datavaluel = (unsigned long)  *(unsigned long*)  datapointer; 
    datavaluef = (float)          *(float*)          datapointer;
    datavalued = (double)         *(double*)         datapointer;


    printf("quad4table: Adjusted Table Pointer = %d [0x%x]\n", 
           (unsigned int) datapointer, 
           (unsigned int) datapointer);
    printf("quad4table: Table Value (char) = [%c] [0x%x]\n", 
           (unsigned char) datavaluec, 
           (unsigned char) datavaluec);
    printf("quad4table: Table Value (short) = [%d] [0x%x]\n", 
           (unsigned int) datavalues, 
           (unsigned int) datavalues);
    printf("quad4table: Table Value (long) = %ld [0x%lx]\n", 
           (unsigned long) datavaluel, 
           (unsigned long) datavaluel);
    printf("quad4table: Table Value (float) = %f\n", 
           datavaluef);
    printf("quad4table: Table Value (double) = %f\n", 
           datavalued);

    printf("quad4table: Table Value = %s\n", datapointer);

    strncpy (buffer1, datapointer, 30); 

    printf("quad4table: Buffer1 String = %s\n", datapointer);

 
    return;
}


/*    End of File:    drvHart.c */
