/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
**  File: drvRtp.c
*/

#include <ctype.h>

#include <vxWorks.h>
#include <types.h>
#include <in.h>
#include <bootLib.h>
#include <inetLib.h>
#include <sysLib.h>
#include <semLib.h>
#include <errno.h>
#include <socket.h>
#include <selectLib.h>
#include <fcntl.h>
#include <ioLib.h>
#include <wdLib.h>
#include <rngLib.h>
#include <task_params.h>
#include <stdioLib.h>
#include <vme.h>
#include <string.h>
#include <logLib.h>
#include <remLib.h>
#include <tickLib.h>

#include <dbDefs.h>
#include <drvSup.h>
/*#include <module_types.h>*/
#include <cvnTypes.h>
#include <restable.h>
#include <devType.h>
#include <scan_types.h>
#include <tsDefs.h>
#include <drvRtp.h>				/* NOTE: change here */
#include <drvRtp_proto.h>			/* NOTE: change here */

extern TS_STAMP	  time_stamp;		/* defined in .../src/time/ts_driv.c */

struct
{
	long		number;
	DRVSUPFUN	report;
	DRVSUPFUN	init;
}
	drvRtp = { 2, NULL, init};

unsigned short		last_rtp_command;
unsigned short		last_rtp_data;

/*
**      External Line Level Data Structures (declared in drvcommon.c)
**
*/
	/* array of resource name strings.		*/

static	unsigned char	resnames [NUM_IOBC] [RESNAMELEN];

	/* array of pointers to resource tables 	*/

static	RES_TABLE	*prestable [NUM_IOBC];

	/* array of resource response flags. 		*/
	/* False = not active resource (slave),		*/
	/* True = active resource (master) 		*/

static	unsigned short	resresponse [NUM_IOBC];							
	/* true/false line active flag. calculated	*/
	/* using activity_time below. 			*/

static	unsigned short	line_activity [NUM_IOBC];

	/* assigned value of line_activity prior to	*/
	/* it's recalculation.   			*/

static	unsigned short	lastactivity [NUM_IOBC];

	/* controls output of communication error log	*/
	/* messages (true/false) 			*/

static	unsigned short	modtrace [NUM_IOBC];

	/* This value is set to TRUE if the partner is	*/
	/* reporting a status over the serial link	*/

static	unsigned short	partner_ok [NUM_IOBC];

	/* controls output of certain resource related	*/
	/* log messages (level) 			*/

static	unsigned short	mltrace [NUM_IOBC];

	/* controls output of MASTER/SLAVE arbitration	*/
	/* log messages (binary) 			*/

static	unsigned short	modrestrace [NUM_IOBC];

	/* holds timetick of last activity. used to	*/
	/* set line_activity flag  			*/

static	long		activity_time [NUM_IOBC];

	/* holds max time in seconds that driver will   */
	/* wait to obtain partner state on first        */
	/* attempting to decide if it should be master. */

static  unsigned short  pstatewait [NUM_IOBC];                                                        			
	/* True after database has been completely traversed */

extern	int		firstInitPass;

/*
**      Storage for Rtp Data structures
**
*/
	/* holds timestamp of beginning of major loop.  */

static unsigned int     lastscan [NUM_IOBC];

	/* holds actual time taken to scan plcs.        */

static unsigned int     scantime [NUM_IOBC];

	/* array of pointers to the     		*/
	/* fixed length dynamically allocated PLC       */
	/* control and data structures. extra pointer   */
	/* is for Query structure used independently    */
	/* of the others. it also allows for the array  */
	/* index to match the mod value.                */

static RTPIO		*plcs [NUM_IOBC] [MAX_RTP_ADDR];

static unsigned short	*vme_base [NUM_IOBC];

static unsigned short	*iobc_addr [NUM_IOBC];

static unsigned short	*bus_swch_addr [NUM_IOBC];

	/* controls output of system error messages:	*/
	/* bit mapped controls                          */

static	unsigned short	trace [NUM_IOBC];

	/* holds id and password of the user that       */
	/* spawned the driver task required by host to  */
	/* check file read privileges on PLC load files */

static char		user [40], password [40];

static char    *bools [] =
{
	"TRUE ",
	"FALSE"
};

int			RtptaskId [NUM_IOBC];

static char		*tasknames [] =
{
	"RTP Ch0","RTP Ch1","RTP Ch2","RTP Ch3"
};

	/* line driver group assignments   */

static unsigned short	rtpResGroups [NUM_IOBC] =
{
	0, 0, 0, 0
};


/*	Define the RTP Analog Output channel conversion table	*/

static	unsigned	short	rtp_ao_channel [] =
{
	RTP_AOT_CHAN_0,
	RTP_AOT_CHAN_1,
	RTP_AOT_CHAN_2,
	RTP_AOT_CHAN_3,
	RTP_AOT_CHAN_4,
	RTP_AOT_CHAN_5,
	RTP_AOT_CHAN_6,
	RTP_AOT_CHAN_7
};

/*	Define BO Recorder variables	*/

BO_RECORDER		borecorder [NUM_IOBC] [200];
unsigned char		borecidx [NUM_IOBC];
extern	TS_STAMP	time_stamp;

static SEM_ID           redsemids [NUM_IOBC];
static int		redsize [NUM_IOBC];
static int		Rtpsredperiod [NUM_IOBC];
static int		serredperiod [NUM_IOBC];
static int		Rtpsredport [NUM_IOBC];
static int		serredport [NUM_IOBC];

static	char		partner_alive [NUM_IOBC];
static	long		one_second [NUM_IOBC];
static	unsigned short	ai_points_scanned [NUM_IOBC];
static	unsigned short	bi_points_scanned [NUM_IOBC];
static	unsigned short	ao_points_scanned [NUM_IOBC];
static	unsigned short	bo_points_scanned [NUM_IOBC];
static	unsigned short	ai_scan_count [NUM_IOBC];	/* Number of points scanned this pass */
static	unsigned short	bi_scan_count [NUM_IOBC];	/* Number of points scanned this pass */
static	unsigned short	ao_scan_count [NUM_IOBC];	/* Number of points scanned this pass */
static	unsigned short	bo_scan_count [NUM_IOBC];	/* Number of points scanned this pass */
static	unsigned short	was_requested [NUM_IOBC];
static	unsigned short	is_responding [NUM_IOBC];
static	unsigned short	no_offlines [NUM_IOBC];
static	unsigned short	no_dataerr [NUM_IOBC];
static	unsigned short	cmd_reset_stat [NUM_IOBC];
static	unsigned short	cmd_force_offline [NUM_IOBC];
static	unsigned short	scan_devlimit [NUM_IOBC];
static	unsigned short	scan_deviation [NUM_IOBC];
static	unsigned short	out_requests [NUM_IOBC];
static	unsigned short	in_requests [NUM_IOBC];
static	unsigned short	scan_fault [NUM_IOBC];
static	unsigned short	no_local_activity [NUM_IOBC];
static	unsigned short	no_partner_activity [NUM_IOBC];
static	unsigned short	badoutcnt [NUM_IOBC];
static	unsigned short	is_redundant [NUM_IOBC];
static	unsigned short	ca_slave_want [NUM_IOBC];
static	unsigned short	ai_threshhold [NUM_IOBC];

static	long		tickspersecond;
static	long		p60second;
static	long		p30second;
static	long		p15second;
static	long		p10second;
static	long		p5second;
static	long		p2second;
static	long		p1second;
static	long		phalfsecond;
static	long		pfifthsecond;
static	long		ptenthsecond;


/*	The following variable is used for debugging	*/

char			rtp_test = FALSE;

static	int		rtpresmsgdisable = FALSE;

static	long		scan_delay [NUM_IOBC];

static	RTPSREDUPD	redund [NUM_IOBC];

static	unsigned short	mycount [NUM_IOBC];

extern  short           resGroupLock [];	/* holds line driver group lock flags. */

extern  char           *resGroupTripName [];	/* holds line driver trip name. */

static int checkgroup (unsigned short line, unsigned short *lstate, unsigned short tracenum);



/**********
*
*	This routine is required by iocinit in the EPICS system.
*
**********/
static long init ()
{
	return (0);
}

/********
**
**	This routine outputs information to help with debugging.
**
********/
void rtphelp (void)
{
    printf ("Rtp Driver Control & Display Functions");
    printf ("\n\n");
    printf ("rtptraceon (N) Enable the following debug messages for line N\n");
    printf (" 1) Major Scan loop Info:\n");
    printf ("    Scan time\n"); 
    printf ("    Backplane errors encountered in scan_plc\n");
    printf ("    All parametric errors encountered in the I/O drivers\n");
    printf ("\n");
    printf ("rtptraceoff (N) Disable all debug messages for line N\n");
    printf ("\n");
    printf ("rtprestraceon (N) Enable display of master/slave state\n");
    printf ("\n");
    printf ("rtprestraceoff (N) Disable display of master/slave state\n");
    printf ("\n");

    taskDelay (2 * 60); /* Time for operator to see it and pause the output */

    printf ("rtptrlv (N, X) Enable the following message for line N:                    \n");
    printf (" X = 0  Disable all trace level messages\n");
    printf (" X = 1  Resource definition error in MAIN\n");
    printf ("        Attempts to bring a plc online\n");
    printf (" X = 2  Calls to AODRIVER, BODRIVER\n");
    printf (" X = 10 Calls to AODRIVER, BODRIVER\n");
    printf ("\n");

    taskDelay (2 * 60); /* Time for operator to see it and pause the output */

    printf (" dumplinestats (N) Display line N comm stats\n");
    printf (" clearlinestats (N) Display and clear line N comm stats\n");
    printf (" dumpplcstats (N) Display line N, comm stats\n");
    printf (" clearplcstats (N) Display and clear line N status & errors\n");
    printf (" dump_mod_table (N) Display line N status, errors and data\n");
    printf (" rtpResGroup(N,G)      Assign line N to group G (valid G range is 1 to 16)         \n");
    printf ("\n");
}

/********
**
**	This routine enables logging of comm trace messages
**
********/
void rtptraceon (unsigned short line)
{
	modtrace [line] = TRUE;
}

/********
**
**	This routine disables logging of comm trace messages
**
********/
void rtptraceoff (unsigned short line)
{
	modtrace [line] = FALSE;
}

/********
**
**	This routine enables logging of resource trace messages
**
********/
void rtprestraceon (unsigned short line)
{
	modrestrace [line] = TRUE;
}

/********
**
**	This routine disables logging of resource trace messages
**
********/
void rtprestraceoff (unsigned short line)
{
	modrestrace [line] = FALSE;
}

/********
**
**	This routine controls logging of trace messages
**
********/
void rtptrlv (unsigned short line, unsigned short level)
{
	mltrace [line] = level;
}
void rtpresmsgon ()
{
	rtpresmsgdisable = FALSE;
}
void rtpresmsgoff ()
{
	rtpresmsgdisable = TRUE;
}

int stoprtp (short ln)
{
	taskDelete (RtptaskId [ln]);
	return (1);
}
void rtpsetaithreshhold (unsigned short line, unsigned short value)
{
	if (line < NUM_IOBC)
	{
	    ai_threshhold [line] = value;
	}
	else
	{
	    logMsg ("Line number must be less than %d\n", NUM_IOBC,0,0,0,0,0);
	    logMsg ("Use - rtpsetaithreshhold line, value",0,0,0,0,0,0);
	}
}
int startrtp (unsigned short line,
		unsigned short iobc_rtp_addr,
		unsigned short *iobc_vme_base,
		unsigned short bus_switch_addr,
		unsigned short redundant,
		unsigned short ca_slave_wnt,
		unsigned short port,
		unsigned short Rtpsperiod,
		unsigned short serport,
		unsigned short serperiod
		)
{

	if (line < NUM_IOBC)
	{
	    is_redundant [line] = redundant;
	    ca_slave_want [line] = ca_slave_wnt;

	    RtptaskId [line] = taskSpawn (tasknames [line],
					RTP_PRIORITY,
					RTP_TASK_OPTIONS,
					RTP_STACK,
					rtp_driver,
					line,
					iobc_rtp_addr,
					iobc_vme_base,
					bus_switch_addr,
					port,
					Rtpsperiod,
					serport,
					serperiod
				     0,0,0);

	    logMsg ("Started the RTP driver on line %d\n", line,0,0,0,0,0);

	    remCurIdGet (user, password);
	    return (1);
	}
	else
	{
	    logMsg ("Invalid line number %d\n", line,0,0,0,0,0);
	    return (-1);
	}
}

int rtpResGroup (unsigned short line, unsigned short group)
{
	if (line < NUM_IOBC)
	{
	    if (group <= MAXGROUPNUM)
	    {
		rtpResGroups [line] = group;
		logMsg ("Assigned Line %d to Group %d\n", line, group,0,0,0,0);
		return (1);
	    }
	    else
	    {
		logMsg ("Invalid Group Assignment: line %d: group number must be less than %d.\n", line, MAXGROUPNUM+1,0,0,0,0);
		return (-1);
	    }
	}
	else
	{
	    logMsg ("Invalid Group Assignment: line %d: line number must be less than %d\n", line, NUM_IOBC,0,0,0,0);
	    return (-1);
	}
}

void rtpgrouptable (void)
{
	unsigned short	x, y, i, j, index;

	printf (" Resource Group Lock Manager:  RTP Driver Line to Group Map Table\n");
	printf (" -----------------------------------------------------------------------\n");
	printf ("  Ln:Grp   Ln:Grp   Ln:Grp   Ln:Grp   Ln:Grp   Ln:Grp   Ln:Grp   Ln:Grp \n");

	x = 4;
	y = NUM_IOBC/x;

	for (i = 0; i < y; i++)
	{
	    for (j = 0; j < x; j++)
	    {
		index = i + j * y;
		printf ("  %d%d:%d%d  ",
			index / 10,
			index % 10,
			rtpResGroups [index] / 10,
			rtpResGroups [index] % 10);
	    }
	    printf ("\n");
	}
}

void rtplocktable (void)
{
	unsigned short x, y, i, j, index;

	printf (" Resource Group Lock Manager Global Array\n");
	printf (" ----------------------------------------\n");
	printf ("  Grp:Flg   Grp:Flg   Grp:Flg   Grp:Flg\n");

	x = 4;
	y = MAXGROUPNUM/x;

	for (i = 0; i < y; i++)
	{
	    for (j = 0; j < x; j++)
	    {
		index = i + j * y;
		printf ("  %d%d:%d%d   ",
			(index+1)/10,
			(index+1)%10,
			resGroupLock [index] / 10,
			resGroupLock [index] % 10);
	    }
	    printf ("\n");
	}
}

void rtplockall (void)
{
	unsigned short i;

	printf (" Resource Group Lock Manager Global Array\n");
	printf (" ----------------------------------------\n");
	printf ("  Group  Flag  Trip Name\n");
	for (i = 0; i < MAXGROUPNUM; i++)
	{
	    printf ("   %d%d     %d%d",
 			(i + 1) / 10,
			(i + 1) % 10,
			resGroupLock [i] / 10,
			resGroupLock [i] % 10);
	    if (resGroupTripName[i])
	    {
		printf ("   %s\n", *resGroupTripName[i]);
	    }
	    else
	    {
		printf ("\n");
	    }
	}
}

void rtpsetlock (unsigned short group)
{
	if (group <= MAXGROUPNUM)
	{
	    printf ("Setting Resource Lock Flag for Group %d.\n", group);
	    taskDelay(60);
	    resGroupLock [group - 1] = 1;
	}
	else
	{
	    logMsg ("Invalid Group %d: group number must be less than %d.\n", group, MAXGROUPNUM+1,0,0,0,0);
	}
}

void rtpclrlock (unsigned short group)
{
	if (group <= MAXGROUPNUM)
	{
	    printf ("Clearing Resource Lock Flag for Group %d.\n", group);
	    taskDelay(60);
	    resGroupLock [group -1] = 0;
	}
	else
	{
	    logMsg ("Invalid Group %d: group number must be less than %d.\n", group, MAXGROUPNUM+1,0,0,0,0);
	}
}

void rtpgrouphelp (void)
{
	printf (" RTP Driver Group Related Functions\n");
	printf ("\n");
	printf (" rtpResGroup(N, G)     Assign line N to group G (valid G range is 1 to 16)         \n");
	printf (" rtpgrouptable         Display the RTP's line to group assignment table.    \n");
	printf (" rtplocktable          Display the resource group lock Global Array.               \n");
	printf (" rtplockall            Display the resource group lock and trip name Global Array. \n");
	printf (" rtpsetlock (G)        Set the resource group lock for group G. (1 to 16)          \n");
	printf (" rtpclrlock (G)        Clear the resource group lock for group G. (1 to 16)        \n");
	printf (" rtpgrouphelp          Display this help message.                                  \n");
	printf ("\n");
}






/**********
*
*	RTP_driver
*
*	The purpose of this task is to scan the RTP I/O modules.
*	It will maintain a configuration list supporting online additions
*	and replacement of modules, and maintain a table of most recently
*	scanned values.
*
**********/                                                                                     	
/* not called from within this driver 	*/

void rtp_driver (unsigned short line,
			unsigned short iobc_rtp_addr,
			unsigned short *iobc_vme_addr,
			unsigned short bus_switch_addr,
			unsigned short port,
			unsigned short Rtpsperiod,
			unsigned short serport,
			unsigned short serperiod
			)
{
	unsigned short  mod, cfgchk, cnt, modspolled;
	unsigned short  activity, pstate, lstate, execute, command,i;
	long	        this_time, last_time;
	RTPIO          *pmod;
        unsigned int    temptick;
	unsigned short  group;


	/* Initialize the data structures */

	resresponse [line] = FALSE;
	resnames [line] [0] = 0;

	activity_time [line] = tickGet ();
	lastactivity [line] = line_activity [line] = FALSE;

	tickspersecond = 60;
	p60second = (60 * tickspersecond);
	p30second = (30 * tickspersecond);
	p15second = (15 * tickspersecond);
	p10second = (10 * tickspersecond);
	p5second = (5 * tickspersecond);
	p2second = (2 * tickspersecond);
	p1second = (1 * tickspersecond);
	phalfsecond = (tickspersecond / 2);
	pfifthsecond = (tickspersecond / 5);
	ptenthsecond = (tickspersecond / 10);
	scan_delay [line] = p60second;
	cmd_reset_stat [line] = 0;
	cmd_force_offline [line] = 0;
	ai_threshhold [line] = 400;

	redsize [line] = sizeof (RTPSREDUPD);
	Rtpsredperiod [line] = Rtpsperiod;
	serredperiod [line] = serperiod;
	Rtpsredport [line] = port;
	serredport [line] = serport;
	borecidx [line] = 0;

	/* wait for valid partner state for up to 60 seconds */	

	pstatewait [line] = 60;

        /* debug structures */

	modtrace [line] = FALSE;
	mltrace [line] = 0;
	modrestrace [line] = TRUE;


	sysBusToLocalAdrs (VME_AM_STD_SUP_DATA,  iobc_vme_addr, &vme_base [line]);
	iobc_addr [line] = (iobc_rtp_addr + vme_base [line]);
	if (bus_switch_addr != 0)
	{
	    bus_swch_addr [line] = ((bus_switch_addr * 2) + vme_base [line]);
	}
	else
	{
	    bus_swch_addr [line] = 0;
	}

	/* Attach to the serial redundant task */

	attachSerialRedund (line, 60);

	/* Wait for resource to be defined via an initial call
		to 'register_resource' via driver call.  */

	while ((resnames [line] [0] == 0) || (prestable [line] == NULL))
	{
	    if (mltrace [line] == 1)
	    {
	    	logMsg ( "No Resource Defined for line - %d \n", line,0,0,0,0,0);
	    }
	    taskDelay (600);
	}

	if (is_redundant [line])
	{

	    /* Start out as slave */

	    resClearCaAnswer (prestable [line]);		/* start quiet */
	    resPutLocalState (prestable [line], RES_SLV);
	    if (modrestrace [line])
	    {
		date ();
		logMsg ("1] Initialized as RES_SLV\n",0,0,0,0,0,0);
	    }

	    initRedDriverData (&resnames [line], &redund [line],
				&redsize [line], (FUNCPTR) slave_updateredimage,
				line, Rtpsredperiod [line], port,
				&redsemids [line]);

            /* wait finite time for pstate to be resolved */

            logMsg ("Begin wait for pstate to become valid.\n",0,0,0,0,0,0);
            for (i = 0; i <= pstatewait [line]; i++)
            {
        	resGetStates (prestable [line], &lstate, &pstate);
        	taskDelay (60);
            }
            if (pstate == 0)
	    {
        	logMsg ("Timeout of pstate wait.\n",0,0,0,0,0,0);
            }
	    else
	    {
        	logMsg ("Completion of pstate wait.\n",0,0,0,0,0,0);
	    }
	}

	resGetStates (prestable [line], &lstate, &pstate);

	activity = ((remoteSerialRedund (line) == 'M')
			&& ((pstate == 0) || (pstate == RES_ISOLATED)));

	/* if the partner is not there, be master */

	if ((pstate != RES_MSTR) && (pstate != RES_SLVWANT) && (!activity))
	{
	    resSetCaAnswer (prestable [line]);
	    resPutLocalState (prestable [line], RES_MSTR);

            lastscan [line] = tickGet ();
            scantime [line] = 0;

	    lstate = RES_MSTR;
	    if (modrestrace [line])
	    {
		date ();
	    	logMsg ("2] Became RES_MSTR   P= %x A= %d \n", pstate, activity,0,0,0,0);
	    }
	}

	cfgchk = 0;		/* Init the new config pointer */
	last_time = tickGet (); /* Initialize the time  */

	while (TRUE)
	{
	    activity_time [line] = tickGet ();
	    lastactivity [line] = line_activity [line] = FALSE;

            /************************************/
            /* check for line group disablement */
            /************************************/

            group = rtpResGroups [line];
            if ((group) && (resGroupLock [group-1]))
            {
                if (modrestrace [line])
                {
                    logMsg ("13] Forced to RES_SLV.\n",0,0,0,0,0,0);
                }
                resClearCaAnswer (prestable [line]);
                resPutLocalState (prestable [line], RES_SLV);
                lstate = RES_SLV;
                resresponse [line] = FALSE; 	/* Make Sure driver is not responding   */
                if (modrestrace [line])
                {
                    logMsg ("14] Resource Group Locked     P= %x A= %d \n", pstate, activity,0,0,0,0);
                }
                while (resGroupLock [group-1])
                {
 		   pulseSerialRedund (line, 'S');
                   group = rtpResGroups [line];
                    if (group)
                    {
                        taskDelay (2);
                    }
                    else /* the line has been taken out of the group */
                    {
                        break;
                    }
                }
                if (modrestrace [line])
                {
                    logMsg ("15] Resource Group Unlocked   P= %x A= %d \n", pstate, activity,0,0,0,0);
                    logMsg ("16] Released to RES_SLV.\n",0,0,0,0,0,0);
                }
            }

/******************************************************************************
*
* we arrive here because we are a slave.
* we stay until:
*		asked to be master by partner.
*		if we can reach PLCs, we become master.
*		the partner is not master and there is no activity on the line.
*
******************************************************************************/

	    while (lstate == RES_SLV)
	    {
 		/* make sure driver is not responding 	*/

		resresponse [line] = FALSE;

 		/* clear out commands 			*/

		resPutCommand (prestable [line], 0);

		taskDelay (2); /* slow down the refresh time */

		pulseSerialRedund (line, 'S');
		if ((partner_alive [line] = remoteSerialRedund (line)) != 'M')
		{
		    line_activity [line] = FALSE;
		}
		else
		{
		    line_activity [line] = TRUE;
		}
		if ((partner_alive [line] == 'M') ||
		    (partner_alive [line] == 'W') ||
		    (partner_alive [line] == 'S'))
		{
		    partner_ok [line] = TRUE;
		}
		else
		{
		    partner_ok [line] = FALSE;
		}

		resGetStates (prestable [line], &lstate, &pstate);

		activity = ((remoteSerialRedund (line) == 'M')
			&& ((pstate == 0) || (pstate == RES_ISOLATED)));

		/************************************/
		/* check for line group disablement */

		if (checkgroup (line, &lstate, 17))
		{
		    break;
		}

                /* see if needed by partner to be master.
		   if yes, do it and break out of slave state.   */

		if (pstate == RES_SLVWANTCA)
		{
		    prepresponses (line);
		    resSetCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_MSTR);
		    lstate = RES_MSTR;
		    resresponse [line] = TRUE;

                    lastscan [line] = tickGet ();
                    scantime [line] = 0;

		    while (remoteSerialRedund (line) == 'M')
		    {
			pulseSerialRedund (line, 'M');
			taskDelay (1);
		    }
		    if (modrestrace [line])
		    {
			date ();
			logMsg ("3] Became RES_MSTR  P= %x A= %d \n",
					pstate, activity,0,0,0,0);
		    }
		    break;
		}
                /* If partner is there and wants to give it up,
		   try to take it */

		if (pstate == RES_SLVWANT || remoteSerialRedund (line) == 'W')
		{
		    execute = FALSE;
		    taskDelay (2); /* Time for switch to occur */

		    /* See if anyone is out there and able to respond
			to a loopback command */

		    for (mod = 0; (mod <= MAX_RTP_ADDR); mod++)
		    {
			/************************************/
			/* check for line group disablement */

			if (checkgroup (line, &lstate, 18))
			{
			    break;
			}
			pmod = plcs [line] [mod];

			/* verify pointer validity prior to use */
			/* that data points have been requested */
			/* user has NOT forced plc offline      */

			if (pmod == 0)
			{
			    continue;
			}
			if ((pmod->was_requested) && (!pmod->cmd_force_offline))
			{
			    on_line (line, mod);	

                            /* check cmd_force_offline flag before AND after
				due to asynchronous tasks */

                            if ((pmod->is_responding) &&
				(!pmod->cmd_force_offline))
			    {
				execute = TRUE;
				break;
			    }
			}
		    }
		    /************************************/
		    /* check for line group disablement */

		    if (checkgroup (line, &lstate, 19))
		    {
			break;
		    }
		    /* If someone answered, become master and make sure
			partner really gave it up */

		    if (execute)
		    {
			resPutLocalState (prestable [line], RES_MSTR);
			lstate = RES_MSTR;

                        lastscan [line] = tickGet ();
                        scantime [line] = 0;

			if (modrestrace [line])
			{
			    date ();
			    logMsg ("4] Became RES_MSTR  P= %x A= %d \n",
					pstate, activity,0,0,0,0);
			}
			execute = FALSE;
			activity_time [line] = tickGet ();
			lastactivity [line] = line_activity [line] = FALSE;
			for (i = 0; (i < 60); i++)
			{
			    /* pause before testing */

			    resGetStates (prestable [line], &lstate, &pstate);
			    activity = ((remoteSerialRedund (line) == 'M')
				&& ((pstate == 0) || (pstate == RES_ISOLATED)));
			    if ((pstate != RES_MSTR) && (pstate != RES_SLVWANT)
				&& (!activity) )
			    {
				/* Partner Gave It Up so take CA */

				execute = TRUE;
				break;
			    }
			}
			if (!execute)
			{
			    /* Partner did not give it up go back to slave */

			    resPutLocalState (prestable [line], RES_SLV);
			    lstate = RES_SLV;
			    if (modrestrace [line])
			    {
				date ();
			    	logMsg ("5] Became RES_SLV  P= %x A= %d \n",
						pstate, activity,0,0,0,0);
			    }
			}
			else
			{
			    /* Assume control by taking CA responsibilities */ 

			    prepresponses (line);
			    resSetCaAnswer (prestable [line]);
			    break;
			}
		    }
		}

		/* if the partner is not talking, be master */

		if ((pstate != RES_MSTR) &&  (pstate != RES_SLVWANT) &&
			(!activity) && (remoteSerialRedund (line) != 'W'))
		{
		    resSetCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_MSTR);
		    lstate = RES_MSTR;

                    lastscan [line] = tickGet ();
                    scantime [line] = 0;

		    if (modrestrace [line])
		    {
			date ();
		    	logMsg ("6] Became RES_MSTR P= %x A= %d \n",
					pstate, activity,0,0,0,0);
		    }
		    prepresponses (line);
		}
	    }

/**********
*
* to arrive here, we used to be master, and had been told to relinquish
* master status.
* stay in this loop until the partner takes over, or the command is canceled
*
**********/
	    while (lstate == RES_SLVWANTCA)
	    {
		/************************************/
		/* check for line group disablement */

		if (checkgroup (line, &lstate, 20))
		{
		    break;
		}

 		/* Make Sure driver is not responding   */

		resresponse [line] = FALSE;

		master_updateredimage (line);  /* Get Last update over */

		pulseSerialRedund (line, 'W');

		taskDelay (15);

		if ((partner_alive [line] = remoteSerialRedund (line)) != 'M')
		{
		    line_activity [line] = FALSE;
		}
		else
		{
		    line_activity [line] = TRUE;
		}
		if ((partner_alive [line] == 'M') ||
		    (partner_alive [line] == 'W') ||
		    (partner_alive [line] == 'S'))
		{
		    partner_ok [line] = TRUE;
		}
		else
		{
		    partner_ok [line] = FALSE;
		}
		
		resGetStates (prestable [line], &lstate, &pstate);
		activity = ((remoteSerialRedund (line) == 'M')
			&& ((pstate == 0) || (pstate == RES_ISOLATED)));
		if ((pstate == RES_MSTR) || (pstate == RES_SLVWANT) || activity)
		{
		    resClearCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_SLV);
		    lstate = RES_SLV;
		    if (modrestrace [line])
		    {
			date ();
		    	logMsg ("7] Became RES_SLV  P= %x A= %d \n",
					pstate, activity,0,0,0,0);
		    }
		    for (i = 0; i < 12; i++)
		    {
			pulseSerialRedund (line, 'W');
			taskDelay (10);
		    }
		    break;
		}
		resGetCommand (prestable [line], &command);

		/* clear out commands */

		resPutCommand (prestable [line], 0);

		if (command == RES_CANCELREL)
		{
		    resSetCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_MSTR);
		    lstate = RES_MSTR;
		    if (modrestrace [line])
		    {
			date ();
		    	logMsg ("8] Became RES_MSTR  P= %x A= %d \n",
					pstate, activity,0,0,0,0);
		    }
		    prepresponses (line);
		}
	    }

/**********
*
* we come here because we have gone to, or are looping in, the master state.
* we stay in this loop until:
*	we have been told to give master state to the partner.
*       we break out to RES_SLVWANTCA.
*       we have successfully handed master state to partner.
*
* While master, this task will process all modules sending outputs
* and getting inputs. It will then try to configure requested
* non-responding modules.
*
**********/

	    last_time = tickGet ();		/* initialize gap */
	    while ((lstate == RES_MSTR) || (lstate == RES_SLVWANT))
	    {
		resGetStates (prestable [line], &lstate, &pstate);
		activity = ((remoteSerialRedund (line) == 'M')
			&& ((pstate == 0) || (pstate == RES_ISOLATED)));
		if (((pstate == RES_MSTR) || (activity) ) && (pstate != RES_SLVWANTCA))
		{
		    resClearCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_SLV);
		    lstate = RES_SLV;
		    if (modrestrace [line])
		    {
			date ();
			logMsg ("9] Became RES_SLV  P= %x A= %d \n", pstate, activity,0,0,0,0);
		    }
		    break;
		}
		/* Make sure driver is responding. driver won't */
                /* issue commands if not set.                   */

		resresponse [line] = TRUE;
		resGetCommand (prestable [line], &command);
		resPutCommand (prestable [line], 0);        /* clear out commands */

		pulseSerialRedund (line, 'M');
		if ((partner_alive [line] = remoteSerialRedund (line)) != 'S')
		{
		    line_activity [line] = FALSE;
		}
		else
		{
		    line_activity [line] = TRUE;
		}
		if ((partner_alive [line] == 'M') ||
		    (partner_alive [line] == 'W') ||
		    (partner_alive [line] == 'S'))
		{
		    partner_ok [line] = TRUE;
		}
		else
		{
		    partner_ok [line] = FALSE;
		}
		
                /* if relinquish command sent, change state and
			break out to outer loop */

		if (command == RES_RELINQUISH)
		{
		    master_updateredimage(line);  /* Get Last update over */
		    taskDelay(60);
		    resClearCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_SLVWANTCA);
		    lstate = RES_SLVWANTCA;
		    pulseSerialRedund (line, 'W');
		    if (modrestrace [line])
		    {
			date ();
			logMsg ("14] Became RES_SLVWANTCA  P= %x A= %d \n",
					pstate, activity,0,0,0,0);
		    }
		    break;
		}

		/* scan plcs for data i/o, if allowed */

		modspolled = 0;

		/* call 'scan_plc' for each PLC device, unless user has */
                /* taken it off line or it is not responding.           */

		for (mod = 0; (mod < MAX_RTP_ADDR); mod++)
		{
		    pmod = plcs [line] [mod];	/* shorthand pointer */
		    if (pmod)			/* verify pointer  */
		    {
			/************************************/
			/* check for line group disablement */

			if (checkgroup (line, &lstate, 22))
			{
			    break;
			}
			if (pmod->cmd_force_offline)
			{ 
			    /* if user has turned this plc offline
				 mark it as not responding. */

 			    pmod->is_responding = FALSE;
	    		}
			else if (pmod->is_responding)
			{
                            /* local/partner lockstep recovery */

                            resGetStates (prestable [line], &lstate, &pstate);
                            if (pstate == RES_MSTR)
                            {
                                resClearCaAnswer (prestable [line]);
                                resPutLocalState (prestable [line], RES_SLV);
                                lstate = RES_SLV;
                                if (modrestrace [line])
				{
				    date ();
                                    logMsg ("13] Became RES_SLV  P= %x A= %d\n",
					pstate, activity,0,0,0,0);
				}

				/* random 5 to 12 second delay */

                                taskDelay (((tickGet () % 8) + 5) * 60);
                                break;
                            }
			    modspolled++;
			    scan_plc (pmod, line, FALSE);	/* I/O done here! */
			    if (!pmod->is_responding)
			    {
				date ();
				logMsg ("RTP Rack not responding L %d A %d\n", line, mod,0,0,0,0);
/*				logMsg ("Output Memory Image Cleared !\n");	*/
/*				mark_not_valid (pmod, line);	*/
				++no_offlines [line];
                            }
			}
		    }
		    /* end of single module action */

		    /************************************/
		    /* test for line group disablement. */

		    if (lstate != RES_MSTR)
		    {
			break;
		    }

		    /* if told to relinquish our master state,
			stop plc scanning immediately. */
		    /* don't trash the command, so that we can still
			read it on the next loop */
                    /* through this while statement. */

		    resGetCommand (prestable [line], &command);
		    if (command == RES_RELINQUISH)
		    {
			break;
		    }
		} /* end of scan loop */

                /* measure how long the scan took.
			to be displayed at beginnig of next loop. */

                temptick = tickGet ();
                scantime [line] = temptick - lastscan [line];

		/* try and bring all nonresponding plcs on line, but */
                /* don't bring in more than a specified number.      */

		cnt = 0;

 		/* do we try to bring any in on this pass? */

		pulseSerialRedund (line, 'M');
		for (mod = 0; (mod < MAX_RTP_ADDR); mod++)
		{
		    pmod = plcs [line] [mod];
		    if (pmod == 0)		/* verify pointer */
		    {
			continue;
		    }	
		    if ((!pmod->is_responding) && (pmod->was_requested) &&
			(!pmod->cmd_force_offline))
		    {
			/* loopback testing done here! 	*/

			on_line (line, mod);
			modspolled++;
	            }
		}

		/************************************/
		/* test for line group disablement. */

		if (lstate != RES_MSTR)
		{
		    break;
		}


		/* see if anyone on the line has been responding. */
                /* if not, try to get partner to become slave.    */
		/* Check for no one responding */

		if (!any_responding (line) && is_redundant [line])
		{
		    /* See if it is time to let partner try */

		    this_time = tickGet ();
		    if ((this_time - last_time) > (3 * SLWWINDOW))
		    {
			/* #ticks for slave to try */

			master_updateredimage (line);  /* Get Last update over */
			taskDelay(60);

			/* Switch States to let partner know it's time to try */

			resPutLocalState (prestable [line], RES_SLVWANT);
			if (ca_slave_want [line])
			{
			    resClearCaAnswer(prestable [line]); /* Stop processing blocks */
			}
			lstate = RES_SLVWANT;
			if (modrestrace [line])
			{
			    date ();
			    logMsg ("10] Became RES_SLVWANT  P= %x A= %d \n",
					pstate, activity,0,0,0,0);
			}

			/* Remember the time of this try */

			last_time = tickGet ();
			lastactivity [line] = line_activity [line] = FALSE;
			for (i = 0; (i < SLWWINDOW); i += 10)
			{
			    /* SLWINDOW is the # of ticks for slave to try */

			    /************************************/
			    /* check for line group disablement */

			    if (checkgroup (line, &lstate, 25))
			    {
				break;
			    }

			    taskDelay (10);

			    /* Let partner know that we are alive via serial link */

			    pulseSerialRedund (line, 'W');

		    	    /* If partner picked up let it have control */
			    /* check immediately */

			    resGetStates (prestable [line], &lstate, &pstate);
			    activity = ((remoteSerialRedund (line) == 'M')
				&& ((pstate == 0) || (pstate == RES_ISOLATED)));

			    if ((pstate == RES_MSTR) || (pstate == RES_SLVWANT)
				|| (activity))
			    {
				resClearCaAnswer (prestable [line]);
				resPutLocalState (prestable [line], RES_SLV);
				lstate = RES_SLV;
				if (modrestrace [line])
				{
				    date ();
				    logMsg ("11] Became RES_SLV  P= %x A= %d\n",
						pstate, activity,0,0,0,0);
				}
				taskDelay (240);
				break;
			    }
			}
			/* Go Back to being the Master and take
				control of the hardware */

			if (lstate != RES_SLV)
			{
			    resSetCaAnswer (prestable [line]);
			    resPutLocalState (prestable [line], RES_MSTR);
			    lstate = RES_MSTR;
			    if (modrestrace [line])
			    {
				date ();
			    	logMsg ("12] Became RES_MSTR  P= %x A= %d\n",
						pstate, activity,0,0,0,0);
			    }
			}
		    }
		}

		/* take plcs off hold status. the plc's data are not
			to be held off since now master.   */

		for (mod = 0; (mod < MAX_RTP_ADDR); mod++ )
		{
		    pmod = plcs [line] [mod];
		    if (pmod)			/* verify pointer */
		    {        
			pmod->holdflag = FALSE;
                    }
		}
		if (!any_responding (line))
		{
		    is_responding [line] = FALSE;
		    taskDelay (60);
		}
		else
		{
		    is_responding [line] = TRUE;
		    taskDelay (6);
		}
	    }
	} /* end infinite task loop */
}

/**********
*
* check for line group disablement
*
**********/

static int checkgroup (unsigned short line, unsigned short *lstate, unsigned short tracenum)
{
	unsigned short group;

	group = rtpResGroups [line];
	if ((group) && (resGroupLock [group-1]))
	{
	    *lstate = RES_LOCKED;
	    if (modrestrace [line])
	    {
		logMsg ("%d] Resource Group Lock Detected.\n", tracenum,0,0,0,0,0);
	    }
	    return (TRUE);
	}
	return (FALSE);
}

/**********
*
*	Try to bring a PLC on line
*
**********/

/*	this routine only called from MAIN	*/

static void on_line (unsigned short line, unsigned short mod)
{
	RTPIO			*pmod;


	pmod = plcs [line] [mod];
	if (pmod)			/* verify pointer prior to use */
	{
	    if (mltrace [line] == 1)
	    {
	        logMsg ( "Bring Online line - %d\n", line,0,0,0,0,0);
	    }

	    if (bus_swch_addr [line] != 0)
	    {
		*(bus_swch_addr [line]) = 0;	/* Select Normal operation mode */
		*(bus_swch_addr [line] + 1) = RTP_BUS_MASTER | RTP_BUS_CLEAR;
	    }

	    if (!firstInitPass)
	    {
		return;
	    }
	    if (scan_plc (pmod, line, FALSE))
	    {
	        pmod->is_responding = TRUE;
	    }
        }
}
/**********
*							
*	Mark all plc data elements as VME_TIMEOUT / INVALID
*                                             		
**********/                                             	

/*	called from on_line and prepresponses	*/    	

static void mark_not_valid (RTPIO  *pmod, unsigned short line)
{
	unsigned	short	j, k;

	if (pmod->rtp.rtp_cmd == RTP_EMPTY)
	{
	    return;
	}
	for (j = 0; j < MAX_GATE_ADDR; j++)
	{
	    switch (pmod->rtp.io [j].rtp_type)
	    {
	    case SLOT_EMPTY:
		break;

	    case RTP_AIN:
		for (k = 0; k < MAX_RTP_AIN; k++)
		{
		    pmod->rtp.io [j].pio.ainio [k].is_responding = FALSE;
		    pmod->rtp.io [j].pio.ainio [k].next_scan = 0;
		}
		pmod->rtp.io [j].status = RTP_RESPONDING;
		break;

	    case RTP_GATE_4:
		break;

	    case RTP_AOT_1:
		for (k = 0; k < MAX_RTP_AOT_1; k++)
		{
		    pmod->rtp.io [j].pio.aotio [k].is_responding = FALSE;
		    pmod->rtp.io [j].pio.aotio [k].out_requested = FALSE;
		    pmod->rtp.io [j].pio.aotio [k].out_value = 0;
		    pmod->rtp.io [j].pio.aotio [k].next_scan = 0;
		}
		pmod->rtp.io [j].status = RTP_RESPONDING;
		break;

	    case RTP_DIN_8:
		for (k = 0; k < MAX_RTP_DIN_8; k++)
		{
		    pmod->rtp.io [j].pio.dinio [k].is_responding = FALSE;
		    pmod->rtp.io [j].pio.dinio [k].next_scan = 0;
		}
		pmod->rtp.io [j].status = RTP_RESPONDING;
		break;

	    case RTP_DOT_8:
		for (k = 0; k < MAX_RTP_DOT_8; k++)
		{
		    pmod->rtp.io [j].pio.dotio [k].is_responding = FALSE;
		    pmod->rtp.io [j].pio.dotio [k].out_requested = FALSE;
		    pmod->rtp.io [j].pio.dotio [k].out_value = 0;
		    pmod->rtp.io [j].pio.dotio [k].next_scan = 0;
		}
		pmod->rtp.io [j].status = RTP_RESPONDING;
		break;

	    default:
		if (mltrace [line] == 5)
		{
		    logMsg ("***Error*** Invalid module type in structure\n",0,0,0,0,0,0);
		}
		break;
	    }
	}
}

/**********
*
*	Initialze IOBC line statistics
*
**********/

static void initiobcstats (link)
{
	ai_points_scanned [link] = 0;
	bi_points_scanned [link] = 0;
	ao_points_scanned [link] = 0;
	bo_points_scanned [link] = 0;
	ai_scan_count [link] = 0;
	bi_scan_count [link] = 0;
	ao_scan_count [link] = 0;
	bo_scan_count [link] = 0;
	no_offlines [link] = 0;
	no_dataerr [link] = 0;
	scan_devlimit [link] = 0;
	scan_deviation [link] = 0;
	out_requests [link] = 0;
	in_requests [link] = 0;
	scan_fault [link] = 0;
	no_local_activity [link] = 0;
	no_partner_activity [link] = 0;
	badoutcnt [link] = 0;
}

/**********
*
*	Initialize rack statistics
*
**********/

static void initrackstats (RTPIO *pmod)
{
	int	i;

	pmod->rack_tmouts = 0;
	for (i = 0; i < MAX_GATE_ADDR; i++)
	{
	    pmod->rtp.io [i].tmouts = 0;
	}
}

/**********
*
*	Scan an RTP I/O Rack -  get inputs and set outputs as required
*                                                            	
**********/

/*	this routine only called from MAIN 	*/

unsigned int	rtp_bus_delay = 0;

static	scan_plc (RTPIO *pmod, unsigned short line, unsigned short test)
{
	unsigned short	j, k, t;
	unsigned int	i;
	unsigned short	command_word;
	unsigned short	data_word;
        unsigned short  status;
        unsigned short  isai[MAX_GATE_ADDR],aiisresp[MAX_GATE_ADDR];
	unsigned short	any_ais_responding;
	unsigned long	nextscan;

	if (pmod->rtp.rtp_cmd == RTP_EMPTY)
	{
	    return (FALSE);
	}
	if (bus_swch_addr [line] != 0)
	{
	    *(bus_swch_addr [line]) = 0;	/* Select Normal operation mode */
	    *(bus_swch_addr [line] + 1) = RTP_BUS_MASTER | RTP_BUS_CLEAR;
	}

	if (one_second [line] == 0)
	{
	    one_second [line] = tickGet ();
	}
	else if ((one_second [line] + 60) <= tickGet ())
	{
	    ai_scan_count [line] = ai_points_scanned [line];
	    bi_scan_count [line] = bi_points_scanned [line];
	    ao_scan_count [line] = ao_points_scanned [line];
	    bo_scan_count [line] = bo_points_scanned [line];
	    ai_points_scanned [line] = 0;
	    bi_points_scanned [line] = 0;
	    ao_points_scanned [line] = 0;
	    bo_points_scanned [line] = 0;
	    one_second [line] = tickGet ();
	}
	    
	for (j = 0; j < MAX_GATE_ADDR; j++)
	{
	    switch (pmod->rtp.io [j].rtp_type)
	    {
	    case SLOT_EMPTY:
		break;

	    case RTP_AIN:
		/* Clear notions of which slots have ai gate cards and are responding */
		for (k = 0; k < MAX_GATE_ADDR; k++)
		{
		    isai[k] = FALSE;
		    aiisresp[k] = FALSE;
		}
		nextscan = tickGet ();
		for (k = 0; k < MAX_RTP_AIN; k++)
		{
		    if (pmod->rtp.io [j].pio.ainio [k].was_requested)
		    {

			/* Remember that this slot has a requested AI gate card in it */
			isai[k/MAX_GATE_SIG_4] = TRUE;

			if (pmod->rtp.io [j].pio.ainio [k].next_scan > nextscan)
			{
			    continue;
			}
			if ((scan_deviation [line]= 
				nextscan - pmod->rtp.io [j].pio.ainio [k].next_scan) >
				((scan_devlimit [line]/ 16) + 1))
			{
			    scan_fault [line]= TRUE;
			}
			else
			{
			    scan_fault [line] = FALSE;
			}
			if (pmod->rtp.io [j].pio.ainio [k].scanrate < scan_delay [line] && pmod->rtp.io [j].pio.ainio [k].scanrate != 0)
			{
			    scan_delay [line] = pmod->rtp.io [j].pio.ainio [k].scanrate;
			}
			pmod->rtp.io [j].pio.ainio [k].next_scan = nextscan + pmod->rtp.io [j].pio.ainio [k].scanrate;
			in_requests [line] = (++in_requests [line]) & 0x7fff;
			command_word = RTP_AIN_INT_DIS | RTP_AIN_INT_DEF;
			command_word |= j &  RTP_AIN_CVT_MSK;
			*pmod->rtp.rtp_cmd = last_rtp_command = command_word;

			for (i = 0; i < rtp_bus_delay; i++);

/*			taskDelay (2);	*/

			if (pmod->rtp.io [j].pio.ainio [k].devtype == ai_RTP_AI)
			{
			    data_word = 0;
/*?			    data_word |= RTP_AIN_OTD_ENA;	*/
			    data_word |= RTP_AIN_INP_DIF;
/*			    if (k & 0x01)
			    {
				data_word |= RTP_AIN_INP_SEO;
			    }
			    else
			    {
				data_word |= RTP_AIN_INP_SEE;
			    }
*/
			    data_word |= ((k / MAX_GATE_SIG_4) << RTP_AIN_GAT_SFT);
			    data_word |= ((k % MAX_GATE_SIG_4) << RTP_AIN_SIG_SFT);
			    switch (pmod->rtp.io [j].pio.ainio [k].gain)
			    {
			    case RTP_AIN_GAIN_1:
				break;

			    case RTP_AIN_GAIN_2:
				data_word |= 0x01;
				break;

			    case RTP_AIN_GAIN_4:
				data_word |= 0x02;
				break;

			    case RTP_AIN_GAIN_8:
				data_word |= 0x03;
				break;
			    }

			    *pmod->rtp.rtp_data = last_rtp_data = data_word;

			}
			else
			{
			    /* Process the analog test card points here */

			    data_word = 0;
			    data_word |= RTP_AIN_INP_DIF;
			    data_word |= ((k / MAX_GATE_SIG_4) << RTP_AIN_GAT_SFT);
			    data_word |= ((k % MAX_GATE_SIG_4) << RTP_AIN_SIG_SFT);
			    data_word |= RTP_AIN_CAL_PLUS;
			    data_word |= RTP_AIN_CAL_GAIN1;
			    *pmod->rtp.rtp_data = last_rtp_data = data_word;
			}

			/* read the ai */


			for (i = 0; i < rtp_bus_delay; i++);

/*			taskDelay (5);		*/

			for (t = 0; t < RTP_MAX_TST_CNT; t++)
			{
			    status = *pmod->rtp.rtp_cmd;
			    if (status & RTP_BUS_DIS_MSK)
			    {
				break;
			    }
			    if (status & RTP_DEV_SEL_MSK)
			    {
				break;
			    }
			    if ((status & RTP_DEV_RDY_MSK) == 0)
			    {
				break;
			    }
			    taskDelay (1);
			}
			if ((status & (RTP_BUS_DIS_MSK | RTP_DEV_SEL_MSK | RTP_DEV_RDY_MSK)) == 0)
			{
			    if (test)
			    {
				return (TRUE);
			    }
			    data_word = *(pmod->rtp.rtp_data);
			    if (pmod->rtp.io [j].pio.ainio [k].value < data_word)
			    {
				if ((data_word - pmod->rtp.io [j].pio.ainio [k].value) >= ai_threshhold [line])
				{
				    date ();
				    logMsg ("AI Threshold exceeded... L %d C %d S %d ov %d nv %d\n",
						line, j,
						(((k / MAX_GATE_SIG_4) *100) +
						(k % MAX_GATE_SIG_4)),
						pmod->rtp.io [j].pio.ainio [k].value,
						data_word,0);
				}
			    }
			    else
			    {
				if ((pmod->rtp.io [j].pio.ainio [k].value - data_word) >= ai_threshhold [line])
				{
				    date ();
				    logMsg ("AI Threshold exceeded... L %d C %d S %d ov %d nv %d\n",
						line, j,
						(((k / MAX_GATE_SIG_4) *100) +
						(k % MAX_GATE_SIG_4)),
						pmod->rtp.io [j].pio.ainio [k].value,
						data_word,0);
				}
			    }

			    pmod->rtp.io [j].pio.ainio [k].is_responding = TRUE;
			    pmod->rtp.io [j].pio.ainio [k].value = data_word;
			    ++ai_points_scanned [line];
			}
			else
			{
			    pmod->rtp.io [j].pio.ainio [k].is_responding = FALSE;
			    pmod->rtp.io [j].pio.ainio [k].next_scan = 0;
		    	    ++pmod->rtp.io [j].tmouts;
			    if (mltrace [line] == 2)
			    {
				logMsg ("AI not responding %d %d\n", j, k,0,0,0,0);
			    }
			}
		    }
		    checkoutputs (line);
		}

		/* May not have scanned any AIs so must review the whole list */
		/* Check to see if any points on this ADC are Responding */
		any_ais_responding = 0;

		for (k = 0; k < MAX_RTP_AIN; k++)
		{
		    if (pmod->rtp.io [j].pio.ainio [k].is_responding)
		    {
			/* Remember that this slot has a responding AI gate card in it */
			aiisresp[k/MAX_GATE_SIG_4] = TRUE;
			any_ais_responding = 1;
		    }
		}

		/* Set the status of the ADC Card Accordingly */
		if (any_ais_responding)
		{
		    pmod->rtp.io [j].status = RTP_RESPONDING;
		}
		else
		{
		    pmod->rtp.io [j].status = RTP_NORESPONSE;
		}

		/* 

		** Go back over the slots and set the state for ai gate cards */
		for (k = 0; k < MAX_GATE_ADDR; k++)
		{
		    if (isai[k])
		    {

			if (aiisresp[k])
			{
			    pmod->rtp.io [k].status = RTP_RESPONDING;
			}
			else
			{
			    pmod->rtp.io [k].status = RTP_NORESPONSE;
			}
		    }
		}
		break;

	    case RTP_AOT_1:
		for (k = 0; k < MAX_RTP_AOT_1; k++)
		{
		    if (pmod->rtp.io [j].pio.aotio [k].was_requested)
		    {
			break;
		    }
		}
		if (k == MAX_RTP_AOT_1)
		{
		    continue;
		}

		/* The AO's do not provide read capability */
		/* Just verify that the card is in the rack */

		nextscan = tickGet ();
		if (pmod->rtp.io [j].pio.aotio [k].next_scan > nextscan)
		{
		    continue;
		}
		if ((scan_deviation [line]= nextscan - pmod->rtp.io [j].pio.aotio [k].next_scan) >
		    ((scan_devlimit [line] / 16) + 1))
		{
		    scan_fault [line] = TRUE;
		}
		else
		{
		    scan_fault [line] = FALSE;
		}
		if (pmod->rtp.io [j].pio.aotio [k].scanrate < scan_delay [line] && pmod->rtp.io [j].pio.aotio [k].scanrate != 0)
		{
		    scan_delay [line] = pmod->rtp.io [j].pio.aotio [k].scanrate;
		}
		pmod->rtp.io [j].pio.aotio [k].next_scan = nextscan + pmod->rtp.io [j].pio.aotio [k].scanrate;
		command_word = 0;
		command_word |= RTP_AOT_MOD_RAN;
		command_word |= (j & RTP_AOT_GAT_MSK);
		*pmod->rtp.rtp_cmd = command_word;


		for (i = 0; i < rtp_bus_delay; i++);

/*		taskDelay (1);		*/

		for (t = 0; t < RTP_MAX_TST_CNT; t++)
		{
		    status = *pmod->rtp.rtp_cmd;
		    if (status & RTP_BUS_DIS_MSK)
		    {
			break;
		    }
		    if (status & RTP_DEV_SEL_MSK)
		    {
			break;
		    }
		    if ((status & RTP_DEV_RDY_MSK) == 0)
		    {
			break;
		    }
/*		    taskDelay (1);	*/
		}
		if ((status & (RTP_BUS_DIS_MSK | RTP_DEV_SEL_MSK | RTP_DEV_RDY_MSK)) == 0)
		{
		    if (test)
		    {
			return (TRUE);
		    }
		    for (k = 0; k < MAX_RTP_AOT_1; k++)
		    {
			if (pmod->rtp.io [j].pio.aotio [k].was_requested)
			{
			    pmod->rtp.io [j].pio.aotio [k].is_responding = TRUE;
			}
		    }
		    pmod->rtp.io [j].status = RTP_RESPONDING;
		    ++ao_points_scanned [line];
		}
		else
		{
		    for (k = 0; k < MAX_RTP_AOT_1; k++)
		    {
			pmod->rtp.io [j].pio.aotio [k].is_responding = FALSE;
			pmod->rtp.io [j].pio.aotio [k].next_scan = 0;
		    }
		    pmod->rtp.io [j].status = RTP_NORESPONSE;
		    ++pmod->rtp.io [j].tmouts;
		    if (mltrace [line] == 2)
		    {
			logMsg ("AO not responding %d %d\n", j, k,0,0,0,0);
		    }
		}

		checkoutputs (line);

		break;

	    case RTP_DIN_8:

		/* read the di */

		nextscan = tickGet ();
		for (k = 0; k < MAX_RTP_DIN_8; k++)
		{
		    if (pmod->rtp.io [j].pio.dinio [k].next_scan > nextscan)
		    {
			continue;
		    }
		    else
		    {
			break;
		    }
		}
		if (k == MAX_RTP_DIN_8)
		{
		    continue;
		}
		for (k = 0; k < MAX_RTP_DIN_8; k++)
		{
		    if ((scan_deviation [line] =
			nextscan - pmod->rtp.io [j].pio.dinio [k].next_scan) >
			((scan_devlimit [line] / 16) + 1))
		    {
			scan_fault [line] = TRUE;
		    }
		    else
		    {
			scan_fault [line] = FALSE;
		    }
		}
		for (k = 0; k < MAX_RTP_DIN_8; k++)
		{
		    if (pmod->rtp.io [j].pio.dinio [k].scanrate < scan_delay [line] && pmod->rtp.io [j].pio.dinio [k].scanrate != 0)
		    {
			scan_delay [line] = pmod->rtp.io [j].pio.dinio [k].scanrate;
		    }
		    pmod->rtp.io [j].pio.dinio [k].next_scan =
			nextscan + pmod->rtp.io [j].pio.dinio [k].scanrate;
		}
		in_requests [line] = (++in_requests [line]) & 0x7fff;
		command_word = 0;
		command_word |= RTP_DIN_INT_DIS;
		command_word |= (j & RTP_DIN_GAT_MSK);
		command_word |= RTP_DIN_LO_BANK;
		*pmod->rtp.rtp_cmd = command_word;

		for (i = 0; i < rtp_bus_delay; i++);

		taskDelay (1);

		for (t = 0; t < RTP_MAX_TST_CNT; t++)
		{
		    status = *pmod->rtp.rtp_cmd;
		    if (status & RTP_BUS_DIS_MSK)
		    {
			break;
		    }
		    if (status & RTP_DEV_SEL_MSK)
		    {
			break;
		    }
		    if ((status & RTP_DEV_RDY_MSK) == 0)
		    {
			break;
		    }
		    taskDelay (1);
		}
		if ((status & (RTP_BUS_DIS_MSK | RTP_DEV_SEL_MSK | RTP_DEV_RDY_MSK)) == 0)
		{
		    if (test)
		    {
			return (TRUE);
		    }
		    data_word = *(pmod->rtp.rtp_data);
		    for (k = 0; k < (MAX_RTP_DIN_8 / 2); k++)
		    {
			pmod->rtp.io [j].pio.dinio [k].value = (((data_word << k) & 0x8000) != 0);
			if (pmod->rtp.io [j].pio.dinio [k].was_requested)
			{
			    pmod->rtp.io [j].pio.dinio [k].is_responding = TRUE;
			}
		    }
		    pmod->rtp.io [j].status = RTP_RESPONDING;
		    bi_points_scanned [line] += 16;
		}
		else
		{
		    for (k = 0; k < MAX_RTP_DIN_8; k++)
		    {
			pmod->rtp.io [j].pio.dinio [k].is_responding = FALSE;
			pmod->rtp.io [j].pio.dinio [k].next_scan = 0;
		    }
		    pmod->rtp.io [j].status = RTP_NORESPONSE;
		    ++pmod->rtp.io [j].tmouts;
		    if (mltrace [line] == 2)
		    {
			logMsg ("BI not responding %d %d\n", j, k,0,0,0,0);
		    }
		}

		checkoutputs (line);

		break;

	    case RTP_DOT_8:
		for (k = 0; k < MAX_RTP_DOT_8; k++)
		{
		    if (pmod->rtp.io [j].pio.dotio [k].was_requested)
		    {
			break;
		    }
		}
		if (k == MAX_RTP_DOT_8)
		{
		    continue;
		}

		/* read the do */

		nextscan = tickGet ();
		for (k = 0; k < MAX_RTP_DOT_8; k++)
		{
		    if (pmod->rtp.io [j].pio.dotio [k].next_scan > nextscan)
		    {
		        continue;
		    }
		    else
		    {
			break;
		    }
		}
		if (k == MAX_RTP_DOT_8)
		{
		    continue;
		}
		for (k = 0; k < MAX_RTP_DOT_8; k++)
		{
		    if ((scan_deviation [line] = nextscan - pmod->rtp.io [j].pio.dotio [k].next_scan) >
		    ((scan_devlimit [line] / 16) + 1))
		    {
		        scan_fault [line] = TRUE;
		    }
		    else
		    {
		        scan_fault [line] = FALSE;
		    }
		    if (pmod->rtp.io [j].pio.dotio [k].scanrate < scan_delay [line] && pmod->rtp.io [j].pio.dotio [k].scanrate != 0)
		    {
			scan_delay [line] = pmod->rtp.io [j].pio.dotio [k].scanrate;
		    }
		    pmod->rtp.io [j].pio.dotio [k].next_scan = nextscan + pmod->rtp.io [j].pio.dotio [k].scanrate;
		}
		command_word = 0;
		command_word |= RTP_DOT_INT_DIS;
		command_word |= (j & RTP_DOT_GAT_MSK);
		command_word |= RTP_DOT_LO_BANK;
		*pmod->rtp.rtp_cmd = command_word;


		for (i = 0; i < rtp_bus_delay; i++);

		taskDelay (1);

		for (t = 0; t < RTP_MAX_TST_CNT; t++)
		{
		    status = *pmod->rtp.rtp_cmd;
		    if (status & RTP_BUS_DIS_MSK)
		    {
			break;
		    }
		    if (status & RTP_DEV_SEL_MSK)
		    {
			break;
		    }
		    if ((status & RTP_DEV_RDY_MSK) == 0)
		    {
			break;
		    }
		    taskDelay (1);
		}
		if ((status & (RTP_BUS_DIS_MSK | RTP_DEV_SEL_MSK | RTP_DEV_RDY_MSK)) == 0)
		{
		    if (test)
		    {
			return (TRUE);
		    }
		    for (k = 0; k < MAX_RTP_DOT_8; k++)
		    {
			if (pmod->rtp.io [j].pio.dotio [k].was_requested)
			{
			    pmod->rtp.io [j].pio.dotio [k].is_responding = TRUE;
			}
		    }
		    pmod->rtp.io [j].status = RTP_RESPONDING;
		    bo_points_scanned [line] += 16;
		}
		else
		{
		    for (k = 0; k < MAX_RTP_DOT_8; k++)
		    {
			pmod->rtp.io [j].pio.dotio [k].is_responding = FALSE;
			pmod->rtp.io [j].pio.dotio [k].next_scan = 0;
		    }
		    pmod->rtp.io [j].status = RTP_NORESPONSE;
		    ++pmod->rtp.io [j].tmouts;
		    if (mltrace [line] == 2)
		    {
			logMsg ("BO not responding %d %d\n", j, k,0,0,0,0);
		    }
		}

		checkoutputs (line);

		break;

	    case RTP_GATE_4:
		break;

	    default:
		if (mltrace [line] == 6)
		{
		    logMsg ("***Error*** Invalid module type in structure\n",0,0,0,0,0,0);
		}
		break;
	    }
	}
	if (test)
	{
	    return (FALSE);
	}
	for (j = 0; j < MAX_GATE_ADDR; j++)
	{
	    if ( (pmod->rtp.io [j].status == RTP_RESPONDING) &&
		  (pmod->rtp.io [j].rtp_type != SLOT_EMPTY) )
	    {
		line_activity [line] = TRUE;
		pmod->is_responding = TRUE;
		break;
	    }
	}
	if (j == MAX_GATE_ADDR)
	{
	    pmod->is_responding = FALSE;
	    ++pmod->rack_tmouts;
	    return (FALSE);
	}
	redund [line].mastercnt++;
	mycount [line] = redund [line].mastercnt;
	master_updateredimage (line);
	return (TRUE);
}

/***********
*
*	checkoutputs
*
*	checks output que and returns if empty.
*	if there is something on the que it attempts to process all
*	pending outputs doing imediate retries. If attempts fail the
*	request is dropped
*
**********/
static	void	checkoutputs (unsigned short line)
{
	unsigned short	i, j, k, l, t;
	unsigned short	command_word;
	volatile unsigned short	data_word;
	unsigned short	status;
	RTPIO  	*pmod;

	for (i = 0; i < MAX_RTP_ADDR; i++)
	{
	    pmod = plcs [line] [i];
	    if (pmod == NULL)
	    {
		continue;
	    }
	    if (pmod->rtp.rtp_cmd == NULL)
	    {
		continue;
	    }
	    if (pmod->out_requested == 0)
	    {
		continue;
	    }

	    pmod->out_requested = FALSE;

	    for (j = 0; j < MAX_GATE_ADDR; j++)
	    {
		switch (pmod->rtp.io [j].rtp_type)
		{
		case SLOT_EMPTY:
		    break;

		case RTP_AIN:
		    break;

		case RTP_GATE_4:
		    break;

		case RTP_DIN_8:
		    break;

		case RTP_AOT_1:
		    for (k = 0; k < MAX_RTP_AOT_1; k++)
		    {
			if (pmod->rtp.io [j].pio.aotio [k].out_requested)
			{
			    out_requests [line] = (++out_requests [line]) & 0x7fff;
			    command_word = 0;
			    command_word |= RTP_AOT_MOD_RAN;
			    command_word |= (j & RTP_AOT_GAT_MSK);
			    *pmod->rtp.rtp_cmd = command_word;

			    for (l = 0; l < rtp_bus_delay; i++);

			    taskDelay (1);

			    for (t = 0; t < RTP_MAX_TST_CNT; t++)
			    {
				status = *pmod->rtp.rtp_cmd;
				if (status & RTP_BUS_DIS_MSK)
				{
				    break;
				}
				if (status & RTP_DEV_SEL_MSK)
				{
				    break;
				}
				if ((status & RTP_DEV_RDY_MSK) == 0)
				{
				    break;
				}
				taskDelay (1);
			    }
			    if ((status & (RTP_BUS_DIS_MSK | RTP_DEV_SEL_MSK | RTP_DEV_RDY_MSK)) == 0)
			    {
				pmod->rtp.io [j].pio.aotio [k].is_responding = TRUE;
				pmod->rtp.io [j].pio.aotio [k].out_requested = FALSE;
				data_word = pmod->rtp.io [j].pio.aotio [k].out_value;
				data_word |= rtp_ao_channel [k];
				*(pmod->rtp.rtp_data) = data_word;
				if (mltrace [line] == 7)
				{
				    date ();
				    logMsg ("AO WRITE L %d A %d C %d Cmd 0x%x Data 0x%x\n",
						line,
						i,
						j,
						command_word,
						data_word,
						0);
				}
			    }
			    else
			    {
				pmod->rtp.io [j].pio.aotio [k].is_responding = FALSE;
				pmod->out_requested = TRUE;
			        ++badoutcnt [line];
				if (mltrace [line] == 7)
				{
				    date ();
				    logMsg ("AO WRITE FAILED L %d A %d C %d S %d\n",
						line,
						i,
						j,
						k,0,0
						);
				}
			    }
			}
		    }
		    break;

		case RTP_DOT_8:
		    for (k = 0; k < MAX_RTP_DOT_8; k++)
		    {
			if (pmod->rtp.io [j].pio.dotio [k].out_requested)
			{
			    break;
			}
		    }
		    if (k == MAX_RTP_DOT_8)
		    {
			continue;
		    }
		    out_requests [line] = (++out_requests [line]) & 0x7fff;
		    command_word = 0;
		    command_word |= RTP_DOT_INT_DIS;
		    command_word |= (j & RTP_DOT_GAT_MSK);
		    command_word |= RTP_DOT_LO_BANK;
		    *pmod->rtp.rtp_cmd = command_word;

		    for (l = 0; l < rtp_bus_delay; i++);

		    taskDelay (1);

		    for (t = 0; t < RTP_MAX_TST_CNT; t++)
		    {
			status = *pmod->rtp.rtp_cmd;
			if (status & RTP_BUS_DIS_MSK)
			{
			    break;
			}
			if (status & RTP_DEV_SEL_MSK)
			{
			    break;
			}
			if ((status & RTP_DEV_RDY_MSK) == 0)
			{
			    break;
			}
			taskDelay (1);
		    }
		    if ((status & (RTP_BUS_DIS_MSK | RTP_DEV_SEL_MSK | RTP_DEV_RDY_MSK)) == 0)
		    {
			for (k = 0; k < (MAX_RTP_DOT_8); k++)
			{
			    pmod->rtp.io [j].pio.dotio [k].is_responding = TRUE;
			    pmod->rtp.io [j].pio.dotio [k].out_requested = FALSE;
			}
			data_word = 0;
			for (k = 0; k < (MAX_RTP_DOT_8 / 2); k++)
			{
			    data_word |= (pmod->rtp.io [j].pio.dotio [k].out_value << (15 - k));
			}
			*(pmod->rtp.rtp_data) = data_word;
			if (mltrace [line] == 8)
			{
			    date ();
			    logMsg ("BO WRITE L %d A %d C %d Cmd 0x%x Data 0x%x\n",
					line,
					i,
					j,
					command_word,
					data_word,0
					);
			}
		    }
		    else
		    {
			for (k = 0; k < MAX_RTP_DOT_8; k++)
			{
			    pmod->rtp.io [j].pio.dotio [k].is_responding = FALSE;
			}
			pmod->out_requested = TRUE;
			++badoutcnt [line];
			if (mltrace [line] == 8)
			{
			    date ();
			    logMsg ("BO WRITE FAILED L %d A %d C %d S %d\n",
					line,
					i,
					j,
					k,0,0
					);
			}
		    }
		}
	    }
	}
}

/**********
*
*	AIDRIVER
*
*	This routines main purpose is to return the analog input value for the
*	specified module. For a module value a (-1) will be returned if
*	there is no valid data ready.
*
**********/

/* this routine not called from within this driver */

int rtp_aidriver (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal, short *pvalue,
		unsigned short conversion, unsigned char *resname,
		unsigned short dtype, int gain, int scanrate)
    {
	RTPIO		*pmod;
	unsigned short  blkix;
	unsigned short	val_1;
	unsigned short	val_2;
	unsigned short	val_3;

	if (rtp_test)
	{
	    return (-1);
	}

	if (link >= NUM_IOBC)
	{
	    return (-1);
	}
	if (adapter >= MAX_RTP_ADDR)
	{
	    switch (adapter)
	    {
	    case PSEUDO90:
		switch (signal)
		{
		case 0: /* Were the PLC values requested */
		    *pvalue = was_requested [link];
		    break;

		case 1: /* Is the PLC responding */
		    *pvalue = is_responding [link];
		    break;

		case 3: /* Offline transitions */
		    *pvalue = no_offlines [link] & 0x7fff;
		     break;

		case 4: /* Offline transitions */
		    *pvalue = no_local_activity [link] & 0x7fff;
		   break;

		case 5: /* Offline transitions */
		    *pvalue = no_partner_activity [link] & 0x7fff;
		    break;

		case 6: /* Offline transitions */
		    *pvalue = out_requests [link] & 0x7fff;
		    break;

	   	case 7: /* Offline transitions */
		    *pvalue = in_requests [link] & 0x7fff;
		    break;

		case 8: /* Offline transitions */
		    *pvalue = no_offlines [link] & 0x7fff;
		    break;

		case 9: /* Data Errors / Invalid status from ICM */
		    *pvalue = badoutcnt [link] & 0x7fff;
		    break;

		case 10: /* Offline transitions */
		    *pvalue = no_offlines [link] & 0x7fff;
		    break;

		case 11: /* Offline transitions */
		    *pvalue = ai_scan_count [link] & 0x7fff;
		    break;

		case 12: /* Scan Count */
		    *pvalue = ao_scan_count [link] & 0x7fff;
		    break;

		case 13: /* Scan Count */
		    *pvalue = bi_scan_count [link] & 0x7fff;
		    break;

		case 14: /* Scan Count */
		    *pvalue = bo_scan_count [link] & 0x7fff;
		    break;

		case 15: /* reset statistics */
		    *pvalue = cmd_reset_stat [link];
		    break;

		case 16: /* Force the PLC offline */
		    *pvalue = cmd_force_offline [link];
		    break;

		case 21: /* Scan Deviation limit in MS */
		    *pvalue = scan_devlimit [link];
		    break;

		case 23: /* Scan  deviation */
		    *pvalue = scan_deviation [link];
		    break;

		default:
		    return (-1);
		}
		return (0);

	    case PSEUDO91:
		if (card >= MAX_RTP_ADDR)
		{
		    return (-1);
		}
		pmod = plcs [link] [card];
		if (pmod == NULL)
		{
		    if (! allocate_plc (link, card))
		    {
			return (-1);
		    }
		    pmod = plcs [link] [card];
		}
		switch (signal)
		{
		case 0: /* Were the PLC values requested */
		    *pvalue = pmod->was_requested;
		    break;

		case 1: /* Is the PLC responding */
		    *pvalue = pmod->is_responding;
		    break;

	        case 2: /* Offline transitions */
		    *pvalue = pmod->rack_tmouts & 0x7fff;
		    break;

		default:
		    return (-1);
		}
		return (0);

	    case PSEUDO92:
		if ((card / 1000) >= MAX_RTP_ADDR)
		{
		    return (-1);
		}
		pmod = plcs [link] [card / 1000];
		if (pmod == NULL)
		{
		    if (! allocate_plc (link, (card / 1000)))
		    {
			return (-1);
		    }
		    pmod = plcs [link] [card / 1000];
		}
		switch (signal)
		{
		case 0:
		    *pvalue = pmod->rtp.io [card % 1000].tmouts & 0x7fff;
		    break;

		case 2:
		    switch (card % 1000)
		    {
		    case 0:
			*pvalue = pmod->rtp.io [0].rtp_type;
			break;

		    case 1:
			*pvalue = pmod->rtp.io [1].rtp_type;
			break;

		    case 2:
			*pvalue = pmod->rtp.io [2].rtp_type;
			break;

		    case 3:
			*pvalue = pmod->rtp.io [3].rtp_type;
			break;

		    case 4:
			*pvalue = pmod->rtp.io [4].rtp_type;
			break;

		    case 5:
			*pvalue = pmod->rtp.io [5].rtp_type;
			break;

		    case 6:
			*pvalue = pmod->rtp.io [6].rtp_type;
			break;

		    case 7:
			*pvalue = pmod->rtp.io [7].rtp_type;
			break;

		    case 8:
			*pvalue = pmod->rtp.io [8].rtp_type;
			break;

		    case 9:
			*pvalue = pmod->rtp.io [9].rtp_type;
			break;

		    case 10:
			*pvalue = pmod->rtp.io [10].rtp_type;
			break;

		    case 11:
			*pvalue = pmod->rtp.io [11].rtp_type;
			break;

		    case 12:
			*pvalue = pmod->rtp.io [12].rtp_type;
			break;

		    case 13:
			*pvalue = pmod->rtp.io [13].rtp_type;
			break;

		    case 14:
			*pvalue = pmod->rtp.io [14].rtp_type;
			break;

		    case 15:
			*pvalue = pmod->rtp.io [15].rtp_type;
			break;

		    default:
			return (-1);
		    }
		    return (0);

		default:
		    return (-1);
		}
	    default:
		return (-1);
	    }
	    return (0);
	}
	if (card >= MAX_GATE_ADDR)
	{
	    return (-1);
        }
	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (! allocate_plc (link, adapter))
	    {
		return (-1);
	    }
	    pmod = plcs [link] [adapter];
	}

	pmod->was_requested = TRUE;
	was_requested [link] = TRUE;

	if (resname != 0)
	{ 
	    register_resource (link, pmod, resname, pointname);
	}
	if (vme_base [link] == 0)
	{
	    return (-1);		/* The driver has not been started */	
	}
	if (pmod->rtp.rtp_cmd == NULL)
	{
	    pmod->rtp.rtp_cmd = (vme_base [link] + (adapter * 2));
	    pmod->rtp.rtp_data = ((vme_base [link] + (adapter * 2)) + 1);
	    if ((pmod->rtp.io [card].pio.ainio = (RTP_AIN_IO *)
		calloc (MAX_RTP_AIN, (sizeof (RTP_AIN_IO)))) == 0)
	    {
		logMsg ("1 Unable to allocate AI memory\n",0,0,0,0,0,0);
		return (-1);
	    }
	    pmod->rtp.io [card].rtp_type = RTP_AIN;
	}
	else if (pmod->rtp.rtp_cmd == (vme_base [link] + (adapter * 2)))
	{
	    if (pmod->rtp.io [card].pio.ainio == 0)
	    {
		if ((pmod->rtp.io [card].pio.ainio = (RTP_AIN_IO *)
		    calloc (MAX_RTP_AIN, (sizeof (RTP_AIN_IO)))) == 0)
		{
		    logMsg ("2 Unable to allocate AI memory\n",0,0,0,0,0,0);
		    return (-1);
		}
		pmod->rtp.io [card].rtp_type = RTP_AIN;
	    }
	    if (pmod->rtp.io [card].rtp_type != RTP_AIN)
	    {
		if (mltrace [link] == 4)
		{
		    logMsg ("AIDRIVER - Contains another card type...%s\n",
				pointname,0,0,0,0,0);
		}
		return (-1);
	    }
	}
	else
	{
	    if (mltrace [link] == 25)
	    {
		logMsg ("RTP Command Address Corrupted... %s\n", pointname,0,0,0,0,0);
	    }
	    return (-1);
	}
	if (pmod->rtp.io [card].pio.ainio == 0)
	{
	    logMsg ("3 Unable to allocate AI memory\n",0,0,0,0,0,0);
	    return (-1);
	}
	if ((signal / 100) >= MAX_GATE_ADDR)
	{
	    return (-1);
	}
	if ((signal % 100) >= MAX_GATE_SIG_4)
	{
	    return (-1);
	}

	pmod->rtp.io [signal / 100].rtp_type = RTP_GATE_4;

	blkix = (((signal / 100) * MAX_GATE_SIG_4) + (signal % 100));

	pmod->rtp.io [card].pio.ainio [blkix].devtype = dtype;
	switch (scanrate)
	{
	case P60SECOND:
	    pmod->rtp.io [card].pio.ainio [blkix].scanrate = p60second;
	    break;

	case P30SECOND:
	    pmod->rtp.io [card].pio.ainio [blkix].scanrate = p30second;
	    break;

	case P15SECOND:
	    pmod->rtp.io [card].pio.ainio [blkix].scanrate = p15second;
	    break;

	case P10SECOND:
	    pmod->rtp.io [card].pio.ainio [blkix].scanrate = p10second;
	    break;

	case P5SECOND:
	    pmod->rtp.io [card].pio.ainio [blkix].scanrate = p5second;
	    break;

	case P2SECOND:
	    pmod->rtp.io [card].pio.ainio [blkix].scanrate = p2second;
	    break;

	case PSECOND:
	    pmod->rtp.io [card].pio.ainio [blkix].scanrate = p1second;
	    break;

	case PHALFSECOND:
	    pmod->rtp.io [card].pio.ainio [blkix].scanrate = phalfsecond;
	    break;

	case PFIFTHSECOND:
	    pmod->rtp.io [card].pio.ainio [blkix].scanrate = pfifthsecond;
	    break;

	case PTENTHSECOND:
	    pmod->rtp.io [card].pio.ainio [blkix].scanrate = ptenthsecond;
	    break;

	default:
	    pmod->rtp.io [card].pio.ainio [blkix].scanrate = 0;
	    break;
	}

	pmod->rtp.io [card].pio.ainio [blkix].was_requested = TRUE;
	pmod->rtp.io [card].pio.ainio [blkix].gain = gain;

	if (!resresponse [link])
	{
	    if (mltrace [link] == 3)
	    {
		date ();
		logMsg ("AIDRIVER resresponse = 0, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-1);
	}
	if (pmod->holdflag)
	{
	    if (mltrace [link] == 3)
	    {
		date ();
		logMsg ("AIDRIVER holdflag = 1, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-2);
	}
	if (!pmod->is_responding)
	{
	    if (mltrace [link] == 3)
	    {
		date ();
		logMsg ("AIDRIVER RACK is_responding = 0, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-1);
	}
	if (!pmod->rtp.io [card].pio.ainio [blkix].is_responding)
        {
	    if (mltrace [link] == 3)
	    {
		date ();
		logMsg ("AIDRIVER CARD is_responding = 0, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-1);
	}
	val_3 = pmod->rtp.io [card].pio.ainio [blkix].value_3 =
		pmod->rtp.io [card].pio.ainio [blkix].value_2;
	val_2 = pmod->rtp.io [card].pio.ainio [blkix].value_2 =
		pmod->rtp.io [card].pio.ainio [blkix].value_1;
	val_1 = pmod->rtp.io [card].pio.ainio [blkix].value_1 =
		pmod->rtp.io [card].pio.ainio [blkix].value;

	if (val_1 <= val_2)
	{
	    if (val_2 <= val_3)
	    {
		*pvalue = val_2;
	    }
	    else if (val_1 <= val_3)
	    {
		*pvalue = val_3;
	    }
	    else
	    {
		*pvalue = val_1;
	    }
	}
	else
	{
	    if (val_1 <= val_3)
	    {
		*pvalue = val_1;
	    }
	    else if (val_2 <= val_3)
	    {
		*pvalue = val_3;
	    }
	    else
	    {
		*pvalue = val_2;
	    }
  	}
	if (mltrace [link] == 3)
	{
	    date ();
	    logMsg ("AIDRIVER Value read, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	}
	return (0);
}

/**********
*
*      AODRIVER
*      This routines main purpose is to set requested analog outputs to the
*      specified module.
*
**********/                                                                            	
/* this routine not called from within this driver */

int rtp_aodriver (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal, double dvalue,
		unsigned short value, unsigned short conversion,
		unsigned char *resname)
{
	RTPIO		*pmod;
	unsigned short  blkix;

	if (rtp_test)
	{
	    return (-1);
	}
	if (link >= NUM_IOBC)
	{
	    return (-1);
	}
	if (adapter >= MAX_RTP_ADDR)
	{
	    return (-1);
	}
	if (card >= MAX_GATE_ADDR)
	{
	    return (-1);
        }
	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (! allocate_plc (link, adapter))
	    {
		return (-1);
	    }
	    pmod = plcs [link] [adapter];
	}
	if (card == PSEUDO90)
	{
	    switch (signal)
	    {
	    case 21: /* Scan Deviation in MS */
		scan_devlimit [link] = value;
		break;

	    default:
		return (-1);
	    }
	    return (0);
	}
	if (resname != 0)
	{ 
	    register_resource (link, pmod, resname, pointname);
	}
	if (vme_base [link] == 0)
	{
	    return (-1);		/* The driver has not been started */	
	}
	if (pmod->rtp.rtp_cmd == NULL)
	{
	    pmod->rtp.rtp_cmd = (vme_base [link] + (adapter * 2));
	    pmod->rtp.rtp_data = ((vme_base [link] + (adapter * 2)) + 1);
	    if ((pmod->rtp.io [card].pio.aotio = (RTP_AOT_IO *)
		calloc (MAX_RTP_AOT_1, (sizeof (RTP_AOT_IO)))) == 0)
	    {
		logMsg ("1 Unable to allocate AO memory\n",0,0,0,0,0,0);
		return (-1);
	    }

	    pmod->rtp.io [card].rtp_type = RTP_AOT_1;
	}
	else if (pmod->rtp.rtp_cmd == (vme_base [link] + (adapter * 2)))
	{
	    if (pmod->rtp.io [card].pio.aotio == 0)
	    {
		if ((pmod->rtp.io [card].pio.aotio = (RTP_AOT_IO *)
		    calloc (MAX_RTP_AOT_1, (sizeof (RTP_AOT_IO)))) == 0)
		{
		    logMsg ("2 Unable to allocate AO memory\n",0,0,0,0,0,0);
		    return (-1);
		}

		pmod->rtp.io [card].rtp_type = RTP_AOT_1;
	    }
	    if (pmod->rtp.io [card].rtp_type != RTP_AOT_1)
	    {
		if (mltrace [link] == 4)
		{
		    logMsg ("AODRIVER - Contains another card type...%s\n",
				pointname,0,0,0,0,0);
		}
		return (-1);
	    }
	}
	else
	{
	    if (mltrace [link] == 25)
	    {
		logMsg ("RTP Command Address Corrupted... %s\n", pointname,0,0,0,0,0);
	    }
	    return (-1);
	}
	if (pmod->rtp.io [card].pio.aotio == 0)
	{
	    logMsg ("3 Unable to allocate AO memory\n",0,0,0,0,0,0);
	    return (-1);
	}
	if (signal >= MAX_RTP_AOT_1)
	{
	    return (-1);
	}
	blkix = signal;
	pmod->rtp.io [card].pio.aotio [blkix].was_requested = TRUE;

	if (!resresponse [link])
	{
	    if (mltrace [link] == 7)
	    {
		date ();
		logMsg ("AODRIVER resresponse = 0, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-1);
	}
	if (pmod->holdflag)
	{
	    if (mltrace [link] == 7)
	    {
		date ();
		logMsg ("AODRIVER holdflag = 1, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-2);
	}
	if ((!pmod->is_responding) || (pmod->cmd_force_offline))
	{
	    if (mltrace [link] == 7)
	    {
		date ();
		logMsg ("AODRIVER RACK is_responding = 0, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-1);
	}
	if (!pmod->rtp.io [card].pio.aotio [blkix].is_responding)
	{
	    if (mltrace [link] == 7)
	    {
		date ();
		logMsg ("AODRIVER CARD is_responding = 0, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-1);
	}
	pmod->rtp.io [card].pio.aotio [blkix].out_value = value;
        pmod->rtp.io [card].pio.aotio [blkix].out_requested = TRUE;
	pmod->out_requested = TRUE;
	if (mltrace [link] == 7)
	{
	    date ();
	    logMsg ("AODRIVER Set %s L %d A %d C %d S %d\n",
			pointname, link, adapter, card, signal,0);
	}
	return (0);
}

/**********
*
*      AOREAD
*      This routines main purpose is to return the value last read back
*      from the analog output hardware. This is used to initialize analog
*      outputs.
*
**********/

/* this routine not called from within this driver */

int rtp_aoread (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal,
		unsigned short *pvalue, unsigned short conversion,
		unsigned char *resname, int scanrate)
{
	RTPIO		*pmod;
	unsigned short  blkix;
	unsigned short	i;

	if (rtp_test)
	{
	    return (-1);
	}

	if (link >= NUM_IOBC)
	{
	    return (-1);
	}
	if (adapter >= MAX_RTP_ADDR)
	{
	    return (-1);
	}
	if ((card >= MAX_GATE_ADDR) && ((card != PSEUDO90) || (card != PSEUDO91) || (card != PSEUDO92)))
	{
	    return (-1);
        }
	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (! allocate_plc (link, adapter))
	    {
		return (-1);
	    }
	    pmod = plcs [link] [adapter];
	}
	if (resname != 0)
	{ 
	    register_resource (link, pmod, resname, pointname);
	}
	if (vme_base [link] == 0)
	{
	    return (-1);		/* The driver has not been started */	
	}
	if (pmod->rtp.rtp_cmd == NULL)
	{
	    pmod->rtp.rtp_cmd = (vme_base [link] + (adapter * 2));
	    pmod->rtp.rtp_data = ((vme_base [link] + (adapter * 2)) + 1);
	    if ((pmod->rtp.io [card].pio.aotio = (RTP_AOT_IO *)
		calloc (MAX_RTP_AOT_1, (sizeof (RTP_AOT_IO)))) == 0)
	    {
		logMsg ("4 Unable to allocate AO memory\n",0,0,0,0,0,0);
		return (-1);
	    }
	    pmod->rtp.io [card].rtp_type = RTP_AOT_1;
	}
	else if (pmod->rtp.rtp_cmd == (vme_base [link] + (adapter * 2)))
	{
	    if (pmod->rtp.io [card].pio.aotio == 0)
	    {
		if ((pmod->rtp.io [card].pio.aotio = (RTP_AOT_IO *)
		    calloc (MAX_RTP_AOT_1, (sizeof (RTP_AOT_IO)))) == 0)
		{
		    logMsg ("5 Unable to allocate AO memory\n",0,0,0,0,0,0);
		    return (-1);
		}

		pmod->rtp.io [card].rtp_type = RTP_AOT_1;
	    }
	    if (pmod->rtp.io [card].rtp_type != RTP_AOT_1)
	    {
		if (mltrace [link] == 4)
		{
		    logMsg ("AOREAD - Contains another card type...%s\n",
				pointname,0,0,0,0,0);
		}
		return (-1);
	    }
	}
	else
	{
	    if (mltrace [link] == 25)
	    {
		logMsg ("RTP Command Address Corrupted... %s\n", pointname,0,0,0,0,0);
	    }
	    return (-1);
	}
	if (pmod->rtp.io [card].pio.aotio == 0)
	{
	    logMsg ("6 Unable to allocate AO memory\n",0,0,0,0,0,0);
	    return (-1);
	}
	if (signal >= MAX_RTP_AOT_1)
	{
	    return (-1);
	}
	blkix = signal;

	switch (scanrate)
	{
	case P60SECOND:
	    pmod->rtp.io [card].pio.aotio [blkix].scanrate = p60second;
	    break;

	case P30SECOND:
	    pmod->rtp.io [card].pio.aotio [blkix].scanrate = p30second;
	    break;

	case P15SECOND:
	    pmod->rtp.io [card].pio.aotio [blkix].scanrate = p15second;
	    break;

	case P10SECOND:
	    pmod->rtp.io [card].pio.aotio [blkix].scanrate = p10second;
	    break;

	case P5SECOND:
	    pmod->rtp.io [card].pio.aotio [blkix].scanrate = p5second;
	    break;

	case P2SECOND:
	    pmod->rtp.io [card].pio.aotio [blkix].scanrate = p2second;
	    break;

	case PSECOND:
	    pmod->rtp.io [card].pio.aotio [blkix].scanrate = p1second;
	    break;

	case PHALFSECOND:
	    pmod->rtp.io [card].pio.aotio [blkix].scanrate = phalfsecond;
	    break;

	case PFIFTHSECOND:
	    pmod->rtp.io [card].pio.aotio [blkix].scanrate = pfifthsecond;
	    break;

	case PTENTHSECOND:
	    pmod->rtp.io [card].pio.aotio [blkix].scanrate = ptenthsecond;
	    break;

	default:
	    pmod->rtp.io [card].pio.aotio [blkix].scanrate = p60second;
	    break;
	}
	pmod->rtp.io [card].pio.aotio [blkix].was_requested = TRUE;

	if (!resresponse [link])
	{
	    if (mltrace [link] == 7)
	    {
		date ();
		logMsg ("AOREAD resresponse = 0, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-1);
	}
	if (pmod->holdflag)
	{
	    if (mltrace [link] == 7)
	    {
		date ();
		logMsg ("AOREAD holdflag = 1, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-2);
	}
	if (!pmod->is_responding)
	{
	    if (mltrace [link] == 7)
	    {
		date ();
		logMsg ("AOREAD RACK is_responding = 0, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-1);
	}
	if (!pmod->rtp.io [card].pio.aotio [blkix].is_responding)
        {
	    if (mltrace [link] == 7)
	    {
		date ();
		logMsg ("AOREAD CARD is_responding = 0, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-1);
	}
        *pvalue = pmod->rtp.io [card].pio.aotio [blkix].value;
	if (mltrace [link] == 7)
	{
	    date ();
	    logMsg ("AOREAD Read %s L %d A %d C %d S %d\n",
			pointname, link, adapter, card, signal,0);
	}
	return (0);
}

/**********
*
*	BIDRIVER - This routine is called from the device interface
*	module to read binary inputs.
*
**********/                                                                                     
/* this routine not called from within this driver */

int rtp_bidriver (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal, unsigned long mask,
		unsigned long *pvalue, unsigned char *resname, int scanrate)

    {
	RTPIO 		*pmod;
	unsigned short	i, j;
	unsigned long	tvalue, tmask;

	if (rtp_test)
	{
	    return (-1);
	}

	if (link >= NUM_IOBC)
	{
	    return (-1);
	}

	for (j = 0, tmask = mask; ((tmask & 1) == 0); j++, tmask >>= 1);

	if (adapter >= MAX_RTP_ADDR)
	{
	    switch (adapter)
	    {
	    case PSEUDO90:
		switch (signal)
		{
		case 0: /* Were the PLC values requested */
		    tvalue = was_requested [link];
		    break;

		case 1: /* Is the PLC responding */
		    tvalue = is_responding [link];
		    break;

		case 2: /* Is the reporting reporting correctly */
		    tvalue = line_activity [link];
		    break;

		case 3: /* Is the partner reporting at all */
		    tvalue = partner_ok [link];
		    break;

		case 10: /* Scan Fault / Time Exceeded */
		    tvalue = scan_fault [link];
		    break;

		default:
		    return (-1);
		}
		return (0);

	    case PSEUDO91:
		if (card >= MAX_RTP_ADDR)
		{
		    return (-1);
		}
		pmod = plcs [link] [card];
		if (pmod == NULL)
		{
		    if (! allocate_plc (link, card))
		    {
			return (-1);
		    }
		    pmod = plcs [link] [card];
		}
		switch (signal)
		{
		case 0: /* Were the PLC values requested */
		    tvalue = pmod->was_requested;
		    break;

		case 1: /* Is the PLC responding */
		    tvalue = pmod->is_responding;
		    break;

		case 2: /* Is the PLC responding */
		    tvalue = pmod->is_responding;
		    break;

		case 15: /* reset statistics */
		    tvalue = pmod->cmd_reset_stat;
		    break;

		case 16: /* Force the PLC offline */
		    tvalue = pmod->cmd_force_offline;
		    break;
		}
		break;

	    case PSEUDO92:
		if ((card / 1000) >= MAX_RTP_ADDR)
		{
		    return (-1);
		}
		pmod = plcs [link] [card / 1000];
		if (pmod == NULL)
		{
		    if (! allocate_plc (link, (card / 1000)))
		    {
			return (-1);
		    }
		    pmod = plcs [link] [card / 1000];
		}
		switch (signal)
		{
		case 1:
		    switch (card % 1000)
	 	    {
		    case 0:
			tvalue = pmod->rtp.io [0].status;
			break;

		    case 1:
			tvalue = pmod->rtp.io [1].status;
			break;

		    case 2:
			tvalue = pmod->rtp.io [2].status;
			break;

		    case 3:
			tvalue = pmod->rtp.io [3].status;
			break;

		    case 4:
			tvalue = pmod->rtp.io [4].status;
			break;

		    case 5:
			tvalue = pmod->rtp.io [5].status;
			break;

	 	    case 6:
			tvalue = pmod->rtp.io [6].status;
			break;

		    case 7:
			tvalue = pmod->rtp.io [7].status;
			break;

		    case 8:
			tvalue = pmod->rtp.io [8].status;
			break;

		    case 9:
			tvalue = pmod->rtp.io [9].status;
			break;

		    case 10:
		        tvalue = pmod->rtp.io [10].status;
		        break;

		    case 11:
			tvalue = pmod->rtp.io [11].status;
			break;

		    case 12:
			tvalue = pmod->rtp.io [12].status;
			break;

		    case 13:
			tvalue = pmod->rtp.io [13].status;
			break;

		    case 14:
			tvalue = pmod->rtp.io [14].status;
			break;

		    case 15:
			tvalue = pmod->rtp.io [15].status;
			break;

		    default:
			return (-1);
		    }
		    break;

		default:
		    return (-1);
		}
		break;

	    default:
		return (-1);
	    }
	    if (pmod->holdflag)
	    {
		return (-2);
	    }
	    *pvalue = ((tvalue << j) & mask);
	    return (0);
	}
	if (card >= MAX_GATE_ADDR)
	{
	    return (-1);
        }
	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (! allocate_plc (link, adapter))
	    {
		return (-1);
	    }
	    pmod = plcs [link] [adapter];
	}

	pmod->was_requested = TRUE;
	was_requested [link] = TRUE;

	if (resname != 0)
	{ 
	    register_resource (link, pmod, resname, pointname);
	}
	if (vme_base [link] == 0)
	{
	    return (-1);		/* The driver has not been started */	
	}
	if (pmod->rtp.rtp_cmd == NULL)
	{
	    pmod->rtp.rtp_cmd = (vme_base [link] + (adapter * 2));
	    pmod->rtp.rtp_data = ((vme_base [link] + (adapter * 2)) + 1);
	    if ((pmod->rtp.io [card].pio.dinio = (RTP_DIN_IO *)
		calloc (MAX_RTP_DIN_8, (sizeof (RTP_DIN_IO)))) == 0)
	    {
		logMsg ("1 Unable to allocate BI memory\n",0,0,0,0,0,0);
		return (-1);
	    }
	    pmod->rtp.io [card].rtp_type = RTP_DIN_8;
	}
	else if (pmod->rtp.rtp_cmd == (vme_base [link] + (adapter * 2)))
	{
	    if (pmod->rtp.io [card].pio.dinio == 0)
	    {
		if ((pmod->rtp.io [card].pio.dinio = (RTP_DIN_IO *)
		    calloc (MAX_RTP_DIN_8, (sizeof (RTP_DIN_IO)))) == 0)
		{
		    logMsg ("2 Unable to allocate BI memory\n",0,0,0,0,0,0);
		    return (-1);
		}
		pmod->rtp.io [card].rtp_type = RTP_DIN_8;
	    }
	    if (pmod->rtp.io [card].rtp_type != RTP_DIN_8)
	    {
		if (mltrace [link] == 4)
		{
		    logMsg ("BIDRIVER - Contains another card type...%s\n",
				pointname,0,0,0,0,0);
		}
		return (-1);
	    }
	}
	else
	{
	    if (mltrace [link] == 25)
	    {
		logMsg ("RTP Command Address Corrupted... %s\n", pointname,0,0,0,0,0);
	    }
	    return (-1);
	}
	if (pmod->rtp.io [card].pio.dinio == 0)
	{
	    logMsg ("3 Unable to allocate BI memory\n",0,0,0,0,0,0);
	    return (-1);
	}
	if (signal >= MAX_RTP_DIN_8)
	{
	    return (-1);
	}
	tvalue = 0;
	for (i = 0; ((tmask & 1) != 0); i++)
	{
	    switch (scanrate)
	    {
	    case P60SECOND:
		pmod->rtp.io [card].pio.dinio [signal + i].scanrate = p60second;
		break;

	    case P30SECOND:
		pmod->rtp.io [card].pio.dinio [signal + i].scanrate = p30second;
		break;

	    case P15SECOND:
		pmod->rtp.io [card].pio.dinio [signal + i].scanrate = p15second;
		break;

	    case P10SECOND:
		pmod->rtp.io [card].pio.dinio [signal + i].scanrate = p10second;
		break;

	    case P5SECOND:
		pmod->rtp.io [card].pio.dinio [signal + i].scanrate = p5second;
		break;

	    case P2SECOND:
		pmod->rtp.io [card].pio.dinio [signal + i].scanrate = p2second;
		break;

	    case PSECOND:
		pmod->rtp.io [card].pio.dinio [signal + i].scanrate = p1second;
		break;

	    case PHALFSECOND:
		pmod->rtp.io [card].pio.dinio [signal + i].scanrate = phalfsecond;
		break;

	    case PFIFTHSECOND:
		pmod->rtp.io [card].pio.dinio [signal + i].scanrate = pfifthsecond;
		break;

	    case PTENTHSECOND:
		pmod->rtp.io [card].pio.dinio [signal + i].scanrate = ptenthsecond;
		break;

	    default:
		pmod->rtp.io [card].pio.dinio [signal + i].scanrate = p60second;
		break;
	    }
	    pmod->rtp.io [card].pio.dinio [signal + i].was_requested = TRUE;
	    if (!pmod->rtp.io [card].pio.dinio [signal + i].is_responding)
	    {
		if (! pmod->holdflag)
		{
		    if (mltrace [link] == 7)
		    {
			date ();
			logMsg ("BIDRIVER CARD is_responding = 0, L %d A %d C %d S %d\n",
				link, adapter, card, signal,0,0);
		    }
		    return (-1);
		}
	    }
	    tvalue |= ((pmod->rtp.io [card].pio.dinio [signal + i].value) << i);
	    tmask >>= 1;
	}
	if (!resresponse [link])
	{
	    if (mltrace [link] == 3)
	    {
		date ();
		logMsg ("BIDRIVER resresponse = 0, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-1);
	}
	if (pmod->holdflag)
	{
	    if (mltrace [link] == 3)
	    {
		date ();
		logMsg ("BIDRIVER holdflag = 1, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-2);
	}
	if (!pmod->is_responding)
	{
	    if (mltrace [link] == 3)
	    {
		date ();
		logMsg ("BIDRIVER RACK is_responding = 0, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-1);
	}
	*pvalue = (tvalue << j);
	return (0);
}

/**********
*
*	BODRIVER - This routine is called from the device interface
*	module to set / clear binary outputs.
*
**********/

/* this routine not called from within this driver */

int rtp_bodriver (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal,
		unsigned long value, unsigned long mask,
		unsigned char *resname)
{
	RTPIO 		*pmod;
	unsigned short       i, j;
	unsigned long	tvalue, tmask;

	if (rtp_test)
	{
	    return (-1);
	}

	if (link >= NUM_IOBC)
	{
	    return (-1);
	}
	if (adapter >= MAX_RTP_ADDR)
	{
	    switch (adapter)
	    {
	    case PSEUDO90:
		pmod = plcs [link] [card];
		if (pmod == NULL)
		{
		    if (! allocate_plc (link, card))
		    {
			return (-1);
		    }
		    pmod = plcs [link] [card];
		}
		switch (signal)
		{
		case 0: /* Reset the accumulated statistics */
		    if (value != 0)
		    {
			initiobcstats (link);
		    }
		    break;

		case 1: /* Reset the accumulated statistics */
		    if (value != 0)
		    {
			initrackstats (pmod);
		    }
		    break;

		case 2: /* Reset the accumulated statistics */
		    if (value != 0)
		    {
			initiobcstats (link);
			for (i = 0; i < MAX_RTP_ADDR; i++)
			{
			    pmod = plcs [link] [i];
			    if (pmod == NULL)
			    {
				continue;
			    }
			    else
			    {
				initrackstats (pmod);
			    }
			}
		    }
		    break;

		case 3: /* Force the PLC offline */
		    pmod->cmd_force_offline = (value != 0);
                    if (value != 0)
		    {
                	pmod->is_responding = FALSE;/* only required when asserting cmd_force_offline */
		    }
		    break;

		default:
		    return (-1);
		}
		break;

	    default:
		return (-1);
	    }
	    return (0);
	}
	if ((card >= MAX_GATE_ADDR) &&
		((card != PSEUDO90) || (card != PSEUDO91) || (card != PSEUDO92)))
	{
	    return (-1);
        }
	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (! allocate_plc (link, adapter))
	    {
		return (-1);
	    }
	    pmod = plcs [link] [adapter];
	}

	for (j = 0, tmask = mask, tvalue = value; ((tmask & 1) == 0); j++, tmask >>= 1, tvalue >>= 1);

	pmod->was_requested = TRUE;
	was_requested [link] = TRUE;

	if (resname != 0)
	{ 
	    register_resource (link, pmod, resname, pointname);
	}
	if (vme_base [link] == 0)
	{
	    return (-1);		/* The driver has not been started */	
	}
	if (pmod->rtp.rtp_cmd == NULL)
	{
	    pmod->rtp.rtp_cmd = (vme_base [link] + (adapter * 2));
	    pmod->rtp.rtp_data = ((vme_base [link] + (adapter * 2)) + 1);
	    if ((pmod->rtp.io [card].pio.dotio = (RTP_DOT_IO *)
		calloc (MAX_RTP_DOT_8, (sizeof (RTP_DOT_IO)))) == 0)
	    {
		logMsg ("1 Unable to allocate BO memory\n",0,0,0,0,0,0);
		return (-1);
	    }
	    pmod->rtp.io [card].rtp_type = RTP_DOT_8;
	}
	else if (pmod->rtp.rtp_cmd == (vme_base [link] + (adapter * 2)))
	{
	    if (pmod->rtp.io [card].pio.dotio == 0)
	    {
		if ((pmod->rtp.io [card].pio.dotio = (RTP_DOT_IO *)
		    calloc (MAX_RTP_DOT_8, (sizeof (RTP_DOT_IO)))) == 0)
		{
		    logMsg ("2 Unable to allocate BO memory\n",0,0,0,0,0,0);
		    return (-1);
		}
		pmod->rtp.io [card].rtp_type = RTP_DOT_8;
	    }
	    if (pmod->rtp.io [card].rtp_type != RTP_DOT_8)
	    {
		if (mltrace [link] == 4)
		{
		    logMsg ("BODRIVER - Already contains another card type L %d A %d C %d S %d\n",
				link, adapter, card, signal,0,0);
		}
		return (-1);
	    }
	}
	else
	{
	    if (mltrace [link] == 25)
	    {
		logMsg ("RTP Command Address Corrupted... %s\n", pointname,0,0,0,0,0);
	    }
	    return (-1);
	}
	if (pmod->rtp.io [card].pio.dotio == 0)
	{
	    logMsg ("2 Unable to allocate BO memory\n",0,0,0,0,0,0);
	    return (-1);
	}
	if (signal >= MAX_RTP_DOT_8)
	{
	    return (-1);
	}
	borecorder [link] [borecidx [link]].mask = tmask;
	borecorder [link] [borecidx [link]].value = tvalue;
	borecorder [link] [borecidx [link]].card = card;
	borecorder [link] [borecidx [link]].rack = adapter;
	borecorder [link] [borecidx [link]].previous_mask = 0;
	tsStampToText (&time_stamp, TS_TEXT_MONDDYYYY, borecorder [link] [borecidx [link]].time);
	for (i = 0; i < MAX_RTP_DOT_8; i++)
	{
	    borecorder [link] [borecidx [link]].previous_mask |=
		(pmod->rtp.io [card].pio.dotio [i].out_value << i);
	}
	for (i = 0; ((tmask & 1) != 0); i++)
	{
	    pmod->rtp.io [card].pio.dotio [signal + i].was_requested = TRUE;
	    if (!pmod->rtp.io [card].pio.dotio [signal + i].is_responding)
	    {
		if (! pmod->holdflag)
		{
		    if (mltrace [link] == 8)
		    {
			date ();
			logMsg ("BODRIVER Card is_responding = 0, L %d A %d C %d S %d\n",
				link, adapter, card, signal,0,0);
		    }
		    return (-1);
		}
	    }
	    pmod->rtp.io [card].pio.dotio [signal + i].out_value = (tvalue & 1);
	    pmod->rtp.io [card].pio.dotio [signal + i].out_requested = TRUE;
	    tvalue >>= 1;
	    tmask >>= 1;
	}
	borecorder [link] [borecidx [link]].current_mask = 0;
	for (i = 0; i < MAX_RTP_DOT_8; i++)
	{
	    borecorder [link] [borecidx [link]].current_mask |=
		(pmod->rtp.io [card].pio.dotio [i].out_value << i);
	}
	strcpy (borecorder [link] [borecidx [link]].pointname, pointname);
	if (++borecidx [link] >= 200)
	{
	    borecidx [link] = 0;
	}
	if (!resresponse [link])
	{
	    if (mltrace [link] == 8)
	    {
		date ();
		logMsg ("BODRIVER resresponse = 0, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-1);
	}
	if (pmod->holdflag)
	{
	    if (mltrace [link] == 8)
	    {
		date ();
		logMsg ("BODRIVER holdflag = 1, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-2);
	}
	if (!pmod->is_responding)
	{
	    if (mltrace [link] == 8)
	    {
		date ();
		logMsg ("BODRIVER Rack is_responding = 0, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-1);
	}
	if (mltrace [link] == 8)
	{
	    date ();
	    logMsg ("BODRIVER Set %s L %d A %d C %d S %d\n",
			pointname, link, adapter, card, signal,0);
	}
	pmod->out_requested = TRUE;
	return (0);
}

/**********
*
*      BOREAD
*
*      This routines main purpose is to return the value last read back
*      from the digital output hardware. This is used to initialize digital
*      outputs. This will return a (-1) if the module is the wrong type or
*      the data is not yet valid
*
**********/

/* this routine not called from within this driver */

int rtp_boread (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal,
		unsigned long *pvalue, unsigned long mask,
		unsigned char *resname, int scanrate)
{
	RTPIO		*pmod;
	unsigned short       i, j;
	unsigned long	tvalue, tmask;

	if (rtp_test)
	{
	    return (-1);
	}

	if (link >= NUM_IOBC)
	{
	    return (-1);
	}
	if (adapter >= MAX_RTP_ADDR)
	{
	    switch (adapter)
	    {
	    case PSEUDO90:
		switch (signal)
		{
		case 0: /* Reset the accumulated statistics */
		    break;

		case 1: /* Reset the accumulated statistics */
		    break;

		case 2: /* Reset the accumulated statistics */
		    break;

		case 3: /* Force the PLC offline */
		    break;

		default:
		    return (-1);
		}
		break;

	    default:
		return (-1);
	    }
	    return (0);
	}
	if (card >= MAX_GATE_ADDR)
	{
	    return (-1);
        }
	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (! allocate_plc (link, adapter))
	    {
		return (-1);
	    }
	    pmod = plcs [link] [adapter];
	}

	for (j = 0, tmask = mask; ((tmask & 1) == 0); j++, tmask >>= 1);

	if (card == PSEUDO90)
	{
	    switch (signal)
	    {

	    case 15: /* reset statistics */
		*pvalue = pmod->cmd_reset_stat;
		break;

	    case 16: /* Force the PLC offline */
		*pvalue = pmod->cmd_force_offline;
		break;

	    default:
		return (-1);
	    }
	    return (0);
	}
	pmod->was_requested = TRUE;
	was_requested [link] = TRUE;

	if (resname != 0)
	{
             register_resource (link, pmod, resname, pointname);
	}
	if (vme_base [link] == 0)
	{
	    return (-1);		/* The driver has not been started */	
	}
	if (pmod->rtp.rtp_cmd == NULL)
	{
	    pmod->rtp.rtp_cmd = (vme_base [link] + (adapter * 2));
	    pmod->rtp.rtp_data = ((vme_base [link] + (adapter * 2)) + 1);
	    if ((pmod->rtp.io [card].pio.dotio = (RTP_DOT_IO *)
		calloc (MAX_RTP_DOT_8, (sizeof (RTP_DOT_IO)))) == 0)
	    {
		logMsg ("3 Unable to allocate BO memory\n",0,0,0,0,0,0);
		return (-1);
	    }
	    pmod->rtp.io [card].rtp_type = RTP_DOT_8;
	}
	else if (pmod->rtp.rtp_cmd == (vme_base [link] + (adapter * 2)))
	{
	    if (pmod->rtp.io [card].rtp_type == SLOT_EMPTY)
	    {
		if ((pmod->rtp.io [card].pio.dotio = (RTP_DOT_IO *)
		    calloc (MAX_RTP_DOT_8, (sizeof (RTP_DOT_IO)))) == 0)
		{
		    logMsg ("4 Unable to allocate BO memory\n",0,0,0,0,0,0);
		    return (-1);
		}
		pmod->rtp.io [card].rtp_type = RTP_DOT_8;
	    }
	    else if (pmod->rtp.io [card].rtp_type != RTP_DOT_8)
	    {
		if (mltrace [link] == 4)
		{
		    logMsg ("BOREAD - Contains another card type...%s\n",
				pointname);
		}
		return (-1);
	    }
	}
	else
	{
	    if (mltrace [link] == 25)
	    {
		logMsg ("RTP Command Address Corrupted... %s\n", pointname,0,0,0,0,0);
	    }
	    return (-1);
	}
	if (pmod->rtp.io [card].pio.dotio == 0)
	{
	    logMsg ("5 Unable to allocate BO memory\n",0,0,0,0,0,0);
	    return (-1);
	}
	if (signal >= MAX_RTP_DOT_8)
	{
	    return (-1);
	}
	tvalue = 0;
	for (i = 0; ((tmask & 1) != 0); i++)
	{
	    switch (scanrate)
	    {
	    case P60SECOND:
		pmod->rtp.io [card].pio.dotio [signal + i].scanrate = p60second;
		break;

	    case P30SECOND:
		pmod->rtp.io [card].pio.dotio [signal + i].scanrate = p30second;
		break;

	    case P15SECOND:
		pmod->rtp.io [card].pio.dotio [signal + i].scanrate = p15second;
		break;

	    case P10SECOND:
		pmod->rtp.io [card].pio.dotio [signal + i].scanrate = p10second;
		break;

	    case P5SECOND:
		pmod->rtp.io [card].pio.dotio [signal + i].scanrate = p5second;
		break;

	    case P2SECOND:
		pmod->rtp.io [card].pio.dotio [signal + i].scanrate = p2second;
		break;

	    case PSECOND:
		pmod->rtp.io [card].pio.dotio [signal + i].scanrate = p1second;
		break;

	    case PHALFSECOND:
		pmod->rtp.io [card].pio.dotio [signal + i].scanrate = phalfsecond;
		break;

	    case PFIFTHSECOND:
		pmod->rtp.io [card].pio.dotio [signal + i].scanrate = pfifthsecond;
		break;

	    case PTENTHSECOND:
		pmod->rtp.io [card].pio.dotio [signal + i].scanrate = ptenthsecond;
		break;

	    default:
		pmod->rtp.io [card].pio.dotio [signal + i].scanrate = p60second;
		break;
	    }
	    pmod->rtp.io [card].pio.dotio [signal + i].was_requested = TRUE;
	    if (!pmod->rtp.io [card].pio.dotio [signal + i].is_responding)
	    {
		if (! pmod->holdflag)
		{
		    if (mltrace [link] == 8)
		    {
			date ();
			logMsg ("BOREAD Card is_responding = 0, L %d A %d C %d S %d\n",
				link, adapter, card, signal,0,0);
		    }
		    return (-1);
		}
	    }
	    tvalue |= ((pmod->rtp.io [card].pio.dotio [signal + i].value) << i);
	    tmask >>= 1;
	}
	if (!resresponse [link])
	{
	    if (mltrace [link] == 8)
	    {
		date ();
		logMsg ("BOREAD resresponse = 0, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-1);
	}
	if (pmod->holdflag)
	{
	    if (mltrace [link] == 8)
	    {
		date ();
		logMsg ("BOREAD holdflag = 1, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-2);
	}
	if (!pmod->is_responding)
	{
	    if (mltrace [link] == 8)
	    {
		date ();
		logMsg ("BOREAD Rack is_responding = 0, L %d A %d C %d S %d\n",
			link, adapter, card, signal,0,0);
	    }
	    return (-1);
	}
	if (mltrace [link] == 8)
	{
	    date ();
	    logMsg ("BODRIVER Read %s L %d A %d C %d S %d\n",
			pointname, link, adapter, card, signal,0);
	}
	*pvalue = (tvalue << j);
	return (0);
}

/**********
*
* register_resource
*
*	This routine is used to initialize the resource pointers
*	If this is the first time the resource for this line is being
*	referenced the pointer to the resource table entry will be set
*	and the resource name will be saved. If the resource name has
*	already been set, this routine will verify that all other points
*	for this driver are using the same resource name.
*
**********/

/* this routine called from AIDRIVER, AODRIVER, AOREAD, BIDRIVER, BODRIVER,
*			    BOREAD.
*/                               	

static void  register_resource (unsigned short link, RTPIO *pmod,
                                unsigned char *resname, char *pointname)
{
	if (resnames [link] [0] == 0) /* This is the first reference */
	{
	    if (resGetByName (resname, &prestable [link]) != RES_OK)
	    {
		if (rtpresmsgdisable == FALSE)
		{
		    logMsg ("Resource not found in table.(Point Name = %s, Resource Name = %s)\n",
                                pointname,
				resname ,0,0,0,0);
		}
	    }
	    else
	    {
		strcpy (resnames [link], resname);
	    }
	}
	else
	{
	    if (strcmp (resnames [link], resname) != 0)
	    {
		if (rtpresmsgdisable == FALSE)
		{
		    logMsg ("Resource name[%s] conflict.(Point Name = %s, Resource Name = %s)\n",
				resnames[link],
                                pointname,
				resname ,0,0,0);
		}
	    }
	}
}

/**********
*
*      link_status                                                           	
*                                                                            	
*      The purpose of this routine is to print out the state of each of the  	
*      analog devices lines and the statistics gathered about them.         	
*                                                                           	
**********/

/* not called from within this driver */

static	void link_status (unsigned short line)
{
	if (line > NUM_IOBC - 1)
	{
	    return;
	}
}

/**********
*                                                                              
*	allocate_plc                                                            
*                                                                              
*	The purpose of this routine is to allocate the memory for a single      
*	plc data structure. If not successful, a console error message is  	
*	printed and an error flag is returned to the caller.                    
*	If successful, the newly allocated data structures are initialized.     
*                                                                             	
**********/

static int allocate_plc (unsigned short line, unsigned short mod )
{
	plcs [line] [mod] = (RTPIO *) calloc (1, (sizeof (RTPIO)));
	if (!plcs [line] [mod])
	{
	    logMsg ("Unable to allocate memory",0,0,0,0,0,0);
	    printf (" for RTPIO data structure on line %d\n", line);
	    return (FALSE);
	}
	else
	{
	    logMsg ("Allocated memory for RTPIO data structure on line %d\n", line,0,0,0,0,0);
	    init_plc_tab (line, mod);	     /*	initialize the data structures. */
	}
	return (TRUE);
}


/**********
*
*      init_plc_tab
*
*      The purpose of this routine is to initialize the line tables for the
*      Rtp I/O
*
**********/

/* only called from allocate_plc  */

static	void init_plc_tab (unsigned short line, unsigned short mod)
{
	RTPIO  *pmod;

	pmod = plcs [line] [mod];
	if (!pmod)
	{
		return;
	}
	pmod->is_responding = FALSE;
	pmod->holdflag = FALSE;
	pmod->was_requested = FALSE;
	pmod->cmd_force_offline = FALSE;
	pmod->out_requested = FALSE;

	initplc (pmod);
	initiobcstats (line);
	initrackstats (pmod);
}
/**********
*
* Cold Init a PLC
*
**********/

/* only called from init_line_tabs  */

static void initplc (RTPIO  *pmod)
{
	unsigned short	i;

	for (i = 0; i < MAX_GATE_ADDR; i++)
	{
	    pmod->rtp.io [i].rtp_type = SLOT_EMPTY;
	    pmod->rtp.io [i].status = RTP_RESPONDING;
	}
}

/**********
*
*	prepresponses
*
*	The purpose of this routine is to mark the PLCs as on hold and the
*	data as not valid until read again from the PLC
*
**********/

/* only called from MAIN */

static	void prepresponses (unsigned short line)

{
	unsigned short	mod;
	RTPIO		*pmod;

	for (mod = 0; (mod < MAX_RTP_ADDR); mod++)
	{
	    pmod = plcs [line] [mod];
	    if (pmod)			/* verify pointer */
	    {
	    	pmod->holdflag = TRUE;
/*		mark_not_valid (pmod, line);	*/
            }
	}
}

/**********
*
*	any_responding                                               	
*                                                                   	
*	The purpose of this routine is to see if any units are       	
*	responding on this line                                      	
*                                                                   	
**********/

/* only called from MAIN */

static	int any_responding (unsigned short line)
{
	unsigned short  mod;
	RTPIO		*pmod;

	for (mod = 0; (mod < MAX_RTP_ADDR); mod++)
	{
	    pmod = plcs [line] [mod];
	    if (pmod)			/* verify pointer */
	    {
                /* check cmd_force_offline flag because it is
			set in an asynchronous task */

                if ((pmod->is_responding) && (!pmod->cmd_force_offline))
		{
		    return (TRUE);
		}
	    }
	}
	return (FALSE);
}

static	void    master_updateredimage (unsigned short line)
{
	unsigned short	i, j, k;
	unsigned short	data_word;
	unsigned short	outix = 0;
	RTPIO  	*pmod;

	drvRedTake (redsemids [line]); 

	for (i = 0; i < MAX_RTP_ADDR; i++)
	{
	    pmod = plcs [line] [i];
	    if (pmod == NULL)
	    {
		continue;
	    }

	    /* Fill in the header */

	    redund [line].line = line;
	    redund [line].mastercnt = mycount [line];
	    redund [line].outcount = 0;

	    for (j = 0; j < MAX_GATE_ADDR; j++)
	    {
		switch (pmod->rtp.io [j].rtp_type)
		{
		case SLOT_EMPTY:
		    break;

		case RTP_AIN:
		    break;

		case RTP_GATE_4:
		    break;

		case RTP_DIN_8:
		    break;

		case RTP_AOT_1:
		    break;

		case RTP_DOT_8:
		    redund [line].outs [outix].rack = i;
		    redund [line].outs [outix].card = j;
		    redund [line].outs [outix].type = pmod->rtp.io [j].rtp_type;
		    data_word = 0;
		    for (k = 0; k < (MAX_RTP_DOT_8 / 2); k++)
		    {
			data_word |= (pmod->rtp.io [j].pio.dotio [k].out_value << (15 - k));
		    }
		    redund [line].outs [outix].data = data_word;
		    ++outix;
		    if (outix >= MAXUPDS )
		    {
			logMsg ("Too many output modules defined for redundancy \n",0,0,0,0,0,0);
			return;
		    }
		    break;
		}
	    }
	}
	redund [line].outcount = outix;
	drvRedGive (redsemids [line]); 
}

static	void 	slave_updateredimage (unsigned short line,
			RTPSREDUPD *redbuf, int size)
{
	unsigned short	i, j;
	unsigned short	outix;
	unsigned short	data_word;
	int		lock;
	RTPIO		*pmod;

	if (line != redbuf->line)
	{
	    logMsg ("Slave update line mismatch L %d BL %d\n", line, redbuf->line,0,0,0,0);
	    return;
	}
	outix = redbuf->outcount;
	for (i = 0; i < outix; i++)
	{
	    pmod = plcs [line] [redbuf->outs [i].rack];
	    if (redbuf->outs [i].type == pmod->rtp.io [redbuf->outs [i].card].rtp_type)
	    {
		data_word = redbuf->outs [i].data;
		lock = taskLock ();

		for (j = (MAX_RTP_DOT_8 / 2); j  != 0 ; j--)
		{
		    pmod->rtp.io [redbuf->outs [i].card].pio.dotio [j - 1].out_value =
				(data_word & 0x01);
		    data_word >>= 1;
		}
		taskUnlock ();
	    }
	}
}

/**********
*
*	rtpdata
*
*	The purpose of this routine is to print all of the data points
*	that have been requested by the initUpdate task.
*
**********/

void rtpdata (unsigned short line, unsigned short mod, unsigned short slot)
{
	RTPIO		*pmod;
	unsigned short	k;

	pmod = plcs [line] [mod];
	if (!pmod)
	{
	    logMsg ("Line Not Defined \n",0,0,0,0,0,0);
	    return;
	}
	switch (pmod->rtp.io [slot].rtp_type)
	{
	case SLOT_EMPTY:
	    printf ("Slot %d EMPTY\n", slot);
	    break;

	case RTP_AIN:
	    printf ("Slot %d AIN\n", slot);

	    for (k = 0; k < MAX_RTP_AIN; k++)
	    {
		if (pmod->rtp.io [slot].pio.ainio [k].was_requested)
		{
		    printf ("AI %d: ", k);
		    printf ("Is Responding %d: ", pmod->rtp.io [slot].pio.ainio [k].is_responding);
		    printf ("Scan Rate %ld: ", pmod->rtp.io [slot].pio.ainio [k].scanrate);
		    printf ("Next Scan %ld: ", pmod->rtp.io [slot].pio.ainio [k].next_scan);
		    printf ("Value = %d\n", pmod->rtp.io [slot].pio.ainio [k].value);
		}
	    }
	    break;

	case RTP_GATE_4:
	    printf ("Slot %d AIN GATE\n", slot);
	    break;

	case RTP_AOT_1:
	    printf ("Slot %d AOT_1\n", slot);
	    for (k = 0; k < MAX_RTP_AOT_1; k++)
	    {
		if (pmod->rtp.io [slot].pio.aotio [k].was_requested)
		{
		    printf ("AO %d: ", k);
		    printf ("Is Responding %d: ", pmod->rtp.io [slot].pio.aotio [k].is_responding);
		    printf ("Scan Rate %ld: ", pmod->rtp.io [slot].pio.aotio [k].scanrate);
		    printf ("Next Scan %ld: ", pmod->rtp.io [slot].pio.aotio [k].next_scan);
		    printf ("Value = %d\n", pmod->rtp.io [slot].pio.aotio [k].out_value);
		}
	    }
	    break;

	case RTP_DIN_8:
	    printf ("Slot %d DIN\n", slot);
	    for (k = 0; k < MAX_RTP_DIN_8; k++)
	    {
		if (pmod->rtp.io [slot].pio.dinio [k].was_requested)
		{
		    printf ("DI %d: ", k);
		    printf ("Is Responding %d: ", pmod->rtp.io [slot].pio.dinio [k].is_responding);
		    printf ("Scan Rate %ld: ", pmod->rtp.io [slot].pio.dinio [k].scanrate);
		    printf ("Next Scan %ld: ", pmod->rtp.io [slot].pio.dinio [k].next_scan);
		    printf ("Value = %ld\n", pmod->rtp.io [slot].pio.dinio [k].value);
		}
	    }
	    break;

	case RTP_DOT_8:
	    printf ("Slot %d DOT\n", slot);
	    for (k = 0; k < MAX_RTP_DOT_8; k++)
	    {
		if (pmod->rtp.io [slot].pio.dotio [k].was_requested)
		{
		    printf ("DO %d: ", k);
		    printf ("Is Responding %d: ", pmod->rtp.io [slot].pio.dotio [k].is_responding);
		    printf ("Scan Rate %ld: ", pmod->rtp.io [slot].pio.dotio [k].scanrate);
		    printf ("Next Scan %ld: ", pmod->rtp.io [slot].pio.dotio [k].next_scan);
		    printf ("Value = %ld\n", pmod->rtp.io [slot].pio.dotio [k].out_value);
		}
	    }
	    break;

	default:
	    printf ("Slot %d ERROR\n", slot);
	}
}	    

/**********
*
*	rtpstats
*
*	The purpose of this routine is to dump the plc level
*	statistics for this driver.
*
**********/

void rtpstats (unsigned short line, unsigned short mod)
{
	RTPIO		*pmod;
	unsigned int	i;

	pmod = plcs [line] [mod];
	if (!pmod)
	{
	    logMsg ("Line Not Defined \n",0,0,0,0,0,0);
	    return;
	}

	printf ("Was requested = %d\n", pmod->was_requested);
	printf ("Is responding = %d\n", pmod->is_responding);
	printf ("Off Lines = %d\n", no_offlines [line]);
	printf ("Rack Timeouts = %d\n", pmod->rack_tmouts);
	printf ("Reset Statistics = %d\n", pmod->cmd_reset_stat);
	printf ("Force Offline = %d\n", pmod->cmd_force_offline);
	printf ("Resource Responding = %d \n", resresponse [line]);
	printf ("Hold flag  = %d\n", pmod->holdflag );
	printf ("Scan Fault = %d\n", scan_fault [line]);
	printf ("Scan Deviation Limit = %d\n", scan_devlimit [line]);
	printf ("Scan Deviation = %d\n", scan_deviation [line]);
	printf ("AI Scan Count = %d\n", ai_scan_count [line]);
	printf ("BI Scan Count = %d\n", bi_scan_count [line]);
	printf ("AO Scan Count = %d\n", ao_scan_count [line]);
	printf ("BO Scan Count = %d\n\n", bo_scan_count [line]);
	for (i = 0; i < MAX_GATE_ADDR; i++)
	{
	    printf ("I/O Slot %d Timeouts = %d\n", i, pmod->rtp.io [i].tmouts);
	}

}

/**********
*
*	Print the BO Recorder structures
*
**********/
void rtpborecorder (int line)
{
	int	i, j;

	printf ("	Time		Point Id		IMask	IVal	PMask	CMask	Address	Card\n");
	for (i = borecidx [line], j = 0; j < 200; j++)
	{
	    if (borecorder [line] [i].time [0] == 0)
	    {
		if (++i >= 200)
		{
		    i = 0;
		}
		continue;
	    }
	    borecorder [line] [i].time [25] = 0;
	    printf ("%s %27s\t0x%x\t0x%x\t0x%x\t0x%x\t%d\t%d\n",
			&borecorder [line] [i].time [13],
			borecorder [line] [i].pointname,
			borecorder [line] [i].mask,
			borecorder [line] [i].value,
			borecorder [line] [i].previous_mask,
			borecorder [line] [i].current_mask,
			borecorder [line] [i].rack,
			borecorder [line] [i].card);
	    if (++i >= 200)
	    {
		i = 0;
	    }
	}
}

/*	End of File:	drvRtp.c */
