/*drvrtu.h,v 1.1.1.1 1999/08/03 14:44:33 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
/*	File:	drvRtu.h							       	
/*	Auth:	John L. Dalesio                                        	       	       	
/*	Edit:	04/07/94
/*
/*	This file contains the data structures used by the TIWAY I/O driver.
/*                                                                             	      	*/

#define TRUE	1
#define FALSE	0
#define NULLP	0
#define CR	0x0d

/*	number of lines of modbus 	*/
#define		RTULINES	1
/*	number of modbus plcs per line */

/*                                                                           		
/*	The IOC will assume that the data to be exchanged with each PLC will           	
/*	come from a fixed number of coils, digital inputs, analog inputs               	
/*	or registers. It will read these in groups to minimize the amount of           	
/*	communications overhead. It will also try to build as large a fetch            	
/*	message as possible & required by remembering which consecutive groups         	
/*	have been requested by the SCAN task.                                          	
*/

#define RTUDO		0			/* Card number for coils 		*/
#define RTUDI		1			/* Card number for digital inputs 	*/
#define RTUAI		3			/* Card number for analog inputs 	*/
#define RTUAO		4			/* Card number for registers 		*/
#define RTUAIRATE	32			/* Card Number for AI Scan Rate Control */
#define MPSEUDO		33			/* Card number for pseudo i/o 		*/

#define SMALLRTUDI	8			/* # of Points on a Small DI Card */
#define SMALLRTUDO	8			/* # of Points on a Small DO Card */
#define LARGERTUDI	16			/* # of Points on a Large DI Card */
#define LARGERTUDO	16			/* # of Points on a Large DO Card */
#define AISPERSLOT	9			/* # of Points on an AI Card */
#define AOSPERSLOT	4			/* # of Points on an AO Card */


#define MAXSLOTS	128			/* Maximum Number of slots in a RTU */
#define SLOTSPERCELL	8			/* Number of slots per cell */
#define MAXCELLS	MAXSLOTS/SLOTSPERCELL						/*  to request in one packet        	*/

/*
/*	Conversion types for analog inputs	*/
/*
*/
#define		AICNVINTEGER	10
#define		AICNVDOUBLEINT	11
#define		AICNVFLOAT	12
#define		AICNVDOUBLEFLT	13
#define		AICNVASCIIFLT	14
#define		AICNVSTATUS	15
#define		AICNVTIME	16
#define		AICNVASCII_2	17
#define		AICNVASCII_4	18
#define		AICNVASCII_12	19
#define		AICNVASCII_20	20
#define		MAXCONV		20

/* The following must match the definition in recChoice.ascii */
#define NOCONV 0
#define LINEAR 1
#define ASCFLT 2





/*
/*	Communications I/O Structures
/*
*/

#define		RXSEM		0
#define		TXSEM		1
#define		MAXRXMSG	256
#define		MAXTXMSG	256


typedef	struct {
	unsigned short		protocol;
	unsigned short		rxstate;
	unsigned short		rxtocnt;
	unsigned short		rxhwerrcnt;
	unsigned short		fixedlen;
	unsigned short		rxmsglen;
	unsigned short		rxix;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		rxbuf[MAXRXMSG];
		} 	RXLINE;


typedef	struct {
	unsigned short		protocol;
	unsigned short		txdisable;
	unsigned short		txtocnt;
	unsigned short		txhwerrcnt;
	unsigned short		txmsglen;
	unsigned short		txix;
	SEMAPHORE		*semid;
	unsigned short		*status;
	unsigned char		txbuf[MAXTXMSG];
		} 	TXLINE;
/*
/*	Communications I/O Constants
/*
*/

#define		RXGOOD			1
#define		RXBAD			2
#define		RXTIMEOUT		60	/* 16 ms ticks 				*/
#define		ACTIVITYTIME		300	/* must see activity in 5 seconds 	*/

#define		VARRESP			FALSE
#define		FIXEDRESP		TRUE
#define		ONERETRY		0x8000	/* OR'ed into line number for xfermsg 	*/
#define		RXIGNORE		0
#define		RXADRST			1
#define		RXFNCST			2
#define		RXLENST			3
#define		RXDATAST		4

#define		TXGOOD			1
#define		TXBAD			2

#define		TXTIMEOUT		1255	/* 16 ms ticks: just under 21 seconds 	*/ 

#define		IO_TIMEOUT		3	/* Code returned by timer 		*/

#define		MAXRETRYS		3	/* retransmission trys 			*/

#define		MAXCFGCK		1	/* Try to Bring online each cycle 	*/
#define		BRINGBACKCNT		4	/* Poll loops per cycle 		*/
#define		BRINGBACKTIME		600	/* Seconds per attempt to bring units on line 		*/

#define		XFERGOOD		1
#define		XFERBAD			2

#define		RESNAMELEN		16	/* length of a resource name 		*/
#define		SLWWINDOW		900	/* # of ticks to allow slave to try 	*/



/*
/* Output Queue Management Structures for Modbus Protocol
/*
*/

#define         AOUTTYPE        1
#define         BOUTTYPE        2
#define         OUTQSIZE        128

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  aosignal;
	short           aovalue;
	double		dblaovalue;
	unsigned short	conversion;
		} AOUT, *AOUTPTR;

typedef struct {
	unsigned short  link;
	unsigned short  adapter;
	unsigned short  card;
	unsigned short  bosignal;
	long            bomask;
	long            bovalue;
		} BOUT, *BOUTPTR;

typedef struct {
		unsigned short  otype;
		union   {
			AOUT    aonode;
			BOUT    bonode;
			}out;
	       } OUTQENT;

typedef struct {
		unsigned short	ooutix;
		unsigned short	oinix;
		OUTQENT         outque[OUTQSIZE];
		} OUTPUTQUE;
/*
/*	Internal IOC Structure for MICROSOL RTU
/*
*/
/*	All of the data necessary to process each slot */
/*	There are 8 slots per units and a maximum of 16 units */
/*	The first unit is local I/O to this cell */
/*	the remaining 15 units are extension cells via serial comm */
typedef struct  {
	unsigned short	aival;		/* Analog Input Value */
	unsigned short	ai_data_valid;	/* Ai Has Been read */
	unsigned short	ai_scan_rate;	/* Rate in ticks to scan this point */
	unsigned long	ai_read_time;	/* Last time read */
	unsigned short	ai_need_read;	/* Ai Needs to Be Read */
	unsigned short	ai_freq;	/* Frequency of ai board */
	unsigned short	ai_read_again;	/* Set when this val changes by x% */
    }	AIPOINT;

#define RTUAIRESET	0		/* AI state Reset */
#define RTUAIZEROWAIT	1		/* AI state Read Zero */
#define RTUAIIDLESTATE	2		/* AI state Idle */
#define RTUAISETTLE	3		/* AI state Select Settle */
#define RTUAICONVERT	4		/* AI state Wait for Conversion */


typedef struct  {
	unsigned short	card_type;		/* Type of module in slot */
	unsigned short	is_initialized;		/* Card has been initialized */
						/* this is reset when an error */
						/* occures to force a reset */
	AIPOINT		ais[AISPERSLOT];	/* Analog Inputs */
	unsigned short	aovals[AOSPERSLOT];	/* Analog Outputs */
	unsigned long	divals;			/* Digital Input Values */
	unsigned long	dovals;			/* Digital Output Values */
	unsigned long	aitime;			/* time hack used by ai set */
						/* to expected next action time */
	unsigned short	aipoint;		/* current ai point being processed */
	unsigned short	aistate;		/* state of current ai conversion */
	unsigned short	ai_freq;		/* Frequency of ai board */
    } SLOT;

typedef struct  {
	unsigned short  holdflag;		/*    The RTU data is in hold 			*/
	unsigned short  was_requested;		/* 0  The scan task has accesses this RTU 	*/
	unsigned short	data_valid;		/*    The Cell Data is current */
	unsigned short  is_responding;		/* 1  The RTU is responding 			*/
	unsigned short  no_dataout_msgs;	/* 5  Count Output Msgs 			*/
	unsigned short  no_datain_msgs;		/* 6  Count Data Requests 			*/
	unsigned short  no_rxnoresp;		/* 7  Count No Responses 			*/
	unsigned short  no_txnoresp;		/* 8  Count No Responses 			*/
	unsigned short  no_dataerr;		/* 9  Count Message Echo Errors 		*/
	unsigned short  no_retries;		/* 10 Count Retries 				*/
	unsigned short  no_rxhwerrs;		/* 11 Count Comm Hardware Errors 		*/
	unsigned short  no_txhwerrs;		/* 12 Count Comm Hardware Errors 		*/
	unsigned short  no_crcerrs;		/* 13 Count CRC Errors 				*/
	unsigned short  no_offlines;		/* 14 Count Off Line Transitions		*/
	unsigned short  cmd_reset_stat;		/* 15 Comand to reset stats			*/
	unsigned short  cmd_force_offline;	/* 16 Command to Force Offline 			*/
	unsigned short  badoutcnt;		/* 17 Count Failed output commands 		*/
						/* 18 Back up Mode 1 = Auto 0 = Man		*/
						/* 19 Active Channel 1= Priamry 0 = Backup 	*/

	SLOT		sl[SLOTSPERCELL]; /* Data from each cell */
    } CELL;



/* DEFINITIONS FOR MICROSOL IO */

#define NUMCARDS 8
#define NUMAICHANS 8
#define NUMAOCHANS 4
#define NUMDICHANS 8
#define NUMDICHANSX 16
#define NUMDOCHANS 8

#define RTUDEFAULTAIRATE	180	/* 3 seconds */
#define   CDTYPEAO		0x41
#define   CDTYPEAI60HZ		0x32
#define   CDTYPEAI50HZ		0x33
#define   CDTYPEDO		0x20
#define   CDTYPEDI8		0x10
#define   CDTYPEDI32		0x11
#define   CDTYPE_		-1


#define   SLOT1		(char *)0xf08200	 /*bpslot1,top*/
#define   SLOT2		(char *)0xf08400	 /*bpslot1,btm*/
#define   SLOT3		(char *)0xf08240	 /*bpslot2,top*/
#define   SLOT4		(char *)0xf08440	 /*bpslot2,btm*/
#define   SLOT5		(char *)0xf08280	 /*bpslot3,top*/
#define   SLOT6		(char *)0xf08480	 /*bpslot3,btm*/
#define   SLOT7		(char *)0xf082c0	 /*bpslot4,top*/
#define   SLOT8		(char *)0xf084c0	 /*bpslot4,btm*/

#define   IOA(x) (((x)<<1)+1)

#define  AO_SLOT_ID	IOA(0x10)
#define  AO_BRD_ENABLE	IOA(0x18)  /*write 0xff*/
#define  AO_ACTIVATE	IOA(0x00)  /*read&discard*/
#define  AO_RLY_CTL	IOA(0x0e)  /*write 0x0f*/
#define  AO_RLY_STAT	IOA(0x0c)  /*read=0x0f*/
#define  AO_= -1

#define    AI_SLOT_ID		IOA(0x10)
/*#define    AI_BRD_ENABLE	0x18      /*write 0xff*/
#define    AI_CHAN_SELECT	IOA(0x00)
#define    AI_DISCON		IOA(0x00) /*write 0x00*/
#define    AI_VALUE		IOA(0x08)
#define    AI_BRD_CTL		IOA(0x08)
#define    AI_BRD_STAT		IOA(0x08)
	/*status register masks*/
#define    AISTAT_SLEEPING	0x10
#define    AISTAT_INTEGRATE	0x20
#define    AISTAT_CONVDONE	0x40
#define    AISTAT_COLLISION	0x80
	/*control register masks*/
#define    AICTL_REGDRLSB	0x00      /*data reg LSB*/
#define    AICTL_REGDRMSB	0x04      /*data reg MSB*/
#define    AICTL_REGSTAT	0x02
#define    AICTL_SELZOFF	0x10
#define    AICTL_SELVSRC	0x00
#define    AICTL_GOSLEEP	0x20
#define    AICTL_RUNMODE	0x00
#define    AICTL_RATE50HZ	0x40
#define    AICTL_RATE60HZ	0x00
#define    AICTL_CONVERT	0x80
	/*various useful commands*/
#define    AICTL_INIT_BOARD	(AICTL_RUNMODE|AICTL_REGSTAT|AICTL_SELZOff)
#define    AICTL_START_ZOFF_CNV	(AICTL_RUNMODE|AICTL_REGSTAT|AICTL_CONVERT|AICTL_SELZOFF)
#define    AICTL_START_CHAN_CNV	(AICTL_RUNMODE|AICTL_REGSTAT|AICTL_CONVERT|AICTL_SELVSRC)
#define    AICTL_CONT_ZOFF_CNV	(AICTL_RUNMODE|AICTL_REGSTAT|AICTL_SELZOFF)
#define    AICTL_CONT_CHAN_CNV	(AICTL_RUNMODE|AICTL_REGSTAT|AICTL_SELVSRC)
#define    AICTL_READ_LSB	(AICTL_RUNMODE|AICTL_REGDRLSB)
#define    AICTL_READ_MSB	(AICTL_RUNMODE|AICTL_REGDRMSB)
#define    AI_=		-1
#define	   AISETTLETIME	 2	/* At Least 2 ticks to make sure we get 1 */
#define	   MAXAICVTTIME	 6	/* 6 ticks to get at least 60 ms */

#define    DO_SLOT_ID	IOA(0x10)
#define    DO_VALUE	IOA(0x00)
#define    DO_		-1

#define    DI_SLOT_ID	IOA(0x10)
#define    DI_LVALUE	IOA(0x00)
#define    DI_HVALUE	IOA(0x04)
#define    DI_		-1

#define    SLOT_ID	IOA(0x10)


/* Arcnet Communications Structures */
typedef struct  {
	unsigned char	id;
	unsigned char	physid;
	unsigned char	npts;
     }	IOCFG;

typedef struct {
	IOCFG	cfg[SLOTSPERCELL];
    } IOCFGMSG;

typedef struct {
	unsigned short	slot;
	unsigned short  aivals[AISPERSLOT];
    } AISLOT;

typedef struct {
	unsigned short  slot;
	unsigned short	aovals[AOSPERSLOT];
    } AOSLOT;

typedef struct {
	unsigned short  slot;
	unsigned char	divals[8];
    } DI8SLOT;

typedef struct {
	unsigned short  slot;
	unsigned char	divals[16];
    } DI16SLOT;

typedef struct {
	unsigned short  slot;
	unsigned char	dovals[NUMDOCHANS];
    } DOSLOT;

typedef struct {
	unsigned short  slot;
	unsigned short  point;
	unsigned short  val;
    } AOCMD;

typedef struct {
	unsigned short  slot;
	unsigned short  point;
	unsigned short  val;
    } DOCMD;

typedef struct {
	unsigned short  slot;
	unsigned short  point;
	unsigned long   sec;
	unsigned long	msec;
    } AIRATECMD;

typedef struct {
	unsigned short  slot;
	unsigned short  pad;
	unsigned long   sec;
	unsigned long	msec;
    } DIRATECMD;

typedef struct {
	unsigned short	nslots;
	union {
	    DOSLOT	dosl;
	    DI16SLOT	di16sl;
	    DI8SLOT	di8sl;
	    AISLOT	aisl;
	    AOSLOT	aosl;
	} slot[SLOTSPERCELL];
    } UPDATEMSG;


/* task definitions */

#define RTU_STACK		8000   /* Stack size              */
#define	RTU_PRIORITY		152    /* default priority        */
#define	RTU_TASK_OPTIONS	8      /* Floating point          */


/*	End of File:	drvRtu.h */
