/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/



/****************************************************************************
 *
 *  MODULE NAME            : devSecurity.c
 *  AUTHOR                 : emsteel
 *  DATE                   : 22 Dec 93
 *
 *
 * DESCRIPTION: device support for Security
 *
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             :
 *  REVISION DATE          :
 *  REVISION:
 *
****************************************************************************/

/****************************************************************************
 *
 * Programmer's Notes:
 *    Where noted, change appropriately to create a new device support
 *    module
 *
 *
****************************************************************************/


#include   <vxWorks.h>
#include   <types.h>
#include   <stdioLib.h>

#include   <alarm.h>
#include   <cvtTable.h>
#include   <dbAccess.h>
#include   <dbDefs.h>
#include   <devSup.h>
#include   <link.h>
#include   <module_types.h>
#include   <cvnTypes.h>
#include   <special.h>
#include   <aiRecord.h>
#include   <biRecord.h>
#include   <boRecord.h>
#include   <mbbiRecord.h>
#include   <stringinRecord.h>
#include   <stringoutRecord.h>

#define  IO_TYPE     SECURITY_IO		/* NOTE: change here */
#define  AIDRIVER    security_aidriver		/* NOTE: change here */
#define  BIDRIVER    security_bidriver		/* NOTE: change here */
#define  BODRIVER    security_bodriver		/* NOTE: change here */
#define  BOREAD      security_boread		/* NOTE: change here */
#define  SIDRIVER    security_sidriver		/* NOTE: change here */
#define  SODRIVER    security_sodriver		/* NOTE: change here */

static char *io_name = "Security";		/* NOTE: change here */


/* Create the dset for devAiSecurity */

static long ai_init_record();
static long read_ai();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   read_ai;
   DEVSUPFUN   special_linconv;
} devAiSecurity = {                        /* NOTE: change here */
   6,
   NULL,
   NULL,
   ai_init_record,
   NULL,
   read_ai,
   NULL};

/* Create the dset for devBiSecurity */

static long bi_init_record ();
static long read_bi ();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   read_bi;
} devBiSecurity = {                        /* NOTE: change here */
   5,
   NULL,
   NULL,
   bi_init_record,
   NULL,
   read_bi};

/* Create the dset for devBoSecurity */

static long bo_init_record ();
static long write_bo ();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   write_bo;
} devBoSecurity = {                        /* NOTE: change here */
   5,
   NULL,
   NULL,
   bo_init_record,
   NULL,
   write_bo};

/* Create the dset for devMbbiSecurity */

static long mbbi_init_record ();
static long read_mbbi ();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   read_mbbi;
} devMbbiSecurity = {                      /* NOTE: change here */
   5,
   NULL,
   NULL,
   mbbi_init_record,
   NULL,
   read_mbbi};

/* Create the dset for devSiSecurity */

static long si_init_record();
static long read_stringin();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   read_stringin;
} devSiSecurity = {                      /* NOTE: change here */
   5,
   NULL,
   NULL,
   si_init_record,
   NULL,
   read_stringin};

/* Create the dset for devSoSecurity */

static long write_stringout();

struct {
   long        number;
   DEVSUPFUN   report;
   DEVSUPFUN   init;
   DEVSUPFUN   init_record;
   DEVSUPFUN   get_ioint_info;
   DEVSUPFUN   write_stringout;
} devSoSecurity = {                      /* NOTE: change here */
   5,
   NULL,
   NULL,
   NULL,
   NULL,
   write_stringout};

static long masks [] = {
   0x00000001, 0x00000002, 0x00000004, 0x00000008,
   0x00000010, 0x00000020, 0x00000040, 0x00000080,
   0x00000100, 0x00000200, 0x00000400, 0x00000800,
   0x00001000, 0x00002000, 0x00004000, 0x00008000,
   0x00010000, 0x00020000, 0x00040000, 0x00080000,
   0x00100000, 0x00200000, 0x00400000, 0x00800000,
   0x01000000, 0x02000000, 0x04000000, 0x08000000,
   0x10000000, 0x20000000, 0x40000000, 0x80000000,
};


/********************************** AI ************************************/

static long ai_init_record (pai)
    struct aiRecord	*pai;
{
    struct tateio	*ptateio;
    char		message[128];
    unsigned short	value;
    int			status;
    union u_value 
    {
	short    ivalue;
	long     lvalue;
	float    fvalue;
	double   dvalue;
	char     afvalue[100];
    }  uval;

    switch (pai->inp.type)
    {
	case (IO_TYPE):
	    break;

	default:
	    strcpy (message, pai->name);
	    strcat (message, ": devAi");
	    strcat (message, io_name);
	    strcat (message, " (init_record) Illegal INP field");
	    errMessage (S_db_badField, message);
	    return (S_db_badField);
    }

      /* call driver so that it configures card */

    ptateio = (struct tateio *) &(pai->inp.value);
    status = AIDRIVER (pai->name,ptateio->link, ptateio->adapter, 
		       ptateio->card, ptateio->signal, &uval, 
		       pai->linr, pai->resn);

    if (status == 0)
    {
	switch(pai->linr)
	{
	    case (CNVINTEGER):
		pai->val = uval.ivalue;
		break;

	    case (CNVDOUBLEINT):
		pai->val = uval.lvalue;
		break;

	    case (CNVFLOAT):
		pai->val = uval.fvalue;
		break;

	    case (CNVDOUBLEFLT):
		pai->val = uval.dvalue;
		break;

	    case (CNVASCIIFLT):
		status = -1;

	    case NOCONV:
	    case LINEAR:
	    case CNVSQRROOT:
	    default:
		pai->rval = uval.ivalue;
		break;
	}
    }

    return (status);
}

static long read_ai (pai)
    struct aiRecord	*pai;
{
    struct tateio	*ptateio;
    int			status;
    unsigned short	value;

    union u_value 
    {
	short    ivalue;
	long     lvalue;
	float    fvalue;
	double   dvalue;
	char     afvalue[100];
    }  uval;

    ptateio = (struct tateio *) &(pai->inp.value);
    status = AIDRIVER (pai->name,ptateio->link, ptateio->adapter, 
		       ptateio->card, ptateio->signal, &uval, 
		       pai->linr, 0);

    /* setting status to 2 tells ai's process() not to convert value */

    if (status == 0)
    {
	switch(pai->linr)
	{
	    case (CNVINTEGER):
		pai->val = uval.ivalue;
		status = 2;
		break;

	    case (CNVDOUBLEINT):
		pai->val = uval.lvalue;
		status = 2;
		break;

	    case (CNVFLOAT):
		pai->val = uval.fvalue;
		status = 2;
		break;

	    case (CNVDOUBLEFLT):
		pai->val = uval.dvalue;
		status = 2;
		break;

	    case (CNVASCIIFLT):
		status = -1;

	    case NOCONV:
	    case LINEAR:
	    case CNVSQRROOT:
	    default:
		pai->rval = uval.ivalue;
		break;
	}
    }

    if (status == -1)
    {
	if ((pai -> aenb != ALARM_DISABLED) && (pai->nsev < VALID_ALARM))
	{
	    pai->nsta = READ_ALARM;
	    pai->nsev = VALID_ALARM;
	}
	status = 2;         /*dont convert*/
    }

    return (status);
}


/*********************************** BI ***********************************/

static long bi_init_record (pbi)
    struct biRecord	*pbi;
{
    struct tateio	*ptateio;
    char		message[128];
    int			status = 0;
    unsigned long	value;

    switch (pbi->inp.type)
    {
	case (IO_TYPE) :
	    ptateio = (struct tateio *) &(pbi->inp.value);
	    pbi->mask = masks[((ptateio->signal - 1) % 32)];
	    status = BIDRIVER (pbi->name,ptateio->link, ptateio->adapter, 
			       ptateio->card, ptateio->signal, pbi->mask, 
			       &value, &pbi->resn [0]);
	    if (!status)
		pbi -> rval = value;

	    break;

	default :
	    strcpy (message, pbi->name);
	    strcat (message, ": devBi");
	    strcat (message, io_name);
	    strcat (message, " (init_record) Illegal INP field");
	    errMessage (S_db_badField, message);
	    status = S_db_badField;

    }
    return (status);
}

static long read_bi(pbi)
    struct biRecord	*pbi;
{
    struct tateio	*ptateio;
    int			status;
    unsigned long	value;

    ptateio = (struct tateio *) &(pbi->inp.value);
    status = BIDRIVER (pbi->name,ptateio->link, ptateio->adapter, 
		       ptateio->card, ptateio->signal, pbi->mask, 
		       &value, 0);
    if (status == 0)
    {
	pbi->rval = value;
	return (status);
    }
    else
    {
	if ((pbi -> aenb != ALARM_DISABLED) && (pbi->nsev < VALID_ALARM))
	{
	    pbi->nsta = READ_ALARM;
	    pbi->nsev = VALID_ALARM;
	}
	return (2);
    }
}


/*********************************** BO ***********************************/

static long bo_init_record (pbo)
    struct   boRecord *pbo;
{
    struct tateio	*ptateio;
    char		message[128];
    unsigned long	value;
    int			status = 0;

    switch (pbo->out.type)
    {
	case (IO_TYPE) :
	    ptateio = (struct tateio *) &(pbo->out.value);

	    /* read the value via the modbus driver */

	    pbo->mask = masks [((ptateio->signal - 1) % 32)];
	    status = BOREAD (pbo->name,ptateio->link, ptateio->adapter, 
	   		     ptateio->card, ptateio->signal, &value, 
			     pbo->mask, &pbo->resn [0]);
	    if(status == 0)
	    {
		pbo->rval = value;
	    }
	    break;

	default :
	    strcpy (message, pbo->name);
	    strcat (message, ": devBo");
	    strcat (message, io_name);
	    strcat (message, " (init_record) Illegal OUT field");
	    errMessage (S_db_badField, message);
	    status = S_db_badField;
	    break;
    }
    return (status);
}

static long write_bo (pbo)
    struct boRecord	*pbo;
{
    struct tateio	*ptateio;
    int			status;

    ptateio = (struct tateio *) &(pbo->out.value);
    status = BODRIVER (pbo->name,ptateio->link, ptateio->adapter, 
		       ptateio->card, ptateio->signal, pbo->rval, 
		       pbo->mask, 0);
    if (status != 0)
    {
	if ((pbo -> aenb != ALARM_DISABLED) && (pbo -> nsev < VALID_ALARM))
	{
	    pbo->nsta = WRITE_ALARM;
	    pbo->nsev = VALID_ALARM;
	}
    }
    return (status);
}


/*********************************** MBBI **********************************/

static long mbbi_init_record (pmbbi)
    struct mbbiRecord	*pmbbi;
{
    struct tateio	*ptateio;
    char		message[128];
    int			status = 0;
    unsigned long	value;

    switch (pmbbi->inp.type)
    {
	case (IO_TYPE) :
	    ptateio = (struct tateio *) &(pmbbi->inp.value);
	    pmbbi->shft = ((pmbbi->inp.value.tateio.signal - 1) % 32);
	    pmbbi->mask <<= pmbbi->shft;
	    status = BIDRIVER (pmbbi->name,ptateio->link, ptateio->adapter, 
			       ptateio->card, ptateio->signal, pmbbi->mask, 
			       &value, &pmbbi->resn [0]);
	    if (!status)
		pmbbi -> rval = value;

	    break;

	default :
	    strcpy (message, pmbbi->name);
	    strcat (message, ": devMbbi");
	    strcat (message, io_name);
	    strcat (message, " (init_record) Illegal INP field");
	    errMessage (S_db_badField, message);
	    status = S_db_badField;
	    break;
    }
    return (status);
}

static long read_mbbi (pmbbi)
    struct mbbiRecord	*pmbbi;
{
    struct tateio	*ptateio;
    int			status = 0;
    unsigned long	value;

    ptateio = (struct tateio *) &(pmbbi->inp.value);
    status = BIDRIVER (pmbbi->name,ptateio->link, ptateio->adapter, 
		       ptateio->card, ptateio->signal, pmbbi->mask, 
		       &value, 0);
    if (status == 0)
    {
	pmbbi->rval = value;
    }
    else
    {
	if ((pmbbi -> aenb != ALARM_DISABLED) && (pmbbi->nsev < VALID_ALARM))
	{
	    pmbbi->nsta = READ_ALARM;
	    pmbbi->nsev = VALID_ALARM;
	}
    }
    return (status);
}


/********************************* SI *************************************/

static long si_init_record(pstringin)
    struct stringinRecord	*pstringin;
{
    struct tateio	*ptateio;
    char		message[128];
    int			status = 0;

    switch (pstringin->inp.type)
    {
	case (IO_TYPE) :
	    ptateio = (struct tateio *) &(pstringin->inp.value);
	    status = SIDRIVER (pstringin->name, ptateio->link, ptateio->adapter, 
			       ptateio->card, ptateio->signal, message, 
			       pstringin->rdsz, pstringin->resn);
	    if (!status)
	    {
		pstringin->udf = FALSE;
		strncpy (pstringin -> val, message, sizeof(pstringin -> val));
	    }
	    break;

	default :
	    strcpy (message, pstringin->name);
	    strcat (message, ": devSi:");
	    strcat (message, io_name);
	    strcat (message, " (init_record) Illegal INP field");
	    errMessage (S_db_badField, message);
	    status = S_db_badField;
    }
    return(status);
}

static long read_stringin(pstringin)
    struct stringinRecord	*pstringin;
{
    struct tateio	*ptateio;
    int			status;
    char		message[128];


    ptateio = (struct tateio *) &(pstringin->inp.value);
    status = SIDRIVER (pstringin->name, ptateio->link, ptateio->adapter, 
		       ptateio->card, ptateio->signal, message, 
		       pstringin->rdsz, 0);

    if(status!=0)
    {                 
	if((pstringin -> aenb != ALARM_DISABLED) &&
	   (pstringin -> nsev < VALID_ALARM))
	{
	    pstringin->nsev = VALID_ALARM;
	    pstringin->nsta = READ_ALARM;
	}
    }
    else
    {
	pstringin->udf = FALSE;
	strncpy (pstringin->val, message, sizeof (pstringin -> val));
    }
   
    return(status);
}


/*********************************** SO ************************************/

static long write_stringout(pstringout)
    struct stringoutRecord	*pstringout;
{
    struct tateio	*ptateio;
    char		message[128];
    long		status;

    switch (pstringout->out.type)
    {
	case (IO_TYPE) :

	    /* Format the message String */

	    message[0]= 0;
	    if (pstringout->desc[0] == '+')
	    {
		strcpy (message, &(pstringout -> desc[1]));
	    }
	    strcat (message, pstringout->val);
	    if (pstringout->desc[0] == '-')
	    {
		strcat (message, &(pstringout-> desc[1]));
	    }
	    ptateio = (struct tateio *) &(pstringout->out.value);
	    status = SODRIVER (pstringout->name, ptateio->link, ptateio->adapter,
			       ptateio->card, ptateio->signal, message, 0, 0);
    	    break;

	default :
	    if ((pstringout -> aenb != ALARM_DISABLED) &&
		(pstringout -> nsev < VALID_ALARM))
	    {
		pstringout->nsev = VALID_ALARM;
		pstringout->nsta = SOFT_ALARM;
		if(pstringout->stat!=SOFT_ALARM)
		{
		    strcpy (message, pstringout->name);
		    strcat (message, ": devSo");
		    strcat (message, io_name);
		    strcat (message, " (write_stringout) Illegal OUT field");
		    errMessage(S_db_badField,message);
		}
	    }
	    status = (-1);
    }
    return(status);
}


