/* password.cxx,v 1.1.1.1 1999/08/03 14:45:00 tis4000 Exp
 *
 */

/*
 * password - TIS4000 Application control bar client machines
 *
 * Displays a menu bar for executing TIS4000 applications. If a
 * login has not been established with a server, a login dialog
 * box is displayed.
 * Process also accepts connections for any apps requiring priviledges
 * or tracking, and passes the requests through to the OpPriv server.
 * Multithreaded - passwd menu and GUI are main thread, opPriv in
 * secondary thread
 *
 */

#include <vport.h>		/* for v*HEADER */
#include vstdlibHEADER		/* for random, exit, EXIT_SUCCESS */
#include vapplicationHEADER
#include vmenubarHEADER
#include vmenuHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vmemHEADER
#include vprefHEADER
#include vtextitemHEADER
#include vcharHEADER
#include vbuttonHEADER

#ifdef WIN32
#include <windows.h>
#include <winbase.h>
#include <stdio.h>
#include <process.h>
#endif WIN32

#include "tisgalaxylib.hxx"
#include "textdlg.h"

extern "C"
{
#include "dbDefs.h"
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32
#include "client.h"
#include "OperPriv.h"
}

class AppMenuItem : public vmenuItem
{
    public:
	// macro for overrides from vkindTyped
	vkindTYPED_FULL_DECLARATION(AppMenuItem);

	// macro for the standard versions of the 
	// necessary constructors and destructors
	vloadableINLINE_CONSTRUCTORS(AppMenuItem, vmenuItem);

	// virtual member function overrides
	void ObserveMenuItem(vevent *event);

	// configuration routine
	void SetAppCommand (char *command) { strncpy (appcommand, command, 255); }
    private:
	char	appcommand [255];   /* application command space */
};

class HostItem : public vmenuItem
{
    public:
	// macro for overrides from vkindTyped
	vkindTYPED_FULL_DECLARATION(HostItem);

	// macro for the standard versions of the 
	// necessary constructors and destructors
	vloadableINLINE_CONSTRUCTORS(HostItem, vmenuItem);

	// virtual member function overrides
	void ObserveMenuItem(vevent *event);
};

class LoginItem : public vmenuItem
{
    public:
	// macro for overrides from vkindTyped
	vkindTYPED_FULL_DECLARATION(LoginItem);

	// macro for the standard versions of the 
	// necessary constructors and destructors
	vloadableINLINE_CONSTRUCTORS(LoginItem, vmenuItem);

	// virtual member function overrides
	void ObserveMenuItem(vevent *event);
};

class LogoutItem : public vmenuItem
{
    public:
	// macro for overrides from vkindTyped
	vkindTYPED_FULL_DECLARATION(LogoutItem);

	// macro for the standard versions of the 
	// necessary constructors and destructors
	vloadableINLINE_CONSTRUCTORS(LogoutItem, vmenuItem);

	// virtual member function overrides
	void ObserveMenuItem(vevent *event);
};

class ExitItem : public vmenuItem
{
    public:
	// macro for overrides from vkindTyped
	vkindTYPED_FULL_DECLARATION(ExitItem);

	// macro for the standard versions of the 
	// necessary constructors and destructors
	vloadableINLINE_CONSTRUCTORS(ExitItem, vmenuItem);

	// virtual member function overrides
	void ObserveMenuItem(vevent *event);
};

class AboutItem : public vmenuItem
{
    public:
	// macro for overrides from vkindTyped
	vkindTYPED_FULL_DECLARATION(AboutItem);

	// macro for the standard versions of the 
	// necessary constructors and destructors
	vloadableINLINE_CONSTRUCTORS(AboutItem, vmenuItem);

	// virtual member function overrides
	void ObserveMenuItem(vevent *event);
};

class LoginButton : public vbutton {
public:
  // macro for overrides from vkindTyped
  vkindTYPED_FULL_DECLARATION(LoginButton);
  
  // macro for the standard versions of the 
  // necessary constructors and destructors
  vloadableINLINE_CONSTRUCTORS(LoginButton, vbutton);
  
  // virtual member function overrides
  void ObserveDialogItem(vevent *event);
  
};

class CancelButton : public vbutton {
public:
  // macro for overrides from vkindTyped
  vkindTYPED_FULL_DECLARATION(CancelButton);
  
  // macro for the standard versions of the 
  // necessary constructors and destructors
  vloadableINLINE_CONSTRUCTORS(CancelButton, vbutton);
  
  // virtual member function overrides
  void ObserveDialogItem(vevent *event);
  
};

class PasswordTextitem : public vtextitem
{
    public:
	// the default versions of the necessary constructors/destructor
	vloadableINLINE_CONSTRUCTORS(PasswordTextitem, vtextitem);

	// overrides from vkindTyped
	vkindTYPED_FULL_DECLARATION(PasswordTextitem);

	// virtual member function overrides

	// vtextview
	virtual void ObserveView(vevent *event, int message);
};

class PasswordText : public vtextitemText
{
    public:
	// the default versions of the necessary constructors/destructor
	vloadableINLINE_CONSTRUCTORS(PasswordText, vtextitemText);

	// overrides from vkindTyped
	vkindTYPED_FULL_DECLARATION(PasswordText);

	// virtual member function overrides

	// vkindTyped
	virtual void Destroy();

	// vtext
	virtual void DeleteText(vtextSelection *selection);
	virtual void InsertText(vtextSelection *selection,
			      const vchar *text,
			      long length);

    //private:
	vtext	   *realText;
	vtextSelection *realSelection;
};  

class PasswordDialog : public vdialog
{
    public:
	// the default versions of the necessary constructors/destructor
	vloadableINLINE_CONSTRUCTORS(PasswordDialog, vdialog);

	// overrides from vkindTyped
	vkindTYPED_FULL_DECLARATION(PasswordDialog);

	// virtual member function overrides
	// vloadable
	virtual void LoadInit(vresource resource);
	// vwindow
	virtual void ObserveWindow(vevent *event);

	// access password text item
	PasswordTextitem *GetPasswordItem () { return textitem; }
	
    private:
	LoginButton *loginbutton;	    // tis4000 login button
	CancelButton *cancelbutton;	    // cancel button
	PasswordTextitem *textitem; // Password item
};

class ApplicationDialog : public vdialog
{
    public:
	// the default versions of the necessary constructors/destructor
	vloadableINLINE_CONSTRUCTORS(ApplicationDialog, vdialog);

	// overrides from vkindTyped
	vkindTYPED_FULL_DECLARATION(ApplicationDialog);

	// virtual member function overrides
	// vloadable
	virtual void LoadInit(vresource resource);
	// vwindow
	virtual void ObserveWindow(vevent *event);

	void SetLoggedInAppearance () { appmenu->Enable (); }
	void DisableAppMenu () { appmenu->Disable (); }

    private:
	void ResetAppRead (void);
	char *GetNextApp (char *);

	vmenu *appmenu;		    // tis4000 aplications menu
	FILE *fp;		    // file for reading menu entries
	char menutext[40];	    // displayed to user in appmenu
	char inpstring[300];	    // temporary input space for menu definitions
};

// Global values -
ApplicationDialog *maindialog;
PasswordDialog *pwdialog;
TextDialog *textdialog;

int curUserLevel;
char *userStack [10];
char *passStack [10];
char msg [512];

// macro definitions for vkindtyped overrides
vkindTYPED_FULL_DEFINITION(AppMenuItem, vmenuItem, NULL);
vkindTYPED_FULL_DEFINITION(LoginButton, vbutton, NULL);
vkindTYPED_FULL_DEFINITION(CancelButton, vbutton, NULL);
vkindTYPED_FULL_DEFINITION(HostItem, vmenuItem, NULL);
vkindTYPED_FULL_DEFINITION(LoginItem, vmenuItem, NULL);
vkindTYPED_FULL_DEFINITION(LogoutItem, vmenuItem, NULL);
vkindTYPED_FULL_DEFINITION(ExitItem, vmenuItem, NULL);
vkindTYPED_FULL_DEFINITION(AboutItem, vmenuItem, NULL);

/*
 *  AppMenuItem::ObserveMenuItem
 *
 *  Perform designated command.
 *
 */
char *argv [20];
char cmdtext [300];

void AppMenuItem::ObserveMenuItem(vevent *event)
{
    char *argp;
    static char execdir [200];
    static char savedir [200];
    char *execdirp;
    int argc;

//    system (appcommand);
    strncpy (cmdtext, appcommand, 300);

    /* break command into pieces */
    argp = cmdtext;
    execdir[0] = 0;
    if (*argp == '(')
    {
	execdirp = execdir;
	argp++;
	while (*argp && *argp != ')')
	{
	    *execdirp++ = *argp++;
	}
	*execdirp = 0;
	if (*argp)
	    argp++;
	while (*argp && (*argp == ' ' || *argp == '\t'))
	    argp++;
    }
    for (argc = 0; argp && *argp && argc < 19; argc++)
    {
	if (*argp == '\"')
	{
	    argp++;
	    argv [argc] = argp;
	    argp = strchr (argp, '"');
	    if (argp)
	    {
		*argp = 0;
		argp++;
		while (*argp && (*argp == ' ' || *argp == '\t'))
		    argp++;
	    }
	    else /* missing trailing " */
		break;
	}
	else
	{
	    argv [argc] = argp;
	    if (strchr (argp, ' '))
	    {
		argp = strchr (argp, ' ');
		*argp = 0;
		argp++;
		while (*argp && (*argp == ' ' || *argp == '\t'))
		    argp++;
	    }
	    else
	    {
		argp = strchr (argp, '\t');
		if (argp)
		{
		    *argp = 0;
		    argp++;
		    while (*argp && (*argp == ' ' || *argp == '\t'))
			argp++;
		}
	    }
	}
    }
    argv [argc] = 0;
    if (execdir[0])
    {
	_getcwd (savedir, 200);
	_chdir (execdir);
    }
    _spawnvp (_P_DETACH, argv[0], (const char * const *)argv);
    if (execdir[0])
    {
	_chdir (savedir);
    }
}

/*
 *  LoginButton::ObserveDialogItem
 *
 *  login to priviledge server
 *
 */

void LoginButton::ObserveDialogItem(vevent *event)
{
    PasswordText *text;
    int		len;
    vstr	*str;
    char	*sys;
    char	opername [OPERATOR_SZ+1];
    PasswordDialog   *dialog;
    int		result;
    
    dialog = (PasswordDialog *)DetermineDialog();
        
    text = (PasswordText *) dialog->GetPasswordItem()->GetTextData();

    // Take the text from the real text data object and put it in a string
    len  	= text->realText->GetLength();
    str 	= vstrCreateSized(len + 1);
    str[len] 	= vcharNULL;

    text->realText->GetText(NULL, str, len);

    // Now you have what they really typed - do so what you want here
    vtextitem *oplabel = (vtextitem *)dialog->FindItem(vnameInternGlobalLiteral("OpNameTag"));
    oplabel->GetText ((unsigned char *)opername, sizeof(opername));
    sys = (char *) vcharExportAlloc(vcharScribeSystemExport(str));
    if (curUserLevel)
    {
	result = OperLogoff ();
	if (result != PRIVOK)
	{
	    infoNotice (maindialog, (unsigned char *)"Warning", (unsigned char *)"Could not log off current user, continuing");
	}
    }
    result = OperLogon (opername, sys);
    vstrDestroy(str);
    if (result == PRIVOK)
    {
	userStack [curUserLevel] = (char *)malloc (strlen(opername)+1);
	strcpy (userStack [curUserLevel], opername);
	passStack [curUserLevel] = (char *)malloc (strlen(sys)+1);
	strcpy (passStack [curUserLevel], sys);
	curUserLevel++;
	maindialog->SetLoggedInAppearance ();
	// Change title to new user name
	sprintf (msg, "TIS4000 Control Center (%s)", userStack [curUserLevel-1]);
	maindialog ->SetTitle ((unsigned char *)msg);
	// leave this dialog!
	dialog->Close ();
    }
    else
    {
	infoNotice (maindialog, (unsigned char *)"Warning", (unsigned char *)"Could not log in, check username and re-enter password");
	// reset password to blanks
	dialog->GetPasswordItem()->SetText(NULL);
    }
    vcharExportFree((void *) sys);
}

/*
 *  CancelButton::ObserveDialogItem
 *
 *  cancel login to priviledge server
 *
 */

void CancelButton::ObserveDialogItem(vevent *event)
{
    PasswordDialog   *dialog;
    
    dialog = (PasswordDialog *)DetermineDialog();
    dialog->Close ();
}

/*
 *  HostItem::vmenuItem
 *
 *  Set priviledge server
 *
 */

void HostItem::ObserveMenuItem(vevent *event)
{
    char servername [132];

    // Pop up host dialog
    if (textdialog->Alert ("Enter priviledge server(s)", GetPrivServer (servername, sizeof (servername))))
	SetPrivServer (textdialog->ResultString ());
}

/*
 *  LoginItem::vmenuItem
 *
 *  login to priviledge server
 *
 */

void LoginItem::ObserveMenuItem(vevent *event)
{
    // reset password to blanks
    pwdialog->GetPasswordItem()->SetText(NULL);
    pwdialog->Open ();
}

/*
 *  LogoutItem::vmenuItem
 *
 *  logout of priviledge server
 *
 */

void LogoutItem::ObserveMenuItem(vevent *event)
{
    int		result;
    
    if (curUserLevel == 0)
    {
	// noone is logged on yet
	infoNotice (maindialog, (unsigned char *)"Warning", (unsigned char *)"No one is logged in yet");
    }
    
    result = OperLogoff ();
    if (result == PRIVOK)
    {
	// Disable app menu
	curUserLevel--;
	free (userStack [curUserLevel]);
	free (passStack [curUserLevel]);
	if (curUserLevel)
	{
	    result = OperLogon (userStack [curUserLevel-1], passStack [curUserLevel-1]);
	    if (result == PRIVOK)
	    {
		// Change title to new user name
		sprintf (msg, "TIS4000 Control Center (%s)", userStack [curUserLevel-1]);
		maindialog ->SetTitle ((unsigned char *)msg);
	    }
	    else
	    {
		sprintf (msg, "Could not restore login for %s", userStack [curUserLevel-1]);
		infoNotice (maindialog, (unsigned char *)"Warning", (unsigned char *)msg);
		curUserLevel = 0;
		maindialog ->SetTitle ((unsigned char *)"TIS4000 Control Center (Please log in)");
		maindialog->DisableAppMenu ();
	    }
	}
	else
	{
	    maindialog ->SetTitle ((unsigned char *)"TIS4000 Control Center (Please log in)");
	    maindialog->DisableAppMenu ();
	}
    }
    else
    {
	/* Can't logoff? */
	curUserLevel = 0;
    }
}

/*
 *  ExitItem::vmenuItem
 *
 *  exit from program
 *
 */

void ExitItem::ObserveMenuItem(vevent *event)
{
    OperLogoff ();
    vevent::StopProcessing();
}

/*
 *  AboutItem::vmenuItem
 *
 *  display version and client status of program
 *
 */

static char aboutmsg [1500];

void AboutItem::ObserveMenuItem(vevent *event)
{
    strcpy (aboutmsg, "Version 1.0");
    strcat (aboutmsg, "\n");
    strcat (aboutmsg, GetClientInfo ());
    infoNotice (maindialog, (unsigned char *)"About TIS4000 Control Center", (unsigned char *)aboutmsg);
}

extern "C"
{
extern void privService (void * param); // in privService.c
}

int main(int argc, char *argv[])
{
    unsigned long  privThread;

    vstartup(argc, argv);

    // create thread to handle interface to PrivTrack Server

    privThread = _beginthread(privService, 0, NULL);

    // initialize user login stack
    userStack [0] = 0;
    curUserLevel = 0;

    // Get the resource file and load the application dialog
    vresource res = vapplication::GetCurrent()->GetResources();
    vresource appDlgRes = vresourceGet(res,
			       vnameInternGlobalLiteral("AppDialog"));
    maindialog = new ApplicationDialog(appDlgRes);

    // Get the resource file and load the password dialog
    vresource pwres = vresourceGet(res,
			       vnameInternGlobalLiteral("LogDialog"));
    pwdialog = new PasswordDialog(pwres);

    // Get the resource file and load the text entry dialog
    vresource textDlgRes = vresourceGet(res,
			       vnameInternGlobalLiteral("TextDialogTag"));
    textdialog = new TextDialog(textDlgRes);

    // Open the dialog and start processing
    maindialog->Open();
    vevent::Process();

    // clean up
    delete maindialog;
    delete pwdialog;
    delete textdialog;

    exit(EXIT_SUCCESS);
    return EXIT_FAILURE;
}

vkindTYPED_FULL_DEFINITION(PasswordText, vtextitemText, "myTextDataClass");

/*
 * PasswordText::InsertText
 *
 * Implementation for the vtext::InsertText method.  
 */
void PasswordText::InsertText(vtextSelection *selection, const vchar *text, long length)
{
    vstr    *str, *scan;
    int	    i, start;

    scan = str = vstrCreateSized(length);
    i = length;
    while (i-- > 0)
	*(scan++) = '*';

    // Send-super to insert the "*"'s into the displayed text item
    start = this->GetSelectionStart(selection);
    vtext::InsertText(selection, str, length);

    // Insert the text into our "real" text object
    this->realText->SelectRange(this->realSelection, start, start);
    this->realText->InsertText(this->realSelection, text, length);

    vstrDestroy(str);
}

/*
 * PasswordText::DeleteText
 *
 * Implementation for the vtext::DeleteText method.  
 */
void PasswordText::DeleteText(vtextSelection	*selection)
{
    // Delete the text in our "real" text object
    this->realText->SelectRange(this->realSelection,
			      this->GetSelectionStart(selection),
			      this->GetSelectionEnd(selection));
    this->realText->DeleteText(this->realSelection);
    vtext::DeleteText(selection);
}

/*
 * PasswordText::Destroy
 *
 * Destroy the auxillary text object stored with our PasswordText 
 * text data object
 */
void PasswordText::Destroy()
{
    delete this->realText;
    vtext::Destroy();
}

vkindTYPED_FULL_DEFINITION(PasswordTextitem, vtextitem, NULL);

/*
 * PasswordTextitem::ObserveView
 *
 * Catch when the user hits return or loses focus, and display
 * the password.  This shows how to retrieve the password from 
 * the auxillary storage
 */
void PasswordTextitem::ObserveView(vevent *event, int message)
{
    if (message != vtextviewACCEPT_NOTIFY)
	return;
    // Can add focus change processing here
}


vkindTYPED_FULL_DEFINITION(PasswordDialog, vdialog, NULL);

/*********
 *  PasswordDialog::LoadInit()
 *
 *  This method gets called whenever the dialog is loaded from a resource.
 *  Match our vtextitem subclass with the textitem contained in the resource,
 *  and initialize the text data subclass used by PasswordTextitem.
 *
 **********/
void PasswordDialog::LoadInit(vresource resource)
{
    PasswordText *text;
    vtextitem *oplabel;
    char    labeltext [100];
    unsigned long labellength = 100;

    // make sure that the textitem is loaded as a PasswordTextitem
    textitem = PasswordTextitem::RequestSubLoad(this, vname_Text);

    loginbutton = LoginButton::RequestSubLoad(this,
					    vnameInternGlobalLiteral("LoginTag"));

    cancelbutton = CancelButton::RequestSubLoad(this,
					    vnameInternGlobalLiteral("CancelTag"));

    vdialog::LoadInit(resource);

    oplabel = (vtextitem *)FindItem(vnameInternGlobalLiteral("OpNameTag"));
#ifdef WIN32
    if (GetUserName (labeltext, &labellength))
#else
xxxx add code for other platforms
#endif WIN32
	oplabel->SetTitleScribed (vcharScribeSystem(labeltext));
	oplabel->SetText ((unsigned char *)labeltext);

    // Create the "realText" text data object, this is where the actual
    // text will go
    text = (PasswordText *) textitem->GetTextData();
    text->realText      = new vtext;
    text->realSelection = text->realText->CreateSelection();

    SetPopup (TRUE);
}


/*********
 *  PasswordDialog::ObserveWindow()
 *
 *  The dialog notification procedure gets called whenever any 
 *  major event applies to the dialog as a whole.  We need to
 *  look for the event which tells us that the dialog has
 *  closed.
 *
 **********/
void PasswordDialog::ObserveWindow (vevent *event)
{
    if (event->GetType() == veventWINDOW_STATE && event->IsClose ())
        Close ();
}

vkindTYPED_FULL_DEFINITION(ApplicationDialog, vdialog, NULL);

/*********
 *  ApplicationDialog::LoadInit()
 *
 *  This method gets called whenever the dialog is loaded from a resource.
 *  Match our vtextitem subclass with the textitem contained in the resource,
 *  and initialize the text data subclass used by PasswordTextitem.
 *
 **********/
void ApplicationDialog::LoadInit(vresource resource)
{
    AppMenuItem *appItem;
    char    *apptext;
    char    command [256];

    vmenubar *menubar = vmenubar::RequestSubLoad (this, vnameInternGlobalLiteral("MenuBar"));
    vmenu *filemenu = vmenu::RequestSubLoad (menubar, vnameInternGlobalLiteral("FileMenuTag"));
    HostItem::RequestSubLoad(filemenu, vnameInternGlobalLiteral("HostTag"));
    LoginItem::RequestSubLoad(filemenu, vnameInternGlobalLiteral("LogInTag"));
    LogoutItem::RequestSubLoad(filemenu, vnameInternGlobalLiteral("LogOutTag"));
    ExitItem::RequestSubLoad(filemenu, vnameInternGlobalLiteral("ExitTag"));
    vmenu *helpmenu = vmenu::RequestSubLoad (menubar, vnameInternGlobalLiteral("HelpMenuTag"));
    AboutItem::RequestSubLoad(helpmenu, vnameInternGlobalLiteral("AboutTag"));

    vdialog::LoadInit(resource);

    // initializations
    fp = NULL;

    // the app menu
    appmenu = (vmenu *)menubar->FindMenu (vnameInternGlobalLiteral("AppMenuTag"));

    ResetAppRead ();
    while (appmenu && (apptext = GetNextApp (command)))
    {
	appItem = new AppMenuItem;
	appItem->SetAppCommand (command);
	appItem->SetTitleScribed(vcharScribeSystem(apptext));
	/* disable or enable this app entry according to permissions */
	appmenu->AppendItem(appItem);
    }
    // start out with apps disabled!
    appmenu->Disable ();

    SetTitle ((unsigned char *)"TIS4000 Control Center (Please log in)");
}


/*********
 *  ApplicationDialog::ObserveWindow()
 *
 *  The dialog notification procedure gets called whenever any 
 *  major event applies to the dialog as a whole.  We need to
 *  look for the event which tells us that the dialog has
 *  closed.
 *
 **********/
void ApplicationDialog::ObserveWindow (vevent *event)
{
    if (event->GetType() == veventWINDOW_STATE && event->IsClose ())
	vevent::StopProcessing ();
}

/*********
 *  ApplicationDialog::ResetAppRead()
 *
 *  Resets the position in the ap menu specification file.
 *
 **********/
void ApplicationDialog::ResetAppRead ()
{
    if (fp)
	fclose (fp);

    /* look up requested service, exit if not found */
    fp = fopen ("/tis4000db/tis4000.menu", "rt");
}

/*********
 *  ApplicationDialog::GetNextAppw()
 *
 *  Reads the next app menu item.
 *
 **********/
char *ApplicationDialog::GetNextApp (char *command)
{
    int ix;
    char *sp;
    char *cp;

    if (fp)
    {
        while (!feof (fp))
        {
	    ix = 0;
	    while (!feof (fp))
	    {
		inpstring [ix] = fgetc(fp);
		if ((inpstring [ix] == 0x0d) || (inpstring [ix] == 0x0a) 
		  || (inpstring [ix] == -1))
		{
		    break;
 		}
		ix++;
 		if (ix == 80)
		{
		    inpstring[79] = 0;
		}
	    }
	    inpstring [ix] = 0;

	    /* process input line */
	    sp = inpstring;
	    /* skip blanks */
	    while (*sp && (*sp == ' ' || *sp == 0x09))
		sp++;
	    if ((*sp == 0) || (*sp == '#'))
		continue; /* blank line or comment! */
	    /* process input- line is formatted as
	    ** <menu text> <executable command>
	    */
	    if (*sp == '"')
		cp = strchr (++sp, '"');
	    else
		cp = strchr (sp, ' ');
	    if (cp)
	    {
		*cp = 0;
		strncpy (menutext, sp, 40);
		sp = cp + 1;
		while (*sp && (*sp == ' ' || *sp == 0x09))
		    sp++;
		strncpy (command, sp, 255);
		return menutext;
	    }
	}
    }
    if (fp)
	fclose(fp);
    fp = NULL;
    return NULL;
}
