/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/************************************************************************
*
*    alarmTalk.c
*
*    modifications
*    -------------
*  .01	09-22-96	ems	creation date
*
************************************************************************/


#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <sys/termios.h>
#include <sys/ttold.h>
#include <netinet/in.h>
#include <errno.h>
#include <netdb.h>
#include <fcntl.h>
#include <signal.h>
#include <malloc.h>
#ifndef AXP
#include <net/if.h>
#include <sys/sockio.h>
#include <sys/filio.h>
#endif

#ifdef MIPSEL
#include "error.h"
#endif

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "dbDefs.h"		/* .../share/epicsH */
#include "shareMem.h"		/* .../share/src/alarm */
#include "alarmLib.h"		/* .../share/src/alarm */
#include "cadef.h"		/* .../share/epicsH */
#include "messages.h"		/* .../share/src/alarm */
#include "alarmTalkdb.h"	/* .../share/src/dataVu */
#include "alarmDefs.h"		/* .../share/src/alarm */
#include "alarmvu.h"		/* .../share/src/alarmVuCC */
#include "slmDefs.h"		/* .../share/epicsH */
#include "OperPriv.h"		/* .../share/src/libtis */
#include "tsDefs.h"		/* .../share/epicsH, TS_STAMP_def */
#include "msgMon.h"		/* .../share/epicsH */
#include "tisEncrypt.h"		/* .../shar/epicsH */
#include "alarmTalk.h"

extern ShareHeader *shareHeader;	/* .../share/src/alarm/utils/alarmLib.c */


/*****************************************************************************
*    
*    Static variables
*
******************************************************************************/

int	alarmPrtListenFd = -1;	/* FD to listen for alarmPrinter on */
int	alarmPrtFd = -1;  /* FD for incoming messages from the alarmPrinter */
int	alarmMgrFd = -1;  /* FD for incoming/outgoing messages to alarmMgr */
int	decTalkFd = -1;	  /* FD for incoming/outgoing messages to the decTalk */

/* buffer for alarm msgs from alarmPrtFd */
char	newAlarmMsg[ALARM_MSG_SIZE + 1];

/* buffer to read decTalk msgs in */
char	decTalkRecvBuffer[DECTALK_MSG_SIZE + 1]; 

/* buffer for current decTalk msg */
char	newDecTalkMsg[DECTALK_MSG_SIZE + 1];	

char	decTalkName[64];	/* name of decTalk device */
char	opName[SLM_NAME_SIZE + 1];	/* operator name from secureList */
char	opId[SLM_KEYPADID_SIZE + 1];	/* keypad id entered */
int	alarmAccessed = 0;	/* did we connect to alarmMgr's shared memory */
int	caInited = 0;		/* was channel access init'd */
int	dbOpened = 0;		/* was alarmTalk.dbf opened */
int	settty = 0;		/* flag to reset tty flags */
int	debug = 0;		/* debug switch */
int	mgrAvailable = 0;	/* alarm mgr is available, I can access his
					   shared memory */
int	alarmMsgCnt = 0;	/* ptr into newAlarmMsg[] */
int	decTalkMsgCnt = 0;	/* ptr into decTalkRecvBuffer[] */
int	currentState = 0;	/* state we are in for processing an 
					   alarm msg
					*/
int	timeToAnnunciate = 0;	/* time in seconds to wait for an alarm to be
					   ack'd before we will annunciate it
					*/
int	autoTimeoutCnt = 0;	/* cnt # requests for auto-caller passwd */
int	autoIdCnt = 0;		/* cnt # requests for good password */
int	phoneCallFailed = 0;
int	validPhoneNumber = 0;	/* there is a number to call */
int	autoNumber = 0;		/* which auto number are we dialing */
int	whoWasCalled = NO_CALL_YET;	/* console, page auto */
int	autoState = NO_AUTO_STATE;	/* where are we in auto call sequence */
int	opIdCnt = -1;		/* index into opId[] */
int	firstDbOpen = 1;	/* used to init codebase only once */
int	decWriteFailed = 0;	/* set by writeDecTalk () */
short	decTalkPort;		/* # if decTal is on network */
int	sentToAutoCaller = 0;

struct sockaddr_in alPrtPort;	/* for alarmPrinter */
struct termios	savetty;	/* save tty setting from initDecTalk () */
time_t		deadline;	/* deadline for something to happen from the decTalk */
time_t		configLastModified;	/* last time alarmtlk.dbf was modified */
time_t		secureLastModified;	/* last time secureList was modified */

QUEMSG_PTR	firstActive = (QUEMSG_PTR)0;	/* active alarm msg list */
QUEMSG_PTR	currentAlarm = (QUEMSG_PTR)0;	/* ptr to current alarm msg */
QUEMSG_PTR	saveCurrentAlarm = (QUEMSG_PTR)0;/* save currentAlarm */
QUEMSG_PTR	firstUnAck = (QUEMSG_PTR)0;	/* first unack'd point */
ALARMPT_PTR	alarmTalkPts = (ALARMPT_PTR)0;	/* list of decTalk alarm pts */
ALARMCFG_PTR	currentAlarmConfig;		/* ptr into alarmConfigs[] */
ALARMCFG		alarmConfigs[MAX_ALARM_CFGS];
ACCESSLIST_PTR	accessTree = (ACCESSLIST_PTR)0;

CODE4	cb;			/* codebase */
DATA4	*alarmTalkDb;		/* return from codebase d4open() */

static char	*WhoWasCalled[] = {
		"NO CALL YET",
		"CONSOLE CALL",
		"PAGE CALL",
		"AUTO CALL",
		0 
		};

static char	*states [] = {
		"GET NEXT ALARM",
		"WHO TO CALL",
		"MAKE A CALL",
		"WAIT FOR ANS",
		"WAIT TRY AGAIN",
		"WAIT REPLY",
		"WAIT HANGUP",
		"WAIT AUTO",
		"SEND UNACK",
		"AUTO SENT",
		"AUTO REPEAT",
		"WAIT ACK REPEAT",
		"ACK PTS",
		"WAIT MGR",
		0 
		};

static char	*autoStates [] = {
		"NO AUTO STATE",
		"ID MSG",
		"BLD ACK",
		0 
		};


/*****************************************************************************/

main (int argc, char *argv[])
{
    time_t		start_time;
    fd_set		chanmask;
    int			doingIt = 1;
    int			maxFd = -1;
    int			status;
    int			i;
    struct timeval	selectWait;	/* timeout setting for select() */

    tisOpenLog ("alarmTalk", 6, 2, NULL, NULL);

    start_time = time (0);

    tisSysLog (0,0, "started at %s", ctime (&start_time));

    /* set the DISPLAY environment variable
       to be used when checking on operator
       privileges
    */

    if ((putenv ("DISPLAY=DECTALK:0")) != 0)
    {
	tisSysLog (0,0, "main: putenv failed (%m)\n");
	return (0);
    }

    if (setDecTalkName (argc, argv))
	return (0);

    initSignals ();

    while (1)
    {
	if ((initDecTalk ()) == 0)
	    break;
	tisSysLog (0,0, "main: could not init the decTalk, is it on ?");
	if (decTalkFd > -1)
	{
	    close (decTalkFd);
	    decTalkFd = -1;
	}
	setDecTalkName (argc, argv);
    }

    if (initDecTalkList ())
    {
	wrapItUp ();
	return (0);
    }

    if (initDecTalkPhoneticList ())
    {
	wrapItUp ();
	return (0);
    }

    if (openAlarmTalkDbf ())
    {
	wrapItUp ();
	return (0);
    }

    if (initAccessList ())
    {
	wrapItUp ();
	return (0);
    }

    if (createAlarmPrtSocket ())
    {
	wrapItUp ();
	return (0);
    }

    if (connectToAlarmMgr ())
    {
	wrapItUp ();
	return(0);
    }

    /* now init alarmMgr's shared memory */

    if (startAlarmAccess ())
    {
	wrapItUp ();
	return (0);
    }

    /* finally start chanel access */

    if (startChannelAccess ())
    {
	wrapItUp ();
	return (0);
    }

    tisSysLog (0,0, "alarmTalk is initialized");

    deadline = time (0) + TIMEOUT_DEADBAND;

    while (doingIt)
    {
	/* may have been closed in ackPoints due to send error */
	if (alarmMgrFd == -1)
	{
	    if (alarmAccessed)
	    {
		UnmapSharedMemory ();
		alarmAccessed = 0;
	    }
	    mgrAvailable = 0;
	    if (connectToAlarmMgr ())
	    {
		doingIt = 0;
		break;
	    }
	}
	/* set in writedecTalk */
	if (decWriteFailed)
	{
	    if (settty)
	    {
		tcsetattr (decTalkFd, TCSADRAIN, &savetty);
		ioctl (decTalkFd, TIOCNXCL, &i);
	    }
	    close (decTalkFd);
	    FD_CLR(decTalkFd, &chanmask);
	    decTalkFd = -1;
	    decTalkMsgCnt = 0;
	    memset ((char *)&decTalkRecvBuffer[0], 0, sizeof(decTalkRecvBuffer));
	    memset ((char *)&newDecTalkMsg[0], 0, sizeof (newDecTalkMsg));
	    currentState = GET_NEXT_ALARM_MSG;
	    while (1)
	    {
		setDecTalkName (argc, argv);
		if ((initDecTalk ()) == 0)
		    break;
		tisSysLog (0,0, "main: write failed to decTalk, could not reinit decTalk");
		if (decTalkFd > -1)
		{
		    close (decTalkFd);
		    decTalkFd = -1;
		}
	    }
	    memset ((char *)&decTalkRecvBuffer[0], 0, sizeof(decTalkRecvBuffer));
	    memset ((char *)&newDecTalkMsg[0], 0, sizeof (newDecTalkMsg));
	}
	for (;;)
	{
	    if ((deadline - (time(0))) <= 0)
	    {
		/* decTalk may be dead */
		if (decTalkState ())
		{
		    writeDecTalk (DTHANGUP, "main");
		    if (settty)
		    {
			tcsetattr (decTalkFd, TCSADRAIN, &savetty);
			ioctl (decTalkFd, TIOCNXCL, &i);
		    }
		    close (decTalkFd);
		    FD_CLR(decTalkFd, &chanmask);
		    decTalkFd = -1;
		    decTalkMsgCnt = 0;
		    memset ((char *)&decTalkRecvBuffer[0], 0, sizeof(decTalkRecvBuffer));
		    memset ((char *)&newDecTalkMsg[0], 0, sizeof (newDecTalkMsg));
		    currentState = GET_NEXT_ALARM_MSG;
		    /* if we are currently working on an alarm try to
		       reprocess it
		    */
		    if (saveCurrentAlarm != (QUEMSG_PTR)0)
		    {
			saveCurrentAlarm -> removeFromList = 0;
		    }
		    while (1)
		    {
			setDecTalkName (argc, argv);
			if ((initDecTalk ()) == 0)
			    break;
			tisSysLog (0,0, "main: bad timeout state, could not reinit decTalk");
			if (decTalkFd > -1)
			{
			    close (decTalkFd);
			    decTalkFd = -1;
			}
		    }
		    memset ((char *)&decTalkRecvBuffer[0], 0, sizeof(decTalkRecvBuffer));
		    memset ((char *)&newDecTalkMsg[0], 0, sizeof (newDecTalkMsg));
		}
		else
		    deadline = time (0) + TIMEOUT_DEADBAND;
	    }
	    selectWait.tv_sec = deadline - (time (0));
	    selectWait.tv_usec = 0;
	    maxFd = -1;
	    FD_ZERO(&chanmask);
	    if (decTalkFd != -1)
		FD_SET(decTalkFd, &chanmask);
	    maxFd = decTalkFd;
	    if (alarmPrtFd != -1)
	    {
		FD_SET(alarmPrtFd, &chanmask);
		if (alarmPrtFd > maxFd)
		    maxFd = alarmPrtFd;
	    }
	    if (alarmMgrFd != -1)
	    {
		FD_SET(alarmMgrFd, &chanmask);
		if (alarmMgrFd > maxFd)
		    maxFd = alarmMgrFd;
	    }
	    if (alarmPrtListenFd != -1)
	    {
		FD_SET(alarmPrtListenFd, &chanmask);
		if (alarmPrtListenFd > maxFd)
		    maxFd = alarmPrtListenFd;
	    }

if(debug)
{
tisSysLog(0,0,
"main, select timeout = %d",
selectWait.tv_sec);
}
	    status = select (maxFd + 1, &chanmask, NULL, NULL, &selectWait);

	    if (status == -1)
	    {
if (debug)
{
tisSysLog (0,0,
"main: select error (%m)");
}
		break;
	    }
	    /* timeout from select */
	    if (status == 0)
	    {
if(debug)
{
tisSysLog (0,0,
"main:status = 0");
}
		/* we have timed out, make sure we are not waiting on the decTalk
		   it may be dead, try to restart it
		*/
		if (decTalkState ())
		{
		    writeDecTalk (DTHANGUP, "main");
		    if (settty)
		    {
			tcsetattr (decTalkFd, TCSADRAIN, &savetty);
			ioctl (decTalkFd, TIOCNXCL, &i);
		    }
		    close (decTalkFd);
		    FD_CLR(decTalkFd, &chanmask);
		    decTalkFd = -1;
		    decTalkMsgCnt = 0;
		    memset ((char *)&decTalkRecvBuffer[0], 0, sizeof(decTalkRecvBuffer));
		    memset ((char *)&newDecTalkMsg[0], 0, sizeof (newDecTalkMsg));
		    currentState = GET_NEXT_ALARM_MSG;
		    /* if we are currently working on an alarm try to
		       reprocess it
		    */
		    if (saveCurrentAlarm != (QUEMSG_PTR)0)
		    {
			saveCurrentAlarm -> removeFromList = 0;
		    }
		    while (1)
		    {
			setDecTalkName (argc, argv);
			if ((initDecTalk ()) == 0)
			    break;
			tisSysLog (0,0, "main: timeout state, could not reinit decTalk");
			if (decTalkFd > -1)
			{
			    close (decTalkFd);
			    decTalkFd = -1;
			}
		    }
		    memset ((char *)&decTalkRecvBuffer[0], 0, sizeof(decTalkRecvBuffer));
		    memset ((char *)&newDecTalkMsg[0], 0, sizeof (newDecTalkMsg));
		}
		else
		{
		    deadline = time (0) + TIMEOUT_DEADBAND;
		    break;
		}
	    }
	    /* we are listening for the alarmPrinter to contact us,
	       do an accept so we can read the msgs from the alarmPrinter
	    */
	    if ( (alarmPrtListenFd != -1) && 
		 (FD_ISSET(alarmPrtListenFd, &chanmask)) )
	    {
if(debug)
{
tisSysLog(0,0,
"main, select, listen fired");
}
		if (acceptAlarmPrinter ())
		{
		    doingIt = 0;
		    break;
		}
	    }
	    if ( (alarmPrtFd != -1) &&
		 (FD_ISSET(alarmPrtFd, &chanmask)) )
	    {
if(debug)
{
tisSysLog(0,0,
"main: prt fired");
}
		/* if we loose the connection to the alarmPrinter
		   close the fd, clear it from the mask and trash
		   what we have in the newAlarmMsg buffer, since we
		   can not be guarented of its contents or what
		   we will receive if we can reestablish a connection
		   to the alarmPrinter
		*/
		if (readAlarmMsg ())
		{
		    close (alarmPrtFd);
		    FD_CLR(alarmPrtFd, &chanmask);
		    alarmPrtFd = -1;
		    alarmMsgCnt = 0;
		    memset (newAlarmMsg, 0, sizeof (newAlarmMsg));
		}
	    }
	    if ( (alarmMgrFd != -1) &&
		 (FD_ISSET(alarmMgrFd, &chanmask)) )
	    {
		/* if we loose connection to the alarm mgr, close the fd,
		   clear it from the mask, unmap the shared memory and
		   try to reconnect to alarm mgr and access the shared
		   memory
		*/
		if (readAlarmMgr ())
		{
		    close (alarmMgrFd);
		    FD_CLR(alarmMgrFd, &chanmask);
		    alarmMgrFd = -1;
		    if (alarmAccessed)
		    {
			UnmapSharedMemory ();
			alarmAccessed = 0;
		    }
		    mgrAvailable = 0;
		    if (connectToAlarmMgr ())
		    {
			doingIt = 0;
			break;
		    }
		    if (startAlarmAccess ())
		    {
			doingIt = 0;
			break;
		    }
		}
	    }
	    if ( (decTalkFd != -1) &&
		 (FD_ISSET(decTalkFd, &chanmask)) )
	    {
if(debug)
{
tisSysLog(0,0,
"main: dec fired");
}
		/* if we loose the connection to the decTalk, reset the
		   ports attributes and  turn off exclusive if we opened
		   the decTalk through a serial port, close the
		   fd and clear it from the mask, try to reconnect
		   to the decTalk
		*/
		if (readDecTalk ())
		{
		    if (settty)
		    {
			tcsetattr (decTalkFd, TCSADRAIN, &savetty);
			ioctl (decTalkFd, TIOCNXCL, &i);
		    }
		    close (decTalkFd);
		    FD_CLR(decTalkFd, &chanmask);
		    decTalkFd = -1;
		    decTalkMsgCnt = 0;
		    memset ((char *)&decTalkRecvBuffer[0], 0, 
							sizeof(decTalkRecvBuffer));
		    memset ((char *)&newDecTalkMsg[0], 0, sizeof (newDecTalkMsg));
		    currentState = GET_NEXT_ALARM_MSG;
		    /* if we are currently working on an alarm try to
		       reprocess it
		    */
		    if (saveCurrentAlarm != (QUEMSG_PTR)0)
		    {
			saveCurrentAlarm -> removeFromList = 0;
		    }
		    while (1)
		    {
			setDecTalkName (argc, argv);
			if ((initDecTalk ()) == 0)
			    break;
			tisSysLog (0,0,"main: decTalk read failure, could not reinit decTalk");
			if (decTalkFd > -1)
			{
			    close (decTalkFd);
			    decTalkFd = -1;
			}
		    }
		}
		break;
	    }
	    else
	    {
		/* if the current time is > than the deadline, check to see if
		   we are waiting for the decTalk, if we are something is 
		   wrong so try to reinit it
		*/
		if ((time(0)) > deadline)
		{
		    if (decTalkState ())
		    {
			writeDecTalk (DTHANGUP, "main");
			if (settty)
			{
			    tcsetattr (decTalkFd, TCSADRAIN, &savetty);
			    ioctl (decTalkFd, TIOCNXCL, &i);
			}
		        close (decTalkFd);
		        FD_CLR(decTalkFd, &chanmask);
		        decTalkFd = -1;
		        decTalkMsgCnt = 0;
		        memset ((char *)&decTalkRecvBuffer[0], 0, sizeof(decTalkRecvBuffer));
		        memset ((char *)&newDecTalkMsg[0], 0, sizeof (newDecTalkMsg));
		        currentState = GET_NEXT_ALARM_MSG;
		        /* if we are currently working on an alarm try to
		           reprocess it
		        */
		        if (saveCurrentAlarm != (QUEMSG_PTR)0)
		        {
			    saveCurrentAlarm -> removeFromList = 0;
		        }
			while (1)
			{
			    setDecTalkName (argc, argv);
			    if ((initDecTalk ()) == 0)
				break;
			    tisSysLog (0,0, "main: no dec read, bad timeout state, could not reinit decTalk");
			    if (decTalkFd > -1)
			    {
				close (decTalkFd);
				decTalkFd = -1;
			    }
			}
		        memset ((char *)&decTalkRecvBuffer[0], 0, sizeof(decTalkRecvBuffer));
		        memset ((char *)&newDecTalkMsg[0], 0, sizeof (newDecTalkMsg));
		    }
		    else
		    {
			deadline = time (0) + TIMEOUT_DEADBAND;
			continue;
		    }
		}
	    }
	}  /* end for (;;) */

	if (!doingIt)
	    break;

	processAlarms ();

if (debug)
{
tisSysLog (0,0,
"main: ret from processAlarms");
}
	/* remove any messages from the list that have been
	   flagged for delete
	*/
	if (firstActive != (QUEMSG_PTR)0)
	{
	    removeMsgFromList (&firstActive);

if (debug)
{
tisSysLog (0,0,
"main, ret from removeMsgFromList");
}
	}

    }  /* end while (doingIt) */

    wrapItUp ();
    return (0);
}


/**************************************************
*
*    ackPoints ()
*	send ack request to alarm manager for
*	all unack'd points annunciated in
*	auto-caller mode
*
**************************************************/

int	ackPoints (void)
{
    QUEMSG_PTR		nextMsg;
    AlarmMsg		ackMsg;
    PvData		*pvdata;
    int			bytesSent;

    nextMsg = firstUnAck;
    while (nextMsg != (QUEMSG_PTR)0)
    {
	memset ((char *)&ackMsg, 0, sizeof (ackMsg));
	ackMsg.msgType=  ACK_MSG;
	strncpy (ackMsg.pointName, nextMsg -> pointName, PVNAME_SZ);
	ackMsg.pointName[PVNAME_SZ] = NULL;
	strncpy (ackMsg.name.operatorName, opName, AL_OPERATOR_SIZE);
	ackMsg.name.operatorName[AL_OPERATOR_SIZE] = NULL;
	if (pvdata = (PvData *)FindPV (NULL, nextMsg -> pointName))
	{
	    ackMsg.timeStamp = pvdata -> lastAlarm;
	    bytesSent = send (alarmMgrFd, (char *)&ackMsg, sizeof (ackMsg), 0);
if(debug)
{
tisSysLog(0,0,
"acked %s, bytesSent=%d",
ackMsg.pointName,bytesSent);
}
	    if (bytesSent == -1)
	    {
		tisSysLog (0,0, "ackPoints: send to mgr fail (%m)");
		close (alarmMgrFd);
		alarmMgrFd = -1;
		return (1);
	    }
	 }
	nextMsg = nextMsg -> nextMsg;
    }
    return (0);
}

/***********************************************
*
*    addMsgToActiveList ()
*	caller from addNewAlarmPoint ()
*	add new alarmmsg to active alarm list
*
***********************************************/

void	addMsgToActiveList (QUEMSG_PTR newMsg)
{
    QUEMSG_PTR		currentMsg, prevMsg;

    if (firstActive == (QUEMSG_PTR)0)
    {
	firstActive = newMsg;
    }
    else
    {
	prevMsg = currentMsg = firstActive;
	while (currentMsg != (QUEMSG_PTR)0)
	{
	    prevMsg = currentMsg;
	    currentMsg = currentMsg -> nextMsg;
	}
	prevMsg -> nextMsg = newMsg;
    }
if(debug)
{
tisSysLog(0,0,
"addMsgAct, pt=%s",
newMsg->pointName);
}
    return;
}


/*****************************************************
*
*    addNewAlarmPoint ()
*
*	called from readAlarmMsg ()
*	add a new alarmpoint to the active alarm 
*	point list firstActive
*
*****************************************************/

void	addNewAlarmPoint (void)
{
    int		foundPoint = 0;
    int		samePoint = 0;
    char	*pChar;
    char	pointName[AL_POINTNAME_SIZE + 1];
    ALARMPT_PTR	currentPoint;
    QUEMSG_PTR	newMsg;
    QUEMSG_PTR	currMsg;

    strncpy (pointName, newAlarmMsg + (AL_DATE_SIZE + 2 + AL_TIME_SIZE + 2),
	     AL_POINTNAME_SIZE);
    pointName[AL_POINTNAME_SIZE] = 0;

    /* remove trailing spaces */
    pChar = (char *)&pointName[0];
    while ( ((*pChar) != ' ') && ((*pChar) != NULL) )
	pChar++;
    *pChar = NULL;

    /* see if this pt is in the list of pts we are to report
       that is if the list exists
    */
    if (alarmTalkPts != (ALARMPT_PTR)0)
    {
	currentPoint = alarmTalkPts;
	while (currentPoint != (ALARMPT_PTR)0)
	{
	    if ((strncmp (pointName, currentPoint -> pointName,
			 PVNAME_SZ)) == 0)
	    {
		foundPoint = 1;
		break;
	    }
	    currentPoint = currentPoint -> nextPoint;
	}
	/* point was not in our list, ignore this alarm */
	if (foundPoint == 0)
	    return;
    }
    if ( (newMsg = getMsgMemory ()) == (QUEMSG_PTR)0)
    {
	tisSysLog (0,0, "addNewAlarmPoint: point %s will not be processed", pointName);
    }
    newMsg -> nextMsg = (QUEMSG_PTR)0;
    strncpy (newMsg -> pointName, pointName, PVNAME_SZ);
    newMsg -> pointName[PVNAME_SZ] = NULL;
    if ( (findPvName (newMsg)) != 0)
    {
	tisSysLog (0,0, "addNewAlarmPoint: point %s will not be processed", pointName);
	newMsg -> removeFromList = 1;
    }
    /* time the alarm was received */
    newMsg -> alarmRecvd = time (0);

    currMsg = firstActive;
    while (currMsg != (QUEMSG_PTR)0)
    {
	if ((strcmp (currMsg -> pointName, newMsg -> pointName)) == 0)
	{
	    if ( (currMsg -> alarmSeverity == newMsg -> alarmSeverity) &&
		 (currMsg -> alarmStatus == newMsg -> alarmStatus) &&
		 (currMsg -> lastAlarm.secPastEpoch == newMsg -> lastAlarm.secPastEpoch) &&
		 (currMsg -> lastAlarm.nsec == newMsg -> lastAlarm.nsec) )
	    {
		samePoint = 1;
		break;
	    }
	}
	currMsg = currMsg -> nextMsg;
    }

    /* if alarm has not already been ack'd, or findPvName found it, add it */
    if ((!(newMsg -> ack)) && ((!(newMsg -> removeFromList)))  && (!samePoint))
	addMsgToActiveList (newMsg);
    else
	free ((char *)newMsg);

    return;
}


/**************************************
*
*    alarmTalkSignalHandler ()
*	process SIGNAL interrupts
*
**************************************/

void	alarmTalkSignalHandler (int which_sig)
{
    time_t	end_time;
    int		do_exit = 0;

    end_time = time (0);

    switch (which_sig)
    {
	case SIGINT:
	    tisSysLog (0,0, "exit due to ^C at %s", ctime (&end_time));
	    do_exit = 1;
	    break;

	case SIGTERM:
	    tisSysLog (0,0, "exit due to term at %s", ctime (&end_time));
	    do_exit = 1;
	    break;

	case SIGUSR1:
	    tisSysLog (0,0, "debug on");
	    debug = 1;
	    break;

	case SIGUSR2:
	    tisSysLog (0,0, "debug off");
	    debug = 0;
	    break;

	default:
	    break;
    }

    if (do_exit)
    {
	wrapItUp ();
	exit (0);
    }
    return;
}

/*************************************************************
*
*    buildMsg ()
*	given ptrs PvData and QUEMSG_PTR, fill the msg with
*	data from PvData
*
*************************************************************/

void	buildMsg (QUEMSG_PTR newMsg, PvData *pvdata)
{
    newMsg -> alarmSeverity = pvdata -> unAckedSeverity;
    newMsg -> alarmStatus = pvdata -> unAckedStatus;
    newMsg -> ack = pvdata -> Ack;
    strncpy (newMsg -> descriptor, pvdata -> description, DESCRIPTOR_SZ);
    newMsg -> descriptor[DESCRIPTOR_SZ] = NULL;
    strncpy (newMsg -> value, pvdata -> unAckedValue, AL_VALUE_SIZE);
    newMsg -> value[AL_VALUE_SIZE] = NULL;
    strncpy (newMsg -> egutag, pvdata -> eguTag, AL_EGUTAG_SIZE);
    newMsg -> egutag[AL_EGUTAG_SIZE] = NULL;
    newMsg -> lastAlarm = pvdata -> lastAlarm;
    newMsg -> mgrNotAvailable = 0;

    return;
}

/************************************************************
*
*    buildUnAckList ()
*	build a list of all unack'd alarms that
*	the operator (opName) has permission to ACK
*	the summary list is a list of all points that
*	are in alarm and not ack'd or was in alarm
*	and not ack'd
*	InitOperPriv() in .../share/src/libtis/OperPrvTrk.c
*	CheckOperPriv() and CloseOperPriv() in same
*
************************************************************/

int	buildUnAckList (void)
{
    char	buff[SLM_NAME_SIZE];
    char	pointname[PVNAME_SZ+FLDNAME_SZ+1];
    PvData	*pvdata;
    FILE	*fd;
    QUEMSG_PTR	newMsg, prevMsg;

    /* if the alarm manager is not available */
    if (!mgrAvailable)
    {
	tisSysLog (0,0, "buildUnAckList: mgr not available");
	return (1);
    }

    /* set up the operator to be used by the operator 
       privilege functions
    */
#ifdef	WIN32
    if ((fd = fopen ("\\tis4000\\operators\\logins\\DECTALK", "w+")) == NULL)
#else
    if ((fd = fopen ("/tis4000/operators/logins/DECTALK", "w+")) == NULL)
#endif
    {
	tisSysLog (0,0, "buildUnAckList: open error (%m)");
	return (1);
    }
    fprintf (fd, "%s\n", opName);
    fclose (fd);

    if ((InitOperPriv (buff)) != PRIVOK)
    {
	tisSysLog (0,0, "buildUnAckList: init op privs failed");
	return (1);
    }

    pvdata = shareHeader -> summaryHead;

    while (pvdata != (PvData *)0)
    {
	if (pvdata -> Ack == 0)
	{
	    sprintf (pointname, "%s.ACK", pvdata -> pvName);
	    if ((CheckOperPriv (pointname)) == PRIVOK)
	    {
		if ( (newMsg = getMsgMemory ()) == (QUEMSG_PTR)0)
		{
		    tisSysLog (0,0, "buildUnAckList, list will not be built");
		    CloseOperPriv ();
		    return (1);
		}
		newMsg -> nextMsg = (QUEMSG_PTR)0;
		strncpy (newMsg -> pointName, pvdata -> pvName, PVNAME_SZ);
		newMsg -> pointName[PVNAME_SZ] = NULL;
		buildMsg (newMsg, pvdata);

		if (firstUnAck == (QUEMSG_PTR)0)
		{
		    firstUnAck = newMsg;
		}
		else
		{
		    prevMsg -> nextMsg = newMsg;
		}
		prevMsg = newMsg;
	    }
	}
	pvdata = pvdata -> nextSummaryPv;
    }
    CloseOperPriv ();

if (debug)
{
tisSysLog (0,0,
"buildUnAckList, ret 0");
}
    return (0);
}


/**********************************************************
*
*    convertDecTalkMsg ()
*	convert non-printable characters in a decTalk msg
*
***********************************************************/

char	*convertDecTalkMsg (char *msg)
{
    char	buffin[sizeof (newDecTalkMsg)];
    char	buffout[sizeof (newDecTalkMsg)];
    char	buff[32];
    char	*pbuffin;
    char	*pbuffout;

    memset (buffin, 0, sizeof (buffin));
    memset (buffout, 0, sizeof (buffout));
    strcpy (buffin, msg);
    pbuffin = (char *)&buffin[0];
    pbuffout = (char *)&buffout[0];
    while (*pbuffin)
    {
	if (*pbuffin == '\e')
	{
	    strcat (pbuffout, "ESC");
	    pbuffout += 3;
	}
	else if ( ((*pbuffin) < 040) || ((*pbuffin) > 0176) )
	{
	    sprintf (buff, "%d", *pbuffin);
	    strcat (pbuffout, buff);
	    pbuffout += strlen (buff);
	}
	else
	{
	    *pbuffout = *pbuffin;
	    pbuffout++;
	}
	pbuffin++;
    }
    pbuffout = (char *)&buffout[0];
    return (pbuffout);
}


/***********************************************
*
*    decTalkReset ()
*	do a soft reset of the decTalk
*
***********************************************/

int	decTalkReset (void)
{
    int			maxFd = -1;
    int			state = 0;
    int			status;
    char		*pChar;
    char		*pMsg;
    fd_set		chanmask;
    struct timeval	selectWait;

    /* do a soft terminal reset on the decTalk */
    if (writeDecTalk (DECSTR, "decTalkReset"))
    {
	return (1);
    }

    /* purge any possible winks or garbage comming from the decTalk */
    fflushDecTalkQue ();

    /* control how decTalk sends ESC and keypad sequences
       back to this program
		allows decTalk to send \r at end of its replies
		treat # and * as response terminators
    */
    if (writeDecTalk (DTMASK, "decTalkReset"))
    {
	return (1);
    }

    /* request device status request to be sure decTalk is working */
    if (writeDecTalk (DSR, "decTalkReset"))
    {
	return (1);
    }

    maxFd = decTalkFd;

    while (state >= 0)
    {
	FD_ZERO(&chanmask);
	FD_SET(decTalkFd, &chanmask);

	selectWait.tv_sec = 30;
	selectWait.tv_usec = 0;

	status = select (maxFd + 1, &chanmask, NULL, NULL, &selectWait);
	if (status == -1)
	{
	    tisSysLog (0,0, "decTalkReset: select error (%m)");
	    return (1);
	}
	if (status == 0)
	{
	    tisSysLog (0,0, "decTalkReset: select timeout, problem with decTalk");
	    return (1);
	}
	if (readDecTalk ())
	    return (1);
	while (state >= 0)
	{
	    getNextDecTalkMsg ();
	    /* either there are no more msgs in the buffer, or we did
	       not get a complete msg on the last read, in either case
	       we go back to the select to wait for more
	    */
	    if (newDecTalkMsg[0] == 0)
		break;
	    switch (state)
	    {
		case 0:		/* waiting for normal status from DSR cmd */
		    if ((strcmp (newDecTalkMsg, DSRNORMAL)) == 0)
		    {
			state = 1;
		    }
		    else if ((strcmp (newDecTalkMsg, DSRFAIL)) == 0)
		    {
			state = 3;
		    }
		    else
		    {
			tisSysLog (0,0, "decTalkReset: unexpected status return, (%s)", 
				convertDecTalkMsg (newDecTalkMsg));
			return (1);
		    }
		    break;

		case 1:		/* waiting for DSRFIRST or DSAFTER msg */
		    if ( ((strcmp (newDecTalkMsg, DSRFIRST)) == 0) ||
			 ((strcmp (newDecTalkMsg, DSRAFTER)) == 0) )
		    {
			/* send a device attribute request, decTalk ids itself */
			if (writeDecTalk (DA, "decTalkReset"))
			{
			    return (1);
			}
			state = 2;
		    }
		    else
		    {
			tisSysLog (0,0, "decTalkReset: unexpected 2nd status sequence, (%s)",
				convertDecTalkMsg (newDecTalkMsg));
			return (1);
		    }
		    break;

		case 2:		/* waiting for decTalk to ID itself */
		    if ((strcmp (newDecTalkMsg, DASTAT)) == 0)
		    {
			state = -1;
		    }
		    else
		    {
			tisSysLog (0,0, "decTalkReset: wrong decTalk id, (%s)",
				convertDecTalkMsg (newDecTalkMsg));
			return (1);
		    }
		    break;

		case 3:		/* reason for DSRFAIL return above */
		    pChar = convertDecTalkMsg (newDecTalkMsg);
		    if ((strcmp (newDecTalkMsg, DSRCOMMFAIL)) == 0)
			pMsg = DSRCOMMSG;
		    else if ((strcmp (newDecTalkMsg, DSROVERFLOW)) == 0)
			pMsg = DSROVERFLOWMSG;
		    else if ((strcmp (newDecTalkMsg, DSRNVRFAIL)) == 0)
			pMsg = DSRNVRMSG;
		    else if ((strcmp (newDecTalkMsg, DSRPHONEM)) == 0)
			pMsg = DSRPHONEMSG;
		    else if ((strcmp (newDecTalkMsg, DSRCTRLSEQ)) == 0)
			pMsg = DSRCTRLMSG;
		    else if ((strcmp (newDecTalkMsg, DSRDECTST)) == 0)
			pMsg = DSRDECTSTMSG;
		    else
			pMsg = "unknown DSR failure";
		    tisSysLog (0,0, "decTalkReset: DSR fail because (%s) (%s)", 
				pMsg, pChar);
		    return (1);

		default:
		    break;
	    }
	}
    }

    return (0);
}


/************************************************************************
*
*    decTalkState ()
*	return non 0 if the decTalk is in a given state (currentState)
*
************************************************************************/

int	decTalkState (void)
{
    int		status;

    switch (currentState)
    {
	case WAIT_FOR_AN_ANSWER:
	    status = 1;
	    break;

	case WAIT_FOR_AUTO_REPLY:
	    if (autoState == AUTO_ID_MSG)
		status = 1;
	    else
		status = 0;
	    break;

	case WAIT_FOR_ACK_REPEAT:
	    status = 1;
	    break;

	default:
	    status = 0;
	    break;
    }

if(debug)
{
tisSysLog(0,0,
"decTalkState, currentState=%s,autoState=%s,status=%d",
states[currentState],autoStates[autoState],status);
}
    return (status);
}


/*********************************************************
*
*    findPvName ()
*	called from addNewAlarmPoint ()
*	find a point name in the alarmMgr shared memory
*	and fill in the msg ptr with pertinant data
*
*********************************************************/

int	findPvName (QUEMSG_PTR newMsg)
{
    PvData	*pvdata = (PvData *)0;

    if (!mgrAvailable)
    {
	newMsg -> mgrNotAvailable = 1;
if (debug)
{
tisSysLog (0,0,
"findPvName: alarmMgr not available for %s", 
newMsg -> pointName);
}
	return (0);
    }

    if ( (pvdata = (PvData *)FindPV (NULL, newMsg -> pointName)) != 0)
    {
	buildMsg (newMsg, pvdata);
    }
    else
    {
	tisSysLog (0,0, "findPvName: pv %s not known to alarmMgr", newMsg -> pointName);
	return (1);
    }
    return (0);
}


/*********************************************
*
*    freeUnAckList ()
*	free unacknowledged alarm list
*
********************************************/

void	freeUnAckList (void)
{
    QUEMSG_PTR	prevMsg, currMsg;

    if (firstUnAck != (QUEMSG_PTR) 0)
    {
	currMsg = firstUnAck;
	while (currMsg != (QUEMSG_PTR) 0)
	{
	    prevMsg = currMsg;
	    currMsg = currMsg -> nextMsg;
	    free ((char *)prevMsg);
	}
    }
    firstUnAck = (QUEMSG_PTR)0;
    return;
}


/**************************************************************
*
*    getMsgMemory ()
*	called from addNewAlarmPoint () and buildUnAckList ()
*	return a memory chunk to store a new alarm msg
*	if not on free stack allocate a new chunk
*
**************************************************************/

QUEMSG_PTR	getMsgMemory (void)
{
    QUEMSG_PTR	newMsg = (QUEMSG_PTR)0;

    if ( (newMsg = (QUEMSG_PTR)calloc (1, sizeof (QUEMSG))) == NULL)
    {
	tisSysLog (0,0, "getMsgMemory: no memory available (%m)");
	return ((QUEMSG_PTR)0);
    }
    memset ((char *)newMsg, 0, sizeof (QUEMSG));
    return (newMsg);
}


/************************************************************
*
*    getNextAlarmMsg ()
*	get the next alarm data from the active or firstUnAck
*	alarm list
*
************************************************************/

QUEMSG_PTR	getNextAlarmMsg (QUEMSG_PTR list, int start)
{
    static    QUEMSG_PTR	newMsg = (QUEMSG_PTR)0;
    QUEMSG_PTR	returnMsg;

    if (start)
    {
	newMsg = list;
    }
    returnMsg = newMsg;
    while (newMsg != (QUEMSG_PTR)0)
    {
	if (newMsg -> removeFromList)
	{
	    newMsg = newMsg -> nextMsg;
	    continue;
	}
	if (newMsg -> alarmStatus == NO_ALARM)
	{

if (debug)
{
tisSysLog (0,0,
"getNextAlarmMsg: no alarm %s",
newMsg -> pointName);
}
	    newMsg -> removeFromList = 1;
	    newMsg = newMsg -> nextMsg;
	    continue;
	}
	if ( (newMsg -> alarmSeverity != INFO_ALARM) &&
	     (newMsg -> alarmSeverity != MINOR_ALARM) &&
	     (newMsg -> alarmSeverity != MAJOR_ALARM) )
	{

if (debug)
{
tisSysLog (0,0,
"getNextAlarmMsg: alarm not I,M,M for %s",
newMsg -> pointName);
}
	    newMsg -> removeFromList = 1;
	    newMsg = newMsg -> nextMsg;
	    continue;
	}
	returnMsg = newMsg;
	newMsg = newMsg -> nextMsg;
	break;
    }
    return (returnMsg);
}


/***************************************************
*
*    getNextDecTalkMsg ()
*	extract the next complete message from the 
*	decTalkRecvBuffer[]
*
***************************************************/

void	getNextDecTalkMsg (void)
{
    int	i = 0;

    memset (newDecTalkMsg, 0, sizeof (newDecTalkMsg));
    while (1)
    {
	if (i >= DECTALK_MSG_SIZE)
	    break;
	if (decTalkRecvBuffer[i] == NULL)
	    break;
	if (decTalkRecvBuffer[i] == '\r')
	{
	    memcpy ((char *)&newDecTalkMsg[0], (char *)&decTalkRecvBuffer[0],
		    (char *)&decTalkRecvBuffer[i] - (char *)&decTalkRecvBuffer[0]);
	    memcpy ((char *)&decTalkRecvBuffer[0], (char *)&decTalkRecvBuffer[i+1],
		    DECTALK_MSG_SIZE - (i + 1));
	    decTalkMsgCnt -= (i + 1);
	    memset ((char *)&decTalkRecvBuffer[decTalkMsgCnt], 0,
		    DECTALK_MSG_SIZE - decTalkMsgCnt);
	    break;
	}
	i++;
    }
    return;
}


/************************************************
*
*    initAccessList ()
*	open and store operator/password list
*
************************************************/

int	initAccessList (void)
{
    char		line[SLM_MAX_LINE_SIZE];
    char		*pChar;
    int			status = 0;
    struct stat		statbuff;
    FILE		*fd;
    ACCESSLIST_PTR	pAccess, lastEntry;

    secureLastModified = 0;
    status = stat (SECURELIST_PATH, &statbuff);
    if (status == -1)
    {
	tisSysLog (0,0, "initAccessList: stat on %s failed (%m)",
		SECURELIST_PATH);
	return (1);
    }

    if ((fd = fopen (SECURELIST_PATH,"r")) != 0)
    {
	while (!(feof(fd)))
	{
	    if (fgets (line, SLM_MAX_LINE_SIZE - 1, fd))
	    {
		if (pChar = strchr (line, '\n'))
		    *pChar = NULL;
		if ((pAccess = (ACCESSLIST_PTR)calloc (1, sizeof (ACCESSLIST))) == NULL)
		{
		    tisSysLog (0,0, "initAccessList: fail to alloc memory (%m)");
		    fclose (fd);
		    return (1);
		}
		pAccess -> nextEntry = (ACCESSLIST_PTR)0;
		strcpy (pAccess -> opname, strtok (line, " "));
		strcpy (pAccess -> passwd, strtok (NULL, " "));
		if (accessTree == (ACCESSLIST_PTR)0)
		    accessTree = pAccess;
		else
		    lastEntry -> nextEntry = pAccess;
		lastEntry = pAccess;
		
	    }
	    else if (ferror (fd))
	    {
		tisSysLog (0,0, "initAccesslist: read error (%m)");
		fclose (fd);
		return (1);
	    }
	}
	fclose (fd);
    }
    else
    {
	tisSysLog (0,0, "initAccessList: open error (%m)");
	return (1);
    }

    secureLastModified = statbuff.st_mtime;

if(debug)
{
tisSysLog(0,0,
"initAccessList: return (0)");
}
    return (0);
}


/********************************************************
*
*    initDecTalkList ()
*	read in list of points that are to be reported
*	to the decTalk if it exists
*
********************************************************/

int	initDecTalkList (void)
{
    struct stat		statbuff;
    ALARMPT_PTR		alarmPt;
    ALARMPT_PTR		prevPt;
    char		buff[128];
    char		*pChar;
    int			status;
    int			linecnt = 0;
    FILE		*fd;

    alarmTalkPts = (ALARMPT *)0;
    
    status = stat (TALKPTSLIST_PATH, &statbuff);
    if (status < 0)
    {
	/* its ok if the file does not exist, we will just 
	   process all alarm msgs we receive
	*/
	if (errno == ENOENT)
	    return (0);
	else
	{
	    tisSysLog (0,0, "initDecTalkList: stat error for %s (%m)",
		    TALKPTSLIST_PATH);
	    return (1);
	}
    }
    if ( (fd = fopen (TALKPTSLIST_PATH, "r")) == NULL)
    {
	tisSysLog (0,0, "initDecTalkList: open error on %s (%m)",
		TALKPTSLIST_PATH);
	return (1);
    }
    while (!feof(fd))
    {
	linecnt++;
	if (fgets (buff, PVNAME_SZ + 1, fd))
	{
	    if (pChar = strchr (buff, '\n'))
		*pChar = NULL;
	    if (buff[0] == 0)
		continue;
	    if ((alarmPt = (ALARMPT_PTR)calloc (1, sizeof (ALARMPT))) == 0)
	    {
		tisSysLog (0,0, "initDecTalkList: fail to alloc memory (%m)");
		fclose (fd);
		return (1);
	    }
	    alarmPt -> nextPoint = (ALARMPT_PTR)0;
	    strncpy (alarmPt -> pointName, buff, sizeof (alarmPt -> pointName));
	    alarmPt -> pointName[sizeof(alarmPt -> pointName) - 1] = 0;
	    if (alarmTalkPts == (ALARMPT_PTR)0)
		alarmTalkPts = alarmPt;
	    else
		prevPt -> nextPoint = alarmPt;
	    prevPt = alarmPt;
	}
	else if (ferror (fd))
	{
	    tisSysLog (0,0, "initDecTalkList: read error for %s,line %d (%m)", 
		    TALKPTSLIST_PATH, linecnt);
	    fclose (fd);
	    return (1);
	}
    }
    fclose (fd);

if (debug)
{
tisSysLog (0,0,
"initDecTalkList: return (0)");
}
	
    return (0);
}


/***************************************************************
*
*    initDecTalkPhoneticList ()
*	read in the list of words that need special phonetic
*	translations for decTalk to properly be able to 
*	pronounce them
*
***************************************************************/

int	initDecTalkPhoneticList (void)
{
    char	word[512];
    char	phonetic[512];
    char	msg[512];
    int		numReturned;
    int		maxFd = decTalkFd;
    int		linecnt = 0;
    int		status = 0;
    struct stat	statbuff;
    fd_set	chanmask;
    FILE	*fd;

    status = stat (PHONETICLIST_PATH, &statbuff);
    if (status < 0)
    {
	/* its ok if the file does not exist */
	if (errno == ENOENT)
	    return (0);
	else
	{
	    tisSysLog (0,0, "initDecTalkPhoenticList: stat error for %s (%m)",
		    PHONETICLIST_PATH);
	    return (1);
	}
    }
    if ((fd = fopen (PHONETICLIST_PATH, "r")) != NULL)
    {
	while (!(feof (fd)))
	{
	    memset ((char *)&word[0], 0, sizeof (word));
	    memset ((char *)&phonetic[0], 0, sizeof (phonetic));
	    linecnt++;
	    if ((numReturned = fscanf (fd, "%s\t%s\n", word, phonetic)) == 2)
	    {
		sprintf (msg, "%s%s %s%s\\",
			      DTDICT, word, phonetic, ESC);
		if (writeDecTalk (msg, "initDecTalkPhonetic"))
		{
		    fclose (fd);
		    return (1);
		}
		while (1)
		{
		    FD_ZERO(&chanmask);
		    FD_SET(decTalkFd, &chanmask);
		    status = select (maxFd + 1, &chanmask, NULL, NULL, NULL);
		    if (status == -1)
		    {
			tisSysLog (0,0, "initDecTalkPhoneticList: select error (%m)");
			fclose (fd);
			return (1);
		    }
		    if (readDecTalk ())
		    {
			tisSysLog (0,0, "initDecTalkPhonetics: decTalk read error");
			fclose (fd);
			return (1);
		    }
		    getNextDecTalkMsg ();
		    if (newDecTalkMsg[0] == NULL)
			continue;
		    if ((strcmp (newDecTalkMsg, DTDICTWORDOK)) == 0)
			break;
		    else if ((strcmp (newDecTalkMsg, DTDICTNOROOM)) == 0)
		    {
			tisSysLog (0,0, "initDecTalkPhoneticList: no more room for word %s, line %d",
				word, linecnt);
			fclose (fd);
			return (1);
		    }
		    else if ((strcmp (newDecTalkMsg, DTDICTWORDLONG)) == 0)
		    {
			tisSysLog (0,0, "initDecTalkPhonetic, word %s phonetic %s > 256 chars, line %d",
				word, phonetic, linecnt);
			fclose (fd);
			return (1);
		    }
		    else
		    {
			tisSysLog (0,0, "initDecTalkPhonetic: unknown msg (%s), for word %s, line %d",
				convertDecTalkMsg (newDecTalkMsg), word, linecnt);
			continue;
		    }
		}
	    }
	    else if (ferror (fd))
	    {
		tisSysLog (0,0, "initDecTalkPhoneticList; read error (%m)");
		fclose (fd);
		return (1);
	    }
	    else if (!(feof (fd)))
	    {
		/* skip a blank line */
		if (numReturned > 0)
		{
		    tisSysLog (0,0, "initDecTalkPhonetic: line %d missing word or phonecic",
			    linecnt);
		}
	    }
	}
	fclose (fd);
    }
    else
    {
	tisSysLog (0,0, "initDecTalkPhoneticList, open error (%m)");
	return (1);
    }

if(debug)
{
tisSysLog(0,0,
"initDecTalkPhoneticList: return (0)");
}

    return (0);
}


/*****************************************************
*
*    initSignals ()
*	inititialize the signal handler
*	use sigset for SOLARIS because under SOLARIS
*	the default signal handler will be attached
*	to the program after the my signal handler
*	returns after the 1st call
*
*****************************************************/

void	initSignals (void)
{
#ifdef SOLARIS
/*    sigset (SIGPIPE, alarmTalkSignalHandler); */
    sigset (SIGINT, alarmTalkSignalHandler);
    sigset (SIGTERM, alarmTalkSignalHandler);
    sigset (SIGUSR1, alarmTalkSignalHandler);
    sigset (SIGUSR2, alarmTalkSignalHandler);
#else
/*    signal (SIGPIPE, alarmTalkSignalHandler); */
    signal (SIGINT, alarmTalkSignalHandler);
    signal (SIGTERM, alarmTalkSignalHandler);
    signal (SIGUSR1, alarmTalkSignalHandler);
    signal (SIGUSR2, alarmTalkSignalHandler);
#endif
    return;
}


/******************************************************************
*
*    parseIdMsg ()
*	extract the keypad ID number, we have todo this because
*	the decTalk has been know to send an escape squence in
*	between the keypad sequence, but if that escape sequence
*	is a hangup or wink, let processAlarms know
*
******************************************************************/

int	parseIdMsg (void)
{
    int		status = 0;
    char	*pChar;

    pChar = newDecTalkMsg;

    while ((*pChar) != NULL)
    {
	if ( (isdigit ((*pChar))) ||
	     ((*pChar) == '#') )
	{
	    opIdCnt++;
	    if (opIdCnt >= SLM_KEYPADID_SIZE)
	    {
		break;
	    }
	    opId[opIdCnt] = *pChar;
	    if ((*pChar) == '#')
	    {
		opId[opIdCnt + 1] = NULL;
		status = 0;
		break;
	    }
	    pChar++;
	}
	else
	{
	    if ( (*pChar) == '\e')
	    {
		if ( ((strcmp (pChar, PHONHOOK)) == 0) ||
		     ((strcmp (pChar, PHWINK)) == 0) )
		{
		    status = 2;
		}
		else if ((strcmp (pChar, PHTIMEOUT)) == 0)
		{
		    status = 3;
		}
		else
		{
		    status = 1;
		}
	    }
	    else
	    {
		status = 1;
	    }
	    break;
	}
    }

if(debug)
{
tisSysLog(0,0,
"parseIdMsg, ret, status=%d,newDec=%s,opId=%s",
status,convertDecTalkMsg(newDecTalkMsg),opId);
}
    return (status);
}


/********************************************************
*
*    processAlarms ()
*	process the next alarm in the activeList or
*	set some state while we wait for the decTalk
*	to respond to us
*
********************************************************/

int	processAlarms (void)
{
    int		inState;
    int		foundAlarm;
    int		noReturn = 1;
    int		startMsgSearch = 1;	/* tells getNextAlarmMsg to start
					   at beginning of list
					*/
    int		timeout;
    int		timediff;
    int		sendGreeting;
    int		parseReturn = 0;
    char	cmdbuff[256];
    QUEMSG_PTR	nextMsg = (QUEMSG_PTR)0;
    QUEMSG_PTR	firstGoodAlarm;

if (debug)
{
tisSysLog(0,0,
"processAlarms: enter, currentState = %s", 
states[currentState]);
}
    while (noReturn)
    {
	switch (currentState)
	{
	    case GET_NEXT_ALARM_MSG:	/* get the next alarm msg */
		whoWasCalled = NO_CALL_YET;
		autoNumber = 0;
		autoState = NO_AUTO_STATE;
		phoneCallFailed = 0;
		autoTimeoutCnt = 0;
		autoIdCnt = 0;
		foundAlarm = 0;
		inState = 1;
		firstGoodAlarm = (QUEMSG_PTR)0;
		while (inState)
		{
		    if ((nextMsg = getNextAlarmMsg(firstActive, startMsgSearch)) == (QUEMSG_PTR)0)
		    {
			inState = 0;
			noReturn = 0;
			continue;
		    }
		    /* if msg already was ack'd */
		    if (nextMsg -> ack)
		    {
			nextMsg -> removeFromList = 1;
			startMsgSearch = 0;
			continue;
		    }
		    /* already processed or for some reason we do not want
		       to process this alarm
		    */
		    if (nextMsg -> removeFromList)
		    {
			startMsgSearch = 0;
			continue;
		    }
		    if (readCurrentAlarmConfig (nextMsg))
		    {
			startMsgSearch = 0;
			nextMsg -> removeFromList = 1;
			continue;
		    }
		    if ((strcmp (currentAlarmConfig -> annunciation, NO_ANNUNCIATION)) == 0)
		    {
			startMsgSearch = 0;
			nextMsg -> removeFromList = 1;
			continue;
		    }
		    if ( ((time (0)) - nextMsg -> alarmRecvd) < timeToAnnunciate)
		    {
			startMsgSearch = 0;
			firstGoodAlarm = nextMsg;
			continue;
		    }
		    currentAlarm = nextMsg;
		    saveCurrentAlarm = nextMsg;
		    currentState = WHO_TO_CALL;
		    startMsgSearch = 0;
		    foundAlarm = 1;
		    inState = 0;
		}
		if (!foundAlarm)
		{
		    if ( timeToAnnunciate && (firstGoodAlarm != (QUEMSG_PTR)0) )
		    {
			timediff = timeToAnnunciate - 
					(time (0) - firstGoodAlarm -> alarmRecvd);
			if (timediff <= 0)
			{
			    deadline = time (0) + TIMEOUT_DEADBAND;
			}
			else
			{
			    deadline = time (0) + timediff;
			}
		    }
		    else
		    {
			deadline = time (0) + TIMEOUT_DEADBAND;
		    }
		    noReturn = 0;
		}
if(debug)
{
tisSysLog(0,0,
"processAlarms,leave GET_NEXT_MSG,found=%d,state=%s,who=%s",
foundAlarm,states[currentState],WhoWasCalled[whoWasCalled]);
}
		break;
	    /* processAlarms (cont) */
	    case WHO_TO_CALL:		/* who do we have to call */
if(debug)
{
tisSysLog(0,0,
"processAlarms,WHO_TO_CALL,who=(%s), callFailed=%d, sentAuto=%d",
WhoWasCalled[whoWasCalled], phoneCallFailed, sentToAutoCaller);
}
		switch (whoWasCalled)
		{
		    case NO_CALL_YET:	/* we have not made a call yet */
			if (currentAlarmConfig -> spkrDevice)
			{
			    if (currentAlarmConfig -> conNumber[0] == NULL)
			    {
				whoWasCalled = CONSOLE_CALL;
				currentState = WHO_TO_CALL;
				break;
			    }
			    whoWasCalled = CONSOLE_CALL;
			    currentState = MAKE_A_PHONE_CALL;
			    validPhoneNumber = 1;
			    phoneCallFailed = 0;
			    sentToAutoCaller = 0;
			    break;
			}
			if (currentAlarmConfig -> pageDevice)
			{
			    if (currentAlarmConfig -> pageNumber[0] == NULL)
			    {
				whoWasCalled = PAGE_CALL;
				currentState = WHO_TO_CALL;
				break;
			    }
			    whoWasCalled = PAGE_CALL;
			    currentState = MAKE_A_PHONE_CALL;
			    validPhoneNumber = 1;
			    phoneCallFailed = 0;
			    sentToAutoCaller = 0;
			    break;
			}
			if (currentAlarmConfig -> autoDevice)
			{
			    /*
			    currentAlarm = saveCurrentAlarm;
			    if (sentToAutoCaller)
			    {
				currentAlarm -> removeFromList = 1;
				currentState = GET_NEXT_ALARM_MSG;
				break;
			    }
			    */
			    if (currentAlarmConfig -> autoNumbrs[0][0] == NULL)
			    {
				tisSysLog (0,0, "processAlarms: WHO_TO_CALL, point %s alarm, no auto numbers to phone",
					currentAlarm -> pointName);
				currentAlarm -> removeFromList = 1;
				currentState = GET_NEXT_ALARM_MSG;
				break;
			    }
			    whoWasCalled = AUTO_CALL;
			    autoNumber = 0;
			    validPhoneNumber = 1;
			    currentState = MAKE_A_PHONE_CALL;
			    phoneCallFailed = 0;
			    sentToAutoCaller = 0;
			    break;
			}
			tisSysLog (0,0, "processAlarms: WHO_TO_CALL, point %s, no device to phone",
				currentAlarm -> pointName);
			currentAlarm -> removeFromList = 1;
			currentState = GET_NEXT_ALARM_MSG;
			break;

		    case CONSOLE_CALL:	
			if (phoneCallFailed)
			{
			    currentState = WAIT_TO_TRY_AGAIN;
			    phoneCallFailed = 0;
			    break;
			}
			if (currentAlarmConfig -> pageDevice)
			{
			    if (currentAlarmConfig -> pageNumber[0] == NULL)
			    {
				whoWasCalled = PAGE_CALL;
				currentState = WHO_TO_CALL;
				break;
			    }
			    whoWasCalled = PAGE_CALL;
			    currentState = MAKE_A_PHONE_CALL;
			    validPhoneNumber = 1;
			    phoneCallFailed = 0;
			    sentToAutoCaller = 0;
			    break;
			}
			currentAlarm = saveCurrentAlarm;
			if (currentAlarmConfig -> autoDevice && (!sentToAutoCaller) )
			{
			    if (currentAlarmConfig -> autoNumbrs[0][0] == NULL)
			    {
				tisSysLog (0,0, "processAlarms: CONSOLE_CALL, point %s, no auto numbers",
					currentAlarm -> pointName);
				currentAlarm-> removeFromList = 1;
				currentState = GET_NEXT_ALARM_MSG;
				break;
			    }
			    whoWasCalled = AUTO_CALL;
			    autoNumber = 0;
			    currentState = MAKE_A_PHONE_CALL;
			    validPhoneNumber = 1;
			    phoneCallFailed = 0;
			    sentToAutoCaller = 0;
			    break;
			}
			/* no one else to call done this alarm,
			   process another alarm
			*/
			currentAlarm-> removeFromList = 1;
			currentState = GET_NEXT_ALARM_MSG;
			break;

		    case PAGE_CALL:
			if (phoneCallFailed)
			{
			    currentState = WAIT_TO_TRY_AGAIN;
			    phoneCallFailed = 0;
			    break;
			}
			currentAlarm = saveCurrentAlarm;
			if (currentAlarmConfig -> autoDevice && (!sentToAutoCaller) )
			{
			    if (currentAlarmConfig -> autoNumbrs[0][0] == NULL)
			    {
				tisSysLog (0,0, "processAlarms: PAGE_CALL, point %s, no auto numbers to phone",
					currentAlarm -> pointName);
				if (!validPhoneNumber)
				{
				    tisSysLog (0,0, "processAlarms: PAGE_CALL, point %s, no numbers to call",
					    currentAlarm -> pointName);
				}
				currentAlarm-> removeFromList = 1;
				currentState = GET_NEXT_ALARM_MSG;
				break;
			    }
			    whoWasCalled = AUTO_CALL;
			    autoNumber = 0;
			    currentState = MAKE_A_PHONE_CALL;
			    validPhoneNumber = 1;
			    phoneCallFailed = 0;
			    sentToAutoCaller = 0;
			    break;
			}
			/* no one else to call done this alarm,
			   process another alarm
			*/
			currentAlarm -> removeFromList = 1;
			currentState = GET_NEXT_ALARM_MSG;
			break;

		    case AUTO_CALL:
			currentAlarm = saveCurrentAlarm;
			if (sentToAutoCaller)
			{
			    currentAlarm -> removeFromList = 1;
			    currentState = GET_NEXT_ALARM_MSG;
			    startMsgSearch = 1;
			    break;
			}
			autoNumber++;
			phoneCallFailed = 0;
			if ( (autoNumber >= NUM_AUTO_CALLS) ||
			     (currentAlarmConfig -> autoNumbrs[autoNumber][0] == NULL) )
			{
			    currentState = WAIT_TO_TRY_AGAIN;
			    autoNumber = 0;
			    autoState = NO_AUTO_STATE;
			}
			else
			{
			    currentState = MAKE_A_PHONE_CALL;
			}
			break;
		}
		break;

	    /* processAlarms (cont) */

	    case MAKE_A_PHONE_CALL:	/* reach out and touch someone */
		{
		    char	phoneNumber[PHONENUMBR_SZ + 1];
if(debug)
{
tisSysLog(0,0,
"processAlarms,MAKE_A_PHONE_CALL,who=(%s)",
WhoWasCalled[whoWasCalled]);
}

		    if (whoWasCalled == CONSOLE_CALL)
		    {
			timeout = currentAlarmConfig -> conRedial;
			strncpy (phoneNumber, currentAlarmConfig -> conNumber, PHONENUMBR_SZ);
			phoneNumber[PHONENUMBR_SZ] = NULL;
		    }
		    else if (whoWasCalled == PAGE_CALL)
		    {
			timeout = currentAlarmConfig -> pageRedial;
			strncpy (phoneNumber, currentAlarmConfig -> pageNumber, PHONENUMBR_SZ);
			phoneNumber[PHONENUMBR_SZ] = NULL;
		    }
		    else
		    {
			timeout = currentAlarmConfig -> autoFailAnsTimeout;
			strncpy (phoneNumber, 
				 (char *)&currentAlarmConfig -> autoNumbrs[autoNumber][0], 
				 PHONENUMBR_SZ);
			phoneNumber[PHONENUMBR_SZ] = NULL;
		    }
		    if (timeout == 0)
		    {
			timeout = ANSWER_PHONE_TIMEOUT;
		    }
		    else if (timeout < ANSWER_PHONE_LOW_TIMEOUT)
		    {
			timeout = ANSWER_PHONE_LOW_TIMEOUT;
		    }
		    else if (timeout > ANSWER_PHONE_HIGH_TIMEOUT)
		    {
			timeout = ANSWER_PHONE_HIGH_TIMEOUT;
		    }
		    sprintf (cmdbuff, "%s%sX<%d>%s\\",
				      DTPHONECALL, phoneNumber, timeout, ESC);
		    /* set this to calc new timeout for select statement in main() */
		    deadline = DIAL_TONE_TIMEOUT + timeout + time (0) + TIMEOUT_DEADBAND;
		    writeDecTalk (cmdbuff, "processAlarms");
		    currentState = WAIT_FOR_AN_ANSWER;
		    noReturn = 0;
		    break;
		}

	    /* processAlarms (cont) */

	    case WAIT_FOR_AN_ANSWER:	/* wait till someone answers the phone */
		getNextDecTalkMsg ();
if(debug)
{
tisSysLog(0,0,
"processAlarms,WAIT_FOR_ANS,msg=(%s)",
convertDecTalkMsg(newDecTalkMsg));
}
		/* do not have a complete msg yet */
		if (newDecTalkMsg[0] == NULL)
		{
			deadline = time (0) + TIMEOUT_DEADBAND;
			noReturn = 0;
		}
		else if ((strcmp (newDecTalkMsg, PHOFFHOOK)) == 0)
		{
		    sendGreeting = 1;
		    sendAlarmMsg (sendGreeting);
		    currentState = WAIT_FOR_REPLY;
		    deadline = time (0) + TIMEOUT_DEADBAND + TIMEOUT_DEADBAND;
		    noReturn = 0;
		}
		else if ( ((strcmp (newDecTalkMsg, PHBUSY)) == 0) ||
		          ((strcmp (newDecTalkMsg, PHNOANS)) == 0) ||
			  ((strcmp (newDecTalkMsg, PHNODIAL)) == 0) )
		{
		    currentState = WHO_TO_CALL;
		    phoneCallFailed = 1;
		}
		else if ((strcmp (newDecTalkMsg, PHONHOOK)) == 0)
		{
		    noReturn = 0;
		    if (deadline < time (0))
			deadline = time (0) + TIMEOUT_DEADBAND;
		}
		else
		{
		    tisSysLog (0,0, "processAlarms: WAIT_FOR_ANS, unknown msg (%s)",
			    convertDecTalkMsg(newDecTalkMsg));
		    writeDecTalk (DTHANGUP, "processAlarms");
		    currentState = WAIT_FOR_HANGUP;
		    phoneCallFailed = 1;
		}
		break;

	    /* call failed, no other options but to try again,
	       but wait the designated amount of time
	    */
	    case WAIT_TO_TRY_AGAIN:	
if(debug)
{
tisSysLog(0,0,
"processAlarms,WAIT_TRY_AGAIN,who=(%s)",
WhoWasCalled[whoWasCalled]);
}
		/* clean up any messages still in the decTalk */
		fflushDecTalkQue ();
		if (whoWasCalled == CONSOLE_CALL)
		{
		    timeout = currentAlarmConfig -> conRedial;
		}
		else if (whoWasCalled == PAGE_CALL)
		{
		    timeout = currentAlarmConfig -> pageRedial;
		}
		else	/* auto-dialer */
		{
		    timeout = currentAlarmConfig -> autoRestart;
		}
		currentState = MAKE_A_PHONE_CALL;
		/* we will wait the designated amount of time before
		   we try to redial
		deadline = time (0) + timeout + TIMEOUT_DEADBAND;
		noReturn = 0;
		break;

	    /* processAlarms (cont) */

	    /* have sent alarm msg */
	    case WAIT_FOR_REPLY:	
		getNextDecTalkMsg ();
if(debug)
{
tisSysLog(0,0,
"processAlarms,WAIT_REPLY, msg=(%s)",
convertDecTalkMsg(newDecTalkMsg));
}
		/* have not spoken  entire message yet,
		   wait somemore time or we get a phone off hook
		*/
		if ( (newDecTalkMsg[0] == NULL) ||
		     ((strcmp (newDecTalkMsg, PHOFFHOOK)) == 0) )
		{
		    deadline = time (0) + TIMEOUT_DEADBAND;
		    noReturn = 0;
		}
		else if ( (strcmp (newDecTalkMsg, DTINDEXREPLY)) == 0)
		{
		    if ( (whoWasCalled == CONSOLE_CALL) ||
			 (whoWasCalled == PAGE_CALL) )
		    {
			/* alarm message successfully sent */
			if (whoWasCalled == CONSOLE_CALL)
			    sendHangUpMsg ();
			writeDecTalk (DTHANGUP, "processAlarms");
			currentState = WAIT_FOR_HANGUP;
			phoneCallFailed = 0;
			deadline = time (0) + TIMEOUT_DEADBAND;
			noReturn = 0;
		    }
		    /* it was an auto dialer call */
		    else
		    {
			sentToAutoCaller = 1;
			timeout = sendAutoIdMsg ();
			deadline = time (0) + timeout + TIMEOUT_DEADBAND;
			currentState = WAIT_FOR_AUTO_REPLY;
			autoState = AUTO_ID_MSG;
			autoTimeoutCnt = 0;
			autoIdCnt = 0;
			phoneCallFailed = 0;
			memset ((char *)&opId[0], 0, sizeof (opId));
			opIdCnt = -1;
			noReturn = 0;
		    }
		}
		/* we had a hangup before message was finished,
		   or there was a line current interruption
		*/
		else if ( ((strcmp (newDecTalkMsg, PHONHOOK)) == 0) ||
			  ((strcmp (newDecTalkMsg, PHWINK)) == 0) )
		{
		    writeDecTalk (DTHANGUP, "processAlarms");
		    currentState = WAIT_FOR_HANGUP;
		    phoneCallFailed = 1;
		    /* wait for any decTalk reply if any */
		    deadline = time (0) + TIMEOUT_DEADBAND;
		    noReturn = 0;
		}
		else
		{
		    tisSysLog (0,0, "processAlarms: WAIT_REPLY, unexpected msg (%s)",
			    convertDecTalkMsg (newDecTalkMsg));
		    writeDecTalk (DTSTOP, "processAlarms");
		    writeDecTalk (DTHANGUP, "processAlarms");
		    currentState = WAIT_FOR_HANGUP;
		    phoneCallFailed = 1;
		    deadline = time (0) + TIMEOUT_DEADBAND;
		    noReturn = 0;
		}
		break;
	    /* processAlarms (cont) */

	    /* wait for user to hangup phone */
	    case WAIT_FOR_HANGUP:	
		getNextDecTalkMsg ();
if(debug)
{
tisSysLog(0,0,
"processAlarms,WAIT_HANGUP, msg=(%s)",
convertDecTalkMsg(newDecTalkMsg));
}
		if ( ((strcmp (newDecTalkMsg, PHONHOOK)) == 0) ||
		     ((strcmp (newDecTalkMsg, PHWINK)) == 0) )
		{
		    writeDecTalk (DTSTOP, "processAlarms");
		    fflushDecTalkQue ();
		    validPhoneNumber = 0;
		    /* see if we have to call someone else */
		    currentState = WHO_TO_CALL;
		}
		else
		{
		    deadline = time (0) + TIMEOUT_DEADBAND;
		    noReturn = 0;
		    break;
		}
		break;

	    /* processAlarms (cont) */

	    /* wait for replies during auto dialer */
	    case WAIT_FOR_AUTO_REPLY:	
		getNextDecTalkMsg ();
if(debug)
{
tisSysLog(0,0,
"processAlarms,WAIT_AUTO_REPLY, autoState=%s, msg=(%s)",
autoStates[autoState],convertDecTalkMsg(newDecTalkMsg));
}
		switch (autoState)
		{
		    case AUTO_ID_MSG:
			if ( (newDecTalkMsg[0] == NULL) ||
			     ((strcmp (newDecTalkMsg, PHOFFHOOK)) == 0) )
			{
			    deadline = time (0) + (3 * TIMEOUT_DEADBAND);
			    noReturn = 0;
			}
			/* had a hangup or line interrupt */
			else if ( ((strcmp (newDecTalkMsg, PHONHOOK)) == 0) ||
			     ((strcmp (newDecTalkMsg, PHWINK)) == 0) )
			{
			    writeDecTalk (DTHANGUP, "processAlarms");
			    currentState = WAIT_FOR_HANGUP;
			    phoneCallFailed = 0;
			    autoState = NO_AUTO_STATE;
			    /* wait for any decTalk reply if any */
			    deadline = time (0) + TIMEOUT_DEADBAND;
			    noReturn = 0;
			}
			else if ( (strcmp (newDecTalkMsg, PHTIMEOUT)) == 0)
			{
timeout1:
			    /* they did not respond in time, may try again */
			    if (autoTimeoutCnt == 0)
			    {
				autoTimeoutCnt++;
				timeout = sendFailToEnterIdMsg ();
				deadline = time (0) + timeout + TIMEOUT_DEADBAND;
				noReturn = 0;
			    }
			    else
			    {
				sendFinalFailedIdMsg ();
				writeDecTalk (DTHANGUP, "processAlarms");
				deadline = time (0) + TIMEOUT_DEADBAND + TIMEOUT_DEADBAND;
				currentState = WAIT_FOR_HANGUP;
				autoState = NO_AUTO_STATE;
				phoneCallFailed = 0;
				noReturn = 0;
			    }
			}
			else if ( (isdigit(newDecTalkMsg[0])) ||
				  (newDecTalkMsg[0] == '#') )
			{
			    /* if we have not rev'd the entire id, wait some more time,
			       the decTalk mayhave sent an escape sequence in between
			       the id keying
			    */
			    parseReturn = parseIdMsg ();
			    /* we did not get the entire id, requires the # to end it */
			    if (parseReturn == 1)
			    {
				deadline = time (0) + (3 * TIMEOUT_DEADBAND);
				noReturn = 0;
			    }
			    /* we got a hangup or line interrupt while the ID was
			       being entered
			    */
			    else if (parseReturn == 2)
			    {
				writeDecTalk (DTHANGUP, "processAlarms");
				currentState = WAIT_FOR_HANGUP;
				phoneCallFailed = 0;
				autoState = NO_AUTO_STATE;
				/* wait for any decTalk reply if any */
				deadline = time (0) + TIMEOUT_DEADBAND;
				noReturn = 0;
			    }
			    /* they entered part of their ID, but did not enter the
			       number sign within the time limit
			    */
			    else if (parseReturn == 3)
			    {
				goto timeout1;
			    }
			    /* got the full ID now */
			    else
			    {
				if (verifyAutoId ())
				{
				    sendWaitBuildListMsg ();
				    /* error while trying to build the list */
				    if (buildUnAckList ())
				    {
					sendNoAlarmsMsg (BUILD_ERR_MSG);
					writeDecTalk (DTHANGUP, "processAlarms");
					deadline = time (0) + TIMEOUT_DEADBAND + TIMEOUT_DEADBAND;
					currentState = WAIT_FOR_HANGUP;
					autoState = NO_AUTO_STATE;
					phoneCallFailed = 0;
					freeUnAckList ();
				    }
				    else
				    {
					/* there are no alarms to no reported, or there are
					   no alarms for which the user has privileges to
					   ack 
					*/
					if (firstUnAck == (QUEMSG_PTR)0)
					{
					    sendNoAlarmsMsg (NO_PRIV_ALARMS_MSG);
					    writeDecTalk (DTHANGUP, "processAlarms");
					    deadline = time (0) + TIMEOUT_DEADBAND;
					    currentState = WAIT_FOR_HANGUP;
					    autoState = NO_AUTO_STATE;
					    phoneCallFailed = 0;
					}
					else
					{
					    autoState = BUILD_ACK_LIST;
					    autoTimeoutCnt = 0;
					    deadline = time (0) + TIMEOUT_DEADBAND;
					}
				    }
				    noReturn = 0;
				}
				else
				{
				    /* invalid ID, if 1st try, give them another chance */
				    if (autoIdCnt == 0)
				    {
					autoIdCnt++;
					timeout = sendRepeatIdMsg ();
					deadline = time (0) + timeout + TIMEOUT_DEADBAND;
					noReturn = 0;
				    }
				    /* 2 tries and they are out of here */
				    else
				    {
					sendFinalBadIdMsg ();
					writeDecTalk (DTHANGUP, "processAlarms");
					deadline = time (0) + TIMEOUT_DEADBAND;
					currentState = WAIT_FOR_HANGUP;
					autoState = NO_AUTO_STATE;
					phoneCallFailed = 0;
					freeUnAckList ();
					noReturn = 0;
				    }
				}
			    }
			}
			else
			{
			    tisSysLog (0,0, "processAlarms: WAIT_AUTO_REPLY, AUTO_ID, unexpected msg (%s)",
				    convertDecTalkMsg (newDecTalkMsg));
		    	    writeDecTalk (DTSTOP, "processAlarms");
			    sendUnexpectedMsg (ID_NOT_RECV_MSG);
			    sendHangUpMsg ();
			    writeDecTalk (DTHANGUP, "processAlarms");
			    deadline = time (0) + TIMEOUT_DEADBAND;
			    currentState = WAIT_FOR_HANGUP;
			    autoState = NO_AUTO_STATE;
			    phoneCallFailed = 0;
			    freeUnAckList ();
			    noReturn = 0;
			}
			break;

		    /* processAlarms (cont) */

		    case BUILD_ACK_LIST:
if(debug)
{
tisSysLog(0,0,
"processAlarms,BUILD_ACK_LIST msg=(%s)",
convertDecTalkMsg(newDecTalkMsg));
}
			/* entire wait msg was not spoken yet, 
			   wait somemore time 
			*/
			if ( (newDecTalkMsg[0] == NULL) ||
			     ((strcmp (newDecTalkMsg, PHOFFHOOK)) == 0) ||
			     ((strcmp (newDecTalkMsg, PHTIMEOUT)) == 0) )
			{
			    deadline = time (0) + TIMEOUT_DEADBAND;
			    noReturn = 0;
			}
			else if ( ((strcmp (newDecTalkMsg, PHONHOOK)) == 0) ||
			     ((strcmp (newDecTalkMsg, PHWINK)) == 0) )
			{
			    freeUnAckList ();
			    writeDecTalk (DTHANGUP, "processAlarms");
			    currentState = WAIT_FOR_HANGUP;
			    phoneCallFailed = 0;
			    autoState = NO_AUTO_STATE;
			    /* wait for any decTalk reply if any */
			    deadline = time (0) + TIMEOUT_DEADBAND;
			    noReturn = 0;
			}
			/* the unack list is built and the build msg was
			   completely spoken
			*/
			else if ( (strcmp (newDecTalkMsg, DTINDEXREPLY)) == 0)
			{
			    currentState = SEND_UNACKED;
			    startMsgSearch = 1;
			    break;
			}
			else
			{
			    tisSysLog (0,0, "processAlarms: WAIT_FOR_AUTO,BUILD_ACK, unexpected msg (%s)",
				    convertDecTalkMsg (newDecTalkMsg));
		    	    writeDecTalk (DTSTOP, "processAlarms");
			    sendUnexpectedMsg (UNABLE_TO_SEND_LIST_MSG);
			    sendHangUpMsg ();
			    writeDecTalk (DTHANGUP, "processAlarms");
			    currentState = WAIT_FOR_HANGUP;
			    autoState = NO_AUTO_STATE;
			    phoneCallFailed = 0;
			    deadline = time (0) + TIMEOUT_DEADBAND;
			    noReturn = 0;
			}
			break;

		}  /* end switch (autoState ) */

		break;	/* end case WAIT_FOR_AUTO_REPLY */

	    /* processAlarms (cont) */

	    /* send the list of unacknowleged alarms */
	    case SEND_UNACKED:
if(debug)
{
tisSysLog(0,0,
"processAlarms,SEND_UNACKED");
}
		/* NOTE startMsgSearch was set in case BUILD_ACK_LIST,
			to 1 to start search at start of firstUnAck list
		*/
		inState = 1;
		foundAlarm = 0;
		while (inState)
		{
		    /* if we have annunciated all unacked alarms, see if we 
		       have to repeat the list or ack all points
		    */
		    if ( (nextMsg = getNextAlarmMsg (firstUnAck, startMsgSearch)) == (QUEMSG_PTR)0)
		    {
			currentState = AUTO_REPEAT;
			break;
		    }
		    if (nextMsg -> removeFromList)
		    {
			startMsgSearch = 0;
			continue;
		    }
		    if (readCurrentAlarmConfig (nextMsg))
		    {
			nextMsg -> removeFromList = 1;
			startMsgSearch = 0;
			continue;
		    }
		    foundAlarm = 1;
		    inState = 0;
		    startMsgSearch = 0;
		}
		if (foundAlarm)
		{
		    currentAlarm = nextMsg;
		    sendGreeting = 0;
		    sendAlarmMsg (sendGreeting);
		    currentState = AUTO_CALLER_SENT;
		    deadline = time (0) + TIMEOUT_DEADBAND;
		    noReturn = 0;
		}
		else
		{
		    /* did not find any list to send to auto caller */
		    if (currentState != AUTO_REPEAT)
		    {
			sendNoAlarmsMsg (NO_ALARMS_TOBE_SENT_MSG);
			writeDecTalk (DTHANGUP, "processAlarms");
			deadline = time (0) + TIMEOUT_DEADBAND;
			currentState = WAIT_FOR_HANGUP;
			autoState = NO_AUTO_STATE;
			phoneCallFailed = 0;
			freeUnAckList ();
			noReturn = 0;
		    }
		}
		break;

	    /* processAlarms (cont) */

	    /* wait for the the unacknowleged alarm to be annunciated */
	    case AUTO_CALLER_SENT:
		getNextDecTalkMsg ();
if(debug)
{
tisSysLog(0,0,
"processAlarms,AUTO_CALLER_SENT, msg=(%s)",
convertDecTalkMsg(newDecTalkMsg));
}
		/* have not received the msg yet, give it more time */
		if ( (newDecTalkMsg[0] == NULL) ||
		     ((strcmp (newDecTalkMsg, PHOFFHOOK)) == 0) )
		{
		    deadline = time (0) + TIMEOUT_DEADBAND;
		    noReturn = 0;
		}
		/* hangup or line interrupt */
		else if ( ((strcmp (newDecTalkMsg, PHONHOOK)) == 0) ||
			  ((strcmp (newDecTalkMsg, PHWINK)) == 0) )
		{

		    writeDecTalk (DTHANGUP, "processAlarms");
		    freeUnAckList ();
		    currentState = WAIT_FOR_HANGUP;
		    autoState = NO_AUTO_STATE;
		    phoneCallFailed = 0;
		    deadline = time (0) + TIMEOUT_DEADBAND;
		    noReturn = 0;
		}
		else if ( (strcmp (newDecTalkMsg, DTINDEXREPLY)) == 0)
		{
		    currentAlarm -> removeFromList = 1;
		    currentState = SEND_UNACKED;
		}
		else
		{
		    tisSysLog (0,0, "processAlarms: AUTO_CALLER_SENT, unexpected msg (%s)",
			    convertDecTalkMsg (newDecTalkMsg));
		    writeDecTalk (DTSTOP, "processAlarms");
		    sendUnexpectedMsg (UNABLE_TO_SEND_LIST_MSG);
		    sendHangUpMsg ();
		    writeDecTalk (DTHANGUP, "processAlarms");
		    freeUnAckList ();
		    currentState = WAIT_FOR_HANGUP;
		    autoState = NO_AUTO_STATE;
		    phoneCallFailed = 0;
		    deadline = time (0) + TIMEOUT_DEADBAND;
		    noReturn = 0;
		}
		break;

	    /* processAlarms (cont) */

	    /* entire unacknowledged list was spoken, see if user wants
	       the list repeated or acknowledged
	    */
	    case AUTO_REPEAT:
if(debug)
{
tisSysLog(0,0,
"processAlarms,AUTO_REPEAT");
}
		timeout = sendRepeatAckMsg ();
		deadline = time (0) + timeout + TIMEOUT_DEADBAND;
		autoTimeoutCnt = 0;
		currentState = WAIT_FOR_ACK_REPEAT;
		noReturn = 0;
		break;

	    /* wait for users reply to repeat/ack question */
	    case WAIT_FOR_ACK_REPEAT:
		getNextDecTalkMsg ();
if(debug)
{
tisSysLog(0,0,
"processAlarms,WAIT_FOR_ACK_REPEAT, msg=(%s)",
convertDecTalkMsg(newDecTalkMsg));
}
		/* have not received the msg yet, give it more time */
		if ( (newDecTalkMsg[0] == NULL) ||
		     ((strcmp (newDecTalkMsg, PHOFFHOOK)) == 0) )
		{
		    deadline = time (0) + (3 * TIMEOUT_DEADBAND);
		    noReturn = 0;
		}
		else if ( ((strcmp (newDecTalkMsg, PHONHOOK)) == 0) ||
		     ((strcmp (newDecTalkMsg, PHWINK)) == 0) )
		{
		    writeDecTalk (DTHANGUP, "processAlarms");
		    freeUnAckList ();
		    currentState = WAIT_FOR_HANGUP;
		    autoState = NO_AUTO_STATE;
		    phoneCallFailed = 0;
		    deadline = time (0) + TIMEOUT_DEADBAND;
		    noReturn = 0;
		}
		else if ( (strcmp (newDecTalkMsg, PHTIMEOUT)) == 0)
		{
		    if (autoTimeoutCnt == 0)
		    {
			autoTimeoutCnt++;
			timeout = sendFailRepeatAckMsg ();
			deadline = time (0) + timeout + TIMEOUT_DEADBAND;
			noReturn = 0;
		    }
		    else
		    {
			sendFinalRepeatAckMsg ();
			writeDecTalk (DTHANGUP, "processAlarms");
			deadline = time (0) + TIMEOUT_DEADBAND;
			freeUnAckList ();
			currentState = WAIT_FOR_HANGUP;
			autoState = NO_AUTO_STATE;
			phoneCallFailed = 0;
			noReturn = 0;
		    }
		}
		/* a repeat of the un ack'd list has been requested */
		else if (newDecTalkMsg[0] == '#')
		{
		    /* zero the removeFromList flag */
		    zeroRemoveFlag (firstUnAck);
		    currentState = SEND_UNACKED;
		    startMsgSearch = 1;
		}
		/* an ack point has been requested */
		else if (newDecTalkMsg[0] == '*')
		{
		    if (!mgrAvailable)
		    {
			sendMgrNotAvailableMsg ();
			autoTimeoutCnt = 0;
			deadline = time (0) + TIMEOUT_DEADBAND;
			currentState = WAIT_FOR_MGR;
			noReturn = 0;
		    }
		    else
		    {
			currentState = ACK_POINTS;
		    }
		}
		else
		{
		    tisSysLog (0,0, "processAlarms: WAIT_FOR_ACK_REPEAT, unexpected msg(%s)",
			    convertDecTalkMsg (newDecTalkMsg));
		    writeDecTalk (DTSTOP, "processAlarms");
		    sendUnexpectedMsg (NO_ACK_PTS_MSG);
		    sendHangUpMsg ();
		    writeDecTalk (DTHANGUP, "processAlarms");
		    freeUnAckList ();
		    currentState = WAIT_FOR_HANGUP;;
		    autoState = NO_AUTO_STATE;
		    phoneCallFailed = 0;
		    deadline = time (0) + TIMEOUT_DEADBAND + TIMEOUT_DEADBAND;
		    noReturn = 0;
		}
		break;

	    /* processAlarms (cont) */

	    /* user has chosen to acknowledge the alarms */
	    case ACK_POINTS:
if(debug)
{
tisSysLog(0,0,
"processAlarms,ACK_POINTS");
}
		if (ackPoints ())
		{
		    sendAllPtsNotAckdMsg ();
		}
		else
		{
		    sendAckdPtsMsg ();
		}
		writeDecTalk (DTHANGUP, "processAlarms");
		currentState = WAIT_FOR_HANGUP;
		autoState = NO_AUTO_STATE;
		phoneCallFailed = 0;
		freeUnAckList ();
		deadline = time (0) + TIMEOUT_DEADBAND;
		noReturn = 0;
		break;

	    /* the alarm manager is not available, so we wait */
	    case WAIT_FOR_MGR:
if(debug)
{
tisSysLog(0,0,
"processAlarms,WAIT_FOR_MGR");
}
		if (!mgrAvailable)
		{
		    if (autoTimeoutCnt == 0)
		    {
			autoTimeoutCnt++;
			deadline = time (0) + TIMEOUT_DEADBAND + TIMEOUT_DEADBAND;
			noReturn = 0;
		    }
		    else
		    {
			sendFinalMgrNotAvailableMsg (NO_ACK_PTS_MSG);
			writeDecTalk (DTHANGUP, "processAlarms");
			freeUnAckList ();
			deadline = time (0) + TIMEOUT_DEADBAND + TIMEOUT_DEADBAND;
			currentState = WAIT_FOR_HANGUP;
			autoState = NO_AUTO_STATE;
			phoneCallFailed = 0;
			noReturn = 0;
		    }
		}
		else
		{
		    sendMgrAvailableMsg ();
		    currentState = ACK_POINTS;
		}
		break;
		
	} /* end switch (currentState) */


    }  /* end while (noReturn) */

if (debug)
{
tisSysLog (0,0,
"leave processAlarms: cState=%s,who=%s,callFailed=%d,sentAuto=%d,autoN=%d,autoId=%d,autoS=%s,deadline=%ld",
states[currentState],WhoWasCalled[whoWasCalled],
phoneCallFailed,sentToAutoCaller,autoNumber,autoIdCnt,
autoStates[autoState],deadline);
}

    return (0);
}


/*******************************************************
*
*    removeMsgFromList ()
*	remove all msgs from an alarmmsg list
*
*******************************************************/

void	removeMsgFromList (QUEMSG_PTR *firstOfList)
{
    QUEMSG_PTR	freedMsg, currentMsg, previousMsg;
    int		freeMem = 0;

    currentMsg = previousMsg = *firstOfList;
    while (currentMsg != (QUEMSG_PTR)0)
    {
	previousMsg = currentMsg;
	if (currentMsg -> removeFromList)
	{
	    /* remove msg from list */
	    if (currentMsg == (*firstOfList))
	    {
		*firstOfList = currentMsg -> nextMsg;
	    }
	    else
	    {
		previousMsg -> nextMsg = currentMsg -> nextMsg;
	    }
	    freeMem= 1;
	    freedMsg = currentMsg;
	}
	currentMsg = currentMsg -> nextMsg;
	if (freeMem)
	{
	    freeMem = 0;
	    free ((char *)freedMsg);
	}
    }
    return;
}

/***************************************************
*
*    removeTrailingSpaces ()
*	place a NULL in all TRAILING spaces
*
***************************************************/

void	removeTrailingSpaces (char *str)
{
    char	*pChar;
    int		i;

    pChar = str;
    for (i = (strlen (str) - 1); i >= 0; i--)
    {
	if (isspace (pChar[i]))
	    pChar[i] = NULL;
	else
	    break;
    }
    return;
}


/*****************************************************
*
*   setDecTalkName ()
*	called from main ()
*	check arg and set global char decTalkName[]
*	and global timeToAnnunciate
*
*****************************************************/

int	setDecTalkName (int argc, char *argv[])
{
    int		ttya = 0;
    int		ttyb = 0;
    char	*pChar;

    if (argc < 2)
    {
	tisSysLog (0,0, "usage: alarmTalk <[/dev/tty[a,b]] netname:[nnnn]>  [delaytime]");
	return (1);
    }
    if (strstr (argv[1], "ttya") != NULL)
    {
	ttya = 1;
    }
    else if (strstr (argv[1], "ttyb") != NULL)
    {
	ttyb = 1;
    }
    else if ( (pChar = strchr (argv[1], ':')) != NULL)
    {
	strcpy (decTalkName, argv[1]);
    }
    else
    {
	tisSysLog (0,0, "usage: alarmTalk <[/dev/tty[a,b]] netname:[nnnn]>  [delaytime]");
	return (1);
    }

#ifdef SOLARIS
    if (ttya)
    {
	strcpy (decTalkName, "/dev/cua/a");
    }
    else if (ttyb)
    {
	strcpy (decTalkName, "/dev/cua/b");
    }
#else
    if (ttya)
    {
	strcpy (decTalkName, "/dev/cua0");
    }
    else if (ttyb)
    {
	strcpy (decTalkName, "/dev/cua1");
    }
#endif

    if (argc == 3)
    {
	if ( (strncmp (argv[2], "32767", 5)) > 0)
	{
	    tisSysLog (0,0, "usage: delay time (%s) must be > 0 and < 32767",
		    argv[2]);
	    return (1);
	}
	timeToAnnunciate = atoi (argv[2]);
    }

if (debug)
{
tisSysLog (0,0,
"setDecTalkName: reutrn (0), %s, timeToAnnun = %d",
decTalkName,timeToAnnunciate);
}

    return (0);
}


/****************************************************
*
*    startAlarmAccess ()
*	try to connect to alarmMgr's shared memory
*
****************************************************/

int	startAlarmAccess (void)
{

    if (InitAlarmAccess ())
    {
	alarmAccessed = 1;
	mgrAvailable = 1;
    }
    else
    {
	tisSysLog (0,0, "startAlarmAccess: failed to access alarmMgr's shared memory");
	return (1);
    }

if(debug)
{
tisSysLog(0,0,
"startAlarmAccess: return (0), alarmAccessed=%d,mgrAvail=%d",
alarmAccessed,mgrAvailable);
}

    return (0);
}


/*******************************
*
*    startChannelAccess ()
*
*******************************/

int	startChannelAccess (void)
{
    if (ca_task_initialize () != AGNT_NORMAL)
    {
	tisSysLog (0,0, "startChannelAccess: failed to start ca");
	return (1);
    }
    else
	caInited = 1;

if(debug)
{
tisSysLog(0,0,
"startCA: return (0), caInited=%d",
caInited);
}

    return (0);
}


/**************************************************************
*
*    updateActiveList ()
*	we have reestablished our connection to the alarmMgr
*	shared memory, traverse the active list and fill
*	in the blanks
*
**************************************************************/

void	updateActiveList (void)
{
    QUEMSG_PTR	currentMsg;
    PvData	*pvdata;
    int		removeActive = 0;

    currentMsg = firstActive;
    while (currentMsg != (QUEMSG_PTR)0)
    {
	if (currentMsg -> mgrNotAvailable)
	{
	    if ( (pvdata = (PvData *)FindPV (NULL, currentMsg -> pointName)) != NULL)
	    {
		currentMsg -> alarmSeverity = pvdata -> severity;
		currentMsg -> alarmStatus = pvdata -> status;
		strncpy (currentMsg -> descriptor, pvdata -> description, 
			 DESCRIPTOR_SZ);
		currentMsg -> descriptor[DESCRIPTOR_SZ] = NULL;
		currentMsg -> lastAlarm = pvdata -> lastAlarm;
		currentMsg -> mgrNotAvailable = 0;
	    }
	    else
	    {
		tisSysLog (0,0, "updateActiveList: pv %s not known to alarmMgr",
			 currentMsg -> pointName);
		currentMsg -> removeFromList = 1;
		removeActive = 1;
	    }
	}
	currentMsg = currentMsg -> nextMsg;
    }
    if (removeActive)
	removeMsgFromList (&firstActive);

    return;
}


/************************************************************
*
*    verifyAutoId ()
*	if we have a valid password, then return the
*	operator name
*	NOTE: tisEncrypt is in tislib.a .../share/src/libtis
*	NOTE: if the file secureList has been modified since
*		the last time we used it, free up current list
*		and create a new one
*
*************************************************************/

int	verifyAutoId (void)
{
    char		encryptedPasswd[SLM_KEYPADID_SIZE * 2 + 1];
    char		*pChar;
    int			foundIt = 0;
    int			status = 0;
    struct stat		statbuff;
    ACCESSLIST_PTR	passWds, lastPass;

    status = stat (SECURELIST_PATH, &statbuff);
    if (status == -1)
    {
	tisSysLog (0,0, "verifyAutoId: stat on %s failed (%m)",
		SECURELIST_PATH);
	return (1);
    }
    if (statbuff.st_mtime > secureLastModified)
    {
	passWds = accessTree;
	while (passWds != (ACCESSLIST_PTR)0)
	{
	    lastPass = passWds;
	    passWds = passWds -> nextEntry;
	    free ((void *)lastPass);
	}
	accessTree = (ACCESSLIST_PTR)0;
	if (initAccessList ())
	{
	    tisSysLog (0,0, "verifyAutoId: could not reinit access list");
	    return (1);
	}
    }
    memset (encryptedPasswd, 0, SLM_KEYPADID_SIZE * 2 + 1);

    /* the decTalk sends back the # sign that the user typed
       to end their enter of their id
    */
    if ((pChar = strchr (opId, '#')) != NULL)
	*pChar = NULL;

    /* get the password that was entered */
    strncpy (encryptedPasswd, opId, SLM_KEYPADID_SIZE);
    encryptedPasswd[SLM_KEYPADID_SIZE] = NULL;

    /* encrypt it */
    tisEncrypt (encryptedPasswd, SLM_KEYPADID_SIZE);

    /* try to find in password list */
    passWds = accessTree;
    while (passWds != (ACCESSLIST_PTR)0)
    {
	/* found it, get operator name for this password */
	if ( (strcmp (passWds -> passwd, encryptedPasswd)) == 0)
	{
	    memset (opName, 0, SLM_NAME_SIZE);
	    strncpy (opName, passWds -> opname, SLM_NAME_SIZE);
	    opName[SLM_NAME_SIZE] = NULL;
	    foundIt = 1;
	    break;
	}
	passWds = passWds -> nextEntry;
    }
if (debug)
{
tisSysLog (0,0,
"verifyAutoId: opId=(%s),opName(%s),found=%d",
opId,opName,foundIt);
}
    if (!foundIt)
	return (0);
    return (1);
}


/****************************************
*
*    wrapItUp ()
*	clean up things before exiting
*
****************************************/

void	wrapItUp (void)
{
    int			i;
    time_t		end_time;
    QUEMSG_PTR		prevMsg, currMsg;
    ALARMPT_PTR		prevPoint, currPoint;
    ACCESSLIST_PTR	pAccess, lastAccess;

    if (alarmPrtFd > -1)
	close (alarmPrtFd);
    if (alarmPrtListenFd > -1)
	close (alarmPrtListenFd);
    if (alarmMgrFd > -1)
	close (alarmMgrFd);
    if (decTalkFd > -1)
    {
	/* set port back to original parameters */
	if (settty)
	{
	    /* set after ALL output written to fd has been transmitted */
	    if (tcsetattr (decTalkFd, TCSADRAIN, &savetty) < 0)
	    {
		tisSysLog (0,0, "failed to restore decTalk port (%m)");
	    }
	    /* set exclusive use off for this tty */
	    if (ioctl (decTalkFd, TIOCNXCL, &i) < 0)
	    {
		tisSysLog (0,0, "failed to turn off exclusive use of tty (%m)");
	    }
	}
	close (decTalkFd);
    }
    settty = 0;
    alarmPrtFd = -1;
    alarmPrtListenFd = -1;
    alarmMgrFd = -1;
    decTalkFd = -1;

    /* release alarmMgr's shared memory */
    if (alarmAccessed)
	UnmapSharedMemory ();
    alarmAccessed = 0;

    /* release channel access */
    if (caInited)
	ca_task_exit ();
    caInited = 0;

    /* close the alarmTalk database */
    if (dbOpened)
	d4close (alarmTalkDb);
    dbOpened = 0;
    if (firstActive != (QUEMSG_PTR) 0)
    {
	currMsg = firstActive;
	while (currMsg != (QUEMSG_PTR) 0)
	{
	    prevMsg = currMsg;
	    currMsg = currMsg -> nextMsg;
	    free ((void *)prevMsg);
	}
	firstActive = (QUEMSG_PTR)0;
    }

    freeUnAckList ();

    if (alarmTalkPts != (ALARMPT_PTR) 0)
    {
	currPoint = alarmTalkPts;
	while (currPoint != (ALARMPT_PTR) 0)
	{
	    prevPoint = currPoint;
	    currPoint = currPoint -> nextPoint;
	    free ((void *)prevPoint);
	}
	alarmTalkPts = (ALARMPT_PTR)0;
    }

    if (accessTree != (ACCESSLIST_PTR)0)
    {
	pAccess = accessTree;
	while (pAccess != (ACCESSLIST_PTR)0)
	{
	    lastAccess = pAccess;
	    pAccess = pAccess -> nextEntry;
	    free ((void *)lastAccess);
	}
	accessTree = (ACCESSLIST_PTR)0;
    }

    end_time = time(0);
    tisSysLog (0,0, "ended at %s", ctime (&end_time));

    closelog ();

    return;
}


/***************************************************
*
*   zeroRemoveFlag ()
*	reset the removeFromList flag from list
*
***************************************************/

void	zeroRemoveFlag (QUEMSG_PTR list)
{
    QUEMSG_PTR	nextMsg;

    nextMsg = list;
    while (nextMsg != (QUEMSG_PTR)0)
    {
	nextMsg -> removeFromList = 0;
	nextMsg = nextMsg -> nextMsg;
    }
    return;
}
