/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/************************************************************************
*
*    alarmTalk.h
*
*    modifications
*    -------------
*  .01	09-22-96	ems	creation date
*
************************************************************************/


#ifndef	INCalarmtalkH
#define	INCalarmtalkH

#include "atDefs.h"

#define	DEC_TALK_PORT		8500

/* define typedef structures */

typedef struct alarmPointList
{
    struct alarmPointList	*nextPoint;
    char			pointName[PVNAME_SZ + 1];
} ALARMPT, *ALARMPT_PTR;

typedef struct queueMsg
{
    struct queueMsg	*nextMsg;
    char		pointName[PVNAME_SZ + 1];
    short		alarmSeverity;
    short		alarmStatus;
    short		ack;
    char		descriptor[DESCRIPTOR_SZ + 1];
    char		value[AL_VALUE_SIZE + 1];
    char		egutag[AL_EGUTAG_SIZE + 1];
    TS_STAMP		lastAlarm;
    int			mgrNotAvailable;
    int			removeFromList;
    time_t		alarmRecvd;		/* time alarm msg recv'd */
} QUEMSG, *QUEMSG_PTR;

/* this struct defines the larmTalkDb config record as defined
   in ../share/src/dataVu/alarmTalkDbf.h
*/
typedef struct alarmConfig
{
    char	alarmSevr[ALARMSEVER_SZ + 1];	/* MAJOR,MINOR,INFO */
    char	annunciation[ANNUNCTION_SZ + 1];/* voice to use */
    int		spkrDevice;			/* 1 if send to speakerphone */
    int		pageDevice;			/* 1 if send to page phone */
    int		autoDevice;			/* 1 if auto dial */
    char	conNumber[PHONENUMBR_SZ + 1];	/* console phone number */
    int		conRedial;			/* timeout to redial console */
    char	pageNumber[PHONENUMBR_SZ + 1];	/* page phone number */
    int		pageRedial;			/* timeout to redial pager */
    char	autoNumbrs[NUM_AUTO_CALLS][PHONENUMBR_SZ + 1];	/*  auto-dial numbers */
    int		autoFailAnsTimeout;		/* timeout to dial next auto number */
    int		autoRestart;			/* timeout to restart auto-dial */
    int		idTimeout;			/* timeout waiting for user id */
    int		repackTimeout;			/* timeout waiting for repeat/ack req. */
} ALARMCFG, *ALARMCFG_PTR;

typedef struct accessList
{
    struct accessList	*nextEntry;
    char		passwd[SLM_KEYPADID_SIZE * 2 + 1]; /* encrypted password */
    char		opname[SLM_NAME_SIZE + 1];	   /* operator id */
} ACCESSLIST, *ACCESSLIST_PTR;


/* function prototypes */

int	acceptAlarmPrinter (void);
int	ackPoints (void);
void	addMsgToActiveList (QUEMSG_PTR);
void	addNewAlarmPoint (void);
void	alarmTalkSignalHandler (int);
int	answerAlarmPrinter (void);
void	buildMsg (QUEMSG_PTR, PvData *);
int	buildUnAckList (void);
int	caInit (void);
void	composeAlarmMessage (char *);
int	createAlarmPrtSocket (void);
int	connectToAlarmMgr (void);
char	*convertDecTalkMsg (char *);
int	decTalkReset (void);
int	decTalkState (void);
int	findPvName (QUEMSG_PTR);
void	fflushDecTalkQue (void);
void	freeUnAckList (void);
QUEMSG_PTR getMsgMemory (void);
QUEMSG_PTR getNextAlarmMsg (QUEMSG_PTR, int);
void	getNextDecTalkMsg (void);
int	initAccessList (void);
int	initDecTalk (void);
int	initDecTalkList (void);
int	initDecTalkPhoneticList (void);
void	initSignals (void);
int	openAlarmTalkDbf (void);
int	openNetworkDecTalk (void);
int	openSerialDecTalk (void);
int	operatorHasPriv (char *, char *);
int	parseIdMsg (void);
int	processAlarms (void);
int	readAlarmMgr (void);
int	readAlarmMsg (void);
int	readCurrentAlarmConfig (QUEMSG_PTR);
int	readDecTalk (void);
void	removeMsgFromList (QUEMSG_PTR *);
void	removeTrailingSpaces (char *);
void	sendAckdPtsMsg (void);
void	sendAlarmMsg (int);
void	sendAllPtsNotAckdMsg (void);
int	sendAutoIdMsg (void);
int	sendFailRepeatAckMsg (void);
int	sendFailToEnterIdMsg (void);
void	sendFinalBadIdMsg (void);
void	sendFinalFailedIdMsg (void);
void	sendFinalMgrNotAvailableMsg (char *);
void	sendFinalRepeatAckMsg (void);
void	sendHangUpMsg (void);
void	sendMgrAvailableMsg (void);
void	sendMgrNotAvailableMsg (void);
void	sendNoAlarmsMsg (char *);
int	sendRepeatAckMsg (void);
int	sendRepeatIdMsg (void);
void	sendUnexpectedMsg (char *);
void	sendWaitBuildListMsg (void);
int	setDecTalkName (int, char **);
int	setNoBlock (int);
int	startAlarmAccess (void);
int	startChannelAccess (void);
void	updateActiveList (void);
int	verifyAutoId (void);
void	wrapItUp (void);
int	writeDecTalk (char *, char *);
void	zeroRemoveFlag (QUEMSG_PTR);


/* define a timeout deadband value, to added to timeout needed by
   the decTalk
*/
#define	TIMEOUT_DEADBAND	10	/* in seconds */

/* define the decTalk default wait for a dial tone timeout */
#define	DIAL_TONE_TIMEOUT	30

/* define the decTalk default answer phone timeout and it range */
#define	ANSWER_PHONE_LOW_TIMEOUT	10
#define	ANSWER_PHONE_HIGH_TIMEOUT	120
#define	ANSWER_PHONE_TIMEOUT		30

/* defines for the QUEMSG.whoWasCalled field */
#define	NO_CALL_YET		0
#define	CONSOLE_CALL		1
#define	PAGE_CALL		2
#define	AUTO_CALL		3

/* define the size of the alarm msg buffer */
#define	ALARM_MSG_SIZE		4096

/* define the size of the decTalk msg buffer */
#define	DECTALK_MSG_SIZE	4096

/* define states that we are in when processing an 
   alarm msg, currentState
*/
#define	GET_NEXT_ALARM_MSG	0
#define	WHO_TO_CALL		1
#define	MAKE_A_PHONE_CALL	2
#define	WAIT_FOR_AN_ANSWER	3
#define	WAIT_TO_TRY_AGAIN	4
#define	WAIT_FOR_REPLY		5
#define	WAIT_FOR_HANGUP		6
#define	WAIT_FOR_AUTO_REPLY	7
#define	SEND_UNACKED		8
#define	AUTO_CALLER_SENT	9
#define	AUTO_REPEAT		10
#define	WAIT_FOR_ACK_REPEAT	11
#define	ACK_POINTS		12
#define	WAIT_FOR_MGR		13

/* defines for QUEMSG.autoState, were are we in
   auto call sequence
*/
#define	NO_AUTO_STATE		0
#define	AUTO_ID_MSG		1	/* waiting for return from ID req msg */
#define	BUILD_ACK_LIST		2

#ifdef WIN32
#define	PHONETICLIST_PATH	"\\tis4000db\\decTalkPhonetics.lst"
#define	SECURELIST_PATH		"\\tis4000\\operators\\secureList"
#define	TALKPTSLIST_PATH	"\\tis4000db\\alarmTalkPts.lst"
#else
#define	PHONETICLIST_PATH	"/tis4000db/decTalkPhonetics.lst"
#define	SECURELIST_PATH		"/tis4000/operators/secureList"
#define	TALKPTSLIST_PATH	"/tis4000db/alarmTalkPts.lst"
#endif WIN32


/* define messages to be received and sent to the decTalk */

#define	ESC		"\e"
#define	DTEND		"z\e\\"

/* request a soft terminal reset */
#define	DECSTR		ESC##"[!p"

/* device attribute request */
#define	DA		ESC##"[c"

/* define device status return */
#define	DASTAT		ESC##"[?83;2;3c"

/* define device status request */
#define	DSR		ESC##"[n"

/* define no malfunction return sequence */
#define	DSRNORMAL	ESC##"[n"
#define	DSRFIRST	ESC##"[?21n"
#define	DSRAFTER	ESC##"[?20n"

/* define status malfunction return */
#define	DSRFAIL		ESC##"[3n"

/* defines for DSRFAIL */
#define	DSRCOMMFAIL	ESC##"[?22n"
#define	DSROVERFLOW	ESC##"[?23n"
#define	DSRNVRFAIL	ESC##"[?24n"
#define	DSRPHONEM	ESC##"[?25n"
#define	DSRCTRLSEQ	ESC##"[?26n"
#define	DSRDECTST	ESC##"[?27n"

#define	DSRCOMMSG	"communication failure"
#define	DSROVERFLOWMSG	"input buffer overflow"
#define	DSRNVRMSG	"NVR failed"
#define	DSRPHONEMSG	"phonemic transcription error"
#define	DSRCTRLMSG	"control sequence error"
#define	DSRDECTSTMSG	"DECTST failed"


/* define the load user dictionary command and its returns
   from the decTalk
*/

#define	DTDICT		ESC##"P0;40z"	/* finished in initDecTalkPhoneticList() */
#define	DTDICTWORDOK	ESC##"P;50;"##DTEND
#define	DTDICTNOROOM	ESC##"P;50;1"##DTEND
#define	DTDICTWORDLONG	ESC##"P;50:2"##DTEND

/* defines for the management telephone escape sequences */
#define	DTPHONE		ESC##"P;60;"
#define	DTHANGUP	DTPHONE##"11"##DTEND
#define	DTKEYPAD	DTPHONE##"20"##DTEND
#define	DTNOKEYPAD	DTPHONE##"21"##DTEND
#define	DTTIMEOUT	DTPHONE##"30;"	/* finished in code, timout value follows 30; */
#define	DTPHONECALL	DTPHONE##"40zW" /* finished in code */
#define	DTWINK		DTPHONE##"50"##DTEND
#define	DTNOWINK	DTPHONE##"51"##DTEND

/* defines for answers decTalk will return as 
   as replies to the phone commands
*/
#define	DTREPLY		ESC##"P;70;"
#define	PHONHOOK	DTREPLY##""##DTEND
#define	PHOFFHOOK	DTREPLY##"1"##DTEND
#define	PHTIMEOUT	DTREPLY##"2"##DTEND
#define	PHWINK		DTREPLY##"4"##DTEND
#define	PHNODIAL	DTREPLY##"5"##DTEND
#define	PHBUSY		DTREPLY##"6"##DTEND
#define	PHNOANS		DTREPLY##"7"##DTEND

/* defines to turn on/off the keypad masking for
   "#" and "*" characters as response terminators
*/
#define	DTMASK		ESC##"P;83;3072"##DTEND
#define	DTNOMASK	ESC##"P;83;0"##DTEND

/* defines for speaking control */

#define	DTSTOP		ESC##"P;10"##DTEND
#define	DTSYNC		ESC##"P;11"##DTEND
#define	DTSPEAK		ESC##"P;12;1"##DTEND
#define	DTINDEXQUERY	ESC##"P;22"##DTEND
#define	DTINDEXREPLY	ESC##"P;32;"##DTEND

/* define the voice commands to be used */

#define	STD_MALE_VOICE_CMD	"[:np]"
#define	STD_FEMALE_VOICE_CMD	"[:nb]"
#define	DEEP_MALE_VOICE_CMD	"[:nh]"
#define	DEEP_FEMALE_VOICE_CMD	"[:nr]"


/* define the initial greeting from alarmTalk
   use the standard male voice for the greeting
*/
#define	ALARMTALK_GREETING	\
"[:np] This is the Southern Nevada water system alarm annunciator."

/* define a hang up msg */

#define	ALARMTALK_HANGUP_MSG	\
"[:np] Please hang up the phone."

/* define the user id request message */

#define	AUTO_ID_REQ_MSG		\
"[:np] To hear a list of all [ahnihk'naalihjhd] alarms, enter your \
password followed by the number sign, else hangup the phone."

/* define a fail to enter password msg */

#define	NO_ID_ENTRY		\
"[:np] You failed to enter your password."

/* define a msg that user enter a bad pasword */

#define	BAD_ID_ENTRY		\
"[:np] You entered an invalid password."

/* define a wait msg to build unack'd list */

#define	WAIT_BUILD_LIST_MSG	\
"[:np] Please wait until I build the [ahnihk'naalihjhd] alarm list."

/* define a build unack list error msg */

#define	BUILD_ERR_MSG		\
"[:np] There was an error building the list, alarms will not be listed."

/* define a no privileges alarms message */

#define	NO_PRIV_ALARMS_MSG		\
"[:np] There are no alarms for which you have [ihk'naalihj] privileges."

/* define a no alarms in unack list to sent msg */

#define	NO_ALARMS_TOBE_SENT_MSG		\
"[:np] There were no alarms in the list that I could send."

/* define the repeat/ack msg */

#define	REPEAT_ACK_MSG		\
"[:np] To repeat the list of [ahnihk'naalihjhd] alarms, depress the number key.\
 To [ihk'naalihj] the alarms, depress the star key, else hang up the phone."

/* define a failed to enter the #/* keys */

#define	FAILED_REPEAT_ACK_MSG	\
"[:np] You failed to enter the number key or the star key."

/* define a hang up the phone msg */

#define	HANGUP_THE_PHONE_MSG	\
"[:np] Please hang up the phone."

/* define an unexpected msg from decTalk msg */

#define	UNEXPECTED_MSG		\
"[:np] I have received an unexpected message from the decTalk."

/* define a can not send list msg */

#define	UNABLE_TO_SEND_LIST_MSG	\
"[:np] I can not send the [ahnihk'naalihjhd] alarm list."

/* define an ID not received msg */

#define	ID_NOT_RECV_MSG	\
"[:np] Your password was not received."

/* define a alarmmgr not available msg */

#define	MGR_NOT_AVAILABLE_MSG	\
"[:np] The alarm manager is not available, I will try again."

/* define a final alarm mgr not available msg */

#define	FINAL_MGR_NOT_AVAILABLE_MSG	\
"[:np] The alarm manager is still not available."

/* define points will not be ackd msg */

#define	NO_ACK_PTS_MSG		\
"[:np] The points can not be [ihk'naalihjhd]."

/* define an alarm mgr available msg */

#define MGR_AVAILABLE_MSG	\
"[:np] The alarm manager is now available."

/* define an all points ack'd message */

#define	ALL_PTS_ACKD_MSG	\
"[:np] All points were [ihk'naalihjhd]."

/* define an all points not acknowledged message */

#define	ALL_PTS_NOT_ACKD_MSG	\
"[:np] All points could not be [ihk'naalihjhd],\
 there was a communication problem with the alarm manager."

#endif	INCalarmtalkH
