/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/*****************************************************************
*
*    atvEdit.c
*	all functions to create form and callbacks
*
*    modifications
*    -------------
*  .01	10-21-96	ems	creation date
*
******************************************************************/


#include "atv.h"

/* strings for alarm type toggle buttons */
static char	*alarmSeverities[] = {
		"Major Alarm",
		"Minor Alarm",
		"Info Alarm",
		NULL
		};

/* widgets for other fields */
static Widget	voiceSelections[MAX_ALARM_CFGS];
static Widget	consoleButtons[MAX_ALARM_CFGS];
static Widget	consoleNumbers[MAX_ALARM_CFGS];
static Widget	consoleRedials[MAX_ALARM_CFGS];
static Widget	pagerButtons[MAX_ALARM_CFGS];
static Widget	pagerNumbers[MAX_ALARM_CFGS];
static Widget	pagerRedials[MAX_ALARM_CFGS];
static Widget	autoButtons[MAX_ALARM_CFGS];
static Widget	autoNumbers[MAX_ALARM_CFGS][NUM_AUTO_CALLS];
static Widget	autoRedials[MAX_ALARM_CFGS];
static Widget	autoRestarts[MAX_ALARM_CFGS];
static Widget	autoIdTimeouts[MAX_ALARM_CFGS];
static Widget	autoRepeatAckTimeouts[MAX_ALARM_CFGS];
static Widget	applyButtonWidget;
static Widget	dismissButtonWidget;

static XmString	noVoice, stdMaleVoice, deepMaleVoice, stdFemaleVoice, deepFemaleVoice;

void	atvApply (Widget, XtPointer, XtPointer);
void	atvButtonPushed (Widget, XtPointer, XmToggleButtonCallbackStruct *);
void	atvCreateAlarmFields (int, int, Widget);
void	atvDismiss (Widget, XtPointer, XtPointer);
void	atvDoDismiss (Widget, XtPointer, XtPointer);
void	atvLoadData (int);
void	atvNumberCheck (Widget, char *, XmTextVerifyCallbackStruct *);
void	atvSaveData (int);
void	atvVoiceOptions (Widget, XtPointer, XtPointer);
void	atvRemoveTrailingSpaces (char *);



/**********************************************************
*
*    atvCreateEdit ()
*
**********************************************************/

void	atvCreateEdit (Widget parent)
{
    XmString	title;
    Arg		args[20];
    int		n = 0;
    int		i;
    int		offset;

    /* create the bulletin board widget to which all others are
       attached
    */
    title = XmStringCreateLocalized ("Alarm Talk Maintenance");
    atvEditDialog = XmCreateBulletinBoard (parent, "ATMAINT", args, n);
    XtVaSetValues (atvEditDialog,
		    XmNnoResize, True,
		    XmNautoUnmanage, False,
		    XmNdialogTitle, title, NULL);
    XtManageChild (atvEditDialog);
    XmStringFree (title);

    for (i = 0, offset = 0; i < MAX_ALARM_CFGS; i++, offset += 235)
    {
	atvCreateAlarmFields (i, offset, atvEditDialog);
	atvLoadData (i);
    }
    offset += 5;
    applyButtonWidget = XtVaCreateManagedWidget ("ApplyButton",
						 xmPushButtonWidgetClass,
						 atvEditDialog,
						 XtVaTypedArg, 
						     XmNlabelString,
						     XmRString,
						     "Apply",
						     6,
						 XmNx, 0,
						 XmNy, offset,
						 XmNheight, 29,
						 XmNwidth, 75,
						 NULL);
    XtAddCallback (applyButtonWidget, XmNactivateCallback, atvApply, (XtPointer)0);

    dismissButtonWidget = XtVaCreateManagedWidget ("DismissButton",
						   xmPushButtonWidgetClass,
						   atvEditDialog,
						   XtVaTypedArg, 
						     XmNlabelString,
						     XmRString,
						     "Dismiss",
						     8,
						   XmNx, 100,
						   XmNy, offset,
						   XmNheight, 29,
						   XmNwidth, 75,
						   NULL);
    XtAddCallback (dismissButtonWidget, XmNactivateCallback, atvDismiss, (XtPointer)0);
}


/************************************************************
*
*    atvCreateAlarmFields ()
*	create all widgets for a particular alarm severity
*	bbWidget is the bulletin board widget that all the
*	other widgets will be attached to
*
************************************************************/

void	atvCreateAlarmFields (int idx, int yOffset, Widget bbWidget)
{
    Widget	label;
    XmString	voice;
    Arg		args[20];
    int		n;

    label = XtVaCreateManagedWidget ("severity",
				     xmLabelWidgetClass,
				     bbWidget,
				     XtVaTypedArg, 
					XmNlabelString,
					XmRString,
					alarmSeverities[idx],
					19,
				     XmNx, 20,
				     XmNy, 25 + yOffset,
				     NULL);

    voice           = XmStringCreateLocalized ("Voice:");
    noVoice         = XmStringCreateLocalized (NO_ANNUNCIATION);
    stdMaleVoice    = XmStringCreateLocalized (STD_MALE_VOICE);
    deepMaleVoice   = XmStringCreateLocalized (DEEP_MALE_VOICE);
    stdFemaleVoice  = XmStringCreateLocalized (STD_FEMALE_VOICE);
    deepFemaleVoice = XmStringCreateLocalized (DEEP_FEMALE_VOICE);

    voiceSelections[idx] = XmVaCreateSimpleOptionMenu (bbWidget,
						       "voices",
						       voice, 'V', 0, atvVoiceOptions,
						       XmVaPUSHBUTTON, noVoice, NULL, NULL, NULL,
						       XmVaPUSHBUTTON, stdMaleVoice, NULL, NULL, NULL,
						       XmVaPUSHBUTTON, deepMaleVoice, NULL, NULL, NULL,
						       XmVaPUSHBUTTON, stdFemaleVoice, NULL, NULL, NULL,
						       XmVaPUSHBUTTON, deepFemaleVoice, NULL, NULL, NULL,
						       XmNx, 140,
						       XmNy, 18 + yOffset,
						       NULL);
    XmStringFree (voice);
    XmStringFree (noVoice);
    XmStringFree (stdMaleVoice);
    XmStringFree (deepMaleVoice);
    XmStringFree (stdFemaleVoice);
    XmStringFree (deepFemaleVoice);

    XtManageChild (voiceSelections[idx]);

    consoleButtons[idx] = XtVaCreateManagedWidget ("console",
						    xmToggleButtonWidgetClass,
						    bbWidget,
						    XtVaTypedArg, 
							XmNlabelString, 
							XmRString, 
							"Console",
							8,
						    XmNx, 425,
						    XmNy, 20 + yOffset,
						    XmNindicatorSize, 15,
						    NULL);
    XtAddCallback (consoleButtons[idx], XmNvalueChangedCallback, (XtCallbackProc)atvButtonPushed, 0);

    label = XtVaCreateManagedWidget ("phone",
				     xmLabelWidgetClass,
				     bbWidget,
				     XtVaTypedArg, 
					XmNlabelString,
					XmRString,
					"Phone:",
					7,
				     XmNx, 520,
				     XmNy, 25 + yOffset,
				     NULL);
    n = 0;
    XtSetArg (args[n], XmNx, 575);
    n++;
    XtSetArg (args[n], XmNy, 20 + yOffset);
    n++;
    XtSetArg (args[n], XmNmaxLength, PHONENUMBR_SZ);
    n++;
    XtSetArg (args[n], XmNcolumns, PHONENUMBR_SZ + 1);
    n++;
    consoleNumbers[idx] = XmCreateText (bbWidget, "number", args, n);
    XtManageChild (consoleNumbers[idx]);
    XtAddCallback (consoleNumbers[idx], XmNmodifyVerifyCallback, 
		   (XtCallbackProc)atvNumberCheck, (XtPointer)0 );

    label = XtVaCreateManagedWidget ("redial",
				     xmLabelWidgetClass,
				     bbWidget,
				     XtVaTypedArg, 
					XmNlabelString,
					XmRString,
					"Redial Delay:",
					8,
				     XmNx, 705,
				     XmNy, 25 + yOffset,
				     NULL);
    n = 0;
    XtSetArg(args[n], XmNx, 815);
    n++;
    XtSetArg(args[n], XmNy, 20 + yOffset);
    n++;
    XtSetArg(args[n], XmNmaxLength, TIMEOUT_SZ);
    n++;
    XtSetArg (args[n], XmNcolumns, TIMEOUT_SZ + 1);
    n++;
    consoleRedials[idx] = XmCreateText (bbWidget, "redial", args, n);
    XtManageChild (consoleRedials[idx]);
    XtAddCallback (consoleRedials[idx], XmNmodifyVerifyCallback, 
		   (XtCallbackProc)atvNumberCheck, (XtPointer)0 );

    pagerButtons[idx] = XtVaCreateManagedWidget ("pager",
						 xmToggleButtonWidgetClass,
						 bbWidget,
						 XtVaTypedArg, 
						     XmNlabelString, 
						     XmRString, 
						     "Pager",
						     6,
						 XmNx, 425,
						 XmNy, 55 + yOffset,
						 XmNindicatorSize, 15,
						 NULL);
    XtAddCallback (pagerButtons[idx], XmNvalueChangedCallback, (XtCallbackProc)atvButtonPushed, 0);

    label = XtVaCreateManagedWidget ("phone",
				     xmLabelWidgetClass,
				     bbWidget,
				     XtVaTypedArg, 
					XmNlabelString,
					XmRString,
					"Phone:",
					7,
				     XmNx, 520,
				     XmNy, 60 + yOffset,
				     NULL);
    n = 0;
    XtSetArg (args[n], XmNx, 575);
    n++;
    XtSetArg (args[n], XmNy, 55 + yOffset);
    n++;
    XtSetArg (args[n], XmNmaxLength, PHONENUMBR_SZ);
    n++;
    XtSetArg (args[n], XmNcolumns, PHONENUMBR_SZ + 1);
    n++;
    pagerNumbers[idx] = XmCreateText (bbWidget, "number", args, n);
    XtManageChild (pagerNumbers[idx]);
    XtAddCallback (pagerNumbers[idx], XmNmodifyVerifyCallback, 
		   (XtCallbackProc)atvNumberCheck, (XtPointer)0 );

    label = XtVaCreateManagedWidget ("redial",
				     xmLabelWidgetClass,
				     bbWidget,
				     XtVaTypedArg, 
					XmNlabelString,
					XmRString,
					"Redial Delay:",
					8,
				     XmNx, 705,
				     XmNy, 60 + yOffset,
				     NULL);

    n = 0;
    XtSetArg(args[n], XmNx, 815);
    n++;
    XtSetArg(args[n], XmNy, 55 + yOffset);
    n++;
    XtSetArg(args[n], XmNmaxLength, TIMEOUT_SZ);
    n++;
    XtSetArg (args[n], XmNcolumns, TIMEOUT_SZ + 1);
    n++;
    pagerRedials[idx] = XmCreateText (bbWidget, "redial", args, n);
    XtManageChild (pagerRedials[idx]);
    XtAddCallback (pagerRedials[idx], XmNmodifyVerifyCallback, 
		   (XtCallbackProc)atvNumberCheck, (XtPointer)0 );

    autoButtons[idx] = XtVaCreateManagedWidget ("auto",
						xmToggleButtonWidgetClass,
						bbWidget,
						XtVaTypedArg, 
						    XmNlabelString, 
						    XmRString, 
						    "Auto",
						    5,
						XmNx, 425,
						XmNy, 90 + yOffset,
						XmNindicatorSize, 15,
						NULL);
    XtAddCallback (autoButtons[idx], XmNvalueChangedCallback, (XtCallbackProc)atvButtonPushed, 0);

    label = XtVaCreateManagedWidget ("phone",
				     xmLabelWidgetClass,
				     bbWidget,
				     XtVaTypedArg, 
					XmNlabelString,
					XmRString,
					"Phone:",
					7,
				     XmNx, 520,
				     XmNy, 95 + yOffset,
				     NULL);

    n = 0;
    XtSetArg (args[n], XmNx, 575);
    n++;
    XtSetArg (args[n], XmNy, 90 + yOffset);
    n++;
    XtSetArg (args[n], XmNmaxLength, PHONENUMBR_SZ);
    n++;
    XtSetArg (args[n], XmNcolumns, PHONENUMBR_SZ + 1);
    n++;
    autoNumbers[idx][0] = XmCreateText (bbWidget, "number", args, n);
    XtManageChild (autoNumbers[idx][0]);
    XtAddCallback (autoNumbers[idx][0], XmNmodifyVerifyCallback, 
		   (XtCallbackProc)atvNumberCheck, (XtPointer)0 );

    n = 0;
    XtSetArg (args[n], XmNx, 690);
    n++;
    XtSetArg (args[n], XmNy, 90 + yOffset);
    n++;
    XtSetArg (args[n], XmNmaxLength, PHONENUMBR_SZ);
    n++;
    XtSetArg (args[n], XmNcolumns, PHONENUMBR_SZ + 1);
    n++;
    autoNumbers[idx][1] = XmCreateText (bbWidget, "number", args, n);
    XtManageChild (autoNumbers[idx][1]);
    XtAddCallback (autoNumbers[idx][1], XmNmodifyVerifyCallback, 
		   (XtCallbackProc)atvNumberCheck, (XtPointer)0 );

    n = 0;
    XtSetArg (args[n], XmNx, 805);
    n++;
    XtSetArg (args[n], XmNy, 90 + yOffset);
    n++;
    XtSetArg (args[n], XmNmaxLength, PHONENUMBR_SZ);
    n++;
    XtSetArg (args[n], XmNcolumns, PHONENUMBR_SZ + 1);
    n++;
    autoNumbers[idx][2] = XmCreateText (bbWidget, "number", args, n);
    XtManageChild (autoNumbers[idx][2]);
    XtAddCallback (autoNumbers[idx][2], XmNmodifyVerifyCallback, 
		   (XtCallbackProc)atvNumberCheck, (XtPointer)0 );

    n = 0;
    XtSetArg (args[n], XmNx, 920);
    n++;
    XtSetArg (args[n], XmNy, 90 + yOffset);
    n++;
    XtSetArg (args[n], XmNmaxLength, PHONENUMBR_SZ);
    n++;
    XtSetArg (args[n], XmNcolumns, PHONENUMBR_SZ + 1);
    n++;
    autoNumbers[idx][3] = XmCreateText (bbWidget, "number", args, n);
    XtManageChild (autoNumbers[idx][3]);
    XtAddCallback (autoNumbers[idx][3], XmNmodifyVerifyCallback, 
		   (XtCallbackProc)atvNumberCheck, (XtPointer)0 );

    n = 0;
    XtSetArg (args[n], XmNx, 575);
    n++;
    XtSetArg (args[n], XmNy, 125 + yOffset);
    n++;
    XtSetArg (args[n], XmNmaxLength, PHONENUMBR_SZ);
    n++;
    XtSetArg (args[n], XmNcolumns, PHONENUMBR_SZ + 1);
    n++;
    autoNumbers[idx][4] = XmCreateText (bbWidget, "number", args, n);
    XtManageChild (autoNumbers[idx][4]);
    XtAddCallback (autoNumbers[idx][4], XmNmodifyVerifyCallback, 
		   (XtCallbackProc)atvNumberCheck, (XtPointer)0 );

    n = 0;
    XtSetArg (args[n], XmNx, 690);
    n++;
    XtSetArg (args[n], XmNy, 125 + yOffset);
    n++;
    XtSetArg (args[n], XmNmaxLength, PHONENUMBR_SZ);
    n++;
    XtSetArg (args[n], XmNcolumns, PHONENUMBR_SZ + 1);
    n++;
    autoNumbers[idx][5] = XmCreateText (bbWidget, "number", args, n);
    XtManageChild (autoNumbers[idx][5]);
    XtAddCallback (autoNumbers[idx][5], XmNmodifyVerifyCallback, 
		   (XtCallbackProc)atvNumberCheck, (XtPointer)0 );

    n = 0;
    XtSetArg (args[n], XmNx, 805);
    n++;
    XtSetArg (args[n], XmNy, 125 + yOffset);
    n++;
    XtSetArg (args[n], XmNmaxLength, PHONENUMBR_SZ);
    n++;
    XtSetArg (args[n], XmNcolumns, PHONENUMBR_SZ + 1);
    n++;
    autoNumbers[idx][6] = XmCreateText (bbWidget, "number", args, n);
    XtManageChild (autoNumbers[idx][6]);
    XtAddCallback (autoNumbers[idx][6], XmNmodifyVerifyCallback, 
		   (XtCallbackProc)atvNumberCheck, (XtPointer)0 );

    n = 0;
    XtSetArg (args[n], XmNx, 920);
    n++;
    XtSetArg (args[n], XmNy, 125 + yOffset);
    n++;
    XtSetArg (args[n], XmNmaxLength, PHONENUMBR_SZ);
    n++;
    XtSetArg (args[n], XmNcolumns, PHONENUMBR_SZ + 1);
    n++;
    autoNumbers[idx][7] = XmCreateText (bbWidget, "number", args, n);
    XtManageChild (autoNumbers[idx][7]);
    XtAddCallback (autoNumbers[idx][7], XmNmodifyVerifyCallback, 
		   (XtCallbackProc)atvNumberCheck, (XtPointer)0 );

    label = XtVaCreateManagedWidget ("next",
				     xmLabelWidgetClass,
				     bbWidget,
				     XtVaTypedArg, 
					XmNlabelString,
					XmRString,
					"Dial Next Number Delay:",
					18,
				     XmNx, 520,
				     XmNy, 165 + yOffset,
				     NULL);

    n = 0;
    XtSetArg(args[n], XmNx, 710);
    n++;
    XtSetArg(args[n], XmNy, 160 + yOffset);
    n++;
    XtSetArg(args[n], XmNmaxLength, TIMEOUT_SZ);
    n++;
    XtSetArg (args[n], XmNcolumns, TIMEOUT_SZ + 1);
    n++;
    autoRedials[idx] = XmCreateText (bbWidget, "redial", args, n);
    XtManageChild (autoRedials[idx]);
    XtAddCallback (autoRedials[idx], XmNmodifyVerifyCallback, 
		   (XtCallbackProc)atvNumberCheck, (XtPointer)0 );

    label = XtVaCreateManagedWidget ("restart",
				     xmLabelWidgetClass,
				     bbWidget,
				     XtVaTypedArg, 
					XmNlabelString,
					XmRString,
					"Restart Auto-Dial Delay:",
					19,
				     XmNx, 769,
				     XmNy, 165 + yOffset,
				     NULL);

    n = 0;
    XtSetArg(args[n], XmNx, 969);
    n++;
    XtSetArg(args[n], XmNy, 160 + yOffset);
    n++;
    XtSetArg(args[n], XmNmaxLength, TIMEOUT_SZ);
    n++;
    XtSetArg (args[n], XmNcolumns, TIMEOUT_SZ + 1);
    n++;
    autoRestarts[idx] = XmCreateText (bbWidget, "restart", args, n);
    XtManageChild (autoRestarts[idx]);
    XtAddCallback (autoRestarts[idx], XmNmodifyVerifyCallback, 
		   (XtCallbackProc)atvNumberCheck, (XtPointer)0 );

    label = XtVaCreateManagedWidget ("idrequest",
				     xmLabelWidgetClass,
				     bbWidget,
				     XtVaTypedArg, 
					XmNlabelString,
					XmRString,
					"ID Request Time-Out:",
					21,
				     XmNx, 545,
				     XmNy, 200 + yOffset,
				     NULL);

    n = 0;
    XtSetArg(args[n], XmNx, 710);
    n++;
    XtSetArg(args[n], XmNy, 195 + yOffset);
    n++;
    XtSetArg(args[n], XmNmaxLength, TIMEOUT_SZ);
    n++;
    XtSetArg (args[n], XmNcolumns, TIMEOUT_SZ + 1);
    n++;
    autoIdTimeouts[idx] = XmCreateText (bbWidget, "idrequest", args, n);
    XtManageChild (autoIdTimeouts[idx]);
    XtAddCallback (autoIdTimeouts[idx], XmNmodifyVerifyCallback, 
		   (XtCallbackProc)atvNumberCheck, (XtPointer)0 );

    label = XtVaCreateManagedWidget ("repeatack",
				     xmLabelWidgetClass,
				     bbWidget,
				     XtVaTypedArg, 
					XmNlabelString,
					XmRString,
					"Repeat/Ack Time-Out:",
					21,
				     XmNx, 802,
				     XmNy, 200 + yOffset,
				     NULL);

    n = 0;
    XtSetArg(args[n], XmNx, 969);
    n++;
    XtSetArg(args[n], XmNy, 195 + yOffset);
    n++;
    XtSetArg(args[n], XmNmaxLength, TIMEOUT_SZ);
    n++;
    XtSetArg (args[n], XmNcolumns, TIMEOUT_SZ + 1);
    n++;
    autoRepeatAckTimeouts[idx] = XmCreateText (bbWidget, "restart", args, n);
    XtManageChild (autoRepeatAckTimeouts[idx]);
    XtAddCallback (autoRepeatAckTimeouts[idx], XmNmodifyVerifyCallback, 
		   (XtCallbackProc)atvNumberCheck, (XtPointer)0 );

    label = XtVaCreateManagedWidget("sep",
				    xmSeparatorWidgetClass,
				    bbWidget,
				    XmNx, 0,
				    XmNy,235 + yOffset,
				    XmNwidth,1025,
				    NULL);

    return;
}


/*********************************************************
*
*    atvApply ()
*	callback when Apply button has been pushed
*
*********************************************************/

void	atvApply (Widget w, XtPointer client_data, XtPointer call_data)
{
    int		idx;

    if (atvIsModified)
    {
	for (idx = 0, d4top (alarmTalkDb); 
	     idx < MAX_ALARM_CFGS; 
	     idx++, d4skip (alarmTalkDb, 1))
	{
	   atvSaveData (idx);
	}
    }
    return;
}


/**************************************************************
*
*    atvButtonPushed ()
*	called if one of the toggle buttons has been pushed
*
**************************************************************/

void atvButtonPushed (Widget w, XtPointer client_data, XmToggleButtonCallbackStruct *cbs)
{
    if (cbs -> set)
    {
	atvIsModified = 1;
    }

    return;
}

/*****************************************
*
*    atvDismiss ()
*	callback from DISMISS button
*
*****************************************/

void	atvDismiss (Widget w, XtPointer client_data, XtPointer call_data)
{
    if (atvIsModified)
    {
	atvCreateAreYouSure ("Quit without saving changes ?", "ATV");
	XtRemoveAllCallbacks (atvAreYouSure, XmNokCallback);
	XtAddCallback (atvAreYouSure, XmNokCallback, atvDoDismiss, (XtPointer)0);
	return;
    }
    if (dbIsOpened)
    {
	d4close (alarmTalkDb);
    }
    exit (1);
}

/****************************************
*
*    atvDoDismiss ()
*	get the heck out of Dodge
*
****************************************/

void	atvDoDismiss (Widget w, XtPointer client_data, XtPointer call_data)
{
    if (dbIsOpened)
    {
	d4close (alarmTalkDb);
    }
    exit (1);
}


/*******************************************************
*
*    atvLoadData ()
*	load data from alarmtlk.dbf onto the screen
*
*******************************************************/

void	atvLoadData (int idx)
{
    char	sevr[ALARMSEVER_SZ + 1];
    char	voice[ANNUNCTION_SZ + 1];
    char	button[16];
    char	number[PHONENUMBR_SZ + 1];
    char	timeout[TIMEOUT_SZ + 1];
    Widget	w, op;

    if (idx == MAJOR_CFG)
    {
	sprintf (sevr, "%-*.*s\0", ALARMSEVER_SZ, ALARMSEVER_SZ, MAJOR_ALARM_STR);
    }
    else if (idx == MINOR_CFG)
    {
	sprintf (sevr, "%-*.*s\0", ALARMSEVER_SZ, ALARMSEVER_SZ, MINOR_ALARM_STR);
    }
    else if (idx == INFO_CFG)
    {
	sprintf (sevr, "%-*.*s\0", ALARMSEVER_SZ, ALARMSEVER_SZ, INFO_ALARM_STR);
    }
    else
	return;

    if ( (d4seek (alarmTalkDb, sevr)) != 0)
    {
	d4append_blank (alarmTalkDb);
	d4unlock (alarmTalkDb);
	return;
    }

    strcpy (voice, f4str (annunciation));
    atvRemoveTrailingSpaces (voice);
    if ((strcmp (voice, NO_ANNUNCIATION)) == 0)
    {
	strcpy (button, "button_0");
    }
    else if ((strcmp (voice, STD_MALE_VOICE)) == 0)
    {
	strcpy (button, "button_1");
    }
    else if ((strcmp (voice, DEEP_MALE_VOICE)) == 0)
    {
	strcpy (button, "button_2");
    }
    else if ((strcmp (voice, STD_FEMALE_VOICE)) == 0)
    {
	strcpy (button, "button_3");
    }
    else if ((strcmp (voice, DEEP_FEMALE_VOICE)) == 0)
    {
	strcpy (button, "button_4");
    }
    else
	return;
    XtVaGetValues (voiceSelections[idx], XmNsubMenuId, &w, NULL);
    op = XtNameToWidget (w, button);
    XtVaSetValues (voiceSelections[idx],
		   XmNmenuHistory, op,
		   NULL);

    if ( (f4true (spkrDevice)) == 1)
    {
	XmToggleButtonSetState (consoleButtons[idx], True, False);
    }

    strncpy (number, f4str (conNumber), PHONENUMBR_SZ);
    number[PHONENUMBR_SZ] = 0;
    atvRemoveTrailingSpaces (number);
    XmTextSetString (consoleNumbers[idx], number);

    strncpy (timeout, f4str (conRedial), TIMEOUT_SZ);
    timeout[TIMEOUT_SZ] = '\0';
    atvRemoveTrailingSpaces (timeout);
    XmTextSetString (consoleRedials[idx], timeout);

    if ( (f4true (pageDevice)) == 1)
    {
	XmToggleButtonSetState (pagerButtons[idx], True, False);
    }

    strncpy (number, f4str (pageNumber), PHONENUMBR_SZ);
    number[PHONENUMBR_SZ] = 0;
    atvRemoveTrailingSpaces (number);
    XmTextSetString (pagerNumbers[idx], number);

    strncpy (timeout, f4str (pageRedial), TIMEOUT_SZ);
    timeout[TIMEOUT_SZ] = 0;
    atvRemoveTrailingSpaces (timeout);
    XmTextSetString (pagerRedials[idx], timeout);

    if ( (f4true (autoDevice)) == 1)
    {
	XmToggleButtonSetState (autoButtons[idx], True, False);
    }

    strncpy (number, f4str (autoNumbr1), PHONENUMBR_SZ);
    number[PHONENUMBR_SZ] = 0;
    atvRemoveTrailingSpaces (number);
    XmTextSetString (autoNumbers[idx][0], number);

    strncpy (number, f4str (autoNumbr2), PHONENUMBR_SZ);
    number[PHONENUMBR_SZ] = 0;
    atvRemoveTrailingSpaces (number);
    XmTextSetString (autoNumbers[idx][1], number);

    strncpy (number, f4str (autoNumbr3), PHONENUMBR_SZ);
    number[PHONENUMBR_SZ] = 0;
    atvRemoveTrailingSpaces (number);
    XmTextSetString (autoNumbers[idx][2], number);

    strncpy (number, f4str (autoNumbr4), PHONENUMBR_SZ);
    number[PHONENUMBR_SZ] = 0;
    atvRemoveTrailingSpaces (number);
    XmTextSetString (autoNumbers[idx][3], number);

    strncpy (number, f4str (autoNumbr5), PHONENUMBR_SZ);
    number[PHONENUMBR_SZ] = 0;
    atvRemoveTrailingSpaces (number);
    XmTextSetString (autoNumbers[idx][4], number);

    strncpy (number, f4str (autoNumbr6), PHONENUMBR_SZ);
    number[PHONENUMBR_SZ] = 0;
    atvRemoveTrailingSpaces (number);
    XmTextSetString (autoNumbers[idx][5], number);

    strncpy (number, f4str (autoNumbr7), PHONENUMBR_SZ);
    number[PHONENUMBR_SZ] = 0;
    atvRemoveTrailingSpaces (number);
    XmTextSetString (autoNumbers[idx][6], number);

    strncpy (number, f4str (autoNumbr8), PHONENUMBR_SZ);
    number[PHONENUMBR_SZ] = 0;
    atvRemoveTrailingSpaces (number);
    XmTextSetString (autoNumbers[idx][7], number);

    strncpy (timeout, f4str (autoFailAnsTimeout), TIMEOUT_SZ);
    timeout[TIMEOUT_SZ] = 0;
    atvRemoveTrailingSpaces (timeout);
    XmTextSetString (autoRedials[idx], timeout);

    strncpy (timeout, f4str (autoRestart), TIMEOUT_SZ);
    timeout[TIMEOUT_SZ] = 0;
    atvRemoveTrailingSpaces (timeout);
    XmTextSetString (autoRestarts[idx], timeout);

    strncpy (timeout, f4str (idTimeout), TIMEOUT_SZ);
    timeout[TIMEOUT_SZ] = 0;
    atvRemoveTrailingSpaces (timeout);
    XmTextSetString (autoIdTimeouts[idx], timeout);

    strncpy (timeout, f4str (repackTimeout), TIMEOUT_SZ);
    timeout[TIMEOUT_SZ] = 0;
    atvRemoveTrailingSpaces (timeout);
    XmTextSetString (autoRepeatAckTimeouts[idx], timeout);

    return;
}


/***************************************************************
*
*    atvNumberCheck ()
*	callback from all number fields to test only that a
*	0 - 9 has been entered
*
***************************************************************/

void	atvNumberCheck (Widget w, char *data, XmTextVerifyCallbackStruct *cbs)
{
    if (cbs -> event == NULL)
    {
	return;
    }
    atvIsModified = 1;
#ifdef SOLARIS
    if (cbs -> text ->length == 0)
    {
	return;
    }
#else
    if (cbs -> text -> ptr == NULL)
    {
	return;
    }
#endif
    /* if not 0 - 9, will beep, dont put character in Widget */
    if (!(isdigit (cbs -> text -> ptr[0])))
    {
	cbs -> doit = False;
    }

    return;
}


/********************************************
*
*    atvSaveData ()
*	save data to alarmtlk.dbf
*
********************************************/

void	atvSaveData (int idx)
{
    Boolean	state;
    char	*pChar, *labelString;
    XmString	label;
    Widget	w;

    if (idx == MAJOR_CFG)
	f4assign (alarmSevr, MAJOR_ALARM_STR);
    else if (idx == MINOR_CFG)
	f4assign (alarmSevr, MINOR_ALARM_STR);
    else if (idx == INFO_CFG)
	f4assign (alarmSevr, INFO_ALARM_STR);

    if ( (state = XmToggleButtonGetState (consoleButtons[idx])) == True)
	f4assign_char (spkrDevice, 'T');
    else
	f4assign_char (spkrDevice, 'F');

    XtVaGetValues (voiceSelections[idx],
		   XmNmenuHistory, &w,
		   NULL);
    XtVaGetValues (w,
		   XmNlabelString, &label,
		   NULL);
    XmStringGetLtoR (label, XmSTRING_DEFAULT_CHARSET, &labelString);
    f4assign (annunciation, labelString);
    XtFree (labelString);

    pChar = XmTextGetString (consoleNumbers[idx]);
    f4assign (conNumber, pChar);
    XtFree (pChar);

    pChar = XmTextGetString (consoleRedials[idx]);
    f4assign (conRedial, pChar);
    XtFree (pChar);

    if ( (state = XmToggleButtonGetState (pagerButtons[idx])) == True)
	f4assign_char (pageDevice, 'T');
    else
	f4assign_char (pageDevice, 'F');

    pChar = XmTextGetString (pagerNumbers[idx]);
    f4assign (pageNumber, pChar);
    XtFree (pChar);

    pChar = XmTextGetString (pagerRedials[idx]);
    f4assign (pageRedial, pChar);
    XtFree (pChar);
 
    if ( (state = XmToggleButtonGetState (autoButtons[idx])) == True)
	f4assign_char (autoDevice, 'T');
    else
	f4assign_char (autoDevice, 'F');

    pChar = XmTextGetString (autoNumbers[idx][0]);
    f4assign (autoNumbr1, pChar);
    XtFree (pChar);

    pChar = XmTextGetString (autoNumbers[idx][1]);
    f4assign (autoNumbr2, pChar);
    XtFree (pChar);

    pChar = XmTextGetString (autoNumbers[idx][2]);
    f4assign (autoNumbr3, pChar);
    XtFree (pChar);

    pChar = XmTextGetString (autoNumbers[idx][3]);
    f4assign (autoNumbr4, pChar);
    XtFree (pChar);

    pChar = XmTextGetString (autoNumbers[idx][4]);
    f4assign (autoNumbr5, pChar);
    XtFree (pChar);

    pChar = XmTextGetString (autoNumbers[idx][5]);
    f4assign (autoNumbr6, pChar);
    XtFree (pChar);

    pChar = XmTextGetString (autoNumbers[idx][6]);
    f4assign (autoNumbr7, pChar);
    XtFree (pChar);

    pChar = XmTextGetString (autoNumbers[idx][7]);
    f4assign (autoNumbr8, pChar);
    XtFree (pChar);

    pChar = XmTextGetString (autoRedials[idx]);
    f4assign (autoFailAnsTimeout, pChar);
    XtFree (pChar);

    pChar = XmTextGetString (autoRestarts[idx]);
    f4assign (autoRestart, pChar);
    XtFree (pChar);

    pChar = XmTextGetString (autoIdTimeouts[idx]);
    f4assign (idTimeout, pChar);
    XtFree (pChar);

    pChar = XmTextGetString (autoRepeatAckTimeouts[idx]);
    f4assign (repackTimeout, pChar);
    XtFree (pChar);
    
    atvIsModified = 0;

    return;
}

/****************************************************************
*
*    atvVoiceOptions ()
*	called whenever any of the voice option pulldown menus
*	is called
*
****************************************************************/

void	atvVoiceOptions (Widget w, XtPointer client_data, XtPointer call_data)
{
    atvIsModified = 1;

    return;
}

/************************************
*   
*    atvRemoveTrailingSpaces ()
*
************************************/

void	atvRemoveTrailingSpaces (char *str)
{
    char	*pChar;
    int		i;

    pChar = str;
    for (i = (strlen (str) - 1); i >= 0; i--)
    {
	if (isspace (pChar[i]))
	{
	    pChar[i] = 0;
	}
	else
	    break;
    }
    return;
}
