/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#ifndef _DBASE_H_
#define	_DBASE_H_

/* Entry into the log file */

#ifdef __cplusplus
struct LogEntry {
#else
typedef struct {
#endif
	int		isAck;
	TS_STAMP	logTime;
	GroupName	groupName;
	PointName	pointName;
	Descriptor	descriptor;
	alStatus	status;
	char		severity;
	int		acked;
	int		noPrint;
	Operator	oper;
        Resource        resource;
        IOCName	        stationName;
	EguTag		eguTag;
	PvValue		value;
	TS_STAMP	ackTime;
	Comment		comment;
	char		info1 [AL_INFO_SIZE];
	char		info2 [AL_INFO_SIZE];
	char		info3 [AL_INFO_SIZE];
	char		info4 [AL_INFO_SIZE];
	char		info5 [AL_INFO_SIZE];
#ifdef __cplusplus
};
#else
} LogEntry;
#endif

#define	LOG_ALARM	0
#define	LOG_ACK		1

/* Information returned from a getNextPv call */

#ifdef __cplusplus
struct PvMapInfo {
#else
typedef struct {
#endif
	Descriptor	descriptor;
	GroupName	groupName;
	PointName	pointName;
        Resource        resName;
#ifdef __cplusplus
};
#else
} PvMapInfo;
#endif

/* Information from reslist */

#ifdef __cplusplus
struct ResListInfo {
#else
typedef struct {
#endif
        IOCName        ioc1;
        IOCName        ioc2;
#ifdef __cplusplus
};
#else
} ResListInfo;
#endif

#define VOID	void
extern	void	initializeDatabase (void);
extern	int	createGroupFile (void);
extern	int	openGroupFile (void);
extern	void	closeGroupFile (void);
extern	int	getNextGroup (char *);
extern	void	writeGroupName (char *);
extern	int	getNumGroups (void);
extern	int	createTreeFile (void);
extern	int	openTreeFile (void);
extern	void	closeTreeFile (void);
extern	int	getNextTreeEdge (char *, char *);
extern	void	writeTreeEdge (char *, char *);
extern	int	createInfoFile (void);
extern	int	openInfoFile (void);
extern	void	closeInfoFile (void);
extern	int	getInfo (char *, GroupInfo *);
extern	void	writeInfo (char *, GroupInfo *);
extern	int	openPvMapFile (void);
extern	void	closePvMapFile (void);
extern	int	getNumPVs (void);
extern	int	getNextPv (PvMapInfo *);
extern	int	getPv (char *, char *, char *, char *);
extern	int	getPvPosition (char *);
extern	char	*GetPVPointInfo ( char *, char * );
extern	void	ResetPVPoint ( void );
extern	int	openPdbXrefFile (void);
extern	void	closePdbXrefFile (void);
extern	char	*GetPdbXrefInfo ( char *, char *, char *);
extern	char	*GetDbInfo ( char *, char *, char * );
extern	char	*GetLimitValue (char *, int );
extern	int	openIOCListFile (void);
extern	void	closeIOCListFile (void);
extern	int	getNextIOCEntry (char *, int *);
extern	int	createLogFile (void);
extern	int	openLogFile (void);
extern	void	closeLogFile (void);
extern	int	getLogEntry (char *, LogEntry *, int *);
extern	int	getNextLogEntry (LogEntry *, int *);
extern	void	writeLogEntry (LogEntry *);
extern	int	writeLogMemo (char *, TS_STAMP *, char *msg);
extern	void	makeLogKeyFromTS (char *, TS_STAMP *, int, char *);
extern	void	makeLogKeyFromEntry (LogEntry *, int, char *);
extern	int	setAckInLog (char *, TS_STAMP *, TS_STAMP *);
extern  void    openResListFile (void);
extern  void    closeResListFile (void);
extern  int     getResListEntry (char *, ResListInfo *);		
#ifndef CONFIG
extern	int	createPointStateFile (void);
extern	int	deletePointState (char *);
extern	int	openPointStateFile (void);
extern	void	closePointStateFile (void);
extern	int	getPointState (char *, PvData *, int *);
extern	void	writePointState (PvData *);
#endif
extern	int	createPrinterFile (void);
extern	int	openPrinterFile (void);
extern	void	closePrinterFile (void);
extern	int	getNextPrinterGroup (PrinterGroupEntry *);
extern	void	writePrinterGroup (PrinterGroupEntry *);
extern	int	getNumPrinters (void);
extern	void	stripBlanks (char *s);
extern	void	updateDBaseFile ();

#endif

