#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vlistitemHEADER
#include vbuttonHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "alarmDefs.h"
#include "alarm.h"
#include "dbFormats.c"	// defines structures for databases
}

#include "alcfg.h"
#include "alCfgTreeListClass.hxx"
#include "alarmConfig.hxx"
#include "alCfgMainDialogClass.hxx"
#include "alCfgDatabaseClass.hxx"
#include "alCfgUtils.hxx"


/*******************************************************************
*
*    alCfgDatabaseClass constructor
*
*******************************************************************/

alCfgDatabaseClass::alCfgDatabaseClass (char *dbPath, vresource alCfgres)
{
    int		result;

    initErrors = ALCFG_SUCCESS;
    alCfgvresource = alCfgres;

    if ( (result = alCfgDatabaseInit (dbPath)) != ALCFG_SUCCESS)
	return;

    // store the default database path
    strncpy (dbasePath, dbPath, ALCFG_MAX_DATABASE_PATH);
    dbasePath [ALCFG_MAX_DATABASE_PATH - 1] = 0;
}



/*******************************************************************
*
*    alCfgDatabaseClass destructor
*
*******************************************************************/

alCfgDatabaseClass::~alCfgDatabaseClass (void)
{
    DB_UnMapDatabase (0);
}


/*******************************************************************
*
*    alCfgDatabaseClass function to init the database
*
*******************************************************************/

int alCfgDatabaseClass::alCfgDatabaseInit (char *dbPath)
{
    DB_MapDatabase (dbPath);	// map to current database path

    int createNew = vFALSE;
    if((alCfgTreeTable = DB_MapTable(ALCFG_TREE_FILE)) == 0)

    {
	if ((alCfgTreeTable = DB_CreateTable (ALCFG_TREE_FILE,
					      treeFields,
					      NULL)) == 0)
	{
	    alCfgMessageDialogCenter (NULL, ALCFG_TREEFILE_ERR);
    	    initErrors = ALCFG_ERROR;
	    return(ALCFG_ERROR);
	}
	else
	{
	    createNew = vTRUE;
	}
    }
    parentFld = DB_MapField (alCfgTreeTable, FLD_PARENT);
    childFld  = DB_MapField (alCfgTreeTable, FLD_CHILD);
    if (createNew == vTRUE)
    {
	DB_RECORD_PTR rec = DB_NewRecord (alCfgTreeTable, NULL);
	// MAIN group has no parent, it is a child of its own creation 
	DB_PutFieldNF (rec, parentFld, "");
	DB_PutFieldNF (rec, childFld, ALCFG_MAIN_GROUP);
    }

    if((alCfgInfoTable = DB_MapTable(ALCFG_INFO_FILE)) == 0)

    {
	if ((alCfgInfoTable = DB_CreateTable (ALCFG_INFO_FILE,
					      infoFields,
					      infoTags)) == 0)
	{
	    alCfgMessageDialogCenter (NULL, ALCFG_INFOFILE_ERR);
    	    initErrors = ALCFG_ERROR;
	    return(ALCFG_ERROR);
	}
    }
    mapInfoTableFields ();

    if((alCfgPrinterTable = DB_MapTable(ALCFG_PRINTER_FILE)) == 0)

    {
	if ((alCfgPrinterTable = DB_CreateTable (ALCFG_PRINTER_FILE,
						 alarmPrinterFields,
						 NULL)) == 0)
	{
	    alCfgMessageDialogCenter (NULL, ALCFG_PRINTERFILE_ERR);
    	    initErrors = ALCFG_ERROR;
	    return(ALCFG_ERROR);
	}
    }
    nameFld = DB_MapField (alCfgPrinterTable, FLD_NAME);
    int i;
    char buff[10];
    for (i = 0; i < MAX_PRINTERS_PER_GROUP; i++)
    {
	sprintf (buff, "PTR%d", i + 1);
	printers[i] = DB_MapField (alCfgPrinterTable, buff);
    }

    createNew == vFALSE;
    if((alCfgGroupTable = DB_MapTable(ALCFG_GROUP_FILE)) == 0)

    {
	if ((alCfgGroupTable = DB_CreateTable (ALCFG_GROUP_FILE,
					       groupNameFields,
					       NULL)) == 0)
	{
	    alCfgMessageDialogCenter (NULL, ALCFG_GROUPFILE_ERR);
    	    initErrors = ALCFG_ERROR;
	    return(ALCFG_ERROR);
	}
	createNew = vTRUE;
    }
    grpGrpNameFld = DB_MapField (alCfgGroupTable, FLD_GROUPNAME);
    // if new we at least need MAIN
    if (createNew)
    {
	DB_RECORD_PTR rec = DB_NewRecord (alCfgGroupTable, NULL);
	// there is at least MAIN
	DB_PutFieldNF (rec, grpGrpNameFld, ALCFG_MAIN_GROUP);
    }

    return (ALCFG_SUCCESS);
}


/*******************************************************************
*
*    alCfgDatabaseClass create info table
*
*******************************************************************/

int alCfgDatabaseClass::createInfoTable (void)
{
    DB_UnMapTable (alCfgInfoTable, 0);

    //delete old files (.dbf, mdx)
    char infoPath[ALCFG_MAX_DATABASE_PATH];
#ifndef WIN32
    sprintf (infoPath, "%s/%s.mdx\0", dbasePath, ALCFG_INFO_FILE);
#else
    sprintf (infoPath, "%s\\%s.mdx\0", dbasePath, ALCFG_INFO_FILE);
#endif WIN32
    unlink (infoPath);

#ifndef WIN32
    sprintf (infoPath, "%s/%s.dbf\0", dbasePath, ALCFG_INFO_FILE);
#else
    sprintf (infoPath, "%s\\%s.dbf\0", dbasePath, ALCFG_INFO_FILE);
#endif WIN32
    unlink (infoPath);

    // create new blank file
    if ((alCfgInfoTable = DB_CreateTable (ALCFG_INFO_FILE,
					  infoFields,
					  infoTags)) == 0)
    {
	alCfgMessageDialogCenter (NULL, ALCFG_INFOFILE_ERR);
	return (ALCFG_ERROR);
    }
    mapInfoTableFields ();
    return (ALCFG_SUCCESS);
}


/*******************************************************************
*
*    alCfgDatabaseClass get info data
*
*******************************************************************/

void alCfgDatabaseClass::getInfoData (TREENODE_PTR node)
{
    char groupName[AL_GROUPNAME_SIZE + 1];
    char seekBuff[AL_GROUPNAME_SIZE + 1];
    int  dataType;

    strncpy (groupName, node -> groupName, AL_GROUPNAME_SIZE);
    groupName[AL_GROUPNAME_SIZE] = 0;
    sprintf (seekBuff, "%-*.*s\0", AL_GROUPNAME_SIZE, AL_GROUPNAME_SIZE, groupName);
    DB_RECORD_PTR rec = DB_Seek (alCfgInfoTable, seekBuff);
    if (rec != 0)
    {
	GROUPINFO_PTR info = (GROUPINFO_PTR)(vmemAllocAndClear (sizeof (GROUPINFO)));
	node -> parameters = info;

	DB_GetFieldValue (rec, ascriptFld, info -> autoScript, &dataType);
	info -> autoScript[AL_SCRIPTNAME_SIZE] = 0;
	DB_GetFieldValue (rec, auxascriptFld, info -> auxAutoScript, &dataType);
	info -> auxAutoScript[AL_SCRIPTNAME_SIZE] = 0;
	DB_GetFieldValue (rec, mscriptFld, info -> manualScript, &dataType);
	info -> manualScript[AL_SCRIPTNAME_SIZE] = 0;
	DB_GetFieldValue (rec, auxmscriptFld, info -> auxManualScript, &dataType);
	info -> auxManualScript[AL_SCRIPTNAME_SIZE] = 0;
	DB_GetFieldValue (rec, dispnameFld, info -> displayName, &dataType);
	info -> displayName[AL_SCRIPTNAME_SIZE] = 0;
	DB_GetFieldValue (rec, helptopicFld, info -> helpTopic, &dataType);
	info -> helpTopic[AL_HELPTOPIC_SIZE] = 0;
	DB_GetFieldValue (rec, guidanceFld, info -> guidance, &dataType);
	info -> guidance[AL_GUIDANCE_SIZE] = 0;

	// copy printers for nomal and auxillary mode

	DB_GetFieldValue (rec, majprintFld, &info -> printers[MAJOR_ALARM][0], &dataType);
	info -> printers[MAJOR_ALARM][AL_PATHNAME_SIZE] = 0;
	DB_GetFieldValue (rec, minprintFld, &info -> printers[MINOR_ALARM][0], &dataType);
	info -> printers[MINOR_ALARM][AL_PATHNAME_SIZE] = 0;
	DB_GetFieldValue (rec, infoprintFld, &info -> printers[INFO_ALARM][0], &dataType);
	info -> printers[INFO_ALARM][AL_PATHNAME_SIZE] = 0;
	DB_GetFieldValue (rec, evprintFld, &info -> printers[EVENT_ALARM][0], &dataType);
	info -> printers[EVENT_ALARM][AL_PATHNAME_SIZE] = 0;
	DB_GetFieldValue (rec, normprintFld, &info -> printers[NO_ALARM][0], &dataType);
	info -> printers[NO_ALARM][AL_PATHNAME_SIZE] = 0;

	DB_GetFieldValue (rec, amajprintFld, &info -> auxPrinters[MAJOR_ALARM][0],&dataType);
	info -> auxPrinters[MAJOR_ALARM][AL_PATHNAME_SIZE] = 0;
	DB_GetFieldValue (rec, aminprintFld, &info -> auxPrinters[MINOR_ALARM][0],&dataType);
	info -> auxPrinters[MINOR_ALARM][AL_PATHNAME_SIZE] = 0;
	DB_GetFieldValue (rec, ainfoprintFld, &info -> auxPrinters[INFO_ALARM][0],&dataType);
	info -> auxPrinters[INFO_ALARM][AL_PATHNAME_SIZE] = 0;
	DB_GetFieldValue (rec, aevprintFld, &info -> auxPrinters[EVENT_ALARM][0], &dataType);
	info -> auxPrinters[EVENT_ALARM][AL_PATHNAME_SIZE] = 0;
	DB_GetFieldValue (rec, anormprintFld, &info -> auxPrinters[NO_ALARM][0], &dataType);
	info -> auxPrinters[NO_ALARM][AL_PATHNAME_SIZE] = 0;

	// copy sounds for normal and auxillary mode

	DB_GetFieldValue (rec, majsoundFld, &info -> sounds[MAJOR_ALARM][0], &dataType);
	info -> sounds[MAJOR_ALARM][AL_PATHNAME_SIZE] = 0;
	DB_GetFieldValue (rec, minsoundFld, &info -> sounds[MINOR_ALARM][0], &dataType);
	info -> sounds[MINOR_ALARM][AL_PATHNAME_SIZE] = 0;
	DB_GetFieldValue (rec, infosoundFld, &info -> sounds[INFO_ALARM][0], &dataType);
	info -> sounds[INFO_ALARM][AL_PATHNAME_SIZE] = 0;
	DB_GetFieldValue (rec, evsoundFld, &info -> sounds[EVENT_ALARM][0], &dataType);
	info -> sounds[EVENT_ALARM][AL_PATHNAME_SIZE] = 0;

	DB_GetFieldValue (rec, amajsoundFld, &info -> auxSounds[MAJOR_ALARM][0], &dataType);
	info -> auxSounds[MAJOR_ALARM][AL_PATHNAME_SIZE] = 0;
	DB_GetFieldValue (rec, aminsoundFld, &info -> auxSounds[MINOR_ALARM][0], &dataType);
	info -> auxSounds[MINOR_ALARM][AL_PATHNAME_SIZE] = 0;
	DB_GetFieldValue (rec, ainfosoundFld, &info -> auxSounds[INFO_ALARM][0], &dataType);
	info -> auxSounds[INFO_ALARM][AL_PATHNAME_SIZE] = 0;
	DB_GetFieldValue (rec, aevsoundFld, &info -> auxSounds[EVENT_ALARM][0], &dataType);
	info -> auxSounds[EVENT_ALARM][AL_PATHNAME_SIZE] = 0;

	// copy rates for normal and auxillary modes

	DB_GetFieldValue (rec, majrateFld, &info -> rates[MAJOR_ALARM][0], &dataType);
	info -> rates[MAJOR_ALARM][AL_PATHNAME_SIZE]= 0;
	DB_GetFieldValue (rec, minrateFld, &info -> rates[MINOR_ALARM][0], &dataType);
	info -> rates[MINOR_ALARM][AL_PATHNAME_SIZE]= 0;
	DB_GetFieldValue (rec, inforateFld, &info -> rates[INFO_ALARM][0], &dataType);
	info -> rates[INFO_ALARM][AL_PATHNAME_SIZE]= 0;
	DB_GetFieldValue (rec, evrateFld, &info -> rates[EVENT_ALARM][0], &dataType);
	info -> rates[EVENT_ALARM][AL_PATHNAME_SIZE]= 0;

	DB_GetFieldValue (rec, amajrateFld, &info -> auxRates[MAJOR_ALARM][0], &dataType);
	info -> auxRates[MAJOR_ALARM][AL_PATHNAME_SIZE]= 0;
	DB_GetFieldValue (rec, aminrateFld, &info -> auxRates[MINOR_ALARM][0], &dataType);
	info -> auxRates[MINOR_ALARM][AL_PATHNAME_SIZE]= 0;
	DB_GetFieldValue (rec, ainforateFld, &info -> auxRates[INFO_ALARM][0], &dataType);
	info -> auxRates[INFO_ALARM][AL_PATHNAME_SIZE]= 0;
	DB_GetFieldValue (rec, aevrateFld, &info -> auxRates[EVENT_ALARM][0], &dataType);
	info -> auxRates[EVENT_ALARM][AL_PATHNAME_SIZE]= 0;

	// one more thing to copy

	DB_GetFieldValue (rec, auxselpvFld, info -> auxSelectPV, &dataType);
	info -> auxSelectPV[AL_POINTNAME_SIZE] = 0;
    }
}


/*******************************************************************
*
*    alCfgDatabaseClass load the alarm groups
*
*******************************************************************/

void alCfgDatabaseClass::loadGroupList (vlist *groupList)
{
    DB_RECORD_PTR rec = DB_Top (alCfgGroupTable);
    char groupName[AL_GROUPNAME_SIZE + 1];
    vstr *groupNameStr;
    int dataType;
    int row = 0;

    while (rec)
    {
	DB_GetFieldValue (rec, grpGrpNameFld, groupName, &dataType);
	groupName[AL_GROUPNAME_SIZE] = 0;
	groupNameStr = vstrCloneScribed (vcharScribeSystem (groupName));
	groupList -> SetCellValue (row, 0, (const void *)groupNameStr);
	row++;
	rec = DB_SeekNext (alCfgGroupTable);
    }
    if ((groupList -> GetRowCount ()) > 0)
    {
	groupList -> Sort (vlistROW, 0, NULL);
    }
}


/*******************************************************************
*
*    alCfgDatabaseClass load the alarm printer groups
*
*******************************************************************/

void alCfgDatabaseClass::loadPrinterList (vlist *printerList)
{
    DB_RECORD_PTR rec = DB_Top (alCfgPrinterTable);

    char printerGroup[AL_GROUPNAME_SIZE + 1];
    char printer[AL_PATHNAME_SIZE + 1];
    PRINTER_ENTRY_PTR entry;
    int dataType;
    int row = 0;
    int i;
    vstr *name;

    while (rec)
    {
	entry = (PRINTER_ENTRY_PTR)(vmemAllocAndClear (sizeof (struct printerEntry)));
	DB_GetFieldValue (rec, nameFld, printerGroup, &dataType);
	printerGroup[AL_GROUPNAME_SIZE] = 0;
	name = vstrCloneScribed (vcharScribeSystem (printerGroup));
	vcharCopyBounded (name, entry -> printerGroupName, AL_GROUPNAME_SIZE);
	entry -> printerGroupName[AL_GROUPNAME_SIZE] = 0;
	vstrDestroy (name);

	// load in the printers
	for (i = 0; i < MAX_PRINTERS_PER_GROUP; i++)
	{
	    DB_GetFieldValue (rec, printers[i], printer, &dataType);
	    printer[AL_PATHNAME_SIZE] = 0;
	    name = vstrCloneScribed (vcharScribeSystem (printer));
	    vcharCopyBounded (name, &entry -> groupPrinters[i][0], AL_PATHNAME_SIZE);
	    entry -> groupPrinters[i][AL_PATHNAME_SIZE] = 0;
	    vstrDestroy (name);
	}
	printerList -> SetCellValue (row, 0, (const void *)entry);
	row++;
	rec = DB_SeekNext (alCfgPrinterTable);
    }

    if ((printerList -> GetRowCount ()) > 0)
    {
	printerList -> Sort (vlistROW, 0, alCfgComparePrinters);
    }
}


/*******************************************************************
*
*    alCfgDatabaseClass map info table fields
*
*******************************************************************/

void alCfgDatabaseClass::mapInfoTableFields (void)
{
    infoGrpNameFld = DB_MapField (alCfgInfoTable, FLD_GROUPNAME);
    ascriptFld     = DB_MapField (alCfgInfoTable, FLD_ASCRIPT);
    auxascriptFld  = DB_MapField (alCfgInfoTable, FLD_AUXASCRIPT);
    mscriptFld     = DB_MapField (alCfgInfoTable, FLD_MSCRIPT);
    auxmscriptFld  = DB_MapField (alCfgInfoTable, FLD_AUXMSCRIPT);
    dispnameFld    = DB_MapField (alCfgInfoTable, FLD_DISPNAME);
    helptopicFld   = DB_MapField (alCfgInfoTable, FLD_HELPTOPIC);
    guidanceFld    = DB_MapField (alCfgInfoTable, FLD_GUIDANCE);
    majprintFld    = DB_MapField (alCfgInfoTable, FLD_MAJPRINT);
    minprintFld    = DB_MapField (alCfgInfoTable, FLD_MINPRINT);
    infoprintFld   = DB_MapField (alCfgInfoTable, FLD_INFOPRINT);
    evprintFld     = DB_MapField (alCfgInfoTable, FLD_EVPRINT);
    normprintFld   = DB_MapField (alCfgInfoTable, FLD_NORMPRINT);
    amajprintFld   = DB_MapField (alCfgInfoTable, FLD_AMAJPRINT);
    aminprintFld   = DB_MapField (alCfgInfoTable, FLD_AMINPRINT);
    ainfoprintFld  = DB_MapField (alCfgInfoTable, FLD_AINFOPRINT);
    aevprintFld    = DB_MapField (alCfgInfoTable, FLD_AEVPRINT);
    anormprintFld  = DB_MapField (alCfgInfoTable, FLD_ANORMPRINT);
    majsoundFld    = DB_MapField (alCfgInfoTable, FLD_MAJSOUND);
    majrateFld     = DB_MapField (alCfgInfoTable, FLD_MAJRATE);
    minsoundFld    = DB_MapField (alCfgInfoTable, FLD_MINSOUND);
    minrateFld     = DB_MapField (alCfgInfoTable, FLD_MINRATE);
    infosoundFld   = DB_MapField (alCfgInfoTable, FLD_INFOSOUND);
    inforateFld    = DB_MapField (alCfgInfoTable, FLD_INFORATE);
    evsoundFld     = DB_MapField (alCfgInfoTable, FLD_EVSOUND);
    evrateFld      = DB_MapField (alCfgInfoTable, FLD_EVRATE);
    amajsoundFld   = DB_MapField (alCfgInfoTable, FLD_AMAJSOUND);
    amajrateFld    = DB_MapField (alCfgInfoTable, FLD_AMAJRATE);
    aminsoundFld   = DB_MapField (alCfgInfoTable, FLD_AMINSOUND);
    aminrateFld    = DB_MapField (alCfgInfoTable, FLD_AMINRATE);
    ainfosoundFld  = DB_MapField (alCfgInfoTable, FLD_AINFOSOUND);
    ainforateFld   = DB_MapField (alCfgInfoTable, FLD_AINFORATE);
    aevsoundFld    = DB_MapField (alCfgInfoTable, FLD_AEVSOUND);
    aevrateFld     = DB_MapField (alCfgInfoTable, FLD_AEVRATE);
    auxselpvFld    = DB_MapField (alCfgInfoTable, FLD_AUXSELPV);
}


/*******************************************************************
*
*    alCfgDatabaseClass write the alarm groups
*	close the current file, create a new one, and then write 
*	the list to the new file, remember that we need MAIN at the 
*	least
*
*******************************************************************/

void alCfgDatabaseClass::writeGroupList (vlist *groupList)
{
    // flush file to disk, and close
    DB_UnMapTable (alCfgGroupTable, 0);

    // delete old file
    char groupPath[ALCFG_MAX_DATABASE_PATH];
#ifndef WIN32
    sprintf (groupPath, "%s/%s.dbf\0", dbasePath, ALCFG_GROUP_FILE);
#else
    sprintf (groupPath, "%s\\%s.dbf\0", dbasePath, ALCFG_GROUP_FILE);
#endif WIN32
    unlink (groupPath);

    // create a blank new file
    if ((alCfgGroupTable = DB_CreateTable (ALCFG_GROUP_FILE,
					   groupNameFields,
					   NULL)) == 0)
    {
	alCfgMessageDialogCenter (NULL, ALCFG_GROUPFILE_ERR);
    }
    else
    {
	grpGrpNameFld = DB_MapField (alCfgGroupTable, FLD_GROUPNAME);

	// there is at least MAIN
	DB_RECORD_PTR rec = DB_NewRecord (alCfgGroupTable, NULL);
	DB_PutFieldNF (rec, grpGrpNameFld, ALCFG_MAIN_GROUP);

	vstr *mainGroupName;
	char groupName[AL_GROUPNAME_SIZE + 1];

	mainGroupName = vstrCloneScribed (vcharScribeSystem (ALCFG_MAIN_GROUP));

	vchar *value;
	vlistIterator iterator;
	iterator.Start (groupList);
	while (iterator.Next ())
	{
	    value = (vchar *)(iterator.GetValue ());
	    if ((vcharCompare (value, mainGroupName)) == 0)
	    {
		continue;
	    }
	    vcharExportBounded (vcharScribeSystemExport (value),
				(void *)groupName,
				AL_GROUPNAME_SIZE);
	    groupName[AL_GROUPNAME_SIZE] = 0;
	    DB_RECORD_PTR rec = DB_NewRecord (alCfgGroupTable, NULL);
	    DB_PutFieldNF (rec, grpGrpNameFld, groupName);
	}
	iterator.Finish ();

	vstrDestroy (mainGroupName);
    }
}


/*******************************************************************
*
*    alCfgDatabaseClass write an info record
*
*******************************************************************/

void alCfgDatabaseClass::writeInfoRecord (TREENODE_PTR node)
{
    GROUPINFO_PTR params = node -> parameters;
    DB_RECORD_PTR rec = DB_NewRecord (alCfgInfoTable, NULL);

    DB_PutFieldNF (rec, infoGrpNameFld, node -> groupName);
    DB_PutFieldNF (rec, ascriptFld,    params -> autoScript);
    DB_PutFieldNF (rec, auxascriptFld, params -> auxAutoScript);
    DB_PutFieldNF (rec, mscriptFld,    params -> manualScript);
    DB_PutFieldNF (rec, auxmscriptFld, params -> auxManualScript);
    DB_PutFieldNF (rec, dispnameFld,   params -> displayName);
    DB_PutFieldNF (rec, helptopicFld,  params -> helpTopic);
    DB_PutFieldNF (rec, guidanceFld,   params -> guidance);
    DB_PutFieldNF (rec, majprintFld,   &(params -> printers[MAJOR_ALARM][0]));
    DB_PutFieldNF (rec, minprintFld,   &(params -> printers[MINOR_ALARM][0]));
    DB_PutFieldNF (rec, infoprintFld,  &(params -> printers[INFO_ALARM][0]));
    DB_PutFieldNF (rec, evprintFld,    &(params -> printers[EVENT_ALARM][0]));
    DB_PutFieldNF (rec, normprintFld,  &(params -> printers[NO_ALARM][0]));
    DB_PutFieldNF (rec, amajprintFld,  &(params -> auxPrinters[MAJOR_ALARM][0]));
    DB_PutFieldNF (rec, aminprintFld,  &(params -> auxPrinters[MINOR_ALARM][0]));
    DB_PutFieldNF (rec, ainfoprintFld, &(params -> auxPrinters[INFO_ALARM][0]));
    DB_PutFieldNF (rec, aevprintFld,   &(params -> auxPrinters[EVENT_ALARM][0]));
    DB_PutFieldNF (rec, anormprintFld, &(params -> auxPrinters[NO_ALARM][0]));
    DB_PutFieldNF (rec, majsoundFld,   &(params -> sounds[MAJOR_ALARM][0]));
    DB_PutFieldNF (rec, minsoundFld,   &(params -> sounds[MINOR_ALARM][0]));
    DB_PutFieldNF (rec, infosoundFld,  &(params -> sounds[INFO_ALARM][0]));
    DB_PutFieldNF (rec, evsoundFld,    &(params -> sounds[EVENT_ALARM][0]));
    DB_PutFieldNF (rec, amajsoundFld,  &(params -> auxSounds[MAJOR_ALARM][0]));
    DB_PutFieldNF (rec, aminsoundFld,  &(params -> auxSounds[MINOR_ALARM][0]));
    DB_PutFieldNF (rec, ainfosoundFld, &(params -> auxSounds[INFO_ALARM][0]));
    DB_PutFieldNF (rec, aevsoundFld,   &(params -> auxSounds[EVENT_ALARM][0]));
    DB_PutFieldNF (rec, majrateFld,    &(params -> rates[MAJOR_ALARM][0]));
    DB_PutFieldNF (rec, minrateFld,    &(params -> rates[MINOR_ALARM][0]));
    DB_PutFieldNF (rec, inforateFld,   &(params -> rates[INFO_ALARM][0]));
    DB_PutFieldNF (rec, evrateFld,     &(params -> rates[EVENT_ALARM][0]));
    DB_PutFieldNF (rec, amajrateFld,   &(params -> auxRates[MAJOR_ALARM][0]));
    DB_PutFieldNF (rec, aminrateFld,   &(params -> auxRates[MINOR_ALARM][0]));
    DB_PutFieldNF (rec, ainforateFld,  &(params -> auxRates[INFO_ALARM][0]));
    DB_PutFieldNF (rec, aevrateFld,    &(params -> auxRates[EVENT_ALARM][0]));
    DB_PutFieldNF (rec, auxselpvFld,   params -> auxSelectPV);
}


/*******************************************************************
*
*    alCfgDatabaseClass write the alarm groups
*
*******************************************************************/

void alCfgDatabaseClass::writePrinterList (vlist *printerList)
{
    // flush file to disk and close
    DB_UnMapTable (alCfgPrinterTable, 0);

    // delete old file
    char printerPath[ALCFG_MAX_DATABASE_PATH];
#ifndef WIN32
    sprintf (printerPath, "%s/%s.dbf\0", dbasePath, ALCFG_PRINTER_FILE);
#else
    sprintf (printerPath, "%s\\%s.dbf\0", dbasePath, ALCFG_PRINTER_FILE);
#endif WIN32
    unlink (printerPath);

    // create a blank new file
    if ((alCfgPrinterTable = DB_CreateTable (ALCFG_PRINTER_FILE,
					     alarmPrinterFields,
					     NULL)) == 0)
    {
	alCfgMessageDialogCenter (NULL, ALCFG_PRINTERFILE_ERR);
    }
    else
    {
	// remap to fields
	nameFld = DB_MapField (alCfgPrinterTable, FLD_NAME);
	int i;
	char buff[10];
	for (i = 0; i < MAX_PRINTERS_PER_GROUP; i++)
	{
	    sprintf (buff, "PTR%d", i + 1);
	    printers[i] = DB_MapField (alCfgPrinterTable, buff);
	}

	char printerGroup[AL_GROUPNAME_SIZE + 1];
	char printer[AL_PATHNAME_SIZE + 1];
	PRINTER_ENTRY_PTR entry;
	vlistIterator iterator;
	iterator.Start (printerList);
	while (iterator.Next ())
	{
	    entry = (PRINTER_ENTRY_PTR)(iterator.GetValue ());
	    DB_RECORD_PTR rec = DB_NewRecord (alCfgPrinterTable, NULL);
	    vcharExportBounded (vcharScribeSystemExport (entry -> printerGroupName),
				printerGroup,
				AL_GROUPNAME_SIZE);
	    printerGroup[AL_GROUPNAME_SIZE] = 0;
	    DB_PutFieldNF (rec, nameFld, printerGroup);

	    // output the printers
	    for (i = 0; i < MAX_PRINTERS_PER_GROUP; i++)
	    {
		vcharExportBounded (vcharScribeSystemExport (&entry -> groupPrinters[i][0]),
				    printer,
				    AL_PATHNAME_SIZE);
		printer[AL_PATHNAME_SIZE] = 0;
		DB_PutFieldNF (rec, printers[i], printer);
	    }
	}
	iterator.Finish ();
    }
    
}


/*******************************************************************
*
*    alCfgDatabaseClass write the tree list
*	close the current file, create a new one
*
*******************************************************************/

int alCfgDatabaseClass::writeTreeList (void)
{
    // flush file to disk, and close
    DB_UnMapTable (alCfgTreeTable, 0);

    // delete old file
    char treePath[ALCFG_MAX_DATABASE_PATH];
#ifndef WIN32
    sprintf (treePath, "%s/%s.dbf\0", dbasePath, ALCFG_TREE_FILE);
#else
    sprintf (treePath, "%s\\%s.dbf\0", dbasePath, ALCFG_TREE_FILE);
#endif WIN32
    unlink (treePath);

    if ((alCfgTreeTable = DB_CreateTable (ALCFG_TREE_FILE,
					  treeFields,
					  NULL)) == 0)
    {
	alCfgMessageDialogCenter (NULL, ALCFG_TREEFILE_ERR);
	return(ALCFG_ERROR);
    }
    parentFld = DB_MapField (alCfgTreeTable, FLD_PARENT);
    childFld  = DB_MapField (alCfgTreeTable, FLD_CHILD);

    // we need to write MAIN here
    DB_RECORD_PTR rec = DB_NewRecord (alCfgTreeTable, NULL);
    // MAIN group has no parent, it is a child of its own creation 
    DB_PutFieldNF (rec, parentFld, "");
    DB_PutFieldNF (rec, childFld, ALCFG_MAIN_GROUP);

    return (ALCFG_SUCCESS);
}


