#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vtextitemHEADER
#include vlistitemHEADER
#include vbuttonHEADER
#include vnumHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "alarmDefs.h"
#include "alarm.h"
}

#include "alcfg.h"
#include "alCfgTreeListClass.hxx"
#include "alarmConfig.hxx"
#include "alCfgDatabaseClass.hxx"
#include "alCfgMainDialogClass.hxx"
#include "alCfgInfoDialogClass.hxx"
#include "alCfgUtils.hxx"

// filter function for printer input fields
static int alGuidFilterFunc (filteredText *text, vtextSelection *sel,
			     const vchar **newText, long *length);
static int alHelpFilterFunc (filteredText *text, vtextSelection *sel,
			     const vchar **newText, long *length);
static int alPathFilterFunc (filteredText *text, vtextSelection *sel,
			     const vchar **newText, long *length);
static int alScriptFilterFunc (filteredText *text, vtextSelection *sel,
			       const vchar **newText, long *length);


vkindTYPED_FULL_DEFINITION(alCfgInfoDialogClass, vdialog, 
						"alCfgInfoDialogClass");

/*******************************************************************
*
*    alCfgInfoDialogClass override for constructors init ()
*
*******************************************************************/

void alCfgInfoDialogClass::LoadInit (vresource res)
{
    title = NULL;
    modified = vFALSE;
    newInfo = NULL;
    treeNode = NULL;

    // get the normal mode stuff
    vcontainer *ncontainer = vcontainer::RequestSubLoad (this,
					alcfg_alCfgParamsNormalContainer);

    nMinorPrinter = filteredTextitem::RequestSubLoad (ncontainer,
					alcfg_alCfgParamsNMinorPrt);
    nMajorPrinter = filteredTextitem::RequestSubLoad (ncontainer,
					alcfg_alCfgParamsNMajorPrt);
    nInfoPrinter = filteredTextitem::RequestSubLoad (ncontainer,
					alcfg_alCfgParamsNInfoPrt);
    nEventPrinter = filteredTextitem::RequestSubLoad (ncontainer,
					alcfg_alCfgParamsNEventPrt);
    nNormalPrinter = filteredTextitem::RequestSubLoad (ncontainer,
					alcfg_alCfgParamsNNormalPrt);
    nMinorSound = filteredTextitem::RequestSubLoad (ncontainer,
					alcfg_alCfgParamsNMinorSnd);
    nMajorSound = filteredTextitem::RequestSubLoad (ncontainer,
					alcfg_alCfgParamsNMajorSnd);
    nInfoSound = filteredTextitem::RequestSubLoad (ncontainer,
					alcfg_alCfgParamsNInfoSnd);
    nEventSound = filteredTextitem::RequestSubLoad (ncontainer,
					alcfg_alCfgParamsNEventSnd);
    nMinorRate = spinnerClass::RequestSubLoad (ncontainer,
					alcfg_alCfgParamsNMinorRate);
    nMajorRate = spinnerClass::RequestSubLoad (ncontainer,
					alcfg_alCfgParamsNMajorRate);
    nInfoRate = spinnerClass::RequestSubLoad (ncontainer,
					alcfg_alCfgParamsNInfoRate);
    nEventRate = spinnerClass::RequestSubLoad (ncontainer,
					alcfg_alCfgParamsNEventRate);
    nManScript = filteredTextitem::RequestSubLoad (ncontainer,
					alcfg_alCfgParamsNManScript);
    nAutoScript = filteredTextitem::RequestSubLoad (ncontainer,
					alcfg_alCfgParamsNAutoScript);

    // get the auxiliary mode stuff
    vcontainer *acontainer = vcontainer::RequestSubLoad (this,
					alcfg_alCfgParamsAuxiliaryContainer);

    aMinorPrinter = filteredTextitem::RequestSubLoad (acontainer,
					alcfg_alCfgParamsAMinorPrt);
    aMajorPrinter = filteredTextitem::RequestSubLoad (acontainer,
					alcfg_alCfgParamsAMajorPrt);
    aInfoPrinter = filteredTextitem::RequestSubLoad (acontainer,
					alcfg_alCfgParamsAInfoPrt);
    aEventPrinter = filteredTextitem::RequestSubLoad (acontainer,
					alcfg_alCfgParamsAEventPrt);
    aNormalPrinter = filteredTextitem::RequestSubLoad (acontainer,
					alcfg_alCfgParamsANormalPrt);
    aMinorSound = filteredTextitem::RequestSubLoad (acontainer,
					alcfg_alCfgParamsAMinorSnd);
    aMajorSound = filteredTextitem::RequestSubLoad (acontainer,
					alcfg_alCfgParamsAMajorSnd);
    aInfoSound = filteredTextitem::RequestSubLoad (acontainer,
					alcfg_alCfgParamsAInfoSnd);
    aEventSound = filteredTextitem::RequestSubLoad (acontainer,
					alcfg_alCfgParamsAEventSnd);
    aMinorRate = spinnerClass::RequestSubLoad (acontainer,
					alcfg_alCfgParamsAMinorRate);
    aMajorRate = spinnerClass::RequestSubLoad (acontainer,
					alcfg_alCfgParamsAMajorRate);
    aInfoRate = spinnerClass::RequestSubLoad (acontainer,
					alcfg_alCfgParamsAInfoRate);
    aEventRate = spinnerClass::RequestSubLoad (acontainer,
					alcfg_alCfgParamsAEventRate);
    aManScript = filteredTextitem::RequestSubLoad (acontainer,
					alcfg_alCfgParamsAManScript);
    aAutoScript = filteredTextitem::RequestSubLoad (acontainer,
					alcfg_alCfgParamsAAutoScript);

    // now for some more stuff
    vcontainer *dcontainer = vcontainer::RequestSubLoad (this,
					alcfg_alCfgParamsDisplayContainer);
    displayName = filteredTextitem::RequestSubLoad (dcontainer,
					alcfg_alCfgParamsDisplayName);

    vcontainer *hcontainer = vcontainer::RequestSubLoad (this,
					alcfg_alCfgParamsHelpContainer);
    helpTopic = filteredTextitem::RequestSubLoad (hcontainer,
					alcfg_alCfgParamsHelp);

    vcontainer *mcontainer = vcontainer::RequestSubLoad (this,
					alcfg_alCfgParamsModePVContainer);
    modeSelectPV = filteredTextitem::RequestSubLoad (mcontainer,
					alcfg_alCfgParamsModePV);

    guidance = filteredTextitem::RequestSubLoad (this,
					alcfg_alCfgParamsGuidance);

    // get the butttons
    vcontainer *buttonContainer = vcontainer::RequestSubLoad (this,
					alcfg_alCfgParamsButtonContainer);

    applyButton = alCfgInfoApplyButtonClass::RequestSubLoad (buttonContainer,
					alcfg_alCfgParamsApplyButton);
    dismissButton = alCfgInfoDismissButtonClass::RequestSubLoad (buttonContainer,
					alcfg_alCfgParamsDismissButton);
    clearButton = alCfgInfoClearButtonClass::RequestSubLoad (buttonContainer,
					alcfg_alCfgParamsClearButton);
    // init inherited data
    vdialog::LoadInit (res);

    // store ptrs to dialog in the buttons
    applyButton   -> storeInfoDialog (this);
    dismissButton -> storeInfoDialog (this);
    clearButton   -> storeInfoDialog (this);

    // store filter functions
    nMinorPrinter  -> SetItemFilterFunc (alPathFilterFunc);
    nMajorPrinter  -> SetItemFilterFunc (alPathFilterFunc);
    nInfoPrinter   -> SetItemFilterFunc (alPathFilterFunc);
    nEventPrinter  -> SetItemFilterFunc (alPathFilterFunc);
    nNormalPrinter -> SetItemFilterFunc (alPathFilterFunc);
    aMinorPrinter  -> SetItemFilterFunc (alPathFilterFunc);
    aMajorPrinter  -> SetItemFilterFunc (alPathFilterFunc);
    aInfoPrinter   -> SetItemFilterFunc (alPathFilterFunc);
    aEventPrinter  -> SetItemFilterFunc (alPathFilterFunc);
    aNormalPrinter -> SetItemFilterFunc (alPathFilterFunc);

    nMinorSound -> SetItemFilterFunc (alPathFilterFunc);
    nMajorSound -> SetItemFilterFunc (alPathFilterFunc);
    nInfoSound  -> SetItemFilterFunc (alPathFilterFunc);
    nEventSound -> SetItemFilterFunc (alPathFilterFunc);
    aMinorSound -> SetItemFilterFunc (alPathFilterFunc);
    aMajorSound -> SetItemFilterFunc (alPathFilterFunc);
    aInfoSound  -> SetItemFilterFunc (alPathFilterFunc);
    aEventSound -> SetItemFilterFunc (alPathFilterFunc);

    nManScript  -> SetItemFilterFunc (alScriptFilterFunc);
    nAutoScript -> SetItemFilterFunc (alScriptFilterFunc);
    aManScript  -> SetItemFilterFunc (alScriptFilterFunc);
    aAutoScript -> SetItemFilterFunc (alScriptFilterFunc);

    displayName  -> SetItemFilterFunc (alScriptFilterFunc);
    helpTopic -> SetItemFilterFunc (alHelpFilterFunc);
    // this filter found in .../share/src/tisgalaxylib/textFilters.cxx
    modeSelectPV -> SetItemFilterFunc (pvnameFilter);
    guidance     -> SetItemFilterFunc (alGuidFilterFunc);

    // store ptr to modified flag
    nMinorPrinter  -> SetTextModifiedPtr (&this -> modified);
    nMajorPrinter  -> SetTextModifiedPtr (&this -> modified);
    nInfoPrinter   -> SetTextModifiedPtr (&this -> modified);
    nEventPrinter  -> SetTextModifiedPtr (&this -> modified);
    nNormalPrinter -> SetTextModifiedPtr (&this -> modified);
    aMinorPrinter  -> SetTextModifiedPtr (&this -> modified);
    aMajorPrinter  -> SetTextModifiedPtr (&this -> modified);
    aInfoPrinter   -> SetTextModifiedPtr (&this -> modified);
    aEventPrinter  -> SetTextModifiedPtr (&this -> modified);
    aNormalPrinter -> SetTextModifiedPtr (&this -> modified);

    nMinorSound -> SetTextModifiedPtr (&this-> modified);
    nMajorSound -> SetTextModifiedPtr (&this-> modified);
    nInfoSound  -> SetTextModifiedPtr (&this-> modified);
    nEventSound -> SetTextModifiedPtr (&this-> modified);
    aMinorSound -> SetTextModifiedPtr (&this-> modified);
    aMajorSound -> SetTextModifiedPtr (&this-> modified);
    aInfoSound  -> SetTextModifiedPtr (&this-> modified);
    aEventSound -> SetTextModifiedPtr (&this-> modified);

    nMinorRate -> SetModifiedPtr (&this -> modified);
    nMajorRate -> SetModifiedPtr (&this -> modified);
    nInfoRate  -> SetModifiedPtr (&this -> modified);
    nEventRate -> SetModifiedPtr (&this -> modified);
    aMinorRate -> SetModifiedPtr (&this -> modified);
    aMajorRate -> SetModifiedPtr (&this -> modified);
    aInfoRate  -> SetModifiedPtr (&this -> modified);
    aEventRate -> SetModifiedPtr (&this -> modified);

    nManScript  -> SetTextModifiedPtr (&this -> modified);
    nAutoScript -> SetTextModifiedPtr (&this -> modified);
    aManScript  -> SetTextModifiedPtr (&this -> modified);
    aAutoScript -> SetTextModifiedPtr (&this -> modified);

    displayName  -> SetTextModifiedPtr (&this -> modified);
    helpTopic    -> SetTextModifiedPtr (&this -> modified);
    modeSelectPV -> SetTextModifiedPtr (&this -> modified);
    guidance     -> SetTextModifiedPtr (&this -> modified);

}


/*******************************************************************
*
*    alCfgInfoDialogClass Destroy function
*
*******************************************************************/

void alCfgInfoDialogClass::Destroy (void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }
    if (title != NULL)
    {
	vstrDestroy (title);
    }
    vdialog::Destroy ();
}


/*******************************************************************
*
*    alCfgInfoDialogClass notify function
*
*******************************************************************/

void alCfgInfoDialogClass::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    alCfgInfoDialogClass Open function
*
*******************************************************************/

void alCfgInfoDialogClass::infoOpen (alCfgMainDialogClass *mnDialog, TREENODE_PTR node)
{
    mainDialog = mnDialog;
    app = mainDialog -> getApplication ();
    treeNode = node;

    vresource titlesRes = mainDialog -> getTitlesResource ();
    vstr *formattedTitle = vresourceGetString (titlesRes, alcfg_alCfgParamsTitle);
    vstr *groupName = vstrCloneScribed (vcharScribeSystem (node -> groupName));
    title = vstrCloneScribed (vcharScribeFormat (formattedTitle, groupName));
    vstrDestroy (formattedTitle);
    vstrDestroy (groupName);

    SetTitle (title);
    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);

    loadInfoDialog ();

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open();
    vevent::Process ();
}


/*******************************************************************
*
*    alCfgInfoDialogClass clear function
*
*******************************************************************/

void alCfgInfoDialogClass::clearInfoDialog (void)
{
    // clear out the normal mode stuff
    nMinorPrinter  -> SetTextScribed (vcharScribeSystem (""));
    nMajorPrinter  -> SetTextScribed (vcharScribeSystem (""));
    nInfoPrinter   -> SetTextScribed (vcharScribeSystem (""));
    nEventPrinter  -> SetTextScribed (vcharScribeSystem (""));
    nNormalPrinter -> SetTextScribed (vcharScribeSystem (""));
    nMinorSound    -> SetTextScribed (vcharScribeSystem (""));
    nMajorSound    -> SetTextScribed (vcharScribeSystem (""));
    nInfoSound     -> SetTextScribed (vcharScribeSystem (""));
    nEventSound    -> SetTextScribed (vcharScribeSystem (""));
    nMinorRate     -> SetValue (nMinorRate -> GetMin ());
    nMajorRate     -> SetValue (nMajorRate -> GetMin ());
    nInfoRate      -> SetValue (nInfoRate  -> GetMin ());
    nEventRate     -> SetValue (nEventRate -> GetMin ());
    nManScript     -> SetTextScribed (vcharScribeSystem (""));
    nAutoScript    -> SetTextScribed (vcharScribeSystem (""));

    // clear out the auxiliary mode stuff
    aMinorPrinter  -> SetTextScribed (vcharScribeSystem (""));
    aMajorPrinter  -> SetTextScribed (vcharScribeSystem (""));
    aInfoPrinter   -> SetTextScribed (vcharScribeSystem (""));
    aEventPrinter  -> SetTextScribed (vcharScribeSystem (""));
    aNormalPrinter -> SetTextScribed (vcharScribeSystem (""));
    aMinorSound    -> SetTextScribed (vcharScribeSystem (""));
    aMajorSound    -> SetTextScribed (vcharScribeSystem (""));
    aInfoSound     -> SetTextScribed (vcharScribeSystem (""));
    aEventSound    -> SetTextScribed (vcharScribeSystem (""));
    aMinorRate     -> SetValue (aMinorRate -> GetMin ());
    aMajorRate     -> SetValue (aMajorRate -> GetMin ());
    aInfoRate      -> SetValue (aInfoRate  -> GetMin ());
    aEventRate     -> SetValue (aEventRate -> GetMin ());
    aManScript     -> SetTextScribed (vcharScribeSystem (""));
    aAutoScript    -> SetTextScribed (vcharScribeSystem (""));

    // now the rest of it
    displayName  -> SetTextScribed (vcharScribeSystem (""));
    helpTopic    -> SetTextScribed (vcharScribeSystem (""));
    modeSelectPV -> SetTextScribed (vcharScribeSystem (""));
    guidance     -> SetTextScribed (vcharScribeSystem (""));
}


/*******************************************************************
*
*    alCfgInfoDialogClass load function
*
*******************************************************************/

void alCfgInfoDialogClass::loadInfoDialog (void)
{
    if (treeNode -> parameters == NULL)
    {
	return;
    }
    GROUPINFO_PTR params = treeNode -> parameters;
    vstr *value;

    // do the normal mode stuff

    // do the normal mode printers
    value = vstrCloneScribed (vcharScribeSystem (&(params -> printers[MINOR_ALARM][0])));
    nMinorPrinter -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> printers[MAJOR_ALARM][0])));
    nMajorPrinter -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> printers[INFO_ALARM][0])));
    nInfoPrinter -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> printers[EVENT_ALARM][0])));
    nEventPrinter -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> printers[NO_ALARM][0])));
    nNormalPrinter -> SetText (value);
    vstrDestroy (value);

    // get the normal mode sounds
    value = vstrCloneScribed (vcharScribeSystem (&(params -> sounds[MINOR_ALARM][0])));
    nMinorSound -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> sounds[MAJOR_ALARM][0])));
    nMajorSound -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> sounds[EVENT_ALARM][0])));
    nEventSound -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> sounds[INFO_ALARM][0])));
    nInfoSound -> SetText (value);
    vstrDestroy (value);

    // get the normal mode rates
    value = vstrCloneScribed (vcharScribeSystem (&(params -> rates[MINOR_ALARM][0])));
    if ((vcharLength (value)) == 0)
    {
	nMinorRate -> SetValue (nMinorRate -> GetMin());
    }
    else
    {
	nMinorRate -> SetValue (vnum::ParseInt (value, NULL));
    }
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> rates[MAJOR_ALARM][0])));
    if ((vcharLength (value)) == 0)
    {
	nMajorRate -> SetValue (nMajorRate -> GetMin());
    }
    else
    {
	nMajorRate -> SetValue (vnum::ParseInt (value, NULL));
    }
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> rates[INFO_ALARM][0])));
    if ((vcharLength (value)) == 0)
    {
	nInfoRate -> SetValue (nInfoRate -> GetMin());
    }
    else
    {
	nInfoRate -> SetValue (vnum::ParseInt (value, NULL));
    }
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> rates[EVENT_ALARM][0])));
    if ((vcharLength (value)) == 0)
    {
	nEventRate -> SetValue (nEventRate -> GetMin());
    }
    else
    {
	nEventRate -> SetValue (vnum::ParseInt (value, NULL));
    }
    vstrDestroy (value);

    // get normal mode script values
    value = vstrCloneScribed (vcharScribeSystem (params -> manualScript));
    nManScript -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (params -> autoScript));
    nAutoScript -> SetText (value);
    vstrDestroy (value);

    // get the auxiliary values

    // do the auxiliary mode printers
    value = vstrCloneScribed (vcharScribeSystem (&(params -> auxPrinters[MINOR_ALARM][0])));
    aMinorPrinter -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> auxPrinters[MAJOR_ALARM][0])));
    aMajorPrinter -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> auxPrinters[INFO_ALARM][0])));
    aInfoPrinter -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> auxPrinters[EVENT_ALARM][0])));
    aEventPrinter -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> auxPrinters[NO_ALARM][0])));
    aNormalPrinter -> SetText (value);
    vstrDestroy (value);

    // get the auxiliary mode sounds
    value = vstrCloneScribed (vcharScribeSystem (&(params -> auxSounds[MINOR_ALARM][0])));
    aMinorSound -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> auxSounds[MAJOR_ALARM][0])));
    aMajorSound -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> auxSounds[EVENT_ALARM][0])));
    aEventSound -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> auxSounds[INFO_ALARM][0])));
    aInfoSound -> SetText (value);
    vstrDestroy (value);

    // get the auxiliary mode rates
    value = vstrCloneScribed (vcharScribeSystem (&(params -> auxRates[MINOR_ALARM][0])));
    if ((vcharLength (value)) == 0)
    {
	aMinorRate -> SetValue (aMinorRate -> GetMin());
    }
    else
    {
	aMinorRate -> SetValue (vnum::ParseInt (value, NULL));
    }
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> auxRates[MAJOR_ALARM][0])));
    if ((vcharLength (value)) == 0)
    {
	aMajorRate -> SetValue (aMajorRate -> GetMin());
    }
    else
    {
	aMajorRate -> SetValue (vnum::ParseInt (value, NULL));
    }
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> auxRates[INFO_ALARM][0])));
    if ((vcharLength (value)) == 0)
    {
	aInfoRate -> SetValue (aInfoRate -> GetMin());
    }
    else
    {
	aInfoRate -> SetValue (vnum::ParseInt (value, NULL));
    }
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (&(params -> auxRates[EVENT_ALARM][0])));
    if ((vcharLength (value)) == 0)
    {
	aEventRate -> SetValue (aEventRate -> GetMin());
    }
    else
    {
	aEventRate -> SetValue (vnum::ParseInt (value, NULL));
    }
    vstrDestroy (value);

    // get auxiliary mode script values
    value = vstrCloneScribed (vcharScribeSystem (params -> auxManualScript));
    aManScript -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (params -> auxAutoScript));
    aAutoScript -> SetText (value);
    vstrDestroy (value);

    // get the rest of the fields
    value = vstrCloneScribed (vcharScribeSystem (params -> displayName));
    displayName -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (params -> helpTopic));
    helpTopic -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (params -> auxSelectPV));
    modeSelectPV -> SetText (value);
    vstrDestroy (value);
    value = vstrCloneScribed (vcharScribeSystem (params -> guidance));
    guidance -> SetText (value);
    vstrDestroy (value);

    modified = vFALSE;
}


/*******************************************************************
*
*    alCfgInfoDialogClass save function
*
*******************************************************************/

void alCfgInfoDialogClass::saveInfoDialog (void)
{
    if (modified == vFALSE)
    {
	return;
    }
    newInfo = (GROUPINFO_PTR)(vmemAllocAndClear (sizeof (GROUPINFO)));
    vstr *value;

    // get the normal mode values

    value = nMinorPrinter -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> printers[MINOR_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> printers[MINOR_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = nMajorPrinter -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> printers[MAJOR_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> printers[MAJOR_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = nInfoPrinter -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> printers[INFO_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> printers[INFO_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = nEventPrinter -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> printers[EVENT_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> printers[EVENT_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = nNormalPrinter -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> printers[NO_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> printers[NO_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = nMinorSound -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> sounds[MINOR_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> sounds[MINOR_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = nMajorSound -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> sounds[MAJOR_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> sounds[MAJOR_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = nInfoSound -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> sounds[INFO_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> sounds[INFO_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = nEventSound -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> sounds[EVENT_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> sounds[EVENT_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = nMinorRate -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> rates[MINOR_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> rates[MINOR_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = nMajorRate -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> rates[MAJOR_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> rates[MAJOR_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = nInfoRate -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> rates[INFO_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> rates[INFO_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = nEventRate -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> rates[EVENT_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> rates[EVENT_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = nManScript -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			      newInfo -> manualScript,
			      AL_SCRIPTNAME_SIZE);
    newInfo -> manualScript[AL_SCRIPTNAME_SIZE] = 0;
    vstrDestroy (value);
    value = nAutoScript -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			      newInfo -> autoScript,
			      AL_SCRIPTNAME_SIZE);
    newInfo -> autoScript[AL_SCRIPTNAME_SIZE] = 0;
    vstrDestroy (value);

    // get auxiliary mode values

    value = aMinorPrinter -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> auxPrinters[MINOR_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> auxPrinters[MINOR_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = aMajorPrinter -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> auxPrinters[MAJOR_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> auxPrinters[MAJOR_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = aInfoPrinter -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> auxPrinters[INFO_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> auxPrinters[INFO_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = aEventPrinter -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> auxPrinters[EVENT_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> auxPrinters[EVENT_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = aNormalPrinter -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> auxPrinters[NO_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> auxPrinters[NO_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = aMinorSound -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> auxSounds[MINOR_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> auxSounds[MINOR_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = aMajorSound -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> auxSounds[MAJOR_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> auxSounds[MAJOR_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = aInfoSound -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> auxSounds[INFO_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> auxSounds[INFO_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = aEventSound -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> auxSounds[EVENT_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> auxSounds[EVENT_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = aMinorRate -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> auxRates[MINOR_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> auxRates[MINOR_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = aMajorRate -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> auxRates[MAJOR_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> auxRates[MAJOR_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = aInfoRate -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> auxRates[INFO_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> auxRates[INFO_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = aEventRate -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			&(newInfo -> auxRates[EVENT_ALARM][0]),
			AL_PATHNAME_SIZE);
    newInfo -> auxRates[EVENT_ALARM][AL_PATHNAME_SIZE] = 0;
    vstrDestroy (value);

    value = aManScript -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			      newInfo -> auxManualScript,
			      AL_SCRIPTNAME_SIZE);
    newInfo -> auxManualScript[AL_SCRIPTNAME_SIZE] = 0;
    vstrDestroy (value);

    value = aAutoScript -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			      newInfo -> auxAutoScript,
			      AL_SCRIPTNAME_SIZE);
    newInfo -> auxAutoScript[AL_SCRIPTNAME_SIZE] = 0;
    vstrDestroy (value);

    // get the rest of the stuff

    value = displayName -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			      newInfo -> displayName,
			      AL_SCRIPTNAME_SIZE);
    newInfo -> displayName[AL_SCRIPTNAME_SIZE] = 0;
    vstrDestroy (value);

    value = helpTopic -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			      newInfo -> helpTopic,
			      AL_HELPTOPIC_SIZE);
    newInfo -> helpTopic[AL_HELPTOPIC_SIZE] = 0;
    vstrDestroy (value);

    value = modeSelectPV -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			      newInfo -> auxSelectPV,
			      AL_POINTNAME_SIZE);
    newInfo -> auxSelectPV[AL_POINTNAME_SIZE] = 0;
    vstrDestroy (value);

    value = aAutoScript -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			      newInfo -> auxAutoScript,
			      AL_SCRIPTNAME_SIZE);
    newInfo -> auxAutoScript[AL_SCRIPTNAME_SIZE] = 0;
    vstrDestroy (value);

    value = guidance -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (value),
			      newInfo -> guidance,
			      AL_GUIDANCE_SIZE);
    newInfo -> guidance[AL_GUIDANCE_SIZE] = 0;
    vstrDestroy (value);
}


/*******************************************************************
*
*    alCfgInfoDialogClass update info
*
*******************************************************************/

void alCfgInfoDialogClass::updateParameters (TREENODE_PTR node)
{
    if (newInfo == NULL)
    {
	return;
    }
    if (node -> parameters == NULL)
    {
	node -> parameters = newInfo;
    }
    else
    {
	*node -> parameters = *newInfo;
	vmemFree ((void *)newInfo);
	newInfo = NULL;
    }
    alarmTreeClass *alarmTree = app -> getAlarmTree ();
    alarmTree -> setInfoModified (vTRUE);
}


vkindTYPED_FULL_DEFINITION(alCfgInfoApplyButtonClass, vbutton, 
				"alCfgInfoApplyButtonClass");

/*******************************************************************
*
*    alCfgInfoApplyButtonClass 
*	Observer function
*
*******************************************************************/

void alCfgInfoApplyButtonClass::ObserveDialogItem (vevent *event)
{
    infoDialog -> saveInfoDialog ();
    // close dialog and stop event loop started in infoOpen ()
    infoDialog -> Close();
    vevent::StopProcessing ();
}


vkindTYPED_FULL_DEFINITION(alCfgInfoDismissButtonClass, vbutton, 
				"alCfgInfoDismissButtonClass");

/*******************************************************************
*
*    alCfgInfoDismissButtonClass 
*	Observer function
*
*******************************************************************/

void alCfgInfoDismissButtonClass::ObserveDialogItem (vevent *event)
{
    int modified = infoDialog -> getModifiedFlag ();
    if (modified == vTRUE)
    {
	int result = alCfgMessageDialogCenter (infoDialog, ALCFG_DATAMOD_MSG);
	if (result == ALCFG_FALSE)
	{
	    return;
	}
    }
    // close dialog and stop event loop started in listOpen ()
    infoDialog -> Close();
    vevent::StopProcessing ();
}


vkindTYPED_FULL_DEFINITION(alCfgInfoClearButtonClass, vbutton, 
				"alCfgInfoClearButtonClass");

/*******************************************************************
*
*    alCfgInfoClearButtonClass 
*	Observer function
*
*******************************************************************/

void alCfgInfoClearButtonClass::ObserveDialogItem (vevent *event)
{
    infoDialog -> clearInfoDialog ();
}



/*******************************************************************
*
*    filter functions
*
*******************************************************************/
static int alGuidFilterFunc (filteredText *text, vtextSelection *sel,
			     const vchar **newText, long *length)
{
    return (NCharFilter (text, sel, newText, length, AL_GUIDANCE_SIZE));
}

static int alHelpFilterFunc (filteredText *text, vtextSelection *sel,
			     const vchar **newText, long *length)
{
    return (NCharFilter (text, sel, newText, length, AL_HELPTOPIC_SIZE));
}

static int alPathFilterFunc (filteredText *text, vtextSelection *sel,
			     const vchar **newText, long *length)
{
    return (NCharFilter (text, sel, newText, length, AL_PATHNAME_SIZE));
}

static int alScriptFilterFunc (filteredText *text, vtextSelection *sel,
			       const vchar **newText, long *length)
{
    return (NCharFilter (text, sel, newText, length, AL_SCRIPTNAME_SIZE));
}
