#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vlistitemHEADER
#include vbuttonHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "alarmDefs.h"
#include "OperPriv.h"
}

#include "alcfg.h"
#include "alCfgTreeListClass.hxx"
#include "alarmConfig.hxx"
#include "alCfgDatabaseClass.hxx"
#include "alCfgInfoDialogClass.hxx"
#include "alCfgMainListClass.hxx"
#include "alCfgMainDialogClass.hxx"
#include "alCfgNewItemDialogClass.hxx"
#include "alCfgPrtDialogClass.hxx"
#include "alCfgListDialogClass.hxx"
#include "alCfgUtils.hxx"


vkindTYPED_FULL_DEFINITION(alCfgMainDialogClass, vdialog, NULL);

/*******************************************************************
*
*    alCfgMainDialogClass override for constructors init ()
*
*******************************************************************/

void alCfgMainDialogClass::LoadInit (vresource res)
{
    tisAbout = NULL;
    loadInitErrors = vFALSE;

    // store the main dialog's resource
    mainDialogResource = res;

    // RequestSubLoad returns ptr to uninited dialog object
    // gets init'd when dialog gets init'd by LoadInit()
    vmenubar *menubar = vmenubar::RequestSubLoad (this,
					alcfg_alCfgMainMenuBar);

    // handle the File menu
    vmenu *fileMenu = vmenu::RequestSubLoad (menubar,
					alcfg_alCfgMainFileMenu);

    vmenuItem *quitMenuItem = vmenuItem::RequestSubLoad (fileMenu,
					alcfg_alCfgMainFileQuit);

    
    // handle the Groups Menu
    vmenu *groupsMenu = vmenu::RequestSubLoad (menubar,
					alcfg_alCfgMainGroupsMenu);

    vmenuItem *createGrpMenuItem = vmenuItem::RequestSubLoad (groupsMenu,
					alcfg_alCfgMainGroupsCreate);
    vmenuItem *deleteGrpMenuItem = vmenuItem::RequestSubLoad (groupsMenu,
					alcfg_alCfgMainGroupsDelete);

    // handle the Tree menu
    vmenu *treeMenu = vmenu::RequestSubLoad (menubar,
					alcfg_alCfgMainTreeMenu);

    vmenuItem *insertMenuItem = vmenuItem::RequestSubLoad (treeMenu,
					alcfg_alCfgMainTreeInsert);
    vmenuItem *removeMenuItem = vmenuItem::RequestSubLoad (treeMenu,
					alcfg_alCfgMainTreeRemove);
    vmenuItem *cutMenuItem = vmenuItem::RequestSubLoad (treeMenu,
					alcfg_alCfgMainTreeCut);
    pasteMenuItem = vmenuItem::RequestSubLoad (treeMenu,
					alcfg_alCfgMainTreePaste);

    // handle the Parameters menu
    vmenu *paramsMenu = vmenu::RequestSubLoad (menubar,
					alcfg_alCfgMainParamsMenu);

    vmenuItem *modMenuItem = vmenuItem::RequestSubLoad (paramsMenu,
					alcfg_alCfgMainParamsMod);
    vmenuItem *clearMenuItem = vmenuItem::RequestSubLoad (paramsMenu,
					alcfg_alCfgMainParamsClear);

    // handle the Printers menu
    vmenu *printerMenu = vmenu::RequestSubLoad (menubar,
					alcfg_alCfgMainPrinterMenu);

    vmenuItem *editPrintMenuItem = vmenuItem::RequestSubLoad (printerMenu,
					alcfg_alCfgMainPrinterEdit);
    vmenuItem *createPrintMenuItem = vmenuItem::RequestSubLoad (printerMenu,
					alcfg_alCfgMainPrinterCreate);
    vmenuItem *deletePrintMenuItem = vmenuItem::RequestSubLoad (printerMenu,
					alcfg_alCfgMainPrinterDelete);

    helpMenu = vmenu::RequestSubLoad (menubar,
					alcfg_alCfgMainHelpMenu);

    // load the Main dialogs vlistitem
    mainDialogList = alCfgMainListClass::RequestSubLoad (this,
					alcfg_alCfgMainList);
    // init inherited data
    vdialog::LoadInit (res);

    // check on privileges
    // init privilege checking, see .../share/src/libtis/OperPrvtrk.c
    if ((InitOperPriv (NULL)) != PRIVOK)
    {
	alCfgMessageDialogCenter (NULL, ALCFG_INIT_OPERPRIV_ERR);
	loadInitErrors = vTRUE;
	return;
    }
    // do the groups menu
    int	allDisabled = 0;
    if ((OperPriv (NULL, "*MMI_ALARMCONFIG_GROUPS_CREATE")) == 0)
    {
	createGrpMenuItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMCONFIG_GROUPS_DELETE")) == 0)
    {
	deleteGrpMenuItem -> Disable ();
	allDisabled++;
    }
    if (allDisabled == 2)
    {
	groupsMenu -> Disable ();
    }
    // do the tree menu
    allDisabled = 0;
    if ((OperPriv (NULL, "*MMI_ALARMCONFIG_TREE_INSERT")) == 0)
    {
	insertMenuItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMCONFIG_TREE_REMOVE")) == 0)
    {
	removeMenuItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMCONFIG_TREE_CUT")) == 0)
    {
	cutMenuItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMCONFIG_TREE_PASTE")) == 0)
    {
	pasteMenuItem -> Disable ();
	allDisabled++;
    }
    if (allDisabled == 4)
    {
	treeMenu -> Disable ();
    }
    // do the parameters menu
    allDisabled = 0;
    if ((OperPriv (NULL, "*MMI_ALARMCONFIG_PARAMS_MODIFY")) == 0)
    {
	modMenuItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMCONFIG_PARAMS_CLEAR")) == 0)
    {
	clearMenuItem -> Disable ();
	allDisabled++;
    }
    if (allDisabled == 2)
    {
	paramsMenu -> Disable ();
    }
    // do the printer menu
    allDisabled = 0;
    if ((OperPriv (NULL, "*MMI_ALARMCONFIG_PRINTER_EDIT")) == 0)
    {
	editPrintMenuItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMCONFIG_PRINTER_CREATE")) == 0)
    {
	createPrintMenuItem -> Disable ();
	allDisabled++;
    }
    if ((OperPriv (NULL, "*MMI_ALARMCONFIG_PRINTER_DELETE")) == 0)
    {
	deletePrintMenuItem -> Disable ();
	allDisabled++;
    }
    if (allDisabled == 3)
    {
	printerMenu -> Disable ();
    }

    // close privilege checking, see .../share/src/libtis/OperPrvtrk.c
    CloseOperPriv ();

    mainDialogList -> storeMainDialog (this);

    // store internal names for the file main menu tags
    commandNames[QUIT_MENUITEM] = alcfg_alCfgMainFileQuit;

    // store internal names for the groups menu tags
    commandNames[NEW_GROUP_MENUITEM] = alcfg_alCfgMainGroupsCreate;
    commandNames[DELETE_GROUP_MENUITEM] = alcfg_alCfgMainGroupsDelete;

    // store internal names for the tree menu tags
    commandNames[INSERT_GROUP_MENUITEM] = alcfg_alCfgMainTreeInsert;
    commandNames[REMOVE_GROUP_MENUITEM] = alcfg_alCfgMainTreeRemove;
    commandNames[CUT_SUBTREE_MENUITEM] = alcfg_alCfgMainTreeCut;
    commandNames[PASTE_CLIPBOARD_MENUITEM] = alcfg_alCfgMainTreePaste;

    // store internal names for the parameters menu tags
    commandNames[MOD_PARAMS_MENUITEM] = alcfg_alCfgMainParamsMod;
    commandNames[CLEAR_PARAMS_MENUITEM] = alcfg_alCfgMainParamsClear;

    // store internal names for the printer menu tags
    commandNames[EDIT_PRINTER_MENUITEM] = alcfg_alCfgMainPrinterEdit;
    commandNames[CREATE_PRINTER_MENUITEM] = alcfg_alCfgMainPrinterCreate;
    commandNames[DELETE_PRINTER_MENUITEM] = alcfg_alCfgMainPrinterDelete;

    // create a command space
    vcommandSpace *space = new vcommandSpace;

    // create a dictionary with a default hash size and hash function
    vdict *dict = new vdict (vnameHash);

    int i;

    // populate our dictionary with menu commands
    for (i = QUIT_MENUITEM; i < MENUITEM_COUNT; i++)
    {
	// get an instance of our subclassed selector command
	alCfgMainCommand *command = new alCfgMainCommand;
	// set the commands selector value
	command -> SetSelector(i);
	// store command in the dictionary using the 
	// intern'd tag name
	dict -> SetElem (commandNames[i], command);
    }
    // add dictionary to the command space and give the command
    // space ownership of the dictionary, vname_Command defines
    // the type attribute of the things in the dictionary
    space -> AddDictOwned (vname_Command, dict);

    SetSpaceOwned (space);
}


/*******************************************************************
*
*    alCfgMainDialogClass Destroy function
*
*******************************************************************/

void alCfgMainDialogClass::Destroy (void)
{
    if (tisAbout != NULL)
    {
	delete tisAbout;
    }
    vdialog::Destroy ();
}


/*******************************************************************
*
*    alCfgMainDialogClass notify function
*
*******************************************************************/

void alCfgMainDialogClass::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    alCfgMainDialogClass Open function
*
*******************************************************************/

void alCfgMainDialogClass::Open (void)
{
    BuildAboutMenu ();
    pasteMenuItem -> Disable ();
    mainDialogList -> setButtonUpFlag (vFALSE);
    mainDialogList -> InsertMain ();

    Place (GetRoot (), vrectPLACE_LEFT, vrectPLACE_TOP);

    vdialog::Open ();
}

/*******************************************************************
*
*    alCfgMainDialogClass Build About menuitem function
*
*******************************************************************/

void alCfgMainDialogClass::BuildAboutMenu (void)
{
    vresource appRes = alCfgApp -> getAlCfgvresource ();
    vresource msgRes = vresourceGet (appRes, alcfg_alCfgMessagesDict);
    vstr *verMsg = vresourceGetString (msgRes, alcfg_VersionMsg);
    vresource tisGalRes = alCfgApp -> getTisGalaxyResource ();
    tisAbout = new TisAbout (tisGalRes, verMsg);
    vstrDestroy (verMsg);
    vresource titleRes = vresourceGet (appRes, alcfg_alCfgTitlesDict);
    vstr *aboutTitle = vresourceGetString (titleRes, alcfg_AboutTitle);
    tisAbout -> SetAboutTitle (aboutTitle);
    vstrDestroy (aboutTitle);
    AboutMenuItem *aboutItem = tisAbout -> GetAboutMenuItem ();
    helpMenu -> AppendItem (aboutItem);
}


vkindTYPED_FULL_DEFINITION(alCfgMainCommand, vcommandSelector, NULL);
/*******************************************************************
*
*    vuMainCommand menu selector commands
*
*******************************************************************/

int alCfgMainCommand::Issue (vdict *dict)
{
    int				result;
    int				command;
    alCfgMainDialogClass	*mainDialog;

    mainDialog = (alCfgMainDialogClass *) dict -> GetElem (vname_Dialog);
    alCfgApplication *app = mainDialog -> getApplication ();
    // get applications resource
    vresource appRes = app -> getAlCfgvresource ();
    vresource titlesRes = mainDialog -> getTitlesResource ();
    command = GetSelector ();
    switch (command)
    {
	// File menu items

	case QUIT_MENUITEM:
	{
	    int quit = vFALSE;
	    alarmTreeClass *alarmTree = app -> getAlarmTree ();
	    if ((alarmTree -> clipBoardActive ()) == vTRUE)
	    {
		TREENODE_PTR node = alarmTree -> getClipBoard();
		vstr *group = vstrCloneScribed (vcharScribeSystem (node-> groupName));
		result = alCfgMessageDialogCenter (mainDialog, ALCFG_CLIP_QUIT_MSG, group);
		if (result == ALCFG_TRUE)
		{
		    quit = vTRUE;
		}
		vstrDestroy (group);
	    }
	    else
	    {
		result = alCfgMessageDialogCenter (mainDialog, ALCFG_QUIT_MSG);
		if (result == ALCFG_TRUE)
		{
		    quit = vTRUE;
		}
	    }
	    if (quit == vTRUE)
	    {
		vevent::StopProcessing ();
	    }
	    break;
	}

	// Group menu items

	case NEW_GROUP_MENUITEM:
	{
	    vresource newRes = vresourceGet (appRes, alcfg_alCfgNewItemDialog);
	    alCfgNewItemDialogClass *newDialog = new alCfgNewItemDialogClass (newRes);
	    newDialog -> newOpen (mainDialog, NEW_GROUP_FLAG);
	    delete newDialog;
	    break;
	}

	case DELETE_GROUP_MENUITEM:
	{
	    vresource listRes = vresourceGet (appRes, alcfg_alCfgListDialog);
	    alCfgListDialogClass *listDialog = new alCfgListDialogClass (listRes);
	    listDialog -> listOpen (mainDialog, REMOVE_GROUP_FLAG);
	    delete listDialog;
	    break;
	}

	// Tree menu items

	case INSERT_GROUP_MENUITEM:
	{
	    vresource listRes = vresourceGet (appRes, alcfg_alCfgListDialog);
	    alCfgListDialogClass *listDialog = new alCfgListDialogClass (listRes);
	    listDialog -> listOpen (mainDialog, INSERT_GROUP_FLAG);
	    vchar selectedName[ALCFG_SELECTEDNAME_SIZE + 1];
	    vcharCopyBounded (listDialog -> getSelectedName (), 
			      selectedName, ALCFG_SELECTEDNAME_SIZE);
	    selectedName[ALCFG_SELECTEDNAME_SIZE] = 0;
	    delete listDialog;
	    if ((vcharLength (selectedName)) != 0)
	    {
		alCfgMainListClass *mainDialogList = mainDialog -> getListView ();
		mainDialogList -> InsertGroup (selectedName);
	    }
	    break;
	}

	case REMOVE_GROUP_MENUITEM:
	{
	    alCfgMainListClass *mainDialogList = mainDialog -> getListView ();
	    mainDialogList -> DeleteGroup ();
	    break;
	}

	case CUT_SUBTREE_MENUITEM:
	{
	    alCfgMainListClass *mainDialogList = mainDialog -> getListView ();
	    mainDialogList -> CutGroup ();
	    break;
	}

	case PASTE_CLIPBOARD_MENUITEM:
	{
	    alCfgMainListClass *mainDialogList = mainDialog -> getListView ();
	    mainDialogList -> PasteGroup ();
	    break;
	}

	// Parameters menu items

	case MOD_PARAMS_MENUITEM:
	{
	    alCfgMainListClass *mainDialogList = mainDialog -> getListView ();
	    vlist *list = mainDialogList -> GetList ();
	    vlistSelection *selected = mainDialogList -> GetSelection ();
	    int row, column;
	    list -> GetSelectedCell (selected, &row, &column);
	    if (row == vlistNO_CELLS_SELECTED)
	    {
		alCfgMessageDialogCenter (mainDialog, ALCFG_NO_NODE_SEL_ERR);
	    }
	    else
	    {
		listEntry *entry = (listEntry *)(list -> GetCellValue (row, 0));
		TREENODE_PTR node = entry -> getTreeNode ();
		vresource infoRes = vresourceGet (appRes, alcfg_alCfgParamsDialog);
		alCfgInfoDialogClass *infoDialog = new alCfgInfoDialogClass (infoRes);
		infoDialog -> infoOpen (mainDialog, node);
		infoDialog -> updateParameters (node);
		delete infoDialog;
	    }
	    break;
	}

	case CLEAR_PARAMS_MENUITEM:
	{
	    alCfgMainListClass *mainDialogList = mainDialog -> getListView ();
	    vlist *list = mainDialogList -> GetList ();
	    vlistSelection *selected = mainDialogList -> GetSelection ();
	    int row, column;
	    list -> GetSelectedCell (selected, &row, &column);
	    if (row == vlistNO_CELLS_SELECTED)
	    {
		alCfgMessageDialogCenter (mainDialog, ALCFG_NO_NODE_SEL_ERR);
	    }
	    else
	    {
		listEntry *entry = (listEntry *)(list -> GetCellValue (row, 0));
		TREENODE_PTR node = entry -> getTreeNode ();
		if (node -> parameters != NULL)
		{
		    vstr *group = vstrCloneScribed (vcharScribeSystem (node -> groupName));
		    result = alCfgMessageDialogCenter (mainDialog, 
						       ALCFG_DELETE_PARAMS_REQ,
						       group);
		    if (result == ALCFG_TRUE)
		    {
			vmemFree ((void *)(node -> parameters));
			node -> parameters = NULL;
			alarmTreeClass *alarmTree = app -> getAlarmTree ();
			// will cause the info.dbf to bew rewritten
			alarmTree -> setInfoModified (vTRUE);
		    }
		    vstrDestroy (group);
		}
	    }
	    break;
	}

	// Printer menu items

	case EDIT_PRINTER_MENUITEM:
	{
	    vresource listRes = vresourceGet (appRes, alcfg_alCfgListDialog);
	    alCfgListDialogClass *listDialog = new alCfgListDialogClass (listRes);
	    listDialog -> listOpen (mainDialog, EDIT_PRINTER_FLAG);
	    vchar selectedName[ALCFG_SELECTEDNAME_SIZE + 1];
	    vcharCopyBounded (listDialog -> getSelectedName (), 
			      selectedName, ALCFG_SELECTEDNAME_SIZE);
	    selectedName[ALCFG_SELECTEDNAME_SIZE] = 0;
	    delete listDialog;
	    if ((vcharLength (selectedName)) != 0)
	    {
		vresource prtRes = vresourceGet (appRes, alcfg_alCfgPrtDialog);
		alCfgPrtDialogClass *prtDialog = new alCfgPrtDialogClass (prtRes);
		prtDialog -> prtOpen (mainDialog, selectedName);
		delete prtDialog;
	    }
	    break;
	}

	case CREATE_PRINTER_MENUITEM:
	{
	    vresource newRes = vresourceGet (appRes, alcfg_alCfgNewItemDialog);
	    alCfgNewItemDialogClass *newDialog = new alCfgNewItemDialogClass (newRes);
	    newDialog -> newOpen (mainDialog, NEW_PRINTER_FLAG);
	    delete newDialog;
	    break;
	}

	case DELETE_PRINTER_MENUITEM:
	{
	    vresource listRes = vresourceGet (appRes, alcfg_alCfgListDialog);
	    alCfgListDialogClass *listDialog = new alCfgListDialogClass (listRes);
	    listDialog -> listOpen (mainDialog, REMOVE_PRINTER_FLAG);
	    delete listDialog;
	    break;
	}

	default:
	    return (vFALSE);
    }
    return (vTRUE);
}

