#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vtextitemHEADER
#include vlistitemHEADER
#include vbuttonHEADER

extern "C"
{
#include "alarmDefs.h"
}

#include "alcfg.h"
#include "alCfgTreeListClass.hxx"
#include "alarmConfig.hxx"
#include "alCfgMainDialogClass.hxx"
#include "alCfgNewItemDialogClass.hxx"
#include "alCfgUtils.hxx"

static int newItemFilter (filteredText *text, vtextSelection *sel,
			  const vchar **newText, long *len);


vkindTYPED_FULL_DEFINITION(alCfgNewItemDialogClass, vdialog, 
						"alCfgNewItemDialogClass");

/*******************************************************************
*
*    alCfgNewItemDialogClass override for constructors init ()
*
*******************************************************************/

void alCfgNewItemDialogClass::LoadInit (vresource res)
{
    title = NULL;
    newItem = filteredTextitem::RequestSubLoad (this,
					alcfg_alCfgNewName);
    okButton = alCfgNewItemOkButtonClass::RequestSubLoad (this,
					alcfg_alCfgNewOkButton);
    cancelButton = alCfgNewItemCancelButtonClass::RequestSubLoad (this,
					alcfg_alCfgNewCancelButton);
    // init inherited data
    vdialog::LoadInit (res);

    // store ptrs
    okButton -> storeNewDialog (this);
    cancelButton -> storeNewDialog (this);

    // set the filter 
    newItem -> SetItemFilterFunc (newItemFilter);
}


/*******************************************************************
*
*    alCfgNewItemDialogClass Destroy function
*
*******************************************************************/

void alCfgNewItemDialogClass::Destroy (void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }
    if (title != NULL)
    {
	vstrDestroy (title);
    }
    vdialog::Destroy ();
}


/*******************************************************************
*
*    alCfgNewItemDialogClass notify function
*
*******************************************************************/

void alCfgNewItemDialogClass::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    alCfgNewItemDialogClass Open function
*
*******************************************************************/

void alCfgNewItemDialogClass::newOpen ( alCfgMainDialogClass *mnDialog, int flag)
{
    mainDialog = mnDialog;
    vresource titlesRes = mainDialog -> getTitlesResource ();
    newItemType = flag;
    if (newItemType == NEW_GROUP_FLAG)
    {
	title = vresourceGetString (titlesRes, alcfg_alCfgNewGrpTitle);
    }
    else
    {
	title = vresourceGetString (titlesRes, alcfg_alCfgNewPrtTitle);
    }

    SetTitle (title);
    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open();
    vevent::Process ();
}


static int newItemFilter (filteredText *text, vtextSelection *sel,
			  const vchar **newText, long *len)
{
    int ret = vFALSE;
    vchar aSpace = vcharFromLiteral (' ');

    if (*len > 0)
    {
	*len = 1;
	if (**newText == aSpace)
	{
	    vwindow::Beep ();
	    ret = vFALSE;
	}
	else
	{
	    ret = NCharFilter (text, sel, newText, len, AL_GROUPNAME_SIZE);
	}
    }
    else
    {
	ret = vFALSE;
    }
    return (ret);
}



vkindTYPED_FULL_DEFINITION(alCfgNewItemOkButtonClass, vbutton, 
				"alCfgNewItemOkButtonClass");

/*******************************************************************
*
*    alCfgNewItemOkButtonClass 
*	Observer function
*
*******************************************************************/

void alCfgNewItemOkButtonClass::ObserveDialogItem (vevent *event)
{
    vstr *newThing = newDialog -> newItem -> GetTextAsString ();

    if ((vcharLength (newThing)) > 0)
    {
	alCfgMainDialogClass *mainDialog = newDialog -> getMainDialog ();
	alCfgApplication *app = mainDialog -> getApplication ();
	int itemType = newDialog -> getItemType ();
	if (itemType == NEW_GROUP_FLAG)
	{
	    if ((app -> addToGroupList (newThing)) != 0)
	    {
		alCfgMessageDialogCenter (newDialog, ALCFG_GROUP_EXISTS_ERR, newThing);
		vstrDestroy (newThing);
		newDialog -> newItem -> SetTextScribed(vcharScribeSystem (""));
		return;
	    }
	}
	else
	{
	   if ((app -> addToPrinterList (newThing)) != 0)
	   {
		alCfgMessageDialogCenter (newDialog, ALCFG_PRINTER_EXISTS_ERR, newThing);
		vstrDestroy (newThing);
		newDialog -> newItem -> SetTextScribed(vcharScribeSystem (""));
		return;
	   }
	}
    }
    else
    {
	vstrDestroy (newThing);
    }

    // close dialog and stop event loop started in newOpen ()
    newDialog -> Close();
    vevent::StopProcessing ();
}


vkindTYPED_FULL_DEFINITION(alCfgNewItemCancelButtonClass, vbutton, 
				"alCfgNewItemCancelButtonClass");

/*******************************************************************
*
*    alCfgNewItemCancelButtonClass 
*	Observer function
*
*******************************************************************/

void alCfgNewItemCancelButtonClass::ObserveDialogItem (vevent *event)
{
    // close dialog and stop event loop started in newOpen ()
    newDialog -> Close();
    vevent::StopProcessing ();
}
