#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vtextitemHEADER
#include vlistitemHEADER
#include vbuttonHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "alarmDefs.h"
}

#include "alcfg.h"
#include "alCfgTreeListClass.hxx"
#include "alarmConfig.hxx"
#include "alCfgDatabaseClass.hxx"
#include "alCfgMainDialogClass.hxx"
#include "alCfgPrtDialogClass.hxx"
#include "alCfgUtils.hxx"

// filter function for printer input fields
static int prtFilterFunc (filteredText *text, vtextSelection *sel,
			  const vchar **newText, long *length);


vkindTYPED_FULL_DEFINITION(alCfgPrtDialogClass, vdialog, 
						"alCfgPrtDialogClass");

/*******************************************************************
*
*    alCfgPrtDialogClass override for constructors init ()
*
*******************************************************************/

void alCfgPrtDialogClass::LoadInit (vresource res)
{
    title = NULL;
    modified = vFALSE;

    vcontainer *prt1container = vcontainer::RequestSubLoad (this,
					alcfg_alCfgPrt1Container);
    printers[0] = filteredTextitem::RequestSubLoad (prt1container,
					alcfg_alCfgPrt1);

    vcontainer *prt2container = vcontainer::RequestSubLoad (this,
					alcfg_alCfgPrt2Container);
    printers[1] = filteredTextitem::RequestSubLoad (prt2container,
					alcfg_alCfgPrt2);

    vcontainer *prt3container = vcontainer::RequestSubLoad (this,
					alcfg_alCfgPrt3Container);
    printers[2] = filteredTextitem::RequestSubLoad (prt3container,
					alcfg_alCfgPrt3);

    vcontainer *prt4container = vcontainer::RequestSubLoad (this,
					alcfg_alCfgPrt4Container);
    printers[3] = filteredTextitem::RequestSubLoad (prt4container,
					alcfg_alCfgPrt4);

    vcontainer *prt5container = vcontainer::RequestSubLoad (this,
					alcfg_alCfgPrt5Container);
    printers[4] = filteredTextitem::RequestSubLoad (prt5container,
					alcfg_alCfgPrt5);

    vcontainer *prt6container = vcontainer::RequestSubLoad (this,
					alcfg_alCfgPrt6Container);
    printers[5] = filteredTextitem::RequestSubLoad (prt6container,
					alcfg_alCfgPrt6);

    vcontainer *prt7container = vcontainer::RequestSubLoad (this,
					alcfg_alCfgPrt7Container);
    printers[6] = filteredTextitem::RequestSubLoad (prt7container,
					alcfg_alCfgPrt7);

    vcontainer *prt8container = vcontainer::RequestSubLoad (this,
					alcfg_alCfgPrt8Container);
    printers[7] = filteredTextitem::RequestSubLoad (prt8container,
					alcfg_alCfgPrt8);

    vcontainer *prt9container = vcontainer::RequestSubLoad (this,
					alcfg_alCfgPrt9Container);
    printers[8] = filteredTextitem::RequestSubLoad (prt9container,
					alcfg_alCfgPrt9);

    vcontainer *prt10container = vcontainer::RequestSubLoad (this,
					alcfg_alCfgPrt10Container);
    printers[9] = filteredTextitem::RequestSubLoad (prt10container,
					alcfg_alCfgPrt10);

    vcontainer *buttonContainer = vcontainer::RequestSubLoad (this,
					alcfg_alCfgPrtButtonContainer);

    applyButton = alCfgPrtApplyButtonClass::RequestSubLoad (buttonContainer,
					alcfg_alCfgPrtApplyButton);
    dismissButton = alCfgPrtDismissButtonClass::RequestSubLoad (buttonContainer,
					alcfg_alCfgPrtDismissButton);
    clearButton = alCfgPrtClearButtonClass::RequestSubLoad (buttonContainer,
					alcfg_alCfgPrtClearButton);
    // init inherited data
    vdialog::LoadInit (res);

    // store ptrs
    applyButton -> storePrtDialog (this);
    dismissButton -> storePrtDialog (this);
    clearButton -> storePrtDialog (this);

    // apply filter function, and ptr to modified flag
    int i;
    for (i = 0; i < MAX_PRINTERS_PER_GROUP; i++)
    {
	printers[i] -> SetItemFilterFunc (prtFilterFunc);
	printers[i] -> SetTextModifiedPtr (&this -> modified);
    }
}


/*******************************************************************
*
*    alCfgPrtDialogClass Destroy function
*
*******************************************************************/

void alCfgPrtDialogClass::Destroy (void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }
    if (title != NULL)
    {
	vstrDestroy (title);
    }
    vdialog::Destroy ();
}


/*******************************************************************
*
*    alCfgPrtDialogClass notify function
*
*******************************************************************/

void alCfgPrtDialogClass::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    alCfgPrtDialogClass Open function
*
*******************************************************************/

void alCfgPrtDialogClass::prtOpen ( alCfgMainDialogClass *mnDialog, vchar *prtGrpName)
{
    mainDialog = mnDialog;
    app = mainDialog -> getApplication ();
    prtGroupName = prtGrpName;

    vresource titlesRes = mainDialog -> getTitlesResource ();
    vstr *formattedTitle = vresourceGetString (titlesRes, alcfg_alCfgPrtTitle);
    title = vstrCloneScribed (vcharScribeFormat (formattedTitle, prtGroupName));
    vstrDestroy (formattedTitle);

    SetTitle (title);
    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);

    loadPrtDialog ();

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open();
    vevent::Process ();
}


/*******************************************************************
*
*    alCfgPrtDialogClass clear function
*
*******************************************************************/

void alCfgPrtDialogClass::clearPrtDialog (void)
{
    int i;
    for (i = 0; i < MAX_PRINTERS_PER_GROUP; i++)
    {
	printers[i] -> SetTextScribed (vcharScribeSystem (""));
    }
}


/*******************************************************************
*
*    alCfgPrtDialogClass load function
*
*******************************************************************/

void alCfgPrtDialogClass::loadPrtDialog (void)
{
    PRINTER_ENTRY_PTR entry;
    vlist *prtList = app -> getPrinterList ();
    vlistIterator iterator;
    iterator.Start (prtList);
    while (iterator.Next ())
    {
	entry = (PRINTER_ENTRY_PTR)(iterator.GetValue ());
	if ((vcharCompareBounded (prtGroupName, entry -> printerGroupName, 
						AL_GROUPNAME_SIZE)) == 0)
	{
	    break;
	}
    }
    iterator.Finish ();

    int i;
    for (i = 0; i < MAX_PRINTERS_PER_GROUP; i++)
    {
	printers[i] -> SetText (entry -> groupPrinters[i]);
    }
    prtEntry = entry;
    modified = vFALSE;
}


/*******************************************************************
*
*    alCfgPrtDialogClass save function
*
*******************************************************************/

void alCfgPrtDialogClass::savePrtDialog (void)
{
    if (modified)
    {
	vstr *value;
	int i;
	for (i = 0; i < MAX_PRINTERS_PER_GROUP; i++)
	{
	    value = printers[i] -> GetTextAsString ();
	    vcharCopyBounded (value, &prtEntry -> groupPrinters[i][0], AL_PATHNAME_SIZE);
	    prtEntry -> groupPrinters[i][AL_PATHNAME_SIZE] = 0;
	    vstrDestroy (value);
	}
	app -> setPrinterListModified(vTRUE);
    }
}


vkindTYPED_FULL_DEFINITION(alCfgPrtApplyButtonClass, vbutton, 
				"alCfgPrtApplyButtonClass");

/*******************************************************************
*
*    alCfgPrtApplyButtonClass 
*	Observer function
*
*******************************************************************/

void alCfgPrtApplyButtonClass::ObserveDialogItem (vevent *event)
{
    prtDialog -> savePrtDialog ();
    // close dialog and stop event loop started in prtOpen ()
    prtDialog -> Close();
    vevent::StopProcessing ();
}


vkindTYPED_FULL_DEFINITION(alCfgPrtDismissButtonClass, vbutton, 
				"alCfgPrtDismissButtonClass");

/*******************************************************************
*
*    alCfgPrtDismissButtonClass 
*	Observer function
*
*******************************************************************/

void alCfgPrtDismissButtonClass::ObserveDialogItem (vevent *event)
{
    int modified = prtDialog -> getModFlag ();
    if (modified == vTRUE)
    {
	int result = alCfgMessageDialogCenter (prtDialog, ALCFG_DATAMOD_MSG);
	if (result == ALCFG_FALSE)
	{
	    return;
	}
    }
    // close dialog and stop event loop started in listOpen ()
    prtDialog -> Close();
    vevent::StopProcessing ();
}


vkindTYPED_FULL_DEFINITION(alCfgPrtClearButtonClass, vbutton, 
				"alCfgPrtClearButtonClass");

/*******************************************************************
*
*    alCfgPrtClearButtonClass 
*	Observer function
*
*******************************************************************/

void alCfgPrtClearButtonClass::ObserveDialogItem (vevent *event)
{
    prtDialog -> clearPrtDialog ();
}



/*******************************************************************
*
*    prtFilterFunc
*	filter function for printer input fields
*
*******************************************************************/

static int prtFilterFunc (filteredText *text, vtextSelection *sel,
			  const vchar **newText, long *length)
{
    return (NCharFilter (text, sel, newText, length, AL_PATHNAME_SIZE));
}
