/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/***************************************************************************
 merge_alarmlog

 takes two alarmlog dbase files and merges the two so the 
 two files contain consistant data.
****************************************************************************/

#include <stdio.h>
#include <time.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "alarmDefs.h"
#include "msgMon.h"
#include "resolveLinks.h"

#define TRUE          1
#define FALSE         0
#define PREV_HOURS    24
#define SECS_PER_HOUR 3600

/* function prototypes */
int GetRecTime(D4DATA *FileDesc, char StartBuf[], char StopBuf[] );

/* Field descriptors */

F4FIELD	*sourceIsAckFld, *sourceDateFld, *sourceTimeFld, *sourceAlarmSecsFld;
F4FIELD *sourceAlarmMsecsFld, *sourceGroupNameFld, *sourcePointNameFld, *sourceDescriptorFld;
F4FIELD *sourceStatusFld, *sourceSeverityFld, *sourceAckedFld, *sourceOperatorFld;
F4FIELD *sourceResourceFld, *sourceEguTagFld, *sourceValueFld, *sourceAckSecsFld;
F4FIELD *sourceAckMsecsFld, *sourceCommentFld, *sourceNoPrintFld;

F4FIELD	*destinationIsAckFld, *destinationDateFld, *destinationTimeFld;
F4FIELD *destinationAlarmSecsFld, *destinationAlarmMsecsFld, *destinationGroupNameFld;
F4FIELD	*destinationPointNameFld, *destinationDescriptorFld, *destinationStatusFld;
F4FIELD *destinationSeverityFld, *destinationAckedFld, *destinationOperatorFld;
F4FIELD *destinationResourceFld, *destinationEguTagFld, *destinationValueFld;
F4FIELD *destinationAckSecsFld, *destinationAckMsecsFld, *destinationCommentFld;
F4FIELD *destinationNoPrintFld;

/************************************************************************
*			PUBLIC DECLARATIONS				*
************************************************************************/

char		rbuf[20];		/* buffer for keeping date */
time_t		CurEpoch;
time_t		CutOffEpoch;
struct tm	CurTm;
struct tm	CutOffTm;
int		CutOff;			/* Cutoff point for keeping data records */

/***********************************************************************/
void  openFields(D4DATA *source, D4DATA *destination)
{
	sourceIsAckFld = d4field (source, "ISACK");
	sourceDateFld = d4field (source, "DATE");
	sourceTimeFld = d4field (source, "TIME");
	sourceAlarmSecsFld = d4field (source, "ALARMSECS");
	sourceAlarmMsecsFld = d4field (source, "ALARMMSECS");
	sourceGroupNameFld = d4field (source, "GROUPNAME");
	sourcePointNameFld = d4field (source, "POINTNAME");
	sourceDescriptorFld = d4field (source, "DESCRIPTOR");
	sourceStatusFld = d4field (source, "STATUS");
	sourceSeverityFld = d4field (source, "SEVERITY");
	sourceAckedFld = d4field (source, "ACKED");
	sourceNoPrintFld = d4field (source, "NOPRINT");
	sourceOperatorFld = d4field (source, "OPERATOR");
	sourceResourceFld = d4field (source, "RESOURCE");
	sourceEguTagFld = d4field (source, "EGUTAG");
	sourceValueFld = d4field (source, "VALUE");
	sourceAckSecsFld = d4field (source, "ACKSECS");
	sourceAckMsecsFld = d4field (source, "ACKMSECS");
	sourceCommentFld = d4field (source, "COMMENT");

	d4unlock_all(source);

	destinationIsAckFld = d4field (destination, "ISACK");
	destinationDateFld = d4field (destination, "DATE");
	destinationTimeFld = d4field (destination, "TIME");
	destinationAlarmSecsFld = d4field (destination, "ALARMSECS");
	destinationAlarmMsecsFld = d4field (destination, "ALARMMSECS");
	destinationGroupNameFld = d4field (destination, "GROUPNAME");
	destinationPointNameFld = d4field (destination, "POINTNAME");
	destinationDescriptorFld = d4field (destination, "DESCRIPTOR");
	destinationStatusFld = d4field (destination, "STATUS");
	destinationSeverityFld = d4field (destination, "SEVERITY");
	destinationAckedFld = d4field (destination, "ACKED");
	destinationNoPrintFld = d4field (destination, "NOPRINT");
	destinationOperatorFld = d4field (destination, "OPERATOR");
	destinationResourceFld = d4field (destination, "RESOURCE");
	destinationEguTagFld = d4field (destination, "EGUTAG");
	destinationValueFld = d4field (destination, "VALUE");
	destinationAckSecsFld = d4field (destination, "ACKSECS");
	destinationAckMsecsFld = d4field (destination, "ACKMSECS");
	destinationCommentFld = d4field (destination, "COMMENT");

	d4unlock_all(destination);
}
/***********************************************************************/
static void copyFields ()

    {
	f4assign (destinationIsAckFld, f4str (sourceIsAckFld));
	f4assign (destinationDateFld, f4str (sourceDateFld));
	f4assign (destinationTimeFld, f4str (sourceTimeFld));
	f4assign (destinationAlarmSecsFld, f4str (sourceAlarmSecsFld));
	f4assign (destinationAlarmMsecsFld, f4str (sourceAlarmMsecsFld));
	f4assign (destinationGroupNameFld, f4str (sourceGroupNameFld));
	f4assign (destinationPointNameFld, f4str (sourcePointNameFld));
	f4assign (destinationDescriptorFld, f4str (sourceDescriptorFld));
	f4assign (destinationStatusFld, f4str (sourceStatusFld));
	f4assign (destinationSeverityFld, f4str (sourceSeverityFld));
	f4assign (destinationAckedFld, f4str (sourceAckedFld));
	f4assign (destinationNoPrintFld, f4str (sourceNoPrintFld));
	f4assign (destinationOperatorFld, f4str (sourceOperatorFld));
	f4assign (destinationResourceFld, f4str (sourceResourceFld));
	f4assign (destinationEguTagFld, f4str (sourceEguTagFld));
	f4assign (destinationValueFld, f4str (sourceValueFld));
	f4assign (destinationAckSecsFld, f4str (sourceAckSecsFld));
	f4assign (destinationAckMsecsFld, f4str (sourceAckMsecsFld));
	m4assign (destinationCommentFld, m4str (sourceCommentFld));
    }

/*****************************************************************************
 copy a record from the source db file into the destination db file
*/

void AddToList( D4DATA *destination )
{
char    point[AL_POINTNAME_SIZE];

	d4top(destination);

        memset( point, 0, sizeof point);
	strcpy (point, f4str (destinationPointNameFld));

	if (d4deleted(destination)) {
		copyFields();
	        d4recall(destination);
 	}
	else {
		d4append_start( destination, 0 );	/* '0' => memo fields */
		d4blank( destination );
		copyFields();
		d4append( destination );
		d4flush_all( destination );
#ifndef	CB51
		d4update_header( destination, 1, 1);
#endif	!CB51
	}
	d4unlock_all(destination);
	return;
}
/*******************************************************************************
 copy the records from the source db file that do not have a match in the 
 destination db file, to the destination db file.
*/

void merge_files(D4DATA *source, D4DATA *destination, char start[], char stop[])  
{
    int    found;
    char   source_pt[AL_POINTNAME_SIZE+20];
    char   dest_pt[AL_POINTNAME_SIZE+20];
    char   tag_string[100], source_isAck, dest_isAck;
    char   source_status[AL_STATUS_SIZE+1], source_sev[AL_SEVERITY_SIZE+1];
    char   dest_status[AL_STATUS_SIZE+1], dest_sev[AL_SEVERITY_SIZE+1];
    char   dest_value[AL_VALUE_SIZE+1], source_value[AL_VALUE_SIZE+1];
    long   source_secs, source_msecs, dest_secs, dest_msecs;
    int    seek;
    
    while (!d4eof(source)) {

	if ( GetRecTime(source, start, stop) < 0)  {
	    /* If the record is not with in the start stop date range */
            d4skip(source, 1);
	    continue;
        }

	found = FALSE;

	/* seek to the destination for the source date/time */

	/* Create a key for this record */

        memset( source_pt, 0, sizeof source_pt);
	strcpy (source_pt, f4str (sourcePointNameFld));
	source_secs = f4long (sourceAlarmSecsFld);
	source_msecs = f4long (sourceAlarmMsecsFld);
	source_isAck = f4char (sourceIsAckFld);

	memset(tag_string, 0, sizeof tag_string);
	sprintf (tag_string, "%*ld%*d%-*s%c",
		AL_GMTSECS_SIZE, source_secs,
		AL_GMTMSECS_SIZE, source_msecs,
 		AL_POINTNAME_SIZE, source_pt,
		(char)source_isAck);

        memset( source_status, 0, sizeof source_status);
        strcpy( source_status, f4str(sourceStatusFld));

        memset( source_sev, 0, sizeof source_sev);
        strcpy( source_sev, f4str(sourceSeverityFld));

        memset( source_value, 0, sizeof source_value);
        strcpy( source_value, f4str(sourceValueFld));

	if ( (seek = d4seek( destination, tag_string )) == 0) {

	   /* get the destination field values to determine if a match has been found */	
	
	   dest_secs = f4long (destinationAlarmSecsFld);
	   dest_msecs = f4long (destinationAlarmMsecsFld);
	   dest_isAck = f4char (destinationIsAckFld);

           memset( dest_pt, 0, sizeof dest_pt);
           strcpy( dest_pt, f4str(destinationPointNameFld));

           memset( dest_status, 0, sizeof dest_status);
           strcpy( dest_status, f4str(destinationStatusFld));

           memset( dest_sev, 0, sizeof dest_sev);
           strcpy( dest_sev, f4str(destinationSeverityFld));

           memset( dest_value, 0, sizeof dest_value);
           strcpy( dest_value, f4str(destinationValueFld));

	   while ( (source_secs == dest_secs) &&
	           (source_msecs == dest_msecs) &&
		   (source_isAck == dest_isAck) &&
	           (strcmp(source_pt, dest_pt) == 0) ){

 	      if(( strcmp(source_status, dest_status) == 0) &&
	         ( strcmp(source_sev, dest_sev) ==0) &&
		 ( strcmp(source_value, dest_value) == 0) ){

		  /* a record is considered to match if the alarmsecs, alarmmsecs,
		     isack, pointnames, status, severity, and value are the same. */

	          found = TRUE;

                  /* If a match is found but ONLY ONE is marked for deletion, 
                     undelete it. */

                  if (d4deleted(source)) {
                    if (!d4deleted(destination))
                        d4recall(source);
                  }
                  else  {
	            if (d4deleted(destination))
	                d4recall(destination);
                  }
	          break;
             }
             else {
		   /* while the tags match determine is a match can be found */
		   /* get the field values */

	           d4skip(destination, 1);

         	   memset( dest_pt, 0, sizeof dest_pt);
		   strcpy (dest_pt, f4str (destinationPointNameFld));

		   dest_secs = f4long (destinationAlarmSecsFld);
		   dest_msecs = f4long (destinationAlarmMsecsFld);
		   dest_isAck = f4char (destinationIsAckFld);

                   memset( dest_status, 0, sizeof dest_status);
                   strcpy( dest_status, f4str(destinationStatusFld));

                   memset( dest_sev, 0, sizeof dest_sev);
                   strcpy( dest_sev, f4str(destinationSeverityFld));

                   memset( dest_value, 0, sizeof dest_value);
                   strcpy( dest_value, f4str(destinationValueFld));
             }
           }
	  }
          if (!found)

                   /* If the source record is not found in the destination db file */
                   AddToList(destination);

       d4skip(source, 1);
      d4top(destination);
   }
}

/**************************************************************************************/
int GetRecTime(D4DATA *FileDesc, char StartBuf[], char StopBuf[] )
{
	int	yr;
	char	DateStr[sizeof rbuf];
	char	CmpDateTime[sizeof rbuf];
	char    TimeStr[sizeof rbuf];

	memset( DateStr, 0, sizeof DateStr );
	memset( CmpDateTime, 0, sizeof CmpDateTime );
	strcpy( DateStr, f4str(d4field(FileDesc, "DATE")));
	if ( strcmp(DateStr, "        ") == 0)
	     return -1;
	yr = atoi( &DateStr[6] );
	memset( TimeStr, 0, sizeof TimeStr);
	strcpy( TimeStr, f4str(d4field(FileDesc, "TIME")));
	if (strcmp(TimeStr, "        ") == 0) 
	     return -1;

	/* Now build comparison buffer */
	if( yr >= 70 )  {
		sprintf( CmpDateTime, "19%2.2s%2.2s%2.2s%2.2s%2.2s%2.2s",
		     &DateStr[6], DateStr, &DateStr[3], TimeStr, &TimeStr[3], &TimeStr[6] );
	}
	else  {
		sprintf( CmpDateTime, "20%2.2s%2.2s%2.2s%2.2s",
		     &DateStr[6], DateStr, &DateStr[3], TimeStr, &TimeStr[3], &TimeStr[6] );
	}

	if(((strcmp(CmpDateTime, StartBuf)) >= 0 ) && ((strcmp(CmpDateTime, StopBuf)) <= 0))
		return 0;
	else if ( strcmp(CmpDateTime, StartBuf) < 0)
		return -1;
        else /* CmpDateTime, StopBuf > 0 */
		return 1;
}
/***************************************************************************************/
void get_mergetimes(char *endtime, char *starttime, int hours)

{

    /*Get the current time to obtain stopping time for merging.*/

	time( &CurEpoch );
	CurTm = *(localtime( &CurEpoch ));
	CutOffEpoch = CurEpoch;
	CutOffTm = *(localtime( &CutOffEpoch ));
	memset( endtime, 0, sizeof rbuf );
	if( CutOffTm.tm_year >= 70 )  {
		sprintf( endtime, "19%2.2d%2.2d%2.2d%2.2d%2.2d%2.2d", 
			CutOffTm.tm_year % 100, CutOffTm.tm_mon+1, CutOffTm.tm_mday, 
			CutOffTm.tm_hour, CutOffTm.tm_min, CutOffTm.tm_sec);
	}
	else  {
		sprintf( endtime, "20%2.2d%2.2d%2.2d%2.2d%2.2d%2.2d", 
			CutOffTm.tm_year % 100, CutOffTm.tm_mon+1, CutOffTm.tm_mday, 
			CutOffTm.tm_hour, CutOffTm.tm_min, CutOffTm.tm_sec);
	}

	/* obtain starting time for merging files */

	if (hours)
	    CutOffEpoch = CutOffEpoch - (hours * SECS_PER_HOUR);
        else
	    CutOffEpoch = CutOffEpoch - (PREV_HOURS * SECS_PER_HOUR);

        CutOffTm = *(localtime(&CutOffEpoch));
	memset( starttime, 0, sizeof rbuf);
	if( CutOffTm.tm_year >= 70 )  {
		sprintf( starttime, "19%2.2d%2.2d%2.2d%2.2d0000", 
			CutOffTm.tm_year % 100, CutOffTm.tm_mon+1, CutOffTm.tm_mday, 
			CutOffTm.tm_hour );
	}
	else  {
		sprintf( starttime, "20%2.2d%2.2d%2.2d%2.2d0000", 
			CutOffTm.tm_year % 100, CutOffTm.tm_mon+1, CutOffTm.tm_mday, 
			CutOffTm.tm_hour );
	}
}
    
/************************************************************************************/
void usage()
{
    tisSysLog(0,0, "Usage:MergeAlarm {source} [hours]\twhere:");
    tisSysLog(0,0,"\t\t{File1} = Merge, complete path to alarmlog.dbf, source file;");
    tisSysLog(0,0,"\t\t[hours] = Optional parameter, number of hours to merge, default is 24;");
}

/***********************************************************************************/

void main(int argc, char **argv)

{
    C4CODE               cb;
    D4DATA               *source, *destination;
    T4TAG                *source_tag, *destination_tag;
    char                 starttime[sizeof rbuf], endtime[sizeof rbuf];
    int                  hours = 0;
    char 		 dbname[100];

      tisOpenLog ("MergeAlarm", 6, 2, NULL, NULL);
      if (argc < 2 || argc > 3) {
	  usage();
	  exit(1);
      }
      else {
          if (argc ==3) {
	     hours = atoi(argv[2]);
	     if (!hours) {
		 tisSysLog(0,0,"MergeAlarm : The number of hours to merge must be an integer greater than 0");
		 exit(1);
             }
           }
      }

    /*Get the merge stop and start times.*/

      get_mergetimes(endtime, starttime, hours);

      /* initialize codebase */ 
       d4init(&cb);

#if defined(USECB63) || defined(WIN32)
	cb.errOpen = 0;
	cb.errTagName = 0;
#else
       cb.open_error = 0;
       cb.tag_name_error = 0;
#endif	USECB63

       /* Open files, get and select tags, position to the top of the files */

       memset(dbname, 0, sizeof dbname);
       resolveLinks(argv[1], dbname);

       if ( !(source = d4open(&cb, dbname)) ){
	   tisSysLog(0,0, "MergeAlarm : file %s can not be opened, no merging will be done",
			dbname); 
	   exit(1);
       }

       memset(dbname, 0, sizeof dbname);
#ifdef WIN32
       resolveLinks("\\tis4000db\\alarmlog.dbf", dbname);
#else
       resolveLinks("/tis4000db/alarmlog.dbf", dbname);
#endif WIN32
       if ( !(destination = d4open(&cb, dbname)) ){
	   d4close(source);
	   tisSysLog(0,0, "MergeAlarm : file %s can not be opened, no merging will be done",
			dbname); 
	   exit(1);
       }

       if ( !(source_tag = d4tag(source, "EVENTTIME")) || 
		!(destination_tag = d4tag(destination,"EVENTTIME")) ){
	   d4close(source);
	   d4close(destination);
	   tisSysLog(0,0, "MergeAlarm : invalid file format, no merging will be done"); 
	   exit(1);
       }

       d4tag_select(source, source_tag);
       d4tag_select(destination, destination_tag);

       d4top(source);
       d4top(destination);

       d4unlock_all(source);
       d4unlock_all(destination);

       openFields(source, destination);

       /* merge the source into the destination */

       merge_files( source, destination, starttime, endtime);

       d4flush_all(destination);
       d4close(source);
       d4close(destination);

       exit( 0 );
}
