/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "tsDefs.h"

#include "alarmDefs.h"
#include "shareMem.h"
#include "groupInfo.h"
#include "globals.h"
#include "tree.h"
#include "sharedMem.h"
#include "alarmMgr.h"
#include "dbase.h"

/* Functions defined in this module */

#ifndef	VOID
#define	VOID	void
#endif	VOID
	void	initializePoints (void);
	void	terminatePoints (void);
static	PvData 	*allocatePoint (void);
static	void	freePoint (PvData *);
	void	loadPoints (void);
static	int	sortCompare (const void *, const void *);
	PvData	*loadPoint (char *, char *, char *, char *);
static  void    getStation (char *, char *);
	void	addPoint (TreeNode *, PvData *);
	void	addPointToSummary (PvData *);
	void	removePointByName (char *);
	void	removePoint (PvData *);
	void	removePointFromSummary (PvData *);
	PvData	*findPointInList (char *);
static	int	findCompare (const void *, const void *);
	int	inSummaryList (PvData *);
	void	resetInUseFlags (void);

/* The point (pv) segment is set up as a pool. In order to avoid utilizing the entire */
/* segment a free list and an allocation limit are maintained. When a new point is */
/* needed the free list is checked first, if it is empty then the allocation limit */
/* is bumped and a new point formed. */

static	char		*pointSegmentAllocationLimit;

/* A list of pointers to the PVs is maintained sorted by pvName so that a binary search */
/* can be used. Since points can be added to the list dynamically at run time, a linear list */
/* is maintained at the end of the sorted list. */

static	PvData	**pointList = 0;	/* List of pointers to points. */
static	PvData	**nextPoint;
static	long	numSortedPoints;	/* Number of sorted points. */

/*----------*/
void initializePoints (VOID)		/* Initialize point segment */

    {

	int	num;

	pointSegmentAllocationLimit = (char *)shareHeader->alarmPoints; /* Initialized by sharedMem.C */
	shareHeader->alarmPointCount = 0;
	numSortedPoints = 0;
	if (pointList)
	    free ((char *)pointList);
	num = getNumPVs ();
	pointList = nextPoint = (PvData **) malloc (num * sizeof (PvData *));

    }

/*----------*/
void terminatePoints (VOID)		/* Terminate the points mechanism */

    {

	free ((char *)pointList);

    }

/*----------*/
static PvData *allocatePoint (VOID)		/* Allocate a new point from the pool */

    {

	PvData	*p;

	extendMemoryIfFull (pointSegmentAllocationLimit + sizeof (PvData));
	p = (PvData *)pointSegmentAllocationLimit;
	pointSegmentAllocationLimit += sizeof (PvData);
	memset ((char *)p, 0, sizeof (PvData));
	*nextPoint++ = p;
	shareHeader->alarmPointCount++;

	return p;

    }


/*----------*/
static void freePoint (PvData *p)		/* Put a point on the free list */

    {

	p->parentGroup = NULL;

    }

/*----------*/
void loadPoints (VOID)

    {

	int	cnt, inSummary;

	/* Read through the pvMap file. Any points that have an alarm group */
	/* assigned will be included in the tree.  If a group cannot be found */
	/* in the tree, the point will be ignored. */

	PvMapInfo pvMapInfo;
	cnt = 0;
	while (getNextPv (&pvMapInfo)) {

	    /* Find the node in the tree for this group */

	    TreeNode *p = findNode (pvMapInfo.groupName);
	    if (p) {
		PvData *q = allocatePoint ();
		if (getPointState (pvMapInfo.pointName, q, &inSummary)) {
		    if (inSummary)	/* point belongs on summary. */
			addPointToSummary (q);
		    else
			q->prevSummaryPv = (PvData *)-1;
		} else {
		    stringCopy (q->description, pvMapInfo.descriptor, sizeof (Descriptor));
		    q->Ack = 1;
		    q->prevSummaryPv = (PvData *)-1;
		}
		q->parentGroup = p;
		q->pvName = addNewString (pvMapInfo.pointName);
                stringCopy (q->resource, pvMapInfo.resName, sizeof (Resource));
                getStation (pvMapInfo.resName, q->stationName);
		addPoint (p, q);
	    }	
	}
	numSortedPoints = shareHeader->alarmPointCount;

	/* Point is is now loaded. Sort the list. */

	qsort ((char *)pointList, (int)numSortedPoints, sizeof (PvData *), sortCompare);

    }

/*----------*/
static int sortCompare ( const void *p1, const void *p2)

    {

	PvData *t1 = *(PvData **)p1;
	PvData *t2 = *(PvData **)p2;
	return strcmp (t1->pvName, t2->pvName);

    }

/*----------*/
/* Load a point */
PvData *loadPoint ( char *pointName, char *groupName, char *descriptor, char *resource)

    {

	/* Find the node in the tree for this group */

	TreeNode *p = findNode (groupName);
	if (p) {
	    PvData *q = allocatePoint ();
	    q->pvName = addNewString (pointName);
	    stringCopy (q->description, descriptor, sizeof (Descriptor));
            getStation (resource, q->stationName);
	    q->parentGroup = p;
	    q->Ack = 1;
	    q->prevSummaryPv = (PvData *)-1;
	    addPoint (p, q);
	    return q;
	}

	return NULL;	/* group not found for this point */

    }

/*----------*/
static void getStation (char *resName, char *stationName)

    /* Get the station name associated with a resource.  If both a primary */
    /* and a secondary ioc exist then they are identical except for the last */
    /* character which will be A or B. The station name is this name stripped */
    /* of the last character. If only a single ioc exists for the resource */
    /* then it is the station name. */

    {

        ResListInfo resListInfo;

        if (getResListEntry (resName, &resListInfo)) {
	  strcpy (stationName, resListInfo.ioc1);
	  if (strlen (resListInfo.ioc2)) {
	    stationName [strlen (stationName) - 1] = '\0';
	  }
	} else {
	  strcpy (stationName, "");
	}

    }

/*----------*/
/* Place a point on a tree nodes point list */
void addPoint ( TreeNode *p, PvData *q)

    {

	q->nextPv = p->pvList;
	p->pvList = q;

    }

/*----------*/
void addPointToSummary ( PvData *p)

    {

	PvData	*q;

	/* If list is currently empty just insert new point */

	if (shareHeader->summaryHead == NULL) {
	    p->nextSummaryPv = NULL;
	    p->prevSummaryPv = NULL;
	    shareHeader->summaryHead = p;
	    shareHeader->summaryTail = p;
	    return;
	}

	/* Find the correct place to insert the point.  We will search */
	/* starting at the tail since that is the most likely insertion point. */

	q = shareHeader->summaryTail;
	while (q) {
	    if (TsCmpStampsLT (&(q->lastAlarm), &(p->lastAlarm)))
		break;
	    q = q->prevSummaryPv;
	}

	if (q == NULL) {				/* Insert at head of list */
	    p->prevSummaryPv = NULL;
	    p->nextSummaryPv = shareHeader->summaryHead;
	    shareHeader->summaryHead->prevSummaryPv = p;
	    shareHeader->summaryHead = p;
	} else if (q == shareHeader->summaryTail) {	/* Insert at tail of list */
	    p->nextSummaryPv = NULL;
	    p->prevSummaryPv = shareHeader->summaryTail;
	    shareHeader->summaryTail->nextSummaryPv = p;
	    shareHeader->summaryTail = p;
	} else {					/* Insert somewhere in the middle */
	    p->nextSummaryPv = q->nextSummaryPv;
	    p->prevSummaryPv = q;
	    q->nextSummaryPv->prevSummaryPv = p;
	    q->nextSummaryPv = p;
	}

    }

/*----------*/
void removePointByName (char *pointName)

    {

	PvData *p = findPointInList (pointName);
	if (p == NULL)
	    return;

	removePoint (p);

    }

/*----------*/
void removePoint (PvData *p)	/* Remove point from group */

    {

	/* Find the point in the tree */

	/* Get a pointer to the beginning of the list. Traverse the list and */ 
	/* find the node to be deleted. */

	PvData *l = p->parentGroup->pvList;
	PvData *prev = NULL;
	while (l != p) {
	    prev = l;
	    l = l->nextPv;
	}

	/* Remove the point from the list and free it. */

	if (prev == NULL)
	    p->parentGroup->pvList = l->nextPv;
	else
	    prev->nextPv = l->nextPv;

	freePoint (l);

    }

/*----------*/
void removePointFromSummary (PvData *p)

    {

	/* Remove a point from the summary list */

	if (shareHeader->summaryHead->nextSummaryPv == NULL) {	/* Only point in the list */
	    shareHeader->summaryHead = NULL;
	    shareHeader->summaryTail = NULL;
	} else if (p == shareHeader->summaryHead) {		/* First on list */
	    shareHeader->summaryHead = p->nextSummaryPv;
	    shareHeader->summaryHead->prevSummaryPv = NULL;
	} else if (p == shareHeader->summaryTail) {		/* Last on list */
	    shareHeader->summaryTail = p->prevSummaryPv;
	    p->prevSummaryPv->nextSummaryPv = NULL;
	} else {						/* Internal to list */
	    p->prevSummaryPv->nextSummaryPv = p->nextSummaryPv;
	    p->nextSummaryPv->prevSummaryPv = p->prevSummaryPv;
	}

	/* Setting these to NULL indicates that the point is not currently in */
	/* the summary list.  nextSummaryPv is set to NULL so alarmVu will not */
	/* freak it looking while unlinking is in progress. */

	p->nextSummaryPv = NULL;
	p->prevSummaryPv = (PvData *)-1;

    }

/*----------*/
PvData *findPointInList (char *pointName)

    {

	PvData	**q, **p;
	long	i;

	/* Perform binary search of the sorted point list. */

	p = (PvData **)bsearch (pointName, (char *)pointList,
			(int)numSortedPoints, sizeof (PvData *), findCompare);

	/* If we found it then return a pointer to it. */

	if (p)
	    if ((*p)->parentGroup != NULL)
		return *p;

	/* If we didn't find it then it may be in the linear list following the */
	/* sorted list. */

	q = pointList + numSortedPoints;
	for (i=numSortedPoints ; i<shareHeader->alarmPointCount ; i++, q++)
	    if (strcmp (pointName, (*q)->pvName) == 0)
		if ((*q)->parentGroup != NULL)
		    return *q;

	return NULL;
	
    }

/*----------*/
static int findCompare ( const void *p1, const void *p2)

    {

	PvData *t = *(PvData **)p2;
	return strcmp ((char *)p1, t->pvName);

    }

/*----------*/
int inSummaryList (PvData *p)

    {

	return p->prevSummaryPv != (PvData *)-1;

    }

/*----------*/
void dumpSummaryList (VOID)

    {

	PvData	*p;

	if (debugLevel < 4)
	    return;

	printf ("********** Alarm Summary List **********\n");
	p = shareHeader->summaryHead;
	while (p) {
	    printf ("%s, ack=%d, RTN=%d, stat=%d, sevr=%d, ustat=%d, usevr=%d, lastAck=%ld\n",
			p->pvName, (int)p->Ack, (int)p->RTN, (int)p->status,
			(int)p->severity, (int)p->unAckedStatus, (int)p->unAckedSeverity,
			p->lastAck.secPastEpoch);
	    p = p->nextSummaryPv;
	}

    }
