/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#ifndef _SHAREMEM_H_
#define	_SHAREMEM_H_

#include "alarmDefs.h"

#include <alarm.h>
#include <tsDefs.h>

/* General comments:
**
**	Strings: All static strings are represented as pointers into a general string space setup
**		 at initialization time.
**
**	Trees: The group tree is represented in leftmost-child/right-sibling notation.  The leftMost
**	       child pointer points to the leftmost child of a node's children.  Each child points to
**	       it's sibling to the right.
**
**	PV's: Each group node contains a pointer to the PV's under it.  Each PV maintains a pointer to
**	      the next PV in the list.  Each PV also maintains a pointer to it's parent group.

** WARNING!! This is the signature of the shared memory data structures. It serves to determine if the
** software matches the current structure. Increment this value if you change ANY of the the data
** structures defined in this file
*/

#define	AL_SIGNATURE	123456788

/* This value identifies the first piece of shared memory used by the alarm components. Additional
** pieces are numbered sequentially, i.e. AL_SHR_MEM_KEY+1, AL_SHR_MEM_KEY+2, ... 
*/

#define	AL_SHR_MEM_KEY	0x29834563

/* More definitions to ensure that everyone can map things into the same place in memory. */
#ifndef AXP
#ifdef WIN32
#define AL_SHARED_BASE	(char *)0x8000000	/* base for shared stuff, to allow different programs to share. */
#else
#define AL_SHARED_BASE	(char *)0xc00000	/* base for shared stuff, to allow different programs to share. */
#endif WIN32
#else
#define AL_SHARED_BASE	(char *)0		/* base for shared stuff, to allow different programs to share. */
#endif
#define	AL_INITIAL_SEGMENTS	2		/* Initial number of segments to allocate */
#define	AL_SEGMENT_SIZE		(1024*1024)	/* Size of each segment */
#define	AL_MAX_SEGMENTS		20		/* Maximum number of segments allowed */

#define	AL_TREESIZE	(1024*1024)		/* Size of the tree segment */
#define	AL_STRINGSIZE	(1024*512)		/* Size of the string table segment */

/* These structures contains all items that are used to accumulate data about a group.
** They are grouped into structures so that the recursive group tree labeling functions
** can be written cleaner by passing a single argument containing all applicable information.
*/
struct GroupData {
    short	notRTNnotACK [AL_NUM_SEVR];		/* Alarms not acked and not rtn */
    short	RTNnotACK [AL_NUM_SEVR];		/* Alarms not acked but rtn */
    short	notRTNandACK [AL_NUM_SEVR];		/* Alarms acked but not rtn */		
    short	numAlarmsBySeverity [AL_NUM_SEVR];	/* Total alarms in this group by severity */
    short	numUnackedBySeverity [AL_NUM_SEVR];	/* Total unacked alarms by severity */
    short	curStat [ALARM_NSTATUS];	/* Number of PV in each status */
    TS_STAMP	lastStatusChange;
    short	groupStatus;			/* 0 -> 3 */
    short	groupSeverity;			/* 0 -> 9 */
};

/* These are the parameters associated with each group.  They are grouped in a separate
** structure so that they may be easily passed as an argument to a recursive that will
** propogate them down the tree.
*/

typedef struct {
    char	*autoScript;
    char	*auxAutoScript;
    char	*manualScript;
    char	*auxManualScript;
    char	*displayName;
    char	*helpTopic;
    char	*guidance;
    char	*printers [AL_NUM_SEVR];
    char	*auxPrinters [AL_NUM_SEVR];
    char	*sounds [AL_NUM_SEVR];
    int		soundRates [AL_NUM_SEVR];
    char	*auxSounds [AL_NUM_SEVR];
    int		auxSoundRates [AL_NUM_SEVR];
    char	*auxSelectPV;
} GroupParameters;

/* One of these exists for each PV in the alarm tree. */

struct treeNode;
typedef struct pvData {
    struct pvData	*nextPv;	/* Next PV in the list */
    char		*pvName;	/* Name of the PV */
    struct treeNode	*parentGroup;	/* Group to which this PV belongs */
    short		Ack;		/* Alarm has been ACKed */
    short		RTN;		/* Return to normal */
    short		aenb;		/* Alarm enabled */
    short		oscn;		/* Point is on scan */
    short		status;		/* Most recent status */
    short		severity;	/* Most recent severity */
    short		curRawSevr;	/* Current raw severity */
    short		prevRawSevr;	/* Previous raw severity */
    PvValue		value;		/* Value of the PV when the alarm occurred */
    short		unAckedStatus;	/* Most recent unacked status */
    short		unAckedSeverity;/* Most recent unacked severity */
    PvValue		unAckedValue;	/* Most recent unacked value */
    Resource            resource;       /* resource for this point */ 
    IOCName             stationName;    /* Name of IOC (sort of) */
    Descriptor		description;	/* Description from pvMap */
    EguTag		eguTag;		/* Engineering Units */
    TS_STAMP		lastAlarm;	/* Time of last alarm in alarmlog */
    TS_STAMP		lastAlarmReceived;	/* Time of last alarm from CA */
    TS_STAMP		lastAck;	/* Time of last ACK */
    Operator		op;		/* Operator that ACKed last alarm */
    struct pvData	*nextSummaryPv;	/* Link onto summary list (forward) */
    struct pvData	*prevSummaryPv;	/* Link onto summary list (backward) */
} PvData;

/* This is the alarm tree node.  The tree in an n-ary tree implemented using
** leftmost child/right sibling representation.
*/

typedef struct treeNode {
    char		*groupName;	/* Name of node */
    struct treeNode	*leftChild;	/* Leftmost child */
    struct treeNode	*rightSibling;	/* Right sibling */
    struct treeNode	*parent;	/* Parent node */
    struct GroupData	groupData;	/* Various data for group */
    GroupParameters	groupParms;	/* Parameters for group */
    short		auxValue;	/* value of auxiliary control PV */
    short		chanNum;	/* Channel number in channel list */
    PvData		*pvList;	/* List of PV's in this group */
} TreeNode;

/* This header appears at the beginning of the shared memory segment. It contains pointers and
** sizes of the various structures in shared memory.
*/
typedef struct {
    long	shareSignature;		/* Identifies shared ememory structures */
    long	shareSegs;		/* Number of shared segments */
    TreeNode	*treeRoot;		/* Root node of group tree */
    PvData	*summaryHead;		/* Pointer to first entry in summary list */
    PvData	*summaryTail;		/* Pointer to last entry in summary list */
    PvData	*alarmPoints;		/* Pointer to base of alarmable PV's */
    long	alarmPointCount;	/* Number of PV's in alarm points list */
    int		archivingActive;	/* The archiver is currently running */
} ShareHeader;
 
#endif
