/*alarmLib.c,v 1.1.1.1 1999/08/03 14:41:44 tis4000 Exp */

/****************************************************************/
/*                                                              */
/*    This software is the proprietary property of Tate         */
/*    Integrated Systems, L.P.  Patents and copyrights are      */
/*    pending.                                                  */
/*                                                              */
/*    This software is furnished under a license and may        */
/*    be used and copied only in accordance with the terms      */
/*    of such license and with the inclusion of the above       */
/*    notice. This software or any other copies thereof         */
/*    may not be provided or otherwise made available to        */
/*    any other person. No title to and ownership of the        */
/*    software is hereby transferred.                           */
/*                                                              */
/*    The information in this software is subject to            */
/*    change without notice and should not be construed         */
/*    as a commitment by Tate Integrated Systems L.P. or        */
/*    its third party suppliers.                                */
/*                                                              */
/*    Tate Integrated Systems, L.P. (TIS) and its third         */
/*    party suppliers assume no responsibility for the          */
/*    use or inability to use any of its software. TIS          */
/*    software is provided "as is" without warranty of          */
/*    any kind and TIS expressly disclaims all implied          */
/*    warranties, including but not limited to the implied      */
/*    warranties of merchantability and fitness for a           */
/*    particular purpose.                                       */
/*                                                              */
/*    Notice: Notwithstanding any other lease or license        */
/*    that may pertain to or accompany the delivery of this     */
/*    computer software, the rights of the Government           */
/*    regarding its use, reproduction, and disclosure are       */
/*    as set forth in Section 52.227-19 of the FARS             */
/*    Computer Software-Restricted Rights clause.               */
/*                                                              */
/****************************************************************/

#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>

#ifdef WIN32
#include "NTUnixSubs.h"
#endif WIN32
#if !(defined(MIPSEL) || defined(AXP) || defined (WIN32))
#include <memory.h>
#else
#include <sys/types.h>
#include <sys/stat.h>
extern int lstat ();
extern int readlink ();
#endif

#include "tsDefs.h"
#include "alarmDefs.h"
#include "shareMem.h"
#include "alarmLib.h"
#include "msgMon.h"
#include "alarmLib.h"

static	int	CountGroups ();

#define	SHMSIZE	1024	/* maximum shared memory segment size (in Kbytes) */
#define	SHMMAX	(SHMSIZE * 1024)	/* max shared memory segment size */

ShareHeader	*shareHeader;
TreeNode	*treeRoot;

char	**nameArrayPtr;
char	**pvNameArrayPtr;
int	groupCount;

int InitAlarmAccess ()
{
    return MapSharedMemory ();
}

int MapSharedMemory ()
{
#ifdef WIN32
    int		shmkey = AL_SHR_MEM_KEY;
#else
    key_t	shmkey = AL_SHR_MEM_KEY;
#endif WIN32
    char	*shrMemSegPtr;
    int		numShrMemSegs;
    int 	shrMemId;
    int		*shrMemIds;
    int		i;

    if ((shrMemId = shmget (shmkey, AL_SEGMENT_SIZE, 0666)) == -1)
    {
	tisSysLog (0,0, "Alarm manager memory unavailable");
	return (0);
    }
    else if ((shrMemSegPtr = (char *)(shmat (shrMemId, (char *) AL_SHARED_BASE, 0))) == (char *) -1)
    {
	tisSysLog (0,0, "shmat failed, %d", errno);
	return (0);
    }
    shareHeader = (ShareHeader *)shrMemSegPtr;
    if (shareHeader->shareSignature != AL_SIGNATURE)
    {
	tisSysLog (0,0, "Shared memory is not correect format, %d", errno);
	perror ("Shared memory is not correct format");
	return (0);
    }

    numShrMemSegs = (int)shareHeader->shareSegs;
    treeRoot = shareHeader->treeRoot;
    if ( !(shrMemIds = (int *) calloc (numShrMemSegs, sizeof (int))))
    {
	tisSysLog (0,0, "Out of memory for shared memory ids");
	return (0);
    }
    shrMemIds[0] = shrMemId;
    for (i = 1; i < numShrMemSegs; i++)
    {
	if ((shrMemIds[i] = shmget (shmkey + i, AL_SEGMENT_SIZE, 0666)) == -1)
	{
	    tisSysLog (0,0, "shmget failed, %d", errno);
	    return (0);
	}
	else
	{
	    shrMemSegPtr = (char *)(shmat (shrMemIds[i], (char *)shareHeader+(i*AL_SEGMENT_SIZE), 0));
	    if  (shrMemSegPtr == (char *) -1)
	    {
		tisSysLog (0,0, "shmat failed, %d", errno);
		return (0);
	    }
	}
    }

    return (1);
}

int UnmapSharedMemory ()
{
    int		i;
    int		numShrMemSegs;

    numShrMemSegs = (int)shareHeader->shareSegs;
    for (i = 0; i < numShrMemSegs; i++)
    {
	if (shmdt ((char *)(AL_SHARED_BASE+(i*SHMMAX))) == -1)
	{
	    tisSysLog (0,0, "shmdt failed, %d", errno);
	    return (0);
	}
    }
    shareHeader = NULL;
    return (1);
}

PvData *FindPV (startNode, pvName)
TreeNode *startNode;
char *pvName;
{
    PvData	*pvPtr;
    TreeNode	*rootNode;
    TreeNode	*node;
/*
    if (startNode == 0)
	rootNode = startNode;
    else
	rootNode = treeRoot;
*/
    if (startNode == 0)
	rootNode = treeRoot;
    else
	rootNode = startNode;

    pvPtr = rootNode->pvList;
    while (pvPtr)
    {
	if (strcmp(pvPtr->pvName, pvName) == 0)
	{
	    return (pvPtr);
	}
	pvPtr = pvPtr->nextPv;
    }

    for (node = rootNode->leftChild; node; node = node->rightSibling)
    {
	if (pvPtr = FindPV (node, pvName))
	{
	    return (pvPtr);
	}
    }
    return NULL;
}

TreeNode *FindGroup (rootNode, group)
TreeNode *rootNode;
char *group;
{
    TreeNode	*nodePtr;
    TreeNode	*resultPtr;

    if (rootNode)
	resultPtr = rootNode;
    else
	resultPtr = treeRoot;	

    if (strcmp (resultPtr->groupName, group) == 0)
	return (resultPtr);

    for (nodePtr = resultPtr->leftChild; nodePtr; nodePtr = nodePtr->rightSibling)
    {
	if (resultPtr = FindGroup (nodePtr, group))
	    return (resultPtr);
    }
    return 0;
}

static void ScanGroups (rootNode)
TreeNode *rootNode;
{
    TreeNode	*nodePtr;

    *pvNameArrayPtr = rootNode->groupName;
    pvNameArrayPtr++;
    for (nodePtr = rootNode->leftChild; nodePtr; nodePtr = nodePtr->rightSibling)
    {
	ScanGroups (nodePtr);
    }
}

int GroupCount (rootNode)
TreeNode *rootNode;
{
    groupCount = 0;
    CountGroups (rootNode);
    return groupCount;
}

static int CountGroups (rootNode)
TreeNode *rootNode;
{
    TreeNode	*nodePtr;

    groupCount++;
    for (nodePtr = rootNode->leftChild; nodePtr; nodePtr = nodePtr->rightSibling)
    {
	CountGroups (nodePtr);
    }
    return groupCount;
}

/*********************************************************************************
**
** char **GetSubGroupNames (char * group)
**	Get all Subgroup names (including the root group) in a branch of the group
**	tree. Returns a pointer to an array of pointers.
**
**	NOTE: the returned pointer must be released by the caller, using free(..)!
**
*********************************************************************************/

char **GetSubgroupNames (group)
char *group;
{
    TreeNode	*nodePtr = treeRoot;
    char	**namePtrs;

    if (nodePtr = FindGroup (nodePtr, group))
    {
	namePtrs = (char **)calloc (GroupCount (nodePtr) + 1, sizeof(char *));
	nameArrayPtr = namePtrs;
	if (namePtrs)
	{
	    ScanGroups (nodePtr);
	    *namePtrs = NULL;
	}
	return nameArrayPtr;
    }
    return 0;
}

static void ScanGroupPVs (rootNode)
TreeNode *rootNode;
{
    TreeNode	*nodePtr;
    PvData	*pvPtr;

    /* Get the PVs for this group */

    for (pvPtr=rootNode->pvList ; pvPtr ; pvPtr = pvPtr->nextPv)
	*pvNameArrayPtr++ = pvPtr->pvName;	

    /* Get the PV's for all the subgroups */

    for (nodePtr = rootNode->leftChild; nodePtr; nodePtr = nodePtr->rightSibling)
	ScanGroupPVs (nodePtr);

}

/*********************************************************************************
**
** char **GetPVNames (char * group)
**	Get all PV names in a branch of the group tree. Returns a pointer to an
**	array of pointers.
**
**	NOTE: the returned pointer must be released by the caller, using free(..)!
**
*********************************************************************************/

char **GetPVNames (group)
char *group;
{
    TreeNode	*nodePtr = treeRoot;
    char	**pvNamePtrs;

    if (nodePtr = FindGroup (nodePtr, group))
    {
	pvNamePtrs = (char **)calloc (shareHeader->alarmPointCount + 1, sizeof(char *));
	pvNameArrayPtr = pvNamePtrs;
	if (pvNamePtrs)
	{
	    ScanGroupPVs (nodePtr);
	    *pvNameArrayPtr = NULL;
	}
	return pvNamePtrs;
    }
    return 0;
}
