#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vbuttonHEADER
#include vlistHEADER
#include vlistviewHEADER
#include vprintHEADER

extern "C"
{
#ifdef WIN32
#include <stdio.h>
#include <winsock.h>
// this bit of sillyness is needed because windows #defines AppendMenu
// as #define AppendMenu AppendMenuA, which interfers with galaxy class
// vmenubar::AppendMenu
#undef AppendMenu
#endif WIN32
#include "shareMem.h"
#include "tsDefs.h"
#include "alarmDefs.h"
#include "alarmfil.h"
#include "alarmvu.h"
}

#include "alvu.h"
#include "alarmvu.hxx"
#include "ackManyDialog.hxx"


/*******************************************************************
*
*    AckManyDialog Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(AckManyDialog, vdialog, "AckManyDialog");

/*******************************************************************
*
*    AckManyDialog constructor
*
*******************************************************************/

void AckManyDialog::LoadInit (vresource res)
{
    ackList	 = NULL;
    okToAck      = vFALSE;
    ackManyView  = vlistview::RequestSubLoad (this, alvu_ListView);
    okButton     = AckManyOkButton::RequestSubLoad (this, alvu_OkButton);
    cancelButton = AckManyCancelButton::RequestSubLoad (this, alvu_CancelButton);

    vdialog::LoadInit (res);

    // store ptr to dialog in buttons
    okButton     -> StoreAckManyDialog (this);
    cancelButton -> StoreAckManyDialog (this);
}


/*******************************************************************
*
*    AckManyDialog window observer
*
*******************************************************************/

void AckManyDialog::ObserveWindow (vevent *event)
{
    if (event -> GetType () == veventWINDOW_STATE && event -> IsClose ())
    {
	okToAck = vFALSE;
	Close ();
	vevent::StopProcessing ();
    }
}

/*******************************************************************
*
*    AckManyDialog destructor
*
*******************************************************************/

void AckManyDialog::Destroy ()
{
    if (ackList != NULL)
    {
	ackManyView -> SetList (NULL);
	delete ackList;
    }
    vdialog::Destroy ();
}


/*******************************************************************
*
*    AckManyDialog open function
*
*******************************************************************/

void AckManyDialog::OpenAckMany (alvuApplication *alvuapp, 
				 char *pvname, 
				 TS_STAMP *lastalarm)
{
    app = alvuapp;
    pv = pvname;
    lastAlarm = *lastalarm;

    BuildAckList ();

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open();
    vevent::Process ();
}


/*******************************************************************
*
*    AckManyDialog build list function
*
*******************************************************************/

void AckManyDialog::BuildAckList (void)
{
    int row = 0;
    ackList = new vliststr;
    AlarmLogRecord alarmRecord;
    char txtBuf[200];

    while (Log_RecordByTime(pv, &lastAlarm, &alarmRecord) == RECFOUND)
    {
	// stop loading list when we get a record that is acked.
	if (alarmRecord.acked)
	    break;

	sprintf (txtBuf, "%s %s  %s %-*.*s %-*.*s %-*.*s %-*.*s",
		 alarmRecord.date, alarmRecord.time, alarmRecord.pointName,
		 AL_DESCRIPTOR_SIZE - 4, AL_DESCRIPTOR_SIZE - 4, alarmRecord.descriptor,
		 AL_STATUS_SIZE, AL_STATUS_SIZE, alarmRecord.status,
		 VALACK_SIZE, VALACK_SIZE, alarmRecord.value,
		 AL_EGUTAG_SIZE, AL_EGUTAG_SIZE, alarmRecord.egutag);

	ackList -> SetCellValueScribed (row, 0, vcharScribeSystem (txtBuf));
	row++;
	if (row >= MAX_ACK_TO_DISPLAY)
	{
	    vresource msgRes = app -> messagesResource;
	    vstr *toMany = vresourceGetString (msgRes, alvu_DisplayManyMsg);
	    ackList -> SetCellValueScribed (row, 0, vcharScribeSystem (toMany));
	    vstrDestroy (toMany);
	    break;
	}

	// infinite loop protection
	if ((lastAlarm.secPastEpoch == alarmRecord.acktime.secPastEpoch)
	  && (lastAlarm.nsec == alarmRecord.acktime.nsec))
	    break; 

	lastAlarm = alarmRecord.acktime;
    }

    ackManyView -> SetList (ackList);
    vlistSelection *selected = ackManyView -> GetSelection ();
    ackList -> SetSelectMethod (selected, vlistNO_SELECTION);
}


/*******************************************************************
*
*    AckManyOkButton Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(AckManyOkButton, vbutton, "AckManyOkButton");

void AckManyOkButton::ObserveDialogItem (vevent *event)
{
    ackManyDialog -> okToAck = vTRUE;
    ackManyDialog -> Close ();
    vevent::StopProcessing ();
}



/*******************************************************************
*
*    AckManyCancelButton Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(AckManyCancelButton, vbutton, "AckManyCancelButton");

void AckManyCancelButton::ObserveDialogItem (vevent *event)
{
    ackManyDialog -> okToAck = vFALSE;
    ackManyDialog -> Close ();
    vevent::StopProcessing ();
}

