#ifndef alarmdialogsHXX
#define	alarmdialogsHXX

#ifdef WIN32
#define	PRINT_FILE_TEMP	"C:\\tis4000\\temp\\alarmvuXXXXXX"
#else
#define	PRINT_FILE_TEMP	"/tis4000/temp/alarmvuXXXXXX"
#endif WIN32

enum cntrlValues
{
    MOVECLICK = 0,
    WINCLOSE,
    ACKCLICK,
    NAMECLICK,
    SUMMARYCLICK,
    GROUPCLICK,
    GUIDECLICK,
    DISPLAYCLICK,
    DETAILCLICK,
    COMMANDCLICK,
    ENABLECLICK,
    DISABLECLICK,
    COMMENTCLICK,
    HELPCLICK,
    CAMONITOR,
    UPCLICK,
    TOPCLICK,
    EXITCLICK,
    PAUSECLICK,
    EXPANDCLICK
};

enum infotype
{
    NOINFO = 0,
    TRKINFO,
    LOGINFO
};

typedef struct Sys_Info 
{
    int		type;
    long	key;
    TS_STAMP	eventTime;
} SYS_DATA;

enum popUpMenuItems
{
    SUMMARY_MENUITEM = 0,
    GRP_MENUITEM,
    GUIDANCE_MENUITEM,
    DISPLAY_MENUITEM,
    DETAIL_MENUITEM,
    SCRIPT_MENUITEM,
    ENABLE_MENUITEM,
    DISABLE_MENUITEM,
    COMMENT_MENUITEM,
    HELP_MENUITEM,
    POPUP_COUNT
};


class ListEntry;
class CloseButton;
class UpButton;
class TopButton;
class PauseButton;
class ExitButton;
class AckButton;
class TextItem;
class FileButtonContainer;
class FirstButton;
class PageupButton;
class LineupButton;
class LinedownButton;
class PagedownButton;
class LastButton;
class CancelDataButtonContainer;
class CancelDataButton;


// *** ALARM DIALOG CLASS **********************************************

class AlarmDialog : public vdialog
{
    void	BuildPopupMenu (void);

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(AlarmDialog);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(AlarmDialog, vdialog);

    // vloadable override
    void LoadInit (vresource resource);

    // override 
    void ObserveWindow (vevent *event);

    // override
    int HandleConfigure (vevent *event);

    // override
    void Destroy ();

    void AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt);

    // base class virtuals to be overridden by sub-classes
    virtual void BuildMenus (void);
    virtual int  CtrlFunc (int cmd, int param);
    virtual void SetDialogTitle (void);
    virtual void UpdateBlink (void);
    virtual void WindowsPrint (int whatToPrint);
    virtual void UnixPrint (int whatToPrint);


    // stuff on the GUI
    vmenubar		*alarmMenuBar;
    AldlgListView	*alarmListView;
    CloseButton		*closeButton;
    UpButton		*upButton;
    TopButton		*topButton;
    PauseButton		*pauseButton;
    CancelDataButtonContainer *cancelDataContainer;
    ExitButton		*exitButton;
    vtextitem		*elementCount;
    FileButtonContainer	*fileButtonContainer;

    vmenu		*PopupMenu;	// for Ack button

    // various functions

    void	AddLine (int colorIdx, char *Date, char *Name,
			 char *Desc, TS_STAMP *altime = 0, int Mon = False);
    void	AddLine (int colorIdx, int ackM, char *Date, char *Name,
			 char *Desc, TS_STAMP *altime = 0, int Mon = False);
    int		CheckCancel (void);
    void	ClearLines (void);
    ListEntry	*GetEntryPtr (int row);
    char	*ItemDate (int row);
    char	*ItemDesc (int row);
    char	*ItemName (int row);
    void	MonitorCB (unsigned int ID);
    int		NumItems (void);
    int		Pos (void);
    void	SetBottom (void);
    void	SetPosition (int newPos);
    void	ShowPaused (int val);
    void	StartCancelBox (void);
    void	StopCancelBox (void);

    // inline functions

    int CheckCancelled (void)
	{ return (dataCancelled); }
    int FileBased (void)
	{ return (fileBased); }
    alvuApplication *GetApp (void)
	{ return (app); }
    int IsPaused (void)
	{ return (paused); }
    void SetDataCancelled (int val)
	{ dataCancelled = val; }
    void SetFileItems (int size)
	{ numFileItems = size; }
    void SetFilePos (int pos)
	{ filePos = pos; }
    int Size (void)
	{ return (displaySize); }

    // function for popup menu
    void DoPopup (vevent *event);


protected:

    int			ackable;
    alvuApplication	*app;
    long		checksCalled;
    int			dataCancelled;
    int			displaySize;
    int			fileBased;
    unsigned int	filePos;
    int			hasdate;
    int			hasdesc;
    int			monitored;
    unsigned int	numFileItems;
    int			paused;
    vstr		*title;
    vlist		*dataList;
};


// *** CLOSE BUTTON CLASS *****************************************

class CloseButton : public vbutton
{
    AlarmDialog	*alarmDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (CloseButton);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (CloseButton, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void StoreAlarmDialog (AlarmDialog *dlg) 
	{ alarmDialog = dlg; }
};

// *** UP BUTTON CLASS ********************************************

class UpButton : public vbutton
{
    AlarmDialog	*alarmDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (UpButton);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (UpButton, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void StoreAlarmDialog (AlarmDialog *dlg) 
	{ alarmDialog = dlg; }
};


// *** TOP BUTTON CLASS ******************************************

class TopButton : public vbutton
{
    AlarmDialog	*alarmDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (TopButton);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (TopButton, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void StoreAlarmDialog (AlarmDialog *dlg) 
	{ alarmDialog = dlg; }
};

// *** PAUSE BUTTON CLASS ***************************************

class PauseButton : public vbutton
{
    AlarmDialog	*alarmDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (PauseButton);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (PauseButton, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void StoreAlarmDialog (AlarmDialog *dlg) 
	{ alarmDialog = dlg; }

    // handle blinking
    void UpdateBlink ();
};


// *** EXIT BUTTON CLASS **************************************

class ExitButton : public vbutton
{
    AlarmDialog	*alarmDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (ExitButton);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (ExitButton, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void StoreAlarmDialog (AlarmDialog *dlg) 
	{ alarmDialog = dlg; }
};

// *** CANCEL DATA BUTTON CONTAINER CLASS *************************

class CancelDataButtonContainer : public vcontainer
{

public:

    // overrides for vkindTyped
    vkindTYPED_FULL_DECLARATION(CancelDataButtonContainer);

    // the default versions of necessary constructors/destructors
    vloadableINLINE_CONSTRUCTORS(CancelDataButtonContainer, vcontainer);

    // vloadable override
    void LoadInit (vresource resource);

    CancelDataButton	*cancelDataButton;
};


// *** CANCEL DATA BUTTON CLASS ***********************************

class CancelDataButton : public vbutton
{
    AlarmDialog	*alarmDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (CancelDataButton);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (CancelDataButton, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void StoreAlarmDialog (AlarmDialog *dlg) 
	{ alarmDialog = dlg; }
};

// *** FILE BUTTON CONTAINER CLASS *******************************

class FileButtonContainer : public vcontainer
{

public:

    // overrides for vkindTyped
    vkindTYPED_FULL_DECLARATION(FileButtonContainer);

    // the default versions of necessary constructors/destructors
    vloadableINLINE_CONSTRUCTORS(FileButtonContainer, vcontainer);

    // vloadable override
    void LoadInit (vresource resource);

    FirstButton	    *firstButton;
    PageupButton    *pageupButton;
    LineupButton    *lineupButton;
    LinedownButton  *linedownButton;
    PagedownButton  *pagedownButton;
    LastButton	    *lastButton;
};


// *** FIRST BUTTON CLASS **************************************

class FirstButton : public vbutton
{
    AlarmDialog	*alarmDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (FirstButton);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (FirstButton, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void StoreAlarmDialog (AlarmDialog *dlg) 
	{ alarmDialog = dlg; }
};

// *** PAGEUP BUTTON CLASS **************************************

class PageupButton : public vbutton
{
    AlarmDialog	*alarmDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (PageupButton);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (PageupButton, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void StoreAlarmDialog (AlarmDialog *dlg) 
	{ alarmDialog = dlg; }
};


// *** LINEUP BUTTON CLASS **************************************

class LineupButton : public vbutton
{
    AlarmDialog	*alarmDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (LineupButton);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (LineupButton, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void StoreAlarmDialog (AlarmDialog *dlg) 
	{ alarmDialog = dlg; }
};

// *** LINEDOWN BUTTON CLASS **************************************

class LinedownButton : public vbutton
{
    AlarmDialog	*alarmDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (LinedownButton);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (LinedownButton, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void StoreAlarmDialog (AlarmDialog *dlg) 
	{ alarmDialog = dlg; }
};


// *** PAGEDOWN BUTTON CLASS **************************************

class PagedownButton : public vbutton
{
    AlarmDialog	*alarmDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (PagedownButton);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (PagedownButton, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void StoreAlarmDialog (AlarmDialog *dlg) 
	{ alarmDialog = dlg; }
};

// *** LAST BUTTON CLASS **************************************

class LastButton : public vbutton
{
    AlarmDialog	*alarmDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION (LastButton);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS (LastButton, vbutton);

    // virtual member function override
    void ObserveDialogItem (vevent *event);

    // handle the pointer to the common dialog
    void StoreAlarmDialog (AlarmDialog *dlg) 
	{ alarmDialog = dlg; }
};



// ** CLASS FOR THE MAIN DIALOGS MENUITEMS **************************

class popupMenuItems : public vmenuItem
{
    AlarmDialog	*alarmDialog;
public:
    // macros for standard versions of the
    vkindTYPED_FULL_DECLARATION(popupMenuItems);

    // constructors and destructors
    vloadableINLINE_CONSTRUCTORS(popupMenuItems, vmenuItem);

    // function to handle the alarmDialog
    void StoreAlarmDialog (AlarmDialog *dlg)
	{ alarmDialog = dlg; }
    AlarmDialog *GetAlarmDialog (void)
	{ return (alarmDialog); }
};


// *** ALARM DIALOG CLASSES ******************************************

enum printWhat
{
    PRINT_THIS_WINDOW = 0,
    PRINT_ENTIRE_LIST,
    PRINT_SAVE_AS
};


// *** ALARM SUMMARY DIALOG CLASS ************************************
enum alarmSummMenuItems
{
    ALSUM_THISWIN_ITEM = 0,
    ALSUM_ENTIRESUM_ITEM,
    ALSUM_MODFILT_ITEM,
    ALSUM_SILENCE_ITEM,
    ALSUM_ACKALL_ITEM,
    ALSUM_ACKPAGE_ITEM,
    ALSUM_ITEM_COUNT
};
class AlarmSummaryDialog : public AlarmDialog
{

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(AlarmSummaryDialog);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(AlarmSummaryDialog, AlarmDialog);

    // vloadable override
    void LoadInit (vresource resource);

    // override
    void Destroy ();

    void AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt);

    // override base class funcs
    void BuildMenus (void);
    int  CtrlFunc (int cmd, int param);
    void SetDialogTitle (void);
    void UpdateBlink (void);
    void WindowsPrint (int whatToPrint);
    void UnixPrint (int whatToPrint);

    void	RefreshDisplay (void);
    void	Resync (void);

    Filter		filter;
    TreeNode		*groupLinkPtr;
    unsigned int	pause;

    // place to store internal names for menu items
    const vname		*menuNames [ALSUM_ITEM_COUNT];
};


class alarmSummaryCommands : public vcommandSelector
{
public:
    // macros for standard versions of the
    vkindTYPED_FULL_DECLARATION(alarmSummaryCommands);

    // constructors and destructors
    vloadableINLINE_CONSTRUCTORS(alarmSummaryCommands, vcommandSelector);

    //override virtual issue function
    int Issue(vdict *);
};


// *** DYNAMIC GROUP ALARM DIALOG CLASS ******************************
enum groupMenuItems
{
    ALGRP_THISWIN_ITEM = 0,
    ALGRP_ENTIRESUM_ITEM,
    ALGRP_MODFILT_ITEM,
    ALGRP_DISPLAY_ITEM,
    ALGRP_SCRIPT_ITEM,
    ALGRP_GUIDANCE_ITEM,
    ALGRP_DISABLE_PT_ITEM,
    ALGRP_ENABLE_PT_ITEM,
    ALGRP_DISABLE_GRP_ITEM,
    ALGRP_ENABLE_GRP_ITEM,
    ALGRP_ACKALL_ITEM,
    ALGRP_ACKPAGE_ITEM,
    ALGRP_ITEM_COUNT
};

class DynamicGroupDialog : public AlarmDialog
{

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(DynamicGroupDialog);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(DynamicGroupDialog, AlarmDialog);

    // vloadable override
    void LoadInit (vresource resource);

    // override
    void Destroy ();

    void AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt);

    // override base class funcs
    void BuildMenus (void);
    int  CtrlFunc (int cmd, int param);
    void SetDialogTitle (void);
    void UpdateBlink (void);
    void WindowsPrint (int whatToPrint);
    void UnixPrint (int whatToPrint);

    void	RefreshDisplay (void);
    void	Resync (void);

    Filter		filter;
    TreeNode		*groupLinkPtr;
 
    // place to store internal names for menu items
    const vname		*menuNames [ALGRP_ITEM_COUNT];
};


class dynamicGroupCommands : public vcommandSelector
{
public:
    // macros for standard versions of the
    vkindTYPED_FULL_DECLARATION(dynamicGroupCommands);

    // constructors and destructors
    vloadableINLINE_CONSTRUCTORS(dynamicGroupCommands, vcommandSelector);

    //override virtual issue function
    int Issue(vdict *);
};


// *** LOG FILE ALARM DIALOG CLASS ******************************

class LogFileDialog : public AlarmDialog
{

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(LogFileDialog);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(LogFileDialog, AlarmDialog);

    // vloadable override
    void LoadInit (vresource resource);

    // override
    void Destroy ();

    void AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt);

    // override base class funcs
    int  CtrlFunc (int cmd, int param);
    virtual void SetDialogTitle (void);
    void UpdateBlink (void);
    void WindowsPrint (int whatToPrint);
    void UnixPrint (int whatToPrint);

    void    RefreshDisplay (void);
    void    Resync (void);
    void    SetIndication (void);
    void    ShowCancelled (void);

    vstr	    *displayTitle;
    int		    filePosition;
    Filter	    filter;
    TreeNode	    *groupLinkPtr;
    int		    indexFileType;
    int		    inMove;
    int		    inRefresh;
    char	    lastName [AL_POINTNAME_SIZE + 1];
    TS_STAMP	    lastTime;
    unsigned int    pause;
    vstr	    *printTitle;
    unsigned int    severity;
    char	    startName [AL_POINTNAME_SIZE + 1];
    TS_STAMP	    startTime;
    char	    topName [AL_POINTNAME_SIZE + 1];
    TS_STAMP	    topTime;
 
    AlarmLogRecord alarmLogRecord;
    AlarmLogRecord alarmLogRecords [ALVU_MAXITEMS];
};


// *** ALARM HISTORY DIALOG CLASS ******************************
enum AlrmHistMenuItems
{
    ALRMHIST_THISWIN_ITEM = 0,
    ALRMHIST_ENTIRESUM_ITEM,
    ALRMHIST_MODFILT_ITEM,
    ALRMHIST_ITEM_COUNT
};

class AlarmHistoryDialog : public LogFileDialog
{

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(AlarmHistoryDialog);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(AlarmHistoryDialog, LogFileDialog);

    void AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt);

    // override base class funcs
    void BuildMenus (void);

    // place to store internal names for menu items
    const vname		*menuNames [ALRMHIST_ITEM_COUNT];
};

class alarmHistoryCommands : public vcommandSelector
{
public:
    // macros for standard versions of the
    vkindTYPED_FULL_DECLARATION(alarmHistoryCommands);

    // constructors and destructors
    vloadableINLINE_CONSTRUCTORS(alarmHistoryCommands, vcommandSelector);

    //override virtual issue function
    int Issue(vdict *);
};


// *** ALARM EVENT DIALOG CLASS ******************************
enum AlrmEvntMenuItems
{
    ALRMEVNT_THISWIN_ITEM = 0,
    ALRMEVNT_ENTIRESUM_ITEM,
    ALRMEVNT_MODFILT_ITEM,
    ALRMEVNT_ITEM_COUNT
};

class AlarmEventDialog : public LogFileDialog
{

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(AlarmEventDialog);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(AlarmEventDialog, LogFileDialog);

    void AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt);

    // override base class funcs
    void BuildMenus (void);

    // place to store internal names for menu items
    const vname		*menuNames [ALRMEVNT_ITEM_COUNT];
};

class alarmEventCommands : public vcommandSelector
{
public:
    // macros for standard versions of the
    vkindTYPED_FULL_DECLARATION(alarmEventCommands);

    // constructors and destructors
    vloadableINLINE_CONSTRUCTORS(alarmEventCommands, vcommandSelector);

    //override virtual issue function
    int Issue(vdict *);
};



// *** ALARM WINDOW DIALOG CLASS ******************************
enum AlrmWndwMenuItems
{
    ALRMWNDW_THISWIN_ITEM = 0,
    ALRMWNDW_ENTIRESUM_ITEM,
    ALRMWNDW_CLEAR_ITEM,
    ALRMWNDW_ITEM_COUNT
};

class AlarmWindowDialog : public LogFileDialog
{

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(AlarmWindowDialog);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(AlarmWindowDialog, LogFileDialog);

    void AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt);

    // override base class funcs
    void BuildMenus (void);

    // place to store internal names for menu items
    const vname		*menuNames [ALRMWNDW_ITEM_COUNT];
};

class alarmWindowCommands : public vcommandSelector
{
public:
    // macros for standard versions of the
    vkindTYPED_FULL_DECLARATION(alarmWindowCommands);

    // constructors and destructors
    vloadableINLINE_CONSTRUCTORS(alarmWindowCommands, vcommandSelector);

    //override virtual issue function
    int Issue(vdict *);
};


// *** OPERATOR TRACKING DIALOG CLASS ******************************
enum trackMenuItems
{
    OPTRK_THISWIN_ITEM = 0,
    OPTRK_ENTIRESUM_ITEM,
    OPTRK_MODFILT_ITEM,
    OPTRK_ITEM_COUNT
};

class AlarmTrackDialog : public AlarmDialog
{

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(AlarmTrackDialog);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(AlarmTrackDialog, AlarmDialog);

    // vloadable override
    void LoadInit (vresource resource);

    // override
    void Destroy ();

    void AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt);

    // override base class funcs
    void BuildMenus (void);
    int  CtrlFunc (int cmd, int param);
    void SetDialogTitle (void);
    void WindowsPrint (int whatToPrint);
    void UnixPrint (int whatToPrint);

    void	RefreshDisplay (void);
    void	Resync (void);

    Filter	filter;
    int		filePosition;
    int		reqPosition;
 
    // place to store internal names for menu items
    const vname		*menuNames [OPTRK_ITEM_COUNT];
};


class alarmTrackCommands : public vcommandSelector
{
public:
    // macros for standard versions of the
    vkindTYPED_FULL_DECLARATION(alarmTrackCommands);

    // constructors and destructors
    vloadableINLINE_CONSTRUCTORS(alarmTrackCommands, vcommandSelector);

    //override virtual issue function
    int Issue(vdict *);
};


// *** ALARM REPORTS DIALOG CLASS ******************************
enum reportsMenuItems
{
    RPTS_THISWIN_ITEM = 0,
    RPTS_ENTIRESUM_ITEM,
    RPTS_SAVEAS_ITEM,
    RPTS_MODFILT_ITEM,
    RPTS_ITEM_COUNT
};

class AlarmPointsDialog : public AlarmDialog
{

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(AlarmPointsDialog);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(AlarmPointsDialog, AlarmDialog);

    // vloadable override
    void LoadInit (vresource resource);

    // override
    void Destroy ();

    void AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt);

    // override base class funcs
    void BuildMenus (void);
    int  CtrlFunc (int cmd, int param);
    void SetDialogTitle (void);
    void WindowsPrint (int whatToPrint);
    void UnixPrint (int whatToPrint);

    void	RefreshDisplay (void);
    void	Resync (void);

    Filter	filter;
    int		filePosition;
    RepCriteria	criteria;
 
    // place to store internal names for menu items
    const vname		*menuNames [RPTS_ITEM_COUNT];
};


class alarmPointCommands : public vcommandSelector
{
public:
    // macros for standard versions of the
    vkindTYPED_FULL_DECLARATION(alarmPointCommands);

    // constructors and destructors
    vloadableINLINE_CONSTRUCTORS(alarmPointCommands, vcommandSelector);

    //override virtual issue function
    int Issue(vdict *);
};


// *** ALARM SYSTEMRM DIALOG CLASS ******************************
enum alrmSysMenuItems
{
    ALSYS_THISWIN_ITEM = 0,
    ALSYS_ENTIRESUM_ITEM,
    ALSYS_MODFILT_ITEM,
    ALSYS_ITEM_COUNT
};

class AlarmSystemDialog : public AlarmDialog
{

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(AlarmSystemDialog);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(AlarmSystemDialog, AlarmDialog);

    // vloadable override
    void LoadInit (vresource resource);

    // override
    void Destroy ();

    void AlarmDialogOpen (alvuApplication *alvuApp, Filter *filt);

    // override base class funcs
    void BuildMenus (void);
    int  CtrlFunc (int cmd, int param);
    void SetDialogTitle (void);
    void WindowsPrint (int whatToPrint);
    void UnixPrint (int whatToPrint);

    void    RefreshDisplay (void);
    void    Resync (void);
    void    ShowCancelled (void);

    Filter		filter;
    int			filePosition;
    TreeNode		*groupLinkPtr;
    int			inMove;
    int			inRefresh;
    unsigned int	numEntries;
    int			reqPosition;
    unsigned int	severity;
    SYS_DATA		srcType [ALVU_MAXITEMS + 1];

    // alarmlog values
    int			trkPosition;
    char		startName [AL_POINTNAME_SIZE + 1];
    TS_STAMP		startTime;
    char		topName [AL_POINTNAME_SIZE + 1];
    TS_STAMP		topTime;
    char		lastName [AL_POINTNAME_SIZE + 1];
    TS_STAMP		lastTime;
 
    // place to store internal names for menu items
    const vname		*menuNames [ALSYS_ITEM_COUNT];
};


class alarmSystemCommands : public vcommandSelector
{
public:
    // macros for standard versions of the
    vkindTYPED_FULL_DECLARATION(alarmSystemCommands);

    // constructors and destructors
    vloadableINLINE_CONSTRUCTORS(alarmSystemCommands, vcommandSelector);

    //override virtual issue function
    int Issue(vdict *);
};

#endif alarmdialogsHXX

