/****************************************************************/
/*                                                              */
/*    This software is the proprietary property of Tate         */
/*    Integrated Systems, L.P.  Patents and copyrights are      */
/*    pending.                                                  */
/*                                                              */
/*    This software is furnished under a license and may        */
/*    be used and copied only in accordance with the terms      */
/*    of such license and with the inclusion of the above       */
/*    notice. This software or any other copies thereof         */
/*    may not be provided or otherwise made available to        */
/*    any other person. No title to and ownership of the        */
/*    software is hereby transferred.                           */
/*                                                              */
/*    The information in this software is subject to            */
/*    change without notice and should not be construed         */
/*    as a commitment by Tate Integrated Systems L.P. or        */
/*    its third party suppliers.                                */
/*                                                              */
/*    Tate Integrated Systems, L.P. (TIS) and its third         */
/*    party suppliers assume no responsibility for the          */
/*    use or inability to use any of its software. TIS          */
/*    software is provided "as is" without warranty of          */
/*    any kind and TIS expressly disclaims all implied          */
/*    warranties, including but not limited to the implied      */
/*    warranties of merchantability and fitness for a           */
/*    particular purpose.                                       */
/*                                                              */
/*    Notice: Notwithstanding any other lease or license        */
/*    that may pertain to or accompany the delivery of this     */
/*    computer software, the rights of the Government           */
/*    regarding its use, reproduction, and disclosure are       */
/*    as set forth in Section 52.227-19 of the FARS             */
/*    Computer Software-Restricted Rights clause.               */
/*                                                              */
/****************************************************************/

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "time.h"
#include "ctype.h"
#include "fcntl.h"

#if !(defined(MIPSEL) || defined(AXP))
#include "memory.h"
#else
#include <sys/types.h>
#include <sys/stat.h>
extern int lstat ();
extern int readlink ();
#endif

#if defined(SOLARIS) || defined(WIN32)
#define bzero(b,n)	memset (b, 0, n)
#endif

#include "shareMem.h"

#ifdef USEX11
#include <X11/Intrinsic.h>
#include <X11/Xlib.h>
#endif USEX11

#include "alarm.h"
#include "alarmvu.h"
#ifdef WIN32
#include <winsock.h>
#else
#include <netinet/in.h>
#endif WIN32

#include "OperPriv.h"

#include "tsDefs.h"
#include "alarmDefs.h"
#include "messages.h"
#include "alarmfil.h"
#include "alarmString.h"
/* #include "alarminfo.h" */

#ifndef AXP
extern time_t timezone;
#endif AXP

/*----- log file stuff -----*/
#include "dbFormats.c"

static void AdjustCounts ();
static int RecIsFiltered ();
static void TrkAdjustCounts ();
static int TrkRecIsFiltered ();

C4CODE		codeBaseDesc;
static D4DATA	*filtersDesc;
static D4DATA	*reportsDesc;
static D4DATA	*alarmLogDesc;
static D4DATA	*alarminfDesc;

extern char	*alarmSeverityString[];
extern char	*validSeverityString[];
extern char	*alarmStatusString[];
extern char	**groupNames;
extern char	linePrinter[LINEPRINTER_SIZE+1];
char		alarmvuLogName [100];
char		filterFileName [100];
char		reportFileName [100];
extern int	alarmhndSocket;

extern ShareHeader *shareHeader;
extern TreeNode	*treeRoot;
int		numSortedPoints;
PvData		**pointList;


char	dbfilename [132]; /* storage for data base file name */
F4FIELD	*field_ptr;	/* Handle to field in record */

char    dbfile [100];   /* current DB file */

char	linkedname [100]; /* storage for link-resolved name */
char	filename [100]; /* storage for link-resolved file name */
char	pathpart [100]; /* path part of filename */
char	filepart [100]; /* file part of filename */
char	pathname [100]; /* path part of filename */

/******************************************************************************
 *                                                                            *
 *     static variables                                                       *
 *                                                                            *
 ******************************************************************************/

#ifndef True
#define True	(1==1)
#define False	(1==0)
#endif True

static long GetGMT_Secs ( );

int	logRecCount;
int	logCounts [NUMIXFILES];
int	logRecPosition [NUMIXFILES];
unsigned int	logTagPosition [NUMIXFILES];
Filter	logFilter [NUMIXFILES];
long	lastPointSec;
long	lastPointMsecs;
PointName lastPointName;

int	trkRecCount;
int	trkCounts;
int	trkRecPosition;
unsigned int	trkTagPosition;
Filter	trkFilter;

F4FIELD * alarmsecsFld;
F4FIELD * alarmmsecsFld;
F4FIELD * pointnameFld;
F4FIELD * descriptorFld;
F4FIELD * statusFld;
F4FIELD * severityFld;
F4FIELD * valueFld;
F4FIELD * ackedFld;
F4FIELD * egutagFld;
F4FIELD * acksecsFld;
F4FIELD * ackmsecsFld;
F4FIELD * operatorFld;
F4FIELD * commentFld;
F4FIELD * stationFld;
F4FIELD * info1Fld;
F4FIELD * info2Fld;
F4FIELD * info3Fld;
F4FIELD * info4Fld;
F4FIELD * info5Fld;

F4FIELD_INFO filterFields [] = {
	{ "FILTERNAME",		'C',	AL_FILTERNAME_SIZE,	0 },	/* Name of filter */
	{ "GROUP",		'C',	AL_GROUPNAME_SIZE,	0 },	/* group name */
	{ "SEVRESTR",		'N',	9,			0 },	/* severity restriction */
	{ "ACKRESTR",		'N',	9,			0 },	/* acknowledged restriction */
	{ "GROUPRESTR",		'N',	9,			0 },	/* group alarms only restriction */
	{ "STARTTIME",		'C',	AL_DATE_SIZE+1+AL_TIME_SIZE, 0 }, /* start time */
	{ "ENDTIME",		'C',	AL_DATE_SIZE+1+AL_TIME_SIZE, 0 }, /* end time */
	{ "OPERATOR",		'C',	AL_OPERATOR_SIZE,	0 },	/* operator */
	{ "STARTUPDSP",		'N',	9,			0 },	/* startup display value */
	{ 0, 0, 0, 0 }
};

T4TAG_INFO filterTags [] = {
	{ "FILTERTAG",	"FILTERNAME",	".NOT. DELETED()", 0, 0 },
	{ 0, 0, 0, 0, 0 }
};

F4FIELD_INFO reportFields [] = {
	{ "REPORTNAME",		'C',	REPORTNAME_SIZE,	0 },	/* Name of report */
	{ "RPTITEMS",		'N',	15,			0 },	/* report criteria items */
	{ 0, 0, 0, 0 }
};

T4TAG_INFO reportTags [] = {
	{ "REPORTTAG",	"REPORTNAME",	".NOT. DELETED()", 0, 0 },
	{ 0, 0, 0, 0, 0 }
};

void resolveLinks (char	*orig, char *resolved);

/*----------*/
resolveLink (file)
char *file;
{
    resolveLinks (file, filename);
    return 0;
}


int OpenCodeBase ()
{
	d4init (&codeBaseDesc);
#if defined (WIN32) || defined(USECB63)
	codeBaseDesc.errFieldName = 0;
	codeBaseDesc.errOpen = 0;
	codeBaseDesc.errTagName = 0;
	codeBaseDesc.readLock = 0;
	codeBaseDesc.errCreate = 0;
	codeBaseDesc.errGo = 0;
	codeBaseDesc.errSkip = 0;
#else
	codeBaseDesc.field_name_error = 0;
	codeBaseDesc.open_error = 0;
	codeBaseDesc.tag_name_error = 0;
	codeBaseDesc.read_lock = 0;
	codeBaseDesc.create_error = 0;
	codeBaseDesc.go_error = 0;
	codeBaseDesc.skip_error = 0;
#endif WIN32

#ifdef WIN32
	strcpy (alarmvuLogName, "c:\\tis4000db\\alarmlog.dbf");
#else
	strcpy (alarmvuLogName, "/tis4000db/alarmlog.dbf");
#endif WIN32

	if (Log_AlarmLog () != NO_ERR)
	    return FALSE;
	return (TRUE);
}

int myOperPriv (key)
char *key;
{
    return OperPriv ( &codeBaseDesc, key );
}

int Log_AlarmLog ( )
{
    int i;

    if ( !(alarmLogDesc = d4open (&codeBaseDesc, alarmvuLogName)))
    {
	return (OPEN_ERR);
    }

    alarmsecsFld = d4field (alarmLogDesc, "ALARMSECS");
    alarmmsecsFld = d4field (alarmLogDesc, "ALARMMSECS");
    pointnameFld = d4field (alarmLogDesc, "POINTNAME");
    descriptorFld = d4field (alarmLogDesc, "DESCRIPTOR");
    statusFld = d4field (alarmLogDesc, "STATUS");
    severityFld = d4field (alarmLogDesc, "SEVERITY");
    valueFld = d4field (alarmLogDesc, "VALUE");
    ackedFld = d4field(alarmLogDesc, "ACKED");
    egutagFld = d4field (alarmLogDesc, "EGUTAG");
    acksecsFld = d4field (alarmLogDesc, "ACKSECS");
    ackmsecsFld = d4field (alarmLogDesc, "ACKMSECS");
    operatorFld = d4field (alarmLogDesc, "OPERATOR");
    commentFld = d4field( alarmLogDesc, "COMMENT");
    stationFld = d4field( alarmLogDesc, "STATION");
    info1Fld = d4field( alarmLogDesc, "INFO1");
    info2Fld = d4field( alarmLogDesc, "INFO2");
    info3Fld = d4field( alarmLogDesc, "INFO3");
    info4Fld = d4field( alarmLogDesc, "INFO4");
    info5Fld = d4field( alarmLogDesc, "INFO5");

    d4reccount( alarmLogDesc );	/* d4open() set num_recs to -1, get the real count */
    d4unlock_all (alarmLogDesc);

    /* initialize filters, counts, and positions */
    for (i = 0; i <NUMIXFILES; i++)
    {
	logCounts [i] = 0;
	logTagPosition [i] = 1;
	logRecPosition [i] = 1;
    }
    GetFilter ("default", &logFilter [ALRM_INDEX_FILE]); /* preset and never changed */
/*    GetFilter ("default", &logFilter [ALRMLOG_INDEX_FILE]);
**    GetFilter ("default", &logFilter [ALRMEVT_INDEX_FILE]);
**    AdjustCounts ();
*/
    return (NO_ERR);
}

Log_CloseAlarmLog ( )
{
    return (NO_ERR);
}

unsigned int Log_NumRecords (indexFile)
 int indexFile;
{
    if (indexFile == ALRM_INDEX_FILE)
	return d4reccount (alarmLogDesc); /* can't change the base filter */

    AdjustCounts ();
    return logCounts [indexFile];
}

int Log_RecordByTime ( pointname, eventtime, alarmLogRecordPtr)
    char *pointname;
    TS_STAMP *eventtime;
    AlarmLogRecord *alarmLogRecordPtr;
{
    char *seekpt;
    int  d4stat;

	seekpt = malloc (AL_GMTSECS_SIZE+AL_GMTMSECS_SIZE+AL_POINTNAME_SIZE+1+1); 
	d4tag_select (alarmLogDesc, d4tag (alarmLogDesc, "EVENTTIME"));

	if (pointname)
	{
	    sprintf (seekpt, "%*d%*d%-*.*s",
				AL_GMTSECS_SIZE, eventtime->secPastEpoch + TS_EPOCH_SEC_PAST_1970,
				AL_GMTMSECS_SIZE, eventtime->nsec / 1000000L,
				AL_POINTNAME_SIZE, AL_POINTNAME_SIZE, pointname);
	}
	else
	{
	    sprintf (seekpt, "%*d%*d",
				AL_GMTSECS_SIZE, eventtime->secPastEpoch + TS_EPOCH_SEC_PAST_1970,
				AL_GMTMSECS_SIZE, eventtime->nsec / 1000000L,
				AL_POINTNAME_SIZE, AL_POINTNAME_SIZE, pointname);
	}
	d4stat = d4seek (alarmLogDesc, seekpt);
	free (seekpt);
	if ((d4stat != r4success) && (d4stat != r4after))
	{
	    d4tag_select (alarmLogDesc, NULL);
	    d4unlock_all (alarmLogDesc);
	    return NORECORD;
	}

	LoadRecord (alarmLogRecordPtr);
	d4unlock_all (alarmLogDesc);
	if (d4stat == r4success)
	{
	    return RECFOUND;
	}
	else
	{
	    return RECAFTER;
	}
}

int Log_Record ( recordNum, alarmLogRecordPtr, indexFile )
    unsigned int recordNum;
    AlarmLogRecord *alarmLogRecordPtr;
    int indexFile;
{
    int		inc;
    int		filtstat;
    int		dbaseStatus;
    d4tag_select (alarmLogDesc, d4tag (alarmLogDesc, "EVENTTIME"));

    bzero((char *) alarmLogRecordPtr, sizeof (AlarmLogRecord));

    if (indexFile == ALRM_INDEX_FILE) /* no filters allowed, we can go right to it */
    {
	logRecPosition[indexFile] = recordNum;
	logTagPosition[indexFile] = recordNum;
    }

    /* First check for any shifting in the file */
    if ((d4top (alarmLogDesc) != 0) || (d4skip (alarmLogDesc, logRecPosition[indexFile])!= 0)
      || ((indexFile != ALRM_INDEX_FILE) && (RecIsFiltered (alarmLogDesc, &logFilter[indexFile]) != FILTOK)))
    {
	/* the record set or filter have changed, start at beginning */
	logRecPosition[indexFile] = 1;
	logTagPosition[indexFile] = 1;
    }

    if (logTagPosition[indexFile] > recordNum)
    {
	inc = -1;
    }
    else if (logTagPosition[indexFile] <= recordNum)
    {
	inc = 1;
    }

    dbaseStatus = ((d4top (alarmLogDesc) == 0)
      && (d4skip (alarmLogDesc, logRecPosition[indexFile]-1) == 0));
    while (dbaseStatus)
    {
	filtstat = RecIsFiltered (alarmLogDesc, &logFilter[indexFile]);
	if ((indexFile == ALRM_INDEX_FILE) || (filtstat == FILTOK))
	{
	    if (recordNum == logTagPosition[indexFile])
	    {
		/* we have it, return the data */
		LoadRecord (alarmLogRecordPtr);
		d4unlock_all (alarmLogDesc);
		return True; /* return OK */
	    }
	    else /* adjust our position in the file */
	    {
		logTagPosition[indexFile] += inc;
		logRecPosition[indexFile] += inc;
	    }
	}
	else if ((inc == -1 && (filtstat == FILTOUTOFRANGELOW))
	  || (inc == 1 && (filtstat == FILTOUTOFRANGEHIGH)))
	{
	    break;
	}
	else /* try to find a qualifying record in the proper direction */
	{
	    logRecPosition[indexFile] += inc;
	}
	dbaseStatus = (d4skip (alarmLogDesc, inc) == 0);
    }
    /* Not found */
    d4unlock_all (alarmLogDesc);
    return False;
}

int Log_RecordsByTime ( name, eventtime, alarmLogRecordPtrs, numRecs, indexFile, callbackfunc, arg )
    char *name;
    TS_STAMP *eventtime;
    AlarmLogRecord alarmLogRecordPtrs [];
    int numRecs;
    int indexFile;
    int (*callbackfunc)( );
    void *arg;
{
    int		loopcnt;
    int		retCnt;
    int		recix;
    int		inc;
    int		filtstat;
    int		i;
    int		dbaseStatus;
    TS_STAMP	temptime;

    d4tag_select (alarmLogDesc, d4tag (alarmLogDesc, "EVENTTIME"));

    if (numRecs < 0) { /* collect backward */
	inc = -1;
	loopcnt = -numRecs;
	retCnt = loopcnt;
	recix = loopcnt-1;
	if (name == 0) { /* start at end of file or at filter end time */
	    if ((indexFile != ALRM_INDEX_FILE) && strlen (logFilter[indexFile].end)) {
		/* find first record that is in range */
		temptime.secPastEpoch = GetGMT_Secs (logFilter[indexFile].end) - TS_EPOCH_SEC_PAST_1970;
		temptime.nsec = 0;
#if !(defined(AXP) || defined(SOLARIS) || defined(WIN32))
		if (Log_RecordByTime (name, &temptime, alarmLogRecordPtrs [recix]) == RECAFTER) {
#else
		if (Log_RecordByTime (name, &temptime, &alarmLogRecordPtrs [recix]) == RECAFTER) {
#endif
		    /* now skip back, to a record that is in range */
		    if (d4skip (alarmLogDesc, -1) != 0) {
			return 0;
		    }
		} else {
		    return 0;
		}
	    } else {
		dbaseStatus = d4bottom (alarmLogDesc);
		if (dbaseStatus)
		    return 0;
	    }
	}
	else { /* start at specified record */
#if !(defined(AXP) || defined(SOLARIS) || defined(WIN32))
	    if (Log_RecordByTime (name, eventtime, alarmLogRecordPtrs [recix]) == NORECORD) {
#else
	    if (Log_RecordByTime (name, eventtime, &alarmLogRecordPtrs [recix]) == NORECORD) {
#endif
		dbaseStatus = d4bottom (alarmLogDesc);
		if (dbaseStatus)
		    return 0;
	    }
	}
    }
    else { /* collect forward */
	inc = 1;
	loopcnt = numRecs;
	retCnt = loopcnt;
	recix = 0;
	/* from specified record */
#if !(defined(AXP) || defined(SOLARIS) || defined(WIN32))
	if (Log_RecordByTime (name, eventtime, alarmLogRecordPtrs [0]) == NORECORD) {
#else
	if (Log_RecordByTime (name, eventtime, &alarmLogRecordPtrs [0]) == NORECORD) {
#endif
	    return 0;
	}
    }

    while (loopcnt && !d4deleted (alarmLogDesc))
    {
	filtstat = RecIsFiltered (alarmLogDesc, &logFilter[indexFile]);
	if ((inc == -1 && (filtstat == FILTOUTOFRANGELOW))
	  || (inc == 1 && (filtstat == FILTOUTOFRANGEHIGH)))
	{
	    break;
	}
	if ((indexFile == ALRM_INDEX_FILE) || (filtstat == FILTOK))
	{
	    if (!d4deleted (alarmLogDesc))
	    {
		bzero((char *)&alarmLogRecordPtrs[recix], sizeof (AlarmLogRecord));
		LoadRecord ( &alarmLogRecordPtrs[recix] );
		recix += inc;
		loopcnt --;
	    }
	}
	if (loopcnt && (d4skip (alarmLogDesc, inc) != 0))
	{
	    break;
	}
	if (callbackfunc)
	    if ((*callbackfunc)(arg))
		break;
    }

    if (inc < 0)
    {
	if (loopcnt) /* if we don't have it all, move recs to correct location in array */
	{
	    for (i = 0; i < retCnt-loopcnt; i++)
	    {
		alarmLogRecordPtrs [i] = alarmLogRecordPtrs [i + loopcnt];
	    }
	}
    }
    d4unlock_all (alarmLogDesc);
    return (retCnt - loopcnt);
}

void LoadRecord (alarmLogRecordPtr)
AlarmLogRecord *alarmLogRecordPtr;
{
    static F4FIELD *fieldPtr;
    time_t unixtime;
    struct tm	*curTimePtr;

    memset (alarmLogRecordPtr, 0, sizeof(AlarmLogRecord));
    alarmLogRecordPtr->eventtime.secPastEpoch = f4long (alarmsecsFld) - TS_EPOCH_SEC_PAST_1970;
    alarmLogRecordPtr->eventtime.nsec = f4long (alarmmsecsFld) * 1000000L;
    unixtime = alarmLogRecordPtr->eventtime.secPastEpoch + TS_EPOCH_SEC_PAST_1970;
    curTimePtr = localtime (&unixtime);
    sprintf (alarmLogRecordPtr->date, "%02d/%02d/%02d",
		curTimePtr->tm_mon+1, curTimePtr->tm_mday, curTimePtr->tm_year % 100);
    sprintf (alarmLogRecordPtr->time, "%02d:%02d:%02d",
		curTimePtr->tm_hour, curTimePtr->tm_min, curTimePtr->tm_sec);
    strncpy (alarmLogRecordPtr->pointName, f4str (pointnameFld), AL_POINTNAME_SIZE);
    alarmLogRecordPtr->pointName[AL_POINTNAME_SIZE] = 0;
    trimblanks (alarmLogRecordPtr->pointName);
    strncpy (alarmLogRecordPtr->descriptor, f4str (descriptorFld), AL_DESCRIPTOR_SIZE);
    alarmLogRecordPtr->descriptor[AL_DESCRIPTOR_SIZE] = 0;
    trimblanks (alarmLogRecordPtr->descriptor);
    alarmLogRecordPtr->acked = atoi (f4str (d4field (alarmLogDesc, "ACKED")));
    strncpy (alarmLogRecordPtr->status, f4str (statusFld), AL_STATUS_SIZE);
    alarmLogRecordPtr->status[AL_STATUS_SIZE] = 0;
    trimblanks (alarmLogRecordPtr->status);
    strncpy (alarmLogRecordPtr->severity, f4str (severityFld), AL_SEVERITY_SIZE);
    alarmLogRecordPtr->severity[AL_SEVERITY_SIZE] = 0;
    trimblanks (alarmLogRecordPtr->severity);
    strncpy (alarmLogRecordPtr->value, f4str (valueFld), AL_VALUE_SIZE);
    alarmLogRecordPtr->value[AL_VALUE_SIZE] = 0;
    trimblanks (alarmLogRecordPtr->value);
    strncpy (alarmLogRecordPtr->egutag, f4str (egutagFld), AL_EGUTAG_SIZE);
    alarmLogRecordPtr->egutag[AL_EGUTAG_SIZE] = 0;
    trimblanks (alarmLogRecordPtr->egutag);
    alarmLogRecordPtr->acktime.secPastEpoch = f4long (acksecsFld) - TS_EPOCH_SEC_PAST_1970;
    alarmLogRecordPtr->acktime.nsec = f4long (ackmsecsFld) * 1000000L;
    strncpy (alarmLogRecordPtr->operatorName, f4str (operatorFld), AL_OPERATOR_SIZE);
    alarmLogRecordPtr->operatorName[AL_OPERATOR_SIZE] = 0;
    trimblanks (alarmLogRecordPtr->operatorName);
    if (stationFld)
    {
	strncpy (alarmLogRecordPtr->stationName, f4str (stationFld), AL_IOCNAME_SIZE);
	trimblanks (alarmLogRecordPtr->stationName);
    }
    else
	alarmLogRecordPtr->stationName [0] = 0;
    if (info1Fld)
    {
	strncpy (alarmLogRecordPtr->info1, f4str (info1Fld), AL_INFO_SIZE);
	trimblanks (alarmLogRecordPtr->info1);
	strncpy (alarmLogRecordPtr->info2, f4str (info2Fld), AL_INFO_SIZE);
	trimblanks (alarmLogRecordPtr->info2);
	strncpy (alarmLogRecordPtr->info3, f4str (info3Fld), AL_INFO_SIZE);
	trimblanks (alarmLogRecordPtr->info3);
	strncpy (alarmLogRecordPtr->info4, f4str (info4Fld), AL_INFO_SIZE);
	trimblanks (alarmLogRecordPtr->info4);
	strncpy (alarmLogRecordPtr->info5, f4str (info5Fld), AL_INFO_SIZE);
	trimblanks (alarmLogRecordPtr->info5);
    }
    else
    {
	alarmLogRecordPtr->info1 [0] = 0;
	alarmLogRecordPtr->info2 [0] = 0;
	alarmLogRecordPtr->info3 [0] = 0;
	alarmLogRecordPtr->info4 [0] = 0;
	alarmLogRecordPtr->info5 [0] = 0;
    }

    fieldPtr = d4field( alarmLogDesc, "COMMENT" );		/* Retrieve memo text */
    m4ncpy ( fieldPtr, alarmLogRecordPtr->comment, AL_MAX_MEMO_SIZE);
    alarmLogRecordPtr->comment[AL_MAX_MEMO_SIZE] = 0;
}

int Log_SetFilter ( filter, indexFile )
    Filter *filter;
    int indexFile;
{
    if (indexFile == ALRM_INDEX_FILE)
	return 0; /* can't change the base filter */

    if (memcmp (&logFilter [indexFile], filter, sizeof(Filter)))
    {
	logFilter [indexFile] = *filter;
	logRecCount = 0; /* force a recount */
	/* AdjustCounts (); -- but don't count 'em till we're asked */
    }
    return 0;
}

static void AdjustCounts ()
{
    int status;
    int	filtstat;
    int i;
    PointName pointName;

    d4tag_select (alarmLogDesc, d4tag (alarmLogDesc, "EVENTTIME"));

    /* adjust record counters, if total count is different */
    if (d4reccount (alarmLogDesc) != logRecCount)
    {
	/* if the old last record is still in the same place, just count the new guys */
	logCounts[0] = d4reccount (alarmLogDesc);
	status = d4top (alarmLogDesc);
	if (status == 0)
	{
	    if (logRecCount == 0) /* need to recount */
	    {
		for (i = 1; i < NUMIXFILES; i++)
		{
		    logCounts [i] = 0;
		    logTagPosition [i] = 1;
		    logRecPosition [i] = 1;
		}
	    }
	    else if (d4skip (alarmLogDesc, logRecCount-1) == 0) /* go to old last record */
	    {
		/* is it the same? */
		strncpy (pointName, f4str (pointnameFld), AL_POINTNAME_SIZE);
		trimblanks (pointName);
		if ((lastPointSec != f4long (alarmsecsFld))
		  || (lastPointMsecs != f4long (alarmmsecsFld))
		  || strcmp (lastPointName, pointName))
		{
		    /* no, do a total recount */
		    for (i = 1; i < NUMIXFILES; i++)
		    {
			logCounts [i] = 0;
			logTagPosition [i] = 1;
			logRecPosition [i] = 1;
		    }
		    status = d4top (alarmLogDesc);
		}
		else /* just count the new stuff, start after the old last guy */
		{
		    status = d4skip (alarmLogDesc, 1);
		}
	    }
	    do
	    {
		for (i = 1; i < NUMIXFILES; i++)
		{
		    filtstat = RecIsFiltered (alarmLogDesc, &logFilter[i]);
		    if (filtstat == FILTOK)
			logCounts[i]++;
		}
	    }
	    while ((filtstat != FILTOUTOFRANGEHIGH) && (d4skip (alarmLogDesc, 1) == 0));

	    /* remember last record info, for optimization */
	    logRecCount = logCounts [0];
	    if (d4skip (alarmLogDesc, -1) == 0)
	    {
		lastPointSec = f4long (alarmsecsFld);
		lastPointMsecs = f4long (alarmmsecsFld);
		strncpy (lastPointName, f4str (pointnameFld), AL_POINTNAME_SIZE);
		trimblanks (lastPointName);
	    }
	}
    }
    d4unlock_all (alarmLogDesc);
}

static int RecIsFiltered (desc, filter)
D4DATA *desc;
Filter *filter;
{
    PvData *pvPtr;
    TreeNode	*node;
    PointName pointName;
    char	*sp;
    char	*dp;

    /* test time filters */
    if (strlen(filter->start))
    {
	if (atol (f4str (alarmsecsFld)) < GetGMT_Secs (filter->start))
	    return FILTOUTOFRANGELOW;
    }
    if (strlen(filter->end))
    {
 	if (atol (f4str (alarmsecsFld)) >= GetGMT_Secs (filter->end))
	    return FILTOUTOFRANGEHIGH;
    }

    /* if alarms only, and not in alarm, return true */
    if ((filter->alarmOnly == ALMONLYFILTER)
      && ((*f4str(severityFld) == ' ') || (*f4ptr(severityFld) == 0) || (*f4ptr(severityFld) == 'E')))
	return FILTOUT;

    /* if Acked and acked alarms aren't allowed, return true */
    if ((filter->ack & ACKFILTER) && (*f4ptr(ackedFld) == 'Y'))
	return FILTOUT;

    /* if in alarm (unacked) and unacked alarms aren't allowed, return true */
    if ((filter->ack & UNACKFILTER) && (*f4ptr(ackedFld) == 'N')
      && !((*f4str(severityFld) == ' ') || (*f4ptr(severityFld) == 'E')))
	return FILTOUT;

    /* test severity filters */
    if (((filter->sev & MAJORFILTER) && (*f4ptr(severityFld) == 'M'))
      || ((filter->sev & MINORFILTER) && (*f4ptr(severityFld) == 'm'))
      || ((filter->sev & INFOFILTER) && (*f4ptr(severityFld) == 'I'))
      || ((filter->sev & EVENTFILTER) && (*f4ptr(severityFld) == 'E'))
      || ((filter->sev & NOALARMFILTER) &&
		((*f4ptr(severityFld) == ' ') || (*f4ptr(severityFld) == 0))))
	return FILTOUT;

    dp = (char *)f4str(operatorFld);
    sp = filter->op;
    if (*sp && *dp)
    {
	while (*sp && (*sp == *dp))
	{
	    sp++;
	    dp++;
	}
	if (*sp || (*dp != ' '))
	    return FILTOUT;
    }

    /* Finally, see if the point is descended from our filter group spec */
    strncpy (pointName, f4str (pointnameFld), sizeof (pointName));
    pointName [sizeof (pointName)-1] = 0;
    trimblanks (pointName);
    /* if no group is specified, include the point */
    trimblanks (filter->group);
    if (strlen (filter->group) == 0)
	return FILTOK;
	
    if (pvPtr = FindChannel (treeRoot, pointName))
    {
	for (node = pvPtr->parentGroup; node; node = node->parent)
	{
	    if (strcmp (node->groupName, filter->group) == 0)
		break;
	}
	if (node) /* we are a descendant */
	    return FILTOK;
	else /* no, return filtered status */
	    return FILTOUT;
    }
    else /* point not in alarm tree, ignore it */
	return FILTOUT;
}

#ifndef SLOW_BUT_SURE
static int sortCompare (const void *p1, const void *p2)
{
    return strcmp ((*(PvData **)p1)->pvName, (*(PvData **)p2)->pvName);
}

static int findCompare (const void *p1, const void *p2)
{
    return strcmp ((char *)p1, (*(PvData **)p2)->pvName);
}
#endif SLOW_BUT_SURE

PvData * FindChannel (groupLinkPtr, chanName)
TreeNode	*groupLinkPtr;
char 	*chanName;
{
#ifdef SLOW_BUT_SURE
    PvData	*chanPtr;
    TreeNode	*node;

/* This code searches all groups to find the PV */
    if (groupLinkPtr == NULL)
	groupLinkPtr = treeRoot;
    chanPtr = groupLinkPtr->pvList;
    while (chanPtr)
    {
	if (strcmp(chanPtr->pvName, chanName) == 0)
	{
	    return (chanPtr);
	}
	chanPtr = chanPtr->nextPv;
    }

    for (node = groupLinkPtr->leftChild; node; node = node->rightSibling)
    {
	if (chanPtr = FindChannel(node, chanName))
	{
	    return (chanPtr);
	}
    }
    return NULL;

#else
/* This code will binary search a table of pointers to the points */
    PvData	**p;
    long	i;

    /* if the shape of things has changed, re-sort them */
    if (numSortedPoints != shareHeader->alarmPointCount)
    {
	if (pointList)
	    free (pointList);
	numSortedPoints = shareHeader->alarmPointCount;
	pointList = p = (PvData **) malloc (numSortedPoints * sizeof (PvData *));
	if (pointList)
	{
	    for (i = 0; i < numSortedPoints; i++)
		pointList [i] = &shareHeader->alarmPoints[i];
	}
	qsort ((char *)pointList, (int)numSortedPoints, sizeof (PvData *), sortCompare);
    }

    /* Perform binary search of the sorted point list. */
    p = (PvData **)bsearch (chanName, (char *)pointList,
			(int)shareHeader->alarmPointCount, sizeof (PvData *), findCompare);

    /* If we found it then return a pointer to it. */
    if (p)
	if ((*p)->parentGroup != NULL)
	    return *p;

    return NULL;
#endif SLOW_BUT_SURE
}

TreeNode * FindGroup (groupLinkPtr, groupName)
TreeNode *groupLinkPtr;
char 	*groupName;
{
    TreeNode	*retPtr;
    TreeNode	*node;

    if (groupLinkPtr == NULL)
	groupLinkPtr = treeRoot;
    if (strcmp (groupLinkPtr->groupName, groupName) == 0)
	return groupLinkPtr;

    for (node = groupLinkPtr->leftChild; node; node = node->rightSibling)
    {
	if (retPtr = FindGroup (node, groupName))
	{
	    return (retPtr);
	}
    }
    return NULL;
}

int Log_PrintAlarmLog ( printerFileDesc )
    FILE *printerFileDesc;
{
	if (d4top (alarmLogDesc) == 0)
	{
	    do
	    {
		if (strncmp ("E", f4str (d4field (alarmLogDesc,"SEVERITY")), AL_SEVERITY_SIZE)
		    != 0)
		{
		    Log_PrintRecord (printerFileDesc);
		}
	    }
	    while (d4skip (alarmLogDesc, 1) == 0);
	}
	d4unlock_all (alarmLogDesc);
	return 0;
}

int Log_PrintEventLog ( printerFileDesc )
    FILE *printerFileDesc;
{
	if (d4top (alarmLogDesc) == 0)
	{
	    do
	    {
		if (strncmp ("E", f4str (d4field (alarmLogDesc,"SEVERITY")), AL_SEVERITY_SIZE)
		    == 0)
		{
		    Log_PrintRecord (printerFileDesc);
		}
	    }
	    while (d4skip (alarmLogDesc, 1) == 0);
	}
	d4unlock_all (alarmLogDesc);
	return 0;
}
 
int Log_PrintRecord ( printerFileDesc )
    FILE *printerFileDesc;
{
	fprintf (printerFileDesc, "%s  ", f4str (d4field (alarmLogDesc, "DATE")));
	fprintf (printerFileDesc, "%s  ", f4str (d4field (alarmLogDesc, "TIME")));
	fprintf (printerFileDesc, "%s  ", f4str (pointnameFld));
	fprintf (printerFileDesc, "%s  ", f4str (descriptorFld));
	fprintf (printerFileDesc, "%s  ", f4str (d4field (alarmLogDesc, "TYPE")));
	fprintf (printerFileDesc, "%s  ", f4str (d4field (alarmLogDesc, "VALACK")));
	fprintf (printerFileDesc, "%s  ", f4str (d4field (alarmLogDesc, "TAGOP")));
	fprintf (printerFileDesc, "\n");
	return 0;
}

static int noAlarmInfo = False;
static char displayvalue [AL_SCRIPTNAME_SIZE];
static char guidancevalue [AL_GUIDANCE_SIZE];

void alarminfoInit ( )
{
    char buff[100];

#ifdef WIN32
    strcpy (buff, "c:\\tis4000db\\alarminf.dbf");
#else
    strcpy (buff, "/tis4000db/alarminf.dbf");
#endif WIN32

    if ( !(alarminfDesc = d4open (&codeBaseDesc, buff)))
    {
	noAlarmInfo = True;
    }
}

#if defined(SOLARIS) || defined(WIN32) || defined(AXP) || defined(GNUCOMPILE)
void GetAlarmInfo (char *Name, char **Guidance, char **Display)
#else
void GetAlarmInfo (Name, Guidance, Display)
char *Name;
char **Guidance;
char **Display;
#endif
{
    char *seekpt;
    int d4stat;

    *Display = 0;
    *Guidance = 0;

    if (!alarminfDesc)
	alarminfoInit ();

    if (noAlarmInfo)
    {
	return;
    }

    seekpt = malloc (AL_POINTNAME_SIZE+1); 
    d4tag_select (alarminfDesc, d4tag (alarminfDesc, "NAME"));
    strncpy (seekpt, Name, AL_POINTNAME_SIZE+1);
    padblanks (seekpt, AL_POINTNAME_SIZE);
    d4stat = d4seek (alarminfDesc, seekpt);

    if (d4stat == 0)
    {
	field_ptr = d4field (alarminfDesc, "GUIDANCE");
 	f4ncpy (field_ptr, guidancevalue, AL_GUIDANCE_SIZE);
	trimblanks (guidancevalue);
	field_ptr = d4field (alarminfDesc, "DISPLAY");
 	f4ncpy (field_ptr, displayvalue, AL_SCRIPTNAME_SIZE);
	trimblanks (displayvalue);
	*Display = displayvalue;
	*Guidance = guidancevalue;
    }
    free (seekpt);
    d4unlock_all (alarminfDesc);
}


D4DATA *opTrackDesc;
int	trkNumRecords;

int TrkOpenFile ()
{
    char buff[100];

#ifdef WIN32
    strcpy (buff, "c:\\tis4000db\\op_trak.dbf");
#else
    strcpy (buff, "/tis4000db/op_trak.dbf");
#endif WIN32

    if ( !(opTrackDesc = d4open (&codeBaseDesc, buff)))
    {
	return OPEN_ERR;
    }
    return NO_ERR;
}

void TrkInit ()
{
    trkRecPosition = 0;
    trkTagPosition = 1;
}

int TrkSeek (recordNum)
unsigned int recordNum;
{
    int		filtstat;
    int		inc;
    char	*seekpt;
    int		d4stat = True;
    static char trkdatestr [DATE_SIZE+1];
    static char trktimestr [TIME_SIZE+1];
    static char trknamestr [POINTNAME_SIZE+1];

    /* First check for any shifting in the file */
    if ((trkRecPosition == 0)
      || (d4go (opTrackDesc, trkRecPosition) != 0)
      || (strncmp (trkdatestr, f4str (d4field (opTrackDesc, "TRK_DATE")), sizeof(trkdatestr)) != 0)
      || (strncmp (trktimestr, f4str (d4field (opTrackDesc, "TRK_TIME")), sizeof(trktimestr)) != 0)
      || (strncmp (trknamestr, f4str (d4field (opTrackDesc, "POINTNAME")), sizeof(trknamestr)) != 0))
    {
	/* the record set changed, start at beginning */
	trkTagPosition = 1;
	d4tag_select (opTrackDesc, d4tag (opTrackDesc, "UNIXTIME"));
	seekpt = malloc (10+8+1); 
	strcpy (seekpt, " !        ");
	d4stat = (d4seek (opTrackDesc, seekpt) == r4after);
	free (seekpt);
    }

    if (trkTagPosition > recordNum)
    {
	inc = -1;
    }
    else if (trkTagPosition <= recordNum)
    {
	inc = 1;
    }

    while (d4stat && !d4deleted(opTrackDesc))
    {
	filtstat = TrkRecIsFiltered (opTrackDesc, &trkFilter);
	if (filtstat == FILTOK)
	{
	    if (recordNum == trkTagPosition)
	    {
		/* we have it, return */
		trkRecPosition = d4recno (opTrackDesc);
		strncpy (trkdatestr, f4str (d4field (opTrackDesc, "TRK_DATE")), sizeof(trkdatestr));
		strncpy (trktimestr, f4str (d4field (opTrackDesc, "TRK_TIME")), sizeof(trktimestr));
		strncpy (trknamestr, f4str (d4field (opTrackDesc, "POINTNAME")), sizeof(trknamestr));
		d4unlock_all (opTrackDesc);
		return RECFOUND;
	    }
	    else /* adjust our position in the file */
	    {
		trkTagPosition = trkTagPosition + inc;
	    }
	}
	else if ((inc == -1 && (filtstat == FILTOUTOFRANGELOW))
	  || (inc == 1 && (filtstat == FILTOUTOFRANGEHIGH)))
	{
	    break;
	}
	d4stat = (d4skip (opTrackDesc, inc) == 0);
    }
    trkRecPosition = d4recno (opTrackDesc);
    d4unlock_all (opTrackDesc);
    return NORECORD;
}

void TrkSetFilter (filter)
Filter *filter;
{
    if (memcmp (&trkFilter, filter, sizeof(Filter)))
    {
	trkFilter = *filter;
	trkCounts = 0; /* force a recount */
	TrkAdjustCounts ();
    }
}

static void TrkAdjustCounts ()
{
    int filtstat;
    char *seekpt;
    int d4stat;

    if (trkCounts != 0 && trkNumRecords == d4reccount(opTrackDesc))
	return;

    trkNumRecords = d4reccount(opTrackDesc);
    trkRecPosition = 0;

    seekpt = malloc (10+8+1);
    d4tag_select (opTrackDesc, d4tag (opTrackDesc, "UNIXTIME"));
    strcpy (seekpt, " !        ");
    d4stat = d4seek (opTrackDesc, seekpt);
    free (seekpt);

    trkCounts = 0;
    if (d4stat == r4after)
    {
	do
	{
	    if (!d4deleted(opTrackDesc))
	    {
		filtstat = TrkRecIsFiltered (opTrackDesc, &trkFilter);
		if (filtstat == FILTOK)
		{
		    trkCounts ++;
		}
	    }
	    else
	    {
		filtstat = FILTOUT;
	    }
	}
	while ((filtstat != FILTOUTOFRANGEHIGH) && (d4skip (opTrackDesc, 1) == 0));
    }
    d4unlock_all (opTrackDesc);
}

static int TrkRecIsFiltered (desc, filter)
D4DATA *desc;
Filter *filter;
{
    PvData *pvPtr;
    TreeNode *node;
    Operator opName;
    PointName pointName;
    long trktime;

    /* test time filters */
    if (strlen(filter->start) || strlen(filter->end))
    {
	trktime = TrkEventTime ();

	if (strlen(filter->start))
	{
	    if (trktime < GetGMT_Secs (filter->start))
		return FILTOUTOFRANGELOW;
	}
	if (strlen(filter->end))
	{
 	    if (trktime >= GetGMT_Secs (filter->end))
		return FILTOUTOFRANGEHIGH;
	}
    }

    /* if Ack and acks aren't allowed, return true */
    if ((filter->ack & ACKFILTER) && (strstr (f4str (d4field (opTrackDesc, "FIELDNAME")), "ACK")))
	return FILTOUT;

    /* if not an ack and unacked alarms aren't allowed, return true */
    if ((filter->ack & UNACKFILTER) && (strcmp (f4str (d4field (opTrackDesc, "FIELDNAME")), "ACK") == 0))
	return FILTOUT;

    if (strlen(filter->op) && strlen(f4str (d4field (opTrackDesc, "OPERATOR"))))
    {
	strncpy (opName, f4str (d4field (opTrackDesc, "OPERATOR")), sizeof (opName));
	opName [sizeof (opName)-1] = 0;
	trimblanks (opName);
	if (strcmp (filter->op, opName) != 0)
	    return FILTOUT;
    }

    /* Finally, see if the point is descended from our filter group spec */
    strncpy (pointName, f4str (d4field (opTrackDesc, "POINTNAME")), sizeof (pointName));
    pointName [sizeof (pointName)-1] = 0;
    trimblanks (pointName);
    /* if no pointname in track rec, it is displayable */
    if (pointName[0] == 0)
	return FILTOK;

    /* if no group is specified, include the point */
    trimblanks (filter->group);
    if (strlen (filter->group) == 0)
	return FILTOK;
    if (pvPtr = FindChannel (treeRoot, pointName))
    {
	for (node = pvPtr->parentGroup; node; node = node->parent)
	{
	    if (strcmp (node->groupName, filter->group) == 0)
		break;
	}
	if (node) /* we are a descendant */
	    return FILTOK;
	else /* no, return filtered status */
	    return FILTOUT;
    }
    else /* point not in alarm tree, ignore it */
	return FILTOUT;
}

unsigned int TrkNumEntries ()
{
    TrkAdjustCounts ();
    return trkCounts;
}

long TrkEventTime ()
{
    return f4long (d4field (opTrackDesc, "UNIX_TIME" ));
}

char *TrkDate ()
{
    static char datestr [DATE_SIZE+1];

    strncpy (datestr, f4str (d4field (opTrackDesc, "TRK_DATE")), sizeof(datestr));
    return datestr;
}

char *TrkTime ()
{
    static char timestr [TIME_SIZE+1];

    strncpy (timestr, f4str (d4field (opTrackDesc, "TRK_TIME")), sizeof(timestr));
    return timestr;
}

char *TrkPointname ()
{
    static char namestr [POINTNAME_SIZE+1];

    strncpy (namestr, f4str (d4field (opTrackDesc, "POINTNAME")), sizeof(namestr));
    return namestr;
}

char *TrkDescriptor ()
{
    static char descstr [DESCRIPTOR_SIZE+1];

    strncpy (descstr, f4str (d4field (opTrackDesc, "DESCRIPTOR")), sizeof(descstr));
    return descstr;
}

char *TrkFieldname ()
{
    static char fieldstr [FIELDNAME_SIZE+1];

    strncpy (fieldstr, f4str (d4field (opTrackDesc, "FIELDNAME")), sizeof(fieldstr));
    return fieldstr;
}

char *TrkOldvalue ()
{
    static char valuestr [VALACK_SIZE+1];

    strncpy (valuestr, f4str (d4field (opTrackDesc, "OLD_VALUE")), sizeof(valuestr));
    return valuestr;
}

char *TrkNewvalue ()
{
    static char valuestr [VALACK_SIZE+1];

    strncpy (valuestr, f4str (d4field (opTrackDesc, "NEW_VALUE")), sizeof(valuestr));
    return valuestr;
}

char *TrkOperator ()
{
    static char operstr [AL_OPERATOR_SIZE+1];

    strncpy (operstr, f4str (d4field (opTrackDesc, "OPERATOR")), sizeof(operstr));
    return operstr;
}

void FiltInit ( )
{
#ifdef WIN32
    strcpy (filterFileName, "c:\\tis4000db\\AlarmvuF.dbf");
#else
    strcpy (filterFileName, "/tis4000db/AlarmvuF.dbf");
#endif WIN32

    if ( !(filtersDesc = d4open (&codeBaseDesc, filterFileName)))
    {
	if ( !(filtersDesc = d4create (&codeBaseDesc, filterFileName, filterFields, filterTags)))
	{
	    printf ("~Can't open/create filter file\n");
	    exit (0);
	}
    }
}

void GetFilter (name, filt)
char *name;
Filter *filt;
{
    char *seekpt;
    int d4stat;

    if (!filtersDesc)
	FiltInit ();
    seekpt = malloc (AL_FILTERNAME_SIZE+1); 
    d4tag_select (filtersDesc, d4tag (filtersDesc, "FILTERTAG"));
    strncpy (seekpt, name, AL_FILTERNAME_SIZE+1);
    padblanks (seekpt, AL_FILTERNAME_SIZE);
    d4stat = d4seek (filtersDesc, seekpt);

    if (d4stat == 0)
    {
	field_ptr = d4field (filtersDesc, "GROUP");
 	f4ncpy (field_ptr, filt->group, AL_GROUPNAME_SIZE);
	trimblanks (filt->group);
	field_ptr = d4field (filtersDesc, "SEVRESTR");
	filt->sev = f4int (field_ptr);
	field_ptr = d4field (filtersDesc, "ACKRESTR");
	filt->ack = f4int (field_ptr);
	field_ptr = d4field (filtersDesc, "GROUPRESTR");
	filt->alarmOnly = f4int (field_ptr);
	field_ptr = d4field (filtersDesc, "STARTTIME");
 	f4ncpy (field_ptr, filt->start, AL_DATE_SIZE+1+AL_TIME_SIZE+1);
	trimblanks (filt->start);
	field_ptr = d4field (filtersDesc, "ENDTIME");
 	f4ncpy (field_ptr, filt->end, AL_DATE_SIZE+1+AL_TIME_SIZE+1);
	trimblanks (filt->end);
	field_ptr = d4field (filtersDesc, "OPERATOR");
 	f4ncpy (field_ptr, filt->op, AL_OPERATOR_SIZE);
	trimblanks (filt->op);
	field_ptr = d4field (filtersDesc, "STARTUPDSP");
	filt->startupDisplay = f4int (field_ptr);
	strncpy (filt->name, name, sizeof(filt->name));
    }
    else
    {
	strcpy (filt->group, "");
	filt->ack = 0;
	filt->sev = 0;
	filt->alarmOnly = ALLFILTER;
	strcpy (filt->start, "");
	strcpy (filt->end, "");
	strcpy (filt->op, "");
	filt->startupDisplay = 0;
	strncpy (filt->name, "Default", sizeof(filt->name));
    }
    free (seekpt);
    d4unlock_all (filtersDesc);
    filt->modified = FALSE;
}

void SaveFilter (name, filt)
char *name;
Filter *filt;
{
    char *seekpt;
    int d4stat;

    if (!filtersDesc)
	FiltInit ();
    seekpt = malloc (AL_FILTERNAME_SIZE+1);
    d4tag_select (filtersDesc, d4tag (filtersDesc, "FILTERTAG"));
    strncpy (seekpt, name, AL_FILTERNAME_SIZE+1);
    padblanks (seekpt, AL_FILTERNAME_SIZE);
    d4stat = d4seek (filtersDesc, seekpt);

    if (d4stat != 0)
    {
	d4append_blank (filtersDesc);
    }
    field_ptr = d4field (filtersDesc, "FILTERNAME");
    f4assign_n (field_ptr, seekpt, AL_FILTERNAME_SIZE);
    field_ptr = d4field (filtersDesc, "GROUP");
    f4assign_n (field_ptr, filt->group, AL_GROUPNAME_SIZE);
    field_ptr = d4field (filtersDesc, "SEVRESTR");
    f4assign_int (field_ptr, filt->sev);
    field_ptr = d4field (filtersDesc, "ACKRESTR");
    f4assign_int (field_ptr, filt->ack);
    field_ptr = d4field (filtersDesc, "GROUPRESTR");
    f4assign_int (field_ptr, filt->alarmOnly);
    field_ptr = d4field (filtersDesc, "STARTTIME");
    f4assign_n (field_ptr, filt->start, AL_DATE_SIZE+1+AL_TIME_SIZE);
    field_ptr = d4field (filtersDesc, "ENDTIME");
    f4assign_n (field_ptr, filt->end, AL_DATE_SIZE+1+AL_TIME_SIZE);
    field_ptr = d4field (filtersDesc, "OPERATOR");
    f4assign_n (field_ptr, filt->op, AL_OPERATOR_SIZE);
    field_ptr = d4field (filtersDesc, "STARTUPDSP");
    f4assign_int (field_ptr, filt->startupDisplay);

    d4flush_all (filtersDesc);
#ifndef CB51
    d4update_header (filtersDesc, TRUE, TRUE);
#endif
    d4unlock_all (filtersDesc);
    free (seekpt);
    filt->modified = FALSE;
}

char **filternames;

char **GetFilterNames ()
{
    int i;
    int j;

    if (!filtersDesc)
	FiltInit ();
    d4go (filtersDesc, 1);
    j = 0;
    if (filternames)
	free (filternames);

    filternames = (char **)calloc (d4reccount (filtersDesc) + 1, sizeof(char *));
    if (filternames == 0)
	return filternames;

    for ( i = 1; i <= d4reccount (filtersDesc); i++)
    {
        d4go (filtersDesc, i);
        if (!d4deleted (filtersDesc))
        {
	    if ((filternames[j] = calloc (1, AL_FILTERNAME_SIZE)) != 0)
	    {
		field_ptr = d4field (filtersDesc, "FILTERNAME");
		f4ncpy (field_ptr, filternames[j], AL_FILTERNAME_SIZE);
		trimblanks (filternames[j]);
		j++;
	    }
	}
    }
    d4unlock_all (filtersDesc);
    return filternames;
}

void DeleteFilter (name)
char *name;
{
    char *seekpt;
    int d4stat;

    if (!filtersDesc)
	FiltInit ();
    seekpt = malloc (AL_FILTERNAME_SIZE+1); 
    d4tag_select (filtersDesc, d4tag (filtersDesc, "FILTERTAG"));
    strncpy (seekpt, name, AL_FILTERNAME_SIZE+1);
    padblanks (seekpt, AL_FILTERNAME_SIZE);
    d4stat = d4seek (filtersDesc, seekpt);

    if (d4stat == 0)
    {
	d4delete (filtersDesc);
	d4pack (filtersDesc);
	d4flush_all (filtersDesc);
    }
    d4unlock_all (filtersDesc);
    free (seekpt);
}

void ReportInit ( )
{
#ifdef WIN32
    strcpy (reportFileName, "c:\\tis4000db\\AlarmvuR.dbf");
#else
    strcpy (reportFileName, "/tis4000db/AlarmvuR.dbf");
#endif WIN32

    if ( !(reportsDesc = d4open (&codeBaseDesc, reportFileName)))
    {
	if ( !(reportsDesc = d4create (&codeBaseDesc, reportFileName, reportFields, reportTags)))
	{
	    printf ("~Can't open/create report file\n");
	    exit (0);
	}
    }
}

void GetReport (name, criteria)
char *name;
RepCriteria *criteria;
{
    char *seekpt;
    int d4stat;

    if (!reportsDesc)
	ReportInit ();
    seekpt = malloc (REPORTNAME_SIZE+1); 
    d4tag_select (reportsDesc, d4tag (reportsDesc, "REPORTTAG"));
    strncpy (seekpt, name, REPORTNAME_SIZE+1);
    padblanks (seekpt, REPORTNAME_SIZE);
    d4stat = d4seek (reportsDesc, seekpt);

    if (d4stat == 0)
    {
	field_ptr = d4field (reportsDesc, "RPTITEMS");
 	criteria->reportitems = f4long (field_ptr);
	strncpy (criteria->name, name, sizeof(criteria->name));
    }
    else
    {
	strcpy (criteria->name, "");
	criteria->reportitems = 0;
    }
    free (seekpt);
    d4unlock_all (reportsDesc);
}

void SaveReport (name, criteria)
char *name;
RepCriteria *criteria;
{
    char *seekpt;
    int d4stat;

    if (!reportsDesc)
	ReportInit ();
    seekpt = malloc (REPORTNAME_SIZE+1); 
    d4tag_select (reportsDesc, d4tag (reportsDesc, "REPORTTAG"));
    strncpy (seekpt, name, REPORTNAME_SIZE+1);
    padblanks (seekpt, REPORTNAME_SIZE);
    d4stat = d4seek (reportsDesc, seekpt);

    if (d4stat != 0)
    {
	d4append_blank (reportsDesc);
    }
    field_ptr = d4field (reportsDesc, "REPORTNAME");
    f4assign_n (field_ptr, seekpt, REPORTNAME_SIZE);
    field_ptr = d4field (reportsDesc, "RPTITEMS");
    f4assign_int (field_ptr, criteria->reportitems);

    d4flush_all (reportsDesc);
#ifndef CB51
    d4update_header (reportsDesc, TRUE, TRUE);
#endif
    d4unlock_all (reportsDesc);
    free (seekpt);
}

char **reportnames;

char **GetReportNames ()
{
    int i;
    int j;

    if (!reportsDesc)
	ReportInit ();
    d4go (reportsDesc, 1);
    j = 0;
    if (reportnames)
	free (reportnames);

    reportnames = (char **)calloc (d4reccount (reportsDesc) + 1, sizeof(char *));
    if (reportnames == 0)
	return reportnames;

    for ( i = 1; i <= d4reccount (reportsDesc); i++)
    {
        d4go (reportsDesc, i);
        if (!d4deleted (reportsDesc))
        {
	    if ((reportnames[j] = calloc (1, REPORTNAME_SIZE)) != 0)
	    {
		field_ptr = d4field (reportsDesc, "REPORTNAME");
		f4ncpy (field_ptr, reportnames[j], REPORTNAME_SIZE);
		trimblanks (reportnames[j]);
		j++;
	    }
	}
    }
    d4unlock_all (reportsDesc);
    return reportnames;
}

void DeleteReport (name)
char *name;
{
    char *seekpt;
    int d4stat;

    if (!reportsDesc)
	ReportInit ();
    seekpt = malloc (REPORTNAME_SIZE+1); 
    d4tag_select (reportsDesc, d4tag (reportsDesc, "REPORTTAG"));
    strncpy (seekpt, name, REPORTNAME_SIZE+1);
    padblanks (seekpt, REPORTNAME_SIZE);
    d4stat = d4seek (reportsDesc, seekpt);

    if (d4stat == 0)
    {
	d4delete (reportsDesc);
	d4pack (reportsDesc);
	d4flush_all (reportsDesc);
    }
    d4unlock_all (reportsDesc);
    free (seekpt);
}

static long GetGMT_Secs (timestr)
char *timestr;
{
    char *strptr;
    struct tm mytm;

    memset (&mytm, 0, sizeof(mytm));

    for (strptr = timestr; *strptr == ' '; strptr++)
	/* all the work is in the for statement */ ;
    mytm.tm_mon = atoi (strptr) - 1;
    for (; *strptr && *strptr != '/'; strptr++)
	/* all the work is in the for statement */ ;
    if (*strptr == 0) /* check format */
	return 0;
    strptr++;
    mytm.tm_mday = atoi (strptr);
    for (; *strptr && *strptr != '/'; strptr++)
	/* all the work is in the for statement */ ;
    if (*strptr == 0) /* check format */
	return 0;
    strptr++;
    mytm.tm_year = atoi (strptr);
    /* make this work until 2090! */
    if (mytm.tm_year < 90)
	mytm.tm_year += 100;

    /* move through string to time section! */
    for (; *strptr != ' '; strptr++)
	/* all the work is in the for statement */ ;
    for (; *strptr == ' '; strptr++)
	/* all the work is in the for statement */ ;
    mytm.tm_hour = atoi (strptr);
    for (; *strptr && *strptr != ':'; strptr++)
	/* all the work is in the for statement */ ;
    strptr++;
    mytm.tm_min = atoi (strptr);
    for (; *strptr && *strptr != ':'; strptr++)
	/* all the work is in the for statement */ ;
    strptr++;
    mytm.tm_sec = atoi (strptr);
    mytm.tm_isdst = -1;

#if defined(MIPSEL) || defined(SOLARIS) || defined(WIN32) || defined(AXP)
    return mktime (&mytm);
#else
    return timelocal (&mytm);
#endif
}

char *subGroupNames [100];
char **subNamePtr;

void ScanGroups (groupLink)
TreeNode *groupLink;
{
    TreeNode	*groupPtr;

    *subNamePtr = groupLink->groupName;
    subNamePtr++;
    for (groupPtr = groupLink->leftChild; groupPtr; groupPtr = groupPtr->rightSibling)
    {
	ScanGroups (groupPtr);
    }
}

char **GetSubGroups (group)
char *group;
{
    TreeNode	*groupPtr = treeRoot;

    subNamePtr = subGroupNames;

    if (groupPtr = FindGroup (groupPtr, group))
    {
	ScanGroups (groupPtr);
    }
    *subNamePtr = NULL;
    return subGroupNames;
}

void trimblanks (string) 
char *string;
{
    int last;

	/* strip trailing blanks */
	for (last = strlen(string)-1; (last >= 0) && (string[last] == ' '); last--)
	   	string[last] = 0;
}

long UnixTime (long *timeVal)
{
#ifdef AXP
    time_t tempTime;
    long retTime;

    if (timeVal)
    {
	retTime = time(&tempTime);
        *timeVal = (long)tempTime;
    }
    else
	retTime = time(NULL);
    return retTime;
#else
    return time(timeVal);
#endif AXP
}

void padblanks (string, len) 
char *string;
int len;
{
    int last;

    /* add trailing blanks */
    for (last = strlen(string); (last < len - 1); last++)
	string[last] = ' ';
    string[last] = 0;
}

#ifdef USEX11
int writeableColors;
unsigned long *cells;
int	ncells;

int alloc_color_cells (w, n)
Widget w;
int    n;
{
    Display *dpy = XtDisplay(w);
    int     scr = DefaultScreen (dpy);
    Colormap cmap = DefaultColormap (dpy, scr);

    ncells = n;
    cells = (unsigned long *)malloc (n*sizeof(unsigned long));

    if (!XAllocColorCells (dpy, cmap, TRUE, NULL, 0, cells, n))
    {
        printf ("~Couldn't get the color cells.\n");
      /*  free (cells);
      **  cells = 0;
      */
        writeableColors = 0;
        return 1;
    }
    writeableColors = 1;
    return 1;
}

Pixel get_pixel_from_index ( n )
int    n;
{
    if (cells)
        return (cells [n]);
    else
        return 0;
}

int get_index_from_pixel ( p )
Pixel    p;
{
    int i;

    if (cells)
    {
        for (i = 0; i < ncells; i++)
	{
	    if (cells[i] == p)
		return i;
	}
    }
    return 0;
}

Pixel set_pixel_from_long (w, n, c)
Widget w;
int    n;
long   c;
{
    Display *dpy = XtDisplay(w);
    int     scr = DefaultScreen (dpy);
    Colormap cmap = DefaultColormap (dpy, scr);
    XColor  color;

    if (writeableColors)
    {
        color.pixel = cells [n];
        color.red = Red (c) * 256;
        color.green = Green (c) * 256;
        color.blue = Blue (c) * 256;
        color.flags = DoRed|DoGreen|DoBlue;

        XStoreColor (dpy, cmap, &color);
        return (color.pixel);
    }
    else
    {
        cells [n] = get_pixel_from_long (w, c);
        return (cells [n]);
    }
}

Pixel get_pixel_from_long (w, c)
Widget w;
long   c;
{
    Display *dpy = XtDisplay(w);
    int     scr = DefaultScreen (dpy);
    Colormap cmap = DefaultColormap (dpy, scr);
    XColor  color;

    color.red = Red (c) * 256;
    color.green = Green (c) * 256;
    color.blue = Blue (c) * 256;

    if (XAllocColor (dpy, cmap, &color))
        return (color.pixel);
    else
        return (BlackPixel(dpy, scr));
}

long  get_color_from_pixel (w, p)
Widget	w;
Pixel	p;
{
    Display *dpy = XtDisplay(w);
    int     scr = DefaultScreen (dpy);
    Colormap cmap = DefaultColormap (dpy, scr);
    XColor  color;

    color.pixel = p;
    XQueryColor (dpy, cmap, &color);
    return (Color (((unsigned int)color.red)/256,
		   ((unsigned int)color.green)/256,
		   ((unsigned int)color.blue)/256));
}

void    ForceUpdate ( w )
Widget  w;
{
    Widget  diashell, topshell;
    Window  diawindow, topwindow;
    XtAppContext cxt = XtWidgetToApplicationContext (w);
    Display	*dpy;
    XWindowAttributes   xwa;
    XEvent	event;

    for (diashell = w; !XtIsShell(diashell); diashell = XtParent(diashell))
        /* all the work happened in th for statement */ ;
    for (topshell = diashell; !XtIsTopLevelShell(topshell); topshell = XtParent(topshell))
        /* all the work happened in th for statement */ ;
    if (XtIsRealized(diashell) && XtIsRealized(topshell))
    {
        dpy = XtDisplay(topshell);
        diawindow = XtWindow(diashell);
        topwindow  = XtWindow(topshell);
        while (XGetWindowAttributes (dpy,diawindow,&xwa) && xwa.map_state != IsViewable)
        {
            if (XGetWindowAttributes (dpy,topwindow,&xwa) && xwa.map_state != IsViewable)
                break;
            XtAppNextEvent (cxt,&event);
            XtDispatchEvent (&event);
        }
    }
    XmUpdateDisplay (topshell);
}

#endif USEX11
