#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vtextHEADER
#include vtextviewHEADER
#include vtextitemHEADER
#include vstdargHEADER
#include vprintHEADER
#include vbuttonHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "shareMem.h"
#include "tsDefs.h"
#include "OperPriv.h"
#include "alarmDefs.h"
#include "alarmfil.h"
#include "alarmvu.h"
}

#include "alvu.h"
#include "alarmvu.hxx"
#include "alvuUtils.hxx"

extern int	alarmBlinkLevel;
extern long	backgroundColor;
extern long	SevColor [NUM_SEV_COLORS];
extern long	UnackSevColor [NUM_SEV_COLORS];


/*******************************************************************
*
*    alvuMessageDialogCenter ()
*	passed a message number as defined in vuUtils.hxx
*
*******************************************************************/

int alvuMessageDialogCenter (vdialog *dialog, int msgType, const vname *msgName, ...)
{
    int		result = vTRUE;
    va_list	argptr;

    va_start (argptr, msgName);

    vapplication *alvuApp   = vapplication::GetCurrent ();
    vresource    alvuRes    = alvuApp -> GetResources ();
    vresource    titleRes   = vresourceGet (alvuRes, alvu_TitlesDictionary);
    vresource    msgRes     = vresourceGet (alvuRes, alvu_MessagesDictionary);
    vstr         *infoTitle = vresourceGetString (titleRes, alvu_InfoNotice);
    vstr         *cautionTitle = vresourceGetString (titleRes, alvu_CautionNotice);

    vstr *format = vresourceGetString (msgRes, msgName);
    // NOTE: vstrScribe cleans up vstr *format
    vstr *msg = vstrCloneScribed (vcharScribeFormatScribedVarargs (vstrScribe (format), argptr));

    va_end (argptr);

    if (msgType == ALVU_INFO_NOTICE)
    {
	infoNotice (dialog, infoTitle, msg);
    }
    else
    {
	result = cautionNotice (dialog, cautionTitle, msg);
    }
    vstrDestroy (infoTitle);
    vstrDestroy (cautionTitle);
    vstrDestroy (msg);

    return (result);
}


/*******************************************************************
*
*    ClosePrivTrack ()
*
*******************************************************************/

void ClosePrivTrack (int initedOperPriv, int initedOperTrack)
{
    if (initedOperPriv == vTRUE)
    {
	CloseOperPriv ();
    }
    if (initedOperTrack == vTRUE)
    {
	char operName [81];
	CloseOperTracking (operName);
    }
}


/*******************************************************************
*
*    GetRGBColor ()
*	Red, Green, Blue are defined in alarmfil.h as Macros
*
*******************************************************************/

void GetRGBColor (long color, double *red, double *green, double *blue)
{
    *red = (Red (color)) / 255.0;
    *green = (Green (color)) / 255.0;
    *blue = (Blue (color)) / 255.0;
}


/*******************************************************************
*
*    UpdateBlinkColors ()
*	SevColor and UnackSevColor defined in alarmvu.cxx
*
*******************************************************************/

void UpdateBlinkColors (void)
{
    long red, green, blue;
    int level = ((alarmBlinkLevel == 0) ? ALVU_BLINK_LEVEL_DEFAULT : alarmBlinkLevel);
    int i;

    for (i = 0; i < NUM_SEV_COLORS; i++)
    {
	red   = Red (SevColor [i]);
	green = Green (SevColor [i]);
	blue  = Blue (SevColor [i]);

	(red < Red (backgroundColor)) ?
		(red += (Red (backgroundColor) - red) * level / 8) :
		(red -= (red - Red (backgroundColor)) * level / 8);
	(green < Green (backgroundColor)) ?
		(green += (Green (backgroundColor) - green) * level / 8) :
		(green -= (green - Green (backgroundColor)) * level / 8);
	(blue < Blue (backgroundColor)) ?
		(blue += (Blue (backgroundColor) - blue) * level / 8) :
		(blue -= (blue - Blue (backgroundColor)) * level / 8);
	UnackSevColor [i] = Color (red, green, blue);
    }
}


/*******************************************************************
*
*    ValidTimeSpec ()
*	valid is MM/DD/YY HH:MM:SS
*
*******************************************************************/

static char daytab [2][13] = {
	{0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
	{0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}};

int ValidTimeSpec (char *timeStr)
{
    char *ptr = timeStr;
    char a;
    char month[3];
    char day[3];
    char year[3];
    int leap;

    if ((strlen (timeStr)) != (strlen (ALVU_DATETIME_STRING)))
	return (ALVU_ERROR);

    // MONTH, 01 - 12
    month[0] = *ptr;
    ptr++;
    if (month[0] == '0') 
    {
	if (((*ptr) < '1') || ((*ptr) > '9'))
	    return (ALVU_ERROR);
    }
    else if (month[0] == '1') 
    {
	if (((*ptr) < '0') || ((*ptr) > '2'))
	    return (ALVU_ERROR);
    }
    else
	return (ALVU_ERROR);
    month[1] = *ptr;
    month[2] = 0;
    int monVal = atoi (month);
    ptr++;
    if ((*ptr) != '/')
	return (ALVU_ERROR);
    ptr++;

    // DAY, 01 - 31
    day[0] = *ptr;
    ptr++;
    if (day[0] == '0')
    {
	if (((*ptr) < '1') || ((*ptr) > '9'))
	    return (ALVU_ERROR);
    }
    else if ((day[0] == '1') || (day[0] == '2'))
    {
	if (((*ptr) < '0') || ((*ptr) > '9'))
	    return (ALVU_ERROR);
    }
    else if (day[0] == '3')
    {
	if (((*ptr) < '0') || ((*ptr) > '1'))
	    return (ALVU_ERROR);
    }
    else
	return (ALVU_ERROR);
    day[1] = *ptr;
    day[2] = 0;
    int dayVal = atoi (day);

    ptr++;
    if ((*ptr) != '/')
	return (ALVU_ERROR);
    ptr++;
    year[0] = *ptr;
    ptr++;
    year[1] = *ptr;
    year[2] = 0;
    int yearVal = atoi (year);
    if (yearVal < 80)
    {
	yearVal += 2000;
    }
    else
    {
	yearVal += 1900;
    }
    leap = yearVal%4 == 0 && yearVal%100 != 0 || yearVal%400 == 0;
    // check for proper month/day
    if (dayVal > daytab[leap][monVal])
	return (ALVU_ERROR);

    ptr++;
    // skip space between date and time
    ptr++;

    // HOUR, 00 - 23
    a = *ptr;
    ptr++;
    if ((a == '0') || (a == '1'))
    {
	if (((*ptr) < '0') || ((*ptr) > '9'))
	    return (ALVU_ERROR);
    }
    else if (a == '2')
    {
	if (((*ptr) < '0') || ((*ptr) > '3'))
	    return (ALVU_ERROR);
    }
    else
	return (ALVU_ERROR);

    ptr++;
    if ((*ptr) != ':')
	return (ALVU_ERROR);

    // MINUTE, 00 - 59
    ptr++;
    if (((*ptr) < '0') || ((*ptr) > '5'))
	return (ALVU_ERROR);
    ptr++;
    if (((*ptr) < '0') || ((*ptr) > '9'))
	return (ALVU_ERROR);

    ptr++;
    if ((*ptr) != ':')
	return (ALVU_ERROR);

    // SECOND, 00 - 59
    ptr++;
    if (((*ptr) < '0') || ((*ptr) > '5'))
	return (ALVU_ERROR);
    ptr++;
    if (((*ptr) < '0') || ((*ptr) > '9'))
	return (ALVU_ERROR);

    return (ALVU_SUCCESS);
}


