#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vbuttonHEADER
#include vtextHEADER
#include vtextviewHEADER
#include vprintHEADER

extern "C"
{
#ifdef WIN32
#include <stdio.h>
#include <winsock.h>
// this bit of sillyness is needed because windows #defines AppendMenu
// as #define AppendMenu AppendMenuA, which interfers with galaxy class
// vmenubar::AppendMenu
#undef AppendMenu
#endif WIN32
#include "shareMem.h"
#include "tsDefs.h"
#include "alarmDefs.h"
#include "alarmfil.h"
#include "alarmvu.h"
}

#include "alvu.h"
#include "alarmvu.hxx"
#include "commentDialog.hxx"
#include "utils.hxx"

static int commentFilter (filteredText *text, vtextSelection *sel,
			  const vchar **s, long *len);


/*******************************************************************
*
*    CommentDialog Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(CommentDialog, vdialog, "CommentDialog");

/*******************************************************************
*
*    CommentDialog constructor
*
*******************************************************************/

void CommentDialog::LoadInit (vresource res)
{
    title        = NULL;
    displayEntry = vtextitem::RequestSubLoad (this, alvu_DisplayEntry);
    commentView  = filteredTextitem::RequestSubLoad (this, alvu_CommentView);
    closeButton  = CommentCloseButton::RequestSubLoad (this, alvu_CloseButton);
    saveButton   = CommentSaveButton::RequestSubLoad (this, alvu_SaveButton);
    deleteButton = CommentDeleteButton::RequestSubLoad (this, alvu_DeleteButton);

    vdialog::LoadInit (res);

    // store ptr to dialog in close button
    closeButton  -> StoreCommentDialog (this);
    saveButton   -> StoreCommentDialog (this);
    deleteButton -> StoreCommentDialog (this);

    //set filter for commentView
    commentView -> SetItemFilterFunc (commentFilter);
}


/*******************************************************************
*
*    CommentDialog window observer
*
*******************************************************************/

void CommentDialog::ObserveWindow (vevent *event)
{
    if (event -> GetType () == veventWINDOW_STATE && event -> IsClose ())
    {
	Close ();
    }
}

/*******************************************************************
*
*    CommentDialog destructor
*
*******************************************************************/

void CommentDialog::Destroy ()
{
    if (IsOpen ())
    {
	Close ();
    }
    if (title != NULL)
    {
	vstrDestroy (title);
    }
    vdialog::Destroy ();
}


/*******************************************************************
*
*    CommentDialog open function
*
*******************************************************************/

void CommentDialog::OpenComment (alvuApplication *alvuapp, char *pvname,
				 char *alarminfo, char *comment, 
				 AlarmLogRecord *recptr)
{
    app = alvuapp;

    recPtr = recptr;

    SetDialogTitle (pvname);

    SetEntryAndComment (alarminfo, comment);

    vdialog::Open();
}



/*******************************************************************
*
*    CommentDialog set title function
*
*******************************************************************/

void CommentDialog::SetDialogTitle (char *pvname)
{
    vresource titlesRes = app -> titlesResource;
    vstr	*formattedMsg;
    vchar	pv [AL_POINTNAME_SIZE + 1];

    formattedMsg = vresourceGetString (titlesRes, alvu_GuidanceTitle);
    vcharCopyScribedBounded (vcharScribeSystem (pvname),
			     pv, AL_POINTNAME_SIZE);
    pv [AL_POINTNAME_SIZE] = 0;
    title = vstrCloneScribed (vcharScribeFormat (formattedMsg, pv));
    vstrDestroy (formattedMsg);

    SetTitle (title);
}

/*******************************************************************
*
*    CommentDialog set display and comment entry
*
*******************************************************************/

void CommentDialog::SetEntryAndComment (char *entry, char *comment)
{
    displayEntry -> SetTextScribed (vcharScribeSystem (entry));
    if (comment != NULL)
    {
	commentView -> SetTextScribed (vcharScribeSystem (comment));
    }
    else
    {
	commentView -> SetTextScribed (vcharScribeSystem (""));
    }
}


/*******************************************************************
*
*    CommentCloseButton Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(CommentCloseButton, vbutton, "CommentCloseButton");

void CommentCloseButton::ObserveDialogItem (vevent *event)
{
    commentDialog -> Close ();
}

/*******************************************************************
*
*    CommentOkButton Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(CommentSaveButton, vbutton, "CommentSaveButton");

void CommentSaveButton::ObserveDialogItem (vevent *event)
{
    AlarmLogRecord *recPtr = commentDialog -> GetAlarmLogRec ();
    vstr *comment = commentDialog -> commentView -> GetTextAsString ();
    char commentStr [AL_MAX_MEMO_SIZE + 1];
    vcharExportBounded (vcharScribeSystemExport (comment),
			commentStr,
			AL_MAX_MEMO_SIZE);
    commentStr [AL_MAX_MEMO_SIZE] = 0;
    Send_storeMemoText (commentStr, recPtr);
}


/*******************************************************************
*
*    CommentDeleteButton Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(CommentDeleteButton, vbutton, "CommentDeleteButton");

void CommentDeleteButton::ObserveDialogItem (vevent *event)
{
    AlarmLogRecord *recPtr = commentDialog -> GetAlarmLogRec ();
    Send_removeMemoText (recPtr);
}

/*******************************************************************
*
*    Filter function for the commentView field
*
*******************************************************************/

static int commentFilter (filteredText *text, vtextSelection *sel,
			  const vchar **s, long *len)
{
    return (NCharFilter (text, sel, s, len, AL_MAX_MEMO_SIZE));
}
