#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vbuttonHEADER
#include vmenubarHEADER
#include vlistviewHEADER
#include vcolorHEADER
#include vprintHEADER

extern "C"
{
#ifdef WIN32
#include <stdio.h>
#include <winsock.h>
// this bit of sillyness is needed because windows #defines AppendMenu
// as #define AppendMenu AppendMenuA, which interfers with galaxy class
// vmenubar::AppendMenu
#undef AppendMenu
#endif WIN32
#include "cadef.h"
#include "shareMem.h"
#include "tsDefs.h"
#include "alarm.h"
#include "alarmDefs.h"
#include "alarmfil.h"
#include "alarmvu.h"
#include "bell.h"
extern char	*alarmSeverityString[];
}

#include "alvu.h"
#include "alarmvu.hxx"
#include "aldlgListView.hxx"
#include "alarmDialogs.hxx"
#include "popupDialog.hxx"
#include "textDialog.hxx"
#include "alvuUtils.hxx"
#include "utils.hxx"

// some globals
extern int		    alarmSoundEnabled;
extern long		    CurrUnackSevColor [];
extern long		    SevColor [];
extern int		    alarmSummaryDisabled;

extern AlarmSummaryDialog   *alarmSummaryDialogPtr;
extern DynamicGroupDialog   *dynamicGroupDialogPtr;
extern PopupDialog	    *popAlarmPtr;
extern TextDialog	    *guidanceDialogPtr;


/*******************************************************************
*
*    PopupDialog Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(PopupDialog, vdialog, "PopupDialog");

/*******************************************************************
*
*    PopupDialog constructor
*
*******************************************************************/

void PopupDialog::LoadInit (vresource res)
{

    alarmGroupButton = AlarmGroupButton::RequestSubLoad (this,
					alvu_AlarmGroupButton);
    buttonContainer = ButtonContainer::RequestSubLoad (this,
					alvu_ButtonContainer);
    vdialog::LoadInit (res);

    alarmGroupButton -> StorePopupDialog (this);
    buttonContainer -> silenceButton  -> StorePopupDialog (this);
    buttonContainer -> displayButton  -> StorePopupDialog (this);
    buttonContainer -> scriptButton   -> StorePopupDialog (this);
    buttonContainer -> closeButton    -> StorePopupDialog (this);
    buttonContainer -> summaryButton  -> StorePopupDialog (this);
    buttonContainer -> guidanceButton -> StorePopupDialog (this);
}


/*******************************************************************
*
*    PopupDialog window observer
*
*******************************************************************/

void PopupDialog::ObserveWindow (vevent *event)
{
    if (event -> GetType () == veventWINDOW_STATE && event -> IsClose ())
    {
	Close ();
	DeleteLater ();
    }
}

/*******************************************************************
*
*    PopupDialog destructor
*
*******************************************************************/

void PopupDialog::Destroy ()
{
    if (IsOpen ())
    {
	Close ();
    }
    vdialog::Destroy ();
    popAlarmPtr = NULL;
}


/*******************************************************************
*
*    PopupDialog open function
*
*******************************************************************/

void PopupDialog::OpenPopup (alvuApplication *alvuapp, TreeNode *snode, Filter *origfilter)
{
    app = alvuapp;
    filter = *origfilter;
    groupPtr = snode;
    sprintf (tempString, "%s", snode -> groupName);

    // enable/disable some buttons
    if (alarmSoundEnabled == vFALSE)
    {
	buttonContainer -> silenceButton -> Disable ();
    }
    if (!HaveScript (groupPtr -> groupParms.displayName))
    {
	buttonContainer -> displayButton -> Disable ();
    }
    if (!HaveScript (groupPtr -> groupParms.manualScript) ||
	!HaveScript (groupPtr -> groupParms.auxManualScript))
    {
	buttonContainer -> scriptButton -> Disable ();
    }
    if (!HaveScript (groupPtr -> groupParms.guidance))
    {
	buttonContainer -> guidanceButton -> Disable ();
    }
    // if no privilege to see alarm summary, disable button
    if (alarmSummaryDisabled == vTRUE)
    {
	buttonContainer -> summaryButton -> Disable ();
    }

    vdialog::Open();

    RefreshDisplay ();
}


/*******************************************************************
*
*    PopupDialog refresh function
*
*******************************************************************/

void PopupDialog::RefreshDisplay (void)
{
    vstr *titleStr;
    titleStr = vstrCloneScribed (vcharScribeSystem (groupPtr -> groupName));
    alarmGroupButton -> SetTitle (titleStr);
    vstrDestroy (titleStr);

    char *sev = alarmSeverityString[HighestSeverity (groupPtr -> groupData.numUnackedBySeverity)];
    titleStr = vstrCloneScribed (vcharScribeSystem (sev));
    SetTitle (titleStr);
    vstrDestroy (titleStr);

    vcolorSpec	*spec = new vcolorSpec;
    vcolor	*bgcolor;
    double	red, green, blue;

    GetRGBColor (SevColor [HighestSeverity (groupPtr -> groupData.numUnackedBySeverity)],
		 &red, &green, &blue);
    spec -> FSetRGB (red, green, blue);
    bgcolor = vcolor::Intern (spec);
    SetBackground (bgcolor);
    delete spec;
    if (groupPtr -> groupParms.displayName == NULL)
    {
	buttonContainer -> displayButton -> Disable ();
    }
}


/*******************************************************************
*
*    AlarmGroupButton Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(AlarmGroupButton, vbutton, "AlarmGroupButton");

void AlarmGroupButton::ObserveDialogItem (vevent *event)
{
    Filter  tempFilter = popupDialog -> filter;

    strcpy (tempFilter.group, popupDialog -> groupPtr -> groupName);
    tempFilter.group[sizeof (tempFilter.group)-1] = 0;
    if (dynamicGroupDialogPtr != NULL)
    {
	delete dynamicGroupDialogPtr;
    }
    vresource alvuRes = popupDialog -> app -> alvuResource;
    vresource alarmDlgRes = vresourceGet (alvuRes, alvu_AlarmDialog);
    dynamicGroupDialogPtr = new DynamicGroupDialog (alarmDlgRes);
    dynamicGroupDialogPtr -> AlarmDialogOpen (popupDialog -> app, &tempFilter);
    popupDialog -> DeleteLater ();
}


/*******************************************************************
*
*    ButtonContainer Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(ButtonContainer, vcontainer, "ButtonContainer");

void ButtonContainer::LoadInit (vresource resource)
{
    silenceButton   = PopupSilenceButton::RequestSubLoad (this, alvu_SilenceButton);
    displayButton   = PopupDisplayButton::RequestSubLoad (this, alvu_DisplayButton);
    scriptButton    = PopupScriptButton::RequestSubLoad (this, alvu_ScriptButton);
    closeButton     = PopupCloseButton::RequestSubLoad (this, alvu_CloseButton);
    summaryButton   = PopupSummaryButton::RequestSubLoad (this, alvu_SummaryButton);
    guidanceButton  = PopupGuidanceButton::RequestSubLoad (this, alvu_GuidanceButton);

    vcontainer::LoadInit (resource);
}


/*******************************************************************
*
*    Silence Button Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(PopupSilenceButton, vbutton, "PopupSilenceButton");

void PopupSilenceButton::ObserveDialogItem (vevent *event)
{
    BellTurnOff ();
}


/*******************************************************************
*
*    Display Button Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(PopupDisplayButton, vbutton, "PopupDisplayButton");

void PopupDisplayButton::ObserveDialogItem (vevent *event)
{
    if (HaveScript (popupDialog -> groupPtr -> groupParms.displayName))
    {
	SpawnGroupDisplay (popupDialog -> groupPtr);
    }
    popupDialog -> DeleteLater ();
}


/*******************************************************************
*
*    Script Button Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(PopupScriptButton, vbutton, "PopupScriptButton");

void PopupScriptButton::ObserveDialogItem (vevent *event)
{
    SpawnGroupScript (popupDialog -> groupPtr);
    popupDialog -> DeleteLater ();
}


/*******************************************************************
*
*    Close Button Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(PopupCloseButton, vbutton, "PopupCloseButton");

void PopupCloseButton::ObserveDialogItem (vevent *event)
{
    popupDialog -> Close ();
    popupDialog -> DeleteLater ();
}


/*******************************************************************
*
*    Summary Button Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(PopupSummaryButton, vbutton, "PopupSummaryButton");

void PopupSummaryButton::ObserveDialogItem (vevent *event)
{
    Filter  tempFilter = popupDialog -> filter;

    strcpy (tempFilter.group, popupDialog -> groupPtr -> groupName);
    tempFilter.group[sizeof (tempFilter.group) - 1] = 0;
    if (alarmSummaryDialogPtr != NULL)
    {
	delete alarmSummaryDialogPtr;
    }
    vresource alvuRes = popupDialog -> app -> alvuResource;
    vresource alarmDlgRes = vresourceGet (alvuRes, alvu_AlarmDialog);
    alarmSummaryDialogPtr = new AlarmSummaryDialog (alarmDlgRes);
    alarmSummaryDialogPtr -> AlarmDialogOpen (popupDialog -> app, &tempFilter);
    popupDialog -> DeleteLater ();
}


/*******************************************************************
*
*    Guidance Button Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(PopupGuidanceButton, vbutton, "PopupGuidanceButton");

void PopupGuidanceButton::ObserveDialogItem (vevent *event)
{
    if (guidanceDialogPtr != NULL)
    {
	delete guidanceDialogPtr;
	guidanceDialogPtr = NULL;
    }
    if (popupDialog -> groupPtr -> groupParms.guidance)
    {
	vresource alvuRes = popupDialog -> app -> alvuResource;
	vresource guidRes = vresourceGet (alvuRes, alvu_TextDialog);
	guidanceDialogPtr = new TextDialog (guidRes);
	guidanceDialogPtr -> OpenText (popupDialog -> app,
				       GUIDANCE_TEXT,
				       popupDialog -> groupPtr -> groupParms.guidance,
				       popupDialog -> groupPtr -> groupName);
    }
}

