#ifndef	reportdialogHXX
#define	reportdialogHXX


class ReportBrowseButton;
class ReportCancelButton;
class ReportOpenButton;
class ReportOKButton;
class ReportSaveButton;

// *** REPORT DIALOG CLASS *********************************************

class ReportDialog : public vdialog
{
    alvuApplication	*app;
    RepCriteria		criteria;	//RepCriteria defined in alarmfil.h
    int			result;

public:
    // overrides for vkindTyped
    vkindTYPED_FULL_DECLARATION(ReportDialog);

    // the default versions of necessary constructors/destructors
    vloadableINLINE_CONSTRUCTORS(ReportDialog, vdialog);

    // vloadable override
    void LoadInit (vresource resource);

    // override
    void ObserveWindow (vevent *event);

    // override
    void Destroy ();

    void OpenReport (alvuApplication *app, RepCriteria *criteria);

    filteredTextitem		*reportName;	// REPORTNAME_SIZE
    ReportOpenButton		*openButton;
    ReportBrowseButton		*browseButton;
    vcontrolToggle		*offScan;
    vcontrolToggle		*disabled;
    vcontrolToggle		*initFail;
    vcontrolToggle		*ioFail;
    vcontrolToggle		*badSensor;
    vcontrolToggle		*systemTag;
    vcontrolToggle		*infoTag;
    vcontrolToggle		*manual;
    vcontrolToggle		*pumpFault;
    vcontrolToggle		*valveFault;
    ReportSaveButton		*saveButton;
    ReportOKButton		*okButton;
    ReportCancelButton		*cancelButton;

    // some handy functions
    alvuApplication *GetApp (void)
	{ return (app); }
    RepCriteria *GetCriteria (void)
	{ return (&criteria); }
    int GetResult (void)
	{ return (result); }
    void UpdateCriteria (void);
    void UpdateReportSelections (void);
    void SetResult (int res)
	{ result = res; }

};


// *** BROWSE BUTTON CLASS ***********************************************

class ReportBrowseButton : public vbutton
{
    ReportDialog	*reportDialog;

public:
    // overrides for vkindTyped
    vkindTYPED_FULL_DECLARATION(ReportBrowseButton);

    // the default versions of necessary constructors/destructors
    vloadableINLINE_CONSTRUCTORS(ReportBrowseButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    // class extensions
    void StoreReportDialog (ReportDialog *dlg)
	{ reportDialog = dlg; }
};

// *** CANCEL BUTTON CLASS **********************************************

class ReportCancelButton : public vbutton
{
    ReportDialog	*reportDialog;

public:
    // overrides for vkindTyped
    vkindTYPED_FULL_DECLARATION(ReportCancelButton);

    // the default versions of necessary constructors/destructors
    vloadableINLINE_CONSTRUCTORS(ReportCancelButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    // class extensions
    void StoreReportDialog (ReportDialog *dlg)
	{ reportDialog = dlg; }
};


// *** OPEN BUTTON CLASS **********************************************

class ReportOpenButton : public vbutton
{
    ReportDialog	*reportDialog;

public:

    // overrides for vkindTyped
    vkindTYPED_FULL_DECLARATION(ReportOpenButton);

    // the default versions of necessary constructors/destructors
    vloadableINLINE_CONSTRUCTORS(ReportOpenButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    // class extensions
    void StoreReportDialog (ReportDialog *dlg)
	{ reportDialog = dlg; }
};

// *** OK BUTTON CLASS **********************************************

class ReportOKButton : public vbutton
{
    ReportDialog	*reportDialog;

public:
    // overrides for vkindTyped
    vkindTYPED_FULL_DECLARATION(ReportOKButton);

    // the default versions of necessary constructors/destructors
    vloadableINLINE_CONSTRUCTORS(ReportOKButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    // class extensions
    void StoreReportDialog (ReportDialog *dlg)
	{ reportDialog = dlg; }
};


// *** SAVE BUTTON CLASS **********************************************

class ReportSaveButton : public vbutton
{
    ReportDialog	*reportDialog;

public:
    // overrides for vkindTyped
    vkindTYPED_FULL_DECLARATION(ReportSaveButton);

    // the default versions of necessary constructors/destructors
    vloadableINLINE_CONSTRUCTORS(ReportSaveButton, vbutton);

    // override
    void ObserveDialogItem (vevent *event);

    // class extensions
    void StoreReportDialog (ReportDialog *dlg)
	{ reportDialog = dlg; }
};

#endif	reportdialogHXX

