#ifdef MIPSEL		/* DEK 3/9/94 - the MIPSEL C compiler doesn't support these */
#ifdef __cplusplus	/* DEK 3/9/94 - the C compiler doesn't support these */
#define	volatile
#define	signed
#endif
#endif

extern "C"
{
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <search.h>


#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include <sys/types.h>
#ifdef WIN32
#include <time.h>
#include <process.h>
#else
#include <sys/time.h>
#include <sys/ipc.h>
#ifdef SOLARIS
#include <errno.h>
#else
#include <sys/errno.h>
#endif
#include <sys/msg.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <fcntl.h>
#ifndef AXP
#include <sys/sockio.h>
#include <sys/filio.h>
#endif
#endif WIN32

#include "shareMem.h"
#include "db_access.h"
#include "alarmvu.h"
#include "OperPriv.h"
#include "alarmfil.h"
#include "messages.h"
}

#include "utils.hxx" 

#include <vport.h>		/* for v*HEADER */
#include vbuttonHEADER
#include vprintHEADER
#include "tisgalaxylib.hxx"
#include "alvu.h"
#include "alarmvu.hxx"
#include "alvuUtils.hxx"

/******************************************************************************
 *                                                                            *
 *     extern functions                                                       *
 *                                                                            *
 ******************************************************************************/
#ifdef MIPSEL
extern "C" void		bzero (char *, int);
extern "C" void		bcopy (char *, char *, int);
#endif

#if defined(SOLARIS) || defined(WIN32)
#define bzero(b,n)	memset (b, 0, n)
#define bcopy(s,d,n)	memcpy (d, s, n)
#endif


/******************************************************************************
 *                                                                            *
 *     extern variables                                                       *
 *                                                                            *
 ******************************************************************************/

extern "C"
{
extern TreeNode		*treeRoot;
}

extern int		groupCount;
extern int		channelCount;
extern int		alarmhndSocket;
extern char		*pointOptions[];
extern int		errno;
extern AckGroupPvPtrs	*ackLists;

/******************************************************************************
 *                                                                            *
 *     static variables                                                       *
 *                                                                            *
 ******************************************************************************/

#ifndef True
#define True	(1==1)
#define False	(1==0)
#endif True

static char		**groupNamePtrs;
static TreeNode		**groupPtrs;
static char		**channelNamePtrs;
static PvData		**channelPtrs;
static char		textbuf [120];

/******************************************************************************
 *                                                                            *
 *     static functions                                                       *
 *                                                                            *
 ******************************************************************************/

static int		CountGroups (TreeNode *);
static void		ScanGroups (TreeNode *);
static int		CountChannels (TreeNode *);
static void		ScanChannels (TreeNode *);
static int		groupNameCmp (TreeNode **, TreeNode **);
static int		channelNameCmp (PvData **, PvData **);

/******************************************************************************
 *                                                                            *
 *                                                                            *
 *     GetGroups ( groupPtrsAddr )                                            *
 *     GetChannels ( channelPtrsAddr )                                        *
 *                                                                            *
 *                                                                            *
 *     Inputs                                                                 *
 *                                                                            *
 *     groupPtrsAddr    address for a pointer to an array of group ptrs       *
 *                                                                            *
 *     channelPtrsAddr  address for a pointer to an array of channel ptrs     *
 *                                                                            *
 *                                                                            *
 *     Return value                                                           *
 *                                                                            *
 *     pointers to name arrays                                                *
 *                                                                            *
 *                                                                            *
 *     Description                                                            *
 *                                                                            *
 *     create, scan, and sort groups and channels                             *
 *                                                                            *
 *                                                                            *
 ******************************************************************************/

char **
GetGroups
(
    TreeNode	***groupPtrsAddr
)
{
    char	**groupNameArray;
    TreeNode	**groupArray;

    groupCount = CountGroups (treeRoot);
    if ( !(groupNameArray = (char **) malloc ((groupCount + 1) * sizeof (char *)))
      || !(groupArray = (TreeNode **) malloc ((groupCount + 1) * sizeof (TreeNode *))))
    {
	return (NULL);
    }
    groupPtrs = groupArray;
    ScanGroups (treeRoot);
    *groupPtrs = NULL;  // terminate with null entry

    qsort ((char *)groupArray, groupCount, sizeof (char *), (int (*)(const void *, const void *)) groupNameCmp);

    groupNamePtrs = groupNameArray;
    for (groupPtrs = groupArray; *groupPtrs; groupPtrs++)
    {
	*groupNamePtrs++ = (*groupPtrs)->groupName;
    }
    *groupNamePtrs = NULL; // terminate with null ptr entry
    groupNamePtrs = groupNameArray; // restore return value

    *groupPtrsAddr = groupArray;

    return (groupNameArray);
}

char **
GetChannels
(
    PvData	***channelPtrsAddr
)
{
    char	**channelNameArray;
    PvData	**channelArray;

    channelCount = CountChannels (treeRoot);
    if ( !(channelNameArray = (char **) malloc ((channelCount + 1) * sizeof (char *)))
      || !(channelArray = (PvData **) malloc ((channelCount + 1) * sizeof (PvData *))))
    {
	return (NULL);
    }
    channelPtrs = channelArray;
    ScanChannels (treeRoot);
    *channelPtrs = NULL;

    qsort ((char *)channelArray, channelCount, sizeof (char *),
	   (int (*) (const void *, const void *)) channelNameCmp);

    channelNamePtrs = channelNameArray;
    for (channelPtrs=channelArray; *channelPtrs; channelPtrs++)
    {
	*channelNamePtrs++ = (*channelPtrs)->pvName;
    }
    *channelNamePtrs = NULL;
    channelNamePtrs = channelNameArray;
    channelPtrs = channelArray;

    *channelPtrsAddr = channelArray;

    return (channelNameArray);
}

static int
CountGroups
(
    TreeNode	*groupLink
)
{
    int		count;
    TreeNode	*groupPtr;

    count = 1;
    groupPtr = groupLink->leftChild;
    while (groupPtr)
    {
	count += CountGroups (groupPtr);
	groupPtr = groupPtr->rightSibling;      
    }
    return(count);
}

static void
ScanGroups
(
    TreeNode	*groupLink
)
{
    TreeNode	*groupPtr;

    *groupPtrs++ = groupLink;
    for (groupPtr = groupLink->leftChild; groupPtr; groupPtr = groupPtr->rightSibling)
    {
	ScanGroups (groupPtr);
    }
}

static int
CountChannels
(
    TreeNode	*groupLink
)
{
    int		count;
    TreeNode	*groupPtr;
    PvData	*pvPtr;

    count = 0;
    for (pvPtr = groupLink->pvList; pvPtr; pvPtr = pvPtr->nextPv)
	count++;
    groupPtr = groupLink->leftChild;
    while (groupPtr)
    {
	count += CountChannels (groupPtr);
	groupPtr = groupPtr->rightSibling;      
    }
    return(count);
}

static void
ScanChannels
(
    TreeNode	*groupLink
)
{
    PvData	*chanPtr;
    TreeNode	*groupPtr;

    chanPtr = groupLink->pvList;
    while (chanPtr)
    {
	*channelPtrs++ = chanPtr;
	chanPtr = chanPtr->nextPv;
    }

    for (groupPtr = groupLink->leftChild; groupPtr; groupPtr = groupPtr->rightSibling)
    {
	ScanChannels (groupPtr);
    }
}

static int
groupNameCmp
(
    TreeNode **groupPtr1,
    TreeNode **groupPtr2
)
{
    return (strcmp((*groupPtr1)->groupName, (*groupPtr2)->groupName));
}

static int
channelNameCmp
(
    PvData **channelPtr1,
    PvData **channelPtr2
)
{
    return (strcmp((*channelPtr1)->pvName, (*channelPtr2)->pvName));
}

/******************************************************************************
 *                                                                            *
 *                                                                            *
 *     SpawnScript ( commandPtr )                                             *
 *     SpawnCommand ( commandPtr )                                            *
 *     SpawnDisplay ( displayname )                                           *
 *     SpawnDetail ( pointname )                                              *
 *                                                                            *
 *                                                                            *
 *     Inputs                                                                 *
 *                                                                            *
 *     commandPtr       point to command string                               *
 *                                                                            *
 *                                                                            *
 *     Return value                                                           *
 *                                                                            *
 *     none                                                                   *
 *                                                                            *
 *                                                                            *
 *     Description                                                            *
 *                                                                            *
 *     spawn unix command                                                     *
 *                                                                            *
 *                                                                            *
 ******************************************************************************/
void SpawnGroupScript (TreeNode *group)
{
    char *script;

    if (group->auxValue) {	// In aux mode
	script = group->groupParms.auxManualScript;
    } else {	// In normal mode
	script = group->groupParms.manualScript;
    }
    SpawnCommand ( ExpandScript (script, group, "", 0, 0, "", 0) );
}

void SpawnScript (PvData *p)
{
    char *script;

    if (p->parentGroup->auxValue) {	// In aux mode
	script = p->parentGroup->groupParms.auxManualScript;
    } else {	// In normal mode
	script = p->parentGroup->groupParms.manualScript;
    }
    SpawnCommand ( ExpandScript (script, p->parentGroup, p->pvName,
		   p->status, p->severity, p->value, p->Ack) );
}

int HaveScript (char *script)
{
    return ! ((*script == 0) || (*script == '*'));
}

char *ExpandScript (char *script, TreeNode *group, char *ptName,
		    int ptStatus, int ptSeverity, char *ptValue, int ptAck)
{
    static char	expScript [400];
    int i=0, j=0;

    if ((*script == 0) || (*script == '*'))
	return ("");

    // A script specifier can have the form:
    //
    //	myScript %N %T ...
    //
    // The flags are replaced by the flag followed by a value
    //
    //	myScript %N groupName %T groupStatus ...
    //
    // The following flags are recognized as:
    //
    //	%N	Group Name
    //	%T	Group Status
    //	%S	Group Severity
    //	%n	Point Name
    //	%t	Point Status
    //	%s	Point Severity
    //	%v	Point Value
    //	%a	Ack FLag (1 => needs acking)
    //  %% will be translated to a single %
    //
    while (script [i]) {
	if (script [i] == '%') {	// Found a flag
	    i++;
	    switch (script [i]) {
		case 'N':		// Group Name
		    sprintf (&expScript [j], "%s", group->groupName);
		    j += strlen (&expScript [j]);
		    break;
		case 'T':		// Group Status
		    sprintf (&expScript [j], "%d", group->groupData.groupStatus);
		    j += strlen (&expScript [j]);
		    break;
		case 'S':		// Group Severity
		    sprintf (&expScript [j], "%d", group->groupData.groupSeverity);
		    j += strlen (&expScript [j]);
		    break;
		case 'n':		// Point Name
		    sprintf (&expScript [j], "%s", ptName);
		    j += strlen (&expScript [j]);
		    break;
		case 't':		// Point Status
		    sprintf (&expScript [j], "%d", ptStatus);
		    j += strlen (&expScript [j]);
		    break;
		case 's':		// Point Severity
		    sprintf (&expScript [j], "%d", ptSeverity);
		    j += strlen (&expScript [j]);
		    break;
		case 'v':		// Point Value
		    sprintf (&expScript [j], "\"%s\"", ptValue);
		    j += strlen (&expScript [j]);
		    break;
		case 'a':		// Point is ACKed
		    sprintf (&expScript [j], "%d", ptAck);
		    j += strlen (&expScript [j]);
		    break;
		default:
		    sprintf (&expScript [j], "%", script [i]);
		    j += strlen (&expScript [j]);
		    break;
	    }
	    i++;
	} else {
	    expScript [j++] = script [i++];
	}
    }
    expScript [j] = '\0';
    return expScript;
}

void
SpawnCommand
(
    char	*commandPtr
)
{
	char	txtBuf[BUFSIZ];

	if (*commandPtr)
	{
	    sprintf (txtBuf, "csh -c \"%s\" &", commandPtr);
	    system (txtBuf);
	}
}

void SpawnGroupDisplay (TreeNode *group)
{
    char *script;

    script = group->groupParms.displayName;
    if ((*script) && (*script == '!'))
    {
	if (*(script+1) == '-')
	{
	    SpawnCommand ( ExpandScript (script+2, group, "", 0, 0, "", 0) );
	    exit (0);
	}
	else
	{
	    SpawnCommand ( ExpandScript (script+1, group, "", 0, 0, "", 0) );
	}
    }
    else if (*script)
    {
	if (*(script+1) == '-')
	{
	    SpawnDisplay ( ExpandScript (script+1, group, "", 0, 0, "", 0) );
	    exit (0);
	}
	else
	{
	    SpawnDisplay ( ExpandScript (script, group, "", 0, 0, "", 0) );
	}
    }
}

void SpawnPointDisplay (PvData *p)
{
    char *script;
    char *guidptr;

    GetAlarmInfo (p->pvName, &guidptr, &script);
    if (script == 0)
	script = p->parentGroup->groupParms.displayName;

    if ((*script) && (*script == '!'))
    {
	if (*(script+1) == '-')
	{
	    SpawnCommand ( ExpandScript (script+2, p->parentGroup, p->pvName,
			   p->status, p->severity, p->value, p->Ack) );
	    exit (0);
	}
	else
	{
	    SpawnCommand ( ExpandScript (script+1, p->parentGroup, p->pvName,
			   p->status, p->severity, p->value, p->Ack) );
	}
    }
    else if (*script)
    {
	if (*(script+1) == '-')
	{
	    SpawnDisplay ( ExpandScript (script+1, p->parentGroup, p->pvName,
			   p->status, p->severity, p->value, p->Ack) );
	    exit (0);
	}
	else
	{
	    SpawnDisplay ( ExpandScript (script, p->parentGroup, p->pvName,
			   p->status, p->severity, p->value, p->Ack) );
	}
    }
}

void
SpawnDisplay
(
    char	*displayname
)
{

	if (displayname)
	{
#ifndef WIN32
	    char	txtBuf[BUFSIZ];
	    sprintf (txtBuf, "csh -c \"dm %s &\"", displayname);
	    system (txtBuf);
#else
	    static char savedir[_MAX_PATH];
	    _getcwd (savedir, _MAX_PATH);
	    _chdir ("\\tis4000\\graphics");
	    _spawnlp (_P_DETACH, "dm", "dm", displayname, NULL);
	    _chdir (savedir);
#endif WIN32
	}
}

void
SpawnDetail
(
    char	*pointname
)
{

	if (pointname)
	{
#ifndef WIN32
	    char	txtBuf[BUFSIZ];
	    sprintf (txtBuf, "csh -c \"launchtype %s &\"", pointname);
	    system (txtBuf);
#else
	    static char savedir[_MAX_PATH];
	    _getcwd (savedir, _MAX_PATH);
	    _chdir ("\\tis4000\\graphics");
	    _spawnlp (_P_DETACH, "launchtype", "launchtype", pointname, NULL);
	    _chdir (savedir);
#endif WIN32
	}
}

/* Initialization of tisLib access */
int InitPrivTrack ()
{
    int status;
    static char	nameBuff[80+1];

    GetOperName (nameBuff);
    if ((status = InitOperPriv (nameBuff)) == PRIVOK)
    {
	status = InitOperTracking (nameBuff);
    }
    return status;
}

/* Utility program for priviledges */
int CheckPriv (char *chan)
{
    int status;

    status = CheckOperPriv (chan);
    return status;
}

/* Utility program for priviledges, with tracking */
int CheckTrackPriv (char *chan, char *OldValue, char *NewValue)
{
    int status;

    status = CheckOperPriv (chan);
    if (status == PRIVOK)
	WriteOperTracking( chan, OldValue, NewValue );
    return status;
}

/******************************************************************************
 *                                                                            *
 *                                                                            *
 *     DisablePoint ( channelName )                                           *
 *     EnablePoint ( channelName )                                            *
 *     AckAlarm ( channelName, time )                                         *
 *                                                                            *
 *     Inputs                                                                 *
 *                                                                            *
 *     channelName      pointer to channel name                               *
 *                                                                            *
 *     Return value                                                           *
 *                                                                            *
 *     none                                                                   *
 *                                                                            *
 *                                                                            *
 *     Description                                                            *
 *                                                                            *
 *     disable/enable/ack alarm point	                                      *
 *                                                                            *
 *                                                                            *
 ******************************************************************************/

int
DisablePoint
(
PvData	*pvptr
)
{
    char	testchannel [AL_POINTNAME_SIZE + 5];
    int		privStat;
    AckGroupPvPtrs *ackgroupptr;

    strncpy (testchannel, pvptr->pvName, AL_POINTNAME_SIZE);
    strcat (testchannel, ".AENB");
    if ((privStat = CheckPriv (testchannel)) != PRIVOK)
    {
#ifdef PRIV_DIAG
	alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, 
				 alvu_DisablePrivErr,privStat);
#endif PRIV_DIAG
	return False;
    }
    if (BeginMsgGroup (&ackgroupptr) == False)
	return False;

    if (AddMsgItem (&ackgroupptr, pvptr, DISABLE_MSG) == False)
    {
	return False;
    }
    EndMsgGroup (&ackgroupptr);

    return True;
}

int
EnablePoint
(
PvData	*pvptr
)
{
    char	testchannel [AL_POINTNAME_SIZE + 5];
    int		privStat;
    AckGroupPvPtrs *ackgroupptr;

    strncpy (testchannel, pvptr->pvName, AL_POINTNAME_SIZE);
    strcat (testchannel, ".AENB");
    if ((privStat = CheckPriv (testchannel)) != PRIVOK)
    {
#ifdef PRIV_DIAG
	alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, 
				 alvu_EnablePrivErr,privStat);
#endif PRIV_DIAG
	return False;
    }
    if (BeginMsgGroup (&ackgroupptr) == False)
    {
	return False;
    }
    if (AddMsgItem (&ackgroupptr, pvptr, ENABLE_MSG) == False)
    {
	return False;
    }
    EndMsgGroup (&ackgroupptr);
    return True;
}

int
AckAlarm
(
PvData	*pvptr
)
{
    char	testchannel [AL_POINTNAME_SIZE + 5];
    int		privStat;
    AckGroupPvPtrs *ackgroupptr;

    strncpy (testchannel, pvptr->pvName, AL_POINTNAME_SIZE);
    strcat (testchannel, ".ACK");
    privStat = CheckPriv (testchannel);
    if (privStat == PRIVDBERR)
    {
	int result = alvuMessageDialogCenter (NULL, ALVU_CAUTION_NOTICE, 
					      alvu_AckDeletedMsg);
	if (result == vTRUE)
	{
	    return True; /* We don't need two error messages! */
	}
    }
    else if (privStat != PRIVOK)
    {
#ifdef PRIV_DIAG
	alvuMessageDialogCenter (NULL, ALVU_INFO_NOTICE, 
				 alvu_AckPrivErr,privStat);
#endif PRIV_DIAG
	return False;
    }

    if (BeginMsgGroup (&ackgroupptr) == False)
	return False;
    if (((pvptr->unAckedSeverity != NO_ALARM) && !pvptr->Ack) || 
	(privStat == PRIVDBERR))
    {
	if (AddMsgItem (&ackgroupptr, pvptr, ACK_MSG) == False)
	{
	    return False;
    	}
    }
    EndMsgGroup (&ackgroupptr);
    return True;
}

int
AckGroups
(
    TreeNode	*group
)
{
    AckGroupPvPtrs *ackgroupptr;

    if (BeginMsgGroup (&ackgroupptr) == False)
	return False;
    if (AddMsgItem (&ackgroupptr, group, ACK_GROUP_MSG) == False)
    {
	return False;
    }
    EndMsgGroup (&ackgroupptr);

    return AckGroup (group);
}

int
AckGroup
(
    TreeNode	*group
)
{
    TreeNode	*tmpgroup;
    PvData	*pvptr;
    /* although this routine is recursive, these next variables
    ** are unused by the time of recursion */
    AckGroupPvPtrs *ackgroupptr;

    /* this routine is called with the AckGroup message already sent
    */

    /* Acknowledge any alarm pvs in this node, and then do subgroups */
    if (BeginMsgGroup (&ackgroupptr) == False)
	return False;

    for (pvptr = group->pvList; pvptr; pvptr = pvptr->nextPv)
    {
	if ((pvptr->unAckedSeverity != NO_ALARM) && !pvptr->Ack)
	{
	    if (AddMsgItem (&ackgroupptr, pvptr, ACK_MSG) == False)
	    {
		return False;
	    }
	}
    }
    EndMsgGroup (&ackgroupptr);

    for (tmpgroup = group->leftChild; tmpgroup; tmpgroup = tmpgroup->rightSibling)
    {
	if (AckGroup (tmpgroup) == False)
	    return False;
    }

    return True;
}

int
DisableGroups
(
    TreeNode	*group
)
{
    return DisableGroup (group);
}

int
DisableGroup
(
    TreeNode	*group
)
{
    TreeNode	*tmpgroup;
    PvData	*pvptr;
    /* although this routine is recursive, these next variables
    ** are unused by the time of recursion */
    AckGroupPvPtrs *ackgroupptr;

    /* this routine is called with the DisableGroup message already sent
    */

    /* Disable any alarm pvs in this node */
    if (BeginMsgGroup (&ackgroupptr) == False)
    {
	return False;
    }
    /* Acknowledge any alarm pvs in this node, and then do subgroups */
    for (pvptr = group->pvList; pvptr; pvptr = pvptr->nextPv)
    {
	if (AddMsgItem (&ackgroupptr, pvptr, DISABLE_MSG) == False)
	{
	    return False;
	}
    }
    EndMsgGroup (&ackgroupptr);

    for (tmpgroup = group->leftChild; tmpgroup; tmpgroup = tmpgroup->rightSibling)
    {
	if (DisableGroup (tmpgroup) == False)
	    return False;
    }

    return True;
}

int
EnableGroups
(
    TreeNode	*group
)
{
    return EnableGroup (group);
}

int
EnableGroup
(
    TreeNode	*group
)
{
    TreeNode	*tmpgroup;
    PvData	*pvptr;
    /* although this routine is recursive, these next variables
    ** are unused by the time of recursion */
    AckGroupPvPtrs *ackgroupptr;

    /* this routine is called with the AckGroup message already sent
    */
    /* Disable any alarm pvs in this node */
    if (BeginMsgGroup (&ackgroupptr) == False)
    {
	return False;
    }
    /* Acknowledge any alarm pvs in this node, and then do subgroups */
    for (pvptr = group->pvList; pvptr; pvptr = pvptr->nextPv)
    {
	if (AddMsgItem (&ackgroupptr, pvptr, ENABLE_MSG) == False)
	{
	    return False;
	}
    }
    EndMsgGroup (&ackgroupptr);

    for (tmpgroup = group->leftChild; tmpgroup; tmpgroup = tmpgroup->rightSibling)
    {
	if (EnableGroup (tmpgroup) == False)
	    return False;
    }

    return True;
}


int Send_storeMemoText ( char memoBuf[], AlarmLogRecord *recptr )
{
    AlarmMemo	*alarmMemo;
    AckGroupPvPtrs *ackgroupptr;

    if ((alarmMemo = (AlarmMemo	*)malloc (sizeof(AlarmMemo))) == 0)
	return False;

    bzero ((char *)alarmMemo, sizeof (alarmMemo));
    alarmMemo->msgType = ALARM_COMMENT_UPDATE;
    strncpy( alarmMemo->pointName, recptr->pointName, sizeof(alarmMemo->pointName));
    alarmMemo->timeStamp = recptr->eventtime;
    strncpy( alarmMemo->memoText, memoBuf, sizeof(alarmMemo->memoText) );

    if (BeginMsgGroup (&ackgroupptr) == False)
    {
	return False;
    }
    if (AddMsgItem (&ackgroupptr, alarmMemo, ALARM_COMMENT_UPDATE) == False)
    {
	return False;
    }
    EndMsgGroup (&ackgroupptr);
    return (TRUE);
}


int Send_removeMemoText ( AlarmLogRecord *recptr )
{
    return (Send_storeMemoText ( "", recptr));
}

/******************************************************************************
 *                                                                            *
 *                                                                            *
 *     CountGroupEntries ( groupLinkPtr, filtermask )                         *
 *                                                                            *
 *                                                                            *
 *     Inputs                                                                 *
 *                                                                            *
 *     groupLinkPtr     group link pointer                                    *
 *                                                                            *
 *     filter       filter mask		                                      *
 *                                                                            *
 *     severity         alarm severity                                        *
 *                                                                            *
 *                                                                            *
 *     Return value                                                           *
 *                                                                            *
 *     none                                                                   *
 *                                                                            *
 *                                                                            *
 *     Description                                                            *
 *                                                                            *
 *     count number of entries in group using filter mask                     *
 *                                                                            *
 *                                                                            *
 ******************************************************************************/

int CountGroupEntries
(
    TreeNode		*groupLinkPtr,
    Filter		*filter
)
{       
    int			numEntries;
    TreeNode		*groupPtr;
    PvData		*pvPtr;

    numEntries = 0;
    for (groupPtr = groupLinkPtr->leftChild; groupPtr; groupPtr = groupPtr->rightSibling)
    {
	if ((filter == 0) || !IsGroupFiltered (groupPtr, filter))
	{
	    numEntries++;
	}
    }
    for (pvPtr = groupLinkPtr->pvList; pvPtr; pvPtr = pvPtr->nextPv)
    {
	if ((filter == 0) || !IsChannelFiltered (pvPtr, filter))
	{
	    numEntries++;
	}
    }

    return (numEntries);
}

/******************************************************************************
 *                                                                            *
 *                                                                            *
 *     IsGroupFiltered ( groupLinkPtr, filter )   		              *
 *                                                                            *
 *     IsChannelFilter ( chanLinkPtr, filter )				      *
 *                                                                            *
 *                                                                            *
 *     Inputs                                                                 *
 *                                                                            *
 *     groupLinkPtr     group link pointer                                    *
 *     chanLinkPtr      channel link pointer                                  *
 *                                                                            *
 *     filter       	filter mask                                           *
 *                                                                            *
 *     severity         alarm severity                                        *
 *                                                                            *
 *                                                                            *
 *     Return value                                                           *
 *                                                                            *
 *     TRUE if yes, FALSE if no                                               *
 *                                                                            *
 *                                                                            *
 *     Description                                                            *
 *                                                                            *
 *     check if a group or channel should be filtered                         *
 *                                                                            *
 *                                                                            *
 ******************************************************************************/

int IsGroupFiltered
(
    TreeNode		*groupLinkPtr,
    Filter		*filter
)
{       
    TreeNode	*node;
    TS_STAMP	tstamp;

    // if alarms only, and not in alarm, return true
    if ((HighestSeverity (groupLinkPtr->groupData.numAlarmsBySeverity) == NO_ALARM)
      && (filter->alarmOnly == ALMONLYFILTER))
    {
	return True;
    }

    // if Acked and acked alarms aren't allowed, return true
    if ((HighestSeverity (groupLinkPtr->groupData.numUnackedBySeverity) == NO_ALARM)
      && (filter->ack & ACKFILTER))
    {
	return True;
    }

    // if in alarm (unacked) and unacked alarms aren't allowed, return true
    if ((HighestSeverity (groupLinkPtr->groupData.numUnackedBySeverity) != NO_ALARM)
      && (filter->ack & UNACKFILTER))
    {
	return True;
    }

    // test severity filters
    if (((HighestSeverity (groupLinkPtr->groupData.numAlarmsBySeverity) == MAJOR_ALARM)
	 && (filter->sev & MAJORFILTER))
      || ((HighestSeverity (groupLinkPtr->groupData.numAlarmsBySeverity) == MINOR_ALARM)
	 && (filter->sev & MINORFILTER))
      || ((HighestSeverity (groupLinkPtr->groupData.numAlarmsBySeverity) == INFO_ALARM)
	 && (filter->sev & INFOFILTER))
      || ((HighestSeverity (groupLinkPtr->groupData.numAlarmsBySeverity) == EVENT_ALARM)
	 && (filter->sev & EVENTFILTER)))
    {
	return True;
    }

    // test time filters
    if (strlen(filter->start))
    {
	GetGMT_TS_STAMP (filter->start, &tstamp);
	if (groupLinkPtr->groupData.lastStatusChange.secPastEpoch < tstamp.secPastEpoch)
	    return True;
    }
    if (strlen(filter->end))
    {
	GetGMT_TS_STAMP (filter->end, &tstamp);
	if (groupLinkPtr->groupData.lastStatusChange.secPastEpoch >= tstamp.secPastEpoch)
	    return True;
    }
/* -- this would require us to walk the whole subtree! ---
**    if (strlen(filter->op) && strlen(groupLinkPtr->op))
**    {
**	if (strcmp (filter->op, groupLinkPtr->op) == 0)
**	    return True;
**    }
*/
    // Finally, see if the group is descended from our filter group spec
    //   -if group is blank, all elements are visible.
    if (strlen(filter->group) == 0)
	return False;
    for (node = groupLinkPtr->parent; node; node = node->parent)
    {
	if (strcmp (node->groupName, filter->group) == 0)
	    break;
    }
    if (node) // we are a descendant
	return False;
    else // no, return filtered status
	return True;
}

int IsChannelFiltered
(
    PvData		*pvPtr,
    Filter		*filter
)
{
    return (IsChannelFiltered (pvPtr, filter, pvPtr->severity));
}

int IsChannelFiltered
(
    PvData		*pvPtr,
    Filter		*filter,
    int			severity
)
{
    TreeNode	*node;
    TS_STAMP	tstamp;

    // if alarms only, and not in alarm, return true
    if (((severity == NO_ALARM) || (severity == EVENT_ALARM)) && (filter->alarmOnly == ALMONLYFILTER))
	return True;

    // if Acked and acked alarms aren't allowed, return true
    if ((pvPtr->Ack) && (filter->ack & ACKFILTER))
	return True;

    // if in alarm (unacked) and unacked alarms aren't allowed, return true
    if (((severity != NO_ALARM) && (severity != EVENT_ALARM))
      && !pvPtr->Ack && (filter->ack & UNACKFILTER))
	return True;

    // test severity filters
    if (((severity == MAJOR_ALARM) && (filter->sev & MAJORFILTER))
      || ((severity == MINOR_ALARM) && (filter->sev & MINORFILTER))
      || ((severity == INFO_ALARM) && (filter->sev & INFOFILTER))
      || ((severity == EVENT_ALARM) && (filter->sev & EVENTFILTER)))
	return True;

    // test time filters
    if (strlen(filter->start))
    {
	GetGMT_TS_STAMP (filter->start, &tstamp);
	if (pvPtr->lastAlarm.secPastEpoch < tstamp.secPastEpoch)
	    return True;
    }
    if (strlen(filter->end))
    {
	GetGMT_TS_STAMP (filter->end, &tstamp);
	if (pvPtr->lastAlarm.secPastEpoch >= tstamp.secPastEpoch)
	    return True;
    }

    if (strlen(filter->op) && strlen(pvPtr->op))
    {
	if (strcmp (filter->op, pvPtr->op) == 0)
	    return True;
    }

    // Finally, see if the point is descended from our filter group spec
    //   -if group is blank, all elements are visible.
    if (strlen(filter->group) == 0)
	return False;
    for (node = pvPtr->parentGroup; node; node = node->parent)
    {
	if (strcmp (node->groupName, filter->group) == 0)
	    break;
    }
    if (node) // we are a descendant
	return False;
    else // no, return filtered status
	return True;
}

int GetGMT_TS_STAMP (char *timestr, TS_STAMP *tstamp)
{
    char *strptr;
    struct tm mytm;

    memset (&mytm, 0, sizeof(mytm));

    for (strptr = timestr; *strptr == ' '; strptr++)
	/* all the work is in the for statement */ ;
    mytm.tm_mon = atoi (strptr) - 1;
    for (; *strptr && *strptr != '/'; strptr++)
	/* all the work is in the for statement */ ;
    if (*strptr == 0) // check format
	return False;
    strptr++;
    mytm.tm_mday = atoi (strptr);
    for (; *strptr && *strptr != '/'; strptr++)
	/* all the work is in the for statement */ ;
    if (*strptr == 0) // check format
	return False;
    strptr++;
    mytm.tm_year = atoi (strptr);
    // make this work until 2090!
    if (mytm.tm_year < 90)
	mytm.tm_year += 100;

    // move through string to time section!
    for (; *strptr != ' '; strptr++)
	/* all the work is in the for statement */ ;
    for (; *strptr == ' '; strptr++)
	/* all the work is in the for statement */ ;
    mytm.tm_hour = atoi (strptr);
    for (; *strptr && *strptr != ':'; strptr++)
	/* all the work is in the for statement */ ;
    strptr++;
    mytm.tm_min = atoi (strptr);
    for (; *strptr && *strptr != ':'; strptr++)
	/* all the work is in the for statement */ ;
    strptr++;
    mytm.tm_sec = atoi (strptr);

#if defined(MIPSEL) || defined(SOLARIS) || defined(AXP) || defined(WIN32)
    tstamp->secPastEpoch = mktime (&mytm) - TS_EPOCH_SEC_PAST_1970;
#else
    tstamp->secPastEpoch = timelocal (&mytm) - TS_EPOCH_SEC_PAST_1970;
#endif
    tstamp->nsec = 0;
    return True;
}

/******************************************************************************
 *                                                                            *
 *                                                                            *
 *     TCPRecv ( socketDesc, buffer, length )                                 *
 *                                                                            *
 *                                                                            *
 *     Inputs                                                                 *
 *                                                                            *
 *     socketDesc       socket decriptor                                      *
 *                                                                            *
 *     buffer           address of buffer to receive message                  *
 *                                                                            *
 *     length           number of bytes to receive                            *
 *                                                                            *
 *                                                                            *
 *     Return value                                                           *
 *                                                                            *
 *     number of bytes received, or -1 if error                               *
 *                                                                            *
 *                                                                            *
 *     Description                                                            *
 *                                                                            *
 *     fill out and show the point options menu listbox                       *
 *                                                                            *
 *                                                                            *
 ******************************************************************************/

int
TCPRecv
(
    int		socketDesc,
    char	*buffer,
    int		length
)
{       
	int		bytesRead, totalRead;
	static char	*partialBuffer;
	static int	partialBufferSize = 0;

	bzero( buffer, length );
	if (partialBufferSize)
	{
	    bcopy (partialBuffer, buffer, partialBufferSize);
	    free (partialBuffer);
	    totalRead = partialBufferSize;
	    partialBufferSize = 0;
	}
	else
	{
	    totalRead = 0;
	}

	while (length-totalRead > 0)
	{
	    bytesRead = recv (socketDesc, &buffer[totalRead], length-totalRead, 0);
	    if( bytesRead <= 0 )  break;
	    totalRead += bytesRead;
	    if ((length - totalRead) <= 0)
	    {
		return (totalRead);
	    }
	}

#ifdef WIN32
	errno = WSAGetLastError ();
	if (errno == WSAEWOULDBLOCK && totalRead > 0)
#else
	if (errno == EWOULDBLOCK && totalRead > 0)
#endif WIN32
	{
	    partialBuffer = (char *)malloc (totalRead);
	    partialBufferSize = totalRead;
	}
	return (bytesRead);
}

/******************************************************************************
 *                                                                            *
 *                                                                            *
 *                                                                            *
 *     StripWhitespace ( stringPtr )                                          *
 *                                                                            *
 *                                                                            *
 *     Inputs                                                                 *
 *                                                                            *
 *     stringPtr        pointer to string                                     *
 *                                                                            *
 *                                                                            *
 *     Return value                                                           *
 *                                                                            *
 *     pointer to string                                                      *
 *                                                                            *
 *                                                                            *
 *     Description                                                            *
 *                                                                            *
 *     strip leading and trailing whitespace from a string                    *
 *                                                                            *
 *                                                                            *
 ******************************************************************************/

char *
StripWhitespace
(
    char	*stringPtr
)
{
	int	i;

	for (i=strlen (stringPtr)-1; i > 0 && isspace (*(stringPtr + i)); i--)
	{
	    *(stringPtr + i) = 0;
	}
	for (i=0; isspace (*(stringPtr + i)); i++)
	    ;
	strcpy (stringPtr, stringPtr + i);
	return (stringPtr);
}

int HighestSeverity (short sevr[])
{
	if (sevr[MAJOR_ALARM])
	    return(MAJOR_ALARM);
	else if (sevr[MINOR_ALARM])
	    return(MINOR_ALARM);
	else if (sevr[EVENT_ALARM])
	    return(EVENT_ALARM);
	else if (sevr[INFO_ALARM])
	    return(INFO_ALARM);
	return(NO_ALARM);
}

/*
**	utilities to manage message blocks out to the alarmMgr.
*/
int BeginMsgGroup (AckGroupPvPtrs **ackgroupptr)
{
    *ackgroupptr = (AckGroupPvPtrs *)malloc (sizeof (AckGroupPvPtrs));
    if (*ackgroupptr == 0)
    {
	printf ("~No memory for ack data available.\n");
	/* no point in continuing */
	return False;
    }

    (*ackgroupptr)->link = 0;
    (*ackgroupptr)->total = 0;
    (*ackgroupptr)->current = 0;
    return True;
}

int AddMsgItem (AckGroupPvPtrs **ackgroupptr, void *itemptr, int itemtype)
{
    AckGroupPvPtrs *ap;

    (*ackgroupptr)->pvptrs [(*ackgroupptr)->total] = itemptr;
    (*ackgroupptr)->msgtype [(*ackgroupptr)->total] = itemtype;
    (*ackgroupptr)->total++;
    if ((*ackgroupptr)->total >= MAXACKPVPTRS)
    {
	if (ackLists)
	{
	    for (ap=ackLists; ap->link; ap = (AckGroupPvPtrs *)ap->link)
		;
	    ap->link = *ackgroupptr;
	} else {
	    ackLists = *ackgroupptr;
	}
	*ackgroupptr = (AckGroupPvPtrs *)malloc (sizeof (AckGroupPvPtrs));
	if (*ackgroupptr == 0)
	{
	    printf ("~Error sending ack point message to alarm manager\n");
	    /* no point in continuing */
	    return False;
	}
	(*ackgroupptr)->link = 0;
	(*ackgroupptr)->total = 0;
	(*ackgroupptr)->current = 0;
    }
    return True;
}

int EndMsgGroup (AckGroupPvPtrs **ackgroupptr)
{
    AckGroupPvPtrs *ap;

    if ((*ackgroupptr)->total)
    {
	if (ackLists)
	{
	    for (ap=ackLists; ap->link; ap = (AckGroupPvPtrs *)ap->link)
		;
	    ap->link = *ackgroupptr;
	} else {
	    ackLists = *ackgroupptr;
	}
    }
    else
	free ((char *)*ackgroupptr);
    return True;
}
