/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
#include "db_access.h"			/* for real DBF_XXXX defs */
#include "cadef.h"			/* EPICS channel access defs */
#include "lstLib.h"			/* linked list library */

#define MIN_TIMEOUT	(0.1)		/* minimum channel access timeout in seconds */
#define STRING_VALUE_LENGTH	(34)	/* max size of string I/O value */
/*
	The following structure will be allocated when the process variable name supplied
	with the macro is NOT found. When this happens the channel ID for that
	variable will be used to add the point to the event handler list. The user variable
	that may be supplied with the 'ca_add_event' function will be the address of this
	ENTRY. On subsequent invocations of the macro, the table will be searched
	for the supplied process variable. When found, the value stored in the table will
	be returned to the spreadsheet. The event handler will be called asyncronously to
	123 any time the value of the points in its list have changed. The 'get' function
	will block on a 'ca_pend_event' before fetching the value. This means that the
	spreadsheet will lock-up until something happens!

	The functions and data described here have been reworked to fit the format for 
	'C' functions in the Applix environment. GCT 7-Jul-95
*/
typedef	union
	{
	char	byte_value;				/* union of possible values */
	int	long_value;				/* based on type from event handler */
	short	short_value;
	float	float_value;
	} CA_VALUE;
#define	CA_VALUE_SIZE	sizeof(CA_VALUE)

typedef struct table_entry
	{
	NODE		next;				/* next node in list or 0 if last */
	char		var[PVNAME_SZ+FLDNAME_SZ+1];
	CA_VALUE	value;
	CA_VALUE	lastvalue;			/* equal to current if changed */
	char		string[STRING_VALUE_LENGTH];	/* for those who return strings */
	char		laststring[STRING_VALUE_LENGTH]; /* for those who return strings */
	int		type;				/* returned type of value as DBF_XXXX */
	chid		channel_id;			/* channel ID for 'ca_put' */
	int		valid;				/* if 1 then data has been received */
	} ENTRY;					/* table entry structure */
#define	ENTRY_SIZE (sizeof(ENTRY))			/* for malloc */

#define	RED_STR_SIZE	4
#define	SECS_PER_DAY	86400.0
#define	SECS_PER_HOUR	3600
#define	SECS_PER_MINUTE	60
#define MAX_SAMPLE_COUNT 60
#define NUMSEV 5
#define NUMALM 22
#define NUMQUAL 5
#define	MAXSTRING	80
#define	MAXPATHLEN	1024
#define SEVMASK     0xF8000000
#define QUALMASK    0x07C00000
#define ALMMASK     0x003FFFFF
#define EDITBIT		0x04000000
#define QERRBIT		0x02000000
#define CONNERRBIT	0x01000000
#define	APPLIX_MINUTE	(0.0006944444)	/* one minute in Applix time */
#define NAME_ARG 0
#define VALUE_ARG 1
typedef enum
        {
        SAMPLE,
        AVERAGE,
        MINIMUM,
        MAXIMUM
        }REDUCTION_TYPE;
