/*bldCvtTable.c,v 1.1.1.1 1999/08/03 14:41:55 tis4000 Exp */

/*  @(#)bldCvtTable.c	1.6	6/7/91
 *	Author: Robert Zieman (ANL)
 *	Date:	6/03/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	mm-dd-yy	iii	Comment
 * .02	mm-dd-yy	iii	Comment
 * 	...
 */

#include <stdio.h>
#include <ctype.h>
#include <sys/file.h>		/* SUN support for disk io */
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <choice.h>
#include <cvtTable.h>
#include <sdrHeader.h>
#include <getSdrLine.h>
#include <errno.h>

void	*malloc();
void	*calloc();

#define TRUE 1
#define FALSE 0
#define MAX_BREAKS 100
#define SAME 0
#define BYTES50		50
#define BYTES20		20
#define SCAN_DIM		1
#define SCAN_STR1_2		2
#define SCAN_3END		3

char            inFile[BYTES50];
char            choiceSdr[BYTES50];
char            cvtSdr[BYTES50];
int             errno;

struct sdrHeader sdrHeader;
struct arrBrkTable *cvtTable = NULL;
struct choiceSet *choiceCvt = NULL;
struct brkInt   brkInt[MAX_BREAKS];

static struct inLineInfo {	/* temp storage for getSdrLine input fields */
    char            dim_keywd[BYTES20];
    int             dim_size;
    int             dim_entries;
    int             choice_num;
    char            brkName[BYTES50];
}               inLineInfo;

struct brkCreateInfo {
    float           engLow;	/* Lowest value desired: engineering units */
    float           engHigh;	/* Highest value desired: engineering units */
    long            rawLow;	/* Raw value for EngLow			 */
    long            rawHigh;	/* Raw value for EngHigh		 */
    float           accuracy;	/* accuracy desired in engineering units */
    float           tblEngFirst;/* First table value: engineering units */
    float           tblEngLast;	/* Last table value: engineering units	 */
    float           tblEngDelta;/* Change per table entry: eng units	 */
    long            nTable;	/* number of table entries 		 */
    /* (last-first)/delta + 1		 */
    float          *pTable;	/* addr of data table			 */
}               brkCreateInfo;


/*
 * BLDCVTTABLE
 *
 * builds SDR files
 *
 * INPUT:	cvtTable.ascii
 *
 * OUTPUT:	choiceCvt.sdr
 * 		cvtTable.sdr
 */

main(argc, argv)
    int             argc;
    char          **argv;
{
    FILE           *inFile_fp;
    FILE           *choiceSdr_fp;
    FILE           *cvtTable_fp;
    int             len;
    int             rtnval;
    int             no_fields;
    int             i;
    int             j;
    int             Icnt;
    long            Size1 = 0;
    long            Size2 = 0;
    long            SSize1 = 0;
    long            SSize2 = 0;
    long            no_breaks;
    char          **paptmp;
    char          **paptmp2;
    float          *ptemp;
    char           *addr1;
    char           *addr2;
    char           *addrS1;
    char           *addrS2;
    struct arrBrkTable *pTcvtTable;
    struct choiceSet *pTchoiceCvt;
    struct brkCreateInfo *pbrk = &brkCreateInfo;
    struct lineIn  *plineIn = NULL;
    struct inLineInfo *pInfo = &inLineInfo;
    len = 0;

    /* check cmd line */
    if (!cmd_CvtTable(argc, argv))
	return -1;

    /* allocate space for the getSdrLine input structure */
    if ((plineIn = (struct lineIn *)
	 calloc(1, (unsigned) sizeof(struct lineIn))) == NULL) {
	printf("bldCvtTable: calloc error\n");
	return (-1);
    }
    plineIn->fname = inFile;

    if (chk_align() < 0)
	return (-1);

    if ((inFile_fp = fopen((caddr_t) plineIn->fname, "r")) == NULL) {
	printf("bldCvtTable: fopen error errno=%i\nfile=%s\n", errno, plineIn->fname);
	return (-1);

    }
    if ((getSdrLine(inFile_fp, plineIn, NO_QUOTES)) < 0)
	return (-1);
    if (plineIn->eof == EOF) {
	printf("bldCvtTable: Error - Unexpected EOF");
	return (-1);
    }
    if (inScan(SCAN_DIM, pInfo, plineIn, pbrk) < 0)
	return (-1);

    if ((pTcvtTable = (struct arrBrkTable *)
	 calloc(1, (unsigned) sizeof(struct arrBrkTable))) == NULL) {
	printf("bldCvtTable: calloc error \n");
	return (-1);
    }
    Size1 = sizeof(struct arrBrkTable);
    if ((paptmp = (caddr_t*)
	 calloc(1, (unsigned) pInfo->dim_size * sizeof(caddr_t))) == NULL) {
	printf("bldCvtTable: calloc error \n");
	return (-1);
    }
    Size1 += pInfo->dim_size * sizeof(caddr_t);
    pTcvtTable->number = pInfo->dim_size;
    pTcvtTable->papBrkTable = (struct brkTable **) paptmp;
    if ((pTchoiceCvt = (struct choiceSet *)
	 calloc(1, (unsigned) sizeof(struct choiceSet))) == NULL) {
	printf("bldCvtTable: calloc error \n");
	return (-1);
    }
    Size2 = sizeof(struct choiceSet);
    if ((paptmp2 = (caddr_t*)
	 calloc(1, (unsigned) pInfo->dim_size * sizeof(caddr_t))) == NULL) {
	printf("bldCvtTable: calloc error \n");
	return (-1);
    }
    Size2 += pInfo->dim_size * sizeof(caddr_t);
    pTchoiceCvt->number = pInfo->dim_size;
    pTchoiceCvt->papChoice = (caddr_t*) paptmp2;
    Icnt = 0;
    FOREVER {
	if ((getSdrLine(inFile_fp, plineIn, TWO_QUOTES)) < 0)
	    return (-1);

	if (plineIn->eof == EOF) {
	    break;
	}
	if (inScan(SCAN_STR1_2, pInfo, plineIn, pbrk) < 0)
	    return (-1);
	len = strlen(pInfo->brkName);
	if ((pTchoiceCvt->papChoice[pInfo->choice_num] = (caddr_t)
	     calloc(1, (unsigned) len + 1)) == NULL) {
	    printf("bldCvtTable: calloc error \n");
	    return (-1);
	}
	strcpy((caddr_t) pTchoiceCvt->papChoice[pInfo->choice_num], pInfo->brkName);
	SSize2 += len + 1;
	/* No therm tables for choice_num < 2 */
	if ((pInfo->choice_num > 1) && (pInfo -> choice_num < 10)) {
	    if (inScan(SCAN_3END, pInfo, plineIn, pbrk) < 0)
		return (-1);
	    if ((pTcvtTable->papBrkTable[pInfo->choice_num] = (struct brkTable *)
		 calloc(1, (unsigned) sizeof(struct brkTable))) == NULL) {
		printf("bldCvtTable: calloc error \n");
		return (-1);
	    }
	    Size1 += sizeof(struct brkTable);
	    /* number of entries expected */
	    no_fields = (pbrk->tblEngLast - pbrk->tblEngFirst)
		/ pbrk->tblEngDelta + 1;
	    pbrk->nTable = no_fields;
	    /* allocate space for input therm data table */
	    if ((ptemp =
		 (float *) calloc(1, (unsigned) pbrk->nTable * sizeof(float))) == NULL) {
		printf("bldCvtTable: calloc error (pbrk->pTable)\n");
		return (-1);
	    }
	    plineIn->pnext = NULL;
	    /* read the data */
	    if ((rtnval = read_data(inFile_fp, ptemp,
				    no_fields, pInfo->brkName, plineIn)) < 0)
		return (-1);
	    if (rtnval < no_fields) {
		printf("bldCvtTable: Error - Too few data entries in %s \n",
		       pInfo->brkName);
		printf("line=%d	%s\n", plineIn->line_no, plineIn->fname);
		return (-1);
	    }
	    pbrk->pTable = ptemp;
	    /* create a break point table */
	    if ((create_break(pbrk, brkInt, (long) MAX_BREAKS, &no_breaks)) < 0)
		return (-1);
	    len = strlen(pInfo->brkName);
	    if ((pTcvtTable->papBrkTable[pInfo->choice_num]->name
		 = (caddr_t) calloc(1, (unsigned) len + 1)) == NULL) {
		printf("bldCvtTable: calloc error \n");
		return (-1);
	    }
	    SSize1 += len + 1;
	    strcpy((caddr_t) pTcvtTable->papBrkTable[pInfo->choice_num]->name,
		   pInfo->brkName);
	    pTcvtTable->papBrkTable[pInfo->choice_num]->number = no_breaks;
	    pTcvtTable->papBrkTable[pInfo->choice_num]->rawLow = pbrk->rawLow;
	    pTcvtTable->papBrkTable[pInfo->choice_num]->rawHigh = pbrk->rawHigh;
	    if ((pTcvtTable->papBrkTable[pInfo->choice_num]->papBrkInt
		 = (struct brkInt **)
		 calloc(1, (unsigned) no_breaks * sizeof(caddr_t))) == NULL) {
		printf("bldCvtTable: calloc error \n");
		return (-1);
	    }
	    Size1 += no_breaks * sizeof(caddr_t);
	    for (i = 0; i < no_breaks; i++) {
		/* allocate  SK structures in a loop */
		if ((pTcvtTable->papBrkTable[pInfo->choice_num]->papBrkInt[i]
		     = (struct brkInt *)
		     calloc(1, (unsigned) sizeof(struct brkInt))) == NULL) {
		    printf("bldCvtTable: calloc error \n");
		    return (-1);
		}
		Size1 += sizeof(struct brkInt);
		bcopy((caddr_t) &brkInt[i],
		      (caddr_t) pTcvtTable->papBrkTable[pInfo->choice_num]->papBrkInt[i],
		      sizeof(struct brkInt));
	    }			/* end for */
	    free((caddr_t) ptemp);
	}
	Icnt++;
	if (Icnt >= pInfo->dim_entries)
	    break;
    }
    if ((getc(inFile_fp)) != EOF) {
	printf("bldRecChoice: Error - extra character(s) in file %s\n", inFile);
	return (-1);
    }
    if (Icnt != pInfo->dim_entries) {
	printf("bldCvtTable: Error - Number of tables != Dimension Entries");
	return (-1);
    }
    /* allocate, init and write the contiguous area for cvtTable.sdr */
    SSize1 = (SSize1 + sizeof(long) - 1) / sizeof(long) * sizeof(long);
    Size1 += SSize1;
    if ((cvtTable = (struct arrBrkTable *) calloc(1, (unsigned) Size1)) == NULL) {
	printf("bldCvtTable: calloc error (struct arrBrkTable )\n");
	return (-1);
    }
    addr1 = (caddr_t) cvtTable + sizeof(struct arrBrkTable);
    cvtTable->number = pTcvtTable->number;
    cvtTable->papBrkTable = (struct brkTable **) addr1;
    addr1 += cvtTable->number * sizeof(long);
    addrS1 = (caddr_t) cvtTable + Size1 - SSize1;
    for (i = 0; i < pTcvtTable->number; i++) {
	if (pTcvtTable->papBrkTable[i] != NULL) {
	    cvtTable->papBrkTable[i] = (struct brkTable *) addr1;
	    addr1 += sizeof(struct brkTable);
	    cvtTable->papBrkTable[i]->name = addrS1;
	    strcpy((caddr_t) addrS1, pTcvtTable->papBrkTable[i]->name);
	    cvtTable->papBrkTable[i]->number = pTcvtTable->papBrkTable[i]->number;
	    cvtTable->papBrkTable[i]->rawLow = pTcvtTable->papBrkTable[i]->rawLow;
	    cvtTable->papBrkTable[i]->rawHigh = pTcvtTable->papBrkTable[i]->rawHigh;
	    cvtTable->papBrkTable[i]->papBrkInt = (struct brkInt **) addr1;
	    addr1 += cvtTable->papBrkTable[i]->number * sizeof(caddr_t);
	    /* note: always a minimum of two break points */
	    for (j = 0; j < pTcvtTable->papBrkTable[i]->number; j++) {
		cvtTable->papBrkTable[i]->papBrkInt[j] = (struct brkInt *) addr1;
		bcopy((caddr_t) pTcvtTable->papBrkTable[i]->papBrkInt[j],
		      (caddr_t) cvtTable->papBrkTable[i]->papBrkInt[j],
		      sizeof(struct brkInt));
		addr1 += sizeof(struct brkInt);
	    }
	    addrS1 += strlen(cvtTable->papBrkTable[i]->name) + 1;
	}
    }

    /* write output file. First, the cvtTable */
    if ((cvtTable_fp = fopen((caddr_t) cvtSdr, "w")) == NULL) {
	printf("bldCvtTable: fopen errno=%i rtnval=%i\nfile=%s\n",
	       errno, rtnval, cvtSdr);
	return (-1);
    }

    fprintf(cvtTable_fp, "cvtTable\n");
    fprintf(cvtTable_fp, "%d\n", cvtTable -> number);
    for (i = 0; i < cvtTable -> number; i++)
	{
	if (!(cvtTable -> papBrkTable[i]))
	    {
	    fprintf(cvtTable_fp, "\n");
	    continue;
	    }
	fprintf(cvtTable_fp, "%s\n", cvtTable -> papBrkTable[i] -> name);
	fprintf(cvtTable_fp, "%d\n", cvtTable -> papBrkTable[i] -> number);
	fprintf(cvtTable_fp, "%d\n", cvtTable -> papBrkTable[i] -> rawLow);
	fprintf(cvtTable_fp, "%d\n", cvtTable -> papBrkTable[i] -> rawHigh);
	for (j = 0; j < cvtTable -> papBrkTable[i] -> number; j++)
	    {
	    fprintf(cvtTable_fp, "%d\n", cvtTable -> papBrkTable[i] -> papBrkInt[j] -> raw);
	    fprintf(cvtTable_fp, "%f\n", cvtTable -> papBrkTable[i] -> papBrkInt[j] -> slope);
	    fprintf(cvtTable_fp, "%f\n", cvtTable -> papBrkTable[i] -> papBrkInt[j] -> eng);
	    }
	}
    if ((fclose(cvtTable_fp)) == EOF) {
	printf("bldCvtTable: fclose error. errno=%i\nfile=%s\n",
	       errno, cvtSdr);
	return (-1);
    }
    /* allocate, init and write the contiguous area for choiceCvt.sdr */
    SSize2 = (SSize2 + sizeof(long) - 1) / sizeof(long) * sizeof(long);
    Size2 += SSize2;
    if ((choiceCvt = (struct choiceSet *) calloc(1, (unsigned) Size2)) == NULL) {
	printf("bldCvtTable: calloc error \n");
	return (-1);
    }
    addr2 = (caddr_t) choiceCvt + sizeof(struct choiceSet);
    choiceCvt->number = pTchoiceCvt->number;
    choiceCvt->papChoice = (caddr_t*) addr2;
    addr2 += choiceCvt->number * sizeof(long);
    addrS2 = (caddr_t) choiceCvt + Size2 - SSize2;
    for (i = 0; i < choiceCvt->number; i++) {
	if (pTchoiceCvt->papChoice[i] != NULL) {
	    choiceCvt->papChoice[i] = (caddr_t) addrS2;
	    strcpy((caddr_t) addrS2, (caddr_t) pTchoiceCvt->papChoice[i]);
	    addrS2 += strlen(choiceCvt->papChoice[i]) + 1;
	}
    }

    /* write output file, second, choiceCvt */
    if ((choiceSdr_fp = fopen((caddr_t) choiceSdr, "w")) == NULL) {
	printf("bldCvtTable: fopen errno=%i rtnval=%i\nfile=%s\n",
	       errno, rtnval, choiceSdr);
	return (-1);
    }

    fprintf(choiceSdr_fp, "choiceCvt\n");
    fprintf(choiceSdr_fp, "%d\n", choiceCvt -> number);
    for (i = 0; i < choiceCvt -> number; i++)
	if (!(choiceCvt -> papChoice[i]))
	    fprintf(choiceSdr_fp, "\n");
	else
	    fprintf(choiceSdr_fp, "%s\n", choiceCvt -> papChoice[i]);

    if ((fclose(choiceSdr_fp)) == EOF) {
	printf("bldCvtTable: fclose error. errno=%i\nfile=%s\n",
	       errno, choiceSdr);
	return (-1);
    }
    return (0);
}				/* end of bldCvtTable.c */
UnixErrMessage(mess)
    char           *mess;
{
    printf("%s\n", mess);
}
/*
 *
 * CHK_ALIGN
 *
 * checks alignment of structures
 *
 */
int
chk_align()
{
    if ((sizeof(struct choiceSet) % 4)) {
	printf("bldCvtTable: struct %s size alignment error\n", "choiceSet");
	return (-1);
    }
    if ((sizeof(struct brkInt) % 4)) {
	printf("bldCvtTable: struct %s size alignment error\n", "brkInt");
	return (-1);
    }
    if ((sizeof(struct brkTable) % 4)) {
	printf("bldCvtTable: struct %s size alignment error\n", "brkTable");
	return (-1);
    }
    if ((sizeof(struct arrBrkTable) % 4)) {
	printf("bldCvtTable: struct %s size alignment error\n", "arrBrkTable");
	return (-1);
    }
    return (0);
}
/*
 * CMD_CVTTABLE
 *
 * checks command line arg usage
 *
 */
int
cmd_CvtTable(argc, argv)
    int             argc;
    char          **argv;
{
    extern char    *optarg;
    struct stat    *buf;
    int             c,
                    err;
    short           aflg,
                    oflg,
                    Oflg;
    aflg = oflg = Oflg = err = 0;
    if ((buf = (struct stat *)
	 calloc(1, (unsigned) sizeof(struct stat))) == NULL) {
	printf("bldCvtTable: calloc error\n");
	return 0;
    }
    while ((c = getopt(argc, argv, "a:o:O:")) != -1)
	switch (c) {
	case 'a':
	    aflg++;
	    strcpy(inFile, optarg);
	    if (stat(inFile, buf) == -1) {
		printf("bldCvtTable: Can't stat file=%s errno=%i\n",
		       inFile, errno);
		err++;
	    }
	    break;
	case 'O':
	    oflg++;
	    strcpy(choiceSdr, optarg);
	    break;
	case 'o':
	    Oflg++;
	    strcpy(cvtSdr, optarg);
	    break;
	}
    if (!aflg || !oflg || !Oflg || err) {
	printf("Usage: \n");
	printf(" <bldCvtTable -a cvtTable.ascii -o cvtTable.sdr -O choiceCvt.sdr >\n");
	return 0;
    }
    return 1;
}
/*
 * INSCAN
 *
 * convert and check requested field(s) format
 *
 * return	-1	error
 *		0	success
 */
int
inScan(scan_type, pInfo, plineIn, pbrk)
    int             scan_type;
    struct inLineInfo *pInfo;
    struct lineIn  *plineIn;
    struct brkCreateInfo *pbrk;
{
    int             err = 0;
    switch (scan_type) {
    case SCAN_DIM:
	if ((sscanf(plineIn->line, "%3s%i%i",
		    pInfo->dim_keywd,
		    &pInfo->dim_size,
		    &pInfo->dim_entries)) != 3) {
	    err = 1;
	    break;
	}
	if ((strcmp("DIM", pInfo->dim_keywd)) != SAME)
	    err = 2;
	break;
    case SCAN_STR1_2:
	if ((sscanf(plineIn->line, "%i \"%50[^\"]\"",
		    &pInfo->choice_num,
		    pInfo->brkName)) != 2)
	    err = 1;
	break;
    case SCAN_3END:
	if ((sscanf(plineIn->line, "%*i \"%*[^\"]\"%f%li%f%li%f%f%f%f",
		    &pbrk->engLow,
		    &pbrk->rawLow,
		    &pbrk->engHigh,
		    &pbrk->rawHigh,
		    &pbrk->accuracy,
		    &pbrk->tblEngFirst,
		    &pbrk->tblEngLast,
		    &pbrk->tblEngDelta)) != 8)
	    err = 1;
	break;
    default:
	printf("bldCvtTable: inScan(scan_type) not defined internally\n");
	printf("bldCvtTable: line=%i\n", plineIn->line_no);
	return (-1);
	break;
    }
    if (err == 1) {
	printf("bldCvtTable: Illegal field format file=%s\n", plineIn->fname);

	printf("bldCvtTable: line=%i\n", plineIn->line_no);
	return (-1);
    }
    if (err == 2) {
	printf("bldCvtTable: Error on DIM control line - line=%i\n", plineIn->line_no);
	printf("bldCvtTable: Illegal field format file=%s\n", plineIn->fname);
	return (-1);
    }
    return (0);
}
/*
 * GETSTRING
 *
 * copies next field from file to buffer
 *
 * return	(-1,0)	error (includes EOF as error),	success
 *
 */
static int
getstring(fp, pbuff, plineIn)
    FILE           *fp;
    char           *pbuff;
    struct lineIn  *plineIn;
{
    char           *Wptr;
    int             no_byts;
    int             anotherfield;
    /* if line empty, try to get another */
    if (plineIn->pnext == NULL) {
	if ((getSdrLine(fp, plineIn, NO_QUOTES)) < 0)
	    return (-1);
	if (plineIn->eof == EOF) {
	    printf("bldCvtTable: Error - Unexpected EOF");
	    return (-1);
	}
    }
    /* copy field string to buff */
    sscanf(plineIn->pnext, "%s%n", pbuff, &no_byts);
    /* step to next field */
    plineIn->pnext += no_byts;
    /* test rest of line empty */
    Wptr = plineIn->pnext;
    anotherfield = 0;
    while (*Wptr != '\n') {
	if (isspace(*Wptr++)) {
	    continue;
	} else {
	    anotherfield = 1;
	    break;
	}
    }
    if (!anotherfield)
	plineIn->pnext = NULL;
    return (0);
}
/*
 * READ_DATA
 *
 * reads data table to an array
 *
 * return	(-1,num) error, number of fields in table
 */
int
read_data(fp, pTempArr, nflds, pbrkName, plineIn)
    FILE           *fp;
    float          *pTempArr;
    int             nflds;
    char           *pbrkName;
    struct lineIn  *plineIn;
{
    char            buff[80];
    float           tmp;
    int             num = 0;
    FOREVER {
	if ((getstring(fp, buff, plineIn)) < 0)
	    return (-1);

	if ((strncmp(buff, "ENDTABLE", 8)) == SAME)
	    break;

	if ((sscanf(buff, "%f", &tmp)) != 1) {
	    printf("bldCvtTable: Error - Unexpected field\n");
	    printf("line=%d	%s\n", plineIn->line_no, plineIn->fname);
	    return (-1);
	}
	num++;
	if (num > nflds) {
	    printf("bldCvtTable: Error - Too many data entries in %s \n", pbrkName);
	    printf("line=%d	%s\n", plineIn->line_no, plineIn->fname);
	    return (-1);
	}
	*pTempArr++ = tmp;
    }
    return (num);
}
/*
 * CREATE_BREAK
 *
 * Create a breakpoint array from table
 *
 * return	(-1,0)	error,success
 *
 */

int
create_break(pbci, pabrkInt, max_breaks, n_breaks)
    struct brkCreateInfo *pbci;
    struct brkInt  *pabrkInt;
    long            max_breaks;
    long           *n_breaks;
{
    struct brkInt  *pbrkInt;
    float          *table = pbci->pTable;
    long            ntable = pbci->nTable;
    double          ilow,
                    ihigh,
                    tbllow,
                    tblhigh,
                    slope,
                    offset;
    int             ibeg,
                    iend,
                    i,
                    inc,
                    imax,
                    n;
    double          rawBeg,
                    engBeg,
                    rawEnd,
                    engEnd,
                    engCalc,
                    engActual,
                    error;
    int             valid,
                    all_ok,
                    expanding;
    /* make checks to ensure that brkCreateInfo makes sense */
    if (pbci->engLow >= pbci->engHigh) {
	UnixErrMessage("create_break: engLow >= engHigh");
	return (-1);
    }
    if ((pbci->engLow < pbci->tblEngFirst)
	    || (pbci->engHigh > pbci->tblEngLast)) {
	UnixErrMessage("create_break: engLow > engHigh");
	return (-1);
    }
    if (pbci->tblEngDelta <= 0.0) {
	UnixErrMessage("create_break: tblEngDelta <= 0.0");
	return (-1);
    }
    if (ntable < 3) {
	UnixErrMessage("raw data must have at least 3 elements");
	return (-1);
    }
/***************************************************************************
			 Convert Table to raw values
 *
 * raw and table values are assumed to be related by an equation of the form:
 *
 * raw = slope*table + offset
 *
 * The following algorithm converts each table value to raw units
 *
 * 1) Finds the locations in Table corresponding to engLow and engHigh
 *    Note that these locations need not be exact integers
 * 2) Interpolates to obtain table values corresponding to engLow and enghigh
 *    we now have the equations:
 *    rawLow = slope*tblLow + offset
 *    rawHigh = slope*tblHigh + offset
 * 4) Solving these equations for slope and offset gives:
 *    slope=(rawHigh-rawLow)/(tblHigh-tblLow)
 *    offset=rawHigh-slope*tblHigh
 * 5) for each table value set table[i]=table[i]*slope+offset
 *************************************************************************/
    /* Find engLow in Table  and then compute tblLow */
    ilow = (pbci->engLow - pbci->tblEngFirst) / (pbci->tblEngDelta);
    i = (int) ilow;
    if (i >= ntable - 1)
	i = ntable - 2;
    tbllow = table[i] + (table[i + 1] - table[i]) * (ilow - (float) i);
    /* Find engHigh in Table and then compute tblHigh */
    ihigh = (pbci->engHigh - pbci->tblEngFirst) / (pbci->tblEngDelta);
    i = (int) ihigh;
    if (i >= ntable - 1)
	i = ntable - 2;
    tblhigh = table[i] + (table[i + 1] - table[i]) * (ihigh - (float) i);
    /* compute slope and offset */
    slope = (pbci->rawHigh - pbci->rawLow) / (tblhigh - tbllow);
    offset = pbci->rawHigh - slope * tblhigh;
    /* convert table to raw units */
    for (i = 0; i < ntable; i++)
	table[i] = table[i] * slope + offset;

/*****************************************************************************
 *		Now create break point table
 *
 * The algorithm does the following:
 *
 * It finds one breakpoint interval at a time. For each it does the following:
 *
 * 1) Use a relatively large portion of the remaining table as an interval
 * 2) It attempts to use the entire interval as a breakpoint interval
 *	Success is determined by the following algorithm:
 *	  a) compute the slope using the entire interval
 *        b) for each table entry in the interval determine the eng value
 *	     using the slope just determined.
 *	  c) compare the computed value with eng value associated with table
 *        d) if all table entries are within the accuracy desired then success.
 * 3) If successful then attempt to expand the interval and try again.
 *    Note that it is expanded by up to 1/10 of the table size.
 * 4) If not successful reduce the interval by 1 and try again.
 *    Once the interval is being decreased it will never be increased again.
 * 5) The algorithm will ultimately fail or will have determined the optimum
 *    breakpoint interval
 *************************************************************************/

    /* Must start with table entry corresponding to engLow; */
    i = ilow;
    if (i >= ntable - 1)
	i = ntable - 2;
    rawBeg = table[i] + (table[i + 1] - table[i]) * (ilow - (float) i);
    engBeg = pbci->engLow;
    ibeg = (int) (ilow + 1.5);	/* Make sure that ibeg-ilow is at least .5 */
    /* start first breakpoint interval */
    n = 1;
    pbrkInt = pabrkInt;
    pbrkInt->raw = rawBeg;
    pbrkInt->eng = engBeg;
    /* determine next breakpoint interval */
    while ((engBeg <= pbci->engHigh) && (ibeg < ntable - 1)) {
	/* determine next interval to try. Up to 1/10 full range */
	iend = ibeg;
	inc = (int) ((ihigh - ilow) / 10.0);
	if (inc < 1)
	    inc = 1;
	valid = TRUE;
	/* keep trying intervals until cant do better */
	expanding = TRUE;	/* originally we are trying larger and larger
				 * intervals */
	while (valid) {
	    imax = iend + inc;
	    if (imax >= ntable) {
		/* don't go past end of table */
		imax = ntable - 1;
		inc = ntable - iend - 1;
		expanding = FALSE;
	    }
	    if (imax > (int) (ihigh + 1.0)) {	/* Don't go to far past
						 * engHigh */
		imax = (int) (ihigh + 1.0);
		inc = (int) (ihigh + 1.0) - iend;
		expanding = FALSE;
	    }
	    if (imax <= ibeg)
		break;		/* failure */
	    rawEnd = table[imax];
	    engEnd = pbci->tblEngFirst + (float) imax *(pbci->tblEngDelta);
	    slope = (engEnd - engBeg) / (rawEnd - rawBeg);
	    all_ok = TRUE;
	    for (i = ibeg + 1; i <= imax; i++) {
		engCalc = engBeg + slope * (table[i] - rawBeg);
		engActual = pbci->tblEngFirst + ((float) i) * (pbci->tblEngDelta);
		error = engCalc - engActual;
		if (error < 0.0)
		    error = -error;
		if (error >= pbci->accuracy) {
		    if (inc > 1) {
			/* we will be trying smaller intervals */
			expanding = FALSE;
			/* just decrease inc and let while(valid) try again */
			inc--;
		    } else {
			/* This is the best we can do.			 */
			/* We either failed or found optimal interval	 */
			valid = FALSE;
		    }
		    all_ok = FALSE;
		    break;
		}
	    }			/* end for */
	    if (all_ok) {
		iend = imax;
		/* if not expanding we found interval */
		if (!expanding)
		    break;
		/* will automatically try larger interval */
	    }
	}			/* end while(valid) */
	/* either we failed or optimal interval has been found */
	if ((iend <= ibeg) && (iend < (int) ihigh)) {
	    UnixErrMessage("Could not meet accuracy criteria");
	    return (-1);
	}
	pbrkInt->slope = slope;
	/* get ready for next breakpoint interval */
	if (n++ >= max_breaks) {
	    UnixErrMessage("Break point table too large");
	    return (-1);
	}
	ibeg = iend;
	pbrkInt++;
	rawBeg = rawEnd;
	engBeg = engEnd;
	pbrkInt->raw = rawBeg;
	pbrkInt->eng = engBeg+(pbrkInt->raw-rawBeg)*slope;
    }
    pbrkInt->slope = 0.0;
    *n_breaks = n;
    return (0);
}
