/*bldDevSup.c,v 1.1.1.1 1999/08/03 14:41:55 tis4000 Exp */

/*  @(#)bldDevSup.c	1.7	6/7/91
 *	Author: Robert Zieman (ANL)
 *	Date:	6/03/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	mm-dd-yy	iii	Comment
 * .02	mm-dd-yy	iii	Comment
 * 	...
 */

#include <stdio.h>
#include <ctype.h>
#include <sys/file.h>		/* SUN support for disk io */
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <choice.h>
#include <devSup.h>
#include <dbRecType.h>
#include <sdrHeader.h>
#include <getSdrLine.h>

#ifndef NULL
#define NULL 0
#endif

#ifndef FALSE
#define FALSE 0
#define TRUE  1
#endif

#define SAME 0
#define BYTES50		50
#define BYTES20		20
#define SCAN_DIM		1
#define SCAN_RECTYPE		2
#define SCAN_CHOICELINES		3

extern char    *calloc();
char            inFile1[BYTES50];
char            inFile2[BYTES50];
char            outFile1[BYTES50];
char            outFile2[BYTES50];
int             errno;

struct sdrHeader sdrHeader;
extern struct devChoiceRec *choiceDev;
extern struct recDevSup *devSup;

static struct inLineInfo {	/* temp storage for input fields */
    char            dim_keywd[BYTES20];
    int             dim_size;
    int             dim_entries;
    char            RecTyp_keywd[BYTES20];
    char            RecTyp_name[BYTES50];
    int             choice_size;
    int             choice_entries;
    int             choice_num;
    long            link_type;
    char            dsetName_str[BYTES50];
    char            choice_str[BYTES50];
}               inLineInfo;
/*
 * BLDDEVSUP
 *
 * INPUT:	devSup.ascii
 *       	dbRecType.sdr
 * OUTPUT:	devSup.sdr
 *              choiceDev.sdr
 */

main(argc, argv)
    int             argc;
    char          **argv;
{

    FILE           *inFile1_fp;
    FILE           *outFile1_fp;
    FILE           *outFile2_fp;
    FILE	   *outFile3_fp;
    int             len;
    int             rtnval = 0;
    short           found;
    short           i;
    short           j;
    short           Icnt;
    short           Jcnt;
    short           rec_no;
    long            size1;
    long            Ssize1 = 0;
    long            size2;
    long            Ssize2 = 0;
    char           *addr1;
    char           *addrS1;
    char           *addr2;
    char           *addrS2;
    char          **pApDsetName;
    struct lineIn  *plineIn;
    struct recDevSup *pTdevSup;
    struct devChoiceRec *pTchoiceDev;
    struct inLineInfo *pInfo = &inLineInfo;
    struct devChoiceSet **pApDevChoiceSet;
    struct devChoiceSet *pDevChoiceSet;
    struct devChoice **pApDevChoice;
    struct devChoice *pDevChoice;
    struct devSup **pApDevSup;
    struct devSup  *pDevSup;

    /* allocate space for the getSdrLine input structure */
    if ((plineIn = (struct lineIn *)
	 calloc(1, sizeof(struct lineIn))) == NULL) {
	printf("bldDevSup: calloc error\n");
	return (-1);
    }
    plineIn->fname = inFile1;

    /* check cmd line */
    if (!cmd_DevSup(argc, argv))
	return -1;

    if (chk_align() < 0)
	return (-1);

    /* load ascii record type names */
    if (sdrLoad(inFile2))
	return (-1);

    if ((inFile1_fp = fopen((caddr_t) inFile1, "r")) == NULL) {
	printf("bldDevSup: fopen error errno=%i\nfile=%s\n", errno, inFile1);
	return (-1);
    }
    if ((getSdrLine(inFile1_fp, plineIn, NO_QUOTES)) < 0)
	return (-1);
    if (plineIn->eof == EOF) {
	printf("bldDevSup: Error - Unexpected EOF\n");
	printf("line=%d	%s\n", plineIn->line_no, inFile1);
	return (-1);
    }
    if (inScan(SCAN_DIM, plineIn, pInfo) < 0)
	return (-1);

    if (pInfo->dim_size > dbRecType->number) {
	printf("bldDevSup: Error - DIM > dbRecType->number\n");
	return (-1);
    }
    if (pInfo->dim_entries > pInfo->dim_size) {
	printf("bldDevSup: Error - dim_entries > dim_size\n");
	return (-1);
    }
    /*
     * allocate temp space for structures and pointer arrays as needed.
     */
    if ((pTchoiceDev = (struct devChoiceRec *)
	 calloc(1, sizeof(struct devChoiceRec))) == NULL) {
	printf("bldDevSup: calloc error \n");
	return (-1);
    }
    size2 = sizeof(struct devChoiceRec);
    pTchoiceDev->number = dbRecType->number;

    /* allocate temp structure space */
    if ((pTdevSup = (struct recDevSup *)
	 calloc(1,  sizeof(struct recDevSup))) == NULL) {
	printf("bldDevSup: calloc error \n");
	return (-1);
    }
    size1 = sizeof(struct recDevSup);
    pTdevSup->number = dbRecType->number;

    /* allocate papDevChoiceSet ptr array */
    if ((pTchoiceDev->papDevChoiceSet = (struct devChoiceSet **)
       calloc(1, (unsigned)dbRecType->number * sizeof(caddr_t))) == NULL) {
	printf("bldDevSup: calloc error \n");
	return (-1);
    }
    pApDevChoiceSet = pTchoiceDev->papDevChoiceSet;
    size2 += dbRecType->number * sizeof(long);

    /* allocate papDevSup ptr array */
    if ((pTdevSup->papDevSup = (struct devSup **)
       calloc(1, (unsigned)(dbRecType->number * sizeof(caddr_t)))) == NULL) {
	printf("bldDevSup: calloc error \n");
	return (-1);
    }
    pApDevSup = pTdevSup->papDevSup;
    size1 += dbRecType->number * sizeof(long);

    Icnt = 0;
    /* search for next rectype */
    FOREVER {
	if ((getSdrLine(inFile1_fp, plineIn, TWO_QUOTES)) < 0)
	    return (-1);
	if (plineIn->eof == EOF)
	    break;
	if (inScan(SCAN_RECTYPE, plineIn, pInfo) < 0)
	    return (-1);
	/* Verify record type name */
	found = FALSE;
	for (i = 0; i < dbRecType->number; i++) {
	    if (dbRecType->papName[i] != NULL) {
		if (strcmp(dbRecType->papName[i],
			   pInfo->RecTyp_name) == SAME) {
		    found = TRUE;
		    rec_no = i;
		    break;
		}
	    }
	}
	if (!found) {
	    printf("bldDevSup: Error - Record type '%s' not defined\n",
		   pInfo->RecTyp_name);
	    return (-1);
	}
	if (pApDevSup[rec_no] != NULL) {
	    printf("bldDevSup: Error - RecType %s is already defined\n",
		   pInfo->RecTyp_name);
	    printf("line=%d	%s\n", plineIn->line_no, inFile1);
	    return (-1);
	}
	/* allocate struct devChoiceSet */
	if ((pApDevChoiceSet[rec_no] = (struct devChoiceSet *)
	     calloc(1, sizeof(struct devChoiceSet))) == NULL) {
	    printf("bldDevSup: calloc error \n");
	    return (-1);
	}
	pDevChoiceSet = pApDevChoiceSet[rec_no];
	size2 += sizeof(struct devChoiceSet);
	pDevChoiceSet->number = pInfo->choice_size;
	/* allocate struct devSup */
	if ((pApDevSup[rec_no] = (struct devSup *)
	     calloc(1,  sizeof(struct devSup))) == NULL) {
	    printf("bldDevSup: calloc error \n");
	    return (-1);
	}
	pDevSup = pApDevSup[rec_no];
	size1 += sizeof(struct devSup);
	pDevSup->number = pInfo->choice_size;
	/* allocate papDevChoice ptr array */
	if ((pDevChoiceSet->papDevChoice = (struct devChoice **)
	     calloc(1,  (unsigned)(pInfo->choice_size * sizeof(caddr_t))))
		== NULL) {
	    printf("bldDevSup: calloc error \n");
	    return (-1);
	}
	pApDevChoice = pDevChoiceSet->papDevChoice;
	size2 += pInfo->choice_size * sizeof(long);
	/* allocate dsetName ptr array */
	if ((pDevSup->dsetName = (caddr_t *)
	     calloc(1,  (unsigned)(pInfo->choice_size * sizeof(caddr_t)))) == NULL) {
	    printf("bldDevSup: calloc error \n");
	    return (-1);
	}
	pApDsetName = pDevSup->dsetName;
	size1 += pInfo->choice_size * sizeof(long);
	/* reserve space for ptr array (dset) to be init'd at runtime */
	size1 += pInfo->choice_size * sizeof(long);
	Jcnt = 0;
	/* search for next choice line */
	FOREVER {
	    if ((getSdrLine(inFile1_fp, plineIn, FOUR_QUOTES)) < 0)
		return (-1);
	    if (plineIn->eof == EOF) {
		printf("bldDevSup: Error - Unexpected EOF\n");
		printf("line=%d	%s\n", plineIn->line_no, inFile1);
		return (-1);
	    }
	    if (inScan(SCAN_CHOICELINES, plineIn, pInfo) < 0)
		return (-1);
	    if (pInfo->choice_num > pInfo->choice_size) {
		printf("bldDevSup: Error - choice_num > choice_size\n");
		return (-1);
	    }
	    /* check for previous definition */
	    if (pApDsetName[pInfo->choice_num] != NULL) {
		printf("bldDevSup: Error - choice_num=%d is already defined\n",
		       pInfo->choice_num);
		printf("line=%d	%s\n", plineIn->line_no, inFile1);
	    }
	    /* allocate struct devChoice */
	    if ((pApDevChoice[pInfo->choice_num] = (struct devChoice *)
		 calloc(1,  sizeof(struct devChoice))) == NULL) {
		printf("bldDevSup: calloc error \n");
		return (-1);
	    }
	    pDevChoice = pApDevChoice[pInfo->choice_num];
	    pDevChoice->link_type = pInfo->link_type;
	    size2 += sizeof(struct devChoice);
	    len = strlen(pInfo->choice_str);
	    /* allocate pchoice string array */
	    if ((pDevChoice->pchoice = (caddr_t)
		 calloc(1,  (unsigned)(len + 1))) == NULL) {
		printf("bldDevSup: calloc error \n");
		return (-1);
	    }
	    strcpy((caddr_t) pDevChoice->pchoice, pInfo->choice_str);
	    Ssize2 += len + 1;
	    len = strlen(pInfo->dsetName_str);
	    /* allocate dsetName string array */
	    if ((pApDsetName[pInfo->choice_num] = (caddr_t)
		 calloc(1, (unsigned)(len + 1))) == NULL) {
		printf("bldDevSup: calloc error \n");
		return (-1);
	    }
	    /* copy name */
	    strcpy((caddr_t) pApDsetName[pInfo->choice_num], pInfo->dsetName_str);
	    Ssize1 += len + 1;
	    Jcnt++;
	    if (Jcnt >= pInfo->choice_entries)
		break;
	}
	if (Jcnt != pInfo->choice_entries) {
	    printf("bldDevSup: Error - Number of choice_entries != Dimension Entries\n");
	    return (-1);
	}
	Icnt++;
	if (Icnt >= pInfo->dim_entries)
	    break;
    }
    if ((getc(inFile1_fp)) != EOF) {
	printf("bldDevSup: Error - extra character(s) in file %s\n", inFile1);
	return (-1);
    }
    if (Icnt != pInfo->dim_entries) {
	printf("bldDevSup: Error - Number of RecTypes != Dimension Entries\n");
	return (-1);
    }
    /* allocate a contiguous area for the devSup structure */
    Ssize1 = (Ssize1 + sizeof(long) - 1) / sizeof(long) * sizeof(long);
    size1 += Ssize1;
    if ((devSup = (struct recDevSup *) calloc(1, (unsigned)size1)) == NULL) {
	printf("bldDevSup: calloc error (struct recDevSup)\n");
	return (-1);
    }
    addr1 = (caddr_t) devSup + sizeof(struct recDevSup);
    devSup->papDevSup = (struct devSup **) addr1;
    devSup->number = pTdevSup->number;
    addr1 += dbRecType->number * sizeof(caddr_t);
    addrS1 = (caddr_t) devSup + size1 - Ssize1;
    /* loop thru structure, insert pointers and data */
    for (i = 0; i < dbRecType->number; i++) {
	if (pTdevSup->papDevSup[i] != NULL) {
	    devSup->papDevSup[i] = (struct devSup *) addr1;
	    devSup->papDevSup[i]->number = pTdevSup->papDevSup[i]->number;
	    addr1 += sizeof(struct devSup);
	    devSup->papDevSup[i]->dsetName = (caddr_t *) addr1;
	    addr1 += devSup->papDevSup[i]->number * sizeof(caddr_t);
	    devSup->papDevSup[i]->papDset = (struct dset **) addr1;
	    addr1 += devSup->papDevSup[i]->number * sizeof(caddr_t);
	    for (j = 0; j < devSup->papDevSup[i]->number; j++) {
		if (pTdevSup->papDevSup[i]->dsetName[j] != NULL) {
		    devSup->papDevSup[i]->dsetName[j] = addrS1;
		    len = strlen(pTdevSup->papDevSup[i]->dsetName[j]);
		    bcopy(pTdevSup->papDevSup[i]->dsetName[j],
			  devSup->papDevSup[i]->dsetName[j], len + 1);
		    addrS1 += len + 1;
		}
	    }
	}
    }

    /* allocate a contiguous area for the choiceDev structure */
    Ssize2 = (Ssize2 + sizeof(long) - 1) / sizeof(long) * sizeof(long);
    size2 += Ssize2;

    if ((choiceDev = (struct devChoiceRec *) calloc(1,(unsigned) size2)) == NULL) {
	printf("bldDevSup: calloc error (struct devChoiceRec )\n");
	return (-1);
    }
    addr2 = (caddr_t) choiceDev + sizeof(struct devChoiceRec);
    choiceDev->papDevChoiceSet = (struct devChoiceSet **) addr2;
    choiceDev->number = dbRecType->number;
    addr2 += dbRecType->number * sizeof(caddr_t);
    addrS2 = (caddr_t) choiceDev + size2 - Ssize2;
    /* loop thru structure, insert pointers and data */
    for (i = 0; i < dbRecType->number; i++) {
	if (pTchoiceDev->papDevChoiceSet[i]) {
	    choiceDev->papDevChoiceSet[i] = (struct devChoiceSet *) addr2;
	    choiceDev->papDevChoiceSet[i]->number = pTchoiceDev->papDevChoiceSet[i]->number;
	    addr2 += sizeof(struct devChoiceSet);
	    choiceDev->papDevChoiceSet[i]->papDevChoice = (struct devChoice **) addr2;
	    addr2 += choiceDev->papDevChoiceSet[i]->number * sizeof(caddr_t);
	    for (j = 0; j < pTchoiceDev->papDevChoiceSet[i]->number; j++) {
		if (pTchoiceDev->papDevChoiceSet[i]->papDevChoice[j]) {
		    choiceDev->papDevChoiceSet[i]->papDevChoice[j] = (struct devChoice *) addr2;
		    choiceDev->papDevChoiceSet[i]->papDevChoice[j]->link_type =
			pTchoiceDev->papDevChoiceSet[i]->papDevChoice[j]->link_type;
		    choiceDev->papDevChoiceSet[i]->papDevChoice[j]->pchoice
			= (caddr_t) addrS2;
		    len = strlen(pTchoiceDev->papDevChoiceSet[i]->papDevChoice[j]->pchoice);
		    strcpy((caddr_t)
		    choiceDev->papDevChoiceSet[i]->papDevChoice[j]->pchoice,
			   pTchoiceDev->papDevChoiceSet[i]->papDevChoice[j]->pchoice);
		    addrS2 += len + 1;
		    addr2 += sizeof(struct devChoice);
		}
	    }
	}
    }

    /* write the SDR files. First, devSup */
    if ((outFile1_fp = fopen((caddr_t) outFile1, "w")) == NULL) {
	printf("bldDevSup: fopen errno=%i \nfile=%s\n",
	       errno, outFile1);
	return (-1);
    }

    fprintf(outFile1_fp, "devSup\n");
    fprintf(outFile1_fp, "%d\n", devSup -> number);
    for (i = 0; i < devSup -> number; i++)
	{
	if (!(devSup -> papDevSup[i]))
	    {
	    fprintf(outFile1_fp, "0\n");
	    continue;
	    }
	else
	    {
	    fprintf(outFile1_fp, "%d\n", devSup -> papDevSup[i] -> number);
	    for (j = 0; j < devSup -> papDevSup[i] -> number; j++)
		fprintf(outFile1_fp, "%s\n", devSup -> papDevSup[i] -> dsetName[j]);
	    }
	}

    if ((fclose(outFile1_fp)) == EOF) {
	printf("bldDevSup: fclose error. errno=%i\nfile=%s\n",
	       errno, outFile1);
	return (-1);
    }

    /* Second, choiceDev */
    if ((outFile2_fp = fopen((caddr_t) outFile2, "w")) == NULL) {
	printf("bldDevSup: fopen errno=%i rtnval=%i\nfile=%s\n",
	       errno, rtnval, outFile2);
	return (-1);
    }

    fprintf(outFile2_fp, "choiceDev\n");
    fprintf(outFile2_fp, "%d\n", choiceDev -> number);
    for (i = 0; i < choiceDev -> number; i++)
	{
	if (!(choiceDev -> papDevChoiceSet[i]))
	    {
	    fprintf(outFile2_fp, "0\n");
	    continue;
	    }
	else
	    {
	    fprintf(outFile2_fp, "%d\n", choiceDev -> papDevChoiceSet[i] -> number);
	    for (j = 0; j < choiceDev -> papDevChoiceSet[i] -> number; j++)
		{
		if (!(choiceDev -> papDevChoiceSet[i] -> papDevChoice[j]))
		    {
		    fprintf(outFile2_fp, "-1\n");
		    fprintf(outFile2_fp, "\n");
		    continue;
		    }
		else
		    {
		    fprintf(outFile2_fp, "%d\n", choiceDev -> papDevChoiceSet[i] -> papDevChoice[j] -> link_type);
		    fprintf(outFile2_fp, "%s\n", choiceDev -> papDevChoiceSet[i] -> papDevChoice[j] -> pchoice);
		    }
		}
	    }
	}

    if ((fclose(outFile2_fp)) == EOF) {
	printf("bldDevSup: fclose error. errno=%i\nfile=%s\n",
	       errno, outFile2);
	return (-1);
    }

    /* Third, ../epicsH/devType.h */
    if ((outFile3_fp = fopen("../epicsH/devType.h", "w")) == NULL ) {
	perror("bldDevSup on devType.h");
	return (-1);
    }

    fprintf(outFile3_fp, "/****************************************************************/\n");
    fprintf(outFile3_fp, "/*								*/\n");
    fprintf(outFile3_fp, "/*		BOARD TYPE SYMBOLS FOR THE TIS-4000		*/\n");
    fprintf(outFile3_fp, "/*								*/\n");
    fprintf(outFile3_fp, "/*	WARNING! WARNING! WARNING! WARNING! WARNING! WARNING!	*/\n");
    fprintf(outFile3_fp, "/*								*/\n");
    fprintf(outFile3_fp, "/*	DO NOT EDIT THIS FILE! IT IS AUTOMAGICALLY CREATED BY	*/\n");
    fprintf(outFile3_fp, "/*	THE bldDevSup PROGRAM AS PART OF CREATING THE SDR!	*/\n");
    fprintf(outFile3_fp, "/*								*/\n");
    fprintf(outFile3_fp, "/****************************************************************/\n\n");


    for (i = 0; i < choiceDev -> number; i++)
	{
	if (!(choiceDev -> papDevChoiceSet[i]))
	    continue;
	for (j = 0; j < choiceDev -> papDevChoiceSet[i] -> number; j++)
	    if (choiceDev -> papDevChoiceSet[i] -> papDevChoice[j])
		{
		char	tempBuff[50];
		char	*bytePtr;

		strcpy(tempBuff, choiceDev -> papDevChoiceSet[i] -> papDevChoice[j] -> pchoice);
		while (bytePtr = (char *) strchr(tempBuff, ' '))
		    *bytePtr = '_';
		while (bytePtr = (char *) strchr(tempBuff, '-'))
		    *bytePtr = '_';
		fprintf(outFile3_fp, "#define %s_%s\t%d\n", dbRecType -> papName[i], tempBuff, j);
		}
	}
    if ((fclose(outFile3_fp)) == EOF) {
	perror("bldDevSup closing devType.h");
	return (-1);
    }

    return (0);
}				/* end of  bldDevSup.c */
/*
 *
 * CHK_ALIGN
 *
 * checks alignment of input header structures
 *
 */
chk_align()
{
    if ((sizeof(struct recType) % 4)) {
	printf("bldDevSup: struct %s size alignment error\n", "recType");
	return (-1);
    }
    if ((sizeof(struct recDevSup) % 4)) {
	printf("bldDevSup: struct %s size alignment error\n", "recDevSup");
	return (-1);
    }
    if ((sizeof(struct devSup) % 4)) {
	printf("bldDevSup: struct %s size alignment error\n", "devSup");
	return (-1);
    }
    if ((sizeof(struct dset) % 4)) {
	printf("bldDevSup: struct %s size alignment error\n", "dset");
	return (-1);
    }
    if ((sizeof(struct devChoiceRec) % 4)) {
	printf("bldDevSup: struct %s size alignment error\n", "devChoiceRec");
	return (-1);
    }
    if ((sizeof(struct devChoiceSet) % 4)) {
	printf("bldDevSup: struct %s size alignment error\n", "devChoiceSet");
	return (-1);
    }
    if ((sizeof(struct devChoice) % 4)) {
	printf("bldDevSup: struct %s size alignment error\n", "devChoice");
	return (-1);
    }
    return (0);
}
/*
 * CMD_DEVSUP
 *
 * checks command line arg usage
 *
 * return (1,0)	success, fail
 */
cmd_DevSup(argc, argv)
    int             argc;
    char          **argv;
{
    extern char    *optarg;
    struct stat    *buf;
    int             c;
    int             err = 0;
    short           aflg = 0;
    short           bflg = 0;
    short           oflg = 0;
    short           Oflg = 0;

    if ((buf = (struct stat *)
	 calloc(1, sizeof(struct stat))) == NULL) {
	printf("bldDevSup: calloc error\n");
	return (0);
    }
    while ((c = getopt(argc, argv, "a:b:o:O:")) != -1)
	switch (c) {
	case 'a':
	    aflg++;
	    strcpy(inFile1, optarg);
	    if (stat(inFile1, buf) == -1) {
		printf("bldDevSup: Can't stat file=%s errno=%i\n",
		       inFile1, errno);
		err++;
	    }
	    break;
	case 'b':
	    bflg++;
	    strcpy(inFile2, optarg);
	    if (stat(inFile2, buf) == -1) {
		printf("bldDevSup: Can't stat file=%s errno=%i\n",
		       inFile2, errno);
		err++;
	    }
	    break;
	case 'o':
	    oflg++;
	    strcpy(outFile1, optarg);
	    break;
	case 'O':
	    Oflg++;
	    strcpy(outFile2, optarg);
	    break;
	}
    if (!aflg || !bflg || !oflg || !Oflg || err) {
	printf("Usage: \n");
	printf(" <bldDevSup -a devSup.ascii -b dbRecType.sdr -o devSup.sdr -O choiceDev.sdr >\n");
	return 0;
    }
    return 1;
}
/*
 * INSCAN
 *
 * return:	(0,-1) success, failure
 */
inScan(scan_type, plineIn, pInfo)
    int             scan_type;
    struct lineIn  *plineIn;
    struct inLineInfo *pInfo;
{
    int             err = 0;
    switch (scan_type) {
    case SCAN_DIM:
	if ((sscanf(plineIn->line, "%3s%i%i",
		    pInfo->dim_keywd,
		    &pInfo->dim_size,
		    &pInfo->dim_entries)) != 3) {
	    err = 1;
	    break;
	}
	if ((strcmp("DIM", pInfo->dim_keywd)) != SAME)
	    err = 3;
	break;
    case SCAN_RECTYPE:
	if ((sscanf(plineIn->line, "%7s \"%50[^\"]\"%i%i",
		    pInfo->RecTyp_keywd,
		    pInfo->RecTyp_name,
		    &pInfo->choice_size,
		    &pInfo->choice_entries)) != 4) {
	    err = 1;
	    break;
	}
	if ((strcmp("RECTYPE", pInfo->RecTyp_keywd)) != SAME)
	    err = 2;
	break;
    case SCAN_CHOICELINES:
	if ((sscanf(plineIn->line, "%i%ld \"%50[^\"]\" \"%50[^\"]",
		    &pInfo->choice_num,
		    &pInfo->link_type,
		    pInfo->dsetName_str,
		    pInfo->choice_str)) != 4)
	    err = 1;
	break;
    default:
	printf("bldDevSup: inScan(scan_type) not defined internally\n");
	printf("line=%i\n", plineIn->line_no);
	return (-1);
	break;
    }
    if (err == 1) {
	printf("bldDevSup: Illegal field format file=%s\n", inFile1);
	printf("line=%i\n", plineIn->line_no);
	return (-1);
    }
    if (err == 2) {
	printf("bldDevSup: Error on RECTYP control line - line=%i\n", plineIn->line_no);
	printf("Illegal field format file=%s\n", inFile1);
	return (-1);
    }
    if (err == 3) {
	printf("bldDevSup: Error on DIM control line - line=%i\n", plineIn->line_no);
	printf("Illegal field format file=%s\n", inFile1);
	return (-1);
    }
    return (0);
}
