/*bldGblChoice.c,v 1.1.1.1 1999/08/03 14:41:56 tis4000 Exp */

/*  @(#)bldGblChoice.c	1.8	6/7/91
 *	Author: Robert Zieman (ANL)
 *	Date:	6/03/91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	mm-dd-yy	iii	Comment
 * .02	mm-dd-yy	iii	Comment
 * 	...
 */

#include <stdio.h>
#include <ctype.h>
#include <sys/file.h>		/* SUN support for disk io */
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <choice.h>
#include <sdrHeader.h>
#include <getSdrLine.h>

#define SAME 0

extern char    *calloc();
int             errno;
struct sdrHeader sdrHeader;
struct arrChoiceSet *choiceGbl = NULL;
struct arrChoiceSet *pTchoiceGbl = NULL;
struct lineIn  *plineIn = NULL;

char            inFile[50];
char            outFile[50];
/*
 * BLDGBLCOICE.C
 *
 * Builds a self defining record file
 *
 * INPUT:     choiceGbl.ascii
 *
 * OUTPUT     choiceGbl.sdr
 *
 */

main(argc, argv)
    int             argc;
    char          **argv;
{

    FILE           *inFile_fp,
                   *fopen();
    FILE           *outFile_fp;
    int             no_byts;

    long            inx_num,
                    choice_num,
                    inx_size,
                    inx_entries,
                    choice_size,
                    choice_entries,
                    SSize,
                    Size,
                    len,
                    i,
                    j;
    char            choice_name[MAXBUF],
                    kwrd_inx[MAXBUF],
                    kwrd_dim[MAXBUF],
                   *addr1,
                   *addr2;
    int
                    inx,
                    chx;
    Size = SSize = 0;

    /* check cmd line usage */
    if (!cmd_choiceGbl(argc, argv))
	return -1;

    /* allocate space for the getSdrLine input structure */
    if ((plineIn = (struct lineIn *)
	 calloc(1, (unsigned) sizeof(struct lineIn))) == NULL) {
	printf("bldGblChoice: calloc error\n");
	return (-1);
    }
    if (chk_align() < 0)
	return (-1);

    if ((inFile_fp = fopen((char *) inFile, "r")) == NULL) {
	printf("bldGblChoice: fopen error errno=%i\nfile=%s\n", errno, inFile);
	return (-1);
    }
    plineIn->fname = inFile;

    if ((inFile_fp = fopen((char *) inFile, "r")) == NULL) {
	printf("bldGblChoice: fopen error errno=%i\nfile=%s\n", errno, inFile);
	return (-1);
    }
    /* expect Dim sdrLine */
    if ((getSdrLine(inFile_fp, plineIn, NO_QUOTES)) < 0)
	return (-1);

    if (plineIn->eof == EOF) {
	printf("bldGblChoice: Error - DIM control line not found:");
	printf("	%s\n", inFile);
	return (-1);
    }
    if ((sscanf(plineIn->line, "%s %li %li", kwrd_dim, &inx_size, &inx_entries)) != 3) {
	printf("bldGblChoice: illegal field in file %s\nline=%i\n",
	       inFile, plineIn->line_no);
	return (-1);
    }
    if ((strcmp("DIM", kwrd_dim)) != SAME) {
	printf("bldGblChoice: DIM control line not found at line=%i\n",
	       plineIn->line_no);
	return (-1);
    }
    /* allocate space as needed for structures and pointer arrays */
    if ((pTchoiceGbl = (struct arrChoiceSet *)
	 calloc(1, (unsigned) sizeof(struct arrChoiceSet))) == NULL) {
	printf("bldGblChoice: calloc error\n");
	return (-1);
    }
    Size = sizeof(struct arrChoiceSet);

    /* allocate papChoiceSet ptr array */
    if ((pTchoiceGbl->papChoiceSet = (struct choiceSet **)
	 calloc(1, (unsigned) inx_size * sizeof(char *))) == NULL) {
	printf("bldGblChoice: calloc error\n");
	return (-1);
    }
    Size += inx_size * sizeof(char *);

    inx = 0;
    /* expect index sdrLine */
    FOREVER {
	if ((getSdrLine(inFile_fp, plineIn, NO_QUOTES)) < 0)
	    return (-1);
	if (plineIn->eof == EOF)
	    break;
	if ((sscanf(plineIn->line, "%s%li%li%li",
		 kwrd_inx, &inx_num, &choice_size, &choice_entries)) != 4) {
	    printf("bldGblChoice: illegal field in file %s\n", inFile);
	    printf("bldGblChoice: line=%i\n", plineIn->line_no);
	    return (-1);
	}
	if ((strcmp("INDEX", kwrd_inx)) != SAME) {
	    printf("bldGblChoice: INDEX control line not found on line=%i\n",
		   plineIn->line_no);
	    return (-1);
	}
	if (inx_num > inx_size) {
	    printf("bldGblChoice: Error - inx_num > inx_size\nline=%s\n",
		   plineIn->line);
	    return (-1);
	}
	/* allocate struct choiceSet */
	if ((pTchoiceGbl->papChoiceSet[inx_num] = (struct choiceSet *)
	     calloc(1, (unsigned) sizeof(struct choiceSet))) == NULL) {
	    printf("bldGblChoice: calloc error\n");
	    return (-1);
	}
	Size += sizeof(struct choiceSet);

	/* allocate papChoice ptr array */
	if ((pTchoiceGbl->papChoiceSet[inx_num]->papChoice = (char **)
	     calloc(1, (unsigned) choice_size * sizeof(char *))) == NULL) {
	    printf("bldGblChoice: calloc error\n");
	    return (-1);
	}
	Size += choice_size * sizeof(char *);

	chx = 0;
	/* expect choice sdrLine */
	FOREVER {
	    if ((getSdrLine(inFile_fp, plineIn, TWO_QUOTES)) < 0)
		return (-1);

	    if (plineIn->eof == EOF) {
		printf("bldGblChoice: Error - Unexpected EOF\n");
		printf("	%s\n", inFile);
		return (-1);
	    }
	    if ((sscanf(plineIn->line, "%li \"%n",
			&choice_num, &no_byts)) != 1) {
		printf("bldGblChoice: illegal field in file %s\n", inFile);
		printf("bldGblChoice: line=%i\n", plineIn->line_no);
		return (-1);
	    }
	    nextInputFld(plineIn, &no_byts);
	    if ((plineIn->pnext[0]) == '"') {
		choice_name[0] = '\0';
		no_byts = 1;
	    } else if ((sscanf(plineIn->pnext, "%[^\"]%n",
			       choice_name,&no_byts)) != 1) {
		printf("bldGblChoice: illegal field in file %s\n", inFile);
		printf("bldGblChoice: line=%i\n", plineIn->line_no);
		return (-1);
	    }
	    nextInputFld(plineIn, &no_byts);
#if 0
/* OLD */
	    if ((sscanf(plineIn->line, "%li \"%[^\"]",
			&choice_num, choice_name)) != 2) {
		printf("bldGblChoice: illegal field in file %s\n", inFile);
		printf("bldGblChoice: line=%i\n", plineIn->line_no);
		return (-1);
	    }
#endif
	    if (choice_num > choice_size) {
		printf("bldGblChoice: Error - choice_num > choice_size\n line=%s\n",
		       plineIn->line);
		return (-1);
	    }
	    len = strlen(choice_name);
	    SSize += len + 1;
	    /* allocate space for the choice_name string */
	    if ((pTchoiceGbl->papChoiceSet[inx_num]->papChoice[choice_num] = (char *)
		 calloc(1, (unsigned) len + 1)) == NULL) {
		printf("bldGblChoice: calloc error\n");
		return (-1);
	    }
	    strcpy(pTchoiceGbl->papChoiceSet[inx_num]->papChoice[choice_num],
		   choice_name);
	    chx++;
	    /* test for end of choice loop */
	    if (chx == choice_entries)
		break;

	}
	pTchoiceGbl->papChoiceSet[inx_num]->number = choice_size;

	/* check for match choice_entries */
	if (chx != choice_entries) {
	    printf("bldGblChoice: Error - choice_entries != actual entries \n");
	    return (-1);
	}
	inx++;
	/* test for end of index loop */
	if (inx == inx_entries)
	    break;
    }
    if ((getc(inFile_fp)) != EOF) {
	printf("bldGblChoice: Error - extra character(s) in file %s\n", inFile);
	return (-1);
    }
    /* check for match inx_entries */
    if (inx != inx_entries) {
	printf("bldGblChoice: Error - inx_entries != actual index entries \n");
	return (-1);
    }
    /* make sure we have padding to long word alignment for strings */
    SSize = ((SSize + sizeof(long) - 1) / sizeof(long)) * sizeof(long);
    Size += SSize;
    if ((choiceGbl = (struct arrChoiceSet *)
	 calloc(1, (unsigned) Size)) == NULL) {
	printf("bldGblChoice: calloc error\n");
	return (-1);
    }
    /* initialize addr1 to first byte past struct arrChoiceSet */
    addr1 = (char *) choiceGbl + sizeof(struct arrChoiceSet);

    choiceGbl->number = inx_size;
    choiceGbl->papChoiceSet = (struct choiceSet **) addr1;

    addr1 = (char *) addr1 + (inx_size * sizeof(char *));

    /* initialize addr2 to place the strings at the end of the stucture */
    addr2 = (char *) choiceGbl + Size - SSize;

    /* insert pointers and data */
    for (i = 0; i < choiceGbl->number; i++) {
	if (pTchoiceGbl->papChoiceSet[i]) {
	    choiceGbl->papChoiceSet[i] = (struct choiceSet *) addr1;
	    addr1 += sizeof(struct choiceSet);
	    choiceGbl->papChoiceSet[i]->papChoice = (char **) addr1;
	    choiceGbl->papChoiceSet[i]->number
		= pTchoiceGbl->papChoiceSet[i]->number;
	    addr1 += choiceGbl->papChoiceSet[i]->number * sizeof(char *);
	    for (j = 0; j < choiceGbl->papChoiceSet[i]->number; j++) {
		if (pTchoiceGbl->papChoiceSet[i]->papChoice[j]) {
		    choiceGbl->papChoiceSet[i]->papChoice[j] = (char *) addr2;
		    len = strlen(pTchoiceGbl->papChoiceSet[i]->papChoice[j]);
		    strcpy(choiceGbl->papChoiceSet[i]->papChoice[j],
			   pTchoiceGbl->papChoiceSet[i]->papChoice[j]);
		    addr2 += len + 1;
		}
	    }
	}
    }

    /* write output file */
    if ((outFile_fp = fopen((char *) outFile, "w")) == NULL) {
	printf("bldGblChoice: fopen errno=%i \nfile=%s\n",
	       errno, outFile);
	return (-1);
    }

    /* write the header */
    fprintf(outFile_fp, "choiceGbl\n");

    /* write the number of choice sets */
    fprintf(outFile_fp, "%d\n", choiceGbl -> number);

    /* write each choice set */
    for (i = 0; i < choiceGbl -> number; i++)
	{
	/* Write the number in this choice set and each choice */
	fprintf(outFile_fp, "%d\n", choiceGbl -> papChoiceSet[i] -> number);
	for (j = 0; j < choiceGbl -> papChoiceSet[i] -> number; j++)
	    if (choiceGbl -> papChoiceSet[i] -> papChoice[j])
		fprintf(outFile_fp, "%s\n", choiceGbl -> papChoiceSet[i] -> papChoice[j]);
	    else
		fprintf(outFile_fp, "\n");
	}
    fclose(outFile_fp);

    return (0);
}				/* end of bldGblChoice.c  */
/*
 * CMD_CHOICEGBL
 *
 * checks command line arg usage
 *
 * return (1,0) success, fail
 *
 */
cmd_choiceGbl(argc, argv)
    int             argc;
    char          **argv;
{
    extern char    *optarg;
    struct stat    *buf;
    int             c,
                    err;
    short           aflg,
                    oflg;
    aflg = oflg = err = 0;

    if ((buf = (struct stat *)
	 calloc(1, (unsigned) sizeof(struct stat))) == NULL) {
	printf("bldGblChoice: calloc error\n");
	return (0);
    }
    while ((c = getopt(argc, argv, "a:o:")) != -1)
	switch (c) {
	case 'a':
	    aflg++;
	    strcpy(inFile, optarg);
	    if (stat(inFile, buf) == -1) {
		printf("bldGblChoice: Can't stat file=%s errno=%i\n",
		       inFile, errno);
		err++;
	    }
	    break;
	case 'o':
	    oflg++;
	    strcpy(outFile, optarg);
	    break;
	}

    if (!aflg || !oflg || err) {
	printf("Usage: \n");
	printf("    <bldGblChoice -a choiceGbl.ascii -o choiceGbl.sdr>\n");
	return 0;
    }
    return 1;
}
/*
 *
 * CHK_ALIGN
 *
 * checks alignment of structure definitions
 *
 * return (0,-1) success, fail
 */
int
chk_align()
{
    if ((sizeof(struct arrChoiceSet) % 4)) {
	printf("bldGblChoice: struct arrChoiceSet size alignment error\n");
	return (-1);
    }
    if ((sizeof(struct choiceSet) % 4)) {
	printf("bldGblChoice: struct choiceSet size alignment error\n");
	return (-1);
    }
    return (0);
}
