/****************************************************************/
/*								*/
/*		CHANNEL ACCESS ALARM UTILITY ROUTINES		*/
/*								*/
/*	The following functions are utilities for the alarm	*/
/*	handler. They perform their services without the	*/
/*	need to establish connections.				*/
/*								*/
/*	ca_ack_alarm:		operator acknowledgement	*/
/*	ca_recv_alarm:		end-to-end receipt with IOC	*/
/*	ca_enable_alarm:	single point alarm enable	*/
/*	ca_disable_alarm:	single point alarm disable	*/
/*								*/
/*	Author:	Gordon Uchenick					*/
/*								*/
/****************************************************************/

#ifdef WIN32
#include <sys/types.h>
#include	<winsock.h>
int sendCAQ(short, char *, int);
#else
#include <types.h>
#include <in.h>
#endif WIN32

#ifdef AXP
#include <string.h>
#endif AXP

#include "tsDefs.h"
#include "dbDefs.h"
#include "agentMsgs.h"
#include "agentErrs.h"


/****************************************************************/
/*								*/
/*		OPERATOR ACKNOWLEDGEMENT			*/
/*								*/
/*	Calling this function indicates that a human operator	*/
/*	has seen the alarm and acknowledged it.			*/
/*								*/
/****************************************************************/

ca_ack_alarm(pvName, stamp)
char		*pvName;
TS_STAMP	*stamp;
{
    short	msgType;
    struct
	{
	CH_ALARM_ACK	ack;
	char		pv[PVNAME_SZ + FLDNAME_SZ + 2];
	} msg;

    msgType = htons(CHAN_ALARM_ACK);
    msg.ack.alarmTime.secPastEpoch = htonl(stamp -> secPastEpoch);
    msg.ack.alarmTime.nsec = htonl(stamp -> nsec);
    msg.ack.alarmNameLen = strlen(pvName) + 1;
    msg.ack.alarmNameLen = htons(msg.ack.alarmNameLen);
    strcpy(msg.pv, pvName);
    sendCAQ(msgType, (char *) &msg, sizeof(CH_ALARM_ACK) 
	    + ntohs(msg.ack.alarmNameLen));
    return (AGNT_NORMAL);
}


/****************************************************************/
/*								*/
/*		ALARM END-TO-END RECEIPT			*/
/*								*/
/*	The alarm manager calls this routine to indicate that	*/
/*	he has received the alarm and logged it to disk. It	*/
/*	does NOT NOT NOT NOT NOT mean that a human operator	*/
/*	has seen and acknowledged the alarm.			*/
/*								*/
/****************************************************************/

ca_recv_alarm(serial, stateLess, pvName, pStamp)
int		serial;
int		stateLess;
char		*pvName;
TS_STAMP	*pStamp;
{
    short		msgType;
    struct
	{
    	CH_ALARM_RECV	alarmRecv;
	char		pv[PVNAME_SZ + FLDNAME_SZ + 2];
	} msg;

    if (!serial)
	return (AGNT_NORMAL);
    msgType = htons(CHAN_ALARM_RECV);
    msg.alarmRecv.alarmSerial = serial;
    msg.alarmRecv.alarmHandle = stateLess;
    msg.alarmRecv.alarmTime.secPastEpoch = htonl(pStamp -> secPastEpoch);
    msg.alarmRecv.alarmTime.nsec = htonl(pStamp -> nsec);
    msg.alarmRecv.alarmNameLen = strlen(pvName) + 1;
    msg.alarmRecv.alarmNameLen = htons(msg.alarmRecv.alarmNameLen);
    strcpy(msg.pv, pvName);
    sendCAQ(msgType, (char *) &msg, sizeof(CH_ALARM_RECV) 
	    + ntohs(msg.alarmRecv.alarmNameLen));
    return (AGNT_NORMAL);
}


/****************************************************************/
/*								*/
/*		CONNECTIONLESS ALARM ENABLE			*/
/*								*/
/*	Calling this function will enable alarms for the	*/
/*	point without a need to build a traditional channel.	*/
/*								*/
/****************************************************************/

ca_enable_alarm(pvName)
char	*pvName;
{
    short	msgType;
    struct
	{
	CH_ALARM_CTL	ctl;
	char		pv[PVNAME_SZ + FLDNAME_SZ + 2];
	} msg;

    msgType = htons(CHAN_ALARM_CTL);
    msg.ctl.alarmCtlFunc = htons(1);
    msg.ctl.alarmNameLen = strlen(pvName) + 1;
    msg.ctl.alarmNameLen = htons(msg.ctl.alarmNameLen);
    strcpy(msg.pv, pvName);
    sendCAQ(msgType, (char *) &msg, sizeof(CH_ALARM_CTL)
	    + ntohs(msg.ctl.alarmNameLen));
    return (AGNT_NORMAL);
}


/****************************************************************/
/*								*/
/*		CONNECTIONLESS ALARM DISABLE			*/
/*								*/
/*	Calling this function will disable alarms for the	*/
/*	point with a need to build a traditional channel.	*/
/*								*/
/****************************************************************/

ca_disable_alarm(pvName)
char	*pvName;
{
    short	msgType;
    struct
	{
	CH_ALARM_CTL	ctl;
	char		pv[PVNAME_SZ + FLDNAME_SZ + 2];
	} msg;

    msgType = htons(CHAN_ALARM_CTL);
    msg.ctl.alarmCtlFunc = 0;
    msg.ctl.alarmNameLen = strlen(pvName) + 1;
    msg.ctl.alarmNameLen = htons(msg.ctl.alarmNameLen);
    strcpy(msg.pv, pvName);
    sendCAQ(msgType, (char *) &msg, sizeof(CH_ALARM_CTL)
	    + ntohs(msg.ctl.alarmNameLen));
    return (AGNT_NORMAL);
}
