/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		GET DISPLAY PARAMETERS FOR A CHANNEL		*/
/*								*/
/*	This routine is used to get information about a channel	*/
/*	in a connectionless manner. The information is just	*/
/*	about the same as the control variants of the fundamen-	*/
/*	tal data types. This method only works through the	*/
/*	callback mechanism. There can only be one active para-	*/
/*	meter callback function at a time.			*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#ifdef vms
#    include <selectvms.h>
#endif

#ifdef WIN32
#    include <sys/types.h>
#    include <winsock.h>
#else
#    include <types.h>
#    include <in.h>
#endif WIN32

#ifdef AXP
#    include <string.h>
#endif

#include "caContext.h"
#include "agentErrs.h"
#include "agentMsgs.h"
#include "dbDefs.h"
#include "paramStructs.h"

ca_get_params (
    char	*chanName,	/* Channel Name */
    short	type,		/* DBR_xxx_PARAM (see paramStructs.h) */
    void	(*paramFunc)(),	/* Callback routine */
    void	*pArg		/* Argument for the callback */
)
{
    short		msgType;
    short		nameLen;
    CATX		*ctx;
    struct
	{
	CH_GETPARAMS	getParams;
	char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
	} msg;

    /*
    Get our context pointer and save the callback function. This
    service is not offered for the vxWorks platform, so we don't
    have to worry about importing one context on top of another.
    */

    ctx = getContext();
    ctx -> paramDataFunc = paramFunc;

    /*
    Now form a message requesting the paramters for the channel.
    */

    msg.getParams.chParmType = htons(type);
#ifndef POINTER64
    msg.getParams.chArgument = (int) pArg;
#else
    msg.getParams.chArgument = TokenFromPointer64 (pArg);
#endif
    nameLen = strlen(chanName) + 1;
    if (nameLen >= (PVNAME_SZ + FLDNAME_SZ + 2))
	return (AGNT_STRTOBIG);
    strcpy(msg.chanName, chanName);
    msg.getParams.chNameLen = htons(nameLen);
    msgType = htons(CHAN_GETPARAMS);
    sendCAQ(msgType, (char *) &msg, sizeof(CH_GETPARAMS) + nameLen);
    return (AGNT_NORMAL);
}
