/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	MANAGE FREE POOL OF CURRENT VALUE DATA STRUCTS		*/
/*								*/
/*	These routines manage a series of pools of structs	*/
/*	so that memory can get reused as necessary.		*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#ifdef WIN32
#include <winsock.h>
#else
#include <sys/time.h>
#include <sys/types.h>
#include <netinet/in.h>
#endif WIN32
#include "db_access.h"
#include "channelControl.h"
#include "caValues.h"
#include "currValues.h"
#include "cvtMgt.h"

static int 	dataSizes[] =
    {
    sizeof(VALUE_STRING),	/* AGNT_STRING */
    sizeof(VALUE_SHORT),	/* AGNT_SHORT or AGNT_INT */
    sizeof(VALUE_FLOAT),	/* AGNT_FLOAT */
    sizeof(VALUE_ENUM),		/* AGNT_ENUM */
    sizeof(VALUE_CHAR),		/* AGNT_CHAR */
    sizeof(VALUE_LONG),		/* AGNT_LONG */
    sizeof(VALUE_DOUBLE)	/* AGNT_DOUBLE */
    };

#define	VALUE_ALLOC	128

static CALINK	*valuePool[sizeof(dataSizes) / sizeof(dataSizes[0])];


/****************************************************************/
/*								*/
/*	ALLOCATE A CURRENT VALUE BLOCK OF A GIVEN TYPE		*/
/*								*/
/****************************************************************/

CALINK *newValue(type, count)
int	type;
int	count;
{
    int		index;
    int		size;
    char	*data;
    CALINK	*currPool;

    size = dataSizes[type];

    /*
    Arrays are a special case. We allocate them when requested and
    don't keep pools.
    */

    if (count != 1)
	{
	currPool = (CALINK *) malloc(sizeof(CALINK));
	currPool -> pData = (void *) malloc(((size - sizeof(VALUE_COMMON)) * count) + sizeof(VALUE_COMMON));
	currPool -> pCHB = 0;
	return (currPool);
	}

    /*
    If the pool is empty, fill it up
    */

    if (!valuePool[type])
	{
	valuePool[type] = (CALINK *) malloc(sizeof(CALINK) * VALUE_ALLOC);
	if (!valuePool[type])
	    return (0);
	currPool = valuePool[type];
	data = (char *) malloc(size * VALUE_ALLOC);
	index = VALUE_ALLOC;
	while (index--)
	    {
	    if (index)
		currPool -> pCHB = (CHB *)(currPool + 1);
	    else
		currPool -> pCHB = 0;
	    currPool -> pData = (void *) data;
	    data += size;
	    }
	}
    
    /*
    Take one off the pool and clear it.
    */

    currPool = valuePool[type];
    valuePool[type] = (CALINK *)(currPool -> pCHB);
    currPool -> pCHB = 0;
    memset(currPool -> pData, 0, size);
    return (currPool);
}


void deleteValue(oldPool, type)
CALINK	*oldPool;
int	type;
{
    /*
    If this entry is for an array, free it immediately. Otherwise, we keep pools
    of value entries.
    */

    if (((VALUE_COMMON *)(oldPool -> pData)) -> nativeCount != 1)
	{
	free(oldPool -> pData);
	free(oldPool);
	return;
	}

    /*
    Push this new one on the front of the appropriate pool.
    */

    oldPool -> pCHB = (CHB *)valuePool[type];
    valuePool[type] = oldPool;
}
