/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	IOC TUNING PARAMETER EDITOR UTILITY PROGRAM		*/
/*								*/
/*	This is a crude (i.e., character based) program for	*/
/*	tuning the parameters that control how sockets are	*/
/*	maintained in the sockMnt component of channel access.	*/
/*	This program allows you to tune any node from any	*/
/*	other node.						*/
/*								*/
/****************************************************************/


#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <agentMsgs.h>
#include <agentPorts.h>
#include <sockTune.h>


main()
{
    char		myNodeName[100];
    char		targNodeName[100];
    char		*eofDetect;
    struct hostent	*hostent;
    struct sockaddr_in	sockAddr;
    SM_PARM_REPLY	smParmReply;
    sockTune		*tuneList;
    int			sockFd;
    short		msgType;
    int			i;
    int			modified;
    int			lineNum;


    gethostname(myNodeName, sizeof(myNodeName));
    for (;;)
	{

	/*
	First, get the name of the node that we are about to tune.
	A blank means this node. The strings "quit" or "exit" mean
	to do that.
	*/

	sockFd = 0;
	while (!sockFd)
	    {
	    printf("Node name? <%s>   ", myNodeName);
	    eofDetect = gets(targNodeName);
	    if (!eofDetect)
		exit(1);
	    if (!targNodeName[0])
		strcpy(targNodeName, myNodeName);
	    else
		{
		if (!strcmp(targNodeName, "quit"))
		    exit(1);
		if (!strcmp(targNodeName, "exit"))
		    exit(1);
		}

	    /*
	    Try to connect to the sockMnt on the node that was given.
	    */

	    hostent = gethostbyname(targNodeName);
	    if (!hostent)
		{
		fprintf(stderr, "Node %s unknown.\n", targNodeName);
		continue;
		}
	    sockAddr.sin_family = AF_INET;
	    sockAddr.sin_addr.s_addr = **((long **)(hostent -> h_addr_list));
	    sockAddr.sin_port = htons(SCM_PORT);
	    if ((sockFd = socket(PF_INET, SOCK_STREAM, 0)) == -1)
		{
		perror("iocTuna");
		sockFd = 0;
		continue;
		}
	    if (connect(sockFd, (struct sockaddr *) &sockAddr, sizeof(sockAddr)) == -1)
		{
		perror("iocTuna");
		close(sockFd);
		sockFd = 0;
		continue;
		}
	    break;
	    }

	/*
	Send the message to get the current list of tuning parameters out of sockMnt.
	*/

	msgType = htons(SOCK_PARM);
	if (send(sockFd, (char *) &msgType, sizeof(msgType), 0) != sizeof(msgType))
	    {
	    perror("iocTuna");
	    close(sockFd);
	    sockFd = 0;
	    continue;
	    }

	/*
	Get the reply. Because this is a crude, character based implementation, we don't
	have to bother protecting ourselves with selects in case of timeouts. Our limit
	for timeout is the user's patience, and the corrective action is his Control-C.
	*/

	if (recvTcp(sockFd, (char *) &msgType, sizeof(msgType)))
	    {
	    sockFd = 0;
	    continue;
	    }
	if (ntohs(msgType) != SOCK_PARM_REPLY)
	    {
	    fprintf(stderr, "iocTuna: Got invalid message type instead of SOCK_PARM_REPLY.\n");
	    close(sockFd);
	    sockFd = 0;
	    continue;
	    }
	if (recvTcp(sockFd, (char *) &smParmReply, sizeof(smParmReply)))
	    {
	    sockFd = 0;
	    continue;
	    }
	smParmReply.smNumSockets = ntohs(smParmReply.smNumSockets);
	tuneList = (sockTune *) malloc(smParmReply.smNumSockets * sizeof(sockTune));
	printf("\tInet Addr\tMax Idle\tMax Quiet\tMax Block\tInq Rate\n");
	for (i = 0; i < smParmReply.smNumSockets; i++)
	    {
	    if (recvTcp(sockFd, (char *)(tuneList + i), sizeof(sockTune)))
		{
		close(sockFd);
		sockFd = 0;
		continue;
		}
	    if (!tuneList[i].stInet)
		strcpy(targNodeName, "<default>");
	    else
		{
		hostent = gethostbyaddr(&tuneList[i].stInet, sizeof(tuneList[i].stInet), AF_INET);
		strcpy(targNodeName, hostent -> h_name);
		trim(targNodeName);
		}
	    printf("%d:\t%-15s\t%6d\t\t%6d\t\t%6d\t\t%6d\n", i + 1, targNodeName,
		    tuneList[i].stMaxIdle, tuneList[i].stMaxQuiet, tuneList[i].stMaxBlock,
		    tuneList[i].stInqRate);
	    }

	/*
	Now let him edit what he wants to.
	*/

	modified = 0;
	printf("\n");
	for (;;)
	    {
	    printf("Function <E = Edit, W = Write, P = Print, Q = Quit, H = Help>?   ");
	    eofDetect = gets(targNodeName);
	    if ((targNodeName[0] == 'E') || (targNodeName[0] == 'e'))
		{
		do
		    {
		    printf("Line number <1 - %d>?   ", smParmReply.smNumSockets);
		    gets(targNodeName);
		    sscanf(targNodeName, "%d", &lineNum);
		    } while ((lineNum < 1) || (lineNum > smParmReply.smNumSockets));
		lineNum--;
	    	if (!tuneList[lineNum].stInet)
		    strcpy(targNodeName, "<default>");
	    	else
		    {
		    hostent = gethostbyaddr(&tuneList[lineNum].stInet, sizeof(tuneList[lineNum].stInet), AF_INET);
		    strcpy(targNodeName, hostent -> h_name);
		    trim(targNodeName);
		    }
		printf("Editing %s\n", targNodeName);
		do
		    {
		    printf("Max Idle?   ");
		    gets(targNodeName);
		    sscanf(targNodeName, "%d", &tuneList[lineNum].stMaxIdle);
		    } while (tuneList[lineNum].stMaxIdle <= 0);
		do
		    {
		    printf("Max Quiet?   ");
		    gets(targNodeName);
		    sscanf(targNodeName, "%d", &tuneList[lineNum].stMaxQuiet);
		    } while (tuneList[lineNum].stMaxQuiet <= 0);
		do
		    {
		    printf("Max Block?   ");
		    gets(targNodeName);
		    sscanf(targNodeName, "%d", &tuneList[lineNum].stMaxBlock);
		    } while ((tuneList[lineNum].stMaxBlock <= 0) 
			    || (tuneList[lineNum].stMaxBlock > 4096));
		do
		    {
		    printf("Inq Rate?   ");
		    gets(targNodeName);
		    sscanf(targNodeName, "%d", &tuneList[lineNum].stInqRate);
		    } while (tuneList[lineNum].stInqRate <= 0);
		modified = 1;
		}
	    else if ((targNodeName[0] == 'W') || (targNodeName[0] == 'w'))
		{
		SM_TUNE		tuneHdr;
		msgType = htons(SOCK_TUNE);

		if (send(sockFd, (char *) &msgType, sizeof(msgType), 0) != sizeof(msgType))
	    	    {
	    	    perror("iocTuna");
	    	    close(sockFd);
	    	    sockFd = 0;
	    	    break;
	    	    }
		tuneHdr.smNumSockets = htons(smParmReply.smNumSockets);
		if (send(sockFd, (char *) &tuneHdr, sizeof(tuneHdr), 0) != sizeof(tuneHdr))
	    	    {
	    	    perror("iocTuna");
	    	    close(sockFd);
	    	    sockFd = 0;
	    	    break;
	    	    }
		for (i = 0; i < smParmReply.smNumSockets; i++)
		    {
		    if (send(sockFd, (char *) (tuneList + i), sizeof(sockTune), 0) != sizeof(sockTune))
	    	    	{
	    	    	perror("iocTuna");
	    	    	close(sockFd);
	    	    	sockFd = 0;
	    	    	break;
	    	    	}
		    }
		modified = 0;
		}
	    else if ((targNodeName[0] == 'P') || (targNodeName[0] == 'p'))
		{
		printf("\tInet Addr\tMax Idle\tMax Quiet\tMax Block\tInq Rate\n");
		for (i = 0; i < smParmReply.smNumSockets; i++)
		    {
		    if (!tuneList[i].stInet)
			strcpy(targNodeName, "<default>");
		    else
			{
			hostent = gethostbyaddr(&tuneList[i].stInet, sizeof(tuneList[i].stInet), AF_INET);
			strcpy(targNodeName, hostent -> h_name);
			trim(targNodeName);
			}
		    printf("%d:\t%s\t%6d\t\t%6d\t\t%6d\t\t%6d\n", i + 1, targNodeName,
			    tuneList[i].stMaxIdle, tuneList[i].stMaxQuiet, tuneList[i].stMaxBlock,
			    tuneList[i].stInqRate);
		    }
		}
	    else if ((targNodeName[0] == 'Q') || (targNodeName[0] == 'q'))
		{
		if (modified)
		    {
		    printf("Data was modified! Really quit without saving?   ");
		    gets(targNodeName);
		    if ((targNodeName[0] == 'Y') || (targNodeName[0] == 'y'))
			{
			if (sockFd)
			    {
			    close(sockFd);
			    sockFd = 0;
			    modified = 0;
			    if (tuneList)
				{
				free(tuneList);
				tuneList = 0;
				}
			    break;
			    }
			}
		    else
		        continue;
		    }
		else
		    {
		    if (sockFd)
			{
			close(sockFd);
			sockFd = 0;
			}
	  	    modified = 0;
		    if (tuneList)
			{
		    	free(tuneList);
			tuneList = 0;
			}
		    break;
		    }
		}
	    else if ((targNodeName[0] == 'H') || (targNodeName[0] == 'h'))
		{
		printf("\n");
		printf("To edit, specify the line number of the node you want to select.\n");
		printf("Parameters have the following meanings:\n");
		printf("\tMax Idle = The amount of time that a line is idle before\n");
		printf("\t\tsending a boomerang, in seconds.\n");
		printf("\tMax Quiet = The number of consecutive boomerangs that have to\n");
		printf("\t\tbe missed for an IOC to be marked as offline.\n");
		printf("\tMax Block = Biggest single logical transmission block, max of %d\n",
				MAX_XMIT_BUFFER);
		printf("\tInq Rate = The number of seconds between UDP invitations to connect.\n");
		printf("\n");
		printf("Write will send parameters back to the originating node for immediate use\n");
		printf("and storage. Print redisplays the list just in case it has scrolled off\n");
		printf("the top of the screen. Quit lets you select the next node to edit.\n");
		printf("\n");
		}
	    }
	}
}

/****************************************************************/
/*								*/
/*	READ BYTES FROM TCP SOCKET UNTIL NBYTES ARE RECVD	*/
/*								*/
/*	No timeout facility is provided. Return a 0 if OK,	*/
/*	a one if there was an error reading the socket.		*/
/*	If there was an error, print an error message and	*/
/*	close the socket.					*/
/*								*/
/****************************************************************/

recvTcp(fd, buff, nbytes)
int	fd;
char	*buff;
int	nbytes;
{
    int		bytesGot;

    while (nbytes)
	{
	bytesGot = recv(fd, buff, nbytes, 0);
	if (bytesGot <= 0)
	    {
	    perror("iocTuna");
	    close(fd);
	    return (1);
	    }
	nbytes -= bytesGot;
	buff += bytesGot;
	}
    return (0);
}

/****************************************************************/
/*								*/
/*	TRIM OUT THE FIRST DOT IN AN INTERNET NAME 		*/
/*								*/
/****************************************************************/

trim(name)
char	*name;
{
    while (*name)
	if (*name == '.')
	    {
	    *name = 0;
	    return(0);
	    }
	else
	    name++;
}
