/*recvCaq.c,v 1.1.1.1 1999/08/03 14:42:04 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	RECEIVE A MESSAGE FROM CHANNEL ACQUISITION		*/
/*								*/
/*	These two routines take care of reading data from our	*/
/*	socket connected to Channel Acquisition and Control.	*/
/*	Every time we read that socket, we attempt to get as	*/
/*	as much data as we can pull out of it. The routine	*/
/*	that parses the input message can then extract bits	*/
/*	and pieces as necessary. A boundary condition can	*/
/*	exist where we may have to read more data than our	*/
/*	original "fill up". This happens when a message gets	*/
/*	split up in the stream. This isn't like to happen,	*/
/*	however, because of the output buffering algorithm	*/
/*	in Channel Acquisition.					*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#define FALSE	0

#ifdef vms
#include <selectvms.h>
#include <ucx$inetdef.h>
#endif vms
#include <bcopy.h>
#include <stdio.h>
#ifdef WIN32
#include <sys/types.h>
#include <winsock.h>
#else
#include <types.h>
#include <socket.h>
#endif WIN32
#include <cadef.h>
#include <caContext.h>
#ifdef vxWorks
#include <selectLib.h>
#include <sockLib.h>
#include <ioLib.h>
#include <sys/times.h>
#else
#ifdef AXP
#ifndef vms
#include <sys/time.h>
#endif
#else
#include <time.h>
#if !(defined(vms) || defined(WIN32))
#include <sys/time.h>
#endif
#endif
#if (defined(ultrix) || defined(AXP))
#include <ioctl.h>
#else
#if !(defined(vms) || defined(WIN32))
#include <filio.h>
#endif vms
#endif ultrix
#endif vxWorks

static struct timeval	maxWait = {5, 0};	/* Five Seconds */

/******
	READ A BIG CHUNK OF DATA OUT OF THE SOCKET

	Return zero on timeout, zero length read, or recv error.
******/

int getMore()
{
    int			bytesGot;
    fd_set		ioMask;
    CATX		*ctx;


    /*
    Get our context
    */

    ctx = getContext();

    /*
    A select is used here on second and subsequent reads to make sure
    that we won't get hung if Channel Acquisition crashes.
    */

    if (ctx -> firstSelDone)
	{
    	ioMask = ctx -> caSocketMask;
	if (select(ctx -> caSocketFd + 1, &ioMask, NULL, NULL, &maxWait) <= 0)
	    {
#ifdef WIN32
	    closesocket(ctx -> caSocketFd);
#else
	    close(ctx -> caSocketFd);
#endif WIN32
	    if (ctx -> regFunc)
		(*(ctx -> regFunc))(ctx -> regFuncArg, ctx -> caSocketFd, FALSE);
	    ctx -> caSocketFd = 0;
	    return (0);
	    }
	}

    /*
    If there is already stuff in the receive buffer, slide it down
    to the start if it's not there already.
    */

    if (ctx -> nchib && ctx -> bufOutPos)
	bcopy(ctx -> caDataBuff + ctx -> bufOutPos, ctx -> caDataBuff, ctx -> nchib);

    /*
    Make sure that our receive buffer is large enough. We don't want to have chanAcq held
    up when he has stuff to send to us.
    */

    if (!(ctx -> caRxTopped))
	{
#ifdef WIN32
        ioctlsocket (ctx -> caSocketFd, FIONREAD, &bytesGot);
#else
	ioctl(ctx -> caSocketFd, FIONREAD, (int)&bytesGot);
#endif WIN32
	if (bytesGot == ctx -> caRxBuff)
	    {
	    ctx -> caRxBuff += 1024;
	    if (setsockopt(ctx -> caSocketFd, SOL_SOCKET, SO_RCVBUF, (char *)&(ctx -> caRxBuff), 
		    sizeof(ctx -> caRxBuff)) == -1)
		ctx -> caRxTopped = 1;
	    }
	}

    /*
    Read as much data into the buffer as is available and will fit.
    */

    if ((bytesGot = recv(ctx -> caSocketFd, ctx -> caDataBuff + ctx -> nchib, 
	    sizeof(ctx -> caDataBuff) - ctx -> nchib, 0)) <= 0)
    	{
#ifdef WIN32
	closesocket(ctx -> caSocketFd);
#else
	close(ctx -> caSocketFd);
#endif WIN32
    	if (ctx -> regFunc)
		(*(ctx -> regFunc))(ctx -> regFuncArg, ctx -> caSocketFd, FALSE);
    	ctx -> caSocketFd = 0;
    	return (0);
    	}

    ctx -> nchib += bytesGot;
    ctx -> bufOutPos = 0;

    return (1);
}


/******
	Extract data from the receive buffer

	Return a zero on error
******/

int recvCAQ(buff, nbytes)
char	*buff;
int	nbytes;
{
    CATX	*ctx;

    /*
    Get our context.
    */

    ctx = getContext();
    
    /*
    If we don't have enough data to satisfy the request, try to get more.
    */

    while (nbytes > ctx -> nchib)
	if (!getMore())
	    return (0);

    /*
    Extract the data, keeping track of where we are.
    */

    bcopy(ctx -> caDataBuff + ctx -> bufOutPos, buff, nbytes);
    ctx -> bufOutPos += nbytes;
    ctx -> nchib -= nbytes;
    return (1);
}
