/*serverInit.c,v 1.1.1.1 1999/08/03 14:42:04 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	COMMON INTIALIZATION FOR CHANNEL ACCESS SERVERS		*/
/*								*/
/*	Many of the things that we do for the three servers	*/
/*	that make up the agent architecture for channel access	*/
/*	are common. This routine avoids code duplication.	*/
/*								*/
/*	Author:	Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <fcntl.h>
#ifdef WIN32
#include <io.h>
#include <sys/locking.h>
#include <process.h>
#include <string.h>
#include <stdio.h>
#include <direct.h>
#else
#include <unistd.h>
#include <file.h>
#include <termios.h>
#endif WIN32
#include <signal.h>
#if defined(AXP)
#include <string.h>
#include <ioctl.h>
#endif

void serverInit(dirPath, lockFile)
char	*dirPath;
char	*lockFile;
{
    int		fd;
    char	buff[10];

/******
	Change our controlling TTY so that we don't pick up any
	stray signals.
******/

#if !(defined(AXP) || defined(WIN32))
    fd = open("/dev/tty", O_RDWR);
    ioctl(fd, TIOCNOTTY, 0);
    close(fd);
#endif

/******
	Move to the defined directory so that the core file
	can be found in case we die an unnatural death.
******/

#ifdef WIN32
    _chdir(dirPath);
#else
    chdir(dirPath);
#endif WIN32
/******
	Make a process group just for ourselves
******/

#ifdef SOLARIS
    setpgrp ();
#else
#ifdef AXP
    setpgrp ();
    setpgid (0,0);
#else
#ifndef WIN32
    setpgrp(getpid (), 0);
#endif WIN32
#endif AXP
#endif SOLARIS

/******
	Insure that we are the only copy of the server running
******/

#ifdef WIN32
    fd = _open(lockFile, _O_RDWR | _O_CREAT, 0640);
    if (fd < 0)
	exit(1);
    if (_locking (fd, _LK_LOCK, 0) == -1)
	exit (0);

/******
	Record our PID in the lock file to make it easy to find
******/

    sprintf(buff, "%d\n", _getpid());
    _write(fd, buff, strlen(buff));
    _close(fd);		/* I think that we keep the lock even if
			   the file is closed. */
#else
    fd = open(lockFile, O_RDWR | O_CREAT, 0640);
    if (fd < 0)
	exit(1);
#ifdef SOLARIS
    if (lockf (fd, F_TLOCK, 0) == -1)
	exit (0);
#else
    if (flock(fd, LOCK_EX | LOCK_NB))
	exit(0);
#endif

/******
	Record our PID in the lock file to make it easy to find
******/

    sprintf(buff, "%d\n", getpid());
    write(fd, buff, strlen(buff));
    close(fd);		/* I think that we keep the lock even if
			   the file is closed. *.

/******
	Finally, set up to ignore all the signals that we aren't
	handling.
******/

    signal(SIGPIPE, SIG_IGN);
#endif WIN32
}
