/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		SOCKET CREATION AND MAINTENANCE			*/
/*								*/
/*	This program handles the socket I/O for the Channel	*/
/*	Acquisition and Maintenance program.			*/
/*								*/
/*	Endian issues: We may be a different endian from the	*/
/*	IOC and also from the name server. We will always be	*/
/*	the same endian as Channel Acquisition and Control.	*/
/*	We may also be a different endian from the socket	*/
/*	tuning utility.						*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#ifndef vxWorks
#include <bcopy.h>
#endif vxWorks

#include <stdio.h>

#ifdef WIN32
#include <winsock.h>
#include <fcntl.h>
#include <errno.h>
#else WIN32
#ifdef vxWorks
#include <vxWorks.h>
#include <netinet/tcp.h>
#include <hostLib.h>
#include <tickLib.h>
#include <ioLib.h>
#include <inetLib.h>
#include <taskLib.h>
#include <selectLib.h>
#include <sys/times.h>
#include <sysLib.h>
#include <sockLib.h>
#include <string.h>
#include <memLib.h>
#include <fioLib.h>
#include <logLib.h>
#else vxWorks
#include <types.h>
#include <socket.h>
#include <in.h>
#include <tcp.h>
#include <netdb.h>
#include <syslog.h>
#ifdef AXP
#include <stdlib.h>
#else AXP
#include <sys/time.h>
#endif AXP
#ifdef SOLARIS
#include <stdlib.h>
#include <fcntl.h>
#include <filio.h>
#endif SOLARIS
#endif vxWorks
#include <errno.h>
#endif WIN32

#include <time.h>

#ifdef WIN32
#include <io.h>
#include "NTUnixSubs.h"
typedef void *caddr_t;
#else
#define SOCKET	int
#define INVALID_SOCKET (-1)
#endif WIN32

#include "inet_ntoa.h"
#include "pathNames.h"
#include "agentPorts.h"
#include "agentMsgs.h"
#include "sockTune.h"
#include "msgMon.h"

#ifdef vxWorks
#include "restable.h"
#endif vxWorks

#ifndef vxWorks
#define	ERROR	-1
#endif vxWorks

#ifdef PRIV_AUDIT
#include "inet_ntoa.h"
#include "sslShared.h"

/* for the auditing and priviledge functions */
#include "iocPriv.h"
#include "auditSystem.h"
extern AUDIT_ID	auditLog;	/* the system log */

static char	X509_dn [X509_MAX_DN]; /* distinguished name from certificate */
#endif PRIV_AUDIT

/****************************************************************/
/*								*/
/*		GLOBAL DATA FOR THIS PROCESS			*/
/*								*/
/****************************************************************/
#define MAX_UTIL_CLIENTS 100


static int		needNameSrv = 0;	/* Are we searching for nameSrv? */
struct in_addr		namAddrs[10];		/* Inet addresses where we can find nameSrv */
int			numNames = 0;		/* Number addresses defined */
#ifndef vxWorks
int			charsPerSecond;		/* For simulation only */
char			scmMobyBuff[512*1024];	/* Moby buffer when things back up! */
int			scmRingIn = 0;		/* Input index in ring buffer */
int			scmRingOut = 0;		/* Output index in ring buffer */
int			scmExtBytes;		/* # Bytes in scmMobyBuff */
fd_set			scmMask;		/* Reference copy, not to be modified */
fd_set			scmWriteFds;		/* Mask for "ready to write" */
fd_set			*scmWriteReady;		/* Pointer to scmWriteFd or NULL */
#ifndef vms
int			cacPid;			/* chanAcq's process ID */
static struct in_addr	traceInet = {0};	/* Inet addr for tracing */
static int		traceBytes = 0;
FILE			*traceFd;
#endif vms
#endif vxWorks

#ifdef USE_UDP
SOCKET			udpFd = INVALID_SOCKET;	/* UDP Socket */
#else USE_UDP
static void kick_TCP (long hostaddr /* in net order */,
		      int port); /* in host order */
#endif USE_UDP
SOCKET			smTcpFd = INVALID_SOCKET;/* TCP Socket */
SOCKET			cacFd = INVALID_SOCKET;	/* TCP Socket to Channel Control */
fd_set			fdMask;			/* For select */
static int		maxFd = 0;		/* Max fd for select */
int			utilityMax;
SOCKET			utilitySockets [MAX_UTIL_CLIENTS];
struct timeval		oneSecond = {5, 0};	/* Select timeout */
struct timeval		pollTimeout = {0, 0};		/* Just looking! */
struct dpaCtl
    {
    char		prefix[16];
    struct in_addr	realInet1;
    struct in_addr	virtInet1;
    SOCKET		sockFd1;
    struct in_addr	realInet2;
    struct in_addr	virtInet2;
    SOCKET		sockFd2;
    int			wanted;
    struct dpaCtl	*nextDPA;
    } *dpaList = 0;

struct sockCtl					/* Our internal control table layout */
    {
    struct sockCtl	*nextSock;		/* Link to next one */
    struct in_addr	Inet;			/* Inet address */
    SOCKET		fd;			/* File descriptor, INVALID_SOCKET = not connected */
    int			lastRx;			/* Last rx time */
    int			lastTx;			/* Last tx time */
    int			lastConn;		/* Last connection attempt time */
    int			bufSize;		/* Transmit buffer size */
    char		*bufLocn;		/* Buffer Location */
    char		*rxTank;		/* A receive buffer */
    int			nchInTank;		/* # bytes in rxTank */
    int			tankOutInd;		/* Output index in rxTank */
    int			currRxBuff;		/* Current Rx Buffer Size */
    short		rxMaxOut;		/* Receive buffer maxed out */
    short		iocDownSent;		/* IOC Down already sent? */
    int			currTxBuff;		/* Current Tx Buffer Size */
    int			boomsSent;		/* Number boomerangs sent */
    int			nextBoom;		/* Time next boomerang should go out */
    sockTune		tuning;			/* Tuning parameters */
    };
struct sockCtl		*socketList = 0;	/* Addr head of list */
int			numIOCs;		/* Number of IOCs we know about */
int			numDPAs;		/* Number of DPAs we know about */
long			lastTable = 0;		/* Last time we traversed table */
long			lastAlq = 0;		/* Last time we sent ALQ message */
long			lastDPA = 0;
static char		rxBuff[MAX_XMIT_BUFFER];
static char		txBuff[MAX_XMIT_BUFFER];
int			txBuffSize = 0;		/* Num chars in txBuff */
sockTune		defTuning;		/* Default socket tuning */
extern int		errno;
char			rxTank[MAX_XMIT_BUFFER];	/* Used for all reads */
static int		nchib;			/* # bytes in rxTank */
int			tankPos;		/* Next output index from rxTank */
int			cacRxSize;		/* Current receive buffer size of cacFd */
int			cacTxSize;		/* Current send buffer size of cacFd */
int			cacRxTopped;		/* Have we maxed out cacRxSize? */
int 			cacTxTopped;		/* Have we maxed out cacTxSize? */

static int		smDebug = 0;
#if defined(vxWorks) || defined(vms)
static int		dpaListBuilt = 0;
#endif

/****************************************************************/
/*								*/
/*			FUNCTION PROTOTYPES			*/
/*								*/
/****************************************************************/

struct sockCtl		*findFd (SOCKET fd);
int			knownIOC (struct in_addr targetAddr);
struct sockCtl		*findInet (struct in_addr targetAddr);
void 			sendCac (char *buff, int nbytes);
void 			recvCac (char *buff, int nbytes);
void 			sendIOC (struct sockCtl *ctlEntry);
void			iocDown (struct sockCtl *ctlEntry);
int			recvIOC (struct sockCtl *ctlEntry, char *buff, int length);
int			recvUtil (SOCKET fd, char *buff, int nbytes);
int			sendUtil (SOCKET fd, char *buff, int nbytes);
void			closeUtil (SOCKET sock);
void			rxFlowControl (struct sockCtl *ctlEntry);
void			cacFlush (void);
void			cacFlowControl (void);
int			fillChanBuffer (void);
void			scmMobySend (void);
int			haveCompleteIocMsg (struct sockCtl *ctlEntry);
int			getIocData (struct sockCtl *ctlEntry);
int			haveCompleteChanMsg (void);
int			sockDebug (int newLevel);
int			getDpaList (SOCKET fd);
void			dumpBlock (char *direction, unsigned char *buff, int nbytes);
int			hdlcAddress (void);

struct in_addr		caLocalInet;

/* external routine (in localAddr.c) */
#ifdef WIN32
extern int local_addr (SOCKET fd, struct sockaddr_in *pAddr);
#else
extern int local_addr (int fd, struct sockaddr_in *pAddr);
#endif WIN32

/****************************************************************/
/*								*/
/*			THE MAIN PROGRAM			*/
/*								*/
/****************************************************************/

#ifdef vxWorks
int sockMnt()
#else
int main()
#endif vxWorks
{
    struct sockaddr_in	portAddr;
    int			optVal = 1;
#ifndef vxWorks
    struct hostent	*hostent;
    NS_IOCL_REPLY	inetListHdr;
#endif vxWorks
    struct sockCtl	*ctlEntry;
    struct dpaCtl	*dpaEntry;
    int			true = 1;
    int			tuningFd;
    int			fdIndex;
    int			portAddrSize;
    short		msgType;
    sockTune		tempTune;
    fd_set		fdList;
    int			anyInputs;
    int			intsize;

/******
	Initialize
******/
#ifdef WIN32
    WSADATA wsaData; 
    WORD    wVersionRequested = MAKEWORD(1, 1);
    int	    err;

    err = WSAStartup(wVersionRequested, &wsaData); 
 
    if (err != 0) 
	return 0; 
#endif WIN32



#ifndef vxWorks
#ifndef vms
    {
    char		*envp;

    if (envp = getenv("sockTrace"))
	{
    	if (hostent = gethostbyname(envp))
	    {
    	    traceInet = *((struct in_addr *)hostent -> h_addr_list[0]);
#ifdef WIN32
	    traceFd = fopen("\\tis4000db\\sockTrace.log", "w");
#else
	    traceFd = fopen("/tis4000db/sockTrace.log", "w");
#endif WIN32
	    }
	}

    if (envp = getenv("sockBaud"))
	charsPerSecond = atoi(envp) / 8;
    }

#ifdef PRIV_AUDIT
    if (sslSharedInit () == 0)
    {
#if !defined(vxWorks) && !defined(vms)
	tisSysLog(0, 0, "sockMnt can't prepare SSL context!");
	exit(1);
#else
	perror("sockMnt can't prepare SSL context!\n");
	taskSuspend (taskIdSelf ());
#endif vxWorks
    }

    /* open default audit trail file */
    if (InitSystemLog (NULL, 0) == NULL)
    {
#if !defined(vxWorks) && !defined(vms)
	tisSysLog(0, 0, "sockMnt can't open default system audit file %s!\nNo auditing will occur\n", systemLogFile);
#else
	perror("sockMnt can't open system audit file %s!\nNo auditing will occur\n", systemLogFile);
#endif vxWorks
    }
#endif PRIV_AUDIT

    tisOpenLog("sockMnt", 2, 2, NULL, NULL);
    serverInit(SMPATH, SMLOCK);
    {
    FILE	*gotcha;
    if ((gotcha = fopen("/tis4000/tasks/caC/chanAcq.lock", "r")) == NULL)
	tisSysLog(0, 0, "sockMnt Can't open /tis4000/tasks/caC/chanAcq.lock: %m");
    else
	{
	fscanf(gotcha, "%d", &cacPid);
	fclose(gotcha);
	}
    }
#endif vms
#endif vxWorks

	/* Set all utility socket entries to INVALID_SOCKET */
	for (fdIndex = 0; fdIndex < MAX_UTIL_CLIENTS; fdIndex++)
	{
	    utilitySockets [fdIndex] = INVALID_SOCKET;
	}

/******
	Get our local Internet address. This is used to avoid sending
	messages to ourselves.
******/

	local_addr(0, &portAddr);
	caLocalInet = portAddr.sin_addr;
#ifdef vxWorks
	if (hdlcAddress())
	    caLocalInet.s_addr = hdlcAddress();
#endif
#if !defined(vxWorks) && !defined(vms)
	tisSysLog(1, 0, "sockMnt: local inet = %s", INET_NTOA(caLocalInet));
#else
	if (smDebug >= 1)
#ifdef vxWorks
	    logMsg("sock local inet = %s\n", (int)INET_NTOA(caLocalInet), 0,0,0,0,0);
#else
	    printf("sock local inet = %s\n", INET_NTOA(caLocalInet));
#endif vxWorks
#endif

    FD_ZERO(&fdMask);
/******
	Create a UDP socket. We'll use this one for catching
	broadcasts from IOCs that have just booted and also
	for inviting IOCs to connect to us.
******/

#ifdef USE_UDP
    udpFd = socket(PF_INET, SOCK_DGRAM, 0);
    maxFd = udpFd;
    FD_SET(maxFd, &fdMask);
    portAddr.sin_family = AF_INET;
    portAddr.sin_addr.s_addr = INADDR_ANY;
    portAddr.sin_port = htons(SCM_PORT);
    if (bind(udpFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) == -1)
	{
#if !defined(vxWorks) && !defined(vms)
	tisSysLog(0, 0, "sockMnt can't bind UDP port: %m");
#else
	perror("sockMnt can't bind UDP port.");
#endif vxWorks

#ifdef vxWorks
	taskSuspend(taskIdSelf());	/* For IOC debugging help */
#else
	exit(1);
#endif vxWorks
	}
#endif USE_UDP

/******
	Connect to the name server. Be persistant about it.
******/

#ifndef vxWorks
#ifndef vms
    smTcpFd = socket(PF_INET, SOCK_STREAM, 0);
    hostent = gethostbyname("localhost");
    portAddr.sin_family = AF_INET;
    portAddr.sin_addr.s_addr = **((long **)(hostent -> h_addr_list));
    portAddr.sin_port = htons(NAM_PORT);
    for (;;)
	{
	if (connect(smTcpFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) != -1)
		break;
#ifdef WIN32
	Sleep (2000);
	closesocket (smTcpFd);
#else
	sleep(2);
	close(smTcpFd);
#endif WIN32
    	smTcpFd = socket(PF_INET, SOCK_STREAM, 0);
	}
    setsockopt(smTcpFd, IPPROTO_TCP, TCP_NODELAY, (char *)&true, sizeof(true));

/******
	Get the list of Internet addresses out of the name server. Allocate the
	initial list in one chunk and set up the Inet address in each of the entries.
******/

    msgType = htons(NAME_IOCL);
    send(smTcpFd, (char *) &msgType, sizeof(msgType), 0);
    recv(smTcpFd, (char *) &msgType, sizeof(msgType), 0);
    recv(smTcpFd, (char *) &inetListHdr, sizeof(inetListHdr), 0);
    numIOCs = ntohs(inetListHdr.nsNumNodes);
    inetListHdr.nsNumNodes = numIOCs;
    socketList = (struct sockCtl *) calloc(numIOCs, sizeof(struct sockCtl));
    ctlEntry = socketList;
    while (inetListHdr.nsNumNodes--)
	{
	recv(smTcpFd, (char *) &(ctlEntry -> Inet), sizeof(ctlEntry -> Inet), 0);
	ctlEntry -> fd = INVALID_SOCKET;
	ctlEntry -> bufLocn = (char *) malloc(MAX_XMIT_BUFFER);
	ctlEntry -> rxTank = (char *) malloc(MAX_XMIT_BUFFER);
	if (inetListHdr.nsNumNodes)
	    ctlEntry -> nextSock = ctlEntry + 1;
	ctlEntry++;
	}

/******
	Get the DPA and Resource lists from the name server.
******/

    getDpaList(smTcpFd);

#ifdef WIN32
    closesocket (smTcpFd);
#else
    close(smTcpFd);
#endif WIN32

#else vms

    /*
    Get the Inet address of places that may have name servers
    */

    for (fdIndex = 0; fdIndex < 10; fdIndex++)
	{
	if (scanf("%s", rxTank) == EOF)
	    break;
	namAddrs[numNames++] = inet_addr(rxTank);
	}
#endif vms
#endif vxWorks

/******
	Open the tuning file and extract the tuning parameters from it.
	Give defaults to those who haven't been tuned at all.
******/

    defTuning = defaultTuning;
#ifndef vxWorks
#ifdef WIN32
    tuningFd = _open("socketTuning", _O_RDWR | _O_CREAT, 0644);
#else
    tuningFd = open("socketTuning", O_RDWR | O_CREAT, 0644);
#endif WIN32

#ifdef WIN32
    while (_read (tuningFd, &tempTune, sizeof(tempTune)) > 0)
#else
    while (read(tuningFd, &tempTune, sizeof(tempTune)) > 0)
#endif WIN32
	{
	if (tempTune.stInet.s_addr == 0)
	    {
	    defTuning = tempTune;
	    continue;
	    }
	for (ctlEntry = socketList; ctlEntry; ctlEntry = ctlEntry -> nextSock)
	    {
	    if (ctlEntry -> Inet.s_addr == tempTune.stInet.s_addr)
		{
		ctlEntry -> tuning = tempTune;
		break;
		}
	    }
	}
    for (ctlEntry = socketList; ctlEntry; ctlEntry = ctlEntry -> nextSock)
	{
	if (ctlEntry -> tuning.stInet.s_addr == 0)
	    {
	    struct in_addr	hisAddr;
	    struct in_addr	myAddr;

	    ctlEntry -> tuning = defTuning;
	    ctlEntry -> tuning.stInet.s_addr = ctlEntry -> Inet.s_addr;
	    hisAddr.s_addr = ctlEntry -> Inet.s_addr;
	    myAddr = caLocalInet;
#ifndef PRIV_AUDIT
	    /* Dont do slow net timeout for CAISO */
	    if (inet_netof(hisAddr) != inet_netof(myAddr))
		{
		ctlEntry -> tuning.stMaxQuiet = 6;
		ctlEntry -> tuning.stMaxIdle = 45;
		}
#endif PRIV_AUDIT
	    }
	}
#ifdef WIN32
    _close (tuningFd);
#else
    close(tuningFd);
#endif WIN32
#endif vxWorks

/******
	Register with Channel Acquisition and Control so he knows that we
	are here.
******/
	
    cacFd = socket(PF_INET, SOCK_STREAM, 0);
    for (;;)
	{
	portAddr.sin_family = AF_INET;
	portAddr.sin_port = htons(CAM_PORT);
	portAddr.sin_addr.s_addr = htonl(0x7f000001);
	if (connect(cacFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) != -1)
		break;
#ifdef vxWorks
	taskDelay(120);
#else
#ifdef WIN32
	Sleep(2000);
#else
	sleep(2);
#endif WIN32
#endif vxWorks
#ifdef WIN32
	closesocket (cacFd);
#else
	close(cacFd);
#endif WIN32
	cacFd = socket(PF_INET, SOCK_STREAM, 0);
	}
    setsockopt(cacFd, IPPROTO_TCP, TCP_NODELAY, (char *)&true, sizeof(true));
#ifndef SOLARIS
#ifdef WIN32
    ioctlsocket (cacFd, FIONBIO, &true);
#else
    ioctl(cacFd, FIONBIO, (int)&true);
#endif WIN32
#else
    sockGrow(cacFd);
    fcntl(cacFd, F_SETFL, O_NONBLOCK);
#endif
    intsize = sizeof(cacRxSize);
    getsockopt(cacFd, SOL_SOCKET, SO_RCVBUF, (char *)&cacRxSize, &intsize);
    intsize = sizeof(cacTxSize);
    getsockopt(cacFd, SOL_SOCKET, SO_SNDBUF, (char *)&cacTxSize, &intsize);
    cacTxTopped = 0;
    cacRxTopped = 0;

    msgType = CHAN_RGST;
    send(cacFd, (char *) &msgType, sizeof(msgType), 0);
#ifndef WIN32
    if (cacFd > maxFd)
	maxFd = cacFd;
#endif WIN32
    FD_SET(cacFd, &fdMask);
#ifndef vxWorks
    FD_ZERO(&scmMask);
    FD_SET(cacFd, &scmMask);
#endif vxWorks

/******
	Create a TCP socket so that IOCs (and others) can connect to it.
******/

    smTcpFd = socket(PF_INET, SOCK_STREAM, 0);
    setsockopt(smTcpFd, SOL_SOCKET, SO_REUSEADDR, (caddr_t) &optVal, sizeof(optVal));
    portAddr.sin_addr.s_addr = INADDR_ANY;
    portAddr.sin_port = htons(SCM_PORT);
    if (bind(smTcpFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) == -1)
	{
#if !defined(vxWorks) && !defined(vms)
	tisSysLog(0, 0, "sockMnt can't bind TCP port: %m");
#else
	perror("sockMnt can't bind TCP port.\n");
#endif vxWorks

#ifdef vxWorks
	taskSuspend(taskIdSelf());	/* For IOC debugging help */
#else
	exit(1);
#endif vxWorks
	}
    listen(smTcpFd, 2);
#ifndef WIN32
    if (smTcpFd > maxFd)
	maxFd = smTcpFd;
#endif WIN32
    FD_SET(smTcpFd, &fdMask);

/******
	We're now ready for the main processing loop. Read messages and react
	to them. Traverse the control table once a second.
******/

    for (;;)
	{

	/*
	Make a local copy of the current mask. Then, drain all the ports.
	What we will do is put a select on with a 1 second timeout. As soon
	as that select returns something, we'll read all of the remaining
	messages. After having drained all of the messages, we'll flush the
	buffer of messages awaiting processing by Channel Acquisition.
	*/

	fdList = fdMask;
#ifdef vxWorks
	anyInputs = select(maxFd + 1, &fdList, NULL, NULL, &oneSecond);
#else
	if (scmExtBytes)
	    {
	    scmWriteReady = &scmWriteFds;
	    bcopy((char *) &scmMask, (char *) &scmWriteFds, sizeof(scmWriteFds));
	    }
	else
	    scmWriteReady = NULL;
	anyInputs = select(maxFd + 1, &fdList, scmWriteReady, NULL, &oneSecond);
	if (scmWriteReady != NULL)
	    if ((cacFd != INVALID_SOCKET) && FD_ISSET(cacFd, &scmWriteFds))
		scmMobySend();
#endif vxWorks
			
	if (anyInputs)
#ifdef vxWorks
	  while (select(maxFd + 1, &fdList, NULL, NULL, &pollTimeout) > 0)
#else
	  if (scmExtBytes)
	      {
	      scmWriteReady = &scmWriteFds;
	      bcopy((char *) &scmMask, (char *) &scmWriteFds, sizeof(scmWriteFds));
	      }
	  else
	      scmWriteReady = NULL;
	  while (select(maxFd + 1, &fdList, scmWriteReady, NULL, &pollTimeout) > 0)
#endif vxWorks
	    {

#ifndef vxWorks
	    if (scmWriteReady)
		if ((cacFd != INVALID_SOCKET) && FD_ISSET(cacFd, &scmWriteFds))
		    {
		    scmMobySend();
		    if (!scmExtBytes)
			scmWriteReady = NULL;
		    }
#endif vxWorks

	    /*
	    Check each fd. Handle the I/O as appropriate.
	    */
#ifdef USE_UDP
	    /*
	    If this is the UDP port, read the message and
	    see if we're interested in this IOC. Remember that
	    there might be more than one TIS 4000 on a single
	    Ethernet.
	    */

	    if ((udpFd != INVALID_SOCKET) && FD_ISSET(udpFd, &fdList))
		{
		char	udpBuff[sizeof(SM_DPAREQ) + sizeof(msgType)];

		FD_CLR (udpFd, &fdList);
#if !defined(vxWorks) && !defined(vms)
		tisSysLog(1, 0, "sockMnt UDP msg received");
#else
		if (smDebug >= 1)
#ifdef vxWorks
	    	    logMsg ("sock UDP msg received\n",0,0,0,0,0,0);
#else
		    printf("sock UDP msg received\n");
#endif vxWorks
#endif
		portAddrSize = sizeof(portAddr);
		recvfrom(udpFd, udpBuff, sizeof(udpBuff), 0, 
			(struct sockaddr *) &portAddr, &portAddrSize);
		msgType = *((short *) udpBuff);
		msgType = htons(msgType);
		if (msgType == SOCK_CONN)
		    {
		
		    /*
		    See if we know about this guy. Ignore him if we don't.
		    He belongs to a different TIS-4000.
		    */

		    if (knownIOC(portAddr.sin_addr))
			{

			/*
			See if this guy was already connected. If so, inform
			Channel Acquisition and Maintenance that this IOC is
			now disconnected. In either case, invite him to connect
			with us.
			*/

			ctlEntry = findInet(portAddr.sin_addr);
			if (ctlEntry -> fd != INVALID_SOCKET && (ctlEntry -> iocDownSent == 0))
			    {
			    CH_IOCDN	iocDownMsg1;

#ifdef PRIV_AUDIT
			    SSL *conn = sslSharedGetConn (ctlEntry -> fd);
			    if (auditLog)
			    {
				GetPeerDN (conn, X509_dn, sizeof(X509_dn));
#ifdef vxWorks
				AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
					taskName(taskIdSelf()), "Prior connection closed");
#else vxWorks

				AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
					"sockMnt", "Prior connection closed");
#endif vxWorks
			    }
			    sslSharedCloseConn (conn);
#else PRIV_AUDIT
#ifdef WIN32
			    closesocket (ctlEntry -> fd);
#else WIN32
			    close(ctlEntry -> fd);
#endif WIN32
#endif PRIV_AUDIT
			    FD_CLR(ctlEntry -> fd, &fdMask);
			    ctlEntry -> fd = INVALID_SOCKET;
			    msgType = CHAN_IOCDN;
			    sendCac((char *) &msgType, sizeof(msgType));
			    iocDownMsg1.chInet = ctlEntry -> Inet;
			    if (iocDownMsg1.chInet.s_addr == htonl(0x7f000001))
			        iocDownMsg1.chInet = caLocalInet;
			    sendCac((char *) &iocDownMsg1, sizeof(iocDownMsg1));
			    ctlEntry -> iocDownSent = 1;
			    *((short *) rxBuff) = htons(SOCK_CONN_REPLY);
			    sendto(udpFd, rxBuff, sizeof(short), 0, (struct sockaddr *)&portAddr, portAddrSize);
#if !defined(vxWorks) && !defined(vms)
			    tisSysLog(1, 0, "sockMnt inviting %s to connect", INET_NTOA(ctlEntry -> Inet));
#else
			    if (smDebug)
#ifdef vxWorks
				logMsg("sock inviting %s to connect\n", (int)INET_NTOA(ctlEntry -> Inet), 0,0,0,0,0);
#else
				printf("sock inviting %s to connect\n", INET_NTOA(ctlEntry -> Inet));
#endif vxWorks
#endif
			    }
			}
		    }

#if defined(vxWorks) || defined(vms)
#ifndef PRIV_AUDIT
		else if (msgType == NAME_PROBE && needNameSrv)
		    {
		    CH_GOTNMSRV		gotNmsrv;

		    /*
		    This code should only ever run under vxWorks or vms.
		    A name server has responded to our probe. Tell chanAcq to
		    connect to him, but first try to extract the DPA and resource
		    lists. Do this before we tell chanAcq about our newly found
		    name server.
		    */

		    if (!dpaListBuilt)
			{
			int	dpaSocket;

			if ((dpaSocket = socket(PF_INET, SOCK_STREAM, 0)) < 0)
			    continue;
			if (connect(dpaSocket, (struct sockaddr *) &portAddr, sizeof(portAddr)) == -1)
			    {
#ifdef WIN32
			    closesocket (dpaSocket);
#else WIN32
			    close(dpaSocket);
#endif WIN32
			    continue;
			    }
			if (getDpaList(dpaSocket))
			    {
			    dpaListBuilt = 1;
			    }
#ifdef WIN32
			closesocket (dpaSocket);
#else WIN32
			close(dpaSocket);
#endif WIN32
			}

		    gotNmsrv.chInet = portAddr.sin_addr;
		    if (smDebug >= 1)
#ifdef vxWorks
			logMsg("sock got name server probe from %s\n",
				(int)INET_NTOA(gotNmsrv.chInet),0,0,0,0,0);
#else vxWorks
			printf("sock got name server probe from %s\n",
				INET_NTOA(gotNmsrv.chInet));
#endif vxWorks
		    msgType = CHAN_GOTNMSRV;
		    sendCac((char *) &msgType, sizeof(msgType));
		    sendCac((char *) &gotNmsrv, sizeof(gotNmsrv));
		    needNameSrv = 0;
		    }
#endif PRIV_AUDIT

#else defined(vxWorks) || defined(vms)
		else if (msgType == BOOMERANG)
		    {
		    msgType = htons(BOOMERANG);
		    sendto(udpFd, (char *) &msgType, sizeof(msgType), 0,
			    (struct sockaddr *)&portAddr, portAddrSize);
#if !defined(vxWorks) && !defined(vms)
		    tisSysLog(2, 0, "sockMnt got UDP boomerang from %s", 
			    INET_NTOA(portAddr.sin_addr));
#else
		    if (smDebug >= 2)
			printf("sockMnt got UDP boomerang from %s\n", 
				INET_NTOA(portAddr.sin_addr.s_addr));
#endif
		    }

		else if (msgType == SOCK_DPAREQ)
		    {
		    struct dpaBuff
		    	{
		    	short		dpaType;
		    	SM_DPAREQ	dpaReq;
		    	} *dpaReqMsg;

		    /*
		    Some other sockMnt wants us to notify a DPA that is running on
		    our host that a connection is desired. Look up the prefix so
		    we can find the DPA.
		    */

		    dpaReqMsg = (struct dpaBuff *) udpBuff;
		    for (dpaEntry = dpaList; dpaEntry; dpaEntry = dpaEntry -> nextDPA)
			{
			if (!strcmp(dpaReqMsg -> dpaReq.dpaPrefix, dpaEntry -> prefix))
			    {
			    struct
				{
				short		dpaType;
				short		padbytes;
				SM_DPACONN	dpaConn;
				} dpaConnMsg;

			    /*
			    Found him. Send a connect request to any DPAs that are on
			    line at the moment.
			    */

			    dpaConnMsg.dpaType = htons(SOCK_DPACONN);
			    dpaConnMsg.dpaConn.dpaInet = portAddr.sin_addr;
			    if ((dpaEntry -> sockFd1 > 0) && (dpaEntry -> realInet1.s_addr == 0x7f000001))
				{
			    	ctlEntry = findInet(dpaEntry -> virtInet1);
				if ((ctlEntry -> bufSize + sizeof(dpaConnMsg)) > ctlEntry -> tuning.stMaxBlock)
				    {
				    sendIOC(ctlEntry);
				    }

				bcopy((char *) &dpaConnMsg, ctlEntry -> bufLocn + ctlEntry -> bufSize, sizeof(dpaConnMsg));
			        ctlEntry -> bufSize += sizeof(dpaConnMsg);
				sendIOC(ctlEntry);
				}
			    if ((dpaEntry -> sockFd2 > 0) && (dpaEntry -> realInet2.s_addr == 0x7f000001))
				{
			    	ctlEntry = findInet(dpaEntry -> virtInet2);
				if ((ctlEntry -> bufSize + sizeof(dpaConnMsg)) > ctlEntry -> tuning.stMaxBlock)
				    {
				    sendIOC(ctlEntry);
				    }

				bcopy((char *) &dpaConnMsg, ctlEntry -> bufLocn + ctlEntry -> bufSize, sizeof(dpaConnMsg));
			        ctlEntry -> bufSize += sizeof(dpaConnMsg);
				sendIOC(ctlEntry);
				}
			    }
			}
		    }
#endif defined(vxWorks) || defined(vms)
		}
#endif USE_UDP
	    /*
	    See if this is a message from Channel Acquisition and Control.
	    */

	    if ((cacFd > 0) && FD_ISSET (cacFd, &fdList))
	      {
	      FD_CLR (cacFd, &fdList);
	      cacFlowControl();
	      fillChanBuffer();
	      while (haveCompleteChanMsg())
		{
		recvCac((char *) &msgType, sizeof(msgType));
		if (msgType == SOCK_FLUSH)
		    {

		    /*
		    Socket flush for everybody
		    */

#if !defined(vxWorks) && !defined(vms)
		    tisSysLog(2, 0, "sockMnt flush");
#else
		    if (smDebug >= 2)
#ifdef vxWorks
			logMsg ("sockMnt flush\n", 0,0,0,0,0,0);
#else
			printf("sockMnt flush\n");
#endif vxWorks
#endif
		    for (ctlEntry = socketList; ctlEntry; ctlEntry = ctlEntry -> nextSock)
			{
			if (ctlEntry -> fd == INVALID_SOCKET)
			    continue;
			if (!(ctlEntry -> bufSize))
			    continue;
			sendIOC(ctlEntry);
			}
		    }

		else if (msgType == CHAN_NEEDNMSRV)
		    {
		    /*
		    chanAcq is informing us that he needs a connection with the
		    name server. We'll find it for him. The reason that we do it
		    instead of him is that we have time based events and he
		    doesn't. All we have to do is set the flag that a name server
		    is wanted and the logic at the end of the loop will take care
		    of probing for one on the network.
		    */

		    needNameSrv = 1;
		    if (smDebug >= 1)
#ifdef vxWorks
			logMsg ("sock sets need name server\n", 0,0,0,0,0,0);
#else
			printf("sock sets need name server\n");
#endif vxWorks
		    }
		else if (msgType == SOCK_WRITE)
		    {

		    /*
		    Socket write. First get the header and then find the control
		    table for this guy. If we can't find it, he's new so we have
		    to create one for him. If there's room, buffer up the message.
		    If there isn't, send out what we have and then buffer up the
		    new one.
		    */

		    SM_WRITE	writeHdr;
		    int		ourIOC;

		    recvCac((char *) &writeHdr, sizeof(writeHdr));

		    /*
		    For some reason, the Microsol unit can't deal with its own
		    explicit Inet address. Therefore, change our own address
		    to 127.0.0.1 so that we'll make the connection.
		    */

		    if (writeHdr.smInet.s_addr == caLocalInet.s_addr)
			writeHdr.smInet.s_addr = htonl(0x7f000001);

#if !defined(vxWorks) && !defined(vms)
		    tisSysLog(2, 0, "sockMnt output message for %s", INET_NTOA(writeHdr.smInet));
#else
		    if (smDebug >= 2)
#ifdef vxWorks
			logMsg ("Output message for %s\n", (int)INET_NTOA(writeHdr.smInet),0,0,0,0,0);
#else
			printf("Output message for %s\n", INET_NTOA(writeHdr.smInet));
#endif vxWorks
#endif

		    ourIOC = knownIOC(writeHdr.smInet);
		    ctlEntry = findInet(writeHdr.smInet);
		    if (ourIOC)
			{
			if ((unsigned)(ctlEntry -> bufSize + writeHdr.smLength) > ctlEntry -> tuning.stMaxBlock)
			    {
			    sendIOC(ctlEntry);
			    }

			/*
			We have to read the data out of the socket immediately to avoid hanging 
			Channel Acquisition and Control.
			*/

			recvCac(ctlEntry -> bufLocn + ctlEntry -> bufSize, writeHdr.smLength);
			ctlEntry -> bufSize += writeHdr.smLength;
			}
		    else
			{
			/*
			This guy is new. Discard the data and tell Channel Acquisition and Control 
			that this guy is down.  Channel Acquisition and Control always assumes that 
			an IOC is up unless we tell him otherwise. This way, we will find out about 
			new IOCs that are added on the fly.
			*/

#if !defined(vxWorks) && !defined(vms)
			tisSysLog(1, 0, "First write to new IOC %s", INET_NTOA(writeHdr.smInet));
#else
			if (smDebug >= 1)
#ifdef vxWorks
			    logMsg ("First write to new IOC %s\n", (int)INET_NTOA(writeHdr.smInet), 0,0,0,0,0);
#else
			    printf("First write to new IOC %s\n", INET_NTOA(writeHdr.smInet));
#endif vxWorks
#endif
			recvCac(ctlEntry -> bufLocn, writeHdr.smLength);
			numIOCs++;

			/*
			Now that we know about this guy, try to get him to connect to us.
			*/

			if (ntohl(ctlEntry -> Inet.s_addr) > 256)
			    {
#ifdef USE_UDP
			    portAddr.sin_family = AF_INET;
			    portAddr.sin_addr.s_addr = ctlEntry -> Inet.s_addr;
			    portAddr.sin_port = htons(RSV_PORT);
			    msgType = htons(SOCK_CONN_REPLY);
			    if (sendto(udpFd, (char *) &msgType, sizeof(msgType), 0, (struct sockaddr *)&portAddr,
				    sizeof(portAddr)) != sizeof(msgType))
#if !defined(vxWorks) && !defined(vms)
				tisSysLog(0, 0, "sockMnt send error for UDP IOC connect invite: %m");
#else
				perror("sockMnt sock send error for UDP IOC connect invite");
#endif
#else  USE_UDP
			    /* connect to target, then close - this will announce our needs */
			    kick_TCP (ctlEntry -> Inet.s_addr, RSV_PORT);
#endif USE_UDP

			/*
			Try to find this new node in the tuning file if there is one.
			*/

#if defined(vms)
#ifdef WIN32
			    tuningFd = _open("socketTuning", _O_RDWR | _O_CREAT, 0644);
#else
			    tuningFd = open("socketTuning", O_RDWR | O_CREAT, 0644);
#endif WIN32
			    if (tuningFd >= 0)
				{
				while (read(tuningFd, &tempTune, sizeof(tempTune)))
				    if (tempTune.stInet == ctlEntry -> Inet)
					{
					ctlEntry -> tuning = tempTune;
					break;
					}
#ifdef WIN32
				_close(tuningFd);
#else
				close(tuningFd);
#endif WIN32
				}
#endif vms
			    }

			else
			    {

			    /*
			    It is a DPA. Now we should set the "wanted" member so
			    that we'll actually connect later.
			    */

			    for (dpaEntry = dpaList; dpaEntry; dpaEntry = dpaEntry -> nextDPA)
				{
				if ((dpaEntry -> virtInet1.s_addr == writeHdr.smInet.s_addr) ||
				    (dpaEntry -> virtInet2.s_addr == writeHdr.smInet.s_addr))
				    {
				    dpaEntry -> wanted = 1;
				    break;
				    }
				}
			    }
			}

		    /*
		    If there isn't a TCP socket available for this IOC, discard the data and
		    then tell channel Acquisition that he's down. This will keep him from sending 
		    us output traffic for that node until we tell him that it came online.
		    */

		    if ((ctlEntry -> fd == INVALID_SOCKET) && (ctlEntry -> iocDownSent == 0))
			{
			CH_IOCDN	iocDownMsg;

			msgType = CHAN_IOCDN;
			sendCac((char *) &msgType, sizeof(msgType));
			iocDownMsg.chInet = writeHdr.smInet;
			if (iocDownMsg.chInet.s_addr == htonl(0x7f000001))
			    iocDownMsg.chInet = caLocalInet;
			sendCac((char *) &iocDownMsg, sizeof(iocDownMsg));
			ctlEntry -> iocDownSent = 1;
#if !defined(vxWorks) && !defined(vms)
			tisSysLog(1, 0, "sockMnt told chanAcq %s is down", INET_NTOA(writeHdr.smInet));
#else
			if (smDebug >= 1)
#ifdef vxWorks
			    logMsg ("sockMnt told chanAcq %s is down\n", (int)INET_NTOA(writeHdr.smInet), 0,0,0,0,0);
#else
			    printf("sockMnt told chanAcq %s is down\n", INET_NTOA(writeHdr.smInet));
#endif vxWorks
#endif
			}
		    }
		}
	    }
	    
	/*
	See if this is someone trying to connect with us.
	*/

	if ((smTcpFd > 0) && FD_ISSET(smTcpFd, &fdList))
	    {

	    SOCKET	newFd;
#ifdef PRIV_AUDIT
	    SSL		*newSSL = NULL;
#endif PRIV_AUDIT

	    FD_CLR (smTcpFd, &fdList);
	    portAddrSize = sizeof(portAddr);
	    if ((newFd = accept(smTcpFd, (struct sockaddr *)&portAddr, &portAddrSize)) == ERROR)
		continue;
#ifdef PRIV_AUDIT
	    if ((newSSL = setup_ssl_accept (newFd)))
	    {
		/* note - this is blocking IO, but won't take longer than 5 seconds */
		if (!accept_ssl_connection_timed (newSSL, 10/*seconds*/))
		{
		    /* failed SSL handshake */
#ifdef vxWorks
		    logMsg("SSL accept error, %s\n",
			    (int)INET_NTOA(portAddr.sin_addr),0,0,0,0,0);
#else vxWorks
		    tisSysLog (0, 0, "SSL accept error, %s\n",
			    INET_NTOA(portAddr.sin_addr));
#endif vxWorks
		    SecurityAlarm (INET_NTOA(portAddr.sin_addr), "Priviledge error, failed SSL handshake", NULL);
		    if (auditLog)
		    {
#ifdef vxWorks
			AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_FAIL, INET_NTOA(portAddr.sin_addr),
				taskName(taskIdSelf()), "Accept SSL failed");
#else vxWorks
			AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_FAIL, INET_NTOA(portAddr.sin_addr),
				"sockMnt", "Accept SSL failed");
#endif vxWorks
		    }
		    sslSharedCloseConn (newSSL);
		    newSSL = NULL;
		}
		else
		{
		    /* validate connection CN */
		    GetPeerCN (newSSL, X509_dn, sizeof(X509_dn));
		    if (GetGeneralPrivHost (X509_dn, "CA_") < 1)					
		    {
			GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
			/* No priviledge to connect! */
#ifdef vxWorks
			logMsg ("CA priviledge error, %s not authorized\n",
				(int)X509_dn,0,0,0,0,0);
#else vxWorks
			tisSysLog (0, 0, "CA priviledge error, %s not authorized\n",X509_dn);
#endif vxWorks
			SecurityAlarm (X509_dn, "Priviledge error, Not authorized", INET_NTOA(portAddr.sin_addr));
			if (auditLog)
			{
			    char msg [80];

			    sprintf (msg, "Client privilege error from %s", INET_NTOA(portAddr.sin_addr));
#ifdef vxWorks
			    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_FAIL, X509_dn,
				    taskName(taskIdSelf()), msg);
#else vxWorks
			    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_FAIL, X509_dn,
				    "sockMnt", msg);
#endif vxWorks
			}
			sslSharedCloseConn (newSSL);
			newSSL = NULL;
		    }
		    else
		    {
			GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
			if (auditLog)
			{
			    char msg [80];

			    sprintf (msg, "Client connected from %s", INET_NTOA(portAddr.sin_addr));
#ifdef vxWorks
			    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_SUCCESS, X509_dn,
				    taskName(taskIdSelf()), msg);
#else vxWorks
			    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_SUCCESS, X509_dn,
				    "sockMnt", msg);
#endif vxWorks
			}
		    }
		}
	    }
	    if (newSSL == NULL)
	    {
		continue;
	    }
#endif PRIV_AUDIT
	    if (portAddr.sin_addr.s_addr == caLocalInet.s_addr)
		portAddr.sin_addr.s_addr = htonl(0x7f000001);	 /* Locals connect as 127.0.0.1 */

#if !defined(vxWorks) && !defined(vms)
	    tisSysLog(1, 0, "sockMnt new TCP connection from %s\n", INET_NTOA(portAddr.sin_addr));
#else !defined(vxWorks) && !defined(vms)
	    if (smDebug >= 1)
#ifdef vxWorks
		logMsg("sockMnt new TCP connection from %s\n", (int)INET_NTOA(portAddr.sin_addr),0,0,0,0,0);
#else vxWorks
		printf("sockMnt new TCP connection from %s\n", INET_NTOA(portAddr.sin_addr));
#endif vxWorks
#endif !defined(vxWorks) && !defined(vms)

#ifndef WIN32
	    if (newFd > maxFd)
		maxFd = newFd;
#endif WIN32
	    setsockopt(newFd, IPPROTO_TCP, TCP_NODELAY, (char *)&true, sizeof(true));

	    /*
	    If this is a host supporting a DPA, then do nothing. We want him to be a utility
	    and then get promoted into becoming an IOC.
	    */

	    for (dpaEntry = dpaList; dpaEntry; dpaEntry = dpaEntry -> nextDPA)
		{
		if (portAddr.sin_addr.s_addr == dpaEntry -> realInet1.s_addr)
		    break;
		if (portAddr.sin_addr.s_addr == dpaEntry -> realInet2.s_addr)
		    break;
		}
	    if (dpaEntry)
		{
		/* put newFd on list of utility sockets */
		for (fdIndex = 0; fdIndex < utilityMax; fdIndex++)
		    {
		    if (utilitySockets [fdIndex] == INVALID_SOCKET)
			break;
		    }
		if (fdIndex < MAX_UTIL_CLIENTS)
		    {
		    if (fdIndex == utilityMax)
			utilityMax++;
		    utilitySockets [fdIndex] = newFd;
		    FD_SET(newFd, &fdMask);
		    }
		else
		    {
#ifdef WIN32
		    closesocket (newFd);
#else WIN32
		    close(newFd);
#endif WIN32
		    }
		continue;
		}

	    /*
	    If this is an IOC, then mark him up and then inform
	    Channel Acquisition and Control. If not an IOC, then
	    just let it be there. Those connections will go away
	    when they are closed at the other end.
	    */

	    /*
	    Don't use findInet -- he'll create a new one!
	    */

	    if (knownIOC(portAddr.sin_addr) && (portAddr.sin_addr.s_addr != caLocalInet.s_addr))
		{
		CH_IOCUP	iocUp;
		int		intSize;

		ctlEntry = findInet(portAddr.sin_addr);
#if !defined(vxWorks) && !defined(vms)
		tisSysLog(1, 0, "sockMnt CONN %s on %d @ %d", INET_NTOA(ctlEntry -> Inet), newFd, time(NULL));
#else !defined(vxWorks) && !defined(vms)
		if (smDebug >= 1)
#ifdef vxWorks
		    logMsg ("sockMnt CONN %s on %d @ %d\n", (int)INET_NTOA(ctlEntry -> Inet), newFd, time(NULL),0,0,0);
#else vxWorks
		    printf("sockMnt CONN %s on %d @ %d\n", INET_NTOA(ctlEntry -> Inet), newFd, time(NULL));
#endif vxWorks
#endif !defined(vxWorks) && !defined(vms)
		ctlEntry -> fd = newFd;
		FD_SET(newFd, &fdMask);
		ctlEntry -> lastRx = time(NULL);
		ctlEntry -> boomsSent = 0;
		ctlEntry -> nextBoom = 0;
		ctlEntry -> bufSize = 0;
		ctlEntry -> nchInTank = 0;
		ctlEntry -> tankOutInd = 0;
		msgType = CHAN_IOCUP;
		sendCac((char *) &msgType, sizeof(msgType));
		iocUp.chInet = ctlEntry -> Inet;
		if (iocUp.chInet.s_addr == htonl(0x7f000001))
		    iocUp.chInet = caLocalInet;
		sendCac((char *) &iocUp, sizeof(iocUp));
		ctlEntry -> iocDownSent = 0;

		/*
		Mark the socket for non-blocking I/O. Do this so we won't hang
		if he dies.
		*/

#ifndef SOLARIS
#ifdef WIN32
		if (ioctlsocket (newFd, FIONBIO, &true))
#else WIN32
		if (ioctl(newFd, FIONBIO, (int)&true))
#endif WIN32
#else SOLARIS
		sockGrow(newFd);
		if (fcntl(newFd, F_SETFL, O_NONBLOCK) == -1)
#endif SOLARIS
#if !defined(vxWorks) && !defined(vms)
		tisSysLog(0, 0, "sockMnt Error setting non-blocking I/O for IOC: %m");
#else !defined(vxWorks) && !defined(vms)
		perror("sockMnt Error setting non-blocking I/O for IOC\n");
#endif !defined(vxWorks) && !defined(vms)

		/*
		Get the current receive and transmit buffer sizes
		*/

		intSize = sizeof(ctlEntry -> currRxBuff);
		getsockopt(ctlEntry -> fd, SOL_SOCKET, SO_RCVBUF, 
			(char *)&(ctlEntry -> currRxBuff), &intSize);
		intSize = sizeof(ctlEntry -> currTxBuff);
		getsockopt(ctlEntry -> fd, SOL_SOCKET, SO_SNDBUF,
			(char *)&(ctlEntry -> currTxBuff), &intSize);
		ctlEntry -> rxMaxOut = 0;
		}
	    else
		{
#ifdef WIN32
		closesocket (newFd);
#else WIN32
		close(newFd);
#endif WIN32
		}
	    }

	for (ctlEntry = socketList; ctlEntry; ctlEntry = ctlEntry -> nextSock)
	    {
	    if (ctlEntry->fd != INVALID_SOCKET
	      && FD_ISSET(ctlEntry->fd, &fdList))
		{

	        FD_CLR (ctlEntry->fd, &fdList);
		/*
		See if this message is from an IOC.
		*/

		rxFlowControl(ctlEntry);

		if (!getIocData(ctlEntry))
		    {
		    iocDown(ctlEntry);
#if !defined(vxWorks) && !defined(vms)
		    tisSysLog(1, 0, "sockMnt kill %s because of I/O error: %m",
			INET_NTOA (ctlEntry -> Inet));
#else !defined(vxWorks) && !defined(vms)
		    if (smDebug >= 1)
#ifdef vxWorks
			logMsg("sock kill %s because of I/O error\n",
			    (int)INET_NTOA (ctlEntry -> Inet),0,0,0,0,0);
#else vxWorks
			printf("sock kill %s because of I/O error\n",
			    INET_NTOA (ctlEntry -> Inet));
#endif vxWorks
#endif !defined(vxWorks) && !defined(vms)
		    }
		else
		    {
#if !defined(vxWorks) && !defined(vms)
		    tisSysLog(2, 0, "sockMnt got msg from %s", INET_NTOA(ctlEntry -> Inet));
#else !defined(vxWorks) && !defined(vms)
		    if (smDebug >= 2)
#ifdef vxWorks
			logMsg("sockMnt got msg from %s\n", (int)INET_NTOA(ctlEntry -> Inet),0,0,0,0,0);
#else vxWorks
			printf("sockMnt got msg from %s\n", INET_NTOA(ctlEntry -> Inet));
#endif vxWorks
#endif !defined(vxWorks) && !defined(vms)
		    }
		    

		while (haveCompleteIocMsg(ctlEntry))
		    {

		    /*
		    From an IOC. Find out what he wants and handle it.
		    At the current time we don't really have to do
		    anything with the message type, but we read it
		    anyway to keep things neat. This is a hook for
		    further expansion as the program evolves in the
		    future.
		    */

		    if (recvIOC(ctlEntry, (char *) &msgType, sizeof(msgType)))
			{
			msgType = ntohs(msgType);
			if (msgType == SOCK_READ)
			    {

			    /*
			    Data to be interpreted by Channel Acquisition.
			    Supply the Inet address and then just forward
			    it to him.
			    */

			    SM_READ		sockRead;
			    int			length;

			    if (recvIOC(ctlEntry, (char *) &sockRead, sizeof(sockRead)))
				{
				sockRead.smInet = ctlEntry -> Inet;

				/*
				chanAcq only knows the real Inet, so we have to undo
				the trick that we did on the way out for the Microsol.
				*/

				if (sockRead.smInet.s_addr == htonl(0x7f000001))
				    sockRead.smInet = caLocalInet;

				length = ntohs(sockRead.smLength);
				if (recvIOC(ctlEntry, rxBuff, length))
				    {
				    sendCac((char *) &msgType, sizeof(msgType));
				    sendCac((char *) &sockRead, sizeof(sockRead));
				    sendCac(rxBuff, length);
				    }
				}
			    }
			}
		    }
		}
    	    }
	/*
	Must be from a utility. See what he wants and handle it.
	*/

	for (fdIndex = 0; fdIndex < utilityMax; fdIndex++)
	    {
	    if (FD_ISSET (utilitySockets[fdIndex], &fdList))
		{
		FD_CLR (utilitySockets[fdIndex], &fdList);
		if (recvUtil(utilitySockets[fdIndex], (char *) &msgType, sizeof(msgType)))
		    {
		    msgType = ntohs(msgType);
		    if (msgType == SOCK_PARM)
			{

			/*
			Utility wants our tuning parameters. Feed them to him.
			*/

			msgType = htons(SOCK_PARM_REPLY);
			if (sendUtil(utilitySockets[fdIndex], (char *) &msgType, sizeof(msgType)))
			    {
			    SM_PARM_REPLY	parmReply;

			    parmReply.smNumSockets = htons((short)(numIOCs + 1)); /* Include default! */
			    if (sendUtil(utilitySockets[fdIndex], (char *) &parmReply, sizeof(parmReply)))
				{
				tempTune = defTuning;
				tempTune.stMaxIdle = htonl(tempTune.stMaxIdle);
				tempTune.stMaxQuiet = htonl(tempTune.stMaxQuiet);
				tempTune.stMaxBlock = htonl(tempTune.stMaxBlock);
				tempTune.stInqRate = htonl(tempTune.stInqRate);
				if (sendUtil(utilitySockets[fdIndex], (char *) &tempTune, sizeof(tempTune)))
				    {
				    ctlEntry = socketList;
				    parmReply.smNumSockets--;
				    while (parmReply.smNumSockets--)
					{
					tempTune = ctlEntry -> tuning;
					tempTune.stInet = tempTune.stInet;
					tempTune.stMaxIdle = htonl(tempTune.stMaxIdle);
					tempTune.stMaxQuiet = htonl(tempTune.stMaxQuiet);
					tempTune.stMaxBlock = htonl(tempTune.stMaxBlock);
					tempTune.stInqRate = htonl(tempTune.stInqRate);
					if (!sendUtil(utilitySockets[fdIndex], (char *) &tempTune, sizeof(tempTune)))
					    break;
	    				ctlEntry = ctlEntry -> nextSock;
					}
				    }
				}
			    }
			}
		    else if (msgType == SOCK_TUNE)
			{

			/*
			Utility has modified tuning parameters. Read, use, and save
			them. The methods used for file I/O are not very efficient,
			but this should happen only on rare circumstance and the
			list of IOCs would not normally be very long.
			*/

			SM_TUNE		sockTuneHdr;
			sockTune	tuneBuff;

			if (recvUtil(utilitySockets[fdIndex], (char *) &sockTuneHdr, sizeof(sockTuneHdr)))
			    {
			    sockTuneHdr.smNumSockets = ntohs(sockTuneHdr.smNumSockets);
#ifdef WIN32
			    tuningFd = _open("socketTuning", _O_RDWR | _O_CREAT, 0644);
#else WIN32
			    tuningFd = open("socketTuning", O_RDWR | O_CREAT, 0644);
#endif WIN32
			    while (sockTuneHdr.smNumSockets--)
				{
				if (!recvUtil(utilitySockets[fdIndex], (char *) &tempTune, sizeof(tempTune)))
					break;
				tempTune.stInet = tempTune.stInet;
				tempTune.stMaxIdle = ntohl(tempTune.stMaxIdle);
				tempTune.stMaxQuiet = ntohl(tempTune.stMaxQuiet);
				tempTune.stMaxBlock = ntohl(tempTune.stMaxBlock);
				tempTune.stInqRate = ntohl(tempTune.stInqRate);
				lseek(tuningFd, 0, SEEK_SET);
				while (read(tuningFd, (char *) &tuneBuff, sizeof(tuneBuff)))
				    {
				    if (tuneBuff.stInet.s_addr == tempTune.stInet.s_addr)
					{
					lseek(tuningFd, -(int)sizeof(tuneBuff), SEEK_CUR);
					break;
					}
				    }
				write(tuningFd, (char *) &tempTune, sizeof(tempTune));

				/*
				That put it in the file. Now also use the memory copy.
				*/

				if (tempTune.stInet.s_addr == 0)
				    defTuning = tempTune;
				else if ((ctlEntry = findInet(tempTune.stInet)))
				    ctlEntry -> tuning = tempTune;
				}
#ifdef WIN32
			    _close(tuningFd);
#else WIN32
			    close(tuningFd);
#endif WIN32
			    }
			}
		    else if (msgType == SOCK_DPAREG)
			{

			/*
			Utility is registering as a DPA. Get his header and then the prefix.
			*/

			SM_DPA		dpaHdr;

			if (recvUtil(utilitySockets[fdIndex], (char *) &dpaHdr, sizeof(dpaHdr)))
			    {
			    char	prefix[16];

			    if (recvUtil(utilitySockets[fdIndex], prefix, ntohs(dpaHdr.dpaPrefixLen)))
				{

				/*
				Find this prefix in our DPA table.
				*/

				for (dpaEntry = dpaList; dpaEntry; dpaEntry = dpaEntry -> nextDPA)
				    {
				    if (!strcmp(prefix, dpaEntry -> prefix))
					{

					/*
					Found it. Get the real Inet address of the other end
					of this socket and then use that to determine which
					virtual Inet address that we should use.
					*/

					struct in_addr	realInet;
					struct in_addr	virtInet;
					CH_IOCUP	iocUp;

#if !defined(vxWorks) && !defined(vms)
					tisSysLog(1, 0, "sockMnt DPA registration for prefix %s", 
						dpaEntry -> prefix);
#else !defined(vxWorks) && !defined(vms)
					if (smDebug >= 1)
#ifdef vxWorks
					    logMsg("sockMnt DPA registration for prefix %s\n",
							(int)dpaEntry -> prefix,0,0,0,0,0);
#else vxWorks
					    printf("sockMnt DPA registration for prefix %s\n",
							dpaEntry -> prefix);
#endif vxWorks
#endif !defined(vxWorks) && !defined(vms)
					portAddrSize = sizeof(portAddr);
					getpeername(utilitySockets[fdIndex], (struct sockaddr *) &portAddr,
						&portAddrSize);
					realInet = portAddr.sin_addr;

					if (realInet.s_addr == dpaEntry -> realInet1.s_addr)
					    {
					    virtInet = dpaEntry -> virtInet1;
					    dpaEntry -> sockFd1 = utilitySockets[fdIndex];
					    }
					else if (realInet.s_addr == dpaEntry -> realInet2.s_addr)
					    {
					    virtInet = dpaEntry -> virtInet2;
					    dpaEntry -> sockFd2 = utilitySockets[fdIndex];
					    }
					
					/*
					Now post the fd for this virtual Inet and then send
					chanAcq an IOCUP transaction.
					*/

					ctlEntry = findInet(virtInet);
					ctlEntry -> fd = utilitySockets[fdIndex];
					utilitySockets[fdIndex] = INVALID_SOCKET;
					msgType = htons(CHAN_IOCUP);
					iocUp.chInet = virtInet;
					sendCac((char *) &msgType, sizeof(msgType));
					sendCac((char *) &iocUp, sizeof(iocUp));

					/*
					Condition the socket and our control table entry for
					the way we normally handle IOCs.
					*/

			    		ctlEntry -> lastRx = time(NULL);
			    		ctlEntry -> boomsSent = 0;
			    		ctlEntry -> nextBoom = 0;
			    		ctlEntry -> bufSize = 0;
			    		ctlEntry -> nchInTank = 0;
			    		ctlEntry -> tankOutInd = 0;
			    		ctlEntry -> iocDownSent = 0;

			    		/*
			    		Mark the socket for non-blocking I/O. Do this so we won't hang
			    		if he dies.
			    		*/
		
#ifndef SOLARIS
#ifdef WIN32
			    		if (ioctlsocket (ctlEntry -> fd, FIONBIO, &true))
#else WIN32
			    		if (ioctl(ctlEntry -> fd, FIONBIO, (int)&true))
#endif WIN32
#else SOLARIS
					sockGrow(ctlEntry -> fd);
					if (fcntl(ctlEntry -> fd, F_SETFL, O_NONBLOCK) == -1)
#endif SOLARIS
					    perror("Error setting non-blocking I/O for IOC");

			    		/*
			    		Get the current receive and transmit buffer sizes
			    		*/

			    		intsize = sizeof(ctlEntry -> currRxBuff);
			    		getsockopt(ctlEntry -> fd, SOL_SOCKET, SO_RCVBUF, 
				    		(char *)&(ctlEntry -> currRxBuff), &intsize);
			    		intsize = sizeof(ctlEntry -> currTxBuff);
			    		getsockopt(ctlEntry -> fd, SOL_SOCKET, SO_SNDBUF,
				    		(char *)&(ctlEntry -> currTxBuff), &intsize);
			    		ctlEntry -> rxMaxOut = 0;
					}
				    }
				}
			    }
			}
		    }
		}
	    }
	  /* If any items remain, we have an orphan fd in the mask! */
	  for (fdIndex = 0; fdIndex < maxFd; fdIndex++)
	    {
#ifndef WIN32
	    if (FD_ISSET (fdIndex, &fdList))
		{
#ifdef vxWorks
		logMsg ("Unhandled fd in select result! fd = %d\n", fdIndex,0,0,0,0,0);
#else vxWorks
		printf ("Unhandled fd in select result! fd = %d\n", fdIndex);
#endif vxWorks
		FD_CLR (fdIndex, &fdMask);
		}
#else WIN32
	    /* WIN32 test for leftovers needs to be based on SOCKET, not fd */
	    /* to be coded! */
#endif WIN32
		
	    }
	  /*
	  Recopy the mask in case we closed any of the sockets.
	  */

	  fdList = fdMask;
	  }

	/*
	If anything is still queued up for Channel Acquisition and Control, ship it out now.
	*/

	if (txBuffSize)
	    cacFlush();

	/*
	Traverse the control table if the time is appropriate.
	We do this every second.
	*/

	if ((time(NULL) - lastTable) > 0)
	    {
	    for (ctlEntry = socketList; ctlEntry; ctlEntry = ctlEntry -> nextSock)
		{
		if (ctlEntry -> fd != INVALID_SOCKET)	/* If connected */
		    {
		    if ((time(NULL) - ctlEntry -> lastTx) > ctlEntry -> tuning.stMaxIdle)
			{

			/*
			We haven't spoken to this guy in too long a time. Send him a boomerang
			for good measure.
			*/

			if ((ctlEntry -> bufSize + sizeof(short)) > ctlEntry -> tuning.stMaxBlock)
			    sendIOC(ctlEntry);
			msgType = htons(BOOMERANG);
			bcopy((char *) &msgType, ctlEntry -> bufLocn + ctlEntry -> bufSize, sizeof(msgType));
		        ctlEntry -> bufSize += sizeof(short);
			sendIOC(ctlEntry);
			ctlEntry -> lastTx = time(NULL);
			}

		    if (ctlEntry -> tuning.stMaxQuiet && 	/* Zero means no timeout! */
				(ctlEntry -> boomsSent > ctlEntry -> tuning.stMaxQuiet))
			{

			/*
			No input traffic for too long. Call him dead.
			*/

#if !defined(vxWorks) && !defined(vms)
			tisSysLog(1, 0, "sockMnt Kill %s, boomerang msgs not returned.", 
				INET_NTOA (ctlEntry -> Inet));
#else !defined(vxWorks) && !defined(vms)
			if (smDebug >= 1)
#ifdef vxWorks
			    logMsg("sockMnt Kill %s, boomerang msgs not returned.\n",
				(int)INET_NTOA (ctlEntry -> Inet),0,0,0,0,0);
#else vxWorks
			    printf("sockMnt Kill %s, boomerang msgs not returned.\n",
				INET_NTOA (ctlEntry -> Inet));
#endif vxWorks
#endif !defined(vxWorks) && !defined(vms)
			iocDown(ctlEntry);
			}
		    else if ((time(NULL) - ctlEntry -> lastRx) > ctlEntry -> tuning.stMaxIdle)
			{

			/*
			This guy has been idle too long. If so, make up a boomerang
			and send him that.
			*/

			if (ctlEntry -> nextBoom > time(NULL))
			    continue;
			ctlEntry -> boomsSent++;
			if ((ctlEntry -> bufSize + sizeof(short)) > ctlEntry -> tuning.stMaxBlock)
			    sendIOC(ctlEntry);
			msgType = htons(BOOMERANG);
			bcopy((char *) &msgType, ctlEntry -> bufLocn + ctlEntry -> bufSize, sizeof(msgType));
		        ctlEntry -> bufSize += sizeof(short);
			sendIOC(ctlEntry);
			ctlEntry -> nextBoom = time(NULL) + ctlEntry -> tuning.stMaxIdle;
			}
		    }
		else			/* Not connected */
		    {

		    /*
		    Ignore this entry if it is our local Internet address. Our name would appear
		    there if the system makes use of the workstation master/slave arbiter. We
		    should also ignore it if the Inet addr is < 256 because it is a virtual one
		    for a DPA.
		    */

#ifndef vxWorks
		    if ((ctlEntry -> Inet.s_addr == caLocalInet.s_addr) || (ntohl(ctlEntry -> Inet.s_addr) < 256))
			continue;
#endif vxWorks

		    /*
		    If the time is appropriate, invite this guy to connect to us.
		    */

		    if ((time(NULL) - ctlEntry -> lastConn) > ctlEntry -> tuning.stInqRate)
			{
#ifdef USE_UDP
			portAddr.sin_family = AF_INET;
			portAddr.sin_addr = ctlEntry -> Inet;
			portAddr.sin_port = htons(RSV_PORT);
			msgType = htons(SOCK_CONN_REPLY);
			sendto(udpFd, (char *) &msgType, sizeof(msgType), 0, (struct sockaddr *)&portAddr, sizeof(portAddr));
#else  USE_UDP
			/* connect to target, then close - this will announce our needs */
			kick_TCP (ctlEntry -> Inet.s_addr, RSV_PORT);
#endif USE_UDP
#if !defined(vxWorks) && !defined(vms)
			tisSysLog(1, 0, "sockMnt invites %s to connect", INET_NTOA(ctlEntry -> Inet));
#else !defined(vxWorks) && !defined(vms)
			if (smDebug >= 1)
#ifdef vxWorks
			    logMsg ("sock invites %s to connect\n", (int)INET_NTOA(ctlEntry -> Inet),0,0,0,0,0);
#else vxWorks
			    printf("sock invites %s to connect\n", INET_NTOA(ctlEntry -> Inet));
#endif vxWorks
#endif !defined(vxWorks) && !defined(vms)
			ctlEntry -> lastConn = time(NULL);
			}
		    }
		}

	    /*
	    If we don't have a connection to the name server, try to find one.
	    */

#if defined(vxWorks) || defined(vms)
	    if (needNameSrv)
		{
#ifdef USE_UDP
#ifdef vxWorks
		/*
		First try for local Name Server.  Handled in chanAcq!
		*/
#endif vxWorks

		/*
		Send a probe to each name server in our list.
		*/

		msgType = htons(NAME_PROBE);
		portAddr.sin_family = AF_INET;
		portAddr.sin_port = htons(NAM_PORT);
		portAddrSize = sizeof(portAddr);
		if (smDebug)
#ifdef vxWorks
		    logMsg ("sock sends name server probes\n", 0,0,0,0,0,0);
#else vxWorks
		    printf("sock sends name server probes\n");
#endif vxWorks
		for (fdIndex = 0; fdIndex < numNames; fdIndex++)
		    {
		    portAddr.sin_addr = namAddrs[fdIndex];
		    sendto(udpFd, (char *) &msgType, sizeof(msgType), 0, (struct sockaddr *)&portAddr, sizeof(portAddr));
		    if (smDebug)
#ifdef vxWorks
			logMsg ("\t%s\n", (int)INET_NTOA(namAddrs[fdIndex]),0,0,0,0,0);
#else vxWorks
			printf("\t%s\n", INET_NTOA(namAddrs[fdIndex]));
#endif vxWorks
		    }
#else USE_UDP
		CH_GOTNMSRV		gotNmsrv;
		int	dpaSocket;

#ifdef vxWorks
		logMsg ("sock attempting connects to name servers\n", 0,0,0,0,0,0);
#else vxWorks
		printf("sock attempting connects to name servers\n");
#endif vxWorks
		if ((dpaSocket = socket(PF_INET, SOCK_STREAM, 0)) < 0)
		    continue;
		/* else try to connect to a name server */
		for (fdIndex = 0; fdIndex < numNames && needNameSrv; fdIndex++)
		    {
		    portAddr.sin_addr = namAddrs[fdIndex];
#ifdef vxWorks
		    logMsg ("\t%s\n", (int)INET_NTOA(namAddrs[fdIndex]),0,0,0,0,0);
#else vxWorks
		    printf("\t%s\n", INET_NTOA(namAddrs[fdIndex]));
#endif vxWorks
		    if (connect(dpaSocket, (struct sockaddr *) &portAddr, sizeof(portAddr)) == -1)
			{
			close(dpaSocket);
			continue;
			}
		    if (getDpaList(dpaSocket))
			{
			dpaListBuilt = 1;
			}
		    close(dpaSocket);
		    needNameSrv = 0;
		    gotNmsrv.chInet = portAddr.sin_addr;
		    if (smDebug >= 1)
			{
#ifdef vxWorks
			logMsg("sock connected to name server %s\n",
				(int)INET_NTOA(gotNmsrv.chInet),0,0,0,0,0);
#else vxWorks
			printf("sock connected to name server %s\n",
				INET_NTOA(gotNmsrv.chInet));
#endif vxWorks
			}
		    }
		if (!needNameSrv)
		    {
		    msgType = CHAN_GOTNMSRV;
		    sendCac((char *) &msgType, sizeof(msgType));
		    sendCac((char *) &gotNmsrv, sizeof(gotNmsrv));
		    }
#endif USE_UDP
		}
#endif defined(vxWorks) || defined(vms)

	    /*
	    If we are missing any DPA connections, request them from their sockMnt
	    */

	    if ((time(NULL) - lastDPA) > 30)
		{
#ifdef USE_UDP
		struct
		    {
		    short	msgType;
		    SM_DPAREQ	dpaReq;
		    } dpaReqMsg;
#endif USE_UDP

		for (dpaEntry = dpaList; dpaEntry; dpaEntry = dpaEntry -> nextDPA)
		    {

		    /*
		    If nobody wanted this DPA, ignore him!
		    */

		    if (!(dpaEntry -> wanted))
			continue;

		    /*
		    If we're not connected, go ahead and request it.
		    */

		    if (dpaEntry -> realInet1.s_addr)
			if (dpaEntry -> sockFd1 <= 0)
			    if (dpaEntry -> realInet1.s_addr != htonl(0x7f000001))
				{
#ifdef USE_UDP
				dpaReqMsg.msgType = htons(SOCK_DPAREQ);
				strcpy(dpaReqMsg.dpaReq.dpaPrefix, dpaEntry -> prefix);
				portAddr.sin_family = AF_INET;
				portAddr.sin_addr = dpaEntry -> realInet1;
				portAddr.sin_port = htons(SCM_PORT);
				sendto(udpFd, (char *) &dpaReqMsg, sizeof(dpaReqMsg), 0, (struct sockaddr *)
					&portAddr, sizeof(portAddr));
#else USE_UDP
				kick_TCP (dpaEntry -> realInet1.s_addr, SCM_PORT);
#endif USE_UDP
				}
		    if (dpaEntry -> realInet2.s_addr)
			if (dpaEntry -> sockFd2 <= 0)
			    if (dpaEntry -> realInet2.s_addr != htonl(0x7f000001))
				{
#ifdef USE_UDP
				dpaReqMsg.msgType = htons(SOCK_DPAREQ);
				strcpy(dpaReqMsg.dpaReq.dpaPrefix, dpaEntry -> prefix);
				portAddr.sin_family = AF_INET;
				portAddr.sin_addr = dpaEntry -> realInet2;
				portAddr.sin_port = htons(SCM_PORT);
				sendto(udpFd, (char *) &dpaReqMsg, sizeof(dpaReqMsg), 0, (struct sockaddr *)
					&portAddr, sizeof(portAddr));
#else USE_UDP
				kick_TCP (dpaEntry -> realInet2.s_addr, SCM_PORT);
#endif USE_UDP
				}
		    }
		lastDPA = time(NULL);
		}

#ifdef vxWorks
	    {
	    struct in_addr partners[10];
	    int		numPartners = 0;
	    int		index;
	    RES_TABLE	*currRes;

	    /*
	    Find redundant partners for alarm queue maintenance for each resource
	    that needs one.
	    */

	    if ((time(NULL) - lastAlq) > 30)
		{
	        for (currRes = resTableHead; currRes; currRes = currRes -> link)
		    {
		    if (!(currRes -> partnerInet.s_addr))
		        continue;
		    if (currRes -> redAlarmFd)
		        continue;
		
		    /*
		    This one needs a partner. See if we already tried the same partner this pass.
		    */

		    for (index = 0; index < numPartners; index++)
		        if (currRes -> partnerInet.s_addr == partners[index].s_addr)
			    break;
		    if (index < numPartners)
		        continue;
		
		    /*
		    This one is new. Remember his Inet addr and send him the message.
		    */

		    partners[numPartners] = currRes -> partnerInet;
		    numPartners++;

		    if (smDebug >= 1)
#ifdef vxWorks
			logMsg ("Sending ALQCONN to %s\n", (int)INET_NTOA(currRes -> partnerInet),0,0,0,0,0);
#else vxWorks
			printf("Sending ALQCONN to %s\n", INET_NTOA(currRes -> partnerInet));
#endif vxWorks
		    msgType = htons(SOCK_ALQCONN);
		    portAddr.sin_family = AF_INET;
		    portAddr.sin_addr = currRes -> partnerInet;
		    portAddr.sin_port = htons(RSV_PORT);
#if 0
 /* Hmmm - looks like this is turned off!! */
		    if (sendto(udpFd, (char *) &msgType, sizeof(msgType), 0, &portAddr, sizeof(portAddr))
			    == ERROR)
			{
			perror("sockMnt sending ALQCONN");
			}
#endif 0
		    }
		lastAlq = time(NULL);
	        }
	    }
#endif vxWorks
	    lastTable = time(NULL);
	    }
	}
    return 0;
}


/****************************************************************/
/*								*/
/*	FIND CONTROL TABLE ENTRY GIVEN FILE DESCRIPTOR		*/
/*								*/
/****************************************************************/

struct sockCtl *findFd(fd)
SOCKET	fd;
{
    struct sockCtl	*currNode;

    for (currNode = socketList; currNode; currNode = currNode -> nextSock)
	{
	if (currNode -> fd == fd)
		return (currNode);
	}
    
    return ((struct sockCtl *) 0);
}


/****************************************************************/
/*								*/
/*	DETERMINE IF AN INTERNET ADDRESS IS KNOWN		*/
/*								*/
/*	0 = no, 1 = yes						*/
/*								*/
/****************************************************************/

int knownIOC(targetAddr)
struct in_addr	targetAddr;
{
    struct sockCtl	*currNode;

    currNode = socketList;
    while (currNode)
	{
	if (currNode -> Inet.s_addr == targetAddr.s_addr)
	    return (1);
	currNode = currNode -> nextSock;
	}
    return (0);
}

/****************************************************************/
/*								*/
/*	FIND CONTROL TABLE ENTRY GIVEN INTERNET ADDRESS		*/
/*								*/
/*	Create a new entry if we don't find the search target.	*/
/*								*/
/****************************************************************/

struct sockCtl *findInet(targetAddr)
struct in_addr	targetAddr;
{
    struct sockCtl	*currNode;

    currNode = socketList;
    if (!currNode)
	{
	currNode = socketList = (struct sockCtl *) calloc(1, sizeof(struct sockCtl));
	socketList -> Inet = targetAddr;
        socketList -> tuning = defTuning;
	socketList -> fd = INVALID_SOCKET;
#ifndef PRIV_AUDIT
	/* Dont do slow net timeout for CAISO */
#ifdef vxWorks
	socketList -> tuning.stMaxQuiet = 6;
	socketList -> tuning.stMaxIdle = 45;
#endif vxWorks
#endif PRIV_AUDIT
	socketList -> bufLocn = (char *) malloc(MAX_XMIT_BUFFER);
	socketList -> rxTank = (char *) malloc(MAX_XMIT_BUFFER);
	numIOCs++;
	return (currNode);
	}

    while (currNode)
	{
	if (currNode -> Inet.s_addr == targetAddr.s_addr)
	    return(currNode);
	if (currNode -> nextSock)
	    currNode = currNode -> nextSock;
	else
	    {
	    currNode -> nextSock = (struct sockCtl *) calloc(1, sizeof(struct sockCtl));
	    currNode -> nextSock -> Inet = targetAddr;
	    currNode -> nextSock -> tuning = defTuning;
	    currNode -> nextSock -> fd = INVALID_SOCKET;
#ifndef PRIV_AUDIT
	    /* Dont do slow net timeout for CAISO */
#ifdef vxWorks
	    currNode -> nextSock -> tuning.stMaxQuiet = 6;
	    currNode -> nextSock -> tuning.stMaxIdle = 45;
#endif vxWorks
#endif PRIV_AUDIT
	    currNode -> nextSock -> bufLocn = (char *) malloc(MAX_XMIT_BUFFER);
	    currNode -> nextSock -> rxTank = (char *) malloc(MAX_XMIT_BUFFER);
	    numIOCs++;
	    return (currNode -> nextSock);
	    }
	}
    return 0;
}


/****************************************************************/
/*								*/
/*	SEND MESSAGE TO CHANNEL ACQUISITION AND CONTROL		*/
/*								*/
/*	Just buffer stuff up. It will get flushed when all the	*/
/*	sockets have been drained.				*/
/*								*/
/****************************************************************/

void sendCac(buff, nbytes)
char	*buff;
int	nbytes;
{
    if ((txBuffSize + nbytes) > sizeof(txBuff))
	cacFlush();
    bcopy(buff, txBuff + txBuffSize, nbytes);
    txBuffSize += nbytes;
}


/****************************************************************/
/*								*/
/*	READ MESSAGE FROM CHANNEL ACQUISITION AND CONTROL	*/
/*								*/
/****************************************************************/

void recvCac(buff, nbytes)
char	*buff;
int	nbytes;
{
    /*
    We don't have to worry about not having enough data in our
    buffer. haveCompleteChanMsg() has done that for us.
    */

    bcopy(rxTank + tankPos, buff, nbytes);
    nchib -= nbytes;
    tankPos += nbytes;
}


/****************************************************************/
/*								*/
/*		SEND A MESSAGE TO AN IOC			*/
/*								*/
/*	It is very important that we never ever hang when we	*/
/*	try to write to an IOC. Therefore, all sockets to the	*/
/*	IOCs have been marked as non blocking. If we ever get	*/
/*	a blocking error, then we will try to increase the	*/
/*	size of the transmit buffer. If we can't then we will	*/
/*	kill the IOC because if he's that far behind then	*/
/*	he's probably down.					*/
/*								*/
/****************************************************************/

#define	NOMINAL_RETRY	250000		/* In microseconds */

void sendIOC(ctlEntry)
struct sockCtl	*ctlEntry;
{
    int		status;
#ifdef PRIV_AUDIT
    SSL		*conn;
#endif PRIV_AUDIT

  if (ctlEntry -> fd == INVALID_SOCKET)
    {
    ctlEntry -> bufSize = 0;
    return;
    }
  if (ctlEntry -> bufSize == 0)
    return;

  while (ctlEntry -> bufSize)
    {
#ifdef PRIV_AUDIT
    conn = sslSharedGetConn (ctlEntry -> fd);
    status = SSL_write (conn, ctlEntry -> bufLocn, ctlEntry -> bufSize);
#else PRIV_AUDIT
    status = send(ctlEntry -> fd, ctlEntry -> bufLocn, ctlEntry -> bufSize, 0);
#endif PRIV_AUDIT
    ctlEntry -> lastTx = time(NULL);

    /*
    Delay here at some nomimal baud rate if that was set up for simulation
    purposes.
    */

#ifndef vxWorks
    if (charsPerSecond)
	{
	fd_set		delaySet;
	struct timeval	delayVal;

	delayVal.tv_sec = ctlEntry -> bufSize / charsPerSecond;
	delayVal.tv_usec = ctlEntry -> bufSize - delayVal.tv_sec * charsPerSecond;
	delayVal.tv_usec =  (long) (((double)delayVal.tv_usec * 1000000.0) / ((double) charsPerSecond));
        FD_ZERO(&delaySet);

	select(1, &delaySet, NULL, NULL, &delayVal);
	}
#endif vxWorks



    if (smDebug)
#ifdef vxWorks
	logMsg ("sock sent %d bytes to %s\n", status, (int)INET_NTOA(ctlEntry -> Inet),0,0,0,0);
#else vxWorks
	printf("sock sent %d bytes to %s\n", status, INET_NTOA(ctlEntry -> Inet));
#endif vxWorks
    if (status == -1)
      {
#ifdef WIN32
	errno = WSAGetLastError ();
	if ((errno == WSAEWOULDBLOCK) || (errno == EAGAIN))
#else WIN32
	if ((errno == EWOULDBLOCK) || (errno == EAGAIN))
#endif WIN32
	{
	ctlEntry -> currTxBuff += ctlEntry -> bufSize;
	if (setsockopt(ctlEntry -> fd, SOL_SOCKET, SO_SNDBUF, (char *)&(ctlEntry -> currTxBuff), 
		sizeof(ctlEntry -> currTxBuff)) == -1)
	    {
#if !defined(vxWorks) && !defined(vms)
	    tisSysLog(1, 0, "sockMnt Kill because of maxout on TX socket to %s\n", 
		    INET_NTOA (ctlEntry -> Inet));
#else !defined(vxWorks) && !defined(vms)
	    if (smDebug >= 1)
#ifdef vxWorks
		logMsg ("Kill because of maxout on TX socket to %s\n",
		    (int)INET_NTOA (ctlEntry -> Inet),0,0,0,0,0);
#else vxWorks
		printf("Kill because of maxout on TX socket to %s\n",
		    INET_NTOA (ctlEntry -> Inet));
#endif vxWorks
#endif !defined(vxWorks) && !defined(vms)
	    iocDown(ctlEntry);
	    return;
	    }
	else
	    ;	/* Don't print this message any more */

	/*
	Now try to send out what we couldn't send on our original attempt
	*/

	continue;
	}
      else
	{
#if !defined(vxWorks) && !defined(vms)
	tisSysLog(1, 0, "sockMnt kill %s because of error on first send: %m",
	    INET_NTOA (ctlEntry -> Inet));
#else !defined(vxWorks) && !defined(vms)
	if (smDebug >= 1)
#ifdef vxWorks
	    logMsg ("Kill %s because of error on first send\n", (int)INET_NTOA (ctlEntry -> Inet),0,0,0,0,0);
#else vxWorks
	    printf("Kill %s because of error on first send", INET_NTOA (ctlEntry -> Inet));
#endif vxWorks
#endif !defined(vxWorks) && !defined(vms)
	iocDown(ctlEntry);
	return;
	}
      }
    else if (status != ctlEntry -> bufSize)
	{
	bcopy(ctlEntry -> bufLocn + status, ctlEntry -> bufLocn, ctlEntry -> bufSize);
	}
    ctlEntry -> bufSize -= status;

#ifndef vxWorks
    if (ctlEntry -> Inet.s_addr == traceInet.s_addr)
	dumpBlock("Send", ctlEntry -> bufLocn, status);
#endif vxWorks
    }
  ctlEntry -> bufSize = 0;
}


/****************************************************************/
/*								*/
/*		DECLARE AN IOC DOWN				*/
/*								*/
/****************************************************************/

void iocDown(ctlEntry)
struct sockCtl	*ctlEntry;
{
    CH_IOCDN	downMsg;
    short	msgType;
#ifdef PRIV_AUDIT
    SSL		*conn;
#endif PRIV_AUDIT

    if (ctlEntry -> fd == 0)
    {
#ifdef vxWorks
	logMsg ("Bad fd = 0",0,0,0,0,0,0);
#else vxWorks
	printf ("Bad fd = 0");
#endif vxWorks
	return;
    }
	
    if (ctlEntry ->fd == INVALID_SOCKET)
	return;

    if (ctlEntry -> iocDownSent == 0)
	{
    	msgType = CHAN_IOCDN;
    	sendCac((char *) &msgType, sizeof(msgType));
    	downMsg.chInet = ctlEntry -> Inet;
        if (downMsg.chInet.s_addr == htonl(0x7f000001))
	    downMsg.chInet = caLocalInet;
    	sendCac((char *) &downMsg, sizeof(downMsg));
    	ctlEntry -> iocDownSent = 1;
	}
#ifdef PRIV_AUDIT
    conn = sslSharedGetConn (ctlEntry -> fd);
    if (auditLog)
    {
	GetPeerDN (conn, X509_dn, sizeof(X509_dn));
#ifdef vxWorks
	AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
		taskName(taskIdSelf()), "Connection closed");
#else vxWorks
	AuditAddEvent (auditLog, AUDIT_LOGOFF, AUDIT_SUCCESS, X509_dn,
		"sockMnt", "Connection closed");
#endif vxWorks
    }
    sslSharedCloseConn (conn);
#else PRIV_AUDIT
    close(ctlEntry -> fd);
#endif PRIV_AUDIT
    /*
    If this was a socket to a DPA, then also clear the fd in that
    table so we can determine that we no longer have a connection.
    */

    if (ntohl(ctlEntry -> Inet.s_addr) < 256)
	{
    	struct dpaCtl	*dpaEntry;

	for (dpaEntry = dpaList; dpaEntry; dpaEntry = dpaEntry -> nextDPA)
	    {
	    if (dpaEntry -> sockFd1 == ctlEntry -> fd)
		{
		dpaEntry -> sockFd1 = INVALID_SOCKET;
		break;
		}
	    else if (dpaEntry -> sockFd2 == ctlEntry -> fd)
		{
		dpaEntry -> sockFd2 = INVALID_SOCKET;
		break;
		}
	    }
	}

    /*
    Now finish cleaning up after the socket we just closed.
    */

    FD_CLR(ctlEntry -> fd, &fdMask);
    ctlEntry -> fd = INVALID_SOCKET;
    ctlEntry -> bufSize = 0;
    ctlEntry -> nchInTank = 0;
    ctlEntry -> tankOutInd = 0;
}


/****************************************************************/
/*								*/
/*		RECEIVE A MESSAGE FROM AN IOC			*/
/*								*/
/*	Return zero on failure, 1 on success			*/
/*								*/
/****************************************************************/

int recvIOC(ctlEntry, buff, length)
struct sockCtl	*ctlEntry;
char		*buff;
int		length;
{
    /*
    Because the buffer has been prequalified, we will always have
    the amount of data that the user wants!
    */

    bcopy(ctlEntry -> rxTank + ctlEntry -> tankOutInd, buff, length);
    ctlEntry -> tankOutInd += length;
    ctlEntry -> nchInTank -= length;
    return (1);
}


/****************************************************************/
/*								*/
/*		RECEIVE MESSAGE FROM A UTILITY			*/
/*								*/
/*	Return a 1 on success, 0 on failure.			*/
/*								*/
/****************************************************************/

int recvUtil(fd, buff, nbytes)
SOCKET	fd;
char	*buff;
int	nbytes;
{
    int	numRead;
    fd_set fdList;

    FD_ZERO(&fdList);
    FD_SET(fd, &fdList);

    while (nbytes)
	{
	if (select(fd + 1, &fdList, NULL, NULL, &oneSecond) <= 0)
	    break;
	if ((numRead = recv(fd, buff, nbytes, 0)) <= 0)
	    break;
	nbytes -= numRead;
	buff += numRead;
	}

    if (nbytes)
	{
	/* this close will only close registered utility sockets */
	closeUtil (fd);
	return (0);
	}
    return (1);
}


/****************************************************************/
/*								*/
/*		SEND A MESSAGE TO A UTILITY			*/
/*								*/
/****************************************************************/

int sendUtil(fd, buff, nbytes)
SOCKET	fd;
char	*buff;
int	nbytes;
{
    if (send(fd, buff, nbytes, 0) == -1)
	return (0);
    else
	return (-1);
}

void closeUtil (SOCKET sock)
{
    int fdIndex;
    
    for (fdIndex = 0; fdIndex < MAX_UTIL_CLIENTS; fdIndex++)
    {
	if (utilitySockets [fdIndex] == sock)
	    break;
    }
    if (fdIndex < MAX_UTIL_CLIENTS)
    {
#ifdef WIN32
	closesocket (utilitySockets [fdIndex]);
#else WIN32
	close(utilitySockets [fdIndex]);
#endif WIN32
	FD_CLR (utilitySockets [fdIndex], &fdMask);
	utilitySockets [fdIndex] = INVALID_SOCKET;
    }
}

/****************************************************************/
/*								*/
/*		RECEIVE FLOW CONTROL FOR TCP SOCKETS		*/
/*								*/
/*	This routine is to be called whenever select() says	*/
/*	that the socket is ready for reading. What we do is	*/
/*	examine the number of bytes in the receive buffer. If	*/
/*	that is equal to the size of the receive buffer, then	*/
/*	we are in danger of the "deadly embrace" or deadlock.	*/
/*	In that case, try to increase the receive buffer size	*/
/*	by 1K bytes. After we've topped out, there is nothing	*/
/*	more that we can do at the current time.		*/
/*								*/
/****************************************************************/

void rxFlowControl(ctlEntry)
struct sockCtl	*ctlEntry;
{
    int		readSize;

    /*
    If we've already maxed out, there's nothing more we can do.
    */

    if (ctlEntry -> rxMaxOut)
	return;

    /*
    Get the number of bytes that are in our input buffer. If its
    not filled to the brim, don't do anything.
    */

#ifdef WIN32
    if (ioctlsocket (ctlEntry -> fd, FIONREAD, &readSize))
#else WIN32
    if (ioctl(ctlEntry -> fd, FIONREAD, (int)&readSize))
#endif WIN32
	return;

    /*
    If the buffer is full, try to bump it up by 1K
    */

    if (readSize < ctlEntry -> currRxBuff)
	return;
    ctlEntry -> currRxBuff += 1024;
    if(setsockopt(ctlEntry -> fd, SOL_SOCKET, SO_RCVBUF, (char *)&(ctlEntry -> currRxBuff), 
	    sizeof(ctlEntry -> currRxBuff)) == -1)
	{
	ctlEntry -> rxMaxOut = 1;
	return;
	}
}

/****************************************************************/
/*								*/
/*		FLUSH THE BUFFER TO CHANNEL ACQUISITION		*/
/*								*/
/*	We dont't want to hang if chanAcq is too busy to keep	*/
/*	up with us. If the I/O can't go out without blocking	*/
/*	us, then we should increase the size of the transmit	*/
/*	buffer by the number of bytes that don't go out. When	*/
/*	the operating system tells us we've hit the limit,	*/
/*	abandon ship because chanAcq is probably permanently	*/
/*	hung.							*/
/*								*/
/****************************************************************/

void cacFlush()
{
    int		bytesSent;

    while (txBuffSize)
    {
#ifndef vxWorks
	if (scmExtBytes)
	{
	    bytesSent = -1;
#ifdef WIN32
	    errno = WSAEWOULDBLOCK;
#else WIN32
	    errno = EWOULDBLOCK;
#endif WIN32
	}
	else
#endif vxWorks
	{
            bytesSent = send(cacFd, txBuff, txBuffSize, 0);
#ifdef WIN32
	    if (bytesSent == -1)
		errno = WSAGetLastError ();
#endif WIN32
	}

	if (bytesSent == -1)
	{
#ifdef WIN32
	    if ((errno == WSAEWOULDBLOCK) || (errno == EAGAIN))
#else WIN32
	    if ((errno == EWOULDBLOCK) || (errno == EAGAIN))
#endif WIN32
	    {
		if (cacTxTopped)
		{
#ifndef vxWorks
		    char	*outBuff = txBuff;

		    if ((txBuffSize + scmExtBytes) > sizeof(scmMobyBuff))
		    {
#ifndef vms
			tisSysLog(0, 0, "sockMnt quits because of moby buffer overflow");
#ifndef WIN32
			kill(cacPid, 3);
#endif WIN32
#endif vms
			exit(1);
		    }
		    scmExtBytes += txBuffSize;
		    while (txBuffSize > 0)
		    {
			bytesSent = txBuffSize;
			if ((scmRingIn + bytesSent) > sizeof(scmMobyBuff))
			{
			    bytesSent = sizeof(scmMobyBuff) - scmRingIn;
			}
			bcopy(outBuff, scmMobyBuff + scmRingIn, bytesSent);
			scmRingIn += bytesSent;
			if (scmRingIn >= sizeof(scmMobyBuff))
			{
			    scmRingIn = 0;
			}
			outBuff += bytesSent;
			txBuffSize -= bytesSent;
		    }
		    return;

#else vxWorks
		    logMsg ("Quit because of logjam to chanAcq\n", 0,0,0,0,0,0);
		    taskSuspend(taskIdSelf());	/* For IOC debugging help */
#endif vxWorks
		}
		cacTxSize += txBuffSize;
		if (setsockopt(cacFd, SOL_SOCKET, SO_SNDBUF, (char *)&cacTxSize, sizeof(cacTxSize)) == -1)
		{
		    cacTxTopped = 1;
		}
		continue;
	    }
	    else
	    {
#ifndef vxWorks
#ifndef vms
		tisSysLog(0, 0, "sockMnt quits because of error sending to chanAcq: %m");
#ifndef WIN32
		kill(cacPid, 3);
#endif WIN32
#endif vms
#else vxWorks
		perror("Error sending to chanAcq");
		taskSuspend(taskIdSelf());	/* For IOC debugging help */
#endif vxWorks
		exit(1);
	    }
	}
	else if (bytesSent != txBuffSize)
	{
	    bcopy(txBuff + bytesSent, txBuff, (txBuffSize - bytesSent));
	}
	txBuffSize -= bytesSent;
    }
}

/****************************************************************/
/*								*/
/*		MANAGE RECEIVE BUFFER SIZE IN CAC SOCKET	*/
/*								*/
/*	We never want to hang up the people that are sending	*/
/*	messages to us. Therefore, if the socket's receive	*/
/*	buffer is completely full any time that select() tells	*/
/*	us that input is present, we should make the buffer 1K	*/
/*	bigger. We keep doing this until the operating system	*/
/*	won't let us do it anymore.				*/
/*								*/
/****************************************************************/

void cacFlowControl()
{
    int		readSize;

    if (cacRxTopped)
	return;
#ifdef WIN32
    ioctlsocket (cacFd, FIONREAD, &readSize);
#else WIN32
    ioctl(cacFd, FIONREAD, (int)&readSize);
#endif WIN32
    if (readSize == cacRxSize)
	{
	cacRxSize += 1024;
	if (setsockopt(cacFd, SOL_SOCKET, SO_SNDBUF, (char *)&cacRxSize, sizeof(cacRxSize)) == -1)
	    cacRxTopped = 1;
	}
}

/****************************************************************/
/*								*/
/*		READ INPUT BUFFER FROM chanAcq			*/
/*								*/
/*	Return a 1 if ok, zero on timeout or error.		*/
/*								*/
/****************************************************************/

int fillChanBuffer()
{
    int		bytesRead;

    /*
    If there's still stuff in the buffer, slide it down to the front
    */

    if (nchib && tankPos)
	bcopy(rxTank + tankPos, rxTank, nchib);

    /*
    Read as much out of the socket as we can. This reduces the number
    of times that we use system services. This makes a BIG difference.
    */

    bytesRead = recv(cacFd, rxTank + nchib, sizeof(rxTank) - nchib, 0);
    if (bytesRead <= 0)
	{
        if (!bytesRead)
	    {
#ifndef vxWorks
#ifndef vms
	    tisSysLog(0, 0, "sockMnt got zero read from chanAcq");
	    abort();
#endif vms
#else vxWorks
	    logMsg ("Zero length receive from chanAcq\n", 0,0,0,0,0,0);
#endif vxWorks
	    }
        else
	    {
#ifndef vxWorks
#ifndef vms
	    tisSysLog(0, 0, "sockMnt got read error from chanAcq: %m");
#ifndef WIN32
	    kill(cacPid, 3);
#endif WIN32
#endif vms
#else vxWorks
	    perror("Read error from chanAcq");
	    taskSuspend(taskIdSelf());	/* For IOC debugging help */
#endif vxWorks
	    }
        exit(1);
	}
    tankPos = 0;
    nchib += bytesRead;
    return (1);
}

/****************************************************************/
/*								*/
/*	SEND AS MUCH DATA AS POSSIBLE OUT OF THE MOBY BUFFER	*/
/*								*/
/****************************************************************/

#ifndef vxWorks
void scmMobySend()
{
    int	writeSize;
    int	bytesSent;
    
    writeSize = scmExtBytes;
    if (writeSize > 4096)
        writeSize = 4096;
    if ((writeSize + scmRingOut) > sizeof(scmMobyBuff))
        writeSize = sizeof(scmMobyBuff) - scmRingOut;
    bytesSent = send(cacFd, scmMobyBuff + scmRingOut, writeSize, 0);
    if (bytesSent == -1)
        {
	if (errno == EAGAIN)
	    {
#ifdef WIN32
	    Sleep(1000);
#else WIN32
	    sleep(1);
#endif WIN32
	    return;
	    }
#ifndef vms
        tisSysLog(0, 0, "sockMnt quits: I/O fail on catch up: %m");
#ifndef WIN32
	    kill(cacPid, 3);
#endif WIN32
	abort();
#endif vms
        exit(1);
        }
    scmExtBytes -= bytesSent;
    if (scmExtBytes > 0)
        {
        scmRingOut += bytesSent;
        if (scmRingOut >= sizeof(scmMobyBuff))
    	    scmRingOut = 0;
        }
    else if (scmExtBytes == 0)
        {
        scmRingOut = 0;
        scmRingIn = 0;
        }
    else
        {
#ifndef vms
        tisSysLog(0, 0, "sockMnt quits: bookkeeping error!");
#ifndef WIN32
	    kill(cacPid, 3);
#endif WIN32
        abort();
#endif vms
	exit(1);
        }
}
#endif vxWorks

/****************************************************************/
/*								*/
/*	DETERMINE IF THE BUFFER CONTAINS A COMPLETE IOC MSG	*/
/*								*/
/****************************************************************/

int haveCompleteIocMsg(ctlEntry)
struct sockCtl	*ctlEntry;
{
    short	readBytes;
    SM_READ	*smRead;

    /*
    See if there's enough data for a message type
    */

    if (ctlEntry -> nchInTank < sizeof(short))
	return (0);

    /*
    Get the message type and see if it is a boomerang. If so, there is a message.
    The upper layer will ignore it, so we can too. However, look even further
    ahead to make sure that we don't give up on the buffer too early.
    */

    bcopy(ctlEntry -> rxTank + ctlEntry -> tankOutInd, (char *) &readBytes, sizeof(readBytes));
    readBytes = ntohs(readBytes);
    if (readBytes == BOOMERANG)
	{
	return (1);
	}


    /*
    It must be a SOCK_READ message.  See if there's enough for the header.
    */

    if (ctlEntry -> nchInTank < (sizeof(short) + sizeof(SM_READ)))
	return (0);

    /*
    See if the number of bytes specified in the header is also
    in the buffer.
    */

    smRead = (SM_READ *)(ctlEntry -> rxTank + ctlEntry -> tankOutInd + sizeof(short));
    bcopy((char *) &(smRead -> smLength), (char *) &readBytes, sizeof(readBytes));
    readBytes = ntohs(readBytes);
    readBytes += (sizeof(short) + sizeof(SM_READ));
    if (readBytes > ctlEntry -> nchInTank)
	return (0);
    else
	return (1);
}

/****************************************************************/
/*								*/
/*		READ DATA FROM AN IOC				*/
/*								*/
/****************************************************************/

int getIocData(ctlEntry)
struct sockCtl	*ctlEntry;
{
    int		bytesGot;
#ifdef PRIV_AUDIT
    SSL		*conn;
#endif PRIV_AUDIT

    /*
    If there's something already in the tank, slide it down
    */

    if (ctlEntry -> nchInTank && ctlEntry -> tankOutInd)
	bcopy(ctlEntry -> rxTank + ctlEntry -> tankOutInd, ctlEntry -> rxTank,
		ctlEntry -> nchInTank);

    /*
    Read as much as we can from the physical socket.
    */

#ifdef PRIV_AUDIT
    conn = sslSharedGetConn (ctlEntry -> fd);
    bytesGot = SSL_read (conn, ctlEntry -> rxTank + ctlEntry -> nchInTank,
			MAX_XMIT_BUFFER - ctlEntry -> nchInTank);
    if (bytesGot <= 0)
    {
	int status = SSL_get_error (conn, bytesGot);
	long err = ERR_get_error ();
	if (err != 0)
	    logMsg ("%s\n", (int)ERR_error_string (err, NULL),0,0,0,0,0);
	switch (status)
	{
	    case SSL_ERROR_WANT_WRITE:
	    case SSL_ERROR_NONE:	
	    case SSL_ERROR_WANT_READ:
	    case SSL_ERROR_WANT_CONNECT:
	    case SSL_ERROR_WANT_X509_LOOKUP:
		/* benign, just want to write/read for state machine */
		/* return OK to caller, no new data */
		return (1);
		break;
	    case SSL_ERROR_SSL:
		if (errno == EWOULDBLOCK)
		{
		    /* benign, just want to write/read for state machine */
		    /* return OK to caller, no new data */
		    return (1);
		}
		break;
	    case SSL_ERROR_SYSCALL:
	    case SSL_ERROR_ZERO_RETURN:
	    default:
		/* default error handling below */
		break;
	}
    }
#else PRIV_AUDIT
    bytesGot = recv(ctlEntry -> fd, ctlEntry -> rxTank + ctlEntry -> nchInTank,
			MAX_XMIT_BUFFER - ctlEntry -> nchInTank, 0);
#endif PRIV_AUDIT

    /*
    Delay here if we are simulating serial links.
    */

#ifndef vxWorks
    if (charsPerSecond)
	{
	fd_set		delaySet;
	struct timeval	delayVal;

	delayVal.tv_sec = bytesGot / charsPerSecond;
	delayVal.tv_usec = bytesGot - delayVal.tv_sec * charsPerSecond;
	delayVal.tv_usec =  (long) (((double)delayVal.tv_usec * 1000000.0) / ((double) charsPerSecond));
        FD_ZERO(&delaySet);

	select(1, &delaySet, NULL, NULL, &delayVal);
	}

    if (ctlEntry -> Inet.s_addr == traceInet.s_addr)
	dumpBlock("Recv", ctlEntry -> rxTank + ctlEntry -> nchInTank, bytesGot);
#endif vxWorks
    if (bytesGot == -1)
	{
#if !defined(vxWorks) && !defined(vms)
	tisSysLog(1, 0, "sockMnt I/O error from %s: %m", INET_NTOA(ctlEntry -> Inet));
#else !defined(vxWorks) && !defined(vms)
	perror("IOC socket error");
#endif !defined(vxWorks) && !defined(vms)
	return (0);
	}
    else if (bytesGot == 0)
	{
	return (0);
	}
    ctlEntry -> nchInTank += bytesGot;
    ctlEntry -> tankOutInd = 0;

    /*
    Also do the required bookkeeping
    */

    ctlEntry -> boomsSent = 0;
    ctlEntry -> nextBoom = 0;
    ctlEntry -> lastRx = time(NULL);
    return (1);
}

/****************************************************************/
/*								*/
/*	DETERMINE IF THE BUFFER CONTAINS A COMPLETE chanAcq MSG	*/
/*								*/
/****************************************************************/

int haveCompleteChanMsg()
{
    short	readBytes;
    SM_WRITE	*smWrite;

    /*
    See if there's enough data for a message type
    */

    if (nchib < sizeof(short))
	return (0);

    /*
    Get the message type and see if it is a flush. If so, there is a message.
    The upper layer will ignore it, so we can too. However, look even further
    ahead to make sure that we don't give up on the buffer too early.
    */

    bcopy(rxTank + tankPos, (char *) &readBytes, sizeof(readBytes));
    if (readBytes == SOCK_FLUSH || readBytes == CHAN_NEEDNMSRV)
	{
	return (1);
	}


    /*
    It must be a SOCK_WRITE message.  See if there's enough for the header.
    */

    if (nchib < (sizeof(short) + sizeof(SM_WRITE)))
	return (0);

    /*
    See if the number of bytes specified in the header is also in the buffer.
    */

    smWrite = (SM_WRITE *)(rxTank + tankPos + sizeof(short));
    bcopy((char *) &(smWrite -> smLength), (char *) &readBytes, sizeof(readBytes));
    readBytes += (sizeof(short) + sizeof(SM_WRITE));
    if (readBytes > nchib)
	return (0);
    else
	return (1);
}

int sockDebug(newLevel)
{
#ifdef vxWorks
    logMsg ("sockMnt debugging changed from level %d to %d\n",
	    smDebug, newLevel, 0,0,0,0);
#else vxWorks
    printf("sockMnt debugging changed from level %d to %d\n",
	    smDebug, newLevel);
#endif vxWorks
    smDebug = newLevel;
    return (0);
}

/****************************************************************/
/*								*/
/*	GET THE LIST OF DPAs AND RESOURCES FROM NAMESRV		*/
/*								*/
/*	Return a 1 for OK, 0 for failure.			*/
/*								*/
/*	WARNING! There is a side effect. If we return a	zero,	*/
/*	then we will have closed the socket!			*/
/*								*/
/****************************************************************/

int getDpaList(fd)
SOCKET	fd;
{
    short		msgType;
    int			nameLen;
    NS_DPAL_REPLY	dpaListHdr;
    NS_RSRQ_REPLY	resListHdr;
    struct dpaCtl	*dpaEntry;

    /*
    Get the list of Data Provider Applications (DPAs) from the name server.
    */

    msgType = htons(NAME_DPAL);
    if (!sendUtil(fd, (char *) &msgType, sizeof(msgType)))
	{
	return (0);
	}
    if (!recvUtil(fd, (char *) &msgType, sizeof(msgType)))
	return (0);
    if (!recvUtil(fd, (char *) &dpaListHdr, sizeof(dpaListHdr)))
	return (0);
    numDPAs = ntohs(dpaListHdr.nsNumDPAs);
    if (numDPAs)
    	dpaList = (struct dpaCtl *) calloc(numDPAs, sizeof(struct dpaCtl));
    dpaEntry = dpaList;
    dpaListHdr.nsNumDPAs = numDPAs;
    dpaListHdr.nsMaxNameLen = ntohs(dpaListHdr.nsMaxNameLen);
    while (dpaListHdr.nsNumDPAs--)
	{
	if (!recvUtil(fd, dpaEntry -> prefix, dpaListHdr.nsMaxNameLen))
		return (0);
	if (!recvUtil(fd, (char *) &(dpaEntry -> virtInet1), sizeof(int)))
		return (0);
	if (!recvUtil(fd, (char *) &(dpaEntry -> virtInet2), sizeof(int)))
		return (0);
	if (dpaListHdr.nsNumDPAs)
	    dpaEntry -> nextDPA = dpaEntry + 1;
	dpaEntry++;
	}

/******
	Get the resource list also. We need that to find DPAs that aren't
	on our local node. Put the real Inet addresses into the DPA table.
******/

    msgType = htons(NAME_RSRQ);
    if (!sendUtil(fd, (char *) &msgType, sizeof(msgType)))
	{
	return (0);
	}
    if (!recvUtil(fd, (char *) &msgType, sizeof(msgType)))
	return (0);
    if (!recvUtil(fd, (char *) &resListHdr, sizeof(resListHdr)))
	return (0);
    resListHdr.nsNumNames = ntohs(resListHdr.nsNumNames);
    nameLen = ntohs(resListHdr.nsMaxNameLen);
    while (resListHdr.nsNumNames--)
	{
	char		resName[16];
	struct in_addr	inetAddr1;
	struct in_addr	inetAddr2;

	if (!recvUtil(fd, resName, nameLen))
		return (0);
	if (!recvUtil(fd, (char *) &inetAddr1, sizeof(inetAddr1)))
		return (0);
	if (inetAddr1.s_addr == caLocalInet.s_addr)
	    inetAddr1.s_addr = htonl(0x7f000001);	 /* Locals connect as 127.0.0.1 */
	if (recvUtil(fd, (char *) &inetAddr2, sizeof(inetAddr2)))
	    if (inetAddr2.s_addr == caLocalInet.s_addr)
		inetAddr2.s_addr = htonl(0x7f000001);

	/*
	Try to find this resource name in the DPA list. If we can't
	find it, then drop this resource on the floor.
	*/

	for (dpaEntry = dpaList; dpaEntry; dpaEntry = dpaEntry -> nextDPA)
	    {
	    if (!strcmp(resName, dpaEntry -> prefix))
		{
		dpaEntry -> realInet1 = inetAddr1;
		dpaEntry -> realInet2 = inetAddr2;
		break;
		}
	    }
	}
    
    return (1);
}


#ifndef vxWorks
#ifndef vms
void dumpBlock(direction, buff, nbytes)
char		*direction;
unsigned char	*buff;
int		nbytes;
{
    struct timeval	tv;
    struct timezone	tz;
    double		timeNow;
    int			bytesDumped = 0;
    int			lastNl;
    int			byte;
    char		graph;

    gettimeofday(&tv, &tz);
    tv.tv_usec /= 1000;
    timeNow = (double) tv.tv_usec;
    timeNow /= 1000.0;
    timeNow += (double) tv.tv_sec;

    traceBytes += nbytes;
    fprintf(traceFd, "\n%s %f:\t%d bytes, %d cumulative\n\t", direction, timeNow, 
	nbytes, traceBytes);

    while (nbytes--)
	{
	lastNl = 0;
	byte = *buff++;
	byte &= 0xff;
	if (isprint(byte))
	    graph = byte;
	else
	    graph = ' ';
	fprintf(traceFd, "%02X%c ", byte, graph);
	bytesDumped++;
	if ((bytesDumped % 25) == 0)
	    {
	    lastNl = 1;
	    fprintf(traceFd, "\n\t");
	    }
	}
    if (!lastNl)
	fprintf(traceFd, "\n");

}
#endif vms
#endif vxWorks

#ifdef vxWorks

/****************************************************************/
/*								*/
/*		GET THE INTERNET ADDRESS OF OUR HDLC LINK	*/
/*								*/
/*	Determine if there is an HDLC link in this machine	*/
/*	and if so, it's IP address. We assume that there is	*/
/*	an HDLC link if in the boot line the gateway addr	*/
/*	is not blank and the network number is different	*/
/*	from the network number of our Ethernet IP addr.	*/
/*	Then, the HDLC address is equal to the gateway IP	*/
/*	addr plus the host number from the Ethernet IP.		*/
/*								*/
/*	Return 0 if there is no link, otherwise the IP.		*/
/*								*/
/****************************************************************/

int hdlcAddress()
{
    char		paramBuff[BOOT_OTHER_LEN];
    char		*hdlcNode;
    char		*colon;
    static BOOT_PARAMS	*bp;

    /*
    If the "other" field blank, then use the "ead" in the
    boot parameters line.
    */

    bp = & sysBootParams;
    if (bp -> other[0] == 0)
	strcpy(paramBuff, bp -> ead);

    /*
    The "other" line is coded. The hdlc address is the
    last parameter on the line. Assume that we start with
    the first nonblank byte after the last comma on the line.
    */

    else
	{
    	hdlcNode = strrchr(bp -> other, ',');
	hdlcNode++;		/* Point past the comma */
	while (*hdlcNode == ' ')
	    hdlcNode++;
    	strcpy(paramBuff, hdlcNode);
	}

    /*
    Wipe out the colon that would indicate the subnet mask.
    */

    if ((colon = strchr(paramBuff, ':')))
	*colon = 0;

    return (inet_addr(paramBuff));
}
#endif vxWorks

#ifndef USE_UDP
/****************************************************************************/
/*									    */
/*	 Attempt to connect (timed), then close to indicate connect-back    */
/****************************************************************************/
static void kick_TCP (long hostaddr /* in net order */,
	       int port) /* in host order */
{
    struct sockaddr_in taddr;
    SOCKET    	socket_fd;
    fd_set	write_fds;
    struct timeval tvtimeout;
    int		true = 1;
#ifdef	SOLARIS
    int		stat;
#endif	SOLARIS

   /* create the TCP socket */
    socket_fd = socket (PF_INET, SOCK_STREAM, 0);
    /* set socket non-blocking! */
#ifdef SOLARIS
    stat = fcntl(socket_fd,F_GETFL,true);
    if(stat >= 0)
      {
	stat |= O_NDELAY;
	stat = fcntl(socket_fd,F_SETFL,stat);
      }
    else
    {
      tisSysLog (0,0, "Can't set non-block on kick_TCP connection: %m");
    }
#else SOLARIS
#ifdef WIN32
    if (ioctlsocket (socket_fd, FIONBIO, &true) == SOCKET_ERROR)
#else
    if (ioctl (socket_fd, FIONBIO, (int)&true) == ERROR)
#endif WIN32
    {
#if !defined(vxWorks) && !defined(vms)
	tisSysLog(0, 0, "Can't set non-block on kick_TCP connection: %m");
#else
	perror("Can't set non-block on kick_TCP connection %d.\n");
#endif vxWorks
    }
#endif SOLARIS
    taddr.sin_family = AF_INET;
    taddr.sin_port = htons (port);
    taddr.sin_addr.s_addr = hostaddr;

    if (connect (socket_fd, (struct sockaddr*)&taddr, sizeof(taddr)) != 0)
    {
#ifdef WIN32
	int localError = WSAGetLastError();
	if ((localError == WSAEINPROGRESS)
	  || (localError == WSAEWOULDBLOCK))
#else
	if (errno == EINPROGRESS)
#endif WIN32
	{
	    /* wait for read file descriptor */
	    FD_ZERO(&write_fds);
	    FD_SET(socket_fd, &write_fds);

	    /* Set timeout to be 0 */
	    tvtimeout.tv_sec = 1;
	    tvtimeout.tv_usec = 0;

	    /* now call select to block until connect, timeout, or error */
	    select (FD_SETSIZE, NULL, &write_fds, NULL, &tvtimeout);
	}
    }
    /* close up and return */
#ifdef WIN32
    closesocket (socket_fd);
#else
    close (socket_fd);
#endif WIN32
}
#endif USE_UDP
