/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		TRANSFORM AGENT DATA STREAMS			*/
/*								*/
/*	This function takes a data stream and transforms it	*/
/*	from one set of embellishments to another. It also	*/
/*	will change the number of elements for array transfer.	*/
/*	Note that we do NOT change the fundamental data type.	*/
/*	We return the number of bytes of data that we copied.	*/
/*	Everything in the input and output buffer is already	*/
/*	in network byte order.					*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#if !(defined(AXP) || defined(vxWorks))
#include <bcopy.h>
#endif

#ifndef NULL
#define NULL	0
#endif

#include <tsDefs.h>
#include <dataTypes.h>
#include <sys/types.h>
#ifdef WIN32
#include <winsock.h>
#else
#include <in.h>
#endif WIN32
#include <choiceGbl.h>
#include <choice.h>

#if defined(AXP) || defined(WIN32) || defined(vxWorks)
#include <string.h>
#endif

#include "tisByteOrder.h"

static short fundamentalSize[] =
    {
    0,				/* AGNT_STRING */
    sizeof(short),		/* AGNT_SHORT and AGNT_INT */
    sizeof(float),		/* AGNT_FLOAT */
    sizeof(short),		/* AGNT_ENUM */
    sizeof(char),		/* AGNT_CHAR */
    sizeof(int),		/* AGNT_LONG (a pre-emptive strike against the DEC Alpha AXP ) */
    sizeof(double),		/* AGNT_DOUBLE */
    };

int transformType(inType, inCount, outType, outCount, inBuff, outBuff, spliceMask)
register short	inType;
short		inCount;
register short	outType;
short		outCount;
register char	*inBuff;
register char	*outBuff;
short		spliceMask;
{
    char	*pStat = NULL;
    char	*pSevr = NULL;
    char	*pAenb = 0;
    char	*pAck = 0;
    char	*pDesc = 0;
    char	*pEgu = 0;
    char	*pTime = NULL;
    short	*pNumStrings = NULL;
    short	numStrings;
    int		grStrLen = 0;
    char	*pHopr = NULL;
    char	*pLopr = NULL;
    char	*pPrec = NULL;
    char	*pDrvh = NULL;
    char	*pDrvl = NULL;
    short	dataSize;
    int		bytesCopied;
    struct choiceSet	*pChoiceSet;
    char	*spliceData = NULL;
    short	spliceChoice = NULL;
    double	spliceDouble;
    float	spliceFloat;
    int		spliceLong;
    short	spliceBits;
    short	alarmAck;
    int		i;
    struct epicsStamp
	{
	int	secPastEpoch;
	int	nsec;
	};
    short	outSevr;

    /*
    First, set the fundamental length. Not relevant if we are passing strings.
    */

    dataSize = fundamentalSize[inType & 0xff];

    /*
    Now find out where all the stuff is in the input. The caller guarantees that
    the output type is a subset of the input type and also that the output count
    is less than or equal to the input count.
    */

    if (inType & AGNT_STAT)
	{
	pStat = inBuff;
	inBuff += sizeof(short);
	pSevr = inBuff;
	bcopy(pSevr, (char *) &spliceBits, sizeof(spliceBits));
	spliceBits = ntohs(spliceBits) & 0xC000;
	inBuff += sizeof(short);
	if ((inType & 0xff) == AGNT_STRING)
	    {
	    pAenb = inBuff;
	    inBuff += 4;
	    pAck = inBuff;
	    inBuff += 4;
	    pDesc = inBuff;
	    inBuff += 28;
	    pEgu = inBuff;
	    inBuff += 8;
	    }
	}

    if (inType & AGNT_TIME)
	{
	pTime = inBuff;
	inBuff += sizeof(struct epicsStamp);
	}

    if (inType & AGNT_GRAPH)
	{
	if ((inType & 0xff) == AGNT_ENUM)
	    {
	    pNumStrings = (short *) inBuff;
	    inBuff += sizeof(short);
	    bcopy((char *)pNumStrings, (char *) &numStrings, sizeof(short));
	    numStrings = ntohs(numStrings);
	    grStrLen = sizeof(short);
	    while (numStrings--)
		{
		do
		    {
		    grStrLen++;
		    } while (*inBuff++);
		}

	    /*
	    Graphic enumeration and control enumeration are really the same thing. Make sure
	    that we only have one of them, regardless of the number of bits that were turned
	    on in the type word passed to us. The data won't occur twice in the input stream.
	    */

	    inType &= ~AGNT_CNTRL;
	    }
	else
	    {
	    pHopr = inBuff;
	    inBuff += dataSize;
	    pLopr = inBuff;
	    inBuff += dataSize;

	    if (((inType & 0xff) == AGNT_FLOAT) || ((inType & 0xff) == AGNT_DOUBLE))
		{
		pPrec = inBuff;
		inBuff += sizeof(short);
		}
	    }
	}

    if (inType & AGNT_CNTRL)
	{
	if ((inType & 0xff) == AGNT_ENUM)
	    {
	    pNumStrings = (short *) inBuff;
	    inBuff += sizeof(short);
	    bcopy((char *)pNumStrings, (char *) &numStrings, sizeof(short));
	    numStrings = ntohs(numStrings);
	    grStrLen = sizeof(short);
	    while (numStrings--)
		{
		do
		    {
		    grStrLen++;
		    } while (*inBuff++);
		}
	    }
	
	else
	    {
	    pDrvh = inBuff;
	    inBuff += dataSize;
	    pDrvl = inBuff;
	    inBuff += dataSize;
	    }
	}

    /*
    inBuff now points to the data. Assemble the output string according to the user's output type.
    */

    bytesCopied = 0;
    if (outType & AGNT_STAT)
	{
	bcopy(pStat, outBuff, sizeof(short));
	outBuff += sizeof(short);
	bytesCopied += sizeof(short);
	bcopy(pSevr, outBuff, sizeof(short));
	if (spliceBits)
	    {
	    short	tempShort;

	    /*
	    Wipe the splicing bits out of SEVR
	    */

	    bcopy(pSevr, (char *) &tempShort, sizeof(tempShort));
	    tempShort = ntohs(tempShort) & 0x3FFF;
	    tempShort = htons(tempShort);
	    bcopy((char *) &tempShort, outBuff, sizeof(tempShort));
	    }
	outBuff += sizeof(short);
	bytesCopied += sizeof(short);
	if ((outType & 0xff) == AGNT_STRING)
	    {
	    
	    /*
	    It is possible that we don't have all the data requested because a string ACK, STAT, or SEVR
	    was spliced onto a non-string VAL. If we don't have any of the data, provide empty strings.
	    */

	    if (pAenb)
		{
	    	bcopy(pAenb, outBuff, 4);
		}
	    else
		*outBuff = 0;
	    outBuff += 4;
	    bytesCopied += 4;

	    if (pAck)
		{
	    	bcopy(pAck, outBuff, 4);	/* This ACK doesn't count if someone opened status string
						   on ACK. The value will contains what is needed! */
		}
	    else
		*outBuff = 0;
	    outBuff += 4;
	    bytesCopied += 4;

	    if (pDesc)
		{
	    	bcopy(pDesc, outBuff, 28);
		}
	    else
		*outBuff = 0;
	    outBuff += 28;
	    bytesCopied += 28;

	    if (pEgu)
		{
	    	bcopy(pEgu, outBuff, 8);
		}
	    else
		*outBuff = 0;
	    outBuff += 8;
	    bytesCopied += 8;
	    }
	}

    if (outType & AGNT_TIME)
	{
	bcopy(pTime, outBuff, sizeof(struct epicsStamp));
	outBuff += sizeof(struct epicsStamp);
	bytesCopied += sizeof(struct epicsStamp);
	}

    if (outType & AGNT_GRAPH)
	{
	if ((outType & 0xff) == AGNT_ENUM)
	    {
	    if (spliceMask & AGNT_SPLICED)
		{
		struct
		    {
		    short	numStrings;
		    char	enumStrs[1024];
		    } enumInfo;

		/*
		We have to supply the strings for enumerations
		*/

		if (spliceMask & AGNT_SPL_STAT)
		    spliceChoice = GBL_ALARM_STAT;
		else if (spliceMask & AGNT_SPL_SEVR)
		    spliceChoice = GBL_ALARM_SEV;
		else if (spliceMask & AGNT_SPL_ACK)
		    spliceChoice = GBL_YES_NO;
		pChoiceSet = GET_PCHOICE_SET(choiceGbl, spliceChoice);
		enumInfo.numStrings = htons((short)pChoiceSet -> number);
		grStrLen = 0;

		for (i = 0; i < pChoiceSet -> number; i++)
		    {
		    strcpy(enumInfo.enumStrs + grStrLen, GET_CHOICE(pChoiceSet, i));
		    grStrLen += strlen(enumInfo.enumStrs + grStrLen) + 1;
		    }
		grStrLen += sizeof(short);
		pNumStrings = (short *) &enumInfo;
		}

	    bcopy((char *)pNumStrings, outBuff, grStrLen);
	    outBuff += grStrLen;
	    bytesCopied += grStrLen;
	    }
	else
	    {
	    double	spliceDouble;
	    float	spliceFloat;
	    short	spliceShort;
	    int		spliceInt;

	    if (spliceMask & AGNT_SPLICED)
		{
		switch (outType & 0xff)
	    	    {
	    	    case AGNT_STRING:
			/** No such thing as graphic string **/
		        break;
	    	    case AGNT_SHORT:
	    	    case AGNT_ENUM:
			spliceShort = htons(100);
			bcopy((char *) &spliceShort, outBuff, sizeof(short));
			outBuff += sizeof(short);
			bytesCopied += sizeof(short);
			spliceShort = 0;
			bcopy((char *) &spliceShort, outBuff, sizeof(short));
			outBuff += sizeof(short);
			bytesCopied += sizeof(short);
			break;
	    	    case AGNT_FLOAT:
			spliceFloat = (float)100;
			htonf(spliceFloat, spliceFloat);
			bcopy((char *) &spliceFloat, outBuff, sizeof(float));
			outBuff += sizeof(float);
			bytesCopied += sizeof(float);
			spliceFloat = (float)0;
			htonf(spliceFloat, spliceFloat); 	/* Dirty zeros? */
			bcopy((char *) &spliceFloat, outBuff, sizeof(float));
			outBuff += sizeof(float);
			bytesCopied += sizeof(float);
			break;
	    	    case AGNT_CHAR:
			*outBuff++ = 100;
			bytesCopied++;
			*outBuff++ = 0;
			bytesCopied++;
			break;
	    	    case AGNT_LONG:
			spliceInt = htonl(100);
			bcopy((char *) &spliceInt, outBuff, sizeof(int));
			outBuff += sizeof(int);
			bytesCopied += sizeof(int);
			spliceInt = 0;
			bcopy((char *) &spliceInt, outBuff, sizeof(int));
			outBuff += sizeof(int);
			bytesCopied += sizeof(int);
			break;
	    	    case AGNT_DOUBLE:
			spliceDouble = 100;
			htond(spliceDouble, spliceDouble);
			bcopy((char *) &spliceDouble, outBuff, sizeof(double));
			outBuff += sizeof(double);
			bytesCopied += sizeof(double);
			spliceDouble = 0;
			htonf(spliceDouble, spliceDouble);	/* Dirty zeros? */
			bcopy((char *) &spliceDouble, outBuff, sizeof(double));
			outBuff += sizeof(double);
			bytesCopied += sizeof(double);
			break;
	    	    }
	        if (((outType & 0xff) == AGNT_FLOAT) || ((outType & 0xff) == AGNT_DOUBLE))
		    {
		    bcopy(pPrec, outBuff, sizeof(short));
		    outBuff += sizeof(short);
		    bytesCopied += sizeof(short);
		    }
		}
	    else
		{
	        bcopy(pHopr, outBuff, dataSize);
	        outBuff += dataSize;
	        bytesCopied += dataSize;
	        bcopy(pLopr, outBuff, dataSize);
	        outBuff += dataSize;
	        bytesCopied += dataSize;
	        if (((outType & 0xff) == AGNT_FLOAT) || ((outType & 0xff) == AGNT_DOUBLE))
		    {
		    bcopy(pPrec, outBuff, sizeof(short));
		    outBuff += sizeof(short);
		    bytesCopied += sizeof(short);
		    }
		}
	    }
	}

    if (outType & AGNT_CNTRL)
	{
	if ((outType & 0xff) == AGNT_ENUM)
	    {
	    if (spliceMask & AGNT_SPLICED)
		{
		struct
		    {
		    short	numStrings;
		    char	enumStrs[1024];
		    } enumInfo;

		/*
		We have to supply the strings for enumerations
		*/

		if (spliceMask & AGNT_SPL_STAT)
		    spliceChoice = GBL_ALARM_STAT;
		else if (spliceMask & AGNT_SPL_SEVR)
		    spliceChoice = GBL_ALARM_SEV;
		else if (spliceMask & AGNT_SPL_ACK)
		    spliceChoice = GBL_YES_NO;
		pChoiceSet = GET_PCHOICE_SET(choiceGbl, spliceChoice);
		enumInfo.numStrings = htons((short)pChoiceSet -> number);
		grStrLen = 0;

		for (i = 0; i < pChoiceSet -> number; i++)
		    {
		    strcpy(enumInfo.enumStrs + grStrLen, GET_CHOICE(pChoiceSet, i));
		    grStrLen += strlen(enumInfo.enumStrs + grStrLen) + 1;
		    }
		grStrLen += sizeof(short);
		pNumStrings = (short *) &enumInfo;
		}
	    bcopy((char *)pNumStrings, outBuff, grStrLen);
	    outBuff += grStrLen;
	    bytesCopied += grStrLen;
	    }
	else
	    {
	    double	spliceDouble;
	    float	spliceFloat;
	    short	spliceShort;
	    int		spliceInt;

	    if (spliceMask & AGNT_SPLICED)
		{
		switch (outType & 0xff)
	    	    {
	    	    case AGNT_STRING:
			/** No such thing as control string **/
		        break;
	    	    case AGNT_SHORT:
	    	    case AGNT_ENUM:
			spliceShort = htons(100);
			bcopy((char *) &spliceShort, outBuff, sizeof(short));
			outBuff += sizeof(short);
			bytesCopied += sizeof(short);
			spliceShort = 0;
			bcopy((char *) &spliceShort, outBuff, sizeof(short));
			outBuff += sizeof(short);
			bytesCopied += sizeof(short);
			break;
	    	    case AGNT_FLOAT:
			spliceFloat = (float)100;
			htonf(spliceFloat, spliceFloat);
			bcopy((char *) &spliceFloat, outBuff, sizeof(float));
			outBuff += sizeof(float);
			bytesCopied += sizeof(float);
			spliceFloat = (float)0;
			htonf(spliceFloat, spliceFloat); 	/* Dirty zeros? */
			bcopy((char *) &spliceFloat, outBuff, sizeof(float));
			outBuff += sizeof(float);
			bytesCopied += sizeof(float);
			break;
	    	    case AGNT_CHAR:
			*outBuff++ = 100;
			bytesCopied++;
			*outBuff++ = 0;
			bytesCopied++;
			break;
	    	    case AGNT_LONG:
			spliceInt = htonl(100);
			bcopy((char *) &spliceInt, outBuff, sizeof(int));
			outBuff += sizeof(int);
			bytesCopied += sizeof(int);
			spliceInt = 0;
			bcopy((char *) &spliceInt, outBuff, sizeof(int));
			outBuff += sizeof(int);
			bytesCopied += sizeof(int);
			break;
	    	    case AGNT_DOUBLE:
			spliceDouble = 100;
			htond(spliceDouble, spliceDouble);
			bcopy((char *) &spliceDouble, outBuff, sizeof(double));
			outBuff += sizeof(double);
			bytesCopied += sizeof(double);
			spliceDouble = 0;
			htonf(spliceDouble, spliceDouble);	/* Dirty zeros? */
			bcopy((char *) &spliceDouble, outBuff, sizeof(double));
			outBuff += sizeof(double);
			bytesCopied += sizeof(double);
			break;
	    	    }
		}

	    else
		{
	        bcopy(pDrvh, outBuff, dataSize);
	        outBuff += dataSize;
	        bytesCopied += dataSize;
	        bcopy(pDrvl, outBuff, dataSize);
	        outBuff += dataSize;
	        bytesCopied += dataSize;
	        }
	    }
	}


    /*
    See if this channel was spliced.
    */

    if (spliceMask & AGNT_SPLICED)
	{

    	/*
    	Make inBuff point to the data we REALLY want to propagate.
    	That will depend on the desired field and data type.
	First, point to the proper raw data field.
	Remember global choice numbers in case we want strings as output.
    	*/

    	if (spliceMask & AGNT_SPL_STAT)
	    {
	    spliceData = pStat;
	    spliceChoice = GBL_ALARM_STAT;
	    }
    	else if (spliceMask & AGNT_SPL_SEVR)
	    {
	    spliceData = (char *) &outSevr;
	    bcopy(pSevr, spliceData, sizeof(outSevr));
	    outSevr &= htons(0x3fff);
	    spliceChoice = GBL_ALARM_SEV;
	    }
    	else if (spliceMask & AGNT_SPL_ACK)
	    {
	    if (spliceBits & 0x4000)
	        alarmAck = htons(1);
	    else
	        alarmAck = 0;
	    spliceData = (char *) &alarmAck;
	    spliceChoice = GBL_YES_NO;
	    }

	switch (outType & 0xff)
	    {
	    case AGNT_STRING:
		pDesc = GET_CHOICE(GET_PCHOICE_SET(choiceGbl, spliceChoice), ntohs(*(short *) spliceData));
		strcpy(outBuff, pDesc);
		return (bytesCopied + strlen(pDesc) + 1);
		break;

    	    case AGNT_SHORT:
	    case AGNT_ENUM:
		bcopy(spliceData, outBuff, sizeof(short));
		return (bytesCopied + sizeof(short));

  	    case AGNT_FLOAT:
		spliceFloat = (float) ntohs(*(short *) spliceData);
		htonf(spliceFloat, spliceFloat);
		bcopy((char *) &spliceFloat, outBuff, sizeof(float));
		return (bytesCopied + sizeof(float));

    	    case AGNT_CHAR:
		*outBuff = (char) ntohs(*(short *) spliceData);
		return (bytesCopied + sizeof(char));

    	    case AGNT_LONG:
		spliceLong = (int) ntohs(*(short *) spliceData);
		spliceLong = htonl(spliceLong);
		bcopy((char *) &spliceLong, outBuff, sizeof(int));
		return (bytesCopied + sizeof(int));

    	    case AGNT_DOUBLE:
		spliceDouble = (double) ntohs(*(short *) spliceData);
		htond(spliceDouble, spliceDouble);
		bcopy((char *) &spliceDouble, outBuff, sizeof(double));
		return (bytesCopied + sizeof(double));
	    }
	}


    /*
    Finally, copy the process variable. We don't do arrays of strings.
    */

    if ((outType & 0xff) == AGNT_STRING)
	{
	strcpy(outBuff, inBuff);
	return (bytesCopied + strlen(inBuff) + 1);
	}

    bcopy(inBuff + (inCount - outCount) * dataSize, outBuff, outCount * dataSize);
    return (bytesCopied + outCount * dataSize);
}
    
