/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include	<stdio.h>
#ifndef AXP
#include	<string.h>
#else
#include	<strings.h>
#endif
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<errno.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

static CODE4		cb;
static char		pathbuff[128];
static char		sysbuff[256];
static struct stat	statbuff;

static int create_subdir (void);
static int convert_pvmap (void);
static int convert_pid (void);

#define	TIS4000DB	"/tis4000db"
#define	DBCNVRT		"t43t44cnvrt"


F4FIELD_INFO	pvmapFieldInfo[] = 
	{
	{"NAME",	'C',	28,	0},
	{"TYPE",	'C',	15,	0},
	{"DESC",	'C',	28,	0},
	{"LOCATION",	'C',	16,	0},
	{"MERGE",	'C',	16,	0},
	{"STYPE",	'C',	10,	0},
	{"SRATE",	'N',	4,	1},
	{"PHASE",	'N',	3,	0},
	{"EVENT",	'N',	4,	0},
	{"ONSCAN",	'L',	1,	0},
	{"ONALARM",	'L',	1,	0},
	{"TAGDEPTH",	'N',	2,	0},
	{"FLNK",	'C',	28,	0},
	{"REV",		'N',	2,	0},
	{"ALARMGROUP",	'C',	16,	0},
	{"PRIVGROUP",	'C',	16,	0},
	{"DEVTYPE",	'C',	23,	0},
	{"SDISABLE",	'C',	28,	0},
	{"DISVALUE",	'N',	3,	0},
	{"RESRCNAME",	'C',	16,	0},
	{"IOC1DNLD",	'L',	1,	0},
	{"IOC2DNLD",	'L',	1,	0},
	{"CHECKSUM",	'C',	4,	0},
	{"TIMESTAMP",	'C',	10,	0},
	{"LOAD1FLAG",	'L',	1,	0},
	{"LOAD2FLAG",	'L',	1,	0},
	{"DELETEFLAG",	'L',	1,	0},
	{"IMPORTNODE",	'C',	16,	0},
	{"IMPORTDATE",	'C',	10,	0},
	{"IMPDELFLAG",	'L',	1,	0},
	{"ONTREND",	'L',	1,	0},
	{0,0,0,0}
	};

T4TAG_INFO	pvmapTagInfo[] = 
	{
	{"NAME",	"NAME",	0,	0,	0},
	{"POINTNAME",	"NAME",	0,	0,	0},
	{0,0,0,0,0}
	};


/* t50 pid record definition, need to change t43 field AUTOCLOSE
   from Logical to Char 10
*/
F4FIELD_INFO	pidFieldInfo[] = 
	{
	{"NAME",	'C',	28,	0},
	{"INITLSETPT",	'N',	10,	4},
	{"REVERSEINI",	'L',	1,	0},
	{"SETPTSPEC",	'C',	33,	0},
	{"SETPTCONST",	'N',	10,	4},
	{"SETPTMS",	'L',	1,	0},
	{"SETPTPROC",	'L',	1,	0},
	{"MVALSPEC",	'C',	33,	0},
	{"MVALMS",	'L',	1,	0},
	{"MVALPROC",	'L',	1,	0},
	{"SETCLAMPHI",	'N',	11,	4},
	{"SETCLAMPLO",	'N',	11,	4},
	{"EGUHIGH",	'N',	11,	4},
	{"EGULOW",	'N',	11,	4},
	{"EGUTAG",	'C',	16,	0},
	{"DISPHIGH",	'N',	11,	4},
	{"DISPLOW",	'N',	11,	4},
	{"PRECISION",	'N',	1,	0},
	{"AUTOMATIC",	'L',	1,	0},
	{"CASFIELD",	'C',	33,	0},
	{"CASCADEMS",	'L',	1,	0},
	{"CASPROCINP",	'L',	1,	0},
	{"CASOPENFLG",	'L',	1,	0},
	{"CASSTPTPCT",	'L',	1,	0},
	{"INCREMENTL",	'L',	1,	0},
	{"OUTPUTCHAN",	'C',	33,	0},
	{"OUTCLAMPHI",	'N',	6,	2},
	{"OUTCLAMPLO",	'N',	7,	2},
	{"OUTREVERSE",	'L',	1,	0},
	{"PROP_BAND",	'N',	7,	3},
	{"RESET",	'N',	7,	3},
	{"RATE",	'N',	7,	3},
	{"ERRSQUARED",	'L',	1,	0},
	{"DELTAERROR",	'L',	1,	0},
	{"TRKMODENA",	'L',	1,	0},
	{"HOLDLOSTMV",	'L',	1,	0},
	{"NOMVOUTVAL",	'N',	7,	3},
	{"DEVLIMIT",	'N',	11,	4},
	{"DEVDEADBND",	'N',	11,	4},
	{"DEVSEVER",	'C',	13,	0},
	{"NOINPSEVER",	'C',	13,	0},
	{"NOMVSEVER",	'C',	13,	0},
	{"NOOUTSEVER",	'C',	13,	0},
	{"NOCCSEVER",	'C',	13,	0},
	{"CLDEADBAND",	'N',	11,	4},
	{"CLHISEVER",	'C',	13,	0},
	{"CLLOSEVER",	'C',	13,	0},
	{"TRKMODSEVR",	'C',	13,	0},
	{"MONDEADBND",	'N',	7,	3},
	{"ARCDEADBND",	'N',	7,	3},
	{"AUTOCLOSE",	'C',	10,	0},
	{"INITACT",	'C',	13,	0},
	{"SETPTRAMP",	'N',	11,	4},
	{"SETPTCROSS",	'L',	1,	0},
	{"SETPTGAP",	'N',	11,	4},
	{"UPDATEISTP",	'L',	1,	0},
	{0,0,0,0}
	};

T4TAG_INFO	pidTagInfo[] = 
	{
	{"NAME",	"NAME",	0,	0,	0},
	{0,0,0,0,0}
	};

int main()
{
    fprintf (stdout, "starting t43 t44/t50 conversion\n\n");
    d4init (&cb);

#if defined(WIN32) || defined(USECB63)
    cb.errGo = 0;
    cb.errOpen = 0;
    cb.errCreate = 0;
    cb.errFieldName = 0;
#else
    cb.go_error = 0;
    cb.open_error = 0;
    cb.create_error = 0;
    cb.field_name_error = 0;
#endif WIN32 or USECB63

    if (create_subdir ())
	return (1);
    if (convert_pvmap ())
	return (1);
    if (convert_pid ())
	return (1);
    d4close_all (&cb);
    fprintf (stdout, "\nt43 t44/t50 conversion finished\n");
    return (0);
}


static int create_subdir (void)
{
    sprintf (pathbuff, "%s/%s", TIS4000DB, DBCNVRT);
    if (stat (pathbuff, &statbuff))
    {
	if (errno == ENOENT)
	{
	    sprintf (sysbuff, "mkdir %s", pathbuff);
	    system (sysbuff);
	    if (stat (pathbuff, &statbuff))
	    {
		fprintf (stdout, "t43t44conv err: failed to create %s\n", pathbuff);
		return (1);
	    }
	    else
	    {
		fprintf (stdout, "t43t44conv: %s created\n", pathbuff);
		return (0);
	    }
	}
	else
	{
	    fprintf (stdout, "t43t44conv err: stat problem (%d) on %s\n", errno, pathbuff);
	    return (1);
	}
    }
    else
    {
	fprintf (stdout, "t43t44conv err: %s already exists\n", pathbuff);
    }
    return (0);
}


static int convert_pid (void)
{
    DATA4	*old_pid, *new_pid;
    FIELD4	*old_outclamplo, *new_outclamplo;
    FIELD4	*old_outclamphi, *new_outclamphi;
    FIELD4	*old_autoclose, *new_autoclose;
    FIELD4	*old_initact, *new_initact;
    FIELD4	*old_setptramp, *new_setptramp;
    FIELD4	*old_setptcross, *new_setptcross;
    FIELD4	*old_setptgap, *new_setptgap;
    FIELD4	*old_updateistp, *new_updateistp;
    int		old_record_width, new_record_width;
    char	*old_rec, *new_rec;
    char	*pold_autoclose, *pold_outclamphi;
    char	*pold_outclamplo, *pnew_outclamplo;

    fprintf (stdout, "...converting pid.dbf\n");

    sprintf (pathbuff, "%s/pid", TIS4000DB);
    if (!(old_pid = d4open (&cb, pathbuff)))
    {
	fprintf (stdout, "t43t44/t50conv err: open error on %s\n", pathbuff);
	return (1);
    }
    d4alias_set (old_pid, "old_pid");

    /* save the old dbase files */
    sprintf (sysbuff, "cp -p %s/pid.dbf %s/%s/pid.dbf.sav", 
		       TIS4000DB, TIS4000DB, DBCNVRT);
    system (sysbuff);
    sprintf (sysbuff, "cp -p %s/pid.mdx %s/%s/pid.mdx.sav", 
		       TIS4000DB, TIS4000DB, DBCNVRT);
    system (sysbuff);

    /* create the new dbase file */
    sprintf (pathbuff, "%s/%s/pid", TIS4000DB,DBCNVRT);
    if (!(new_pid = d4create (&cb, pathbuff, pidFieldInfo, pidTagInfo)))
    {
	fprintf (stdout, "t43t44/t50conv err: failed to create %s\n", pathbuff);
	d4close (old_pid);
	return (1);
    }

    old_record_width = d4record_width (old_pid);
    new_record_width = d4record_width (new_pid);
    if (old_record_width == new_record_width)
    {
	fprintf (stdout, "t43t44/t50conv err: pid has already been converted\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }

    if ((old_outclamphi = d4field (old_pid, "OUTCLAMPHI")) == 0)
    {
	fprintf (stdout, "t43/t44conv err: OUTCLAMPHI not exist in old pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    pold_outclamphi = f4ptr (old_outclamphi);

    if ((new_outclamphi = d4field (new_pid, "OUTCLAMPHI")) == 0)
    {
	fprintf (stdout, "t43/t44conv err: OUTCLAMPHI not exist in new pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }

    if ((old_outclamplo = d4field (old_pid, "OUTCLAMPLO")) == 0)
    {
	fprintf (stdout, "t43/t44conv err: OUTCLAMPLO not exist in old pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    pold_outclamplo = f4ptr (old_outclamplo);

    if ((new_outclamplo = d4field (new_pid, "OUTCLAMPLO")) == 0)
    {
	fprintf (stdout, "t43/t44conv err: OUTCLAMPLO not exist in new pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    pnew_outclamplo = f4ptr (new_outclamplo);

    if ((old_autoclose = d4field (old_pid, "AUTOCLOSE")) == 0)
    {
	fprintf (stdout, "t43/t44conv err: AUTOCLOSE not exist in old pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    if ((f4type (old_autoclose)) != r4log)
    {
	fprintf (stdout, "t43/t44conv err: AUTOCLOSE field is not a logical\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    /* ptr to AUTOCLOSE field in old_pid */
    pold_autoclose = f4ptr (old_autoclose);

    if ((new_autoclose = d4field (new_pid, "AUTOCLOSE")) == 0)
    {
	fprintf (stdout, "t43/t44conv err: AUTOCLOSE not exist in new pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }

    if ((old_initact = d4field (old_pid, "INITACT")) == 0)
    {
	fprintf (stdout, "t43/t44conv err: INITACT not exist in old pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    if ((new_initact = d4field (new_pid, "INITACT")) == 0)
    {
	fprintf (stdout, "t43/t44conv err: INITACT not exist in new pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }

    if ((old_setptramp = d4field (old_pid, "SETPTRAMP")) == 0)
    {
	fprintf (stdout, "t43/t44conv err: SETPTRAMP not exist in old pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    if ((new_setptramp = d4field (new_pid, "SETPTRAMP")) == 0)
    {
	fprintf (stdout, "t43/t44conv err: SETPTRAMP not exist in new pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }

    if ((old_setptcross = d4field (old_pid, "SETPTCROSS")) == 0)
    {
	fprintf (stdout, "t43/t44conv err: SETPTCROSS not exist in old pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    if ((new_setptcross = d4field (new_pid, "SETPTCROSS")) == 0)
    {
	fprintf (stdout, "t43/t44conv err: SETPTCROSS not exist in new pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }

    if ((old_setptgap = d4field (old_pid, "SETPTGAP")) == 0)
    {
	fprintf (stdout, "t43/t44conv err: SETPTGAP not exist in old pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    if ((new_setptgap = d4field (new_pid, "SETPTGAP")) == 0)
    {
	fprintf (stdout, "t43/t44conv err: SETPTGAP not exist in new pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }

    if ((old_updateistp = d4field (old_pid, "UPDATEISTP")) == 0)
    {
	fprintf (stdout, "t43/t44conv err: UPDATEISTP not exist in old pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }
    if ((new_updateistp = d4field (new_pid, "UPDATEISTP")) == 0)
    {
	fprintf (stdout, "t43/t44conv err: UPDATEISTP not exist in new pid\n");
	d4close (old_pid);
	d4close (new_pid);
	return (1);
    }

    old_rec = d4record (old_pid);
    new_rec = d4record (new_pid);

    for (d4top (old_pid); !d4eof (old_pid); d4skip (old_pid, 1))
    {
	/* do not copy if deleted */
	if (d4deleted (old_pid))
	{
	    continue;
	}
	d4append_start (new_pid, 0);
	d4blank (new_pid);
	memcpy (new_rec, old_rec, pold_outclamphi - old_rec);
	f4assign_field (new_outclamphi, old_outclamphi);
	memcpy (pnew_outclamplo, pold_outclamplo, (pold_autoclose - pold_outclamplo));
	if ((f4true (old_autoclose)) == 0)
	{
	    f4assign (new_autoclose, "Automatic");
	}
	else
	{
	    f4assign (new_autoclose, "Manual");
	}
	f4assign_field (new_initact, old_initact);
	f4assign_field (new_setptramp, old_setptramp);
	f4assign_field (new_setptcross, old_setptcross);
	f4assign_field (new_setptgap, old_setptgap);
	f4assign_field (new_updateistp, old_updateistp);

	d4append (new_pid);
	d4flush (new_pid);
    }
    d4close (old_pid);
    d4close (new_pid);

    fprintf (stdout, "...finished pid.dbf\n");
    return (0);
}


static int convert_pvmap (void)
{
    DATA4	*old_pvmap, *new_pvmap;
    FIELD4	*old_name, *new_name;
    FIELD4	*old_type, *new_type; 
    FIELD4	*old_desc, *new_desc; 
    FIELD4	*old_loc, *new_loc; 
    FIELD4	*old_merge, *new_merge; 
    FIELD4	*old_stype, *new_stype; 
    FIELD4	*old_srate, *new_srate; 
    FIELD4	*old_phase, *new_phase; 
    FIELD4	*old_event, *new_event; 
    FIELD4	*old_onscan, *new_onscan; 
    FIELD4	*old_onalarm, *new_onalarm; 
    FIELD4	*old_tagdepth, *new_tagdepth; 
    FIELD4	*old_flnk, *new_flnk; 
    FIELD4	*old_rev, *new_rev; 
    FIELD4	*old_alarmgroup, *new_alarmgroup; 
    FIELD4	*old_privgroup, *new_privgroup; 
    FIELD4	*old_devtype, *new_devtype; 
    FIELD4	*old_sdisable, *new_sdisable; 
    FIELD4	*old_disvalue, *new_disvalue; 
    FIELD4	*old_resrcname, *new_resrcname; 
    FIELD4	*old_ioc1dnld, *new_ioc1dnld; 
    FIELD4	*old_ioc2dnld, *new_ioc2dnld; 
    FIELD4	*old_checksum, *new_checksum; 
    FIELD4	*old_ontrend, *new_ontrend; 
    FIELD4	*old_timestamp, *new_timestamp; 
    FIELD4	*old_load1flag, *new_load1flag; 
    FIELD4	*old_load2flag, *new_load2flag; 
    FIELD4	*old_deleteflag, *new_deleteflag; 
    FIELD4	*old_importnode, *new_importnode; 
    FIELD4	*old_importdate, *new_importdate; 
    FIELD4	*old_impdelflag, *new_impdelflag; 

    int		old_record_width, new_record_width;

    fprintf (stdout, "...converting pvmap.dbf\n");

    sprintf (pathbuff, "%s/pvmap", TIS4000DB);
    if (!(old_pvmap = d4open (&cb, pathbuff)))
    {
	fprintf (stdout, "t43t44conv err: open error on %s\n", pathbuff);
	return (1);
    }
    d4alias_set (old_pvmap, "old_pvmap");

    /* save the old dbase files */
    sprintf (sysbuff, "cp -p %s/pvmap.dbf %s/%s/pvmap.dbf.sav", 
		       TIS4000DB, TIS4000DB, DBCNVRT);
    system (sysbuff);
    sprintf (sysbuff, "cp -p %s/pvmap.mdx %s/%s/pvmap.mdx.sav", 
		       TIS4000DB, TIS4000DB, DBCNVRT);
    system (sysbuff);

    /* create the new dbase file */
    sprintf (pathbuff, "%s/%s/pvmap", TIS4000DB,DBCNVRT);
    if (!(new_pvmap = d4create (&cb, pathbuff, pvmapFieldInfo, pvmapTagInfo)))
    {
	fprintf (stdout, "t43t44conv err: failed to create %s\n", pathbuff);
	d4close (old_pvmap);
	return (1);
    }
    old_record_width = d4record_width (old_pvmap);
    new_record_width = d4record_width (new_pvmap);
    if (old_record_width == new_record_width)
    {
	/* the records might be the same length, but the ONTREND
	 * field must be last (31) for the file to have already been
	 * converted
	 */
	int fldNum;
	fldNum = d4field_number (old_pvmap, "ONTREND");
	if (fldNum == 31)
	{
	    fprintf (stdout, "t43t44conv err: pvmap has already been converted\n");
	    d4close (old_pvmap);
	    d4close (new_pvmap);
	    return (1);
	}
    }

    if ((old_name = d4field (old_pvmap, "NAME")) == 0)
    {
	fprintf (stdout, "t43t44conv err: NAME not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_name = d4field (new_pvmap, "NAME")) == 0)
    {
	fprintf (stdout, "t43t44conv err: NAME not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_type = d4field (old_pvmap, "TYPE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: TYPE not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_type = d4field (new_pvmap, "TYPE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: TYPE not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_desc = d4field (old_pvmap, "DESC")) == 0)
    {
	fprintf (stdout, "t43t44conv err: DESC not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_desc = d4field (new_pvmap, "DESC")) == 0)
    {
	fprintf (stdout, "t43t44conv err: DESC not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_loc = d4field (old_pvmap, "LOCATION")) == 0)
    {
	fprintf (stdout, "t43t44conv err: LOCATION not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_loc = d4field (new_pvmap, "LOCATION")) == 0)
    {
	fprintf (stdout, "t43t44conv err: LOCATION not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_merge = d4field (old_pvmap, "MERGE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: MERGE not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_merge = d4field (new_pvmap, "MERGE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: MERGE not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_stype = d4field (old_pvmap, "STYPE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: STYPE not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_stype = d4field (new_pvmap, "STYPE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: STYPE not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_srate = d4field (old_pvmap, "SRATE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: SRATE not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_srate = d4field (new_pvmap, "SRATE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: SRATE not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_phase = d4field (old_pvmap, "PHASE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: PHASE not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_phase = d4field (new_pvmap, "PHASE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: PHASE not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_event = d4field (old_pvmap, "EVENT")) == 0)
    {
	fprintf (stdout, "t43t44conv err: EVENT not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_event = d4field (new_pvmap, "EVENT")) == 0)
    {
	fprintf (stdout, "t43t44conv err: EVENT not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_onscan = d4field (old_pvmap, "ONSCAN")) == 0)
    {
	fprintf (stdout, "t43t44conv err: ONSCAN not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_onscan = d4field (new_pvmap, "ONSCAN")) == 0)
    {
	fprintf (stdout, "t43t44conv err: ONSCAN not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_onalarm = d4field (old_pvmap, "ONALARM")) == 0)
    {
	fprintf (stdout, "t43t44conv err: ONALARM not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_onalarm = d4field (new_pvmap, "ONALARM")) == 0)
    {
	fprintf (stdout, "t43t44conv err: ONALARM not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_tagdepth = d4field (old_pvmap, "TAGDEPTH")) == 0)
    {
	fprintf (stdout, "t43t44conv err: TAGDEPTH not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_tagdepth = d4field (new_pvmap, "TAGDEPTH")) == 0)
    {
	fprintf (stdout, "t43t44conv err: TAGDEPTH not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_flnk = d4field (old_pvmap, "FLNK")) == 0)
    {
	fprintf (stdout, "t43t44conv err: FLNK not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_flnk = d4field (new_pvmap, "FLNK")) == 0)
    {
	fprintf (stdout, "t43t44conv err: FLNK not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_rev = d4field (old_pvmap, "REV")) == 0)
    {
	fprintf (stdout, "t43t44conv err: REV not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_rev = d4field (new_pvmap, "REV")) == 0)
    {
	fprintf (stdout, "t43t44conv err: REV not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_alarmgroup = d4field (old_pvmap, "ALARMGROUP")) == 0)
    {
	fprintf (stdout, "t43t44conv err: ALARMGROUP not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_alarmgroup = d4field (new_pvmap, "ALARMGROUP")) == 0)
    {
	fprintf (stdout, "t43t44conv err: ALARMGROUP not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_privgroup = d4field (old_pvmap, "PRIVGROUP")) == 0)
    {
	fprintf (stdout, "t43t44conv err: PRIVGROUP not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_privgroup = d4field (new_pvmap, "PRIVGROUP")) == 0)
    {
	fprintf (stdout, "t43t44conv err: PRIVGROUP not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_devtype = d4field (old_pvmap, "DEVTYPE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: DEVTYPE not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_devtype = d4field (new_pvmap, "DEVTYPE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: DEVTYPE not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_sdisable = d4field (old_pvmap, "SDISABLE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: SDISABLE not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_sdisable = d4field (new_pvmap, "SDISABLE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: SDISABLE not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_disvalue = d4field (old_pvmap, "DISVALUE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: DISVALUE not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_disvalue = d4field (new_pvmap, "DISVALUE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: DISVALUE not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_resrcname = d4field (old_pvmap, "RESRCNAME")) == 0)
    {
	fprintf (stdout, "t43t44conv err: RESRCNAME not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_resrcname = d4field (new_pvmap, "RESRCNAME")) == 0)
    {
	fprintf (stdout, "t43t44conv err: RESRCNAME not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_ioc1dnld = d4field (old_pvmap, "IOC1DNLD")) == 0)
    {
	fprintf (stdout, "t43t44conv err: IOC1DNLD not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_ioc1dnld = d4field (new_pvmap, "IOC1DNLD")) == 0)
    {
	fprintf (stdout, "t43t44conv err: IOC1DNLD not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_ioc2dnld = d4field (old_pvmap, "IOC2DNLD")) == 0)
    {
	fprintf (stdout, "t43t44conv err: IOC2DNLD not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_ioc2dnld = d4field (new_pvmap, "IOC2DNLD")) == 0)
    {
	fprintf (stdout, "t43t44conv err: IOC2DNLD not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_checksum = d4field (old_pvmap, "CHECKSUM")) == 0)
    {
	fprintf (stdout, "t43t44conv err: CHECKSUM not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_checksum = d4field (new_pvmap, "CHECKSUM")) == 0)
    {
	fprintf (stdout, "t43t44conv err: CHECKSUM not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_ontrend = d4field (old_pvmap, "ONTREND")) == 0)
    {
	fprintf (stdout, "t43t44conv err: ONTREND not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_ontrend = d4field (new_pvmap, "ONTREND")) == 0)
    {
	fprintf (stdout, "t43t44conv err: ONTREND not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_timestamp = d4field (old_pvmap, "TIMESTAMP")) == 0)
    {
	fprintf (stdout, "t43t44conv err: TIMESTAMP not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_timestamp = d4field (new_pvmap, "TIMESTAMP")) == 0)
    {
	fprintf (stdout, "t43t44conv err: TIMESTAMP not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_load1flag = d4field (old_pvmap, "LOAD1FLAG")) == 0)
    {
	fprintf (stdout, "t43t44conv err: LOAD1FLAG not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_load1flag = d4field (new_pvmap, "LOAD1FLAG")) == 0)
    {
	fprintf (stdout, "t43t44conv err: LOAD1FLAG not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_load2flag = d4field (old_pvmap, "LOAD2FLAG")) == 0)
    {
	fprintf (stdout, "t43t44conv err: LOAD2FLAG not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_load2flag = d4field (new_pvmap, "LOAD2FLAG")) == 0)
    {
	fprintf (stdout, "t43t44conv err: LOAD2FLAG not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_deleteflag = d4field (old_pvmap, "DELETEFLAG")) == 0)
    {
	fprintf (stdout, "t43t44conv err: DELETEFLAG not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_deleteflag = d4field (new_pvmap, "DELETEFLAG")) == 0)
    {
	fprintf (stdout, "t43t44conv err: DELETEFLAG not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_importnode = d4field (old_pvmap, "IMPORTNODE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: IMPORTNODE not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_importnode = d4field (new_pvmap, "IMPORTNODE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: IMPORTNODE not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_importdate = d4field (old_pvmap, "IMPORTDATE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: IMPORTDATE not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_importdate = d4field (new_pvmap, "IMPORTDATE")) == 0)
    {
	fprintf (stdout, "t43t44conv err: IMPORTDATE not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((old_impdelflag = d4field (old_pvmap, "IMPDELFLAG")) == 0)
    {
	fprintf (stdout, "t43t44conv err: IMPDELFLAG not exist in orig pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    if ((new_impdelflag = d4field (new_pvmap, "IMPDELFLAG")) == 0)
    {
	fprintf (stdout, "t43t44conv err: IMPDELFLAG not exist in new pvmap\n");
	d4close (old_pvmap);
	d4close (new_pvmap);
	return (1);
    }

    for (d4top (old_pvmap); !d4eof (old_pvmap); d4skip (old_pvmap, 1))
    {
	/* do not copy if the delete flag is set */
	if (f4true(old_deleteflag))
	    continue;
	d4append_start (new_pvmap, 0);
	d4blank (new_pvmap);
	f4assign_field (new_name, old_name);
	f4assign_field (new_type, old_type);
	f4assign_field (new_desc, old_desc);
	f4assign_field (new_loc, old_loc);
	f4assign_field (new_merge, old_merge);
	f4assign_field (new_stype, old_stype);
	f4assign_field (new_srate, old_srate);
	f4assign_field (new_phase, old_phase);
	f4assign_field (new_event, old_event);
	f4assign_field (new_onscan, old_onscan);
	f4assign_field (new_onalarm, old_onalarm);
	f4assign_field (new_tagdepth, old_tagdepth);
	f4assign_field (new_flnk, old_flnk);
	f4assign_field (new_rev, old_rev);
	f4assign_field (new_alarmgroup, old_alarmgroup);
	f4assign_field (new_privgroup, old_privgroup);
	f4assign_field (new_devtype, old_devtype);
	f4assign_field (new_sdisable, old_sdisable);
	f4assign_field (new_disvalue, old_disvalue);
	f4assign_field (new_resrcname, old_resrcname);
	f4assign_field (new_ioc1dnld, old_ioc1dnld);
	f4assign_field (new_ioc2dnld, old_ioc2dnld);
	f4assign_field (new_checksum, old_checksum);
	f4assign_field (new_ontrend, old_ontrend);
	f4assign_field (new_timestamp, old_timestamp);
	f4assign_field (new_load1flag, old_load1flag);
	f4assign_field (new_load2flag, old_load2flag);
	f4assign_field (new_deleteflag, old_deleteflag);
	f4assign_field (new_importnode, old_importnode);
	f4assign_field (new_importdate, old_importdate);
	f4assign_field (new_impdelflag, old_impdelflag);
	d4append (new_pvmap);
	d4flush (new_pvmap);
    }

    d4close (old_pvmap);
    d4close (new_pvmap);

    fprintf (stdout, "...finished pvmap.dbf\n");
    return (0);
}

