/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include	<stdio.h>
#ifndef AXP
#include	<string.h>
#else
#include	<strings.h>
#endif
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<errno.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include	"tagvudb.h"

static CODE4		cb;
static char		pathbuff[128];
static char		sysbuff[256];
static char		*old_rec, *new_rec;
static struct stat	statbuff;

static int create_subdir (void);
static int convert_tagvu (void);

#define	TIS4000DB	"/tis4000db"
#define	DBCNVRT		"t51tagvucnvrt"


int main()
{
    fprintf (stdout, "starting T51TAGVU conversion\n\n");
    d4init (&cb);

#if defined(WIN32) || defined(USECB63)
    cb.errGo = 0;
    cb.errOpen = 0;
    cb.errCreate = 0;
    cb.errFieldName = 0;
#else
    cb.go_error = 0;
    cb.open_error = 0;
    cb.create_error = 0;
    cb.field_name_error = 0;
#endif WIN32 or USECB63

    if (create_subdir ())
	return (1);
    if (convert_tagvu ())
	return (1);
    d4close_all (&cb);
    fprintf (stdout, "\nT51TAGVU conversion finished\n");
    return (0);
}


static int create_subdir (void)
{
    sprintf (pathbuff, "%s/%s", TIS4000DB, DBCNVRT);
    if (stat (pathbuff, &statbuff))
    {
	if (errno == ENOENT)
	{
	    sprintf (sysbuff, "mkdir %s", pathbuff);
	    system (sysbuff);
	    if (stat (pathbuff, &statbuff))
	    {
		fprintf (stdout, "t51tagvuconv err: failed to create %s\n", pathbuff);
		return (1);
	    }
	    else
	    {
		fprintf (stdout, "t51tagvuconv: %s created\n", pathbuff);
		return (0);
	    }
	}
	else
	{
	    fprintf (stdout, "t51tagvuconv err: stat problem (%d) on %s\n", errno, pathbuff);
	    return (1);
	}
    }
    else
    {
	fprintf (stdout, "t51tagvuconv err: %s already exists\n", pathbuff);
    }
    return (0);
}


static int convert_tagvu (void)
{
    DATA4	*old_tagvu, *new_tagvu;
    FIELD4	*oldptname, *newptname;
    FIELD4	*oldloc, *newloc;
    FIELD4	*olddesc, *newdesc;
    FIELD4	*olddate_apl, *newdate_apl;
    FIELD4	*oldtime_apl, *newtime_apl;
    FIELD4	*oldoper_apl, *newoper_apl;
    FIELD4	*oldreason_apl, *newreason_apl;
    FIELD4	*olddate_rm, *newdate_rm;
    FIELD4	*oldtime_rm, *newtime_rm;
    FIELD4	*oldoper_rm, *newoper_rm;
    FIELD4	*oldreason_rm, *newreason_rm;
    FIELD4	*tagtype;
    int		create_tagvu = 0;

    fprintf (stdout, "...converting tagvu.dbf\n");

    sprintf (pathbuff, "%s/tagvu.dbf", TIS4000DB);
    /* since the odl tagvu created tagvu.dbf, there may not be one
       so all we have to do is create it
    */
    if (stat (pathbuff, &statbuff) != 0)
    {
	if (errno == ENOENT)
	    create_tagvu = 1;
	else
	{
	    fprintf (stdout, "t51tagvuconv err: stat problem (%d) on %s\n", errno, pathbuff);
	    return (1);
	}
    }
    sprintf (pathbuff, "%s/tagvu", TIS4000DB);
    if (create_tagvu == 0)
    {
	if (!(old_tagvu = d4open (&cb, pathbuff)))
	{
	    fprintf (stdout, "t51tagvuconv err: open error on %s\n", pathbuff);
	    return (1);
	}
	if ((tagtype = d4field (old_tagvu, "TAGTYPE")) != 0)
	{
	    fprintf (stdout, "t51tagvuconv err: %s already converted\n", pathbuff);
	    d4close (old_tagvu);
	    return (1);
	}
	d4alias_set (old_tagvu, "old_tagvu");

	/* save the old dbase files */
	sprintf (sysbuff, "cp -p %s/tagvu.dbf %s/%s/tagvu.dbf.sav", 
			   TIS4000DB, TIS4000DB, DBCNVRT);
	system (sysbuff);
	sprintf (sysbuff, "cp -p %s/tagvu.mdx %s/%s/tagvu.mdx.sav", 
			   TIS4000DB, TIS4000DB, DBCNVRT);
	system (sysbuff);
    }

    /* create the new dbase file */
    sprintf (pathbuff, "%s/%s/tagvu", TIS4000DB,DBCNVRT);
    if (!(new_tagvu = d4create (&cb, pathbuff, tagvuFieldInfo, tagvuTagInfo)))
    {
	fprintf (stdout, "t51tagvuconv err: failed to create %s\n", pathbuff);
	if (create_tagvu == 0)
	    d4close (old_tagvu);
	return (1);
    }
    if (create_tagvu)
    {
	d4close (new_tagvu);
	fprintf (stdout, "...finished tagvu.dbf\n");
	return (0);
    }

    old_rec = d4record (old_tagvu);
    new_rec = d4record (new_tagvu);

    if ((oldptname = d4field (old_tagvu, "POINTNAME")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: POINTNAME not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newptname = d4field (new_tagvu, "POINTNAME")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: POINTNAME not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((oldloc = d4field (old_tagvu, "LOCATION")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: LOCATION not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newloc = d4field (new_tagvu, "LOCATION")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: LOCATION not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((olddesc = d4field (old_tagvu, "DESCRIPTOR")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: DESCRIPTOR not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newdesc = d4field (new_tagvu, "DESCRIPTOR")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: DESCRIPTOR not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((olddate_apl = d4field (old_tagvu, "DATE_APL")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: DATE_APL not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newdate_apl = d4field (new_tagvu, "DATE_APL")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: DATE_APL not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((oldtime_apl = d4field (old_tagvu, "TIME_APL")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: TIME_APL not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newtime_apl = d4field (new_tagvu, "TIME_APL")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: TIME_APL not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((oldoper_apl = d4field (old_tagvu, "OPER_APL")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: OPER_APL not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newoper_apl = d4field (new_tagvu, "OPER_APL")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: OPER_APL not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((oldreason_apl = d4field (old_tagvu, "REASON_APL")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: REASON_APL not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newreason_apl = d4field (new_tagvu, "REASON_APL")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: REASON_APL not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((olddate_rm = d4field (old_tagvu, "DATE_RMV")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: DATE_RMV not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newdate_rm = d4field (new_tagvu, "DATE_RMV")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: DATE_RMV not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((oldtime_rm = d4field (old_tagvu, "TIME_RMV")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: TIME_RMV not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newtime_rm = d4field (new_tagvu, "TIME_RMV")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: TIME_RMV not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((oldoper_rm = d4field (old_tagvu, "OPER_RMV")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: OPER_RMV not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newoper_rm = d4field (new_tagvu, "OPER_RMV")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: OPER_RMV not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((oldreason_rm = d4field (old_tagvu, "REASON_RMV")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: REASON_RMV not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((newreason_rm = d4field (new_tagvu, "REASON_RMV")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: REASON_RMV not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    if ((tagtype = d4field (new_tagvu, "TAGTYPE")) == 0)
    {
	fprintf (stdout, "t51tagvuconv err: TAGTYPE not exist\n");
	d4close (old_tagvu);
	d4close (new_tagvu);
	return (1);
    }

    for (d4top (old_tagvu); !d4eof (old_tagvu); d4skip (old_tagvu, 1))
    {
	d4append_start (new_tagvu, 0);
	d4blank (new_tagvu);
	f4assign_field (newptname, oldptname);
	f4assign_field (newloc, oldloc);
	f4assign_field (newdesc, olddesc);
	f4assign_field (newdate_apl, olddate_apl);
	f4assign_field (newtime_apl, oldtime_apl);
	f4assign_field (newoper_apl, oldoper_apl);
	f4assign_field (newreason_apl, oldreason_apl);
	f4assign_field (newdate_rm, olddate_rm);
	f4assign_field (newtime_rm, oldtime_rm);
	f4assign_field (newoper_rm, oldoper_rm);
	f4assign_field (newreason_rm, oldreason_rm);
	f4assign_char (tagtype, 'O');
	d4append (new_tagvu);
	d4flush (new_tagvu);
    }

    d4close (old_tagvu);
    d4close (new_tagvu);

    fprintf (stdout, "...finished tagvu.dbf\n");
    return (0);
}

