#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vbuttonHEADER
#include vcontainerHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER
#include vprefHEADER

extern "C"
{
#ifndef WIN32
#    include <sys/time.h>
#    include <sys/types.h>
#    include <sys/socket.h>
#    include <sys/ioctl.h>
#    include <sys/fcntl.h>
#    include <netinet/in.h>
#    include <netinet/tcp.h>
#    include <netdb.h>
#    include <error.h>
#else
#    include "NTUnixSubs.h"
#    include <winsock.h>
typedef char * caddr_t;
extern char *optarg;
#endif WIN32

#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "cadef.h"
#include "dbDefs.h"
#include "choice.h"
#include "choiceAo.h"
#include "choiceGbl.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vunetwork.hxx"
#include "vuCommonDialogClass.hxx"
#include "vuUtils.hxx"
#include "vuDatabaseClass.hxx"
#include "vuImportClass.hxx"

// use by the dblibrary functions and is declared extern there
// therefore I declare it global here
DB_DATABASE	DB_Database;	

//** used by vuDialogMessageCenter and vuMainMessageCenter
int	vuOutput;
// just what it says used everywhere to disable Apply functions
int	vuReadOnly;
// if cannot create a point, should not be able to update/delete a point
int	vuUpdatePointPriv = vTRUE;
int	vuDeletePointPriv = vTRUE;


/*******************************************************************
*
*    Main ()
*
*******************************************************************/

int main (int argc, char **argv)
{
    int cmd;
    int newargc;
    const vchar **newargv;
    extern char *optarg;

    char dbPath[VU_MAX_PATH];
    char newPath[VU_MAX_PATH];
    char vuImportFile[VU_MAX_PATH];
    char vuPointName[PVNAME_SZ+1];
 #ifdef WIN32
    WSADATA wsaData; 
    WORD    wVersionRequested;
    int err;
#endif WIN32

    dataVuApplication *dataVuApp = NULL;
    // Initialize Galaxy
    vstartup (argc, argv);
#ifdef WIN32
      wVersionRequested = MAKEWORD(1, 1);
      // Tell the user that we couldn't find a useable winsock.dll.
      if ((err = WSAStartup(wVersionRequested, &wsaData)) != 0) 
      {
	  exit(0);
      }
#endif WIN32
      vapplication *app = vapplication::GetCurrent();


    // messages go to NOTICES or the MAIN dialog textitem
    vuOutput = VU_NO_CONSOLE_MSG;

#ifndef WIN32
    strncpy (dbPath, "/tis4000db", VU_MAX_PATH);
    app->SetFont(vfont::FindScribed(vcharScribeLiteral("family:fixed-face:bold-size:14")));
#else
    strncpy (dbPath, "c:\\tis4000db", VU_MAX_PATH);
    app->SetFont(vfont::FindScribed(vcharScribeLiteral("family:Courier-face:medium-size:10")));
#endif WIN32
    // init the database engine
    DB_InitDatabase (argv[0]);
    //
    // Get command line parameters here
    //
    vpref::GetArgs(&newargc,&newargv);
    vuReadOnly = vFALSE;
    while ((cmd = getopt(newargc,(char **)newargv,"rc:p:i:n:")) != EOF)
    {
	switch(cmd)
	{
	    case 'r':
		vuReadOnly = vTRUE;
		break;
	    case 'c':
		if(*optarg != '-')
		{
		    strcpy(newPath,optarg);
		    vuDatabaseClass::vuCreateNewDatabase(vstrCloneScribed(vcharScribeSystem(newPath)));
		    exit (EXIT_SUCCESS);
		    return EXIT_FAILURE;
		}
		else
		fprintf(stderr,"%s: option requires an argument -- %c\n",newargv[0],cmd);
		break;
	    case 'p':
		if(*optarg != '-')
		{
		    strcpy(dbPath,optarg);
		}
		else
		    fprintf(stderr,"%s: option requires an argument -- %c\n",newargv[0],cmd);
		break;
	    case 'i':
		if(*optarg != '-')
		{
		    dataVuApp = new dataVuApplication(argv[0],dbPath);
		    if((dataVuApp->getInitErrors()) != VU_SUCCESS)
		    {
			delete dataVuApp;
			exit (EXIT_FAILURE);
		    }
		    strcpy(vuImportFile,optarg);
		    vuOutput = VU_CONSOLE_MSG;
		    vuImportClass *im = new vuImportClass(dataVuApp->getMainDialogClass(),
					  dataVuApp->getDatabaseClass(),
					  vstrCloneScribed(vcharScribeSystem(vuImportFile)));
		    im->vuImport();
		    delete im;
		    delete dataVuApp;
		    exit (EXIT_SUCCESS);
		    return EXIT_FAILURE;
		}
		else
		{
		    fprintf(stderr,"%s: option requires an argument -- %c\n",newargv[0],cmd);
		}
		break;
	    case 'n':
		if(*optarg != '-')
		{
		    char pointTypeName[16];
		    int dbaseType;
		    strcpy(vuPointName,optarg);
		    dataVuApp = new dataVuApplication(argv[0],dbPath);
		    if((dataVuApp->getInitErrors()) != VU_SUCCESS)
		    {
			delete dataVuApp;
			exit (EXIT_FAILURE);
		    }
		    vuDatabaseClass *db = dataVuApp->getDatabaseClass();
		    vresource vuRes = dataVuApp->getDataVuvresource();
		    vresource commonResource = vresourceGet (vuRes, vunames_vuCommonDialog);
		    // find the record in pvmap and get the point type
		    DB_RECORD_PTR pntRecord = DB_Seek(db->vuPvmapTable,vuPointName);
		    DB_FIELD_PTR pType = DB_MapField(db->vuPvmapTable, FLD_TYPE);
		    DB_GetFieldValue (pntRecord, pType, (void *)pointTypeName, &dbaseType);
		    // convert point type string to an index
		    int pointType = vuGetPointTypeIndex(pointTypeName);
		    vstr *pvname = vstrCloneScribed(vcharScribeSystem (vuPointName));
		    if(db->vuVerifyDb(vuPointName,pointType) != 0)
			vuNewLoadAndGo(dataVuApp->getMainDialogClass(), pvname, 
				       commonResource, pointType);
		    vstrDestroy (pvname);
		}
		else
		{
		    fprintf(stderr,"%s: option requires an argument -- %c\n",newargv[0],cmd);
		}
		break;
	    case '?':
		fprintf(stderr,"Usage is:\n");
		fprintf(stderr,"\tdataVu [-p pathname -n pointname -r -i importfile]\n");
		fprintf(stderr,"\tpathname --- location of Project files.\n");
		fprintf(stderr,"\tpointname -- Name of Process Variable.\n");
		fprintf(stderr,"\timportfile -- name of import text file.\n");
		fprintf(stderr,"\t-r --------- Read only.\n");
		exit(1);
	}
    } // end while (getopt)

    if(dataVuApp == NULL)
    {
	dataVuApp = new dataVuApplication(argv[0],dbPath);
	if((dataVuApp->getInitErrors()) != VU_SUCCESS)
	{
	    delete dataVuApp;
	    exit (EXIT_FAILURE);
	}
    }
    dataVuApp->getMainDialogClass()->vuMainMessageCenter(vunames_vuDataVuInitMsg,
						(vuReadOnly == 1 ? "In read-only mode":""));
    dataVuApp->getMainDialogClass()->Open();	// open main dialog here (deferred)

    // enter main event process loop
    veventProcess ();

    // have exited process loop, start clean up
    delete dataVuApp;

    exit (EXIT_SUCCESS);
    return EXIT_FAILURE;
}


/*******************************************************************
*
*    dataVuApplication constructor
*
*******************************************************************/

dataVuApplication::dataVuApplication (char *programName,char *dbPath)
{

    initErrors = VU_SUCCESS;
    initedTisGalaxy = vFALSE;
    vuMainDialog = NULL;
    CAInit = vFALSE;
    dbClientNotify = NULL;
    dataVuvapplication = vapplication::GetCurrent();
    dataVuvresource = dataVuvapplication->GetResources();
    // create array of defines for tag names in app
    vunamesInternNames ();


    vuDataBase = new vuDatabaseClass (dbPath, dataVuvresource);

    if ( (vuDataBase -> getInitErrors ()) != VU_SUCCESS)
    {
	initErrors = VU_ERROR;
	return;
    }

    // store the default database path
    strncpy (dbasePath, dbPath, VU_MAX_PATH);
    dbasePath [VU_MAX_PATH - 1] = 0;

    // load the main dialog
    vresource mainDialogResource = vresourceGet (dataVuvresource, 
					vunames_vuMainDialog);
    vuMainDialogClass *mainDialog = new vuMainDialogClass (mainDialogResource);

    // store main dialog in application
    vuMainDialog = mainDialog;

    if ((mainDialog -> GetLoadInitError ()) != VU_SUCCESS)
    {
	initErrors = VU_ERROR;
	return;
    }

    vresource titleRes = vresourceGet (dataVuvresource,
					vunames_vuDialogTitles);
    vresource msgRes = vresourceGet (dataVuvresource,
					vunames_vuMessagesDictionary);
    vresource defRes = vresourceGet (dataVuvresource,
					vunames_vuDefaultsDictionary);
    vresource arraysRes = vresourceGet (dataVuvresource,
					vunames_vuArraysDict);

    // init tisGalaxy, see tisgalaxylib.<hxx,cxx>
    tisGalaxyvresource = startTisGalaxy ();
    initedTisGalaxy = vTRUE;

    // store titles,messages and defaults dictionary resources in main dialog
    mainDialog -> storeTitlesResource (titleRes);
    mainDialog -> storeMessagesResource (msgRes);
    mainDialog -> storeDefaultsResource (defRes);
    mainDialog -> storeArraysResource (arraysRes);

    // store ptr to main dialog in the database class
    vuDataBase -> storeMainDialog (mainDialog);

    // store ptr to application in the main dialog
    mainDialog -> storeDataVuApp (this);
    mainDialog -> AddStyle(vwindowSTYLE_GLOBAL);
    //
    // attempt to open a socket to dbClient. This will be used to notify dbClient
    // of point status changes as a result of the Apply or Delete functions
    //
    dbClientNotify = new vuNetwork(mainDialog);
}


/*******************************************************************
*
*    dataVuApplication destructor
*
*******************************************************************/

dataVuApplication::~dataVuApplication (void)
{
    // close down tisGaalxy if used
    if (initedTisGalaxy == vTRUE)
    {
	stopTisGalaxy (tisGalaxyvresource);
    }

    //
    // close the dbClient socket
    //
    if(dbClientNotify != NULL);
        delete dbClientNotify;

    // close out the database
    DB_UnMapDatabase (0);

    // Close out channel access
    if (CAInit == vTRUE)
    {
	ca_task_exit ();
    }

    // delete the database stuff
    delete vuDataBase;

    // delete the main dialog
    if (vuMainDialog != NULL)
	delete vuMainDialog;
}


/*******************************************************************
*
*    dataVuApplication channel access timer start function
*
*******************************************************************/

void dataVuApplication::StartChanAccTimer (void)
{
    chanAccTimer = new ChanAccTimer ();
    // set to restart after timer code finished
    chanAccTimer -> SetRecurrent ();
    // set timer to expire in 15 seconds, 0 nanoseconds
    chanAccTimer -> SetPeriod (15, 0);
    chanAccTimer -> Start ();
}


/*******************************************************************
*
*    application wide classes
*
*******************************************************************/

/*******************************************************************
*
*    class that defines vtimer to do channel access pend io's
*
*******************************************************************/

void ChanAccTimer::ObserveTimer (void)
{
    // allow channel access to process events, if any
    ca_pend_io ((float)0.0);
}


/*******************************************************************
*
*    class that defines a button and a pvname input field
*    override of the virtual ObserverDialogItem function
*    if there is a name in the textbox, then we have to look it up
*    and bring that points point dialog
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuPvNameButtonClass, vbutton, "vuPvNameButtonClass");

void vuPvNameButtonClass::ObserveDialogItem (vevent *event)
{
    vchar		pvnameText [PVNAME_SZ + 1];
    char		pvname [PVNAME_SZ + 1];
    char		seekBuff [PVNAME_SZ + 1];

    // get a pointer to the textitem that may have a pvname in it
    filteredTextitem *pvName = getPvName ();
    if (pvName == NULL)
	return;
    // get text from the textitem
    pvName -> GetText (pvnameText, PVNAME_SZ);
    if (pvnameText[0] != 0)
    {
	// a field name may be part of the name, we can only use the point name
	vcharGetFirstToken (pvnameText, ".");
	// export it to a char[] so we can use it
	vcharExportBounded (vcharScribeSystemExport (pvnameText), pvname, PVNAME_SZ);
	sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvname);
	vuCommonDialogClass *commonDialog = getCommonDialog ();
	if (commonDialog != NULL)
	{
	    int	pointType;
	    int	dbaseType;
	    char	pointTypeName [POINTTYPE_SZ];

	    vuMainDialogClass *mainDialog = commonDialog -> getMainDialog ();
	    dataVuApplication *vuApp = mainDialog -> getApplication ();
	    vuDatabaseClass   *db = vuApp -> getDatabaseClass ();
	    vresource         vuRes = vuApp -> getDataVuvresource ();
	    vresource         commonResource = vresourceGet (vuRes, vunames_vuCommonDialog);

	    // does the point already exist in pvmap
	    DB_RECORD_PTR pntRecord = DB_Seek (db -> vuPvmapTable, seekBuff);
	    if (pntRecord != (DB_RECORD_PTR)0)
	    {
		// map to the point type field in pvmap
		DB_FIELD_PTR pType = DB_MapField(db -> vuPvmapTable, FLD_TYPE);
		DB_GetFieldValue (pntRecord, pType, (void *)pointTypeName, &dbaseType);
		pointType = vuGetPointTypeIndex (pointTypeName);

		if(db->vuVerifyDb(pvname,pointType) != 0)
		    vuNewLoadAndGo (mainDialog, pvnameText, commonResource, pointType);
	    }
	    else
	    {
		vuMessageDialogCenter ((vdialog *)commonDialog , 
						VU_PTNOTINPVMAPERR, pvnameText);
	    }
	}  // end if there is a dialog

    }  // end if there is a pvname
}


/**********************************************************************
*
*    class that defines boardtype menu class
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION(vuBoardTypeMenuClass, vcontrolOptionMenu, "vuBoardTypeMenuClass");

void vuBoardTypeMenuClass::ObserveDialogItem (vevent *event)
{
    // get current value of control
    int controlValue = this -> GetValue ();

    if ( (controlValue >= this -> GetMin ()) &&
	 (controlValue <= this -> GetMax ()) )
    {
	int *mod = vuGetModifiedPtr ();
	if (mod != NULL)
	    vuSetModifiedFlag ();
    }
}

void vuBoardTypeMenuClass::LoadInit (vresource resource)
{
    vcontrolOptionMenu::LoadInit (resource);
    modified = NULL;
}


/*******************************************************************
*
*    vuBuildBoardTypeMenu ()
*	passed the pointType, build a menu from the SDR and attach 
*	it to the control
*    NOTE: pointType + 1, because the point menu is indexed from 0 but
*	   the dbRecType in the SDR, "ai" is index 1, index 0 is NULL
*
*******************************************************************/
void vuBoardTypeMenuClass::vuBuildBoardTypeMenu (int pointType)
{
    int			i;
    int			pType = pointType + 1;
    int			numChoicesFound = 0;
    long		numDevChoices = 0;
    char		*pChoice;
    struct devChoice	**devChoices;
    vmenuItem		*menuItem;

    if (choiceDev == NULL)
	return;

    if (choiceDev -> papDevChoiceSet[pType])
    {
	numDevChoices = choiceDev -> papDevChoiceSet[pType] -> number;
	devChoices = choiceDev -> papDevChoiceSet[pType] -> papDevChoice;

	/* create a new menu to attach to the vcontrol item, we are 
	   creating a popup menu within a dialog
	*/
	vmenu *newMenu = new vmenu;

	for (i = 0; i < numDevChoices; i++)
	{
	    pChoice = devChoices[i] -> pchoice;
	    if (pChoice)
	    {
		menuItem = new vmenuItem;
		menuItem -> SetTitleScribed (vcharScribeSystem ((void *)pChoice));
		newMenu -> AppendItem (menuItem);
		numChoicesFound++;
	    }
	}
	if (numChoicesFound)
	{
	    // attach the menu to the control
    	    this -> SetMenu (newMenu);
	    // set min and max for the control
	    this -> SetMin (0);
	    this -> SetMax (numChoicesFound - 1);
	    // set value so current menu item will be the 1st one
	    this -> SetValue (0);
	}
    }
}


/**********************************************************************
*
*    class that defines a check box observer
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION(vuCheckBoxClass, vcontrolToggle, "vuCheckBoxClass");

void vuCheckBoxClass::ObserveDialogItem (vevent *event)
{
    int *mod = vuGetModifiedPtr ();
    if (mod != NULL)
	vuSetModifiedFlag ();
}

void vuCheckBoxClass::LoadInit (vresource resource)
{
    vcontrolToggle::LoadInit (resource);
    modified = NULL;
}


/**********************************************************************
*
*    class that defines conversion menu
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION(vuConversionMenuClass, vcontrolOptionMenu, 
						"vuConversionMenuClass");

void vuConversionMenuClass::ObserveDialogItem (vevent *event)
{
    // get pointer to menu attached to the control
    vmenu *menu = this -> GetMenu ();

    // get current value of control
    int controlValue = this -> GetValue ();

    if ( (controlValue >= this -> GetMin ()) &&
	 (controlValue <= this -> GetMax ()) )
    {
	int *mod = vuGetModifiedPtr ();
	if (mod != NULL)
	    vuSetModifiedFlag ();
    }
}

void vuConversionMenuClass::LoadInit (vresource resource)
{
    vcontrolOptionMenu::LoadInit (resource);
    modified = NULL;
}


/*******************************************************************
*
*    vuConversionMenuClass ()
*	passed the pointType, build a menu from the SDR and attach 
*	it to the control
*    NOTE: pointType + 1, because the point menu is indexed from 0 but
*	   the dbRecType in the SDR, "ai" is index 1, index 0 is NULL
*
*******************************************************************/
void vuConversionMenuClass::vuBuildConversionMenu (int pointType)
{
    int			i;
    long		numChoices = 0;
    int			numChoicesFound = 0;
    char		**choiceList;
    char		*pChoice;
    vmenuItem		*menuItem;

    switch (pointType)
    {
	case VU_AI:
	    if (choiceCvt == NULL)
		break;
	    numChoices = choiceCvt -> number;
	    choiceList = choiceCvt -> papChoice;
	    break;

	case VU_AO:
	    if (choiceRec == NULL)
		break;
	    numChoices = 
		   choiceRec -> papArrChoiceSet [VU_AO + 1] -> papChoiceSet [REC_AO_LINR] -> number;
	    choiceList = 
		   choiceRec -> papArrChoiceSet [VU_AO + 1] -> papChoiceSet [REC_AO_LINR] -> papChoice;
	    break;

	default:
	    break;
    }
    if (numChoices)
    {
    	/* create a new menu to attach to the vcontrol item, we are 
	   creating a popup menu within a dialog
	*/
	vmenu *newMenu = new vmenu;

	for (i = 0; i < numChoices; i++)
	{
	    pChoice = choiceList[i];
	    if (pChoice)
	    {
		menuItem = new vmenuItem;
		menuItem -> SetTitleScribed (vcharScribeSystem ((void *)pChoice));
		newMenu -> AppendItem (menuItem);
		numChoicesFound++;
	    }
	}
	if (numChoicesFound)
	{
	    // attach the menu to the control
	    this -> SetMenu (newMenu);
	    // set min and max for the control
	    this -> SetMin (0);
	    this -> SetMax (numChoicesFound - 1);
	    // set value so current menu item will be the 1st one
	    this -> SetValue (0);
	}
    }
}


/**********************************************************************
*
*    class that defines the choice global option menu class
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION(vuChoiceGblMenuClass, vcontrolOptionMenu, 
						"vuChoiceGblMenuClass");

void vuChoiceGblMenuClass::ObserveDialogItem (vevent *event)
{

    // get current value of control
    int controlValue = this -> GetValue ();

    if ( (controlValue >= this -> GetMin ()) &&
	 (controlValue <= this -> GetMax ()) )
    {
	int *mod = vuGetModifiedPtr ();
	if (mod != NULL)
	    vuSetModifiedFlag ();
    }
}

void vuChoiceGblMenuClass::LoadInit (vresource resource)
{
    vcontrolOptionMenu::LoadInit (resource);
    modified = NULL;
}


/*******************************************************************
*
*    vuChoiceGblMenuClass build the option menu
*    NOTE: 
*	choiceIndicies are defined in .../share/ascii/choiceGbl.h
*
*******************************************************************/

void vuChoiceGblMenuClass::vuBuildOptionMenu (int choiceIndex)
{
    int			i;
    long		numChoices;
    int			numFound = 0;
    char		**choices;
    vmenuItem		*menuItem;

    if (choiceGbl == NULL)
	return;

    // get the signal types from the SDR
    // see ../epicsH/choice.h and .../share/ascii/choice<ptname>.h

    numChoices = choiceGbl -> papChoiceSet[choiceIndex] -> number;
    choices = choiceGbl -> papChoiceSet[choiceIndex] -> papChoice;

    /* create a new menu to attach to the vcontrol item, we are 
       creating a popup menu within a dialog
    */
    vmenu *newMenu = new vmenu;

    for (i = 0; i < numChoices; i++)
    {
	if (choices[i])
	{
	    numFound++;
	    menuItem = new vmenuItem;
	    menuItem -> SetTitleScribed (vcharScribeSystem ((void *)choices[i]));
	    newMenu -> AppendItem (menuItem);
	}
    }

    // attach the menu to the control
    this -> SetMenu (newMenu);
    // set min and max for the control
    this -> SetMin (0);
    this -> SetMax (numFound - 1);
    // set value so current menu item will be the 1st one
    this -> SetValue (0);
}



/**********************************************************************
*
*    class that defines the choice rec option menu class
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION(vuChoiceRecMenuClass, vcontrolOptionMenu, 
						"vuChoiceRecMenuClass");

void vuChoiceRecMenuClass::ObserveDialogItem (vevent *event)
{
    // get pointer to menu attached to the control
    vmenu *menu = this -> GetMenu ();

    // get current value of control
    int controlValue = this -> GetValue ();

    if ( (controlValue >= this -> GetMin ()) &&
	 (controlValue <= this -> GetMax ()) )
    {
	int *mod = vuGetModifiedPtr ();
	if (mod != NULL)
	    vuSetModifiedFlag ();
    }
}

void vuChoiceRecMenuClass::LoadInit (vresource resource)
{
    vcontrolOptionMenu::LoadInit (resource);
    modified = NULL;
}


/*******************************************************************
*
*    vuChoiceRecMenuClass build the option menu
*    NOTE: pointType + 1, because the point menu is indexed from 0 but
*	   the dbRecType in th SDR, "ai" is index 1, index 0 is NULL
*	choiceIndicies are defined in .../share/ascii/choice<ptname>.h
*
*******************************************************************/

void vuChoiceRecMenuClass::vuBuildOptionMenu (int pointIndex, int choiceIndex)
{
    int			i;
    long		numTypes;
    int			numFound = 0;
    char		**choices;
    vmenuItem		*menuItem;

    if (choiceRec == NULL)
	return;

    // get the signal types from the SDR
    // see ../epicsH/choice.h and .../share/ascii/choice<ptname>.h

    numTypes = 
	choiceRec->papArrChoiceSet[pointIndex + 1]->papChoiceSet[choiceIndex]->number;
    choices = 
	choiceRec->papArrChoiceSet[pointIndex + 1]->papChoiceSet[choiceIndex]->papChoice;

    /* create a new menu to attach to the vcontrol item, we are 
       creating a popup menu within a dialog
    */
    vmenu *newMenu = new vmenu;

    for (i = 0; i < numTypes; i++)
    {
	if (choices[i])
	{
	    numFound++;
	    menuItem = new vmenuItem;
	    menuItem -> SetTitleScribed (vcharScribeSystem ((void *)choices[i]));
	    newMenu -> AppendItem (menuItem);
	}
    }

    // attach the menu to the control
    this -> SetMenu (newMenu);
    // set min and max for the control
    this -> SetMin (0);
    this -> SetMax (numFound - 1);
    // set value so current menu item will be the 1st one
    this -> SetValue (0);
}


/**********************************************************************
*
*    class that overrides filteredTextitem class
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION(vuFilteredTextitem, filteredTextitem, "vuFilteredtextitem");

