#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vlistviewHEADER
#include vlistitemHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vdateHEADER
#include vspinnerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbDefs.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuNavigatorDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vuUtils.hxx"
#include "vuAlmGrpSelDialogClass.hxx"


vkindTYPED_FULL_DEFINITION(vuAlmGrpSelDialogClass, vdialog, "vuAlmGrpSelDialogClass");

/*******************************************************************
*
*    vuAlmGrpSelDialogClass 
*	LoadInit, constructor
*
*******************************************************************/

void vuAlmGrpSelDialogClass::LoadInit(vresource res)
{
    vmemSet ((void *)alarmGroup, 0, sizeof (alarmGroup));

    alarmGrpListitem = vuAlmGrpSelListitem::RequestSubLoad (this,
					vunames_vuAlarmGrpSelListItem);
    okButton = vuAlmGrpSelOKButtonClass::RequestSubLoad (this,
					vunames_vuAlarmGrpSelOkButton);
    dismissButton = vuAlmGrpSelDismissButtonClass::RequestSubLoad (this,
					vunames_vuAlarmGrpSelDismissButton);

    // call base class loadInit to finish the job
    vdialog::LoadInit(res);

    // store the dialog in listitem for later use
    alarmGrpListitem -> vuSetAlmGrpSelDialog (this);

    // store the dialog in the button for later use
    okButton -> vuSetAlmGrpDialog (this);

    // store the dialog in the button for later use
    dismissButton -> vuSetAlmGrpDialog (this);

    // place window in center of screen
    Place(GetRoot(), vrectPLACE_CENTER, vrectPLACE_CENTER);		
}


/*******************************************************************
*
*    vuAlmGrpSelDialogClass notify function
*
*******************************************************************/

void vuAlmGrpSelDialogClass::ObserveWindow(vevent *event)
{
    if ( event -> GetType() == veventWINDOW_STATE && event -> IsClose())
    {
	// zap alarmGroup, so caller of dialog will act on a non selection
	// from the dialog
	vmemSet ((void *)alarmGroup, 0, ALARMGRP_SZ);
	alarmGroup[ALARMGRP_SZ] = 0;
	Close();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuAlmGrpSelDialogClass Destroy function
*	if dialog is open, close it and stop its event loop
*
*******************************************************************/

void vuAlmGrpSelDialogClass::Destroy(void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }
    vliststr *oldList = (vliststr *)alarmGrpListitem->GetList();
    if(oldList != NULL)  oldList->Destroy();

    vdialog::Destroy();
}


/*******************************************************************
*
*    vuAlmGrpSelDialogClass Open function
*	build the alarm group list before we open dialog
*
*******************************************************************/

void vuAlmGrpSelDialogClass::vuAlmGrpSelOpen(vuMainDialogClass *mainDialog)
{
    dataVuApplication	*app = mainDialog -> getApplication();
    vuDatabaseClass	*db = app -> getDatabaseClass();

    vliststr *oldList = (vliststr *)alarmGrpListitem->GetList();
    if(oldList != NULL)  oldList->Destroy();
    // function in vuUtils.cxx
    oldList = vuGetAlarmGroupList (db);
    oldList -> Sort (vlistROW, 0, NULL);
    alarmGrpListitem->SetList(oldList);
    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the alarmGrp after the dialog is gone
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


vkindTYPED_FULL_DEFINITION(vuAlmGrpSelListitem, vlistview, "vuAlmGrpSelListitem");

/*******************************************************************
*
*    vuAlmGrpSelListitem selection notify function
*	SELECTION_HAS_CHANGED = user finally release the mouse
*	button and commits the new selection, if we have a double
*	mouse click, then take same action as APPLY BUTTON
*
*******************************************************************/

void vuAlmGrpSelListitem::ObserveSelect(int state)
{
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	int clickCount = GetLastPointerClickCount ();
	if (clickCount == 2)
	{
	    vchar	*alarmGrp = almGrpSelDialog -> vuGetAlarmGroup ();
	    vlistSelection *selected = GetSelection ();
	    vliststr *list = (vliststr *)GetList ();
	    int row, column;
	    list -> GetSelectedCell (selected, &row, &column);
	    vchar *itemValue = (vchar *)(list -> GetCellValue (row, 0));
	    // copy value to dialogs alarmGroup array for use by dialog caller
	    vcharCopyBounded (itemValue, alarmGrp, ALARMGRP_SZ);
	    // close the dialog and stop the sub event loop
	    // started in vuAlmGrpSelOpen ()
	    almGrpSelDialog -> Close ();
	    vevent::StopProcessing ();
	}
    }
}


vkindTYPED_FULL_DEFINITION(vuAlmGrpSelOKButtonClass, vbutton, 
			   "vuAlmGrpSelOKButtonClass");

/*******************************************************************
*
*    vuAlmGrpSelOKButtonClass 
*	Observer function
*
*******************************************************************/

void vuAlmGrpSelOKButtonClass::ObserveDialogItem (vevent *event)
{
    vchar	*itemValue;
    vchar	*alarmGrp;

    // the pointer to the dialog is stored in the button class
    // use it to get the listitem 
    vuAlmGrpSelListitem *listitem = almGrpSelDialog -> vuGetAlmGrpListitem ();
    // get pointer to alarmGroup in dialog  so it can be set for later use 
    // by caller of dialog
    alarmGrp = almGrpSelDialog -> vuGetAlarmGroup ();
    vlistSelection *selected = listitem -> GetSelection ();
    vliststr *list = (vliststr *)(listitem -> GetList ());
    int row, column;
    list -> GetSelectedCell (selected, &row, &column);
    if (row != vlistNO_CELLS_SELECTED)
    {
	itemValue = (vchar *)(list -> GetCellValue (row, 0));
	vcharCopyBounded (itemValue, alarmGrp, ALARMGRP_SZ);
    }
    else
    {
	vmemSet ((void *)alarmGrp, 0, ALARMGRP_SZ);
    }
    // close the dialog and stop the sub event loop
    // started in vuAlmGrpSelOpen ()
    almGrpSelDialog -> Close ();
    vevent::StopProcessing ();
}



vkindTYPED_FULL_DEFINITION(vuAlmGrpSelDismissButtonClass, vbutton, 
			   "vuAlmGrpSelDismissButtonClass");

/*******************************************************************
*
*    vuAlmGrpSelDismissButtonClass 
*	Observer function
*
*******************************************************************/

void vuAlmGrpSelDismissButtonClass::ObserveDialogItem (vevent *event)
{
    vchar	*alarmGrp;

    alarmGrp = almGrpSelDialog -> vuGetAlarmGroup ();
    // zap alarmGrp, so caller of dialog will act on a non selection
    // from the dialog
    vmemSet ((void *)alarmGrp, 0, ALARMGRP_SZ);
    alarmGrp[ALARMGRP_SZ] = 0;
    // close the dialog and stop the sub event loop
    // started in vuAlmGrpSelOpen ()
    almGrpSelDialog -> Close ();
    vevent::StopProcessing ();
}

