#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vnumHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbFldTypes.h"
#include "choiceGbl.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "dbRecDes.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vuCommonDialogClass.hxx"
#include "vuAoDialogClass.hxx"
#include "vuNavigatorDialogClass.hxx"
#include "vuUtils.hxx"


vkindTYPED_FULL_DEFINITION(vuAoDialogClass, vuCommonDialogClass, "vuAoDialogClass");

/*******************************************************************
*
*    vuAoDialogClass override for constructors init ()
*
*******************************************************************/

void vuAoDialogClass::LoadInit (vresource res)
{
    // load in the ai specific fields
    aoContainer = vuAoContainerClass::RequestSubLoad (this,
				vunames_vuCommonContainer);
    vuCommonDialogClass::LoadInit (res);

    // store dialog modified flag
    aoContainer -> boardType -> vuSetModifiedPtr (&this -> dialogModified);
    aoContainer -> line -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> addr -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> card -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> signal -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> convertType -> vuSetModifiedPtr (&this -> dialogModified);
    aoContainer -> valueLoc -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> output -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> constantValue -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> closedLoop -> vuSetModifiedPtr (&this -> dialogModified);
    aoContainer -> onScan -> vuSetModifiedPtr (&this -> dialogModified);
    aoContainer -> onAlarm -> vuSetModifiedPtr (&this -> dialogModified);
    aoContainer -> incremental -> vuSetModifiedPtr (&this -> dialogModified);
    aoContainer -> monitorDeadband -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> archiveDeadband -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> alarmDeadband -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> rampRate -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> outputGap -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> precision -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> eguTag -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> irLow -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> irHigh -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> eguLow -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> eguHigh -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> displayLow -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> displayHigh -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> clampLow -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> clampHigh -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> alarmDisable -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> alarmDelay -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> readback -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> hihiLimit -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> highLimit -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> lowLimit -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> loloLimit -> SetTextModifiedPtr (&this -> dialogModified);
    aoContainer -> hihiSeverity -> vuSetModifiedPtr (&this -> dialogModified);
    aoContainer -> highSeverity -> vuSetModifiedPtr (&this -> dialogModified);
    aoContainer -> lowSeverity -> vuSetModifiedPtr (&this -> dialogModified);
    aoContainer -> loloSeverity -> vuSetModifiedPtr (&this -> dialogModified);
    aoContainer -> alarmGroup -> SetTextModifiedPtr (&this -> dialogModified);

    // store dialog ptr for vuPvNameButtonClass
    aoContainer -> valueLocButton -> storeCommonDialog ((vuCommonDialogClass *)this);
    aoContainer -> outputButton -> storeCommonDialog ((vuCommonDialogClass *)this);
    aoContainer -> alarmDisableButton -> storeCommonDialog ((vuCommonDialogClass *)this);

    // store dialog ptr for vuPvNameFilter
    aoContainer -> valueLoc -> storeCommonDialog ((vuCommonDialogClass *)this);
    aoContainer -> output -> storeCommonDialog ((vuCommonDialogClass *)this);
    aoContainer -> alarmDisable -> storeCommonDialog ((vuCommonDialogClass *)this);

    // store dialog ptr for vuAlarmGroupFilter
    aoContainer -> alarmGroup -> storeCommonDialog ((vuCommonDialogClass *)this);

    // set the point type
    setPointType (VU_AO);

    // get the title from the resource file, aiTitle must be
    // freed by the class destuctor
    vapplication *app = vapplication::GetCurrent ();
    vresource dialogTitles = vresourceGet (app -> GetResources (),
				vunames_vuDialogTitles);
    aoTitle = vresourceGetString (dialogTitles, 
				vunames_vuAoDialogTitle);
    SetTitle (aoTitle);
}


/*******************************************************************
*
*    vuAoDialogClass Destroy function
*
*******************************************************************/

void vuAoDialogClass::Destroy (void)
{
    vstrDestroy (aoTitle);
    vuCommonDialogClass::Destroy ();
}


/*******************************************************************
*
*    vuAoDialogClass, reset fields, overrides base class 
*	vuCommonDialogClass
*
*******************************************************************/

void vuAoDialogClass::resetDialog (void)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    vchar		epics[FLDNAME_SZ + 1];

    vresource defRes = mainDialog -> getDefaultsResource ();

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0; 
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	// if not a common dialog field
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    dialogItem = vuFindContainerItem (aoContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    if ( (dialogItem -> IsKindOf (&vtextitem::Kind)) == vTRUE)
	    {
		textItem = (vtextitem *)dialogItem;

		// set the default value for the archive deadband field
		if ( (vcharCompareBounded (epics, "ADEL", FLDNAME_SZ)) == 0)
		{
		    vstr *adel = vresourceGetString (defRes, 
						vunames_vuADELdefault);
		    textItem -> SetText (adel);
		    vstrDestroy (adel);
		}
		// set the default value for the Ao's ramp rate field
		else if ( ((vcharCompareBounded (epics, "OROC", FLDNAME_SZ)) == 0) ||
			  ((vcharCompareBounded (epics, "OGAP", FLDNAME_SZ)) == 0) ||
			  ((vcharCompareBounded (epics, "LINE", FLDNAME_SZ)) == 0) ||
			  ((vcharCompareBounded (epics, "ADDR", FLDNAME_SZ)) == 0) ||
			  ((vcharCompareBounded (epics, "CARD", FLDNAME_SZ)) == 0) ||
			  ((vcharCompareBounded (epics, "SIGN", FLDNAME_SZ)) == 0) ||
			  ((vcharCompareBounded (epics, "RBDT", FLDNAME_SZ)) == 0) )
		{
		    vstr *zero = vresourceGetString (defRes, 
						vunames_vuZeroDefault);
		    textItem -> SetText (zero);
		    vstrDestroy (zero);
		}
		else if ( (vcharCompareBounded (epics, "PREC", FLDNAME_SZ)) == 0)
		{
		    vstr *precision = vresourceGetString (defRes, 
						vunames_vuPRECdefault);
		    textItem -> SetText (precision);
		    vstrDestroy (precision);
		}
		else
		{
		    textItem -> SetTextScribed (vcharScribeSystem (""));
		}
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolToggle::Kind)) == vTRUE)
	    {
		toggleItem = (vcontrolToggle *)dialogItem;
		toggleItem -> SetValue (0);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolOptionMenu::Kind)) == vTRUE)
	    {
		optionMenu = (vcontrolOptionMenu *)dialogItem;
		optionMenu -> SetValue (0);
	    }
	}  // end if not in common dialog
    }
    iterator.Finish ();

    // call base class reset function with proper point index
    vuCommonDialogClass::resetDialog ();

    dialogModified = vFALSE;
}


/*******************************************************************
*
*    vuLoadDialog 
*
*******************************************************************/

void vuAoDialogClass::vuLoadDialog (vchar *pointName)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    vchar		epics[FLDNAME_SZ + 1];
    vstr		*finalValue;
    vstr		*referenceString;
    vstr		*constantString;

    char		pvName [PVNAME_SZ + 1];
    DB_TABLE_PTR	lastTable;	// pointer to last table used
    DB_RECORD_PTR	pntRecord;	// codebase record ptr
    int			dbaseType;	// codebase field type
    int			intValue;
    unsigned int	uIntValue;
    long		longValue;
    unsigned long	uLongValue;
    double		doubleValue;
    DTYP_TYPE		dType;		// device type indicator

    vcharExportBounded (vcharScribeSystemExport (pointName),
			pvName,
			PVNAME_SZ);

    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
    pntRecord = DB_Seek (db -> vuPvmapTable, seekBuff);
    if (pntRecord == (DB_RECORD_PTR)0)
    {
	return;
    }
    else
    {
	int ptType;
	DB_GetFieldValue (pntRecord, db -> pvmapTypeField, dbValue, &dbaseType);
	ptType = vuGetPointTypeIndex ((char *)dbValue);
	// the point name in pvname field is NOT an AO
	if (ptType != VU_AO)
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_PTTYPEDIALOGERR,
				   pointName, ptType, VU_AO);
	    vuClearPvNameField ();
	    return;
	}
	// the point name entered in pvname field has been deleted
	if (DB_IsDeleted (db -> vuPvmapTable))
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_PTISDELETEDERR, pointName);
	    vuClearPvNameField ();
	    return;
	}
    }
    // call base class vuLoadDialog first
    vuCommonDialogClass::vuLoadDialog (pointName);

    // since the ao can be a hardware point, find what we are for later use
    // by getting the DTYP (device type) dictionary entry
    referenceString = vstrCloneScribed (vcharScribeSystem ("REFERENCE"));
    constantString = vstrCloneScribed (vcharScribeSystem ("CONSTANT"));
    xrefNode = (xrefNodeClass *)(dict -> GetElem (vnameGetGlobalScribed (vcharScribeSystem ("DTYP"))));
    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
    pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
    DB_GetFieldValue (pntRecord, xrefNode -> xrefField, dbValue, &dbaseType);
    if ( (vcharCompare ((const vchar *)dbValue, referenceString)) == 0)
	dType = VU_REFERENCE;
    else if ( (vcharCompare ((const vchar *)dbValue, constantString)) == 0)
	dType = VU_CONSTANT;
    else
	dType = VU_HARDWARE;
    vstrDestroy (referenceString);
    vstrDestroy (constantString);

    lastTable = NULL;

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0;
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    // only seek if table has changed
	    if (lastTable != xrefNode -> xrefTable)
	    {
		sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
		pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
	    }
	    lastTable = xrefNode -> xrefTable;
	    dialogItem = vuFindContainerItem (aoContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    DB_GetFieldValue (pntRecord, xrefNode -> xrefField, dbValue, &dbaseType);
	    doubleValue = 0.0;
	    // GetField actually returns a number in the char array
	    if ( (dbaseType == 'F') || (dbaseType == 'N') )
	    {
		doubleValue = *((double *)dbValue);
	    }
	    switch (xrefNode -> xrefEPICSFieldType)
	    {
		case DBF_SHORT:
		    intValue = (int)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeInt (intValue));
		    break;

		case DBF_LONG:
		    longValue = (long)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeLong (longValue));
		    break;

		case DBF_ENUM:
		case DBF_USHORT:
		    uIntValue = (unsigned int)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeUInt (uIntValue));
		    break;

		case DBF_ULONG:
		    uLongValue = (unsigned long)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeULong (uLongValue));
		    break;

		case DBF_FLOAT:
		case DBF_DOUBLE:
		    // to output CONSTANT field, device type must be CONSTANT
		    if ( (strncmp (xrefNode -> xrefEpicsName, "CNST", FLDNAME_SZ)) == 0)
		    {
			if (dType != VU_CONSTANT)
			    continue;
		    }
		    finalValue = vstrCloneScribed (vnumScribeFloat (doubleValue));
		    break;

		case DBF_STRING:
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		case DBF_GBLCHOICE:
		case DBF_RECCHOICE:
		case DBF_CVTCHOICE:
		case DBF_DEVCHOICE:
		    if (dbaseType == 'L')
		    {
			int	logical;

			logical = (*((char*)dbValue) ? 
					xrefNode -> xrefTrue : 
					(xrefNode -> xrefTrue ^1));
			finalValue = vstrCloneScribed (vnumScribeInt (logical));
			break;
		    }
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		case DBF_OUTLINK:
		    // device type must be REFERENCE to output the reference
		    if (dType == VU_REFERENCE)
		    {
			DB_GetFieldValue (pntRecord, xrefNode -> xrefRefField, 
								dbValue, &dbaseType);
			finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    }
		    else 
		    {
			finalValue = vstrCloneScribed (vcharScribeSystem (""));
		    }
		    break;

		case DBF_INLINK:
		    // works for all ao inputs, see pdbxref
		    DB_GetFieldValue (pntRecord, xrefNode -> xrefRefField, 
								dbValue, &dbaseType);
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		default:
		    continue;
	    }
	    if ( (dialogItem -> IsKindOf (&vtextitem::Kind)) == vTRUE)
	    {
		textItem = (vtextitem *)dialogItem;
		textItem -> SetTextScribed (vcharScribeSystem (""));
		textItem -> SetText (finalValue);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolToggle::Kind)) == vTRUE)
	    {
		toggleItem = (vcontrolToggle *)dialogItem;
		if (finalValue[0] == '0')
		    toggleItem -> SetValue (0);
		else
		    toggleItem -> SetValue (1);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolOptionMenu::Kind)) == vTRUE)
	    {
		optionMenu = (vcontrolOptionMenu *)dialogItem;
		vuSetOptionMenuValue (optionMenu, finalValue);
		vstrDestroy (finalValue);
	    }
	}  // end if not in commondialog
    }
    iterator.Finish ();

    setDialogModify (vFALSE);
}


/*******************************************************************
*
*    vuSaveDialog 
*
*******************************************************************/

void vuAoDialogClass::vuSaveDialog (vchar *pointName)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    const vchar 	*optionMenuValue;
    vchar		epics[FLDNAME_SZ + 1];
    vstr		*stringValue;
    vstr		*referenceString;
    vstr		*constantString;

    char		pvName [PVNAME_SZ + 1];
    char		vuCurrentBoard [VU_KEY_SIZE];
    DB_FIELD_PTR	putField;
    DB_TABLE_PTR	lastTable;	// pointer to last table used
    DB_RECORD_PTR	pntRecord;	// codebase record ptr
    int			dType;		// device type
    double		doubleValue;
    int			doSave;

    vcharExportBounded (vcharScribeSystemExport (pointName),
			pvName,
			PVNAME_SZ);

    // since the ao can be a hardware point, find what we are for later use
    // by getting the DTYP (device type) dictionary entry
    referenceString = vstrCloneScribed (vcharScribeSystem ("REFERENCE"));
    constantString = vstrCloneScribed (vcharScribeSystem ("CONSTANT"));
    vcharCopyScribedBounded (vcharScribeSystem ("DTYP"), epics, FLDNAME_SZ);
    epics[sizeof (epics) - 1] = 0;
    // find the BOARD (DTYP) field and get its value
    dialogItem = vuFindContainerItem (aoContainer, epics);
    optionMenu = (vcontrolOptionMenu *)dialogItem;
    optionMenuValue = vuGetOptionMenuValue (optionMenu);
    vcharExport (vcharScribeSystemExport (optionMenuValue), (void *)vuCurrentBoard);
    if ( (vcharCompare (optionMenuValue, referenceString)) == 0)
	dType = VU_REFERENCE;
    else if ( (vcharCompare (optionMenuValue, constantString)) == 0)
	dType = VU_CONSTANT;
    else
	dType = VU_HARDWARE;
    vstrDestroy (referenceString);
    vstrDestroy (constantString);

    lastTable = NULL;

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0;
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    doSave = vTRUE;
	    // only seek if table has changed
	    if (lastTable != xrefNode -> xrefTable)
	    {
		sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
		pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
	    }
	    lastTable = xrefNode -> xrefTable;

	    putField = xrefNode -> xrefField;

	    dialogItem = vuFindContainerItem (aoContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    switch (xrefNode -> xrefEPICSFieldType)
	    {
		case DBF_SHORT:
		case DBF_LONG:
		case DBF_ENUM:
		case DBF_USHORT:
		case DBF_ULONG:
		case DBF_FLOAT:
		case DBF_DOUBLE:
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    doubleValue = atof ((const char *)dbValue);
		    if ( (strncmp (xrefNode -> xrefEpicsName, "CNST", FLDNAME_SZ))
										== 0)
		    {
			if (dType != VU_CONSTANT)
			    doubleValue = 0;
		    }
		    DB_PutFieldValue (pntRecord, putField, (void *)&doubleValue);
		    doSave = vFALSE;
		    break;

		case DBF_STRING:
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    break;

		case DBF_GBLCHOICE:
		case DBF_RECCHOICE:
		case DBF_CVTCHOICE:
		case DBF_DEVCHOICE:
		{
		    if (xrefNode -> xrefField -> field_type == 'L')
		    {
			int	logical;
			toggleItem = (vcontrolToggle *)dialogItem;
			logical = toggleItem -> GetValue ();
			*((char *)dbValue) = (logical ? xrefNode -> xrefTrue : 
							    xrefNode -> xrefTrue ^1);
			break;
		    }
		    optionMenu = (vcontrolOptionMenu *)dialogItem;
		    optionMenuValue = vuGetOptionMenuValue (optionMenu);
		    vcharExport (vcharScribeSystemExport (optionMenuValue), dbValue);
		    if ( (strncmp (xrefNode -> xrefEpicsName, "DTYP", FLDNAME_SZ)) == 0)
		    {
			// write to DEVTYPE in pvmap
			DB_PutFieldValue (pntRecord, xrefNode -> xrefField, dbValue);
		    }
		    break;
		}

		case DBF_OUTLINK:
		    // there is only 1 output and that is OUT
		    // always write the boardtype from DTYP into the OUT field
		    DB_PutFieldValue (pntRecord, putField, vuCurrentBoard);
		    // if we are a reference then write OUT into the REFERENCE field
		    if (dType == VU_REFERENCE)
		    {
			putField = xrefNode -> xrefRefField;
			textItem = (vtextitem *)dialogItem;
			stringValue = textItem -> GetTextAsString ();
			vcharExport (vcharScribeSystemExport (stringValue), dbValue);
			vstrDestroy (stringValue);
		    }
		    else
		    {
			// make sure to null out the field
			putField = xrefNode -> xrefRefField;
			vmemSet (dbValue, 0, VU_KEY_SIZE);
		    }
		    break;

		case DBF_INLINK:
		    // for DOL, we must put its value to the OUTPUTREF field
		    // other wise it and other INLINKS are the same 
		    if ( (strncmp (xrefNode -> xrefEpicsName, "DOL ", FLDNAME_SZ)) == 0)
		    {
			putField = xrefNode -> xrefRefField;
		    }
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    break;

		default:
		    continue;

	    }  // end xrefNode -> xrefEPICSFieldType

	    if (doSave == vTRUE)
	    {
		DB_PutFieldValue (pntRecord, putField, dbValue);
	    }

	}  // end if not in commondialog

    }  // end while iterator

    iterator.Finish ();

    // call base class vuSaveDialog
    vuCommonDialogClass::vuSaveDialog (pointName);
}


vkindTYPED_FULL_DEFINITION(vuAoContainerClass, vcontainer, "vuAoContainerClass");

/*******************************************************************
*
*    vuAoContainerClass override for constructor
*
*******************************************************************/

void vuAoContainerClass::LoadInit (vresource res)
{
    vapplication *app = vapplication::GetCurrent ();
    vresource pointDictRes = vresourceGet (app -> GetResources (), 
				vunames_vuPointDictionary);
    vresource realRes = vresourceGet (pointDictRes, vunames_vuAoContainer);

    vcontainer *btcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoDTYPContainer);
    boardType = vuBoardTypeMenuClass::RequestSubLoad (btcontainer,
				vunames_DTYP);

    vcontainer *linecontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoLINEContainer);
    line = filteredTextitem::RequestSubLoad (linecontainer,
				vunames_LINE);

    vcontainer *addrcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoADDRContainer);
    addr = filteredTextitem::RequestSubLoad (addrcontainer,
				vunames_ADDR);

    vcontainer *cardcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoCARDContainer);
    card = filteredTextitem::RequestSubLoad (cardcontainer,
				vunames_CARD);

    vcontainer *signalcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoSIGNContainer);
    signal = filteredTextitem::RequestSubLoad (signalcontainer,
				vunames_SIGN);

    vcontainer *cvrtcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoLINRContainer);
    convertType = vuConversionMenuClass::RequestSubLoad (cvrtcontainer,
				vunames_LINR);

    vcontainer *valcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoDOLContainer);
    valueLocButton = vuPvNameButtonClass::RequestSubLoad (valcontainer,
				vunames_vuAoDOLButton);
    valueLoc = vuFilteredTextitem::RequestSubLoad (valcontainer,
				vunames_DOL);

    vcontainer *outcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoOUTContainer);
    outputButton = vuPvNameButtonClass::RequestSubLoad (outcontainer,
				vunames_vuAoOUTButton);
    output = vuFilteredTextitem::RequestSubLoad (outcontainer,
				vunames_OUT);

    vcontainer *cnstcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoCNSTContainer);
    constantValue = filteredTextitem::RequestSubLoad (cnstcontainer,
				vunames_CNST);

    closedLoop = vuCheckBoxClass::RequestSubLoad (this,
				vunames_OMSL);

    incremental = vuCheckBoxClass::RequestSubLoad (this,
				vunames_OIF);

    onScan = vuCheckBoxClass::RequestSubLoad (this,
				vunames_OSCN);

    onAlarm = vuCheckBoxClass::RequestSubLoad (this,
				vunames_AENB);

    vcontainer *mdbcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoMDELContainer);
    monitorDeadband = filteredTextitem::RequestSubLoad (mdbcontainer,
				vunames_MDEL);

    vcontainer *archcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoADELContainer);
    archiveDeadband = filteredTextitem::RequestSubLoad (archcontainer,
				vunames_ADEL);

    vcontainer *alarmdbcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoHYSTContainer);
    alarmDeadband = filteredTextitem::RequestSubLoad (alarmdbcontainer,
				vunames_HYST);

    vcontainer *rampcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoOROCContainer);
    rampRate = filteredTextitem::RequestSubLoad (rampcontainer,
				vunames_OROC);

    vcontainer *ogapcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoOGAPContainer);
    outputGap = filteredTextitem::RequestSubLoad (ogapcontainer,
				vunames_OGAP);

    vcontainer *preccontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoPRECContainer);
    precision = filteredTextitem::RequestSubLoad (preccontainer,
				vunames_PREC);

    vcontainer *egutagcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoEGUContainer);
    eguTag = filteredTextitem::RequestSubLoad (egutagcontainer,
				vunames_EGU);

    vcontainer *irlocontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoIRLOContainer);
    irLow = filteredTextitem::RequestSubLoad (irlocontainer,
				vunames_IRLO);

    vcontainer *irhicontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoIRHIContainer);
    irHigh = filteredTextitem::RequestSubLoad (irhicontainer,
				vunames_IRHI);

    vcontainer *egulowcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoEGULContainer);
    eguLow = filteredTextitem::RequestSubLoad (egulowcontainer,
				vunames_EGUL);

    vcontainer *eguhicontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoEGUFContainer);
    eguHigh = filteredTextitem::RequestSubLoad (eguhicontainer,
				vunames_EGUF);

    vcontainer *dsplowcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoLOPRContainer);
    displayLow = filteredTextitem::RequestSubLoad (dsplowcontainer,
				vunames_LOPR);

    vcontainer *dsphicontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoHOPRContainer);
    displayHigh = filteredTextitem::RequestSubLoad (dsphicontainer,
				vunames_HOPR);

    vcontainer *clamplowcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoDRVLContainer);
    clampLow = filteredTextitem::RequestSubLoad (clamplowcontainer,
				vunames_DRVL);

    vcontainer *clamphicontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoDRVHContainer);
    clampHigh = filteredTextitem::RequestSubLoad (clamphicontainer,
				vunames_DRVH);

    vcontainer *alarmdiscontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoADISContainer);
    alarmDisableButton = vuPvNameButtonClass::RequestSubLoad (alarmdiscontainer,
				vunames_vuAoADISButton);
    alarmDisable = vuFilteredTextitem::RequestSubLoad (alarmdiscontainer,
				vunames_ADIS);

    vcontainer *alarmdelcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoALDLContainer);
    alarmDelay = filteredTextitem::RequestSubLoad (alarmdelcontainer,
				vunames_ALDL);

    vcontainer *readbackcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoRBDTContainer);
    readback = filteredTextitem::RequestSubLoad (readbackcontainer,
				vunames_RBDT);

    vcontainer *alrmgrpcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoALGPContainer);
    alarmGroup = vuFilteredTextitem::RequestSubLoad (alrmgrpcontainer,
				vunames_ALGP);

    vcontainer *alarmcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuAoAlarmContainer);

    vcontainer *hihilimcontainer = vcontainer::RequestSubLoad (alarmcontainer,
				vunames_vuAoHIHIContainer);
    hihiLimit = filteredTextitem::RequestSubLoad (hihilimcontainer,
				vunames_HIHI);
    vcontainer *hihisevcontainer = vcontainer::RequestSubLoad (alarmcontainer,
				vunames_vuAoHHSVContainer);
    hihiSeverity = vuChoiceGblMenuClass::RequestSubLoad (hihisevcontainer,
				vunames_HHSV);

    vcontainer *highlimcontainer = vcontainer::RequestSubLoad (alarmcontainer,
				vunames_vuAoHIGHContainer);
    highLimit = filteredTextitem::RequestSubLoad (highlimcontainer,
				vunames_HIGH);
    vcontainer *highsevcontainer = vcontainer::RequestSubLoad (alarmcontainer,
				vunames_vuAoHSVContainer);
    highSeverity = vuChoiceGblMenuClass::RequestSubLoad (highsevcontainer,
				vunames_HSV);

    vcontainer *lowlimcontainer = vcontainer::RequestSubLoad (alarmcontainer,
				vunames_vuAoLOWContainer);
    lowLimit = filteredTextitem::RequestSubLoad (lowlimcontainer,
				vunames_LOW);
    vcontainer *lowsevcontainer = vcontainer::RequestSubLoad (alarmcontainer,
				vunames_vuAoLSVContainer);
    lowSeverity = vuChoiceGblMenuClass::RequestSubLoad (lowsevcontainer,
				vunames_LSV);

    vcontainer *lololimcontainer = vcontainer::RequestSubLoad (alarmcontainer,
				vunames_vuAoLOLOContainer);
    loloLimit = filteredTextitem::RequestSubLoad (lololimcontainer,
				vunames_LOLO);
    vcontainer *lolosevcontainer = vcontainer::RequestSubLoad (alarmcontainer,
				vunames_vuAoLLSVContainer);
    loloSeverity = vuChoiceGblMenuClass::RequestSubLoad (lolosevcontainer,
				vunames_LLSV);

    vcontainer::LoadInit (realRes);

    // store ptr to text field for vuPvNameButtonClass
    valueLocButton -> storePvName (valueLoc);
    outputButton -> storePvName (output);
    alarmDisableButton -> storePvName (alarmDisable);

    // set filters
    line -> SetItemFilterFunc (numericFilter);
    addr -> SetItemFilterFunc (numericFilter);
    card -> SetItemFilterFunc (numericFilter);
    signal -> SetItemFilterFunc (numericFilter);

    valueLoc -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *vText = (filteredText *)valueLoc -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    vText -> SetUserPtr ((void *)valueLoc);

    output -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *outText = (filteredText *)output -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    outText -> SetUserPtr ((void *)output);

    constantValue -> SetItemFilterFunc (floatFilter);
    monitorDeadband -> SetItemFilterFunc (floatFilter);
    archiveDeadband -> SetItemFilterFunc (floatFilter);
    alarmDeadband -> SetItemFilterFunc (floatFilter);
    rampRate -> SetItemFilterFunc (floatFilter);
    outputGap -> SetItemFilterFunc (floatFilter);
    precision -> SetItemFilterFunc (numericFilter);
    eguTag -> SetItemFilterFunc (egutagFilter);
    irLow -> SetItemFilterFunc (floatFilter);
    irHigh -> SetItemFilterFunc (floatFilter);
    eguLow -> SetItemFilterFunc (floatFilter);
    eguHigh -> SetItemFilterFunc (floatFilter);
    displayLow -> SetItemFilterFunc (floatFilter);
    displayHigh -> SetItemFilterFunc (floatFilter);
    clampLow -> SetItemFilterFunc (floatFilter);
    clampHigh -> SetItemFilterFunc (floatFilter);

    alarmDisable -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *adText = (filteredText *)alarmDisable -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    adText -> SetUserPtr ((void *)alarmDisable);

    alarmDelay -> SetItemFilterFunc (numericFilter);
    readback -> SetItemFilterFunc (numericFilter);
    hihiLimit -> SetItemFilterFunc (floatFilter);
    highLimit -> SetItemFilterFunc (floatFilter);
    lowLimit -> SetItemFilterFunc (floatFilter);
    loloLimit -> SetItemFilterFunc (floatFilter);

    alarmGroup -> SetItemFilterFunc (vuAlarmGroupFilter);
    filteredText *alarmText = (filteredText *)alarmGroup -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    alarmText -> SetUserPtr ((void *)alarmGroup);

    // build popup menus
    boardType    -> vuBuildBoardTypeMenu (VU_AO);
    convertType  -> vuBuildConversionMenu (VU_AO);
    hihiSeverity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    highSeverity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    lowSeverity  -> vuBuildOptionMenu (GBL_ALARM_SEV);
    loloSeverity -> vuBuildOptionMenu (GBL_ALARM_SEV);

}


/*******************************************************************
*
*    vuAoContainerClass override for destructor
*
*******************************************************************/

void vuAoContainerClass::Destroy (void)
{
    vcontainer::Destroy ();
}

