#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vtextitemHEADER
#include vstartupHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vstrHEADER
#include vnoticeHEADER
#include vlimitsHEADER
#include vlistviewHEADER
#include vliststrHEADER
#include vstdargHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER
#include vrulerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "choice.h"
#include "choiceGbl.h"
#include "dbFldTypes.h"
#include "dbDefs.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "dbRecDes.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuUtils.hxx"
#include "vuDatabaseClass.hxx"
#include "vuExportClass.hxx"


vkindTYPED_FULL_DEFINITION(vuExportClass, tisProgressDialog, NULL);

void vuExportClass::Destroy (void)
{
    fclose(fp);
    if (dbValue != NULL)
	vmemFree (dbValue);
    tisProgressDialog::Destroy ();
}

void vuExportClass::ObserveWindow(vevent *event)
{
    // Start the work process after the dialog has been 
    // successfully drawn so that we don't get any ugly, 
    // half-drawn dialogs on screen.

    if (event -> GetType () == veventWINDOW_STATE && event->IsOpen())
    {
	ExportProgressEvent *event = new ExportProgressEvent;
	event -> SetClientData(this);
	event -> SetPriority(veventPRIORITY_MIN);
	// Start Work
	event -> Post();
    }
}


vbool vuExportClass::DoWork (double *percentCompleted)
{
    DB_RECORD_PTR rec;
    char typeString[PVNAME_SZ+1];
    char paddedname[PVNAME_SZ+1];
    int dataType;

    vbool doWork = ((!IsQuitting ()) && (IsOpen ()) && (listIndex < numPts));

    if (doWork)
    {
	sprintf(paddedname,"%-*.*s", PVNAME_SZ, PVNAME_SZ, selectedPoints[listIndex]);
	// seek to point
	if((rec = DB_Seek(dataBase->vuPvmapTable, paddedname)) != 0)
	{
	    // get point type string
	    DB_GetFieldValue (rec,dataBase->pvmapTypeField,typeString,&dataType);
	    // type first
	    fprintf(fp,"#PT#\t=\t\"%s\"\n",typeString);
	    // then point name
	    fprintf(fp,"NAME\t=\t\"%s\"\n",selectedPoints[listIndex]);
	    // get type index
	    int index = vuGetPointTypeIndex(typeString);
	    // export one record at a time
	    vuExportOneRecord(paddedname,index,typeString);
	    // also do ar_reduc
	    vuExportArReduc(paddedname,index);
	    // space between records
	    fprintf(fp,"\n");
	    
	    // update progress indicator every so often
	    if (listIndex % 25 == 0)
	    {
		*percentCompleted = (double)(listIndex + 1) / (double)numPts;
		if ((*percentCompleted) <= 1.0)
		{
		    UpdateDomain (percentCompleted);
		    char buff[256];
		    sprintf (buff, "Export %d%% complete", (int) ((*percentCompleted) * 100));
		    messageText -> SetTextScribed (vcharScribeSystem (buff));
		}
	    }
	}
	listIndex++;
    }
    return (doWork);
}


/******************************************************************************
*
*   vuExportClass set parameters and list functions
*
*******************************************************************************/

void vuExportClass::SetParameters(vuDatabaseClass *db, vstr *path)
{
    char *fileName;
    dataBase = db;
    fileName = (char *)vcharExportAlloc(vcharScribeSystemExport((vchar *)path));
    fp = fopen(fileName,"w");
    vcharExportFree(fileName);
    vstrDestroy(path);
    dbValue = vmemAllocAndClear(VU_KEY_SIZE + 1);
    listIndex = 0;;
}

void vuExportClass::SetList (char **selectedPts)
{
    selectedPoints = selectedPts;
    numPts = 0;
    int i;
    for (i = 0; selectedPoints[i]; i++)
	numPts++;
}


/************************************************************************
*
*   vuExportClass export record function
*	
************************************************************************/

void vuExportClass::vuExportOneRecord(char *pv,int index,char *typeString)
{
    xrefNodeClass   *xrefNode;		// Xref information node pointer
    vdict	    *dict = dataBase->getXrefDict (index);
    lastTable = NULL;
    vname	    *nameKey;		// used for vdict lookup
    struct fldDes   *vuFldDes;		// ptr to epics field description from SDR
    struct recTypDes	*vuRecTypDes;	// ptr to record type description from SDR
    VU_EXTRAFIELDS	*localExtraFields; // ptr to extra fields not in SDR

	// while there are fields for this point type in the SDR
	vuRecTypDes = GET_PRECTYPDES(index + 1);
	fieldIndex = 0;
	while (vuFldDes = GET_PFLDDES(vuRecTypDes,fieldIndex))
	{
	    // position us in the xref table, if a prompted field or the NOT the NAME field
	    if((vuFldDes -> promptflag) && (strcmp(vuFldDes->fldname,"NAME") != 0))
	    {
		sprintf(key,"%-*.*s\0",FLDNAME_SZ,FLDNAME_SZ,vuFldDes->fldname);
		if((nameKey = (vname *)vnameGetGlobalScribed(vcharScribeSystem(key))) != 0)
		{
		    if((xrefNode = (xrefNodeClass *)(dict -> GetElem(nameKey)))!= 0)
		    {
			vuProcessFields(xrefNode,pv,index);
		    }
		}
	    }
	    fieldIndex++;
	}
	//
	// Do the extra fields which are not in the SDR but are in the XrefNodeClass for
	// the given point type.
	//
	fieldIndex = 0;
	localExtraFields = dataBase->xrefExtraFields [index];

	while (localExtraFields[fieldIndex].xtraName)
	{
	    sprintf(key,"%-*.*s\0",FLDNAME_SZ,FLDNAME_SZ,localExtraFields[fieldIndex].xtraName);
	    nameKey = (vname *)vnameGetGlobalScribed(vcharScribeSystem(key));
	    xrefNode = (xrefNodeClass *)(dict -> GetElem(nameKey));
	    vuProcessFields(xrefNode,pv,index);
	    fieldIndex++;
	}
}


/************************************************************************
*
*   vuExportClass export ar reduct function
*	
************************************************************************/

void vuExportClass::vuExportArReduc(char *pv,int typeIndex)
{
    char	**arReduc = NULL;
    char	pointName[PVNAME_SZ + 1];
    char	*pvname;
    char	*rate;
    char	*method;
    char	*archive;
    char	*blank;
    int		numStrings = 0;
    int		index;

    strncpy (pointName, pv, sizeof (pointName) -1);
    blank = strchr (pointName, ' ');
    if (blank != NULL)
	*blank = 0;
    arReduc = dbFindInArreduc (dataBase->vuArReducTable, pointName, typeIndex+1, &numStrings);
    index = 0;
    if (arReduc != NULL)
    {
	while (arReduc[index])
	{
	    pvname = strtok (arReduc[index], ",");
	    rate   = strtok (NULL, ",");
	    method = strtok (NULL, ",");
	    archive = strtok (NULL, ",");
	    fprintf (fp, "#AR#\t%-*.*s\t%-5.5s\t%-3.3s\t%c\n",
			 PVNAME_SZ, PVNAME_SZ, pvname, rate, method, archive[0]);
	    free ((void *)arReduc[index]);
	    index++;
	}
	free ((void *)arReduc);
    }
}


/************************************************************************
*
*   vuExportClass set the scan type function
*	
************************************************************************/

void vuExportClass::vuSetScanType (DB_TABLE_PTR table,DB_RECORD_PTR rec,char *dbvalue)
{
    DB_FIELD_PTR	rateField;
    double		rate;
    int			type;

    rateField = DB_MapField (table, FLD_SCANRATE);

    if ( (strcmp (dbvalue, "Periodic")) == 0)
    {
	DB_GetFieldValue (rec, rateField, &rate, &type);
	if (rate > 5.0)
	    sprintf (dbvalue, "%2.0f second\0", rate);
	else if (rate > 0.9)
	    sprintf (dbvalue, "%1.0f second\0", rate);
	else if (rate < 0.15)
	    strcpy (dbvalue, "100 ms");
	else if (rate < 0.25)
	    strcpy (dbvalue, "200 ms");
	else
	    strcpy (dbvalue, "500 ms");
    }
}


/************************************************************************
*
*   vuExportClass process fields function
*	
************************************************************************/

void vuExportClass::vuProcessFields(xrefNodeClass *xrefNode,char *pv,int index)
{
    // temp value locations
    int		    intValue;
    unsigned int    uIntValue;
    long	    longValue;
    unsigned long   uLongValue;
    double	    doubleValue;
    DB_RECORD_PTR   pntRecord;		// codebase record ptr
    vstr	    *finalValue;	// final value to be printed
    int		    dbaseType;		// codebase field type
    char	    *valString;
    char	    tempName[PVNAME_SZ+1];
    char	    *blank;
    char	    seekBuff[PVNAME_SZ+1];

    strcpy (tempName, pv);
    if(strlen(xrefNode->xrefOpId))
    {
	if(strlen(xrefNode->xrefOpId))
	{
	    sprintf (key,"%-*.*s%-s\0",PVNAME_SZ,PVNAME_SZ,tempName,xrefNode->xrefOpId);
	    pntRecord = DB_Seek(xrefNode->xrefTable,key);
	}
    }
    else
    {
	if((xrefNode->xrefTable == dataBase->vuIOAddrTable) && (index == VU_PAT))
	{
	    // For PAT, the entry in the IOADDR file will be quadruplicated. 
	    // To avoid this, the point name will have a '1','2','3', or '4' 
	    // appended depending on the EPICS field name.
	    // POUT (pos channel) = 1, NOUT (neg channel) = 2, 
	    // RLIN (rl channel) = 3, INP (fb channel) = 4
	    blank = strchr (tempName,' ');
	    if (blank != NULL)
		*blank = 0;
	    if((strncmp((const char *)xrefNode->xrefEpicsName, "LINE", FLDNAME_SZ) == 0) ||
			(strncmp((const char *)xrefNode->xrefEpicsName, "ADDR", FLDNAME_SZ) == 0) ||
			(strncmp((const char *)xrefNode->xrefEpicsName, "CARD", FLDNAME_SZ) == 0) ||
			(strncmp((const char *)xrefNode->xrefEpicsName, "SIGN", FLDNAME_SZ) == 0))
	    {
		strcat(tempName, "1");
	    }
	    else if((strncmp((const char *)xrefNode->xrefEpicsName, "LIN2", FLDNAME_SZ) == 0) ||
			     (strncmp((const char *)xrefNode->xrefEpicsName, "ADR2", FLDNAME_SZ) == 0) ||
			     (strncmp((const char *)xrefNode->xrefEpicsName, "CRD2", FLDNAME_SZ) == 0) ||
			     (strncmp((const char *)xrefNode->xrefEpicsName, "SIG2", FLDNAME_SZ) == 0))
	    {
		strcat(tempName, "2");
	    }
	    else if((strncmp((const char *)xrefNode->xrefEpicsName, "LIN3", FLDNAME_SZ) == 0) ||
			     (strncmp((const char *)xrefNode->xrefEpicsName, "ADR3", FLDNAME_SZ) == 0) ||
			     (strncmp((const char *)xrefNode->xrefEpicsName, "CRD3", FLDNAME_SZ) == 0) ||
			     (strncmp((const char *)xrefNode->xrefEpicsName, "SIG3", FLDNAME_SZ) == 0))
	    {
		strcat(tempName, "3");
	    }
	    else if((strncmp ((const char *)xrefNode->xrefEpicsName, "LIN4", FLDNAME_SZ) == 0) ||
			      (strncmp((const char *)xrefNode->xrefEpicsName, "ADR4", FLDNAME_SZ) == 0) ||
			      (strncmp((const char *)xrefNode->xrefEpicsName, "CRD4", FLDNAME_SZ) == 0) ||
			      (strncmp((const char *)xrefNode->xrefEpicsName, "SIG4", FLDNAME_SZ) == 0))
	    {
		strcat(tempName, "4");
	    }
	}
	if((xrefNode->xrefTable == dataBase->vuIOAddrTable) && (index == VU_PTO))
	{
	    // For PTO, the entry in the IOADDR file will be duplicated. 
	    // To avoid this, the point name will have a '1','2'
	    // appended depending on the EPICS field name.
	    // POUT (pos channel) = 1, NOUT (neg channel) = 2, 
	    blank = strchr (tempName,' ');
	    if (blank != NULL)
		*blank = 0;
	    if((strncmp ((const char *)xrefNode->xrefEpicsName, "LINE", FLDNAME_SZ) == 0) ||
			 (strncmp((const char *)xrefNode->xrefEpicsName, "ADDR", FLDNAME_SZ) == 0) ||
			 (strncmp((const char *)xrefNode->xrefEpicsName, "CARD", FLDNAME_SZ) == 0) ||
			 (strncmp((const char *)xrefNode->xrefEpicsName, "SIGN", FLDNAME_SZ) == 0))
	    {
		strcat (tempName, "1");
	    }
	    else if((strncmp ((const char *)xrefNode->xrefEpicsName, "LIN2", FLDNAME_SZ) == 0) ||
			      (strncmp((const char *)xrefNode->xrefEpicsName, "ADR2", FLDNAME_SZ) == 0) ||
			      (strncmp((const char *)xrefNode->xrefEpicsName, "CRD2", FLDNAME_SZ) == 0) ||
			      (strncmp((const char *)xrefNode->xrefEpicsName, "SIG2", FLDNAME_SZ) == 0))
	    {
		strcat (tempName, "2");
	    }
	}
	sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, tempName);
	pntRecord = DB_Seek(xrefNode->xrefTable,seekBuff);
    }
    lastTable = xrefNode -> xrefTable;
    DB_GetFieldValue (pntRecord, xrefNode -> xrefField, dbValue, &dbaseType);
    doubleValue = 0.0;
    // GetField actually returns a number in the char array
    if ( (dbaseType == 'F') || (dbaseType == 'N') )
    {
	doubleValue = *((double *)dbValue);
    }
    switch (xrefNode -> xrefEPICSFieldType)
    {
	case DBF_SHORT:
	    intValue = (int)doubleValue;
	    finalValue = vstrCloneScribed(vnumScribeInt (intValue));
	    break;
	case DBF_LONG:
	    longValue = (long)doubleValue;
	    finalValue = vstrCloneScribed(vnumScribeLong (longValue));
	    break;
	case DBF_ENUM:
	case DBF_USHORT:
	    uIntValue = (unsigned int)doubleValue;
	    finalValue = vstrCloneScribed(vnumScribeUInt (uIntValue));
	    break;
	case DBF_ULONG:
	    uLongValue = (unsigned long)doubleValue;
	    finalValue = vstrCloneScribed(vnumScribeULong (uLongValue));
	    break;
	case DBF_FLOAT:
	case DBF_DOUBLE:
	    finalValue = vstrCloneScribed(vnumScribeFloat (doubleValue));
	    break;
	case DBF_STRING:
	    finalValue = vstrCloneScribed(vcharScribeSystem (dbValue));
	    break;
	case DBF_GBLCHOICE:
	case DBF_RECCHOICE:
	case DBF_CVTCHOICE:
	case DBF_DEVCHOICE:
	    if(dbaseType == 'L')
	    {
		int logical;
		logical = (*((char*)dbValue) ? xrefNode -> xrefTrue : (xrefNode -> xrefTrue ^1));
		sprintf((char *)dbValue,"%c\0",(logical ? 'T' : 'F'));
		finalValue = vstrCloneScribed (vcharScribeSystem(dbValue));
		break;
	    }
	    if ((strncmp(xrefNode->xrefEpicsName,"SCAN",FLDNAME_SZ)) == 0)
	    {
		vuSetScanType(xrefNode -> xrefTable, pntRecord, (char*)dbValue);
	    }
	    finalValue = vstrCloneScribed(vcharScribeSystem (dbValue));
	    break;
	case DBF_INLINK:
	case DBF_OUTLINK:
	    if(!strncmp(xrefNode->xrefEpicsName,"SDIS",FLDNAME_SZ))
	    {
		;
	    }
	    if(xrefNode->xrefRefField)
	    {
		DB_GetFieldValue(pntRecord,xrefNode->xrefRefField,dbValue,&dbaseType);
	    }
	    finalValue = vstrCloneScribed(vcharScribeSystem (dbValue));
	    break;
	case DBF_FWDLINK:
	    finalValue = vstrCloneScribed(vcharScribeSystem (dbValue));
	    break;
	default:
	    return;
    }
    valString = (char *)vcharExportAlloc(vcharScribeSystemExport((vchar *)finalValue));
    fprintf(fp,"%s\t=\t\"%s\"\n",xrefNode->xrefEpicsName,valString);
    vcharExportFree(valString);
    vstrDestroy(finalValue);
}


/**************************************************************
*
* ProgressEvent::Handle()
*
*   This is the basic work procedure.
*   It is called to do one iteration's worth of work.
*   It then reposts its continuation event so that it will get 
*   called again
*
***************************************************************/

int ExportProgressEvent::Handle()
{
    double  percentCompleted = 0.0;
    // Get iteration value
    vuExportClass *progressDlg = (vuExportClass *)this->GetClientData();

    // call users work function
    vbool doWork = progressDlg -> DoWork (&percentCompleted);

    // If we're not done, repost ourself.
    if (doWork && (percentCompleted <= 1.0))
    {
	// still more work to do
	Post();
    }
    else if (!progressDlg -> IsQuitting())
    {
	// Our work is done
	progressDlg -> Close();
    }
    else if (progressDlg -> IsQuitting ())
    {
	// the cancel button has been pressed
	progressDlg -> Close ();
    }
    else
    {
	// The user has quit from the progress dialog from the window
	// manager level. We don't want to update the bar any more so
	// do nothing here.
    }
   
    // return TRUE to indicate that we handled this event.
    return (vTRUE);
}
