#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vcharHEADER
#include vnameHEADER
#include vresourceHEADER
#include vdialogHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vlistviewHEADER
#include vliststrHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbDefs.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vuIocEditDialogClass.hxx"
#include "vuUtils.hxx"
#include <memory.h>


vkindTYPED_FULL_DEFINITION(vuIocEditDialogClass, vdialog, "vuIocEditDialogClass");

/*******************************************************************
*
*    vuIocEditDialogClass 
*	LoadInit, constructor
*
*******************************************************************/

void vuIocEditDialogClass::LoadInit(vresource res)
{
    // resource name
    vcontainer *container = vcontainer::RequestSubLoad (this,
					vunames_vuIOCContainer1);	
    iocName = filteredTextitem::RequestSubLoad (container,
					vunames_vuIocEditName);

    // and the list
    iocList = vuIocListClass::RequestSubLoad(this,
					vunames_vuIocEditNameList);
    iocList->vuSetIocEditDialog(this);

    // then the toggle buttons
    hostToggle = vcontrolToggle::RequestSubLoad(this,
					vunames_vuIocEditIsHost);
    syncToggle = vcontrolToggle::RequestSubLoad(this,
					vunames_vuIocEditOKSync);

    // finally the buttons
    buttonContainer = vcontainer::RequestSubLoad(this, 
					vunames_vuIOCContainer2);
    okButton = vuIocOkButtonClass::RequestSubLoad(buttonContainer,
					vunames_vuIocEditOK);
    dismissButton = vuIocDismissButtonClass::RequestSubLoad(buttonContainer,
					vunames_vuIocEditDismiss);
    delButton = vuIocDelButtonClass::RequestSubLoad(buttonContainer,
					vunames_vuIocEditDelete);

    okButton -> vuSetIocEditDialog(this);
    dismissButton -> vuSetIocEditDialog(this);
    delButton -> vuSetIocEditDialog(this);

    // call base class loadInit to finish the job
    vdialog::LoadInit(res);
    //
    // handle readonly state for the apply and delete buttons
    //
    if(vuReadOnly == vTRUE)
    {
	okButton->Disable();
	delButton->Disable();
    }
}


/*******************************************************************
*
*    vuIocEditDialogClass 
*	open function
*
*******************************************************************/

void vuIocEditDialogClass::vuIocEditOpen(vuMainDialogClass *mDialog)
{
    mainDialog = mDialog;
    dataVuApplication *app = mainDialog->getApplication();
    db = app->getDatabaseClass();
    char **iocListStr = NULL;
    vliststr *iocListData;
    vlistSelection *selection;

    // destroy any existing listdata
    iocListData = (vliststr *)iocList -> GetList();	
    if(iocListData != NULL)
	iocListData -> Destroy();

    // then create a new IOC list.
    iocListData = vuGetIocList(db);
	//  tell list view what to look at
    iocList -> SetList(iocListData);
	// set so only one iocource name can be selected
    selection = iocList -> GetSelection();
    iocListData -> SetSelectMethod(selection, vlistONE_ONLY_SELECTION);

    pIoc    = DB_MapField (db->vuIocListTable, FLD_IOCNAME);
    pOkSync = DB_MapField (db->vuIocListTable, FLD_NOSYNC);
    pIsHost = DB_MapField (db->vuIocListTable, FLD_ISHOST);

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process();
}
/*******************************************************************
*
*    vuIocEditDialogClass notify function
*
*******************************************************************/

void vuIocEditDialogClass::ObserveWindow(vevent *event)
{
    if ( event -> GetType() == veventWINDOW_STATE && event -> IsClose())
    {
	Close();
	vevent::StopProcessing();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuIocEditDialogClass Destroy function
*	if dialog is open, close it and stop its event loop
*
*******************************************************************/

void vuIocEditDialogClass::Destroy(void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }
    vliststr *listdata;
    listdata = (vliststr *)iocList -> GetList();
    if(listdata) 
	listdata->Destroy();
    vdialog::Destroy();
}

vkindTYPED_FULL_DEFINITION(vuIocOkButtonClass, vbutton, "vuIocOkButtonClass");
/*******************************************************************
*
*    vuIocOkButtonClass 
*	Observer function
*
*******************************************************************/
void vuIocOkButtonClass::ObserveDialogItem (vevent *event)
{
    char iocname[IOCNAME_SZ + 1];
    DB_RECORD_PTR pntRecord;

    vstr *str1 = ioceditDialog -> iocName -> GetTextAsString();
    vcharExportBounded (vcharScribeSystemExport (str1),
			iocname,
			IOCNAME_SZ);
    iocname[IOCNAME_SZ] = 0;
    vstrDestroy (str1);

    if(strlen(iocname))
    {
	int newRecord = vFALSE;
	// Seek to record for this ioc name. If found then just write
	// out new data. If not then create new record and then fill it in
	// and add to list.
	if((pntRecord = DB_Seek(ioceditDialog -> db -> vuIocListTable, iocname)) == 0)
	{
	    pntRecord = DB_NewRecord (ioceditDialog -> db -> vuIocListTable, NULL);
	    DB_PutFieldValue(pntRecord, ioceditDialog -> pIoc, iocname);
	    newRecord = vTRUE;
	    ioceditDialog->db->IocListHasChanged(vTRUE);
	}
	char state = (char)ioceditDialog -> syncToggle -> GetValue() & 1;
	DB_PutFieldValue(pntRecord,ioceditDialog -> pOkSync, &state);
	state = (char)ioceditDialog -> hostToggle -> GetValue () & 1;
	DB_PutFieldValue(pntRecord, ioceditDialog -> pIsHost, &state);

	// if a new record, after adding stuff to the table, add to the list
	// and select the new row, this will cause the lists observer to fire
	if (newRecord == vTRUE)
	{
	    vliststr *iocListData = (vliststr *)(ioceditDialog -> iocList -> GetList ());
	    int rowCount = iocListData -> GetRowCount ();
	    iocListData -> SetCellValueScribed(rowCount, 0, vcharScribeSystem(iocname));
	    vlistSelection *selection = ioceditDialog -> iocList -> GetSelection ();
	    iocListData -> SelectRow (selection, rowCount);
	}
    }
}


vkindTYPED_FULL_DEFINITION(vuIocDismissButtonClass, vbutton, "vuIocDismissButtonClass");

/*******************************************************************
*
*    vuIocDismissButtonClass 
*	Observer function
*
*******************************************************************/

void vuIocDismissButtonClass::ObserveDialogItem (vevent *event)
{
    // close dialog and stop event loop started in vuNavOpen ()
    ioceditDialog -> Close();
    vevent::StopProcessing();
}


vkindTYPED_FULL_DEFINITION(vuIocDelButtonClass, vbutton, "vuIocDelButtonClass");

/*******************************************************************
*
*    vuIocDelButtonClass 
*	Observer function
*
*******************************************************************/

void vuIocDelButtonClass::ObserveDialogItem (vevent *event)
{
    char iocname[IOCNAME_SZ + 1];
    DB_RECORD_PTR pntRecord;

    vstr *str1 = ioceditDialog -> iocName -> GetTextAsString();
    vcharExportBounded (vcharScribeSystemExport (str1),
			iocname,
			IOCNAME_SZ);

    if(strlen(iocname))
    {
	// Seek to record for this ioc name. If found then just delete it
	if((pntRecord = DB_Seek(ioceditDialog -> db -> vuIocListTable, iocname)) == (DB_RECORD_PTR)1)
	{
	    DB_DeleteRecord(ioceditDialog -> db -> vuIocListTable, pntRecord);

	    // also delete it from the ioc list
	    vliststr *iocListData = (vliststr *)ioceditDialog -> iocList -> GetList();
	    vlistIterator iterator;
	    int row;
	    vchar *rowValue;
	    iterator.Start (iocListData);
	    while (iterator.Next())
	    {
		row = iterator.GetRow ();
		rowValue = (vchar *)(iterator.GetValue());
		if ((vcharCompare (str1, rowValue)) == 0)
		{
		    iocListData -> RemoveRow (row);
		    int rowCount = iocListData -> GetRowCount ();
		    if (rowCount > 0)
		    {
			vlistSelection *selected = ioceditDialog -> iocList -> GetSelection ();
			iocListData -> SelectRow (selected, 0);
		    }
		    break;
		}
	    }
	    iterator.Finish();
	    ioceditDialog->db->IocListHasChanged(vTRUE);
	}
    }

    vstrDestroy(str1);
}


vkindTYPED_FULL_DEFINITION(vuIocListClass, vlistview, "vuIocListClass");

/*******************************************************************
*
*    vuIocListClass selection notify function
*	SELECTION_HAS_CHANGED = user finally released the mouse
*	button to establish the new selection.
*
*******************************************************************/

void vuIocListClass::ObserveSelect(int state)
{
    int row,column;
    char ioc[IOCNAME_SZ + 1];
    const vchar *iocName;
    char ishost,oksync;
    int ftype;
    DB_RECORD_PTR pntRecord;

    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	vliststr *ptListData = (vliststr *)GetList();
	vlistSelection *listSelection = GetSelection();
	ptListData -> GetSelectedCell (listSelection, &row, &column);
	if (row == vlistNO_CELLS_SELECTED)
	{
	    return;
	}
	iocName = (const vchar *) (ptListData -> GetCellValue(row, 0));
	ioceditDialog -> iocName -> SetText (iocName);
	//
	// Find record in ioclist.dbf and fill in other fields
	//
	vcharExportBounded (vcharScribeSystemExport (iocName),
			    ioc,
			    IOCNAME_SZ);
	ioc[IOCNAME_SZ] = 0;
	char seekBuff[IOCNAME_SZ + 1];
	sprintf (seekBuff, "%-*.*s\0", IOCNAME_SZ, IOCNAME_SZ, ioc);
	pntRecord = DB_Seek(ioceditDialog -> db -> vuIocListTable, seekBuff);
	DB_GetFieldValue(pntRecord, ioceditDialog -> pIsHost, &ishost, &ftype);
	DB_GetFieldValue(pntRecord, ioceditDialog -> pOkSync, &oksync, &ftype);
	ioceditDialog -> hostToggle -> SetValue((int)(ishost & 1));
	ioceditDialog -> syncToggle -> SetValue((int)(oksync & 1));
    }
}
