#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vlistviewHEADER
#include vliststrHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vnumHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER
#include vrulerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbDefs.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"

#include "vuMkPtsDialogClass.hxx"
#include "vuUtils.hxx"


vkindTYPED_FULL_DEFINITION(vuMkPtsDialogClass, vuFindDialogClass, 
						"vuMkPtsDialogClass");

/*******************************************************************
*
*    vuMkPtsDialogClass 
*	LoadInit, constructor
*
*******************************************************************/

void vuMkPtsDialogClass::LoadInit(vresource res)
{
    // set to NULL
    dialogTitle = NULL;

    mkPtsContainer = vuMkPtsContainerClass::RequestSubLoad (this,
				vunames_vuFindCommonContainer);

    vcontainer *buttonContainer = vcontainer::RequestSubLoad (this,
				vunames_vuFindButtonContainer);
    
    dismissButton = vuMkPtsDismissButtonClass::RequestSubLoad (buttonContainer,
				vunames_vuFindDismissButton);
    findButton = vuMkPtsFindButtonClass::RequestSubLoad (buttonContainer,
				vunames_vuFindFindButton);

    // call base class loadInit to finish the job
    vuFindDialogClass::LoadInit(res);

    // set pointer to navigator dialog in these items

    mkPtsContainer -> markButton -> setMkPtsDialog (this);
    mkPtsContainer -> unmarkButton -> setMkPtsDialog (this);
    mkPtsContainer -> ruler -> setMkPtsDialog (this);
    findButton -> setMkPtsDialog (this);
    dismissButton -> setMkPtsDialog (this);
	//
	// handle readonly state for the apply and delete buttons
	//
	if(vuReadOnly == vTRUE)
	{
		mkPtsContainer -> markButton->Disable();
		mkPtsContainer -> unmarkButton->Disable();
	}

}


/*******************************************************************
*
*    vuMkPtsDialogClass notify function
*
*******************************************************************/

void vuMkPtsDialogClass::ObserveWindow(vevent *event)
{
    if ( event -> GetType() == veventWINDOW_STATE && event -> IsClose())
    {
	Close ();
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    vuMkPtsDialogClass Destroy function
*	if dialog is open, close it and stop its event loop
*
*******************************************************************/

void vuMkPtsDialogClass::Destroy(void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }
    // get the list attached to the listview and destroy it, also set the 
    // foundList in the find class to NULL because the base class Destroy will
    // also try to destroy it if it is not NULL
    vliststr *listdata = (vliststr *)(mkPtsContainer -> pointList -> GetList ());
    if (listdata != NULL)
    {
	mkPtsContainer -> pointList -> SetList (NULL);
	listdata -> Destroy ();
	foundList = NULL;
    }

    vstrDestroy (dialogTitle);

    vuFindDialogClass::Destroy();
}


/*******************************************************************
*
*    vuMkPtsDialogClass Open function
*
*******************************************************************/

void vuMkPtsDialogClass::vuFindOpen(vchar *label,vuMainDialogClass *mainDialg,int multiSelect)
{

    mainDialog = mainDialg;

    // set points found label
    mkPtsContainer -> pointsFoundLabel -> SetTitleScribed (vcharScribeSystem (" "));
	//mkPtsContainer->markButton->Disable();
	//mkPtsContainer->unmarkButton->Disable();

    // call base class open
    vuFindDialogClass::vuFindOpen (label, mainDialg,multiSelect);

    if (dialogTitle == NULL)
    {
	vresource titleRes  = mainDialog -> getTitlesResource ();
	dialogTitle = vresourceGetString (titleRes, vunames_vuMkPtsDialogTitle);
	SetTitle (dialogTitle);
    }

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open();
    vevent::Process ();
}


vkindTYPED_FULL_DEFINITION(vuMkPtsRulerClass, vruler, "vuMkPtsRulerClass");

/*******************************************************************
*
*    vuMkPtsRulerClass  notify function
*
*******************************************************************/

void vuMkPtsRulerClass::ObserveRuler(vevent *event, vrulerItem *item, int message)
{
    if (message == vrulerITEM_SELECT)
    {
	vuMkPtsContainerClass *mkptsContainer = mkptsDialog -> getMkPtsContainer ();
	vliststr *list = (vliststr *)(mkptsContainer -> pointList -> GetList ());
	if (list == NULL)
	{
	    return;
	}
	vlistSelection *selection = mkptsContainer -> pointList -> GetSelection ();
	list -> UnselectAll (selection);
	int index = item -> GetIndex ();

	// have clicked on PVNAME, so select entire list
	if (index == 0)
	{
	    list -> SelectAll (selection);
	}
	// else select one of the otehr columns
	else
	{
	    list -> SelectColumn (selection, index);
	}
    }
}


vkindTYPED_FULL_DEFINITION(vuMkPtsMarkButtonClass, vbutton, 
					"vuMkPtsMarkButtonClass");

/*******************************************************************
*
*    vuMkPtsMarkButtonClass 
*	Observer function
*
*******************************************************************/

void vuMkPtsMarkButtonClass::ObserveDialogItem (vevent *event)
{
    vuMkPtsContainerClass *mkptsContainer = mkptsDialog -> getMkPtsContainer ();
    vliststr *list = (vliststr *)(mkptsContainer -> pointList -> GetList ());
    if (list == NULL)
    {
	return;
    }
    vlistSelection *selection = mkptsContainer -> pointList -> GetSelection ();
    int row, column;
    list -> GetSelectedCell (selection, &row, &column);
    if (row == vlistNO_CELLS_SELECTED)
    {
	vuMessageDialogCenter (mkptsDialog, VU_NOTHING_SELECTED_MSG);
	return;
    }
    vuMainDialogClass   *mainDialog = mkptsDialog -> getMainDialog ();
    vresource titleRes = mainDialog -> getTitlesResource ();
    vstr  *markedStr   = vresourceGetString (titleRes, vunames_vuMkPtsMarkedTitle);
    vstr  *spaceStr    = vstrCloneScribed (vcharScribeLiteral (" "));
    vstr  *asterickStr = vstrCloneScribed (vcharScribeLiteral ("*"));

    VERIFY_INFO verifyInfo;
    char pvname[PVNAME_SZ + 1];
    char seekBuff[PVNAME_SZ + 1];
    vchar *cellValue;
    vlistIterator iterator;
    int doBoth;
    dataVuApplication *app = mainDialog -> getApplication();
    vuDatabaseClass *db = app -> getDatabaseClass();

    mkptsDialog -> Block ();

    iterator.StartWithSelection (selection);
    while (iterator.Next ())
    {
	doBoth = vFALSE;
	row = iterator.GetRow ();
	cellValue = (vchar *)(list -> GetCellValue (row, 0));
	if ((list -> IsCellSelected (selection, row, 0)) == vTRUE)
	{
	    doBoth = vTRUE;
	}
	vcharExportBounded (vcharScribeSystemExport (cellValue),
			    pvname,
			    PVNAME_SZ);
	pvname[PVNAME_SZ] = 0;
	sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvname);
	DB_GetVerifyInfo (db -> vuPvmapTable, seekBuff, &verifyInfo, NULL);
	if (((list -> IsCellSelected (selection, row, 1)) == vTRUE) || 
	    (doBoth == vTRUE) )
	{
	    cellValue = (vchar *)(list -> GetCellValue (row, 1));
	    if ((vcharSearchChar ((const vchar *)cellValue, '*')) != NULL)
	    {
		list -> SetCellValueScribed (row, 1, vcharScribeFormat (markedStr, asterickStr));
	    }
	    else
	    {
		list -> SetCellValueScribed (row, 1, vcharScribeFormat (markedStr, spaceStr));
	    }
	    verifyInfo.viPVStatus &= ~VI_LOADED_1;
	}
	if (((list -> IsCellSelected (selection, row, 2)) == vTRUE) || 
	    (doBoth == vTRUE) )
	{
	    cellValue = (vchar *)(list -> GetCellValue (row, 2));
	    if ((vcharSearchChar ((const vchar *)cellValue, '*')) != NULL)
	    {
		list -> SetCellValueScribed (row, 2, vcharScribeFormat (markedStr, asterickStr));
	    }
	    else
	    {
		list -> SetCellValueScribed (row, 2, vcharScribeFormat (markedStr, spaceStr));
	    }
	    verifyInfo.viPVStatus &= ~VI_LOADED_2;
	}
	DB_PutVerifyInfo (db -> vuPvmapTable, seekBuff, &verifyInfo, NULL);
    }
    iterator.Finish ();
    list -> UnselectAll (selection);

    mkptsDialog -> Unblock ();

    vstrDestroy (markedStr);
    vstrDestroy (spaceStr);
    vstrDestroy (asterickStr);
}


vkindTYPED_FULL_DEFINITION(vuMkPtsUnmarkButtonClass, vbutton, 
					"vuMkPtsUnmarkButtonClass");

/*******************************************************************
*
*    vuMkPtsUnmarkButtonClass 
*	Observer function
*
*******************************************************************/

void vuMkPtsUnmarkButtonClass::ObserveDialogItem (vevent *event)
{
    vuMkPtsContainerClass *mkptsContainer = mkptsDialog -> getMkPtsContainer ();
    vliststr *list = (vliststr *)(mkptsContainer -> pointList -> GetList ());
    if (list == NULL)
    {
	return;
    }
    vlistSelection *selection = mkptsContainer -> pointList -> GetSelection ();
    int row, column;
    list -> GetSelectedCell (selection, &row, &column);
    if (row == vlistNO_CELLS_SELECTED)
    {
	vuMessageDialogCenter (mkptsDialog, VU_NOTHING_SELECTED_MSG);
	return;
    }
    vuMainDialogClass   *mainDialog = mkptsDialog -> getMainDialog ();
    vresource titleRes = mainDialog -> getTitlesResource ();
    vstr  *unmarkedStr = vresourceGetString (titleRes, vunames_vuMkPtsUnmarkedTitle);
    vstr  *spaceStr    = vstrCloneScribed (vcharScribeLiteral (" "));
    vstr  *asterickStr = vstrCloneScribed (vcharScribeLiteral ("*"));

    VERIFY_INFO verifyInfo;
    char pvname[PVNAME_SZ + 1];
    char seekBuff[PVNAME_SZ + 1];
    vchar *cellValue;
    vlistIterator iterator;
    int doBoth;
   
    dataVuApplication *app = mainDialog->getApplication();
    vuDatabaseClass *db = app -> getDatabaseClass();

    mkptsDialog -> Block ();

    iterator.StartWithSelection (selection);
    while (iterator.Next ())
    {
	doBoth = vFALSE;
	row = iterator.GetRow ();
	cellValue = (vchar *)(list -> GetCellValue (row, 0));
	if ((list -> IsCellSelected (selection, row, 0)) == vTRUE)
	{
	    doBoth = vTRUE;
	}
	vcharExportBounded (vcharScribeSystemExport (cellValue),
			    pvname,
			    PVNAME_SZ);
	pvname[PVNAME_SZ] = 0;
	sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvname);
	DB_GetVerifyInfo (db -> vuPvmapTable, seekBuff, &verifyInfo, NULL);
	if (((list -> IsCellSelected (selection, row, 1)) == vTRUE) || 
	    (doBoth == vTRUE) )
	{
	    cellValue = (vchar *)(list -> GetCellValue (row, 1));
	    if ((vcharSearchChar ((const vchar *)cellValue, '*')) != NULL)
	    {
		list -> SetCellValueScribed (row, 1, vcharScribeFormat (unmarkedStr, asterickStr));
	    }
	    else
	    {
		list -> SetCellValueScribed (row, 1, vcharScribeFormat (unmarkedStr, spaceStr));
	    }
	    verifyInfo.viPVStatus |= VI_LOADED_1;
	}
	if (((list -> IsCellSelected (selection, row, 2)) == vTRUE) || 
	    (doBoth == vTRUE) )
	{
	    cellValue = (vchar *)(list -> GetCellValue (row, 2));
	    if ((vcharSearchChar ((const vchar *)cellValue, '*')) != NULL)
	    {
		list -> SetCellValueScribed (row, 2, vcharScribeFormat (unmarkedStr, asterickStr));
	    }
	    else
	    {
		list -> SetCellValueScribed (row, 2, vcharScribeFormat (unmarkedStr, spaceStr));
	    }
	    verifyInfo.viPVStatus |= VI_LOADED_2;
	}
	DB_PutVerifyInfo (db -> vuPvmapTable, seekBuff, &verifyInfo, NULL);
    }
    iterator.Finish ();
    list -> UnselectAll (selection);

    mkptsDialog -> Unblock ();

    vstrDestroy (unmarkedStr);
    vstrDestroy (spaceStr);
    vstrDestroy (asterickStr);
}


vkindTYPED_FULL_DEFINITION(vuMkPtsDismissButtonClass, vbutton,
					"vuMkPtsDismissButtonClass");

/*******************************************************************
*
*    vuMkPtsDismissButtonClass 
*	Observer function
*
*******************************************************************/

void vuMkPtsDismissButtonClass::ObserveDialogItem (vevent *event)
{

    // close dialog and stop event loop started in vuFindOpen ()
    mkptsDialog -> Close ();
    vevent::StopProcessing ();
}


vkindTYPED_FULL_DEFINITION(vuMkPtsFindButtonClass, vbutton, 
						"vuMkPtsFindButtonClass");

/*******************************************************************
*
*    vuMkPtsFindButtonClass 
*	Observer function
*
*******************************************************************/

void vuMkPtsFindButtonClass::ObserveDialogItem (vevent *event)
{
    vuMkPtsContainerClass *mkptsContainer = mkptsDialog -> getMkPtsContainer ();
    vuMainDialogClass   *mainDialog = mkptsDialog -> getMainDialog ();

    // destroy the old point name list since a new one has 
    // been requested
    vliststr *listdata = (vliststr *)(mkptsContainer -> pointList -> GetList ());
    if (listdata)
    {
	mkptsContainer -> pointList -> SetList (NULL);
	listdata -> Destroy ();
	// make sure that the list in the find class is set to NULL
	// after we destroy it
	mkptsDialog -> setFoundList (NULL);
    }

    // set Point Found label to nothing
    vresource msgRes = mainDialog -> getMessagesResource ();
    mkptsContainer -> pointsFoundLabel -> SetTitleScribed (vcharScribeSystem (" "));

    // ask base class to get a list of points
    mkptsDialog -> vuFindList ();

    // get the list base class found and attach it to our listview
    listdata = mkptsDialog -> getFoundList ();
    int numInList = mkptsDialog -> getNumInList ();
    if (listdata != NULL)
    {
	mkptsDialog -> Block ();

	vresource titleRes  = mainDialog -> getTitlesResource ();
	vstr  *markedStr = vresourceGetString (titleRes, vunames_vuMkPtsMarkedTitle);
	vstr  *unmarkedStr = vresourceGetString (titleRes, vunames_vuMkPtsUnmarkedTitle);
	vstr  *spaceStr = vstrCloneScribed (vcharScribeLiteral (" "));
	vstr  *asterickStr = vstrCloneScribed (vcharScribeLiteral ("*"));
	int   row;
	vchar *listValue;
	char  pvname[PVNAME_SZ + 1];
	char  seekBuff[VU_KEY_SIZE + 1];
	char  resource[RESOURCE_SZ + 1];
	char  iocName1[IOCNAME_SZ + 1];
	char  iocName2[IOCNAME_SZ + 1];
	int   dataType;
	DB_RECORD_PTR rec, resRec;
	VERIFY_INFO   verifyInfo;
	dataVuApplication *app = mainDialog->getApplication();
	vuDatabaseClass *db = app -> getDatabaseClass();
	vlistIterator iterator;

	iterator.Start (listdata);
	while (iterator.Next())
	{
	    row = iterator.GetRow ();
	    listValue = (vchar *)iterator.GetValue ();
	    vcharExportBounded (vcharScribeSystemExport (listValue),
				pvname,
				PVNAME_SZ);
	    pvname[PVNAME_SZ] = 0;
	    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvname);
	    rec = DB_Seek (db -> vuPvmapTable, seekBuff);
	    DB_GetFieldValue (rec, mkptsDialog -> pRes, resource, &dataType);
	    memset ((void *)iocName1, 0, sizeof (iocName1));
	    memset ((void *)iocName2, 0, sizeof (iocName2));
	    sprintf (seekBuff, "%-*.*s\0", RESOURCE_SZ, RESOURCE_SZ, resource);
	    if ((resRec = DB_Seek (db -> vuResListTable, seekBuff)) != 0)
	    {
		DB_GetFieldValue (resRec, mkptsDialog -> pIoc1, iocName1, &dataType);
		DB_GetFieldValue (resRec, mkptsDialog -> pIoc2, iocName2, &dataType);
	    }
	    DB_GetVerifyInfo (db -> vuPvmapTable, pvname, &verifyInfo, NULL);
	    if ((VI_IS_LOADED1 (verifyInfo.viPVStatus)) == 0)
	    {
		if (iocName1[0] == 0)
		{
		listdata -> SetCellValueScribed (row, 1, vcharScribeFormat (markedStr, asterickStr));
		}
		else
		{
		listdata -> SetCellValueScribed (row, 1, vcharScribeFormat (markedStr, spaceStr));
		}
	    }
	    else
	    {
		if (iocName1[0] == 0)
		{
		listdata -> SetCellValueScribed (row, 1, vcharScribeFormat (unmarkedStr, asterickStr));
		}
		else
		{
		listdata -> SetCellValueScribed (row, 1, vcharScribeFormat (unmarkedStr, spaceStr));
		}
	    }
	    if ((VI_IS_LOADED2 (verifyInfo.viPVStatus)) == 0)
	    {
		if (iocName2[0] == 0)
		{
		listdata -> SetCellValueScribed (row, 2, vcharScribeFormat (markedStr, asterickStr));
		}
		else
		{
		listdata -> SetCellValueScribed (row, 2, vcharScribeFormat (markedStr, spaceStr));
		}
	    }
	    else
	    {
		if (iocName2[0] == 0)
		{
		listdata -> SetCellValueScribed (row, 2, vcharScribeFormat (unmarkedStr, asterickStr));
		}
		else
		{
		listdata -> SetCellValueScribed (row, 2, vcharScribeFormat (unmarkedStr, spaceStr));
		}
	    }
	}
	iterator.Finish ();

	vuMkPtsContainerClass *mkptsContainer = mkptsDialog -> getMkPtsContainer ();
	mkptsContainer -> pointList -> SetList (listdata);
	//mkptsContainer->markButton->Disable();
	//mkptsContainer->unmarkButton->Disable();
	int i;
	int itemCount = mkptsContainer -> ruler -> GetItemCount ();;
	vrulerItem *rulerItem;
	int rulerItemWidth;
	for (i = 0; i < itemCount; i++)
	{
	    rulerItem = mkptsContainer -> ruler -> GetItemAt (i);
	    rulerItemWidth = rulerItem -> GetWidth ();
	    mkptsContainer -> pointList -> SetColumnWidth (i, rulerItemWidth);
	}
	vlistSelection *selection = mkptsContainer -> pointList -> GetSelection ();
	listdata -> SetSelectMethod (selection, vlistFREE_SELECTION);

	vstrDestroy (markedStr);
	vstrDestroy (unmarkedStr);
	vstrDestroy (spaceStr);
	vstrDestroy (asterickStr);
	mkptsDialog -> Unblock ();
    }
    vstr *foundMsg = vresourceGetString (msgRes, vunames_vuPtsFoundMsg);
    mkptsContainer -> pointsFoundLabel -> SetTitleScribed (vcharScribeFormat (foundMsg, 
							   numInList));
    vstrDestroy (foundMsg);
}



vkindTYPED_FULL_DEFINITION(vuMkPtsContainerClass, vcontainer, 
						"vuMkPtsContainerClass");

/*******************************************************************
*
*    vuMkPtsContainerClass 
*	override for constructor
*	We get the real resource since what comes in is the resource
*	for the find dialog and we need the resource for the
*	container
*
*******************************************************************/

void vuMkPtsContainerClass::LoadInit (vresource res)
{
    vapplication *app = vapplication::GetCurrent ();
    vresource findDictRes = vresourceGet (app -> GetResources (),
					vunames_vuFindDictionary);
    vresource realRes = vresourceGet (findDictRes, 
					vunames_vuMkPtsContainer);

    pointList = vlistview::RequestSubLoad (this,
					vunames_vuMkPtsPointList);
    pointsFoundLabel = vdialogLabelItem::RequestSubLoad (this,
					vunames_vuMkPtsFoundLabel);

    vcontainer *buttonContainer = vcontainer::RequestSubLoad (this,
				vunames_vuMkPtsButtonContainer);
    
    markButton = vuMkPtsMarkButtonClass::RequestSubLoad (buttonContainer,
				vunames_vuMkPtsMarkButton);
    
    unmarkButton = vuMkPtsUnmarkButtonClass::RequestSubLoad (buttonContainer,
				vunames_vuMkPtsUnmarkButton);

    ruler = vuMkPtsRulerClass::RequestSubLoad (this,
				vunames_vuMkPtsRuler);

    vcontainer::LoadInit (realRes);
}


/*******************************************************************
*
*    vuMkPtsContainerClass 
*	override 
*
*******************************************************************/

void vuMkPtsContainerClass::Destroy (void)
{
    vcontainer::Destroy ();
}

