#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vcharHEADER
#include vnameHEADER
#include vresourceHEADER
#include vdialogHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vlistviewHEADER
#include vliststrHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbDefs.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vuResEditDialogClass.hxx"
#include "vuUtils.hxx"
#include <memory.h>


vkindTYPED_FULL_DEFINITION(vuResEditDialogClass, vdialog, "vuResEditDialogClass");

/*******************************************************************
*
*    vuResEditDialogClass 
*	LoadInit, constructor
*
*******************************************************************/

void vuResEditDialogClass::LoadInit(vresource res)
{
    // resolve the text entry fields
    vcontainer *container = vcontainer::RequestSubLoad (this,
					vunames_vuResEditContainer1);
    resName = filteredTextitem::RequestSubLoad (container,
					vunames_vuResEditResourceName);
    container = vcontainer::RequestSubLoad (this,
					vunames_vuResEditContainer2);
    resIOC1 = filteredTextitem::RequestSubLoad (container,
					vunames_vuResEditIOCName1);
    container = vcontainer::RequestSubLoad (this,
					vunames_vuResEditContainer3);
    resIOC2 = filteredTextitem::RequestSubLoad (container,
					vunames_vuResEditIOCName2);
    // and the list
    resList = vuResListClass::RequestSubLoad(this,
					vunames_vuResEditResourceList);
    resList->vuSetResEditDialog(this);

    // then the toggle button
    dpaToggle = vcontrolToggle::RequestSubLoad(this,
					vunames_vuResEditDPA);
    // finally the buttons
    container = vcontainer::RequestSubLoad(this, 
					vunames_vuResEditContainer);
    okButton = vuResOkButtonClass::RequestSubLoad(container,
					vunames_vuResEditOK);
    dismissButton = vuResDismissButtonClass::RequestSubLoad(container,
					vunames_vuResEditDismiss);
    delButton = vuResDelButtonClass::RequestSubLoad(container,
					vunames_vuResEditDelete);
    okButton -> vuSetResEditDialog(this);
    dismissButton -> vuSetResEditDialog(this);
    delButton -> vuSetResEditDialog(this);

    // call base class loadInit to finish the job
    vdialog::LoadInit(res);
    //
    // handle readonly state for the apply and delete buttons
    //
    if(vuReadOnly == vTRUE)
    {
	okButton->Disable();
	delButton->Disable();
    }
}

/*******************************************************************
*
*    vuResEditDialogClass open function
*
*******************************************************************/

void vuResEditDialogClass::vuResEditOpen(vuMainDialogClass *mDialog)
{
    mainDialog = mDialog;
    dataVuApplication *app = mainDialog -> getApplication ();
    db = app -> getDatabaseClass ();

    // RESOURCE LIST
    // destroy any existing listdata
    vliststr *resListData = (vliststr *)(resList -> GetList());	
    if(resListData != NULL)
    {
	resListData -> Destroy ();
    }
    // then create a new resource list with DPA's
    resListData = vuGetResList (db, VU_ALL_RESOURCES);
  
    //  tell list view what to look at
    resList -> SetList (resListData);

    // set so only one resource name can be selected
    vlistSelection *selection = resList -> GetSelection ();
    resListData -> SetSelectMethod (selection, vlistONE_ONLY_SELECTION);

    pRes   = DB_MapField (db -> vuResListTable, FLD_RESOURCE);
    pIOC1  = DB_MapField (db -> vuResListTable, FLD_IOC1);
    pIOC2  = DB_MapField (db -> vuResListTable, FLD_IOC2);
    pDPA   = DB_MapField (db -> vuResListTable, FLD_DPA);

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process();
}


/*******************************************************************
*
*    vuResEditDialogClass notify function
*
*******************************************************************/

void vuResEditDialogClass::ObserveWindow(vevent *event)
{
    if ( event -> GetType() == veventWINDOW_STATE && event -> IsClose())
    {
	Close();
	vevent::StopProcessing();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuResEditDialogClass Destroy function
*	if dialog is open, close it and stop its event loop
*
*******************************************************************/

void vuResEditDialogClass::Destroy(void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }
    vliststr *listdata;
    listdata = (vliststr *)resList -> GetList();
    if(listdata != NULL) 
    {
	listdata->Destroy();
    }
    vdialog::Destroy();
}

vkindTYPED_FULL_DEFINITION(vuResOkButtonClass, vbutton, "vuResOkButtonClass");
/*******************************************************************
*
*    vuResOkButtonClass 
*	Observer function
*
*******************************************************************/
void vuResOkButtonClass::ObserveDialogItem (vevent *event)
{
    char resname[RESOURCE_SZ + 1], ioc1name[IOCNAME_SZ + 1],ioc2name[IOCNAME_SZ + 1];
    DB_RECORD_PTR pntRecord;

    vstr *str1 = reseditDialog -> resName -> GetTextAsString ();
    vcharExportBounded(vcharScribeSystemExport(str1), resname, RESOURCE_SZ);
    resname [RESOURCE_SZ] = 0;

    if(strlen (resname))
    {
	vstr *str2 = reseditDialog -> resIOC1 -> GetTextAsString();
	vcharExportBounded (vcharScribeSystemExport (str2), ioc1name, IOCNAME_SZ);
	ioc1name [IOCNAME_SZ] = 0;
	vstrDestroy(str2);

	str2 = reseditDialog -> resIOC2-> GetTextAsString();
	vcharExportBounded (vcharScribeSystemExport (str2), ioc2name, IOCNAME_SZ);
	ioc2name [IOCNAME_SZ] = 0;
	vstrDestroy(str2);

	// Seek to record for this resource name. If found then just write
	// out new data. If not then create new record and then fill it in
	// also add to the list.
	char seekBuff [RESOURCE_SZ + RESOURCE_SZ];
	int result;
	sprintf (seekBuff, "%-*.*s\0", RESOURCE_SZ, RESOURCE_SZ, resname);
	if((pntRecord = DB_Seek(reseditDialog -> db -> vuResListTable, seekBuff)) == 0)
	{
	    result = vuMessageDialogCenter (reseditDialog, VU_RESEDIT_NEW_REQ, str1);
	    if (result == VU_FALSE)
	    {
		vstrDestroy (str1);
		return;
	    }
	    pntRecord = DB_NewRecord (reseditDialog -> db -> vuResListTable, 0);
	    DB_PutFieldValue (pntRecord, reseditDialog -> pRes, resname);
	    vliststr *resListData = (vliststr *)(reseditDialog -> resList -> GetList());
	    resListData -> SetCellValueScribed(resListData -> GetRowCount(),
					       0,
					       vcharScribeSystem(resname));
	    reseditDialog -> mainDialog -> vuMainMessageCenter (vunames_ReseditNewMsg, str1);
	    reseditDialog->db->ResListHasChanged(vTRUE);
	    
	}
	else
	{
	    result = vuMessageDialogCenter (reseditDialog, VU_RESEDIT_UPDATE_REQ, str1);
	    if (result == VU_FALSE)
	    {
		vstrDestroy (str1);
		return;
	    }
	    reseditDialog -> mainDialog -> vuMainMessageCenter (vunames_ReseditUpdateMsg, str1);
	}
	vstrDestroy (str1);
	DB_PutFieldValue (pntRecord, reseditDialog -> pIOC1, ioc1name);
	DB_PutFieldValue (pntRecord, reseditDialog -> pIOC2, ioc2name);
	char state = (char)(reseditDialog -> dpaToggle -> GetValue() & 1);
	DB_PutFieldValue (pntRecord, reseditDialog -> pDPA, &state);
	vuMessageDialogCenter (reseditDialog, VU_CA_DBC_RESTART_MSG);
    }
    else
    {
	vuMessageDialogCenter (reseditDialog, VU_RESEDIT_RESNAME_ERR);
	vstrDestroy (str1);
	return;
    }
}

vkindTYPED_FULL_DEFINITION(vuResDismissButtonClass, vbutton, "vuResDismissButtonClass");
/*******************************************************************
*
*    vuResDismissButtonClass 
*	Observer function
*
*******************************************************************/
void vuResDismissButtonClass::ObserveDialogItem (vevent *event)
{
    // close dialog and stop event loop started in vuResEditOpen ()
    reseditDialog -> Close();
    vevent::StopProcessing();
}

vkindTYPED_FULL_DEFINITION(vuResDelButtonClass, vbutton, "vuResDelButtonClass");
/*******************************************************************
*
*    vuResDelButtonClass 
*	Observer function
*
*******************************************************************/
void vuResDelButtonClass::ObserveDialogItem (vevent *event)
{
    char resname[RESOURCE_SZ + 1];
    DB_RECORD_PTR pntRecord;

    vstr *str1 = reseditDialog -> resName -> GetTextAsString();
    vcharExportBounded(vcharScribeSystemExport(str1), resname, RESOURCE_SZ);
    resname [RESOURCE_SZ] = 0;

    if(strlen(resname))
    {
	// if record is in resource table
	if((pntRecord = DB_Seek(reseditDialog -> db -> vuResListTable, resname)) == 
								(DB_RECORD_PTR)1)
	{
	    int result = vuMessageDialogCenter (reseditDialog, VU_RESEDIT_DEL_REQ, str1);
	    if (result == VU_FALSE)
	    {
		vstrDestroy (str1);
		return;
	    }
	    DB_DeleteRecord(reseditDialog -> db -> vuResListTable, pntRecord);
	    reseditDialog -> mainDialog -> vuMainMessageCenter (vunames_ReseditDelMsg, str1);

	    // delete from resource list
	    vliststr *list = (vliststr *)(reseditDialog -> resList -> GetList ());
	    vlistIterator iterator;
	    vchar *value;
	    int row;
	    iterator.Start (list);
	    while (iterator.Next ())
	    {
		row = iterator.GetRow ();
		value = (vchar *)(iterator.GetValue ());
		if ((vcharCompare (value, str1)) == 0)
		{
		    list -> RemoveRow (row);
		    if ( (list -> GetRowCount ()) == 0)
		    {
			reseditDialog -> resName -> SetTextScribed (vcharScribeSystem (""));
			reseditDialog -> resIOC1 -> SetTextScribed (vcharScribeSystem (""));
			reseditDialog -> resIOC2 -> SetTextScribed (vcharScribeSystem (""));
			reseditDialog -> dpaToggle -> SetValue (0);
		    }
		    else
		    {
			vlistSelection *selected = reseditDialog -> resList -> GetSelection ();
			list -> SelectRow (selected, 0);
		    }
		    break;
		}
	    }
	    iterator.Finish ();
	    vstrDestroy (str1);
	    reseditDialog->db->ResListHasChanged(vTRUE);
	}
	else
	{
	    vuMessageDialogCenter (reseditDialog, VU_RESEDIT_DEL_ERR, str1);
	    vstrDestroy (str1);
	    return;
	}
    }
    else
    {
	vuMessageDialogCenter (reseditDialog, VU_RESEDIT_RESNAME_ERR);
	vstrDestroy (str1);
	return;
    }
}


vkindTYPED_FULL_DEFINITION(vuResListClass, vlistview, "vuResListClass");
/*******************************************************************
*
*    vuResListClass selection notify function
*	SELECTION_HAS_CHANGED = user finally released the mouse
*	button to establish the new selection.
*
*******************************************************************/

void vuResListClass::ObserveSelect(int state)
{
    DB_RECORD_PTR pntRecord;
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	int	row,column;
	char	ioc1[IOCNAME_SZ + 1],ioc2[IOCNAME_SZ + 1],dpa;
	char	resName [RESOURCE_SZ + 1];
	int	ftype;
	const vchar *resourceName;

	vliststr *ptListData = (vliststr *)(GetList ());
	vlistSelection *listSelection = GetSelection ();
	ptListData -> GetSelectedCell (listSelection, &row, &column);
	if (row == vlistNO_CELLS_SELECTED)
	{
	    return;
	}
	resourceName = (const vchar *)ptListData -> GetCellValue (row, 0);
	reseditDialog -> resName -> SetText(resourceName);

	// Find record in reslist.dbf and fill in other textfields

	vcharExportBounded(vcharScribeSystemExport(resourceName), resName, RESOURCE_SZ);
	resName [RESOURCE_SZ] = 0;
	char seekbuff [RESOURCE_SZ + RESOURCE_SZ];
	sprintf (seekbuff, "%-*.*s\0", RESOURCE_SZ, RESOURCE_SZ, resName);
	pntRecord = DB_Seek (reseditDialog -> db-> vuResListTable, seekbuff);
	DB_GetFieldValue(pntRecord, reseditDialog -> pIOC1, ioc1, &ftype);
	DB_GetFieldValue(pntRecord, reseditDialog -> pIOC2, ioc2, &ftype);
	DB_GetFieldValue(pntRecord, reseditDialog -> pDPA, &dpa, &ftype);

	reseditDialog -> resIOC1 -> SetTextScribed (vcharScribeSystem (ioc1));
	reseditDialog -> resIOC2 -> SetTextScribed (vcharScribeSystem (ioc2));
	reseditDialog -> dpaToggle -> SetValue ((int)(dpa & 1));
    }
}
