#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vlistviewHEADER
#include vliststrHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vnumHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbDefs.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vuScriptScheduleDialogClass.hxx"
#include "vuUtils.hxx"

// filter functions unique to script schedular
static int vuScrSchIdFilter (filteredText *text, vtextSelection *sel,
			     const vchar **string, long *len);
static int vuScrSchNameFilter (filteredText *text, vtextSelection *sel,
			       const vchar **string, long *len);
static int vuScrSchDescFilter (filteredText *text, vtextSelection *sel,
			       const vchar **string, long *len);
static int vuScrSchParamFilter (filteredText *text, vtextSelection *sel,
			        const vchar **string, long *len);

// some arrays to be used by vuMapFields

static char	*dayNames [] = 
	{
	"SUN", "MON", "TU", "WED", "TH", "FRI", "SAT", 0
	};

static char	*monthNames [] = 
	{
	"JAN", "FEB", "MAR", "APR", "MAY", "JUN", "JUL", "AUG", "SEP", "OCT", "NOV", "DEC", 0
	};

// defines for the DAY OF MONTH database entries
#define	VU_FIRSTWDOM	"FIRSTWDOM"
#define	VU_LASTWDOM	"LASTWDOM"
#define	VU_LASTDOM	"LASTDOM"


/*******************************************************************
*
*	SCRIPT SCHEDULE DIALOG CLASS
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuScrSchDialogClass, vdialog, 
						"vuScrSchDialogClass");

/*******************************************************************
*
*    vuScrSchDialogClass 
*	LoadInit, constructor
*
*******************************************************************/

void vuScrSchDialogClass::LoadInit(vresource res)
{
    dialogModified = vFALSE;

    selScrDialog = NULL;

    scriptID = vuScrSchIdFilteredTextitem::RequestSubLoad (this,
					vunames_vuScrSchID);

    vcontainer *shcontainer = vcontainer::RequestSubLoad (this,
					vunames_vuScrSchShellContainer);
    shell = optionMenuClass::RequestSubLoad (shcontainer,
					vunames_vuScrSchShell);

    vcontainer *nmcontainer = vcontainer::RequestSubLoad (this,
					vunames_vuScrSchNameContainer);
    name = filteredTextitem::RequestSubLoad (nmcontainer,
					vunames_vuScrSchName);

    vcontainer *dccontainer = vcontainer::RequestSubLoad (this,
					vunames_vuScrSchDescContainer);
    description = filteredTextitem::RequestSubLoad (dccontainer,
					vunames_vuScrSchDesc);

    useParam[0] = checkBoxClass::RequestSubLoad (this, vunames_vuScrSchUseP1);
    useParam[1] = checkBoxClass::RequestSubLoad (this, vunames_vuScrSchUseP2);
    useParam[2] = checkBoxClass::RequestSubLoad (this, vunames_vuScrSchUseP3);
    useParam[3] = checkBoxClass::RequestSubLoad (this, vunames_vuScrSchUseP4);
    useParam[4] = checkBoxClass::RequestSubLoad (this, vunames_vuScrSchUseP5);
    useParam[5] = checkBoxClass::RequestSubLoad (this, vunames_vuScrSchUseP6);
    useParam[6] = checkBoxClass::RequestSubLoad (this, vunames_vuScrSchUseP7);
    useParam[7] = checkBoxClass::RequestSubLoad (this, vunames_vuScrSchUseP8);
    useParam[8] = checkBoxClass::RequestSubLoad (this, vunames_vuScrSchUseP9);
    useParam[9] = checkBoxClass::RequestSubLoad (this, vunames_vuScrSchUseP10);

    vcontainer *p1container = vcontainer::RequestSubLoad (this,
					vunames_vuScrSchP1Container);
    params[0] = filteredTextitem::RequestSubLoad (p1container,
					vunames_vuScrSchP1);

    vcontainer *p2container = vcontainer::RequestSubLoad (this,
					vunames_vuScrSchP2Container);
    params[1] = filteredTextitem::RequestSubLoad (p2container,
					vunames_vuScrSchP2);

    vcontainer *p3container = vcontainer::RequestSubLoad (this,
					vunames_vuScrSchP3Container);
    params[2] = filteredTextitem::RequestSubLoad (p3container,
					vunames_vuScrSchP3);

    vcontainer *p4container = vcontainer::RequestSubLoad (this,
					vunames_vuScrSchP4Container);
    params[3] = filteredTextitem::RequestSubLoad (p4container,
					vunames_vuScrSchP4);

    vcontainer *p5container = vcontainer::RequestSubLoad (this,
					vunames_vuScrSchP5Container);
    params[4] = filteredTextitem::RequestSubLoad (p5container,
					vunames_vuScrSchP5);

    vcontainer *p6container = vcontainer::RequestSubLoad (this,
					vunames_vuScrSchP6Container);
    params[5] = filteredTextitem::RequestSubLoad (p6container,
					vunames_vuScrSchP6);

    vcontainer *p7container = vcontainer::RequestSubLoad (this,
					vunames_vuScrSchP7Container);
    params[6] = filteredTextitem::RequestSubLoad (p7container,
					vunames_vuScrSchP7);

    vcontainer *p8container = vcontainer::RequestSubLoad (this,
					vunames_vuScrSchP8Container);
    params[7] = filteredTextitem::RequestSubLoad (p8container,
					vunames_vuScrSchP8);

    vcontainer *p9container = vcontainer::RequestSubLoad (this,
					vunames_vuScrSchP9Container);
    params[8] = filteredTextitem::RequestSubLoad (p9container,
					vunames_vuScrSchP9);

    vcontainer *p10container = vcontainer::RequestSubLoad (this,
					vunames_vuScrSchP10Container);
    params[9] = filteredTextitem::RequestSubLoad (p10container,
					vunames_vuScrSchP10);

    vcontainer *tmcontainer = vcontainer::RequestSubLoad (this,
					vunames_vuScrSchTimeContainer);
    timeValue = hhmmTimeClass::RequestSubLoad (tmcontainer,
					vunames_vuScrSchTime);

    dayOfWeek = groupContainerClass::RequestSubLoad (this,
					vunames_vuScrSchDOWContainer);

    dayOfMonth = vuScrSchDomClass::RequestSubLoad (this,
					vunames_vuScrSchDOMContainer);

    dayOfMonthValue = vuScrSchDomValueClass::RequestSubLoad (this,
					vunames_vuScrSchDOMValue);

    monthOfYear = groupContainerClass::RequestSubLoad (this,
					vunames_vuScrSchMOYContainer);

    vcontainer *btcontainer = vcontainer::RequestSubLoad (this,
					vunames_vuScrSchButtonContainer);

    applyButton = vuScrSchApplyButtonClass::RequestSubLoad (btcontainer,
					vunames_vuScrSchApplyButton);

    dismissButton = vuScrSchDismissButtonClass::RequestSubLoad (btcontainer,
					vunames_vuScrSchDismissButton);

    deleteButton = vuScrSchDeleteButtonClass::RequestSubLoad (btcontainer,
					vunames_vuScrSchDeleteButton);

    newButton = vuScrSchNewButtonClass::RequestSubLoad (btcontainer,
					vunames_vuScrSchNewButton);

    gotoButton = vuScrSchGoToButtonClass::RequestSubLoad (btcontainer,
					vunames_vuScrSchGoToButton);

    nextButton = vuScrSchNextButtonClass::RequestSubLoad (btcontainer,
					vunames_vuScrSchNextButton);

    prevButton = vuScrSchPrevButtonClass::RequestSubLoad (btcontainer,
					vunames_vuScrSchPrevButton);

    // call base class loadInit to finish the job
    vdialog::LoadInit(res);

    // store ptr to dialog in buttons
    applyButton -> storeScrSchDialog (this);
    dismissButton -> storeScrSchDialog (this);
    deleteButton -> storeScrSchDialog (this);
    newButton -> storeScrSchDialog (this);
    gotoButton -> storeScrSchDialog (this);
    nextButton -> storeScrSchDialog (this);
    prevButton -> storeScrSchDialog (this);
    //
    // handle readonly state for the apply and delete buttons
    //
    if(vuReadOnly == vTRUE)
    {
	applyButton->Disable();
	deleteButton->Disable();
    }

    // store ptr to dialog
    scriptID -> storeScrSchDialog (this);
    dayOfMonth -> SetUserPtr ((void *)this);
    dayOfMonthValue -> SetUserPtr ((void *)this);

    // set filters
    scriptID -> SetItemFilterFunc (vuScrSchIdFilter);
    filteredText *idtext = (filteredText *)scriptID -> GetTextData ();
    // remember ptr to textitem for later use
    idtext -> SetUserPtr ((void *)scriptID);

    name -> SetItemFilterFunc (vuScrSchNameFilter);
    description -> SetItemFilterFunc (vuScrSchDescFilter);

    int i;

    for (i = 0; i < 10; i++)
    {
	params[i] -> SetItemFilterFunc (vuScrSchParamFilter);
    }

    timeValue -> SetItemFilterFunc (hhmmTimeFilter);
    filteredText *tvtext = (filteredText *)timeValue -> GetTextData ();
    // remember ptr to textitem for later use
    tvtext -> SetUserPtr ((void *)timeValue);

    // store dialogModified ptr
    scriptID -> SetTextModifiedPtr (&this -> dialogModified);
    shell -> SetModifiedPtr (&this -> dialogModified);
    name -> SetTextModifiedPtr (&this -> dialogModified);
    description -> SetTextModifiedPtr (&this -> dialogModified);
    for (i = 0; i < 10; i++)
    {
	useParam[i] -> SetModifiedPtr (&this -> dialogModified);
    }
    for (i = 0; i < 10; i++)
    {
	params[i] -> SetTextModifiedPtr (&this -> dialogModified);
    }

    timeValue -> SetTextModifiedPtr (&this -> dialogModified);
    dayOfWeek -> SetModifiedPtr (&this -> dialogModified);
    dayOfMonth -> SetModifiedPtr (&this -> dialogModified);
    dayOfMonthValue -> SetModifiedPtr (&this -> dialogModified);
    monthOfYear -> SetModifiedPtr (&this -> dialogModified);

    // place window in center of screen
    Place(GetRoot(),vrectPLACE_CENTER,vrectPLACE_CENTER);
}


/*******************************************************************
*
*    vuScrSchDialogClass notify function
*
*******************************************************************/

void vuScrSchDialogClass::ObserveWindow(vevent *event)
{
    if ( event -> GetType() == veventWINDOW_STATE && event -> IsClose())
    {
	int result = VU_TRUE;
	if (dialogModified == vTRUE)
	{
	    result = vuMessageDialogCenter (this, VU_MOD_DISMISS_MSG);
	}
	if (result == VU_TRUE)
	{
	    Close();
	    vevent::StopProcessing ();
	}
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuScrSchDialogClass Destroy function
*	if dialog is open, close it and stop its event loop
*
*******************************************************************/

void vuScrSchDialogClass::Destroy(void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }

    if (selScrDialog != NULL)
    {
	delete selScrDialog;
    }

    vdialog::Destroy();
}


/*******************************************************************
*
*    vuScrSchDialogClass map database fields function
*
*******************************************************************/

void vuScrSchDialogClass::vuMapFields (void)
{
    dataVuApplication *vuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = vuApp -> getDatabaseClass ();

    scIdField = DB_MapField (db -> vuScriptTable, "SCRIPTID");
    scNameField = DB_MapField (db -> vuScriptTable, "PROCESS");
    scDescField = DB_MapField (db -> vuScriptTable, "DESC");
    scShellField = DB_MapField (db -> vuScriptTable, "SHELL");
    scHourField = DB_MapField (db -> vuScriptTable, "HOUR");
    scMinField = DB_MapField (db -> vuScriptTable, "MINUTE");
    scDOMField = DB_MapField (db -> vuScriptTable, "DOM");
    scUseField = DB_MapField (db -> vuScriptTable, "USE");

    char paramName[8];
    int  i;

    for (i = 0; i < 10; i++)
    {
	sprintf (paramName, "%P%d\0", i);
	scParamsField[i] = DB_MapField (db -> vuScriptTable, paramName);
    }
    for (i = 0; i < 7; i++)
    {
	scDOWField[i] = DB_MapField (db -> vuScriptTable, dayNames[i]);
    }
    for (i = 0; i < 12; i++)
    {
	scMOYField[i] = DB_MapField (db -> vuScriptTable, monthNames[i]);
    }
}


/*******************************************************************
*
*    vuScrSchDialogClass Open function
*
*******************************************************************/

void vuScrSchDialogClass::vuScrSchOpen(vuMainDialogClass *mainDialg)
{
    mainDialog = mainDialg;

    vuMapFields ();

    dayOfMonthValue -> buildMenu ();

    vuNewScript ();

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    vuScrSchDialogClass new script  function
*
*******************************************************************/

void vuScrSchDialogClass::vuNewScript(void)
{
    scriptID -> SetTextScribed (vcharScribeSystem (""));
    shell -> SetValue (0);
    name -> SetTextScribed (vcharScribeSystem (""));
    description -> SetTextScribed (vcharScribeSystem (""));
    int i;
    for (i = 0; i < 10; i++)
    {
	useParam[i] -> SetValue (0);
    }
    for (i = 0; i < 10; i++)
    {
	params[i] -> SetTextScribed (vcharScribeSystem (""));
    }
    timeValue -> SetTextScribed (vcharScribeLiteral ("00:00"));

    vcontrolToggle	*toggle;

    for (i = 0; i < dayOfWeek -> GetItemCount (); i++)
    {
	toggle = (vcontrolToggle *)dayOfWeek -> GetItemAt (i);
	toggle -> SetValue (0);
    }

    const vchar	*toggleTitle;

    for (i = 0; i < dayOfMonth -> GetItemCount (); i++)
    {
	toggle = (vcontrolToggle *)dayOfMonth -> GetItemAt (i);
	toggleTitle = toggle -> GetTitle ();
	if (toggleTitle[0] == 'S')
	{
	    dayOfMonth -> SetSelection (toggle);
	    break;
	}
    }

    dayOfMonthValue -> Enable ();
    dayOfMonthValue -> SetValue (0);

    for (i = 0; i < monthOfYear -> GetItemCount (); i++)
    {
	toggle = (vcontrolToggle *)monthOfYear -> GetItemAt (i);
	toggle -> SetValue (1);
    }

    dialogModified = vFALSE;
}


/*******************************************************************
*
*    vuScrSchDialogClass load script function
*
*******************************************************************/

void vuScrSchDialogClass::vuLoadScript (vchar *scriptId)
{
    dataVuApplication *vuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = vuApp -> getDatabaseClass ();
    DB_RECORD_PTR rec;
    char script[SCRIPTID_SZ + 1];

    vcharExportBounded (vcharScribeSystemExport (scriptId),
			script,
			SCRIPTID_SZ);
    script[SCRIPTID_SZ] = 0;

    if ( (rec = DB_Seek (db -> vuScriptTable, script)) != 0)
    {
	vcontrolToggle	*toggle;
	char		text[80];
	char		trueFalse;
	vstr		*vStr1, *vStr3;
	int		type, i;
	double		hour, minute;

	// id field
	scriptID -> SetTextScribed (vcharScribeSystem (script));

	// name (process) field
	DB_GetFieldValue (rec, scNameField, text, &type);
	name -> SetTextScribed (vcharScribeSystem (text));

	// description field
	DB_GetFieldValue (rec, scDescField, text, &type);
	description -> SetTextScribed (vcharScribeSystem (text));

	// shell value
	DB_GetFieldValue (rec, scShellField, text, &type);
	if ((strlen (text)) == 0)
	    strcpy (text, "exec");
	vStr1 = vstrCloneScribed (vcharScribeSystem (text));
	shell -> SetOptionMenuValue (vStr1);
	vstrDestroy (vStr1);

	// get use parameter toggles
	// 1st turn them all off
	for (i = 0; i < 10; i++)
	{
	    useParam[i] -> SetValue (0);
	}
	DB_GetFieldValue (rec, scUseField, text, &type);
	i = 0;
	while (text[i])
	{
	    useParam[text[i] - '0'] -> SetValue (1);
	    i++;
	}

	// get parameter fields
	for (i = 0; i < 10; i++)
	{
	    DB_GetFieldValue (rec, scParamsField[i], text, &type);
	    params[i] -> SetTextScribed (vcharScribeSystem (text));
	}

	// get time of day
	DB_GetFieldValue (rec, scHourField, &hour, &type);
	DB_GetFieldValue (rec, scMinField, &minute, &type);
	vStr3 = vstrCloneScribed (vcharScribeFormatLiteral ("%02d:%02d",
				  (int)hour, (int)minute));
	timeValue -> SetText (vStr3);
	vstrDestroy (vStr3);

	// get day of week
	for (i = 0; i < 7; i++)
	{
	    DB_GetFieldValue (rec, scDOWField[i], &trueFalse, &type);
	    toggle = (vcontrolToggle *)dayOfWeek -> GetItemAt (i);
	    toggle -> SetValue ((int)trueFalse);
	}

	// get day of month choice
	DB_GetFieldValue (rec, scDOMField, text, &type);
	dayOfMonthValue -> Disable ();
	if ( (strcmp (text, VU_LASTDOM)) == 0)
	{
	    toggle = (vcontrolToggle *)dayOfMonth -> GetItemAt (3);
	    dayOfMonth -> SetSelection (toggle);
	}
	else if ( (strcmp (text, VU_LASTWDOM)) == 0)
	{
	    toggle = (vcontrolToggle *)dayOfMonth -> GetItemAt (2);
	    dayOfMonth -> SetSelection (toggle);
	}
	else if ( (strcmp (text, VU_FIRSTWDOM)) == 0)
	{
	    toggle = (vcontrolToggle *)dayOfMonth -> GetItemAt (1);
	    dayOfMonth -> SetSelection (toggle);
	}
	else
	{
	    toggle = (vcontrolToggle *)dayOfMonth -> GetItemAt (0);
	    dayOfMonth -> SetSelection (toggle);
	    vStr1 = vstrCloneScribed (vcharScribeSystem (text));
	    if ( (vStr1[0] == 0) || (vStr1[0] == ' ') )
		i = 0;
	    else
		i = vnum::ParseInt (vStr1, NULL);
	    if ( (i > 0) && (i < 32) )
	    {
		dayOfMonthValue -> SetValue (i - 1);
	    }
	    else
	    {
		dayOfMonthValue -> SetValue (0);
	    }
	    vstrDestroy (vStr1);
	    dayOfMonthValue -> Enable ();
	}

	// get months of year
	for (i = 0; i < 12; i++)
	{
	    DB_GetFieldValue (rec, scMOYField[i], &trueFalse, &type);
	    toggle = (vcontrolToggle *)monthOfYear -> GetItemAt (i);
	    toggle -> SetValue ((int)trueFalse);
	}
    }

    dialogModified = vFALSE;
}


/*******************************************************************
*
*    vuScrSchApplyButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuScrSchApplyButtonClass, vbutton, 
					"vuScrSchApplyButtonClass");

void vuScrSchApplyButtonClass::ObserveDialogItem (vevent *event)
{
    vchar scriptID[SCRIPTID_SZ + 2];
    scrSchDialog -> scriptID -> GetText (scriptID, SCRIPTID_SZ + 1);
    if ( (vcharLength (scriptID)) == 0)
    {
	vuMessageDialogCenter (scrSchDialog, VU_SCRIPT_NO_ID_MSG);
	return;
    }
    char script[SCRIPTID_SZ + 1];
    vcharExportBounded (vcharScribeSystemExport (scriptID),
			script,
			SCRIPTID_SZ);
    script[SCRIPTID_SZ] = 0;

    double hour = 0;
    double minute = 0;
    vchar timeVal[HHMMTIME_SZ + 2];
    scrSchDialog -> timeValue -> GetText (timeVal, HHMMTIME_SZ + 1);
    if (timeVal[0] != 0)
    {
	vchar *hr = vcharGetFirstToken (timeVal, ":");
	vchar *min = vcharGetNextToken ("");
	if (hr[0] != 0)
	{
	    hour = vnum::ParseFloat (hr, NULL);
	}
	if (min[0] != 0)
	{
	    minute = vnum::ParseFloat (min, NULL);
	}
	if ( (hour < 0) || (hour > 23) )
	{
	    vuMessageDialogCenter (scrSchDialog, VU_HOUR_ERR_MSG);
	    return;
	}
	if ( (minute < 0) || (minute > 59) )
	{
	    vuMessageDialogCenter (scrSchDialog, VU_MINUTE_ERR_MSG);
	    return;
	}
    }

    vuMainDialogClass *mainDialog = scrSchDialog -> getMainDialog ();
    dataVuApplication *vuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = vuApp -> getDatabaseClass ();
    DB_RECORD_PTR rec;

    int result;
    if ( (rec = DB_Seek (db -> vuScriptTable, script)) == 0)
    {
	result = vuMessageDialogCenter (scrSchDialog, VU_SCRIPT_NEW_REQ_MSG, scriptID);
	if (result == VU_FALSE)
	{
	    return;
	}
	rec = (DB_RECORD_PTR)DB_NewRecord (db -> vuScriptTable, NULL);
	mainDialog -> vuMainMessageCenter (vunames_vuScriptCreateMsg, scriptID);
	vuSelScrDialogClass *selScrDialog = scrSchDialog -> getSelScrDialog ();
	// add to selection list if it exists
	if (selScrDialog != NULL)
	{
	    vliststr *list = (vliststr *)(selScrDialog -> selScrList -> GetList ());
	    int rowCount = list -> GetRowCount ();
	    list -> SetCellValue (rowCount, 0, (const void *)scriptID);
	    list -> Sort (vlistROW, 0, NULL);
	}
    }
    else
    {
	result = vuMessageDialogCenter (scrSchDialog, VU_SCRIPT_MOD_REQ_MSG, scriptID);
	if (result == VU_FALSE)
	{
	    return;
	}
	mainDialog -> vuMainMessageCenter (vunames_vuScriptModMsg, scriptID);
    }

    vstr *dialogText;
    const vchar *optionMenuValue;
    char text[128];

    // save script id to database
    DB_PutFieldValue (rec, scrSchDialog -> scIdField, script);

    // save shell name 
    optionMenuValue = scrSchDialog -> shell -> GetOptionMenuValue ();
    if ( (vcharCompare (optionMenuValue, "exec")) == 0)
    {
	strcpy (text, "   ");
    }
    else
    {
	vcharExportBounded (vcharScribeSystemExport (optionMenuValue),
			    text,
			    3);
	text[3] = 0;
    }
    DB_PutFieldValue (rec, scrSchDialog -> scShellField, text);

    // save script name
    dialogText = scrSchDialog -> name -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (dialogText),
			text,
			SCRIPTPROC_SZ);
    text[SCRIPTPROC_SZ] = 0;
    DB_PutFieldValue (rec, scrSchDialog -> scNameField, text);
    vstrDestroy (dialogText);

    // save script description
    dialogText = scrSchDialog -> description -> GetTextAsString ();
    vcharExportBounded (vcharScribeSystemExport (dialogText),
			text,
			SCRIPTDESC_SZ);
    text[SCRIPTDESC_SZ] = 0;
    DB_PutFieldValue (rec, scrSchDialog -> scDescField, text);
    vstrDestroy (dialogText);

    // write parameter fields and the use field
    int  i, j;
    char use[11];
    int  toggleValue;
    char logical;
    vcontrolToggle *toggle;

    for (i = 0, j = 0; i < 10; i++)
    {
	dialogText = scrSchDialog -> params[i] -> GetTextAsString ();
	vcharExportBounded (vcharScribeSystemExport (dialogText),
			    text,
			    SCRIPTPARAM_SZ);
	text[SCRIPTPARAM_SZ] = 0;
	DB_PutFieldValue (rec, scrSchDialog -> scParamsField[i], text);
	vstrDestroy (dialogText);

	// get the use toggle
	toggleValue = scrSchDialog -> useParam[i] -> GetValue ();
	if (toggleValue)
	{
	    use[j] = i + '0';
	    j++;
	}
    }
    use[j] = 0;
    DB_PutFieldValue (rec, scrSchDialog -> scUseField, use);

    // write the hour and minutes fields
    DB_PutFieldValue (rec, scrSchDialog -> scHourField, &hour);
    DB_PutFieldValue (rec, scrSchDialog -> scMinField, &minute);

    // write the days of the week
    for (i = 0; i < 7; i++)
    {
	toggle = (vcontrolToggle *)scrSchDialog -> dayOfWeek -> GetItemAt (i);
	logical = (char)toggle -> GetValue ();
	DB_PutFieldValue (rec, scrSchDialog -> scDOWField[i], &logical);
    }

    // write out day of month
    strcpy (text, "         ");	// in case nothing selected
    vdialogItem *dom = scrSchDialog -> dayOfMonth -> GetSelection ();
    if (dom != NULL)
    {
	i = scrSchDialog -> dayOfMonth -> GetItemIndex (dom);
	switch (i)
	{
	    case 0:	// selected day of month
		optionMenuValue = scrSchDialog -> dayOfMonthValue -> GetOptionMenuValue ();
		vcharExportBounded (vcharScribeSystemExport (optionMenuValue),
				    text,
				    10);
		text[10] = 0;
		break;

	    case 1:	// first day of week in month
		strcpy (text, VU_FIRSTWDOM);
		break;

	    case 2:	// last day of week in month
		strcpy (text, VU_LASTWDOM);
		break;

	    case 3:	// last day of month
		strcpy (text, VU_LASTDOM);
		break;
	}
    }
    DB_PutFieldValue (rec, scrSchDialog -> scDOMField, text);

    // write months of year
    for (i = 0; i < 12; i++)
    {
	toggle = (vcontrolToggle *)scrSchDialog -> monthOfYear -> GetItemAt (i);
	logical = (char)toggle -> GetValue ();
	DB_PutFieldValue (rec, scrSchDialog -> scMOYField[i], &logical);
    }

    scrSchDialog -> setDialogModify (vFALSE);
}


/*******************************************************************
*
*    vuScrSchDismissButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuScrSchDismissButtonClass, vbutton, 
					"vuScrSchDismissButtonClass");

void vuScrSchDismissButtonClass::ObserveDialogItem (vevent *event)
{
    int modified = scrSchDialog -> getDialogModify ();
    int result = VU_TRUE;

    if (modified == vTRUE)
    {
	result = vuMessageDialogCenter (scrSchDialog, VU_MOD_DISMISS_MSG);
    }
    if (result == VU_TRUE)
    {
	scrSchDialog -> Close ();
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    vuScrSchDeleteButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuScrSchDeleteButtonClass, vbutton, 
					"vuScrSchDeleteButtonClass");

void vuScrSchDeleteButtonClass::ObserveDialogItem (vevent *event)
{
    vstr *scriptID = scrSchDialog -> scriptID -> GetTextAsString ();
    if ( (vcharLength (scriptID)) == 0)
    {
	vuMessageDialogCenter (scrSchDialog, VU_SCRIPT_NO_ID_MSG);
	vstrDestroy (scriptID);
	return;
    }
    int result = vuMessageDialogCenter (scrSchDialog,
					VU_SCRIPT_DEL_REQ_MSG,
					scriptID);
    if (result == VU_TRUE)
    {
	char script[SCRIPTID_SZ + 1];
	vcharExportBounded (vcharScribeSystemExport (scriptID),
			    script,
			    SCRIPTID_SZ);
	script[SCRIPTID_SZ] = 0;

	vuMainDialogClass *mainDialog = scrSchDialog -> getMainDialog ();
	dataVuApplication *vuApp = mainDialog -> getApplication ();
	vuDatabaseClass *db = vuApp -> getDatabaseClass ();
	DB_RECORD_PTR rec;

	if ( (rec = DB_Seek (db -> vuScriptTable, script)) != 0)
	{
	    DB_DeleteRecord (db -> vuScriptTable, rec);
	    // log this with main
	    mainDialog -> vuMainMessageCenter (vunames_vuScriptDelMsg,
					       scriptID);

	    // update script id list from script select dialog, because its list
	    // is used for the Next and Prev button options

	    vuSelScrDialogClass *selScrDialog = scrSchDialog -> getSelScrDialog ();
	    if (selScrDialog != NULL)
	    {
		vliststr *list = (vliststr *)(selScrDialog -> selScrList -> GetList ());
		vlistIterator listIterator;
		int row;
		vchar *listScriptId;
		listIterator.Start ((vlist *)list);
		while (listIterator.Next ())
		{
		    row = listIterator.GetRow ();
		    listScriptId = (vchar *)(list -> GetCellValue (row, 0));
		    if ( (vcharCompareBounded (listScriptId, scriptID, SCRIPTID_SZ)) == 0)
		    {
			list -> RemoveRow (row);
			break;
		    }
		}
		listIterator.Finish ();
	    }
	}
	else
	{
	    vuMessageDialogCenter (scrSchDialog, VU_SCRIPT_NOT_EXIST_MSG, scriptID);
	}
    }
    vstrDestroy (scriptID);
}


/*******************************************************************
*
*    vuScrSchNewButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuScrSchNewButtonClass, vbutton, 
					"vuScrSchNewButtonClass");

void vuScrSchNewButtonClass::ObserveDialogItem (vevent *event)
{
    scrSchDialog -> vuNewScript ();
}


/*******************************************************************
*
*    vuScrSchGoToButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuScrSchGoToButtonClass, vbutton, 
					"vuScrSchGoToButtonClass");

void vuScrSchGoToButtonClass::ObserveDialogItem (vevent *event)
{
    vuSelScrDialogClass *selScrDialog = scrSchDialog -> getSelScrDialog ();
    if (selScrDialog == NULL)
    {
	vuMainDialogClass *mainDialog = scrSchDialog -> getMainDialog ();
	dataVuApplication *vuApp = mainDialog -> getApplication ();
	vresource vuRes = vuApp -> getDataVuvresource ();
	vresource selscrRes = vresourceGet (vuRes, vunames_vuSelectScriptDialog);
	selScrDialog = new vuSelScrDialogClass (selscrRes);
	scrSchDialog -> storeSelScrDialog (selScrDialog);
    }
    selScrDialog -> vuSelScrOpen (scrSchDialog);
    vchar *scriptId = selScrDialog -> vuGetScriptId ();
    if (scriptId[0] != 0)
    {
	scrSchDialog -> vuLoadScript (scriptId);
    }
}


/*******************************************************************
*
*    vuScrSchNextButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuScrSchNextButtonClass, vbutton, 
					"vuScrSchNextButtonClass");

void vuScrSchNextButtonClass::ObserveDialogItem (vevent *event)
{
    vuSelScrDialogClass *selScrDialog = scrSchDialog -> getSelScrDialog ();
    if (selScrDialog == NULL)
    {
	vuMessageDialogCenter (scrSchDialog, VU_NO_LIST_ERR);
	return;
    }
    vliststr *list = (vliststr *)(selScrDialog -> selScrList -> GetList ());
    vlistSelection *selection = selScrDialog -> selScrList -> GetSelection ();
    int row, column;
    list -> GetSelectedCell (selection, &row, &column);
    if (row != vlistNO_CELLS_SELECTED)
    {
	int numRows = list -> GetRowCount ();
	if ( (row + 1) < numRows)
	{
	    if ( (scrSchDialog -> getDialogModify ()) == vTRUE)
	    {
		int result = vuMessageDialogCenter (scrSchDialog, VU_MOD_DISMISS_MSG);
		if (result == VU_FALSE)
		{
		    return;
		}
	    }
	    list -> UnselectRow (selection, row);
	    row++;
	    list -> SelectRow (selection, row);
	    vchar *scriptId = (vchar *)(list -> GetCellValue (row, 0));
	    scrSchDialog -> vuLoadScript (scriptId);
	}
	else
	{
	    vuMessageDialogCenter (scrSchDialog, VU_ENDOFLIST_MSG);
	}
    }
    else
    {
	vuMessageDialogCenter (scrSchDialog, VU_SCRIPT_NO_SELECTED_MSG);
    }
}


/*******************************************************************
*
*    vuScrSchPrevButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuScrSchPrevButtonClass, vbutton, 
					"vuScrSchPrevButtonClass");

void vuScrSchPrevButtonClass::ObserveDialogItem (vevent *event)
{
    vuSelScrDialogClass *selScrDialog = scrSchDialog -> getSelScrDialog ();
    if (selScrDialog == NULL)
    {
	vuMessageDialogCenter (scrSchDialog, VU_NO_LIST_ERR);
	return;
    }
    vliststr *list = (vliststr *)(selScrDialog -> selScrList -> GetList ());
    vlistSelection *selection = selScrDialog -> selScrList -> GetSelection ();
    int row, column;
    list -> GetSelectedCell (selection, &row, &column);
    if (row != vlistNO_CELLS_SELECTED)
    {
	if ( (row - 1) >= 0)
	{
	    if ( (scrSchDialog -> getDialogModify ()) == vTRUE)
	    {
		int result = vuMessageDialogCenter (scrSchDialog, VU_MOD_DISMISS_MSG);
		if (result == VU_FALSE)
		{
		    return;
		}
	    }
	    list -> UnselectRow (selection, row);
	    row--;
	    list -> SelectRow (selection, row);
	    vchar *scriptId = (vchar *)(list -> GetCellValue (row, 0));
	    scrSchDialog -> vuLoadScript (scriptId);
	}
	else
	{
	    vuMessageDialogCenter (scrSchDialog, VU_STARTOFLIST_MSG);
	}
    }
    else
    {
	vuMessageDialogCenter (scrSchDialog, VU_SCRIPT_NO_SELECTED_MSG);
    }
}


/**********************************************************************
*
*    class that defines vuScrSchDomClass
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION(vuScrSchDomClass, exclusiveContainerClass, 
						"vuScrSchDomClass");

void vuScrSchDomClass::ObserveDialogItem (vevent *event)
{
    // call base class, to set modified flag
    exclusiveContainerClass::ObserveDialogItem (event);

    // if Selected Day of Week is not selected disable the 
    // dayOfWeekValue menu
    vuScrSchDialogClass *scrSchDialog = (vuScrSchDialogClass *)GetUserPtr ();
    vcontrolToggle *toggle = (vcontrolToggle *)scrSchDialog -> dayOfMonth -> GetSelection ();
    if (toggle != NULL)
    {
	const vchar *toggleTitle = toggle -> GetTitle ();
	if (toggleTitle[0] != 'S')
	{
	    scrSchDialog -> dayOfMonthValue -> SetValue (0);
	    scrSchDialog -> dayOfMonthValue -> Disable ();
	}
	else
	{
	    scrSchDialog -> dayOfMonthValue -> Enable ();
	}
    }
}



/**********************************************************************
*
*    class that defines vuScrSchDomValueClass
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION(vuScrSchDomValueClass, optionMenuClass, 
						"vuScrSchDomValueClass");

/**********************************************************************
*
*	vuScrSchDomValueClass observe function
*
**********************************************************************/

void vuScrSchDomValueClass::ObserveDialogItem (vevent *event)
{
    // call base class, to set modified flag
    optionMenuClass::ObserveDialogItem (event);

    // get current value of control, the day of month selected
    int controlValue = this -> GetValue ();

    if ( (controlValue >= this -> GetMin ()) &&
	 (controlValue <= this -> GetMax ()) )
    {
	vuScrSchDialogClass *scrSchDialog = (vuScrSchDialogClass *)GetUserPtr ();
	int itemCount = scrSchDialog -> monthOfYear -> GetItemCount ();
	vcontrolToggle *month;
	int i;

	// if day of month is the 28th, 29th, 30th, 31st
	// NOTE: control value is index value 0-30 corresponding to
	//	 possible days in month 1-31
	if (controlValue > 26)
	{
	    vdate *date = new vdate;
	    date -> SetCurrent ();
	    int year = date -> GetYear ();
	    int daysInMonth;
	    for (i = 0; i < itemCount; i++)
	    {
		daysInMonth = vdate::GetDaysInMonth (i, year);
		month = (vcontrolToggle *)scrSchDialog -> monthOfYear -> GetItemAt (i);
		if ( (controlValue + 1) > daysInMonth)
		{
		    month -> SetValue (0);
		    month -> Disable ();
		}
		else
		{
		    month -> Enable ();
		}
	    }
	    delete date;
	}
	// else all months can be enabled
	else
	{
	    for (i = 0; i < itemCount; i++)
	    {
		month = (vcontrolToggle *)scrSchDialog -> monthOfYear -> GetItemAt (i);
		if ( (month -> IsEnabled ()) == vFALSE)
		{
		    month -> Enable ();
		}
	    }
	}
    }
}


/***********************************************************************
*
*    vuScrSchDomDialog build days of month menu function
*
***********************************************************************/

void vuScrSchDomValueClass::buildMenu (void)
{
    int		i;
    vmenu	*newMenu = new vmenu;
    vmenuItem	*menuItem;

    for (i = 1; i < 32; i++)
    {
	menuItem = new vmenuItem;
	menuItem -> SetTitleScribed (vnum::ScribeInt (i));
	newMenu -> AppendItem (menuItem);
    }

    // attach the menu to the control
    this -> SetMenu (newMenu);
    // set min and max for the control
    this -> SetMin (0);
    this -> SetMax (30);
    // set value so current menu item will be the 1st one
    this -> SetValue (0);
    return;
}



/*******************************************************************
*
*	SELECT SCRIPT DIALOG CLASS
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuSelScrDialogClass, vdialog, 
						"vuSelSCrDialogClass");

/*******************************************************************
*
*    vuSelScrDialogClass 
*	LoadInit, constructor
*
*******************************************************************/

void vuSelScrDialogClass::LoadInit(vresource res)
{
    selScrList = vuSelScrListClass::RequestSubLoad (this,
					vunames_vuSelScrList);

    applyButton = vuSelScrApplyButtonClass::RequestSubLoad (this,
					vunames_vuSelScrApplyButton);

    dismissButton = vuSelScrDismissButtonClass::RequestSubLoad (this,
					vunames_vuSelScrDismissButton);

    // call base class loadInit to finish the job
    vdialog::LoadInit(res);

    selScrList -> storeSelScrDialog (this);
    applyButton -> storeSelScrDialog (this);
    dismissButton -> storeSelScrDialog (this);

}


/*******************************************************************
*
*    vuScrSelDialogClass notify function
*
*******************************************************************/

void vuSelScrDialogClass::ObserveWindow(vevent *event)
{
    if ( event -> GetType() == veventWINDOW_STATE && event -> IsClose())
    {
	// zap scriptId so caller will act on NULL string
	vmemSet ((void *)scriptId, 0, SCRIPTID_SZ);
	scriptId[SCRIPTID_SZ] = 0;
	Close();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuSelScrDialogClass Destroy function
*	if dialog is open, close it and stop its event loop
*
*******************************************************************/

void vuSelScrDialogClass::Destroy(void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }
    vliststr *list = (vliststr *)(selScrList -> GetList ());
    selScrList -> SetList (NULL);
    list -> Destroy ();

    vdialog::Destroy();
}



/*******************************************************************
*
*    vuSelScrDialogClass Open function
*
*******************************************************************/

void vuSelScrDialogClass::vuSelScrOpen(vuScrSchDialogClass *scrSchDlg)
{
    scrSchDialog = scrSchDlg;
    vuMainDialogClass *mainDialog = scrSchDialog -> getMainDialog ();
    dataVuApplication *vuApp = mainDialog -> getApplication ();
    vuDatabaseClass *db = vuApp -> getDatabaseClass ();


    vliststr *list = (vliststr *)(selScrList -> GetList ());

    // destroy old list if it exists
    if (list != NULL)
    {
	selScrList -> SetList (NULL);
	list -> Destroy ();
    }

    // new list to populate
    vliststr *newList = new vliststr;

    int type;
    int row = 0;
    char scIdName[SCRIPTID_SZ + 1];

    DB_RECORD_PTR rec = DB_Top (db -> vuScriptTable);
    while (rec)
    {
	DB_GetFieldValue (rec, scrSchDialog -> scIdField, scIdName, &type);
	scIdName[SCRIPTID_SZ] = 0;
	newList -> SetCellValueScribed (row, 0, vcharScribeSystem (scIdName));
	row++;
	rec = DB_SeekNext (db -> vuScriptTable);
    }

    newList -> Sort (vlistROW, 0, NULL);
    selScrList -> SetList (newList);
    vlistSelection *listSelection = selScrList -> GetSelection ();
    // only one selection at a time
    newList -> SetSelectMethod (listSelection, vlistONE_ONLY_SELECTION);

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    // place window in center of screen
    Place(scrSchDlg, vrectPLACE_CENTER, vrectPLACE_CENTER);
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    vuScrSelApplyButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuSelScrApplyButtonClass, vbutton, 
					"vuSelScrApplyButtonClass");

void vuSelScrApplyButtonClass::ObserveDialogItem (vevent *event)
{
    vchar *scriptId = selScrDialog -> vuGetScriptId ();
    int row, column;
    const vchar *itemValue;

    vliststr *list = (vliststr *)selScrDialog -> selScrList -> GetList ();
    vlistSelection *listSelection = selScrDialog -> selScrList -> GetSelection ();
    list -> GetSelectedCell (listSelection, &row, &column);
    if (row != vlistNO_CELLS_SELECTED)
    {
	itemValue = (const vchar *)(list -> GetCellValue (row, column));
	// copy value to script dialog
	vcharCopyBounded (itemValue, scriptId, SCRIPTID_SZ);
    }
    // no item selected, then set to zip, so caller 
    // processes NULL string
    else
    {
	vmemSet ((void *)scriptId, 0, SCRIPTID_SZ);
    }
    // close dialogand stop sub event loop
    selScrDialog -> Close ();
    vevent::StopProcessing ();
}


/*******************************************************************
*
*    vuSelScrDismissButtonClass Observer function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuSelScrDismissButtonClass, vbutton, 
					"vuSelScrDismissButtonClass");

void vuSelScrDismissButtonClass::ObserveDialogItem (vevent *event)
{
    vchar *scriptId = selScrDialog -> vuGetScriptId ();
    // zap it so caller of dialog will act on a NULL string
    vmemSet ((void *)scriptId, 0, SCRIPTID_SZ);
    selScrDialog -> Close ();
    vevent::StopProcessing ();
}


/*******************************************************************
*
*	SCRIPT SELECT LIST DIALOG CLASS
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuSelScrListClass, vlistview,
					"vuSelScrListClass")

/******************************************************************
*
*    vuSelScrListClass observe function
*	SELECTIN_HAS_CHANGED = user has release the mouse button
*	and commits the new selection, if we have a double mouse
*	click, then take same action as APPLY BUTTON
*
******************************************************************/

void vuSelScrListClass::ObserveSelect (int state)
{
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	int clickCount = GetLastPointerClickCount ();
	if (clickCount == 2)
	{
	    vchar *scriptId = selScrDialog -> vuGetScriptId ();
	    int row, column;
	    const vchar *itemValue;

	    vliststr *list = (vliststr *)GetList ();
	    vlistSelection *listSelection = GetSelection ();
	    list -> GetSelectedCell (listSelection, &row, &column);
	    itemValue = (const vchar *)(list -> GetCellValue (row, column));
	    // copy value to script dialog
	    vcharCopyBounded (itemValue, scriptId, SCRIPTID_SZ);
	    // close dialogand stop sub event loop
	    selScrDialog -> Close ();
	    vevent::StopProcessing ();
	}
    }
}


/**********************************************************************
*
*    class that overrides filteredTextitem class
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION(vuScrSchIdFilteredTextitem, filteredTextitem, 
						"vuScrSchIdFilteredTextitem");

/**********************************************************************
*
*    vuScrSchIfFilteredTextitem observe view override
*	we are looking for a loss of focus on the script id field
*
**********************************************************************/

void vuScrSchIdFilteredTextitem::ObserveView (vevent *event, int message)
{    if (message == vtextviewACCEPT_NOTIFY)
    {
	vdialogItemList	*list = this -> GetItemListIn ();
	vobjectAttributed *obj = list -> GetFocusItem ();
	if ( ((vobjectAttributed *)this) != obj)
	{
	    vdialogItemList *list = GetItemListIn ();
	    if (list != NULL)
	    {
		vdialog *dialog = list -> DetermineDialog ();
		if (dialog != NULL)
		{
		    vchar scriptId [SCRIPTID_SZ + 1];
		    GetText (scriptId, SCRIPTID_SZ);
		    if (scriptId[0] != 0)
		    {
			vuScrSchDialogClass *scrSchDlg = (vuScrSchDialogClass *)dialog;
			scrSchDlg -> vuLoadScript (scriptId);
		    }
		}
	    }
	}
    }
}


/**********************************************************************
*
*	filter functions unique to the script schedular
*
**********************************************************************/

static int vuScrSchIdFilter (filteredText *text, vtextSelection *sel,
			     const vchar **string, long *len)
{
    int		ret;
    vchar	aSpace;

    if (*len > 0)
    {
	aSpace = vcharFromLiteral (' ');
	if (**string == aSpace)
	{
	    long textLen = text -> GetLength ();
	    if (textLen != 0)
	    {
		vwindow::Beep ();
		ret = vFALSE;
	    }
	    else
	    {
		vuScrSchIdFilteredTextitem *textItem = 
					(vuScrSchIdFilteredTextitem *)text -> GetUserPtr ();
		if (textItem != NULL)
		{
		    int ok = vFALSE;
		    vuScrSchDialogClass *scrSchDialog = textItem -> getScrSchDialog ();
		    if (scrSchDialog != NULL)
		    {
    			vuSelScrDialogClass *selScrDialog = scrSchDialog -> getSelScrDialog ();
			if (selScrDialog == NULL)
			{
			    vuMainDialogClass *mainDialog = scrSchDialog -> getMainDialog ();
			    dataVuApplication *vuApp = mainDialog -> getApplication ();
			    vresource vuRes = vuApp -> getDataVuvresource ();
			    vresource selscrRes = vresourceGet (vuRes, 
							vunames_vuSelectScriptDialog);
			    selScrDialog = new vuSelScrDialogClass (selscrRes);
			    scrSchDialog -> storeSelScrDialog (selScrDialog);
			}
			selScrDialog -> vuSelScrOpen (scrSchDialog);
			const vchar *scriptId = (const vchar *)(selScrDialog -> vuGetScriptId ());
			vchar *newString = ((vchar *)(*string));
			if (scriptId[0] != 0)
			{
			    *len = vcharLength (scriptId);
			    vcharCopyBounded (scriptId, newString, SCRIPTID_SZ);
			    ret = NCharFilter (text, sel, string, len, SCRIPTID_SZ - 1);
			    ok = vTRUE;
			}
			else
			{
			    ok = vFALSE;
			}
		    }
		    if (ok == vFALSE)
		    {
			vwindow::Beep ();
			ret = vFALSE;
		    }
		}
		else
		{
		    vwindow::Beep ();
		    ret = vFALSE;
		}
	    }
	}
	else
	{
	    ret = NCharFilter (text, sel, string, len, SCRIPTID_SZ - 1);
	}
    }
    else
    {
	vwindow::Beep ();
	ret = vFALSE;
    }

    return (ret);
}


static int vuScrSchNameFilter (filteredText *text, vtextSelection *sel,
			       const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, SCRIPTPROC_SZ - 1) );
}

static int vuScrSchDescFilter (filteredText *text, vtextSelection *sel,
			       const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, SCRIPTDESC_SZ - 1) );
}

static int vuScrSchParamFilter (filteredText *text, vtextSelection *sel,
			        const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, SCRIPTPARAM_SZ - 1) );
}

