#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vnumHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbFldTypes.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "dbRecDes.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vuCommonDialogClass.hxx"
#include "vuStringoutDialogClass.hxx"
#include "vuNavigatorDialogClass.hxx"
#include "vuUtils.hxx"


vkindTYPED_FULL_DEFINITION(vuStringoutDialogClass, vuCommonDialogClass, "vuStringoutDialogClass");

/*******************************************************************
*
*    vuStringoutDialogClass override for constructors init ()
*
*******************************************************************/

void vuStringoutDialogClass::LoadInit (vresource res)
{
    // load in the ai specific fields
    stringoutContainer = vuStringoutContainerClass::RequestSubLoad (this,
				vunames_vuCommonContainer);
    vuCommonDialogClass::LoadInit (res);

    // store dialog modified flag
    stringoutContainer -> boardType -> vuSetModifiedPtr (&this -> dialogModified);
    stringoutContainer -> line -> SetTextModifiedPtr (&this -> dialogModified);
    stringoutContainer -> addr -> SetTextModifiedPtr (&this -> dialogModified);
    stringoutContainer -> card -> SetTextModifiedPtr (&this -> dialogModified);
    stringoutContainer -> signal -> SetTextModifiedPtr (&this -> dialogModified);
    stringoutContainer -> valueLoc -> SetTextModifiedPtr (&this -> dialogModified);
    stringoutContainer -> output -> SetTextModifiedPtr (&this -> dialogModified);
    stringoutContainer -> onScan -> vuSetModifiedPtr (&this -> dialogModified);
    stringoutContainer -> outputOnChange -> vuSetModifiedPtr (&this -> dialogModified);
    stringoutContainer -> closedLoop -> vuSetModifiedPtr (&this -> dialogModified);
    stringoutContainer -> stringValue -> SetTextModifiedPtr (&this -> dialogModified);

    // store dialog ptr for vuPvNameButtonClass
    stringoutContainer -> valueLocButton -> storeCommonDialog ((vuCommonDialogClass *)this);
    stringoutContainer -> outputButton -> storeCommonDialog ((vuCommonDialogClass *)this);

    // store dialog ptr for vuPvNameFilter
    stringoutContainer -> valueLoc -> storeCommonDialog ((vuCommonDialogClass *)this);
    stringoutContainer -> output -> storeCommonDialog ((vuCommonDialogClass *)this);

    // set the point type
    setPointType (VU_STRINGOUT);

    // get the title from the resource file, biTitle must be
    // freed by the class destuctor
    vapplication *app = vapplication::GetCurrent ();
    vresource dialogTitles = vresourceGet (app -> GetResources (),
				vunames_vuDialogTitles);
    stringoutTitle = vresourceGetString (dialogTitles, 
				vunames_vuStringoutDialogTitle);
    SetTitle (stringoutTitle);

    // trend button is not needed 
    vuCommonTrendButtonClass *trendbutton = getTrendButton ();
    trendbutton -> Disable ();
}


/*******************************************************************
*
*    vuStringoutDialogClass Destroy function
*
*******************************************************************/

void vuStringoutDialogClass::Destroy (void)
{
    vstrDestroy (stringoutTitle);
    vuCommonDialogClass::Destroy ();
}


/*******************************************************************
*
*    vuStringoutDialogClass, reset fields, overrides base class 
*	vuCommonDialogClass
*
*******************************************************************/

void vuStringoutDialogClass::resetDialog (void)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    vchar		epics[FLDNAME_SZ + 1];

    vresource defRes = mainDialog -> getDefaultsResource ();

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0; 
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	// if not a common dialog field
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    dialogItem = vuFindContainerItem (stringoutContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    if ( (dialogItem -> IsKindOf (&vtextitem::Kind)) == vTRUE)
	    {
		textItem = (vtextitem *)dialogItem;

		// set the default value for the Bo's line field
		if ( ((vcharCompareBounded (epics, "LINE", FLDNAME_SZ)) == 0) ||
		     ((vcharCompareBounded (epics, "ADDR", FLDNAME_SZ)) == 0) ||
		     ((vcharCompareBounded (epics, "CARD", FLDNAME_SZ)) == 0) ||
		     ((vcharCompareBounded (epics, "SIGN", FLDNAME_SZ)) == 0) )
		{
		    vstr *zero = vresourceGetString (defRes, 
						vunames_vuZeroDefault);
		    textItem -> SetText (zero);
		    vstrDestroy (zero);
		}
		else
		{
		    textItem -> SetTextScribed (vcharScribeSystem (""));
		}
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolToggle::Kind)) == vTRUE)
	    {
		toggleItem = (vcontrolToggle *)dialogItem;
		toggleItem -> SetValue (0);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolOptionMenu::Kind)) == vTRUE)
	    {
		optionMenu = (vcontrolOptionMenu *)dialogItem;
		optionMenu -> SetValue (0);
	    }
	}  // end if not in common dialog
    }
    iterator.Finish ();

    // call base class reset function with proper point index
    vuCommonDialogClass::resetDialog ();

    dialogModified = vFALSE;
}


/*******************************************************************
*
*    vuLoadDialog 
*
*******************************************************************/

void vuStringoutDialogClass::vuLoadDialog (vchar *pointName)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    vchar		epics[FLDNAME_SZ + 1];
    vstr		*finalValue;
    vstr		*referenceString;
    vstr		*constantString;

    char		pvName [PVNAME_SZ + 1];
    DB_TABLE_PTR	lastTable;	// pointer to last table used
    DB_RECORD_PTR	pntRecord;	// codebase record ptr
    int			dbaseType;	// codebase field type
    int			intValue;
    unsigned int	uIntValue;
    long		longValue;
    unsigned long	uLongValue;
    double		doubleValue;
    DTYP_TYPE		dType;		// device type indicator

    vcharExportBounded (vcharScribeSystemExport (pointName),
			pvName,
			PVNAME_SZ);

    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
    pntRecord = DB_Seek (db -> vuPvmapTable, pvName);
    if (pntRecord == (DB_RECORD_PTR)0)
    {
	return;
    }
    else
    {
	int ptType;
	DB_GetFieldValue (pntRecord, db -> pvmapTypeField, dbValue, &dbaseType);
	ptType = vuGetPointTypeIndex ((char *)dbValue);
	// the point name in pvname field is NOT a STRINGOUT
	if (ptType != VU_STRINGOUT)
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_PTTYPEDIALOGERR,
				   pointName, ptType, VU_STRINGOUT);
	    vuClearPvNameField ();
	    return;
	}
	// the point name entered in pvname field has been deleted
	if (DB_IsDeleted (db -> vuPvmapTable))
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_PTISDELETEDERR, pointName);
	    vuClearPvNameField ();
	    return;
	}
    }
    // call base class vuLoadDialog first
    vuCommonDialogClass::vuLoadDialog (pointName);

    // since the stringout can be a hardware point, find what we are for later use
    // by getting the DTYP (device type) dictionary entry
    referenceString = vstrCloneScribed (vcharScribeSystem ("REFERENCE"));
    constantString = vstrCloneScribed (vcharScribeSystem ("CONSTANT"));
    xrefNode = (xrefNodeClass *)(dict -> GetElem (
				vnameGetGlobalScribed (vcharScribeSystem ("DTYP"))));
    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
    pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
    DB_GetFieldValue (pntRecord, xrefNode -> xrefField, dbValue, &dbaseType);
    if ( (vcharCompare ((const vchar *)dbValue, referenceString)) == 0)
	dType = VU_REFERENCE;
    else if ( (vcharCompare ((const vchar *)dbValue, constantString)) == 0)
	dType = VU_CONSTANT;
    else
	dType = VU_HARDWARE;
    vstrDestroy (referenceString);
    vstrDestroy (constantString);

    lastTable = NULL;

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0;
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    // only seek if table has changed
	    if (lastTable != xrefNode -> xrefTable)
		{
		    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
			pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
		}
	    lastTable = xrefNode -> xrefTable;
	    dialogItem = vuFindContainerItem (stringoutContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    DB_GetFieldValue (pntRecord, xrefNode -> xrefField, dbValue, &dbaseType);
	    doubleValue = 0.0;
	    // GetField actually returns a number in the char array
	    if ( (dbaseType == 'F') || (dbaseType == 'N') )
	    {
		doubleValue = *((double *)dbValue);
	    }
	    switch (xrefNode -> xrefEPICSFieldType)
	    {
		case DBF_SHORT:
		    intValue = (int)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeInt (intValue));
		    break;

		case DBF_LONG:
		    longValue = (long)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeLong (longValue));
		    break;

		case DBF_ENUM:
		case DBF_USHORT:
		    uIntValue = (unsigned int)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeUInt (uIntValue));
		    break;

		case DBF_ULONG:
		    uLongValue = (unsigned long)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeULong (uLongValue));
		    break;

		case DBF_FLOAT:
		case DBF_DOUBLE:
		    // to output CONSTANT field, device type must be CONSTANT
		    if ( (strncmp (xrefNode -> xrefEpicsName, "CNST", FLDNAME_SZ)) 
										== 0)
		    {
			if (dType != VU_CONSTANT)
			    continue;
		    }
		    finalValue = vstrCloneScribed (vnumScribeFloat (doubleValue));
		    break;

		case DBF_STRING:
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		case DBF_GBLCHOICE:
		case DBF_RECCHOICE:
		case DBF_CVTCHOICE:
		case DBF_DEVCHOICE:
		    if (dbaseType == 'L')
		    {
			int	logical;

			logical = (*((char*)dbValue) ? 
					xrefNode -> xrefTrue : 
					(xrefNode -> xrefTrue ^1));
			finalValue = vstrCloneScribed (vnumScribeInt (logical));
			break;
		    }
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		case DBF_OUTLINK:
		    // device type must be REFERENCE to output the reference
		    if (dType == VU_REFERENCE)
		    {
			DB_GetFieldValue (pntRecord, xrefNode -> xrefRefField, 
								dbValue, &dbaseType);
			finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    }
		    else
		    {
			finalValue = vstrCloneScribed (vcharScribeSystem (""));
		    }
		    break;

		case DBF_INLINK:
		    // works for all bo inputs, see pdbxref
		    DB_GetFieldValue (pntRecord, xrefNode -> xrefRefField, 
								dbValue, &dbaseType);
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		default:
		    continue;
	    }
	    if ( (dialogItem -> IsKindOf (&vtextitem::Kind)) == vTRUE)
	    {
		textItem = (vtextitem *)dialogItem;
		textItem -> SetTextScribed (vcharScribeSystem (""));
		textItem -> SetText (finalValue);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolToggle::Kind)) == vTRUE)
	    {
		toggleItem = (vcontrolToggle *)dialogItem;
		if (finalValue[0] == '0')
		    toggleItem -> SetValue (0);
		else
		    toggleItem -> SetValue (1);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolOptionMenu::Kind)) == vTRUE)
	    {
		optionMenu = (vcontrolOptionMenu *)dialogItem;
		vuSetOptionMenuValue (optionMenu, finalValue);
		vstrDestroy (finalValue);
	    }
	}  // end if not in commondialog
    }
    iterator.Finish ();

    setDialogModify (vFALSE);
}


/*******************************************************************
*
*    vuSaveDialog 
*
*******************************************************************/

void vuStringoutDialogClass::vuSaveDialog (vchar *pointName)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    const vchar 	*optionMenuValue;
    vchar		epics[FLDNAME_SZ + 1];
    vstr		*stringValue;
    vstr		*referenceString;
    vstr		*constantString;

    char		pvName [PVNAME_SZ + 1];
    char		vuCurrentBoard [VU_KEY_SIZE];
    DB_FIELD_PTR	putField;
    DB_TABLE_PTR	lastTable;	// pointer to last table used
    DB_RECORD_PTR	pntRecord;	// codebase record ptr
    int			dType;		// device type
    double		doubleValue;
    int			doSave;

    vcharExportBounded (vcharScribeSystemExport (pointName),
			pvName,
			PVNAME_SZ);

    // since the stringout can be a hardware point, find what we are for later use
    // by getting the DTYP (BOARD) value from the dialog
    referenceString = vstrCloneScribed (vcharScribeSystem ("REFERENCE"));
    constantString = vstrCloneScribed (vcharScribeSystem ("CONSTANT"));
    vcharCopyScribedBounded (vcharScribeSystem ("DTYP"), epics, FLDNAME_SZ);
    epics[sizeof (epics) - 1] = 0;
    // find the BOARD (DTYP) field and get its value
    dialogItem = vuFindContainerItem (stringoutContainer, epics);
    optionMenu = (vcontrolOptionMenu *)dialogItem;
    optionMenuValue = vuGetOptionMenuValue (optionMenu);
    vcharExport (vcharScribeSystemExport (optionMenuValue), (void *)vuCurrentBoard);
    if ( (vcharCompare (optionMenuValue, referenceString)) == 0)
	dType = VU_REFERENCE;
    else if ( (vcharCompare (optionMenuValue, constantString)) == 0)
	dType = VU_CONSTANT;
    else
	dType = VU_HARDWARE;
    vstrDestroy (referenceString);
    vstrDestroy (constantString);

    lastTable = NULL;

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0;
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    doSave = vTRUE;
	    // only seek if table has changed
	    if (lastTable != xrefNode -> xrefTable)
		{
		    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
			pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
		}
	    lastTable = xrefNode -> xrefTable;

	    putField = xrefNode -> xrefField;

	    dialogItem = vuFindContainerItem (stringoutContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    switch (xrefNode -> xrefEPICSFieldType)
	    {
		case DBF_SHORT:
		case DBF_LONG:
		case DBF_ENUM:
		case DBF_USHORT:
		case DBF_ULONG:
		case DBF_FLOAT:
		case DBF_DOUBLE:
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    doubleValue = atof ((const char *)dbValue);
		    if ( (strncmp (xrefNode -> xrefEpicsName, "CNST", FLDNAME_SZ)) 
										== 0)
		    {
			if (dType != VU_CONSTANT)
			    doubleValue = 0;
		    }
		    DB_PutFieldValue (pntRecord, putField, (void *)&doubleValue);
		    doSave = vFALSE;
		    break;

		case DBF_STRING:
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    break;

		case DBF_GBLCHOICE:
		case DBF_RECCHOICE:
		case DBF_CVTCHOICE:
		case DBF_DEVCHOICE:
		{
		    if (xrefNode -> xrefField -> field_type == 'L')
		    {
			int	logical;
			toggleItem = (vcontrolToggle *)dialogItem;
			logical = toggleItem -> GetValue ();
			*((char *)dbValue) = (logical ? xrefNode -> xrefTrue : 
								xrefNode -> xrefTrue ^1);
			break;
		    }
		    optionMenu = (vcontrolOptionMenu *)dialogItem;
		    optionMenuValue = vuGetOptionMenuValue (optionMenu);
		    vcharExport (vcharScribeSystemExport (optionMenuValue), dbValue);
		    if ( (strncmp (xrefNode -> xrefEpicsName, "DTYP", FLDNAME_SZ)) == 0)
		    {
			// write to DEVTYPE in pvmap
			DB_PutFieldValue (pntRecord, xrefNode -> xrefField, dbValue);
		    }
		    break;
		}

		case DBF_OUTLINK:
		    // there is only 1 output and that is OUT
		    // always write the boardtype from DTYP into the OUT field
		    DB_PutFieldValue (pntRecord, putField, vuCurrentBoard);
		    // if we are a reference then write OUT into the REFERENCE field
		    if (dType == VU_REFERENCE)
		    {
			putField = xrefNode -> xrefRefField;
			textItem = (vtextitem *)dialogItem;
			stringValue = textItem -> GetTextAsString ();
			vcharExport (vcharScribeSystemExport (stringValue), dbValue);
			vstrDestroy (stringValue);
		    }
		    else
		    {
			// make sure to null out the field
			putField = xrefNode -> xrefRefField;
			vmemSet (dbValue, 0, VU_KEY_SIZE);
		    }
		    break;

		case DBF_INLINK:
		    // for DOL, we must put its value to the OUTPUTREF field
		    // other wise it and other INLINKS are the same 
		    if ( (strncmp (xrefNode -> xrefEpicsName, "DOL ", FLDNAME_SZ)) == 0)
		    {
			putField = xrefNode -> xrefRefField;
		    }
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    break;

		default:
		    continue;

	    }  // end xrefNode -> xrefEPICSFieldType

	    if (doSave == vTRUE)
	    {
		DB_PutFieldValue (pntRecord, putField, dbValue);
	    }

	}  // end if not in commondialog

    }  // end while iterator

    iterator.Finish ();

    // call base class vuSaveDialog
    vuCommonDialogClass::vuSaveDialog (pointName);
}


vkindTYPED_FULL_DEFINITION(vuStringoutContainerClass, vcontainer, "vuStringoutContainerClass");

/*******************************************************************
*
*    vuStringoutContainerClass override for constructor
*
*******************************************************************/

void vuStringoutContainerClass::LoadInit (vresource res)
{
    vapplication *app = vapplication::GetCurrent ();
    vresource pointDictRes = vresourceGet (app -> GetResources (), 
				vunames_vuPointDictionary);
    vresource realRes = vresourceGet (pointDictRes, vunames_vuStringoutContainer);

    vcontainer *btcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringoutDTYPContainer);
    boardType = vuBoardTypeMenuClass::RequestSubLoad (btcontainer,
				vunames_DTYP);

    vcontainer *linecontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringoutLINEContainer);
    line = filteredTextitem::RequestSubLoad (linecontainer,
				vunames_LINE);

    vcontainer *addrcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringoutADDRContainer);
    addr = filteredTextitem::RequestSubLoad (addrcontainer,
				vunames_ADDR);

    vcontainer *cardcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringoutCARDContainer);
    card = filteredTextitem::RequestSubLoad (cardcontainer,
				vunames_CARD);

    vcontainer *signalcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringoutSIGNContainer);
    signal = filteredTextitem::RequestSubLoad (signalcontainer,
				vunames_SIGN);

    vcontainer *valcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringoutDOLContainer);
    valueLocButton = vuPvNameButtonClass::RequestSubLoad (valcontainer,
				vunames_vuStringoutDOLButton);
    valueLoc = vuFilteredTextitem::RequestSubLoad (valcontainer,
				vunames_DOL);

    vcontainer *outcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringoutOUTContainer);
    outputButton = vuPvNameButtonClass::RequestSubLoad (outcontainer,
				vunames_vuStringoutOUTButton);
    output = vuFilteredTextitem::RequestSubLoad (outcontainer,
				vunames_OUT);

    onScan = vuCheckBoxClass::RequestSubLoad (this,
				vunames_OSCN);

    closedLoop = vuCheckBoxClass::RequestSubLoad (this,
				vunames_OMSL);

    outputOnChange = vuCheckBoxClass::RequestSubLoad (this,
				vunames_OCHG);

    vcontainer *stvalcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuStringoutVALContainer);
    stringValue = filteredTextitem::RequestSubLoad (stvalcontainer,
				vunames_VAL);

    vcontainer::LoadInit (realRes);

    // store ptr to text field for vuPvNameButtonClass
    valueLocButton -> storePvName (valueLoc);
    outputButton -> storePvName (output);

    // set filters
    line -> SetItemFilterFunc (numericFilter);
    addr -> SetItemFilterFunc (numericFilter);
    card -> SetItemFilterFunc (numericFilter);
    signal -> SetItemFilterFunc (numericFilter);
    stringValue -> SetItemFilterFunc (stringValFilter);

    valueLoc -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *valText = (filteredText *)valueLoc -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    valText -> SetUserPtr ((void *)valueLoc);

    output -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *outText = (filteredText *)output -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    outText -> SetUserPtr ((void *)output);

    // build popup menus
    boardType -> vuBuildBoardTypeMenu (VU_STRINGOUT);

}

/*******************************************************************
*
*    vuStringoutContainerClass override for destructor
*
*******************************************************************/

void vuStringoutContainerClass::Destroy (void)
{
    vcontainer::Destroy ();
}
