#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vnumHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER
#include vlistHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbFldTypes.h"
#include "choiceGbl.h"
#include "choiceUdb.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "dbRecDes.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vuCommonDialogClass.hxx"
#include "vuUdbDialogClass.hxx"
#include "vuNavigatorDialogClass.hxx"
#include "vuUtils.hxx"

static int vuUdbProgramNameSpecialFilter (filteredText *text, vtextSelection *sel,
					  const vchar **string, long *len);
static vlistitemList *vuUdbGetProgramNames (vuDatabaseClass *db, int useCfgTable);
static int vuUdbProgramNameFilter (filteredText *text, vtextSelection *sel,
				   const vchar **string, long *len);
static int vuUdbLabelFilter (filteredText *text, vtextSelection *sel,
			     const vchar **string, long *len);


/*******************************************************************
*
*    vuUdbDialogClass 
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbDialogClass, vuCommonDialogClass, 
						"vuUdbDialogClass");

/*******************************************************************
*
*    vuUdbDialogClass override for constructors init ()
*
*******************************************************************/

void vuUdbDialogClass::LoadInit (vresource res)
{
    // load in the acc specific fields
    udbContainer = vuUdbContainerClass::RequestSubLoad (this,
				vunames_vuCommonContainer);
    vuCommonDialogClass::LoadInit (res);

    // set the point type
    setPointType (VU_UDB);

    // get the title from the resource file, udbTitle must be
    // freed by the class destuctor
    vapplication *app = vapplication::GetCurrent ();
    vresource dialogTitles = vresourceGet (app -> GetResources (),
				vunames_vuDialogTitles);
    udbTitle = vresourceGetString (dialogTitles, 
				vunames_vuUdbDialogTitle);
    SetTitle (udbTitle);

    // start the labels list
    dialogLabels = new vlist;

    // set the labels modifed flag
    udbLabelsModified = vFALSE;

    // load up the the other UDB dialogs
    vresource pointDictRes = vresourceGet (app -> GetResources (), 
				vunames_vuPointDictionary);

    vresource statusDialogRes = vresourceGet (pointDictRes, vunames_vuUdbStatusDialog);
    statusDialog = new vuUdbStatusDialogClass (statusDialogRes);

    vresource ioDialogRes = vresourceGet (pointDictRes, vunames_vuUdbIODialog);
    ioDialog = new vuUdbIODialogClass (ioDialogRes);

    vresource constantDialogRes = vresourceGet (pointDictRes, 
					vunames_vuUdbConstantsDialog);
    constantsDialog = new vuUdbConstantsDialogClass (constantDialogRes);

    vresource limitsDialogRes = vresourceGet (pointDictRes, vunames_vuUdbLimitsDialog);
    limitsDialog = new vuUdbLimitsDialogClass (limitsDialogRes);

    vresource miscDialogRes = vresourceGet (pointDictRes, vunames_vuUdbMiscDialog);
    miscDialog = new vuUdbMiscDialogClass (miscDialogRes);

    // store pointers to these dialogs in there respective buttons
    udbContainer -> statusButton -> storeStatusDialog (statusDialog);
    udbContainer -> ioButton -> storeIODialog (ioDialog);
    udbContainer -> constantsButton -> storeConstantsDialog (constantsDialog);
    udbContainer -> limitsButton -> storeLimitsDialog (limitsDialog);
    udbContainer -> miscButton -> storeMiscDialog (miscDialog);

    // store ptr to udb dialog in these buttons
    udbContainer -> ioButton -> storeUdbDialog (this);
    udbContainer -> constantsButton -> storeUdbDialog (this);
    udbContainer -> limitsButton -> storeUdbDialog (this);

    // store ptr to dialog for vuUdbProgramNameButtonClass
    udbContainer -> udbProgramNameButton -> storeUdbDialog (this);

    // store ptr to this dialog for vuUdbProgramNameSpecialFilter
    udbContainer -> udbProgramName -> storeCommonDialog ((vuCommonDialogClass *)this);

    // store dialog modified flag for the udb container fields
    udbContainer -> udbProgramName -> SetTextModifiedPtr (&this -> dialogModified);
    udbContainer -> onScan -> vuSetModifiedPtr (&this -> dialogModified);
    udbContainer -> onAlarm -> vuSetModifiedPtr (&this -> dialogModified);

    // store dialog modified flag for the udb status dialog fields
    statusDialog -> status1 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> status2 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> status3 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> status4 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> status5 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> status6 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> status7 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> status8 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> status9 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> status10 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> status11 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> status12 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> status1Severity -> vuSetModifiedPtr (&this -> dialogModified);
    statusDialog -> status2Severity -> vuSetModifiedPtr (&this -> dialogModified);
    statusDialog -> status3Severity -> vuSetModifiedPtr (&this -> dialogModified);
    statusDialog -> status4Severity -> vuSetModifiedPtr (&this -> dialogModified);
    statusDialog -> status5Severity -> vuSetModifiedPtr (&this -> dialogModified);
    statusDialog -> status6Severity -> vuSetModifiedPtr (&this -> dialogModified);
    statusDialog -> status7Severity -> vuSetModifiedPtr (&this -> dialogModified);
    statusDialog -> status8Severity -> vuSetModifiedPtr (&this -> dialogModified);
    statusDialog -> status9Severity -> vuSetModifiedPtr (&this -> dialogModified);
    statusDialog -> status10Severity -> vuSetModifiedPtr (&this -> dialogModified);
    statusDialog -> status11Severity -> vuSetModifiedPtr (&this -> dialogModified);
    statusDialog -> status12Severity -> vuSetModifiedPtr (&this -> dialogModified);
    statusDialog -> cmd1 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> cmd2 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> cmd3 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> cmd4 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> cmd5 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> cmd6 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> cmd7 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> cmd8 -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> cosvSeverity -> vuSetModifiedPtr (&this -> dialogModified);
    statusDialog -> undefinedSeverity -> vuSetModifiedPtr (&this -> dialogModified);
    statusDialog -> alarmGroup -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> alarmDelay -> SetTextModifiedPtr (&this -> dialogModified);
    statusDialog -> alarmDisable -> SetTextModifiedPtr (&this -> dialogModified);

    // store ptr to dialog for vuPvNameButtonClass for status dialog
    statusDialog -> alarmDisableButton -> storeCommonDialog ((vuCommonDialogClass *)this);

    // store ptr to dialog for vuPvNameFilter for status dialog
    statusDialog -> alarmDisable -> storeCommonDialog ((vuCommonDialogClass *)this);

    // store ptr to dialog for vuAlarmGroupFilter for status dialog
    statusDialog -> alarmGroup -> storeCommonDialog ((vuCommonDialogClass *)this);

    // store dialog modified flag for the udb io dialog fields
    ioDialog -> input1 -> SetTextModifiedPtr (&this -> dialogModified);
    ioDialog -> input2 -> SetTextModifiedPtr (&this -> dialogModified);
    ioDialog -> input3 -> SetTextModifiedPtr (&this -> dialogModified);
    ioDialog -> input4 -> SetTextModifiedPtr (&this -> dialogModified);
    ioDialog -> input5 -> SetTextModifiedPtr (&this -> dialogModified);
    ioDialog -> input6 -> SetTextModifiedPtr (&this -> dialogModified);
    ioDialog -> input7 -> SetTextModifiedPtr (&this -> dialogModified);
    ioDialog -> input8 -> SetTextModifiedPtr (&this -> dialogModified);
    ioDialog -> output1 -> SetTextModifiedPtr (&this -> dialogModified);
    ioDialog -> output2 -> SetTextModifiedPtr (&this -> dialogModified);
    ioDialog -> output3 -> SetTextModifiedPtr (&this -> dialogModified);
    ioDialog -> output4 -> SetTextModifiedPtr (&this -> dialogModified);
    ioDialog -> output5 -> SetTextModifiedPtr (&this -> dialogModified);
    ioDialog -> output6 -> SetTextModifiedPtr (&this -> dialogModified);
    ioDialog -> output7 -> SetTextModifiedPtr (&this -> dialogModified);
    ioDialog -> output8 -> SetTextModifiedPtr (&this -> dialogModified);
    ioDialog -> prc1 -> vuSetModifiedPtr (&this -> dialogModified);
    ioDialog -> prc2 -> vuSetModifiedPtr (&this -> dialogModified);
    ioDialog -> prc3 -> vuSetModifiedPtr (&this -> dialogModified);
    ioDialog -> prc4 -> vuSetModifiedPtr (&this -> dialogModified);
    ioDialog -> prc5 -> vuSetModifiedPtr (&this -> dialogModified);
    ioDialog -> prc6 -> vuSetModifiedPtr (&this -> dialogModified);
    ioDialog -> prc7 -> vuSetModifiedPtr (&this -> dialogModified);
    ioDialog -> prc8 -> vuSetModifiedPtr (&this -> dialogModified);
    ioDialog -> mxs1 -> vuSetModifiedPtr (&this -> dialogModified);
    ioDialog -> mxs2 -> vuSetModifiedPtr (&this -> dialogModified);
    ioDialog -> mxs3 -> vuSetModifiedPtr (&this -> dialogModified);
    ioDialog -> mxs4 -> vuSetModifiedPtr (&this -> dialogModified);
    ioDialog -> mxs5 -> vuSetModifiedPtr (&this -> dialogModified);
    ioDialog -> mxs6 -> vuSetModifiedPtr (&this -> dialogModified);
    ioDialog -> mxs7 -> vuSetModifiedPtr (&this -> dialogModified);
    ioDialog -> mxs8 -> vuSetModifiedPtr (&this -> dialogModified);

    // store ptr to dialog for vuPvNameButtonClass for io dialog
    ioDialog -> input1Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> input2Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> input3Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> input4Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> input5Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> input6Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> input7Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> input8Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> output1Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> output2Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> output3Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> output4Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> output5Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> output6Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> output7Button -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> output8Button -> storeCommonDialog ((vuCommonDialogClass *)this);

    // store ptr to dialog for vuPvNameFilter for io dialog
    ioDialog -> input1 -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> input2 -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> input3 -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> input4 -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> input5 -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> input6 -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> input7 -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> input8 -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> output1 -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> output2 -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> output3 -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> output4 -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> output5 -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> output6 -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> output7 -> storeCommonDialog ((vuCommonDialogClass *)this);
    ioDialog -> output8 -> storeCommonDialog ((vuCommonDialogClass *)this);

    // add io button ptrs to dialogLabels list, MUST BE DONE IN THIS ORDER
    vuUdbAddLabel ((vdialogItem *)(ioDialog -> input1Button), "vuUdbI1");
    vuUdbAddLabel ((vdialogItem *)(ioDialog -> input2Button), "vuUdbI2");
    vuUdbAddLabel ((vdialogItem *)(ioDialog -> input3Button), "vuUdbI3");
    vuUdbAddLabel ((vdialogItem *)(ioDialog -> input4Button), "vuUdbI4");
    vuUdbAddLabel ((vdialogItem *)(ioDialog -> input5Button), "vuUdbI5");
    vuUdbAddLabel ((vdialogItem *)(ioDialog -> input6Button), "vuUdbI6");
    vuUdbAddLabel ((vdialogItem *)(ioDialog -> input7Button), "vuUdbI7");
    vuUdbAddLabel ((vdialogItem *)(ioDialog -> input8Button), "vuUdbI8");
    vuUdbAddLabel ((vdialogItem *)(ioDialog -> output1Button), "vuUdbO1");
    vuUdbAddLabel ((vdialogItem *)(ioDialog -> output2Button), "vuUdbO2");
    vuUdbAddLabel ((vdialogItem *)(ioDialog -> output3Button), "vuUdbO3");
    vuUdbAddLabel ((vdialogItem *)(ioDialog -> output4Button), "vuUdbO4");
    vuUdbAddLabel ((vdialogItem *)(ioDialog -> output5Button), "vuUdbO5");
    vuUdbAddLabel ((vdialogItem *)(ioDialog -> output6Button), "vuUdbO6");
    vuUdbAddLabel ((vdialogItem *)(ioDialog -> output7Button), "vuUdbO7");
    vuUdbAddLabel ((vdialogItem *)(ioDialog -> output8Button), "vuUdbO8");

    // store dialog modified flag for the udb constants dialog fields
    constantsDialog -> float1 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> float2 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> float3 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> float4 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> float5 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> float6 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> float7 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> float8 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> float9 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> float10 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> float11 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> float12 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> double1 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> double2 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> double3 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> double4 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> double5 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> double6 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> ulong1 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> ulong2 -> SetTextModifiedPtr (&this -> dialogModified);
    constantsDialog -> float1Limit -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> float2Limit -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> float3Limit -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> float4Limit -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> float5Limit -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> float6Limit -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> float7Limit -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> float8Limit -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> float9Limit -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> float10Limit -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> float11Limit -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> float12Limit -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> ulong1Limit -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> ulong2Limit -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> chkbx1 -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> chkbx2 -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> chkbx3 -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> chkbx4 -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> chkbx5 -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> chkbx6 -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> chkbx7 -> vuSetModifiedPtr (&this -> dialogModified);
    constantsDialog -> chkbx8 -> vuSetModifiedPtr (&this -> dialogModified);

    // add constant label ptrs to dialogLabels list, MUST BE DONE IN THIS ORDER
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> float1Label), "FLOAT01");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> float2Label), "FLOAT02");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> float3Label), "FLOAT03");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> float4Label), "FLOAT04");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> float5Label), "FLOAT05");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> float6Label), "FLOAT06");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> float7Label), "FLOAT07");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> float8Label), "FLOAT08");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> float9Label), "FLOAT09");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> float10Label), "FLOAT10");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> float11Label), "FLOAT11");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> float12Label), "FLOAT12");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> double1Label), "DOUBLE01");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> double2Label), "DOUBLE02");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> double3Label), "DOUBLE03");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> double4Label), "DOUBLE04");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> double5Label), "DOUBLE05");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> double6Label), "DOUBLE06");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> ulong1Label), "ULONG01");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> ulong2Label), "ULONG02");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> chkbx1Label), "vuUdbBox1");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> chkbx2Label), "vuUdbBox2");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> chkbx3Label), "vuUdbBox3");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> chkbx4Label), "vuUdbBox4");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> chkbx5Label), "vuUdbBox5");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> chkbx6Label), "vuUdbBox6");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> chkbx7Label), "vuUdbBox7");
    vuUdbAddLabel ((vdialogItem *)(constantsDialog -> chkbx8Label), "vuUdbBox8");

    // store dialog modified flag for the udb limits dialog fields
    limitsDialog -> tempFloat1Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempFloat2Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempFloat3Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempFloat4Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempFloat5Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempFloat6Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempFloat7Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempFloat8Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempFloat9Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempFloat10Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempFloat11Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempFloat12Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempFloat13Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempFloat14Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempFloat15Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempFloat16Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempDouble1Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempDouble2Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempDouble3Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempDouble4Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempUlong1Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempUlong2Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempUlong3Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempUlong4Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempUlong5Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempUlong6Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempUlong7Limit -> vuSetModifiedPtr (&this -> dialogModified);
    limitsDialog -> tempUlong8Limit -> vuSetModifiedPtr (&this -> dialogModified);

    // add limit label ptrs to dialogLabels list, MUST BE DONE IN THIS ORDER
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempFloat1Label), "vuFloat1");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempFloat2Label), "vuFloat2");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempFloat3Label), "vuFloat3");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempFloat4Label), "vuFloat4");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempFloat5Label), "vuFloat5");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempFloat6Label), "vuFloat6");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempFloat7Label), "vuFloat7");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempFloat8Label), "vuFloat8");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempFloat9Label), "vuFloat9");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempFloat10Label), "vuFloat10");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempFloat11Label), "vuFloat11");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempFloat12Label), "vuFloat12");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempFloat13Label), "vuFloat13");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempFloat14Label), "vuFloat14");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempFloat15Label), "vuFloat15");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempFloat16Label), "vuFloat16");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempUlong1Label), "vuUlong1");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempUlong2Label), "vuUlong2");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempUlong3Label), "vuUlong3");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempUlong4Label), "vuUlong4");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempUlong5Label), "vuUlong5");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempUlong6Label), "vuUlong6");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempUlong7Label), "vuUlong7");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempUlong8Label), "vuUlong8");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempDouble1Label), "vuDouble1");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempDouble2Label), "vuDouble2");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempDouble3Label), "vuDouble3");
    vuUdbAddLabel ((vdialogItem *)(limitsDialog -> tempDouble4Label), "vuDouble4");

    // store dialog modified flag for the udb misc dialog fields
    miscDialog -> initialMode -> vuSetModifiedPtr (&this -> dialogModified);
    miscDialog -> mode01 -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> mode02 -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> mode03 -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> mode04 -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> mode05 -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> mode06 -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> mode07 -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> mode08 -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> initialFunction -> vuSetModifiedPtr (&this -> dialogModified);
    miscDialog -> func01 -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> func02 -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> func03 -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> func04 -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> limit1Low -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> limit2Low -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> limit3Low -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> limit1High -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> limit2High -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> limit3High -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> limit1Precision -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> limit2Precision -> SetTextModifiedPtr (&this -> dialogModified);
    miscDialog -> limit3Precision -> SetTextModifiedPtr (&this -> dialogModified);

}


/*******************************************************************
*
*    vuUdbDialogClass Destroy function
*
*******************************************************************/

void vuUdbDialogClass::Destroy (void)
{
    vstrDestroy (udbTitle);
    // delete the other UDB dialogs
    delete statusDialog;
    delete ioDialog;
    delete constantsDialog;
    delete limitsDialog;
    delete miscDialog;
    // delete the dialogLabels list
    if (dialogLabels != NULL)
    {
	vlistIterator listIterator;
	VU_UDBLABEL_PTR labelPtr;
	listIterator.Start (dialogLabels);
	while (listIterator.Next ())
	{
	    labelPtr = (VU_UDBLABEL_PTR)(listIterator.GetValue ());
	    vmemFree ((void *)labelPtr);
	}
	listIterator.Finish ();

	delete dialogLabels;
    }
    vuCommonDialogClass::Destroy ();
}


/*******************************************************************
*
*    vuUdbDialogClass add dialog item to labels list
*
*******************************************************************/

void vuUdbDialogClass::vuUdbAddLabel (vdialogItem *dlgItem, char *widgetId)
{
    VU_UDBLABEL_PTR	newLabel;
    vchar		newTitle[UDBLABEL_SZ + 1];
    vchar		*ptr;

    newLabel = (VU_UDBLABEL_PTR)(vmemAllocAndClear (sizeof (struct udbLabels)));
    newLabel -> dialogItem = dlgItem;
    vcharCopyBounded (dlgItem -> GetTitle (), newTitle, UDBLABEL_SZ);
    newTitle[UDBLABEL_SZ] = 0;
    ptr = newTitle;
    // get past any leading spaces
    while (*ptr)
    {
	if ((*ptr) == ' ')
	    ptr++;
	else
	    break;
    }
    strncpy (newLabel -> widgetId, (const char *)widgetId, UDBLABEL_SZ);
    newLabel -> widgetId[UDBLABEL_SZ] = 0;
    vcharCopyBounded (ptr, newLabel -> defaultLabel, UDBLABEL_SZ);
    newLabel -> defaultLabel[UDBLABEL_SZ] = 0;
    vcharCopyBounded (ptr, newLabel -> newLabel, UDBLABEL_SZ);
    newLabel -> newLabel[UDBLABEL_SZ] = 0;
    int cellCount = dialogLabels -> GetRowCount ();
    dialogLabels -> SetCellValue (cellCount, 0, (const void *)newLabel);
}


/*******************************************************************
*
*    vuUdbDialogClass modify dialog labels function
*	see .../share/src/dataVu/udbcfg.h for field defs
*	NOTE: the dialogLabels list is in the same order as are
*		fields in the udbcfg.dbf record
*
*******************************************************************/

void vuUdbDialogClass::vuUdbModLabels (void)
{
    vstr	*udbPgName;

    if (udbLabelsModified == vTRUE)
	return;

    // get the udb program name from the dialog
    udbPgName = udbContainer -> udbProgramName -> GetTextAsString ();
    if ( (vcharLength (udbPgName)) > 0)
    {
	vuMainDialogClass *mainDialog = getMainDialog ();
	dataVuApplication *vuApp=  mainDialog -> getApplication ();
	vuDatabaseClass   *db = vuApp -> getDatabaseClass ();
	char		pgName [UDBPROGRAMNAME_SZ + 1];
	DB_RECORD_PTR	rec;

	vcharExportBounded (vcharScribeSystemExport (udbPgName),
			    (void *)pgName, UDBPROGRAMNAME_SZ);

	// if this program name exists in the udb config database file
	if ( (rec = DB_Seek (db -> vuUdbCfgTable, pgName)) != 0)
	{
	    vlistIterator	iterator;
	    int			row;
	    int			type;
	    VU_UDBLABEL_PTR	listValue;
	    char		labelName [UDBLABEL_SZ + 1];
	    char		newLabel [UDBLABEL_SZ + 1];
	    char		fieldName [12];
	    DB_FIELD_PTR	field;

	    // given the label from the file, mod the label in the dialog
	    iterator.Start (dialogLabels);
	    while (iterator.Next())
	    {
		row = iterator.GetRow ();
		listValue = (VU_UDBLABEL_PTR)(iterator.GetValue ());
		// list rows indexed from 0, field name LABEL1 -> LABEL72
		sprintf (fieldName, "LABEL%d\0", row + 1);
		field = DB_MapField (db -> vuUdbCfgTable, fieldName);
		DB_GetFieldValue (rec, field, labelName, &type);
		if (strlen (labelName))
		{
		    sprintf (newLabel, "%*.*s\0", UDBLABEL_SZ, UDBLABEL_SZ,
						  labelName);
		    listValue -> dialogItem -> 
				SetTitleScribed (vcharScribeSystem (newLabel));
		}
	    }
	    iterator.Finish ();
	    // set flag so resetDialog() will know to reset the dialog
	    // labels to their default values
	    udbLabelsModified = vTRUE;
	}
    }
    vstrDestroy (udbPgName);
}


/*******************************************************************
*
*    vuUdbDialogClass, reset fields, overrides base class 
*	vuCommonDialogClass
*
*******************************************************************/

void vuUdbDialogClass::resetDialog (void)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    vchar		epics[FLDNAME_SZ + 1];
    int			firstTimeFlag;
    int			inDialogFlag;

    vresource defRes = mainDialog -> getDefaultsResource ();

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0; 
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	// if not a common dialog field
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    firstTimeFlag = xrefNode -> getFirstTimeFlag ();
	    // only need to do this once
	    if (firstTimeFlag == vTRUE)
	    {
		xrefNode -> setFirstTimeFlag (vFALSE);
		// find which dialog this field is in
		dialogItem = vuFindContainerItem (udbContainer, epics);
		if (dialogItem == NULL)
		{
		    dialogItem = vuFindDialogItem (statusDialog, epics);
		    if (dialogItem == NULL)
		    {
			dialogItem = vuFindDialogItem (ioDialog, epics);
			if (dialogItem == NULL)
			{
			    dialogItem = vuFindDialogItem (constantsDialog, epics);
			    if (dialogItem == NULL)
			    {
				dialogItem = vuFindDialogItem (limitsDialog, epics);
				if (dialogItem == NULL)
				{
				    dialogItem = vuFindDialogItem (miscDialog, epics);
				    if (dialogItem == NULL)
				    {
					xrefNode -> setInDialogFlag (VU_UDB_NOT_IN_DIALOG);
					continue;
				    }
				    else
				    {
					xrefNode -> setInDialogFlag (VU_UDB_IN_MISC_DIALOG);
				    }
				}
				else
				{
				    xrefNode -> setInDialogFlag (VU_UDB_IN_LIMITS_DIALOG);
				}
			    }
			    else
			    {
				xrefNode -> setInDialogFlag (VU_UDB_IN_CONSTANTS_DIALOG);
			    }
			}
			else
			{
			    xrefNode -> setInDialogFlag (VU_UDB_IN_IO_DIALOG);
			}
		    }
		    else
		    {
			xrefNode -> setInDialogFlag (VU_UDB_IN_STATUS_DIALOG);
		    }
		}
		else
		{
		    xrefNode -> setInDialogFlag (VU_UDB_IN_MAIN_DIALOG);
		}
	    }
	    else
	    {
		inDialogFlag = xrefNode -> getInDialogFlag ();
		dialogItem = NULL;
		switch (inDialogFlag)
		{
		    case VU_UDB_NOT_IN_DIALOG:
			break;

		    case VU_UDB_IN_MAIN_DIALOG:
			dialogItem = vuFindContainerItem (udbContainer, epics);
			break;

		    case VU_UDB_IN_STATUS_DIALOG:
			dialogItem = vuFindDialogItem (statusDialog, epics);
			break;

		    case VU_UDB_IN_IO_DIALOG:
			dialogItem = vuFindDialogItem (ioDialog, epics);
			break;

		    case VU_UDB_IN_CONSTANTS_DIALOG:
			dialogItem = vuFindDialogItem (constantsDialog, epics);
			break;

		    case VU_UDB_IN_LIMITS_DIALOG:
			dialogItem = vuFindDialogItem (limitsDialog, epics);
			break;

		    case VU_UDB_IN_MISC_DIALOG:
			dialogItem = vuFindDialogItem (miscDialog, epics);
			break;

		    default:
			break;
		}
	    }
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    if ( (dialogItem -> IsKindOf (&vtextitem::Kind)) == vTRUE)
	    {
		textItem = (vtextitem *)dialogItem;
		textItem -> SetTextScribed (vcharScribeSystem (""));
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolToggle::Kind)) == vTRUE)
	    {
		toggleItem = (vcontrolToggle *)dialogItem;
		toggleItem -> SetValue (0);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolOptionMenu::Kind)) == vTRUE)
	    {
		optionMenu = (vcontrolOptionMenu *)dialogItem;
		optionMenu -> SetValue (0);
	    }
	}  // end if not in common dialog
    }
    iterator.Finish ();

    if (udbLabelsModified)
    {
	// reset labels to defaults
	VU_UDBLABEL_PTR label;
	vlistIterator	listIterator;
	listIterator.Start (dialogLabels);
	while (listIterator.Next ())
	{
	    label = (VU_UDBLABEL_PTR)(listIterator.GetValue ());
	    label -> dialogItem -> SetTitleScribed (vcharScribeFormatLiteral (
					"%*.*s\0", 
					UDBLABEL_SZ, UDBLABEL_SZ,
					label -> defaultLabel));

	}
	listIterator.Finish ();
    }
    udbLabelsModified = vFALSE;

    // call base class reset function with proper point index
    vuCommonDialogClass::resetDialog ();

    dialogModified = vFALSE;
}


/*******************************************************************
*
*    vuLoadDialog 
*
*******************************************************************/

void vuUdbDialogClass::vuLoadDialog (vchar *pointName)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    vchar		epics[FLDNAME_SZ + 1];
    vstr		*finalValue;

    char		pvName [PVNAME_SZ + 1];
    DB_TABLE_PTR	lastTable;	// pointer to last table used
    DB_RECORD_PTR	pntRecord;	// codebase record ptr
    int			dbaseType;	// codebase field type
    int			intValue;
    unsigned int	uIntValue;
    long		longValue;
    unsigned long	uLongValue;
    double		doubleValue;
    int			inDialogFlag;

    vcharExportBounded (vcharScribeSystemExport (pointName),
			pvName,
			PVNAME_SZ);

    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
    pntRecord = DB_Seek (db -> vuPvmapTable, seekBuff);
    if (pntRecord == (DB_RECORD_PTR)0)
    {
	return;
    }
    else
    {
	int ptType;
	DB_GetFieldValue (pntRecord, db -> pvmapTypeField, dbValue, &dbaseType);
	ptType = vuGetPointTypeIndex ((char *)dbValue);
	// the point name in pvname field is NOT an ACCUM
	if (ptType != VU_UDB)
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_PTTYPEDIALOGERR,
				   pointName, ptType, VU_UDB);
	    vuClearPvNameField ();
	    return;
	}
	// the point name entered in pvname field has been deleted
	if (DB_IsDeleted (db -> vuPvmapTable))
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_PTISDELETEDERR, pointName);
	    vuClearPvNameField ();
	    return;
	}
    }
    // call base class vuLoadDialog first
    vuCommonDialogClass::vuLoadDialog (pointName);

    lastTable = NULL;

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0;
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    // only seek if table has changed
	    if (lastTable != xrefNode -> xrefTable)
		{
		    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
			pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
		}
	    lastTable = xrefNode -> xrefTable;

	    inDialogFlag = xrefNode -> getInDialogFlag ();
	    dialogItem = NULL;
	    switch (inDialogFlag)
	    {
		case VU_UDB_NOT_IN_DIALOG:
		    break;

		case VU_UDB_IN_MAIN_DIALOG:
		    dialogItem = vuFindContainerItem (udbContainer, epics);
		    break;

		case VU_UDB_IN_STATUS_DIALOG:
		    dialogItem = vuFindDialogItem (statusDialog, epics);
		    break;

		case VU_UDB_IN_IO_DIALOG:
		    dialogItem = vuFindDialogItem (ioDialog, epics);
		    break;

		case VU_UDB_IN_CONSTANTS_DIALOG:
		    dialogItem = vuFindDialogItem (constantsDialog, epics);
		    break;

		case VU_UDB_IN_LIMITS_DIALOG:
		    dialogItem = vuFindDialogItem (limitsDialog, epics);
		    break;

		case VU_UDB_IN_MISC_DIALOG:
		    dialogItem = vuFindDialogItem (miscDialog, epics);
		    break;

		default:
		    break;
	    }
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    DB_GetFieldValue (pntRecord, xrefNode -> xrefField, dbValue, &dbaseType);
	    doubleValue = 0.0;
	    // GetField actually returns a number in the char array
	    if ( (dbaseType == 'F') || (dbaseType == 'N') )
	    {
		doubleValue = *((double *)dbValue);
	    }
	    switch (xrefNode -> xrefEPICSFieldType)
	    {
		case DBF_SHORT:
		    intValue = (int)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeInt (intValue));
		    break;

		case DBF_LONG:
		    longValue = (long)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeLong (longValue));
		    break;

		case DBF_ENUM:
		case DBF_USHORT:
		    uIntValue = (unsigned int)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeUInt (uIntValue));
		    break;

		case DBF_ULONG:
		    uLongValue = (unsigned long)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeULong (uLongValue));
		    break;

		case DBF_FLOAT:
		case DBF_DOUBLE:
		    finalValue = vstrCloneScribed (vnumScribeFloat (doubleValue));
		    break;

		case DBF_STRING:
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		case DBF_GBLCHOICE:
		case DBF_RECCHOICE:
		case DBF_CVTCHOICE:
		case DBF_DEVCHOICE:
		    if (dbaseType == 'L')
		    {
			int	logical;

			// special case these are defined as logicals in the
			// database but as choice menus in SDR
			// NOTE: GetFieldValue return 0/1 for False/True
			if ( ((strncmp ((const char *)epics, "CK1 ", FLDNAME_SZ)) == 0) ||
			     ((strncmp ((const char *)epics, "CK2 ", FLDNAME_SZ)) == 0) ||
			     ((strncmp ((const char *)epics, "CK3 ", FLDNAME_SZ)) == 0) ||
			     ((strncmp ((const char *)epics, "CK4 ", FLDNAME_SZ)) == 0) ||
			     ((strncmp ((const char *)epics, "CK5 ", FLDNAME_SZ)) == 0) ||
			     ((strncmp ((const char *)epics, "CK6 ", FLDNAME_SZ)) == 0) ||
			     ((strncmp ((const char *)epics, "CK7 ", FLDNAME_SZ)) == 0) ||
			     ((strncmp ((const char *)epics, "CK8 ", FLDNAME_SZ)) == 0) )
			{
			    logical = *((char *)dbValue) & 1;
			    strcpy ((char *)dbValue, (logical == 1 ? "YES" : "NO"));
			    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
			}
			else
			{

			    logical = (*((char*)dbValue) ? 
					xrefNode -> xrefTrue : 
					(xrefNode -> xrefTrue ^1));
			    finalValue = vstrCloneScribed (vnumScribeInt (logical));
			}
			break;
		    }
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		case DBF_INLINK:
		case DBF_OUTLINK:
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		default:
		    continue;
	    }
	    if ( (dialogItem -> IsKindOf (&vtextitem::Kind)) == vTRUE)
	    {
		textItem = (vtextitem *)dialogItem;
		textItem -> SetTextScribed (vcharScribeSystem (""));
		textItem -> SetText (finalValue);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolToggle::Kind)) == vTRUE)
	    {
		toggleItem = (vcontrolToggle *)dialogItem;
		if (finalValue[0] == '0')
		    toggleItem -> SetValue (0);
		else
		    toggleItem -> SetValue (1);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolOptionMenu::Kind)) == vTRUE)
	    {
		optionMenu = (vcontrolOptionMenu *)dialogItem;
		vuSetOptionMenuValue (optionMenu, finalValue);
		vstrDestroy (finalValue);
	    }
	}  // end if not in commondialog
    }
    iterator.Finish ();

    setDialogModify (vFALSE);
}


/*******************************************************************
*
*    vuSaveDialog 
*
*******************************************************************/

void vuUdbDialogClass::vuSaveDialog (vchar *pointName)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    const vchar 	*optionMenuValue;
    vchar		epics[FLDNAME_SZ + 1];
    vstr		*stringValue;

    char		pvName [PVNAME_SZ + 1];
    DB_FIELD_PTR	putField;
    DB_TABLE_PTR	lastTable;	// pointer to last table used
    DB_RECORD_PTR	pntRecord;	// codebase record ptr
    double		doubleValue;
    int			doSave;
    int			inDialogFlag;

    vcharExportBounded (vcharScribeSystemExport (pointName),
			pvName,
			PVNAME_SZ);

    lastTable = NULL;

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0;
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    doSave = vTRUE;
	    // only seek if table has changed
	    if (lastTable != xrefNode -> xrefTable)
		{
		    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
			pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
		}
	    lastTable = xrefNode -> xrefTable;

	    putField = xrefNode -> xrefField;

	    inDialogFlag = xrefNode -> getInDialogFlag ();
	    dialogItem = NULL;
	    switch (inDialogFlag)
	    {
		case VU_UDB_NOT_IN_DIALOG:
		    break;

		case VU_UDB_IN_MAIN_DIALOG:
		    dialogItem = vuFindContainerItem (udbContainer, epics);
		    break;

		case VU_UDB_IN_STATUS_DIALOG:
		    dialogItem = vuFindDialogItem (statusDialog, epics);
		    break;

		case VU_UDB_IN_IO_DIALOG:
		    dialogItem = vuFindDialogItem (ioDialog, epics);
		    break;

		case VU_UDB_IN_CONSTANTS_DIALOG:
		    dialogItem = vuFindDialogItem (constantsDialog, epics);
		    break;

		case VU_UDB_IN_LIMITS_DIALOG:
		    dialogItem = vuFindDialogItem (limitsDialog, epics);
		    break;

		case VU_UDB_IN_MISC_DIALOG:
		    dialogItem = vuFindDialogItem (miscDialog, epics);
		    break;

		default:
		    break;
	    }
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    switch (xrefNode -> xrefEPICSFieldType)
	    {
		case DBF_SHORT:
		case DBF_LONG:
		case DBF_ENUM:
		case DBF_USHORT:
		case DBF_ULONG:
		case DBF_FLOAT:
		case DBF_DOUBLE:
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    doubleValue = atof ((const char *)dbValue);
		    DB_PutFieldValue (pntRecord, putField, (void *)&doubleValue);
		    doSave = vFALSE;
		    break;

		case DBF_STRING:
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    break;

		case DBF_GBLCHOICE:
		case DBF_RECCHOICE:
		case DBF_CVTCHOICE:
		case DBF_DEVCHOICE:
		{
		    if (xrefNode -> xrefField -> field_type == 'L')
		    {
			int	logical;

			// special case these are defined as logicals in the
			// database but as choice menus in SDR
			if ( ((strncmp ((const char *)epics, "CK1 ", FLDNAME_SZ)) == 0) ||
			     ((strncmp ((const char *)epics, "CK2 ", FLDNAME_SZ)) == 0) ||
			     ((strncmp ((const char *)epics, "CK3 ", FLDNAME_SZ)) == 0) ||
			     ((strncmp ((const char *)epics, "CK4 ", FLDNAME_SZ)) == 0) ||
			     ((strncmp ((const char *)epics, "CK5 ", FLDNAME_SZ)) == 0) ||
			     ((strncmp ((const char *)epics, "CK6 ", FLDNAME_SZ)) == 0) ||
			     ((strncmp ((const char *)epics, "CK7 ", FLDNAME_SZ)) == 0) ||
			     ((strncmp ((const char *)epics, "CK8 ", FLDNAME_SZ)) == 0) )
			{
			    optionMenu = (vcontrolOptionMenu *)dialogItem;
			    optionMenuValue = vuGetOptionMenuValue (optionMenu);
			    *((char *)dbValue) = (optionMenuValue[0] == 'Y') ? 1 : 0;
			}
			else
			{
			    toggleItem = (vcontrolToggle *)dialogItem;
			    logical = toggleItem -> GetValue ();
			    *((char *)dbValue) = (logical ? xrefNode -> xrefTrue : 
								xrefNode -> xrefTrue ^1);
			}
			break;
		    }
		    optionMenu = (vcontrolOptionMenu *)dialogItem;
		    optionMenuValue = vuGetOptionMenuValue (optionMenu);
		    vcharExport (vcharScribeSystemExport (optionMenuValue), dbValue);
		    break;
		}

		case DBF_INLINK:
		case DBF_OUTLINK:
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    break;

		default:
		    continue;

	    }  // end xrefNode -> xrefEPICSFieldType

	    if (doSave == vTRUE)
	    {
		DB_PutFieldValue (pntRecord, putField, dbValue);
	    }

	}  // end if not in commondialog

    }  // end while iterator

    iterator.Finish ();

    // call base class vuSaveDialog
    vuCommonDialogClass::vuSaveDialog (pointName);
}


/*******************************************************************
*
*    vuUdbContainerClass
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbContainerClass, vcontainer, 
						"vuUdbContainerClass");

/*******************************************************************
*
*    vuUdbContainerClass override for constructor
*
*******************************************************************/

void vuUdbContainerClass::LoadInit (vresource res)
{
    vapplication *app = vapplication::GetCurrent ();
    vresource pointDictRes = vresourceGet (app -> GetResources (), 
				vunames_vuPointDictionary);
    vresource realRes = vresourceGet (pointDictRes, vunames_vuUdbContainer);


    vcontainer *dnamcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbDNAMContainer);
    udbProgramNameButton = vuUdbProgramNameButtonClass::RequestSubLoad (dnamcontainer,
				vunames_vuUdbDNAMButton);
    udbProgramName = vuFilteredTextitem::RequestSubLoad (dnamcontainer,
				vunames_DNAM);

    onScan = vuCheckBoxClass::RequestSubLoad (this,
				vunames_OSCN);

    onAlarm = vuCheckBoxClass::RequestSubLoad (this,
				vunames_AENB);

    statusButton = vuUdbStatusButtonClass::RequestSubLoad (this,
				vunames_vuUdbStatusButton);

    ioButton = vuUdbIOButtonClass::RequestSubLoad (this,
				vunames_vuUdbIOButton);

    constantsButton = vuUdbConstantsButtonClass::RequestSubLoad (this,
				vunames_vuUdbConstantButton);

    limitsButton = vuUdbLimitsButtonClass::RequestSubLoad (this,
				vunames_vuUdbLimitsButton);

    miscButton = vuUdbMiscButtonClass::RequestSubLoad (this,
				vunames_vuUdbMiscButton);

    vcontainer::LoadInit (realRes);

    // set filters

    udbProgramName -> SetItemFilterFunc (vuUdbProgramNameSpecialFilter);
    filteredText *nameText = (filteredText *)udbProgramName -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    nameText -> SetUserPtr ((void *)udbProgramName);

    // store ptr to udbProgramName in udbProgramNameButton
    udbProgramNameButton -> storeProgramName (udbProgramName);

}

/*******************************************************************
*
*    vuUdbContainerClass override for destructor
*
*******************************************************************/

void vuUdbContainerClass::Destroy (void)
{
    vcontainer::Destroy ();
}


/******************************************************************
*
*    vuUdbProgramNameButtonClass observer
*	open the udb config dialog and wait here until finished
*	the open will wait before returning here
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbProgramNameButtonClass, vbutton, 
					"vuUdbProgramNameButtonClass");

void vuUdbProgramNameButtonClass::ObserveDialogItem (vevent *event)
{
    vuMainDialogClass *mainDialog = udbDialog -> getMainDialog ();
    dataVuApplication *vuApp = mainDialog -> getApplication ();
    vresource vuRes = vuApp -> getDataVuvresource ();
    vresource vuUdbConfigRes = vresourceGet (vuRes, vunames_vuUdbConfigDialog);
    vuUdbConfigDialogClass *configDialog = 
					new vuUdbConfigDialogClass(vuUdbConfigRes);
    configDialog -> vuUdbConfigDialogOpen (vuApp, programNameText, udbDialog);
    delete configDialog;
}


/******************************************************************
*
*    vuUdbStatusButtonClass observer
*	open the status dialog and wait here until finished
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbStatusButtonClass, vbutton, 
					"vuUdbStatusButtonClass");

void vuUdbStatusButtonClass::ObserveDialogItem (vevent *event)
{
    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    statusDialogClass -> Place (vwindow::GetRoot (), vrectPLACE_CENTER, 
				vrectPLACE_CENTER);
    statusDialogClass -> Open();
    MakeTopMostDialog (statusDialogClass, NULL);
    vevent::Process ();
}


/******************************************************************
*
*    vuUdbIOButtonClass observer
*	open the IO dialog and wait here until finished
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbIOButtonClass, vbutton, 
					"vuUdbIOButtonClass");

void vuUdbIOButtonClass::ObserveDialogItem (vevent *event)
{
    udbDialogClass -> vuUdbModLabels ();

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    ioDialogClass -> Place (vwindow::GetRoot (), vrectPLACE_CENTER, 
			    vrectPLACE_CENTER);
    ioDialogClass -> Open();
    MakeTopMostDialog (ioDialogClass, NULL);
    vevent::Process ();
}


/******************************************************************
*
*    vuUdbConstantsButtonClass observer
*	open the constants dialog and wait here until finished
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbConstantsButtonClass, vbutton, 
					"vuUdbConstantsButtonClass");

void vuUdbConstantsButtonClass::ObserveDialogItem (vevent *event)
{
    udbDialogClass -> vuUdbModLabels ();

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    constantsDialogClass -> Place (vwindow::GetRoot (), vrectPLACE_CENTER, 
				   vrectPLACE_CENTER);
    constantsDialogClass -> Open();
    MakeTopMostDialog (constantsDialogClass, NULL);
    vevent::Process ();
}


/******************************************************************
*
*    vuUdbLimitsButtonClass observer
*	open the limits dialog and wait here until finished
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbLimitsButtonClass, vbutton, 
					"vuUdbLimitsButtonClass");

void vuUdbLimitsButtonClass::ObserveDialogItem (vevent *event)
{
    udbDialogClass -> vuUdbModLabels ();

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    limitsDialogClass -> Place (vwindow::GetRoot (), vrectPLACE_CENTER, 
				vrectPLACE_CENTER);
    limitsDialogClass -> Open();
    MakeTopMostDialog (limitsDialogClass, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    vuUdbMiscButtonClass observer
*	open the miscellaneous dialog and wait here until finished
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbMiscButtonClass, vbutton, 
					"vuUdbMiscButtonClass");

void vuUdbMiscButtonClass::ObserveDialogItem (vevent *event)
{
    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    miscDialogClass -> Place (vwindow::GetRoot (), vrectPLACE_CENTER, 
			      vrectPLACE_CENTER);
    miscDialogClass -> Open();
    MakeTopMostDialog (miscDialogClass, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    vuUdbStatusDialogClass
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbStatusDialogClass, vdialog, 
					"vuUdbStatusDialogClass");

/*******************************************************************
*
*    vuUdbStatusDialogClass override for constructors init ()
*
*******************************************************************/

void vuUdbStatusDialogClass::LoadInit (vresource res)
{

    vcontainer *ss01container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSS01Container);
    status1 = filteredTextitem::RequestSubLoad (ss01container,
				vunames_SS01);

    vcontainer *ss02container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSS02Container);
    status2 = filteredTextitem::RequestSubLoad (ss02container,
				vunames_SS02);

    vcontainer *ss03container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSS03Container);
    status3 = filteredTextitem::RequestSubLoad (ss03container,
				vunames_SS03);

    vcontainer *ss04container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSS04Container);
    status4 = filteredTextitem::RequestSubLoad (ss04container,
				vunames_SS04);

    vcontainer *ss05container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSS05Container);
    status5 = filteredTextitem::RequestSubLoad (ss05container,
				vunames_SS05);

    vcontainer *ss06container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSS06Container);
    status6 = filteredTextitem::RequestSubLoad (ss06container,
				vunames_SS06);

    vcontainer *ss07container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSS07Container);
    status7 = filteredTextitem::RequestSubLoad (ss07container,
				vunames_SS07);

    vcontainer *ss08container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSS08Container);
    status8 = filteredTextitem::RequestSubLoad (ss08container,
				vunames_SS08);

    vcontainer *ss09container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSS09Container);
    status9 = filteredTextitem::RequestSubLoad (ss09container,
				vunames_SS09);

    vcontainer *ss10container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSS10Container);
    status10 = filteredTextitem::RequestSubLoad (ss10container,
				vunames_SS10);

    vcontainer *ss11container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSS11Container);
    status11 = filteredTextitem::RequestSubLoad (ss11container,
				vunames_SS11);

    vcontainer *ss12container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSS12Container);
    status12 = filteredTextitem::RequestSubLoad (ss12container,
				vunames_SS12);

    vcontainer *sv01container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSV01Container);
    status1Severity = vuChoiceGblMenuClass::RequestSubLoad (sv01container,
				vunames_SV01);

    vcontainer *sv02container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSV02Container);
    status2Severity = vuChoiceGblMenuClass::RequestSubLoad (sv02container,
				vunames_SV02);

    vcontainer *sv03container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSV03Container);
    status3Severity = vuChoiceGblMenuClass::RequestSubLoad (sv03container,
				vunames_SV03);

    vcontainer *sv04container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSV04Container);
    status4Severity = vuChoiceGblMenuClass::RequestSubLoad (sv04container,
				vunames_SV04);

    vcontainer *sv05container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSV05Container);
    status5Severity = vuChoiceGblMenuClass::RequestSubLoad (sv05container,
				vunames_SV05);

    vcontainer *sv06container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSV06Container);
    status6Severity = vuChoiceGblMenuClass::RequestSubLoad (sv06container,
				vunames_SV06);

    vcontainer *sv07container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSV07Container);
    status7Severity = vuChoiceGblMenuClass::RequestSubLoad (sv07container,
				vunames_SV07);

    vcontainer *sv08container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSV08Container);
    status8Severity = vuChoiceGblMenuClass::RequestSubLoad (sv08container,
				vunames_SV08);

    vcontainer *sv09container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSV09Container);
    status9Severity = vuChoiceGblMenuClass::RequestSubLoad (sv09container,
				vunames_SV09);

    vcontainer *sv10container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSV10Container);
    status10Severity = vuChoiceGblMenuClass::RequestSubLoad (sv10container,
				vunames_SV10);

    vcontainer *sv11container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSV11Container);
    status11Severity = vuChoiceGblMenuClass::RequestSubLoad (sv11container,
				vunames_SV11);

    vcontainer *sv12container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbSV12Container);
    status12Severity = vuChoiceGblMenuClass::RequestSubLoad (sv12container,
				vunames_SV12);

    vcontainer *cst1container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCST1Container);
    cmd1 = filteredTextitem::RequestSubLoad (cst1container,
				vunames_CST1);

    vcontainer *cst2container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCST2Container);
    cmd2 = filteredTextitem::RequestSubLoad (cst2container,
				vunames_CST2);

    vcontainer *cst3container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCST3Container);
    cmd3 = filteredTextitem::RequestSubLoad (cst3container,
				vunames_CST3);

    vcontainer *cst4container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCST4Container);
    cmd4 = filteredTextitem::RequestSubLoad (cst4container,
				vunames_CST4);

    vcontainer *cst5container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCST5Container);
    cmd5 = filteredTextitem::RequestSubLoad (cst5container,
				vunames_CST5);

    vcontainer *cst6container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCST6Container);
    cmd6 = filteredTextitem::RequestSubLoad (cst6container,
				vunames_CST6);

    vcontainer *cst7container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCST7Container);
    cmd7 = filteredTextitem::RequestSubLoad (cst7container,
				vunames_CST7);

    vcontainer *cst8container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCST8Container);
    cmd8 = filteredTextitem::RequestSubLoad (cst8container,
				vunames_CST8);

    vcontainer *cosvcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCOSVContainer);
    cosvSeverity = vuChoiceGblMenuClass::RequestSubLoad (cosvcontainer,
				vunames_COSV);

    vcontainer *unsvcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbUNSVContainer);
    undefinedSeverity = vuChoiceGblMenuClass::RequestSubLoad (unsvcontainer,
				vunames_UNSV);

    vcontainer *algpcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbALGPContainer);
    alarmGroup = vuFilteredTextitem::RequestSubLoad (algpcontainer,
				vunames_ALGP);

    vcontainer *aldlcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbALDLContainer);
    alarmDelay = filteredTextitem::RequestSubLoad (aldlcontainer,
				vunames_ALDL);

    vcontainer *adiscontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbADISContainer);
    alarmDisableButton = vuPvNameButtonClass::RequestSubLoad (adiscontainer,
				vunames_vuUdbADISButton);
    alarmDisable = vuFilteredTextitem::RequestSubLoad (adiscontainer,
				vunames_ADIS);

    dismissButton = vuUdbStatusDismissButtonClass::RequestSubLoad (this,
				vunames_vuUdbStatusDialogDismissButton);

    vdialog::LoadInit (res);

    // store ptr to text field for vuPvNameButtonClass
    alarmDisableButton -> storePvName (alarmDisable);

    // set filters
    status1 -> SetItemFilterFunc (udbStringsFilter);
    status2 -> SetItemFilterFunc (udbStringsFilter);
    status3 -> SetItemFilterFunc (udbStringsFilter);
    status4 -> SetItemFilterFunc (udbStringsFilter);
    status5 -> SetItemFilterFunc (udbStringsFilter);
    status6 -> SetItemFilterFunc (udbStringsFilter);
    status7 -> SetItemFilterFunc (udbStringsFilter);
    status8 -> SetItemFilterFunc (udbStringsFilter);
    status9 -> SetItemFilterFunc (udbStringsFilter);
    status10 -> SetItemFilterFunc (udbStringsFilter);
    status11 -> SetItemFilterFunc (udbStringsFilter);
    status12 -> SetItemFilterFunc (udbStringsFilter);
    cmd1 -> SetItemFilterFunc (udbStringsFilter);
    cmd2 -> SetItemFilterFunc (udbStringsFilter);
    cmd3 -> SetItemFilterFunc (udbStringsFilter);
    cmd4 -> SetItemFilterFunc (udbStringsFilter);
    cmd5 -> SetItemFilterFunc (udbStringsFilter);
    cmd6 -> SetItemFilterFunc (udbStringsFilter);
    cmd7 -> SetItemFilterFunc (udbStringsFilter);
    cmd8 -> SetItemFilterFunc (udbStringsFilter);
    alarmDelay -> SetItemFilterFunc (numericFilter);

    alarmDisable -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *adisText = (filteredText *)alarmDisable -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    adisText -> SetUserPtr ((void *)alarmDisable);

    alarmGroup -> SetItemFilterFunc (vuAlarmGroupFilter);
    filteredText *alarmText = (filteredText *)alarmGroup -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    alarmText -> SetUserPtr ((void *)alarmGroup);

    // store dialog pointer in dismiss button
    dismissButton -> storeStatusDialog (this);

    // build popup menus
    status1Severity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    status2Severity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    status3Severity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    status4Severity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    status5Severity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    status6Severity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    status7Severity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    status8Severity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    status9Severity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    status10Severity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    status11Severity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    status12Severity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    cosvSeverity -> vuBuildOptionMenu (GBL_ALARM_SEV);
    undefinedSeverity -> vuBuildOptionMenu (GBL_ALARM_SEV);

}


/*******************************************************************
*
*    vuUdbStatusDialogClass notify function
*
*******************************************************************/

void vuUdbStatusDialogClass::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose ())
    {
	Close ();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuUdbStatusDismissButtonClass notify function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbStatusDismissButtonClass, vbutton, 
					"vuUdbStatusDismissButtonClass");

void vuUdbStatusDismissButtonClass::ObserveDialogItem (vevent *event)
{
    statusDialogClass -> Close ();
    vevent::StopProcessing ();
}



/*******************************************************************
*
*    vuUdbIODialogClass
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbIODialogClass, vdialog, 
					"vuUdbIODialogClass");

/*******************************************************************
*
*    vuUdbIODialogClass override for constructors init ()
*
*******************************************************************/

void vuUdbIODialogClass::LoadInit (vresource res)
{

    vcontainer *inp1container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbINP1Container);
    input1Button = vuPvNameButtonClass::RequestSubLoad (inp1container,
				vunames_vuUdbINP1Button);
    input1 = vuFilteredTextitem::RequestSubLoad (inp1container,
				vunames_INP1);

    prc1 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PRC1);

    mxs1 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_MXS1);

    vcontainer *inp2container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbINP2Container);
    input2Button = vuPvNameButtonClass::RequestSubLoad (inp2container,
				vunames_vuUdbINP2Button);
    input2 = vuFilteredTextitem::RequestSubLoad (inp2container,
				vunames_INP2);

    prc2 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PRC2);

    mxs2 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_MXS2);

    vcontainer *inp3container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbINP3Container);
    input3Button = vuPvNameButtonClass::RequestSubLoad (inp3container,
				vunames_vuUdbINP3Button);
    input3 = vuFilteredTextitem::RequestSubLoad (inp3container,
				vunames_INP3);

    prc3 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PRC3);

    mxs3 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_MXS3);

    vcontainer *inp4container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbINP4Container);
    input4Button = vuPvNameButtonClass::RequestSubLoad (inp4container,
				vunames_vuUdbINP4Button);
    input4 = vuFilteredTextitem::RequestSubLoad (inp4container,
				vunames_INP4);

    prc4 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PRC4);

    mxs4 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_MXS4);

    vcontainer *inp5container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbINP5Container);
    input5Button = vuPvNameButtonClass::RequestSubLoad (inp5container,
				vunames_vuUdbINP5Button);
    input5 = vuFilteredTextitem::RequestSubLoad (inp5container,
				vunames_INP5);

    prc5 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PRC5);

    mxs5 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_MXS5);

    vcontainer *inp6container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbINP6Container);
    input6Button = vuPvNameButtonClass::RequestSubLoad (inp6container,
				vunames_vuUdbINP6Button);
    input6 = vuFilteredTextitem::RequestSubLoad (inp6container,
				vunames_INP6);

    prc6 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PRC6);

    mxs6 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_MXS6);

    vcontainer *inp7container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbINP7Container);
    input7Button = vuPvNameButtonClass::RequestSubLoad (inp7container,
				vunames_vuUdbINP7Button);
    input7 = vuFilteredTextitem::RequestSubLoad (inp7container,
				vunames_INP7);

    prc7 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PRC7);

    mxs7 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_MXS7);

    vcontainer *inp8container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbINP8Container);
    input8Button = vuPvNameButtonClass::RequestSubLoad (inp8container,
				vunames_vuUdbINP8Button);
    input8 = vuFilteredTextitem::RequestSubLoad (inp8container,
				vunames_INP8);

    prc8 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PRC8);

    mxs8 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_MXS8);

    vcontainer *out1container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbOUT1Container);
    output1Button = vuPvNameButtonClass::RequestSubLoad (out1container,
				vunames_vuUdbOUT1Button);
    output1 = vuFilteredTextitem::RequestSubLoad (out1container,
				vunames_OUT1);

    pin1 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PIN1);

    vcontainer *out2container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbOUT2Container);
    output2Button = vuPvNameButtonClass::RequestSubLoad (out2container,
				vunames_vuUdbOUT2Button);
    output2 = vuFilteredTextitem::RequestSubLoad (out2container,
				vunames_OUT2);

    pin2 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PIN2);

    vcontainer *out3container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbOUT3Container);
    output3Button = vuPvNameButtonClass::RequestSubLoad (out3container,
				vunames_vuUdbOUT3Button);
    output3 = vuFilteredTextitem::RequestSubLoad (out3container,
				vunames_OUT3);

    pin3 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PIN3);

    vcontainer *out4container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbOUT4Container);
    output4Button = vuPvNameButtonClass::RequestSubLoad (out4container,
				vunames_vuUdbOUT4Button);
    output4 = vuFilteredTextitem::RequestSubLoad (out4container,
				vunames_OUT4);

    pin4 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PIN4);

    vcontainer *out5container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbOUT5Container);
    output5Button = vuPvNameButtonClass::RequestSubLoad (out5container,
				vunames_vuUdbOUT5Button);
    output5 = vuFilteredTextitem::RequestSubLoad (out5container,
				vunames_OUT5);

    pin5 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PIN5);

    vcontainer *out6container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbOUT6Container);
    output6Button = vuPvNameButtonClass::RequestSubLoad (out6container,
				vunames_vuUdbOUT6Button);
    output6 = vuFilteredTextitem::RequestSubLoad (out6container,
				vunames_OUT6);

    pin6 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PIN6);

    vcontainer *out7container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbOUT7Container);
    output7Button = vuPvNameButtonClass::RequestSubLoad (out7container,
				vunames_vuUdbOUT7Button);
    output7 = vuFilteredTextitem::RequestSubLoad (out7container,
				vunames_OUT7);

    pin7 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PIN7);

    vcontainer *out8container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbOUT8Container);
    output8Button = vuPvNameButtonClass::RequestSubLoad (out8container,
				vunames_vuUdbOUT8Button);
    output8 = vuFilteredTextitem::RequestSubLoad (out8container,
				vunames_OUT8);

    pin8 = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PIN8);

    dismissButton = vuUdbIODismissButtonClass::RequestSubLoad (this,
				vunames_vuUdbIODialogDismissButton);

    vdialog::LoadInit (res);

    // store ptr to text field for vuPvnameButtonClass
    input1Button -> storePvName (input1);
    input2Button -> storePvName (input2);
    input3Button -> storePvName (input3);
    input4Button -> storePvName (input4);
    input5Button -> storePvName (input5);
    input6Button -> storePvName (input6);
    input7Button -> storePvName (input7);
    input8Button -> storePvName (input8);
    output1Button -> storePvName (output1);
    output2Button -> storePvName (output2);
    output3Button -> storePvName (output3);
    output4Button -> storePvName (output4);
    output5Button -> storePvName (output5);
    output6Button -> storePvName (output6);
    output7Button -> storePvName (output7);
    output8Button -> storePvName (output8);

    // set filters

    input1 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *inp1Text = (filteredText *)input1 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    inp1Text -> SetUserPtr ((void *)input1);

    input2 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *inp2Text = (filteredText *)input2 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    inp2Text -> SetUserPtr ((void *)input2);

    input3 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *inp3Text = (filteredText *)input3 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    inp3Text -> SetUserPtr ((void *)input3);

    input4 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *inp4Text = (filteredText *)input4 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    inp4Text -> SetUserPtr ((void *)input4);

    input5 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *inp5Text = (filteredText *)input5 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    inp5Text -> SetUserPtr ((void *)input5);

    input6 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *inp6Text = (filteredText *)input6 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    inp6Text -> SetUserPtr ((void *)input6);

    input7 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *inp7Text = (filteredText *)input7 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    inp7Text -> SetUserPtr ((void *)input7);

    input8 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *inp8Text = (filteredText *)input8 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    inp8Text -> SetUserPtr ((void *)input8);

    output1 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *out1Text = (filteredText *)output1 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    out1Text -> SetUserPtr ((void *)output1);

    output2 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *out2Text = (filteredText *)output2 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    out2Text -> SetUserPtr ((void *)output2);

    output3 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *out3Text = (filteredText *)output3 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    out3Text -> SetUserPtr ((void *)output3);

    output4 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *out4Text = (filteredText *)output4 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    out4Text -> SetUserPtr ((void *)output4);

    output5 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *out5Text = (filteredText *)output5 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    out5Text -> SetUserPtr ((void *)output5);

    output6 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *out6Text = (filteredText *)output6 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    out6Text -> SetUserPtr ((void *)output6);

    output7 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *out7Text = (filteredText *)output7 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    out7Text -> SetUserPtr ((void *)output7);

    output8 -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *out8Text = (filteredText *)output8 -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    out8Text -> SetUserPtr ((void *)output8);

    // store dialog pointer in dismiss button
    dismissButton -> storeIODialog (this);

}


/*******************************************************************
*
*    vuUdbIODialogClass notify function
*
*******************************************************************/

void vuUdbIODialogClass::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose ())
    {
	Close ();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuUdbIODismissButtonClass notify function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbIODismissButtonClass, vbutton, 
					"vuUdbIODismissButtonClass");

void vuUdbIODismissButtonClass::ObserveDialogItem (vevent *event)
{
    ioDialogClass -> Close ();
    vevent::StopProcessing ();
}



/*******************************************************************
*
*    vuUdbConstantsDialogClass
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbConstantsDialogClass, vdialog, 
					"vuUdbConstantsDialogClass");

/*******************************************************************
*
*    vuUdbConstantsDialogClass override for constructors init ()
*
*******************************************************************/

void vuUdbConstantsDialogClass::LoadInit (vresource res)
{

    vcontainer *flt1container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbN01Container);
    float1 = filteredTextitem::RequestSubLoad (flt1container,
				vunames_N01);
    float1Label = vdialogLabelItem::RequestSubLoad (flt1container,
				vunames_vuUdbN01Label);
    vcontainer *fl01container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFL01Container);
    float1Limit = vuChoiceRecMenuClass::RequestSubLoad (fl01container,
				vunames_FL01);

    vcontainer *flt2container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbN02Container);
    float2 = filteredTextitem::RequestSubLoad (flt2container,
				vunames_N02);
    float2Label = vdialogLabelItem::RequestSubLoad (flt2container,
				vunames_vuUdbN02Label);
    vcontainer *fl02container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFL02Container);
    float2Limit = vuChoiceRecMenuClass::RequestSubLoad (fl02container,
				vunames_FL02);

    vcontainer *flt3container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbN03Container);
    float3 = filteredTextitem::RequestSubLoad (flt3container,
				vunames_N03);
    float3Label = vdialogLabelItem::RequestSubLoad (flt3container,
				vunames_vuUdbN03Label);
    vcontainer *fl03container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFL03Container);
    float3Limit = vuChoiceRecMenuClass::RequestSubLoad (fl03container,
				vunames_FL03);

    vcontainer *flt4container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbN04Container);
    float4 = filteredTextitem::RequestSubLoad (flt4container,
				vunames_N04);
    float4Label = vdialogLabelItem::RequestSubLoad (flt4container,
				vunames_vuUdbN04Label);
    vcontainer *fl04container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFL04Container);
    float4Limit = vuChoiceRecMenuClass::RequestSubLoad (fl04container,
				vunames_FL04);

    vcontainer *flt5container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbN05Container);
    float5 = filteredTextitem::RequestSubLoad (flt5container,
				vunames_N05);
    float5Label = vdialogLabelItem::RequestSubLoad (flt5container,
				vunames_vuUdbN05Label);
    vcontainer *fl05container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFL05Container);
    float5Limit = vuChoiceRecMenuClass::RequestSubLoad (fl05container,
				vunames_FL05);

    vcontainer *flt6container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbN06Container);
    float6 = filteredTextitem::RequestSubLoad (flt6container,
				vunames_N06);
    float6Label = vdialogLabelItem::RequestSubLoad (flt6container,
				vunames_vuUdbN06Label);
    vcontainer *fl06container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFL06Container);
    float6Limit = vuChoiceRecMenuClass::RequestSubLoad (fl06container,
				vunames_FL06);

    vcontainer *flt7container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbN07Container);
    float7 = filteredTextitem::RequestSubLoad (flt7container,
				vunames_N07);
    float7Label = vdialogLabelItem::RequestSubLoad (flt7container,
				vunames_vuUdbN07Label);
    vcontainer *fl07container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFL07Container);
    float7Limit = vuChoiceRecMenuClass::RequestSubLoad (fl07container,
				vunames_FL07);

    vcontainer *flt8container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbN08Container);
    float8 = filteredTextitem::RequestSubLoad (flt8container,
				vunames_N08);
    float8Label = vdialogLabelItem::RequestSubLoad (flt8container,
				vunames_vuUdbN08Label);
    vcontainer *fl08container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFL08Container);
    float8Limit = vuChoiceRecMenuClass::RequestSubLoad (fl08container,
				vunames_FL08);

    vcontainer *flt9container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbN09Container);
    float9 = filteredTextitem::RequestSubLoad (flt9container,
				vunames_N09);
    float9Label = vdialogLabelItem::RequestSubLoad (flt9container,
				vunames_vuUdbN09Label);
    vcontainer *fl09container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFL09Container);
    float9Limit = vuChoiceRecMenuClass::RequestSubLoad (fl09container,
				vunames_FL09);

    vcontainer *flt10container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbN10Container);
    float10 = filteredTextitem::RequestSubLoad (flt10container,
				vunames_N10);
    float10Label = vdialogLabelItem::RequestSubLoad (flt10container,
				vunames_vuUdbN10Label);
    vcontainer *fl10container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFL10Container);
    float10Limit = vuChoiceRecMenuClass::RequestSubLoad (fl10container,
				vunames_FL10);

    vcontainer *flt11container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbN11Container);
    float11 = filteredTextitem::RequestSubLoad (flt11container,
				vunames_N11);
    float11Label = vdialogLabelItem::RequestSubLoad (flt11container,
				vunames_vuUdbN11Label);
    vcontainer *fl11container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFL11Container);
    float11Limit = vuChoiceRecMenuClass::RequestSubLoad (fl11container,
				vunames_FL11);

    vcontainer *flt12container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbN12Container);
    float12 = filteredTextitem::RequestSubLoad (flt12container,
				vunames_N12);
    float12Label = vdialogLabelItem::RequestSubLoad (flt12container,
				vunames_vuUdbN12Label);
    vcontainer *fl12container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFL12Container);
    float12Limit = vuChoiceRecMenuClass::RequestSubLoad (fl12container,
				vunames_FL12);

    vcontainer *db1container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbD01Container);
    double1 = filteredTextitem::RequestSubLoad (db1container,
				vunames_D01);
    double1Label = vdialogLabelItem::RequestSubLoad (db1container,
				vunames_vuUdbD01Label);
    vcontainer *dl01container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbDL01Container);
    double1Limit = vuChoiceRecMenuClass::RequestSubLoad (dl01container,
				vunames_DL01);

    vcontainer *db2container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbD02Container);
    double2 = filteredTextitem::RequestSubLoad (db2container,
				vunames_D02);
    double2Label = vdialogLabelItem::RequestSubLoad (db2container,
				vunames_vuUdbD02Label);
    vcontainer *dl02container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbDL02Container);
    double2Limit = vuChoiceRecMenuClass::RequestSubLoad (dl02container,
				vunames_DL02);

    vcontainer *db3container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbD03Container);
    double3 = filteredTextitem::RequestSubLoad (db3container,
				vunames_D03);
    double3Label = vdialogLabelItem::RequestSubLoad (db3container,
				vunames_vuUdbD03Label);
    vcontainer *dl03container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbDL03Container);
    double3Limit = vuChoiceRecMenuClass::RequestSubLoad (dl03container,
				vunames_DL03);

    vcontainer *db4container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbD04Container);
    double4 = filteredTextitem::RequestSubLoad (db4container,
				vunames_D04);
    double4Label = vdialogLabelItem::RequestSubLoad (db4container,
				vunames_vuUdbD04Label);
    vcontainer *dl04container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbDL04Container);
    double4Limit = vuChoiceRecMenuClass::RequestSubLoad (dl04container,
				vunames_DL04);

    vcontainer *db5container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbD05Container);
    double5 = filteredTextitem::RequestSubLoad (db5container,
				vunames_D05);
    double5Label = vdialogLabelItem::RequestSubLoad (db5container,
				vunames_vuUdbD05Label);
    vcontainer *dl05container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbDL05Container);
    double5Limit = vuChoiceRecMenuClass::RequestSubLoad (dl05container,
				vunames_DL05);

    vcontainer *db6container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbD06Container);
    double6 = filteredTextitem::RequestSubLoad (db6container,
				vunames_D06);
    double6Label = vdialogLabelItem::RequestSubLoad (db6container,
				vunames_vuUdbD06Label);
    vcontainer *dl06container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbDL06Container);
    double6Limit = vuChoiceRecMenuClass::RequestSubLoad (dl06container,
				vunames_DL06);

    vcontainer *ul1container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbU01Container);
    ulong1 = filteredTextitem::RequestSubLoad (ul1container,
				vunames_U01);
    ulong1Label = vdialogLabelItem::RequestSubLoad (ul1container,
				vunames_vuUdbU01Label);
    vcontainer *ul01container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbUL01Container);
    ulong1Limit = vuChoiceRecMenuClass::RequestSubLoad (ul01container,
				vunames_UL01);

    vcontainer *ul2container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbU02Container);
    ulong2 = filteredTextitem::RequestSubLoad (ul2container,
				vunames_U02);
    ulong2Label = vdialogLabelItem::RequestSubLoad (ul2container,
				vunames_vuUdbU02Label);
    vcontainer *ul02container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbUL02Container);
    ulong2Limit = vuChoiceRecMenuClass::RequestSubLoad (ul02container,
				vunames_UL02);

    vcontainer *chk1container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCK1Container);
    chkbx1 = vuChoiceGblMenuClass::RequestSubLoad (chk1container,
				vunames_CK1);
    chkbx1Label = vdialogLabelItem::RequestSubLoad (chk1container,
				vunames_vuUdbCK1Label);

    vcontainer *chk2container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCK2Container);
    chkbx2 = vuChoiceGblMenuClass::RequestSubLoad (chk2container,
				vunames_CK2);
    chkbx2Label = vdialogLabelItem::RequestSubLoad (chk2container,
				vunames_vuUdbCK2Label);

    vcontainer *chk3container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCK3Container);
    chkbx3 = vuChoiceGblMenuClass::RequestSubLoad (chk3container,
				vunames_CK3);
    chkbx3Label = vdialogLabelItem::RequestSubLoad (chk3container,
				vunames_vuUdbCK3Label);

    vcontainer *chk4container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCK4Container);
    chkbx4 = vuChoiceGblMenuClass::RequestSubLoad (chk4container,
				vunames_CK4);
    chkbx4Label = vdialogLabelItem::RequestSubLoad (chk4container,
				vunames_vuUdbCK4Label);

    vcontainer *chk5container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCK5Container);
    chkbx5 = vuChoiceGblMenuClass::RequestSubLoad (chk5container,
				vunames_CK5);
    chkbx5Label = vdialogLabelItem::RequestSubLoad (chk5container,
				vunames_vuUdbCK5Label);

    vcontainer *chk6container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCK6Container);
    chkbx6 = vuChoiceGblMenuClass::RequestSubLoad (chk6container,
				vunames_CK6);
    chkbx6Label = vdialogLabelItem::RequestSubLoad (chk6container,
				vunames_vuUdbCK6Label);

    vcontainer *chk7container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCK7Container);
    chkbx7 = vuChoiceGblMenuClass::RequestSubLoad (chk7container,
				vunames_CK7);
    chkbx7Label = vdialogLabelItem::RequestSubLoad (chk7container,
				vunames_vuUdbCK7Label);

    vcontainer *chk8container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbCK8Container);
    chkbx8 = vuChoiceGblMenuClass::RequestSubLoad (chk8container,
				vunames_CK8);
    chkbx8Label = vdialogLabelItem::RequestSubLoad (chk8container,
				vunames_vuUdbCK8Label);

    dismissButton = vuUdbConstantsDismissButtonClass::RequestSubLoad (this,
				vunames_vuUdbConstantsDialogDismissButton);

    vdialog::LoadInit (res);

    // set filters

    float1 -> SetItemFilterFunc (floatFilter);
    float2 -> SetItemFilterFunc (floatFilter);
    float3 -> SetItemFilterFunc (floatFilter);
    float4 -> SetItemFilterFunc (floatFilter);
    float5 -> SetItemFilterFunc (floatFilter);
    float6 -> SetItemFilterFunc (floatFilter);
    float7 -> SetItemFilterFunc (floatFilter);
    float8 -> SetItemFilterFunc (floatFilter);
    float9 -> SetItemFilterFunc (floatFilter);
    float10 -> SetItemFilterFunc (floatFilter);
    float11 -> SetItemFilterFunc (floatFilter);
    float12 -> SetItemFilterFunc (floatFilter);
    double1 -> SetItemFilterFunc (floatFilter);
    double2 -> SetItemFilterFunc (floatFilter);
    double3 -> SetItemFilterFunc (floatFilter);
    double4 -> SetItemFilterFunc (floatFilter);
    double5 -> SetItemFilterFunc (floatFilter);
    double6 -> SetItemFilterFunc (floatFilter);
    ulong1 -> SetItemFilterFunc (numericFilter);
    ulong2 -> SetItemFilterFunc (numericFilter);

    // build option menus

    float1Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    float2Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    float3Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    float4Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    float5Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    float6Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    float7Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    float8Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    float9Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    float10Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    float11Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    float12Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    double1Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    double2Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    double3Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    double4Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    double5Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    double6Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    ulong1Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    ulong2Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    chkbx1 -> vuBuildOptionMenu (GBL_YES_NO);
    chkbx2 -> vuBuildOptionMenu (GBL_YES_NO);
    chkbx3 -> vuBuildOptionMenu (GBL_YES_NO);
    chkbx4 -> vuBuildOptionMenu (GBL_YES_NO);
    chkbx5 -> vuBuildOptionMenu (GBL_YES_NO);
    chkbx6 -> vuBuildOptionMenu (GBL_YES_NO);
    chkbx7 -> vuBuildOptionMenu (GBL_YES_NO);
    chkbx8 -> vuBuildOptionMenu (GBL_YES_NO);

    // store dialog pointer in dismiss button
    dismissButton -> storeConstantsDialog (this);
}


/*******************************************************************
*
*    vuUdbConstantsDialogClass notify function
*
*******************************************************************/

void vuUdbConstantsDialogClass::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose ())
    {
	Close ();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuUdbConstantsDismissButtonClass notify function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbConstantsDismissButtonClass, vbutton, 
					"vuUdbConstantsDismissButtonClass");

void vuUdbConstantsDismissButtonClass::ObserveDialogItem (vevent *event)
{
    constantsDialogClass -> Close ();
    vevent::StopProcessing ();
}



/*******************************************************************
*
*    vuUdbLimitsDialogClass
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbLimitsDialogClass, vdialog, 
					"vuUdbLimitsDialogClass");

/*******************************************************************
*
*    vuUdbLlmitsDialogClass override for constructors init ()
*
*******************************************************************/

void vuUdbLimitsDialogClass::LoadInit (vresource res)
{

    vcontainer *ft01container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFT01Container);
    tempFloat1Limit = vuChoiceRecMenuClass::RequestSubLoad (ft01container,
				vunames_FT01);
    tempFloat1Label = vdialogLabelItem::RequestSubLoad (ft01container,
				vunames_vuUdbFT01Label);

    vcontainer *ft02container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFT02Container);
    tempFloat2Limit = vuChoiceRecMenuClass::RequestSubLoad (ft02container,
				vunames_FT02);
    tempFloat2Label = vdialogLabelItem::RequestSubLoad (ft02container,
				vunames_vuUdbFT02Label);

    vcontainer *ft03container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFT03Container);
    tempFloat3Limit = vuChoiceRecMenuClass::RequestSubLoad (ft03container,
				vunames_FT03);
    tempFloat3Label = vdialogLabelItem::RequestSubLoad (ft03container,
				vunames_vuUdbFT03Label);

    vcontainer *ft04container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFT04Container);
    tempFloat4Limit = vuChoiceRecMenuClass::RequestSubLoad (ft04container,
				vunames_FT04);
    tempFloat4Label = vdialogLabelItem::RequestSubLoad (ft04container,
				vunames_vuUdbFT04Label);

    vcontainer *ft05container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFT05Container);
    tempFloat5Limit = vuChoiceRecMenuClass::RequestSubLoad (ft05container,
				vunames_FT05);
    tempFloat5Label = vdialogLabelItem::RequestSubLoad (ft05container,
				vunames_vuUdbFT05Label);

    vcontainer *ft06container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFT06Container);
    tempFloat6Limit = vuChoiceRecMenuClass::RequestSubLoad (ft06container,
				vunames_FT06);
    tempFloat6Label = vdialogLabelItem::RequestSubLoad (ft06container,
				vunames_vuUdbFT06Label);

    vcontainer *ft07container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFT07Container);
    tempFloat7Limit = vuChoiceRecMenuClass::RequestSubLoad (ft07container,
				vunames_FT07);
    tempFloat7Label = vdialogLabelItem::RequestSubLoad (ft07container,
				vunames_vuUdbFT07Label);

    vcontainer *ft08container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFT08Container);
    tempFloat8Limit = vuChoiceRecMenuClass::RequestSubLoad (ft08container,
				vunames_FT08);
    tempFloat8Label = vdialogLabelItem::RequestSubLoad (ft08container,
				vunames_vuUdbFT08Label);

    vcontainer *ft09container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFT09Container);
    tempFloat9Limit = vuChoiceRecMenuClass::RequestSubLoad (ft09container,
				vunames_FT09);
    tempFloat9Label = vdialogLabelItem::RequestSubLoad (ft09container,
				vunames_vuUdbFT09Label);

    vcontainer *ft10container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFT10Container);
    tempFloat10Limit = vuChoiceRecMenuClass::RequestSubLoad (ft10container,
				vunames_FT10);
    tempFloat10Label = vdialogLabelItem::RequestSubLoad (ft10container,
				vunames_vuUdbFT10Label);

    vcontainer *ft11container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFT11Container);
    tempFloat11Limit = vuChoiceRecMenuClass::RequestSubLoad (ft11container,
				vunames_FT11);
    tempFloat11Label = vdialogLabelItem::RequestSubLoad (ft11container,
				vunames_vuUdbFT11Label);

    vcontainer *ft12container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFT12Container);
    tempFloat12Limit = vuChoiceRecMenuClass::RequestSubLoad (ft12container,
				vunames_FT12);
    tempFloat12Label = vdialogLabelItem::RequestSubLoad (ft12container,
				vunames_vuUdbFT12Label);

    vcontainer *ft13container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFT13Container);
    tempFloat13Limit = vuChoiceRecMenuClass::RequestSubLoad (ft13container,
				vunames_FT13);
    tempFloat13Label = vdialogLabelItem::RequestSubLoad (ft13container,
				vunames_vuUdbFT13Label);

    vcontainer *ft14container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFT14Container);
    tempFloat14Limit = vuChoiceRecMenuClass::RequestSubLoad (ft14container,
				vunames_FT14);
    tempFloat14Label = vdialogLabelItem::RequestSubLoad (ft14container,
				vunames_vuUdbFT14Label);

    vcontainer *ft15container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFT15Container);
    tempFloat15Limit = vuChoiceRecMenuClass::RequestSubLoad (ft15container,
				vunames_FT15);
    tempFloat15Label = vdialogLabelItem::RequestSubLoad (ft15container,
				vunames_vuUdbFT15Label);

    vcontainer *ft16container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFT16Container);
    tempFloat16Limit = vuChoiceRecMenuClass::RequestSubLoad (ft16container,
				vunames_FT16);
    tempFloat16Label = vdialogLabelItem::RequestSubLoad (ft16container,
				vunames_vuUdbFT16Label);

    vcontainer *td01container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbTD01Container);
    tempDouble1Limit = vuChoiceRecMenuClass::RequestSubLoad (td01container,
				vunames_TD01);
    tempDouble1Label = vdialogLabelItem::RequestSubLoad (td01container,
				vunames_vuUdbTD01Label);

    vcontainer *td02container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbTD02Container);
    tempDouble2Limit = vuChoiceRecMenuClass::RequestSubLoad (td02container,
				vunames_TD02);
    tempDouble2Label = vdialogLabelItem::RequestSubLoad (td02container,
				vunames_vuUdbTD02Label);

    vcontainer *td03container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbTD03Container);
    tempDouble3Limit = vuChoiceRecMenuClass::RequestSubLoad (td03container,
				vunames_TD03);
    tempDouble3Label = vdialogLabelItem::RequestSubLoad (td03container,
				vunames_vuUdbTD03Label);

    vcontainer *td04container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbTD04Container);
    tempDouble4Limit = vuChoiceRecMenuClass::RequestSubLoad (td04container,
				vunames_TD04);
    tempDouble4Label = vdialogLabelItem::RequestSubLoad (td04container,
				vunames_vuUdbTD04Label);

    vcontainer *ud01container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbUD01Container);
    tempUlong1Limit = vuChoiceRecMenuClass::RequestSubLoad (ud01container,
				vunames_UD01);
    tempUlong1Label = vdialogLabelItem::RequestSubLoad (ud01container,
				vunames_vuUdbUD01Label);

    vcontainer *ud02container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbUD02Container);
    tempUlong2Limit = vuChoiceRecMenuClass::RequestSubLoad (ud02container,
				vunames_UD02);
    tempUlong2Label = vdialogLabelItem::RequestSubLoad (ud02container,
				vunames_vuUdbUD02Label);

    vcontainer *ud03container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbUD03Container);
    tempUlong3Limit = vuChoiceRecMenuClass::RequestSubLoad (ud03container,
				vunames_UD03);
    tempUlong3Label = vdialogLabelItem::RequestSubLoad (ud03container,
				vunames_vuUdbUD03Label);

    vcontainer *ud04container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbUD04Container);
    tempUlong4Limit = vuChoiceRecMenuClass::RequestSubLoad (ud04container,
				vunames_UD04);
    tempUlong4Label = vdialogLabelItem::RequestSubLoad (ud04container,
				vunames_vuUdbUD04Label);

    vcontainer *ud05container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbUD05Container);
    tempUlong5Limit = vuChoiceRecMenuClass::RequestSubLoad (ud05container,
				vunames_UD05);
    tempUlong5Label = vdialogLabelItem::RequestSubLoad (ud05container,
				vunames_vuUdbUD05Label);

    vcontainer *ud06container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbUD06Container);
    tempUlong6Limit = vuChoiceRecMenuClass::RequestSubLoad (ud06container,
				vunames_UD06);
    tempUlong6Label = vdialogLabelItem::RequestSubLoad (ud06container,
				vunames_vuUdbUD06Label);

    vcontainer *ud07container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbUD07Container);
    tempUlong7Limit = vuChoiceRecMenuClass::RequestSubLoad (ud07container,
				vunames_UD07);
    tempUlong7Label = vdialogLabelItem::RequestSubLoad (ud07container,
				vunames_vuUdbUD07Label);

    vcontainer *ud08container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbUD08Container);
    tempUlong8Limit = vuChoiceRecMenuClass::RequestSubLoad (ud08container,
				vunames_UD08);
    tempUlong8Label = vdialogLabelItem::RequestSubLoad (ud08container,
				vunames_vuUdbUD08Label);

    dismissButton = vuUdbLimitsDismissButtonClass::RequestSubLoad (this,
				vunames_vuUdbLimitsDialogDismissButton);

    vdialog::LoadInit (res);

    // store dialog pointer in dismiss button
    dismissButton -> storeLimitsDialog (this);

    // build popup menus
    tempFloat1Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempFloat2Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempFloat3Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempFloat4Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempFloat5Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempFloat6Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempFloat7Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempFloat8Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempFloat9Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempFloat10Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempFloat11Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempFloat12Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempFloat13Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempFloat14Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempFloat15Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempFloat16Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempDouble1Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempDouble2Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempDouble3Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempDouble4Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempUlong1Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempUlong2Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempUlong3Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempUlong4Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempUlong5Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempUlong6Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempUlong7Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
    tempUlong8Limit -> vuBuildOptionMenu (VU_UDB, REC_UDB_LIM);
}


/*******************************************************************
*
*    vuUdbLimitsDialogClass notify function
*
*******************************************************************/

void vuUdbLimitsDialogClass::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose ())
    {
	Close ();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuUdbLimitsDismissButtonClass notify function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbLimitsDismissButtonClass, vbutton, 
					"vuUdbLimitsDismissButtonClass");

void vuUdbLimitsDismissButtonClass::ObserveDialogItem (vevent *event)
{
    limitsDialogClass -> Close ();
    vevent::StopProcessing ();
}


/*******************************************************************
*
*    vuUdbMiscDialogClass
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbMiscDialogClass, vdialog, 
					"vuUdbMiscDialogClass");

/*******************************************************************
*
*    vuUdbMiscDialogClass override for constructors init ()
*
*******************************************************************/

void vuUdbMiscDialogClass::LoadInit (vresource res)
{

    vcontainer *imdecontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbIMDEContainer);
    initialMode = vuChoiceRecMenuClass::RequestSubLoad (imdecontainer,
				vunames_IMDE);

    vcontainer *mst1container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbMST1Container);
    mode01 = filteredTextitem::RequestSubLoad (mst1container,
				vunames_MST1);

    vcontainer *mst2container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbMST2Container);
    mode02 = filteredTextitem::RequestSubLoad (mst2container,
				vunames_MST2);

    vcontainer *mst3container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbMST3Container);
    mode03 = filteredTextitem::RequestSubLoad (mst3container,
				vunames_MST3);

    vcontainer *mst4container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbMST4Container);
    mode04 = filteredTextitem::RequestSubLoad (mst4container,
				vunames_MST4);

    vcontainer *mst5container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbMST5Container);
    mode05 = filteredTextitem::RequestSubLoad (mst5container,
				vunames_MST5);

    vcontainer *mst6container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbMST6Container);
    mode06 = filteredTextitem::RequestSubLoad (mst6container,
				vunames_MST6);

    vcontainer *mst7container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbMST7Container);
    mode07 = filteredTextitem::RequestSubLoad (mst7container,
				vunames_MST7);

    vcontainer *mst8container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbMST8Container);
    mode08 = filteredTextitem::RequestSubLoad (mst8container,
				vunames_MST8);

    vcontainer *ifuncontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbIFUNContainer);
    initialFunction = vuChoiceRecMenuClass::RequestSubLoad (ifuncontainer,
				vunames_IFUN);

    vcontainer *fst1container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFST1Container);
    func01 = filteredTextitem::RequestSubLoad (fst1container,
				vunames_FST1);

    vcontainer *fst2container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFST2Container);
    func02 = filteredTextitem::RequestSubLoad (fst2container,
				vunames_FST2);

    vcontainer *fst3container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFST3Container);
    func03 = filteredTextitem::RequestSubLoad (fst3container,
				vunames_FST3);

    vcontainer *fst4container = vcontainer::RequestSubLoad (this,
				vunames_vuUdbFST4Container);
    func04 = filteredTextitem::RequestSubLoad (fst4container,
				vunames_FST4);

    vcontainer *l1locontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbL1LOContainer);
    limit1Low = filteredTextitem::RequestSubLoad (l1locontainer,
				vunames_L1LO);

    vcontainer *l2locontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbL2LOContainer);
    limit2Low = filteredTextitem::RequestSubLoad (l2locontainer,
				vunames_L2LO);

    vcontainer *l3locontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbL3LOContainer);
    limit3Low = filteredTextitem::RequestSubLoad (l3locontainer,
				vunames_L3LO);

    vcontainer *l1hicontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbL1HIContainer);
    limit1High = filteredTextitem::RequestSubLoad (l1hicontainer,
				vunames_L1HI);

    vcontainer *l2hicontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbL2HIContainer);
    limit2High = filteredTextitem::RequestSubLoad (l2hicontainer,
				vunames_L2HI);

    vcontainer *l3hicontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbL3HIContainer);
    limit3High = filteredTextitem::RequestSubLoad (l3hicontainer,
				vunames_L3HI);

    vcontainer *l1prcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbL1PRContainer);
    limit1Precision = filteredTextitem::RequestSubLoad (l1prcontainer,
				vunames_L1PR);

    vcontainer *l2prcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbL2PRContainer);
    limit2Precision = filteredTextitem::RequestSubLoad (l2prcontainer,
				vunames_L2PR);

    vcontainer *l3prcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuUdbL3PRContainer);
    limit3Precision = filteredTextitem::RequestSubLoad (l3prcontainer,
				vunames_L3PR);

    dismissButton = vuUdbMiscDismissButtonClass::RequestSubLoad (this,
				vunames_vuUdbMiscDismissButton);

    vdialog::LoadInit (res);

    // store dialog pointer in dismiss button
    dismissButton -> storeMiscDialog (this);

    // set filters
    mode01 -> SetItemFilterFunc (udbStringsFilter);
    mode02 -> SetItemFilterFunc (udbStringsFilter);
    mode03 -> SetItemFilterFunc (udbStringsFilter);
    mode04 -> SetItemFilterFunc (udbStringsFilter);
    mode05 -> SetItemFilterFunc (udbStringsFilter);
    mode06 -> SetItemFilterFunc (udbStringsFilter);
    mode07 -> SetItemFilterFunc (udbStringsFilter);
    mode08 -> SetItemFilterFunc (udbStringsFilter);
    func01 -> SetItemFilterFunc (udbStringsFilter);
    func02 -> SetItemFilterFunc (udbStringsFilter);
    func03 -> SetItemFilterFunc (udbStringsFilter);
    func04 -> SetItemFilterFunc (udbStringsFilter);
    limit1Low -> SetItemFilterFunc (floatFilter);
    limit2Low -> SetItemFilterFunc (floatFilter);
    limit3Low -> SetItemFilterFunc (floatFilter);
    limit1High -> SetItemFilterFunc (floatFilter);
    limit2High -> SetItemFilterFunc (floatFilter);
    limit3High -> SetItemFilterFunc (floatFilter);
    limit1Precision -> SetItemFilterFunc (numericFilter);
    limit2Precision -> SetItemFilterFunc (numericFilter);
    limit3Precision -> SetItemFilterFunc (numericFilter);

    // build option menus

    initialMode -> vuBuildOptionMenu (VU_UDB, REC_UDB_IMDE);
    initialFunction -> vuBuildOptionMenu (VU_UDB, REC_UDB_IFUN);
}


/*******************************************************************
*
*    vuUdbMiscDialogClass notify function
*
*******************************************************************/

void vuUdbMiscDialogClass::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose ())
    {
	Close ();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuUdbMiscDismissButtonClass notify function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbMiscDismissButtonClass, vbutton, 
					"vuUdbMiscDismissButtonClass");

void vuUdbMiscDismissButtonClass::ObserveDialogItem (vevent *event)
{
    miscDialogClass -> Close ();
    vevent::StopProcessing ();
}


/*******************************************************************
*
*    vuUdbProgramNamesDialogClass
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbProgramNamesDialogClass, vdialog, 
					"vuUdbProgramNamesDialogClass");

/*******************************************************************
*
*    vuUdbProgramNamesDialogClass override for constructors init ()
*
*******************************************************************/

void vuUdbProgramNamesDialogClass::LoadInit (vresource res)
{
    namesList = vuUdbProgramNamesListClass::RequestSubLoad (this,
				vunames_vuUdbNamesDialogList);
    okButton = vuUdbProgramNamesOKButtonClass::RequestSubLoad (this,
				vunames_vuUdbNamesDialogOKButton);
    dismissButton = vuUdbProgramNamesDismissButtonClass::RequestSubLoad (this,
				vunames_vuUdbNamesDialogDismissButton);
    vdialog::LoadInit (res);

    // store ptr to dialog in these
    namesList -> storeNamesDialog (this);
    okButton ->  storeNamesDialog (this);
    dismissButton ->  storeNamesDialog (this);
}


/*******************************************************************
*
*    vuUdbProgramNamesDialogClass notify function
*
*******************************************************************/

void vuUdbProgramNamesDialogClass::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose ())
    {
	vmemSet ((void *)udbProgramName, 0, UDBPROGRAMNAME_SZ);
	Close ();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuUdbProgramNamesDialogClass destroy function
*
*******************************************************************/

void vuUdbProgramNamesDialogClass::Destroy (void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }
    namesList -> Destroy ();
    vdialog::Destroy();
}


/*******************************************************************
*
*    vuUdbProgramNamesDialogClass open function
*	it first must build the list of UDB program names
*
*******************************************************************/

void vuUdbProgramNamesDialogClass::vuUdbNamesDialogOpen (dataVuApplication *vuApp)
{
    vuDatabaseClass	*db = vuApp -> getDatabaseClass ();

    vlistitemList *pgNamesList = vuUdbGetProgramNames (db, 0);
    namesList -> SetList (pgNamesList);
    namesList -> SetSelectMethod (vlistONE_ONLY_SELECTION);
    int rowCount = pgNamesList -> GetRowCount ();
    if (rowCount > 0)
    {
	namesList -> SortItems (NULL);
    }

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    Place (vwindow::GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    vuUdbProgramNamesListClass notify function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbProgramNamesListClass, vlistitem, 
					"vuUdbProgramNamesListClass");

void vuUdbProgramNamesListClass::ObserveSelect (int state)
{

    if (state == vlistitemSELECTION_HAS_CHANGED)
    {
	int clickCount = GetLastPointerClickCount ();
	if (clickCount == 2)
	{
	    vchar *pgName = namesDialog -> getUdbProgramName ();
	    int selectedItem = GetSelectedItem ();
	    const vchar *itemValue =  GetItemValue (selectedItem);
	    // copy value to program name array for use by caller
	    vcharCopyBounded (itemValue, pgName, UDBPROGRAMNAME_SZ);
	    namesDialog -> Close ();
	    vevent::StopProcessing ();
	}
    }

}

/*******************************************************************
*
*    vuUdbProgramNamesOKButtonClass notify function
*	get the name that is selected, if there is one, and store
*	it in the names dialogs program names array
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbProgramNamesOKButtonClass, vbutton, 
					"vuUdbProgramNamesOKButtonClass");

void vuUdbProgramNamesOKButtonClass::ObserveDialogItem (vevent *event)
{
    vchar *pgName = namesDialog -> getUdbProgramName ();
    int selectedItem = namesDialog -> namesList -> GetSelectedItem ();

    if (selectedItem != vlistitemNO_CELLS_SELECTED)
    {
	const vchar *itemValue = 
		namesDialog -> namesList -> GetItemValue (selectedItem);
	vcharCopyBounded (itemValue, pgName, UDBPROGRAMNAME_SZ);
    }
    else
    {
	vmemSet ((void *)pgName, 0, UDBPROGRAMNAME_SZ);
    }

    namesDialog -> Close ();
    vevent::StopProcessing ();
}


/*******************************************************************
*
*    vuUdbProgramNamesDismissButtonClass notify function
*	zap names dialog program names array
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbProgramNamesDismissButtonClass, vbutton, 
					"vuUdbProgramNamesDismissButtonClass");

void vuUdbProgramNamesDismissButtonClass::ObserveDialogItem (vevent *event)
{
    vchar *pgName = namesDialog -> getUdbProgramName ();

    vmemSet ((void *)pgName, 0, UDBPROGRAMNAME_SZ);

    namesDialog -> Close ();
    vevent::StopProcessing ();
}


/*******************************************************************
*
*    vuUdbConfigDialogClass
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbConfigDialogClass, vdialog, 
					"vuUdbConfigDialogClass");

/*******************************************************************
*
*    vuUdbConfigDialogClass override for constructors init ()
*
*******************************************************************/

void vuUdbConfigDialogClass::LoadInit (vresource res)
{
    pgnamesList = vuUdbCfgPgNamesListClass::RequestSubLoad (this,
					vunames_vuUdbConfigNamesList);
    labelsList = vuUdbCfgLabelsListClass::RequestSubLoad (this,
					vunames_vuUdbConfigLabelsList);
    vcontainer *namecontainer = vcontainer::RequestSubLoad (this,
					vunames_vuUdbConfigNameContainer);
    programName = filteredTextitem::RequestSubLoad (namecontainer,
					vunames_vuUdbConfigProgramName);
    vcontainer *defcontainer = vcontainer::RequestSubLoad (this,
					vunames_vuUdbConfigDefaultsContainer);
    defaultLabel = vtextitem::RequestSubLoad (defcontainer,
					vunames_vuUdbConfigDefault);
    vcontainer *newcontainer = vcontainer::RequestSubLoad (this,
					vunames_vuUdbConfigNewLabelContainer);
    newLabel = filteredTextitem::RequestSubLoad (newcontainer,
					vunames_vuUdbConfigNewLabel);

    vcontainer *buttoncontainer = vcontainer::RequestSubLoad (this,
					vunames_vuUdbConfigButtonContainer);
    applyButton = vuUdbCfgApplyButtonClass::RequestSubLoad (buttoncontainer,
					vunames_vuUdbConfigApplyButton);
    dismissButton = vuUdbCfgDismissButtonClass::RequestSubLoad (buttoncontainer,
					vunames_vuUdbConfigDismissButton);
    deleteButton = vuUdbCfgDeleteButtonClass::RequestSubLoad (buttoncontainer,
					vunames_vuUdbConfigDeleteButton);
    defaultButton = vuUdbCfgDefaultButtonClass::RequestSubLoad (buttoncontainer,
					vunames_vuUdbConfigDefaultButton);
    replaceButton = vuUdbCfgReplaceButtonClass::RequestSubLoad (buttoncontainer,
					vunames_vuUdbConfigReplaceButton);
    vdialog::LoadInit (res);

    // store ptr to dialog in these
    pgnamesList -> storeConfigDialog (this);
    labelsList -> storeConfigDialog (this);
    applyButton -> storeConfigDialog (this);
    dismissButton -> storeConfigDialog (this);
    deleteButton -> storeConfigDialog (this);
    defaultButton -> storeConfigDialog (this);
    replaceButton -> storeConfigDialog (this);

    // set some filters
    programName -> SetItemFilterFunc (vuUdbProgramNameFilter);
    newLabel -> SetItemFilterFunc (vuUdbLabelFilter);

    // init
    cfgModified = vFALSE;
}


/*******************************************************************
*
*    vuUdbConfigDialogClass notify function
*
*******************************************************************/

void vuUdbConfigDialogClass::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose ())
    {
	int result = VU_TRUE;

	if (cfgModified == vTRUE)
	{
	    result = vuMessageDialogCenter (this, VU_MOD_DISMISS_MSG);
	}
	if (result == VU_TRUE)
	{
	    Close ();
	    vevent::StopProcessing ();
	}
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    vuUdbConfigDialogClass destroy function
*
*******************************************************************/

void vuUdbConfigDialogClass::Destroy (void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }

    vdialog::Destroy();
}


/*******************************************************************
*
*    vuUdbConfigDialogClass open function
*
*******************************************************************/

void vuUdbConfigDialogClass::vuUdbConfigDialogOpen (dataVuApplication *vuApp,
						    vuFilteredTextitem *pgName,
						    vuUdbDialogClass *udbDialogClass)
{
    int selectedItem = -1;

    // store in dialog for future use
    database = vuApp -> getDatabaseClass ();

    // store UDB dialog for later use
    udbDialog = udbDialogClass;

    // build the program names list
    pgnamesList -> buildCfgNamesList ();

    // get the program name if there is one
    vstr *stringValue = pgName -> GetTextAsString ();
    int rowCount = pgnamesList -> GetItemCount ();
    // if there is a list
    if (rowCount > 0)
    {
	int i;
	int foundIt = vFALSE;
	const vchar *itemValue;
	// see if this program name exists in the list
	// if it does select it, this will cause the lists
	// ObserveSelect to execute
	for (i = 0; i < rowCount; i++)
	{
	    itemValue = pgnamesList -> GetItemValue (i);
	    if ( (vcharCompareBounded (stringValue, itemValue, UDBPROGRAMNAME_SZ)) == 0)
	    {
		selectedItem = i;
		foundIt = vTRUE;
		break;
	    }
	}
	// if not there select the 1st in the list,
	if (foundIt == vFALSE)
	{
	    selectedItem = 0;
	}
    }
    vstrDestroy (stringValue);

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    Place (vwindow::GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    // this should cause the names list ObserveSelect to fire
    // NOTE: done here because the dialog MUST be open for the select
    //		to fire 
    if (selectedItem >= 0)
    {
	pgnamesList -> SelectItem (selectedItem);
    }
    vevent::Process ();
}


/*******************************************************************
*
*    vuUdbConfigDialogClass function to load the udb's dialogLabels
*	list newLabel with the default labels
*
*******************************************************************/

void vuUdbConfigDialogClass::vuUdbLoadDefaultConfig (void)
{
    vlist *list = udbDialog -> vuUdbGetDialogLabelsList ();
    vlistIterator	iterator;
    VU_UDBLABEL_PTR	label;

    iterator.Start (list);
    while (iterator.Next ())
    {
	label = (VU_UDBLABEL_PTR)(iterator.GetValue ());
	vcharCopyBounded ((const vchar *)label -> defaultLabel, 
			  label -> newLabel, UDBLABEL_SZ);
	label -> newLabel[UDBLABEL_SZ] = 0;
    }
    iterator.Finish ();
}


/*******************************************************************
*
*    vuUdbConfigDialogClass function to load the udb's dialogLabels
*	list newLabel with the values based on the current selected
*	udb program name
*
*******************************************************************/

void vuUdbConfigDialogClass::vuUdbLoadCurrentConfig (void)
{
    DB_RECORD_PTR	rec;

    // get the current selected udb program name
    const vchar *selectedProgramName = 
	pgnamesList -> GetItemValue (pgnamesList -> GetSelectedItem());

    char pgName[UDBPROGRAMNAME_SZ + 1];
    vcharExportBounded (vcharScribeSystemExport (selectedProgramName),
			(void *)pgName, 
			UDBPROGRAMNAME_SZ);

    vlist *list = udbDialog -> vuUdbGetDialogLabelsList ();

    vuDatabaseClass *db = getDatabase ();

    if ( (rec = DB_Seek (db -> vuUdbCfgTable, pgName)) != 0)
    {
	DB_FIELD_PTR	field;
	vlistIterator	iterator;
	VU_UDBLABEL_PTR	listValue;
	char		fieldName [12];
	char		labelName [UDBLABEL_SZ + 1];
	int		type;
	int		row;
	iterator.Start (list);
	while (iterator.Next ())
	{
	    row = iterator.GetRow ();
	    listValue = (VU_UDBLABEL_PTR)(iterator.GetValue ());
	    // list rows indexed from 0, field name LABEL1 -> LABEL72
	    sprintf (fieldName, "LABEL%d\0", row + 1);
	    field = DB_MapField (db -> vuUdbCfgTable, fieldName);
	    DB_GetFieldValue (rec, field, labelName, &type);
	    if (strlen (labelName))
	    {
		vcharCopyScribedBounded (vcharScribeSystem (labelName),
					 listValue -> newLabel,
					 UDBLABEL_SZ);
		listValue -> newLabel[UDBLABEL_SZ] = 0;
	    }
	}
	iterator.Finish ();
    }
    // if udb program name is not in the config file
    // we will use the defaults
    else
    {
	vuUdbLoadDefaultConfig ();
    }
}


/*******************************************************************
*
*    vuUdbCfgPgNamesListClass
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbCfgPgNamesListClass, vlistitem, 
					"vuUdbCfgPgNamesListClass");

/*******************************************************************
*
*    vuUdbCfgPgNamesListClass notify function
*
*******************************************************************/

void vuUdbCfgPgNamesListClass::ObserveSelect (int state)
{

    if (state == vlistitemSELECTION_HAS_CHANGED)
    {
	int selectedItem = GetSelectedItem ();
	if (selectedItem == vlistitemNO_CELLS_SELECTED)
	{
	    return;
	}
	const vchar *itemValue = GetItemValue (selectedItem);

	// update the program name field with the current
	// selected program name
	configDialog -> programName -> SetText (itemValue);

	// build the labels list based on the current 
	// selected program name
	configDialog -> labelsList -> buildCfgLabelsList (1);

    }
}


/*******************************************************************
*
*    vuUdbCfgPgNamesListClass build list function
*
*******************************************************************/

void vuUdbCfgPgNamesListClass::buildCfgNamesList (void)
{
    vuDatabaseClass	*db = configDialog -> getDatabase ();

    vlistitemList *pgNamesList = vuUdbGetProgramNames (db, 1);
    SetList (pgNamesList);

    SetSelectMethod (vlistONE_ONLY_SELECTION);
    int rowCount = pgNamesList -> GetRowCount ();
    if (rowCount > 0)
    {
	SortItems (NULL);
    }
}


/*******************************************************************
*
*    vuUdbCfgLabelsListClass
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbCfgLabelsListClass, vlistitem, 
					"vuUdbCfgLabelsListClass");

/*******************************************************************
*
*    vuUdbCfgLabelsListClass notify function
*
*******************************************************************/

void vuUdbCfgLabelsListClass::ObserveSelect (int state)
{
    if (state == vlistitemSELECTION_HAS_CHANGED)
    {
	vchar *defLbl, *newLbl;
	vchar listVal[80];
	int   i;
	int selectedItem = GetSelectedItem ();
	if (selectedItem == vlistitemNO_CELLS_SELECTED)
	{
	    return;
	}
	const vchar *listValue = GetItemValue (selectedItem);
	vcharCopyBounded (listValue, listVal, sizeof (listVal));
	// see buildCfgLabelList for format of list string
	defLbl = vcharGetFirstToken (listVal, "|");
	newLbl = vcharGetNextToken ("\0");

	// clip off trailing spaces on default value
	for (i = (vcharLength ((const vchar *)defLbl)) - 1; i >=0; i--)
	{
	    if (defLbl[i] != ' ')
	    {
		defLbl[i+1] = 0;
		break;
	    }
	}

	// clip off any leading spaces on new label
	while ((*(newLbl)) == ' ')
	    newLbl++;

	// now set the default and new label fields in the
	// config dialog
	configDialog -> defaultLabel -> SetText ((const vchar *)defLbl);
	configDialog -> newLabel -> SetText ((const vchar *)newLbl);
    }
}


/*******************************************************************
*
*    vuUdbCfgLabelsListClass build list function
*
*******************************************************************/

void vuUdbCfgLabelsListClass::buildCfgLabelsList (int useCurrentConfig)
{
    if (useCurrentConfig)
    {
	// load in the current label list based on the current selected
	// udbprogram name
	configDialog -> vuUdbLoadCurrentConfig ();
    }
    else
    {
	configDialog -> vuUdbLoadDefaultConfig ();
    }

    // delete the current list from the label list item
    vliststr *labelList = GetListstr ();
    int rowCount = labelList -> GetRowCount ();
    if (rowCount > 0)
    {
	// set listitem list ptr to NULL
//	SetList (NULL);
	// delete the formally attached list
//	labelList ->  Destroy () ;
	// create a new list for the listitem
	vlistitemList *newList = CreateList ();
	// attach it to the listitem
	SetList (newList);
    }

    VU_UDBLABEL_PTR	labelValue;
    vlistIterator	iterator;
    vuUdbDialogClass	*udbDlg = configDialog -> getUdbDialog ();
    vlist *list = udbDlg -> vuUdbGetDialogLabelsList ();

    iterator.Start (list);
    while (iterator.Next ())
    {
	labelValue = (VU_UDBLABEL_PTR)iterator.GetValue ();
	// add this to the label items list
	AppendScribed (vcharScribeFormatLiteral ("%-*.*s | %s\0",
			UDBLABEL_SZ, UDBLABEL_SZ,
			labelValue -> defaultLabel,
			labelValue -> newLabel) );
    }
    iterator.Finish ();
    SetSelectMethod (vlistONE_ONLY_SELECTION);
    // select the first item in the list,
    // this should cause the listitem's ObserveSelect to fire
    SelectItem (0);
}


/*******************************************************************
*
*    vuUdbCfgApplyButtonClass notify function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbCfgApplyButtonClass, vbutton, 
					"vuUdbCfgApplyButtonClass");

void vuUdbCfgApplyButtonClass::ObserveDialogItem (vevent *event)
{
    int	newConfig = vFALSE;

    // do you really want to update question
    int result = vuMessageDialogCenter (configDialog, VU_UDB_CFG_UPDATE_MSG);
    if (result == VU_FALSE)
    {
	return;
    }
    vstr *pgName = configDialog -> programName -> GetTextAsString();
    // a you dummy there is no program name message
    if ( (vcharLength (pgName)) == 0)
    {
	vuMessageDialogCenter (configDialog, VU_UDB_CFG_NO_PGNAME_ERR);
	vstrDestroy (pgName);
	return;
    }
    DB_RECORD_PTR	rec;
    DB_FIELD_PTR	pgnameField;
    vuDatabaseClass	*db = configDialog -> getDatabase ();
    vuUdbDialogClass	*udbDlg = configDialog -> getUdbDialog ();
    vuMainDialogClass	*mainDialog = udbDlg -> getMainDialog ();
    char pgname [UDBPROGRAMNAME_SZ + 1];

    vcharExportBounded (vcharScribeSystemExport ((const vchar *)pgName),
			    (void *)pgname, UDBPROGRAMNAME_SZ);
    pgname[UDBPROGRAMNAME_SZ] = 0;

    if ( (rec = DB_Seek (db -> vuUdbCfgTable, pgname)) == 0)
    {
	// create new record in config file, add program name to record
	rec = (DB_RECORD_PTR)DB_NewRecord (db -> vuUdbCfgTable, 0);
        pgnameField = DB_MapField (db -> vuUdbCfgTable, FLD_UDBPGNAME);
	DB_PutFieldValue (rec, pgnameField, pgname);

	// put msg up on dataVu as to what's happening
	mainDialog -> vuMainMessageCenter (vunames_vuUdbNewCfgMsg, pgName);

	newConfig = vTRUE;
    }
    else
    {
	// put up modify msg on dataVu
	mainDialog -> vuMainMessageCenter (vunames_vuUdbCfgModMsg, pgName);
    }

    vlistIterator iterator;
    vlist *list = udbDlg -> vuUdbGetDialogLabelsList ();
    VU_UDBLABEL_PTR	labelValue;
    char		label[UDBLABEL_SZ + 1];
    char		fieldName[UDBLABEL_SZ + 1];
    DB_FIELD_PTR	field;
    int			row;

    iterator.Start (list);
    while (iterator.Next ())
    {
	row = iterator.GetRow ();
	labelValue = (VU_UDBLABEL_PTR)iterator.GetValue ();
	vcharExportBounded (vcharScribeSystemExport (labelValue -> newLabel),
			    (void *)label, UDBLABEL_SZ);
	label[UDBLABEL_SZ] = 0;
	// list rows indexed from 0, field name WIDGET1 - WIDGET72, LABEL1 -> LABEL72
	sprintf (fieldName, "WIDGET%d\0", row + 1);
	field = DB_MapField (db -> vuUdbCfgTable, fieldName);
	DB_PutFieldValue (rec, field, labelValue -> widgetId);
	sprintf (fieldName, "LABEL%d\0", row + 1);
	field = DB_MapField (db -> vuUdbCfgTable, fieldName);
	DB_PutFieldValue (rec, field, label);
    }
    iterator.Finish ();

    // clear config mod flag
    configDialog -> vuUdbSetCfgModFlag (vFALSE);
    // clear UDB labels mod flag to cause update of dialog labels
    udbDlg -> vuUdbSetLabelsFlag (vFALSE);

    if (newConfig == vTRUE)
    {
	// add name to program names list
	configDialog -> pgnamesList -> AppendScribed (vcharScribe((const vchar *)pgName) );
	configDialog -> pgnamesList -> SortItems (NULL);
    }

    vstrDestroy (pgName);
}


/*******************************************************************
*
*    vuUdbCfgDismissButtonClass notify function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbCfgDismissButtonClass, vbutton, 
					"vuUdbCfgDismissButtonClass");

void vuUdbCfgDismissButtonClass::ObserveDialogItem (vevent *event)
{
    int result = VU_TRUE;
    int mod = configDialog -> vuUdbGetCfgModFlag ();

    if (mod == vTRUE)
    {
	result = vuMessageDialogCenter (configDialog, VU_MOD_DISMISS_MSG);
    }
    if (result == VU_TRUE)
    {
	configDialog -> Close ();
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    vuUdbCfgDeleteButtonClass notify function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbCfgDeleteButtonClass, vbutton, 
					"vuUdbCfgDeleteButtonClass");

void vuUdbCfgDeleteButtonClass::ObserveDialogItem (vevent *event)
{
    vstr *pgName = configDialog -> programName -> GetTextAsString ();
    if ( (vcharLength (pgName)) == 0)
    {
	vuMessageDialogCenter (configDialog, VU_UDB_CFG_NO_PGNAME_ERR);
	vstrDestroy (pgName);
	return;
    }
    DB_RECORD_PTR	rec;
    vuDatabaseClass	*db = configDialog -> getDatabase ();
    vuUdbDialogClass	*udbDlg = configDialog -> getUdbDialog ();
    vuMainDialogClass	*mainDialog = udbDlg -> getMainDialog ();
    char		pgname [UDBPROGRAMNAME_SZ + 1];

    vcharExportBounded (vcharScribeSystemExport ((const vchar *)pgName),
		        (void *)pgname, 
			UDBPROGRAMNAME_SZ);
    pgname[UDBPROGRAMNAME_SZ] = 0;

    // if not in udbcfg table, give message
    if ( (rec = DB_Seek (db -> vuUdbCfgTable, pgname)) == 0)
    {
	vuMessageDialogCenter (configDialog,
			       VU_UDB_CFG_NO_RECORD_ERR, 
			       pgName);
	vstrDestroy (pgName);
	return;
    }
    // ask if really want to delete pgName config
    int result = vuMessageDialogCenter (configDialog, 
					VU_UDB_CFG_DELETE_MSG,
					pgName);
    if (result == VU_TRUE)
    {
	DB_DeleteRecord (db -> vuUdbCfgTable, rec);
	mainDialog -> vuMainMessageCenter (vunames_vuUdbCfgDeletingMsg, pgName);
	configDialog -> programName -> SetTextScribed (vcharScribeSystem (""));
	configDialog -> defaultLabel -> SetTextScribed (vcharScribeSystem (""));
	configDialog -> newLabel -> SetTextScribed (vcharScribeSystem (""));
	configDialog -> vuUdbSetCfgModFlag (vFALSE);

	// delete the program name from the names list
	int numRows = configDialog -> pgnamesList -> GetItemCount ();
	int row;
	const vchar *itemValue;
	for (row = 0; row < numRows; row++)
	{
	    itemValue = configDialog -> pgnamesList -> GetItemValue (row);
	    if ( (vcharCompareBounded (itemValue, pgName, UDBPROGRAMNAME_SZ)) == 0)
	    {
		configDialog -> pgnamesList -> RemoveItem (row);
		if (numRows > 1)
		{
		    // this will cause the names list ObserveSelect to fire
		    configDialog -> pgnamesList -> SelectItem (0);
		}
		break;
	    }
	}
    }

    vstrDestroy (pgName);
}


/*******************************************************************
*
*    vuUdbCfgDefaultButtonClass notify function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbCfgDefaultButtonClass, vbutton, 
					"vuUdbCfgDefaultButtonClass");

void vuUdbCfgDefaultButtonClass::ObserveDialogItem (vevent *event)
{
    // build the new list based on the defaults
    configDialog -> labelsList -> buildCfgLabelsList (0);
    configDialog -> vuUdbSetCfgModFlag (vTRUE);
}


/*******************************************************************
*
*    vuUdbCfgReplaceButtonClass notify function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(vuUdbCfgReplaceButtonClass, vbutton, 
					"vuUdbCfgReplaceButtonClass");

void vuUdbCfgReplaceButtonClass::ObserveDialogItem (vevent *event)
{
    int selectedListItem = configDialog -> labelsList -> GetSelectedItem ();
    if (selectedListItem == vlistitemNO_CELLS_SELECTED)
    {
	vuMessageDialogCenter (configDialog, VU_UDB_CFG_LABEL_SELECT_ERR);
	return;
    }
    // get values from the 2 fields
    vstr *deflabel = configDialog -> defaultLabel -> GetTextAsString ();
    vstr *newlabel = configDialog -> newLabel -> GetTextAsString ();
    if (vcharLength ((const vchar *)newlabel) == 0)
    {
	vuMessageDialogCenter (configDialog, VU_UDB_CFG_NO_NEWLABEL_ERR);
	vstrDestroy (deflabel);
	vstrDestroy (newlabel);
	return;
    }
    // change the value in the labels list
    configDialog -> labelsList -> SetItemValueScribed ( 
			vcharScribeFormatLiteral ("%-*.*s | %s\0",
				  UDBLABEL_SZ, UDBLABEL_SZ, 
				  deflabel, newlabel), selectedListItem);

    // replace the new label value in the udb labels list
    vuUdbDialogClass *udbDlg = configDialog -> getUdbDialog ();
    vlist *list = udbDlg -> vuUdbGetDialogLabelsList ();
    VU_UDBLABEL_PTR labelValue = (VU_UDBLABEL_PTR)(list -> GetCellValue (selectedListItem, 0));
    vcharCopyBounded ((const vchar *)newlabel, labelValue -> newLabel, UDBLABEL_SZ);
    configDialog -> vuUdbSetCfgModFlag (vTRUE);
    vstrDestroy (deflabel);
    vstrDestroy (newlabel);
}


/*******************************************************************
*
*    vuUdbProgramNameSpecialFilter ()
*	special filter function for the udb program name field
*	if a space is typed as the first character, the
*	UDB Program Name dialog is brought up so the user can make 
*	a selection
*
*******************************************************************/

static int vuUdbProgramNameSpecialFilter (filteredText *text, vtextSelection *sel,
					  const vchar **string, long *len)
{
    int		ret;
    vchar	aSpace;

    if (*len > 0)
    {
	aSpace = vcharFromLiteral (' ');
	if (**string == aSpace)
	{
	    long textLen = text -> GetLength ();
	    if (textLen != 0)
	    {
		vwindow::Beep ();
		ret = vFALSE;
	    }
	    else
	    {
		vuFilteredTextitem *textItem = (vuFilteredTextitem *)text -> GetUserPtr ();
		if (textItem != NULL)
		{
		    vuCommonDialogClass *commonDialog = textItem -> getCommonDialog ();
		    int	ok = vFALSE;
		    if (commonDialog != NULL)
		    {
			vuMainDialogClass *mainDialog = commonDialog -> getMainDialog ();
			dataVuApplication *vuApp = mainDialog -> getApplication ();
			vresource vuRes = vuApp -> getDataVuvresource ();
			vresource vuUdbNamesRes = vresourceGet (vuRes, 
						vunames_vuUdbNamesDialog);
			vuUdbProgramNamesDialogClass *udbNamesDialog = 
					new vuUdbProgramNamesDialogClass(vuUdbNamesRes);
			udbNamesDialog -> vuUdbNamesDialogOpen (vuApp);
			vchar *progName = udbNamesDialog -> getUdbProgramName ();
			vchar *newString = ((vchar *)(*string));
			if (progName[0] != 0)
			{
			    *len = vcharLength (progName);
			    vcharCopyBounded (progName, newString, UDBPROGRAMNAME_SZ);
			    ret = NCharFilter (text, sel, string, len, UDBPROGRAMNAME_SZ - 1);
			    ok = vTRUE;
			}
			else
			{
				ok = vFALSE;
			}
			delete udbNamesDialog;
		    } 
		    if (ok == vFALSE)
		    {
			vwindow::Beep ();
			ret = vFALSE;
		    }
		}
		else
		{
		    vwindow::Beep ();
		    ret = vFALSE;
		}
	    }
	}
	else
	{
	    ret = NCharFilter (text, sel, string, len, UDBPROGRAMNAME_SZ - 1);
	}
    }
    else
    {
	vwindow::Beep ();
	ret = vFALSE;
    }

    return (ret);
}


/*******************************************************************
*
*    some simple filters for the udb config dialog
*    vuUdbProgramNameFilter ()
*    vuUdbLabelFilter ()
*
*******************************************************************/

static int vuUdbProgramNameFilter (filteredText *text, vtextSelection *sel,
				   const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, UDBPROGRAMNAME_SZ - 1) );
}

static int vuUdbLabelFilter (filteredText *text, vtextSelection *sel,
				   const vchar **string, long *len)
{
    return (NCharFilter (text, sel, string, len, UDBLABEL_SZ - 1) );
}


/*******************************************************************
*
*    vuUdbGetProgramNames function
*	returns a list if UDB program names to the caller
*
*******************************************************************/

static vlistitemList *vuUdbGetProgramNames (vuDatabaseClass *db, int useCfgTable)
{
    DB_TABLE_PTR	table;
    DB_RECORD_PTR	rec;
    DB_FIELD_PTR	pgnameField;
    char		pgName [UDBPROGRAMNAME_SZ + 1];
    vstr		*vstrpgName;
    vchar		*value;
    vlistIterator	iterator;
    int			foundDup;
    int			type;

    if (useCfgTable)
    {
	table = db -> vuUdbCfgTable;
    }
    else
    {
	table = db -> vuUdbTable;
    }
    vlistitemList *list = new vlistitemList ();
    pgnameField = DB_MapField (table, FLD_UDBPGNAME);
    rec = DB_Top (table);
    int row = 0;
    while (rec)
    {
	DB_GetFieldValue (rec, pgnameField, pgName, &type);
	pgName [UDBPROGRAMNAME_SZ] = 0;

	// the same udb program name can be used by multiple points
	// only want it in the list once
	foundDup = vFALSE;
	vstrpgName = vstrCloneScribed (vcharScribeSystem (pgName));
	iterator.Start (list);
	while (iterator.Next())
	{
	    value = (vchar *)(iterator.GetValue ());
	    if (vcharCompareBounded (vstrpgName, value, UDBPROGRAMNAME_SZ) == 0)
	    {
		foundDup = vTRUE;
		break;
	    }
	}
	iterator.Finish ();
	vstrDestroy (vstrpgName);

	if (foundDup == vFALSE)
	{
	    list -> SetCellValueScribed (row, 0, vcharScribeSystem (pgName));
	    row++;
	}
	rec = DB_SeekNext (table);
    }
    return (list);
}

