#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vcontainerHEADER
#include vbuttonHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vmemHEADER
#include vnumHEADER
#include vlistitemHEADER
#include vdateHEADER
#include vspinnerHEADER

extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "dbFldTypes.h"
#include "choiceGbl.h"
#include "choiceWaveform.h"
#include "DBDefines.h"
#include "DBCVerifyDefs.h"
#include "DBProto.h"
#include "dbRecDes.h"
}

#include "vunames.h"
#include "dataVu.hxx"
#include "vuMainDialogClass.hxx"
#include "vuDatabaseClass.hxx"
#include "vuCommonDialogClass.hxx"
#include "vuWaveformDialogClass.hxx"
#include "vuNavigatorDialogClass.hxx"
#include "vuUtils.hxx"


vkindTYPED_FULL_DEFINITION(vuWaveformDialogClass, vuCommonDialogClass, "vuWaveformDialogClass");

/*******************************************************************
*
*    vuWaveformDialogClass override for constructors init ()
*
*******************************************************************/

void vuWaveformDialogClass::LoadInit (vresource res)
{
    // load in the ai specific fields
    waveformContainer = vuWaveformContainerClass::RequestSubLoad (this,
				vunames_vuCommonContainer);
    vuCommonDialogClass::LoadInit (res);

    // store dialog modified flag
    waveformContainer -> boardType -> vuSetModifiedPtr (&this -> dialogModified);
    waveformContainer -> line -> SetTextModifiedPtr (&this -> dialogModified);
    waveformContainer -> addr -> SetTextModifiedPtr (&this -> dialogModified);
    waveformContainer -> card -> SetTextModifiedPtr (&this -> dialogModified);
    waveformContainer -> signal -> SetTextModifiedPtr (&this -> dialogModified);
    waveformContainer -> reference -> SetTextModifiedPtr (&this -> dialogModified);
    waveformContainer -> constantValue -> SetTextModifiedPtr (&this -> dialogModified);
    waveformContainer -> onScan -> vuSetModifiedPtr (&this -> dialogModified);
    waveformContainer -> processInput -> vuSetModifiedPtr (&this -> dialogModified);
    waveformContainer -> maximizeSeverity -> vuSetModifiedPtr (&this -> dialogModified);
    waveformContainer -> rearm -> vuSetModifiedPtr (&this -> dialogModified);
    waveformContainer -> monitorCount -> SetTextModifiedPtr (&this -> dialogModified);
    waveformContainer -> elementCount -> SetTextModifiedPtr (&this -> dialogModified);
    waveformContainer -> precision -> SetTextModifiedPtr (&this -> dialogModified);
    waveformContainer -> eguTag -> SetTextModifiedPtr (&this -> dialogModified);
    waveformContainer -> displayLow -> SetTextModifiedPtr (&this -> dialogModified);
    waveformContainer -> displayHigh -> SetTextModifiedPtr (&this -> dialogModified);
    waveformContainer -> valueType -> vuSetModifiedPtr (&this -> dialogModified);
    waveformContainer -> triggerCount -> vuSetModifiedPtr (&this -> dialogModified);
    waveformContainer -> sampleRate -> vuSetModifiedPtr (&this -> dialogModified);

    // store dialog ptr for vuPvNameButtonClass
    waveformContainer -> referenceButton -> storeCommonDialog ((vuCommonDialogClass *)this);

    // store dialog ptr for vuPvNameFilter
    waveformContainer -> reference -> storeCommonDialog ((vuCommonDialogClass *)this);

    // set the point type
    setPointType (VU_WAVEFORM);

    // get the title from the resource file, aiTitle must be
    // freed by the class destuctor
    vapplication *app = vapplication::GetCurrent ();
    vresource dialogTitles = vresourceGet (app -> GetResources (),
				vunames_vuDialogTitles);
    waveformTitle = vresourceGetString (dialogTitles, 
				vunames_vuWaveformDialogTitle);
    SetTitle (waveformTitle);

    // trend button is not needed 
    vuCommonTrendButtonClass *trendbutton = getTrendButton ();
    trendbutton -> Disable ();
}


/*******************************************************************
*
*    vuWaveformDialogClass Destroy function
*
*******************************************************************/

void vuWaveformDialogClass::Destroy (void)
{
    vstrDestroy (waveformTitle);
    vuCommonDialogClass::Destroy ();
}


/*******************************************************************
*
*    vuWaveformDialogClass, reset fields, overrides base class 
*	vuCommonDialogClass
*
*******************************************************************/

void vuWaveformDialogClass::resetDialog (void)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    vchar		epics[FLDNAME_SZ + 1];

    vresource defRes = mainDialog -> getDefaultsResource ();

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0; 
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	// if not a common dialog field
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    dialogItem = vuFindContainerItem (waveformContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    if ( (dialogItem -> IsKindOf (&vtextitem::Kind)) == vTRUE)
	    {
		textItem = (vtextitem *)dialogItem;

		// set the default value for the Waveform's line field
		if ( ((vcharCompareBounded (epics, "LINE", FLDNAME_SZ)) == 0) ||
		     ((vcharCompareBounded (epics, "ADDR", FLDNAME_SZ)) == 0) ||
		     ((vcharCompareBounded (epics, "CARD", FLDNAME_SZ)) == 0) ||
		     ((vcharCompareBounded (epics, "SIGN", FLDNAME_SZ)) == 0) )
		{
		    vstr *zero = vresourceGetString (defRes, 
						vunames_vuZeroDefault);
		    textItem -> SetText (zero);
		    vstrDestroy (zero);
		}
		// set the default value for the Waveform's precision field
		else if ( (vcharCompareBounded (epics, "PREC", FLDNAME_SZ)) == 0)
		{
		    vstr *precision = vresourceGetString (defRes, 
						vunames_vuPRECdefault);
		    textItem -> SetText (precision);
		    vstrDestroy (precision);
		}
		else
		{
		    textItem -> SetTextScribed (vcharScribeSystem (""));
		}
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolToggle::Kind)) == vTRUE)
	    {
		toggleItem = (vcontrolToggle *)dialogItem;
		toggleItem -> SetValue (0);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolOptionMenu::Kind)) == vTRUE)
	    {
		optionMenu = (vcontrolOptionMenu *)dialogItem;
		optionMenu -> SetValue (0);
	    }
	}  // end if not in common dialog
    }
    iterator.Finish ();

    // call base class reset function with proper point index
    vuCommonDialogClass::resetDialog ();

    dialogModified = vFALSE;
}


/*******************************************************************
*
*    vuLoadDialog 
*
*******************************************************************/

void vuWaveformDialogClass::vuLoadDialog (vchar *pointName)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    vchar		epics[FLDNAME_SZ + 1];
    vstr		*finalValue;
    vstr		*referenceString;
    vstr		*constantString;

    char		pvName [PVNAME_SZ + 1];
    DB_TABLE_PTR	lastTable;	// pointer to last table used
    DB_RECORD_PTR	pntRecord;	// codebase record ptr
    int			dbaseType;	// codebase field type
    int			intValue;
    unsigned int	uIntValue;
    long		longValue;
    unsigned long	uLongValue;
    double		doubleValue;
    DTYP_TYPE		dType;		// device type indicator

    vcharExportBounded (vcharScribeSystemExport (pointName),
			pvName,
			PVNAME_SZ);

    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
    pntRecord = DB_Seek (db -> vuPvmapTable, seekBuff);
    if (pntRecord == (DB_RECORD_PTR)0)
    {
	return;
    }
    else
    {
	int ptType;
	DB_GetFieldValue (pntRecord, db -> pvmapTypeField, dbValue, &dbaseType);
	ptType = vuGetPointTypeIndex ((char *)dbValue);
	// the point name in pvname field is NOT an WAVEFORM
	if (ptType != VU_WAVEFORM)
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_PTTYPEDIALOGERR,
				   pointName, ptType, VU_WAVEFORM);
	    vuClearPvNameField ();
	    return;
	}
	// the point name entered in pvname field has been deleted
	if (DB_IsDeleted (db -> vuPvmapTable))
	{
	    vuMessageDialogCenter ((vdialog *)this, VU_PTISDELETEDERR, pointName);
	    vuClearPvNameField ();
	    return;
	}
    }
    // call base class vuLoadDialog first
    vuCommonDialogClass::vuLoadDialog (pointName);

    // since the waveform can be a hardware point, 
    // find what we are for later use by getting the 
    // DTYP (device type) dictionary entry
    referenceString = vstrCloneScribed (vcharScribeSystem ("REFERENCE"));
    constantString = vstrCloneScribed (vcharScribeSystem ("CONSTANT"));
    xrefNode = (xrefNodeClass *)(dict -> GetElem (
			vnameGetGlobalScribed (vcharScribeSystem ("DTYP"))));
    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
    pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
    DB_GetFieldValue (pntRecord, xrefNode -> xrefField, dbValue, &dbaseType);
    if ( (vcharCompare ((const vchar *)dbValue, referenceString)) == 0)
	dType = VU_REFERENCE;
    else if ( (vcharCompare ((const vchar *)dbValue, constantString)) == 0)
	dType = VU_CONSTANT;
    else
	dType = VU_HARDWARE;
    vstrDestroy (referenceString);
    vstrDestroy (constantString);

    lastTable = NULL;

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0;
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    // only seek if table has changed
	    if (lastTable != xrefNode -> xrefTable)
		{
		    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
			pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
		}
	    lastTable = xrefNode -> xrefTable;
	    dialogItem = vuFindContainerItem (waveformContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    DB_GetFieldValue (pntRecord, xrefNode -> xrefField, dbValue, &dbaseType);
	    doubleValue = 0.0;
	    // GetField actually returns a number in the char array
	    if ( (dbaseType == 'F') || (dbaseType == 'N') )
	    {
		doubleValue = *((double *)dbValue);
	    }
	    switch (xrefNode -> xrefEPICSFieldType)
	    {
		case DBF_SHORT:
		    intValue = (int)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeInt (intValue));
		    break;

		case DBF_LONG:
		    longValue = (long)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeLong (longValue));
		    break;

		case DBF_ENUM:
		case DBF_USHORT:
		    uIntValue = (unsigned int)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeUInt (uIntValue));
		    break;

		case DBF_ULONG:
		    uLongValue = (unsigned long)doubleValue;
		    finalValue = vstrCloneScribed (vnumScribeULong (uLongValue));
		    break;

		case DBF_FLOAT:
		case DBF_DOUBLE:
		    // to output CONSTANT field, device type must be CONSTANT
		    if ( (strncmp (xrefNode -> xrefEpicsName, "CNST", FLDNAME_SZ)) == 0)
		    {
			if (dType != VU_CONSTANT)
			    continue;
		    }
		    finalValue = vstrCloneScribed (vnumScribeFloat (doubleValue));
		    break;

		case DBF_STRING:
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		case DBF_GBLCHOICE:
		case DBF_RECCHOICE:
		case DBF_CVTCHOICE:
		case DBF_DEVCHOICE:
		    if (dbaseType == 'L')
		    {
			int	logical;

			logical = (*((char*)dbValue) ? 
					xrefNode -> xrefTrue : 
					(xrefNode -> xrefTrue ^1));
			finalValue = vstrCloneScribed (vnumScribeInt (logical));
			break;
		    }
		    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		case DBF_INLINK:
		    if ( (strncmp (xrefNode -> xrefEpicsName, "INP ", FLDNAME_SZ)) == 0)
		    {
			// device type must be REFERENCE to output the reference
			if (dType == VU_REFERENCE)
			{
			    DB_GetFieldValue (pntRecord, xrefNode -> xrefRefField, 
					      dbValue, &dbaseType);
			    finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
			}
			else 
			{
			    finalValue = vstrCloneScribed (vcharScribeSystem (""));
			}
		    }
		    else
			finalValue = vstrCloneScribed (vcharScribeSystem (dbValue));
		    break;

		default:
		    continue;
	    }
	    if ( (dialogItem -> IsKindOf (&vtextitem::Kind)) == vTRUE)
	    {
		textItem = (vtextitem *)dialogItem;
		textItem -> SetTextScribed (vcharScribeSystem (""));
		textItem -> SetText (finalValue);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolToggle::Kind)) == vTRUE)
	    {
		toggleItem = (vcontrolToggle *)dialogItem;
		if (finalValue[0] == '0')
		    toggleItem -> SetValue (0);
		else
		    toggleItem -> SetValue (1);
		vstrDestroy (finalValue);
	    }
	    else if ( (dialogItem -> IsKindOf (&vcontrolOptionMenu::Kind)) == vTRUE)
	    {
		optionMenu = (vcontrolOptionMenu *)dialogItem;
		vuSetOptionMenuValue (optionMenu, finalValue);
		vstrDestroy (finalValue);
	    }
	}  // end if not in commondialog
    }
    iterator.Finish ();

    setDialogModify (vFALSE);
}


/*******************************************************************
*
*    vuSaveDialog 
*
*******************************************************************/

void vuWaveformDialogClass::vuSaveDialog (vchar *pointName)
{
    vuMainDialogClass	*mainDialog = getMainDialog ();
    dataVuApplication	*app  = mainDialog -> getApplication ();
    vuDatabaseClass	*db   = app -> getDatabaseClass ();
    vdict		*dict = db -> getXrefDict (pointType);
    vdictIterator	iterator;
    vname		*epicsName;
    xrefNodeClass	*xrefNode;
    vdialogItem		*dialogItem;
    vtextitem		*textItem;
    vcontrolToggle	*toggleItem;
    vcontrolOptionMenu	*optionMenu;
    const vchar 	*optionMenuValue;
    vchar		epics[FLDNAME_SZ + 1];
    vstr		*stringValue;
    vstr		*referenceString;
    vstr		*constantString;

    char		pvName [PVNAME_SZ + 1];
    char		vuCurrentBoard [VU_KEY_SIZE];
    DB_FIELD_PTR	putField;
    DB_TABLE_PTR	lastTable;	// pointer to last table used
    DB_RECORD_PTR	pntRecord;	// codebase record ptr
    int			dType;		// device type
    double		doubleValue;
    int			doSave;

    vcharExportBounded (vcharScribeSystemExport (pointName),
			pvName,
			PVNAME_SZ);

    // since the waveform can be a hardware point, 
    // find what we are for later use by getting the 
    // DTYP (BOARD) value from the dialog
    referenceString = vstrCloneScribed (vcharScribeSystem ("REFERENCE"));
    constantString = vstrCloneScribed (vcharScribeSystem ("CONSTANT"));
    vcharCopyScribedBounded (vcharScribeSystem ("DTYP"), epics, FLDNAME_SZ);
    epics[sizeof (epics) - 1] = 0;
    // find the BOARD (DTYP) field and get its value
    dialogItem = vuFindContainerItem (waveformContainer, epics);
    optionMenu = (vcontrolOptionMenu *)dialogItem;
    optionMenuValue = vuGetOptionMenuValue (optionMenu);
    vcharExport (vcharScribeSystemExport (optionMenuValue), (void *)vuCurrentBoard);
    if ( (vcharCompare (optionMenuValue, referenceString)) == 0)
	dType = VU_REFERENCE;
    else if ( (vcharCompare (optionMenuValue, constantString)) == 0)
	dType = VU_CONSTANT;
    else
	dType = VU_HARDWARE;
    vstrDestroy (referenceString);
    vstrDestroy (constantString);

    lastTable = NULL;

    iterator.Start (dict);
    while ( iterator.Next ())
    {
	epicsName = (vname *)iterator.GetKey ();
	vmemSet (epics, ' ', sizeof (epics));
	vcharCopyBounded (epicsName, epics, FLDNAME_SZ);
	epics[sizeof (epics) - 1] = 0;
	xrefNode  = (xrefNodeClass *)iterator.GetValue ();
	if ( (xrefNode -> getInCommonDialog ()) == vFALSE)
	{
	    doSave = vTRUE;
	    // only seek if table has changed
	    if (lastTable != xrefNode -> xrefTable)
		{
		    sprintf (seekBuff, "%-*.*s\0", PVNAME_SZ, PVNAME_SZ, pvName);
			pntRecord = DB_Seek (xrefNode -> xrefTable, seekBuff);
		}
	    lastTable = xrefNode -> xrefTable;

	    putField = xrefNode -> xrefField;

	    dialogItem = vuFindContainerItem (waveformContainer, epics);
	    if (dialogItem == NULL)
		continue;
	    if ( (dialogItem -> IsEnabled ()) == vFALSE)
		continue;
	    switch (xrefNode -> xrefEPICSFieldType)
	    {
		case DBF_SHORT:
		case DBF_LONG:
		case DBF_ENUM:
		case DBF_USHORT:
		case DBF_ULONG:
		case DBF_FLOAT:
		case DBF_DOUBLE:
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    doubleValue = atof ((const char *)dbValue);
		    if ( (strncmp (xrefNode -> xrefEpicsName, "CNST", FLDNAME_SZ)) == 0)
		    {
			if (dType != VU_CONSTANT)
			    doubleValue = 0;
		    }
		    DB_PutFieldValue (pntRecord, putField, (void *)&doubleValue);
		    doSave = vFALSE;
		    break;

		case DBF_STRING:
		    textItem = (vtextitem *)dialogItem;
		    stringValue = textItem -> GetTextAsString ();
		    vcharExport (vcharScribeSystemExport (stringValue), dbValue);
		    vstrDestroy (stringValue);
		    break;

		case DBF_GBLCHOICE:
		case DBF_RECCHOICE:
		case DBF_CVTCHOICE:
		case DBF_DEVCHOICE:
		{
		    if (xrefNode -> xrefField -> field_type == 'L')
		    {
			int	logical;
			toggleItem = (vcontrolToggle *)dialogItem;
			logical = toggleItem -> GetValue ();
			*((char *)dbValue) = (logical ? xrefNode -> xrefTrue : 
								xrefNode -> xrefTrue ^1);
			break;
		    }
		    optionMenu = (vcontrolOptionMenu *)dialogItem;
		    optionMenuValue = vuGetOptionMenuValue (optionMenu);
		    vcharExport (vcharScribeSystemExport (optionMenuValue), dbValue);
		    if ( (strncmp (xrefNode -> xrefEpicsName, "DTYP", FLDNAME_SZ)) == 0)
		    {
			// write to DEVTYPE in pvmap
			DB_PutFieldValue (pntRecord, xrefNode -> xrefField, dbValue);
		    }
		    break;
		}

		case DBF_INLINK:
			if ( (strncmp (xrefNode -> xrefEpicsName, "INP ", FLDNAME_SZ)) 
										== 0)
			{
			    // always write the boardtype from DTYP into the INP field
			    DB_PutFieldValue (pntRecord, putField, vuCurrentBoard);
			    // if we are a reference then write INP into the REFERENCE 
			    // field
			    if (dType == VU_REFERENCE)
			    {
				putField = xrefNode -> xrefRefField;
				textItem = (vtextitem *)dialogItem;
				stringValue = textItem -> GetTextAsString ();
				vcharExport (vcharScribeSystemExport (stringValue), 
									    dbValue);
				vstrDestroy (stringValue);
			    }
			    else
			    {
				// make sure to null out the field
				putField = xrefNode -> xrefRefField;
				vmemSet (dbValue, 0, VU_KEY_SIZE);
			    }
			}
			else
			{
			    textItem = (vtextitem *)dialogItem;
			    stringValue = textItem -> GetTextAsString ();
			    vcharExport (vcharScribeSystemExport (stringValue), 
									    dbValue);
			    vstrDestroy (stringValue);
			}
		    break;

		default:
		    continue;

	    }  // end xrefNode -> xrefEPICSFieldType

	    if (doSave == vTRUE)
	    {
		DB_PutFieldValue (pntRecord, putField, dbValue);
	    }

	}  // end if not in commondialog

    }  // end while iterator

    iterator.Finish ();

    // call base class vuSaveDialog
    vuCommonDialogClass::vuSaveDialog (pointName);
}


vkindTYPED_FULL_DEFINITION(vuWaveformContainerClass, vcontainer, "vuWaveformContainerClass");

/*******************************************************************
*
*    vuWaveformContainerClass override for constructor
*
*******************************************************************/

void vuWaveformContainerClass::LoadInit (vresource res)
{
    vapplication *app = vapplication::GetCurrent ();
    vresource pointDictRes = vresourceGet (app -> GetResources (), 
				vunames_vuPointDictionary);
    vresource realRes = vresourceGet (pointDictRes, vunames_vuWaveformContainer);

    vcontainer *btcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuWaveDTYPContainer);
    boardType = vuBoardTypeMenuClass::RequestSubLoad (btcontainer,
				vunames_DTYP);

    vcontainer *linecontainer = vcontainer::RequestSubLoad (this,
				vunames_vuWaveLINEContainer);
    line = filteredTextitem::RequestSubLoad (linecontainer,
				vunames_LINE);

    vcontainer *addrcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuWaveADDRContainer);
    addr = filteredTextitem::RequestSubLoad (addrcontainer,
				vunames_ADDR);

    vcontainer *cardcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuWaveCARDContainer);
    card = filteredTextitem::RequestSubLoad (cardcontainer,
				vunames_CARD);

    vcontainer *signalcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuWaveSIGNContainer);
    signal = filteredTextitem::RequestSubLoad (signalcontainer,
				vunames_SIGN);

    vcontainer *egutagcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuWaveEGUContainer);
    eguTag = filteredTextitem::RequestSubLoad (egutagcontainer,
				vunames_EGU);

    vcontainer *refcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuWaveINPContainer);
    referenceButton = vuPvNameButtonClass::RequestSubLoad (refcontainer,
				vunames_vuWaveINPButton);
    reference = vuFilteredTextitem::RequestSubLoad (refcontainer,
				vunames_INP);

    vcontainer *cnstcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuWaveCNSTContainer);
    constantValue = filteredTextitem::RequestSubLoad (cnstcontainer,
				vunames_CNST);

    onScan = vuCheckBoxClass::RequestSubLoad (this,
				vunames_OSCN);

    processInput = vuCheckBoxClass::RequestSubLoad (this,
				vunames_PRCI);

    maximizeSeverity = vuCheckBoxClass::RequestSubLoad (this,
				vunames_MAXS);

    rearm = vuCheckBoxClass::RequestSubLoad (this,
				vunames_RARM);

    vcontainer *mccontainer = vcontainer::RequestSubLoad (this,
				vunames_vuWaveMCNTContainer);
    monitorCount = filteredTextitem::RequestSubLoad (mccontainer,
				vunames_MCNT);

    vcontainer *nelmcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuWaveNELMContainer);
    elementCount = filteredTextitem::RequestSubLoad (nelmcontainer,
				vunames_NELM);

    vcontainer *preccontainer = vcontainer::RequestSubLoad (this,
				vunames_vuWavePRECContainer);
    precision = filteredTextitem::RequestSubLoad (preccontainer,
				vunames_PREC);

    vcontainer *dsplowcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuWaveLOPRContainer);
    displayLow = filteredTextitem::RequestSubLoad (dsplowcontainer,
				vunames_LOPR);

    vcontainer *dsphicontainer = vcontainer::RequestSubLoad (this,
				vunames_vuWaveHOPRContainer);
    displayHigh = filteredTextitem::RequestSubLoad (dsphicontainer,
				vunames_HOPR);

    vcontainer *ftvlcontainer = vcontainer::RequestSubLoad (this,
				vunames_vuWaveFTVLContainer);
    valueType = vuChoiceGblMenuClass::RequestSubLoad (ftvlcontainer,
				vunames_FTVL);

    vcontainer *ptsscontainer = vcontainer::RequestSubLoad (this,
				vunames_vuWavePTSSContainer);
    triggerCount = vuChoiceRecMenuClass::RequestSubLoad (ptsscontainer,
				vunames_PTSS);

    vcontainer *ratecontainer = vcontainer::RequestSubLoad (this,
				vunames_vuWaveRATEContainer);
    sampleRate = vuChoiceRecMenuClass::RequestSubLoad (ratecontainer,
				vunames_RATE);

    vcontainer::LoadInit (realRes);

    // store ptr to text field for vuPvNameButtonClass
    referenceButton -> storePvName (reference);

    // set filters
    line -> SetItemFilterFunc (numericFilter);
    addr -> SetItemFilterFunc (numericFilter);
    card -> SetItemFilterFunc (numericFilter);
    signal -> SetItemFilterFunc (numericFilter);

    reference -> SetItemFilterFunc (vuPvNameFilter);
    filteredText *refText = (filteredText *)reference -> GetTextData ();
    // remember ptr to textitem for later use by the filter
    refText -> SetUserPtr ((void *)reference);

    constantValue -> SetItemFilterFunc (floatFilter);
    monitorCount -> SetItemFilterFunc (numericFilter);
    elementCount -> SetItemFilterFunc (numericFilter);
    precision -> SetItemFilterFunc (numericFilter);
    displayLow -> SetItemFilterFunc (floatFilter);
    displayHigh -> SetItemFilterFunc (floatFilter);

    // build popup menus
    boardType    -> vuBuildBoardTypeMenu (VU_WAVEFORM);
    valueType   -> vuBuildOptionMenu (GBL_FTYPE);
    triggerCount -> vuBuildOptionMenu (VU_WAVEFORM, REC_WAVEFORM_PRETRIGGER);
    sampleRate -> vuBuildOptionMenu (VU_WAVEFORM, REC_WAVEFORM_RATE);
}

/*******************************************************************
*
*    vuWaveformContainerClass override for destructor
*
*******************************************************************/

void vuWaveformContainerClass::Destroy (void)
{
    vcontainer::Destroy ();
}

