/*
    This is the Network class for all dataVu network operations. These include:

	1. Online point additions.
	2. Online point deletions.
	3. Remote unit reset.
	4. dbClient notification of point additions/modifications
*/
#include <vport.h>
#include vkindHEADER
#include vbuttonHEADER
#include vlistviewHEADER
#include vlistitemHEADER
#include vliststrHEADER
#include vstdargHEADER
#include vlistitemHEADER
extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include	<time.h>
#include	<stdio.h>
#include	<signal.h>
#include	<stdlib.h>
#ifndef WIN32
#include	<sys/time.h>
#include	<sys/types.h>
#include	<sys/socket.h>
#include	<sys/ioctl.h>
#include	<sys/fcntl.h>
#include	<netinet/in.h>
#include	<netinet/tcp.h>
#include	<netdb.h>
#include	<error.h>
/* there is a man page for this, but not in any .h
 * since C++ needs the prot, I have to resort to this
 * NOTE: on later solaris, it is prot'd in /usr/include/unistd.h
 * ems 22-jul-99
 */
extern int gethostname (char *, int);
#else
#include	<io.h>
#include	<winsock.h>
typedef char * caddr_t;
#endif WIN32
#include	"dbFldTypes.h"
#include	"dbDefs.h"
#include	"DBDefines.h"
#include	"DBCVerifyDefs.h"
#include	"DBProto.h"
#include	"dbRecDes.h"
#include	"dbRecType.h"
#include	"redmsgs.h"
#ifdef PRIV_AUDIT
#include	"inet_ntoa.h"
#include	"auditSystem.h"
extern AUDIT_ID	auditLog;	/* the system log */

#include "sslShared.h"
#endif PRIV_AUDIT
}

#include	"vunames.h"
#include	"dataVu.hxx"
#include	"vuMainDialogClass.hxx"
#include	"vunetwork.hxx"

extern char *sys_errlist[ ];
extern int errno;
static vdialogItemNoteProc byebye(vdialogItem *item,vevent *ev);

static struct timeval	purgeTimeOut = {10, 0};
static struct timeval	resetTimeOut = {10, 0};

#ifdef USE_UDP
vuNetwork::vuNetwork(vuMainDialogClass *md,char *dest)
{
    mainDialog = md;
    dataVuApplication *dataVuApp = mainDialog->getApplication();
    vuRes = dataVuApp->getDataVuvresource();
    vuNetError = vFALSE;
    strcpy(destName,dest);
    vuSocket = INVALID_SOCKET;
    if(ent = gethostbyname(destName))			/* get INET address for this IOC */
    {
	addr = *(long *)*ent -> h_addr_list;
	destAddr.sin_family = AF_INET;
	destAddr.sin_addr.s_addr	= addr;
	destAddr.sin_port	= htons(DBSERVER_UDP_PORT);
	vuSocket = socket(AF_INET,SOCK_DGRAM,0);
    }
    else
    {
	vuNetError = vTRUE;
    }
}
#endif USE_UDP

vuNetwork::vuNetwork(vuMainDialogClass *md)	// obtain socket to "localhost" and connect
{
    int stat;

    mainDialog = md;
    dataVuApplication *dataVuApp = mainDialog->getApplication();
    vuRes = dataVuApp->getDataVuvresource();
    vuNetError = vFALSE;
    rxCount = 0;
    rxNextChar = 0;
    gethostname(destName,VU_NETWORK_MSG_SIZE);		/* get my network name */
    vuSocket = INVALID_SOCKET;
    if(ent = gethostbyname(destName))			/* get INET address for me */
    {
	addr = *(long *)*ent -> h_addr_list;
#ifdef PRIV_AUDIT
	{
	    SSL *newConn;
	    if ((newConn = connect_ssl_connection_timed (addr, DBSERVER_UDP_PORT, 5)) == NULL)
	    {
		sprintf(outMsg,"Can't open SSL connection");
		mainDialog->vuMainMessageCenter(vunames_NetworkCantConnect,destName,outMsg);
		vuSocket = INVALID_SOCKET;
		vuNetError = vTRUE;
	    }
	    else
		vuSocket = SSL_get_fd (newConn);
	}
#else PRIV_AUDIT
	destAddr.sin_family = AF_INET;
	destAddr.sin_addr.s_addr = addr;
	destAddr.sin_port = htons(DBSERVER_UDP_PORT);
	vuSocket = socket(AF_INET,SOCK_STREAM,0);
	if((stat = connect(vuSocket,(struct sockaddr *)&destAddr,sizeof(struct sockaddr))) < 0)
	{
#ifdef WIN32
	    sprintf(outMsg,"winsock error # %d",WSAGetLastError());
#else WIN32
	    sprintf(outMsg,"%s",sys_errlist[errno]);
#endif WIN32
	    mainDialog->vuMainMessageCenter(vunames_NetworkCantConnect,destName,outMsg);
#ifdef WIN32
	    closesocket (vuSocket);
#else WIN32
	    close(vuSocket);
#endif WIN32
	    vuSocket = INVALID_SOCKET;
	    vuNetError = vTRUE;
	}
#endif PRIV_AUDIT
    }
    else
    {
	vuNetError = vTRUE;
    }
}
//
// close socket when instance is deleted
//
vuNetwork::~vuNetwork()
{
    if(vuSocket != INVALID_SOCKET)
    {
#ifdef PRIV_AUDIT
	sslSharedCloseConn (sslSharedGetConn(vuSocket));
#else PRIV_AUDIT
#ifdef WIN32
	closesocket (vuSocket);
#else WIN32
	close(vuSocket);
#endif WIN32
#endif PRIV_AUDIT
    }
}
//
// member functions to actually do some work
//
int vuNetwork::checkError(void)					    // return error flag
{
    return(vuNetError);
}

#ifdef USE_UDP
int vuNetwork::sendReset(void)		   // reset "dest" IOC
{
    short notify = htons(DBC_MSG_RESETNOTIFY);
    struct fd_set readFds;
    short msgType;
    struct sockaddr_in iocaddr;
    int len,stat;

    len = sizeof(struct sockaddr);
    vuNetError = vFALSE;
    if(sendto(vuSocket,(char *)&notify,2,0,(struct sockaddr *)&destAddr,sizeof(destAddr)) < 0)
    {
#ifdef WIN32
	sprintf(outMsg,"winsock error # %d",WSAGetLastError());
#else
	sprintf(outMsg,"%s",sys_errlist[errno]);
#endif
	mainDialog->vuMainMessageCenter(vunames_NetworkCantSend,destName,outMsg);
	vuNetError = vTRUE;
    }
    else
    {
	mainDialog->vuMainMessageCenter(vunames_NetworkIocNotified,destName);
/*
Wait for IOC to reply to reset message.
*/
	FD_ZERO(&readFds);
	FD_SET(vuSocket,&readFds);
	stat = select(vuSocket+1,&readFds,NULL,NULL,&resetTimeOut);
	if(stat == 0)
	{
#ifdef WIN32
	    sprintf(outMsg,"winsock error # %d",WSAGetLastError());
#else WIN32
	    sprintf(outMsg,"%s",sys_errlist[errno]);
#endif WIN32
	    mainDialog->vuMainMessageCenter(vunames_NetworkResetAckTimeout,destName,outMsg);
	    vuNetError = vTRUE;
	}
	else if(stat < 0)
	{
#ifdef WIN32
	    sprintf(outMsg,"winsock error # %d",WSAGetLastError());
#else WIN32
	    sprintf(outMsg,"%s",sys_errlist[errno]);
#endif WIN32
	    mainDialog->vuMainMessageCenter(vunames_NetworkSelectError,outMsg);
	}
	else
	{
	    len = sizeof(iocaddr);
	    recvfrom(vuSocket,(char *)&msgType,2,0,(struct sockaddr *)&iocaddr,&len);
	    mainDialog->vuMainMessageCenter(vunames_NetworkResetAck,destName);
	}
    }
    return(0);
}
#else USE_UDP
int vuNetwork::sendReset(char *iocname)	   // reset "dest" IOC
{
    char msg[80];
    short msgType;
    fd_set readFds;
    int inetAddress;

    if((ent = gethostbyname(iocname)))
    {
	inetAddress = *(long *)*ent -> h_addr_list;
	if(sendToLocalClient(htons(UTIL_MSG_RESETNOTIFY),(char *)&inetAddress,sizeof(inetAddress)) == 0)
	{
	    sprintf(outMsg,"%s notified for RESET\n",iocname);
	    mainDialog->vuMainMessageCenter(vunames_NetworkCantSend,destName,outMsg);
	    /*
	    Wait for dbClient to reply to reset message.
	    */
	    FD_ZERO(&readFds);
	    FD_SET(vuSocket,&readFds);
	    rxCount = 0;					/* Initialize rxBuffer indicies */
	    rxNextChar = 0;
	    if (select (vuSocket+1, &readFds, NULL, NULL, &resetTimeOut) > 0)
	    {
		recvFromLocalClient();			/* get as many bytes as possible */
		readTcp((char *)&msgType,sizeof(msgType));/* read first 2 for message type */
		if(ntohs(msgType) == DBC_MSG_RESETACK)
		    sprintf(msg,"IOC %s has acknowledged a Reset\0",iocname);
		else
		    sprintf(msg,"Invalid message...IOC %s may not have been Reset\0",iocname);
	    }
	    else
		sprintf(msg,"IOC %s may not have been Reset(timeout)\0",iocname);
	    mainDialog->vuMainMessageCenter(vunames_NetworkCantSend,destName,outMsg);
	}
	else
	{
	    sprintf(msg,"IOC %s may not have been Reset(I/O error)\0",iocname);
	    mainDialog->vuMainMessageCenter(vunames_NetworkCantSend,destName,outMsg);
	}
    }
    else
    {
	sprintf(msg,"IOC %s not in network database\n\0",iocname);
	mainDialog->vuMainMessageCenter(vunames_NetworkCantSend,destName,outMsg);
    }
    return(0);
}
#endif USE_UDP
/*
	Tell destination IOC to delete the given point. Wait for reply. If error
	then state reason. If other points reference this one display them in a
	dialog box and show it to the user.
*/
#ifdef USE_UDP
void vuNetwork::sendDelete(char *pvname)			 // delete "pvname" from "dest"
{
    int stat;
    int len;
    char inMsg[VU_NETWORK_MSG_SIZE],*pinMsg;
    char *poutMsg;
    struct fd_set readFds;
    struct	_notify
    {
	short	type;
	char	name[PVNAME_SZ+1];
    } notify;
    short	msgType,msgStatus,msgCount;

    pinMsg = &inMsg[0];
    poutMsg = &outMsg[0];
    notify.type = htons(DBC_MSG_OLDELETE);
    strcpy(notify.name,pvname);
    if((stat = sendto(vuSocket,(char *)&notify,sizeof(struct _notify),0,(struct sockaddr *)&destAddr,(int)sizeof(destAddr))) < 0)
    {
#ifdef WIN32
	sprintf(outMsg,"winsock error # %d",WSAGetLastError());
#else WIN32
	sprintf(outMsg,"%s",sys_errlist[errno]);
#endif WIN32
	mainDialog->vuMainMessageCenter(vunames_NetworkCantSend,destName,outMsg);
	vuNetError = vTRUE;
	return;
    }
/*
Wait here for the incomming status message from the IOC. The part
of the message beyond the message type is dependant on the message
type.
*/
    len = sizeof(destAddr);
    FD_ZERO(&readFds);
    FD_SET(vuSocket,&readFds);
    stat = select(vuSocket + 1, &readFds, NULL, NULL, &purgeTimeOut);
    if(stat == 0)
    {
	mainDialog->vuMainMessageCenter(vunames_NetworkSelectTimeout);
    }
    else if(stat < 0)
    {
	// io error
#ifdef WIN32
	sprintf(outMsg,"winsock error # %d",WSAGetLastError());
#else WIN32
	sprintf(outMsg,"%s",sys_errlist[errno]);
#endif WIN32
	mainDialog->vuMainMessageCenter(vunames_NetworkSelectError,outMsg);
    }
    else
    {
	recvfrom(vuSocket,(char *)pinMsg,VU_NETWORK_MSG_SIZE,0,(struct sockaddr *)&destAddr,&len);
	msgType = ntohs(*((short *)pinMsg));
	pinMsg += 2;
	msgStatus = ntohs(*((short *)pinMsg));
	switch(msgStatus)
	{
	    case DELETION_OK:
		mainDialog->vuMainMessageCenter(vunames_NetworkDeleteOK,destName,pvname);
		break;
	    case DELETION_ONSCAN:
		mainDialog->vuMainMessageCenter(vunames_NetworkNoDelOnscan,destName,pvname);
		break;
	    case DELETION_OPEN:
		mainDialog->vuMainMessageCenter(vunames_NetworkNoDelOpen,destName,pvname);
		break;
	    case DELETION_NOREC:
		mainDialog->vuMainMessageCenter(vunames_NetworkNoRecord,destName,pvname);
		break;
	    case DELETION_REF:
		mainDialog->vuMainMessageCenter(vunames_NetworkNoDelRef,destName,pvname);
		pinMsg += 2;
		msgCount = ntohs(*((short *)pinMsg));
		pinMsg += 2;
		poutMsg += strlen(outMsg);
		vresource findRes = vresourceGet (vuRes, vunames_vuInUseDialog);
		vuInUseDialogClass *inuseDialog = new vuInUseDialogClass(findRes);
		// loop here to extract point names from the stream and append
		// them to the list
		while(--msgCount >= 0)
		{
		    inuseDialog->addToList(pinMsg);
		    pinMsg += PVNAME_SZ; 
		}
		inuseDialog->Open();
		vevent::Process();
		delete inuseDialog;
		break;
	}
    }
}
#else USE_UDP
void vuNetwork::sendDelete(char *iocname, char *pvname)		 // delete "pvname" from "dest"
{
    char	paddedName[PVNAME_SZ+1];
    int		i, inetAddress[2], stat;
    char	index;
    fd_set	readFds;
    DBC_OLDELETE del;
    DBC_DELSTATUS delStat;
    short	msgType, msgCount, msgStatus, refs;

    if((ent = gethostbyname(iocname)))
    {
	del.originatorFd = *(long *)*ent -> h_addr_list;
	inetAddress[i] = del.originatorFd;
	strncpy (del.nameToDelete, pvname, PVNAME_SZ);
	if((sendToLocalClient(htons(UTIL_MSG_OLDELETE), (char *)&del, sizeof(DBC_OLDELETE)) < 0))
	{
	    sprintf (outMsg, "Can't send delete message to dbClient\n\0");
	    mainDialog->vuMainMessageCenter(vunames_NetworkNoRecord,destName,outMsg);
	}
    }
    /*
    Wait here for the incomming status message from the IOC. The part
    of the message beyond the message type is dependant on the message
    type.
    */
    rxCount = 0;					/* Initialize rxBuffer indicies */
    rxNextChar = 0;
    while(1)
    {
	FD_ZERO (&readFds);
	FD_SET (vuSocket, &readFds);
	stat = select(vuSocket + 1, &readFds, NULL, NULL, &purgeTimeOut);
	if(stat == 0)
	{
	    sprintf(outMsg,"Receive from dbClient for %s deletion timed out.\n\0",pvname);
	    mainDialog->vuMainMessageCenter(vunames_NetworkNoRecord,destName,outMsg);
	    rxCount = 0;					/* Initialize rxBuffer indicies */
	    rxNextChar = 0;
	    return;
	}
	if(stat < 0)
	{
	    sprintf(outMsg,"Error from select: %s\n\0",sys_errlist[errno]);
	    mainDialog->vuMainMessageCenter(vunames_NetworkNoRecord,destName,outMsg);
	    rxCount = 0;					/* Initialize rxBuffer indicies */
	    rxNextChar = 0;
	    return;
	}
	recvFromLocalClient();
	if(haveCompleteMsg())
	    break;
    }
    readTcp((char *)&msgType,2);
    if(ntohs(msgType) != DBC_MSG_DELSTATUS)
    {
	sprintf(outMsg,"Invalid delete response\n\0");
	mainDialog->vuMainMessageCenter(vunames_NetworkNoRecord,destName,outMsg);
	rxCount = 0;					/* Initialize rxBuffer indicies */
	rxNextChar = 0;
	return;
    }
    readTcp((char *)&delStat,sizeof(DBC_DELSTATUS));
    msgStatus = ntohs(delStat.deletionStatus);
    refs = ntohs(delStat.numRefs);
    if(delStat.originatorFd == inetAddress[0])
	index = 0;
    else if(delStat.originatorFd == inetAddress[1])
	index = 1; 
    switch(msgStatus)
    {
	case DELETION_OK:
	    mainDialog->vuMainMessageCenter(vunames_NetworkDeleteOK,iocname,pvname);
	    break;
	case DELETION_ONSCAN:
	    mainDialog->vuMainMessageCenter(vunames_NetworkNoDelOnscan,iocname,pvname);
	    break;
	case DELETION_OPEN:
	    mainDialog->vuMainMessageCenter(vunames_NetworkNoDelOpen,iocname,pvname);
	    break;
	case DELETION_NOREC:
	    mainDialog->vuMainMessageCenter(vunames_NetworkNoRecord,iocname,pvname);
	    break;
	case DELETION_REF:
	    mainDialog->vuMainMessageCenter(vunames_NetworkNoDelRef,iocname,pvname);
	    readTcp((char *)&msgCount, sizeof(msgCount));
	    vresource findRes = vresourceGet (vuRes, vunames_vuInUseDialog);
	    vuInUseDialogClass *inuseDialog = new vuInUseDialogClass(findRes);
	    // loop here to extract point names from the stream and append
	    // them to the list
	    while(--msgCount >= 0)
	    {
		readTcp (paddedName, PVNAME_SZ);
		inuseDialog->addToList(paddedName);
	    }
	    inuseDialog->Open();
	    vevent::Process();
	    delete inuseDialog;
	    break;
    }
}
#endif USE_UDP

/*
	Send on line addition/modification message to destination IOC
*/
#ifdef USE_UDP
int vuNetwork::sendAdd(void)			     // tell "dest" that there are new points
{
    short notify = htons(DBC_MSG_OLDNOTIFY);

    if(sendto(vuSocket,(char *)&notify,2,0,(struct sockaddr *)&destAddr,sizeof(destAddr)) < 0)
    {
	vuNetError = vTRUE;
#ifdef WIN32
	sprintf(outMsg,"winsock error # %d",WSAGetLastError());
#else WIN32
	sprintf(outMsg,"%s",sys_errlist[errno]);
#endif WIN32
	mainDialog->vuMainMessageCenter(vunames_NetworkCantSend,destName,outMsg);
    }
    else
    {
	vuNetError = vFALSE;
	mainDialog->vuMainMessageCenter(vunames_NetworkAddNotify,destName);
    }
    return(0);
}
#else
int vuNetwork::sendAdd(char *iocname)		     // tell "iocname" that there are new points
{
    int inetAddress;
    char msg[80];

    /*
    Execute special CalISO script here
    GCT 21-Jan-1999
    */
    system("/tis4000/tasks/GenIOCNames.sh");
    if((ent = gethostbyname(iocname)))
    {
	inetAddress = *(long *)*ent -> h_addr_list;
	if(sendToLocalClient (htons(UTIL_MSG_OLDNOTIFY), (char *)&inetAddress, sizeof(inetAddress)) < 0)
	{
	    sprintf(msg,"Error notifying dbClient about online\npoint addition for %s\n-- %s\n\0",iocname,sys_errlist[errno]);
	    mainDialog->vuMainMessageCenter (vunames_NetworkCantSend, iocname, msg);
	}
	else
	{
	    mainDialog->vuMainMessageCenter(vunames_NetworkAddNotify, iocname);
	}
    }
    else
    {
	sprintf(msg,"IOC %s not in network database\n\0",iocname);
	mainDialog->vuMainMessageCenter (vunames_NetworkCantSend, iocname, msg);
    }
    return 0;
}
#endif USE_UDP

void vuNetwork::sendNotify(char *name)	    						// notify dbClient
{
    if (vuSocket == INVALID_SOCKET)
	return;
    /*
	we have a socket at this point so use it.
    */
    short msgType = htons(UTIL_MSG_VERIFYNOW);
#ifdef PRIV_AUDIT
    if (SSL_write (sslSharedGetConn (vuSocket), (char *)&mType, sizeof(mType)) == sizeof(mType))
    {
	if(size != 0)
	{
	    if (SSL_write (sslSharedGetConn (vuSocket), (char *)msg, size) != size)
	    {
		sprintf(outMsg,"SSL write error");
		mainDialog->vuMainMessageCenter(vunames_NetworkCantSend,destName,outMsg);
		vuNetError = vTRUE;
	    }
	}
    }
    else
    {
	sprintf(outMsg,"SSL write error");
	mainDialog->vuMainMessageCenter(vunames_NetworkCantSend,destName,outMsg);
	vuNetError = vTRUE;
    }
#else PRIV_AUDIT
    if(send(vuSocket,(char *)&msgType,2,0) <= 0)
    {
#ifdef WIN32
	sprintf(outMsg,"winsock error # %d",WSAGetLastError());
#else
	sprintf(outMsg,"%s",sys_errlist[errno]);
#endif
	mainDialog->vuMainMessageCenter(vunames_NetworkCantSend,destName,outMsg);
	vuNetError = vTRUE;
	return;
    }
    sprintf(outMsg,"%-*.*s\0",PVNAME_SZ,PVNAME_SZ,name);
    if(send(vuSocket,(char *)outMsg,PVNAME_SZ,0) <= 0)
    {
#ifdef WIN32
	sprintf(outMsg,"winsock error # %d",WSAGetLastError());
#else
	sprintf(outMsg,"%s",sys_errlist[errno]);
#endif
	mainDialog->vuMainMessageCenter(vunames_NetworkCantSend,destName,outMsg);
	vuNetError = vTRUE;
    }
#endif PRIV_AUDIT
}

#ifndef USE_UDP

/*
  This function is a generic one used to communicate with dbClient. This is because dataVu
  will no longer talk directly with any IOC/RTU via UDP. The messages which will be
  transmitted from here used to go directly to the IOC/RTU. They will now go to dbClient
  first and then relayed to the IOC/RTU.

*/
int vuNetwork::sendToLocalClient (short mType, char *msg, int size)
{
    if (vuSocket == INVALID_SOCKET)
	return(-1);
    /*
	we have a socket at this point so use it.
    */
#ifdef PRIV_AUDIT
    if (SSL_write (sslSharedGetConn (vuSocket), (char *)&mType, sizeof(mType)) == sizeof(mType))
    {
	if(size != 0)
	{
	    if (SSL_write (sslSharedGetConn (vuSocket), (char *)msg, size) != size)
	    {
		sslSharedCloseConn (sslSharedGetConn (vuSocket));
		vuSocket = INVALID_SOCKET;
		return(-1);
	    }
	}
    }
    else
    {
	sslSharedCloseConn (sslSharedGetConn (vuSocket));
	vuSocket = INVALID_SOCKET;
	return(-1);
    }
#else PRIV_AUDIT
    /* need to check error return */
    if(send(vuSocket,(char *)&mType,sizeof(mType),0) == sizeof(mType))
    {
	if(size != 0)
	{
	    if(send(vuSocket,(char *)msg,size,0) != size)
	    {
#ifdef WIN32
		closesocket (vuSocket);
#else WIN32
		close(vuSocket);
#endif WIN32
		vuSocket = INVALID_SOCKET;
		return(-1);
	    }
	}
    }
    else
    {
#ifdef WIN32
	closesocket (vuSocket);
#else WIN32
	close(vuSocket);
#endif WIN32
	vuSocket = INVALID_SOCKET;
    }
#endif PRIV_AUDIT
return(0);
}

/*
  Read the vuSocket. Use SSL when needed. -1 is an error
*/
int vuNetwork::recvFromLocalClient (void)
{
    int readSize,stat;

#ifdef PRIV_AUDIT
    SSL *conn = sslSharedGetConn (vuSocket);
#endif
    if(vuSocket == INVALID_SOCKET)
	return(-1);
    errno = 0;
    if(rxNextChar)
    {
	if(rxCount)
	{
	    memcpy (&rxBuffer[0], &rxBuffer[rxNextChar], rxCount);
	}
	rxNextChar = 0;
    }
    readSize = DBC_NETMSG_SIZE - rxCount;
#ifdef PRIV_AUDIT
    if ((stat = SSL_read (conn, rxBuffer, readSize)) < 0)
    {
	int status = SSL_get_error (conn, stat);
	long err = ERR_get_error();
	if (err != 0)
	    return(-1);
	switch (status)
	{
	    case SSL_ERROR_NONE:
	    case SSL_ERROR_WANT_READ:
	    case SSL_ERROR_WANT_WRITE:
	    case SSL_ERROR_WANT_X509_LOOKUP:
	    case SSL_ERROR_WANT_CONNECT:
		/* benign, just want to write/read for state machine */
		return(-1);
	    case SSL_ERROR_SSL:
		if (errno == EWOULDBLOCK)
		    return(-1);
	    case SSL_ERROR_SYSCALL:
	    case SSL_ERROR_ZERO_RETURN:
	    default:
		return(-1);
	}
    }
#else PRIV_AUDIT
    if ((stat = recv(vuSocket, rxBuffer, readSize, 0)) <= 0)
    {
	return(-1);		/* Drop the whole thing on error or 0 read */
    }
#endif PRIV_AUDIT
    rxCount += stat;		/* number actually read into buffer */
    return(stat);
}

int  vuNetwork::readTcp(char *buff,int nbytes)
{
    memcpy (buff, &rxBuffer[rxNextChar], nbytes);
    rxNextChar += nbytes;
    rxCount -= nbytes;
    return (nbytes);
}

/*
  Because of the volume of messages from dbClient it is neccassary to impliment
  some sort of complete message detection similar to dbClient. This will be used by
  DBC_MSG_RESETACK & DBC_MSG_DELSTATUS for now
*/
int  vuNetwork::haveCompleteMsg()
{
    short	mType;
    int		mSize;
    DBC_DELSTATUS del;

    if(rxCount < sizeof(mType))
	return(0);
    memcpy (&mType, &rxBuffer[rxNextChar], sizeof(mType));
    mSize = DBC_NETMSG_SIZE+1;
    /*
    switch on the message type. All message are considered complete if the number of bytes
    in the buffer (rxCount) is greater that or equal to the total message size PLUS
    the size of the message type.
    */
    switch(ntohs(mType))
    {
	case DBC_MSG_RESETACK:
	    mSize = sizeof(mType);
	    break;
	case DBC_MSG_DELSTATUS:
	    memcpy ((char *)&del, &rxBuffer[rxNextChar], sizeof(DBC_DELSTATUS));
	    mSize = sizeof(mType) + sizeof(DBC_DELSTATUS) + (ntohs(del.numRefs) * PVNAME_SZ);
	    break;
    }
    if(mSize > rxCount)
	return(0);
    else
	return(mSize);	/* return message size. this includes the message type */
}

#endif USE_UDP

vkindTYPED_FULL_DEFINITION(vuInUseDialogClass, vdialog, "vuInUseDialogClass");

/*******************************************************************
*
*    vuInUseDialogClass 
*	LoadInit, constructor
*
*******************************************************************/

void vuInUseDialogClass::LoadInit(vresource res)
{
    iocList = vlistitem::RequestSubLoad(this,vunames_vuInUseList);
    dismissButton = vbutton::RequestSubLoad(this,vunames_vuInUseDismiss);
    dismissButton->SetObserveDialogItemProc((vdialogItemNoteProc)byebye);
// call base class loadInit to finish the job
    vdialog::LoadInit(res);
}
void vuInUseDialogClass::addToList(char *item)
{
    iocList->AppendScribed(vcharScribeSystem(item));
}
static vdialogItemNoteProc byebye(vdialogItem *item,vevent *ev)
{
    item->DetermineDialog()->Destroy();
    return(0);
}
