/*iocInit.c,v 1.1.1.1 1999/08/03 14:42:34 tis4000 Exp */

/* iocInit.c	ioc initialization */ 
/* share/src/db @(#)iocInit.c	1.19     8/12/91 */
/*
 *      Author:		Marty Kraimer
 *      Date:		06-01-91
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  07-20-91	rac	print release data; set env params
 * .02	08-06-91	mrk	parm string length test changed to warning
 *				with continue
 * .03	04-22-94	ems	in initDatabase check both init and in2 flags
 *				before calling points init_record
 * .04	05-12-95	ems	comment out call to coreRelease(), stupid and not needed
 * .05	07-10-96	ems	added SysAlmStr[80]
 */

#include	<vxWorks.h>
#include	<types.h>
#include	<lstLib.h>
#include	<memLib.h>
#include	<sysLib.h>
#include	<symLib.h>
#include	<sysSymTbl.h>	/* for sysSymTbl*/
#include	<a_out.h>	/* for N_TEXT */
#include	<stdioLib.h>
#include	<strLib.h>
#include	<ctype.h>
#include	<taskLib.h>
#include	<stdio.h>
#include	<fioLib.h>
#include	<ioLib.h>
#include	<wdLib.h>	/* required by callbacks.h */
#include	<logLib.h>
#include	<netinet/in.h>

/* used to be a compiler command line option */
#define	MODULE_TYPES_INIT

#include	"fast_lock.h"
#include	"choice.h"
#include	"dbDefs.h"
#include	"dbCommon.h"
#include	"link.h"
#include	"db_field_log.h"
#include	"dbAccess.h"
#include	"dbFldTypes.h"
#include 	"dbRecDes.h"
#include 	"dbRecType.h"
#include	"dbRecords.h"
#include	"devSup.h"
#include	"drvSup.h"
#include	"errMdef.h"
#include	"recSup.h"
#include	"envDefs.h"
#include	"scanLock.h"
#include	"task_params.h"
#include	"ts_driv.h"
#include	"dbserver.h"
#include	"restable.h"
#include	"resmgr.h"
#include	"callbacks.h"
#include	"dbScan.h"
#include	"error.h"
#include	"epicsEnvParams.h"
#include	"resGroupTask.h"
#include	"serverControl.h"
#include	"rsrv.h"
#include	"cadef.h"

typedef unsigned char	UTINY;

   /* global buffer pointer to be used by process stats routines and rec types,
      see .../share/src/rec/procstat.c
   */
int	firstInitPass = 0;
long *process_cnts = NULL;

/* global buffer pointer for stringin alarm strings when
   board type is SYSINFO, it is init'd to NULL in recStringin:initialize()
*/

char	SysAlmStr[80];

static initialized=FALSE;

struct driver
{
   int driver_loaded;
   char driver_name[64];
};
static struct driver *drivers;

/* define forward references*/
static long initProcessStats();
static long initDrvSup();
static long initRecSup();
static long initDevSup();
static long getResources (char *fname);

/*
A set of three structures so we can stall the scan
tasks when necessary.
*/

FAST_LOCK	LOCK_PERIODIC;
FAST_LOCK	LOCK_EVENT;
FAST_LOCK	LOCK_IOINTR;


int iocInit (pfilename,pResourceFilename)
char * pfilename;
char * pResourceFilename;
{
    long status;
    int	 priority;
    char name[40];
    long rtnval;
    void (*pdbUserExit)();
    UTINY type;

    if(initialized) {
	printf("iocInit can only be called once\n");
	return(-1);
    }
    /*coreRelease();*/
    epicsSetEnvParams();
    if ((status=getResources(pResourceFilename))) {
	printf("iocInit aborting because getResources failed\n");
	return(-1);
    }
    printf("getResources completed\n");
    FASTLOCKINIT(&LOCK_PERIODIC);
    FASTLOCKINIT(&LOCK_EVENT);
    FASTLOCKINIT(&LOCK_IOINTR);
    initialized = TRUE;
    /* enable interrupt level 5 and 6 */
#if CPU_FAMILY==MC680X0
    sysIntEnable(5);
    sysIntEnable(6);
#endif
#if !(defined (PRIV_AUDIT))
    /* time functions must be started earlier for token/secure systems! */
    ts_init(); printf("Time Stamp Driver Initialized\n");
#endif PRIV_AUDIT
    if(dbServerInit()==0) printf("dbServer Initialized\n");
    redTableAlloc();
    drivers = NULL;
    if(initDrvSup()==0) printf("Drivers Initialized\n");
    if(initRecSup()==0) printf("Record Support Initialized\n");
    if(initDevSup()==0) printf("Device Support Initialized\n");
    if (drivers)
	free (drivers);
    if (initProcessStats() == 0) printf ("Process stats Initialized\n");

    /*
    During init, we may create traffic for dbserver. That will get
    lost in the finite message queue (see redqlib.c) if our priority 
    is more important than his. Reduce our priority for the amount of
    time we spend in initDatabase. Put it back when we are done.
    */

    if ((taskPriorityGet(taskIdSelf(), &priority)) == ERROR)
	printf("taskPG error in iocInit\n");
    if ((taskPrioritySet(taskIdSelf(), DBSERVER_PRI + 1)) == ERROR)
	printf("taskPS 1 error in iocInit\n");
    if(initDatabase(0)==0) printf("Database Initialized\n");
/* Not available at load time! (must move to startup.cmd
    initResGroupScan();
*/

    /*
    Indicate that we've made at least one trip through the database.
    */

    firstInitPass = 1;

    if ((taskPrioritySet(taskIdSelf(), priority)) == ERROR)
	printf("taskPS 2 error in iocInit\n");

    if(startRedundancyTasks()==0) printf("Redundancy tasks Initialized\n");

#if !(defined (PRIV_AUDIT))
    if(resMgrInit()==0) printf("resMgr Initialized\n");
#endif PRIV_AUDIT

    /* if user exit exists call it */
    strcpy(name,"_");
    strcat(name,"dbUserExit");
    rtnval = symFindByName(sysSymTbl,name,(char **)&pdbUserExit,&type);
    if(rtnval == OK && ((type & N_TEXT) != 0)) {
	(*pdbUserExit)();
	printf("User Exit was called\n");
    }
    scan_init(); printf("Scanners Initialized\n");
    rsrv_init(); printf("Channel Access Servers Initialized\n");
    taskDelay(2);	/* give rsrvTask a chance to get going */
    caClientInit (); printf("Channel Access Client Services Initialized\n");
    printf("iocInit: All initialization complete\n");

    return(0);
}

/* #include	<module_types.h> */


static long initDrvSup() /* Locate all driver support entry tables */
{
    char	*pname;
    char	name[40];
    int		i, j;
    UTINY	type;
    char	message[100];
    long	status;
    long	rtnval=0;
    
    if(!drvSup) {
	status = S_drv_noDrvSup;
	errMessage(status,"No device drivers are defined (drvSup is NULL)");
	return(status);
    }
    if ((drivers = (struct driver *)calloc (drvSup -> number, sizeof (struct driver))) == NULL)
    {
	errMessage (-1L, "initDrvSup, drivers calloc failed");
    }
    for(i=0; i< (drvSup->number); i++) {
	if(!(pname = drvSup->drvetName[i])) continue;

	   /* remember driver name, in lower case */

	if (drivers)
	{
	    strncpy (drivers[i].driver_name, pname, sizeof (drivers[i].driver_name)-1);
	    for (j = 0; j < strlen (drivers[i].driver_name); j++)
	        drivers[i].driver_name[j] = tolower (drivers[i].driver_name[j]);
	}
	strcpy(name,"_");
	strcat(name,pname);
	rtnval = symFindByName(sysSymTbl,name,(char **)&(drvSup->papDrvet[i]),&type);
	if( rtnval!=OK || ( (type&N_TEXT) == 0) ) {
	    /* ems 15feb94
	    strcpy(message, "Driver -");
	    strcat(message, pname);
	    strcat(message, "- not installed");
	    status = S_drv_noDrvet;
	    errMessage(status,message);
	    if(rtnval==OK) rtnval=status;
	    */
	    continue;
	}
	strcpy(message, "Driver -");
	strcat(message, pname);
	strcat(message, "- installed");
	logMsg("%s",(int)message,0,0,0,0,0);

	   /* remember that this driver was loaded */

	if (drivers)
	    drivers[i].driver_loaded = 1;

	status = (*(drvSup->papDrvet[i]->init))();
	/* if(rtnval==OK) rtnval=status; */
	if (!status)
	    printf (" and initialized\n");
	else
	    printf ("\n");

    }
    return(0);
}

static long initRecSup()
{
    char	name[40];
    int		i;
    UTINY	type;
    char	message[100];
    long	status;
    long	rtnval=0; /*rtnval will be 0 or first error found*/
    int		nbytes;
    
    if(!dbRecType) {
	status = S_rectype_noRecs;
	errMessage(status,"dbRecType is NULL, i.e. no record types defined");
	return(status);
    }
    nbytes = sizeof(struct recSup) + dbRecType->number*sizeof(caddr_t);
    recSup = (struct recSup *)calloc(1,nbytes);
    recSup->number = dbRecType->number;
    (long)recSup->papRset = (long)recSup + (long)sizeof(struct recSup);
    for(i=0; i< (recSup->number); i++) {
	if(dbRecType->papName[i] == NULL)continue;
	strcpy(name,"_");
	strcat(name,dbRecType->papName[i]);
	strcat(name,"RSET");
	rtnval = symFindByName(sysSymTbl,name,(char **)&(recSup->papRset[i]),&type);
	if( rtnval!=OK || ( (type&N_TEXT) == 0) ) {
	    strcpy(message,"record support entry table not found for ");
	    strcat(message,name);
	    status = S_rec_noRSET;
	    errMessage(status,message);
	    if(rtnval==OK)rtnval=status;
	    continue;
	}
	if(!(recSup->papRset[i]->init)) continue;
	else {
	    status = (*(recSup->papRset[i]->init))();
	    if(rtnval==OK)rtnval=status;
	}
    }
    return(rtnval);
}

static long initDevSup() /* Locate all device support entry tables */
{
    char	*pname;
    char	*rectype;
    char	name[40];
    int		i,j;
    UTINY	type;
    char	message[100];
    long	status;
    long	rtnval=0; /*rtnval will be 0 or first error found*/
    struct devSup	*pdevSup;
    
    if(!devSup)
    {
	status = S_dev_noDevSup;
	errMessage(status,"No device support is defined (devSup is NULL)");
	return(status);
    }
    for(i=0; i< (devSup->number); i++)
    {
	/* devSup indexed same as dbRecType (ptrs to ptype names) */
	rectype = dbRecType -> papName[i];

	if((pdevSup = devSup->papDevSup[i]) == NULL)
	    continue;
	for(j=0; j < (pdevSup->number); j++)
	{
	    if(!(pname = pdevSup->dsetName[j]))
		continue;
	    strcpy(name,"_");
	    strcat(name,pname);
	    rtnval = (long)symFindByName(sysSymTbl,name,
		(char **)&(pdevSup->papDset[j]),&type);
	    if( rtnval!=OK || ( (type&N_TEXT) == 0) )
	    {
		/* only output a message if the driver was loaded
		*/
		if (drivers)
		{
		    char devname[64];
		    char *pdev, *pdrv;
		    int  l, no_msg, not_found, len_ptype;

		    strncpy (devname, pname, sizeof (devname) - 1);
		    devname[sizeof (devname) - 1] = 0;
		    for (l = 0; l < strlen (devname); l++)
			devname[l] = tolower (devname[l]);

		    /* special case where ptype part of device name
		     * does not match actual ptype str
		    */
		    if ( ((strcmp (rectype, "stringin")) == 0) ||
			 ((strcmp (rectype, "stringout")) == 0) ||
			 ((strcmp (rectype, "waveform")) == 0) )
		    {
			len_ptype = 2;
		    }
		    else
		    {
			len_ptype = strlen (rectype);
		    }
		    /* devname has form devxxxdrivername
		     * skip the dev and ptype to get the driver
		    */
		    pdev = devname + 3 + len_ptype;
		    no_msg = 0;
		    not_found = 1;
		    for (l = 0; l < drvSup -> number; l++)
		    {
			/* get just driver name, from drvmodicon */

			pdrv = drivers[l].driver_name + 3;
			if ((strcmp (pdev, pdrv)) == 0)
			{
			    not_found = 0;
			    if (drivers[l].driver_loaded == 0)
				no_msg = 1;
			    break;
			}
		    }
		    if (no_msg || not_found)
			continue;
		}
		strcpy(message,"Device support entry table not found for ");
		strcat(message,pname);
		status = S_dev_noDSET;
		errMessage(status,message);
		if(rtnval==OK)rtnval=status;
		continue;
	    }
	    if(!(pdevSup->papDset[j]->init)) continue;
	    status = (*(pdevSup->papDset[j]->init))();
	    if(rtnval==OK)rtnval=status;
	}
    
	/* Now initialize dset for each record */
	/* This code moved to initDatabase() by Gordon */
    }
    return(rtnval);
}

long initDatabase(restart)
int	restart;
{
    char	name[PVNAME_SZ+FLDNAME_SZ+2];
    short	i,j,k;
    char	message[120];
    long	status;
    long	rtnval=0; /*rtnval will be 0 or first error found*/
    short	nset=0;
    struct recLoc	*precLoc;
    struct rset		*prset;
    struct recTypDes	*precTypDes;
    struct fldDes	*pfldDes;
    struct dbCommon	*precord;
    struct dbAddr	dbAddr;
    struct link		*plink;
    struct recChain	*pchain;
    
    status = S_record_noRecords;
    if(!dbRecords) {
	status = S_record_noRecords;
	errMessage(status,"No database records are defined");
	return(status);
    }

    if (restart)
	nset = dbScanPvt.nset;

    for(i=0; i< (dbRecords->number); i++) {
	if(!(precLoc = dbRecords->papRecLoc[i]))
	    continue;
	if(!(prset=GET_PRSET(i))) {
	    strcpy(name,dbRecType->papName[i]);
	    strcat(name,"RSET");
	    strcpy(message,"record support entry table not found for ");
	    strcat(message,name);
	    status = S_rec_noRSET;
	    errMessage(status,message);
	    if(rtnval==OK)
		rtnval = status;
	    continue;
	}
	precTypDes = dbRecDes->papRecTypDes[i];
	for(j=0, pchain = (struct recChain *) precLoc -> pFirst;
	    j<precLoc->no_records;
	    j++, pchain = pchain -> next) {
	        /* If NAME is null then skip this record*/
		precord = (struct dbCommon *) pchain -> pRecord;
		if(!(precord->name[0])) continue;

	        /* initialize mlok and mlis*/
		if (!(precord -> mlok.ppend))
			FASTLOCKINIT(&precord->mlok);
		/*
		Change by tate. Remove the next two lines of code because
		it is possible to come in here for an existing point and
		we don't want to screw up the monitor list. Records that
		are new are guaranteed to have zeros in all non-prompted
		fields, and all that these routines do is set stuff to 
		zero anyway.

		lstInit(&(precord->mlis));
		precord->pact=FALSE;
		*/

		/* Create a new lock set if this record isn't already in one */

		if (precord -> lset == 0)
			{
			nset++;
			precord -> lset = nset;
			precord -> lsad = (struct scanLock *) calloc(1,sizeof(struct scanLock));
			precord -> lsad -> pFirstRec = precord;
			FASTLOCKINIT(&(precord -> lsad -> lock));
			}

		/* Convert all PV_LINKs to DB_LINKs or CA_LINKs*/
		for(k=0; k<precTypDes->no_links; k++) {
		    pfldDes = precTypDes->papFldDes[precTypDes->link_ind[k]];
		    plink = (struct link *)((char *)precord + pfldDes->offset);
		    if((plink -> type == DB_LINK) || (plink->type == PV_LINK)) {
		        char	savePvName[PVNAME_SZ];
			char	saveFldName[FLDNAME_SZ];
			strncpy(savePvName, plink -> value.pv_link.pvname, PVNAME_SZ);
			strncpy(saveFldName, plink -> value.pv_link.fldname, FLDNAME_SZ);
			strncpy(name,plink->value.pv_link.pvname,PVNAME_SZ);
			name[PVNAME_SZ]=0;
			strcat(name,".");
			strncat(name,plink->value.pv_link.fldname,FLDNAME_SZ);
			if (plink -> type == DB_LINK)
				{
				/*dbAddr.precord = (caddr_t) precord;*/
				dbAddr.precord = ((struct dbAddr *)(plink -> value.db_link.pdbAddr))
						-> precord;
				dbAddr.no_elements = ((struct dbAddr *)(plink -> value.db_link.pdbAddr)) 
						-> no_elements;
				}

			else 
				{
				if(dbNameToAddr(name,&dbAddr) == 0) 
				    {
			            plink->type = DB_LINK;
			            strncpy(plink -> value.db_link.pvname, savePvName, PVNAME_SZ);
			            strncpy(plink -> value.db_link.fldname, saveFldName, FLDNAME_SZ);
			            plink->value.db_link.pdbAddr =
				        (caddr_t)calloc(1,sizeof(struct dbAddr));
			            *((struct dbAddr *)(plink->value.db_link.pdbAddr))=dbAddr;
				    }
				else
				    {
			            /*This will be replaced by channel access call*/
				    dbAddr.precord = 0;
			            strncpy(message,precord->name,PVNAME_SZ);
			            message[PVNAME_SZ]=0;
			            strcat(message,".");
			            strncat(message,pfldDes->fldname,FLDNAME_SZ);
			            strcat(message,": link process variable =");
			            strcat(message,name);
			            strcat(message," not found");
			            status = S_db_notFound;
			            errMessage(status,message);
			            if(rtnval==OK) rtnval=status;
			            }
				}
			if (dbAddr.precord)
				{

				/*
				Now check the lock set number pointed to by this link.
				Don't include INLINKS that are NOT process passive. In
				that case we look but don't touch, so don't include them
				in our lock set.
				*/

				if (!(pfldDes -> field_type == DBF_INLINK 
				   && !(plink -> value.db_link.process_passive)
				   && (dbAddr.no_elements <= 1)))
					{
					struct dbCommon *targRec;
					struct dbCommon *oldHead;

					targRec = (struct dbCommon *) dbAddr.precord;
					if (targRec -> lset != precord -> lset)
						{

						/*
						They don't match.
						*/

						if (targRec -> lset == 0)
							{

							/*
							Other record doesn't belong to one
							yet. Make him join ours.
							*/

							targRec -> lset = precord -> lset;
							targRec -> lsad = precord -> lsad;
							targRec -> lsnr = targRec -> lsad -> pFirstRec;
							targRec -> lsad -> pFirstRec = targRec;
							}
						
						else
							{

							/*
							Other record is already a member of another
							set. Merge that set with the one for this
							record.
							*/

							oldHead = precord -> lsad -> pFirstRec;
							precord -> lsad -> pFirstRec = targRec -> lsad ->
												pFirstRec;
							FASTLOCKFREE(&(targRec -> lsad -> lock));
							free(targRec -> lsad);
							targRec = precord -> lsad -> pFirstRec;
							for (;;)
								{
								targRec -> lset = precord -> lset;
								targRec -> lsad = precord -> lsad;
								if (targRec -> lsnr)
									targRec = targRec -> lsnr;
								else
									{
									targRec -> lsnr = oldHead;
									break;
									}
								}
							}
						}
					}
				}
			}
		    }

		/* Initialize the device support entry table pointer */
		if (!(precord -> dset))
			precord->dset=(struct dset *)GET_PDSET(devSup -> papDevSup[i],precord->dtyp);
		if(!(prset->init_record)) continue;

		/* check both initialization flags before calling the points
		   init_record, see each point type, ems 04/22/94
		*/
		if ((!(precord -> init)) && (!(precord -> ini2)))
			status = (*(prset->init_record))(precord);
		if(rtnval==OK)rtnval=status;
	}
    }

    dbScanPvt.nset = nset;

    /* If this is a database restart (i.e., after an online download that
     * has either added points to the database or changed resources, etc.)
     * then we have to get the world ready to recreate the lock sets. That
     * means we have to free the old ones and reappend all the lock set numbers.
     */

     if (restart)
	resFreeChains();

    /* Now determine which resources are involved in each of the lock sets*/

    for(i=0; i<dbRecords->number; i++) {
	if(!(precLoc = dbRecords->papRecLoc[i]))continue;
	precTypDes = dbRecDes->papRecTypDes[i];
	for(j=0, pchain = (struct recChain *)precLoc->pFirst;
	    j<precLoc->no_records;
	    j++, pchain = pchain -> next) {
	        /* If NAME is null then skip this record*/
		precord = (struct dbCommon *) (pchain -> pRecord);
		if(!(precord->name[0]))
			continue;
		resAppendLock(precord -> radr, precord -> lsad);
		precord -> lsad -> resDefined = 0;
		precord -> lsad -> resGot = 0;
	}
    }    

    resLockInit();
    return(rtnval);
}

#define MAX 128
#define SAME 0
static char    *cvt_str[] = {
    "DBF_STRING",
    "DBF_SHORT",
    "DBF_LONG",
    "DBF_FLOAT",
    "DBF_DOUBLE"
};
#define CVT_COUNT (sizeof(cvt_str) / sizeof(char*))

static long getResources(fname) /* Resource Definition File interpreter */
    char           *fname;
{
    int             fd;
    int             len;
    int             len2;
    int             lineNum = 0;
    int             i = 0;
    int             found = 0;
    int             cvType = 0;
    char            buff[MAX + 1];
    char            name[40];
    char            s1[MAX];
    char            s2[MAX];
    char            s3[MAX];
    char            message[100];
    long            rtnval = 0;
    UTINY           type;
    char           *pSymAddr;
    short           n_short;
    long            n_long;
    float           n_float;
    double          n_double;
    if (!fname) return (0);
    if ((fd = open(fname, READ, 0x0)) < 0) {
	errMessage(0L, "getResources: No such Resource file");
	return (-1);
    }
    while ((len = fioRdString(fd, buff, MAX)) != EOF) {
	lineNum++;
	if (len < 2)
	    goto CLEAR;
	if (len >= MAX) {
	    sprintf(message,
		    "getResources: Line too long - line=%d", lineNum);
	    errMessage(-1L, message);
	    return (-1);
	}
	for (i = 0; i < len; i++) {
	    if (buff[i] == '!') {
		goto CLEAR;
	    }
	}
	/* extract the 3 fields as strings */
	if ((sscanf(buff, "%s %s %[^\n]", s1, s2, s3)) != 3) {
	    sprintf(message,
		    "getResources: Not enough fields - line=%d", lineNum);
	    errMessage(-1L, message);
	    return (-1);
	}
	found = 0;
	len2 = strlen(s2);
	for (i = 0; i < CVT_COUNT; i++) {
	    if ((strncmp(s2, cvt_str[i], len2)) == SAME) {
		found = 1;
		cvType = i;
		break;
	    }
	}
	if (!found) {
	    sprintf(message,
		    "getResources: Field 2 not defined - line=%d", lineNum);
	    errMessage(-1L, message);
	    return (-1);
	}
	strcpy(name, "_");
	strcat(name, s1);
	rtnval = symFindByName(sysSymTbl, name, &pSymAddr, &type);
	if (rtnval != OK || ((type & N_TEXT) == 0)) {
	    sprintf(message,
		  "getResources: Symbol name not found - line=%d", lineNum);
	    errMessage(0L, message);
	    return (-1);
	}
	switch (cvType) {
	case 0:		/* DBF_STRING */
	    len = strlen(s3);
	    len2 = strlen((char *)pSymAddr);
	    if (len > len2) {
		sprintf(message,
			"getResources: Warning, string might exceed previous reserved space - line=%d",
			lineNum);
		errMessage(-1L, message);
	    }
	    strncpy(pSymAddr, s3, len + 1);
	    break;
	case 1:		/* DBF_SHORT */
	    if ((sscanf(s3, "%hd", &n_short)) != 1) {
		sprintf(message,
		      "getResources: conversion failed - line=%d", lineNum);
		errMessage(0L, message);
	        return (-1);
	    }
	    *(short *) pSymAddr = n_short;
	    break;
	case 2:		/* DBF_LONG */
	    if ((sscanf(s3, "%ld", &n_long)) != 1) {
		sprintf(message,
		      "getResources: conversion failed - line=%d", lineNum);
		errMessage(0L, message);
	        return (-1);
	    }
	    *(long *) pSymAddr = n_long;
	    break;
	case 3:		/* DBF_FLOAT */
	    if ((sscanf(s3, "%e", &n_float)) != 1) {
		sprintf(message,
		      "getResources: conversion failed - line=%d", lineNum);
		errMessage(0L, message);
	        return (-1);
	    }
	    *(float *) pSymAddr = n_float;
	    break;
	case 4:		/* DBF_DOUBLE */
	    if ((sscanf(s3, "%le", &n_double)) != 1) {
		sprintf(message,
		      "getResources: conversion failed - line=%d", lineNum);
		errMessage(0L, message);
	        return (-1);
	    }
	    *(double *) pSymAddr = n_double;
	    break;
	default:
	    sprintf(message,
		 "getResources: switch default reached - line=%d", lineNum);
	    errMessage(-1L, message);
	    return (-1);
	    break;
	}
CLEAR:	bzero(buff, MAX);
	bzero(s1, MAX);
	bzero(s2, MAX);
	bzero(s3, MAX);
    }
    close(fd);
    return (0);
}

static long initProcessStats ()
{
    if ((process_cnts = 
	(long *)(calloc (1, dbRecType -> number * sizeof (long)))) == NULL)
    {
	logMsg ("initProcessStats failed calloc\n",0,0,0,0,0,0);
	return (-1L);
    }

    return (0);
}
