#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>

#if defined(SOLARIS) || defined(WIN32)
#include <sys/stat.h>
#endif

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#define	PT_TYPE_FLD	"POINTTYPE"
#define	PT_TYPE_FLD_SZ	15
#define	FLDNAME_FLD	"FIELDNAME"
#define	FLDNAME_FLD_SZ	6
#define	FLD_NAME_FLD	"FIELD_NAME"
#define	FLD_NAME_FLD_SZ	10
#define	TABLE_FLD	"TABLE"
#define	TABLE_FLD_SZ	20
#define	TRUESTATE_FLD	"TRUE_STATE"
#define	TRUESTATE_FLD_SZ 1
#define	OPERANDID_FLD	"OPERANDID"
#define	OPERANDID_FLD_SZ 20
#define	HWFLAG_FLD	"HWFLAG"
#define	HWFLAG_FLD_SZ	1
#define	REF_FLD		"REFFIELD"
#define	REF_FLD_SZ	10
#define	CONST_FLD	"CONSTFIELD"
#define	CONST_FLD_SZ	10

static FIELD4INFO xrefFieldInfo [] = 
{
    { PT_TYPE_FLD,	'C',	PT_TYPE_FLD_SZ,		0 },
    { FLDNAME_FLD,	'C',	FLDNAME_FLD_SZ,		0 },
    { FLD_NAME_FLD,	'C',	FLD_NAME_FLD_SZ,	0 },
    { TABLE_FLD,	'C',	TABLE_FLD_SZ,		0 },
    { TRUESTATE_FLD,	'L',	TRUESTATE_FLD_SZ,	0 },
    { OPERANDID_FLD,	'C',	OPERANDID_FLD_SZ,	0 },
    { HWFLAG_FLD,	'L',	HWFLAG_FLD_SZ,		0 },
    { REF_FLD,		'C',	REF_FLD_SZ,		0 },
    { CONST_FLD,	'C',	CONST_FLD_SZ,		0 },
    { 0, 0, 0, 0 }
};

TAG4INFO xrefTagInfo [] = 
{
    { "XREF",	"POINTTYPE+FIELDNAME",	0, 0, 0 },
    { 0, 0, 0, 0, 0 }
};

#define	XREFNAME	"pdbxref"
#define	XREFTEXT	"pdbxref.txt"
#define	END_OF_LINE	0
#define	VALID_FIELD	1

/* codebase stuff */
static CODE4	cb;
static DATA4	*xrefData;
static FIELD4	*pointtypeFld, *fldnameFld, *fld_nameFld, *tableFld;
static FIELD4	*truestateFld, *operandidFld, *hwflagFld, *refFld;
static FIELD4	*constFld;


/* the functions */
static int	createXrefFile (void);
static int	getNextField (void);
static char	*getNextLine (void);
static void	initCodeBase (void);
static void	loadDatabase (void);
static int	mapFields (void);
static int	openXrefText (void);
static void	outputDatabaseFields (void);

static FILE	*fd;
static int	error = 0;
static char	fieldBuff[64];
static char	inBuff[256];
static char	*inBuffPtr;
static int	line = 0;
static char	xreftextPath[256];

static char	type[PT_TYPE_FLD_SZ + 1];
static char	field[FLDNAME_FLD_SZ + 1];
static char	fld[FLD_NAME_FLD_SZ + 1];
static char	table[TABLE_FLD_SZ + 1];
static char	truestate[TRUESTATE_FLD_SZ + 1];
static char	operandid[OPERANDID_FLD_SZ + 1];
static char	hwflag[HWFLAG_FLD_SZ + 1];
static char	reffld[REF_FLD_SZ + 1];
static char	constfld[CONST_FLD_SZ + 1];



void main (int argc, char **argv)
{
    if (argc != 2)
    {
	fprintf (stdout, "xreconvert usage: xrefconvert xreftextpath\n");
	exit (1);
    }
    fprintf (stdout, "xrefconvert started\n");

    strcpy (xreftextPath, argv[1]);

    initCodeBase ();

    if (openXrefText ())
    {
	fprintf (stdout, "xrefconvert, finished with error\n");
	exit (1);
    }

    if (createXrefFile ())
    {
	fprintf (stdout, "xrefconvert, finished with error\n");
	fclose (fd);
	exit (1);
    }

    if (mapFields ())
    {
	fprintf (stdout, "xrefconvert, finished with error\n");
	d4close (xrefData);
	fclose (fd);
	exit (1);
    }

    loadDatabase ();

    fprintf (stdout, "xrefconvert finished\n");

    exit (0);
}


static int createXrefFile (void)
{
    int res = 0;

    if ((xrefData = d4create (&cb, XREFNAME, xrefFieldInfo, xrefTagInfo)) == NULL)
    {
	fprintf (stdout, "xrefconvert: failed to create %s (%d)\n",
			 XREFNAME, cb.errorCode);
	res = 1;
    }

    return (res);
}

static void initCodeBase (void)
{
    code4init (&cb);

    cb.safety = 0;	/* overwrite file if it exists */
    cb.errOpen = 0;	/* tell codebase not to error on open */
    cb.errCreate = 0;	/* tell codebase not to error on create */
    cb.errFieldName = 0; /* tell codebase not to error on field name */
}


static int getNextField (void)
{
    int		firstChar = 1;
    char	*fieldBuffPtr;

    memset (fieldBuff, 0, sizeof (fieldBuff));
    fieldBuffPtr = &fieldBuff[0];

    while (1)
    {
	if ((*inBuffPtr) == ',')
	{
	    inBuffPtr++;
	    break;
	}
	if ((*inBuffPtr) == '\n')
	{
	    if (firstChar)
	    {
		return (END_OF_LINE);
	    }
	    else
	    {
		return (VALID_FIELD);
	    }
	}
	firstChar = 0;
	if ((*inBuffPtr) == '\"')
	{
	    inBuffPtr++;
	    continue;
	}
	*fieldBuffPtr++ = *inBuffPtr++;
    }

    return (VALID_FIELD);
}


static char *getNextLine (void)
{
    char	*ptr;

    line++;
    ptr = fgets (inBuff, sizeof(inBuff)-1, fd);
    if (ptr == NULL)
    {
	if (!feof (fd))
	{
	    error = 1;
	}
    }
    else
    {
	inBuffPtr = &inBuff[0];
    }

    return (ptr);
}

static void loadDatabase (void)
{
    int	nextFieldRes;
    int fldNum;

    while ((getNextLine ()) != NULL)
    {
	if (error)
	{
	    fprintf (stdout, "format error line %d\n",line);
	    exit (1);
	}
	fldNum = -1;
	while ((nextFieldRes = getNextField ()) == VALID_FIELD)
	{
	    fldNum++;
	    switch (fldNum)
	    {
		case 0:
		    memset (type, 0, sizeof (type));
		    strncpy (type, fieldBuff, sizeof (type)-1);
		    type[sizeof(type) -1] = 0;
		    break;
		case 1:
		    memset (field, 0, sizeof (field));
		    strncpy (field, fieldBuff, sizeof (field)-1);
		    field[sizeof(field) -1] = 0;
		    break;
		case 2:
		    memset (fld, 0, sizeof (fld));
		    strncpy (fld, fieldBuff, sizeof (fld)-1);
		    fld[sizeof(fld) -1] = 0;
		    break;
		case 3:
		    memset (table, 0, sizeof (table));
		    strncpy (table, fieldBuff, sizeof (table)-1);
		    table[sizeof(table) -1] = 0;
		    break;
		case 4:
		    memset (truestate, 0, sizeof (truestate));
		    strncpy (truestate, fieldBuff, sizeof (truestate)-1);
		    truestate[sizeof(truestate) -1] = 0;
		    break;
		case 5:
		    memset (operandid, 0, sizeof (operandid));
		    strncpy (operandid, fieldBuff, sizeof (operandid)-1);
		    operandid[sizeof(operandid) -1] = 0;
		    break;
		case 6:
		    memset (hwflag, 0, sizeof (hwflag));
		    strncpy (hwflag, fieldBuff, sizeof (hwflag)-1);
		    hwflag[sizeof(hwflag) -1] = 0;
		    break;
		case 7:
		    memset (reffld, 0, sizeof (reffld));
		    strncpy (reffld, fieldBuff, sizeof (reffld)-1);
		    reffld[sizeof(reffld) -1] = 0;
		    break;
		case 8:
		    memset (constfld, 0, sizeof (constfld));
		    strncpy (constfld, fieldBuff, sizeof (constfld)-1);
		    constfld[sizeof(constfld) -1] = 0;
		    break;
	    }
	}
	outputDatabaseFields ();
    }
    d4flush (xrefData);
    d4unlock (xrefData);
    d4close (xrefData);
}


static int mapFields (void)
{
    int	res = 0;

    if ((pointtypeFld = d4field (xrefData, PT_TYPE_FLD)) == NULL)
    {
	fprintf (stdout, "xrefconvert error: failed to find %s\n",PT_TYPE_FLD);
	return (1);
    }

    if ((fldnameFld = d4field (xrefData, FLDNAME_FLD)) == NULL)
    {
	fprintf (stdout, "xrefconvert error: failed to find %s\n",FLDNAME_FLD);
	return (1);
    }

    if ((fld_nameFld = d4field (xrefData, FLD_NAME_FLD)) == NULL)
    {
	fprintf (stdout, "xrefconvert error: failed to find %s\n",FLD_NAME_FLD);
	return (1);
    }

    if ((tableFld = d4field (xrefData, TABLE_FLD)) == NULL)
    {
	fprintf (stdout, "xrefconvert error: failed to find %s\n",TABLE_FLD);
	return (1);
    }

    if ((truestateFld = d4field (xrefData, TRUESTATE_FLD)) == NULL)
    {
	fprintf (stdout, "xrefconvert error: failed to find %s\n",TRUESTATE_FLD);
	return (1);
    }

    if ((operandidFld = d4field (xrefData, OPERANDID_FLD)) == NULL)
    {
	fprintf (stdout, "xrefconvert error: failed to find %s\n",OPERANDID_FLD);
	return (1);
    }

    if ((hwflagFld = d4field (xrefData, HWFLAG_FLD)) == NULL)
    {
	fprintf (stdout, "xrefconvert error: failed to find %s\n",HWFLAG_FLD);
	return (1);
    }

    if ((refFld = d4field (xrefData, REF_FLD)) == NULL)
    {
	fprintf (stdout, "xrefconvert error: failed to find %s\n",REF_FLD);
	return (1);
    }

    if ((constFld = d4field (xrefData, CONST_FLD)) == NULL)
    {
	fprintf (stdout, "xrefconvert error: failed to find %s\n",CONST_FLD);
	return (1);
    }

    return (res);
}


static int openXrefText (void)
{
    int res = 0;
    char  xrefpath[256];

    sprintf (xrefpath, "%s/%s", xreftextPath, XREFTEXT);

    if ((fd = fopen (xrefpath, "r")) == NULL)
    {
	fprintf (stdout, "xrefconvert error: failed to open %s\n", xrefpath);
	return (1);
    }
    return (res);
}


static void outputDatabaseFields (void)
{
    d4appendStart (xrefData, 0);
    d4blank (xrefData);
    f4assign (pointtypeFld, type);
    f4assign (fldnameFld, field);
    f4assign (fld_nameFld, fld);
    f4assign (tableFld, table);
    f4assign (truestateFld, truestate);
    f4assign (operandidFld, operandid);
    f4assign (hwflagFld, hwflag);
    f4assign (refFld, reffld);
    f4assign (constFld, constfld);
    d4append (xrefData);
    d4flush (xrefData);
}
