/*dbe2.c,v 1.1.1.1 1999/08/03 14:42:38 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <vxWorks.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <string.h>
#include <ioLib.h>
#include <vmLib.h>
#include <logLib.h>
#include <errnoLib.h>
#include <taskLib.h>
#include <ioLib.h>
#include <sysLib.h>
#include <envLib.h>

#include "sdrHeader.h"
#include "dbDefs.h"
#include "dbRecDes.h"
#include "dbCommon.h"

/* function prototypes */
int	protectBBM (void);
int	unprotectBBM (void);
int	sizeBBM (void);

#define MAXBINS 256
typedef unsigned long ui32;
typedef unsigned short ui16;
typedef unsigned char uchar;
typedef struct _sl { struct _sl *l; } sl;
typedef struct _bv { int n, b, m, *v; } bvec;
typedef struct _bn { sl l; void *d; } bin;
typedef struct _nd { char key[PVNAME_SZ+1]; ui16 type, crc; ui32 recn; } node;

#define guSDR_NTYPES	33

int	protectionDisabled = 0;

static int lsb( int x )
{
    return x & -x;
}

static int nbits( int x )
{
    register int n = 0;
    /**/
    while( x )
    ++n, x ^= lsb( x );
    return n;
}

static int lg2( int x )
{
    register int n = 0;
    /**/
    while( x )
	++n, x >>= 1;
    return n - 1;
}

static bvec *newbv( int n )
{
    bvec *p = (bvec*) calloc( 1, sizeof(bvec) );
    int t = nbits( -1 );
    /**/
    p-> n = n, p-> b = lg2( t ), p-> m = t - 1;
    p-> v = (int*) calloc( (n + p-> m) >> p-> b, sizeof(int) );
    return p;
}

static void set( int n, bvec *p )
{
    if( (unsigned) n < p-> n )
	p-> v[ n >> p-> b ] |= ((unsigned) (1 << p-> m) >> (n & p-> m));
}

static int tst( int n, bvec *p )
{
    return ((unsigned) n < p-> n) ?
	(p-> v[ n >> p-> b ] & ((unsigned) (1 << p-> m) >> (n & p-> m)) ? 1 : 0) : 0;
}

#define MAXPR 1024
static bvec *pr;
static int isprime( ui32 n )
{
    register ui32 i;
    /**/
    for( i = 2; i < MAXPR; ++i )
	if( ! tst( i, pr ) && (n % i) == 0 )
	  return 0;
	else if( i*i > n ) break;
    /**/
    return 1;
}

static ui32 nprime( ui32 n )
{
    register int i, j;
    /**/
    if( n > (MAXPR*MAXPR) ) return -1;
    if( pr == 0 )
    {
	pr = newbv( MAXPR );
	for( i = 2; i < MAXPR / 2; ++i )
	{
	    if( ! tst( i, pr ) )
		for( j = i + i; j < MAXPR; j += i )
		    set( j, pr );
	}
    }
    /**/
    while( n < MAXPR )
	if( ! tst( n, pr ) )
	    return n; else ++n;
    /**/
    while( n < (MAXPR*MAXPR) )
	if( isprime( n ) )
	    return n; else ++n;
    /**/
    return -1;
}

static ui32 khash( uchar *k, int l, ui32 M )
{
    register ui32 h = 0;
    /**/
    if( l & 1 )
	h = ((ui16) *k << 16) % M, --l, ++k;
    l >>= 1;
    while( l-- )
	h = ((h + (((ui16) k[0] << 8) + k[1])) * nprime((l+1)<<2)) % M, k += 2;
    /**/
    return h;
}

static ui32 oM, klen;
static bin **bins;
static void inithash( ui32 n, ui32 kl )
{
    if( bins == 0 )
    {
	oM = nprime( n ), klen = kl;
	bins = (bin**) calloc( oM, sizeof(bin*) );
    }
}

static void hput( uchar *k, void *d )
{
    register ui32 hk = khash( k, klen, oM );
    register bin *pb = (bin*) calloc( 1, sizeof(bin) );
    /**/
    pb-> d = d, pb-> l.l = (sl*) bins[ hk ], bins[ hk ] = (bin*) pb;
}

static void *hfind( uchar *k, int (*m)() )
{
    register ui32 hk = khash( k, klen, oM );
    register bin *pb = bins[ hk ];
    /**/
    while( pb != 0 )
	if( ! m( k, pb-> d ) )
	    return pb-> d;
	else
	    pb = (bin*) pb-> l.l;
    /**/
    return 0;
}

static bin *hbin( uchar *k, int *phk )
{
    register bin *pb;
    register int kl = strlen( k );
    /**/
    if( kl < klen )
	memset( k + kl, 0, klen - kl );
    *phk = khash( k, klen, oM ), pb = bins[ *phk ];
    /**/
    while( pb != 0 )
	if( ! strcmp( (char*) k, pb-> d ) )
	    break;
	else
	    pb = (bin*) pb-> l.l;
    /**/
    return pb;
}

static node *hwalk( uchar *k )
{
    register bin *pb = 0;
    int bn = 0;
    /**/
    if( *k != 0 )
	if( (pb = (bin*) hbin( k, &bn )-> l.l) == 0 )
	    ++bn;
    /**/
    if( pb == 0 )
	while( bn < oM && (pb = bins[ bn ]) == 0 )
	    ++bn;
    /**/
    return pb != 0 ? (node*) pb-> d : 0;
}

static char e2ftemp [50];
static char e2fname [50];
static char *e2fwork;
static int e2fd = -1;
static ui32 numrecs, numpnts;
static sl *empty[guSDR_NTYPES+1]; /*types are 1-based*/

static int e2flush()
{
    close( e2fd ), e2fd = -1;
    return open( e2fwork, UPDATE, 0 );
}

int InitE2DB()
{
    register int i;
    char *cp;

    if ((cp = getenv("BBM:")) != NULL)
    {
	sprintf (e2ftemp, "%se2db.tmp", cp);
	sprintf (e2fname, "%se2db.dat", cp);
    }
    else
    {
	sprintf (e2ftemp, "BBM:e2db.tmp");
	sprintf (e2fname, "BBM:e2db.dat");
    }

    /**/
    if( e2fd != -1 )
	return TRUE;
    inithash( MAXBINS, PVNAME_SZ );
    /**/

    unprotectBBM();

    remove( e2ftemp );
    if( (e2fd = open( e2fwork = e2fname, UPDATE, 0 )) < 0 )
    {
	if( (e2fd = open( e2fwork = e2ftemp, UPDATE|O_CREAT, 0 )) < 0 )
	{
	    logMsg( "open '%s': 0x%x\n", (int)e2ftemp, errnoGet(),0,0,0,0 ), taskSuspend(0);
	}
	/**/
	numrecs = numpnts = 0;
	write( e2fd, (char *)&numrecs, sizeof(numrecs) );
	/**/
	if( (e2fd = e2flush()) < 0 )
	{
	    logMsg( "flush(I) '%s': 0x%x\n", (int)e2fwork, errnoGet(),0,0,0,0 ), taskSuspend(0);
	}

	protectBBM();

	return FALSE;
    }

    /**/
    if( lseek( e2fd, 0L, SEEK_SET ) < 0 ||
      read( e2fd, (char *)&numrecs, sizeof(numrecs) ) < sizeof(numrecs) )
    {
	logMsg( "read(1) '%s': 0x%x\n", (int)e2fwork, errnoGet(),0,0,0,0 ), taskSuspend(0);
    }
    /**/
    for( i=0; i < numrecs; ++i )
    {
	register node *pn;
	register ui32 recn = lseek( e2fd, 0L, SEEK_CUR );
	ui16 type, len, crc;
	uchar pkey[PVNAME_SZ];
	/**/
	if( recn < 0 ||
          read( e2fd, (char *)&type, sizeof(type) ) < sizeof(type) ||
          read( e2fd, pkey, sizeof(pkey) ) < sizeof(pkey) )
	{
	    logMsg( "read(2) '%s': 0x%x\n", (int)e2fwork, errnoGet(),0,0,0,0 ), taskSuspend(0);
	}
	/**/
	len = dbRecDes-> papRecTypDes[ type ]-> rec_size;
	if( *pkey != 0 )
	{
	    char rec[ len -= PVNAME_SZ ];
	    /**/
	    if( read( e2fd, rec, len ) < len || /*already read key*/
	      read( e2fd, (char *)&crc, sizeof(crc) ) < sizeof(crc) )
	    {
		logMsg( "read(3) '%s': 0x%x\n",(int) e2fwork, errnoGet(),0,0,0,0 ), taskSuspend(0);
	    }
	    /*validate crc*/
	    if( hfind( pkey, strcmp ) != 0 )
		return FALSE/*kill db file?*/;
	    if( (pn = (node*) calloc( 1, sizeof(node) )) == 0 )
		return ERROR/*kill db file?*/;
	    /**/
	    sprintf( pn-> key, "%.*s", (int)sizeof(pkey), pkey );
	    pn-> type = type, pn-> crc = crc, pn-> recn = recn;
	    hput( pkey, pn ), ++numpnts;
	}
	else
	{
	    register bin *pb = (bin*) calloc( 1, sizeof(bin) );
	    /**/
	    if( pb == 0 || (pn = (node*) calloc( 1, sizeof(node) )) == 0 )
		return ERROR/*kill db file?*/;
	    /**/
	    lseek( e2fd, (len - PVNAME_SZ) + sizeof(crc), SEEK_CUR );
	    memset( pn-> key, 0, sizeof(pn-> key) );
	    pn-> type = type, pn-> crc = -1, pn-> recn = recn;
	    pb-> d = pn, pb-> l.l = empty[ type ], empty[ type ] = &pb-> l;
	}
    }

    protectBBM();

    logMsg( "BBM DB loaded (numrecs: %lu / numpnts: %lu)\n", numrecs, numpnts,0,0,0,0 );
	return TRUE;
}

void ActivateE2DB()
{
    if( e2fd != -1 )
    {

	unprotectBBM();

	close( e2fd ), e2fd = -1;
	if (rename( e2ftemp, e2fname ) == ERROR)
	    perror("Error renaming file");
	if( (e2fd = open( e2fwork = e2fname, UPDATE, 0 )) < 0 )
	{
	    logMsg( "flush(A) '%s': 0x%x\n", (int)e2fwork, errnoGet(),0,0,0,0 ), taskSuspend(0);
	}

	protectBBM();

	logMsg( "BBM DB activated\n",0,0,0,0,0,0 );
    }
}

void DelE2DB()
{
    register bin *pb;
    register int i;
    /**/

    unprotectBBM();

    if( e2fd != -1 )
	close( e2fd ), e2fd = -1;
    remove( e2fname );

    protectBBM();

    numpnts = numrecs = 0;
    /**/
    for( i=1; i <= guSDR_NTYPES; ++i )
	while( (pb = (bin*) empty[i]) != 0 )
	{
	    empty[i] = pb-> l.l;
	    cfree( (char*) pb-> d );
	    cfree( (char*) pb );
	}
    /**/
    for( i=0; i < oM; ++i )
	while( (pb = bins[i]) != 0 )
	{
	    bins[i] = (bin*) pb-> l.l;
	    cfree( (char*) pb-> d );
	    cfree( (char*) pb );
	}
    cfree( (char*) bins );
    logMsg( "BBM DB deleted\n",0,0,0,0,0,0 );
}

static node *avail( ui16 type )
{
    register bin *pb = (bin*) empty[ type ];
    register node *pn;
    /**/
    if( pb != 0 )
    {
	pn = pb-> d, empty[ type ] = pb-> l.l;
	cfree( (char*) pb );
    }
    else
    {
	if( (pn = (node*) calloc( 1, sizeof(node) )) != 0 )
	{
	    memset( pn-> key, 0, sizeof(pn-> key) );
	    pn-> type = type, pn-> crc = -1, pn-> recn = lseek( e2fd, 0L, SEEK_END );
	}
    }
    return pn;
}

ui16 GetType( char *pkey )
{
    register node *pn = hwalk( (uchar*) pkey );
    /**/
    if( pn != 0 )
    {
	memset( pkey, 0, PVNAME_SZ ), strcpy( pkey, pn-> key );
	return pn-> type;
    }
    /**/
    return 0;
}

void GetRecord( char *pkey, short type, char *buff )
{
    register node *pn;
    register int kl = strlen( pkey );
    register int len = dbRecDes-> papRecTypDes[ type ]-> rec_size;
    /**/
    if( kl < klen )
	memset( pkey + kl, 0, klen - kl );
    if( (pn = hfind( (uchar*) pkey, strcmp )) == 0 )
    {
	logMsg( "can't happen (GR)!\n",0,0,0,0,0,0 ), taskSuspend(0);
    }
    /**/
    /*logMsg( "GetRecord ='%.*s' @%lu (..%d)\n", PVNAME_SZ, (int)pkey, pn-> recn, len,0,0 );*/

    unprotectBBM();

    if( lseek( e2fd, pn-> recn + sizeof(type), SEEK_SET ) < 0 )
    {
	logMsg( "lseek(GR) '%s': 0x%x\n", (int)e2fwork, errnoGet(),0,0,0,0 ), taskSuspend(0);
    }
    if( read( e2fd, buff, len ) < len ) /*'full' record including key*/
    {
	logMsg( "read(GR) '%s': 0x%x\n", (int)e2fwork, errnoGet(),0,0,0,0 ), taskSuspend(0);
    }

    protectBBM();

}

/* consider records opaque - they are still in net std order in the DB */
void PutRecord( char *pkey, short type, char *buff )
{
    register node *pn;
    register int kl = strlen( pkey );
    register int len = dbRecDes-> papRecTypDes[ type ]-> rec_size, new;
    /**/
    if( kl < klen )
	memset( pkey + kl, 0, klen - kl );
    if( (new = ((pn = hfind( (uchar*) pkey, strcmp )) == 0)) )
	pn = avail( (ui16) type );
    /**/
    memcpy( pn-> key, pkey, klen ), memcpy( buff, pkey, klen );
    pn-> crc = -1; /*update...*/
    /**/

    unprotectBBM();

    if( lseek( e2fd, pn-> recn, SEEK_SET ) < 0 )
    {
	logMsg( "lseek(PR) '%s': 0x%x\n", (int)e2fwork, errnoGet(),0,0,0,0 ), taskSuspend(0);
    }

    if (((struct dbCommon *)buff)->lset) /* net std order doesn't matter here */
    {
	logMsg("LSET IS SET, suspending task\n",0,0,0,0,0,0);
	taskSuspend(0);
    }
    if( write( e2fd, (char *)&type, sizeof(type) ) < 0 ||
      write( e2fd, buff, len ) < 0 || /*'full' record including key*/
      write( e2fd, (char *)&pn-> crc, sizeof(pn-> crc) ) < 0 )
    {
	logMsg( "write(PR) '%s': 0x%x\n", (int)e2fwork, errnoGet(),0,0,0,0 ), taskSuspend(0);
    }
    /**/
    if( new )
    {
	hput( (uchar*) pkey, pn ), ++numpnts, ++numrecs;
	if( lseek( e2fd, 0L, SEEK_SET ) < 0 ||
	  write( e2fd, (char *)&numrecs, sizeof(numrecs) ) < 0 )
	{
	    logMsg( "write(PR) '%s': 0x%x\n", (int)e2fwork, errnoGet(),0,0,0,0 ), taskSuspend(0);
	}
    }
    /**/
    if( (e2fd = e2flush()) < 0 )
    {
	logMsg( "flush(PR) '%s': 0x%x\n", (int)e2fwork, errnoGet(),0,0,0,0 ), taskSuspend(0);
    }
    logMsg( "PutRecord (%d:%s) ='%.*s' (..%d)\n", 
	 numrecs, (int)(new?"new":"old"), PVNAME_SZ, (int)pkey, len,0 );

    protectBBM();

}

void DelRecord( char *pkey )
{
    register int kl = strlen( pkey );
    register ui32 hk;
    register bin **ppb, *pb;
    register node *pn = NULL;
    /**/
    if( kl < klen )
	memset( pkey + kl, 0, klen - kl );
    hk = khash( (uchar*) pkey, klen, oM );
    ppb = &bins[ hk ], pb = *ppb;
    /**/
    while( pb != 0 )
    {
	pn = (node*) pb-> d;
	if( ! strcmp( pkey, pn-> key ) ) break;
	else ppb = (bin**) &pb-> l.l, pb = *ppb;
    }
    if( pb == 0 )
    {
	logMsg( "can't happen (DR)!\n",0,0,0,0,0,0), taskSuspend(0);
    }
    *ppb = (bin*) pb-> l.l;
    /**/
    memset( pn-> key, 0, sizeof(pn-> key) );

    unprotectBBM();

    if( lseek( e2fd, pn-> recn + sizeof(pn-> type), SEEK_SET ) < 0 ||
      write( e2fd, pn-> key, sizeof(pn-> key) ) < 0 )
    {
	logMsg( "write(DR) '%s': 0x%x\n", (int)e2fwork, errnoGet(),0,0,0,0 ), taskSuspend(0);
    }
    pn-> crc = -1;
    pb-> l.l = empty[ pn-> type ], empty[ pn-> type ] = &pb-> l;
    --numpnts;
    /**/
    if( (e2fd = e2flush()) < 0 )
    {
	logMsg( "flush(PR) '%s': 0x%x\n", (int)e2fwork, errnoGet(),0,0,0,0 ), taskSuspend(0);
    }
logMsg( "DelRecord ='%.*s'\n", PVNAME_SZ, (int)pkey,0,0,0,0 );

    protectBBM();

}

void e2dbgShowPoint( char *pkey )
{
    register ui32 hk = khash( (uchar*) pkey, klen, oM );
    register node *pn = hfind( (uchar*) pkey, strcmp );
    /**/
    if( pn != 0 )
	printf( "'%-*.*s' bin(%3lu) type(%2hd) rec#:%8lu\n", PVNAME_SZ, PVNAME_SZ, pkey, hk, pn-> type, pn-> recn );
    else
	printf( "'%-*.*s' not found!\n", PVNAME_SZ, PVNAME_SZ, pkey );
}

void e2dbgShowPoints()
{
    char pName[PVNAME_SZ];
    short pType;
    /**/
    memset( pName, 0, sizeof(pName) );
    printf( "%ld points active\n", numpnts );
    printf( "%ld records in file\n", numrecs );
    while( (pType = GetType( pName )) != 0 )
	e2dbgShowPoint( pName );
}

void e2dbgDumpFile()
{
    char *fmt = "%8u %-48.48s %-16.16s\n";
    ui32 nrecs, recp;
    int fd;
    int n = 0;
    uchar hb[48+1], tx[16+1], c;
    /**/
    unprotectBBM();

    fd = open(e2fwork, READ, 0);
    if( fd < 0 )
    {
	printf( "** unable to open file\n" );
	return;
    }
    /**/
    if( lseek( fd, 0L, SEEK_SET ) < 0 ||
      read( fd, (char *)&nrecs, sizeof(nrecs) ) < sizeof(nrecs) )
    {
	printf( "** error: %x\n", errnoGet() ); return;
    }
    /**/
    recp = sizeof(nrecs), hb[0] = tx[0] = 0;
    while( read( fd, &c, 1 ) == 1 )
    {
	sprintf( hb + strlen(hb), " %02x", c );
	sprintf( tx + strlen(tx), "%c", c<32||c>126 ? '.' : c );
	if( ++n == 16 )
	    printf( fmt, recp, hb, tx ), hb[0] = tx[0] = n = 0, recp += 16;
    }
    if( n )
	printf( fmt, recp, hb, tx );
    /**/
    printf( "#recs: %lu (%u bytes)\n", nrecs, lseek( fd, 0L, SEEK_CUR ));
    close( fd );

    protectBBM();

}

void e2dbgStats( BOOL tablestats )
{
    register int i, k;
    /**/
    printf( "#pnts: %lu / #recs: %lu\n", numpnts, numrecs );
    for( k=0,i=1; i <= guSDR_NTYPES; ++i )
    {
	register int j = 0;
	register bin *pb = (bin*) empty[i];
	while( pb != 0 )
	    ++j, pb = (bin*) pb-> l.l;
	if( j )
	    printf( "#spare (type %2d): %d\n", i, j ), k += j;
    }
    if( k != numrecs - numpnts )
	printf( "** spare count mismatch: %d,%ld\n", k, numrecs - numpnts );
    /**/
    for( k=0,i=0; i < oM; ++i )
    {
	register bin *pb = bins[i];
	while( pb != 0 )
	    ++k, pb = (bin*) pb-> l.l;
    }
    if( k != numpnts )
	printf( "** record count mismatch: %d,%ld\n", k, numpnts );
    else
    {
	register float dist = (float) numpnts / oM, phit = dist / numpnts;
	register int m, hi = 0, lo = oM;
	/**/
	printf( "hash stats:\n" );
	for( i=0; i < oM; ++i )
	{
	    register int j = 0;
	    register bin *pb = bins[i];
	    while( pb != 0 )
		++j, pb = (bin*) pb-> l.l;
	    /**/
	    if( j )
	    {
		if( j < lo ) lo = j;
		if( j > hi ) hi = j;
		/**/
		if( tablestats )
		    printf( "bin #%3d: %3d (%7.3f%%) (%7.3f*d) (%7.3f*h)\n",
			   i, j, j * 100. / numpnts, (float) j / dist, ((float) j / numpnts) - phit );
	    }
	}
	m = lo + ((hi - lo) >> 1);
	/**/
	printf( "lo: %d / hi: %d / avg: %d / oM: %ld\n", lo, hi, m, oM );
    }
}

#if ((CPU == I80386) || (CPU == I80486))
int protectBBM()
{
    return (0);
}

int unprotectBBM()
{
    return (0);
}


int disableBBMProtect()
{
    return (0);
}


int enableBBMProtect()
{
    return (0);
}

int sizeBBM()
{
    return (0);
}
#else
int protectBBM()
{
    if (!bbmSize)
	sizeBBM();
    if (protectionDisabled)
	return (0);
    vmStateSet(NULL, sysMemTop(), 0x400000, VM_STATE_MASK_WRITABLE, VM_STATE_WRITABLE_NOT);
    return (0);
}

int unprotectBBM()
{
    if (!bbmSize)
	sizeBBM();
    vmStateSet(NULL, sysMemTop(), 0x400000, VM_STATE_MASK_WRITABLE, VM_STATE_WRITABLE);
    return (0);
}


int disableBBMProtect()
{
    printf("BBM protection now disabled.\n");
    protectionDisabled = 1;
    unprotectBBM();
    return (0);
}


int enableBBMProtect()
{
    printf("BBM protection now enabled.\n");
    protectionDisabled = 0;
    return (0);
}

int sizeBBM()
{
    char	*probe;
    int		contents;

    probe = (char *) sysMemTop();

    while (vxMemProbe(probe, VX_READ, sizeof(contents), (char *) &contents) != ERROR)
	{
	probe += 0x100000;
	bbmSize += 0x100000;
	}

    return (0);
}

#endif cpu == I80386
