/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*============================================================================*/
/*									      */
/*									      */
/*	Tate Integrated Systems, LP					      */
/*									      */
/*									      */
/*	Author:			Greg Tomalesky				      */
/*	Date:			29-Jun-1994				      */
/*									      */
/*	Function Name:		DB_GetVerifyInfo			      */
/*	Function Description:						      */
/*									      */
/*									      */
/*									      */
/*									      */
/*	Input Parameters:	PrimaryKey   - primary key in record.         */
/*				TableHandle  - pointer to table structure.    */
/*				verify_info  - pointer to VERIFY_INFO struct  */
/*									      */
/*	Output Parameters:	none					      */
/*									      */
/*	Return Values or error codes:	error		0 bad record	      */
/*					success:	record handle	      */
/*									      */
/*	Modification History:	29-Jun-1994	First time		      */
/*									      */
/*									      */
/*									      */
/*									      */
/*============================================================================*/
#include <DBDefines.h>
#include <DBCVerifyDefs.h>
#include <DBCDefines.h>
#include <dbRecType.h>

int	DB_GetVerifyInfo(TableHandle,pvName,verify_info,import_info)
DB_TABLE_PTR	TableHandle;
char		*pvName;
VERIFY_INFO_PTR	verify_info;
IMPORT_INFO_PTR	import_info;
{
int		stat,typelen;
char	tb[16];
short	index;
char		*ts;
F4FIELD		*timeStamp,*downLoad1,*downLoad2,*deleted,*type,*name,*impdel,*impts,*impname;
F4FIELD		*crc;	/* 8-Jan-97 added crc to inter-host message */

if (TableHandle == 0)
  return(0);

timeStamp	= d4field(TableHandle->table,VI_FLD_TIMESTAMP);
downLoad1	= d4field(TableHandle->table,VI_FLD_LOAD1FLAG);
downLoad2	= d4field(TableHandle->table,VI_FLD_LOAD2FLAG);
deleted		= d4field(TableHandle->table,VI_FLD_DELETEFLAG);
type		= d4field(TableHandle->table,VI_FLD_POINTTYPE);
name		= d4field(TableHandle->table,VI_FLD_NAME);
crc		= d4field(TableHandle->table,VI_FLD_CHECKSUM);

if(pvName)
  {
    if((stat = d4seek(TableHandle->table,pvName)) != 0)
      {
		d4unlock_all(TableHandle->table);
		error4set(&DB_CodeBase,0);
		return(0);
      }
  }
verify_info->viPVStatus = 0;										/* clear before setting bits */
verify_info->viPVStatus |= (f4true(downLoad1) == 1 ? VI_LOADED_1 : 0);	/* set status flags */
verify_info->viPVStatus |= (f4true(downLoad2) == 1 ? VI_LOADED_2 : 0);
verify_info->viPVStatus |= (f4true(deleted)   == 1 ? VI_DELETED  : 0);
/*
  get the CRC for this stuff now
*/
strcpy(tb,f4str(crc));
verify_info->viCRC = (short)strtol(tb,NULL,16);
gethostname(&verify_info->viHostName[0],DBC_IOCNAMESIZE);			/* get host name */
verify_info->viTimeStamp = (time_t)atol(f4ptr(timeStamp));			/* get time stamp */
strncpy(&verify_info->viPVName[0],f4ptr(name),PVNAME_SZ);			/* get point name */
typelen = f4len(type);
ts = f4str(type);
for(index = 1;GET_PRECTYPE(index);index++)
  {
    sprintf(tb,"%-*.*s\0",typelen,typelen,GET_PRECTYPE(index));
    if(!strncmp(ts,tb,typelen))
      break;
  }
if((index < VU_AI) || (index > VU_MAX_TYPES))
  {
    d4unlock_all(TableHandle -> table);
    return(0);				/* point type not found return error */
  }
verify_info->viTypeIndex = index;
/*
	if the 'import_info' pointer is not NULL then
	retrieve the data from Pvmap.
*/
if(import_info)
  {
    impdel		= d4field(TableHandle->table,VI_FLD_IMPDELFLAG);
    impts		= d4field(TableHandle->table,VI_FLD_IMPORTDATE);
    impname		= d4field(TableHandle->table,VI_FLD_IMPORTNODE);
    import_info->viImportDate = (time_t)atol(f4ptr(impts));	       		/* get remote time stamp */
    strncpy(&import_info->viImportNode[0],f4ptr(impname),DBC_IOCNAMESIZE);	/* get point name */
    import_info->viImpDelFlag = f4true(impdel);		       			/* get remote deletion status flag */
  }
d4unlock_all(TableHandle -> table);
return(1);
}
