/*DBInitDatabase.c,v 1.1.1.1 1999/08/03 14:42:40 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*============================================================================*/
/*									      */
/*									      */
/*	Tate Integrated Systems, LP					      */
/*									      */
/*									      */
/*	Author:			Greg Tomalesky				      */
/*	Date:			14-Jan-1993				      */
/*									      */
/*	Function Name:		DB_InitDatabase				      */
/*	Function Description:						      */
/*									      */
/*									      */
/*									      */
/*									      */
/*	Input Parameters:	None.					      */
/*									      */
/*	Output Parameters:	none					      */
/*									      */
/*	Return Values or error codes: none.				      */
/*									      */
/*	Modification History:	05-Jan-1993	First time		      */
/*									      */
/*									      */
/*									      */
/*									      */
/*============================================================================*/
#include <DBDefines.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#ifdef WIN32
#include <process.h>
#endif WIN32

FILE	*dbLog;		/* Log file pointer */
int	dbPid;		/* PID of current program */
int	dbJournal;	/* verbosity level */
char	dbPName[MAXPATHLEN];	/* program name from argv[0] */
int     lockFd;		/* fd of lock file */

char	DB_DatabasePath[MAXPATHLEN];
C4CODE	DB_CodeBase;
static	char first_time = 1;			/* so d4init only called once */

void	DB_InitDatabase(pgmname)
char	*pgmname;
{
if(first_time)
	{
	char	*jour;
	d4init(&DB_CodeBase);
	DB_CodeBase.errFieldName = 0;
	DB_CodeBase.errOpen = 0;
	DB_CodeBase.errTagName = 0;
	DB_CodeBase.readLock = 0;
#ifndef CB51
	DB_CodeBase.wait = 0;
#endif
	DB_CodeBase.errCreate = 0;
	DB_CodeBase.errGo = 0;	/* 18-apr-1996 */
	DB_CodeBase.errOff = 1;	/* 19-Apr-1996 turn off error display */
	first_time = 0;
/*
	Attempt to open log file as /tis4000db/dbaseLog.log. If the file is there
	then OK else set dbLog to zero. Get the PID of the calling process here and
	the verbosity level as set by the DBJOURNAL environment variable.
*/
	dbLog = fopen("/tis4000db/dbaseLog.log","r+");	/* if error then dbLog will be NULL */
	if(dbLog) fseek(dbLog,0,2);			/* position to end of file */
#ifdef WIN32
	dbPid = _getpid();				/* calling process PID */
#else
	dbPid = getpid();				/* calling process PID */
#endif WIN32
	strcpy(dbPName,pgmname);			/* save program name too */
	dbJournal = 0;					/* default level is "don't print anything" */
	if(jour = getenv("DBJOURNAL"))			/* get value if present */
		dbJournal = atoi(jour);
/* The following open will attempt to open the lock file. This file's */
/* sole purpose is to give the lockf() function something to lock. In the */
/* vuCreateRecords() & DB_CrFullRec() functions, this file is locked */
/* during file extension (d4append) and unlocked afterwards. If either */
/* dataVu or dbClient detects a lock it waits until it can get the lock */
/* and then proceeds to update (extend) the database files. */
	lockFd = open("/tis4000db/dbLock.lock",(O_CREAT|O_RDWR),0666);
	}
return;
}
/* This function will attempt to lock the dbLock.lock file. If this fails because */
/* somebody already has the file locked (EACCES) then the lock operation will be */
/* retried until it is successful or the number of retries has been exceeded. */
int dbLockTest(int retries)
{
#ifdef WIN32
	return(1);
#else
int stat;
if(retries <= 0) retries = 1;
while(retries)
  {
    stat = lockf(lockFd,F_TLOCK,0);
    if(stat == 0)
      return(1);		/* got the lock...OK to proceed */
    else
      {
		if((errno == EACCES) || (errno == ENOLCK))
		  {
		    retries--;		/* try again */
		    sleep(1);		/* sleep for a second */
		    continue;
		  }
      }
  }
return(0);			/* retries exceeded */
#endif WIN32
}
/* This function will unlock the dbLock file. This would be done after */
/* all file I/O has been completed on the database files. */
int dbUnLock()
{
#ifndef WIN32
return(lockf(lockFd,F_ULOCK,0));
#else
return(1);
#endif
}

