/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "descDataTable.h"
#include "dbRecType.h"
#include "msgMon.h"

static char separator[] = "\t\v\r\n\f";
static int  linenum = 0;

int processLimit(limitStruct *currLimit, typeDesc *currDesc);

typeDesc **bldDesDataTable()
{
    typeDesc		**pointerTable;
    typeDesc		*currDesc;
    char		buff[132];
    char		*pChar;
    char		*pstr;
    FILE		*fd;
    int			i;
    int			readFlag;

    /*
    The first thing to do is to open the text file. If we can't do that,
    then there is no point in continuing.
    */

    if ((fd = fopen("/tis4000/tasks/caDataDesc.txt", "r")) == NULL)
	{
	perror("Can't open /tis4000/tasks/caDataDesc.txt");
	return (NULL);
	}

    /*
    Now figure out the largest number of structures that we will
    ever have.
    */

    pointerTable = (typeDesc **) calloc(dbRecType -> number, sizeof(typeDesc *));

    /*
    Now read lines out of the source file and discard them until we
    find a "TYPE" command.
    */

    pChar = 0;
    readFlag = 0;
    for (;;)
	{
	if (!pChar)
	{
	    if (fgets(buff, sizeof(buff), fd) == NULL)
		break;
	    readFlag = 1;
	    linenum++;
	}

	/*
	If the first character on the line is a '#', then it is a comment.
	Ignore the line.
	*/

	if (buff[0] == '#')
	{
	    readFlag = 0;
	    continue;
	}

	/*
	Find the first non-blank in the line. If there are none,
	then the line is blank and should also be ignored.
	*/

	if (readFlag)
	    pChar = strtok(buff, separator);
	else
	    pChar = strtok(NULL, separator);
	if (!pChar)
	{
	    readFlag = 0;
	    continue;
	}

	/*
	We should be pointed at a "TYPE" command. If not, abandon processing.
	*/

	if (strcmp(pChar, "TYPE"))
	    {
	    tisSysLog (0, 0, "TYPE command not found when expected. Run abandoned.\n");
	    fclose (fd);
	    return (NULL);
	    }

	/*
	It's OK so far. Now find the point type.
	*/

	pChar = strtok(NULL, separator);
	if (!pChar)
	    {
	    tisSysLog (0, 0, "TYPE command with no operand. Run abandoned.\n");
	    fclose (fd);
	    return (NULL);
	    }

	/*
	Find the point type.
	*/

	for (i = 0; i < dbRecType -> number; i++)
	    {
	    if ( (pstr = GET_PRECTYPE(i)) == NULL)
		continue;
	    if ( !strcmp(pChar, pstr))
		break;
	    }
	if (i >= dbRecType -> number)
	    {
	    tisSysLog (0, 0, "TYPE command with unknown point type.\n");
	    fclose (fd);
	    return (NULL);
	    }

	/*
	Allocate a type header for this point type.
	*/

	currDesc = (typeDesc *) calloc(1, sizeof(typeDesc));
	pointerTable[i] = currDesc;

	/*
	Now read and process the various commands that we need to describe
	this point type.
	*/
	readFlag = 0;
	for (;;)
	    {
	    enumField 		*newEnum;

	    if (!pChar)
	    {
	        if (fgets(buff, sizeof(buff), fd) == NULL)
		    break;
		readFlag = 1;
		linenum++;
	    }

	    /*
	    Same tests for comments or blank lines as used above.
	    */

	    if (buff[0] == '#')
	    {
		readFlag= 0;
		continue;
	    }
	    if (readFlag)
		pChar = strtok(buff, separator);
	    else
		pChar = strtok(NULL, separator);
	    if (!pChar)
	    {
		readFlag = 0;
		continue;
	    }

	    /*
	    If we see a TYPE command, we are done here and should move
	    on to the next point type.
	    */

	    if (!strcmp(pChar, "TYPE"))
	    {
		buff[strlen(pChar)] = '\t';
		break;
	    }

	    /*
	    Now see if the command is one that we know about.
	    */

	    if (!strcmp(pChar, "DEFCTLLOW"))
		{

		/*
		Default Control Low.
		*/

		currDesc -> defaultControlLow = (limitStruct *) calloc(1, sizeof(limitStruct));
		if (processLimit(currDesc -> defaultControlLow, currDesc))
		{
		    fclose (fd);
		    return (NULL);
		}
		pChar = 0;

		}

	    else if (!strcmp(pChar, "DEFCTLHIGH"))
		{

		/*
		Default Control High.
		*/

		currDesc -> defaultControlHigh = (limitStruct *) calloc(1, sizeof(limitStruct));
		if (processLimit(currDesc -> defaultControlHigh, currDesc))
		{
		    fclose (fd);
		    return (NULL);
		}
		pChar = 0;

		}

	    else if (!strcmp(pChar, "DEFDISPLOW"))
		{

		/*
		Default Display  Low.
		*/

		currDesc -> defaultDisplayLow = (limitStruct *) calloc(1, sizeof(limitStruct));
		if (processLimit(currDesc -> defaultDisplayLow, currDesc))
		{
		    fclose (fd);
		    return (NULL);
		}
		pChar = 0;

		}

	    else if (!strcmp(pChar, "DEFDISPHIGH"))
		{

		/*
		Default Display High.
		*/

		currDesc -> defaultDisplayHigh = (limitStruct *) calloc(1, sizeof(limitStruct));
		if (processLimit(currDesc -> defaultDisplayHigh, currDesc))
		{ 
		    fclose (fd);
		    return (NULL);
		}
		pChar = 0;

		}

	    else if (!strcmp(pChar, "DEFPREC"))
		{

		/*
		Default Precision.
		*/

		currDesc -> defaultPrec = (limitStruct *) calloc(1, sizeof(limitStruct));
		if (processLimit(currDesc -> defaultPrec, currDesc))
		{
		    fclose (fd);
		    return (NULL);
		}
		pChar = 0;

		}

	    else if (!strcmp(pChar, "ENUM"))
		{

		/*
		Enumeration List. Allocate a structure and then put it into
		the chain at the head.
		*/

		newEnum = (enumField *) calloc(1, sizeof(enumField));
		newEnum -> nextEnum = currDesc -> enumList;
		currDesc -> enumList = newEnum;

		/*
		Find the field name and the number of fields.
		*/

		pChar = strtok(NULL, separator);
		if (!pChar)
		    {
		    tisSysLog (0, 0, "ENUM statement without field name. Run abandoned.\n");
		    fclose (fd);
		    return (NULL);
		    }

		strncpy(newEnum -> fieldName, pChar, sizeof(newEnum -> fieldName));
		pChar = strtok(NULL, separator);
		if (!pChar)
		    {
		    tisSysLog (0, 0, "ENUM statement without number of fields. Run abandoned.\n");
		    fclose (fd);
		    return (NULL);
		    }
		newEnum -> numStrings = atoi(pChar);

		/*
		Now find the individual fields that we need. We ignore blank lines and
		comments here, too.
		*/

		for (i = 0; i < newEnum -> numStrings; )	/* Notice: no increment on purpose! */
		    {
		    if (fgets(buff, sizeof(buff), fd) == NULL)
			{
			tisSysLog (0, 0, "EOF for ENUM strings when not expected. Run abandoned.\n");
			fclose (fd);
			return (NULL);
			}
		    readFlag = 1;
		    linenum++;
		    if (buff[0] == '#')
		    {
			readFlag = 0;
			continue;
		    }
		    pChar = strtok(buff, separator);
		    if (!pChar)
		    {
			readFlag = 0;
			continue;
		    }

		    /*
		    This must be a field name.
		    */

		    strncpy(newEnum -> refField[i], pChar, 4);
		    i++;					/* Increment because we found one! */
		    }

		/*
		Indicate that we should read another command line.
		*/

		buff[strlen(pChar)] = '\t';
		pChar = 0;
		readFlag = 0;
		}

	    else if (!strcmp(pChar, "EXCEPTION"))
		{
		numericField		*numException;

		/*
		Numeric Exception List field. We'll ignore blank lines and comments.
		Process the commands that we know. If we see one we don't know,
		then assume it's another major command. Break out into the main
		loop for the type with pChar pointing to the key word.
		*/

		numException = (numericField *) calloc(1, sizeof(numericField));
		numException -> nextException = currDesc -> exceptions;
		currDesc -> exceptions = numException;
		/*
		Find the field name
		*/

		pChar = strtok(NULL, separator);
		if (!pChar)
		    {
		    tisSysLog (0, 0, "EXCEPTION statement without field name. Run abandoned.\n");
		    fclose (fd);
		    return (NULL);
		    }

		strncpy(numException -> fieldName, pChar, sizeof(numException -> fieldName));

		for (;;)
		    {
		    pChar = 0;
		    if (fgets(buff, sizeof(buff), fd) == NULL)
			break;
		    linenum++;

		    /*
		    Handle comments and blank lines.
		    */

		    if (buff[0] == '#')
			continue;
		    pChar = strtok(buff, separator);
		    if (!pChar)
			continue;

		    /*
		    Look for a command.
		    */

		    if (!strcmp(pChar, "CTLLOW"))
			{
			numException -> controlLow = (limitStruct *) calloc(1, sizeof(limitStruct));
			if (processLimit(numException -> controlLow, currDesc))
			{
			    fclose (fd);
			    return (NULL);
			}
			pChar = 0;
			}

		    else if (!strcmp(pChar, "CTLHIGH"))
			{
			numException -> controlHigh = (limitStruct *) calloc(1, sizeof(limitStruct));
			if (processLimit(numException -> controlHigh, currDesc))
			{
			    fclose (fd);
			    return (NULL);
			}
			pChar = 0;
			}

		    else if (!strcmp(pChar, "DISPLOW"))
			{
			numException -> displayLow = (limitStruct *) calloc(1, sizeof(limitStruct));
			if (processLimit(numException -> displayLow, currDesc))
			{
			    fclose (fd);
			    return (NULL);
			}
			pChar = 0;
			}

		    else if (!strcmp(pChar, "DISPHIGH"))
			{
			numException -> displayHigh = (limitStruct *) calloc(1, sizeof(limitStruct));
			if (processLimit(numException -> displayHigh, currDesc))
			{
			    fclose (fd);
			    return (NULL);
			}
			pChar = 0;
			}
		    else if (!strcmp(pChar, "PREC"))
			{
			numException -> precision = (limitStruct *) calloc(1, sizeof(limitStruct));
			if (processLimit(numException -> precision, currDesc))
			{
			    fclose (fd);
			    return (NULL);
			}
			pChar = 0;
			}
		    else
			{

			/*
			Unknown command. Get into the outer loop.
			*/
			buff[strlen(pChar)] = '\t';
			break;
			}
		    }
		}
	    }

	if (feof(fd))
	    break;
	}
    fclose (fd);
    return (pointerTable);
}

/************************************************************************/
/*									*/
/*	PROCESS A COMMAND LINE THAT CONTAINS A LIMIT OR PRECISION	*/
/*									*/
/************************************************************************/
int processLimit(currLimit, currDesc)
limitStruct	*currLimit;
typeDesc	*currDesc;
{
    char	*currToken;
    char	*nextToken;

    /*
    The next token should be "REF", "CONST", or "SPECL"
    They all take operands.
    */

    currToken = strtok(NULL, separator);
    if (!currToken)
	{
	tisSysLog (0, 0, "Limit statement missing an operator.\n");
	return(1);
	}
    nextToken = strtok(NULL, separator);
    if (!nextToken)
	{
	tisSysLog (0, 0, "Limit statement missing an operand.\n");
	return(1);
	}

    if (!strcmp(currToken, "REF"))
	{

	/*
	A reference to another field in the same record.
	*/

	currLimit -> limitType = 1;
	strncpy(currLimit -> reference, nextToken, sizeof(currLimit -> reference));

	/*
	Allocate a slot number to hold this value.
	*/

	currLimit -> tableSlot = currDesc -> tableSize++;
	}

    else if (!strcmp(currToken, "CONST"))
	{

	/*
	A constant.
	*/

	currLimit -> limitType = 0;
	currLimit -> hardConst = atof(nextToken);
	}

    else if (!strcmp(currToken, "SPECL"))
	{

	/*
	Special Processing.
	*/

	currLimit -> limitType = 2;
	currLimit -> backDoor = atoi(nextToken);
	/*
	Allocate a slot number to hold this value.
	*/

	currLimit -> tableSlot = currDesc -> tableSize++;
	}

    else
	{
	tisSysLog (0, 0, "Invalid operand type -%s- in a limit statement.\n", currToken);
	return(1);
	}
    return (0);
}
