/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include	<stdio.h>
#include	<string.h>
#ifdef WIN32
#include	<time.h>
#else
#include	<sys/time.h>
#include	<sys/param.h>
#include	<types.h>
#endif WIN32
#include	<sys/stat.h>
#include	<memory.h>
#include	<errno.h>
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include	"db_access.h"	/* for def of MAX_STRING_SIZE */
#include	"dbDefs.h"
#include	"dbRecType.h"
#include	"dbRecDes.h"
#include	"choice.h"
#include	"choiceGbl.h"
#include	"devSup.h"
#include	"drvSup.h"
#include	"cvtTable.h"
#include	"recSup.h"
#include	"special.h"
#include	"dbRecords.h"
#include	"dbFldTypes.h"
#include	"descDataTable.h"
#include	"fieldAttributes.h"
#include	"msgMon.h"

/* sizes of field types as defined in dbFldTypes.h
   NOTE: this must agree with same in devCom/devSoft.c
*/
int sizeofTypes[] = {0,1,1,2,2,4,4,4,8,2};

typeDesc	**descTable;

SPECL_PASS	udbpass;
UDBPVS_PTR	udbpvs;

#ifdef EDSTEST
void main (void)
{
    time_t elap;

    tisOpenLog ("edstest", 6, 2, NULL, NULL);
    elap = time(0);
    tisSysLog(0,0,"Start time %s",ctime(&elap));
    if (FA_InitAttributes ())
	tisSysLog (0,0,"init failed\n");
    elap = time(0);
    tisSysLog(0,0,"End time %s",ctime(&elap));
    d4close_all (&FA_CodeBase);
}
#endif


/***************************************************************
*
*    FA_convert_dbAddr ()
*	certain point types need to modify the dbAddr struct
*
***************************************************************/

void	FA_convert_dbAddr (struct dbAddr *pdbAddr, char *pvname)
{
    char	*pstr;
    char	*waveformFieldType  = (char *)0;
    char	pv[PVNAME_SZ+FLDNAME_SZ+1];
    char        fname[256];
    int		lenFieldType;
    int		i;
    DATA4	*waveformDb;
    FIELD4	*waveformNELM;
    FIELD4	*waveformFTVL;

    pstr = GET_PRECTYPE (pdbAddr -> record_type);
    if ( (strncmp (pstr, "waveform", 8)) == 0)
    {
	/* if the choice table is not loaded, error */
	if (choiceGbl == NULL)
	{
	    tisSysLog (0,0,"convert_dbaddr: choiceGbl not loaded\n");
	    return;
	}
	sprintf(fname,"%swaveform",FA_DATABASEPATH);
	if ( (waveformDb = getFile(fname,pdbAddr -> record_type )) == 0)
	{
	    tisSysLog (0,0,"FA_convert_dbAddr: failed to open %s for %s\n",fname,pvname);
	    return;
	}
	if ( (waveformNELM = d4field (waveformDb, "NUMELMENTS")) == 0)
	{
	    tisSysLog (0,0,"FA_convert_dbAddr: failed to find \"NUMELMENTS\" for %s\n",
		     pvname);
	    return;
	}
	if ( (waveformFTVL = d4field (waveformDb, "FIELDTYPE")) == 0)
	{
	    tisSysLog (0,0,"FA_convert_dbAddr: failed to find \"FIELDTYPE\" for %s\n",
		     pvname);
	    return;
	}
	strncpy (pv, pvname, PVNAME_SZ+FLDNAME_SZ);
	if ( (pstr = strrchr (pv, '.')) != NULL)
	{
	    *pstr = 0;
	}
	if ( (d4seek (waveformDb, pv)) != 0)
	{
	    tisSysLog (0,0,"FA_convert_dbAddr: failed to find %s\n", pv);
	    return;
	}
	/* now tiptoe through the choice list and try to match what is in
	   the database, the value if i, becomes the field type and is also
	   the index into the array of sizes (see fieldAttributes.h)
	*/
	lenFieldType=  f4len (waveformFTVL);
	if ((waveformFieldType = (char *)malloc (lenFieldType + 1)) == NULL)
	{
	    tisSysLog (0,0,"FA_convert_dbAddr: malloc failed for %s\n",pvname);
	    return;
	}
	memset (waveformFieldType, ' ', lenFieldType + 1);
	strncpy (waveformFieldType, f4str (waveformFTVL), lenFieldType);
	*(waveformFieldType + lenFieldType) = 0;
	FA_removeTrailingSpaces (waveformFieldType);
	
	for (i = 0; i < choiceGbl -> papChoiceSet[GBL_FTYPE] -> number; i++)
	{
	    if ( (strcmp (waveformFieldType, 
			  choiceGbl -> papChoiceSet[GBL_FTYPE] -> papChoice[i])) == 0)
	    {
		pdbAddr -> field_type = i;
		pdbAddr -> dbr_field_type = i;
		if (i == 0)
		{
		    pdbAddr -> field_size = MAX_STRING_SIZE;
		}
		else
		{
		    pdbAddr -> field_size = sizeofTypes[i];
		}
		pdbAddr -> no_elements = f4long (waveformNELM);
		break;
	    }
	}
	if (waveformFieldType != (char *)0)
	    free ((void *)waveformFieldType);
    }
    return;
}


/**************************************************************
*
*    getAttributes ()
*	aruments given:
*	    char *pvname - pointname.FLD, if no FLD assume VAL
*	    int  pointType - index into sdr *dbRecType
*	    int  maxString - max num string that can be returned
*			     for an ENUM
*	    char **stringList - Array of ptrs to store ENUM strings
*	    int  longestString - max length of any string, including NULL
*	arguments returned:
*	    short *nativeType - defined in epicsH/dataTypes.h
*	    short *nativeCount - native array count, 1 if not an array
*
*	    these fields applicable if datatype id INT, SHORT, FLOAT
*	    CHAR, DOUBLE, prec applicable only to FLOAT and DOUBLE:
*	    
*	    double *hopr - display high limit
*	    double *lopr - display low limit
*	    double *drvh - upper control limit
*	    double *drvl - lower control limit
*	    short  *prec - display precision
*
*	return the field attributes as requested
*
**************************************************************/

int	getAttributes (char *pvname, int pointType, int	maxString,
		       char stringList[][32], int longestString,
		       unsigned short crc,unsigned timestamp,
		       short *nativeType, short *nativeCount,
		       double *hopr, double *lopr, double *drvh,
		       double *drvl, short *prec,short *numStrings)
{
    struct dbAddr	dbAddr;
    struct choiceSet	*pchoiceSet;
    struct arrChoiceSet	*parrChoiceSet;
    struct devChoice	*pdevChoice;
    struct devChoiceSet	*pdevChoiceSet;
    int			no_strs;
    int			i;
    char *str;
    double dblPrec;
    int nstr;

    *hopr = 0;
    *lopr = 0;
    *drvh = 0;
    *drvl = 0;
    *prec = 0;
    *numStrings = 0;

    if (FA_dbNameToAddr (pvname, pointType, &dbAddr))
    {
	return (-1);
    }

    switch (dbAddr.dbr_field_type)
    {
	/* nothing to do with a string */
	case DBR_STRING:
	    break;

	case DBR_ENUM:
	    if (dbAddr.field_type == DBF_ENUM)
	    {
	      *numStrings = 0;
	      FA_FetchData(descTable,pvname,(short)pointType,
			   crc,timestamp,lopr,hopr,drvl,drvh,
			   &dblPrec,&str,&nstr);
	      *prec = (short)dblPrec;
	      if(str)
		{
		  if(nstr > maxString) nstr = maxString;
		  for(i = 0;i < nstr; i++)
		    {
		      strncpy(&stringList[i][0],str,longestString);	/* copy string into supplied array */
		      stringList[i][longestString-1] = 0;		/* leave space for NULL */
		      if((*str != 0) ||
			 (!strncmp(GET_PRECTYPE (pointType),"bo",2)) ||
			 (!strncmp(GET_PRECTYPE (pointType),"bi",2))) *numStrings = i + 1;
		      str += (FA_STRING_SIZE+1);       			/* point to next string */
		    }
		}
	    }
	    else
	    {
		switch (dbAddr.field_type)
		{
		    case DBF_GBLCHOICE:
			pchoiceSet = GET_PCHOICE_SET(choiceGbl,dbAddr.choice_set);
			goto choice_common;
		    case DBF_CVTCHOICE:
			pchoiceSet = choiceCvt;
			goto choice_common;
		    case DBF_RECCHOICE:
			parrChoiceSet = GET_PARR_CHOICE_SET(choiceRec,dbAddr.record_type);
			pchoiceSet = GET_PCHOICE_SET(parrChoiceSet,dbAddr.choice_set);
choice_common:
			if (pchoiceSet == NULL)
			{
			    tisSysLog (0,0,"getAttributes: choice set for %s does not exist\n",
				     pvname);
			    return (-1);
			}
			if (pchoiceSet -> number <= maxString)
			    no_strs = pchoiceSet -> number;
			else
			    no_strs = maxString;
			for (i = 0; i < no_strs; i++)
			{
			    if (pchoiceSet -> papChoice[i] == NULL)
				stringList[i][0] = 0;
			    else
			    {
				strncpy (&stringList[i][0], pchoiceSet -> papChoice[i], longestString);
				stringList[i][longestString-1] = 0;
			    }
			}
			*numStrings = (short)no_strs;
			break;

		    case DBF_DEVCHOICE:
			pdevChoiceSet = GET_PDEV_CHOICE_SET(choiceDev,dbAddr.record_type);
			if (pdevChoiceSet == NULL)
			{
			    tisSysLog (0,0,"getAttributes: dev choice for %s does not exist\n",
				     pvname);
			    return (-1);
			}
			if (pdevChoiceSet -> number < maxString)
			    no_strs = pdevChoiceSet -> number;
			else
			    no_strs = maxString;
			for (i = 0; i < no_strs; i++)
			{
			    pdevChoice = GET_DEV_CHOICE(pdevChoiceSet,i);
			    if ( (pdevChoice == NULL) || (pdevChoice -> pchoice == NULL) )
				stringList[i][0] = 0;
			    else
			      {
				strncpy (&stringList[i][0], pdevChoice -> pchoice, longestString);
				stringList[i][longestString-1] = 0;
			      }
			}
			*numStrings = (short)no_strs;
			break;

		    default:
			tisSysLog (0,0,"getAttributes: for %s unknown enum choice %d\n",
				        pvname, dbAddr.field_type);
			return (-1);
		}
	    }
	    break;

	case DBR_CHAR:
	case DBR_UCHAR:
	case DBR_SHORT:
	case DBR_USHORT:
	case DBR_LONG:
	case DBR_ULONG:
	case DBR_FLOAT:
	case DBR_DOUBLE:
	   FA_FetchData(descTable,pvname,(short)pointType,
		        crc,timestamp,lopr,hopr,drvl,drvh,
			&dblPrec,&str,&nstr);
	   *prec = (short)dblPrec;
	   break;

	default:
	    tisSysLog (0,0,"getAttributes: unknown field type for %s, %d\n",
		     pvname, dbAddr.dbr_field_type);
	    return (-1);
    }
    *nativeType = dbAddr.dbr_field_type;
    *nativeCount = (short)dbAddr.no_elements;
    return (0);
}


/*************************************************
*    
*    FA_initAttributes ()
*	init the SDR and the database
*
**************************************************/

int	FA_InitAttributes (void)
{
    char	filepath[MAXPATHLEN];
    int		status;
#ifdef WIN32
    struct _stat statbuff;
#else
    struct stat	statbuff;
#endif WIN32

    sprintf (filepath, "%sdefault.dctsdr", FA_DATABASEPATH);
    /* if the SDR is already loaded, dont do it again */
    if (dbRecType == NULL)
    {
	if (sdrLoad (filepath))
	    return (1);
    }
    if ((descTable = bldDesDataTable ()) == NULL)
	return (1);
    if (FA_InitDatabase ())
	return (1);
    udbpass = PASS_1;
    udbpvs = (UDBPVS_PTR)0;
    sprintf (filepath, "%s%s", FA_DATABASEPATH, DBTREENAME);
    /* if the file exists an dis not 0 length, just read it in */
#ifdef WIN32
    status = _stat (filepath, &statbuff);
#else
    status = stat (filepath, &statbuff);
#endif WIN32
    if ( (status == 0) && (statbuff.st_size != 0) )
    {
	if (FA_ReadTree (filepath))
	    return (1);
    }
    else
    {
	/* file does not exist or is zero length, create tree and write it */
	if (FA_InitAvlTree (descTable))
	    return (1);
	if (FA_WriteTree (filepath))
	    return (1);
    }
    return (0);
}


/**************************************************************
*
*    FA_dbNameToAddr ()
*	given a pointname and a pointer to a struct dbAddr
*	fill in the info that is needed
*
**************************************************************/

int	FA_dbNameToAddr (char *ptname, int pointType, struct dbAddr *pdbAddr)
{
    char		pvname[PVNAME_SZ + FLDNAME_SZ + 1];
    char		fldName[FLDNAME_SZ];
    char		*pFldname;
    short		top, bottom, test;
    short		*sortFldInd;
    int			compare;
    unsigned long	*sortFldName;
    struct fldDes	*pfldDes;
    struct recTypDes	*precTypDes;

    strncpy (pvname, ptname, PVNAME_SZ + FLDNAME_SZ);
    pvname[PVNAME_SZ + FLDNAME_SZ] = 0;
    if ((pFldname = strrchr (pvname, '.')) != NULL)
    {
	pFldname++;
	sprintf (fldName, "%-*.*s", FLDNAME_SZ, FLDNAME_SZ, pFldname);
    }
    /* default field is VAL, if one is not present */
    else
    {
	memcpy (fldName, "VAL ", FLDNAME_SZ);
    }

    /* record decsription must be there */
    if (dbRecDes == NULL)
    {
	tisSysLog (0,0,"FA_dbNameToAddr: dbRecDes not loaded\n");
	return (1);
    }

    /* point type must valid */
    if ( (pointType < 0) || (pointType >= dbRecDes -> number))
    {
	tisSysLog (0,0,"FA_dbNameToAddr: pointType value %d for %s\n", pointType, ptname);
	return (1);
    }

    precTypDes = dbRecDes -> papRecTypDes[pointType];
    sortFldName = precTypDes -> sortFldName;
    sortFldInd = precTypDes -> sortFldInd;

    /* binary search through ordered field names */
    top = precTypDes -> no_fields - 1;
    bottom = 0;
    test = (top + bottom) / 2;
    while (1)
    {
	compare = memcmp ((void *)&sortFldName[test], (void *)&fldName[0], FLDNAME_SZ);
	if (compare == 0)
	{
	    pfldDes = precTypDes -> papFldDes[sortFldInd[test]];
	    break;
	}
	else if (compare > 0)
	{
	    top = test - 1;
	    if (top < bottom)
	    {
		tisSysLog (0,0,"FA_dbNameToAddr: field %4.4s not found for %s, ptype %d\n", 
			       fldName, ptname, pointType);
		return (1);
	    }
	    test = (top + bottom) / 2;
	}
	else
	{
	    bottom = test + 1;
	    if (top < bottom)
	    {
		tisSysLog (0,0,"FA_dbNameToAddr: field %4.4s not found for %s, ptype %d\n", 
			       fldName, ptname, pointType);
		return (1);
	    }
	    test = (top + bottom) / 2;
	}
    }
    pdbAddr -> precord        = (caddr_t)0;
    pdbAddr -> pfield         = (caddr_t)0;
    pdbAddr -> record_type    = pointType;
    pdbAddr -> pfldDes        = (caddr_t)pfldDes;
    pdbAddr -> field_type     = pfldDes -> field_type;
    pdbAddr -> dbr_field_type = pfldDes -> dbr_field_type;
    pdbAddr -> field_size     = pfldDes -> size;
    pdbAddr -> choice_set     = pfldDes -> choice_set;
    pdbAddr -> special        = pfldDes -> special;
    pdbAddr -> no_elements    = 1;
    if (pfldDes -> special == SPC_DBADDR)
    {
	FA_convert_dbAddr (pdbAddr, ptname);
    }

    return (0);
}


/**************************************************************
*
*    FA_removeTrailingSpaces
*	place a NULL in all trailing spaces
*
**************************************************************/

void	FA_removeTrailingSpaces (char *str)
{
    char	*pChar;
    int		i;

    pChar = str;
    for (i = (strlen (str) - 1); i >= 0; i--)
    {
	if (isspace (pChar[i]))
	    pChar[i] = 0;
	else
	    break;
    }
    return;
}
