/*devRes.c,v 1.1.1.1 1999/08/03 14:42:44 tis4000 Exp */

/****************************************************************/
/*                                                              */
/*   This software is the proprietary property of Tate          */
/*   Integrated Systems, L.P.  Patents and copyrights are       */
/*   pending.                                                   */
/*                                                              */
/*   This software is furnished under a license and may         */
/*   be used and copied only in accordance with the terms       */
/*   of such license and with the inclusion of the above        */
/*   notice. This software or any other copies thereof          */
/*   may not be provided or otherwise made available to         */
/*   any other person. No title to and ownership of the         */
/*   software is hereby transferred.                            */
/*                                                              */
/*   The information in this software is subject to             */
/*   change without notice and should not be construed          */
/*   as a commitment by Tate Integrated Systems L.P. or         */
/*   its third party suppliers.                                 */
/*                                                              */
/*   Tate Integrated Systems, L.P. (TIS) and its third          */
/*   party suppliers assume no responsibility for the           */
/*   use or inability to use any of its software. TIS           */
/*   software is provided "as is" without warranty of           */
/*   any kind and TIS expressly disclaims all implied           */
/*   warranties, including but not limited to the implied       */
/*   warranties of merchantability and fitness for a            */
/*   particular purpose.                                        */
/*                                                              */
/*   Notice: Notwithstanding any other lease or license         */
/*   that may pertain to or accompany the delivery of this      */
/*   computer software, the rights of the Government            */
/*   regarding its use, reproduction, and disclosure are        */
/*   as set forth in Section 52.227-19 of the FARS              */
/*   Computer Software-Restricted Rights clause.                */
/*                                                              */
/****************************************************************/


/****************************************************************************
 *
 *  MODULE NAME            : devRes.c
 *  AUTHOR                 : emsteel
 *  DATE                   : 22 Dec 93
 *
 *
 * DESCRIPTION: device support for Res
 *
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 29 Mar 94
 *  REVISION:  make forwardfunc defs agree with actual func def
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 19 Apr 94
 *  REVISION: for mbbo fix msg fro mbbi to mbbo
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 29 June 94
 *  REVISION:	added stringin, tsDefs.h, time_stamp
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 21 June 95
 *  REVISION:	added ai, bi, bo, find_symbols(), ahdlclinestats(), 
 *		ahdlcunitstats()
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 11 Jul 96
 *  REVISION:	added SysAlmStr[] to SI if signal is 3
 *
****************************************************************************/

/****************************************************************************
 *
 * Programmer's Notes:
 *    This is a compendium of all dev**Res.c files in .../share/src/dev
 *    Where noted, change appropriately to create a new device support
 *    module
 *
 *
****************************************************************************/


#include	<vxWorks.h>
#include	<inetLib.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<symLib.h>
#include	<sysSymTbl.h>     /* for sysSymTbl */
#include 	<string.h>

#include	"alarm.h"
#include	"cvtTable.h"
#include	"dbDefs.h"
#include	"dbCommon.h"
#include	"db_field_log.h"
#include	"link.h"
#include	"dbAccess.h"
#include	"devSup.h"
#include	"special.h"
#include	"rec/aiRecord.h"
#include	"rec/biRecord.h"
#include	"rec/boRecord.h"
#include	"rec/mbbiRecord.h"
#include	"rec/mbboRecord.h"
#include	"rec/stringinRecord.h"
#include	"restable.h"
#include	"tsDefs.h"
#include	"ahdlcstats.h"
#include	"error.h"
#include	"serialTask.h"

/****************************************************************************************
 *   Resource I/O Name definitions							
 *	MODIFICATION BY GORDON: Have the upper and lower case versions of the
 *	names in the list. The list MUST have the names in the same order in
 *	two halves of the list. See the routine match_resource() in this
 *	source file for further details.
 */

static char  *resources[] = {"TIME", "DATE", "SYSMSG", "RESGROUP", "AHDLC","SERIAL",
			     "time", "date", "sysmsg", "resgroup", "ahdlc","serial" };
static short resnamelen[] = {  4,       4,       6,         8,        5,	6,
			       4,	4,	 6,	    8,	      5,	6   };


#define rTIME		1
#define rDATE		2
#define rSYSMSG		3
#define rRESGROUP	4
#define rAHDLC		5
#define rSERIAL		6
#define NUMRESOURCES	12
/****************************************************************************************/

typedef unsigned char UTINY;

extern TS_STAMP	time_stamp;			/* defined in .../src/time/ts_driv.c */
extern char	  	SysAlmStr[80];		/* defined in .../src/db/iocInit.c */
extern short	resGroupLock[16];		/* defined in .../src/redund/resgrouptask.c */
extern char	  	*resGroupTripName[16];	/* defined in .../src/redund/resgrouptask.c */

short	  		devSerialPort = -1;
unsigned short	DEVDEBUG = FALSE;
static ahDevStats *ahdlcdevstats = NULL;	/* defined in epicsH/ahdlstats.h */
static		  void (*ahdlcfailover)() = NULL;
static		  void (*ahdlcinitstats)() = NULL;
static char 	  *io_name = "Res";	/* NOTE: change here */
static int 	  find_symbols ();
int		  ahdlclinestats ();
int		  ahdlcunitstats ();

/* Create the dset for devAiRes */
static int match_resource (char	*name);

static long ai_init_record ();
static long read_ai ();
static void ahlineheader ();
static void ahlinestats (int i);
int ahdlcunitstats (int line);
static void ahunitheader ();
static void ahunitstats (int line, int i);

struct {
    long        number;
    DEVSUPFUN   report;
    DEVSUPFUN   init;
    DEVSUPFUN   init_record;
    DEVSUPFUN   get_ioint_info;
    DEVSUPFUN   read_ai;
    DEVSUPFUN   special_linconv;
} devAiRes = {					/* NOTE: change here */
    6,
    NULL,
    NULL,
    ai_init_record,
    NULL,
    read_ai,
    NULL
};

/* Create the dset for devBiRes */

static long bi_init_record ();
static long read_bi ();

struct {
    long        number;
    DEVSUPFUN   report;
    DEVSUPFUN   init;
    DEVSUPFUN   init_record;
    DEVSUPFUN   get_ioint_info;
    DEVSUPFUN   read_bi;
} devBiRes = {					/* NOTE: change here */
    5,
    NULL,
    NULL,
    bi_init_record,
    NULL,
    read_bi
};

/* Create the dset for devBoRes */

static long bo_init_record ();
static long write_bo ();

struct {
    long        number;
    DEVSUPFUN   report;
    DEVSUPFUN   init;
    DEVSUPFUN   init_record;
    DEVSUPFUN   get_ioint_info;
    DEVSUPFUN   write_bo;
} devBoRes = {					/* NOTE: change here */
    5,
    NULL,
    NULL,
    bo_init_record,
    NULL,
    write_bo
};

/* Create the dset for devMbbiRes */

static long mbbi_init_record ();
static long read_mbbi ();

struct {
    long        number;
    DEVSUPFUN   report;
    DEVSUPFUN   init;
    DEVSUPFUN   init_record;
    DEVSUPFUN   get_ioint_info;
    DEVSUPFUN   read_mbbi;
} devMbbiRes = {                          /* NOTE: change here */
    5,
    NULL,
    NULL,
    mbbi_init_record,
    NULL,
    read_mbbi
};

/* Create the dset for devMbboRes */

static long mbbo_init_record ();
static long write_mbbo ();

struct {
    long        number;
    DEVSUPFUN   report;
    DEVSUPFUN   init;
    DEVSUPFUN   init_record;
    DEVSUPFUN   get_ioint_info;
    DEVSUPFUN   write_mbbo;
} devMbboRes = {                          /* NOTE: change here */
    5,
    NULL,
    NULL,
    mbbo_init_record,
    NULL,
    write_mbbo
};

/* Create the dset for devSiRes */

static long si_init_record ();
static long read_stringin ();

struct {
    long        number;
    DEVSUPFUN   report;
    DEVSUPFUN   init;
    DEVSUPFUN   init_record;
    DEVSUPFUN   get_ioint_info;
    DEVSUPFUN   read_stringin;
} devSiRes = {                          /* NOTE: change here */
    5,
    NULL,
    NULL,
    si_init_record,
    NULL,
    read_stringin
};



/********************************** AI *********************************/

static long ai_init_record (pai)
   struct aiRecord   *pai;
{

    struct tateio	*ptateio;
    char		message[128];
    char		*colon;
    long		status = 0;

    /* ai.inp must be RES_IO */

    switch (pai->inp.type)
    {
	case (RES_IO) :
	    strcpy (message, pai->name);
	    strcat (message, ": devAi");
	    strcat (message, io_name);
	    strcat (message, " (init_record) ");

	    if ((colon = (char *) index(pai -> name, ':')) == NULL)
	    {
		strcat (message, "No colon in name");
		if (DEVDEBUG) errMessage (S_db_badField, message);
		return (S_db_badField);
	    }
	    colon++;
	    if (strlen (colon) == 0)
	    {
		strcat (message, "No resource name after colon");
		if (DEVDEBUG) errMessage (S_db_badField, message);
		return (S_db_badField);
	    }
	    if (strncmp (colon, "AHDLC", 5) == 0)
	    {

	        if (find_symbols (message))
	        {
		    if (DEVDEBUG) errMessage (S_db_badField, message);
		    return (S_db_badField);
	        }

	        ptateio = (struct tateio *)&(pai -> inp.value);
	        /* line value */
	        if ((ptateio -> link < 0) || (ptateio -> link > 63))
	        {
		    strcat (message, "line not 0-63");
		    if (DEVDEBUG) errMessage (S_db_badField, message);
		    return (S_db_badField);
	        }
	        if ((ptateio -> card < 0) || (ptateio -> card > 1))
	        {
		    strcat (message, "card not 0-1");
		    if (DEVDEBUG) errMessage (S_db_badField, message);
		    return (S_db_badField);
	        }
	        if (ptateio -> card == 1)
	        {
		    /* address */
		    if ((ptateio -> adapter < 1) || (ptateio -> adapter > 32))
		    {
		        strcat (message, "adapter not 1-32");
		        if (DEVDEBUG) errMessage (S_db_badField, message);
		        return (S_db_badField);
		    }
	        }

	        if ((ptateio -> signal < 1) || (ptateio -> signal > 15))
 	        {
 		    strcat (message, "signal not 1-15");
 		    if (DEVDEBUG) errMessage (S_db_badField, message);
 		    return (S_db_badField);
 	        }

	        if ((status = read_ai(pai)) == -1)
	        {
		    strcat (message, "data not valid");
		    if (DEVDEBUG) errMessage (-1L, message);
		    return (-1L);
	        }

	    } else {


		/* Add New Resources before here !!!!! */

		strcat (message, "Resource Name not Known");
		if (DEVDEBUG) errMessage (S_db_badField, message);
		return (S_db_badField);
	    }

	    break;

	default :
	    strcpy (message, pai->name);
	    strcat (message, ": devAi");
	    strcat (message, io_name);
	    strcat (message, " (init_record) Illegal INP field");
	    if (DEVDEBUG) errMessage (S_db_badField, message);
	    return (S_db_badField);
    }
    return (status);
}


static long read_ai (pai)
    struct aiRecord   *pai;
{
    char		message[128];
    int			status = 0;
    ahLineStats		*ahLineStats;
    ahUnitStats		*ahUnitStats;
    struct tateio	*ptateio;
    char		*colon;

    /* ai.inp must be RES_IO */

    ptateio = (struct tateio *)&(pai -> inp.value);
    switch (pai->inp.type)
    {
	case (RES_IO) :


	    /* Make a pointer to the Resource Type*/

	    colon = (char *) index(pai -> name, ':');
	    colon++;


	    if (strncmp (colon, "AHDLC", 5) == 0)
	    {

	        if (ahdlcdevstats -> data_valid[ptateio -> link] == 0)
	        {
		    status = -1;
		    break;
	        }
	        switch (ptateio -> card)
	        {
		    case 0:
		        ahLineStats = &(ahdlcdevstats -> linestats[ptateio -> link]);
		        switch (ptateio -> signal)
		        {
			    case 1:
			        pai -> rval = ahLineStats -> ipackets;
			        break;

			    case 2:
			        pai -> rval = ahLineStats -> opackets;
			        break;

			    case 3:
			        pai -> rval = ahLineStats -> ierrors;
			        break;

			    case 4:
			        pai -> rval = ahLineStats -> oerrors;
			        break;

			    case 5:
			        pai -> rval = ahLineStats -> txbufs;
			        break;

			    case 6:
			        pai -> rval = ahLineStats -> rxbufs;
			        break;

			    case 7:
			        pai -> rval = ahLineStats -> txbufs_avail;
			        break;

			    case 8:
			        pai -> rval = ahLineStats -> rxbufs_avail;
			        break;

			    case 9:
			        pai -> rval = ahLineStats -> txbuf_size;
			        break;

			    case 10:
			        pai -> rval = ahLineStats -> rxbuf_size;
			        break;

			    case 11:
			        pai -> rval = ahLineStats -> txframes;
			        break;

			    case 12:
			        pai -> rval = ahLineStats -> txtimeouts;
			        break;

			    case 13:
			        pai -> rval = ahLineStats -> rxframes;
			        break;

			    case 14:
			        pai -> rval = ahLineStats -> rxtimeouts;
			        break;

			    case 15:
			        pai -> rval = ahLineStats -> rxhwerrors;
			        break;

			    default:
			        status = -1;
			        break;

		        }
		        break;

		    case 1:
		        ahUnitStats = &(ahdlcdevstats -> unitstats[ptateio -> link][ptateio -> adapter - 1]);
		        switch (ptateio -> signal)
		        {
			    case 1:
			        pai -> rval = ahUnitStats -> inet_addr;
			        break;

			    case 2:
			        pai -> rval = ahUnitStats -> online_count;
			        break;

			    case 3:
			        pai -> rval = ahUnitStats -> crc_errors;
			        break;

			    case 4:
			        pai -> rval = ahUnitStats -> frame_errors;
			        break;

			    case 5:
			        pai -> rval = ahUnitStats -> format_errors;
			        break;

			    case 6:
			        pai -> rval = ahUnitStats -> rxhw_errors;
			        break;

			    case 7:
			        pai -> rval = ahUnitStats -> ipackets;
			        break;

			    case 8:
			        pai -> rval = ahUnitStats -> opackets;
			        break;

			    case 9:
			        pai -> rval = ahUnitStats -> ierrors;
			        break;

			    case 10:
			        pai -> rval = ahUnitStats -> oerrors;
			        break;

			    case 11:
			        pai -> rval = ahUnitStats -> txframes;
			        break;

			    case 12:
			        pai -> rval = ahUnitStats -> txtimeouts;
			        break;

			    case 13:
			        pai -> rval = ahUnitStats -> rxframes;
			        break;

			    case 14:
			        pai -> rval = ahUnitStats -> rxtimeouts;
			        break;

			    case 15:
			        pai -> rval = ahUnitStats -> rxhwerrors;
			        break;

			    default:
			        status = -1;
 
			        break;
		        }
		        break;

		    default:
		        status = -1;
		        break;
	        }

	        break;

	    } else {   /* Not AHDLC */

		/* Put more Resource Types in here !!!!! */

		status = -1;
                break;
	    }

	default :
	    if (pai->nsev < VALID_ALARM)
	    {
		pai->nsev = VALID_ALARM;
		pai->nsta = SOFT_ALARM;
		if (pai->stat != SOFT_ALARM)
		{
		    strcpy (message, pai->name);
		    strcat (message, ": devAi");
		    strcat (message, io_name);
		    strcat (message, " (read_ai) Illegal INP field");
		    if (DEVDEBUG) errMessage (S_db_badField, message);
		    status = S_db_badField;
		}
	    }
	    break;
    }
    return (status);
}


/********************************** BI *********************************/

static long bi_init_record (pbi)
   struct biRecord   *pbi;
{

    struct tateio	*ptateio;
    char		message[128];
    char		*colon;
    long		status = 0;
    short		restype;

    /* bi.inp must be RES_IO */

    ptateio = (struct tateio *)&(pbi -> inp.value);

    switch (pbi->inp.type)
    {
	case (RES_IO) :
	    strcpy (message, pbi->name);
	    strcat (message, ": devBi");
	    strcat (message, io_name);
	    strcat (message, " (init_record) ");

	    if ((colon = (char *) index(pbi -> name, ':')) == NULL)
	    {
		strcat (message, "No colon in name");
		if (DEVDEBUG) errMessage (S_db_badField, message);
		return (S_db_badField);
	    }
	    colon++;
	    if (strlen (colon) == 0)
	    {
		strcat (message, "No resource name after colon");
		if (DEVDEBUG) errMessage (S_db_badField, message);
		return (S_db_badField);
	    }

	    restype = match_resource (colon);

	    if (restype == rRESGROUP)
	    {
	        if ((ptateio -> signal < 1) || (ptateio -> signal > 16))
	            return (-1);

	        if ((status = read_bi(pbi)) == -1)
	        {
		    strcat (message, "data not valid");
		    if (DEVDEBUG) errMessage (-1L, message);
		    return (-1L);
		}
		return (0);
	    }

	    if (restype == rAHDLC)
	    {
	        if (find_symbols (message))
	        {
		    if (DEVDEBUG) errMessage (S_db_badField, message);
		    return (S_db_badField);
	        }

	        ptateio = (struct tateio *)&(pbi -> inp.value);
	        /* line value */
	        if ((ptateio -> link < 0) || (ptateio -> link > 63))
	        {
		    strcat (message, "line not 0-63");
		    if (DEVDEBUG) errMessage (S_db_badField, message);
		    return (S_db_badField);
	        }
	        if ((ptateio -> card < 0) || (ptateio -> card > 1))
	        {
		    strcat (message, "card not 0-1");
		    if (DEVDEBUG) errMessage (S_db_badField, message);
		    return (S_db_badField);
	        }
	        if (ptateio -> card == 1)
	        {
		    /* address */
		    if ((ptateio -> adapter < 1) || (ptateio -> adapter > 32))
		    {
		        strcat (message, "adapter not 1-32");
		        if (DEVDEBUG) errMessage (S_db_badField, message);
		        return (S_db_badField);
		    }
		    if (ptateio -> signal != 1)
		    {
 		        strcat (message, "signal not 1");
 		        if (DEVDEBUG) errMessage (S_db_badField, message);
 		        return (S_db_badField);
	            }
		} else {
 		    if ((ptateio -> signal < 1) || (ptateio -> signal > 4))
 		    {
 		        strcat (message, "signal not 1-4");
 		        if (DEVDEBUG) errMessage (S_db_badField, message);
 		        return (S_db_badField);
 		    }
		}

  	        if ((status = read_bi(pbi)) == -1)
	        {
		    strcat (message, "data not valid");
		    if (DEVDEBUG) errMessage (-1L, message);
		    return (-1L);
	        }
	    } else {

		/* Add New Resources before here !!!!! */

		strcat (message, "Resource Name not Known");
		if (DEVDEBUG) errMessage (S_db_badField, message);
		return (S_db_badField);
	    }

	    break;

	default :
	    strcpy (message, pbi->name);
	    strcat (message, ": devBi");
	    strcat (message, io_name);
	    strcat (message, " (init_record) Illegal INP field");
	    if (DEVDEBUG) errMessage (S_db_badField, message);
	    return (S_db_badField);
    }
    return (0);
}


static long read_bi (pbi)
    struct biRecord   *pbi;
{
    char		message[128];
    int			status = 0;
    ahLineStats		*ahLineStats;
    ahUnitStats		*ahUnitStats;
    struct tateio	*ptateio;
    char		*colon;
    short		restype;

    /* bi.inp must be RES_IO */

    ptateio = (struct tateio *)&(pbi -> inp.value);
    switch (pbi->inp.type)
    {
	case (RES_IO) :

	    colon = (char *) index(pbi -> name, ':');
	    colon++;

	    restype = match_resource (colon);

	    if (restype == rRESGROUP)
	    {
	        if ((ptateio -> signal < 1) || (ptateio -> signal > 16))
	            return (-1);
	        pbi -> rval = resGroupLock [ptateio -> signal-1];
		return (0);
	    }

	    if (restype == rAHDLC)
	    {

	        if (ahdlcdevstats -> data_valid[ptateio -> link] == 0)
	        {
		    status = -1;
		    break;
	        }
	        switch (ptateio -> card)
	        {
		    case 0:
		        ahLineStats = &(ahdlcdevstats -> linestats[ptateio -> link]);
		        switch (ptateio -> signal)
		        {
			    case 1:
			        pbi -> rval = ahLineStats -> line_active;
			        break;

			    case 2:
			        pbi -> rval = ahLineStats -> master_slave;
			        break;

			    case 3:
			        pbi -> rval = ahLineStats -> partner_alive;
			        break;

			    case 4:
			        pbi -> rval = ahLineStats -> compression;
			        break;

			    case 5:
			        pbi -> rval = ahLineStats -> line_responding;
			        break;

			    default:
			        status = -1;
			        break;

		        }
		        break;

		    case 1:
		        ahUnitStats = &(ahdlcdevstats -> unitstats[ptateio -> link][ptateio -> adapter - 1]);
		        switch (ptateio -> signal)
		        {
			    case 1:
			        pbi -> rval = ahUnitStats -> online;
			        break;

			    default:
			        status = -1;
			        break;
		        }
		        break;

		    default:
		        status = -1;
		        break;
	        }

	 	break;

	    } else { /* Not AHDLC */

		/* Put more Resource Types in here !!!!! */

		status = -1;
                break;
            }

	default :
	    if (pbi->nsev < VALID_ALARM)
	    {
		pbi->nsev = VALID_ALARM;
		pbi->nsta = SOFT_ALARM;
		if (pbi->stat != SOFT_ALARM)
		{
		    strcpy (message, pbi->name);
		    strcat (message, ": devBi");
		    strcat (message, io_name);
		    strcat (message, " (read_bi) Illegal INP field");
		    if (DEVDEBUG) errMessage (S_db_badField, message);
		    status = S_db_badField;
		}
	    }
	    break;
    }
    return (status);
}


/********************************** BO *********************************/

static long bo_init_record (pbo)
   struct boRecord   *pbo;
{

    struct tateio	*ptateio;
    char		message[128];
    char		*colon;
    short		restype;

    /* bo.inp must be RES_IO */

    ptateio = (struct tateio *)&(pbo -> out.value);

    switch (pbo -> out.type)
    {
	case (RES_IO) :
	    strcpy (message, pbo->name);
	    strcat (message, ": devBo");
	    strcat (message, io_name);
	    strcat (message, " (init_record) ");

	    if ((colon = (char *) index(pbo -> name, ':')) == NULL)
	    {
		strcat (message, "No colon in name");
		if (DEVDEBUG) errMessage (S_db_badField, message);
		return (S_db_badField);
	    }
	    colon++;
	    if (strlen (colon) == 0)
	    {
		strcat (message, "No resource name after colon");
		if (DEVDEBUG) errMessage (S_db_badField, message);
		return (S_db_badField);
	    }
	    restype = match_resource (colon);

	    if (restype == rRESGROUP)
	    {
	        if ((ptateio -> signal < 1) || (ptateio -> signal > 16))
	            return (-1);
	        pbo -> mask = 1;
	        pbo -> rval = 0;
		  return (0);
	    }

	    if (restype == rAHDLC)
	    {

	        if (find_symbols (message))
	        {
		    if (DEVDEBUG) errMessage (S_db_badField, message);
		    return (S_db_badField);
	        }

	        ptateio = (struct tateio *)&(pbo -> out.value);
	        /* line value */
	        if ((ptateio -> link < 0) || (ptateio -> link > 63))
	        {
		    strcat (message, "line not 0-63");
		    if (DEVDEBUG) errMessage (S_db_badField, message);
		    return (S_db_badField);
	        }
	        if ((ptateio -> card < 0) || (ptateio -> card > 1))
	        {
		    strcat (message, "card not 0-1");
		    if (DEVDEBUG) errMessage (S_db_badField, message);
		    return (S_db_badField);
	        }
	        if ((ptateio -> signal < 1) || (ptateio -> signal > 2))
	        {
		    strcat (message, "signal not 1");
		    if (DEVDEBUG) errMessage (S_db_badField, message);
		    return (S_db_badField);
	        }
	        if ((ptateio -> card == 1) && (ptateio -> signal == 2))
	        {
		    if ((ptateio -> adapter < 1) || (ptateio -> adapter > 32))
		    {
		        strcat (message, "adapter not 1-32");
		        if (DEVDEBUG) errMessage (S_db_badField, message);
		        return (S_db_badField);
		    }
	        }

	        pbo -> mask = 1;
	        pbo -> rval = 0;

	        break;
	    }
	    if (restype == rSERIAL)
	    {
		if ((ptateio -> signal < 0) || (ptateio -> signal > 15))
		{
			strcat (message, "signal not 0-15");
			if (DEVDEBUG) errMessage (S_db_badField, message);
			return (S_db_badField);
		}
		if ((devSerialPort != ptateio -> signal) && (devSerialPort != -1))
		{
			strcat (message, "redefinition of access to serial port status by");
			if (DEVDEBUG) errMessage (S_db_badField, message);
			return (S_db_badField);
		}
		attachSerialRedund(ptateio -> signal ,60);	
		devSerialPort = ptateio -> signal,
		pbo -> mask = 1;
		pbo -> rval = 0;
		return (0);
		break;
	    } else {

		/* Add New Resources before here !!!!! */

		strcat (message, "Resource Name not Known");
		if (DEVDEBUG) errMessage (S_db_badField, message);
		return (S_db_badField);
	    }


	default :
	    strcpy (message, pbo->name);
	    strcat (message, ": devBo");
	    strcat (message, io_name);
	    strcat (message, " (init_record) Illegal INP field");
	    if (DEVDEBUG) errMessage (S_db_badField, message);
	    return (S_db_badField);
    }
    return (0);
}


static long write_bo (pbo)
    struct boRecord   *pbo;
{
    char		message[128];
    int			status = 0;
    struct tateio	*ptateio;
    char		*colon;
    short		restype;
    unsigned char	mchar;

    /* bo.inp must be RES_IO */

    ptateio = (struct tateio *)&(pbo -> out.value);
    switch (pbo -> out.type)
    {
	case (RES_IO) :


	    colon = (char *) index(pbo -> name, ':');
	    colon++;

	    restype = match_resource (colon);

	    if (restype == rRESGROUP)
	    {
	        if ((ptateio -> signal < 1) || (ptateio -> signal > 16))
	            return (-1);
	        resGroupLock [ptateio -> signal-1] = pbo -> rval;
		return (0);
	    }
	    if (restype == rAHDLC)
	    {


	        if (ahdlcdevstats -> data_valid[ptateio -> link] == 0)
	        {
		    status = -1;
		    break;
	        }
	        if (pbo -> rval == 1)
	        {
		    switch (ptateio -> card)
		    {
		         case 0:
			    if (ptateio -> signal == 1)
			    {
			        /* switch line from MASTER to SLAVE */
			        (*ahdlcfailover) (ptateio -> link);
			    }
			    else if (ptateio -> signal == 2)
			    {
			        /* init ahdlcdevstats->linestats[link] */
			        ahdlcinitstats (ptateio -> link, (short)0);
			    }
			    break;

		        case 1:
			    if (ptateio -> signal == 2)
			    {
			        /* init ahdlcdevstats -> unitstats[link][adapater] */
			        ahdlcinitstats (ptateio -> link, ptateio -> adapter);
			    }
			    break;

		        default:
			    status = -1;
			    break;
		    }
	        }
	        break;
	    }
	    if (restype == rSERIAL)
	    {
		if (pbo -> rval == 0)
			mchar = 'S';
		else
			mchar = 'M';

		pulseSerialRedund(devSerialPort, mchar);
		break;
	    } else { /* Not Serial */

		/* Put more Resource Types in here !!!!! */

		status = -1;
                break;
            }

	default :
	    if (pbo->nsev < VALID_ALARM)
	    {
		pbo->nsev = VALID_ALARM;
		pbo->nsta = SOFT_ALARM;
		if (pbo->stat != SOFT_ALARM)
		{
		    strcpy (message, pbo->name);
		    strcat (message, ": devBo");
		    strcat (message, io_name);
		    strcat (message, " (write_bo) Illegal INP field");
		    if (DEVDEBUG) errMessage (S_db_badField, message);
		    status = S_db_badField;
		}
	    }
	    break;
    }
    return (status);
}


/********************************** MBBI *********************************/

static long mbbi_init_record (pmbbi)
    struct mbbiRecord   *pmbbi;
{
    char  message[128];
    char  *colon;
    short	restype;

    /* mbbi.inp must be RES_IO */

    pmbbi -> init = 1;
    switch (pmbbi->inp.type)
    {
	case (RES_IO) :
	    strcpy (message, pmbbi->name);
	    strcat (message, ": devMbbi");
	    strcat (message, io_name);
	    strcat (message, " (init_record) ");
	    if ((colon = (char *) index(pmbbi -> name, ':')) == NULL)
	    {
	    }
	    colon++;

	    restype = match_resource (colon);

	    if (restype == rSERIAL)
	    {
		if (devSerialPort == -1)
		{
			strcat (message, "Serial port number not set by BO yet");
			if (DEVDEBUG) errMessage (S_db_badField, message);
			return (S_db_badField);
		} else {

	        read_mbbi(pmbbi);
	        break;
		}
	    }


	    if (strlen (colon) == 0)
	    {
		strcat (message, "No resource name after colon");
		if (DEVDEBUG) errMessage (S_db_badField, message);
		return (S_db_badField);
	    }
	    if (resGetByName (colon, (RES_TABLE **)&(pmbbi -> dpvt)) == RES_OK)
	    {
	        read_mbbi(pmbbi);
	        break;
	    } else {

		/* Add New Resources before here !!!!! */

		strcat (message, "Resource Name not Known");
		if (DEVDEBUG) errMessage (S_db_badField, message);
		return (S_db_badField);
	    }

	default :
	    strcpy (message, pmbbi->name);
	    strcat (message, ": devMbbi");
	    strcat (message, io_name);
	    strcat (message, " (init_record) Illegal INP field");
	    if (DEVDEBUG) errMessage (S_db_badField, message);
	    return (S_db_badField);
    }
    return (0);
}


static long read_mbbi (pmbbi)
    struct mbbiRecord   *pmbbi;
{
    char        message[128];
    int         status;
    short       cstate;
    char  	*colon;
    unsigned char	mchar;
    unsigned short restype;
    /* mbbi.inp must be RES_IO */

    switch (pmbbi->inp.type)
    {
	case (RES_IO) :
	    colon = (char *) index(pmbbi -> name, ':');
	    colon++;

	    restype = match_resource (colon);

	    if (restype == rSERIAL)
	    {

		mchar = remoteSerialRedund(pmbbi->inp.value.tateio.signal);
	        pmbbi -> rval = 0;
		if (mchar == 'S')
	        pmbbi -> rval = 1;
		if (mchar == 'M')
	        pmbbi -> rval = 2;
			
		status = 0; 
	      break;

	    }

	    if (resGetByName (colon, (RES_TABLE **)&(pmbbi -> dpvt)) == RES_OK)
	    {

	        switch (pmbbi->inp.value.tateio.signal)
	        {
		    case 0:
		        resGetStates ((RES_TABLE *)pmbbi -> dpvt, &cstate, 0);
		        break;

		    case 1:
		        resGetStates ((RES_TABLE *)pmbbi -> dpvt, 0, &cstate);
		        break;

		    default:
		        return (-1);
	        }
	        cstate &= ~RES_CHANGED;   /* Remove changed bit */
	        pmbbi -> rval = cstate;
		  status = 0;
	        break;
	    } else { 

		/* Put more Resource Types in here !!!!! */

		status = -1;
                break;
            }


	default :
	    if (pmbbi->nsev < VALID_ALARM)
	    {
		pmbbi->nsev = VALID_ALARM;
		pmbbi->nsta = SOFT_ALARM;
		if (pmbbi->stat != SOFT_ALARM)
		{
		    strcpy (message, pmbbi->name);
		    strcat (message, ": devMbbi");
		    strcat (message, io_name);
		    strcat (message, " (read_mbbi) Illegal INP field");
		    if (DEVDEBUG) errMessage (S_db_badField,message);
		}
	    }
	    status = -1;
	    break;
    }
    return (status);
}


/*********************************** MBBO ********************************/

static long mbbo_init_record (pmbbo)
    struct mbboRecord *pmbbo;
{
    char  message[128];
    char  *colon;

    strcpy (message, pmbbo -> name);
    strcat (message, ": devMbbo");
    strcat (message, io_name);
    strcat (message, " (mbbo_init) ");
    pmbbo -> init = 1;

    switch (pmbbo -> out.type)
    {
	case (RES_IO):
	    if ((colon = (char *) index(pmbbo -> name, ':')) == NULL)
	    {
		strcat (message, "No colon in name");
		if (DEVDEBUG) errMessage (S_db_badField, message);
		return (S_db_badField);
	    }
	    colon++;
	    if (strlen(colon) == NULL)
	    {
		strcat (message, "No resource name after colon");
		if (DEVDEBUG) errMessage (S_db_badField, message);
		return (S_db_badField);
	    }
	    if (resGetByName (colon, (RES_TABLE **)&(pmbbo -> dpvt)) == RES_OK)
	    {
	        pmbbo -> rval = 0;
	        return (0);
	    } else {

		/* Add New Resources before here !!!!! */

		strcat (message, "Resource Name not Known");
		if (DEVDEBUG) errMessage (S_db_badField, message);
		return (S_db_badField);
	    }
	default:
	    strcat (message, "Illegal I/O type");
	    if (DEVDEBUG) errMessage (S_db_badField, message);
	    return (S_db_badField);
    }
}


static long write_mbbo (pmbbo)
    struct mbboRecord   *pmbbo;
{

    char  message[128];
    char  *colon;
    int         status;

    switch (pmbbo->out.type)
    {
	case (RES_IO) :
	    colon = (char *) index(pmbbo -> name, ':');
	    colon++;
	    if (resGetByName (colon, (RES_TABLE **)&(pmbbo -> dpvt)) == RES_OK)
	    {
		resPutCommand((RES_TABLE *)pmbbo -> dpvt, (unsigned short)(pmbbo -> rval));
		status = 0;
		break;
	    } else {

		/* Put more Resource Types in here !!!!! */

		status = -1;
                break;
            }


	default :
	    if (pmbbo->nsev < VALID_ALARM)
	    {
		pmbbo->nsev = VALID_ALARM;
		pmbbo->nsta = SOFT_ALARM;
		if (pmbbo->stat != SOFT_ALARM)
		{
		    strcpy (message, pmbbo->name);
		    strcat (message, ": devMbbo");
		    strcat (message, io_name);
		    strcat (message, " (write_mbbo) Illegal INP field");
		    if (DEVDEBUG) errMessage (S_db_badField,message);
		}
	    }
	    status = -1;
	    break;
    }
    return (status);
}



/********************************** SI *********************************/

static long si_init_record (pstringin)
    struct stringinRecord   *pstringin;
{
    struct tateio	*ptateio;
    char		message[128];
    char		*colon;
    long		status = 0;
    short		strtype;

    /* stringin.inp must be RES_IO */

    ptateio = (struct tateio *)&(pstringin -> inp.value);
    switch (pstringin->inp.type)
    {
	case (RES_IO) :
	    strcpy (message, pstringin->name);
	    strcat (message, ": devSi");
	    strcat (message, io_name);
	    strcat (message, " (init_record) ");

	    ptateio = (struct tateio *)&(pstringin -> inp.value);

	    if ((colon = (char *) index(pstringin -> name, ':')) == NULL)
	    {
		strcat (message, "No colon in name");
		if (DEVDEBUG) errMessage (S_db_badField, message);
		status = S_db_badField;
		break;
	    }
	    colon++;
	    if (strlen (colon) == 0)
	    {
		strcat (message, "No Resource name after colon");
		if (DEVDEBUG) errMessage (S_db_badField, message);
		status = S_db_badField;
		break;
	    }

	    strtype = match_resource (colon);

	    switch ( strtype )
	    {
		case rTIME:
		case rDATE:
		    if (time_stamp.secPastEpoch < 110000) 
		    {
		        /* make sure val is NULL if can't get DATE/TIME, 
		           recStringin:init_record may let block init if tagged, 
		           off scan or in manual
		        */
		        bzero (pstringin -> val, sizeof (pstringin -> val));
		        strcat (message, "TIME not set");
		        if (DEVDEBUG) errMessage (-1, message);
		        status = -1;
		        break;
		    }
		    status = read_stringin(pstringin);
		    break;

		case rSYSMSG:

		    bzero (pstringin -> val, sizeof (pstringin -> val));
		    status = read_stringin(pstringin);
		    break;

		case rRESGROUP:

	            if ((ptateio -> signal < 1) || (ptateio -> signal > 16))
	            {
		        strcat (message, "signal not 1");
		        if (DEVDEBUG) errMessage (S_db_badField, message);
		        return (S_db_badField);
	            }
		    bzero (pstringin -> val, sizeof (pstringin -> val));
		    status = read_stringin(pstringin);
		    break;


			/* Add New Resources before here !!!!! */

		default:

		    strcat (message, "Resource Name not Known");
		    if (DEVDEBUG) errMessage (S_db_badField, message);
		    return (S_db_badField);
	    }

	    break;

	default :
	    strcpy (message, pstringin->name);
	    strcat (message, ": devSi");
	    strcat (message, io_name);
	    strcat (message, " (init_record) Illegal INP field");
	    if (DEVDEBUG) errMessage (S_db_badField, message);
	    status = S_db_badField;
	    break;
    }
    return (status);
}


static long read_stringin (pstringin)
   struct stringinRecord   *pstringin;
{
    struct tateio	*ptateio;
    char	        message[128];
    int			status = 0;
    char		*colon;
    short		strtype;

    /* stringin.inp must be RES_IO */

    ptateio = (struct tateio *)&(pstringin -> inp.value);
    switch (pstringin->inp.type)
    {

	case (RES_IO) :
	    ptateio = (struct tateio *)&(pstringin -> inp.value);
	    bzero (message, sizeof(message));
	    colon = (char *) index(pstringin -> name, ':');
	    colon++;

	    strtype = match_resource (colon);

	    /* returns DATE and TIME in format
	       Mon dd, yyyy hh:mm:ss.nano-secs
	    */
	    tsStampToText (&time_stamp, TS_TEXT_MONDDYYYY, message);

	    switch ( strtype )
	    {
		case rTIME:

		    bzero (pstringin -> val, sizeof (pstringin -> val));
		    strncpy (pstringin -> val, &message[13], sizeof (pstringin -> val) - 1);
		    break;

		case rDATE:

		    bzero (pstringin -> val, sizeof (pstringin -> val));
		    strncpy (pstringin -> val, message, 12);
		    break;

		case rSYSMSG:

		    bzero (pstringin -> val, sizeof (pstringin -> val));
		    strncpy (pstringin -> val, SysAlmStr, sizeof (pstringin -> val) - 1);
		    break;

		case rRESGROUP:

	            if ((ptateio -> signal < 1) || (ptateio -> signal > 16))
			return (-1); 

		    bzero (pstringin -> val, sizeof (pstringin -> val));
		    strncpy (pstringin -> val, resGroupTripName[ptateio -> signal-1], sizeof (pstringin -> val) - 1);
		    break;


			/* Add New Resources before here !!!!! */

		default:

		    strcpy (message, pstringin->name);
		    strcat (message, ": devSi");
		    strcat (message, io_name);
		    strcat (message, " (read_stringin) ");
		    strcat (message, "Not a  Known name after colon");
		    if (DEVDEBUG) errMessage (S_db_badField, message);
		    status = S_db_badField;
		    break;
	    }

	    break;

	default :
	    if (pstringin->nsev < VALID_ALARM)
	    {
		pstringin -> nsev = VALID_ALARM;
		pstringin -> nsta = SOFT_ALARM;
		if (pstringin -> stat != SOFT_ALARM)
		{
		    strcpy (message, pstringin->name);
		    strcat (message, ": devSi");
		    strcat (message, io_name);
		    strcat (message, " (read_stringin) Illegal INP field");
		    if (DEVDEBUG) errMessage (S_db_badField,message);
		}
	    }
	    break;
    }
    return (status);
}


/**************************** Match the Resource *****************************/
static int match_resource (name)
    char	*name;
{
	/*
	MODIFICATION BY GORDON. Make the list of resources all upper case
	names followed by the same names in lower case. If we find a match
	in the second half of the list, fold the index that we return back
	into the first half. WARNING! YOU MUST BE VERY CAREFUL WHEN YOU ADD
	NAMES TO THIS LIST! THE LIST MUST, BY DEFINITION, HAVE AN EVEN NUMBER
	OF ENTRIES!
	*/

	short	i;
	for (i=0;(i<NUMRESOURCES);i++)
	{
	    if (strncmp (name, resources[i], resnamelen[i]) == 0)
		{
		i++;
		if ( i > (NUMRESOURCES >> 1))
			i -= (NUMRESOURCES >> 1);
	      return (i);
		}
	}
	return (0);
}


/**************************** find_symbols ***********************************/

static int find_symbols (message)
    char	*message;
{
    long	rtnval;
    UTINY	type;
    char	symName[32];      /* name of the symbol name */

    if (ahdlcdevstats == NULL)
    {
	strcpy (symName, "_");
	strcat (symName, "ahdlcdevstats");
	if ((rtnval = symFindByName (sysSymTbl, symName, (char **)&ahdlcdevstats, &type)) != OK)
   	{
	    strcat (message, "no ahdlcdevstats");
	    return (1);
	}
    }
    if (ahdlcfailover == NULL)
    {
	strcpy (symName, "_");
	strcat (symName, "ahdlcfailover");
	if ((rtnval = symFindByName (sysSymTbl, symName, (char **)&ahdlcfailover, &type)) != OK)
   	{
	    strcat (message, "no ahdlcfailover");
	    return (1);
	}
    }
    if (ahdlcinitstats == NULL)
    {
	strcpy (symName, "_");
	strcat (symName, "ahdlcinitstats");
	if ((rtnval = symFindByName (sysSymTbl, symName, (char **)&ahdlcinitstats, &type)) != OK)
	{
	    strcat (message, "no ahdlcinitstats");
	    return (1);
	}
    }

    return (0);
}


/********************************** linestats *****************************************/
int ahls (line,reset)

    int	line,reset;

{
    ahUnitStats	*punitstats;
    int		i;

    if (ahdlcdevstats == NULL)
    {
	printf ("linestats, no ahdlcdevstats\n");
	return (0);
    }

    if (ahdlcdevstats -> data_valid == 0)
    {
	printf ("linestats, data not valid\n");
	return (0);
    }

    ahlineheader();

    ahlinestats(line);

    ahunitheader ();

    for (i = 0; i < 32; i++)
    {
	punitstats = &(ahdlcdevstats -> unitstats[line][i]);
	if (punitstats -> inet_addr == 0)
	{
		continue;
	}
	ahunitstats(line,i);

    }

    if (reset != 0)
    {
	ahdlcinitstats (line, (short)0);
    }

    return (0);
}

int ahdlclinestats (reset)
    int reset;
{
    int		i;

    if (ahdlcdevstats == NULL)
    {
	printf ("linestats, no ahdlcdevstats\n");
	return (0);
    }

    if (ahdlcdevstats -> data_valid == 0)
    {
	printf ("linestats, data not valid\n");
	return (0);
    }

    ahlineheader();

    for (i = 0; i < 64; i++)
    {
	ahlinestats(i);
        if (reset != 0)
        {
	    ahdlcinitstats ( i, (short)0);
        }

    }

    return (0);
}

static void ahlineheader ()
{

	printf ("LINE ACT M/S    RESP IPAC     OPAC     IERR     OERR     TXFRM     TXTO     RXFRM    RXTO     RXHWER \n");
	printf ("--------------------------------------------------------------------------------------------------------------------\n");

}

static void ahlinestats(i)
    int		i;
{
    ahLineStats	*plinestats;
    char	ipackets[16];
    char	opackets[16];
    char	ierrors[16];
    char	oerrors[16];
    char	txbufs[16];
    char	rxbufs[16];
    char	txbufs_avail[16];
    char	rxbufs_avail[16];
    char	txbuf_size[16];
    char	rxbuf_size[16];
    char	txframes[16];
    char	txtimeouts[16];
    char	rxframes[16];
    char	rxtimeouts[16];
    char	rxhwerrs[16];
    char	line[16];

	plinestats = &(ahdlcdevstats -> linestats[i]);
	sprintf (line,         "%d", i);
	sprintf (ipackets,     "%d", plinestats -> ipackets);
	sprintf (opackets,     "%d", plinestats -> opackets);
	sprintf (ierrors,      "%d", plinestats -> ierrors);
	sprintf (oerrors,      "%d", plinestats -> oerrors);
	sprintf (txbufs,       "%d", plinestats -> txbufs);
	sprintf (rxbufs,       "%d", plinestats -> rxbufs);
	sprintf (txbufs_avail, "%d", plinestats -> txbufs_avail);
	sprintf (rxbufs_avail, "%d", plinestats -> rxbufs_avail);
	sprintf (txbuf_size,   "%d", plinestats -> txbuf_size);
	sprintf (rxbuf_size,   "%d", plinestats -> rxbuf_size);
	sprintf (txframes,     "%d", plinestats -> txframes);
	sprintf (txtimeouts,   "%d", plinestats -> txtimeouts);
	sprintf (rxframes,     "%d", plinestats -> rxframes);
	sprintf (rxtimeouts,   "%d", plinestats -> rxtimeouts);
	sprintf (rxhwerrs,     "%d", plinestats -> rxhwerrors);

	printf ("%-4.4s %-3.3s %-6.6s %-3.3s %-8.8s %-8.8s %-8.8s %-8.8s %-8.8s %-8.8s %-8.8s %-8.8s %-8.8s\n",
		line,
		plinestats -> line_active == 0 ? "NO" : "YES",
		plinestats -> master_slave == 0 ? "SLAVE" : "MASTER",
		plinestats -> line_responding == 0 ? "NO" : "YES",
		ipackets, opackets, ierrors, oerrors,
		txframes, txtimeouts, rxframes, rxtimeouts, rxhwerrs);
}

/********************************** unitstats ****************************************/

int ahdlcunitstats (line)
    int		line;
{
    int		i;
    ahUnitStats	*punitstats;

    if (ahdlcdevstats == NULL)
    {
	printf ("unitstats, no ahdlcdevstats");
	return (0);
    }

    if (ahdlcdevstats -> data_valid == 0)
    {
	printf ("unitstats, data not valid\n");
	return (0);
    }

    if ((line < 0) || (line > 63))
    {
	printf ("unitstats, line not 0-63\n");
	return (0);
    }

    ahunitheader ();

    for (i = 0; i < 32; i++)
    {
	punitstats = &(ahdlcdevstats -> unitstats[line][i]);
	if (punitstats -> inet_addr == 0)
	{
		continue;
	}
	ahunitstats(line,i);

    }

    return (0);
}

static void ahunitheader ()
{
    printf ("UNIT ONL INET   OLCNT  CRCERR FRAMERR FORMERR RXHWERR IPAC   OPAC   IERRS  OERRS  TXFRM  TXTO   RXFRM  RXTO   RXHWER \n");
    printf ("---------------------------------------------------------------------------------------------------------------------\n");

}

static void ahunitstats(line,i)
    int		line,i;
{
    ahUnitStats	*punitstats;
    char	inetaddr[16];
    char	onlinecnt[16];
    char	crcerr[16];
    char	framerr[16];
    char	formaterr[16];
    char	rxhwerr[16];
    char	ipackets[16];
    char	opackets[16];
    char	ierrors[16];
    char	oerrors[16];
    char	txframes[16];
    char	txtimeouts[16];
    char	rxframes[16];
    char	rxtimeouts[16];
    char	rxhwerrs[16];
    char	unit[16];

	punitstats = &(ahdlcdevstats -> unitstats[line][i]);
	sprintf (unit,      "%d", i+1);
	sprintf (inetaddr,  "%ld", punitstats -> inet_addr);
	sprintf (onlinecnt, "%d", punitstats -> online_count);
	sprintf (crcerr,    "%d", punitstats -> crc_errors);
	sprintf (framerr,   "%d", punitstats -> frame_errors);
	sprintf (formaterr, "%d", punitstats -> format_errors);
	sprintf (rxhwerr,   "%d", punitstats -> rxhw_errors);
	sprintf (ipackets,  "%d", punitstats -> ipackets);
	sprintf (opackets,  "%d", punitstats -> opackets);
	sprintf (ierrors,   "%d", punitstats -> ierrors);
	sprintf (oerrors,   "%d", punitstats -> oerrors);
	sprintf (txframes,  "%d", punitstats -> txframes);
	sprintf (txtimeouts,"%d", punitstats -> txtimeouts);
	sprintf (rxframes,  "%d", punitstats -> rxframes);
	sprintf (rxtimeouts,"%d", punitstats -> rxtimeouts);
	sprintf (rxhwerrs,  "%d", punitstats -> rxhwerrors);
 

	printf ("%-4.4s %-3.3s %-6.6s %-6.6s %-6.6s %-6.6s  %-6.6s  %-6.6s  %-6.6s %-6.6s %-6.6s %-6.6s %-6.6s %-6.6s %-6.6s %-6.6s %-6.6s \n",
		unit,
		punitstats -> online ==0 ? "NO" : "YES",
		inetaddr, onlinecnt, crcerr, framerr, formaterr,
		rxhwerr, ipackets, opackets, ierrors, oerrors,
		txframes, txtimeouts, rxframes, rxtimeouts, rxhwerrs);

}
