/***********************************
*
*	File:	serialFile.c
*	Auth:	JLD
*	Edit:	JLD	10/26/97
*
************************************/
#include <vxWorks.h>
#include <taskLib.h>
#include <ioLib.h>
#include <sysLib.h>
#include <logLib.h>
#include <sioLib.h>
#include <stdio.h>
#include <serialFile.h>

/******	Tuning/sizing parameters ******/

#define TASK_PRI		200
#define TASK_STACK		2048

/****** Private Data ******/


LINESTATUS	serialFileTable[MAXLINES];


/****** Function Declarations ******/

void serialFile(short line);
static STATUS valid_mode( char *p, int *baud, int *hwcfg );
/****** Task Names ******/

static	char		*stasknames [] =

			{"SerFile Ch0", "SerFile Ch1", "SerFile Ch2", "SerFile Ch3",
			 "SerFile Ch4", "SerFile Ch5", "SerFile Ch6", "SerFile Ch7",
			 "SerFile Ch8", "SerFile Ch9", "SerFile Ch10", "SerFile Ch11",
			 "SerFile Ch12", "SerFile Ch13", "SerFile Ch14", "SerFile Ch15",
			 "SerFile CH16", "SerFile Ch17", "SerFile Ch18", "SerFile Ch19",
			 "SerFile CH20", "SerFile Ch21", "SerFile Ch22", "SerFile Ch23",
			 "SerFile CH24", "SerFile Ch25", "SerFile Ch26", "SerFile Ch27",
			 "SerFile CH28", "SerFile Ch29", "SerFile Ch30", "SerFile Ch31"
			};

/**********
*
*		INITIALIZE THE SERIAL PORT FOR THE LINE
*
**********/

int initSerialFile(short line, char *portName, char *mode,
				   FUNCPTR rxcallback, int rxcallbackarg,
				   FUNCPTR txcallback, int txcallbackarg,
				   short ticks)

{
    int	baud, hwcfg;



    if (( line >= MAXLINES ) || ( serialFileTable[line].inuse ))
		return (ERROR);


    /* check the mode string and convert to pattern for setting hardware options */
    if ((valid_mode( mode, &baud, &hwcfg ) == ERROR)) {
		logMsg("Can't open port for serial line = %d, port %s, invalid mode %s\n",
			line,(int) portName,(int) mode,0,0,0);
		return (ERROR);
    }

    /*    Open the serial port    */

    if ((serialFileTable[line].fd = open(portName, O_RDWR, 0777)) == ERROR) {
		logMsg("Can't open port for serial line = %d, port %s\n",
			line,(int) portName,0,0,0,0);
		return (ERROR);
    }

    /* Get the pointer to the serial channel for setup and modem control */

    serialFileTable[line].pchan = sysSerialChanGet(line);
    /* have the vx works driver ignore modem control we'll do it ourselves */
    hwcfg |= CLOCAL;
    sioIoctl(serialFileTable[line].pchan, SIO_HW_OPTS_SET, (int *)hwcfg);

    /*    Save the upate rate   */
    if ((ticks == 0) || (ticks>100)) {
		serialFileTable[line].ticks = 1;
    } else {
		serialFileTable[line].ticks = ticks;
    }

 	ioctl(serialFileTable[line].fd, FIOBAUDRATE, baud);
   /*   Launch the redundant update task    */

    serialFileTable[line].rxcallback = rxcallback;
    serialFileTable[line].rxcallbackarg = rxcallbackarg;
    serialFileTable[line].txcallback = txcallback;
    serialFileTable[line].txcallbackarg = txcallbackarg;
    serialFileTable[line].inuse = TRUE;
    serialFileTable[line].txcount = 0;
    serialFileTable[line].txdone = TRUE;

   if (taskSpawn(stasknames[line], TASK_PRI, VX_FP_TASK, TASK_STACK,(FUNCPTR) serialFile,
		line, 0, 0, 0, 0, 0, 0, 0, 0, 0) == ERROR) {
		logMsg("Can't spawn task for serial line = %d, port %s\n",
				line,(int) portName,0,0,0,0);
		return (ERROR);
    }

	logMsg("Opened port for serial line = %d, port %s, fd = %d\n",
				line,(int) portName,serialFileTable[line].fd,0,0,0);
	return (serialFileTable[line].fd);
}


/**********
*
*		THE SERIALFILE TASK
*		Watches a serial port and does rxcallbacks while there are characters buffered
*
**********/
int	sfwait = TRUE;
void serialFile(short line)
{
 
    int			bytesAvail;

    for (;;) {


		/*   Wait for characters to be available */
		bytesAvail = 0;
		while (!bytesAvail) {
			taskDelay(serialFileTable[line].ticks);
			ioctl(serialFileTable[line].fd, FIONREAD,(int)  &bytesAvail);
		}
		while (bytesAvail) {
		    /* Allow Driver to consume data */
		    (*serialFileTable[line].rxcallback)(serialFileTable[line].rxcallbackarg);
		    ioctl(serialFileTable[line].fd, FIONREAD,(int) &bytesAvail);
		}
    }
}

/**********
*
*		UTILITY ROUTINES FOR DRIVERS
*
**********/

/**********
*
*		GET THE SERIAL DATA
*
**********/			
int readfile_rx_char( unsigned short line, unsigned char *rxbyte,
		     unsigned short *rxstat, unsigned short *rxerrmask)
{
    int			bytesAvail;
	char		databyte;

	/* This Routine should only be accessed by a call back from above	*/
	/* Otherwise it will result in an underflow of data being available	*/
			
			
	*rxerrmask = 0xff;
	*rxstat = 0xff;

	if (( line >= MAXLINES ) || ( !serialFileTable[line].inuse )) {
		return (FALSE);
	}
	ioctl(serialFileTable[line].fd, FIONREAD, (int) &bytesAvail);
	if (!bytesAvail) {
		logMsg("Serial File Read underflow line = %d\n",line,0,0,0,0,0);
		return (FALSE);
	}
	read(serialFileTable[line].fd, &databyte, 1);
	*rxbyte = databyte;
	*rxerrmask = 0xff;
	*rxstat = 0x0;

	return (TRUE);

}

/**********
*
*		 PROCESS SERIAL DATA WRITE ENABLE
*			clead the write finished flag
*			start callin the txcall back until a txdisable is called
*
**********/
int	enablefiletx (short line)
{

	if (( line >= MAXLINES ) || ( !serialFileTable[line].inuse )) {
		logMsg("Serial File Enable Tx to unopened line = %d\n",line,0,0,0,0,0);
		return (FALSE);
	}
	serialFileTable[line].txdone = FALSE;
	while (!serialFileTable[line].txdone) {
		    (*serialFileTable[line].txcallback)(serialFileTable[line].txcallbackarg);
	}
	write(serialFileTable[line].fd, serialFileTable[line].txbuf, serialFileTable[line].txcount);
	serialFileTable[line].txcount = 0;
	return (TRUE);
}
/**********
*
*		 PROCESS SERIAL DATA WRITE DISABLE
*			set the finish write flag
*
**********/	
int	disablefiletx (short line)
{

	if (( line >= MAXLINES ) || ( !serialFileTable[line].inuse )) {
		logMsg("Serial File Disable Tx to unopened line = %d\n",line,0,0,0,0,0);
		return (FALSE);
	}
	serialFileTable[line].txdone = TRUE;
	return (TRUE);
}
/**********
*
*		 PROCESS SERIAL DATA WRITE
*			buffer the data until the buffer is full then do a write
*			and reset the index
*
**********/			
int writefile_tx_char( unsigned short line, unsigned char txbyte,
		     unsigned short *txstat, unsigned short *txerrmask)
{


	/* This Routine should only be accessed by a call back from above	*/
	/* Otherwise it will result in an underflow of data being available	*/
			
			
	*txerrmask = 0xff;
	*txstat = 0xff;

	if (( line >= MAXLINES ) || ( !serialFileTable[line].inuse )) {
		logMsg("Serial File Write to unopened line = %d\n",line,0,0,0,0,0);
		return (FALSE);
	}
	serialFileTable[line].txbuf[serialFileTable[line].txcount++] = txbyte;
	if (serialFileTable[line].txcount >= TXBUFSIZE) {
		write(serialFileTable[line].fd, serialFileTable[line].txbuf, serialFileTable[line].txcount);
		serialFileTable[line].txcount = 0;
	}
	*txstat = 0x0;
	return (TRUE);
}

/*****************************************
*
* valid_mode - verifies the contents of an ASCII, null-terminated 'DOS' mode string
*
* This routine scans a standard 'DOS' format mode string and verifies the
* individual components as supported. It produces a packed configuration
* descriptor in <*cfg> as a natural by-product.
*
* RETURNS: OK if the mode string represents a valid combination, ERROR otherwise.
*
* ARGUMENTS
* .IP "char *p" 20
* Address of mode string to be evaluated.
* .IP "int *baud"
* Address of word for baud rate.
* .IP "int *hwcfg"
* Address of word for packed equivalent of mode string.
* .LP
*
* NOMANUAL
*/
static int valid_baudrates[] = {115200, 57600, 38400, 19200, 9600, 4800, 2400, 1200, 300 };
#define BAUDMAX	9
static USHORT valid_charsizes[] = { 8, 7, 6, 5 };
static USHORT hwcharsize [] = {CS8, CS7, CS6, CS5 };
#define CSIZEMAX	4
static USHORT valid_parity[]    = { 'N', 'O', 'E' };
static USHORT hwparity[] = { 0, PARODD, PARENB };
#define PARITYMAX	3
static USHORT valid_stops[]     = { 1, 2 };
static USHORT hwstops[] = { 0,STOPB };
#define STOPSMAX	2

static STATUS valid_mode( char *p, int *baud, int *hwcfg )
{
  register int bd, pr, cs, sb;
  int wbaudrate=0, wcharsize=0, wstopbits=0;
  UCHAR wparity=0;
  if( sscanf( p, "%d,%c,%d,%d",   &wbaudrate, &wparity, &wcharsize, &wstopbits ) == 4 )
  {
    for( bd=0; bd < BAUDMAX; ++bd ) if( wbaudrate == valid_baudrates[ bd ] )
    {
      for( pr=0; pr < PARITYMAX; ++pr ) if( (wparity & ~0x20) == valid_parity[ pr ] )
      {
        for( cs=0; cs < CSIZEMAX; ++cs ) if( wcharsize == valid_charsizes[ cs ] )
        {
          for( sb=0; sb < STOPSMAX; ++sb ) if( wstopbits == valid_stops[ sb ] )
          {
	    *baud = wbaudrate;
            *hwcfg =  (hwparity[pr]) | (hwcharsize[cs]) | (hwstops[sb]);
            return OK;
          }
        }
      }
    }
  }
  return ERROR;
}



/****************************
*
*	Stup for serlib for now
*
*****************************/

struct _line **linedefs;			/*serLib.c*/
struct _hwline **hwlinedefs;


STATUS serLibInit( int x )
{
	return (FALSE);
}
STATUS serOpen( int x)
{
	return (FALSE);
}
STATUS serClose( int x)
{
	return (FALSE);
}
STATUS serWrite( int x, UCHAR y )
{
	return (FALSE);
}
STATUS serWriteBlk( int x, UCHAR* y, int z)
{
	return (FALSE);
}
STATUS serRead( int x)
{
	return (FALSE);
}
STATUS serReadBlk( int w, UCHAR* x , int y, UCHAR z )
{
	return (FALSE);
}
STATUS serIoctl( int x, int y )
{
	return (FALSE);
}
void serMapShow( )
{
	return;
}

