/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <vxWorks.h>
#include <taskLib.h>
#include <ioLib.h>
#include <stdio.h>

#include "serialTask.h"

/******	Tuning/sizing parameters ******/

#define MAX_NULLS		128
#define TASK_PRI		110
#define TASK_STACK		2048


/****** Private Data ******/

static char		strobeBuff[MAX_NULLS];
static int		ioSize = 50;	/* Default 50 bytes per xmit */
static int		ioDelay = 50;	/* Default 50 ticks between xmits */


/****** Function Declarations ******/

void redundStrobe(int serialFd, int index);



/****************************************************************/
/*								*/
/*		INITIALIZE THE SERIAL HANDSHAKE TASK		*/
/*								*/
/****************************************************************/

int
initRedundStrobe(portName, index)
char		*portName;		/* Device name */
int		index;
{
    int		serialFd;

    /*
    Open the serial port
    */

    if ((serialFd = open(portName, O_RDWR, 0x666)) == ERROR)
	{
	perror("Can't open serial line for redundancy");
	return (ERROR);
	}

    /*
    Launch the tomato head strobe task
    */

    if (taskSpawn("redundStrobe", TASK_PRI, VX_FP_TASK, TASK_STACK, (FUNCPTR)redundStrobe,
		serialFd, index, 0, 0, 0, 0, 0, 0, 0, 0) == ERROR)
	return (ERROR);
    return (!ERROR);
}


/****************************************************************/
/*								*/
/*		THE REDUNDANT UPDATE TASK			*/
/*								*/
/****************************************************************/

void redundStrobe(int serialFd, int index)
{
    int		localState;
    long	myPid;

    for (;;)
	{
	localState = localSerialRedund(index);
	if ((localState == 'M') || (localStrobeSerialRedund(index) && (localState != 'T')))
	    write(serialFd, strobeBuff, ioSize);
	else
	    myPid = taskIdSelf();
        taskDelay(ioDelay);
	}
}


/****************************************************************/
/*								*/
/*		UTILITY ROUTINES FOR DRIVERS			*/
/*								*/
/****************************************************************/

int
setStrobeLength(newLen)
int newLen;
{
	if (newLen <= 0 || newLen > sizeof(strobeBuff))
		{
		printErr("Invalid new length ignored.\n");
		return (ERROR);
		}
	printErr("Strobe length changed from %d to %d\n", ioSize, newLen);
	ioSize = newLen;
	return (!ERROR);
}

int
setStrobeDelay(newDelay)
int newDelay;
{
	if (newDelay < 0)
		{
		printErr("Invalid new delay ignored.\n");
		return (ERROR);
		}
	printErr("Strobe delay changed from %d to %d\n", ioDelay, newDelay);
	ioDelay = newDelay;
	return (!ERROR);
}
