#include <in.h>
#include "dsk.h"

extern DPA_INFO	dpaInfo;

AVL_TREE  *dskTree;
char	appName[100];
char	currentGroup[PVNAME_SZ+FLDNAME_SZ+1];
void	dskOpen(),dskRead(),dskWrite(),sigHandler(),dskClose();
void	*memberHandle;
void	*updateHandle;
int	signals[] = {SIGPIPE,SIGINT,SIGHUP,0};
int  appSocket;

main(int argc,char *argv[])
{
char	*slash;

slash = (char *)strrchr(argv[0],'/');
if(slash)
  {
    slash++;
    strcpy(appName,slash);
  }
else
  strcpy(appName,argv[0]);
tisOpenLog ("dsk", 6, 2, NULL, NULL);
dskTree = avl_init(dskEntryCompare);		/* init the AVL tree library */
/*
	Init DPA sub-system
*/
if((appSocket = dpaInit(appName,dskOpen,dskRead,0,0,sigHandler,signals)) < 0)
  exit(1);
dpaSetResMode(ALWAYS_MASTER);
dpaConfigure(appName);
/*
  Loop here and display sizes for open channels every 10 seconds. Also check
  for alarm conditions and send alarms if necassary.
*/
while(1)
  {
    if(dpaEvent(5.0) == DPA_ZEROREAD)			/* Five (5) second scan rate */
      {
	tisSysLog(0,2, "Zero length read from hsrvTask");
	exit(2);
      }
    avl_traverse(dskTree,INORDER,dskDisplay,(void *)0);
  }
}
/*
  Handle signals
*/
void	sigHandler(int whichone)
{
switch(whichone)
  {
  case SIGPIPE:
    tisSysLog (0,0, "Recieved SIGPIPE on socket %d",appSocket);
    break;
  case SIGINT:
  case SIGHUP:
    tisSysLog (0,0, "DSK:exiting due to ^C or -INT");
    break;
  default:
    tisSysLog (0,0, "Unknown signal %d",whichone);
  }
avl_empty(dskTree,avl_freenode);
exit(1);
}
int  dskEntryCompare(void * e1, void *e2)
  {
    return(strcmp( ((DSK_ENTRY_PTR)e1)->dskPath,((DSK_ENTRY_PTR)e2)->dskPath));
  }
/*
  This function returns the size of the selected item. If the item is
  a file then the size of the file in bytes is returned. If the item
  is a path name, the value returned is the percent in-use for the 
  file system the path is on. If an error occurs while trying to obtain
  the information, then the size is set to -1. I used -1 because zero
  could be a valid file size(empty file). This also allows the user to
  set the LowLow alarm limit to 0 and the severity to major to signify
  a missing file.
*/
void dskItemSize(DSK_ENTRY_PTR item)
{
int st;
double totalBlocks,freeBlocks;
struct stat buff;
#ifdef SOLARIS
struct statvfs fsbuff;
#else
struct statfs fsbuff;
#endif
if(item->dskPathType == DSK_VOL)
  {
#ifdef SOLARIS
    st = statvfs(item->dskPath,&fsbuff);
    totalBlocks = fsbuff.f_blocks;
    freeBlocks = fsbuff.f_bavail;
#else
    st = statfs(item->dskPath,&fsbuff);
    totalBlocks = fsbuff.f_blocks;
    freeBlocks = fsbuff.f_bavail;
#endif
    if(st < 0)
      {
/* 	tisSysLog(0,0,"Error getting 'statfs' for ...<%s>...%m",&item->dskPath[0]); */
	item->dskSize = -1.0;
      }
    else
      item->dskSize = 100.0 - ((freeBlocks/totalBlocks)*100.0);	/* display as percentage in-use */
  }
if(item->dskPathType == DSK_FILE)
  {
    if(stat(item->dskPath,&buff) < 0)
      {
/* 	tisSysLog(0,0,"Error getting 'stat' for ...<%s>...%m",&item->dskPath[0]); */
	item->dskSize = -1.0;
      }
    else
      item->dskSize = buff.st_size;
  }
}
