/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <hsmain.h>
FD_INFO  fdInfo;
char hsVerbose = V_INFO;
char *sendBuff = 0;					/* calloc'ed area for message transmission */
char *sendOut;
int sendBuffSize = 0;					/* size of area */
int sendBuffIndex = 0;					/* index into area */

long	hsKeepalive;					/* boomarang rate */
long	hsExchange;					/* file info exchange rate */
int hsCount;						/* number of messages sent */

int hsListenSocket;
int numToConnect,listenOnly,numConnected = 0;
int maxConnected = 0;
/*
	Message level strings
*/
char	*msgLevel[] =
{
  "Silent",
  "Informational",
  "Status",
  "DEBUG!"
};
AVL_TREE  *fileTree;
HOST_NODE hostParent[HSYNC_MAX_HOSTS];

HOST_NODE_PTR lastHost = 0;

main(int argc,char **argv)
{
  struct timeval maxWait,currTime,exitTime;
  int status;
  struct sockaddr_in	sockAddr;
  int		len,srcCheckCounter=HSYNC_SRC_CHECK_TIMEOUT;
  time_t srcListTime = 0;				/* zero to start */
  time_t checkTimer = 0;
  int		addr,cmd,i,newfd;
  struct stat stBuff;
#ifdef	BLAST_IT
  char rcpCmd[MAXPATHLEN];
#endif	BLAST_IT

  extern	char	*optarg;
  extern	int	optind;

  tisOpenLog("hsync", 6, 2, NULL, NULL);
#ifndef SOLARIS
  signal(SIGINT,hsSignalHandler);	       	/* handle ^C from keyboard */
  signal(SIGHUP,hsSignalHandler);	       	/* handle task termination */
  signal(SIGUSR1,hsSignalHandler);		/* increment verbose flag */
  signal(SIGUSR2,hsSignalHandler);		/* decrement verbose flag */
  signal(SIGPIPE,hsSignalHandler);		/* handle broken pipe signals */
  signal(SIGCHLD,hsSignalHandler);		/* handle child return signals */
#else
  sigset(SIGINT,hsSignalHandler);	       	/* handle ^C from keyboard */
  sigset(SIGHUP,hsSignalHandler);	       	/* handle task termination */
  sigset(SIGUSR1,hsSignalHandler);		/* increment verbose flag */
  sigset(SIGUSR2,hsSignalHandler);		/* decrement verbose flag */
  sigset(SIGPIPE,hsSignalHandler);		/* handle broken pipe signals */
  sigset(SIGCHLD,hsSignalHandler);		/* handle child return signals */
#endif
  hsExchange = HS_POLL_TIMER;
  hsCount = HS_MESSAGE_COUNT;
  while((cmd = getopt(argc,argv,"v:x:c:")) != -1)
    {
      switch(cmd)
	{
	case 'v':
	  hsVerbose = V_DEBUG;
	  if(*optarg != '-')
	    {
	      hsVerbose = abs(atoi(optarg));	/* set initial verbosity level */
	      break;
	    }
	  break;
	case 'x':
	  if(*optarg != '-')
	    {
	      hsExchange = abs(atoi(optarg));	/* set exchange file info rate */
	    }
	  else
	    {
	      fprintf(stderr,"Exchange rate must have a value.\n");
	      exit(1);
	    }
	  break;
	case 'c':
	  if(*optarg != '-')
	    {
	      hsCount = abs(atoi(optarg));	/* set exchange file info rate */
	    }
	  else
	    {
	      fprintf(stderr,"Message count must have a value.\n");
	      exit(1);
	    }
	  break;
	case '?':
	  fprintf(stderr,"Usage is:\n");
	  fprintf(stderr,"\thsync -v level -x xrate -c count host1 [host2]\n");
	  fprintf(stderr,"\tlevel - debug level (%s).\n",msgLevel[hsVerbose]);
	  fprintf(stderr,"\txrate - exchange rate (default %d).\n", HS_POLL_TIMER);
	  fprintf(stderr,"\tcount - message count (default %d).\n", HS_MESSAGE_COUNT);
	  fprintf(stderr,"host1 [host2] - 1 (or 2) hosts to connect with for synch info\n");
	  exit(1);
	}
    }
  /* get the rest of the command line args. These will be the hosts designated for
     connection. Max=2(for now!).
     */
  for(i=0;i < HSYNC_MAX_HOSTS;i++)
    {
      hostParent[i].hostName[0] = 0;
      hostParent[i].hostFD = -1;
      hostParent[i].hostState = HOST_NONET;
      hostParent[i].hostCharCount = 0;
      hostParent[i].hostNextChar = 0;
      hostParent[i].hostAddr = 0;
    }
  for(numToConnect = 0,i = optind;i < argc;i++)
    strcpy(&hostParent[numToConnect++].hostName[0],argv[i]);
  if(numToConnect == 0) listenOnly = 1;
  fileTree = avl_init(hsFileCompare);
  if(hsInitNetwork() < 0)
    {
      VERBOSE(V_DEBUG)(0,0,"Network init error...%m");
      exit(1);
    }
  if(hsBuildFileTree(fileTree) < 0)    		   	/* build file tree return time of hsync.src file */
    {
      VERBOSE(V_DEBUG)(0,0,"Error building file list...%m");
      exit(1);
    }
  sendBuff = (char *)calloc(1,sendBuffSize+1);
  sendOut = sendBuff;
  stat(HSYNC_FILE_LIST,&stBuff);			/* get new time AFTER file is read & closed */
  srcListTime = stBuff.st_mtime;

  hsInitializeHosts(hostParent,numToConnect);      		/* init hostParent array; */
  VERBOSE(V_INFO)(0,0,"hsync started...exchange=%d secs.,count=%d messages/block",
									hsExchange,hsCount);

  for(i = 0;i < numToConnect;i++)
    hsConnectToHost(&hostParent[i]);
  while(1)
    {
      maxWait.tv_sec = HS_POLL_TIMER;
      maxWait.tv_usec = 0;
      while((maxWait.tv_sec >= 0) && (maxWait.tv_usec >= 0))
	{
	  gettimeofday(&currTime,0);
	  memcpy(&fdInfo.allFds,&fdInfo.myFds,sizeof(fd_set));
	  status = select(fdInfo.maxFd+1,&fdInfo.allFds,0,0,&maxWait);
	  if(status == 0)
	    break;
	  else if(status < 0)
	    {
	      VERBOSE(V_DEBUG)(0,0,"Error from select...%m");
	    }
	  /*
	    A socket has something for us to read. Find out which one and
	    read it.
	    */
	  else
	    {
	      if((hsListenSocket != -1) && FD_ISSET(hsListenSocket,&fdInfo.allFds))   	/* listening for connects */
		{
		  len = sizeof(sockAddr);
		  if(( newfd = accept(hsListenSocket,(struct sockaddr *)&sockAddr,&len)) < 0)
		    break;
		  addr = sockAddr.sin_addr.s_addr;
		  hsHandleConnect(addr,newfd);		/* perform accept */
		}
/*
	Must have been one from a HOST that I am connected
	to already.

	3-Dec-96 Try up to the maximum number of connections
	we know about.
*/
	      else
		{
		  for(i = 0;i < maxConnected;i++)
		    {
		      hsIncomingMsg(fileTree,&hostParent[i]);
		    }
		}
	    }
/*
	adjust timer
*/
	  gettimeofday(&exitTime,0);
	  exitTime.tv_sec -= currTime.tv_sec;
	  exitTime.tv_usec -= currTime.tv_usec;
	  if(exitTime.tv_usec < 0)
	    {
	      exitTime.tv_usec += 1000000;
	      exitTime.tv_sec--;
	    }
	  maxWait.tv_sec -= exitTime.tv_sec;
	  maxWait.tv_usec -= exitTime.tv_usec;
	  if(maxWait.tv_usec < 0)
	    {
	      maxWait.tv_usec += 1000000;
	      maxWait.tv_sec--;
	    }
    }/* inner while */
      if(--srcCheckCounter <= 0)
	{
	  srcCheckCounter = HSYNC_SRC_CHECK_TIMEOUT;
	  if(stat(HSYNC_FILE_LIST,&stBuff) != 1)
	    {
	      if(stBuff.st_mtime > srcListTime)
		{
		  VERBOSE(V_INFO)(0,0,"Re-reading list file");
		  avl_empty(fileTree,avl_freenode);	/* empty and delete tree */
		  free(sendBuff);
		  sendBuffSize = 0;
		  sendBuffIndex = 0;
		  if(hsBuildFileTree(fileTree) < 0)
		    {
		      close(hsListenSocket);
		      for(i = 0;i < numToConnect;i++)
			hsCloseSockets(&hostParent[i]);
		      VERBOSE(V_INFO)(0,0,"Error re-building file list...%m");
		      exit(1);
		    }
		  sendBuff = (char *)calloc(1,sendBuffSize+1);
		  sendOut = sendBuff;
		  stat(HSYNC_FILE_LIST,&stBuff);	/* get new time AFTER file is read & closed */
		  srcListTime = stBuff.st_mtime;
#ifdef BLAST_IT
		  for(i = 0;i < maxConnected;i++)
		    {
		      if(hostParent[i].hostFD != -1)
			{
			  sprintf(rcpCmd,"rcp -p %s %s:%s &",HSYNC_FILE_LIST,hostParent[i].hostName,HSYNC_FILE_LIST);
			  VERBOSE(V_DEBUG)(0,0,"RCP command %s\n",rcpCmd);
			  if(system(rcpCmd) < 0)
			    VERBOSE(V_DEBUG)(0,0,"Error from SYSTEM/RCP: %m");
			}
		    }
#endif
		}
	    }
	  else
	    {
	      close(hsListenSocket);
	      for(i = 0;i < numToConnect;i++)
		hsCloseSockets(&hostParent[i]);
	      VERBOSE(V_INFO)(0,0,"Error checking the file list...exiting!...%m");
	      exit(1);
	    }
	}
      sendBuff = sendOut;			/* reset pointer for output buffer */
      /*
	See if it is time to check the file Ds & Ts. Use same rate as "src" file check
      */
      if(--checkTimer <= 0)
	{
	  checkTimer = HSYNC_SRC_CHECK_TIMEOUT;
	  VERBOSE(V_STATUS)(0,0,"Checking local files.");
	  avl_traverse(fileTree,INORDER,hsGetFileInfo,0);
	}
      /*
	File info messages are made from hosts which do not
	specify another host to connect to.
      */
      if(numToConnect == 0)
	{
	  for(i = 0;i < maxConnected;i++)
	    {
	      if((hostParent[i].hostFD != -1) && 
		 ((time(0) - hostParent[i].hostExchangeTime) > hsExchange))
		{
		  hsSendFileInfo(&hostParent[i]);
		}
	    }
	}
      for(i = 0;i < numToConnect;i++)
	hsConnectToHost(&hostParent[i]);
    }
}
void hsConnectToHost(HOST_NODE_PTR node)
{
  struct	sockaddr_in	hostaddr;
  int len,trueFlag = 1;;

  if(node->hostFD == -1)
    {
      VERBOSE(V_DEBUG)(0,0,"Attempting to connect to %s",node->hostName);
      hostaddr.sin_family      	= AF_INET;
      hostaddr.sin_addr.s_addr	= htonl(node->hostAddr);
      hostaddr.sin_port		= htons(HSYNC_UDP_PORT);
      node->hostFD = socket(AF_INET,SOCK_STREAM,0);
      len = sizeof(struct sockaddr);
      if(connect(node->hostFD,(struct sockaddr *)&hostaddr,len) != -1)
	{
        maxSocket(node->hostFD);
	  hsSetNoBlock(node);
	  setsockopt(node -> hostFD,IPPROTO_TCP,TCP_NODELAY,(char *)&trueFlag,sizeof(trueFlag));
	  node->hostState = HOST_OK;		/* OK to proceed */
	  node->hostRxTime = time(0);
	  node->hostExchangeTime = time(0);
	  if(node->hostFD > fdInfo.maxFd) fdInfo.maxFd = node->hostFD;
	  FD_SET(node->hostFD,&fdInfo.myFds);
	  node->hostCharCount = 0;
	  node->hostNextChar = 0;
	  node->hostSendIndex = 0;
	  node->hostPass = HOST_PASS1;			/* start with pass one */
	  node->hostSendStatus = HS_SEND_INFO;		/* start out by sending info */
	  numConnected++;
	  if(numConnected > maxConnected) maxConnected = numConnected;
	  VERBOSE(V_DEBUG)(0,0,"Connected to %s",node->hostName);
	}
      else
	{
	  VERBOSE(V_DEBUG)(0,0,"Error from HOST %s connect...%m",node->hostName);
	  close(node->hostFD);
	  FD_CLR(node->hostFD,&fdInfo.myFds);
	  node->hostFD = -1;
	  node->hostState = HOST_NONET;
	}
    }
}

/*
  Compare only the file names. The paths will be used in the 'rcp'
  operations.
*/
int hsFileCompare(void *f1,void *f2) /* compare function for file names */
{
char *n1,*n2;
if((n1 = strrchr(((FILE_NODE_PTR)f1)->fileName,'/')) == 0) 
	n1 = &((FILE_NODE_PTR)f1)->fileName[0];
if((n2 = strrchr(((FILE_NODE_PTR)f2)->fileName,'/')) == 0) 
	n2 = &((FILE_NODE_PTR)f2)->fileName[0];
return(strcmp(n1,n2));
}

void hsSignalHandler(int whichone)
{
time_t	end_time;
int i,who;
switch(whichone)
  {
  case SIGINT:
  case SIGHUP:
    close(hsListenSocket);
    for(i = 0;i < numToConnect;i++)
      hsCloseSockets(&hostParent[i]);
    end_time = time(0);
    VERBOSE(V_INFO)(0,0,"Exiting due to ^C: or SIGHUP %s",ctime(&end_time));
    exit(1);
  case SIGUSR1:
    if(++hsVerbose > V_DEBUG) hsVerbose = V_DEBUG;
    tisSysLog(0,0,"Message level now at %s\n",msgLevel[hsVerbose]);
    break;
  case SIGUSR2:
    if(--hsVerbose < 0) hsVerbose = 0;
    tisSysLog(0,0,"Message level now at %s\n",msgLevel[hsVerbose]);
    break;
  case SIGPIPE:
    if(lastHost != 0)
      hsTcpError(lastHost);
    break;
  case SIGCHLD:
    wait(&who);
    if(who != 0)
      VERBOSE(V_STATUS)(0,0,"Child process (RCP) exiting... %08X",who);
    break;
  }
}

void hsCloseSockets(HOST_NODE *node)
{
close(node->hostFD);
}

/*
  Find a host given its inet addr. Return the index in the hostInfo array of a found
  host or the next available slot.
*/
int hsyncFindHost(int addr)
{
  int i;
  for(i = 0;i < HSYNC_MAX_HOSTS;i++)
    {
      if(addr == hostParent[i].hostAddr)
	return(i);
      else if(hostParent[i].hostAddr == 0)
	return(i);
    }
return(-1);		/* this should NOT happen but return -1 anyway */
}
