/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

#ifndef	INChsmainH
#define	INChsmainH

#ifndef AXP
#include	<sys/param.h>
#else
#include	<getpath.h>
#endif

#include	<string.h>
#include	<stdio.h>
#include	<signal.h>
#include	<sys/time.h>
#include	<sys/types.h>
#include	<sys/socket.h>
#include	<sys/ioctl.h>
#include	<sys/fcntl.h>
#include	<netinet/in.h>
#include	<netinet/tcp.h>
#include	<netdb.h>
#include	<error.h>
#include	<errno.h>
#include	<syslog.h>
#include	<sys/stat.h>

#include	"avl.h"
#include	"msgMon.h"

#define	HSYNC_UDP_PORT		9595

#define IOCNAMESZ 16
#define HOST_BUFF_SIZE 4096
#define FILE_NAME_SIZE 32
#define HS_POLL_TIMER 20
#define HS_MESSAGE_COUNT 20
#define	V_DEBUG	3
#define	V_STATUS	2
#define	V_INFO	1
#define	VERBOSE(x) if(hsVerbose >= x) tisSysLog
#ifdef	WIN32
#define HSYNC_FILE_LIST "\\tis4000db\\hsync.src"
#else
#define HSYNC_FILE_LIST "/tis4000db/hsync.src"
#endif	WIN32
#define HSYNC_SRC_CHECK_TIMEOUT 3
#define HSYNC_DBF_FILE_EXT ".dbf"
#define HSYNC_MDX_FILE_EXT ".mdx"
#define HSYNC_MAX_HOSTS 64
#define HSYNC_SEND_SIZE 1000
/*
  Message type codes
*/
enum MSG_TYPES {HS_MSG_INFO=1,HS_MSG_REQUEST,HS_MSG_WAKEUP,HS_MSG_BOOMERANG};
/*
  Host states
*/
typedef enum {HOST_NONET,HOST_OK,HOST_SENDING} HOST_STATE;
typedef enum {HOST_PASS1,HOST_PASS2} HOST_PASS;
typedef enum {FILE_DELETED,FILE_VALID} FILE_STATUS;
typedef enum {HS_SEND_INFO,HS_NO_SEND_INFO} SEND_STATUS;
/* The HOST_NODE structure contains info about each host in the system. The */
/* info is obtained from the name server. */
/*
   GCT 25-Mar-1997

   The new transmit scheme for hsync is to send large portions of the file information
   for each connected host. Also since all hosts get the same info this scheme will only
   obtain the file date/time information once per pass. The current scheme gets the info
   for each file every time for every host.
*/

typedef struct
{
  HOST_PASS	hostPass;
  char		hostName[IOCNAMESZ+1];
  unsigned int 	hostAddr;
  int		hostFD;
  HOST_STATE	hostState;
  char		hostBuff[HOST_BUFF_SIZE];
  int		hostNextChar;
  int		hostCharCount;
  time_t       	hostRxTime;
  time_t	hostTxTime;
  time_t	hostExchangeTime;
  int		hostSendIndex;
  SEND_STATUS	hostSendStatus;
} HOST_NODE,*HOST_NODE_PTR;

#define HOST_NODE_SZ sizeof(HOST_NODE)

/* The FILE_NODE structure contains information about each file in the file */
/* list file. The file list file contains the full path name and file name */
/* of the file(s) to be synchronized between systems. */

typedef struct
{
  char 	fileName[MAXPATHLEN];
  char 	fileRemoteHost[IOCNAMESZ+1];
  HOST_NODE_PTR   fileRecentHost;
  time_t fileMaxTime;
  time_t fileLocalTime;
  FILE_STATUS  	fileStatus;	/* 1 means file in list...0 means deleted from list */
} FILE_NODE,*FILE_NODE_PTR;

#define FILE_NODE_SIZE sizeof(FILE_NODE)

typedef struct
{
  fd_set	myFds;
  fd_set	allFds;
  int		maxFd;
} FD_INFO,*FD_INFO_PTR;

/*
  message structures
*/
typedef struct
{
  time_t infoTime;
  int    infoNameSize;
} INFO_MSG;

typedef struct
{
  int recNameSize;			/* the size is followed by the file name */
} REC_MSG;

/* These are the prototypes for the support functions in hsync. */

/*
  Functions from hsInit.c
*/
int hsInitNetwork(void);		/* initialize network stuff */
int hsInitializeHosts(HOST_NODE [],int); /* initialize host array */
int hsBuildFileTree(AVL_TREE *);	/* read file list file...add to tree...check file time  */
int hsyncFindHost(int addr);		/* find index into hostParent array given addr */
/*
  Functions from hsMain.c
*/
void hsPollHost(AVL_TREE *,HOST_NODE_PTR); /* attempt to talk to hosts */
void hsCheckFile(HOST_NODE_PTR,FILE_NODE_PTR);     	/* check mod time of files in tree send if necassary*/
int hsFileCompare(void *,void *);	/* compare function for file names */
void hsSignalHandler(int);		/* handle the usual signals */
void hsCloseSockets(HOST_NODE_PTR);	/* close all sockets */
/*
  Functions from hsIOUtils.c
*/
void hsIncomingMsg(AVL_TREE *,HOST_NODE_PTR);	/* check for incoming messages from hosts */
int  hsReadChars(HOST_NODE_PTR,char *,int); /* read TCP buffer for given host */
int  hsWriteChars(HOST_NODE_PTR,char *,int); /* send chars to host */
int  hsHaveCompleteMsg(HOST_NODE_PTR); 	/* check for complete messages */
int  hsReadFD(HOST_NODE_PTR);		/* drain FD for this host */
int  hsTcpError(HOST_NODE_PTR);		/* in case of an error... */
void hsHandleConnect(int addr,int newfd);    /* other host has connected */
void hsSendBoomerang(HOST_NODE_PTR);	/* see if host is talking */
void hsSendFileInfo(HOST_NODE_PTR);	/* send file info to hosts */
void hsDoAccept(int ,HOST_NODE_PTR,int);/* loop for HOST address...assign fd when found */
void hsSetNoBlock(HOST_NODE_PTR);	/* set non-blocking I/O System sesitive! */
void hsConnectToHost(HOST_NODE_PTR node); /* Attempt to connect to host from command line */
void hsGetFileInfo(void *p1,void *finfo);     	/* referesh file D&T  */
char *hsGet20(HOST_NODE_PTR node,int *size,int *count);	/* return pointer to & size of up to hsCount complete messages */
int hsFillSendBuff(char *buff,FILE_NODE_PTR finfo);	/* fill buffer with INFO_MSG stuff return size of message */
/*
  externs
*/
extern FD_INFO fdInfo;
extern int hsListenSocket,hsUDPSocket;
extern char	hsVerbose;
extern struct dbdatabase	DB_Database;
extern int numToConnect;
extern int numConnected;
extern int maxConnected;
extern HOST_NODE hostParent[];
extern HOST_NODE_PTR lastHost;
extern char *sendBuff;
extern char *sendOut;
extern int sendBuffSize;
extern int sendBuffIndex;
extern int hsCount;

#endif	INChsmainH
