/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
  rerouted.c - TIS gateway route management daemon
  1993.06.25,ktb

  de-kai-ized 8/11/95 dpd

  -d 	enable debugging
  -p #  set UDP port id

  SIGUSR1	set the debug flag
  SIGUSR2	reset the debug flag
  SIGTERM	EXIT gracefully
*/

#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <malloc.h>
#include <netdb.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <varargs.h>
#include <netinet/in.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <net/if.h>
#include <net/route.h>

/*
 * this is defined in stdio.h, but due to something not
 * being defined, the compiler does not see it
 */
extern FILE *popen (char *, char *);

#include "reroute.h"
#include "msgMon.h"

/* Functions defined in this module */

static	void	set_debug ();
static	void	reset_debug ();
static	char	*dot_addr ();
static	void	term ();
static	int	svc_routechange ();
static	void	run_active_tables ();
static	unsigned long makeBin ();
static	int	findNetStatus ();

static	int	debug;
static	int	pid, lockFd, gatewayFd;
static	char	*lockfname = "/tis4000/tasks/rerouted.lock";
static	int	portid = GATEPORT;
static	int	si = 0;
static	u_long	stk[512];
static	char	dotAddr1 [16], dotAddr2 [16], dotAddr3 [16];
static	char	*progName;
static	int	running;

/* Network table */

#define	MAX_NETS	100
static	struct nettable {
	u_long		net;
	u_long		gateway;
	u_long		status;
} netTable [MAX_NETS];
static	int	numNets = 0;

/*----------*/
main (int argc, char *argv[])

    {

	fd_set	readfds;
	int	optval;
	struct sockaddr_in sin;
	int	c, i;
	extern	char	*optarg;

	pid = getpid ();
	progName = argv [0];

	tisOpenLog ("rerouted", 6, 2, NULL, NULL);
	tisSysLog (0,0, "rerouted strated with pid %d", pid);

	/* Process command line options */

	debug = 0;
	while ((c = getopt (argc, argv, "dp:")) != -1)
	    switch (c) {
		case 'd':	/* Debugging on */
		    debug = 1;
		    break;
		case 'p':
		    portid = atoi (optarg);
		    break;
	    }

	/* Do things that make us a daemon */

#ifdef SOLARIS
	setpgrp();
#else
#ifdef AXP
	setpgrp();
	setpgid(0,0);
#else
	setpgrp( 0, pid);
#endif
#endif

	lockFd = open (lockfname, O_RDWR|O_CREAT|O_TRUNC, 0644);
	if (lockFd == -1) {
	    tisSysLog (0,0, "rerouted(%d): Unable to create lock file (%s)",
				pid, lockfname);
	    exit (-1);
	}
#ifdef SOLARIS
	if (lockf (lockFd, F_TLOCK, 0) == -1)
	    exit(0);
#else
	if (flock (lockFd, LOCK_EX|LOCK_NB))
	    exit(0);
#endif

	i = open( "/dev/tty", O_RDWR );
	ioctl(i, TIOCNOTTY, 0 );

	/* Install various signal handlers */

	signal (SIGQUIT, SIG_IGN);
	signal (SIGTSTP, SIG_IGN);
	signal (SIGTERM, term);
	signal (SIGUSR1, set_debug);
	signal (SIGUSR2, reset_debug);

	/* Set up the UDP port for receiving broadcast messages from gateways */

	memset ((char*)&sin, 0, sizeof (struct sockaddr_in));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl (INADDR_ANY);
	sin.sin_port = htons (portid);

	if ((gatewayFd = socket(PF_INET, SOCK_DGRAM, 0)) < 0 ) {
	    tisSysLog (0,0, "Unable to create socket (pid %d)", pid);
	    exit (-1);
	}

	if (setsockopt (gatewayFd, SOL_SOCKET, SO_REUSEADDR, (const char *)&optval, sizeof(int)) < 0 ) {
	    tisSysLog (0,0, "Unable to set socket to SO_REUSEADDR (pid %d)", pid);
	    exit (-1);
	}

	if (bind (gatewayFd, (struct sockaddr*)&sin, sizeof (struct sockaddr_in)) < 0 ) {
	    tisSysLog (0,0, "Unable to bind socket (pid %d)", pid);
	    exit (-1);
	}

	optval = 1;
	if (setsockopt(gatewayFd, SOL_SOCKET, SO_BROADCAST, (const char *)&optval, sizeof(int)) < 0 ) {
	    tisSysLog (0,0, "Unable to set socket to SO_BRAODCAST (pid %d)", pid);
	    exit (-1);
	}



	/* Process messages coming from gateways */

	running = 1;
	while (running) {

	    FD_ZERO (&readfds);
	    FD_SET (gatewayFd, &readfds);

	    if (select (gatewayFd + 1, &readfds, NULL, NULL, NULL) < 0) {
		if( errno == EINTR )
		    continue;
		tisSysLog (0,0, "Error on select (pid %d)", pid);
		exit (-1);
	    }

	    if (FD_ISSET (gatewayFd, &readfds)) {
		if (svc_routechange (gatewayFd) < 0) {
        	    tisSysLog (0,0, "Error changing routes (pid %d)", pid);
		    exit (-1);
		}
	    }

	}

	/* Wrap things up */

	close (gatewayFd);
#ifdef SOLARIS
	lockf(lockFd, F_ULOCK, 0);
#else
	flock (lockFd, LOCK_UN|LOCK_NB);
#endif
	close (lockFd);

	tisSysLog (0,0,  "Terminated (pid 5d)" , pid);

	exit(0);

    }

/*----------*/
static void set_debug ()

    {

	debug = 1;
	tisSysLog (0,0, "Debugging on (pid %d)", pid);

    }

/*----------*/
static void reset_debug ()

    {

	debug = 0;
	tisSysLog (0,0,  "Debugging off (pid %d)", pid);

    }

/*----------*/
static char *dot_addr ( int ip, char *buff)

    {

	sprintf (buff, "%u.%u.%u.%u", (ip>>24)&0xff, (ip>>16)&0xff, (ip>>8)&0xff, ip&0xff );
	return buff;

    }

/*----------*/
static void term (int sig)

    {

	running = 0;

    }

/*----------*/
static int svc_routechange (int fd)

    {

	struct sockaddr_in	from;
	int 			alen = sizeof(struct sockaddr_in);
	struct _linkmsg		stat;
	struct _linkmsg		*currMsg;
	u_long			gatewayAddr, **ipaddr, ipnet;
	struct rtentry		rt;
	struct hostent		*ndb;
	char			msgBuff[1024];
	int			bytesGot;

	/* Check for a message from a gateway */

	bytesGot = recvfrom (fd, msgBuff, sizeof(msgBuff), 0, (struct sockaddr*) &from, &alen);
	if (bytesGot < 0)
	    {
	    return -1;
	    }

    for (currMsg = (struct _linkmsg *) msgBuff; bytesGot > 0; currMsg++, bytesGot -= sizeof(struct _linkmsg))
      {

	/* Grab the current message segment. */

	stat = *currMsg;

	/* Change fields to host byte ordering */

	stat.LinkIPAddr = ntohl (stat.LinkIPAddr);
	stat.SubnetMask = ntohl (stat.SubnetMask);
	gatewayAddr = ntohl (from.sin_addr.s_addr);

	/* Log this message if in debug mode */

	if (debug) {
	    dot_addr (gatewayAddr, dotAddr1);
	    dot_addr (stat.LinkIPAddr, dotAddr2);
	    dot_addr (stat.SubnetMask, dotAddr3);
	    tisSysLog (0,0,  "Message from gateway %s: Net %s:%s (pid %d)",
			dotAddr1, dotAddr2, dotAddr3, pid);
	}
#if 0
dot_addr (gatewayAddr, dotAddr1);
dot_addr (stat.LinkIPAddr, dotAddr2);
dot_addr (stat.SubnetMask, dotAddr3);
printf("\tMessage from gateway %s: Net %s:%s\n", dotAddr1, dotAddr2, dotAddr3);
#endif

	/* Check if an update is needed */

	if (!findNetStatus (stat.LinkIPAddr & stat.SubnetMask, gatewayAddr)) {
	    if (debug)
		tisSysLog (0,0, "Status has not changed (pid %d)", pid);
#if 0
printf("\t\tStatus unchanged.\n");
#endif
	    continue;

	}

	ipnet = (stat.LinkIPAddr & stat.SubnetMask);

	memset (&rt, 0, sizeof (rt));
	rt.rt_flags = RTF_UP | RTF_GATEWAY | RTF_HOST;
	rt.rt_dst.sa_family = AF_INET;
	rt.rt_gateway.sa_family = AF_INET;

	if (debug)
	    tisSysLog (0,0,  "Scanning active tables...(pid %d)", pid);

	/* Build a list of routes to be deleted */

	run_active_tables (ipnet, stat.SubnetMask, gatewayAddr);

	 if (debug)
	    tisSysLog (0,0,  "%d matches found (pid %d)", si/2,pid );

	/* Unconditionlly delete all routes involving the gateway */

	if (1) {
	    while (si) {

		/* Pick up the gateway and the destination off the stack */

		memcpy (&rt.rt_gateway.sa_data [2], &stk [--si], sizeof (u_long));
		memcpy (&rt.rt_dst.sa_data [2], &stk [--si], sizeof (u_long));

		/* If the route does not involve the gateway we got the message from, ignore it */

#if 0
		/* Test by Gordon. I don't think this statement belongs in here.  We want to delete any
		route involving the destination that just reported, regardless of where the old
		gateway was. Take this case: We had a route but it died. A new guy picks up the destination.
		Shouldn't we delete the old route?
		*/
		if (stk [si+1] != from.sin_addr.s_addr)
		    continue;
#endif

		if (debug) {
		    dot_addr (stk[si], dotAddr1);
		    dot_addr (stk[si+1], dotAddr2);
		    tisSysLog (0,0, "Trying to delete %s:%s for gateway/host (pid %d)",
				dotAddr1, dotAddr2,pid);
		}
#if 0
dot_addr (stk[si], dotAddr1);
dot_addr (stk[si+1], dotAddr2);
printf("\t\tTrying to delete %s:%s for gateway/host\n", dotAddr1, dotAddr2);
#endif

		/* Try to delete the route with both gateway and destination host specified. */

		if (ioctl (gatewayFd, SIOCDELRT, &rt) < 0) {

		    /* Didn't delete, try it for just the gateway */

		    rt.rt_flags ^= RTF_HOST;
		    if (debug) {
			dot_addr (stk[si], dotAddr1);
			dot_addr (stk[si+1], dotAddr2);
			tisSysLog (0,0, "Failed, trying to delete %s:%s for gateway only (pid %d)",
					dotAddr1, dotAddr2, pid);
		    }
		    if (ioctl (gatewayFd, SIOCDELRT, &rt) < 0)
			tisSysLog (0,0, "Unable to delete route! (pid %d)", pid);
		    else
			if (debug)
			    tisSysLog (0,0,  "Route deleted (pid %d)", pid);
		} else 
		    if (debug)
			tisSysLog (0,0,  "Route deleted (pid %d)", pid);
	    }

	    /* The status is always up here! */

		sethostent (1);
		rt.rt_flags = RTF_UP | RTF_GATEWAY | RTF_HOST;
		memcpy (&rt.rt_gateway.sa_data [2], &from.sin_addr.s_addr, sizeof (u_long));
		while ((ndb = gethostent()) != 0) {
		    for (ipaddr = (u_long**) ndb->h_addr_list; *ipaddr; ++ipaddr) {
			if (**ipaddr == stat.LinkIPAddr)
			    break;
			if ((**ipaddr & stat.SubnetMask) == ipnet) {
			    if (debug) {
				dot_addr (**ipaddr, dotAddr1);
				dot_addr (from.sin_addr.s_addr, dotAddr2);
				tisSysLog (0,0,  "Adding host=%s(%s), gateway=%s (pid %d)",
					ndb->h_name, dotAddr1, dotAddr2, pid);
			    }
#if 0
dot_addr (**ipaddr, dotAddr1);
dot_addr (from.sin_addr.s_addr, dotAddr2);
printf("\t\tAdding host=%s(%s), gateway=%s\n", ndb->h_name, dotAddr1, dotAddr2);
#endif
			    if (gatewayFd > 0) {
				memcpy (&rt.rt_dst.sa_data [2], *ipaddr, sizeof (u_long));
				if (ioctl (gatewayFd, SIOCADDRT, &rt) < 0)
				    {
				    tisSysLog (0,0, "Route could not be added (pid %d)", pid);
				    }
			    }
			}
		    }
		}
		endhostent();

	}
      }
	return 0;
    }


/*----------*/
static void run_active_tables (u_long net, u_long subnetMask, u_long gatewayAddr)

    {

	FILE		*f;
	char		buf [100], destBuf [100], gatewayBuf [100];
	unsigned long	destination, gateway;

	/* Run the netstat command and get it's result through a file descriptor */

	f = popen ("netstat -nr", "r");

	/* Throw away all lines until we see a line that begins with a numeric digit. */

	do
	    {
	    fgets (buf, sizeof (buf), f);
	    } while (!isdigit(buf[0]));

	/* Get the first two fields (destination and gateway). Place them on the */
	/* stack if they are of interest */

	while (fgets (buf, sizeof (buf), f)) {

	    /* Get destination and gateway */

	    strcpy (destBuf, strtok (buf, " "));
	    strcpy (gatewayBuf, strtok (NULL, " "));
	    if (strcmp (destBuf, "default") == 0)
		destination = 0;
	    else
		destination = makeBin (destBuf);
	    gateway = makeBin (gatewayBuf);

	    if (debug) {
		dot_addr (destination, dotAddr1);
		dot_addr (gateway, dotAddr2);
		tisSysLog (0,0,  "%-16s %-16s (pid %d)",
				dotAddr1, dotAddr2, pid);
	    }

	    if ((destination & subnetMask) == net)  {
		stk [si++] = destination;
		stk[si++] = gateway;
		if (debug) {
		    dot_addr (destination, dotAddr1);
		    dot_addr (gateway, dotAddr2);
		    tisSysLog (0,0,  "Stacking %s %s (pid %d)",
					dotAddr1, dotAddr2, pid);
		}
	    }

	}
	pclose (f);

    }

/*----------*/
static unsigned long makeBin (s)

    char	*s;

    {

	int	a, b, c, d;

	a = atoi (strtok (s, "."));
	b = atoi (strtok (NULL, "."));
	c = atoi (strtok (NULL, "."));
	d = atoi (strtok (NULL, "."));

	return (a << 24) | (b << 16) | (c << 8) | d;

    }

/*----------*/
static int findNetStatus (net, gateway)

    u_long	net, gateway;

    {

	/* Look up a network and compare it's status with the last status. */
	/* If they differ, update the current status in the table and return 1 */
	/* indicating that the routing tables have to be updated. If the status */
	/* has not changed, return 0.  If a network is not in the table, insert */
	/* it and return 1 so the new net will be added to the table. */

	int	i;

	for (i=0 ; i<numNets ; i++)
	    if (netTable [i].net == net)
		break;

	if (i == numNets) {	/* Not found */
	    netTable [numNets].net = net;
	    netTable [numNets].gateway = gateway;
	    netTable [numNets].status = 1;
	    numNets++;
	    return 1;
	} else if (netTable [i].gateway != gateway) {
	    netTable [i].gateway = gateway;
	    return 1;
	} else {
	    return 0;
	}

    }
