/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include "ims.h"
#include "msgMon.h"

FILE  *imsControlFile;
int  caSocket,maxFd;
fd_set  myFds,saveFds;
CONTROL_INFO   imsControl;

int main(int argc,char *argv[])
{
FILE *pidFile;
int status,i;
char cmd;
char controlTableName[MAXPATHLEN];
extern	int	optind; 

imsControl.imsExportPoints = (AVL_TREE *)0;
imsControl.imsImportPoints = (AVL_TREE *)0;

tisOpenLog("ims", 6, 2, NULL, NULL);
#ifdef SOLARIS
sigset(SIGHUP,imsSignalHandler);
sigset(SIGINT,imsSignalHandler);
sigset(SIGUSR1,imsSignalHandler);
sigset(SIGCHLD,imsSignalHandler);
#else
signal(SIGHUP,imsSignalHandler);
signal(SIGINT,imsSignalHandler);
signal(SIGUSR1,imsSignalHandler);
signal(SIGCHLD,imsSignalHandler);
#endif
strcpy(controlTableName,"/tis4000/tasks/imsControlTable");
pidFile = fopen("/tis4000/tasks/ims.pid","w");
fprintf(pidFile,"%d\n",getpid());
fclose(pidFile);
if((cmd = getopt (argc, argv, "c:")) != -1)
  {
    switch (cmd)
      {
      case 'c':
	strcpy(controlTableName,argv[optind-1]);		/* get control file name & path from command line */
	break;
      case '?':
	fprintf(stderr,"Usage is:\n");
	fprintf(stderr,"ims [-c control file path & name]...default:/tis4000/tasks/imsControlTable\n");
	exit(0);
      default:
	break;
      }
  }

if(ca_task_initialize() != ECA_NORMAL)		/* try Channel Access */
  {
    tisSysLog(0,0,"Channel Access is not available");
    exit(1);
  }
if((imsControlFile = fopen(controlTableName,"r")) == 0)	/* try the control file */
  {
    tisSysLog(0,0,"Error opening imsControlTable: %m");
    exit(2);
  }
if(sdrLoad("/tis4000db/default.dctsdr",0) < 0)	/* try the SDR */
  {
    fclose(imsControlFile);
    tisSysLog(0,0,"Error loading SDR tables");
    exit(3);
  }
if(!imsParseControl(&imsControl))		/* parse the control file */
  {
    tisSysLog(0,0,"Error parsing imsControlTable");
    fclose(imsControlFile);
    exit(4);
  }
if(!imsParseExport(&imsControl))
  {
    tisSysLog(0,0,"Error parsing %s",&imsControl.imsExportFileName[0]);
    fclose(imsControlFile);
    exit(5);
  }
fclose(imsControlFile);

while(1)					/* loop for events fromCA */
  {
    sleep(1);
    if(imsControl.imsExportRate != 0)
      {
	if(--imsControl.imsExportRate <= 0)
	  {
	    imsControl.imsExportRate = imsControl.imsRateSave;
	    imsParseExport(&imsControl);
	    avl_traverse(imsControl.imsExportPoints,INORDER,imsWriteExportPoints,(void *)0);
	  }
      }
    imsWriteImportFile(&imsControl);
    status = ca_pend_io(0.01);		/* get data from CA */
    for(i=0;i<100;i++)
      {
	if(status != ECA_TIMEOUT)
	  status = ca_pend_io(0.01);
	else
	  break;
      }
  }
}
/*
  Handle signals
*/
void	imsSignalHandler(int whichone)
{
int pid,retStatus;
switch(whichone)
  {
  case SIGUSR1:
    tisSysLog(0,0,"Async export file read");
    imsParseExport(&imsControl);
    avl_traverse(imsControl.imsExportPoints,INORDER,imsWriteExportPoints,(void *)0);
    break;
  case SIGINT:
  case SIGHUP:
    tisSysLog(0,0,"IMS:exiting due to ^C or -INT\n");
    if(imsControl.imsImportPoints)
      avl_empty(imsControl.imsImportPoints,avl_freenode);
    if(imsControl.imsExportPoints)
      avl_empty(imsControl.imsExportPoints,avl_freenode);
    unlink("/tis4000/tasks/ims.pid");
    ca_task_exit();
    exit(0);
    break;
  case SIGCHLD:
    pid = wait(&retStatus);
    tisSysLog(0,0,"Child process pid = %d exiting with status = %08X\n",pid,retStatus);
    break;
  default:
    tisSysLog(0,0,"\nUnknown signal %d\n",whichone);
  }
}

