/*
** NTUnixSubs.c - Missing Unix functions
**
*/

#include <stdio.h>
#include <string.h>
#include <sys/timeb.h>
#include "NTUnixSubs.h"	/* this is where you find the missing defs */

int strcasecmp (char *s1,char *s2)
{
    while (*s1 && *s2)
    {
	if (toupper(*s1) != toupper(*s2))
	    return *s1 - *s2;
	s1++, s2++;
    }
    return 0;
}

int strncasecmp (char *s1,char *s2, int len)
{
    while (*s1 && *s2 && len)
    {
	if (toupper(*s1) != toupper(*s2))
	    return *s1 - *s2;
	s1++, s2++, len--;
    }
    return 0;
}

char *optarg;
int optopt;
int optind = 1;
int opterr = 1;

int getopt (int argc, char **argv, char *optstring)
{
    char *optptr;

    /* skip past program name */
    if (optind == 0)
	optind++;

    if ((optind < argc) && (*argv[optind] == '-'))
    {
	if (*(argv[optind]+1) == '-')
	{
	    optind++;
	    return -1;
	}
	optptr = strchr (optstring, *(argv[optind]+1));
	if (optptr)
	{
	    optopt = *(argv[optind]+1);
	    if (*(optptr+1) == ':')
	    {
		/* if there is immediate text after option, return address of remainder */
		if (*(argv[optind]+2))
		{
		    optarg = argv[optind]+2;
		}
		/* else if there are additional arguments, return address of next arg */
		else if (optind + 1 < argc)
		{
		    optarg = argv[++optind]; /* and advance index */
		}
		else
		{
		    optarg = NULL;
		}
	    }
	    optind++;
	    return *optptr;
	}
	else
	{
	    if (opterr)
		fprintf (stderr, "Unknown option '%c'\n", *(argv[optind]+1));
	    optind++;
	    return '?';
	}
    }
    else
    {
	return -1;
    }
}

int NTgettimeofday (struct timeval *tv, struct timezone *tz)
{
   struct _timeb timebuffer;

   _ftime( &timebuffer );

   tv->tv_sec = timebuffer.time;
   tv->tv_usec = timebuffer.millitm * 1000;
   if (tz)
   {
       tz->tz_minuteswest = timebuffer.timezone;
       tz->tz_dsttime = timebuffer.dstflag; /* this is not quite right, should be DST type as in Unix */
   }
   return 0;
}

int inet_netof (struct in_addr in)
{
    if (IN_CLASSA(in.s_addr))
	return (in.s_addr & IN_CLASSA_NET);
    else if (IN_CLASSB(in.s_addr))
	return (in.s_addr & IN_CLASSB_NET);
    else if (IN_CLASSC(in.s_addr))
	return (in.s_addr & IN_CLASSC_NET);
    else
	return (in.s_addr);
}

int shmNum = 0;
int shmKey[10];
HANDLE shmHandle[10];
void *shmBase[10];
int shmLen[10];
// Security descriptor to grant access to everyone
SECURITY_ATTRIBUTES SA_ShMem;
PSECURITY_DESCRIPTOR pSD_ShMem;

int shmget (int shm_key, int shm_len, int shmFlag)
{
    int i;

    for (i = 0; i < shmNum; i++)
    {
	if (shmKey[i] == shm_key)
	    break;
    }
    if (i < 10)
    {
	if (i >= shmNum)
	{
	    /* rescan for blanks */
	    for (i = 0; i < shmNum; i++)
	    {
		if (shmKey[i] == 0)
		    break;
	    }
	    if (i >= shmNum)
	    {
		shmNum++;
	    }
	}
	shmKey[i] = shm_key;
	shmLen[i] = shm_len;
	return (0x12340000+i);
    }
    else
    {
	return -1;
    }
}

void *shmat (int shmId, void *loc, int shmFlag)
{
    char shmName [40];
    int i;

    i = shmId - 0x12340000;
    if (i < 0 || i >= shmNum)
    {
	perror ("Shared memory ID doesn't exist!\n");
	return (void *)-1;
    }

    sprintf (shmName, "SHM%d", shmKey[i]);

    if (shmHandle[i] == NULL)
    {
	shmHandle[i] = OpenFileMapping (
			    ((shmFlag & SHM_RDONLY) == 0 ? FILE_MAP_WRITE : FILE_MAP_READ), /* access flags */
			    TRUE,	    /* inherit handle */
			    shmName);  /* name of shared region */
	if (shmHandle[i] == NULL)
	{
	    int shm_error = GetLastError();

	    perror ("Can't open file, creating it\n");
	    //
	    // Create a NULL security descriptor. This allows full access to the
	    // shared memory by user mode apps.
	    //
	    if (pSD_ShMem == NULL)
	    {
		pSD_ShMem = (PSECURITY_DESCRIPTOR)LocalAlloc(LPTR, SECURITY_DESCRIPTOR_MIN_LENGTH);
		if (pSD_ShMem != NULL)
		{
		    if (InitializeSecurityDescriptor(pSD_ShMem, SECURITY_DESCRIPTOR_REVISION)
		      && SetSecurityDescriptorDacl(pSD_ShMem, TRUE, (PACL)NULL, FALSE))
		    {
			SA_ShMem.nLength              = sizeof(SA_ShMem);
			SA_ShMem.lpSecurityDescriptor = pSD_ShMem;
			SA_ShMem.bInheritHandle       = TRUE;
		    }
		    else
		    {
			// Do something on error...
			LocalFree (pSD_ShMem);
			pSD_ShMem = NULL;
		    }
		}
	    }
	    shmHandle[i] = CreateFileMapping ((HANDLE)0xFFFFFFFF, /* paging space file only */
    					(pSD_ShMem)?&SA_ShMem: NULL, /* No security */
					((shmFlag & SHM_RDONLY) == 0 ? PAGE_READWRITE : PAGE_READONLY), /* access flags */
					0,	    /* high order share size */
					shmLen[i],    /* low order share size */
					shmName);  /* name of shared region */
	}
    }

    if (shmHandle[i])
    {
	if (loc == NULL)
	{
	    shmBase[i] = MapViewOfFile (shmHandle[i],    /* handle of file mapping */
					((shmFlag & SHM_RDONLY) == 0 ? FILE_MAP_ALL_ACCESS : FILE_MAP_READ), /* access flags */
					0,  0,	    /* file offset */
					shmLen[i]);   /* number of bytes to map */
	}
	else
	{
	    shmBase[i] = MapViewOfFileEx (shmHandle[i],    /* handle of file mapping */
				    ((shmFlag & SHM_RDONLY) == 0 ? FILE_MAP_ALL_ACCESS : FILE_MAP_READ), /* access flags */
				    0,  0,	/* file offset */
				    shmLen[i],	/* number of bytes to map */
				    (char *)loc);	/* desired mapping address */
	}
	if (shmBase[i])
	{
	    return shmBase[i];
	}
	else
	{
	    int shm_error = GetLastError();
	    perror ("Can't map to new file\n");
	    CloseHandle(shmHandle[i]);
	    return (void *)-1;
	}
    }
    else
    {
	int shm_error = GetLastError();
	perror ("Can't open or create shared file\n");
	return (void *)-1;
    }
}

int shmdt (void *shm_addr)
{
    int i;

    for (i = 0; i < shmNum; i++)
    {
	if (shm_addr == shmBase[i])
	    break;
    }

    if (i >= shmNum)
    {
	perror ("Shared memory location doesn't exist!\n");
	return -1;
    }
    else
    {

	UnmapViewOfFile(shmBase[i]);
	CloseHandle (shmHandle[i]);
	shmBase[i] = NULL;
	return 0;
    }
}

int shmctl (int shm_key, int flags, void *shm_buf)
{
    /* not a valid call on NT, just ignore */
    return 0;
}

/* tis4000 semaphore mechanisms */
int semNum = 0;
int semKey[10];
HANDLE semHandle[10];

int sem_create (int sem_key, int initval)
{
    char semName [40];
    int i;

    for (i = 0; i < semNum; i++)
    {
	if (semKey[i] == sem_key)
	{
	    perror ("semaphore already exists\n");
	    return -1;
	}
    }
    for (i = 0; i < 10; i++)
    {
	if (semKey[i] == 0)
	{
	    break;
	}
    }
    if (i < 10)
    {
	if (i >= semNum)
	    semNum = i + 1;
	sprintf (semName, "SEM%d", sem_key);
	semKey[i] = sem_key;
	if (pSD_ShMem == NULL)
	{
	    pSD_ShMem = (PSECURITY_DESCRIPTOR)LocalAlloc(LPTR, SECURITY_DESCRIPTOR_MIN_LENGTH);
	    if (pSD_ShMem != NULL)
	    {
		if (InitializeSecurityDescriptor(pSD_ShMem, SECURITY_DESCRIPTOR_REVISION)
		  && SetSecurityDescriptorDacl(pSD_ShMem, TRUE, (PACL)NULL, FALSE))
		{
		    SA_ShMem.nLength              = sizeof(SA_ShMem);
		    SA_ShMem.lpSecurityDescriptor = pSD_ShMem;
		    SA_ShMem.bInheritHandle       = TRUE;
		}
		else
		{
		    // Do something on error...
		    LocalFree (pSD_ShMem);
		    pSD_ShMem = NULL;
		}
	    }
	}
	semHandle[i] = CreateSemaphore(
				(pSD_ShMem)?&SA_ShMem: NULL,// pointer to security attributes 
				initval,	// initial count 
				1000,		// maximum count 
				semName 	// pointer to semaphore-object name  
			       );
	if (semHandle[i])
	{
	    return 0x12340000 + i;
	}
	else
	{
	    perror ("semaphore space exhausted\n");
	    return -1;
	}
    }
    else
	return -1;
}

int sem_open (int sem_key)
{
    char semName [40];
    int i;

    for (i = 0; i < semNum; i++)
    {
	if (semKey[i] == sem_key)
	{
	    perror ("semaphore already exists\n");
	    return -1;
	}
    }
    for (i = 0; i < 10; i++)
    {
	if (semKey[i] == 0)
	{
	    break;
	}
    }
    if (i < 10)
    {
	if (i >= semNum)
	    semNum = i + 1;
	sprintf (semName, "SEM%d", sem_key);
	semKey[i] = sem_key;
	semHandle[i] = OpenSemaphore(
				SEMAPHORE_ALL_ACCESS,	// access flag 
				FALSE,		// inherit flag 
				semName 	// pointer to semaphore-object name
			       );
	if (semHandle[i])
	{
	    return 0x12340000 + i;
	}
	else
	{
	    perror ("semaphore space exhausted\n");
	    return -1;
	}
    }
    else
	return -1;
}

int sem_rm (int id)
{
    int i;

    i = id - 0x12340000;
    if (i < 0 || i >= semNum)
    {
	perror ("Semaphore ID doesn't exist!\n");
	return -1;
    }
    CloseHandle (semHandle[i]);
    semKey[i] = 0;
    return 0;
}

int sem_close (int id)
{
    int i;

    i = id - 0x12340000;
    if (i < 0 || i >= semNum)
    {
	perror ("Semaphore ID doesn't exist!\n");
	return -1;
    }
    CloseHandle (semHandle[i]);
    semKey[i] = 0;
    return 0;
}

int sem_signal (int id)
{
    int i;

    i = id - 0x12340000;
    if (i < 0 || i >= semNum)
    {
	perror ("Semaphore ID doesn't exist!\n");
	return -1;
    }

    if (ReleaseSemaphore (semHandle[i], 1, NULL))
    {
	/* success */
	return 0;
    }
    else
    {
	/* failed */
	i = GetLastError ();
	return -1;
    }
}

int sem_wait (int id)
{
    int i;

    i = id - 0x12340000;
    if (i < 0 || i >= semNum)
    {
	perror ("Semaphore ID doesn't exist!\n");
	return -1;
    }

    if (WaitForSingleObject (semHandle[i], INFINITE) == WAIT_FAILED)
    {
	/* failed */
	i = GetLastError ();
	return -1;
    }
    else
    {
	/* succeeded */
	return 0;
    }
}
