/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
**	pdfHelpUtils.c - alarm object display formats
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <pdfHelpUtils.h>

#ifndef TRUE
#define TRUE (1==1)
#endif
#ifndef FALSE
#define FALSE (1!=1)
#endif

#ifdef	WIN32
static char *helpDefFile = {"\\tis4000\\documents/helpDefs.def"};
static char *pdfDir = {"\\tis4000\\documents"};
#else
static char *helpDefFile = {"/tis4000/documents/helpDefs.def"};
static char *pdfDir = {"/tis4000/documents"};
#endif	WIN32

static char *pdfReader = {"acroread"};
static char inpstring [MAXINP];
static char cmdline [MAXINP];

static char *getElement (char **ptr);

#ifdef DEBUG_TEST
main ()
{
    int i;
    char **helpstrings;

    helpstrings = getHelpItems ("alarmvu", "summary");
    if (helpstrings)
    {
	i = 0;
	while (helpstrings[i])
	{
	    printf ("helpstrings %d = %s\n", i, helpstrings[i]);
	    selHelpItem ("alarmvu", "summary", helpstrings[i]);
	    i++;
	}
	freeHelpItems (helpstrings);
    }
    printf ("finished\n");
}
#endif DEBUG_TEST

/*
**	getHelpItems (char *AppName, char *ScreenName) - open help spec file,
**	consisting of text in the format:
**		>AppName ScreenName
**		   HelpLabel PDFFile
**	where AppName is the Application name, ScreenName is the Screen whos help menu
**	is being defined,
**	      HelpLabel is the text that will appear on the help menu,
**	      and PDFFile is the help display to bring up.
**	The HelpLabel must be entered in quotes if spaces are desired.
**
**	for example,
**		>alarmvu summary
**		    "Display Info" alSumDisp.pdf
**		    "Commands" alSumCmds.pdf
**
**	RETURNS:
**		an array of string pointers to the help items for the screen
*/
char ** getHelpItems (char *AppName, char *ScreenName)
{
    int ix;
    char *sp;
    FILE *fp;
    char *cp1;
    char *cp2;
    int returnStringIx;
    char **returnStringPtr = NULL;

    fp = fopen (helpDefFile, "rt");
    if (fp)
    {
        while (!feof (fp))
        {
            /* get a line of text from input */
	    ix = 0;
            while (!feof (fp))
            {
                inpstring [ix] = fgetc(fp);
                if ((inpstring [ix] == 0x0d) || (inpstring [ix] == 0x0a) 
		  || (inpstring [ix] == -1))
		{
                    break;
                }
		ix++;
                if (ix == (MAXINP - 1))
		{
		    if (returnStringPtr)
			freeHelpItems (returnStringPtr);
		    printf ("Line too long in Help spec file \n");
                    return NULL; /* bad string in file. */
		}
            }
	    inpstring [ix] = 0;

	    /* process input line */
	    sp = inpstring;
	    while (*sp && (*sp == ' ' || *sp == 0x09))
		sp++;
	    if ((*sp == 0) || (*sp == '#'))
		continue; /* blank line or comment! */

	    /* Get first item on line, either app name or help text */
	    if (*sp == '>')
	    {
		/* if we already parsed a name, a new name means quit */
		if (returnStringPtr)
		    break;
		sp++;
		cp1 = getElement (&sp);
		cp2 = getElement (&sp);
		if (cp1 && (strcmp (cp1, AppName) == 0)
		  && cp2 && (strcmp (cp2, ScreenName) == 0))
		{
		    returnStringPtr = calloc (MAXHELPITEMS, sizeof(char *));
		    returnStringIx = 0;
		}
	    }
	    else if (returnStringPtr) /* line must be a help spec line */
	    {
		cp1 = getElement (&sp);
		returnStringPtr [returnStringIx] = malloc (strlen (cp1) + 1);
		if (returnStringPtr [returnStringIx])
		    strcpy (returnStringPtr [returnStringIx], cp1);
		else /* out of memory! */
		    return (returnStringPtr);
		returnStringIx++;
	    }
	}
        fclose (fp);
    } /* end of opened file code */
    return returnStringPtr;
}

void selHelpItem (char *AppName, char *ScreenName, char *HelpItem)
{
    int ix;
    char *sp;
    FILE *fp;
    char *cp1;
    char *cp2;
    int inHelpSect = FALSE;

    fp = fopen (helpDefFile, "rt");
    if (fp)
    {
        while (!feof (fp))
        {
            /* get a line of text from input */
	    ix = 0;
            while (!feof (fp))
            {
                inpstring [ix] = fgetc(fp);
                if ((inpstring [ix] == 0x0d) || (inpstring [ix] == 0x0a) 
		  || (inpstring [ix] == -1))
		{
                    break;
                }
		ix++;
                if (ix == MAXINP - 1)
		{
		    printf ("Line too long in Help spec file \n");
                    return; /* bad string in file. */
		}
            }
	    inpstring [ix] = 0;

	    /* process input line */
	    sp = inpstring;
	    while (*sp && (*sp == ' ' || *sp == 0x09))
		sp++;
	    if ((*sp == 0) || (*sp == '#'))
		continue; /* blank line or comment! */

	    /* Get first item on line, either app name or help text */
	    if (*sp == '>')
	    {
		sp++;
		cp1 = getElement (&sp);
		cp2 = getElement (&sp);
		if (cp1 && (strcmp (cp1, AppName) == 0)
		  && cp2 && (strcmp (cp2, ScreenName) == 0))
		{
		    inHelpSect = TRUE;
		}
		else
		{
		    if (inHelpSect)
		    {
			/* not found! */
			printf ("Help item not found\n");
			fclose (fp);
			return;
		    }
		}
	    }
	    else if (inHelpSect) /* line must be a help spec line */
	    {
		cp1 = getElement (&sp);
		if (strcmp (cp1, HelpItem) == 0)
		{
		    cp2 = getElement(&sp);
		    if (cp2)
		    {
			fclose (fp);
			sprintf (cmdline, "cd %s; %s %s &\n", pdfDir, pdfReader, cp2);
#ifdef DEBUG_TEST
			printf (cmdline);
#endif DEBUG_TEST
			system (cmdline);
			return;
		    }
		}
	    }
	}
        fclose (fp);
    } /* end of opened file code */
}

void freeHelpItems (char **helpStringPtr)
{
    int i;

    for (i = 0; i < MAXHELPITEMS; i++)
	if (helpStringPtr[i])
	    free (helpStringPtr[i]);
    free (helpStringPtr);
}

static char *getElement (char **ptr)
{
    char * retp = NULL;

    /* skip white space */
    while (*ptr && (**ptr == ' ' || **ptr == 0x09))
	(*ptr)++;
    if (**ptr == '"')
    {
	(*ptr)++;
	retp = *ptr;
	while (**ptr && **ptr != '"')
	    (*ptr)++;
	if (**ptr)
	{
	    **ptr = 0;
	    (*ptr)++;
	}
    }
    else
    {
	retp = *ptr;
	while (**ptr && (**ptr != ':')
	  && (**ptr != ' ') && (**ptr != 0x09))
	    (*ptr)++;
	if (**ptr)
	{
	    **ptr = 0;
	    (*ptr)++;
	}
    }
    return retp;
}
