/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	COMMON INTIALIZATION FOR CHANNEL ACCESS SERVERS		*/
/*								*/
/*	Many of the things that we do for the three servers	*/
/*	that make up the agent architecture for channel access	*/
/*	are common. This routine avoids code duplication.	*/
/*								*/
/*	Author:	Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/file.h>
#include <sys/types.h>
#include <termios.h>
#include <signal.h>
#include <string.h>

#include "msmain.h"
#include "msgMon.h"

void serverInit(appName, dirPath, lockFile)
char	*appName;
char	*dirPath;
char	*lockFile;
{
    int		fd;
    char	buff[10];

/******
	Change our controlling TTY so that we don't pick up any
	stray signals.
******/

#ifndef SOLARIS
    fd = open("/dev/tty", O_RDWR);
    ioctl(fd, TIOCNOTTY, 0);
    close(fd);
#endif

/******
	Move to the defined directory so that the core file
	can be found in case we die an unnatural death.
******/

    if (chdir(dirPath) != 0) {
	tisSysLog (0,0, "%s(%d): Unable to find %s lock directory (%s)",
			appName, getpid (), appName, dirPath);
	if (mkdir (dirPath, 0775) == 0) {
	    tisSysLog (0,0, "%s(%d): Lock directory %s created.",
			appName, getpid (), dirPath);
	    chdir (dirPath);
	} else {
	    tisSysLog (0,0, "%s(%d): Lock directory %s could not be created.",
			appName, getpid (), dirPath);
	    exit (-1);
	}
    }

/******
	Make a process group just for ourselves
******/

#ifdef AXP
    tcsetpgrp (0, getpid ());
#else
#ifdef SOLARIS
    setpgrp ();
#else
    setpgrp(0, getpid());
#endif
#endif

/******
	Record our PID in the lock file to make it easy to find
******/

    sprintf(buff, "%d\n", getpid());
    write(fd, buff, strlen(buff));
    close(fd);		/* I think that we keep the lock even if
			   the file is closed. */

}

