/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	APPLICATIONS INTERFACE TO TIS 4000 LOG FACILITY		*/
/*								*/
/*	These routines provide the hooks for an application	*/
/*	to use the TIS 4000 logging facility.			*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <stdio.h>
#include <signal.h>
#ifdef WIN32
#else
#include <syslog.h>
#endif WIN32
#include <sys/types.h>
#include <sys/stat.h>
#if defined(AXP) || defined(WIN32)
#include <string.h>
#endif AXP or WIN32

#ifdef	WIN32
/* Can't use header file! func prototype uses vararg (so you don't need all 10 params),
**   but routine is defined for 10 parameters! */
#include "msgMon.h"
#endif	WIN32

static int	currLogLevel;
static int	logCeiling;
static void	*oldUsr1;
static void	*oldUsr2;
static void	incrLevel();
static void	decrLevel();
static char	progName[64];
static int	logCrossLevel;
static FILE	*logFd;

#ifdef WIN32
#include <windows.h>
HANDLE eventH;

#define LOG_ERR		EVENTLOG_ERROR_TYPE
#define LOG_LOCAL1	0
#endif WIN32

/****************************************************************/
/*								*/
/*		CONNECT TO THE MESSAGE SYSTEM			*/
/*								*/
/*	This routine also establishes the signal handlers used	*/
/*	to increment and decrement the current debug level.	*/
/*	If the user provides signal handlers, then they will	*/
/*	be used. If he doesn't, then default ones will be	*/
/*	provided.						*/
/*								*/
/****************************************************************/

int tisOpenLog(char *ident, int ceiling, int crossOver,
	       void (*sigUsr1)(), void (*sigUsr2)())
{
    logCeiling = ceiling;
    logCrossLevel = crossOver;
    strcpy(progName, ident);
#ifdef WIN32
#else
    if (!sigUsr1)
	sigUsr1 = incrLevel;
    if (!sigUsr2)
	sigUsr2 = decrLevel;
    oldUsr1 = signal(SIGUSR1, sigUsr1);
    oldUsr2 = signal(SIGUSR2, sigUsr2);
#endif WIN32
    tisSysLog(0, 0, "Program %s started", ident);
    return (0);
}

/****************************************************************/
/*								*/
/*		DEFAULT HANDLERS FOR USR1 & USR2 SIGNALS	*/
/*								*/
/****************************************************************/

static void incrLevel()
{
    tisBumpLogLevel(1);
}

static void decrLevel()
{
    tisBumpLogLevel(-1);
}


/****************************************************************/
/*								*/
/*			LOG A MESSAGE				*/
/*								*/
/*	Maximum frequence feature not implemented at the	*/
/*	current time.						*/
/*								*/
/****************************************************************/

int tisSysLog(level, maxFreq, message, par1, par2, par3, par4, 
	  par5, par6, par7, par8, par9, par10)
int	level;
int	maxFreq;
char	*message;
#ifdef	AXP
void*	par1;
void*	par2;
void*	par3;
void*	par4;
void*	par5;
void*	par6;
void*	par7;
void*	par8;
void*	par9;
void*	par10;
#else
int	par1;
int	par2;
int	par3;
int	par4;
int	par5;
int	par6;
int	par7;
int	par8;
int	par9;
int	par10;
#endif AXP
{

    if (level > currLogLevel)
	return (0);

    if (logFd)
    {
	fprintf(logFd, message, par1, par2, par3, par4, par5, par6, 
		par7, par8, par9, par10);
	fprintf(logFd, "\n");
	return (0);
    }
#ifdef WIN32
    {
	char reportMsg [300];
	char	* addrReportMsg[1] = {reportMsg};

	eventH = RegisterEventSource (NULL, progName);

	sprintf (reportMsg, message, par1, par2, par3, par4, par5, par6, 
		par7, par8, par9, par10);

        ReportEvent (eventH,	/* handle of event source */
            EVENTLOG_ERROR_TYPE,/* event type */
            1,                  /* event category */
            0,                  /* event ID */
            NULL,               /* current user's SID */
            1,                  /* strings in lpszStrings */
            0,                  /* no bytes of raw data */
            addrReportMsg,      /* array of error strings */
            NULL);              /* no raw data */

            DeregisterEventSource (eventH);
    }
#else
    if (progName[0] != 0 )
    {
	char msgBuff[512];
	sprintf (msgBuff, "%s: %s", progName, message);
	syslog(LOG_ALERT | LOG_LOCAL1, msgBuff, par1, par2, par3, par4,
	   par5, par6, par7, par8, par9, par10);
    }
    else
    {
	syslog(LOG_ALERT | LOG_LOCAL1, message, par1, par2, par3, par4,
	   par5, par6, par7, par8, par9, par10);
    }
#endif WIN32
    return (0);
}

/****************************************************************/
/*								*/
/*	ROUTINES TO FETCH CURR LEVEL AND TO BUMP IT		*/
/*								*/
/****************************************************************/

int tisGetLogLevel()
{
    return (currLogLevel);
}


int tisBumpLogLevel(int incr)
{
    currLogLevel += incr;
    if (currLogLevel > logCeiling)
	{
	currLogLevel = logCeiling;
	return (currLogLevel);
	}
    else if (currLogLevel < 0)
	{
	currLogLevel = 0;
	return (currLogLevel);
	}
#ifdef WIN32
#else
    syslog(LOG_ALERT | LOG_LOCAL1, "Debug level for %s changed to %d", progName, currLogLevel);
#endif WIN32

    /*
    See if it is time to either create or close a file.
    */

    if (incr > 0)
	{
	if (currLogLevel == logCrossLevel)
	    {
	    char	fileName1[100];
	    char	fileName2[100];
#ifdef WIN32
	    struct _stat	statBuff;
	    strcpy(fileName1, "\\tis4000db\\");
#else
	    struct stat	statBuff;
	    strcpy(fileName1, "/tis4000db/");
#endif WIN32

	    strcat(fileName1, progName);
	    strcat(fileName1, ".log.1");
#ifdef WIN32
	    if (_stat(fileName1, &statBuff) == 0)
#else
	    if (stat(fileName1, &statBuff) == 0)
#endif WIN32
		{
#ifdef WIN32
	    	strcpy(fileName2, "\\tis4000db\\");
#else
	    	strcpy(fileName2, "/tis4000db/");
#endif WIN32
	    	strcat(fileName2, progName);
	    	strcat(fileName2, ".log.2");
		rename(fileName1, fileName2);
		}
	    strcpy(fileName2, fileName1);
#ifdef WIN32
	    strcpy(fileName1, "\\tis4000db\\");
#else
	    strcpy(fileName1, "/tis4000db/");
#endif WIN32
	    strcat(fileName1, progName);
	    strcat(fileName1, ".log.0");
#ifdef WIN32
	    if (_stat(fileName1, &statBuff) == 0)
#else
	    if (stat(fileName1, &statBuff) == 0)
#endif WIN32
#ifdef WIN32
	    if (_stat(fileName1, &statBuff) == 0)
#else
	    if (stat(fileName1, &statBuff) == 0)
#endif WIN32
		rename(fileName1, fileName2);
	    strcpy(fileName2, fileName1);
#ifdef WIN32
	    strcpy(fileName1, "\\tis4000db\\");
#else
	    strcpy(fileName1, "/tis4000db/");
#endif WIN32
	    strcat(fileName1, progName);
	    strcat(fileName1, ".log");
#ifdef WIN32
	    if (_stat(fileName1, &statBuff) == 0)
#else
	    if (stat(fileName1, &statBuff) == 0)
#endif WIN32
		rename(fileName1, fileName2);
	    logFd = fopen(fileName1, "w");
	    if (logFd == NULL)
		{
#ifdef WIN32
#else
		syslog(LOG_ALERT | LOG_LOCAL1, "Can't create %s: %m", fileName1);
#endif WIN32
		return (0);
		}
#ifdef WIN32
#else
	    syslog(LOG_ALERT | LOG_LOCAL1, "Program %s now logging to %s", progName, fileName1);
#endif WIN32
	    }
	}
    else if (incr < 0)
	{
	if (currLogLevel < logCrossLevel)
	    {
	    if (logFd)
		{
		fclose(logFd);
		logFd = 0;
		}
	    }
	}

    return (currLogLevel);
}


/****************************************************************/
/*								*/
/*		DISCONNECT FROM THE MESSAGING SYSTEM		*/
/*								*/
/*	Also, restore the USR1 and USR2 signals to their	*/
/*	previous values.					*/
/*								*/
/****************************************************************/

void tisCloseLog()
{
    tisSysLog(0, 0, "Program %s disconnected from messaging facility", progName);
#ifdef WIN32
#else
    signal(SIGUSR1, oldUsr1);
    signal(SIGUSR2, oldUsr2);
#endif WIN32
}
