/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#ifdef WIN32
#include <process.h>
#include "NTUnixSubs.h"
#endif WIN32

#include "mlist.h"
#include "trendDefs.h"
#include "arSnapRecord.h"
#include "ar_schedule.h"
#include "dbase.h"
#include "utils.h"
#include "dbRequests.h"
#include "rb.h"
#ifdef ORACLESNAP
#include "oracleSnap.h"
#include "oracleSnapLib.h"
#endif ORACLESNAP
#include "utils.h"
#include "msgMon.h"
#include "resolveLinks.h"

#define STATISTICS

/* Functions defined in this module */

	void	readArchiveConfig (void);
	void	archiveAndClean (int);
#ifndef	ORACLESNAP
static	void	copyFields (void);
static	void	makeArNotesKey (char *, int, int, int, int, int, char *);
#endif	!ORACLESNAP
static	void	archiveAndCleanNotes (void);
static  void    buildPointList (void);
static  void    freePointList (void);
static  int     inPointList (char *);
static  int     searchCompare (const void *, const void *);
#ifdef ORACLESNAP
static int ReduceSnapFile (OracleDesc *desc, char *pointName, int method, REDUC_SCHEDULE *schedule);
#endif ORACLESNAP

/* External functions */

#ifndef	STANDALONE_ARCHIVE
extern  void    doWriteRequests (void);
#endif	STANDALONE_ARCHIVE

/* Archiving and cleaning information */

char	archiveFileName [100];			/* Name of the file to place archive data into */
int	runAtHour;				/* Hour at which to run archiver */
int	archiveOlderThan;			/* Archive records older than this many days */
int	cleanOlderThan;				/* Clean records older than this many days */
int     archiveEnabled, cleanEnabled;
time_t  archiveTime, cleanTime, timeNow;
#ifdef AVM
int     targetSize;				/* no new archive data when file reaches this size */
int     maxDuration;				/* life of archive file in days */
int     diskSpace;				/* no archive file created if disk space is less than this */
int     numDirectories;				/* number of directories in list */
char    **dirList;				/* array of pointers to directory strings */ 
#endif AVM

/* List of points to be archived */

static  PointMethName   *pointList;
static  int    numPoints;

/* external data */

extern	int		debugLevel;
extern	int		currentHour;		/* The current hour */
extern	int		archiveInProgress;
extern	int		running;
extern	int		dbOpen;
#ifndef	STANDALONE_ARCHIVE
extern	RingBuf		*requestQueue;
#endif	STANDALONE_ARCHIVE
extern  int             myPid;

#ifdef ORACLESNAP
OracleDesc		*oTrendDesc;	/* Oracle trend data descriptor */
OracleDesc		*oArchDesc;	/* Oracle Archive data descriptor */
SnapRec			snapRec;	/* Oracle interface data record */
SnapRec			newRec;		/* Oracle interface data record */
#else
extern	CODE4		cb;
extern	FIELD4INFO	snapFields [];
extern	TAG4INFO	snapTags [];
extern	FIELD4INFO	notesFields [];
extern	TAG4INFO	notesTags [];

extern	DATA4	*arSnapDesc;
extern	FIELD4	*pointNameFld, *eguTagFld, *dispLowFld, *dispHighFld, *yearFld, *dayOfYearFld, *hourFld;
extern	FIELD4	*offsetMinFld, *sampleCntFld, *redMethodFld, *rateFld, *valueFld [BUCKETS_PER_RECORD], *statusFld [BUCKETS_PER_RECORD]; 

	DATA4	*archive;
	FIELD4	*archivePointNameFld, *archiveEguTagFld, *archiveDispLowFld, *archiveDispHighFld;
	FIELD4	*archiveYearFld, *archiveDayOfYearFld, *archiveHourFld;
	FIELD4	*archiveOffsetMinFld, *archiveSampleCntFld, *archiveRedMethodFld, *archiveRateFld;
	FIELD4	*archiveValueFld [BUCKETS_PER_RECORD], *archiveStatusFld [BUCKETS_PER_RECORD];  

extern	DATA4	*arNotesDesc;
extern	FIELD4	*notesNameFld, *notesYearFld, *notesDayOfYearFld, *notesHourFld;
extern	FIELD4	*notesMinFld, *notesSecFld, *notesNoteFld;
#endif ORACLESNAP

extern int errno;

/*----------*/
void readArchiveConfig (void)

    {

	FILE		*f;
	char		str [100];
#ifdef AVM
	int   num;
#endif AVM

	if (debugLevel >= 1)
	    tisSysLog (0, 2, "Reading archive configuration file.\n");

#ifdef AVM
#ifdef WIN32
	f = fopen ("\\tis4000db\\archconfig.txt", "r");
#else
	f = fopen ("/tis4000db/archconfig.txt", "r");
#endif WIN32
	if (f == NULL)
	  {
	    tisSysLog (0, 2, "Unable to find %s -- No archiving or cleaning will be performed. Error=%d\n",
		    "/tis4000db/archconfig.txt", errno);
	    archiveEnabled = 0;
	    cleanEnabled = 0;
	    return;
	  }

	fgets (str, 100, f);
	runAtHour = atoi (str);

	fgets (str, 100, f);
	targetSize = atoi (str);

	fgets (str, 100, f);
	maxDuration = atoi (str);

	fgets (str, 100, f);
	diskSpace = atoi (str);

	fgets (str, 100, f);
	cleanOlderThan = atoi (str);

	fgets (str, 100, f);
	archiveOlderThan = atoi (str);

	fgets (str, 100, f);
	archiveEnabled = (strncasecmp (str, "enable", 6) == 0);

	fgets (str, 100, f);
	cleanEnabled = (strncasecmp (str, "enable", 6) == 0);

	fgets (str, 100, f);
	num = atoi (str);
	numDirectories = num;
	if(num)
	  {
	    dirList = (char **)calloc(numDirectories,sizeof(char *));
	    while(--num >= 0)
	      {
		fgets(str,100,f);
		if (strlen(str))
		    str[strlen(str)-1] = 0;
		dirList[num] = calloc(1,strlen(str)+1);
		strcpy(dirList[num],str);
	      }
	  }

	fclose (f);
#else
	f = fopen (ARCHIVE_INFO_FILE, "r");
	if (f == NULL) {
	    tisSysLog (0, 2, "Unable to find %s -- No archiving or cleaning will be performed.\n",
			ARCHIVE_INFO_FILE);
	    archiveEnabled = 0;
	    cleanEnabled = 0;
	    return;
	}

	fgets (archiveFileName, 100, f);
	if (strlen (archiveFileName))
	    archiveFileName [strlen (archiveFileName) - 1] = '\0';

	fgets (str, 100, f);
	archiveOlderThan = atoi (str);

	fgets (str, 100, f);
	cleanOlderThan = atoi (str);

	fgets (str, 100, f);
	runAtHour = atoi (str);

	fgets (str, 100, f);
	archiveEnabled = (strncasecmp (str, "enable", 6) == 0);

	fgets (str, 100, f);
	cleanEnabled = (strncasecmp (str, "enable", 6) == 0);

	fclose (f);

#endif AVM
    }

/*----------*/
void archiveAndClean (int currentHour)


    {

	time_t	   clock;	
	int	   stat;
	PointName  pointName;
	int        method;
	int	   interval;
#ifdef ORACLESNAP
	time_t	   currentDay;
	struct tm *tmblock;
	int        archive;
	CreateParams createParams;
	REDUC_SCHEDULE schedule;	/* place for current schedule info */
#else
	time_t     recTime;
	int	   year, dayOfYear, hour, min, stat2, i;
	char	   buf [100], key [100], archPath [200], pathName [200];
	long	   numRecs, cnt;
	PointMethName  pointMethName;
	TAG4	   *tag;
#endif ORACLESNAP

	if (!archiveEnabled && !cleanEnabled)
	    return;

	/* Check if it is time to do an archive and clean operation */

	if (currentHour != runAtHour)
	    return;

	/* Get the start time of the current day so that we can compare to it */
	timeNow = prevMidnite (time (0));

#ifdef ORACLESNAP
	dbOpen = TRUE;

	/* reduce data where required. */
	/* for each point in ar_reduc file */
	if (openArReduc () == 0)
	{
	    tisSysLog (0, 2, "Unable to open ar_reduc.dbf for reduce processing.\n");
	    return;
	}

	/* GetArReducSchedule skips any records that do not have a text rate field */
	while (getArReducSchedule (pointName, &method, &schedule, &archive)) {
	    /* reduce trend_data area */
	    oTrendDesc = MakeSnapDesc ("TISTREND", "TREND_DATA");
	    if (!OpenSnapTable (oTrendDesc, FALSE)) {
		tisSysLog (0, 2, "Unable to open ar_snap or ar_last.\n");
		free (oTrendDesc);
		exit (-1);
	    }
	    ReduceSnapFile (oTrendDesc, pointName, method, &schedule);
	    free (oTrendDesc);
	    /* TODO: reduce each online archive data area (from catlib) only if archive flag set */
	}
	closeArReduc ();
	/* If both archiving and cleaning are disabled then we can leave now. */
#endif ORACLESNAP

	if (debugLevel >= 1) {
	    clock = time (0);
#ifdef WIN32
	    tisSysLog (0, 2, "Archiver activated at %s.\n", ctime (&clock));
#else
	    tisSysLog (0, 2, "Archiver activated at %s.\n", ctime (&clock));
#endif WIN32
	}
	archiveInProgress = TRUE;

	/* Check if the archive file exists. If it does not then create it. */

#ifdef ORACLESNAP
	if (archiveEnabled) {
	    createParams.targetSize = targetSize;
	    createParams.diskSpace = diskSpace;
	    createParams.numDirectories = numDirectories;
	    createParams.dirList = dirList;

	    oArchDesc = MakeSnapDesc ("TISARCHIVE", "ARCHIVE_DATA");
	    if (!OpenSnapTable (oArchDesc, &createParams)) {
		tisSysLog (0, 2, "Unable to open/create archive table.\n");
		return;
	    }

	    /* Build a table of all the points to be archived. */

	    buildPointList ();

	}

	if (debugLevel >= 1)
	    tisSysLog (0, 2, "Time now is %s\n", ctime (&timeNow));

	/* Get the archive and clean threshold times */

	if (archiveEnabled) {
	    archiveTime = timeNow - (long)(archiveOlderThan * SECONDS_PER_DAY);
	    if (debugLevel >= 1)
		tisSysLog (0, 2, "Records older than %s will be archived.\n",
						ctime (&archiveTime));
	}

	if (cleanEnabled) {
	    cleanTime = timeNow - (long)(cleanOlderThan * SECONDS_PER_DAY);
	    if (debugLevel >= 1)
	    	tisSysLog (0, 2, "Records older than %s will be cleaned.\n",
						ctime (&cleanTime));
	}

	/* Open the ar_snap file. It is automatically positioned at the beginning */

	oTrendDesc = MakeSnapDesc ("TISTREND", "TREND_DATA");
	if (!OpenSnapTable (oTrendDesc, FALSE)) {
	    tisSysLog (0, 2, "Unable to open ar_snap or ar_last.\n");
	    free (oTrendDesc);
	    free (oArchDesc);
	    exit (-1);
	}

	/* Scan through allowable ar_snap records and perform archiving
	** and cleaning operations. Fetch records less than archive time
	*/

	/* get eldest date in trend_data, start archiving records stamped before the next day */
	stat = GetSnapRecsByTime (oTrendDesc, 0, archiveTime, &snapRec);
        if (stat)
	{   
	    tmblock = gmtime ((time_t *)&snapRec.rec_time);
	    currentDay = snapRec.rec_time - (tmblock->tm_hour*3600 + tmblock->tm_min*60 + tmblock->tm_sec)
			 + 86400 /* 1 day in seconds */;

	    /* Transfer each day of data for each point into the archive */
	    while (currentDay <= archiveTime) {
		if (openArReduc () == 0)
		{
		    tisSysLog (0, 2, "Unable to open ar_reduc.dbf for archive processing.\n");
		    return;
		}
		while (getArReducRecord (pointName, &method, &interval, &archive)) {
		    if (archive) {
			if (!ArchiveSnapPoint (oTrendDesc, oArchDesc, pointName,
					    int2methodString(method), currentDay)) {
			    /* delete records from this day from archive file */
			    DeleteSnapRecsByTime (oArchDesc, currentDay - 86400, currentDay);
			    /* close this archive file */
			    DropArchiveSynonym ();
			    free (oArchDesc);
			    /* create a new one */
			    oArchDesc = MakeSnapDesc ("TISARCHIVE", "ARCHIVE_DATA");
			    OpenSnapTable (oArchDesc, &createParams);
			    /*resume archiving for this day */
			    closeArReduc ();
			    if (openArReduc () == 0)
			    {
				tisSysLog (0, 2, "Unable to reopen ar_reduc.dbf for archive processing.\n");
				return;
			    }
			}
		    }
		}
		closeArReduc ();
		/* delete archived records from trend space */
		if (openArReduc () == 0)
		{
		    tisSysLog (0, 2, "Unable to open ar_reduc.dbf for delete after archive processing.\n");
		    return;
		}
		while (getArReducRecord (pointName, &method, &interval, &archive)) {
		    if (archive) {
			DeleteSnapPointRecs (oTrendDesc, pointName, int2methodString(method), currentDay);
		    }
		}
		/* store records while archiving */
#ifndef	STANDALONE_ARCHIVE
		if (!rbEmpty (requestQueue)) {
		     doWriteRequests ();
		}
#endif	STANDALONE_ARCHIVE
		currentDay += 86400; /* add 1 days worth of seconds */
		closeArReduc ();
	    }
	    UpdateArchiveCatalog ();
	    EndSnapRecByTime (oTrendDesc);
	}

	/* Check to see if it is time to clean out records */
	if (cleanEnabled) {
	    stat = GetSnapRecsByTime (oTrendDesc, 0, cleanTime, &snapRec);
            if (stat) {   
		tmblock = gmtime ((time_t *)&snapRec.rec_time);
		currentDay = snapRec.rec_time - (tmblock->tm_hour*3600 + tmblock->tm_min*60 + tmblock->tm_sec)
			 + 86400 /* 1 day in seconds */;

		/* Remove one day of data for any points in the archive */
		while (currentDay <= cleanTime) {
		    DeleteSnapRecsByTime (oTrendDesc, 0, currentDay);
		    /* store records while cleaning */
#ifndef	STANDALONE_ARCHIVE
		    if (!rbEmpty (requestQueue)) {
			doWriteRequests ();
		    }
#endif	STANDALONE_ARCHIVE
		    currentDay += 86400; /* add 1 days worth of seconds */
		}
	    }
	}
	free (oTrendDesc);
	free (oArchDesc);
#else
	if (archiveEnabled) {

	    sprintf (archPath, "/tisarch/%s", archiveFileName);
	    resolveLinks (archPath, pathName);
	    archive = d4open (&cb, pathName);
	    if (archive == NULL) {
		archive = d4create (&cb, pathName, snapFields, snapTags);
		if (archive == NULL) {
		    tisSysLog (0, 2, "Unable to create archive file %s\n", archPath);
		    return;
		}
	    }

	    /* Open all the fields in the archive file */

	    archivePointNameFld = d4field (archive, "POINTNAME");
	    archiveEguTagFld = d4field (archive, "EGUTAG");
	    archiveDispLowFld = d4field (archive, "DISP_LOW");
	    archiveDispHighFld = d4field (archive, "DISP_HIGH");
	    archiveYearFld = d4field (archive, "YEAR");
	    archiveDayOfYearFld = d4field (archive, "DAYOFYEAR");
	    archiveHourFld = d4field (archive, "HOUR");
	    archiveOffsetMinFld = d4field (archive, "OFFSET_MIN");
	    archiveSampleCntFld = d4field (archive, "SAMPLE_CNT");
	    archiveRedMethodFld = d4field (archive, "RED_METHOD");
	    archiveRateFld = d4field (archive, "RATE");
	    for (i=0 ; i<BUCKETS_PER_RECORD ; i++) {
		sprintf (buf, "VALUE%d", i + 1); 
		archiveValueFld [i] = d4field (archive, buf);
		sprintf (buf, "STATUS%d", i + 1); 
		archiveStatusFld [i] = d4field (archive, buf);
	    }
	    d4tag_select (archive, NULL);

	    /* Build a table of all the points to be archived. */

	    buildPointList ();

	}

	/* Get the current time so that we can compare to it */

	timeNow = time (0);
	if (debugLevel >= 1)
	    tisSysLog (0, 2, "Time now is %s\n", ctime (&timeNow));

	/* Get the archive and clean threshold times */

	if (archiveEnabled) {
	    archiveTime = timeNow - (long)(archiveOlderThan * SECONDS_PER_DAY);
	    if (debugLevel >= 1)
		tisSysLog (0, 2, "Records older than %s will be archived.\n",
								ctime (&archiveTime));
	}

	if (cleanEnabled) {
	    cleanTime = timeNow - (long)(cleanOlderThan * SECONDS_PER_DAY);
	    if (debugLevel >= 1)
	    	tisSysLog (0, 2, "Records older than %s will be cleaned.\n",
								ctime (&cleanTime));
	}

	/* Open the ar_snap file. It is automatically positioned at the beginning */

	if (!openArSnap ()) {
	    tisSysLog (0, 2, "Unable to open ar_snap or ar_last.\n");
	    exit (-1);
	}
	dbOpen = TRUE;

	/* Get the tag used by dbServer */

	tag = d4tag (arSnapDesc, "METH_RATE");

	/* Scan through ar_snap and perform archiving and cleaning operations. */

	numRecs = d4reccount (arSnapDesc);
	for (cnt=1 ; cnt<=numRecs ; cnt++) {

	    /* Service any queued requests */
#ifndef	STANDALONE_ARCHIVE
	    if (((cnt % 100) == 0) && !rbEmpty (requestQueue)) {
		d4tag_select (arSnapDesc, tag);
		doWriteRequests ();
		d4tag_select (arSnapDesc, NULL);
	    }
#endif	STANDALONE_ARCHIVE

	    /* Get to the next record to be considered */

	    stat = d4go (arSnapDesc, cnt);
	    if (stat != 0) {
		closeArSnap ();
		dbOpen = FALSE;
		if (archiveEnabled)
		    d4close (archive);
		tisSysLog (0, 2, "Error during d4go -- archiving terminated\n");
		return;
	    }
	    d4unlock_all (arSnapDesc);

	    /* Skip over deleted entries. */

	    strcpy (pointName, f4str (pointNameFld));
	    if (d4deleted (arSnapDesc) || (pointName [0] == ' '))
		continue;

	    /* Create a key for this record */

	    year = f4int (yearFld);
	    dayOfYear = f4int (dayOfYearFld);
	    hour = f4int (hourFld);
	    min = f4int (offsetMinFld);
	    arSnapTime2ts (year, dayOfYear, hour, min, &recTime);
	    stripBlanks (pointName);
	    method = methodString2Int (f4str (redMethodFld));
	    if (f4len (rateFld) == 2)
		interval = (int)(atof (f4str (rateFld)) * 60);
	    else
		interval = (int)(atof (f4str (rateFld)));
	    makeArSnapKey (pointName, method, interval, year, dayOfYear, hour, min, key);

	    /* Check to see if it is time to archive this record */
	    /* We archive records that are complete as of the archive & clean cutoff times. We therefore */
	    /* use the time of the beginning of the next record for the comparison. */

	    strcpy (pointMethName, pointName);
	    strcat (pointMethName, "|");
	    strcat (pointMethName, int2methodString (method));
	    if (archiveEnabled && inPointList (pointMethName)) {
		if (recTime < archiveTime) {
		    if ((stat2 = d4seek (archive, key)) == 0) {	/* Overwrite existing record */
			d4blank (archive);	/* Forces record changed flag to be set */
			copyFields ();
		    } else {	/* Create new record */
			d4append_start (archive, 0);
			copyFields ();
			d4append (archive);
		    }
		    d4flush_all (archive);
		    d4unlock_all (archive);
		}
	    }

	    /* Check to see if it is time to clean this record */

	    if (cleanEnabled) {
		if (recTime < cleanTime) {
		    d4blank (arSnapDesc);		/* Blank the record so that it will float to the top. */
		    d4delete (arSnapDesc);		/* Mark for deletion */
		    updateDBaseFile (arSnapDesc);	/* Force an update to place blanked record in the index */
		}
	    }

	    if (!running)
		break;

	}

	closeArSnap ();
	dbOpen = FALSE;
	if (archiveEnabled) {
	    updateDBaseFile (archive);
	    d4close (archive);
	}
#endif ORACLESNAP

	/* Archive and clean the notes file. */

	archiveAndCleanNotes ();

	if (debugLevel >= 1) {
	    clock = time (0);
	    tisSysLog (0, 2, "Archiver finished at %s.\n", ctime (&clock));
	}
	archiveInProgress = FALSE;

	if (archiveEnabled)
	    freePointList ();
    }

/*----------*/
#ifndef	ORACLESNAP
static void copyFields (void)

    {
	int	i;

	f4assign (archivePointNameFld, f4str (pointNameFld));
	f4assign (archiveEguTagFld, f4str (eguTagFld));
	f4assign (archiveDispLowFld, f4str (dispLowFld));
	f4assign (archiveDispHighFld, f4str (dispHighFld));
	f4assign_int (archiveYearFld, f4int (yearFld));
	f4assign_int (archiveDayOfYearFld, f4int (dayOfYearFld));
	f4assign_int (archiveHourFld, f4int (hourFld));
	f4assign_int (archiveOffsetMinFld, f4int (offsetMinFld));
	f4assign_int (archiveSampleCntFld, f4int (sampleCntFld));
	f4assign (archiveRedMethodFld, f4str (redMethodFld));
	/*
	  GCT 25-Oct-1996 changing code from f4assign_int to f4assign
	  to copy rate as left justified string rather than an integer
	  which is right justified. This will allow the key created by
	  makeArSnapKey() to find the record when asked. Also, the 
	  rate field contents are fetched using f4str to get it as
	  a string.
	*/
	f4assign (archiveRateFld, f4str (rateFld));
	for (i=0 ; i<BUCKETS_PER_RECORD ; i++) {
	    f4assign (archiveValueFld [i], f4str (valueFld [i]));
	    f4assign (archiveStatusFld [i], f4str (statusFld [i]));
	}
    }
#endif !ORACLESNAP

/*----------*/
static void archiveAndCleanNotes (void)

    {

	int		stat;
	PointName	pointName;
	long		cnt;
#ifdef ORACLESNAP
	char		*notepoint;
	long		notetime;
	char		*notetext;
#else
	DATA4		*newNotesDesc;
	FIELD4		*newNotesNameFld, *newNotesYearFld, *newNotesDayOfYearFld; 
	FIELD4		*newNotesHourFld;
	FIELD4		*newNotesMinFld, *newNotesSecFld, *newNotesNoteFld;	
	time_t	        recTime;
	int		year, dayOfYear, hour, min, sec;
	char		notesName [100], notesPath [100], pathName [100], key [100];
	long		numRecs;
#endif ORACLESNAP

	/* Open the ar_snap file. It is automatically positioned at the beginning */

#ifdef ORACLESNAP
	oTrendDesc = MakeSnapDesc ("TISTREND", "TREND_DATA");
	if (!OpenSnapTable (oTrendDesc, FALSE)) {
	    free (oTrendDesc);
	    return;
	}

	/* Check if the archive file exists. If it does not then create it. */

	if (!archiveEnabled) {
	    return;
	}

	/* Scan through ar_snapn and perform archiving and cleaning operations. */

	if (archiveTime > cleanTime)
	    stat = GetSnapNoteByTime (oTrendDesc, &notepoint,
			    0, archiveTime, &notetime, &notetext);
	else
	    stat = GetSnapNoteByTime (oTrendDesc, &notepoint,
			    0, cleanTime, &notetime, &notetext);
	for (cnt=1 ; stat ; cnt++) {

	    /* Service any queued requests */
#ifndef STANDALONE_ARCHIVE
	    if (((cnt % 100) == 0) && !rbEmpty (requestQueue)) {
		EndSnapNoteByTime (oTrendDesc);
		doWriteRequests ();
		if (archiveTime > cleanTime)
		    stat = GetSnapNoteByTime (oTrendDesc, &notepoint,
				    0, archiveTime, &notetime, &notetext);
		else
		    stat = GetSnapNoteByTime (oTrendDesc, &notepoint,
				    0, cleanTime, &notetime, &notetext);
	    }
#endif	STANDALONE_ARCHIVE

	    /* Check to see if it is time to archive this record */
	    strncpy (pointName, notepoint, sizeof(pointName));
	    stripBlanks (pointName);

	    if (archiveEnabled && inPointList (pointName)) {
		if (notetime < archiveTime) {
		    AddSnapNote (oArchDesc, notepoint, notetime, notetext);
		    DeleteSnapNote (oTrendDesc, notepoint, notetime);
		}
	    }

	    if (!running)
		break;

	    /* Get to the next record to be considered */

	    stat = NextSnapNoteByTime (oTrendDesc, &notepoint, &notetime, &notetext);

	}

	EndSnapNoteByTime (oTrendDesc);

	/* Clean out of date records */

	if (cleanEnabled) {
	    DeleteSnapNoteByTime (oTrendDesc, 0, cleanTime);
	}

	free (oTrendDesc);
	free (oArchDesc);
#else
	if (!openArNotesFile ())
	    return;

	/* Check if the archive file exists. If it does not then create it. */

	if (archiveEnabled) {
	    strcpy (notesName, archiveFileName);
	    *strchr (notesName, '.') = '\0';
	    sprintf (notesPath, "/tisarch/%sn", notesName);
	    resolveLinks (notesPath, pathName);
	    newNotesDesc = d4open (&cb, pathName);
	    if (newNotesDesc == NULL) {
		newNotesDesc = d4create (&cb, pathName, notesFields, notesTags);
		if (newNotesDesc == NULL) {
		    tisSysLog (0, 2, "Unable to create archive notes file %s\n",notesPath);
		    return;
		}
	    }
	    newNotesNameFld = d4field (newNotesDesc, "POINTNAME");
	    newNotesYearFld = d4field (newNotesDesc, "YEAR");
	    newNotesDayOfYearFld = d4field (newNotesDesc, "DAYOFYEAR");
	    newNotesHourFld = d4field (newNotesDesc, "HOUR");
	    newNotesMinFld = d4field (newNotesDesc, "MINUTE");
	    newNotesSecFld = d4field (newNotesDesc, "SECOND");
	    newNotesNoteFld = d4field (newNotesDesc, "NOTE");
	}

	/* Scan through ar_snapn and perform archiving and cleaning operations. */

	numRecs = d4reccount (arNotesDesc);
	for (cnt=1 ; cnt<=numRecs ; cnt++) {

	    /* Service any queued requests after every 100 records. */
#ifndef STANDALONE_ARCHIVE
	    if (((cnt % 100) == 0) && !rbEmpty (requestQueue))
		doWriteRequests ();
#endif	STANDALONE_ARCHIVE

	    /* Get to the next record to be considered */

	    stat = d4go (arNotesDesc, cnt);
	    if (stat != 0) {
		closeArNotesFile ();
		if (archiveEnabled)
		    d4close (newNotesDesc);
		printf ("Error during d4go -- notes archiving terminated\n");
		return;
	    }

	    /* Skip over deleted entries. */

	    stringCopy (pointName, f4str (notesNameFld), POINTNAME_SIZE+1);
	    if (d4deleted (arNotesDesc) || (pointName [0] == ' '))
		continue;

	    /* Create a key for this record */

	    year = f4int (notesYearFld);
	    dayOfYear = f4int (notesDayOfYearFld);
	    hour = f4int (notesHourFld);
	    min = f4int (notesMinFld);
	    sec = f4int (notesSecFld);
	    arNotesTime2ts (year, dayOfYear, hour, min, sec, &recTime);
	    stripBlanks (pointName);
	    makeArNotesKey (pointName, year, dayOfYear, hour, min, sec, key);

	    /* Check to see if it is time to archive this record */

	    if (archiveEnabled && inPointList (pointName)) {
		if (recTime < archiveTime) {
		    if (d4seek (newNotesDesc, key) == 0) {	/* Overwrite existing record */
			f4assign (newNotesNameFld, f4str (notesNameFld));
			f4assign_int (newNotesYearFld, f4int (notesYearFld));
			f4assign_int (newNotesDayOfYearFld, f4int (notesDayOfYearFld));
			f4assign_int (newNotesHourFld, f4int (notesHourFld));
			f4assign_int (newNotesMinFld, f4int (notesMinFld));
			f4assign_int (newNotesSecFld, f4int (notesSecFld));
			f4memo_assign (newNotesNoteFld, f4memo_str (notesNoteFld));
		    } else {	/* Create new record */
			d4append_start (newNotesDesc, 0);
			f4assign (newNotesNameFld,  f4str (notesNameFld));
			f4assign_int (newNotesYearFld, f4int (notesYearFld));
			f4assign_int (newNotesDayOfYearFld, f4int (notesDayOfYearFld));
			f4assign_int (newNotesHourFld, f4int (notesHourFld));
			f4assign_int (newNotesMinFld, f4int (notesMinFld));
			f4assign_int (newNotesSecFld, f4int (notesSecFld));
			f4memo_assign (newNotesNoteFld, f4memo_str (notesNoteFld));
			d4append (newNotesDesc);
		    }
		}
	    }

	    /* Check to see if it is time to clean this record */

	    if (cleanEnabled) {
		if (recTime < cleanTime) {
		    d4blank (arNotesDesc);		/* Blank the record so that it will float to the top. */
		    d4delete (arNotesDesc);		/* Mark for deletion */
		    updateDBaseFile (arNotesDesc);	/* Force an update to place blanked record in the index */
		}
	    }

	    if (!running)
		break;

	}

	closeArNotesFile ();
	if (archiveEnabled) {
	    updateDBaseFile (newNotesDesc);
	    d4close (newNotesDesc);
	}
#endif ORACLESNAP

    }

/*----------*/
#ifndef	ORACLESNAP
static void makeArNotesKey (char *pointName, int year, int dayOfYear, int hour, int min,
			    int secs, char *key)

    {

	sprintf (key, "%-*s%*d%*d%*d%*d%2d",
			POINTNAME_SIZE, pointName,
			YEAR_SIZE, year,
			DAY_SIZE, dayOfYear,
			HOUR_SIZE, hour,
			MIN_SIZE, min,
			secs);

    }
#endif	!ORACLESNAP

/*----------*/
static void buildPointList (void)

    {

        PointMethName   pointName;
	int         interval, method, archive;

        /* Open the ar_reduc file and determine how many records it has. */
        /* Allocate enough memory to hold a list of all points in ar_reduc. */
        /* This is a worst-case size. */

	if (openArReduc () == 0)
	{
	    tisSysLog (0, 2, "Unable to open ar_reduc.dbf to get pointlist.\n");
	    return;
	}
	pointList = (PointMethName *) malloc (sizeof (PointMethName) * getArReducSize ());
	numPoints = 0;

	/* Read through ar_reduc. Any point that has the archive bit set */
	/* is added to the list */

	while (getArReducRecord (pointName, &method, &interval, &archive)) {
	    if (archive) {
	        strcpy (pointList [numPoints], pointName);
		strcat (pointList [numPoints], "|");
		strcat (pointList [numPoints++], int2methodString(method));
	    }
	}

	closeArReduc ();

	/* Now sort the list for later binary searching */

	qsort (pointList, numPoints, sizeof (PointMethName), searchCompare);

    }

/*----------*/
static void freePointList (void)

    {

        free ((void *)pointList);

    }

/*----------*/
static int inPointList (char *pointName)

    {

        return bsearch (pointName, pointList, numPoints, sizeof (PointMethName), searchCompare) != NULL;

    } 

/*----------*/
static int searchCompare (const void *key, const void *entry)

    {
	char *cp;
	int retval;

	/* if the key doesn't have a method */
	if (strchr (key, '|') == NULL)
	{
	    /* compare the strings without method */
	    cp = strchr (entry, '|');
	    if (cp)
		*cp = 0;
	    retval = strcmp ((char *)key, (char *)entry);
	    if (cp)
		*cp = '|';
	    return retval;
	}
	else
	{
	    /* just compare the strings */
	    return strcmp ((char *)key, (char *)entry);
	}
    }

#ifdef ORACLESNAP
/*----------*/
/* routine to reduce all records according to the provided schedule */
static int ReduceSnapFile (OracleDesc *desc, char *pointName, int method, REDUC_SCHEDULE *schedule)
{
    int stat;
    int srcix;
    int secsIntoBucket;
    float lastValue;
    time_t srcts;
    time_t dstts;
    int i, bix;
    time_t archiveTime;

#ifdef STATISTICS
    int scanrecs;
    int delrecs;
    int reducrecs;
    int skiprecs;
    struct timeval startreduce;
    struct timeval startget;
    struct timeval startnext;
    struct timeval endtime;
    struct timeval avgnext = {0,0};
    char nullstr = 0;
#endif STATISTICS

    /* Get first record in archive */
#ifdef STATISTICS
    gettimeofday (&startreduce, &nullstr);
    gettimeofday (&startget, &nullstr);
#endif STATISTICS
    stat = GetSnapRecord (desc, pointName, int2methodString(method), 0, &snapRec);
#ifdef STATISTICS
    gettimeofday (&endtime, &nullstr);
    startget.tv_sec = endtime.tv_sec - startget.tv_sec;
    startget.tv_usec = endtime.tv_usec - startget.tv_usec;
    if (startget.tv_usec < 0)
    {
	startget.tv_usec += 1000000;
	startget.tv_sec --;
    }
#endif STATISTICS
    srcix = 0;
    srcts = snapRec.rec_time;
#ifdef STATISTICS
    scanrecs = 1;
    delrecs = 0;
    reducrecs = 0;
    skiprecs = 0;
#endif STATISTICS
    i = schedule->numschedule-1;
    /* adjust dsts to strt new reduc series */
    dstts = 0;
    archiveTime = timeNow - (long)(schedule->schedule[i].age * SECONDS_PER_DAY);
    /* reduce until no more records (break out early if end of reduc schedule seen) */
    while (stat)
    {
	if (debugLevel >= 1)
		tisSysLog (0, 2, "Records older than %s will be reduced.\n",
								ctime (&archiveTime));
	while (snapRec.rec_time >= archiveTime && (i > 0))
	{
	    if (dstts && (newRec.sample_cnt != 0))
	    {
		UpdateSnapRecord (desc, &newRec);
		dstts = 0;
	    }
	    /* move down toward later reductions */
	    i--;
	    /* adjust dsts to strt new reduc series */
	    dstts = 0;
	    archiveTime = timeNow - (long)(schedule->schedule[i].age * SECONDS_PER_DAY);
	}
	/* if end of schedules, break out of loop (entry 0 is base reduction) */
	if (i == 0)
	    break;
	/* delete this record, this is a 'clean' indication */
	if (schedule->schedule[i].age == 0)
	{
	    DeleteSnapRecord (desc, &snapRec);
#ifdef STATISTICS
	    delrecs++;
#endif STATISTICS
	}
	else if (snapRec.rate < schedule->schedule[i].interval) /* do records with higher rates */
	{ /* reduce a record ! */
	    /* if necessary, start a new destination record */
	    if (dstts == 0)
	    {
		newRec = snapRec;
		newRec.rate = schedule->schedule[i].interval;
		newRec.rec_time = recordStartTimeA (newRec.rate, snapRec.rec_time);
		dstts = newRec.rec_time;
		newRec.sample_cnt = 0;
		/* fill to reduc time with 'C' values */
		bix = bucketIndexA (newRec.rec_time, newRec.rate, snapRec.rec_time);
		while (newRec.sample_cnt < bix)
		{
		    /* any gaps represent Connect errors */
		    newRec.value[newRec.sample_cnt] = 0;
		    newRec.status[newRec.sample_cnt][0] = 'C';
		    newRec.status[newRec.sample_cnt][1] = ' ';
		    newRec.status[newRec.sample_cnt][2] = ' ';
		    newRec.sample_cnt++;
		    dstts += newRec.rate;
		}
#ifdef STATISTICS
		reducrecs++;
#endif STATISTICS
	    }
	    /* initialize current cell */
	    newRec.value[newRec.sample_cnt] = 0;
	    newRec.status[newRec.sample_cnt][0] = 0;
	    /* reduce from record data until this sample is filled, then proceed to outer loop */
	    while (stat && (srcts < (dstts + newRec.rate)))
	    {
		/* Use the appropriate reduction method to integrate the new sample */
		switch (method)
		{
		    case REDUC_MIN:
			if ((newRec.status[newRec.sample_cnt][0] == 0)
			  || (snapRec.value[srcix] < newRec.value[newRec.sample_cnt]))
			    newRec.value[newRec.sample_cnt] = snapRec.value[srcix];
			break;
		    case REDUC_MAX:
			if ((newRec.status[newRec.sample_cnt][0] == 0)
			  || (snapRec.value[srcix] > newRec.value[newRec.sample_cnt]))
			    newRec.value[newRec.sample_cnt] = snapRec.value[srcix];
			break;
		    case REDUC_SMP:
			if (newRec.status[newRec.sample_cnt][0] == 0)
			    newRec.value[newRec.sample_cnt] = snapRec.value[srcix];
			break;
		    case REDUC_AVG:
			secsIntoBucket = secondsIntoBucketA (newRec.rec_time, newRec.rate, srcts);
			if (secsIntoBucket == 0)
			{
			    newRec.value[newRec.sample_cnt] = snapRec.value[srcix];
			    lastValue = snapRec.value[srcix];
			    break;
			}
		        if (newRec.status[newRec.sample_cnt][0] == 0)
			{
			    newRec.value[newRec.sample_cnt] = snapRec.value[srcix];
			}
			else
			{
			    /* remove excess contribution for previous value */
			    /* by calculating the average for this far into bucket */
			    newRec.value[newRec.sample_cnt] = ((newRec.value[newRec.sample_cnt] * newRec.rate) -
					(lastValue * (newRec.rate - secsIntoBucket))) /
					secsIntoBucket;
			}
			/* recalculate average for total interval */
			newRec.value[newRec.sample_cnt] = ((newRec.value[newRec.sample_cnt] * secsIntoBucket) +
			          (snapRec.value[srcix] * (newRec.rate - secsIntoBucket))) /
				  newRec.rate;
			lastValue = snapRec.value[srcix];
			break;
		}
		if ((newRec.status[newRec.sample_cnt][0] == 0)
		  || letter2severity (snapRec.status[srcix][1]) > letter2severity (newRec.status[newRec.sample_cnt][1])) {
		    newRec.status[newRec.sample_cnt][1] = snapRec.status[srcix][1];
		    newRec.status[newRec.sample_cnt][2] = snapRec.status[srcix][2];
		}
		if (newRec.status[newRec.sample_cnt][0] == 0)
		    newRec.status[newRec.sample_cnt][0] = snapRec.status[srcix][0];
		else if (newRec.status[newRec.sample_cnt][0] != snapRec.status[srcix][0])
		{
		    switch (snapRec.status[snapRec.sample_cnt][0])
		    {
			case '*': /* if new value is *, the record is a star */
			    newRec.status[newRec.sample_cnt][0] = snapRec.status[srcix][0];
			    break;
			case 'C': /* if new value is C, and the record is not star (and record is not C already) it's a q */
			    if (newRec.status[newRec.sample_cnt][0] != '*')
				newRec.status[newRec.sample_cnt][0] = 'q';
			case 'q': /* if new value is q, and the record is not star (and record is not q already) it's a q */
			    if (newRec.status[newRec.sample_cnt][0] != '*')
		   		newRec.status[newRec.sample_cnt][0] = 'q';
		    }
		}
		srcts += snapRec.rate;
		srcix++;
		if (srcix == snapRec.sample_cnt)
		{
		    /* delete the reduced record */
		    DeleteSnapRecord (desc, &snapRec);
		    /* get the next record */
#ifdef STATISTICS
		    gettimeofday (&startnext, &nullstr);
#endif STATISTICS
		    stat = NextSnapRecord (desc, &snapRec);
#ifdef STATISTICS
		    gettimeofday (&endtime, &nullstr);
		    startnext.tv_sec = endtime.tv_sec - startnext.tv_sec;
		    startnext.tv_usec = endtime.tv_usec - startnext.tv_usec;
		    if (startnext.tv_usec < 0)
		    {
			startnext.tv_usec += 1000000;
			startnext.tv_sec --;
		    }
		    avgnext.tv_usec = startnext.tv_usec + avgnext.tv_usec;
		    avgnext.tv_sec = startnext.tv_sec + avgnext.tv_sec;
		    if (avgnext.tv_usec >= 1000000)
		    {
			avgnext.tv_usec -= 1000000;
			avgnext.tv_sec++;
		    }
#endif STATISTICS
		    srcix = 0;
		    srcts = snapRec.rec_time;
#ifdef STATISTICS
		    scanrecs++;
		    delrecs++;
#endif STATISTICS
		}
	    }
	    newRec.sample_cnt++;
	    dstts += newRec.rate;
	    /* close current reduced record if necessary */
	    if ((dstts - newRec.rec_time >= 24*60*60) || (newRec.sample_cnt == 60))
	    {
		UpdateSnapRecord (desc, &newRec);
		dstts = 0;
	    }
	}
	else
	{
	    /* get the next record */
#ifdef STATISTICS
	    gettimeofday (&startnext, &nullstr);
#endif STATISTICS
	    stat = NextSnapRecord (desc, &snapRec);
#ifdef STATISTICS
	    gettimeofday (&endtime, &nullstr);
	    startnext.tv_sec = endtime.tv_sec - startnext.tv_sec;
	    startnext.tv_usec = endtime.tv_usec - startnext.tv_usec;
	    if (startnext.tv_usec < 0)
	    {
		startnext.tv_usec += 1000000;
		startnext.tv_sec --;
	    }
	    avgnext.tv_usec = startnext.tv_usec + avgnext.tv_usec;
	    avgnext.tv_sec = startnext.tv_sec + avgnext.tv_sec;
	    if (avgnext.tv_usec >= 1000000)
	    {
		avgnext.tv_usec -= 1000000;
		avgnext.tv_sec++;
	    }
#endif STATISTICS
	    srcix = 0;
	    if (stat)
		srcts = snapRec.rec_time;
	    else
	    {
		srcts = 0x7fffffff;
		snapRec.rate = 5;
	    }
#ifdef STATISTICS
	    scanrecs++;
	    skiprecs++;
#endif STATISTICS
 	}
    }
    if (dstts && (newRec.sample_cnt != 0))
    {
	UpdateSnapRecord (desc, &newRec);
	dstts = 0;
    }
    EndSnapRecord (desc);
#ifdef STATISTICS
    gettimeofday (&endtime, &nullstr);
    startreduce.tv_sec = endtime.tv_sec - startreduce.tv_sec;
    startreduce.tv_usec = endtime.tv_usec - startreduce.tv_usec;
    if (startreduce.tv_usec < 0)
    {
	startreduce.tv_usec += 1000000;
	startreduce.tv_sec --;
    }
    avgnext.tv_sec = avgnext.tv_sec/scanrecs;
    avgnext.tv_usec = (avgnext.tv_usec + ((avgnext.tv_sec%scanrecs) * 1000000))/scanrecs;
    printf ("Reduce point %s|%s: Scan %d, Delete %d, Skip %d, Reduced %d in %d.%6.6d secs (%d.%6.6d %d.%6.6d)\n",
		pointName, int2methodString(method), scanrecs, delrecs, skiprecs, reducrecs,
		startreduce.tv_sec, startreduce.tv_usec, startget.tv_sec, startget.tv_usec,
		avgnext.tv_sec, avgnext.tv_usec);
#endif STATISTICS
}
#endif ORACLESNAP
