/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/* dBASE (codeBase) related functions */

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <math.h>
#include <sys/stat.h>
#include <time.h>
#ifdef WIN32
#endif WIN32
#include "msgMon.h"
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "mlist.h"
#include "trendDefs.h"
#include "arSnapRecord.h"
#include "utils.h"
#include "ar_schedule.h"	/* schedule access definitions */
#ifdef ORACLESNAP
#include "oracleSnap.h"
#include "oracleSnapLib.h"
#endif ORACLESNAP

/* Functions defined in this module */

	void	initCodebase (void);
	int	openArReduc (void);
	long	getArReducSize (void);
	int	getArReducRecord (char *, int *, int *, int *);
	int	getArReducSchedule (char *, int *, REDUC_SCHEDULE *, int *);
	void	closeArReduc (void);
	int	openArSnap (void);
	void	closeArSnap (void);
#ifndef	ORACLESNAP
static	void	getArRecord (ArSnapRecord *);
static	void	putArRecord (ArSnapRecord *);
static  char    *rate2string (int);
#endif	!ORACLESNAP
	int	getArSnapRecord (char *, int, int, time_t, ArSnapRecord *);
	int	putArSnapRecord (ArSnapRecord *);
	int	openArNotesFile (void);
	void	closeArNotesFile (void);
	void	makeArSnapKey (char *, int, int, int, int, int, int, char *);
	void	arSnapTime2ts (int, int, int, int, time_t *);
	void	arNotesTime2ts (int, int, int, int, int, time_t *);
	void	ts2arSnapTime (time_t, int *, int *, int *, int *);
	int     methodString2Int (char *);
	void	updateDBaseFile (DATA4 *);

/* External functions */

extern	void	resolveLinks ();

/* External data */

extern	int    debugLevel;
extern  int    myPid;
        
/* codeBase related data */

CODE4	cb;

DATA4	*arReducDesc;
FIELD4	*pointNameFld, *reducMethodFld, *reducIntervalFld, *reducArchiveFld;

DATA4	*arArchDesc;
FIELD4	*archNameFld, *archMethodFld, *archYearFld, *archDayOfYearFld, *archHourFld, *archMinFld, *archRateFld;

static  int    rateFieldSize;

/* default unknown schedule to 60 second rate, no age reductions */
static	REDUC_SCHEDULE defaultSchedule = { "", 1, { 0, 60 }};

#ifdef ORACLESNAP
extern CreateParams defCreateParams;
OracleDesc *oSnapDesc;
#else
DATA4	*arSnapDesc;
FIELD4	*pointNameFld, *eguTagFld, *dispLowFld, *dispHighFld, *yearFld, *dayOfYearFld, *hourFld;
FIELD4	*offsetMinFld, *sampleCntFld, *redMethodFld, *rateFld;
FIELD4  *valueFld [BUCKETS_PER_RECORD], *statusFld [BUCKETS_PER_RECORD];  

DATA4	*arNotesDesc;
FIELD4	*notesNameFld, *notesYearFld, *notesDayOfYearFld, *notesHourFld, *notesMinFld, *notesSecFld, *notesNoteFld;

/* ar_snap field and tag definitions */
FIELD4INFO	snapFields[] =	{

	/* Field name,		Type,	Width,		Decimals */
	{ "POINTNAME",		'C',	POINTNAME_SIZE,	        0},
	{ "EGUTAG",		'C',	EGUTAG_SIZE,	        0},
	{ "DISP_LOW",		'C',	LIMITS_SIZE, 	        0},
	{ "DISP_HIGH",		'C',	LIMITS_SIZE,	        0},
	{ "YEAR",		'N',	YEAR_SIZE,	        0},
	{ "DAYOFYEAR",		'N',	DAY_SIZE,	        0},
	{ "HOUR",		'N',	HOUR_SIZE,  	        0},
	{ "OFFSET_MIN",		'N',	MIN_SIZE, 	        0},
	{ "SAMPLE_CNT",		'N',	SAMPLECNT_SIZE,         0},
	{ "RED_METHOD",		'C',	METHOD_SIZE,	        0},
	{ "RATE",		'C',	RATE_SIZE,              0},
	{ "VALUE1",		'C',	VALUE_SIZE,		0},
	{ "STATUS1",		'C',	STATUS_SIZE,		0},
	{ "VALUE2",		'C',	VALUE_SIZE,		0},
	{ "STATUS2",		'C',	STATUS_SIZE,		0},
	{ "VALUE3",		'C',	VALUE_SIZE,		0},
	{ "STATUS3",		'C',	STATUS_SIZE,		0},
	{ "VALUE4",		'C',	VALUE_SIZE,		0},
	{ "STATUS4",		'C',	STATUS_SIZE,		0},
	{ "VALUE5",		'C',	VALUE_SIZE,		0},
	{ "STATUS5",		'C',	STATUS_SIZE,		0},
	{ "VALUE6",		'C',	VALUE_SIZE,		0},
	{ "STATUS6",		'C',	STATUS_SIZE,		0},
	{ "VALUE7",		'C',	VALUE_SIZE,		0},
	{ "STATUS7",		'C',	STATUS_SIZE,		0},
	{ "VALUE8",		'C',	VALUE_SIZE,		0},
	{ "STATUS8",		'C',	STATUS_SIZE,		0},
	{ "VALUE9",		'C',	VALUE_SIZE,		0},
	{ "STATUS9",		'C',	STATUS_SIZE,		0},
	{ "VALUE10",		'C',	VALUE_SIZE,		0},
	{ "STATUS10",		'C',	STATUS_SIZE,		0},
	{ "VALUE11",		'C',	VALUE_SIZE,		0},
	{ "STATUS11",		'C',	STATUS_SIZE,		0},
	{ "VALUE12",		'C',	VALUE_SIZE,		0},
	{ "STATUS12",		'C',	STATUS_SIZE,		0},
	{ "VALUE13",		'C',	VALUE_SIZE,		0},
	{ "STATUS13",		'C',	STATUS_SIZE,		0},
	{ "VALUE14",		'C',	VALUE_SIZE,		0},
	{ "STATUS14",		'C',	STATUS_SIZE,		0},
	{ "VALUE15",		'C',	VALUE_SIZE,		0},
	{ "STATUS15",		'C',	STATUS_SIZE,		0},
	{ "VALUE16",		'C',	VALUE_SIZE,		0},
	{ "STATUS16",		'C',	STATUS_SIZE,		0},
	{ "VALUE17",		'C',	VALUE_SIZE,		0},
	{ "STATUS17",		'C',	STATUS_SIZE,		0},
	{ "VALUE18",		'C',	VALUE_SIZE,		0},
	{ "STATUS18",		'C',	STATUS_SIZE,		0},
	{ "VALUE19",		'C',	VALUE_SIZE,		0},
	{ "STATUS19",		'C',	STATUS_SIZE,		0},
	{ "VALUE20",		'C',	VALUE_SIZE,		0},
	{ "STATUS20",		'C',	STATUS_SIZE,		0},
	{ "VALUE21",		'C',	VALUE_SIZE,		0},
	{ "STATUS21",		'C',	STATUS_SIZE,		0},
	{ "VALUE22",		'C',	VALUE_SIZE,		0},
	{ "STATUS22",		'C',	STATUS_SIZE,		0},
	{ "VALUE23",		'C',	VALUE_SIZE,		0},
	{ "STATUS23",		'C',	STATUS_SIZE,		0},
	{ "VALUE24",		'C',	VALUE_SIZE,		0},
	{ "STATUS24",		'C',	STATUS_SIZE,		0},
	{ "VALUE25",		'C',	VALUE_SIZE,		0},
	{ "STATUS25",		'C',	STATUS_SIZE,		0},
	{ "VALUE26",		'C',	VALUE_SIZE,		0},
	{ "STATUS26",		'C',	STATUS_SIZE,		0},
	{ "VALUE27",		'C',	VALUE_SIZE,		0},
	{ "STATUS27",		'C',	STATUS_SIZE,		0},
	{ "VALUE28",		'C',	VALUE_SIZE,		0},
	{ "STATUS28",		'C',	STATUS_SIZE,		0},
	{ "VALUE29",		'C',	VALUE_SIZE,		0},
	{ "STATUS29",		'C',	STATUS_SIZE,		0},
	{ "VALUE30",		'C',	VALUE_SIZE,		0},
	{ "STATUS30",		'C',	STATUS_SIZE,		0},
	{ "VALUE31",		'C',	VALUE_SIZE,		0},
	{ "STATUS31",		'C',	STATUS_SIZE,		0},
	{ "VALUE32",		'C',	VALUE_SIZE,		0},
	{ "STATUS32",		'C',	STATUS_SIZE,		0},
	{ "VALUE33",		'C',	VALUE_SIZE,		0},
	{ "STATUS33",		'C',	STATUS_SIZE,		0},
	{ "VALUE34",		'C',	VALUE_SIZE,		0},
	{ "STATUS34",		'C',	STATUS_SIZE,		0},
	{ "VALUE35",		'C',	VALUE_SIZE,		0},
	{ "STATUS35",		'C',	STATUS_SIZE,		0},
	{ "VALUE36",		'C',	VALUE_SIZE,		0},
	{ "STATUS36",		'C',	STATUS_SIZE,		0},
	{ "VALUE37",		'C',	VALUE_SIZE,		0},
	{ "STATUS37",		'C',	STATUS_SIZE,		0},
	{ "VALUE38",		'C',	VALUE_SIZE,		0},
	{ "STATUS38",		'C',	STATUS_SIZE,		0},
	{ "VALUE39",		'C',	VALUE_SIZE,		0},
	{ "STATUS39",		'C',	STATUS_SIZE,		0},
	{ "VALUE40",		'C',	VALUE_SIZE,		0},
	{ "STATUS40",		'C',	STATUS_SIZE,		0},
	{ "VALUE41",		'C',	VALUE_SIZE,		0},
	{ "STATUS41",		'C',	STATUS_SIZE,		0},
	{ "VALUE42",		'C',	VALUE_SIZE,		0},
	{ "STATUS42",		'C',	STATUS_SIZE,		0},
	{ "VALUE43",		'C',	VALUE_SIZE,		0},
	{ "STATUS43",		'C',	STATUS_SIZE,		0},
	{ "VALUE44",		'C',	VALUE_SIZE,		0},
	{ "STATUS44",		'C',	STATUS_SIZE,		0},
	{ "VALUE45",		'C',	VALUE_SIZE,		0},
	{ "STATUS45",		'C',	STATUS_SIZE,		0},
	{ "VALUE46",		'C',	VALUE_SIZE,		0},
	{ "STATUS46",		'C',	STATUS_SIZE,		0},
	{ "VALUE47",		'C',	VALUE_SIZE,		0},
	{ "STATUS47",		'C',	STATUS_SIZE,		0},
	{ "VALUE48",		'C',	VALUE_SIZE,		0},
	{ "STATUS48",		'C',	STATUS_SIZE,		0},
	{ "VALUE49",		'C',	VALUE_SIZE,		0},
	{ "STATUS49",		'C',	STATUS_SIZE,		0},
	{ "VALUE50",		'C',	VALUE_SIZE,		0},
	{ "STATUS50",		'C',	STATUS_SIZE,		0},
	{ "VALUE51",		'C',	VALUE_SIZE,		0},
	{ "STATUS51",		'C',	STATUS_SIZE,		0},
	{ "VALUE52",		'C',	VALUE_SIZE,		0},
	{ "STATUS52",		'C',	STATUS_SIZE,		0},
	{ "VALUE53",		'C',	VALUE_SIZE,		0},
	{ "STATUS53",		'C',	STATUS_SIZE,		0},
	{ "VALUE54",		'C',	VALUE_SIZE,		0},
	{ "STATUS54",		'C',	STATUS_SIZE,		0},
	{ "VALUE55",		'C',	VALUE_SIZE,		0},
	{ "STATUS55",		'C',	STATUS_SIZE,		0},
	{ "VALUE56",		'C',	VALUE_SIZE,		0},
	{ "STATUS56",		'C',	STATUS_SIZE,		0},
	{ "VALUE57",		'C',	VALUE_SIZE,		0},
	{ "STATUS57",		'C',	STATUS_SIZE,		0},
	{ "VALUE58",		'C',	VALUE_SIZE,		0},
	{ "STATUS58",		'C',	STATUS_SIZE,		0},
	{ "VALUE59",		'C',	VALUE_SIZE,		0},
	{ "STATUS59",		'C',	STATUS_SIZE,		0},
	{ "VALUE60",		'C',	VALUE_SIZE,      	0},
	{ "STATUS60",		'C',	STATUS_SIZE,		0},
	{0, 0, 0, 0}		/* NULL-record terminator */
};


TAG4INFO	snapTags [] =	{
	{ "METH_RATE",	"STR(YEAR,4,0)+STR(DAYOFYEAR,3,0)+STR(HOUR,2,0)+STR(OFFSET_MIN,2,0)+POINTNAME+RED_METHOD+RATE+STR(RECNO(),8,0)", 0, 0, 0},
	{ "METHOD_REC",	"POINTNAME+RED_METHOD+STR(YEAR,4,0)+STR(DAYOFYEAR,3,0)+STR(HOUR,2,0)+STR(OFFSET_MIN,2,0)+STR(RECNO(),8,0)", 0, 0, 0},
	{0, 0, 0, 0, 0}		/* NULL-record terminator */
};

FIELD4INFO	notesFields [] = {
   /* field name,   type, width, decimals */
   { "POINTNAME",   'C',    POINTNAME_SIZE, 0 },
   { "YEAR",	    'N',    YEAR_SIZE,      0 },
   { "DAYOFYEAR",   'N',    DAY_SIZE,       0 },
   { "HOUR",	    'N',    HOUR_SIZE,      0 },
   { "MINUTE",	    'N',    MIN_SIZE,       0 },
   { "SECOND",	    'N',    SECOND_SIZE,    0 },
   { "NOTE",	    'M',    10,    0 },
   { 0, 0, 0, 0 },
};

TAG4INFO notesTags [] = {
   /* name,	    expression,	filter, unique,	     descending */
   { "MEMO_REC", "POINTNAME+STR(YEAR,4)+STR(DAYOFYEAR,3)+STR(HOUR,2)+STR(MINUTE,2)+STR(SECOND,2)",	"", 0, 0 },
   { 0, 0, 0, 0, 0 },
};
#endif ORACLESNAP

/*----------*/
void initCodebase (void)

    {

	d4init (&cb);
#if defined(WIN32) || defined(USECB63)
	cb.errOpen = 0;
	cb.errCreate = 0;
	cb.errFieldName = 0;
#else
	cb.open_error = 0;
	cb.create_error = 0;
	cb.field_name_error = 0;
#endif WIN32
#ifdef ORACLESNAP
	OpenOracle ();
#endif ORACLESNAP
    }

/*----------*/
int openArReduc (void)	/* Open the ar_reduc database and interesting fields */

    {

	char	pathName [200];

	resolveLinks (AR_REDUC_FILE, pathName);
	arReducDesc = d4open (&cb, pathName);	
	if (arReducDesc == NULL)
	    return 0;

	/* Open all the fields */

	pointNameFld = d4field (arReducDesc, "POINTNAME");
	if (pointNameFld == NULL)
	    return 0;

	reducMethodFld = d4field (arReducDesc, "METHOD");
	if (reducMethodFld == NULL)
	    return 0;

	reducIntervalFld = d4field (arReducDesc, "RATE");
	if (reducIntervalFld == NULL)
	    return 0;
        rateFieldSize = f4len (reducIntervalFld);

	reducArchiveFld = d4field (arReducDesc, "ARCHIVE");

	/* Read the first record */

	return d4top (arReducDesc) == 0;

    }

/*----------*/
long getArReducSize (void)

    {

	return d4reccount (arReducDesc);

    }
	
/*----------*/
int getArReducRecord (char *name, int *method, int *interval, int *archive)

    {

	char	schedName [SCHED_SIZE+1];

	if (d4eof (arReducDesc))
	    return 0;

	/* Get the name since a zero length name indicates a deleted record */
	
	stringCopy (name, f4str (pointNameFld), sizeof (PointName));
	stripBlanks (name);

	/* Skip any deleted records */
	    
	while (d4deleted (arReducDesc) || (strlen (name) == 0)) {
	    d4skip (arReducDesc, 1L);
	    stringCopy (name, f4str (pointNameFld), sizeof (PointName));
	    stripBlanks (name);
	    if (d4eof (arReducDesc))
		return 0;
	}

	*method = methodString2Int (f4str (reducMethodFld));

	/* This is for backward compatability. If the field has a length */
	/* of 2 then the interval is in minutes. Otherwise it is already */
	/* in seconds. */

	if (rateFieldSize == 2)
	    *interval = (int)(atof (f4str (reducIntervalFld)) * 60.0);
        else
	{
	    if (isalpha (*f4str (reducIntervalFld))) {
		strncpy (schedName, f4str (reducIntervalFld), SCHED_SIZE);
		schedName [SCHED_SIZE] = 0;
		*interval = GetBasicRate (schedName);
		if (*interval == 0)
		{
		    /* bad schedule name in ar_reduc file! */
		    printf ("Unknown schedule name %s in ar_reduc file\n", schedName);
		    *interval = defaultSchedule.schedule[0].interval;
		}
	    } else {
		*interval = (int)atof (f4str (reducIntervalFld));
	    }
	}

	/* If an ARCHIVE field exists, get it. Otherwise default to true. */

	if (reducArchiveFld != NULL)
	    *archive = f4true (reducArchiveFld);
	else
	    *archive = TRUE;

	d4skip (arReducDesc, 1L);

	return 1;

    }

/*----------*/
int getArReducSchedule (char *name, int *method, REDUC_SCHEDULE *schedule, int *archive)

    {

	char	schedName [SCHED_SIZE];

	/* NOTE: schedule unsupported for old style ar_reduc files */
	if (d4eof (arReducDesc) || (rateFieldSize == 2))
	    return 0;

	/* Get the name since a zero length name indicates a deleted record */
	
	stringCopy (name, f4str (pointNameFld), sizeof (PointName));
	stripBlanks (name);

	/* Skip any deleted records, and records with numeric rate fields */
	    
	while (d4deleted (arReducDesc) || (strlen (name) == 0)
	  || (!isalpha (*f4str (reducIntervalFld)))) {
	    d4skip (arReducDesc, 1L);
	    stringCopy (name, f4str (pointNameFld), sizeof (PointName));
	    stripBlanks (name);
	    if (d4eof (arReducDesc))
		return 0;
	}

	*method = methodString2Int (f4str (reducMethodFld));

	strncpy (schedName, f4str (reducIntervalFld), SCHED_SIZE);
	schedName [SCHED_SIZE] = 0;
	if (!GetSchedule (schedName, schedule))
	{
	    /* bad schedule name in ar_reduc file */
	    printf ("Unknown schedule name %s in ar_reduc file\n", schedName);
	    *schedule = defaultSchedule;
	}

	/* If an ARCHIVE field exists, get it. Otherwise default to true. */

	if (reducArchiveFld != NULL)
	    *archive = f4true (reducArchiveFld);
	else
	    *archive = TRUE;

	d4skip (arReducDesc, 1L);

	return 1;

    }

/*----------*/
void closeArReduc (void)	/* Close ar_reduc database and perform any cleanup */

    {

	d4close (arReducDesc);

    }

/*----------*/
int openArSnap (void)

    {
#ifdef ORACLESNAP

	oSnapDesc = MakeSnapDesc ("TISTREND", "TREND_DATA");
	if (oSnapDesc != NULL)
	    OpenSnapTable (oSnapDesc, &defCreateParams);
	return (oSnapDesc != NULL);

#else

	int	i;
	char	buf [20], pathName [200];
	TAG4	*t;

	/* Open the ar_snap database.  If it does not exist then create it and return */
	/* false indicating that it did not previously exist. */

	resolveLinks (ARSNAP_FILE, pathName);
	arSnapDesc = d4open (&cb, pathName);
	if (arSnapDesc == NULL) {
	    arSnapDesc = d4create (&cb, pathName, snapFields, snapTags);
	    if (arSnapDesc == NULL) {
		tisSysLog (0, 2, "Unable to create %s", pathName); 
		return 0;
	    }
	}
	t = d4tag (arSnapDesc, "METH_RATE");
	d4tag_select (arSnapDesc, t);
	d4top (arSnapDesc);
	d4unlock_all (arSnapDesc);

	/* Open up all the fields */

	pointNameFld = d4field (arSnapDesc, "POINTNAME");
	eguTagFld = d4field (arSnapDesc, "EGUTAG");
	dispLowFld = d4field (arSnapDesc, "DISP_LOW");
	dispHighFld = d4field (arSnapDesc, "DISP_HIGH");
	yearFld = d4field (arSnapDesc, "YEAR");
	dayOfYearFld = d4field (arSnapDesc, "DAYOFYEAR");
	hourFld = d4field (arSnapDesc, "HOUR");
	offsetMinFld = d4field (arSnapDesc, "OFFSET_MIN");
	sampleCntFld = d4field (arSnapDesc, "SAMPLE_CNT");
	redMethodFld = d4field (arSnapDesc, "RED_METHOD");
	rateFld = d4field (arSnapDesc, "RATE");
	rateFieldSize = f4len (rateFld);

	for (i=0 ; i<BUCKETS_PER_RECORD ; i++) {
	    sprintf (buf, "VALUE%d", i + 1); 
	    valueFld [i] = d4field (arSnapDesc, buf);
	    sprintf (buf, "STATUS%d", i + 1); 
	    statusFld [i] = d4field (arSnapDesc, buf);
	}

	return 1;

#endif ORACLESNAP
    }

/*----------*/
void closeArSnap (void)

    {
#ifdef ORACLESNAP

#else

	updateDBaseFile (arSnapDesc);
	d4close (arSnapDesc);

#endif ORACLESNAP
    }

/*----------*/
#ifndef	ORACLESNAP
static void getArRecord (ArSnapRecord *rec) /* Copy data to structure */

    {

	int	i;
	char	buf [50];
        Status  status;

	stringCopy (rec->pointName, f4str (pointNameFld), POINTNAME_SIZE+1);
	stripBlanks (rec->pointName);
	stringCopy (rec->eguTag, f4str (eguTagFld), EGUTAG_SIZE+1);
	stripBlanks (rec->eguTag);
	rec->dispLow = f4double (dispLowFld);
	rec->dispHigh = f4double (dispHighFld);
	rec->year = f4int (yearFld);
	rec->dayOfYear = f4int (dayOfYearFld);
	rec->hour = f4int (hourFld);
	rec->offsetMin = f4int (offsetMinFld);
	rec->sampleCnt = f4int (sampleCntFld);
        rec->redMethod = methodString2Int (f4str (redMethodFld));

	if (f4len (rateFld) == 2)
	    rec->rate = (int)atof (f4str (rateFld)) * 60;
	else
	    rec->rate = (int)atof (f4str (rateFld));

	for (i=0 ; i<BUCKETS_PER_RECORD ; i++) {
	    memcpy (status, f4str (statusFld [i]), sizeof (Status));
	    rec->quality [i] = status [0];
	    rec->severity [i] = letter2severity (status [1]);
	    rec->status [i] = letter2status (status [2]); /* GCT 6-Aug-1996 swap status & severity */
	    stringCopy (buf, f4str (valueFld [i]), 10);
	    stripBlanks (buf);
	    if (strlen (buf) != 0) {
		rec->value [i] = atof (buf);
		rec->count [i] = 1;
	    } else {
		rec->value [i] = 0;
		rec->count [i] = 0;
	    }
	}

    }
#endif !ORACLESNAP

/*----------*/
#ifndef	ORACLESNAP
static void putArRecord (ArSnapRecord *rec) /* Copy the internal structure to the current record */

    {
	int	i;
	char	buf [100], method [4];
        Status  status;

	d4blank (arSnapDesc);
	f4assign (pointNameFld, rec->pointName);
	f4assign (eguTagFld, rec->eguTag);
	f4assign_double (dispLowFld, rec->dispLow);
	f4assign_double (dispHighFld, rec->dispHigh);
	f4assign_int (yearFld, rec->year);
	f4assign_int (dayOfYearFld, rec->dayOfYear);
	f4assign_int (hourFld, rec->hour);
	f4assign_int (offsetMinFld, rec->offsetMin);
	f4assign_int (sampleCntFld, rec->sampleCnt);

	strcpy (method, int2methodString (rec->redMethod));
	f4assign (redMethodFld, method);

	f4assign (rateFld, rate2string (rec->rate));

	for (i=0 ; i<rec->sampleCnt ; i++) {
	    if (rec->quality [i] != 'C') {
		sprintf (buf, "%.9lf", rec->value [i]);
		f4assign (valueFld [i], buf);
	    }
	    status [0] = rec->quality [i];
	    status [1] = (rec->quality [i] != 'C') ? severity2letter (rec->severity [i]) : ' '; /* GCT 6-Aug-1996 swap status & severity */
	    status [2] = (rec->quality [i] != 'C') ? status2letter (rec->status [i]) : ' ';
	    f4assign (statusFld [i], status);    
	}

    }
#endif !ORACLESNAP

/*----------*/
int getArSnapRecord (char *pointName, int method, int interval,
		     time_t ts, ArSnapRecord *rec)

    {
#ifdef ORACLESNAP
	int	i;
        Status  status;
	SnapRec snaprec;

	if (GetSnapRecord (oSnapDesc, pointName, int2methodString (method), ts, &snaprec))
	{
	    stringCopy (rec->pointName, snaprec.pointname, POINTNAME_SIZE+1);
	    stripBlanks (rec->pointName);
	    stringCopy (rec->eguTag, snaprec.egutag, EGUTAG_SIZE+1);
	    stripBlanks (rec->eguTag);
	    rec->dispLow = snaprec.disp_low;
	    rec->dispHigh = snaprec.disp_high;
	    ts2arSnapTime (snaprec.rec_time, &rec->year, &rec->dayOfYear, &rec->hour, &rec->offsetMin);
	    rec->sampleCnt = snaprec.sample_cnt;
            rec->redMethod = methodString2Int (snaprec.red_method);
	    rec->rate = snaprec.rate;

	    for (i=0 ; i < BUCKETS_PER_RECORD ; i++) {
		memcpy (status, snaprec.status[i], sizeof (Status));
		rec->quality [i] = status [0];
		rec->severity [i] = letter2severity (status [1]);
		rec->status [i] = letter2status (status [2]);
		if (i < rec->sampleCnt) {
		    rec->value [i] = snaprec.value[i];
		    rec->count [i] = 1;
		} else {
		    rec->value [i] = 0;
		    rec->count [i] = 0;
		}
	    }
	    return 1;
	}
	else
	{
	    return 0;
	}
#else

	int	year, dayOfYear, hour, min, stat;
	char	key [100];

	/* Convert ts to discrete values so that a key can be constructed */

	ts2arSnapTime (ts, &year, &dayOfYear, &hour, &min);
	makeArSnapKey (pointName, method, interval, year, dayOfYear, hour, min, key);
	stat = d4seek (arSnapDesc, key);
	d4unlock_all (arSnapDesc);
	if (stat != 0)
	    return 0;
	getArRecord (rec);
	return 1;

#endif ORACLESNAP
    }

/*----------*/
int putArSnapRecord (ArSnapRecord *rec)

    {
#ifdef ORACLESNAP
	int	i;
        char  status[4];
	SnapRec snaprec;

	/* ignore attempts to save NULL records! */
	if (strlen (rec->pointName) == 0)
	{
		tisSysLog (0, 2, "Attempt to put empty record"); 
		return 1;
	}
	stringCopy (snaprec.pointname, rec->pointName, POINTNAME_SIZE+1);
	stringCopy (snaprec.egutag, rec->eguTag, EGUTAG_SIZE+1);
	snaprec.disp_low = rec->dispLow;
	snaprec.disp_high = rec->dispHigh;
	arSnapTime2ts (rec->year, rec->dayOfYear, rec->hour, rec->offsetMin, &snaprec.rec_time);
	snaprec.sample_cnt = rec->sampleCnt;
	stringCopy (snaprec.red_method, int2methodString (rec->redMethod), sizeof (snaprec.red_method));
	snaprec.rate = rec->rate;

	for (i=0 ; i < BUCKETS_PER_RECORD ; i++) {
	    if (i < (snaprec.sample_cnt))
	    {
		if (rec->quality [i] != 'C') {
		    snaprec.value[i] = rec->value [i];
		} else {
		    snaprec.value[i] = 0.;
		}
		status [0] = rec->quality [i];
		status [1] = (rec->quality [i] != 'C') ? severity2letter (rec->severity [i]) : ' ';
		status [2] = (rec->quality [i] != 'C') ? status2letter (rec->status [i]) : ' ';
		status [3] = 0;
	    } else {
		snaprec.value[i] = 0.;
		strcpy (snaprec.status[i], "   ");
	    }
	    stringCopy (snaprec.status[i], status, sizeof (status));
	}
	UpdateSnapRecord (oSnapDesc, &snaprec);
	return 1;
	
#else

	char		key [100];
	int	        stat;

	/* Convert ts to discrete values so that a key can be constructed */

	makeArSnapKey (rec->pointName, rec->redMethod, rec->rate,
		       rec->year, rec->dayOfYear, rec->hour,
			rec->offsetMin, key);

	/* If record currently exists in file, replace it with the new one. If the */
	/* record does not currently exist in file. We will look to see if a blank record exists */
	/* in the file and use it if we find one. Blank records also have the delete flag set */
	/* in case an offline pack is performed. */

	stat = d4seek (arSnapDesc, key);
	if (stat == 0) {	/* Found it */
	    putArRecord (rec);
	} else {	/* Not in file */
	    stat = d4top (arSnapDesc);
	    if ((stat == 0) && (*f4str (pointNameFld) == ' ')) {	/* Found a blank record */
	               putArRecord (rec);
		d4recall (arSnapDesc);
	    } else {	/* Append a new record to the file */
		d4go_eof (arSnapDesc);
		d4append_start (arSnapDesc, 0);
		putArRecord (rec);
		d4append (arSnapDesc);
	    }
	}

 	updateDBaseFile (arSnapDesc);

	return 1;

#endif ORACLESNAP
    }

/*----------*/
int openArNotesFile (void)

    {
#ifdef ORACLESNAP

	/* no-op with oracle DB */

#else

	char	pathName [200];
	TAG4	*t;

	/* Open the ar_snapn database. */

	resolveLinks (ARSNAPN_FILE, pathName);
	arNotesDesc = d4open (&cb, pathName);
	if (arNotesDesc == NULL)
	    return 0;

	t = d4tag (arNotesDesc, "MEMO_REC");
	d4tag_select (arNotesDesc, t);
	d4top (arNotesDesc);
	d4unlock_all (arNotesDesc);

	/* Open up all the fields */

	notesNameFld = d4field (arNotesDesc, "POINTNAME");
	notesYearFld = d4field (arNotesDesc, "YEAR");
	notesDayOfYearFld = d4field (arNotesDesc, "DAYOFYEAR");
	notesHourFld = d4field (arNotesDesc, "HOUR");
	notesMinFld = d4field (arNotesDesc, "MINUTE");
	notesSecFld = d4field (arNotesDesc, "SECOND");
	notesNoteFld = d4field (arNotesDesc, "NOTE");

	return 1;

#endif ORACLESNAP
    }

/*----------*/
void closeArNotesFile (void )

    {
#ifdef ORACLESNAP

	/* no-op with oracle DB */

#else

	updateDBaseFile (arNotesDesc);
	d4close (arNotesDesc);

#endif ORACLESNAP
    }

/*----------*/
#ifndef	ORACLESNAP
void makeArSnapKey (char *pointName, int method, int interval,
		    int year, int dayOfYear, int hour, int min,
		    char *key)
    
    {
	sprintf (key, "%*d%*d%*d%*d%-*s%*s%-*s",
			YEAR_SIZE, year,
			DAY_SIZE, dayOfYear,
			HOUR_SIZE, hour,
			MIN_SIZE, min,
			POINTNAME_SIZE, pointName,
			METHOD_SIZE, int2methodString (method),
			rateFieldSize, rate2string (interval));

    }
#endif !ORACLESNAP

/*----------*/
void arSnapTime2ts (int year, int dayOfYear, int hour, int min, time_t *ts)

    {

#ifndef AXP
	extern time_t	timezone;
#endif AXP
	int		i, leap, day;
	static int	day_tab[2][13] = {
		{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
		{31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}
	};
	struct tm	tim;

	leap = (( year % 4 ) == 0 ) && (( year % 100 ) != 0 ) || (( year % 400 ) == 0 );

	day = dayOfYear + 1;
	for( i=0 ; day > day_tab [leap][i] ; i++ )
		day -= day_tab [leap][i];

	tim.tm_year = year - 1900;
	tim.tm_yday = dayOfYear;
	tim.tm_hour = hour;
	tim.tm_min = min;
	tim.tm_sec = 0;
	tim.tm_mon = i;
	tim.tm_mday = day;
	tim.tm_isdst = 0;
#if !(defined(SOLARIS) || defined(WIN32))
	tim.tm_gmtoff = 0;
#endif

#if (defined(WIN32) || defined(MIPSEL) || defined(SOLARIS) || defined (AXP))
	*ts = mktime (&tim) - timezone;
#else
	*ts = timegm (&tim);
#endif

    }

/*----------*/
void arNotesTime2ts (int year, int dayOfYear, int hour, int min,
		     int sec, time_t *ts)

    {

#ifndef AXP
	extern time_t	timezone;
#endif AXP
	int		i, leap, day;
	static int	day_tab[2][13] = {
		{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
		{31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}
	};
	struct tm	tim;

	leap = (( year % 4 ) == 0 ) && (( year % 100 ) != 0 ) || (( year % 400 ) == 0 );

	day = dayOfYear + 1;
	for( i=0 ; day > day_tab [leap][i] ; i++ )
		day -= day_tab [leap][i];

	tim.tm_year = year - 1900;
	tim.tm_yday = dayOfYear;
	tim.tm_hour = hour;
	tim.tm_min = min;
	tim.tm_sec = sec;
	tim.tm_mon = i;
	tim.tm_mday = day;
	tim.tm_isdst = 0;
#if !(defined(SOLARIS) || defined(WIN32))
	tim.tm_gmtoff = 0;
#endif

#if (defined(WIN32) || defined(MIPSEL) || defined(SOLARIS) || defined (AXP))
	*ts = mktime (&tim) - timezone;
#else
	*ts = timegm (&tim);
#endif

    }

/*----------*/
void ts2arSnapTime (time_t ts, int *pYear, int *pDayOfYear, int *pHour, int *pMin)

    {

	struct tm	*pTm;

	/* time_t to struct tm */

	pTm = gmtime (&ts);

	*pYear = pTm->tm_year + 1900;
	*pDayOfYear = pTm->tm_yday;
	*pHour = pTm->tm_hour;
	*pMin = pTm->tm_min;
    }

/*----------*/
#ifndef	ORACLESNAP
static char *rate2string (int rate)

    {
        static  char    str [10];

        /* If we are operating in compatability mode then rate is expressed */
	/* in minutes. Otherwise the rate is expressed in seconds. */

	if (f4len (rateFld) == 2) {
	  if (rate == 30) {
	    strcpy (str, ".5");
	  } else {
	    sprintf (str, "%d", rate / 60);
	  }
	} else {
	  sprintf (str, "%d", rate);
	}

	return str;

    }
#endif !ORACLESNAP

/*----------*/
void updateDBaseFile (DATA4 *data)

    {
#ifdef ORACLESNAP

	/* no-op with oracle DB */

#else
	d4flush (data);
	d4unlock (data);

#endif ORACLESNAP
    }

#ifdef TRENDUPDATES

#include "trendupdate.h"

/* find and install update info for the current time */
int GetUpdates (PointNodePtr   (findPoint) (char *))
{
    time_t	recordTime;
    time_t	fileTime;
    struct tm	*tm_struct;
    char	fileName [13];		/* 8.3 file name size */
    char	journalName [100];	/* path + 8.3 file name size */
    char	pathName [200];		/* resolved links file name */
    int		journalValueIx;
    char	pointName [POINTNAME_SIZE + 1];
    short	dataIx;
    PointNodePtr p;
    TrendPtr	t;
    D4DATA	*journal;	/* Codebase file descriptor */
    char	updateStatus;
    char	seekpt [11];
    int		methIndex = methodString2Int ("SMP");

    recordTime = recordStartTimeA (300, time (0));

    fileTime = prevMidnite (recordTime);
    tm_struct = gmtime (&fileTime);

    /* Open journal file (return if none)*/
    sprintf (fileName, "TU%02.2d%02.2d%02.2d",
		tm_struct->tm_mon+1, tm_struct->tm_mday, tm_struct->tm_year % 100);
    sprintf (journalName, "%s/%s", JOURNAL_PATH, fileName);
    /* Open the database.  If it does not exist then create it */
    resolveLinks (journalName, pathName);
    journal = d4open (&cb, pathName);
    if (journal == NULL)
    {
	return;
    }

    /* locate any records with current information */
    d4tag_select (journal, d4tag (journal, "TIME"));
    d4top (journal);
    sprintf (seekpt, "%10d", recordTime);
    if (d4seek (journal, seekpt) != r4success)
    {
	d4close (journal);
	return 0;
    }
    /* Get base value field */
    journalValueIx = d4field_number (journal, "VALUE1");
    while (!d4eof (journal))
    {
	if (f4long (d4field (journal, "TIME")) == recordTime)
	{
	    /* this is a current record! Process any new info */
	    strcpy (pointName, f4str (d4field (journal, "POINTNAME")));

	    /* Lookup the point in the tree. If found process sample trend */
	    stripBlanks (pointName);
	    p = findPoint (pointName);
	    if (p != NULL)
	    {
		t = p->trend [methIndex];
		if (t != NULL && (f4long (d4field (journal, "TIME")) == t->startTime))
		{
		    for (dataIx = 0; dataIx < 60; dataIx++)
		    {
			updateStatus = *f4str (d4field_j (journal, journalValueIx + (dataIx * 2) + 1));
			if (updateStatus != ' ')
			{
			    /* update into current record!  (only overwrite C or F data) */
			    if ((t->sampleCnt <= dataIx) || (t->quality [dataIx] == 'C')
			      || (t->quality [dataIx] == 'F') || (t->quality [dataIx] == 'Q')
			      || (t->quality [dataIx] == 'q'))
			    {
				t->value [dataIx]
				   = f4double (d4field_j (journal, journalValueIx + (dataIx * 2)));
				t->quality [dataIx] = '!';
				t->severity [dataIx] = -1;
				t->status [dataIx] = -1;
			    }
			}
	    	    }
	        }
	    }
	}
	d4skip (journal, 1);
    }
    d4unlock_all (journal);
    d4close (journal);
    return 0;
}

#endif TRENDUPDATES
