/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/* Write the key for each record to stdout */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#ifdef	WIN32
#  include <process.h>
#endif	WIN32
/*#include <syslog.h>*/
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32
#include "mlist.h"
#include "trendDefs.h"
#include "msgMon.h"

C4CODE  cb;
D4DATA	*arSnapDesc;
F4FIELD	*pointNameFld, *eguTagFld, *dispLowFld, *dispHighFld, *yearFld, *dayOfYearFld, *hourFld;
F4FIELD	*offsetMinFld, *sampleCntFld, *redMethodFld, *rateFld, *valueFld [60], *statusFld [60];  

	void	stripBlanks ();
static	void	openSnap ();
static	void	closeSnap ();

extern	C4CODE	cb;

/*----------*/
void main (int argc, char *argv[])

    {

	int		year, dayOfYear, hour, min, stat, sampleCnt, cnt;
	char		pointName [50], method [40];
	char		recPointName [50], recMethod [4];
	char		key [50], value [10], status [4];
	double          recInterval, interval;

	tisOpenLog ("dumpSnap", 6, 2, NULL, NULL);

	if (argc != 5) {
	    printf ("Usage: dumpSnap <fileName> <pointName> <reductionMethod> <reductionRate>\n");
	    printf ("    - for <fileName> means /tis4000db/ar_snap\n");
	    exit (-1);
	}

	/* Get point name, method, and interval from command line */

	strcpy (pointName, argv [2]);
	strcpy (method, argv [3]);
	for (cnt=0 ; cnt<(int)(strlen (method)) ; cnt++)
	    method [cnt] = toupper (method [cnt]);
	interval = atof (argv [4]);

	d4init (&cb);
#if defined(USECB63) || defined(WIN32) 
	cb.errOpen = 0;
	cb.errCreate = 0;
#else
	cb.open_error = 0;
	cb.create_error = 0;
#endif USECB63

#ifdef WIN32
	openSnap ((strcmp (argv [1], "-") == 0) ? "\\tis4000db\\ar_snap" : argv [1]);
#else
	openSnap ((strcmp (argv [1], "-") == 0) ? "/tis4000db/ar_snap" : argv [1]);
#endif WIN32

	/* Seek to the first record that matches our pointName and method */

	sprintf (key, "%-*s%-*s", POINTNAME_SIZE, pointName, METHOD_SIZE, method);
	if ((stat = d4seek (arSnapDesc, key)) != 0) {
	    tisSysLog (0,0,"point %s not found in file.\n", pointName);
	    closeSnap (arSnapDesc);
	    exit (-1);
	}

	while (!d4eof (arSnapDesc)) {

	    /* See if this is a record we want */

	    recInterval = atof (f4str (rateFld));
	    if (recInterval == interval) {

		/* Get the date and time for the record */

		year = f4int (yearFld);
		dayOfYear = f4int (dayOfYearFld);
		hour = f4int (hourFld);
		min = f4int (offsetMinFld);

		/* Dump out all the buckets */

		sampleCnt = f4int (sampleCntFld);
		for (cnt=0 ; cnt < sampleCnt ; cnt++) {
	
		    strcpy (value, f4str (valueFld [cnt]));
		    strcpy (status, f4str (statusFld [cnt]));

		    printf ("%-*d %-*d %-*d %-*d %-2d %-*s %-*s\n",
				YEAR_SIZE, year,
				DAY_SIZE, dayOfYear,
				HOUR_SIZE, hour,
				MIN_SIZE, min,
				cnt + 1,
				9, value,
				3, status);

		}

	    }

	    stat = d4skip (arSnapDesc, 1L);

	    /* If we are not at the correct pointName and method then we have found */
	    /* all the record of interest. */

	    strcpy (recPointName, f4str (pointNameFld));
	    stripBlanks (recPointName);
	    strcpy (recMethod, f4str (redMethodFld));
	    if (strcmp (recMethod, method) != 0)
		break;
	    if (strcmp (recPointName, pointName) != 0)
		break;
	
	}

	closeSnap ();

    }

/*----------*/
static void openSnap (fileName)

    char	*fileName;

    {

	int	i;
	char	buf [20];
	T4TAG	*t;

	/* Open the ar_snap database.  If it does not exist then create it and return */
	/* false indicating that it did not previously exist. */

	arSnapDesc = d4open (&cb, fileName);
	if (arSnapDesc == NULL) {
	    tisSysLog (0,0,"Unable to open %s.\n", fileName);
	    exit (1);
	}
	t = d4tag (arSnapDesc, "METHOD_REC");
	d4tag_select (arSnapDesc, t);
	d4top (arSnapDesc);
	d4unlock_all (arSnapDesc);

	/* Open up all the fields */

	pointNameFld = d4field (arSnapDesc, "POINTNAME");
	eguTagFld = d4field (arSnapDesc, "EGUTAG");
	dispLowFld = d4field (arSnapDesc, "DISP_LOW");
	dispHighFld = d4field (arSnapDesc, "DISP_HIGH");
	yearFld = d4field (arSnapDesc, "YEAR");
	dayOfYearFld = d4field (arSnapDesc, "DAYOFYEAR");
	hourFld = d4field (arSnapDesc, "HOUR");
	offsetMinFld = d4field (arSnapDesc, "OFFSET_MIN");
	sampleCntFld = d4field (arSnapDesc, "SAMPLE_CNT");
	redMethodFld = d4field (arSnapDesc, "RED_METHOD");
	rateFld = d4field (arSnapDesc, "RATE");
	for (i=0 ; i<60 ; i++) {
	    sprintf (buf, "VALUE%d", i + 1); 
	    valueFld [i] = d4field (arSnapDesc, buf);
	    sprintf (buf, "STATUS%d", i + 1); 
	    statusFld [i] = d4field (arSnapDesc, buf);
	}

    }

/*----------*/
static void closeSnap ()

    {

	d4close (arSnapDesc);

    }

/*----------*/
void stripBlanks (buf)	/* Strip blanks from a string */

    char	*buf;

    {

	while (*buf) {
	    if (*buf == ' ') {
		*buf = '\0';
		return;
	    }
	    buf++;
	}
	
    }

