/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
   File:   mergeupdates.c - trend update server, records data from
		 IOC historic trend function.

   Author: Grant Irani
   Date:   10/27/97
*/

#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include <fcntl.h>
#include <dirent.h>

#if !(defined(MIPSEL) || defined(AXP))
#include "memory.h"
#else
#include <sys/types.h>
#include <sys/stat.h>
extern int lstat ();
extern int readlink ();
#endif

#include <pwd.h>
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "trendSizes.h"
#include "trendupdate.h"
#include "msgMon.h"

#define LOG_OPTIONSI	(LOG_INFO | LOG_LOCAL1)
#define LOG_OPTIONSD	(LOG_DEBUG | LOG_LOCAL1)
#define LOG_OPTIONSE	(LOG_ERR | LOG_LOCAL1)

static	C4CODE	cb;		/* Codebase status configuration block */
static	D4DATA	*data;		/* Codebase file descriptors */
static	D4DATA	*journal;	/* Codebase file descriptors */
static	time_t	journalTime;
static	char	FilePath[40];

int		myPid;

int		remote;		/* indicate whether remote or local merge */
int		reset;		/* indicate whether reset merged flags */

static int ReadUpdates (void);
static int ResetUpdates (void);

int main (int argc, char **argv)
{

    tisOpenLog ("mergeupdates", 6, 2, NULL, NULL);

    remote = 0;
    reset = 0;
    if (argc == 2 && strcmp(argv[1], "remote") == 0)
	remote = 1;
    else if (argc == 2 && strcmp(argv[1], "reset") == 0)
	reset = 1;

    d4init( &cb );		/* Initialize the Codebase settings. */
#ifdef USECB63
    cb.errGo = 0;
    cb.errCreate = 0;
    cb.errOpen = 0;
    cb.errTagName = 0;
    cb.errFieldName = 0;
#else
    cb.go_error = 0;
    cb.create_error = 0;
    cb.open_error = 0;
    cb.tag_name_error = 0;
    cb.field_name_error = 0;
#endif USECB63

    myPid = getpid ();

#ifdef WIN32
    data = d4open (&cb, "\\tis4000db\\ar_snap.dbf");
#else
    data = d4open (&cb, "/tis4000db/ar_snap.dbf");
#endif WIN32
    if (data == 0)
    {
    	tisSysLog ("Can't open /tis4000db/ar_snap.dbf\n");
    }
    else
    {
	d4tag_select (data, d4tag (data, "METHOD_REC"));
	if (!reset)
	    ReadUpdates ();
	else
	    ResetUpdates ();
    }
    closelog ();
}

static int ReadUpdates (void)
{
    short	dataIx, i;
    int		dataLen;
    int		journalValueIx;
    double 	trendvalue;
    int		appending;
    time_t	snapTime;
    time_t	recTime;
    time_t	fileTime;
    char	pointName [POINTNAME_SIZE];
    char	journalName [100];	/* path + 8.3 file name size */
    char	fileName [13];	/* 8.3 file name size */
    char	seekpt [POINTNAME_SIZE+3+4+3+2+2+1]; /* enough to hold search with time info */
    char	updateStatus;
    struct tm	*tm_struct;
    DIR		*journalDir;
    struct dirent *dirEnt;
    int		dataValueIx;
    int		recCnt;
    char	recQuality;

    /* scan all files in update directory */
    if (journalDir = opendir(JOURNAL_PATH))
    {
	while (dirEnt = readdir(journalDir))
	{
	    if ((strlen (dirEnt->d_name) == 12)
	      && (strncmp (dirEnt->d_name, "TU", 2) == 0)
	      && (strchr (dirEnt->d_name, '.'))
	      && (strcmp (strchr (dirEnt->d_name, '.'), ".dbf") == 0))
	    {
		/* should be a journal file! */
		sprintf (journalName, "%s%s", JOURNAL_PATH, dirEnt->d_name);
		journal = d4open (&cb, journalName);
		if (journal)
		{
		    d4tag_select (journal, d4tag (journal, "POINT_TIME"));
		    d4top (journal);
		    /* read journal, update and mark updated info */
		    journalValueIx = d4field_number (journal, "VALUE1");
		    while (!d4eof (journal))
		    {
			strcpy (pointName, f4str (d4field (journal, "POINTNAME")));
			recTime = f4long (d4field (journal, "TIME"));
			/* prescan for any updates required in record */
			for (dataIx = 0; dataIx < 60; dataIx++)
			{
			    updateStatus = *f4str (d4field_j (journal, journalValueIx + (dataIx * 2) + 1));
 			    if (updateStatus == '0' || (remote && updateStatus == '1'))
				break;	/* local update required! */
			}
			if (dataIx < 60)
			{
			    /* seek in DB, and update record */
			    tm_struct = gmtime (&recTime);
			    sprintf (seekpt, "%-*s%3s%4d%3d%2d%2d", POINTNAME_SIZE, pointName, "SMP",
				     tm_struct->tm_year + 1900, tm_struct->tm_yday, tm_struct->tm_hour, tm_struct->tm_min);
			    if (d4seek (data, seekpt) != r4success)
			    {
				/* New record required! */
				appending = 1;
				d4append_start (data, 0);
				d4blank (data);
				f4assign (d4field (data, "POINTNAME"), pointName);
				f4assign_int (d4field (data, "YEAR"), tm_struct->tm_year + 1900);
				f4assign_int (d4field (data, "DAYOFYEAR"), tm_struct->tm_yday);
				f4assign_int (d4field (data, "HOUR"), tm_struct->tm_hour);
				f4assign_int (d4field (data, "OFFSET_MIN"), tm_struct->tm_min);
				f4assign_int (d4field (data, "RATE"), 300);
				f4assign_long (d4field (data, "SAMPLE_CNT"), 0);
				/* Don't know what to do to get following fields! */
				f4assign (d4field (data, "EGUTAG"), f4str (d4field (journal, "EGUTAG"))); 
				f4assign_double (d4field (data, "DISP_LOW"),
						 f4double (d4field (journal, "DISP_LOW"))); 
				f4assign_double (d4field (data, "DISP_HIGH"),
						 f4double (d4field (journal, "DISP_HIGH"))); 
				f4assign (d4field (data, "RED_METHOD"), "SMP"); 
				dataValueIx = d4field_number (data, "VALUE1");
				for (i = 0; i < dataIx; i++)
				{
				    /* fill start with bad data status */
				    f4assign (d4field_j (data, dataValueIx + (dataIx * 2)), "0");
				    f4assign (d4field_j (data, dataValueIx + (dataIx * 2) + 1), "C  ");
				}
			    }
			    else
			    {
				appending = 0;
			    }
			    /* remember number of values in current target record */
			    recCnt = f4long (d4field (data, "SAMPLE_CNT"));
			    for (; dataIx < 60; dataIx++)
			    {
			        updateStatus = *f4str (d4field_j (journal, journalValueIx + (dataIx * 2) + 1));
				if (updateStatus == '0' || (remote && updateStatus == '1'))
				{
				    dataValueIx = d4field_number (data, "VALUE1");
				    /* Get quality value from current target record */
				    recQuality = *f4str (d4field_j (data, dataValueIx + (dataIx * 2) + 1));
				    /* if adding to length of record or record value is poor quality */
				    if (recCnt < dataIx || recQuality == 'C' || recQuality == 'F')
				    {
					/* update with value from journal */
					f4assign (d4field_j (data, dataValueIx + (dataIx * 2)),
					     f4str (d4field_j (journal, journalValueIx + (dataIx * 2))) );
					f4assign (d4field_j (data, dataValueIx + (dataIx * 2) + 1),
								 "!  "); /* update status */
				    }
				    /* Indicate processed locally in journal */
				    if (remote)
					f4assign (d4field_j (journal, journalValueIx + (dataIx * 2) + 1), "2");
				    else
					f4assign (d4field_j (journal, journalValueIx + (dataIx * 2) + 1), "1");
				    /* if extending record, backfill C quality, and adjust record length */
				    if (recCnt <= dataIx)
				    {
					for (i = recCnt; i < dataIx; i++)
					    f4assign (d4field_j (data, dataValueIx + (i * 2) + 1),
								 "C  "); /* update status for gap! */
					f4assign_long (d4field (data, "SAMPLE_CNT"), dataIx + 1);
					recCnt = dataIx + 1;
				    }
				}
			    }
			    if (appending)
			    {
				d4append (data);
			    }
		            d4flush_all (data);
		            d4flush_all (journal);
			}
			d4skip (journal, 1);
		    }
		    d4close (journal);
		}
	    }
	}
	closedir(journalDir);
    }
}

static int ResetUpdates (void)
{
    short	dataIx, i;
    int		dataLen;
    int		journalValueIx;
    double 	trendvalue;
    int		appending;
    time_t	snapTime;
    time_t	recTime;
    time_t	fileTime;
    char	pointName [POINTNAME_SIZE];
    char	journalName [100];	/* path + 8.3 file name size */
    char	fileName [13];	/* 8.3 file name size */
    char	seekpt [POINTNAME_SIZE+3+4+3+2+2+1]; /* enough to hold search with time info */
    char	updateStatus;
    struct tm	*tm_struct;
    DIR		*journalDir;
    struct dirent *dirEnt;
    int		dataValueIx;

    /* scan all files in update directory */
    if (journalDir = opendir(JOURNAL_PATH))
    {
	while (dirEnt = readdir(journalDir))
	{
	    if ((strlen (dirEnt->d_name) == 12)
	      && (strncmp (dirEnt->d_name, "TU", 2) == 0)
	      && (strchr (dirEnt->d_name, '.'))
	      && (strcmp (strchr (dirEnt->d_name, '.'), ".dbf") == 0))
	    {
		/* should be a journal file! */
		sprintf (journalName, "%s%s", JOURNAL_PATH, dirEnt->d_name);
		journal = d4open (&cb, journalName);
		if (journal)
		{
		    d4tag_select (journal, d4tag (journal, "POINT_TIME"));
		    d4top (journal);
		    /* read journal, update and mark updated info */
		    journalValueIx = d4field_number (journal, "VALUE1");
		    while (!d4eof (journal))
		    {
			/* Scan for any updates required in record */
			for (dataIx = 0; dataIx < 60; dataIx++)
			{
			    updateStatus = *f4str (d4field_j (journal, journalValueIx + (dataIx * 2) + 1));
 			    if ((updateStatus == '1') || (updateStatus == '2'))
			    {
				/* Indicate not recorded locally in journal */
				f4assign (d4field_j (journal, journalValueIx + (dataIx * 2) + 1), "0");
			    }
			}
		        d4flush_all (journal);
			d4skip (journal, 1);
		    }
		    d4close (journal);
		}
	    }
	}
	closedir(journalDir);
    }
}
