/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/* General purpose utility for manipulating trending files.
/*	- determining ar_snap growth behavior (-s)
/*	- packing trending files (p)
/*	- determining free space in the trend file (-f)
*/


#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <syslog.h>
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "mlist.h"
#include "trendDefs.h"
#include "utils.h"
#include "arSnapRecord.h"
#include "dbase.h"

extern	D4DATA	*arSnapDesc, *arNotesDesc;
extern	F4FIELD	*pointNameFld;
 

/* Functions defined in this module */

static	int	snapSize (void);
static	int	packFiles (void);
static	int	findSpace (void);

int     myPid;

/*----------*/
main (int argc, char *argv[])

    {

	char	c;

        myPid = getpid ();

	initCodebase ();

	while ((c = getopt (argc, argv, "spf")) != -1)
	    switch (c) {
		case 's':	/* Determine file growth characteristics */
		    return snapSize ();
		case 'p':	/* Pack trending files */
		    return packFiles ();
		case 'f':	/* Determine ar_snap utilization */
		    return findSpace ();
	    }

	return 0;

    }

/*----------*/
static int snapSize ()

    {
	PointName	name;
	int		method, interval, archive, records;
	long		size, recSize;

	if (!openArReduc ())
	    return -1;
	if (!openArSnap ())
	    return -1;
	recSize = d4record_width (arSnapDesc);    
	closeArSnap ();

	size = 0;
	while (getArReducRecord (name, &method, &interval, &archive)) {
	    records = getRecordsPerDay (interval);
	    size += records * recSize;
	}

	closeArReduc ();
  
	printf ("%ld\n", size);

	return 0;

    }

/*----------*/
static int packFiles (void)

    {

	int	status;

	/* Pack the ar_snap file */

	openArSnap ();
	status = d4pack (arSnapDesc);
	closeArSnap ();
	if (status != 0)
	    return -1;

	/* pack the notes file if it exists */

	if (!openArNotesFile ())
	    return 0;
	status = d4pack (arNotesDesc);
	closeArNotesFile ();
	return status == 0 ? 0 : -1;

    }

/*----------*/
static int findSpace (void)

    {

	int		stat;
	long		totalRecs, freeRecs;
        PointName	pointName;

	if (!openArSnap ())
	    return -1;

	totalRecs = d4reccount (arSnapDesc);
	freeRecs = 0;
	while (!d4eof (arSnapDesc)) {

	    /* Count over deleted entries. */

	    strcpy (pointName, f4str (pointNameFld));
	    if (d4deleted (arSnapDesc) || (pointName [0] == ' ')) {
		stat = d4skip (arSnapDesc, 1L);
		freeRecs++;;
	    } else
		break;
	}

	closeArSnap ();

	printf ("%ld %ld\n", totalRecs, freeRecs);

	return 0;

    }



