/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
   File:   trendupdate.c - trend update server, records data from
		 IOC historic trend function.

   Author: Grant Irani
   Date:   10/27/97
*/

#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include <fcntl.h>

#include "d4all.h"

#if !(defined(MIPSEL) || defined(AXP))
#include "memory.h"
#else
extern int lstat ();
extern int readlink ();
#endif
#include <sys/types.h>
#include <sys/stat.h>

#include "server.h"
#include "dbDefs.h"
#include "trendSizes.h"
#include "trendupdate.h"
#include "trendupdatemsgs.h"
#include "plcTrendLib.h"

#define LOG_OPTIONSI	(LOG_INFO | LOG_LOCAL1)
#define LOG_OPTIONSD	(LOG_DEBUG | LOG_LOCAL1)
#define LOG_OPTIONSE	(LOG_ERR | LOG_LOCAL1)

static	C4CODE	cb;		/* Codebase status configuration block */
static	D4DATA	*journal;	/* Codebase file descriptors */
static	time_t	journalTime;
static	char	FilePath[40];

SERVER	*server;		/* the server control structure */
int		myPid;
UpdateRspUnion	rspmsg;

static int GetUpdateResp (UpdateMsg *reqmsg);
static int create_subdir (void);


int main (argc, argv)
int	argc;
char**	argv;
{
#ifdef TEST
    UpdateMsg	testMsg;
    char	pointName [POINTNAME_SIZE];
    short	value;
    short	dataIx;
    short	bit;
    double 	trendvalue;
    double	hopr;
    double	lopr;
    char	eguTag [EGUTAG_SIZE];
#endif TEST
    UpdateMsgUnion *reqmsg;
    int		msglen;

    d4init( &cb );		/* Initialize the Codebase settings. */
#ifdef USECB63
    cb.errGo = 0;
    cb.errCreate = 0;
    cb.errOpen = 0;
    cb.errTagName = 0;
    cb.errFieldName = 0;
#else
    cb.go_error = 0;
    cb.create_error = 0;
    cb.open_error = 0;
    cb.tag_name_error = 0;
    cb.field_name_error = 0;
#endif USECB63

#ifdef TEST
    if (argc > 1)
    {
	for (value = 0; value < 4096; value++)
	{
	    printf ("Value %d(%x) ", value, value);
	    strcpy (pointName, "nwwaioca");
	    if (StartUpdateGroup (pointName, 1, 1))
	    {
		while (NextUpdatePoint (pointName, &dataIx, &bit))
		{
		    if ( bit == -1) /* analog */
		    {
			trendvalue = GetTrendData (pointName,
						 value,
						 PCT_ANALOG,
						 &hopr,
						 &lopr,
						 eguTag);
			printf ("A %s    => %f ", pointName, trendvalue);
		    }
		    else /* digital */
		    {
			trendvalue = GetTrendData (pointName,
					 	 (value & (0x8000 >> (bit-1))) != 0,
						 PCT_DIGITAL,
						 &hopr,
						 &lopr,
						 eguTag);
			printf ("D %s(%d) => %f ", pointName, bit, trendvalue);
		    }
		}
	    }
	    printf ("\n");
	}
    }
    else
    {
	testMsg.type = htons (UPDATE_INFO);
	testMsg.dataLen = htons (3);
	testMsg.snapTime = htonl (time (0) - (300 * 2));
	testMsg.line = htons (1);
	testMsg.address = htons (1);
	value = (ntohl (testMsg.snapTime) / 300) & 0xfff;
	testMsg.data[0] = htons (value);
	testMsg.data[1] = htons ((short)(value + 1));
	testMsg.data[2] = htons ((short)(value + 2));
	testMsg.data[3] = htons ((short)(value + 3));
	testMsg.data[4] = htons ((short)(value + 4));
	testMsg.data[5] = htons ((short)(value + 5));
	testMsg.data[6] = htons ((short)(value + 6));
	testMsg.data[7] = htons ((short)(value + 7));
	testMsg.data[8] = htons ((short)(value + 8));
	testMsg.data[9] = htons ((short)(value + 9));
	testMsg.data[10] = htons ((short)(value + 10));
	testMsg.data[11] = htons ((short)(value + 11)); /* this word is digitals */
	strcpy ((char *)&testMsg.data[3], "nwwaioca");

	msglen = GetUpdateResp (&testMsg);
	printf ("Finished, response len = %d\n", msglen);
    }
    return;
#endif TEST

    myPid = getpid ();

    /* open server as synchronous I/O */
    if ((server = OpenServer ("TrendUpdate", TRUE, NULL)) == NULL)
	exit (0);

    syslog(LOG_OPTIONSI, "trendupdate(%d): Trend update server running", myPid);

#ifdef DEBUG
    SetSocketNum (server, 8002);
#endif DEBUG

    while (reqmsg = GetReq (server))
    {
#ifndef DEBUG
	syslog	(LOG_OPTIONSD, "trendupdate(%d): Processing request %d",
		 myPid, ((UpdateMsg *)reqmsg)->type);
#endif
	((UpdateMsg *)reqmsg)->type = ntohs (((UpdateMsg *)reqmsg)->type);
	switch (((UpdateMsg *)reqmsg)->type)
	{
	    case UPDATE_INFO:
		msglen = GetUpdateResp ((UpdateMsg *)reqmsg);
		break;
	}
	SendRsp (server, (u_char *)&rspmsg, msglen);
	free (reqmsg);
    }
#ifndef DEBUG
    syslog(LOG_OPTIONSD, "trendupdate(%d): Null message from GetReq", myPid);
#endif

    /* cleanup any open files, etc... */
    CloseSession (server);
    closelog ();
}

static int GetUpdateResp (UpdateMsg *reqmsg)
{
    UpdateRsp	*updrsp = (UpdateRsp *)&rspmsg;
    short	dataIx;
    int		dataLen;
    int		fieldNum;
    short	bit;
    double 	trendvalue;
    int		appending;
    time_t	snapTime;
    time_t	recTime;
    time_t	fileTime;
    int		line;
    int		address;
    char	iocName [IOCNAME_SIZE];
    char	pointName [POINTNAME_SIZE];
    char	journalName [100];	/* path + 8.3 file name size */
    char	fileName [13];	/* 8.3 file name size */
    char	pathName [200];		/* resolved links file name */
    char	seekpt [POINTNAME_SIZE+10+1]; /* enough to hold search with time info */
    char	valbuf [10];
    struct tm	*tm_struct;
    double	hopr;
    double	lopr;
    char	eguTag [EGUTAG_SIZE];

    /* unpack message data */
    dataLen = ntohs (reqmsg->dataLen);
    snapTime = ntohl (reqmsg->snapTime);
    line = ntohs (reqmsg->line);
    address = ntohs (reqmsg->address);
    strcpy (iocName, (char *)&reqmsg->data[dataLen]);
#ifdef TEST
    tm_struct = gmtime (&snapTime);
    printf ("Update report from %s - %2.2d/%2.2d/%2.2d %2.2d:%2.2d:%2.2d\n",
		iocName,
		tm_struct->tm_mon+1, tm_struct->tm_mday, tm_struct->tm_year % 100,
		tm_struct->tm_hour, tm_struct->tm_min, tm_struct->tm_sec);
#endif TEST
    
    /* place converted values in to journal file */
    recTime = recordStartTimeA (300, snapTime);
    fileTime = prevMidnite (snapTime);
    tm_struct = gmtime (&fileTime);

    /* build response message, assuming the worst */
    updrsp->type = ntohs (UPDATE_INFO);
    updrsp->value = ntohs (0);

    /* set up in correct journal file */
    if (journalTime && journalTime != fileTime)
    {
	/* if current journal file open, but for wrong time, close it */
	d4close (journal);
	journalTime = 0;
    }
    if (journalTime == 0)
    {
	/* Open journal file (create if necessary)*/
	sprintf (fileName, "TU%02.2d%02.2d%02.2d",
		tm_struct->tm_mon+1, tm_struct->tm_mday, tm_struct->tm_year % 100);
	sprintf (journalName, "%s%s", JOURNAL_PATH, fileName);
	/* Open the database.  If it does not exist then create it */
	resolveLinks (journalName, pathName);
	journal = d4open (&cb, pathName);
	if (journal == NULL)
	{
	    if (create_subdir ())
	    {
		journal = d4create (&cb, pathName, journalFields, journalTags);
		if (journal == NULL) {
		    syslog (LOG_OPTIONSE, "trendupdate(%d): Unable to create %s",
			myPid, pathName);
		}
	    }
	}
	d4tag_select (journal, d4tag (journal, "POINT_TIME"));
	d4top (journal);
	d4unlock_all (journal);
    }

    /* write journal */
    if (journal)
    {
	if (StartUpdateGroup (iocName, line, address))
	{
	    while (NextUpdatePoint (pointName, &dataIx, &bit))
	    {
		if (dataIx >= dataLen)
		{  /* dataIx more than report provides! */
		    syslog (LOG_OPTIONSE, "trendupdate(%d): IOC %s inconsistent data definition.",
			myPid, iocName); 
		    continue; /* continue processing! */
		}

		if ( bit == -1) /* analog */
		{
		    trendvalue = GetTrendData (pointName,
						 ntohs(reqmsg->data [dataIx]),
						 PCT_ANALOG,
						 &hopr,
						 &lopr,
						 eguTag);
#ifdef TEST
		    printf ("Storing analog %s - raw data=%d, converted data = %f\n",
				 pointName, ntohs(reqmsg->data [dataIx]), trendvalue);
#endif TEST
		}
		else /* digital */
		{
		    trendvalue = GetTrendData (pointName,
					 	 (ntohs(reqmsg->data [dataIx]) & (0x8000 >> (bit-1))) != 0,
						 PCT_DIGITAL,
						 &hopr,
						 &lopr,
						 eguTag);
#ifdef TEST
		    printf ("Storing digital %s - raw data=%x bit# %d value %d, converted data = %f\n",
				 pointName, ntohs(reqmsg->data [dataIx]), bit,
				 (ntohs(reqmsg->data [dataIx]) & (0x8000 >> (bit-1))) != 0,
				 trendvalue);
#endif TEST
		}
		/* Store trendvalue in trend data as follows */
		sprintf (seekpt, "%-*s%10d", POINTNAME_SIZE, pointName, recTime);
		if (d4seek (journal, seekpt) != r4success)
		{
		    appending = TRUE;
		    d4append_start (journal, 0);
		    d4blank (journal);
		    f4assign (d4field (journal, "POINTNAME"), pointName);
		    f4assign_long (d4field (journal, "TIME"), recTime);
		    f4assign (d4field (journal, "EGUTAG"), eguTag);
		    f4assign_double (d4field (journal, "DISP_LOW"), lopr);
		    f4assign_double (d4field (journal, "DISP_HIGH"), hopr);
		}
		else
		{
		    appending = FALSE;
		}
		fieldNum = d4field_number (journal, "VALUE1") + (2 * bucketIndexA (recTime, 300, snapTime));
		sprintf (valbuf, "%.9lf", trendvalue);
		f4assign (d4field_j (journal, fieldNum++), valbuf);
		f4assign_int (d4field_j (journal, fieldNum), 0); /* not recorded yet */

		if (appending)
		{
		    d4append (journal);
		}
	    }
	}
        d4flush_all (journal);
	updrsp->value = ntohs (1);
    }

    return (sizeof(UpdateRsp));
}

static int create_subdir (void)
{
    char	sysbuff[256];
    struct stat	statbuff;

    if (stat (JOURNAL_PATH, &statbuff))
    {
	if (errno == ENOENT)
	{
	    sprintf (sysbuff, "mkdir %s", JOURNAL_PATH);
	    system (sysbuff);
	    if (stat (JOURNAL_PATH, &statbuff))
	    {
		syslog (LOG_OPTIONSE, "trendupdate(%d): failed to create %s\n",
				myPid, JOURNAL_PATH);
		return (0);
	    }
	    else
	    {
	        syslog (LOG_OPTIONSI, "trendupdate(%d): %s created\n",
				myPid, JOURNAL_PATH);
		return (1);
	    }
	}
	else
	{
	    syslog (LOG_OPTIONSE, "trendupdate(%d): stat problem (%d) on %s\n",
				myPid, errno, JOURNAL_PATH);
	    return (0);
	}
    }
    return (1);
}

