/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/




#ifndef TRUE
#define TRUE (1==1)
#define FALSE (1 == 0)
#endif

/* xbase descriptor structure */
#define TABLESPACENAMESIZE	60
#define TABLENAMESIZE		60

#define TREND_SPACE		0
#define ARCHIVE_SPACE		1
#define TEMP_SPACE		2

struct createParms {
    int     targetSize;		/* no new archive data when file reaches this size */
    int     diskSpace;		/* no archive file created if disk space is less than this */
    int     numDirectories;	/* number of directories in list */
    char    **dirList;		/* array of pointers to directory strings */ 
};
typedef struct createParms CreateParams;

struct xBaseData {
    C4CODE	*cb;
    D4DATA	*data;
    D4DATA	*notes;
    int	    ref;
    char    fileName [AVM_TABLENAME_SIZE];
    char    pathName [AVM_MOUNTPNT_SIZE];
};
typedef struct xBaseData XBaseData;

struct xBaseDesc {
    XBaseData *xbaseFile;
    int	    snapspace;
    char    rec_pointname [30];
    char    rec_dbred[4];
    long    rec_time;
    long    lastpointtime;
    char    notes_rec_pointname[30];
    long    notes_rec_time;
    char    currentnote [1024];
    long    currentnotetime;
};
typedef struct xBaseDesc XBaseDesc;
	
/* Force default file to something other than ar_snap */
void GetDefaultFile (char **path, char **file);

/* Force default file to something other than ar_snap */
void SetDefaultFile (char *path, char *file);

/* drop the synonym for the archive table (forces a new one) */
int DropArchiveSynonym (void);

/* update the archive data catalog entry time for the archive synonym */
int UpdateArchiveCatalog (void);

/* open access to codebase (xBase). */
int OpenXBase ( void );

/* create a descriptor object to access data */
XBaseDesc *MakeSnapDesc ( char *filename, char *tablename );

/* close access to oracle*/
int CloseOracle ( void );

/* open and maybe create an empty snap or archive table in the xbase db */
int OpenSnapTable ( XBaseDesc *desc, CreateParams *createVals );

/* create an empty archive file in the xBase db */
int CreateArchTable (XBaseDesc *desc, int size, char *path);

/* create an empty table in the xbase db */
int CreateSnapTable (XBaseDesc *desc);

/* search for record in the xbase db, and fetch first item. */
int StartPointSet (char *pointname, char *dbred, time_t when, time_t end, SnapRec *rec);

/* get the next record */
int NextPointSet (SnapRec *rec, time_t end);

/* close the cursor (noop in xBase) */
int EndPointSet ( );

/* retrieve list of offline files from previous query, NULL means no offline data required */
char *GetOfflineTables ( );

/* search for record in the xbase db, and fetch first item. */
int GetSnapRecord (XBaseDesc *desc, char *pointname, char *dbred, time_t when, SnapRec *rec);

/* get the next record */
int NextSnapRecord (XBaseDesc *desc, SnapRec *rec);

/* close the cursor (noop in xBase) */
int EndSnapRecord (XBaseDesc *desc);

/* open a cursor in the xbase db, and fetch first item. */
int GetSnapRecsByTime (XBaseDesc *desc, time_t start,
		       time_t end, SnapRec *rec);

/* get the next record */
int NextSnapRecByTime (XBaseDesc *desc, SnapRec *rec);

/* close the cursor (noop in xBase) */
int EndSnapRecByTime (XBaseDesc *desc);

/* add to the ar_snap table from the provided data */
int AddSnapRecord (XBaseDesc *desc, SnapRec *rec);

/* update the ar_snap table from the provided data */
int UpdateSnapRecord (XBaseDesc *desc, SnapRec *rec);

/* remove the data from the ar_snap table */
int DeleteSnapRecord (XBaseDesc *desc, SnapRec *rec);

/* remove records before times from the ar_snap table */
int DeleteSnapRecsByTime (XBaseDesc *desc, time_t start, time_t end);

/* remove records for given point before time from the ar_snap table */
int DeleteSnapPointRecs (XBaseDesc *desc, char *pointName, char *reduc, time_t when);

int ArchiveSnapPoint (XBaseDesc *srcDesc, XBaseDesc *destDesc,
		      char *pointName, char *reduc, time_t end);

/* create an empty ar_snap notes table in the xbase db */
int OpenSnapNotesTable ( XBaseDesc *desc, CreateParams *createVals );

/* open a cursor in the xbase db, and fetch first item.
** closes old cursor if open. */
int GetSnapNote (XBaseDesc *desc, char *pointname, time_t *when, char **note);

/* get the next note from the cursor */
int NextSnapNote (XBaseDesc *desc, time_t *when, char **note);

/* get the note previous to the specified time */
int PrevSnapNote (XBaseDesc *desc, char *pointname, time_t *when, char **note);

/* close the cursor (noop in xBase) */
int EndSnapNote (XBaseDesc *desc);

/* open a cursor in the xbase db, and fetch first item.
** closes old cursor if open. */
int GetSnapNoteByTime (XBaseDesc *desc, char **pointname, time_t start,
		       time_t end, time_t *when, char **note);

/* get the next note from the cursor */
int NextSnapNoteByTime (XBaseDesc *desc, char **pointname, time_t *when, char **note);

/* close the cursor (noop in xBase) */
int EndSnapNoteByTime (XBaseDesc *desc);

/* add the ar_snap note from the provided data */
int AddSnapNote (XBaseDesc *desc, char *pointname, time_t when, char *note);

/* update the ar_snap note from the provided data */
int UpdateSnapNote (XBaseDesc *desc, char *pointname, time_t when, char *note);

/* remove the note from the ar_snap table */
int DeleteSnapNote (XBaseDesc *desc, char *pointname, time_t when);

/* remove records between times from the ar_snap table */
int DeleteSnapNoteByTime (XBaseDesc *desc, time_t start, time_t end);

/* Note operations with archive catalog support */
int AddSetNote (char *pointname, time_t when, char *text);
int DeleteSetNote (char *pointname, time_t when);
int PrevSetNote (char *pointname, time_t *when, char **text);
int StartNoteSet (char *pointname, time_t *when, char **note);
int NextNoteSet (time_t *when, char **note);
int EndNoteSet ( void );

/* Archive database/tablespace manipulation */
int UnloadArchiveData (char *table);
int MoveArchiveData (char *table, char *olddbpath, char *newdbpath);
int LoadArchiveData (char *table);

