/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	    MESSAGE STRUCTURES FOR THE ODSS INTERFACE		*/
/*								*/
/*	Hathaway Industrial Automation				*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

/****** Typedefs for unambiguous integer sizes The following	******/
/****** definitions work for Sun, DEC, Intel, and Motorola	******/
/****** using the GNU compiler suite and most other compilers.	******/

#ifndef	INCodssH
#define	INCodssH

typedef int		int32;
typedef unsigned int	uint32;
typedef short		int16;
typedef unsigned short	uint16;
typedef char		int8;
typedef unsigned char	uint8;


/****** The TIS 4000 interface will accept connections on the	******/
/******	following TCP port 					******/
#define ODSS_PORT	8600

/****** All messages start with a 16 bit integer giving the	******/
/****** message type. The following is the set of messages	******/
/****** known to the interface.					******/
typedef enum
{
 ODSS_BUILD=0,		/*	(ODSS_BASE + 0x00)	Channel Build */
 ODSS_BUILD_ACK,       	/*	(ODSS_BASE + 0x01)	Channel Build Ack */
 ODSS_CHAN_CONN,       	/*	(ODSS_BASE + 0x02)	Channel Connection Status Change */
 ODSS_MONITOR,		/*	(ODSS_BASE + 0x03)	Establish or Clear Monitor */
 ODSS_MONDATA,		/*	(ODSS_BASE + 0x04)	Unsolicited Monitor Data */
 ODSS_READ,		/*	(ODSS_BASE + 0x05)	Read Request */
 ODSS_READ_DATA,       	/*	(ODSS_BASE + 0x06)	Read Data */
 ODSS_WRITE,		/*	(ODSS_BASE + 0x07)	Write Request */
 ODSS_WRITE_ACK,       	/*	(ODSS_BASE + 0x08)	Write Ack/Nak */
 ODSS_HIST_READ,       	/*	(ODSS_BASE + 0x09)	Historical Data Read */
 ODSS_HIST_DATA,       	/*	(ODSS_BASE + 0x0a)	Historical Data Return */
 ODSS_CHAN_CLEAR,	/*	(ODSS_BASE + 0x0b)	Channel Close */
 ODSS_BOOMERANG,       	/*	(ODSS_BASE + 0x0c)	A Ping function */
 ODSS_REV,		/*	(ODSS_BASE + 0x0d)	Revision Level */
 ODSS_BUILD_NACK,	/* Negative build ack */
 ODSS_API_OPENED,	/* reply to successful ODSS_REV */
 ODSS_REVISION_MISMATCH,/* reply if rev number mismatch */
 ODSS_MONITOR_NACK,	/* reply if monitor cannot be established */
 ODSS_READ_NACK,	/* reply if read fails */
 ODSS_HIST_NACK,	/* reply if numrequested > 600 or illegal req */
 ODSS_CHAN_CLEAR_NACK,	/* reply if chan clear failed */
 ODSS_CHAN_CLEAR_ACK,	/* reply if clear OK */
 ODSS_WRITE_NACK	/* reply if write failed */
} ODSS_MESSAGE_TYPES;

/****** ODSS_BOOMERANG Message					******/
/****** The TIS 4000 service echos this message to the client	******/
/****** to indicate that the service is up and running.		******/

typedef struct
{
  int32	ODSShandle;
} ODSS_BOOMERANG_MSG;


/****** ODSS_REV Message					******/
/****** This message is sent by the client immediately after	******/
/****** connecting to the TIS 4000 service to declare the	******/
/****** client's current protocol revision level.		******/

typedef struct
{
  int32	ODSSRev;
} ODSS_REV_MSG;

#define ODSS_CURRENT_REV	1


/****** ODSS_BUILD Message. This message is used to create a	******/
/****** channel between a field in a TIS 4000 database record	******/
/****** and the ODSS software. Once created, a channel can be	******/
/****** read, monitored, or written.				******/

typedef struct
{
  int32	ODSShandle;			/* Supplied by ODSS */
  int32	nameLength;			/* Length of name, including
					   null terminator */
} ODSS_BUILD_MSG;

/****** This structure is followed by the channel name. A null	******/
/****** terminator byte is required.				******/



/****** ODSS_BUILD_ACK Message. This message is returned to the	******/
/****** ODSS from the TIS 4000. It indicates that the channel	******/
/****** connection has been built. When this message has been	******/
/****** received, the channel has been connected to the master	******/
/****** RTU.							******/

typedef struct
{
  int32	ODSShandle;			/* References ODSS_BUILD_MSG */
  int16	nativeType;			/* Data Type for field:
					   0 = string
					   1 = 16 bit integer
					   2 = 32 bit float
					   3 = Enumeration
					   4 = Character
					   5 = 32 bit integer
					   6 = 64 bit float */
  int16	nativeCount;			/* Native count for field, always 1
					   except for .VAL in waveforms */
} ODSS_BUILD_ACK_MSG;

typedef struct
{
  int32 ODSShandle;
} ODSS_BUILD_NACK_MSG;


/****** ODSS_CHAN_CONN Message. This message is sent from the	******/
/****** TIS 4000 to the ODSS when the connection status changes ******/

typedef struct
{
  int32	ODSShandle;			/* References ODSS_BUILD_MSG */
  int16	connState;			/* New connection state:
					   0 = Connection Lost
					   1 = Connection Restored */
  int16	filler;				/* Make struct size a multiple of 4 */
} ODSS_CHAN_CONN_MSG;



/****** ODSS_MONITOR Message. This message is sent from the	******/
/****** ODSS to the TIS 4000 to establish or clear a monitor.	******/
/****** When a monitor is first created, an ODSS_MONDATA msg	******/
/****** will be transmitted from the TIS 4000 to the ODSS as	******/
/****** soon as possible. Once a monitor has been established,	******/
/****** its mask cannot be changed. To change a mask, it is	******/
/****** necessary to remove the monitor and then establish it	******/
/****** again with the new mask. There is no response to a	******/
/****** ODSS_MONITOR_MSG that removes a monitor.		******/

typedef struct
{
  int32	ODSShandle;			/* References ODSS_BUILD_MSG */
  int16	eventMask;			/* Masked Conditions to report:
					   (Conditions may be combined)
					   0 = None (remove monitor)
					   1 = Value (Any change for discrete
					   points, beyond MDEL for 
					   analogs)
					   2 = Log (.VAL only, reported when
					   value written to trend data)
					   4 = Alarm (Change in STAT) */
  int16	filler;				/* Make struct size a multiple of 4 */
} ODSS_MONITOR_MSG;



/******	ODSS_MONDATA Message. This message is sent from the	******/
/****** TIS 4000 to the ODSS whenever a field in the database	******/
/****** changes in accordance with the mask that was supplied	******/
/****** when the monitor was established.			******/

typedef struct
{
  int32	ODSShandle;			/* References ODSS_BUILD_MSG */
  int16	status;				/* Point Status. See list at the
					   end of this document. */
  int16	severity;			/* Point Severity. See list at the
					   end of this document. */
  int16	dataType;			/* Data Type. See ODSS_BUILD_ACK_MSG */
  int16	dataLength;			/* Number of bytes of data to follow */
} ODSS_MONDATA_MSG;

/****** This structure is followed in the TCP stream by the	******/
/****** data. All data is in network standard byte order (i.e.,	******/
/****** big endian). All floating point values conform to IEEE	******/
/****** Standard 754. The 32 bit floating point values have 8	******/
/****** exponent bits with a bias of 127 and 23(+1) mantissa	******/
/****** bits. The 64 bit floating point values have 11 exponent	******/
/****** bits with a bias of 1023 and 52(+1) mantissa bits.	******/



/****** ODSS_READ Message. This message is sent from the ODSS	******/
/****** to the TIS 4000 to request that a value be read on a	******/
/****** channel that has been built.				******/

typedef struct
{
  int32	ODSShandle;			/* References ODSS_BUILD_MSG */
  int16	timeout;			/* Read timeout in seconds */
  int16	filler;				/* Make struct size a multiple of 4 */
} ODSS_READ_MSG;



/****** ODSS_READ_DATA Message. This message is sent from the	******/
/****** TIS 4000 to the ODSS to give the results for a previous	******/
/****** ODSS_READ messasge.					******/

typedef struct
{
  int32	ODSShandle;			/* References ODSS_BUILD_MSG */
  int16	readState;			/* Results of the read operation:
					   0 = read successful
					   1 = timed out
					   2 = Rejected (not connected) */
  int16	status;				/* Point Status. See list at the
					   end of this document. */
  int16	severity;			/* Point Severity. See list at the
					   end of this document. */
  int16	dataType;			/* Data Type. See ODSS_BUILD_ACK_MSG */
  int16	dataLength;			/* Number of bytes of data to follow */
  int16	filler;				/* Make struct size a multiple of 4 */
} ODSS_READ_DATA_MSG;

typedef struct
{
  int32 ODSShandle;
} ODSS_READ_NACK_MSG;

/****** This structure is followed in the TCP stream by the	******/
/****** data. All data is in network standard byte order (i.e.,	******/
/****** big endian). All floating point values conform to IEEE	******/
/****** standard 754. See the description of the ODSS_MONDATA	******/
/****** message for a description of the standard format.	******/



/****** ODSS_WRITE Message. This message is sent from the ODSS	*****/
/****** to the TIS 4000 to write data into a channel that has	*****/
/****** previously been created.				*****/

typedef struct
{
  int32	ODSShandle;			/* References ODSS_BUILD_MSG */
  int16	timeout;			/* Write verify timeout in seconds
					   0 = no verify wanted */
  int16	dataType;			/* Must correspond to the native type */
  int16	dataLength;			/* Number of bytes of data to follow */
  int16	filler;				/* Make struct size a multiple of 4 */
} ODSS_WRITE_MSG;

/****** This structure is followed in the TCP stream by the	******/
/****** data. All data is in network standard byte order (i.e.,	******/
/****** big endian. All floating point values conform to IEEE	******/
/****** standard 754. See the description for the ODSS_MONDATA	******/
/****** message for a description of the standard format. Also,	******/
/****** string data must be null terminated. The null must be	******/
/****** included in the dataLength structure member.		******/



/****** ODSS_WRITE_ACK Message. This message is sent from the	******/
/****** TIS 4000 to the ODSS to give the status of a write.	******/
/****** This message will not be sent if the ODSS_WRITE message	******/
/****** had timeout equal to zero.				******/

typedef struct
{
  int32	ODSShandle;			/* References ODSS_BUILD_MSG */
  int16	status;				/* Write Status
					   0 = Verified
					   1 = Written, but verify timed out
					   2 = Rejected, channel offline */
  int16	filler;				/* Make struct size a multiple of 4 */
} ODSS_WRITE_ACK_MSG;



/****** ODSS_CHAN_CLEAR Message. This message is sent from the	******/
/****** ODSS to the TIS 4000 to close a channel. There is no	******/
/****** response to this message.				******/

typedef struct
{
  int32	ODSShandle;			/* References ODSS_BUILD_MSG */
} ODSS_CHAN_CLEAR_MSG;



/****** ODSS_HIST_READ Message. This message is sent from the	******/
/****** ODSS to the TIS 4000 to retrieve historical data.	******/

typedef struct
{
  int32	collRate;			/* Trend collection rate in seconds */
  char	dbReduc[4];			/* Trend File Reduction Method
					   "SMP" for Sample
					   "MAX" for Maximum
					   "MIN" for Minimum
					   "AVG" for Average */
  char	dataReduc[4];			/* Reduction to use if data must be
					   further reduced. See dbReduc */
  int32	secsPerSample;			/* Desired rate, seconds per sample */
  int32	startTime;			/* Seconds since 00:00:00, 1/1/1970 */
  int32	numSamples;			/* Number of samples to get */
  int16	nameLen;			/* Length of point name, including null */
  int16	filler;				/* Make struct size a multiple of 4 */
} ODSS_HIST_READ_MSG;


/****** This message is followed in the input stream by the	******/
/****** point name, null terminated.				******/



/****** ODSS_HIST_DATA Nesage. This data is in response to the	******/
/****** ODSS_HIST_READ message.					******/

typedef struct
{
  int32	numReturned;			/* Number of samples returned */
  int32	dataLength;			/* Number of bytes of data returned */
} ODSS_HIST_DATA_MSG;

/****** This message is followed in the TCP stream by a series	******/
/****** of numReturned 64 bit floating point numbers. This is	******/
/****** the list of values. Another series of numReturned 64	******/
/****** bit floating point numbers with each one containing	******/
/******	the number of actual trend data points that have been	******/
/****** combined into the corresponding entry in the value	******/
/****** list. Finally, the stream also contains a series of	******/
/****** numReturned 3 byte entries containing the status of	******/
/****** each corresponding data entry. These three byte codes	******/
/****** are the same as displayed by trendVu. See the manual	******/
/****** for that program for further information. Therefore, 	******/
/****** dataLength will always be 19 * numReturned, but is 	******/
/****** included here for convenience.				******/ 


/******	Status Codes in ODSS_MONDATA and ODSS_READ_DATA messages:
	00 = No alarm
	01 = Read
	02 = Write
	03 = Hihi
	04 = High
	05 = Lolo
	06 = Low
	07 = Signal
	08 = ROC
	09 = State
	10 = COS
	11 = Open T/C
	12 = Timeout
	13 = HW Limit
	14 = Calc
	15 = Scan
	16 = Link
	17 = Soft
	18 = Bad UDB
	19 = UDF
	20 = Inhibit
	21 = Tagged
	22 = Cmd not allowed
	23 = Volume
	24 = Hihi Level
	25 = Temperature
	26 = Pressure
	27 = BSW
	28 = Tank Valve
	29 = Valve Open
	30 = Valve Close
	31 = Valve Torque Lag
	32 = Valve Fault
	33 = Valve Position
	34 = Automatic
	35 = Vibration
	36 = Pump Start
	37 = Pump Stop
	38 = POT Drop Ready
	39 = Permissive
	40 = Uncommanded
	41 = Threshold Exceeded
	42 = MV Error
	43 = Output Error
	44 = Deviation
	45 = Tracking Mode
	46 = Setpoint Hi Clamp
	47 = Setpt Low Clamp
	48 = Output Hi Clamp
	49 = Output Lo Clamp
	50 = Manual Override
	51 = Setpoint Error
	52 = Output Mode Error
	53 = Parameter
	54 = High Level
	55 = Low Level
	56 = Lolo Level
	57 = Hihi Temperature
	58 = High Temperature
	59 = Low Temperature
	60 = Lolo Temperature
	61 = No Init

/****** Severity Code is ODSS_MONDATA and ODSS_READ_DATA messages:

	00 = No alarm
	01 = Informational
	02 = Event
	03 = Minor
	04 = Major
	05 = Valid
******/

#endif	INCodssH
