
/******************************************************************************\
*       This is a part of the Microsoft Source Code Samples. 
*       Copyright (C) 1993-1995 Microsoft Corporation.
*       All rights reserved. 
*       This source code is only intended as a supplement to 
*       Microsoft Development Tools and/or WinHelp documentation.
*       See these sources for detailed information regarding the 
*       Microsoft samples programs.
\******************************************************************************/

#include <windows.h>
#include <ddeml.h>

#define IDM_ABOUT 100
#define IDM_BLOCKALLCBS 200
#define IDM_ENABLEONECB 201
#define IDM_BLOCKNEXTCB 202
#define IDM_TERMNEXTCB  203
#define IDM_RUNAWAY     204
#define IDM_RENDERDELAY 206
#define IDM_UNBLOCKALLCBS 209
#define IDM_HELP        210
#define IDM_CONTEXT     211
#define IDM_APPOWNED    212
#define IDM_EXIT		213
#define IDM_BOOMERANG	214

#define IDS_BADLENGTH   1

#define MAX_TOPIC       20
#define MAX_COMMENT     80
#define MAX_EXEC        40
#define CCHARS          80
#define CLINES          40

#define UM_CHGDATA      (WM_USER + 435)

#define CSYSTEMITEMS 4
#define CTESTITEMS  4
#define CTOPICS     2
#define ODT_DEFAULT_HOST "localhost"
#define ODT_ODSS_NAME "/tis4000/tasks/odss&"
#define ODT_DEFAULT_PORT 8600
#define ODT_READ_TIMEOUT 10
#define ODT_WRITE_TIMEOUT 0
#define ODT_ACK_TIMEOUT 20
#define ODT_MAX_BUFF_SIZE 20000
#define ODT_STOP_MASK 0
#define ODT_VALUE_MASK 1
#define ODT_TREND_MASK 2
#define ODT_ALARM_MASK 4
#define ODT_CHAN_NAME_SIZE 34						/* name + . + field + \0 */
#define ODT_TIMER 1
#define ODT_BOOM_RESPONSE_TIME 30					/* time to wait before server declared down */
#define SECS_PER_DAY 86400
#define SECS_PER_MINUTE 60
#define SECS_PER_HOUR 3600

#include "dialog.h"

/****** Structrues ******/

// predefined format list item

typedef struct {
    ATOM atom;
    PSTR sz;
} FORMATINFO;
#define CFORMATS 3

extern FORMATINFO aFormats[];

typedef struct _XFERINFO {
    UINT wType;
    UINT wFmt;
    HCONV hConv;
    HSZ hszTopic;
    HSZ hszItem;
    HDDEDATA hData;
    DWORD lData1;
    DWORD lData2;
} XFERINFO;
typedef XFERINFO *PXFERINFO;

typedef struct _ITEMLIST {
    HSZ hszItem;
    HDDEDATA (*npfnCallback)(PXFERINFO, WORD);
    PSTR pszItem;
} ITEMLIST;

typedef struct _TOPICLIST {
    HSZ hszTopic;
    ITEMLIST *pItemList;
    UINT cItems;
    PSTR pszTopic;
} TOPICLIST;

typedef struct {    // used to passinfo to/from TextDataEntryDlgProc and
    HDDEDATA hDdeData;
    UINT wFmt;
    HSZ hszItem;
} XACT;
/* local structures */
typedef enum
{
  ODT_NOCONN=0,
  ODT_CONN
} ODT_CONN_STATE;

typedef enum
{
  ODT_BUILDACK_PENDING,
  ODT_BUILD_ACKED,
  ODT_READ_PENDING,
  ODT_READ_COMPLETE
} ODT_ACK_STATE;

typedef struct _chan_info
{
  char infoChannelName[ODT_CHAN_NAME_SIZE];		/* from command line */
  int32 infoHandle;								/* actually a pointer to this node */
  ODT_ACK_STATE infoAckState;					/* state of acks for timeout checks */
  ODT_CONN_STATE infoConnState;					/* state of connection to CA */
  time_t infoTime;								/* time of read post or build */
  int16 infoNativeType;							/* returned from server */
  int16 infoNativeCount;						/* returned from server */
  int16 infoMask;								/* mask for monitors */
  int16 infoCmd;								/* command we are doing */
  WORD	infoFmt;								/* */
  UINT	owner;									/* application "owns" data handle...must free it later */
  HDDEDATA ddeHandle;							/* this must be freed upon exit */
  char  infoData[80];							/* so we can write after a build ACK */
} ODT_CHAN_INFO, *ODT_CHAN_INFO_PTR;
#define CHAN_INFO_SIZE sizeof(ODT_CHAN_INFO)




// GLOBALS

extern DWORD idInst;
extern CONVCONTEXT CCFilter;
extern HANDLE hInst;
extern HWND hwndServer;
extern RECT rcRand;
extern RECT rcCount;
extern RECT rcComment;
extern RECT rcExec;
extern RECT rcConnCount;
extern RECT rcRndrDelay;
extern RECT rcRunaway;
extern RECT rcAllBlock;
extern RECT rcNextAction;
extern RECT rcHugeSize;
extern RECT rcAppowned;
extern BOOL fAllBlocked;
extern BOOL fAllEnabled;
extern BOOL fEnableOneCB;
extern BOOL fBlockNextCB;
extern BOOL fTermNextCB;
extern BOOL fAppowned;
extern WORD cRunaway;
extern WORD RenderDelay;
extern DWORD count;
extern HSZ hszAppName;
extern CHAR szClass[];
extern CHAR szTopic[MAX_TOPIC];
extern CHAR szServer[MAX_TOPIC];
extern CHAR szComment[MAX_COMMENT];
extern CHAR szExec[MAX_EXEC];
extern CHAR *pszComment;
extern WORD seed;
extern WORD cyText;
extern WORD cServers;
extern HDDEDATA hDataHelp[CFORMATS];
extern HDDEDATA hDataCount[CFORMATS];
extern HDDEDATA hDataRand[CFORMATS];
extern HDDEDATA hDataHuge[CFORMATS];
extern DWORD cbHuge;

extern CHAR szDdeHelp[];
extern FORMATINFO aFormats[CFORMATS];
extern ITEMLIST SystemTopicItemList[CSYSTEMITEMS];
extern ITEMLIST ServerTopicItemList[CTESTITEMS];
extern TOPICLIST topicList[CTOPICS];

extern clientLastTxTime;
extern clientLastRxTime;
extern clientBoomTime;
extern clientNextChar;
extern clientCount;
extern char clientReadBuff[];
extern odssFd;
extern maxFd;
extern struct timeval maxWait;
extern fd_set fdMask;
extern AVL_TREE *odtTree;
extern double *hValues;
extern double *hCounts;
extern unsigned char *hStatus;


// FUNCTIONS


BOOL InitApplication(HANDLE);
BOOL InitInstance(HANDLE, INT);
INT FAR DoDialog(LPSTR lpTemplateName, DLGPROC lpDlgProc, LPARAM param, BOOL fRememberFocus);
LONG  APIENTRY MainWndProc(HWND, UINT, WPARAM, LONG);
BOOL  APIENTRY About(HWND, UINT, WPARAM, LONG);
BOOL  APIENTRY RenderDelayDlgProc(HWND, UINT, WPARAM, LONG);
BOOL  APIENTRY SetTopicDlgProc(HWND, UINT, WPARAM, LONG);
BOOL  APIENTRY SetServerDlgProc(HWND, UINT, WPARAM, LONG);
BOOL  APIENTRY ContextDlgProc(HWND, UINT, WPARAM, LONG);
VOID PaintServer(HWND hwnd);
VOID DrawTextLine(HDC hdc, RECT *prcClip, RECT *prcText, PSTR psz);
HDDEDATA CALLBACK DdeCallback(UINT wType, UINT wFmt, HCONV hConv, HSZ hszTopic,
        HSZ hszItem, HDDEDATA hData, DWORD lData1, DWORD lData2);
HDDEDATA TopicListXfer(PXFERINFO pXferInfo, WORD iFmt);
HDDEDATA ItemListXfer(PXFERINFO pXferInfo, WORD iFmt);
HDDEDATA sysFormatsXfer(PXFERINFO pXferInfo, WORD iFmt);
HDDEDATA TestRandomXfer(PXFERINFO pXferInfo, WORD iFmt);
HDDEDATA HelpXfer(PXFERINFO pXferInfo, WORD iFmt);
HDDEDATA ProcessGetValue(PXFERINFO pXferInfo,WORD iFmt);
HDDEDATA ProcessGetHTrend(PXFERINFO pXferInfo,WORD iFmt);
HDDEDATA ProcessGetHStatus(PXFERINFO pXferInfo,WORD iFmt);
int odssConnect(char *host,short port);
VOID Hszize(VOID);
VOID UnHszize(VOID);
VOID Delay(DWORD delay, BOOL fModal);
void odtReadServer(void);
short odtCompleteMsg(void);
int odtReadSocket(void);
void odtReadBytes(char *buff,int size);
void odtNetError(void);
void odtSendToServer(short msgtype,char *msg,int size);
int odtWrite(char *buff,int nbytes);
void odtBoomerang(void);
int32 odtGetHandle(char *channel);
int32 odtMakeHandle(char *channel);
int odtChanCompare(ODT_CHAN_INFO_PTR ci1,ODT_CHAN_INFO_PTR ci2);
int32 odtBuildChannel(char *channel);
int16 odtFormatData(char *valueString,char *data,int16 type);
void odtFormatOutput(char *data,ODT_CHAN_INFO_PTR pInfo,int size);
void odtCheckAck(void *p1,ODT_CHAN_INFO_PTR node);
void odtFreeDDEHandles(void *p1,ODT_CHAN_INFO_PTR node);
int32 odtParseTime(char *timeString);
