/*
	This file contains code from the ODT test program. It is used
	here to form the client size of the odssdde server.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <odss.h>
#include <avl.h>
#include "server.h"                 /* specific to this program              */
//#include "huge.h"
#include <time.h>
#include "NTUnixSubs.h"
#include <winsock.h>
typedef char * caddr_t;
#include <cadef.h>
#include <signal.h>
#include "msgMon.h"

void odtReadServer(void)
{
	int status;
	unsigned short type,msgType;
	short mSize;

	status = odtReadSocket();					/* drain the FD */
	if(status > 0)
	{
		while(mSize = odtCompleteMsg())
		{					       			/* if complete message is in buffer */
			odtReadBytes((char *)&msgType,2);      			/* get the message type */
			type = ntohs(msgType);
/* 	printf("got message type %02X\n",type); */
			switch(type)
			{
			case ODSS_BOOMERANG:
				{
					ODSS_BOOMERANG_MSG msg;
					time_t now;
					struct tm *tms;
					now = time(0);
					tms = localtime(&now);
					odtReadBytes((char *)&msg,sizeof(msg));
					if((ntohl(msg.ODSShandle) != 0x1234) || ((time(0) - clientBoomTime) > ODT_BOOM_RESPONSE_TIME))
						sprintf(szComment,"Boomerang error or server not responding");
					else
						sprintf(szComment,"ODSS response: %02d:%02d:%02d",tms->tm_hour,tms->tm_min,tms->tm_sec);
					InvalidateRect(hwndServer, &rcComment, TRUE);
					break;
				}
			case ODSS_BUILD_ACK:
				{
					ODT_CHAN_INFO_PTR pInfo;
					ODSS_BUILD_ACK_MSG msg;
					odtReadBytes((char *)&msg,sizeof(msg));
					pInfo = (ODT_CHAN_INFO_PTR)ntohl(msg.ODSShandle);
					pInfo->infoNativeType = ntohs(msg.nativeType);
					pInfo->infoNativeCount = ntohs(msg.nativeCount);
					pInfo->infoAckState = ODT_BUILD_ACKED;
					break;
				}
			case ODSS_CHAN_CONN:
				{
					ODT_CHAN_INFO_PTR pInfo;
					ODSS_CHAN_CONN_MSG msg;
					int16 state;
					odtReadBytes((char *)&msg,sizeof(msg));
					pInfo = (ODT_CHAN_INFO_PTR)ntohl(msg.ODSShandle);
					state = ntohs(msg.connState);
					if(state == 1)
					{
						pInfo->infoConnState = ODT_CONN;
						switch(pInfo->infoCmd)
						{
						case ODSS_READ:
							{
								ODSS_READ_MSG readMsg;
								readMsg.ODSShandle = msg.ODSShandle;
								readMsg.timeout = htons(ODT_READ_TIMEOUT);
								odtSendToServer(ODSS_READ,(char *)&readMsg,sizeof(readMsg));
								break;
							}
						case ODSS_MONITOR:
							{
								ODSS_MONITOR_MSG monMsg;
								monMsg.ODSShandle = msg.ODSShandle;
								monMsg.eventMask = htons(pInfo->infoMask);
								odtSendToServer(ODSS_MONITOR,(char *)&monMsg,sizeof(monMsg));
							}
						}
					}
					else
					{
						pInfo->infoConnState = ODT_NOCONN;
					}
					break;
				}
			case ODSS_READ_DATA:
				{
					int16 size;
					ODT_CHAN_INFO_PTR pInfo;
					ODSS_READ_DATA_MSG rMsg;
					char data[ODT_MAX_BUFF_SIZE];
					odtReadBytes((char *)&rMsg,sizeof(ODSS_READ_DATA_MSG));	/* get the message body */
					size = ntohs(rMsg.dataLength);
					odtReadBytes((char *)data,size);				/* get the data */
					pInfo = (ODT_CHAN_INFO_PTR)ntohl(rMsg.ODSShandle);
					odtFormatOutput(data,pInfo,size);
					sprintf(szComment,"Data for %s...%s",pInfo->infoChannelName,pInfo->infoData);
					InvalidateRect(hwndServer, &rcComment, TRUE);
					break;
				}
			case ODSS_MONDATA:
				{
				  int16 size;
				  ODT_CHAN_INFO_PTR pInfo;
				  ODSS_MONDATA_MSG mMsg;
				  char data[ODT_MAX_BUFF_SIZE];
				  /**/
				  odtReadBytes((char *)&mMsg,sizeof(ODSS_MONDATA_MSG));	/* get the message body */
				  size = ntohs(mMsg.dataLength);
				  odtReadBytes((char *)data,size);				/* get the data */
				  pInfo = (ODT_CHAN_INFO_PTR)ntohl(mMsg.ODSShandle);
				  odtFormatOutput(data,pInfo,size);				/* format to output */
				  break;
				}
			  case ODSS_HIST_DATA:
				{
				  ODSS_HIST_DATA_MSG hdMsg;
				  int32 size,num,i;
				  unsigned char statBytes[4];
				  odtReadBytes((char *)&hdMsg,sizeof(ODSS_HIST_DATA_MSG));
				  size = ntohl(hdMsg.dataLength);
				  num = ntohl(hdMsg.numReturned);
				  if(hValues == 0) hValues = (double *)calloc(num,sizeof(double));
				  if(hCounts == 0) hCounts = (double *)calloc(num,sizeof(double));
				  if(hStatus == 0) hStatus = (unsigned char *)calloc(num,3);
				  odtReadBytes((char *)hValues,sizeof(double)*num);
				  odtReadBytes((char *)hCounts,sizeof(double)*num);
				  odtReadBytes((char *)hStatus,3*num);
				  for(i=0;i<num;i++)
				  {
					  if(hStatus[i*3] == 0xff)		/* check for No Data */
						  strcpy(statBytes,"---");
					  else
						  sprintf(statBytes,"%3.3s\0",&hStatus[i*3]);
					  printf("Value = %lf,Count = %lf,status = %s\nodt> ",hValues[i],hCounts[i],statBytes);
				  }
				}
			}
		}
	}
}
int16 odtFormatData(char *valueString,char *data,int16 type)
{
  int16 len;
  float tempf;
  double tempd;
  short temps;
  int tempi;

  len = 0;
  switch(type)
    {
    case OLD_DBF_CHAR:
      data[0] = valueString[0];
      len = 1;
      break;
    case OLD_DBF_STRING:
      strcpy(data,valueString);
      len = strlen(valueString);
      break;
    case OLD_DBF_FLOAT:
      tempf = (float)atof(valueString);
      memcpy(data,(char *)&tempf,sizeof(float));
      len = sizeof(float);
      break;
    case OLD_DBF_DOUBLE:
      tempd = atof(valueString);
      memcpy(data,(char *)&tempd,sizeof(double));
      len = sizeof(double);
      break;
    case OLD_DBF_SHORT:
    case OLD_DBF_ENUM:
      temps = (short)atoi(valueString);
      memcpy(data,(char *)&temps,sizeof(short));
      len = sizeof(short);
      break;
    case OLD_DBF_LONG:
      tempi = (int)atoi(valueString);
      memcpy(data,(char *)&tempi,sizeof(int));
      len = sizeof(int);
      break;
    default:
      len = 0;
    }
  return(len);
}
void odtFormatOutput(char *data,ODT_CHAN_INFO_PTR pInfo,int size)	/* format to output */
{
  switch(pInfo->infoNativeType)
    {
    case OLD_DBF_DOUBLE:
      sprintf(pInfo->infoData,"%lf\0",*(double *)data);
      break;
    case OLD_DBF_FLOAT:
      sprintf(pInfo->infoData,"%f\0",*(float *)data);
      break;
    case OLD_DBF_STRING:
      strncpy(pInfo->infoData,data,size);
      break;
    case OLD_DBF_SHORT:
    case OLD_DBF_ENUM:
      sprintf(pInfo->infoData,"%d\0",*(short *)data);
      break;
    case OLD_DBF_LONG:
      sprintf(pInfo->infoData,"%d\0",*(int *)data);
      break;
    case OLD_DBF_CHAR:
      pInfo->infoData[0] = data[0];
      break;
    }
}
short odtCompleteMsg(void)
{
  short	mType;
  int mSize;
  ODSS_READ_DATA_MSG rMsg;
  ODSS_MONDATA_MSG mMsg;
  ODSS_HIST_DATA_MSG hdMsg;

  if(clientCount < sizeof(mType))
    return(0);
  memcpy(&mType,(char *)(clientReadBuff+clientNextChar),sizeof(mType));
  mSize = ODT_MAX_BUFF_SIZE+1;
  /*
    switch on the message type. All message are considered complete if the number of bytes
    in the buffer (ioc_rxCount) is greater that or equal to the total message size PLUS
    the size of the message type.
  */
  switch(ntohs(mType))
    {
    case ODSS_BOOMERANG:
      mSize = sizeof(ODSS_BOOMERANG_MSG) + sizeof(mType);
      break;
    case ODSS_BUILD_ACK:
      mSize = sizeof(ODSS_BUILD_ACK_MSG) + sizeof(mType);
      break;
    case ODSS_CHAN_CONN:
      mSize = sizeof(ODSS_CHAN_CONN_MSG) + sizeof(mType);
      break;
    case ODSS_WRITE_ACK:
      mSize = sizeof(ODSS_WRITE_ACK_MSG) + sizeof(mType);
      break;
    case ODSS_READ_DATA:
      if(clientCount >= (sizeof(ODSS_READ_DATA_MSG) + sizeof(mType)))
	{
	  short temp;
	  memcpy((char *)&rMsg,(char *)(clientReadBuff+clientNextChar+2),sizeof(ODSS_READ_DATA_MSG));
	  temp = rMsg.dataLength;
	  mSize = sizeof(ODSS_READ_DATA_MSG) + ntohs(temp) + sizeof(mType);
	}
      break;
    case ODSS_MONDATA:
      if(clientCount >= (sizeof(ODSS_MONDATA_MSG) + sizeof(mType)))
	{
	  short temp;
	  memcpy((char *)&mMsg,(char *)(clientReadBuff+clientNextChar+2),sizeof(ODSS_MONDATA_MSG));
	  temp = mMsg.dataLength;
	  mSize = sizeof(ODSS_MONDATA_MSG) + ntohs(temp) + sizeof(mType);
	}
      break;
    case ODSS_HIST_DATA:
      if(clientCount >= (sizeof(ODSS_HIST_DATA_MSG) + sizeof(mType)))
	{
	  int32 temp;
	  memcpy((char *)&hdMsg,(char *)(clientReadBuff+clientNextChar+2),sizeof(ODSS_HIST_DATA_MSG));
	  temp = hdMsg.dataLength;
	  mSize = sizeof(ODSS_HIST_DATA_MSG) + ntohl(temp) + sizeof(mType);
	}
      break;
    default:
      printf("Invalid message %04X from ODSS server...exiting\n",mType);
      odtNetError();
    }
  if(mSize > clientCount)
    return(0);
  else
    return(mSize);
}
int odtReadSocket(void)
{
  int		readSize,stat;

  if(odssFd == -1)
    return(-1);
  errno = 0;
/*
	If we are already some bytes into the buffer then slide them down to the
	begining and zero the next char pointer.
*/
  if(clientNextChar)
    {
      if(clientCount)
         {
           memcpy((char *)clientReadBuff,(char *)(clientReadBuff+clientNextChar),clientCount);
         }
      clientNextChar = 0;
    }
  readSize = ODT_MAX_BUFF_SIZE - clientCount;
  if ((stat = recv(odssFd, (char *)(clientReadBuff+clientCount+clientNextChar), readSize, 0)) < 0)
    {
      perror("Read error in \"odtReadSocket\"");
      odtNetError();		/* GCT 2-Jul-1996 drop the whole thing on error */
    }
  if(stat == 0)					/* zero length read */
    {
      perror("Zero-length read in \"odtReadSocket\"");
      odtNetError();
    }
  clientCount += stat;			/* number actually read into buffer */
  clientLastRxTime = time(0);		/* time of this recieve */
  return(stat);
}
void odtReadBytes(char *buff,int size)
{
  memcpy(buff,(char *)(clientReadBuff+clientNextChar),size);
  clientNextChar += size;
  clientCount -= size;
}
void odtSendToServer(short msgtype,char *msg,int size)
{
  short mtype = htons(msgtype);
  odtWrite((char *)&mtype,sizeof(short));
  odtWrite(msg,size);
}
int odtWrite(char *buff,int nbytes)
{
  int	buffSize;
  int	numSent;
  int	intSize = sizeof(int);
  int	localError;

  if(odssFd == -1)
    return(-1);
  numSent = send(odssFd,buff,nbytes,0);
  clientLastTxTime = time(0);
  if (numSent == -1)
  {
	  localError = WSAGetLastError();
	  if (localError == WSAEWOULDBLOCK)
	  {
		  getsockopt(odssFd, SOL_SOCKET, SO_SNDBUF, (char *)&buffSize, &intSize);
		  buffSize += nbytes;
		  if (setsockopt(odssFd, SOL_SOCKET, SO_SNDBUF, (char *)&buffSize, sizeof(buffSize)) == -1)
		  {
	      //perror("odt: Setsockopt error in \"odtWrite\"  on socket");
			  odtNetError();
		  }
	  }
      else
	  {
	  //perror("odt: Send error in \"odtWrite\" on socket");
		  odtNetError();
	  }
	  return (odtWrite(buff, nbytes));
  }
  else if (numSent != nbytes)
  {
      buff += numSent;
      nbytes -= numSent;
      return (odtWrite(buff, nbytes));
  }
  return(1);
}
void odtNetError(void)
{
	int lastError = WSAGetLastError();
	close(odssFd);
	sprintf(szComment,"WSA error %d",lastError);
	InvalidateRect(hwndServer, &rcComment, TRUE);
	exit(lastError);
}
/*
  return my handle for this channel name. If no handle exists then return
  zero to force an ODSS_BUILD_MSG to be sent
*/
int32 odtGetHandle(char *channel)
{
  ODT_CHAN_INFO info;

  if(channel)
    {
      strcpy((char *)&info.infoChannelName,channel);
      return((int32)avl_findsym(odtTree,&info));	/* either it's there or its not */
    }
  else
    return(0);
}
/*
  Make handle from channel name. The handle is actually the address
  of the node in the tree for the given point name.
*/
int32 odtMakeHandle(char *channel)
{
  ODT_CHAN_INFO_PTR pInfo;

  pInfo = (ODT_CHAN_INFO_PTR)avl_newnode(CHAN_INFO_SIZE);		/* create new node */
  strcpy(pInfo->infoChannelName,channel);      	/* put in node */
  if(avl_insert(odtTree,pInfo) != NULL)
    {
      avl_freenode(pInfo);
      pInfo = 0;
    }
  else
    {
      pInfo->infoHandle = (int32)pInfo;
      pInfo->infoConnState = ODT_NOCONN;		/* not connected yet */
      pInfo->infoAckState = ODT_BUILDACK_PENDING;	/* waiting for build ACK */
    }
  return((int32)pInfo);
}

int odtChanCompare(ODT_CHAN_INFO_PTR ci1,ODT_CHAN_INFO_PTR ci2)
{
  return(strcmp(ci1->infoChannelName,ci2->infoChannelName));
}
void odtCheckAck(void *p1,ODT_CHAN_INFO_PTR node)
{
	if(((time(0) - node->infoTime) > ODT_ACK_TIMEOUT) && (node->infoAckState == ODT_BUILDACK_PENDING))
	{
		avl_delete(odtTree,node);
		sprintf(szComment,"Build ack timed-out for \"%s\"",node->infoChannelName);
		InvalidateRect(hwndServer, &rcComment, TRUE);
	}
}
void odtFreeDDEHandles(void *p1,ODT_CHAN_INFO_PTR node)
{
	if(node->ddeHandle) DdeFreeDataHandle(node->ddeHandle);
}
/*
	These functions do all the message work
*/
void odtBoomerang(void)
{
  ODSS_BOOMERANG_MSG msg;

  clientBoomTime = time(0);
  msg.ODSShandle = htonl(0x1234);
  odtSendToServer(ODSS_BOOMERANG,(char *)&msg,sizeof(msg));
}
int32 odtBuildChannel(char *channel)
{
  int32 handle;
  int32 size;

  struct
  {
    ODSS_BUILD_MSG msg;
    char name[ODT_CHAN_NAME_SIZE];
  } nameMsg;
  ODT_CHAN_INFO_PTR pInfo;

  handle = 0;
  if(channel)
    {
      handle = odtMakeHandle(channel);
      nameMsg.msg.ODSShandle = htonl(handle);
      nameMsg.msg.nameLength = htonl(strlen(channel)+1);
      sprintf((char *)&nameMsg.name,"%s\0",channel);
      size = strlen(channel)+1;
      odtSendToServer(ODSS_BUILD,(char *)&nameMsg,(sizeof(ODSS_BUILD_MSG)+size));
      pInfo = (ODT_CHAN_INFO_PTR)handle;
      pInfo->infoTime = time(0);
    }
  return(handle);
}
