#ifndef	INCodtmainH
#define	INCodtmainH

#define	ODT_DEFAULT_PORT	8600

#define ODT_DEFAULT_HOST "localhost"
#define ODT_ODSS_NAME "/tis4000/tasks/odss&"
#define ODT_READ_TIMEOUT 10
#define ODT_WRITE_TIMEOUT 0
#define ODT_ACK_TIMEOUT 20
#define ODT_MAX_BUFF_SIZE 20000
#define ODT_STOP_MASK 0
#define ODT_VALUE_MASK 1
#define ODT_TREND_MASK 2
#define ODT_ALARM_MASK 4
#define ODT_CHAN_NAME_SIZE 34			/* name + . + field + \0 */
#define ODT_TIMER 1
#define ODT_BOOM_RESPONSE_TIME 30		/* time to wait before server declared down */
#define SECS_PER_DAY 86400
#define SECS_PER_MINUTE 60
#define SECS_PER_HOUR 3600
/* command strings */
#define ODT_COMMAND_PROMPT "odt> "
#define ODT_COMMAND_QUIT "QUIT"
#define ODT_COMMAND_GET "GET"
#define ODT_COMMAND_PUT "PUT"
#define ODT_COMMAND_MON "MON"
#define ODT_COMMAND_STOP "STOP"
#define ODT_COMMAND_CLOSE "CLOSE"
#define ODT_COMMAND_HIST "HIST"
#define ODT_COMMAND_BOOM "BOOM"
#define ODT_COMMAND_DUMP "DUMP"
#define ODT_COMMAND_PAUSE "PAUSE"
#define ODT_COMMAND_TREND "TREND"
#define ODT_COMMAND_ALARM "ALARM"

typedef enum
{
  ODT_NOCONN=0,
  ODT_CONN
} ODT_CONN_STATE;

typedef enum
{
  ODT_BUILDACK_PENDING,
  ODT_BUILD_ACKED,
  ODT_READ_PENDING,
  ODT_READ_COMPLETE
} ODT_ACK_STATE;

/* local structures */
typedef struct _chan_info
{
  char infoChannelName[ODT_CHAN_NAME_SIZE];		/* from command line */
  int32 infoHandle;			/* actually a pointer to this node */
  ODT_ACK_STATE infoAckState;		/* state of acks for timeout checks */
  ODT_CONN_STATE infoConnState;		/* state of connection to CA */
  time_t infoTime;			/* time of read post or build */
  int16 infoNativeType;			/* returned from server */
  int16 infoNativeCount;		/* returned from server */
  int16 infoMask;			/* mask for monitors */
  int16 infoCmd;			/* command we are doing */
  char  infoData[80];			/* so we can write after a build ACK */
} ODT_CHAN_INFO, *ODT_CHAN_INFO_PTR;
#define CHAN_INFO_SIZE sizeof(ODT_CHAN_INFO)

/* function prototypes */
void odtSignalHandler(int whichOne);
int odtNetworkInit(char *hostname,short port);
int32 odtBuildChannel(char *channel);
int odtGetChannelValue(char *channel);
int odtPutChannelValue(char *channel,char *value);
int odtMonitorChannel(char *channel,short mask);
int odtCloseChannel(char *channel);
int odtBoomerang(void);
int odtSendToServer(short msgtype,char *msg,int size);
int odtWrite(char *msg,int size);
int32 odtGetHandle(char *channel);
int odtChanCompare(void *ci1,void *ci2);
int32 odtMakeHandle(char *channel);
void odtDumpNode(void *p1,void *node);
int16 odtFormatData(char *valueString,char *data,int16 type);
int odtNetError();
int odtReadServer(void);
short odtCompleteMsg(void);
int odtReadSocket(void);
void odtReadBytes(char *buff,int size);
int odtFormatOutput(char *data,ODT_CHAN_INFO_PTR pInfo,int size);	/* format to output */
void odtGetMulti(void *p1,void *node);
void odtCloseMulti(void *p1,void *node);
void odtCheckAck(void *p1,void *node);
void odtProcessCommands(char *cmdline);
int odtGetHistData(char *hist);
int32 odtParseTime(char *timeString);

#endif	INCodtmainH
