#ifndef	opnoteHXX
#define	opnoteHXX

extern "C"
{
#include "opNoteDb.h"
}
#include "textFilters.hxx"
#include "tisgalaxylib.hxx"
#include "opNoteUtils.hxx"

class opNoteMainDialog ;
class opNoteMainTextItem ;
class opNoteTimer;
class opNoteSearchDialog ;
class opNoteEditDialog ;
class opNoteKeyDialog ;
class opNoteKeyClassComboBox ;
class opNoteKeyValueComboBox ;

// enum defs
enum {
    OPNO_ERROR = -1,
    OPNO_SUCCESS,
    OPNO_INFO_NOTICE,
    OPNO_CAUTION_NOTICE,
    OPNO_NUMMISCENUMS
};


class opNoteApplication
{
    vapplication	*opNoteVApplication;
    vresource		opNoteResource;
    vresource		messagesResource;
    vresource		titlesResource;
    vresource		tisGalaxyResource;
    char		currentOperator[OP_SIZE + 1];
    Filter		currentFilter;
    OpNoteRecord	currentSelectedRecord;
    int			initError;
    // is there a record selected in the main dialogs textview
    int			isSelectedRecord;
    int			dbopened;
    int			initedTisGalaxy;

public:

    // class constructor
    opNoteApplication (void);
    // class destructor
    ~opNoteApplication (void);

    // function to return the resources
    vresource getOpNoteResource (void) 
	{ return (opNoteResource); }
    vresource getMessagesResource (void)
	{ return (messagesResource); }
    vresource getTitlesResource (void)
	{ return (titlesResource); }
    vresource getGalaxyRes (void)
	{ return (tisGalaxyResource); }

    int hasInitErrors (void)
	{ return (initError); }

    // function to deal with current operator
    char *getCurrentOperator (void)
	{ return (currentOperator); }

    // func to update the current selected record
    void setCurrentSelectedRecord (OpNoteRecord *rec)
	{ currentSelectedRecord = *rec; }
    OpNoteRecord *getCurrentSelectedRecord (void)
	{ return (&currentSelectedRecord); }

    // functions to deal with whether a record is selected
    void setIsSelectedRecord (int TrueFalse)
	{ isSelectedRecord = TrueFalse; }
    int getIsSelectedRecord (void)
	{ return (isSelectedRecord); }

    // functions to deal with the current record filter
    void getCurrentFilter (Filter *filter)
	{ *filter = currentFilter; }
    void updateCurrentFilter (Filter *filter)
	{ currentFilter = *filter; }

};


/*** MAIN DIALOG *******************************************************************/

// enum values used for the vcommandSelector::Issue
enum
{
    EXIT_MENUITEM = 0,
    SEARCH_MENUITEM,
    NEW_MENUITEM,
    MODIFY_MENUITEM,
    DELETE_MENUITEM,
    KEY_MENUITEM,
    MENUITEM_COUNT
};

class opNoteMainDialog : public vdialog
{
    // pointer to application
    opNoteApplication	*opNoteApp;

    // resource for the main dialog
    vresource		mainDialogResource;

    // pointers to items in the main dialog
    vmenu		*helpMenu;
    opNoteMainTextItem	*mainTextItem;

    // pointers to other dialogs
    opNoteSearchDialog	*searchDialog;
    opNoteEditDialog    *editDialog;
    opNoteKeyDialog	*keyDialog;

    // place to store internal names for main dialog menu items
    const vname		*commandNames[MENUITEM_COUNT];

    int			loadInitErrors;
    TisAbout		*tisAbout;
    void		BuildAboutMenu (void);

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(opNoteMainDialog);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(opNoteMainDialog, vdialog);

    // vwindow overide
    void ObserveWindow (vevent *event);

    // vloadable override
    void LoadInit (vresource resource);

    // override 
    void Destroy ();

    // open 
    void OpenMain (opNoteApplication *opNoteApp);

    // functions to the application
    void storeOpNoteApp (opNoteApplication *app)
	{ opNoteApp = app; }
    opNoteApplication *getApplication (void)
	{ return (opNoteApp); }

    // return the opNote applications resource
    vresource getOpNoteResource (void)
	{ return (opNoteApp -> getOpNoteResource()); }

    // return the main dialog's resource
    vresource getMainDialogResource (void)
	{ return (mainDialogResource); }

    // functions for the searchDialog
    void storeSearchDialog (opNoteSearchDialog *srchD)
	{searchDialog = srchD;}
    opNoteSearchDialog *getSearchDialog (void)
	{ return (searchDialog); }

    // functions for the editDialog
    void storeEditDialog (opNoteEditDialog *editD)
	{ editDialog = editD; }
    opNoteEditDialog *getEditDialog (void)
	{ return (editDialog); }

    // functions for the keyDialog
    void storeKeyDialog (opNoteKeyDialog *keyD)
	{ keyDialog = keyD; }
    opNoteKeyDialog *getKeyDialog (void)
	{ return (keyDialog); }

    // return the mainTextItem
    opNoteMainTextItem *getMainTextItem (void)
	{return (mainTextItem); }

    int getLoadInitErrors (void)
	{ return (loadInitErrors); }
};


class  opNoteMainCommand : public vcommandSelector
{
public:
    // macros for standard versions of the
    // constructors and destructors
    vkindTYPED_FULL_DECLARATION(opNoteMainCommand);
    vloadableINLINE_CONSTRUCTORS(opNoteMainCommand, vcommandSelector);

    //override virtual issue function
    int Issue(vdict *);
};


class opNoteMainTextItem : public vtextitem
{
    // pointer to the main dialog
    opNoteMainDialog	*mainDialog;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(opNoteMainTextItem);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(opNoteMainTextItem, vtextitem);

    //override virtual notify function
    void ObserveView(vevent *event, int message);


    // add new member functions
    void storeMainDialog (opNoteMainDialog *main)
	{ mainDialog = main; }
    void setCurrentSelection (void);
    int updateCurrentSelectedRecord (vtext *text, long startLine);
    void redisplayData (void);

    // ptr to timer
    opNoteTimer	*opnoteTimer;
};


/*** OPNOTE TIMER CLASS **************************************************************/

class opNoteTimer : public vtimer
{
    opNoteMainTextItem	*mainText;

public:

    // virtual overrride
    void ObserveTimer (void);

    // handle mainText
    void StoreMainText (opNoteMainTextItem *mt)
	{ mainText = mt; }

    // start timer sequence
    void StartMainTimer (void);
};


/*** SEARCH DIALOG *******************************************************************/

// defines for search filter dialog status
#define	NO_SEARCH_STATUS	0
#define	SEARCH_OK		1
#define	SEARCH_CANCEL		2

enum
{
    APPLY_BUTTON,
    NEW_BUTTON,
    CANCEL_BUTTON,
    SEARCHITEM_COUNT
};

class opNoteSearchDialog : public vdialog
{
    vresource	searchDialogResource;

    // pointers to items in search dialog
    vtextitem	*oper;
    vtextitem	*startDate;
    vtextitem	*startTime;
    vtextitem	*endDate;
    vtextitem	*endTime;
    vtextitem	*searchString;
    opNoteControl *keyClass;
    vcontrol	*keyValue; 
    vbutton	*applyButton;
    vbutton	*newButton;
    vbutton	*cancelButton;

    // place to store internal names for search dialog buttons
    const vname	*commandNames[SEARCHITEM_COUNT];

    // search copy of filter
    Filter	searchFilter;

    // search dialog status
    unsigned short	searchDialogStatus;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(opNoteSearchDialog);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(opNoteSearchDialog, vdialog);

    // vwindow override
    void ObserveWindow (vevent *event);

    // vloadable override
    void LoadInit (vresource resource);

    // override
    void Destroy ();
    
    // store resource for possible later use
    void storeSearchDialogResource (vresource res)
	{ searchDialogResource = res; }

    // set search dialog status
    void setSearchDialogStatus (unsigned short status)
	{ searchDialogStatus = status; }

    // get the dialog's status
    unsigned short getSearchDialogStatus ()
	{ return (searchDialogStatus); }

    // start the search dialog
    void startOpNoteSearch (void);

    // save data from search dialog
    int verifySearchParams (void);

    // reset the fields on the dialog
    void resetDialogFields (void);

    // update the filter for the app
    void updateSearchFilter (Filter *filter);
};


class  opNoteSearchCommand : public vcommandSelector
{
public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(opNoteSearchCommand);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(opNoteSearchCommand, vcommandSelector);

    //override virtual issue function
    int Issue(vdict *);
};


/*** EDIT DIALOG *******************************************************************/

// defines for edit filter dialog status
#define	NO_EDIT_STATUS		0
#define	EDIT_OK			1
#define	EDIT_CANCEL		2

enum
{
    SAVE_EDIT_BUTTON,
    CANCEL_EDIT_BUTTON,
    EDITITEM_COUNT
};

class opNoteEditDialog : public vdialog
{
    // pointer to application
    opNoteApplication	*App;

    vresource	editDialogResource;

    // pointers to items in edit dialog
    vtextitem	*oper;
    vtextitem	*editDate;
    vtextitem	*editTime;
    opNoteControl *keyClass;
    vcontrol	*keyValue; 
    vtextitem	*editText;
    vbutton	*saveButton;
    vbutton	*cancelButton;

    // place to store internal names for search dialog buttons
    const vname	*commandNames[EDITITEM_COUNT];

    // edit dialog status
    unsigned short	editDialogStatus;

    // the edited record
    OpNoteRecord	editedRecord;

    // remember daylight savings time
    int			daylightSavingsTime;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(opNoteEditDialog);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(opNoteEditDialog, vdialog);

    // vwindow override
    void ObserveWindow (vevent *event);

    // vloadable override
    void LoadInit (vresource resource);

    // override
    void Destroy ();

    // functions to the application
    void storeOpNoteApp (opNoteApplication *app) 
	{ App = app; }
    opNoteApplication *getApplication (void) 
	{ return (App); }
    
    // store resource for possible later use
    void storeEditDialogResource (vresource res) 
	{ editDialogResource = res; }

    // functions for edit dialog status
    void setEditDialogStatus (unsigned short status) 
	{ editDialogStatus = status; }
    unsigned short getEditDialogStatus ()
	{ return (editDialogStatus); }

    // start the edit dialog
    void startOpNoteEdit (OpNoteRecord *currentSelectedRecord);

    // save data from edit dialog
    void applyEdit (void);

    // return the edited record
    void getEditedRecord (OpNoteRecord *rec)
	{ *rec = editedRecord; }

    // functions to process daylightSavingstime
    void setDayLightSavingsTime (int dlst)
	{ daylightSavingsTime = dlst; }
    int  getDayLightSavingsTime (void)
	{ return (daylightSavingsTime); }
};


class  opNoteEditCommand : public vcommandSelector
{
public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(opNoteEditCommand);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(opNoteEditCommand, vcommandSelector);

    //override virtual issue function
    int Issue(vdict *);
};


/*** KEY DIALOG *******************************************************************/

enum
{
    ADD_KEY_BUTTON,
    DELETE_KEY_BUTTON,
    CLOSE_KEY_BUTTON,
    KEYITEM_COUNT
};

class opNoteKeyDialog : public vdialog
{
    vresource	keyDialogResource;

    // pointers to items in key dialog
    opNoteKeyClassComboBox	*keyClassCombo;
    opNoteKeyValueComboBox	*keyValueCombo;
    vbutton			*addButton;
    vbutton			*deleteButton;
    vbutton			*closeButton;

    // place to store internal names for search dialog buttons
    const vname	*commandNames[KEYITEM_COUNT];

    // flag if deleting entire key type
    int		deletingKeyType;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(opNoteKeyDialog);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(opNoteKeyDialog, vdialog);

    // vwindow override
    void ObserveWindow (vevent *event);

    // vloadable override
    void LoadInit (vresource resource);

    // override
    void Destroy ();
    
    // store resource for possible later use
    void storeKeyDialogResource (vresource res)
	{ keyDialogResource = res; }

    // start the key dialog
    void startOpNoteKey (void);

    // return the key dialog's comboboxes
    opNoteKeyClassComboBox *getKeyClassCombo (void)
	{ return (keyClassCombo); }
    opNoteKeyValueComboBox *getKeyValueCombo (void)
	{ return (keyValueCombo); }

    // function to add/delete from the key list
    int addKey (void);
    int deleteKey (void);

    // functions to handle deletingKeyType
    void setDeletingKeyType (int value)
	{ deletingKeyType = value; }
    int getDeletingKeyType (void)
	{ return (deletingKeyType); }
};


class  opNoteKeyCommand : public vcommandSelector
{
public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(opNoteKeyCommand);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(opNoteKeyCommand, vcommandSelector);

    //override virtual issue function
    int Issue(vdict *);
};

class opNoteKeyClassComboBox : public vcombobox
{
    char    lastSelected[KEY_SIZE + 1];
    int	    modifying;

public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(opNoteKeyClassComboBox);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(opNoteKeyClassComboBox, vcombobox);

    // vloadable override 
    void LoadInit (vresource resource);

    // override observer
    void ObserveCombobox (vevent *event, int message);

    // extend the combobox definition
    void buildKeyClassCombo (void);
    char *getLastSelected (void)
	{ return (lastSelected); }
    void putLastSelected (char *pStr)
    { 
	strncpy (lastSelected, pStr, KEY_SIZE); 
	lastSelected[KEY_SIZE] = 0;
    }
    void setModifying (void)
	{ modifying = 1; }
    void clearModifying (void)
	{ modifying = 0; }
    int  getModifying (void)
	{ return (modifying); }
};

class opNoteKeyValueComboBox : public vcombobox
{
    int    modifying;
public:

    // overrides from vkindTyped
    vkindTYPED_FULL_DECLARATION(opNoteKeyValueComboBox);

    // the default versions of necessary constructors/destructor
    vloadableINLINE_CONSTRUCTORS(opNoteKeyValueComboBox, vcombobox);

    // extend the combobox definition
    void buildKeyValueCombo (char *pStr);
    void setModifying (void)
	{ modifying = 1; }
    void clearModifying (void)
	{ modifying = 0; }
    int  getModifying (void)
	{ return (modifying); }
};

#endif opnoteHXX
