#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32
#include <time.h>
#include <ctype.h>
#include <fcntl.h>

#if !(defined(MIPSEL) || defined(AXP))
#include <memory.h>
#else
#include <sys/types.h>
#include <sys/stat.h>
#endif

#include "opNoteDb.h"

static char *KeyDefFile = {"/tis4000db/OpKeyDefs.dat"};
static char *KeyTempFile = {"/tis4000db/OpKeyTemp.dat"};
static char inpstring [MAXKEYLEN];
static char outstring [MAXKEYLEN];
static char *getElement (char **ptr);


/* codebase variables */
static C4CODE	codeBaseDesc;
static D4DATA	*notes;

/* Notes file structures */

/* define database field structure */
static F4FIELD_INFO notes_field_info[] = 
{
   /* field name,   type, width, decimals */
   { "TIME",	    'C',    10,    0 },
   { "YEAR",	    'N',    4,     0 },
   { "DAYOFYEAR",   'N',    3,     0 },
   { "HOUR",	    'N',    2,     0 },
   { "MINUTE",	    'N',    2,     0 },
   { "SECOND",	    'N',    2,     0 },
   { "OPERATOR",    'C',    30,    0 },
   { "KEYTYPE",	    'C',    32,    0 },
   { "KEYVALUE",    'C',    32,    0 },
   { "NOTE",	    'M',    1024,    0 },
   { 0, 0, 0, 0 },
} ;

/* define multiple tags */
static T4TAG_INFO notes_tag_info[] = 
{
   /* name,	    expression,	filter, unique,	     descending */
   { "MEMO_TIME", "TIME+OPERATOR",	"", 0, 0 },
   { 0,  0,  0, 0, 0 },
};

int openNotefile (char *dbfile, int create)
{
    d4init (&codeBaseDesc);
#if defined(USECB63) || defined(WIN32)
    codeBaseDesc.errGo = 0;
    codeBaseDesc.errCreate = 0;
    codeBaseDesc.errOpen = 0;
    codeBaseDesc.errTagName = 0;
    codeBaseDesc.errFieldName = 0;
    codeBaseDesc.errOff = 1;		/* disable codebase error msgs */
#else
    codeBaseDesc.go_error = 0;
    codeBaseDesc.create_error = 0;
    codeBaseDesc.open_error = 0;
    codeBaseDesc.tag_name_error = 0;
    codeBaseDesc.field_name_error = 0;
#   ifdef CB51
	codeBaseDesc.read_only = 1;
#   endif CB51
#endif USECB63

    /* create name from dbfile name */
    /* will create new database, and overwrite */
    codeBaseDesc.safety = 0;  
#if defined(USECB63) || defined(WIN32)
    codeBaseDesc.readLock = 0;
#else
    codeBaseDesc.read_lock = 0;
#endif USECB63

    notes = d4open (&codeBaseDesc, dbfile);
    if ((notes == 0) && create)
    {
        if ((notes = d4create (&codeBaseDesc, dbfile, notes_field_info, 
				notes_tag_info)) == 0)
        {
            return DBNOTESERR;
        }
    }
    return DBOK;
}

void RelNotes ( )
{
    if (notes)
    {
        d4unlock_all (notes);
    }
}

int CloseNotes ()
{
    if (notes)
	d4close (notes);
    notes = 0;
    return 1;
}

/* static storage for current note values */
static OpNoteRecord currentrec;

/* helper routine */
static OpNoteRecord *GetFilteredRecord ( Filter *filt, long dir )
{
    OpNoteRecord *retrec = NULL;
    int d4stat;

    while (retrec == NULL)
    {
	/* load record */
	f4ncpy (d4field (notes, "OPERATOR"), currentrec.op, OP_SIZE);
	currentrec.op [OP_SIZE] = 0;
	trimblanks (currentrec.op);
	currentrec.time = atol (f4str (d4field (notes, "TIME")));
	f4ncpy (d4field (notes, "KEYTYPE"), currentrec.keytype, KEY_SIZE);
	currentrec.keytype[KEY_SIZE] = 0;
	trimblanks (currentrec.keytype);
	f4ncpy (d4field (notes, "KEYVALUE"), currentrec.keyvalue, KEY_SIZE);
	currentrec.keyvalue[KEY_SIZE] = 0;
	trimblanks (currentrec.keyvalue);
	strncpy (currentrec.text, m4ptr (d4field (notes, "NOTE")), MAX_NOTE);
	currentrec.text[MAX_NOTE] = 0;
	trimblanks (currentrec.text);
	if ((filt == NULL)
	  || ( ((strlen(filt->op) == 0) || (strncmp (filt->op, currentrec.op, 
		 strlen(filt->op)) == 0))
	    && ((filt->start == 0) || (filt->start <= currentrec.time))
	    && ((filt->end == 0) || (filt->end > currentrec.time))
	    && ((filt->keytype == 0) || (strncmp (filt->keytype, currentrec.keytype, 
		 strlen(filt->keytype)) == 0))
	    && ((filt->keyvalue == 0) || (strncmp (filt->keyvalue, currentrec.keyvalue, 
		 strlen(filt->keyvalue)) == 0))
	    && ((strlen(filt->search_text) == 0) || (strstr (currentrec.text, 
			filt->search_text))) ))

	{
	    retrec = &currentrec;
	}
	else
	{
	    /* stop searching when times get out of range */
	    if (((dir < 0) && ((filt->start != 0) && (filt->start > currentrec.time)))
	      || ((dir > 0) && ((filt->end != 0) && (filt->end <= currentrec.time))))
		break;

            /* skip a record in the correct direction */
	    d4stat = d4skip (notes, dir);
	    
	    /* stop if no more data */
	    if ((d4stat != 0) && (d4stat != r4after))
		break;
        }
    }
    return retrec;
}

OpNoteRecord *GetNote ( Filter *filt, long time )
{
    char *seekpt;
    int  d4stat;
    OpNoteRecord *retrec = NULL;

    if (notes)
    {
	seekpt = (char *)malloc (TIME_SIZE+OP_SIZE+1); 
        d4tag_select (notes, d4tag (notes, "MEMO_TIME"));

        sprintf (seekpt, "%-10.10d%-*.*s", time, OP_SIZE, OP_SIZE, filt->op);
        d4stat = d4seek (notes, seekpt);

        /* if positioned after, return the time from the record */
        if ((d4stat == 0) || (d4stat == r4after))
        {
	    retrec = GetFilteredRecord (filt, 1L);
        }
        free (seekpt);
    }
    return retrec;
}

OpNoteRecord *NextNote ( Filter *filt, long time )
{
    /* search for note after current record. */
    int  d4stat;
    OpNoteRecord *retrec = NULL;

    if (notes)
    {
	d4stat = d4skip (notes, 1L);

	/* read from here until filter is OK */
	if ((d4stat == 0) || (d4stat == r4after))
	{
	    retrec = GetFilteredRecord (filt, 1L);
	}
    }
    return retrec;
}


OpNoteRecord *GetPrevNote ( Filter *filt, long time )
{
    /* search for note before current record. */
    int  d4stat;
    OpNoteRecord *retrec = NULL;

    if (notes)
    {
	d4stat = d4skip (notes, -1L);

	/* read from here until filter is OK */
	if ((d4stat == 0) || (d4stat == r4after))
	{
	    retrec = GetFilteredRecord (filt, -1L);
	}
    }
    return retrec;
}

int SaveNote ( OpNoteRecord *rec, char *dbfile )
{
    struct tm *tmblock;
    char timebuf [11];
    int stat;
    int appending;
    Filter tempfilter;
    OpNoteRecord *rp;

    tempfilter.start = 0;
    tempfilter.end = 0;
    strcpy (tempfilter.op, rec->op);
    strcpy (tempfilter.keytype, rec->keytype);
    strcpy (tempfilter.keyvalue, rec->keyvalue);
    tempfilter.search_text[0] = 0;

    /* create a new record */

    /* create notes file if necessary */
    if (notes == 0)
    {
        if (openNotefile (dbfile, 1) != DBOK)
        {
	    return (0);
        }
    }
    else if ((rp = GetNote (&tempfilter, rec->time)) && 
	     (rp->time == rec->time))
    {
        /* note exists at this time, delete it */
	d4blank (notes);
	d4delete (notes);
	d4flush_all (notes);
    }

    stat = d4top (notes);
    if (((stat == 0) && !d4deleted(notes))
      || (stat == r4eof))
    {
	/* Couldn't find a blank record */
	appending = 1;
	d4append_start (notes, 1);
    }
    else
    {
	appending = 0;
    }
    d4blank (notes);
    f4assign (d4field (notes, "OPERATOR"), rec->op);
    f4assign (d4field (notes, "KEYTYPE"), rec->keytype);
    f4assign (d4field (notes, "KEYVALUE"), rec->keyvalue);
    sprintf (timebuf, "%-10.10d", rec->time);
    f4assign (d4field (notes, "TIME"), timebuf);
    tmblock = gmtime ((time_t *)&rec->time);
    f4assign_int (d4field (notes, "YEAR"), tmblock->tm_year+1900);
    f4assign_int (d4field (notes, "DAYOFYEAR"), tmblock->tm_yday);
    f4assign_int (d4field (notes, "HOUR"), tmblock->tm_hour);
    f4assign_int (d4field (notes, "MINUTE"), tmblock->tm_min);
    f4assign_int (d4field (notes, "SECOND") , tmblock->tm_sec);
    m4assign (d4field (notes, "NOTE"), rec->text);

    if (appending)
    {	/* Restore updated record */
	d4append (notes);
    }
    else
    {
	d4recall (notes);
    }
    d4flush_all (notes);
#ifndef CB51
    d4update_header (notes, 1, 1);
#endif
    d4unlock_all (notes);
    return 1;
}


int DeleteNote ( OpNoteRecord *rec )
{
    Filter tempfilter;

    tempfilter.start = 0;
    tempfilter.end = 0;
    strcpy (tempfilter.op, rec->op);
    strcpy (tempfilter.keytype, rec->keytype);
    strcpy (tempfilter.keyvalue, rec->keyvalue);
    tempfilter.search_text[0] = 0;

    /* delete a record */

    /* no database open */
    if (notes == 0)
    {
        return 0;
    }
    else if (GetNote (&tempfilter, rec->time)->time == rec->time)
    {
	d4blank (notes);
	d4delete (notes);
	d4flush_all (notes);
    }
    return 1;
}

char **GetKeytypes (char *errmsg )
{
    int ix;
    char *sp;
    FILE *fp;
    char *cp1;
    int returnStringIx = 0;
    char **returnStringPtr = NULL;

    fp = fopen (KeyDefFile, "rt");
    if (fp)
    {
        while (!feof (fp))
        {
            /* get a line of text from input */
	    ix = 0;
            while (!feof (fp))
            {
                inpstring [ix] = fgetc(fp);
                if ((inpstring [ix] == 0x0d) || (inpstring [ix] == 0x0a) 
		  || (inpstring [ix] == -1))
		{
                    break;
                }
		ix++;
                if (ix == (MAXKEYLEN - 1))
		{
		    if (returnStringPtr)
			freeItems (returnStringPtr);
		    sprintf (errmsg, "Too many values for type \n");
                    return NULL; /* bad string in file. */
		}
            }
	    inpstring [ix] = 0;

	    /* process input line */
	    sp = inpstring;
	    while (*sp && (*sp == ' ' || *sp == 0x09))
		sp++;
	    if ((*sp == 0) || (*sp == '#'))
		continue; /* blank line or comment! */

	    /* Get first item on line, key type name, check for type 
	       we have already have
	    */
	    cp1 = getElement (&sp);
	    if (cp1)
	    {
		if (returnStringPtr == NULL)
		{
		    returnStringPtr = (char **)calloc (MAXKEYITEMS, sizeof(char *));
		    if (returnStringIx == MAXKEYITEMS)
		    {
			sprintf (errmsg, 
				 "Warning: Unable to allocate memory for key types\n");
			return (returnStringPtr);
		    }
		    returnStringIx = 0;
		}
		if (returnStringPtr) /* add type to list */
		{
		    /* see if we already have it */
		    for (ix = 0; ix < returnStringIx; ix++)
			if (strcmp (cp1, returnStringPtr[ix]) == 0)
			    break;
		    if (ix == returnStringIx)
		    {
			returnStringPtr [returnStringIx] = 
						(char *)malloc (strlen (cp1) + 1);
			if (returnStringPtr [returnStringIx])
			    strcpy (returnStringPtr [returnStringIx], cp1);
			else /* out of memory! */
			{
			    sprintf (errmsg, 
				  "Warning: Unable to allocate memory for type '%s'\n", 
				  cp1);
			    return (returnStringPtr);
			}
			returnStringIx++;
			if (returnStringIx == MAXKEYITEMS)
			{
			    sprintf (errmsg, 
				     "Warning: Key type limit reached at type '%s'\n", 
				     cp1);
			    return (returnStringPtr);
			}
		    }
		}
	    }
	}
        fclose (fp);
    } /* end of opened file code */
    return returnStringPtr;
}

char **GetKeyValues ( char *keytype, char *errmsg )
{
    int ix;
    char *sp;
    FILE *fp;
    char *cp1;
    char *cp2;
    int returnStringIx = 0;
    char **returnStringPtr = NULL;

    fp = fopen (KeyDefFile, "rt");
    if (fp)
    {
        while (!feof (fp))
        {
            /* get a line of text from input */
	    ix = 0;
            while (!feof (fp))
            {
                inpstring [ix] = fgetc(fp);
                if ((inpstring [ix] == 0x0d) || (inpstring [ix] == 0x0a) 
		  || (inpstring [ix] == -1))
		{
                    break;
                }
		ix++;
                if (ix == (MAXKEYLEN - 1))
		{
		    if (returnStringPtr)
			freeItems (returnStringPtr);
		    sprintf (errmsg, "Too many values for type \n");
                    return NULL; /* bad string in file. */
		}
            }
	    inpstring [ix] = 0;

	    /* process input line */
	    sp = inpstring;
	    while (*sp && (*sp == ' ' || *sp == 0x09))
		sp++;
	    if ((*sp == 0) || (*sp == '#'))
		continue; /* blank line or comment! */

	    /* Get first item on line, key type name, check for our type */
	    cp1 = getElement (&sp);
	    if (cp1 && (strcmp (cp1, keytype) == 0))
	    {
		if (returnStringPtr == NULL)
		{
		    returnStringPtr = (char **)calloc (MAXKEYITEMS, sizeof(char *));
		    if (returnStringIx == MAXKEYITEMS)
		    {
			sprintf (errmsg, 
			       "Warning: Unable to allocate memory for key type '%s'\n", 
			        keytype);
			return (returnStringPtr);
		    }
		    returnStringIx = 0;
		}
		if (returnStringPtr) /* add value to list */
		{
		    cp2 = getElement (&sp);
		    returnStringPtr [returnStringIx] = (char *)malloc (strlen (cp2) + 1);
		    if (returnStringPtr [returnStringIx])
			strcpy (returnStringPtr [returnStringIx], cp2);
		    else /* out of memory! */
		    {
			sprintf (errmsg, 
			      "Warning: Unable to allocate memory for key value '%s'\n", 
			       cp2);
			return (returnStringPtr);
		    }
		    returnStringIx++;
		    if (returnStringIx == MAXKEYITEMS)
		    {
			sprintf (errmsg, 
				"Warning: Key value limit reached for type '%s'\n", 
				keytype);
			return (returnStringPtr);
		    }
		}
	    }
	}
        fclose (fp);
    } /* end of opened file code */
    return returnStringPtr;
}

int WriteKeyValue ( char *keytype, char *keyvalue, char *msg )
{
    int ix;
    char *sp;
    FILE *fp;
    FILE *ofp;
    char *cp1;
    char *cp2;
    char out [KEY_SIZE + 1 + KEY_SIZE + 1];
    int written = 0;

    fp = fopen (KeyDefFile, "rt");
    ofp = fopen (KeyTempFile, "wt");
    if (ofp)
    {
        while (fp && !feof (fp))
        {
            /* get a line of text from input */
	    ix = 0;
            while (fp && !feof (fp))
            {
                inpstring [ix] = fgetc(fp);
                if ((inpstring [ix] == 0x0d) || (inpstring [ix] == 0x0a) 
		  || (inpstring [ix] == -1))
		{
                    break;
                }
		ix++;
                if (ix == (MAXKEYLEN - 1))
		{
		    sprintf (msg, "Too many characters in input line\n");
		    fclose (fp);
 		    fclose (ofp);
		    unlink (KeyTempFile);
                    return 0; /* bad string in file. */
		}
            }
	    inpstring [ix] = 0;

	    /* process input line */
	    sp = inpstring;
	    strcpy (outstring, inpstring);
	    while (*sp && (*sp == ' ' || *sp == 0x09))
		sp++;
	    if ((*sp == 0) || (*sp == '#'))
	    {
		if (ix != 0)
		{
		    fputs (outstring, ofp);
		    fputs ("\n", ofp);
		}
		continue; /* blank line or comment! */
	    }

	    /* guard for duplicate type and value */
	    cp1 = getElement (&sp);
	    cp2 = getElement (&sp);
	    if ((cp1 && (strcmp (cp1, keytype) == 0))
	      && (cp2 && (strcmp (cp2, keyvalue) == 0)))
	    {
		/* redundant keytype/value - warn and drop */
		sprintf (msg, "Warning: duplicate key\n");
		fclose (fp);
 		fclose (ofp);
		unlink (KeyTempFile);
		return 0; /* bad string in file. */
	    }
	    else if ((!written) && ((cp1 && (strcmp (keytype, cp1) <= 0))
	      && (cp2 && (strcmp (keyvalue, cp2) < 0))))
	    {
		/* proper location for new record, write it */
		sprintf (out, "%s:%s\n", keytype, keyvalue);
		fputs (out, ofp);
		written = 1;
	    }
	    /* Write out work record */
	    fputs (outstring, ofp);
	    fputs ("\n", ofp);
	}
	if (fp)
	    fclose (fp);
	if (!written)
	{
	    /* If rec not written yet, write it out */
	    sprintf (out, "%s:%s\n", keytype, keyvalue);
	    fputs (out, ofp);
	    written = 1;
	}
	fclose (ofp);
	/* rename the temporary file over the source file */
	rename (KeyTempFile, KeyDefFile);
    }
    if (fp)
	fclose (fp);
    return 1;
}

int DeleteKeyType ( char *keytype, char *msg)
{
    int ix;
    char *sp;
    FILE *fp;
    FILE *ofp;
    char *cp1;

    fp = fopen (KeyDefFile, "rt");
    ofp = fopen (KeyTempFile, "wt");
    if (fp && ofp)
    {
        while (fp && !feof (fp))
        {
            /* get a line of text from input */
	    ix = 0;
            while (fp && !feof (fp))
            {
                inpstring [ix] = fgetc(fp);
                if ((inpstring [ix] == 0x0d) || (inpstring [ix] == 0x0a) 
		  || (inpstring [ix] == -1))
		{
                    break;
                }
		ix++;
                if (ix == (MAXKEYLEN - 1))
		{
		    sprintf (msg, "Too many characters in input line\n");
		    fclose (fp);
 		    fclose (ofp);
		    unlink (KeyTempFile);
                    return 0; /* bad string in file. */
		}
            }
	    inpstring [ix] = 0;

	    /* process input line */
	    strcpy (outstring, inpstring);
	    sp = inpstring;
	    while (*sp && (*sp == ' ' || *sp == 0x09))
		sp++;
	    if ((*sp == 0) || (*sp == '#'))
	    {
		if (ix != 0)
		{
		    fputs (outstring, ofp);
		    fputs ("\n", ofp);
		}
		continue; /* blank line or comment! */
	    }

	    /* check type */
	    cp1 = getElement (&sp);
	    if (cp1 && (strcmp (cp1, keytype) == 0))
	    {
		/* found the matching entry, don't copy this line */
	    }
	    else
	    {
	        /* Write out work record */
		fputs (outstring, ofp);
		fputs ("\n", ofp);
	    }
	}
	fclose (fp);
	fclose (ofp);
	/* rename the temporary file over the source file */
	rename (KeyTempFile, KeyDefFile);
    }
    if (fp)
	fclose (fp);
    return 1;
}

int DeleteKeyValue ( char *keytype, char *keyvalue, char *msg)
{
    int ix;
    char *sp;
    FILE *fp;
    FILE *ofp;
    char *cp1;
    char *cp2;

    fp = fopen (KeyDefFile, "rt");
    ofp = fopen (KeyTempFile, "wt");
    if (fp && ofp)
    {
        while (fp && !feof (fp))
        {
            /* get a line of text from input */
	    ix = 0;
            while (fp && !feof (fp))
            {
                inpstring [ix] = fgetc(fp);
                if ((inpstring [ix] == 0x0d) || (inpstring [ix] == 0x0a) 
		  || (inpstring [ix] == -1))
		{
                    break;
                }
		ix++;
                if (ix == (MAXKEYLEN - 1))
		{
		    sprintf (msg, "Too many characters in input line\n");
		    fclose (fp);
 		    fclose (ofp);
		    unlink (KeyTempFile);
                    return 0; /* bad string in file. */
		}
            }
	    inpstring [ix] = 0;

	    /* process input line */
	    strcpy (outstring, inpstring);
	    sp = inpstring;
	    while (*sp && (*sp == ' ' || *sp == 0x09))
		sp++;
	    if ((*sp == 0) || (*sp == '#'))
	    {
		if (ix != 0)
		{
		    fputs (outstring, ofp);
		    fputs ("\n", ofp);
		}
		continue; /* blank line or comment! */
	    }

	    /* check type and value */
	    cp1 = getElement (&sp);
	    cp2 = getElement (&sp);
	    if ((cp1 && (strcmp (cp1, keytype) == 0))
	      && (cp2 && (strcmp (cp2, keyvalue) == 0)))
	    {
		/* found the matching entry, don't copy this line */
	    }
	    else
	    {
	        /* Write out work record */
		fputs (outstring, ofp);
		fputs ("\n", ofp);
	    }
	}
	fclose (fp);
	fclose (ofp);
	/* rename the temporary file over the source file */
	rename (KeyTempFile, KeyDefFile);
    }
    if (fp)
	fclose (fp);
    return 1;
}

void freeItems (char **stringPtr)
{
    int i;

    for (i = 0; i < MAXKEYITEMS; i++)
	if (stringPtr[i])
	    free (stringPtr[i]);
    free (stringPtr);
}

static char *getElement (char **ptr)
{
    char *retp = NULL;

    /* skip white space */
    while (*ptr && (**ptr == ' ' || **ptr == 0x09))
	(*ptr)++;
    if (**ptr == '"')
    {
	(*ptr)++;
	retp = *ptr;
	while (**ptr && **ptr != '"')
	    (*ptr)++;
	if (**ptr)
	{
	    **ptr = 0;
	    (*ptr)++;
	}
    }
    else
    {
	retp = *ptr;
	while (**ptr && (**ptr != ':')
	  && (**ptr != ' ') && (**ptr != 0x09))
	    (*ptr)++;
	if (**ptr)
	{
	    **ptr = 0;
	    (*ptr)++;
	}
    }
    return retp;
}


void trimblanks (char *string) 
{
    int last;

    /* strip trailing blanks */
    for (last = strlen(string)-1; 
	 (last >= 0) && (string[last] == ' '); 
	 last--)
	string[last] = 0;
}

int	getPrivilege (char *privilege)
{
    return (OperPriv (&codeBaseDesc, privilege));
}
