#include <vport.h>

/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vbuttonHEADER
#include vtextitemHEADER
#include vstartupHEADER
#include vtimeHEADER
#include vmemHEADER
#include vcontrolHEADER
#include vtimerHEADER
extern "C"
{
#include "opNoteDb.h"
}
#include "opno.h"
#include "opNoteUtils.hxx"
#include "opNote.hxx"


vkindTYPED_FULL_DEFINITION(opNoteEditDialog, vdialog, NULL);

/*******************************************************************
*
*    opNoteEditDialog override for constructors init ()
*
*******************************************************************/

void opNoteEditDialog::LoadInit (vresource res)
{
    int		i;

    // store the search dialog's resource
    editDialogResource = res;

    // set dialog status
    setEditDialogStatus (NO_EDIT_STATUS);

    oper = vtextitem::RequestSubLoad (this, opno_editOperatorTag);
    editDate = vtextitem::RequestSubLoad (this, opno_editDateTag);
    editTime = vtextitem::RequestSubLoad (this, opno_editTimeTag);
    keyClass = opNoteControl::RequestSubLoad (this, opno_editKeyClassTag);
    keyValue = vcontrolOptionMenu::RequestSubLoad (this, opno_editKeyValueTag);
    keyClass -> setKeyValue (keyValue);
    editText = vtextitem::RequestSubLoad (this, opno_editTextTag);
    saveButton = vbutton::RequestSubLoad (this, opno_editSaveButtonTag);
    cancelButton = vbutton::RequestSubLoad (this, opno_editCancelButtonTag);

    // init inherited data
    vdialog::LoadInit (res);

    // store internal names for the edit button tags
    commandNames[SAVE_EDIT_BUTTON]  = opno_editSaveButtonTag;
    commandNames[CANCEL_EDIT_BUTTON] = opno_editCancelButtonTag;

    // create a command space
    vcommandSpace *space = new vcommandSpace;

    // create a dictionary with a default hash size and hash function
    vdict *dict = new vdict (vnameHash);

    // populate our dictionary with menu commands
    for (i = SAVE_EDIT_BUTTON; i < EDITITEM_COUNT; i++)
    {
	// get an instance of our subclassed selector command
	opNoteEditCommand *command = new opNoteEditCommand;
	// set the commands selector value
	command -> SetSelector(i);
	// store command in the dictionary using the 
	// intern'd tag name
	dict -> SetElem (commandNames[i], command);
    }
    // add dictionary to the command space and give the command
    // space ownership of the dictionary, vname_Command defines
    // the type attribute of the things in the dictionary
    space -> AddDictOwned (vname_Command, dict);

    SetSpaceOwned (space);

}


/*******************************************************************
*
*    opNoteEditDialog destroy function
*
*******************************************************************/

void opNoteEditDialog::Destroy (void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }
    vdialog::Destroy ();
}

/*******************************************************************
*
*    opNoteEditDialog notify function
*
*******************************************************************/

void opNoteEditDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose ())
    {
	setEditDialogStatus (EDIT_CANCEL);
	Close ();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    opNoteEditDialog start function
*
*******************************************************************/

void opNoteEditDialog::startOpNoteEdit (OpNoteRecord *currentSelectedRecord)
{
    struct tm		*tim;
    time_t		now;
    vchar		*pvChar;
    vmenu		*keyMenu;
    vmenuItem		*menuItem;

    // build the keyClass menu
    keyClass -> buildKeyClassMenu (0);

    // called to create a new record
    if (currentSelectedRecord == NULL )
    {
	opNoteApplication *opNoteApp = getApplication ();
	char *currOperator = opNoteApp -> getCurrentOperator ();
	oper -> SetTextScribed (vcharScribeSystem ((vchar *)currOperator));

	now = time (0);
	tim = localtime (&now);

	// build key value menu based on 1st item in key class menu
	keyMenu = keyClass -> GetMenu ();
	int menuItemCount = keyMenu -> GetItemCount ();
	if (menuItemCount > 0)
	{
	    menuItem = keyMenu -> GetItemAt (0);
	    pvChar = (vchar *)menuItem -> GetTitle ();
	    keyClass -> buildKeyValueMenu (pvChar);

	    // make sure to enable these 2 menus
	    keyClass -> Enable ();
	    keyValue -> Enable ();
	}

	// make sure the text area is clear
	editText -> SetTextScribed (vcharScribeSystem (""));
    }
    // called to edit an existing record
    else
    {
	oper -> SetTextScribed (vcharScribeSystem ((vchar *)&currentSelectedRecord -> op[0]));
	tim = localtime (&currentSelectedRecord -> time);

	// set keyClass to value in currentSelectedRecord
	keyMenu = keyClass -> GetMenu ();
	int maxMenuItems = keyClass -> GetMax ();
	int foundIt = 0;
	int i;
	const vchar *menuItemTitle;
	for (i = 0; i <= maxMenuItems; i++)
	{
	    menuItem = keyMenu -> GetItemAt (i);
	    menuItemTitle = menuItem -> GetTitle ();
	    if ( (strncmp (currentSelectedRecord -> keytype, (char *)menuItemTitle,
			   KEY_SIZE)) == 0)
	    {
		foundIt = 1;
		break;
	    }
	}
	if (foundIt == 0)
	    i = 0;
	keyClass -> SetValue (i);
	menuItem = keyMenu -> GetItemAt (i);
	pvChar = (vchar *)menuItem -> GetTitle ();
	keyClass -> buildKeyValueMenu (pvChar);

	// set the keyValue to the value in the currentSelectedRecord
	keyMenu = keyValue -> GetMenu ();
	maxMenuItems = keyValue -> GetMax ();
	foundIt = 0;
	for (i = 0; i < maxMenuItems; i++)
	{
	    menuItem = keyMenu -> GetItemAt (i);
	    menuItemTitle = menuItem -> GetTitle ();
	    if ( (strncmp (currentSelectedRecord -> keyvalue, (char *)menuItemTitle,
			   KEY_SIZE)) == 0)
	    {
		foundIt = 1;
		break;
	    }
	}
	if (foundIt == 0)
	    i = 0;
	keyValue -> SetValue (i);

	// do not allow user to modify these 2
	keyClass -> Disable ();
	keyValue -> Disable ();

	// place text in the text field for modification
	editText -> SetTextScribed (vcharScribeSystem (currentSelectedRecord -> text));
    }
    char buff[32];
    sprintf (buff, "%02d/%02d/%4d", tim -> tm_mon+1, tim -> tm_mday, tim -> tm_year + 1900);
    editDate -> SetTextScribed (vcharScribeSystem (buff));
    sprintf (buff, "%02d:%02d:%02d", tim -> tm_hour, tim -> tm_min, tim -> tm_sec);
    editTime -> SetTextScribed (vcharScribeSystem (buff));

    // be it a new record or a modification of an existing record
    // the user can not modify these fields
    oper -> Disable ();
    editDate -> Disable ();
    editTime -> Disable ();
    setDayLightSavingsTime (tim -> tm_isdst);
    setEditDialogStatus (NO_EDIT_STATUS);
    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
    vdialog::Open ();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    opNoteEdit function to apply edits to record
*
*******************************************************************/

void opNoteEditDialog::applyEdit (void)
{
    vchar	dateStr[16];
    vchar	timeStr[16];
    vchar	op[OP_SIZE + 1];
    vchar	note[MAX_NOTE + 1];
    vchar	*pvChar;
    char	key[KEY_SIZE + 1];
    struct tm	tim;

    vmemSet ((void *)&tim, 0, sizeof (struct tm));
    vmemSet ((void *)&editedRecord, 0, sizeof (OpNoteRecord));
    vmemSet ((void *)&dateStr[0], 0, sizeof (dateStr));
    vmemSet ((void *)&timeStr[0], 0, sizeof (timeStr));
    vmemSet ((void *)&op[0], 0, sizeof (op));
    vmemSet ((void *)&note[0], 0, sizeof (note));

    oper -> GetText (op, OP_SIZE);
    vcharExportBounded (vcharScribeSystemExport (op), editedRecord.op, OP_SIZE);
    oper -> GetText ((vchar *)&editedRecord.op[0], OP_SIZE);
    editDate -> GetText (dateStr, sizeof (dateStr));
    extractDate (dateStr, &tim);
    editTime -> GetText (timeStr, sizeof (timeStr));
    extractTime (timeStr, &tim);
    tim.tm_isdst = getDayLightSavingsTime ();
#if (defined(MIPSEL) || defined(SOLARIS) || defined(AXP) || defined(WIN32))
    editedRecord.time = mktime (&tim);
#else
    editedRecord.time = timegm (&tim);
#endif

    // get the key class value
    int controlValue = keyClass -> GetValue ();
    // get the menu attached to the control item
    vmenu *keyMenu = keyClass -> GetMenu ();
    // return a ptr to the string that is currently displayed on the menu
    vmenuItem *menuItem = keyMenu -> GetItemAt (controlValue);
    pvChar = (vchar *)(menuItem -> GetTitle ());
    vcharExportBounded (vcharScribeSystemExport (pvChar), key, KEY_SIZE);
    strncpy (editedRecord.keytype, key, KEY_SIZE);

    /* get the key value */
    controlValue = keyValue -> GetValue ();
    keyMenu = keyValue -> GetMenu ();
    menuItem = keyMenu -> GetItemAt (controlValue);
    pvChar = (vchar *)(menuItem -> GetTitle ());
    vcharExportBounded (vcharScribeSystemExport (pvChar), key, KEY_SIZE);
    strncpy (editedRecord.keyvalue, key, KEY_SIZE);

    // get the note string
    editText -> GetText (note, MAX_NOTE);
    vcharExportBounded (vcharScribeSystemExport (note), editedRecord.text, MAX_NOTE);
}


vkindTYPED_FULL_DEFINITION(opNoteEditCommand, vcommandSelector, NULL);

/*******************************************************************
*
*    opNoteEditCommand menu selector commands
*
*******************************************************************/

int opNoteEditCommand::Issue (vdict *dict)
{
    opNoteEditDialog *editDialog = (opNoteEditDialog *) dict -> GetElem (vname_Dialog);
    switch (GetSelector ())
    {
	case SAVE_EDIT_BUTTON:
	    editDialog -> applyEdit ();
	    editDialog -> setEditDialogStatus (EDIT_OK);
	    editDialog -> Close ();
	    // stops the event loop that was created when the
	    // search dialog was Opened
	    vevent::StopProcessing ();
	    break;

	case CANCEL_EDIT_BUTTON:
	    editDialog -> setEditDialogStatus (EDIT_CANCEL);
	    editDialog -> Close ();
	    vevent::StopProcessing ();
	    break;

	default:
	    return (vFALSE);
    }
    return (vTRUE);
}
