#include <vport.h>

/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vbuttonHEADER
#include vtextitemHEADER
#include vstartupHEADER
#include vtimeHEADER
#include vmemHEADER
#include vcontrolHEADER
#include vcomboboxHEADER
#include vtimerHEADER
extern "C"
{
#include "opNoteDb.h"
}
#include "opno.h"
#include "opNoteUtils.hxx"
#include "opNote.hxx"


vkindTYPED_FULL_DEFINITION(opNoteKeyDialog, vdialog, NULL);

/*******************************************************************
*
*    opNoteKeyDialog override for constructors init ()
*
*******************************************************************/

void opNoteKeyDialog::LoadInit (vresource res)
{
    int		i;

    // store the search dialog's resource
    keyDialogResource = res;

    keyClassCombo = opNoteKeyClassComboBox::RequestSubLoad (this, opno_keyKeyClassTag);
    keyValueCombo = opNoteKeyValueComboBox::RequestSubLoad (this, opno_keyKeyValueTag);
    addButton = vbutton::RequestSubLoad (this, opno_keyAddButtonTag);
    deleteButton = vbutton::RequestSubLoad (this, opno_keyDeleteButtonTag);
    closeButton = vbutton::RequestSubLoad (this, opno_keyCloseButtonTag);

    // init inherited data
    vdialog::LoadInit (res);

    // store internal names for the edit button tags
    commandNames[ADD_KEY_BUTTON]  = opno_keyAddButtonTag;
    commandNames[DELETE_KEY_BUTTON] = opno_keyDeleteButtonTag;
    commandNames[CLOSE_KEY_BUTTON] = opno_keyCloseButtonTag;

    // create a command space
    vcommandSpace *space = new vcommandSpace;

    // create a dictionary with a default hash size and hash function
    vdict *dict = new vdict (vnameHash);

    // populate our dictionary with menu commands
    for (i = ADD_KEY_BUTTON; i < KEYITEM_COUNT; i++)
    {
	// get an instance of our subclassed selector command
	opNoteKeyCommand *command = new opNoteKeyCommand;
	// set the commands selector value
	command -> SetSelector(i);
	// store command in the dictionary using the 
	// intern'd tag name
	dict -> SetElem (commandNames[i], command);
    }
    // add dictionary to the command space and give the command
    // space ownership of the dictionary, vname_Command defines
    // the type attribute of the things in the dictionary
    space -> AddDictOwned (vname_Command, dict);

    SetSpaceOwned (space);
}


/*******************************************************************
*
*    opNoteKeyDialog destroy function
*
*******************************************************************/

void opNoteKeyDialog::Destroy (void)
{
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }
    vdialog::Destroy ();
}

/*******************************************************************
*
*    opNoteKeyDialog notify function
*
*******************************************************************/

void opNoteKeyDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose ())
    {
	Close ();
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    opNoteKeyDialog start function
*
*******************************************************************/

void opNoteKeyDialog::startOpNoteKey (void)
{
    char	classStr [KEY_SIZE + 1];

    setDeletingKeyType (vFALSE);

    // build the keyClass combo
    keyClassCombo -> clearModifying ();
    keyClassCombo -> buildKeyClassCombo ();
    int valueCount = keyClassCombo -> GetValueCount ();
    keyValueCombo -> clearModifying ();
    if (valueCount > 0)
    {
	keyClassCombo -> SelectValueAt (0);

	// build the key value combo based on the 1st item in the 
	//key class combo
	vcomboboxValue *value = keyClassCombo -> GetValueAt (0);
	const vchar *keyClassTitle = value -> GetTitle ();
	vcharExportBounded (vcharScribeSystemExport (keyClassTitle), classStr, KEY_SIZE);
	keyValueCombo -> buildKeyValueCombo (classStr);
    }

    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
    vdialog::Open ();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    opNoteKeyDialog add Key function
*
*******************************************************************/

int opNoteKeyDialog::addKey (void)
{
    char	message[MAXERRMSG];
    char	classStr [KEY_SIZE + 1];
    char	valueStr [KEY_SIZE + 1];

    opNoteKeyClassComboBox *keyClass = getKeyClassCombo ();
    opNoteKeyValueComboBox *keyValue = getKeyValueCombo ();

    vcomboboxValue *classEntryBoxValue = keyClass -> GetEntryboxValue (keyClass -> GetEntrybox ());
    const vchar *classEntryBoxTitle = classEntryBoxValue -> GetTitle ();
    vcharExportBounded (vcharScribeSystemExport (classEntryBoxTitle), classStr, KEY_SIZE);

    vcomboboxValue *valueEntryBoxValue = keyValue -> GetEntryboxValue (keyValue -> GetEntrybox ());
    const vchar *valueEntryBoxTitle = valueEntryBoxValue -> GetTitle ();
    vcharExportBounded (vcharScribeSystemExport (valueEntryBoxTitle), valueStr, KEY_SIZE);

    trimblanks (classStr);
    trimblanks (valueStr);
    vmemSet ((void *)&message[0], 0, sizeof (message));
    if ( (classStr[0] == 0) || (valueStr[0] == 0) )
    {
	opnoMessageCenter (this, OPNO_INFO_NOTICE, opno_AddKeyErr);
	return (vFALSE);
    }
    else if ( (WriteKeyValue (classStr, valueStr, message)) == 0)
    {
	vstr *msg = vstrCloneScribed (vcharScribeSystem (message));
	opnoMessageCenter (this, OPNO_INFO_NOTICE, opno_WriteKeyValueErr, msg);
	vstrDestroy (msg);
	return (vFALSE);
    }
    return (vTRUE);
}


/*******************************************************************
*
*    opNoteKeyDialog delete Key function
*
*******************************************************************/

int opNoteKeyDialog::deleteKey (void)
{
    char	message[MAXERRMSG];
    char	classStr [KEY_SIZE + 1];
    char	valueStr [KEY_SIZE + 1];

    opNoteKeyClassComboBox *keyClass = getKeyClassCombo ();
    opNoteKeyValueComboBox *keyValue = getKeyValueCombo ();

    vcomboboxValue *classEntryBoxValue = 
			keyClass -> GetEntryboxValue (keyClass -> GetEntrybox ());
    const vchar *classEntryBoxTitle = classEntryBoxValue -> GetTitle ();
    vcharExportBounded (vcharScribeSystemExport (classEntryBoxTitle), classStr, KEY_SIZE);
    vcomboboxValue *valueEntryBoxValue = 
			keyValue -> GetEntryboxValue (keyValue -> GetEntrybox ());
    const vchar *valueEntryBoxTitle = valueEntryBoxValue -> GetTitle ();
    vcharExportBounded (vcharScribeSystemExport (valueEntryBoxTitle), 
		        valueStr, 
			KEY_SIZE);

    trimblanks (classStr);
    trimblanks (valueStr);
    vmemSet ((void *)&message[0], 0, sizeof (message));
    if ( ((classStr[0] == 0) && (valueStr[0] == 0) ) ||
	 ((classStr[0] == 0) && (valueStr[0] != 0) ) )
    {
	opnoMessageCenter (this, OPNO_INFO_NOTICE, opno_DeleteKeyErr);
	return (vFALSE);
    }
/*    if ( (classStr[0] != 0) && (valueStr[0] == 0) )
    {
	sprintf (message, "Are you sure you want to delete the entire %s key class", 
			   classStr);
	result = deleteNotice (this, message);
	if (result == vTRUE)
	{
	    if ( (DeleteKeyType (classStr, message)) == 0)
	    {
		dataNotice (this, DELETE_MSG, message);
		return (vFALSE);
	    }
	    else
		setDeletingKeyType (vTRUE);
	}
    }
    else*/ if ( (DeleteKeyValue (classStr, valueStr, message)) == 0)
    {
	vstr *msg = vstrCloneScribed (vcharScribeSystem (message));
	opnoMessageCenter (this, OPNO_INFO_NOTICE, opno_DeleteKeyValueErr, msg);
	vstrDestroy (msg);
	return (vFALSE);
    }
    return (vTRUE);
}


vkindTYPED_FULL_DEFINITION(opNoteKeyCommand, vcommandSelector, NULL);

/*******************************************************************
*
*    opNoteKeyCommand menu selector commands
*
*******************************************************************/

int opNoteKeyCommand::Issue (vdict *dict)
{
    int			index;
    int			error;
    int			selectedIndex;
    int			valueCount;
    char		classStr [KEY_SIZE + 1];
    const vchar		*valueTitle;
    vcomboboxValue	*comboValue;
    vcomboboxValue	*classEntryBoxValue;
    const vchar		*classEntryBoxTitle;

    opNoteKeyDialog *keyDialog = (opNoteKeyDialog *) dict -> GetElem (vname_Dialog);
    opNoteKeyClassComboBox *keyClass = keyDialog -> getKeyClassCombo ();
    opNoteKeyValueComboBox *keyValue = keyDialog -> getKeyValueCombo ();

    switch (GetSelector ())
    {
	case ADD_KEY_BUTTON:
	    error = 0;
	    if ( (keyDialog -> addKey ()) == vFALSE)
	    {
		opnoMessageCenter (keyDialog, OPNO_INFO_NOTICE, opno_AddNotDoneMsg);
		error = 1;
	    }
	    classEntryBoxValue = 
		    keyClass -> GetEntryboxValue (keyClass -> GetEntrybox ());
	    classEntryBoxTitle = classEntryBoxValue -> GetTitle ();
	    vcharExportBounded (vcharScribeSystemExport (classEntryBoxTitle), 
			        classStr, 
				KEY_SIZE);
	    selectedIndex = keyClass -> GetSelectedIndex ();
	    keyClass -> setModifying ();
	    keyClass -> buildKeyClassCombo ();

	    // if no error on add, set class selected index to value in entry box
	    if (error == 0)
	    {
		valueCount = keyClass -> GetValueCount ();

		// search for the string
		for (index = 0; index < valueCount; index++)
		{
		    comboValue = keyClass -> GetValueAt (index);
		    valueTitle = comboValue -> GetTitle();
		    if ( (vcharCompareBounded (classStr, valueTitle, KEY_SIZE)) == 0)
		    {
			keyClass -> SelectValueAt (index);
			break;
		    }
		}
	    }
	    else
	    {
		// build key value combo based on the old class selected item
		keyClass -> SelectValueAt (selectedIndex);
		comboValue = keyClass -> GetValueAt (selectedIndex);
		valueTitle = comboValue -> GetTitle ();
		vcharExportBounded (vcharScribeSystemExport (valueTitle),
				    classStr,
				    KEY_SIZE);
	    }
	    keyValue -> buildKeyValueCombo (classStr);
	    keyClass -> clearModifying ();
	    break;

	case DELETE_KEY_BUTTON:
	    error = 0;
	    keyDialog -> setDeletingKeyType (vFALSE);
	    if ( (keyDialog -> deleteKey ()) == vFALSE)
	    {
		opnoMessageCenter (keyDialog, OPNO_INFO_NOTICE, opno_DeleteNotDoneMsg);
		error = 1;
	    }
	    classEntryBoxValue = keyClass -> GetEntryboxValue (keyClass -> GetEntrybox ());
	    classEntryBoxTitle = classEntryBoxValue -> GetTitle ();
	    vcharExportBounded (vcharScribeSystemExport (classEntryBoxTitle), 
				classStr, KEY_SIZE);
	    selectedIndex = keyClass -> GetSelectedIndex ();
	    valueCount = keyValue -> GetValueCount ();
	    keyClass -> setModifying ();
	    keyClass -> buildKeyClassCombo ();

	    // if no error on add, set class selected index to value in entry box
	    if (error == 0)
	    {
		// if deleted an entire key class or have
		// deleted the last key value set index to 0
		if ( ((keyDialog -> getDeletingKeyType ()) == vTRUE) ||
		     (valueCount == 1) )
		{
		    keyClass -> SelectValueAt (0);
		    // build the key value combo based on the 1st item in the 
		    //key class combo
		    vcomboboxValue *keyValue = keyClass -> GetValueAt (0);
		    const vchar *keyClassTitle = keyValue -> GetTitle ();
		    vcharExportBounded (vcharScribeSystemExport (keyClassTitle), 
					classStr, KEY_SIZE);
		}
		else
		{
		    valueCount = keyClass -> GetValueCount ();
		    // search for the string
		    for (index = 0; index < valueCount; index++)
		    {
			comboValue = keyClass -> GetValueAt (index);
			valueTitle = comboValue -> GetTitle();
			if ( (vcharCompareBounded (classStr, valueTitle, KEY_SIZE)) == 0)
			{
			    keyClass -> SelectValueAt (index);
			    break;
			}
		    }
		}

	    }
	    else
	    {
		// build key value combo based on the old class selected item
		keyClass -> SelectValueAt (selectedIndex);
		comboValue = keyClass -> GetValueAt (selectedIndex);
		valueTitle = comboValue -> GetTitle ();
		vcharExportBounded (vcharScribeSystemExport (valueTitle),
				    classStr,
				    KEY_SIZE);
	    }
	    keyValue -> buildKeyValueCombo (classStr);
	    keyClass -> clearModifying ();
	    break;

	case CLOSE_KEY_BUTTON:
	    keyDialog -> Close ();
	    vevent::StopProcessing ();
	    break;

	default:
	    return (vFALSE);
    }
    return (vTRUE);
}


vkindTYPED_FULL_DEFINITION(opNoteKeyClassComboBox, vcombobox, NULL);

/*******************************************************************
*
*    opNoteKeyClassComboBox LoadInit override
*
*******************************************************************/

void opNoteKeyClassComboBox::LoadInit (vresource res)
{
    vcombobox::LoadInit (res);
    vmemSet ((void *)lastSelected, 0, sizeof (lastSelected));
}   

/*******************************************************************
*
*    opNoteKeyClassComboBox Observe function
*
*******************************************************************/

void opNoteKeyClassComboBox::ObserveCombobox (vevent *event, int message)
{
    if ( (message != vcomboboxNOTIFY_SELECT) &&
	 (message != vcomboboxNOTIFY_ACCEPT) )
	return;

    int index = GetSelectedIndex ();
    int valueCount = GetValueCount ();
    int amModifying = getModifying ();
    if ( amModifying)
	return;
    if ( (index >= 0) && (index < valueCount) )
    {
	char		keyClass[KEY_SIZE + 1];
	
	opNoteKeyDialog *keyDialog = (opNoteKeyDialog *)(DetermineDialog ());
	opNoteKeyValueComboBox *valueCombo = keyDialog -> getKeyValueCombo ();

	vcomboboxValue *entryBoxValue = GetEntryboxValue (GetEntrybox ());
	const vchar *entryBoxTitle = entryBoxValue -> GetTitle ();
	vcharExportBounded (vcharScribeSystemExport (entryBoxTitle), 
			    keyClass,
			    KEY_SIZE);
	char *last = getLastSelected ();

	if ( (strncmp (keyClass, last, KEY_SIZE)) != 0)
	{
	    deleteComboBoxList (valueCombo);
	    valueCombo -> buildKeyValueCombo (keyClass);
	    putLastSelected (keyClass);
	}
    }
}


/*******************************************************************
*
*    opNoteKeyClassComboBox build list function
*
*******************************************************************/

void opNoteKeyClassComboBox::buildKeyClassCombo (void)
{
    char	errmsg[MAXERRMSG];
    char	**list = NULL;

    vmemSet ((void *)errmsg, 0, sizeof (errmsg));
    list = GetKeytypes (errmsg);
    if (errmsg[0] != 0)
    {
	opNoteKeyDialog *keyDialog = (opNoteKeyDialog *)(DetermineDialog ());
	vstr *msg = vstrCloneScribed (vcharScribeSystem (errmsg));
	opnoMessageCenter (keyDialog, OPNO_INFO_NOTICE, opno_KeyClassErr, msg);
	vstrDestroy (msg);
    }
    int valueCnt = GetValueCount ();
    int comboWasOpen = vFALSE;

    // if the combo is open,close it make the build faster
    if ( (comboWasOpen = IsOpen ()) == vTRUE)
	Close ();
    if (valueCnt > 0)
	deleteComboBoxList (this);

    if (list)
    {
	int	cnt;

	// count number of entries in list
	// NOTE: intentional ; at end of for()
	for (cnt = 0; list[cnt]; cnt++);


	// build new combo list
	int		i;
	vcomboboxValue	*value;
	for (i = 0; i < cnt; i++)
	{
	    value = new vcomboboxValue;
	    value -> SetTitleScribed (vcharScribeSystem (list[i]));
	    AppendValue (value);
	    free ((void *)list[i]);
	}
	free ((void *)list);
	SelectValueAt (0);
    }
    if (comboWasOpen == vTRUE)
	Open ();
    vcomboboxValue *entryBoxValue = GetEntryboxValue (GetEntrybox ());
    entryBoxValue -> SetTitleScribed (vcharScribeSystem (""));
}


vkindTYPED_FULL_DEFINITION(opNoteKeyValueComboBox, vcombobox, NULL);

/*******************************************************************
*
*    opNoteKeyValueComboBox build function
*
*******************************************************************/

void opNoteKeyValueComboBox::buildKeyValueCombo (char *pStr)
{
    char	errmsg[MAXERRMSG];
    char	**list = NULL;

    vmemSet ((void *)errmsg, 0, sizeof (errmsg));
    list = GetKeyValues (pStr, errmsg);
    if (errmsg[0] != 0)
    {
	opNoteKeyDialog *keyDialog = (opNoteKeyDialog *)(DetermineDialog ());
	vstr *vStr = vstrCloneScribed (vcharScribeSystem (pStr));
	vstr *msg = vstrCloneScribed (vcharScribeSystem (errmsg));
	opnoMessageCenter (keyDialog, OPNO_INFO_NOTICE, opno_KeyValueErr, vStr, msg);
	vstrDestroy (vStr);
	vstrDestroy (msg);
    }

    int valueCnt = GetValueCount ();
    int comboWasOpen = vFALSE;

    // if the combo is open,close it make the build faster
    if ( (comboWasOpen = IsOpen ()) == vTRUE)
	Close ();
    if (valueCnt > 0)
	deleteComboBoxList (this);

    if (list)
    {
	int	cnt;

	// count number of entries in list
	// NOTE: intentional ; at end of for()
	for (cnt = 0; list[cnt]; cnt++);

	// build new combo list
	int		i;
	vcomboboxValue	*value;
	for (i = 0; i < cnt; i++)
	{
	    value = new vcomboboxValue;
	    value -> SetTitleScribed (vcharScribeSystem ((void *)list[i]));
	    AppendValue (value);
	    free ((void *)list[i]);
	}
	free ((void *)list);
	SelectValueAt (0);
    }
    if (comboWasOpen == vTRUE)
	Open ();
}
