#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vbuttonHEADER
#include vtextitemHEADER
#include vstartupHEADER
#include vtimeHEADER
#include vmemHEADER
#include vnoticeHEADER
#include vcontrolHEADER
#include vtimerHEADER
extern "C"
{
#include "opNoteDb.h"
}
#include "opno.h"
#include "opNoteUtils.hxx"
#include "opNote.hxx"

static int	doStrCmp (char **a, char **b);

static int	dayTable[2][12] = {
	{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
	{31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}
	};

/*******************************************************************
*
*    opnoMessageCenter ()
*	handle messages to user
*
*******************************************************************/

int opnoMessageCenter (vdialog *dialog, int msgType, const vname *msgName, ...)
{
    int		result = vTRUE;
    va_list	argptr;

    va_start (argptr, msgName);

    vapplication *opnoApp   = vapplication::GetCurrent ();
    vresource    opnoRes    = opnoApp -> GetResources ();
    vresource    titleRes   = vresourceGet (opnoRes, opno_titlesDictionary);
    vresource    msgRes     = vresourceGet (opnoRes, opno_messagesDictionary);
    vstr         *infoTitle = vresourceGetString (titleRes, opno_InfoNotice);
    vstr         *cautionTitle = vresourceGetString (titleRes, opno_CautionNotice);

    vstr *format = vresourceGetString (msgRes, msgName);
    // NOTE: vstrScribe cleans up vstr *format
    vstr *msg = vstrCloneScribed (vcharScribeFormatScribedVarargs (vstrScribe (format), argptr));

    va_end (argptr);

    if (msgType == OPNO_INFO_NOTICE)
    {
	infoNotice (dialog, infoTitle, msg);
    }
    else
    {
	result = cautionNotice (dialog, cautionTitle, msg);
    }
    vstrDestroy (infoTitle);
    vstrDestroy (cautionTitle);
    vstrDestroy (msg);

    return (result);
}


/*******************************************************************
*
*    dateTimeErrorNotice ()
*	handle messages to user about dates and times
*
*******************************************************************/
void dateTimeErrorNotice (vdialog *dialog, DateTimeErrors msgType)
{
    vapplication *opnoApp  = vapplication::GetCurrent ();
    vresource    opnoRes   = opnoApp -> GetResources ();
    vresource    arraysRes = vresourceGet (opnoRes, opno_arraysDictionary);
    // dtRes is a resource for an array of date/time error msgs
    vresource    dtRes     = vresourceGet (arraysRes, opno_dateTimeErrorMsgs);
    int msgCount = vresourceCountComponents (dtRes);

    if ( (msgType < OPDATE_OK) || (msgType >= MAX_DATETIME_ERRORS) )
    {
	opnoMessageCenter (dialog, OPNO_INFO_NOTICE,
			   opno_BadDateTimeErr, (int)msgType);
	return;
    }
    vstr *errMsg = vresourceGetStringAt (dtRes, (int)msgType);
    opnoMessageCenter (dialog, OPNO_INFO_NOTICE, opno_DateTimeError, errMsg);
    vstrDestroy (errMsg);
    return;
}


/*******************************************************************
*
*    doStrCmp ()
*	argument to qsort
*
*******************************************************************/

int	doStrCmp (char **a, char **b)
{
    return (strcmp (*a, *b));
}


/*******************************************************************
*
*    delete list function for a vcombobox
*
*******************************************************************/

void	deleteComboBoxList (vcombobox *comboBox)
{
    vcomboboxValue	*value;
    int			valueCnt;
    int			i;

    valueCnt = comboBox -> GetValueCount ();

    for (i = valueCnt - 1; i >= 0; i--)
    {
	value = comboBox -> GetValueAt (i);
	comboBox -> DeleteValueAt (i);
	delete value;
    }
    // wipe out the entry box also
//    value = comboBox -> GetEntryboxValue (comboBox -> GetEntrybox ());
//   value -> SetTitleScribed (vcharScribeSystem (""));
}



/*******************************************************************
*
*    extractDate ()
*
*******************************************************************/

DateTimeErrors	extractDate (vchar *dateString, struct tm *tim)
{
    int		isLeap = 0;
    int		year;
    int		dayMonth;
    char	date[TIME_SIZE + 1];

    vcharExportBounded (vcharScribeSystemExport (dateString), date, sizeof (date));
    year = atoi (date + 6);
    tim -> tm_year = year - 1900;
    if (tim -> tm_year < 0)
	return (OPYEAR_ERROR);
    isLeap = year%4 == 0 && year%100 != 0 || year%400 == 0;
    tim -> tm_mon = atoi (date) - 1;
    if ( (tim -> tm_mon < 0) || (tim -> tm_mon > 11) )
	return (OPMONTH_ERROR);
    tim -> tm_mday = atoi (date + 3);
    dayMonth = dayTable[isLeap][tim -> tm_mon];
    if ( (tim -> tm_mday < 0) || (tim -> tm_mday > dayMonth) )
	return (OPDAY_ERROR);
    return (OPDATE_OK);
}


/*******************************************************************
*
*    extractTime ()
*
*******************************************************************/

DateTimeErrors	extractTime (vchar *timeString, struct tm *tim)
{
    char	time[TIME_SIZE];

    if (timeString[0] == 0)
    {
	tim -> tm_hour = 0;
	tim -> tm_min  = 0;
	tim -> tm_sec  = 0;
    }
    else
    {
	vcharExportBounded (vcharScribeSystemExport (timeString), 
			    time, sizeof (time));
	tim -> tm_hour = atoi (time);
	if ( (tim -> tm_hour < 0) || (tim -> tm_hour > 23) )
	    return (OPHOUR_ERROR);
	tim -> tm_min = atoi (time + 3);
	if ( (tim -> tm_hour < 0) || (tim -> tm_min > 59) )
	    return (OPMIN_ERROR);
	if ( (vcharLength (timeString)) > 6)
	{
	    tim -> tm_sec = atoi (time + 6);
	    if ( (tim -> tm_sec < 0) || (tim -> tm_sec > 59) )
		return (OPSEC_ERROR);
	}
	else
	    tim -> tm_sec = 0;
    }
    return (OPDATE_OK);
}


vkindTYPED_FULL_DEFINITION(opNoteControl, vcontrolOptionMenu, NULL);
/*******************************************************************
*
*    notify override function for opNoteControl class
*
*******************************************************************/
void opNoteControl::Notify (vevent *event)
{
    // get pointer to menu attached to the control
    vmenu *menu = this -> GetMenu ();

    // get current value of control
    int controlValue = this -> GetValue ();

    if ( (controlValue >= this -> GetMin ()) &&
	 (controlValue <= this -> GetMax ()) )
    {
	vmenuItem *menuItem = menu -> GetItemAt (controlValue);
	vchar *pvChar = (vchar *)menuItem -> GetTitle ();
	if (pvChar)
	{
	    buildKeyValueMenu (pvChar);
	}
    }
}


/*******************************************************************
*
*    buildKeyClassMenu ()
*	build a popup menu and attach it to a vcontrol object
*	GetKeytypes can be found in opNoteDb.c
*	NOTE: the menu attribute for key class is destroyed when
*		it is changed, so vmenuDestroy is not needed
*
*******************************************************************/

void opNoteControl::buildKeyClassMenu (int useDefault)
{
    char	**list = NULL;
    char	errmsg[MAXERRMSG];
    int		cnt = 0;
    int		x;
    vmenuItem	*menuItem;
    vdialog	*dialog = NULL;

    /* create a new menu to attach to the vcontrol item, we are creating a
       popup menu within a dialog
    */
    vmenu *newMenu = new vmenu;
    
    if (useDefault)
    {
	/* set the default menu item */
	menuItem = new vmenuItem;
	menuItem -> SetTitleScribed (vcharScribeSystem (KEY_CLASS_DEFAULT));
	newMenu -> AppendItem (menuItem);
    }

    vmemSet ((void *)&errmsg[0], 0, sizeof (errmsg));
    list = GetKeytypes (errmsg);
    if (errmsg[0] != 0)
    {
	vstr *msg = vstrCloneScribed (vcharScribeSystem (errmsg));
	opnoMessageCenter (NULL, OPNO_INFO_NOTICE, opno_KeyClassErr, msg);
	vstrDestroy (msg);
    }
    if (list)
    {
	/* count the number of entries in the list
	   NOTE: intentional ; after for ()
	*/
	for (cnt = 0; list[cnt]; cnt++);

	qsort ((char *)list, cnt, sizeof (char *), 
		(int (*)(const void *, const void *)) doStrCmp);
	/* build the menu list for newMenu */
	for (x = 0; x < cnt; x++)
	{
	    menuItem = new vmenuItem;
	    menuItem -> SetTitleScribed (vcharScribeSystem ((void *)list[x]));
	    newMenu -> AppendItem (menuItem);
	    free ((void *)list[x]);
	}

	free ((void *)list);
    }
    // attach the menu to the control
    SetMenu (newMenu);
    // set min and max for the control
    SetMin (0);
    if (useDefault)
	SetMax (cnt);
    else
	SetMax (cnt - 1);
    // set value so current menu item will be the 1st one
    SetValue (0);
}


/*******************************************************************
*
*    buildKeyValueMenu ()
*	build a popup menu and attach it to a vcontrol object
*	NOTE: the menu attribute for key value is destroyed when
*		it is changed, so vmenuDestroy is not needed
*
*******************************************************************/

void opNoteControl::buildKeyValueMenu (vchar *pvChar)
{
    vmenuItem	*menuItem;
    int		cnt = -1;
    int		x;
    char	**list = NULL;
    char	errmsg[MAXERRMSG];

    // extract the keyValue ptr from the data area of the keyClass control
    vcontrol *keyVal =  getKeyValue ();

    /* create a new menu to attach to the vcontrol item, we are creating a
       popup menu within a dialog
    */
    vmenu *newMenu = new vmenu;
    vstr  *classDef = vstrCloneScribed (vcharScribeSystem (KEY_CLASS_DEFAULT));
    if ( (vcharCompare (pvChar, classDef)) == 0)
    {
	menuItem = new vmenuItem;
	menuItem -> SetTitleScribed (vcharScribeSystem (KEY_VALUE_DEFAULT));
	newMenu -> AppendItem (menuItem);
	// set min and max for the control
	keyVal -> SetMin (0);
	keyVal -> SetMax (0);
    }
    else
    {
	vmemSet ((void *)&errmsg[0], 0, sizeof (errmsg));
	char pStr[KEY_SIZE + 1];
	vcharExportBounded (vcharScribeSystemExport (pvChar), (void *)&pStr[0], KEY_SIZE);
	list = GetKeyValues (pStr, errmsg);
	if (errmsg[0] != 0)
	{
	    vstr *vStr = vstrCloneScribed (vcharScribeSystem (pStr));
	    vstr *msg = vstrCloneScribed (vcharScribeSystem (errmsg));
	    opnoMessageCenter (NULL, OPNO_INFO_NOTICE, opno_KeyValueErr, vStr, msg);
	    vstrDestroy (vStr);
	    vstrDestroy (msg);
	}
	if (list)
	{

	    /* count the number of entries in the list
	       NOTE: intentional ; after for ()
	    */
	    for (cnt = 0; list[cnt]; cnt++);

	    qsort ((char *)list, cnt, sizeof (char *), 
		   (int (*)(const void *, const void *)) doStrCmp);

	    // build the menu list for newMenu
	    for (x = 0; x < cnt; x++)
	    {
		menuItem = new vmenuItem;
		menuItem -> SetTitleScribed (vcharScribeSystem ((void *)list[x]));
		newMenu -> AppendItem (menuItem);
		free ((void *)list[x]);
	    }
	    // set min and max for the control
	    keyVal -> SetMin (0);
	    keyVal -> SetMax (cnt - 1);

	    free ((void *)list);
	}
    }
    vstrDestroy (classDef);
    // attach the menu to the control
    keyVal -> SetMenu (newMenu);
    // set value so current menu item will be the 1st one
    keyVal -> SetValue (0);
}

