/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#ifdef SOLARIS
#include <string.h>
#else
#include <strings.h>
#endif

#include <Gina/Gina.h>

extern "C"
{
#include <stdlib.h>
#include "avl.h"
#include "ctype.h"
#include "logutils.h"
#include "msgMon.h"

#ifndef MIPSEL
#include "memory.h"
#endif
}
#include <unistd.h>

#include "loginShell.h"
#include "operDefs.h"

// System error logins options
int operCompare(OPERATOR_ENTRY_PTR,OPERATOR_ENTRY_PTR);

#define	LOG_OPTIONS	LOG_ALERT | LOG_LOCAL1

class LoginDocument : public GnDocument {
public:
    void	CreateWindows(int new_width, int new_height);    
protected:
    LoginDialog *loginDialog;
};

class LoginApplication : public GnApplication {
public:
		LoginApplication (void);
    GnDocument	*CreateDocument ( );
    char	*app_class () { return "tisLogin" ; }
    char	*app_signature() { return "tisLogin" ; }
    char	*app_file_type() { return ".tisLogin" ; }
    void	IdleAction ();
private:
    GnDocument	*document;
    int		titlesInitialized;
};

LoginDocument		*theLoginDocument;
LoginShell		*theLoginShell;

// The path to the login file

char	loginFilePath [200];

// List of possible operators

int		numOperators;
char		displayHost [50];
int		menuChange;
AVL_TREE  *operInfoTree;
//----------
LoginApplication :: LoginApplication (void)

    {

	FILE *f;
	char  line[SLM_MAX_LINE_SIZE];

	titlesInitialized = 0;


	tisOpenLog("tisLogin", 6, 2, NULL, NULL);

	// Make sure the /tis4000/operators directory exists

	strcpy (loginFilePath, "/tis4000/operators");
	if (access (loginFilePath, R_OK | W_OK) != 0) {
	    tisSysLog (0,0, "/tis4000/operators directory not found!");
	    exit (-1);
	}

	// If this is the first time a login has ever occurred then it is likely that
	// /tis4000/operators/logins does not exist. Create it if it doesn't.

	strcat (loginFilePath, "/logins");
	if (access (loginFilePath, R_OK | W_OK) != 0) {
	    if (mkdir (loginFilePath, 0700) != 0) {
		tisSysLog (0,0, "Unable to create /tis4000/operators/logins");
		exit (-1);
	    }
	}

	// If a login file does not exist for this host. Create one with the "nooperator"
	// operator.  This indicates that no one is currently logged in.

	strcat (loginFilePath, "/");
	strcpy (displayHost, getDisplayHost ());
	strcat (loginFilePath, displayHost);
	if (access (loginFilePath, R_OK | W_OK) != 0) {
	    if ((f = fopen (loginFilePath, "w")) == NULL) {
		tisSysLog (0,0, "Unable to create %s",loginFilePath);
		exit (-1);
	    }
	    fputs ("nooperator\n", f);
	    fclose (f);
	}

	// Build a tree of names and passwords
	operInfoTree = avl_init(operCompare);
	numOperators = 0;
	OPERATOR_ENTRY_PTR  pEntry;
	if((f = fopen("/tis4000/operators/secureList","r")) != 0)
	  {
	    while((fgets(line,SLM_MAX_LINE_SIZE-1,f)) != 0)
	      {
		line[strlen(line)-1] = 0;			/* drop \n and NULL terminate */
		pEntry = (OPERATOR_ENTRY_PTR)avl_newnode(OPERATOR_ENTRY_SIZE);	/* create new node */
		strcpy(pEntry->name,strtok(line," "));		/* get name */
		strcpy(pEntry->password,strtok(NULL," "));	/* then keypad ID(password) */
		strtok(NULL," ");				/* skip card ID */
		strtok(NULL," ");				/* and logout time */
		strtok(NULL," ");				/* and GLOBAL flag */
		pEntry->opi = (!strcmp(strtok(NULL," "),"NOOPI") ? 0 : 1); /* to get the OPI flag */
		avl_insert(operInfoTree,pEntry);
		numOperators++;
	      }
	    fclose(f);
	  }
    }
int operCompare(OPERATOR_ENTRY_PTR p1,OPERATOR_ENTRY_PTR p2)
{
return(strcmp(p1->name,p2->name));
}
//----------
GnDocument *LoginApplication :: CreateDocument ( )

    {

	LoginDocument *document;

	// Initialize things

	document = new LoginDocument;

	theLoginDocument = document;
	theLoginShell = NULL;

	return document;

    }

//----------
void LoginApplication :: IdleAction ()

    {

	if (!titlesInitialized) {
	    theLoginDocument->GetMainShell ()->setR_title ("TIS4000 Operator Login");
	    theLoginDocument->GetMainShell ()->setR_iconName("tisLogin");
	    theLoginDocument->GetMainShell ()->SetValues();
	    titlesInitialized = 1;
	}

    }

//----------
void LoginDocument :: CreateWindows(int /*new_width*/, int /*new_height*/)

    {

	theLoginShell = new LoginShell (this, 500, 800);

	SetMainShell (theLoginShell);
	GetMainShell ()->create(APPLICATION->GetApplicationShell(), "tisLogin");

    }

//----------
int main (unsigned int argc, char ** argv)
{
  LoginApplication app;
  int	c;
  // Parse the command line options
  menuChange = 1;
  while ((c = getopt (argc, argv, "n")) != -1)
    switch (c) {
    case 'n':	// No menu change support
      menuChange = 0;
      break;
    }
  // setup fallback resources
  app.AddFallbackResource("tisLogin*fontList: 8x13bold");
  app.AddFallbackResource("tisLogin*background: gray");
  app.AddFallbackResource("tisLogin*foreground: black");
  // Set up the idle timeout
  app.SetIdleTimeout (1000);
  app.Run ( 1, argv );
}

