/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#ifdef SOLARIS
#include <string.h>
#else
#include <strings.h>
#endif

#include <Gina/Gina.h>

extern "C"
{
#include "logutils.h"
#include "ctype.h"
#include "logutils.h"
#include "msgMon.h"

#ifndef MIPSEL
#include "memory.h"
#endif
}

#include "userMaintShell.h"
#include "operDefs.h"

// System error userMaints options

#define	LOG_OPTIONS	LOG_ALERT | LOG_LOCAL1

class UserMaintDocument : public GnDocument {
public:
    void	CreateWindows(int new_width, int new_height);    
protected:
    MaintDiag *maintDiag;
};

class UserMaintApplication : public GnApplication {
public:
		UserMaintApplication (void);
    GnDocument	*CreateDocument ( );
    char	*app_class () { return "tisusermaint" ; }
    char	*app_signature() { return "tisusermaint" ; }
    char	*app_file_type() { return ".tisusermaint" ; }
    void	IdleAction ();
private:
    GnDocument	*document;
    int		titlesInitialized;
};

UserMaintDocument	*theUserMaintDocument;
UserMaintShell		*theUserMaintShell;

// List of possible operators

OperatorList	operList;
int		numOperators;

// Functions

static int	compar (void *e1, void *e2);

char	displayHost [50];

//----------
UserMaintApplication :: UserMaintApplication (void)

    {

	FILE *f;

	titlesInitialized = 0;

	// Make sure the /tis4000/operators directory exists. If it doesn't create it and
	// put an empty tisOperators file in it.

	if (access ("/tis4000/operators", R_OK | W_OK) != 0) {
	    tisSysLog (0,0, "tisUserMaint(%d): /tis4000/operators directory not found - created.", getpid ());
	    mkdir ("/tis4000/operators", 0700);
	    f = fopen ("/tis4000/operators/tisOperators", "w");
	    fclose (f);
	}

	// Build a list of names and passwords

	f = fopen ("/tis4000/operators/tisOperators", "r");
	numOperators = 0;
	while (fscanf (f, "%s %s", operList [numOperators].name, operList [numOperators].password) == 2) {
	    operList [numOperators].deleted = 0;
	    numOperators++;
	}
	fclose (f);

	// Get the display Host

	strcpy (displayHost, getDisplayHost ());

    }

//----------
GnDocument *UserMaintApplication :: CreateDocument ( )

    {

	UserMaintDocument *document;

	// Initialize things

	document = new UserMaintDocument;

	theUserMaintDocument = document;
	theUserMaintShell = NULL;

	return document;

    }

//----------
void UserMaintApplication :: IdleAction ()

    {

	if (!titlesInitialized) {
	    theUserMaintDocument->GetMainShell ()->setR_title ("TIS4000 Operator Maintenance");
	    theUserMaintDocument->GetMainShell ()->setR_iconName("tisUserMaint");
	    theUserMaintDocument->GetMainShell ()->SetValues();
	    titlesInitialized = 1;
	}

    }

//----------
void UserMaintDocument :: CreateWindows(int /*new_width*/, int /*new_height*/)

    {

	theUserMaintShell = new UserMaintShell (this, 500, 800);

	SetMainShell (theUserMaintShell);
	GetMainShell ()->create(APPLICATION->GetApplicationShell(), "shell");

    }

//----------
int main (unsigned int /*argc*/, char ** argv)

    {

    UserMaintApplication app;

    // Set up the idle timeout

    app.SetIdleTimeout (1000);

    app.Run ( 1, argv );
}

//----------
int addUser (char *userName)	// Add a new user to the list

    {

	// Check to see if the user is already in the list

	for (int i=0 ; i<numOperators ; i++)
	    if ((strcmp (userName, operList [i].name) == 0) && !operList [i].deleted)
		return 0;

	// Insert the new user

	strcpy (operList [numOperators].name, userName);
	strcpy (operList [numOperators].password, "");
	operList [numOperators].deleted = 0;
	numOperators++;

	// Sort the list in ascending order

	qsort (operList, numOperators, sizeof (OperatorEntry), compar);

	return 1;

    }

//----------
static int compar (void *e1, void *e2)

    {

	return strcmp (((OperatorEntry *)e1)->name, ((OperatorEntry *)e2)->name);

    }

//-----------
void deleteUser (char *userName)

    {

	// Find the user in the list

	for (int i=0 ; i<numOperators ; i++)
	    if (strcmp (userName, operList [i].name) == 0)
		break;

	operList [i].deleted = 1;

    }

//-----------
void setPassword (char *userName, char *password)

    {

	// Find the user in the list

	for (int i=0 ; i<numOperators ; i++)
	    if (strcmp (userName, operList [i].name) == 0)
		break;

	strcpy (operList [i].password, password);

    }
