/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
#undef DEBUG

#include <stdio.h>
#include <Xlib.h>      			/* the X library 		*/
#include <Xutil.h>
#include <db_access.h>			/* db types and structs		*/
#include <alarm.h>
#include <cadef.h>
#include <dldef.h>			/* display list defines 	*/
#include <dlinfo.h>			/* more display list defines 	*/
#include <font_table.h>			/* defines for font table	*/
#include <dl_context.h>			/* defines dl context		*/
#include <color_map.h>
#include <dynamic_def.h>		/* rti dynamic defines		*/
#include <op_gsdl_def.h>		/* com_table element def	*/

#ifdef AXP
#include <string.h>
#endif

extern void *tisMalloc();
extern void *tisCalloc();
extern void tisFree();

static  int  isNumeric (char *);

#define XDEG_PER_REV	(360*64)
#define BITS_PER_INT	(sizeof(int) * 8)

/* THIS MUST BE AT LEAST AS BIG AS MAX_STRING_SIZE IN DB_ACCESS! */
#define MAX_CHARS	MAX_STRING_SIZE

/*----------*/
static void refreshRectangle (Display *d, Window w, XRectangle rec)
{
  static XExposeEvent  theEvent;

  theEvent.type = Expose;
  theEvent.display = d;
  theEvent.window = w;
  theEvent.x = rec.x;
  theEvent.y = rec.y;
  theEvent.width = rec.width;
  theEvent.height = rec.height;
  XSendEvent (d, w, (Bool)0, 0L, (XEvent *)&theEvent);
}

/***********************************************************************
**
**	UPDATE_DYNAMIC_GRAPHICS
**	
**	Deb Kerstiens 12/90
**
**	Purpose : This routine draws the dynamic graphic-type objects from a display list
**		into an X Window.  It is done via two passes through the display and run-time
**		lists.  The first pass draws errors found, clears anything to be erased, redraws
**		static elems affected by clear, and marks any objects that should be drawn.  
**		The second pass draws the dynamic objects flagged for drawing in the first pass.
**		
**	Input :
**	    pdlc	: pointer to Display context structure 
**	    pdl_index	: color map index from display list
**	    pdl		: ptr to display list info
**	    pdyn	: ptr to the rti_dynamic list
**	    ndyn	: count of dynamic elements to display
**	    prec	: area to be updated (for expose)
**	    pDm_llist_el: ptr to dm llist el for static, nwords stat
**	    expose_flag	: indicates whether expose event or
**				regular update pass
**	    dis_bg	: index into dl_color_map for display background
**	Output :
**	    none
**	Returns
**	    none
**
**	Notes: 	
**		There are parallel lists (the display list and the run-time list) and a
**		graphic elem from each is processed in each iteration of the for loop.
**		Attribute elems (fill, line-style, etc.)are included in the display
**		list, but not in the run-time list.  Therefore these elements are not 
**		included in the count of graphics elems on the lists and will be handled as
**		they occur(0,1 or more) in an iteration until the next graphic element from
**		the display list is encountered and processed.  For efficiency the 
**		rti_dynamic_graphic struct could be expanded to include storage for this
**		info and filled in at initialization and never processed again.    		
**
**		The two pass approach was necessary for clearing the errors drawn as this
**		caused a clear of an already-drawn dynamic object if more than one object 
**		was in the same geographic location, i.e. clear error then draw obj, clear
**		error again, then draw second obj--first obj was affected by second clear.
**
**		Clipping and bounds checking can be turned off by setting the
**		XRectangle pointer to nil.
**
**
**	Modification:
**		DK	2/91	added fill_zero or fill_not_zero modification (processing
**				of static part of display list for area needed)
**		DK	2/91	modifed zero, not zero  as visibility mod (not color mode)
**				to correspond to changes in the display list
**		DK	9/91	added error checking and display(dmDisplayPvErr)
**				modified set_fg to use get_fgbg;
**				reorganized code for better readability--created 
**				dmDraw_dyn from previously in-line code
**		DK	10/91	changed text to use draw_fixed_??_string to avoid
**				changing sizes and fonts
**		DK	10/91	adding error checking above caused problem in erase logic--
**				fixed by restructuring to make two passes through list.
**				First pass does errors, clears necessary areas, redraws
**				static, and marks objects that should be drawn.  Second pass
**				draws dynamics marked for update in first pass.
**				pass
**		DK	10/91	fixed text to align as specified in display list(left, center,
**				etc.)
**
**********************************************************************/

void update_dynamic_graphics(pdlc, pStatic_dlc, pdl_index, pdl, 
	prti, cnt, prec, pDm_llist_el, expose_flag, dis_bg)
struct dl_context		*pdlc;
struct dl_context		*pStatic_dlc;
#ifndef AXP
int 				*pdl_index;
#else
unsigned long			*pdl_index;
#endif
void 				*pdl;
struct rti_dynamic_graphic 	*prti;
unsigned short			cnt;
struct dm_llist_el		*pDm_llist_el;
XRectangle			*prec;
int				expose_flag;
unsigned short			dis_bg;	
{
 	float				cvrt_to_float();
	void				*dis_ptr;
	unsigned short			i;
	struct rti_dynamic_graphic	*pRti_dyn; 
	XGCValues			svGC_values;

	/* save gc vals to put back when done */
	XGetGCValues(pdlc->d, pdlc->gc, GCBackground|GCForeground|GCLineWidth|GCLineStyle, &svGC_values);

	/* for each element on the display list */
	for ( dis_ptr = pdl, pRti_dyn = prti, i = 0;  
	      i < cnt; 
	      dis_ptr = ((short *)dis_ptr + ((struct dl_element_hdr *)dis_ptr)->nwords), i++, pRti_dyn++){
		int			*pSet_mod;	/* ptr to set modified array */
		struct dl_object_hdr	*objptr; 
 		int			x_1, x_2, y_1, y_2, w, h;
		unsigned short		pv_err;		/* indicates no ca cnxn, db_type errs */
		XRectangle		rec;
#ifndef AXP
		int	rule_fg;
#else
		long	rule_fg;
#endif

		if(dl_element_class[((struct dl_element_hdr *)dis_ptr)->type] != dl_class_graphic){
			if(((struct dl_element_hdr *)dis_ptr)->type == dl_dynamic_attribute)
				rule_fg = ((struct dl_dynamic_attribute *)dis_ptr)->attr.param.rulecolorfg;

			pRti_dyn--;	/* count only the objects, not the attributes */
			i--;
			continue;	/* drop thru to next iteration of for */
		}

		objptr = &(((struct dl_anygo *)dis_ptr)->object);
		x_1 = objptr->x;
		y_1 = objptr->y;
		x_2 = (int)(objptr->x + objptr->width);
		y_2 = (int)(objptr->y + objptr->height);
		if (((pRti_dyn->fill == dl_fill_outline) || (((struct dl_anygo *)dis_ptr)->element.type == dl_line)) &&
			(((struct dl_anygo *)dis_ptr)->element.type != dl_text)) {
			int	lw	= (pRti_dyn->width == 0) ? 1 : pRti_dyn->width;
			x_1 -= lw;
			y_1 -= lw;
			x_2 += lw;
			y_2 += lw;
		}
		rec.x		= x_1;
		rec.y		= y_1;
		rec.width	= x_2 - x_1;
		rec.height	= y_2 - y_1;

		pSet_mod = (pRti_dyn->pComel->pMod_bits);
		/* check for valid ca connection-- not valid if marked as not connected or no valid data yet */
		if (pRti_dyn->pComel->pChan)
			pv_err = dmChkPvStat(pRti_dyn->pComel, 0, pRti_dyn->not_conn);
		else
			pv_err = 0;

		if(expose_flag){
			/* update exposed area */
			if( x_2 < prec->x ) {
			    continue;
			}
    			if( x_1 > (prec->x + (int)prec->width) ) {
			    continue;
			}
    			if( y_2 < prec->y ) {
			    continue;
			}
    			if( y_1 > (prec->y + (int)prec->height) ) {
			    continue;
			}
			if(pv_err){
				dmDisplayPvErr(objptr,pv_err, pRti_dyn->pComel, pdlc, pRti_dyn->pLast_val);
				pRti_dyn->not_conn = pv_err;
				continue;
			}
			else if (!pRti_dyn->pComel->pChan) {
				pRti_dyn->update = TRUE;
				continue;
			}
		}
		else{
			if (!pRti_dyn->pComel->pChan) {
				if (pRti_dyn->clrrule_has_changed) {
					pRti_dyn->update = TRUE;
				}
				continue;
			}
			if(pv_err){
				/* if same error as last time, no graphics */
				if(pRti_dyn->not_conn && (pRti_dyn->not_conn == pv_err)){
					continue;
				}
				else{	
					/* display error condition */
					dmDisplayPvErr(objptr, pv_err, pRti_dyn->pComel, pdlc, pRti_dyn->pLast_val);
					pRti_dyn->not_conn = pv_err;
					continue;
				}
			}
			else{
				/* we have connection and data area */
				/* check for modification */
				/* if the dynamic chan has not been modified, no graphics */
				if((!(pSet_mod[pRti_dyn->mod_intix] & ~clear_mod_bit_mask[pRti_dyn->mod_bitix])) &&
				!pRti_dyn->clrrule_has_changed){
					continue;
				}
			}
		}

		/* if we made it here, we need to update the display -- either exposed or modified */
		if(pv_err == dmConnection_okay){
			float	fval;

			/* update background objects if obj was drawn last time but not now */
			if(!pRti_dyn->pLast_val)
				pRti_dyn->pLast_val = 
					(float *)tisCalloc(__FILE__, __LINE__, sizeof(float), 1);
			fval = cvrt_to_float(pRti_dyn->pComel, ca_field_type(pRti_dyn->pComel->pChan));
			dmSave_float(&pRti_dyn->pLast_val, fval);		/* save current value */
			if( ((pRti_dyn->vis == dl_vismod_zero) && (fval != 0.0)) ||
			    ((pRti_dyn->vis == dl_vismod_not_zero) && (fval == 0.0)) ||
				((pRti_dyn->clrrule_has_changed) && (rule_fg == invisibleColor)) ){
				/* clear the area and redraw the elements that clear affected */
				/* clear modified bit */
				/* reset connection flag */

				XClearArea(pdlc->d, pdlc->w, (int)rec.x, (int)rec.y, rec.width, rec.height, FALSE);
				draw_static(pStatic_dlc,pDm_llist_el->pstatic, pDm_llist_el->pdl_hdr->nwords_sta, &rec, FALSE);
				update_mon_objs(pdlc, pDm_llist_el -> pdl_color_map_index, pDm_llist_el -> pdl_mon, pDm_llist_el -> prti_mon,
					pDm_llist_el -> pdl_hdr -> nmon, &rec, pDm_llist_el, TRUE);
				pSet_mod[pRti_dyn->mod_intix] =
				    pSet_mod[pRti_dyn->mod_intix] & clear_mod_bit_mask[pRti_dyn->mod_bitix];
				pRti_dyn->not_conn = pv_err;
				if (rule_fg == invisibleColor)
					pRti_dyn->clrrule_has_changed = 0;
				continue;
			}
			if(pRti_dyn->not_conn){
				/* clear out error rectangle*/
  				/* draw the elements that clear affected */
				XClearArea(pdlc->d, pdlc->w, (int)rec.x, (int)rec.y, rec.width, rec.height, FALSE);
				draw_static(pStatic_dlc,pDm_llist_el->pstatic, pDm_llist_el->pdl_hdr->nwords_sta, &rec, FALSE);
				update_mon_objs(pdlc, pDm_llist_el -> pdl_color_map_index,
						pDm_llist_el -> pdl_mon, pDm_llist_el -> prti_mon,
						pDm_llist_el -> pdl_hdr -> nmon, &rec, pDm_llist_el, TRUE);

/*				refreshRectangle (pdlc->d, pdlc->w, rec);
*/			}
			pRti_dyn->update = TRUE;
			pSet_mod[pRti_dyn->mod_intix] =
			    pSet_mod[pRti_dyn->mod_intix] & clear_mod_bit_mask[pRti_dyn->mod_bitix];
			pRti_dyn->not_conn = pv_err;
		} /* end if connection okay */
	} /* end first pass */

	/* for each element on the display list, run-time list */
	for ( dis_ptr = pdl, pRti_dyn = prti, i = 0;  
	      i < cnt; 
	      dis_ptr = ((short *)dis_ptr + ((struct dl_element_hdr *)dis_ptr)->nwords)){
#ifndef AXP
		int	rule_fg;
#else
		long	rule_fg;
#endif
		/*
		 *  update the GC from the attributes in display list if basic attribute
 		 *	   else fill in dynamic info
		 *  Note:  it has been guaranteed that the first item in the
		 *      display list will be an attribute
		 */
  		if(dl_element_class[((struct dl_element_hdr *)dis_ptr)->type] != dl_class_graphic){
			if(((struct dl_element_hdr *)dis_ptr)->type == dl_dynamic_attribute){
				rule_fg = ((struct dl_dynamic_attribute *)dis_ptr)->attr.param.rulecolorfg;
			}
		}
		else { /* do dynamic draw */
			if(pRti_dyn->update){
				XGCValues	GC_values;

				GC_values.line_width = pRti_dyn->width;
				switch(pRti_dyn->style){
				    default:
				    case dl_edge_solid:	GC_values.line_style = LineSolid;     break;
				    case dl_edge_dash:	GC_values.line_style = LineOnOffDash; break;
				}
				XChangeGC(pdlc->d, pdlc->gc, GCLineWidth|GCLineStyle, &GC_values);
				pdlc->fill_style = pRti_dyn->fill;
				if (set_fg(pdlc, pdl_index, pRti_dyn, pRti_dyn->clr, pRti_dyn->clrmod, rule_fg))
					dmDraw_dyn(pdlc, &(((struct dl_anygo *)dis_ptr)->object), dis_ptr);
				pRti_dyn->update = FALSE;
				pRti_dyn->clrrule_has_changed = 0;
#ifdef DEBUG
printf("~d clearing (%s) at:0x%x\n", pRti_dyn->pComel->pChan ? ca_name(pRti_dyn->pComel->pChan) : "", &pRti_dyn->clrrule_has_changed);
#endif
			}
			i++;
			pRti_dyn++;
		}
	}/* end second pass */

	/* set GC back so everything is like it was when we came in */
	XChangeGC(pdlc->d, pdlc->gc, GCBackground|GCForeground|GCLineWidth|GCLineStyle, &svGC_values);

}



/************************************************************************************************
*
*	get_fgbg
*
*	DK 12/90
*
*	Get_fgbg gets indexes into the hardware color map to set foreground and background
*	    indexes.  Color selected depends on color mode set at configuration time,
*	    i.e. static, based on alarm condition, etc. 
*
*	Note:   if status and severity change positions in db structs 
*		in db_access.h, this MUST BE MODIFIED!!!
*
*
*	Input	: pdl_index	ptr to the color map from the display list
*		  prti_el	ptr to run-time struct for this controller
*		  static_fg	if static color mode, use this index for fg
*		  static_bg	if static color mode, use this index for bg
*		  clrmod	enum indicates color mode
*		  rule_fg	if rule color mode, use this index for fg
*		  fg		ptr to foreground color index value
*		  bg		ptr to background color index value
*	Output	: fg		
*		  bg		
*	Return: 0 if OK, -1 if error
*/ 



get_fgbg(pdl_index, pComel, static_fg, static_bg, clrmod, rule_fg, rule_bg, fg, bg)
#ifndef AXP /* this is unmaintainable code from the lab! */
int			*pdl_index;	/* ptr to intensity or color hardware map for display_list*/
struct gsdl_com_record	*pComel;
int			static_fg;
int			static_bg;
int			clrmod;
int			rule_fg;
int			rule_bg;
unsigned		*fg;
unsigned		*bg;
#else
unsigned long		*pdl_index;	/* ptr to intensity or color hardware map for display_list*/
struct gsdl_com_record	*pComel;
unsigned long		static_fg;
unsigned long		static_bg;
int			clrmod;
unsigned long		rule_fg;
unsigned long		rule_bg;
unsigned long		*fg;
unsigned long		*bg;
#endif
{
	/*
	 * set up fg color based on color modifier type
	 */
	switch(clrmod){
	    case dl_clrmod_static:	/* if we have a static foreground val */
		*fg = COLOR_LOOKUP(pdl_index, static_fg);
		*bg = COLOR_LOOKUP(pdl_index, static_bg);
		break;

	    case dl_clrmod_rule:	/* if we have a color rule foreground val */
		switch(rule_fg){
		    case BAD_COLOR_DATA:	return -1;
		    case USE_DEFAULT_COLOR:	*fg = COLOR_LOOKUP(pdl_index, static_fg);	break;
		    default:			*fg = COLOR_LOOKUP(pdl_index, rule_fg);	break;
		}
		switch(rule_bg){
		    case BAD_COLOR_DATA:	return -1;
		    case USE_DEFAULT_COLOR:	*bg = COLOR_LOOKUP(pdl_index, static_bg);	break;
		    default:			*bg = COLOR_LOOKUP(pdl_index, rule_bg);	break;
		}
		break;
	}
	return 0;
}


/***********************************************************************
*
*	set_fgbg
*	
*	Deb Kerstiens 9/90
*
*	Purpose : sets up the gc for foreground and background to draw the
*		object.  The foreground color is set according to the clrmod
*		specified when the element was created: static, alarm, discrete,
*		etc.  
*	Input :
*	    pdl_context	: dl_context struct for display id, gc 
*	    pdl_index	: color map for display list els
*	    prti_el	: ptr to run-time element 
*	    static_fg	: foreground color spec'd. at creation of display
*	    static_bg	: background color spec'd. at creation of display
*	    clrmod	: static, alarm, discrete, determines color choice
*
*	Output :
*	    pdlc	: modifies pdl_context->gc 
*	Returns:	    
*	    none
*
**********************************************************************/


set_fgbg(pdl_context, pdl_index, prti_mon, static_fg, static_bg, clrmod, rule_fg, rule_bg)
#ifndef AXP
struct dl_context	*pdl_context;		/* ptr to intensity or color hardware map for display_list*/ 
int			*pdl_index;		/* ptr to intensity or color hardware map for rti*/
struct rti_update_mon	*prti_mon;
int			static_fg;
int			static_bg;
int			clrmod;
int			rule_fg;
int			rule_bg;
#else
struct dl_context	*pdl_context;		/* ptr to intensity or color hardware map for display_list*/ 
unsigned long		*pdl_index;		/* ptr to intensity or color hardware map for rti*/
struct rti_update_mon	*prti_mon;
unsigned long		static_fg;
unsigned long		static_bg;
int			clrmod;
unsigned long		rule_fg;
unsigned long		rule_bg;
#endif
{
	XGCValues gc_vals;

 	if(get_fgbg(pdl_index, prti_mon->pgeom_comel, static_fg, static_bg, clrmod, rule_fg, rule_bg,
	    &gc_vals.foreground, &gc_vals.background) < 0){
		gc_vals.foreground = WhitePixel(pdl_context->d, the_screen);
		gc_vals.background = BlackPixel(pdl_context->d, the_screen);
	}

	if (gc_vals.foreground != invisibleColor) {
		/* set the fg, bg colors in the gc */
		XChangeGC(pdl_context->d, pdl_context->gc, GCForeground|GCBackground, &gc_vals);
		return TRUE;
	} else
	    return FALSE;
}


/***********************************************************************
*
*	set_fg
*	
*	Deb Kerstiens 2/91
*
*	Purpose : sets up the gc for foreground to draw the object.  
*		The foreground color is set according to the clrmod
*		specified when the element was created: static, alarm, 
*		discrete, etc.  
*	Input :
*	    pdl_context	: dl_context struct for display id, gc 
*	    pdl_index	: color map for display list els
*	    prti_el	: ptr to run-time element 
*	    static_fg	: foreground color spec'd. at creation of display
*	    clrmod	: static, alarm, discrete, determines color choice
*
*	Output : 
*	    pdlc	: modifies pdl_context->gc 
*	Returns:	    
*	    none
*
**********************************************************************/


set_fg(pdl_context, pdl_index, prti_el, static_fg, clrmod, rule_fg)
#ifndef AXP
struct dl_context		*pdl_context; 
int				*pdl_index;		/* ptr to intensity or color hardware map for rti*/
struct rti_dynamic_graphic	*prti_el;
int				static_fg;
int				clrmod;
int				rule_fg;
#else
struct dl_context		*pdl_context; 
unsigned long			*pdl_index;		/* ptr to intensity or color hardware map for rti*/
struct rti_dynamic_graphic	*prti_el;
unsigned long			static_fg;
int				clrmod;
unsigned long			rule_fg;
#endif
{
	XGCValues gc_vals;

	if(get_fgbg(pdl_index, prti_el->pComel, static_fg, static_fg, clrmod, rule_fg, USE_DEFAULT_COLOR,
	    &gc_vals.foreground, &gc_vals.background) < 0){
		gc_vals.foreground = WhitePixel(pdl_context->d, the_screen);
 	}

	if (gc_vals.foreground != invisibleColor) {
		/* set only the fg color in the gc */
		XChangeGC(pdl_context->d, pdl_context->gc, GCForeground, &gc_vals);
		return TRUE;
	} else
		return FALSE;
}




/***********************************************************************
*
*	dmDraw_dyn
*	
*	Deb Kerstiens 9/91
*
*	Purpose : draw the dynamic object defined in the display list.
*		gc has already been set up with attributes specified:
*		GCLineWidth,GCLineStyle,GCForeground
*		This code was originally in-line.  
* 
*	Input :
*	    pdlc	: dl_context struct for display id, gc 
*	    objptr	: ptr to object_hdr:x,y,w,h
*	    elemptr	: ptr to the display list element hdr
*
*	Output : 
*	    none
*	Returns:	    
*	    none
*
**********************************************************************/

dmDraw_dyn(pdlc, objptr, pDis)
struct dl_context *pdlc;
struct dl_object_hdr *objptr;
void *pDis;
{
    register struct dl_element_hdr *elemptr = 
		&(((struct dl_anygo *)pDis)->element);

	/* draw the object */
    	switch(elemptr->type){

      	    case dl_user: 
      	    case dl_rec: /* rectangle */
      	    { 

	    	/* 
	    	draw the rectangle 
	    	*/
	    	if(pdlc->fill_style == dl_fill_outline){
			XDrawRectangle(	
				pdlc->d,
				pdlc->w,
				pdlc->gc,
				objptr->x,
				objptr->y,
				objptr->width, 
				objptr->height);
	    	}else
			XFillRectangle(	pdlc->d,
				pdlc->w,
				pdlc->gc,
				objptr->x,
				objptr->y,
				objptr->width, 
				objptr->height);
	    	break;
      	    }

	    case dl_line:{ /* line */
		if(pdlc->fill_style == dl_fill_outline)
			XDrawLines(
				pdlc->d,
				pdlc->w,
				pdlc->gc,
				(XPoint *) &(((struct dl_line *)pDis)->point[0]),
				(int) ((struct dl_line *)pDis)->npoints,
				CoordModeOrigin);
		else
			XFillPolygon(
				pdlc->d,
				pdlc->w,
				pdlc->gc,
				(XPoint *) &(((struct dl_line *)pDis)->point[0]),
				(int) ((struct dl_line *)pDis)->npoints,
				Complex,
				CoordModeOrigin);
		break;
	    }

	    case dl_oval:{ /* oval */
		/* 
		draw the oval
		*/
		if(pdlc->fill_style == dl_fill_outline){
			XDrawArc(
				pdlc->d,
				pdlc->w,
				pdlc->gc,
				objptr->x,
				objptr->y,
				objptr->width, 
				objptr->height,
				0,
				XDEG_PER_REV);
		}
		else
			XFillArc(
				pdlc->d,
				pdlc->w,
				pdlc->gc,
				objptr->x,
				objptr->y,
				objptr->width,
				objptr->height,
				0,
				XDEG_PER_REV);
		break;

	    }
	    case dl_arc:{ /* arc */
		register struct dl_arc *parc = (struct dl_arc *) pDis;

		if(pdlc->fill_style == dl_fill_outline){
			XDrawArc(
				pdlc->d,
				pdlc->w,
				pdlc->gc,
				parc->object.x,
				parc->object.y,
				parc->object.width,
				parc->object.height,
				parc->begin,
				parc->path);
		}
		else
			XFillArc(
				pdlc->d,
				pdlc->w,
				pdlc->gc,
				parc->object.x,
				parc->object.y,
				parc->object.width,
				parc->object.height,
				parc->begin,
				parc->path);
	
		break;
	    }

	    case dl_text:{ /* text */
		char 			*display_str;
		unsigned 		display_str_length;
		int			status;
		XRectangle		box;

		register struct dl_text *ptext = (struct dl_text *) pDis;

		/* convert string index to string pointer */
		status = string_index_to_display_str(pdlc->pstring_table,
						(ptext->textix), 
						&display_str, 
						&display_str_length);
		if(status < 0)
			abort();

		DLOBJ_TO_XREC(ptext->object, box);

		/*
	 	* using size of extent rectangle, determine appropriate font and draw
	 	*/
		if(box.width>box.height)
			draw_fixed_horizontal_string(
				pdlc->d, 
				pdlc->w, 
				pdlc->gc, 
				&box, 
				pdlc->pfont_table, 
				display_str, 
				display_str_length,
				ptext->align);
		else
			draw_fixed_vertical_string(
				pdlc->d, 
				pdlc->w, 
				pdlc->gc, 
				&box, 
				pdlc->pfont_table, 
				display_str, 
				display_str_length,
				ptext->align);

		break;
	    }

	    default:
		/* should return error or abort? */
		printf("~Type not matched\n");
		break;
	} /* end switch */

    return;
}/* end dmDraw_dyn */


/***********************************************************************
*
*	update_color_rules
*
**********************************************************************/

update_color_rules(pdm_llist_el)
struct dm_llist_el	*pdm_llist_el;
{
	struct rti_color_rules		*pRti_cr;
	int				i;

	for(i = 0, pRti_cr = pdm_llist_el->prti_color_rules ; i < pdm_llist_el->ncolor_rule_objs; i++, pRti_cr++)
		if(pRti_cr->prule)
			execute_color_rule(pdm_llist_el, pRti_cr);
}


/***********************************************************************
*
*	execute_color_rule
*
**********************************************************************/


execute_color_rule(pdm_llist_el, pRti_cr)
struct dm_llist_el	*pdm_llist_el;
struct rti_color_rules	*pRti_cr;
{
	struct	{
		int	true;
		int	loper;
	} 			results[MAX_RULE_LINES];
	int			line;
	struct dl_color_rule	*prule	= pRti_cr->prule;
#ifndef AXP
	int			last_fg	= *pRti_cr->prulecolorfg;
	int			last_bg = *pRti_cr->prulecolorbg;
#else
	unsigned long		last_fg	= *pRti_cr->prulecolorfg;
	unsigned long		last_bg = *pRti_cr->prulecolorbg;
#endif
      	char	have [MAX_CHARS], want [MAX_CHARS];
	double	diff;
	int	result;
	void	stripBlanks ();
        double	atof ();
	/*
	 * pass one: collect data, watching for bad data which invalidates the rule
	 */
	for(line = 0; line < MAX_RULE_LINES; line++){
		struct gsdl_com_record	*pComel	= pRti_cr->line[line].pComel;
		if(!pdm_llist_el->pdl_strings[(int)(prule->info[line].chan)][0]){
			/* line is empty, treat it like a TRUE AND */
			results[line].true  = TRUE;
			results[line].loper = dl_clrrule_and;

		} else if(dmChkPvStat(pComel, 0, 0) != dmConnection_okay){

			/* any bad piece of data invalidates the rule execution */
			*pRti_cr->prulecolorfg		= BAD_COLOR_DATA;
			*pRti_cr->prulecolorbg		= BAD_COLOR_DATA;
			if ((*pRti_cr->prulecolorfg != last_fg) || (*pRti_cr->prulecolorbg != last_bg))
				{
				*pRti_cr->pclrrule_has_changed = 1;
#ifdef DEBUG
printf("~setting (%s) at:0x%x BAD ", ca_name(pComel->pChan), pRti_cr->pclrrule_has_changed);
#endif
				}
#ifdef DEBUG
printf("~changed:%d fg:%d bg:%d\n",*pRti_cr->pclrrule_has_changed,*pRti_cr->prulecolorfg,*pRti_cr->prulecolorbg);
#endif
			return;
		} else {
			/* compare the field value (have) to what we want for the TRUE condition */

#ifdef DEBUG
printf("~checking (%s) at 0x%x line:%d ",ca_name(pComel->pChan),pRti_cr->pclrrule_has_changed,line);
#endif

			/* Get the wanted value and strip the leading and trailing blanks */

			strcpy (want, pdm_llist_el->pdl_strings[(int)(prule->info[line].value)]);
			stripBlanks (want);

			/* Get the current value and strip the leading and trailing blanks */

			cvrt_to_str(pComel, ca_field_type(pComel->pChan), have, -1);
			stripBlanks (have);
#ifdef DEBUG
printf("~have:(%s) want:(%s)\n",have,want);
#endif
			/* If the values are numeric then convert both to double to perform the comparison. */
			/* Otherwise compare the values as strings. */

			switch (ca_field_type(pComel->pChan)) {
			    case DBR_FLOAT:
			    case DBR_DOUBLE:
			    case DBR_LONG:
			    case DBR_SHORT:
			        if (isNumeric (want)) {
				    diff = atof (have) - atof (want);
				    break;
				}
			    default:
				diff = strcmp (have, want);
				break;
			}

			switch (prule->info [line].operation) {
			    case dl_clrrule_eq:
				result = diff == 0.0;
				break;
			    case dl_clrrule_ne:
				result = diff != 0.0;
				break;
			    case dl_clrrule_le:
				result = diff <= 0.0;
				break;
			    case dl_clrrule_lt:
				result = diff < 0.0;
				break;
			    case dl_clrrule_ge:
				result = diff >= 0.0;
				break;
			    case dl_clrrule_gt:
				result = diff > 0.0;
				break;
			}

			results[line].true  = result;
			results[line].loper = prule->info[line].connector;
		}
	}

	/*
	 * pass two: convert andnot operators to and
	 *	NOTE: edd insures that last line is dl_clrrule_use
	 */
	for(line = 0; line < MAX_RULE_LINES; line++){
		if(results[line].loper == dl_clrrule_andnot){
			results[line+1].true = !results[line+1].true;
			results[line].loper = dl_clrrule_and;
		}
	}

	/*
	 * pass three: select the rules color
	 */
	line = 0;
	while(line < MAX_RULE_LINES){
		if (results[line].loper == dl_clrrule_use && results[line].true){
			*pRti_cr->prulecolorfg		= prule->info[line].clr;
			*pRti_cr->prulecolorbg		= prule->bg_enable ? prule->default_bg : USE_DEFAULT_COLOR;
			if ((*pRti_cr->prulecolorfg != last_fg) || (*pRti_cr->prulecolorbg != last_bg))
				{
				*pRti_cr->pclrrule_has_changed = 1;
#ifdef DEBUG
printf("~setting (%s) at:0x%x GOOD ", ca_name(pRti_cr->line[line].pComel->pChan), pRti_cr->pclrrule_has_changed);
#endif
				}
#ifdef DEBUG
printf("~line %d is TRUE changed at:0x%x:%d fg:%d bg:%d\n", line, pRti_cr->pclrrule_has_changed,*pRti_cr->pclrrule_has_changed,*pRti_cr->prulecolorfg,*pRti_cr->prulecolorbg);
#endif

			return;
		} else if(results[line].loper == dl_clrrule_use && !results[line].true){
			line++;

		} else if(results[line].loper == dl_clrrule_and && results[line].true){
			line++;

		} else if(results[line].loper == dl_clrrule_and && !results[line].true){
			while(results[line].loper == dl_clrrule_and && line < MAX_RULE_LINES)
				line++;
			line++;
		}
	}
	*pRti_cr->prulecolorfg		= prule->fg_enable ? prule->default_fg : USE_DEFAULT_COLOR;
	*pRti_cr->prulecolorbg		= prule->bg_enable ? prule->default_bg : USE_DEFAULT_COLOR;

	if ((*pRti_cr->prulecolorfg != last_fg) || (*pRti_cr->prulecolorbg != last_bg))
				{
				*pRti_cr->pclrrule_has_changed = 1;

#ifdef DEBUG
printf("~setting (%s) at:0x%x DEFAULT ", pRti_cr->line[line-1].pComel->pChan ? ca_name(pRti_cr->line[line-1].pComel->pChan) : "", pRti_cr->pclrrule_has_changed);
#endif
				}
#ifdef DEBUG
printf("~taking defaults changed:%d fg:%d bg:%d\n",*pRti_cr->pclrrule_has_changed,*pRti_cr->prulecolorfg,*pRti_cr->prulecolorbg);
#endif
}

/*----------*/
void stripBlanks (s)

    char	*s;

    {

	char	*p = s, temp [100];
	int	i;

	/* Skip over the leading blanks */

	while (*p)
	    if (*p != ' ')
		break;
	    else
		p++;

	strcpy (temp, p);

	/* Strip any trailing blanks */

	i = strlen (temp);
	if (i != 0) {
	    p = &temp [i - 1];
	    while (*p)
		if (*p != ' ')
		    break;
		else
		    p--;
	    *(p + 1) = '\0';
	}

	strcpy (s, temp);
	
    }

/*----------*/
static int isNumeric (char *str)  /* Return 1 if string is a valid number */

    { 

      int  points = 0;

      while (*str) {
	if (*str == '.')
	  points++;
	else if (!isdigit (*str))
	  return 0;
	str++;
      }

      return points <= 1;

    }
