/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*	@(#)pfoPlot.c	1.4 12/2/91
 *	Author:	Roger A. Cole
 *	Date:	03-04-91
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	03-04-91	rac	initial version
 * .02	07-31-91	rac	installed in SCCS
 * .03  08-22-91	rac	used new sydxxx calls
 * .04	09-30-91	rac	use new calling sequence for sydChanOpen
 * .05	12-02-91	rac	use new calling sequence for sydInputStoreInSet
 *
 * make options
 *	-DXWINDOWS	makes a version for X11
 *	-DPFO_TEST	makes a program for testing the routines
 *
 */
/*+/mod***********************************************************************
* TITLE	pfoPlot - plotting for OPI routines
*
* DESCRIPTION
*	pfo... routines provide high-level plotting capability for the EPICS
*	operator interface (OPI) runtime.  The data to be plotted are
*	in the form supplied by the "gsd_..." routines, which get
*	synchronous data.
*
* QUICK REFERENCE
* PFO_WIN pPfo;		pointer to plot window structure
*
*     void  pfoAxisAutoRange(pAxis                                         )
*     void  pfoAxisClose(pAxis                                             )
* PFO_AXIS *pfoAxisOpen(pPfo, isX, fill, fg, pGsd, pDbrGr, dbrGrType       )
*     void  pfoWinAutoRange(pPfo                                           )
*     void  pfoWinClose(pPfo                                               )
*  PFO_WIN *pfoWinOpen(pDisp, window, fg, bg, nSamp, title, xTitle, yTitle )
*     void  pfoWinPlotSample(pPfo                                          )
*     void  pfoWinRepaint(pPfo, dl_plot_style, axisType                    )
*     void  pfoWinRestart(pPfo, dl_plot_style, axisType                    )
*     void  pfoWinSetAttr(pPfo, attr, value, pArg                          )
*               attr = PFO_ATTR_{FG1,FG2}
*     void  pfoWinWritePSFile(pPfo, dl_plot_style, axisType, filePath      )
*
* BUGS
* o	features not presently supported:
*	zoom, pan, clipping
*	cross-hair axes
*	color PostScript output
*	scaling and orientation of PostScript output
*	suppressing axis labels and/or annotations
*	caller specified axis scaling, number of axis divisions, etc.
*
*-***************************************************************************/
#include <memLeak.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#ifdef AXP
#include <string.h>
#endif
#include <cadef.h>
#include <db_access.h>
#include "gsd_sync_defs.h"
#include <dldef.h>
#include <pprPlotDefs.h>
#include <sydDefs.h>
#include <sydPlotDefs.h>
#define PFO_PRIVATE
#include <pfoPlotDefs.h>

/*/subhead pfo test routines---------------------------------------------------
*
*----------------------------------------------------------------------------*/
#if defined PFO_TEST

struct gsd_sync_data gsd1, *pGsd1=&gsd1, gsd2, *pGsd2=&gsd2;

PFO_WIN *pPfo;
PFO_AXIS *pAxis1, *pAxis2;

void monitor();

static void
pfoSleep(seconds, usec)
int	seconds;	/* I number of seconds (added to usec) to sleep */
int	usec;		/* I number of micro-sec (added to sec) to sleep */
{
#ifndef vxWorks
    usleep((unsigned)(seconds*1000000 + usec));
#else
    int		ticks;
    static int ticksPerSec=0;
    static int usecPerTick;

    if (ticksPerSec == 0) {
	ticksPerSec = sysClkRateGet();
	usecPerTick = 1000000 / ticksPerSec;
    }

    ticks = seconds*ticksPerSec + usec/usecPerTick + 1;
    taskDelay(ticks);
#endif
}

main()
{
    Display	*pDisp;		/* pointer to X server connection */
    int		screenNo;	/* screen number */
    Window	rootWindow;	/* the root window on display */
    unsigned long fg, bg;
    XEvent	anEvent;	/* a window event structure */
    Window	plotWindow;	/* the plot window on the display */
    XSizeHints	sizeHints;	/* defaults for position and size */

    chid	pCh1,pCh2;
    struct dbr_gr_double gr1, *pGr1=&gr1, gr2, *pGr2=&gr2;
    PFO_CHAN	*pChan1, *pChan2;
    int		monitoring=0;
    int		x,y,width,height;
    long	stat;
    chtype	grType1, grType2;
    unsigned long fg1, fg2, fgWin, smithPix1, smithPix2;
    Visual	*pVisual;
    Colormap	cmap;
    XColor	color, rgbDb;
    int	isMono;
    int		nSamp, plot, axis;
    enum dl_plot_style plotType;

#if 0
    SEVCHK(ca_search("~ts2:wave0", &pCh1), "search1");
    SEVCHK(ca_search("~waveCompAve", &pCh2), "search2");
#else
    SEVCHK(ca_search("~rai_2000", &pCh1), "search1");
    SEVCHK(ca_search("~rao_2000", &pCh2), "search2");
#endif
    SEVCHK(ca_pend_io(4.), "~pend after search");
    grType1 = dbf_type_to_DBR_GR(ca_field_type(pCh1));
    grType2 = dbf_type_to_DBR_GR(ca_field_type(pCh2));
    SEVCHK(ca_array_get(grType1, 1, pCh1, pGr1), "~gr get 1");
    SEVCHK(ca_array_get(grType2, 1, pCh2, pGr2), "~gr get 2");
    SEVCHK(ca_pend_io(4.), "~pend after array_get for GR");
    pGsd1->pName = ca_name(pCh1);
    pGsd1->time_type = dbf_type_to_DBR_TIME(ca_field_type(pCh1));
    pGsd1->count = ca_element_count(pCh1);
    pGsd1->pChid = pCh1;
    pGsd1->pSdata =
		(void *)GenMalloc(dbr_size_n(pGsd1->time_type, pGsd1->count));
    pGsd2->pName = ca_name(pCh2);
    pGsd2->time_type = dbf_type_to_DBR_TIME(ca_field_type(pCh2));
    pGsd2->count = ca_element_count(pCh2);
    pGsd2->pChid = pCh2;
    pGsd2->pSdata =
		(void *)GenMalloc(dbr_size_n(pGsd2->time_type, pGsd2->count));

    if ((pDisp = XOpenDisplay((char *)NULL)) == NULL)
	abort();
    screenNo = DefaultScreen(pDisp);
    rootWindow = DefaultRootWindow(pDisp);
    sizeHints.x = x = 100;
    sizeHints.y = y = 100;
    sizeHints.width = width = 200;
    sizeHints.height = height = 200;
    sizeHints.flags = PSize | PPosition;
    plotWindow = XCreateSimpleWindow(pDisp, rootWindow, x, y, width, height, 5,
		BlackPixel(pDisp, screenNo), WhitePixel(pDisp, screenNo));

    XSetStandardProperties(pDisp, plotWindow, "win title", "icon title",
		None, 0, NULL, &sizeHints);
    XSelectInput(pDisp, plotWindow,
		ExposureMask |
		ButtonPressMask | ButtonReleaseMask);
    XMapWindow(pDisp, plotWindow);

    pVisual = DefaultVisual(pDisp, screenNo);
    if (pVisual->class != GrayScale && pVisual->class != StaticGray) {
	isMono = 0;
	cmap = DefaultColormap(pDisp, screenNo);
	XAllocNamedColor(pDisp, cmap, "maroon", &color, &rgbDb);
	fgWin = color.pixel;
	XAllocNamedColor(pDisp, cmap, "gold", &color, &rgbDb);
	fg1 = color.pixel;
	XAllocNamedColor(pDisp, cmap, "cyan", &color, &rgbDb);
	fg2 = color.pixel;
	XAllocNamedColor(pDisp, cmap, "red", &color, &rgbDb);
	smithPix1 = color.pixel;
	XAllocNamedColor(pDisp, cmap, "green", &color, &rgbDb);
	smithPix2 = color.pixel;
    }
    else {
	isMono = 1;
	fgWin = BlackPixel(pDisp, screenNo);
	fg1 = BlackPixel(pDisp, screenNo);
	fg2 = BlackPixel(pDisp, screenNo);
	smithPix1 = BlackPixel(pDisp, screenNo);
	smithPix2 = BlackPixel(pDisp, screenNo);
    }
    printf("~plot type 1:line 2:mark\n");
    printf("~axis type 1:XY 2:XYY 3:Y 4:YY 5:IMP 6:ADM 7:IMM 8:TY 9:TYY\n");
    printf("~nSamp, numeric plot_type and axis_type: ");
    fflush(stdout);
    if (fscanf(stdin, "%d %d %d", &nSamp, &plot, &axis) != 3)
	return;
    if (plot == 1)
	plotType = dl_line_plot;
    else
	plotType = dl_point_plot;
    pPfo = pfoWinOpen(pDisp, plotWindow, fgWin, WhitePixel(pDisp, screenNo),
	    nSamp, "top title", "bottom title", "left title", 1);
    pAxis1 = pfoAxisOpen(pPfo, 0, 1, fg1, pGsd1, pGr1, grType1);
    assert(pAxis1 != NULL);
    pAxis2 = pfoAxisOpen(pPfo, 1, 1, fg2, pGsd2, pGr2, grType2);
    assert(pAxis2 != NULL);
    pfoWinSetAttr(pPfo, PFO_ATTR_FG1, 0, &smithPix1);
    pfoWinSetAttr(pPfo, PFO_ATTR_FG2, 0, &smithPix2);
    while (1) {
	XFlush(pDisp);
	ca_pend_event(.01);
	if (XPending(pDisp) != 0) {
	    XNextEvent(pDisp, &anEvent);
	    if (anEvent.type == ButtonRelease &&
				anEvent.xbutton.button == Button3) {
		if (pPfo != NULL) {
		    pfoWinWritePSFile(pPfo, "pfo.ps");
		    (void)printf("~PostScript written to pfo.ps\n");
		}
		if (pPfo != NULL)
		    pfoWinClose(pPfo);
		GenFree((char *)pGsd1->pSdata);
		GenFree((char *)pGsd2->pSdata);
		XCloseDisplay(pDisp);
		break;
	    }
	    else if (anEvent.type == ButtonRelease &&
				anEvent.xbutton.button == Button2) {
		if (pPfo != NULL)
		    pfoWinRestart(pPfo, plotType, axis);
	    }
	    else if (anEvent.type == ButtonRelease &&
				anEvent.xbutton.button == Button1) {
		if (pPfo != NULL) {
		    pfoWinAutoRange(pPfo);
		    pfoWinRepaint(pPfo, plotType, axis);
		}
	    }
	    else if (anEvent.type == Expose && anEvent.xexpose.count == 0) {
		if (!monitoring) {
		    monitoring = 1;
		    SEVCHK(ca_add_array_event(pGsd1->time_type,
				pGsd1->count, pCh1, monitor, (void *)pAxis1,
				0., 0., 0., NULL), "~add event 1");
		    SEVCHK(ca_add_array_event(pGsd2->time_type,
				pGsd2->count, pCh2, monitor, (void *)pAxis2,
				0., 0., 0., NULL), "~add event 2");
		}
		pfoWinRepaint(pPfo, plotType, axis);
	    }
	}
	pfoSleep(0, 50000);
    }
}
void
monitor(arg)
struct event_handler_args arg;
{
    PFO_AXIS	*pAxis = (PFO_AXIS *)arg.usr;
    struct gsd_sync_data *pGsd = pAxis->pfoChan.pGsd;
    SYD_SPEC	*pSspec = pAxis->pPfo->pSspec;
    int		sampleNum;
    int		nBytes;
    long	stat;
    int		i;
    short	*pShrt;

/*-----------------------------------------------------------------------------
*	for now, store the data into the place implied by the pGsd.  This
*	simulates the form in which data will eventually be fed to the
*	sydXxx routines.
*
*	When a value for the second channel is received, force the time
*	stamp of the first one to be the same, then process the two as a
*	synchronous sample.
*----------------------------------------------------------------------------*/
    pShrt = &((struct dbr_time_short *)arg.dbr)->value;
    nBytes = dbr_size_n(pGsd->time_type, pGsd->count);
    bcopy((char *)arg.dbr, (char *)pGsd->pSdata, nBytes);
    pGsd->svalid = 1;
    if (pAxis == pAxis1)
	return;
    tsLocalTime(&((struct dbr_time_double *)pAxis1->pfoChan.pGsd->pSdata)->stamp);
    tsLocalTime(&((struct dbr_time_double *)pAxis2->pfoChan.pGsd->pSdata)->stamp);
    pfoWinPlotSample(pPfo);
}
#endif  	/* defined PFO_TEST */

/*+/subr**********************************************************************
* NAME	pfoAxisAutoRange - re-range channel's axis to fit min and max values
*
* DESCRIPTION
*	Change the scaling for the axis for a channel based on the
*	minimum and maximum values encountered for the axis.
*
*	Rescaling sets the value at the axis origin to the minimum value
*	for the channel, and the value at the end of the axis to the
*	maximum value for the channel.
*
*	In order to get the plotting grid redrawn with the new scaling,
*	it is necessary to call pfoWinRepaint or pfoWinRestart.
*
*	Calling this routine has no effect on the data (if any) stored
*	in the buffers.  Thus, it is possible to replot stored data
*	on the rescaled grid.
*
* RETURNS
*	void
*
* SEE ALSO
*	pfoWinAutoRange
*
*-*/
void
pfoAxisAutoRange(pAxis)
PFO_AXIS *pAxis;	/* I channel to rerange */
{
    sydPlotAxisAutoRange(pAxis->pSlave);
}

/*+/subr**********************************************************************
* NAME	pfoAxisClose - close a channel plot structure
*
* DESCRIPTION
*	Closes a plot channel.
*
* RETURNS
*	void
*
* BUGS
* o	isn't presently available directly to users of pfoPlot.c--sydPlot
*	doesn't presently support closing an individual channel
*
*-*/
static void
pfoAxisClose(pAxis)
PFO_AXIS *pAxis;	/* I channel to rerange */
{
    long	stat;

    stat = sydChanClose(pAxis->pPfo->pSspec, pAxis->pfoChan.pSchan);
    assert(stat == S_syd_OK);
    GenFree((char *)pAxis);
}

/*+/subr**********************************************************************
* NAME	pfoAxisOpen - open a channel plot structure
*
* DESCRIPTION
*	Adds a channel to the list of channels in a pfo window structure.
*
*	A circular storage buffer is established to hold the number of
*	values specified in the pfoWinOpen call.  Calls to pfoWinPlotSample
*	result in plotting sample values and also storing them for use
*	in subsequent repaint operations--e.g., on expose or resize.
*
*	This routine does no drawing.  Thus, it can be called regardless
*	of whether the window for plotting is visible.  Typical usage is
*	to set up all channels and then call pfoWinRepaint (assuming that
*	the window is visible prior to calling the latter).
*
*	Most of the activities for pfoPlot are governed by the arguments
*	supplied in this call.  Both the synchronous data structure and the
*	graphics buffer must contain valid information when this routine
*	is called.  (The synchronous data structure buffer pointer is
*	ignored, however, on this call--even if it contains valid data
*	for the channel.)
*
*	It is assumed that the synchronous data pointer will always point
*	to a valid data structure, and that the structure will indicate
*	whether valid data is in the buffer contained in the structure.
*
*	The pDbrGr pointer is used to obtain graphics information for the
*	channel, including high and low operating range, engineering units
*	text, and state labels for enumerated channels.  The buffer pointed
*	to by pDbrGr can be one of the DBR_GR_... types or one of the
*	DBR_CTRL_... types; the specific type must be supplied.
*
*	Most plotting done on behalf of the channel is done using the fg
*	pixel value.  This will include axis annotations and actual lines
*	or marks corresponding to plotted data.  If the pfoWinOpen call
*	indicated that the display is monochrome, plotting will still be
*	done using the fg pixel value, but, in addition, line plots will
*	be done with dashed lines, with each channel having its own
*	dashed line pattern.
*
*	Channels for which the `isX' argument is true will be plotted
*	along the horizontal axis.  Other channels are plotted along the
*	vertical axis.  The mix of channels determines the general type
*	of plot which will be done:
*
*	o   all Y channels--results in a `position vs value' plot, with
*	    channel value along the vertical axis.
*
*	    The horizontal axis is calibrated by `bin number', with its
*	    extent matching the count for the channel with the largest
*	    array dimension.  If several channels are being plotted,
*	    channels with a smaller array dimension will occupy only part
*	    of the horizontal axis.
*
*	    Scalar channels are treated special.  If all the channels
*	    are scalar, then the horizontal axis is set up as a 60 second
*	    time line.  Then the time stamps from the values are used
*	    for horizontal placement; when the end of the horizontal axis
*	    is reached, plotting wraps around to the beginning.
*
*	    If some of the channels are array and some are scalar, then
*	    the horizontal axis is calibrated by bin number and the
*	    array channels are plotted as described above.  Each new
*	    sample for a scalar channel is treated as belonging to the
*	    next available bin; when the last bin has been used, plotting
*	    wraps around.
*
*	o   one X channel and one or more Y channels--results in a `value
*	    vs value' plot.  Each Y channel is plotted against the X
*	    channel; plotting uses the fg pixel value specified for the
*	    Y channel.  Either scalar channels or array channels can be
*	    used.
*
*	    The horizontal axis is calibrated according to the value range
*	    of the X channel.
*
*	    The x vs y plotting is affected by the counts for the X
*	    channel and the Y channel.  If both channels are scalars,
*	    then a single point results.  If both are arrays with equal
*	    counts, then that number of points results.  If the counts
*	    are unequal, then the number of points to plot is the minimum
*	    of the two counts.
*
*	o   pairs of X and Y channels--not implemented yet
*
* RETURNS
*	PFO_AXIS *, or
*	NULL if an error occurs
*
* BUGS
* o	there is no provision for the caller to specify number of
*	intervals, values at axis endpoints, annotations, etc.
* o	assumes that the graphics buffer is either DBR_GR_... or DBR_CTRL_...
* o	assumes that the data buffer is DBR_TIME_...
* o	assumes that the layout of both DBR_GR_... and DBR_CTRL_... buffers
*	with regard to HOPR, LOPR, EGU, PREC, and ENUM information is same
*
* SEE ALSO
*	pfoWinOpen
*
*-*/
PFO_AXIS *
pfoAxisOpen(pPfo, isX, fill, fg, pGsd, pDbrGr, dbrGrType)
PFO_WIN *pPfo;		/* I pointer to plot window structure */
int	isX;		/* I 1 if channel is for x axis, else 0 */
int	fill;		/* I 1 to fill in missing values from prior, else 0 */
unsigned long fg;	/* I foreground pixel value for channel */
struct gsd_sync_data *pGsd;/* I pointer to sync data buf for channel */
union db_access_val *pDbrGr;/* I pointer to buffer containing graphics info */
chtype	dbrGrType;	/* I DBR type of pDbrGr--DBR_GR_... or DBR_CTRL... */
{
    PFO_AXIS	*pAxis;
    double	charHt, charHtX;
    double	minDbl, maxDbl;
    long	stat;
    enum sydChanSync sync;

    pAxis = (PFO_AXIS *)GenMalloc(sizeof(PFO_AXIS));
    if (pAxis == NULL) {
	(void)printf("~pfoAxisOpen: couldn't get memory\n");
	return NULL;
    }

    if (fill)		sync = SYD_SY_FILLED;
    else		sync = SYD_SY_NONF;
    pAxis->pfoChan.pGsd = pGsd;
    pAxis->pfoChan.pDbrGr = pDbrGr;
    pAxis->pfoChan.dbrGrType = dbrGrType;

    stat = sydChanOpen(pPfo->pSspec, NULL, pGsd->pName,sync,&pAxis->pfoChan,0);
    if (stat != S_syd_OK) {
	(void)printf("~pfoAxisOpen: error from sydChanOpen\n");
	GenFree((char *)pAxis);
	return NULL;
    }
    pAxis->pfoChan.pSchan = pPfo->pSspec->pChanTail;

    pAxis->pSlave = sydPlotChanAdd(pPfo->pMstr, pAxis->pfoChan.pSchan);
    assertAlways(pAxis->pSlave != NULL);
    stat = sydPlotAxisSetAttr(pAxis->pSlave, SYD_PLATTR_FG, 0, &fg);
    assertAlways(stat == S_syd_OK);
    stat = sydPlotAxisSetAttr(pAxis->pSlave, SYD_PLATTR_BG, 0, &pPfo->bg);
    assertAlways(stat == S_syd_OK);

    pAxis->isAnX = isX;
    if (isX) {
	DoubleListAppend(pAxis, pPfo->pXAxisHead, pPfo->pXAxisTail);
	stat = sydPlotAxisSetAttr(pAxis->pSlave, SYD_PLATTR_XCHAN, 1, NULL);
	assertAlways(stat == S_syd_OK);
    }
    else
	DoubleListAppend(pAxis, pPfo->pYAxisHead, pPfo->pYAxisTail);

    pAxis->pPfo = pPfo;

    stat = sydSampleSetFree(pPfo->pSspec);
    assert(stat == S_syd_OK);
    return pAxis;
}

/*+/subr**********************************************************************
* NAME	pfoWinAutoRange - re-range all channels for a plot window
*
* DESCRIPTION
*	Change the scaling for each axes for a plot window based on the
*	minimum and maximum values encountered for the axis.
*
*	Rescaling sets the value at the axis origin to the minimum value
*	for the channel, and the value at the end of the axis to the
*	maximum value for the channel.
*
*	In order to get the plotting grid redrawn with the new scaling,
*	it is necessary to call pfoWinRepaint or pfoWinRestart.
*
*	Calling this routine has no effect on the data (if any) stored
*	in the buffers.  Thus, it is possible to replot stored data
*	on the rescaled grid.
*
* RETURNS
*	void
*
* SEE ALSO
*	pfoAxisAutoRange
*
*-*/
void
pfoWinAutoRange(pPfo)
PFO_WIN *pPfo;		/* I pointer to plot window structure */
{
    PFO_AXIS	*pAxis;
    pAxis = pPfo->pXAxisHead;
    while (pAxis != NULL) {
	pfoAxisAutoRange(pAxis);
	pAxis = pAxis->pNext;
    }
    pAxis = pPfo->pYAxisHead;
    while (pAxis != NULL) {
	pfoAxisAutoRange(pAxis);
	pAxis = pAxis->pNext;
    }
}

/*+/subr**********************************************************************
* NAME	pfoWinClose - close a plot window
*
* DESCRIPTION
*	Closes a plot window structure, wrapping up plot operations, closing
*	plotting channels which haven't already been closed, etc.
*
*	The private gc's used by pfoPlot are free'd, but the window supplied
*	in the pfoWinOpen call is not free'd.
*
* RETURNS
*	void
*
* SEE ALSO
*	pfoWinOpen
*
*-*/
void
pfoWinClose(pPfo)
PFO_WIN *pPfo;		/* I pointer to plot window structure */
{
    PFO_AXIS	*pAxis;

    while (pPfo->pXAxisHead != NULL) {
	pAxis = pPfo->pXAxisHead;
	DoubleListRemove(pAxis, pPfo->pXAxisHead, pPfo->pXAxisTail);
	pfoAxisClose(pAxis);
    }
    while (pPfo->pYAxisHead != NULL) {
	pAxis = pPfo->pYAxisHead;
	DoubleListRemove(pAxis, pPfo->pYAxisHead, pPfo->pYAxisTail);
	pfoAxisClose(pAxis);
    }
    if (pPfo->pMstr != NULL) {
	sydPlotDone(pPfo->pMstr, 1);
	pPfo->pMstr = NULL;
    }
    if (pPfo->pSspec != NULL)
	sydClose(pPfo->pSspec);
    XFree((caddr_t)(pPfo->gc));
    GenFree((char *)pPfo);
    
}

/*+/subr**********************************************************************
* NAME	pfoWinOpen - open a plot window
*
* DESCRIPTION
*	Creates a plot window structure and returns a handle to be used
*	in subsequent plotting operations in the window.
*
*	The foreground pixel value is used in plotting the grid and the
*	titles along the top, bottom, and left edges of the plot window.
*
*	This routine does no actual plotting.  The grid, titles, and
*	annotations are plotted when pfoWinRepaint is called.  Data plotting
*	occurs for calls to pfoWinRepaint and pfoWinPlotSample.
*
* RETURNS
*	PFO_WIN *, or
*	NULL if an error occurs
*
* BUGS
* o	there is no mechanism to suppress automatic x and y labels and
*	annotations, or to supply caller's own x and y labels and annotations
*
* SEE ALSO
*	pfoWinSetAttr, pfoWinRepaint, pfoWinRestart, pfoWinWritePSFile
*
*-*/
PFO_WIN *
pfoWinOpen(pDisp, window, fg, bg, nSamples, title, xTitle, yTitle)
Display	*pDisp;		/* I pointer to the X display */
Window	window;		/* I the X window in which to plot */
unsigned long fg;	/* I foreground pixel value */
unsigned long bg;	/* I background pixel value */
int	nSamples;	/* I desired size of circular buffer for samples */
char	*title;		/* I title along top of plot, or NULL */
char	*xTitle;	/* I title along bottom of plot, or NULL */
char	*yTitle;	/* I title along left side of plot, or NULL */
{
    PFO_WIN	*pPfo;		/* pointer to pfo structure */
    long	stat;
    long	sydOpenPFO();

    if (strlen(title) >= sizeof(pPfo->title)) {
	(void)printf("~pfoWinOpen: title too long\n");
	return NULL;
    }
    if (strlen(yTitle) >= sizeof(pPfo->yTitle)) {
	(void)printf("~pfoWinOpen: yTitle too long\n");
	return NULL;
    }
    if (strlen(xTitle) >= sizeof(pPfo->xTitle)) {
	(void)printf("~pfoWinOpen: xTitle too long\n");
	return NULL;
    }
    if ((pPfo = (void *)GenMalloc(sizeof(PFO_WIN))) == NULL) {
	(void)printf("~pfoWinOpen: couldn't get memory\n");
	return NULL;
    }
    pPfo->pXAxisHead = NULL;
    pPfo->pXAxisTail = NULL;
    pPfo->pYAxisHead = NULL;
    pPfo->pYAxisTail = NULL;
    pPfo->binNum = 0;
    pPfo->pSspec = NULL;
    pPfo->pMstr = NULL;
    pPfo->nSamples = nSamples;

    pPfo->gc = XCreateGC(pDisp, window, 0, NULL);
    XSetForeground(pDisp, pPfo->gc, fg);
    XSetBackground(pDisp, pPfo->gc, bg);
    pPfo->fg = fg;
    pPfo->bg = bg;
    stat = sydOpenPFO(&pPfo->pSspec, NULL);
    if (stat != S_syd_OK) {
	(void)printf("~pfoWinOpen: error from sydOpen\n");
	goto pfoWinOpen_err;
    }
    pPfo->pMstr = &pPfo->master;
    stat = sydPlotInitUW(pPfo->pMstr,
		pPfo->pSspec, pDisp, window, pPfo->gc, fg, bg);
    if (stat != S_syd_OK) {
	(void)printf("~pfoWinOpen: error from sydPlotInitUW\n");
	pPfo->pMstr = NULL;
	goto pfoWinOpen_err;
    }
    stat = sydPlotSetTitles(pPfo->pMstr, title, yTitle, xTitle, NULL);
    assertAlways(stat == S_syd_OK);
    strcpy(pPfo->title, title);
    strcpy(pPfo->yTitle, yTitle);
    strcpy(pPfo->xTitle, xTitle);
    pPfo->pDisp = pDisp;
    pPfo->window = window;

    return pPfo;

pfoWinOpen_err:
    pfoWinClose(pPfo);
    return NULL;
}

/*+/subr**********************************************************************
* NAME	pfoWinPlotSample - plot a synchronous sample
*
* DESCRIPTION
*	Plots a synchronous sample.  The nature of the plotting depends
*	both on the setup implied by the calls to pfoAxisOpen and by the
*	plot type arguments in the most recent call to pfoWinRepaint.
*
*	If all channels in the synchronous sample have their .svalid flag
*	equal to 0, then the sample is ignored.
*
*	This routine plots only a single sample.  If multiple samples are
*	available, then this routine must be called multiple times.
*
* RETURNS
*	void
*
* BUGS
* o	doesn't erase `stale' marks
*
* SEE ALSO
*	pfoWinRepaint, pfoWinRestart, pfoWinWritePSFile
*
* NOTES
* 1.	If the number of samples (specified by `nSamp' in the pfoWinOpen call)
*	is 0, then the data area is erased before plotting, so that only the
*	current sample appears on the screen.
*
*-*/
void
pfoWinPlotSample(pPfo)
PFO_WIN *pPfo;		/* I pointer to plot window structure */
{
    SYD_SPEC	*pSspec=pPfo->pSspec;
    long	stat;

    stat = sydInputGet(pSspec, NULL);
    if (stat == S_syd_OK || stat == S_syd_partial) {
	sydInputStoreInSet(pSspec, 0);
	if (pPfo->nSamples == 0)
	    sydPlotEraseSamples(pPfo->pMstr);
	sydPlotSamples(pPfo->pMstr, pSspec->lastData, pSspec->lastData, 1);
    }
}

/*+/subr**********************************************************************
* NAME	pfoWinRepaint - repaint a plot window
*
* DESCRIPTION
*	Repaints the plot window.  This routine (or pfoWinRestart) should
*	be called for the initial expose for the window, as well as when
*	the window is subsequently exposed or resized.  Several operations
*	are performed:
*
*	o   the window is erased
*	o   the appropriate plotting background is drawn, with titles,
*	    axis labels, and axis annotations, as needed
*	o   if the circular buffers contain any data, it is plotted
*
*	Calling this routine has no effect on the data (if any) stored
*	in the buffers.
*
*       On color displays, if no alternate foreground pixel values have been
*       set, then the overlays are drawn using the color of the plot window.
*
* RETURNS
*	void
*
* BUGS
* o	multiple X vs multiple Y isn't implemented
*
* SEE ALSO
*	pfoWinRestart, pfoWinSetAttr, pfoWinWritePSFile
*
* NOTES
* 1.	Plot style terminology differs slightly between this routine and
*	OPI display lists: dl_point_plot selects a plot which draws a
*	plot mark at each data point, rather than drawing a pixel-sized
*	point at each data point.
* 2.	No formalism has been established for axis type; instead, numeric
*	codes are used on an interim basis:
*	1   x vs y plotting; each y channel has its own grid
*	2   x vs y plotting; all y channels share a single grid
*	3   bin vs y plotting; each y channel has its own grid.  The main
*	    intent is to support waveform plotting.  This type can be used
*	    for scalars, however--each sample is given its own `bin'.
*	4   bin vs y plotting; all y channels share a single grid.  Similar
*	    to type 3, above.
*	5   time vs y plotting; each y channel has its own grid
*	6   time vs y plotting; all y channels share a single grid
*	7   x vs y plotting, on a smith chart impedance overlay, which has
*           circles tangent on the right.  If PFO_ATTR_FG1 has been
*           used to set an alternate foreground pixel value, then, on
*           color displays, the overlay is drawn using that pixel value.
*	    This isn't an actual smith chart--just an overlay for x,y data.
*	8   x vs y plotting, on a smith chart admittance overlay, which has
*           circles tangent on the left.  If PFO_ATTR_FG2 has been
*           used to set an alternate foreground pixel value, then, on
*           color displays, the overlay is drawn using that pixel value.
*	    This isn't an actual smith chart--just an overlay for x,y data.
*	9   x vs y plotting, on a smith chart impedance overlay, which is
*           a combination of the impedance overlay on top of the
*           admittance overlay.  On color displays when alternate
*           foreground pixel values have been specified, the overlays
*           are drawn as described above.  On monochrome displays, or
*           when no alternate foreground pixel values are specified,
*           the impedance overlay is drawn with a solid line and the
*           admittance overlay is drawn with a dashed line.
*	    This isn't an actual smith chart--just an overlay for x,y data.
*
*-*/
void
pfoWinRepaint(pPfo, style, axis)
PFO_WIN *pPfo;		/* I pointer to plot window structure */
enum dl_plot_style style;/* I plot style */
int	axis;
{
    long	stat;
    int		nSamp;

    if (style == dl_line_plot) {
	stat = sydPlotSetAttr(pPfo->pMstr, SYD_PLATTR_LINE, 1, NULL);
	assertAlways(stat == S_syd_OK);
	stat = sydPlotSetAttr(pPfo->pMstr, SYD_PLATTR_MARK, 0, NULL);
	assertAlways(stat == S_syd_OK);
    }
    else if (style == dl_point_plot) {
	stat = sydPlotSetAttr(pPfo->pMstr, SYD_PLATTR_LINE, 0, NULL);
	assertAlways(stat == S_syd_OK);
	stat = sydPlotSetAttr(pPfo->pMstr, SYD_PLATTR_MARK, 1, NULL);
	assertAlways(stat == S_syd_OK);
    }

    stat = sydPlotSetAttr(pPfo->pMstr, SYD_PLATTR_WRAP, 0, NULL);
    assertAlways(stat == S_syd_OK);

    if (pPfo->pSspec->dataDim == 0) {
	if ((nSamp = pPfo->nSamples) <= 1)
	    nSamp = 1;
	stat = sydSampleSetAlloc(pPfo->pSspec, nSamp);
	assertAlways(stat == S_syd_OK);
    }
    if      (axis == 1)
	pPfo->pMstr->plotAxis = SYD_PLAX_XY;
    else if (axis == 2)
	pPfo->pMstr->plotAxis = SYD_PLAX_XYY;
    else if (axis == 3)
	pPfo->pMstr->plotAxis = SYD_PLAX_Y;
    else if (axis == 4)
	pPfo->pMstr->plotAxis = SYD_PLAX_YY;
    else if (axis == 5)
	pPfo->pMstr->plotAxis = SYD_PLAX_SMITH_IMP;
    else if (axis == 6)
	pPfo->pMstr->plotAxis = SYD_PLAX_SMITH_ADM;
    else if (axis == 7)
	pPfo->pMstr->plotAxis = SYD_PLAX_SMITH_IMM;
    else if (axis == 8) {
	pPfo->pMstr->plotAxis = SYD_PLAX_TY;
	stat = sydPlotSetAttr(pPfo->pMstr, SYD_PLATTR_WRAP, 1, NULL);
	assertAlways(stat == S_syd_OK);
    }
    else if (axis == 9) {
	pPfo->pMstr->plotAxis = SYD_PLAX_TYY;
	stat = sydPlotSetAttr(pPfo->pMstr, SYD_PLATTR_WRAP, 1, NULL);
	assertAlways(stat == S_syd_OK);
    }

    pPfo->style = style;
    pPfo->axis = axis;

    stat = sydPlotWinReplot(pPfo->pMstr);
    assert(stat == S_syd_OK);
}

/*+/subr**********************************************************************
* NAME	pfoWinRestart - restart sampling for a plot window
*
* DESCRIPTION
*	Restarts plotting after emptying the circular buffers.  The additional
*	functions are those provided by pfoWinRepaint.
*
* RETURNS
*	void
*
* BUGS
* o	multiple X vs multiple Y isn't implemented
*
* SEE ALSO
*	pfoWinRepaint, pfoWinSetAttr
*
*-*/
void
pfoWinRestart(pPfo, style, axis)
PFO_WIN *pPfo;		/* I pointer to plot window structure */
enum dl_plot_style style;/* I plot style */
int	axis;
{
    long	stat;
    int		nSamp;

    if ((nSamp = pPfo->nSamples) <= 1)
	nSamp = 1;
    stat = sydSampleSetAlloc(pPfo->pSspec, nSamp);
    assert(stat == S_syd_OK);
    sydInputResetSampled(pPfo->pSspec);
    pfoWinRepaint(pPfo, style, axis);
}

/*+/subr**********************************************************************
* NAME	pfoWinSetAttr - set plot window attributes
*
* DESCRIPTION
*	Set special attributes for the plot window.  Calls to this routine
*	are a little strange:  Depending on the attribute, either a value or
*	a pointer is required.  Two arguments exist in the call, one for a
*	value, the other for a pointer; the unused argument is to be set to
*	0 or NULL, respectively.  The appropriate form is shown for each
*	attribute.
*
*	pfoWinSetAttr(pPfo, PFO_ATTR_FG1, 0, pFgPixelValue);  sets the
*	    color to be used in drawing the smith chart impedance overlay.
*	pfoWinSetAttr(pPfo, PFO_ATTR_FG2, 0, pFgPixelValue);  sets the
*	    color to be used in drawing the smith chart admittance overlay.
*
* RETURNS
*	void
*
*-*/
void
pfoWinSetAttr(pPfo, attr, value, pArg)
PFO_WIN *pPfo;		/* I pointer to plot window structure */
PFO_ATTR attr;		/* I attribute selector--one of PFO_ATTR_xxx */
int	value;		/* I value for attribute */
void	*pArg;		/* I pointer for value for attribute */
{
    long	stat;

    if (attr == PFO_ATTR_FG1) {
	stat = sydPlotSetAttr(pPfo->pMstr, SYD_PLATTR_FG1, 0, pArg);
	assert(stat == S_syd_OK);
    }
    else if (attr == PFO_ATTR_FG2) {
	stat = sydPlotSetAttr(pPfo->pMstr, SYD_PLATTR_FG2, 0, pArg);
	assert(stat == S_syd_OK);
    }
}

/*+/subr**********************************************************************
* NAME	pfoWinWritePSFile - write a PostScript file containing the plot
*
* DESCRIPTION
*	Writes a PostScript file containing the plot currently being
*	displayed in the plot window.  Only the data presently in the
*	circular buffers will be plotted.
*
*	The size of plot on paper is similar to size on screen.  There
*	presently is no capability to specify scaling or orientation of
*	the PostScript plot.
*
*	Only monochrome PostScript output is presently supported.  Colored
*	data lines will be rendered with dashed line patterns.
*
* RETURNS
*	void
*
* BUGS
* o	text
*
* SEE ALSO
*	pfoWinOpen, pfoWinRepaint
*
*-*/
void
pfoWinWritePSFile(pPfo, filePath)
PFO_WIN *pPfo;		/* I pointer to plot window structure */
char	*filePath;	/* I path for file to contain PostScript plot */
{
    long	stat;
    int		nSamp;
    SYD_PL_MSTR master;
    SYD_PL_MSTR *pMaster;
    PFO_AXIS	*pAxis;
    SYD_PL_SLAVE *pSlave;

    pMaster = &master;
    pMaster->width = pPfo->pMstr->width;
    pMaster->height = pPfo->pMstr->height;
    stat = sydPlotInit(pMaster,
			pPfo->pSspec, PPR_WIN_POSTSCRIPT, filePath, NULL, 0);
    if (stat != S_syd_OK) {
	assertAlways(0);
    }
    stat = sydPlotSetTitles(pMaster,
			pPfo->title, pPfo->yTitle, pPfo->xTitle, NULL);
    assertAlways(stat == S_syd_OK);

    if (pPfo->style == dl_line_plot) {
	stat = sydPlotSetAttr(pMaster, SYD_PLATTR_LINE, 1, NULL);
	assertAlways(stat == S_syd_OK);
	stat = sydPlotSetAttr(pMaster, SYD_PLATTR_MARK, 0, NULL);
	assertAlways(stat == S_syd_OK);
    }
    else if (pPfo->style == dl_point_plot) {
	stat = sydPlotSetAttr(pMaster, SYD_PLATTR_LINE, 0, NULL);
	assertAlways(stat == S_syd_OK);
	stat = sydPlotSetAttr(pMaster, SYD_PLATTR_MARK, 1, NULL);
	assertAlways(stat == S_syd_OK);
    }


    pMaster->plotAxis = pPfo->pMstr->plotAxis;
    if (pMaster->plotAxis == SYD_PLAX_TY || pMaster->plotAxis == SYD_PLAX_TYY)
	stat = sydPlotSetAttr(pMaster, SYD_PLATTR_WRAP, 1, NULL);
    else
	stat = sydPlotSetAttr(pMaster, SYD_PLATTR_WRAP, 0, NULL);
    assertAlways(stat == S_syd_OK);

    pAxis = pPfo->pXAxisHead;
    while (pAxis != NULL) {
	pSlave = sydPlotChanAdd(pMaster, pAxis->pfoChan.pSchan);
	assertAlways(pSlave != NULL);
	stat = sydPlotAxisSetAttr(pSlave, SYD_PLATTR_XCHAN, 1, NULL);
	assertAlways(stat == S_syd_OK);
	pSlave->originVal = pAxis->pSlave->originVal;
	pSlave->extentVal = pAxis->pSlave->extentVal;
	pAxis = pAxis->pNext;
    }
    pAxis = pPfo->pYAxisHead;
    while (pAxis != NULL) {
	pSlave = sydPlotChanAdd(pMaster, pAxis->pfoChan.pSchan);
	assertAlways(pSlave != NULL);
	pSlave->originVal = pAxis->pSlave->originVal;
	pSlave->extentVal = pAxis->pSlave->extentVal;
	pAxis = pAxis->pNext;
    }

    stat = sydPlotWinLoop(pMaster);
    assert(stat == S_syd_OK);
}
