#include "opiStringDefs.h"
extern	char	*opiStringTab[];
/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
#undef DEBUG
#define	XSYNC	1		/* set to 0 to turn off xsync */
/*
 *    Display_manager.c contains routines 
 *    Deb Kerstiens 7/90
 *
 *
 *	Modification History
 *	DK	08/22/91	fixed conversion to double in getStripDouble
 *	DK	10/16/91	added macro replacement for static text to 
 *				setup_graphics()
 *	DK	10/16/91	fixed visibility check on visibilityNotify
 *				event so will update when changed from fully
 *				to partially obscured
 *	lrd	10/27/91	modified the XWM Hints to set text input on focus - this
 *				accomadates changes mase by Jeff for gizmos
 *	DK	12/09/91	took update_plots() out of if(iconified) loop to
 *				allow update of plots even when iconified
 *	DK	1/27/92		fixed bug in x_evhandler for ClientMessage event
 *				to screen out all but wm_delete and wm_save_yourself
 *	DK	1/28/92		modified to eliminate operator input if no display
 *				specified or if invalid name on first display--
 *				dm now exits on these conditions so can run properly
 *				from a script
 *	DK	04/15/92	added time check to avoid updating plots too often
 *				(this used too much cpu and slowed the iocs and dm
 *				down to a crawl)--now updates once per sec; slowed 
 *				strip chart updates down to about 5 Hz as well.
 *				this is a temporary fix until the synchronous data
 *				mechanism used for plotting is upgraded
 *	DK	04/20/92	added code in delete_plots that caused deleting a 
 *				display with plot window to crash dm
 *
 */

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>
#ifdef WIN32
#include <winsock.h>
#include <time.h>
#include <io.h>
#include <fcntl.h>
#else
#include <sys/file.h>
#include <sys/time.h>
#endif WIN32

/* the X library */
#include <X11/Xlib.h>
#ifdef WIN32
#include <X11/XlibXtra.h>
#endif WIN32
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include "cadef.h"
#include "dldef.h"
#include "dlinfo.h"
#include "dl_context.h"
#include "font_table.h"
#include "gizmo.h"
#include "op_gsdl_def.h"
#include "dynamic_def.h"    	/* info for dynamic graphical updates-- must be after dldef.h ,dl_context.h */
#include "message_butt.h"		/*defines for type of message button */
#include "GraphX.h"

#ifdef TOKENPTRS
#include <tokenSupport.h>
#endif

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32
#include "OperPriv.h"

extern void *tisMalloc();
extern void *tisCalloc();
extern void *tisRealloc();
extern void tisFree();
#ifdef WIN32
Window	Xfdlabel;
#endif WIN32
extern void *fdmgr_init();
extern void *gsd_sync_init();

#define BORDER_WIDTH	1		/* constant for window setup */
#define SEC_TIME_OUT	20		/* struct timeval's timeout of 100000 usec */
#define USEC_TIMEOUT	0
#define NUM_TRACES	7		/* max number of traces for strip plots */
#define SYNC_TIMEOUT	1.0		/* timeout in seconds for synchronous reads */


#ifdef GTA
void	*pfdctx;			/* file descriptor manager id */
#else
XtAppContext appContext;
#endif GTA

struct gsdl_com_llist_controller	com_llist_controller;
struct gsdl_com_llist_element		*pOld_llist_start = NULL;
int 	PrivCheckOverride = FALSE;		/* for the -p flag */
int 	TestTimer;				/* for the -t flag */
time_t 	NextTestTime;				/* for the -t flag */
static struct timeval dm_capoll = {3,0};
static struct timeval dm_blinkpoll = {0,1000000};
static long update_rate = 34;	/* update each screen no faster than this value (in milliseconds) */
Atom	proto[2];			/* global for setting window manger protocols */
int	invisibleColor = DEFAULT_INVISIBLE_COLOR; /* Pixel number used to represent the invisible color */
extern	int	displayAdded;

extern void gsdl_fd_register ();
int	opi_x_error_handler();		/* route xerrors here to get coredump */

void *read_display_lst (struct dm_llist_el *pdm_llist_el);
int setup_window (struct dm_llist_el *pDm_llist_el);
void load_display_strings (struct dm_llist_el *pdm_llist_el);
int free_display_strings (struct dm_llist_el *pdm_llist_el);
int setup_graphics (struct dm_llist_el *pdm_llist_el, struct dm_llist_el *pfirst_llist_el);
int init_color_rule_instance (struct dm_llist_el	*pdm_llist_el,
			    int			*pcnt_list,	
			    struct rti_color_rules *pRti_cr,
			    int			rule,
			    int			args,
			    unsigned int	*rulechan,
			    void		*rulecolorfg,
			    void		*rulecolorbg);
void *build_rti_color_rules (struct dm_llist_el *pDm_llist_el, int *pcnt_list);
void *build_rti_dynamic_list (unsigned short cnt, struct dm_llist_el *pdm_llist_el, int *pcnt_list);
void *build_rti_mon_list (unsigned short cnt, struct dm_llist_el *pdm_llist_el, int *pcnt_list);
void *build_rti_ctrl_list (unsigned short cnt, struct dm_llist_el *pdm_llist_el, int *pcnt_list);
void *build_rti_plot_list (unsigned short cnt, struct dm_llist_el *pdm_llist_el);
void *string_index_to_display_str (char **pstring_table, unsigned short textix, char **pdis_str, unsigned *pstr_len);
void x_evhandler (struct dm_llist_hdr *pdm_llist_hdr);
int setup_strip (struct dm_llist_el *pdm_llist_el, struct dl_context *pdlc, struct rti_update_plot *pPlot_el);
double getStripDouble (int index, caddr_t *userarg);
int init_display_mgr (struct dm_llist_hdr *pdm_llist_hdr,
		  struct gsdl_com_llist_controller *pcom_llist_controller,
		  int argc, char *argv[]);
int Usage (void);
void dm_CaException (struct exception_handler_args arg);
int delete_display (struct dm_llist_hdr *pDm_llist_hdr,struct dm_llist_el *pDm_llist_el);
void delete_plots (Display *display, XContext win_context,
		   struct rti_update_plot *pRti_plot, unsigned short nplots);
int proc_plot (struct rti_update_plot *pRti_plot, XEvent the_event);
int dmSetup_cmap (struct dm_llist_el *pDm_llist_el);
int dmSetup_crules (struct dm_llist_el *pDm_llist_el);
int add_sting_table_strings_to_list (LIST *pst, struct dm_llist_el *pDm_llist_el);
int fixup_rulechan_strings (struct dm_llist_el *pDm_llist_el, int rule, int args,
			unsigned int *rulechan, int this_channel);
char *get_pc_string (struct dl_hdr *pdl_hdr, unsigned int channel);
char *get_rule_string (struct dm_llist_el *pDm_llist_el, int rule, int line);
static void dm_capollfunc (void	*pfdctx);
static void dm_blinkpollfunc (struct dm_llist_hdr *pdm_llist_hdr);
int kludgeblinker (void);
int convert_string_index_to_ptr (void);

long 	lastLicenseCheck = 0;

char	*opiStringTab[OPI_NUM_STRINGS];	/* OPI_NUM_STRINGS defined in stringDefs.h */
	
struct dm_llist_hdr	dm_llist_hdr;

/* CodeBase stuff to support the launch facility */

C4CODE	codebase;
D4DATA	*launchData;
F4FIELD	*nextFld, *prevFld, *fileFld, *paramFld;
T4TAG	*tag;

extern int	caHasUpdated;

struct dm_llist_el	*pDm_llist_pos = NULL;	/* current display in callup stack */
int			maxDisplays;		/* maximum number of displays allowed */
int			numDisplays;		/* number of active displays */
/*----------*/
main(argc, argv)
int	argc;
char	*argv[];
	{
	struct dm_llist_el	*pdm_llist_el;
	struct dm_llist_el	*pPrev_el;
	struct dm_llist_el	*pDm_llist_el;
	XRectangle		rec;		/* x,y,ht,wd struct */
	struct dl_object_hdr	*pobj;
#ifdef WIN32
	int			tzp;
#else
	struct timezone		tzp;
#endif WIN32
	float			update_elapsed;
	float			strip_elapsed;
	float			plot_elapsed;
	struct timeval		cur_time;
	struct timeval		tp;
	static struct timeval	update_time = {0,0};
	static struct timeval	strip_time = {0,0};
	static struct timeval	plot_time = {0,0};
	static struct timeval	timeout = {SEC_TIME_OUT, USEC_TIMEOUT};
	char			path [200], str [200];
	int			firstTime = 1;

#ifdef WIN32

	/* handle WinSocket initialization */
	WSADATA wsaData; 
	WORD    wVersionRequested = MAKEWORD(1, 1);
	int	err;
	extern int _fmode;

	_fmode = _O_BINARY;

	err = WSAStartup(wVersionRequested, &wsaData); 
	if (err != 0)
	{
		/* Tell the user that we couldn't find a useable */ 
		/* winsock.dll.     */ 
		printf ("WinSocket initialization error %d\n", err);
		return -1;
	}


#ifdef Xm
	HCLXmInit();		/* HCL - initialize the Xm DLL */
#endif Xm
	/* initialize XDK environment */
	HCLXlibInit();
	HCLXtInit();
	HCLXtInit();
#endif WIN32
	/*
	Handle TIS licensing
	*/

#if !(defined(ultrix) || defined(WIN32))
	tisHardLicense();
	xLicense();
	lastLicenseCheck = time(NULL);
#endif

/*
	GCT	21-Nov-1994

	The following function loads in the translated strings created with the
	'strConv' utility. If OPISTRINGTABLE environment variable exits, get the
	string table from there.
*/
	
	if (getenv ("OPISTRINGTABLE"))
	    tisLoadStrings (getenv ("OPISTRINGTABLE"), opiStringTab, OPI_NUM_STRINGS);
	else
#if !(defined(AXP) || defined(SOLARIS) || defined(WIN32))
	    tisLoadStrings("/home/tis4000/tasks/opiStringTable",opiStringTab,OPI_NUM_STRINGS);
#else
	    tisLoadStrings("/tis4000/tasks/opiStringTable",opiStringTab,OPI_NUM_STRINGS);
#endif

	/* Initialize the codeBase system and open the launch database. */

	d4init (&codebase);
#ifdef CB60
	codebase.errOpen = 0;
	codebase.errTagName = 0;
#else
	codebase.open_error = 0;
	codebase.tag_name_error = 0;
#endif CB60
#ifdef WIN32
	strcpy (path, "\\tis4000db\\grphicdb");
#else
	resolveLinks ("/tis4000db/grphicdb", path);
#endif WIN32
	if (launchData = d4open (&codebase, path)) {
	    nextFld = d4field (launchData, "NEXT");
	    prevFld = d4field (launchData, "PREV");
	    fileFld = d4field (launchData, "FILE");
	    paramFld = d4field (launchData, "PARAM");
	    tag = d4tag (launchData, "TAG");
	    if (tag == NULL) {
		printf ("grphicdb database does not contain the TAG tag -- launch facility not available.\n");
		d4close (launchData);
		launchData = NULL;
	    } else {
		d4tag_select (launchData, tag);
	    }
	}

	/* get maxDisplays value from environment, else use 128 */
	if (getenv ("OPIMAXDISPLAYS"))
	{
	    maxDisplays = atoi (getenv ("OPIMAXDISPLAYS"));
	    if (maxDisplays <= 1)
		maxDisplays = 2; /* must allow at least two displays! */
	}
	else
	{
	    maxDisplays = 128;
	}

	init_display_mgr(&dm_llist_hdr,&com_llist_controller, argc, argv);

	if (TestTimer)
	    NextTestTime = time(0) + TestTimer;

	for (;;)
		{
		/* if test time period is not zero, and time delay expired, invoke first related display of top display */
		if (TestTimer && time(0) > NextTestTime) {
		    struct rti_rel_disp		*pRti_rel_disp = pDm_llist_pos->prti_rel_disp;
		    struct dl_related_display	*pDl =
			    (struct dl_related_display *)((short *)pDm_llist_pos->pdl_hdr + pDm_llist_pos->pdl_hdr->ord);
		    char				**pDl_str_table	= pDm_llist_pos->pdl_strings;
		    int				i;

		    for (i = 0; i < pDm_llist_pos->pdl_hdr->nrd; i++, pRti_rel_disp++) {
			int	name_index = (int)(pDl->display[i].name);
			if (strlen(pDl_str_table[name_index]) > 0) {
			    add_related_display (pRti_rel_disp, pRti_rel_disp->pgz, i);
			    break;
			}
		    }
		    NextTestTime = time(0) + TestTimer;
		}


		/* process events that have occurred since last time (add new displays, */
		/* setup for screen updates based on db change)*/
		/* if new display requested, add it to com and dm llists */
		/* two loops so can begin ca stuff then build rti dynamic info list */
		/* loop one : read display file and init com_llist struct */
		gettimeofday(&cur_time, &tzp);
		update_elapsed = (float)((cur_time.tv_sec + (.000001 * cur_time.tv_usec)) - 
				(update_time.tv_sec + (.000001 * update_time.tv_usec))); 
		strip_elapsed = (float)((cur_time.tv_sec + (.000001 * cur_time.tv_usec)) -
				(strip_time.tv_sec + (.000001 * strip_time.tv_usec))); 
		plot_elapsed = (float)((cur_time.tv_sec + (.000001 * cur_time.tv_usec)) -
				(plot_time.tv_sec + (.000001 * plot_time.tv_usec))); 

		if (dm_llist_hdr.new_llist_flag)
			{
			dm_llist_hdr.new_llist_flag = 0;

			/* add new display(s) to display_man_llist */
			pPrev_el = pdm_llist_el = dm_llist_hdr.pNew_llist;

			/* skip to first new el, then begin adding */
			while (pdm_llist_el != NULL)
				{ /* need to add new display */
				if (pdm_llist_el->pdl_hdr == NULL)
					{/* read in display list */
					if ((pdm_llist_el->pdl_hdr = (struct dl_hdr *)read_display_lst(pdm_llist_el))== NULL)
						{
						printf(opiStringTab[STR_Couldnt_read_file_Hsn       ], pdm_llist_el->display_name);
						/* delete el from list here if bad display name,read err, etc.--add retry?? */
						/* if first el on list, ask for new display name */
						if (pdm_llist_el == dm_llist_hdr.pNew_llist)
							{
							printf(opiStringTab[STR_No_display_name_Hs_exiting_d], pdm_llist_el->display_name);
							NormalExit(1);
							}
		  				  else
							{
							pPrev_el->pNext_el = pdm_llist_el->pNext_el;
							tisFree(pdm_llist_el);
							pdm_llist_el = pPrev_el->pNext_el;
							}
						}
					else
						{
						/* we have a new display */
						/* set current display pointer to the new display */
						pDm_llist_pos = pdm_llist_el;
						numDisplays ++;
						/* add an element to com llist */
						/* if not first el on list */
						if (pdm_llist_el != dm_llist_hdr.pNew_llist)
							{
							pdm_llist_el->pcom_el = (struct gsdl_com_llist_element *)
							tisCalloc(__FILE__, __LINE__, 1, sizeof(struct gsdl_com_llist_element));
							pPrev_el->pcom_el->pNext_element = pdm_llist_el->pcom_el; 
							}
						/* init new com_llist entry */     
						pdm_llist_el->pcom_el->pDl_hdr = pdm_llist_el->pdl_hdr;

						/* set up table of pointers to strings */
						load_display_strings(pdm_llist_el);

						/* setup dl color rules for the display */
						if (dmSetup_crules(pdm_llist_el) == -1)
							continue;

						/* call comm (ca) setup here */
						com_llist_controller.new_llist_flag = 1;
						gsdl_llist_main(&com_llist_controller, &pOld_llist_start, pdm_llist_el->ppMacro_pairs);

						/* setup window, get ptr to dyn and stat from display list, build rti dynamic list */
						if ((setup_graphics(pdm_llist_el, dm_llist_hdr.pNew_llist)) < 0)
							{
							delete_display(&(dm_llist_hdr), pdm_llist_el);
							continue;
							}

						/* Check if the operator has the privilege to run this graphic */

						
						gsdl_main_macro(str, pdm_llist_el->pdl_strings[(int)(pdm_llist_el->pdl_hdr->privname)],
							dm_llist_hdr.pNew_llist->ppMacro_pairs, sizeof (str));
						if (strlen(str) > 0) {
						    int priv = OperPriv (&codebase, str);
						    if (priv <= 0) {
							char *pchoices[2], msg [100], *msgs [2];
							pchoices[0] = opiStringTab[STR_ok                          ];
							pchoices[1] = 0;
							strcpy (msg, opiStringTab[STR_Insufficient_priviledge_to_d]);
							msgs [0] = msg;
							msgs [1] = 0;
						perform_dialog(DefaultScreenOfDisplay(dm_llist_hdr.pNew_llist->dl_context.d),
								600, 300, 21,
								msgs,
								pchoices, 21);
							delete_display(&(dm_llist_hdr), pdm_llist_el);
							continue;
						    }
						}

						/* Run the automatic startup program if one. */

						gsdl_main_macro(str, pdm_llist_el->pdl_strings[(int)(pdm_llist_el->pdl_hdr->startprogram)],
							dm_llist_hdr.pNew_llist->ppMacro_pairs, sizeof (str));
						if (strlen (str) > 0) {
						    system (str);
						}
 
 						/* set up so dm won't be killed if window deleted */
						XSetWMProtocols(pdm_llist_el->dl_context.d, pdm_llist_el->dl_context.w, proto, 2);
			
						setup_related_displays(&dm_llist_hdr, &(pdm_llist_el->dl_context), pdm_llist_el,
							(short *)pdm_llist_el->pdl_hdr + pdm_llist_el->pdl_hdr->ord, 
							pdm_llist_el->pdl_hdr->nrd,
							pdm_llist_el->pdl_color_map_index);
			
						prestart_related_displays(&dm_llist_hdr, &(pdm_llist_el->dl_context), pdm_llist_el,
							(short *)pdm_llist_el->pdl_hdr + pdm_llist_el->pdl_hdr->ord, 
							pdm_llist_el->pdl_hdr->nrd,
							pdm_llist_el->pdl_color_map_index);

						setup_execute_scripts(&dm_llist_hdr, &(pdm_llist_el->dl_context), pdm_llist_el,
							(short *)pdm_llist_el->pdl_hdr + pdm_llist_el->pdl_hdr->oes,
							pdm_llist_el->pdl_hdr->nes,
							pdm_llist_el->pdl_color_map_index);
			
						setup_kill_displays(&dm_llist_hdr, &(pdm_llist_el->dl_context), pdm_llist_el,
							(short *)pdm_llist_el->pdl_hdr + pdm_llist_el->pdl_hdr->okd,
							pdm_llist_el->pdl_hdr->nkd,
						  pdm_llist_el->pdl_color_map_index);

						ca_pend_io(0.0);
						pPrev_el = pdm_llist_el;
						pdm_llist_el = pdm_llist_el->pNext_el;
						}    
					}
				else
					{ /* just keep moving along */
					pPrev_el = pdm_llist_el;
					pdm_llist_el = pdm_llist_el->pNext_el;
					}
				}/* end while */
			} /* end if new els to add */

		/* get precise time (for update throttle) */
		gettimeofday(&tp, &tzp);

		/* need to check to see if we need this--it was part of the original code */
		/* Xpending does a flush */
		displayAdded = 0;
		if(XPending(dm_llist_hdr.pNew_llist->dl_context.d)) {
 		   x_evhandler(&dm_llist_hdr);
		}
		/* if over display limit, kill 2nd display (1st display is sacred)
		** if over limit */
		while (numDisplays > maxDisplays)
		{
			dmExit_display(&dm_llist_hdr, dm_llist_hdr.pNew_llist->pNext_el, -1); /* <- note the -1! */
    		}
							
#ifdef DEBUG
fprintf(stderr,opiStringTab[STR_calling_fdmgr_pend_eventHld_], time(0));
#endif
		/* If a new display is added we don't want to potentially block on */
		/* this fdmgr_pend_event. */

#ifdef DEBUG
fprintf(stderr,opiStringTab[STR_ca_pend_io_atHldn           ], time(0));
#endif
		ca_pend_io(0.0);	/* get the callbacks executed if this is a ca event */

		if ((!displayAdded || firstTime)) {
		    if (caHasUpdated) {
		      timeout.tv_sec = 0;
		      timeout.tv_usec = 0;
		    } else {
		      if (TestTimer) {
			timeout.tv_sec = NextTestTime - time(0);
			if (timeout.tv_sec < 0) timeout.tv_sec = 0;
			timeout.tv_usec = 0;
		      } else {
			timeout.tv_sec = 20;
			timeout.tv_usec = 0;
		      }
		    }
		      
#if defined(WIN32)
		    ca_pend_io (0.25);
#elif (defined(GTA))
		    fdmgr_pend_event(pfdctx, &timeout);
#else
		    ca_pend_io (timeout.tv_sec);
#endif GTA
		    firstTime = 0;
		}
	
		caHasUpdated = 0;
		pDm_llist_el = dm_llist_hdr.pNew_llist;
		while ((pDm_llist_el != NULL) &&(pDm_llist_el->pdl_hdr != NULL))
			{
			/*if this display is recently modified, iconified or obscured , don't do graphic updates */
			float elapsed = (float)((tp.tv_sec + (.000001 * tp.tv_usec))
				 - (pDm_llist_el->update_time.tv_sec + (.000001 * pDm_llist_el->update_time.tv_usec))); 
			if ((elapsed >= ((float)update_rate)/1000) && pDm_llist_el->visible)
				{
				pDm_llist_el->update_time = tp;
				pobj = &(pDm_llist_el->pdl_hdr->object);
				rec.x		= /* pobj->x */ 0;
				rec.y		= /* pobj->y */ 0;
				rec.width	= pobj->width;
				rec.height	= pobj->height;

#ifdef DEBUG
fprintf(stderr,opiStringTab[STR_tcalling_update_color_rulesH], time(0));
#endif
				update_color_rules(pDm_llist_el);

#ifdef DEBUG
fprintf(stderr,opiStringTab[STR_tcalling_update_dynamic_grap], time(0));
#endif
				update_dynamic_graphics(&(pDm_llist_el->dl_context),
				  &(pDm_llist_el->static_dlcontext),
				  pDm_llist_el->pdl_color_map_index,
				  pDm_llist_el->pdl_dynamic, 
				  pDm_llist_el->prti_dynamic, 
				  pDm_llist_el->pdl_hdr->ndyn, &rec,
				  pDm_llist_el, FALSE,
				  pDm_llist_el->pdl_hdr->bclr);

#ifdef DEBUG
fprintf(stderr,opiStringTab[STR_tcalling_update_mon_objsHldn], time(0));
#endif
				update_mon_objs(&(pDm_llist_el->dl_context),
				  pDm_llist_el->pdl_color_map_index,
				  pDm_llist_el->pdl_mon, 
				  pDm_llist_el->prti_mon, 
				  pDm_llist_el->pdl_hdr->nmon, &rec,
				  pDm_llist_el, FALSE);
    
#ifdef DEBUG
fprintf(stderr,opiStringTab[STR_tcalling_update_controllersH], time(0));
#endif
				update_controllers(&(pDm_llist_el->dl_context),
				  pDm_llist_el->pdl_color_map_index,
				  pDm_llist_el->pdl_ctrl, 
				  pDm_llist_el->prti_ctrl, 
				  pDm_llist_el->pdl_hdr->nctrl, &rec,
				  pDm_llist_el, FALSE);
				}

			update_plots(&(pDm_llist_el->dl_context),
			  pDm_llist_el->pdl_color_map_index,
			  pDm_llist_el->pdl_plot, 
			  pDm_llist_el->prti_plot, 
			  pDm_llist_el->pdl_hdr->nplot, &rec, FALSE, &pfdctx,
			  strip_elapsed, plot_elapsed);

			pDm_llist_el = pDm_llist_el->pNext_el;
			}
		if (update_elapsed > UPDATE_TIME)
			update_time = cur_time;
		if (strip_elapsed > UPDATE_STRIP_TIME)
			strip_time = cur_time;
		if (plot_elapsed > UPDATE_PLOT_TIME)
			plot_time = cur_time;
		}
	}

/*----------*/
void *read_display_lst(pdm_llist_el)
struct dm_llist_el *pdm_llist_el;
	{
	int 	fd;
	int 	file_len;
	char 	*display_ptr;

#ifdef WIN32
	if ((fd = open_display_list_file(getenv("EPICS_DISPLAY_PATH"), pdm_llist_el->display_path,
				sizeof(pdm_llist_el->display_path), pdm_llist_el->display_name, _O_BINARY|_O_RDONLY)) == -1)
#else
	if ((fd = open_display_list_file(getenv("EPICS_DISPLAY_PATH"), pdm_llist_el->display_path,
				sizeof(pdm_llist_el->display_path), pdm_llist_el->display_name, O_RDONLY)) == -1)
#endif WIN32
		return NULL;

	/* get the file length in bytes */
#ifdef SOLARIS
	file_len = lseek(fd,0L,SEEK_END);	/* get file len */
	lseek(fd,0L,SEEK_SET);				/* back to beginning */
#elif defined(WIN32)
	file_len = _lseek(fd, 0L, SEEK_END);	/* get file len */
	_lseek(fd, 0L, SEEK_SET);				/* back to beginning */
#else
	file_len = lseek(fd,0L,L_XTND);		/* get file len */
	lseek(fd,0L,L_SET);					/* back to beginning */
#endif
	/* allocate storage for display list and read it in */
	if ((display_ptr = tisCalloc(__FILE__, __LINE__, 1, file_len)) == NULL)
		{
		printf(opiStringTab[STR_error_in_file_readcalloc_fai]);
#ifdef WIN32
		_close(fd);
#else
		close(fd);
#endif WIN32
		return NULL;
		}
#ifdef WIN32
	if ((_read(fd,display_ptr,file_len)) != file_len)
#else
	if ((read(fd,display_ptr,file_len)) != file_len)
#endif WIN32
  		{
		printf(opiStringTab[STR_error_in_file_readnum_bytes_]);
#ifdef WIN32
		_close(fd);
#else
		close(fd);
#endif WIN32
		return NULL;
		}
#ifdef WIN32
	_close(fd);
#else
	close(fd);
#endif WIN32
	if (valid_dl_test((struct dl_hdr *)display_ptr) < 0)
		return NULL;

	return display_ptr;
	}


int setup_window(pDm_llist_el)
struct dm_llist_el	*pDm_llist_el;
	{
	Window			root_window;
	XSizeHints		size_hints;		/* hints to display mgr */
	XGCValues		GC_values;
	struct dl_hdr		*pDl_hdr = pDm_llist_el->pdl_hdr;
	struct dl_context	*pDl_context = &(pDm_llist_el->dl_context);
	unsigned long		*pDl_color_map	= pDm_llist_el->pdl_color_map_index;
	int			foreground = COLOR_LOOKUP(pDl_color_map, pDl_hdr->clr);
	int			background = COLOR_LOOKUP(pDl_color_map, pDl_hdr->bclr);
	int			iconize;

	the_screen = DefaultScreen(pDm_llist_el->dl_context.d);
	root_window = DefaultRootWindow(pDm_llist_el->dl_context.d);

	/*
	 * force these to reasonable values
	 */
	foreground = (foreground == invisibleColor) ?
			WhitePixel(pDm_llist_el->dl_context.d, the_screen) : foreground;
	background = (background == invisibleColor) ?
			BlackPixel(pDm_llist_el->dl_context.d, the_screen) : background;
	
	/* Set up the window manager size hints according to display list info. */
	size_hints.x		= pDl_hdr->object.x;
	size_hints.y		= pDl_hdr->object.y;
	size_hints.width	= pDl_hdr->object.width;
	size_hints.height	= pDl_hdr->object.height;
	size_hints.flags	= PSize|PPosition;

	/* Create a window of fixed size, origin, and borderwidth. */
	pDl_context->w = XCreateSimpleWindow(pDl_context->d, root_window, size_hints.x, size_hints.y,
	  size_hints.width, size_hints.height, BORDER_WIDTH, foreground, background);

#ifdef WIN32
	if (Xfdlabel == 0)
	    Xfdlabel = pDl_context->w;
#endif WIN32
	/* Set the usual properties. */
	XSetStandardProperties(pDl_context->d, pDl_context->w, pDm_llist_el->display_name, pDm_llist_el->display_name,
	  None, NULL, 0, &size_hints);

	/* Select exposure events and button press events. */
	if (pDm_llist_el->pdl_hdr->exitaction ==  2 /* auto-exit */ )
	{
		/* set LeaveNotify */
		XSelectInput(pDl_context->d, pDl_context->w,
		  ExposureMask|ButtonReleaseMask|StructureNotifyMask|VisibilityChangeMask|ButtonPressMask|Button3MotionMask|LeaveWindowMask);
	}
	else
	{
		XSelectInput(pDl_context->d, pDl_context->w,
		  ExposureMask|ButtonReleaseMask|StructureNotifyMask|VisibilityChangeMask|ButtonPressMask|Button3MotionMask);
	}

	/*
	 *	Allow the window manager to set focus on this window and its children.
	 *	This is placed here to get the input to be accepted on text widgets by LRD 10-27-91
	 */
	{
	XWMHints	wmh;

	wmh.window_group = root_window;
	wmh.initial_state = NormalState;
	wmh.input = True;
	wmh.flags = InputHint;
	XSetWMHints(pDl_context->d, pDl_context->w, &wmh);
	}

	XMapWindow(pDl_context->d,pDl_context->w);    /* Tell the server to make the window visible. */

	iconize = *(pDm_llist_el->display_name + strlen (pDm_llist_el->display_name) + 1) == 0x7F;
	if (iconize) {
		XIconifyWindow (pDl_context->d, pDl_context->w, DefaultScreen (pDl_context->d));
	}
	else if (pDm_llist_el->pdl_hdr->exitaction == 2 /* window is autoexit */ )
	{
		/* warp pointer to center of new window */
		XWarpPointer (pDl_context->d, None, root_window, 0, 0, 0, 0,
			size_hints.x + size_hints.width/2, size_hints.y + size_hints.height/2);
 	}
	/* setup default gc for dynamics */
	GC_values.foreground = foreground;
	GC_values.background = background;
	GC_values.join_style = JoinRound;
	GC_values.cap_style = CapProjecting;
	GC_values.fill_rule = WindingRule;
	pDl_context->gc = XCreateGC(pDl_context->d, pDl_context->w,
	  GCForeground|GCBackground|GCJoinStyle|GCCapStyle|GCFillRule, &GC_values);
 
	/* set up dl_context for displaying static objects */
	pDm_llist_el->static_dlcontext = *pDl_context;
	/* get a new gc to avoid copying and restoring values between dyn and stat*/
	pDm_llist_el->static_dlcontext.gc = XCreateGC(pDl_context->d, pDl_context->w,
	  GCForeground|GCBackground|GCCapStyle|GCFillRule, &GC_values);

	XFlush(pDl_context->d);

	return 0;
	}



void load_display_strings(pdm_llist_el)
struct dm_llist_el *pdm_llist_el;
	{
	unsigned short	i;
	short			*pdl_string;

	if (pdm_llist_el->pdl_hdr->nstr <= 0)
		return;

	pdm_llist_el->dl_context.pstring_table =	/* set this for use in draw_static */
	pdm_llist_el->pdl_strings = (char **)tisCalloc(__FILE__, __LINE__, pdm_llist_el->pdl_hdr->nstr, sizeof(char *));

	pdl_string = (short *)(pdm_llist_el->pdl_hdr) + pdm_llist_el->pdl_hdr->ostr;
	for (i = 0; i < pdm_llist_el->pdl_hdr->nstr; i++)
		{
		char	*ptr = (char *)pdl_string + sizeof(struct dl_string);
      
		if (strlen(ptr) > 0 && pdm_llist_el->ppMacro_pairs != NULL)
			{
			char	buf[500];

			gsdl_main_macro(buf, ptr, pdm_llist_el->ppMacro_pairs, sizeof buf);
			pdm_llist_el->pdl_strings[i] = strcpy((char *)tisCalloc(__FILE__, __LINE__, strlen(buf)+1, sizeof (char)), buf);
			}
		else
			pdm_llist_el->pdl_strings[i] = ptr;
#ifdef DEBUG
printf(opiStringTab[STR_stringHdHsn                 ], i, pdm_llist_el->pdl_strings[i] ? pdm_llist_el->pdl_strings[i] : "none");
#endif
		pdl_string += ((struct dl_element_hdr *)pdl_string)->nwords;
		}
	}


int free_display_strings(pdm_llist_el)
struct dm_llist_el *pdm_llist_el;
	{
	if (pdm_llist_el->pdl_hdr->nstr > 0)
		{
		/*
		if (pdm_llist_el->ppMacro_pairs != NULL)
		*/
			{
			int		i;
			
			for (i = 0; i < pdm_llist_el->pdl_hdr->nstr; i++)
				if (strlen(pdm_llist_el->pdl_strings[i]) > 0 && pdm_llist_el->ppMacro_pairs != NULL)
					tisFree(pdm_llist_el->pdl_strings[i]);
			}
		}
	tisFree(pdm_llist_el->pdl_strings);
	return 0;
	}


int setup_graphics(pdm_llist_el, pfirst_llist_el)
struct dm_llist_el *pdm_llist_el, *pfirst_llist_el;
	{
	int		*pcnt_list; 

	pdm_llist_el->dl_context.d = pfirst_llist_el->dl_context.d;

	/* setup dl color map for the display, must be done before dynamic obj setup */
	if (dmSetup_cmap(pdm_llist_el) == -1)
		return -1;

	if (pdm_llist_el->pdl_hdr->nwords_sta > 0)	/* locate static part of display list */
		{
		pdm_llist_el->pstatic = (short *)(pdm_llist_el->pdl_hdr) + 
		pdm_llist_el->pdl_hdr->osta;
		}

	/* create list for setting up set_modified bit indexes */
	pcnt_list = (int *)tisCalloc(__FILE__, __LINE__, pdm_llist_el->pdl_hdr->npc, sizeof(int));

	if (pdm_llist_el->pdl_hdr->ndyn > 0)	/* locate dynamic part of display list */
		{
		pdm_llist_el->pdl_dynamic = (short *)(pdm_llist_el->pdl_hdr) + 
						pdm_llist_el->pdl_hdr->odyn;
		/* build list of static objs with dynamic color */
		pdm_llist_el->prti_dynamic = build_rti_dynamic_list(pdm_llist_el->pdl_hdr->ndyn,
								pdm_llist_el, pcnt_list);
		}

	if (pdm_llist_el->pdl_hdr->nmon > 0)  	/* build list of monitor-type objects */
		{
		pdm_llist_el->pdl_mon = (short *)(pdm_llist_el->pdl_hdr) +
						pdm_llist_el->pdl_hdr->omon;
		pdm_llist_el->prti_mon = build_rti_mon_list(pdm_llist_el->pdl_hdr->nmon,
							pdm_llist_el,pcnt_list);
		}

	if(pdm_llist_el->pdl_hdr->nctrl > 0)	/* build list of controller-type objects */
		{
		pdm_llist_el->pdl_ctrl =  (short *)(pdm_llist_el->pdl_hdr) +
						pdm_llist_el->pdl_hdr->octrl;
		pdm_llist_el->prti_ctrl = build_rti_ctrl_list(pdm_llist_el->pdl_hdr->nctrl,
						pdm_llist_el, pcnt_list);
		}

	/* build list of color rule objects */
	pdm_llist_el->prti_color_rules = build_rti_color_rules(pdm_llist_el, pcnt_list);

	/* setup for window creation, fill in dl_context,GC_values for this window */
	setup_window(pdm_llist_el);

	/* Fill in info and set up for selection on window in X event handler */
	pdm_llist_el->evwin.plot_win = FALSE;
	pdm_llist_el->evwin.puser = pdm_llist_el;
	XSaveContext(pdm_llist_el->dl_context.d, pdm_llist_el->dl_context.w,
			pdm_llist_el->win_context, (XPointer)&(pdm_llist_el->evwin));
 
	if(pdm_llist_el->pdl_hdr->nplot > 0)	/* build list of plot-type objects */
		{
		pdm_llist_el->pdl_plot =  (short *)(pdm_llist_el->pdl_hdr) +
						pdm_llist_el->pdl_hdr->oplot;
		pdm_llist_el->prti_plot = build_rti_plot_list(pdm_llist_el->pdl_hdr->nplot, 
								pdm_llist_el);
		}
	tisFree(pcnt_list);
	return 0;
	}

#define BITS_PER_INT	(sizeof(int) * 8)

int init_color_rule_instance(pdm_llist_el, pcnt_list, pRti_cr, rule, args, rulechan, rulecolorfg, rulecolorbg)
struct dm_llist_el	*pdm_llist_el;
int			*pcnt_list;	
struct rti_color_rules	*pRti_cr;
int			rule;
int			args;
unsigned int		*rulechan;
void			*rulecolorfg;
void			*rulecolorbg;
	{
	struct dl_color_rule	*prule	= pdm_llist_el->dl_context.pcrules->pdl_crules;
	unsigned short		i;

	pRti_cr->prulecolorfg	= rulecolorfg;
	pRti_cr->prulecolorbg	= rulecolorbg;
	pRti_cr->prule			= NULL;

	for (i = 0; i < pdm_llist_el->dl_context.pcrules->nrules; i++, prule++)	
		if(strcmp(pdm_llist_el->pdl_strings[(int)(prule->name)], pdm_llist_el->pdl_strings[rule]) == 0)
			{
			pRti_cr->prule = prule;
			break;
			}
	if (pRti_cr->prule == NULL)
		{
		printf(opiStringTab[STR_unresolved_color_ruleHd_in_i], rule);
		*pRti_cr->prulecolorfg 		= (char)BAD_COLOR_DATA;
		*pRti_cr->prulecolorbg 		= (char)BAD_COLOR_DATA;
		return -1;
		}
	for (i = 0; i < MAX_RULE_LINES; i++)
		{
		pRti_cr->line[i].pComel = &(pdm_llist_el->pcom_el->pCom_table[rulechan[i]]);
		(pcnt_list[rulechan[i]])++;
		}
	return 0;
	}

void *build_rti_color_rules(pDm_llist_el, pcnt_list)
struct dm_llist_el	*pDm_llist_el;
int					*pcnt_list;	
	{
	unsigned int				i;
	struct dl_dynamic_attribute	*pdyn;
	struct dl_element_hdr		*pelem;
	struct rti_color_rules		*pRti_color_rule;
	struct rti_color_rules		*pRti_cr;

	/*
	 * create the color rule instance list
	 */
	pRti_cr =
	pRti_color_rule = (struct rti_color_rules *)tisCalloc(__FILE__, __LINE__, pDm_llist_el->ncolor_rule_objs, sizeof(struct rti_color_rules));

	/* for each element on the dynamic_graphic display list */
	pdyn = (struct dl_dynamic_attribute *)((short *)(pDm_llist_el->pdl_hdr) + pDm_llist_el->pdl_hdr->odyn);
	for (i = 0; i < pDm_llist_el->pdl_hdr->ndyn; /* see i++ below */ )
		{
		switch (pdyn->element.type)
		  {
		  case dl_basic_attribute:
			break;
		  case dl_dynamic_attribute:
		  	{
			struct dl_dyn_param	*p	= &pdyn->attr.param;

			if (pdyn->attr.mod.clrmod == dl_clrmod_rule && *pDm_llist_el->pdl_strings[(int)(p->clrrule)])
				{
				if (init_color_rule_instance(pDm_llist_el, pcnt_list, pRti_cr, (int)p->clrrule,
					    (int)p->clrargs, (unsigned int *)&p->rulechan[0],
				  &p->rulecolorfg, &p->rulecolorbg) < 0)
					pdyn->attr.mod.clrmod = dl_clrmod_static;
				pRti_cr->pclrrule_has_changed = &pDm_llist_el->prti_dynamic[i].clrrule_has_changed;
#ifdef DEBUG
printf(opiStringTab[STR_d_p0xHxn                    ],pRti_cr->pclrrule_has_changed);
#endif
				pRti_cr++;
				}
			}
			break;
		  case dl_rec:
		  case dl_line:
		  case dl_oval:
		  case dl_arc:
		  case dl_text:
			i++;		/* increment count only for graphic els */
			break;
		  }
		pdyn = (struct dl_dynamic_attribute *)((short *)pdyn + ((struct dl_element_hdr *)pdyn)->nwords);
		}

	/* for each element on the monitor display list */
	pelem = (struct dl_element_hdr *)((short *)(pDm_llist_el->pdl_hdr) + pDm_llist_el->pdl_hdr->omon);
	for (i = 0; i < pDm_llist_el->pdl_hdr->nmon; i++)
		{
		unsigned char	*clrmod;
		int				clrrule;
		int				clrargs;
		unsigned int	*rulechan;
		void			*rulecolorfg;
		void			*rulecolorbg;

		switch (pelem->type)
		  {
		  case dl_bar:
			clrmod		= &((struct dl_bar *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_bar *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_bar *)pelem)->clrargs);
			rulechan	= (unsigned int *)&(((struct dl_bar *)pelem)->rulechan[0]);
			rulecolorfg	= &((struct dl_bar *)pelem)->rulecolorfg;
			rulecolorbg	= &((struct dl_bar *)pelem)->rulecolorbg;
			break;
		  case dl_meter:
			clrmod		= &((struct dl_meter *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_meter *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_meter *)pelem)->clrargs);
			rulechan	= (unsigned int *)&(((struct dl_meter *)pelem)->rulechan[0]);
			rulecolorfg	= &((struct dl_meter *)pelem)->rulecolorfg;
			rulecolorbg	= &((struct dl_meter *)pelem)->rulecolorbg;
			break;
		  case dl_tank:
			clrmod		= &((struct dl_tank *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_tank *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_tank *)pelem)->clrargs);
			rulechan	= (unsigned int *)&(((struct dl_tank *)pelem)->rulechan[0]);
			rulecolorfg	= &((struct dl_tank *)pelem)->rulecolorfg;
			rulecolorbg	= &((struct dl_tank *)pelem)->rulecolorbg;
			break;
		  case dl_indicator:
			clrmod		= &((struct dl_indicator *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_indicator *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_indicator *)pelem)->clrargs);
			rulechan	= (unsigned int *)&(((struct dl_indicator *)pelem)->rulechan[0]);
			rulecolorfg	= &((struct dl_indicator *)pelem)->rulecolorfg;
			rulecolorbg	= &((struct dl_indicator *)pelem)->rulecolorbg;
			break;
		  case dl_text_update:
			clrmod		= &((struct dl_text_update *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_text_update *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_text_update *)pelem)->clrargs);
			rulechan	= (unsigned int *)&(((struct dl_text_update *)pelem)->rulechan[0]);
			rulecolorfg	= &((struct dl_text_update *)pelem)->rulecolorfg;
			rulecolorbg	= &((struct dl_text_update *)pelem)->rulecolorbg;
			break;
		  }
		if (*clrmod == dl_clrmod_rule && *pDm_llist_el->pdl_strings[clrrule])
			{
			if (init_color_rule_instance(pDm_llist_el, pcnt_list, pRti_cr, clrrule, clrargs, rulechan, rulecolorfg, rulecolorbg)
			  < 0)
				*clrmod = dl_clrmod_static;
			pRti_cr->pclrrule_has_changed = &pDm_llist_el->prti_mon[i].clrrule_has_changed;
#ifdef DEBUG
printf(opiStringTab[STR_m_p0xHxn                    ],pRti_cr->pclrrule_has_changed);
#endif
			pRti_cr++;
			}
		pelem = (struct dl_element_hdr *)((short *)pelem + pelem->nwords);
	}

	/* for each element on the controller display list */
	pelem = (struct dl_element_hdr *)((short *)(pDm_llist_el->pdl_hdr) + pDm_llist_el->pdl_hdr->octrl);
	for (i = 0; i < pDm_llist_el->pdl_hdr->nctrl; i++)
		{
		unsigned char	*clrmod;
		int				clrrule;
		int				clrargs;
		unsigned int	*rulechan;
		void			*rulecolorfg;
		void			*rulecolorbg;

		switch (pelem->type)
		  {
		  case dl_message_button:
			clrmod		= &((struct dl_message_button *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_message_button *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_message_button *)pelem)->clrargs);
			rulechan	= (unsigned int *)&(((struct dl_message_button *)pelem)->rulechan[0]);
			rulecolorfg	= &((struct dl_message_button *)pelem)->rulecolorfg;
			rulecolorbg	= &((struct dl_message_button *)pelem)->rulecolorbg;
			break;
		  case dl_valuator:
			clrmod		= &((struct dl_valuator *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_valuator *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_valuator *)pelem)->clrargs);
			rulechan	= (unsigned int *)&(((struct dl_valuator *)pelem)->rulechan[0]);
			rulecolorfg	= &((struct dl_valuator *)pelem)->rulecolorfg;
			rulecolorbg	= &((struct dl_valuator *)pelem)->rulecolorbg;
			break;
		  case dl_button:
			clrmod		= &((struct dl_button *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_button *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_button *)pelem)->clrargs);
			rulechan	= (unsigned int *)&(((struct dl_button *)pelem)->rulechan[0]);
			rulecolorfg	= &((struct dl_button *)pelem)->rulecolorfg;
			rulecolorbg	= &((struct dl_button *)pelem)->rulecolorbg;
			break;
		  case dl_choice_button:
			clrmod		= &((struct dl_choice_button *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_choice_button *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_choice_button *)pelem)->clrargs);
			rulechan	= (unsigned int *)&(((struct dl_choice_button *)pelem)->rulechan[0]);
			rulecolorfg	= &((struct dl_choice_button *)pelem)->rulecolorfg;
			rulecolorbg	= &((struct dl_choice_button *)pelem)->rulecolorbg;
			break;
		  case dl_menu:
			clrmod		= &((struct dl_menu *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_menu *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_menu *)pelem)->clrargs);
			rulechan	= (unsigned int *)&(((struct dl_menu *)pelem)->rulechan[0]);
			rulecolorfg	= &((struct dl_menu *)pelem)->rulecolorfg;
			rulecolorbg	= &((struct dl_menu *)pelem)->rulecolorbg;
			break;
		  case dl_text_entry:
			clrmod		= &((struct dl_text_entry *)pelem)->clrmod;
				clrrule		= (int)(((struct dl_text_entry *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_text_entry *)pelem)->clrargs);
			rulechan	= (unsigned int *)&(((struct dl_text_entry *)pelem)->rulechan[0]);
			rulecolorfg	= &((struct dl_text_entry *)pelem)->rulecolorfg;
			rulecolorbg	= &((struct dl_text_entry *)pelem)->rulecolorbg;
			break;
		  }
		if (*clrmod == dl_clrmod_rule && *pDm_llist_el->pdl_strings[clrrule])
			{
			if (init_color_rule_instance(pDm_llist_el, pcnt_list, pRti_cr, clrrule, clrargs, rulechan, rulecolorfg, rulecolorbg)
			  < 0)
				*clrmod = dl_clrmod_static;
			pRti_cr->pclrrule_has_changed = &pDm_llist_el->prti_ctrl[i].clrrule_has_changed;
#ifdef DEBUG
printf(opiStringTab[STR_c_p0xHxn                    ],pRti_cr->pclrrule_has_changed);
#endif
			pRti_cr++;
			}
		pelem = (struct dl_element_hdr *)((short *)pelem + pelem->nwords);
		}
#ifdef DEBUG
printf(opiStringTab[STR_Hd_color_rule_instancesn    ],pRti_cr - pRti_color_rule);
#endif
	return (void *)pRti_color_rule;
	}




/*******************************************************************************************
 *
 *    build_rti_dynamic_list
 *    11/90	DK
 *
 *	Build_rti_dynamic_list sets up the run-time structures needed for dynamic graphic
 *		objects and sets up ptrs to both configuration information from the display  
 *		list and the information from relevant channel access connections.  This list 
 *		will be used by the display manager to update the display for the dynamic
 *		graphic objects.
 *	Note :	dynamic graphic objects are those that have color, fill, etc. modified in a
 *		discrete fashion based on the value of a process variable.
 *	Input 	: cnt		number of dynamic graphic objs defined for the display
 *		  pdm_llist_el	ptr to the display list information for the display
 *		  pcnt_list	ptr to list of connection counts for each process vaiable on
 *			the display to use for setting up set modified bits
 *	Output	: none
 *	Returns	: ptr to run-time list of dynamic graphic information
 */
void *build_rti_dynamic_list(cnt, pdm_llist_el, pcnt_list)
unsigned short		cnt;
struct dm_llist_el	*pdm_llist_el;
int			*pcnt_list;	
	{
	unsigned short				i;
	struct rti_dynamic_graphic	*pRti_el;
	unsigned short			bit_ix;
	struct dl_dynamic_attribute	*pdyn = (struct dl_dynamic_attribute *)(pdm_llist_el->pdl_dynamic);
	struct rti_dynamic_graphic	*pRti_dynamic_graphics	= (struct rti_dynamic_graphic *)tisCalloc(__FILE__, __LINE__, cnt,sizeof(struct rti_dynamic_graphic));

	/* for each element on the display list */
	/* set up comm_list and indexes for set modified bits */
	for (pRti_el = pRti_dynamic_graphics, i = 0; i < cnt;
	  pdyn = (struct dl_dynamic_attribute *)((short *)pdyn + ((struct dl_element_hdr *)pdyn)->nwords))
		{
		unsigned short	last_clr;
		unsigned short	last_width;
		unsigned short	last_style;
		unsigned short	last_fill;
		unsigned short	last_clrmod;
		unsigned short	last_vis;
		unsigned int	last_clrrule;
		unsigned int	last_chan;

		/* get indexes into set modified table */
		/* set ptrs to chids- save a few levels of indirection later */
		switch (pdyn->element.type)
		  {
	      case dl_basic_attribute:
			last_clr		= ((struct dl_basic_attribute *)pdyn)->attr.clr;
			last_width		= ((struct dl_basic_attribute *)pdyn)->attr.width;
			last_style		= ((struct dl_basic_attribute *)pdyn)->attr.style;
			last_fill		= ((struct dl_basic_attribute *)pdyn)->attr.fill;
			break;
	      case dl_dynamic_attribute:
			last_clrmod		= ((struct dl_dynamic_attribute *)pdyn)->attr.mod.clrmod;

			last_vis		= ((struct dl_dynamic_attribute *)pdyn)->attr.mod.vis;
			last_clrrule	= (int)((struct dl_dynamic_attribute *)pdyn)->attr.param.clrrule;
			last_chan		= (int)pdyn->attr.param.chan;	    /* save ix for process var from attr */
			break;
	      case dl_rec:
	      case dl_line:
	      case dl_oval:
	      case dl_arc:
	      case dl_text:
			pRti_el->clr		= (char)last_clr;
			pRti_el->width		= (unsigned char)last_width;
			pRti_el->style		= (unsigned char)last_style;
			pRti_el->fill		= (unsigned char)last_fill;
			pRti_el->clrmod		= (unsigned char)last_clrmod;
			pRti_el->vis		= (unsigned char)last_vis;
			pRti_el->clrrule	= last_clrrule;
			pRti_el->pComel		= &(pdm_llist_el->pcom_el->pCom_table[last_chan]);
			pRti_el->mod_intix	= (bit_ix = (pcnt_list[last_chan])++) / BITS_PER_INT;
			pRti_el->mod_bitix	=  bit_ix - (pRti_el->mod_intix*BITS_PER_INT);
			i++;		/* increment count only for graphic els */
			pRti_el++;
			break;
		  default:
			printf(opiStringTab[STR_in_build_dynamic_graphicsbad]);
		  }
		}
	return (void *)pRti_dynamic_graphics;
	}


/*******************************************************************************************
 *
 *    build_rti_mon_list
 *    10/90	DK
 *
 *	Build_rti_mon_list sets up the run-time structures needed for monitor-type objects
 *		and sets up ptrs to both configuration information from the display list 
 *		and the information from relevant channel access connections.  This list 
 *		will be used by the display manager to update the display for the monitors.
 *	Input 	: cnt		number of monitors defined for the display
 *		  pdm_llist_el	ptr to the display list information for the display
 *		  pcnt_list 	ptr to list of connection counts for each process vaiable on
 *			the display to use for setting up set modified bits
 *	Output	: none
 *	Returns	: ptr to run-time list of monitor information
 */
void *build_rti_mon_list(cnt, pdm_llist_el, pcnt_list)
unsigned short		cnt;			/* number els on dynamic part of display list */
struct dm_llist_el	*pdm_llist_el;	/* pointer to dynamic part of display list */
int					*pcnt_list;
	{
	/* cnt_list to assign set_modified index to each ref to channel */
	unsigned short			i, j;
	struct rti_update_mon	*prti_el;
	struct dl_element_hdr	*pdyn		= (struct dl_element_hdr *)(pdm_llist_el->pdl_mon);
	struct dl_monitor_hdr	*pdl_mon;
	unsigned short			bit_ix;
	struct dl_pc			*pDl_pc;
	struct rti_update_mon	*prti_mons	= (struct rti_update_mon *)tisCalloc(__FILE__, __LINE__, cnt,sizeof(struct rti_update_mon));

	/* set up comm_list and indexes for set modified bits */
	for (prti_el = prti_mons,i = 0; i < cnt; i++,prti_el++,
	  pdyn = (struct dl_element_hdr *)((short *)pdyn + ((struct dl_element_hdr *)pdyn)->nwords))
		{
		/* set ptr to dl_dynamic part */
		prti_el->pdl_mon  = pdyn;
		pdl_mon = &(((struct dl_anymo *)prti_el->pdl_mon)->monitor);
		pDl_pc = (struct dl_pc *)(((short *)pdm_llist_el->pdl_hdr) + pdm_llist_el->pdl_hdr->opc);
		for (j = 0; j < (int)(pdl_mon->rdbk); j++)
			pDl_pc = (struct dl_pc *)((short *)pDl_pc + ((struct dl_element_hdr *)pDl_pc)->nwords);

		/* get indexes into set modified table */
		/* set ptrs to chids- save a few levels of indirection later */
		switch (pdyn->type)
		  {
		  case dl_bar:
		  case dl_meter:
		  case dl_tank:
		  case dl_indicator:
		  case dl_text_update:
			prti_el->g_type.rectangle.hdr.chan_name			= (char **)(pDl_pc + 1);
			prti_el->g_type.rectangle.hdr.foregroundcolor	= COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pdl_mon->clr);
			prti_el->pgeom_comel = &(pdm_llist_el->pcom_el->pCom_table[(int)(pdl_mon->rdbk)]);
			bit_ix = (pcnt_list[(int)(pdl_mon->rdbk)])++;
			prti_el->geom_intix = bit_ix / BITS_PER_INT;
			prti_el->geom_modix = bit_ix - (prti_el->geom_intix*BITS_PER_INT);
			break;
		  default:
			printf(opiStringTab[STR_in_build_monbad_case_n      ]);
		  }
		}
	return (void *)prti_mons;
	}


/*******************************************************************************************
 *
 *    build_rti_ctrl_list
 *    10/90	DK
 *
 *	Build_rti_ctrl_list sets up the run-time structures needed for controllers(gizmos)
 *		and sets up ptrs to both configuration information from the display list 
 *		and the information from relevant channel access connections.  This list 
 *		will be used by the display manager to update the display for the controllers.
 *	Input 	: cnt		number of controllers defined for the display
 *		  pdm_llist_el	ptr to the display list information for the display
 *		  pcnt_list	ptr to list of connection counts for each process vaiable on
 *			the display to use for setting up set modified bits
 *	Output	: none
 *	Returns	: ptr to run-time list of controller information
 */
void *build_rti_ctrl_list(cnt, pdm_llist_el, pcnt_list)
unsigned			short cnt;		/* number ctrl els on display list */
struct dm_llist_el	*pdm_llist_el;	/* pointer to display list info */
int					*pcnt_list;
	{
	/* cnt_list to assign set_modified index to each ref to channel */
	unsigned short			i;
	struct rti_ctrl_info	*prti_el;
	struct dl_element_hdr	*pdyn		= (struct dl_element_hdr *)(pdm_llist_el->pdl_ctrl);
	struct dl_control_hdr	*pdl_ctrl_hdr;
	unsigned short			bit_ix;
	struct rti_ctrl_info	*prti_ctrls	= (struct rti_ctrl_info *)tisCalloc(__FILE__, __LINE__, cnt,sizeof(struct rti_ctrl_info));

	/* set up comm_list and indexes for set modified bits */
	for (prti_el = prti_ctrls,i = 0; i < cnt; i++,prti_el++,
	  pdyn = (struct dl_element_hdr *)((short *)pdyn + ((struct dl_element_hdr *)pdyn)->nwords))
		{
		/* set ptr to dl part */
		prti_el->pdl_ctrl  = pdyn;
		/* get indexes into set modified table */
		/* set ptrs to chids- save a few levels of indirection later */
		switch (pdyn->type)
		  {
		  case dl_message_button:
			if ((prti_el->pExt = (void *)tisCalloc(__FILE__, __LINE__, 3, sizeof (char *))) == NULL)
				printf(opiStringTab[STR_Couldnt_calloc_in_build_rti_]);
			else
				{
				((char **)prti_el->pExt)[0] = pdm_llist_el->pdl_strings[(int)(((struct dl_message_button *)pdyn)->press_msg)];
				((char **)prti_el->pExt)[1] = pdm_llist_el->pdl_strings[(int)(((struct dl_message_button *)pdyn)->release_msg)];
				if (strlen(((char **)prti_el->pExt)[0]) > 0)
					prti_el->user = (strlen(((char **)prti_el->pExt)[1]) > 0) ? PRESS_RELEASE : PRESS;
				else
					prti_el->user = PRESS_RELEASE_DB;
				}
				/* fall thru to complete init */
		  case dl_valuator:
		  case dl_button:
		  case dl_choice_button:
		  case dl_menu:
		  case dl_text_entry:
			pdl_ctrl_hdr = &(((struct dl_anyco *)prti_el->pdl_ctrl)->control);
			prti_el->pgeom_comel = &(pdm_llist_el->pcom_el->pCom_table[(int)(pdl_ctrl_hdr->ctrl)]);
			bit_ix = (pcnt_list[(int)(pdl_ctrl_hdr->ctrl)])++;
			prti_el->geom_intix = bit_ix / BITS_PER_INT;
			prti_el->geom_modix = bit_ix - (prti_el->geom_intix*BITS_PER_INT);
			break;
		  default:
			printf(opiStringTab[STR_in_build_ctrlbad_case_n     ]);
		  }
		}
	return (void *)prti_ctrls;
	}


/*******************************************************************************************
 *
 *    build_rti_plot_list
 *    11/90	DK
 *
 *	Build_rti_plot_list sets up the run-time structures needed for plots and sets
 *		up ptrs to both configuration information from the display list and the
 *		information from relevant channel access connections.  Note: ca 
 *		connections are done using the synchronous data routines--not the ones
 *		used by the rest of the dynamic elements.
 *		This list will be used by the display manager to update the display.
 *	Input 	: cnt		number of plots defined for the display
 *		  pdm_llist_el	ptr to the display list information for plots
 *	Output	: none
 *	Returns	: ptr to run-time list of plot information
 */
void *build_rti_plot_list(cnt, pdm_llist_el)
unsigned short		cnt;			/* number els on dynamic part of display list */
struct dm_llist_el	*pdm_llist_el;	/* pointer to dynamic part of display list */
	{
	unsigned short			i;
	struct rti_update_plot	*prti_el;
	struct dl_element_hdr	*pdyn		= (struct dl_element_hdr *)(pdm_llist_el->pdl_plot);
	struct rti_update_plot	*prti_plots	= (struct rti_update_plot *)tisCalloc(__FILE__, __LINE__, cnt, sizeof(struct rti_update_plot));

	/* set up comm_list and indexes for set modified bits */
	for (prti_el = prti_plots,i = 0; i < cnt;i++,prti_el++,
	  pdyn = (struct dl_element_hdr *)((short *)pdyn + ((struct dl_element_hdr *)pdyn)->nwords))
	  	{
		/* set ptr to dl_plot(strip, etc.) struct */
		prti_el->pDl_plot = pdyn;
		/* get indexes into set modified table */
		/* set ptrs to chids- save a few levels of indirection later */
		switch (pdyn->type)
		  {
		  case dl_plot:				setup_xyplot(pdm_llist_el, &(pdm_llist_el->dl_context),prti_el, &pfdctx);	break;
		  case dl_smith_overlay:	setup_smith(pdm_llist_el, &(pdm_llist_el->dl_context),prti_el, &pfdctx);	break;
		  case dl_surface_plot:		printf(opiStringTab[STR_surface_not_implemented_in_b]);								break;
		  case dl_strip:			setup_strip(pdm_llist_el,&(pdm_llist_el->dl_context), prti_el);				break;
		  default:					printf(opiStringTab[STR_in_build_monbad_case_n      ]);										break;
		  }
		}
	return (void *)prti_plots;
	}

void *string_index_to_display_str(pstring_table, textix, pdis_str, pstr_len)
char			**pstring_table;
unsigned short	textix;
char			**pdis_str;
unsigned		*pstr_len;
	{
    *pdis_str = pstring_table[textix];
    *pstr_len = strlen(*pdis_str);
    return 0;
	}
    


/*******************************************************************************************
 *
 *	x_evhandler
 *
 *	8/90	DK
 *	10/90	DK	added filter_gizmo_events to pass on processing to gizmo code
 *	11/90	DK	modified to include strip chart Xevent handling
 *	12/90	DK	added context management to allow processing of events from 
 *			multiple windows, esp. plot windows
 *	1/91	DK	modified to handle deletion of displays
 *	3/91	DK	modified to handle cartesian and smith plots
 *	4/91	DK	added ClientMessage event check for deleting top-level windows
 *				using window manager
 *	4/91	DK	added UmapNofify, MapNotify, Visibility Notify for avoiding
 *				graphic updates when window is iconized or totally obscured
 *
 *	x_evhandler processes all Xevents on the queue.  This includes the controller
 *	events(valuator, text_updates, menus, etc.) that are handled by filter gizmo
 *	events and the events registered for strip charts and plots that are child windows.
 *	Expose events will cause all display elements within the exposed area of the
 *	drawable(window) to be updated.
 *	Note : the name x_evhandler may be misleading --
 *	x_evhandler is called during each pass of the display manager's update loop and
 *	thus is really synchronous as opposed to event driven.
 *
 *	Input	: pdm_llist_hdr-- ptr to display manager's display header defined and 
 *		accessed for each rti display window
 *
 *	Output	: graphical updates to display window according to events in Xevent queue
 *	
 *	Returns	: nothing
 */
void x_evhandler(pdm_llist_hdr)
struct dm_llist_hdr *pdm_llist_hdr;
	{
	XEvent			the_event;		/* X event struct */
	caddr_t			pEvwin;
	struct dm_llist_el	*pDm_llist_el;

	/* The Xevent loop.
	 * Note that the first event will be an exposure event which will cause the first graphics operation. Exit on a button up.
	 */
	/* process as long as events on queue-- must check XPending since XNext blocks */
	while (XPending(pdm_llist_hdr->pNew_llist->dl_context.d))
		{
		XNextEvent(pdm_llist_hdr->pNew_llist->dl_context.d,&the_event);
		if (filter_gizmo_events(&the_event))
			/* if event for gizmo, pass it on for processing by gizmo code */
			continue;

		if (XFindContext(the_event.xany.display,the_event.xany.window, pdm_llist_hdr->pNew_llist->win_context, &pEvwin) != 0)
			/* if not part of display man's context, do nothing */
			continue;

		if (((struct event_on_window *)pEvwin)->plot_win)
			{
			/* if not main display or gizmo, plot window */
			struct rti_update_plot *pRti_plot = (struct rti_update_plot *)((struct event_on_window *)pEvwin)->puser;
			if (XFindContext(the_event.xany.display, pRti_plot->window, pdm_llist_hdr->pNew_llist->win_context, &pEvwin) != 0)
				{
				printf(opiStringTab[STR_in_xev_no_context_for_this_p]);
				continue;
				}
			proc_plot(pRti_plot, the_event);
			continue; /* don't do rest of updates for main display window */
			}
		pDm_llist_el = ((struct dm_llist_el *)((struct event_on_window *)pEvwin)->puser);

		/* handle events for display */
		
		switch (the_event.type)
		  {
		  case ClientMessage:
		  	{
			/* intercept kill client from Window Manager */
			XClientMessageEvent	*pCl	= (XClientMessageEvent *) &the_event;
			/* look for wm_delete(proto[0]) */
			if ((unsigned)(pCl->data.l[0])== proto[0])
				{
				/* get rid of any associated plot windows, etc. */
				if (pDm_llist_el->prti_plot)
					{
					delete_plots(pDm_llist_el->dl_context.d, pdm_llist_hdr->pNew_llist->win_context,
							pDm_llist_el->prti_plot, pDm_llist_el->pdl_hdr->nplot);
					}
				/* keep further events for this element from being processed after window deletion */
				XDeleteContext(pDm_llist_el->dl_context.d, pDm_llist_el->dl_context.w,
							pdm_llist_hdr->pNew_llist->win_context);
				/* delete from dm_llist and free associated  memory */
				delete_display(pdm_llist_hdr, pDm_llist_el);
				}
			}
			break;
		  case UnmapNotify:
			pDm_llist_el->visible = FALSE;
			break;
		  case MapNotify:
			pDm_llist_el->visible = TRUE;
			break;
		  case VisibilityNotify:
			pDm_llist_el->visible = TRUE;
			break;
		  case ButtonRelease:
			break;
		  case ButtonPress:
			if (the_event.xbutton.button == Button3)
			{
				/* the following line prevents any sub-menus for auto-exit displays */
				if (pDm_llist_el->pdl_hdr->exitaction !=  2 /* auto-exit */ )
				    do_dmMenu(&the_event, pDm_llist_el);
			}
			break;
		  case Expose:	     				/* On expose events, draw elements from display list */
			{
			XRectangle	rec;			/* x, y, ht, wd struct */
			unsigned	dl_size;		/* size of display list */
			
			rec.x		= the_event.xexpose.x;
			rec.y		= the_event.xexpose.y;
			rec.width	= the_event.xexpose.width;
			rec.height	= the_event.xexpose.height;

			dl_size		= pDm_llist_el->pdl_hdr->nwords_sta;

			draw_static(&(pDm_llist_el->static_dlcontext), pDm_llist_el->pstatic, dl_size, &rec, FALSE);

			update_color_rules(pDm_llist_el);

			update_dynamic_graphics(&(pDm_llist_el->dl_context), &(pDm_llist_el->static_dlcontext),
					pDm_llist_el->pdl_color_map_index, pDm_llist_el->pdl_dynamic,
					pDm_llist_el->prti_dynamic, pDm_llist_el->pdl_hdr->ndyn,
					&rec, pDm_llist_el, TRUE, pDm_llist_el->pdl_hdr->bclr);
			  
			update_mon_objs(&(pDm_llist_el->dl_context), pDm_llist_el->pdl_color_map_index,
					pDm_llist_el->pdl_mon, pDm_llist_el->prti_mon,
					pDm_llist_el->pdl_hdr->nmon, &rec, pDm_llist_el, TRUE);
			  
			update_controllers(&(pDm_llist_el->dl_context), pDm_llist_el->pdl_color_map_index,
					pDm_llist_el->pdl_ctrl, pDm_llist_el->prti_ctrl,
					pDm_llist_el->pdl_hdr->nctrl, &rec, pDm_llist_el, TRUE);
			break;
			}
		  case LeaveNotify:
			{
			if (((the_event.xcrossing.detail == NotifyNonlinear)
			    || (the_event.xcrossing.detail == NotifyNonlinearVirtual)
			    || (the_event.xcrossing.detail == NotifyAncestor))
			  && (the_event.xcrossing.subwindow == 0))
				{
				/* kill the display */
				dmExit_display(pdm_llist_hdr, pDm_llist_el, NULL); /* <- note the NULL! */
				}
			break;
			}
		  default:
			break;	/* Ignore everything else. */
		  }
		}
	}


#define NPTS			50
#define LO				0.0
#define HI				10.0
#define FREQ			1.0
#undef	SYNC_TIMEOUT
#define SYNC_TIMEOUT	.1
#define STRINGSIZE		60
#define SECSPERUSEC		.000001
#define SECSPERMSEC		.001
#define SECSPERSEC		1
#define SECSPERMIN		60
#define SECSPERHR		(SECSPERMIN * 60)
#define SECSPERDAY		(24 * SECSPERHR)

/*******************************************************************************************
 *
 *	setup_strip
 *
 *	11/90	DK
 *
 *	Setup_strip initializes both the rti_update_plot struct for the strip
 *		chart and the Strip struct (for the strip routines to use info
 *		from the display list).  It initiates a synchronous read for
 *		the set of process variables.  The inital draw of the strip
 *		chart is done here.
 *
 *	Note:  we're coming in the "back door" here using the Strip struct to
 *		set our colors and fonts after flagging that we'll be doing
 *		so (in the setup for the  stripInit call).  
 *
 *	Input	: pdm_llist_el-- ptr to display manager's llist el for this display 
 *		: pdlc -- ptr to the dl_context struct for this display
 *		: pPlot_el-- ptr to the run-time struct for this strip chart
 *	Output	: none
 *	
 *	Returns	: nothing
 *
 */
int setup_strip(pdm_llist_el,pdlc, pPlot_el)
struct dm_llist_el *pdm_llist_el;
struct dl_context *pdlc;
struct rti_update_plot *pPlot_el;
	{
	unsigned short			index, i,j;
	double					getStripDouble();
	unsigned				ix;
	struct dl_pc			*pDl_pc;
	char					titleFont[STRINGSIZE], axesFont[STRINGSIZE];
	struct dl_object_hdr	*pRec	= &(((struct dl_anypo *)pPlot_el->pDl_plot)->object);
	unsigned long			color[7];
	unsigned short			cnt;	/* num of traces for this strip chart */
	struct dl_strip			*pDl_strip;
	static double			secsperunit[] =
		{
		SECSPERDAY,
		SECSPERHR,
		SECSPERMIN,
		SECSPERSEC,
		SECSPERMSEC,
		SECSPERUSEC
 		};
	int				saveSize;	/* number of elements to maintain for draw */

	pDl_strip = ((struct dl_strip *)pPlot_el->pDl_plot);
	index = (int)(pDl_strip->plotcom.title);
	if (pDl_strip->delay == 0)
	    {
	    pDl_strip->delay = 1;	/* set 0 delay between updates to 1 second! */
	    pDl_strip->units = 3;
	    }

	/* create a window for the strip chart with info from display list (x,y,ht,wd,clr,bclr)*/
	/* get plot coords (x,y,h,w ) from display list*/
	pPlot_el->window= XCreateSimpleWindow(pdlc->d, pdlc->w, pRec->x,pRec->y, pRec->width,pRec->height, BORDER_WIDTH,
	  WhitePixel(pdlc->d,the_screen), BlackPixel(pdlc->d,the_screen));

	/* Select exposure events and button press events. */
	XSelectInput(pdlc->d, pPlot_el->window, ExposureMask|ButtonPressMask|ButtonReleaseMask|Button3MotionMask);

	/* Tell the server to make the window visible. */
	XMapWindow(pdlc->d,pPlot_el->window);

	/* set up for handling events on window id  */
	pPlot_el->evwin.plot_win = TRUE;
	pPlot_el->evwin.puser = pPlot_el;
	XSaveContext(pdlc->d,pPlot_el->window, pdm_llist_el->win_context, (XPointer)&(pPlot_el->evwin));

	/* init the strip chart */
	pPlot_el->pPlot = (caddr_t)stripInit(pdlc->d, the_screen, pPlot_el->window);
	if (pPlot_el->pPlot == NULL)
		{
		printf(opiStringTab[STR_Couldnt_init_strip_chartn   ]);
		return 0;
		}

	for (i = 0, cnt = 0; i < 7; i++)
		{
		/* get process variable names for ca (sync read) */
		ix = (int)(pDl_strip->pen[i].chan);
		pDl_pc = (struct dl_pc *)(((short *)pdm_llist_el->pdl_hdr) + pdm_llist_el->pdl_hdr->opc);
		for (j = 0; j < ix;j++)
			{
			pDl_pc = (struct dl_pc *)((short *)pDl_pc + ((struct dl_element_hdr *)pDl_pc)->nwords);
			}
		if (strlen((char *)(pDl_pc + 1)) > 0)
			{
			/* set up data fetch/conversion */
#ifdef OLD_STRIP
			pPlot_el->getPV_vals[cnt] = getStripDouble;
			pPlot_el->pPlot_data[cnt] = (double *)tisCalloc(__FILE__, __LINE__, 1, (unsigned)NPTS *sizeof(double));
#else
			char  name_str[36];
			strncpy (name_str, (char *)(pDl_pc + 1), sizeof(name_str));
			if(pdm_llist_el->ppMacro_pairs != NULL){
		    	    gsdl_main_macro(name_str, (char *)(pDl_pc + 1), pdm_llist_el->ppMacro_pairs, 36);
			}
			pPlot_el->getPV_vals[cnt] = NULL;
			saveSize = ((struct _Strip *)(pPlot_el->pPlot))->dataWidth;
			pPlot_el->pPlot_data[cnt] = (double *)tisCalloc(__FILE__, __LINE__,
									1, (unsigned)saveSize *sizeof(double));
			/* expand macro names! */
			*(strip_channel **)&(pPlot_el->getPV_vals[cnt]) = OpenStripChannel (name_str,
						 saveSize,
						 (int)(secsperunit[pDl_strip->units] * (double)pDl_strip->delay),
						 pPlot_el->pPlot_data[cnt]);
#endif OLD_STRIP
			pPlot_el->pYlim[cnt] = &(pdm_llist_el->pcom_el->pCom_table[ix]);
			pPlot_el->stripRange[cnt].minVal = LO;
			pPlot_el->stripRange[cnt].maxVal = HI;
			pPlot_el->ysync_data[cnt].pName	 = (char *)(pDl_pc + 1);
			color[cnt++] = COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pDl_strip->pen[i].clr);
			}
		}

	pPlot_el->num_traces = cnt;
	((Strip *)pPlot_el->pPlot)->setOwnColors = TRUE;	/* setup to select colors from the display list by index */

	/* get and set proper fonts */
	sprintf(titleFont,"-*-*-bold-r-*-*-%d-*-*-*-m-*-iso8859-1", GraphX_TitleFontSize(((Strip *)pPlot_el->pPlot)));
	sprintf(axesFont,"-*-*-medium-r-*-*-%d-*-*-*-m-*-iso8859-1", GraphX_AxesFontSize(((Strip *)pPlot_el->pPlot)));

	/* set up for file descriptor management */
	((Strip *)pPlot_el->pPlot)->pfdctx = (caddr_t)pfdctx;
	/* setup Strip struct */
	stripSet(((Strip *)pPlot_el->pPlot),
	  cnt,
	  saveSize,
	  pPlot_el->pPlot_data,
	  pPlot_el->stripRange,
	  secsperunit[pDl_strip->units] * (double)pDl_strip->delay,
	  pPlot_el->getPV_vals,
	  pdm_llist_el->pdl_strings[index], 
	  titleFont,
	  pdm_llist_el->pdl_strings[(int)(pDl_strip->plotcom.xlabel)], 
	  pdm_llist_el->pdl_strings[(int)(pDl_strip->plotcom.ylabel)],
	  axesFont,
	  NULL, NULL,
	  NULL,
	  StripExternal);

	/* fill in color indexes for strip plot */
	((Strip *)pPlot_el->pPlot)->forePixel = COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pDl_strip->plotcom.clr);
	((Strip *)pPlot_el->pPlot)->backPixel = COLOR_LOOKUP(pdm_llist_el->pdl_color_map_index, pDl_strip->plotcom.bclr);
	for (i = 0; i < cnt; i++)
		{
		((Strip *)pPlot_el->pPlot)->dataPixel[i] = color[i];
		}
	/* set up user data ptr for indexing proper strip chart when data fetch/conversion routines are called */
	((Strip *)pPlot_el->pPlot)->userData = (caddr_t)pPlot_el;

#ifdef OLD_STRIP
	/* set up channel access for synchronous reads, need a ctrl ptr for each strip chart */
	pPlot_el->pSync_ctrl = gsd_sync_init(pPlot_el->ysync_data, pPlot_el->num_traces, &pfdctx, pdm_llist_el->ppMacro_pairs);
	gsd_sync_read(NEXTSET_SYNC_DATA, pPlot_el->pSync_ctrl, pPlot_el->num_traces, &pfdctx, SYNC_TIMEOUT);
#endif OLD_STRIP
	return 0;
	}


/*******************************************************************************************
 *
 *	getStripDouble
 *
 *	11/90	DK
 *
 *	getStripDouble fetches and converts the ca data for the appropriate
 *		strip chart and trace number (0-6).  If a channel is not
 *		connected it will be shown as end of range.  If it has
 *		disconnected, it will remain the last known val.  This 
 *		probably should be changed later.
 *
 *
 *	Input	: index-- trace index (0-6) of this pv 
 *		: userarg-- ptr to rti_update_plot struct for this strip chart
 *	Output	: none
 *	
 *	Returns	: double representation of the value of the pv specified for this trace
 *
 */
double getStripDouble(index, userarg)
int		index;
caddr_t	*userarg;
	{
    struct rti_update_plot	*pStrip_el	= (struct rti_update_plot *)userarg;

    /* get data from ca if we have some--should check TYPE_NOT_CONN, may want to indicate not conn with white ?? */

    if (((pStrip_el->init_bits & (1 << index)) == (1 << index)))
		{	/* !TYPENOTCONN */
		if (dbr_type_is_FLOAT(pStrip_el->ysync_data[index].time_type))
			return (double)(((struct dbr_time_float*)pStrip_el->ysync_data[index].pSdata)->value);

		else if (dbr_type_is_DOUBLE(pStrip_el->ysync_data[index].time_type))
			return ((struct dbr_time_double *)pStrip_el->ysync_data[index].pSdata)->value;

		else if (dbr_type_is_SHORT(pStrip_el->ysync_data[index].time_type))
		    return (double)(((struct dbr_time_short *)pStrip_el->ysync_data[index].pSdata)->value);

		else if (dbr_type_is_CHAR(pStrip_el->ysync_data[index].time_type))
		    return (double)(((struct dbr_time_char *)pStrip_el->ysync_data[index].pSdata)->value);

		else if (dbr_type_is_ENUM(pStrip_el->ysync_data[index].time_type))
			return (double)(((struct dbr_time_enum *)pStrip_el->ysync_data[index].pSdata)->value);

		else if (dbr_type_is_LONG(pStrip_el->ysync_data[index].time_type))
			return (double)(((struct dbr_time_long *)pStrip_el->ysync_data[index].pSdata)->value);
		}
    return pStrip_el->stripRange[index].minVal - (double)(1.0);		/* drop out of range -- pen up might be better */
	}	


/********************************************************************************************
 *
 *	init_display_mgr
 *
 *		set up llist hdrs, initing them
 *		init channel access 
 *		set up fdmgr to allow processing of "select" events by X, ca, etc.
 *		set up font table
 *		init context for processing X events based on window id
 *
 */
int init_display_mgr(pdm_llist_hdr, pcom_llist_controller, argc, argv)
struct dm_llist_hdr			*pdm_llist_hdr;
struct gsdl_com_llist_controller	*pcom_llist_controller;
int					argc;
char					*argv[];
	{		
	int		x_fd;
	int		stat;
	void		dm_CaException();
	extern char	*optarg;
	extern int	optind;
	char		c;

	/* initialize com_llist_controller and display manager llist */
	pcom_llist_controller->new_llist_flag = 1;
	pcom_llist_controller->pNew_llist = (struct gsdl_com_llist_element *)tisCalloc(__FILE__, __LINE__, 1, sizeof(struct gsdl_com_llist_element));

	/* initialize display_man_llist */
	pdm_llist_hdr->new_llist_flag = 1;
	pdm_llist_hdr->pNew_llist = (struct dm_llist_el *)tisCalloc(__FILE__, __LINE__, 1, sizeof(struct dm_llist_el));
	pdm_llist_hdr->pNew_llist->pNext_el	= NULL;
	pdm_llist_hdr->pNew_llist->pcom_el	= pcom_llist_controller->pNew_llist;

	/* set up file descriptor mgr and initialize channel access */

	if (ca_task_initialize() != ECA_NORMAL) {
#ifdef WIN32
	    lprintf ("Channel Access is not running!\n");
#else
	    printf ("Channel Access is not running!\n");
#endif WIN32
	    exit (-1);
	}

	/* add exception handler to avoid aborts from CA server */
	/* may want to pass more info--do we need multiple handlers ??? */
	stat = ca_add_exception_event(dm_CaException, NULL);
	if (stat != ECA_NORMAL)
		printf(opiStringTab[STR_Bad_stat_from_ca_add_excepti]);

	/* set up file descriptor manager(fdmgr) for single context X,CA selects */  
	pfdctx = (void *) fdmgr_init();
	if (!pfdctx)
		{
		printf(opiStringTab[STR_ca_task_initialize_fdmgr_ini]);
		abort();
		}
	/* get CA file descriptor for fdmgr */
	SEVCHK(ca_add_fd_registration(gsdl_fd_register, pfdctx), opiStringTab[STR_gsdl_macro_ca_task_initializ]);
	fdmgr_add_timeout(pfdctx,&dm_capoll, dm_capollfunc,pfdctx);

	/* open display, get X file descriptor for fdmgr */
	if ((pdm_llist_hdr->pNew_llist->dl_context.d = XOpenDisplay(NULL)) == NULL)
		{
		printf(opiStringTab[STR_in_Hs___cant_open_displayn  ],argv[0]);
		abort();
		}
#if XSYNC
	XSetErrorHandler(opi_x_error_handler);
	XSynchronize(pdm_llist_hdr->pNew_llist->dl_context.d, (Bool) TRUE);
	printf(opiStringTab[STR_in_dm_XSynch_is_onn         ]);
#endif
	/* get context for window event management */
	if((pdm_llist_hdr->pNew_llist->win_context = XUniqueContext()) == 0)
		{
		printf(opiStringTab[STR_in_display_mgr_couldnt_get_c]);
		abort();
		}

	/* get up file descriptor mgr so can use both ca and X -- may need changes here if more than one display */
	x_fd = ConnectionNumber(pdm_llist_hdr->pNew_llist->dl_context.d);
	fdmgr_add_fd(pfdctx,x_fd,x_evhandler,pdm_llist_hdr);

	/* do some command line parsing */
	while ((c = getopt(argc, argv, "pt:u:")) != -1)
		switch (c)
		  {
		  case 'p':	PrivCheckOverride = TRUE;	break;
		  case 't':	TestTimer = atoi (optarg);	break;
		  case 'u':	update_rate = atoi (optarg);	break;
		  case '?':	Usage();			break;
		  }
	argv += optind;
	if (optind++ < argc)
		strcpy(pdm_llist_hdr->pNew_llist->display_name, *argv++);	/* get that display name */
	else
		Usage();

	if (optind++ < argc)
		{
		pdm_llist_hdr->pNew_llist->ppMacro_pairs = (char **)tisCalloc(__FILE__, __LINE__, MAX_PAIR_STRS, sizeof(char *));	/* get those macro args */
		get_macro_pairs(*argv++, pdm_llist_hdr->pNew_llist->ppMacro_pairs, MAX_PAIR_STRS, DefaultScreenOfDisplay(pdm_llist_hdr->pNew_llist->dl_context.d));
		}
	else
		pdm_llist_hdr->pNew_llist->ppMacro_pairs = NULL;

	/* load fonts, set up font table if not done yet */
	pdm_llist_hdr->pNew_llist->dl_context.pfont_table = init_font_table(pdm_llist_hdr->pNew_llist->dl_context.d);
	if (pdm_llist_hdr->pNew_llist->dl_context.pfont_table == NULL)
		{
		printf(opiStringTab[STR_Cant_load_fontsabortingn    ]);
		abort();
		}

	/* find it if color or bw */
	pdm_llist_hdr->pNew_llist->dl_context.color = color_visual_test(pdm_llist_hdr->pNew_llist->dl_context.d,the_screen);
	fdmgr_add_timeout(pfdctx,&dm_blinkpoll, dm_blinkpollfunc,pdm_llist_hdr);

	/* set up protocol to avoid kill client on delete of individual windows */
	proto[0] = XInternAtom(pdm_llist_hdr->pNew_llist->dl_context.d, "WM_DELETE_WINDOW",False);
	if (proto[0] == None)
		abort();
	proto[1] = XInternAtom(pdm_llist_hdr->pNew_llist->dl_context.d, "WM_SAVE_YOURSELF",False);
	if (proto[1] == None)
		abort();

	init_dmMenu(pdm_llist_hdr->pNew_llist->dl_context.d, pdm_llist_hdr);
	dmInitPlotMenu(pdm_llist_hdr->pNew_llist->dl_context.d, pdm_llist_hdr);
	dmInitStripMenu(pdm_llist_hdr->pNew_llist->dl_context.d, NULL);
	return 0;
	}

int Usage()
	{
    /*
     * it's really: "Usage : dm [-p] display_name [macro_replacement_string]\n"
     *
     * but, let's keep -p our little secret! -t is also a secret!
     */
    printf(opiStringTab[STR_Usage__dm_display_name_macro]);
    exit(1);
    return 0;
	}


/***************************************************************************************
 *
 *	dm_CaException
 *
 *	DK 1/91
 *
 *	Purpose	:  handle exceptions coming from CA server to avoid abort on errors
 *		from gets, puts, etc.
  *
 *	Input	:  arg		struct from CA containing info about error
 *	
 *	Output	: nothing
 *
 *	Returns	: nothing
 *
 */
void dm_CaException(arg)
struct exception_handler_args	arg;
	{
    chid	pCh;

    pCh = arg.chid;
    if (pCh != NULL)
		printf(opiStringTab[STR_error_from_CA_on_Hsn        ], ca_name(pCh));
    printf("%s\n", ca_message(arg.stat));
	}


int delete_display(pDm_llist_hdr,pDm_llist_el)
struct dm_llist_hdr  	*pDm_llist_hdr;
struct dm_llist_el		*pDm_llist_el;
	{
	struct dm_llist_el	*pDm_llist_tmp;
	int			i;

	/* need to lock to avoid reentrancy probs for vx */
	pDm_llist_tmp = pDm_llist_hdr->pNew_llist;
	if (pDm_llist_tmp == pDm_llist_el)
		NormalExit(1);	/* first display on list */

	while (pDm_llist_tmp->pNext_el != pDm_llist_el)
		pDm_llist_tmp = pDm_llist_tmp->pNext_el;

	if (pDm_llist_el->pcom_el != NULL)
		{	/* set up for deletion in com llist */
		pDm_llist_tmp->pcom_el->pNext_element = pDm_llist_el->pcom_el->pNext_element;
		com_llist_controller.new_llist_flag = 1;
		gsdl_llist_main(&com_llist_controller, &pOld_llist_start, NULL);
		tisFree(pDm_llist_el->pcom_el);	/* free mem from com llist */
		}
	delete_controllers(pDm_llist_el->prti_ctrl, pDm_llist_el->pdl_hdr->nctrl);
	delete_related_displays(pDm_llist_el->prti_rel_disp, pDm_llist_el->pdl_hdr->nrd);
	delete_execute_scripts(pDm_llist_el->prti_exe_script, pDm_llist_el->pdl_hdr->nes);
	delete_kill_displays(pDm_llist_el->prti_kill_display, pDm_llist_el->pdl_hdr->nkd);

	XFreeGC(pDm_llist_el->dl_context.d, pDm_llist_el->dl_context.gc);
 	XFreeGC(pDm_llist_el->dl_context.d, pDm_llist_el->static_dlcontext.gc);
	XDestroyWindow(pDm_llist_el->dl_context.d, pDm_llist_el->dl_context.w);	/* get rid of display window */

	if (pDm_llist_el->dl_context.pcmap)		delete_cmap(pDm_llist_el->dl_context.d, pDm_llist_el->dl_context.pcmap);
	if (pDm_llist_el->dl_context.pcrules)	delete_crules(pDm_llist_el->dl_context.pcrules);
	if (pDm_llist_el->prti_dynamic)
		{
		if (pDm_llist_el->pdl_hdr)
			{
			/*
			A leak here! loop along them and free the pLast_val member!
			*/

			struct rti_dynamic_graphic *pRti_dyn = pDm_llist_el->prti_dynamic;

			for (i = 0; i < pDm_llist_el->pdl_hdr->ndyn; i++, pRti_dyn++)
				tisFree(pRti_dyn -> pLast_val);
			}
		tisFree(pDm_llist_el->prti_dynamic);
		}
	if (pDm_llist_el->prti_color_rules)		tisFree(pDm_llist_el->prti_color_rules);
	if (pDm_llist_el->prti_mon)
		{
		if (pDm_llist_el -> pdl_hdr)
			{
			/*
			Another leak here! loop along them and free the pLast_val member!
			*/

			struct rti_update_mon	*prti_mon = pDm_llist_el->prti_mon;

			for (i = 0; i < pDm_llist_el->pdl_hdr->nmon; i++, prti_mon++)
				tisFree(prti_mon -> pLast_val);
			}
		tisFree(pDm_llist_el->prti_mon);
		}
	if (pDm_llist_el->prti_ctrl)			tisFree(pDm_llist_el->prti_ctrl);
	if (pDm_llist_el->prti_plot)			tisFree(pDm_llist_el->prti_plot);
	if (pDm_llist_el->prti_rel_disp)		tisFree(pDm_llist_el->prti_rel_disp);
	if (pDm_llist_el->prti_exe_script)		tisFree(pDm_llist_el->prti_exe_script);
	if (pDm_llist_el->prti_kill_display)	tisFree(pDm_llist_el->prti_kill_display);
	
	free_display_strings(pDm_llist_el);
	if (pDm_llist_el->ppMacro_pairs)		tisFree(pDm_llist_el->ppMacro_pairs);
	if (pDm_llist_el->pdl_hdr)				tisFree(pDm_llist_el->pdl_hdr);

	/* free mem from dm llist */
	pDm_llist_tmp->pNext_el= pDm_llist_el->pNext_el;
	tisFree(pDm_llist_el);
	return 0;
	}


void delete_plots(display, win_context, pRti_plot, nplots)
Display					*display;
XContext				win_context;
struct rti_update_plot	*pRti_plot;
unsigned short			nplots;
	{
	unsigned short	i, j;

	for (i = 0; i < nplots; i++, pRti_plot++)
		{
		switch (((struct dl_element_hdr *)pRti_plot->pDl_plot)->type)
		  {
		  case dl_strip:
		  	{
			Strip *pStrip = (Strip *)pRti_plot->pPlot;
			fdmgr_clear_timeout(pStrip->pfdctx, pStrip->palarm);
#ifdef OLD_STRIP
			gsd_sync_clear(pRti_plot->pSync_ctrl, pRti_plot->num_traces);
#else
			for (j = 0; j < pRti_plot->num_traces; j++)
				StopMonitor (*(strip_channel **)&(pRti_plot->getPV_vals[j]));
#endif OLD_STRIP
			stripTerm(pStrip);
			pRti_plot->pSync_ctrl = NULL;
			break;
			}
		  case dl_plot:
		  case dl_smith_overlay:
			gsd_sync_clear(pRti_plot->pSync_ctrl, pRti_plot->num_traces);
			pRti_plot->pSync_ctrl = NULL;
			pfoWinClose(pRti_plot->pPlot);
			break;
		  }
		/* free associated memory and set NULL so plot won't be updated */ 
		for (j = 0; j < pRti_plot->num_traces; j++)
			{
			if (pRti_plot->pPlot_data[j])
				tisFree(pRti_plot->pPlot_data[j]);
			}
		XDeleteContext(display, pRti_plot->window, win_context);	/* keep remaining x events off of the deleted window */
		XDestroyWindow(display, pRti_plot->window);	/* get rid of plot window */
		}
	}


/***************************************************************************************
 *
 *
 *	proc_plot
 *
 *	DK 3/91
 *
 *	Purpose	:  handle Xevents for plot windows
 *
 *	Input	:  pRti_plot-- ptr to run-time plot struct
 *		:  pEvent-- ptr to Xevent struct
 *	
 *	Output	: updates plots if necessary
 *
 *	Returns	: nothing
 *
 */
int proc_plot(pRti_plot, the_event)
struct rti_update_plot	*pRti_plot;
XEvent					the_event;
	{
	switch (((struct dl_anypo *)pRti_plot->pDl_plot)->element.type)
	  {
	  case dl_strip:
        switch (the_event.type)
		  {
		  case Expose:
	    	init_strip(pRti_plot);
	    	stripRefresh((Strip *)pRti_plot->pPlot);
	    	break;
		  case ConfigureNotify:
			/* the window has been resized */
			stripResize((Strip *)pRti_plot->pPlot);
			stripDraw((Strip *)pRti_plot->pPlot);
			break;
		  case ButtonPress:
			if (the_event.xbutton.button == Button3)
			dmStripMenu(&the_event, (Strip *)pRti_plot->pPlot);
			break;
		  }
		break;
	  case dl_plot:
        if (!((pRti_plot->init_bits == 0x0) || (pRti_plot->rti_plottype == 0)))
 	    	switch (the_event.type)
			  {
	    	  case Expose:
				if (the_event.xexpose.count == 0)
					do_plot(pRti_plot, 1, 0);
				break;
			  case ButtonPress:
				if (the_event.xbutton.button == Button3)
					dmPlotMenu(&the_event, pRti_plot);
				break;
			  }
		break;
	  case dl_smith_overlay:
		if (!((pRti_plot->init_bits == 0x0) || (pRti_plot->rti_plottype == 0)))
			switch (the_event.type)
			  {
			  case Expose:
				if(the_event.xexpose.count == 0)
					do_smith(pRti_plot, 1, 0);
				break;
			  case ButtonPress:
				if (the_event.xbutton.button == Button3)
					dmPlotMenu(&the_event, pRti_plot);
				break;
			  }	
		break;
	  }
	return 0;
	}


/***************************************************************************************
 *
 *
 *	dmSetup_cmap
 *
 *	DK 3/91
 *
 *	Purpose	: set up color maps for a particular display
 *
 *	Notes	: will use local colormap if specified in dl_hdr at configuration time,
 *		else will try to open remote display list file specified in dl_hdr and
 *		use color map defined there.
 *
 *	Input	: pDm_llist_el-- ptr to runtime struct for a display 
 *	
 *	Output	: fills in color map ptrs in pDm_llist_el
 *
 *	Memory	: create_cmap allocs memory for color maps
 *
 *	Returns	: 0 if color maps setup okay
 *		  -1 if error in setting up color maps
 *
 */
int dmSetup_cmap(pDm_llist_el)
struct dm_llist_el	*pDm_llist_el;
	{
	struct dl_color_map *pDl_colormap = (struct dl_color_map *)((short *)(pDm_llist_el->pdl_hdr) 
						+ pDm_llist_el->pdl_hdr->ocmap);

	pDm_llist_el->dl_context.color = color_visual_test(pDm_llist_el->dl_context.d,the_screen);

	/* get dl color map */
	/* if there's a local color map, use it--*/
	/* else try to get one from display specified in dl_hdr */

	if (pDl_colormap->element.type != dl_color_map)
	    {
	    /* if we have a remote color map */
	    if (strlen(pDm_llist_el->pdl_strings[(int)(pDm_llist_el->pdl_hdr->cmap)]) > 0)
		{
	    	pDm_llist_el->dl_context.pcmap = 		
				load_cmap_from_display_file(pDm_llist_el->dl_context.d,
				pDm_llist_el->pdl_strings[(int)(pDm_llist_el->pdl_hdr->cmap)], 
				getenv("EPICS_DISPLAY_PATH"), FALSE);
			if (!pDm_llist_el->dl_context.pcmap)
				return -1;
		}
		else
		    return -1;
	    }
	else
	    {
		pDm_llist_el->pdl_color_map = pDl_colormap;

		/* create color map for display list els */

		if((pDm_llist_el->dl_context.pcmap =
			create_cmap(pDm_llist_el->dl_context.d, pDm_llist_el->pdl_color_map + 1,
			pDm_llist_el->pdl_color_map->ncolors, 0, !(pDm_llist_el->dl_context.color), 
			pDm_llist_el->dl_context.color)) == NULL){
  		 	return -1;
	    }
	}

	/* set ptr to appropriate array of pixels (color or bw) */
	pDm_llist_el->pdl_color_map_index = pDm_llist_el->dl_context.color ?
			pDm_llist_el->dl_context.pcmap->pcolor_pixels : 
			pDm_llist_el->dl_context.pcmap->pinten_pixels;
	return 0;
	}


/***************************************************************************************
 *
 *
 *	dmSetup_crules
 *
 *	stolen from DK 1/93
 *
 *	Purpose	: set up color rules for a particular display
 *
 *	Notes	: will use local colorrules if specified in dl_hdr at configuration time,
 *		else will try to open remote display list file specified in dl_hdr and
 *		use color rules defined there.
 *
 *	Input	: pDm_llist_el-- ptr to runtime struct for a display 
 *	
 *	Output	: fills in color rules ptrs in pDm_llist_el
 *
 *	Memory	: create_crules allocs memory for color rules
 *
 *	Returns	: 0 if color rules setup okay
 *		  -1 if error in setting up color rules
 *
 */
int dmSetup_crules(pDm_llist_el)
struct dm_llist_el	*pDm_llist_el;
	{
	unsigned int				i;
	struct dl_dynamic_attribute	*pdyn;
	struct dl_element_hdr		*pelem;
	struct dl_color_rule_table	*pDl_colorrules	= (struct dl_color_rule_table *)((short *)(pDm_llist_el->pdl_hdr) +
								   pDm_llist_el->pdl_hdr->ocrules);

	/* get dl color rules */
	if (pDl_colorrules->element.type != dl_color_rule_table)
		{
		/* there are no local color rules, try to get one from display specified in dl_hdr */
		if (strlen(pDm_llist_el->pdl_strings[(int)(pDm_llist_el->pdl_hdr->cmap)]) == 0)
			return -1;	/* there are no local rules and no remote color rule display name, go figure! */

		else
			{
			LIST	st;	/* list for crule strings out of other display */

			lstInit(&st);
			pDm_llist_el->dl_context.pcrules = load_crules_from_display_file(
			  pDm_llist_el->pdl_strings[(int)(pDm_llist_el->pdl_hdr->cmap)], getenv("EPICS_DISPLAY_PATH"), &st);
			add_sting_table_strings_to_list(&st, pDm_llist_el);
			if (!pDm_llist_el->dl_context.pcrules )
				return -1;
			}
		}
	else
		{
		/* there are local color rules */
		pDm_llist_el->pdl_color_rules = pDl_colorrules;

		/* create them */
		pDm_llist_el->dl_context.pcrules = create_crules(pDm_llist_el->pdl_color_rules + 1, pDm_llist_el->pdl_color_rules->nrules, 0);
		if(pDm_llist_el->dl_context.pcrules == NULL)
			return -1;
		}

	/*
	 * go thru the display elements and resolve the color rule references
	 * for each element that specifies a color rule
	 *	 combine dm args and element args with rule channel strings to set up channel strings
	 */
	pDm_llist_el->ncolor_rule_objs = 0;

	/* for each element on the dynamic_graphic display list */
	pdyn = (struct dl_dynamic_attribute *)((short *)(pDm_llist_el->pdl_hdr) + pDm_llist_el->pdl_hdr->odyn);
	for (i = 0; i < pDm_llist_el->pdl_hdr->ndyn; /* see i++ below */ )
		{
		switch (pdyn->element.type)
		  {
		  case dl_basic_attribute:
			break;
		  case dl_dynamic_attribute:
			if (pdyn->attr.mod.clrmod == dl_clrmod_rule && *pDm_llist_el->pdl_strings[(int)(pdyn->attr.param.clrrule)])
				{
				fixup_rulechan_strings(pDm_llist_el, (int)pdyn->attr.param.clrrule, (int)pdyn->attr.param.clrargs,
				  (unsigned int *)&pdyn->attr.param.rulechan[0], (int)pdyn->attr.param.chan);
				}
			break;
		  case dl_rec:
		  case dl_line:
		  case dl_oval:
		  case dl_arc:
		  case dl_text:
			i++;		/* increment count only for graphic els */
			break;
		  default:
			printf(opiStringTab[STR_in_dmSetup_crules__bad_casen]);
			return -1;
		  }
		pdyn = (struct dl_dynamic_attribute *)((short *)pdyn + ((struct dl_element_hdr *)pdyn)->nwords);
		}

	/* for each element on the monitor display list */
	pelem = (struct dl_element_hdr *)((short *)(pDm_llist_el->pdl_hdr) + pDm_llist_el->pdl_hdr->omon);
	for (i = 0; i < pDm_llist_el->pdl_hdr->nmon; i++)
		{
		int				clrmod;
		int				clrrule;
		int				clrargs;
		unsigned int	*rulechan;
		int				channel;

		switch (pelem->type)
		  {
		  case dl_bar:
			clrmod		= ((struct dl_bar *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_bar *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_bar *)pelem)->clrargs);
			rulechan	= (unsigned int *)(&((struct dl_bar *)pelem)->rulechan[0]);
			channel		= (int)(((struct dl_bar *)pelem)->monitor.rdbk);
			break;
		  case dl_meter:
			clrmod		= ((struct dl_meter *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_meter *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_meter *)pelem)->clrargs);
			rulechan	= (unsigned int *)(&((struct dl_meter *)pelem)->rulechan[0]);
			channel		= (int)(((struct dl_meter *)pelem)->monitor.rdbk);
			break;
		  case dl_tank:
			clrmod		= ((struct dl_tank *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_tank *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_tank *)pelem)->clrargs);
			rulechan	= (unsigned int *)(&((struct dl_tank *)pelem)->rulechan[0]);
			channel		= (int)(((struct dl_tank *)pelem)->monitor.rdbk);
			break;
		  case dl_indicator:
			clrmod		= ((struct dl_indicator *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_indicator *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_indicator *)pelem)->clrargs);
			rulechan	= (unsigned int *)(&((struct dl_indicator *)pelem)->rulechan[0]);
			channel		= (int)(((struct dl_indicator *)pelem)->monitor.rdbk);
			break;
		  case dl_text_update:
			clrmod		= ((struct dl_text_update *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_text_update *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_text_update *)pelem)->clrargs);
			rulechan	= (unsigned int *)(&((struct dl_text_update *)pelem)->rulechan[0]);
			channel		= (int)(((struct dl_text_update *)pelem)->monitor.rdbk);
			break;
		  default:
			printf(opiStringTab[STR_in_dmSetup_crules__bad_casen]);
			return -1;
		  }
		if (clrmod == dl_clrmod_rule && *pDm_llist_el->pdl_strings[clrrule])
			fixup_rulechan_strings(pDm_llist_el, clrrule, clrargs, rulechan, channel);

		pelem = (struct dl_element_hdr *)((short *)pelem + pelem->nwords);
		}

	/* for each element on the controller display list */
	pelem = (struct dl_element_hdr *)((short *)(pDm_llist_el->pdl_hdr) + pDm_llist_el->pdl_hdr->octrl);
	for (i = 0; i < pDm_llist_el->pdl_hdr->nctrl; i++)
		{
		int				clrmod;
		int				clrrule;
		int				clrargs;
		unsigned int	*rulechan;
		int				channel;

		switch (pelem->type)
		  {
		  case dl_message_button:
			clrmod		= ((struct dl_message_button *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_message_button *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_message_button *)pelem)->clrargs);
			rulechan	= (unsigned int *)(&((struct dl_message_button *)pelem)->rulechan[0]);
			channel		= (int)(((struct dl_message_button *)pelem)->control.ctrl);
			break;
		  case dl_valuator:
			clrmod		= ((struct dl_valuator *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_valuator *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_valuator *)pelem)->clrargs);
			rulechan	= (unsigned int *)(&((struct dl_valuator *)pelem)->rulechan[0]);
			channel		= (int)(((struct dl_valuator *)pelem)->control.ctrl);
			break;
		  case dl_button:
			clrmod		= ((struct dl_button *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_button *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_button *)pelem)->clrargs);
			rulechan	= (unsigned int *)(&((struct dl_button *)pelem)->rulechan[0]);
			channel		= (int)(((struct dl_button *)pelem)->control.ctrl);
			break;
		  case dl_choice_button:
			clrmod		= ((struct dl_choice_button *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_choice_button *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_choice_button *)pelem)->clrargs);
			rulechan	= (unsigned int *)(&((struct dl_choice_button *)pelem)->rulechan[0]);
			channel		= (int)(((struct dl_choice_button *)pelem)->control.ctrl);
			break;
		  case dl_menu:
			clrmod		= ((struct dl_menu *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_menu *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_menu *)pelem)->clrargs);
			rulechan	= (unsigned int *)(&((struct dl_menu *)pelem)->rulechan[0]);
			channel		= (int)(((struct dl_menu *)pelem)->control.ctrl);
			break;
		  case dl_text_entry:
			clrmod		= ((struct dl_text_entry *)pelem)->clrmod;
			clrrule		= (int)(((struct dl_text_entry *)pelem)->clrrule);
			clrargs		= (int)(((struct dl_text_entry *)pelem)->clrargs);
			rulechan	= (unsigned int *)(&((struct dl_text_entry *)pelem)->rulechan[0]);
			channel		= (int)(((struct dl_text_entry *)pelem)->control.ctrl);
			break;
		  default:
			printf(opiStringTab[STR_in_dmSetup_crules__bad_casen]);
			return -1;
		  }
		if (clrmod == dl_clrmod_rule && *pDm_llist_el->pdl_strings[clrrule])
			fixup_rulechan_strings(pDm_llist_el, clrrule, clrargs, rulechan, channel);

		pelem = (struct dl_element_hdr *)((short *)pelem + pelem->nwords);
		}
	return 0;
	}


int add_sting_table_strings_to_list(pst, pDm_llist_el)
LIST				*pst;
struct dm_llist_el	*pDm_llist_el;
	{
	struct dl_hdr				*pdl_hdr	= pDm_llist_el->pdl_hdr;
	struct color_rules			*prules		= pDm_llist_el->dl_context.pcrules;
	struct dl_color_rule		*pdlcr;
	NODE						*pnode;
	unsigned int				rule, line, i;
	
	/*
	 * realloc the string table and add the stings in the string list
	 */
	pDm_llist_el->dl_context.pstring_table = pDm_llist_el->pdl_strings = 
	  (char **)tisRealloc(__FILE__, __LINE__, pDm_llist_el->pdl_strings, 
			(pdl_hdr->nstr + stringCount(pst)) * sizeof(char *));

	for (pnode = lstFirst(pst), i = pdl_hdr->nstr; pnode != NULL; pnode = lstNext(pnode), i++)
		{
		char	buf[500];
		char	*ptr = ((StringItem *)pnode)->string;
      
		if (strlen(ptr) > 0 && pDm_llist_el->ppMacro_pairs != NULL)
			{
			gsdl_main_macro(buf, ptr, pDm_llist_el->ppMacro_pairs, sizeof buf);
			ptr = &buf[0];
			}
		pDm_llist_el->pdl_strings[i] = strcpy((char *)tisCalloc(__FILE__, __LINE__, strlen(ptr) + 1, sizeof (char)), ptr);
		}

	/*
	 * resolve the string indices in the color rules
	 */
	pdlcr = pDm_llist_el->dl_context.pcrules->pdl_crules;
	for (rule = 0; rule < prules->nrules; rule++, pdlcr++)
		{
#ifndef TOKENPTRS
		pdlcr->name = (char *)(pdl_hdr->nstr + stringGetIndex(pst, pdlcr->name));
#else
		pdlcr->name = pdl_hdr->nstr + stringGetIndex(pst, dmPointer64fromToken (pdlcr->name));
#endif
		for (line = 0; line < MAX_RULE_LINES; line++)
			{
#ifndef TOKENPTRS
			pdlcr->info[line].chan  = (char *)(pdl_hdr->nstr + stringGetIndex(pst, pdlcr->info[line].chan));
			pdlcr->info[line].value = (char *)(pdl_hdr->nstr + stringGetIndex(pst, pdlcr->info[line].value));
#else
			pdlcr->info[line].chan  = pdl_hdr->nstr
					 + stringGetIndex(pst, dmPointer64fromToken (pdlcr->info[line].chan));
			pdlcr->info[line].value = pdl_hdr->nstr
					 + stringGetIndex(pst, dmPointer64fromToken (pdlcr->info[line].value));
#endif
			}
		}

	/*
	 * update the string count
	 * delete these strings to get rid of the storage space
	 */
	pdl_hdr->nstr += stringCount(pst);
	stringDeleteList(pst);
	return 0;
	}

int fixup_rulechan_strings(pDm_llist_el, rule, args, rulechan, this_channel)
struct dm_llist_el	*pDm_llist_el;
int	                rule;
int			args;
unsigned int		*rulechan;
int			this_channel;
	{
	char	*get_pc_string(), *get_rule_string();
	int	i, line;
	char	**macro_pairs;
	char	*this_channel_name = get_pc_string(pDm_llist_el->pdl_hdr, this_channel);

	pDm_llist_el->ncolor_rule_objs++;	/* count it */

	/*
	 * get the argument string broken up into pairs
	 * apply the pairs to the channel string and overwrite the place holder string
	 */
	macro_pairs = (char **)tisCalloc(__FILE__, __LINE__, MAX_PAIR_STRS*2, sizeof (char *));	/* reserve space for 2 sets */
	i = 0;
	/* do the dm args first */
	if (pDm_llist_el->ppMacro_pairs)
		for (; i < MAX_PAIR_STRS && pDm_llist_el->ppMacro_pairs[i]; i++)
			macro_pairs[i] = pDm_llist_el->ppMacro_pairs[i];
	/* now the builtin "name" variable, but don't include field name */
	if (this_channel_name)
		{
		char	*p;
		
		macro_pairs[i++] = "name";
		macro_pairs[i++] = strcpy(tisCalloc(__FILE__, __LINE__, strlen(this_channel_name) + 1, sizeof (char)), this_channel_name);
		for (p = macro_pairs[i-1]; *p; p++)
			if (*p == '.')
				{
				*p = 0; /* kill off the .XXX field name */
				break;
				}
		}
	/* now the elements args */
	get_macro_pairs(pDm_llist_el->pdl_strings[args], &macro_pairs[i], MAX_PAIR_STRS*2 - i, DefaultScreenOfDisplay(pDm_llist_el->dl_context.d));

	for (line = 0; line < MAX_RULE_LINES; line++)
		{
		char			resolved_name[db_name_dim];
		char			*template	= get_rule_string(pDm_llist_el, rule, line);
		struct dl_pc	*pDl_pc;
		unsigned short	pc_index;

		gsdl_main_macro(resolved_name, template, macro_pairs, sizeof resolved_name);

		pDl_pc = (struct dl_pc *)( ((short *)(pDm_llist_el->pdl_hdr)) + pDm_llist_el->pdl_hdr->opc);
		for (pc_index = 0; pc_index < pDm_llist_el->pdl_hdr->npc; pc_index++)
			{
			if (*(char *)(pDl_pc + 1) != '.' && !strcmp(resolved_name, (char *)(pDl_pc + 1)))
				break;
			pDl_pc = (struct dl_pc *) (((short *) pDl_pc) + pDl_pc->element.nwords);
			}
		if (pc_index != pDm_llist_el->pdl_hdr->npc)
			rulechan[line] = pc_index;	/* reassign the channel to a pre-existing instance */
		else
			strcpy(get_pc_string(pDm_llist_el->pdl_hdr, rulechan[line]), resolved_name);
		}
	if (this_channel_name)
		tisFree(macro_pairs[i-1]);
	tisFree(macro_pairs);

	return 0;
}

char *get_pc_string(pdl_hdr, channel)
struct dl_hdr	*pdl_hdr;
unsigned int	channel;
	{
	struct dl_pc	*pDl_pc = (struct dl_pc *)( ((short *)pdl_hdr) + pdl_hdr->opc);

	assert(channel < pdl_hdr->npc);
	while (channel-- > 0)
		pDl_pc = (struct dl_pc *) (((short *) pDl_pc) + pDl_pc->element.nwords);
	return (char *)(pDl_pc + 1);
	}

char *get_rule_string(pDm_llist_el, rule, line)
struct dm_llist_el	*pDm_llist_el;
int					rule;
int					line;
	{
	struct dl_color_rule	*prule	= pDm_llist_el->dl_context.pcrules->pdl_crules;
	unsigned int			i;

	for (i = 0; i < pDm_llist_el->dl_context.pcrules->nrules; i++, prule++)	
		if (strcmp(pDm_llist_el->pdl_strings[(int)(prule->name)], pDm_llist_el->pdl_strings[rule]) == 0)
			return pDm_llist_el->pdl_strings[(int)(prule->info[line].chan)];

	printf(opiStringTab[STR_bad_ruleHd_in_get_rule_strin]);
	return "";
	}



static void dm_capollfunc(pfdctx)	/* checks with channel access every 5 secs to see if we've lost a connection */
void	*pfdctx;
	{
    ca_pend_event(.00001);
    fdmgr_add_timeout(pfdctx, &dm_capoll, dm_capollfunc, pfdctx);
	}


static void dm_blinkpollfunc(pdm_llist_hdr)
struct dm_llist_hdr	*pdm_llist_hdr;
	{
	static	int			toggle = 0;
	struct dm_llist_el	*pDm_llist_el = pdm_llist_hdr->pNew_llist;

	toggle ^= 1;
	while ((pDm_llist_el != NULL) &&(pDm_llist_el->pdl_hdr != NULL))
		{
		if ( pDm_llist_el->visible)
			{
			struct dl_context	*pdlc = &pDm_llist_el->dl_context;

			do_blink(pdlc->d, pdlc->pcmap, toggle);
			}
		pDm_llist_el = pDm_llist_el->pNext_el;
		}
	fdmgr_add_timeout(pfdctx, &dm_blinkpoll, dm_blinkpollfunc, pdm_llist_hdr);
	}

int kludgeblinker()
	{
	printf(opiStringTab[STR_dm_called_klugeblinker_whyn ]);
	return 0;
	}

int convert_string_index_to_ptr()
	{
	printf(opiStringTab[STR_dm_called_convert_string_ind]);
	return 0;
	}
	
