#include "opiStringDefs.h"
extern	char	*opiStringTab[];
/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
#include <X11/Xlib.h>  			/* the X library 		*/
#ifdef WIN32
#include <X11/XlibXtra.h>
#include <winsock.h>
#endif WIN32
#include <X11/Xutil.h>
#include <db_access.h>			/* db types and structs		*/
#include "alarm.h"
#include "cadef.h"
#include "dldef.h"			/* display list defines 	*/
#include "dlinfo.h"			/* more display list defines 	*/
#include "font_table.h"			/* defines for font table	*/
#include "dl_context.h"			/* defines dl context		*/
#include "gizmo.h"			/* defs for gizmos 		*/
#include "dynamic_def.h"		/* rti dynamic defines		*/
#include "op_gsdl_def.h"		/* com_table element def	*/
#include "draw_strings.h"
#include "tisXext.h"

#ifdef AXP
#include <string.h>
#endif

#define F_TO_STR_MAX	10000000	/* max num accepted by f_to_str() 			*/
#define SCI_NOT			1	/* indicate regular e-notation 				*/
#define BITS_PER_INT	(sizeof(int) * 8)

/* THIS NUMBER MUST BE AT LEAST AS BIG AS THE MAX_STRING_SIZE VALUE INDB_ACCESS */
#define MAX_CHARS		MAX_STRING_SIZE

extern void *tisMalloc();
extern void *tisCalloc();
extern void tisFree();

char *dmErrStrs[] =
	{
	0,
	0,
	0,
	0,
	0,
	0,
	0
	};

/***********************************************************************
**
**	MONITOR_OBJS.C
**	
**	Deb Kerstiens 10/90
**
**	Purpose : These routines draw the dl_monitor-type objects from a display list
**		into an X Window.
**	Note: The index into the font table should be determined for each
**		object only once and used on repeated display.  Should this be
**		saved in the run-time part of the struct or saved in the display
**		list at configure time ??
**
**
**	Modification History:
**
**	08/12/91	DK	added valid_data flag check as part of not_conn check
**	08/30/91	DK	added txt msgs to indicate why data not displayed to
**				the white rectangle draw
**	09/17/91	DK	added setup of default limits 0.0-100.0 to indicators,
**				meters, and bars
**	09/18/91	DK	removed not_conn flag from set_fgbg, removed 
**				status_to_color() since it is not used
**	10/03/91	DK	added units to text updates
**	10/04/91	DK	fixed bug in dmDisplayPvErr for ""
**				channel name 
**	10/04/91	DK	added invalid db_type error to meters, bars, 
**				and indicators for db_type = string --no
**				longer draw the string
**	10/16/91	DK	added ca_name() to init routine params for macro 
**				substitution of channel label type
**	10/23/91	DK	fixed bug in call to draw_sring --rec -> &rec
**	10/24/91	DK	changed dmDisplayPvErr to display 3 strings instead of 2
**	10/31/91	DK	modified dmDisplayPvErr to display as many of the 3 possible
**				error information strings as it can --before, it tried to 
**				display 3 and might get none because the boxes were too small
**	1/31/92		DK	fixed constants in dmSave_str() to be proper
**				db defines
**
**********************************************************************/



/***********************************************************************
*
*	update_mon_objs
*	
*	Deb Kerstiens 10/90
*
*	Purpose : Draws each monitor object on the display list.  
*		For each el:
*			determine if it needs to be drawn (exposed or modified)
*			if so, draw either error state (not connected, etc.) 
*				or specific object (bar, indicator, etc.)
*		
*
*	Input :
*	    pdl_context : pointer to Display context structure 
*	    pdl_index	: color map index from display list
*	    pdl		: ptr to display list info
*	    pdyn	: ptr to the rti_dynamic list
*	    ndyn	: count of dynamic elements to display
*	    prec	: area to be updated (for expose)
*	    expose_flag	: indicates whether expose event or
*				regular update pass
*	Output :
*	    none
*	Returns
*	    none
*
*
*
**********************************************************************/

void update_mon_objs(pdlc, pdl_index, pdl, prti, nmon, prec, pDm_llist_el, expose_flag)
struct dl_context	*pdlc;
#ifndef AXP
int 			*pdl_index;
#else
unsigned long		*pdl_index;
#endif
void 			*pdl;
struct rti_update_mon 	*prti;
unsigned short		nmon;
XRectangle		*prec;
struct dm_llist_el	*pDm_llist_el;
int			expose_flag;	
	{
	void			*dis_ptr;
	unsigned short		i;
	struct rti_update_mon	*prti_mon; 
	unsigned short		pv_err;			/* indicates no ca cnxn */
	unsigned short		dmChkPvStat();
	short			db_ftype;		/* native field type of chans*/
	int			*pset_modgeom;		/* ptr to set modified array */

    /* for each element on the display list */
    for (dis_ptr = pdl, prti_mon = prti, i = 0; i < nmon; dis_ptr = ((short *)dis_ptr + ((struct dl_element_hdr *)dis_ptr)->nwords), prti_mon++, i++)
		{
    	struct dl_object_hdr	*objptr = &(((struct dl_anygo *)dis_ptr)->object);

    	pset_modgeom = (prti_mon->pgeom_comel->pMod_bits);
    	pv_err = dmChkPvStat(prti_mon->pgeom_comel, prti_mon->init_flag, prti_mon->not_conn);
    	if (expose_flag) {  /* update exposed area */
    	    if ((objptr->x+objptr->width) < prec->x) {
	    	continue;
	    }
    	    if (objptr->x > (prec->x + (int)(prec->width))) {
	    	continue;
	    }
    	    if ((objptr->y+objptr->height) < prec->y) {
	    	continue;
	    }
	    if (objptr->y > (prec->y + (int)(prec->height))) {
	    	continue;
	    }
	    if (pv_err) {
		/* we can't display a real value for this object, display error state */
	    	dmDisplayPvErr(objptr,pv_err, prti_mon->pgeom_comel, pdlc, prti_mon->pLast_val);
	    	prti_mon->init_flag = FALSE;
	    }
    	} else {	/* this is not an expose, check for modification */
            if (pv_err) {
    	    	/* if the error state is the same as last time, no graphics */
    	    	if (prti_mon->not_conn && (prti_mon->not_conn == pv_err))
		    continue;
	    	else {	/* display error condition */
		    dmDisplayPvErr(objptr, pv_err, prti_mon->pgeom_comel, pdlc, prti_mon->pLast_val);
		    prti_mon->init_flag = FALSE;
	    	}
	    } else {	/* we have connection and data area, check for modification */
			/* if the dynamic chan has not been modified, no graphics */
	    	if (!(pset_modgeom[prti_mon->geom_intix] & ~clear_mod_bit_mask[prti_mon->geom_modix]) && !prti_mon->clrrule_has_changed)
		    continue;
	    }
    	}
     	/* if we made it here, we need to update the display -- either exposed or modified */
     	if (pv_err == dmConnection_okay) {

#ifndef AXP
	    int		clr;
	    int		bclr;
	    int		clrmod;
	    int		rulecolorfg;
	    int		rulecolorbg;
#else
	    unsigned long clr;
	    unsigned long bclr;
	    int		clrmod;
	    unsigned long rulecolorfg;
	    unsigned long rulecolorbg;
#endif

	    if (prti_mon->clrrule_has_changed)
		expose_flag = 1;
	    prti_mon->clrrule_has_changed = 0;

	    db_ftype = ca_field_type(prti_mon->pgeom_comel->pChan);
	    switch(((struct dl_element_hdr *)dis_ptr)->type) {
		case dl_bar:
		    clr	= ((struct dl_bar *)prti_mon->pdl_mon)->monitor.clr;
		    bclr = ((struct dl_bar *)prti_mon->pdl_mon)->monitor.bclr;
		    clrmod = ((struct dl_bar *)prti_mon->pdl_mon)->clrmod;
		    rulecolorfg = ((struct dl_bar *)prti_mon->pdl_mon)->rulecolorfg;
		    rulecolorbg = ((struct dl_bar *)prti_mon->pdl_mon)->rulecolorbg;
		    break;
		case dl_meter:
		    clr = ((struct dl_meter *)prti_mon->pdl_mon)->monitor.clr;
		    bclr = ((struct dl_meter *)prti_mon->pdl_mon)->monitor.bclr;
		    clrmod = ((struct dl_meter *)prti_mon->pdl_mon)->clrmod;
		    rulecolorfg = ((struct dl_meter *)prti_mon->pdl_mon)->rulecolorfg;
		    rulecolorbg = ((struct dl_meter *)prti_mon->pdl_mon)->rulecolorbg;
		    break;
		case dl_indicator:
		    clr = ((struct dl_indicator *)prti_mon->pdl_mon)->monitor.clr;
		    bclr = ((struct dl_indicator *)prti_mon->pdl_mon)->monitor.bclr;
		    clrmod = ((struct dl_indicator *)prti_mon->pdl_mon)->clrmod;
		    rulecolorfg = ((struct dl_indicator *)prti_mon->pdl_mon)->rulecolorfg;
		    rulecolorbg = ((struct dl_indicator *)prti_mon->pdl_mon)->rulecolorbg;
		    break;
		case dl_text_update:
		    clr = ((struct dl_text_update *)prti_mon->pdl_mon)->monitor.clr;
		    bclr = ((struct dl_text_update *)prti_mon->pdl_mon)->monitor.bclr;
		    clrmod = ((struct dl_text_update *)prti_mon->pdl_mon)->clrmod;
		    rulecolorfg = ((struct dl_text_update *)prti_mon->pdl_mon)->rulecolorfg;
		    rulecolorbg = ((struct dl_text_update *)prti_mon->pdl_mon)->rulecolorbg;
		    break;
	    }
    	    if (set_fgbg(pdlc, pdl_index, prti_mon, clr, bclr, clrmod, rulecolorfg, rulecolorbg))
		switch(((struct dl_element_hdr *)dis_ptr)->type) {
		    case dl_bar:
			draw_bar(pdlc, objptr, prti_mon, db_ftype, &pv_err, expose_flag);
			break;
		    case dl_meter:
			draw_meter(pdlc, objptr, prti_mon, db_ftype, &pv_err, expose_flag);
		 	break;
		    case dl_indicator:
			draw_indicator(pdlc, objptr, prti_mon, db_ftype, &pv_err, expose_flag);
			break;
		    case dl_text_update:
			rti_update_text(pdlc, ((struct dl_anygo *)dis_ptr)->object, db_ftype,
					prti_mon, &pv_err, COLOR_LOOKUP(pdl_index, (short)bclr));
			break;
		}
	    else if (expose_flag) {
		XRectangle rec;
		rec = *(XRectangle *)objptr;
		rec.width++;
		rec.height++;
		XClearArea(pdlc->d, pdlc->w, rec.x, rec.y, rec.width, rec.height, FALSE);
		draw_static(&(pDm_llist_el->static_dlcontext), pDm_llist_el->pstatic, pDm_llist_el->pdl_hdr->nwords_sta, &rec, FALSE);
	    }
	}
	if (pset_modgeom)
   	    pset_modgeom[prti_mon->geom_intix] = pset_modgeom[prti_mon->geom_intix] & clear_mod_bit_mask[prti_mon->geom_modix];
   	prti_mon->not_conn = pv_err;
    }
    XFlush(pdlc->d);
}



/***********************************************************************
**
**	cvrt_to_str
**	
**	Deb Kerstiens 8/90
**
**	Purpose : This routine builds an ascii string representing the
**			value of the data.   
**	Note:	channel access setup is done using dbr_*_ctrl structs.
**		This routine depends upon that fact and must be modified
**		if different.  (Ref op_comm.c )
**		The convert for double should be changed to do something
**		like f_to_str for efficiency.
**	
**
**	Input :
**	    pcom_el	: pointer to a com record structure 
**	    field_type	: native type of the data stored in the struct
**	    display_str	: storage for the string to be filled in( must
**				be large enough to hold data since
**				covnersion routines don't clamp)
**	    format	: specifies data presentation(scientific, float, 
**				integer,octal,hex) type
**
**	Output :
**	    display_str	: ascii representation of the data 
**	Returns:
**	    none
**
**********************************************************************/


cvrt_to_str(pcom_el, field_type, display_str, format)
struct gsdl_com_record	*pcom_el;
short					field_type;
char					*display_str;
enum dl_string_format	format;
	{
	switch (field_type)
	  {
	  case OLD_DBR_STRING:
		strcpy(display_str, ((struct dbr_sts_string *)pcom_el->pData_struct)->value);
		break;
	  case OLD_DBR_FLOAT:
	  	{
		struct dbr_ctrl_float	*pctrl_float = ((struct dbr_ctrl_float *)pcom_el->pData_struct);
#ifdef AXP
		/* gcc on AXP generates a signal 0 if we use 'pctrl_float->value' */
		float local_float = pctrl_float->value;
		/* clamp value to prevent overflow in convert buffer! */
		if (local_float > 1.0E29 || local_float < -1.0E29)
		{
		    abort ();
		}
#else
		/* clamp value to prevent overflow in convert buffer! */
		if (pctrl_float->value > 1.0E29 || pctrl_float->value < -1.0E29)
		{
		    abort ();
		}
#endif
		/* clamp precision to reasonable value! */
		if (pctrl_float->precision > 9)
		{
		    pctrl_float->precision = 9;
		}
		switch (format)
		  {
		  case dl_format_e:
#ifndef AXP
			if (pctrl_float->value > F_TO_STR_MAX) 
			    sprintf(display_str, "% 1.*e", pctrl_float->precision, pctrl_float->value);
			else 
			    f_to_estr(pctrl_float->value, display_str, pctrl_float->precision, SCI_NOT);
#else
			if (local_float > F_TO_STR_MAX) 
			    sprintf(display_str, "% 1.*e", pctrl_float->precision, local_float);
			else 
			    f_to_estr(local_float, display_str, pctrl_float->precision, SCI_NOT);
#endif
			break;
		  case dl_format_g:
#ifndef AXP
			sprintf(display_str, "% .*g", pctrl_float->precision, pctrl_float->value);
#else
			sprintf(display_str, "% .*g", pctrl_float->precision, local_float);
#endif
			break;
		  case dl_format_f:
		  default:		    
#ifndef AXP
			if (pctrl_float->value > F_TO_STR_MAX) 
				sprintf(display_str, "% .*f", pctrl_float->precision, pctrl_float->value);
			else
				ff_to_str(pctrl_float->value, display_str, pctrl_float->precision);
#else
			if (local_float > F_TO_STR_MAX) 
				sprintf(display_str, "% .*f", pctrl_float->precision, local_float);
			else
				ff_to_str(local_float, display_str, pctrl_float->precision);
#endif
			break;
		  }
		break;
	    }
	  case OLD_DBR_DOUBLE:
	  	{
		struct dbr_ctrl_double	*pctrl_double = ((struct dbr_ctrl_double *)pcom_el->pData_struct);
		char					*format_string;

		/* clamp value to prevent overflow in convert buffer! */
		if (!_isnan(pctrl_double->value) && (pctrl_double->value > 1.0E29 || pctrl_double->value < -1.0E29))
		{
		    abort ();
		}
		/* clamp precision to reasonable value! */
		if (pctrl_double->precision > 9)
		{
		    pctrl_double->precision = 9;
		}
		switch (format)
		  {
		  case dl_format_e:	format_string = "% 1.*e";	break;
		  case dl_format_g:	format_string = "% .*g";	break;
		  default:
		  case dl_format_f:	format_string = "% .*f";	break;
		  }
		if (_isnan(pctrl_double->value))
		    sprintf(display_str, "Nan");
		else
		    sprintf(display_str, format_string, pctrl_double->precision, pctrl_double->value);
		break;
	    }
	  case OLD_DBR_ENUM:
		{
		struct dbr_ctrl_enum	*pctrl_enum = ((struct dbr_ctrl_enum *)pcom_el->pData_struct);
		
		if((pctrl_enum->value >= 0) && (pctrl_enum->value <= (pctrl_enum->no_str - 1)))			
		    strcpy(display_str, pctrl_enum->strs[pctrl_enum->value]);
		else
			strcpy(display_str, opiStringTab[STR_INVALID_VALUE               ]);
		break;
	    }
	  case OLD_DBR_LONG:
		{
		struct dbr_ctrl_long		*pctrl_int = ((struct dbr_ctrl_long *)pcom_el->pData_struct);
		char					*format_string;

		switch (format)
		  {
		  case dl_format_o:	format_string = "%oo";	break;
		  case dl_format_x:	format_string = "%xx";	break;
		  default:
		  case dl_format_i:	format_string = "%d";	break;
		  }
		sprintf(display_str, format_string, pctrl_int->value);
		break;
	    }
	  case OLD_DBR_SHORT:
		{
		struct dbr_ctrl_int		*pctrl_int = ((struct dbr_ctrl_int *)pcom_el->pData_struct);
		char					*format_string;

		switch (format)
		  {
		  case dl_format_o:	format_string = "%oo";	break;
		  case dl_format_x:	format_string = "%xx";	break;
		  default:
		  case dl_format_i:	format_string = "%d";	break;
		  }
		sprintf(display_str, format_string, pctrl_int->value);
		break;
	    }
	  case OLD_DBR_CHAR:
		{
		struct dbr_ctrl_char	*pctrl_char = ((struct dbr_ctrl_char *)pcom_el->pData_struct);
		char					*format_string;
		
		switch (format)
		  {
		  case dl_format_o:	format_string = "%oo";	break;
		  case dl_format_x:	format_string = "%xx";	break;
		  default:
		  case dl_format_i:	format_string = "%d";	break;
		  }
		sprintf(display_str, format_string, pctrl_char->value);
		break;
	    }
	  default:	/* printf("Unknown data type in cvrt_to_str()\n");*/
		return -1;
	  }
	return 0;	
	}

	


/***********************************************************************
**
**	cvrt_to_float
**	
**	Deb Kerstiens 8/90
**
**	Purpose : Cvrt_to_float produces the float representation of the  
**			pv value from the db stored in the com_el struct.
**	Note:	channel access setup is done using dbr_*_ctrl structs.
**		This routine depends upon that fact and must be modified
**		if different.  (Ref op_comm.c )
**
**	Input :
**	    pcom_el	: pointer to a com record structure 
**	    field_type	: native type of the data stored in the struct
**
**	Output :
**	    none 
**	Returns:	    
**	    fval	: float value of data
**
**********************************************************************/

float cvrt_to_float(pcom_el, field_type)
struct gsdl_com_record	*pcom_el;
short					field_type;
	{	
	switch (field_type)
	  {
	  case OLD_DBR_FLOAT:	return ((struct dbr_ctrl_float *)pcom_el->pData_struct)->value;
	  case OLD_DBR_DOUBLE:	return (float)(((struct dbr_ctrl_double *)pcom_el->pData_struct)->value);
	  case OLD_DBR_ENUM:	return (float)(((struct dbr_ctrl_enum *)pcom_el->pData_struct)->value);
	  case OLD_DBR_LONG:	return (float)(((struct dbr_ctrl_long *)pcom_el->pData_struct)->value);
	  case OLD_DBR_SHORT:	return (float)(((struct dbr_ctrl_int *)pcom_el->pData_struct)->value);
	  case OLD_DBR_CHAR:	return (float)(((struct dbr_ctrl_char *)pcom_el->pData_struct)->value);
	  default:			return (float)0.0;
	  }
	}




/***********************************************************************
**
**	get_hoprnlopr
**	
**	Deb Kerstiens 8/90
**
**	Purpose : This routine fills in values for high operating 
**			and low operating limits from db.
**	Note:	channel access setup is done using dbr_*_ctrl structs.
**		This routine depends upon that fact and must be modified
**		if different.   
**	
**
**	Input :
**	    pcom_el	: pointer to a com record structure 
**	    field_type	: native type of the data stored in the struct
**	    pHopr	: pointer to high operating limit
**	    pLopr	: pointer to low operating limit
**	    pPrec	: pointer to precision 
**	    
**	Output :
**	    pHopr	: fills in value from db struct
**	    pLopr	: fills in value from db struct
**	    pPrec	: fills in value from db struct
**	Returns:
**	    none
**
**********************************************************************/


get_hoprnlopr(pcom_el, field_type, pHopr, pLopr, pPrec)
struct gsdl_com_record	*pcom_el;
short					field_type;
float					*pHopr;
float					*pLopr;
unsigned short			*pPrec;
	{
	switch (field_type)
	  {
	  case OLD_DBR_FLOAT:
		*pHopr = ((struct dbr_ctrl_float *)pcom_el->pData_struct)->upper_disp_limit;
		*pLopr = ((struct dbr_ctrl_float *)pcom_el->pData_struct)->lower_disp_limit;
		*pPrec = ((struct dbr_ctrl_float *)pcom_el->pData_struct)->precision;
		break;
	  case OLD_DBR_DOUBLE:
		*pHopr = (float)((struct dbr_ctrl_double *)pcom_el->pData_struct)->upper_disp_limit;
		*pLopr = (float)((struct dbr_ctrl_double *)pcom_el->pData_struct)->lower_disp_limit;
		*pPrec = ((struct dbr_ctrl_double *)pcom_el->pData_struct)->precision;
		break;
	  case OLD_DBR_ENUM:
		*pLopr = (float)0.0;
		*pHopr = (float)((((struct dbr_ctrl_enum *)pcom_el->pData_struct)->no_str) - 1.0);	/* high is number of enum values - 1 */
		*pPrec = 0;
		break;
	  case OLD_DBR_LONG:
		*pHopr = (float)((struct dbr_ctrl_long *)pcom_el->pData_struct)->upper_disp_limit;
		*pLopr = (float)((struct dbr_ctrl_long *)pcom_el->pData_struct)->lower_disp_limit;
		*pPrec = 0;
		break;
	  case OLD_DBR_SHORT:
		*pHopr = ((struct dbr_ctrl_int *)pcom_el->pData_struct)->upper_disp_limit;
		*pLopr = ((struct dbr_ctrl_int *)pcom_el->pData_struct)->lower_disp_limit;
		*pPrec = 0;
		break;
	  case OLD_DBR_CHAR:
		*pHopr = ((struct dbr_ctrl_char *)pcom_el->pData_struct)->upper_disp_limit;
		*pLopr = ((struct dbr_ctrl_char *)pcom_el->pData_struct)->lower_disp_limit;
		*pPrec = 0;
		break;
	  default:
/*		printf("Unknown data type in get_hoprnlopr()\n");*/
		break;
	  }
	return 0;
	}


/***********************************************************************
**
**	draw_bar
**	
**	Deb Kerstiens 8/90
**
**	Purpose : Draw_bar initializes and handles the graphic routines
**		that update a bar-type monitor
**	Note:	channel access setup is done using dbr_*_ctrl structs.
**		This routine depends upon that fact and must be modified
**		if different.  (Ref op_comm.c )
**
**	Input :
**	    pdlc	: pointer to display context info 
**	    objptr	: pointer to dl_header(x,y,h,w)
**	    pdl_index	: color map index for display list
**	    pPv_err	: ptr to error status for process var (conn,etc.)
**	    db_ftype	: native db type of the pv
**	    prti_mon	: pointer to rti monitor struct
**	    field_type	: native type of the data stored in the struct
**
**	Output :
**	    none 
**	Returns:	    
**	    nothing
**
**********************************************************************/

draw_bar(pdlc, objptr, prti_mon, db_ftype, pPv_err, expose_flag)
struct dl_context		*pdlc;
struct dl_object_hdr	*objptr;
struct rti_update_mon	*prti_mon;
short					db_ftype;
unsigned short			*pPv_err, expose_flag;
	{
    float			fval = (float)0.0;

    if (db_ftype == OLD_DBR_STRING)
		{
		/* string type, indicate invalid db_type */
		if (expose_flag || !prti_mon->init_flag)
			{
			/* first time thru or expose - do something here to indicate icky-poo */
			*pPv_err = dmInvalid_dbtype;
			dmDisplayPvErr(objptr, *pPv_err, prti_mon->pgeom_comel, pdlc, prti_mon->pLast_val);
			prti_mon->init_flag = TRUE;
			}
		/*rti_drawtext(pdlc, objptr, db_ftype, prti_mon, pPv_err);*/
		return 0;
		}

    /* draw bar */
    fval = cvrt_to_float(prti_mon->pgeom_comel, db_ftype);
    dmSave_float(&prti_mon->pLast_val, fval);
    if (!prti_mon->init_flag)
		{
    	/* initializing, set up function calls */
		setup_rec_funcs(prti_mon);
		get_hoprnlopr(prti_mon->pgeom_comel,db_ftype, &(prti_mon->hopr),&(prti_mon->lopr), &(prti_mon->g_type.rectangle.hdr.precision));
		if (!(prti_mon->hopr > prti_mon->lopr))
			{
			prti_mon->hopr = (float)100.0;
			prti_mon->lopr = (float)0.0;
			}
		(*(prti_mon->init_func)) (pdlc, objptr, prti_mon->lopr,prti_mon->hopr, 0, fval, &(prti_mon->g_type.rectangle), 
		  (short)((struct dl_bar *)prti_mon->pdl_mon)->label, ca_name(prti_mon->pgeom_comel->pChan));
		prti_mon->init_flag = TRUE;
		}

    else if (prti_mon->not_conn || expose_flag)
		{
		/* shown as not connected last time thru (new connection) or expose event - init graphics and draw */
		(*(prti_mon->init_func)) (pdlc, objptr, prti_mon->lopr,prti_mon->hopr, 1, fval, &(prti_mon->g_type.rectangle),
		  (short)((struct dl_bar *)prti_mon->pdl_mon)->label, ca_name(prti_mon->pgeom_comel->pChan));	
		}
    else
		{
		/* updating - draw to indicate current value */
		(*(prti_mon->upd_func))(pdlc, objptr, fval, prti_mon->lopr, prti_mon->hopr, &(prti_mon->g_type.rectangle));
		}
	}


/***********************************************************************
**
**	setup_rec_funcs
**	
**	Deb Kerstiens 8/90
**
**	Purpose :  Setup the appropriate routines to be called for the
**			bar type specified
**
**
**	Input :
**	    prti_mon	: pointer to rti monitor struct
**
**	Output :
**	    none 
**	Returns:	    
**	    nothing
**
**
**********************************************************************/

setup_rec_funcs(prti_mon)
struct rti_update_mon	*prti_mon;
	{
	int				init_rec_l(), FillRecLtToRt();
	int				init_rec_b(), fill_recv();
	int				init_rec_r(), FillRecRtToLt();
	int				init_rec_t(), fill_rect();
	int				init_rec_m(), FillRecFromMid();
	int				init_rec_midv(), FillRecFromMidv();
	struct dl_bar	*pDl_bar = prti_mon->pdl_mon;

	if (pDl_bar->fillmod == dl_zero_fill)
		{
		switch (pDl_bar->direction)
		  {
		  case dl_increase_right:	prti_mon->init_func = init_rec_m;		prti_mon->upd_func = FillRecFromMid;	break;
		  case dl_increase_up:		prti_mon->init_func = init_rec_midv;	prti_mon->upd_func = FillRecFromMidv;	break;
		  case dl_increase_left:	prti_mon->init_func = init_rec_m;		prti_mon->upd_func = FillRecFromMid;	break;
		  case dl_increase_down:	prti_mon->init_func = init_rec_midv;	prti_mon->upd_func = FillRecFromMidv;	break;
		  case dl_increase_radial:	prti_mon->init_func = init_rec_m;		prti_mon->upd_func = FillRecFromMid;	break;	
		  default:					printf(opiStringTab[STR_geom_type_not_matchedn      ]);			break;
		  }
		}
	else
		{
		switch (pDl_bar->direction)
		  {
		  case dl_increase_right:	prti_mon->init_func = init_rec_l;		prti_mon->upd_func = FillRecLtToRt;		break;
		  case dl_increase_up:		prti_mon->init_func = init_rec_b;		prti_mon->upd_func = fill_recv;			break;
		  case dl_increase_left:	prti_mon->init_func = init_rec_r;		prti_mon->upd_func = FillRecRtToLt;		break;
		  case dl_increase_down:	prti_mon->init_func = init_rec_t;		prti_mon->upd_func = fill_rect;			break;
		  case dl_increase_radial:	prti_mon->init_func = init_rec_m;		prti_mon->upd_func = FillRecFromMid;	break;	
		  default:					printf(opiStringTab[STR_geom_type_not_matchedn      ]);			break;
		  }
		}
	return 0;
	}	



/***********************************************************************
**
**	rti_update_text
**	
**	Deb Kerstiens 8/90
**
**	Purpose : Rti_update_text initializes and handles the graphic routines
**		that update a text-type monitor
**	Note:	channel access setup is done using dbr_*_ctrl structs.
**		This routine depends upon that fact and must be modified
**		if different.  (Ref op_comm.c )
**
**	Input :
**	    pdlc	: pointer to display context info 
**	    obj		: dl_header struct (x,y,h,w)
**	    pPv_err	: ptr to error status for process var (conn,etc.)
**	    db_ftype	: native db type of the pv
**	    prti_mon	: pointer to rti monitor struct
**	    db_ftype	: native type of the data stored in the struct
**	    bg_color	: color to use when drawind background box
**
**	Output :
**	    none 
**	Returns:	    
**	    nothing
**
**********************************************************************/

rti_update_text(pdlc, obj, db_ftype, prti_mon, pPv_err, bg_color)
struct dl_context	*pdlc;
struct dl_object_hdr 	obj;
struct rti_update_mon 	*prti_mon;
short 			db_ftype;
unsigned short 		*pPv_err;
int			bg_color;
	{
    char 		display_str[MAX_CHARS];
    XRectangle		xrec;
    enum string_format	frmt;
    XGCValues 		gc_vals;
    static GC 		color_box_gc;
    static unsigned char gc = FALSE;	

    if (cvrt_to_str(prti_mon->pgeom_comel, db_ftype, display_str, ((struct dl_text_update *)prti_mon->pdl_mon)->format) < 0)
		{
		/* invalid string, display error  Note we don't set pv_err--will check again next time through */
		dmDisplayPvErr(&obj, dmInvalid_data, prti_mon->pgeom_comel, pdlc, prti_mon->pLast_val);
		return 0;
    	}

    dmSave_str(&prti_mon->pLast_val, display_str);		/* save current value */
    if (!prti_mon->init_flag)
		{
		int		draw_fixed_horizontal_string(), draw_fixed_vertical_string();

		switch (((struct dl_text_update *)prti_mon->pdl_mon)->align)
		  {
		  case dl_horiz_left:
		  case dl_horiz_right:
		  case dl_horiz_center:	prti_mon->upd_func = draw_fixed_horizontal_string;	break;
		  case dl_vert_top:
		  case dl_vert_bottom:
		  case dl_vert_center:	prti_mon->upd_func = draw_fixed_vertical_string;	break;
		  default:				abort();
		  }
		}
    gc_vals.foreground = bg_color;
    if (!gc)
		{
		color_box_gc = XCreateGC(pdlc->d, pdlc->w, GCForeground, &gc_vals);
		gc = TRUE;
    	}
    else
		XChangeGC(pdlc->d, color_box_gc, GCForeground, &gc_vals);	/* set the fg color in the gc */

    /* draw background box */
    obj.width  = obj.width  ? obj.width  : 1;	/* ensure some size */
    obj.height = obj.height ? obj.height : 1;
    XFillRectangle(pdlc->d, pdlc->w, color_box_gc, obj.x, obj.y, obj.width-1, obj.height-1);
#if 0
    if ((strchr(ca_name(prti_mon->pgeom_comel->pChan), '.') == NULL) || (strstr(ca_name(prti_mon->pgeom_comel->pChan), ".VAL") != NULL))
		dmGet_ctrlunits(prti_mon->pgeom_comel,db_ftype, display_str);
#endif

    DLOBJ_TO_XREC(obj, xrec);
    switch (((struct dl_text_update *)prti_mon->pdl_mon)->align)
	  {
	  case dl_horiz_left:
	  case dl_vert_top:		frmt = box_begin;	break;
	  case dl_horiz_right:
	  case dl_vert_bottom:	frmt = box_end;		break;
	  case dl_horiz_center:
	  case dl_vert_center:	frmt = box_center;	break;
	  default:				abort();
	  }
    (*(prti_mon->upd_func))(pdlc->d, pdlc->w, pdlc->gc, &xrec, pdlc->pfont_table, display_str, strlen(display_str), frmt);
	}




/***********************************************************************
**
**	draw_meter
**	
**	Deb Kerstiens 8/90
**
**	Purpose : Draw_meter initializes and handles the graphic routines
**		that update a meter-type monitor
**	Note:	channel access setup is done using dbr_*_ctrl structs.
**		This routine depends upon that fact and must be modified
**		if different.  (Ref op_comm.c )
**
**	Input :
**	    pdlc	: pointer to display context info 
**	    objptr	: pointer to dl_header(x,y,h,w)
**	    pPv_err	: ptr to error status for process var (conn,etc.)
**	    db_ftype	: native db type of the pv
**	    prti_mon	: pointer to rti monitor struct
**	    field_type	: native type of the data stored in the struct
**
**	Output :
**	    none 
**	Returns:	    
**	    nothing
**
**********************************************************************/

draw_meter(pdlc, objptr, prti_mon, db_ftype, pPv_err, expose_flag)
struct dl_context		*pdlc;
struct dl_object_hdr	*objptr;
struct rti_update_mon	*prti_mon;
short					db_ftype;
unsigned short			*pPv_err;
	{
    float	fval = (float)0.0;

    if (db_ftype == OLD_DBR_STRING)
		{
    	/* string, indicate invalid db_type */
		/*rti_drawtext(pdlc, objptr, db_ftype, prti_mon, pPv_err);*/
		if (expose_flag || !prti_mon->init_flag)
			{
			/* first time thru or expose - do something here to indicate icky-poo */
			*pPv_err = dmInvalid_dbtype;
			dmDisplayPvErr(objptr, *pPv_err, prti_mon->pgeom_comel, pdlc, prti_mon->pLast_val);
			prti_mon->init_flag = TRUE;
			}
		return 0;
		}

    /* draw dynamic value type object */
    fval = cvrt_to_float(prti_mon->pgeom_comel,db_ftype);
    dmSave_float(&prti_mon->pLast_val, fval);    /* save current value */

    if (!prti_mon->init_flag)
		{
		/* initializing, set up function calls, tables */
		get_hoprnlopr(prti_mon->pgeom_comel,db_ftype, &(prti_mon->hopr),&(prti_mon->lopr), &(prti_mon->g_type.meter.hdr.precision));
		if(!(prti_mon->hopr > prti_mon->lopr ))
			{
			prti_mon->hopr = (float)100.0;
			prti_mon->lopr = (float)0.0;
			}
		prti_mon->init_flag = TRUE;	
		init_meter (pdlc, objptr, prti_mon->lopr,prti_mon->hopr, 0, fval, &(prti_mon->g_type.meter), 
		  (short)((struct dl_meter *)prti_mon->pdl_mon)->label, ca_name(prti_mon->pgeom_comel->pChan));
		}
    else if(prti_mon->not_conn || expose_flag)
		{ 
		/* shown as not connected last time thru (new connection) or expose event    init graphics and draw */
    	/* draw and fill the rectangle */
    	/* do initial draw */
    	init_meter(pdlc, objptr, prti_mon->lopr,prti_mon->hopr, 1, fval, &(prti_mon->g_type.meter), 
		  (short)((struct dl_meter *)prti_mon->pdl_mon)->label, ca_name(prti_mon->pgeom_comel->pChan));
	    }
    else
		/* we're just updating - draw to indicate current value */
		update_meter(pdlc, fval, &(prti_mon->g_type.meter));
	}



/***********************************************************************
**
**	setup_rec_funcs
**	
**	Deb Kerstiens 8/90
**
**	Purpose :  Setup the appropriate routines to be called for the
**			bar type specified
**
**
**	Input :
**	    prti_mon	: pointer to rti monitor struct
**
**	Output :
**	    none 
**	Returns:	    
**	    nothing
**
**
**********************************************************************/

setup_indicator_funcs(prti_mon)
struct rti_update_mon *prti_mon;
	{
	int		init_indicatorL(), update_indicatorL();
	int		init_indicatorR(), update_indicatorR();
	int		init_indicatorU(), update_indicatorU();
	int		init_indicatorD(), update_indicatorD();
	struct dl_indicator	*pDl_indicator = prti_mon->pdl_mon;

	switch (pDl_indicator->direction)
	  {
	  case dl_pointer_left:		prti_mon->init_func = init_indicatorL;	prti_mon->upd_func = update_indicatorL;	break;
	  case dl_pointer_right:	prti_mon->init_func = init_indicatorR;	prti_mon->upd_func = update_indicatorR;	break;
	  case dl_pointer_up:		prti_mon->init_func = init_indicatorU;	prti_mon->upd_func = update_indicatorU;	break;
	  case dl_pointer_down:		prti_mon->init_func = init_indicatorD;	prti_mon->upd_func = update_indicatorD;	break;
	  default:					printf(opiStringTab[STR_Type_not_matched_in_init_ind]);							break;	
	  }
	return 0;
	}



/***********************************************************************
**
**	draw_indicator
**	
**	Deb Kerstiens 8/90
**
**	Purpose : Draw_indicator initializes and handles the graphic routines
**		that update a indicator-type monitor
**	Note:	channel access setup is done using dbr_*_ctrl structs.
**		This routine depends upon that fact and must be modified
**		if different.  (Ref op_comm.c )
**
**	Input :
**	    pdlc	: pointer to display context info 
**	    objptr	: pointer to dl_header(x,y,h,w)
**	    pPv_err	: ptr to error status for process var (conn,etc.)
**	    db_ftype	: native db type of the pv
**	    prti_mon	: pointer to rti monitor struct
**	    field_type	: native type of the data stored in the struct
**
**	Output :
**	    none 
**	Returns:	    
**	    nothing
**
**********************************************************************/


draw_indicator(pdlc, objptr, prti_mon, db_ftype, pPv_err, expose_flag)
struct dl_context		*pdlc;
struct dl_object_hdr	*objptr;
struct rti_update_mon	*prti_mon;
short					db_ftype;
unsigned short			*pPv_err;
	{
	float	fval = (float)0.0;

	db_ftype = ca_field_type(prti_mon->pgeom_comel->pChan);
	if (db_ftype == OLD_DBR_STRING)
		{
		/* string, indicate invalid db_type */
		/*rti_drawtext(pdlc, objptr, db_ftype, prti_mon, pPv_err);*/
	    if (expose_flag || !prti_mon->init_flag)
			{
			/* first time thru or expose - do something here to indicate icky-poo */
	    	*pPv_err = dmInvalid_dbtype;
	    	dmDisplayPvErr(objptr, *pPv_err, prti_mon->pgeom_comel, pdlc, prti_mon->pLast_val);
	    	prti_mon->init_flag = TRUE;
			}
		return 0;
		}

	/* draw dynamic value type object */
 	fval = cvrt_to_float(prti_mon->pgeom_comel, db_ftype);
	dmSave_float(&prti_mon->pLast_val, fval);		/* save current value */

	if (!prti_mon->init_flag)
		{
   		/* initializing, set up function calls, tables */
	    setup_indicator_funcs(prti_mon);
	    get_hoprnlopr(prti_mon->pgeom_comel,db_ftype, &(prti_mon->hopr),&(prti_mon->lopr), &(prti_mon->g_type.indicator.hdr.precision));
	    if (!(prti_mon->hopr > prti_mon->lopr))
			{
			/* if not valid limits, set default */
			prti_mon->hopr = (float)100.0;
			prti_mon->lopr = (float)0.0;
	    	}
	    prti_mon->init_flag = TRUE;	
	    (*(prti_mon->init_func))(pdlc, objptr, prti_mon->lopr,prti_mon->hopr, 0, fval, &(prti_mon->g_type.indicator), 
		  (short)((struct dl_indicator *)prti_mon->pdl_mon)->label, ca_name(prti_mon->pgeom_comel->pChan));
		}
	else if (prti_mon->not_conn || expose_flag)
		{ 
		/* shown as not connected last time thru (new connection) or expose event */
		/* init graphics and draw */
 	    /* draw and fill the rectangle */
	    (*(prti_mon->init_func))(pdlc, objptr, prti_mon->lopr,prti_mon->hopr, 1, fval, &(prti_mon->g_type.indicator), 
		  (short)((struct dl_indicator *)prti_mon->pdl_mon)->label, ca_name(prti_mon->pgeom_comel->pChan));
		}
	else
		{
		/* updating - draw to indicate current value */
	    fval = cvrt_to_float(prti_mon->pgeom_comel,db_ftype);
	    (*(prti_mon->upd_func)) (pdlc, fval, &(prti_mon->g_type.indicator));
		}
	}


#define DEF_FRMT	-1	/* flag to take default format type */


/***********************************************************************
**
**	rti_drawtext
**	
**	Deb Kerstiens 8/90
**
**	Purpose : This routine is used to draw text string if field type is
**		 DBR_STRING and the object type selected is not compatible 
**		(meter, bar, etc.--anything but text_update )
** 
**	
**
**	Input :
**	    pdlc	: pointer to display context info 
**	    objptr	: pointer to dl_header(x,y,h,w)
**	    db_ftype	: native db type of the pv
**	    prti_mon	: pointer to rti monitor struct
**	    pPv_err	: pointer to error statie od
**
**	Output :
**	    none
**	Returns:
**	    none
**
**********************************************************************/

rti_drawtext(pdlc, objptr, db_ftype, prti_mon, pPv_err)
struct dl_context		*pdlc;
struct dl_object_hdr	*objptr;
struct rti_update_mon	*prti_mon;
short					db_ftype;
unsigned short			*pPv_err;
	{
	int 			ycentered;
	int			nchars;
	XFontStruct  		*pfont;
	char 			display_str[MAX_CHARS];
	char 			str[MAX_CHARS];

    pfont = pdlc->pfont_table->height_to_pfinfo[min(objptr->height, MAXFONTHEIGHT)];	/* get proper font*/
    if (pfont >= (XFontStruct *)NULL)
		{
		XSetFont(pdlc->d, pdlc->gc, pfont->fid);
		/* set up to center text vertically */
		ycentered = ((objptr->height - pfont->max_bounds.ascent - pfont->max_bounds.descent)/2) + pfont->max_bounds.ascent;
		}
    else
		{
		/* no font -- print rectangle to indicate they will know something was intended */
		objptr->width  = objptr->width  ? objptr->width  : 1;	/* ensure some size */
		objptr->height = objptr->height ? objptr->height : 1;
		XFillRectangle(	pdlc->d, pdlc->w, pdlc->gc, objptr->x, objptr->y, objptr->width-1, objptr->height-1);
		return 0;
		}
    if (cvrt_to_str(prti_mon->pgeom_comel, db_ftype, display_str, DEF_FRMT) < 0)
		{
		dmDisplayPvErr(objptr, dmInvalid_data, prti_mon->pgeom_comel, pdlc, prti_mon->pLast_val);
		return 0;
    	}

    /* clip to object boundary and draw the text */
    nchars = objptr->width / pfont->max_bounds.width;
    sprintf(str,"%-*.*s", min(MAX_CHARS-1, nchars+1), min(((int)strlen(display_str)+1), nchars+1), display_str);
    tisDrawImageString(pdlc->d, pdlc->w, pdlc->gc, objptr->x, objptr->y + ycentered, str, strlen(str));
    return 0;
	}



/***********************************************************************
*
*	dmChkPvStat
*	
*	Deb Kerstiens 09/91
*
*	Purpose : check for connection, invalid data etc. so that an
*		error condition may be displayed if necessary		
*
*	Input :
*	    pComel	: ptr to pv information for the element
*	    init_flag	: indicates if the display el has been inited yet
*  	    pv_err	: last error state of this el (conn, not_conn,etc.)
*			 *note-- only needed for invalid range condition
*
*	Output :
*	    none
*	Returns
*	    none
*
**********************************************************************/
unsigned short dmChkPvStat(pComel, init_flag, last_err)
struct gsdl_com_record	*pComel;
unsigned short			init_flag;
unsigned short			last_err;
	{
	short pv_state;

    /* check for valid ca connection-- not valid 
	if marked as not connected or no data returned from monitor */

    if (pComel->pChan )
		{
/*printf ("pChan=%lx, name=%s, state=%d\n", (long)(pComel->pChan),
	ca_name(pComel->pChan),(int) ca_state(pComel->pChan)); */
		if((pv_state = ca_state(pComel->pChan))!= cs_conn)
	    	if( pv_state == cs_never_conn)
	    		return dmNever_connected;
	    	else
 	    		return dmDisconnected;
    	}
    else
		return dmNever_connected;

    /* unlikely state--we get here before ca callback happens */
    if ( !pComel->valid_data)
		return dmInvalid_data;

	/* we still have an invalid range */
    /* if !init_flag, we have a new connect and need to try again*/
    else if ((last_err==dmInvalid_range) && init_flag)
		return dmInvalid_range;
    else
		return dmConnection_okay;
	}


/***********************************************************************
*
*	dmDisplayPvErr
*	
*	Deb Kerstiens 09/91
*
*	Purpose :  display a white box with text to indicate that the
*		element is in error state, i.e.  not connected, disconnected,
*		invalid data, etc.		
*
*	Input :
*	    objptr	: ptr to display coordinate info
*	    pv_err	: id of error to display in text
*	    pComel	: ptr to pv information for the element
*	    pdlc	: ptr to get info for display, window 
*	    pLast_val	: ptr to last value for an object
*
*	Output :
*	    none
*	Returns
*	    none
*
**********************************************************************/

dmDisplayPvErr(objptr, pv_err, pComel, pdlc, pLast_val)
struct dl_object_hdr	*objptr;
unsigned short			pv_err;
struct gsdl_com_record	*pComel;
struct dl_context		*pdlc;
void					*pLast_val;
	{
    short			num;
    XRectangle		rec1, rec2, rec3; 
    char			err_str[50];
/*
	this code used to be
    static char		disconstr[] = opiStringTab[STR_No_val                      ];
    static char		nochanstr[] = opiStringTab[STR_No_chan_name                ];

	GCT 21-Nov-1994

	The above init code was changed to avoid compiler error after conversion.
*/
    static char		*disconstr;
    static char		*nochanstr;

    static GC		dmErrbox_gc = NULL;
    static GC		dmErrstr_gc = NULL;
/*
	GCT	21-Nov-1994

	This initialization was moved in here to avoid compiler errors after conversion.
*/
	disconstr = opiStringTab[STR_No_val                      ];
	nochanstr = opiStringTab[STR_No_chan_name                ];
	dmErrStrs[0] = opiStringTab[STR_OKAY                        ];
	dmErrStrs[1] = opiStringTab[STR_Not_connected               ];
	dmErrStrs[2] = opiStringTab[STR_Disconnected                ];
	dmErrStrs[3] = opiStringTab[STR_No_data_available           ];
	dmErrStrs[4] = opiStringTab[STR_Invalid_range               ];
	dmErrStrs[5] = opiStringTab[STR_Invalid_db_type             ];
	dmErrStrs[6] = 0;

    if (dmErrbox_gc == NULL)
		{
		XGCValues gc_box;
		
		gc_box.foreground = WhitePixel(pdlc->d, the_screen);
		dmErrbox_gc = XCreateGC(pdlc->d, pdlc->w, GCForeground, &gc_box);
    	}
    if (dmErrstr_gc == NULL)
		{
		XGCValues gc_str;
		
		gc_str.foreground = BlackPixel(pdlc->d, the_screen);
		dmErrstr_gc = XCreateGC(pdlc->d, pdlc->w, GCForeground, &gc_str);
    	}

    /* indicate not yet connected or lost connection */
    objptr->width  = objptr->width  ? objptr->width  : 1;	/* ensure some size */
    objptr->height = objptr->height ? objptr->height : 1;
    XFillRectangle(pdlc->d, pdlc->w, dmErrbox_gc, objptr->x, objptr->y, objptr->width-1,objptr->height-1);
    if (objptr->height > objptr->width)
		{
		rec1.x		= objptr->x;
		rec1.y		= objptr->y;
		rec1.width	= objptr->width/3;
		rec1.height	= objptr->height;
		if (rec1.width > pdlc->pfont_table->min_font_height)
			{
			rec2 = rec1;
		    rec2.x = rec2.x + rec2.width;
		    rec3 = rec2;
		    rec3.x += rec3.width;
		    num = 3;
			}
		else
			{
	    	rec1.width = objptr->width / 2;
	    	if (rec1.width > pdlc->pfont_table->min_font_height)
				{
	    	    rec2 = rec1;
	    	    rec2.x = rec2.x + rec2.width;
				num = 2;
	    		}
		    else
				{
				rec1.width = objptr->width;
				num = 1;
	    		}
			}
    	}
    else
		{
		rec1.x		= objptr->x;
		rec1.y		= objptr->y;
		rec1.width	= objptr->width;
		rec1.height	= objptr->height/3;
		if (rec1.height> pdlc->pfont_table->min_font_height)
			{
	    	rec2 = rec1;
	    	rec2.y = rec2.y + rec2.height;
	    	rec3 = rec2;
	    	rec3.y += rec3.height;
	    	num = 3;
			}
		else
			{
			rec1.height= objptr->height/ 2;
	    	if (rec1.height> pdlc->pfont_table->min_font_height)
				{
	        	rec2 = rec1;
	        	rec2.y = rec2.y + rec2.height;
				num = 2;
	    		}
		    else
				{
				rec1.height= objptr->height;
				num = 1;
				}
			}
    	}
    if (pComel->pChan != NULL)
		{
    	sprintf(err_str,"%s", ca_name(pComel->pChan),dmErrStrs[pv_err]);
      	draw_string(pdlc->d, pdlc->w, dmErrstr_gc, &rec1, pdlc->pfont_table, err_str, strlen(err_str), box_center);
      	if (num >= 2)
			draw_string(pdlc->d, pdlc->w, dmErrstr_gc, &rec2, pdlc->pfont_table, dmErrStrs[pv_err], strlen(dmErrStrs[pv_err]), box_center);
	   }
    else
     	draw_string(pdlc->d, pdlc->w, dmErrstr_gc, &rec1, pdlc->pfont_table, nochanstr, strlen(nochanstr), box_center);

    if (num == 3)
		{
    	if (pv_err == dmDisconnected)
			{
			if (pLast_val != NULL)
				{
				if (ca_field_type(pComel->pChan) == OLD_DBF_STRING)
					sprintf(err_str,"%-*s ", strlen(pLast_val), (char *)pLast_val);
	        	else
		    		sprintf(err_str,"%f ", *((float *)(pLast_val)));
  	    	    draw_string(pdlc->d, pdlc->w, dmErrstr_gc, &rec3, pdlc->pfont_table, err_str, strlen(err_str), box_center);
	    		}
	    	else
	    		draw_string(pdlc->d, pdlc->w, dmErrstr_gc, &rec3, pdlc->pfont_table, disconstr, strlen(disconstr), box_center);
		   	}
    	else
		    draw_string(pdlc->d, pdlc->w, dmErrstr_gc, &rec3, pdlc->pfont_table, disconstr, strlen(disconstr), box_center);
    	}
    return 0;
	}




/***********************************************************************
*
*	dmSave_float
*	
*	Deb Kerstiens 09/91
*
*	Purpose : If not done allocates storage for a float.  Saves the
*		current value as the "last_val".	
*
*	Input :
*	    pLast_float	: rti list el ptr to last value
*	    cur_val	: current value to save 
*
*	Output :
*	    sets pLast_float, *pLast_float
*	Returns
*	    none
*
**********************************************************************/


dmSave_float(ppLast_float, cur_val)
float	**ppLast_float;
float	cur_val;
	{
	if (*ppLast_float == NULL)
		*ppLast_float = (float *)tisCalloc(__FILE__, __LINE__, sizeof (float), 1);
	**ppLast_float = (float)cur_val;
	return 0;
	}




/***********************************************************************
*
*	dmSave_str
*	
*	Deb Kerstiens 09/91
*
*	Purpose : If not done allocates storage for a string.  Saves the
*		current value as the "last_val".	
*
*	Input :
*	    pLast_str	: rti list el ptr to last value
*	    cur_val	: current value to save 
*
*	Output :
*	    sets pLast_str, *pLast_str
*	Returns
*	    none
*
**********************************************************************/
dmSave_str(ppLast_str, cur_str)
char	**ppLast_str;
char	*cur_str;
	{
	if (*ppLast_str == NULL)
		*ppLast_str = (char *)tisCalloc(__FILE__, __LINE__, MAX_STRING_SIZE, sizeof(char));

	if (strlen(cur_str) > MAX_STRING_SIZE - 1)
		{
		strncpy(*ppLast_str, cur_str, MAX_STRING_SIZE - 1);
		*ppLast_str[MAX_STRING_SIZE - 1] = '\0';
		}
	else
		strcpy( *ppLast_str, cur_str);
	return 0;
	}




/***********************************************************************
**
**	dmGet_ctrlunits
**	
**	Deb Kerstiens 10/91
**
**	Purpose : This routine fills in the string representing units for
**			the process variable referenced 
**	Note:	channel access setup is done using dbr_*_ctrl structs.
**		This routine depends upon that fact and must be modified
**		if different.   
**	
**
**	Input :
**	    pcom_el	: pointer to a com record structure 
**	    field_type	: native type of the data stored in the struct
**	    unit_str	: string representing units
**	    
**	Output :
**	    unit_str	: fills in units from db struct
**	Returns:
**	    none
**
**********************************************************************/


dmGet_ctrlunits(pcom_el,field_type, string)
struct gsdl_com_record	*pcom_el;
short					field_type;
char					*string;
	{
    char	unit_str[db_units_dim+1];
	
	switch (field_type)
	  {
	  default:
	  case OLD_DBF_ENUM:
	  case OLD_DBF_STRING:	/* do nothing --no units for this type */	return 0;
	  case OLD_DBF_FLOAT:	sprintf(unit_str, " %s", ((struct dbr_ctrl_float *)pcom_el->pData_struct)->units);	break;
	  case OLD_DBF_DOUBLE:	sprintf(unit_str, " %s", ((struct dbr_ctrl_double *)pcom_el->pData_struct)->units);	break;
	  case OLD_DBF_SHORT:	sprintf(unit_str, " %s", ((struct dbr_ctrl_short *)pcom_el->pData_struct)->units);	break;
	  case OLD_DBF_CHAR:	sprintf(unit_str, " %s", ((struct dbr_ctrl_char *)pcom_el->pData_struct)->units);	break;
	  }			
	strcat(string, unit_str);
	return 0;
	}



