#include "opiStringDefs.h"
extern	char	*opiStringTab[];
/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *
 *	E L E M E N T . C
 *
 *	Author	Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *
 *	display list element specific stuff
 */
#ifdef AXP
#include <stdlib.h>
#endif
#ifdef TOKENPTRS
#include <tokenSupport.h>
#endif

#include <dldef.h>
#include <dlinfo.h>
#include <dlcvrt.h>
#include <edd.h>
#include <db_access.h>

static struct sinfo	*find_break_by_class();
static struct sinfo	*find_break();
static void 		close_retained_structure();

void *tisMalloc();
void *tisCalloc();
void tisFree();


/*
 *	CLOSE_ELEMENT
 */
void close_element(pdue, peue, in_dl, retain_strings)
due	*pdue;
eue	*peue;
int	in_dl;
int	retain_strings;
	{
	eue_sel 			*psel = PEUE_TO_PSEL(peue);
	struct sinfo		*psinfo;
	enum dl_elements	type = peue->elem.dl_anygo.element.type;

	/*
	 * remove from the select list
	 */
	if (peue->sel && in_dl)
		{
		lstDelete(&pdue->selections, (struct node *)psel);
		peue->sel = FALSE;
		}

	if (in_dl)
		lstDelete(&pdue->dl, &peue->node);

	if (peue->pg)
		free_gizmo(peue->pg);

	pdue->dl_mdt = TRUE;
	psinfo = dl_elements_ginfo[peue->elem.dl_anygo.element.type].sinfo;
	}



/*
 *	REOPEN_ELEMENT
 *
 *	take strings from edd'd string table and place
 *	them back in the display list's string tables
 *
 *	(used by the undo facility)
 */
void reopen_element(pdue, peue)
due		*pdue;
eue		*peue;
	{
	eue_sel 			*psel	= PEUE_TO_PSEL(peue);
	enum dl_elements	type	= peue->elem.dl_anygo.element.type;
	struct sinfo		*psinfo	= dl_elements_ginfo[type].sinfo;

	pdue->dl_mdt = TRUE;
	open_structure(psinfo, &peue->elem.dl_anygo.element, pdue, TRUE);

	/* add auxillary stuff if graphic
	 */
	if (dl_element_class[type] == dl_class_graphic)
		{
		open_structure(dl_sta_attr_sinfo, &peue->sta, pdue, TRUE);
		if (dynamic_graphical_element(peue))
			open_structure(dl_dyn_param_sinfo, &peue->dyn.param, pdue, TRUE);
		open_structure(dl_dyn_mod_sinfo, &peue->dyn.mod, pdue, TRUE);
		}
	}


/*
 *	CREATE_ELEMENT
 */
eue *create_element(pdue, pgo, elem_size)
due				*pdue;
struct dl_anygo	*pgo;
unsigned 		elem_size;
	{
	eue_sel		*psel;

	/*psel = (eue_sel *) calloc(1, sizeof(*psel));*/
	psel = (eue_sel *) tisCalloc(__FILE__, __LINE__, 1, sizeof(*psel));
	assert(psel);

#if defined(SOLARIS) || defined(WIN32)
	memcpy(&psel->eue.elem, pgo, elem_size);
#else
	bcopy(pgo, &psel->eue.elem, elem_size);
#endif
	/*
	 * Insert in the display list
	 */
	lstAdd(&pdue->dl, &psel->eue.node);
	pdue->dl_mdt = TRUE;

	return &psel->eue;
	}




/*
 *	DYNAMIC_GRAPHICAL_ELEMENT()
 *
 *	test to see if this is a dynamic graphical element
 */
dynamic_graphical_element(peue)
eue	*peue;
	{
	return (peue->dyn.mod.vis != dl_vismod_static) || (peue->dyn.mod.clrmod != dl_clrmod_static);
	}


/*
 *	TRANSFORM_ELEMENT
 *
 *	(change the type of an element)
 */
void transform_element(pdue, type, peue)
due					*pdue;
enum dl_elements	type;
eue					*peue;
	{
	unsigned			dl_size_new	= WDRND(dl_elements_ginfo[type].size)<<1;
	enum dl_elements	oldtype		= peue->elem.dl_anygo.element.type;
	struct sinfo		*psinfo;
	struct sinfo		*psinfo_after_header;

	if (type == oldtype)
		return;

	psinfo = dl_elements_ginfo[oldtype].sinfo;
	if (dl_element_class[type] != dl_element_class[oldtype])
		{
		if (dl_element_class[type] == dl_class_graphic)
			{
#if defined(SOLARIS) || defined(WIN32)
			memset(&peue->sta, 0, sizeof(peue->sta));
			memset(&peue->dyn, 0, sizeof(peue->dyn));
#else
			bzero((char *)&peue->sta, sizeof(peue->sta));
			bzero((char *)&peue->dyn, sizeof(peue->dyn));
#endif
			open_structure(dl_sta_attr_sinfo, &peue->sta, pdue, FALSE);
			open_structure(dl_dyn_mod_sinfo, &peue->dyn.mod, pdue, FALSE);
			if (dynamic_graphical_element(peue))
				open_structure(dl_dyn_param_sinfo, &peue->dyn.param, pdue, FALSE);
			}
		psinfo_after_header = find_break(psinfo, dl_anygo_sinfo);
		}
	else
		psinfo_after_header = find_break_by_class(type, psinfo);

	if (psinfo_after_header)
		/*
		 * make certain object extensions (not shared) default to nill
		 */
#if defined(SOLARIS) || defined(WIN32)
		memset (((char *) &peue->elem) + psinfo_after_header->offset, 0, sizeof(peue->elem) - psinfo_after_header->offset);
#else
		bzero(((char *) &peue->elem) + psinfo_after_header->offset, sizeof(peue->elem) - psinfo_after_header->offset);
#endif
	peue->elem.dl_anygo.element.type = type;
	if (type == dl_line)
		{
		peue->elem.dl_line.npoints = 2;
		peue->elem.dl_line.point[0].x    = peue->elem.dl_line.object.x;
		peue->elem.dl_line.point[0].y    = peue->elem.dl_line.object.y;
		peue->elem.dl_line.point[1].x    = peue->elem.dl_line.object.x + peue->elem.dl_line.object.width;
		peue->elem.dl_line.point[1].y    = peue->elem.dl_line.object.y + peue->elem.dl_line.object.height;
		peue->elem.dl_line.orgpoint[0].x = 0;
		peue->elem.dl_line.orgpoint[0].y = 0;
		peue->elem.dl_line.orgpoint[1].x = 1000;
		peue->elem.dl_line.orgpoint[1].y = 1000;
		}

	psinfo = dl_elements_ginfo[type].sinfo;
	if (dl_element_class[type] != dl_element_class[oldtype])
		psinfo_after_header = find_break(psinfo, dl_anygo_sinfo);
	else
		psinfo_after_header = find_break_by_class(type, psinfo);

	if (psinfo_after_header)
		open_structure(psinfo_after_header, &peue->elem, pdue, FALSE);

	/*
	 * fix the element word count and update the total words count
	 */
	pdue->dl_hdr.nwords_dspy -= peue->elem.dl_anygo.element.nwords;
	peue->elem.dl_anygo.element.nwords = WDRND(dl_size_new);
	pdue->dl_hdr.nwords_dspy += peue->elem.dl_anygo.element.nwords;
	}


/*
 *	FIND BREAK BY CLASS
 *	find the structure info pointer beginning after a common header
 */
static struct sinfo *find_break_by_class(type, psinfo)
enum dl_elements	type;
struct sinfo		*psinfo;
	{
	switch (dl_element_class[type])
	  {
	  case dl_class_graphic:	return find_break(psinfo, dl_anygo_sinfo);
	  case dl_class_control:	return find_break(psinfo, dl_anyco_sinfo);
	  case dl_class_monitor:	return find_break(psinfo, dl_anymo_sinfo);
	  case dl_class_plot:		return find_break(psinfo, dl_anypo_sinfo);
	  default:
	  case dl_class_related_display:
	  case dl_class_execute_script:
	  case dl_class_kill_display:
		/*
		 * Invalid condition since there is only one member of these classes
		 */
		assert(FALSE);
	  }
	}


/*
 *	FIND BREAK
 *	find the structure info pointer beginning after a common header
 */
static struct sinfo *find_break(psinfo, psinfo_hdr)
struct sinfo	*psinfo;
struct sinfo	*psinfo_hdr;
	{
	for ( ; psinfo->name; psinfo++, psinfo_hdr++)
		{
		if (!psinfo_hdr->name)
			return psinfo;

		assert(psinfo->type == psinfo_hdr->type);
		assert(psinfo->nelements == psinfo_hdr->nelements);
		assert(psinfo->offset == psinfo_hdr->offset);
		}
	return NULL;
	}




/*
 *	ASSIGN_FINAL_INDEX
 */
struct assign_final_index_args
	{
	void	*pold;
	void	*pnew;
	due		*pdue;
	};

static void assign_final_index_action(pargs, psinfo, offset)
struct assign_final_index_args	*pargs;
struct sinfo					*psinfo;
register unsigned				offset;
	{
#ifndef TOKENPTRS
	char	**pnew = (char **)((char *)(pargs->pnew) + offset);
	char	**pold = (char **)((char *)(pargs->pold) + offset);
#else
	int	*pnew = (int *)((char *)(pargs->pnew) + offset);
	int	*pold = (int *)((char *)(pargs->pold) + offset);
#endif

	switch (psinfo->type)
	  {
#ifndef TOKENPTRS
	  case dl_string_index:	*(int *)pnew = stringGetIndexAndIncCount(&pargs->pdue->st, *pold);	break;
	  case dl_pc_index:	*(int *)pnew = stringGetIndexAndIncCount(&pargs->pdue->ct, *pold);	break;
#else
	  case dl_string_index:	*pnew = stringGetIndexAndIncCount(&pargs->pdue->st,
						 dmPointer64fromToken (*pold));
				break;
	  case dl_pc_index:	*pnew = stringGetIndexAndIncCount(&pargs->pdue->ct,
						 dmPointer64fromToken (*pold));
				break;
#endif
	  }
	}

void assign_final_index(pdue, psinfo, pold, pnew)
due				*pdue;
struct sinfo	*psinfo;
void			*pold;
void			*pnew;
	{
	struct assign_final_index_args	args;

	args.pold	= pold;
	args.pnew	= pnew;
	args.pdue	= pdue;
	traverse_psinfo(psinfo, 0, assign_final_index_action, &args);
	}


/*
 *	REMAP_STRING_PTRS
 */
struct remap_string_ptrs_args
	{
	void	*pstructure;
	due		*pdue;
	};

static void remap_string_ptrs_action(pargs, psinfo, offset)
struct remap_string_ptrs_args	*pargs;
struct sinfo					*psinfo;
register unsigned				offset;
	{
#ifndef TOKENPTRS
	char	**pstructure = (char **)((char *)(pargs->pstructure) + offset);
#else
	int	*pstructure = (int *)((char *)(pargs->pstructure) + offset);
#endif

	switch (psinfo->type)
	  {
#ifndef TOKENPTRS
	  case dl_string_index:	*pstructure = stringGetPtr(&pargs->pdue->st, *pstructure);	break;
	  case dl_pc_index:	*pstructure = stringGetPtr(&pargs->pdue->ct, *pstructure);	break;
#else
	  case dl_string_index:	*pstructure = dmTokenFromPointer64 (stringGetPtr(&pargs->pdue->st,
								 dmPointer64fromToken(*pstructure)));
				break;
	  case dl_pc_index:	*pstructure = dmTokenFromPointer64 (stringGetPtr(&pargs->pdue->ct,
								 dmPointer64fromToken(*pstructure)));
				break;
#endif
	  }
	}

void remap_string_ptrs(pdue, psinfo, pstructure)
due				*pdue;
struct sinfo	*psinfo;
void			*pstructure;
	{
	struct remap_string_ptrs_args	args;

	args.pstructure	= pstructure;
	args.pdue		= pdue;
	traverse_psinfo(psinfo, 0, remap_string_ptrs_action, &args);
	}


/*
 *	CHECK_STRING_OFFSETS
 */
static void check_string_offsets_action(pargs, psinfo, offset)
void			*pargs;
struct sinfo	*psinfo;
unsigned		offset;
	{
	switch (psinfo->type)
	  {
	  case dl_string_index:
	  case dl_pc_index:
		if ((offset % 4) != 0)
			printf(opiStringTab[STR_bad_string_offset_Hd_in_stru], offset, psinfo->name, psinfo->description);
		break;
	  }
	}

void check_string_offsets(psinfo)
struct sinfo	*psinfo;
	{
	traverse_psinfo(psinfo, 0, check_string_offsets_action, 0);
	}


/*
 *	CONVERT_STRING_INDEX_TO_PTR
 */
struct convert_string_index_to_ptr_args
	{
	char	*pstructure;
	char	**pstr;
	char	**ppc;
	};

static void convert_string_index_to_ptr_action(pargs, psinfo, offset)
struct convert_string_index_to_ptr_args	*pargs;
struct sinfo							*psinfo;
unsigned								offset;
	{
#ifndef TOKENPTRS
	char	**pfield	= (char **)(pargs->pstructure + offset);
#else
	int	*pfield	= (int *)(pargs->pstructure + offset);
#endif

	switch (psinfo->type)
	  {
#ifndef TOKENPTRS
	  case dl_string_index:	*pfield = pargs->pstr[*(int *)pfield];	break;
	  case dl_pc_index:	*pfield = pargs->ppc[*(int *)pfield];	break;
#else
	  case dl_string_index:	*pfield = dmTokenFromPointer64 (pargs->pstr[*(int *)pfield]);
				break;
	  case dl_pc_index:	*pfield = dmTokenFromPointer64 (pargs->ppc[*(int *)pfield]);
				break;
#endif
	  }
	}

void convert_string_index_to_ptr(psinfo, pstructure, ppc, pstr)
struct sinfo	*psinfo;
char			*pstructure;
char			**pstr;
char			**ppc;
	{
	struct convert_string_index_to_ptr_args	args;

	args.pstructure	= pstructure;
	args.pstr		= pstr;
	args.ppc		= ppc;
	traverse_psinfo(psinfo, 0, convert_string_index_to_ptr_action, &args);
	}


/*
 *	STRING_INDEX_TO_DISPLAY_STR
 *	
 *	inefficient, but allows static draw routine to be shared
 *	between the editor and runtime codes
 */
int string_index_to_display_str(pstring_table, textix, ppdisplay_str, pdisplay_str_length)
char		**pstring_table;
#if !defined(TOKENPTRS) && defined(AXP)
char		*textix;
#else
unsigned	textix;
#endif
char		**ppdisplay_str;
unsigned 	*pdisplay_str_length;
	{
#ifndef TOKENPTRS
	return stringName(pstring_table, textix, ppdisplay_str, pdisplay_str_length);
#else
	/* the string table param is unused, but the routine expects a list */
	return stringName((LIST *)pstring_table, dmPointer64fromToken(textix), ppdisplay_str, pdisplay_str_length);
#endif
	}


/*
 *	COMPUTE_BOUNDS
 *	(compute the bounds of an object taking into account that some graphic
 *	objects may have a line width other than one pixel wide)
 */
void compute_bounds(peue, pbounds)
eue			*peue;
XRectangle	*pbounds;
	{
	DLOBJ_TO_XREC(peue->elem.dl_anygo.object, *pbounds);

	if (dl_element_class[peue->elem.dl_anygo.element.type] == dl_class_graphic)
		{
		if(peue->elem.dl_anygo.element.type == dl_text)
			return;

		if (peue->sta.fill == dl_fill_outline)
			{
			int	lwd2	= peue->sta.width >> 1;
			int lw		= peue->sta.width * 1;

			if (lw == 0)
				lw++;		/* zero pixel width lines are still one pixel wide */

			pbounds->x		-= lwd2;
			pbounds->y		-= lwd2;
			pbounds->width	+= lw;
			pbounds->height	+= lw;
			}
		}
	}


/*
 *	OPEN_STRUCTURE
 */
struct open_structure_args
	{
	due		*pdue;
	char	*pstructure;
	};
	

void open_structure(psinfo, pstructure, pdue)
struct sinfo	*psinfo;
char			*pstructure;
due				*pdue;
	{
	static void open_structure_action();
	struct open_structure_args	args;

	args.pstructure	= pstructure;
	args.pdue	= pdue;
	traverse_psinfo(psinfo, 0, open_structure_action, &args);
	}


static void open_structure_action(pargs, psinfo, offset, elementno)
struct open_structure_args	*pargs;
struct sinfo				*psinfo;
register unsigned			offset;
unsigned					elementno;
	{
	char								*pfield;
	long								value;
	enum edd_valuator_constraint_type	constraint;
	int									match;
	int									status;
	char								*pinit;
	unsigned							length;

	pfield = pargs->pstructure + offset;
		
	switch (psinfo->type)
	  {
	  case dl_string_index:
		if ((strcmp(psinfo->name, opiStringTab[STR_clrrule                     ]) == 0) && (pargs->pstructure != (char *)&pargs->pdue->dl_hdr))
#ifndef TOKENPTRS
			stringName(&pargs->pdue->st, pargs->pdue->dl_hdr.clrrule, &pinit, &length);
#else
			stringName(&pargs->pdue->st, dmPointer64fromToken(pargs->pdue->dl_hdr.clrrule), &pinit, &length);
#endif
		else if ((strcmp(psinfo->name, opiStringTab[STR_clrargs                     ]) == 0))
			pinit = opiStringTab[STR_CCGname                     ];
		else
			pinit = "";

#ifndef TOKENPTRS
		*(char **)pfield = stringInsert(&pargs->pdue->st, pinit);
#else
		*(int *)pfield = dmTokenFromPointer64 (stringInsert(&pargs->pdue->st, pinit));
#endif
		break;

	  case dl_pc_index:
		if (strcmp(psinfo->name, opiStringTab[STR_rulechan                    ]) == 0)
			pinit = "";
		else
			/*
			 * use the default process variable stored as a string in the header
			 */
#ifndef TOKENPTRS
			stringName(&pargs->pdue->ct, pargs->pdue->dl_hdr.pv, &pinit, &length);
#else
			stringName(&pargs->pdue->ct, dmPointer64fromToken(pargs->pdue->dl_hdr.pv), &pinit, &length);
#endif

#ifndef TOKENPTRS
		*(char **)pfield = stringInsert(&pargs->pdue->ct, pinit);
#else
		*(int *)pfield = dmTokenFromPointer64 (stringInsert(&pargs->pdue->ct, pinit));
#endif
		break;

	  case dl_color_index:
		*(char *)pfield = pargs->pdue->dl_hdr.clr;
		break;

	  case dl_bckgnd_color_index:
		*(char *)pfield = pargs->pdue->dl_hdr.bclr;
		break;
	  }

	match = FALSE;
	switch (psinfo->input)
	  {
	  case edd_menu:
	  case edd_button:
		if (psinfo->pinfo == dl_dynamic_color_mod_einfo)
			{
			value = pargs->pdue->dl_hdr.clrmod;
			match = TRUE;
			}
		else if (psinfo->pinfo == dl_fill_style_einfo)
			{
			value = pargs->pdue->dl_hdr.fill;
			match = TRUE;
			}
		else if (psinfo->pinfo == dl_edge_style_einfo){
			value = pargs->pdue->dl_hdr.style;
			match = TRUE;
			}
		break;

	  case edd_valuator:
		constraint = (enum edd_valuator_constraint_type) psinfo->pinfo;
		if (constraint == edd_line_width)
			{
			value = pargs->pdue->dl_hdr.width;
			match = TRUE;
			}
		break;
	  }

	if (match)
		{
		status = long_to_dl(pargs->pstructure, psinfo, value, elementno);
		assert(status >= 0);
		}
	}



/*
 *	CHANGE_ELEMENT()
 */
void
change_element(pdue, peue, pneweue)
due	*pdue;
eue	*peue;
eue	*pneweue;
	{
	enum dl_elements	newtype;
	enum dl_elements	oldtype;
	int					old_dyn;
	int					new_dyn;
	gizmo				*pg;
	int					x,y;

	pg = peue->pg;
	if (pg)
		{
		x = peue->pg->obj.x;
		y = peue->pg->obj.y;
		}

	old_dyn = dynamic_graphical_element(peue);
	new_dyn = dynamic_graphical_element(pneweue);

	newtype = pneweue->elem.dl_anygo.element.type;
	oldtype = peue->elem.dl_anygo.element.type;

	/*
	 * special post processing for type change
	 */
	if (newtype != oldtype)
		{
		int type_ok	= FALSE;

		if (pg)
			{
			/*
			 *	clear the old gizmo before the transformation to avoid inconsistent data structures (The free might call back)
			 *	NOTE: Clear here could cause a distressing period when neither the old or new form are on the screen!
			 */
			free_gizmo(pg);
			peue->pg = NULL;
			}
		switch (dl_element_class[newtype])
		  {
		  case dl_class_graphic:		type_ok = (newtype != dl_anygo);	break;
		  case dl_class_control:		type_ok = (newtype != dl_anyco);	break;
		  case dl_class_monitor:		type_ok = (newtype != dl_anymo);	break;
		  case dl_class_plot:			type_ok = (newtype != dl_anypo);	break;
		  case dl_class_related_display:
		  case dl_class_execute_script:
		  case dl_class_kill_display:	type_ok = TRUE;						break;
		  default:															break;
		  }
		if (!type_ok)
			{
			sprintf(pdue->pmsg, opiStringTab[STR_Sorry_cant_transform_element]);
			redraw_msg(pdue);
			return;
			}
		/*
		 * transform element (default new fields to nill), pop up a new form, and delete old form
		 */
		transform_element(pdue, newtype, peue);
		}
	else if (new_dyn != old_dyn)
		{
		/*
		 * special post processing for dyn graphical change
		 */
		if (pg)
			{
			/* see note above */
			free_gizmo(pg);
			peue->pg = NULL;
			}
		peue->dyn.mod = pneweue->dyn.mod;
		if (new_dyn)
			{
			peue->dyn.mod = pneweue->dyn.mod;
#if defined(SOLARIS) || defined(WIN32)
			memset(&peue->dyn.param, 0, sizeof(peue->dyn.param));
#else
			bzero((char *)&peue->dyn.param, sizeof(peue->dyn.param));
#endif
			open_structure(dl_dyn_param_sinfo, &peue->dyn.param, pdue, FALSE);
			}
		}
	else
		*peue = *pneweue;

	/*
	 * 	If the gizmo was deleted above
	 */
	if (pg != peue->pg)
		modify_element(pdue, peue, x, y);
	}
	
