#include "opiStringDefs.h"
extern	char	*opiStringTab[];
/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *	S A V E . C
 *
 *	Author	Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *	save a display file
 *
 *
 *	History
 *	-------
 *	.01 joh 012391	Added save to a text file (report) popup.
 *	.02 joh	040991	Fixed problem where dyn attr write
 *			was comparing edit string index with
 *			compacted string index.
 *
 * 	Improvements
 * 	------------
 * 	01	In the display file write sub take out sections which 
 *		are io dependent or rewrite to use UNIX STDIO only
 */

#include <stdio.h>
#ifdef NeXT
#include <types.h>
#endif
#ifdef WIN32
#include	<io.h>
#include <direct.h>
#else
#include 	<file.h>
#endif WIN32
#include <fcntl.h>
#if (defined(SOLARIS) || defined(WIN32))
#include	<string.h>
#else
#include	<strings.h>
#endif
#include <errno.h>
#ifdef WIN32
#include <stdlib.h>
int		errno;
#else
extern		errno;
#endif WIN32

#ifdef TOKENPTRS
#include <tokenSupport.h>
#endif

#include <X11/Xlib.h>
#ifdef WIN32
#include <X11/XlibXtra.h>
#endif WIN32
#include <X11/Xutil.h>

#include <edd.h>
#include <dlinfo.h>
#include <db_access.h>

extern void *tisMalloc();
extern void *tisCalloc();
extern void tisFree();

static int	write_sta_attr();
static int	write_dyn_attr();
static int	make_segment();
static void	save_popup_finish();
static void	save_text_popup_finish();
static		write_class();


/*
 *	GM_SAVE()
 *	(see edd_menus.h)
 */
void gm_save(rename_flag, pdue, pbe)
int				rename_flag;
due				*pdue;
XButtonEvent	*pbe;
	{
	int			status;
	char		pname[sizeof(pdue->pname)+sizeof(pdue->ppath)];
	char		pbackup[sizeof(pdue->pname)+sizeof(pdue->ppath)+1];
	char		ptmpname[sizeof(pdue->pname)+sizeof(pdue->ppath)+sizeof(TMP_EXTENSION)+1];
	enum 		choices{cancel, replace, terminate, nchoices};
	static char	*pmsgs[3];
	char 		*pchoices[nchoices];
	XRectangle 	obj;
	int			namesize;

	if (rename_flag)
		{
		int choice;

		status = edd_path_cat(pdue, pdue->psnpath, pdue->psnname, pname, sizeof(pname));
		if (status == ERROR)
			return;

		pmsgs[0] = opiStringTab[STR_That_file_exists            ];
		pmsgs[1] = opiStringTab[STR_Do_you_want_to_replace_its_c];
		pmsgs[2] = 0;
		/*
		 * are they replacing an important file
		 */
#ifdef WIN32
		/* this touches the file! */
		status = _open(pname, _O_BINARY|_O_CREAT|_O_EXCL|_O_RDONLY);
#else
		status = open(pname, O_CREAT|O_EXCL|O_RDONLY);
#endif WIN32
		if (status < 0)
			{
			if (errno == EEXIST)
				{
				pchoices[replace]	= opiStringTab[STR_replace                     ];
				pchoices[cancel]	= opiStringTab[STR_cancel                      ];
				pchoices[terminate]	= 0;

				obj.x		= pbe->x_root;
				obj.y		= pbe->y_root;
				obj.width	= (pdue->gizmo_width<<1);
				obj.height	= (pdue->gizmo_height<<1);

				choice = perform_dialog(
					pdue->s, 
					pbe->x_root, 
					pbe->y_root, 
					pdue->gizmo_height, 
					pmsgs, 
					pchoices, 
					pdue->gizmo_height);
				if (choice == cancel)
					return;
				}
			}
		else
			_close (status);
		}
	else
		{
		if (pdue->pname[0] == 0)
			{
			gm_save_to_new_file(NULL, pdue, pbe);
			return;
			}
		status = edd_path_cat(pdue, pdue->ppath, pdue->pname, pname, sizeof(pname));
		if (status == ERROR)
			return;
		}

	namesize = strlen(pname);
	if (namesize+sizeof(TMP_EXTENSION)+1>sizeof(ptmpname))
		{
		sprintf(pdue->pmsg, opiStringTab[STR_Name_to_large_to_create_a_tm]);
		redraw_msg(pdue);
		return;
		}

	/*
	 * write to a temporary file so that if it fails or they 
	 * interrupt the old file will still exist
	 */
	strncpy(ptmpname, pname, sizeof(ptmpname)-1);
	strncat(ptmpname, TMP_EXTENSION, sizeof(ptmpname)-namesize-1);
	status = save_display(pdue, ptmpname);
	if (status < 0)
		return;

	/*
	 * make a backup copy
	 */
#ifdef WIN32
	_unlink (strcat(strcpy(pbackup, pname), "%"));
#endif WIN32
	rename(pname, pbackup);

	/*
	 * now that we have success rename the temp file to their file
	 */
	rename(ptmpname, pname);
	sprintf(pdue->pmsg, opiStringTab[STR_Wrote_to_Hs                 ], pname);
	redraw_msg(pdue);

	pdue->dl_mdt = FALSE;
	if (rename_flag)
		{
		strcpy(pdue->pname, pdue->psnname);
		strcpy(pdue->ppath, pdue->psnpath);
		}
	}


/*
 * 	SAVE_DISPLAY()
 *	save a display to a named file
 */
long save_display(pdue, pname)
due		*pdue;
char	*pname;
	{
	struct dl_hdr	tmphdr;
	edd_stream		stream;
	int				status;

	stream.io_info.pfile = fopen(pname, "w+b");
	if (!stream.io_info.pfile)
		{
		sprintf(pdue->pmsg, opiStringTab[STR_File_access_problems_during_], pname);
		redraw_msg(pdue);
		return ERROR;
		}
	stream.type = use_stdio;
	stream.nbytes = 0;

	/*
	 * skip past the display list header
	 */
#ifdef WIN32
	status = fseek(stream.io_info.pfile, sizeof(pdue->dl_hdr), SEEK_SET);
#else
	status = fseek(stream.io_info.pfile, sizeof(pdue->dl_hdr), L_SET);
#endif WIN32
	assert(status >= 0);
	stream.nbytes += sizeof(pdue->dl_hdr);

	status = make_segment(pdue, &stream);
	if (status < 0)
		{
error_exit:
		sprintf(pdue->pmsg, opiStringTab[STR_Unable_to_write_out_display_]);
		redraw_msg(pdue);
		fclose(stream.io_info.pfile);
#ifdef WIN32
		_unlink(pname);
#else
		unlink(pname);
#endif WIN32
		return ERROR;
		}

	/*
	 * Write out the display list header (after we know the number of words in the display)
	 */
	rewind(stream.io_info.pfile);
	tmphdr = pdue->dl_hdr;
	assign_final_index(pdue, dl_hdr_sinfo, &pdue->dl_hdr, &tmphdr);
	tmphdr.nstr	= stringCount(&pdue->st);
	tmphdr.npc	= stringCount(&pdue->ct);
	status = fwrite(&tmphdr, sizeof(tmphdr), 1, stream.io_info.pfile);
	if (status != 1)
		goto error_exit;

	fclose(stream.io_info.pfile);
	return OK;
	}


/*
 *	make_segment()
 */
int	counting_strings;

static int make_segment(pdue, pstream)
due			*pdue;
edd_stream	*pstream;
	{
	eue							*peue;
	struct dl_basic_attribute	sta_attr;
	struct dl_dynamic_attribute	dyn_attr;
	unsigned					ndyn;
	unsigned					nsta;
	char						*pstring;
	unsigned					length;
	int							display_includes_color_map;

#ifndef TOKENPTRS
	stringName(&pdue->st, pdue->dl_hdr.cmap, &pstring, &length);
#else
	stringName(&pdue->st, dmPointer64fromToken(pdue->dl_hdr.cmap), &pstring, &length);
#endif
	display_includes_color_map = (length == 0);

	/*
	 * Write out the process connections and strings
	 *	NOTE:	as the segments are written, assign_final_index() will count the string usage, therefore
	 *			these writes (with zeroed counts) serve to reserve space, they are rewritten later with valid counts
	 */
	counting_strings = 1;
	stringZeroCounts(&pdue->ct);
	assign_clrrule_channels(pdue);
	pdue->dl_hdr.opc = WDRND(ftell(pstream->io_info.pfile));
	stringTraverse(&pdue->ct, write_pc, pstream);

	stringZeroCounts(&pdue->st);
	pdue->dl_hdr.ostr = WDRND(ftell(pstream->io_info.pfile));
	stringTraverse(&pdue->st, write_string, pstream);
	counting_strings = 0;

	/*
	 * Write out the color table
	 */
	pdue->dl_hdr.ocmap = WDRND(ftell(pstream->io_info.pfile));
	if (display_includes_color_map)
		if (write_cmap(pdue->dlc.pcmap, pstream) < 0)
			return ERROR;

	pdue->dl_hdr.nwords_cmap = WDRND(ftell(pstream->io_info.pfile))-pdue->dl_hdr.ocmap;

	/*
	 * Write out the color rule table
	 */
	pdue->dl_hdr.ocrules = WDRND(ftell(pstream->io_info.pfile));
	if (display_includes_color_map)
		{
		struct color_rules	tmpcrules;
		int					size	= pdue->dlc.pcrules->nrules * sizeof(struct dl_color_rule);
		unsigned int		i;

		tmpcrules.nrules		= pdue->dlc.pcrules->nrules;
		tmpcrules.pdl_crules	= (struct dl_color_rule *)
			tisCalloc(__FILE__, __LINE__, 1, size);
		tmpcrules.ppopg			= 0;
		assert(tmpcrules.pdl_crules != NULL);

		memcpy(tmpcrules.pdl_crules, pdue->dlc.pcrules->pdl_crules, size);
		for ( i = 0; i < pdue->dlc.pcrules->nrules; i++)
			assign_final_index(pdue, dl_color_rule_sinfo, pdue->dlc.pcrules->pdl_crules + i, tmpcrules.pdl_crules + i);

		if (write_crules(&tmpcrules, pstream) < 0)
			return ERROR;
		/*free(tmpcrules.pdl_crules);*/
		tisFree(tmpcrules.pdl_crules);
		}
	pdue->dl_hdr.nwords_crules = WDRND(ftell(pstream->io_info.pfile))-pdue->dl_hdr.ocrules;

	/*
	 * Write out all the display list static elements
	 */
	nsta = 0;
	pdue->dl_hdr.osta = WDRND(ftell(pstream->io_info.pfile));
	
	for (peue = (eue *) pdue->dl.node.next; peue; peue = (eue *) peue->node.next)
		{
		union dl_elements_union ago;
		unsigned 				type	= peue->elem.dl_anygo.element.type;
		
		if (dl_element_class[type] != dl_class_graphic)
			continue;

		/*
		 * runtime OPI requires the dynamic graphical to be in 
		 * a seperate list
		 */
		if (dynamic_graphical_element(peue))
			continue;
#ifdef DEBUG
		dump_element(&peue->elem, &pdue->ct, &pdue->st);
#endif
		nsta++;
		if (write_sta_attr(pdue, pstream, peue, &sta_attr, nsta==1) == ERROR)
			return ERROR;

		ago = peue->elem;
		assign_final_index(pdue, dl_elements_ginfo[type].sinfo, &peue->elem, &ago);

		if (edd_write(pstream, &ago, WDRND(dl_elements_ginfo[type].size)<<1) == ERROR)
			return ERROR;
		}
	pdue->dl_hdr.nwords_sta = WDRND(ftell(pstream->io_info.pfile))-pdue->dl_hdr.osta;

	/*
	 * Write out all the display list dynamic elements
	 */
	ndyn = 0;
	pdue->dl_hdr.odyn = WDRND(ftell(pstream->io_info.pfile));

	for (peue = (eue *) pdue->dl.node.next; peue; peue = (eue *) peue->node.next)
		{
		union dl_elements_union ago;
		unsigned 				type	= peue->elem.dl_anygo.element.type;
		
		if (dl_element_class[type] != dl_class_graphic)
			continue;

		/*
		 * runtime OPI requires the dynamic graphical to be in a seperate list
		 */
		if (!dynamic_graphical_element(peue))
			continue;
#ifdef DEBUG
		dump_element(&peue->elem, &pdue->ct, &pdue->st);
#endif

		ndyn++;
		if (write_sta_attr(pdue, pstream, peue, &sta_attr, ndyn==1) == ERROR)
			return ERROR;
		if (write_dyn_attr(pdue, pstream, peue, &dyn_attr, ndyn==1) == ERROR)
			return ERROR;

		ago = peue->elem;
		assign_final_index(pdue, dl_elements_ginfo[type].sinfo, &peue->elem, &ago);

		if (edd_write(pstream, &ago, WDRND(dl_elements_ginfo[type].size)<<1) == ERROR)
			return ERROR;
		}
	pdue->dl_hdr.ndyn = ndyn;

	/*
	 * Write out monitor elements
	 * Write out control elements
	 * Write out plot elements
	 * Write out related display elements
	 * Write out execute script elements
	 * Write out kill display elements
	 */
	if (write_class(pdue, pstream, dl_class_monitor, &pdue->dl_hdr.nmon, &pdue->dl_hdr.omon) == ERROR)
		return ERROR;

	if (write_class(pdue, pstream, dl_class_control, &pdue->dl_hdr.nctrl, &pdue->dl_hdr.octrl) == ERROR)
		return ERROR;

	if (write_class(pdue, pstream, dl_class_plot, &pdue->dl_hdr.nplot, &pdue->dl_hdr.oplot) == ERROR)
		return ERROR;

	if (write_class(pdue, pstream, dl_class_related_display, &pdue->dl_hdr.nrd, &pdue->dl_hdr.ord) == ERROR)
		return ERROR;

	if (write_class(pdue, pstream, dl_class_execute_script, &pdue->dl_hdr.nes, &pdue->dl_hdr.oes) == ERROR)
		return ERROR;

	if (write_class(pdue, pstream, dl_class_kill_display, &pdue->dl_hdr.nkd, &pdue->dl_hdr.okd) == ERROR)
		return ERROR;

	/*
	 * that's it, record the number of bytes in the header
	 */
	pdue->dl_hdr.nwords_dspy = WDRND(ftell(pstream->io_info.pfile));
	
	/*
	 * now, rewrite the strings to get the string counts into the file
	 */
#ifdef WIN32
	fseek(pstream->io_info.pfile, (unsigned long)(pdue->dl_hdr.opc<<1), SEEK_SET);
#else
	fseek(pstream->io_info.pfile, (unsigned long)(pdue->dl_hdr.opc<<1), L_SET);
#endif WIN32
	stringTraverse(&pdue->ct, write_pc, pstream);

	return OK;
	}


assign_clrrule_channels(pdue)
due	*pdue;
	{
	eue	*peue;

	reset_rulechan_id();
	for (peue = (eue *) pdue->dl.node.next; peue; peue = (eue *) peue->node.next)
		{
		union dl_elements_union *pelem	= &peue->elem;

		switch (pelem->dl_anygo.element.type)
		  {
		  case dl_rec:
		  case dl_line:
		  case dl_oval:
		  case dl_arc:
		  case dl_text:			
			reassign_rulechan(pdue, peue->dyn.mod.clrmod, peue->dyn.param.clrrule, &peue->dyn.param.rulechan[0]);
			break;
		  case dl_bar:
		  	{
			struct dl_bar	*p	= (struct dl_bar *)pelem;
			
			reassign_rulechan(pdue, p->clrmod, p->clrrule, &p->rulechan[0]);
			break;
			}
		  case dl_meter:
		  	{
			struct dl_meter	*p	= (struct dl_meter *)pelem;
			
			reassign_rulechan(pdue, p->clrmod, p->clrrule, &p->rulechan[0]);
			break;
			}
		  case dl_tank:
		  	{
			struct dl_tank	*p	= (struct dl_tank *)pelem;
			
			reassign_rulechan(pdue, p->clrmod, p->clrrule, &p->rulechan[0]);
			break;
			}
		  case dl_indicator:
		  	{
			struct dl_indicator	*p	= (struct dl_indicator *)pelem;
			
			reassign_rulechan(pdue, p->clrmod, p->clrrule, &p->rulechan[0]);
			break;
			}
		  case dl_text_update:
		  	{
			struct dl_text_update	*p	= (struct dl_text_update *)pelem;
			
			reassign_rulechan(pdue, p->clrmod, p->clrrule, &p->rulechan[0]);
			break;
			}
		  case dl_message_button:
		  	{
			struct dl_message_button	*p	= (struct dl_message_button *)pelem;
			
			reassign_rulechan(pdue, p->clrmod, p->clrrule, &p->rulechan[0]);
			break;
			}
		  case dl_valuator:
		  	{
			struct dl_valuator	*p	= (struct dl_valuator *)pelem;
			
			reassign_rulechan(pdue, p->clrmod, p->clrrule, &p->rulechan[0]);
			break;
			}
		  case dl_button:
		  	{
			struct dl_button	*p	= (struct dl_button *)pelem;
			
			reassign_rulechan(pdue, p->clrmod, p->clrrule, &p->rulechan[0]);
			break;
			}
		  case dl_choice_button:
		  	{
			struct dl_choice_button	*p	= (struct dl_choice_button *)pelem;
			
			reassign_rulechan(pdue, p->clrmod, p->clrrule, &p->rulechan[0]);
			break;
			}
		  case dl_menu:
		  	{
			struct dl_menu	*p	= (struct dl_menu *)pelem;
			
			reassign_rulechan(pdue, p->clrmod, p->clrrule, &p->rulechan[0]);
			break;
			}
		  case dl_text_entry:
		  	{
			struct dl_text_entry	*p	= (struct dl_text_entry *)pelem;
			
			reassign_rulechan(pdue, p->clrmod, p->clrrule, &p->rulechan[0]);
			break;
			}
		  }
		}
	return 0;
	}

int	next_rulechan_id	= 0;

reset_rulechan_id()
	{
	next_rulechan_id = 0;
	return 0;
	}


reassign_rulechan(pdue, clrmod, clrrule, prulechan)
due				*pdue;
unsigned char	clrmod;
#ifndef TOKENPTRS
char		*clrrule;
char		**prulechan;
#else
unsigned int	clrrule;
int		*prulechan;
#endif
	{
	int		i;
	char	buf[50];
	int		uses_colorrule;

	if ( clrmod == dl_clrmod_rule )
		{ /* needs rule mode AND non-zero length rule name string */
		char		*pstring;
		unsigned	length;

#ifndef TOKENPTRS
		stringName(&pdue->st, clrrule, &pstring, &length);
#else
		stringName(&pdue->st, dmPointer64fromToken (clrrule), &pstring, &length);
#endif
		uses_colorrule = (length != 0);
		}
	else
		uses_colorrule = 0;

	for (i = 0; i < MAX_RULE_LINES; i++, prulechan++)
		{
		if (uses_colorrule)
			sprintf(buf, ".%d", next_rulechan_id++ );
		else
			buf[0] = 0;	/* set to null string if no color rule */
#ifndef TOKENPTRS
		*prulechan = stringInsert(&pdue->ct, buf);
#else
		*prulechan = dmTokenFromPointer64 (stringInsert(&pdue->ct, buf));
#endif
		}
	return 0;
	}


/*
 *	WRITE_CLASS
 *	Write out all elements within a class
 */
static write_class(pdue, pstream, class, pnitems, poffset)
due						*pdue;
edd_stream				*pstream;
enum dl_element_class	class;
unsigned short			*pnitems;
unsigned int			*poffset;
	{
	eue		*peue;

	*pnitems = 0;
	*poffset = WDRND(ftell(pstream->io_info.pfile));
	for (peue = (eue *) pdue->dl.node.next; peue; peue = (eue *) peue->node.next)
		{
		union dl_elements_union ago;
		unsigned 				type	= peue->elem.dl_anygo.element.type;
		
		if (dl_element_class[type] != class)
			continue;
#ifdef DEBUG
		dump_element(&peue->elem, &pdue->ct, &pdue->st);
#endif

		(*pnitems)++;

		/*
		 * reorder pc indexes so they are contiguous (dont do it in the elements under edit however)
		 */
		ago = peue->elem;
		assign_final_index(pdue, dl_elements_ginfo[type].sinfo, &peue->elem, &ago);

		/*
		 * write out this element
		 */
		if (edd_write(pstream, &ago, WDRND(dl_elements_ginfo[type].size)<<1) == ERROR)
			return ERROR;
		}
	return OK;
	}


/*
 *	WRITE_STA_ATTR
 */
static write_sta_attr(pdue, pstream, peue, pattr_elem, init)
due							*pdue;
edd_stream					*pstream;
eue							*peue;
struct dl_basic_attribute	*pattr_elem;
int							init;
	{
	struct dl_basic_attribute	tmp_attr_elem;

	if (init)
		{
		/*
		 *	init the structure and allways write it to the dl
		 */
		pattr_elem->element.type	= dl_basic_attribute;
		pattr_elem->element.nwords	= WDRND(sizeof(*pattr_elem));
		}
	tmp_attr_elem.attr = peue->sta;
	assign_final_index(pdue, dl_sta_attr_sinfo, &peue->sta, &tmp_attr_elem.attr);

	if (!init)
		/*
		 *	write it to the dl only on changes
		 */
#if (defined(SOLARIS) || defined(WIN32))
		if (memcmp(&pattr_elem->attr, &tmp_attr_elem.attr, sizeof(pattr_elem->attr)) == 0)
#else
		if (bcmp((char *)&pattr_elem->attr, (char *)&tmp_attr_elem.attr, sizeof(pattr_elem->attr)) == 0)
#endif
			return OK;

	pattr_elem->attr = tmp_attr_elem.attr;

	return edd_write(pstream, pattr_elem, pattr_elem->element.nwords<<1);
	}


/*
 *	WRITE_DYN_ATTR
 */
static write_dyn_attr(pdue, pstream, peue, pattr_elem, init)
due				*pdue;
edd_stream			*pstream;
eue				*peue;
struct dl_dynamic_attribute	*pattr_elem;
int				init;
	{
	struct dl_dynamic_attribute	tmp_attr_elem;

	if (init)
		{
		/*
		 *	init the structure and allways write it to the dl
		 */
		pattr_elem->element.type	= dl_dynamic_attribute;
		pattr_elem->element.nwords	= WDRND(sizeof(*pattr_elem));
		}
	tmp_attr_elem.attr = peue->dyn;
	assign_final_index(pdue,dl_dyn_attr_sinfo, &peue->dyn, &tmp_attr_elem.attr);
	
	if (!init)
		/*
		 *	write it to the dl only on changes
		 */
#if defined(SOLARIS) || defined(WIN32)
		if (memcmp(&pattr_elem->attr, &tmp_attr_elem.attr, sizeof(pattr_elem->attr)) == 0)
#else
		if ( bcmp((char *)&pattr_elem->attr, (char *)&tmp_attr_elem.attr, sizeof(tmp_attr_elem.attr)) == 0)
#endif
			return OK;

	pattr_elem->attr = tmp_attr_elem.attr;

	return edd_write(pstream, pattr_elem, pattr_elem->element.nwords<<1);
	}


/*
 *	WRITE_PC
 */
void write_pc(pstream, p)
edd_stream	*pstream;
StringItem	*p;
	{
	struct dl_pc	pc;
	unsigned		length = max(db_name_dim, strlen(p->string)) + 1;	/* including the null and allowing for macro expansion */

	pc.element.type		= dl_pc;
	pc.element.nwords	= WDRND(sizeof(pc) + length);
	pc.count			= p->count;
	if (!counting_strings && p->count == 0)
		*(p->string) = 0;
	
	assert(edd_write(pstream, &pc, sizeof(pc)) != ERROR);
	assert(edd_write(pstream, p->string, (pc.element.nwords<<1) - sizeof(pc)) != ERROR);
	}


/*
 *	WRITE_STRING
 */
void write_string(pstream, p)
edd_stream	*pstream;
StringItem	*p;
	{
	struct dl_string	str;
	unsigned			length = strlen(p->string) + 1;	/* including the null */

	str.element.type	= dl_string;
	str.element.nwords	= WDRND(sizeof(str) + length);
	str.count			= p->count;
	if (!counting_strings && p->count == 0)
		*(p->string) = 0;

	assert(edd_write(pstream, &str, sizeof(str)) != ERROR);
	assert(edd_write(pstream, p->string, (str.element.nwords<<1) - sizeof(str)) != ERROR);
	}


/*
 * GM_SAVE_TO_NEW_FILE
 * (see edd_menus.h)
 */
void gm_save_to_new_file(pdummy, pdue, pbe)
void			*pdummy;
due				*pdue;
XButtonEvent	*pbe;
	{
	XRectangle		obj;

	if (pdue->psavenewg)
		{
		move_gizmo(pdue->psavenewg, pbe->x_root, pbe->y_root);
		return;
		}

	/*
	 * set the default display path as the current wd
	 * set the default display name
	 */
	if (pdue->psnname[0] == 0)
		strcpy(pdue->psnname, DEFAULT_DISPLAY_NAME);
	if (pdue->psnpath[0] == 0)
#ifdef SOLARIS
		getcwd(pdue->psnpath, sizeof(pdue->psnpath));
#else
#ifdef WIN32
		_getcwd(pdue->psnpath, sizeof(pdue->psnpath));
#else
		getwd(pdue->psnpath);
#endif WIN32
#endif
	obj.x		= pbe->x_root;
	obj.y		= pbe->y_root;
	obj.width	= (pdue->gizmo_width<<1);
	obj.height	= (pdue->gizmo_height<<1);
	pdue->psavenewg = create_file_name_entry_gizmo(
				pdue->s, 
				pdue->w,
				&obj, 
				opiStringTab[STR_New_File_Name_and_Path_Entry],
				pdue->psnpath, 
				sizeof(pdue->psnpath), 
				pdue->psnname, 
				sizeof(pdue->psnname), 
				save_popup_finish,
				pdue,
				pdue->gizmo_height);
	set_gizmo_color(pdue->psavenewg, pdue->foreground.pixel, pdue->background.pixel);
	}

/*
 *	SAVE_POPUP_FINISH()
 */
static void save_popup_finish(pdue, execute, pbe)
due				*pdue;
int				execute;
XButtonEvent	*pbe;
	{
	free_gizmo(pdue->psavenewg);
	pdue->psavenewg = NULL;
	if (execute)
		{
		gm_save(TRUE, pdue, pbe);
		XClearArea(pdue->d, pdue->w, 0, 0, 0, 0, True); /* send expose event toforce the editor header to redraw*/
		}
	}


/*
 * INIT_TEXT_FILE_NAME  (see edd_menus.h)
 */
void init_text_file_name(pdue)
due	*pdue;
	{
	char	*p;

	strcpy(pdue->ptpath, pdue->ppath);
	strcpy(pdue->ptname, (pdue->pname[0] == 0) ? DEFAULT_DISPLAY_NAME :  pdue->pname);

	/*
	 *	strip off existing extension if any and add the default extension extension
 	 */
#if (defined(SOLARIS) || defined(WIN32))
	if (p = strrchr(pdue->ptname, '.'))
#else
	if (p = rindex(pdue->ptname, '.'))
#endif
		*p = 0;
	strcat(pdue->ptname, TEXT_DL_NAME_EXTENSION);
	}



/*
 * GM_SAVE_TO_TEXT_FILE
 * (see edd_menus.h)
 */
void gm_save_to_text_file(param, pdue, pbe)
void			*param;
due				*pdue;
XButtonEvent	*pbe;
	{
	XRectangle	obj;

	if (pdue->psavetextg)
		{
		move_gizmo(pdue->psavetextg, pbe->x_root, pbe->y_root);
		return;
		}

	if (pdue->ptname[0] == 0)
		init_text_file_name(pdue);

	obj.x		= pbe->x_root;
	obj.y		= pbe->y_root;
	obj.width	= (pdue->gizmo_width<<1);
	obj.height	= (pdue->gizmo_height<<1);
	pdue->psavetextg = create_file_name_entry_gizmo(
				pdue->s, 
				pdue->w,
				&obj, 
				opiStringTab[STR_Report_File_Name_and_Path_En],
				pdue->ptpath, 
				sizeof(pdue->ptpath), 
				pdue->ptname, 
				sizeof(pdue->ptname), 
				save_text_popup_finish,
				pdue,
				pdue->gizmo_height);
	set_gizmo_color(pdue->psavetextg, pdue->foreground.pixel, pdue->background.pixel);			
	}


/*
 *	SAVE_TEXT_POPUP_FINISH()
 */
static void save_text_popup_finish(pdue, execute, pbe)
due				*pdue;
int				execute;
XButtonEvent	*pbe;
	{
	FILE 	*poutfile;
	char	pname[sizeof(pdue->pname)+sizeof(pdue->ppath)+sizeof(TMP_EXTENSION)+1];
	char	paname[sizeof(pdue->ptname)+sizeof(pdue->ptpath)+1];

	if (execute)
		{
		if (strcmp(pdue->ptname, pdue->pname) == 0)
			{
			sprintf(pdue->pmsg, opiStringTab[STR_Report_and_executable_file_n]);
			redraw_msg(pdue);
			return;
			}

		if (edd_path_cat(pdue, pdue->ptpath, pdue->ptname, paname, sizeof(paname)) == ERROR)
			return;

		poutfile = fopen(paname, "w");
		if (!poutfile)
			{
			sprintf(pdue->pmsg, opiStringTab[STR_File_creation_problems_Hs   ], pname);
			redraw_msg(pdue);
			gm_save_to_text_file(NULL, pdue, pbe);
			return;
			}

		/*	
		 * update temporary disk binary dl
		 */
		if (edd_path_cat(pdue, pdue->ppath, pdue->pname, pname, sizeof(pname)) == ERROR)
			return;

		if (strlen(pname)+sizeof(TMP_EXTENSION)+1>sizeof(pname))
			{
			sprintf(pdue->pmsg, opiStringTab[STR_Name_to_large_to_create_a_tm]);
			redraw_msg(pdue);
			return;
			}
		strncat(pname, TMP_EXTENSION, sizeof(pname)-strlen(pname)-1);
		
		if (save_display(pdue, pname) < 0)
			return;

		/*	
		 * dump from disk binary to disk ASCII
		 */
		dump_display(pname, poutfile, pdue->pname);
		fclose(poutfile);
		sprintf(pdue->pmsg, opiStringTab[STR_Reported_to_Hs              ], paname);
		redraw_msg(pdue);
#ifdef WIN32
		_unlink(pname);
#else
		unlink(pname);
#endif WIN32
		}
	free_gizmo(pdue->psavetextg);
	pdue->psavetextg = NULL;
	}
