/*popup_gz.c,v 1.1.1.1 1999/08/03 14:43:57 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *	P O P U P _ G Z . C
 *	
 *	Author: Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *	popup wrapper for another gizmo
 *
 *	History
 *	.01 100491 joh	set input hint true and allow olwm to set 
 *			focus to the popup as the mouse moves
 *	.02 120291 joh	no longer includes vxWorks.h
 *
 */
#include 	<stdio.h>
#include 	<Xlib.h>
#include 	<Xutil.h>
#ifdef AXP
#include	<string.h>
#endif
#include 	<gizmo.h>

typedef struct
	{
	gizmo			*pg;
	gizmo			*pmenug;
	char			*ptitle;
	XRectangle		title_obj;
	unsigned short	title_length;
	unsigned short	margin;
	} popup_gizmo_ext;


#define ANY_BUTTON	(Button1Mask|Button2Mask|Button3Mask|Button4Mask|Button5Mask)

static void 	redraw_popup_gz();
static void 	popup_redirect();


/*
 *	POPUP_INIT
 */
void popup_init(pgg)
gizmo_global *pgg;
	{
	}


/*
 *	POPUP_FREE
 */
void popup_free(pg)
gizmo	*pg;
	{
	XUnmapEvent		event;
	popup_gizmo_ext	*ppge = (popup_gizmo_ext *)(pg+1);

	/*
	 * Sending a synthetic unmap notify when unmapping a top level window
 	 * See section L.4.1.14 O'Reilly Vol 0, ICCCM chapter, section L.4.1.14
	 */
	event.type				= UnmapNotify;
	event.event				= pg->parent;
	event.window			= pg->w;
	event.from_configure	= False;
	XSendEvent(pg->pgg->d, pg->parent, False, SubstructureRedirectMask|SubstructureNotifyMask, (XEvent *)&event);
	free_gizmo(ppge->pg);
	free_gizmo(ppge->pmenug);
	}


/*
 *	CREATE_POPUP_GIZMO
 */
gizmo *create_popup_gizmo(s, leader_window, ptitle, x, y, margin, func, param, pchoices, ext_size)
Screen		*s;	/* parent's screen */
Window		leader_window;
char		*ptitle;
int			x;
int			y;
unsigned	margin;
void		(*func)();
void		*param;
char		**pchoices;
unsigned	ext_size;	/* size of the application extension */
	{
	gizmo			*pg;
	gizmo			*pmenug;
	popup_gizmo_ext	*ppge;
	XRectangle		obj;		
	unsigned		size	= sizeof(popup_gizmo_ext) + ext_size;

	/*
	 * init size which will change when the gizmo is added later
	 */
	obj.x		= x;
	obj.y		= y;
	obj.width	= GIZMO_BORDER_WIDTH<<2;
	obj.height	= GIZMO_BORDER_WIDTH<<2;

	/*
	 * button press and release selected for an auto fgrab
	 */
	pg = create_gizmo(gz_popup, s, RootWindowOfScreen(s), &obj, func, param, size, ExposureMask);
	if(!pg)
		return NULL;

	pmenug = create_choice_gizmo(s, pg->w, &obj, ok_to_operate, popup_redirect, pg, pchoices, gz_stack_row, 0);
	if (!pmenug)
		{
		free_gizmo(pg);
		return NULL;
		}
	XStoreName(pg->pgg->d, pg->w, ptitle);

	/*
	 * Tell the window manager that I will reassign focus between subwindows
	 */
	XSetWMProtocols(pg->pgg->d, pg->w, &pg->pgg->XA_WM_TAKE_FOCUS, 1);

	/*
	 * Allow the window manager to set focus on this window and its children.
	 */
	{
	XWMHints	wmh;

	wmh.window_group = leader_window;
	wmh.initial_state = NormalState;
	wmh.input = True;
	wmh.flags = WindowGroupHint|StateHint|InputHint;
	XSetWMHints(pg->pgg->d, pg->w, &wmh);
	}

	/*
	 * minimal decoration on popups and identify the group leader
	 */
#ifndef MIPSEL
	XSetTransientForHint(pg->pgg->d, pg->w, leader_window);
#endif
	ppge = (popup_gizmo_ext *)(pg+1);
	ppge->margin = margin;
	ppge->pmenug = pmenug;
	ppge->ptitle = ptitle;
	ppge->title_length = strlen(ptitle);
	if(ext_size)
		pg->pae = (void *) (ppge+1);

	assert(XSaveContext(pg->pgg->d, pg->w,  gizmo_ctx, (XPointer)pg) == 0);
	assert(XSaveContext(pg->pgg->d, pg->tw, gizmo_ctx, (XPointer)pg) == 0);
	return pg;
	}


/*
 *	POPUP_GIZMO_ADD_GIZMO
 */
void popup_gizmo_add_gizmo(pg, paddg)
gizmo	*pg;
gizmo	*paddg;
	{
	popup_gizmo_ext		*ppge = (popup_gizmo_ext *)(pg+1);
	XRectangle			gizmo_obj;
	XRectangle			popup_obj;		
	XRectangle			menu_obj;		
	XRectangle			title_obj;

	title_obj.x			= ppge->margin;
	title_obj.y			= ppge->margin;
	title_obj.width		= paddg->obj.width;
	title_obj.height	= ppge->margin;
	ppge->title_obj = title_obj;

	gizmo_obj = paddg->obj;
	gizmo_obj.x	= ppge->margin;
	gizmo_obj.y = ppge->margin*2 + title_obj.height;
	move_resize_gizmo(paddg, &gizmo_obj);
	set_gizmo_color(paddg, pg->fg, pg->bg);

	ppge->pg = paddg;

	menu_obj.x = ppge->margin;
	menu_obj.y = ppge->margin*3 + title_obj.height + paddg->obj.height;
	menu_obj.width = paddg->obj.width;
	menu_obj.height = title_obj.height;
	move_resize_gizmo(ppge->pmenug, &menu_obj);
	set_gizmo_color(ppge->pmenug, pg->fg, pg->bg);

	popup_obj = pg->obj;
	popup_obj.width = ppge->margin*2 + paddg->obj.width;
	popup_obj.height = ppge->margin*4 + title_obj.height + paddg->obj.height + menu_obj.height;
	move_resize_gizmo(pg, &popup_obj);

	XMapWindow(pg->pgg->d, paddg->w);
	XMapWindow(pg->pgg->d, pg->w);
	}


/*
 *	POPUP_GIZMO_RETURN_PGIZMO
 */
gizmo *popup_gizmo_return_pgizmo(pg)
gizmo	*pg;
	{
	return ((popup_gizmo_ext *)(pg+1))->pg;
	}


/*
 *	POPUP_GIZMO_RETURN_PMENUGIZMO
 */
gizmo *popup_gizmo_return_pmenugizmo(pg)
gizmo	*pg;
	{
	return ((popup_gizmo_ext *)(pg+1))->pmenug;
	}


/*
 *	POPUP_ACTION
 */
void popup_action(pevent, pg)
XEvent	*pevent;
gizmo	*pg;
	{
	switch (pevent->xany.type)
	  {
	  case Expose:	redraw_popup_gz(pg);	break;
	  }
	}



/*
 *	REDRAW_POPUP_GZ
 */
static void redraw_popup_gz(pg)
gizmo *pg;
	{
	popup_gizmo_ext	*ppge	= (popup_gizmo_ext *)(pg+1);
	Display 		*d		= pg->pgg->d;
	Window 			w		= pg->w;

	if (!pg->mapped)
		return;

	if (!ppge->pg)
		return;

	XSetForeground(pg->pgg->d, pg->pgg->fg_GC, pg->fg);
	draw_horizontal_string(d, w, pg->pgg->fg_GC, &ppge->title_obj, pg->pgg->pft, ppge->ptitle, ppge->title_length, box_center);
	}


/*
 *	POPUP_REDIRECT
 *
 *	(force the gizmo id to be for the pop gizmo)
 */
static void popup_redirect(ppopg, pmenug, value, pbe)
gizmo			*ppopg;
gizmo			*pmenug;
int				value;
XButtonEvent	*pbe;
	{
	if (ppopg->func)
		(*ppopg->func)(ppopg->arg, ppopg, value, pbe);
	}


/*
 *	POPUP_GIZMO_COLOR_CHANGE
 */
void popup_gizmo_color_change(pg)
gizmo	*pg;
	{
	popup_gizmo_ext	*ppge = (popup_gizmo_ext *)(pg+1);

	set_gizmo_color(ppge->pmenug, pg->fg, pg->bg);
	if (ppge->pg)
		set_gizmo_color(ppge->pg, pg->fg, pg->bg);

	if (pg->mapped)
		redraw_popup_gz(pg);
	}

