/*textedit_gz.c,v 1.1.1.1 1999/08/03 14:43:58 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *	T E X T E D I T _ G Z . C
 *	
 *	Author: Jeffrey O. Hill 7-90
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *	text input gizmo 
 *
 *	History
 *	.01 100891 joh	Use ICCCM kbd focus mode `locally active' and let the 
 *			window manager set the focus.
 *
 *	Improvements
 *	.01	 to many redraws may occur on machines without save under
 *		(the menu item redraws and the menu caused expose redraws)
 */


#include 	<stdio.h>
#include 	<Xlib.h>
#include 	<Xutil.h>
#include	<keysym.h>
#ifdef AXP
#include	<string.h>
#endif
#include	<plain_menus.h>
#include	<font_table.h>
#include	<gizmo.h>

/*
 *	NOTE: string under edit stored immediately following this structure
 */
typedef struct
	{
	unsigned short	maxchars;
	unsigned short	nchars;
	unsigned short	insert;		/* where the next character is inserted */
	unsigned short	first_drawn;	/* first char drawn in the window */
	Font		font;
	unsigned short	ycentered;
	unsigned short	char_width;
	unsigned short	nchar_window;	/* max chars in view */
	unsigned	focus:1;	/* gizmo has the kb focus */
	unsigned	kb_active:1;	/* expect kbd input */
	unsigned	mdt:1;		/* true if notify pending */
	XPoint		insert_marker_points[4];
	int		(*permissive_func)();
	int		permissive_result;
	} textedit_gizmo_ext;

#define SCROLL_WIDTH(PTGE) (1+(PTGE->nchar_window>>3))

char  *texteditmc[] =
	{       
	"type here", 
	"enter",
	"shift right",
	"shift left",
	"clear",
	0
	};

static void tem_set_insert();
static void tem_enter();
static void tem_right();
static void tem_left();
static void tem_clear();
struct jump_table    texteditjt[] =
	{
	{tem_set_insert,	(void *) 0},
	{tem_enter, 		(void *) 0},
	{tem_right, 		(void *) 0},
	{tem_left, 		(void *) 0},
	{tem_clear, 		(void *) 0},
	{0,			0}
	};
struct menu_configuration       textedit_mc =
	{
	"textedit menu",
	texteditmc,
	texteditjt
	};

static void 	redraw_textedit_gz();
static void 	resize_textedit_gizmo();
/*static int 	real_focus_event();	defined but never used */

int	exitKeyPressed;


/*
 *	TEXTEDIT_INIT
 */
void textedit_init(pgg)
gizmo_global	*pgg;
	{
	init_menu(pgg->d, &textedit_mc);
/*
	GCT	27-Nov-1994

	Added here to prevent compiler errors
*/
	texteditmc[0] = "~type here";
	texteditmc[1] = "~enter";
	texteditmc[2] = "~shift right";
	texteditmc[3] = "~shift left";
	texteditmc[4] = "~clear";
	texteditmc[5] = 0;
	}



/*
 *	TEXTEDIT_ACTION
 */
void 	textedit_action(pevent, pg)
XEvent	*pevent;
gizmo	*pg;
	{
	textedit_gizmo_ext	*ptge = (textedit_gizmo_ext *)(pg+1);
	char			*gz_string = (char *)(ptge+1);
	int			mot = FALSE;

	switch (pevent->type)
	  {
	  case EnterNotify:
		if ((ptge->permissive_result = (*ptge->permissive_func)(pg->arg, pg, FALSE)) == FALSE)
			break;
		if (pevent->xcrossing.mode != NotifyNormal)
			break;
		XSetWindowBorder(pg->pgg->d, pg->w, pg->fg);
		XDefineCursor(pg->pgg->d, pg->tw, pg->pgg->cursor);
		if (pevent->xcrossing.focus && ptge->kb_active == FALSE)
			{
			ptge->kb_active = TRUE;
			mot = TRUE;
			}
		break;
	  case LeaveNotify:
		if (ptge->permissive_result == FALSE)
			break;
		if(pevent->xcrossing.mode != NotifyNormal)
			break;
		XSetWindowBorder(pg->pgg->d, pg->w, pg->bg);
		XUndefineCursor(pg->pgg->d, pg->tw);
		if (!ptge->focus && ptge->kb_active == TRUE)
			{
			ptge->kb_active = FALSE;
			if (ptge->mdt)
				{
				tem_enter(NULL, pg);
				/*
				 * allways complete all labor for the gizmo prior to
				 * calling their routine since they may delete this
				 * gizmo while in it!
				 */
				return;
				}
			mot = TRUE;
			}
		break;
	  case KeyPress:
		{
		unsigned 		length;
		char			string[32];
		KeySym			keysym;
		XComposeStatus	status;

		if (ptge->permissive_result == FALSE)
			break;
		if(!ptge->font)
			break;

		length = XLookupString(	(XKeyEvent *)pevent, string, sizeof(string)-1, &keysym, &status);
/*		if(!length)
			break;
*/
		if (length == 0) {	/* cursor control keys */
		    switch (keysym) {
			case XK_Left:
			    if (ptge->insert > 0) {
				ptge->insert--;
				if (ptge->insert < ptge->first_drawn)
				    ptge->first_drawn--;
				mot = TRUE;
				ptge->mdt = TRUE;
			    }
			    break;
			case XK_Right:
			    if (ptge->insert < ptge->nchars) {
				if (ptge->insert - ptge->first_drawn >= ptge->nchar_window)
				    ptge->first_drawn++;
				ptge->insert++;
				mot = TRUE;
				ptge->mdt = TRUE;
			    }
			    break;
			case XK_Home:
			case XK_Up:
			    ptge->insert = 0;
			    ptge->first_drawn = 0;
			    mot = TRUE;
			    ptge->mdt = TRUE;
			    break;
			case XK_End:	/* Exit the current popup */
			    if (ptge->permissive_result == FALSE)
				break;
			    XSetWindowBorder(pg->pgg->d, pg->w, pg->bg);
			    XUndefineCursor(pg->pgg->d, pg->tw);
			    exitKeyPressed = 1;
			    if (!ptge->focus && ptge->kb_active == TRUE) {
				ptge->kb_active = FALSE;
				if (ptge->mdt) {
				    tem_enter(NULL, pg);
				    /*
				     * allways complete all labor for the gizmo prior to
				     * calling their routine since they may delete this
				     * gizmo while in it!
				    */
				    return;
				}
				mot = TRUE;
			    }
			    break;
			case XK_Down:
			    ptge->insert = ptge->nchars;
			    if (ptge->nchars - ptge->first_drawn > ptge->nchar_window)
				ptge->first_drawn = ptge->nchars - ptge->nchar_window;
			    mot = TRUE;
			    ptge->mdt = TRUE;
			    break;
		    }
		    break;
		}

		string[length] = NULL;

		switch (keysym)
		  {
		  case XK_Return:
		  case XK_Linefeed:
		  case XK_KP_Enter:
			tem_enter(NULL, pg);
			/*
			 * allways complete all labor for the gizmo prior to
			 * calling their routine since they may delete this
			 * gizmo while in it!
			 */
			return;
		  case XK_BackSpace:
		  case XK_Delete:
			ptge->mdt = TRUE;
			/* remove last entered character */
			if (!ptge->insert)
				{
				XBell(pevent->xany.display, 50);
				break;
				}
			strcpy(&gz_string[ptge->insert-1], &gz_string[ptge->insert]);
			ptge->insert--;
			ptge->nchars--;
			if (ptge->first_drawn)
				--ptge->first_drawn;
			mot = TRUE;
			break;
		  default:
			/* looking for normal printing characters */
			if ((keysym>=XK_KP_Space && keysym<=XK_KP_9) || (keysym>=XK_space && keysym<=XK_asciitilde))
				{
				char	*ps;
				char	*pd;
				
				if (length + ptge->nchars >= ptge->maxchars)
					{
					XBell(pevent->xany.display, 50);
					break;
					}
				for(ps = &gz_string[ptge->nchars], pd = ps + length; ps >= &gz_string[ptge->insert]; ps--, pd--)
					*pd = *ps;
				for(ps = string, pd = &gz_string[ptge->insert]; ps < &string[length]; ps++, pd++)
					*pd = *ps;
				ptge->nchars += length;
				ptge->insert += length;
				if (ptge->insert - ptge->first_drawn > (unsigned int)(ptge->nchar_window))
					ptge->first_drawn++;
				ptge->mdt = TRUE;
				mot = TRUE;
				}
		  }
		break;
		}
	  case ButtonPress:
		if ((ptge->permissive_result = (*ptge->permissive_func)(pg->arg, pg, TRUE)) == FALSE)
			break;
		if (pevent->xbutton.button == MENU_BUTTON)
			post_menu(pevent,&textedit_mc,pg);
		else if(pevent->xbutton.button == SELECT_BUTTON)
			tem_set_insert(	NULL, pg, pevent);
		break;
	  case FocusIn:
		if (ptge->permissive_result == FALSE)
			break;
        switch (pevent->xfocus.detail)
          {
		  case NotifyPointer:
			ptge->kb_active = TRUE;
			break;
		  case NotifyNonlinear:
		  case NotifyAncestor:
		  case NotifyInferior:
			ptge->kb_active = TRUE;
			mot= TRUE;
			ptge->focus = TRUE;
			break;
		  case NotifyVirtual:
		  case NotifyNonlinearVirtual:
		  default:
			break;
		  }	
		break;

	  case FocusOut:
		if (ptge->permissive_result == FALSE)
			break;
        switch (pevent->xfocus.detail)
          {
          case NotifyNonlinear:
          case NotifyInferior:
			ptge->kb_active = FALSE;
			/* no break */
          case NotifyAncestor:
			ptge->focus = FALSE;
			if (ptge->mdt)
				{
				tem_enter(NULL, pg);
				/*
				 * allways complete all labor for the gizmo prior to calling their routine since they may delete 
				 * this gizmo while in it!
				 */
				return;
				}
			mot = TRUE;
			break;
          case NotifyPointer:
			ptge->kb_active = FALSE;
			break;
          case NotifyVirtual:
          case NotifyNonlinearVirtual:
          default:
			break;
          }
		break;
	  case ConfigureNotify:
		resize_textedit_gizmo(pg);
		break;
	  default:
		break;
	  }

	if ((pevent->type==Expose)||mot)
		redraw_textedit_gz(pg);
	}



/*
 *	real_focus_event
 *	defined but never called
 */
/*
static int 
real_focus_event(pxev)
XEvent	*pxev;
	{
	if (pxev->type != FocusIn && pxev->type != FocusOut)
		return FALSE;

	switch (pxev->xfocus.detail)
	  {
	  case NotifyNonlinear:
	  case NotifyAncestor:
	  case NotifyInferior:
		return TRUE;
		break;
	  case NotifyPointer:
	  case NotifyVirtual:
	  case NotifyNonlinearVirtual:
	  default:
		break;
	  }	
	return FALSE;
	}
*/
	

/*
 *	REDRAW_TEXTEDIT_GZ
 */
static void redraw_textedit_gz(pg)
gizmo *pg;
	{
	Display			*d = pg->pgg->d;
	Window			w = pg->w;
	textedit_gizmo_ext	*ptge	= (textedit_gizmo_ext *)(pg+1);

	if (!pg->mapped)
		return;

	XSetForeground(pg->pgg->d, pg->pgg->fg_GC, pg->fg);
	XSetBackground(pg->pgg->d, pg->pgg->fg_GC, pg->bg);
	XSetForeground(pg->pgg->d, pg->pgg->grey_fg_GC, pg->fg);
	XSetBackground(pg->pgg->d, pg->pgg->grey_fg_GC, pg->bg);
	XClearWindow(d, w);
	if (!ptge->mdt)
		XDrawLine(d, w, pg->pgg->grey_fg_GC, 0, ptge->ycentered+1, pg->obj.width, ptge->ycentered+1);

	if (ptge->font)
		{
		GC	tmpgc = (ptge->mdt) ? pg->pgg->grey_fg_GC : pg->pgg->fg_GC;

		XSetFont(d, tmpgc, ptge->font);
		tisDrawString(d, w, tmpgc, 0, ptge->ycentered, ((char *)(ptge+1)) + ptge->first_drawn, ptge->nchars-ptge->first_drawn);
		if (ptge->kb_active)
			{
			tmpgc = pg->pgg->fg_GC;
			ptge->insert_marker_points[0].x = ((short)(ptge->char_width * (ptge->insert-ptge->first_drawn))) - (ptge->char_width>>1);
			XDrawLines(d, w, pg->pgg->fg_GC, ptge->insert_marker_points, NELEMENTS(ptge->insert_marker_points), CoordModePrevious);
			}
		}
	else
		{
		XDrawLine(d, w, pg->pgg->grey_fg_GC, 0, 0, pg->obj.width, pg->obj.height);
		XDrawLine(d, w, pg->pgg->grey_fg_GC, pg->obj.width, 0, 0, pg->obj.height);
		}
	}



/*
 *	CREATE_TEXTEDIT_GIZMO
 */
gizmo *create_textedit_gizmo(s, w, pobj, perm_func, func, arg, maxchars, ext_size)
Screen		*s;		/* parent's screen */
Window		w;		/* parent window */
XRectangle	*pobj;		
int			(*perm_func)();
void		(*func)();
void		*arg;
unsigned	maxchars;
unsigned	ext_size;	/* size of the application extension */
	{
	gizmo				*pg;
	textedit_gizmo_ext	*ptge;
	unsigned			text_size = ALIGN(maxchars + 1);

	pg = create_gizmo(gz_textedit, s, w, pobj, func, arg, sizeof(textedit_gizmo_ext) + text_size + ext_size,
	  PLAIN_MENU_MIN_EVENT_MASK|KeyPressMask|ButtonPressMask|EnterWindowMask|LeaveWindowMask|ExposureMask|FocusChangeMask);
	if (!pg)
		return NULL;

	ptge = (textedit_gizmo_ext *) (pg+1);

	ptge->maxchars = maxchars;
	/* install a NULL string to start with */
	ptge->insert = 0;
	ptge->nchars = 0;
	*((char *)(ptge+1)) = NULL;
	ptge->permissive_func = perm_func;
	ptge->permissive_result = FALSE;
	if(ext_size)
		pg->pae = (void *) (text_size + (char *)(ptge+1));

	resize_textedit_gizmo(pg);
	assert(XSaveContext(pg->pgg->d, pg->w,  gizmo_ctx, (XPointer)pg) == 0);
	assert(XSaveContext(pg->pgg->d, pg->tw, gizmo_ctx, (XPointer)pg) == 0);

	return pg;
	}


/*
 *	RESIZE_TEXTEDIT_GIZMO
 *
 */
static void resize_textedit_gizmo(pg)
gizmo		*pg;
	{
	textedit_gizmo_ext	*ptge = (textedit_gizmo_ext *)(pg+1);
	unsigned			index;
	XFontStruct			*pfi;

	index = min(NELEMENTS(pg->pgg->pft->height_to_pfinfo)-1, pg->obj.height);
	pfi = pg->pgg->pft->height_to_pfinfo[index];
	if (pfi)
		{
		ptge->font = pfi->fid;
		ptge->ycentered = (((unsigned int)(pg->obj.height) - pfi->max_bounds.ascent - pfi->max_bounds.descent)>>1)
			+ pfi->max_bounds.ascent;
		ptge->char_width = pfi->max_bounds.width;
		ptge->nchar_window = pg->obj.width/ptge->char_width;
		/* first x point set prior to each redraw_textedit_gz */
		ptge->insert_marker_points[0].y = ptge->ycentered + pfi->max_bounds.descent;
		/* the rest are offsets */
		ptge->insert_marker_points[1].x = (ptge->char_width>>1);
		ptge->insert_marker_points[1].y = 0;
		ptge->insert_marker_points[2].x = 0;
		ptge->insert_marker_points[2].y = -((short)(pfi->max_bounds.ascent + pfi->max_bounds.descent));
		ptge->insert_marker_points[3].x = (ptge->char_width>>1);
		ptge->insert_marker_points[3].y = 0;
		}
	}


/*
 *	SET_TEXTEDIT_GIZMO
 */
int set_textedit_gizmo(pg, pinitstr)
gizmo	*pg;
char	*pinitstr;
	{
	textedit_gizmo_ext	*ptge = (textedit_gizmo_ext *)(pg+1);
	unsigned			length;

	if (ptge->kb_active || ptge->mdt)
		return ERROR;	
	else
		{
		if (strcmp((char *)(ptge+1), pinitstr))
			{
			length = strlen(pinitstr);
			assert(length < ptge->maxchars);
			ptge->insert = length;
			ptge->nchars = length;
			strcpy((char *)(ptge+1), pinitstr);
			redraw_textedit_gz(pg);
			}
		}
	return OK;
	}


/*
 *	TEXTEDIT_GIZMO_RETURN_PBUF
 */
char *textedit_gizmo_return_pbuf(pg)
gizmo		*pg;
	{
	textedit_gizmo_ext	*ptge = (textedit_gizmo_ext *)(pg+1);

	return (char *)(ptge+1);
	}



/*
 *	TEM_SET_INSERT
 *
 */
static void tem_set_insert(param, pg, pbe)
void		*param;
gizmo		*pg;
XButtonEvent	*pbe;
	{
	textedit_gizmo_ext	*ptge = (textedit_gizmo_ext *)(pg+1);

	if (!ptge->font)
		return;

	ptge->insert = min((unsigned int)(ptge->first_drawn) + (pbe->x + ((unsigned int)(ptge->char_width)>>1))/
	  (unsigned int)(ptge->char_width), (unsigned int)(ptge->nchars));
	redraw_textedit_gz(pg);
	}




/*
 *	TEM_RIGHT
 */
static void tem_right(param, pg, pbe)
void			*param;
gizmo			*pg;
XButtonEvent	*pbe;
	{
	textedit_gizmo_ext	*ptge = (textedit_gizmo_ext *)(pg+1);
	
	if (!ptge->font)
		return;

	/* dont overflow unsigned */
	if (ptge->nchars <= ptge->nchar_window)
		return;

	ptge->first_drawn = min((unsigned int)(ptge->first_drawn+SCROLL_WIDTH(ptge)),
				(unsigned int)(ptge->nchars - ptge->nchar_window));
	redraw_textedit_gz(pg);
	}


/*
 *	TEM_LEFT
 */
static void tem_left(param, pg, pbe)
void			*param;
gizmo			*pg;
XButtonEvent	*pbe;
	{
	textedit_gizmo_ext	*ptge = (textedit_gizmo_ext *)(pg+1);

	if (!ptge->font)
		return;

	/* the if prevents unsigned overflow (scrolling behind char zero) */
	if ((unsigned int)(ptge->first_drawn) > SCROLL_WIDTH(ptge))
		ptge->first_drawn = ptge->first_drawn - SCROLL_WIDTH(ptge);
	else
		ptge->first_drawn = 0;
	redraw_textedit_gz(pg);
	}


/*
 *	TEM_CLEAR
 */
static void tem_clear(param, pg, pbe)
void			*param;
gizmo			*pg;
XButtonEvent	*pbe;
	{
	textedit_gizmo_ext	*ptge		= (textedit_gizmo_ext *)(pg+1);
	char				*gz_string	= (char *)(ptge+1);

	ptge->first_drawn = 0;
	ptge->insert = 0;
	ptge->nchars = 0;
	ptge->mdt = TRUE;
	gz_string[0] = NULL;
	redraw_textedit_gz(pg);
	}


/*
 *	TEM_ENTER
 *
 */
static void tem_enter(param, pg)
void	*param;
gizmo	*pg;
	{
	textedit_gizmo_ext	*ptge		= (textedit_gizmo_ext *)(pg+1);
	char				*gz_string	= (char *)(ptge+1);

	ptge->mdt = FALSE;

	/*
	 * consensus is to loose focus after enter
	 */
	if (ptge->focus)
   		XSetInputFocus(pg->pgg->d, PointerRoot, RevertToPointerRoot, CurrentTime);
	else
		redraw_textedit_gz(pg);

	/*
	 * allways complete all labor for the gizmo prior to calling their routine since they may delete this
	 * gizmo while in it!
	 */
	if(pg->func)
		(pg->func)(pg->arg, pg, gz_string);
	}


/*
 * 	TEXTEDIT_FREE
 * 	runs when this gizmo is deleted
 *
 *	NOTE: does nothing for now since they will not wish for changes
 *	to be saved unless they explicitly take action to do so.
 */
void textedit_free(pg)
gizmo   *pg;
	{

#ifdef SAVE_CHANGES_ON_EXIT
	/*
	 * NOTE: do not enable this unless changes are made to edd to prevent
	 * crash when element dyn or type fields change and gizmo is deleted
	 * for element transition.
	 */
	if (pg->func)	
		(pg->func)(pg->arg, pg, gz_string);
#endif
	}
