/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *	V A L U A T O R _ G Z . C
 *
 *	Author: Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *	analog value input (control) gizmo
 *
 *	History
 *	.00	07xx90	joh	written
 *	.01	121990	lrd	make cursor end up on value on 
 *					button release
 *	.02	121990	lrd	send change continuously
 *	.03	030791	joh	added left, right, and up pointer directions
 *	.04	030791	joh	added scale and precision as options
 *	.05	031991	joh	only erases box around pointer after move
 *	.06	04xx91	joh	gizmo highlight stays on when button
 *				down ptr leaves the gizmo box
 *	.07	041191 	joh	added setup form (popup)
 *	.08	042491 	joh	added send change continuously option
 *					to setup form
 *	.09	080591	joh	a set of within 2% outside the limit 
 *				now clips to the limit
 *	.10	091991	joh	now uses %f not %g at GTA's request
 *	.11	112791	joh	draw pointer last to avoid text overlap 
 *	.12	120291	joh	no longer includes vxWorks.h
 */
#include 	<stdio.h>
#include 	<Xlib.h>
#include 	<Xutil.h>
#ifdef AXP
#include	<string.h>
#endif
#include	<keysym.h>
#include	<plain_menus.h>
#include	<font_table.h>
#include	<gizmo.h>
#include	<dldef.h>

/*
 * setup form inards
 */
enum 	columns
	{
	description_col,
	gizmo_col,
	last_col
	};
	
enum 	rows
	{
	ptr_row,
	incr_row,
	delta_row,
	continuous_row,
	last_row
	};

struct setup_ext
	{
	gizmo	*pg[last_row];
	};

enum ptr_scale 
	{
	ps_coarse, 
	ps_medium, 
	ps_fine, 
	ps_xfine, 
	ps_nitems
	};
	
static char	*pptr_choices[ps_nitems+1] =
	{
	"coarse",
	"medium",
	" fine ",
	"xfine ",
	0
	};

enum incr_enbl
	{
	ie_disable, 
	ie_enable, 
	ie_nitems
	};
	
static char	*pincr_choices[ie_nitems+1] =
	{
	"disable",
	"enable",
	0
	};

enum continuous_enbl
	{
	cp_disable, 
	cp_enable, 
	cp_nitems
	};
	
static char	*pcontinuous_choices[ie_nitems+1] =
	{
	"release",
	"motion",
	0
	};
	
enum direction
	{
	dir_none,
	dir_up,
	dir_down,
	};

#define NTICKS 8
#define DEFAULT_PRECISION 4
#define VALUATOR_POPUP_BORDER_MM	5
#define VALUATOR_KB_POPUP_MAXCHAR 	32
#define VALUATOR_KB_POPUP_WIDTH_MM	60
#define VALUATOR_KB_POPUP_HEIGHT_MM	30
#define VALUATOR_SETUP_POPUP_MAXCHAR 	32
#define VALUATOR_GIZMO_WIDTH_MM		150
#define VALUATOR_GIZMO_HEIGHT_MM	8
#define MAX_VALUATOR_PRECISION		10
#define VALUATOR_EV_MASK \
(PLAIN_MENU_MIN_EVENT_MASK|ButtonPressMask|ButtonReleaseMask|EnterWindowMask|LeaveWindowMask|Button1MotionMask|ExposureMask)


#define PTR_DIVIDE 2	/* use a power of two */

typedef struct
	{
	float		ul;
	float		ll;
	float		value;
	float		last_value;
	float		initial_value;
	float 		ptr_scale_factor[ps_nitems];
	float		custom_incr;		/* ciustom incr/decr delta */
	enum ptr_scale	ptr_scale;		/* the delta for ptr pixel cross */
	enum incr_enbl	custom_incr_enbl;	/* custom incr/decr enable */
	XFontStruct	*pfi;
	gizmo		*pkbg;
	gizmo		*psug;
	XPoint		value_marker_points[4];
	unsigned short	nchars;
	short		last_pos;		/* last mouse position */
	short		ptr_dir;		/* the pointer direction */
	short		ptrsz;			/* dim of square ptr bounds */
	short		precision;		/* max num digits right of decimal */
	short		label_type;
	short		ptr_inside;		/* pointer in gz's window */
#define PTR_PLUS 0
#define PTR_MINUS 1
	short		ptr_sign;		/* vertical sliders must invert */
	short		select_button_down;	/* button is depressed */
	short		continuous;		/* send all updates */
	enum direction	direction;		/* dir_none, dir_up or dir_down */
	enum direction	last_direction;		/* dir_none, dir_up or dir_down */
	int		(*permissive_func)();
	int		permissive_result;
	} valuator_gizmo_ext;


char  *valuatormc[] =
	{
	"increment",
	"decrement",
	"keyboard",
	"coarse",
 	"medium", 
 	"fine", 
 	"xfine", 
	"setup",
	0
	};

static void kb_popup_update();
       void vm_modify();
static void vm_set_ptr_scale();
static void vm_incr();
static void vm_kb_popup();
static void kb_popup_finish();
static void update_valuator();
static void valuator_expose_redraw();
static void draw_scale();
static void compute_polygon_bounds();
static void resize_valuator_gizmo();
static void vm_setup_popup();
static void setup_popup_finish();
static void setup_popup_update();
static int  internal_set_valuator_gizmo();
static int  position_to_ivalue();

struct jump_table    valuatorjt[] =
	{
	{vm_incr,			(void *) 1},
	{vm_incr,			(void *) -1},
	{vm_kb_popup,		NULL},
	{vm_set_ptr_scale,	(void *) ps_coarse},
	{vm_set_ptr_scale,	(void *) ps_medium},
	{vm_set_ptr_scale,	(void *) ps_fine},
	{vm_set_ptr_scale,	(void *) ps_xfine},
	{vm_setup_popup,	NULL},
	{0, 0}
	};
	
struct menu_configuration	valuator_mc =
	{
	"valuator menu",
	valuatormc,
	valuatorjt
	};


/*
 *	VALUATOR INIT
 */
void valuator_init(pgg)
gizmo_global	*pgg;
	{
	init_menu(pgg->d, &valuator_mc);
	pptr_choices[0] = "~coarse";
	pptr_choices[1] = "~medium";
	pptr_choices[2] = "~ fine ";
	pptr_choices[3] = "~xfine ";
	pptr_choices[4] = 0;

	pincr_choices[0] = "~disable";
	pincr_choices[1] = "~enable";
	pincr_choices[2] = 0;

	pcontinuous_choices[0] = "~release";
	pcontinuous_choices[1] = "~motion";
	pcontinuous_choices[2] = 0;

	valuatormc[0] = "~increment";
	valuatormc[1] = "~decrement";
	valuatormc[2] = "~keyboard";
	valuatormc[3] = "~coarse";
	valuatormc[4] = "~medium"; 
	valuatormc[5] = "~fine";
	valuatormc[6] = "~xfine"; 
	valuatormc[7] = "~setup";
	valuatormc[8] = 0;
	}


/*
 *	CREATE_VALUATOR_GIZMO
 */
gizmo *create_valuator_gizmo(s, w, pobj, perm_func, func, arg, ll, ul, ptr_dir, ext_size)
Screen		*s;		/* parent's screen */
Window		w;		/* parent window  */
XRectangle	*pobj;		
int		(*perm_func)();
void		(*func)();
void		*arg;
double		ll;
double		ul;
int		ptr_dir;	/* the pointer direction */
unsigned	ext_size;	/* size of the application extension */
	{
	gizmo			*pg;
	valuator_gizmo_ext	*pvge;

	/* user will expect left to right increase on sliders */
	if (ul < ll)
		return NULL;

	pg = create_gizmo(gz_valuator, s, w,pobj, func, arg, sizeof(valuator_gizmo_ext) + ext_size, VALUATOR_EV_MASK);
	if (!pg)
		return NULL;

	pvge = (valuator_gizmo_ext *) (pg+1);
	if (ext_size)
		pg->pae = (void *) (pvge+1);

	pvge->ll		= ll;
	pvge->ul		= ul;
	pvge->ptr_dir		= ptr_dir;
	pvge->precision		= DEFAULT_PRECISION;
	pvge->ptr_scale		= ps_coarse;
	pvge->custom_incr	= (ul - ll)/1000.0; /* .1 % */
	pvge->custom_incr_enbl	= ie_disable;
	pvge->continuous	= TRUE;
	pvge->value		=
	pvge->last_value	=
	pvge->initial_value	= (ul-ll)/2.0 + ll;
	pvge->direction		= dir_none;
	pvge->permissive_func	= perm_func;
	pvge->permissive_result	= FALSE;
	pvge->label_type 	= dl_valuator_label_limits;
	resize_valuator_gizmo(pg);

	assert(XSaveContext(pg->pgg->d, pg->w,  gizmo_ctx, (XPointer)pg) == 0);
	assert(XSaveContext(pg->pgg->d, pg->tw, gizmo_ctx, (XPointer)pg) == 0);
	return pg;
	}



/*
 * 	RESIZE_VALUATOR_GIZMO
 */
static void resize_valuator_gizmo(pg)
gizmo *pg;
	{
	valuator_gizmo_ext 	*pvge = (valuator_gizmo_ext *) (pg+1);
	XRectangle		*pobj = &pg->obj;
	unsigned		ptrsz;
	unsigned		ptrszd2;
	unsigned		ptrszd4;
	unsigned 		index;
	XFontStruct 		*pfi;

	index = min(NELEMENTS(pg->pgg->pft->height_to_pfinfo)-1, (pg->obj.height>>1) );
	pvge->pfi = pfi = pg->pgg->pft->height_to_pfinfo[index];
	if (pfi)
		/* nchars in each of the 3 fields */
		pvge->nchars = (unsigned int)(pg->obj.width) / (3 * pfi->max_bounds.width);

	ptrsz = min((unsigned int)(pobj->width)/PTR_DIVIDE, pobj->height/PTR_DIVIDE);
	if (ptrsz>1)
		ptrsz -= 2;
	pvge->ptrsz = ptrsz;
	ptrszd2 = ptrsz>>1;
	ptrszd4 = ptrszd2>>1;
	switch (pvge->ptr_dir)
	  {
	  case gz_point_down:
		pvge->value_marker_points[0].x = pobj->width>>1;
		pvge->value_marker_points[0].y = pobj->height-1;
		/* the rest are offsets */
		pvge->value_marker_points[1].x = - ptrszd2;
		pvge->value_marker_points[1].y = - ptrsz;
		pvge->value_marker_points[2].x = ptrsz;
		pvge->value_marker_points[2].y = 0;
		pvge->value_marker_points[3].x = -ptrszd2;
		pvge->value_marker_points[3].y = ptrsz;
		pvge->ptr_sign = PTR_PLUS;
		break;
	  case gz_point_up:
		pvge->value_marker_points[0].x = pobj->width>>1;
		pvge->value_marker_points[0].y = 1;
		/* the rest are offsets */
		pvge->value_marker_points[1].x = - ptrszd2;
		pvge->value_marker_points[1].y = ptrsz;
		pvge->value_marker_points[2].x = ptrsz;
		pvge->value_marker_points[2].y = 0;
		pvge->value_marker_points[3].x = -ptrszd2;
		pvge->value_marker_points[3].y = -ptrsz;
		pvge->ptr_sign = PTR_PLUS;
		break;
	  case gz_point_right:
		pvge->value_marker_points[0].x = pobj->width-1;
		pvge->value_marker_points[0].y = pobj->height>>1;
		/* the rest are offsets */
		pvge->value_marker_points[1].x = - ptrsz;
		pvge->value_marker_points[1].y = - ptrszd2;
		pvge->value_marker_points[2].x = 0;
		pvge->value_marker_points[2].y = ptrsz;
		pvge->value_marker_points[3].x = ptrsz;
		pvge->value_marker_points[3].y = - ptrszd2;
		pvge->ptr_sign = PTR_MINUS;
		break;
	  case gz_point_left:
		pvge->value_marker_points[0].x = 1;
		pvge->value_marker_points[0].y = pobj->height>>1;
		/* the rest are offsets */
		pvge->value_marker_points[1].x = ptrsz;
		pvge->value_marker_points[1].y = - ptrszd2;
		pvge->value_marker_points[2].x = 0;
		pvge->value_marker_points[2].y = ptrsz;
		pvge->value_marker_points[3].x = - ptrsz;
		pvge->value_marker_points[3].y = - ptrszd2;
		pvge->ptr_sign = PTR_MINUS;
		break;
	  default:
		assert(FALSE);
	  }

	pvge->ptr_scale_factor[ps_coarse] = (pvge->ul - 
		pvge->ll)/position_to_ivalue(pvge,pobj->width,pobj->height);
	pvge->ptr_scale_factor[ps_medium] = pvge->ptr_scale_factor[ps_coarse]/10.0;
	pvge->ptr_scale_factor[ps_fine] = pvge->ptr_scale_factor[ps_medium]/10.0;
	pvge->ptr_scale_factor[ps_xfine] = pvge->ptr_scale_factor[ps_fine]/10.0;
	}




/*
 * 	VALUATOR_FREE
 */
void valuator_free(pg)
gizmo *pg;
	{
	valuator_gizmo_ext 	*pvge = (valuator_gizmo_ext *) (pg+1);

	if (pvge->pkbg)
		free_gizmo(pvge->pkbg);
	if (pvge->psug)
		free_gizmo(pvge->psug);
	}


/*
 * 	SET_VALUATOR_GIZMO
 */
int set_valuator_gizmo(pg, value)
gizmo	*pg;
double	value;
	{
	return (((valuator_gizmo_ext *)(pg+1))->select_button_down) ? ERROR : internal_set_valuator_gizmo(pg, value);
	}


/*
 * 	INTERNAL_SET_VALUATOR_GIZMO
 */
static int internal_set_valuator_gizmo(pg, value)
gizmo	*pg;
double	value;
	{
	valuator_gizmo_ext 	*pvge	= (valuator_gizmo_ext *) (pg+1);
	int			status	= OK;
	
	if (value > pvge->ul)
		{
		if (value - pvge->ul < (pvge->ul - pvge->ll)/2.0)
			{
			value = pvge->ul;
			status = ERROR;
			}
		else
			return ERROR;
		}
	else if (value < pvge->ll)
		{
		if (pvge->ll - value < (pvge->ul - pvge->ll)/2.0)
			{
			value = pvge->ll;
			status = ERROR;
			}
		else
			return ERROR;
		}

	if (pvge->value != value)
		{
		pvge->last_value = pvge->value;
		pvge->value = value;
		pvge->last_direction = pvge->direction;
		pvge->direction = (pvge->value > pvge->last_value) ? dir_up : dir_down;
		update_valuator(pg);
		}

	return status;
	}


/*
 * 	SET_VALUATOR_GIZMO_PRECISION
 */
int set_valuator_gizmo_precision(pg, precision)
gizmo		*pg;
unsigned	precision; /* max digits to the right of decimal pt */
	{
	valuator_gizmo_ext 	*pvge = (valuator_gizmo_ext *) (pg+1);

	/* clip  */
	if (precision > MAX_VALUATOR_PRECISION)
		return ERROR;

	pvge->precision = precision;
	valuator_expose_redraw(pg);

	return OK;
	}


/*
 * 	SET_VALUATOR_GIZMO_LABEL_TYPE
 */
int set_valuator_gizmo_label_type (pg, label_type)
gizmo	*pg;
int	label_type;
	{
	valuator_gizmo_ext 	*pvge = (valuator_gizmo_ext *) (pg+1);

	pvge->label_type = label_type;
	valuator_expose_redraw(pg);

	return OK;
	}



/*
 * VALUATOR_ACTION
 */
void 	valuator_action(pevent, pg)
XEvent		*pevent;
gizmo		*pg;
	{
	valuator_gizmo_ext 	*pvge = (valuator_gizmo_ext *) (pg+1);
	Display			*d = pevent->xany.display;
	Window			w = pevent->xany.window;

	switch (pevent->type)
	  {
	  case EnterNotify:
		if ((pvge->permissive_result = (*pvge->permissive_func)(pg->arg, pg, FALSE)) == FALSE)
			break;
		/*
		 * ptr is allways snapped back to inside when they release the button
		 */
		pvge->ptr_inside = TRUE;
		if (!pvge->select_button_down)
			{
			XSetWindowBorder(pg->pgg->d, pg->w, pg->fg);
			XDefineCursor(pg->pgg->d, pg->tw, pg->pgg->cursor);
			}
		break;
	  case LeaveNotify:
		if (pvge->permissive_result == FALSE)
			break;
		/*
		 * ptr is allways snapped back to inside when they release the button
		 */
		pvge->ptr_inside = FALSE;
		if (!pvge->select_button_down)
			{
			XSetWindowBorder(pg->pgg->d, pg->w, pg->bg);
			XUndefineCursor(pg->pgg->d, pg->tw);
			}
		break;
	  case ButtonRelease:
		if (pvge->permissive_result == FALSE)
			break;
		if (pevent->xbutton.button == SELECT_BUTTON)
			{
			pvge->select_button_down = FALSE;
			/* Move pointer onto current value */
			XWarpPointer(d, None, w, 0, 0, 0, 0, pvge->value_marker_points[0].x, pvge->value_marker_points[0].y);
			pvge->direction = dir_none;

			/* call their action routine */
			if (pg->func)
				(*pg->func)(pg->arg, pg, pvge->value);
			/*
			 * allways complete all labor for the gizmo prior to calling their routine since they may delete this gizmo while in
			 * it!
			 */
			}
		break;
	  case ButtonPress:
		if ((pvge->permissive_result = (*pvge->permissive_func)(pg->arg, pg, TRUE)) == FALSE)
			break;
		if (pevent->xbutton.button == MENU_BUTTON) 
			post_menu(pevent, &valuator_mc, pg);
		else if (pevent->xbutton.button == SELECT_BUTTON)
			{
			pvge->select_button_down = TRUE;
			/* Move pointer onto current value */
			XWarpPointer(d, None, w, 0, 0, 0, 0, pvge->value_marker_points[0].x, pvge->value_marker_points[0].y);
			pvge->last_pos = position_to_ivalue(pvge, pvge->value_marker_points[0].x, pvge->value_marker_points[0].y);
			pvge->initial_value = pvge->value;
			pvge->direction = dir_none;
			}
		break;
	  case MotionNotify:
		if (pvge->permissive_result == FALSE)
			break;
		if (pevent->xmotion.state & Button1Mask)
			{
			int new;
			int change;

			new = position_to_ivalue(pvge, pevent->xmotion.x, pevent->xmotion.y);
			change = new - pvge->last_pos;

			if ((change >  2) || (change < -2))
				{
				double value;

				value = pvge->value;
				if (pvge->ptr_sign == PTR_PLUS)
					value += change*pvge->ptr_scale_factor[pvge->ptr_scale];
				else
					value -= change*pvge->ptr_scale_factor[pvge->ptr_scale];
				internal_set_valuator_gizmo(pg, value);
				pvge->last_pos = new;

				/* call their action routine */
				if (pg->func && pvge->continuous)
					(*pg->func)(pg->arg, pg, pvge->value);
				}
			}
		break;
	  case Expose:
		valuator_expose_redraw(pg);
		break;
	  case ConfigureNotify:
		resize_valuator_gizmo(pg);
		break;
	  }
	}



/*
 * GET_VALUATOR_INITIAL_VALUE
 */
float get_valuator_initial_value(pg)
gizmo	*pg;
	{
	return	((valuator_gizmo_ext *) (pg+1))->initial_value;
	}

/*
 * VALUATOR_DIRECTION_CHANGED
 */
valuator_direction_changed(pg)
gizmo	*pg;
	{
	valuator_gizmo_ext 	*pvge	= (valuator_gizmo_ext *) (pg+1);

	return (pvge->last_direction != dir_none) && (pvge->direction ^ pvge->last_direction);
	}


/*
 * POSITION_TO_IVAL
 */
static int position_to_ivalue(pvge, x, y)
valuator_gizmo_ext 	*pvge;
int					x;
int					y;
	{
	if(pvge->ptr_dir == gz_point_down || pvge->ptr_dir == gz_point_up)
		return x;
	else if(pvge->ptr_dir == gz_point_right || pvge->ptr_dir == gz_point_left)
		return y;
	else
		assert(FALSE);
	}


/*
 * UPDATE_VALUATOR
 */
static
void update_valuator(pg)
gizmo		*pg;
{
	valuator_gizmo_ext 	*pvge = (valuator_gizmo_ext *) (pg+1);
	XRectangle			cvrec;		/* current value rec */
	double				pfs;
	XRectangle 			cliprec;

	if (!pg->mapped)
		return;

	XSetForeground(pg->pgg->d, pg->pgg->fg_GC, pg->fg);
	XSetBackground(pg->pgg->d, pg->pgg->fg_GC, pg->bg);
	XSetForeground(pg->pgg->d, pg->pgg->grey_fg_GC, pg->fg);
	XSetBackground(pg->pgg->d, pg->pgg->grey_fg_GC, pg->bg);
	compute_polygon_bounds(pvge->value_marker_points, NELEMENTS(pvge->value_marker_points), &cliprec);

	XClearArea(	pg->pgg->d, pg->w, cliprec.x, cliprec.y, cliprec.width, cliprec.height, False);
	if (pvge->pfi)
		XSetFont(pg->pgg->d, pg->pgg->fg_GC, pvge->pfi->fid);

	pfs = (pvge->value-pvge->ll) / (pvge->ul - pvge->ll);
	switch (pvge->ptr_dir)
	  {
	  case gz_point_down:
	  case gz_point_up:
		cvrec.width = pg->obj.width/3;
		cvrec.height = pg->obj.height - pvge->ptrsz;
		cvrec.y	= (pvge->ptr_dir == gz_point_down) ? 0 : pvge->ptrsz;
		cvrec.x = cvrec.width;
		pvge->value_marker_points[0].x = pg->obj.width*pfs;
		break;
	  case gz_point_right:
	  case gz_point_left:
		cvrec.width = pg->obj.width - pvge->ptrsz;
		cvrec.height = pg->obj.height/3;
		cvrec.x = (pvge->ptr_dir == gz_point_right) ? 0 : pvge->ptrsz;
		cvrec.y = cvrec.height;
		pvge->value_marker_points[0].y = pg->obj.height - (int)(pg->obj.height*pfs);
		break;
	  default:
		assert(FALSE);
	  }

	if (pvge->label_type > dl_valuator_label_outline) {
	    /* clamp precision to reasonable value! */
	    if (pvge->precision > 9) {
		pvge->precision = 9;
	    }
	    sprintf(pg->pgg->sprintf_buf, "% .*f", pvge->precision, pvge->value);
	    XClearArea (pg->pgg->d, pg->w, cvrec.x, cvrec.y, cvrec.width, cvrec.height, False);
	    draw_horizontal_string(pg->pgg->d, pg->w, pg->pgg->fg_GC, &cvrec, pg->pgg->pft, pg->pgg->sprintf_buf,
			strlen(pg->pgg->sprintf_buf), box_begin);
	}

	/* draw new pointer */

	XFillPolygon(pg->pgg->d, pg->w, pg->pgg->grey_fg_GC, pvge->value_marker_points,
			NELEMENTS(pvge->value_marker_points), Convex, CoordModePrevious);
	/*
	 * redraw only those ticks which have been erased (draw ticks ontop of the pointer so they are easier to read)
	 */
	if (pvge->label_type > dl_valuator_label_none)
		draw_scale(pg, &cliprec);
}



/*
 *	EXPOSE_REDRAW
 */
static
void valuator_expose_redraw(pg)
gizmo	*pg;
	{
	valuator_gizmo_ext 	*pvge =	(valuator_gizmo_ext *) (pg+1);
	XRectangle		ulrec;
	XRectangle		llrec;
	int			x1, y1, x2, y2;
	int			llfmt, ulfmt;

	XSetForeground(pg->pgg->d, pg->pgg->fg_GC, pg->fg);
	if(!pg->mapped)
		return;

	switch(pvge->ptr_dir)
	  {
	  case gz_point_down:
		llrec.width = ulrec.width = pg->obj.width/3;
		llrec.height = ulrec.height = pg->obj.height - pvge->ptrsz;
		ulrec.y = llrec.y = 0;
		llrec.x = 0;
		ulrec.x = pg->obj.width - ulrec.width;
		llfmt = box_begin;
		ulfmt = box_end;
		x1 = 0;
		y1 = pg->obj.height-1;
		x2 = pg->obj.width-1;
		y2 = pg->obj.height-1;
		break;
	  case gz_point_up:
		llrec.width = ulrec.width = pg->obj.width/3;
		llrec.height = ulrec.height = pg->obj.height - pvge->ptrsz;
		ulrec.y = llrec.y = pvge->ptrsz;
		llrec.x = 0;
		ulrec.x = pg->obj.width - ulrec.width;
		llfmt = box_begin;
		ulfmt = box_end;
		x1 = 0;
		y1 = 0;
		x2 = pg->obj.width-1;
		y2 = 0;
		break;
	  case gz_point_right:
		ulrec.width = llrec.width = pg->obj.width - pvge->ptrsz;
		ulrec.height = llrec.height = pg->obj.height/8;
		ulrec.x = llrec.x = 0;
		ulrec.y = 0;
		llrec.y = pg->obj.height - ulrec.height;
		llfmt = box_begin;
		ulfmt = box_begin;
		x1 = pg->obj.width-1;
		y1 = 0;
		x2 = pg->obj.width-1;
		y2 = pg->obj.height-1;
		break;
	  case gz_point_left:
		ulrec.width = llrec.width = pg->obj.width - pvge->ptrsz;
		ulrec.height = llrec.height = pg->obj.height/8;
		ulrec.x = llrec.x = pvge->ptrsz;
		ulrec.y = 0;
		llrec.y = pg->obj.height - ulrec.height;
		llfmt = box_begin;
		ulfmt = box_begin;
		x1 = 0;
		y1 = 0;
		x2 = 0;
		y2 = pg->obj.height-1;
		break;
	  default:
		assert(FALSE);
	  }

	/* Draw the decorations. Each is cumulative. 1=>scale, 2=>scale+limits, 3=>scale+ limits+value */

	switch (pvge->label_type) {
	    case dl_valuator_label_limits:		/* Draw limits and current value */
		/* clamp precision to reasonable value! */
		if (pvge->precision > 9) {
		    pvge->precision = 9;
		}
		sprintf(pg->pgg->sprintf_buf, "% .*f", pvge->precision, pvge->ll);
		draw_horizontal_string(pg->pgg->d, pg->w, pg->pgg->fg_GC, &llrec, pg->pgg->pft,
			pg->pgg->sprintf_buf, strlen(pg->pgg->sprintf_buf), llfmt);
		sprintf(pg->pgg->sprintf_buf, "% .*f", pvge->precision, pvge->ul);
		draw_horizontal_string(pg->pgg->d, pg->w, pg->pgg->fg_GC, &ulrec, pg->pgg->pft,
			pg->pgg->sprintf_buf, strlen(pg->pgg->sprintf_buf), ulfmt);
	    case dl_valuator_label_outline:		/* Draw the scale */
		XDrawLine(pg->pgg->d, pg->w, pg->pgg->fg_GC, x1, y1, x2, y2);
		draw_scale(pg, NULL);
	    case dl_valuator_label_none:		/* Draw the pointer */
		update_valuator (pg);
	}
}


/*
 *	COMPUTE_CLIP
 *	(compute the bounds of an offset type polygon description)
 */
static void compute_polygon_bounds(ppoints, npoints, pcliprec)
XPoint		*ppoints;
unsigned	npoints;
XRectangle	*pcliprec;
	{
	unsigned 	i;
	int		x1, y1, x2, y2;
	int 		x, y;
	
	assert(npoints >= 1);

	x = x1 = x2 = ppoints[0].x;
	y = y1 = y2 = ppoints[0].y;
	for (i = 1; i < npoints; i++)
		{
		x += ppoints[i].x;
		y += ppoints[i].y;

		x1= min(x1, x);
		y1= min(y1, y);
		x2= max(x2, x);
		y2= max(y2, y);
		}
	pcliprec->x =x1;
	pcliprec->y =y1;
	pcliprec->width = x2-x1;
	pcliprec->height = y2-y1;
	}


/*
 *	DRAW_SCALE
 */
static
void draw_scale(pg, pcliprec)
gizmo		*pg;
XRectangle	*pcliprec;
	{
	valuator_gizmo_ext 	*pvge = (valuator_gizmo_ext *) (pg+1);
	int					x1, y1, x2, y2;
	int					ticksz;
	int					tickszd2;
	int					i;

	ticksz = pvge->ptrsz>>1;
	tickszd2 = ticksz>>1;
	for (i = 0; i <= NTICKS; i++)
		{
		int lclticksz;

		lclticksz =  i&1 ? tickszd2 : ticksz;

		/*
		 * NOTE: 
		 *	x1 must be <= x2 and
		 *	y1 must be <= y2
		 * for the clipping below to work properly !
		 */
		switch (pvge->ptr_dir)
		  {
		  case gz_point_down:
			x2 = x1 = (((unsigned int)(pg->obj.width)-1)*i)/NTICKS;
			y2 = pg->obj.height-1;
			y1 = y2 - lclticksz;
			break;
		  case gz_point_up:
			x2 = x1 = (((unsigned int)(pg->obj.width)-1)*i)/NTICKS;
			y1 = 0;
			y2 = lclticksz;
			break;
		  case gz_point_right:
			y2 = y1 = (((unsigned int)(pg->obj.height)-1)*i)/NTICKS;
			x2 = pg->obj.width-1;
			x1 = x2 - lclticksz;
			break;
		  case gz_point_left:
			y2 = y1 = (((unsigned int)(pg->obj.height)-1)*i)/NTICKS;
			x1 = 0;
			x2 = lclticksz;
			break;
		  default:
			assert(FALSE);
		  }

		/*
		 * clip ticks which dont need to be redrawn
		 */
		if (pcliprec)
			if (pcliprec->x > x2 || pcliprec->y > y2 ||
				pcliprec->x + (unsigned int)(pcliprec->width) < x1 || pcliprec->y + (unsigned int)(pcliprec->height) < y1)
				continue;

		XDrawLine(pg->pgg->d, pg->w, pg->pgg->fg_GC, x1, y1, x2, y2);
		}
	}



/*
 *	VM_SET_SCALE
 */
static
void vm_set_ptr_scale(param, pg, pbe)
void			*param;
gizmo			*pg;
XButtonEvent	*pbe;
	{
	valuator_gizmo_ext 	*pvge = (valuator_gizmo_ext *) (pg+1);

	pvge->ptr_scale = *(int *)&param;
	}


/*
 *	VM_INCR
 */
static void vm_incr(param, pg, pbe)
void			*param;
gizmo			*pg;
XButtonEvent	*pbe;
{
	valuator_gizmo_ext 	*pvge	= (valuator_gizmo_ext *) (pg+1);
	int			sign	= *(int *)& param;
	double			incr;
	double			scale;

	incr = (pvge->custom_incr_enbl == ie_enable) ? pvge->custom_incr : pvge->ptr_scale_factor[pvge->ptr_scale];
	internal_set_valuator_gizmo(pg, pvge->value + incr*sign);
	pvge->direction = (pvge->last_direction == dir_down) ? dir_up : dir_down; /* force these updates */

	/*
	 * call their action routine
	 * 
	 * allways complete all labor for the gizmo prior to calling their
	 * routine since they may delete this gizmo while in it!
	 */
	(*pg->func)(pg->arg, pg, pvge->value);
	}




/*
 *	VM_KB_POPUP
 *
 */
static void vm_kb_popup(param, pg, pbe)
void		*param;
gizmo		*pg;
XButtonEvent	*pbe;
{
	valuator_gizmo_ext 	*pvge = (valuator_gizmo_ext *) (pg+1);
	gizmo			*ptextg;
	XRectangle		bounds;
	int			border;
	static char 		*kb_popup_choices[] = {"quit", 0};

	if (pvge->pkbg)
		{
		move_gizmo(pvge->pkbg, pbe->x_root, pbe->y_root);
		return;
        }

	kb_popup_choices[0] = "~quit";
	kb_popup_choices[1] = 0;

	border = (VALUATOR_POPUP_BORDER_MM * WidthOfScreen(pg->pgg->s))/WidthMMOfScreen(pg->pgg->s);
	pvge->pkbg = create_popup_gizmo(pg->pgg->s, pg->parent, "~Keyboard Entry", pbe->x_root, pbe->y_root,
	  border, kb_popup_finish, pg, kb_popup_choices, 0);
	set_gizmo_color(pvge->pkbg, pg->fg, pg->bg);

	bounds.x = 0;
	bounds.y = 0;
	bounds.width = (VALUATOR_KB_POPUP_WIDTH_MM*WidthOfScreen(pg->pgg->s))/WidthMMOfScreen(pg->pgg->s);
	bounds.height = (VALUATOR_KB_POPUP_HEIGHT_MM*HeightOfScreen(pg->pgg->s))/HeightMMOfScreen(pg->pgg->s);
	ptextg = create_textedit_gizmo(pg->pgg->s, pvge->pkbg->w, &bounds, ok_to_operate, kb_popup_update, pg,
	  VALUATOR_SETUP_POPUP_MAXCHAR, 0);
	set_gizmo_color(ptextg, pg->fg, pg->bg);
	sprintf(pg->pgg->sprintf_buf, "% .*f", pvge->precision, pvge->value);
	set_textedit_gizmo(ptextg, pg->pgg->sprintf_buf);
	popup_gizmo_add_gizmo(pvge->pkbg, ptextg);
	XMapWindow(pg->pgg->d, pvge->pkbg->w);
	}


/*
 *	KB_POPUP_UPDATE
 */
static void kb_popup_update(pvaluatorg, ptexteditg, pvalue)
gizmo	*pvaluatorg;
gizmo	*ptexteditg;
char	*pvalue;
	{
	double 	fval;
	int		status;

	valuator_gizmo_ext 	*pvge = (valuator_gizmo_ext *) (pvaluatorg+1);

	status = sscanf(pvalue, "%le", &fval);
	if (status == 1)
		{
		status = internal_set_valuator_gizmo(pvaluatorg, fval);
		if (status < 0)
			outOfRangeDialog(pvaluatorg->pgg->s, TRUE, (float)pvge->ll, (float)pvge->ul, NULL);
		else
			{
			pvge->direction = (pvge->last_direction == dir_down) ? dir_up : dir_down; /* force these updates */
			/* call their action routine */
			if(pvaluatorg->func)
				(*pvaluatorg->func)(pvaluatorg->arg, pvaluatorg, pvge->value);
			}
		}
	else
		outOfRangeDialog(pvaluatorg->pgg->s, TRUE, (float)pvge->ll, (float)pvge->ul, "~bad format");
	}


/*
 *	KB_POPUP_FINISH
 *
 */
static void kb_popup_finish(pvaluatorg, ptexteditg, value)
gizmo	*pvaluatorg;
gizmo	*ptexteditg;
int	value;
	{
	valuator_gizmo_ext 	*pvge = (valuator_gizmo_ext *) (pvaluatorg+1);

	free_gizmo(pvge->pkbg);
	pvge->pkbg = NULL;
	}



/*
 *	VM_SETUP_POPUP
 *
 */
static void vm_setup_popup(param, pg, pbe)
void			*param;
gizmo			*pg;
XButtonEvent	*pbe;
	{
	Screen				*s = pg->pgg->s;
	valuator_gizmo_ext 	*pvge = (valuator_gizmo_ext *) (pg+1);
	gizmo				*ptextg;
	gizmo				*pformg;
	XRectangle			bounds;
	int					border;
	struct setup_ext	*pse;

	static char		*setup_popup_choices[] =
		{
		"quit",
		0
		};

	if (pvge->psug)
		{
		move_gizmo(pvge->psug, pbe->x_root, pbe->y_root);
		return;
        }

	setup_popup_choices[0] = "~quit";
	setup_popup_choices[1] = 0;

	border = (VALUATOR_POPUP_BORDER_MM*WidthOfScreen(s))/WidthMMOfScreen(s);
	pvge->psug = create_popup_gizmo(s, pg->parent, "~Valuator Setup", pbe->x_root, pbe->y_root, border, setup_popup_finish,
	  pg, setup_popup_choices, sizeof(struct setup_ext));
	pse = pvge->psug->pae;
	pformg = create_form_gizmo(s, pvge->psug->w, 0, 0, last_col, last_row, border, 0);

	/*
	 * position set later
	 */
	bounds.x = 0;
	bounds.y = 0;
	bounds.width = (VALUATOR_GIZMO_WIDTH_MM * WidthOfScreen(s))/WidthMMOfScreen(s);
	bounds.height = (VALUATOR_GIZMO_HEIGHT_MM * HeightOfScreen(s))/HeightMMOfScreen(s);

	/*
	 *	ptr row
	 */
	form_gizmo_add_string(pformg, description_col, ptr_row, bounds.height, "~Gain");
	pse->pg[ptr_row] = create_choice_gizmo(pformg->pgg->s, pformg->w, &bounds, ok_to_operate, setup_popup_update,
	  pg, pptr_choices, gz_stack_row, 0);
	set_choice_gizmo(pse->pg[ptr_row], pvge->ptr_scale);
	form_gizmo_add_gizmo(pformg, gizmo_col, ptr_row, pse->pg[ptr_row]);

	/*
	 *	incr row
	 */
	form_gizmo_add_string(pformg, description_col, incr_row, bounds.height, "~Custom Increment/Decrement");
	pse->pg[incr_row] = create_choice_gizmo(pformg->pgg->s, pformg->w, &bounds, ok_to_operate, setup_popup_update,
	  pg, pincr_choices, gz_stack_row, 0);
	set_choice_gizmo(pse->pg[incr_row], pvge->custom_incr_enbl);
	form_gizmo_add_gizmo(pformg, gizmo_col, incr_row, pse->pg[incr_row]);

	/*
	 *	delta row
	 */
	form_gizmo_add_string(pformg, description_col, delta_row, bounds.height, "~Custom Increment/Decrement Value");
	pse->pg[delta_row] = create_textedit_gizmo(pformg->pgg->s, pformg->w, &bounds, ok_to_operate, setup_popup_update, pg,
	  VALUATOR_KB_POPUP_MAXCHAR, 0);
	assert(pse->pg[delta_row] != NULL);

	sprintf(pg->pgg->sprintf_buf, "% .*f", pvge->precision, pvge->custom_incr);
	set_textedit_gizmo(pse->pg[delta_row], pg->pgg->sprintf_buf);
	form_gizmo_add_gizmo(pformg, gizmo_col, delta_row, pse->pg[delta_row]);

	/*
	 *	continuous row
	 */
	form_gizmo_add_string(pformg, description_col, continuous_row, bounds.height, "~Pointer Update Mode");
	pse->pg[continuous_row] = create_choice_gizmo(pformg->pgg->s,  pformg->w, &bounds, ok_to_operate, setup_popup_update,
	  pg, pcontinuous_choices, gz_stack_row, 0);
	set_choice_gizmo(pse->pg[continuous_row], pvge->continuous);
	form_gizmo_add_gizmo(pformg, gizmo_col, continuous_row, pse->pg[continuous_row]);
	resize_form_gizmo(pformg);

	/*
	 * place the form on the popup
	 */
	popup_gizmo_add_gizmo(pvge->psug, pformg);
	set_gizmo_color(pvge->psug, pg->fg, pg->bg);
	XMapWindow(pg->pgg->d, pvge->psug->w);
	}


/*
 *	SETUP_POPUP_UPDATE
 *
 */
static void setup_popup_update(pvaluatorg, pg, value)
gizmo	*pvaluatorg;
gizmo	*pg;
void	*value;
	{
	valuator_gizmo_ext 	*pvge = (valuator_gizmo_ext *) (pvaluatorg+1);
	struct setup_ext	*pse = (struct setup_ext *) (pvge->psug->pae);
	double 				fval;

	if (pg == pse->pg[ptr_row])
		pvge->ptr_scale = *(int *)& value;
		
	else if (pg == pse->pg[incr_row])
		pvge->custom_incr_enbl = *(int *)& value;
		
	else if (pg == pse->pg[delta_row])
		if (sscanf(value, "%le", &fval) == 1)
			pvge->custom_incr = fval;
		else
			printf(	"~custom incr kb entry invalid [%s]\n", value);
	else if(pg == pse->pg[continuous_row])
		pvge->continuous =  *(int *)& value;
	}


/*
 *	SETUP_POPUP_FINISH
 *
 */
static void setup_popup_finish(pvaluatorg, ppopupg, value)
gizmo	*pvaluatorg;
gizmo	*ppopupg;
int	value;
	{
	valuator_gizmo_ext 	*pvge = (valuator_gizmo_ext *) (pvaluatorg+1);

	free_gizmo(pvge->psug);
	pvge->psug = NULL;
	}
