/*3D.h,v 1.1.1.1 1999/08/03 14:43:58 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *
 * 7 Jan 1990 - MDA
 *
 * algorithm and shaded polygon drawing code adapted from: 
 *		Digital Equipment Corporation
 *  		Nick Williams, IED/Reading
 *
 *
 */	

#include	<stdio.h>
#include	<math.h>

#include	<X11/X.h>
#include	<X11/Xlib.h>
#include	<X11/Xutil.h>

#ifndef min
#  define min(a,b) ((a) < (b) ? a : b)
#endif
#ifndef max
#  define max(a,b) ((a) > (b) ? a : b)
#endif

#define	MAX_COLORS	100
#define	RGB_DEFAULT	100
#define	RGB_SCALE_MAX	100
#define	MAXCORNERS	4	/* max vertices per facet (quadrilateral mesh)*/
#define	MAXDISTANCE	100.0
#define	PI		3.14159
#define	DTR		PI / 180.0
#define	RTD		180.0 / PI
#define SOLID		0
#define SHADED		1


/* function macros */
#define 	ABS(A)		((A) >= 0.0 ? (A) : -(A))


/* structures for 3D */
 
struct depths {				/* structure  used to depth */
	float zmax;			/* sort facets		    */
	int facet;
	};


struct _3d {

  Display *display;
  Window window;
  XWindowAttributes wind_struct;
  GC gc;
  Pixmap pixmap;			/* pixmap backing store */

  unsigned long forePixel, backPixel, dataPixel;

  float xMin, xMax, xDelta;		/* need these to go from normalized  */
  float yMin, yMax, yDelta;		/*  data [0,1] and back again        */
  float zMin, zMax, zDelta;


/* 3D data structures */

  int nverts;				/* total number of vertices */
  int nfacets;				/* total number of facets */
  int draw_mode;			/* SOLID or SHADED */
  int **facets;				/* vertices in facet */
  int *ncorners;			/* number of vertices / facet */
  float **vertices;			/* 3D vertex positions */
  float **out_verts;			/* xformed 3D vertex positions */
  float obj_box[2][3];			/* object bounding box */
  float screen_box[2][3];		/* screen space bounding box */

  struct depths *facet_depth;


/* colour map and parameters */

  Colormap map;
  XColor *colours;
  int red_level;
  int blue_level;
  int green_level;
  int num_colours;
  unsigned long pixels[MAX_COLORS];


/* Matrix which maps world coordinates to ndc coordinates */
  float world_to_ndc[4][4];

/* Matrix which maps normalised device coor. to  screen pixel coor. */
  float ndc_to_screen[4][4];	

/* Matrix product of ndc_to_screen x world_to_ndc 
	(maps world coordinates to screen pixels */
  float world_to_screen[4][4];

/* Concatenated viewing transformation - maps world coordinates to screen 
	pixels i.e. the whole viewing pipeline */
  float cvt[4][4];

/* Current transformation matrix - operates on world coordinates producing 
	transformed world coordinates */
  float ctm[4][4];

/*  parameters for rotation, translation, and scaling parameters */
  float	rotation[3];
  float	translation[3];
  float	scaling[3];

};

typedef struct _3d ThreeD;




/*** if ANSI-C 
extern ThreeD *init3d();
extern void meshToFacets3d(ThreeD *s3d, int xSize, int ySize, 
	double *mesh, double *xArray, double *yArray);
extern void rotate3d(ThreeD *s3d, float x, float y, float z);
extern void setPixmap3d(ThreeD *s3d, Pixmap pixmap);
extern void setStyle3d(ThreeD *s3d, int style);
extern void set3d(ThreeD *s3d, Display *display, Window drawWindow);
extern void draw3d(ThreeD *s3d);
extern void term3d(ThreeD *s3d);
***/

extern ThreeD *init3d();
extern void meshToFacets3d();
extern void rotate3d();
extern void setPixmap3d();
extern void setStyle3d();
extern void set3d();
extern void draw3d();
extern void term3d();
