/*Graph.h,v 1.1.1.1 1999/08/03 14:43:59 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 * Graph structure definition
 */

#include <X11/Xlib.h>

#include "GraphXMacros.h"


#define GRAPH_MARGIN		0.15	/* window fraction for text, &c */
#define GRAPH_DRAWINGAREA	0.7	/* 2*MARGIN + DRAWINGAREA == 1 */


/* must define the filename of the temporary xwd file for printing */

#ifndef XWD_TEMP_FILE_STRING
#  define XWD_TEMP_FILE_STRING    "./graph.xwd"
#endif


#ifndef MAXCURVES
#  define MAXCURVES 10	/* maximum number of traces or curves per plot */
#endif

typedef enum {
    GraphLine,		/* XY plot (conected line sequence) type */
    GraphPoint,		/* XY plot (scatter or point) type */
    GraphBar		/* 1D bar chart (vertical - with `binning' ) type */
} GraphType;

typedef enum {
    GraphInternal,		/* make private copies of all  data */
    GraphExternal		/* use the passed-in storage for all data */
} GraphStorageMode;

typedef struct {
    double	   x;
    double	   y;
} XYdataPoint;

typedef enum {
    GraphDefaultRange,	/* do own range calculation: based on data */
    GraphSpecRange	/* use user specified ranges */
} GraphRangeType;

typedef struct {
    double minVal;
    double maxVal;
} GraphRange;



struct _Graph {
    Display	    *display;
    int		     screen;
    Window	     window;
    Pixmap	     pixmap;
    GC		     gc;
    GraphStorageMode storageMode;

    int		     nBuffers;		/* the # of "plots" in this graph */
    XYdataPoint    **dataBuffer;	/* an array of data buffers */
    int		     bufferSize;
    int             *nValues;		/* an arry of data buffer counters */

    GraphRangeType   xRangeType[MAXCURVES];
    GraphRange       xRange[MAXCURVES];
    GraphRangeType   yRangeType[MAXCURVES];
    GraphRange       yRange[MAXCURVES];

    GraphType	     type;		/* the enumerated type of Graph */
    int		     nBins[MAXCURVES];	/* nBins to use (for GraphBar only)*/
    XPoint	    *devicePoints;	/* device coordinates for dataPoints */

    BOOLEAN	     setOwnColors;	/* if TRUE user will fill in pixels */
    unsigned long    forePixel; 
    unsigned long    backPixel;
    unsigned long   *dataPixel;		/* an arry of pixel values for data */

    BOOLEAN	     dontLinearScaleX;   /* if TRUE don't make nice X scales */
    BOOLEAN	     dontLinearScaleY;   /* if TRUE don't make nice Y scales */
    unsigned long    w;
    unsigned long    h;

    /* these (X0,Y0), (X1,Y1) are relative to main window/pixmap */
    unsigned long    dataX0;            /* upper left corner - data region X */
    unsigned long    dataY0;            /* upper left corner - data region Y */
    unsigned long    dataX1;            /* lower right corner - data region X */
    unsigned long    dataY1;            /* lower right corner - data region Y */


    BOOLEAN	     setOwnFonts;	/* if TRUE user will fill in fonts */
    char	    *title;
    char	    *titleFont;
    XFontStruct	    *titleFontStruct;
    int		     titleFontHeight;
    int		     titleFontWidth;
    char	    *xAxisLabel;
    char	    *yAxisLabel;
    char	    *axesFont;
    XFontStruct	    *axesFontStruct;
    int		     axesFontHeight;
    int		     axesFontWidth;

    BOOLEAN         legend;
    char           *legendTitle;
    char          **legendArray;
    Window          legendWindow;

    caddr_t	    userData;		/* pointer to arbitrary user data */

    };
typedef struct _Graph Graph;


extern Graph *graphInit();
extern void graphDraw();
extern void graphPrint();
extern void graphRefresh();
extern void graphResize();
extern void graphSet();
extern void graphSetRange();
extern void graphSetRangeDefault();
extern void graphSetBins();
extern void graphSetDisplayType();
extern void graphSetDataColor();
extern void graphTerm();

extern void graphSetLegend();
extern void graphDrawLegend();

