/*Seql.h,v 1.1.1.1 1999/08/03 14:44:00 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 * Seql structure definition
 */

#include <X11/Xlib.h>

#include "GraphXMacros.h"


#define SEQL_MARGIN		0.15	/* window fraction for text, &c */
#define SEQL_DRAWINGAREA	0.7	/* 2*MARGIN + DRAWINGAREA == 1 */


/* must define the filename of the temporary xwd file for printing */

#ifndef XWD_TEMP_FILE_STRING
#  define XWD_TEMP_FILE_STRING	"./seql.xwd"
#endif


#ifndef MAXCURVES
#  define MAXCURVES 10	/* maximum number of traces or curves per plot */
#endif


typedef enum {
    SeqlPolygon,	/* sequential data filled polygon type */
    SeqlLine,		/* sequential data line type */
    SeqlPoint		/* sequential data point type */
} SeqlType;

typedef enum {
    SeqlInternal,	/* make private copies of all sequential data */
    SeqlExternal	/* use the passed-in storage for all data */
} SeqlStorageMode;

typedef enum {
    SeqlDefaultRange,	/* do own range calculation: based on data */
    SeqlSpecRange	/* use user specified ranges */
} SeqlRangeType;

typedef struct {
    double minVal;
    double maxVal;
} SeqlRange;


struct _Seql {
    Display	    *display;
    int		    screen;
    Window	    window;
    Pixmap	    pixmap;
    GC		    gc;
    SeqlStorageMode storageMode;

    int             nBuffers;		/* the # of "plots" in this seql */
    double	  **dataBuffer;		/* an array of data buffers */
    int		    bufferSize;
    int            *nValues;		/* an array of data buffer counters */

    SeqlRangeType   xRangeType[MAXCURVES];
    SeqlRange	    xRange[MAXCURVES];
    SeqlRangeType   yRangeType[MAXCURVES];
    SeqlRange	    yRange[MAXCURVES];

    SeqlType	    type;		/* the enumerated type of Seq'l plot */
    XPoint	   *devicePoints;	/* device coordinates for dataPoints */

    BOOLEAN	    setOwnColors;	/* if TRUE user will fill in pixels */
    unsigned long   forePixel; 
    unsigned long   backPixel;
    unsigned long  *dataPixel;		/* an array of pixel values for data */

    BOOLEAN	    dontLinearScaleX;	/* if TRUE don't make nice X scales */
    BOOLEAN	    dontLinearScaleY;	/* if TRUE don't make nice Y scales */
    unsigned long   w;
    unsigned long   h;

    /* these (X0,Y0), (X1,Y1) are relative to main window/pixmap */
    unsigned long    dataX0;		/* upper left corner - data region X */
    unsigned long    dataY0;		/* upper left corner - data region Y */
    unsigned long    dataX1;		/* lower right corner - data region X */
    unsigned long    dataY1;		/* lower right corner - data region Y */

    BOOLEAN	    setOwnFonts;	/* if TRUE user will fill in fonts */
    char	   *title;
    char	   *titleFont;
    XFontStruct	   *titleFontStruct;
    int		    titleFontHeight;
    int		    titleFontWidth;
    char	   *xAxisLabel;
    char	   *yAxisLabel;
    char	   *axesFont;
    XFontStruct	   *axesFontStruct;
    int		    axesFontHeight;
    int		    axesFontWidth;

    BOOLEAN	    legend;
    char	   *legendTitle;
    char          **legendArray;
    Window	    legendWindow;

    caddr_t	    userData;		/* pointer to arbitrary user data */

  };

typedef struct _Seql Seql;


extern Seql *seqlInit();
extern void seqlDraw();
extern void seqlPrint();
extern void seqlRefresh();
extern void seqlResize();
extern void seqlSet();
extern void seqlSetRange();
extern void seqlSetRangeDefault();
extern void seqlSetDisplayType();
extern void seqlSetDataColor();
extern void seqlTerm();

extern void seqlSetLegend();
extern void seqlDrawLegend();
