/*Strip.h,v 1.1.1.1 1999/08/03 14:44:00 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 * Strip structure definition
 */

#include <X11/Xlib.h>

#ifdef GTA
#ifndef WIN32
#  include <sys/time.h>
#endif WIN32
#else
#  include <X11/Intrinsic.h>
#endif

#include "GraphXMacros.h"
#include "ca_strip.h"

#define STRIP_MARGIN            0.125    /* window fraction for text, &c */
#define STRIP_DRAWINGAREA       0.75     /* 2*MARGIN + DRAWINGAREA == 1 */


#ifdef GTA
extern caddr_t fdmgr_add_timeout();
extern void fdmgr_clear_timeout();
#endif


/* must define the filename of the temporary xwd file for printing */

#ifndef XWD_TEMP_FILE_STRING
#  define XWD_TEMP_FILE_STRING    "./strip.xwd"
#endif



#ifndef MAXCURVES
#  define MAXCURVES 10		/* maximum number of curves/traces per plot */
#endif


typedef enum {
    StripInternal,		/* make private copies of all strip data */
    StripExternal		/* use the passed-in storage for all data */
} StripStorageMode;

 
typedef struct {
    double minVal;
    double maxVal;
} StripRange;


 
struct _Strip {
    Display	   *display;
    int		    screen;
    Window	    window;
    Pixmap	    pixmap;
    GC		    gc;
    StripStorageMode storageMode;

    int		    nChannels;		/* the # of channels in this strip */
    double	  **dataBuffer;		/* an array of data buffers */
    int             bufferSize;
    unsigned long  *bufferPtr;		/* an array of data buffer pointers */
    int		   *nValues;		/* an array of data buffer counters */

    StripRange      valueRange[MAXCURVES];

    double	    samplingInterval;	/* sampling interval in secs */
#ifdef OLD_STRIP
    double	 (**getValue)();	/* array of ptrs. to fns. 
					   returning sampled value */
#else
    strip_channel **caStrip;		/* A ca_strip data control */

#endif
    BOOLEAN	    paused;		/* state variable: is strip paused? */

    /* these (X0,Y0), (X1,Y1) are relative to main window/pixmap */
    unsigned long    dataX0;		/* upper left corner - data region X */
    unsigned long    dataY0;		/* upper left corner - data region Y */
    unsigned long    dataX1;		/* lower right corner - data region X */
    unsigned long    dataY1;		/* lower right corner - data region Y */
    unsigned long    dataWidth;		/* width of data region */
    unsigned long    dataHeight;	/* height of data region */

    Pixmap	    dataPixmap;		/* another pixmap for sliding data */
    double	    adjustedInterval;	/* adjusted sampling interval in secs */

    BOOLEAN	    setOwnColors;	/* if TRUE user will fill in pixels */
    unsigned long   forePixel; 
    unsigned long   backPixel;
    unsigned long   *dataPixel;		/* array of pixel values for data */

    unsigned long   w;
    unsigned long   h;

    BOOLEAN	    setOwnFonts;	/* if TRUE user will fill in fonts */
    char	   *title;
    char	   *titleFont;
    XFontStruct	   *titleFontStruct;
    int		    titleFontHeight;
    int		    titleFontWidth;
    char	   *xAxisLabel;
    char	   *yAxisLabel;
    char	   *axesFont;
    XFontStruct	   *axesFontStruct;
    int		    axesFontHeight;
    int		    axesFontWidth;

    BOOLEAN         legend;
    char           *legendTitle;
    char          **legendArray;
    Window          legendWindow;

    caddr_t	    userData;		/* pointer to arbitrary user data */

#ifdef GTA
    caddr_t	    pfdctx;		/* pointer to file descriptor context */
    caddr_t	    palarm;		/* pointer to fdm `alarm' */
#else
    XtAppContext    appContext;		/* default or user-created AppContext */
    XtIntervalId    timeoutId;
#endif

    };
typedef struct _Strip Strip;


extern Strip *stripInit();
extern void stripDraw();
extern void stripPrint();
extern void stripRefresh();
extern void stripResize();
extern void stripSet();
extern void stripSetDataColor();
extern void stripSetRange();
extern void stripSetInterval();
extern void stripPause();
extern void stripResume();
extern void stripTerm();

extern void stripSetLegend();
extern void stripDrawLegend();

#ifdef GTA
#else
extern void stripSetAppContext();
#endif
