/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
#ifdef TOKENPTRS
/* Replace this definition file with token type references in mirror file */
#include "dldef_tokens.h"

/* Prevent further processing */
#define INCdldefh
#endif

#ifndef INCdldefh
#define INCdldefh

/*
 *	D L D E F . H
 *
 *	OPI display list definition
 *	(The source of dldef.h is dldef.m4 - changes to dldef.h will
 *	be superseded)
 *
 *	Author 	Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 */

/*
 * NOTE: the display list elements currently contain nothing larger
 * than a short. If this changes RISC alignment issues will need to be
 * revisited.
 */


/* Standard include info */

/*
 * display list version
 */
#define DL_MAJOR_VERSION 1
#define DL_MINOR_VERSION 22

/*
 * magic number identifying properly byte swaped dl segments
 */
#define DL_MAGIC 0x12345678

/* round from bytes to words */
#define WDRND(A) ( ((A)>>1) + (((A)&3) ? (((A)&2)? 1 : 2) : 0) )

#define FALSE				0
#define TRUE				1
#define OK				0
#ifndef	ERROR
#	define ERROR			(-1)
#endif
#define OFFSET(structure, member)	((int) &(((structure *) 0) -> member))
#define MEMBER_SIZE(structure, member)	(sizeof (((structure *) 0) -> member))
#ifndef	NELEMENTS
#    define NELEMENTS(array)		(sizeof (array) / sizeof ((array) [0]))
#endif	NELEMENTS
#define FOREVER				for (;;)
#define max(x, y)			(((x) < (y)) ? (y) : (x))
#define min(x, y)			(((x) < (y)) ? (x) : (y))

#define XREC_TO_DLOBJ(XREC, DLOBJ)\
{\
(DLOBJ).x = (XREC).x;\
(DLOBJ).y = (XREC).y;\
(DLOBJ).width = (XREC).width;\
(DLOBJ).height = (XREC).height;\
}

#define DLOBJ_TO_XREC(DLOBJ, XREC)\
{\
(XREC).x = (DLOBJ).x;\
(XREC).y = (DLOBJ).y;\
(XREC).width = (DLOBJ).width;\
(XREC).height = (DLOBJ).height;\
}

#define DL_MINCOLOR 0
#define DL_MIDCOLOR (DL_MAXCOLOR>>1)
#define DL_MAXCOLOR ((unsigned char) ~0)

#define MAX_NPOINTS	50
#define MAX_RULE_LINES	32

enum dl_data_type { 
	dl_char,
	dl_uchar,
	dl_pchar,
	dl_short,
	dl_ushort,
	dl_pshort,
	dl_int,
	dl_uint,
	dl_pint,
	dl_float,
	dl_pfloat,
	dl_double,
	dl_pdouble,
	dl_field,
	dl_ufield,
	dl_struct,
	dl_pc_index,
	dl_string_index,
	dl_color_index,
	dl_bckgnd_color_index,
 };

enum edd_user_input { 
	edd_none,
	edd_valuator,
	edd_text_edit,
	edd_message,
	edd_menu,
	edd_button,
	edd_palette,
	edd_form,
	edd_subform,
 };

enum edd_valuator_constraint_type { 
	edd_no_constraint,
	edd_x_pos,
	edd_y_pos,
	edd_width,
	edd_height,
	edd_x_pos_on_root,
	edd_y_pos_on_root,
	edd_width_on_root,
	edd_height_on_root,
	edd_begin_angle,
	edd_line_width,
	edd_intensity,
	edd_path_angle,
 };

enum dl_dynamic_color_mod { 
	dl_clrmod_static,
	dl_clrmod_rule,
 };

enum dl_color_rule_connector { 
	dl_clrrule_use,
	dl_clrrule_and,
	dl_clrrule_andnot,
};

enum dl_color_rule_operation {
	dl_clrrule_eq,
	dl_clrrule_ne,
	dl_clrrule_ge,
	dl_clrrule_gt,
	dl_clrrule_le,
	dl_clrrule_lt,
};

enum dl_dynamic_visibility_mod { 
	dl_vismod_static,
	dl_vismod_not_zero,
	dl_vismod_zero,
 };

enum dl_dynamic_io_type { 
	dl_io_monitor,
	dl_io_control,
 };

enum dl_dynamic_geom_mod { 
	dl_geommod_right,
	dl_geommod_up,
	dl_geommod_left,
	dl_geommod_down,
	dl_geommod_radial,
 };

enum dl_dynamic_direction { 
	dl_increase_right,
	dl_increase_left,
	dl_increase_up,
	dl_increase_down,
	dl_increase_radial,
 };

enum dl_pointer_direction { 
	dl_pointer_down,
	dl_pointer_up,
	dl_pointer_right,
	dl_pointer_left,
 };

enum dl_button_stacking { 
	dl_stack_column,
	dl_stack_row,
	dl_stack_row_column,
 };

enum dl_dynamic_fill_mod { 
	dl_edge_fill,
	dl_zero_fill,
 };

enum dl_dynamic_label_type { 
	dl_label_none,
	dl_label_outline,
	dl_label_limits,
	dl_label_channel,
 };

enum dl_valuator_label_type { 
	dl_valuator_label_none,
	dl_valuator_label_outline,
	dl_valuator_label_limits,
 };

enum dl_fill_style { 
	dl_fill_outline,
	dl_fill_solid,
 };

enum dl_edge_style { 
	dl_edge_solid,
	dl_edge_dash,
 };

enum dl_plot_style { 
	dl_point_plot,
	dl_line_plot,
	dl_fill_plot,
 };

enum dl_time_units { 
	dl_day,
	dl_hour,
	dl_min,
	dl_sec,
	dl_msec,
	dl_usec,
 };

enum dl_string_alignment { 
	dl_horiz_left,
	dl_horiz_right,
	dl_horiz_center,
	dl_vert_top,
	dl_vert_bottom,
	dl_vert_center,
 };

enum dl_string_format { 
	dl_format_f,
	dl_format_e,
	dl_format_g,
	dl_format_i,
	dl_format_x,
	dl_format_o,
 };

enum dl_element_class { 
	dl_class_attribute,
	dl_class_graphic,
	dl_class_control,
	dl_class_monitor,
	dl_class_misc,
	dl_class_internal,
	dl_class_plot,
	dl_class_related_display,
	dl_class_execute_script,
	dl_class_kill_display,
 };

enum dl_on_off { 
	dl_off,
	dl_on,
 };

enum dl_elements { 
	dl_anygo,
	dl_rec,
	dl_line,
	dl_oval,
	dl_arc,
	dl_text,
	dl_user,
	dl_anymo,
	dl_bar,
	dl_tank,
	dl_meter,
	dl_indicator,
	dl_text_update,
	dl_symbol,
	dl_anypo,
	dl_plot,
	dl_smith_overlay,
	dl_surface_plot,
	dl_strip,
	dl_anyco,
	dl_valuator,
	dl_text_entry,
	dl_button,
	dl_choice_button,
	dl_message_button,
	dl_menu,
	dl_2d_control,
	dl_related_display,
	dl_execute_script,
	dl_kill_display,
	dl_group,
	dl_pc,
	dl_string,
	dl_color_map,
	dl_color_rule_table,
	dl_basic_attribute,
	dl_dynamic_attribute,
 };

struct dl_element_hdr{
	unsigned char	type;	/* element type */ 
	char		pad;	/* available */ 
	unsigned short	nwords;	/* shrt wds to nxt elmnt */ 
}; 

struct dl_object_hdr{
	short	x;		/* horizontal position */ 
	short	y;		/* vertical position */ 
	short	width;		/* horizontal size */ 
	short	height;		/* vertical size */ 
}; 

struct dl_point{
	short	x;		/* horizontal position */ 
	short	y;		/* vertical position */ 
}; 

struct dl_plot_hdr{
	short		pad;		/* available */
	char		*title;		/* plot title */ 
	char		*xlabel;	/* x axis title */ 
	char		*ylabel;	/* y axis title */ 
	char		*package;	/* plot package */ 
	unsigned char	clr;		/* foreground color */ 
	unsigned char	bclr;		/* background color */ 
}; 

struct dl_control_hdr{
	short		pad;		/* available */ 
	char 		*ctrl;		/* channel to control */ 
	unsigned char	clr;		/* foreground color */ 
	unsigned char	bclr;		/* background color */ 
}; 

struct dl_monitor_hdr{
	short		pad;		/* available */ 
	char		*rdbk;		/* channel to monitor */ 
	unsigned char	clr;		/* foreground color */ 
	unsigned char	bclr;		/* background color */ 
}; 

struct dl_sta_attr{
	unsigned char	clr;		/* color */ 
	unsigned char	style;		/* edge style */ 
	unsigned char	fill;		/* fill style */ 
	unsigned char	width;		/* line width */ 
}; 

struct dl_dyn_mod{
	unsigned char	clrmod;		/* color modifier */ 
	unsigned char	vis;		/* visibility modifier */ 
}; 

struct dl_dyn_param{
	unsigned char	rulecolorfg;	/* result of color rule execution */ 
	unsigned char	rulecolorbg;	/* unused result of color rule execution */ 
	char 		*chan;		/* channel for dynamics */ 
	char 		*rulechan [MAX_RULE_LINES];	/* channels for dynamic color rules */ 
	char 		*clrrule;	/* color rule name */ 
	char 		*clrargs;	/* color rule arguments */ 
}; 

struct dl_dyn_attr{
	struct dl_dyn_mod 	mod;	/* dynamic modifiers */ 
	struct dl_dyn_param	param;	/* dynamic parameters */ 
}; 

struct dl_color{
	unsigned char	r;		/* red intensity */ 
	unsigned char	g;		/* green intensity */ 
	unsigned char	b;		/* blue intensity */ 
	unsigned char	inten;		/* greyscale intensity */ 
	unsigned short	blink;		/* blink enable */ 
	short		pad;		/* available */ 
}; 

struct dl_color_rule_info{
	char		*chan;		/* channel */ 
	char 		*value;		/* value */ 
	unsigned char	connector;	/* logical connector */
	unsigned char	operation;	/* relational operator */
	unsigned char	clr;		/* color */ 
	char		pad;		/* available */ 
}; 

struct dl_color_rule{
	char				*name;			/* rule name */ 
	struct dl_color_rule_info	info [MAX_RULE_LINES];	/* rule information */ 
	unsigned short  		fg_enable;		/* foreground enable */ 
	unsigned short  		bg_enable;		/* background enable */ 
	unsigned char  			default_fg;		/* foreground color */ 
	unsigned char  			default_bg;		/* background color */ 
}; 

struct dl_pen{
	char		*chan;		/* pen channel */ 
	unsigned char	clr;		/* pen color */ 
	char		pad1;		/* available */ 
	char		pad2;		/* available */ 
	char		pad3;		/* available */ 
}; 

struct dl_trace{
	char		*xdata;		/* x data channel */ 
	char		*ydata;		/* y data channel */ 
	unsigned char	data_clr;	/* data color */ 
	char		pad1;		/* available */ 
	char		pad2;		/* available */ 
	char		pad3;		/* available */ 
}; 

struct dl_rd_info{
	char		*label;		/* menu label */ 
	char		*name;		/* display name */ 
	char		*args;		/* display arguments */ 
}; 

struct dl_es_info{
	char		*label;		/* menu label */ 
	char		*path;		/* script path */ 
	char		*args;		/* script arguments */ 
}; 

struct dl_hdr{
	unsigned int		magic;		/* display list magic value */ 
	unsigned short		majv;		/* display list major version */ 
	unsigned short		mnrv;		/* display list minor version */ 
	unsigned short		ndyn;		/* number of dynamic graphical */ 
	unsigned short		npc;		/* number of process connections */ 
	unsigned short		nstr;		/* number of strings */ 
	unsigned short		nmon;		/* number of monitor elements */ 
	unsigned short		nctrl;		/* number of control elements */ 
	unsigned short		nplot;		/* number of plot elements */ 
	unsigned short		nrd;		/* number of related display elements */ 
	unsigned short		nes;		/* number of execute script elements */ 
	unsigned short		nkd;		/* number of kill display elements */ 
	struct dl_object_hdr	object;		/* object header */ 
	unsigned char		clr;		/* foreground color */ 
	unsigned char		bclr;		/* background color */ 
	unsigned int		nwords_dspy;	/* shrt wds in display */ 
	unsigned int		nwords_sta;	/* shrt wds in static element portion */ 
	unsigned int		nwords_cmap;	/* shrt wds in clr map portion */ 
	unsigned int		nwords_crules;	/* shrt wds in clr rules portion */ 
	unsigned int		odyn;		/* shrt wd offset to dynamics */ 
	unsigned int		osta;		/* shrt wd offset to statics */ 
	unsigned int		omon;		/* shrt wd offset to monitors */ 
	unsigned int		octrl;		/* shrt wd offset to controls */ 
	unsigned int		oplot;		/* shrt wd offset to plots */ 
	unsigned int		ord;		/* shrt wd offset to related displays */ 
	unsigned int		oes;		/* shrt wd offset to execute scripts */ 
	unsigned int		okd;		/* shrt wd offset to kill displays */ 
	unsigned int		opc;		/* shrt wd offset to proc conns */ 
	unsigned int		ostr;		/* shrt wd offset to strings */ 
	unsigned int		ocmap;		/* shrt wd offset to clr map */ 
	unsigned int		ocrules;	/* shrt wd offset to clr rules */ 
	unsigned char		style;		/* default edge style */ 
	unsigned char		fill;		/* default fill style */ 
	unsigned char		width;		/* default line width */ 
	unsigned char		clrmod;		/* default color modifier */ 
	unsigned char		vismod;		/* visibility modifier */ 
	unsigned char		exitaction;	/* 0=> terminate, 1=>iconize */ 
	short			pad2;		/* available */ 
	char			*clrrule;	/* default color rule name */ 
	char			*pv;		/* default channel */ 
	char			*cmap;		/* external color palette display name */
	char			*startprogram;	/* Script to be run at startup */
	char			*privname;	/* Priv needed to run this display */
}; 

struct dl_anygo{
	struct dl_element_hdr	element;	/* element header */ 
	struct dl_object_hdr	object;		/* object header */ 
	short			groupid;	/* id if in a group */ 
}; 

struct dl_rec{
	struct dl_element_hdr	element;	/* element header */ 
	struct dl_object_hdr	object;		/* object header */ 
	short			groupid;	/* id if in a group */ 
}; 

struct dl_line{
	struct dl_element_hdr	element;	/* element header */ 
	struct dl_object_hdr	object;		/* object header */ 
	short			groupid;	/* id if in a group */ 
	unsigned short		npoints;	/* n points following header */ 
	struct dl_point		point [MAX_NPOINTS]; /* point parameters */ 
	struct dl_point		orgpoint [MAX_NPOINTS]; /* original point parameters */ 
}; 

struct dl_oval{
	struct dl_element_hdr	element;	/* element header */ 
	struct dl_object_hdr	object;		/* object header */ 
	short			groupid;	/* id if in a group */ 
}; 

struct dl_arc{
	struct dl_element_hdr	element;	/* element header */ 
	struct dl_object_hdr	object;		/* object header */ 
	short			groupid;	/* id if in a group */ 
	short			begin;		/* begin angle (64's of a degree) */ 
	short			path;		/* path length angle (64's of a degree) */ 
}; 

struct dl_text{
	struct dl_element_hdr	element;	/* element header */ 
	struct dl_object_hdr	object;		/* object header */ 
	short			groupid;	/* id if in a group */ 
	short			pad;		/* available */ 
	char			*textix;	/* text string */ 
	unsigned short		align;		/* text alignment */ 
}; 

struct dl_user{
	struct dl_element_hdr	element;	/* element header */ 
	struct dl_object_hdr	object;		/* object header */ 
	short			groupid;	/* id if in a group */ 
	short			pad;		/* available */ 
	char			*param [10];	/* user parameters */ 
}; 

struct dl_anymo{
	struct dl_element_hdr	element;	/* element header */ 
	struct dl_object_hdr	object;		/* object header */ 
	short			groupid;	/* id if in a group */ 
	struct dl_monitor_hdr	monitor;	/* monitor header */ 
}; 

struct dl_bar{
	struct dl_element_hdr	element;	/* element header */ 
	struct dl_object_hdr	object;		/* object header */ 
	short			groupid;	/* id if in a group */ 
	struct dl_monitor_hdr	monitor;	/* monitor header */ 
	unsigned char		label;		/* label type */ 
	unsigned char		clrmod;		/* color modifier */ 
	char			*clrrule;	/* color rule name */ 
	char			*clrargs;	/* color rule arguments */ 
	char			*rulechan [MAX_RULE_LINES]; /* channels for dynamic color rules */ 
	unsigned char		direction;	/* direction of increase */ 
	unsigned char		fillmod;	/* fill modifier */ 
	unsigned char		rulecolorfg;	/* result of color rule execution */ 
	unsigned char		rulecolorbg;	/* result of color rule execution */ 
}; 

struct dl_tank{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_monitor_hdr monitor; /* monitor header */ 
	unsigned char  label; /* label type */ 
	unsigned char  clrmod; /* color modifier */ 
	char *  clrrule; /* color rule name */ 
	char *  clrargs; /* color rule arguments */ 
	char *  rulechan [MAX_RULE_LINES]; /* channels for dynamic color rules */ 
	unsigned char  fillmod; /* fill modifier */ 
	unsigned char  direction; /* direction of increase */ 
	unsigned char  rulecolorfg; /* result of color rule execution */ 
	unsigned char  rulecolorbg; /* result of color rule execution */ 
}; 

struct dl_meter{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_monitor_hdr monitor; /* monitor header */ 
	unsigned char  label; /* label type */ 
	unsigned char  clrmod; /* color modifier */ 
	char *  clrrule; /* color rule name */ 
	char *  clrargs; /* color rule arguments */ 
	char *  rulechan [MAX_RULE_LINES]; /* channels for dynamic color rules */ 
	unsigned char  rulecolorfg; /* result of color rule execution */ 
	unsigned char  rulecolorbg; /* result of color rule execution */ 
}; 

struct dl_indicator{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_monitor_hdr monitor; /* monitor header */ 
	unsigned char  label; /* label type */ 
	unsigned char  clrmod; /* color modifier */ 
	char *  clrrule; /* color rule name */ 
	char *  clrargs; /* color rule arguments */ 
	char *  rulechan [MAX_RULE_LINES]; /* channels for dynamic color rules */ 
	unsigned char  direction; /* pointer direction */ 
	char  pad; /* available */ 
	unsigned char  rulecolorfg; /* result of color rule execution */ 
	unsigned char  rulecolorbg; /* result of color rule execution */ 
}; 

struct dl_text_update{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_monitor_hdr monitor; /* monitor header */ 
	unsigned char  clrmod; /* color modifier */ 
	char  pad; /* available */ 
	char *  clrrule; /* color rule name */ 
	char *  clrargs; /* color rule arguments */ 
	char *  rulechan [MAX_RULE_LINES]; /* channels for dynamic color rules */ 
	unsigned char  align; /* text alignment */ 
	unsigned char  format; /* text format */ 
	unsigned char  rulecolorfg; /* result of color rule execution */ 
	unsigned char  rulecolorbg; /* result of color rule execution */ 
}; 

struct dl_symbol{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_monitor_hdr monitor; /* monitor header */ 
	short  pad; /* available */ 
	char *  tbd [10]; /* symbol parameters */ 
}; 

struct dl_anypo{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_plot_hdr plotcom; /* common plot parameters */ 
}; 

struct dl_plot{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_plot_hdr plotcom; /* common plot parameters */ 
	unsigned char  style; /* plot style */ 
	unsigned char  erase_oldest; /* erase oldest sample */ 
	unsigned short  count; /* sample save count */ 
	unsigned short  pad; /* available */ 
	struct dl_trace trace [8]; /* trace parameters */ 
}; 

struct dl_smith_overlay{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_plot_hdr plotcom; /* common plot parameters */ 
	unsigned char  igrid; /* impedance grid */ 
	unsigned char  agrid; /* admittance grid */ 
	unsigned char  igrid_clr; /* impedance grid color */ 
	unsigned char  agrid_clr; /* admittance grid color */ 
	unsigned char  data_clr; /* data color */ 
	char  pad; /* available */ 
	char *  xdata; /* x data channel */ 
	char *  ydata; /* y data channel */ 
}; 

struct dl_surface_plot{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_plot_hdr plotcom; /* common plot parameters */ 
	short  pad1; /* available */ 
	char *  data; /* 2D Z data channel */ 
	unsigned char  data_clr; /* data color */ 
	char  pad2; /* available */ 
	short  dis; /* viewing distance */ 
	short  xyangle; /* xy viewing angle */ 
	short  zangle; /* z viewing angle */ 
}; 

struct dl_strip{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_plot_hdr plotcom; /* common plot parameters */ 
	unsigned short  delay; /* delay between samples */ 
	unsigned char  units; /* delay units */ 
	unsigned char  pad1; /* available */ 
	unsigned char  pad2; /* available */ 
	unsigned char  pad3; /* available */ 
	struct dl_pen pen [8]; /* pen parameters */ 
}; 

struct dl_anyco{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_control_hdr control; /* control header */ 
}; 

struct dl_valuator{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_control_hdr control; /* control header */ 
	unsigned char  label; /* label type */ 
	unsigned char  clrmod; /* color modifier */ 
	char *  clrrule; /* color rule name */ 
	char *  clrargs; /* color rule arguments */ 
	char *  rulechan [MAX_RULE_LINES]; /* channels for dynamic color rules */ 
	unsigned char  direction; /* pointer direction */ 
	char  pad; /* available */ 
	unsigned char  rulecolorfg; /* result of color rule execution */ 
	unsigned char  rulecolorbg; /* result of color rule execution */ 
}; 

struct dl_text_entry{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_control_hdr control; /* control header */ 
	unsigned char  clrmod; /* color modifier */ 
	char  pad; /* available */ 
	char *  clrrule; /* color rule name */ 
	char *  clrargs; /* color rule arguments */ 
	char *  rulechan [MAX_RULE_LINES]; /* channels for dynamic color rules */ 
	unsigned char  rulecolorfg; /* result of color rule execution */ 
	unsigned char  rulecolorbg; /* result of color rule execution */ 
}; 

struct dl_button{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_control_hdr control; /* control header */ 
	unsigned char  clrmod; /* color modifier */ 
	char  pad1; /* available */ 
	char *  clrrule; /* color rule name */ 
	char *  clrargs; /* color rule arguments */ 
	char *  rulechan [MAX_RULE_LINES]; /* channels for dynamic color rules */ 
	unsigned char  rulecolorfg; /* result of color rule execution */ 
	unsigned char  rulecolorbg; /* result of color rule execution */ 
	unsigned char  stacking; /* stacking modifier */ 
	char  pad2; /* available */ 
}; 

struct dl_choice_button{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_control_hdr control; /* control header */ 
	unsigned char  clrmod; /* color modifier */ 
	char  pad1; /* available */ 
	char *  clrrule; /* color rule name */ 
	char *  clrargs; /* color rule arguments */ 
	char *  rulechan [MAX_RULE_LINES]; /* channels for dynamic color rules */ 
	unsigned char  rulecolorfg; /* result of color rule execution */ 
	unsigned char  rulecolorbg; /* result of color rule execution */ 
	unsigned char  stacking; /* stacking modifier */ 
	char  pad2; /* available */ 
}; 

struct dl_message_button{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_control_hdr control; /* control header */ 
	unsigned short  pad1; /* available */ 
	char *  press_msg; /* press message */ 
	char *  release_msg; /* release message */ 
	unsigned char  clrmod; /* color modifier */ 
	unsigned char  pad2; /* available */ 
	unsigned short  behavior; /* momentary/latched */ 
	char *  clrrule; /* color rule name */ 
	char *  clrargs; /* color rule arguments */ 
	char *  rulechan [MAX_RULE_LINES]; /* channels for dynamic color rules */ 
	unsigned char  rulecolorfg; /* result of color rule execution */ 
	unsigned char  rulecolorbg; /* result of color rule execution */ 
};

struct dl_menu{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_control_hdr control; /* control header */ 
	unsigned char  clrmod; /* color modifier */ 
	char  pad1; /* available */ 
	unsigned char  rulecolorfg; /* result of color rule execution */ 
	unsigned char  rulecolorbg; /* result of color rule execution */ 
	unsigned short  pad2; /* available */ 
	char *  clrrule; /* color rule name */ 
	char *  clrargs; /* color rule arguments */ 
	char *  rulechan [MAX_RULE_LINES]; /* channels for dynamic color rules */ 
}; 

struct dl_2d_control{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	struct dl_control_hdr control; /* control header */ 
	unsigned short  pad; /* available */ 
	char *  yctrl; /* y control channel */ 
	unsigned char  clr; /* foreground color */ 
	unsigned char  bclr; /* background color */ 
	unsigned char  label; /* label type */ 
	unsigned char  clrmod; /* color modifier */ 
	char *  clrrule; /* color rule name */ 
	char *  clrargs; /* color rule arguments */ 
	char *  rulechan [MAX_RULE_LINES]; /* channels for dynamic color rules */ 
}; 

struct dl_related_display{
	struct dl_element_hdr	element;	/* element header */ 
	struct dl_object_hdr 	object;		/* object header */ 
	short			groupid;	/* id if in a group */ 
	unsigned char		selfkill;	/* self kill */ 
	unsigned char		pad;		/* available */ 
	struct dl_rd_info	display [8];	/* display info */ 
	unsigned char		clr;		/* foreground color */ 
	unsigned char		bclr;		/* background color */ 
}; 

struct dl_execute_script{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	unsigned short  pad; /* available */ 
	struct dl_es_info script [8]; /* script info */ 
	unsigned char  clr; /* foreground color */ 
	unsigned char  bclr; /* background color */ 
}; 

struct dl_kill_display{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
	short  groupid; /* id if in a group */ 
	unsigned short  pad; /* available */ 
	char *  label; /* button label */ 
	unsigned char  clr; /* foreground color */ 
	unsigned char  bclr; /* background color */ 
}; 

struct dl_group{
	struct dl_element_hdr element; /* element header */ 
	struct dl_object_hdr object; /* object header */ 
}; 

struct dl_pc{
	struct dl_element_hdr element; /* element header */ 
	unsigned short  count; /* n dynamic elements on this conn */ 
}; 

struct dl_string{
	struct dl_element_hdr element; /* element header */ 
	unsigned short  count; /* n references to this string */ 
}; 

struct dl_color_map{
	struct dl_element_hdr element; /* element header */ 
	unsigned short  ncolors; /* n colors following header */ 
}; 

struct dl_color_rule_table{
	struct dl_element_hdr element; /* element header */ 
	unsigned short  nrules; /* n color rules following header */ 
	unsigned short  pad; /* available */ 
}; 

struct dl_basic_attribute{
	struct dl_element_hdr element; /* element header */ 
	struct dl_sta_attr attr; /* static attributes */ 
}; 

struct dl_dynamic_attribute{
	struct dl_element_hdr element; /* element header */ 
	struct dl_dyn_attr attr; /* dynamic attributes */ 
}; 

union dl_elements_union {
	struct dl_anygo dl_anygo;
	struct dl_rec dl_rec;
	struct dl_line dl_line;
	struct dl_oval dl_oval;
	struct dl_arc dl_arc;
	struct dl_text dl_text;
	struct dl_user dl_user;
	struct dl_anymo dl_anymo;
	struct dl_bar dl_bar;
	struct dl_tank dl_tank;
	struct dl_meter dl_meter;
	struct dl_indicator dl_indicator;
	struct dl_text_update dl_text_update;
	struct dl_symbol dl_symbol;
	struct dl_anypo dl_anypo;
	struct dl_plot dl_plot;
	struct dl_smith_overlay dl_smith_overlay;
	struct dl_surface_plot dl_surface_plot;
	struct dl_strip dl_strip;
	struct dl_anyco dl_anyco;
	struct dl_valuator dl_valuator;
	struct dl_text_entry dl_text_entry;
	struct dl_button dl_button;
	struct dl_choice_button dl_choice_button;
	struct dl_message_button dl_message_button;
	struct dl_menu dl_menu;
	struct dl_2d_control dl_2d_control;
	struct dl_related_display dl_related_display;
	struct dl_execute_script dl_execute_script;
	struct dl_kill_display dl_kill_display;
	struct dl_group dl_group;
	struct dl_pc dl_pc;
	struct dl_string dl_string;
	struct dl_color_map dl_color_map;
	struct dl_color_rule_table dl_color_rule_table;
	struct dl_basic_attribute dl_basic_attribute;
	struct dl_dynamic_attribute dl_dynamic_attribute;
};
#endif
