/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/* 
 *	dynamic_def.h
 *
 *	defines for the structures needed for dynamic elements at run-time
 */


#include <gsd_sync_defs.h>	/*defs for synchronous data--may want to move */
#include <event_on_window.h>
#include <color_map.h>
#ifdef WIN32
#include <time.h>
#else
#include <sys/time.h>
#endif WIN32

#ifndef min
#define min(a,b) ((a) < (b)? (a):(b))
#endif
#ifndef FALSE
#define FALSE			0
#endif
#ifndef TRUE
#define TRUE			1
#endif
#define MAX_REL_DISPLAYS	8
#define MAX_EXE_SCRIPTS		8
#define	MAX_PAIR_STRS		((40 * 2) + 1)	/* 40 pairs + 1 for termination */
#define UPDATE_TIME		20.0
#define UPDATE_STRIP_TIME	.19
#define UPDATE_PLOT_TIME	.49

#define	USE_DEFAULT_COLOR	254
#define	BAD_COLOR_DATA		253

/*#define PLOT_NAME_DIM		(db_name_dim + 2)*/

int	the_screen;
enum	dmErrs
	{
	dmConnection_okay,
	dmNever_connected,
	dmDisconnected,
	dmInvalid_data,
	dmInvalid_range,
	dmInvalid_dbtype,
	};


/* struct for table for meters */
struct pixels
	{			/*x,y coord. for each % of full_value*/
	short	x1;
	short	y1;
	};

struct label_index
	{
	char	string[9];
	int	len;
	int	width;
	int	starting_point_x;
	int	starting_point_y;
	};

struct channel_label
	{
	char	string[15];
	int	len;
	int	width;
	int	starting_point_x;
	int	starting_point_y;
	};

struct header
	{     /* this info is included at front of each obj structure */
	char		**chan_name;
	unsigned short	precision;
	unsigned long	foregroundcolor;
	};
 
/* here are all the structures needed for object filling */
struct meter
	{
	struct header		hdr;
	int			old_degree;		/*last percent displayed*/		
	int			center_x;		/*center x position of indicator */
	int			center_y;		/*center y position of indicator */
	unsigned short		meter_width;
	unsigned short		meter_height;
	short			meter_x;
	short			meter_y;
	short			label_height_meter;
	float			full_range;		/*full range the meter spans*/
	float			low_val;
	short			outofrange;
	struct label_index	out_of_range;
	XRectangle		channel_name_rectangle;
	XRectangle		rightlimit_rectangle;
	XRectangle		leftlimit_rectangle;
	XSegment		ticks[7];
	struct label_index	left_limit;
	struct label_index	right_limit;
	struct channel_label	channel_name;
	short			label_flag;			
	struct pixels		*ppixels;
	unsigned short		precision;
	};

struct rectangle
	{
	struct header		hdr;
	int			old_x;			/*last x pixel value filled to*/
	short			label_flag;		
	float			range;			/* full range of rectangle */
	int			inside_x;		/*x,y pixel values inside rec outline*/
	int			inside_y;		
	int			new_height;		/*height inside rec, not incl outline*/
	short			outofrange;
	unsigned long		old_foreground;   	/* last foreground color */
	XRectangle		channel_name_rectangle;
	XRectangle		leftlimit_rectangle;
	XRectangle		rightlimit_rectangle;
	struct label_index	out_of_range;
	XSegment		ticks[9];
	struct label_index	left_limit;
	struct label_index	right_limit;
	struct channel_label	channel_name;
	XFontStruct		*pfont;
	unsigned short		precision;
	};

struct rec_mid
	{
	struct header		hdr;
	int			old_x;			/*last x pixel value filled to */
	int			positive_pixel_width;	/*pixel width of positive side of rec*/
	int			negative_pixel_width;	/*pixel width of negative side of rec*/
	int			pos_x;			/* left-most x pixel value of pos side*/
	int			pos_y;			/*left-most y pixel value of pos side */
	int			neg_x;			/*left-most x pixel val of neg side*/
	int			neg_y;			/*left-most y pixel val of neg side*/
	int			inside_height;		/*height inside of rec, not incl outline*/
	char			*font;			/*font size for labeling*/
	int			zero_point_x;		/*x pixel value at 0*/
	int			low_val;
	int			high_val;
	int			old_val;
	int			positive_pixels;
	int			negative_pixels;
	int			zero_point;
	short			label_flag;
	short			outofrange;
	XRectangle		channel_name_rectangle;
	XRectangle		limits_rectangle;
	XSegment		ticks[17];
	struct label_index	out_of_range;
	struct label_index	left_limit;
	struct label_index	right_limit;
	struct label_index	zero_index;
	struct channel_label	channel_name;
	unsigned short		precision;
	};

struct rec_midv
	{
	struct header		hdr;
	int			old_y;
	int			positive_pixel_height;
	int			negative_pixel_height;
	int			pos_x;
	int			pos_y;
	int			neg_x;
	int			neg_y;
	int			inside_width;
	int			zero_point_y;
	int			low_val;
	int			high_val;
	short			outofrange;
	short			label_flag;
	XRectangle		channel_name_rectangle;
	XRectangle		toplimit_rectangle;
	XRectangle		bottomlimit_rectangle;
	XRectangle		zeroindex_rectangle;
	XSegment		ticks[17];
	struct label_index	bottom_limit;
	struct label_index	top_limit;
	struct label_index	zero_index;
	struct label_index	out_of_range;
	struct channel_label	channel_name;
	unsigned short		precision;
	};	

/* struct for table for circles */
struct circle_pixels
	{		/*x,y coord. for each % of full_value*/
	short	x1;
	short	y1;
	short	x2;
	};

struct circle
	{
	struct header		hdr;
	int			old_percent;		/*last percent displayed*/
	/*float			full_range;*/		/*full range the circle spans*/
	char			*font;			/*font size for labeling*/
	int			x;		
	int			y;
	int			width;
	int			height;
	struct circle_pixels	*pcircle_pixels;
/* the following variables are for tanks */
	short			old_y;
	float			full_range;
	short			bottom_y;
	short			left_x;
	short			new_width;
	GC			pix_gc;
	Pixmap			pixid;
	};

struct rectangle_v
	{
	struct header		hdr;
	int			old_y;			/*last y pixel width filled to */
	float			full_range;		/*full range the rectangle spans*/
	char			*font;			/*font size for labeling*/
	int			bottom_y;		/*bottom y pixel value of rec,no outline*/
	int			left_x;			/*left-most x pixel value,not on outline*/
	int			new_width;		/*width to fill - not incl. outline*/
	int			rec_x;
	short			outofrange;
	short			label_flag;
	XRectangle		channel_name_rectangle;
	XRectangle		toplimit_rectangle;
	XRectangle		bottomlimit_rectangle;
	XSegment		ticks[9];
	struct label_index	out_of_range;
	struct label_index	bottom_limit;
	struct label_index	top_limit;
	struct channel_label	channel_name;
	unsigned short		precision;
	};

struct circle_conc
	{
	struct header		hdr;
	int			old_whereonradius;	/*last whereonradius value*/
	float			full_range;		/*full range this circle spans*/
	char			*font;			/*font size for labeling*/
	int			radius;			/*radius of circle*/
	int			current_x;   		/*current x,y coord. of box enclosing last fill circle*/
	int			current_y;    
	int			current_width; 		/*current width and height of box enclosing last fill*/
	int			current_height;         /*circle*/
	int			x;
	int			y;
	int			width;
	int			height;
	};

struct indicator
	{
	struct header		hdr;
	int			old_x;
	int			y1;
	int			y2;
	float			full_range;
	float			low_val;
	int			width;
	int			origin_x;
	int			origin_y;
	int			run;
	int			new_height;
	int			space;
	int			half_space;
	int			indicator_height;
	short			label_flag;
	XRectangle		indicator_rectangle;
	XRectangle		channel_name_rectangle;
	XRectangle		leftlimit_rectangle;
	XRectangle		rightlimit_rectangle;
	XRectangle		toplimit_rectangle;  /* for vertical indicator */
	XRectangle		bottomlimit_rectangle;  /* for vertical indicator */
	Pixmap			indicator_save_pixmap;
	Region			region;
	Pixmap			stipple_pixmap;
	int			index[9];
	XSegment		ticks[9];
	XPoint			points_array[4];
	short			outofrange;
	struct label_index	out_of_range;
	struct label_index	left_limit;
	struct label_index	right_limit;
	struct label_index	bottom_limit;  /* for vertical indicator */
	struct label_index	top_limit;  /* for vertical indicator */
	struct channel_label	channel_name;
	XFontStruct		*pfont;
	int			ycentered;
	};

struct tank
	{
	struct header		hdr;
	unsigned short		old_y;          /* last y value displayed */
	float			full_range;     /* full range of possible values*/
	unsigned short		bottom_y;       /* y pixel value at bottom of tank */
	unsigned short		left_x;         /* x pixel value at far left of tank*/
	unsigned short		new_width;   	/*new width after resizing to bounding box*/
	unsigned short		new_height; 	/*new height after resizing to bounding box*/
	unsigned short		curve_width;	/*curve width, for horizontal tanks only */
	unsigned short		curve_height;	/* curve height, for vertical tanks only*/
	GC			pix_gc;		/* gc for pixmap of tank */
	Pixmap			pixid;		/* pixmap id for pixmap of tank */
	XArc			arcs[2];	/* data for redrawing arcs*/
	XSegment		lines[2];	/* data for redrawing rectangle lines*/
	XSegment		ticks[7];	/* data for redrawing tick marks*/
	XRectangle		channel_name_rectangle;
	struct label_index	low_index;
	struct label_index	high_index;
	struct channel_label	channel_name; 
	XFontStruct		*pfont;
	}; 

union graphic_type
	{
	struct rectangle	rectangle;
	struct circle		circle;
	struct meter		meter;
	struct circle_conc	circle_conc;
	struct rectangle_v	rec_v;
	struct rec_mid		rec_mid;
	struct rec_midv		rec_midv;
	struct indicator	indicator;
	struct tank		tank;
	};



struct dm_llist_el
	{
	struct dm_llist_el 	*pNext_el;
	struct dl_context 	dl_context;
	struct dl_context	static_dlcontext;
	XGCValues 		GC_values;
	unsigned short		visible;	/* indicates if window is visible and should be updated */
	unsigned long		*pdl_color_map_index;	/* ptr to hw color or inten map for dl */
	struct dl_hdr 		*pdl_hdr;		/* ptr to display list hdr */
	void 			*pstatic;		/* ptr to static prt of display list */
	void 			*pdl_dynamic;		/* ptr to dynamic ptr of "      "    */
	struct rti_color_rules	*prti_color_rules;	/* ptr to run time dynamic color rule list */
	unsigned int		ncolor_rule_objs;	/* number of objects using color rules */
	struct rti_dynamic_graphic  *prti_dynamic;	/* ptr to run time dynamic info list */
	void 			*pdl_mon;
	struct rti_update_mon	*prti_mon;
	void			*pdl_ctrl;
	struct rti_ctrl_info 	*prti_ctrl;
	void			*pdl_plot;		/* ptr to plot section of display list */
	struct rti_update_plot	*prti_plot;		/* ptr to plot section of run-time plot info*/
	struct dl_related_display *pDl_related_display;	/* ptr to dl related display info */
	struct rti_rel_disp	*prti_rel_disp;		/* ptr to run time related display el list */
	struct dl_execute_script *pDl_execute_script;	/* ptr to dl execute script info */
	struct rti_exe_script	*prti_exe_script;	/* ptr to run time execute script el list */
	struct dl_kill_display	*pDl_kill_display; 	/* ptr to dl kill display info */
	struct rti_kill_display	*prti_kill_display;	/* ptr to run time kill display el list */
	struct gsdl_com_llist_element *pcom_el; 	/* ptr to corr. comm_llist */
	char			**pdl_strings;		/* array of pointers to strings in dl*/
	struct dl_color_map 	*pdl_color_map;		/* ptr to color map from dl */
	struct dl_color_rule_table *pdl_color_rules;	/* ptr to color rules from dl */
	char 			display_name[50];	/* display name for read */
	char 			display_path[100];	/* display path that the read got */
	char			**ppMacro_pairs;
	struct event_on_window 	evwin;			/* struct for window-based Xevent hndling */
	XContext 		win_context;		/* also useful for checking if display is already there */
	struct timeval		update_time;		/* last time screen was drawn */
	};

struct dm_llist_hdr
	{
	unsigned short   	new_llist_flag;		/* flags add or delete to list*/
	struct dm_llist_el	*pNew_llist;		/* ptr to el 1 on dm_llist */
	}; 
     
struct rti_dynamic_graphic
	{
	struct gsdl_com_record	*pComel;
	unsigned short 		mod_bitix;
	unsigned short 		mod_intix;
	unsigned short		not_conn;
	unsigned short		update;
	void			*pLast_val;	/* ptr to data for last value */
	char			clr;		/* color */ 
	unsigned char		style;		/* line style */ 
	unsigned char		fill;		/* fill style */ 
	unsigned char		width;		/* line width */ 
	unsigned char		clrmod;		/* color modifier */ 
	unsigned char		vis;		/* visibility modifier */ 
	unsigned int		clrrule;	/* color rule name */ 
	int			clrrule_has_changed;
	};

struct rti_color_rules
	{
	char			*prulecolorfg;
	char			*prulecolorbg;
	int			*pclrrule_has_changed;
	struct dl_color_rule	*prule;
	struct color_rule_line_info
		{
		struct gsdl_com_record	*pComel;
		} line[MAX_RULE_LINES];
	};

struct rti_update_mon
	{
	void			*pdl_mon;		/* pointer to related dyn info in dl */
	struct gsdl_com_record	*pgeom_comel;		/* pointer to channel table entry */
	unsigned short 		not_conn;		/* indicates no ca cnxn		  */
	unsigned short 		init_flag;		/* inited if true */
	int			(*init_func)();		/* function to call for initialization */
	int 			(*upd_func)();		/* function to call when updating */		
	/* do we need local hopr,lopr? (for pan,zoom ) */
	float 			hopr,lopr;
	unsigned short 		geom_modix;		/* cnl bit index for set modified */
	unsigned short 		geom_intix;		/* cnl array index for set modified */
	void 			*pLast_val;		/* last value		*/
	float 			sig_delta;		/* change great enough to display */
	/* put info for X calls for the object */
	union graphic_type 	g_type;			/* struct for appropriate type */
	int			clrrule_has_changed;
	};

struct rti_ctrl_info
	{
	void			*pdl_ctrl;		/* pointer to related dyn info in dl */
	struct gsdl_com_record	*pgeom_comel;		/* pointer to channel table entry */
	unsigned short 		not_conn;		/* indicates no ca cnxn		  */
	unsigned short 		init_flag;		/* inited if true */	
	/* do we need local hopr,lopr? (for pan,zoom ) */
	float 			hopr,lopr;
	unsigned short 		geom_modix;		/* cnl bit index for set modified */
	unsigned short 		geom_intix;		/* cnl array index for set modified */
	int			user;			/* user defined for specific object type */
	gizmo			*pgz;			/* ptr to gizmo struct	*/
	void 			*pExt;			/* ptr to mem alloced */
	void			*pLast_val;		/* ptr to data for last value */
	int			clrrule_has_changed;
	};

struct stripRange
	{
	double minVal;
	double maxVal;
	};

struct rti_update_plot
	{
	void			*pDl_plot;	/* pointer to related dyn info in dl */
	struct 	gsd_sync_data	ysync_data[8];	/* ptr to process variable info for ys*/
	struct gsd_sync_data	xsync_data[8];	/* ptr to process variable info for x */
	void			*pSync_ctrl;	/* ptr for synchronous reads */
	/* should we store type to keep from doing ca_field_type each time data is copied?*/
	struct gsdl_com_record	*pYlim[8];	/* pointer to channel table entry */
	struct gsdl_com_record	*pXlim[8];	/* pointer to channel table entry */
	double			*pPlot_data[8]; /* ptrs to data area for plot routines */
	struct stripRange	stripRange[8];	/* min, max range vals for each pv */
	caddr_t			pPlot;		/* ptr to struct for specific plot type */
	unsigned short		num_traces;	/* number of elements(pvs) to plot */
	Window			window;		/* id for plot window */
	double 			(*getPV_vals[8])();	/* ptrs to functions to get/cvrt data */
	unsigned char		init_bits;	/* inited if true */
	unsigned char		xinit_bits;	
	unsigned long		data_clrs[8];
	struct event_on_window 	evwin;		/* struct for window-based Xevent hndling */
	unsigned short		rti_plottype;	/* xyy or yy(waveform) */
	unsigned long		*pDl_colors;	/* ptr to display list colors */
	char			*xname[8];
	char			*yname[8];
	};


struct rti_rel_disp
	{
	struct dl_related_display	*pDl_related_display;		/* pointer to related dyn info in dl */
	struct dm_llist_hdr		*pDm_llist_hdr; 		/* ptr to dm list hdr */
	struct dm_llist_el		*pDm_llist_el;			/* ptr to this displays list element */
	char				*pDisplays[MAX_REL_DISPLAYS];	/* str ptrs for display names */
	char				*pMenustr[MAX_REL_DISPLAYS+1];	/* str ptrs for menu labels */
	char				*pArgList[MAX_REL_DISPLAYS+1];	/* str ptrs for arg lists */
	char				**ppMacros[MAX_REL_DISPLAYS];	/* str ptrs for arg value pairs */
	gizmo				*pgz;				/* ptr to gizmo struct */
	};

struct rti_exe_script
	{
	struct dl_execute_script	*pDl_exe_script;		/* pointer to exe script dyn info in dl */
	struct dm_llist_el		*pDm_llist_el;			/* ptr to dm list el */
	char				*pScripts[MAX_EXE_SCRIPTS];	/* str ptrs for script names */
	char				*pMenustr[MAX_EXE_SCRIPTS+1];	/* str ptrs for menu labels */
	char				*pArgList[MAX_EXE_SCRIPTS+1];	/* str ptrs for arg lists */
	char				**ppMacros[MAX_EXE_SCRIPTS];	/* str ptrs for arg value pairs */
	gizmo				*pgz;				/* ptr to gizmo struct */
	};

struct rti_kill_display
	{
	struct dl_kill_display		*pDl_kill_display;	/* pointer to exe script dyn info in dl */
	struct dm_llist_hdr		*pDm_llist_hdr; 	/*ptr to dm list hdr */
	struct dm_llist_el		*pDm_llist_el;		/*ptr to dm list el */
	gizmo				*pgz;			/*ptr to gizmo struct */
	};

struct rti_update_color
	{
	struct dl_dynamic 	*pdl_dyn;	/* pointer to related dyn info in dl */
	struct gsdl_com_record	*pgeom_comel;	/* pointer to channel table entry */
	unsigned short 		not_conn;	/* indicates no ca cnxn		  */
	unsigned short 		geom_modix;	/* cnl bit index for set modified */
	unsigned short 		geom_intix;	/* cnl array index for set modified */
	float 			last_val;	/* last value		*/
	float 			last_fg;	/* last foreground val  */
	float 			last_bg;	/* last background val  */
	};


	
static clear_mod_bit_mask[] =
	{
	0xfffffffe,
	0xfffffffd,
	0xfffffffb,
	0xfffffff7,
	0xffffffef,
	0xffffffdf,
	0xffffffbf,
	0xffffff7f,
	0xfffffeff,
	0xfffffdff,
	0xfffffbff,
	0xfffff7ff,
	0xffffefff,
	0xffffdfff,
	0xffffbfff,
	0xffff7fff,
	0xfffeffff,
	0xfffdffff,
	0xfffbffff,
	0xfff7ffff,
	0xffefffff,
	0xffdfffff,
	0xffbfffff,
	0xff7fffff,
	0xfeffffff,
	0xfdffffff,
	0xfbffffff,
	0xf7ffffff,
	0xefffffff,
	0xdfffffff,
	0xbfffffff,
	0x7fffffff
	};



#define OKAY		0		/* index for green in color table   */
#define WARNING 	1		/* index for yellow in color table  */
#define ALARM		2		/* index for red in color table     */
#define NOTCONN_COLOR	3		/* index for not connected--white   */
#define DEFAULT_BG	4		/* index for default bg color-black */

struct strip_elems
	{
	void *pdata;
	short field_type;
	};

struct dyn_strip_elems
	{
	void *pdata;
	chid chid;
	};

