/*palette_menu.c,v 1.1.1.1 1999/08/03 14:44:05 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *	P A L E T T E _ M E N U . C
 *
 *	Author: Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *
 *
 *	OPI editor color palette menus
 *	.01 joh 121191	no longer includes vxWorks.h
 *
 *
 */
#include        <stdio.h>
#include        <Xlib.h>           
#include        <Xutil.h>

#ifdef AXP
#include <stdlib.h>
#endif

void *tisMalloc();
void *tisCalloc();
void tisFree();

#include        <gizmo.h>
#include        <color_map.h>

#ifndef INCpalette_menush
#include	<palette_menus.h>
#endif

#if     !defined(NULL)
#       define NULL            0
#endif

#if     !defined(FALSE) || (FALSE!=0)
#       define FALSE           0
#endif

#if     !defined(TRUE) || (TRUE!=1)
#       define TRUE            1
#endif

#if     !defined(OK) || (OK!=0)
#       define OK              0
#endif

#if     !defined(ERROR) || (ERROR!=(-1))
#       define ERROR           (-1)
#endif

#if     !defined(NELEMENTS)
#       define NELEMENTS(ARRAY) \
                (sizeof (ARRAY) / sizeof ((ARRAY) [0]))
#endif

#if     !defined(OFFSET)
#       define OFFSET(STRUCT_DECL, FIELD) \
                ((int) &(((STRUCT_DECL *) 0) -> FIELD))
#endif

#if     !defined(max)
#       define max(x, y)       (((x) < (y)) ? (y) : (x))
#endif

#if     !defined(min)
#       define min(x, y)       (((x) < (y)) ? (x) : (y))
#endif


#define MENU_BORDER_WIDTH 2

#define COLOR_BOX_SIZE 30
#define MAX_COLOR_BOX_ROWS 5
#define COLOR_INDEX_TO_X(INDEX)		((INDEX)/MAX_COLOR_BOX_ROWS)
#define COLOR_INDEX_TO_Y(INDEX)		((INDEX)%MAX_COLOR_BOX_ROWS)
#define POINT_TO_COLOR_INDEX(X,Y)	((X)*MAX_COLOR_BOX_ROWS+(Y))



/* 
 *
 *	root for each window tree
 *
 */
typedef struct{
	Display			*d;
	GC 			foreground_GC;  	
	GC 			background_GC;  
	Window			w;
	short			x_root;
	short			y_root;
	unsigned char		selection;
	unsigned short		npixels;
	long			*ppixels;
} palette;

extern	int	invisibleColor;

/*
 *	BEWARE:
 *	You must also change event masks to move the
 *	the menu button.
 *
 */
#define MENUBUTTON 3


#define ANYWINDOW None 
static XContext palette_menu_ctx;

static redraw_action();







/*
 *
 *
 *	INIT_PALETTE_MENUS
 *
 *
 *
 */
void init_palette_menus(d)
Display				*d;
{
	XGCValues 		GC_values; 
	palette 		*ppal = NULL;
	XSetWindowAttributes 	xswa;


	if(!palette_menu_ctx)
		palette_menu_ctx = XUniqueContext();

	XFindContext(d, ANYWINDOW, palette_menu_ctx, (caddr_t *)&ppal);
	if(!ppal){

		ppal = (palette  *) tisCalloc(__FILE__, __LINE__, 1, sizeof(palette));
			/*calloc(1, sizeof(palette));*/
		if(!ppal)
			abort();

		ppal->d = d;

  		GC_values.foreground = WhitePixel(d, DefaultScreen(d));
  		GC_values.background = BlackPixel(d, DefaultScreen(d));
  		GC_values.line_style = LineOnOffDash;
  		ppal->foreground_GC = 
			XCreateGC(	d, 
					DefaultRootWindow(d),
					GCForeground|GCBackground|GCLineStyle,
					&GC_values);
		
  		GC_values.foreground = BlackPixel(d, DefaultScreen(d));
  		GC_values.background = WhitePixel(d, DefaultScreen(d));
  		ppal->background_GC = 
			XCreateGC(	d, 
					DefaultRootWindow(d),
					GCForeground|GCBackground,
					&GC_values);

		if(XSaveContext(d, ANYWINDOW, palette_menu_ctx, (XPointer)ppal))
			abort();

		xswa.background_pixel = BlackPixel(d,DefaultScreen(d));
		xswa.border_pixel = WhitePixel(d,DefaultScreen(d));
		xswa.event_mask = 	ButtonPressMask|
				ButtonReleaseMask|ButtonMotionMask|
				EnterWindowMask|LeaveWindowMask;
		xswa.bit_gravity = NorthWestGravity;
		xswa.override_redirect = TRUE;
		xswa.save_under = TRUE;

		/*
		NOTE: X bombs if you create a zero size window 
		*/
  		ppal->w = XCreateWindow(
			d,
			DefaultRootWindow(d),
			0,	/* dont know x yet */
			0,	/* dont know y yet */
		     	1,	/* dont know ht yet */
			1,	/* dont know wd yet */
			MENU_BORDER_WIDTH,
			CopyFromParent,	/* take depth from parent */
			InputOutput,
			CopyFromParent,
			CWBackPixel|CWBorderPixel|CWEventMask|CWBitGravity|
				CWOverrideRedirect|CWSaveUnder,
			&xswa);

	
	}
}




/*
 *
 *
 *	POST_PALETTE_MENU
 *
 *
 *
 */
int
post_palette_menu(pevent, ppixels, npixels)
XEvent				*pevent;
long				*ppixels;
unsigned			npixels;
{  	
	int 			status;
	XEvent			event;
	int			x, y;
	int			ht, wd;
	palette		 	*ppal = NULL;
	unsigned int			selection;
	int			color_slot;

	if(npixels == 0)
		return ERROR;

	XFindContext(pevent->xany.display, ANYWINDOW, palette_menu_ctx, (caddr_t *)&ppal);
	if(!ppal)
		return ERROR;

	npixels++;	/* add one for the Invisible color */
	ht = min(npixels,MAX_COLOR_BOX_ROWS)*COLOR_BOX_SIZE;
	wd = ( (npixels-1)/MAX_COLOR_BOX_ROWS + 1 )*COLOR_BOX_SIZE;

	/*
	 *	Force menu to come up on screen
	 *
	 */
	ppal->x_root = min(	pevent->xbutton.x_root,
				DisplayWidth(	ppal->d, DefaultScreen(ppal->d))
					- wd - (MENU_BORDER_WIDTH<<2)
				);
	ppal->y_root = min(	pevent->xbutton.y_root, 
				DisplayHeight(	ppal->d, DefaultScreen(ppal->d))
					- ht - (MENU_BORDER_WIDTH<<2)
				);
	ppal->x_root = max(	ppal->x_root, 
				0);
	ppal->y_root = max(	ppal->y_root, 
				0);

	ppal->npixels = npixels;
	ppal->ppixels = ppixels;

	/*
	 * Move window under the pointer
	 */
	XMoveResizeWindow(	ppal->d, 
				ppal->w, 
				ppal->x_root,
				ppal->y_root,
				wd,
				ht);

	/* Move pointer onto current selection */
	color_slot = (ppal->selection == invisibleColor) ? (ppal->npixels - 1) : ppal->selection;
	x = COLOR_INDEX_TO_X(color_slot)*COLOR_BOX_SIZE + (COLOR_BOX_SIZE>>1);
	y = COLOR_INDEX_TO_Y(color_slot)*COLOR_BOX_SIZE + (COLOR_BOX_SIZE>>1);
	XWarpPointer(	ppal->d, 
			None, 
			ppal->w,
			0,0,0,0,
			x,
			y);

	XMapRaised(ppal->d, ppal->w);
	redraw_action(ppal);

	while(TRUE){
		do
			kludgeblinker();
		while (!XPending(ppal->d));
		XNextEvent(ppal->d, &event);
	
		switch(event.type){
		case ButtonRelease:
			if(event.xbutton.button != MENUBUTTON)
				continue;

			/*
			 * Allways leave the pointer where they had it if
			 * leaving the root menu do this before the pointer
			 * ungrab to avoid race
			 */
			XWarpPointer(	ppal->d, 
					None, 
					DefaultRootWindow(ppal->d),
					0,0,0,0,
					ppal->x_root,
					ppal->y_root);

			XUnmapWindow(	ppal->d, ppal->w);

			/*
			 * only execute if they moused up in the menu or they
			 * released the button prior to mapping the menu
			 * (they want to quickly take the default)
			 */
			if(event.xany.window == ppal->w)
				return ppal->selection;
			else
				return ERROR;
	
		case MotionNotify:
			if(!(event.xmotion.state & Button3Mask))
				continue;
			if(event.xany.window != ppal->w)
				continue;

			selection = POINT_TO_COLOR_INDEX(	
					event.xmotion.x / COLOR_BOX_SIZE, 
					event.xmotion.y / COLOR_BOX_SIZE);

			if(selection <0 || selection >= (unsigned int)(ppal->npixels))
				continue;

			if(selection == ppal->selection || (selection == (ppal->npixels-1) && ppal->selection == invisibleColor))
				continue;

			color_slot = (ppal->selection == invisibleColor) ? (ppal->npixels - 1) : ppal->selection;
			x = COLOR_INDEX_TO_X(color_slot)*COLOR_BOX_SIZE;
			y = COLOR_INDEX_TO_Y(color_slot)*COLOR_BOX_SIZE;

			XDrawRectangle(	ppal->d, 
					ppal->w,
					ppal->background_GC,
					x,
					y,
					COLOR_BOX_SIZE,
					COLOR_BOX_SIZE);

			ppal->selection = (selection != (ppal->npixels-1)) ? selection : invisibleColor;

			x = COLOR_INDEX_TO_X(selection)*COLOR_BOX_SIZE;
			y = COLOR_INDEX_TO_Y(selection)*COLOR_BOX_SIZE;

			XDrawRectangle(	ppal->d, 
					ppal->w,
					ppal->foreground_GC,
					x,
					y,
					COLOR_BOX_SIZE,
					COLOR_BOX_SIZE);
			break;

		case EnterNotify:
			if(event.xany.window != ppal->w)
				break;

			color_slot = (ppal->selection == invisibleColor) ? (ppal->npixels - 1) : ppal->selection;
			x = COLOR_INDEX_TO_X(color_slot)*COLOR_BOX_SIZE;
			y = COLOR_INDEX_TO_Y(color_slot)*COLOR_BOX_SIZE;

			XDrawRectangle(	
					ppal->d, 
					ppal->w,
					ppal->foreground_GC,
					x,
					y,
					COLOR_BOX_SIZE,
					COLOR_BOX_SIZE);
			break;

		case LeaveNotify:
			if(event.xany.window != ppal->w)
				continue;

			color_slot = (ppal->selection == invisibleColor) ? (ppal->npixels - 1) : ppal->selection;
			x = COLOR_INDEX_TO_X(color_slot)*COLOR_BOX_SIZE;
			y = COLOR_INDEX_TO_Y(color_slot)*COLOR_BOX_SIZE;

			XDrawRectangle(	
					ppal->d, 
					ppal->w,
					ppal->background_GC,
					x,
					y,
					COLOR_BOX_SIZE,
					COLOR_BOX_SIZE);
			break;
		}
	}
}






/*
 *
 *
 *	REDRAW_ACTION
 *
 *
 *
 */
static
redraw_action(ppal)
palette			*ppal;
{
	Display		*d = ppal->d;
	Window		w = ppal->w;
	GC		gc = ppal->foreground_GC;

	int	 	x,y;
	int		i;

	for(i=0; i< ((unsigned int)(ppal->npixels) - 1); i++){

		/* would be faster if done incrementally */
		x = COLOR_INDEX_TO_X(i)*COLOR_BOX_SIZE;
		y = COLOR_INDEX_TO_Y(i)*COLOR_BOX_SIZE;

		XSetForeground(d, gc, ppal->ppixels[i]);
                XFillRectangle( d,
 				w,
 				gc,
				x,
 				y,
				COLOR_BOX_SIZE,
				COLOR_BOX_SIZE);
		if(i == ppal->selection){
			XSetForeground(d, gc, WhitePixel(d, DefaultScreen(d)));
			XDrawRectangle(	ppal->d, 
					ppal->w,
					ppal->foreground_GC,
					x,
					y,
					COLOR_BOX_SIZE,
					COLOR_BOX_SIZE);
		}
	}
	
	/* now do the Invisible Color target */
	x = COLOR_INDEX_TO_X(i)*COLOR_BOX_SIZE;
	y = COLOR_INDEX_TO_Y(i)*COLOR_BOX_SIZE;

	XSetForeground(d, gc, BlackPixel(d, DefaultScreen(d)));
	XFillRectangle(d,
 				w,
 				gc,
				x,
 				y,
				COLOR_BOX_SIZE,
				COLOR_BOX_SIZE);
	XSetForeground(d, gc, WhitePixel(d, DefaultScreen(d)));
	tisDrawString(d, w, gc, x+5, y+COLOR_BOX_SIZE-5, "~INV", 3);
	if (ppal->selection == invisibleColor)
		XDrawRectangle(ppal->d, 
					ppal->w,
					ppal->foreground_GC,
					x,
					y,
					COLOR_BOX_SIZE,
					COLOR_BOX_SIZE);
	XSetForeground(d, gc, WhitePixel(d, DefaultScreen(d)));

	return;
}
		
