/*bitfield.c,v 1.1.1.1 1999/08/03 14:44:07 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *
 *	B I T F I E L D . C
 *
 *	Author	Jeffrey O. Hill
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *
 *	Manipulate C bitfields
 *
 *
 *
 *
 * 	History
 *	.01 040291 joh	Improved documentation
 *	.02 040291 joh	now prints warning if bits are clipped 
 *			during a write.
 *	 
 *
 *	Notes
 *	.01	This module is architecture specific. 
 *		It should work on any architecture which orders bits 
 *		in decreasing order and characters in increasing
 *		order such as the 68000 or SPARC.
 *
 *		Additional code would need to be written
 *		if this must run on a machine which orders both bits 
 *		and characters in increasing order such
 *		as the VAX or the Intel 80xx.
 *
 *	.02	I have attempted to make this module compiler 
 *		independent as long as the architecture issues 
 *		described above are addressed. map_bitfield()
 *		finds the bitfield at runtime allowing this
 *		to work independent of different compiler field 
 *		position allocations.
 *
 *	.03	map_bitfield() allways returns bitfield offsets which 
 *		are less than NBBY*sizeof(short)
 *
 *	Testing
 *	.01	This module has been tested on the 68000
 *
 *	Improvements
 *	.01	test this under the SPARC architecture
 *
 */
#include <dldef.h>
#include <dlinfo.h>
#include <stdio.h>


#define NBBY 8	/* nuber of bits per byte */



/*
 *	MAP_BITFIELD
 *
 *	Find the address and bit offset of a bit field
 *
 *	structure should be cleared except for the
 *	field which should all of its bits set 
 *	prior to calling map_bitfield().
 */
map_bitfield(pstruct,size,poffset,pfoffset)
unsigned short 	*pstruct;	/* ptr to struct with field set */
unsigned 	size;		/* n bytes in this struct */
unsigned short  *poffset;	/* byte offset returned */
unsigned short	*pfoffset;	/* field offset returned */
{
	unsigned short 		*pfd;
	register unsigned short field;
	register unsigned 	offset;

	pfd = &pstruct[size/sizeof(*pstruct)-1];
	while(pfd>=pstruct){
		if(field = *pfd){
			for(offset=0;!(field&1);field=field>>1)
				offset++;
			*poffset = (unsigned) 
				( ((char *)pfd) - ((char *)pstruct) );
			*pfoffset = offset;
			return OK;
		}
		pfd--;
	}

	return ERROR;		
}



/*
 *	READ_BITFIELD
 *
 *	read an unsigned integer from a bit field
 *
 * 	
 */
read_bitfield(pds, psinfo, ival)
void		*pds;
struct sinfo 	*psinfo;
long		*ival;
{
	if(psinfo->nelements <= NBBY*sizeof(short)){
		short	*pwd = (short *) ( ((char *) pds) + psinfo->offset );
		short 	mask;
		short	wd;
		short	nbits0;
		short	nbits1;

		nbits0 = min(NBBY*sizeof(*pwd)-psinfo->fieldoffset, 
			psinfo->nelements);
		if(nbits0 > 0){
			mask = ~(~0 << nbits0);
			wd = *pwd >> psinfo->fieldoffset;
			*ival = mask & wd;
		}

		nbits1 = psinfo->nelements - nbits0;
		if(nbits1 > 0){
			mask = ~(~0 << nbits1) << nbits0;
			wd = *pwd << nbits0;
			*(pwd+1) = (~mask & *(pwd+1)) | (wd & mask);

			*ival = (*ival & ~mask) | (wd & mask);
		}
	}
	else if(psinfo->nelements <= NBBY*sizeof(int)){
		int	*pwd = (int *) (( (char *) pds) + psinfo->offset );
		int 	mask;
		int	wd;
		short	nbits0;
		short	nbits1;

		nbits0 = min(NBBY*sizeof(*pwd)-psinfo->fieldoffset, 
				psinfo->nelements);
		if(nbits0 > 0){
			mask = ~(~0 << nbits0);
			wd = *pwd >> psinfo->fieldoffset;
			*ival = mask & wd;
		}

		nbits1 = psinfo->nelements - nbits0;
		if(nbits1 > 0){
			mask = ~(~0 << nbits1) << nbits0;
			wd = *pwd << nbits0;
			*(pwd+1) = (~mask & *(pwd+1)) | (wd & mask);

			*ival = (*ival & ~mask) | (wd & mask);
		}
	}else
		abort();
	return 0;
}



/*
 *	WRITE_BITFIELD
 *
 *	write an unsigned integer to a bit field.
 *	(prevents overflows)
 *
 */
write_bitfield(pds, psinfo, ival)
void		*pds;
struct sinfo 	*psinfo;
long		ival;
{
	{
		int mask;
		int value;

		if(ival>=0)
			value = ival;
		else
			value = -ival;

		mask = ~(~0 << psinfo->nelements);
		if(mask & value != value){
			printf("write_bitfield(): warning- bits clipped\n");
		}
	}

	if(psinfo->nelements <= NBBY*sizeof(short)){
		short	*pwd = (short *) ( ((char *) pds) + psinfo->offset );
		short	wd;
		short	nbits0;
		short	nbits1;
		short 	mask;

		nbits0 = min(NBBY*sizeof(*pwd)-psinfo->fieldoffset, 
				psinfo->nelements);
		if(nbits0 > 0){
			mask = ~(~0 << nbits0) << psinfo->fieldoffset;
			wd = (short)(ival << psinfo->fieldoffset);
#ifdef DEBUG
			printf("%x - %x - %x\n", mask, wd, *pwd);
#endif
			*pwd = (~mask & *pwd) | (wd & mask);
#ifdef DEBUG
			printf("%x - %x - %x\n", mask, wd, *pwd);
#endif
		}

		nbits1 = psinfo->nelements - nbits0;
		if(nbits1 > 0){
			mask = ~(~0 << nbits1);
			wd = (short)(ival >> (nbits0));
			*(pwd+1) = (~mask & *(pwd+1)) | (wd & mask);
		}
	}
	else if(psinfo->nelements <= NBBY*sizeof(int)){
		int	*pwd = (int *) (( (char *)pds ) + psinfo->offset );
		int 	mask;
		int	wd;
		short	nbits0;
		short	nbits1;

		nbits0 = min(NBBY*sizeof(*pwd)-psinfo->fieldoffset, 
				psinfo->nelements);
		if(nbits0 > 0){
			mask = ~(~0 << nbits0) << psinfo->fieldoffset;
			wd = ival << psinfo->fieldoffset;
			*pwd = (~mask & *pwd) | (wd & mask);
		}

		nbits1 = psinfo->nelements - nbits0;
		if(nbits1 > 0){
			mask = ~(~0 << nbits1);
			wd = ival >> (nbits0);
			*(pwd+1) = (~mask & *(pwd+1)) | (wd & mask);
		}
	}else
		abort();
	return 0;
}
