/*draw_strings.c,v 1.1.1.1 1999/08/03 14:44:07 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *	D R A W _ S T R I N G S
 *
 * 	ROUTINES FOR DRAWING A STRING WITHIN A BOUNDING BOX
 *
 *	Author: Jeffrey O. Hill 08-31-90
 *		hill@luke.lanl.gov
 *		505 665 1831
 *
 *	History
 *	.01 032691 joh	Moved to a seperate source file
 *	.02 032791 joh	Added fixed string drawing routines so continuously
 *			updated strings dont flicker.
 *	.03 102291 joh	Visually indicate clipped strings
 *	.04 102291 joh	comment out .03 as requested
 */
#define USE_BEST_FONT

#include <stdio.h>
#include <dldef.h>
#include <X11/Xlib.h>      			/* the X library 		*/
#include <X11/Xutil.h>
#include "font_table.h"			/* defines for font table	*/
#include "draw_strings.h"
#include "tisXext.h"

static void	draw_vert_fixed_font();
static void	draw_horiz_fixed_font();
static void	mark_clipped_strings();


/*
 *
 * 	DRAW_STRING
 *	joh 08-31-90
 *	draw string within a bounding box
 *	(choose between vertical or horizontal formats)
 *
 *	all of these arguments instead of pdlc in case draw_sring()
 *	is needed by gizmos which dont know about pdlc
 */
draw_string(d, w, gc, prec, pfont_table, pstring, length, format)
Display			*d;
Window			w;
GC			gc;
XRectangle		*prec;
font_table 		*pfont_table;
char			*pstring;
unsigned		length;
enum string_format	format;	
{
	if(prec->width > prec->height)
		draw_horizontal_string(
				d, 
				w, 
				gc, 
				prec, 
				pfont_table, 
				pstring, 
				length,
				format);
	else
		draw_vertical_string(
				d, 
				w, 
				gc, 
				prec, 
				pfont_table, 
				pstring, 
				length,
				format);
	return 0;
}



/*
 *
 * 	DRAW_HORIZONTAL_STRING
 *	joh 08-31-90
 *	draw string within a bounding box
 * 	- choose the best font based on the string length
 *	and the width and height of the box
 *
 *	all of these arguments instead of pdlc in case draw_sring()
 *	is needed by gizmos which dont know about pdlc
 */
draw_horizontal_string(d, w, gc, prec, pfont_table, pstring, length, format)
Display			*d;
Window			w;
GC			gc;
XRectangle		*prec;
font_table 		*pfont_table;
char			*pstring;
int			length;
enum string_format	format;	
{
	XFontStruct  	*pfont = NULL;
	unsigned	best_width;

	if(prec->width == 0 || prec->height == 0)
		return 0;

	if(length == 0){
		pstring = "____";
		length = 4;
	}

	/*
	 * Binary search for the best font
	 */
	best_width = min((unsigned int)(prec->width)/length, MAXFONTWIDTH);
	best_width = max(best_width, MINFONTWIDTH);
	while(TRUE){

		pfont = pfont_table->width_to_pfinfo[best_width];
		if(pfont)
			break;

		best_width += best_width;
		best_width = min(best_width, MAXFONTWIDTH);
	}

	if(pfont->max_bounds.ascent + pfont->max_bounds.descent > (int)(prec->height)){
		int tmpht = min(prec->height, MAXFONTHEIGHT);
		pfont = pfont_table->height_to_pfinfo[tmpht];
	}

	draw_horiz_fixed_font(d, w, gc, prec, pfont, pstring, length, format);

	return 0;
}


/*
 *
 * 	DRAW_FIXED_HORIZONTAL_STRING
 *	joh 08-31-90
 *	draw string within a bounding box
 * 	- choose the best font based on the height of the box
 *
 *	all of these arguments instead of pdlc in case draw_sring()
 *	is needed by gizmos which dont know about pdlc
 */
draw_fixed_horizontal_string(d, w, gc, prec, pfont_table, pstring, length, format)
Display			*d;
Window			w;
GC			gc;
XRectangle		*prec;
font_table 		*pfont_table;
char			*pstring;
int			length;
enum string_format	format;	
{
	XFontStruct  	*pfont = NULL;

	if(prec->width == 0 || prec->height == 0)
		return 0;

	if(length == 0){
		pstring = "____";
		length = 4;
	}

	pfont = pfont_table->height_to_pfinfo[min(prec->height, MAXFONTHEIGHT)];

	draw_horiz_fixed_font(d, w, gc, prec, pfont, pstring, length, format);
	return 0;
}


/*
 *
 * 	DRAW_HORIZ_FIXED_FONT
 *	joh 08-31-90
 *	draw string within a bounding box
 *
 *	all of these arguments instead of pdlc in case draw_sring()
 *	is needed by gizmos which dont know about pdlc
 */
static void
draw_horiz_fixed_font(d, w, gc, prec, pfont, pstring, length, format)
Display			*d;
Window			w;
GC			gc;
XRectangle		*prec;
XFontStruct  		*pfont;
char			*pstring;
int			length;
enum string_format	format;	
{
	register int 	ycentered;
	register int	nchars = 0;
	int		diff;
	int		clip;

	if(pfont)
		nchars = (unsigned int)(prec->width) / pfont->max_bounds.width;

	if(pfont && nchars){
		unsigned xcentered;

	  	XSetFont(d, gc, pfont->fid);
	  	/* set up to center text vertically */
	  	ycentered = (((unsigned int)(prec->height) - (int)pfont->max_bounds.ascent - 
			(int)pfont->max_bounds.descent)/2) + 
			(int)pfont->max_bounds.ascent;


		diff = nchars - length;
		/* clip to object boundary */
		if(diff<0){
			length = nchars;
			diff=0;
			clip = TRUE;
		}
		else{
			clip = FALSE;
		}
			
		if(format == box_end){
			xcentered = (unsigned int)(prec->width) % pfont->max_bounds.width
				+ diff * (int)pfont->max_bounds.width;
		}
		else if(format == box_center){
			xcentered = (((unsigned int)(prec->width) % 
					pfont->max_bounds.width)>>1)+
				(diff>>1) *(int)pfont->max_bounds.width;

			/* if odd */
			if(diff & 1)
				xcentered += (pfont->max_bounds.width>>1);
		}
		else{
			xcentered = 0;
		}

	  	/* draw the text */
	  	tisDrawString(
			d,
			w,
			gc, 
			prec->x + xcentered,
			prec->y + ycentered,
			pstring, 
			length);

		if(clip){
			mark_clipped_strings(d, w, gc, prec, TRUE, FALSE);
		}
		
	}
	else{
#ifdef DRAW_BOX_IF_NO_FONT
		/* 
		 *	so they will know somthing was intended
		 */
		XFillRectangle(	d,
				w,
				gc,
				prec->x,
				prec->y,
				prec->width, 
				prec->height);
#endif
	}
}



/*
 *
 * 	DRAW_VERTICAL_STRING
 *	joh 10-22-90
 *	draw vertical string within a bounding box
 * 	- choose the best font based on the string length
 *	and the width and height of the box
 *
 *	all of these arguments instead of pdlc in case 
 *	draw_vertical_string() is needed 
 *	by gizmos which dont know about the dlc
 */
draw_vertical_string(d, w, gc, prec, pfont_table, pstring, length, format)
Display			*d;
Window			w;
GC			gc;
XRectangle		*prec;
font_table 		*pfont_table;
char			*pstring;
int			length;
enum string_format	format;	
{
	XFontStruct  		*pfont = NULL;
	unsigned 		best_height;

	if(prec->width == 0 || prec->height == 0)
		return 0;

	if(length == 0){
		pstring = "____";
		length = 4;
	}

	/*
	 * Binary search for the best font
	 */
	best_height = min((unsigned int)(prec->height)/length, MAXFONTHEIGHT);
	best_height = max(best_height, MINFONTHEIGHT);
	while(TRUE){

		pfont = pfont_table->height_to_pfinfo[best_height];
		if(pfont)
			break;

		best_height += best_height;
		best_height = min(best_height, MAXFONTHEIGHT);
	}

	if(pfont->max_bounds.width > (int)prec->width){
		int tmpwd = min(prec->width, MAXFONTWIDTH);
		pfont = pfont_table->width_to_pfinfo[tmpwd];
	}

	draw_vert_fixed_font(d, w, gc, prec, pfont, pstring, length, format);

	return 0;
}


/*
 *
 * 	DRAW_FIXED_VERTICAL_STRING
 *	joh 10-22-90
 *	draw vertical string within a bounding box
 * 	- choose the best font based on the width of the box
 *
 *	all of these arguments instead of pdlc in case 
 *	draw_vertical_string() is needed 
 *	by gizmos which dont know about the dlc
 */
draw_fixed_vertical_string(d, w, gc, prec, pfont_table, pstring, length, format)
Display			*d;
Window			w;
GC			gc;
XRectangle		*prec;
font_table 		*pfont_table;
char			*pstring;
int			length;
enum string_format	format;	
{
	XFontStruct  		*pfont = NULL;

	if(prec->width == 0 || prec->height == 0)
		return 0;

	if(length == 0){
		pstring = "____";
		length = 4;
	}

	pfont = pfont_table->width_to_pfinfo[min(prec->width, MAXFONTWIDTH)];

	draw_vert_fixed_font(d, w, gc, prec, pfont, pstring, length, format);

	return 0;		
}


/*
 *
 * 	DRAW_VERT_FIXED_FONT
 *	joh 10-22-90
 *	draw vertical string within a bounding box
 *
 *	all of these arguments instead of pdlc in case 
 *	draw_vertical_string() is needed 
 *	by gizmos which dont know about the dlc
 */
static void
draw_vert_fixed_font(d, w, gc, prec, pfont, pstring, length, format)
Display			*d;
Window			w;
GC			gc;
XRectangle		*prec;
XFontStruct  		*pfont;
char			*pstring;
int			length;
enum string_format	format;	
{
	register int		nchars = 0;
	unsigned		yinc;

	/* clip to object boundary */
	if(pfont){
		yinc = pfont->max_bounds.ascent + pfont->max_bounds.descent;
		nchars = prec->height / yinc;
	}

	if(pfont && nchars){
		unsigned	yoff;
		unsigned	xoff;
		char		*eos;
		int		diff;

	  	XSetFont(d, gc, pfont->fid);

		eos = pstring + length;

		xoff = prec->x;
		yoff = prec->y + (int)pfont->max_bounds.ascent;

		diff =  nchars - length;

		if(diff >= 0){
			xoff += (unsigned int)(prec->width - pfont->max_bounds.width)>>1;

			if(format == box_end){
				yoff += diff*yinc;
				yoff += prec->height%yinc;
			}
			else if(format == box_center){
				yoff += (prec->height%yinc)>>1;
				yoff += (diff>>1)*yinc;
				/* if odd */
				if(diff&1)
					yoff += (yinc>>1);
			}
		}

		while(pstring<eos){
			char		*eol;
			unsigned	ll;
	
			ll = min(nchars, length);
			eol = pstring+ll;

	  		/* draw the text */
			for(; pstring < eol; pstring++){
	  			tisDrawString(
					d,
					w,
					gc, 
					xoff,
					yoff,
					pstring, 
					1);
				yoff += yinc;
			}
			length -= ll;
			xoff += (int)pfont->max_bounds.width;

			if(xoff + pfont->max_bounds.width > 
				(unsigned int)(prec->width + prec->x)){
				break;
			}

			yoff = prec->y + (int)pfont->max_bounds.ascent;
		}


		if(pstring<eos){
			mark_clipped_strings(d, w, gc, prec, FALSE, TRUE);
		}
	}
	else{
#ifdef DRAW_BOX_IF_NO_FONT
		/* 
		 *	so they will know somthing was intended
		 */
		XFillRectangle(	d,
				w,
				gc,
				prec->x,
				prec->y,
				prec->width, 
				prec->height);
#endif
	}
}



/*
 *	MARK_CLIPPED_STRINGS
 * 	draw an X through strings which are clipped
 */
static void
mark_clipped_strings(d, w, gc, prec, horiz, vert)
Display			*d;
Window			w;
GC			gc;
XRectangle		*prec;
int			horiz;
int			vert;
{
#	define MARKER_WIDTH 3

#ifdef MARK_CLIPPED_STRINGS

	if(prec->width > MARKER_WIDTH*2 && vert){
		XDrawLine(
			d,
			w,
			gc,
			prec->x + prec->width,
			prec->y,
			prec->x + prec->width, 
			prec->y + prec->height);
		XDrawLine(
			d,
			w,
			gc,
			prec->x,
			prec->y,
			prec->x, 
			prec->y + prec->height);
		XDrawLine(
			d,
			w,
			gc,
			prec->x + prec->width-MARKER_WIDTH+1,
			prec->y,
			prec->x + prec->width-MARKER_WIDTH+1, 
			prec->y + prec->height);
		XDrawLine(
			d,
			w,
			gc,
			prec->x + MARKER_WIDTH-1,
			prec->y,
			prec->x + MARKER_WIDTH-1, 
			prec->y + prec->height);
	}

	if(prec->height > MARKER_WIDTH*2 && horiz){
		XDrawLine(
			d,
			w,
			gc,
			prec->x,
			prec->y,
			prec->x + prec->width, 
			prec->y);
		XDrawLine(
			d,
			w,
			gc,
			prec->x,
			prec->y + prec->height,
			prec->x + prec->width, 
			prec->y + prec->height);
		XDrawLine(
			d,
			w,
			gc,
			prec->x,
			prec->y + MARKER_WIDTH-1,
			prec->x + prec->width, 
			prec->y + MARKER_WIDTH-1);
		XDrawLine(
			d,
			w,
			gc,
			prec->x,
			prec->y + prec->height - MARKER_WIDTH+1,
			prec->x + prec->width, 
			prec->y + prec->height - MARKER_WIDTH+1);
	}
#endif
}
