/*open_display.c,v 1.1.1.1 1999/08/03 14:44:08 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 * O P E N _ D I S P L A Y
 * 
 * Author: 	Jeffrey O. Hill 
 *		hill@lanl.gov 
 *		(505) 665 1831
 * 
 * 
 * History
 * .01 032691 joh	written
 * 
 * 
 * 
 */
#include <stdio.h>
#include <dldef.h>
#include <errno.h>

#ifdef AXP
#include	<string.h>
#endif

char *getenv();
char *strchr();
char *strrchr();

#ifdef UNIX
#define errnoSet(E) (errno = (E))
extern errno;
#endif


#ifdef TESTING
#include <fcntl.h>
main()
{
	char 	*ppath;
	int	fd;

	ppath = getenv("LD_DISPLAY_PATH");
	if(ppath == NULL)
		return -1;

	
	fd = open_display_list_file("test.dl", ppath, O_RDONLY);


	if(fd>=0)
		printf("found it\n");
}
#endif



/*
 *	OPEN_DISPLAY_LIST_FILE()
 *	(attempt to open a file using a path string of the same format 
 *	as the "PATH" environment variable)
 *
 *	2nd and 3rd arguments can be omitted by setting them to NULL and 0
 *	respectively.
 *
 *	ppath_used is left unmodified if no file could be opened.
 *	Otherwise, ppath_used will contain the actual path used 
 *	to open a file.
 */
int open_display_list_file(pld_display_path, ppath_used, 
	length_path_used, pfile_name, flags, mode)
char 	*pld_display_path;	/* str folling the format of "PATH" env var */
char	*ppath_used;		/* pointer to where env path used is wriiten */
int	length_path_used;	/* length of above char array */
char 	*pfile_name; 		/* nxt 3 are ident to args used by open() */
int	flags;
int	mode;
{
	int 	fd;
	char	*pcolon = NULL;
	char	*pslash;
	int	name_length;
	int 	path_length;
	char	pname[128];

	if(pfile_name == NULL){
		errnoSet(EINVAL);
		return -1;
	}

	name_length = strlen(pfile_name);
	if(name_length+1 > sizeof(pname)){
		errnoSet(EINVAL);
		return -1;
	}

	if(name_length == 0){
		errnoSet(EINVAL);
		return -1;
	}

	pslash = strrchr(pfile_name, '/');
	if(pslash!=NULL || pld_display_path==NULL){
#ifdef WIN32
		fd = _open(pfile_name, flags, mode);
#else
		fd = open(pfile_name, flags, mode);
#endif WIN32
		if(ppath_used && fd>0){
			if(pslash){
				path_length = pslash-pfile_name;
			}
			else{
				path_length = 0;
			}

			if(path_length+1>length_path_used){
#ifdef WIN32
				_close(fd);
#else
				close(fd);
#endif WIN32
				errnoSet(EINVAL);
				return -1;
			}
			memcpy(ppath_used, pfile_name, path_length);
			ppath_used[path_length] = NULL;
		}
		/*
		 * errno is set by open
		 *
		 */
		return fd;
	}

	while(TRUE){

		if(pcolon){
			pld_display_path = pcolon+1;
 		}
			
		pcolon = strchr(pld_display_path, ':');
		if(pcolon){	
			path_length = pcolon-pld_display_path;
		}
		else{
			path_length = strlen(pld_display_path);
		}


		if(path_length+name_length+2 > sizeof(pname))
			continue;
		strncpy(
			pname, 
			pld_display_path, 
			path_length);

		if(pld_display_path[path_length-1] != '/'){
			pname[path_length] = '/';
			path_length++;
		}
			
		strncpy(
			&pname[path_length], 
			pfile_name,
			name_length+1);

#ifdef DEBUG
		printf("Attempting to open [%s]\n", pname);
#endif
#ifdef WIN32
		fd = _open(pname, flags, mode);
#else
		fd = open(pname, flags, mode);
#endif WIN32
		if(fd >= 0){
			if(ppath_used){
				if(path_length+1>length_path_used){
#ifdef WIN32
					_close(fd);
#else
					close(fd);
#endif WIN32
					errnoSet(EINVAL);
					return -1;
				}
				memcpy(ppath_used, pname, path_length);
				ppath_used[path_length] = NULL;
			}
			return fd;
		}

		/*
		 * errno is set by open
		 *
		 */
		if(pcolon == NULL){
			return -1;
		}
	}
}
