/*string_table.c,v 1.1.1.1 1999/08/03 14:44:08 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
 *	S T R I N G _ T A B L E . C
 *
 *
 *	A library for holding and sorting strings.
 */
#include <stdio.h>
#if defined(AXP) || defined(WIN32) || defined(SOLARIS)
#include	<string.h>
#include	<stdlib.h>
#else
#endif

#include <dldef.h>
#include <string_table.h>

void stringInitList(pst)
LIST	*pst;
	{
	lstInit(pst);
	}
	
	
void stringZeroCounts(pst)
LIST	*pst;
	{
	StringItem	*p;

	for (p = (StringItem *)lstFirst(pst); p; p = (StringItem *)lstNext((NODE*)p))
		p->count = 0;
	}


char *stringInsert(pst, ptr)
LIST	*pst;
char	*ptr;
	{
	/*
	 * allocate a StringItem and storage space for the given string
	 * add the StringItem onto the given list and copy the string
	 */
	StringItem	*p;
			
	for (p = (StringItem *)lstFirst(pst); p; p = (StringItem *)lstNext((NODE*)p))
		if (strcmp(ptr, p->string) == 0)
			/* already in here */
			return p->string;

	p = (StringItem *)calloc(1, sizeof (StringItem));
	p->string = calloc(strlen(ptr)+1, sizeof (char));
	lstAdd(pst, &p->node);
	return strcpy(p->string, ptr);
	}


int stringCount(pst)
LIST	*pst;
	{
	return lstCount(pst);
	}


void stringDeleteList(pst)
LIST	*pst;
	{
	StringItem	*p;

	/*
	 * free string space and then list space
	 */
	for (p = (StringItem *)lstFirst(pst); p; p = (StringItem *)lstNext((NODE*)p))
		free(p->string);
	lstFree(pst);
	}


int stringName(pst, ptr, ppstring, pstring_len)
LIST		*pst;
char		*ptr;
char		**ppstring;
unsigned	*pstring_len;
	{
	*ppstring = ptr;
	*pstring_len = strlen(ptr);
	return OK;
	}


int stringGetIndex(pst, ptr)
LIST	*pst;
char	*ptr;
	{
	StringItem	*p;
	int			index;

	/*
	 * run the list looking for a match and return the index when found
	 */
	for (p = (StringItem *)lstFirst(pst), index = 0; p; p = (StringItem *)lstNext((NODE*)p), index++)
		if (strcmp(ptr, p->string) == 0)
			return index;
	return 0;
	}


int stringGetIndexAndIncCount(pst, ptr)
LIST	*pst;
char	*ptr;
	{
	StringItem	*p;
	int			index;

	/*
	 * run the list looking for a match and return the index when found
	 */
	for (p = (StringItem *)lstFirst(pst), index = 0; p; p = (StringItem *)lstNext((NODE*)p), index++)
		if (strcmp(ptr, p->string) == 0)
			{
			p->count++;
			return index;
			}
	return 0;
	}


char *stringGetPtr(pst, ptr)
LIST	*pst;
char	*ptr;
	{
	StringItem	*p;

	/*
	 * run the list looking for a match and return the strings pointer when found
	 */
	for (p = (StringItem *)lstFirst(pst); p; p = (StringItem *)lstNext((NODE*)p))
		if (strcmp(ptr, p->string) == 0)
			return p->string;
	return NULL;
	}


void stringTraverse(pst, func, arg)
LIST	*pst;
void	(*func)();
void	*arg;
	{
	StringItem	*p;

	/*
	 * run the list calling the function
	 */
	for (p = (StringItem *)lstFirst(pst); p; p = (StringItem *)lstNext((NODE*)p))
		(*func)(arg, p);
	}


void stringSearchAndExecute(pst, ptr, func, arg)
LIST	*pst;
char	*ptr;
void	(*func)();
void	*arg;
	{
	StringItem	*p;

	/*
	 * run the list looking for a match and return the strings pointer when found
	 */
	for (p = (StringItem *)lstFirst(pst); p; p = (StringItem *)lstNext((NODE*)p))
		if (strcmp(ptr, p->string) == 0)
			{
			(*func)(arg, p);
			return;
			}
	}

void stringPrintList(pst)
LIST	*pst;
	{
	StringItem	*p;

	for (p = (StringItem *)lstFirst(pst); p; p = (StringItem *)lstNext((NODE*)p))
		printf("%03d %s\n", p->count, p->string);
	}
