/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

#include  <stdio.h>
#include  <db_access.h>
#include  <dbDefs.h>
#include  <time.h>
#include  <alarmlist.h>

static struct alarm *page_head, *page_tail, *print_head, *print_tail, 
		    *signal_head, *signal_tail, *retry_head, *retry_tail;
/*************************************************************************************/
void init_alarmlist()
{
    page_head = NULL;
    page_tail = NULL;
    print_head = NULL;
    print_tail = NULL;
    signal_head = NULL;
    signal_tail = NULL;
    retry_head = NULL;
    retry_tail = NULL;
}

/***********************************************************************************/
struct alarm *construct(pointname, pchid, message, message_syntax, error, area, timeoccured, 
			timetopage, timetoprint, timetosignal, value, desc, alarmset) 

char    pointname[PVNAME_SZ + 1];
chid    pchid;
char    message[MAX_MESSAGE + 1];
char    message_syntax[MAX_MESSAGE + 1];
char    error[MAX_ERROR + 1];
int     area;
time_t  timeoccured;
time_t  timetopage;
time_t  timetoprint;
time_t  timetosignal;
char    value[MAX_STRING_SIZE + 1];
char    desc[MAX_DESCRIPTION+1];
int     alarmset;

{
   struct alarm *node = (struct alarm *)malloc(sizeof(struct alarm));

   strcpy(node->pointname,pointname); 
   node->pchid = pchid;
   strcpy(node->message, message);
   strcpy(node->message_syntax, message_syntax);
   strcpy(node->error, error);
   node->area = area;
   node->time_of_occurance =  timeoccured;
   node->time_to_page = timetopage;
   node->time_to_print = timetoprint;
   node->time_to_signal = timetosignal;
   strcpy(node->value, value);
   strcpy(node->desc, desc);
   node->critical_alarm_set = alarmset;
   node->prev_page = NULL;	      
   node->next_page = NULL;	      
   node->prev_print = NULL;	      
   node->next_print = NULL;	      
   node->prev_signal = NULL;
   node->next_signal = NULL;
   node->prev_retry = NULL;
   node->next_retry = NULL;

   return node;
}


/***********************************************************************************/
int get_node_data(node, pointname, pchid, message, message_syntax, error, area, timeoccured, timetopage, timetoprint, timetosignal, value, desc) 

struct alarm   *node;
char           pointname[PVNAME_SZ + 1];
chid           *pchid;
char           message[MAX_MESSAGE + 1];
char           message_syntax[MAX_MESSAGE + 1];
char           error[MAX_ERROR];
int            *area;
time_t         *timeoccured;
time_t         *timetopage;
time_t         *timetoprint;
time_t         *timetosignal;
char           value[MAX_STRING_SIZE + 1];
char           desc[MAX_DESCRIPTION+1];

{

if (node != NULL) {

   strcpy(pointname, node->pointname); 
   *pchid = node->pchid;
   strcpy(message, node->message);
   strcpy(message_syntax, node->message_syntax);
   strcpy(error, node->error);
   *area = node->area;
   *timeoccured = node->time_of_occurance;
   *timetopage = node->time_to_page;
   *timetoprint = node->time_to_print;
   *timetosignal = node->time_to_signal;
   strcpy(value, node->value);
   strcpy(desc, node->desc);

   return TRUE;
}

   return FALSE;
}



/************************************************************************************/
struct alarm *update_status(node, value, desc, add_page, dial_out_timer, print_out_timer,                            signal_timer)

struct alarm *node;
char         value[MAX_STRING_SIZE + 1];
char         desc[MAX_DESCRIPTION + 1];
int          *add_page;
time_t       dial_out_timer;
time_t       print_out_timer;
time_t       signal_timer;

{
if (node != NULL) {

  *add_page = FALSE;

  strcpy(node->value, value);
  strcpy(node->desc, desc);
  if (!node->critical_alarm_set) {
     node->time_of_occurance= time(NULL);
     node->time_to_page = time(NULL) + dial_out_timer;
     node->time_to_print = time(NULL) + print_out_timer;
     node->time_to_signal = time(NULL) + signal_timer;
     node->critical_alarm_set = TRUE;
     *add_page = TRUE;
  }
}
  return node;
}
/************************************************************************************/
struct alarm *update_message(node, message)
struct alarm *node;
char         message[MAX_MESSAGE+1];
{

  if (node != NULL)
      strcpy(node->message, message);

  return node;
}

/************************************************************************************/
void reset_alarm(node)
struct alarm *node;
{

  if (node != NULL)
      node->critical_alarm_set = FALSE;      

}
/************************************************************************************/
struct alarm *update_error(node, error)
struct alarm *node;
char         error[MAX_ERROR+1];
{

   if (node != NULL)
      strcpy(node->error, error);

   return node;
}

/************************************************************************************/
struct alarm *get_page_data(time_to_page, message_syntax, description, value)

time_t    *time_to_page;
char      message_syntax[MAX_MESSAGE + 1];
char      description[MAX_DESCRIPTION + 1];
char      value[MAX_STRING_SIZE + 1];

{

    if (page_head != NULL) {

       strcpy(message_syntax, page_head->message_syntax);
       *time_to_page = page_head->time_to_page;
       strcpy(value, page_head->value);
       strcpy(description, page_head->desc);
    }
    return page_head;
}
/************************************************************************************/
struct alarm *get_signal_data(time_to_signal)

time_t    *time_to_signal;
{
    if (signal_head != NULL) {
       *time_to_signal = signal_head->time_to_signal;
    }
    return signal_head;
}
/************************************************************************************/

struct alarm *get_print_data(pointname, message, time_occured, time_to_print)

char          pointname[PVNAME_SZ + 1];
char          message[MAX_MESSAGE + 1];
time_t        *time_occured;
time_t        *time_to_print;

{
    if (print_head != NULL){

       strcpy(pointname, print_head->pointname);
       strcpy(message, print_head->message);
       *time_occured = print_head->time_of_occurance;
       *time_to_print = print_head->time_to_print;
    }
    return print_head;
}
/************************************************************************************/

struct alarm *get_retry_data(current, pchid)
struct alarm *current;
chid   *pchid;

{

   /* return the node pointed to by current */

   if (current == NULL) {

   /* start at the head of the list */

       if (retry_head != NULL)
           *pchid = retry_head->pchid;

       return retry_head;
   }
   else {

   /* return the next one in the list */

       if (current->next_retry != NULL)
          *pchid = current->next_retry->pchid;

       return current->next_retry;
   }
}

/************************************************************************************/
void add_to_page(node)
struct alarm *node;
{
    if (node != NULL){

	   if (page_head == NULL){ 
	       page_head = node;
	       page_tail = node;
           }
           else {
	       page_tail->next_page = node;
	       node->prev_page = page_tail;
	       page_tail = node;
          }
    }
};

/************************************************************************************/
struct alarm *remove_from_page(node)
struct alarm *node;
{

	       if (node == page_head) {
		   if (page_head == page_tail){
		       page_head = NULL;
		       page_tail = NULL;
                   }
		   else {
		       page_head = node->next_page;
		       page_head->prev_page = NULL;
		       node->next_page = NULL;
                   }
               }
	       else if (node == page_tail) {
		   page_tail = node->prev_page;
		   page_tail->next_page = NULL;
		   node->prev_page = NULL;
               }
	       else if ( (node != NULL) &&
			 (node->next_page != NULL) &&
			 (node->prev_page != NULL) ){
		  node->prev_page->next_page = node->next_page;
		  node->next_page->prev_page = node->prev_page;
		  node->next_page = NULL;
		  node->prev_page = NULL;
               }
     return node;
};
/***********************************************************************************/
void add_to_print(node)
struct alarm *node;
{
    if (node != NULL) {

	   if (print_head == NULL){ 
	       print_head = node;
	       print_tail = node;
           }
           else {
	       print_tail->next_print = node;
	       node->prev_print = print_tail;
	       print_tail = node;
           }
    }
};

/************************************************************************************/
void remove_from_print(node)
struct alarm *node;
{

	       if (node == print_head) {
		   if (print_head == print_tail){
		       print_head = NULL;
		       print_tail = NULL;
                   }
		   else {
		       print_head = node->next_print;
		       print_head->prev_print = NULL;
		       node->next_print = NULL;
                  }
               }
	       else if (node == print_tail) {
		   print_tail = node->prev_print;
		   print_tail->next_print = NULL;
		   node->prev_print = NULL;
               }
	       else if ( (node != NULL) &&
			 (node->next_print != NULL) &&
			 (node->prev_print != NULL) ){
		  node->prev_print->next_print = node->next_print;
		  node->next_print->prev_print = node->prev_print;
		  node->next_print = NULL;
		  node->prev_print = NULL;
               }
};
/***********************************************************************************/
void add_to_signal(node)
struct alarm *node;
{
    if (node != NULL) {

	   if (signal_head == NULL){ 
	       signal_head = node;
	       signal_tail = node;
           }
           else {
	       signal_tail->next_signal = node;
	       node->prev_signal = signal_tail;
	       signal_tail = node;
           }
    }
};

/************************************************************************************/
void remove_from_signal(node)
struct alarm *node;
{

	       if (node == signal_head) {
		   if (signal_head == signal_tail){
		       signal_head = NULL;
		       signal_tail = NULL;
                   }
		   else {
		       signal_head = node->next_signal;
		       signal_head->prev_signal = NULL;
		       node->next_signal = NULL;
                  }
               }
	       else if (node == signal_tail) {
		   signal_tail = node->prev_signal;
		   signal_tail->next_signal = NULL;
		   node->prev_signal = NULL;
               }
	       else if ( (node != NULL) &&
			 (node->prev_signal != NULL) &&
			 (node->next_signal != NULL) ){
		  node->prev_signal->next_signal = node->next_signal;
		  node->next_signal->prev_signal = node->prev_signal;
		  node->prev_signal = NULL;
		  node->next_signal = NULL;
               }
};
/***********************************************************************************/
void add_to_retry(node)
struct alarm *node;
{
    if (node != NULL) {

	   if (retry_head == NULL){ 
	       retry_head = node;
	       retry_tail = node;
           }
           else {
	       retry_tail->next_retry = node;
	       node->prev_retry = retry_tail;
	       retry_tail = node;
           }
   }
};

/************************************************************************************/
void remove_from_retry(node)
struct alarm *node;
{
	       if (node == retry_head) {
		   if (retry_head == retry_tail){
		       retry_head = NULL;
		       retry_tail = NULL;
                   }
		   else {
		       retry_head = node->next_retry;
		       retry_head->prev_retry = NULL;
                  }
               }
	       else if (node == retry_tail) {
		   retry_tail = node->prev_retry;
		   retry_tail->next_retry = NULL;
		   node->prev_retry = NULL;
               }
	       else {
		  node->prev_retry->next_retry = node->next_retry;
		  node->next_retry->prev_retry = node->prev_retry;
		  node->prev_retry = NULL;
               }
};
/*************************************************************************/
int    remove_point_from_signal(pointname)
char   pointname[PVNAME_SZ+1];
{
struct alarm *current;
int     found = FALSE;

    /*  Remove all alarms for pointname currently in the signal list. */

       for (current = signal_head; current != NULL && !found; current = current->next_signal){

           if ( strcmp(current->pointname, pointname) == 0) {
	       
	       if (current == signal_head) {
		   if (signal_head == signal_tail){
		       signal_head = NULL;
		       signal_tail = NULL;
                   }
		   else {
		       signal_head = current->next_signal;
		       signal_head->prev_signal = NULL;
		       current->next_signal = NULL;
                   }
               }
	       else if (current == signal_tail) {
		   signal_tail = current->prev_signal;
		   signal_tail->next_signal = NULL;
		   current->prev_signal = NULL;
               }
	       else {
		  current->prev_signal->next_signal = current->next_signal;
		  current->next_signal->prev_signal = current->prev_signal;
		  current->next_signal = NULL;
		  current->prev_signal = NULL;
               }
	       found =  TRUE;
           }
       }
}
/***********************************************************************************/
int    remove_point_from_page(pointname)
char   pointname[PVNAME_SZ + 1];
{
struct alarm *current;
int     found = FALSE;

    /*  Remove all alarms for pointname currently in the page list. */

       for (current = page_head; current != NULL && !found; current = current->next_page){

           if ( strcmp(current->pointname, pointname) == 0) {

	       if (current == page_head) {
		   if (page_head == page_tail){
		       page_head = NULL;
		       page_tail = NULL;
                   }
		   else {
		       page_head = current->next_page;
		       page_head->prev_page = NULL;
		       current->next_page = NULL;
                   }
               }
	       else if (current == page_tail) {
		   page_tail = current->prev_page;
		   page_tail->next_page = NULL;
		   current->prev_page = NULL;
               }
	       else {
		  current->prev_page->next_page = current->next_page;
		  current->next_page->prev_page = current->prev_page;
		  current->next_page = NULL;
		  current->prev_page = NULL;
               }
	       found =  TRUE;
           }
       }
}

/***********************************************************************************/
int    remove_point_from_print(pointname)
char   pointname[PVNAME_SZ + 1];
{
struct alarm *current;
int     found = FALSE;


    /*  Remove all alarms for pointname currently in the print list. */
    found = FALSE; 
    for (current = print_head; current != NULL && !found; current = current->next_print){

           if ( strcmp(current->pointname, pointname) == 0) {

	       if (current == print_head) {
		   if (print_head == print_tail){
		       print_head = NULL;
		       print_tail = NULL;
                   }
		   else {
		       print_head = current->next_print;
		       print_head->prev_print = NULL;
		       current->next_print = NULL;
                   }
               }
	       else if (current == print_tail) {
		   print_tail = current->prev_print;
		   print_tail->next_print = NULL;
		   current->prev_print = NULL;
               }
	       else {
		  current->prev_print->next_print = current->next_print;
		  current->next_print->prev_print = current->prev_print;
		  current->next_print = NULL;
		  current->prev_print = NULL;
               }
	       found = TRUE;
           }
       }
}

/***********************************************************************************/
