/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*******************************************************************************
*    Digital Pager Interface - Real Time Component
*
*    KAG    3/94
*******************************************************************************/
#include   <errno.h>
#include  <stdio.h>
#include  <db_access.h>
#include  <dbDefs.h>
#include  <time.h>
#include  <alarmlist.h>
#include  <pagerdefs.h>
#include  <cadef.h>
#include  <alarm.h>
#include  <string.h>
#include  <sys/types.h>
#include  <sys/time.h>
#include  <syslog.h>

#define   STDIN       0
#define   STDOUT      1

C4CODE    cb;
time_t    dial_out_timer, print_out_timer, signal_timer;
int       page_available = TRUE;
int       signal_on = FALSE;
int       print_available = FALSE;

/********************************************************************************************/
void message_print(print_dev, message)
char *print_dev;
char *message;
{
FILE      *stream;
char      printstring[PRINT_MESSAGE + 50];

   /* write to the line printer */
if(0) {
   if (strstr( print_dev, "/dev/")) {
       if (stream = fopen(print_dev, "w")) { 
          fprintf(stream, "%s\n", message);
          fflush(stream);
          fclose(stream);
       }
   }
}

if (print_available) {

   if (print_dev[0] == 'P' || print_dev[0] == 'p') {
       sprintf(printstring, "echo \"%s\" | lpr -P%s 2>/dev/console\n", message, &print_dev[1]);
       system( printstring);

   }
   else {
   /* append to a file */

       if (stream = fopen(&print_dev[1], "a")) {
          fprintf(stream, "%s", message);
          fclose(stream);
       }
   }
}

}
/******************************************************************************************/

void build_message(message, descriptor, value)


char   message[MAX_MESSAGE+1];
char   descriptor[MAX_DESCRIPTION+1];
char   value[MAX_STRING_SIZE+1];

     /* build the message string to be sent to the pager.  Expand where indicated. */
{
char    temp[MAX_MESSAGE+1], message_piece[2];
char    expression;
int     i, x, length;

     temp[0] = '\0';

     if (strcmp(message,"                                                                                ") == 0) {

	/* The message string is blank, fill the message with the description and value */

	length = strlen(descriptor);
        if (length > MAX_MESSAGE) 
            length = MAX_MESSAGE;
        strncat(temp, descriptor, length);

	length = MAX_MESSAGE - length;

        if (length > strlen(value))
            length = strlen(value);
        strncat(temp, value, length);
     }
     else {

	/* Parse the message string and expand where indicated */

        for (i=0 , x = 0; x < MAX_MESSAGE && i <= MAX_MESSAGE  ; ++i, x = x+length) {
     
	   if ( ((i+1) <= MAX_MESSAGE) && (message[i] == '~') && 
		( ((expression=message[i+1]) == '1') || (expression == '2')) ){
		    
           /* if the string contains the escape character "~" followed by a 1 or 2 the message
	      needs expanding. */

	       if (i+2 <= MAX_MESSAGE && isdigit(message[i+2])){

		   /* If the message string contains an integer value immediately following the expression.
		      This value describes the number of characters that will be concatenated to the
		      temporary message string. */
	     
	           if (i+3 <= MAX_MESSAGE && isdigit(message[i+3])) {
		      message_piece[0] = message[i+2];
		      message_piece[1] = '\0';
		      length = atoi(message_piece);
		      length = length * 10;
		      message_piece[0] = message[i+3];
	              length = length + atoi(message_piece);
		      i = i+3;
                   }
	           else {
		      message_piece[0] = message[i+2];
		      message_piece[1] = '\0';
	              length = (atoi(message_piece));
                      i = i+2;
		   }
               }
               else {

		  /* otherwise the entire description/value will be concatenated to the message string */

	           length = MAX_MESSAGE - x;
		   i = i+1;
               }

	       if (length + x > MAX_MESSAGE)
	           length = MAX_MESSAGE - x;


	       if (expression == '1') {
                   if (length > strlen(descriptor) )
		       length = strlen(descriptor);
		   strncat(temp, descriptor, length);
               }
	       else if (expression == '2') {
		   if (length > strlen(value) )
		       length = strlen(value);
		   strncat(temp, value, length);
               }
          }
	  else {
	      temp[x] = message[i];
	      length = 1;
	      temp[x+length] = '\0';
          } 
        }
     }      
     strcpy(message, temp);
}

/************************************************************************************/

void  page_for_acknowledgement()
{
int           more_to_page=TRUE;

struct alarm  *alarm_node;
time_t        time_to_page;
char          message[MAX_MESSAGE + 1], description[MAX_DESCRIPTION+1], value[MAX_STRING_SIZE+1];

     while (more_to_page) {

        alarm_node = get_page_data(&time_to_page, message, description, value);

	if (alarm_node != NULL && time_to_page <= time(NULL) ) {

            /* The alarm has remained on the list for longer than the dial out time, get
	       message data */
       
	    build_message(message, description, value);
	    update_message(alarm_node, message); 

            /* dial the phone number using the asynchronous port connected to the pager interface,
               and send out the alarm message to the pager. */

if (1) {
            if (page_available) 
                 write(STDOUT, alarm_node, sizeof(struct alarm));
}

	    alarm_node = remove_from_page(alarm_node);
	    add_to_print(alarm_node);
       }
       else
	   more_to_page = FALSE;
    } 
}
/******************************************************************************************/

void   print_alarm_message(print_dev)
char   *print_dev;

{
char         pointname[PVNAME_SZ + 1], message[MAX_MESSAGE + 1], print_message[PRINT_MESSAGE+1], current[50], occured[50];
int          more_to_print = TRUE;
struct alarm *alarm_node;
time_t       time_occured, time_to_print;
time_t       epoch;

    /* The security notification timer has expired, print the message on the security 
       printer and remove message from list. */

    while (more_to_print) {
	  
       alarm_node = get_print_data(pointname, message, &time_occured, &time_to_print); 

       if (alarm_node != NULL && time_to_print <= time(&epoch) ){

          sprintf(current, "%s", ctime(&epoch));
          sprintf(occured,"%s", ctime(&time_occured)); 

          sprintf (print_message, "\n%s     Critical alarm occured for point %s, on %s     Alarm not acknowledged within security notification timer.\n     Page message : %s\n     This alarm cleared from pager system.",current, pointname, occured, message);

	  
          message_print(print_dev, print_message);

          remove_from_print(alarm_node);
       }
       else
	   more_to_print = FALSE;
   }
}

/***************************************************************************************/
void   signal(pchid)
chid   pchid;
{
struct alarm *alarm_node;
time_t       time_to_signal;
time_t       epoch;
short        on, off;

    /* check for expired signal timer on unacknowledged alarms */

       alarm_node = get_signal_data(&time_to_signal); 

       if (alarm_node != NULL && time_to_signal <= time(&epoch)){

	   /* turn signal on the time has expired */
	   if (!signal_on) {
	      signal_on = TRUE;
	      on = 1;
	      ca_put(DBR_SHORT, pchid, &on);
	      ca_flush_io(1.0);
           }
       }
       else 
	   if (signal_on){
	   /* turn signal off the timer has not expired for any unacknowledged alarms*/
           signal_on = FALSE;
	   off = 0;
	   ca_put(DBR_SHORT, pchid, &off);
	   ca_flush_io(1.0);
       }
}

/*****************************************************************************************/
void   update_alarm_node(alarmargs)
struct event_handler_args   alarmargs;

{
char          pointname[PVNAME_SZ+1], value[MAX_STRING_SIZE+1], desc[MAX_DESCRIPTION+1];
int           add_to_list;
struct alarm  *alarm_node;
char          ack[5];

    if (strcmp (((struct dbr_sts_string *)alarmargs.dbr)->ack, "NO") == 0) {
    
       /* The event has been acknowledged, "NO" indicates that it is not waiting to
          be acknowledged, remove all events for the point from the list
	  and reset the list */
   
       strcpy(pointname, ((struct alarm *)alarmargs.usr)->pointname);

       remove_point_from_page(pointname);
       remove_point_from_print(pointname);
       remove_point_from_signal(pointname);

       reset_alarm( (struct alarm *)alarmargs.usr );

    }
    else if (((struct dbr_sts_string *)alarmargs.dbr)->severity == MAJOR_ALARM) { 

       /* The alarm message is of critical (major) severity add it to the list */
       
       alarm_node = update_status((struct alarm *)alarmargs.usr, 
				  ((struct dbr_sts_string *)alarmargs.dbr)->value, 
				  ((struct dbr_sts_string *)alarmargs.dbr)->desc, 
				  &add_to_list, dial_out_timer, print_out_timer, 
				  signal_timer);

       if (add_to_list) {
	   add_to_page(alarm_node);
	   add_to_signal(alarm_node);
       }

   }
}
/**************************************************************************************/
int retry_connect()

{
chid          pchid;
struct alarm  *alarm_node = NULL;
int           more_to_retry = FALSE;
evid          eventid;

       while ( (alarm_node = get_retry_data(alarm_node, &pchid)) != NULL) {

	     if (ca_state(pchid) == currOpen) {

	         ca_add_masked_array_event(DBR_STS_STRING, 1, pchid, update_alarm_node, 
                      alarm_node, (float)0, (float)0, (float)0, &eventid, DBE_ALARM);

                 remove_from_retry(alarm_node);
             } 
	     else
	       
		more_to_retry = TRUE;
      }
      if (!more_to_retry)
	  return FALSE;
      else
	  return TRUE;
}
/**************************************************************************************/
void validate_input(param, ptype, printer_dev)
char *param;
char *ptype;
char *printer_dev;
{
char  error_message[MAX_ERROR+1];


     for (; *param != '\0'; ++param)
	 if (!isdigit(*param)) {
	    sprintf(error_message, "\nERROR, INVALID %s.", ptype);
	    message_print(printer_dev, error_message);
	    exit(0);
         }
}
/*****************************************************************************************/
     main(argc, argv)

     int    argc;
     char   *argv[];
     
     /* The command line to start the process will contain three parameters.  The dial out timer,
	the security notification timer, and the name of the security printer device */
     {
   char          pointname[PVNAME_SZ+1], printer_dev[DEV_NAME +1], tempname[PVNAME_SZ+1], 
		 *param[4], error_message[MAX_ERROR+1];
   chid          pchid, S_pchid;
   evid          eventid;
   int           to_parent[2], to_child[2], pid, done, i;
   int           signal_available = FALSE;
   float         check_status;
   D4DATA        *pointgrp;
   T4TAG         *tag;
   F4FIELD       *ptname_fld, *area_fld, *message_fld;
   struct alarm  *alarm_node, *alarmnode=(struct alarm *)malloc(sizeof(struct alarm));
   struct timeval    timeout;
   fd_set            chanmask;
   time_t        epoch;
   int           temp;
   int           reconnect = FALSE, success;
   char          fileToOpen[1000];
   short         off;
   char          *theerror = NULL;


   if (argc >= 4) {
      strncpy(printer_dev, argv[4], DEV_NAME);
      if (printer_dev[0] == 'F' || printer_dev[0] == 'f' ||
	  printer_dev[0] == 'P' || printer_dev[0] == 'p' )
	  print_available = TRUE;
      else
	 system("echo \"WARNING, INVALID PRINT DEVICE, PAGER WILL NOT IMPLEMENT PRINTING\" >/dev/console");
      if (argc != 9) {
         message_print(printer_dev,"\nERROR, PAGER PROCESS EXITING INVALID COMMAND LINE."); 
         exit(0);
      }
   }
   else
      exit(0);

   timeout.tv_sec = 1;
   timeout.tv_usec = 0;

     dial_out_timer = atoi(argv[1]);
     if (dial_out_timer <= 0)
	 validate_input(argv[1], "DIAL OUT TIMER, MUST BE A POSITIVE INTEGER", printer_dev);
     print_out_timer = atoi(argv[2]);
     if (print_out_timer <= 0)
	 validate_input(argv[2], "PRINT OUT TIMER, MUST BE A POSITIVE INTEGER", printer_dev);
     check_status = atoi(argv[3])+0.0;
     if (check_status < 2) 
	 validate_input("invalid", "CHECK STATUS, MUST BE GREATER THAN OR EQUAL TO 2", printer_dev);
     signal_timer = atoi(argv[8]);
     if (signal_timer <= 0)
	 validate_input(argv[8], "SIGNAL TIMER, MUST BE A POSITIVE INTEGER", printer_dev);
     param[0] = (char *)malloc(sizeof(char)*20);
     param[1] = (char *)malloc(sizeof(char)*DEV_NAME + 1);
     param[2] = (char *)malloc(sizeof(char)*MAX_PHONE + 1);
     param[3] = 0;
     sprintf(param[0],"%s", "pagerinterface");
     strncpy(param[1], argv[5], DEV_NAME); /*modem dev */
     strncpy(param[2], argv[6], DEV_NAME); /* modem phone */
     validate_input(param[2], "PAGER SERVICE NUMBER", printer_dev);
     /* Spawn the child process responsible for actual pager interface */

if (1) {
       pipe(to_parent); 
         pipe(to_child);

     if (fork() == 0) { 
	/* child process */

        close(0);        /* close standard in */
        dup(to_child[0]);    /* dup pipe read to standard in */
        close(1);      /* close old standard output */
        dup(to_parent[1]);    /* dup pipe write to standard out */

        close(to_parent[1]);  /* clean up */
        close(to_child[0]);
        close(to_parent[0]); 
        close(to_child[1]);
        execvp("/tis4000/tasks/pagerinterface", param);    /* start the child process */
	perror(theerror);
        if (errno) exit();
     } 

     close(1);        /* close standard out */
     dup(to_child[1]);      /* dup pipe write to standard out */
     close(0);        /* close standard in */
     dup(to_parent[0]);     /* dup pipe read to standard in */

     close(to_child[1]);    /* clean up on this end */
     close(to_parent[0]);
     close(to_child[0]);
     close(to_parent[1]); 
}
     /* clean up parameters passed to child process */

     for (i=0; i<=2; ++i)
	 free(param[i]);

     /* initialize the pointers used in alarmlist */

     init_alarmlist();

     /* initialize codebase  */
     d4init(&cb);

     cb.go_error = 0;
     cb.open_error = 0;

     /* Get pointnames, areas, and message string data needed for channel access */
  
     translate_db_link("ptgroup.dbf", fileToOpen);

     if (! (pointgrp = d4open(&cb, fileToOpen)) ) {
	   sprintf(error_message, "\n********************************************************************************\n%sERROR! THE POINTGRP TABLE DOES NOT EXIST PAGING PROCESS EXITING\n********************************************************************************",ctime(&epoch));

		       message_print(printer_dev, error_message);
           
            exit(0);
     }
     e4exit_test(&cb);
     d4pack(pointgrp);
     tag = d4tag(pointgrp, "POINTNAME");
     d4tag_select(pointgrp, tag);
     d4top(pointgrp);
     ptname_fld = d4field(pointgrp, "POINTNAME");
     area_fld = d4field(pointgrp, "AREANUM");
     message_fld = d4field(pointgrp, "MESSAGE");

     /* Setup channel access for all points specified in point grouping table. */
     ca_task_initialize();

     /* connect with signal */

     if ( (ca_search(argv[7], &S_pchid) == ECA_NORMAL) &&
	  (ca_pend_io(3.0) == ECA_NORMAL) ) {

		/* assure that connected */
		if (ca_state(S_pchid) == currOpen) {
		    signal_available = TRUE;
		    off = 0;
		    ca_put(DBR_SHORT, S_pchid, &off);
	            ca_flush_io(1.0);
                } 
      }

     for (; !d4eof(pointgrp);) {

       if (strcmp( f4str(d4field(pointgrp,"TAG")), " ") != 0) { 
	 f4ncpy(ptname_fld, pointname, sizeof(pointname)); 

	 /* get rid of trailing blanks returned from database */
	 for (i = 0; i < PVNAME_SZ && pointname[i] != ' '; ++i)
	      tempname[i] = pointname[i];
         tempname[i] = '\0';
	 sprintf(pointname, "%s", tempname);

	 success = FALSE;

         if (ca_search(pointname, &pchid) == ECA_NORMAL) { 

	     alarm_node = construct(pointname, pchid, " ", f4str(message_fld), " ", 
	                   f4int(area_fld), 0, 0, 0, 0, " ", " ",FALSE);

	    if (ca_pend_io(1.0) == ECA_NORMAL) {

	      if (ca_state(pchid) == currOpen) {

	          ca_add_masked_array_event(DBR_STS_STRING, 1, pchid, update_alarm_node, 
         		alarm_node, (float)0, (float)0, (float)0, &eventid, DBE_ALARM);

                  success = TRUE;

              } 
           }
	   if (!success) {
	       /* could not connect to point add to retry */
              add_to_retry(alarm_node);
	      reconnect = TRUE;
           }
         }
       }
       d4skip(pointgrp, 1);
        
     }

     d4close(pointgrp);

        for (;;)  {     /* forever loop */

  	   ca_pend_event(check_status);

	   page_for_acknowledgement();

         if (page_available) {
	   
	      /* determine if any alarm messages for points could not be sent to pager.  If
	         this occured remove from print list message and send to printer*/

            done = FALSE;
	      while (!done) {

                FD_ZERO(&chanmask);

                FD_SET(STDIN, &chanmask);

                if (select(STDIN+1, &chanmask, NULL, NULL, &timeout) ) {

                 if (FD_ISSET(STDIN, &chanmask)){

	            if ((temp = read(STDIN, alarmnode, sizeof(struct alarm))) == 0) {

		       /* the child process died */

		       syslog(LOG_ALERT | LOG_DAEMON, "PAGERINTERFACE PROCESS HAS EXITED");

		       time(&epoch);
		       sprintf(error_message, "\n********************************************************************************\n%sERROR! THE PAGER INTERFACE PROCESS HAS TERMINATED NO ALARMS CAN BE SENT TO PAGER\n********************************************************************************",ctime(&epoch));

		       message_print(printer_dev, error_message);

		       done = TRUE;

		       page_available = FALSE;

		       print_out_timer = dial_out_timer;

                    }
                    else {

		       /* the message could not be sent to the pager, send
			  it to the printer immediatly and remove from print
			  list */
			  
                     
		     message_print(printer_dev, alarmnode->error);

		    remove_point_from_print(alarmnode->pointname); 

                    }
                 }
                }
	        else
		  done = TRUE;
              }
           }

	   print_alarm_message(printer_dev);

	   if (signal_available)
	       signal(S_pchid);
           else
	       if (ca_state(S_pchid) == currOpen){
	           signal_available = TRUE;
	           off = 0;
	           ca_put(DBR_SHORT, S_pchid, &off);
	           ca_flush_io(1.0);
               }
	  

	   if (reconnect) 
	       reconnect = retry_connect();
        } 
}
