/*pidtuning.c,v 1.1.1.1 1999/08/03 14:44:11 tis4000 Exp */

/*pidtuning.c,v 1.1.1.1 1999/08/03 14:44:11 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/***********************************************************************************************
     loop_tuning.c

     This file contains the functionality identified from the software functional specification for
     PID Loop Tuning Package Interface for the TIS 4000, Revision A.1.

     Inputs:	pointname for the pidblock 
		the number of samples to be collected
		the interval between the samples

     Outputs:   An ascii file containing a 5 line header, and three columns of plant data, where each
		row contains a set of data.  The first column contains the time of the plant data set, 
		the second column contains a number representing the Process Variable, and the third 
		column contains a number representing the Controller Output.

     K.Grimes   March 1994

***************************************************************************************************/
#include  <stdio.h>
#include  <db_access.h>
#include  <dbDefs.h>
#include  <time.h>
#include  <d4all.h>
#include  <cadef.h>
#include  <alarm.h>
#include  <string.h>
#include  <scan_types.h>
#include  "time_utility.h"

#define   TRUE              1
#define   FALSE             0

#define   MAX_MESSAGE     100 
#define   MIN_SAMPLE       33
#define   MAX_SAMPLE    10000 
#define   MIN_INTERVAL     .1
#define   MAX_INTERVAL     10
#define   MAX_FILENAME     11
#define   MAX_FILES        20

struct  filedata {
	     char    timestamp[TIMESTRING_SIZE + 1];
             float   process_variable;
	     float   controller_output;
} *point;

int i; 
int done=FALSE;
char  filename[] = "nextnum.dat";
char  tempfile[]= "temp.dat";


/************************************************************************************
    function      : connection_handler_function 
*************************************************************************************/
void connection_handler_function(handlerargs)

struct connection_handler_args handlerargs;
{

    if (handlerargs.op == CA_OP_CONN_DOWN)

	done = TRUE;
}


/************************************************************************************
    function      :  get_block_data
*************************************************************************************/
void  get_block_data(alarmargs)

struct event_handler_args    alarmargs;
{
float    *temp;

   if (strcmp( (char *)alarmargs.usr, ".MVAL") == 0) {                       

       temp = (float *)alarmargs.dbr;

       point[i].process_variable = *temp;
   }
   else if (strcmp( alarmargs.usr, ".VAL") == 0) {

      temp = (float *)alarmargs.dbr;

      point[i].controller_output = *temp;
  }
}

/****************************************************************************
   function   :  get_filename

   description:  A total of MAX_FILES can be associated with this software at any one time.  
		 To uniquely identify these files a number is concatinated onto the filename.  The 
		 current number is found in filename.  When MAX_FILES is reached file "1" will
		 be written, overwritting any currently existing file "1".
*****************************************************************************/
		 
void  get_filename(pidfilename)

char pidfilename[MAX_FILENAME + 1];
{
FILE *stream;
int   current, next, value;
int   done = FALSE;

    if ( (stream = fopen(filename, "r")) == NULL) {
         stream = fopen(filename, "w");
	 current = 1;
	 next = current+1;
	 fprintf(stream, "%d", next);
	 fclose(stream);
    }
    else {
         if ( (value = fscanf(stream,"%d", &current) ) != EOF) {
           if ( current == MAX_FILES)
              next = 1;
           else
	      next = current+1;
           fclose(stream);
         }
         if ( (stream = fopen(filename, "w")) != NULL) {
           fprintf(stream, "%d", next);
           fclose(stream);
         }
    }
    sprintf(pidfilename,"pvco%d.asc", current);
}

/************************************************************************************
    function      :  convert_file 

    description   :  converts a unix text file to a dos text file.
*************************************************************************************/
void  convert_file()

{
char  pidfilename[MAX_FILENAME + 1];
char  command_line[40];

      get_filename(pidfilename);

      sprintf(command_line, "unix2dos %s %s",tempfile, pidfilename);

      system(command_line);
}

/************************************************************************************
    function      :  write_header
*************************************************************************************/
FILE *write_header(pointname, num_samples, interval, starttime, message)

char    pointname[PVNAME_SZ + 1];
int     num_samples;
float   interval;
int     starttime;
char    *message;

{
FILE *stream;
char pidfilename[MAX_FILENAME + 1];

    stream = fopen(tempfile, "w");
   
    /* Write the file header */

    fprintf(stream, "%s", asctime(localtime(&starttime)));
    fprintf(stream, "pointname : %s          sample count : %d      sample interval : %7.2f\n",pointname,
		     num_samples, interval);
    fprintf(stream, "%s\n",message);
    fprintf(stream, "TIME          PV         CO\n");       

    return stream;
}

/************************************************************************************
    function      :  write_data
*************************************************************************************/
void write_data(i, interval, stream)

int     i; /*num samples */
float   interval;
FILE    *stream;

{
int     x, actual=0, total, required;
char    current[TIMESTRING_SIZE+1];

    if ( interval >= 1 )
       for (x=0; x < i; ++x) 
           fprintf(stream, "%s      %6.3f     %6.3f\n",point[x].timestamp, 
		   point[x].process_variable, point[x].controller_output);

    else {

       /* fill in any missed time with the last set of data collected for the second. */

       if (interval > .45)
	  required = 2;
       else if (interval > .15)
	  required = 5;
       else /* interval >= .1 */
	  required = 10;

       for (x=0, actual=0; actual < i;) {

            sprintf(current, "%s", point[x].timestamp);
	    total = 0;
	    
	    /* while the number of samples written to the file is less then the total
	    number of samples requested and the dates are equal, write the data to
	    the file  */

	    while (actual < i && strcmp(current, point[x].timestamp) == 0 ) {
                fprintf(stream, "%s      %6.3f     %6.3f\n",point[x].timestamp, 
	                point[x].process_variable, point[x].controller_output);
		++x;
		++total;
		++actual;
            }

            /* if all samples for the current second were not collected fill the previous 
	    data in for the current */

	    while( total < required && actual < i) {
               fprintf(stream, "%s      %6.3f     %6.3f\n",point[x-1].timestamp, 
	               point[x-1].process_variable, point[x-1].controller_output);
	       ++total;
	       ++actual;
            }
       }
    }

		   
	   
    fclose(stream);
}

/***********************************************************************************
    function      : missed_seconds

    description   : if the interval is less than 1 determines if any seconds were 
		    missed while getting data.  If time was missed the data sets are 
		    invalid.
************************************************************************************/
int missed_seconds(interval, samples)
float interval;
int   samples;
{
int   i, second_1, second_2, hour, minute;
int   missed = FALSE;
char  *temp = NULL, timestring[TIMESTRING_SIZE];

   if (interval < 1.0) {

   sprintf(timestring,"%s", point[0].timestamp);

   parse_time_string(timestring, &hour, &minute, &second_1);

       for (i = 1; i < samples && !missed; ++i) {
	  
	  sprintf(timestring, "%s", point[i].timestamp);

          parse_time_string(timestring, &hour, &minute, &second_2);

	  if  (second_1 != second_2) {
	  
	      if (second_1 == 59) {
		 if (second_2 != 0)
		     missed = TRUE;
              }
	      else if (second_2 != (second_1 + 1))
		 missed = TRUE;
          }
	  second_1 = second_2;
       }
   }
   return missed;
}

/************************************************************************************
    function      :  main
*************************************************************************************/
main(argc, argv)
int   argc;
char *argv[];

{
int        num_samples, success=TRUE, value, starttime=0, theinterval, second, 
	   atempts=0, retries=0, total=0, samples=0;
char       message[MAX_MESSAGE+1];
char       pointname[PVNAME_SZ+1], field[PVNAME_SZ + 6];
chid       val_pchid, mval_pchid, ct_pchid, scan_pchid;
evid       val_evid, mval_evid, ct_evid;
FILE       *stream;
double      interval;
struct tm  *currenttime;
time_t     epoch;


   ca_task_initialize();

   strcpy(message,"\n");
   
   time(&starttime);
  
   if (argc >= 2)
      strcpy(pointname, argv[1]);
   else
      strcpy(pointname, " ");

   if (argc >= 3) {
      num_samples = atoi(argv[2]);
      samples = num_samples;
   }
   else 
      num_samples = 0;

   if (argc >= 4) { 
      interval = atof(argv[3]);
   }
   else {
      
      /* the interval is set to the scan time defined in the PID block */

     sprintf(field, "%s%s", pointname, ".SCAN");
     if (ca_search(field, &scan_pchid) == ECA_NORMAL) {

       if (ca_pend_io(1.0) == ECA_NORMAL) {

         value = ca_get(DBR_INT, scan_pchid, &theinterval); 

         if ( ca_pend_io(1.0) == ECA_NORMAL) {

            switch (theinterval) {

               case P10SECOND : 
		       interval = 10;
		       break;

               case P5SECOND  : 
		       interval =  5;
		       break;

               case P2SECOND  : 
		       interval =  2;
		       break;

               case PSECOND   : 
		       interval =  1;
		       break;

               default        : 
		       interval =  0;
		       break;
             }
         }
	 else 
	    interval = 0;
      }
      else 
	 interval = 0;
   }  
   else
      interval = 0;
}

   if (argc >= 3) {

      if (num_samples < MIN_SAMPLE || num_samples > MAX_SAMPLE) {

	strcpy(message, "ERROR : CAN NOT EXECUTE PID LOOP DATA CAPTURE UTILITY INVALID SAMPLE COUNT\n");
	stream = write_header(pointname, num_samples, interval, starttime, message);
	write_data(0, interval, stream);
	convert_file();
	success = FALSE;
      } 

      else 
	if (interval < MIN_INTERVAL || interval > MAX_INTERVAL) {
           
	   strcpy(message, "ERROR : CAN NOT EXECUTE PID LOOP DATA CAPTURE UTILITY INVALID INTERVAL\n");
	   stream = write_header(pointname, num_samples, interval, starttime, message);
	   write_data(0, interval, stream);
	   convert_file();
	   success = FALSE;
        }
      else {
	   if (interval == 0) {
		   strcpy(message, "ERROR : CAN NOT EXECUTE PID LOOP DATA CAPTURE UTILITY INVALID INTERVAL\n"); 
		   stream = write_header(pointname, num_samples, interval, starttime, message);
		   write_data(0, interval, stream);
		   convert_file();
		   success = FALSE;
           } 
	
	   else if (interval < 1 ) {
	        
	       if (interval > .05 && interval <= .15) 
		   interval = .1;
               else if (interval >.15 && interval <= .25)
		   interval = .2;
               else if (interval > .45 && interval <= .55)
		   interval = .5;
               else {
		   strcpy(message, "ERROR : CAN NOT EXECUTE PID LOOP DATA CAPTURE UTILITY INVALID INTERVAL\n"); 
		   stream = write_header(pointname, num_samples, interval, starttime, message);
		   write_data(0, interval, stream);
		   convert_file();
		   success = FALSE;
               }
           }	
	   else {
               /* interval is between 1 and 10, get rid of any fractional input */
	       i = interval;
	       interval = i;
	   }
      }
  }
  else {

     success = FALSE;
     sprintf(message, "ERROR : CAN NOT EXECUTE PID LOOP DATA CAPTURE UTILITY INVALID COMMAND LINE.\n");
     stream = write_header(pointname, num_samples, interval, starttime, message);
     write_data(0, interval ,stream);
     convert_file();
  }

    if (success) {

       /* set up for channel access */

       sprintf(field, "%s.VAL", pointname);
       
       if (ca_build_and_connect(field, TYPENOTCONN, 1, &val_pchid, 0, connection_handler_function, 0) == ECA_NORMAL) {

          sprintf(field, "%s.MVAL", pointname);
          
	  if (ca_build_and_connect(field, TYPENOTCONN, 1, &mval_pchid, 0, connection_handler_function, 0) == ECA_NORMAL) {

            if ( (ca_pend_io(1.0) == ECA_NORMAL)  &&
	       ( (ca_field_type(val_pchid) != TYPENOTCONN) && (ca_state(val_pchid) == cs_conn)) &&
	       ( (ca_field_type(mval_pchid) != TYPENOTCONN) && (ca_state(mval_pchid) == cs_conn)) ) {

                /* declare an array big enough to hold all the collected data. */

                point = (struct filedata *)calloc(num_samples, sizeof(struct filedata));
             
		/* start collecting data on the first full second */

                currenttime = (struct tm *)malloc(sizeof(struct tm));
		time(&epoch);
		currenttime = localtime(&epoch);
		second = currenttime->tm_sec;
     

		while (second == currenttime->tm_sec) {
		      time(&epoch);
		      currenttime = localtime(&epoch);
                }

                free(currenttime);

		/* set the number of retries allowed for a timeout return on pend envent */
                if (interval >= 1)
		    retries = 2;
                else if (interval > .45) /*.5*/
	            retries = 3;
                else if (interval > .15) /* .2 */
	            retries = 5;
                else /* interval >= .1 */
	            retries = 10;

		/* Collect a plant data set for the number of samples requested. */

                for ( i = 0; i < num_samples && !done; ++i) {

                   if( (ca_array_get_callback(DBR_FLOAT, 1, val_pchid, get_block_data, ".VAL")) &&
                       (ca_array_get_callback(DBR_FLOAT, 1, mval_pchid, get_block_data, ".MVAL")) ){

		     get_time_string(point[i].timestamp);

		     if (ca_pend_event(interval) != ECA_NORMAL) {

			++total;

                        ++atempts;

			if (atempts > retries) {
			   /* lost connection with point */

		           sprintf(message,"WARNING : UNABLE TO GATHER ALL REQUESTED SAMPLES FOR %s.\n", pointname);
			   done = TRUE;

                           i = i - 1;    /* adjust i to reflect correct sample count */
			   num_samples = i;
                        }
			else if (interval < 1) {

			       /* If an interval is missed fill in the dataset with the previously collected value */
			       point[i].process_variable = point[i-1].process_variable;
			       point[i].controller_output = point[i-1].controller_output;
			}
			else
			   /* the timestamp can be used only write the data actually collected. */
			       --i;
                    }
		    else
		        atempts = 0;
                } 
		}
		if ( (samples == num_samples) && (((total * 100) / num_samples) >= 5) )
		   strcpy(message,"WARNING: DATA IS SUSPECT, MORE THAN 5% OF CONTIGUOUS SAMPLES WERE NOT AVAILABLE.\n");

		if (missed_seconds(interval, num_samples) ) {
		    strcpy(message, "ERROR : DATA IS SUSPECT, UNABLE TO COLLECT REQUESTED SAMPLES.\n");
		    stream = write_header(pointname, num_samples, interval, starttime, message);
		    write_data(0, interval, stream);
		}
		else {

		   /* write the ascii file for the output */
		   stream = write_header(pointname, num_samples, interval, starttime, message);
	            write_data(i, interval, stream);
                }

		convert_file();

  	        free(point);

		ca_task_exit();
           }
	   else {
              success = FALSE;
              strcpy(message, "ERROR : CAN NOT EXECUTE PID LOOP DATA CAPTURE UTILITY UNABLE TO CONNECT TO POINT.\n");
	      stream = write_header(pointname, num_samples, interval, starttime, message);
	      write_data(0, interval, stream);

	      convert_file();
           }
         }
       }
   }
}
