/*watchdog.c,v 1.1.1.1 1999/08/03 14:44:12 tis4000 Exp */
/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/****************************************************************/
/*								*/
/*	    PLATFORM INDEPENDENT WATCHDOG ROUTINES              */
/*								*/
/****************************************************************/

/*  File: watchdog.c                                                                    
**  Edit:  01/03/95     WEP: created                                                    
**                         :
*/
/* task definitions */

#include <wdLib.h>
#include <taskLib.h>
#include <vxWorks.h>
#include <logLib.h>
#include <tickLib.h>
#include <string.h>
#include <stdio.h>

#include "platform.h"

#define   MAXWATCHDOG           100    /* maximum number of tasks monitored by the watch dog time */

#define   HARDWAREOFF    1
#define   HARDWAREON     2
#define   HARDWAREONSET  3
#define   HARDWAREOFFSET 4

static short global_watchdog_enable,   /* TRUE: test tasklist for timeout. FALSE: don't test tasklist */
             global_watchdog_abort,    /* TRUE: normal opperation. FALSE: cause wdt task to return.   */
             global_watchdogHW_control; /* 0: no action. 1: enable use of wdt hw. 2: disable wdt hw.   */

static FUNCPTR  tisEnableWatchDogHW;
static FUNCPTR  tisPetWatchDogHW;
static FUNCPTR  tisDisableWatchDogHW;
static int      tisRateMask;

static int      watchdog_delay;
static WDOG_ID  watchdog_id;
static STATUS   watchdog_status;
static int      watchdog_hemlock;
static int      watchdog_tick;
static int      watchdog_petcount;
static int      watchdog_tableid;

static int      watchdog_installed;

typedef struct {
        short  valid;
        short  used;
        short  enabled;
        short  interval;
        short  max;
        long   lastpulsetick;
        long   taskId;
        long   average32;
               } TASKWDC;

/* static variables are zeroed by C compiler */
static TASKWDC  tiswatchdogtable[MAXWATCHDOG];

void watchdog_timeout(int dummy);

int  tisWatchDogAttach(int task_id, short interval, short *wdId);
int  tisWatchDogPulse(short wdId);
int  tisWatchDogDetach(int task_id);

int  tisWatchDogDisable(int task_id);
int  tisWatchDogEnable(int task_id);

int  tisWatchDogAbort(void);
void tisWatchDogList(void);

int  wdt100(void);
int  wdt10(void);
int  wdt(int n);
int  wdtdummy(int priority, short interval, short delay);
void dummytask(short interval, short delay);
int  wdtabort(void);
int  wdtteston(void);
int  wdttestoff(void);
void clearallwdt(void);
void clearwdt(int index);
void dumpwdt(void);
static void printwatchdoghdr (unsigned char *buffer, unsigned short cnt);

/*****************************************
**       watchdog task                  **
*****************************************/

void tisWatchDog(int (*enableWatchDogHW)(int),
                 int (*disableWatchDogHW)(),
                 void (*petWatchDogHW)(),
                 int   rateMask,
                 int   delay)
{
    int      task_id;
    int      i;

    short    table_id;
    short   *ptable_id;

    if (watchdog_installed)
    {
        logMsg("watchdog aleady installed. Aborting this attempted watchdog installation.\n",0,0,0,0,0,0);
        return;
    }
    watchdog_installed++;
    watchdog_delay = delay;

    /* clear out the table */
    for (i = 0; i < MAXWATCHDOG; i++)
    {
        tiswatchdogtable[i].valid = 0;
        tiswatchdogtable[i].used = 0;
    }

    /* initialize pointers for VxWorks watchdog timer service routine */
    tisEnableWatchDogHW  = (FUNCPTR) enableWatchDogHW;
    tisPetWatchDogHW     = (FUNCPTR) petWatchDogHW;
    tisDisableWatchDogHW = (FUNCPTR) disableWatchDogHW;
    ptable_id = &table_id;
    task_id = taskIdSelf();

    tisRateMask          = rateMask;

    if ((watchdog_id = wdCreate()) == NULL)
    {
        logMsg("Error: failed wdCreate of watchdog timer. Aborting watchdog installation.\n",0,0,0,0,0,0);
        return;
    }

    /* enable global watch dog timer logic */
    global_watchdog_enable = TRUE;
    /* enable global watch dog timer logic */
    global_watchdog_abort = FALSE;
    /* enable global watch dog timer logic */
    global_watchdogHW_control = HARDWAREON;

/****************************************************************************************/
/* this section is here to let us measure the performance of the VxWorks watchdog timer */
/****************************************************************************************/
    if (tisWatchDogAttach(task_id, 1000, ptable_id) != OK)
    {
        logMsg("Test: attempt at self attachment failed.\n",0,0,0,0,0,0);
        (*disableWatchDogHW)();
        task_id = taskIdSelf();
        tisWatchDogDetach(task_id);
        return;
    }
    if (tisWatchDogPulse(table_id) != OK)
    {
        logMsg("Test: 1st attempt at self pulsing failed.\n",0,0,0,0,0,0);
        (*disableWatchDogHW)(rateMask);
        task_id = taskIdSelf();
        tisWatchDogDetach(task_id);
        return;
    }
    watchdog_tableid = table_id;
    tisWatchDogDisable(task_id);
/****************************************************************************************/
/****************************************************************************************/


    if((*enableWatchDogHW)(rateMask) != OK)
    {
        /* attempt to turn it off */
        (*disableWatchDogHW)();
        logMsg("Error: failed initialization of watchdog hardware. Aborting watchdog installation.\n",0,0,0,0,0,0);
        return;
    }
    /***************************************/
/*    (*disableWatchDogHW)();
    global_watchdogHW_control = HARDWAREOFF;
*/
    /***************************************/

    /* get operating system to pet the watchdog hardware timer - task can too easily be suspended. */
    if((watchdog_status = wdStart(watchdog_id, 1 /*watchdog_delay*/, (FUNCPTR)watchdog_timeout, 0)) == ERROR)
    {
        logMsg("Error: failed wdStart of watchdog timer. Aborting watchdog installation.\n",0,0,0,0,0,0);
        if (wdDelete(watchdog_id) == ERROR)
        {
            logMsg("Error: failed wdDelete of watchdog timer. Aborting watchdog installation.\n",0,0,0,0,0,0);
            return;
        }
        return;
       (*disableWatchDogHW)();
       global_watchdogHW_control = HARDWAREOFF;
    }
    logMsg("Task-Monitoring WatchDog Function successfully installed.\n",0,0,0,0,0,0);
    return;
}


/*****************************************
**      watchdog timout function        **
*****************************************/

void watchdog_timeout(int dummy)
{
    short    i;
    long     thisinterval;
    int      task_id;
    char     *task_name;
    int      status;

    watchdog_tick++;
    watchdog_hemlock = FALSE;
    /* do it. don't bother checking results. */
    tisWatchDogPulse(watchdog_tableid);

    switch(global_watchdogHW_control)
    {
        case HARDWAREONSET:
            global_watchdogHW_control = HARDWAREON;
            (*tisEnableWatchDogHW)(tisRateMask);
            break;
        case HARDWAREOFFSET:
            global_watchdogHW_control = HARDWAREOFF;
            (*tisDisableWatchDogHW)();
            break;
    }

    if (global_watchdog_enable == TRUE)
    {
        status = OK;
        for (i = 0; i < MAXWATCHDOG; i++)
        {
            if (tiswatchdogtable[i].valid && tiswatchdogtable[i].enabled)
            {
                thisinterval = tickGet() - tiswatchdogtable[i].lastpulsetick;
                if (thisinterval > tiswatchdogtable[i].interval)
                {
                    task_id = tiswatchdogtable[i].taskId;
                    task_name = taskName(task_id);
                    logMsg("Watch Dog Timer failed for Task Name %s, Task Id 0x%x\n",
                            (int)task_name, task_id,0,0,0,0);
                    status = ERROR;
                    watchdog_hemlock = TRUE; /* stop hardware petting */
                    break;
                }
            }
        }
    }

    if ((!watchdog_hemlock) && (global_watchdogHW_control != HARDWAREOFF))
    {
        /* pet the hardware */
        (*tisPetWatchDogHW)();
        watchdog_petcount++;
    }

    if (global_watchdog_abort)
    {
        /* requested, disable the hardware and exit. */
        (*tisDisableWatchDogHW)();
        /* clear flag to allow watchdog to be re-started. */
        watchdog_installed = 0;
        if (wdDelete(watchdog_id) == ERROR)
        {
            logMsg("Error: wdDelete.\n",0,0,0,0,0,0);
            return;
        }
        logMsg("WatchDog Terminated.\n",0,0,0,0,0,0);
        global_watchdog_abort = 2;       /* let abort program know its done. */
        return;
    }
    else
    {
        /* restart the watchdog */
        wdStart(watchdog_id, watchdog_delay, (FUNCPTR)watchdog_timeout, 0);
        return;
    }
}

/*********************************************
**    attach a task to the watchdog table   **
*********************************************/

int  tisWatchDogAttach(int task_id, short interval, short *wdId)
{
    short i;
    char  *task_name;

    task_name = taskName(task_id);
    if (task_name == 0)
    {
        logMsg("Watch Dog Timer Attachment failed for Task Id 0x%x (null task name)\n",
                task_id,0,0,0,0,0);
        return(ERROR);
    }

    for (i = 0; i < MAXWATCHDOG; i++)
    {

        if (!tiswatchdogtable[i].valid)
        {
            logMsg("Watch Dog Timer Attached for Task Name %s, Task Id 0x%x\n",
                    (int)task_name, task_id,0,0,0,0);
            tiswatchdogtable[i].used = TRUE;
            tiswatchdogtable[i].valid = TRUE;
            tiswatchdogtable[i].enabled = TRUE;
            tiswatchdogtable[i].interval = interval;
            tiswatchdogtable[i].lastpulsetick = tickGet();
            tiswatchdogtable[i].taskId = task_id;
            tiswatchdogtable[i].average32 = 0;
            tiswatchdogtable[i].max = 0;
            *wdId = i;
            return(OK);
        }
        else
        {
            if (tiswatchdogtable[i].taskId == task_id)
            {
                logMsg("Watch Dog Timer Re-attached for Task Name %s, Task Id 0x%x\n",
                        (int)task_name, task_id,0,0,0,0);
                tiswatchdogtable[i].valid = TRUE;
                tiswatchdogtable[i].enabled = TRUE;
                tiswatchdogtable[i].interval = interval;
                tiswatchdogtable[i].lastpulsetick = tickGet();
                tiswatchdogtable[i].taskId = task_id;
                tiswatchdogtable[i].average32 = 0;
                tiswatchdogtable[i].max = 0;
                *wdId = i;
                return(OK);
            }
        }
    }
    logMsg("Watch Dog Timer Attachment Failed for Task Name %s, Task Id 0x%x\n",
            (int)task_name, task_id,0,0,0,0);
    return(ERROR);
}




/********************************************************
**    set the time for an previously attached task     **
********************************************************/

int  tisWatchDogSet(int task_id, short interval)
{
    short i;
    char  *task_name;

    task_name = taskName(task_id);
    if (task_name == 0)
    {
        logMsg("Watch Dog Timer Set failed for Task Id 0x%x (null task name)\n",
                task_id,0,0,0,0,0);
        return(ERROR);
    }

    for (i = 0; i < MAXWATCHDOG; i++)
    {

        if ((tiswatchdogtable[i].valid) && (tiswatchdogtable[i].taskId == task_id))
        {
            logMsg("Watch Dog Timer Set for Task Name %s, Task Id 0x%x\n",
                    (int)task_name, task_id,0,0,0,0);
            tiswatchdogtable[i].interval = interval;
            tiswatchdogtable[i].lastpulsetick = tickGet();
            tiswatchdogtable[i].taskId = task_id;
            tiswatchdogtable[i].average32 = 0;
            tiswatchdogtable[i].max = 0;
            return(OK);
        }
    }
    logMsg("Watch Dog Timer Set Failed for Task Name %s, Task Id 0x%x\n",
            (int)task_name, task_id,0,0,0,0);
    return(ERROR);
}




/*************************************************
**      pulse the task's watchdog table entry   **
*************************************************/

int  tisWatchDogPulse(short wdId)
{
    long  thistime;
    long  diff;

    if ((wdId >= 0) && (wdId < MAXWATCHDOG))
    {
        thistime = tickGet();
        diff = thistime - tiswatchdogtable[wdId].lastpulsetick;
        tiswatchdogtable[wdId].lastpulsetick = thistime;
        tiswatchdogtable[wdId].average32 += (diff - ((tiswatchdogtable[wdId].average32+16)/32));
        if (diff > tiswatchdogtable[wdId].max)
        {
            tiswatchdogtable[wdId].max = diff;
        }
        return(OK);
    }
    return(ERROR);
}



/**********************************************
**   detach a task from the watchdog table   **
**********************************************/

int  tisWatchDogDetach(int task_id)
{
    short i;
    char  *task_name;

    task_name = taskName(task_id);

    for (i = 0; i < MAXWATCHDOG; i++)
    {
        if ((tiswatchdogtable[i].valid) && (tiswatchdogtable[i].taskId == task_id))
        {
            tiswatchdogtable[i].valid = FALSE;
            logMsg("Watch Dog Timer Detached for Task Name %s, Task Id 0x%x\n",
                    (int)task_name, task_id,0,0,0,0);
            return(OK);
        }
    }
    logMsg("Watch Dog Timer Detachment failed for Task Name %s Task Id 0x%x\n",
            (int)task_name, task_id,0,0,0,0);
    return(ERROR);
}



/**********************************************
**    disable a task in the watchdog table   **
**                                           **
**     or                                    **
**                                           **
**    globally disable watchdog table        **
**********************************************/

int  tisWatchDogDisable(int task_id)
{
    short i;
    char  *task_name;

    if (task_id == 0)
    {
        global_watchdog_enable = FALSE;
        logMsg("Watch Dog Monitoring Globally Disabled.\n",0,0,0,0,0,0);

        return(OK);
    }

    if (task_id == 1)
    {
        global_watchdogHW_control = HARDWAREOFFSET;
        global_watchdog_enable = FALSE;
        logMsg("Watch Dog Monitoring Globally Disabled.\n",0,0,0,0,0,0);
        logMsg("Watch Dog Hardware Timer Disabled.\n",0,0,0,0,0,0);

        return(OK);
    }

    task_name = taskName(task_id);

    for (i = 0; i < MAXWATCHDOG; i++)
    {
        if ((tiswatchdogtable[i].valid) && (tiswatchdogtable[i].taskId == task_id))
        {
            tiswatchdogtable[i].enabled = FALSE;
            logMsg("Watch Dog Monitoring Disabled for Task Name %s, Task Id 0x%x\n",
            (int)task_name, task_id,0,0,0,0);
            return(OK);
        }
    }
    logMsg("Watch Dog Monitoring Disablement failed for Task Name %s, Task Id 0x%x\n",
            (int)task_name, task_id,0,0,0,0);
    return(ERROR);
}


/**********************************************
**    enable a task in the watchdog table    **
**                                           **
**     or                                    **
**                                           **
**    globally disable watchdog table        **
**********************************************/

int  tisWatchDogEnable(int task_id)
{
    short i;
    char  *task_name;

    if (task_id == 0)
    {
        global_watchdog_enable = TRUE;

        /* update time for all valid tasks, any or all of which may have been suspended */
        for (i = 0; i < MAXWATCHDOG; i++)
        {
            if (tiswatchdogtable[i].valid)
            {
/*
                tiswatchdogtable[i].lastpulsetick = tickGet();
                tiswatchdogtable[i].enabled = TRUE;
*/
            }
        }
        logMsg("Watch Dog Monitoring Globally Enabled.\n",0,0,0,0,0,0);
        return(OK);
    }

    if (task_id == 1)
    {
        global_watchdog_enable = TRUE;
        global_watchdogHW_control = HARDWAREONSET;

        /* update time for all valid tasks, any or all of which may have been suspended */
        for (i = 0; i < MAXWATCHDOG; i++)
        {
            if (tiswatchdogtable[i].valid)
            {
/*
                tiswatchdogtable[i].lastpulsetick = tickGet();
                tiswatchdogtable[i].enabled = TRUE;
*/
            }
        }
        logMsg("Watch Dog Monitoring Globally Enabled.\n",0,0,0,0,0,0);
        logMsg("Watch Dog Hardware Timer Enabled.\n",0,0,0,0,0,0);
        return(OK);
    }

    task_name = taskName(task_id);

    for (i = 0; i < MAXWATCHDOG; i++)
    {
        if ((tiswatchdogtable[i].valid) && (tiswatchdogtable[i].taskId == task_id))
        {
/*
            tiswatchdogtable[i].lastpulsetick = tickGet();
*/
            tiswatchdogtable[i].enabled = TRUE;
            logMsg("Watch Dog Monitoring Enabled for Task Name %s, Task Id 0x%x\n",
                    (int)task_name, task_id,0,0,0,0);
            return(OK);
        }
    }
    logMsg("Watch Dog Monitoring Enablement failed for Task Name %s, Task Id 0x%x\n",
            (int)task_name, task_id,0,0,0,0);
    return(ERROR);
}

/*********************************************
**    abort the watchdog task               **
*********************************************/

int  tisWatchDogAbort(void)
{
    global_watchdog_abort = TRUE;
    taskDelay(60);
    if (global_watchdog_abort == 2)
    {
        logMsg("Watch Dog Abort Completed.\n",0,0,0,0,0,0);
        return(OK);
    }
    else
    {
        logMsg("Watch Dog Abort Failed.\n",0,0,0,0,0,0);
        return(ERROR);
    }
}

/*****************************************
**      spawn 100 dummy tasks           **
*****************************************/

int wdt100(void)
{
    int i;
    int priority, save;
    int *ppriority;
    int task_id;

    ppriority = &priority;
    task_id = taskIdSelf();
    taskPriorityGet(task_id, ppriority);
    save = priority;
    taskPrioritySet(task_id, 255);

    logMsg("*************** wdt100 Begins ****************\n",0,0,0,0,0,0);
    for (i=0; i <100; i++)
    {
        wdtdummy(100+i, 200+(60*i), 100+(2*i));
        taskDelay(2);
    }
    save = 1;
    taskPrioritySet(task_id, save);
    return(OK);
} 

/*****************************************
**      spawn 10 dummy tasks            **
*****************************************/

int wdt(int n)
{
    int i;
    int priority, save;
    int *ppriority;
    int task_id;

    ppriority = &priority;
    task_id = taskIdSelf();
    taskPriorityGet(task_id, ppriority);
    save = priority;
    taskPrioritySet(task_id, 255);

    logMsg("*************** wdt(%d) Begins ****************\n", n,0,0,0,0,0);
    for (i=0; i <n; i++)
    {
        wdtdummy(100+i, 200+(60*i), 100+(2*i));
        taskDelay(2);
    }
    save = 1;
    taskPrioritySet(task_id, save);
    return(OK);
} 
/*****************************************
**      spawn 10 dummy tasks            **
*****************************************/

int wdt10(void)
{
    int i;
    int priority, save;
    int *ppriority;
    int task_id;

    ppriority = &priority;
    task_id = taskIdSelf();
    taskPriorityGet(task_id, ppriority);
    save = priority;
    taskPrioritySet(task_id, 255);

    logMsg("*************** wdt100 Begins ****************\n",0,0,0,0,0,0);
    for (i=0; i <10; i++)
    {
        wdtdummy(100+i, 200+(60*i), 100+(2*i));
        taskDelay(2);
    }
    save = 1;
    taskPrioritySet(task_id, save);
    return(OK);
} 

/*****************************************
**      spawn a dummy task              **
*****************************************/

int wdtdummy(int priority, short interval, short delay)
{
    static char taskhead[] = "wdtask";
    static char tasknum[4];
    static char taskname[20];

    static int  tasknumber;

/*     if (tasknumber > 99) return(ERROR);*/

    strcpy(taskname, taskhead);
    tasknum[0] = '0'+(tasknumber/100);
    tasknum[1] = '0'+((tasknumber%100)/10);
    tasknum[2] = '0'+(tasknumber%10);
    tasknum[3] = 0;
    strcat(taskname, tasknum);

    tasknumber++;

    taskSpawn (taskname,
               priority,
               0,
               2000,
               (FUNCPTR)dummytask,
               interval,
               delay, 0,0,0,0,0,0,0,0);
    return (OK);
}

/*****************************************
** dummy task to loadup watchdog table  **
*****************************************/

void dummytask(short interval, short delay)
{
    int     task_id;
    short   table_id;
    short  *ptable_id;

    ptable_id = &table_id;
    task_id = taskIdSelf();

    if (tisWatchDogAttach(task_id, interval, ptable_id) != OK)
    {
        logMsg("Test: attempt at dummy task attachment failed.\n",0,0,0,0,0,0);
        tisWatchDogDetach(task_id);
        return;
    }

    if (tisWatchDogPulse(table_id) != OK)
    {
        logMsg("Test: 1st attempt at dummy task pulsing failed.\n",0,0,0,0,0,0);
        tisWatchDogDetach(task_id);
        return;
    }

    while (1)
    {
        if (tisWatchDogPulse(table_id) != OK)
        {
            logMsg("Test: Nth attempt at dummy task pulsing failed.\n",0,0,0,0,0,0);
            tisWatchDogDetach(task_id);
            return;
        }
        taskDelay(delay);
    }
}


/*********************************************
**    abort the watchdog task               **
*********************************************/

int  wdtabort(void)
{
    global_watchdog_abort = TRUE;
    return(OK);
}

/*********************************************
**    wdthwon:turn on watchdog hardware     **
*********************************************/

int  wdthwon(void)
{
    global_watchdogHW_control = HARDWAREONSET;
    return(OK);
}


/*********************************************
**    wdthwoff: turn off watchdog hardware  **
*********************************************/

int  wdthwoff(void)
{
    global_watchdogHW_control = HARDWAREOFFSET;
    return(OK);
}


/*******************************************************
**  clear all watchdog table entry's statistics       **
*******************************************************/

void clearallwdt(void)
{
    int i;

    /* update time for all valid tasks, any or all of which may have been suspended */
    for (i = 0; i < MAXWATCHDOG; i++)
    {
        if (tiswatchdogtable[i].valid)
        {
            tiswatchdogtable[i].max = 0;
            tiswatchdogtable[i].average32 = 0;
        }
    }
    logMsg("All Valid WatchDog Statistics cleared.\n",0,0,0,0,0,0);
    return;
}

/*******************************************************
**  clear a single watchdog table entry's statistics  **
*******************************************************/

void clearwdt(int index)
{
    tiswatchdogtable[index].max = 0;
    tiswatchdogtable[index].average32 = 0;
}


/**********************************************
**    dump watchdog table                    **
**********************************************/

void  tisWatchDogList(void)
{
    unsigned	short i, addr, linecnt ;
    unsigned	char  buffer[200];
                long  task_id;
                char *task_name;
                int   task_priority;
                int  *ppriority;

    ppriority = &task_priority;

    printf("\n");
    printf (" Watch Dog Task Monitor Installed = %d\n", watchdog_installed);
    printf (" Global Enable Flag               = %d\n", global_watchdog_enable);
    printf("\n");

    linecnt = 1;
    printwatchdoghdr(buffer, linecnt);
    addr = 0;
    for (i=0; (i < MAXWATCHDOG); i++)
    {
        if (tiswatchdogtable[i].valid == TRUE)
        {
            taskDelay(6);
            task_id = tiswatchdogtable[i].taskId;
            task_name = taskName(task_id);
            if (strlen(task_name) == 0) continue;
            taskPriorityGet(task_id, ppriority);
            printf(" %2d    %2d    %8lx   %16s     %3d        %1x        %4d %7d %7ld",
                    i,
                    tiswatchdogtable[i].valid,
                    task_id,
                    task_name,
                    task_priority,
                    tiswatchdogtable[i].enabled,
                    tiswatchdogtable[i].interval,
                    tiswatchdogtable[i].max,
                   (tiswatchdogtable[i].average32+16)/32);

	    printf("\n");
	    linecnt += 1;
	    if ( (linecnt >= 17) && ((i +1) < MAXWATCHDOG) )
	    {
		linecnt = 1;
		printwatchdoghdr(buffer, linecnt);
	    }
	    addr += 1;		
	}
    }
}

void dumpwdt(void)
{
    unsigned	short i, addr, linecnt ;
    unsigned	char  buffer[200];
                long  task_id;
                char *task_name;
                int   task_priority;
                int  *ppriority;

    ppriority = &task_priority;

    printf("\n");
    printf (" Watch Dog Task Monitor Installed = %d\n", watchdog_installed);
    printf (" Global Enable Flag               = %d\n", global_watchdog_enable);
    printf (" Global Abort Flag                = %d\n", global_watchdog_abort);
    printf (" Global Hardware Control Flag     = %d\n", global_watchdogHW_control);
    printf (" (1 = HW disabled, 2= HW enabled)\n");

    printf("\n");
    printf (" Watchdog Delay     = %d\n", watchdog_delay);
    printf (" Watchdog Id        = %d\n", (int)watchdog_id);
    printf (" Watchdog Status    = %d\n", watchdog_status);
    printf (" Watchdog Hemlock   = %d\n", watchdog_hemlock);
    printf (" Watchdog Tick      = %d\n", watchdog_tick);
    printf (" Watchdog Pet Count = %d\n", watchdog_petcount);

    linecnt = 1;
    printwatchdoghdr(buffer, linecnt);
    addr = 0;
    for (i=0; (i < MAXWATCHDOG); i++)
    {
        if (tiswatchdogtable[i].used == TRUE)
        {
            taskDelay(6);
            task_id = tiswatchdogtable[i].taskId;
            task_name = taskName(task_id);
            taskPriorityGet(task_id, ppriority);
            printf(" %2d    %2d    %8lx   %16s     %3d        %1x        %4d %7d %7ld",
                    i,
                    tiswatchdogtable[i].valid,
                    task_id,
                    task_name,
                    task_priority,
                    tiswatchdogtable[i].enabled,
                    tiswatchdogtable[i].interval,
                    tiswatchdogtable[i].max,
                   (tiswatchdogtable[i].average32+16)/32);

	    printf("\n");
	    linecnt += 1;
	    if ( (linecnt >= 17) && ((i +1) < MAXWATCHDOG) )
	    {
		linecnt = 1;
		printwatchdoghdr(buffer, linecnt);
	    }
	    addr += 1;		
	}
    }
}

/**************************************
**    dump watchdog table header     **
**************************************/

static void printwatchdoghdr (unsigned char *buffer, unsigned short cnt)
{	
	taskDelay(2);
	sprintf(buffer,"                                   Watch Dog Table          ");
	printf( "%s\n", buffer);
	cnt += 1;
        sprintf(buffer,
               " ID:  Valid  -TaskId-  --- TaskName ---   Priority  Enabled  Interval Maximum Average(32) ");
 	printf( "%s\n", buffer);
 	cnt += 1;
}

/*	End of File:	mv167.c */
