/*datamove.c,v 1.1.1.1 1999/08/03 14:44:17 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************************
 *
 *  MODULE NAME            : datamove.c 
 *  AUTHOR                 : emsteel
 *  DATE                   : 08 Oct 93
 *
 *
 *  DESCRIPTION: functions for the datamove task
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 03 Jan 94
 *  REVISION:  set ack to 0 when set init to TRUE
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 29 Mar 94
 *  REVISION:  make forward func def agree with actual def
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 03 Aug 94
 *  REVISION:	datamove(), instead of calling ca_add_event in the main task 
 *		loop, after ichd allocated, create a connection handler to be 
 *		called when input is connected or reconnected
 *
 *		dmInputConnectHandler() new func, if connected and no event 
 *		id then	create one, if successful set INIT and INI2 init flags,
 *		if we are reconnected then force a call to the event handler 
 *		to fire a monitor
 *
 *		dmEventHandler(), if datamove point does not own resource,
 *		do not post VAL, only ca_put to output if there is an output
 *		chid (ochd), we are connected and we are MASTER for our 
 *		resource
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 18 Sep 96
 *  REVISION:	added READ/WRITE alarms, connection handler for the output
 *		channel, and in the event handler for input channel, if the
 *		arg chid != the input chid (ichd), then close it and return
 *
****************************************************************************/

/****************************************************************************
 *
 * Programmer's Notes:
 *
 *    cadef.h - see .../share/epicsH
 *       def of chid, evid, ca_name, ca_message etc
 *    datamoveMsgQId - see initialize () in recDatamove.c
 *    datemoveSem    - see initialize () in recDatamove.c
 *
****************************************************************************/


#include <vxWorks.h>
#include <inetLib.h>
#include <errno.h>
#include <msgQLib.h>
#include <semLib.h>
#include <taskLib.h>
#include <string.h>
#include <stdio.h>

#include "alarm.h"
#include "cadef.h"    /* channel access */
#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#include "dbRecType.h"
#include "rec/datamoveRecord.h"
#include "error.h"
#include "restable.h"
#include "db_addr.h"
#include "dbEvent.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

MSG_Q_ID datamoveMsgQId;    /* for message queueing */
SEM_ID   datamoveSem;       /* for semaphores       */

static void dmInputConnectHandler (struct connection_handler_args connArgs);
static void dmEventHandler (struct event_handler_args evArgs);
static void dmOutputConnectHandler (struct connection_handler_args connArgs);
int	dmCaAvail = 0;
static unsigned short  dmtype;


/****************************************************************************
 *
 *  MODULE NAME            : datamove 
 *  AUTHOR                 : emsteel
 *  DATE                   : 08 Oct 93
 *
 *  DESCRIPTION: driver for the datamove task
 *               numEvents - each time a datamove record is init'd it posts
 *                           its address into datamoveMsgQId, if 
 *                           ca_add_event () for record is OK then 
 *                           numEvents++
 *               if no events, then wait forever for something to do
 *               a monitor event is only established for the input channel
 *                  name and not the output channel
 *
****************************************************************************/

void datamove (void)
{
    struct datamoveRecord    *pdatamove;
    int     numEvents = 0;
    long    status;
    chid    pCh;             /* channel id ptr */
    char    errmsg[128];
    char    *pstr;

    /* Keep trying until we can initialize */

    while (ca_task_initialize() != ECA_NORMAL)
	taskDelay(300);
    dmCaAvail = 1;


    /* dmtype is needed by process_cnts[] */

    for (dmtype = 1; dmtype < 100; dmtype++)
    {
	pstr = GET_PRECTYPE (dmtype);
	if (strncmp (pstr, "datamove", 8) == 0)
	    break;
    }

    for (;;)
    {
	if (numEvents)
	{
	    ca_pend_event (1.0);
	}
	else
	{
	    if ((semTake (datamoveSem, WAIT_FOREVER)) == ERROR)
	    {
		sprintf (errmsg, "datamove task, semTake error");
		errMessage (-1L, errmsg);
	    }
	}
	while ( (msgQReceive (datamoveMsgQId, (char *)&pdatamove,
			      sizeof (pdatamove), NO_WAIT) ) != ERROR)
	{
            /* if there is an input channel established,
               and input channel name has changed, then clear old 
               old event id, old channel id and make ready to 
               establish a new channel
            */

	    if (pdatamove -> ichd)
	    {
		if ( (strncmp (pdatamove -> inam, 
			       ca_name ((chid)pdatamove -> ichd), 34) ) != 0)
		{
		    if ( (status = ca_clear_event ((evid)pdatamove -> evid) ) !=
                                                              ECA_NORMAL)
		    {
			sprintf (errmsg, "datamove,clear event for %s in %s (%s)",
				          pdatamove -> inam, pdatamove -> name,
					  ca_message (status));
			errMessage (-1L, errmsg);
		    }
		    else
		    {
			numEvents--;
			pdatamove -> evid = 0;
		    }
		    ca_flush_io ();
		    if ((status = ca_clear_channel ((chid)pdatamove -> ichd)) !=
                                                              ECA_NORMAL)
		    {
			sprintf (errmsg, "datamove,clear chan for %s in %s (%s)",
					  pdatamove -> inam, pdatamove -> name,
					  ca_message (status));
			errMessage (-1L, errmsg);
		    }
		    else
			pdatamove -> ichd = 0;
		    ca_flush_io ();

		}
	    }

            /* if there is no channel id ptr, try to establish one */

	    if (!pdatamove -> ichd)
	    {
		if ( (status = ca_build_and_connect (pdatamove -> inam, 
						     AGNT_NOCONN,
						     0,
						     &pCh,
						     0,
						     dmInputConnectHandler,
						     (void*)pdatamove) ) != ECA_NORMAL)
		{
		    sprintf (errmsg, "datamove,ca_build of %s in %s (%s)",
				      pdatamove -> inam, pdatamove -> name,
				      ca_message (status));
		    errMessage (-1L, errmsg);
		}
		else
		{
		    pdatamove -> ichd = (void*)pCh;
		    numEvents++;
		}
	    }

            /* if there is an ouput channel established, 
               if output channel name has changed, then clear old channel,
                  and make ready to establish a new channel
            */

	    if (pdatamove -> ochd)
	    {
		if ((strncmp (pdatamove -> onam,
                          ca_name ((chid)pdatamove -> ochd), 34)) != 0)
		{
		    if ((status = ca_clear_channel ((chid)pdatamove -> ochd)) !=
                                                               ECA_NORMAL)
		    {
			sprintf (errmsg, "datamove,clear ochan for %s in %s (%s)",
					  pdatamove -> onam, pdatamove -> name,
					  ca_message (status));
			errMessage (-1L, errmsg);
		    }
		    else
			pdatamove -> ochd = 0;
		    ca_flush_io ();
		}
	    }

            /* if there is no output channel id and there is an 
               output channel name, try to establish one
            */

	    if ( (!pdatamove -> ochd) && pdatamove -> onam[0])
	    {
		if ( (status = ca_build_and_connect (pdatamove -> onam, 
						     AGNT_NOCONN,
						     0,
						     &pCh,
						     0,
						     dmOutputConnectHandler,
						     (void*)pdatamove) ) != ECA_NORMAL)
		{
		    sprintf (errmsg, "datamove,ca_build of %s for %s (%s)",
				      pdatamove -> onam, pdatamove -> name,
				      ca_message (status));
		    errMessage (-1L, errmsg);
		}
		else
		    pdatamove -> ochd = (void*)pCh;
	    } 

	} /* end while msgQReceive () */

    } /* end for (;;) */

} /* end func datamove () */


/****************************************************************************
 *
 *  MODULE NAME            : dmEventHandler
 *  AUTHOR                 : emsteel
 *  DATE                   : 11 Oct 93
 *
 *  DESCRIPTION: 
 *              called every time an event is processed
 *
 *  NOTE:
 *       LAST_TYPE def'd in ../share/epicsH/db_access.h
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 13 Dec 93
 *  REVISION: cast input value to a var of its type then take its address
 *            to pass to ca_put ()
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 20 Oct 94
 *  REVISION:	added call to tsLocalTime, do it here and not in 
 *		recDatamove:process() as is the normal place
 *
****************************************************************************/

static void dmEventHandler (struct event_handler_args evArgs)
{
    struct datamoveRecord *pdatamove;
    int			status;
    chtype		odbrType = 0;    /* output DBR_xxx */
    short		ashort;
    float		afloat;
    char		achar;
    long		along;
    double		adouble;
    void		*avoid = NULL;
    char		errmsg[128];
    unsigned short	master = 0;
    unsigned short	stat;
    unsigned short	sevr;

    pdatamove = (struct datamoveRecord *)(evArgs.usr);

    /* we were called by channel that is not mine, somehow
       I am also connected to another channel, so close
       it and get out
    */
    if (((void *)evArgs.chid) != pdatamove -> ichd)
    {
	ca_clear_channel ((chid)evArgs.chid);
	return;
    }

    /* if record has not been initalized (init_record) do nothing */

    if ((!pdatamove -> init) || (!pdatamove -> ini2) )
	return;

    /* get time stamp for last time record was processed */

    tsLocalTime (&pdatamove -> lptm);

    /* if record is tagged, off scan or in manual mode, do nothing */

    if (pdatamove -> tagd || (pdatamove -> oscn == OFF_SCAN) ||
			     (pdatamove -> mode != NORMAL) )
	return;

    /* update number of process times */
    if (process_stats)
	process_cnts[dmtype]++;

    /* find out if this datamove points' resource is master */

    resGetCaAnswer (pdatamove -> radr, &master);

    stat = pdatamove -> stat;
    sevr = pdatamove -> sevr;

      /* if there is an output channel, gets it's DBR_xxx type */

    if (pdatamove -> ochd)
    {
	odbrType = ca_field_type ((chid)pdatamove -> ochd);
	/* if the output is not connected */
	if (ca_state ((chid)pdatamove -> ochd) != cs_conn)
	{
	    pdatamove -> stat = WRITE_ALARM;
	    pdatamove -> sevr = VALID_ALARM;
	}
    }
    else
    {
	pdatamove -> stat = NO_ALARM;
	pdatamove -> sevr = NO_ALARM;
    }
    if ( ((stat != pdatamove -> stat) ||
	  (sevr != pdatamove -> sevr) ) &&
	 master)
    {
	db_post_events ((struct dbCommon *)pdatamove, (caddr_t)&pdatamove -> stat, DBE_VALUE);
	db_post_events ((struct dbCommon *)pdatamove, (caddr_t)&pdatamove -> sevr, DBE_VALUE);
    }

    /* put event value into record VAL field and post the event
       1st convert input type to a double
       if there is an output channel, convert input type to
       output type
    */
   
    switch (evArgs.type)
    {
	case OLD_DBR_SHORT:    /* DBR_INT is the same */
	case OLD_DBR_ENUM:
	    ashort = *((short *)evArgs.dbr);
	    pdatamove -> val = (double)ashort;
	    if (master)
	    	db_post_events ((struct dbCommon *)pdatamove, (caddr_t)&pdatamove -> val, DBE_VALUE);
	    if (pdatamove -> ochd)
	    {
		switch (odbrType)
		{
		    case OLD_DBR_SHORT:    /* DBR_INT is the same */
		    case OLD_DBR_ENUM:
			avoid = evArgs.dbr;
			break;

		    case OLD_DBR_FLOAT:
			afloat = (float)ashort;
			avoid = (void *)(&afloat);
			break;

		    case OLD_DBR_CHAR:
			achar = (char)ashort;
			avoid = (void *)(&achar);
			break;

		    case OLD_DBR_LONG:
			along = (long)ashort;
			avoid = (void *)(&along);
			break;

		    case OLD_DBR_DOUBLE:
			adouble = (double)ashort;
			avoid = (void *)(&adouble);
			break;

		    case TYPENOTCONN:
			return;

		    default:
			sprintf (errmsg, "datamoveEH, illegal output DBR %d for %s",
					  odbrType, pdatamove -> name);
			errMessage (-1L, errmsg);
			return;
		}
	    }
	    break;

	case OLD_DBR_FLOAT:
	    afloat = *((float *)evArgs.dbr);
	    pdatamove -> val = (double)afloat;
	    if (master)
	    	db_post_events ((struct dbCommon *)pdatamove, (caddr_t)&pdatamove -> val, DBE_VALUE);
	    if (pdatamove -> ochd)
	    {
		switch (odbrType)
		{
		    case OLD_DBR_SHORT:    /* DBR_INT is the same */
		    case OLD_DBR_ENUM:
			ashort = (short)afloat;
			avoid = (void *)(&ashort);
			break;

		    case OLD_DBR_FLOAT:
			avoid = evArgs.dbr;
			break;

		    case OLD_DBR_CHAR:
			achar = (char)afloat;
			avoid = (void *)(&achar);
			break;

		    case OLD_DBR_LONG:
			along = (long)afloat;
			avoid = (void *)(&along);
			break;

		    case OLD_DBR_DOUBLE:
			adouble =(double)afloat;
			avoid = (void *)(&adouble);
			break;

		    case TYPENOTCONN:
			return;

		    default:
			sprintf (errmsg, "datamoveEH, illegal output DBR %d for %s",
					  odbrType, pdatamove -> name);
			errMessage (-1L, errmsg);
			return;
		}
	    }
	    break;

	case OLD_DBR_CHAR:
	    achar = *((char *)evArgs.dbr);
	    pdatamove -> val = (double)achar;
	    if (master)
	    	db_post_events ((struct dbCommon *)pdatamove, (caddr_t)&pdatamove -> val, DBE_VALUE);
	    if (pdatamove -> ochd)
	    {
		switch (odbrType)
		{
		    case OLD_DBR_SHORT:    /* DBR_INT is the same */
		    case OLD_DBR_ENUM:
			ashort = (short)achar;
			avoid = (void *)(&ashort);
			break;

		    case OLD_DBR_FLOAT:
			afloat = (float)achar;
			avoid = (void *)(&afloat);
			break;

		    case OLD_DBR_CHAR:
			avoid = evArgs.dbr;
			break;

		    case OLD_DBR_LONG:
			along = (long)achar;
			avoid = (void *)(&along);
			break;

		    case OLD_DBR_DOUBLE:
			adouble = (double)achar;
			avoid = (void *)(&adouble);
			break;

		    case TYPENOTCONN:
			return;

		    default:
			sprintf (errmsg, "datamoveEH, illegal output DBR %d for %s",
					  odbrType, pdatamove -> name);
			errMessage (-1L, errmsg);
			return;
		}
	    }
	    break;

	case OLD_DBR_LONG:
	    along = *((long *)evArgs.dbr);
	    pdatamove -> val = (double)along;
	    if (master)
	    	db_post_events ((struct dbCommon *)pdatamove, (caddr_t)&pdatamove -> val, DBE_VALUE);
	    if (pdatamove -> ochd)
	    {
		switch (odbrType)
		{
		    case OLD_DBR_SHORT:    /* DBR_INT is the same */
		    case OLD_DBR_ENUM:
			ashort = (short)along;
			avoid = (void *)(&ashort);
			break;

		    case OLD_DBR_FLOAT:
			afloat = (float)along;
			avoid = (void *)(&afloat);
			break;

		    case OLD_DBR_CHAR:
			achar = (char)along;
			avoid = (void *)(&achar);
			break;

		    case OLD_DBR_LONG:
			avoid = evArgs.dbr;
			break;

		    case OLD_DBR_DOUBLE:
			adouble = (double)along;
			avoid = (void *)(&adouble);
			break;

		    case TYPENOTCONN:
			return;

		    default:
			sprintf (errmsg, "datamoveEH, illegal output DBR %d for %s",
					  odbrType, pdatamove -> name);
			errMessage (-1L, errmsg);
			return;
		}
	    }
	    break;

	case OLD_DBR_DOUBLE:
	    adouble = *((double *)evArgs.dbr);
	    pdatamove -> val = adouble;
	    if (master)
	    	db_post_events ((struct dbCommon *)pdatamove, (caddr_t)&pdatamove -> val, DBE_VALUE);
	    if (pdatamove -> ochd)
	    {
		switch (odbrType)
		{
		    case OLD_DBR_SHORT:    /* DBR_INT is the same */
		    case OLD_DBR_ENUM:
			ashort = (short)adouble;
			avoid = (void *)(&ashort);
			break;

		    case OLD_DBR_FLOAT:
			afloat = (float)adouble;
			avoid = (void *)(&afloat);
			break;

		    case OLD_DBR_CHAR:
			achar = (char)adouble;
			avoid = (void *)(&achar);
			break;

		    case OLD_DBR_LONG:
			along = (long)adouble;
			avoid = (void *)(&along);
			break;

		    case OLD_DBR_DOUBLE:
			avoid = evArgs.dbr;
			break;

		    case TYPENOTCONN:
			return;

		    default:
			sprintf (errmsg, "datamoveEH, illegal output DBR %d for %s",
					  odbrType, pdatamove -> name);
			errMessage (-1L, errmsg);
			return;
		}
	    }
	    break;

	default:
	    sprintf (errmsg, "datamoveEH, illegal input DBR %ld for %s",
			      evArgs.type, pdatamove -> name);
	    errMessage (-1L, errmsg);
	    return;

    } /* end switch (evArgs.type) */

    /* if there is an output channel and we are still connected and
       this datamove points resource is master, send value
    */

    if (pdatamove -> ochd &&  
        ((ca_state ((chid)pdatamove -> ochd)) == cs_conn) && 
	master)
    {
	if ((status = ca_put (odbrType, (chid)pdatamove -> ochd,
                            avoid)) != ECA_NORMAL)
	{
	    sprintf (errmsg, "datamoveEH,ca_put,for %s in %s (%s)",
			      pdatamove -> onam, pdatamove -> name,
			      ca_message (status));
	    errMessage (-1L, errmsg);
	}
	else
	    ca_flush_io ();
    }

    return;

} /* end func dmEventHandler () */



/****************************************************************************
 *
 *  MODULE NAME            : dmInputConnectHandler
 *  AUTHOR                 : emsteel
 *  DATE                   : 28 July 94
 *
 *  DESCRIPTION: 
 *		called when connect happens for input channel
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : 
 *  REVISION DATE          :
 *  REVISION:
 *
****************************************************************************/

static void dmInputConnectHandler (struct connection_handler_args connArgs)
{
    struct datamoveRecord *pdatamove;
    chtype	fld_type;
    char	errmsg[128];
    long	status;
    evid	pEv;
    chid	pCh;
    unsigned short	master = 0;
    unsigned short	stat;
    unsigned short	sevr;

    pCh = connArgs.chid;
    pdatamove = (struct datamoveRecord *)ca_puser (pCh);
    stat = pdatamove -> stat;
    sevr = pdatamove -> sevr;

    resGetCaAnswer (pdatamove -> radr, &master);

    if (connArgs.op == CA_OP_CONN_UP)
    {
	fld_type = ca_field_type (pCh);
	if (!pdatamove -> evid)
	{
	    if ( (status = ca_add_event (fld_type, pCh, dmEventHandler,
					 (void*)pdatamove, &pEv) ) != ECA_NORMAL)
	    {
		sprintf (errmsg, "datamove: add event %s for %s (%s)",
				  pdatamove -> inam, pdatamove -> name,
				  ca_message (status) );
		errMessage (-1L, errmsg);
		return;
	    }
	    else
	    {
		pdatamove -> evid = (void *)pEv;
		ca_flush_io ();

		/* input is ok, set init flags */

		pdatamove -> init = 1;
		pdatamove -> ini2 = 1;
	    }
	}
	else
	{
	    if ( (status = ca_array_get_callback (fld_type, 1, pCh, 
						  dmEventHandler,
						  (void *)pdatamove) ) != ECA_NORMAL)
	    {
		sprintf (errmsg, "datamove, get_callback, %s for %s (%s)",
				  pdatamove -> inam, pdatamove -> name,
				  ca_message (status) );
		return;
	    }
	}
	pdatamove -> stat = NO_ALARM;
	pdatamove -> sevr = NO_ALARM;
	/* if there is an output channel and it is not connected */
	if (pdatamove -> ochd &&
	   (ca_state ((chid)pdatamove -> ochd) != cs_conn) )
	{
	    pdatamove -> stat = WRITE_ALARM;
	    pdatamove -> sevr = VALID_ALARM;
	}
    }
    /* if we have lost the input channel */
    else if (connArgs.op == CA_OP_CONN_DOWN)
    {
	pdatamove -> stat = READ_ALARM;
	pdatamove -> sevr = VALID_ALARM;
    }

    if ( ((stat != pdatamove -> stat) ||
	  (sevr != pdatamove -> sevr)) &&
	 master)
    {
	db_post_events ((struct dbCommon *)pdatamove, (caddr_t)&pdatamove -> stat, DBE_VALUE);
	db_post_events ((struct dbCommon *)pdatamove, (caddr_t)&pdatamove -> sevr, DBE_VALUE);
    }

    return;

} /* end func dmInputConnectHandler () */




/****************************************************************************
 *
 *  MODULE NAME            : dmOutputConnectHandler
 *  AUTHOR                 : emsteel
 *  DATE                   : 16 Sep 96
 *
 *  DESCRIPTION: 
 *		called when connect happens for output channel
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : 
 *  REVISION DATE          :
 *  REVISION:
 *
****************************************************************************/

static void dmOutputConnectHandler (struct connection_handler_args connArgs)
{
    struct datamoveRecord *pdatamove;
    chid	pCh;
    unsigned short	master = 0;
    unsigned short	stat;
    unsigned short	sevr;

    pCh = connArgs.chid;
    pdatamove = (struct datamoveRecord *)ca_puser (pCh);
    stat = pdatamove -> stat;
    sevr = pdatamove -> sevr;

    resGetCaAnswer (pdatamove -> radr, &master);

    if (connArgs.op == CA_OP_CONN_UP)
    {
	if (pdatamove -> stat == WRITE_ALARM)
	{
	    pdatamove -> stat = NO_ALARM;
	    pdatamove -> sevr = NO_ALARM;
	}
    }
    else if (connArgs.op == CA_OP_CONN_DOWN)
    {
	if (pdatamove -> stat != READ_ALARM)
	{
	    pdatamove -> stat = WRITE_ALARM;
	    pdatamove -> sevr = VALID_ALARM;
	}
    }
    if ( ((stat != pdatamove -> stat) ||
	  (sevr != pdatamove -> sevr) ) &&
	 master)
    {
	db_post_events ((struct dbCommon *)pdatamove, (caddr_t)&pdatamove -> stat, DBE_VALUE);
	db_post_events ((struct dbCommon *)pdatamove, (caddr_t)&pdatamove -> sevr, DBE_VALUE);
    }

    return;

} /* end func dmOutputConnectHandler () */

