/*recBo.c,v 1.1.1.1 1999/08/03 14:44:18 tis4000 Exp */

/* recBo.c - Record Support Routines for Binary Output records */
/*
 *      Original Author: Bob Dalesio
 *      Current Author:  Marty Kraimer
 *      Date:            7-14-89
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  12-9-88         lrd     lock the record during processing
 * .02  12-15-88        lrd     process the forward scan link
 * .03  12-23-88        lrd     Alarm on locked MAX_LOCKED times
 * .04  01-13-89        lrd     delete db_read_bo and db_write_bo
 * .05  01-20-89        lrd     fixed vx includes
 * .06  03-29-89        lrd     make hardware errors MAJOR
 *                              remove hw severity spec from database
 *                              add continuous control
 * .07  04-06-89        lrd     add monitor detection
 * .08  05-03-89        lrd     removed process mask from arg list
 * .09  08-16-89        lrd     add ability to do softchannel momentary
 * .10  08-17-89        lrd     add soft channel support
 * .11  01-31-90        lrd     add the plc_flag arg to the ab_bodriver call
 * .12  03-21-90        lrd     add db_post_events for RVAL and RBV
 * .13  04-02-90        ba/lrd  add monitor handling for momentary outputs
 *                              remove rbv arg to ab_bodriver
 * .14  04-05-90        lrd     make momentary output handling a task
 *                              as the watchdog runs at interrupt level
 * .15  04-11-90        lrd     make locals static
 * .16  05-02-90        lrd     fix initial value set in the DOL field
 * .17  10-10-90	mrk	Changes for record and device support
 * .18  03-12-93	ems	deleted ack from redundancy
 * .19  01-03-94	ems	ack = 0 in init_record
 *				in process, if not init'd do nothing, also
 *				  check for init before write not needed
 * .20  01-05-94	ems	added special func to handle SPC_CNST
 * .21  03-01-94	ems	added support for process stats
 * .22	03-29-94	ems	make forward func def agree with actual func def
 * .23	04-21-94	ems	added INI2 to namelist, chg NUMFLDS to NUMBOFLDS
 *				rewrite init_record
 * .24  06-10-94	ems	move call to tsLocalTime before call to device
 *				if on scan and !pact
 * .25	06-13-94	ems	added supprt for maximize severity
 * .26	08-18-94	ems	monitor(), remove DBE_LOG form ack
 *				only set DBE_LOG on val if tenb (trending) is enabled
 *				only set DBE_VALUE on rval and rbv
 * .27	10-12-94	ems	add TIME to redundancy, fpas chg'd to frps, added fpps
 *				(1st process pass), do time stamp when stat or sevr chg,
 *				see monitor(),and not in process(),set lalm to val when 
 *				stat or sevr chg,in process() if fpps (first process)
 *				init lalm to val and if time stamp is 0, init it also
 *				added lval to test for COS alarm instead of lalm which
 *				is now used as above
 * .28	10-20-94	ems	chg time to latm (last alarm time) added 
 *				added lptm (last process time)
 * .29	01-13-95	wep	add mlst to redundancy, init_record(), init lptm and latm
 * .30	01-18-95	ems	monitor(), bug, when mlst != val, 
 *				monitor_mask |= DBE_VALUE not = DBE_VALUE, wipes out 
 *				the DBE_ALARM bit set above
 * .31	03-15-95	ems	monitor(), do not set ACK if sevr is EVENT_ALARM
 * .32	09-11-95	ems	monitor(), archive when stat/sevr chg's also
 * .33	09-12-95	ems	monitor(), see mod for handling VALID_ALARM
 * .34	01-31-96	ems	added support for output on chg only OCHG
 * .35	02-05-96	ems	fixed momemtayr output to work not only in SUPERVISORY
 *				but also in CLOSED_LOOP and redundantly, deleted RBV and
 *				ORBV, no longer used
 * .36	04-11-96	ems	added code to support alarm delay
 * .37	06-19-96	ems	added code for alarm disable ADIS
 */

#include	<vxWorks.h>
#include	<inetLib.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<lstLib.h>
#include	<strLib.h>
#include	<wdLib.h>
#include	<memLib.h>
#include	<logLib.h>
#include	<taskLib.h>
#include	<string.h>

#include	"alarm.h"
#include	"dbDefs.h"
#include	"dbCommon.h"
#include	"link.h"
#include	"db_field_log.h"
#include	"dbAccess.h"
#include	"dbFldTypes.h"
#include	"dbRecType.h"
#include	"devSup.h"
#include	"errMdef.h"
#include	"recSup.h"
#include	"special.h"
#include	"rec/boRecord.h"
#include	"redflds.h"
#include	"redque.h"
#include	"boredfldidx.h"
#include	"restable.h"
#include	"callbacks.h"
#include	"db_addr.h"
#include	"recGbl.h"
#include	"dbEvent.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table*/

#define report NULL
#define initialize NULL
static long init_record ();
static long process ();
static long special ();
static long get_value ();
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
#define get_units NULL
#define get_precision NULL
static long get_enum_str ();
static long get_enum_strs ();
static long put_enum_str ();
static long get_graphic_double ();
static long get_control_double ();
#define get_alarm_double NULL

struct rset boRSET = {
	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double
	};

struct bodset {		/* binary output dset */
	long	     number;
	DEVSUPFUN    dev_report;
	DEVSUPFUN    init;
	DEVSUPFUN    init_record;    /*returns:(0,2)=>(success,success no convert*/
	DEVSUPFUN    get_ioint_info;
	DEVSUPFUN    write_bo;       /*returns:(-1,0,1) => 
				      (failure, success, don't continue)*/
	DEVSUPFUN    readback_bo;
	};

/* The following definitions are required for redundancy */

static struct	fldDes	**mylist = 0;

static char	*namelist [] = {
		"STAT",
		"SEVR",
		"VAL",
		"LALM",
		"INI2",
		"LATM",
		"LVAL",
		"MLST",
		"LOVL",
		"CBFL",
		"LIVL",
		"OVAL",
		"HHLL",
		"RTCT",
		0
		};

static	unsigned	short	type;

static void boCallback ();
static void alarm ();
static void monitor ();
static void test_ochg ();
static void boReadback ();

/* defines for the callback flag CBFL 
   NOTE: if you change the meaning of these
	 defines, check .../share/src/redund/freeMemory.c
*/
#define	CALLBACK_NOT_USED	0
#define	CALLBACK_NOT_ACTIVE	1
#define	CALLBACK_ACTIVE		2
#define	CALLBACK_COMPLETED	3



/*------------------------------------------------------------------------*/

static long init_record(pbo)

    struct	boRecord	*pbo;
{
    struct	bodset		*pdset;
    long 			status = 0;
    CALLBACK_PTR		pcallback;
    unsigned	short		i;
    char			*pstr;
    char			redbuf [256];
    int				ini2_flag = 0;
    int				do_convert;

	if (!(mylist)) 
	{
	    for (type = 1; type < 100; type++) 
	    {
		pstr = GET_PRECTYPE (type);
		if (strcmp (pstr, "bo") == 0) 
		    break;
	    }
	    mylist = redFldAlloc (type, NUMBOFLDS);

	    if (!mylist)
	    {
		recGblRecordError (S_rec_Alloc, (struct dbCommon *)pbo,
				   "bo, init_record, mylist failed");
		return (S_rec_Alloc);
	    }

	    for (i = 0; i < NUMBOFLDS; i++) 
		redFldAdd (type, i, namelist [i]);
	}

        if (!(pdset = (struct bodset *) (pbo->dset))) 
	{
	    recGblRecordError (S_dev_noDSET, (struct dbCommon *)pbo, "bo: init_record");
	    return (S_dev_noDSET);
        }

        /* must have  write_bo functions defined */

        if ( (pdset->number < 5) || (pdset->write_bo == NULL) ) 
	{
	    recGblRecordError (S_dev_missingSup, (struct dbCommon *)pbo, "bo: init_record");
	    return (S_dev_missingSup);
        }

	if (!(pbo -> dpvt)) 
	{
	    pcallback = (CALLBACK_PTR) (calloc (1, sizeof (CALLBACK)));
	    if (!pcallback)
	    {
		recGblRecordError (S_rec_Alloc, (struct dbCommon *)pbo,
				   "bo, init_record, pcallback failed");
		return (S_rec_Alloc);
	    }
	    pbo->dpvt = (caddr_t) pcallback;
	    pcallback->callback = boCallback;
	    pcallback->priority = priorityMedium;
	    if (dbNameToAddr (pbo->name, &(pcallback->dbAddr))) 
	    {
	        logMsg("dbNameToAddr failed in init_record for recBo\n", 0,0,0,0,0,0);
	        exit(1);
	    }
	}

	pbo -> init = TRUE;
	pbo -> fpps = 1;
	pbo -> nsta = NO_ALARM;
	pbo -> nsev = NO_ALARM;
	pbo -> msev = NO_ALARM;
	pbo -> cbfl = CALLBACK_NOT_USED;

	/* init alarm delay variables */
	pbo -> larf = 0;
	pbo -> hhll = 0;
	pbo -> hilo = 0;
	pbo -> dvct = 0;
	pbo -> rtct = 0;

	/* init readback variables */
	pbo -> rbfn = boReadback;

	/* init last process time and last alarm time */
	if (pbo -> lptm.secPastEpoch == 0)
	    tsLocalTime (&pbo -> lptm);
	if (pbo -> latm.secPastEpoch == 0)
	    tsLocalTime (&pbo -> latm);

	do_convert = 1;

	/* do not allow conversion if the board type is CONSTANT and output
	   is CONSTANT or REFERENCE, this was done in devSoft.c
	*/

        if (pbo -> dtyp <= 1)
	{
	    switch (pbo -> out.type)
	    {
		case (CONSTANT):
		case (PV_LINK):
		case (DB_LINK):
		case (CA_LINK):
		    do_convert = 0;
		    break;
	    }
        }

        if( pdset->init_record ) 
	{
	    status = (*pdset->init_record) (pbo, process);
	    if (status)
	    {
		pbo -> nsta = NO_INIT_ALARM;
		pbo -> nsev = VALID_ALARM;
	    }
	    else
	    {
		ini2_flag = 1;
		if (do_convert)
		{
		    if(pbo->rval == 0) 
		    {
		        if (pbo -> invt)
		    	    pbo -> val = 1;
		        else
			    pbo->val = 0;
		    } 
		    else 
		    {
		        if (pbo -> invt)
			    pbo -> val = 0;
		        else
			    pbo->val = 1;
		    }
		}
		pbo->udf = FALSE;
	    }
        }
	else
	    ini2_flag = 1;
	
	/* init last value, use by alarm() to test COS alarm */

	pbo -> oval = pbo -> livl = pbo -> lovl = pbo -> lval = pbo -> val;

	if (resGetCaAnswer(pbo -> radr, &i) == RES_OK)
	{
	    if (i)
	    {
		/* post STAT, SEVR and VAL (monitor), if we will finally init, 
		   because we have read a value from the input device, set INI2,
		   send all redund fields to slave, when slave receives INI2, 
		   it will be fully init'd also
		*/

		alarm (pbo);

	    	/* check for maximize severity */

	    	if (pbo -> msev > pbo -> nsev)
	    	{
		    pbo -> nsev = pbo -> msev;
		    pbo -> nsta = LINK_ALARM;
	    	}

		monitor (pbo);

		if (ini2_flag)
		{
		    pbo -> ini2 = 1;

		    redQInit ((struct dbCommon *)pbo, redbuf, sizeof (redbuf), type);
		    redQField (redbuf, mylist [STAT]);
		    redQField (redbuf, mylist [SEVR]);
		    redQField (redbuf, mylist [VAL]);
		    redQField (redbuf, mylist [LALM]);
		    redQField (redbuf, mylist [INI2]);
		    redQField (redbuf, mylist [LVAL]);
		    redQField (redbuf, mylist [LOVL]);
		    redQField (redbuf, mylist [CBFL]);
		    redQField (redbuf, mylist [LIVL]);
		    redQField (redbuf, mylist [OVAL]);
		    redQSend (redbuf);
		}
	    }
	}

        return (status);
}


/*------------------------------------------------------------------------*/

static void alarm (pbo)
    struct boRecord	*pbo;
{
	unsigned short	val = pbo->val;
	int		adis = 0;

	adis = AlarmDisableLink ((void *)pbo);
        if ( (pbo->aenb == ALARM_DISABLED) || (adis != 0) )
	{ 
	    if (pbo -> nsta != NO_INIT_ALARM)
	    {
                pbo->nsta = INHIBIT_ALARM;
                pbo->nsev = NO_ALARM;
	    }
            return;
        }

        /* check for udf alarm */
        if(pbo->udf == TRUE )
	{
            if (pbo->nsev < VALID_ALARM) 
	    {
                pbo->nsta = UDF_ALARM;
                pbo->nsev = VALID_ALARM;
            }
        }
	
	if (pbo -> nsev == VALID_ALARM)
		return;

        /* check for  state alarm */

        if (val == 0) 
	{
            if (pbo->nsev < pbo->zsv) 
	    {
                pbo->nsta = STATE_ALARM;
                pbo->nsev = pbo->zsv;
            }
        } 
	else 
	{
            if (pbo->nsev < pbo->osv)
	    {
                pbo->nsta = STATE_ALARM;
                pbo->nsev = pbo->osv;
            }
        }

	/* check if time to output STATE alarm */

	AlarmStates ((void *)pbo, 0);

        /* check for cos alarm */

	if (val != pbo -> lval)
	{
	    if (pbo -> nsev < pbo -> cosv)
	    {
		pbo -> nsta = COS_ALARM;
		pbo -> nsev = pbo -> cosv;
	    }
	}
	pbo -> lval = val;

	/* check for return to normal status */

	AlarmStates ((void *)pbo, 1);

        return;
}


/*------------------------------------------------------------------------*/

static void boCallback(pcallback)
    CALLBACK_PTR	pcallback;
{
    struct boRecord	*pbo = (struct boRecord *)(pcallback->dbAddr.precord);
    short		master;

    if (resGetCaAnswer (pbo -> radr, &master) == RES_OK)
    {
	if (master)
	{
	    dbScanLock((caddr_t)pbo);
	    pbo -> cbfl = CALLBACK_COMPLETED;
	    (void)process(&(pcallback->dbAddr));
	    dbScanUnlock((caddr_t)pbo); 
	}
	/* my resource is SLAVE */
	else
	{
	    if (pbo -> cbfl == CALLBACK_ACTIVE)
		pbo -> cbfl = CALLBACK_COMPLETED;
	}
    }
}


/*------------------------------------------------------------------------*/

static long get_value(pbo,pvdes)
    struct boRecord		*pbo;
    struct valueDes	*pvdes;
{
    pvdes->field_type = DBF_ENUM;
    pvdes->no_elements=1;
    (unsigned short *)(pvdes->pvalue) = &pbo->val;
    return(0);
}

/*------------------------------------------------------------------------*/

static long get_enum_str(paddr,pstring)
    struct dbAddr *paddr;
    char	  *pstring;
{
    struct boRecord	*pbo=(struct boRecord *)paddr->precord;
    unsigned short      *field = (unsigned short *)paddr -> pfield;

    if((*field) == 0)
    {
	strncpy(pstring,pbo->znam,sizeof(pbo->znam));
	pstring[sizeof(pbo->znam)] = 0;
    } 
    else if((*field) == 1)
    {
	strncpy(pstring,pbo->onam,sizeof(pbo->onam));
	pstring[sizeof(pbo->onam)] = 0;
    } 
    else 
    {
	strcpy(pstring,"Illegal_Value");
    }
    return(0);
}

/*------------------------------------------------------------------------*/

static long get_enum_strs(paddr,pes)
    struct dbAddr *paddr;
    struct dbr_enumStrs *pes;
{
    struct boRecord	*pbo=(struct boRecord *)paddr->precord;

    pes->no_str = 2;
    bzero((char *)pes->strs[0],sizeof(pes->strs));
    strncpy(pes->strs[0],pbo->znam,sizeof(pbo->znam));
    strncpy(pes->strs[1],pbo->onam,sizeof(pbo->onam));
    return(0);
}


/*------------------------------------------------------------------------*/

static long get_graphic_double (paddr, pgd)

    struct	dbAddr		*paddr;
    struct	dbr_grDouble	*pgd;
{
    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {

	case ( (int)&(((struct boRecord *)0) -> rbdt) ):
	    pgd -> upper_disp_limit = 9999;
	    pgd -> lower_disp_limit = 0;
	    break;

	default:
	    pgd -> upper_disp_limit = 2;
	    pgd -> lower_disp_limit = 0;
	    break;
    }

    return (0);
}


/*------------------------------------------------------------------------*/

static long get_control_double (paddr, pcd)

    struct	dbAddr		*paddr;
    struct	dbr_ctrlDouble	*pcd;
{
    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct boRecord *)0) -> rbdt) ):
	    pcd -> upper_ctrl_limit = 9999;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	default:
	    pcd -> upper_ctrl_limit = 2;
	    pcd -> lower_ctrl_limit = 0;
	    break;
    }

    return (0);
}


/*------------------------------------------------------------------------*/

static void monitor (pbo)
    struct boRecord	*pbo;
{
	unsigned short	monitor_mask;
        short           stat,sevr,nsta,nsev;
	int		set_ack = 1;

        /* get previous stat and sevr  and new stat and sevr*/

        stat = pbo->stat;
        sevr = pbo->sevr;
        nsta = pbo->nsta;
        nsev = pbo->nsev;

        /*set current stat and sevr*/

        pbo->stat = nsta;
        pbo->sevr = nsev;
        pbo->nsta = 0;
        pbo->nsev = 0;

	/* if we are coming out of VALID_ALARM and our new stat/sevr are
	   the same as when we went into VALID_ALARM, do not set ack
	*/
	if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	     (nsta == pbo -> psta) && (nsev == pbo -> psev) )
	{
	    set_ack = 0;
	}
	/* if we are going into VALID_ALARM, and out previous stat was not
	   VALID_ALARM, remember previous non VALID_ALARM stat/sevr
	   and we do not want to set ACK,
	*/
	if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
	{
	    set_ack = 0;
	    pbo -> psta = stat;
	    pbo -> psev = sevr;
	    db_post_events ((struct dbCommon *)pbo, (caddr_t)&pbo -> psta, DBE_VALUE);
	    db_post_events ((struct dbCommon *)pbo, (caddr_t)&pbo -> psev, DBE_VALUE);
	}

        monitor_mask = 0;

	/*
	Extra Patch: If the user defines his output to have
	the same alarm severity for both states and the value
	is monitored with only the DBE_ALARM mask, then he'll
	never see that state change. This patch will avoid
	that problem.

	Extra Extra Patch: If we have a passive BO with only
	COS alarming, then we will get the first alarm on the
	first state change and then no others because stat and
	sevr never change even though the value has.
	*/

	if (((nsta == STATE_ALARM) || (pbo -> nsta == COS_ALARM)) &&
	    (pbo -> mlst != pbo -> val) )
		stat = ~nsta;

        /* alarm condition changed this scan */

        if ((stat != nsta) || (sevr != nsev))
	{
            /* post events for alarm condition change*/

            monitor_mask = DBE_ALARM;

	    /* when alarm status chg's, get current time and
	       save VAL in lalm (last value when alarm chg'd)
	    */
	    tsLocalTime (&pbo -> latm);
	    pbo -> lalm = pbo -> val;

            /* post stat and sevr fields */

            db_post_events ((struct dbCommon *)pbo, (caddr_t)&pbo->stat, DBE_VALUE);
            db_post_events ((struct dbCommon *)pbo, (caddr_t)&pbo->sevr, DBE_VALUE);
            if ((pbo->sevr != NO_ALARM)  && (set_ack) &&
		(pbo->sevr != EVENT_ALARM) ) 
	    {
                if (pbo->ack == 0)
		{
                    pbo->ack = 1;           /* new unacknowledged alarm */
                    db_post_events ((struct dbCommon *)pbo, (caddr_t)&pbo->ack, DBE_VALUE);
                }
            }
        }
        /* check for value change */
        if (pbo->mlst != pbo->val)
	{
            /* post events for value change and archive change */
            monitor_mask |= DBE_VALUE;
        }

	if (pbo -> tenb)
	{
	    /* if VAL has chagned or there has been a change in alarm status */
	    if ( (pbo -> mlst != pbo -> val) || (monitor_mask & DBE_ALARM) )
		monitor_mask |= DBE_LOG;
	}

	if (monitor_mask & (DBE_VALUE | DBE_LOG))
	    pbo -> mlst = pbo -> val;

        /* send out monitors connected to the value field */
        if (monitor_mask)
	{
            db_post_events((struct dbCommon *)pbo, (caddr_t)&pbo->val,monitor_mask);
        }
	if(pbo->oraw!=pbo->rval) 
	{
	    db_post_events((struct dbCommon *)pbo, (caddr_t)&pbo->rval,DBE_VALUE);
	    pbo->oraw = pbo->rval;
	}
        return;
}


/*------------------------------------------------------------------------*/

static long process (paddr)

    struct dbAddr	*paddr;
{
    struct boRecord	*pbo   = (struct boRecord *) (paddr->precord);
    struct bodset	*pdset = (struct bodset *) (pbo->dset);
    CALLBACK_PTR	pcallback;
    long		status = 0;
    unsigned short	stat   = pbo -> stat;
    unsigned short	sevr   = pbo -> sevr;
    unsigned short	val    = pbo -> val;
    unsigned short	lalm   = pbo -> lalm;
    unsigned short	lval   = pbo -> lval;
    TS_STAMP		latm   = pbo -> latm;
    double		mlst   = pbo -> mlst;
    unsigned short	lovl   = pbo -> lovl;
    short		cbfl   = pbo -> cbfl;
    unsigned short	livl   = pbo -> livl;
    unsigned short      oval   = pbo -> oval;
    short		hhll   = pbo -> hhll;
    short		rtct   = pbo -> rtct;
    unsigned short	ival;
    int			wait_time;
    int			write_output = 1;
    short		icbf = CALLBACK_NOT_USED;
    char		redbuf [256];

	   /* if not init'd, do nothing */

	if ((!pbo -> init) || (!pbo -> ini2))
	    return (status);

	if ((pdset == NULL) || (pdset->write_bo == NULL)) 
	{
	    pbo->pact = TRUE;
	    recGblRecordError (S_dev_missingSup, (struct dbCommon *)pbo, "write_bo");
	    return (S_dev_missingSup);
	}
	/* get time stamp for last time record was processed */

	tsLocalTime (&pbo -> lptm);

	pbo -> msev = NO_ALARM;

	if (pbo -> tagd)
	{
	    pbo -> nsta = TAGGED_ALARM;
	    pbo -> nsev = VALID_ALARM;
	}
        else if (pbo->oscn == OFF_SCAN) 
	{
            pbo -> nsta = SCAN_ALARM;
            pbo -> nsev = VALID_ALARM;
	}
	else 
	{
	    pbo -> nsta = NO_ALARM;
	    pbo -> nsev = NO_ALARM;

            if (!pbo->pact) 
	    {
	        /* update number of process times */
	        if (process_stats)
	           process_cnts[type]++;	

		if ((pbo->dol.type == DB_LINK) && (pbo->omsl == CLOSED_LOOP))
		{
		    long options = 0;
		    long nRequest = 1;

		    pbo->pact = TRUE;
		    status = dbGetLink (&pbo->dol.value.db_link, 
					(struct dbCommon *)pbo,
					DBR_USHORT, (caddr_t)&ival, 
					&options, &nRequest);
		    pbo->pact = FALSE;
		    if (status == 0) 
		    {
			pbo->val = ival;
			pbo->udf = FALSE;
		    }
		    else 
		    {
			if (pbo->nsev < VALID_ALARM) 
			{
			    pbo->nsev = VALID_ALARM;
			    pbo->nsta = LINK_ALARM;
			}
		    }
		}
		if (pbo -> high > 0)
		{
		    if (pbo -> val == 0)
		    {
			pbo -> oval = pbo -> val;
			icbf = CALLBACK_NOT_ACTIVE;
		    }
		    else	/* val != 0 */
		    {
			if (pbo -> livl == 0)
			{
			    pbo -> oval = pbo -> val;
			    icbf = CALLBACK_ACTIVE;
			}
			else	/* livl != 0 */
			{
			    if (pbo -> oval == 1)
			    {
				/* callback completed, turn off output */
				if (pbo -> cbfl == CALLBACK_COMPLETED)
				{
				    if ( (pbo -> dol.type == DB_LINK) && 
					 (pbo -> omsl == CLOSED_LOOP) )
				    {
					pbo -> oval = 0;
					icbf = CALLBACK_NOT_ACTIVE;
				    }
				    else
				    {
					pbo -> val = pbo -> oval = 0;
					icbf = CALLBACK_NOT_ACTIVE;
				    }
				}
				else
				{
				    icbf = pbo -> cbfl;
				}
			    }
			    else	/* oval != 0 */
			    {
			        icbf = CALLBACK_NOT_ACTIVE;
			    }
			}
		    }
		}
		else
		{
		    pbo -> oval = pbo -> val;
		    pbo -> cbfl = CALLBACK_NOT_USED;
		    icbf = CALLBACK_NOT_USED;
		}
		pbo -> livl = pbo -> val;

		if ( pbo->mask != 0 ) 
		{
		    if (pbo->oval == 0) 
		    {
			if (pbo -> invt)
			    pbo -> rval = pbo -> mask;
			else
			    pbo->rval = 0;
		    } 
		    else 
		    {
			if (pbo -> invt)
			    pbo -> rval = 0;
			else
			    pbo->rval = pbo->mask;
		    }
		}
	    } /* end !pact */

	    if (status == 0) 
	    {
		test_ochg (pbo, &write_output);
		if (write_output)
		{
		    status = (*pdset->write_bo) (pbo); /* write the new value */
		    pbo -> lovl = pbo -> oval;
		}
 	    }
	    pbo->pact = TRUE;

	    /* status is one if an asynchronous record is being processed*/

	    if(status == 1) 
		return(0);
	    
	    if (status == -2) 
	    {
		pbo->pact = FALSE;
		return (0);
	    }

	    /* test if need to start a momentary callback */
	    if (pbo -> high > 0)
	    {
		switch (icbf)
		{
		    case CALLBACK_NOT_USED:
			pbo -> cbfl = CALLBACK_NOT_USED;
			break;

		    case CALLBACK_NOT_ACTIVE:
			if (pbo -> cbfl == CALLBACK_ACTIVE)
			{
			    pcallback = (CALLBACK_PTR) (pbo -> dpvt);
			    wdCancel (pcallback -> wd_id);
			}
			pbo -> cbfl = CALLBACK_NOT_ACTIVE;
			break;

		    case CALLBACK_ACTIVE:
			if (pbo -> cbfl == CALLBACK_ACTIVE)
			    break;
			wait_time = (int) ((pbo->high) * vxTicksPerSecond);
			pcallback = (CALLBACK_PTR) (pbo -> dpvt);
			if (pcallback->wd_id == NULL) 
			{
			    pcallback->wd_id = wdCreate ();
			}
			wdStart (pcallback->wd_id, wait_time, (FUNCPTR)callbackRequest, (int)pcallback);
			pbo -> cbfl = CALLBACK_ACTIVE;
			break;
		    default:
			pbo -> cbfl = CALLBACK_NOT_USED;
			break;

		}
	
	    }

	    /* check for alarms */

	    alarm(pbo);

	    /* check for maximize severity */

	    if (pbo -> msev > pbo -> nsev)
	    {
		pbo -> nsev = pbo -> msev;
		pbo -> nsta = LINK_ALARM;
	    }
	}

	/* if the time stamp is 0, init and post val with
	   alarm bit set, this makes sure that proper alarm
	   messages are set if IOC was down and alarm status
	   may have chg'd
	*/
	if (pbo -> latm.secPastEpoch == 0)
	{
	    tsLocalTime (&pbo -> latm);
	    db_post_events ((struct dbCommon *)pbo, (char *)&pbo -> val, DBE_VALUE | DBE_ALARM);
	}

	/* if it is the 1st process pass 
	   init last value when alarm chg'd (lalm)
	*/

	if (pbo -> fpps)
	{
	    pbo -> fpps = 0;
	    pbo -> lalm = pbo -> val;
	}

	/* check event list */

	monitor(pbo);

	redQInit ((struct dbCommon *)pbo, redbuf, sizeof (redbuf), type);

	if (stat != pbo->stat) 
	    redQField (redbuf, mylist [STAT]);
	
	if (sevr != pbo->sevr) 
	    redQField (redbuf, mylist [SEVR]);
	
	if (val != pbo->val) 
	    redQField (redbuf, mylist [VAL]);
	
	if (lalm != pbo->lalm) 
	    redQField (redbuf, mylist [LALM]);
	
	if (TsCmpStampsNE (&latm, &pbo -> latm) )
	    redQField (redbuf, mylist [LATM]);

	if (lval != pbo -> lval)
	    redQField (redbuf, mylist [LVAL]);
	
	if (mlst != pbo -> mlst)
	    redQField (redbuf, mylist [MLST]);
	
	if (lovl != pbo -> lovl)
	    redQField (redbuf, mylist [LOVL]);
	
	if (cbfl != pbo -> cbfl)
	    redQField (redbuf, mylist [CBFL]);
	
	if (livl != pbo -> livl)
	    redQField (redbuf, mylist [LIVL]);
	
	if (oval != pbo -> oval)
	    redQField (redbuf, mylist [OVAL]);
	
	if (hhll != pbo -> hhll)
	    redQField (redbuf, mylist [HHLL]);
	
	if (rtct != pbo -> rtct)
	    redQField (redbuf, mylist [RTCT]);
	
	redQSend (redbuf);

	/* process the forward scan link record */

	if (pbo->flnk.type == DB_LINK)  
	{
	    dbScanPassive ((struct dbAddr *)pbo->flnk.value.db_link.pdbAddr);
	}
	pbo->pact = FALSE;
	return (status);
}


/*------------------------------------------------------------------------*/

static long put_enum_str(paddr,pstring)
    struct dbAddr *paddr;
    char          *pstring;
{
    struct boRecord     *pbo=(struct boRecord *)paddr->precord;
    unsigned short      *field = (unsigned short *)paddr -> pfield;

    if(strncmp(pstring,pbo->znam,sizeof(pbo->znam))==0)
	*field = 0;
    else  if(strncmp(pstring,pbo->onam,sizeof(pbo->onam))==0) 
	*field = 1;
    else 
	return(S_db_badChoice);
    return(0);
}


/*------------------------------------------------------------------------*/

static long special (paddr,after)
    struct dbAddr *paddr;
    int           after;
{
    struct boRecord     *pbo = (struct boRecord *)(paddr->precord);
    CALLBACK_PTR	pcallback;
    int                 special_type = paddr->special;
    int			wait_time;
    short		master;
    static unsigned short save_cnst;
    static unsigned short save_cbfl;

    /* if dtyp and out are not CONSTANT, do not let them
       update CNST
    */

    if(!after) 
    {
	save_cnst = pbo -> cnst;
	save_cbfl = pbo -> cbfl;
	return(0);
    }

    switch(special_type) 
    {

    	case (SPC_CNST):
	    if ((pbo -> dtyp <= 1) && (pbo -> out.type == CONSTANT))
            {
	        pbo -> out.value.value = pbo -> cnst;
                pbo -> rval = pbo -> val = pbo -> cnst;
            }
	    else
	        pbo -> cnst = save_cnst;
            break;

	case SPC_BO_VAL:
	    /* if output chg is set and I am in SUPERVISORY mode 
	       (NOTE: dol!=link and omsl=closed_loop acts like supervisory), and 
	       my output channel is not a reference, and my resource is MASTER , 
	       set LOVL to this value, this will ensure that the VAL the user 
	       input will be output
	    */
	    if ( (pbo -> ochg == 0) || 
		 ( (pbo -> dol.type == DB_LINK) && (pbo -> omsl == CLOSED_LOOP) ) ||
		 ( (pbo -> dtyp <= 1) && (pbo -> out.type == DB_LINK) ) )
		break;
	    if (resGetCaAnswer (pbo -> radr, &master) == RES_OK)
	    {
		if (master)
		{
		    pbo -> lovl = 2;
		}
	    }
	    break;

	case SPC_BO_CBFL:

	    /* SLAVE must be init'd before we do this */
	    if ( (!pbo -> init) || (!pbo -> ini2) )
		break;

	    /* the callback flag has been received by the SLAVE from the
	       MASTER, stop/start a callback
	    */
	    pcallback = (CALLBACK_PTR)(pbo -> dpvt);

	    /* a callback was active and we got another active,
	       cancel the curretn callback and start a new one
	    */
	    if ( (pbo -> cbfl == CALLBACK_ACTIVE) && 
		 (save_cbfl == CALLBACK_ACTIVE) )
	    {
		wdCancel (pcallback -> wd_id);
		wait_time = (int) ((pbo->high) * vxTicksPerSecond);
		if (pcallback->wd_id == NULL) 
		{
		    pcallback->wd_id = wdCreate ();
		}
		wdStart (pcallback->wd_id, wait_time, (FUNCPTR)callbackRequest, (int)pcallback);
	    }
	    /* new active callback on MASTER, so start on here
	       on the SLAVE 
	    */
	    else if (pbo -> cbfl == CALLBACK_ACTIVE)
	    {
		wait_time = (int) ((pbo->high) * vxTicksPerSecond);
		if (pcallback->wd_id == NULL) 
		{
		    pcallback->wd_id = wdCreate ();
		}
		wdStart (pcallback->wd_id, wait_time, (FUNCPTR)callbackRequest, (int)pcallback);
	    }
	    /* we had an active callbeack on the SLAVE and now we
	       don't so cancel it
	    */
	    else if (save_cbfl == CALLBACK_ACTIVE)
	    {
		wdCancel (pcallback -> wd_id);
	    }
	    break;

        default:
            recGblDbaddrError(S_db_badChoice,paddr,"bo: special");
            return(S_db_badChoice);
    }

    return (0);
}


/*------------------------------------------------------------------------*/

static void test_ochg (pbo, write_output)
    struct boRecord	*pbo;
    int			*write_output;
{
    long		status;
    long		options = 0;
    long		nRequests = 1;
    unsigned short	lovl;

    if (pbo -> ochg == 0)
	return;

    /* if my output is a reference, then get its value, and set lovl
       (last val output) to the refs value. this protects us from the
       case where the output may be a field that has been changed
       from a graphic
    */
    if ( (pbo -> dtyp <= 1) && (pbo -> out.type == DB_LINK) )
    {
	status = dbGetLink (&pbo -> out.value.db_link, 
			    (struct dbCommon *)pbo, 
			    DBR_ENUM, (caddr_t)&lovl, 
			    &options, &nRequests);
	if (status)
	{
	    *write_output = 0;
	    pbo -> nsev = VALID_ALARM;
	    pbo -> nsta = LINK_ALARM;
	    return;
	}
	if (pbo -> invt)
	{
	    if (lovl)
		lovl = 0;
	    else
		lovl = 1;
	}
	else
	{
	    if (lovl)
		lovl = 1;
	    else
		lovl = 0;
	}
	pbo -> lovl = lovl;
    }
    if (pbo -> lovl == pbo -> oval)
	*write_output = 0;
    return;
}


/*------------------------------------------------------------------------*/

static void	boReadback (struct boRecord *pbo)
{
    struct bodset	*pdset;
    unsigned long	rval = pbo -> rval;
    long		status = 0;

    pdset = (struct bodset *)pbo -> dset;

    /* no use if we don't have the ability to readback the output */
    if (pdset -> readback_bo == NULL)
	return;

    /* if the bo is a momentary, then the callback will
       write to the output
    */
    if (pbo -> cbfl == CALLBACK_ACTIVE)
	return;

    status = (*pdset ->readback_bo)(pbo);
    if (status == 0)
    {
	/* if there was a change */
	if (rval != pbo -> rval)
	{
	    unsigned short	stat   = pbo -> stat;
	    unsigned short	sevr   = pbo -> sevr;
	    unsigned short	val    = pbo -> val;
	    unsigned short	lalm   = pbo -> lalm;
	    unsigned short	lval   = pbo -> lval;
	    TS_STAMP		latm   = pbo -> latm;
	    double		mlst   = pbo -> mlst;
	    unsigned short	livl   = pbo -> livl;
	    unsigned short      oval   = pbo -> oval;
	    short		hhll   = pbo -> hhll;
	    short		rtct   = pbo -> rtct;
	    char		redbuf [256];

	    /* update last process time */
	    tsLocalTime (&pbo -> lptm);
	    if (pbo -> rval == 0)
	    {
		if (pbo -> invt)
		    pbo -> val = 1;
		else
		    pbo -> val = 0;
	    }
	    else
	    {
		if (pbo -> invt)
		    pbo -> val = 0;
		else
		    pbo -> val = 1;
	    }
	    pbo -> livl = pbo -> val;
	    alarm (pbo);
	    monitor (pbo);

	    /* do redundant updates */
	    redQInit ((struct dbCommon *)pbo, redbuf, sizeof (redbuf), type);

	    if (stat != pbo->stat) 
		redQField (redbuf, mylist [STAT]);
	
	    if (sevr != pbo->sevr) 
		redQField (redbuf, mylist [SEVR]);
	
	    if (val != pbo->val) 
		redQField (redbuf, mylist [VAL]);
	
	    if (lalm != pbo->lalm) 
		redQField (redbuf, mylist [LALM]);

	    if (TsCmpStampsNE (&latm, &pbo -> latm) )
		redQField (redbuf, mylist [LATM]);

	    if (lval != pbo -> lval)
		redQField (redbuf, mylist [LVAL]);
	
	    if (mlst != pbo -> mlst)
		redQField (redbuf, mylist [MLST]);
	
	    if (livl != pbo -> livl)
		redQField (redbuf, mylist [LIVL]);
	
	    if (oval != pbo -> oval)
		redQField (redbuf, mylist [OVAL]);
	
	    if (hhll != pbo -> hhll)
		redQField (redbuf, mylist [HHLL]);
	
	    if (rtct != pbo -> rtct)
		redQField (redbuf, mylist [RTCT]);
	
	    redQSend (redbuf);
	}
    }

    return;
}


