/*recCalc.c,v 1.1.1.1 1999/08/03 14:44:18 tis4000 Exp */

/* recCalc.c - Record Support Routines for Calculation records */
/*
 *      Original Author: Julie Sander and Bob Dalesio
 *      Current  Author: Marty Kraimer
 *      Date:            7-27-87
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  5-18-88         lrd     modified modulo and power to avoid math library
 * .02  5-19-88         lrd     modified absolute value to avoid math library
 *                              defined unary math lib routines as doubles
 *                              removed include math.h
 *                              stopped loading dinglers math routines (ml)
 *                              wrote a random number generator to return a
 *                                      double between 0 and 1
 * .03  12-09-88        lrd     fixed modulo not to perform zero division
 * .04  12-12-88        lrd     lock the record while processing
 * .05  12-13-88        lrd     made an alarm for math error
 * .06  12-15-88        lrd     Process the forward scan link
 * .07  12-23-88        lrd     Alarm on locked MAX_LOCKED times
 * .08  01-11-89        lrd     Add Right and Left Shift
 * .09  02-01-89        lrd     Add Trig functions
 * .10  03-14-89        lrd     fix true on C question mark operator
 * .11  03-29-89        lrd     make hardware errors MAJOR
 *                              remove hw severity spec from database
 * .12  04-06-89        lrd     add monitor detection
 * .13  05-03-89        lrd     removed process mask from arg list
 * .14  06-05-89        lrd     check for negative square root
 * .15  08-01-89        lrd     full range of exponentiation using pow(x,y)
 * .16  04-04-90        lrd     fix post events for read and calc alarms
 *                              fix neg base raised to integer exponent
 * .17  04-06-90        lrd     change conditional to check for 0 and non-zero
 *                              instead of 0 and 1 (more 'C' like)
 * .18  10-10-90	mrk	Made changes for new record support
 *				Note that all calc specific details moved
 *				to libCalc
 * .19  03-12-93	ems	deleted ack from redundancy
 * .20  01-03-94	ems	ack = 0 in init_record
 *				in process, if not init'd do nothing
 *				  commented out check for init before do fetch
 *				  not needed
 * .21	3-01-94		ems	added support for process stats
 * .22  03-10-94	ems	alarm chgs, if val is past alarm setting, set that
 *				alarm condition, else if val was in that alarm and
 *				it is still in hyst range of alarm value, then stay
 *				in alarm, remove return from high alarm, let user
 *				determine which is more important.
 * .23	03-29-94	ems	make forward func defs agree with actual def
 * .24  04-04-94	ems	init roci to 0, redund cant handle -1, rework bumping of
 *				roci in alarm, make sure does not exceed MAX_ROC_VALEUS
 * .25	04-26-94	ems	added INI2 to namelist, chg NUMFLDS to NUMCALCFLDS
 *				rewrite init_record, get_control_double, if HYST
 *				set low to 0 and high to hopr - lopr
 * .26	06-13-94	ems	added support for maximize severity
 * .27	08-08-94	ems	added field type to get_graphic_double and
 *				get_control_double
 *				monitor(), rm DBE_LOG from ack, only archive VAL if
 *				tenb (trending) enabled, only set DBE_VALUE on 
 *				current input values
 * .28	10-12-94	ems	add TIME to redundancy, fpas chg'd to frps, added fpps
 *				(1st process pass), do time stamp when stat or sevr chg,
 *				see monitor(),and not in process(),set lalm to val when 
 *				stat or sevr chg,in process() if fpps (first process)
 *				init lalm to val and if time stamp is 0, init it also
 * .29	10-20-94	ems	chg time to latm (last alarm time) added 
 *				added lptm (last process time)
 * .30	03-15-95	ems	monitor(), do not set ACK if sevr is EVENT_ALARM
 * .31	03-24-95	ems	arg to resGetCaAnswer needs to be a short not int
 * .32	09-11-95	ems	monitor(), archive when stat/sevr chg's also
 * .33	09-12-95	ems	monitor(), see mod for handling VALID_ALARM
 * .34	11-17-95	ems	added arg of max num bytes topost fix, new error 7
 *				means expression to large, deleted error return from
 *				postfix, 9,10,11 no longer returned by postfix
 * .35	02-06-96	ems	added code for deviation alarm
 *				always update rate of chg buffer as soon as alarms are enabled
 *				    so will always have latest data, also check if rate of chg
 *				    value is > 0.0 before process roc alarm
 *				remove return from HIHI and LOLO alarms, so will always check 
 *				    all alarm sevrs
 *				alarm(), check for VALID and return, also set nsta & nsev =
 *				NO_ALARM BEFORE do rate of hcg alarm NOT after
 * .36	04-03-96	ems	added code to support alarm delay
 * .37	06-19-96	ems	added code for alarm disable ADIS
 * .38	07-02-96	ems	alarm(), update rocb if roc > 0.0 or frps is set
 * .39	07-02-96	ems	process(), compare msev >= to nsev rather than >, so msev
 *				can be expressed if one of the inputs are bad
 */

#include	<vxWorks.h>
#include	<inetLib.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<lstLib.h>
#include	<string.h>
#include	<memLib.h>
#include	<stdlib.h>
#include	<logLib.h>

#include	"alarm.h"
#include	"dbDefs.h"
#include	"dbCommon.h"
#include	"link.h"
#include	"db_field_log.h"
#include	"dbAccess.h"
#include	"dbFldTypes.h"
#include	"dbRecType.h"
#include	"errMdef.h"
#include	"recSup.h"
#include	"special.h"
#include	"rec/calcRecord.h"
#include	"redflds.h"
#include	"redque.h"
#include	"calcredfldidx.h"
#include	"restable.h"
#include	"db_addr.h"
#include	"recGbl.h"
#include	"dbEvent.h"
#include	"calcs.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table*/

#define report NULL
#define initialize NULL
static long init_record ();
static long process ();
static long special ();
static long get_value ();
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
static long get_units ();
static long get_precision ();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double ();
static long get_control_double ();
static long get_alarm_double ();

struct rset calcRSET = {
	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double
	};

/* The following definitions are required for redundancy */

static struct	fldDes	**mylist = 0;

static char	*namelist [] = {
		"STAT",
		"SEVR",
		"VAL",
		"ROCI",
		"ROCC",
		"LALM",
		"INI2",
		"LATM",
		"MLST",
		"HHLL",
		"HILO",
		"RTCT",
		"LARF",
		"DVCT",
		0
		};

static	unsigned	short	type;

static void alarm ();
static void monitor ();
static int fetch_values ();

#define ARG_MAX 12


/*------------------------------------------------------------------------*/

static long init_record (pcalc)

    struct	calcRecord	*pcalc;

{
    long		status;
    struct	link	*plink;
    int			i;
    short		master;
    double		*pvalue;
    short		error_number;
    char		rpbuf [sizeof (pcalc -> rpcl)];
    char		*pstr;
    char		redbuf [256];
    int			ini2_flag = 0;
    char		msgBuff [256];

	if (!(mylist)) 
	{
	    for (type = 1; type < 100; type++) 
	    {
		pstr = GET_PRECTYPE (type);
		if (strcmp (pstr, "calc") == 0) 
		    break;
	    }
	    mylist = redFldAlloc (type, NUMCALCFLDS);

	    if (!mylist)
	    {
		recGblRecordError (S_rec_Alloc, (struct dbCommon *)pcalc,
				   "calc, init_record, mylist failed");
		return (S_rec_Alloc);
	    }

	    for (i = 0; i < NUMCALCFLDS; i++) 
		redFldAdd (type, i, namelist [i]);
	}

	/* Allocate rate of change value buffer */

	if (!(pcalc->rocb)) 
	{
	    pcalc->rocb = (char *)calloc (1, (MAX_ROC_VALUES * sizeof (float)));
	    if (!pcalc -> rocb)
	    {
		recGblRecordError (S_rec_Alloc, (struct dbCommon *)pcalc,
				   "calc, init_record, rocb failed");
		return (S_rec_Alloc);
	    }
	    pcalc->rocc = 0;
	    pcalc->roci = 0;
	}
	pcalc -> init = TRUE;
	pcalc -> frps = TRUE;
	pcalc -> fpps = 1;
	pcalc -> nsta = NO_ALARM;
	pcalc -> nsev = NO_ALARM;
	pcalc -> msev = NO_ALARM;

	/* init alarm delay variables */
	pcalc -> larf = 0;
	pcalc -> hhll = 0;
	pcalc -> hilo = 0;
	pcalc -> dvct = 0;
	pcalc -> rtct = 0;

	/* init last process time and last alarm time */

	if (pcalc->lptm.secPastEpoch == 0)
	{
	    tsLocalTime (&pcalc->lptm);
	}
	if (pcalc->latm.secPastEpoch == 0)
	{
	    tsLocalTime (&pcalc->latm);
	}
        plink = &pcalc->inpa;
        pvalue = &pcalc->a;
        for (i = 0; i < ARG_MAX; i++, plink++, pvalue++) 
	{
            if (plink->type == CONSTANT) 
		*pvalue = plink->value.value;
        }

        status = postfix (pcalc->calc, rpbuf, &error_number, sizeof (rpbuf));
        if (status) 
	{
	    sprintf (msgBuff, "%s expression error: ",pcalc -> name);
	    if (error_number == 4)
		strcat (msgBuff, "operand needed\n");
	    else if (error_number == 5)
		strcat (msgBuff, "operator expected\n");
	    else if (error_number == 6)
		strcat (msgBuff, "umatched parens\n");
	    else if (error_number == 7)
		strcat (msgBuff, "expression to large\n");
	    else if (error_number == 8)
		strcat (msgBuff, "illegal expression element\n");
	    else
	    {
		sprintf (redbuf, "unknown error %d\n", error_number); 
		strcat (msgBuff, redbuf);
	    }
	    logMsg(msgBuff,0,0,0,0,0,0);
	    pcalc -> nsta = NO_INIT_ALARM;
	    pcalc -> nsev = VALID_ALARM;
	}
	else
	{
            bcopy (rpbuf, pcalc->rpcl, sizeof (pcalc->rpcl));
	    ini2_flag = 1;
	}

	if (resGetCaAnswer(pcalc -> radr, &master) == RES_OK)
	{
	    if (master)
	    {
		/* post STAT, SEVR and VAL (monitor), if we will finally init, 
		   because we have read a value from the input device, set INI2,
		   send all redund fields to slave, when slave receives INI2, 
		   it will be fully init'd also
		*/

		monitor (pcalc);

		if (ini2_flag)
		{
		    pcalc -> ini2 = 1;

		    redQInit ((struct dbCommon *)pcalc, redbuf, sizeof (redbuf), type);
		    redQField (redbuf, mylist [STAT]);
		    redQField (redbuf, mylist [SEVR]);
		    redQField (redbuf, mylist [VAL]);
		    redQField (redbuf, mylist [ROCI]);
		    redQField (redbuf, mylist [ROCC]);
		    redQField (redbuf, mylist [LALM]);
		    redQField (redbuf, mylist [INI2]);
		    redQSend (redbuf);
		}
	    }
	}
	return (0);
}


/*------------------------------------------------------------------------*/

static long process (paddr)
    struct	dbAddr	*paddr;
{
    struct calcRecord	*pcalc = (struct calcRecord *) (paddr->precord);
    unsigned short	stat   = pcalc -> stat;
    unsigned short	sevr   = pcalc -> sevr;
    double		val    = pcalc -> val;
    short		roci   = pcalc -> roci;
    short		rocc   = pcalc -> rocc;
    double		lalm   = pcalc -> lalm;
    double		mlst   = pcalc -> mlst;
    TS_STAMP		latm   = pcalc -> latm;
    short		hhll   = pcalc -> hhll;
    short		hilo   = pcalc -> hilo;
    short		rtct   = pcalc -> rtct;
    unsigned long	larf   = pcalc -> larf;
    short		dvct   = pcalc -> dvct;
    int			val_sent = 0;
    char		redbuf [256];

    /* if not init'd, do nothing */

    if ((!pcalc -> init) || (!pcalc -> ini2))
	return (0);

    /* get time stamp for last time record was processed */

    tsLocalTime (&pcalc -> lptm);

    pcalc -> msev = NO_ALARM;

    if (pcalc -> tagd)
    {
	pcalc -> nsta = TAGGED_ALARM;
	pcalc -> nsev = VALID_ALARM;
    }
    else if (pcalc -> oscn == OFF_SCAN)
    {
	pcalc -> nsta = SCAN_ALARM;
	pcalc -> nsev = VALID_ALARM;
    }
    else
    {
	pcalc -> nsta = NO_ALARM;
	pcalc -> nsev = NO_ALARM;

	/* update number of process times */
	if (process_stats)
	    process_cnts[type]++;	

	if (pcalc->mode == NORMAL) 
	{
	    pcalc->pact = TRUE;
	    if (fetch_values (pcalc) == 0) 
	    {
		if (calcPerform (&pcalc->a, &pcalc->val, pcalc->rpcl)) 
		{
		    if (pcalc->nsev < VALID_ALARM) 
		    {
			pcalc->nsta = CALC_ALARM;
			pcalc->nsev = VALID_ALARM;
		    }
		} 
		else 
		{
		    pcalc->udf = FALSE;
		}
	    }
	}

	/* check for alarms */
		
	alarm (pcalc);

	/* check for maximize severity */

	if ( (pcalc -> msev > NO_ALARM) && (pcalc -> msev >= pcalc -> nsev))
	{
	    pcalc -> nsev = pcalc -> msev;
	    pcalc -> nsta = LINK_ALARM;
	}
    }

    /* if the time stamp is 0, init and post val with
       alarm bit set, this makes sure that proper alarm
       messages are set if IOC was down and alarm status
       may have chg'd
    */
    if (pcalc -> latm.secPastEpoch == 0)
    {
	tsLocalTime (&pcalc -> latm);
	db_post_events ((struct dbCommon *)pcalc, (caddr_t)&pcalc -> val, DBE_VALUE | DBE_ALARM);
    }

    /* if it is the 1st process pass 
       init last value when alarm chg'd (lalm)
    */

    if (pcalc -> fpps)
    {
	pcalc -> fpps = 0;
	pcalc -> lalm = pcalc -> val;
    }

    /* check event list */

    monitor (pcalc);

    redQInit ((struct dbCommon *)pcalc, redbuf, sizeof (redbuf), type);

    if (stat != pcalc->stat) 
	redQField (redbuf, mylist [STAT]);
	
    if (sevr != pcalc->sevr) 
	redQField (redbuf, mylist [SEVR]);
	
    if (val != pcalc->val) 
    {
	val_sent = 1;
	redQField (redbuf, mylist [VAL]);
    }
	
    if (roci != pcalc->roci) 
    {
	if (!val_sent)
	    redQField (redbuf, mylist [VAL]);
	redQField (redbuf, mylist [ROCI]);
    }
    if (rocc != pcalc->rocc) 
	redQField (redbuf, mylist [ROCC]);
	
    if (lalm != pcalc->lalm) 
	redQField (redbuf, mylist [LALM]);
	
    if (TsCmpStampsNE (&latm, &pcalc -> latm) )
	redQField (redbuf, mylist [LATM]);
	
    if (mlst != pcalc->mlst)
	redQField (redbuf, mylist [MLST]);

    if (hhll != pcalc -> hhll)
	redQField (redbuf, mylist [HHLL]);

    if (hilo != pcalc -> hilo)
	redQField (redbuf, mylist [HILO]);

    if (rtct != pcalc -> rtct)
	redQField (redbuf, mylist [RTCT]);

    if (larf != pcalc -> larf)
	redQField (redbuf, mylist [LARF]);

    if (dvct != pcalc -> dvct)
	redQField (redbuf, mylist [DVCT]);

    redQSend (redbuf);

    /* process the forward scan link record */

    if (pcalc->flnk.type == DB_LINK) 
	dbScanPassive ((struct dbAddr *)pcalc->flnk.value.db_link.pdbAddr);

    pcalc->pact = FALSE;

    return (0);
}


/*------------------------------------------------------------------------*/

static long special (paddr, after)

    struct	dbAddr	*paddr;
    int			after;
{
	long			status;
	struct	calcRecord  	*pcalc = (struct calcRecord *) (paddr->precord);
	int			special_type = paddr->special;
	short			error_number;
	char			rpbuf [sizeof(pcalc -> rpcl)];
	char			msgBuff [256];
	char			redbuf [256];

	if (!after)
	    return(0);
	
	switch (special_type) 
	{
	    case (SPC_CALC):
		status = postfix (pcalc->calc, rpbuf, &error_number, sizeof (rpbuf));
		if (status) 
		{
	    	    sprintf (msgBuff, "%s expression error: ",pcalc -> name);
	    	    if (error_number == 4)
			strcat (msgBuff, "operand needed\n");
	    	    else if (error_number == 5)
			strcat (msgBuff, "operator expected\n");
	    	    else if (error_number == 6)
			strcat (msgBuff, "umatched parens\n");
	    	    else if (error_number == 7)
			strcat (msgBuff, "expression to large\n");
	    	    else if (error_number == 8)
			strcat (msgBuff, "illegal expression element\n");
	    	    else
	    	    {
			sprintf (redbuf, "unknown error %d\n", error_number); 
			strcat (msgBuff, redbuf);
	    	    }
	    	    logMsg(msgBuff,0,0,0,0,0,0);
		    return (status);
	        }
	        else
	        {
	            bcopy (rpbuf, pcalc->rpcl, sizeof (pcalc->rpcl));
	            db_post_events ((struct dbCommon *)pcalc, pcalc->calc, DBE_VALUE);
	        }
	        return (0);

	    default:
	        recGblDbaddrError (S_db_badChoice, paddr, "calc: special");
	        return (S_db_badChoice);
	}
}


/*------------------------------------------------------------------------*/

static long get_value (pcalc, pvdes)

	struct	calcRecord	*pcalc;
	struct	valueDes	*pvdes;
{
	pvdes->field_type = DBF_DOUBLE;
	pvdes->no_elements = 1;
	(double *) (pvdes->pvalue) = &pcalc->val;
	return (0);
}

/*------------------------------------------------------------------------*/

static long get_units (paddr, units)

    struct	dbAddr	*paddr;
    char		*units;
{
	struct	calcRecord	*pcalc = (struct calcRecord *)paddr->precord;

	strncpy (units, pcalc->egu, sizeof (pcalc->egu));
	return (0);
}

/*------------------------------------------------------------------------*/

static long get_precision (paddr, precision)
    struct	dbAddr	*paddr;
    long		*precision;
{
	struct	calcRecord	*pcalc = (struct calcRecord *) paddr->precord;

	*precision = pcalc->prec;
	return (0);
}

/*------------------------------------------------------------------------*/

static long get_alarm_double (paddr, pad)

    struct	dbAddr		*paddr;
    struct	dbr_alDouble	*pad;
{
	struct	calcRecord	*pcalc = (struct calcRecord *) paddr->precord;

	pad->upper_alarm_limit = pcalc->hihi;
	pad->upper_warning_limit = pcalc->high;
	pad->lower_warning_limit = pcalc->low;
	pad->lower_alarm_limit = pcalc->lolo;
	return (0);
}


/*------------------------------------------------------------------------*/

static long get_graphic_double (paddr, pgd)

    struct	dbAddr		*paddr;
    struct	dbr_grDouble	*pgd;
{
    struct	calcRecord	*pcalc = (struct calcRecord *) paddr->precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct calcRecord *)0) -> hyst) ):
	case ( (int)&(((struct calcRecord *)0) -> roc) ):
	case ( (int)&(((struct calcRecord *)0) -> adel) ):
	case ( (int)&(((struct calcRecord *)0) -> mdel) ):
	case ( (int)&(((struct calcRecord *)0) -> devl) ):
	{
	    float    span;

	    span = pcalc -> hopr - pcalc -> lopr;
	    if (span < 0.0)
	        span = -span;
	    pgd -> upper_disp_limit = span;
	    pgd -> lower_disp_limit = 0;
	    break;
	}

	case ( (int)&(((struct calcRecord *)0) -> rocn) ):
	    pgd -> upper_disp_limit = 60;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct calcRecord *)0) -> aldl) ):
	    pgd -> upper_disp_limit = 9999;
	    pgd -> lower_disp_limit = 0;
	    break;

	default:
	    pgd -> upper_disp_limit = pcalc -> hopr;
	    pgd -> lower_disp_limit = pcalc -> lopr;
	    break;
    }

    return (0);
}


/*------------------------------------------------------------------------*/

static long get_control_double (paddr, pcd)

    struct	dbAddr		*paddr;
    struct	dbr_ctrlDouble	*pcd;
{
    struct	calcRecord	*pcalc = (struct calcRecord *) paddr->precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct calcRecord *)0) -> hyst) ):
	case ( (int)&(((struct calcRecord *)0) -> roc) ):
	case ( (int)&(((struct calcRecord *)0) -> adel) ):
	case ( (int)&(((struct calcRecord *)0) -> mdel) ):
	case ( (int)&(((struct calcRecord *)0) -> devl) ):
	{
	    float    span;

	    span = pcalc -> hopr - pcalc -> lopr;
	    if (span < 0.0)
	        span = -span;
	    pcd -> upper_ctrl_limit = span;
	    pcd -> lower_ctrl_limit = 0;
	    break;
	}

	case ( (int)&(((struct calcRecord *)0) -> rocn) ):
	    pcd -> upper_ctrl_limit = 60;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	default:
	    pcd -> upper_ctrl_limit = pcalc -> hopr;
	    pcd -> lower_ctrl_limit = pcalc -> lopr;
	    break;
    }

    return (0);
}


/*------------------------------------------------------------------------*/

static void alarm (pcalc)
    struct	calcRecord	*pcalc;
{
    double		val = pcalc -> val;
    register float	hval, lval;
    register short	i, inx;
    int			adis = 0;
    ALARM_LIMITS	alarm_limits;

    /* add value to the rate of change buffer, 
       if there is a rate of chg limit or 1st pass is set
    */

    if ( (pcalc -> roc > 0.0) || (pcalc -> frps == FALSE) )
    {
	pcalc -> roci++;
	if (pcalc->roci >= MAX_ROC_VALUES) 
	    pcalc->roci = 0;
	    
	if (pcalc->rocc < MAX_ROC_VALUES) 
	    pcalc->rocc++;
	    
	if (pcalc -> frps)
	{
	    pcalc -> frps = FALSE;
	    *(float *) (pcalc -> rocb) = val;
	}
	*(float *) (pcalc->rocb + (pcalc->roci * sizeof (float))) = val;
    }

    /* check for alarm inhibit */

    adis = AlarmDisableLink ((void *)pcalc);

    if ( (pcalc->aenb == ALARM_DISABLED) || (adis != 0) )
    {
	if (pcalc -> nsta != NO_INIT_ALARM)
	{
	    pcalc->nsta = INHIBIT_ALARM;
	    pcalc->nsev = NO_ALARM;
	}
	return;
    }

    if (pcalc->udf == TRUE ) 
    {
	if (pcalc->nsev < VALID_ALARM) 
	{
	    pcalc->nsev = VALID_ALARM;
	    pcalc->nsta = UDF_ALARM;
	}
	return;
    }

    if ( (pcalc->rcsv != NO_ALARM) && (pcalc->rocb) && (pcalc -> roc > 0.0) ) 
    {
	if (pcalc->rocc >= 2) 
	{
	    inx = pcalc->roci;
	    hval = lval =  *(float *) (pcalc->rocb + (inx * sizeof (float)));
	    for (i = 0; 
		 i < ((pcalc -> rocc < pcalc -> rocn) ? pcalc -> rocc : pcalc -> rocn); 
		 i++) 
	    {
		if (--inx < 0) 
		    inx = MAX_ROC_VALUES - 1;
		    
		if ((*((float *) (pcalc->rocb + (inx * sizeof (float))))) > hval) 
		    hval = *((float *) (pcalc->rocb + (inx * sizeof (float))));
	    
		if ((*((float *) (pcalc->rocb + (inx * sizeof (float))))) < lval) 
		    lval = *((float *) (pcalc->rocb + (inx * sizeof (float))));
		    
	    }
	    if ((hval - lval) > pcalc->roc) 
	    {
		pcalc->nsev = pcalc->rcsv;
		pcalc->nsta = ROC_ALARM;
	    }
	}
    }

    /* set up for limit alarms */

    alarm_limits.hihi = pcalc -> hihi;
    alarm_limits.high = pcalc -> high;
    alarm_limits.low  = pcalc -> low;
    alarm_limits.lolo = pcalc -> lolo;
    alarm_limits.devl = pcalc -> devl;
    alarm_limits.devs = pcalc -> devs;
    alarm_limits.hyst = pcalc -> hyst;
    alarm_limits.hhsv = pcalc -> hhsv;
    alarm_limits.hsv  = pcalc -> hsv;
    alarm_limits.lsv  = pcalc -> lsv;
    alarm_limits.llsv = pcalc -> llsv;
    alarm_limits.dvsv = pcalc -> dvsv;

    AlarmLimits (&alarm_limits, val, (void *)pcalc);

    /* do a return to normal, check for alarm delay ALDL only
       on the listed alarms
    */

    if ( ((pcalc -> nsta == NO_ALARM) && (pcalc -> nsev == NO_ALARM)) &&
	 ((pcalc -> stat == HIHI_ALARM) || (pcalc -> stat == HIGH_ALARM) ||
	  (pcalc -> stat == LOW_ALARM)  || (pcalc -> stat == LOLO_ALARM) ||
	  (pcalc -> stat == DEVIATION_ALARM) ) )
    {
	if (pcalc -> aldl > 0)
	{
	    if (pcalc -> rtct < ALDL_MAX)
		pcalc -> rtct++;

	    /* delay return to normal so, continue last alarm */
	    if (pcalc -> rtct < pcalc -> aldl)
	    {
		pcalc -> nsta = pcalc -> stat; 
		pcalc -> nsev = pcalc -> sevr;
	    }
	}
    }
    else
	pcalc -> rtct = 0;

    return;
}


/*------------------------------------------------------------------------*/

static void monitor (pcalc)

    struct	calcRecord	*pcalc;
{
	unsigned short	monitor_mask;
	double		delta;
        short           stat, sevr, nsta, nsev;
	double		*pnew;
	double		*pprev;
	int		i;
	int		set_ack = 1;

        /* get previous stat and sevr  and new stat and sevr*/

        stat = pcalc->stat;
        sevr = pcalc->sevr;
        nsta = pcalc->nsta;
        nsev = pcalc->nsev;

        /*set current stat and sevr*/

        pcalc->stat = nsta;
        pcalc->sevr = nsev;
        pcalc->nsta = 0;
        pcalc->nsev = 0;

	/* if we are coming out of VALID_ALARM and our new stat/sevr are
	   the same as when we went into VALID_ALARM, do not set ack
	*/
	if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	     (nsta == pcalc -> psta) && (nsev == pcalc -> psev) )
	{
	    set_ack = 0;
	}
	/* if we are going into VALID_ALARM, and out previous stat was not
	   VALID_ALARM, remember previous non VALID_ALARM stat/sevr
	   and we do not want to set ACK,
	*/
	if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
	{
	    set_ack = 0;
	    pcalc -> psta = stat;
	    pcalc -> psev = sevr;
	    db_post_events ((struct dbCommon *)pcalc, (caddr_t)&pcalc -> psta, DBE_VALUE);
	    db_post_events ((struct dbCommon *)pcalc, (caddr_t)&pcalc -> psev, DBE_VALUE);
	}

        monitor_mask = 0;

        /* alarm condition changed this scan */

        if (stat != nsta || sevr != nsev) 
	{
            /* post events for alarm condition change*/

            monitor_mask = DBE_ALARM;

	    /* when alarm status chg's, get current time and
	       save VAL in lalm (last value when alarm chg'd)
	    */
	    tsLocalTime (&pcalc -> latm);
	    pcalc -> lalm = pcalc -> val;

            /* post stat and nsev fields */

            db_post_events ((struct dbCommon *)pcalc, (caddr_t)&pcalc->stat, DBE_VALUE);
            db_post_events ((struct dbCommon *)pcalc, (caddr_t)&pcalc->sevr, DBE_VALUE);
	    if ((pcalc->sevr != NO_ALARM)&& (set_ack) &&
		(pcalc->sevr != EVENT_ALARM) )
	    {
		if (pcalc->ack == 0)
		{
		    pcalc->ack = 1;		/* new unacknowledged alarm */
	       	    db_post_events ((struct dbCommon *)pcalc, (caddr_t)&pcalc->ack, DBE_VALUE);
		}
	    }
        }
        /* check for value change */

        delta = pcalc->mlst - pcalc->val;
        if (delta < 0.0) 
	    delta = -delta;
	
        if (delta > pcalc->mdel) 
	{
	    /* post events for value change */

            monitor_mask |= DBE_VALUE;

            /* update last value monitored */

            pcalc->mlst = pcalc->val;
        }

        /* check for archive change  only if trending (tenb) enabled */

	if (pcalc -> tenb)
	{
            delta = pcalc->alst - pcalc->val;

            if (delta < 0.0) 
	    	delta = -delta;
	
	    /* if exceed archive deadband or there has been a change in alarm status */
            if ( (delta > pcalc->adel) || (monitor_mask & DBE_ALARM) )
	    {
            	/* post events on value field for archive change */

            	monitor_mask |= DBE_LOG;

            	/* update last archive value monitored */

            	pcalc->alst = pcalc->val;
	    }
        }

        /* send out monitors connected to the value field */

        if (monitor_mask) 
	{
            db_post_events ((struct dbCommon *)pcalc, (caddr_t)&pcalc->val, monitor_mask);

	    /* check all input fields for changes*/

	    for(i = 0, pnew = &pcalc->a, pprev = &pcalc->la;
		i < ARG_MAX; 
		i++, pnew++, pprev++) 
	    {

		if (*pnew != *pprev) 
		{
		    db_post_events ((struct dbCommon *)pcalc, (caddr_t)pnew, DBE_VALUE);
		    *pprev = *pnew;
		}
	    }
        }
        return;
}


/*------------------------------------------------------------------------*/

static int fetch_values (pcalc)

    struct	calcRecord	*pcalc;
{
	struct	link	*plink;		/* structure of the link field  */
	double		*pvalue;
	long		options, nRequest;
	int		i;
	long		status;

	for (i = 0, plink = &pcalc->inpa, pvalue = &pcalc->a;
	    i < ARG_MAX; 
	    i++, plink++, pvalue++) 
	{ 
	    if (plink->type != DB_LINK) 
		continue;
	    
	    options = 0;
	    nRequest = 1;
	    status = dbGetLink (&plink->value.db_link, 
				(struct dbCommon *)pcalc, 
				DBR_DOUBLE, (caddr_t)pvalue, 
				&options, &nRequest);
	    if (status != 0) 
	    {
		if (pcalc->nsev < VALID_ALARM) 
		{
		    pcalc->nsev = VALID_ALARM;
		    pcalc->nsta = LINK_ALARM;
		}
		return (-1);
	    }
	}
	return (0);
}

