/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/*	Modifications
*
*  .01	04-15=96	ems	created, initial entry for Alarm Limits
*				and Alarm States
*  .02	06-19-96	ems	added AlarmDisableLink()
*
*/


#include	<vxWorks.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<lstLib.h>
#include	<math.h>

#include	"alarm.h"
#include	"cvtTable.h"
#include	"dbDefs.h"
#include	"dbCommon.h"
#include	"link.h"
#include	"db_field_log.h"
#include	"dbAccess.h"
#include	"dbFldTypes.h"
#include	"dbRecType.h"
#include	"devSup.h"
#include	"errMdef.h"
#include	"recSup.h"
#include	"special.h"


/*****************************************************************************
*
*    NOTES:
*	LARF_xxxxx are defined in alarm.h
*	ALDL_MAX is defiend in alarm.h
*
*****************************************************************************/

void AlarmLimits (limits, val, rec)
    ALARM_LIMITS	*limits;
    double		val;
    void		*rec;
{
    struct dbCommon	*prec = (struct dbCommon *)rec;
    float	tempfl;

    /* if VAL is > than alarm severity or the block was in the alarm
       condition last scan, and VAL is greater than the alarm limit
       minus the alarm deadband, stay in the alarm unless alarm
       delay is in effect
    */

    /* HIHI alarm test */

    if (limits -> hhsv > NO_ALARM)
    {
	if ( (val > limits -> hihi) ||
	     ( (prec -> stat == HIHI_ALARM) &&
	       (val > (limits -> hihi - limits -> hyst)) ) )
	{
	    if (prec -> aldl > 0)
	    {
		/* if we were counting LOLO, 0 first, same counter
		   is used for LOLO and HIHI
		*/
		if (prec -> larf & LARF_LOLO)
		    prec -> hhll = 0;
		if (prec -> hhll < ALDL_MAX)
		    prec -> hhll++;
		prec -> larf |= LARF_HIHI;

		/* if count is > than alarm delay and the current alarm
		   severity is less than the HIHI alarm severity
		   set HIHI
		*/
		if ( (prec -> hhll >= prec -> aldl) &&
		     (prec -> nsev < limits -> hhsv) )
		{
		    prec -> nsta = HIHI_ALARM;
		    prec -> nsev = limits -> hhsv;
		}
	    }
	    else
	    {
		if (prec -> nsev < limits -> hhsv)
		{
		    prec -> nsta = HIHI_ALARM;
		    prec -> nsev = limits -> hhsv;
		}
	    }
	}
	else
	{
	    if (prec -> aldl > 0)
	    {
		/* remove the HIHI alarm bit */
		prec -> larf &= ~LARF_HIHI;

		/* since hhll ocunts both HIHI and LOLO, if LOLO is 
		   not set, zero hhll counter
		*/
		if (!(prec -> larf & ((unsigned long)LARF_LOLO)))
		    prec -> hhll = 0;
	    }
	}
    }
    else
    {
	if (prec -> aldl > 0)
	{
	    prec -> larf &= ~LARF_HIHI;
	    if (!(prec -> larf & ((unsigned long)LARF_LOLO)))
		prec -> hhll = 0;
	}
    }

    /* LOLO alarm test */

    if (limits -> llsv > NO_ALARM)
    {
	if ( (val < limits -> lolo) ||
	     ( (prec -> stat == LOLO_ALARM) &&
	       (val < (limits -> lolo + limits -> hyst)) ) )
	{
	    if (prec -> aldl > 0)
	    {
		if (prec -> larf & LARF_HIHI)
		    prec -> hhll = 0;
		if (prec -> hhll < ALDL_MAX)
		    prec -> hhll++;
		prec -> larf |= LARF_LOLO;
		if ( (prec -> hhll >= prec -> aldl) &&
		     (prec -> nsev < limits -> llsv) )
		{
		    prec -> nsta = LOLO_ALARM;
		    prec -> nsev = limits -> llsv;
		}
	    }
	    else
	    {
		if (prec -> nsev < limits -> llsv)
		{
		    prec -> nsta = LOLO_ALARM;
		    prec -> nsev = limits -> llsv;
		}
	    }
	}
	else
	{
	    if (prec -> aldl > 0)
	    {
		prec -> larf &= ~LARF_LOLO;
		if (!(prec -> larf & ((unsigned long)LARF_HIHI)))
		    prec -> hhll = 0;
	    }
	}
    }
    else
    {
	if (prec -> aldl > 0)
	{
	    prec -> larf &= ~LARF_LOLO;
	    if (!(prec -> larf & ((unsigned long)LARF_HIHI)))
		prec -> hhll = 0;
	}
    }

    /* HIGH alarm test */

    if (limits -> hsv > NO_ALARM)
    {
	if ( (val > limits -> high) ||
	     ( (prec -> stat == HIGH_ALARM) &&
	       (val > (limits -> high - limits -> hyst)) ) )
	{
	    if (prec -> aldl > 0)
	    {
		if (prec -> larf & LARF_LOW)
		    prec -> hilo = 0;
		if (prec -> hilo < ALDL_MAX)
		    prec -> hilo++;
		prec -> larf |= LARF_HIGH;
		if ( (prec -> hilo >= prec -> aldl) &&
		     (prec -> nsev < limits -> hsv) )
		{
		    prec -> nsta = HIGH_ALARM;
		    prec -> nsev = limits -> hsv;
		}
	    }
	    else
	    {
		if (prec -> nsev < limits -> hsv)
		{
		    prec -> nsta = HIGH_ALARM;
		    prec -> nsev = limits -> hsv;
		}
	    }
	}
	else
	{
	    if (prec -> aldl > 0)
	    {
		prec -> larf &= ~LARF_HIGH;
		if (!(prec -> larf & ((unsigned long)LARF_LOW)))
		    prec -> hilo = 0;
	    }
	}
    }
    else
    {
	if (prec -> aldl > 0)
	{
	    prec -> larf &= ~LARF_HIGH;
	    if (!(prec -> larf & ((unsigned long)LARF_LOW)))
		prec -> hilo = 0;
	}
    }

    /* LOW alarm test */

    if (limits -> lsv > NO_ALARM)
    {
	if ( (val < limits -> low) ||
	     ( (prec -> stat == LOW_ALARM) &&
	       (val < (limits -> low + limits -> hyst)) ) )
	{
	    if (prec -> aldl > 0)
	    {
		if (prec -> larf & LARF_HIGH)
		    prec -> hilo = 0;
		if (prec -> hilo < ALDL_MAX)
		    prec -> hilo++;
		prec -> larf |= LARF_LOW;
		if ( (prec -> hilo >= prec -> aldl) &&
		     (prec -> nsev < limits -> lsv) )
		{
		    prec -> nsta = LOW_ALARM;
		    prec -> nsev = limits -> lsv;
		}
	    }
	    else
	    {
		if (prec -> nsev < limits -> lsv)
		{
		    prec -> nsta = LOW_ALARM;
		    prec -> nsev = limits -> lsv;
		}
	    }
	}
	else
	{
	    if (prec -> aldl > 0)
	    {
		prec -> larf &= ~LARF_LOW;
		if (!(prec -> larf & ((unsigned long)LARF_HIGH)))
		    prec -> hilo = 0;
	    }
	}
    }
    else
    {
	if (prec -> aldl > 0)
	{
	    prec -> larf &= ~LARF_LOW;
	    if (!(prec -> larf & ((unsigned long)LARF_HIGH)))
		prec -> hilo = 0;
	}
    }

    /* DEVIATION alarm test */

    if (limits -> devl > 0.0)
    {
	if (limits -> dvsv > NO_ALARM)
	{
	    tempfl = val - limits -> devs;
	    if (tempfl < 0.0)
		tempfl = -tempfl;
	    if ( (tempfl >= limits -> devl) ||
		 ( (prec -> stat == DEVIATION_ALARM) &&
		   (tempfl >= (limits -> devl - limits -> hyst)) ) )
	    {
		if (prec -> aldl > 0)
		{
		if (prec -> dvct < ALDL_MAX)
		    prec -> dvct++;
		prec -> larf |= LARF_DEV;
		if ( (prec -> dvct >= prec -> aldl) &&
		     (prec -> nsev < limits -> dvsv) )
		{
		    prec -> nsta = DEVIATION_ALARM;
		    prec -> nsev = limits -> dvsv;
		}
		}
		else
		{
		    if (prec -> nsev < limits -> dvsv)
		    {
			prec -> nsta = DEVIATION_ALARM;
			prec -> nsev = limits -> dvsv;
		    }
		}
	    }
	    else
	    {
		if (prec -> aldl > 0)
		{
		    prec -> larf &= ~LARF_DEV;
		    prec -> dvct = 0;
		}
	    }
	}
	else
	{
	    if (prec -> aldl > 0)
	    {
		prec -> larf &= ~LARF_DEV;
		prec -> dvct = 0;
	    }
	}
    }
    else
    {
	if (prec -> aldl > 0)
	{
	    prec -> larf &= ~LARF_DEV;
	    prec -> dvct = 0;
	}
    }
}


/*****************************************************************
*
*    AlarmStates
*	check if need to output a STATE alarm taking into account
*	alarm delay (ALDL)
*    NOTE:
*	rtn = 1, do return to normal code
*
******************************************************************/

void AlarmStates (rec, rtn)
    void	*rec;
    int		rtn;
{
    struct dbCommon	*prec = (struct dbCommon *)rec;

    if (rtn == 0)
    {
	if (prec -> nsta == STATE_ALARM)
	{
	    if (prec -> aldl > 0)
	    {
		if (prec -> hhll < ALDL_MAX)
		    prec -> hhll++;
	    }
	    /* if the count is less than the alarm delay
	       do not output a state alarm yet
	    */
	    if (prec -> hhll < prec -> aldl)
	    {
		prec -> nsta = NO_ALARM;
		prec -> nsev = NO_ALARM;
	    }
	}
	else
	{
	    if (prec -> aldl > 0)
	    {
		prec -> hhll = 0;
	    }
	}
    }
    else
    {
	if ( (prec -> nsta == NO_ALARM) && (prec -> stat == STATE_ALARM) )
	{
	    if (prec -> aldl > 0)
	    {
		if (prec -> rtct < ALDL_MAX)
		    prec -> rtct++;
		if (prec -> rtct < prec -> aldl)
		{
		    prec -> nsta = prec -> stat;
		    prec -> nsev = prec -> sevr;
		}
	    }
	}
	else
	{
	    if (prec -> aldl > 0)
	    {
		prec -> rtct = 0;
	    }
	}
    }
    return;
}


/*****************************************************************
*
*    AlarmDisableLink
*	check if alarm disable link is defined and is non-zero
*    NOTE:
*	rtn = 1, disable alarms
*
******************************************************************/

int AlarmDisableLink (precord)
    void	*precord;
{
    struct dbCommon	*prec = (struct dbCommon *)precord;
    long		status = 0;
    long		nRequests = 1;
    long		options = 0;
    short		ini2;
    double		adis = 0;

    /* of cource if there is no link */
    if (prec -> adis.type != DB_LINK)
	return (0);

    ini2 = ((struct dbCommon *)
	    (((struct dbAddr*)(prec->adis.value.db_link.pdbAddr))->precord))->ini2;

    /* if ADIS is not fully init'd, do not allow disable of alarms */
    if (ini2 == 0)
	return (0);

    status = dbGetLink (&prec -> adis.value.db_link, prec,
			DBR_DOUBLE, (caddr_t)&adis,
			&options, &nRequests);
    if (status)
    {
	if (prec -> nsev < VALID_ALARM)
	{
	    prec -> nsta = LINK_ALARM;
	    prec -> nsev = VALID_ALARM;
	}
	return (0);
    }
    if (adis != 0)
	return (1);
    return (0);
}

