/*recDatamove.c,v 1.1.1.1 1999/08/03 14:44:19 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************************
 *
 *  MODULE NAME            : recDatamove.c 
 *  AUTHOR                 : emsteel
 *  DATE                   : 08 Oct 93
 *
 *  DESCRIPTION: Functions needed to process a datamove record
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 29 Mar 94
 *  REVISION: make forward func defs agree with actual func defs
 *
****************************************************************************/

/****************************************************************************
 *
 * Programmer's Notes:
 *
 *    initRecSup - see .../src/db/iocInit.c
 *    datamoveMsgQId - declared in .../src/rec/datamove.c
 *    datamoveSem - declared in .../src/rec/datamove.c
 *    args to taskSpawn - see .../epicsH/task_params.h
 *
****************************************************************************/


#include <vxWorks.h>
#include <inetLib.h>
#include <types.h>
#include <stdioLib.h>
#include <lstLib.h>
#include <msgQLib.h>
#include <taskLib.h>
#include <string.h>
#include <semLib.h>
#include <logLib.h>

#include "alarm.h"
#include "cvtTable.h"
#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#include "dbAccess.h"
#include "dbFldTypes.h"
#include "dbRecType.h"
#include "devSup.h"
#include "errMdef.h"
#include "recSup.h"
#include "special.h"
#include "rec/datamoveRecord.h"
#include "redflds.h"
#include "redque.h"
#include "datamoveredfldidx.h"
#include "restable.h"
#include "task_params.h"
#include "error.h"
#include "db_addr.h"
#include "recGbl.h"
#include "dbEvent.h"
#include "datamove.h"

   /* Create RSET - Record Support Entry Table*/

#define report NULL
static long initialize ();
static long init_record ();
static long process ();
static long special ();
static long get_value ();
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
#define get_units NULL
#define get_precision NULL
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
#define get_graphic_double NULL
#define get_control_double NULL
#define get_alarm_double NULL

struct rset datamoveRSET = {

   RSETNUMBER,
   report,
   initialize,
   init_record,
   process,
   special,
   get_value,
   cvt_dbaddr,
   get_array_info,
   put_array_info,
   get_units,
   get_precision,
   get_enum_str,
   get_enum_strs,
   put_enum_str,
   get_graphic_double,
   get_control_double,
   get_alarm_double
   };
             
   /* The folowing definitions are for redundancy */
 
static struct fldDes   **mylist = 0;

static  char *namelist [] = {
      "STAT",
      "SEVR",
      "VAL",
      0
      };

static unsigned short type = 0;


/****************************************************************************
 *
 *  MODULE NAME            : initialize 
 *  AUTHOR                 : emsteel
 *  DATE                   : 08 Oct 93
 *
 *  CALLING MODULE(S)      : initRecSup in iocInit
 *
 *
 *  DESCRIPTION: called once by initRecSup for all datamove records
 *               get index for datamove point (type)
 *               set up a msgQ to store ptrs to all datamove records
 *               set up semaphore between tasks
 *               spawn the datamove task
 *
****************************************************************************/

static long initialize ()
{
   unsigned short    i;
   int               taskId;
   STATUS            status;
   char              *pstr;
   struct datamoveRecord *pdatamove = NULL;

   for (i = 0, type = 1; type < 100; type++)
   {
      pstr = GET_PRECTYPE (type);
      if (strncmp (pstr, "datamove", 8) == 0)
      {
         i = 1;
         break;
      }
   }
   if (!i)
   {
      recGblRecordError (S_db_noRecType, (struct dbCommon *)pdatamove, "datamove, init_record");
      return (S_db_noRecType);
   }
   if ((datamoveMsgQId = msgQCreate (10, sizeof (pdatamove),
                                     MSG_Q_FIFO)) == NULL)
   {
      errMessage (-1L, "datamove, initialize, failed to create msgQ");
      return (-1L);
   }

   if ((datamoveSem = semBCreate (SEM_Q_PRIORITY, SEM_EMPTY)) == NULL)
   {
      errMessage (-1L, "datamove, initialize, failed to create semaphore");
      return (-1L);
   }

   if ((taskId = taskSpawn (DATAMOVE_NAME, DATAMOVE_PRI, DATAMOVE_OPT,
                            DATAMOVE_STACK, (FUNCPTR)datamove, 0,0,0,0,0,0,0,0,0,0)) == ERROR)
   {
       errMessage (-1L, "datamove, initialize, failed to start task");
       if ((status = msgQDelete (datamoveMsgQId)) == ERROR)
           errMessage (-1L, "datamove, failed to release msgQ");
       return (-1L);
   }

   logMsg ("datamove initialized\n",0,0,0,0,0,0);
   return (0);
}


/****************************************************************************
 *
 *  MODULE NAME            : init_record 
 *  AUTHOR                 : emsteel
 *  DATE                   : 08 Oct 93
 *
 *          ARGUMENT(S)    : pdatamove - ptr to datamove record 
 *
 *  DESCRIPTION: initialize this record
 *               init redundancy list
 *               send the address to the datamove task msg queue
 *               do a semGive, this will give the datamove task a kick in
 *                  butt and let it know it has some work to do
 *               normally the init flag would be set by this routine,
 *                  however, it will be set by the datamove task when
 *                  a successful ca_add_event () for this record has been
 *                  accomplished
 *
****************************************************************************/

static long init_record (pdatamove)
    struct datamoveRecord   *pdatamove;

{
   unsigned short   i;
   char     *pstr;
   char     errmsg[128];

   if (!(mylist))
   {
      if (!type)
      {
         for (i = 0, type = 1; type < 100; type++)
         {
            pstr = GET_PRECTYPE (type);
            if (strncmp (pstr, "datamove", 8) == 0)
            {
               i = 1;
               break;
            }
         }
         if (!i)
         {
            recGblRecordError (S_db_noRecType, (struct dbCommon *)pdatamove, 
                               "datamove, init_record");
            return (S_db_noRecType);
         }
      }
      mylist = redFldAlloc (type, NUMDATAMOVEFLDS);

      for (i = 0; i < NUMDATAMOVEFLDS; i++)
         redFldAdd (type, i, namelist [i]);
   }

   /*
   Don't go any further until the Channel Access task is ready and
   the name server has already been found.
   */

   if (dmCaAvail == 0)
	return (-1L);

   if (!pdatamove -> inam[0])
   {
      recGblRecordError (S_db_badField, (struct dbCommon *)pdatamove, 
                         "datamove: init_record, no input");
      return (S_db_badField);
   }

/*
	DANGER! DANGER! WARNING! WARNING! DANGER! DANGER!
	Following code removed by Gordon as a test.
*/

#if 0
   pdatamove -> ichd = 0;
   pdatamove -> ochd = 0;
   pdatamove -> evid = 0;
#endif

   /* init last process time and last alarm time */

   if (pdatamove->lptm.secPastEpoch == 0)
   {
	tsLocalTime (&pdatamove->lptm);
   }
   if (pdatamove->latm.secPastEpoch == 0)
   {
	tsLocalTime (&pdatamove->latm);
   }
   if ((msgQSend (datamoveMsgQId, (char *)&pdatamove, sizeof (pdatamove),
                  NO_WAIT, MSG_PRI_NORMAL)) == ERROR)
   {
      sprintf (errmsg, "datamove: init_record %s, msgQSend failed %0x",
                        pdatamove -> name,  errno);
      errMessage (-1L, errmsg);
      return (-1L);
   }

   if ((semGive (datamoveSem)) == ERROR)
   {
      sprintf (errmsg, "datamove, init %s, semGive failed", 
                        pdatamove -> name);
      errMessage (-1L, errmsg);
      return (-1L);
   }

   return (0);
}


/****************************************************************************
 *
 *  MODULE NAME            : process 
 *  AUTHOR                 : emsteel
 *  DATE                   : 08 Oct 93
 *
 *          ARGUMENT(S)    : paddr - ptr to record structure packet
 *
 *  DESCRIPTION: called to process a datamove record
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 03 Jan 94
 *  REVISION:  process(), if not init'd do nothing
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 20 Oct 94
 *  REVISION:	rm tsLocalTime call, do it in the event handler, see
 *		datamove.c
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 16 Sep 96
 *  REVISION:	gut process(), remove tagd and scan alarm and redundancy
 *		no meaning and tagd or scan alarm could overwrite the
 *		READ/WRITE alarm set by datamove.c
 *
****************************************************************************/

static long process (paddr)
    struct dbAddr   *paddr;
{
    long	status = 0;
    struct datamoveRecord   *pdatamove =
                                 (struct datamoveRecord *) (paddr->precord);

      /* if not init'd do nothing */

   if ((!pdatamove -> init) || (!pdatamove -> ini2))
      return (status);


   pdatamove->pact = FALSE;

      /* process the forward scan link record */

   if (pdatamove->flnk.type == DB_LINK)
      dbScanPassive ((struct dbAddr *)pdatamove->flnk.value.db_link.pdbAddr);

   return (status);
}


/****************************************************************************
 *
 *  MODULE NAME            : get_value 
 *  AUTHOR                 : emsteel
 *  DATE                   : 08 Oct 93
 *
****************************************************************************/

static long get_value(pdatamove,pvdes)
   struct   datamoveRecord   *pdatamove;
   struct   valueDes   *pvdes;
{
   pvdes->field_type = DBF_DOUBLE;
   pvdes->no_elements = 1;
   (double *) (pvdes->pvalue) = &pdatamove->val;
   return (0);
}


/****************************************************************************
 *
 *  MODULE NAME            : special 
 *  AUTHOR                 : emsteel
 *  DATE                   : 08 Oct 93
 *
 *          ARGUMENT(S)    : paddr - ptr to record struct packet
 *                           after - not used
 *
 *  DESCRIPTION: special processing when the input channel name INAM
 *               and the output channel name ONAM change
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 06 Jan 94
 *  REVISION: check for !after
 *
****************************************************************************/

static long special (paddr,after)
   struct   dbAddr   *paddr;
   int               after;
{
   struct datamoveRecord *pdatamove =
                                   (struct datamoveRecord *)(paddr->precord);
   int              special_type = paddr->special;
   char             errmsg[128];

   if (!after)
      return (0);

   switch (special_type)
   {
      case SPC_DM_INAM:
      case SPC_DM_ONAM:
	 /*
	 If the datamover task can't use CA yet, don't both him. Just set the
	 init bit back to zero. The initUpdate task (dbo_init.c) will keep
	 trying to reinit until it succeeds.
	 */

	 if (dmCaAvail == 0)
		{
		pdatamove -> init = 0;
		pdatamove -> ini2 = 0;
		return (-1);
		}
         if ((msgQSend (datamoveMsgQId, (char *)&pdatamove, sizeof (pdatamove),
                        NO_WAIT, MSG_PRI_NORMAL)) == ERROR)
         {
            sprintf (errmsg, "datamove: special for %s, msgQSend failed %d", 
                              pdatamove -> name, errno);
            errMessage (-1L, errmsg);
            return (-1L);
         }

         if ((semGive (datamoveSem)) == ERROR)
         {
            sprintf (errmsg, "datamove: special for %s, semGive failed",
                              pdatamove -> name);
            errMessage (-1L, errmsg);
            return (-1L);
         }
         break;

      default:
         recGblDbaddrError (S_db_badChoice, paddr, "datamove: special");
         return (S_db_badChoice);
   }

   return (0);
}
