/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/***********************************************************************
* recDdc.c - Record Support Routines for Digital Drum Controller
*
*      Author: emsteel
*      Date:   16 Apr 96
*
* Modification Log
* ----------------
* .01	01-03-94	ems	created
* .02	06-19-96	ems	added code for alarm disable ADIS
*
***********************************************************************/


#include	<vxWorks.h>
#include	<inetLib.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<lstLib.h>
#include	<string.h>
#include	<tickLib.h>
#include	<logLib.h>
#include	<stdlib.h>

#include	"alarm.h"
#include	"dbDefs.h"
#include	"dbCommon.h"
#include	"link.h"
#include	"db_field_log.h"
#include	"dbAccess.h"
#include	"dbFldTypes.h"
#include	"dbRecType.h"
#include	"devSup.h"
#include	"errMdef.h"
#include	"special.h"
#include	"recSup.h"
#include	"rec/ddcRecord.h"
#include	"redflds.h"
#include	"redque.h"
#include	"ddcredfldidx.h"
#include	"restable.h"
#include	"recGbl.h"
#include	"okToUse.h"
#include	"dbcSub.h"
#include	"error.h"
#include	"db_addr.h"
#include	"dbEvent.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table*/

#define report NULL
#define initialize NULL
static long init_record ();
static long process ();
static long special ();
static long get_value ();
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
#define get_units NULL
static long get_precision ();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double ();
static long get_control_double ();
#define get_alarm_double NULL

struct rset ddcRSET = {
	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double,
	};

struct ddcdset {		/* digital drum controller dset */
	long		number;
	DEVSUPFUN	dev_report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record;
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	write_ddc;
	};

/* The following definitions are required for redundancy */

static struct	fldDes	**mylist = 0;

static char	*namelist [] = {
		"STAT",
		"SEVR",
		"VAL",
		"LALM",
		"INI2",
		"LATM",
		"MLST",
		"OUTX",
		"PSIN",
		"LTCK",
		"LOVL",
		"LRST",
		"LDTM",
		"LMOD",
		"MSTN",
		"MODE",
		0
		};

static unsigned short	type;

/* define errmessage size */
#define	DDC_ERR_MSG		64

/* define the MAX number of outputs */

#define	DDC_MAX_OUTPUTS		32

/* these 5 sets of defines must agree with defs 
   in .../share/ascii/choiceRec.ascii
*/

/* defines for OMOD, operating mode */

#define	DDC_STEPPED_OPMODE	0
#define	DDC_TIMED_OPMODE	1

/* defines for TMOD, timing mode */

#define	DDC_REL_TIMEMODE	0
#define	DDC_ABS_TIMEMODE	1

/* defines for LTED, edge type */

#define	DDC_LEADING_EDGE	0
#define	DDC_TRAILING_EDGE	1

/* defines for DCTY, DDC type */

#define	DDC_PRIMARY		0
#define	DDC_SECONDARY		1
#define	DDC_DOWNSTREAM		2

/* defines fro LTED, edge type */

#define LEADING_EDGE    0
#define TRAILING_EDGE   1

/* defines for IACT, init action field */

#define	DDC_INIT_LAST_MODE	0
#define	DDC_INIT_AUTOMATIC	1
#define	DDC_INIT_MANUAL		2

/* defines for MODE must agree with defs in 
   .../share/ascii/choiceGbl.ascii
*/
#define	DDC_AUTOMATIC	0
#define	DDC_MANUAL	1

/* defines for limits on VAL field */

#define	DDC_VAL_LOWER_LIMIT	0
#define	DDC_VAL_UPPER_LIMIT	255

/* defines for the NUMB (num bits) field */

#define	DDC_NUMB_LOWER_LIMIT	1
#define	DDC_NUMB_UPPER_LIMIT	8

/* defines for the NUMW (num words) field */

#define	DDC_NUMW_LOWER_LIMIT	1
#define	DDC_NUMW_UPPER_LIMIT	32

static void alarm (struct ddcRecord *);
static void bit_split_val (struct ddcRecord *);
static int  check_step_values (struct ddcRecord *);
static void compute_abs_step (struct ddcRecord *, unsigned long);
static int  convert_time (struct ddcRecord *, char *, unsigned long *, char *);
static void do_drum_io (struct ddcRecord *, int);
static void find_start_abs_outx (struct ddcRecord *, TS_STAMP);
static void get_next_output_time (struct ddcRecord *, unsigned long *);
static void get_tot_steps (struct ddcRecord *, short*);
static void monitor (struct ddcRecord *);
static void set_next_outx (struct ddcRecord *);
static int  special_convert_time (struct dbAddr *, char *, int);


/*******************************************************
*
*    init routine, called until a ddc point is init'd
*
*******************************************************/

static long init_record (pddc)
    struct ddcRecord	*pddc;
{
    struct ddcdset	*pdset;
    long		status = 0;
    unsigned short	i = 0;
    unsigned short	master;
    unsigned long	rval;
    unsigned short	local_mode;
    unsigned short	save_mode = pddc -> mode;
    short		reset_value = 0;
    short		primary_topw;
    char		redbuf [512];
    char		*pstr;
    int			dont_init = 0;
    int			ini2_flag = 0;
    int			disable_timing = 0;
    int			use_val = 0;
    TS_STAMP		timenow;
    long		options = 0;
    long		nRequests = 1;
    char		errmsg[DDC_ERR_MSG];

    if (!(mylist)) 
    {
	for (type = 1; type < 100; type++) 
	{
	    pstr = GET_PRECTYPE (type);
	    if (strncmp (pstr, "ddc", 3) == 0) 
	    {
		i = 1;
		break;
	    }
	}
	
	if (i == 0)
	{
	    recGblRecordError(S_db_noRecType, (struct dbCommon *)pddc, "ddc:init_record");
	    return (S_db_noRecType);
	}

	mylist = redFldAlloc (type, NUMDDCFLDS);

	if (!mylist)
	{
	    recGblRecordError (S_rec_Alloc, (struct dbCommon *)pddc, 
			       "ddc:init_record, mylist failed");
	    return (S_rec_Alloc);
	}

	for (i = 0; i < NUMDDCFLDS; i++) 
	    redFldAdd (type, i, namelist [i]);
    }

    /* init hardware mask */

    pddc -> mask = 0;
    for (i = 0; i < pddc -> numb; i++)
    {
	pddc -> mask <<= 1;	/* shift left 1 bit */
	pddc -> mask |= 1;	/* set low order bit */
    }

    /* init alarm delay variables */
    pddc -> larf = 0;
    pddc -> hhll = 0;
    pddc -> hilo = 0;
    pddc -> dvct = 0;
    pddc -> rtct = 0;

    pddc -> fpps = 1;
    pddc -> nsta = NO_ALARM;
    pddc -> nsev = NO_ALARM;
    pddc -> msev = NO_ALARM;
    pddc -> outx = 0;			/* current output word index */
    pddc -> psin = 0;			/* previous step input */
    pddc -> val = DDC_VAL_LOWER_LIMIT;
    pddc -> rval = 0;
    pddc -> shft = 0;			/* mask shift value, set by driver */
    pddc -> lrst = 0;			/* last reset value */
    pddc -> ltck = tickGet ();
    pddc -> ldtm = 0;			/* last disable time flag */

    /* init last alarm time and last process time */
    if (pddc -> latm.secPastEpoch == 0)
	tsLocalTime (&pddc -> latm);
    if (pddc -> lptm.secPastEpoch == 0)
	tsLocalTime (&pddc -> lptm);

    if (pddc -> dcty == DDC_DOWNSTREAM)
    {
	pddc -> init = 1;
	ini2_flag = 1;
    }
    else
    {
	if (!(pdset = (struct ddcdset *) (pddc -> dset))) 
	{
	    recGblRecordError (S_dev_noDSET, (struct dbCommon *)pddc, "ddc: init_record");
	    return (S_dev_noDSET);
	}

	/* must have write_ddc function defined */

	if ((pdset -> number < 5) || (pdset -> write_ddc == NULL)) 
	{
	    recGblRecordError (S_dev_missingSup, (struct dbCommon *)pddc, "ddc: init_record");
	    return (S_dev_missingSup);
	}

	if ( (pddc -> numb < DDC_NUMB_LOWER_LIMIT) || 
	     (pddc -> numb > DDC_NUMB_UPPER_LIMIT) )
	{
	    sprintf (errmsg, "ddc:init_record,num bits range %d-%d",
			     DDC_NUMB_LOWER_LIMIT, DDC_NUMB_UPPER_LIMIT);
	    recGblRecordError (S_db_badField, (struct dbCommon *)pddc, errmsg);
	    return (S_db_badField);
	}

	if (check_step_values (pddc))
	{
	    return (S_db_badField);
	}

	if (pddc -> dcty == DDC_PRIMARY)
	{
	    if (pddc -> omod == DDC_TIMED_OPMODE)
	    {
		if (pddc -> tmod == DDC_REL_TIMEMODE)
		    pddc -> notm = pddc -> ltck;
		else	/* absolute time mode */
		{
		    tsLocalTime(&timenow);
		    if (timenow.secPastEpoch < 110000) 
		    {
			logMsg ("DDC for %s cannot init until Time is set in IOC\n",
				(int)pddc -> name, 0,0,0,0,0);
			return (0);
		    }
		    pddc -> notm = timenow.secPastEpoch;
		}
	    }
	}
	else if (pddc -> dcty == DDC_SECONDARY)
	{
	    if (pddc -> prcd.type != DB_LINK)
	    {
		recGblRecordError (S_db_badField, (struct dbCommon *)pddc, "ddc:init_record,no PRCD");
		return (S_db_badField);
	    }
	    primary_topw = ((struct ddcRecord*)
			    (((struct dbAddr*)(pddc->prcd.value.db_link.pdbAddr))->precord))->topw;
	    if (pddc -> topw != primary_topw)
	    {
		sprintf (errmsg, "ddc:init_record,topw(%d) != primary(%d)",
				  pddc->topw, primary_topw);
		recGblRecordError (S_db_badField, (struct dbCommon *)pddc, errmsg);
		return (S_db_badField);
	    }
	}

	pddc -> init = 1;

	dont_init = 0;

	if ( pdset -> init_record ) 
	{
	    status = (*pdset -> init_record) (pddc, process);
	    if (status)
	    {
		pddc -> nsta = NO_INIT_ALARM;
		pddc -> nsev = VALID_ALARM;
	    }
	    else
	    {
		rval = pddc -> rval;
		rval >>= pddc -> shft;
		pddc -> val = rval;
		bit_split_val (pddc);
		if (pddc -> dcty == DDC_PRIMARY)
		{
		    switch (pddc -> iact)
		    {
			case DDC_INIT_LAST_MODE:
			    break;

			case DDC_INIT_AUTOMATIC:
			    if (pddc -> mode != DDC_AUTOMATIC)
				pddc -> mode = DDC_AUTOMATIC;
			    break;

			case DDC_INIT_MANUAL:
			    if (pddc -> mode != DDC_MANUAL)
				pddc -> mode = DDC_MANUAL;
			    break;
		    }
		    /* if MODE has chg'd, must do CRC update to keep 
		       dbClient happy with the checksum
		    */
		    if (save_mode != pddc -> mode)
		    {
			pddc -> rcrc = DBC_CRCRecord ((char*)pddc, type);
		    }
		    pddc -> lmod = pddc -> mode;
		    local_mode = pddc -> mode;

		    if ( (pddc -> spin.type == DB_LINK) &&
			 (pddc -> omsl == CLOSED_LOOP) &&
			 (!dont_init) )
		    {
			if (!(okToUse(&pddc -> spin)) )
			    dont_init = 1;
			else
			{
			    /* get initial value for last step input value */
			    status = dbGetLink (&pddc -> spin.value.db_link, 
					        (struct dbCommon *)pddc,
						DBR_USHORT, (caddr_t)&pddc -> psin, 
						&options, &nRequests);
			    if (status)
			    {
				errMessage(-1, "ddc:init_record, Get error spin");
				dont_init = 1;
			    }
			}
		    }
		    if ( (pddc -> rsln.type == DB_LINK) &&
			 (pddc -> omsl == CLOSED_LOOP) &&
			 (!dont_init) )
		    {
			if (!(okToUse(&pddc -> rsln)) )
			    dont_init = 1;
			else
			{
			    /* get reset value */
			    status = dbGetLink (&pddc -> rsln.value.db_link, 
					        (struct dbCommon *)pddc,
						DBR_SHORT, (caddr_t)&reset_value, 
						&options, &nRequests);
			    if (status)
			    {
				errMessage(-1, "ddc:init_record, Get error rsln");
				dont_init = 1;
			    }
			    else
			    {
				if (reset_value == pddc -> resv)
				    pddc -> lrst = reset_value;
			    }
			}
		    }
		    if (!dont_init)
		    {
			pddc -> outx = 0;
			if ( (pddc -> omod == DDC_TIMED_OPMODE) &&
			     (reset_value != pddc -> resv) &&
			     (local_mode == DDC_AUTOMATIC) )
			{
			    if ( ((pddc -> lted == LEADING_EDGE) &&
				  (pddc -> psin != 0)) ||
				 ((pddc -> lted == TRAILING_EDGE) &&
				  (pddc -> psin == 0)) )
			    {
				disable_timing = 1;
			    }
			    if ( (pddc -> tmod == DDC_REL_TIMEMODE) &&
				 (disable_timing == 0) )
			    {
				pddc -> notm += pddc -> ot01;
			    }
			    else if ( (pddc -> tmod == DDC_ABS_TIMEMODE) &&
				      (disable_timing == 0) )
			    {
				find_start_abs_outx (pddc, timenow);
			    }
			}
		    }	/* end !dont_init */
		}
		else	/* we are secondary */
		{
		    if (!(okToUse (&pddc -> prcd)) )
			dont_init = 1;
		    else
		    {
			local_mode  = ((struct ddcRecord*)
				      (((struct dbAddr*)(pddc->prcd.value.db_link.pdbAddr))->precord))->mode;
			pddc -> outx = ((struct ddcRecord*)
				       (((struct dbAddr*)(pddc->prcd.value.db_link.pdbAddr))->precord))->outx;
		    }
		}
		if (!dont_init)
		{
		    if ( (!pddc -> tagd) && (pddc -> oscn == ON_SCAN) &&
			 (local_mode == DDC_AUTOMATIC) )
		    {
			do_drum_io (pddc, use_val);
		    }
		    pddc -> mstn = pddc -> outx + 1;
		    ini2_flag = 1;
		}
		else
		{
		    pddc -> nsta = NO_INIT_ALARM;
		    pddc -> nsev = VALID_ALARM;
		}
	    }
	}
	else  /* end if pdset -> init_record, let it init */
	{
	    ini2_flag = 1;
	}

    } /* end PRIMARY/SECONDARY */

    /* if we are MASTER post monitors and send to redundant partner */

    if (resGetCaAnswer (pddc -> radr, &master) == RES_OK)
    {
	if (master)
	{

	    /* post monitors, if we will finally init, 
	       set INI2, send all redund fields to slave, when slave 
	       receives INI2, it will be fully init'd also
	    */

	    monitor (pddc);

	    if (ini2_flag)
	    {
		pddc -> ini2 = 1;

		redQInit ((struct dbCommon *)pddc, redbuf, sizeof (redbuf), type);

		redQField (redbuf, mylist [STAT]);
		redQField (redbuf, mylist [SEVR]);
		redQField (redbuf, mylist [VAL]);
		redQField (redbuf, mylist [INI2]);
		redQField (redbuf, mylist [MLST]);
		redQField (redbuf, mylist [OUTX]);
		redQField (redbuf, mylist [PSIN]);
		redQField (redbuf, mylist [LTCK]);
		redQField (redbuf, mylist [LOVL]);
		redQField (redbuf, mylist [LRST]);
		redQField (redbuf, mylist [LDTM]);
		redQField (redbuf, mylist [LMOD]);
		redQField (redbuf, mylist [MSTN]);
		redQField (redbuf, mylist [MODE]);

		redQSend (redbuf);

		db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ob01, DBE_VALUE);
		db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ob02, DBE_VALUE);
		db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ob03, DBE_VALUE);
		db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ob04, DBE_VALUE);
		db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ob05, DBE_VALUE);
		db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ob06, DBE_VALUE);
		db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ob07, DBE_VALUE);
		db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ob08, DBE_VALUE);
		db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> psin, DBE_VALUE);
		db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> mstn, DBE_VALUE);
		db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> outx, DBE_VALUE);
		db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> mode, DBE_VALUE);
	    }
	}
    }

    return (0);
}


/********************
*
*    handle alarms 
*
********************/

static void alarm (pddc)
    struct ddcRecord	*pddc;
{
    int	adis = 0;

    /* check for alarm inhibit */

    adis = AlarmDisableLink ((void *)pddc);
    if ( (pddc -> aenb == ALARM_DISABLED) || (adis != 0) )
    {
	if (pddc -> nsta != NO_INIT_ALARM)
	{
	    pddc -> nsta = INHIBIT_ALARM;
	    pddc -> nsev = NO_ALARM;
	}
	return;
    }

    if(pddc -> udf == TRUE ) 
    {
	if (pddc -> nsev < VALID_ALARM) 
	{
	    pddc -> nsta = UDF_ALARM;
	    pddc -> nsev = VALID_ALARM;
	}
	return;
    }
}


/**************************************************************
*
*    split VAL into its bit components and store in bit fields
*    called fro init_record(), do_drum_io(), special()
*
**************************************************************/

static void bit_split_val (pddc)
    struct ddcRecord	*pddc;
{
    unsigned short	mask;
    short		*shptr;
    int			i;

    shptr = &pddc -> ob01;
    mask = 1;

    for (i = 0; i < DDC_NUMB_UPPER_LIMIT; i++)
    {
	if (pddc -> val & mask)
	    *shptr = 1;
	else
	    *shptr = 0;
	shptr++;
	mask <<= 1;
    }

    return;
}


/*******************************************
*
*    check validity of output words
*    accumulate total output words
*    check validity of NUMW
*    if (DDC_PRIMARY)
*	convert time strings
*    traverse downstream cascade
*    called from init_record()
*
********************************************/

static int check_step_values (pddc)
    struct ddcRecord	*pddc;
{
    struct ddcRecord	*dsddc;
    char		*pstr;
    unsigned long	*ulptr;
    unsigned short	*usptr;
    char		errmsg[DDC_ERR_MSG];
    char		emsg[DDC_ERR_MSG];
    int			i;
    int			bad_value = 0;

    dsddc = pddc;
    dsddc -> topw = 0;
    while (dsddc)
    {
	if ( (dsddc -> numw < DDC_NUMW_LOWER_LIMIT) || 
	     (dsddc -> numw > DDC_NUMW_UPPER_LIMIT) )
	{
	    sprintf (errmsg, "ddc:init_record,num output words range %d-%d",
			      DDC_NUMW_LOWER_LIMIT, DDC_NUMW_UPPER_LIMIT);
	    recGblRecordError (S_db_badField, (struct dbCommon *)dsddc, errmsg);
	    return (1);
	}

	/* check on validity of output values */
	bad_value = 0;
	usptr = &dsddc -> ow01;
	for (i = 0; i < dsddc -> numw; i++)
	{
	    /* if out of range 0-255 */
	    if ((*usptr) & 0xff00)
	    {
		sprintf (errmsg, "ddc:init_record,output word %d,range %d-%d",
				  i+1, DDC_VAL_LOWER_LIMIT, DDC_VAL_UPPER_LIMIT);
		recGblRecordError (S_db_badField, (struct dbCommon *)dsddc, errmsg);
		bad_value = 1;
	    }
	    else
	    {
		pddc -> topw++;
	    }
	    usptr++;
	}
	if (bad_value)
	    return (1);

	/* if I am PRIMARY ddc block and we are TIMED MODE
	   then convert time strings 
	*/
	if ( (pddc -> dcty == DDC_PRIMARY) &&
	     (pddc -> omod == DDC_TIMED_OPMODE) )
	{
	    bad_value = 0;
	    pstr = &dsddc -> ts01[0];
	    ulptr = &dsddc -> ot01;
	    for (i = 0; i < dsddc -> numw; i++)
	    {
		if (convert_time (dsddc, pstr, ulptr, emsg))
		{
		    sprintf (errmsg, "ddc:init_record,time string %d, %s",
				     i+1, emsg);
		    recGblRecordError (S_db_badField, (struct dbCommon *)dsddc, errmsg);
		    bad_value = 1;
		}
		pstr += sizeof (dsddc -> ts01);
		ulptr++;
	    }
	}

	if (dsddc -> dscd.type == DB_LINK)
	{
	    dsddc = (struct ddcRecord*)
		    (((struct dbAddr*)(dsddc->dscd.value.db_link.pdbAddr))->precord);
	}
	else
	    dsddc = 0;
    }
    if (bad_value)
	return (1);
    return (0);
}


/***************************************************
*
*    called from find_start_abs_outx() to find
*    where our current time falls in the time
*    sequence
*
***************************************************/

static void compute_abs_step (pddc, timenow)
    struct ddcRecord	*pddc;
    unsigned long	timenow;
{
    TS_STAMP		start_of_day, midnite;
    struct ddcRecord	*dsddc;
    unsigned long	*ulptr;
    unsigned long	last_output_time;
    short		i;
    short		totw;
    char		timebuff[32];
    char		*ptimebuff;
    short		last_outx = 0;
    short		first = 1;

    strcpy (timebuff, "00:00:00");
    ptimebuff = &timebuff[0];
    tsTextToStamp (&start_of_day, &ptimebuff);
    strcpy (timebuff, "23:59:59");
    tsTextToStamp (&midnite, &ptimebuff);
    midnite.secPastEpoch += 1;

    dsddc = pddc;

    if (pddc -> frst)
    {
	i = 1;
	last_outx = 1;
	ulptr = &pddc -> ot02;
	totw = 1;
    }
    else
    {
	i = 0;
	last_outx = 0;
	ulptr = &pddc -> ot01;
	totw = 0;
    }
    last_output_time = *ulptr;

    for ( ; i < pddc -> topw; i++, totw++)
    {
	if (totw >= dsddc -> numw)
	{
	    totw = 0;
	    dsddc = (struct ddcRecord*)
		    (((struct dbAddr*)(dsddc->dscd.value.db_link.pdbAddr))->precord);
	    ulptr = &dsddc -> ot01;
	}
	if (first)
	{
	    first = 0;
	    ulptr++;
	    continue;
	}
	if (last_output_time < (*ulptr))
	{
	    if ( (timenow > last_output_time) && (timenow < (*ulptr)) )
	    {
		pddc -> outx = last_outx;
		pddc -> notm = start_of_day.secPastEpoch + (*ulptr);
		break;
	    }
	}
	else	/* have gone past midnite */
	{
	    if ( (timenow > last_output_time) || (timenow < (*ulptr)) )
	    {
		pddc -> outx = last_outx;
		if (timenow > last_output_time)
		    pddc -> notm = midnite.secPastEpoch + (*ulptr);
		else
		    pddc -> notm = start_of_day.secPastEpoch + (*ulptr);
		break;
	    }
	}

	last_outx = i;
	last_output_time = *ulptr;
	ulptr++;
    }

    return;
}


/*******************************************************
*
*    convert a time output string to an unsigned long
*	calc total seconds
*	if TMOD is absolute, we ignore decimal secs
*	if TMOD is relative convert secs to clock
*	ticks and add in decimal value
*	NOTE: 6 clock ticks per 0.1 sec
*	called from check_step_values() and
*	special_convert_time()
*
*******************************************************/

static int  convert_time (pddc, time_string, ulptr, perr)
    struct ddcRecord	*pddc;
    char		*time_string;
    unsigned long	*ulptr;
    char		*perr;
{
    int		i;
    int		num_parts;
    int		k;
    int		len;
    int		num_decimals;
    int		in_decimal;
    char	decimal[2];
    char	parts[3][3];
    int		hms[3];

    len = strlen (time_string);
    memset (&parts[0], 0, sizeof (parts));
    memset (decimal, 0, sizeof (decimal));
    memset ((char *)&hms[0], 0, sizeof (hms));
    num_decimals = 0;
    in_decimal = 0;
    num_parts = 0;
    k = 0;
    for (i = 0; i < len; i++)
    {
	switch ((int)time_string[i])
	{
	    case ':':
		if (in_decimal)
		{
		    strcpy (perr, ": after decimal");
		    return (1);
		}
		num_parts++;
		k = 0;
		if (num_parts > 2)
		{
		    strcpy (perr, "too many :");
		    return (1);
		}
		break;

	    case '.':
		if (in_decimal)
		{
		    strcpy (perr, "more than 1 decimal pt");
		    return (1);
		}
		in_decimal = 1;
		break;

	    case '0':
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
	    case '8':
	    case '9':
		if (in_decimal)
		{
		    if (num_decimals >= 1)
		    {
			strcpy (perr, "to many decimals");
			return (1);
		    }
		    decimal[0] = time_string[i];
		    num_decimals++;
		}
		else
		{
		    if (k > 1)
		    {
			strcpy (perr, "to many chars in part");
			return (1);
		    }
		    parts[num_parts][k] = time_string[i];
		    k++;
		}
		break;

	    default:
		strcpy (perr, "illegal char");
		return (1);
	}
    }

    for (i = num_parts, k = 0; i >= 0; i--, k++)
	hms[2-i] = atoi (parts[k]);

    if ( (hms[0] < 0) || (hms[0] > 23) )
    {
	strcpy (perr, "hrs range 0-23");
	return (1);
    }
    if ( (hms[1] < 0) || (hms[1] > 59) )
    {
	strcpy (perr, "min range 0-59");
	return (1);
    }
    if ( (hms[2] < 0) || (hms[2] > 59) )
    {
	strcpy (perr, "sec range 0-59");
	return (1);
    }

    /* calc total seconds */
    *ulptr = (hms[0] * 3600) + (hms[1] * 60) + hms[2];

    if (pddc -> tmod == DDC_REL_TIMEMODE)
    {
	*ulptr = ((*ulptr) * vxTicksPerSecond) + (atoi(decimal) * 6);
    }

    return (0);
}


/************************************************
*
*    perform the I/O, assume that OUTX is at the 
*    correct spot
*    get the use first flag from the primary
*    called from init_record() and process()
*
*************************************************/

static void do_drum_io (pddc, use_val)
    struct ddcRecord	*pddc;
    int			use_val;
{
    struct ddcdset	*pdset = (struct ddcdset *)(pddc -> dset);
    struct ddcRecord	*dsddc;
    short		i;
    short		totw;
    short		output_value = 0;
    unsigned short	*usptr;

    if (use_val)
	output_value = pddc -> val;
    else
    {
	dsddc = pddc;
	usptr = &dsddc -> ow01;
	totw = 0;
	for (i = 0; i < pddc -> topw; i++, totw++)
	{
	    if (totw >= dsddc -> numw)
	    {
		totw = 0;
		dsddc = (struct ddcRecord*)
			(((struct dbAddr*)(dsddc->dscd.value.db_link.pdbAddr))->precord);
		usptr = &dsddc -> ow01;
	    }
	    if (i == pddc -> outx)
	    {
		output_value = *usptr;
		break;
	    }
	    usptr++;
	}
    }
    pddc -> rval = output_value;
    pddc -> rval <<= pddc -> shft;
    pddc -> val = output_value;
    pddc -> lovl = pddc -> val;
    bit_split_val (pddc);
    (*pdset -> write_ddc)(pddc);

    return;
}


/************************************************************
*
*    called from init_record, given the current epics time
*	get seconds since midnight of the current day
*	find the output index (OUTX) and next output time
*	    (NOTM) based on above time
*	also called from process()
*    NOTE: since only called when we are PRIMARY, just get
*	frst from pddc
*
************************************************************/

static void find_start_abs_outx (pddc, timenow)
    struct ddcRecord	*pddc;
    TS_STAMP		timenow;
{
    struct tsDetail	tsD;
    struct ddcRecord	*dsddc;
    TS_STAMP		start_of_day, midnite;
    unsigned long	time_now;	/* seconds since 00:00:00 of today */
    unsigned long	start_time;
    unsigned long	end_time = 0;
    unsigned long	*ulptr;
    short		totw;
    short		last_outx;
    char		timebuff[32];
    char		*ptimebuff;
    int			i;

    if (pddc -> frst)
    {
	start_time = pddc -> ot02;
	ulptr = &pddc -> ot02;
	i = 1;
	last_outx = 1;
	totw = 1;
    }
    else
    {
	start_time = pddc -> ot01;
	ulptr = &pddc -> ot01;
	i = 0;
	last_outx = 0;
	totw = 0;
    }

    dsddc = pddc;

    for ( ; i < pddc -> topw; i++, totw++)
    {
	if (totw >= dsddc -> numw)
	{
	    totw = 0;
	    dsddc = (struct ddcRecord*)
		    (((struct dbAddr*)(dsddc->dscd.value.db_link.pdbAddr))->precord);
	    ulptr = &dsddc -> ot01;
	}
	end_time = *ulptr;
	last_outx = i;
	ulptr++;
    }

    strcpy (timebuff, "00:00:00");
    ptimebuff = &timebuff[0];
    tsTextToStamp (&start_of_day, &ptimebuff);
    strcpy (timebuff, "23:59:59");
    tsTextToStamp (&midnite, &ptimebuff);
    midnite.secPastEpoch += 1;
	
    tsStampToLocal (timenow, &tsD);
    time_now = (tsD.hours * 3600) + (tsD.minutes * 60) + tsD.seconds;

    if (start_time < end_time)
    {
	/* if current time is outside of time range */
	if ( (time_now < start_time) || (time_now >= end_time) )
	{
	    pddc -> outx = last_outx;
	    if (time_now < start_time)
	    {
		pddc -> notm = start_of_day.secPastEpoch + start_time;
	    }
	    else
	    {
		pddc -> notm = midnite.secPastEpoch + start_time;
	    }
	}
	else
	{
	    compute_abs_step (pddc, time_now);
	}
    }
    else	/* start >= end, spans more than 1 day */
    {
	/* if current time is outside of time range */
	if ( (time_now >= end_time) && (time_now < start_time) )
	{
	    pddc -> outx = last_outx;
	    pddc -> notm = start_of_day.secPastEpoch + start_time;
	}
	else
	{
	    compute_abs_step (pddc, time_now);
	}
    }
    return;
}


/******************************************************
*
*    get the next output time
*    for RELATIVE
*	just return the time
*    for ABSOLUTE
*	added midnite in epics seconds to the
*	time found
*    called from process ()
*    NOTE: since only called if we are PRIMARY, get
*	frst from pddc
*
******************************************************/

static void get_next_output_time (pddc, output_time)
    struct ddcRecord	*pddc;
    unsigned long	*output_time;
{
    struct ddcRecord	*dsddc;
    TS_STAMP		midnite;
    int			i;
    short		totw;
    short		outx;
    short		found_it = 0;
    short		past_midnite = 0;
    short		pmid = 0;
    unsigned long	last_output_time;
    unsigned long	*current_output_time;
    char		timebuff[32];
    char		*ptimebuff;

    dsddc = pddc;
    totw = 0;
    if (pddc -> frst)
    {
	i = 1;
	last_output_time = pddc -> ot02;
	current_output_time = &pddc -> ot02;
    }
    else
    {
	i = 0;
	last_output_time = pddc -> ot01;
	current_output_time = &pddc -> ot01;
    }

    outx = pddc -> outx;

    /* if we are ABSOLUTE timing, then we want to get the
       output time at index (outx) + 1
    */
    if (pddc -> tmod == DDC_ABS_TIMEMODE)
    {
	outx++;
    }

    for (; i < pddc -> topw; i++, totw++)
    {
	if (totw >= dsddc -> numw)
	{
	    totw = 0;
	    dsddc = (struct ddcRecord*)
		    (((struct dbAddr*)(dsddc->dscd.value.db_link.pdbAddr))->precord);
	    current_output_time = &dsddc -> ot01;
	}

	/* did we pass midnite, 2200 -> 0200 on last loop, but if next is 0400
	   then past_midnite is set to 0, but we remember that we passed
	   midnite in pmid if we go off the end of the list of times
	*/
	if (((*(current_output_time)) < last_output_time) &&
	    (dsddc -> tmod == DDC_ABS_TIMEMODE) )
	{
	    past_midnite = 1;
	    pmid = 1;
	}
	else
	    past_midnite = 0;
	if (i == outx)
	{
	    if (dsddc -> tmod == DDC_REL_TIMEMODE)
	    {
		*output_time = *current_output_time;
	    }
	    /* we are ABSOLUTE timing */
	    else
	    {
		/* if we passed midnite, then get end of this
		   day, add 1 second since cant use 24:00:00
		*/
		if (past_midnite)
		{
			strcpy (timebuff, "23:59:59");
			ptimebuff = &timebuff[0];
			tsTextToStamp(&midnite, &ptimebuff);
			midnite.secPastEpoch += 1;
		}
		else
		{
			strcpy (timebuff, "00:00:00");
			ptimebuff = &timebuff[0];
			tsTextToStamp(&midnite, &ptimebuff);
		}
		*output_time = midnite.secPastEpoch + (*current_output_time);
	    }
	    found_it = 1;
	    break;
	}
	last_output_time = *current_output_time;
	current_output_time++;
    }
    /* we have not found the time because we are timing ABSOLUTE and local outx
       was set to 1 past the end, so the next time is the 1st/2nd in
       the list, the only differnece is did we pass midnite in our search,
       this dicatates how much epics time to add to the start time
    */
    if (!found_it)
    {
	if (pmid)	/* did we pass midnite */
	{
	    strcpy (timebuff, "00:00:00");
	    ptimebuff = &timebuff[0];
	    tsTextToStamp(&midnite, &ptimebuff);
	}
	else
	{
	    strcpy (timebuff, "23:59:59");
	    ptimebuff = &timebuff[0];
	    tsTextToStamp(&midnite, &ptimebuff);
	    midnite.secPastEpoch += 1;
	}
	if (pddc -> frst)
	{
	    current_output_time = &pddc -> ot02;
	}
	else
	{
	    current_output_time = &pddc -> ot01;
	}
	*output_time = midnite.secPastEpoch + (*current_output_time);
    }

    return;
}


/********************************************
*
*    return a default precision of 0
*
*********************************************/

static long get_precision (paddr, precision)

    struct dbAddr	*paddr;
    long		*precision;
{
    *precision = 0;
    return (0);
}

/**********************************
*
*    return the block value
*
**********************************/

static long get_value(pddc, pvdes)

    struct ddcRecord	*pddc;
    struct valueDes	*pvdes;
{
    pvdes->field_type = DBF_USHORT;
    pvdes->no_elements = 1;
    (unsigned short *) (pvdes->pvalue) = &pddc->val;
    return (0);
}



/***********************************************
*
*    return limits for certain graphic field
*
***********************************************/

static long get_graphic_double (paddr, pgd)
    struct dbAddr		*paddr;
    struct dbr_grDouble		*pgd;
{
    struct ddcRecord    *pddc = (struct ddcRecord *) paddr->precord;
    short		tot_steps;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct ddcRecord *)0) -> ow01) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow02) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow03) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow04) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow05) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow06) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow07) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow08) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow09) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow10) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow11) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow12) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow13) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow14) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow15) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow16) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow17) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow18) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow19) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow20) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow21) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow22) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow23) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow24) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow25) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow26) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow27) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow28) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow29) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow30) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow31) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow32) ):
	    pgd -> upper_disp_limit = DDC_VAL_UPPER_LIMIT;
	    pgd -> lower_disp_limit = DDC_VAL_LOWER_LIMIT;
	    break;

	case ( (int)&(((struct ddcRecord *)0) -> numb) ):
	    pgd -> upper_disp_limit = DDC_NUMB_UPPER_LIMIT;
	    pgd -> lower_disp_limit = DDC_NUMB_LOWER_LIMIT;
	    break;

	case ( (int)&(((struct ddcRecord *)0) -> numw) ):
	    pgd -> upper_disp_limit = DDC_NUMW_UPPER_LIMIT;
	    pgd -> lower_disp_limit = DDC_NUMW_LOWER_LIMIT;
	    break;

	case ( (int)&(((struct ddcRecord *)0) -> resv) ):
	    pgd -> upper_disp_limit = 32767;
	    pgd -> lower_disp_limit = -32768;
	    break;

	case ( (int)&(((struct ddcRecord *)0) -> mstn) ):
	    get_tot_steps (pddc, &tot_steps);
	    pgd -> upper_disp_limit = tot_steps;
	    pgd -> lower_disp_limit = 1;
	    break;
	
	case ( (int)&(((struct ddcRecord *)0) -> val) ):
	    pgd -> upper_disp_limit = DDC_VAL_UPPER_LIMIT;
	    pgd -> lower_disp_limit = DDC_VAL_LOWER_LIMIT;
	    break;
	
	case ( (int)&(((struct ddcRecord *)0) -> srst) ):
	    pgd -> upper_disp_limit = 1;
	    pgd -> lower_disp_limit = 0;
	    break;

    }
    return (0);
}


/*******************************************
*
*    return field limits to a graphic
*
*******************************************/

static long get_control_double (paddr, pcd)
    struct dbAddr		*paddr;
    struct dbr_ctrlDouble	*pcd;
{
    struct ddcRecord	*pddc = (struct ddcRecord *) paddr->precord;
    short		tot_steps;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct ddcRecord *)0) -> ow01) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow02) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow03) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow04) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow05) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow06) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow07) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow08) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow09) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow10) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow11) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow12) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow13) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow14) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow15) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow16) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow17) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow18) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow19) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow20) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow21) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow22) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow23) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow24) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow25) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow26) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow27) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow28) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow29) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow30) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow31) ):
	case ( (int)&(((struct ddcRecord *)0) -> ow32) ):
	    pcd -> upper_ctrl_limit = DDC_VAL_UPPER_LIMIT;
	    pcd -> lower_ctrl_limit = DDC_VAL_LOWER_LIMIT;
	    break;

	case ( (int)&(((struct ddcRecord *)0) -> numb) ):
	    pcd -> upper_ctrl_limit = DDC_NUMB_UPPER_LIMIT;
	    pcd -> lower_ctrl_limit = DDC_NUMB_LOWER_LIMIT;
	    break;

	case ( (int)&(((struct ddcRecord *)0) -> numw) ):
	    pcd -> upper_ctrl_limit = DDC_NUMW_UPPER_LIMIT;
	    pcd -> lower_ctrl_limit = DDC_NUMW_LOWER_LIMIT;
	    break;

	case ( (int)&(((struct ddcRecord *)0) -> resv) ):
	    pcd -> upper_ctrl_limit = 32767;
	    pcd -> lower_ctrl_limit = -32768;
	    break;

	case ( (int)&(((struct ddcRecord *)0) -> mstn) ):
	    get_tot_steps (pddc, &tot_steps);
	    pcd -> upper_ctrl_limit = tot_steps;
	    pcd -> lower_ctrl_limit = 1;
	    break;
	
	case ( (int)&(((struct ddcRecord *)0) -> val) ):
	    pcd -> upper_ctrl_limit = DDC_VAL_UPPER_LIMIT;
	    pcd -> lower_ctrl_limit = DDC_VAL_LOWER_LIMIT;
	    break;
	
	case ( (int)&(((struct ddcRecord *)0) -> srst) ):
	    pcd -> upper_ctrl_limit = 1;
	    pcd -> lower_ctrl_limit = 0;
	    break;
    }
    return (0);
}


/*************************************
*
*    get the total number of steps
*    called fro get_control_double ()
*    and get_display_double ()
*
*************************************/

static void get_tot_steps (pddc, tot_steps)
    struct ddcRecord	*pddc;
    short		*tot_steps;
{
    struct ddcRecord	*dsddc;

    dsddc = pddc;
    *tot_steps = 0;
    while (dsddc)
    {
	*tot_steps += dsddc -> numw;
	if (dsddc -> dscd.type == DB_LINK)
	{
	    dsddc = (struct ddcRecord*)
		    (((struct dbAddr*)(dsddc->dscd.value.db_link.pdbAddr))->precord);
	}
	else
	    dsddc = 0;
    }
    return;
}


/******************************************
*
*    handle alarm, val and trend monitors
*
******************************************/

static void monitor (pddc)
    struct ddcRecord	*pddc;
{
    unsigned short	monitor_mask;
    short		stat,sevr,nsta,nsev;
    int			set_ack = 1;
	
    /* get previous stat and sevr and new stat and sevr*/

    stat = pddc -> stat;
    sevr = pddc -> sevr;
    nsta = pddc -> nsta;
    nsev = pddc -> nsev;

    /*set current stat and sevr*/

    pddc -> stat = nsta;
    pddc -> sevr = nsev;
    pddc -> nsta = 0;
    pddc -> nsev = 0;

    /* if we are coming out of VALID_ALARM and our new stat/sevr are
       the same as when we went into VALID_ALARM, do not set ack
    */
    if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	 (nsta == pddc -> psta) && (nsev == pddc -> psev) )
    {
	set_ack = 0;
    }
    /* if we are going into VALID_ALARM, and out previous stat was not
       VALID_ALARM, remember previous non VALID_ALARM stat/sevr
       and we do not want to set ACK,
    */
    if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
    {
	set_ack = 0;
	pddc -> psta = stat;
	pddc -> psev = sevr;
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> psta, DBE_VALUE);
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> psev, DBE_VALUE);
    }

    monitor_mask = 0;
	

    /* alarm condition changed this scan */

    if ((stat != nsta) || (sevr != nsev)) 
    {
	/* post events for alarm condition change*/

	monitor_mask = DBE_ALARM;

	/* when alarm status chg's, get current time and
	   save VAL in lalm (last value when alarm chg'd)
	*/
	tsLocalTime (&pddc -> latm);
	pddc -> lalm = pddc -> val;

	/* post stat and nsev fields */

	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> stat, DBE_VALUE);
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> sevr, DBE_VALUE);
	if ( (pddc -> sevr != NO_ALARM) && (set_ack) &&
	     (pddc -> sevr != EVENT_ALARM) )
	{
	    if (pddc -> ack == 0) 
	    {
		pddc -> ack = 1;           /* new unacknowledged alarm */
		db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ack, DBE_VALUE);
	    }
	}
    }

    /* check for value change */

    if (pddc -> val != pddc -> mlst) 
    {
	/* post events for value change */

	monitor_mask |= DBE_VALUE;

	/* update last value monitored */

	pddc -> mlst = pddc -> val;
    }

    /* check for archive change if tenb (trending) is enabled */

    if (pddc -> tenb)
    {
	/* if val has chg'd or there has been a change in alarm status */
	if ( (pddc -> val != pddc -> alst) || (monitor_mask & DBE_ALARM) )
	{
	    /* post events on value field for archive change */

	    monitor_mask |= DBE_LOG;

	    /* update last archive value monitored */

	    pddc -> alst = pddc -> val;
	}
    }

    /* send out monitors connected to the value field */

    if (monitor_mask)
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> val, monitor_mask);

    return;
}


/***************************
*
*    process a DDC block
*
***************************/

static long process (paddr)
    struct dbAddr	*paddr;
{
    struct ddcRecord	*pddc   = (struct ddcRecord *) (paddr -> precord);
    struct ddcdset	*pdset = (struct ddcdset *) (pddc -> dset);
    long		status = 0;
    unsigned short	stat   = pddc -> stat;
    unsigned short	sevr   = pddc -> sevr;
    short		val    = pddc -> val;
    double		lalm   = pddc -> lalm;
    TS_STAMP		latm   = pddc -> latm;
    double		mlst   = pddc -> mlst;
    short		outx   = pddc -> outx;
    unsigned short	psin   = pddc -> psin;
    unsigned long	ltck   = pddc -> ltck;
    unsigned short	lovl   = pddc -> lovl;
    short		lrst   = pddc -> lrst;
    short		ldtm   = pddc -> ldtm;
    unsigned short	lmod   = pddc -> lmod;
    short		mstn   = pddc -> mstn;
    short		sstp   = pddc -> sstp;
    int			disable_timing  = 0;
    int			use_val         = 0;
    int			reset_flag      = 0;
    int			step_trans_flag = 0;
    unsigned short	step_input      = 0;
    long		options         = 0;
    long		nRequests       = 1;
    short		reset_value     = 0;
    short		primary_outx;
    unsigned long	ctck;
    unsigned long	delta_ticks;
    unsigned long	output_time;
    unsigned short	mode;
    char		redbuf[512];

    /* if not init'd, do nothing */

    if ( (!pddc -> init) || (!pddc -> ini2) )
	return (status);

    /* nothing to process if we are a downstream block
       since all we hold is data for either a PRIMARY
       or SECONDARY DDC
    */
    if (pddc -> dcty == DDC_DOWNSTREAM)
	return (status);

    if ((pdset == NULL) || (pdset -> write_ddc == NULL)) 
    {
	pddc -> pact = TRUE;
	recGblRecordError (S_dev_missingSup, (struct dbCommon *)pddc,"write_ddc");
	return (S_dev_missingSup);
    }

    /* get time stamp for last time record was processed */

    tsLocalTime (&pddc -> lptm);

    pddc->msev = NO_ALARM;

    ctck = tickGet ();
    delta_ticks = ctck - pddc -> ltck;
    pddc -> ltck = ctck;

    /* off normal operations (use VAL and use RVAL) */

    if (pddc -> tagd)
    {
	pddc -> nsta = TAGGED_ALARM;
	pddc -> nsev = VALID_ALARM;
    }
    else if (pddc -> oscn == OFF_SCAN)
    {
	pddc -> nsta = SCAN_ALARM;
	pddc -> nsev = VALID_ALARM;	
    }
    else
    {
	pddc -> nsta = NO_ALARM;
	pddc -> nsev = NO_ALARM;

	/* update number of process times */

	if (process_stats)
	    process_cnts[type]++;

	pddc -> pact = TRUE;

	if (pddc -> dcty == DDC_PRIMARY)
	{
	    mode = pddc -> mode;
	    if (pddc -> spin.type == DB_LINK)
	    {
		status = dbGetLink (&pddc -> spin.value.db_link, 
				    (struct dbCommon *)pddc,
				    DBR_USHORT, (caddr_t)&step_input,
				    &options, &nRequests);
		if (status)
		{
		    if (pddc -> nsev < VALID_ALARM)
		    {
			pddc -> nsev = VALID_ALARM;
			pddc -> nsta = LINK_ALARM;
		    }
		}
		else
		{
		    if (step_input != pddc -> psin)
		    {
			if (pddc -> omod == DDC_STEPPED_OPMODE)
			{
			    if ( ((pddc -> psin == 0) && 
				  (pddc -> lted == DDC_LEADING_EDGE)) ||
				 ((pddc -> psin == 1) &&
				  (pddc -> lted == DDC_TRAILING_EDGE)) )
			    {
				step_trans_flag = 1;
			    }
			}
		    }
		    pddc -> psin = step_input;
		}
	    }
	    /* handle the supervisory step command SSTP */
	    if ( (pddc -> spin.type != DB_LINK) || (pddc -> omsl == SUPERVISORY) )
	    {
		if (pddc -> sstp != 0)
		{
		    step_trans_flag = 1;
		}
		else
		{
		    step_trans_flag = 0;
		}
	    }
	    if (pddc -> omod == DDC_STEPPED_OPMODE)
		pddc -> sstp = 0;
	}
	else	/* are secondary */
	{
	    /* get my mode from the primary */
	    mode  = ((struct ddcRecord*)
		     (((struct dbAddr*)(pddc->prcd.value.db_link.pdbAddr))->precord))->mode;
	}  /* end SECONADRY */

	if (mode == DDC_AUTOMATIC)
	{
	    if (pddc -> dcty == DDC_PRIMARY)
	    {
		if ( (pddc -> rsln.type == DB_LINK) &&
		     (pddc -> omsl == CLOSED_LOOP) )
		{
		    status = dbGetLink (&pddc -> rsln.value.db_link, 
				        (struct dbCommon *)pddc,
					DBR_SHORT, (caddr_t)&reset_value,
					&options, &nRequests);
		    if (status)
		    {
			if (pddc -> nsev < VALID_ALARM)
			{
			    pddc -> nsev = VALID_ALARM;
			    pddc -> nsta = LINK_ALARM;
			}
		    }
		    else
		    {
			if (reset_value == pddc -> resv)
			{
			    reset_flag = 1;
			}
		    }
		}
		else
		{
		    if (pddc -> srst)
		    {
			reset_flag = 1;
		    }
		}
		if (reset_flag)
		{
		    pddc -> outx = 0;
		    if (outx != pddc -> outx)
		    {
			do_drum_io (pddc, use_val);
		    }
		}
		else
		{
		    if (pddc -> omod == DDC_STEPPED_OPMODE)
		    {
			if (step_trans_flag)
			{
			    set_next_outx (pddc);
			    do_drum_io (pddc, use_val);
			}
		    }
		    else  /* TIMED */
		    {
			/* for timed mode SPIN (step input) can be
			   used as a permissive, under the following
			   conditions
			*/
			if ( (pddc -> spin.type == DB_LINK) &&
			     (pddc -> omsl == CLOSED_LOOP) &&
			     (pddc -> psin == 0) )
			{
			    disable_timing = 1;
			}
			/* if SPIN is not defined, but we are in
			   SUPERVISORY, then check on SSTP, 
			   supervisory step increase
			*/
			if ( (pddc -> spin.type != DB_LINK) &&
			     (pddc -> omsl == SUPERVISORY) )
			{
			    if (pddc -> sstp > 0)
			    {
				disable_timing = 0;
			    }
			}
			if (!disable_timing)
			{
			    if (pddc -> tmod == DDC_REL_TIMEMODE)
			    {
				/* if coming out of reset, get RELATIVE
				   time to stay in step 1 (outx 0) if we are in
				   step 1, else get the time for which ever step
				   we are in
				*/
				if ( (reset_flag == 0) && (pddc -> lrst != 0) )
				{
				    if (pddc -> outx == 0)
				    {
					pddc -> notm = pddc -> ltck + pddc -> ot01;
				    }
				    else
				    {
					get_next_output_time (pddc, &output_time);
					pddc -> notm = pddc -> ltck + output_time;
				    }
				}
				/* how long to stay in step we were left in in
				   MANUAL mode
				*/
				if (pddc -> lmod != mode)
				{
				    if (pddc -> outx == 0)
				    {
					pddc -> notm = pddc -> ltck + pddc -> ot01;
				    }
				    else
				    {
					get_next_output_time (pddc, &output_time);
					pddc -> notm = pddc -> ltck + output_time;
				    }
				}
				/* should we go to the next step */
				if (pddc -> ltck >= pddc -> notm)
				{
				    /* go to next step, output its value, and
				       calculate time to stay in this step
				    */
				    set_next_outx (pddc);
				    do_drum_io (pddc, use_val);
				    get_next_output_time (pddc, &output_time);
				    pddc -> notm = pddc -> ltck + output_time;
				}
			    }
			    else  /* we are ABSOLUTE time */
			    {
				/* if coming out of reset, disable_timing, or
				   manual mode
				*/
				if ( ((reset_flag == 0) && (pddc -> lrst != 0)) ||
				     ((pddc -> ldtm != 0) && (!disable_timing)) ||
				     (pddc -> lmod != pddc -> mode) )
				{
				    find_start_abs_outx (pddc, pddc -> lptm);
				    do_drum_io (pddc, use_val);
				}
				/* time to step, output value at current step
				   and calc time to next step
				*/
				if (pddc -> lptm.secPastEpoch >= pddc -> notm)
				{
				    set_next_outx (pddc);
				    do_drum_io (pddc, use_val);
				    get_next_output_time (pddc, &output_time);
				    pddc -> notm = output_time;
				}
			    }
			}
			/* we are disable timing, if we are RELATIVE
			   then update NOTM, we are holding the timer
			*/
			else
			{
			    if (pddc -> tmod == DDC_REL_TIMEMODE)
			    {
				pddc -> notm += delta_ticks;
			    }
			}
		    }  /* end TIMED */

		}  /* end !reset */

	    }  /* end AUTOMATIC/PRIMARY */

	    else   /* we are AUTOMATIC and SECONDARY */
	    {
		primary_outx = ((struct ddcRecord*)
				(((struct dbAddr*)(pddc -> prcd.value.db_link.pdbAddr))->precord))->outx;
		if (primary_outx != pddc -> outx)
		{
		    pddc -> outx = primary_outx;
		    do_drum_io (pddc, use_val);
		}
	    }
	}  /* end AUTOMATIC mode */

	else	/* we are in MANUAL mode */
	{
	    if (pddc -> dcty == DDC_PRIMARY)
	    {
		/* if out of limit, set to -1 to cause to post */
		if ( (pddc -> mstn < 1) ||
		     (pddc -> mstn > pddc -> topw) )
		{
		    pddc -> mstn = -1;
		}
		else
		{
		    if ((pddc -> mstn - 1) != pddc -> outx)
		    {
			pddc -> outx = pddc -> mstn - 1;
			do_drum_io (pddc, use_val);
		    }
		}
		/* if VAL is out of 0-255 range */
		if (pddc -> val & 0xff00)
		{
		    pddc -> val = pddc -> lovl;
		}
		if (pddc -> val != pddc -> lovl)
		{
		    use_val = 1;
		    do_drum_io (pddc, use_val);
		}
	    }
	    else  /* we are MANUAL/SECONDARY */
	    {
		primary_outx = ((struct ddcRecord*)
				(((struct dbAddr*)(pddc -> prcd.value.db_link.pdbAddr))->precord))->outx;
		if (primary_outx != pddc -> outx)
		{
		    pddc -> outx = primary_outx;
		    do_drum_io (pddc, use_val);
		}
		/* if VAL is out of 0-255 range */
		if (pddc -> val & 0xff00)
		{
		    pddc -> val = pddc -> lovl;
		}
		if (pddc -> val != pddc -> lovl)
		{
		    use_val = 1;
		    do_drum_io (pddc, use_val);
		}
	    }
	}	/* end MANUAL */

	if (pddc -> dcty == DDC_PRIMARY)
	{
	    pddc -> mstn = pddc -> outx + 1;
	    pddc -> lmod = mode;
	    pddc -> ldtm = disable_timing;
	    pddc -> lrst = reset_flag;
	}

	alarm (pddc);

    }	/* end ON_SCAN */

    /* if the time stamp is 0, init and post val with
       alarm bit set, this makes sure that proper alarm
       messages are set if IOC was down and alarm status
       may have chg'd
    */
    if (pddc -> latm.secPastEpoch == 0)
    {
	tsLocalTime (&pddc -> latm);
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> val, DBE_VALUE | DBE_ALARM);
    }

    /* if it is the 1st process pass 
       init last value when alarm chg'd (lalm)
    */

    if (pddc -> fpps)
    {
	pddc -> fpps = 0;
	pddc -> lalm = pddc -> val;
    }

    /* check event list */

    monitor (pddc);

    redQInit ((struct dbCommon *)pddc, redbuf, sizeof (redbuf), type);

    if (stat != pddc->stat) 
	redQField (redbuf, mylist [STAT]);
	
    if (sevr != pddc->sevr) 
	redQField (redbuf, mylist [SEVR]);
	
    if (val != pddc->val) 
    {
	redQField (redbuf, mylist [VAL]);
    }
	
    if (lalm != pddc->lalm) 
	redQField (redbuf, mylist [LALM]);

    if (TsCmpStampsNE (&latm, &pddc -> latm) )
	redQField (redbuf, mylist [LATM]);

    if (mlst != pddc -> mlst)
	redQField (redbuf, mylist [MLST]);

    if (outx != pddc -> outx)
    {
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> outx, DBE_VALUE);
	redQField (redbuf, mylist [OUTX]);
    }

    if (psin != pddc -> psin)
    {
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> psin, DBE_VALUE);
	redQField (redbuf, mylist [PSIN]);
    }

    if (ltck != pddc -> ltck)
	redQField (redbuf, mylist [LTCK]);

    if (lovl != pddc -> lovl)
	redQField (redbuf, mylist [LOVL]);

    if (lrst != pddc -> lrst)
	redQField (redbuf, mylist [LRST]);

    if (ldtm != pddc -> ldtm)
	redQField (redbuf, mylist [LDTM]);

    if (lmod != pddc -> lmod)
	redQField (redbuf, mylist [LMOD]);

    if (mstn != pddc -> mstn)
    {
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> mstn, DBE_VALUE);
	redQField (redbuf, mylist [MSTN]);
    }

    redQSend (redbuf);
    
    if ( (val != pddc -> val) || (lovl != pddc -> lovl) )
    {
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ob01, DBE_VALUE);
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ob02, DBE_VALUE);
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ob03, DBE_VALUE);
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ob04, DBE_VALUE);
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ob05, DBE_VALUE);
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ob06, DBE_VALUE);
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ob07, DBE_VALUE);
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> ob08, DBE_VALUE);
    }

    if (sstp != pddc -> sstp)
	db_post_events ((struct dbCommon *)pddc, (caddr_t)&pddc -> sstp, DBE_VALUE);

    /* process the forward scan link record */

    if (pddc -> flnk.type == DB_LINK) 
	dbScanPassive ((struct dbAddr *)pddc -> flnk.value.db_link.pdbAddr);

    pddc->pact = FALSE;
	
    return (status);
}


/***********************************************
*
*    adjust the output index OUTX
*    called from process ()
*    NOTE: since only called if we are PRIMARY, 
*	get frst from pddc
*
***********************************************/

static void set_next_outx (pddc)
    struct ddcRecord	*pddc;
{
    pddc -> outx++;

    /* if we are at the end, reset back to start,
       but take intoaccount that the 1st index
       is used for reset only
    */
    if (pddc -> outx >= pddc -> topw)
    {
	if (pddc -> frst)
	{
	    pddc -> outx = 1;
	}
	else
	{
	    pddc -> outx = 0;
	}
    }
    return;
}


/**********************
*
*    handle specials
*
**********************/

static long special (paddr, after)
    struct	dbAddr	*paddr;
    int			after;
{
    struct ddcRecord  		*pddc = (struct ddcRecord *)(paddr -> precord);
    int				special_type = paddr -> special;
    long			status = 0;
    static char			time_string[sizeof (pddc -> ts01)];


    /* we have been called before the new time string
       has been placed in the record, we will save the 
       original in case the new one is in error
    */
    if (!after)
    {
	if ( (pddc -> dcty == DDC_PRIMARY) &&
	     (special_type == SPC_DDC_TSXX) )
	    special_convert_time (paddr, time_string, after);
    }

    switch (special_type) 
    {
	case SPC_DDC_VAL:	/* split val to OB01-OB08 */
	    bit_split_val (pddc);
	    break;

	case SPC_DDC_LTCK:	/* have SLAVE get its own last tick count */
	    pddc -> ltck = tickGet ();
	    break;

	case SPC_DDC_TSXX:
	    if (pddc -> dcty == DDC_PRIMARY)
	    {
		if (special_convert_time (paddr, time_string, after))
		    status = S_db_badField;
	    }
	    break;

	default:
	    recGblDbaddrError (S_db_badChoice, paddr, "ddc: special");
	    status = S_db_badChoice;
	    break;
    }

    return (status);
}


/*****************************************************************
*
*    called from special () to convert a new time string into an
*    output time, but restore original values if the new string
*    has a bad format
*
*****************************************************************/

static int special_convert_time (paddr, time_string, after)
    struct dbAddr	*paddr;
    char		*time_string;
    int			after;
{
    struct ddcRecord	*pddc = (struct ddcRecord *)(paddr -> precord);
    int			offset_time_start;
    int			offset_time_string;
    int			offset_time_diff;
    unsigned long	*time_value;
    char		errmsg[DDC_ERR_MSG];
    char		emsg[DDC_ERR_MSG];

    offset_time_start = ((int)&(((struct ddcRecord *)0) -> ts01));
    offset_time_string = ((struct fldDes *)paddr -> pfldDes) -> offset;
    offset_time_diff = offset_time_string - offset_time_start;
    time_value = (&pddc -> ot01) + (offset_time_diff / sizeof (pddc -> ts01));

    /* save the value of the original string before the new 
       time string is placed in the record
    */
    if (!after)
    {
	strcpy (time_string, (&pddc -> ts01[0]) + offset_time_diff);
    }
    /* the new string has been placed in the record, convert it to
       a ulong, but if the time string is bad restore the original
       value of the string
    */
    else
    {
	if (convert_time (pddc, (char*)(&pddc -> ts01[0] + offset_time_diff),
			  time_value, emsg))
	{
	    strcpy ((char*)(&pddc -> ts01[0] + offset_time_diff), time_string);
	    db_post_events ((struct dbCommon *)pddc, (&pddc -> ts01[0] + offset_time_diff), DBE_VALUE);
	    sprintf (errmsg, "ddc:special, %s", emsg);
	    recGblRecordError (S_db_badField, (struct dbCommon *)pddc, errmsg);
	    return (1);
	}
    }

    return (0);
}


