/*recLeadlag.c,v 1.1.1.1 1999/08/03 14:44:19 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
  
/* recLeadlag.c - Record Support Routines for Leadlag records */
/*
**	Author:		John L. Dalesio
*/
/*	Date:		4/24/93
 *
 * Modification Log
 * ----------------
 * .01 01-03-94		ems	set ack to 0 in init_record
 *				in process, if not init'd do nothing
 * .02 01-05-94		ems	in monitor check for VALID_ALARM before
 *				set ACK
 * .03  03-01-94	ems	added support for process stats
 *				removed ack = 0
 * .04	03-21-94	ems	removed LLMD,OUT0,DI from redundancy,
 *				reworked leadlag()
 * .05	03-29-94	ems	make forward func def agree with actual func def
 * .06	05-10-94	ems	rewrite init_record, set nsta&nsev=NO_ALARM in 
 *				process(), not in leadlag()
 * .07	06-05-94	ems	added special() to handle tim0
 * .08	06-13-94	ems	added support fo rmaximize severity (msev)
 *				move call to tsLocalTime inside on scan
 * .09	07-12-94	ems	in leadlag, if timedif == 0, return
 * .10  08-08-94	ems	added field types to get_graphic_double and
 *				get_control_double
 * .11	08-18-94	ems	monitor(), rm DBE_LOG from ack
 *				only archive val if tenb (trending) is enabled
 * .12	10-13-94	ems	add TIME to redundancy, added fpps
 *				(1st process pass), do time stamp when stat or sevr chg,
 *				see monitor(),and not in process(),added lalm,
 *				set lalm to val when stat or sevr chg,in process() if 
 *				fpps (first process) init lalm to val and if time stamp 
 *				is 0, init it also
 * .13	10-20-94	ems	chg time to latm (last alarm time) added 
 *				added lptm (last process time)
 * .14	12-20-94	ems	removed lalm and fpps, not needed
 * .15	03-15-95	ems	monitor(), do not set ACK if sevr is EVENT_ALARM
 * .16	09-11-95	ems	monitor(), archive when stat/sevr chg's also
 * .17	09-12-95	ems	monitor(), see mod for handling VALID_ALARM
 */

#include	<vxWorks.h>
#include	<inetLib.h>
#include        <tickLib.h>    /* for tickGet () */
#include	<types.h>
#include	<stdioLib.h>
#include	<lstLib.h>
#include	<math.h>
#include	<string.h>
#include	<memLib.h>

#include	"alarm.h"
#include	"cvtTable.h"
#include	"dbDefs.h"
#include	"dbCommon.h"
#include	"link.h"
#include	"db_field_log.h"
#include	"dbAccess.h"
#include	"dbFldTypes.h"
#include	"dbRecType.h"
#include	"devSup.h"
#include	"errMdef.h"
#include	"recSup.h"
#include	"special.h"
#include	"rec/leadlagRecord.h"
#include	"redflds.h"
#include	"redque.h"
#include	"leadlagredfldidx.h"
#include	"restable.h"
#include	"db_addr.h"
#include	"recGbl.h"
#include	"dbEvent.h"
#include	"okToUse.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table*/

#define report NULL
#define initialize NULL
static long init_record();
static long process();
static long special();
static long get_value();
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
static long get_units();
static long get_precision();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double();
static long get_control_double();
#define get_alarm_double NULL

struct rset leadlagRSET = {

	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double
	};


/* The folowing definitions are for redundancy */
 
static struct fldDes	**mylist = 0;

static	char *namelist [] = {
		"STAT",
		"SEVR",
		"VAL",
		"ILST",
		"TIM0",
		"INI2",
		"LATM",
		"MLST",
		0
		};

#define STABLE		0
#define LEADINC		1
#define	LEADDEC		2
#define	LAGDEC		3
#define	LAGINC		4


static	unsigned	short	type;

static void clamp_ll ();
static long leadlag ();
static void monitor ();


/*------------------------------------------------------------------------*/

static long init_record (pll)

    struct leadlagRecord	*pll;

{
    double		value;
    long		options, nRequest;
    unsigned	short	i;
    char		*pstr;
    char		redbuf [256];
    int			ini2_flag = 0;

    if (!(mylist)) 
    {
	for (type = 1; type < 100; type++) 
	{
	    pstr = GET_PRECTYPE (type);
	    if (strncmp (pstr, "leadlag", 7) == 0)
		break;
	    
	}
	mylist = redFldAlloc (type, NUMLLFLDS);

	if (!mylist)
	{
	    recGblRecordError (S_rec_Alloc, (struct dbCommon *)pll,
				   "leadlag, init_record, mylist failed");
	    return (S_rec_Alloc);
	}

	for (i = 0; i < NUMLLFLDS; i++) 
	    redFldAdd (type, i, namelist [i]);
    }

    pll -> init = TRUE;
    pll -> nsta = NO_ALARM;
    pll -> nsev = NO_ALARM;
    pll -> msev = NO_ALARM;

    /* init last process time and last alarm time */

    if (pll->lptm.secPastEpoch == 0)
    {
	tsLocalTime (&pll->lptm);
    }
    if (pll->latm.secPastEpoch == 0)
    {
	tsLocalTime (&pll->latm);
    }
    if (pll -> inp.type == DB_LINK)
    {
	/* if we are tagged, off scan or in manual mode we will let
	   the block init without getting VAL
	*/
	if ( (!pll -> tagd) && (pll -> oscn != OFF_SCAN) &&
	     (pll -> mode == NORMAL) )
	{
	    options = 0;
	    nRequest = 1;

	    if ( (dbGetLink (&pll->inp.value.db_link, 
			     (struct dbCommon *)pll, DBR_DOUBLE,
		             (caddr_t)&value, 
			     &options, &nRequest) == NULL) &&
		 (okToUse (&pll -> inp)) )
	    {
		pll->val = pll->ilst = value;
		clamp_ll (pll);

		/* fetch the start time */
		pll->tim0 = tickGet();

		ini2_flag = 1;
	    }
	    else
	    {
		pll -> nsta = NO_INIT_ALARM;
		pll -> nsev = VALID_ALARM;
	    }
	}
	else
	    ini2_flag = 1;
    }
    else
    {
	pll -> nsta = NO_INIT_ALARM;
	pll -> nsev = VALID_ALARM;
    }



    if (resGetCaAnswer(pll -> radr, &i) == RES_OK)
    {
	if (i)
	{
	    /* post STAT, SEVR and VAL (monitor), if we will finally init, 
	       because we have read a value from the input, set INI2,
	       send all redund fields to slave, when slave receives INI2, 
	       it will be fully init'd also
	    */

	    /* check for maximize severity */

	    if (pll -> msev > pll -> nsev)
	    {
		pll -> nsev = pll -> msev;
		pll -> nsta = LINK_ALARM;
	    }

	    monitor (pll);

	    if (ini2_flag)
	    {
		pll -> ini2 = 1;

		redQInit ((struct dbCommon *)pll, redbuf, sizeof (redbuf), type);
		redQField (redbuf, mylist [STAT]);
		redQField (redbuf, mylist [SEVR]);
		redQField (redbuf, mylist [VAL]);
		redQField (redbuf, mylist [ILST]);
		redQField (redbuf, mylist [TIM0]);
		redQField (redbuf, mylist [INI2]);
		redQSend (redbuf);
	    }
	}
    }

    return (0);
}


/*------------------------------------------------------------------------*/

static long process (paddr)

    struct dbAddr	*paddr;

    {
	struct leadlagRecord *pll = (struct leadlagRecord *) (paddr->precord);
	char		redbuf [256];

	long		status = 0;
	unsigned short	stat   = pll->stat;
	unsigned short	sevr   = pll->sevr;
	double		val    = pll->val;
	float		ilst   = pll->ilst;
	unsigned long	tim0   = pll->tim0;
	float		mlst   = pll->mlst;
	TS_STAMP	latm   = pll->latm;

	   /* if not init'd do nothing */

	if ((!pll -> init) || (!pll -> ini2))
            return (status);

	/* get time stamp for last time record was processed */

	tsLocalTime (&pll -> lptm);

	/* Initialize the new status to the last */
	pll->nsta = pll->stat;
	pll->nsev = pll->sevr;
	pll->msev = NO_ALARM;

        /* off normal operations (use VAL and use RVAL) */

	if (pll -> tagd)
	{
	    pll->nsta = TAGGED_ALARM;
	    pll->nsev = VALID_ALARM;
	}
        else if (pll->oscn == OFF_SCAN)
	{
            pll->nsta = SCAN_ALARM;
	    pll->nsev = VALID_ALARM;
	}
	else
	{
	    pll -> nsta = NO_ALARM;
	    pll -> nsev = NO_ALARM;

	    /* update number of process times */
	    if (process_stats)
	       process_cnts[type]++;	

	    pll -> pact = TRUE;
	    if (pll -> mode == NORMAL) 
		status = leadlag(pll);

	    /* check for maximize severity */

	    if (pll -> msev > pll -> nsev)
	    {
		pll -> nsev = pll -> msev;
		pll -> nsta = LINK_ALARM;
	    }
	}

	/* if the time stamp is 0, init and post val with
	   alarm bit set, this makes sure that proper alarm
	   messages are set if IOC was down and alarm status
	   may have chg'd
	*/
	if (pll -> latm.secPastEpoch == 0)
	{
	    tsLocalTime (&pll -> latm);
	    db_post_events ((struct dbCommon *)pll, (caddr_t)&pll -> val, DBE_VALUE | DBE_ALARM);
	}

	/* check event list */

	monitor (pll);

	pll->pact = FALSE;

	redQInit ((struct dbCommon *)pll, redbuf, sizeof (redbuf), type);

	if (stat != pll->stat) 
	    redQField (redbuf, mylist [STAT]);
	
	if (sevr != pll->sevr) 
	    redQField (redbuf, mylist [SEVR]);
	
	if (val != pll->val) 
	    redQField (redbuf, mylist [VAL]);
	
	if (ilst != pll->ilst) 
	    redQField (redbuf, mylist [ILST]);
	
	if (tim0 != pll->tim0) 
	    redQField (redbuf, mylist [TIM0]);
	
	if (TsCmpStampsNE (&latm, &pll -> latm) )
	    redQField (redbuf, mylist [LATM]);
	
	if (mlst != pll->mlst)
	    redQField (redbuf, mylist [MLST]);

	redQSend (redbuf);


	/* process the forward scan link record */

	if (pll->flnk.type == DB_LINK) 
	    dbScanPassive ((struct dbAddr *)pll->flnk.value.db_link.pdbAddr);
	
	return (status);
}


/*------------------------------------------------------------------------*/

/*
**	Lead Lag Algotithm
*/
static long leadlag(struct leadlagRecord *pll)

    {
	long		status;
	float		timedif, gain;
	unsigned long	time;
	double		ivalue, ioabsdif, leadt, lagt, temp;
	long		options, nRequest;


	options = 0;
	nRequest = 1;
	status = dbGetLink (&pll->inp.value.db_link, 
			    (struct dbCommon *)pll, DBR_DOUBLE,
		            (caddr_t)&ivalue, 
			    &options, &nRequest);
	if (status != 0) 
	{
	    pll->nsev  = VALID_ALARM;
	    pll->nsta  = LINK_ALARM;
	    return(status);
	}

	/* fetch the start time */
	time = tickGet();

	/* Get the seconds of time difference since started moving*/
	timedif = (((float)time) - ((float)pll->tim0)) / 60.0;

	pll->tim0 = time;

	ioabsdif = (pll->val - ivalue);
	if (ioabsdif < 0)
	    ioabsdif = -ioabsdif;

	/* If close enough just set it */
	if (ioabsdif < pll->nois) 
	{
	    pll->val = pll->ilst = ivalue;
	    clamp_ll (pll);
	    return(0);
	}

	/* if not time has elapsed (< 1 tick), no need to go on */
	if (timedif == 0)
	    return (0);

	ioabsdif = ivalue - pll -> ilst;
	if (ioabsdif < 0)
	    ioabsdif = -ioabsdif;
	
	if ((pll -> lead == 0) || (ioabsdif < pll -> nois))
	    gain = 0;
	else
	    gain = pll -> gain;

	temp = ivalue - pll -> val;

	leadt = pll -> lead / timedif;
	lagt  = pll -> lag / timedif;

	if ((pll -> lead != 0) || (pll -> lag != 0))
	{
	    if (pll -> lead != 0)
	    {
		pll -> val = pll -> val + (gain * (ivalue - pll -> ilst));
		if ((gain == 0) || (pll -> nois <= 0.0))
		    pll -> val = pll -> val + (temp / exp (leadt));
	    }
	    if (pll -> lag != 0)
		pll -> val = pll -> val + (temp / exp (lagt));
	}
	else
	    pll -> val = ivalue;

	/* Set the last value to this value */
	pll->ilst = ivalue;

	clamp_ll (pll);

	return(0);
    }


/*------------------------------------------------------------------------*/

static void clamp_ll (pll)
    struct leadlagRecord    *pll;
{
    if (pll -> val > pll -> hopr)
	pll -> val = pll -> hopr;
    else if (pll -> val < pll -> lopr)
	pll -> val = pll -> lopr;
    return;
}

/*------------------------------------------------------------------------*/

static long get_value(pll,pvdes)

    struct	leadlagRecord	*pll;
    struct	valueDes	*pvdes;
{
	pvdes->field_type = DBF_DOUBLE;
	pvdes->no_elements = 1;
	(double *) (pvdes->pvalue) = &pll->val;
	return (0);
}

/*------------------------------------------------------------------------*/

static long get_units(paddr,units)

    struct dbAddr *paddr;
    char	  *units;
{
    struct	leadlagRecord	*pll = (struct leadlagRecord *) paddr->precord;

	strncpy (units, pll->egu, sizeof (pll->egu));
	return (0);
}

/*------------------------------------------------------------------------*/

static long get_precision (paddr, precision)

    struct	dbAddr	*paddr;
    long		*precision;
{
	struct leadlagRecord	*pll = (struct leadlagRecord *) paddr->precord;

	*precision = pll->prec;
	return (0);
}


/*------------------------------------------------------------------------*/

static long get_graphic_double(paddr,pgd)

    struct	dbAddr		*paddr;
    struct	dbr_grDouble	*pgd;
{
    struct leadlagRecord	*pll = (struct leadlagRecord *) paddr->precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct leadlagRecord *)0) -> gain) ):
	case ( (int)&(((struct leadlagRecord *)0) -> nois) ):
	case ( (int)&(((struct leadlagRecord *)0) -> adel) ):
	case ( (int)&(((struct leadlagRecord *)0) -> mdel) ):
	{
	    float    span;

	    span = pll -> hopr - pll -> lopr;
	    if (span < 0.0)
		span = -span;
	    pgd -> upper_disp_limit = span;
	    pgd -> lower_disp_limit = 0;
	    break;
	}

	case ( (int)&(((struct leadlagRecord *)0) -> lead) ):
	case ( (int)&(((struct leadlagRecord *)0) -> lag) ):
	    pgd -> upper_disp_limit = 86400;	/* seconds/day */
	    pgd -> lower_disp_limit = 0;
	    break;

	default:
	    pgd -> upper_disp_limit = pll -> hopr;
	    pgd -> lower_disp_limit = pll -> lopr;
	    break;
    }
    return (0);
}


/*------------------------------------------------------------------------*/

static long get_control_double(paddr,pcd)

    struct	dbAddr		*paddr;
    struct	dbr_ctrlDouble	*pcd;
{
    struct leadlagRecord *pll = (struct leadlagRecord *) paddr->precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct leadlagRecord *)0) -> gain) ):
	case ( (int)&(((struct leadlagRecord *)0) -> nois) ):
	case ( (int)&(((struct leadlagRecord *)0) -> adel) ):
	case ( (int)&(((struct leadlagRecord *)0) -> mdel) ):
	{
	    float    span;

	    span = pll -> hopr - pll -> lopr;
	    if (span < 0.0)
		span = -span;
	    pcd -> upper_ctrl_limit = span;
	    pcd -> lower_ctrl_limit = 0;
	    break;
	}

	case ( (int)&(((struct leadlagRecord *)0) -> lead) ):
	case ( (int)&(((struct leadlagRecord *)0) -> lag) ):
	    pcd -> upper_ctrl_limit = 86400;	/* seconds/day */
	    pcd -> lower_ctrl_limit = 0;
	    break;

	default:
	    pcd -> upper_ctrl_limit = pll -> hopr;
	    pcd -> lower_ctrl_limit = pll -> lopr;
	    break;
    }
    return (0);
}


/*------------------------------------------------------------------------*/

static void monitor (pll)

    struct	leadlagRecord	*pll;
{
	unsigned short	monitor_mask;
	double		delta;
	short		stat, sevr, nsta, nsev;
	int		set_ack = 1;

	/* get previous stat and sevr  and new stat and sevr*/

	stat = pll->stat;
	sevr = pll->sevr;
	nsta = pll->nsta;
	nsev = pll->nsev;

	/*set current stat and sevr*/

	pll->stat = nsta;
	pll->sevr = nsev;
	pll->nsta = 0;
	pll->nsev = 0;

	/* if we are coming out of VALID_ALARM and our new stat/sevr are
	   the same as when we went into VALID_ALARM, do not set ack
	*/
	if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	     (nsta == pll -> psta) && (nsev == pll -> psev) )
	{
	    set_ack = 0;
	}
	/* if we are going into VALID_ALARM, and out previous stat was not
	   VALID_ALARM, remember previous non VALID_ALARM stat/sevr
	   and we do not want to set ACK,
	*/
	if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
	{
	    set_ack = 0;
	    pll -> psta = stat;
	    pll -> psev = sevr;
	    db_post_events ((struct dbCommon *)pll, (caddr_t)&pll -> psta, DBE_VALUE);
	    db_post_events ((struct dbCommon *)pll, (caddr_t)&pll -> psev, DBE_VALUE);
	}

	monitor_mask = 0;

	/* alarm condition changed this scan */

	if ((stat != nsta) || (sevr != nsev)) 
	{
	    /* post events for alarm condition change*/

	    monitor_mask = DBE_ALARM;

	    /* when alarm status chg's, get current time
	    */
	    tsLocalTime (&pll -> latm);

	    /* post stat and nsev fields */

	    db_post_events ((struct dbCommon *)pll, (caddr_t)&pll->stat, DBE_VALUE);
	    db_post_events ((struct dbCommon *)pll, (caddr_t)&pll->sevr, DBE_VALUE);
            if ((pll->sevr != NO_ALARM) && (set_ack) &&
		(pll->sevr != EVENT_ALARM) )
	    {
                if (pll->ack == 0) 
		{
                    pll->ack = 1;           /* new unacknowledged alarm */
                    db_post_events((struct dbCommon *)pll, (caddr_t)&pll->ack, DBE_VALUE);
                }
	    }
	}
	/* check for value change */

	delta = pll->mlst - pll->val;
	if (delta < 0.0) 
	    delta = -delta;
	
	if (delta > pll->mdel) 
	{
	    /* post events for value change */

	    monitor_mask |= DBE_VALUE;

	    /* update last value monitored */

	    pll->mlst = pll->val;
	}

	/* check for archive change if tenb (trending) enabled) */

	if (pll -> tenb)
	{
	    delta = pll->alst - pll->val;
	    if (delta < 0.0) 
	    	delta = -delta;
       
	    /* if exceed archive deadband or there has been a change in alarm status */
	    if ( (delta > pll->adel) || (monitor_mask & DBE_ALARM) )
	    {
	    	/* post events on value field for archive change */

	    	monitor_mask |= DBE_LOG;

	    	/* update last archive value monitored */

	    	pll->alst = pll->val;
	    }
	}

	/* send out monitors connected to the value field */

	if (monitor_mask) 
	    db_post_events ((struct dbCommon *)pll, (caddr_t)&pll->val, monitor_mask);
	
	return;
}


/*------------------------------------------------------------------------*/

static long special (paddr, after)
    struct dbAddr    *paddr;
    int              after;
{
    struct leadlagRecord *pll = (struct leadlagRecord *) (paddr -> precord);
    int			special_type = paddr -> special;

    if (!after)
	return (0);

    switch (special_type)
    {
	case SPC_LL_TIM0:
	    pll -> tim0 = tickGet();
	    break;

	default:
	    recGblDbaddrError (S_db_badChoice, paddr, "leadlag: special");
	    return (S_db_badChoice);
    }

    return (0);
}  

