/*recPid.c,v 1.1.1.1 1999/08/03 14:44:20 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/* recPid.c - Record Support Routines for PID records */
/*
 *      Author: John L. Dalesio
 *      Date:   7-23-93
 *
 *	Tate Integrateed Systems (TIS-4000 v4.02)
 *
 *	Copyright 1993 Tate Integrated Systems
 * Modification Log:
 * -----------------
 * .01  01-03-94	ems	set ack to 0 in init_record
 * .02  01-05-94	ems	in monitor, check VALID_ALARM 
 *				before set ACK
 * .03  01-24-94	ems	big mods to get_graphic_double
 *				and get_control_double, deleted
 *				LSP from namelist, no such field in rec
 * .04  01-25-94	ems	1-deleted def's for alarms, use alarm.h
 *				2-deleted E_IO_INTERRUPT & gts_trigger_counter
 *				  not needed
 *				3-deleted lot of calls for redunadancy in
 *				  init_pidrecord, not needed
 *				4-don't init rec if dbGetLink for output value
 *				  fails or output not init'd
 *				5-don't alarm if no output link in 
 *				  init_pidrecord
 *				6-add SPAN,ZERO to redundancy, create special to
 *				  recalc if EGUF,EGUL change
 *				7-add IACT init action
 *				8-add ramp_setpoint()
 *				9-add setpoint gap
 * .05  02-14-94	ems	added ISTP to redund, update initial setpt (istp)
 *				with setpt (astp) if update (ustp) is true
 * .05	02-28-94	ems	remove ack=0 from init, only db_post_events from
 *				init_record if are MASTER, deleted ALST,MLST from
 *				redundancy, not needed
 * .06  03-01-94	ems	added support for process stats
 * .07  03-15-94	ems	call monitor in init to update VAL, in processPid
 * 				move call for forward link process to end
 * .07	03-23-94	ems	make processPid static
 * .08	03-29-94	ems	make forward func def agree with actual func def
 *				make proper casts in get_control_double, del dopid(),
 *				not used
 * .09	04-24-94	ems	added INI2 to namelist
 * .10	06-13-94	ems	added support for maximize severity (msev)
 *				moved call to tsLocalTime inside on scan
 * .11	07-12-94	ems	init_pidrecord, if eguf-egul == 0, set span = 0
 *				special, same chg
 * .12	08-08-94	ems	added more cases to get_graphic_double,
 *				get_control_double, kp, chg upper from 1000 to 5000
 *				rm DBE_LOG from all except VAL
 * .13	10-12-94	ems	add TIME to redundancy, added fpps
 *				(1st process pass), do time stamp when stat or sevr chg,
 *				see monitor(),and not in process(),set lalm to val when 
 *				stat or sevr chg,in process() if fpps (first process)
 *				init lalm to val and if time stamp is 0, init it also
 * .14	10-20-94	ems	chg time to latm (last alarm time) added 
 *				added lptm (last process time)
 * .15	02-08-95	ems	use MVAL's MDEL and ADEL to for posting and archiving of
 *				MVAL and ASTP, new fields LMMV(last monitored MVAL), LAMV
 *				(last archived MVAL), LMAS(last monitored ASTP) and LAAS
 *				(last archived ASTP), LMMV and LMAS are redundant
 *				For ASTP, only use the MVAL's MDEl and ADEL if the
 *				setpoint is a LINK and we are in closed cascade loop 
 *				because this situation can produce the the most 
 *				flucuations, else post and archive if changed
 * .16	02-15-95	ems	if SPTL is a link (astp gets its value here) and are in
 *				CLOSED CASCADE, user can alter ASTP from graphic, but
 *				text update field does not reflect ASTP value in record
 *				since LMAS (last monitored ASTP) does not chg
 *				special sets LMAS to new ASTP, so when block next 
 *				processes the true ASTP will be posted
 * .17	02-16-95	ems	doPid(), if downstream mode != LINK was doing a 
 *				db_post_events
 *				for MODE, bad move
 * .18	03-15-95	ems	monitor(), do not set ACK if sevr is EVENT_ALARM
 * .19	03-20-95	ems	doPid(), on loss of MV, do not set VAL to HVAL unless
 *				in AUTO, also if in TRACKING continue to track output
 *				on MV loss, new func do_tracking()
 * .20	03-24-95	ems	arg to resGetCaAnswer needs to be a short not int
 * .21	09-11-95	ems	archive when stat/sevr chg's also
 * .22	09-12-95	ems	monitor(), see mod for handling VALID_ALARM
 * .23	10-25-95	ems	added support to chg for ACLO (auto close)
 * .24	04-08-96	ems	added code to support alarm delay
 * .25	06-19-96	ems	added code for alarm disable ADIS
 * .26	07-01-96	ems	added sstp, save setpoint
 */


#include	<vxWorks.h>
#include	<inetLib.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<lstLib.h>
#include	<string.h>
#include	<math.h>
#include	<tickLib.h>

#include	"alarm.h"
#include	"cvtTable.h"
#include	"dbDefs.h"
#include	"dbCommon.h"
#include	"link.h"
#include	"db_field_log.h"
#include	"dbAccess.h"
#include	"dbRecType.h"
#include	"dbFldTypes.h"
#include	"devSup.h"
#include	"errMdef.h"
#include	"recSup.h"
#include	"special.h"
#include	"rec/pidRecord.h"
#include	"redflds.h"
#include	"redque.h"
#include	"pidredfldidx.h"
#include	"restable.h"
#include	"recGbl.h"
#include	"db_addr.h"
#include	"dbEvent.h"
#include	"okToUse.h"
#include	"dbcSub.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table*/

#define report NULL
#define initialize NULL
static long init_pidrecord();
static long processPid();
static long special ();
static long get_value();
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
static long get_units();
static long get_precision();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double();
static long get_control_double();
static long get_alarm_double();

struct rset pidRSET={
	RSETNUMBER,
	report,
	initialize,
	init_pidrecord,
	processPid,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double};

/* The following definitions are required for redundancy */

static struct	fldDes	**mylist = 0;
static unsigned short	pointtypeindex;

/* NOTE: spelling and order of namelist MUST match that in
	 .../share/epicsH/pidredfldidx.h
	 spelling must also match that found in
	 .../share/ascii/pidrecord.ascii
*/
static char	*namelist [] = {
		"STAT",
		"SEVR",
		"ASTP",
		"MVAL",
		"NMV",
		"NSP",
		"CT",
		"ERR",
		"ERSQ",
		"DERR",
		"DVAL",
		"VAL",
		"PVAL",
		"LALM",
		"LMDE",
		"MODE",
		"SPAN",
		"ZERO",
		"AM",
		"USCS",
		"SPT",
		"FGCR",
		"SNDF",
		"GACT",
		"NGAP",
		"CSP",
		"ISTP",
		"INI2",
		"LATM",
		"MLST",
		"LMAS",
		"LMMV",
		"DVCT",
		"HLOC",
		"HLSC",
		0
		};
/*
   NOTE: various defines must agree with values in 
	 .../share/ascii/choiceRec.ascii
*/
#define	CH_NO			0
#define CH_YES			1

#define PID_CASCLOSED		CH_NO
#define PID_CASOPEN		CH_YES

#define PID_NO_REV_INIT		CH_NO
#define	PID_REV_INIT		CH_YES

#define	PID_CAS_EGU		CH_NO
#define PID_CAS_PCT		CH_YES

#define	PID_FWD_ACT		CH_NO
#define	PID_REV_ACT		CH_YES

#define	PID_NO_ERRSQ		CH_NO
#define	PID_ERRSQ		CH_YES

#define	PID_NO_FB_ENABLED	CH_NO
#define	PID_FB_ENABLED		CH_YES

/* NOTE:
	these 3 values must agree with what is
	in .../ascii/choiceRec.ascii for
	REC_PID_AUTOCLOSE
*/
#define	PID_ACLO_MANUAL		0
#define	PID_ACLO_AUTOMATIC	1
#define PID_ACLO_LASTMODE	2

#define	PID_NO_HOLD_ON_LOSS	CH_NO
#define	PID_HOLD_ON_LOSS	CH_YES


#define	PID_INCREMENTAL		1
#define	PID_POSITIONAL		0
#define	PID_MODE_AUTO		0
#define	PID_MODE_MANUAL		1
#define	PID_MODE_TRACKING	2
#define MODE_SUPERVISORY	0
#define	MODE_CLOSED_LOOP	1

#define	CONTROL_VALID_PERCENT	.1

/* defines for the init action flag IACT 
   NOTE: order must agree with that in 
	 .../share/ascii/choiceRec.ascii */

#define	INIT_LAST_MODE		0
#define INIT_MAN_OPEN		1
#define INIT_MAN_CLOSE		2
#define INIT_AUTO_OPEN		3
#define INIT_AUTO_CLOSE		4

/* defines for the clamp mask used for alarm delay */

#define	DO_HIGH_OUTPUT_CLAMP	1
#define	DO_LOW_OUTPUT_CLAMP	2
#define	DO_HIGH_SETPT_CLAMP	4
#define	DO_LOW_SETPT_CLAMP	8


static void alarm (struct pidRecord *, unsigned short);
static void clamp_setpoint (struct pidRecord *, unsigned short *);
static long doPid (struct pidRecord *, unsigned short *);
static void do_tracking (struct pidRecord *, unsigned short *);
static void get_setpoint ();
static void monitor ();
static void ramp_setpoint ();


/******************************* init_pidrecord ********************************/

static long init_pidrecord(ppid)
    struct pidRecord	*ppid;
{
    long            options,nRequest;
    float           oval;
    int		    i;
    short	    master;
    float	    mval;
    char	    *pstr;	
    char	    redbuf[400];
    int	            ini2_flag = 0;
    unsigned short  clamp_mask = 0;

    /* Redundant Update Fields */
    unsigned short	am   = ppid -> am;
    unsigned short	uscs = ppid -> uscs;

    /* Connect redundant update fields */
    if (!(mylist)) 
    {
        for (pointtypeindex = 1; pointtypeindex < 100; pointtypeindex++) 
	{
	    pstr = GET_PRECTYPE (pointtypeindex);
	    if (strncmp (pstr, "pid", 3) == 0) 
	        break;
	}
	mylist = redFldAlloc (pointtypeindex, NUMPIDFLDS);

	if (!mylist)
	{
	    recGblRecordError (S_rec_Alloc, (struct dbCommon *)ppid,
			   "pid, init_record, mylist failed");
	    return (S_rec_Alloc);
	}

	for (i = 0; i < NUMPIDFLDS; i++) 
	    redFldAdd (pointtypeindex, i, namelist [i]);
    }

    ppid -> init = TRUE;
    ppid -> fpps = 1;
    ppid -> udf = FALSE;
    ppid -> nsta = NO_ALARM;
    ppid -> nsev = NO_ALARM;
    ppid -> msev = NO_ALARM;

    /* init alarm delay variables */
    ppid -> larf = 0;
    ppid -> hhll = 0;
    ppid -> hilo = 0;
    ppid -> dvct = 0;
    ppid -> rtct = 0;
    ppid -> hloc = 0;
    ppid -> hlsc = 0;

    /* init last process time and last alarm time */

    if (ppid->lptm.secPastEpoch == 0)
    {
	tsLocalTime (&ppid->lptm);
    }
    if (ppid->latm.secPastEpoch == 0)
    {
	tsLocalTime (&ppid->latm);
    }
    /* the measured value must be a link */
    if ((ppid->mvl.type == DB_LINK) ) 
    {
        options=0;
        nRequest=1;

	/* if we can get the measured value from the link, and 
	   the link record is ok to use
	*/
        if( (dbGetLink(&(ppid->mvl.value.db_link),
		       (struct dbCommon *)ppid,
		       DBR_FLOAT, (caddr_t)&mval,
		       &options,&nRequest) == NULL) && 
	    (okToUse(&ppid->mvl)) ) 
	{

	    /* initialize output value */
    	    ppid -> pval = 0;
    	    ppid -> val = 0;

	    /* fetch the output value */
    	    if (ppid -> rout.type == DB_LINK)
	    {
                options = 0;
                nRequest = 1;
                if( (dbGetLink(&(ppid->rout.value.db_link),
			       (struct dbCommon *)ppid,
			       DBR_FLOAT, (caddr_t)&oval,
			       &options,&nRequest) == NULL) && 
		    (okToUse(&ppid->rout) ) )
	        {
	    	    if (ppid -> velo == PID_POSITIONAL)
	    	    {
		        ppid -> pval = (oval /100);	/* previous output value 0 - 1 */
		        ppid -> val = oval;		/*  output value 0 - 100% */
	            }
	        }
	        else
	        {
	            ppid -> nsta = NO_INIT_ALARM;
	            ppid -> nsev = VALID_ALARM;
	        }
	    }
	    if (ppid -> nsta == NO_ALARM)
	    {
    	    	/* set up constants */
		if ((ppid -> eguf - ppid -> egul) == 0)
		    ppid -> span = 0;
		else
    	    	    ppid->span = 1 / (ppid->eguf - ppid->egul);
    	    	ppid->zero = ppid->egul * ppid->span;

    	    	/* normailze setpoint gap, can not go neagtive */
    	    	ppid->ngap = ppid->sptg * ppid -> span;    
    	    	if (ppid -> ngap < 0.0)
		    ppid -> ngap = -ppid -> ngap;

    	    	ppid->astp = ppid->istp;	/* active setpoint = initial setpoint */
	    	ppid->i = 0;			/* last integral contribution */
	    	ppid->err = 0;
	    	ppid->lerr = 0;
	    	ppid->ersq = 0;
	    	ppid->lesq = 0;
	    	ppid->derr = 0;
	    	ppid->lder = 0;
	    	ppid->dval = 0;
	    	ppid->ldvl = 0;
	    	ppid->fgcr = 0;
	    	ppid->gact = 0;
	    	ppid->csp  = 0;

	    	/* Time Management, init current clock ticks */
	    	ppid->ct = tickGet();

	    	/* initial measured value */
	    	ppid->mval = mval;			

	    	/* initial measured value 0 - 1 */
	    	ppid->nmv = (mval * ppid->span) - ppid->zero;

	    	/* init AM (Pid mode selected) and USCS (Cascade select) */
	    	switch (ppid -> iact)
	    	{
	            case INIT_LAST_MODE:
		        break;

		    case INIT_MAN_OPEN:	/* manual with open cascade */
	    	        ppid -> am = PID_MODE_MANUAL;
	    	        ppid -> uscs = PID_CASOPEN;
	    	        break;

		    case INIT_MAN_CLOSE:	/* manula with close cascade */
	    	        ppid -> am = PID_MODE_MANUAL;
	    	        ppid -> uscs = PID_CASCLOSED;
	    	        break;

		    case INIT_AUTO_OPEN:	/* automatic with open cascade */
	    	        ppid -> am = PID_MODE_AUTO;
	    	        ppid -> uscs = PID_CASOPEN;
	    	        break;

		    case INIT_AUTO_CLOSE:	/* automatic with close cascade */
	    	        ppid -> am = PID_MODE_AUTO;
	    	        ppid -> uscs = PID_CASCLOSED;
	    	        break;

		    default:
	    	    {
	                char errmsg [128];
	                sprintf (errmsg, "pid: iact for %d", ppid -> iact);
	                recGblRecordError (S_db_badChoice, (struct dbCommon *)ppid, errmsg);
	                break;
	    	    }
    	        }
	        ppid->lmde = ppid->mode = ppid -> am;

		if (ppid->rini == PID_REV_INIT)
		{
	    	    ppid->astp = mval;
		}
		else
		{
	    	    get_setpoint (ppid);
	    	    clamp_setpoint (ppid, &clamp_mask);
		}
		ppid -> sstp = ppid ->astp;

		/* set point used in setpt ramp */
		ppid->spt = ppid->astp;	

		/* normalize the setpoint  0 - 1 */
		ppid -> nsp = (ppid -> astp * ppid -> span) - ppid -> zero;

		/* Reverse calculate the setpoint we are using to show upstream 
	           when we are out of control 
		*/
		if (ppid->cpct == PID_CAS_PCT) /* Turn the Setpoint to EGU */
	    	    ppid->csp = ((ppid->astp * ppid->span) - ppid->zero) * 100;
		else
	    	    ppid->csp = ppid->astp;
		ini2_flag = 1;
	    }
	}

	/* could not get the measured value, dont't init */
	else
	{
	    ppid -> nsta = NO_INIT_ALARM;
	    ppid -> nsev = VALID_ALARM;
	}
    }

    /* the measured value was not a link, don't init */
    else
    {
	ppid -> nsta = NO_INIT_ALARM;
	ppid -> nsev = VALID_ALARM;
    }


    if (ini2_flag)
    {
        if ((am != ppid -> am) || (uscs != ppid -> uscs))
        {
	    redQInit ((struct dbCommon *)ppid, redbuf, sizeof (redbuf), pointtypeindex);

	       /* these are both prompted fields, recalc checksum
	          do it here and not in redund update below, if we are
	          MASTER then this will handle redundant update, but
	          we must let dbClient know of the changes to these 
	          prompted fields regardless whether we are MASTER
	       */

	    ppid -> rcrc = DBC_CRCRecord ((char *)ppid, pointtypeindex);
	    if (am != ppid -> am)
	    {
	        redQField (redbuf, mylist [AM]);
	        db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid -> am, DBE_VALUE);
	    }
            if (uscs != ppid -> uscs)
            {
	        redQField (redbuf, mylist [USCS]);
	        db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid -> uscs, DBE_VALUE);
            }
            redQSend (redbuf);
        }
    }
	    

    /* Process Redundant Updates */
    if (resGetCaAnswer(ppid -> radr, &master) == RES_OK) 
    {
        if (master) 
	{
	   /* post STAT, SEVR and VAL (monitor), if we will finally init, 
   	      because we have read a value from the input device, set INI2,
	      send all redund fields to slave, when slave receives INI2, 
	      it will be fully init'd also
	   */

	    /* check for maximize severity */

	    if (ppid -> msev > ppid -> nsev)
	    {
		ppid -> nsev = ppid -> msev;
		ppid -> nsta = LINK_ALARM;
	    }

            monitor (ppid);

	    if (ini2_flag)
	    {
		ppid -> ini2 = 1;

	        redQInit ((struct dbCommon *)ppid, redbuf, sizeof (redbuf), pointtypeindex);
		redQField (redbuf, mylist [STAT]);
		redQField (redbuf, mylist [SEVR]);

    	        db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->astp,DBE_VALUE);
	        redQField (redbuf, mylist [ASTP]);

	        db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->spt,DBE_VALUE);
	        redQField (redbuf, mylist [SPT]);
	    
	        db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->mval,DBE_VALUE);
	        redQField (redbuf, mylist [MVAL]);

	        redQField (redbuf, mylist [NMV]);

	        redQField (redbuf, mylist [NSP]);

    	        db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->ct,DBE_VALUE);
	        redQField (redbuf, mylist [CT]);

    	        db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->err,DBE_VALUE);
	        redQField (redbuf, mylist [ERR]);

    	        db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->ersq,DBE_VALUE);
	        redQField (redbuf, mylist [ERSQ]);
	    
    	        db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->derr,DBE_VALUE);
	        redQField (redbuf, mylist [DERR]);
	    
    	        db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->dval,DBE_VALUE);
	        redQField (redbuf, mylist [DVAL]);
	    
	        redQField (redbuf, mylist [VAL]);

    	        db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->pval,DBE_VALUE);
	        redQField (redbuf, mylist [PVAL]);

    	        db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->lmde,DBE_VALUE);
	        redQField (redbuf, mylist [LMDE]);

    	        db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid -> mode, DBE_VALUE);
	        redQField (redbuf, mylist [MODE]);
	    
    	        db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->span,DBE_VALUE);
	        redQField (redbuf, mylist [SPAN]);
	    
    	        db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->zero,DBE_VALUE);
	        redQField (redbuf, mylist [ZERO]);
	    
	        redQField (redbuf, mylist [FGCR]);

	        redQField (redbuf, mylist [NGAP]);

	        redQField (redbuf, mylist [GACT]);

	        redQField (redbuf, mylist [CSP]);

	        redQField (redbuf, mylist [INI2]);
	        redQSend (redbuf);
	    }
	}
    }
    return(0);
}


/********************************** alarm ************************************/

static void alarm (ppid, clamp_mask)
    struct pidRecord	*ppid;
    unsigned short	clamp_mask;
{
    double	val = ppid -> val;
    float	temp,ndevl,nhyst;
    int		adis = 0;

    /* check for alarm inhibit */

    adis = AlarmDisableLink ((void *)ppid);
    if ( (ppid->aenb == ALARM_DISABLED) || (adis != 0) )
    {
	if (ppid -> nsta != NO_INIT_ALARM)
	{
	    ppid->nsta = INHIBIT_ALARM;
	    ppid->nsev = NO_ALARM;
	}
	return;
    }

    if (ppid->udf == TRUE ) 
    {
	if (ppid->nsev < VALID_ALARM) 
	{
	    ppid->nsev = VALID_ALARM;
	    ppid->nsta = UDF_ALARM;
	}
	return;
    }

    /* check for a deviation alarm */

    if (ppid -> devl != 0.0)
    {
	nhyst = (ppid->hyst * ppid->span);
	if (nhyst < 0.0)
	    nhyst = -nhyst;
	ndevl = (ppid->devl * ppid->span);
	if (ndevl < 0.0)
	    ndevl = -ndevl;
	if (ppid -> dvsv > NO_ALARM)
	{
	    temp = (ppid->nmv - ppid->nsp);
	    if (temp < 0 ) 
		temp = -temp;
	    if ( (temp >= ndevl) || 
		 ( (ppid->stat == DEVIATION_ALARM) &&
		   (temp >= (ndevl-nhyst)) )  )
	    {
		if (ppid -> aldl > 0)
		{
		    /* check for alarm delay */
		    if (ppid -> dvct < ALDL_MAX)
			ppid -> dvct++;

		    /* if count is > than alarm delay and the current 
		       alarm severity is less than the DEV severity
		       set DEVIATION
		    */
		    if ( (ppid -> dvct >= ppid -> aldl) &&
			 (ppid -> nsev < ppid -> dvsv) )
		    {
			ppid -> nsta = DEVIATION_ALARM;
			ppid -> nsev = ppid -> dvsv;
		    }
		}
		else
		{
		    if (ppid -> nsev < ppid -> dvsv)
		    {
			ppid -> nsta = DEVIATION_ALARM;
			ppid -> nsev = ppid -> dvsv;
		    }
		}
	    }
	    else
	    {
		if (ppid -> aldl > 0)
		{
		    /* remove the DEVIATION alarm bit
		       and set deviation counter to 0
		    */
		    ppid -> larf &= ~LARF_DEV;
		    ppid -> dvct = 0;
		}
	    }
	}
	else
	{
	    if (ppid -> aldl > 0)
	    {
		ppid -> larf &= ~LARF_DEV;
		ppid -> dvct = 0;
	    }
	}
    }
    else
    {
	if (ppid -> aldl > 0)
	{
	    ppid -> larf &= ~LARF_DEV;
	    ppid -> dvct = 0;
	}
    }

    /* check for HIGH output clamp alarm and HIGH setpoint
       clamp alarm
    */
    if (ppid -> chsv > NO_ALARM)
    {
	if ( (clamp_mask & DO_HIGH_OUTPUT_CLAMP) ||
	     ( (ppid -> stat == HIGH_OUTPUT_CLAMP) &&
	       (val >= (ppid -> outh - ppid -> cdel)) ) )
	{
	    if (ppid -> aldl > 0)
	    {
		/* if we were counting low output clamp, 
		   zero counter since same counter used for
		   HIGH and LOW clamp alarm
		*/
		if (ppid -> larf & LARF_OUTPUT_LOW)
		    ppid -> hloc = 0;
		if (ppid -> hloc < ALDL_MAX)
		    ppid -> hloc++;
		ppid -> larf |= LARF_OUTPUT_HIGH;
		if ( (ppid -> hloc >= ppid ->aldl) &&
		     (ppid -> nsev < ppid -> chsv) )
		{
		    ppid -> nsta = HIGH_OUTPUT_CLAMP;
		    ppid -> nsev = ppid -> chsv;
		}
	    }
	    else
	    {
		if (ppid -> nsev < ppid -> chsv)
		{
		    ppid -> nsta = HIGH_OUTPUT_CLAMP;
		    ppid -> nsev = ppid -> chsv;
		}
	    }
	}
	else
	{
	    if (ppid -> aldl > 0)
	    {
		ppid -> larf &= ~LARF_OUTPUT_HIGH;
		if (!(ppid -> larf & ((unsigned long)LARF_OUTPUT_LOW)))
		    ppid -> hloc = 0;
	    }
	}
	if (clamp_mask & DO_HIGH_SETPT_CLAMP)
	{
	    if (ppid -> aldl > 0)
	    {
		if (ppid -> larf & LARF_SETPT_LOW)
		    ppid -> hlsc = 0;
		if (ppid -> hlsc < ALDL_MAX)
		    ppid -> hlsc++;
		ppid -> larf |= LARF_SETPT_HIGH;
		if ( (ppid -> hlsc >= ppid -> aldl) &&
		     (ppid -> nsev < ppid -> chsv) )
		{
		    ppid -> nsta = HIGH_SETPT_CLAMP;
		    ppid -> nsev = ppid -> chsv;
 		}
	    }
	    else
	    {
		if (ppid -> nsev < ppid -> chsv)
		{
		    ppid -> nsta = HIGH_SETPT_CLAMP;
		    ppid -> nsev = ppid -> chsv;
		}
	    }
	}
	else
	{
	    if (ppid -> aldl > 0)
	    {
		ppid -> larf &= ~LARF_SETPT_HIGH;
		if (!(ppid -> larf & ((unsigned long)LARF_SETPT_LOW)))
		    ppid -> hlsc = 0;
	    }
	}
    }
    else
    {
	if (ppid -> aldl > 0)
	{
	    ppid -> larf &= ~(LARF_OUTPUT_HIGH | LARF_SETPT_HIGH);
	    if (!(ppid -> larf & ((unsigned long)LARF_OUTPUT_LOW)))
		ppid -> hloc = 0;
	    if (!(ppid -> larf & ((unsigned long)LARF_SETPT_LOW)))
		ppid -> hlsc = 0;
	}
    }
    if (ppid -> clsv > NO_ALARM)
    {
	if ( (clamp_mask & DO_LOW_OUTPUT_CLAMP) ||
	     ( (ppid -> stat == LOW_OUTPUT_CLAMP) &&
	       (val <= (ppid -> outl + ppid -> cdel)) ) )
	{
	    if (ppid -> aldl > 0)
	    {
		if (ppid -> larf & LARF_OUTPUT_HIGH)
		    ppid -> hloc = 0;
		if (ppid -> hloc < ALDL_MAX)
		    ppid -> hloc++;
		ppid -> larf |= LARF_OUTPUT_LOW;
		if ( (ppid -> hloc >= ppid ->aldl) &&
		     (ppid -> nsev < ppid -> clsv) )
		{
		    ppid -> nsta = LOW_OUTPUT_CLAMP;
		    ppid -> nsev = ppid -> clsv;
		}
	    }
	    else
	    {
		if (ppid -> nsev < ppid -> clsv)
		{
		    ppid -> nsta = LOW_OUTPUT_CLAMP;
		    ppid -> nsev = ppid -> clsv;
		}
	    }
	}
	else
	{
	    if (ppid -> aldl > 0)
	    {
		ppid -> larf &= ~LARF_OUTPUT_LOW;
		if (!(ppid -> larf & ((unsigned long)LARF_OUTPUT_HIGH)))
		    ppid -> hloc = 0;
	    }
	}
	if (clamp_mask & DO_LOW_SETPT_CLAMP)
	{
	    if (ppid -> aldl > 0)
	    {
		if (ppid -> larf & LARF_SETPT_HIGH)
		    ppid -> hlsc = 0;
		if (ppid -> hlsc < ALDL_MAX)
		    ppid -> hlsc++;
		ppid -> larf |= LARF_SETPT_LOW;
		if ( (ppid -> hlsc >= ppid -> aldl) &&
		     (ppid -> nsev < ppid -> clsv) )
		{
		    ppid -> nsta = LOW_SETPT_CLAMP;
		    ppid -> nsev = ppid -> clsv;
		}
	    }
	    else
	    {
		if (ppid -> nsev < ppid -> clsv)
		{
		    ppid -> nsta = LOW_SETPT_CLAMP;
		    ppid -> nsev = ppid -> clsv;
		}
	    }
	}
	else
	{
	    if (ppid -> aldl > 0)
	    {
		ppid -> larf &= ~LARF_SETPT_LOW;
		if (!(ppid -> larf & ((unsigned long)LARF_SETPT_HIGH)))
		    ppid -> hlsc = 0;
	    }
	}
    }
    else
    {
	if (ppid -> aldl > 0)
	{
	    ppid -> larf &= ~(LARF_OUTPUT_LOW | LARF_SETPT_LOW);
	    if (!(ppid -> larf & ((unsigned long)LARF_OUTPUT_HIGH)))
		ppid -> hloc = 0;
	    if (!(ppid -> larf & ((unsigned long)LARF_SETPT_HIGH)))
		ppid -> hlsc = 0;
	}
    }

    /* do a return to normal, check for alarm delay ALDL only
       on the listed alarms
    */

    if ( ((ppid -> nsta == NO_ALARM) && (ppid -> nsev == NO_ALARM)) &&
	 ((ppid -> stat == DEVIATION_ALARM) || 
	  (ppid -> stat == HIGH_OUTPUT_CLAMP) ||
	  (ppid -> stat == LOW_OUTPUT_CLAMP) ||
	  (ppid -> stat == HIGH_SETPT_CLAMP) ||
	  (ppid -> stat == LOW_SETPT_CLAMP)) )
    {
	if (ppid -> aldl > 0)
	{
	    if (ppid -> rtct < ALDL_MAX)
		ppid -> rtct++;

	    /* delay return to normal so, continue last alarm */
	    if (ppid -> rtct < ppid -> aldl)
	    {
		ppid -> nsta = ppid -> stat;
		ppid -> nsev = ppid -> sevr;
	    }
	}
    }
    else
    {
	if (ppid -> aldl > 0)
	{
	    ppid -> rtct = 0;
	}
    }

    return;
}


/******************************** clamp_setpoint ******************************/

static void clamp_setpoint (ppid, clamp_mask)
    struct pidRecord    *ppid;
    unsigned short	*clamp_mask;
{

	/* clamp the setpoint in case someone set it too high */
    if (ppid->astp >= ppid->sthc)
    {
	ppid->astp = ppid->sthc;
	if (ppid->nsev < ppid->chsv) 
	{
	    *clamp_mask |= DO_HIGH_SETPT_CLAMP;
	}
    }
    else if (ppid->astp <= ppid->stlc)
    {
	ppid->astp = ppid->stlc;
	if (ppid->nsev < ppid->clsv) 
	{
	    *clamp_mask |= DO_LOW_SETPT_CLAMP;
	}
    }
    return;
}


/********************************** doPid *************************************/

/*
 *	Perform the Algorithm
 *
*/
long doPid (ppid, clamp_mask)
    struct pidRecord	*ppid;
    unsigned short	*clamp_mask;
{
        long            options,nRequest;
	unsigned short	mode;	/* downstream record mode */

	float		nomax,nomin,tlastnormmv,pgain,pgmul,igain,dgain,temp;
	unsigned long	thistime;
	float		deltatime = 0.0;
	float		deltafix = 0.0;
	char		casm2name[40];	/* buffer to build name of ds mode */
	int		no_elements;
	float		mval,oval,lastderiv,thisderiv,tout,deltav,olditerm;
	float		noval = 0.0;
	unsigned short	closeinit,needdeltafix;
	unsigned short	incontrol;

            /* Time Management, get current clock ticks */
	thistime = tickGet(); 

            /* Compute the delta time since last processing */
	deltatime = (thistime - ppid->ct) / (float)vxTicksPerSecond;
	ppid->ct = thistime;

	/* Save the last mode  */
	ppid->lmde = ppid->mode;


	mode = ppid->am;   /* operator controls mode directly */
	ppid->csmd = PID_MODE_AUTO;

	/* fetch downstream mode - this record uses this output value as its input */	
	if (ppid->dscs.type == DB_LINK) 
	{
	    options=0;
            nRequest=1;
	    if( (dbGetLink(&(ppid->dscs.value.db_link),
			   (struct dbCommon *)ppid,
			   DBR_USHORT, (caddr_t)(&ppid->csm1),
			   &options, &nRequest) != NULL) ||
		(!okToUse(&ppid->dscs) )  )
	    {
		if (ppid->nsev<ppid->nccs) 
		{
		    ppid->nsta = OUT_MODE_ERROR;
		    ppid->nsev = ppid->nccs;
               	}
		mode = PID_MODE_TRACKING;
		ppid->csmd = PID_MODE_TRACKING;
            } 
	    else 
	    {
	        if (strncmp(ppid->dscs.value.db_link.fldname,"MODE",4) == 0) 
		{
		    /* Down Stream must be an input point type */
	            if(ppid->csm1 == PID_MODE_MANUAL) 
		    {
		        /* Enter Tracking Mode */
		        mode = PID_MODE_TRACKING;
		        ppid->csmd = PID_MODE_TRACKING;
		    }
		}
	        else if (strncmp(ppid->dscs.value.db_link.fldname,"OMSL",4) == 0) 
		{
		    /* Down Stream must be an output point type */
	            if(ppid->csm1 == MODE_SUPERVISORY) 
		    {
	        	/* Enter Tracking Mode */
		    	mode = PID_MODE_TRACKING;
		    	ppid->csmd = PID_MODE_TRACKING;
		    }
		}
		else if (strncmp(ppid->dscs.value.db_link.fldname,"USCS",4) == 0) 
		{
		    struct dbAddr   addr;
		    long	    options = 0;
		    long	    nRequest = 1;

		    /* Down Stream must be a pid point type */
		   
	            strncpy(casm2name,ppid->dscs.value.db_link.pvname,PVNAME_SZ);
	            strcat(casm2name,".MODE");
	            no_elements = 1;
	            if ( (dbNameToAddr(casm2name, &addr) != 0) ||
			 (dbGetField(&addr, DBR_USHORT,(caddr_t)(&ppid->csm2),
				       &options, &nRequest, NULL) != 0) ) 
	            {
		        if (ppid->nsev<ppid->nccs) 
			{
			    ppid->nsta = OUT_MODE_ERROR;
			    ppid->nsev = ppid->nccs;
                        }
		        ppid->csmd = PID_MODE_TRACKING;
		        mode = PID_MODE_TRACKING;
		    } 
		    else 
		    {
		        if( (ppid->csm1 != PID_CASCLOSED) || 
			    (ppid->csm2 != PID_MODE_AUTO) ) 
			{
		            /* Enter Tracking Mode */
			    ppid->csmd = PID_MODE_TRACKING;
			    mode = PID_MODE_TRACKING;
		        }
		    }

              	}
		else
		{
		    if (ppid -> csm1 == 0)
		    {
		        /* Enter Tracking Mode */
		        mode = PID_MODE_TRACKING;
		        ppid->csmd = PID_MODE_TRACKING;
		    }
		}
	    }
        }

	/* Go into Manual Mode with tracking */
	if ( (ppid->lmde != PID_MODE_TRACKING) && (mode == PID_MODE_TRACKING) ) 
	{
	    /* if aclo not manual or auto then it is last mode
	       so we do not chg am
	    */
	    if (ppid -> aclo == PID_ACLO_MANUAL)
		ppid -> am = PID_MODE_MANUAL;
	    else if (ppid -> aclo == PID_ACLO_AUTOMATIC)
		ppid -> am = PID_MODE_AUTO;
	}

	/* If leaving tracking and autoclose go into automatic */ 
	if ( (ppid->lmde == PID_MODE_TRACKING) && (mode != PID_MODE_TRACKING) )
	{
	    /* if aclo not manual or auto then it is last mode
	       so we do not chg am
	    */
	    if (ppid -> aclo == PID_ACLO_MANUAL)
		ppid -> am = PID_MODE_MANUAL;
	    else if (ppid -> aclo == PID_ACLO_AUTOMATIC)
		ppid -> am = PID_MODE_AUTO;
	}

	/* Set this Pass Mode */
	if (mode == PID_MODE_TRACKING) 
	    ppid->mode = PID_MODE_TRACKING;
	else 
	    ppid->mode = ppid->am;

	/* fetch the controlled value */
	if (ppid->mvl.type != DB_LINK) 
	{	/* no controlled value */
	    if (ppid->nsev<ppid->ncsv) 
	    {
		ppid->nsta = PID_ALM_NOMV;
		ppid->nsev = ppid->ncsv;
            }
	    if ( (ppid->hold == PID_HOLD_ON_LOSS) &&
		 (ppid -> mode == PID_MODE_AUTO) )
	    {	/* hold output */
		ppid->val = ppid->hval;
		ppid->pval = ppid->val/100;
		db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->val,DBE_VALUE);
	    }
	    return(0);
	}

        options=0;
        nRequest=1;
	if( (dbGetLink(&(ppid->mvl.value.db_link),
		       (struct dbCommon *)ppid,
		       DBR_FLOAT, (caddr_t)&mval,
		       &options, &nRequest) != NULL) ||
	     (!okToUse(&ppid->mvl)) ) 
	{
	    if (ppid->nsev<ppid->ncsv) 
	    {
		ppid->nsta = PID_ALM_NOMV;
		ppid->nsev = ppid->ncsv;
            }
	    if ( (ppid->hold == PID_HOLD_ON_LOSS) &&
		 (ppid -> mode == PID_MODE_AUTO) )
	    {		/* hold output */
		ppid->val = ppid->hval;
		ppid->pval = ppid->val/100;
		db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->val,DBE_VALUE);
	    }
            /* Go Into MANUAL Mode on loss of input 
	       unless we are in tracking mode, then continue to
	       track the output
	    */
	    if (ppid -> mode != PID_MODE_TRACKING)
	    	ppid->mode = ppid->am = PID_MODE_MANUAL;
	    else
		do_tracking (ppid, clamp_mask);
            return(0);
	}

        /* Have A New Measured Value */
	ppid->mval  = mval;
	tlastnormmv = ppid -> nmv;

	/* normalize the measured value */
	ppid -> nmv = (ppid -> mval * ppid -> span) - ppid -> zero;

	/* Process Tracking Mode Exit if AUTOEXIT*/
        closeinit = FALSE;
	if ((ppid->lmde == PID_MODE_TRACKING) &&
	    (ppid->mode == PID_MODE_AUTO) )
	{
	    closeinit = TRUE;
        }

	/* Process Manual Mode Exit */
	if ( (ppid->lmde == PID_MODE_MANUAL) &&
	     (ppid->mode == PID_MODE_AUTO) ) 
	{
	    closeinit = TRUE;
        }


	/* Transition to AUTOMATIC Processing */
	if (closeinit) 
	{
	    ppid->i = 0;	/* last integral contribution */
	    ppid->err = 0; 	/* Error */
	    ppid->lerr = 0;	/* Last Error */
	    ppid->ersq = 0;	/* Error Squared */
	    ppid->lesq = 0;	/* Last Error Squared */
	    ppid->derr = 0;	/* Delta Error */
	    ppid->lder = 0;	/* Last Delta Error */
	    ppid->dval = 0;	/* Delta Measurement */
	    ppid->ldvl = 0;	/* Last Delta Measurement */
	    ppid->fgcr = 0;	/* just crossed into setpoint gap */
	    ppid->gact = 0;	/* in gap, gap is active */

	    ppid->pval = ppid->val/100;


            /* Set the SP = MV if revinit is selected */
	    
	    if (ppid->rini == PID_REV_INIT)
	    {
		ppid -> spt = ppid -> astp = mval;
		ppid -> nsp = (ppid -> astp * ppid -> span) - ppid -> zero;
	    }
         
	    if ((ppid->velo == PID_POSITIONAL) && (ppid->rout.type == DB_LINK))  
	    {
	        /* fetch the output value and set our output to it */
                /* So it does not move as we initialize		   */
	        options=0;
	        nRequest=1;
		if( (dbGetLink(&(ppid->rout.value.db_link),
			       (struct dbCommon *)ppid,
			       DBR_FLOAT, (caddr_t)&oval,
			       &options, &nRequest) != NULL) ||
		    (!okToUse(&ppid->rout)) )
		{
		    if (ppid->nsev < ppid->nosv) 
		    {
		        ppid->nsta = PID_ALM_NOOUT;
		        ppid->nsev = ppid->nosv;
		    }
		    return(0);
                }
		ppid->pval = (oval /100);	/* previous output value 0 - 1 */
		ppid->val = oval;
	    }
	}

	get_setpoint (ppid);
	clamp_setpoint (ppid, clamp_mask);

	ramp_setpoint (ppid, deltatime);

	/* Reverse calculate the setpoint we are using to show upstream 
	   when we are out of control 
	*/
	if (ppid->cpct == PID_CAS_PCT) /* Turn the Setpoint to EGU */
	    ppid->csp = ((ppid->astp * ppid->span) - ppid->zero) * 100;
	else
	    ppid->csp = ppid->astp;


	/* Normalize the Output Clamps */
	nomax = (ppid->outh/100);
	nomin = (ppid->outl/100);


	/* If Current Mode is Tracking Process it Here */
	if ( (ppid->mode == PID_MODE_TRACKING) ||
	     ( (ppid->lmde == PID_MODE_TRACKING) &&
	       (ppid->mode == PID_MODE_MANUAL) )  ) 
	{
	    do_tracking (ppid, clamp_mask);
	    return (0);
	}

	/* Process Manual mode */
	if (ppid->mode == PID_MODE_MANUAL) 
	{
	    ppid->pval = ppid->val/100;
	    if (ppid->pval >= nomax) 
	    {
		ppid->pval = nomax;
		ppid->val = ppid->pval*100;
		if (ppid->nsev < ppid->chsv) 
		{
		    *clamp_mask |= DO_HIGH_OUTPUT_CLAMP;
        	}
	    } 
	    else 
	    { 
		if (ppid->pval <= nomin) 
		{
		    ppid->pval = nomin;
		    ppid->val = ppid->pval*100;
		    if (ppid->nsev < ppid->clsv) 
		    {
			*clamp_mask |= DO_LOW_OUTPUT_CLAMP;
		    }
		}
            }
	    return (0);
        }

	/* This is where we process Automatic Mode */

	/* Compute the gains */
	/*Proportional */
	if (ppid->kp == 0) 
	{
	    pgain = 0;
	    pgmul = 1;
	} 
	else 
	{
	    pgain = (100/ppid->kp);
	    pgmul = pgain;
	}
	/*Integral*/
	if (ppid->ki == 0) 
	    igain = 0;
	else 
	    igain = (pgmul*(deltatime/(60*ppid->ki)));

	/* Derivative */
	if (ppid->kd == 0) 
	    dgain = 0;
	else 
	    dgain = (pgmul*((ppid->kd*deltatime)/60));



	/* Update the "last" values before processing the new ones */

        /* These Values should all be normalize between 0 and 1 */
	ppid->ldvl = ppid->dval;	/* Delta MV */
	ppid->lerr = ppid->err;		/* Error */
	ppid->lder = ppid->derr;	/* Delta Error */
	ppid->lesq = ppid->ersq;	/* Error Squared */

	/* Compute the "this time" values */
	ppid->dval = (tlastnormmv-ppid->nmv);
	ppid->err = (ppid->nsp-ppid->nmv);
	ppid->derr = (ppid->err-ppid->lerr);

	/* We Need A special Case for Error Squared to Preserve the sign */
	ppid->ersq = ((ppid->err*ppid->err)*100);
	if (ppid->err < 0) ppid->ersq = -ppid->ersq;


	/* gap the output if there is a setpoint gap */

	if (ppid -> sptg != 0)
	{
	       /* if we are inside the gap */
	    temp =  ppid -> err;
	    if (temp < 0.0)
		temp = -temp;
	    if (temp <= ppid -> ngap)
	    {
		   /* just crossed into the gap, remember which side
		      of setpoint we are on. sndf == 1 if below setpoint */
		if (ppid -> fgcr == 0)
		{
		    ppid -> fgcr = 1;
		    ppid -> sndf = (ppid -> err > 0);
		}
		   /* don't cross setpoint or change output, stay on this
		      side of the setpoint */
		if (ppid -> sptc == NO)
		{
		    if (ppid -> velo == PID_INCREMENTAL)
		    {
			ppid -> val = 0;
			ppid -> pval = 0;
	    		if (ppid->pval >= nomax) 
	    		{
			    ppid->pval = nomax;
			    ppid->val = ppid->pval*100;
			    if (ppid->nsev < ppid->chsv) 
			    {
		    		*clamp_mask |= DO_HIGH_OUTPUT_CLAMP;
        		    }
	    		} 
	    		else 
	    		{ 
			    if (ppid->pval <= nomin) 
			    {
		    		ppid->pval = nomin;
		    		ppid->val = ppid->pval*100;
		    		if (ppid->nsev < ppid->clsv) 
		    		{
				    *clamp_mask |= DO_LOW_OUTPUT_CLAMP;
		    		}
			    }
            		}
		    }
		    return (0);
		}

		   /* we crossed the setpoint, don't change the output
		      even if we go back and forth over the setpoint 
		      or err (temp) is 0.0, the 0.0005 is used instead of
		      0.0 since the floating point process may not give us
		      0.0, this puts us within the proper accuracy*/
		if ( ((ppid -> err > 0) != ppid -> sndf) || 
		     (ppid -> gact) ||
		     (temp < 0.0005) )
		{
		    if (ppid -> velo == PID_INCREMENTAL)
		    {
			ppid -> val = 0;
			ppid -> pval = 0;
	    		if (ppid->pval >= nomax) 
	    		{
			    ppid->pval = nomax;
			    ppid->val = ppid->pval*100;
			    if (ppid->nsev < ppid->chsv) 
			    {
		    		*clamp_mask |= DO_HIGH_OUTPUT_CLAMP;
        		    }
	    		} 
	    		else 
	    		{ 
			    if (ppid->pval <= nomin) 
			    {
		    		ppid->pval = nomin;
		    		ppid->val = ppid->pval*100;
		    		if (ppid->nsev < ppid->clsv) 
		    		{
				    *clamp_mask |= DO_LOW_OUTPUT_CLAMP;
		    		}
			    }
            		}
		    }
		    ppid -> gact = 1;
		    return (0);
		}
	    }
	    else
	    {
		ppid -> fgcr = 0;
		ppid -> gact = 0;
	    }
	}

	/* If Feedback is enabled check to see if we are out of control */
	/* If the output is more than .1% from our last setting we assume */
	/* the output is being controled from somewhere else  */
	incontrol = TRUE;
	if ( (ppid->velo == PID_POSITIONAL) &&
	     (ppid->optr == PID_FB_ENABLED) && 
	     (ppid->rout.type == DB_LINK) ) 
	{
	    options=0;
	    nRequest=1;
	    if ( (dbGetLink(&(ppid->rout.value.db_link),
			    (struct dbCommon *)ppid,
			    DBR_FLOAT, (caddr_t)&oval,
			    &options, &nRequest) != NULL) ||
		 (!okToUse(&ppid->rout)) )
	    {
		if (ppid->nsev < ppid->nosv)  
		{
                    ppid->nsta = PID_ALM_NOOUT;
                    ppid->nsev = ppid->nosv;
		}
                return (0);
	    }

	    deltav = oval - ppid->val;
	    noval = oval/100; /* !!! */
	    if (deltav < 0) deltav = -deltav;
	    if (deltav > CONTROL_VALID_PERCENT) incontrol = FALSE;
	}

	/* Compute the P I & D terms */

	/* Proportional Term */
	if (ppid->esqr == PID_NO_ERRSQ)
	    /* Use Straight computation */
	    ppid->p = (pgain * ppid->derr);
	else
	    /* Used the Squared change in error */
	    ppid->p = pgain*(ppid->ersq-ppid->lesq);

	/* Integral Term */

	olditerm = ppid->i;

	if (ppid->esqr == PID_NO_ERRSQ)
	    /* Use the Straight Computation */
	    ppid->i = (igain*ppid->err);
	else
	    /* Use the Squared Error */
	    ppid->i = (igain*ppid->ersq);

	/* adjust for feed back */
/*!!!	if (incontrol) olditerm = 0; */

	/* Derivative term */
	if (ppid->velo != PID_POSITIONAL) 
	{
	    /* Process the incremental form of the algorithm */
	    if (ppid->dmod) 
	    {
		/* Process the Delta Measurement Mode Algorithm */
		ppid->d = dgain * (ppid->dval-ppid->ldvl);
	    } 
	    else 
	    {
		/* Process the Delta Error Mode Algorithm */
		ppid->d = dgain * (ppid->derr-ppid->lder);
	    }

	    /* add up the terms and clamp the result */
	    ppid->pval = ppid->p+ppid->i+ppid->d;

	    if (ppid->ract == PID_REV_ACT) 
	    {
		ppid->pval = (-ppid->pval);
	    }

	    if (ppid->pval >= nomax) 
	    {
		ppid->pval = nomax;
	    } 
	    else 
	    { 
		if (ppid->pval <= nomin) 
		{
		    ppid->pval = nomin;
		}
	    }
 	    ppid->val = ppid->pval*100;

	} 
	else 
	{
	    /* If I get This far the algorithm is positional */
	    /*Process the positional form of the algorithm */

            /* Compute this pass and last pass derivative contribution */ 
	    if (ppid->dmod) 
	    {
	        /* Process the Delta Measurement Mode Algorithm */
		lastderiv = dgain * (ppid->ldvl);
		thisderiv = dgain * (ppid->dval);
            } 
	    else 
	    {
	        /* Process the Delta Error Mode Algorithm */
		lastderiv = dgain * (ppid->lder);
		thisderiv = dgain * (ppid->derr);
	    }

	    /* We may need to fudge the last deriv terms if we have 
	       reached the limits 
	    */
	    needdeltafix = FALSE;

	    /* Compute the intermediate result with out the last deriv term */
	    if (ppid->ract != PID_REV_ACT) 
	    {
	        tout = ppid->pval+ppid->p+ppid->i-lastderiv;
	    } 
	    else 
	    {
		tout = ppid->pval-ppid->p-ppid->i+lastderiv;
	    }

            /* Clamp the results */
 	    if (tout >= nomax) 
	    {
		tout = nomax;
	    } 
	    else 
	    {
		if (tout <= nomin) 
		{
		    tout = nomin;
		}
	    }

	    /* Now try to fit this term in */
	    if ((thisderiv>0) && (ppid->ract != PID_REV_ACT) )
	    {
	        /* Moving in a positive Direction */
	        if ((tout+thisderiv)>=nomax) 
		{
                    /* It Does not fit change the dterm and clamp the output */
		    deltafix = (nomax-tout)/dgain;
		    needdeltafix = TRUE;
		    ppid->pval = nomax;
	        } 
		else 
		{
		    /* It Does fit Use It */
		    ppid->pval = tout+thisderiv;
	        }
	    }
	    if ((thisderiv<=0) && (ppid->ract != PID_REV_ACT) )
	    {	    
                /* Moving in a negative direction */
	        if (((tout+thisderiv) <= nomin) && (dgain != 0)) 
		{
		    /* It Does not fit change the dterm and clamp the output */
		    deltafix = (nomin-tout)/dgain;
		    needdeltafix = TRUE;
		    ppid->pval = nomin;
	        } 
		else 
		{
		    /* It does fit use it */
		    ppid->pval = tout+thisderiv;
	        }
	    }
	    if ((thisderiv<0) && (ppid->ract == PID_REV_ACT) )
	    {
	        /* Moving in a positive Direction */
	        if ((tout-thisderiv)>=nomax) 
		{
                    /* It Does not fit change the dterm and clamp the output */
		    deltafix = -((nomax-tout)/dgain);
		    needdeltafix = TRUE;
		    ppid->pval = nomax;
	        } 
		else 
		{
		    /* It Does fit Use It */
		    ppid->pval = tout-thisderiv;
	        }
	    }
	    if ((thisderiv>=0) && (ppid->ract == PID_REV_ACT) )
	    {	    
                /* Moving in a negative direction */
	        if (((tout-thisderiv) <= nomin) && (dgain != 0)) 
		{
		    /* It Does not fit change the dterm and clamp the output */
		    deltafix = -((nomin-tout)/dgain);
		    needdeltafix = TRUE;
		    ppid->pval = nomin;
	        } 
		else 
		{
		    /* It does fit use it */
		    ppid->pval = tout-thisderiv;
	        }
	    }

	    /* if the derivative term did not fit fudge it back to what did */
	    if (needdeltafix) 
	    {
	        /* Which factor did we use */
		if (ppid->dmod) 
		{
	            /* Process the Delta Measurement Mode Algorithm */
		    ppid->dval = deltafix;
	        } 
		else 
		{
	            /* Process the Delta Error Mode Algorithm */
		    ppid->derr = deltafix;
	        }
	    }
	}

	/* if this integral contribution brings it closer use it */
	if (!incontrol) 
	{	
	    if (ppid->ract == PID_REV_ACT) 
		olditerm = -olditerm;
	    if (fabs(ppid -> pval - noval - olditerm) < fabs(ppid -> pval - noval))
		ppid->pval = ppid->pval - olditerm;
	}


	ppid->val = (ppid->pval*100);


	/* Check for High Output Clamping */
	if ((ppid->nsev < ppid->chsv) && (ppid->pval >= nomax)) 
	{
	    *clamp_mask |= DO_HIGH_OUTPUT_CLAMP;
        }
	/* Check for Low OutPut Clamping */
	if ((ppid->nsev < ppid->clsv) && (ppid->pval<= nomin)) 
	{
	    *clamp_mask |= DO_LOW_OUTPUT_CLAMP;
	}

	/* Done Processing go deal with alarms monitors and redundancy */
	return (0);

}


/************************************* do_tracking *****************************/

static void do_tracking (ppid, clamp_mask)
    struct pidRecord	*ppid;
    unsigned short	*clamp_mask;
{
    float oval;
    float nomax;
    float nomin;
    long  options = 0;
    long  nRequest = 1;

    /* Normalize the Output Clamps */

    nomax = (ppid -> outh / 100);
    nomin = (ppid -> outl / 100);

    if ( (ppid -> mode == PID_MODE_TRACKING) && (ppid -> nsev < ppid -> tmsv) ) 
    {
	ppid -> nsta = PID_ALM_TRACKING;
	ppid -> nsev = ppid -> tmsv;
    }

    if (ppid -> rout.type == DB_LINK) 
    {
        /* fetch the output value and set our output to it */
        /* So it does not move as we initialize		   */
        if( (dbGetLink(&(ppid -> rout.value.db_link),
		       (struct dbCommon *)ppid,
		       DBR_FLOAT, (caddr_t)&oval,
		       &options, &nRequest) != NULL) || 
	    (!okToUse(&ppid -> rout)) ) 
	{
	    if (ppid -> nsev < ppid -> nosv) 
	    {
		ppid -> nsta = PID_ALM_NOOUT;
		ppid -> nsev = ppid -> nosv;
	    }
	    return;
	}
	ppid -> pval = (oval / 100);	/* previous output value 0 - 1 */
	ppid -> val = oval;
    }
    ppid -> pval = ppid -> val / 100;
    if (ppid -> pval >= nomax) 
    {
	ppid -> pval = nomax;
	ppid -> val = ppid -> pval * 100;
	if (ppid -> nsev < ppid -> chsv) 
	{
	    *clamp_mask |= DO_HIGH_OUTPUT_CLAMP;
        }
    } 
    else 
    { 
	if (ppid -> pval <= nomin) 
	{
	    ppid -> pval = nomin;
	    ppid -> val = ppid -> pval * 100;
	    if (ppid -> nsev < ppid -> clsv) 
	    {
		*clamp_mask |= DO_LOW_OUTPUT_CLAMP;
	    }
        }
    }

    return;
}


/*************************** get_alarm_double *******************************/

static long get_alarm_double(paddr,pad)
    struct dbAddr *paddr;
    struct dbr_alDouble	*pad;
{
    struct pidRecord	*ppid=(struct pidRecord *)paddr->precord;

    pad->upper_alarm_limit = ppid->sthc;
    pad->upper_warning_limit = ppid->sthc;
    pad->lower_warning_limit = ppid->stlc;
    pad->lower_alarm_limit = ppid->stlc;
    return(0);
}

/*************************** get_precision ***********************************/

static long get_precision(paddr,precision)
    struct dbAddr *paddr;
    long	  *precision;
{
    struct pidRecord	*ppid=(struct pidRecord *)paddr->precord;

    *precision = ppid->prec;
    return(0);
}

/**************************** get_units *************************************/

static long get_units(paddr,units)
    struct dbAddr *paddr;
    char	  *units;
{
    struct pidRecord	*ppid=(struct pidRecord *)paddr->precord;

    strncpy(units,ppid->egu,sizeof(ppid->egu));
    return(0);
}

/***************************** get_value ************************************/

static long get_value(ppid,pvdes)
    struct pidRecord		*ppid;
    struct valueDes	*pvdes;
{
    pvdes->field_type = DBF_DOUBLE;
    pvdes->no_elements=1;
    (double *)(pvdes->pvalue) = &ppid->val;
    return(0);
}


/*********************** get_graphic_double *********************************/

static long get_graphic_double(paddr,pgd)
    struct dbAddr *paddr;
    struct dbr_grDouble	*pgd;
{
    struct pidRecord *ppid = (struct pidRecord *)paddr->precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct pidRecord *)0) -> val) ):
	case ( (int)&(((struct pidRecord *)0) -> outh) ):
	case ( (int)&(((struct pidRecord *)0) -> outl) ):
	case ( (int)&(((struct pidRecord *)0) -> cdel) ):
	case ( (int)&(((struct pidRecord *)0) -> adel) ):
	case ( (int)&(((struct pidRecord *)0) -> mdel) ):
	case ( (int)&(((struct pidRecord *)0) -> hval) ):
	case ( (int)&(((struct pidRecord *)0) -> alst) ):
	case ( (int)&(((struct pidRecord *)0) -> mlst) ):
	    pgd -> upper_disp_limit = 100;
	    if (ppid -> velo == PID_POSITIONAL)
	        pgd -> lower_disp_limit = 0;
	    else
	        pgd -> lower_disp_limit = -100;
	    break;

	case ( (int)&(((struct pidRecord *)0) -> kp) ):
	    pgd -> upper_disp_limit = 5000;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct pidRecord *)0) -> kd) ):
	case ( (int)&(((struct pidRecord *)0) -> ki) ):
	    pgd -> upper_disp_limit = 100;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct pidRecord *)0) -> eguf) ):
	case ( (int)&(((struct pidRecord *)0) -> egul) ):
	    pgd -> upper_disp_limit = ppid -> eguf;
	    pgd -> lower_disp_limit = ppid -> egul;
	    break;

        case ( (int)&(((struct pidRecord *)0) -> devl) ):
        case ( (int)&(((struct pidRecord *)0) -> hyst) ):
        case ( (int)&(((struct pidRecord *)0) -> srmp) ):
        case ( (int)&(((struct pidRecord *)0) -> sptg) ):
	{
	    float    span;

	    span = ppid -> hopr - ppid -> lopr;
	    if (span < 0)
		span = -span;
	    pgd -> upper_disp_limit = span;
	    pgd -> lower_disp_limit = 0;

	    break;
	}

	case ( (int)&(((struct pidRecord *)0) -> err) ):
	case ( (int)&(((struct pidRecord *)0) -> dval) ):
	case ( (int)&(((struct pidRecord *)0) -> pval) ):
	case ( (int)&(((struct pidRecord *)0) -> zero) ):
	    pgd -> upper_disp_limit = 1;
	    pgd -> lower_disp_limit = -1;
	    break;

	case ( (int)&(((struct pidRecord *)0) -> ersq) ):
	    pgd -> upper_disp_limit = 100;
	    pgd -> lower_disp_limit = -100;
	    break;

	case ( (int)&(((struct pidRecord *)0) -> span) ):
	case ( (int)&(((struct pidRecord *)0) -> ngap) ):
	    pgd -> upper_disp_limit = 1;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct pidRecord *)0) -> lmde) ):
	    pgd -> upper_disp_limit = 2;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct pidRecord *)0) -> derr) ):
	    pgd -> upper_disp_limit = 2;
	    pgd -> lower_disp_limit = -2;
	    break;

	default:
	    pgd -> upper_disp_limit = ppid -> hopr;
	    pgd -> lower_disp_limit = ppid -> lopr;
	    break;
    }
    return(0);
}

/***************************** get_control_double ****************************/

static long get_control_double(paddr,pcd)
    struct dbAddr *paddr;
    struct dbr_ctrlDouble *pcd;
{
    struct pidRecord *ppid = (struct pidRecord *)paddr->precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct pidRecord *)0) -> val) ):
	    pcd -> upper_ctrl_limit = ppid -> outh;
	    pcd -> lower_ctrl_limit = ppid -> outl;
	    break;

	case ( (int)&(((struct pidRecord *)0) -> outh) ):
	case ( (int)&(((struct pidRecord *)0) -> outl) ):
	case ( (int)&(((struct pidRecord *)0) -> cdel) ):
	case ( (int)&(((struct pidRecord *)0) -> adel) ):
	case ( (int)&(((struct pidRecord *)0) -> mdel) ):
	case ( (int)&(((struct pidRecord *)0) -> hval) ):
	    pcd -> upper_ctrl_limit = 100;
	    if (ppid -> velo == PID_POSITIONAL)
	        pcd -> lower_ctrl_limit = 0;
	    else
	        pcd -> lower_ctrl_limit = -100;
	break;
	case ( (int)&(((struct pidRecord *)0) -> kp) ):
	    pcd -> upper_ctrl_limit = 5000;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct pidRecord *)0) -> kd) ):
	case ( (int)&(((struct pidRecord *)0) -> ki) ):
	    pcd -> upper_ctrl_limit = 100;
    	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct pidRecord *)0) -> astp) ):
	case ( (int)&(((struct pidRecord *)0) -> istp) ):
	    pcd -> upper_ctrl_limit = ppid -> sthc;
    	    pcd -> lower_ctrl_limit = ppid -> stlc;
	    break;

	case ( (int)&(((struct pidRecord *)0) -> sthc) ):
	case ( (int)&(((struct pidRecord *)0) -> stlc) ):
	    pcd -> upper_ctrl_limit = ppid -> hopr;
    	    pcd -> lower_ctrl_limit = ppid -> lopr;
	    break;

	case ( (int)&(((struct pidRecord *)0) -> devl) ):
	case ( (int)&(((struct pidRecord *)0) -> hyst) ):
	case ( (int)&(((struct pidRecord *)0) -> sptg) ):
        case ( (int)&(((struct pidRecord *)0) -> srmp) ):
	{
	    float    span;

	    span = ppid -> hopr - ppid -> lopr;
	    if (span < 0.0)
	        span = -span;
	    pcd -> upper_ctrl_limit = span;
	    pcd -> lower_ctrl_limit = 0;
	    break;
	}

	default:
	    pcd -> upper_ctrl_limit = ppid -> eguf;
	    pcd -> lower_ctrl_limit = ppid -> egul;
	    break;
    }

    return(0);
}


/******************************** get_setpoint *********************************/

static void get_setpoint (ppid)
    struct pidRecord    *ppid;
{
    long options  = 0;
    long nRequest = 1;

	/* Time to decide where our set point is coming from */
    if ((ppid->stpl.type == DB_LINK) && (ppid->uscs == PID_CASCLOSED)) 
    {
	   /* Casacded and closed -> go get it */
	if( (dbGetLink(&(ppid->stpl.value.db_link),
		       (struct dbCommon *)ppid,
		       DBR_FLOAT, (caddr_t)(&ppid->csp),
		       &options, &nRequest) != NULL) ||
	    (!okToUse(&ppid->stpl)) )
	{
	    if (ppid->nsev < ppid->nssv) 
	    {
		ppid->nsta = SETPT_ERROR;
		ppid->nsev = ppid->nssv;
	    }

	}
	else 
	{

	    if (ppid->cpct == PID_CAS_PCT)    /* Turn the Setpoint to EGU */
		ppid->astp = (((ppid->csp/100)*(ppid->eguf-ppid->egul))+ppid->egul);
	    else
		ppid->astp = ppid->csp;
	}
    }

    return;
}


/******************************** monitor ************************************/

static void monitor (ppid)
    struct pidRecord	*ppid;

    {
	unsigned short	monitor_mask;
        short           stat, sevr, nsta, nsev;
	double 		delta;
	int		set_ack = 1;

        /* get previous stat and sevr  and new stat and sevr*/

	stat = ppid->stat;
	sevr = ppid->sevr;
	nsta = ppid->nsta;
	nsev = ppid->nsev;

        /*set current stat and sevr*/

	ppid->stat = nsta;
	ppid->sevr = nsev;
	ppid->nsta = 0;
	ppid->nsev = 0;

	/* if we are coming out of VALID_ALARM and our new stat/sevr are
	   the same as when we went into VALID_ALARM, do not set ack
	*/
	if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	     (nsta == ppid -> psta) && (nsev == ppid -> psev) )
	{
	    set_ack = 0;
	}
	/* if we are going into VALID_ALARM, and out previous stat was not
	   VALID_ALARM, remember previous non VALID_ALARM stat/sevr
	   and we do not want to set ACK,
	*/
	if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
	{
	    set_ack = 0;
	    ppid -> psta = stat;
	    ppid -> psev = sevr;
	    db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid -> psta, DBE_VALUE);
	    db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid -> psev, DBE_VALUE);
	}

        monitor_mask = 0;

        /* alarm condition changed this scan */

        if (stat != nsta || sevr != nsev) 
	{
            /* post events for alarm condition change*/

            monitor_mask = DBE_ALARM;

	    /* when alarm status chg's, get current time and
	       save VAL in lalm (last value when alarm chg'd)
	    */
	    tsLocalTime (&ppid -> latm);
	    ppid -> lalm = ppid -> val;

            /* post stat and nsev fields */

	    db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->stat, DBE_VALUE);
	    db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->sevr, DBE_VALUE);
	    if ((ppid->sevr != NO_ALARM) && (set_ack) &&
		(ppid->sevr != EVENT_ALARM) ) 
	    {
		if (ppid->ack == 0)
		{
		    ppid->ack = 1;		/* new unacknowledged alarm */
		    db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->ack, DBE_VALUE);
		}
	    }
        }
        /* check for value change */

	delta = ppid->mlst - ppid->val;
        if (delta < 0.0) 
	    delta = -delta;
	
	if (delta > ppid->mdel) 
	{
	    /* post events for value change */

            monitor_mask |= DBE_VALUE;

            /* update last value monitored */

	    ppid->mlst = ppid->val;
        }

        /* check for archive change if tenb (trending) is enabled */

	if (ppid -> tenb)
	{
	    delta = ppid->alst - ppid->val;

            if (delta < 0.0) 
	    	delta = -delta;
	
	    /* if exceed archive deadband or there has been a change in alarm status */
	    if ( (delta > ppid->adel) || (monitor_mask & DBE_ALARM) )
	    {
            	/* post events on value field for archive change */

            	monitor_mask |= DBE_LOG;

            	/* update last archive value monitored */

	    	ppid->alst = ppid->val;
	    }
        }

        /* send out monitors connected to the value field */

        if (monitor_mask) 
	{
            db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->val, monitor_mask);
	}
	return;
    }



/****************************** processPid ***********************************/
/*
 *	Process the PID Algorithm
 *
*/
static long processPid (paddr)
    struct dbAddr	*paddr;
{

    struct		pidRecord	*ppid = (struct pidRecord *) (paddr->precord);
    long		status = 0;
    char		redbuf[400];
    char		pvname[PVNAME_SZ + FLDNAME_SZ + 4];
    char		*p;
    struct dbAddr	dbAddr;
    long		options = 0;
    long		nRequest = 1;
    float		mval_mdel;
    float		mval_adel;
    double		delta;
    unsigned short	monitor_mask;

    /* Fields for redundant update */
    unsigned short	stat = ppid -> stat;
    unsigned short	sevr = ppid -> sevr;
    float		mval = ppid -> mval;
    float		nmv  = ppid -> nmv;
    float		nsp  = ppid -> nsp;
    unsigned long	ct   = ppid -> ct;
    float		err  = ppid -> err;
    float		ersq = ppid -> ersq;
    float		derr = ppid -> derr;
    float		dval = ppid -> dval;
    double		val  = ppid -> val;
    float		pval = ppid -> pval;
    float		lalm = ppid -> lalm;
    float		mlst = ppid -> mlst;
    short		lmde = ppid -> lmde;
    unsigned short	mode = ppid -> mode;
    unsigned short	am   = ppid -> am;
    float		spt  = ppid -> spt;
    short		fgcr = ppid -> fgcr;
    unsigned short	sndf = ppid -> sndf;
    short		gact = ppid -> gact;
    float		csp  = ppid -> csp;
    float		istp = ppid -> istp;
    TS_STAMP		latm = ppid -> latm;
    short		dvct = ppid -> dvct;
    short		hloc = ppid -> hloc;
    short		hlsc = ppid -> hlsc;
    unsigned short	clamp_mask;


    /* Do Not Process unless initialized */
    if ((!ppid->init) || (!ppid->ini2))
	return (0);

    /* get time stamp for last time record was processed */

    tsLocalTime (&ppid -> lptm);

    ppid -> msev = NO_ALARM;
    clamp_mask = 0;

    if (ppid -> tagd) 
    {
	ppid -> nsta = TAGGED_ALARM;
	ppid -> nsev = VALID_ALARM;
    }
    else if (ppid -> oscn == OFF_SCAN) 
    {
	ppid -> nsta = SCAN_ALARM;
	ppid -> nsev = VALID_ALARM;
    }
    else 
    {
	ppid -> nsta = NO_ALARM;
	ppid -> nsev = NO_ALARM;

        /* update number of process times */
        if (process_stats)
	    process_cnts[pointtypeindex]++;	

	/* process the pid pidulation */
	ppid->pact = TRUE;
	status = doPid(ppid, &clamp_mask);
        if(status==1) 
	{
	    ppid->pact = FALSE;
	    return(0);
        }

	/* check for alarms */
	alarm (ppid, clamp_mask);

	/* check for maximize severity */

    	if (ppid -> msev > ppid -> nsev)
    	{
	    ppid -> nsev = ppid -> msev;
	    ppid -> nsta = LINK_ALARM;
    	}
    }

    /* if the time stamp is 0, init and post val with
       alarm bit set, this makes sure that proper alarm
       messages are set if IOC was down and alarm status
       may have chg'd
    */
    if (ppid -> latm.secPastEpoch == 0)
    {
	tsLocalTime (&ppid -> latm);
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid -> val, DBE_VALUE | DBE_ALARM);
    }

    /* if it is the 1st process pass 
       init last value when alarm chg'd (lalm)
    */

    if (ppid -> fpps)
    {
	ppid -> fpps = 0;
	ppid -> lalm = ppid -> val;
    }

    /* check event list */
    monitor(ppid);

    ppid->pact=FALSE;
	
    /* these are prompted fields, recalc checksum
       do it here and not in redund update below, if we are
       MASTER then this will handle redundant update, but
       we must let dbClient know of the changes to these 
       prompted fields regardless whether we are MASTER
    */
    if ( (am != ppid -> am) || (ppid -> sstp != ppid -> astp) ||
	 (istp != ppid -> astp) )
    {
	int      update = 0;

	if (am != ppid -> am)
	    update++;

	/* if setpt has changed or initial setpt does not equal
		current setpt,
	   if we are not reverse init, since this sets astp to
		meeasued val
	   and we are CASCADE OPEN or our setpt is not a link
	   and we are to update initial setpt from current
		setpt
	*/
	if ( (ppid -> sstp != ppid -> astp) || (istp != ppid -> astp) )
	{
	    if ( (ppid -> rini != PID_REV_INIT) &&
		 ( (ppid -> uscs == PID_CASOPEN) ||
		   (ppid -> stpl.type != DB_LINK) ) &&
		 (ppid -> ustp) )
	    {
		ppid -> istp = ppid -> astp;
		update++;
	    }
	}

	if (update)
	{
	    ppid -> rcrc = DBC_CRCRecord ((char *)ppid, pointtypeindex);
	}
    }

    redQInit ((struct dbCommon *)ppid, redbuf, sizeof (redbuf), pointtypeindex);

    if (stat != ppid->stat)
	redQField (redbuf, mylist [STAT]);
	
    if (sevr != ppid->sevr)
	redQField (redbuf, mylist [SEVR]);

    /* get MVAL's MDEL (monitor deadband) and ADEL (archive deadband) */

    strncpy (pvname, ppid -> mvl.value.db_link.pvname, sizeof (pvname));
    if ( (p = (char *)index (pvname, '.')) != NULL)
	*p = NULL;
    strcat (pvname, ".MDEL");
    if ( ((dbNameToAddr (pvname, &dbAddr)) != 0) ||
	 ((dbGetField (&dbAddr, DBR_FLOAT, (caddr_t)&mval_mdel,
	               &options, &nRequest, NULL)) != 0) )
    {
	mval_mdel = 0;
    }
    p = (char *)index (pvname, '.');
    *p = NULL;
    strcat (pvname, ".ADEL");
    if ( ((dbNameToAddr (pvname, &dbAddr)) != 0) ||
	 ((dbGetField (&dbAddr, DBR_FLOAT, (caddr_t)&mval_adel,
		       &options, &nRequest, NULL)) != 0) )
    {
	mval_adel = 0;
    }

    monitor_mask = 0;
    delta = ppid -> lmmv - ppid -> mval;
    if (delta < 0.0)
    delta = -delta;
    if (delta > mval_mdel)
    {
	monitor_mask |= DBE_VALUE;
	ppid -> lmmv = ppid -> mval;
    }
    if (ppid -> tenb)
    {
	delta = ppid -> lamv - ppid -> mval;
	if (delta < 0.0)
	    delta = -delta;
	/* if exceed archive deadband or there has been a change in alarm status */
	if  ( (delta > mval_adel) || (stat != ppid -> stat) || 
	      (sevr != ppid -> sevr) )
	{
	    monitor_mask |= DBE_LOG;
	    ppid -> lamv = ppid -> mval;
	}
    }
    if (monitor_mask)
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid -> mval, monitor_mask);
		
    if (mval != ppid->mval)
	redQField (redbuf, mylist [MVAL]);

    /* only use deadbands if we are getting our setpt from from an upstream
       block and we are inclosed cascade
    */
    monitor_mask = 0;
    if ( (ppid -> stpl.type == DB_LINK) && (ppid -> uscs == PID_CASCLOSED) )
    {
	delta = ppid -> lmas - ppid -> astp;
	if (delta < 0.0)
	    delta = -delta;
	if (delta > mval_mdel)
	{
	    monitor_mask |= DBE_VALUE;
	    ppid -> lmas = ppid -> astp;
	}
	if (ppid -> tenb)
	{
	    delta = ppid -> laas - ppid -> astp;
	    if (delta < 0.0)
		delta = -delta;
	    /* if exceed archive deadband or there has been a change in alarm status */
	    if  ( (delta > mval_adel) || (stat != ppid -> stat) || 
		  (sevr != ppid -> sevr) )
	    {
		monitor_mask |= DBE_LOG;
		ppid -> laas = ppid -> astp;
	    }
	}
    }
    else
    {
	if (ppid -> lmas != ppid -> astp)
	{
	    monitor_mask |= DBE_VALUE;
	    ppid -> lmas = ppid -> astp;
	}
	if (ppid -> tenb)
	{
	    /* if setpt has changed or there has been a change in alarm status */
	    if ( (ppid -> laas != ppid -> astp) || (stat != ppid -> stat) ||
		 (sevr != ppid -> sevr) )
	    {
		monitor_mask |= DBE_LOG;
		ppid -> laas = ppid -> astp;
	    }
	}
    }
    if (monitor_mask)
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid -> astp, monitor_mask);

    /* we always want to redundantly update ASTP and MVAL if they
       have changed
    */
    if (ppid -> sstp != ppid -> astp)
    {
	redQField (redbuf, mylist [ASTP]);
	ppid -> sstp = ppid -> astp;
    }

    if (nmv != ppid->nmv)
	redQField (redbuf, mylist [NMV]);
		
    if (nsp != ppid->nsp)
	redQField (redbuf, mylist [NSP]);

    if (ct != ppid->ct)
    {
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->ct,DBE_VALUE);
	redQField (redbuf, mylist [CT]);
    }
		
    if (err != ppid->err)
    {
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->err,DBE_VALUE);
	redQField (redbuf, mylist [ERR]);
    }
		
    if (ersq != ppid->ersq)
    {
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->ersq,DBE_VALUE);
	redQField (redbuf, mylist [ERSQ]);
    }
		
    if (derr != ppid->derr)
    {
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->derr,DBE_VALUE);
	redQField (redbuf, mylist [DERR]);
    }
		
    if (dval != ppid->dval)
    {
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->dval,DBE_VALUE);
	redQField (redbuf, mylist [DVAL]);
    }
		
    if (val != ppid->val)
	redQField (redbuf, mylist [VAL]);
		
    if (pval != ppid->pval)
    {
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->pval,DBE_VALUE);
	redQField (redbuf, mylist [PVAL]);
    }
		
    if (lalm != ppid->lalm)
	redQField (redbuf, mylist [LALM]);
		
    if (lmde != ppid->lmde)
    {
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->lmde,DBE_VALUE);
	redQField (redbuf, mylist [LMDE]);
    }
                
    if (mode != ppid->mode)
    {
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->mode,DBE_VALUE);
	redQField (redbuf, mylist [MODE]);
    }
                
    if (spt != ppid->spt) 
    {
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->spt,DBE_VALUE);
	redQField (redbuf, mylist [SPT]);
    }
    if (am != ppid -> am)
    {
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid -> am, DBE_VALUE);
	redQField (redbuf, mylist [AM]);
    }
                
    if (fgcr != ppid->fgcr) 
    {
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->fgcr,DBE_VALUE);
	redQField (redbuf, mylist [FGCR]);
    }
                
    if (sndf != ppid->sndf) 
    {
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->sndf,DBE_VALUE);
	redQField (redbuf, mylist [SNDF]);
    }
                
    if (gact != ppid->gact) 
    {
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid->gact,DBE_VALUE);
	redQField (redbuf, mylist [GACT]);
    }
                
    if (csp != ppid->csp) 
	redQField (redbuf, mylist [CSP]);

    if (istp != ppid -> istp)
    {
	redQField (redbuf, mylist [ISTP]);
	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid -> istp, DBE_VALUE);
    }
               
    if (TsCmpStampsNE (&latm, &ppid -> latm) )
	redQField (redbuf, mylist [LATM]);

    if (mlst != ppid->mlst)
	redQField (redbuf, mylist [MLST]);

    if (dvct != ppid->dvct)
	redQField (redbuf, mylist [DVCT]);

    if (hloc != ppid->hloc)
	redQField (redbuf, mylist [HLOC]);

    if (hlsc != ppid->hlsc)
	redQField (redbuf, mylist [HLSC]);
	
    redQSend (redbuf);

    /* process the forward scan link record */

    if (ppid->flnk.type==DB_LINK)
	dbScanPassive((struct dbAddr *)ppid->flnk.value.db_link.pdbAddr);

    return(status);
}


/***************************** ramp_setpoint **********************************/

static void ramp_setpoint (ppid, deltatime)
    struct pidRecord	*ppid;
    float		deltatime;
{
    float    ramprate;
    float    temp;

       /* calc ramprate in egus/sec */
    ramprate = deltatime * ppid -> srmp;

	/* cannot allow a negative ramprate */
    if (ramprate < 0.0)
	ramprate = -ramprate;

    temp = ppid -> astp - ppid -> spt;
    if (temp < 0.0)
	temp = -temp;
    if ( (temp <= ramprate) || (ppid -> srmp == 0) )
    {
	   /* normailize the setpoint */
	ppid -> nsp = (ppid -> astp * ppid -> span) - ppid -> zero;
	ppid -> spt = ppid -> astp;
    }
    else
    {
	if (ppid -> astp < ppid -> spt)
	{
	    ppid -> nsp = ((ppid -> spt - ramprate) * ppid -> span) - 
			   ppid -> zero;
	    ppid -> spt = ppid -> spt - ramprate;
	}
	else
	{
	    ppid -> nsp = ((ppid -> spt + ramprate) * ppid -> span) - 
			   ppid -> zero;
	    ppid -> spt = ppid -> spt + ramprate;
	}
    }

    return;
}


/************************************* special *******************************/

static long special (paddr, after)
    struct dbAddr    *paddr;
    int              after;
{
    struct pidRecord	*ppid = (struct pidRecord *) (paddr -> precord);
    int			special_type = paddr -> special;
    short		master = 0;
    int			gap_crossing_changed = 0;
    char		redbuf[400];
    float		span = ppid -> span;
    float		zero = ppid -> zero;
    float		ngap = ppid -> ngap;
    short		fgcr = ppid -> fgcr;
    short		gact = ppid -> gact;
    float		lmas = ppid -> lmas;

    if (!after)
	return (0);

    resGetCaAnswer (ppid -> radr, &master);

    switch (special_type)
    {
        case SPC_PID_SPAN:
	    if ((ppid -> eguf - ppid -> egul) == 0)
		ppid -> span = 0;
	    else
		ppid -> span = 1 / (ppid -> eguf - ppid -> egul);
	    ppid -> zero = ppid -> egul * ppid -> span;
	    ppid -> ngap = ppid -> sptg * ppid -> span;
		/* can not let it go negative */
	    if (ppid -> ngap < 0.0)
		ppid -> ngap = -ppid -> ngap;
	    if (span != ppid -> span)
	    	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid -> span, DBE_VALUE);
	    if (zero != ppid -> zero)
	    	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid -> zero, DBE_VALUE);
	    if (ngap != ppid -> ngap)
	    	db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid -> ngap, DBE_VALUE);
	    break;

	case SPC_PID_GAP:
	    ppid -> ngap = ppid -> sptg * ppid -> span;
		/* can not let it go negative */
	    if (ppid -> ngap < 0.0)
		ppid -> ngap = -ppid -> ngap;
	    db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid -> ngap, DBE_VALUE);
	    break;

	case SPC_PID_SPTC:
	    gap_crossing_changed = 1;
	    break;

	case SPC_PID_CT:
	    ppid -> ct = tickGet();
	    break;

	/* the astp may have been chg'd by a graphic even though it is
	   gotten from a downstream point, we must update LMAS
	   (last astp monitored) so that when the astp is finally 
	   read from STPL, it will be posted, also note below
	   LMAS is redundant
	*/
	case SPC_PID_ASTP:
	    if ( (ppid->stpl.type == DB_LINK) && 
		 (ppid->uscs == PID_CASCLOSED) ) 
	    { 
		ppid -> lmas = ppid -> astp;
	    }
	    /* if we are SLAVE, save setpoint, astp, in sstp */
	    if (!master)
		ppid -> sstp = ppid -> astp;
	    break;


	default:
	    recGblDbaddrError (S_db_badChoice, paddr, "pid: special");
	    return (S_db_badChoice);
    }

    /* if the gap or gap crossing has changed,
       if we are in AUTOMATIC mode and are now in the gap,
       if setpoint crossing is NO, then it doesn't matter about FGCR
	   and GACT, treat this as 1st time in gap
           reset first crossing and run gap algorithm
       else if SPTC is YES, first gap crossing is 0, we were not in 
	   the gap, now we are, run gap algorithm
       else SPTC == YES and FGCR == 1 and gap active == 0,
	   we are doing setpoint crossing, we were in the gap and
	   still are, but we have not crossed the setpoint, by
	   running the gap algorithm, we remember what side of
	   the setpoint we are on, forget where we have been we
	   can't know that now
    */
	
    if ( (ngap != ppid -> ngap) || gap_crossing_changed)
    {
    	float	terr = ppid -> err;

	if (terr < 0.0)
	    terr = -terr;
	if ( (ppid -> mode == PID_MODE_AUTO) && (terr <= ppid -> ngap) )
	{
	    if ( (ppid -> sptc == NO) || (ppid -> fgcr == 0) ||
		 (ppid -> gact == 0) )
	    {
		ppid -> fgcr = 0;
		ppid -> gact = 0;
		if (fgcr != ppid -> fgcr)
		    db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid -> fgcr, DBE_VALUE);
		if (gact != ppid -> gact)
		    db_post_events ((struct dbCommon *)ppid, (caddr_t)&ppid -> gact, DBE_VALUE);
	    }
	}
    }

    if (master)
    {
	redQInit ((struct dbCommon *)ppid, redbuf, sizeof (redbuf), pointtypeindex);
	if (span != ppid -> span)
	    redQField (redbuf, mylist [SPAN]);
	if (zero != ppid -> zero)
	    redQField (redbuf, mylist [ZERO]);
	if (ngap != ppid -> ngap)
	    redQField (redbuf, mylist [NGAP]);
	if (fgcr != ppid -> fgcr)
	    redQField (redbuf, mylist [FGCR]);
	if (gact != ppid -> gact)
	    redQField (redbuf, mylist [GACT]);
	if (lmas != ppid -> lmas)
	    redQField (redbuf, mylist [LMAS]);
	redQSend (redbuf);
    }

    return (0);
}  

