/*recPump.c,v 1.1.1.1 1999/08/03 14:44:20 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************************
 *
 *  MODULE NAME            : recPump.c 
 *  AUTHOR                 : emsteel
 *  DATE                   : 11 May 93
 *
 *
 *  DESCRIPTION: record support routines for a Pump record
 *
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 02 Mar 94
 *  REVISION: added process stats code
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 29 Mar 94
 *  REVISION:  make forward func def agree with actual func def
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 10 May 94
 *  REVISION:  added cmd fields to redund
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 13 June 94
 *  REVISION:	added support for maximize severity, moved call to tsLocalTime
 *		inside on scan
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 31 Aug 94
 *  REVISION:	added get_graphic_double() and get_control_double()
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 13 Oct 94
 *  REVISION:	added TIME to namelist, monitor(), do tsLocalTime if status
 *		chgs, do not do in process()
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 20 Oct 94
 *  REVISION:	chg time to latm (last alarm time)
 *		added lptm (last process time)
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 6 Feb 95
 *  REVISION:	use VAL to store PASV, so it can be trended
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 Feb 95
 *  REVISION:	made VAL redundant
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 09 Feb 96
 *  REVISION:	added code to support accumulating pump running time
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 12 Feb 96
 *  REVISION:	added code to support minimum off time (MOTM)
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 20 Jun 96
 *  REVISION:	adde code for alarm disable ADIS
 *
****************************************************************************/

/****************************************************************************
 *
 * Programmer's Notes:
 *       1) values for ppump->tstk[] are the alarm values for each sensor
 *          PRESSURE, TEMPERATURE, VIBRATION found in ../share/epicsH/alarm.h
 *
****************************************************************************/


#include <vxWorks.h>
#include <inetLib.h>
#include <tickLib.h>       /* for tickGet () */
#include <types.h>
#include <stdioLib.h>
#include <lstLib.h>
#include <string.h>

#include "alarm.h"
#include "choice.h"
#include "cvtTable.h"
#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#include "dbAccess.h"
#include "dbFldTypes.h"
#include "dbRecType.h"
#include "devSup.h"
#include "errMdef.h"
#include "recSup.h"
#include "special.h"
#include "rec/pumpRecord.h"
#include "redflds.h"
#include "redque.h"
#include "pumpredfldidx.h"
#include "tsDefs.h"          
#include "tankfarm.h"
#include "restable.h"
#include "db_addr.h"
#include "recGbl.h"
#include "dbEvent.h"
#include "okToUse.h"
#include "dbcSub.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table */

#define report NULL
#define initialize NULL
static long init_record ();
static long process ();
static long special ();
#define get_value NULL
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
#define get_units NULL
static long get_precision ();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL 
static long get_graphic_double ();
static long get_control_double ();
#define get_alarm_double NULL

struct rset pumpRSET = {

    RSETNUMBER,
    report,
    initialize,
    init_record,
    process,
    special,
    get_value,
    cvt_dbaddr,
    get_array_info,
    put_array_info,
    get_units,
    get_precision,
    get_enum_str,
    get_enum_strs,
    put_enum_str,
    get_graphic_double,
    get_control_double,
    get_alarm_double
    };


/* The folowing definitions are for redundancy */
 
static struct fldDes **mylist = 0;

   /* NOTE: the order and spelling in namelist must be the
            same as in pumpredfldidx.h
   */

static char *namelist [] = {
               "STAT",
               "SEVR",
               "PUST",
               "SWST",
               "PASV",
               "SWSV",
               "LMOD",
               "ESTP",
               "EXON",
               "EXOF",
               "TTIM",
               "SPED",
               "SMOD",
               "LMEN",
               "LCMD",
               "LSTA",
               "LSEV",
               "PLTR",
	       "PCMD",
	       "INI2",
	       "CMDS",
	       "CSTR",
	       "CSTP",
	       "STRS",
	       "STPS",
	       "LATM",
	       "VAL",
	       "SVRT",
	       "ACRT",
	       "LPRT",
	       "LSVL",
               0
               };

   /* structure to save some of the values in the
      record at start of processing, most are used
      for redunadancy checking
   */

struct pumpred
{
    unsigned short	stat;
    unsigned short	sevr;
    double		val;
    char		pust[18];
    char		swst[18];
    char		pcmd[18];
    short		pasv;
    short		swsv;
    short		lmod;
    short		estp;
    double		exon;
    double		exof;
    double		ttim;
    double		sped;
    unsigned long	tcks;
    unsigned short	smod;
    unsigned short	lmen;
    short		lcmd;
    unsigned short	lsta;
    unsigned short	lsev;
    unsigned short	pltr;
    short		ini2;
    char		cmds[18];
    short		cstr;
    short		cstp;
    short		strs;
    short		stps;
    TS_STAMP		latm;
    unsigned short	lsvl;
};

/* define data structure used in various places, set in process */

struct pump_data
{
    short    permissive;          /* set by read_permissive_pv () */
    short    pump_line_mode;      /* pump is single or dual bit mode */
    short    pressure_tripped;    /* set by read_pressure_pv () */
    short    run_confirmed;       /* set by read_run_confirm_pv () */
    short    run_confirm_installed; /* set by read_run_confirm_pv () */
    short    temperature_tripped; /* set by read_temperature_pv () */
    short    remote_local_trans;  /* = 1 if going from remote to local */
    short    vibration_tripped;   /* set by read_vibration_pv () */
};

static unsigned short type;          /* index to pump type for redundancy */

/* defines for pump_line_mode */

#define	SINGLE_BIT_MODE		1
#define	DUAL_BIT_MODE		2

/* defines for learn mode */

#define	PUMP_NOT_LEARNING	0
#define	PUMP_START_LEARNING	1
#define	PUMP_STOP_LEARNING	2

/* defines for pump commands
   NOTE: these must have these values to
         coincide with pumpRecord.ascii
*/

#define	PUMP_START_CMD		1
#define	PUMP_STOP_CMD		2

/* defines for trip alarm stack */

#define	DELETE_TRIP_ALARM	0
#define	ADD_TRIP_ALARM		1

static void alarm (struct pumpRecord *, int);
static long check_current_pump_status (struct pumpRecord *, int *, double, struct pump_data *);
static long check_line_mode (struct pumpRecord *, struct pump_data *);
static long check_local_tracking (struct pumpRecord *, double, struct pump_data *);
static long check_trips (struct pumpRecord *, int *);
static long emergency_stop (struct pumpRecord *, short);
static void monitor (struct pumpRecord *);
static void post_events (struct pumpRecord *, struct pumpred *);
static long read_pressure_pv (struct pumpRecord *, struct pump_data *);
static long read_rl_switch (struct pumpRecord *, struct pump_data *);
static long read_run_confirm_pv (struct pumpRecord *, struct pump_data *);
static long read_sensors (struct pumpRecord *, struct pump_data *);
static long read_temperature_pv (struct pumpRecord *, struct pump_data *);
static long read_vibration_pv (struct pumpRecord *, struct pump_data *);
static void reset_pump_commands (struct pumpRecord *);
static long run_commands (struct pumpRecord *, int *, struct pump_data *);
static void save_pump (struct pumpRecord *, struct pumpred *);
static long talk_to_pump (struct pumpRecord *, unsigned short *, unsigned short *, short);
static void trip_alarms (struct pumpRecord *, int, int);
static void update_redun (struct pumpRecord *, struct pumpred *);



/****************************************************************************
 *
 *  MODULE NAME            : init_record 
 *  AUTHOR                 : emsteel
 *  DATE                   : 04 May 93
 *
 *  ARGUMENT(S)            : ppump - ptr to pump record  
 *
 *  DESCRIPTION: initialize a pump record
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 dec 93
 *  REVISION:  init PCMD (last command), TF_INIT_CMD defined in
 *             .../share/epicsH/tankfarm.h
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 03 Jan 94
 *  REVISION:  set ack to 0
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 28 Feb 94
 *  REVISION: removed INIT from redundant update, not needed, 
 *            removed ack = 0 from init
 *            only db_post_events if MASTER, use init flag to test, we are
 *              MASTER when process is called, and we will not process unless
 *              the init flag is set
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 13 Jan 94
 *  REVISION:	init lptm and latm
 *
****************************************************************************/

static long init_record (ppump)
   struct pumpRecord	*ppump;

{
    long              status = 0;
    unsigned short    i;
    char              *pstr;
    struct pumpred    pump_old;
    struct pump_data  pumpData;
    int		      ini2_flag = 0;

    if (!(mylist))
    {
         /* find record type for pump */
       for (type = 1, i = 0; type < 100; type++)
       {
          pstr = GET_PRECTYPE (type);     
          if (strncmp (pstr, "pump", 4) == 0)
          {
             i = 1;
             break;
          }
       }

       if (i == 0)
       {
          recGblRecordError (S_db_noRecType, (struct dbCommon *)ppump, 
				"pump:init_record");
          return (S_db_noRecType);
       }
         /* allocate memory for mylist */
       mylist = redFldAlloc (type, NUMPUMPFLDS);

       if (!mylist)
       {
          recGblRecordError (S_rec_Alloc, (struct dbCommon *)ppump,
			   "pump, init_record, mylist failed");
  	  return (S_rec_Alloc);
       }

         /* add fields in namelist to mylist */
       for (i = 0; i < NUMPUMPFLDS; i++)
            redFldAdd (type, i, namelist [i]);
    }

    ppump -> init = TRUE;
    ppump -> nsta = NO_ALARM;
    ppump -> nsev = NO_ALARM;
    ppump -> msev = NO_ALARM;
    if (ppump -> latm.secPastEpoch == 0)
	tsLocalTime (&ppump -> latm);
    if (ppump -> lptm.secPastEpoch == 0)
	tsLocalTime (&ppump -> lptm);

    /* save current pump values */
    save_pump (ppump, &pump_old);

    /* init some values */
    ppump -> lmod = PUMP_NOT_LEARNING;        /* learn mode */
    ppump -> ttim = 0;                        /* travel timer */
    bzero (ppump -> cmds, sizeof (ppump -> cmds));
    ppump -> cstr = 0;
    ppump -> cstp = 0;
    ppump -> strs = 0;
    ppump -> stps = 0;
    ppump -> prec = 2;                        /* display precision */
    ppump -> smod = ppump -> mode;            /* save mode */
    ppump -> lsta = NO_ALARM;
    ppump -> lsev = NO_ALARM;
    ppump -> tstk[0] = NO_ALARM;              /* set trip stack to NO_ALARM */
    ppump -> tstk[1] = NO_ALARM;
    ppump -> tstk[2] = NO_ALARM;
    strncpy (ppump -> pcmd, TF_INIT_CMD, sizeof (ppump -> pcmd));
    ppump -> pltr = ppump -> lten;            /* rememnber current local tracking */
    ppump -> lsvl = 0;				/* last stop value */
    ppump -> ralm = 0;				/* reset alarm switch */

    /* init tick counter for acrt to current */
    ppump -> tcks = tickGet ();
    /* svrt is saved in pump.dbf, so init acrt with it */
    ppump -> acrt = ppump -> svrt;
    /* force post events on acrt */
    ppump -> lprt = ppump -> acrt + 99;
       
      /* read the sensors/switches */
    status = read_sensors (ppump, &pumpData);
    if (!status)
    {
	int dont_init = 0;

	/* what state is the pump currently in, if no run_confirmed PV is installed
	   must assume pump is stopped, whether it realy is or not
	*/
	if (pumpData.run_confirm_installed)
	{
	    if (pumpData.run_confirmed)
	    {
		strncpy (ppump -> pust, ppump -> ston, sizeof (ppump -> pust));
		ppump -> pasv = PUMP_RUNNING;
		ppump -> lcmd = PUMP_RUNNING;
	    }
	    else
	    {
		strncpy (ppump -> pust, ppump -> stof, sizeof (ppump -> pust));
		ppump -> pasv = PUMP_STOPPED;
		ppump -> lcmd = PUMP_STOPPED;
	    }
	}
	else
	{
	    strncpy (ppump -> pust, ppump -> stof, sizeof (ppump -> pust));
	    ppump -> pasv = PUMP_STOPPED;
	    ppump -> lcmd = PUMP_STOPPED;
	}
	ppump -> val = ppump -> pasv;
	
	/* if output PVs are not init'd then we can not init */

	if (!okToUse (&ppump -> ponp) )
	    dont_init = 1;
	if (pumpData.pump_line_mode == DUAL_BIT_MODE)
	{
	    if (!okToUse (&ppump -> pofp) )
		dont_init = 1;
	}
	if (dont_init)
	{
	    ppump -> nsta = NO_INIT_ALARM;
	    ppump -> nsev = VALID_ALARM;
	}
	else
	    ini2_flag = 1;
    }
    else
    {
	ppump -> nsta = NO_INIT_ALARM;
	ppump -> nsev = VALID_ALARM;
    }

      /* do redundancy if not in slave mode */

    if (resGetCaAnswer (ppump -> radr, &i) == RES_OK)
    {
	if (i)
	{
	    /* check for maximize severity */

	    if (ppump -> msev > ppump -> nsev)
	    {
		ppump -> nsev = ppump -> msev;
		ppump -> nsta = LINK_ALARM;
	    }

	    monitor (ppump);

	    if (ini2_flag)
	    {
		ppump -> ini2 = 1;
		post_events (ppump, &pump_old);
		update_redun (ppump, &pump_old);
	    }
	}
    }

    return (0);

} /* end func init_records () */



/****************************************************************************
 *
 *  MODULE NAME            : alarm
 *  AUTHOR                 : emsteel
 *  DATE                   : 28 Oct 93
 *
 *  CALLING MODULE:        : process
 *  ARGUMENT(S)            : ppump - ptr to pump record  
 *
 *  DESCRIPTION: perpetuate pump alarms if needed
 *               if alarm_reset or nsev >= sevr then normal processing will
 *                  handle
 *               if nsta == a start or stop alarm, set last time alarm 
 *                  lsta and lsev
 *		 if ralm, reset alarm is non-zero, and stat is one of these
 *		    choices, then the operator has "acknowledged" them so we
 *		    can clear them
 *
****************************************************************************/

static void alarm (ppump, alarm_reset)
    struct pumpRecord    *ppump;
    int			alarm_reset;
{
    int	adis = 0;

    adis = AlarmDisableLink ((void *)ppump);
    if ( (ppump -> aenb == ALARM_DISABLED) || (adis != 0) )
    {
	ppump -> nsta = INHIBIT_ALARM;
	ppump -> nsev = NO_ALARM;
	return;
    }
    if (alarm_reset)
      return;
    if (ppump -> nsev >= ppump -> sevr)
    {
	if ((ppump -> nsta == PUMP_START) || (ppump -> nsta == PUMP_STOP))
	{
	    ppump -> lsta = ppump -> nsta;
	    ppump -> lsev = ppump -> nsev;
	}
	return;
    }
    if (ppump -> ralm)
    {
	ppump -> ralm = 0;
	if ( (ppump -> stat == PUMP_START) ||
	     (ppump -> stat == PUMP_STOP)  ||
	     (ppump -> stat == CMD_NOT_ALLOWED) ||
	     (ppump -> stat == TF_UNCOMMANDED) ||
	     (ppump -> stat == AUTOMATIC) ||
	     (ppump -> stat == MANUAL_OVERRIDE) ||
	     (ppump -> stat == LINK_ALARM) )
	    return;
    }
    ppump -> nsta = ppump -> stat;
    ppump -> nsev = ppump -> sevr;
   
    return;

} /* end func alarm () */



/****************************************************************************
 *
 *  MODULE NAME            : check_current_pump_status
 *  AUTHOR                 : emsteel
 *  DATE                   : 03 June 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : ppump - ptr to pump record
 *
 *  DESCRIPTION: update the pump state
 *  NOTE:
 *       delta_time = diff in  secs tween last scan and this scan, see
 *         process (), delta_time is static global this source
 *       ppump -> lmen = learn mode enable
 *         if ppump -> lmod == PUMP_NOT_LEARNING, then set exon/exof to
 *         ppump -> ttim, else only if ppump -> ttim > current exon/exof
 *       if pump was stopped etc check change of state if there indicators
 *         that give new state
 *
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 29 Nov 92
 *  REVISION:  if pump is starting and exceeds EXON then stop the pump by
 *             sending a 0 to the start pv and a 1 to the stop pv, if single
 *             bit, the 0 to the start pv will stop the pump
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 17 dec 93
 *  REVISION:  only do this if in REMOTE not in LOCAL, status will be 
 *             updated in run_commands() by the LOCAL test
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 2 May 94
 *  REVISION: removed db_post_events calls, done in post_events()
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 02 April 95
 *  REVISION:	if pump is running and loose run_confirmed, if LRCF is
 *		set, then call emergency_stop()
 *
****************************************************************************/

static long check_current_pump_status (ppump, alarm_reset, delta_time, pumpData)
    struct pumpRecord    *ppump;
    int			*alarm_reset;
    double		delta_time;
    struct pump_data	*pumpData;
{
    unsigned short	start_val;
    unsigned short	stop_val;
    long		status      = 0;
    int			uncommanded = 0;      /* pump started/stopped without
						 being commanded */

    if (ppump -> swsv == RL_LOCAL)
	return (status);

    switch (ppump -> pasv)
    {
	case PUMP_STOPPED:
	    if (pumpData -> run_confirm_installed)
	    {
		if (pumpData -> run_confirmed)
		{
		    strncpy (ppump -> pust, ppump -> ston,
			     sizeof (ppump -> pust));
		    ppump -> pasv = PUMP_RUNNING;
		    if (ppump -> lcmd == PUMP_STOPPED)
			uncommanded++;
		    else
		    {
			ppump -> lcmd = PUMP_RUNNING;
			*alarm_reset = 1;
			ppump -> lsta = ppump -> lsev = NO_ALARM;
		    }
		}
	    }
	    if ( (pumpData -> pump_line_mode == DUAL_BIT_MODE) &&
		 (ppump -> lsvl != 0) )
	    {
		start_val = 0;
		stop_val = 0;
		status = talk_to_pump (ppump, &start_val, &stop_val,
				       pumpData -> pump_line_mode);
	    }
	    ppump -> ttim = 0;
	    break;

	case PUMP_RUNNING:
	    if (pumpData -> run_confirm_installed)
	    {
		/* we have lost run confirmed */
		if (pumpData -> run_confirmed == 0)
		{
		    if (ppump -> lcmd == PUMP_RUNNING)
			uncommanded++;
		    else
		    {
			if (ppump -> motm > 0)
			{
			    strncpy (ppump -> pust, ppump -> stmo,
				     sizeof (ppump -> pust) );
			    ppump -> pasv = PUMP_MIN_OFF;
			}
			else
			{
			    strncpy(ppump -> pust, ppump -> stof,
				    sizeof (ppump -> pust));
			    ppump -> pasv = PUMP_STOPPED;
			}
			ppump -> lcmd = PUMP_STOPPED;
			*alarm_reset = 1;
			ppump -> lsta = ppump -> lsev = NO_ALARM;
		    }
		    /* do emergency stop on pump if told to do so */
		    if (ppump -> lrcf)
			emergency_stop (ppump, pumpData -> pump_line_mode);
		    else
		    {
			if (ppump -> motm > 0)
			{
			    strncpy (ppump -> pust, ppump -> stmo,
				     sizeof (ppump -> pust) );
			    ppump -> pasv = PUMP_MIN_OFF;
			}
			else
			{
			    strncpy(ppump -> pust, ppump -> stof,
				    sizeof (ppump -> pust));
			    ppump -> pasv = PUMP_STOPPED;
			}
		    }
		}
		ppump -> ttim = 0;
	    }
	    break;

	case PUMP_STARTING:
	    ppump -> ttim += delta_time;
	    if (pumpData -> run_confirm_installed)
	    {
		if (pumpData -> run_confirmed)
		{
		    if (pumpData -> pump_line_mode == DUAL_BIT_MODE)
		    {
			start_val = 0;
			stop_val = 0;
			status = talk_to_pump (ppump, &start_val, &stop_val, 
					       pumpData -> pump_line_mode);
		    }
		    strncpy (ppump -> pust, ppump -> ston,
			     sizeof (ppump -> pust));
		    ppump -> pasv = PUMP_RUNNING;
		    ppump -> lcmd = PUMP_RUNNING;
		    if (ppump -> lmen)
		    {
			if (ppump -> lmod & PUMP_START_LEARNING)
			{
			    if (ppump -> ttim > ppump -> exon)
				ppump -> exon = ppump -> ttim;
			}
			else
			{
			    ppump -> lmod |= PUMP_START_LEARNING;
			    ppump -> exon = ppump -> ttim;
			}
		    }
		    ppump -> ttim = 0;
		    *alarm_reset = 1;
		    ppump -> lsta = ppump -> lsev = NO_ALARM;
		}
		else if (ppump -> exon && (!ppump -> lmen))
		{
		    if (ppump -> ttim >
			(ppump -> exon + 
			((ppump -> exon * (double)ppump -> sadb)/100) ) )
		    {
			TankFarmPostAlarm ((struct dbCommon *)ppump, 
					    PUMP_START, ppump -> stsv);
			/* stop the pump regardless whether single or dual
			   bit, safety precaution for pump motor
			   ems, 11/29/93
			*/
			start_val = 0;
			stop_val = 1;
			status = talk_to_pump (ppump, &start_val, &stop_val, 
					       pumpData -> pump_line_mode);
			strncpy (ppump -> pust, ppump -> stfa,
			sizeof (ppump -> pust));
			ppump -> pasv = PUMP_FAULT;
			ppump -> ttim = 0;
		    }
		}
	    }
	    else if (ppump -> exon)
	    {
		if (ppump -> ttim >= ppump -> exon)
		{
		    if (pumpData -> pump_line_mode == DUAL_BIT_MODE)
		    {
			start_val = 0;
			stop_val = 0;
			status = talk_to_pump (ppump, &start_val, &stop_val, 
					       pumpData -> pump_line_mode);
		    }
		    strncpy (ppump -> pust, ppump -> ston,
			     sizeof (ppump -> pust));
		    ppump -> pasv = PUMP_RUNNING;
		    ppump -> lcmd = PUMP_RUNNING;
		    ppump -> ttim = 0;
		    *alarm_reset = 1;
		    ppump -> lsta = ppump -> lsev = NO_ALARM;
		}
	    }     
	    break;

	case PUMP_STOPPING:
	    ppump -> ttim += delta_time;
	    if (pumpData -> run_confirm_installed)
	    {
		if (pumpData -> run_confirmed == 0)
		{
		    if (pumpData -> pump_line_mode == DUAL_BIT_MODE)
		    {
			start_val = 0;
			stop_val = 0;
			status = talk_to_pump (ppump, &start_val, &stop_val, 
			                       pumpData -> pump_line_mode);
		    }
		    if (ppump -> motm > 0)
		    {
			strncpy (ppump -> pust, ppump -> stmo,
				 sizeof (ppump -> pust) );
			ppump -> pasv = PUMP_MIN_OFF;
		    }
		    else
		    {
			strncpy (ppump -> pust, ppump -> stof,
				 sizeof (ppump -> pust) );
			ppump -> pasv = PUMP_STOPPED;
		    }
		    ppump -> lcmd = PUMP_STOPPED;
		    if (ppump -> lmen)
		    {
			if (ppump -> lmod & PUMP_STOP_LEARNING)
			{
			    if (ppump -> ttim > ppump -> exof)
			        ppump -> exof = ppump -> ttim;
			}
			else
			{
			    ppump -> lmod |= PUMP_STOP_LEARNING;
			    ppump -> exof = ppump -> ttim;
			}
		    }
		    ppump -> ttim = 0;
		    *alarm_reset = 1;
		    ppump -> lsta = ppump -> lsev = NO_ALARM;
		}
		else if (ppump -> exof && (!ppump -> lmen))
		{
		    if (ppump -> ttim >
			(ppump -> exof + 
			((ppump -> exof * ppump -> stdb)/100) ) )
		    {
			TankFarmPostAlarm ((struct dbCommon *)ppump, 
					   PUMP_STOP, ppump -> spvs);
			if (pumpData -> pump_line_mode == DUAL_BIT_MODE)
			{
			    start_val = 0;
			    stop_val = 0;
			    status = talk_to_pump (ppump, &start_val, &stop_val, 
						   pumpData -> pump_line_mode);
			}
			strncpy (ppump -> pust, ppump -> stfa,
				 sizeof (ppump -> pust));
			ppump -> pasv = PUMP_FAULT;
			ppump -> ttim = 0;
		    }
		}
	    }
	    else if (ppump -> exof)
	    {
		if (ppump -> ttim >= ppump -> exof)
		{
		    if (pumpData -> pump_line_mode == DUAL_BIT_MODE)
		    {
			start_val = 0;
			stop_val = 0;
			status = talk_to_pump (ppump, &start_val, &stop_val,
					       pumpData -> pump_line_mode);
		    }
		    if (ppump -> motm > 0)
		    {
			strncpy (ppump -> pust, ppump -> stmo,
				 sizeof (ppump -> pust) );
			ppump -> pasv = PUMP_MIN_OFF;
		    }
		    else
		    {
			strncpy (ppump -> pust, ppump -> stof,
				 sizeof (ppump -> pust));
			ppump -> pasv = PUMP_STOPPED;
		    }
		    ppump -> lcmd = PUMP_STOPPED;
		    ppump -> ttim = 0;
		    *alarm_reset = 1;
		    ppump -> lsta = ppump -> lsev = NO_ALARM;
		}
	    }
	    break;

	case PUMP_FAULT:
	    if (pumpData -> run_confirm_installed)
	    {
		if (pumpData -> run_confirmed)
		{
		    strncpy (ppump -> pust, ppump -> ston,
			     sizeof (ppump -> pust));
		    ppump -> pasv = PUMP_RUNNING;
		    if (ppump -> lcmd != PUMP_STOPPING)
		    {
			ppump -> lcmd = PUMP_RUNNING;
			*alarm_reset = 1;
			ppump -> lsta = ppump -> lsev = NO_ALARM;
		    }
		    ppump -> ttim = 0;
		}
		else
		{
		    strncpy (ppump -> pust, ppump -> stof,
			     sizeof (ppump -> pust));
		    ppump -> pasv = PUMP_STOPPED;
		    if (ppump -> lcmd != PUMP_STARTING)
		    {
			ppump -> lcmd = PUMP_STOPPED;
			*alarm_reset = 1;
			ppump -> lsta = ppump -> lsev = NO_ALARM;
		    }
		    ppump -> ttim = 0;
		}
		if (pumpData -> pump_line_mode == DUAL_BIT_MODE)
		{
		    start_val = 0;
		    stop_val = 0;
		    status = talk_to_pump (ppump, &start_val, &stop_val,
					   pumpData -> pump_line_mode);
		}
	    }
	    break;

	case PUMP_MIN_OFF:
	    if (pumpData -> run_confirm_installed)
	    {
		if (pumpData -> run_confirmed)
		{
		    strncpy (ppump -> pust, ppump -> ston,
			     sizeof (ppump -> pust) );
		    ppump -> pasv = PUMP_RUNNING;
		    if (ppump -> lcmd == PUMP_STOPPED)
			uncommanded++;
		    else
		    {
			ppump -> lcmd = PUMP_RUNNING;
			*alarm_reset = 1;
			ppump -> lsta = ppump -> lsev = NO_ALARM;
		    }
		    ppump -> ttim = 0;
		    break;
		}
	    }
	    if ( (pumpData -> pump_line_mode == DUAL_BIT_MODE) &&
		 (ppump -> lsvl != 0) )
	    {
		start_val = 0;
		stop_val = 0;
		status = talk_to_pump (ppump, &start_val, &stop_val,
				       pumpData -> pump_line_mode);
	    }
	    ppump -> ttim += delta_time;	
	    /* motm is in minutes, so divided accumulated off time (TTIM)
	       in seconds by 60 to convert to minutes
	    */
	    if ((ppump -> ttim/60.0) >= ppump -> motm)
	    {
		ppump -> ttim = 0;
		strncpy (ppump -> pust, ppump -> stof,
			 sizeof (ppump -> pust) );
		ppump -> pasv = PUMP_STOPPED;
		ppump -> lcmd = PUMP_STOPPED;
		*alarm_reset = 1;
		ppump -> lsta = ppump -> lsev = NO_ALARM;
	    }
	    break;
	default:
	    break;

    } /* end switch (pasv) */

   if (uncommanded)
   {
      TankFarmPostAlarm ((struct dbCommon *)ppump, TF_UNCOMMANDED,
                                                              MAJOR_ALARM);
   }

   return (status);

} /* end func check_current_pump_status () */



/****************************************************************************
 *
 *  MODULE NAME            : check_line_mode 
 *  AUTHOR                 : emsteel
 *  DATE                   : 11 May 93
 *
 *  CALLING MODULE(S)      : read_sensors
 *          ARGUMENT(S)    : ppump - ptr to pump record
 *
 *  DESCRIPTION: 1 - check for valid line mode
 *               2 - set  pump_line_mode accordingly
 *
 *  NOTE: you can not have a close pv without the open pv, but you
 *        can have an open pv without the close (SINGLE_BIT_MODE)
 *       
****************************************************************************/

static long check_line_mode (ppump, pumpData)
   struct pumpRecord	*ppump;
   struct pump_data	*pumpData;
{
   long  status    = 0;
   short line_mode = 0;

      /* check for open pump pv */

   if (ppump -> ponp.type != CONSTANT)
   {
      if (ppump -> ponp.type == DB_LINK)
      {
         line_mode = 1;
      }
   }
      /* check for close pump pv */

   if (ppump -> pofp.type != CONSTANT)
   {
      if (ppump -> pofp.type == DB_LINK)
      {
         line_mode |= 2;
      }
   }

   if (line_mode == 1)
      pumpData -> pump_line_mode = SINGLE_BIT_MODE;
   else if (line_mode == 3)
      pumpData -> pump_line_mode = DUAL_BIT_MODE;
   else
   {
      pumpData -> pump_line_mode = 0;
      recGblRecordError (S_tank_Line_Err, (struct dbCommon *)ppump, 
				"pump:check_line_mode");
      return (S_tank_Line_Err);
   }

   return (status);
   
} /* end func check_line_mode () */


/****************************************************************************
 *
 *  MODULE NAME            : check_local_tracking
 *  AUTHOR                 : emsteel
 *  DATE                   : 07 Feb 97
 *
 *  CALLING MODULE(S)      : process
 *
 *  DESCRIPTION:  we are in local mode, track our status and our outputs
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : 
 *  REVISION DATE          :
 *  REVISION:  
 *
****************************************************************************/

static long check_local_tracking (ppump, delta_time, pumpData)
    struct pumpRecord    *ppump;
    double		delta_time;
    struct pump_data	*pumpData;
{
    long		status = 0;
    short		save_pump_state = ppump -> pasv;
    unsigned short	start_val = 0;
    unsigned short	stop_val = 0;
    int			talktopump = 0;
    int			first_motm = 0;

    if (pumpData -> run_confirm_installed)
    {
	if (pumpData -> run_confirmed)
	{
	    if (ppump -> pasv != PUMP_RUNNING)
	    {
		strncpy (ppump -> pust, ppump -> ston, sizeof (ppump -> pust));
		ppump -> pasv = PUMP_RUNNING;
		ppump -> lcmd = PUMP_RUNNING;
		ppump -> ttim = 0;
	    }
	    if (pumpData -> pump_line_mode == SINGLE_BIT_MODE)
	    {
		if (ppump -> lten)
		{
		    if ( ( (save_pump_state != PUMP_STARTING) &&
			   (save_pump_state != PUMP_RUNNING) ) ||
			 (pumpData -> remote_local_trans) ||
			 (ppump -> pltr != ppump -> lten) )
		    {
			start_val = 1;
			stop_val = 0;
			talktopump++;
		    }
		}
	    }
	}
	else
	{
	    if  ( (ppump -> pasv != PUMP_STOPPED) &&
		  (ppump -> pasv != PUMP_MIN_OFF) )
	    {
		if ( (ppump -> motm > 0) &&
		     ( (ppump -> pasv == PUMP_RUNNING) || 
		     (ppump -> pasv == PUMP_STOPPING) ) )
		{
		    first_motm = 1;
		    strncpy (ppump -> pust, ppump -> stmo, sizeof (ppump -> pust));
		    ppump -> pasv = PUMP_MIN_OFF;
		    ppump -> ttim = 0;
		}
		else
		{
		    strncpy (ppump -> pust, ppump -> stof, sizeof (ppump -> pust));
		    ppump -> pasv = PUMP_STOPPED;
		}
		ppump -> lcmd = PUMP_STOPPED;
	    }
	    if ( (ppump -> pasv == PUMP_MIN_OFF) && (first_motm == 0) )
	    {
		ppump -> ttim += delta_time;
		/* motm is in minutes, so divided accumulated off time (TTIM)
		   in seconds by 60 to convert to minutes
		*/
		if ((ppump -> ttim/60.0) >= ppump -> motm)
		{
		    ppump -> ttim = 0;
		    strncpy (ppump -> pust, ppump -> stof,
		    sizeof (ppump -> pust) );
		    ppump -> pasv = PUMP_STOPPED;
		    ppump -> lcmd = PUMP_STOPPED;
		}
	    }
	    if (pumpData -> pump_line_mode == SINGLE_BIT_MODE)
	    {
		if (ppump -> lten)
		{
		    if ( ( (save_pump_state != PUMP_STOPPING) &&
			   (save_pump_state != PUMP_STOPPED) ) ||
			 (pumpData -> remote_local_trans) ||
			 (ppump -> pltr != ppump -> lten) )
		    {
			start_val = 0;
			stop_val = 1;
			talktopump++;
		    }
		}
	    }
	} 
    }  /* end run_confirmed installed */
    else
    {
	if (ppump -> pasv == PUMP_MIN_OFF)
	{
	    ppump -> ttim += delta_time;
	    /* motm is in minutes, so divided accumulated off time (TTIM)
	       in seconds by 60 to convert to minutes
	    */
	    if ((ppump -> ttim/60.0) >= ppump -> motm)
	    {
		ppump -> ttim = 0;
		strncpy (ppump -> pust, ppump -> stof, sizeof (ppump -> pust) );
		ppump -> pasv = PUMP_STOPPED;
		ppump -> lcmd = PUMP_STOPPED;
	    }
	}
	else if ( (ppump -> motm > 0) && (ppump -> pasv == PUMP_STOPPING) )
	{
		ppump -> ttim = 0;
		strncpy (ppump -> pust, ppump -> stmo, sizeof (ppump -> pust) );
		ppump -> pasv = PUMP_MIN_OFF;
	}
    }
    if ( (pumpData -> pump_line_mode == DUAL_BIT_MODE) &&
	     (pumpData -> remote_local_trans) )
    {
	start_val = 0;
	stop_val = 0;
	talktopump++;
    }
    ppump -> pltr = ppump -> lten;    /* update last local tracking */

    if (talktopump)
	status = talk_to_pump (ppump, &start_val, &stop_val, pumpData -> pump_line_mode);

    return (status);

} /* end func check_local_tracking */
 


/****************************************************************************
 *
 *  MODULE NAME            : check_trips 
 *  AUTHOR                 : emsteel
 *  DATE                   : 02 June 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : ppump - ptr to pump record
 *
 *  DESCRIPTION: alarm on oldest tripped sensor, return non 0 to process ()
 *               to initiate emergency shutdown if any trips enabled
 *               if not tripped now but was previous scan , reset alarm
 *                  and if there was a time alarm, resetablish time alarm 
 *
****************************************************************************/

static long check_trips (ppump, alarm_reset)
   struct pumpRecord    *ppump;
   int			*alarm_reset;
{
   long  trip_enabled = 0;
   int   i;

   if (ppump -> tstk[0] != NO_ALARM)
   {
      for (i = 2; i >=0; i--)
      {
         if (ppump -> tstk[i] != NO_ALARM)
         {
            TankFarmPostAlarm ((struct dbCommon *)ppump, ppump -> tstk[i],
                                                               MAJOR_ALARM);
            switch (ppump -> tstk[i])
            {
               case PRESSURE:
                  if (ppump -> ptre)
                     trip_enabled = 1;
                  break;

               case TEMPERATURE:
                  if (ppump -> ttre)
                     trip_enabled = 1;
                  break;
case VIBRATION:
                  if (ppump -> vtre)
                     trip_enabled = 1;
                  break;

               default:
                  break;
            }
            break;
         }
      }
   }
   else
   {
      if ( (ppump -> stat == PRESSURE   ) ||
           (ppump -> stat == TEMPERATURE) ||
           (ppump -> stat == VIBRATION  ) )
      {
         if (ppump -> lsta != NO_ALARM)
            TankFarmPostAlarm ((struct dbCommon *)ppump, ppump -> lsta,
                                                            ppump -> lsev);
         *alarm_reset = 1;
      }
   }

   return (trip_enabled);

} /* end func check_trips () */


/****************************************************************************
 *
 *  MODULE NAME            : emergency_stop 
 *  AUTHOR                 : emsteel
 *  DATE                   : 05 May 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : ppump - ptr to pump record
 *
 *  DESCRIPTION: if an emergency stop has been initiated, then handle
 *               the pump accordingly
 *  NOTE: the value (start_val, stop_val) has been defined as an
 *        unsigned short, since it assumed that the start/stop pv are binary
 *        records and their val fields are defined as DBR_ENUM which
 *        translates to unsigned short
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 08 Mar 94
 *  REVISION: do nothing if pump is also in a FAULT condition or it is in a
 *            STOP ALARM
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 2 May 94
 *  REVISION: removed db_post_events calls, done in post_events()
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 03 April 95
 *  REVISION:	test lcmd (last commnad)  instead of pasv (current status)
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 07 Feb 97
 *  REVISION:	if in local, only output CMD_NOT_ALLOWED if estp is set
 *
****************************************************************************/

static long emergency_stop (ppump, pump_line_mode)
   struct pumpRecord *ppump;
   short	     pump_line_mode;
{
    long           status = 0;
    unsigned short start_val;
    unsigned short stop_val;

    /* do nothing if already stopping or stopped */

    if ( (ppump -> lcmd == PUMP_STOPPED) || (ppump -> lcmd == PUMP_STOPPING) )
	return (status);

    /* we must be in remote mode to do anything software wise */

    if (ppump -> swsv != RL_REMOTE)
    {
	if (ppump -> estp)
	    TankFarmPostAlarm ((struct dbCommon *)ppump, CMD_NOT_ALLOWED, INFO_ALARM);
	return (status);
    }

    start_val = 0;
    stop_val = 1;
    status = talk_to_pump (ppump, &start_val, &stop_val, pump_line_mode);
    strncpy (ppump -> pust, ppump -> stsp, sizeof (ppump -> pust));
    ppump -> pasv = PUMP_STOPPING;
    ppump -> lcmd = PUMP_STOPPING;
    ppump -> ttim = 0;

    return (status);

} /* end func emergency_stop () */


/****************************************************************************
 *
 *  MODULE NAME            : get_control_double 
 *  AUTHOR                 : emsteel
 *  DATE                   : 31 Aug 94
 *
 *
 *  DESCRIPTION: return control limits
 *
****************************************************************************/

static long get_control_double(paddr,pcd)
    struct	dbAddr		*paddr;
    struct	dbr_ctrlDouble	*pcd;
{
    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct pumpRecord *)0) -> exon) ):
	case ( (int)&(((struct pumpRecord *)0) -> exof) ):
	    pcd -> upper_ctrl_limit = 86400;	/* secs/dasy */
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct pumpRecord *)0) -> sadb) ):
	case ( (int)&(((struct pumpRecord *)0) -> stdb) ):
	    pcd -> upper_ctrl_limit = 100;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct pumpRecord *)0) -> sped) ):
	    pcd -> upper_ctrl_limit = 1000000;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct pumpRecord *)0) -> estp) ):
	    pcd -> upper_ctrl_limit = 1;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct pumpRecord *)0) -> val) ):
	    pcd -> upper_ctrl_limit = 4;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct pumpRecord *)0) -> acrt) ):
	    pcd -> upper_ctrl_limit = 99999;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct pumpRecord *)0) -> motm) ):
	    pcd -> upper_ctrl_limit = 9999.9;
	    pcd -> lower_ctrl_limit = 0;
	    break;
    }

    return (0);

} /* end func get_control_double () */


/****************************************************************************
 *
 *  MODULE NAME            : get_graphic_double 
 *  AUTHOR                 : emsteel
 *  DATE                   : 31 Aug 94
 *
 *
 *  DESCRIPTION: return graphic limits
 *
****************************************************************************/

static long get_graphic_double(paddr,pgd)
    struct	dbAddr		*paddr;
    struct	dbr_grDouble	*pgd;
{
    struct pumpRecord    *ppump = (struct pumpRecord *) paddr->precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct pumpRecord *)0) -> pasv) ):
	case ( (int)&(((struct pumpRecord *)0) -> lcmd) ):
	case ( (int)&(((struct pumpRecord *)0) -> val) ):
	    pgd -> upper_disp_limit = 4;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct pumpRecord *)0) -> swsv) ):
	case ( (int)&(((struct pumpRecord *)0) -> estp) ):
	    pgd -> upper_disp_limit = 1;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct pumpRecord *)0) -> lmod) ):
	    pgd -> upper_disp_limit = 2;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct pumpRecord *)0) -> exon) ):
	case ( (int)&(((struct pumpRecord *)0) -> exof) ):
	    pgd -> upper_disp_limit = 86400;	/* secs/day */
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct pumpRecord *)0) -> sped) ):
	    pgd -> upper_disp_limit = 1000000;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct pumpRecord *)0) -> sadb) ):
	case ( (int)&(((struct pumpRecord *)0) -> stdb) ):
	    pgd -> upper_disp_limit = 100;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct pumpRecord *)0) -> ttim) ):
	{
	    double on_time = 0;
	    double off_time = 0;

	    on_time = ppump -> exon + ((ppump -> exon * ppump -> sadb) / 100);
	    off_time = ppump -> exof + ((ppump -> exof * ppump -> stdb) / 100);
	    pgd -> upper_disp_limit = (on_time > off_time) ? on_time : off_time;
	    pgd -> lower_disp_limit = 0;
	    break;
	}

	case ( (int)&(((struct pumpRecord *)0) -> acrt) ):
	    pgd -> upper_disp_limit = 99999;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct pumpRecord *)0) -> motm) ):
	    pgd -> upper_disp_limit = 9999.9;
	    pgd -> lower_disp_limit = 0;
	    break;
    }

    return (0);

} /* end func get_graphic_double () */


/****************************************************************************
 *
 *  MODULE NAME            : get_precision 
 *  AUTHOR                 : emsteel
 *  DATE                   : 01 July 93
 *
 *
 *  DESCRIPTION: get display precision
 *  NOTE:
 *       ppump -> prec set to default 2 in init_record ()
 *
****************************************************************************/

static long get_precision (paddr, precision)
    struct dbAddr     *paddr;
    long              *precision;
{
    struct pumpRecord *ppump = (struct pumpRecord *)paddr -> precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct pumpRecord *)0) -> acrt) ):
	    *precision = ppump -> prrt;
	    break;

	case ( (int)&(((struct pumpRecord *)0) -> motm) ):
	    *precision = 1;
	    break;

	default:
	    *precision = ppump -> prec;
	    break;
    }

   return (0);

} /* end func get_precision () */



/****************************************************************************
 *
 *  MODULE NAME            : monitor 
 *  AUTHOR                 : emsteel
 *  DATE                   : 11 May 93
 *
 *  CALLING MODULE(S)      : process ()
 *
 *  DESCRIPTION: check event list
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 05 Jan 94
 *  REVISION:  check VALID_ALARM before set ACK
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 2 May 94
 *  REVISION: removed db_post_events for VAL, VAL not used
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 30 Aug 94
 *  REVISION:	rm DBE_LOG from ack
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 15 Mar 95
 *  REVISION:	do not set ack if sevr is EVENT_ALARM
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 11 Sept 95
 *  REVISION:	archive if stat/sevr also chg's
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 12 Sept 95
 *  REVISION:	chg way handle VALID_ALARM
 *
****************************************************************************/

static void monitor (ppump)
   struct pumpRecord *ppump;
{
   short		stat, sevr, nsta, nsev;
   unsigned short	monitor_mask;
   int			set_ack = 1;

      /* get previous stat and sevr and new stat and sevr */

   stat = ppump -> stat;
   sevr = ppump -> sevr;
   nsta = ppump -> nsta;
   nsev = ppump -> nsev;

      /* set current stat and sevr */

   ppump -> stat = nsta;
   ppump -> sevr = nsev;
   ppump -> nsta = 0;
   ppump -> nsev = 0;

    /* if we are coming out of VALID_ALARM and our new stat/sevr are
       the same as when we went into VALID_ALARM, do not set ack
    */
    if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	 (nsta == ppump -> psta) && (nsev == ppump -> psev) )
    {
	set_ack = 0;
    }
    /* if we are going into VALID_ALARM, and out previous stat was not
       VALID_ALARM, remember previous non VALID_ALARM stat/sevr
       and we do not want to set ACK,
    */
    if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
    {
	set_ack = 0;
	ppump -> psta = stat;
	ppump -> psev = sevr;
	db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> psta, DBE_VALUE);
	db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> psev, DBE_VALUE);
    }

   monitor_mask = 0;

      /* did alarm condition change this scan */

   if ((stat != nsta) || (sevr != nsev))
   {
      monitor_mask = DBE_ALARM;

      /* when alarm status chg's, get current time and
         save VAL in lalm (last value when alarm chg'd)
      */
      tsLocalTime (&ppump -> latm);

         /* post stat and sevr fields */

      db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> stat, DBE_VALUE);
      db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> sevr, DBE_VALUE);
      if ((ppump -> sevr != NO_ALARM) && (set_ack) &&
	  (ppump -> sevr != EVENT_ALARM) )
      {
         if (ppump -> ack == 0)
         {
            ppump -> ack = 1;
            db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> ack, DBE_VALUE);
         }
      }
   }

    if (ppump -> val != ppump -> pasv)
    {
	monitor_mask |= DBE_VALUE;
    }

    if (ppump -> tenb)
    {
	if ( (ppump -> val != ppump -> pasv) || (monitor_mask & DBE_ALARM) )
	    monitor_mask |= DBE_LOG;
    }

    if (monitor_mask & (DBE_VALUE | DBE_LOG))
	ppump -> val = ppump -> pasv;

    if (monitor_mask)
	db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> val, monitor_mask);

   return;

} /* end func monitor () */


/****************************************************************************
 *
 *  MODULE NAME            : post_events 
 *  AUTHOR                 : emsteel
 *  DATE                   : 2 May 94
 *
 *
 *  DESCRIPTION: called fro init_record and process to post any events 
 *		 where a value has changed
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : 
 *  REVISION DATE          :
 *  REVISION: 
 *
****************************************************************************/

static void post_events (ppump, ppump_old)
    struct pumpRecord	*ppump;
    struct pumpred	*ppump_old;
{

    if (strncmp (ppump -> pust, ppump_old -> pust, sizeof (ppump -> pust)) != 0)
	db_post_events ((struct dbCommon *)ppump, ppump -> pust, DBE_VALUE);
    
    if (strncmp (ppump -> swst, ppump_old -> swst, sizeof (ppump -> swst)) != 0)
	db_post_events ((struct dbCommon *)ppump, ppump -> swst, DBE_VALUE);
    
    if (strncmp (ppump -> pcmd, ppump_old -> pcmd, sizeof (ppump -> pcmd)) != 0)
	db_post_events ((struct dbCommon *)ppump, ppump -> pcmd, DBE_VALUE);

    if (ppump -> pasv != ppump_old -> pasv)
	db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> pasv, DBE_VALUE);

    if (ppump -> swsv != ppump_old -> swsv)
	db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> swsv, DBE_VALUE);

    if (ppump -> lmod != ppump_old -> lmod)
	db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> lmod, DBE_VALUE);

    if (ppump -> exon != ppump_old -> exon)
	db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> exon, DBE_VALUE);

    if (ppump -> exof != ppump_old -> exof)
	db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> exof, DBE_VALUE);

    if (ppump -> ttim != ppump_old -> ttim)
	db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> ttim, DBE_VALUE);

    if (ppump -> tcks != ppump_old -> tcks)
	db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> tcks, DBE_VALUE);

    if (ppump -> lmen != ppump_old -> lmen)
	db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> lmen, DBE_VALUE);

    if (ppump -> lcmd != ppump_old -> lcmd)
	db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> lcmd, DBE_VALUE);

    if (ppump -> lsvl != ppump_old -> lsvl)
	db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> lsvl, DBE_VALUE);

    
    return;

} /* end func post_events () */


/****************************************************************************
 *
 *  MODULE NAME            : process 
 *  AUTHOR                 : emsteel
 *  DATE                   : 11 May 93
 *
 *
 *  DESCRIPTION: called when time to process a pump record
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 08 Mar 94
 *  REVISION: added PERMISSIVE_ALM if permissive is dropped
 *            if previous alarm was VALID, then set alarm_reset to clear
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 2 May 94
 *  REVISION: removed calls to db_post_events, done in post_events,
 *	      added call to post_events before call to update_redun
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 12 Aug 94
 *  REVISION:	added code to process the forward link flnk
 *
****************************************************************************/

static long process (paddr)
    struct dbAddr *paddr;
{
    struct pumpRecord   *ppump     = (struct pumpRecord *) (paddr -> precord);
    struct pumpred      pump_old;
    struct pump_data    pumpData;
    long                status        = 0;
    int                 alarm_reset = 0;		/* 1 = reset previous alarm */
    double              delta_time;          /* time in secs from last scan */

      /* do nothing if we did not init */

    if ((!ppump -> init) || (!ppump -> ini2))
	return (0);

   /* get time stamp for last time record was processed */

    tsLocalTime (&ppump -> lptm);

      /* save the current state of the tank for redundancy check */

    save_pump (ppump, &pump_old);

      /* get system tick counter to use for time elapse */

    ppump -> tcks = tickGet ();

      /* compute diff in secs tween last scan and this one */

    delta_time = TFTicksDiffInSeconds (&ppump -> tcks, &pump_old.tcks);

      /* init new alarm status to previous */

    ppump -> nsta = ppump -> stat;
    ppump -> nsev = ppump -> sevr;
    ppump -> msev = NO_ALARM;

      /* off normal operations */

    if (ppump -> tagd)
    {
	ppump -> nsta = TAGGED_ALARM;
	ppump -> nsev = VALID_ALARM;
    }
    else if (ppump -> oscn == OFF_SCAN)
    {
	ppump -> nsta = SCAN_ALARM;
	ppump -> nsev = VALID_ALARM;
    }
    else
    {
	if (ppump -> sevr == VALID_ALARM)
	    alarm_reset = 1;
	ppump -> nsta = NO_ALARM;
	ppump -> nsev = NO_ALARM;

      /* update number of process times */
	if (process_stats)
	    process_cnts[type]++;	

         /* if automatic or manual mode and alarm on condition
            then generate an alarm, if changed
         */

	if (ppump -> smod != ppump -> mode)
	{ 
	    if ((ppump -> mode == TF_AUTOMATIC) && ppump -> alau)
	    {
		TankFarmPostAlarm ((struct dbCommon *)ppump, AUTOMATIC,
				    ppump -> ausv);
	    }
	    else if ((ppump -> mode == TF_MANUAL) && ppump -> alma)
	    {
		TankFarmPostAlarm ((struct dbCommon *)ppump, MANUAL_OVERRIDE,
				    ppump -> masv);
	    }
	    ppump -> smod = ppump -> mode;
	}
         /* read all sensors and switches */

	if (!(status = read_sensors (ppump, &pumpData)))
	{
            /* check cmds in progress, update timers */

	    check_current_pump_status (ppump, &alarm_reset, delta_time, &pumpData);

            /* if emergency stop has been indicated */

	    if (ppump -> estp)
	    {
		emergency_stop (ppump, pumpData.pump_line_mode);
		reset_pump_commands (ppump);
	    }
	    else if ((status = check_trips (ppump, &alarm_reset)) != 0)
	    {
		emergency_stop (ppump, pumpData.pump_line_mode);
		reset_pump_commands (ppump);
	    }
	    else if (pumpData.permissive == FALSE)
	    {
		TankFarmPostAlarm ((struct dbCommon *)ppump,
				    PERMISSIVE_ALM, MAJOR_ALARM);
		emergency_stop (ppump, pumpData.pump_line_mode);
		reset_pump_commands (ppump);
	    }
	    else
	    {
                /* if the pump was in a permissive alarm last scan, now it is not
                   then cause alarm condition to reset itself
                */
                  
		if (ppump -> stat == PERMISSIVE_ALM)
		    alarm_reset = 1;

                /* set pump to active */

		ppump -> pact = TRUE;

		if (ppump -> lmen)
		{
                    /* if learn mode is enabled but there is no run
                       confirm pv installed, learning is not possible
                    */

		    if (!pumpData.run_confirm_installed)
		    {
			TankFarmPostAlarm ((struct dbCommon *)ppump,
					    CMD_NOT_ALLOWED, INFO_ALARM);
			ppump -> lmen = 0;
			ppump -> lmod = PUMP_NOT_LEARNING;
		    }
		}
		else
		{
                    /* if pump was learning and now is not, reset 1st
                       learn indicator
                    */
		    ppump -> lmod = PUMP_NOT_LEARNING;
		}

		status = run_commands (ppump, &alarm_reset, &pumpData);
	    }

	    if (ppump -> swsv == RL_LOCAL)
		status = check_local_tracking (ppump, delta_time, &pumpData);

	    /* if the pump was running and still is accumulate 
	       running time
	    */
	    if ( (pump_old.pasv == PUMP_RUNNING) && 
		 (ppump -> pasv == PUMP_RUNNING) )
	    {
		/* if we are in remote or are in local and there is a
		   run confirmed installed
		*/
		if ( (ppump -> swsv == RL_REMOTE) ||
		     ( (ppump -> swsv == RL_LOCAL) &&
		       (pumpData.run_confirm_installed) ) )
		{
		    /* accumulate in hours */
		    ppump -> acrt += delta_time / 3600;
		}
	    }

	    alarm (ppump, alarm_reset);

	} /* end read_sensors */

	/* check for maximize severity */

	if (ppump -> msev > ppump -> nsev)
	{
	    ppump -> nsev = ppump -> msev;
	    ppump -> nsta = LINK_ALARM;
	}

    } /* end else (oscn != OFF_SCAN) */

      /* check event list */

    monitor (ppump);

    ppump -> pact = FALSE;

      /* post any value changes for monitors */

    post_events (ppump, &pump_old);

      /* now check for redundancy update */

    update_redun (ppump, &pump_old);

    /* process the forward scan link record */

    if (ppump -> flnk.type == DB_LINK) 
	dbScanPassive ((struct dbAddr *)ppump -> flnk.value.db_link.pdbAddr);

    return (status);
   
} /* end func process () */



/****************************************************************************
 *
 *  MODULE NAME            : read_permissive_pv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 02 June 93
 *
 *  CALLING MODULE(S)      : read_sensors
 *          ARGUMENT(S)    : ppump - ptr to pump record
 *
 *  DESCRIPTION: read the permissive pv, set global permissive accordingly
 *               if pv not installed, set permissive to TRUE
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 2 May 94
 *  REVISION: if ini2 not set (pump not fully init'd, check if PV is init'd
 *
****************************************************************************/

static long read_permissive_pv (ppump, pumpData)
   struct pumpRecord    *ppump;
   struct pump_data	*pumpData;
{
    long status    = 0;
    long options   = 0;
    long nRequests = 1;

    pumpData -> permissive = TRUE;

    if (ppump -> perp.type != CONSTANT)
    {
	if (ppump -> perp.type == DB_LINK)
        {
            status = dbGetLink (&ppump -> perp.value.db_link,
                                (struct dbCommon *)ppump, DBR_ENUM,
                                (caddr_t)&pumpData -> permissive, &options, 
				&nRequests);
	    if (status)
            {
                TankFarmPostAlarm ((struct dbCommon *)ppump, LINK_ALARM,
                                                        VALID_ALARM);
            }
	    else if (!ppump -> ini2)
	    {
	        if (!okToUse (&ppump -> perp) )
	    	    status = -1;
            }
	}
    }

    return (status);

} /* end func read_permissive_pv () */



/****************************************************************************
 *
 *  MODULE NAME            : read_pressure_pv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 02 June 93
 *
 *  CALLING MODULE(S)      : read_sensors
 *          ARGUMENT(S)    : ppump - ptr to pump record
 *
 *  DESCRIPTION: read the pressure pv and set global pressure_tripped
 *                   accordingly
 *               if sensor is tripped, added to trip alarm stack, else
 *                   delete it from the stack
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 2 May 94
 *  REVISION: if ini2 not set (pump not fully init'd, check if PV is init'd
 *
****************************************************************************/

static long read_pressure_pv (ppump, pumpData)
   struct pumpRecord    *ppump;
   struct pump_data	*pumpData;
{
    long status    = 0;
    long options   = 0;
    long nRequests = 1;

    pumpData -> pressure_tripped = 0;

    if (ppump -> prsp.type != CONSTANT)
    {
        if (ppump -> prsp.type == DB_LINK)
        {
	    status = dbGetLink (&ppump -> prsp.value.db_link,
                             (struct dbCommon *)ppump, DBR_ENUM,
                             (caddr_t)&pumpData -> pressure_tripped,
                             &options, &nRequests);
            if (status)
            {
            	TankFarmPostAlarm ((struct dbCommon *)ppump, LINK_ALARM,
                                                              VALID_ALARM);
		return (status);
            }
	    else if (!ppump -> ini2)
	    {
		if (!okToUse (&ppump -> prsp) )
		    return (-1);
	    }
        }
    }

    if (pumpData -> pressure_tripped)
        trip_alarms (ppump, PRESSURE, ADD_TRIP_ALARM);
    else
        trip_alarms (ppump, PRESSURE, DELETE_TRIP_ALARM);

    return (status);

} /* end func read_pressure_pv () */



/****************************************************************************
 *
 *  MODULE NAME            : read_rl_switch 
 *  AUTHOR                 : emsteel
 *  DATE                   : 11 May 93
 *
 *  CALLING MODULE(S)      : read_sensors
 *          ARGUMENT(S)    : ppump - ptr to a pump record
 *
 *
 *  DESCRIPTION: set RL switch state (swst, swsv)
 *               if the RL switch is not installed, then set RL status
 *                  to RL_REMOTE
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 16 dec 93
 *  REVISION:  added remote_local_trans code
 *             turn off timers if transition from remote to local
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 2 May 94
 *  REVISION: if ini2 not set (pump not fully init'd, check if PV is init'd
 *	      remove db_post_events of swst and swsv, done in post_events()
 *
****************************************************************************/

static long read_rl_switch (ppump, pumpData)
   struct pumpRecord	*ppump;
   struct pump_data	*pumpData;
{
    long     nRequests = 1;
    long     options   = 0;
    long     status    = 0;
    unsigned short curr_rl;          /* returned by the RL switch */

    pumpData -> remote_local_trans = 0;

    if (ppump -> rolp.type != CONSTANT)
    {
        if (ppump -> rolp.type == DB_LINK)
	{
            /* get value from sensor */
	    status = dbGetLink (&ppump -> rolp.value.db_link, 
                                (struct dbCommon *)ppump, DBR_ENUM, 
                                (caddr_t)&curr_rl, &options, &nRequests);
            if (status)
            {
            	TankFarmPostAlarm ((struct dbCommon *)ppump, LINK_ALARM,
                                                               VALID_ALARM);
		return (status);
            }
	    else if (!ppump -> ini2)
	    {
		if (!okToUse (&ppump -> rolp) )
		    return (-1);
	    }
            if (curr_rl == 1)
            {
                strncpy (ppump -> swst, ppump -> swre,
                                                    sizeof (ppump -> swst));
                ppump -> swsv = RL_REMOTE;
            }
            else
            {
              /* are we transitioning from remote to local */

                if (ppump -> swsv == RL_REMOTE)
                    pumpData -> remote_local_trans = 1;

                strncpy (ppump -> swst, ppump -> swlo,
                                                    sizeof (ppump -> swst));
                ppump -> swsv = RL_LOCAL;
            }
	}
    }
    else
    {
        strncpy (ppump -> swst, ppump -> swre, sizeof (ppump -> swst));
        ppump -> swsv = RL_REMOTE;
    }

    /* turn off timer if going from remote to local, but not if we are
       already in PUMP_MIN_OFF, because we will continue to time
       for this state even in LOCAL
    */

    if ( (pumpData -> remote_local_trans) && (ppump -> pasv != PUMP_MIN_OFF) )
       ppump -> ttim = 0;

    return (status);
   
} /* end func read_rl_switch () */



/****************************************************************************
 *
 *  MODULE NAME            : read_run_confirm_pv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 02 June 93
 *
 *  CALLING MODULE(S)      : read_sensors
 *          ARGUMENT(S)    : ppump - ptr to pump record
 *
 *  DESCRIPTION: read the run confirm pv, set globals run_confirmed
 *               and run_confirm_installed accordingly
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 2 May 94
 *  REVISION: if ini2 not set (pump not fully init'd, check if PV is init'd
 *
****************************************************************************/

static long read_run_confirm_pv (ppump, pumpData)
   struct pumpRecord    *ppump;
   struct pump_data	*pumpData;
{
    long status    = 0;
    long options   = 0;
    long nRequests = 1;

    pumpData -> run_confirmed = 0;
    pumpData -> run_confirm_installed = 0;

    if (ppump -> rucp.type != CONSTANT)
    {
        if (ppump -> rucp.type == DB_LINK)
        {
            status = dbGetLink (&ppump -> rucp.value.db_link,
                                (struct dbCommon *)ppump, DBR_ENUM,
                                (caddr_t)&pumpData -> run_confirmed, &options,
                                &nRequests);
            if (status)
            {
                TankFarmPostAlarm ((struct dbCommon *)ppump, LINK_ALARM,
                                                              VALID_ALARM);
		return (status);
            }
	    else if (!ppump -> ini2)
	    {
		if (!okToUse (&ppump -> rucp) )
		    return (-1);
	    }
            pumpData -> run_confirm_installed = 1;
        }
    }

    return (status);

} /* end func read_run_confirm_pv () */



/****************************************************************************
 *
 *  MODULE NAME            : read_sensors 
 *  AUTHOR                 : emsteel
 *  DATE                   : 02 June 93
 *
 *  CALLING MODULE(S)      : init_record, process
 *          ARGUMENT(S)    : ppump - ptr to pump record
 *
 *  DESCRIPTION: read various switches and sensors
 *
****************************************************************************/

static long read_sensors (ppump, pumpData)
   struct pumpRecord    *ppump;
   struct pump_data	*pumpData;
{
    long status = 0;

    if ((status = check_line_mode (ppump, pumpData)) != 0)
        return (status);

    if ((status = read_rl_switch (ppump, pumpData)) != 0)
        return (status);

    if ((status = read_permissive_pv (ppump, pumpData)) != 0)
        return (status);

    if ((status = read_temperature_pv (ppump, pumpData)) != 0)
        return (status);

    if ((status = read_vibration_pv (ppump, pumpData)) != 0)
        return (status);

    if ((status = read_pressure_pv (ppump, pumpData)) != 0)
        return (status);

    if ((status = read_run_confirm_pv (ppump, pumpData)) != 0)
        return (status);

    return (status);

} /* end func read_sensors () */



/****************************************************************************
 *
 *  MODULE NAME            : read_temperature_pv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 02 June 93
 *
 *  CALLING MODULE(S)      : read_sensors
 *          ARGUMENT(S)    : ppump - ptr to pump record
 *
 *  DESCRIPTION: read the temperature pv if it exists, set 
 *               temperature_tripped accordingly
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 2 May 94
 *  REVISION: if ini2 not set (pump not fully init'd, check if PV is init'd
 *
****************************************************************************/

static long read_temperature_pv (ppump, pumpData)
   struct pumpRecord    *ppump;
   struct pump_data	*pumpData;
{
    long status    = 0;
    long options   = 0;
    long nRequests = 1;

    pumpData -> temperature_tripped = 0;

    if (ppump -> tmpp.type != CONSTANT)
    {
        if (ppump -> tmpp.type == DB_LINK)
        {
            status = dbGetLink (&ppump -> tmpp.value.db_link,
                                (struct dbCommon *)ppump, DBR_ENUM,
                                (caddr_t)&pumpData -> temperature_tripped, &options,
                                &nRequests);
            if (status)
            {
                TankFarmPostAlarm ((struct dbCommon *)ppump, LINK_ALARM,
                                                              VALID_ALARM);
		return (status);
            }
	    else if (!ppump -> ini2)
	    {
		if (!okToUse (&ppump -> tmpp) )
		    return (-1);
	    }
        }
    }
   
    if (pumpData -> temperature_tripped)
        trip_alarms (ppump, TEMPERATURE, ADD_TRIP_ALARM);
    else
        trip_alarms (ppump, TEMPERATURE, DELETE_TRIP_ALARM);

    return (status);

} /* end func read_temperature_pv () */



/****************************************************************************
 *
 *  MODULE NAME            : read_vibration_pv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 02 June 93
 *
 *  CALLING MODULE(S)      : read_sensors
 *          ARGUMENT(S)    : ppump - ptr to pump record
 *
 *  DESCRIPTION: read the vibration pv and set vibration_tripped
 *               accordingly, if pv not installed set to 0;
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 2 May 94
 *  REVISION: if ini2 not set (pump not fully init'd, check if PV is init'd
 *
****************************************************************************/

static long read_vibration_pv (ppump, pumpData)
   struct pumpRecord    *ppump;
   struct pump_data	*pumpData;
{
    long status    = 0;
    long options   = 0;
    long nRequests = 1;

    pumpData -> vibration_tripped = 0;

    if (ppump -> vibp.type != CONSTANT)
    {
        if (ppump -> vibp.type == DB_LINK)
        {
            status = dbGetLink (&ppump -> vibp.value.db_link,
                                (struct dbCommon *)ppump, DBR_ENUM,
                                (caddr_t)&pumpData -> vibration_tripped, &options,
                                &nRequests);
            if (status)
            {
                TankFarmPostAlarm ((struct dbCommon *)ppump, LINK_ALARM,
                                                              VALID_ALARM);
		return (status);
            }
	    else if (!ppump -> ini2)
	    {
		if (!okToUse (&ppump -> vibp) )
		    return (-1);
	    }
	}
    }

    if (pumpData -> vibration_tripped)
        trip_alarms (ppump, VIBRATION, ADD_TRIP_ALARM);
    else
        trip_alarms (ppump, VIBRATION, DELETE_TRIP_ALARM);
  
    return (status);

} /* end func read_vibration_pv () */



/****************************************************************************
 *
 *  MODULE NAME            : reset_pump_commands 
 *  AUTHOR                 : emsteel
 *  DATE                   : 30 July 93
 *
 *          ARGUMENT(S)    : ppump - ptr to pump record
 *
 *  DESCRIPTION: reset pump commands and post changes
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 dec 93
 *  REVISION:  deleted db_post_events for command fields, not needed
 *             added FirstSpecialDigitalCmd
 *
****************************************************************************/

static void reset_pump_commands (ppump)
   struct pumpRecord    *ppump;
{
   struct tfcmd   tfcmd;

   tfcmd.NumCmds                = 2;
   tfcmd.FirstDigitalCmd        = &ppump -> cstr;
   tfcmd.FirstSpecialDigitalCmd = &ppump -> strs;
   tfcmd.FirstCmdString         = ppump -> sstr;
   tfcmd.CmdString              = ppump -> cmds;

   ResetTFCommand (&tfcmd);

   return;

} /* end func reset_pump_commands () */



/****************************************************************************
 *
 *  MODULE NAME            : run_commands 
 *  AUTHOR                 : emsteel
 *  DATE                   : 02 June 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : ppump - ptr to pump record
 *
 *  DESCRIPTION: process a command if any
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 dec 93
 *  REVISION:  added code to support special digital command and update of
 *             PCMD (last command)
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 16 dec 93
 *  REVISION:  if local  always update status
 *             if local tracking enabled, only send signal to pump if
 *                single bit mode and
 *                status change and/or
 *                transition from remote to local and/or
 *                transition from no local tracking to tracking
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 Feb 94
 *  REVISION:  if local and DUAL_BIT_MODE, drop output bits when transition
 *             to local regardless whether local tracking is enabled
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 2 May 94
 *  REVISION:  remove calls to db_post_events, done in post_events()
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 07 Feb 97
 *  REVISION:	removed local tracking code to check_local_tracking ()
 *
****************************************************************************/

static long run_commands (ppump, alarm_reset, pumpData)
    struct pumpRecord    *ppump;
    int			*alarm_reset;
    struct pump_data	*pumpData;
{
    struct tfcmd	tfcmd;
    int			cmd;
    int			cmd_type;
    unsigned short	start_val       = 0;
    unsigned short	stop_val        = 0;
    long		status          = 0;
    int			talktopump      = 0;

    tfcmd.NumCmds                = 2;
    tfcmd.FirstDigitalCmd        = &ppump -> cstr;
    tfcmd.FirstSpecialDigitalCmd = &ppump -> strs;
    tfcmd.FirstCmdString         = ppump -> sstr;
    tfcmd.CmdString              = ppump -> cmds;

    cmd = GetTFCommand (&tfcmd, &cmd_type);

    if (ppump -> swsv == RL_REMOTE)
    {
	if (cmd != NO_TF_COMMAND)
	{
	    if ((ppump -> mode == TF_AUTOMATIC) || (ppump -> mode == TF_MANUAL))
	    {
		if (cmd == ILLEGAL_TF_STRING_CMD)
		    TankFarmPostAlarm ((struct dbCommon *)ppump, 
				       CMD_NOT_ALLOWED, INFO_ALARM);
		else if ( ( (ppump -> mode == TF_AUTOMATIC) &&
			    ( (cmd_type == TF_STRING_CMD) ||
			      (cmd_type == TF_SPECIAL_DIGITAL_CMD) ) ) ||
			  ( (ppump -> mode == TF_MANUAL) &&
			    (cmd_type == TF_DIGITAL_CMD) ) )
		{
		    switch (cmd)
		    {
			case PUMP_START_CMD:
			    /* can not start the pump if timimg for a 
			       minimum off time
			    */
			    if (ppump -> pasv == PUMP_MIN_OFF)
			    {
				TankFarmPostAlarm ((struct dbCommon *)ppump,
						   CMD_NOT_ALLOWED, INFO_ALARM);
				break;
			    }
			    ppump -> ttim = 0;
			    strncpy (ppump -> pust, ppump -> stst,
				     sizeof (ppump -> pust));
			    ppump -> pasv = PUMP_STARTING;
			    ppump -> lcmd = PUMP_STARTING;
			    *alarm_reset = 1;
			    ppump -> lsta = ppump -> lsev = NO_ALARM;
			    start_val = 1;
			    stop_val = 0;
			    talktopump++;
  
			    /* update last command */

			    if (ppump -> sstr[0])
				strncpy (ppump -> pcmd, ppump -> sstr, 
					 sizeof (ppump -> pcmd));
			    else
				strncpy (ppump -> pcmd, TF_UNDEF_CMD, 
					 sizeof (ppump -> pcmd));
			    break;
   
			case PUMP_STOP_CMD:
			    if ( (ppump -> motm <= 0) ||
				 ( (ppump -> pasv != PUMP_MIN_OFF) &&
				   (ppump -> pasv != PUMP_STOPPED) ) )
			    {
				ppump -> ttim = 0;
				if ( (ppump -> motm > 0) &&
				     (ppump -> pasv == PUMP_STARTING) )
				{
				    strncpy (ppump -> pust, ppump -> stof,
					     sizeof (ppump -> pust));
				    ppump -> pasv = PUMP_STOPPED;
				    ppump -> lcmd = PUMP_STOPPED;
				}
				else
				{
				    strncpy (ppump -> pust, ppump -> stsp,
					     sizeof (ppump -> pust));
				    ppump -> pasv = PUMP_STOPPING;
				    ppump -> lcmd = PUMP_STOPPING;
				}
			    }
			    *alarm_reset = 1;
			    ppump -> lsta = ppump -> lsev = NO_ALARM;
			    start_val = 0;
			    stop_val = 1;
			    talktopump++;
  
			    if (ppump -> sstp[0])
				strncpy (ppump -> pcmd, ppump -> sstp, 
					 sizeof (ppump -> pcmd));
			    else
				strncpy (ppump -> pcmd, TF_UNDEF_CMD, 
					 sizeof (ppump -> pcmd));
			    break;
   
			default:
			    TankFarmPostAlarm ((struct dbCommon *)ppump,
					       CMD_NOT_ALLOWED, INFO_ALARM);
		    }
		}
		else
		{
		    TankFarmPostAlarm ((struct dbCommon *)ppump, CMD_NOT_ALLOWED, 
				       INFO_ALARM);
		}
	    }
	    else if (ppump -> mode == TF_LOCKED)
	    {
		TankFarmPostAlarm ((struct dbCommon *)ppump, CMD_NOT_ALLOWED,
				   INFO_ALARM);
	    }
	} /* end cmd != NO_TF_CMD */

    } /* end if RL_REMOTE */

    else   /* are local */
    {
	if (cmd != NO_TF_COMMAND)
	    TankFarmPostAlarm ((struct dbCommon *)ppump, CMD_NOT_ALLOWED,
			       INFO_ALARM);
    } 

    if (talktopump)
	status = talk_to_pump (ppump, &start_val, &stop_val, pumpData -> pump_line_mode);

     reset_pump_commands (ppump);

     return (status);

} /* end func run_commands () */



/****************************************************************************
 *
 *  MODULE NAME            : save_pump 
 *  AUTHOR                 : emsteel
 *  DATE                   : 11 May 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : ppump - ptr to pump record
 *                           ppump_old - ptr to pump redunadancy struct
 *
 *  DESCRIPTION: save values from the pump record for reduandancy check
 *               at end of process.
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 2 May 94
 *  REVISION: deleted val and time (not neeeded) added pcmd
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 5 May 94
 *  REVISION: added ini2
 *
****************************************************************************/

static void save_pump (ppump, ppump_old)
   struct pumpRecord      *ppump;
   struct pumpred         *ppump_old;
{
   ppump_old -> stat = ppump -> stat;
   ppump_old -> sevr = ppump -> sevr;
   ppump_old -> val = ppump -> val;
   strncpy (ppump_old -> pust, ppump -> pust, sizeof (ppump_old -> pust));
   strncpy (ppump_old -> swst, ppump -> swst, sizeof (ppump_old -> swst));
   strncpy (ppump_old -> pcmd, ppump -> pcmd, sizeof (ppump_old -> pcmd));
   ppump_old -> pasv = ppump -> pasv;
   ppump_old -> swsv = ppump -> swsv;
   ppump_old -> lmod = ppump -> lmod;
   ppump_old -> estp = ppump -> estp;
   ppump_old -> exon = ppump -> exon;
   ppump_old -> exof = ppump -> exof;
   ppump_old -> ttim = ppump -> ttim;
   ppump_old -> sped = ppump -> sped;
   ppump_old -> tcks = ppump -> tcks;
   ppump_old -> smod = ppump -> smod;
   ppump_old -> lmen = ppump -> lmen;
   ppump_old -> lcmd = ppump -> lcmd;
   ppump_old -> lsta = ppump -> lsta;
   ppump_old -> lsev = ppump -> lsev;
   ppump_old -> pltr = ppump -> pltr;
   ppump_old -> ini2 = ppump -> ini2;
   strncpy (ppump_old -> cmds, ppump -> cmds, sizeof (ppump_old -> cmds));
   ppump_old -> cstr = ppump -> cstr;
   ppump_old -> cstp = ppump -> cstp;
   ppump_old -> strs = ppump -> strs;
   ppump_old -> stps = ppump -> stps;
   ppump_old -> latm = ppump -> latm;
   ppump_old -> lsvl = ppump -> lsvl;

   return;
   
} /* end func save_pump () */



/****************************************************************************
 *
 *  MODULE NAME            : special
 *  AUTHOR                 : emsteel
 *  DATE                   : 09 Feb 96
 *
 *  DESCRIPTION: run code on fields when updated by MASTER
 *
****************************************************************************/

static long special (paddr, after)
    struct	dbAddr	*paddr;
    int			after;
{
    struct pumpRecord  	*ppump = (struct pumpRecord *)(paddr->precord);
    int           	special_type = paddr->special;
    unsigned short	master;

    if (!after)
    {
       return (0);
    }

    switch (special_type) 
    {
	case (SPC_PUMP_TICKS):		/* on fields TTIM & ACRT */
	    if (resGetCaAnswer (ppump -> radr, &master) == RES_OK)
	    {
		if (master)
		    break;
		else
		{
		    ppump -> tcks = tickGet ();
		}
	    }
	    break;

	default:
	    recGblDbaddrError (S_db_badChoice, paddr, "pump: special");
	    return (S_db_badChoice);
    }
    return (0);
}


/****************************************************************************
 *
 *  MODULE NAME            : talk_to_pump 
 *  AUTHOR                 : emsteel
 *  DATE                   : 06 May 93
 *
 *  CALLING MODULE(S)      : various
 *          ARGUMENT(S)    : ppump - ptr to pump record
 *                           start_val - value to send to start pv
 *                           stop_val - value to send to stop pv
 *
 *  DESCRIPTION: set start and stop pump pvs, output the 0 value first
 *		 to ensure that we do not have both outputs up even if it is
 *		 for a brief instance
 *
****************************************************************************/

static long talk_to_pump (ppump, start_val, stop_val, pump_line_mode)
    struct pumpRecord	*ppump;
    unsigned short	*start_val;
    unsigned short      *stop_val;
    short		pump_line_mode;
{
    long   status    = 0;
    long   nRequests = 1;
    double sped;

    /* save the last stop value issued */
    ppump -> lsvl = *stop_val;

    if ( (pump_line_mode == SINGLE_BIT_MODE) || (ppump -> sped != 0) )
    {
	if (ppump -> sped == 0)
	{
	    status = dbPutLink (&ppump -> ponp.value.db_link,
				(struct dbCommon *)ppump, DBR_ENUM,
				(caddr_t)start_val, nRequests);
	    if (status)
	    {
		TankFarmPostAlarm ((struct dbCommon *)ppump,
				   LINK_ALARM, VALID_ALARM);
		return (status);
	    }
	}
	else
	{
	    if ((*start_val) == 1)
		sped = ppump -> sped;
	    else
		sped = (double)(*start_val);
	    status = dbPutLink (&ppump -> ponp.value.db_link,
				(struct dbCommon *)ppump, DBR_DOUBLE,
				(caddr_t)&sped, nRequests);
	    if (status)
	    {
		TankFarmPostAlarm ((struct dbCommon *)ppump, LINK_ALARM, 
				   VALID_ALARM);
		return (status);
	    }
	}
	/* in case have speed control and a stop command */
	if (pump_line_mode == DUAL_BIT_MODE)
	{
	    status = dbPutLink (&ppump -> pofp.value.db_link,
				(struct dbCommon *)ppump, DBR_ENUM,
				(caddr_t)stop_val, nRequests);
	    if (status)
	    {
		TankFarmPostAlarm ((struct dbCommon *)ppump, LINK_ALARM, 
				   VALID_ALARM);
		return (status);
	    }
	}
    }
       /* if dual bit mode, set stop pump state */

    else /* dual bit mode */
    {
	if ((*start_val) > 0)
	{
	    status = dbPutLink (&ppump -> pofp.value.db_link,
				(struct dbCommon *)ppump, DBR_ENUM,
				(caddr_t)stop_val, nRequests);
	    if (status)
	    {
		TankFarmPostAlarm ((struct dbCommon *)ppump, LINK_ALARM, 
				   VALID_ALARM);
		return (status);
	    }
	    status = dbPutLink (&ppump -> ponp.value.db_link,
				(struct dbCommon *)ppump, DBR_ENUM,
				(caddr_t)start_val, nRequests);
	    if (status)
	    {
		TankFarmPostAlarm ((struct dbCommon *)ppump, LINK_ALARM, 
				   VALID_ALARM);
		return (status);
	    }
	}
	else /* stop_val >= 0 */
	{
	    status = dbPutLink (&ppump -> ponp.value.db_link,
				(struct dbCommon *)ppump, DBR_ENUM,
				(caddr_t)start_val, nRequests);
	    if (status)
	    {
		TankFarmPostAlarm ((struct dbCommon *)ppump, LINK_ALARM, 
				   VALID_ALARM);
		return (status);
	    }
	    status = dbPutLink (&ppump -> pofp.value.db_link,
				(struct dbCommon *)ppump, DBR_ENUM,
				(caddr_t)stop_val, nRequests);
	    if (status)
	    {
		TankFarmPostAlarm ((struct dbCommon *)ppump, LINK_ALARM, 
				   VALID_ALARM);
		return (status);
	    }
	}
    }

    return (status);
   
} /* end func talk_to_pump () */



/****************************************************************************
 *
 *  MODULE NAME            : trip_alarms 
 *  AUTHOR                 : emsteel
 *  DATE                   : 02 Nov 93
 *
 *  CALLING MODULE(S)      : read_pressure_pv, read_temperature_pv
 *                           read_vibration_pv
 *          ARGUMENT(S)    : ppump - ptr to pump record
 *                           type  - type of sensor (PRESSURE etc)
 *                           add   - 1 = add to stack, 0 = delete
 *
 *  DESCRIPTION: either added the trip alarm to the stack or delete it
 *               add new at head of stack
 *
****************************************************************************/

static void trip_alarms (ppump, type, add)
   struct pumpRecord    *ppump;
   int                  type;
   int                  add;
{
   int i;
   int found;

   for (i = 0, found = -1; i < 3; i++)
   {
      if (ppump -> tstk[i] == type)
      {
         found = i;
         break;
      }
   }

   if (add)
   {
      if (found < 0)
      {
         bcopy ((char *)(&ppump -> tstk[0]), (char *)(&ppump -> tstk[1]), 4);
         ppump -> tstk[0] = type;
      }
   }
   else
   {
      if (found > -1)
      {
         if (found != 2)
         {
            bcopy ( (char *) (&ppump -> tstk[found + 1]),
                    (char *) (&ppump -> tstk[found]),
                    4 - (found * 2) );
         }
         ppump -> tstk [2] = NO_ALARM;
      }
   }

   return;

} /* end func trip_alarms () */



/****************************************************************************
 *
 *  MODULE NAME            : update_redun 
 *  AUTHOR                 : emsteel
 *  DATE                   : 02 June 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : ppump - ptr to pump record
 *                           ppump_old - ptr to pump before process
 *
 *  DESCRIPTION: have any of the fields deemed as being redundant, changed
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 15 Feb 94
 *  REVISION: handle, LMEN, EXON, EXOF whether we are redundant or not,
 *            these are prompted fields and we must let dbClient know if thay
 *            change, dbserver will know if we are redunant
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 Feb 94
 *  REVISION:  deleted val (not needed) added pcmd
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 5 May 94
 *  REVISION: added ini2
 *
****************************************************************************/

static void update_redun (ppump, ppump_old)
    struct pumpRecord    *ppump;
    struct pumpred       *ppump_old;
{
    char		redbuf[REDQ_MAX_MSGLEN];
    int			svrt_int, acrt_int;
    int			update_svrt = 0;
    int			update_acrt = 0;
    float		temp;
    unsigned short	monitor_mask;

    /* if 1 hour has elapsed since we saved the accumulated
       pump running time, it is time then to save it up
       to pump.dbf
    */
    svrt_int = ppump -> svrt;
    acrt_int = ppump -> acrt;
    if (svrt_int != acrt_int)
    {
	ppump -> svrt = ppump -> acrt;
	update_svrt = 1;
    }

    if ( (ppump -> lmen != ppump_old -> lmen) ||
	 (ppump -> exon != ppump_old -> exon) ||
	 (ppump -> exof != ppump_old -> exof) ||
	 (update_svrt) )
    {
         /* recalc check sum */

      ppump -> rcrc = DBC_CRCRecord ((char *)ppump, type);
    }

    redQInit ((struct dbCommon *)ppump, redbuf, sizeof (redbuf), type);

    if (ppump -> stat != ppump_old -> stat)
	redQField (redbuf, mylist[STAT]);

    if (ppump -> sevr != ppump_old -> sevr)
	redQField (redbuf, mylist[SEVR]);

    if (strncmp (ppump -> pust, ppump_old -> pust,
		 sizeof (ppump -> pust)) != 0)
	redQField (redbuf, mylist[PUST]);

    if (strncmp (ppump -> swst, ppump_old -> swst,
		 sizeof (ppump -> pust)) != 0)
	redQField (redbuf, mylist[SWST]);

    if (ppump -> pasv != ppump_old -> pasv)
	redQField (redbuf, mylist[PASV]);

    if (ppump -> swsv != ppump_old -> swsv)
	redQField (redbuf, mylist[SWSV]);

    if (ppump -> lmod != ppump_old -> lmod)
	redQField (redbuf, mylist[LMOD]);

    if (ppump -> estp != ppump_old -> estp)
	redQField (redbuf, mylist[ESTP]);

    if (ppump -> exon != ppump_old -> exon)
	redQField (redbuf, mylist[EXON]);

    if (ppump -> exof != ppump_old -> exof)
	redQField (redbuf, mylist[EXOF]);

    if (ppump -> ttim != ppump_old -> ttim)
	redQField (redbuf, mylist[TTIM]);

    if (ppump -> sped != ppump_old -> sped)
	redQField (redbuf, mylist[SPED]);

    if (ppump -> smod != ppump_old -> smod)
	redQField (redbuf, mylist[SMOD]);

    if (ppump -> lmen != ppump_old -> lmen)
	redQField (redbuf, mylist[LMEN]);

    if (ppump -> lcmd != ppump_old -> lcmd)
	redQField (redbuf, mylist[LCMD]);

    if (ppump -> lsta != ppump_old -> lsta)
	redQField (redbuf, mylist[LSTA]);

    if (ppump -> lsev != ppump_old -> lsev)
	redQField (redbuf, mylist[LSEV]);

    if (ppump -> pltr != ppump_old -> pltr)
	redQField (redbuf, mylist[PLTR]);

    if (strncmp (ppump -> pcmd, ppump_old -> pcmd,
		 sizeof (ppump -> pcmd)) != 0)
	redQField (redbuf, mylist[PCMD]);

    if (ppump -> ini2 != ppump_old -> ini2)
	redQField (redbuf, mylist[INI2]);

    if (strncmp (ppump -> cmds, ppump_old -> cmds,
                                              sizeof (ppump -> cmds)) != 0)
	redQField (redbuf, mylist[CMDS]);

    if (ppump -> cstr != ppump_old -> cstr)
	redQField (redbuf, mylist[CSTR]);

    if (ppump -> cstp != ppump_old -> cstp)
	redQField (redbuf, mylist[CSTP]);

    if (ppump -> strs != ppump_old -> strs)
	redQField (redbuf, mylist[STRS]);

    if (ppump -> stps != ppump_old -> stps)
	redQField (redbuf, mylist[STPS]);

    if (TsCmpStampsNE (&ppump -> latm, &ppump_old-> latm) )
	redQField (redbuf, mylist [LATM]);

    if (ppump -> val != ppump_old -> val)
	redQField (redbuf, mylist[VAL]);
	
    /* is it time to redundantly update the accumulated
       pump running time, compare abs value to last
       post ACRT, update is based on the precision
       the user has selected for ACRT
    */
    temp = ppump -> acrt - ppump -> lprt;
    if (temp < 0.0)
	temp = -temp;
    /* 
    ** .001 hrs */
    if (ppump -> prrt >= 3)
    {
	if (temp >= 0.001)
	    update_acrt = 1;
    }
    /* .01 hrs */
    else if (ppump -> prrt == 2)
    {
	if (temp >= 0.01)
	    update_acrt = 1;
    }
    /* .1 hrs (6 mins) */
    else if (ppump -> prrt == 1)
    {
	if (temp >= 0.1)
	    update_acrt = 1;
    }
    /* 1 hr */
    else
    {
	if (temp >= 1.0)
	    update_acrt = 1;
    }
    if (update_acrt)
    {
	ppump -> lprt = ppump -> acrt;
	redQField (redbuf, mylist [ACRT]);
	redQField (redbuf, mylist [LPRT]);
	monitor_mask = DBE_VALUE;
	if (ppump -> tenb)
	    monitor_mask |= DBE_LOG;
	db_post_events ((struct dbCommon *)ppump, (caddr_t)&ppump -> acrt, monitor_mask);
    }

    if (update_svrt)
	redQField (redbuf, mylist[SVRT]);

    if (ppump -> lsvl != ppump_old -> lsvl)
	redQField (redbuf, mylist[LSVL]);

    redQSend (redbuf);

} /* end func update_redun () */

