/*recState.c,v 1.1.1.1 1999/08/03 14:44:22 tis4000 Exp */

/* recState.c - Record Support Routines for State records */
/*
 *      Original Author: Bob Dalesio
 *      Current Author:  Marty Kraimer
 *      Date:            10-10-90 
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01  10-10-90	mrk	extensible record and device support
 * .02  01-04-94	ems	added init_record to set INIT to 1 and
 *				ACK = 0;
 *				in process, if not init'd, do nothing
 * .03  03-01-94	ems	added support for process stats
 *                              removed ack = 0
 * .04	03-29-94	ems	make forward func def agree with actual func def
 * .05	05-13-94	ems	deleted monitor, not needed since no alarms are
 *				set, oval is now updated in process()
 * .06	10-20-94	ems	added lptm (last process time)
 * .07  01-13-95        wep     init lptm and latm
 */

#include	<vxWorks.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<lstLib.h>
#include	<string.h>

#include	"dbDefs.h"
#include	"dbCommon.h"
#include	"link.h"
#include	"db_field_log.h"
#include	"dbAccess.h"
#include	"dbFldTypes.h"
#include	"dbRecType.h"
#include	"devSup.h"
#include	"errMdef.h"
#include	"recSup.h"
#include	"rec/stateRecord.h"
#include	"db_addr.h"
#include	"dbEvent.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table*/
#define report NULL
#define initialize NULL
static long init_record ();
static long process();
#define special NULL
static long get_value();
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
#define get_units NULL
#define get_precision NULL
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
#define get_graphic_double NULL
#define get_control_double NULL
#define get_alarm_double NULL

struct rset stateRSET={
	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double };

static unsigned short	type;


static long init_record (pstate)
    struct stateRecord    *pstate;
{
    char *pstr;

    /* find record type */
    for (type = 1; type < 100; type++)
    {
	pstr = GET_PRECTYPE (type);
	if (strncmp (pstr, "state", 5) == 0)
	    break;
    }

    pstate -> init = 1;
    pstate -> ini2 = 1;

    /* init last process time and last alarm time */
    if (pstate -> lptm.secPastEpoch == 0)
        tsLocalTime (&pstate -> lptm);
    if (pstate -> latm.secPastEpoch == 0)
        tsLocalTime (&pstate -> latm);

    return (0);
}


static long process(paddr)
    struct dbAddr	*paddr;
{
	struct stateRecord	*pstate=(struct stateRecord *)(paddr->precord);

	   /* if not init'd, do nothing */

	if ((!pstate -> init) || (!pstate -> ini2))
	    return (0);

	/* get time stamp for last time record was processed */

	tsLocalTime (&pstate -> lptm);

	/* update number of process times */
	if (process_stats)
	   process_cnts[type]++;	

	pstate->udf = FALSE;
        pstate->pact=TRUE;

	if(strncmp(pstate->oval,pstate->val,sizeof(pstate->val))) 
	{
	    db_post_events ((struct dbCommon *)pstate, &(pstate->val[0]),DBE_VALUE);
	    strncpy (pstate->oval,pstate->val,sizeof(pstate->oval));
	}

        /* process the forward scan link record */
        if (pstate->flnk.type==DB_LINK) 
	    dbScanPassive ((struct dbAddr *)pstate->flnk.value.db_link.pdbAddr);

        pstate->pact=FALSE;
	return(0);
}

static long get_value(pstate,pvdes)
    struct stateRecord             *pstate;
    struct valueDes     *pvdes;
{
    pvdes->field_type = DBF_STRING;
    pvdes->no_elements=1;
    pvdes->pvalue = (caddr_t)(&pstate->val[0]);
    return(0);
}

