/*recValve.c,v 1.1.1.1 1999/08/03 14:44:23 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************************
 *
 *  MODULE NAME            : recValve.c
 *  AUTHOR                 : emsteel
 *  DATE                   : 04 May 93
 *
 *
 *  DESCRIPTION: record support routines for a Valve record
 *
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 02 Mar 94
 *  REVISION: added process stats code
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 29 Mar 94
 *  REVISION:  make forward func def agree with actual func def
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 05 May 94
 *  REVISION:  rewrite init_record, consolidate all db_post_events in
 *		post_events()
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 June 94
 *  REVISION:	added support for maximize severity (msev), move call to
 *		tsLocalTime inside on scan
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 01 Sep 94
 *  REVISION:	added get_control_double() and get_graphic_double()
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 13 Oct 94
 *  REVISION:	added TIME to namelist, monitor(), do tsLocalTime if status
 *		chgs, do not do in process()
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 20 Oct 94
 *  REVISION:	chg time to latm (last alarm time)
 *		added lptm (last process time)
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 6 Feb 95
 *  REVISION:	make VAL = VASV so valve status can be trended
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 Feb 95
 *  REVISION:	made VAL redundant
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 26 Mar 96
 *  REVISION:	made static globals part of struct valve_vars to make code
 *		reentrant
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 20 Jun 96
 *  REVISION:	added code for alarm disable ADIS
 *
****************************************************************************/

/****************************************************************************
 *
 * Programmer's Notes:
 *
 * VALVE_OPEN, VALVE_CLOSED etc see tankfarm.h for these defines
 *
****************************************************************************/



#include <vxWorks.h>
#include	<inetLib.h>
#include <tickLib.h>       /* for tickGet () */
#include <types.h>
#include <stdioLib.h>
#include <lstLib.h>
#include <logLib.h>
#include <string.h>

#include "alarm.h"
#include "choice.h"
#include "cvtTable.h"
#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#include "dbAccess.h"
#include "dbFldTypes.h"
#include "dbRecType.h"
#include "devSup.h"
#include "errMdef.h"
#include "recSup.h"
#include "special.h"
#include "rec/valveRecord.h"
#include "redflds.h"
#include "redque.h"
#include "valveredfldidx.h"
#include "tsDefs.h"          
#include "tankfarm.h"
#include "restable.h"
#include "db_addr.h"
#include "dbEvent.h"
#include "recGbl.h"
#include "okToUse.h"
#include "dbcSub.h"

/* defined as global in .../share/db/iocInit.c */
extern long *process_cnts;
/* defined as global in .../share/src/procstat.c */
extern int process_stats;

/* Create RSET - Record Support Entry Table */

#define report NULL
#define initialize NULL
static long init_record();
static long process();
#define special NULL
#define get_value NULL
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
#define get_units NULL
static long get_precision();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double ();
static long get_control_double ();
#define get_alarm_double NULL

struct rset valveRSET = {

   RSETNUMBER,
   report,
   initialize,
   init_record,
   process,
   special,
   get_value,
   cvt_dbaddr,
   get_array_info,
   put_array_info,
   get_units,
   get_precision,
   get_enum_str,
   get_enum_strs,
   put_enum_str,
   get_graphic_double,
   get_control_double,
   get_alarm_double
   };


/* The folowing definitions are for redundancy */
 
static struct fldDes **mylist = 0;

   /* NOTE: the order and spelling in namelist must be the
            same as in valveredfldidx.h
   */

static char *namelist [] = {
		"STAT",
		"SEVR",
		"VAST",
		"SWST",
		"VASV",
		"SWSV",
		"LMOD",
		"ESIN",
		"EXOP",
		"EXCL",
		"BSIN",
		"VCUR",
		"BTIM",
		"TTIM",
		"FSRC",
		"SMOD",
		"LMEN",
		"LCMD",
		"PLTR",
		"PCMD",
		"MLTL",
		"INI2",
		"CMDS",
		"COPN",
		"CCLS",
		"CREV",
		"CJGO",
		"CJGC",
		"CPOS",
		"OPNS",
		"CLSS",
		"REVS",
		"JGOS",
		"JGCS",
		"POSS",
		"LATM",
		"VAL",
		0
		};

   /* structure to save some of the values in the
      record at start of processing, most are used
      for redunadancy checking
   */

struct valvered
{
   unsigned short    stat;
   unsigned short    sevr;
   double            val;
   char              vast[18];
   char              swst[18];
   char		     pcmd[18];
   short             vasv;
   short             swsv;
   short             lmod;
   short             esin;
   double            exop;
   double            excl;
   short             bsin;
   double            vcur;
   double            btim;
   double            ttim;
   unsigned long     tcks;
   short             fsrc;
   unsigned short    smod;
   unsigned short    lmen;
   short             lcmd;
   unsigned short    pltr;
   short             ini2;
   double            mltl;
   double            vpos;
   char              cmds[18];
   short             copn;
   short             ccls;
   short             crev;
   short             cjgo;
   short             cjgc;
   short             cpos;
   short             opns;
   short             clss;
   short             revs;
   short             jgos;
   short             jgcs;
   short             poss;
   TS_STAMP          latm;
};

struct valve_vars
{
    int     alarm_reset;		/* 1 = reset previous valv alarm condition */
    int     close_indicators;		/* set by check_current_valve_status () */
    short   close_limit_active;		/* returned by the close limit switch */
    short   close_torque_active;	/* returned by close torque switch */
    short   current_valve_state;	/* set by check_current_valve_status () */
    double  delta_time;			/* diff in secs between this scan and 
					   last scan */
    short   drive_position_pv_installed;  /* set by read_drive_position_pv () */
    short   fault_mbbi;			/* set by read_mbbi_fault_pv () */
    short   limit_switches_installed;	/* set by read_limit_switches () */
    int     open_indicators;		/* set by check_current_valve_status()*/
    short   open_limit_active;		/* returned by the open limit switch */
    short   open_torque_active;		/* returned by open torque switch */
    short   permissive;			/* set by read_permissive_pv () */
    short   position_pv_installed;	/* set by read_position_pv () */
    short   remote_local_trans;		/* = 1 on transition from remote to local */
    short   torque_switches_installed;	/* set by read_torque_switches () */
    short   valve_line_mode;		/* valve is single or dual bit mode */
};

static unsigned short type;           /* index to valve type for redundancy */

/* defines for valve_line_mode */

#define  NO_LINE_MODE      0
#define  SINGLE_BIT_MODE   1
#define  DUAL_BIT_MODE     2

/* defines for learn mode */

#define  VALVE_NOT_LEARNING        0
#define  VALVE_OPEN_LEARNING       1
#define  VALVE_CLOSE_LEARNING      2

/* defines for valve commands
   NOTE: they must have these values to
         coincide with valveRecord.ascii
*/

#define  VALVE_OPEN_CMD       1
#define  VALVE_CLOSE_CMD      2
#define  VALVE_REVERSE_CMD    3
#define  VALVE_JOG_OPEN_CMD   4
#define  VALVE_JOG_CLOSE_CMD  5
#define  VALVE_POSITION_CMD   6

/* defines for valve faults */

#define NO_VALVE_FAULT              0
#define VALVE_TIME_FAULT            1
#define VALVE_MBBI_FAULT            2
#define VALVE_LIMIT_TORQUE_FAULT    4

/* defines for valve torque/limit switches */

#define VALVE_SWITCH_NOT_ACTIVE     0
#define VALVE_SWITCH_ACTIVE         1

static void alarm (struct valveRecord *, 
		   struct valve_vars *);
static void check_current_valve_status (struct valveRecord *, 
					struct valve_vars *);
static long check_line_mode (struct valveRecord *, 
			     struct valve_vars *);
static long check_local_tracking (struct valveRecord *, 
				  struct valve_vars *);
static void check_switch_status (struct valveRecord *, 
				 struct valve_vars *);
static long drive_valve_to_position ();
static long emergency_stop (struct valveRecord *, 
			    struct valve_vars *);
static long get_drive_position_pv (struct valveRecord *, 
				   struct valve_vars *);
static void monitor ();
static void post_events ();
static long read_limit_switches (struct valveRecord *, 
				 struct valve_vars *);
static long read_mbbi_fault_pv (struct valveRecord *, 
				struct valve_vars *);
static long read_permissive_pv (struct valveRecord *, 
				struct valve_vars *);
static long read_position_pv (struct valveRecord *, 
			      struct valve_vars *);
static long read_rl_switch (struct valveRecord *, 
			    struct valve_vars *);
static long read_sensors (struct valveRecord *, 
			  struct valve_vars *);
static long read_torque_switches (struct valveRecord *, 
				  struct valve_vars *);
static void reset_valve_commands ();
static long run_commands (struct valveRecord *, 
			  struct valve_vars *);
static void save_valve ();
static long talk_to_valve (struct valveRecord *,
			   unsigned short *,
			   unsigned short *,
			   struct valve_vars *);
static void update_redun ();


/****************************************************************************
 *
 *  MODULE NAME            : init_record 
 *  AUTHOR                 : emsteel
 *  DATE                   : 04 May 93
 *
 *  ARGUMENT(S)            : pvalve - ptr to valve record  
 *
 *  DESCRIPTION: initialize a valve record
 *  NOTE:  current_valve_state is static global this source
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 dec 93
 *  REVISION:  init PCMD (last command). TF_INIT_CMD defined in 
 *             .../share/epicsH/tankfarm.h
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 22 dec 93
 *  REVISION:  init fields before call to read_sensors ()
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 04 Jan 94
 *  REVISION:  set ack to 0
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 28 Feb 94
 *  REVISION: remove ack = 0, and INIT from redund
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 13 jan 95
 *  REVISION:	init latm and lptm
 *
****************************************************************************/

static long init_record (pvalve)
   struct valveRecord	*pvalve;

{
    long		status;
    unsigned short	i;
    char		*pstr;
    int			ini2_flag = 0;
    struct valvered	valve_old;
    struct valve_vars	valveVals;

    if (!(mylist))
    {
         /* find record type for valve */
        for (type = 1, i = 0; type < 100; type++)
        {
            pstr = GET_PRECTYPE (type);     
            if (strncmp (pstr, "valve", 5) == 0)
            {
                i = 1;
                break;
            }
        }

        if (i == 0)
        {
            recGblRecordError (S_db_noRecType, (struct dbCommon *)pvalve, "valve:init_record");
            return (S_db_noRecType);
        }
         /* allocate memory for mylist */
        mylist = redFldAlloc (type, NUMVALVEFLDS);

	if (!mylist)
	{
	    recGblRecordError (S_rec_Alloc, (struct dbCommon *)pvalve, 
                               "valve, init_record, mylist failed");
	    return (S_rec_Alloc);
	}

         /* add fields in namelist to mylist */
        for (i = 0; i < NUMVALVEFLDS; i++)
           redFldAdd (type, i, namelist [i]);
    }

    pvalve -> init = TRUE;
    pvalve -> nsta = NO_ALARM;
    pvalve -> nsev = NO_ALARM;
    pvalve -> msev = NO_ALARM;
    /* init last alarm time and last process time */
    if (pvalve -> latm.secPastEpoch == 0)
	tsLocalTime (&pvalve -> latm);
    if (pvalve -> lptm.secPastEpoch == 0)
	tsLocalTime (&pvalve -> lptm);

    /* save current valve values */
    save_valve (pvalve, &valve_old);

    pvalve -> lmod = VALVE_NOT_LEARNING;   /* learn mode */
    pvalve -> bsin = FALSE;                /* bad state indicator */
    pvalve -> btim = 0;                    /* bad timer */
    pvalve -> esin = 0;                    /* emergency stop input */
    pvalve -> ttim = 0;                    /* travel timer */
    bzero (pvalve -> cmds, sizeof (pvalve -> cmds)); 
    pvalve -> copn = 0;
    pvalve -> ccls = 0;
    pvalve -> crev = 0;
    pvalve -> cjgo = 0;
    pvalve -> cjgc = 0;
    pvalve -> cpos = 0;
    pvalve -> opns = 0;
    pvalve -> clss = 0;
    pvalve -> revs = 0;
    pvalve -> jgos = 0;
    pvalve -> jgcs = 0;
    pvalve -> poss = 0;
    pvalve -> prec = 0;                    /* display precision */
    pvalve -> fsrc = NO_VALVE_FAULT;
    pvalve -> smod = pvalve -> mode;       /* save mode */
    strncpy (pvalve -> pcmd, TF_INIT_CMD, sizeof (pvalve -> pcmd));
    pvalve -> pltr = pvalve -> lten;       /* init previous local tracking */
    pvalve -> ralm = 0;			   /* reset alarms */

      /* check for valid line mode */

    status = read_sensors (pvalve, &valveVals);
    if (!status)
    {
	int dont_init = 0;

	pvalve -> vasv = VALVE_UNKNOWN;
	valveVals.alarm_reset = 0;

	check_current_valve_status (pvalve, &valveVals);

	switch (valveVals.current_valve_state)
	{
	    case VALVE_OPENED:
		strncpy (pvalve -> vast, pvalve -> stop, sizeof (pvalve -> vast));
		pvalve -> vasv = VALVE_OPENED;
		pvalve -> lcmd = VALVE_OPENED;
		break;

	    case VALVE_CLOSED:
		strncpy (pvalve -> vast, pvalve -> stcl, sizeof (pvalve -> vast));
		pvalve -> vasv = VALVE_CLOSED;
		pvalve -> lcmd = VALVE_CLOSED;
		break;

	    default:
		strncpy (pvalve -> vast, pvalve -> stun, sizeof (pvalve -> vast));
		pvalve -> vasv = VALVE_UNKNOWN;
		pvalve -> lcmd = -1;
		break;
	}
	pvalve -> val = pvalve -> vasv;

	/* check on valve outputs before initing valve */

	if (valveVals.valve_line_mode != NO_LINE_MODE)
	{
	    if (!okToUse (&pvalve -> vopp) )
		dont_init = 1;
	    if (valveVals.valve_line_mode == DUAL_BIT_MODE)
	    {
		if (!okToUse (&pvalve -> vclp) )
		    dont_init = 1;
	    }
	}
	if (valveVals.drive_position_pv_installed)
	{
	    if (!okToUse (&pvalve -> drop) )
		dont_init = 1;
	}
	if (dont_init)
	{
	    pvalve -> nsta = NO_INIT_ALARM;
	    pvalve -> nsev = VALID_ALARM;
	}
	else
	    ini2_flag = 1;
    }
    else
    {
	pvalve -> nsta = NO_INIT_ALARM;
	pvalve -> nsev = VALID_ALARM;
    }


      /* do redundancy if not in slave mode */

    if (resGetCaAnswer (pvalve -> radr, &i) == RES_OK)
    {
	if (i)
	{
	    /* check for maximize severity */

	    if (pvalve -> msev > pvalve -> nsev)
	    {
		pvalve -> nsev = pvalve -> msev;
		pvalve -> nsta = LINK_ALARM;
	    }

	    monitor (pvalve);

	    if (ini2_flag)
	    {
		pvalve -> ini2 = 1;
		post_events (pvalve, &valve_old);
		update_redun (pvalve, &valve_old);
	    }
	}
    }

    return (0);

} /* end func init_records () */



/****************************************************************************
 *
 *  MODULE NAME            : alarm
 *  AUTHOR                 : emsteel
 *  DATE                   : 26 Oct 93
 *
 *  CALLING MODULE         : process
 *  ARGUMENT(S)            : pvalve - ptr to valve record  
 *
 *  DESCRIPTION: perpetuate valve alarms if needed
 *               if alarm_reset or current severity >= previous severity then
 *                  normal processing will handle
 *
****************************************************************************/

static void alarm (pvalve, valveVals)
    struct valveRecord	*pvalve;
    struct valve_vars	*valveVals;
{
    int	adis = 0;

    adis = AlarmDisableLink ((void *)pvalve);
    if ( (pvalve -> aenb == ALARM_DISABLED) || (adis != 0) )
    {
	pvalve -> nsta = INHIBIT_ALARM;
	pvalve -> nsev = NO_ALARM;
	return;
    }
    if (valveVals -> alarm_reset)
	return;
    if (pvalve -> nsev >= pvalve -> sevr)
	return;
    if (pvalve -> ralm)
    {
	pvalve -> ralm = 0;
	if ( (pvalve -> stat == VALVE_OPEN_ALM) ||
	     (pvalve -> stat == VALVE_CLOSE_ALM)  ||
	     (pvalve -> stat == VALVE_POSITION)  ||
	     (pvalve -> stat == VALVE_TORQUE_LAG)  ||
	     (pvalve -> stat == CMD_NOT_ALLOWED) ||
	     (pvalve -> stat == TF_UNCOMMANDED) ||
	     (pvalve -> stat == AUTOMATIC) ||
	     (pvalve -> stat == MANUAL_OVERRIDE) ||
	     (pvalve -> stat == LINK_ALARM) )
	    return;
    }

    pvalve -> nsta = pvalve -> stat;
    pvalve -> nsev = pvalve -> sevr;

    return;

} /* end func alarm () */



/****************************************************************************
 *
 *  MODULE NAME            : check_current_valve_status 
 *  AUTHOR                 : emsteel
 *  DATE                   : 27 May 93
 *
 *  CALLING MODULE(S)      : init_record, process
 *          ARGUMENT(S)    : pvalve - ptr to valve record
 *
 *  DESCRIPTION: is the valve opened, closed or unknown
 *               the stop indicators govern which set of switches
 *               are to be used to indicate opened/closed
 *
 *               olsi - open limit stop indicator
 *               otsi - open torrque stop indicator
 *               clsi - close limit stop indicator
 *               ctsi - close torque stop indicator
 *               open_limit_active and close_limit active set by
 *                   read_limit_switches ()
 *               open_torque_active and close_torque_active set by
 *                   read_torque_switches ()
 *               update if commands in progress
 *               alarm_reset, set to 1 to cause valve alarms not to continue
 *                   see alarm ()
 *  NOTE: delta_time = diff in secs between last scan and current scan
 *        current_valve_state is static global this source
 *        open_indicators/close_indicators static global the source
 *        if pvalve -> lmen = learn mode enable
 *          if pvalve -> lmod == VALVE_NOT_LEARNING, then set exop/excl to
 *          pvalve -> ttim, else only if pvalve -> ttim is greater than
 *          current exop/excl, also update pvalve -> mltl (limit/torque
 *          lag time also
 *        without indicators (otsi etc) if a valve has exceeded its expected
 *          open/close time, then the valve must be considered opened/closed
 *        with indicators, if the valve has exceeded expected open/close
 *          time, then the valve is in a FAULT state
 *        if the valve was in a fault state due to MBBI fault and is not
 *          now, update valves status
 *        if the valve was in an open/close time fault or in a limit/torque
 *          time fault and is now opened/closed, update valves status
 *        if valve moved when it should not then generate an TF_UNCOMMANDED
 *          (uncommanded valve) alarm
 *        if only open limit indicator, if active, valve is opened else
 *          it is considered closed
 *        if only close limit indicator, if active, valve is closed else
 *          it is considered opened, however if the open limit inidicator
 *          is also active then the valve is in an UNKNOWN state
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 17 dec 93
 *  REVISION: if in LOCAL, then get current_valve_state, but do not update
 *              valve status or timers, etc. valve status will be updated in
 *              run_commands() in LOCAL logic
 *            deleted drop bits from POSITION cases, not needed since
 *              drive to position does not use start/stop pv
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 22 dec 93
 *  REVISION: also if not init'd yet get current_valve_state but do not
 *              update, see 17 dec revision
 *            also if VALVE_FAULT and TIME_FAULT or LIMIT_TORQUE, do not
 *              drop bits, already done when FAULT was set 
 *
****************************************************************************/

static void check_current_valve_status (pvalve, valveVals)
   struct valveRecord	*pvalve;
   struct valve_vars	*valveVals;
{
   int reset                  = 0;
   int uncommanded            = 0;    /* valve uncommanded movement */
   int single_indicator       = 0;    /* only olsi or clsi */
   int            drop_valve_bits = 0;
   double         dbvalue;       /* deadband value */

   valveVals -> current_valve_state = VALVE_UNKNOWN;
   valveVals -> open_indicators = valveVals -> close_indicators = 0;

   if (pvalve -> olsi && (!pvalve -> otsi) && 
        (!pvalve -> clsi) && (!pvalve -> ctsi) )
   {
      if (valveVals -> open_limit_active)
         valveVals -> current_valve_state = VALVE_OPENED;
      else
         valveVals -> current_valve_state = VALVE_CLOSED;
      single_indicator++;
   }
   else if (pvalve -> clsi && (!pvalve -> ctsi) && 
             (!pvalve -> olsi) && (!pvalve -> otsi) )
   {
      if (valveVals -> close_limit_active)
         valveVals -> current_valve_state = VALVE_CLOSED;
      else
         valveVals -> current_valve_state = VALVE_OPENED;
      single_indicator++;
   }
   else
   {
      if (pvalve -> olsi && pvalve -> otsi)
      {
         if (valveVals -> open_limit_active && 
	     valveVals -> open_torque_active)
            valveVals -> current_valve_state = VALVE_OPENED;
         valveVals -> open_indicators++;
      }
      else if (pvalve -> olsi)
      {
         if (valveVals -> open_limit_active)
            valveVals -> current_valve_state = VALVE_OPENED;
         valveVals -> open_indicators++;
      }
      else if (pvalve -> otsi)
      {
         if (valveVals -> open_torque_active)
            valveVals -> current_valve_state = VALVE_OPENED;
         valveVals -> open_indicators++;
      }
      if (pvalve -> clsi && pvalve -> ctsi)
      {
         if (valveVals -> close_limit_active && valveVals -> close_torque_active)
            valveVals -> current_valve_state = VALVE_CLOSED;
         valveVals -> close_indicators++;
      }
      else if (pvalve -> clsi)
      {
         if (valveVals -> close_limit_active)
         {
            if (valveVals -> open_limit_active)
               valveVals -> current_valve_state = VALVE_UNKNOWN;
            else
               valveVals -> current_valve_state = VALVE_CLOSED;
         }
         valveVals -> close_indicators++;
      }
      else if (pvalve -> ctsi)
      {
         if (valveVals -> close_torque_active)
            valveVals -> current_valve_state = VALVE_CLOSED;
         valveVals -> close_indicators++;
      }
   }

   if ((pvalve -> swsv == RL_LOCAL) || (pvalve -> ini2 == 0))
      return;

 switch (pvalve -> vasv)
   {
      case VALVE_OPENED:
         if (valveVals -> open_indicators || single_indicator)
         {
            if (valveVals -> current_valve_state == VALVE_UNKNOWN)
            {
               strncpy (pvalve -> vast, pvalve -> stun,
                                                  sizeof (pvalve -> vast));
               uncommanded++;
            }
            else if (valveVals -> current_valve_state == VALVE_CLOSED)
            {
               strncpy (pvalve -> vast, pvalve -> stcl,
                                                  sizeof (pvalve -> vast));
               if (pvalve -> lcmd == VALVE_OPENED)
                  uncommanded++;
               else
               {
                  pvalve -> lcmd = VALVE_CLOSED;
                  valveVals -> alarm_reset = 1;
               }
            }
            pvalve -> vasv = valveVals -> current_valve_state;
            reset++;
         }

         break;

      case VALVE_CLOSED:
         if (valveVals -> open_indicators || single_indicator)
         {
            if (valveVals -> current_valve_state == VALVE_UNKNOWN)
            {
               strncpy (pvalve -> vast, pvalve -> stun,
                                                  sizeof (pvalve -> vast));
               uncommanded++;
            }
            else if (valveVals -> current_valve_state == VALVE_OPENED)
            {
               strncpy (pvalve -> vast, pvalve -> stop,
                                                  sizeof (pvalve -> vast));
               if (pvalve -> lcmd == VALVE_CLOSED)
                  uncommanded++;
               else
               {
                  pvalve -> lcmd = VALVE_OPENED;
                  valveVals -> alarm_reset = 1;
               }
            }
            pvalve-> vasv = valveVals -> current_valve_state;
            reset++;
         }
         break;

     case VALVE_OPENING:

         pvalve -> ttim += valveVals -> delta_time;

         if (valveVals -> current_valve_state == VALVE_OPENED)
         {
            strncpy (pvalve -> vast, pvalve -> stop,
                                                  sizeof (pvalve -> vast));
            pvalve -> vasv = valveVals -> current_valve_state;
            if (pvalve -> lmen)
            {
               if (pvalve -> lmod & VALVE_OPEN_LEARNING)
               {
                  if (pvalve -> ttim > pvalve -> exop)
                     pvalve -> exop = pvalve -> ttim;
                  if (pvalve -> bsin == TRUE)
                  {
                     if (pvalve -> btim > pvalve -> mltl)
                        pvalve -> mltl = pvalve -> btim;
                  }
               }
               else if (valveVals -> open_indicators)
               {
                  pvalve -> lmod |= VALVE_OPEN_LEARNING;
                  pvalve -> exop = pvalve -> ttim;
             
                     /* if not already CLOSE LEARNING */


                   if (!(pvalve -> lmod & VALVE_CLOSE_LEARNING))
                   {
                      if (pvalve -> bsin == TRUE)
                         pvalve -> mltl = pvalve -> btim;
                   }
                  else
                  {
                     if (pvalve -> bsin == TRUE)
                     {
                        if (pvalve -> btim > pvalve -> mltl)
                           pvalve -> mltl = pvalve -> btim;
                     }
                  }
               }
            }
            if (valveVals -> valve_line_mode == DUAL_BIT_MODE)
            {
               drop_valve_bits++;
            }
            pvalve -> lcmd = VALVE_OPENED;
            reset++;
            valveVals -> alarm_reset = 1;
         }
         else
         {
               /* valve still opening or valve is closed,
                  is there an expected open time
                  only check if learn mode is not enabled */

            if (!pvalve -> lmen)
            {
                     /* add in open deadband time as a % of expected
                        open time */

               if (valveVals -> open_indicators || single_indicator)
                  dbvalue = (pvalve -> exop * (double)pvalve -> opdb)/100.0;
               else
                  dbvalue = 0.0;

               if (pvalve -> ttim > (pvalve -> exop + dbvalue))
               {
                  if (valveVals -> current_valve_state == VALVE_UNKNOWN)
                  {
                    if (valveVals -> open_indicators || single_indicator)
                     {
                        TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                            VALVE_OPEN_ALM, pvalve -> opsv);
                        strncpy (pvalve -> vast, pvalve -> stfa,
                                                   sizeof (pvalve -> vast));
                        pvalve -> vasv = VALVE_FAULT;
                        pvalve -> fsrc = VALVE_TIME_FAULT;
                        pvalve -> ttim = 0;
                        pvalve -> btim = 0;
                        pvalve -> bsin = FALSE;
                     }
                     else 
                     {
                        strncpy (pvalve -> vast, pvalve -> stop,
                                                   sizeof (pvalve -> vast));
                        pvalve -> vasv = VALVE_OPENED;
                        pvalve -> lcmd = VALVE_OPENED;
                        reset++;
                        valveVals -> alarm_reset = 1;
                     }
                  }
                     /* valve is closed */

                  else
                  {
                     strncpy (pvalve -> vast, pvalve -> stcl,
                                                  sizeof (pvalve -> vast));
                     pvalve -> vasv = valveVals -> current_valve_state;
                     TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                            VALVE_OPEN_ALM, pvalve -> opsv);
                     reset++;
                  }
                  if (valveVals -> valve_line_mode == DUAL_BIT_MODE)
                  {
                     drop_valve_bits++;
                  }
               }
            }
         }
         break;

     case VALVE_CLOSING:

         pvalve -> ttim += valveVals -> delta_time;

         if (valveVals -> current_valve_state == VALVE_CLOSED)
         {
            strncpy (pvalve -> vast, pvalve -> stcl,
                                                  sizeof (pvalve -> vast));
            pvalve -> vasv = valveVals -> current_valve_state;
            if (pvalve -> lmen)
            {
               if (pvalve -> lmod & VALVE_CLOSE_LEARNING)
               {
                  if (pvalve -> ttim > pvalve -> excl)
                     pvalve -> excl = pvalve -> ttim;
                  if (pvalve -> bsin == TRUE)
                  {
                     if (pvalve -> btim > pvalve -> mltl)
                        pvalve -> mltl = pvalve -> btim;
                  }
               }
               else if (valveVals -> close_indicators)
               {
                  pvalve -> lmod |= VALVE_CLOSE_LEARNING;
                  pvalve -> excl = pvalve -> ttim;

                     /* if not OPEN LEARNING */

                  if (!(pvalve -> lmod & VALVE_OPEN_LEARNING))
                  {
                     if (pvalve -> bsin == TRUE)
                        pvalve -> mltl = pvalve -> btim;
                  }
                  else
                  {
                     if (pvalve -> btim > pvalve -> mltl)
                        pvalve -> mltl = pvalve -> btim;
                  }
               }
            }
            if (valveVals -> valve_line_mode == DUAL_BIT_MODE)
            {
               drop_valve_bits++;
            }
            pvalve -> lcmd = VALVE_CLOSED;
            reset++;
            valveVals -> alarm_reset = 1;
         }
         else
         {
            if (!pvalve -> lmen)
            {
                  /* add in close deadband if there are close indicators */

               if (valveVals -> close_indicators || single_indicator)
                  dbvalue = (pvalve -> excl * (double)pvalve -> cldb)/100.0;
               else
                  dbvalue = 0.0;

               if (pvalve -> ttim > (pvalve -> excl + dbvalue))
               {
                  if (valveVals -> current_valve_state == VALVE_UNKNOWN)
                  {
                     if (valveVals -> close_indicators || single_indicator)
                     {
                        TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                           VALVE_CLOSE_ALM, pvalve -> clsv);
                        strncpy (pvalve -> vast, pvalve -> stfa,
                                                   sizeof (pvalve -> vast));
                        pvalve -> vasv = VALVE_FAULT;
                        pvalve -> fsrc = VALVE_TIME_FAULT;
                        pvalve -> ttim = 0;
                        pvalve -> btim = 0;
                        pvalve -> bsin = FALSE;
                     }
                     else
                     {
                        strncpy (pvalve -> vast, pvalve -> stcl,
                                                   sizeof (pvalve -> vast));
                        pvalve -> vasv = VALVE_CLOSED;
                        pvalve -> lcmd = VALVE_CLOSED;
                        reset++;
                        valveVals -> alarm_reset = 1;
                     }
                  }
                     /* valve is opened */

                  else
                  {
                     strncpy (pvalve -> vast, pvalve -> stop,
                                                   sizeof (pvalve -> vast));
                     pvalve -> vasv = VALVE_OPENED;
                     TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                           VALVE_CLOSE_ALM, pvalve -> clsv);
                     reset++;
                  }
                  if (valveVals -> valve_line_mode == DUAL_BIT_MODE)
                  {
                     drop_valve_bits++;
                  }
               }
            }
         }
         break;

     case VALVE_JOG_OPEN:
         if (valveVals -> current_valve_state == VALVE_OPENED)
         {
            strncpy (pvalve -> vast, pvalve -> stop,
                                                  sizeof (pvalve -> vast));
            pvalve -> vasv = valveVals -> current_valve_state;
            if (valveVals -> valve_line_mode == DUAL_BIT_MODE)
            {
               drop_valve_bits++;
            }
            pvalve -> lcmd = VALVE_OPENED;
            reset++;
            valveVals -> alarm_reset = 1;
         }
         else
         {
            pvalve -> ttim += valveVals -> delta_time;
   
               /* if have reached jog open time */
      
            if (pvalve -> ttim >= pvalve -> vjot)
            {
               if (valveVals -> current_valve_state == VALVE_UNKNOWN)
               {
                  strncpy (pvalve -> vast, pvalve -> stun,
                                                  sizeof (pvalve -> vast));
                  pvalve -> vasv = VALVE_UNKNOWN;
                  valveVals -> alarm_reset = 1;
               }
               else
               {
                  strncpy (pvalve -> vast, pvalve -> stcl,
                                                  sizeof (pvalve -> vast));
                  pvalve -> vasv = VALVE_CLOSED;
                  pvalve -> lcmd = VALVE_CLOSED;
                  TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                           VALVE_OPEN_ALM, pvalve -> opsv);
               }
                  /* stop the valve */
   
               if (valveVals -> valve_line_mode == DUAL_BIT_MODE)
               {
                  drop_valve_bits++;
               }
               reset++;
   
            }
         }
         break;

     case VALVE_JOG_CLOSE:
         if (valveVals -> current_valve_state == VALVE_CLOSED)
         {
            strncpy (pvalve -> vast, pvalve -> stcl,
                                                  sizeof (pvalve -> vast));
            pvalve -> vasv = valveVals -> current_valve_state;
            if (valveVals -> valve_line_mode == DUAL_BIT_MODE)
            {
               drop_valve_bits++;
            }
            pvalve -> lcmd = VALVE_CLOSED;
            reset++;
            valveVals -> alarm_reset = 1;
         }
         else
         {
            pvalve -> ttim += valveVals -> delta_time;
   
               /* if have reached jog open time */
      
            if (pvalve -> ttim >= pvalve -> vjct)
            {
               if (valveVals -> current_valve_state == VALVE_UNKNOWN)
               {
                  strncpy (pvalve -> vast, pvalve -> stun,
                                                   sizeof (pvalve -> vast));
                  pvalve -> vasv = VALVE_UNKNOWN;
                  valveVals -> alarm_reset = 1;
               }
               else
               {
                  strncpy (pvalve -> vast, pvalve -> stop,
                                                  sizeof (pvalve -> vast));
                  pvalve -> vasv = VALVE_OPENED;
                  pvalve -> lcmd = VALVE_OPENED;
                  TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                          VALVE_CLOSE_ALM, pvalve -> clsv);
               }
                  /* stop the valve */
   
               if (valveVals -> valve_line_mode == DUAL_BIT_MODE)
               {
                  drop_valve_bits++;
               }
               reset++;
            }
         }
         break;

     case VALVE_POSITION_OPENING:
         if (valveVals -> current_valve_state == VALVE_OPENED)
         {
            strncpy (pvalve -> vast, pvalve -> stop,
                                                  sizeof (pvalve -> vast));
            pvalve -> vasv = valveVals -> current_valve_state;
            pvalve -> lcmd = VALVE_OPENED;
            reset++;
            valveVals -> alarm_reset = 1;
         }
         else
         {
               /* if the current valve position >= desired position */
   
            if (pvalve -> vcur >= pvalve -> vpos)
            {
               strncpy (pvalve -> vast, "Positioned",
                                                   sizeof (pvalve -> vast));
               pvalve -> vasv = VALVE_POSITIONED;
               reset++;
               valveVals -> alarm_reset = 1;
            }
               /* did we exceed expected open time */

            else
            {
               pvalve -> ttim += valveVals -> delta_time;

                  /* add in open deadband time as a % of expected open time */

               if (valveVals -> open_indicators || single_indicator)
                  dbvalue = (pvalve -> exop * (double)pvalve -> opdb)/100.0;
               else
                  dbvalue = 0.0;

               if (pvalve -> ttim > (pvalve -> exop + dbvalue))
               {
                  if (valveVals -> current_valve_state == VALVE_UNKNOWN)
                  {
                     if (valveVals -> open_indicators || single_indicator)
                     {
                        TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                             VALVE_OPEN_ALM, pvalve -> opsv);
                        strncpy (pvalve -> vast, pvalve -> stfa,
                                                   sizeof (pvalve -> vast));
                        pvalve -> vasv = VALVE_FAULT;
                        pvalve -> fsrc = VALVE_TIME_FAULT;
                        pvalve -> ttim = 0;
                        pvalve -> btim = 0;
                        pvalve -> bsin = FALSE;
                     }
                     else 
                     {
                        strncpy (pvalve -> vast, pvalve -> stop,
                                                   sizeof (pvalve -> vast));
                        pvalve -> vasv = VALVE_OPENED;
                        valveVals -> current_valve_state = VALVE_OPENED;
                        pvalve -> lcmd = VALVE_OPENED;
                        reset++;
                        valveVals -> alarm_reset = 1;
                     }
                  }
                  else
                  {
                     TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                            VALVE_OPEN_ALM, pvalve -> opsv);
                     strncpy (pvalve -> vast, pvalve -> stcl,
                                                  sizeof (pvalve -> vast));
                     pvalve -> vasv = VALVE_CLOSED;
                     pvalve -> lcmd = VALVE_CLOSED;
                     reset++;
                  }
               }
            }
         }
         break;

     case VALVE_POSITION_CLOSING:
         if (valveVals -> current_valve_state == VALVE_CLOSED)
         {
            strncpy (pvalve -> vast, pvalve -> stcl,
                                                  sizeof (pvalve -> vast));
            pvalve -> vasv = valveVals -> current_valve_state;
            pvalve -> lcmd = VALVE_CLOSED;
            reset++;
            valveVals -> alarm_reset = 1;
         }
         else
         {
            if (pvalve -> vcur <= pvalve -> vpos)
            {
               strncpy (pvalve -> vast, "Positioned",
                                                   sizeof (pvalve -> vast));
               pvalve -> vasv = VALVE_POSITIONED;
               reset++;
               valveVals -> alarm_reset = 1;
            }
               /* did we exceed expected closing time */

            else
            {
                  /* add in close deadband time as a % of expected
                     close time */

               pvalve -> ttim += valveVals -> delta_time;
               if (valveVals -> close_indicators || single_indicator)
                  dbvalue = (pvalve -> excl * (double)pvalve -> cldb)/100.0;
               else
                  dbvalue = 0.0;
               if (pvalve -> ttim > (pvalve -> excl + dbvalue))
               {
                  if (valveVals -> current_valve_state == VALVE_UNKNOWN)
                  {
                     if (valveVals -> close_indicators || single_indicator)
                     {
                        TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                           VALVE_CLOSE_ALM, pvalve -> clsv);
                        strncpy (pvalve -> vast, pvalve -> stfa,
                                                   sizeof (pvalve -> vast));
                        pvalve -> vasv = VALVE_FAULT;
                        pvalve -> fsrc = VALVE_TIME_FAULT;
                        pvalve -> ttim = 0;
                        pvalve -> btim = 0;
                        pvalve -> bsin = FALSE;
                     }
                     else
                     {
                        strncpy (pvalve -> vast, pvalve -> stcl,
                                                   sizeof (pvalve -> vast));
                        pvalve -> vasv = VALVE_CLOSED;
                        pvalve -> lcmd = VALVE_CLOSED;
                        valveVals -> current_valve_state = VALVE_CLOSED;
                        reset++;
                        valveVals -> alarm_reset = 1;
                     }
                  }
                  else
                  {
                     strncpy (pvalve -> vast, pvalve -> stop,
                                                   sizeof (pvalve -> vast));
                     pvalve -> vasv = VALVE_OPENED;
                     pvalve -> lcmd = VALVE_OPENED;
                     TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                           VALVE_CLOSE_ALM, pvalve -> clsv);
                     reset++;
                  }
               }
            }
         }
         break;

    case VALVE_FAULT:

         if ((pvalve -> fsrc == VALVE_MBBI_FAULT) &&
             (valveVals -> fault_mbbi == NO_VALVE_FAULT) )
         {
            pvalve -> fsrc = NO_VALVE_FAULT;
            if (valveVals -> current_valve_state == VALVE_OPENED)
            {
               strncpy (pvalve -> vast, pvalve -> stop,
                                                  sizeof (pvalve -> vast));
               pvalve -> lcmd = VALVE_OPENED;
            }
            else if (valveVals -> current_valve_state == VALVE_CLOSED)
            {
               strncpy (pvalve -> vast, pvalve -> stcl,
                                                  sizeof (pvalve -> vast));
               pvalve -> lcmd = VALVE_CLOSED;
            }
            else
            {
               strncpy (pvalve -> vast, pvalve -> stun,
                                                  sizeof (pvalve -> vast));
               pvalve -> lcmd = -1;
            }
            pvalve -> vasv = valveVals -> current_valve_state;
            if (valveVals -> valve_line_mode == DUAL_BIT_MODE)
            {
               drop_valve_bits++;
            }
               /* clear out any command that may have been entered while
                  valve was in FAULT state
               */
            reset_valve_commands (pvalve);
            reset++;
            valveVals -> alarm_reset = 1;
         }
         else if ( (pvalve -> fsrc == VALVE_TIME_FAULT) ||
                   (pvalve -> fsrc == VALVE_LIMIT_TORQUE_FAULT) )
         {
            if (valveVals -> current_valve_state == VALVE_OPENED)
            {
               strncpy (pvalve -> vast, pvalve -> stop,
                                                   sizeof (pvalve -> vast));
               pvalve -> vasv = VALVE_OPENED;
               if (pvalve -> lcmd == VALVE_CLOSING)
                  uncommanded++;
               else
               {
                  pvalve -> lcmd = VALVE_OPENED;
                  valveVals -> alarm_reset = 1;
               }
               reset++;
            }
            else if (valveVals -> current_valve_state == VALVE_CLOSED)
            {
               strncpy (pvalve -> vast, pvalve -> stcl,
                                                   sizeof (pvalve -> vast));
               pvalve -> vasv = VALVE_CLOSED;
               if (pvalve -> lcmd == VALVE_OPENING)
                  uncommanded++;
               else
               {
                  pvalve -> lcmd = VALVE_CLOSED;
                  valveVals -> alarm_reset = 1;
               }
               reset++;
            }
         }

         break;

    case VALVE_UNKNOWN:
         if (valveVals -> current_valve_state == VALVE_CLOSED)
         {
            strncpy (pvalve -> vast, pvalve -> stcl,
                                                   sizeof (pvalve -> vast));
            pvalve -> vasv = VALVE_CLOSED;
            if (valveVals -> valve_line_mode == DUAL_BIT_MODE)
            {
               drop_valve_bits++;
            }
            reset++;
            if (pvalve -> lcmd == VALVE_OPENING)
               uncommanded++;
            else
            {
               pvalve -> lcmd = VALVE_CLOSED;
               valveVals -> alarm_reset = 1;
            }
         }
         else if (valveVals -> current_valve_state == VALVE_OPENED)
         {
            strncpy (pvalve -> vast, pvalve -> stop,
                                                   sizeof (pvalve -> vast));
            pvalve -> vasv = VALVE_OPENED;
            if (valveVals -> valve_line_mode == DUAL_BIT_MODE)
            {
               drop_valve_bits++;
            }
            reset++;
            if (pvalve -> lcmd == VALVE_CLOSING)
               uncommanded++;
            else
            {
               pvalve -> lcmd = VALVE_OPENED;
               valveVals -> alarm_reset = 1;
            }
         }
         pvalve -> vasv = valveVals -> current_valve_state;
         break;

      case VALVE_POSITIONED:
         if ( valveVals -> current_valve_state != VALVE_UNKNOWN)
         {
            if (valveVals -> current_valve_state == VALVE_CLOSED)
            {
               strncpy (pvalve -> vast, pvalve -> stcl,
                                                   sizeof (pvalve -> vast));
               if (pvalve -> lcmd == VALVE_OPENING)
                  uncommanded++;
               else
               {
                  pvalve -> lcmd = VALVE_CLOSED;
                  valveVals -> alarm_reset = 1;
               }
            }
            else
            {
               strncpy (pvalve -> vast, pvalve -> stop,
                                                   sizeof (pvalve -> vast));
               if (pvalve -> lcmd == VALVE_CLOSING)
                  uncommanded++;
               else
               {
                  pvalve -> lcmd = VALVE_OPENED;
                  valveVals -> alarm_reset = 1;
               }
            }
            pvalve -> vasv = valveVals -> current_valve_state;
            reset++;
         }
         break;

      default:
         break;

   } /* end switch (vasv) */

 if (drop_valve_bits)
   {
      unsigned short open_val = 0;
      unsigned short close_val = 0;
      long status = 0;

      status = talk_to_valve (pvalve, &open_val, &close_val, valveVals);
   }
   if (reset)
   {
      pvalve -> ttim = 0;
      pvalve -> btim = 0;
      pvalve -> fsrc = NO_VALVE_FAULT;
      pvalve -> bsin = FALSE;
   }
   if (uncommanded)
   {
      TankFarmPostAlarm ((struct dbCommon *)pvalve, TF_UNCOMMANDED,
                                                               MAJOR_ALARM);
   }
   
   return;

} /* end func check_current_valve_status () */


/****************************************************************************
 *
 *  MODULE NAME            : check_line_mode 
 *  AUTHOR                 : emsteel
 *  DATE                   : 04 May 93
 *
 *  CALLING MODULE(S)      : init_record, process
 *          ARGUMENT(S)    : pvalve - ptr to valve record
 *          GLOBAL(S)      : valve_line_mode (static this file)
 *
 *  DESCRIPTION: 1 - check for valid line mode
 *               2 - set  valve_line_mode accordingly
 *
****************************************************************************/

static long check_line_mode (pvalve, valveVals)
   struct valveRecord   *pvalve;
   struct valve_vars	*valveVals;
{
   long  status    = 0;
   short line_mode = 0;

      /* check for open valve pv */

   if (pvalve -> vopp.type != CONSTANT)
   {
      if (pvalve -> vopp.type == DB_LINK)
      {
         line_mode = 1;
      }
   }
      /* check for close valve pv */

   if (pvalve -> vclp.type != CONSTANT)
   {
      if (pvalve -> vclp.type == DB_LINK)
      {
         line_mode |= 2;
      }
   }

   if (line_mode == 1)
      valveVals -> valve_line_mode = SINGLE_BIT_MODE;
   else if (line_mode == 3)
      valveVals -> valve_line_mode = DUAL_BIT_MODE;
   else
      valveVals -> valve_line_mode = NO_LINE_MODE;

   return (status);
   
} /* end func check_line_mode () */


/****************************************************************************
 *
 *  MODULE NAME            : check_local_tracking
 *  AUTHOR                 : emsteel
 *  DATE                   : 05 May 97
 *
 *  CALLING MODULE(S)      : process
 *
 *  DESCRIPTION:  we are in local mode, track our status and our outputs
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : 
 *  REVISION DATE          :
 *  REVISION:  
 *
****************************************************************************/

static long check_local_tracking (pvalve, valveVals)
    struct valveRecord	*pvalve;
    struct valve_vars	*valveVals;
{
    unsigned short open_val;                 /* value for open pv */
    unsigned short close_val;                /* value for close pv */
    short	   save_valve_state = pvalve -> vasv;
    long	   status = 0;

    if (valveVals -> current_valve_state == VALVE_OPENED)
    {
	if (pvalve -> vasv != VALVE_OPENED)
	{
	    strncpy (pvalve -> vast, pvalve -> stop, sizeof (pvalve -> vast));
	    pvalve -> vasv = VALVE_OPENED;
	    pvalve -> lcmd = VALVE_OPENED;
	}
	/* and we are in single bit mode and local tracking is enabled */

	if (valveVals -> valve_line_mode == SINGLE_BIT_MODE)
	{
	    if (pvalve -> lten)
	    {
		if ( ( (save_valve_state != VALVE_OPENING) &&
		     (save_valve_state != VALVE_OPENED) ) ||
		     valveVals -> remote_local_trans ||
		     (pvalve -> pltr != pvalve -> lten) )
		{
		    open_val = 1;
		    close_val = 0;
		    status = talk_to_valve (pvalve, &open_val, &close_val,
					    valveVals);
		}
	    }
	}
    }
    else if (valveVals -> current_valve_state == VALVE_CLOSED)
    {
	if (pvalve -> vasv != VALVE_CLOSED)
	{
	    strncpy (pvalve -> vast, pvalve -> stcl, sizeof (pvalve -> vast));
	    pvalve -> vasv = VALVE_CLOSED;
	    pvalve -> lcmd = VALVE_CLOSED;
	}
	if (valveVals -> valve_line_mode == SINGLE_BIT_MODE)
	{
	    if (pvalve -> lten)
	    {
		if ( ( (save_valve_state != VALVE_CLOSING) &&
		     (save_valve_state != VALVE_CLOSED) ) ||
		     valveVals -> remote_local_trans ||
		     (pvalve -> pltr != pvalve -> lten) )
		{
		    open_val = 0;
		    close_val = 1;
		    status = talk_to_valve (pvalve, &open_val, &close_val,
					    valveVals);
		}
	    }
	}
    }
    if ( (valveVals -> valve_line_mode == DUAL_BIT_MODE) &&
	 (valveVals -> remote_local_trans) )
    {
	open_val = 0;
	close_val = 0;
	status = talk_to_valve (pvalve, &open_val, &close_val, valveVals);
    }
    pvalve -> pltr = pvalve -> lten;

    return (status);

} /* end func check_local_tracking () */



/****************************************************************************
 *
 *  MODULE NAME            : check_switch_status 
 *  AUTHOR                 : emsteel
 *  DATE                   : 10 May 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : pvalve
 *
 *  DESCRIPTION: check for bad relationship between limit switches and
 *                   torque switches
 *               there will be a lag pvalve -> mltl (in secs) between the
 *                   time a limit switch is active and its corresponding
 *                   torque switch to become active
 *                   if the valve is stuck then the torque switch may 
 *                   become active before the limit switch
 *  NOTE:
 *       pvalve -> otsi - open torque stop indicator
 *       pvalve -> olsi - open limit stop indicator
 *       pvalve -> clsi - close limit stop indicator
 *       pvalve -> ctsi - close torque stop indicator
 *       both indicators for open/close must be set to test limit/torque
 *          lag time
 *       only applicable if valve is OPENING or CLOSING
 *
****************************************************************************/

static void check_switch_status (pvalve, valveVals)
    struct valveRecord	*pvalve;
    struct valve_vars	*valveVals;
{
   unsigned short open_val;
   unsigned short close_val;
   long     status = 0;

   if ( (!valveVals -> torque_switches_installed) || 
	(!valveVals -> limit_switches_installed) )
      return;

   if ((pvalve -> vasv != VALVE_OPENING) && (pvalve -> vasv != VALVE_CLOSING))
      return;

   if ((!pvalve -> mltl) && (pvalve -> lmen == VALVE_NOT_LEARNING))
      return;
   if ((pvalve -> vasv == VALVE_OPENING) && ((!pvalve -> otsi)  ||
                                             (!pvalve -> olsi) ) )
      return;
   if ((pvalve -> vasv == VALVE_CLOSING) && ((!pvalve -> ctsi)  ||
                                             (!pvalve -> clsi) ) )
      return;

   if (pvalve -> bsin == FALSE)
   {
         /* if bad relatioship between switches, start to count
            the time when this condition exists
         */

      switch (pvalve -> vasv)
      {
         case VALVE_OPENING:
            if ( (valveVals -> open_limit_active && 
		  (!valveVals -> open_torque_active) ) ||
                 ( (!valveVals -> open_limit_active) && 
		   valveVals -> open_torque_active) )
            {
               pvalve -> bsin = TRUE;
               pvalve -> btim = 0;
            }
            break;

         case VALVE_CLOSING:
            if ( (valveVals -> close_limit_active && 
		  (!valveVals -> close_torque_active) ) ||
                 ( (!valveVals -> close_limit_active) && 
		   valveVals -> close_torque_active) )
            {
               pvalve -> bsin = TRUE;
               pvalve -> btim = 0;
            }
            break;

         default:
            break;
      }
   }
   else if ((valveVals -> open_limit_active == valveVals -> open_torque_active) &&
            (valveVals -> close_limit_active == valveVals -> close_torque_active))
   {
      pvalve -> bsin  = FALSE;
      pvalve -> btim = 0;
   }
   else
   {
         /* increment the timer and test against the limit,
            if in learn mode (pvalve -> lmen) do not test against
            pvalve -> mltl (limit/torque lag value) */

      pvalve -> btim += valveVals -> delta_time;
      if ((pvalve -> btim > pvalve -> mltl) && 
          (pvalve -> lmen == VALVE_NOT_LEARNING) )
      {
         TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                         VALVE_TORQUE_LAG, pvalve -> ltsv);
         if (valveVals -> valve_line_mode == DUAL_BIT_MODE)
         {
            open_val = 0;
            close_val = 0;
            status = talk_to_valve (pvalve, &open_val, &close_val, valveVals);
         }
         strncpy (pvalve -> vast, pvalve -> stfa, sizeof (pvalve -> vast));
         pvalve -> vasv = VALVE_FAULT;
         pvalve -> fsrc = VALVE_LIMIT_TORQUE_FAULT;
         pvalve -> ttim = 0;
         pvalve -> btim = 0;
         pvalve -> bsin = FALSE;
      }
   }

   return;

} /* end func check_switch_status () */



/****************************************************************************
 *
 *  MODULE NAME            : drive_valve_to_position 
 *  AUTHOR                 : emsteel
 *  DATE                   : 28 May 93
 *
 *  CALLING MODULE(S)      : run_commands
 *          ARGUMENT(S)    : pvalve - ptr to valve record
 *                           vpos - double ptr for valve position
 *
 *  DESCRIPTION: move the valve to postion vpos
 *
****************************************************************************/

static long drive_valve_to_position (pvalve, vpos)
   struct valveRecord   *pvalve;
   double               *vpos;
{
   long status    = 0;
   long nRequests = 1;

   status = dbPutLink (&pvalve -> drop.value.db_link,
                       (struct dbCommon *)pvalve, DBR_DOUBLE,
                       (caddr_t)vpos, nRequests);
   if (status)
   {
      TankFarmPostAlarm ((struct dbCommon *)pvalve, LINK_ALARM, VALID_ALARM);
   }

   return (status);

} /* end func drive_valve_to_position () */



/****************************************************************************
 *
 *  MODULE NAME            : emergency_stop 
 *  AUTHOR                 : emsteel
 *  DATE                   : 05 May 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : pvalve - ptr to valve record
 *
 *  DESCRIPTION: if an emergency stop has been initiated, then handle
 *               the valve accordingly
 *  NOTE: the value (open_val, close_val) has been defined as an
 *        unsigned short, since it assumed that the open/close pv are binary
 *        records and their val fields are defined as DBR_ENUM which
 *        translates to unsigned short
 *  NOTE: see choice.h for the defs of GET_... macros
 *  NOTE: if the valve is already closed/closing/positioned at 0%
 *          or opened/opening/positioned at 100% do not repeat
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 04 Jan 94
 *  REVISION: chg choiceInd from short to unsigned short to agree with what
 *            esac is 
 *
****************************************************************************/

static long emergency_stop (pvalve, valveVals)
   struct valveRecord   *pvalve;
   struct valve_vars	*valveVals;
{
   long                 status     = 0;
   unsigned short       open_val;
   unsigned short       close_val;
   short                choice_set = REC_VALVE_ES;    /* which choice set */
   unsigned short       choiceInd  = pvalve -> esac;  /* index into set   */
   struct arrChoiceSet  *parrChoiceSet;               /* ptr to valve sets */
   struct choiceSet     *pchoiceSet;                  /* ptr to set */
   char                 *pchoice_es;                  /* ptr to stop action */

   if (   (!(parrChoiceSet = GET_PARR_CHOICE_SET (choiceRec, type)))
       || (!(pchoiceSet = GET_PCHOICE_SET (parrChoiceSet, choice_set)))
       || (!(pchoice_es = GET_CHOICE (pchoiceSet, choiceInd))) )
   {
      recGblRecordError (S_db_badChoice, (struct dbCommon *)pvalve, "valve:estop action (1)");
      return (S_db_badChoice);
   }

      /* emergency stop action is to open the valve */

   if (strncmp (pchoice_es, "open", 4) == 0)
   {
      if (  ( (pvalve -> vasv == VALVE_OPENING) ||
              (pvalve -> vasv == VALVE_OPENED)  ||
              (pvalve -> vasv == VALVE_FAULT)   ||
              (pvalve -> stat == VALVE_OPEN_ALM) ||
              (pvalve -> nsta == VALVE_OPEN_ALM) ) ||
            ( ( (pvalve -> vasv == VALVE_POSITIONED) &&
                (pvalve -> vcur == 100) ) ||
              (pvalve -> vasv == VALVE_POSITION_OPENING) )  )
         return (status);
      open_val = 1;
      close_val = 0;
   }
      /* emergency stop action is to close the valve */

   else if (strncmp (pchoice_es, "close", 5) == 0)
   {
      if (  ( (pvalve -> vasv == VALVE_CLOSING) ||
              (pvalve -> vasv == VALVE_CLOSED)  ||
              (pvalve -> vasv == VALVE_FAULT)   ||
              (pvalve -> stat == VALVE_CLOSE_ALM) ||
              (pvalve -> nsta == VALVE_CLOSE_ALM) ) ||
            ( ( (pvalve -> vasv == VALVE_POSITIONED) &&
                (pvalve -> vcur == 0) ) ||
              (pvalve -> vasv == VALVE_POSITION_CLOSING) )  )
         return (status);
      open_val = 0;
      close_val = 1;
   }
   else
   {
      recGblRecordError (S_db_badChoice, (struct dbCommon *)pvalve, "valve:estop action (2)");
      return (S_db_badChoice);
   }

      /* we must be in remote mode to do anything software wise */

   if (pvalve -> swsv != RL_REMOTE)
   {
      TankFarmPostAlarm ((struct dbCommon *)pvalve, CMD_NOT_ALLOWED,
                                                                INFO_ALARM);
      return (status);
   }
   if ( (valveVals -> valve_line_mode == NO_LINE_MODE) && 
	(!valveVals -> drive_position_pv_installed) )
   {
      TankFarmPostAlarm ((struct dbCommon *)pvalve, CMD_NOT_ALLOWED,
                                                                INFO_ALARM);
      return (status);
   }

   if (valveVals -> valve_line_mode != NO_LINE_MODE)
   {
      status = talk_to_valve (pvalve, &open_val, &close_val, valveVals);
      if (status)
      {
         strncpy (pvalve -> vast, pvalve -> stun, sizeof (pvalve -> vast));
         pvalve -> vasv = VALVE_UNKNOWN;
         pvalve -> lcmd = -1;
      }
      else
      {
         if (open_val == 0)
         {
            strncpy (pvalve -> vast, pvalve -> stct, sizeof (pvalve -> vast));
            pvalve -> vasv = VALVE_CLOSING;
            pvalve -> lcmd = VALVE_CLOSING;
         }
         else
         {
            strncpy (pvalve -> vast, pvalve -> stot, sizeof (pvalve -> vast));
            pvalve -> vasv = VALVE_OPENING;
            pvalve -> lcmd = VALVE_OPENING;
         }
      }
   }
   else 
   {
      double vpos;

      if (open_val)
      { 
         vpos = 100;
         pvalve -> vpos = 100;
         strncpy (pvalve -> vast, pvalve -> stot, sizeof (pvalve -> vast));
         pvalve -> vasv = VALVE_POSITION_OPENING;
         pvalve -> lcmd = VALVE_OPENING;
      }
      else
      {
         vpos = 0;
         pvalve -> vpos = 0;
         strncpy (pvalve -> vast, pvalve -> stct, sizeof (pvalve -> vast));
         pvalve -> vasv = VALVE_POSITION_CLOSING;
         pvalve -> lcmd = VALVE_CLOSING;
      }
      status = drive_valve_to_position (pvalve, &vpos);
   }

   pvalve -> ttim = 0;
   pvalve -> btim = 0;
   pvalve -> fsrc = NO_VALVE_FAULT;
   pvalve -> bsin = FALSE;

   return (status);

} /* end func emergency_stop () */



/****************************************************************************
 *
 *  MODULE NAME            : get_drive_position_pv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 27 May 93
 *
 *  CALLING MODULE(S)      : read_sensors
 *          ARGUMENT(S)    : pvalve - ptr to valve record
 *
 *  DESCRIPTION: if the valve drive position pv is installed, then set
 *               drive_position_pv to 1
 *
****************************************************************************/

static long get_drive_position_pv (pvalve, valveVals)
   struct valveRecord   *pvalve;
   struct valve_vars	*valveVals;
{
   long status = 0;

   valveVals -> drive_position_pv_installed = 0;

   if ((pvalve -> drop.type != CONSTANT) &&
       (pvalve -> drop.type == DB_LINK))
      valveVals -> drive_position_pv_installed = 1;

   return (status);

} /* end func get_drive_position_pv () */



/****************************************************************************
 *
 *  MODULE NAME            : get_control_double 
 *  AUTHOR                 : emsteel
 *  DATE                   : 1 Sep 94
 *
 *
 *  DESCRIPTION: return control limits
 *
****************************************************************************/

static long get_control_double(paddr,pcd)
    struct	dbAddr		*paddr;
    struct	dbr_ctrlDouble	*pcd;
{
    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct valveRecord *)0) -> vjot) ):
	case ( (int)&(((struct valveRecord *)0) -> vjct) ):
	case ( (int)&(((struct valveRecord *)0) -> exop) ):
	case ( (int)&(((struct valveRecord *)0) -> excl) ):
	case ( (int)&(((struct valveRecord *)0) -> mltl) ):
	    pcd -> upper_ctrl_limit = 86400;	/* secs/day */
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct valveRecord *)0) -> disl) ):
	case ( (int)&(((struct valveRecord *)0) -> dish) ):
	case ( (int)&(((struct valveRecord *)0) -> hopr) ):
	case ( (int)&(((struct valveRecord *)0) -> lopr) ):
	case ( (int)&(((struct valveRecord *)0) -> vpos) ):
	case ( (int)&(((struct valveRecord *)0) -> opdb) ):
	case ( (int)&(((struct valveRecord *)0) -> cldb) ):
	    pcd -> upper_ctrl_limit = 100;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct valveRecord *)0) -> esin) ):
	    pcd -> upper_ctrl_limit = 1;
	    pcd -> lower_ctrl_limit = 0;
	    break;

	case ( (int)&(((struct valveRecord *)0) -> val) ):
	    pcd -> upper_ctrl_limit = 10;
	    pcd -> lower_ctrl_limit = 0;
	    break;
    }

    return (0);

} /* end func get_control_double () */


/****************************************************************************
 *
 *  MODULE NAME            : get_graphic_double 
 *  AUTHOR                 : emsteel
 *  DATE                   : 1 Sep 94
 *
 *
 *  DESCRIPTION: return graphic limits
 *
****************************************************************************/

static long get_graphic_double(paddr,pgd)
    struct	dbAddr		*paddr;
    struct	dbr_grDouble	*pgd;
{
    struct valveRecord    *pvalve = (struct valveRecord *) paddr->precord;

    switch ( ((struct fldDes *)paddr -> pfldDes) -> offset )
    {
	case ( (int)&(((struct valveRecord *)0) -> vjot) ):
	case ( (int)&(((struct valveRecord *)0) -> vjct) ):
	case ( (int)&(((struct valveRecord *)0) -> exop) ):
	case ( (int)&(((struct valveRecord *)0) -> excl) ):
	case ( (int)&(((struct valveRecord *)0) -> mltl) ):
	    pgd -> upper_disp_limit = 86400;	/* secs/day */
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct valveRecord *)0) -> disl) ):
	case ( (int)&(((struct valveRecord *)0) -> dish) ):
	case ( (int)&(((struct valveRecord *)0) -> hopr) ):
	case ( (int)&(((struct valveRecord *)0) -> lopr) ):
	case ( (int)&(((struct valveRecord *)0) -> vpos) ):
	case ( (int)&(((struct valveRecord *)0) -> opdb) ):
	case ( (int)&(((struct valveRecord *)0) -> cldb) ):
	case ( (int)&(((struct valveRecord *)0) -> vcur) ):
	    pgd -> upper_disp_limit = 100;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct valveRecord *)0) -> vasv) ):
	case ( (int)&(((struct valveRecord *)0) -> val) ):
	    pgd -> upper_disp_limit = 10;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct valveRecord *)0) -> swsv) ):
	case ( (int)&(((struct valveRecord *)0) -> esin) ):
	case ( (int)&(((struct valveRecord *)0) -> bsin) ):
	    pgd -> upper_disp_limit = 1;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct valveRecord *)0) -> lmod) ):
	    pgd -> upper_disp_limit = 2;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct valveRecord *)0) -> lcmd) ):
	    pgd -> upper_disp_limit = 3;
	    pgd -> lower_disp_limit = -1;
	    break;

	case ( (int)&(((struct valveRecord *)0) -> fsrc) ):
	    pgd -> upper_disp_limit = 4;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct valveRecord *)0) -> btim) ):
	    pgd -> upper_disp_limit = pvalve -> mltl;
	    pgd -> lower_disp_limit = 0;
	    break;

	case ( (int)&(((struct valveRecord *)0) -> ttim) ):
	{
	    double open_time = 0;
	    double close_time = 0;

	    open_time = pvalve -> exop + ((pvalve -> exop * pvalve -> opdb) / 100);
	    close_time = pvalve -> excl + ((pvalve -> excl * pvalve -> cldb) / 100);
	    pgd -> upper_disp_limit = (open_time > close_time) ? open_time : close_time;
	    pgd -> lower_disp_limit = 0;
	    break;
	}
    }
    return (0);

} /* end func get_graphic_double () */


/****************************************************************************
 *
 *  MODULE NAME            : get_precision 
 *  AUTHOR                 : emsteel
 *  DATE                   : 01 July 93
 *
 *
 *  DESCRIPTION: get display precision
 *  NOTE:
 *        see init_record, prec set to default of 2
 *
****************************************************************************/

static long get_precision (paddr, precision)
   struct dbAddr     *paddr;
   long              *precision;
{
   struct valveRecord *pvalve = (struct valveRecord *)paddr -> precord;

   *precision = pvalve -> prec;

   return (0);

} /* end func get_precision () */



/****************************************************************************
 *
 *  MODULE NAME            : monitor 
 *  AUTHOR                 : emsteel
 *  DATE                   : 04 May 93
 *
 *  CALLING MODULE(S)      : process ()
 *
 *  DESCRIPTION: check event list
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 05 Jan 94
 *  REVISION:  check VALID_ALARM before set ACK
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 04 May 94
 *  REVISION: deleted val, not used by valve
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 30 Aug 94
 *  REVISION:	rm DBE_LOG from ack
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 15 Mar 95
 *  REVISION:	do not set ack if sevr is EVENT_ALARM
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 11 Sept 95
 *  REVISION:	archive if stat/sevr also chg's
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 12 Sept 95
 *  REVISION:	chg way handle VALID_ALARM
 *
****************************************************************************/

static void monitor (pvalve)
   struct valveRecord   *pvalve;
{
   short		stat, sevr, nsta, nsev;
   unsigned short	monitor_mask;
   int			set_ack = 1;

      /* get previous stat and sevr and new stat and sevr */

   stat = pvalve -> stat;
   sevr = pvalve -> sevr;
   nsta = pvalve -> nsta;
   nsev = pvalve -> nsev;

      /* set current stat and sevr */

   pvalve -> stat = nsta;
   pvalve -> sevr = nsev;
   pvalve -> nsta = 0;
   pvalve -> nsev = 0;

    /* if we are coming out of VALID_ALARM and our new stat/sevr are
       the same as when we went into VALID_ALARM, do not set ack
    */
    if ( (nsev != VALID_ALARM) && (sevr == VALID_ALARM) &&
	 (nsta == pvalve -> psta) && (nsev == pvalve -> psev) )
    {
	set_ack = 0;
    }
    /* if we are going into VALID_ALARM, and out previous stat was not
       VALID_ALARM, remember previous non VALID_ALARM stat/sevr
       and we do not want to set ACK,
    */
    if ( (nsev == VALID_ALARM) && (sevr != VALID_ALARM) )
    {
	set_ack = 0;
	pvalve -> psta = stat;
	pvalve -> psev = sevr;
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> psta, DBE_VALUE);
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> psev, DBE_VALUE);
    }

   monitor_mask = 0;

      /* did alarm condition change this scan */

   if ((stat != nsta) || (sevr != nsev))
   {
      monitor_mask = DBE_ALARM;

      /* when alarm status chg's, get current time */
      tsLocalTime (&pvalve -> latm);

         /* post stat and sevr fields */

      db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> stat, DBE_VALUE);
      db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> sevr, DBE_VALUE);
      if ((pvalve -> sevr != NO_ALARM) && (set_ack) &&
	  (pvalve -> sevr != EVENT_ALARM) )
      {
         if (pvalve -> ack == 0)
         {
            pvalve -> ack = 1;
            db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> ack, DBE_VALUE);
         }
      }
   }

    if (pvalve -> val != pvalve -> vasv)
    {
	monitor_mask |= DBE_VALUE;
    }

    if (pvalve -> tenb)
    {
	if ( (pvalve -> val != pvalve -> vasv) || (monitor_mask & DBE_ALARM) )
	    monitor_mask |= DBE_LOG;
    }

    if (monitor_mask & (DBE_VALUE | DBE_LOG))
	pvalve -> val = pvalve -> vasv;

    if (monitor_mask)
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> val, monitor_mask);

   return;

} /* end func monitor () */


/****************************************************************************
 *
 *  MODULE NAME            : post_events 
 *  AUTHOR                 : emsteel
 *  DATE                   : 04 May 94
 *
 *
 *  DESCRIPTION: called to post value changes for monitors
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : 
 *  REVISION DATE          :
 *  REVISION: 
 *
****************************************************************************/

static void post_events (pvalve, pvalve_old)
    struct valveRecord	*pvalve;
    struct valvered	*pvalve_old;
{

    if (strncmp (pvalve -> pcmd, pvalve_old -> pcmd, sizeof (pvalve -> pcmd)) != 0)
	db_post_events ((struct dbCommon *)pvalve, pvalve -> pcmd, DBE_VALUE);

    if (strncmp (pvalve -> vast, pvalve_old -> vast, sizeof (pvalve -> vast)) != 0)
	db_post_events ((struct dbCommon *)pvalve, pvalve -> vast, DBE_VALUE);

    if (strncmp (pvalve -> swst, pvalve_old -> swst, sizeof (pvalve -> swst)) != 0)
	db_post_events ((struct dbCommon *)pvalve, pvalve -> swst, DBE_VALUE);

    if (pvalve -> vasv != pvalve_old -> vasv)
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> vasv, DBE_VALUE);

    if (pvalve -> lcmd != pvalve_old -> lcmd)
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> lcmd, DBE_VALUE);

    if (pvalve -> swsv != pvalve_old -> swsv)
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> swsv, DBE_VALUE);

    if (pvalve -> exop != pvalve_old -> exop)
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> exop, DBE_VALUE);

    if (pvalve -> excl != pvalve_old -> excl)
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> excl, DBE_VALUE);

    if (pvalve -> ttim != pvalve_old -> ttim)
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> ttim, DBE_VALUE);

    if (pvalve -> btim != pvalve_old -> btim)
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> btim, DBE_VALUE);

    if (pvalve -> mltl != pvalve_old -> mltl)
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> mltl, DBE_VALUE);

    if (pvalve -> lmod != pvalve_old -> lmod)
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> lmod, DBE_VALUE);

    if (pvalve -> fsrc != pvalve_old -> fsrc)
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> fsrc, DBE_VALUE);

    if (pvalve -> bsin != pvalve_old -> bsin)
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> bsin, DBE_VALUE);

    if (pvalve -> tcks != pvalve_old -> tcks)
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> tcks, DBE_VALUE);

    if (pvalve -> esin != pvalve_old -> esin)
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> esin, DBE_VALUE);

    if (pvalve -> lmen != pvalve_old -> lmen)
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> lmen, DBE_VALUE);

    if (pvalve -> vpos != pvalve_old -> vpos)
	db_post_events ((struct dbCommon *)pvalve, (caddr_t)&pvalve -> vpos, DBE_VALUE);

    return;

} /* end func post_events () */


/****************************************************************************
 *
 *  MODULE NAME            : process 
 *  AUTHOR                 : emsteel
 *  DATE                   : 04 May 93
 *
 *
 *  DESCRIPTION: called when time to process a valve record
 *  NOTE:
 *       delta_time = static global, used various places
 *       open_indicators and close_indicators are global and set by
 *          check_current_valve_status ()
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 8 Mar 94
 *  REVISION: added PERMISSIVE_ALM if valve dropped permissive
 *            if previous alarm was VALID, set alarm_reset to clear
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 12 Aug 94
 *  REVISION:	added code to process the forward link flnk
 *
****************************************************************************/

static long process (paddr)
   struct dbAddr  *paddr;
{
   struct valveRecord   *pvalve = (struct valveRecord *) (paddr -> precord);
   struct valvered      valve_old;
   struct valve_vars	valveVals;
   long                 status = 0;

      /* do nothing if we did not init */

   if ((!pvalve -> init) || (!pvalve -> ini2))
      return (0);

    /* get time stamp for last time record was processed */

    tsLocalTime (&pvalve -> lptm);

      /* save the current state of the tank for redundancy check */

   save_valve (pvalve, &valve_old);

      /* get system tick counter to use for time elapse */

   pvalve -> tcks = tickGet ();

      /* compute diff in secs tween last scan and this one */

   valveVals.delta_time = TFTicksDiffInSeconds (&pvalve -> tcks, &valve_old.tcks);

      /* init new alarm status to previous */

   pvalve -> nsta = pvalve -> stat;
   pvalve -> nsev = pvalve -> sevr;
   pvalve -> msev = NO_ALARM;

      /* off normal operations */

   if (pvalve -> tagd)
   {
      pvalve -> nsta = TAGGED_ALARM;
      pvalve -> nsev = VALID_ALARM;
   }
   else if (pvalve -> oscn == OFF_SCAN)
   {
      pvalve -> nsta = SCAN_ALARM;
      pvalve -> nsev = VALID_ALARM;
   }
   else
   {
      valveVals.alarm_reset = 0;
      if (pvalve -> sevr == VALID_ALARM)
        valveVals.alarm_reset = 1;
      pvalve -> nsta = NO_ALARM;
      pvalve -> nsev = NO_ALARM;

      /* update number of process times */
      if (process_stats)
         process_cnts[type]++;	

         /* if automatic or manual mode and alarm on condition
            then generate alarm, only if changed
         */

      if (pvalve -> mode != pvalve -> smod)
      {
         if ((pvalve -> mode == TF_AUTOMATIC) && (pvalve -> alau))
         {
            TankFarmPostAlarm ((struct dbCommon *)pvalve, AUTOMATIC,
                                                          pvalve -> ausv);
         }
         else if ((pvalve -> mode == TF_MANUAL) && (pvalve -> alma))
         {
            TankFarmPostAlarm ((struct dbCommon *)pvalve, MANUAL_OVERRIDE,
                                                          pvalve -> masv);
         }
         pvalve -> smod = pvalve -> mode;
      }
         /* read all valve switches and sensors */

      if (!(status = read_sensors (pvalve, &valveVals)))
      {
         if (valveVals.fault_mbbi == NO_VALVE_FAULT)
         {
               /* get current valve state, check cmds in progress,
                  update timers
               */

            check_current_valve_status (pvalve, &valveVals);

               /* check for limit/torque time lag */

            check_switch_status (pvalve, &valveVals);

               /* if emergency stop has been indicated */
   
            if (pvalve -> esin)
            {
               emergency_stop (pvalve, &valveVals);
               reset_valve_commands (pvalve);
               pvalve -> esin = 0;
            }
   
               /* if permission to move the valve has been denied
                  take emergency stop action
               */
   
            else if (valveVals.permissive == FALSE)
            {
               TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                            PERMISSIVE_ALM, MAJOR_ALARM);
               emergency_stop (pvalve, &valveVals);
               reset_valve_commands (pvalve);
            }
   
            else
            {
                  /* if we were in PERMISSIVE alarm, now we are
                     not, so set to clear it
                  */
               if (pvalve -> stat == PERMISSIVE_ALM)
                  valveVals.alarm_reset = 1;

                  /* set valve to active */

               pvalve -> pact = TRUE;

               if (pvalve -> lmen)
               {
                     /* if learn mode is enabled but there are no
                        indicators then learning is not possible */

                  if ((!valveVals.open_indicators) && (!valveVals.close_indicators))
                  {
                     TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                             CMD_NOT_ALLOWED, INFO_ALARM);
                     pvalve -> lmen = 0;
                     pvalve -> lmod = VALVE_NOT_LEARNING;
                  }
               }
               else
               {
                  /* if valve was learning, and now is not, reset 1st
                     learn indicator
                  */
   
                  pvalve -> lmod = VALVE_NOT_LEARNING;
               }
   
               status = run_commands (pvalve, &valveVals);
            }

	    if (pvalve -> swsv == RL_LOCAL)
		status = check_local_tracking (pvalve, &valveVals);

            alarm (pvalve, &valveVals);

         } /* end fault_mbbi == NO_VALVE_FAULT */

      } /* end status == 0 */

      /* check for maximize severity */

      if (pvalve -> msev > pvalve -> nsev)
      {
         pvalve -> nsev = pvalve -> msev;
         pvalve -> nsta = LINK_ALARM;
      }

   } /* end else (oscn != OFF_SCAN) */

      /* check event list */

   monitor (pvalve);

   pvalve -> pact = FALSE;

      /* post value changes for monitors */

   post_events (pvalve, &valve_old);

      /* now check for redundancy update */

   update_redun (pvalve, &valve_old);

   /* process the forward scan link record */

   if (pvalve -> flnk.type == DB_LINK) 
      dbScanPassive ((struct dbAddr *)pvalve -> flnk.value.db_link.pdbAddr);

   return (status);
   
} /* end func process () */



/****************************************************************************
 *
 *  MODULE NAME            : read_limit_switches 
 *  AUTHOR                 : emsteel
 *  DATE                   : 04 May 93
 *
 *  CALLING MODULE(S)      : init_record
 *                           process
 *          ARGUMENT(S)    : pvalve - ptr to a valve record
 *
 *  DESCRIPTION: if there are limit switches, set valve state accordingly
 *
****************************************************************************/

static long read_limit_switches (pvalve, valveVals)
   struct valveRecord	*pvalve;
   struct valve_vars	*valveVals;
{
   long     nRequests      = 1;
   long     options        = 0;
   long     status         = 0;
   short    limit_switches = 0;
   short    active         = VALVE_SWITCH_NOT_ACTIVE;

   valveVals -> open_limit_active = VALVE_SWITCH_NOT_ACTIVE;
   valveVals -> close_limit_active = VALVE_SWITCH_NOT_ACTIVE;
   valveVals -> limit_switches_installed = 0;

      /* check on the open/close limit switches */

   if (pvalve -> olsp.type != CONSTANT)
   {
      if (pvalve -> olsp.type == DB_LINK)
      {
            /* get value from sensor */
         status = dbGetLink (&pvalve -> olsp.value.db_link, 
                             (struct dbCommon *)pvalve, DBR_ENUM, 
                             (caddr_t)&active, &options, &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)pvalve, LINK_ALARM,
                                                               VALID_ALARM);
            return (status);
         }
	 else if (!pvalve -> ini2)
	 {
	    if (!okToUse (&pvalve -> olsp) )
		return (-1);
	 }
         if (active)
            valveVals -> open_limit_active = VALVE_SWITCH_ACTIVE;
         limit_switches = 1;
      }
   }


   if (pvalve -> clsp.type != CONSTANT)
   {
      if (pvalve -> clsp.type == DB_LINK)
      {
         active = VALVE_SWITCH_NOT_ACTIVE;
            /* get value from sensor */
         status = dbGetLink (&pvalve -> clsp.value.db_link, 
                             (struct dbCommon *)pvalve, DBR_ENUM, 
                             (caddr_t)&active, &options, &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)pvalve, LINK_ALARM,
                                                               VALID_ALARM);
            return (status);
         }
	 else if (!pvalve -> ini2)
	 {
	    if (!okToUse (&pvalve -> clsp) )
		return (-1);
	 }
         if (active)
            valveVals -> close_limit_active = VALVE_SWITCH_ACTIVE;
         limit_switches += 2;
      }
   }

   if (limit_switches == 3)
      valveVals -> limit_switches_installed = 1;

   return (status);
   
} /* end func read_limit_switches () */



/****************************************************************************
 *
 *  MODULE NAME            : read_mbbi_fault_pv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 27 May 93
 *
 *  CALLING MODULE(S)      : read_sensors ()
 *          ARGUMENT(S)    : pvalve - ptr to valve record
 *
 *  DESCRIPTION: read the fault mbbi pv
 *               if it returns a non 0, then
 *                  1 - set valve state to fault
 *                  2 - gen alarm
 *                  3 - return non 0 status
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 04 May 94
 *  REVISION: deleted dp_post_events, added reset_valve_commands
 *
****************************************************************************/

static long read_mbbi_fault_pv (pvalve, valveVals)
   struct valveRecord   *pvalve;
   struct valve_vars	*valveVals;
{
   long              status    = 0;
   unsigned short    fault_val;
   long              options   = 0;
   long              nRequests = 1;

   valveVals -> fault_mbbi = NO_VALVE_FAULT;

   if (pvalve -> fmbp.type != CONSTANT)
   {
      if ((pvalve -> fmbp.type = DB_LINK))
      {
         status = dbGetLink (&pvalve -> fmbp.value.db_link,
                             (struct dbCommon *)pvalve, DBR_ENUM,
                             (caddr_t)&fault_val, &options, &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)pvalve, LINK_ALARM,
                                                               VALID_ALARM);
            return (status);
         }
	 else if (!pvalve -> ini2)
	 {
	    if (!okToUse (&pvalve -> fmbp) )
		return (-1);
	 }
         if (fault_val)
         {
            valveVals -> fault_mbbi = VALVE_MBBI_FAULT;
            strncpy (pvalve -> vast, pvalve -> stfa,
                                                sizeof (pvalve -> vast));
            pvalve -> vasv = VALVE_FAULT;
            pvalve -> fsrc = VALVE_MBBI_FAULT;
            pvalve -> ttim = 0;
            pvalve -> btim = 0;
	    reset_valve_commands (pvalve);
            TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                           VALVE_FAULT_ALM, MAJOR_ALARM);
            return (S_tank_ValveFault);
         }
      }
   }

   return (status);

} /* end func read_mbbi_fault_pv () */



/****************************************************************************
 *
 *  MODULE NAME            : read_permissive_pv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 28 May 93
 *
 *  CALLING MODULE(S)      : read_sensors
 *          ARGUMENT(S)    : pvalve - ptr to valve record
 *
 *  DESCRIPTION: read the permissive pv, set global permissive accordingly
 *               if pv not installed, set permissive to TRUE
 *
****************************************************************************/

static long read_permissive_pv (pvalve, valveVals)
   struct valveRecord   *pvalve;
   struct valve_vars	*valveVals;
{
   long  status    = 0;
   long  options   = 0;
   long  nRequests = 1;

   valveVals -> permissive = TRUE;

   if (pvalve -> perp.type != CONSTANT)
   {
      if (pvalve -> perp.type == DB_LINK)
      {
         status = dbGetLink (&pvalve -> perp.value.db_link,
                             (struct dbCommon *)pvalve, DBR_ENUM,
                             (caddr_t)&valveVals -> permissive, &options, 
			     &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)pvalve, LINK_ALARM,
                                                               VALID_ALARM);
            valveVals -> permissive = FALSE;
         }
	 else if (!pvalve -> ini2)
	 {
	    if (!okToUse (&pvalve -> perp) )
		return (-1);
	 }
      }
   }

   return (status);

} /* end func read_permissive_pv () */



/****************************************************************************
 *
 *  MODULE NAME            : read_position_pv 
 *  AUTHOR                 : emsteel
 *  DATE                   : 06 May 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : pvalve - ptr to valve record
 *
 *  DESCRIPTION: read the position pv if it exists and get the current valve
 *               position
 *
****************************************************************************/

static long read_position_pv (pvalve, valveVals)
   struct valveRecord   *pvalve;
   struct valve_vars	*valveVals;
{
   long     status    = 0;
   long     nRequests = 1;
   long     options   = 0;
   double   curr_valve_position = pvalve -> vcur;

   valveVals -> position_pv_installed = 0;

   if (pvalve -> vpop.type != CONSTANT)
   {
      if (pvalve -> vpop.type == DB_LINK)
      {
         status = dbGetLink (&pvalve -> vpop.value.db_link,
                             (struct dbCommon *)pvalve, DBR_DOUBLE,
                             (caddr_t)&curr_valve_position,
                             &options, &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)pvalve, LINK_ALARM,
                                                               VALID_ALARM);
	    return (status);
         }
	 else if (!pvalve -> ini2)
	 {
	    if (!okToUse (&pvalve -> vpop) )
		return (-1);
	 }
         pvalve -> vcur = curr_valve_position;
         valveVals -> position_pv_installed = 1;
      }
   }

   return (status);
   
} /* end func read_position_pv () */



/****************************************************************************
 *
 *  MODULE NAME            : read_rl_switch 
 *  AUTHOR                 : emsteel
 *  DATE                   : 04 May 93
 *
 *  CALLING MODULE(S)      : init_record, check_rl_status
 *          ARGUMENT(S)    : pvalve - ptr to a valve record
 *
 *
 *  DESCRIPTION: set RL switch state (swst, swsv)
 *               if the RL switch is not installed, the status is REMOTE
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 16 dec 93
 *  REVISION:  added remote_local_trans code
 *
****************************************************************************/

static long read_rl_switch (pvalve, valveVals)
   struct valveRecord   *pvalve;
   struct valve_vars	*valveVals;
{
   long     nRequests       = 1;
   long     options         = 0;
   long     status          = 0;
   unsigned short curr_rl   = 0;             /* returned by the ROL switch */

   valveVals -> remote_local_trans = 0;

   if (pvalve -> rolp.type != CONSTANT)
   {
      if (pvalve -> rolp.type == DB_LINK)
      {
            /* get value from sensor */
         status = dbGetLink (&pvalve -> rolp.value.db_link, 
                             (struct dbCommon *)pvalve, DBR_ENUM, 
                             (caddr_t)&curr_rl, &options, &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)pvalve, LINK_ALARM,
                                                               VALID_ALARM);
            strncpy (pvalve -> swst, pvalve -> swre,
                                                   sizeof (pvalve -> swst));
            pvalve -> swsv = RL_REMOTE;
	    return (status);
         }
	 else if (!pvalve -> ini2)
	 {
	    if (!okToUse (&pvalve -> rolp) )
		return (-1);
	 }
         if (curr_rl == 1)
         {
            strncpy (pvalve -> swst, pvalve -> swre,
                                                sizeof (pvalve -> swst));
            pvalve -> swsv = RL_REMOTE;
         }
         else
         {
               /* if transition from remote to local */

            if (pvalve -> swsv == RL_REMOTE)
               valveVals -> remote_local_trans = 1;

            strncpy (pvalve -> swst, pvalve -> swlo,
                                                sizeof (pvalve -> swst));
            pvalve -> swsv = RL_LOCAL;
         }
      }
   }
   else
   {
      strncpy (pvalve -> swst, pvalve -> swre, sizeof (pvalve -> swst));
      pvalve -> swsv = RL_REMOTE;
   }

      /* turn off timers if going from REMOTE to LOCAL */
   if (valveVals -> remote_local_trans)
   {
      pvalve -> ttim = 0;
      pvalve -> btim = 0;
      pvalve -> bsin = FALSE;
      pvalve -> fsrc = NO_VALVE_FAULT;
   }

   return (status);
   
} /* end func read_rl_switch () */



/****************************************************************************
 *
 *  MODULE NAME            : read_sensors 
 *  AUTHOR                 : emsteel
 *  DATE                   : 27 May 93
 *
 *  CALLING MODULE(S)      : init_record, process
 *          ARGUMENT(S)    : pvalve - ptr to valve record
 *
 *  DESCRIPTION: read various switches and sensors
 *
****************************************************************************/

static long read_sensors (pvalve, valveVals)
   struct valveRecord	*pvalve;
   struct valve_vars	*valveVals;
{
   long  status = 0;

   if ((status = read_mbbi_fault_pv (pvalve, valveVals)) != 0)
      return (status);

   if ((status = check_line_mode (pvalve, valveVals)) != 0)
      return (status);

   if ((status = read_rl_switch (pvalve, valveVals)) != 0)
      return (status);

   if ((status = read_permissive_pv (pvalve, valveVals)) != 0)
      return (status);

   if ((status = read_limit_switches (pvalve, valveVals)) != 0)
      return (status);

   if ((status = read_torque_switches (pvalve, valveVals)) != 0)
      return (status);

   if ((status = read_position_pv (pvalve, valveVals)) != 0)
      return (status);

   if ((get_drive_position_pv (pvalve, valveVals)) != 0)
      return (status);

   return (status);

} /* end func read_sensors () */



/****************************************************************************
 *
 *  MODULE NAME            : read_torque_switches 
 *  AUTHOR                 : emsteel
 *  DATE                   : 04 May 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : pvalve - ptr to a valve record
 *
 *
 *  DESCRIPTION: if there are torque switches, set variables to indicate
 *
****************************************************************************/

static long read_torque_switches (pvalve, valveVals)
   struct valveRecord   *pvalve;
   struct valve_vars	*valveVals;
{
   long     nRequests          = 1;
   long     options            = 0;
   long     status             = 0;
   short    switches_installed = 0;
   short    active             = VALVE_SWITCH_NOT_ACTIVE;

   valveVals -> open_torque_active = VALVE_SWITCH_NOT_ACTIVE;
   valveVals -> close_torque_active = VALVE_SWITCH_NOT_ACTIVE;
   valveVals -> torque_switches_installed = 0;

      /* check on the open/close torque switches */

   if (pvalve -> otsp.type != CONSTANT)
   {
      if (pvalve -> otsp.type == DB_LINK)
      {
            /* get value from sensor */
         status = dbGetLink (&pvalve -> otsp.value.db_link, 
                             (struct dbCommon *)pvalve, DBR_ENUM, 
                             (caddr_t)&active, &options, &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)pvalve, LINK_ALARM,
                                                               VALID_ALARM);
            return (status);
         }
	 else if (!pvalve -> ini2)
	 {
	    if (!okToUse (&pvalve -> otsp) )
		return (-1);
	 }
         if (active)
            valveVals -> open_torque_active = VALVE_SWITCH_ACTIVE;
         switches_installed = 1;
      }
   }

   if (pvalve -> ctsp.type != CONSTANT)
   {
      if (pvalve -> ctsp.type == DB_LINK)
      {
         active = VALVE_SWITCH_NOT_ACTIVE;
            /* get value from sensor */
         status = dbGetLink (&pvalve -> ctsp.value.db_link, 
                             (struct dbCommon *)pvalve, DBR_ENUM, 
                             (caddr_t)&active, &options, &nRequests);
         if (status)
         {
            TankFarmPostAlarm ((struct dbCommon *)pvalve, LINK_ALARM,
                                                               VALID_ALARM);
            return (status);
         }
	 else if (!pvalve -> ini2)
	 {
	    if (!okToUse (&pvalve -> ctsp) )
		return (-1);
	 }
         if (active)
            valveVals -> close_torque_active = VALVE_SWITCH_ACTIVE;
         switches_installed += 2;
      }
   }

   if (switches_installed == 3)
      valveVals -> torque_switches_installed = 1;

   return (status);
   
} /* end func read_torque_switches () */


/****************************************************************************
 *
 *  MODULE NAME            : reset_valve_commands 
 *  AUTHOR                 : emsteel
 *  DATE                   : 28 July 93
 *
 *          ARGUMENT(S)    : pvalve - ptr to valve record
 *
 *  DESCRIPTION: reset valve commands
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 dec 93
 *  REVISION:  deleted db_post_events for command fields, not needed
 *             added special digital command
 *
****************************************************************************/

static void reset_valve_commands (pvalve)
struct   valveRecord    *pvalve;
{
   struct tfcmd   tfcmd;

   tfcmd.NumCmds                = 6;
   tfcmd.FirstDigitalCmd        = &pvalve -> copn;
   tfcmd.FirstSpecialDigitalCmd = &pvalve -> opns;
   tfcmd.FirstCmdString         = pvalve -> sopn;
   tfcmd.CmdString              = pvalve -> cmds;

   ResetTFCommand (&tfcmd);

   return;

} /* end func reset_valve_commands () */



/****************************************************************************
 *
 *  MODULE NAME            : run_commands 
 *  AUTHOR                 : emsteel
 *  DATE                   : 05 May 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : pvalve - ptr to valve record
 *
 *  DESCRIPTION: process a command if any
 *  NOTE:
 *       any valid command will cause the previous valve alarm to be reset
 *       reverse only makes sense when state is opening or closing
 *          not for jog or position
 *       can not jog a valve open/close if line_mode is SINGLE_BIT_MODE
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 dec 93
 *  REVISION:  added code to handle special digital commands
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 17 dec 93
 *  REVISION:  if local tracking enabled, only send signal to valve if
 *                change in status
 *                transition from remote to local
 *                transition from no local tracking to tracking
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 14 Feb 94
 *  REVISION: if local and if DUAL_BIT_MODE, drop output bits when
 *            transition to local regardless if local tracking is
 *            enabled
 *
****************************************************************************/

static long run_commands (pvalve, valveVals)
   struct valveRecord	*pvalve;
   struct valve_vars	*valveVals;
{
   unsigned short open_val;                 /* value for open pv */
   unsigned short close_val;                /* value for close pv */
   long           status    = 0;
   double         vpos;
   struct tfcmd   tfcmd;
   int            cmd;
   int            cmd_type;
   int            reset     = 0;


   tfcmd.NumCmds                = 6;
   tfcmd.FirstDigitalCmd        = &pvalve -> copn;
   tfcmd.FirstSpecialDigitalCmd = &pvalve -> opns;
   tfcmd.FirstCmdString         = pvalve -> sopn;
   tfcmd.CmdString              = pvalve -> cmds;

   cmd = GetTFCommand (&tfcmd, &cmd_type);

   if (pvalve -> swsv == RL_REMOTE)
   {
      if (cmd != NO_TF_COMMAND)
      {
         if (   (pvalve -> mode == TF_AUTOMATIC)
             || (pvalve -> mode == TF_MANUAL)   )
         {
   
            if (cmd == ILLEGAL_TF_STRING_CMD)
               TankFarmPostAlarm ((struct dbCommon *)pvalve, CMD_NOT_ALLOWED,
                                                               INFO_ALARM);
            else if ( ((pvalve -> mode == TF_AUTOMATIC) &&
                       ((cmd_type == TF_STRING_CMD) ||
                        (cmd_type == TF_SPECIAL_DIGITAL_CMD))) ||
                      ((pvalve -> mode == TF_MANUAL) &&
                       (cmd_type == TF_DIGITAL_CMD)) )
            {
               switch (cmd)
               {
                  case VALVE_OPEN_CMD:
                  case VALVE_JOG_OPEN_CMD: 
      
                     reset++;

                     if (valveVals -> valve_line_mode == NO_LINE_MODE)
                     {
                        TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                               CMD_NOT_ALLOWED, INFO_ALARM);
                        break;
                     }

                     if ((cmd == VALVE_JOG_OPEN_CMD) &&
                         (valveVals -> valve_line_mode == SINGLE_BIT_MODE) )
                     {
                        TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                               CMD_NOT_ALLOWED, INFO_ALARM);
                        break;
                     }

                        /* update valve status to OPENING */
      
                     strncpy (pvalve -> vast, pvalve -> stot,
                                                   sizeof (pvalve -> vast));
                     if (cmd == VALVE_OPEN_CMD)
                     {
                        pvalve -> vasv = VALVE_OPENING;
  
                           /* update PCMD, last command */

                        if (pvalve -> sopn[0])
                           strncpy (pvalve -> pcmd, pvalve -> sopn, 
                                                   sizeof (pvalve -> pcmd));
                        else
                           strncpy (pvalve -> pcmd, TF_UNDEF_CMD, 
                                                   sizeof (pvalve -> pcmd));
                     }
                     else
                     {
                        pvalve -> vasv = VALVE_JOG_OPEN;
  
                        if (pvalve -> sjop[0])
                           strncpy (pvalve -> pcmd, pvalve -> sjop, 
                                                   sizeof (pvalve -> pcmd));
                        else
                           strncpy (pvalve -> pcmd, TF_UNDEF_CMD, 
                                                   sizeof (pvalve -> pcmd));
                     }
                     pvalve -> lcmd = VALVE_OPENING;
      
      
                        /* tell valve to open */
      
                     open_val = 1;
                     close_val = 0;
                     status = talk_to_valve (pvalve, &open_val, &close_val, 
					     valveVals);
                     valveVals -> alarm_reset = 1;
                     break;
         
                  case VALVE_CLOSE_CMD:
                  case VALVE_JOG_CLOSE_CMD:
      
                     reset++;

                     if (valveVals -> valve_line_mode == NO_LINE_MODE)
                     {
                        TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                               CMD_NOT_ALLOWED, INFO_ALARM);
                        break;
                     }

                     if ((cmd == VALVE_JOG_CLOSE_CMD) &&
                         (valveVals -> valve_line_mode == SINGLE_BIT_MODE) )
                     {
                        TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                               CMD_NOT_ALLOWED, INFO_ALARM);
                        break;
                     }

                        /* set valve status to CLOSING */
      
                     strncpy (pvalve -> vast, pvalve -> stct,
                                                   sizeof (pvalve -> vast));
                     if (cmd == VALVE_CLOSE_CMD)
                     {
                        pvalve -> vasv = VALVE_CLOSING;
  
                        if (pvalve -> scls[0])
                           strncpy (pvalve -> pcmd, pvalve -> scls, 
                                                   sizeof (pvalve -> pcmd));
                        else
                           strncpy (pvalve -> pcmd, TF_UNDEF_CMD, 
                                                   sizeof (pvalve -> pcmd));
                     }
                     else
                     {
                        pvalve -> vasv = VALVE_JOG_CLOSE;
  
                        if (pvalve -> sjcl[0])
                           strncpy (pvalve -> pcmd, pvalve -> sjcl, 
                                                   sizeof (pvalve -> pcmd));
                        else
                           strncpy (pvalve -> pcmd, TF_UNDEF_CMD, 
                                                   sizeof (pvalve -> pcmd));
                     }
                     pvalve -> lcmd = VALVE_CLOSING;

                         /* tell valve to close */
      
                     open_val = 0;
                     close_val = 1;
                     status = talk_to_valve (pvalve, &open_val, &close_val, 
					     valveVals);
                     valveVals -> alarm_reset = 1;
                     break;
         
                  case VALVE_REVERSE_CMD:
      
                     if (valveVals -> valve_line_mode == NO_LINE_MODE)
                     {
                        TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                               CMD_NOT_ALLOWED, INFO_ALARM);
                        break;
                     }

                     open_val = 0;
                     close_val = 0;
      
                     if (pvalve -> vasv == VALVE_OPENING) 
                     {
                        strncpy (pvalve -> vast, pvalve -> stct,
                                                    sizeof (pvalve -> vast));
                        pvalve -> vasv = VALVE_CLOSING;
                        pvalve -> lcmd = VALVE_CLOSING;
                        open_val = 0;
                        close_val = 1;
                     }
                     else if (pvalve -> vasv == VALVE_CLOSING)
                     {
                        strncpy (pvalve -> vast, pvalve -> stot,
                                                    sizeof (pvalve -> vast));
                        pvalve -> vasv = VALVE_OPENING;
                        pvalve -> lcmd = VALVE_OPENING;
                        open_val = 1;
                        close_val = 0;
                     }
  
                     if (pvalve -> srev[0])
                        strncpy (pvalve -> pcmd, pvalve -> srev, 
                                                sizeof (pvalve -> pcmd));
                     else
                        strncpy (pvalve -> pcmd, TF_UNDEF_CMD, 
                                                sizeof (pvalve -> pcmd));

                     if (open_val || close_val)
                     {
                        status = talk_to_valve (pvalve, &open_val, &close_val,
						valveVals);

                           /* reset travel timer and fault indicator
                              only for status opening/closing not for
                              jog open/close or position open/close */
      
                        reset++;
                        valveVals -> alarm_reset = 1;
                     }
                     else
                     {
                        TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                                CMD_NOT_ALLOWED, INFO_ALARM);
                     }
                     break;
         
                  case VALVE_POSITION_CMD: 
      
                        /* if drive position pv is installed, and the pv
                           to read the current valve position is installed */
      
                     if (valveVals -> drive_position_pv_installed && 
			 valveVals -> position_pv_installed)
                     {
                        reset++;

                        vpos = pvalve -> vpos;
      
                        if ((vpos < 0.0) || (vpos > 100.0))
                        {
                           TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                                VALVE_POSITION, MAJOR_ALARM);
                           break;
                        }
                           /* new position can not be less than low limit
                              or higher than high limit
                           */
      
                        if (pvalve -> lopr)
                        {
                           if (vpos < pvalve -> lopr)
                           {
                              TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                                VALVE_POSITION, MAJOR_ALARM);
                              break;
                           }
                        }
                        if (pvalve -> hopr)
                        {
                           if (vpos > pvalve -> hopr)
                           {
                              TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                                VALVE_POSITION, MAJOR_ALARM);
                              break;
                           }
                        }
      
                        if (vpos == pvalve -> vcur)
                        {
                           TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                                CMD_NOT_ALLOWED, INFO_ALARM);
                        }
                        else
                        {
                           if (vpos > pvalve -> vcur)
                           {
                              strncpy (pvalve -> vast, pvalve -> stot,
                                                    sizeof (pvalve -> vast));
                              pvalve -> vasv = VALVE_POSITION_OPENING;
                              pvalve -> lcmd = VALVE_OPENING;
                           }
                           else 
                           {
                              strncpy (pvalve -> vast, pvalve -> stct,
                                                    sizeof (pvalve -> vast));
                              pvalve -> vasv = VALVE_POSITION_CLOSING;
                              pvalve -> lcmd = VALVE_CLOSING;
                           }
                           status = drive_valve_to_position (pvalve, &vpos);
                           valveVals -> alarm_reset = 1;
  
                           if (pvalve -> spos[0])
                              strncpy (pvalve -> pcmd, pvalve -> spos, 
                                                      sizeof (pvalve -> pcmd));
                           else
                              strncpy (pvalve -> pcmd, TF_UNDEF_CMD, 
                                                      sizeof (pvalve -> pcmd));
                        }
                     }
                     else
                     {
                        TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                                CMD_NOT_ALLOWED, INFO_ALARM);
                     }
                     break;
         
                  default:
                     TankFarmPostAlarm ((struct dbCommon *)pvalve,
                                                CMD_NOT_ALLOWED, INFO_ALARM);
               } /* end switch (cmd) */

            } /* end else if */

            else
            {
               TankFarmPostAlarm ((struct dbCommon *)pvalve, CMD_NOT_ALLOWED, 
                                                               INFO_ALARM);
            }

         } /* end if AUTOMATIC or MANUAL */

         else if (pvalve -> mode == TF_LOCKED)
         {
            TankFarmPostAlarm ((struct dbCommon *)pvalve, CMD_NOT_ALLOWED,
                                                                 INFO_ALARM);
         }

      } /* end cmd != NO_TF_COMMAND */

   } /* end if RL_REMOTE */

      /* are local */
   else
   {
      if (cmd != NO_TF_COMMAND)
         TankFarmPostAlarm ((struct dbCommon *)pvalve, CMD_NOT_ALLOWED,
                                                                  INFO_ALARM);
   }

   if (reset)
   {
      pvalve -> ttim = 0;
      pvalve -> btim = 0;
      pvalve -> fsrc = NO_VALVE_FAULT;
      pvalve -> bsin = FALSE;
   }

   reset_valve_commands (pvalve);

   return (status);

} /* end func run_commands () */



/****************************************************************************
 *
 *  MODULE NAME            : save_valve 
 *  AUTHOR                 : emsteel
 *  DATE                   : 04 May 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : pvalve - ptr to valve record
 *                           pvalve_old - ptr to valve redundancy struct
 *
 *  DESCRIPTION: save values from the valve record for redundancy check
 *               at end of process
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 28 Feb 94
 *  REVISION:  deleted init, not needed
 *
****************************************************************************/

static void save_valve (pvalve, pvalve_old)
   struct  valveRecord  *pvalve;
   struct  valvered     *pvalve_old;
{
   pvalve_old -> stat = pvalve -> stat;
   pvalve_old -> sevr = pvalve -> sevr;
   pvalve_old -> val = pvalve -> val;
   strncpy (pvalve_old -> vast, pvalve -> vast, sizeof (pvalve_old -> vast));
   strncpy (pvalve_old -> swst, pvalve -> swst, sizeof (pvalve_old -> swst));
   strncpy (pvalve_old -> pcmd, pvalve -> pcmd, sizeof (pvalve_old -> pcmd));
   pvalve_old -> vasv = pvalve -> vasv;
   pvalve_old -> swsv = pvalve -> swsv;
   pvalve_old -> lmod = pvalve -> lmod;
   pvalve_old -> esin = pvalve -> esin;
   pvalve_old -> exop = pvalve -> exop;
   pvalve_old -> excl = pvalve -> excl;
   pvalve_old -> bsin = pvalve -> bsin;
   pvalve_old -> vcur = pvalve -> vcur;
   pvalve_old -> btim = pvalve -> btim;
   pvalve_old -> ttim = pvalve -> ttim;
   pvalve_old -> tcks = pvalve -> tcks;
   pvalve_old -> fsrc = pvalve -> fsrc;
   pvalve_old -> smod = pvalve -> smod;
   pvalve_old -> lmen = pvalve -> lmen;
   pvalve_old -> lcmd = pvalve -> lcmd;
   pvalve_old -> pltr = pvalve -> pltr;
   pvalve_old -> mltl = pvalve -> mltl;
   pvalve_old -> ini2 = pvalve -> ini2;
   pvalve_old -> vpos = pvalve -> vpos;
   strncpy (pvalve_old -> cmds, pvalve -> cmds, sizeof (pvalve_old -> cmds));
   pvalve_old -> copn = pvalve -> copn;
   pvalve_old -> ccls = pvalve -> ccls;
   pvalve_old -> crev = pvalve -> crev;
   pvalve_old -> cjgo = pvalve -> cjgo;
   pvalve_old -> cjgc = pvalve -> cjgc;
   pvalve_old -> cpos = pvalve -> cpos;
   pvalve_old -> opns = pvalve -> opns;
   pvalve_old -> clss = pvalve -> clss;
   pvalve_old -> revs = pvalve -> revs;
   pvalve_old -> jgos = pvalve -> jgos;
   pvalve_old -> jgcs = pvalve -> jgcs;
   pvalve_old -> poss = pvalve -> poss;
   pvalve_old -> latm = pvalve -> latm;
   
   return;

} /* end func save_valve () */



/****************************************************************************
 *
 *  MODULE NAME            : talk_to_valve 
 *  AUTHOR                 : emsteel
 *  DATE                   : 06 May 93
 *
 *  CALLING MODULE(S)      : various
 *          ARGUMENT(S)    : pvalve - ptr to valve record
 *                           open_val - value to send to open pv
 *                           close_val - value to send to close pv
 *
 *  DESCRIPTION: set open and close valve pvs, if dual bit valve, make sure
 *		 we write the 0 value out first to prevent possibility of
 *		 having the open and close up even for a brief time
 *
****************************************************************************/

static long talk_to_valve (pvalve, open_val, close_val, valveVals)
   struct valveRecord   *pvalve;
   unsigned short       *open_val;
   unsigned short       *close_val;
   struct valve_vars	*valveVals;
{
   long  status = 0;
   long  nRequests = 1;

    if (valveVals -> valve_line_mode == SINGLE_BIT_MODE)
    {
	status = dbPutLink (&pvalve -> vopp.value.db_link,
			    (struct dbCommon *)pvalve, DBR_ENUM,
			    (caddr_t)open_val, nRequests);
	if (status)
	{
	    TankFarmPostAlarm ((struct dbCommon *)pvalve, LINK_ALARM, VALID_ALARM);
	    return (status);
	}
    }
      /* if dual bit mode, set close valve state */

    else if (valveVals -> valve_line_mode == DUAL_BIT_MODE)
    {
	if ((*open_val) > 0)
	{
	    status = dbPutLink (&pvalve -> vclp.value.db_link,
				(struct dbCommon *)pvalve, DBR_ENUM,
				(caddr_t)close_val, nRequests);
	    if (status)
	    {
		TankFarmPostAlarm ((struct dbCommon *)pvalve, LINK_ALARM, 
				   VALID_ALARM);
		return (status);
	    }
	    status = dbPutLink (&pvalve -> vopp.value.db_link,
			        (struct dbCommon *)pvalve, DBR_ENUM,
			        (caddr_t)open_val, nRequests);
	    if (status)
	    {
	        TankFarmPostAlarm ((struct dbCommon *)pvalve, LINK_ALARM, VALID_ALARM);
	        return (status);
	    }
	}
	else /* close_val >= 0 */
	{
	    status = dbPutLink (&pvalve -> vopp.value.db_link,
			        (struct dbCommon *)pvalve, DBR_ENUM,
			        (caddr_t)open_val, nRequests);
	    if (status)
	    {
	        TankFarmPostAlarm ((struct dbCommon *)pvalve, LINK_ALARM, VALID_ALARM);
	        return (status);
	    }
	    status = dbPutLink (&pvalve -> vclp.value.db_link,
				(struct dbCommon *)pvalve, DBR_ENUM,
				(caddr_t)close_val, nRequests);
	    if (status)
	    {
		TankFarmPostAlarm ((struct dbCommon *)pvalve, LINK_ALARM, 
				   VALID_ALARM);
		return (status);
	    }
	}
    }
    else
    {
	logMsg ("Valve error: talk_to_valve, %s, no line mode\n", 
		 (int)pvalve->name,0,0,0,0,0);
	status = 1;
    }

   return (status);
   
} /* end func talk_to_valve () */



/****************************************************************************
 *
 *  MODULE NAME            : update_redun 
 *  AUTHOR                 : emsteel
 *  DATE                   : 04 May 93
 *
 *  CALLING MODULE(S)      : process
 *          ARGUMENT(S)    : pvalve - ptr to valve record at end of process
 *                           pvalve_old - ptr to valve record before process
 *
 *  DESCRIPTION: have any of the fields deemed as being redundant, changed
 *
 ***************************** REVISION HISTORY *****************************
 *
 *  REVISED BY             : emsteel
 *  REVISION DATE          : 15 Feb 94
 *  REVISION: handle, LMEN, EXON, EXOF whether we are redundant or not,
 *            these are prompted fields and we must let dbClient know if thay
 *            change, dbserver will know if we are redunant, recal checksum
 *
****************************************************************************/

static void update_redun (pvalve, pvalve_old)
   struct valveRecord   *pvalve;
   struct valvered      *pvalve_old;
{
   char  redbuf[REDQ_MAX_MSGLEN];

   if ( (pvalve -> lmen != pvalve_old -> lmen) ||
        (pvalve -> exop != pvalve_old -> exop) ||
        (pvalve -> excl != pvalve_old -> excl) ||
	(pvalve -> mltl != pvalve_old -> mltl) )
   {
      redQInit ((struct dbCommon *)pvalve, redbuf, sizeof (redbuf), type);

      pvalve -> rcrc = DBC_CRCHostRecord ((char *)pvalve, type);

      if (pvalve -> lmen != pvalve_old -> lmen)
         redQField (redbuf, mylist[LMEN]);

      if (pvalve -> exop != pvalve_old -> exop)
         redQField (redbuf, mylist[EXOP]);

      if (pvalve -> excl != pvalve_old -> excl)
         redQField (redbuf, mylist[EXCL]);

      if (pvalve -> mltl != pvalve_old -> mltl)
         redQField (redbuf, mylist[MLTL]);

      redQSend (redbuf);
   }

   redQInit ((struct dbCommon *)pvalve, redbuf, sizeof (redbuf), type);

   if (pvalve -> stat != pvalve_old -> stat)
      redQField (redbuf, mylist[STAT]);
   
   if (pvalve -> sevr != pvalve_old -> sevr)
      redQField (redbuf, mylist[SEVR]);
   
   if (pvalve -> val != pvalve_old -> val)
      redQField (redbuf, mylist[VAL]);

   if (strncmp (pvalve -> vast, pvalve_old -> vast,
                                               sizeof (pvalve -> vast)) != 0)
      redQField (redbuf, mylist[VAST]);

   if (strncmp (pvalve -> swst, pvalve_old -> swst,
                                               sizeof (pvalve -> swst)) != 0)
      redQField (redbuf, mylist[SWST]);

   if (strncmp (pvalve -> pcmd, pvalve_old -> pcmd,
                                               sizeof (pvalve -> pcmd)) != 0)
      redQField (redbuf, mylist[PCMD]);

   if (pvalve -> vasv != pvalve_old -> vasv)
      redQField (redbuf, mylist[VASV]);

   if (pvalve -> swsv != pvalve_old -> swsv)
      redQField (redbuf, mylist[SWSV]);

   if (pvalve -> lmod != pvalve_old -> lmod)
      redQField (redbuf, mylist[LMOD]);

   if (pvalve -> esin != pvalve_old -> esin)
      redQField (redbuf, mylist[ESIN]);

   if (pvalve -> bsin != pvalve_old -> bsin)
      redQField (redbuf, mylist[BSIN]);

   if (pvalve -> vcur != pvalve_old -> vcur)
      redQField (redbuf, mylist[VCUR]);

   if (pvalve -> btim != pvalve_old -> btim)
      redQField (redbuf, mylist[BTIM]);

   if (pvalve -> ttim != pvalve_old -> ttim)
      redQField (redbuf, mylist[TTIM]);

   if (pvalve -> fsrc != pvalve_old -> fsrc)
      redQField (redbuf, mylist[FSRC]);

   if (pvalve -> smod != pvalve_old -> smod)
      redQField (redbuf, mylist[SMOD]);

   if (pvalve -> lcmd != pvalve_old -> lcmd)
      redQField (redbuf, mylist[LCMD]);

   if (pvalve -> pltr != pvalve_old -> pltr)
      redQField (redbuf, mylist[PLTR]);

   if (pvalve -> ini2 != pvalve_old -> ini2)
      redQField (redbuf, mylist[INI2]);

   if (strncmp (pvalve -> cmds, pvalve_old -> cmds,
                                               sizeof (pvalve -> cmds)) != 0)
      redQField (redbuf, mylist[CMDS]);

   if (pvalve -> copn != pvalve_old -> copn)
      redQField (redbuf, mylist[COPN]);

   if (pvalve -> ccls != pvalve_old -> ccls)
      redQField (redbuf, mylist[CCLS]);

   if (pvalve -> crev != pvalve_old -> crev)
      redQField (redbuf, mylist[CREV]);

   if (pvalve -> cjgo != pvalve_old -> cjgo)
      redQField (redbuf, mylist[CJGO]);

   if (pvalve -> cjgc != pvalve_old -> cjgc)
      redQField (redbuf, mylist[CJGC]);

   if (pvalve -> cpos != pvalve_old -> cpos)
      redQField (redbuf, mylist[CPOS]);

   if (pvalve -> opns != pvalve_old -> opns)
      redQField (redbuf, mylist[OPNS]);

   if (pvalve -> clss != pvalve_old -> clss)
      redQField (redbuf, mylist[CLSS]);

   if (pvalve -> revs != pvalve_old -> revs)
      redQField (redbuf, mylist[REVS]);

   if (pvalve -> jgos != pvalve_old -> jgos)
      redQField (redbuf, mylist[JGOS]);

   if (pvalve -> jgcs != pvalve_old -> jgcs)
      redQField (redbuf, mylist[JGCS]);

   if (pvalve -> poss != pvalve_old -> poss)
      redQField (redbuf, mylist[POSS]);

   if (TsCmpStampsNE (&pvalve -> latm, &pvalve_old -> latm) )
      redQField (redbuf, mylist [LATM]);
	
   redQSend (redbuf);

   return;

} /* end func update_redun () */

