/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <vxWorks.h>
#include	<inetLib.h>
#include <types.h>
#include <stdioLib.h>
#include <lstLib.h>
#include <taskLib.h>
#include <logLib.h>
#include <string.h>

#include "dbDefs.h"
#include "restable.h"
#include "resmgr.h"

static int taskId;
static int delayTicks;
int ResourceTest (char *resname, int delay_secs);


/**********************************************************************************/

int StartResTest (char *resname, int delay_secs)
{
    if (resname == NULL)
    {
	logMsg ("StartResTest: usage: StartResTest resname delay_secs\n",0,0,0,0,0,0);
    }
    else if ( (taskId = taskSpawn ("res_test", 215, VX_FP_TASK | VX_STDIO, 
				    2048, (FUNCPTR)ResourceTest, (int)resname, 
				    delay_secs,0,0,0,0,0,0,0,0)) == ERROR)
    {
	logMsg ("StartResTest: taskSpawn failed (%d)\n", errno,0,0,0,0,0);
    }

    return (0);
}


/**********************************************************************************/

int StopResTest (void)
{
    if (taskDelete (taskId) == ERROR)
    {
	logMsg ("StopResTest: taskDelete failed (%d)\n", errno,0,0,0,0,0);
    }

    return (0);
}


/**********************************************************************************/

int ResetResDelay (int secs)
{
    if (secs > 0)
    {
	delayTicks = secs * vxTicksPerSecond;
    }
    return (0);
}


/**********************************************************************************/

int ResourceTest (char *resname, int delay_secs)
{
    short	master;
    int		foundResource = 0;
    char	resName[32];
    RES_TABLE	*resEntry = resTableHead;

    if (delay_secs <= 0)
    {
	delayTicks = 30 * vxTicksPerSecond;
    }
    else
    {
	delayTicks = delay_secs * vxTicksPerSecond;
    }

    strcpy (resName, resname);
    logMsg ("RT, res=%s,dt=%d,re=%0x\n", (int)resName, delayTicks, (int)resEntry,0,0,0);

    while (resEntry)
    {
	if ( (strcmp (resName, resEntry -> resourceName)) == 0)
	{
	    foundResource = 1;
	    break;
	}
	resEntry = resEntry -> link;
    }
    if (foundResource == 0)
    {
	logMsg ("ResourceTest: failed to find resource %s\n", (int)resName,0,0,0,0,0);
	taskSuspend (taskIdSelf());
    }
    for (;;)
    {
	taskDelay (delayTicks);
	master = 0;
	if ( (resGetCaAnswer (resEntry, &master)) == RES_OK)
	{
	    if (master)
	    {
		rmSlave (resName);
	    }
	    else
	    {
		rmMaster (resName);
	    }
	}
    }
    taskSuspend (taskIdSelf());
}
