/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/



/****************************************************************************
 *
 *  MODULE NAME            : tcntReset
 *  AUTHOR                 : emsteel
 *  DATE                   : 22 Feb 95
 *                                     
 *
 *  DESCRIPTION: reset the TCNT (transition counter) field for all BI's whose
 *		 resource is master
 *
****************************************************************************/

/****************************************************************************
 *
 * Programmer's Notes:
 *	1) args to taskSpawn are found in .../share/epicsH/task_params.h
 *	2) vxTicksPerSecond defined in .../share/epicsH/dbDefs.h
 *	3) TS_STAMP defined in .../share/epicsH/tsDefs.h
 *	4) tsDetail defined in .../share/epicsH/tsDefs.h
 *	5) dbGetField() and dbPutField, see .../share/src/db/dbAccess.c
 *
****************************************************************************/

#include <vxWorks.h>
#include <inetLib.h>
#include <types.h>
#include <stdioLib.h>
#include <lstLib.h>
#include <taskLib.h>
#include <logLib.h>
#include <string.h>

#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#include "dbAccess.h"
#include "dbRecType.h"
#include "task_params.h"
#include "dbRecords.h"
#include "restable.h"
#include "rec/biRecord.h"
#include "db_addr.h"
#include "dbEvent.h"

void tcnt_reset (void);
int update_tcnt_reset (int mode, char *ptname);

static int	trTaskId;
static int	reset_mode = 0;
static int	new_mode = -1;
static int	new_async = 0;
static char	manual_reset_name[PVNAME_SZ + FLDNAME_SZ + 4];
static struct dbAddr	dbAddr;
static struct dbAddr	newdbAddr;


/****************************************************************************
 *
 *  MODULE NAME            : start_tcnt_reset
 *  AUTHOR                 : emsteel
 *  DATE                   : 22 Feb 95
 *
 *          ARGUMENT(S)    : mode - type of reset
 *			     ptname - ptname if mode is 0
 *
 *  DESCRIPTION: start the task to reset the TCNT field in all BI records
 *
****************************************************************************/

int start_tcnt_reset (mode, ptname)
    int		mode;
    char	*ptname;
{
    long	status = 0;

    if ( (mode < 0) || (mode > 3) )
    {
	logMsg ("start_tcnt_reset, mode range is 0 - 3 (%d)\n",mode,0,0,0,0,0);
	return (1);
    }
    bzero (manual_reset_name, sizeof (manual_reset_name) );
    if (mode == 0)
    {
	if (ptname[0] == NULL)
	{
	    logMsg ("start_tcnt_reset, mode 0 requires a pointname\n",0,0,0,0,0,0);
	    return (1);
	}
	strncpy (manual_reset_name, ptname, sizeof (manual_reset_name) );
	if ( (status = dbNameToAddr (manual_reset_name, &dbAddr)) != 0)
	{
	    logMsg ("start_tcnt_reset, %s not in IOC\n",(int)ptname,0,0,0,0,0);
	    return (1);
	}
    }
    reset_mode = mode;
    if ((trTaskId = taskSpawn (TCNT_RESET_NAME, TCNT_RESET_PRI,
				TCNT_RESET_OPT, TCNT_RESET_STACK,
				(FUNCPTR)tcnt_reset,0,0,0,0,0,
				0,0,0,0,0)) == ERROR)
    {
	logMsg ("failed to start tcnt_reset task(%d)\n", errno,0,0,0,0,0);
	return (1);
    }

    return (0);
}


/****************************************************************************
 *
 *  MODULE NAME            : tcnt_reset
 *  AUTHOR                 : emsteel
 *  DATE                   : 22 Feb 95
 *
 *  DESCRIPTION: the task to reset the TCNT field in all BI records
 *
****************************************************************************/

void tcnt_reset (void)
{
    unsigned short	last_async_reset = 0;
    unsigned short	curr_async_reset = 0;
    int			reset;
    long		status;
    long		options = 0;
    long		nRequest = 1;
    TS_STAMP		last_etime;		/* last epics time */
    TS_STAMP		curr_etime;		/* current epics time */
    struct tsDetail	last_time;		/* last epics time in local time */
    struct tsDetail	curr_time;		/* current epcsi time in local time */
    struct dbAddr	tcnt_dbAddr;
    float		tcnt = 0;
    short		answer;
    char		*p;
    char		pvname[PVNAME_SZ+FLDNAME_SZ+4];
    struct recLoc	*recLoc;
    struct recChain	*chain;
    struct biRecord	*pbi;
    int			rectype;
    unsigned short	monitor_mask;

    /* wait for the time to be set in the IOC */
    for (;;)
    {
	tsLocalTime (&last_etime);
	if (last_etime.secPastEpoch < 110000)
	    taskDelay (10 * vxTicksPerSecond);
	else
	    break;
    }

    /* find the record type for a BI */
    for (rectype = 1; rectype < 100; rectype++)
    {
	p = GET_PRECTYPE(rectype);
	if (strcmp (p, "bi") == 0)
	    break;
    }

    logMsg ("tcnt_reset is ready\n",0,0,0,0,0,0);

    for (;;)
    {
	if (new_mode > -1)
	{
	    reset_mode = new_mode;
	    new_mode = -1;
	}
	if (new_async)
	{
	    dbAddr = newdbAddr;
	    new_async = 0;
	}

	reset = 0;

	/* get the current epics time */
	tsLocalTime (&curr_etime);

	/* convert both epics times into local times */
	tsStampToLocal (last_etime, &last_time);
	tsStampToLocal (curr_etime, &curr_time);

	switch (reset_mode)
	{
	    case 0:		/* manual mode	*/
		status = 0;
		if ( (status = dbGetField (&dbAddr, DBR_ENUM, 
					   (caddr_t)&curr_async_reset,
					   &options, &nRequest, NULL)) != 0)
		{
		    logMsg ("tcnt_reset, failed to get asyn value\n",0,0,0,0,0,0);
		    break;
		}
		if ( (last_async_reset == 0) && curr_async_reset)
		    reset++;
		last_async_reset = curr_async_reset;
		break;

	    case 1:		/* daily reset	*/
		if (last_time.dayOfWeek != curr_time.dayOfWeek)
		    reset++;
		break;

	    case 2:		/* weekly reset */
		/* if yesterday was Saturday and today is Sunday */
		if ( (last_time.dayOfWeek == 6) && (curr_time.dayOfWeek == 0) )
		    reset++;
		break;

	    case 3:		/* monthly reset */
		if (last_time.monthNum != curr_time.monthNum)
		    reset++;
		break;

	    default:
		break;
	}
	last_time = curr_time;
	if (reset)
	{
	    if (!(recLoc = GET_PRECLOC(rectype)))
	    {
		logMsg("tcnt_reset, can not find BI\n",0,0,0,0,0,0);
		taskSuspend (taskIdSelf ());
	    }
	    for (chain = (struct recChain *) (recLoc -> pFirst);
		 chain;
		 chain = chain -> next)
	    {
		/* no ptr to record */
		if (!(pbi = (struct biRecord *) (chain -> pRecord)))
		    continue;

		/* no record name */
		if (!(pbi -> name[0]) )
		    continue;
		
		/* record not fully init'd yet */
		if ( (!(pbi -> init)) || (!(pbi -> ini2)) )
		    continue;

		/* if resource is not master, record not processing */
		resGetCaAnswer(pbi -> radr, &answer);
		if (!answer)
		    continue;

		/* if already 0 (zero) don't bother */
		if (!pbi -> tcnt)
		    continue;

		strncpy (pvname, pbi -> name, sizeof (pvname) );
		if ( (p = (char *)index (pvname, '.')) != NULL)
		    *p = NULL;
		strcat (pvname, ".TCNT");
		if ( (status = dbNameToAddr (pvname, &tcnt_dbAddr)) != 0)
		{
		    logMsg("tcnt_reset, dbNameToAddr fail, %s\n",(int)pvname,0,0,0,0,0);
		    continue;
		}
		if ( (status = dbPutField (&tcnt_dbAddr, DBR_FLOAT,
					   (caddr_t)&tcnt, nRequest)) != 0)
		{
		    logMsg("tcnt_reset, dbPutField fail, %s\n",(int)pvname,0,0,0,0,0);
		    continue;
		}

		monitor_mask = DBE_VALUE;
		/* if trending is enabled */
		if (pbi -> tenb)
		    monitor_mask |= DBE_LOG;
		db_post_events ((struct dbCommon *)pbi, (caddr_t)&pbi -> tcnt, monitor_mask);

		/* throttle to 20/second */
		taskDelay(3);
	    }
	}

	/* wait 1 minute before trying again */
	taskDelay (60 * vxTicksPerSecond);
    }
}


/****************************************************************************
 *
 *  MODULE NAME            : update_tcnt_reset
 *  AUTHOR                 : emsteel
 *  DATE                   : 22 Feb 95
 *
 *          ARGUMENT(S)    : mode - type of reset
 *			     ptname - ptname if mode is 0
 *
 *  DESCRIPTION: change the mode/manual pvname for tcnt_reset task
 *
****************************************************************************/

int update_tcnt_reset (int mode, char *ptname)
{
    char pvname[PVNAME_SZ + FLDNAME_SZ + 4];
    struct dbAddr	chgdbAddr;
    long		status = 0;
    int			reset_name = 0;
    
    if ( (mode < 0) || (mode > 3) )
    {
	logMsg ("update_tcnt_reset, mode range is 0 - 3 (%d)\n", mode,0,0,0,0,0);
	return (1);
    }
    if (mode == 0)
    {
	if (ptname[0] == NULL)
	{
	    logMsg ("update_tcnt_reset, mode 0 requires a pointname\n",0,0,0,0,0,0);
	    return (1);
	}
	strncpy (pvname, ptname, sizeof (pvname) );
	if ( (status = dbNameToAddr (pvname, &chgdbAddr) ) != 0)
	{
	    logMsg ("update_tcnt_reset, %s is not in IOC\n",(int)pvname,0,0,0,0,0);
	    return (1);
	}
	reset_name++;
    }
    new_mode = mode;
    if (reset_name)
    {
	strncpy (manual_reset_name, pvname, sizeof (manual_reset_name) );
	newdbAddr = chgdbAddr;
	new_async = 1;
    }

    return (0);
}

