/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

#include <stdio.h>
#include <search.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <socket.h>
#include <in.h>
#include <tcp.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/errno.h>
#include <string.h>
#ifdef SOLARIS
#include <sys/filio.h>
#include <file.h>
#endif
#include <d4all.h>
#include <redCbDefs.h>

/* Various local data */

static	int 	slaveFd, slaveConnected, connectPending;
static	struct	sockaddr_in	portAddr; 
static	char	slaveName [100], myName [100];
static	long	nextConnectTryTime;
static	C4CODE	*myCodeBase;
static	int	appendInProgress;

/* This is a list of all the open files.  It is used to send the slave the list of */
/* open files when it come alive. */

static	MasterFileDesc	openFiles [MAX_OPEN_FILES];
static	int		numOpenFiles;

/* Pointer to handle translation table */

#define	MAX_HANDLES	1000
static	void	*pTable [MAX_HANDLES];
static	size_t	numHandles = 0;

static	int	myPID;
static	int	redundActive;
static	int	inhibited;
	int	noInetd = 0;

/* Functions defined in this module */

	int	red_d4append (D4DATA *);
	int	red_d4append_blank (D4DATA *);
	int	red_d4append_start (D4DATA *, int);
	void	red_d4blank (D4DATA *);
	int	red_d4bottom (D4DATA *);
	int	red_d4close (D4DATA *);
	void	red_d4close_all (C4CODE *);
	D4DATA	*d4create (C4CODE *, char *, F4FIELD_INFO *, T4TAG_INFO *);
	void	red_d4delete (D4DATA *);
	F4FIELD	*red_d4field (D4DATA *, char *);
	int	red_d4flush_all (D4DATA *);
	int	red_d4go (D4DATA *, long);
	void	red_d4init (C4CODE *);
	int	red_d4lock_file (D4DATA *);
	D4DATA	*red_d4open (C4CODE *, char *);
	int	red_d4pack (D4DATA *);
	int	red_d4seek (D4DATA *, char *);
	int	red_d4skip (D4DATA *, long);
	void	red_d4tag_select (D4DATA *, T4TAG *);
	int	red_d4top (D4DATA *);
	int	red_d4unlock_all (D4DATA *);
	void	red_e4exit_test (C4CODE *);
	void	red_f4assign (F4FIELD *, char *);
	void	red_f4assign_double (F4FIELD *, double);
	void	red_f4assign_int (F4FIELD *, int);
	void	red_f4assign_long (F4FIELD *, long);
	void	red_f4assign_n (F4FIELD *, char *, int);
	void	red_RedundancyOn (void);
	void	red_RedundancyOff (void);
static	int	redundancyActive (D4DATA *);

	void	red_Terminate (void);
static	int	getSlaveName (void);
static	void	tryToCreateSlave (void);
static	void	sigio_func ();
static	int	sendFileListToSlave (void);
static	void	copyFileToSlave ();
static	void	makeSendFileMsg (int i, FILE_DESC_MSG *);
static	int	fileHasMDX (char *);
static	int	fileHasDBT (char *);
static	void	sendFiles (char *, int, int);
static	void	sendFile (char *);
static	void	sendHandle (int, D4DATA *);
static	void	sendRequest(uint32, char *, uint32);
static	void	sendRequestAndData (uint32, char *, uint32, char *, uint32);
static	void	sendBytes(char *, uint32);
static	int	getSendRequest (void);
static	int	getBytes (char *, int);
static	void	insertFile (char *, D4DATA *, uint32);
static	void	removeFile (D4DATA *);
static	void	insertField (D4DATA *, char *, F4FIELD *, uint32);
static	F4FIELD	*fieldOpen (D4DATA *, char *);
static	void	setTag (D4DATA *, T4TAG *);
static	int	findFile (D4DATA *);
static	void	setRecNumbers (void);
static	void	trimCpy (char *, char *);
static	void	strlwr (char *);
static	void	sigPipeHandler (int);
	void	inhibitRedundancy (void);
	void	allowRedundancy (void);
static	uint32	makeDataHandle (D4DATA *);
static	uint32	makeFieldHandle (F4FIELD *);
static	uint32	makeHandle (void *);
static	int	searchCompare (const void *, const void *);
static	int	sortCompare (const void *, const void *);

extern	int	Trace, Debug, Verbose, Redundancy;
extern	int	errno;
extern	char	*sys_errlist [];

/*----------*/
int red_d4append (data)

    D4DATA	*data;

    {

	int	result;

	if (Trace)
	    printf ("redCbLib(%d): Got d4append(%x)\n", myPID, data);

	tryToCreateSlave ();

	if (redundancyActive (data)) {
	    sendHandle (D4APPEND, data);
	}

	result = d4append (data);
	appendInProgress = 0;
	openFiles [findFile (data)].appendInProgress = 0;
	return result;

    }

/*----------*/
int red_d4append_blank (data)

    D4DATA	*data;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got d4append_blank(%x)\n", myPID, data);

	tryToCreateSlave ();

	if (redundancyActive (data)) {
	    sendHandle (D4APPEND_BLANK, data);
	}

	return d4append_blank (data);

    }

/*----------*/
int red_d4append_start (data, useMemoEntries)

    D4DATA	*data;
    int32	useMemoEntries;

    {

	D4APPEND_START_MSG	request;

	if (Trace)
	    printf ("redCbLib(%d): Got d4append_start(%x,%d)\n",
			myPID, data, useMemoEntries);

	tryToCreateSlave ();

	appendInProgress = 1;
	openFiles [findFile (data)].appendInProgress = 1;

	if (redundancyActive (data)) {
	    request.data = htonl (makeDataHandle (data));
	    request.useMemoEntries = htonl (useMemoEntries);
	    sendRequest(D4APPEND_START, (char *) &request, sizeof(request));
	}

	return d4append_start (data, useMemoEntries);

    }

/*----------*/
void red_d4blank (data)

    D4DATA	*data;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got d4blank(%x)\n", myPID, data);

	tryToCreateSlave ();

	if (redundancyActive (data)) {
	    sendHandle (D4BLANK, data);
	}

	d4blank (data);

    }

/*----------*/
int red_d4bottom (data)

    D4DATA	*data;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got d4bottom(%x)\n", myPID, data);

	tryToCreateSlave ();

	if (redundancyActive (data))
	    sendHandle (D4BOTTOM, data);

	return d4bottom (data);   

    }

/*----------*/
int red_d4close (data)

    D4DATA	*data;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got d4close(%x)\n", myPID, data);

	tryToCreateSlave ();

	if (redundancyActive (data)) {
	    sendHandle (D4CLOSE, data);
	}

	removeFile (data);
	return d4close (data);

    }

/*----------*/
void red_d4close_all (code_base)

    C4CODE	*code_base;

    {

	MsgHeader	msg;

	if (Trace)
	    printf ("redCbLib(%d): Got d4close_all\n", myPID);

	tryToCreateSlave ();

	sendRequest (D4CLOSE_ALL, (char *)&msg, sizeof (msg));

	numOpenFiles = 0;
	d4close_all (code_base);

    }

/*----------*/
D4DATA *red_d4create (code_base, name, field_info, tag_info)

    C4CODE	 *code_base;
    char	 *name;
    F4FIELD_INFO *field_info;
    T4TAG_INFO	 *tag_info;

    {

	F4FIELD_INFO	*fieldList;
	T4TAG_INFO	*tagList;
	D4CREATE_MSG	request;
	F4FIELD_MSG	fieldReq;
	T4TAG_MSG	tagReq;
	D4DATA		*data;
	uint32		handle;
	uint32		numFields, numTags;

	if (Trace)
	    printf ("redCbLib(%d): Got d4create(\"%s\")\n", myPID, name);

	tryToCreateSlave ();

	data = d4create (code_base, name, field_info, tag_info);
	if (data == NULL)
	    return NULL;

	handle = makeDataHandle (data);
	insertFile (name, data, handle);

	if (redundancyActive (data)) {

	    if (slaveConnected) {
	
		/* Count the number of fields */

		fieldList = field_info;
		numFields = 0;
		while (fieldList -> name) {
		    fieldList++;
		    numFields++;
		}
		request.numFields = htonl (numFields);

		/* Count the number of tags */
    
		tagList = tag_info;
		numTags = 0;
		if (tagList)
		    while (tagList->name) {
			tagList++;
			numTags++;
		    }
		request.numTags = htonl (numTags);

		/* Send the file name, number of fields, and number of tags */

		strcpy(request.name, name);
		request.data = htonl (makeDataHandle (data));
		sendRequest(D4CREATE, (char *) &request, sizeof(request));

		/* Send the field descriptors */

		while (numFields--) {
		    strcpy(fieldReq.name, field_info->name);
		    fieldReq.type = htonl (field_info->type);
		    fieldReq.len = htonl (field_info->len);
		    fieldReq.dec = htonl (field_info->dec);
		    sendRequest (FIELD_DESC, (char *) &fieldReq, sizeof(fieldReq));
		    field_info++;
		}

		/* Send the tags */

		while (numTags--) {
		    strcpy(tagReq.name, tag_info->name);
		    strcpy(tagReq.expression, tag_info->expression);
		    if (tag_info->filter)
			strcpy (tagReq.filter, tag_info->filter);
		    else
			strcpy (tagReq.filter, "");
		    tagReq.unique = htonl (tag_info->unique);
		    tagReq.descending = htonl (tag_info->descending);
		    sendRequest (TAG_DESC, (char *) &tagReq, sizeof (tagReq));
		    tag_info++;
		}
	    }

	}

	if (Trace)
	    printf ("redCbLib(%d): d4create(\"%s\") returns %x\n", myPID, name, data);

	return data;

    }

/*----------*/
void red_d4delete (data)

    D4DATA	*data;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got d4delete(%x)\n", myPID, data);

	tryToCreateSlave ();

	if (redundancyActive (data))
	    sendHandle (D4DELETE, data);

	d4delete (data);

    }

/*----------*/
F4FIELD *red_d4field (data, name)

    D4DATA	*data;
    char	*name;

    {

	F4FIELD		*field;
	int		i, j;
	uint32		fieldHandle;

	if (Trace)
	    printf ("redCbLib(%d): Got d4field (%x,\"%s\")\n", myPID, data, name);

	tryToCreateSlave ();

	field = fieldOpen (data, name);
	if (field == NULL) {
	    field = d4field (data, name);
	    fieldHandle = makeFieldHandle (field);
	    insertField (data, name, field, fieldHandle);
	}

	if (redundancyActive (data)) {
	    D4FIELD_MSG	request;
	    request.data = htonl (makeDataHandle (data));
	    request.masterHandle = htonl (fieldHandle);
	    sendRequestAndData(D4FIELD, (char *) &request, sizeof(request),
				name, strlen(name) + 1);
	}

	if (Trace)
	    printf ("redCbLib(%d): d4field ()\n", myPID);

	return field;

    }

/*----------*/
int red_d4flush_all (data)

    D4DATA	*data;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got d4flush_all(%x)\n", myPID, data);

	tryToCreateSlave ();

	if (redundancyActive (data)) {
	    sendHandle (D4FLUSH_ALL, data);
	}

	return d4flush_all (data);

    }

/*---------*/
int red_d4go (data, record_number)

    D4DATA	*data;
    long	record_number;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got d4go(%x,%ld)\n", myPID, data, record_number);

	tryToCreateSlave ();

	if (redundancyActive (data)) {
	    D4GO_MSG request;
	    request.data = htonl (makeDataHandle (data));
	    request.record_number = htonl ((uint32)record_number);
	    sendRequest(D4GO, (char *) &request, sizeof(request));
	}

	return d4go (data, record_number);

    }

/*----------*/
void red_d4init (code_base)

    C4CODE	*code_base;

    {

	myPID = getpid ();

	if (Trace)
	    printf ("redCbLib(%d): Got d4init\n", myPID);

	signal (SIGPIPE, sigPipeHandler);

	d4init (code_base);
	myCodeBase = code_base;
	numOpenFiles = 0;

	/* Get name of slave so we can connect to it later */

	if (Redundancy)
	    if (!getSlaveName ())
		exit (-1);

	/* Slave is not yet connected.  If not using redundancy, this will remain */
	/* zero thus permanently preventing any attempt to send messages to slave. */

	redundActive = 1;	/* Used to set the redundancy active bit for each file */
	slaveConnected = 0;
	connectPending = 0;
	nextConnectTryTime = time (0) - 1;	/* Force initial try */

	tryToCreateSlave ();


    }

/*----------*/
int red_d4lock_file (data)

    D4DATA	*data;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got lock_file(%x)\n", myPID, data);

	tryToCreateSlave ();

	if (redundancyActive (data))
	    sendHandle (D4LOCK_FILE, data);

	return d4lock_file (data);

    }

/*----------*/
D4DATA *red_d4open (code_base, name)

    C4CODE	*code_base;
    char	*name;

    {

	D4DATA	*data;
	uint32	handle;
	int	result;

	if (Trace)
	    printf ("redCbLib(%d): Got d4open (\"%s\")\n", myPID, name);

	tryToCreateSlave ();

	data = d4open (code_base, name);
	if (data == NULL)
	   return NULL;

	handle = makeDataHandle (data);
	insertFile (name, data, handle);

	if (redundancyActive (data)) {

	    /* If a slave was created successfully, then it may be copying the file from */
	    /* here to itself.  We have to wait for a message from the slave indicating */
	    /* that the file is finished being copied before we can continue. */

	    if (slaveConnected) {

		D4OPEN_MSG	request;
		int		copied;

		request.data = htonl (makeDataHandle (data));
		request.hasMDX = htonl (fileHasMDX (name));
		request.hasDBT = htonl (fileHasDBT (name));		
		sendRequestAndData (D4OPEN, (char *) &request, sizeof(request),
				name, strlen (name) + 1);

		result = getSendRequest ();
		if (result == -1) {
		    if (Verbose)
			printf ("redCbLib(%d): Slave server not available\n", myPID);		    
		} else if (result > 0) {
		    sendFiles (name, ntohl (request.hasMDX), ntohl (request.hasDBT));
		}
	    }
	}

	return data;

    }

/*----------*/
int red_d4pack (data)

    D4DATA	*data;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got d4pack\n", myPID);

	tryToCreateSlave ();

	if (redundancyActive (data))
	    sendHandle (D4PACK, data);

	return d4pack (data);
	
    }

/*----------*/
int red_d4seek (data, str)

    D4DATA	*data;
    char	*str;

    {

	D4SEEK_MSG request;

	if (Trace)
	    printf ("redCbLib(%d): Got d4seek(\"%s\")\n", myPID, str);

	tryToCreateSlave ();

	if (redundancyActive (data)) {
	    request.data = htonl (makeDataHandle (data));
	    sendRequestAndData(D4SEEK, (char *) &request, sizeof(request),
				str, strlen(str) + 1);
	}

	return d4seek (data, str);

    }

/*----------*/
int red_d4skip (data, num_records)

    D4DATA	*data;
    long	num_records;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got d4skip(%ld)\n", myPID, num_records);

	tryToCreateSlave ();

	if (redundancyActive (data)) {
	    D4SKIP_MSG request;
	    request.data = htonl (makeDataHandle (data));
	    request.num_records = htonl ((uint32)num_records);
	    sendRequest(D4SKIP, (char *) &request, sizeof(request));
	}

	return d4skip (data, num_records);

    }

/*----------*/
void red_d4tag_select (data, tag)

    D4DATA	*data;
    T4TAG	*tag;

    {

	uint32	tagHandle;

	if (Trace)
	    printf ("redCbLib(%d): Got d4tag_select(%x,%x)\n", myPID, data, tag);

	tryToCreateSlave ();

	if (redundancyActive (data)) {
	    D4TAG_SELECT_MSG request;
	    if (tag)
		strcpy (request.tagName, tag->alias);
	    request.data = htonl (makeDataHandle (data));
	    request.tag = htonl (makeDataHandle ((D4DATA *)tag));
	    sendRequest(D4TAG_SELECT, (char *) &request, sizeof(request));
	}

	if (tag)
	    setTag (data, tag);

	d4tag_select (data, tag);

}

/*----------*/
int red_d4top (data)

    D4DATA	*data;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got d4top(%x)\n", myPID, data);

	tryToCreateSlave ();

	if (redundancyActive (data))
	    sendHandle (D4TOP, data);

	return d4top (data);

    }

/*----------*/
int red_d4unlock_all (data)

    D4DATA	*data;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got d4unlock_all(%x)\n", myPID, data);

	tryToCreateSlave ();

	if (redundancyActive (data))
	    sendHandle (D4UNLOCK_ALL, data);

	return d4unlock_all (data);

    }

/*----------*/
void red_f4assign (field, str)

    F4FIELD	*field;
    char	*str;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got f4assign(%x,\"%s\")\n", myPID, field, str);

	tryToCreateSlave ();

	if (redundancyActive (field->data)) {
	    F4ASSIGN_MSG request;
	    request.field = htonl (makeFieldHandle (field));
	    sendRequestAndData(F4ASSIGN, (char *) &request, sizeof(request),
				str, strlen(str) + 1);
	}

	f4assign (field, str);
	
}

/*----------*/
void red_f4assign_double (field, value)

    F4FIELD	*field;
    double	value;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got f4assign_double(%x,%10.2lf)\n", myPID, field, value);

	tryToCreateSlave ();

	if (redundancyActive (field->data)) {
	    F4ASSIGN_DOUBLE_MSG request;
	    request.field = htonl (makeFieldHandle (field));
	    sprintf (request.value , "%f", value);
	    sendRequest(F4ASSIGN_DOUBLE, (char *) &request, sizeof(request));
	}

	f4assign_double (field, value);

}

/*----------*/
void red_f4assign_int (field, value)

    F4FIELD	*field;
    int		value;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got f4assign_int(%x,%d)\n", myPID, field, value);

	tryToCreateSlave ();

	if (redundancyActive (field->data)) {
	    F4ASSIGN_INT_MSG request;
	    request.field = htonl (makeFieldHandle (field));
	    request.value = htonl ((uint32)value);
	    sendRequest(F4ASSIGN_INT, (char *) &request, sizeof(request));
	}

	f4assign_int (field, value);

    }

/*----------*/
void red_f4assign_long (field, value)

    F4FIELD	*field;
    long	value;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got f4assign_long(%x,%ld)\n", myPID, field, value);

	tryToCreateSlave ();

	if (redundancyActive (field->data)) {
	    F4ASSIGN_LONG_MSG request;
	    request.field = htonl (makeFieldHandle (field));
	    request.value = htonl ((uint32)value);
	    sendRequest(F4ASSIGN_LONG, (char *) &request, sizeof(request));
	}

	f4assign_long (field, value);

}

/*----------*/
void red_f4assign_n (field, str, str_len)

    F4FIELD	*field;
    char	*str;
    int		str_len;

    {

	if (Trace)
	    printf ("redCbLib(%d): Got f4assign_n(%x,\"%s\",%d)\n", myPID, field, str, str_len);

	tryToCreateSlave ();

	if (redundancyActive (field->data)) {
	    F4ASSIGN_N_MSG request;
	    request.field = htonl (makeFieldHandle (field));
	    request.stringLength = htonl (str_len);
	    sendRequestAndData(F4ASSIGN_N, (char *) &request, sizeof(request), str, str_len);
	}

	f4assign_n (field, str, str_len);

}

/*----------*/
void red_RedundancyOn (void)	/* Turn on redundancy */

    {

	redundActive = 1;

    }

/*----------*/
void red_RedundancyOff (void)	/* Turn off redundancy */

    {

	redundActive = 0;

    }

/*----------*/
static int redundancyActive (handle)	/* Check if redundancy is on for a file */

    D4DATA		*handle;

    {
	int	i;

	if (handle == NULL)
	    return 0;

	i = findFile (handle);
	if (i == -1) {
	    printf ("redCbLib(%d): Failure looking up handle %x\n", myPID, handle);
	    exit (1);
	}

	return openFiles [i].redundActive;

    }

/*----------*/
void red_Terminate (void)	/* Perform an orderly shutdown */

    {

	close (slaveFd);

    }

/*----------*/
static int getSlaveName (void)

    {

	D4DATA	*data;
	F4FIELD	*name1, *name2;
	char	*p;

	gethostname(myName, sizeof(myName));
	p = strchr (myName, '.');
	if (p)
	    *p = '\0';
	if (data = d4open(myCodeBase, "/tis4000db/reslist")) {
	    d4top(data);
	    name1 = d4field(data, "PRIMARYIOC");
	    name2 = d4field(data, "SECNDRYIOC");
	    while (!d4eof(data)) {
		trimCpy(slaveName, f4str(name1));
		if (!strcmp(myName, slaveName)) {
		    trimCpy(slaveName, f4str(name2));
		    break;
		}
		trimCpy(slaveName, f4str(name2));
		if (!strcmp(myName, slaveName)) {
		    trimCpy(slaveName, f4str(name1));
		    break;
		}
		strcpy (slaveName, "");
		d4skip(data, (long) 1);
	    }
	    d4close(data);
	}

	if (strlen (slaveName) > 0) {
	    if (Verbose)
		printf ("redCbLib(%d): I am %s, my slave is %s\n", myPID, myName, slaveName);
	    return 1;
	} else {
	    printf ("redCbLib(%d): Ubnable to find my name(%s) in reslist\n", myPID, myName);
	    return 0;
	}
    }

/*----------*/
static void tryToCreateSlave (void)		/* Try to create a slave */ 

    {

	struct	servent		*sp;
	struct	hostent		*hp;
	int			partnerInet;
	int			sendBufferSize = SEND_BLOCK_SIZE;
	int			true = 1, false = 0, result;
	fd_set			fdSet;
	struct timeval		timeOut;

	/* We never try to create a slave while in the process or appending a */
	/* new record. */

	if (appendInProgress)
	    return;

	/* Check if already connected */

	if (!Redundancy || slaveConnected)
	    return;

	/* If time to retry connect, do it. */

	if (time (0) >= nextConnectTryTime) {

	    if (Verbose)
		printf ("redCbLib(%d): Attempting to create slave on %s\n", myPID, slaveName);

	    if (connectPending)
		close (slaveFd);

	    /* Initialize a socket for communication with slave.  It is */
	    /* non-blocking so that we won't hang around waiting for a connect. */

	    slaveFd = socket(PF_INET, SOCK_STREAM, 0);
	    setsockopt (slaveFd, IPPROTO_TCP, TCP_NODELAY, (char *)&true, sizeof(true));
	    setsockopt (slaveFd, SOL_SOCKET, SO_SNDBUF, (char *)&sendBufferSize, sizeof (sendBufferSize));
	    setsockopt (slaveFd, SOL_SOCKET, SO_KEEPALIVE, (char *)&true, sizeof (true));
	    fcntl (slaveFd, FNDELAY, 1);
	    ioctl (slaveFd, FIONBIO, &true);

	    /* Build an address for connection to slave */

	    memset ((char *)&portAddr, 0, sizeof (portAddr));
	    portAddr.sin_family = AF_INET;
	    hp = gethostbyname (slaveName);
	    partnerInet = ntohl (**((long **)(hp -> h_addr_list)));
	    portAddr.sin_addr.s_addr = htonl (partnerInet);

	    /* "redcb" is the name of the redundant slave service as registered with inetd */

	    if (noInetd) {
		portAddr.sin_port = htons (REDCBPORT);
	    } else {
		if ((sp = getservbyname ("redcb", "tcp")) == NULL) {
		    printf ("redCbLib(%d): Unable to find service redcb/tcp\n", myPID);
		    exit (1);
		} else
		    portAddr.sin_port = sp->s_port;
	    }

	    result = connect(slaveFd, (struct sockaddr *) &portAddr, sizeof(portAddr));
	    if (result == -1) {
		if (errno == EINPROGRESS) {
		    if (Verbose)
			printf ("redCbLib(%d): Connection to %s pending\n", myPID, slaveName);
		    connectPending = 1;
		} else
		    if (Verbose)
			printf ("redCbLib(%d): Connect attempt failed, %s\n", myPID, sys_errlist [errno]);
	    } else {
		if (Verbose)
		    printf ("redCbLib(%d): Connected to slave on %s\n", myPID, slaveName);
		slaveConnected = 1;
		if (!sendFileListToSlave ())
		    slaveConnected = 0;
		connectPending = 0;
	    }

	    nextConnectTryTime = time (0) + PARTNER_CONNECT_INTERVAL;
	}

	/* If a connect is pending, check for connection complete. */

	if (connectPending) {
	    FD_ZERO (&fdSet);
	    FD_SET (slaveFd, &fdSet);
	    timeOut.tv_sec = 0;
	    timeOut.tv_usec = 0;
	    result = select(slaveFd + 1, NULL, &fdSet, NULL, &timeOut);
	    if (result > 0) {
		char buf [2];
		result = read (slaveFd, buf, 1);
		if ((result == -1) && (errno == ECONNREFUSED)) {
		    if (Verbose)
			printf ("redCbLib(%d): Connection to %s refused\n", myPID, slaveName);
		    connectPending = 0;
		    nextConnectTryTime = time (0) + PARTNER_CONNECT_INTERVAL;
		    return;
		}
		if ((errno == 0) || (errno == EWOULDBLOCK)) {
		    if (Verbose)
			printf ("redCbLib(%d): Connected to %s\n", myPID, slaveName);
		    connectPending = 0;
		    slaveConnected = 1;
		    if (!sendFileListToSlave ()) {
			slaveConnected = 0;
		    }	
		}
	    }
	}
    }

/*----------*/
static int sendFileListToSlave (void)  /* Send copies of open files to slave */

    {

	FILE			*scriptFile;
	MASTER_LIST_MSG		masterList;
	FILE_DESC_MSG		fileRequest;
	int			i, numCopied, result;

	/* Slave has just come on line. Pass through the list of files. During */
	/* this pass, do a d4flush_all() to get all pending record updates written */
	/* to the disk, a d4update_header() to update the header, and call */
	/* d4unlock_all() to make sure that we hold no locks on the file. */

	/* Flush, update, and unlock all files */

	for (i=0 ; i<numOpenFiles ; i++) {
	    d4flush_all (openFiles [i].masterHandle);
	    d4unlock_all (openFiles [i].masterHandle);
	}

	/* Set the current record number in each open file */

	setRecNumbers ();

	/* Send the file descriptor for each file */

	masterList.numFiles = htonl (numOpenFiles);
	strcpy (masterList.masterName, myName);
	sendRequest (MASTER_LIST, (char *) &masterList, sizeof(masterList));
	numCopied = 0;
	for (i=0 ; i<numOpenFiles ; i++) {
	    openFiles [i].hasMDX = fileHasMDX (openFiles [i].fileName);
	    openFiles [i].hasDBT = fileHasDBT (openFiles [i].fileName);
	    makeSendFileMsg (i, &fileRequest);
	    if (Verbose)
		/**/printf ("redCbLib(%d): Sending query for %s\n", myPID, openFiles [i].fileName);
	    sendRequest (SEND_FILE, (char *)&fileRequest, sizeof (fileRequest));
	    result = getSendRequest ();
	    if (Verbose)
		/**/printf ("redCbLib(%d): Slave server responded with %d\n", myPID, result);
	    if (result == -1) {
		if (Verbose)
		    printf ("redCbLib(%d): Slave server not available\n", myPID);
		close (slaveFd);
		slaveConnected = 0; 
		return 0;
	    }
	    if (result == 1)
		sendFiles (openFiles [i].fileName, openFiles [i].hasMDX,
				openFiles [i].hasDBT);
	    numCopied++;
	}

	return 1;

    }

/*----------*/
static void makeSendFileMsg (int i, FILE_DESC_MSG *msg)

    {

	int	j;

	for (j=0 ; j<openFiles [i].numFields ; j++) {
	    strcpy (msg->fieldName [j], openFiles [i].fieldName [j]);
	    msg->fieldHandles [j] = htonl (openFiles [i].slaveFields [j]);
	}
	strcpy (msg->fileName, openFiles [i].fileName);
	msg->fileHandle = htonl (openFiles [i].slaveHandle);
	msg->numFields = htonl (openFiles [i].numFields);
	msg->currentRecord = htonl (openFiles [i].currentRecord);
	msg->appendInProgress = htonl (openFiles [i].appendInProgress);
	msg->hasMDX = htonl (openFiles [i].hasMDX);
	msg->hasDBT = htonl (openFiles [i].hasDBT);

    }

/*----------*/
static int fileHasMDX (fileName)	/* Check if a file has an associated index file */

    char	*fileName;

    {

	char	rawFileName [200], *p, fName [200];

	/* Strip the extension from the filename */

	strcpy (rawFileName, fileName);
	p = strrchr (rawFileName, '.');
	if (p)
	    *p = '\0';
	
	/* Check if an index file exists */

	sprintf (fName, "%s.mdx", rawFileName);
	return access (fName, 0) == 0;
		    
    }

/*----------*/
static int fileHasDBT (fileName)	/* Check if a file has an associated memo file */

    char	*fileName;

    {

	char	rawFileName [200], *p, fName [200];

	/* Strip the extension from the filename */

	strcpy (rawFileName, fileName);
	p = strrchr (rawFileName, '.');
	if (p)
	    *p = '\0';
	
	/* Check if a memo file exists */

	sprintf (fName, "%s.dbt", rawFileName);
	return access (fName, 0) == 0;
		    
    }

/*----------*/
static void sendFiles (fileName, hasMDX, hasDBT)	/* Send DBF, MDX, DBT files to slave */

    char	*fileName;
    int		hasMDX, hasDBT;

    {

	char	rawFileName [200], *p, fName [200];

	/* Send the data file */

	sendFile (fileName);

	/* Strip the extension from the filename */

	strcpy (rawFileName, fileName);
	p = strrchr (rawFileName, '.');
	if (p)
	    *p = '\0';
	
	/* If an index file exists, send it */

	if (hasMDX) {
	    sprintf (fName, "%s.mdx", rawFileName);
	    sendFile (fName);
	}

	/* If a memo file exists, send it */

	if (hasDBT) {
	    sprintf (fName, "%s.dbt", rawFileName);
	    sendFile (fName);
	}
		    
    }

/*----------*/
static void sendFile (fileName)		/* Send a single file prefixed with a header for each block */

    char	*fileName;

    {

	int		fd, bytesLeft, bytesRead;
	struct stat	statbuf;
	FILE_SEND_MSG	fileMsg;

	fd = open (fileName, O_RDONLY);

	fstat (fd, &statbuf);
	bytesLeft = statbuf.st_size;

	/* Send the file */

	if (Verbose)
	    printf ("redCbLib(%d): File %s sending to the slave\n", myPID, fileName);

	while (bytesLeft > 0) {
	    bytesRead = read (fd, fileMsg.block, SEND_BLOCK_SIZE);
	    bytesLeft -= bytesRead;
	    fileMsg.bytesLeft = htonl (bytesLeft);
	    fileMsg.blockSize = htonl (bytesRead);
	    sendRequest (SEND_FILE, (char *)&fileMsg, sizeof (fileMsg));
	}

	close (fd);
  
	if (Verbose)
	    printf ("redCbLib(%d): File %s was sent to the slave\n", myPID, fileName);

    }

/*----------*/
static void sendHandle (msgType, masterHandle)

    int		msgType;
    D4DATA	*masterHandle;

    {

	HANDLE_ONLY_MSG	msg;

	msg.data = htonl (makeDataHandle (masterHandle));
	sendRequest (msgType, (char *)&msg, sizeof (msg));

    }	

/*----------*/
static void sendRequest (msgType, request, requestSize)

    uint32	msgType;
    char	*request;
    uint32	requestSize;

    {

	MsgHeader	*header;

	if (!slaveConnected || inhibited)
	    return;

	header = (MsgHeader *)request;

	/* Optimization: Copy message header and request to a single buffer */
	/* and send it out in one call to sendBytes */

	header->msgMagic = htonl (MAGIC_NUMBER);
	header->msgSize = htonl (requestSize);
	header->msgType = htonl (msgType);
	sendBytes ((char *)request, requestSize);

	if (Debug)
	    printf ("redCbLib(%d): sendRequest: msgType = %x(%d)\n", myPID, msgType, msgType & 0xff);

    }

/*----------*/
static void sendRequestAndData (msgType, request, requestSize, data, dataSize)

    uint32	msgType;
    char	*request, *data;
    uint32	requestSize, dataSize;

    {

	char		outBuf [1000];
	MsgHeader	*header;

	if (!slaveConnected || inhibited)
	    return;

	/* Optimization: Copy message type, request, and data to a single buffer */
	/* and send it out in one call to sendBytes */

	header = (MsgHeader *)request;
	header->msgMagic = htonl (MAGIC_NUMBER);
	header->msgSize = htonl (requestSize + dataSize);
	header->msgType = htonl (msgType);
	memcpy (outBuf, request, requestSize);
	memcpy (&outBuf [requestSize], data, dataSize);

	sendBytes (outBuf, requestSize + dataSize);

	if (Debug)
	    printf ("redCbLib(%d): sendRequestAndData: msgType = %x(%d)\n", myPID, msgType, msgType);

    }

/*----------*/
static void sendBytes (buff, nbytes)

    char	*buff;
    uint32	nbytes;

    {

	int		result, bytesSent, bytesLeft, i;
	fd_set		writeSet, exceptSet;
	struct timeval	timeOut;
	char		*bufPtr;

	if (!slaveConnected)
	    return;

	if (Debug)
	    printf ("redCbLib(%d): Sending %d bytes on fd=%d\n", myPID, nbytes, slaveFd);

	timeOut.tv_sec = 5;
	timeOut.tv_usec = 0;

	bytesLeft = nbytes;
	bufPtr = buff;
	while (bytesLeft > 0) {

	    FD_ZERO (&writeSet);
	    FD_SET (slaveFd, &writeSet);
	    FD_ZERO (&exceptSet);
	    FD_SET (slaveFd, &exceptSet);

	    result = select (slaveFd + 1, NULL, &writeSet, &exceptSet, &timeOut);
	    if ((result == -1) && (errno == EPIPE)) {
		if (Verbose)
		    printf (
			"redCbLib(%d): Slave disconnected\n", myPID);
		slaveConnected = 0;
		close (slaveFd);
		return;
	    }
	    if (result >= 1) {
		if (FD_ISSET (slaveFd, &writeSet)) {
		    bytesSent = send (slaveFd, bufPtr, bytesLeft, 0);
/*printf ("sending %d bytes, result = %d, errno = %d (%s)\n", bytesLeft, bytesSent, errno, sys_errlist [errno]);
for (i=0 ; i<bytesLeft ; i++)
    printf ("%x ", (int)(*(bufPtr + i) & 0xff));
printf ("\n");
		    /* For some strange reason SOLARIS on X86 returns a -1 with errno == EINVAL */
		    /* even though the send succeeded.  All I can do is detect this and assume */
		    /* everything was sent. This makes me very nervous. */

		    if ((bytesSent == -1) && (errno != EINVAL))
			continue;
		    else
			bytesSent = bytesLeft;

		    bufPtr += bytesSent;
		    bytesLeft -= bytesSent;
		} else if (FD_SET (slaveFd, &exceptSet)) {
		    slaveConnected = 0;
		    close (slaveFd);
		    nextConnectTryTime = time (0) + PARTNER_CONNECT_INTERVAL;
		    if (Verbose)
			printf ("redCbLib(%d): Slave disconnected\n", myPID);
		    return;
		}
	    } else {	/* Slave has disappeared */
		slaveConnected = 0;
		close (slaveFd);
		nextConnectTryTime = time (0) + PARTNER_CONNECT_INTERVAL;
		if (Verbose)
		    printf ("redCbLib(%d): Slave disconnected.\n", myPID);
		return;
	    }
	}

	if (Debug) {
	    printf ("sendBytes: ");
	    while (nbytes--)
		printf ("%02.2x ", ((int)(*buff++)) & 0xFF);
	    printf ("\n");
	}

}

/*----------*/
static int getSendRequest (void)	/* Wait for send request. Return true if send is desired */

    {

	SEND_REQUEST_MSG	msg;

	/* If a send request is in the buffer, eat it and return true. If not then synchronization */
	/* has been lost. Close the socket so that the server can restart */

	if (getBytes ((char *)&msg, sizeof (msg)) != sizeof (msg))
	    return -1;

	if (ntohl (msg.header.msgMagic) != MAGIC_NUMBER) {
	    if (Verbose)
		printf ("redCbLib(%d): Expected SEND_REQUEST from slave, received bad magic number.",
				myPID);
	    close (slaveFd);
	    slaveConnected = 0;
	    return -1;
	}

	if (ntohl (msg.header.msgType) != SEND_REQUEST) {
	    if (Verbose)
		printf ("redCbLib(%d): Expected SEND_REQUEST from slave, received %d", 
					myPID, ntohl (msg.header.msgType));
	    close (slaveFd);
	    slaveConnected = 0;
	    return -1;
	}

	return ntohl (msg.sendFiles);

    }

/*----------*/
static int getBytes (buff, nbytes)

    char	*buff;
    int		nbytes;

    {

	fd_set	singleFd;
	int	numRead, result, bytesLeft, i;
	struct timeval	timeOut;
	char	*bufPtr;

	timeOut.tv_sec = 300;
	timeOut.tv_usec = 0;

	bufPtr = buff;
	bytesLeft = nbytes;
	while (bytesLeft > 0) {

	    FD_ZERO(&singleFd);
	    FD_SET(slaveFd, &singleFd);

	    result = select (slaveFd + 1, &singleFd, NULL, NULL, &timeOut);

	    if (result <= 0) {
		if (Debug)
		    printf ("redCbLib(%d): Error getting message from slave -- slave disconnected\n", 
				myPID);
		close (slaveFd);
		slaveConnected = 0;
		return -1;
	    }

	    if (FD_ISSET (slaveFd, &singleFd)) {
		numRead = recv (slaveFd, bufPtr, bytesLeft, 0);

		if (numRead <= 0) {
		    if (Debug)
			printf ("redCbLib(%d): Error getting message from slave -- slave disconnected\n", 
				myPID);
		    close (slaveFd);
		    slaveConnected = 0;
		    return -1;
		}

		if (Debug) {
		    printf ("getBytes: ");
		    for (i=0 ; i<numRead ; i++)
			printf ("%02.2x ", ((int)(*(bufPtr+i))) & 0xFF);
		    printf ("\n");
		}

		bytesLeft -= numRead;
		bufPtr += numRead;
	    }
	}

	return nbytes;
    }

/*----------*/
static void insertFile (name, data, handle)	/* Insert file into open file list */

    char	*name;
    D4DATA	*data;
    uint32	handle;

    {

	char	fileName [100], *p;

	strcpy (fileName, name);

	if (numOpenFiles == MAX_OPEN_FILES)
	    printf ("redCbLib(%d): Too many open files", myPID);

	/* If the file does not have a .dbf extension, place one on the file */

	p = strrchr (fileName, '.');
	if (p) {
	    if (strcasecmp (p, ".dbf") != 0) {
		strcat (fileName, ".dbf");
	    }
	} else
	    strcat (fileName, ".dbf");
		
	strcpy (openFiles [numOpenFiles].fileName, fileName);
	strlwr (openFiles [numOpenFiles].fileName);
	openFiles [numOpenFiles].masterHandle = data;
	openFiles [numOpenFiles].slaveHandle = handle;
	openFiles [numOpenFiles].numFields = 0;
	strcpy (openFiles [numOpenFiles].currentTag, "");
	openFiles [numOpenFiles].redundActive = redundActive;
	numOpenFiles++;

    }

/*----------*/
static void removeFile (data)	/* Remove a file from the open file list */

    D4DATA	*data;

    {

	int	i;

	i = findFile (data);

	if (i != -1) {
	    while (i < (numOpenFiles - 1)) {
		openFiles [i] = openFiles [i+1];
		i++;
	    }
	    numOpenFiles--;
	}

    }

/*----------*/
static void insertField (data, name, field, fieldHandle)	/* Insert a field into a file's field list */

    D4DATA	*data;
    char	*name;
    F4FIELD	*field;
    uint32	fieldHandle;

   {

	int	i;

	i = findFile (data);

	if (i != -1) {
	    strcpy (openFiles [i].fieldName [openFiles [i].numFields], name);
	    openFiles [i].masterFields [openFiles [i].numFields] = field;
	    openFiles [i].slaveFields [openFiles [i].numFields] = fieldHandle;
	    openFiles [i].numFields++;
	} else {
	    printf ("redCbLib(%d): File is not open for opening of field: %s\n", myPID, field);
	    exit (1);
	}
    	    
   }

/*----------*/
static F4FIELD *fieldOpen (data, name)		/* Check if a field is already open */

    D4DATA	*data;
    char	*name;

    {

	int	i, j;

	i = findFile (data);
	if (i != -1) {
	    for (j=0 ; j<openFiles [i].numFields ; j++) {
		if (strcmp (name, openFiles [i].fieldName [j]) == 0)
		    return openFiles [i].masterFields [j];
	    }
	    return 0;
	} else {
	    printf ("redCbLib(%d): Attempt to open field for an unopened file, data=%x\n", myPID, data);
	    exit (1);
	}

    }

/*----------*/
static void setTag (data, tag)

    D4DATA	*data;
    T4TAG	*tag;

    {

	int	i;

	for (i=0 ; i<numOpenFiles ; i++)
	    if (openFiles [i].masterHandle == data)
		break;

	if (i != numOpenFiles)
	    strcpy (openFiles [i].currentTag, tag->alias);

    }

/*----------*/
static int findFile (data)

    D4DATA	*data;

    {

	int	i;

	for (i=0 ; i<numOpenFiles ; i++)
	    if (openFiles [i].masterHandle == data)
		return i; 
	return -1;

    }

/*----------*/
static void setRecNumbers (void)

    {

	int	i;

	for (i=0 ; i<numOpenFiles ; i++) {
	    openFiles [i].currentRecord = d4recno (openFiles [i].masterHandle);
	    if (openFiles [i].appendInProgress)
		openFiles [i].currentRecord--;
	}

    }

/*----------*/
static void trimCpy (dest, source)  /* Copy string stopping at first blank */

    char	*dest;
    char	*source;

    {

	while ((*dest = *source)) {
	    if (*dest == ' ') {
		*dest = 0;
		return;
	    }
	    dest++;
	    source++;
	}

    }

/*----------*/
static void strlwr (str)

    char	*str;

    {

	while (*str) {
	    *str = tolower (*str);
	    str++;
	}

    }

/*----------*/
static void sigPipeHandler (sig)

    int		sig;

    {


    }

/*----------*/
void inhibitRedundancy ()	/* Inhibit all redundancy */

    {

	inhibited = 1;

    }

/*----------*/
void allowRedundancy ()

    {

	inhibited = 0;

    }

/*----------*/
static uint32 makeDataHandle (D4DATA *data)

    {

	return makeHandle ((void *)data);

    }

/*----------*/
static uint32 makeFieldHandle (F4FIELD *field)

    {

	return makeHandle ((void *)field);

    }

/*----------*/
static uint32 makeHandle (void *ptr)

    {

	/* Get handle associated with a pointer. If none, create one */

	void	**p;

	/* Check if handle is already in the list */

	p = (void **) lsearch ((char *)&ptr, (char *)pTable, &numHandles, sizeof (void *), searchCompare);
	return (uint32)(p - pTable); 

    }

/*----------*/
static int searchCompare (const void *key, const void *elem)

    {

	return *(void **)key - *(void **)elem;

    }

