/*adjust.c,v 1.1.1.1 1999/08/03 14:44:27 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/****************************************************************/
/*								*/
/*		SDR RECORD ADJUSTMENT ROUTINES			*/
/*								*/
/*	The following set of routines is adapted, fairly	*/
/*	directly from libCom/sdrLoad.c. The major difference	*/
/*	here is that these routines do not mind loading new	*/
/*	sdr data on top of old. This could happen when the	*/
/*	downloading process gets interrupted and then starts	*/
/*	over. This is very unlikely in a pure Ethernet		*/
/*	environment, but could well happen when voice grade	*/
/*	comm lines are used.					*/
/*								*/
/****************************************************************/

#include <vxWorks.h>
#include <types.h>
#include <ctype.h>

#include <dbDefs.h>
#include <sdrHeader.h>
#include <choice.h>
#include <dbRecType.h>
#include <dbRecDes.h>
#include <dbRecords.h>
#include <devSup.h>
#include <drvSup.h>
#include <cvtTable.h>
#include <errMdef.h>
#include <error.h>

struct choiceSet *choiceCvt = NULL;
struct arrChoiceSet *choiceGbl = NULL;
struct choiceRec *choiceRec = NULL;
struct arrBrkTable *cvtTable = NULL;
struct recDes  *dbRecDes = NULL;
struct pvd     *dbPvd = NULL;
struct recType *dbRecType = NULL;
struct recHeader *dbRecords = NULL;
struct drvSup  *drvSup = NULL;
struct recSup  *recSup = NULL;
struct devChoiceRec *choiceDev = NULL;
struct recDevSup *devSup = NULL;

/*
 * ADJ_DBRECTYPE
 *
 * (Also handles the dbRecords pointer because we aren't going to
 * download the database records via the sdr mechanism.)
 */

static adj_dbRecType(ptmp)
char           *ptmp;
{
    int		i;
    int		offset;

    if (dbRecType)
	free(dbRecType);
    dbRecType = (struct recType *) ptmp;
    (long) dbRecType->papName += (long) dbRecType;
    for (i = 0; i < dbRecType->number; i++) {
	if (dbRecType->papName[i] != NULL)
	    (long) dbRecType->papName[i] += (long) dbRecType;
    }

    if (dbRecords)
	{
	for (i = 0; i < dbRecords -> number; i++)
		{
		if (dbRecords -> papRecLoc[i] != NULL)
			{
			struct recChain	*currStruct;
			struct recChain *nextStruct;
			currStruct = (struct recChain *) dbRecords -> papRecLoc[i] -> pFirst;
			while (currStruct)
				{
				nextStruct = currStruct -> next;
				free(currStruct);
				currStruct = nextStruct;
				}
			free(dbRecords -> papRecLoc[i]);
			}
		}
	free(dbRecords);
	}

    /* 
    contiguous space must be allocated for the header and pointer array 
    */

    offset = sizeof(struct recHeader) + dbRecType->number * sizeof(caddr_t);
    if ((dbRecords = (struct recHeader *) malloc((unsigned) offset)) == NULL) {
	    printErr("malloc error for database records\n");
	    return (ERROR);
    }

    bzero((caddr_t) dbRecords , (int) offset);
    (long)dbRecords->papRecLoc = (long) ((long)dbRecords + sizeof(struct recHeader));
    dbRecords->number = dbRecType->number;
    return (!ERROR);
}

/*
 * ADJ_DBRECDES
 */
static adj_dbRecDes(ptmp)
char           *ptmp;
{
    int		i;
    int		j;
    if (dbRecType == NULL) {
	return (ERROR);
    }
    if (dbRecDes)
	free((caddr_t) dbRecDes);

    dbRecDes = (struct recDes *) ptmp;
    (long) dbRecDes->papRecTypDes += (long) dbRecDes;
    for (i = 0; i < dbRecDes->number; i++) {
	if (dbRecDes->papRecTypDes[i] != NULL) {
	    (long) dbRecDes->papRecTypDes[i] += (long) dbRecDes;
	    (long) dbRecDes->papRecTypDes[i]->papFldDes += (long) dbRecDes;
	    (long) dbRecDes->papRecTypDes[i]->sortFldName += (long) dbRecDes;
	    (long) dbRecDes->papRecTypDes[i]->sortFldInd += (long) dbRecDes;
	    (long) dbRecDes->papRecTypDes[i]->link_ind += (long) dbRecDes;
	    for (j = 0; j < dbRecDes->papRecTypDes[i]->no_fields; j++) {
		if (dbRecDes->papRecTypDes[i]->papFldDes != NULL) {
		    if (dbRecDes->papRecTypDes[i]->papFldDes[j] != NULL) {
			(long) dbRecDes->papRecTypDes[i]->papFldDes[j]
			    += (long) dbRecDes;
		    }
		}
	    }
	}
    }
    return (!ERROR);
}
/*
 * ADJ_CHOICEGBL
 */
static adj_choiceGbl(ptmp)
char           *ptmp;
{
    int		i;
    int		j;
    if (choiceGbl)
	free((caddr_t) choiceGbl);

    choiceGbl = (struct arrChoiceSet *) ptmp;
    (long) choiceGbl->papChoiceSet += (long) choiceGbl;
    for (i = 0; i < choiceGbl->number; i++) {
	/*******************************************************/
	/* ALERT! I think this is an error. It appears to make */
	/* sense to me if (and only if) there are no "holes"   */
	/* in the global choice list. This is the case if you  */
	/* look in choiceGbl.ascii. If you want to be able to  */
	/* have a sparse list, then the following line should  */
	/* be moved to be the first thing that happens if the  */
	/* entry is not null (i.e., switch the order of the    */
	/* following two lines.				       */
	/*******************************************************/
	(long) choiceGbl->papChoiceSet[i] += (long) choiceGbl;
	if (choiceGbl->papChoiceSet[i] != NULL) {
	    (long) choiceGbl->papChoiceSet[i]->papChoice += (long) choiceGbl;
	    for (j = 0; j < choiceGbl->papChoiceSet[i]->number; j++) {
		if (choiceGbl->papChoiceSet[i]->papChoice[j] != NULL) {
		    (long) choiceGbl->papChoiceSet[i]->papChoice[j]
			+= (long) choiceGbl;
		}
	    }
	}
    }
    return (!ERROR);
}
/*
 * ADJ_CHOICECVT
 */
static adj_choiceCvt(ptmp)
char           *ptmp;
{
    int		i;
    if (choiceCvt)
	free((caddr_t) choiceCvt);
    choiceCvt = (struct choiceSet *) ptmp;
    (long) choiceCvt->papChoice += (long) choiceCvt;
    for (i = 0; i < choiceCvt->number; i++) {
	if (choiceCvt->papChoice[i] != NULL) {
	    (long) choiceCvt->papChoice[i] += (long) choiceCvt;
	}
    }
    return (!ERROR);
}
/*
 * ADJ_CHOICEREC
 */
static adj_choiceRec(ptmp)
char           *ptmp;
{
    int		i;
    int		j;
    int		k;
    if (!dbRecType) {
	return (ERROR);
    }

    if (choiceRec)
	free((caddr_t) choiceRec);

    choiceRec = (struct choiceRec *) ptmp;
    (long) choiceRec->papArrChoiceSet += (long) choiceRec;
    for (i = 0; i < dbRecType->number; i++) {
	if (choiceRec->papArrChoiceSet[i] != NULL) {
	    (long) choiceRec->papArrChoiceSet[i] += (long) choiceRec;
	    (long) choiceRec->papArrChoiceSet[i]->papChoiceSet
		+= (long) choiceRec;
	    for (j = 0; j < choiceRec->papArrChoiceSet[i]->number; j++) {
		if (choiceRec->papArrChoiceSet[i]->papChoiceSet[j] != NULL) {
		    (long) choiceRec->papArrChoiceSet[i]->papChoiceSet[j]
			+= (long) choiceRec;
		    (long) choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice
			+= (long) choiceRec;
		    for (k = 0; k < choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->number; k++) {
			if (choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice[k] != NULL) {
			    (long) choiceRec->papArrChoiceSet[i]->papChoiceSet[j]->papChoice[k]
				+= (long) choiceRec;
			}
		    }
		}
	    }
	}
    }
    return (!ERROR);
}
/*
 * ADJ_CHOICEDEV
 */
static adj_choiceDev(ptmp)
char           *ptmp;
{
    int		i;
    int		j;

    if (!dbRecType) {
	return (ERROR);
    }
    
    if (choiceDev)
	free((caddr_t) choiceDev);
    choiceDev = (struct devChoiceRec *) ptmp;
    (long) choiceDev->papDevChoiceSet += (long) choiceDev;
    for (i = 0; i < dbRecType->number; i++) {
	if (choiceDev->papDevChoiceSet[i] != NULL) {
	    (long) choiceDev->papDevChoiceSet[i] += (long) choiceDev;
	    (long) choiceDev->papDevChoiceSet[i]->papDevChoice
		+= (long) choiceDev;
	    for (j = 0; j < choiceDev->papDevChoiceSet[i]->number; j++) {
		if (choiceDev->papDevChoiceSet[i]->papDevChoice[j] != NULL) {
		    (long) choiceDev->papDevChoiceSet[i]->papDevChoice[j]
			+= (long) choiceDev;
		    (long) choiceDev->papDevChoiceSet[i]->papDevChoice[j]->pchoice
			+= (long) choiceDev;
		}
	    }
	}
    }
    return (!ERROR);
}
/*
 * ADJ_DEVSUP
 */
static adj_devSup(ptmp)
char           *ptmp;
{
    int		i;
    int		j;

    if (!dbRecType) {
	return (!ERROR);
    }

    if (devSup)
	free((caddr_t) devSup);

    devSup = (struct recDevSup *) ptmp;
    (long) devSup->papDevSup += (long) devSup;
    for (i = 0; i < dbRecType->number; i++) {
	if (devSup->papDevSup[i] != NULL) {
	    (long) devSup->papDevSup[i] += (long) devSup;
	    (long) devSup->papDevSup[i]->dsetName += (long) devSup;
	    (long) devSup->papDevSup[i]->papDset += (long) devSup;
	    for (j = 0; j < devSup->papDevSup[i]->number; j++) {
		if (devSup->papDevSup[i]->dsetName[j] != NULL) {
		    (long) devSup->papDevSup[i]->dsetName[j]
			+= (long) devSup;
		}
	    }
	}
    }
    return (!ERROR);
}
/*
 * ADJ_CVTTABLE
 */
static adj_cvtTable(ptmp)
char           *ptmp;
{
    int		i;
    int		j;

    if (cvtTable)
	free((caddr_t) cvtTable);

    cvtTable = (struct arrBrkTable *) ptmp;
    (long) cvtTable->papBrkTable += (long) cvtTable;
    for (i = 0; i < cvtTable->number; i++) {
	if (cvtTable->papBrkTable[i] != NULL) {
	    (long) cvtTable->papBrkTable[i] += (long) cvtTable;
	    (long) cvtTable->papBrkTable[i]->papBrkInt += (long) cvtTable;
	    (long) cvtTable->papBrkTable[i]->name += (long) cvtTable;
	    for (j = 0; j < cvtTable->papBrkTable[i]->number; j++) {
		(long) cvtTable->papBrkTable[i]->papBrkInt[j] += (long) cvtTable;
	    }
	}
    }
    return (!ERROR);
}
/*
 * ADJ_DRVSUP
 *
 * Load and adjust SDR structure
 *
 */
static adj_drvSup(ptmp)
char           *ptmp;
{
    int		i;
    if (drvSup)
	free((caddr_t) drvSup);

    drvSup = (struct drvSup *) ptmp;
    (long) drvSup->drvetName += (long) drvSup;
    (long) drvSup->papDrvet += (long) drvSup;
    for (i = 0; i < drvSup->number; i++) {
	if (drvSup->drvetName[i] != NULL)
	    (long) drvSup->drvetName[i] += (long) drvSup;
    }
    return (!ERROR);
}
/*
 * ADJ_RECSUP
 */
static adj_recSup(ptmp)
char           *ptmp;
{
    if (recSup)
	free((caddr_t) recSup);

    recSup = (struct recSup *) ptmp;
    return (!ERROR);
}
/* Note: function placement must agree with sdrHeader.h */
int (*adjFun[]) () = {

    adj_dbRecType,
    NULL,
    adj_dbRecDes,
    NULL,
    adj_choiceGbl,
    adj_choiceCvt,
    adj_choiceRec,
    adj_choiceDev,
    adj_devSup,
    adj_cvtTable,
    adj_drvSup,
    adj_recSup,
};


/****************************************************************/
/*								*/
/*		FREE ALL DATABASE MEMORY			*/
/*								*/
/*	This routine frees all of the memory that has been	*/
/*	allocated to store the database. This routine is 	*/
/*	called when it has been decided to abort a download.	*/
/*	By freeing all of the memory before starting over,	*/
/*	we end up with a clean slate.				*/
/*								*/
/****************************************************************/

int deleteDatabase(void)
{
    /*
    Record support.
    */

    if (recSup)
	{
	free(recSup);
	recSup = 0;
	}

    /*
    Driver support.
    */

    if (drvSup)
	{
	free(drvSup);
	drvSup = 0;
	}

    /*
    Conversion Table
    */

    if (cvtTable)
	{
	free(cvtTable);
	cvtTable = 0;
	}

    /*
    Device Support
    */

    if (devSup)
	{
	free(devSup);
	devSup = 0;
	}

    /*
    Device Choice
    */

    if (choiceDev)
	{
	free(choiceDev);
	choiceDev = 0;
	}

    /*
    Record Choice
    */

    if (choiceRec)
	{
	free(choiceRec);
	choiceRec = 0;
	}

    /*
    Conversion Choice
    */

    if (choiceCvt)
	{
	free(choiceCvt);
	choiceCvt = 0;
	}

    /*
    Global Choice
    */

    if (choiceGbl)
	{
	free(choiceGbl);
	choiceGbl = 0;
	}

    /*
    Record Description
    */

    if (dbRecDes)
	{
	free(dbRecDes);
	dbRecDes = 0;
	}

    /*
    Record Types
    */

    if (dbRecType)
	{
	free(dbRecType);
	dbRecType = 0;
	}

    /*
    The database records themselves
    */

    if (dbRecords)	
	{
	int	i;

	for (i = 0; i < dbRecords -> number; i++)
	    {
	    if (dbRecords -> papRecLoc[i])
		{
		struct recChain	*currChain;
		struct recChain	*nextChain;

		currChain = (struct recChain *)(dbRecords -> papRecLoc[i] -> pFirst);

		while (currChain)
		    {
		    nextChain = currChain -> next;
		    free(currChain);
		    currChain = nextChain;
		    }
		free(dbRecords -> papRecLoc[i]);
		}
	    }
	free(dbRecords);
	dbRecords = 0;
	}

    /*
    And finally, the pvd
    */

    if (dbPvd)
	{
	free(dbPvd);
	dbPvd = 0;
	pvd_create();
	}
    return (0);
}
