/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*============================================================================*/
/*									      */
/*									      */
/*	Tate Integrated Systems, LP					      */
/*									      */
/*									      */
/*	Author:			Greg Tomalesky				      */
/*	Date:			22-Jan-1993				      */
/*									      */
/*	Module Name:		DBCCRCFunctions.c			      */
/*	Module Description:						      */
/*									      */
/*									      */
/*	Modification History:	First try	22-Jan-1993		      */
/*									      */
/*									      */
/*									      */
/*									      */
/*									      */
/*============================================================================*/

#ifdef WIN32
typedef char *caddr_t;
#include <memory.h>
#include <string.h>
#include <stdio.h>
#include <winsock.h>
#else
#ifdef vxWorks
#include <string.h>
#include <stdio.h>
#endif vxWorks
#include	<types.h>
#include	<error.h>
#include	<in.h>
#endif WIN32

#include	<math.h>

#include	"dbDefs.h"
#include	"dbRecType.h"
#include	"dbRecDes.h"
#include	"dbRecords.h"
#include	"dbFldTypes.h"
#include	"choice.h"
#include	"link.h"
#include	"dbcSub.h"
#include	"tisByteOrder.h"

#define NEW_ROUNDING_METHOD 1

#ifdef NEW_ROUNDING_METHOD
typedef union
  {
    double doubleValue;
    unsigned int    intValue[2];
  } NEW_DOUBLE;
double tempDouble;
#endif NEW_ROUNDING_METHOD

/*
	Database pointers.
*/
#ifndef WIN32
extern	ERRSYMTAB_ID	errSymTbl;			/* defined in error.h */
#endif WIN32

double	dbRound (double, int);
#ifndef vxWorks
double olddbRound (double, int);
#endif
double atof ();

static double place[] =
	{
	1.0,
	10.0,
	100.0,
	1000.0,
	10000.0,
	100000.0,
	1000000.0
	};

char Target[PVNAME_SZ] = {0};
static int docmp = 0;
static struct fldDes   	*DBC_FldDes;
/* static short realIOC = -1; *//* ems 23 sep 98 */


/*============================================================================*/
/*									      */
/*									      */
/*	Tate Integrated Systems, LP					      */
/*									      */
/*									      */
/*	Author:			Greg Tomalesky				      */
/*	Date:			21-Jan-1993				      */
/*									      */
/*	Function Name:		DBC_MakeCRC				      */
/*	Function Description:						      */
/*									      */
/*									      */
/*									      */
/*									      */
/*	Input Parameters:	none.					      */
/*									      */
/*	Output Parameters:	none.					      */
/*									      */
/*	Return Values or error codes:					      */
/*									      */
/*	Modification History:						      */
/*									      */
/*									      */
/*									      */
/*									      */
/*============================================================================*/
unsigned short	DBC_MakeCRC (char *buf, int len, unsigned short	seed)
{
    unsigned short	checksum;
    int			i,j;
    unsigned short	flag;
    unsigned char	tempch;

    checksum = seed;				/* use previous value. (0xffff to start) */
    for(j = 0;j < len;j++)
    {
	tempch = buf[j];
	flag = tempch;
	checksum ^= tempch;
	for(i = 0;i < 8; i++)
	{
	    flag = checksum & 1;
	    checksum >>= 1;
	    if(flag)
		checksum ^= (unsigned short)0x0a001;
	}
    }
    if(docmp)
    {
	/*       FIELD   oldCRC  newCRC */
	printf("<%*.*s> <%4.4X> <%4.4X> ",FLDNAME_SZ,FLDNAME_SZ,DBC_FldDes->fldname,seed,checksum);
	for(j = 0;j < len;j++)
	    printf("%02X ",(unsigned char)buf[j]);
	printf("\n");
    }
    return(checksum);
}

/*============================================================================*/
/*									      */
/*									      */
/*	Tate Integrated Systems, LP					      */
/*									      */
/*									      */
/*	Author:			Greg Tomalesky				      */
/*	Date:			21-Jan-1993				      */
/*									      */
/*	Function Name:		DBC_CRCHostRecord			      */
/*	Function Description:						      */
/*									      */
/*									      */
/*									      */
/*									      */
/*	Input Parameters:	char	*buf - pointer to db record (in local */
/*								host form!)   */
/*				int	type_index - point type number	      */
/*									      */
/*	Output Parameters:	none.					      */
/*									      */
/*	Return Values or error codes:	return the CRC			      */
/*									      */
/*	Modification History:						      */
/*									      */
/*									      */
/*									      */
/*									      */
/*============================================================================*/
unsigned short	DBC_CRCHostRecord(buf,type_index)
char	*buf;
int	type_index;
{
    struct recTypDes	*DBC_RecTypDes;
    int			field_index,size,prec;
    char		*bptr;
    short		stemp;
    int			itemp;
    float		ftemp;
    double		dtemp;
    unsigned short	crc;
    struct link		*lptr,temp_link;
    long		link_type;

    DBC_RecTypDes = GET_PRECTYPDES(type_index);			/* point to type descriptor in SDR */
    field_index = 0;
    crc = 0xffff;						/* CRC seed value */
    docmp = 0;							/* 6-Dec-96 compare point names with */
    if(!strcmp(Target,buf))		       			/* target point. if equal then print info */
    {
	printf("Target point %s\n",Target);
	docmp = 1;
    }
    while((DBC_FldDes = GET_PFLDDES(DBC_RecTypDes,field_index))) /* loop to get all desc's */
    {
	if(DBC_FldDes -> promptflag)				/* CRC only for prompted fields */
	{
	    bptr = buf + DBC_FldDes -> offset;
	    size = DBC_FldDes -> size;
	    switch(DBC_FldDes -> field_type)
	    {
		case DBF_INLINK:
		case DBF_OUTLINK:
		case DBF_FWDLINK:
		    lptr = (struct link *)bptr;	/* point to link field */
		    link_type = lptr -> type;
		    itemp = htonl(lptr -> type);
		    bptr = (char *) &itemp;
		    size = sizeof(itemp);
		    if (link_type >= SAB_IO)
		    {
			temp_link.value.tateio.link = htons(lptr -> value.tateio.link);
			temp_link.value.tateio.adapter = htons(lptr -> value.tateio.adapter);
			temp_link.value.tateio.card = htons(lptr -> value.tateio.card);
			temp_link.value.tateio.signal = htons(lptr -> value.tateio.signal);
			crc = DBC_MakeCRC((char *)&(temp_link.value.tateio), sizeof(struct tateio), crc);
		    }
		    else
		    {
			switch(link_type)
			{
			    case DB_LINK:
				itemp = htonl(PV_LINK);
				/*
					Copy the link structure to a temp area and change CA_LINK & DB_LINK
					back to PV_LINK so we can calc the CRCs over the proper areas of the union.
				*/
				temp_link.value.pv_link.process_passive = htons(lptr->value.db_link.process_passive);
				temp_link.value.pv_link.maximize_sevr = htons(lptr->value.db_link.maximize_sevr);
				crc = DBC_MakeCRC((char *)&temp_link.value.pv_link.process_passive, sizeof(short), crc);
				crc = DBC_MakeCRC((char *)&temp_link.value.pv_link.maximize_sevr, sizeof(short), crc);
				size = strlen(lptr -> value.db_link.pvname);
				if (size > PVNAME_SZ)
				    size = PVNAME_SZ;
				crc = DBC_MakeCRC(lptr -> value.db_link.pvname, size, crc);
				size = strlen(lptr -> value.db_link.fldname);
				if (size > FLDNAME_SZ)
				    size = FLDNAME_SZ;
				crc = DBC_MakeCRC(lptr -> value.db_link.fldname, size, crc);
				break;

			    case PV_LINK:
				temp_link.value.pv_link.process_passive = htons(lptr->value.pv_link.process_passive);
				temp_link.value.pv_link.maximize_sevr = htons(lptr->value.pv_link.maximize_sevr);
				crc = DBC_MakeCRC((char *)&temp_link.value.pv_link.process_passive, sizeof(short), crc);
				crc = DBC_MakeCRC((char *)&temp_link.value.pv_link.maximize_sevr, sizeof(short), crc);
				size = strlen(lptr -> value.pv_link.pvname);
				if (size > PVNAME_SZ)
					size = PVNAME_SZ;
				crc = DBC_MakeCRC(lptr -> value.pv_link.pvname, size, crc);
				size = strlen(lptr -> value.pv_link.fldname);
				if (size > FLDNAME_SZ)
					size = FLDNAME_SZ;
				crc = DBC_MakeCRC(lptr -> value.pv_link.fldname, size, crc);
				break;

			    case CONSTANT:
				switch(type_index)	/* digital point types where CNST is a short */
				{
				    case 3:
				    case 4:
				    case 7:
				    case 8:
					prec = 0;	/* drop all digits to right of DP */
					break;
				    default:
					prec = 3;
				}
				*((unsigned long *)&ftemp) = *((unsigned long *)&lptr->value.value);
#ifdef NEW_ROUNDING_METHOD
				tempDouble = dbRound((double)ftemp, prec);
				crc = DBC_MakeCRC((char *)&tempDouble, sizeof(double), crc);
#else
				ftemp = (float)dbRound((double)ftemp, prec);
				*((unsigned long *)&ftemp) = htonl(*((unsigned long *)&ftemp));
				crc = DBC_MakeCRC((char *)&ftemp, sizeof(float), crc);
#endif NEW_ROUNDING_METHOD
				break;

			    case VME_IO:
				crc = DBC_MakeCRC((char *)&(lptr -> value.vmeio), sizeof(struct vmeio), crc);
				break;

			    case CAMAC_IO:
				temp_link.value.camacio.branch = htons(lptr -> value.camacio.branch);
				temp_link.value.camacio.crate = htons(lptr -> value.camacio.crate);
				temp_link.value.camacio.slot = htons(lptr -> value.camacio.slot);
				temp_link.value.camacio.channel = htons(lptr -> value.camacio.channel);
				crc = DBC_MakeCRC((char *)&(temp_link.value.camacio), sizeof(struct camacio), crc);
				break;

			    case AB_IO:
				temp_link.value.abio.link = htons(lptr -> value.abio.link);
				temp_link.value.abio.adapter = htons(lptr -> value.abio.adapter);
				temp_link.value.abio.card = htons(lptr -> value.abio.card);
				temp_link.value.abio.signal = htons(lptr -> value.abio.signal);
				temp_link.value.abio.plc_flag = htons(lptr -> value.abio.plc_flag);
				crc = DBC_MakeCRC((char *)&(temp_link.value.abio), sizeof(struct abio), crc);
				break;

			    case GPIB_IO:
				temp_link.value.gpibio.link = htons(lptr -> value.gpibio.link);
				temp_link.value.gpibio.addr = htons(lptr -> value.gpibio.addr);
				crc = DBC_MakeCRC((char *)&(temp_link.value.gpibio.link), sizeof(short), crc);
				crc = DBC_MakeCRC((char *)&(temp_link.value.gpibio.addr), sizeof(short), crc);
				size = strlen(lptr -> value.gpibio.parm);
				if (size > sizeof(lptr -> value.gpibio.parm))
					size = sizeof(lptr -> value.gpibio.parm);
				crc = DBC_MakeCRC(lptr -> value.gpibio.parm, size, crc);
				break;

			    case BITBUS_IO:
				crc = DBC_MakeCRC(&(lptr -> value.bitbusio.link), sizeof(char), crc);
				crc = DBC_MakeCRC(&(lptr -> value.bitbusio.node), sizeof(char), crc);
				crc = DBC_MakeCRC(&(lptr -> value.bitbusio.port), sizeof(char), crc);
				crc = DBC_MakeCRC(&(lptr -> value.bitbusio.signal), sizeof(char), crc);
				size = strlen(lptr -> value.bitbusio.parm);
				if (size > sizeof(lptr -> value.bitbusio.parm))
					size = sizeof(lptr -> value.bitbusio.parm);
				crc = DBC_MakeCRC(lptr -> value.bitbusio.parm, size, crc);
				break;

			    case INST_IO:
				size = strlen(lptr -> value.instio.string);
				if (size > sizeof(lptr -> value.instio.string))
					size = sizeof(lptr -> value.instio.string);
				crc = DBC_MakeCRC(lptr -> value.instio.string, size, crc);
				break;

			    case BBGPIB_IO:
				crc = DBC_MakeCRC(&(lptr -> value.bbgpibio.link), sizeof(char), crc);
				crc = DBC_MakeCRC(&(lptr -> value.bbgpibio.bbaddr), sizeof(char), crc);
				crc = DBC_MakeCRC(&(lptr -> value.bbgpibio.gpibaddr), sizeof(char), crc);
				size = strlen(lptr -> value.bbgpibio.parm);
				if (size > sizeof(lptr -> value.bbgpibio.parm))
					size = strlen(lptr -> value.bbgpibio.parm);
				crc = DBC_MakeCRC(lptr -> value.bbgpibio.parm, size, crc);
				break;
			}
			break;
		    }
		case DBF_STRING:
		    size = strlen(bptr);
		    break;

		case DBF_FLOAT:	/* GCT 07-Jun-94 must use 'htonl' for MIPSEL compatability */
		    ftemp = *(float *)bptr;
#ifdef NEW_ROUNDING_METHOD
		    tempDouble = dbRound((double)ftemp,DBC_FldDes -> promptflag);
		    bptr = (char *)&tempDouble;
		    size = sizeof(double);
#else
		    ftemp = dbRound((double)ftemp,DBC_FldDes -> promptflag);
		    *((unsigned long *)&ftemp) = htonl(*((unsigned long *)&ftemp));
		    bptr = (char *)&ftemp;
		    size = sizeof(float);
#endif
		    break;

		case DBF_DOUBLE:
		    dtemp = *(double *)bptr;
#ifdef NEW_ROUNDING_METHOD
		    tempDouble = dbRound(dtemp,DBC_FldDes -> promptflag);
		    bptr = (char *)&tempDouble;
#else
		    dtemp = dbRound(dtemp,DBC_FldDes -> promptflag);
		    htond (dtemp, dtemp);
		    bptr = (char *)&dtemp;
#endif NEW_ROUNDING_METHOD
		    size = sizeof(double);
		    break;

		case DBF_SHORT:
		case DBF_USHORT:
		case DBF_ENUM:
		case DBF_GBLCHOICE:
		case DBF_CVTCHOICE:
		case DBF_RECCHOICE:
		case DBF_DEVCHOICE:
		    stemp = htons(*(short *)bptr);
		    bptr = (char *) &stemp;
		    break;

		case DBF_LONG:
		case DBF_ULONG:
		    itemp = htonl(*(int *)bptr);
		    bptr = (char *) &itemp;
		    break;

		default:
			break;
	    }
	    crc = DBC_MakeCRC(bptr,size,crc);
	}
	field_index++;
    }
    return(crc);
}

/*============================================================================*/
/*									      */
/*									      */
/*	Tate Integrated Systems, LP					      */
/*									      */
/*									      */
/*	Author:			Greg Tomalesky				      */
/*	Date:			21-Jan-1993				      */
/*									      */
/*	Function Name:		DBC_CRCRecord				      */
/*	Function Description:						      */
/*									      */
/*									      */
/*									      */
/*									      */
/*	Input Parameters:	char	*buf - pointer to db record (in net   */
/*							standard form!)       */
/*				int	type_index - point type number	      */
/*									      */
/*	Output Parameters:	none.					      */
/*									      */
/*	Return Values or error codes:	return the CRC			      */
/*									      */
/*	Modification History:						      */
/*									      */
/*									      */
/*									      */
/*									      */
/*============================================================================*/
unsigned short	DBC_CRCRecord(buf,type_index)
char	*buf;
int	type_index;
{
    struct recTypDes	*DBC_RecTypDes;
    int			field_index,size,prec;
    char		*bptr;
    int			itemp;
    float		ftemp;
    double		dtemp;
    unsigned short	crc;
    struct link		*lptr;
    long		link_type;

    DBC_RecTypDes = GET_PRECTYPDES(type_index);			/* point to type descriptor in SDR */
    field_index = 0;
    crc = 0xffff;							/* CRC seed value */
    docmp = 0;							/* 6-Dec-96 compare point names with */
    if(!strcmp(Target,buf))		       				/* target point. if equal then print info */
    {
	printf("Target point %s\n",Target);
	docmp = 1;
    }
    while((DBC_FldDes = GET_PFLDDES(DBC_RecTypDes,field_index)))	/* loop to get all desc's */
    {
	if(DBC_FldDes -> promptflag)				/* CRC only for prompted fields */
	{
	    bptr = buf + DBC_FldDes -> offset;
	    size = DBC_FldDes -> size;
	    switch(DBC_FldDes -> field_type)
	    {
		case DBF_INLINK:
		case DBF_OUTLINK:
		case DBF_FWDLINK:
		    lptr = (struct link *)bptr;	/* point to link field */
		    link_type = ntohl(lptr -> type);
		    itemp = lptr -> type;
		    bptr = (char *) &itemp;
		    size = sizeof(itemp);
		    if (link_type >= SAB_IO)
		    {
			crc = DBC_MakeCRC((char *)&(lptr->value.tateio), sizeof(struct tateio), crc);
		    }
		    else
		    {
			switch(link_type)
			{
			    case DB_LINK:
				itemp = htonl(PV_LINK);
				/*
					Copy the link structure to a temp area and change CA_LINK & DB_LINK
					back to PV_LINK so we can calc the CRCs over the proper areas of the union.
				*/
				crc = DBC_MakeCRC((char *)&(lptr->value.pv_link.process_passive), sizeof(short), crc);
				crc = DBC_MakeCRC((char *)&(lptr->value.pv_link.maximize_sevr), sizeof(short), crc);
				size = strlen(lptr -> value.db_link.pvname);
				if (size > PVNAME_SZ)
				    size = PVNAME_SZ;
				crc = DBC_MakeCRC(lptr -> value.db_link.pvname, size, crc);
				size = strlen(lptr -> value.db_link.fldname);
				if (size > FLDNAME_SZ)
				    size = FLDNAME_SZ;
				crc = DBC_MakeCRC(lptr -> value.db_link.fldname, size, crc);
				break;

			    case PV_LINK:
				crc = DBC_MakeCRC((char *)&(lptr->value.pv_link.process_passive), sizeof(short), crc);
				crc = DBC_MakeCRC((char *)&(lptr->value.pv_link.maximize_sevr), sizeof(short), crc);
				size = strlen(lptr -> value.pv_link.pvname);
				if (size > PVNAME_SZ)
					size = PVNAME_SZ;
				crc = DBC_MakeCRC(lptr -> value.pv_link.pvname, size, crc);
				size = strlen(lptr -> value.pv_link.fldname);
				if (size > FLDNAME_SZ)
					size = FLDNAME_SZ;
				crc = DBC_MakeCRC(lptr -> value.pv_link.fldname, size, crc);
				break;

			    case CONSTANT:
				switch(type_index)	/* digital point types where CNST is a short */
				{
				    case 3:
				    case 4:
				    case 7:
				    case 8:
					prec = 0;	/* drop all digits to right of DP */
					break;
				    default:
					prec = 3;
				}
				*((unsigned long *)&ftemp) = ntohl(*((unsigned long *)&lptr->value.value));
#ifdef NEW_ROUNDING_METHOD
				tempDouble = dbRound((double)ftemp, prec);
				crc = DBC_MakeCRC((char *)&tempDouble, sizeof(double), crc);
#else
				ftemp = (float)dbRound((double)ftemp, prec);
				*((unsigned long *)&ftemp) = htonl(*((unsigned long *)&ftemp));
				crc = DBC_MakeCRC((char *)&ftemp, sizeof(float), crc);
#endif NEW_ROUNDING_METHOD
				break;

			    case VME_IO:
				crc = DBC_MakeCRC((char *)&(lptr -> value.vmeio), sizeof(struct vmeio), crc);
				break;

			    case CAMAC_IO:
				crc = DBC_MakeCRC((char *)&(lptr->value.camacio), sizeof(struct camacio), crc);
				break;

			    case AB_IO:
				crc = DBC_MakeCRC((char *)&(lptr->value.abio), sizeof(struct abio), crc);
				break;

			    case GPIB_IO:
				crc = DBC_MakeCRC((char *)&(lptr->value.gpibio.link), sizeof(short), crc);
				crc = DBC_MakeCRC((char *)&(lptr->value.gpibio.addr), sizeof(short), crc);
				size = strlen(lptr -> value.gpibio.parm);
				if (size > sizeof(lptr -> value.gpibio.parm))
					size = sizeof(lptr -> value.gpibio.parm);
				crc = DBC_MakeCRC(lptr -> value.gpibio.parm, size, crc);
				break;

			    case BITBUS_IO:
				crc = DBC_MakeCRC(&(lptr -> value.bitbusio.link), sizeof(char), crc);
				crc = DBC_MakeCRC(&(lptr -> value.bitbusio.node), sizeof(char), crc);
				crc = DBC_MakeCRC(&(lptr -> value.bitbusio.port), sizeof(char), crc);
				crc = DBC_MakeCRC(&(lptr -> value.bitbusio.signal), sizeof(char), crc);
				size = strlen(lptr -> value.bitbusio.parm);
				if (size > sizeof(lptr -> value.bitbusio.parm))
					size = sizeof(lptr -> value.bitbusio.parm);
				crc = DBC_MakeCRC(lptr -> value.bitbusio.parm, size, crc);
				break;

			    case INST_IO:
				size = strlen(lptr -> value.instio.string);
				if (size > sizeof(lptr -> value.instio.string))
					size = sizeof(lptr -> value.instio.string);
				crc = DBC_MakeCRC(lptr -> value.instio.string, size, crc);
				break;

			    case BBGPIB_IO:
				crc = DBC_MakeCRC(&(lptr -> value.bbgpibio.link), sizeof(char), crc);
				crc = DBC_MakeCRC(&(lptr -> value.bbgpibio.bbaddr), sizeof(char), crc);
				crc = DBC_MakeCRC(&(lptr -> value.bbgpibio.gpibaddr), sizeof(char), crc);
				size = strlen(lptr -> value.bbgpibio.parm);
				if (size > sizeof(lptr -> value.bbgpibio.parm))
					size = strlen(lptr -> value.bbgpibio.parm);
				crc = DBC_MakeCRC(lptr -> value.bbgpibio.parm, size, crc);
				break;
			}
			break;
		    }
		case DBF_STRING:
		    size = strlen(bptr);
		    break;

		case DBF_FLOAT:	/* GCT 07-Jun-94 must use 'htonl' for MIPSEL compatability */
		    *((unsigned long *)&ftemp) = ntohl(*((unsigned long *)bptr));
#ifdef NEW_ROUNDING_METHOD
		    tempDouble = dbRound(ftemp, DBC_FldDes -> promptflag);
		    bptr = (char *)&tempDouble;
		    size = sizeof(double);
#else
		    ftemp = dbRound((double)ftemp, DBC_FldDes -> promptflag);
		    *((unsigned long *)&ftemp) = htonl(*((unsigned long *)&ftemp));
		    bptr = (char *)&ftemp;
		    size = sizeof(float);
#endif
		    break;

		case DBF_DOUBLE:
		    ntohd (*(double *)bptr, dtemp);
#ifdef NEW_ROUNDING_METHOD
		    tempDouble = dbRound(dtemp,DBC_FldDes -> promptflag);
		    bptr = (char *)&tempDouble;
#else
		    dtemp = dbRound(dtemp,DBC_FldDes -> promptflag);
		    bptr = (char *)&dtemp;
#endif NEW_ROUNDING_METHOD
		    htond (dtemp, dtemp);
		    size = sizeof(double);
		    break;

		case DBF_SHORT:
		case DBF_USHORT:
		case DBF_ENUM:
		case DBF_GBLCHOICE:
		case DBF_CVTCHOICE:
		case DBF_RECCHOICE:
		case DBF_DEVCHOICE:
		    break;

		case DBF_LONG:
		case DBF_ULONG:
		    break;

		default:
			break;
	    }
	    crc = DBC_MakeCRC(bptr,size,crc);
	}
	field_index++;
    }
    return(crc);
}
/*
	This function will dbRound to 'num' digits
*/
#ifdef NEW_ROUNDING_METHOD
double dbRound(double x,int num)
{
    NEW_DOUBLE val;
    int 	leftPart,rightPart;
    double	wholeNumber;
    double	fraction;

    leftPart = (int)x;
    wholeNumber = leftPart;
    fraction = x - wholeNumber;
    if (fraction > 0.0)
	rightPart = (int)(fraction * place[num] + 0.5);
    else if(fraction < 0.0)
	rightPart = (int)(fraction * place[num] - 0.5);
    else /* fraction == 0.0 */
	rightPart = (int)fraction;
    val.intValue[0] = htonl(leftPart);
    val.intValue[1] = htonl(rightPart);
    return(val.doubleValue);
}
#else
double dbRound(double x, int num)
{
    double  rounded;
    char    buffer[32];

    if (realIOC < 0)
    {
#ifdef vxWorks
	realIOC = strcmp(sysModel(), "MicroSol XCell");
#else
	realIOC = 1;
#endif
    }
    if (realIOC)
    {
	int		leftPart;
	double	wholeNumber;
	double	fraction;

	leftPart = x;
	wholeNumber = (double) leftPart;
	fraction = x - wholeNumber;
	if (fraction > 0.0)
	    leftPart = fraction * place[num] + 0.5;
	else
	    leftPart = fraction * place[num] - 0.5;
	fraction = leftPart;
	fraction /= place[num];
	rounded = wholeNumber + fraction;
    }
    else
    {
	sprintf(buffer, "%20.*f", num, x);
	rounded = atof(buffer);
    }

    return (rounded);
}
#endif NEW_ROUNDING_METHOD

#ifndef vxWorks
unsigned short	oldDBC_CRCRecord(buf,type_index)
char	*buf;
int	type_index;
{
    struct recTypDes	*DBC_RecTypDes;
    int			field_index,size,prec;
    char		*bptr;
    int			itemp;
    float		ftemp;
    double		dtemp;
    unsigned short	crc;
    struct link		*lptr,temp_link;
    long		link_type;

    DBC_RecTypDes = GET_PRECTYPDES(type_index);			/* point to type descriptor in SDR */
    field_index = 0;
    crc = 0xffff;							/* CRC seed value */
    docmp = 0;							/* 6-Dec-96 compare point names with */
    if(!strcmp(Target,buf))		       				/* target point. if equal then print info */
    {
	printf("Target point %s\n",Target);
	docmp = 1;
    }
    while(DBC_FldDes = GET_PFLDDES(DBC_RecTypDes,field_index))	/* loop to get all desc's */
    {
	if(DBC_FldDes -> promptflag)				/* CRC only for prompted fields */
	{
	    bptr = buf + DBC_FldDes -> offset;
	    size = DBC_FldDes -> size;
	    switch(DBC_FldDes -> field_type)
	    {
		case DBF_INLINK:
		case DBF_OUTLINK:
		case DBF_FWDLINK:
		    lptr = (struct link *)bptr;	/* point to link field */

		    itemp = lptr -> type;
		    link_type = ntohl(lptr -> type);
		    bptr = (char *) &itemp;
		    size = sizeof(itemp);
		    if (link_type >= SAB_IO)
		    {
			crc = DBC_MakeCRC((char *)&(lptr -> value.tateio), sizeof(struct tateio), crc);
		    }
		    else
		    {
			switch(link_type)
			{
			    case DB_LINK:
				itemp = htonl(PV_LINK);
				/*
					Copy the link structure to a temp area and change CA_LINK & DB_LINK
					back to PV_LINK so we can calc the CRCs over the proper areas of the union.
				*/
				temp_link.value.pv_link.process_passive = lptr->value.db_link.process_passive;
				temp_link.value.pv_link.maximize_sevr = lptr->value.db_link.maximize_sevr;
				crc = DBC_MakeCRC((char *)&lptr->value.pv_link.process_passive, sizeof(short), crc);
				crc = DBC_MakeCRC((char *)&lptr->value.pv_link.maximize_sevr, sizeof(short), crc);
				size = strlen(lptr -> value.db_link.pvname);
				if (size > PVNAME_SZ)
					size = PVNAME_SZ;
				crc = DBC_MakeCRC(lptr -> value.db_link.pvname, size, crc);
				size = strlen(lptr -> value.db_link.fldname);
				if (size > FLDNAME_SZ)
					size = FLDNAME_SZ;
				crc = DBC_MakeCRC(lptr -> value.db_link.fldname, size, crc);
				break;

			    case PV_LINK:
				temp_link.value.pv_link.process_passive = lptr->value.pv_link.process_passive;
				temp_link.value.pv_link.maximize_sevr = lptr->value.pv_link.maximize_sevr;
				crc = DBC_MakeCRC((char *)&lptr->value.pv_link.process_passive, sizeof(short), crc);
				crc = DBC_MakeCRC((char *)&lptr->value.pv_link.maximize_sevr, sizeof(short), crc);
				size = strlen(lptr -> value.pv_link.pvname);
				if (size > PVNAME_SZ)
					size = PVNAME_SZ;
				crc = DBC_MakeCRC(lptr -> value.pv_link.pvname, size, crc);
				size = strlen(lptr -> value.pv_link.fldname);
				if (size > FLDNAME_SZ)
					size = FLDNAME_SZ;
				crc = DBC_MakeCRC(lptr -> value.pv_link.fldname, size, crc);
				break;

			    case CONSTANT:
				switch(type_index)	/* digital point types where CNST is a short */
					{
					case 3:
					case 4:
					case 7:
					case 8:
						prec = 0;	/* drop all digits to right of DP */
						break;
					default:
						prec = 3;
					}

				memcpy((char *)&ftemp,(char *)&lptr->value.value,sizeof(float));
				*((unsigned long *)&ftemp) = ntohl(*((unsigned long *)&ftemp));
				ftemp = (float)olddbRound((double)ftemp,prec);
				*((unsigned long *)&ftemp) = htonl(*((unsigned long *)&ftemp));
				crc = DBC_MakeCRC((char *)&ftemp, sizeof(float), crc);
				break;

			    case VME_IO:
				crc = DBC_MakeCRC((char *)&(lptr -> value.vmeio), sizeof(struct vmeio), crc);
				break;

			    case CAMAC_IO:
				crc = DBC_MakeCRC((char *)&(lptr -> value.camacio), sizeof(struct camacio), crc);
				break;

			    case AB_IO:
				crc = DBC_MakeCRC((char *)&(lptr -> value.abio), sizeof(struct abio), crc);
				break;

			    case GPIB_IO:
				crc = DBC_MakeCRC((char *)&(lptr -> value.gpibio.link), sizeof(short), crc);
				crc = DBC_MakeCRC((char *)&(lptr -> value.gpibio.addr), sizeof(short), crc);
				size = strlen(lptr -> value.gpibio.parm);
				if (size > sizeof(lptr -> value.gpibio.parm))
					size = sizeof(lptr -> value.gpibio.parm);
				crc = DBC_MakeCRC(lptr -> value.gpibio.parm, size, crc);
				break;

			case BITBUS_IO:
				crc = DBC_MakeCRC(&(lptr -> value.bitbusio.link), sizeof(char), crc);
				crc = DBC_MakeCRC(&(lptr -> value.bitbusio.node), sizeof(char), crc);
				crc = DBC_MakeCRC(&(lptr -> value.bitbusio.port), sizeof(char), crc);
				crc = DBC_MakeCRC(&(lptr -> value.bitbusio.signal), sizeof(char), crc);
				size = strlen(lptr -> value.bitbusio.parm);
				if (size > sizeof(lptr -> value.bitbusio.parm))
				    size = sizeof(lptr -> value.bitbusio.parm);
				crc = DBC_MakeCRC(lptr -> value.bitbusio.parm, size, crc);
				break;

			    case INST_IO:
				size = strlen(lptr -> value.instio.string);
				if (size > sizeof(lptr -> value.instio.string))
					size = sizeof(lptr -> value.instio.string);
				crc = DBC_MakeCRC(lptr -> value.instio.string, size, crc);
				break;

			    case BBGPIB_IO:
				crc = DBC_MakeCRC(&(lptr -> value.bbgpibio.link), sizeof(char), crc);
				crc = DBC_MakeCRC(&(lptr -> value.bbgpibio.bbaddr), sizeof(char), crc);
				crc = DBC_MakeCRC(&(lptr -> value.bbgpibio.gpibaddr), sizeof(char), crc);
				size = strlen(lptr -> value.bbgpibio.parm);
				if (size > sizeof(lptr -> value.bbgpibio.parm))
					size = strlen(lptr -> value.bbgpibio.parm);
				crc = DBC_MakeCRC(lptr -> value.bbgpibio.parm, size, crc);
				break;
			}
			break;
		    }
		case DBF_STRING:
		    size = strlen(bptr);
		    break;

		case DBF_FLOAT:	/* GCT 07-Jun-94 must use 'htonl' for MIPSEL compatability */
		    memcpy((char *)&ftemp,bptr,sizeof(float));
		    *((unsigned long *)&ftemp) = ntohl(*((unsigned long *)&ftemp));
		    ftemp = (float)olddbRound((double)ftemp,DBC_FldDes -> promptflag);
		    *((unsigned long *)&ftemp) = htonl(*((unsigned long *)&ftemp));
		    bptr = (char *)&ftemp;
		    size = sizeof(float);
		    break;

		case DBF_DOUBLE:
		    ntohd(*(double *)bptr, dtemp);	/* convert back to local float rep. MIPSEL ONLY */
		    dtemp = olddbRound (dtemp, DBC_FldDes -> promptflag);
		    bptr = (char *)&dtemp;
		    htond(dtemp, dtemp);
		    size = sizeof(double);
		    break;

		case DBF_SHORT:
		case DBF_USHORT:
		case DBF_ENUM:
		case DBF_GBLCHOICE:
		case DBF_CVTCHOICE:
		case DBF_RECCHOICE:
		case DBF_DEVCHOICE:
		    break;

		case DBF_LONG:
		case DBF_ULONG:
		    break;

		default:
		    break;
	    }
	    crc = DBC_MakeCRC(bptr,size,crc);
	}
	field_index++;
    }
    return(crc);
}
/*
	This function will olddbRound to 'num' digits
*/
double olddbRound(x,num)
double	x;
int	num;
{
    int	trunc;

    if (x < 0)
	trunc = (int)(x * place[num] - 0.5);
    else
	trunc = (int)(x * place[num] + 0.5);
    return (((double) trunc) / place[num]);
}
#endif vxWorks
