/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/************************************************************************/
/*									*/
/*			IOC DATABASE SERVER TASK			*/
/*									*/
/*	This task performs several database services on behalf of the	*/
/*	IOC. Listed in chronological order of development, they are:	*/
/*									*/
/*	Initial implementation:						*/
/*		1.	"On the fly" construction of the resource table	*/
/*		2.	Connection managment with redundant partners	*/
/*		3.	Update of prompted data fields between partners	*/
/*		4.	Synchronization between partners.		*/
/*									*/
/************************************************************************/

/****** MOD LOG
01/08/92	GMU	410-581-0422	uchenick@tate.com
		Original entry
******/

/************************************************************************/
/*									*/
/*			INCLUDE FILES					*/
/*									*/
/************************************************************************/

#include <vxWorks.h>
#include <symLib.h>
#include <configAll.h>		/* For getting address of the boot line */
#include <sockLib.h>
#include <sys/socket.h>
#include <in.h>
#include <net/if.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <rngLib.h>
#include <bootLib.h>
#include <rebootLib.h>
#include <netinet/tcp.h>
#include <tickLib.h>
#include <sysLib.h>
#include <wdLib.h>
#include <selectLib.h>
#include <hostLib.h>
#include <inetLib.h>
#include <semLib.h>
#include <ioLib.h>
#include <logLib.h>
#include <envLib.h>

#include "inet_ntoa.h"
#include "restable.h"
#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#include "task_params.h"	/* For getting task priorities */
#include "dbRecords.h"
#include "redmsgs.h"
#include "systime.h"
#include "redflds.h"
#include "dbRecType.h"
#include "redque.h"
#include "special.h"
#include "dbAccess.h"
#include "recSup.h"
#include "sdrHeader.h"
#include "dbserver.h"
#include "alarm.h"
#include "scanLock.h"
#include "bcopy.h"
#include "callbacks.h"
#include "../db/dbScan.h"
#include "pvd.h"
#include "redqlib.h"
#include "dbcSub.h"
#include "sdrLoad.h"
#include "dbPvd.h"
#include "redund.h"

#include "tisByteOrder.h"
#ifdef PRIV_AUDIT
/* for the auditing and priviledge functions */
#include "sslShared.h"
#include "iocPriv.h"
#include "auditSystem.h"
extern AUDIT_ID	auditLog;	/* the system log */
extern int RunScript (char *scriptFile);

static char X509_dn [X509_MAX_DN]; /* distinguished name from certificate */
#endif PRIV_AUDIT

/************************************************************************/
/*									*/
/*			GLOBAL DATA WE DEFINE				*/
/*									*/
/************************************************************************/
/*#define	CHECK_REFS*/
/*#define		OLD_SDR 1*/
RES_TABLE	*resTableHead = 0;
#define		LOCAL_MEM_LOCAL_ADRS	0	/* vxWorks hid this too deep */
struct fldDes	***papRedFldList;
RING_ID		redundQueue;
FAST_LOCK	dbsrvLock;
int		redQueueSize = REDQ_KBYTES * 1024;
SEM_ID		dbServerSem;
SEM_ID		oldHoldSem;
struct fldDes	**valList;	/* Pointers for VAL descriptors */
extern SYMTAB_ID sysSymTbl;
struct in_addr	dbEmulIP;
int		startTime;	/* Time at download completion GCT 2Oct1998 */


/************************************************************************/
/*									*/
/*			GLOBAL DATA WE REFERENCE			*/
/*									*/
/************************************************************************/

extern (*adjFun[])();

/************************************************************************/
/*									*/
/*			TUNING PARAMETERS				*/
/*									*/
/************************************************************************/

#define RESCHED_TICKS	  30	/* # ticks between runs */
#define MAX_DISCONN	 300	/* Must connect within 5 secs or start over */
#define HEARTBEAT_RATE    60	/* Send heartbeat if line idle for 1 second */
#define CARDIAC_ARREST	 600	/* Line is dead if no activity 10 seconds */
#define	MAX_UDP_MSG	1024	/* # bytes in biggest msg, UDP protocol */
#define MAX_TCP_MSG	2048	/* # bytes in biggest msg, TCP protocol */
#define DLD_DELAY	 120	/* dbClient responds in 120 secs or we start over */
#define SYNC_TRAVERSE	  15	/* Traverse sync database in 15 minutes */
#define CHKSUM_HOLDOFF	   5	/* minutes to wait before first checksum broadcast GCT 2Oct1998 */


/************************************************************************/
/*									*/
/*			OUR PRIVATE DATA				*/
/*									*/
/************************************************************************/
struct olAdd
	{
	short		recType;
	short		junk;		/* Force good alignment */
	struct recChain	newRec;
	};

/* NOTE:struct dbClient *clientList was defined here, since 
 *	this .h included in iocInit, linker gave an already 
 *	defined error def was moved to dbserver.c where it belongs
 */
struct dbClient
	{
	struct in_addr	clientInet;
	int		clientFd;
	struct olAdd	*firstOlAdd;
	struct dbClient	*next;
	};

struct dbClient		*clientList = 0;
static int		bcastSocket;	/* fd for our UDP socket */
static int		tcpListenSocket; /* listen for kick_TCP */
static fd_set		bcastReadFds;
				/* fd mask for select(), UDP socket only */
static fd_set		tcpFds;
				/* fd mask for all TCP sockets */
static SEM_ID		ioSem;
static struct timeval	selectPoll = {0, 0};
				/* Return immediately from select() */
struct in_addr		ourInetAddr;	/* Our Internet address */
static struct in_addr	bcastAddress;	/* The broadcast address for the IOC */
static int		maxTcpFd = -1;
static int		resTrace = 0;
static struct fldDes	**rocbList;	/* Pointers for ROCB descriptors */
static int		currTcpFd;
static char		tcpBuff[MAX_TCP_MSG];
static char		dbclientBuff[MAX_TCP_MSG];
static int		currTcpSize = 0;
static int		syncTime;	/* Time last sync message sent */
static int		totPnts;	/* Total points in the database */
static int		syncDelta;	/* # Ticks between sync messages */
static int		syncPts;	/* # Points per sync message */
static int		syncType;	/* Last point type sync'd */
static struct recChain	*syncNext;	/* Next point to be sync'd */
static int		ioFlag = 0;	/* 0 means I/O task is selecting */
static char		traceName[40] = "";
static char		rxTank[4096];	/* Place to receive redundant updates */
static short		nchib = 0;	/* # bytes active in rxTank */
static short		tankOutPos;	/* Next output position in rxTank */
static short		firstSelDone;	/* Avoids duplicate selects */
void	(*dbActivateFunc)();
void	(*dbPutRecordFunc)();
void	(*dbDelRecordFunc)();
void	(*dbGetRecordFunc)();

/************************************************************************/
/*									*/
/*			FUNCTION PROTOTYPES				*/
/*									*/
/************************************************************************/

/* this is in ../ca/iocinf.h, but that file needs other ca stuff
 * so I declare it here as extern, ems
*/
extern int  local_addr (int s, struct sockaddr_in *plcladdr);

int	dbGetLocalClient (void);
int	dbGetRemoteClient (int bcast);
void	deleteOlAdds (struct dbClient *client);

void	closeTcpStream (int fd);
int	dbDBDownload (int fd,void (*e2putfunc)());
int	dbSDRDownload (int fd);
int	dbServerTask (void);
int	flushTcp (void);
int	getPartnerData (int fd);
int	npendSend (int fd, char *buff, int nbytes);
int	recvPartner (int fd, char *buffer, int msgSize);
int	recvTcp (int fd, char *buff, int nbytes);
void	redundIO (void);
void rejectDeletion(short code,int from,int origFd,short refs);
int	sendClients (char *buffer,int length);
int	sendPartner (int fd, char *buffer, int msgSize);
int	sendToClients (char *buffer,int length);
int	tcpErr (int fd);
void	tis_fd_clr (int fd, fd_set *fdMask);
void	tis_fd_set (int fd, fd_set *fdMask);

/************************************************************************/
/*									*/
/*				INITIALIZE				*/
/*									*/
/************************************************************************/

int dbServerInit(void)
{
int localFd;
int remoteFd = -1;
int downloadFd;
SYM_TYPE	symtemp;
struct sockaddr_in setPort;

#ifdef PRIV_AUDIT
    sslSharedInit ();
#endif PRIV_AUDIT

	FD_ZERO(&tcpFds);
/******
	Create a socket for sending and receiving connection messages.
	This socket is UDP protocol with broadcasts enabled. Also bind
	it to the well known port number for dbServer.
******/
#ifdef USE_UDP
	if ((bcastSocket = socket(AF_INET, SOCK_DGRAM, 0)) == ERROR)
		{
		logMsg ("Can't create UDP socket: %d\n", errno,0,0,0,0,0);
		return (-1);
		}

	/*
	The following code segment uses a trick that will be employed
	many times in this task. When a chunk of memory is needed for
	a code segment and is not needed later on, I put that code segment
	and its temporary variables within braces. This makes those temporary
	variables disappear after the closing brace, thereby allowing
	the memory that they occupy to be reused by later code segments
	that pull the same trick. This is a habit that I picked up back
	in the bad old days when memory was an extremely scarce resource.
	*/

	{
	int optionValue = 1;		/* The temporary variable */
	if (setsockopt(bcastSocket, SOL_SOCKET, SO_BROADCAST,
		     (caddr_t) &optionValue, sizeof(optionValue)) == ERROR)
		{
		logMsg ("Can't set broadcast option: %d\n", errno,0,0,0,0,0);
		return (-1);
		}
	}
	{
	setPort.sin_family = AF_INET;
	setPort.sin_addr.s_addr = htonl(INADDR_ANY);
	setPort.sin_port = htons(DBSERVER_UDP_PORT);
	if (bind(bcastSocket, (struct sockaddr *) &setPort, sizeof(setPort))
							== ERROR)
		{
		logMsg ("Can't bind dbServer's UDP port: %d\n", errno,0,0,0,0,0);
		return (-1);
		}	
	}
/******
	We will be using the select() function to get the status of only
	the UDP port in a single call. Set up the masks that we will need
	to make that call. Also clear the fd mask for TCP ports because
	none have been created yet.
******/

	FD_ZERO(&bcastReadFds);
	tis_fd_set(bcastSocket, &bcastReadFds);

    /*
		Create a TCP port to listen host connections.
    */
#else USE_UDP
    tcpListenSocket = socket(AF_INET, SOCK_STREAM, 0);
    setPort.sin_family = AF_INET;
    setPort.sin_addr.s_addr = htonl(INADDR_ANY);
    setPort.sin_port = htons(8592);
    if (bind(tcpListenSocket, (struct sockaddr *) &setPort, sizeof(setPort)) == ERROR)
		perror("Can't bind dbserver listen TCP port");
    listen(tcpListenSocket, 2);
    FD_SET(tcpListenSocket, &tcpFds);
    if (tcpListenSocket > maxTcpFd)
		maxTcpFd = tcpListenSocket;
#endif USE_UDP
/******
	Now get our Internet address. We will need that to compare with the
	other end of TCP connections to see who should accept() and who
	should connect()
******/
	{
	struct sockaddr_in our_inet;	

	local_addr(bcastSocket, &our_inet);
	ourInetAddr = our_inet.sin_addr;
	}

/******
	Now get our broadcast address. Find first ifnet supporting broadcast
	and use that for broadcast.
******/

	{
	struct ifnet *lkup_net;
	extern struct ifnet * ifnet;

	for (lkup_net = ifnet; lkup_net; lkup_net = lkup_net->if_next)
	    {
	    if ((lkup_net->if_flags & IFF_BROADCAST) == IFF_BROADCAST)
	 	{
		bcastAddress = ((struct sockaddr_in *)&(lkup_net->if_addrlist->ifa_broadaddr))->sin_addr;
		break;	/* only accept first one found */
		}
	    }
	logMsg ("Using %s as broadcast address\n", (int)INET_NTOA(bcastAddress),0,0,0,0,0);
	}

/******
	Start the download process. This is written such that it can
	be started over and over again in case the link between us and
	dbClient gets severed or if he crashes in the middle of this.
******/
/*
	Get address of functions to put new records into EEROM database.
	this will be zero on systems that do not support the local
	database.
*/
symFindByName(sysSymTbl,"_ActivateE2DB",(char **)&dbActivateFunc,&symtemp);
symFindByName(sysSymTbl,"_PutRecord",(char **)&dbPutRecordFunc,&symtemp);
symFindByName(sysSymTbl,"_DelRecord",(char **)&dbDelRecordFunc,&symtemp);
symFindByName(sysSymTbl,"_GetRecord",(char **)&dbGetRecordFunc,&symtemp);
for (;;)
	{
	downloadFd = -1;
	while (downloadFd == -1)
	    {
	    if((localFd = dbGetLocalClient()) < 0)		/* try for local client (dbEmulator) */
		{
		remoteFd = dbGetRemoteClient(bcastSocket);	/* if absent then try remote client (dbClient)*/
		downloadFd = remoteFd;				/* and use it */
		}
	    else
		downloadFd = localFd;
	    }

	if(dbSDRDownload(downloadFd) < 0)			/* ask for SDR table from client */
		continue;
/*
	If we cannot allocate memory for the pvd table, it's not much
	use in going on, pvd_create will give a malloc msg
	   ems 03-08-94
*/
	if (pvd_create())
	    exit(-1);
/*
	Now load down the database records
*/
	totPnts = 0;
	for (;;)
		{
		totPnts = dbDBDownload(downloadFd, (downloadFd == localFd ? 0 : dbPutRecordFunc));
		if ((totPnts != 0) && (totPnts != ERROR))
			break;

		/*
		Remember that the fd we used is possibly GONE!
		*/

		if (totPnts == ERROR)
			{
			struct dbClient		*currClient;

			currClient = clientList;
			while (currClient)
				{
				if (currClient -> clientFd == downloadFd)
					{
					currClient -> clientFd = -1;
					break;
					}
				currClient = currClient -> next;
				}
			}

		/*
		Now, go get another one.
		*/
		while (remoteFd == -1)
		    {
		    remoteFd = dbGetRemoteClient(bcastSocket);	/* if absent then try remote client (dbClient)*/
		    }
		downloadFd = remoteFd;
		localFd = -1;
		if(dbRecDes == 0)
			{
			if(dbSDRDownload(downloadFd) < 0)
				break;
			}
		}
	
	if ((totPnts != 0) && (totPnts != ERROR))				/* got some points then we're done */
		break;


	}/* end of main for(;;) loop */

	if(downloadFd == remoteFd && dbActivateFunc != 0)
		dbActivateFunc();
/*
	Done loading. Calculate the database synchronization parameters.
	2Oct1998 Get time to be used for all network traffic delays GCT
*/
	startTime = tickGet();
#ifdef USE_UDP
	if (bcastSocket > maxTcpFd)
		maxTcpFd = bcastSocket;
#endif
	if (totPnts)
		{
		/*
			Prep syncPts to tell us the maximum number of points in each CRC transmission. For
			3-May-1999 this will be the number sent in a TCP message(NO Broadcast!). The
			DBC_CRCINFO structure is 12 bytes and is found in redmsgs.h
		*/
		syncPts = (sizeof(tcpBuff) - sizeof(short) - sizeof(short)) / sizeof(DBC_CRCINFO);
		if (syncPts > totPnts)
			syncPts = totPnts;
		/*syncDelta = SYNC_TRAVERSE * 60 * 60 / ((totPnts + syncPts - 1) / syncPts);*/
		syncDelta = CHKSUM_HOLDOFF * 60 * 60;		/* 5 minute wait before CRC broadcast */
		syncType = dbRecType -> number + 1;
		syncNext = 0;
		}


/******
	Now create the resource table by making a pass over the database.
******/

	{
	int		type;
	RES_TABLE	*tableHead = resTableHead;
	
	rocbList = (struct fldDes **) calloc(dbRecords -> number, sizeof(struct fldDes *));
	valList = (struct fldDes **) calloc(dbRecords -> number, sizeof(struct fldDes *));

	for (type = 0; type < dbRecords -> number; type++)
		{
		struct recLoc	*precLoc;
		struct recChain	*pchain;
		int		recNum;

		/*
		While we are here looping along by point type, see if
		this point type has an ROCI field. If it does, then
		get the address of the ROCB and VAL descriptors, also.
		*/

		if (pvdGetFld(type, "ROCI") != NULL)
			{
			rocbList[type] = pvdGetFld(type, "ROCB");
			}
		valList[type] = pvdGetFld(type, "VAL ");

		if (!(precLoc = GET_PRECLOC(type)))
			continue;
		for (recNum = 0, pchain = (struct recChain *) (precLoc -> pFirst); 
				recNum < precLoc -> no_records; recNum++, pchain = pchain -> next)
			{
			struct dbCommon	*recHeader;
			RES_TABLE	*resTbl;
			
			recHeader = (struct dbCommon *) pchain -> pRecord;
			if (recHeader -> name[0] == 0)
				continue;	/* Empty record */
			
			/*
			Calculate the CRC of the record and store it in the record itself
			*/

			recHeader -> rcrc = DBC_CRCHostRecord((char *) recHeader, type);

			/*
			We can't use resGetByName because we don't want to set
			resTableHead until the entire table has been constructed. The
			reason for not setting that is we don't want to be fooling
			around with links while other tasks might be trying to follow
			them. Otherwise, we'd have to get involved with locking the
			table and that is more trouble than it's worth at this time.
			*/

			resTbl = tableHead;
			for (;;)
				{
				if (!resTbl)
					{
					tableHead = (RES_TABLE *)calloc(1, sizeof(RES_TABLE));
					resTbl = tableHead;
					}
				else
					{
					if (strcmp(recHeader -> resn, resTbl -> resourceName))
						{
						if (resTbl -> link)
							{
							resTbl = resTbl -> link;
							continue;
							}
						else
							{
							resTbl -> link = (RES_TABLE *) calloc(1, sizeof(RES_TABLE));
							resTbl = resTbl -> link;
							}
						}
					else
						break;
					}
				
				strcpy(resTbl -> resourceName, recHeader -> resn);
				break;
				}
			recHeader -> radr = resTbl;
			}
		}
	resTableHead = tableHead;	/* Only set after we're all done! */
	}


/******
	Now create the memory queue that other tasks will use to tell us
	that they have modified the database.
******/

	if ((redundQueue = rngCreate(redQueueSize)) == NULL)
		{
		logMsg ("Can't create update queue, errno = %d.\n", errno,0,0,0,0,0);
		return (-1);
		}
	FASTLOCKINIT(&dbsrvLock);

/******
	Create the semaphores for dbServer and his I/O slave.
******/

	if ((dbServerSem = semBCreate(SEM_Q_PRIORITY, SEM_EMPTY)) == NULL)
		{
		logMsg ("Can't create dbServerSem.\n",0,0,0,0,0,0);
		return (-1);
		}
	if ((ioSem = semBCreate(SEM_Q_PRIORITY, SEM_EMPTY)) == NULL)
		{
		logMsg ("Can't create ioSem.\n",0,0,0,0,0,0);
		return (-1);
		}

/******
	Create the semaphore for mutual exclusion with initUpdate task
******/

	if ((oldHoldSem = semMCreate(SEM_Q_FIFO)) == NULL)
		{
		logMsg ("Can't create holdoff semaphore\n",0,0,0,0,0,0);
		return(-1);
		}

/******
	Done except for spawning off the two tasks dedicated to
	redundancy. Don't do that yet until we have had a chance
	to initialize the database.
******/

	return(0);
}

int startRedundancyTasks(void)
{
/******
	Drain the UDP socket. There's nothing there of any use.
******/

#ifdef USE_UDP	
	for (;;)
		{
		fd_set	ourFd;
		int	fromSize;
		int	udpLength;
		struct sockaddr findClient;
		bcopy((char *) &bcastReadFds, (char *) &ourFd, sizeof(ourFd));
		if (select(bcastSocket + 1, &ourFd, NULL, NULL, &selectPoll) <= 0)
			break;
		fromSize = sizeof(findClient);
		if ((udpLength = recvfrom(bcastSocket, (char *) &tcpBuff, sizeof(tcpBuff),
				0, &findClient, &fromSize)) == ERROR)
			break;
		if (!udpLength)
			{
			logMsg ("Zero length UDP packet while draining socket.\n",0,0,0,0,0,0);
			}
		}
#endif
/******
	Spawn the I/O server task
******/

	{
	if (taskSpawn("redundIO", DBSERVER_PRI - 1, VX_FP_TASK, 2000, 
		      (FUNCPTR)redundIO,0,0,0,0,0,0,0,0,0,0) == ERROR)
		{
		logMsg ("Can't spawn redundIO task\n",0,0,0,0,0,0);
		return (-1);
		}
	}

/******
	Finally, spawn the actual dbServer task.
******/

	if (taskSpawn(DBSERVER_NAME, DBSERVER_PRI, VX_FP_TASK, 12000, 
		      (FUNCPTR)dbServerTask,0,0,0,0,0,0,0,0,0,0) == ERROR)
		{
		logMsg ("Can't spawn dbServerTask.\n",0,0,0,0,0,0);
		return (-1);
		}

/******
	Initialization complete
******/

	return(0);
}

/****************************************************************/
/*								*/
/*			THE DATABASE SERVER TASK		*/
/*								*/
/*	This is the database server task. It expects to run	*/
/*	in conjunction with the periodic scan task, so they	*/
/*	are set up to run at the same priority, at least	*/
/*	for the time being.					*/
/*								*/
/****************************************************************/

int dbServerTask(void)
{
	long	lastTable = 0;

	syncTime = tickGet();

for (;;)
	{
/******
	Check the UDP port for an incoming message of any type
******/
#ifdef USE_UDP
if (ioFlag)
	{
	fd_set readFds;
	bcopy((char *) &bcastReadFds, (char *) &readFds, sizeof(bcastReadFds));
	while (select(bcastSocket + 1, &readFds, NULL, NULL, &selectPoll) > 0)
		{

		/*
		Message is pending. Read one word so we know what type
		it is. That will tell us what to do next.
		*/

		short			msgType;
		char			*udpBuff = tcpBuff;
		struct sockaddr_in	from;
		int			fromSize = sizeof(from);
		int			udpLength;

		if ((udpLength = recvfrom(bcastSocket, udpBuff, sizeof(tcpBuff), 0,
				(struct sockaddr *)&from, &fromSize)) == ERROR)
			logMsg ("Read error on UDP socket: \n", errno,0,0,0,0,0);
		else
			{
			if (!udpLength)
				{
				logMsg ("Zero length UDP message from 0x%X\n", 
					 htonl(from.sin_addr.s_addr),0,0,0,0,0);
				continue;
				}

			/*
			On a broadcast, we can hear ourselves. Drop any messages
			that we originated.
			*/

			if (from.sin_addr.s_addr == ourInetAddr.s_addr)
				continue;
			msgType = ntohs(*((short *) udpBuff));
			if (msgType == RED_CONNREQ)
				{

				/********************************/
				/*				*/
				/*	CONNECTION REQUEST	*/
				/*				*/
				/********************************/

				RED_CONNREQ_HDR	*reqHdr;
				RED_CONNREQ_REQ *reqList;
				int		numReplies = 0;
				struct
					{
					short		    replyType;
					RED_CONNREPLY_HDR   replyHdr;
					RED_CONNREPLY_REPLY replyList
							[(MAX_UDP_MSG
							- sizeof(msgType)
							- sizeof(RED_CONNREQ_HDR))
							/ sizeof(RED_CONNREQ_REQ)];
					} replyMsg;

				reqHdr = (RED_CONNREQ_HDR *) (udpBuff + sizeof(msgType));
				reqList = (RED_CONNREQ_REQ *) (udpBuff + sizeof(msgType) + sizeof(RED_CONNREQ_HDR));
if (resTrace)
	logMsg ("Connection request, %d names\n", htons(reqHdr -> redNumRequests),0,0,0,0,0);

				/*
				Try to find each resource in the list of names
				that was sent to us, putting our information
				into the message as we go along. When we do find
				a match, also put his port number into our table.
				*/

				reqHdr -> redNumRequests = ntohs(reqHdr -> redNumRequests);
				for (; reqHdr -> redNumRequests--; reqList++)
					{
					RES_TABLE	*resEntry;
					if (resGetByName(reqList -> redReqName,
							&resEntry) == RES_OK)
						{
						if (resEntry -> driverPort == 0)
							continue;
						strcpy(replyMsg.replyList[numReplies].redReplyName,
								resEntry -> resourceName);
						replyMsg.replyList[numReplies].redReplyPort =
							htons(resEntry -> driverPort);
						resEntry -> partnerInet = from.sin_addr;
						resEntry -> partnerPort =
							ntohs(reqList -> redReqPort);
						numReplies++;
						}
					}

				/*
				Send the reply message, if there's anything in it.
				*/

				if (numReplies)
					{
#ifdef USE_UDP
					int	errStat;
#endif
if (resTrace)
	logMsg ("...sent reply for %d resources\n", numReplies,0,0,0,0,0);
					replyMsg.replyType = htons(RED_CONNREPLY);
					replyMsg.replyHdr.redNumReplies = htons(numReplies);
#ifdef USE_UDP
					errStat =
					sendto(bcastSocket, (char *) &replyMsg,
						sizeof(replyMsg.replyType)
							+ sizeof(RED_CONNREPLY_HDR)
							+ sizeof(RED_CONNREPLY_REPLY) * numReplies,
						0,
						(struct sockaddr *)&from,
						fromSize);
					if (errStat == ERROR)
						logMsg ("Couldn't reply, errno = %d\n", 
						         errno,0,0,0,0,0);
#endif
					}
				}
			else if (msgType == RED_CONNREPLY)
				{

				/********************************/
				/*				*/
				/*	CONNECTION REPLY	*/
				/*				*/
				/********************************/

				RED_CONNREPLY_HDR	*replyHdr;
				RED_CONNREPLY_REPLY	*replyList;

				/*
				This message was directed specifically to us. Therefore, we
				should find every resource in it. Processing of this message
				consists of posting the TCP port number on the partner machine
				into our resource table. We also clear a bunch of structure
				members that will cause the dbServer <-> dbServer connection
				to get established during a later scan of the table.

				An interesting complication happens when this message is a
				duplicate of one we have already seen. In that case, ignore it.
				This situation may have occurred because many request messages
				may have been buffered on the other node.
				*/

				replyHdr = (RED_CONNREPLY_HDR *) (udpBuff + sizeof(msgType));
				replyList = (RED_CONNREPLY_REPLY *) (udpBuff + sizeof(msgType)
						    + sizeof(RED_CONNREPLY_HDR));
				replyHdr -> redNumReplies = ntohs(replyHdr -> redNumReplies);
if (resTrace)
	logMsg ("Connection Reply, %d names.\n", replyHdr -> redNumReplies,0,0,0,0,0);
				
				for (; replyHdr -> redNumReplies--; replyList++)
					{
					RES_TABLE	*resEntry;

					if (resGetByName(replyList -> redReplyName,
							 &resEntry) == RES_OK)
						{
						if (resEntry -> partnerInet.s_addr == from.sin_addr.s_addr
						 && resEntry -> partnerPort == ntohs(replyList -> redReplyPort))
							continue;
						resEntry -> partnerInet = from.sin_addr;
						resEntry -> partnerPort =
							ntohs(replyList -> redReplyPort);
#if 0
	Take out the next three lines. This is so the resMgr can recover
	when he trashes his connection but we dont trash ours.

						resEntry -> localDbPort = 0;
						resEntry -> dbServerPort = 0;
						resEntry -> dbServerFd = 0;
#endif
						}
					else
						logMsg ("Can't find resource %s\n",
							 (int)(replyList -> redReplyName),0,0,0,0,0);
					}
				}
			else if (msgType == RED_IOCDBREQ)
				{

				/****************************************/
				/*					*/
				/*	DBSERVER CONNECTION REQUEST	*/
				/*					*/
				/****************************************/

				/*
				Processing for this message consists of finding each
				table row that has the same partner Internet address
				as the message originator's. For each such entry, we
				copy the TCP port number from the message to the table.
				*/

				RES_TABLE		*resEntry = resTableHead;
				RED_IOCLINKREQ		*linkReply;

if (resTrace)
	logMsg ("dbServer Connection Request.\n",0,0,0,0,0,0);
				linkReply = (RED_IOCLINKREQ *) (udpBuff + sizeof(msgType));

				while (resEntry)
					{
					if (resEntry -> partnerInet.s_addr == from.sin_addr.s_addr)
						{
						resEntry -> dbServerPort = ntohs(linkReply ->
								redDbServerPort);
						}
					resEntry = resEntry -> link;
					}
				}
			else if (msgType == DBC_MSG_WAKEUP)
				{
				/*
				Another dbClient wants to register with us.
				*/
				struct dbClient	*currClient;
				int		newFd;
				int		newEntry = 0;
#ifdef PRIV_AUDIT
				SSL		*newSSL;
#endif PRIV_AUDIT

				currClient = clientList;
				while (currClient)
					{
					if (currClient -> clientInet.s_addr == from.sin_addr.s_addr)
						break;
					if (currClient -> next)
						currClient = currClient -> next;
					else
						{
						currClient -> next = (struct dbClient *)
							calloc(1, sizeof(struct dbClient));
						currClient = currClient -> next;
						currClient -> clientInet =
							from.sin_addr;
						newEntry = 1;
						break;
						}
					}
				if (currClient -> clientFd)
					{
					continue;
					}
#ifdef PRIV_AUDIT
				newSSL = connect_ssl_connection_timed (from.sin_addr.s_addr,
							ntohs(from.sin_port), 15/*seconds*/);
				if (newSSL)
				    {
				    GetPeerCN (newSSL, X509_dn, sizeof(X509_dn));
				    if (GetGeneralPrivHost (X509_dn, "DB_") < 1)					
					{
					GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
					/* No priviledge to connect! */
					logMsg("DB priviledge error, %s not authorized\n",
						(int)X509_dn,0,0,0,0,0);
					SecurityAlarm (X509_dn, "DB priviledge error, Not authorized", NULL);
					if (auditLog)
					    {
					    char msg [80];

					    sprintf (msg, "Client privilege error from %s", INET_NTOA(from.sin_addr));
					    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_FAIL, X509_dn,
						    taskName(taskIdSelf()), msg);
					    }
					sslSharedCloseConn (newSSL);
					}
				    else
					{
					GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
					if (auditLog)
					    {
					    char msg [80];

					    sprintf (msg, "Client connected from %s", INET_NTOA(from.sin_addr));
					    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_SUCCESS, X509_dn,
						    taskName(taskIdSelf()), msg);
					    }
					newFd = SSL_get_fd (newSSL);
					currClient -> clientFd = newFd;
					tis_fd_set(newFd, &tcpFds);
					if (newFd > maxTcpFd)
					    maxTcpFd = newFd;
					if (newEntry)
					    {
					    msgType = htons(DBC_MSG_SYMPURGE);
					    if (npendSend(newFd, (char *) &msgType,
							    sizeof(msgType)) == ERROR)
						    tcpErr(newFd);
					    }
					}
				    }
#else PRIV_AUDIT
				if ((newFd = socket(AF_INET, SOCK_STREAM,0)) != ERROR)
					{
					int	true = 1;

					if (connect(newFd, (struct sockaddr *)&from, sizeof(from)) != ERROR)
						{
						ioctl(newFd, FIONBIO, (int)&true);
						currClient -> clientFd = newFd;
						tis_fd_set(newFd, &tcpFds);
						if (newFd > maxTcpFd)
							maxTcpFd = newFd;
						if (newEntry)
							{
							msgType = htons(DBC_MSG_SYMPURGE);
							if (npendSend(newFd, (char *) &msgType,
									sizeof(msgType)) == ERROR)
								tcpErr(newFd);
							}
						}
					else
						close(newFd);
					}
#endif PRIV_AUDIT
				}
							
			else if (msgType == DBC_MSG_CRCRESPONSE)
				{
				/*
				Some other IOC is broadcasting CRCs. Ignore this.
				*/
				}
			else if (msgType == DBC_MSG_AWAKE)
				{
				/*
				Some other IOC just woke up. Ignore this. dbClient will handle it.
				*/
				}
			else if (msgType == RED_UPDREQ)
				{
				/*
				Slave wants an update for this point. Do this only if
				we can find the record and we are master. Call redQNPSend
				to send all the fields except the prompted fields, sending
				prompted fields even though they have not changed, causes
				a broadcast to all dbClients for these fields, If by chance
				they have changed and was not caught by the scan task, the
				slave will see a check sum mismatch and request all fields
				mods 01-20-95 ems
				*/

				RED_ASKUPDT	*updtMsg;
				short		recType;
				short		recNum;
				struct recLoc	*precLoc;
				struct dbCommon	*recHdr = 0;

				updtMsg = (RED_ASKUPDT *) (udpBuff + sizeof(msgType));
if (resTrace)
	logMsg ("Redundant update requested by slave for %s\n", 
		 (int)(updtMsg -> redPointName),0,0,0,0,0);
				if (pvdGetRec(updtMsg -> redPointName, &recType, &recNum) < 0)
					continue;	/* No such point */
				precLoc = GET_PRECLOC(recType);
				pvdGetAddr(updtMsg -> redPointName,(char **)&recHdr);
				resGetCaAnswer(recHdr -> radr, &recNum);
				if (!recNum)
					continue;	/* Don't own the resource */
				redQInit(recHdr, tcpBuff, sizeof(tcpBuff), recType);
				redQNPSend(tcpBuff);
				}
			else
				{
				logMsg ("Received unknown msg type %d on UDP port.\n", 
					 msgType,0,0,0,0,0);
				}
			}
		}
	}
#endif USE_UDP
/******
	Check for I/O completion on any of our TCP connections
******/

if (ioFlag)
	{
	fd_set	readFds;
	int	numFds;
	bcopy((char *) &tcpFds, (char *) &readFds, sizeof(tcpFds));

	while ((numFds = select(maxTcpFd + 1, &readFds, NULL, NULL, &selectPoll)) > 0)
		{
		int	fdIndex;
#ifndef USE_UDP
		/*
			GCT 11-May-1999

			dbClient is trying to connect to to us by connecting on this port/socket. If
			I get the connection then close the socket and do the real connect myself.
		*/
		if(FD_ISSET(tcpListenSocket,&readFds))
		{
		/*
			Another dbClient wants to register with us.
		*/
			struct dbClient	*currClient;
			int		newFd;
			int		newEntry = 0;
			struct sockaddr_in from;
			int fromLength = sizeof(from);
			short msgType;
#ifdef PRIV_AUDIT
			SSL		*newSSL;
#endif PRIV_AUDIT
			currClient = clientList;
			newFd = accept(tcpListenSocket, (struct sockaddr *)&from,&fromLength);
			logMsg("dbClient has connect indication on %d\n",(int)newFd,0,0,0,0,0);
			while (currClient)
				{
				if (currClient -> clientInet.s_addr == from.sin_addr.s_addr)
					break;
				if (currClient -> next)
					currClient = currClient -> next;
				else
					{
					currClient -> next = (struct dbClient *)
						calloc(1, sizeof(struct dbClient));
					currClient = currClient -> next;
					currClient -> clientInet =
						from.sin_addr;
					newEntry = 1;
					break;
					}
				}
			close(newFd);
			if (currClient -> clientFd == 0) /* If already have a connection then skip this */
				{
#ifdef PRIV_AUDIT
			newSSL = connect_ssl_connection_timed (from.sin_addr.s_addr,
						DBSERVER_UDP_PORT, 10/*seconds*/);
			if (newSSL)
			    {
			    GetPeerCN (newSSL, X509_dn, sizeof(X509_dn));
			    if (GetGeneralPrivHost (X509_dn, "DB_") < 1)					
				{
					GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
					/* No priviledge to connect! */
					logMsg("DB priviledge error, %s not authorized\n",
						(int)X509_dn,0,0,0,0,0);
					SecurityAlarm (X509_dn, "DB priviledge error, Not authorized", NULL);
					if (auditLog)
					    {
					    char msg [80];

					    sprintf (msg, "Client privilege error from %s", INET_NTOA(from.sin_addr));
					    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_FAIL, X509_dn,
						    taskName(taskIdSelf()), msg);
					    }
					sslSharedCloseConn (newSSL);
					newSSL = NULL;
					}
				    else
					{
					GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
					if (auditLog)
					    {
					    char msg [80];

					    sprintf (msg, "Client connected from %s", INET_NTOA(from.sin_addr));
					    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_SUCCESS, X509_dn,
						    taskName(taskIdSelf()), msg);
					    }
					newFd = SSL_get_fd (newSSL);
					currClient -> clientFd = newFd;
					tis_fd_set(newFd, &tcpFds);
logMsg ("Client connected from %s, fd=%d, newSSL=%lx\n", (int)INET_NTOA(from.sin_addr), newFd, (int)newSSL, 0,0,0);
					if (newFd > maxTcpFd)
					    maxTcpFd = newFd;
					if (newEntry)
					    {
					    msgType = htons(DBC_MSG_SYMPURGE);
					    if (npendSend(newFd, (char *) &msgType,
							    sizeof(msgType)) == ERROR)
						    tcpErr(newFd);
					    }
					}
				    }
#else PRIV_AUDIT
				if ((newFd = socket(AF_INET, SOCK_STREAM,0)) != ERROR)
					{
					int	true = 1;

					from.sin_port = htons(DBSERVER_UDP_PORT);
					if (connect(newFd, (struct sockaddr *)&from, sizeof(from)) != ERROR)
						{
						ioctl(newFd, FIONBIO, (int)&true);
						currClient -> clientFd = newFd;
						tis_fd_set(newFd, &tcpFds);
						if (newFd > maxTcpFd)
							maxTcpFd = newFd;
						if (newEntry)
							{
							msgType = htons(DBC_MSG_SYMPURGE);
							if (npendSend(newFd, (char *) &msgType,
									sizeof(msgType)) == ERROR)
								tcpErr(newFd);
							}
						}
					else
						close(newFd);
					}
#endif PRIV_AUDIT
			}/* end of check for existing dbClient connection */
		}/* end of tcpListenSocket check */
#endif USE_UDP
		for (fdIndex = 0; fdIndex <= maxTcpFd; fdIndex++)
			{
			RES_TABLE	*resEntry;
			if (!FD_ISSET(fdIndex, &readFds) || (fdIndex == tcpListenSocket))
				continue;
			
			/*
			Message pending on this fd. Find any resource table entry with
			this fd in it to see if we are connected. If we aren't, then
			we have to issue the accept() call. If we are, then handle the
			message.
			*/
			if (resFindServerFd(fdIndex, &resEntry) == RES_OK)
				{
				if (!resEntry -> dbServerConnected)
					{
					
					/*
					Not connected. Issue the accept() call which
					gives us a new fd. Close the old one because
					we don't need it anymore. Then, propagate the
					new fd and the updated status of dbServerConnected
					throughout the rest of the table.
					*/

					int newFd;
					struct sockaddr_in connAddr;
					int connAddrLen = sizeof(connAddr);
					newFd = accept(resEntry -> dbServerFd, (struct sockaddr *)&connAddr,
									&connAddrLen);
					if (newFd > 0)
					    {
					    RES_TABLE 	*otherEntry;
					    int		true = 1;

#ifdef PRIV_AUDIT
					    SSL		*newConn;
					    if ((newConn = setup_ssl_accept (newFd)) != NULL)
						{
						if (accept_ssl_connection_timed (newConn, 10/*seconds*/))
						    {
						    /* validate user privilege */
						    GetPeerCN (newConn, X509_dn, sizeof(X509_dn));
						    if (newConn && (GetGeneralPrivHost (X509_dn, "DB_") < 1))
							{
							char msg [80];

							GetPeerDN (newConn, X509_dn, sizeof(X509_dn));
							sprintf (msg, "db connect privilege error from %s", INET_NTOA(connAddr.sin_addr));
							SecurityAlarm (X509_dn, msg, 0);
							if (auditLog)
							    {
							    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_FAIL, X509_dn,
								taskName(taskIdSelf()), msg);
							    }
							sslSharedCloseConn (newConn);
							return NULL;
							}
						    else
							{
							GetPeerDN (newConn, X509_dn, sizeof(X509_dn));
							if (auditLog)
							    {
							    char msg [80];

							    sprintf (msg, "db connect from %s.", INET_NTOA(connAddr.sin_addr));
							    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_SUCCESS, X509_dn,
									   taskName(taskIdSelf()), msg);
							    }
							ioctl(newFd, FIONBIO, (int)&true);
							close(resEntry -> dbServerFd);
							tis_fd_clr(resEntry -> dbServerFd, &tcpFds);
							tis_fd_set(newFd, &tcpFds);
							if (newFd > maxTcpFd)
							    maxTcpFd = newFd;
							otherEntry = 0;
							while (resFindEquivDbServer(resEntry, &otherEntry)
										== RES_OK)
							    {
							    otherEntry -> dbServerFd = newFd;
							    otherEntry -> dbServerConnected = 1;
							    otherEntry -> disconnectTime = 0;
							    otherEntry -> lastRxHeartbeat = tickGet();
							    }
							resEntry -> dbServerFd = newFd;
							resEntry -> dbServerConnected = 1;
							resEntry -> disconnectTime = 0;
if (resTrace)
    logMsg ("...Socket established by accept() with redundant partner\n",0,0,0,0,0,0);
							}
						    }
						else
						    {
						    printErr ("dbServer: error engaging SSL with %s\n", INET_NTOA(connAddr.sin_addr));
						    sslSharedCloseConn (newConn);
						    newConn = NULL;
						    }
						}
					    else
						{
						printErr ("dbServer: SSL build failed for %s - status = 0x%x\n",
							  INET_NTOA(connAddr.sin_addr), errno);
						close (newFd);
						}
#else PRIV_AUDIT
					    ioctl(newFd, FIONBIO, (int)&true);
					    close(resEntry -> dbServerFd);
					    tis_fd_clr(resEntry -> dbServerFd, &tcpFds);
					    tis_fd_set(newFd, &tcpFds);
					    if (newFd > maxTcpFd)
						maxTcpFd = newFd;
					    otherEntry = 0;
					    while (resFindEquivDbServer(resEntry, &otherEntry)
								    == RES_OK)
						{
						otherEntry -> dbServerFd = newFd;
						otherEntry -> dbServerConnected = 1;
						otherEntry -> disconnectTime = 0;
						otherEntry -> lastRxHeartbeat = tickGet();
						}
					    resEntry -> dbServerFd = newFd;
					    resEntry -> dbServerConnected = 1;
					    resEntry -> disconnectTime = 0;
if (resTrace)
    logMsg ("...Socket established by accept() with redundant partner\n",0,0,0,0,0,0);
#endif PRIV_AUDIT
					    }
					else
					    logMsg ("Accept failed, errno = %d\n", 
						     errno,0,0,0,0,0);
					}
				else
				    {
				    firstSelDone = 0;
				    if (getPartnerData(resEntry -> dbServerFd) == ERROR)
					{
if (resTrace)
	logMsg ("Error initial read from partner\n",0,0,0,0,0,0);
					continue;
					}
				    firstSelDone = 1;
				    while (nchib)
					{
					/*
					Must be a redundancy message, or an error signifying
					a disconnect. Read the message header, get the status,
					and then act accordingly.
					*/

					short msgType;
					if (recvPartner(resEntry -> dbServerFd,
							(char *) &msgType, sizeof(msgType)) == ERROR)
						{
if (resTrace)
	logMsg ("Error reading msgType from partner\n",0,0,0,0,0,0);
						continue;
						}
					
					if (msgType == RED_UPDATE)
						{

						/*
						Redundant update. Get the header so we get the
						number of fields and the common update data.
						*/

						char		name[PVNAME_SZ];
						RED_DBUPDATE	dbUpdate;
						short		size;
						short		offset;
						struct dbCommon	*recHdr;
						struct recLoc	*precLoc;
						short		recType;
						short		numFields;
						short		dumping = 0;

						if (recvPartner(resEntry -> dbServerFd,
								(char *) &dbUpdate,
								sizeof(dbUpdate)) == ERROR)
							{
if (resTrace)
	logMsg ("Error reading redundant update hdr from partner\n",0,0,0,0,0,0);
							continue;
							}
						if (recvPartner(resEntry -> dbServerFd,
								name,
								dbUpdate.nameLen) == ERROR)
							{
if (resTrace)
	logMsg ("Error reading redundant point name from partner\n",0,0,0,0,0,0);
							continue;
							}
						
						/*
						Translate point name to type and record number.
						*/

						if (resTrace && !strcmp(name, traceName))
							logMsg ("Update for %s\n", 
							         (int)traceName,0,0,0,0,0);

						if (traceName[0])
						    if (!strcmp(traceName, name))
							{
							logMsg ("RUPDT for %s\n", 
								 (int)name,0,0,0,0,0);
							dumping = 1;
							}

						if (pvdGetRec(name, &recType, &offset) < 0)
							{
							if (resTrace)
							    logMsg ("Can't find %s in PVD.\n",
								     (int)name,0,0,0,0,0);
							/*
							Drain out the rest of the message.
							*/

							while (dbUpdate.numFields--)
								{
								char buff[80];
								short nextra;
								if (recvPartner(resEntry -> dbServerFd,
										(char *) &size,
										sizeof(short)) == ERROR)
									{
if (resTrace)
	logMsg ("Error reading size during message discard from partner\n",0,0,0,0,0,0);
									break;
									}
								if (size & ROCI_UPDATE)
									nextra = 3;
								else if (size & SPECIAL_PROC)
									nextra = 1;
								else
									nextra = 0;
								while (nextra--)
									{
									if (recvPartner(resEntry ->
										dbServerFd,
										(char *) &offset,
										sizeof(offset))
										== ERROR)

										{
if (resTrace)
	logMsg ("Error reading extra words during message discard from partner\n",0,0,0,0,0,0);
										break;
										}
									}
								if (recvPartner(resEntry -> dbServerFd,
										(char *) &offset,
										sizeof(short)) == ERROR)
									{
if (resTrace)
	logMsg ("Error reading offset during message discard from partner\n",0,0,0,0,0,0);
									break;
									}
								if (recvPartner(resEntry -> dbServerFd,
										buff,
										size & 0x3fff) == ERROR)
									{
if (resTrace)
	logMsg ("Error reading value during message discard from partner\n",0,0,0,0,0,0);
									break;
									}
								}
							continue;
							}

						/*
						Get the address of the record.
						*/

						precLoc = GET_PRECLOC(recType);
						pvdGetAddr(name, (char **)&recHdr);

						/*
						Now, put in the alarm ack field
						*/

						recHdr -> ack = dbUpdate.ack;

						/*
						Update each field that is in the message
						*/

						numFields = dbUpdate.numFields;
						while (dbUpdate.numFields--)
							{
							short	rocbOffset;
							short	valOffset;
							short	valLength;
							short	specialProcess;
							struct dbAddr recAddr;
							struct rset *prset;
							RECSUPFUN pspecial = 0;

							specialProcess = 0;
							if (recvPartner(resEntry -> dbServerFd,
									(char *) &size,
									sizeof(short)) == ERROR)
								{
if (resTrace)
	logMsg ("Error reading size from partner\n",0,0,0,0,0,0);
								break;
								}
							
							/*
							See if there is anything unique about this
							particular update.
							*/


							if (size & ROCI_UPDATE)
								{

								/*
								Update to ROCI. There are three more
								words to grab off the input stream. They
								are the offset of the ROCB field, the
								offset of the VAL field, and the size of
								the VAL field.
								*/

								if (recvPartner(resEntry -> dbServerFd,
									(char *) &rocbOffset,
									sizeof(rocbOffset)) == ERROR)
									{
if (resTrace)
	logMsg ("Error reading ROCB offset from partner\n",0,0,0,0,0,0);
									break;
									}
								if (recvPartner(resEntry -> dbServerFd,
									(char *) &valOffset,
									sizeof(valOffset)) == ERROR)
									{
if (resTrace)
	logMsg ("Error reading VAL offset from partner\n",0,0,0,0,0,0);
									break;
									}
								if (recvPartner(resEntry -> dbServerFd,
									(char *) &valLength,
									sizeof(valLength)) == ERROR)
									{
if (resTrace)
	logMsg ("Error reading VAL length from partner\n",0,0,0,0,0,0);
									break;
									}
								}
							else if (size & SPECIAL_PROC)
								{

								/*
								This point has special processing. Get
								the special processing number.
								*/

								if (recvPartner(resEntry -> dbServerFd,
									(char *) &specialProcess,
									sizeof(specialProcess)) == ERROR)
										{
if (resTrace)
	logMsg ("Error reading special processing # from partner\n",0,0,0,0,0,0);
										break;
										}

								/*
								For special processing, we need a phony
								struct dbAddr. The only members that are
								necessary are the record address and the
								special processing number.
								NOT! We also need the point type.
								*/

								if (specialProcess && (specialProcess != SPC_NOMOD))
									{
									recAddr.precord = (caddr_t) recHdr;
									recAddr.special = specialProcess;
									recAddr.record_type = recType;
									prset = GET_PRSET(recType);

									if (specialProcess < 100)
										{
										if (specialProcess == SPC_SCAN)
											{
											semTake(oldHoldSem, WAIT_FOREVER);
											stallScanTasks();
											delete_from_scan_list(&recAddr);
											resumeScanTasks();
											semGive(oldHoldSem);
											}
										}
									else
										{
										if (prset && (pspecial = prset -> special))
											(*pspecial)(&recAddr,0);
										}
									}
								}


							if (recvPartner(resEntry -> dbServerFd,
									(char *) &offset,
									sizeof(short)) == ERROR)
								{
if (resTrace)
	logMsg ("Error reading offset from partner\n",0,0,0,0,0,0);
								break;
								}
/****************
	During development, the following code was used to make sure that
	we never did a redundant update that wasn't in the proper place
	and was a proper size for that place. However, this code eats a
	lot of CPU time and was pulled from the production versions.
****************/

						if (dumping)
							{
							struct fldDes *fld;
							struct recTypDes *des;
							int i;

							des = GET_PRECTYPDES(recType);
							for (i = 0; i < des -> no_fields; i++)
								{
								fld = GET_PFLDDES(des, i);
								if (offset != fld -> offset)
									continue;
								if ((size & 0x3fff) != fld -> size)
									{
									logMsg ("\nUpdate for %s at offset %d has size %d, should be %d\n",
										 (int)name,
										 offset,
										 size & 0x3fff,
										 fld -> size,0,0);
									}
								logMsg ("    %4.4s: 0x", 
									 (int)(fld -> fldname),
									 0,0,0,0,0);
								break;
								}
							if (i >= des -> no_fields)
								{
								logMsg ("Update for %s at offset %d is WRONG!\n",
									 (int)name,
									 offset,0,0,0,0);

								}
							}
/***************/
							if (size & SPECIAL_PROC)
								dbScanLock((caddr_t)recHdr);
							if (recvPartner(resEntry -> dbServerFd,
									((char *)recHdr) + offset,
									size & 0x3fff) == ERROR)
								{
								if (size & SPECIAL_PROC)
									dbScanUnlock((caddr_t)recHdr);
if (resTrace)
	logMsg ("Error reading value from partner\n",0,0,0,0,0,0);
								break;
								}
							if (dumping)
								{
								unsigned char *dumpLoc = ((char *) recHdr) + offset;
								int	nbytes = size & 0x3fff;

								while (nbytes--)
									logMsg ("%02X", *dumpLoc++,
										 0,0,0,0,0);
								logMsg ("\n",0,0,0,0,0,0);
								}
							if (size & SPECIAL_PROC)
								dbScanUnlock((caddr_t)recHdr);
							
							/*
							Now the first phase is done. Check for any specials that
							may have occurred.
							*/

							if (size & ROCI_UPDATE)
								{

								/*
								Perform the logical function:
									ROCB[ROCI] = VAL;
								Remember that size is the length
								of ROCI.
								*/

								float	*rocLoc;
								int	rocIndex = 0;
								size &= 0x3fff;
								bcopy((char*) recHdr + rocbOffset,
									(char *) &rocLoc,
									sizeof(rocLoc));
								bcopy((char *) recHdr + offset,
									(char *) &rocIndex 
										+ sizeof(int)
										- size,
									size);
								if ((rocIndex < 0) || (rocIndex > 59))
									rocIndex = 0;
								if (valLength == sizeof(float))
									{
									float *valLoc = (float *)((char *) recHdr + valOffset);
									rocLoc[rocIndex] = *valLoc;
									}
								else
									{
									double *valLoc = (double *)((char *) recHdr + valOffset);
									rocLoc[rocIndex] = *valLoc;
									}
								}
							else if (specialProcess && (size & SPECIAL_PROC))
								{
								recAddr.pfield = (caddr_t)((char *)recHdr + offset);
								if (specialProcess < 100)
									{
									if (specialProcess == SPC_SCAN)
										{
										semTake(oldHoldSem, WAIT_FOREVER);
										stallScanTasks();
										add_to_scan_list(&recAddr, 0xffff);
										resumeScanTasks();
										semGive(oldHoldSem);
										}
									}
								else
									(*pspecial)(&recAddr, 1);
								}
							}

						/*
						Recalculate the CRC if any prompted fields
						have changed.
						*/

						if (dbUpdate.promptMod)
							recHdr -> rcrc = DBC_CRCHostRecord((char *) recHdr,
										       recType);
						
						/*
						Always compare our CRC with the sender's. If not
						the same, then request a fresh copy of the record.
						*/

						if (recHdr -> rcrc != dbUpdate.crc)
							{
							struct recTypDes	*pRecTypDes;

							/*
							First look at the number of fields we updated.
							If that's equal to # prompted fields - number
							of links, then this was a total record update.
							Don't ask for another update or we'll get into
							an infinite thrash doing this.
							*/


							pRecTypDes = (struct recTypDes *) 
									GET_PRECTYPDES( recType);
							if (numFields == (pRecTypDes -> no_prompt -
									pRecTypDes -> no_links))
								{
								logMsg ("dbServer: %s sync recovery fail!\n",
								         (int)(recHdr -> name),
									 0,0,0,0,0);
								continue;
								}
							msgType = RED_CRCFAIL;
							currTcpFd = resEntry -> dbServerFd;
							currTcpSize = 0;
							sendPartner(resEntry -> dbServerFd,
								    (char *) &msgType,
								    sizeof(msgType));
							sendPartner(resEntry -> dbServerFd,
								    recHdr -> name,
								    sizeof(recHdr -> name));
							flushTcp();
							}

						/*
						If the dbEmulator is loaded and there were prompted
						fields in the record, put the record into the local
						copy of the database.
						*/

						if (dbPutRecordFunc && dbUpdate.promptMod)
							{
							char			*recBuff;
							int			recSize;
    							struct recTypDes	*precTypDes;
							struct fldDes		*pfldDes;
							int			fieldIndex;
							int			offset;
							char			*oldRec;

							/*
							Modification by Gordon: Read the old record
							off the disk. Copy all the prompted fields
							that are not DBF_INLINK, DBF_OUTLINK,
							DBF_FWDLINK, or DBF_NOACCESS from the
							live database to the record read from disk.
							Then, also copy the CRC!
							*/

							recSize = dbRecDes -> papRecTypDes[ recType ] -> rec_size;
							recBuff = (char *) malloc(recSize);
							dbGetRecordFunc(recHdr, recType, recBuff);
							precTypDes = (struct recTypDes *) GET_PRECTYPDES(recType);
							oldRec = (char *) recHdr;

							for (fieldIndex = 0; fieldIndex < precTypDes -> no_fields; fieldIndex++)
							    {
							    pfldDes = GET_PFLDDES(precTypDes, fieldIndex);
							    if (!(pfldDes -> promptflag))
								continue;
							    if (pfldDes -> field_type == DBF_INLINK)
								continue;
							    if (pfldDes -> field_type == DBF_OUTLINK)
								continue;
							    if (pfldDes -> field_type == DBF_FWDLINK)
								continue;
							    if (pfldDes -> field_type == DBF_NOACCESS)
								continue;
							    offset = pfldDes -> offset;
							    bcopy(oldRec + offset, recBuff + offset, pfldDes -> size);
							    }

							((struct dbCommon *) recBuff) -> rcrc = recHdr -> rcrc;
							dbPutRecordFunc(recHdr, recType, recBuff);
							free(recBuff);
							}

						/*
						Now set the flag that indicates that we have
						had a redundant update for this record.
						*/

						recHdr -> rdur = 1;
						}
					else if (msgType == RED_KEEPALIVE)
						{

						/*
						Detected heartbeat from partner. This message
						is only sent if he had absolutely nothing
						else to say. Do nothing.
						*/

						;
						}
					else if (msgType == RED_CRCFAIL)
						{

						/*
						Read the point name and use it to get record
						type and record number.
						*/

						char 			recName[28];
						short			recType;
						short			offset;
						struct dbCommon		*recHdr;
						struct recLoc		*precLoc;
						struct recTypDes	*precTypeDes;
						struct fldDes		*fld;
						short			fieldIndex = 0;
						RED_DBUPDATE		dbUpdate;

						if (recvPartner(resEntry -> dbServerFd, recName,
								sizeof(recName)) == ERROR)
							{
if (resTrace)
	logMsg ("Error reading CRC fail name from partner\n",0,0,0,0,0,0);
							continue;
							}
						if (pvdGetRec(recName, &recType, &offset) < 0)
							{
							logMsg ("Can't find %s on CRCFAIL\n",
								 (int)recName,0,0,0,0,0);
							continue;
							}
						
						/*
						Now find the header for the point type and the
						address of the record.
						*/

						precTypeDes = (struct recTypDes *) GET_PRECTYPDES(recType);
						
						precLoc = GET_PRECLOC(recType);
						pvdGetAddr(recName, (char **)&recHdr);
						
						/*
						Build a redundant update message header and send
						it to our partner who requested it.
						*/

						dbUpdate.nameLen = strlen(recName) + 1;
						recHdr -> rcrc = DBC_CRCHostRecord((char *) recHdr, recType);
						dbUpdate.crc = recHdr -> rcrc;
						dbUpdate.ack = recHdr -> ack;
						dbUpdate.promptMod = 1;
						dbUpdate.numFields = precTypeDes -> no_prompt;
						dbUpdate.numFields -= precTypeDes -> no_links;
						/********************************************************/
						/* ALERT! We must be more gentle about this when	*/
						/* we allow for link modification!			*/
						/********************************************************/
						msgType = RED_UPDATE;
						currTcpFd = resEntry -> dbServerFd;
						currTcpSize = 0;
						sendPartner(resEntry -> dbServerFd, (char *) &msgType,
								sizeof(msgType));
						sendPartner(resEntry -> dbServerFd,
								(char *) &dbUpdate,
								sizeof(dbUpdate));
						sendPartner(resEntry -> dbServerFd,
								recName,
								dbUpdate.nameLen);
						
						/*
						Perform this update just as if it were a big
						redundant update. This is a little different from
						the message queue processing logic because we
						get the values from the record instead of the
						message queue and we also know that we will never
						see the ROCI field.
						*/

						for (; dbUpdate.numFields; fieldIndex++)
							{
							fld = GET_PFLDDES(precTypeDes, fieldIndex);
							if (!fld -> promptflag)
								continue;
							/********************************************************/
							/* ALERT! Be smarter than this when we allow link	*/
							/* modification!					*/
							/* ******************************************************/
							if (fld -> field_type == DBF_INLINK)
								continue;
							if (fld -> field_type == DBF_OUTLINK)
								continue;
							if (fld -> field_type == DBF_FWDLINK)
								continue;
							dbUpdate.numFields--;

							msgType = fld -> size;
							if (fld -> special || fld -> promptflag)
								msgType |= SPECIAL_PROC;
							sendPartner(resEntry -> dbServerFd,
									(char *) &msgType,
									sizeof(msgType));
							if (msgType & SPECIAL_PROC)
								{
								msgType = fld -> special;
								sendPartner(resEntry -> dbServerFd,
									(char *) &msgType,
									sizeof(msgType));
								}
							msgType = fld -> offset;
							sendPartner(resEntry -> dbServerFd,
								(char *) &msgType, sizeof(msgType));
							sendPartner(resEntry -> dbServerFd,
									(char *) recHdr +
										fld -> offset,
									fld -> size);
							}
						flushTcp();
						}
					else
						{
						logMsg ("Got TCP undefined msg type %d\n", 
						         msgType,0,0,0,0,0);
						}


					}
				    }
				}

			else
				{
				/*
				It wasn't from a redundant partner. It must be from a dbClient.
				*/

				short msgType;
				if (recvTcp(fdIndex, (char *) &msgType, sizeof(msgType)) == ERROR)
					{
					tcpErr(fdIndex);
					continue;
					}
				msgType = ntohs(msgType);
				if (msgType == DBC_MSG_BOOMERANG)
					{
					msgType = htons(msgType);
					if (npendSend(fdIndex, (char *) &msgType, sizeof(msgType)) == ERROR)
						tcpErr(fdIndex);
					}
				else if (msgType == DBC_MSG_SYMREQUEST)
					{
					/*
					Symbol table request. Read the header.
					*/

					struct recChain	*chain;
					char		*buff = tcpBuff;
					DBC_SYMREQ	request;
					short		type;
					short		number;
					short		*symCount;
					struct dbCommon	*hdr;
					if (recvTcp(fdIndex, (char *) &request, sizeof(request)) == ERROR)
						{
						tcpErr(fdIndex);
						continue;
						}
					if (request.symLastName[0])
						{
						/*
						Find the struct recChain for this point.
						*/

						pvdGetRec(request.symLastName, &type, &number);
						pvdGetAddr(request.symLastName, &buff);
						chain = (struct recChain *) (buff - sizeof(struct recChain));

						/*
						Symbol table reporting starts with the next one in line.
						*/

						chain = chain -> next;
						}
					else
						{
						type = 0;
						chain = 0;
						}
					
					/*
					Report the requested number of symbols.
					*/

					number = 0;
					buff = tcpBuff;
					/*
					  Place msgType and symbol count in same buffer as symbol/value
					  pairs. This will minimize the use of mbuffs
					*/
					*((short *)buff) = htons(DBC_MSG_SYMRESPONSE);
					buff += sizeof(short);
					symCount = (short *)buff;/* point to place where number of syms will be stored */
					*((short *)buff) = htons(number);
					buff += sizeof(number);
					request.numSyms = ntohs(request.numSyms);
					if (request.numSyms > totPnts)
						request.numSyms = totPnts;

					while (request.numSyms)
						{
						while (!chain)
							{
							type++;
							if (type >= dbRecType -> number)
								break;
							if (dbRecords -> papRecLoc[type])
								chain = (struct recChain *)
									(dbRecords -> papRecLoc[type] -> pFirst);
							}

						if (!chain)
							break;
						if (!(chain -> pRecord[0])) /* GCT 11/14/94 use [0] to check for deleted point */
							{
							chain = chain -> next;
							continue;
							}

						number++;
						request.numSyms--;
						hdr = (struct dbCommon *)(chain -> pRecord);
						strcpy(buff, hdr -> name);
						buff += sizeof(hdr -> name);
						*((long *) buff) = (long) hdr;
						buff += sizeof(long);
						chain = chain -> next;
						}
					*symCount = htons(number);/* store number of syms ready to send here */
					if(npendSend(fdIndex,
						     tcpBuff,
						     (number*(PVNAME_SZ+sizeof(long)))+sizeof(number)+sizeof(msgType)) == ERROR)
						{
						tcpErr(fdIndex);
						continue;
						}
					}
				
				else if (msgType == DBC_MSG_RECREQUEST)
					{

					DBC_UPLINKREQ	request;
					short			type;
					short			offset;
					struct recTypDes	*precTypeDes;
					struct recLoc		*precLoc;
					struct dbCommon		*recHdr;
					short			fieldIndex = 0;
					DBC_DBMODHDR		modHdr;
					struct fldDes		*fld;
					char			*buff;

					/*
					Client has detected a CRC mismatch with one of our records.
					Read the header to get the name.
					*/
					if (recvTcp(fdIndex, (char *) &request, sizeof(request)) == ERROR)
						{
						tcpErr(fdIndex);
						continue;
						}
					/*
					Get the name and treat it almost like a CRC failure from
					the partner IOC.
					*/


					if (pvdGetRec(request.requestName, &type, &offset) < 0)
						{
						logMsg ("Can't find %s on REC_REQUEST\n",
							 (int)(request.requestName),
							 0,0,0,0,0);
						continue;
						}
					precTypeDes = (struct recTypDes *) GET_PRECTYPDES(type);
					precLoc = GET_PRECLOC(type);
					pvdGetAddr(request.requestName, (char **)&recHdr);

					/*
					Pretend that this is just one great big uplink field mod.
					*/
					buff = tcpBuff;
					msgType = htons(DBC_MSG_DBMODREQ);
					*((short *)buff) = msgType;
					buff += sizeof(msgType);
					strcpy(modHdr.dbmodName, request.requestName);
/*
	Put the point type in the mod message. This will allow dbClient to drain the socket
	in the event of a field mod for a point that is not in the database on the host.

	GCT 10-May-1994

*/
					modHdr.dbmodType = htons(type);
					modHdr.dbmodNFields = precTypeDes -> no_prompt - precTypeDes -> no_links;
					recHdr -> rcrc = DBC_CRCHostRecord((char *) recHdr, type);
					modHdr.dbmodCRC = htons(recHdr -> rcrc);
					modHdr.dbmodPointId = (long) recHdr;
					modHdr.dbmodTimeStamp = htonl((int)(recHdr -> tstp));
/*

	The following code segment loops through the SDR totalling up the sizes
	of each field name and the size of its cooresponding value. This is done
	so we can include this size in the message header.

	GCT 15-Sep-1994
*/
					modHdr.dbmodSize = 0;
					for (; modHdr.dbmodNFields; fieldIndex++)
						{
						fld = GET_PFLDDES(precTypeDes, fieldIndex);
						if (!(fld -> promptflag))
							continue;
						/*************************************************/
						/* ALERT! Be smarter than this when we allow for */
						/* link modification.				 */
						/*************************************************/
						if (fld -> field_type == DBF_INLINK)
							continue;
						if (fld -> field_type == DBF_OUTLINK)
							continue;
						if (fld -> field_type == DBF_FWDLINK)
							continue;
						modHdr.dbmodNFields--;
						modHdr.dbmodSize += FLDNAME_SZ + fld->size;		/* accumulate sizes for all fields sent */
						}
					modHdr.dbmodSize = htons(modHdr.dbmodSize);
					modHdr.dbmodNFields = htons (precTypeDes -> no_prompt - precTypeDes -> no_links);
					bcopy((char *)&modHdr,buff,sizeof(modHdr));
					buff += sizeof(modHdr);
					/* must convert back to host order for loop */
					modHdr.dbmodNFields = ntohs (modHdr.dbmodNFields);
					for (fieldIndex = 0; modHdr.dbmodNFields; fieldIndex++)
						{/* for all fields */
						fld = GET_PFLDDES(precTypeDes, fieldIndex);
						if (!(fld -> promptflag))
							continue;
						/*************************************************/
						/* ALERT! Be smarter than this when we allow for */
						/* link modification.				 */
						/*************************************************/
						if (fld -> field_type == DBF_INLINK)
							continue;
						if (fld -> field_type == DBF_OUTLINK)
							continue;
						if (fld -> field_type == DBF_FWDLINK)
							continue;
						modHdr.dbmodNFields--;
						bcopy(fld->fldname,buff,FLDNAME_SZ);
						buff += FLDNAME_SZ;
						bcopy((char *) recHdr + fld -> offset, buff, fld -> size);
						switch (fld -> field_type)
							{
							case DBF_FLOAT:
								htonf(*(float *)buff, *(float *)buff);
								break;
							case DBF_DOUBLE:
								htond(*(double *)buff, *(double *)buff);
								break;
							case DBF_LONG:
							case DBF_ULONG:
								*(int *)buff = htonl(*(int *)buff);
								break;
							case DBF_SHORT:
							case DBF_USHORT:
							case DBF_ENUM:
							case DBF_GBLCHOICE:
							case DBF_CVTCHOICE:
							case DBF_RECCHOICE:
							case DBF_DEVCHOICE:
								*(short *)buff = htons(*(short *)buff);
								break;
							}
						buff += fld->size;/* point to next field position within tcpBuff */
					  	}/* end of for all fields */
					if (npendSend(fdIndex, tcpBuff, (buff - tcpBuff)) == ERROR)
						{
						tcpErr(fdIndex);
						break;
						}
					}
					/* Slide this one tab stop to the left ! */
					else if (msgType == DBC_MSG_OLDRESP)
						{
						short		recType;
						struct olAdd	*olAddRec;
						DBC_DWNLD_OK	unmarkMsg;

						/*
						Online download response. Get the number of records
						dbClient promises to send.
						*/
						DBC_DWNLD_RESP	response;
						struct dbClient	*client;

						if (recvTcp(fdIndex, (char *) &response,
								sizeof(response)) == ERROR)
							{
							tcpErr(fdIndex);
							continue;
							}
						/*
						Get a pointer to this client's control block.
						*/

						client = clientList;
						while (client)
							{
							if (client -> clientFd == fdIndex)
								break;
							client = client -> next;
							}
						if (!client)
							{
							logMsg ("Online response from unknown dbClient?\n",
								 0,0,0,0,0,0);
							tcpErr(fdIndex);
							continue;
							}

						/*
						If any records are coming, read them. Store them in the
						queue for this client.
						*/

						response.msgNumberSent = ntohs(response.msgNumberSent);
						if (response.msgNumberSent)
						    {
						    DBC_DWNLD_REQUEST	request;

						    while (response.msgNumberSent--)
							{
						    	/*
						    	Get the record type.
						    	*/
	
						    	if (recvTcp(fdIndex, (char *) &recType, sizeof(recType)) == ERROR)
								{
								tcpErr(fdIndex);
								break;
								}
						
						    	/*
						    	Allocate a structure to hold the record. This is allocated
						    	as a struct olAdd plus enough space for the record. Eventually
						    	we will use the struct recChain that is embedded inside the
						    	struct olAdd for new records. We'll be freeing the entire
						    	thing for record overwrites.
						    	*/

							recType = ntohs(recType);
						    	olAddRec = (struct olAdd *) malloc(sizeof(struct olAdd) +
							    	dbRecDes -> papRecTypDes[recType] -> rec_size);
						    	olAddRec -> recType = recType;
						    	olAddRec -> newRec.pRecord = (caddr_t)((char *) olAddRec +
							    	sizeof(struct olAdd));
						    	if (recvTcp(fdIndex, (char *) olAddRec -> newRec.pRecord,
							    	dbRecDes -> papRecTypDes[recType] -> rec_size) == ERROR)
								{
								tcpErr(fdIndex);
								break;
								}
						    	olAddRec -> newRec.next = (struct recChain *) (client -> firstOlAdd);
						    	client -> firstOlAdd = olAddRec;

							/*
							Remember the record name so we can continue
							*/

							strcpy(request.msgLastName, 
								((struct dbCommon *)(olAddRec -> newRec.pRecord)) -> name);
							logMsg ("Online download: %s\n", 
								 (int)(request.msgLastName),
								 0,0,0,0,0);
						    	}

						    /*
						    Request the next batch of records
						    */

						    if (!(client -> clientFd))
							continue;	/* Lost connection during record read loop */
						    msgType = htons(DBC_MSG_OLDREQ);
						    if (npendSend(fdIndex, (char *) &msgType, sizeof(msgType))
								== ERROR)
							{
							tcpErr(fdIndex);
							continue;
							}
						    request.msgStatus = htons(DSTAT_OK);
						    if (npendSend(fdIndex, (char *) &request, sizeof(request))
								== ERROR)
							{
							tcpErr(fdIndex);
							continue;
							}
						    }
						else
						    {
						    struct olAdd	*nextOlAdd;

						    /*
						    All records have been loaded. Now, stop the scan tasks
						    and put them into the database as appropriate.
						    */

						    logMsg ("Online download complete.\n",
							     0,0,0,0,0,0);
						    semTake(oldHoldSem, WAIT_FOREVER);
						    stallScanTasks();

						    olAddRec = client -> firstOlAdd;
						    while (olAddRec)
							{
							struct recLoc		*loc;
							short 			oldType;
							short			number;
							struct dbCommon		*recHdr;
							struct dbAddr		dbAddr;
							struct recChain		*current;
							DBC_SYMRESPNAME		respName;

							nextOlAdd = (struct olAdd *) (olAddRec -> newRec.next);
							olAddRec -> newRec.next = 0;
							recType = olAddRec -> recType;

							/*
							Write to EEROM database if present.
							*/

							if (dbPutRecordFunc)
							    dbPutRecordFunc(olAddRec -> newRec.pRecord, recType,
									olAddRec -> newRec.pRecord);
							
							recHdr = (struct dbCommon *) (olAddRec -> newRec.pRecord);
							localEndian((char *)recHdr, recType);
							recHdr -> rcrc = DBC_CRCHostRecord((char *) recHdr, recType);

							/*
							Get ready to store this record.
							*/

							loc = dbRecords -> papRecLoc[recType];
							if (!loc)
								{

								/*
								First record ever of this type
								*/

								loc = (struct recLoc *)calloc(1, sizeof(struct recLoc));
								dbRecords -> papRecLoc[recType] = loc;
								loc -> record_type = recType;
								loc -> rec_size = dbRecDes -> papRecTypDes[recType] ->
									rec_size;
								}
							
							/*
							See if this record exists
							*/

							if (pvdGetRec(recHdr -> name, &oldType, &number) < 0)
							    {
							    struct recChain	*chain;

							    /*
							    Brand new record. Put it into the database at the
							    end of the chain for this type. We don't want to
							    rearrange point numbers in case someone else may
							    have remembered one.
							    */

							    chain = &(olAddRec -> newRec);
							    dbAddr.record_type = recType;
							    dbAddr.precord = (caddr_t) (chain -> pRecord);
							    add_to_scan_list(&dbAddr, 0xffff);

							    current = (struct recChain *) (loc -> pFirst);
							    while (current)
								{
								if (!(current -> next))
									{
									current -> next = &(olAddRec -> newRec);
									number = loc -> no_records;
									break;
									}
								current = current -> next;
								}
							    if (!current)
								loc -> pFirst = (caddr_t) &(olAddRec -> newRec);
							    loc -> no_records++;
							    totPnts++;
							    add_pv(recHdr -> name, recType, number);
							    pvdSetAddr(recHdr -> name, olAddRec -> newRec.pRecord);
							    }
							else
							    {
							    int update_scan = 0;
							    struct dbCommon	*oldHdr;

							    /*
							    Existing record. Copy the new record on top of the old
							    one and then free the new record.
							    */


							    pvdGetAddr(recHdr -> name, (char **)&oldHdr); 
							    freeRecMem(oldHdr, recType);

							    /* if the scan rate has changed, then delete
							       from scan list, below we will add it to the
							       new scan list
							    */
							    if ((recHdr -> scan != oldHdr -> scan) 
									|| (recHdr -> phas != oldHdr -> phas))
								{
								update_scan = 1;
								dbAddr.record_type = recType;
								dbAddr.precord = (caddr_t) oldHdr;
								delete_from_scan_list (&dbAddr);
								}

							    /*
							    Copy the mlok and mlis fields from the
							    old record to the new one just in case
							    there are locks and/or channels open on
							    the record while we are playing with it.
							    */

							    bcopy((char *)&(oldHdr -> mlok), (char *)
								    &(recHdr -> mlok), sizeof(recHdr -> mlok));
							    bcopy((char *)&(oldHdr -> mlis), (char *)
								    &(recHdr -> mlis), sizeof(recHdr -> mlis));

							    /* 
							    If a member of a lockset, remove it and close up the links
							    */

							    if (oldHdr -> lset != 0)
								{

								/* if we are first in link, new link is next */

								    if (oldHdr -> lsad -> pFirstRec == oldHdr)
									oldHdr -> lsad -> pFirstRec = oldHdr -> lsnr;
									/* find us and relink */
								    else
								    {
								        struct dbCommon *currRec;
								        struct dbCommon *prevRec;

									currRec = oldHdr -> lsad -> pFirstRec;
									prevRec = currRec;
									while (currRec != oldHdr)
									{
									    prevRec = currRec;
									    currRec = currRec -> lsnr;
									}
									prevRec -> lsnr = oldHdr -> lsnr;
								    }
								}

							    /*
							    Now copy the record itself
							    */

							    bcopy(recHdr, oldHdr, loc -> rec_size);
							    free(olAddRec);
							    recHdr = oldHdr;
							    if (update_scan)
							        {
							        update_scan = 0;
							        add_to_scan_list(&dbAddr, 0xffff);
							        }
							    }

							/*
							Now see if a new resource has to be added
							*/

							if (resGetByName(recHdr -> resn, &resEntry)
									!= RES_OK)
								{
								/*
								Create a new resource
								*/

								resEntry = resTableHead;
								while (resEntry)
									{
									if (!(resEntry -> link))
										{
										resEntry -> link = (RES_TABLE *)
											calloc(1, sizeof(RES_TABLE));
										resEntry = resEntry -> link;
										break;
										}
									resEntry = resEntry -> link;
									}
								strcpy(resEntry -> resourceName, recHdr -> resn);
								}

							/*
							Establish resource table addr in new record
							*/

							recHdr -> radr = resEntry;

							/*
							Update the symbol table in all online clients
							*/

							msgType = htons(DBC_MSG_NEWSYM);
							sendClients((char *) &msgType, sizeof(msgType));
							msgType = htons(1);	/* # Symbols */
							sendClients((char *) &msgType, sizeof(msgType));
							strcpy(respName.symName, recHdr -> name);
							respName.symTrans = htonl(((int)recHdr));
							sendClients((char *) &respName, sizeof(respName));

							/*
							Set pointer to next record to be added
							*/

							olAddRec = nextOlAdd;
							}

						    initDatabase(1);	/* This is a restart */
						    resumeScanTasks();
						    semGive(oldHoldSem);

						    /*
						    Recalculate database synchronization parameters.
						    */

						    if (totPnts)
						        {
		/*
			Prep syncPts to tell us the maximum number of points in each CRC transmission. For
			3-May-1999 this will be the number sent in a TCP message(NO Broadcast!). The
			DBC_CRCINFO structure is 12 bytes and is found in redmsgs.h
		*/
								syncPts = (sizeof(tcpBuff) - sizeof(short) - sizeof(short)) / sizeof(DBC_CRCINFO);
						        if (syncPts > totPnts)
							    syncPts = totPnts;
						        syncDelta = SYNC_TRAVERSE * 60 * 60 /
							    ((totPnts + syncPts - 1) / syncPts);
						        syncType = dbRecType -> number + 1;
						        syncNext = 0;
						        syncTime = tickGet();
						        }
						    /*
						    Now tell the dbClient that sent us the points that it's OK
						    to unmark them now.
						    */

						    msgType = htons(DBC_MSG_OKTOUNMARK);
						    if (npendSend(fdIndex, (char *) &msgType, sizeof(msgType)) == ERROR)
							    {
							    tcpErr(fdIndex);
							    continue;
							    }
						    unmarkMsg.ackStatus = 0;
						    if (npendSend(fdIndex, (char *) &unmarkMsg, sizeof(unmarkMsg)) == ERROR)
							    {
							    tcpErr(fdIndex);
							    continue;
							    }

						    /*
						    Clear pointer in dbClient control block because we're done.
						    */

						    client -> firstOlAdd = 0;
						    }
						}
	/*
		GCT 4-May-1999

		The following message handler is in response to a DBC_MSG_RESETNOTIFY message
		from dbClient. It has been relocated to the TCP side of things as part of the
		effort to rid ourselves of UDP messages. The acknowledgment contains the fd
		from the dataVu which originally sent the message to the dbClient I got this
		message from.
		
			The 'reboot' function produces the same effect as pushing the reset button.
	*/
				else if(msgType == DBC_MSG_RESETNOTIFY)
					{
						DBC_RESETREQ reset;
					/*
					Format a reset acknowledge message and send it.
					*/
						logMsg ("Received DBC_MSG_RESETNOTIFY message\n",0,0,0,0,0,0);
						if (recvTcp(fdIndex, (char *) &reset, sizeof(DBC_RESETREQ)) == ERROR)
							{
							tcpErr(fdIndex);
							continue;
							}
						msgType = htons(DBC_MSG_RESETACK);
					    if (npendSend(fdIndex, (char *) &msgType, sizeof(msgType)) == ERROR)
						    {
							tcpErr(fdIndex);
							continue;
						    }
					    if (npendSend(fdIndex, (char *) &reset.originatorFd, sizeof(reset.originatorFd)) == ERROR)
						    {
							tcpErr(fdIndex);
							continue;
						    }
					taskDelay(10 * vxTicksPerSecond);
					sysReboot();
					taskSuspend(taskIdSelf());
					}
			else if(msgType == DBC_MSG_OLDNOTIFY)
				{
				/*
				The utility program for online download has been run. Look for
				the corresponding dbClient and let him know we are ready to start.
				*/

				struct dbClient		*client = clientList;
				DBC_DWNLD_REQUEST	request;

				while(client)
					{
					if (client -> clientFd == fdIndex)
						break;
					client = client -> next;
					}
				if (!client)
					{
					logMsg ("Online downline notify from unknown workstation.\n",0,0,0,0,0,0);
					continue;
					}
				
				/*
				If dbClient isn't online, we can't send the message. He'll have to
				try again.
				*/

				if (!(client -> clientFd))
					continue;

				/*
				Format an online download request and send it.
				*/

				msgType = htons(DBC_MSG_OLDREQ);
				if (npendSend(fdIndex, (char *) &msgType, sizeof(msgType)) == ERROR)
					{
					tcpErr(fdIndex);
					continue;
					}

				request.msgLastName[0] = 0;
				request.msgStatus = DSTAT_OK;
				if (npendSend(fdIndex, (char *) &request, sizeof(request)) == ERROR)
					{
					tcpErr(fdIndex);
					continue;
					}

				/*
				Wipe out any online adds that may have been pending from this guy.
				*/

				deleteOlAdds(client);
				logMsg ("Online download requested.\n",0,0,0,0,0,0);
#ifdef PRIV_AUDIT
				/* launch online.cmd script to pick up releted dataset information */
				if (getenv("ONLINEADD"))
				    RunScript (getenv ("ONLINEADD"));
#endif PRIV_AUDIT
				}
			else if (msgType == DBC_MSG_OLDELETE)
				{
				/*
				Online Record Deletion
				*/

				DBC_OLDELETE		deleteMsg;
				DBC_DELRESP			dresp;
				short				recType;
				short				recNum;
				struct dbCommon		*recHdr;
				struct recLoc		*precLoc;
				short				ndups;
				short				type;
				struct dbAddr		dbAddr;
				int dataVu;
				char *pBuff = tcpBuff;
				/*
				Find this record
				*/
				if (recvTcp(fdIndex, (char *) &deleteMsg, sizeof(DBC_OLDELETE)) == ERROR)
					{
					tcpErr(fdIndex);
					continue;
					}
				dataVu = ntohl(deleteMsg.originatorFd);	/* get FD of dataVu */
				if (pvdGetRec(deleteMsg.nameToDelete, & recType, &recNum) < 0)
					{
					rejectDeletion(DELETION_NOREC, fdIndex,dataVu,0);
					continue;
					}
				precLoc = GET_PRECLOC(recType);
				pvdGetAddr(deleteMsg.nameToDelete, (char **)&recHdr);
#ifdef	DEL_OFF_SCAN
				/*							don't check for off scan GCT 12-May-1994
				Check for record off scan
				*/
				if (recHdr -> oscn != OFF_SCAN)
					{
					rejectDeletion(DELETION_ONSCAN,fdIndex,dataVu,0);
					continue;
					}
#endif				
				/*
				Check for open channels
				*/

				if (recHdr -> mlis.count)
					{
					rejectDeletion(DELETION_OPEN,fdIndex,dataVu,0);
					continue;
					}
				/*
				Check to see if there is any other record in the same lock set
				*/

				ndups = 0;
				for (type = 0; type < dbRecords -> number; type++)
					{
					struct recChain	*pchain;
					struct dbCommon	*otherRec;
					if (!(precLoc = GET_PRECLOC(type)))
						continue;
					for (pchain = (struct recChain *) (precLoc -> pFirst);
							pchain; pchain = pchain -> next)
						{
						if (!(otherRec = (struct dbCommon *)(pchain -> pRecord)))
							continue;
						if (!(otherRec -> name[0]))
							continue;
						if (otherRec -> lset == recHdr -> lset)
							if (otherRec != recHdr)
								{
								strcpy(pBuff + ndups * PVNAME_SZ,otherRec->name);
								ndups++;
								}
						}
					}
				if (ndups)
				{
					/*
						GCT 6-May-1999

						No need to notify dbClient/dataVu if this deletion is performed
						on the slave IOC in a redundant configuration
					*/
					resGetCaAnswer(recHdr->radr, &type);
					if (type)
					{
						rejectDeletion(DELETION_REF,fdIndex,dataVu,ndups);
						if (npendSend(fdIndex,(char *)pBuff,ndups*PVNAME_SZ) == ERROR)
							{
							tcpErr(fdIndex);
							continue;
							}
						continue;
					}
				}
/*
	Reject deletion if this record is referenced by any other record.
	Do this by traversing the entire database looking for DB_LINK &
	PV_LINK link types whose PV matches this one. Build list into reply
	message as above.

	GCT	08-sep-1994
*/
#ifdef	CHECK_REFS
				ndups = 0;
    			for(i=0; i< (dbRecords->number); i++)
					{
					struct recChain	*pchain;
					struct dbCommon	*precord;
					if(!(precLoc = dbRecords->papRecLoc[i]))continue;
					precTypDes = dbRecDes->papRecTypDes[i];
					for(j=0, pchain = (struct recChain *) precLoc -> pFirst;
					    j<precLoc->no_records;
					    j++, pchain = pchain -> next)
						{
						precord = (struct dbCommon *) pchain -> pRecord;
						if(!(precord->name[0])) continue;	        /* If NAME is null then skip this record*/
						for(k=0; k<precTypDes->no_links; k++)
							{
						    pfldDes = precTypDes->papFldDes[precTypDes->link_ind[k]];
						    plink = (struct link *)((char *)precord + pfldDes->offset);
						    if((plink -> type == DB_LINK) || (plink->type == PV_LINK))
								{
								if(!strncmp(deleteMsg->nameToDelete, plink -> value.pv_link.pvname, PVNAME_SZ))
									{
									if(strncmp(precord->name,deleteMsg->nameToDelete, PVNAME_SZ))
										{
										strcpy(udpBuff + 3 * sizeof(short) + ndups * PVNAME_SZ,precord -> name);
										ndups++;
										}
									}
								}
							}
						}
					}
				if (ndups)
					{
					short *msg = (short *) udpBuff;
					*msg++ = htons(DBC_MSG_DELSTATUS);
					*msg++ = htons(DELETION_REF);
					*msg++ = htons(ndups);
#ifdef USE_UDP
					sendto(bcastSocket, udpBuff, 3 * sizeof(short) +
							ndups * PVNAME_SZ, 0, &from, sizeof(from));
#endif
					continue;
					}
#endif
				/*
				Now delete the record. First, take it out of the scan list.
				*/

				strcpy(dresp.deletedName, recHdr -> name);
				recHdr -> name[0] = 0;
				semTake(oldHoldSem, WAIT_FOREVER);
				stallScanTasks();
				dbAddr.precord = (caddr_t) recHdr;
				delete_from_scan_list(&dbAddr);
				resumeScanTasks();
				semGive(oldHoldSem);

				/*
				Now for some cleanup
				*/

				freeRecMem(recHdr, recType);
				delete_from_pvd((char *)&dresp.deletedName);
				rejectDeletion(DELETION_OK,fdIndex,dataVu,0);
				recType = htons(DBC_MSG_OLDELRESP);
				dresp.deletedSym = htonl((long) recHdr);
				resGetCaAnswer(recHdr -> radr, &type);
				if (type)
					{
					sendClients((char *)&recType, sizeof(recType));
					sendClients((char *)&dresp, sizeof(dresp));
					}
				}
				else
					{
					logMsg ("Received unknown msg type %d from dbClient\n",msgType,0,0,0,0,0);
					}
				}

			}
		/*
		Recopy the fd mask prior to selecting again. An I/O error or a new
		connection may have caused us to change the mask.
		*/

		bcopy((char *) &tcpFds, (char *) &readFds, sizeof(tcpFds));
		}
	}

/******
	Check the incoming queue for notifications of database updates
******/

	{
	char		queueMsg[REDQ_MAX_MSGLEN];
	currTcpFd = 0;
	currTcpSize = 0;

	while (!rngIsEmpty(redundQueue))
		{
		struct dbServerQHdr *queHdr;
		RED_DBUPDATE	    dbUpdate;
		RES_TABLE	    *resEntry;
		char		    *sourcePtr;
		short		    msgType = RED_UPDATE;
		short		    clientMsg;
		int		    msgSize;
		DBC_DBMODHDR	    modHdr;
		char		    *buff;

		/*
		The first thing in the ring buffer is the size of the message.
		Get that.
		*/

		rngBufGet(redundQueue, (char *) &msgSize, sizeof(msgSize));

		/*
		Now get a message that long. Remember to first deduct the size of the
		integer that contains the message size!
		*/

		msgSize -= sizeof(msgSize);
		rngBufGet(redundQueue, queueMsg, msgSize);

		/*
		Got the message to process. If the resource for this point
		is not connected, throw the message on the floor. If it
		is connected, then build the header of the message for
		our partner. Send it to him.
		*/

		queHdr = (struct dbServerQHdr *) queueMsg;
		resEntry = queHdr -> recordAddr -> radr;
		
		/*
		Now send the changes for all prompted fields up to all dbClients
		that have registered with us.
		*/
		buff = dbclientBuff;
		if (queHdr -> promptFlag)
			{
			int	numFields;

			clientMsg = htons(DBC_MSG_DBMODREQ);
			/*sendClients((char *) &clientMsg, sizeof(clientMsg));*/
			*((short *)buff) = clientMsg;	/* fill big buffer before transmit */
			buff += sizeof(clientMsg);
			strcpy(modHdr.dbmodName, queHdr -> recordAddr -> name);
/*
	Send point type with field mod. This is so dbClient can drain the socket
	for a point that has been deleted on the host.

	GCT 10-May-1994

*/
			modHdr.dbmodType = htons(queHdr -> pointType);
			modHdr.dbmodNFields = 0;    /* count them as we add them! (remember to htons!) */
			modHdr.dbmodCRC = htons(queHdr -> recordAddr -> rcrc);
			modHdr.dbmodPointId = (long) (queHdr -> recordAddr);
			modHdr.dbmodTimeStamp = htonl((int)(queHdr -> recordAddr -> tstp));
/*

	The following code segment loops through the SDR totalling up the sizes
	of each field name and the size of its cooresponding value. This is done
	so we can include this size in the message header.

	GCT 15-Sep-1994
*/
			numFields = queHdr -> numFields;
			modHdr.dbmodSize = 0;
			sourcePtr = queueMsg + sizeof(struct dbServerQHdr);
			while (numFields--)
				{
				struct fldDes	*fld;
				bcopy((char *) sourcePtr, (char *) &fld, sizeof(fld));
				sourcePtr += sizeof(fld);
				if (!(fld -> promptflag))
					{
					sourcePtr += fld -> size;
					continue;
					}
				modHdr.dbmodSize += (FLDNAME_SZ + fld->size);
				sourcePtr += fld -> size;
				modHdr.dbmodNFields++;
			}
			modHdr.dbmodNFields = htons (modHdr.dbmodNFields);
			modHdr.dbmodSize = htons(modHdr.dbmodSize);
			bcopy((char *)&modHdr,buff,sizeof(modHdr));	/* fill big buffer before transmit */
			buff += sizeof(modHdr);
			/*sendClients((char *) &modHdr, sizeof(modHdr));*/
			sourcePtr = queueMsg + sizeof(struct dbServerQHdr);
			numFields = queHdr -> numFields;
			/*
			  in this loop fill the big buffer until all requested fields are in there
			  and then transmit
			*/
			while (numFields--)
				{
				struct fldDes	*fld;
				bcopy((char *) sourcePtr, (char *) &fld, sizeof(fld));
				sourcePtr += sizeof(fld);
				if (!(fld -> promptflag))
					{
					sourcePtr += fld -> size;
					continue;
					}
				bcopy(fld->fldname,buff,FLDNAME_SZ);
				buff += FLDNAME_SZ;	/* copy field name into buffer */
				/*sendClients(fld -> fldname, FLDNAME_SZ);*/
				bcopy(sourcePtr, buff, fld -> size);	/* copy field value into buffer */
				switch (fld -> field_type)
					{
					case DBF_FLOAT:
						htonf(*(float *)buff, *(float *)buff);
						break;
					case DBF_DOUBLE:
						htond(*(double *)buff, *(double *)buff);
						break;
					case DBF_LONG:
					case DBF_ULONG:
						*(int *)buff = htonl(*(int *)buff);
						break;
					case DBF_SHORT:
					case DBF_USHORT:
					case DBF_ENUM:
					case DBF_GBLCHOICE:
					case DBF_CVTCHOICE:
					case DBF_RECCHOICE:
					case DBF_DEVCHOICE:
						*(short *)buff = htons(*(short *)buff);
						break;
					}
				sourcePtr += fld -> size;
				buff += fld->size;	/* bump by size of field value */
				}
			/*
			  Now send the big buffer. (buff - dbClientBuff) is the transmit size
			*/
			sendClients(dbclientBuff, (buff - dbclientBuff));
			}

		if (!resEntry -> dbServerConnected)
			continue;
		if (!currTcpFd)
			currTcpFd = resEntry -> dbServerFd;
		sendPartner(resEntry -> dbServerFd, (char *) &msgType, sizeof(msgType));
		dbUpdate.nameLen = strlen(queHdr -> recordAddr -> name) + 1;
		dbUpdate.crc = queHdr -> recordAddr -> rcrc;
		dbUpdate.ack = queHdr -> recordAddr -> ack;
		dbUpdate.promptMod = queHdr -> promptFlag;
if (resTrace)
	/*logMsg ("Scan task modified %s, %d fields\n", 
		   (int)(queHdr -> recordAddr -> name), 
		   dbUpdate.numFields,0,0,0,0);*/;
		dbUpdate.numFields = queHdr -> numFields;
		sendPartner(resEntry -> dbServerFd, (char *) &dbUpdate, sizeof(dbUpdate));
		sendPartner(resEntry -> dbServerFd, queHdr -> recordAddr -> name,
				dbUpdate.nameLen);

		/*
		We sent the header. Now, build up and send the stuff for each field in the message.
		*/

		sourcePtr = queueMsg + sizeof(struct dbServerQHdr);
		while (dbUpdate.numFields--)
			{
			short		msgWord;	/* A place to put a 16 bit word */
			struct fldDes	*fld;

			/*
			Get the field length and ship it.
			*/

			bcopy((char *) sourcePtr, (char *) &fld, sizeof(fld));
			sourcePtr += sizeof(fld);
			msgWord = fld -> size;

			/*
			Check here for special processing
			*/

			if (fld -> special)
				msgWord |= SPECIAL_PROC;
			else
				{
				/*
				Check to see if the field is ROCI. If so, then
				set appropriate bit in the length word. We'll
				figure out what more to send later.
				*/

				if (!strncmp(fld -> fldname, "ROCI", sizeof(fld -> fldname)))
					msgWord |= ROCI_UPDATE;
				}

			sendPartner(resEntry -> dbServerFd, (char *) &msgWord, sizeof(msgWord));

			if (msgWord & SPECIAL_PROC)
				{
				msgWord = fld -> special;
				sendPartner(resEntry -> dbServerFd, (char *) &msgWord,
								sizeof(msgWord));
				}
			else if (msgWord & ROCI_UPDATE)
				{

				/*
				The order of words to ship is the offset of the ROCB field, the offset 
				of the VAL field, and the size of the VAL field. The rest of the logic
				will ship the offset of ROCI and its new value.
				*/

				msgWord = rocbList[queHdr -> pointType] -> offset;
				sendPartner(resEntry -> dbServerFd, (char *) &msgWord, sizeof(msgWord));
				msgWord = valList[queHdr -> pointType] -> offset;
				sendPartner(resEntry -> dbServerFd, (char *) &msgWord, sizeof(msgWord));
				msgWord = valList[queHdr -> pointType] -> size;
				sendPartner(resEntry -> dbServerFd, (char *) &msgWord, sizeof(msgWord));
						}


			/*
			Now ship the offset
			*/

			msgWord = fld -> offset;
			sendPartner(resEntry -> dbServerFd, (char *) &msgWord, sizeof(msgWord));

			/*
			Now ship the value.
			****************************************************
			* WE MADE THE RULE THAT REDUNDANT PARTNERS MUST	   *
			* HAVE PROCESSORS WITH THE SAME ENDIAN		   *
			****************************************************
			*/

			sendPartner(resEntry -> dbServerFd, sourcePtr, fld -> size);
			sourcePtr += fld -> size;
			}

		
		}
	flushTcp();
	}

/******
	Traverse the resource table in case some connection
	work needs to be done
******/

if ((tickGet() - lastTable) >= 60)
	{
	RES_TABLE	*resEntry;
	struct
		{
		short			reqType;
		RED_CONNREQ_HDR		reqHdr;
		RED_CONNREQ_REQ		reqList
						[(MAX_UDP_MSG
						- sizeof(short)
						- sizeof(RED_CONNREQ_HDR))
						/ sizeof(RED_CONNREQ_REQ)];
		} reqMsg;
	int		numReqs = 0;

	lastTable = tickGet();
	for (resEntry = resTableHead; resEntry; resEntry = resEntry -> link)
		{

		/************************************************************************/
		/*									*/
		/*	If a driver hasn't registered this resource yet or if		*/
		/*	one has declared this resource to be nonredundant, there	*/
		/*	is nothing left to do for this row.				*/
		/*									*/
		/************************************************************************/

		if (resEntry -> driverPort == 0 || resEntry -> private)
			continue;

		/************************************************************************/
		/*									*/
		/*		SEE IF CONNECTION ATTEMPTS HAVE TO START OVER		*/
		/*									*/
		/************************************************************************/

		if (resEntry -> disconnectTime)
			{
			if ((tickGet() - resEntry -> disconnectTime) > MAX_DISCONN)
				{

				/*
				if resMgr is also disconnected, then we should start over
				*/

				if (!resEntry -> partnerConnected)
					{

					/*
					This connection is in bad shape. Start over by
					clearing the partnerInet field for every resource
					that used this particular connection.
					*/

					RES_TABLE	*otherEntry = resTableHead;

					for (; otherEntry; otherEntry = otherEntry -> link)
						if (otherEntry -> dbServerFd == resEntry ->
									dbServerFd)
							otherEntry -> partnerInet.s_addr = 0;
					
					closeTcpStream(resEntry -> dbServerFd);
if (resTrace)
	logMsg ("Trashed TCP connection for %s\n", 
	         (int)(resEntry -> resourceName),0,0,0,0,0);
					}
				}
			}


		/************************************************************************/
		/*									*/
		/*	See if this entry was registered but still needs a partner.	*/
		/*									*/
		/************************************************************************/

		if ((resEntry -> partnerInet.s_addr == 0) || !(resEntry -> partnerConnected))
			{

			/*
			It needs a partner. Age the entry. If it is the proper
			time for a broadcast, put this resource into a request
			message so that we can ship it out after traversing
			the entire table or when the message buffer overflows,
			whichever comes first.
			*/

			if (--resEntry -> bcastCounter > 0)
				continue;
			resEntry -> bcastsMade++;
			resEntry -> bcastCounter = resEntry -> bcastRate;
			if ((sizeof(short) + sizeof(RED_CONNREQ_HDR) +
				(sizeof(RED_CONNREQ_REQ) * (numReqs + 1))) > MAX_UDP_MSG)
				{
				struct sockaddr_in bcastAddr;
				reqMsg.reqType = htons(RED_CONNREQ);
				reqMsg.reqHdr.redNumRequests = htons(numReqs);
				bcastAddr.sin_family = AF_INET;
				bcastAddr.sin_addr.s_addr = htonl(INADDR_BROADCAST);
				bcastAddr.sin_port = htons(DBSERVER_UDP_PORT);
#ifdef USE_UDP
				if (sendto(bcastSocket,
					   (char *) &reqMsg,
					   sizeof(short) + sizeof(RED_CONNREQ_HDR) +
						sizeof(RED_CONNREQ_REQ) * numReqs,
					   0,
					   (struct sockaddr *) &bcastAddr,
					   sizeof(bcastAddr)) == ERROR)
						{
						logMsg ("Can't broadcast: %d\n", 
							 errno,0,0,0,0,0);
						}
#endif
				numReqs = 0;
				}
			strcpy(reqMsg.reqList[numReqs].redReqName, resEntry -> resourceName);
			reqMsg.reqList[numReqs].redReqPort = htons(resEntry -> driverPort);
if (resTrace)
	logMsg ("Looking for a partner for %s on port %d\n", 
		 (int)(resEntry -> resourceName), 
		 resEntry -> driverPort,0,0,0,0);

			numReqs++;
			continue;
			}

		/************************************************************************/
		/*									*/
		/*	See if we have to create a TCP socket for conversing with	*/
		/*	the dbServer task in a partner machine.				*/
		/*									*/
		/************************************************************************/

		if (resEntry -> localDbPort == 0)
			{

			/*
			Don't have one for this resource. See if we can use a connection
			that was already made for another resource.
			*/

			RES_TABLE	*otherEntry = resTableHead;

			while (otherEntry)
				{
				if (otherEntry -> partnerInet.s_addr == resEntry -> partnerInet.s_addr
				 && otherEntry -> localDbPort != 0)
					break;
				otherEntry = otherEntry -> link;
				}
			if (otherEntry)
				{

				/*
				There is one we can use
				*/

if (resTrace)
	logMsg ("Borrowing dbServer port for %s from %s.\n", 
		 (int)(resEntry -> resourceName), 
		 (int)(otherEntry -> resourceName),0,0,0,0);

				resEntry -> localDbPort = otherEntry -> localDbPort;
				resEntry -> dbServerFd = otherEntry -> dbServerFd;
				resEntry -> dbServerPort = otherEntry -> dbServerPort;
				resEntry -> dbServerConnected = otherEntry -> dbServerConnected;
				resEntry -> disconnectTime = otherEntry -> disconnectTime;
				resEntry -> lastRxHeartbeat = otherEntry -> lastRxHeartbeat;
				}
			else
				{
				/*
				We have to create one. When we do so, processing for other
				table rows will cause them to copy this one, if appropriate.
				*/

if (resTrace)
	logMsg ("Creating TCP port for %s\n", (int)(resEntry -> resourceName),0,0,0,0,0);
				if ((resEntry -> dbServerFd = socket(AF_INET, SOCK_STREAM, 0))
								== ERROR)
					{
					logMsg ("Can't create local dbServer socket.\n", 
						 errno,0,0,0,0,0);
					resEntry -> localDbPort = 0;
					}
				else
					{
					struct sockaddr_in partnerAddr;
					int sockSize = sizeof(partnerAddr);
					char true = 1;
					struct
						{
						short		msgType;
						RED_IOCLINKREQ	req;
						} reqMsg;

					/*ioctl(resEntry -> dbServerFd, FIONBIO, (int)&true);*/
					setsockopt(resEntry -> dbServerFd, IPPROTO_TCP,
						TCP_NODELAY, &true, sizeof(true));
					if (resEntry -> dbServerFd > maxTcpFd)
						maxTcpFd = resEntry -> dbServerFd;
					resEntry -> disconnectTime = tickGet();
					partnerAddr.sin_family = AF_INET;
					partnerAddr.sin_addr.s_addr = htonl(INADDR_ANY);
					partnerAddr.sin_port = htons(0);
					bind(resEntry -> dbServerFd, (struct sockaddr *) &partnerAddr,
									sizeof(partnerAddr));
					getsockname(resEntry -> dbServerFd, (struct sockaddr *)&partnerAddr,
							&sockSize);
					resEntry -> localDbPort = ntohs(partnerAddr.sin_port);

					/*
					If we have the lower Internet address of the pair,
					then we will be the server. If so, set up the port
					to listen to one connection.
					*/

					if (ntohl(resEntry -> partnerInet.s_addr) > ntohl(ourInetAddr.s_addr))
						{
if (resTrace)
	logMsg ("...Our Inet addr is lower so we will listen!\n",0,0,0,0,0,0);
						listen(resEntry -> dbServerFd, 1);
						tis_fd_set(resEntry -> dbServerFd, &tcpFds);
						}
					else
if (resTrace)
	logMsg ("...Our Inet addr is greater so we will connect!\n",0,0,0,0,0,0);

					/*
					Tell our redundant partner that we are waiting to
					connect with him.
					*/

					reqMsg.msgType = htons(RED_IOCDBREQ);
					reqMsg.req.redDbServerPort = htons(resEntry -> localDbPort);
					partnerAddr.sin_family = AF_INET;
					partnerAddr.sin_addr = resEntry -> partnerInet;
					partnerAddr.sin_port = htons(DBSERVER_UDP_PORT);
#ifdef USE_UDP
					sendto(bcastSocket, (char *) &reqMsg, sizeof(short) +
						sizeof(RED_IOCLINKREQ), 0, (struct sockaddr *)&partnerAddr,
						sizeof(partnerAddr));
#endif
if (resTrace)
	logMsg ("...and told my partner to meet me on port %d\n", 
		 resEntry -> localDbPort,0,0,0,0,0);
					continue;
					}
				}
			}

		/****************************************************************/
		/*								*/
		/*	See if we know the TCP port number in our partner	*/
		/*								*/
		/****************************************************************/

		if (resEntry -> dbServerPort == 0)
			{

			/*
			Before we ask our redundant partner for his port number,
			see if any other resources have one we can borrow.
			*/

			RES_TABLE	*otherEntry = resTableHead;

			while (otherEntry)
				{
				if (otherEntry -> partnerInet.s_addr == resEntry -> partnerInet.s_addr
					&& otherEntry -> dbServerPort != 0)
						break;
				otherEntry = otherEntry -> link;
				}
			if (otherEntry)
				{
				resEntry -> dbServerPort = otherEntry -> dbServerPort;
				resEntry -> dbServerFd = otherEntry -> dbServerFd;
				resEntry -> dbServerConnected = otherEntry -> dbServerConnected;
				resEntry -> disconnectTime = otherEntry -> disconnectTime;
				resEntry -> lastRxHeartbeat = otherEntry -> lastRxHeartbeat;
if (resTrace)
	logMsg ("Borrowing tcp port for %s from %s\n", 
		 (int)(resEntry -> resourceName), 
		 (int)(otherEntry -> resourceName),
		 0,0,0,0);
				}
			else
				{

				/*
				Ask redundant partner for his dbServer port number
				*/

				struct
					{
					short		msgType;
					RED_IOCLINKREQ	req;
					} reqMsg;
				struct sockaddr_in	partner;

				reqMsg.msgType = htons(RED_IOCDBREQ);
				reqMsg.req.redDbServerPort = htons(resEntry -> localDbPort);
				partner.sin_family = AF_INET;
				partner.sin_addr = resEntry -> partnerInet;
				partner.sin_port = htons(DBSERVER_UDP_PORT);
#ifdef USE_UDP
				sendto(bcastSocket, (char *) &reqMsg, sizeof(short) +
					sizeof(RED_IOCLINKREQ), 0, (struct sockaddr *)&partner, sizeof(partner));
#endif
if (resTrace)
	logMsg ("Asking partner for TCP connection\n",0,0,0,0,0,0);
				continue;
				}
			}
		
		/************************************************************************/
		/*									*/
		/*	Now, if we still aren't connected at this point, try to		*/
		/*	make the TCP connection if our Internet address is greater	*/
		/*	than our redundant partner's. If our address is lower, we	*/
		/*	already issued the listen() call when we created the port.	*/
		/*									*/
		/************************************************************************/

		if (resEntry -> dbServerConnected == 0)
			{

			/*
			Before we try a connect, let's see if another resource to
			the same node is already connected.
			*/

			RES_TABLE	*otherEntry;
			otherEntry = resTableHead;
			while (otherEntry)
				{
				if (otherEntry != resEntry
				 && otherEntry -> partnerInet.s_addr == resEntry -> partnerInet.s_addr
				 && otherEntry -> dbServerPort == resEntry -> dbServerPort
				 && otherEntry -> dbServerConnected)
					{
					resEntry -> dbServerConnected = 1;
					resEntry -> disconnectTime = 0;
					resEntry -> lastRxHeartbeat = otherEntry -> lastRxHeartbeat;
					resEntry -> lastTxHeartbeat = otherEntry -> lastTxHeartbeat;
if (resTrace)
	logMsg ("Found equivalent connection for %s from %s.\n", 
		 (int)(resEntry -> resourceName), 
		 (int)(otherEntry -> resourceName),
		 0,0,0,0);
					break;
					}
				otherEntry = otherEntry -> link;
				}
			
			if (otherEntry)
				continue;
			if (ntohl(ourInetAddr.s_addr) > ntohl(resEntry -> partnerInet.s_addr))
				{
#ifdef PRIV_AUDIT
				SSL *newSSL;
#endif PRIV_AUDIT
				struct sockaddr_in partner;
				partner.sin_family = AF_INET;
				partner.sin_addr = resEntry -> partnerInet;
				partner.sin_port = htons(resEntry -> dbServerPort);

if (resTrace)
	logMsg ("Attempting connect on port %d\n", 
		 resEntry -> localDbPort,0,0,0,0,0);
#ifdef PRIV_AUDIT
				newSSL = connect_ssl_connection_timed (partner.sin_addr.s_addr,
							ntohs(partner.sin_port), 10/*seconds*/);
				if (newSSL)
				    {
				    GetPeerCN (newSSL, X509_dn, sizeof(X509_dn));
				    if (GetGeneralPrivHost (X509_dn, "DB_") < 1)					
					{
					/* No priviledge to connect! */
					char msg [80];

					GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
					sprintf (msg, "DB privilege error from %s", INET_NTOA(partner.sin_addr));
					logMsg("DB priviledge error, %s not authorized. From %s\n",
						(int)X509_dn,(int)INET_NTOA(partner.sin_addr),0,0,0,0);
					SecurityAlarm (X509_dn, "DB priviledge error, Not authorized", INET_NTOA(partner.sin_addr));
					if (auditLog)
					    {
					    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_FAIL, X509_dn,
						    taskName(taskIdSelf()), msg);
					    }
					sslSharedCloseConn (newSSL);
					newSSL = NULL;
					}
				    else
					{
					GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
					if (auditLog)
					    {
					    char msg [80];

					    sprintf (msg, "DB connect from %s", INET_NTOA(partner.sin_addr));
					    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_SUCCESS, X509_dn,
						    taskName(taskIdSelf()), msg);
					    }
					}
				    }
				if (newSSL)
					{
					closeTcpStream(resEntry -> dbServerFd);
					resEntry -> dbServerFd = SSL_get_fd (newSSL);
#else PRIV_AUDIT
				if (connect(resEntry -> dbServerFd, (struct sockaddr *)&partner, sizeof(partner))
								!= ERROR)
					{
					int	true = 1;

					/*
					Connection is made. Now set our dbServerConnected
					flag and do the same for every other table row
					that was waiting for the same connection. Also,
					update the mask of fd's for the select service.
					*/

					ioctl(resEntry -> dbServerFd, FIONBIO, (int)&true);
#endif PRIV_AUDIT
					otherEntry = resTableHead;
					resEntry -> dbServerConnected = 1;
					resEntry -> disconnectTime = 0;
					resEntry -> lastRxHeartbeat = tickGet();

if (resTrace)
	logMsg ("...And it worked!\n",0,0,0,0,0,0);
					while (otherEntry)
						{
						if (otherEntry -> partnerInet.s_addr == resEntry -> partnerInet.s_addr
						 && otherEntry -> dbServerPort == resEntry -> dbServerPort)
							{
							otherEntry -> dbServerConnected = 1;
							otherEntry -> disconnectTime = 0;
							otherEntry -> dbServerFd = resEntry -> dbServerFd;
							otherEntry -> lastRxHeartbeat = tickGet();
							}
						otherEntry = otherEntry -> link;
						}
					tis_fd_set(resEntry -> dbServerFd, &tcpFds);
					}
				else
					{
					logMsg ("Connect failed, errno = %d\n", errno,0,0,0,0,0);

					/*
					Connect failed, so we should destroy this socket and go back
					one phase in our connection logic.
					*/

					closeTcpStream(resEntry -> dbServerFd);
					}

				}
			else
				{

				/*
				Ask redundant partner for his dbServer port number
				*/

				struct
					{
					short		msgType;
					RED_IOCLINKREQ	req;
					} reqMsg;
				struct sockaddr_in	partner;

				reqMsg.msgType = htons(RED_IOCDBREQ);
				reqMsg.req.redDbServerPort = htons(resEntry -> localDbPort);
				partner.sin_family = AF_INET;
				partner.sin_addr = resEntry -> partnerInet;
				partner.sin_port = htons(DBSERVER_UDP_PORT);
#ifdef USE_UDP
				sendto(bcastSocket, (char *) &reqMsg, sizeof(short) +
					sizeof(RED_IOCLINKREQ), 0, (struct sockaddr *)&partner, sizeof(partner));
#endif
if (resTrace)
	logMsg ("We don't connect so sending DBREQ message.\n",0,0,0,0,0,0);
				continue;
				}
			}
		else
			{

			/************************************************/
			/*						*/
			/*	HANDLE INPUT AND OUTPUT HEARTBEATS	*/
			/*						*/
			/************************************************/
				
				if ((tickGet() - resEntry -> lastRxHeartbeat) > CARDIAC_ARREST)
					{
					closeTcpStream(resEntry -> dbServerFd);
if (resTrace)
	logMsg ("TCP link timed out on fd %d\n", resEntry -> dbServerFd,0,0,0,0,0);
					}
				else if ((tickGet() - resEntry -> lastTxHeartbeat) > HEARTBEAT_RATE)
					{
					short	msgType = htons(RED_KEEPALIVE);
					currTcpFd = resEntry -> dbServerFd;
					currTcpSize = 0;
					sendPartner(resEntry -> dbServerFd, (char *) &msgType, sizeof(msgType));
					flushTcp();
					}
			}
		}

	{
	struct dbClient		*currClient;
	char			true = 1;
	int			newsock;
#ifdef PRIV_AUDIT
	SSL			*newSSL;
#else PRIV_AUDIT
	struct sockaddr_in	findClient;
#endif PRIV_AUDIT

	/*
	If we once had a connection to dbClient and lost it,
	try to get it back again.
	*/

	currClient = clientList;
	while (currClient)
		{
		if (currClient -> clientInet.s_addr == dbEmulIP.s_addr)
			{
			if (currClient -> clientFd)
				currClient = 0;
			break;
			}
		currClient = currClient -> next;
		}
	
	if (currClient)
		{
#ifdef PRIV_AUDIT
		newSSL = connect_ssl_connection_timed (dbEmulIP.s_addr,
					DBSERVER_UDP_PORT, 10/*seconds*/);
		if (newSSL)
		    {
		    GetPeerCN (newSSL, X509_dn, sizeof(X509_dn));
		    if (GetGeneralPrivHost (X509_dn, "DB_") < 1)					
			{
			GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
			/* No priviledge to connect! */
			logMsg("DB priviledge error, %s not authorized\n",
				(int)X509_dn,0,0,0,0,0);
			SecurityAlarm (X509_dn, "DB priviledge error, Not authorized", NULL);
			if (auditLog)
			    {
			    char msg [80];

			    sprintf (msg, "DB privilege error with %s", INET_NTOA(dbEmulIP));
			    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_FAIL, X509_dn,
				    taskName(taskIdSelf()), msg);
			    }
			sslSharedCloseConn (newSSL);
			newSSL = NULL;
			}
		    else
			{
			GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
			if (auditLog)
			    {
			    char msg [80];

			    sprintf (msg, "DB connected with %s", INET_NTOA(dbEmulIP));
			    AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_SUCCESS, X509_dn,
				    taskName(taskIdSelf()), msg);
			    }
			newsock = currClient -> clientFd = SSL_get_fd (newSSL);
			setsockopt(newsock, IPPROTO_TCP, TCP_NODELAY, &true, sizeof(true));
			if (newsock > maxTcpFd)
				maxTcpFd = newsock;
			tis_fd_set(newsock, &tcpFds);
			}
		    }
#else PRIV_AUDIT
		newsock = socket(PF_INET, SOCK_STREAM, 0);
		findClient.sin_family = AF_INET;
		findClient.sin_addr = dbEmulIP;
		findClient.sin_port = htons(DBSERVER_UDP_PORT);
		if (connect(newsock,(struct sockaddr *)&findClient, sizeof(findClient)) == ERROR)
			close(newsock);
		else
			{
			currClient -> clientFd = newsock;
			setsockopt(newsock, IPPROTO_TCP, TCP_NODELAY, &true, sizeof(true));
			if (newsock > maxTcpFd)
				maxTcpFd = newsock;
			tis_fd_set(newsock, &tcpFds);
			}
#endif PRIV_AUDIT
		}
	}


	
	/*
	If there are still any resource connection requests buffered up, we have to
	ship them out now that we have finished traversing the table.
	*/

	if (numReqs)
		{
		struct sockaddr_in bcastAddr;
		reqMsg.reqType = htons(RED_CONNREQ);
		reqMsg.reqHdr.redNumRequests = htons(numReqs);
		bcastAddr.sin_family = AF_INET;
		bcastAddr.sin_addr = bcastAddress;
		bcastAddr.sin_port = htons(DBSERVER_UDP_PORT);
#ifdef USE_UDP
		if (sendto(bcastSocket, (char *) &reqMsg,
			   sizeof(short) + sizeof(RED_CONNREQ_HDR) +
			   sizeof(RED_CONNREQ_REQ) * numReqs, 0,
			   (struct sockaddr *) &bcastAddr, sizeof(bcastAddr)) == ERROR)
				{
				logMsg ("Can't broadcast: %d\n", errno,0,0,0,0,0);
				}
#endif USE_UDP
		}

	}

	/*
	See if it is time to synchronize
	*/

	if (totPnts && ((tickGet() - syncTime) > syncDelta))
		{
		struct crcHeader
			{
			short	msgType;
			short	numSums;
			} *headLoc = (struct crcHeader *) &tcpBuff;

		syncDelta = SYNC_TRAVERSE * 60 * 60 / ((totPnts + syncPts - 1) / syncPts);
		/*
		Start by clearing out the header.
		*/

		headLoc -> msgType = htons(DBC_MSG_CRCRESPONSE);

		/*
		Get the next list of sums to send
		*/

		headLoc->numSums = findSums(tcpBuff + sizeof(struct crcHeader));
		if (headLoc->numSums)
			{
			int			size;
			size = sizeof(struct crcHeader) + headLoc->numSums * sizeof(DBC_CRCINFO);
			headLoc->numSums = htons(headLoc->numSums);
			sendClients(tcpBuff,size);
			}
		syncTime = tickGet();
		}



	/*
	Let the I/O server task know we are now looking for work.
	Then, wait for something to happen.
	*/

	{
	if (ioFlag)
		{
		ioFlag = 0;
		semGive(ioSem);
		}
	semTake(dbServerSem, WAIT_FOREVER);
	}
	}
}


/****************************************************************/
/*								*/
/*		SEND A MESSAGE TO REDUNDANT DBSERVER		*/
/*								*/
/*	This routine buffers up the messages until either the	*/
/*	buffer overflows or the fd changes.			*/
/****************************************************************/

int sendPartner(int fd, char *buffer, int msgSize)
{
	int	bytesMoved;

	/*
	Ignore send if fd is zero. Than means we had a send
	error earlier in the transaction and the stream was
	close because of it.
	*/

	if (fd == 0)
		return (ERROR);
	if (fd != currTcpFd)
		{
		flushTcp();
		currTcpFd = fd;
		}
	while (msgSize)
		{
		if (msgSize > (MAX_TCP_MSG - currTcpSize))
			bytesMoved = MAX_TCP_MSG - currTcpSize;
		else
			bytesMoved = msgSize;
		bcopy(buffer, tcpBuff + currTcpSize, bytesMoved);
		buffer += bytesMoved;
		currTcpSize += bytesMoved;
		msgSize -= bytesMoved;
		if (currTcpSize == MAX_TCP_MSG)
			if((flushTcp()) == ERROR)	return(ERROR);
		}
	return (!ERROR);
}


/****************************************************************/
/*								*/
/*			FLUSH THE TCP STREAM			*/
/*								*/
/****************************************************************/

int flushTcp(void)
{
  RES_TABLE	*resEntry;
  RES_TABLE	*otherEntry = 0;
  int		bytesSent;
  /*
    Try to send the entire buffer and see how much went out
  */
  if ((currTcpSize == 0) || (currTcpFd == 0))
    {
      currTcpSize = 0;
      return (ERROR);
    }
  resFindServerFd(currTcpFd, &resEntry);
  while(currTcpSize > 0)
    {
#ifdef PRIV_AUDIT
      bytesSent = SSL_write (sslSharedGetConn (currTcpFd), tcpBuff, currTcpSize);
#else PRIV_AUDIT
      bytesSent = send(currTcpFd, tcpBuff, currTcpSize, 0);
#endif PRIV_AUDIT
      if(bytesSent == ERROR)
	{
	  if((errno == EWOULDBLOCK) || (errno == EAGAIN))
	    {
	      taskDelay(10);
	      continue;
	    }
	  else
	    {
	      logMsg("Error writing to redundant partner: %d\n",errno,0,0,0,0,0);
	      closeTcpStream(currTcpFd);
	      return (ERROR);
	    }
	}
      else if (bytesSent == 0)
	{
	  logMsg("Zero bytes sent to redundant partner: %d\n",errno,0,0,0,0,0);
	  closeTcpStream(currTcpFd);
	  return (ERROR);
	}
      else if (bytesSent != currTcpSize)
	{
	  /*logMsg("Short write! Wanted %d, got %d\n", currTcpSize, bytesSent,0,0,0,0);*/
	  bcopy(tcpBuff + bytesSent, tcpBuff, currTcpSize);
	}
      currTcpSize -= bytesSent;
    }
  while (resFindEquivDbServer(resEntry, &otherEntry) == RES_OK)
    otherEntry -> lastTxHeartbeat = tickGet();
  resEntry -> lastTxHeartbeat = tickGet();
  return (!ERROR);
}

/****************************************************************/
/*								*/
/*		RECEIVE MESSAGE FROM REDUNDANT DBSERVER		*/
/*								*/
/****************************************************************/

int recvPartner(int fd, char *buffer, int msgSize)
{
	while (nchib < msgSize)
		if (getPartnerData(fd) == ERROR)
			{
			nchib = 0;
if (resTrace)
	logMsg ("Error reading additional data from partner\n",0,0,0,0,0,0);
			return (ERROR);
			}
	
	bcopy(rxTank + tankOutPos, buffer, msgSize);
	nchib -= msgSize;
	tankOutPos += msgSize;
	return (!ERROR);
}

/****************************************************************/
/*								*/
/*		EMPTY RECEIVE BUFFER FROM REDUNDANT PARTNER	*/
/*								*/
/****************************************************************/

int getPartnerData(int fd)
{
	RES_TABLE	*resEntry;
	RES_TABLE	*otherEntry = 0;
	fd_set		fdSel;
	struct timeval	timeOut = {1,0};
	int		bytesGot;

	if (fd == 0)
		{
		return (ERROR);
		}
	if (firstSelDone)
	    {
	    FD_ZERO(&fdSel);
	    FD_SET(fd, &fdSel);
	    if (select(fd + 1, &fdSel, NULL, NULL, &timeOut) <= 0)
	        {
    		if (resTrace)
		    logMsg ("TCP message timed out after select on fd %d\n", fd,0,0,0,0,0);
	        closeTcpStream(fd);
	        return (ERROR);
	 	}
	    }

	if (nchib)
	    bcopy(rxTank + tankOutPos, rxTank, nchib);

#ifdef PRIV_AUDIT
	if ((bytesGot = SSL_read (sslSharedGetConn (fd), rxTank + nchib, sizeof(rxTank) - nchib)) == ERROR)
	    {
	    long err;

	    int status = SSL_get_error (sslSharedGetConn (fd), bytesGot);
	    err = ERR_get_error ();
	    if (err != 0)
		logMsg ("%s\n", (int)ERR_error_string (err, NULL),0,0,0,0,0);
	    switch (status)
		{
		case SSL_ERROR_WANT_WRITE:
		case SSL_ERROR_NONE:
		case SSL_ERROR_WANT_READ:
		case SSL_ERROR_WANT_X509_LOOKUP:
		case SSL_ERROR_WANT_CONNECT:
		    /* benign, just want to write/read for state machine */
		    return (!ERROR);
		case SSL_ERROR_SSL:
		    if (errno == EWOULDBLOCK)
			return (!ERROR);
		case SSL_ERROR_SYSCALL:
		case SSL_ERROR_ZERO_RETURN:
		default:
		    printf ("Socket receive error %d\n", errno);
		    closeTcpStream(fd);
		    return (ERROR);
		}
	    }
#else PRIV_AUDIT
	if ((bytesGot = recv(fd, rxTank + nchib, sizeof(rxTank) - nchib, 0)) == ERROR)
	    {
	    if (resTrace)
		logMsg ("TCP stream error on fd %d\n", fd,0,0,0,0,0);
	    closeTcpStream(fd);
	    return (ERROR);
	    }
#endif PRIV_AUDIT

	if (!bytesGot)
	    {
	    closeTcpStream(fd);
	    return(ERROR);
	    }
	nchib += bytesGot;
	tankOutPos = 0;

	resFindServerFd(fd, &resEntry);
	while (resFindEquivDbServer(resEntry, &otherEntry) == RES_OK)
		otherEntry -> lastRxHeartbeat = tickGet();
	resEntry -> lastRxHeartbeat = tickGet();
	return (!ERROR);
}


/****************************************************************/
/*								*/
/*		CLOSE BROKEN TCP/IP STREAM			*/
/*								*/
/****************************************************************/

void closeTcpStream(int fd)
{
	RES_TABLE	*resEntry;
	RES_TABLE	*otherEntry = 0;

	if (fd == 0)
		return;
#ifdef PRIV_AUDIT
	sslSharedCloseConn (sslSharedGetConn (fd));
#else PRIV_AUDIT
	close(fd);
#endif PRIV_AUDIT
	tis_fd_clr(fd, &tcpFds);
	currTcpFd = 0;
	currTcpSize = 0;

	/*
	Now find all of the entries in the resource table
	that used this resource and mark them as disconnected.
	*/

	resFindServerFd(fd, &resEntry);
	while (resFindEquivDbServer(resEntry, &otherEntry) == RES_OK)
		{
		otherEntry -> localDbPort = 0;
		otherEntry -> dbServerPort = 0;
		otherEntry -> dbServerConnected = 0;
		otherEntry -> dbServerFd = 0;
		otherEntry -> disconnectTime = 0;
		}
	resEntry -> localDbPort = 0;
	resEntry -> dbServerPort = 0;
	resEntry -> dbServerConnected = 0;
	resEntry -> dbServerFd = 0;
	resEntry -> disconnectTime = 0;
}

/****************************************************************/
/*								*/
/*		TURN TRACE VARIABLE ON OR OFF			*/
/*								*/
/****************************************************************/

void resTraceOn(void)
{
	resTrace = 1;
}

void resTraceOff(void)
{
	resTrace = 0;
}

/********************************************************************************/
/*										*/
/*	Create the global pointer that is used to manage lists of fields	*/
/*	requiring redundant updates.						*/
/*										*/
/********************************************************************************/

void redTableAlloc(void)
{
	papRedFldList = (struct fldDes ***) calloc(dbRecType -> number + 1,
				sizeof(struct fldDes **));
}

/********************************************************************************/
/*										*/
/*		READ TCP/IP MESSAGE MAKING SURE THAT WE GET ALL OF IT!		*/
/*										*/
/********************************************************************************/

int recvTcp(int fd, char *buff, int nbytes)
{
	int		bytesWanted = nbytes;
	int		bytesGot;
	struct timeval	timeOut = {DLD_DELAY, 0};
	fd_set 		fdSel;
#ifdef PRIV_AUDIT
	SSL		*sslConn = sslSharedGetConn (fd);

	if (!sslConn)
	    return (ERROR);
#endif PRIV_AUDIT

	FD_ZERO(&fdSel);
	tis_fd_set(fd, &fdSel);

	while (nbytes > 0)
		{
#ifdef PRIV_AUDIT
		/* Must read BEFORE select in case data is in BIO */
		if ((bytesGot = SSL_read (sslConn,
					buff, nbytes)) > 0)
			{
			nbytes -= bytesGot;
			buff += bytesGot;
			continue;
			}
#endif PRIV_AUDIT
		if ((bytesGot = select(fd + 1, &fdSel, NULL, NULL, &timeOut)) <= 0)
			{
			return (ERROR);
			}
#ifdef PRIV_AUDIT
		if ((bytesGot = SSL_read (sslConn,
				    buff, nbytes)) == ERROR)
		    {
		    long err;

		    int status = SSL_get_error (sslConn, bytesGot);
		    err = ERR_get_error ();
		    if (err != 0)
			logMsg ("%s\n", (int)ERR_error_string (err, NULL),0,0,0,0,0);
		    switch (status)
			{
			case SSL_ERROR_WANT_WRITE:
			case SSL_ERROR_NONE:
			case SSL_ERROR_WANT_READ:
			case SSL_ERROR_WANT_CONNECT:
			case SSL_ERROR_WANT_X509_LOOKUP:
			    /* benign, just want to write/read for state machine */
			    continue;
			case SSL_ERROR_SSL:
			    if (errno == EWOULDBLOCK)
				continue;
			case SSL_ERROR_SYSCALL:
			case SSL_ERROR_ZERO_RETURN:
			default:
			    return (ERROR);
			}
		    }
#else PRIV_AUDIT
		if ((bytesGot = recv(fd, buff, nbytes, 0)) == ERROR)
			{
			return (ERROR);
			}
#endif PRIV_AUDIT
		if (!bytesGot)
			{
			return (ERROR);
			}
		nbytes -= bytesGot;
		buff += bytesGot;
		}
	return (bytesWanted);
}

/****************************************************************/
/*								*/
/*		SEND A MESSAGE TO ALL dbClients			*/
/*								*/
/****************************************************************/

int sendClients(char *buffer,int length)
{
	struct dbClient	*client;

	for (client = clientList; client; client = client -> next)
		{
		if (!(client -> clientFd))
			continue;
		if (npendSend(client -> clientFd, buffer, length) == ERROR)
			{
#ifdef PRIV_AUDIT
			sslSharedCloseConn (sslSharedGetConn (client -> clientFd));
#else PRIV_AUDIT
			close(client -> clientFd);
#endif PRIV_AUDIT
			tis_fd_clr(client -> clientFd, &tcpFds);
			client -> clientFd = 0;
			}
		}
	return (0);
}
/****************************************************************/
/*								*/
/*		SEND A MESSAGE TO ALL dbClients	via		*/
/*		directed UDP socket				*/
/*								*/
/*		27-Jun-1994 G. Tomalesky			*/
/*								*/
/****************************************************************/
#ifndef	BCAST_SUMS
int sendToClients(char *buffer,int length)
{
	struct dbClient	*client;
	struct	sockaddr_in	addr;

	for (client = clientList; client; client = client -> next)
		{
		if (!(client -> clientFd))
			continue;
		addr.sin_family	= AF_INET;
		addr.sin_addr	= client->clientInet;
		addr.sin_port	= htons(DBSERVER_UDP_PORT);
		sendto(bcastSocket, buffer, length, 0, (struct sockaddr *)&addr,sizeof(addr));
		}
	return (0);
}
#endif
/****************************************************************/
/*								*/
/*		FIND RECORDS FOR DATABASE SYNCHRONIZATION	*/
/*								*/
/****************************************************************/

int findSums(char *buff)
{
	int		numDone = 0;
	int		numToDo = syncPts;
	struct dbCommon	*recHdr;
	short		answer;
	short		wrap = 0;
	DBC_CRCINFO *crcinfo = (DBC_CRCINFO *)buff;

	while (numToDo)
		{
		while (!syncNext)
			{
			syncType++;
			if (syncType >= dbRecType -> number)
				{
				if (wrap)
					return (numDone);
				wrap = 1;
				syncType = 0;
				}
			if (dbRecords -> papRecLoc[syncType])
				syncNext = (struct recChain *)(dbRecords ->
						papRecLoc[syncType] -> pFirst);
			}
		/* test contents of record pointer. if 0 then record has been deleted */
		if (!(syncNext -> pRecord[0]))
			{
			syncNext = syncNext -> next;
			continue;
			}
		
		numToDo--;
		recHdr = (struct dbCommon *)(syncNext -> pRecord);
		resGetCaAnswer(recHdr -> radr, &answer);
		if (answer)
			{
			/*
				3-May-1999 GCT. The new CRC transmission scheme reqires a lnown size of
				message. using a structure such as DBC_CRCINFO will allow me to provide
				one. It also makes it easier to see what is going on.
			*/
			crcinfo->crcKey = htonl((int)recHdr);
			crcinfo->crcTimeStamp = htonl((int)(recHdr->tstp));
			crcinfo->crcCRC = htons(recHdr -> rcrc);
			crcinfo++;
			numDone++;
			}
		syncNext = syncNext -> next;
		}
	return (numDone);
}

/****************************************************************/
/*								*/
/*		HANDLE I/O ERROR ON SOCKET TO dbClient		*/
/*								*/
/****************************************************************/

int tcpErr(int fd)
{
	struct dbClient		*client = clientList;

	tis_fd_clr(fd, &tcpFds);
	while (client)
		{
		if (client -> clientFd == fd)
			{
#ifdef PRIV_AUDIT
			sslSharedCloseConn (sslSharedGetConn (fd));
#else PRIV_AUDIT
			close(fd);
#endif PRIV_AUDIT
			client -> clientFd = 0;
			deleteOlAdds(client);
			return (!ERROR);
			}
		client = client -> next;
		}
	return (ERROR);
}

void rejectDeletion(short code,int from,int origFd,short refs)
{
	DBC_DELSTATUS ds;
	short msgType;

	msgType = htons(DBC_MSG_DELSTATUS);
	ds.deletionStatus = htons(code);
	ds.originatorFd = htonl(origFd);
	ds.numRefs = htons(refs);
	if(npendSend(from,(char *)&msgType,sizeof(msgType)) == ERROR)
		{
		tcpErr(from);
		return;
		}
	if(npendSend(from,(char *)&ds,sizeof(DBC_DELSTATUS)) == ERROR)
		{
		tcpErr(from);
		return;
		}
}

/****************************************************************/
/*								*/
/*	Customized versions of FD_SET and FD_CLR		*/
/*								*/
/****************************************************************/


void tis_fd_set(int fd, fd_set *fdMask)
{
	FD_SET(fd, fdMask);
}

void tis_fd_clr(int fd, fd_set *fdMask)
{
	FD_CLR(fd, fdMask);
}


/****************************************************************/
/*								*/
/*		SOCKET I/O SERVER TASK				*/
/*								*/
/****************************************************************/

void redundIO(void)
{
	struct timeval ioWait = {1,0};
	fd_set	ourFds;
	int	numChans;
	int	i;
	int	limit;

	limit = sizeof(fd_set) / sizeof(fd_mask);
	for (;;)
		{
		for (i = 0; i < limit; i++)
			ourFds.fds_bits[i] = bcastReadFds.fds_bits[i] | tcpFds.fds_bits[i];
		if ((numChans = select(maxTcpFd + 1, &ourFds, NULL, NULL, &ioWait)) >= 0)
			{
			ioFlag = 1;
			semGive(dbServerSem);
			semTake(ioSem, WAIT_FOREVER);
			}
		else
			logMsg ("Select error in redundIO, %d\n", errno,0,0,0,0,0);
		}
}

int setTraceName(char *name)
{
	strcpy(traceName, name);
	return (1);
}

/****************************************************************/
/*								*/
/*		REPORT DATABASE SIZE AND SOURCE NODE		*/
/*								*/
/****************************************************************/

int showDownload(void)
{
	char	hostName[100];

	/*
	The host that downloaded us is always the first one in clientList.
	*/

	if (hostGetByAddr(clientList -> clientInet.s_addr, hostName) == ERROR)
		strcpy(hostName, "<node name not available>");
	logMsg ("%d points downloaded from address %s (%s)\n",
		 totPnts,
		 (int)((char *)INET_NTOA(clientList -> clientInet)),
		 (int)hostName,0,0,0);
	return (0);
}
/*
	These functions are meant to make the operations of SDR & Database downloading
	somewhat modular so that I can call them when needed. Both functions will have
	the network socket as a parameter. Zero will be an error return for both. The 
	database download functions will return the number of points actually downoaded.
	
	The socket might be from to different clients. One (or both) may be for the local
	client (dbEmulator) or if different the remote client (dbClient) socket would also
	be used.
	
	The dbGetDBSocket function attempts to establish a socket on which to communicate
	with a database client. It first trys to talk to the local client (dbEmulator) task.
	If this fails it trys a remote client (dbClient).

	05-Apr-1994	GCT
*/
int	dbGetLocalClient(void)
{
    int	download;
#ifndef PRIV_AUDIT
    struct sockaddr_in	findClient;
    int	true = 1;
    int	status;
#endif PRIV_AUDIT
    char opt = 1;
    /*
	    Attempt to connect with local client first. This will fail on IOCs that do not have
	    EAPROMS. If succesful then make this fd the first in the clientList.

	    GCT 07-Apr-1994
    */
#ifdef PRIV_AUDIT
    SSL *newSSL = connect_ssl_connection_timed (hostGetByName("localhost"),
			    DBSERVER_UDP_PORT, 10/*seconds*/);
    download = -1;
    if (newSSL)
	{
	GetPeerCN (newSSL, X509_dn, sizeof(X509_dn));
	if (GetGeneralPrivHost (X509_dn, "DB_") < 1)					
	    {
	    /* No priviledge to connect! */
	    GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
	    logMsg("DB priviledge error, %s not authorized\n",
		    (int)X509_dn,0,0,0,0,0);
	    SecurityAlarm (X509_dn, "DB priviledge error, Not authorized", NULL);
	    if (auditLog)
		{
		AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_FAIL, X509_dn,
			taskName(taskIdSelf()), "Client privilege error with localhost");
		}
	    sslSharedCloseConn (newSSL);
	    newSSL = NULL;
	    }
	else
	    {
	    GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
	    if (auditLog)
		{
		AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_SUCCESS, X509_dn,
			taskName(taskIdSelf()), "Client connect to localhost");
		}
	    setsockopt (SSL_get_fd (newSSL), IPPROTO_TCP, TCP_NODELAY, &opt, sizeof(opt));
	    }
	}
    if (newSSL)
	{
	download = SSL_get_fd (newSSL);
#else PRIV_AUDIT
    download = socket(PF_INET, SOCK_STREAM, 0);
    setsockopt(download, IPPROTO_TCP, TCP_NODELAY, &opt, sizeof(opt));
    findClient.sin_family = AF_INET;
    findClient.sin_addr.s_addr = hostGetByName("localhost");
    findClient.sin_port = htons(DBSERVER_UDP_PORT);
    if(connect(download, (struct sockaddr *)&findClient, sizeof(findClient)) == ERROR)
	{
	close(download);		/* can't connect...try the remote client */
	download = -1;
	}
    else
	{
	/*
	Got a connection. Remember it for when we start
	sending updates to dbClient.
	*/
	status = ioctl(download, FIONBIO, (int)&true);
	if (status == ERROR)
	    perror("dbserver can't set non blocking on socket to dbEmulator");
#endif PRIV_AUDIT
	if (download > maxTcpFd)
		{
		maxTcpFd = download;
		tis_fd_set(download, &tcpFds);
		}
	if (!clientList)
		{
		clientList = (struct dbClient *) calloc(1, sizeof(struct dbClient));
		}
	else
		{
		bzero((char *) clientList, sizeof(struct dbClient));
		}
	clientList -> clientInet.s_addr = hostGetByName("localhost");
	dbEmulIP = clientList -> clientInet;
	clientList -> clientFd = download;
	}
    return(download);			/* got local client...return with it */
}
int	dbGetRemoteClient(int bcast)
{
#ifdef USE_UDP
    short		msgType;
#endif
    struct sockaddr_in	findClient;
    fd_set		ourFd;
    int			download;
    int			bcastTime;
    struct dbClient	*prev,*curr;
    struct timeval	dldDelay = {DLD_DELAY, 0};
    int newFd;
#ifndef PRIV_AUDIT
    char opt = 1;
#else PRIV_AUDIT
    SSL			*newSSL;
#endif PRIV_AUDIT

    /*
	    Start by finding a dbClient that is willing to work with us.
    */
	download = -1;
#ifdef USE_UDP
	msgType = htons(DBC_MSG_AWAKE);
	findClient.sin_family = AF_INET;
	findClient.sin_addr = bcastAddress;
	findClient.sin_port = htons(DBSERVER_UDP_PORT);
	sendto(bcast, (char *) &msgType, sizeof(msgType), 0,
		(struct sockaddr *) &findClient, sizeof(findClient));
	logMsg ("Broadcasting.\n",0,0,0,0,0,0);
#endif USE_UDP
	bcastTime = tickGet();
#ifdef USE_UDP
	bcopy((char *) &bcastReadFds, (char *) &ourFd, sizeof(ourFd));
#else USE_UDP
	bcopy((char *) &tcpFds, (char *) &ourFd, sizeof(ourFd));
#endif USE_UDP
	dldDelay.tv_sec = 5;
	while (select(maxTcpFd + 1, &ourFd, NULL, NULL, &dldDelay) > 0)
	    {
	    int	fromSize = sizeof(findClient);
#ifdef USE_UDP
		int	udpLength;
	    if ((udpLength = recvfrom(bcast, (char *) &msgType, sizeof(msgType),
			    0, (struct sockaddr *)&findClient, &fromSize)) == ERROR)
		break;
	    if (!udpLength)
		{
			logMsg ("Zero length UDP message ignored.\n",0,0,0,0,0,0);
			continue;
		}
	    if (ntohs(msgType) == DBC_MSG_WAKEUP)
		{
#else USE_UDP
		newFd = accept(tcpListenSocket, (struct sockaddr *)&findClient,&fromSize);
		close(newFd);
#endif USE_UDP
#ifdef PRIV_AUDIT
		newSSL = connect_ssl_connection_timed (findClient.sin_addr.s_addr,
					DBSERVER_UDP_PORT, 10/*seconds*/);
		if (newSSL)
		    {
				GetPeerCN (newSSL, X509_dn, sizeof(X509_dn));
				if (GetGeneralPrivHost (X509_dn, "DB_") < 1)					
				{
			/* No priviledge to connect! */
					GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
					logMsg("DB priviledge error, %s not authorized\n",
						(int)X509_dn,0,0,0,0,0);
					SecurityAlarm (X509_dn, "DB priviledge error, Not authorized", NULL);
					if (auditLog)
					{
						char msg [80];

						sprintf (msg, "Client privilege error with %s", INET_NTOA(findClient.sin_addr));
						AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_FAIL, X509_dn,
							taskName(taskIdSelf()), msg);
					}
				sslSharedCloseConn (newSSL);
				newSSL = NULL;
				}
				else
				{
					char opt = 1;
					GetPeerDN (newSSL, X509_dn, sizeof(X509_dn));
					if (auditLog)
					{
						char msg [80];
	
						sprintf (msg, "Client connected with %s", INET_NTOA(findClient.sin_addr));
						AuditAddEvent (auditLog, AUDIT_LOGON, AUDIT_SUCCESS, X509_dn,
						    taskName(taskIdSelf()), msg);
				    }
					download = SSL_get_fd (newSSL);
					setsockopt (download, IPPROTO_TCP, TCP_NODELAY, &opt, sizeof(opt));
				}
		    }
#else PRIV_AUDIT
		download = socket(AF_INET, SOCK_STREAM, 0);
		setsockopt(download, IPPROTO_TCP, TCP_NODELAY, &opt, sizeof(opt));
		findClient.sin_port = htons(DBSERVER_UDP_PORT);
		if (connect(download, (struct sockaddr *)&findClient, sizeof(findClient)) == ERROR)
		{
		    close(download);
		    download = -1;
		}
		else
#endif PRIV_AUDIT
		{
		    int	true = 1;

		    /*
		    Got a connection. Remember it for when we start
		    sending updates to dbClient.
		    */

		    ioctl(download, FIONBIO, (int)&true);
		    if (download > maxTcpFd)
			{
				maxTcpFd = download;
			}
/*
Add this client to my list of clients I have heard from. This is little
more complex than it used to be because the remote might not be the first
client I have connected with. If it is then this should work as before. If
a local client is present then it would be first and the remote would be
second.
*/
		    if (!clientList)
			{
				clientList = (struct dbClient *) calloc(1, sizeof(struct dbClient));
				curr = clientList;
			}
		    else
			{
			/*bzero((char *) clientList, sizeof(struct dbClient));*/
				prev = 0;
				curr = clientList;
				while(curr)
			    {
					if (curr -> clientInet.s_addr == findClient.sin_addr.s_addr)
						break;
					prev = curr;
					curr = curr->next;
			    }
			if (!curr)
			    {
					curr = (struct dbClient *) calloc(1, sizeof(struct dbClient));
					prev->next = curr;
			    }
			}
		    curr -> clientInet.s_addr = findClient.sin_addr.s_addr;
		    curr -> clientFd = download;
		    break;
		    }
#ifdef USE_UDP
		}
	    else
		{
		if ((tickGet() - bcastTime) * 60 > DLD_DELAY)
			break;
		}
#endif USE_UDP
	}
	if (download != -1)
	    {
	    tis_fd_set(download, &tcpFds);
	    }
	return(download);	/* return with socket status */
}

int	dbSDRDownload(int fd)
{
    char    filename[50];
    char    *cp;

    /*
	    Download the sdr records.
    */
    void		(*dbEmulatorFunc)();
    SYM_TYPE	symtemp;
    logMsg ("SDR Load Starting.\n",0,0,0,0,0,0);
    fd = 1;

    /*
    If the dbEmulator is loaded, then do NOT load the SDR now. That has already been
    done in the startup.cmd!
    */
    if (symFindByName(sysSymTbl, "_dbEmulator", (char **)&dbEmulatorFunc, &symtemp) == ERROR)
	sdrLoad("/tis4000db/default.dctsdr");
    else
	{
	if ((cp = getenv ("SDR:")) != NULL)
	    {
	    sprintf (filename, "%sDEFAULT.SDR", cp);
	    sdrLoad(filename);
	    }
	else
	    {
	    sdrLoad("SDR:DEFAULT.SDR");
	    }
	}
    return(fd);
}
int	dbDBDownload(int fd,void (*e2putfunc)())
{
    DBC_DWNLD_REQUEST	request;
    short	msgType;
    struct dbCommon		*recHdr;
    unsigned short localCRC;
    request.msgLastName[0] = 0;
    request.msgStatus = DSTAT_OK;
    logMsg ("Database Load Starting.\n",0,0,0,0,0,0);

    for (;;)
	{
	struct recLoc	*loc;
	struct recChain	*chain;
	short		recType;
	short		numRecs;

	/*
	Request the next record from dbClient.
	*/

	msgType = htons(DBC_MSG_DWNLDREQ);
	if (npendSend(fd, (char *) &msgType, sizeof(msgType)) == ERROR)
		{
#ifdef PRIV_AUDIT
		sslSharedCloseConn (sslSharedGetConn (fd));
#else PRIV_AUDIT
		close(fd);
#endif PRIV_AUDIT
		tis_fd_clr(fd, &tcpFds);
		fd = -1;
		break;
		}
	if (npendSend(fd, (char *) &request, sizeof(request)) == ERROR)
		{
#ifdef PRIV_AUDIT
		sslSharedCloseConn (sslSharedGetConn (fd));
#else PRIV_AUDIT
		close(fd);
#endif PRIV_AUDIT
		tis_fd_clr(fd, &tcpFds);
		fd = -1;
		break;
		}

	/*
	Get the response to our request.
	*/

	if (recvTcp(fd, (char *) &msgType, sizeof(msgType)) == ERROR)
		{
#ifdef PRIV_AUDIT
		sslSharedCloseConn (sslSharedGetConn (fd));
#else PRIV_AUDIT
		close(fd);
#endif PRIV_AUDIT
		tis_fd_clr(fd, &tcpFds);
		fd = -1;
		break;
		}
	if (ntohs(msgType) != DBC_MSG_DWNLDRESP)
		{
		logMsg ("Badly formed record fd response\n",0,0,0,0,0,0);
#ifdef PRIV_AUDIT
		sslSharedCloseConn (sslSharedGetConn (fd));
#else PRIV_AUDIT
		close(fd);
#endif PRIV_AUDIT
		tis_fd_clr(fd, &tcpFds);
		fd = -1;
		break;
		}
	if (recvTcp(fd, (char *) &numRecs, sizeof(numRecs)) == ERROR)
		{
#ifdef PRIV_AUDIT
		sslSharedCloseConn (sslSharedGetConn (fd));
#else PRIV_AUDIT
		close(fd);
#endif PRIV_AUDIT
		tis_fd_clr(fd, &tcpFds);
		fd = -1;
		break;
		}
	
	/*
	Get the number of records that he promises to send. A zero means that
	we are done.
	*/

	numRecs = ntohs(numRecs);
	if (numRecs == 0)
		break;
	while (numRecs--)
		{
		/*
		Get the record type so we know the length
		*/

		if (recvTcp(fd, (char *) &recType, sizeof(recType)) == ERROR)
			{
#ifdef PRIV_AUDIT
			sslSharedCloseConn (sslSharedGetConn (fd));
#else PRIV_AUDIT
			close(fd);
#endif PRIV_AUDIT
			tis_fd_clr(fd, &tcpFds);
			fd = -1;
			break;
			}
		
		/*
		Create a home for this record
		*/

		recType = ntohs(recType);
		loc = dbRecords -> papRecLoc[recType];
		if (!loc)
			{
			loc = (struct recLoc *) calloc(1, sizeof(struct recLoc));
			dbRecords -> papRecLoc[recType] = loc;
			loc -> record_type = recType;
			loc -> rec_size = dbRecDes -> papRecTypDes[recType] -> rec_size;
			}
		chain = (struct recChain *) calloc(1, sizeof(struct recChain) + loc -> rec_size);
		chain -> next = (struct recChain *) (loc -> pFirst);
		chain -> pRecord = (caddr_t)((char *)chain + sizeof(struct recChain));
		loc -> pFirst = (caddr_t) chain;
		if (recvTcp(fd, (char *) chain -> pRecord, loc -> rec_size) == ERROR)
			{
#ifdef PRIV_AUDIT
			sslSharedCloseConn (sslSharedGetConn (fd));
#else PRIV_AUDIT
			close(fd);
#endif PRIV_AUDIT
			tis_fd_clr(fd, &tcpFds);
			fd = -1;
			break;
			}
		/*
		  GCT 11-Jun-1997

		  Re-compute CRCs and compare with value contained in the record just
		  received. If the same then proceed as normal otherwise abort the
		  download and delete whatever database has been created so far.
		*/
		recHdr = (struct dbCommon *)chain -> pRecord;
		localCRC = DBC_CRCRecord((char *)recHdr,recType);	/* re-compute CRC */
		if(localCRC != ntohs(recHdr->rcrc))			/* compare with one in record */
			{
#ifdef PRIV_AUDIT
			sslSharedCloseConn (sslSharedGetConn (fd));
#else PRIV_AUDIT
			close(fd);
#endif PRIV_AUDIT
			tis_fd_clr(fd, &tcpFds);
			fd = -1;
			logMsg ("CRC mis-match for %s old=%04X new=%04X\n",
			         (int)((((struct dbCommon *)(chain -> pRecord)) -> name)),
				 ntohs(recHdr->rcrc),
				 localCRC,0,0,0);
			break;
			}
		logMsg ("%s\n",(int)((((struct dbCommon *)(chain -> pRecord)) -> name)),
			 0,0,0,0,0);
/*
	Add to the EEROM database in net std order!
*/
		if(e2putfunc != 0)
			e2putfunc(chain->pRecord, recType, chain->pRecord);

		localEndian((char *) chain -> pRecord, recType);
		add_pv((char *)(((struct dbCommon *) (chain -> pRecord)) -> name), recType, 
					loc -> no_records);
		pvdSetAddr((char *)(((struct dbCommon *) (chain -> pRecord)) -> name), chain -> pRecord);
		loc -> no_records++;
		loc -> no_used++;
		strcpy(request.msgLastName, ((struct dbCommon *)(chain -> pRecord)) -> name);
		totPnts++;
		}
	if (fd == -1)
		break;
	}
if(fd == -1)
	{
	logMsg ("Database download aborted\n",0,0,0,0,0,0);
	deleteDatabase();
	totPnts = ERROR;
	}
else
	logMsg ("Database Load Complete\n",0,0,0,0,0,0);
return(totPnts);
}

/****************************************************************/
/*								*/
/*		SET THE SIZE OF THE REDUNDANT UPDATE QUEUE	*/
/*								*/
/*	This can only be done before iocInit is called.		*/
/*	The size is specified in K bytes. The utility routine	*/
/*	will multiply by 1024 for you.				*/
/*								*/
/****************************************************************/

int setRedQueueSize(int size)
{
	redQueueSize = size * 1024;
	return (0);
}


/****************************************************************/
/*								*/
/*		SEND NON-PENDED MESSAGE TO DBCLIENT		*/
/*								*/
/****************************************************************/

int npendSend(int fd, char *buff, int nbytes)
{
  int	totalSent = 0;
  int	bytesSent;
  /*
    Try to send the entire buffer and see how much went out
  */
  while(nbytes > 0)
    {
#ifdef PRIV_AUDIT
      bytesSent = SSL_write (sslSharedGetConn(fd), &buff[totalSent], nbytes);
#else PRIV_AUDIT
      bytesSent = send(fd, &buff[totalSent], nbytes, 0);
#endif PRIV_AUDIT
      if (bytesSent == ERROR)
	{
	  if((errno == EWOULDBLOCK) || (errno == EAGAIN))
	    {
	      taskDelay(10);
	      continue;
	    }
	  else
	    {
	      logMsg("Error writing to dbClient: %d\n",errno,0,0,0,0,0);
	      return (ERROR);
	    }
	}
      else if (bytesSent == 0)
	{
	  logMsg("Zero bytes sent to dbClient: %d\n",errno,0,0,0,0,0);
	  return (ERROR);
	}
      totalSent += bytesSent;
      nbytes -= bytesSent;
    }
  return (!ERROR);}

/****************************************************************/
/*								*/
/*	DELETE ALL RECORDS PENDING ONLINE ADDITION		*/
/*								*/
/*	We have been queueing up records to be online added	*/
/*	to the database. Something went wrong with the		*/
/*	connection, so now it is time to discard all of the	*/
/*	records so that the process can start over as a		*/
/*	complete entity. We don't want to have half the job	*/
/*	done.							*/
/*								*/
/****************************************************************/

void deleteOlAdds(struct dbClient *client)
{
    struct olAdd	*currAdd;
    struct olAdd	*nextAdd;

    currAdd = client -> firstOlAdd;
    while(currAdd)
	{
	nextAdd = (struct olAdd *) currAdd -> newRec.next;
	free(currAdd);
	currAdd = nextAdd;
	}
    client -> firstOlAdd = 0;
}
