/*freeMemory.c,v 1.1.1.1 1999/08/03 14:44:27 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/****************************************************************/
/*								*/
/*		FREE DYAMICALLY ACQUIRED RECORD MEMORY		*/
/*								*/
/*	The purpose of this routine is to free the memory	*/
/*	that may have been dynamically acquired during record	*/
/*	initializaiton. When records are modified by adding	*/
/*	dynamically acquired buffers, this routine will have	*/
/*	to know about them.					*/
/*								*/
/****************************************************************/

/******************** REVISION HISTORY *************************
*
*  REVISED BY 		: ems
*  REVISION DATE 	: 08 dec 93
*  REVISION: added deadtime, pto, sigchar and td,
*
*  REVISED BY 		: ems
*  REVISION DATE 	: 11 Jan 94
*  REVISION: deleted compress, histogram, sub, not supported
*
*  REVISED BY 		: ems
*  REVISION DATE 	: 22 Feb 96
*  REVISION:	added bo, dpvt to waveform, pat
*
****************************************************************/

/* NOTE: future, free up dbAddr structs that were alloc'd as
	 part of db_link struct, see .../epicsH/link.h
*/


#include <vxWorks.h>
#include <wdLib.h>
#include <types.h>
#include <memLib.h>
#include <string.h>
#include <stdlib.h>

#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#include "dbAccess.h"	/* ems, needed for deadtime */
#include "dbRecType.h"
#include "callbacks.h"
#include "redund.h"
#include "rec/aiRecord.h"
#include "rec/biRecord.h"
#include "rec/boRecord.h"
#include "rec/calcRecord.h"
#include "rec/deadtimeRecord.h"
#include "rec/patRecord.h"
#include "rec/ptoRecord.h"
#include "rec/selRecord.h"
#include "rec/sigcharRecord.h"
#include "rec/tdRecord.h"
#include "rec/waveformRecord.h"

int freeRecMem(struct dbCommon *recHdr, short recType)
{
	char		*label = dbRecType -> papName[recType];
	CALLBACK_PTR	pcallback;

	if (!label)
		return (ERROR);

	if (!strcmp(label, "ai"))
		{
		struct aiRecord *ai = (struct aiRecord *) recHdr;
		if (ai -> rocb)
			free(ai -> rocb);
		}
	else if (!strcmp(label, "bi"))
		{
		struct biRecord *bi = (struct biRecord *) recHdr;
		if (bi -> dpvt)
			free(bi -> dpvt);
		}
	else if (!strcmp(label, "bo"))
		{
		    struct boRecord *bo = (struct boRecord *) recHdr;
		    if (bo -> dpvt)
		    {
			pcallback = (CALLBACK_PTR) (bo -> dpvt);
			if (pcallback -> wd_id != 0)
			{
			    wdCancel (pcallback -> wd_id);
			    wdDelete (pcallback -> wd_id);
			}
			/* if the callback is active, call the 
			   callback routine
			*/
			if (bo -> cbfl == 2)
			    (*pcallback -> callback)(pcallback);
			free(bo -> dpvt);
		    }
		}
	else if (!strcmp(label, "calc"))
		{
		struct calcRecord *calc = (struct calcRecord *) recHdr;
		if (calc -> rocb)
			free(calc -> rocb);
		}
	else if (!strcmp(label, "sel"))
		{
		struct selRecord *sel = (struct selRecord *) recHdr;
		if (sel -> rocb)
			free(sel -> rocb);
		}
	else if (!strcmp(label, "waveform"))
		{
		struct waveformRecord *waveform = (struct waveformRecord *) recHdr;
		if (waveform -> bptr)
			free(waveform -> bptr);
		if (waveform -> val)
			free(waveform -> val);
		if (waveform -> dpvt)
			free(waveform -> dpvt);
		}
	else if (!strcmp (label, "pto"))
		{
		    struct ptoRecord *pto = (struct ptoRecord *) recHdr;
		    if (pto -> dpvt)
		    {
			pcallback = (CALLBACK_PTR) (pto -> dpvt);
			if (pcallback -> wd_id != 0)
			{
			    wdCancel (pcallback -> wd_id);
			    wdDelete (pcallback -> wd_id);
			}
			free (pto -> dpvt);
		    }
		}
	else if (!strcmp (label, "deadtime"))
		{
		struct deadtimeRecord *dt = (struct deadtimeRecord *) recHdr;
		if (dt -> detb)
			free (dt -> detb);
		if (dt -> stab)
			free (dt -> stab);
		if (dt -> sevb)
			free (dt -> sevb);
		}
	else if (!strcmp (label, "td"))
		{
		    struct tdRecord *td = (struct tdRecord *) recHdr;
		    if (td -> tdcb)
		    {
			pcallback = (CALLBACK_PTR) (td -> tdcb);
			if (pcallback -> wd_id != 0)
			{
			    wdCancel (pcallback -> wd_id);
			    wdDelete (pcallback -> wd_id);
			}
			free (td -> tdcb);
		    }
		}
	else if (!strcmp (label, "sigchar"))
		{
		struct sigcharRecord *sig = (struct sigcharRecord *) recHdr;
		if (sig -> rocb)
			free (sig -> rocb);
		}
	else if (!strcmp(label, "pat"))
		{
		    struct patRecord *pat = (struct patRecord *) recHdr;
		    if (pat -> dpvt)
		    {
			pcallback = (CALLBACK_PTR) (pat -> dpvt);
			if (pcallback -> wd_id != 0)
			{
			    wdCancel (pcallback -> wd_id);
			    wdDelete (pcallback -> wd_id);
			}
			free(pat -> dpvt);
		    }
		}
	return (!ERROR);
}
