/*pvdOnTheFly.c,v 1.1.1.1 1999/08/03 14:44:28 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/****************************************************************/
/*								*/
/*	ROUTINES TO CREATE THE PROCESS VARIABLE DIRECTORY	*/
/*								*/
/****************************************************************/


#include <string.h>
#include <stdio.h>
#include <fioLib.h>
#include <stdlib.h>
#include <memLib.h>

#include "dbPvd.h"
#include "dbRecType.h"
#include "dbCommon.h"
#include "redund.h"
#include "pvd.h"

static int findslot(void);

int add_pv(char *input,short type,short record_no)
{
	struct pvd_entry	*pentry;
	short		hshval;
	unsigned short	curr;
	short		slot;

        if (!dbPvd)
                return(-4);

	/* get the hash pointer */
	hshval = pvdHash(input);

	/* get an available slot in the process variable directory */
	if ((slot = findslot()) < 0) {
		printErr("The pvd is full!\n");
		return(-3);
	}	
	/* chain new PV slot to the collision list in the hash table */
	curr = dbPvd->hash_entry[hshval];
	dbPvd->hash_entry[hshval] = slot;
	
	/* place process variable in directory */
	pentry = &dbPvd->entry[slot];
	pentry->next = curr;
	pentry->record_type = type;
	pentry->record_no = record_no;
	strncpy(pentry->name,input,PVNAME_SZ);

	return(0);
}

static int findslot(void)
{
	short	i;
	struct pvd_entry	*pentry;
	

	/* get an available elt entry */
	pentry = dbPvd->entry;
	for (i = 0;i < MAX_PV; i++,pentry++)
		if (!(pentry->name[0])) return(i);
	
	return(-1);
}

int	pvd_create(void)
{
	struct pvd_entry	entry;
	struct pvd_entry	*pentry = &entry;
	int		i;

	if (!dbPvd){
		if ((dbPvd = (struct pvd *)malloc( sizeof(struct pvd))) == NULL){
			printErr("Can't malloc pvd!\n");
			return(-1);
		}
	}

	/* initialize the hash table */

	for (i = 0; i < HASH_NO; i++)
		dbPvd->hash_entry[i] = NO_ENTRY;
	
	pentry->name[0] = 0;
	pentry->record_type = 0;
	pentry->record_no = 0;
	pentry->next = NO_ENTRY;
	for (i = 0; i < MAX_PV; i++)
		bcopy((char *)pentry, (char *)&dbPvd->entry[i],
				sizeof(struct pvd_entry));

	return(0);
}

int delete_from_pvd(char *name)
{
	struct pvd_entry	*pentry;
	struct pvd	*ppvd;		/* set to dbPvd */
	short		hshval;		/* hash value for pv	*/
	short	prevptr;	/* pointer to previous pv */
	short	currptr;	/* pointer to current pv */
	short		fwdptr;		/* pointer to next pv	*/
	short		i;

	ppvd = dbPvd;

	/* get the hash pointer */
	hshval = pvdHash(name);

	/* find the process variable in the table */
	currptr = ppvd->hash_entry[hshval];
	prevptr = NO_ENTRY;
	for (i = strlen((char *)name); i < PVNAME_SZ; i++)	/* NULL fill the name*/
			 *(name + i) = 0;
	while ((strncmp(name,ppvd->entry[currptr].name, PVNAME_SZ) != 0)
	  && (currptr != NO_ENTRY)){
		prevptr = currptr;
		currptr = ppvd->entry[currptr].next;
	}
	
	fwdptr = ppvd->entry[currptr].next;

	/* unlink pv from hash chain */
	if (prevptr == NO_ENTRY){
		ppvd->hash_entry[hshval] = fwdptr;
	}else{
		ppvd->entry[prevptr].next = fwdptr;
	}
	
	/* delete the pv record */
	pentry = &ppvd->entry[currptr];
	strncpy(pentry->name,"                            ",PVNAME_SZ);
	pentry->record_type = 0;
	pentry->record_no = 0;
	pentry->next = NO_ENTRY;
	
	return(0);
}
