/*redqlib.c,v 1.1.1.1 1999/08/03 14:44:28 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/****************************************************************/
/*								*/
/*		REDUNDANT UPDATE MESSAGE QUEUE LIBRARY		*/
/*								*/
/*	This library consists of some small routines that	*/
/*	make it easier for the record processing routines	*/
/*	and channel access to put redundant updates into	*/
/*	the queue that informs the dbServer task about the	*/
/*	change so that he can propagate it to all the other	*/
/*	places that it has to go.				*/
/*								*/
/****************************************************************/

/****** MOD LOG
01/30/91	GMU	410-581-0422	uchenick@tate.com
			Original entry.
03/18/94	ems	redQSend, if rduf set, do not trash what
			is in buffer, just add to it making sure
			do not duplicate data
01/20/95	ems	added redQNPSend
08/14/1998	GCT	made function declarations ANSI std.
******/

/****************************************************************/
/*								*/
/*			INCLUDE FILES				*/
/*								*/
/****************************************************************/

#include <vxWorks.h>
#include <inetLib.h>
#include <types.h>
#include <errno.h>
#include <semLib.h>
#include <string.h>
#include <stdio.h>
#include <fioLib.h>
#include <tickLib.h>
#include <logLib.h>

#include "dbDefs.h"
#include "dbCommon.h"
#include "dbRecType.h"
#include "redque.h"
#include "redflds.h"
#include "restable.h"

extern SEM_ID	dbServerSem;
static int	redundantUpdates = 0;
static int	redundantFails = 0;
static int firstTime1 = 1;	       	/* these are used to gate the logMsg delay notification */
static int firstTime2 = 1;
#define THREE_MINUTES	(3*60*60)	/* 60 ticks/sec * 60 sec/min * 3 min */
extern int startTime;			/* defined in dbserver.c */
int sendDelay = THREE_MINUTES;		/* this is global so it can be tuned in the startup.cmd file */

/****************************************************************/
/*								*/
/*		INITIALIZE THE QUEUE BUFFER			*/
/*								*/
/****************************************************************/

int redQInit(struct dbCommon *pRecord, char *buffer,int buffSize,int recType)
{
	struct redQBuffHdr	*buffHeader;

	if (!buffer)
		return (-1);	/* Null buffer pointer */
	buffHeader = (struct redQBuffHdr *) buffer;
	if (!pRecord)
		return (-1);	/* Null record pointer */
	buffHeader -> recordAddr = pRecord;
	if (buffSize <= 0)
		return (-1);	/* Invalid size */
	buffHeader -> totalSize = buffSize
		- sizeof (buffHeader -> totalSize);

	if (!GET_PRECTYPE(recType))
		return (-1);	/* Invalid record type */
	buffHeader -> pointType = recType;

	buffHeader -> currentSize = sizeof(buffHeader -> recordAddr)
			+ sizeof(buffHeader -> currentSize)
			+ sizeof(buffHeader -> pointType)
			+ sizeof(buffHeader -> promptFlag)
			+ sizeof(buffHeader -> numFields);
	buffHeader -> numFields = 0;
	buffHeader -> promptFlag = 0;
	return (0);
}


/****************************************************************/
/*								*/
/*		INSERT FIELD INTO THE CURRENT MESSAGE		*/
/*								*/
/****************************************************************/

int redQField(char *buffer, struct fldDes *pfldDes)
{
	struct redQBuffHdr	*buffHeader;

	if (!buffer)
		return (-1);	/* Null buffer pointer */
	buffHeader = (struct redQBuffHdr *) buffer;
	if (!pfldDes)
		return (-1);	/* Null field pointer */
	
	/*
	First see if there is room in the buffer to stash
	pfldDes and the new value.
	*/

	if ((buffHeader -> currentSize + sizeof(pfldDes)
			+ pfldDes -> size) > buffHeader -> totalSize)
		{
		printErr("Redundant update Q overflow while processing %s\n",
			buffHeader -> recordAddr -> name);
		return (-1);	/* Buffer overflow */
		}
	
	/*
	Now, stick the pfldDes and the value onto the end of what we've
	already accumulated. Use bcopy() so that we won't have problems 
	with alignment.
	*/

	bcopy((char *) &pfldDes, 
	      ((char *) &buffHeader -> currentSize) + buffHeader -> currentSize,
	      sizeof(pfldDes));
	buffHeader -> currentSize += sizeof(pfldDes);
	bcopy((char *) buffHeader -> recordAddr + pfldDes -> offset,
	      ((char *) &buffHeader -> currentSize) + buffHeader -> currentSize,
	      pfldDes -> size);
	buffHeader -> currentSize += pfldDes -> size;
	
	/*
	Set the prompt flag, if necessary, and then bump the number of fields
	in the message.
	*/

	if (pfldDes -> promptflag)
		buffHeader -> promptFlag++;
	buffHeader -> numFields++;

	return (0);
}


/****************************************************************/
/*								*/
/*		SEND A MESSAGE TO DBSERVER			*/
/*								*/
/****************************************************************/

int redQSend(char *buffer)
{
	struct redQBuffHdr	*buffHeader;

	/*
	    Return if 3 minutes have not gone by from download completion. startTime
	    is defined in dbserver.c and is set immediately after a database download.

	    GCT 2-Oct-1998
	*/
	if(firstTime1 > 0)
	{
	      firstTime1--;
	      logMsg("Waiting %d min. for first redQSend\n",sendDelay/3600,0,0,0,0,0);
	}
	if((tickGet() - startTime) < sendDelay)
	{
	      return 0;
	}
	if(firstTime2 > 0)
	{
	      firstTime2--;
	      logMsg("redQSend's allowed after %d min. delay\n",sendDelay/3600,0,0,0,0,0);
	}

	if (!buffer)
		return (-1);	/* Invalid buffer pointer */
	
	/*
	Examine the "Must do a full update the next pass" flag.
	If this flag is set, do not trash what is already in buffer,
	simply add to it making sure what you add is not already in the
	buffer.
	*/

	redundantUpdates++;
	buffHeader = (struct redQBuffHdr *) buffer;
	if (!(((RES_TABLE *)(buffHeader -> recordAddr -> radr)) -> caAnswer))
	    return(1);
	if (buffHeader -> recordAddr -> rduf)
	{
	    struct fldDes    **currField;
	    int              orig_flds = buffHeader -> numFields;
	    int              fld_cnt;
	    char             *fldptr;

	    currField = papRedFldList[buffHeader -> pointType];
	    while (*currField)
	    {
	        fldptr = buffer + sizeof (struct redQBuffHdr);
		fld_cnt = orig_flds;
	    	if ( (*currField) -> promptflag == 0) 
		  {
		  while (fld_cnt--)
		    {
		    if ((*currField) == *((struct fldDes **)fldptr))
			break;
		    fldptr += (sizeof ((struct fldDes *)fldptr) + 
			      (*((struct fldDes **)fldptr)) -> size);
		     }
		  if (fld_cnt < 0)
		      redQField(buffer, *currField);
		  }
		currField++;
	    }
	}

	/*
	Send the message. If we can't, set the "Must do a full update
	the next pass" flag. If we were able to send the message and
	that flag was set, clear it.
	*/
	if (!buffHeader -> numFields)
		return (0);	/* Nothing was posted. Don't send it */
	FASTLOCK(&dbsrvLock);
	if (rngFreeBytes(redundQueue) < buffHeader -> currentSize)
		{
		buffHeader -> recordAddr -> rduf = 1;
		redundantFails++;
		}
	else
		{
		rngBufPut(redundQueue, (char *) &(buffHeader -> currentSize),
			buffHeader -> currentSize);
		if (buffHeader -> recordAddr -> rduf)
			buffHeader -> recordAddr -> rduf = 0;
		if (rngFreeBytes(redundQueue) <= (rngNBytes(redundQueue) >> 2))
			semGive(dbServerSem);
		}
	FASTUNLOCK(&dbsrvLock);
	return (0);
}


/****************************************************************/
/*								*/
/*		SEND A MESSAGE TO DBSERVER, BUT DO NOT SEND	*/
/*		ANY PROMPTED FIELDS				*/
/*								*/
/****************************************************************/

int redQNPSend(char *buffer)
{
	struct redQBuffHdr	*buffHeader;
	struct fldDes		**currField;
	int			orig_flds;
	int			fld_cnt;
	char			*fldptr;

	if (!buffer)
		return (-1);	/* Invalid buffer pointer */

	buffHeader = (struct redQBuffHdr *) buffer;
	if (!(((RES_TABLE *)(buffHeader -> recordAddr -> radr)) -> caAnswer))
	    return(1);
	orig_flds = buffHeader -> numFields;
	currField = papRedFldList[buffHeader -> pointType];
	while (*currField)
	{
	    if ( (*currField) -> promptflag != 0)
	    {
		currField++;
		continue;
	    }
	    fldptr = buffer + sizeof (struct redQBuffHdr);
	    fld_cnt = orig_flds;
	    while (fld_cnt--)
	    {
		if ((*currField) == *((struct fldDes **)fldptr))
		    break;
		fldptr += (sizeof ((struct fldDes *)fldptr) + 
			      (*((struct fldDes **)fldptr)) -> size);
	    }
	    if (fld_cnt < 0)
		redQField(buffer, *currField);
	    currField++;
	}

	/*
	Send the message. If we can't, set the "Must do a full update
	the next pass" flag. If we were able to send the message and
	that flag was set, clear it.
	*/
	if (!buffHeader -> numFields)
		return (0);	/* Nothing was posted. Don't send it */
	FASTLOCK(&dbsrvLock);
	if (rngFreeBytes(redundQueue) < buffHeader -> currentSize)
		{
		buffHeader -> recordAddr -> rduf = 1;
		redundantFails++;
		}
	else
		{
		rngBufPut(redundQueue, (char *) &(buffHeader -> currentSize),
			buffHeader -> currentSize);
		if (buffHeader -> recordAddr -> rduf)
			buffHeader -> recordAddr -> rduf = 0;
		if (rngFreeBytes(redundQueue) <= (rngNBytes(redundQueue) >> 2))
			semGive(dbServerSem);
		}
	FASTUNLOCK(&dbsrvLock);
	return (0);
}

/****************************************************************/
/*								*/
/*	POST SEMAPHORE TO DBSERVER IF QUEUE ISN'T EMPTY		*/
/*								*/
/*	This routine is to be called at the end of a list	*/
/*	by any one of the scan tasks. If the message Queue	*/
/*	isn't empty, then post the semaphore to make dbServer	*/
/*	process the queue.					*/
/*								*/
/****************************************************************/

void redQPost(void)
{
	if (!rngIsEmpty(redundQueue))
		semGive(dbServerSem);
}

/****************************************************************/
/*								*/
/*		PRINT AND CLEAR REDUNDANT UPDATE COUNT		*/
/*								*/
/****************************************************************/

int printRedUpdateCount(void)
{
	printf("%d Redundant Updates\n", redundantUpdates);
	redundantUpdates = 0;
	return (0);
}


/****************************************************************/
/*								*/
/*		PRINT AND CLEAR REDUNDANT UPDATE FAILS		*/
/*								*/
/****************************************************************/

int printRedUpdateFails(void)
{
	printf("%d Redundant Update Failures\n", redundantFails);
	redundantFails = 0;
	return (0);
}
