/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		CHANNEL ACCESS Data Provider Application server	*/
/*								*/
/*	This task resides in the IOC. It services two kinds	*/
/*	of events. They are the arrival of a message from	*/
/*	one of our sockets and the presence of data in the	*/
/*	event queue. An I/O slave task will wake us up when	*/
/*	a message arrives by giving a semaphore. The routine	*/
/*	that writes into the event queue will also set that	*/
/*	same semaphore.						*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/*								*/
/*	The following code has been modified to remove, via	*/
/*	#ifdef-#endif,code that enables this process to run	*/
/*	under vxWorks. Additional code is for workstation	*/
/*	platforms and is contained within #ifndef-#endif.	*/
/*								*/
/*	Author: Greg Tomalesky					*/
/*								*/
/****************************************************************/

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <in.h>
#include <tcp.h>

#include "dataTypes.h"
#include "restable.h"
#include "dbDefs.h"
#include "serverControl.h"
#include "agentMsgs.h"
#include "agentPorts.h"
#include "dpa.h"
#include "msgMon.h"

#ifndef	ERROR
#define	ERROR	(-1)
#endif
#define	V_DEBUG		3
#define	V_STATUS	2
#define	V_INFO		1
#define V_NOVERBOSE	0
#define	VERBOSE(x)		if(hsrvVerbose >= x) tisSysLog
#define SESS 1
#define SOCK_RETRY_INTERVAL  50 
#define NEW_DRRDATA 1
/*
	Global Data
*/
typedef struct alq
    {
    CH_ALARM		alarmMsg;
    char		alarmData[150];
    struct alq		*nextAlq;
    } ALARM_QUE;
ALARM_QUE	*freeAlarms = 0;	/* Free chain */
ALARM_QUE	*alarmHead = 0;		/* Head of Queue */
ALARM_QUE	*alarmTail = 0;		/* Tail of Queue */
int    		maxAlarmQue = 100;	/* Max Queue depth */
int    		alarmQueSize = 0;	/* Current Queue size */
int    		alarmSerial = 0;	/* Last serial # issued */
fd_set		rsrvFds;			/* Fd mask, all sockets */


char   	*hsrvEnvString = 0;	/* could be set to DPAHOST if present */
int    	hsrvAppFd = -1;        	/* for app communications */
int    	hsrvUdpFd = -1;		/* Our UDP port */
int    	hsrvTcpFd = -1;		/* Out TCP port */
int    	hsrvMaxFd = -1;		/* Max fd used */
int    	hsrvInet;	       	/* My Inet addr */
int    	hsrvVerbose = V_INFO; 	/* verbose flag 0 = no messages */
static int	debugging = 0;
CCB    	*firstClient = 0;
char   	buff[DPA_MAX_CHARS];   	/* General Get/Put buffer */
char   	appName[100];		/* hold application(parent) name here */
int sockRetry = SOCK_RETRY_INTERVAL;
RES_TABLE	hsrvResource;  	/* Only one resource...Name of application as registered with SockMnt */

#define	RSRV_MAX_CHARS	8192 /* was 4096 */
int	rxCount;
int	rxNext;
static char	*rxBuff = 0;
/*
	External Data
*/


/*
	Function Prototypes
*/
CCB    	*findCCB();
CCB    	*rsrvFindFd();
int     rsrvReadFd();
int    	rsrvCompleteMsg();
int    	hsrvReadChars();
void	hsrvSignalHandler(int);
char	*addrToString(unsigned int addr);
void	hsrvCloseFd(int fd,fd_set *mask);
void	hsrvFdSet(int fd,fd_set *mask);

/*
	Message level strings
*/
char	*msgLevel[] =
	{
	"Silent",
	"Informational",
	"Status",
	"DEBUG!"
	};
char	*resMode[]=
	{
	"SLAVE!",
	"MASTER!"
	};

/****************************************************************/
/*								*/
/*		CHANNEL ACCESS SERVER TASK			*/
/*								*/
/****************************************************************/
int	main(int argc, char *argv[])		/* this is main for workstations */
{
  fd_set       	myFds;
  short  	msgType;
  struct sockaddr_in	portAddr;
  int	inetAddr;
  int  	fdIndex;
  CCB  	*currCCB,*ccb1,*ccb2;
  MCB  	*pMCB;
  int  	addrSize;
  int  	newFd;
  int  	optVal = 1;
  int 	selectStat;
  struct timeval	maxWait,currTime,exitTime;

#ifdef SOLARIS
  sigset(SIGINT,hsrvSignalHandler);			/* handle ^C from keyboard */
  sigset(SIGHUP,hsrvSignalHandler);			/* handle ^C from keyboard */
  sigset(SIGUSR1,hsrvSignalHandler);			/* increment verbose flag */
  sigset(SIGUSR2,hsrvSignalHandler);			/* decrement verbose flag */
#else
  sigset(SIGINT,hsrvSignalHandler);			/* handle ^C from keyboard */
  signal(SIGHUP,hsrvSignalHandler);			/* handle ^C from keyboard */
  signal(SIGUSR1,hsrvSignalHandler);			/* increment verbose flag */
  signal(SIGUSR2,hsrvSignalHandler);			/* decrement verbose flag */
#endif
  tisOpenLog("hsrvTask",6,2,NULL,NULL);
  strcpy(appName,argv[1]);
  VERBOSE(V_NOVERBOSE)(0,0,"Starting for %s on port %s",argv[1],argv[2]);
  hsrvInit(argv[2]);				/* invoke as function on host */
  if(rxBuff == 0)
    rxBuff = (char *)calloc(1,RSRV_MAX_CHARS);
  strcpy(hsrvResource.resourceName,argv[1]);			/* save resource (application) name here */
  for (;;)			/* first BIG for loop */
    {
/*
If the I/O flag is set, check the sockets.
*/
      bcopy((char *) &rsrvFds, (char *) &myFds, sizeof(myFds));
      maxWait.tv_sec = 0;				/* 1/10 second timeout */
      maxWait.tv_usec = 100000;
      while((maxWait.tv_sec >= 0) && (maxWait.tv_usec >= 0))
	{
	  gettimeofday(&currTime,0);
	  bcopy((char *) &rsrvFds, (char *) &myFds, sizeof(myFds));
	  selectStat = select(hsrvMaxFd + 1, &myFds, NULL, NULL, &maxWait);
	  if(selectStat == 0)
	    {
	      break;
	    }
	  else if(selectStat < 0)
	    {
	      VERBOSE(V_DEBUG)(0,0,"Select error:%m");
	    }
	  else
	    {
	      for (fdIndex = 0; fdIndex <= hsrvMaxFd; fdIndex++)		/* fd check loop */
		{
		  if (!FD_ISSET(fdIndex, &myFds))
		    continue;
	    /*
	    See if there is any activity on the UDP port.
	    */
		  if (fdIndex == hsrvUdpFd)
		    {
		/*
		Read just the first two bytes. We aren't interested in any more than that.
		*/
		      addrSize = sizeof(portAddr);
		      msgType = -1;
		      if (recvfrom(hsrvUdpFd, (char *) &msgType, sizeof(msgType), 0,
				   (struct sockaddr *) &portAddr, &addrSize) > 0)
			{
		    /*
		    Handle according to the message type
		    */
			  msgType = ntohs(msgType);
			  VERBOSE(V_DEBUG)(0,0,"Received message type 0x%04X from UDP socket %d",
					   msgType,hsrvUdpFd);
			  if (msgType == SOCK_CONN_REPLY)
			    {
			/*
			A client wants us to connect to him. Get his Internet address and try 
			to connect.
			*/
			      newFd = socket(PF_INET, SOCK_STREAM, 0);
			      inetAddr = ntohl(portAddr.sin_addr.s_addr);
			      portAddr.sin_family = AF_INET;
			      portAddr.sin_addr.s_addr = htonl(inetAddr);
			      portAddr.sin_port = htons(SCM_PORT);
			      if (connect(newFd, (struct sockaddr *) &portAddr, sizeof(portAddr)) ==
				  ERROR)
				{
			    /*
			    Can't connect. Wait until we are invited again.
			    */
				  hsrvCloseFd(newFd,&rsrvFds);
				  continue;
				}
			      if (!(currCCB = findCCB(inetAddr)))
				{
			    /*
			    This guy is new. Create a CCB for him and put him in the chain.
			    */
				  currCCB = (CCB *) calloc(1, sizeof(CCB));
				  currCCB -> nextClient = (CCB *) firstClient;
				  firstClient = currCCB;
				  currCCB -> clientInet.s_addr = inetAddr;
				  currCCB -> clientFd = newFd;
				}
			      else
				{
				/*
				We've seen him before. Close the old socket and replace it with
				this one. He must have crashed and then restarted.
				*/
				  if (currCCB -> clientFd > 0)
				    {
				      hsrvCloseFd(currCCB -> clientFd,&rsrvFds);
				    }
				  currCCB -> clientFd = newFd;
				}
			    /*
			    Make sure that TCP won't hang onto buffers for this new socket. We want
			    the system to look fast.
			    */
			      setsockopt(newFd, IPPROTO_TCP, TCP_NODELAY, (char *)&optVal, sizeof(optVal));
			      hsrvFdSet(newFd, &rsrvFds);
			    /*
			    Set up for flow control
			    */
			      addrSize = sizeof(currCCB -> clientTxBuff);
			      getsockopt(newFd, SOL_SOCKET, SO_SNDBUF,(char *)&(currCCB -> clientTxBuff),
					 &addrSize);
			      addrSize = sizeof(currCCB -> clientRxBuff);
			      getsockopt(newFd, SOL_SOCKET, SO_RCVBUF, (char *)&(currCCB -> clientRxBuff),
					 &addrSize);
			      currCCB -> clientRxTopped = 0;
			    }
			  else
/*
	Got a BOOMERANG from a host I 'pinged' earlier. He must be there so
	now I can connect to him. His Inet address is given in 'portAddr'
	so just get a new socket and store in CCB.
*/ 
			    if(msgType == BOOMERANG)
			      {
				inetAddr = ntohl(portAddr.sin_addr.s_addr);
				currCCB = findCCB(inetAddr);
				if(currCCB->clientFd == -1)	/* GCT 31-Oct-1977 connect only once */
				  {
				    currCCB->clientFd = socket(PF_INET, SOCK_STREAM, 0);
				    if(connect(currCCB->clientFd,(struct sockaddr *)&portAddr,addrSize) < 0)
				      {
					VERBOSE(V_DEBUG)(0,0,"SockMnt connect error: %m");
					hsrvCloseFd(currCCB->clientFd,&rsrvFds);
					currCCB->clientFd = -1;
				      }
				    else
				      {
					/*
					  Connect OK! send the registration prefix code and add socket to mask.
					*/
					int	true = 1;
					SM_DPA	rgMsg;
					VERBOSE(V_DEBUG)(0,0,"Established local connection with %s on socket %d", addrToString(inetAddr),currCCB->clientFd);
					setsockopt(currCCB->clientFd,IPPROTO_TCP,TCP_NODELAY,(char *)&true,sizeof(true));
					hsrvFdSet(currCCB->clientFd,&rsrvFds);
					msgType = htons(SOCK_DPAREG);
					rgMsg.dpaPrefixLen = strlen(appName) + 1;
					send(currCCB->clientFd,(char *)&msgType,sizeof(msgType),0);
					send(currCCB->clientFd,(char *)&rgMsg,sizeof(SM_DPA),0);
					send(currCCB->clientFd,(char *)appName,rgMsg.dpaPrefixLen,0);
				      }
				  }
			      }
			    else
			      {
			        VERBOSE(V_DEBUG)(0,0,"Got unknown msg type %d on UDP.", ntohs(msgType));
			      }
			}		/* end of recvfrom if */
		      else
			{
			  VERBOSE(V_DEBUG)(0,0,"Recv error on UDP");
			}
		    }
/*
	Code to handle incomming messages from my application
	task.
*/
		  else if(fdIndex == hsrvAppFd)
		    {
		      CH_BLDACK	aMsg;
		      DPA_DRRDATA	rMsg;
		      CH_SINGMON	smMsg;
		      CH_MONDATA	mdMsg;
		      CH_CLOSE	clMsg;
		      CH_CLOSEACK	cakMsg;
		      CH_TRDATA	trMsg;
		      CH_PARMREPLY	prMsg;
		      CH_ALARM	alMsg;
		      CH_RESMODE	resMsg;
		      static char	*buff = 0;
		      MCB		*pmcb;
		      CCB		*ccb;
		      if(buff == 0)
			buff = (char *)calloc(1,DPA_MAX_CHARS);
		      else
			memset(buff,0,DPA_MAX_CHARS);
/*
	Drain socket as much as possible. Check for complete messages and
	process them as they are encountered until no more remain.
*/
		      if(rsrvReadFd(hsrvAppFd) > 0)
			{
			  while(rsrvCompleteMsg())
			    {
			      hsrvReadChars(&msgType,2);
/*
	switch on messaqge type. These two messages have and additional
	member. This new member is a pointer to the CCB that was in effect
	when the CH_BUILD message was sent to the application. The application
	simply send this pointer back untouched so I can use it here.
*/
			      switch(msgType)
				{
				case CHAN_BLDACK:
				  hsrvReadChars(&aMsg,sizeof(CH_BLDACK));
/*
	Find the MCB for this point.
*/
				  VERBOSE(V_STATUS)(0,0, "Received BLDACK from APP...chHandle=%08X...Resource is %s, Count=%d, Type=%04X",
						    aMsg.chHandle,resMode[aMsg.currState & 1],aMsg.chElementCount,aMsg.chDataType);
				  ccb = firstClient;
				  while(ccb)
				    {
				      pmcb = ccb -> firstChan;
				      while (pmcb)
					{
					  if(pmcb->handle == aMsg.chHandle)
					    {
					      pmcb -> fldCntl = (FCB *)aMsg.chIocHandle;	/* pAddr @app */
					      pmcb -> dataType = aMsg.chDataType;
					      pmcb -> numElem = aMsg.chElementCount;
					      pmcb -> chanMask = 0;
					      hsrvResource.caAnswer = aMsg.currState;		/* current Resource state */
					      aMsg.chIocHandle = (int)pmcb;		/* restore handles for ACK message */
					      VERBOSE(V_STATUS)(0,0, "Sending BLDACK to CA...chHandle=%08X, pMCB=%08X, fldCntl=%08X",
								aMsg.chHandle,pmcb,pmcb->fldCntl);
					      sendClient(ccb,msgType,&aMsg,sizeof(CH_BLDACK),0,0,&myFds);
					      break;
					    }
					  pmcb = pmcb -> nextClient;
					}
				      ccb = ccb ->nextClient;
				    }
				  break;
				case CHAN_DRRDATA:
				  /*
				    GCT 30-Sep-1997
				    Use new DPA_DRRDATA message. This message now contains the chAppHandle
				    which is required in the CH_DRRDATA return message. It is after the embedded
				    CH_DRRDATA message structure so I did NOT need to change anything as far as 
				    CA is concerned. This chAppHandle is already available in the rsrvTask program
				    but because I had to separate the request from the response I had to provide
				    the handle in this manner. This whole excersize was to avoid using a possibly
				    zero currCCB pointer.
				  */
				  hsrvReadChars(&rMsg,sizeof(DPA_DRRDATA));
				  hsrvReadChars(buff,rMsg.drr.chDataLen);
				  VERBOSE(V_STATUS)(0,0, "Received DRRDATA from APP...chHandle=%08X, chCaHandle=%08X, Count=%d, Type=%04X",
						    rMsg.drr.chHandle,rMsg.drr.chCaHandle,rMsg.drr.chCount,rMsg.drr.chDataType);
				  ccb = firstClient;
				  while(ccb)
				    {
				      pmcb = ccb -> firstChan;
				      while(pmcb)
					{
					  if(rMsg.drr.chHandle == pmcb->handle)
					    {
					      rMsg.drr.chHandle = (int)rMsg.appHandle;
					      sendClient(ccb,msgType,&rMsg,sizeof(CH_DRRDATA),buff,rMsg.drr.chDataLen,&myFds);
					      VERBOSE(V_DEBUG)(0,0, "Sending  DRRDATA to CA...chHandle=%08X, chCaHandle=%08X, Count=%d, Type=%04X",
								rMsg.drr.chHandle,rMsg.drr.chCaHandle,rMsg.drr.chCount,rMsg.drr.chDataType);
					    }
					  pmcb = pmcb->nextClient;
					}
				      ccb = ccb->nextClient;
				    }
				  break;
				case CHAN_SINGMON:
				  hsrvReadChars(&smMsg,sizeof(CH_SINGMON));
				  hsrvReadChars(buff,smMsg.chDataLen);
				  VERBOSE(V_STATUS)(0,0, "Received SINGMON from APP...chHandle=%08X, chCaHandle=%08X, Count=%d, Type=%04X, Length=%d",
						    smMsg.chHandle,smMsg.chCaHandle,smMsg.chCount,smMsg.chDataType,smMsg.chDataLen);
				  ccb = firstClient;
				  while(ccb)
				    {
				      pmcb = ccb -> firstChan;
				      while(pmcb)
					{
					  if(smMsg.chHandle == pmcb->handle)
					    {
					      sendClient(ccb,msgType,&smMsg,sizeof(CH_SINGMON),buff,smMsg.chDataLen,&myFds);
					      VERBOSE(V_DEBUG)(0,0, "Sending  SINGMON to CA...chHandle=%08X, chCaHandle=%08X, Count=%d, Type=%04X, Length=%d",
								smMsg.chHandle,smMsg.chCaHandle,smMsg.chCount,smMsg.chDataType,smMsg.chDataLen);
					    }
					  pmcb = pmcb->nextClient;
					}
				      ccb = ccb->nextClient;
				    }
				  break;
				case CHAN_MONDATA:
				  hsrvReadChars(&mdMsg,sizeof(CH_MONDATA));
				  hsrvReadChars(buff,mdMsg.chDataLen);
				  ccb = firstClient;
				  VERBOSE(V_STATUS)(0,0, "Received MONDATA from APP...chHandle=%08X,Count=%d Type=%04X, Length=%d",
						   mdMsg.chHandle,mdMsg.chCount,mdMsg.chDataType,mdMsg.chDataLen);
				  while(ccb)
				    {
				      pmcb = ccb -> firstChan;
				      while(pmcb)
					{
					  if(mdMsg.chHandle == pmcb->handle)
					    {
					      sendClient(pmcb->client,msgType,&mdMsg,sizeof(CH_MONDATA),buff,mdMsg.chDataLen,&myFds);
					      VERBOSE(V_DEBUG)(0,0, "Sending  MONDATA to CA...chHandle=%08X,Count=%d Type=%04X, Length=%d",
								mdMsg.chHandle,mdMsg.chCount,mdMsg.chDataType,mdMsg.chDataLen);
					    }
					  pmcb = pmcb->nextClient;
					}
				      ccb = ccb->nextClient;
				    }
				  break;
				case CHAN_TRDATA:
				  hsrvReadChars(&trMsg,sizeof(CH_TRDATA));
				  hsrvReadChars(buff,trMsg.chNameLen+trMsg.chDataLen);
				  VERBOSE(V_STATUS)(0,0, "Received TRDATA from APP for...%s, Type = %04X",
						    buff,trMsg.chDataType);
				  ccb = firstClient;
				  while(ccb)
				    {
				      if(ccb -> trendHandler)
					{
					  sendClient(ccb,msgType,&trMsg,sizeof(CH_TRDATA),buff,trMsg.chNameLen+trMsg.chDataLen,&myFds);
					  VERBOSE(V_STATUS)(0,0, "Sending TRDATA to CA for...%s, Type = %04X"
							    ,buff,trMsg.chDataType);
					}
				      ccb = ccb->nextClient;
				    }
				  break;
				case CHAN_PARMREPLY:
				  hsrvReadChars(&prMsg,sizeof(CH_PARMREPLY));
				  VERBOSE(V_STATUS)(0,0,"Received Param reply");
				  msgType = htons(msgType);
				  ccb = firstClient;
				  while (ccb)
				    {
				      if (ccb -> clientFd != -1)
					{
					  VERBOSE(V_STATUS)(0,0,"Sending Parameter reply to...%08X",ccb);
					  sendClient(ccb,msgType,&prMsg,sizeof(CH_PARMREPLY),0,0,&myFds);
					}
				      ccb = ccb -> nextClient;
				    }
				  break;
				case CHAN_ALARM:
				  hsrvReadChars(&alMsg,sizeof(CH_ALARM));
				  hsrvReadChars(buff,alMsg.alarmDataLen);
				  alMsg.alarmSerial = ++alarmSerial;
				  VERBOSE(V_STATUS)(0,0,"Received Alarm from App");
				  sendNewAlarm(&alMsg, buff);
				  break;
				case CHAN_RESMODE:
/*
	Sent by App to inform us of a resource mode change.
*/
				  hsrvReadChars(&resMsg,sizeof(CH_RESMODE));
				  hsrvResource.caAnswer = resMsg.chResStatus;
				  VERBOSE(V_STATUS)(0,0, "Received resource response for %s as %s",
						    resMsg.chResName,resMode[resMsg.chResStatus & 1]);
				  msgType = htons(msgType);
				  ccb = firstClient;
				  while (ccb)
				    {
				      if (ccb -> clientFd != -1)
					{
					  VERBOSE(V_STATUS)(0,0,"Sending resource mode change to...%08X",ccb);
					  sendClient(ccb, msgType, (char *) &resMsg, sizeof(resMsg),(char *) 0, 0, &myFds);
					}
				      ccb = ccb -> nextClient;
				    }
				  break;
				case CHAN_CLOSE:
				  hsrvReadChars(&clMsg,sizeof(CH_CLOSE));
				  ccb = firstClient;
				  while (ccb)
				    {
				      pmcb = ccb -> firstChan;
				      while (pmcb)
					{
					  if(pmcb->handle == clMsg.chHandle)
					    {
					      VERBOSE(V_STATUS)(0,0, "Received CLOSE from App for chHandle=%08X,chAppHandle=%08X",
								pmcb->handle,clMsg.chAppHandle);
					      closeMCB(pmcb);
					      msgType = CHAN_CLOSEACK;
					      cakMsg.chHandle = clMsg.chHandle;
					      sendClient(ccb,msgType,&cakMsg,sizeof(CH_CLOSEACK),0,0,&myFds);
					      break;
					    }
					  pmcb = pmcb->nextClient;
					}
				      ccb = ccb -> nextClient;
				    }
				  break;
				}
			    }
			}
		    }				/* END of app receive if */
		  else
		    {
/*
	Must be a message from sockMnt. Find the CCB and process the message.
*/
		      int savefd;
		      currCCB = rsrvFindFd(fdIndex);
		      if(!currCCB)
			{
			  FD_CLR(fdIndex, &rsrvFds);
			  continue;		/* no CCB then continue */
			}
/*
	Get the message type to find out what kind of message we have. If zero length read then
	sockMnt must have gone away. Close my socket to him and set the value to -1. I will
	check up on him after my select times out.

	GCT 21-dec-1995
*/
		      savefd = currCCB->clientFd;
		      if (!clientRecv((char *) &msgType, sizeof(msgType), currCCB, &myFds))
			{
			  VERBOSE(V_STATUS)(0,0, "SockMnt dis-connect...socket %d has been closed",
					    savefd);
			  continue;
			}
		      msgType = ntohs(msgType);
		      if(msgType != 0)
			{
			  VERBOSE(V_DEBUG)(0,0,"Received message type 0x%04X on TCP socket %d",msgType,fdIndex);
			}
		      if (msgType == BOOMERANG)	/* moby-mega-maxi IF statement */
			{
/*
	A boomerang. If there's anything queued up for this client, just
	send that. If nothing, generate a boomerang.
*/
			  if (currCCB -> txBuffSize == 0)
			    {
			      currCCB -> txBuffSize = sizeof(msgType);
			      *(short *) (currCCB -> txBuff) = htons(BOOMERANG);
			    }
			  clientSend(currCCB, &myFds);
			}
		      else if (msgType == CHAN_SESS_INQ)
			{
			  CH_SESS_REPLY	chSessReply;
			    /*
			    Client sends this just after he gets started. He wants our session ID to
			    see if all the channels we had during a previous connection are still
			    valid. Just reply. If the numbers match, he'll do nothing. If they don't
			    match (i.e., he crashed), then he'll send a separate transaction to
			    set the session ID.
			    */
			  VERBOSE(V_DEBUG)(0,0,"Received Session inquery for %d...",currCCB -> clientSession);
			  chSessReply.chSession = currCCB -> clientSession;
			  msgType = htons(CHAN_SESS_REPLY);
			  sendClient(currCCB, msgType, (char *) &chSessReply, sizeof(chSessReply),(char *) 0, 0, &myFds);
			}
		      else if (msgType == CHAN_BUILD)
			{
			  CH_BUILD	chanBuild;
			  char		chanName1[PVNAME_SZ + FLDNAME_SZ + 2];
			  MCB		*pMCB;
			    /*
			    Channel Build. Get the message header and the channel name.
			    */
			  if (!clientRecv((char *) &chanBuild, sizeof(chanBuild), currCCB, &myFds))
			    continue;
			  if (!clientRecv(chanName1, ntohs(chanBuild.chNameLen), currCCB, &myFds))
			    continue;
			    /*
			    It is possible to get duplicate requests as a part of error recovery.
			    If so, just return the original handle in an acknowledgement.
			    */
			  pMCB = currCCB -> firstChan;
			  while (pMCB)
			    {
			      if(chanBuild.chHandle == pMCB -> handle)
				{
				  break;
				}
			      pMCB = pMCB -> nextClient;
			    }
			  if (pMCB)
			    continue;
			  VERBOSE(V_DEBUG)(0,0,"Received build request for %s...%08X",chanName1,chanBuild.chHandle);
/*
  Create a new MCB for this channel. The address of the MCB will become
  the handle that we send back to the client. Fill in the rest of the
  static data.
*/
			  pMCB = (MCB *)newMCB();
			  pMCB -> handle = chanBuild.chHandle;
			  pMCB -> dataType = ntohs(chanBuild.chType);
			  pMCB -> numElem = ntohs(chanBuild.chCount);
			  pMCB -> client = currCCB;
			    /*
			    Put this new MCB at the head of the chain for this client.
			    */
			  pMCB -> nextClient = currCCB -> firstChan;
			  currCCB -> firstChan = pMCB;
			  if (pMCB -> nextClient)
			    pMCB -> nextClient -> prevClient = pMCB;
/*
	Pass the CHAN_BUILD request directly to the application. The CHAN_BLDACK message
	will be returned later.
*/
			  if((strlen(chanName1) & 1) == 0)
			    chanBuild.chNameLen++;
			  VERBOSE(V_STATUS)(0,0, "Sending BUILD to APP for %s...%08X",
					    chanName1,chanBuild.chHandle);
			  send(hsrvAppFd,(char *)&msgType,sizeof(msgType),0);
			  send(hsrvAppFd,(char *)&chanBuild,sizeof(chanBuild),0);
			  send(hsrvAppFd,(char *)chanName1,chanBuild.chNameLen,0);
			}
		      else if (msgType == CHAN_CLOSE)
			{
			  CH_CLOSE		chanClose;
			    /*
			    Channel Close. Get the handle so we know what he's talking about.
			    */
			  if (clientRecv((char *) &chanClose, sizeof(chanClose), currCCB, &myFds))
			    {
					/*
					If we ever send the workstation a message with a handle that he
					doesn't know about, he'll tell us to close it. Just to be safe, make
					sure that we actually have the handle he's talking about before we
					try to delete it! Send an acknowledgement even if we didn't find
					the channel just in case the workstation is holding on to resources
					pending our close ack.
					*/
			      pMCB = currCCB -> firstChan;
			      while (pMCB)
				{
				  if (pMCB -> handle == chanClose.chHandle)
				    break;
				  pMCB = pMCB -> nextClient;
				}
			      if (pMCB)
				{
				  chanClose.chAppHandle = (int)pMCB->fldCntl;
				  VERBOSE(V_DEBUG)(0,0, "Sending CLOSE to App for chHandle=%08X,chAppHandle=%08X",
						   chanClose.chHandle,chanClose.chAppHandle);
				  send(hsrvAppFd,(char *)&msgType,sizeof(msgType),0);
				  send(hsrvAppFd,(char *)&chanClose,sizeof(CH_CLOSE),0);
				  pMCB -> fldCntl = 0; 		/* mark for delete */
				}
			    }
			}
/*
	Send the CHAN_DIRREAD mesage directly to the application. The response will come back
	later and be handled above.
*/
		      else if (msgType == CHAN_DIRREAD)
			{
			  CH_DIRREAD		dirRead;
			  MCB				*pMCB;
			    /*
			    Direct Read. Get the requested data and then package it up for shipment.
			    */
			    if (clientRecv((char *) &dirRead, sizeof(dirRead), currCCB, &myFds))
			      {
				pMCB = (MCB *)dirRead.chHandle;
				if(!validMCB(pMCB,currCCB,&myFds))
				  continue;				/* if bad then go no further! */
				dirRead.chHandle = (int)pMCB->handle;		/* pAddr @app was chCaHandle*/
				send(hsrvAppFd,(char *)&msgType,sizeof(msgType),0);
				send(hsrvAppFd,(char *)&dirRead,sizeof(dirRead),0);
				VERBOSE(V_STATUS)(0,0, "Sending DIRREAD to App...chHandle = %08X,chCaHandle = %08X, Count=%d, Type=%04X",
						  dirRead.chHandle,dirRead.chCaHandle,dirRead.chNumElem,dirRead.chDataType);
			      }
			}
		      else if (msgType == CHAN_SESS_SET)
			{
			  CH_SESS_SET		sessSet;
			    /*
			    A client has sent us a session handle. If it doesn't agree with the
			    one that we previously had for him, then he crashed and we should close
			    all of the channels that we had for him.
			    */
			  if (clientRecv((char *) &sessSet, sizeof(sessSet), currCCB, &myFds))
			    {
			      VERBOSE(V_DEBUG)(0,0,"Setting session number from %d to %d",
						currCCB->clientSession,sessSet.chSession);
			      if (currCCB -> clientSession != sessSet.chSession)
				{
				  currCCB -> clientSession = sessSet.chSession;
				  while (currCCB -> firstChan)
				    closeMCB(currCCB -> firstChan);
				}
			    }
			}
		      else if (msgType == CHAN_TYPE)
			{
			  CH_TYPE      	chType;
			    /*
			    Set the data type, count, and mask in a user's MCB. Also, update the
			    overall mask in the FCB. We don't need to lock the monitor list because
			    we aren't affecting the linkage.
			    */
			  if (clientRecv((char *) &chType, sizeof(chType), currCCB, &myFds))
			    {
			      pMCB = (MCB *) chType.chHandle;
			      if(!validMCB(pMCB,currCCB,&myFds))
				continue;				/* if bad then go no further! */
			      pMCB -> dataType = chType.chType;
			      pMCB -> numElem = chType.chCount;
			      pMCB -> chanMask = chType.chMask;
			      VERBOSE(V_DEBUG)(0,0, "Sending CH_TYPE info to App for %08X, Type=%04X, Count=%d",
					       pMCB->handle,chType.chType,chType.chCount);
			      chType.chHandle = (int)pMCB->handle; /* send clients handle */
			      send(hsrvAppFd,(char *)&msgType,sizeof(msgType),0);		/* send to APP no return */
			      send(hsrvAppFd,(char *)&chType,sizeof(CH_TYPE),0);
			    }
			}
		      else if (msgType == CHAN_MONADD)
			{
			  CH_MONADD      	monAdd;
			    /*
			    Set the data type, count, and mask in a user's MCB. Also, update the
			    overall mask in the FCB. We don't need to lock the monitor list because
			    we aren't affecting the linkage.
			    */
			  if (clientRecv((char *) &monAdd, sizeof(monAdd), currCCB, &myFds))
			    {
			      pMCB = (MCB *) monAdd.chHandle;
			      pMCB -> dataType = monAdd.chType;
			      pMCB -> numElem = monAdd.chCount;
			      pMCB -> chanMask = monAdd.mask;
			      VERBOSE(V_DEBUG)(0,0, "Sending CH_MONADD info to App for %08X, Type=%04X, Count=%d",
					       pMCB->handle,monAdd.chType,monAdd.chCount);
			      monAdd.chHandle = (int)pMCB->handle; /* send clients handle */
			      send(hsrvAppFd,(char *)&msgType,sizeof(msgType),0);		/* send to APP no return */
			      send(hsrvAppFd,(char *)&monAdd,sizeof(CH_MONADD),0);
			    }
			}
		      else if (msgType == CHAN_MONREQ)
			{
			  CH_MONREQ    	monReq;
			    /*
			    Request for an immediate monitor report for a point.
			    */
			  if (clientRecv((char *) &monReq, sizeof(monReq), currCCB, &myFds))
			    {
			      pMCB = (MCB *)monReq.chHandle;
			      monReq.chHandle = (int)pMCB->fldCntl/*handle*/;
			      /* monReq.chCaHandle = (int)pMCB->fldCntl; */			/* should be pAddr @ app */
			      VERBOSE(V_DEBUG)(0,0,"Received MONREQ from socket %d",fdIndex);
			      if(!validMCB(pMCB,currCCB,&myFds))
				continue;				/* if bad then go no further! */
			      VERBOSE(V_STATUS)(0,0, "Sending MONREQ to APP...chHandle=%08X, chCaHandle=%08X, Count=%d",
						monReq.chHandle,monReq.chCaHandle,monReq.chCount);
			      send(hsrvAppFd,(char *)&msgType,sizeof(msgType),0);
			      send(hsrvAppFd,(char *)&monReq,sizeof(monReq),0);
			    }
			}
		      else if (msgType == CHAN_DIRPUT)
			{
			  CH_DIRPUT		dirPut;
			    /*
			    Direct Put.
			    */
			  if (clientRecv((char *) &dirPut, sizeof(dirPut), currCCB, &myFds))
			    {
			      if (clientRecv(buff, ntohs(dirPut.chDataSize), currCCB, &myFds))
				{
/*
	Pass the CHAN_DIRPUT request directly to the application. There is no return message
*/
				  pMCB = (MCB *)dirPut.chHandle;
				  if(!validMCB(pMCB,currCCB,&myFds))
				    continue;				/* if bad then go no further! */
				  dirPut.chCaHandle = (int)pMCB->fldCntl;			/* should be pAddr @ app */
				  VERBOSE(V_STATUS)(0,0, "Sending DIRPUT to APP...chHandle = %08X, chCaHandle = %08X",
						    dirPut.chHandle,dirPut.chCaHandle);
				  send(hsrvAppFd,(char *)&msgType,sizeof(msgType),0);
				  send(hsrvAppFd,(char *)&dirPut,sizeof(dirPut),0);
				  if(dirPut.chDataSize & 1) dirPut.chDataSize++;	/* force even size */
				  send(hsrvAppFd,buff,dirPut.chDataSize,0);
				}
			    }
			}
		      else if (msgType == CHAN_RESMODE)
			{
			  CH_RESMODE		resModeInq;
/*
	Resource mode. Normally we send this message to a workstation, but he can also send
	it to us to inquire our current state of affairs with respect to this resource.
	Look up the name in the resource table, get the current state, and then send it back
	to the requester.
*/
			  if (clientRecv((char *) &resModeInq, sizeof(resModeInq), currCCB, &myFds))
			    {
			      VERBOSE(V_DEBUG)(0,0, "CHAN_RESMODE inquiry for %s as %s",
					       resModeInq.chResName,resMode[resModeInq.chResStatus & 1]);
			      resModeInq.chResStatus = hsrvResource.caAnswer;
			      msgType = htons(msgType);
			      sendClient(currCCB, msgType, (char *) &resModeInq, sizeof(resModeInq),(char *) 0, 0, &myFds);
			    }
			}
		      else if (msgType == CHAN_RESINQ)
			{
			  CH_RESMODE		resModeReply;
			    /*
			    Resource mode inquiry. The workstation can send this message to us to ask about
			    the state of all of our resources. He would usually do this when a connection
			    is resumed after a disruption of the communications facility.
			    */
			  VERBOSE(V_DEBUG)(0,0, "CHAN_RESINQ for resource %s",
					   hsrvResource.resourceName);
			  msgType = htons(CHAN_RESMODE);
			  strcpy(resModeReply.chResName, hsrvResource.resourceName);
			  resModeReply.chResStatus = hsrvResource.caAnswer;
			  sendClient(currCCB, msgType, (char *) &resModeReply, sizeof(resModeReply),(char *) 0, 0, &myFds);
			}
		      else if (msgType == CHAN_BLDRDMON)
			{
			  CH_BLDRDMON  	chBldRdMon;
			  char 	chanName[256]; 		/* big enough for DPA */
			    /*
			    This transaction is really three for one. We create a channel, do
			    a direct read for the control variant of the native data type with
			    the native count, and then generate a single monitor report for
			    the status variant of the native data type with the native count.
			    The purpose of this transaction is to be able to pop up graphics
			    in dm without waiting for at least two and possibly three round
			    trips between the workstation and the IOC.

			    Start processing by getting the message header and the channel name.
			    A lot of this code is very similar to the processing for the
			    CHAN_BUILD, CHAN_DIRREAD, and CHAN_MONREQ messages. However, it
			    wasn't close enough to warrant making those parts of the overall
			    code modules to be called from both here and there.
			    */
			  if (!clientRecv((char *) &chBldRdMon, sizeof(chBldRdMon), currCCB, &myFds))
			    continue;
			  memset(chanName,0,256);      	/* zero fill chan Name */
			  if (!clientRecv(chanName, ntohs(chBldRdMon.chNameLen), currCCB, &myFds))
			    continue;
			  VERBOSE(V_DEBUG)(0,0, "Received build/read/monitor request for %s...%08X",
					   chanName,chBldRdMon.chHandle);
			  pMCB = currCCB -> firstChan;
			  while (pMCB)
			    {
			      if(chBldRdMon.chHandle == pMCB -> handle)
				break;
			      pMCB = pMCB -> nextClient;
			    }
			    /*
			    Allocate a new MCB if we aren't piggybacking on an old one. Fill out
			    the information that we know so far.
			    */
			  if (!pMCB)
			    {
			      pMCB = (MCB *)newMCB();
			      pMCB -> handle = chBldRdMon.chHandle;
			      pMCB -> numElem = 1;
			      pMCB -> client = currCCB;
				/*
				This MCB is new, so put it at the head of the chain for this client.
				*/
			      pMCB -> nextClient = currCCB -> firstChan;
			      currCCB -> firstChan = pMCB;
			      if (pMCB -> nextClient)
				pMCB -> nextClient -> prevClient = pMCB;
			    }
			  VERBOSE(V_DEBUG)(0,0, "Sending BLDRDMON to APP for %s -- pMCB=%08X, chHandle=%08X",
					    chanName,pMCB,chBldRdMon.chHandle);
			  send(hsrvAppFd,(char *)&msgType,sizeof(msgType),0);
			  if((strlen(chanName) & 1) == 0)
			    chBldRdMon.chNameLen++;
			  send(hsrvAppFd,(char *)&chBldRdMon,sizeof(chBldRdMon),0);
			  send(hsrvAppFd,chanName,chBldRdMon.chNameLen,0);
			}
		      else if (msgType == CHAN_HANDLERS)
			{
			  CH_HANDLERS  	handlerMsg;
			  int  	oldTrendHandler;
			  int  	oldAlarmHandler;
			    /*
			    A workstation is sending us this message to update (or confirm)
			    his handler status for trends and alarms.
			    */
			  if (!clientRecv((char *) &handlerMsg, sizeof(handlerMsg), currCCB, &myFds))
			    continue;
			  oldTrendHandler = currCCB -> trendHandler;
			  oldAlarmHandler = currCCB -> alarmHandler;
			  currCCB -> trendHandler = handlerMsg.chTrendHandler;
			  currCCB -> alarmHandler = handlerMsg.chAlarmHandler;
			  VERBOSE(V_DEBUG)(0,0,"registering trend handler...<%d>",
						handlerMsg.chTrendHandler);
			  VERBOSE(V_DEBUG)(0,0,"registering alarm handler...<%d>",
						handlerMsg.chAlarmHandler);
			    /*
			    If adding a new alarm handler, dump the alarm Queue at him and then start
			    the database traversals.
			    */
			  if (handlerMsg.chAlarmHandler && !oldAlarmHandler)
			    {
			      ALARM_QUE	*currAlarm;
			      currAlarm = alarmHead;
			      msgType = htons(CHAN_ALARM);
			      while (currAlarm)
				{
				  sendClient(currCCB, msgType, (char *) &(currAlarm->alarmMsg),sizeof(CH_ALARM),
					     (char *)&(currAlarm->alarmData),ntohs(currAlarm->alarmMsg.alarmDataLen),
					     &myFds);
				  currAlarm = currAlarm -> nextAlq;
				}
			    }
/*
	If adding a new trend handler, tell the application to
	spoon feed him some data to get started.
*/
			  if (handlerMsg.chTrendHandler && !oldTrendHandler)
			    {
			      msgType = CHAN_HANDLERS;
			      send(hsrvAppFd,(char *)&msgType,sizeof(msgType),0);
			      send(hsrvAppFd,(char *)&handlerMsg,sizeof(CH_HANDLERS),0);
			    }
			}
		      else if (msgType == CHAN_GETPARAMS)
			{
			  CH_GETPARAMS	paramReq;
			  char		chanName[PVNAME_SZ + FLDNAME_SZ + 2];
			    /*
			    Request for parameters. Get the header and the channel name.
			    */
			  if (!clientRecv((char *) &paramReq, sizeof(paramReq), currCCB, &myFds))
			    continue;
			  if (!clientRecv(chanName, ntohs(paramReq.chNameLen), currCCB, &myFds))
			    continue;
			  VERBOSE(V_STATUS)(0,0,"Requesting parameters for %s...<%d>",
						chanName);
			  send(hsrvAppFd,(char *)&msgType,sizeof(msgType),0);
			  send(hsrvAppFd,(char *)&paramReq,sizeof(CH_GETPARAMS),0);
			  send(hsrvAppFd,chanName,paramReq.chNameLen,0);
			}
		      else if (msgType == CHAN_ALARM_RECV)
			{
			  CH_ALARM_RECV	alarmRecv;
			  ALARM_QUE    	*currAlq;
			  ALARM_QUE    	*prevAlq;
			  char 	chanName[PVNAME_SZ + FLDNAME_SZ + 2];
			    /*
			    A workstation alarm manager is telling us that he has received an
			    alarm. Delete it from our queue.
			    */
			  if (!clientRecv((char *) &alarmRecv, sizeof(alarmRecv), currCCB, &myFds))
			    continue;
			  if (!clientRecv(chanName, ntohs(alarmRecv.alarmNameLen), currCCB, &myFds))
			    continue;
			  currAlq = alarmHead;
			  prevAlq = 0;
			  while (currAlq)
			    {
			      if (currAlq -> alarmMsg.alarmSerial == alarmRecv.alarmSerial)
				break;
			      prevAlq = currAlq;
			      currAlq = currAlq -> nextAlq;
			    }
			    /*
			    It may be possible that some other workstation already acknowledged this
			    alarm. If we can't find the serial number in the queue, then silently
			    ignore it.
			    */
			  if (!currAlq)
			    {
			      continue;
			    }
			    /*
			    Take this one out of the queue, updating our head and tail pointers. Put
			    this one on the free list.
			    */
			  if (prevAlq)
			    prevAlq -> nextAlq = currAlq -> nextAlq;
			  else
			    alarmHead = currAlq -> nextAlq;
			  if (alarmTail == currAlq)
			    {
			      if (prevAlq)
				alarmTail = prevAlq;
			      else
				alarmTail = alarmHead;
			    }
			  currAlq -> nextAlq = freeAlarms;
			  freeAlarms = currAlq;
			}
		      else if (msgType == CHAN_ALARM_ACK)
			{
			  CH_ALARM_ACK	alarmAck;
			  char 	pv[PVNAME_SZ + FLDNAME_SZ + 2];
			    /*
			    An operator has acknowledged an alarm. Turn the ACK field from 1 to zero.
			    */
			  if (!clientRecv((char *) &alarmAck, sizeof(alarmAck), currCCB, &myFds))
			    continue;
			  if (!clientRecv(pv, ntohs(alarmAck.alarmNameLen), currCCB, &myFds))
			    continue;
			  VERBOSE(V_STATUS)(0,0, "Received alarm ACK for %s...<%d>", pv);
			  if((strlen(pv) & 1) == 0)
			    alarmAck.alarmNameLen++;
			  send(hsrvAppFd,(char *)&msgType,sizeof(msgType),0);
			  send(hsrvAppFd,(char *)&alarmAck,sizeof(CH_ALARM_ACK),0);
			  send(hsrvAppFd,pv,alarmAck.alarmNameLen,0);
			    }
		      else if (msgType == CHAN_ALARM_CTL)
			{
			  CH_ALARM_CTL	alarmCtl;
			  char 	pv[PVNAME_SZ + FLDNAME_SZ + 2];
			    /*
			    Turn alarming on or off for a single point.
			    */
			  if (!clientRecv((char *) &alarmCtl, sizeof(alarmCtl), currCCB, &myFds))
			    continue;
			  if (!clientRecv(pv, ntohs(alarmCtl.alarmNameLen), currCCB, &myFds))
			    continue;
			  if((strlen(pv) & 1) == 0)
			    alarmCtl.alarmNameLen++;
			  send(hsrvAppFd,(char *)&msgType,sizeof(msgType),0);
			  send(hsrvAppFd,(char *)&alarmCtl,sizeof(CH_ALARM_CTL),0);
			  send(hsrvAppFd,pv,alarmCtl.alarmNameLen,0);
			}
		      else if(msgType == SOCK_DPACONN)
			{
			  SM_DPACONN	dpaConn;
			  CCB		*ccb;
			  short	paddbytes;
			  if (!clientRecv((char *) &paddbytes, sizeof(short), currCCB, &myFds))	/* strip pad bytes first */
			    continue;
			  if (!clientRecv((char *) &dpaConn, sizeof(SM_DPACONN), currCCB, &myFds))	/* then read message */
			    continue;
			  portAddr.sin_family = AF_INET;
			  portAddr.sin_addr.s_addr = dpaConn.dpaInet.s_addr;
			  portAddr.sin_port = htons(SCM_PORT);
			  VERBOSE(V_DEBUG)(0,0,"Recieved SOCK_DPACONN message from %s",
						addrToString(dpaConn.dpaInet.s_addr));
/*
	Add to CCB list anyway so we can check later.
*/
			  if((ccb = findCCB(dpaConn.dpaInet)) == 0)
			    {
			      ccb = (CCB *) calloc(1, sizeof(CCB));
			      ccb -> nextClient = (struct cCl *) firstClient;
			      firstClient = ccb;
			      ccb -> clientInet = dpaConn.dpaInet;
			      ccb -> clientSession = 0;
			    }
			  else
			    {
			      if(ccb->clientFd != -1)
				{
				  hsrvCloseFd(ccb->clientFd,&rsrvFds);
				  ccb->clientFd = -1;
				}
			    }
			  ccb->clientFd = socket(PF_INET, SOCK_STREAM, 0);
			  if(connect(ccb->clientFd,(struct sockaddr *)&portAddr,addrSize) < 0)
			    {
			      VERBOSE(V_DEBUG)(0,0,"SockMnt connect error: %m");
			      hsrvCloseFd(ccb->clientFd,&rsrvFds);
			      ccb->clientFd = -1;
			    }
			  else
			    {
/*
Connect OK! send the registration prefix code and add socket to mask.
*/
			      int	true = 1;
			      SM_DPA	rgMsg;
			      VERBOSE(V_STATUS)(0,0, "Established remote connection with %s on socket %d",
						addrToString(dpaConn.dpaInet.s_addr),ccb->clientFd);
			      setsockopt(ccb->clientFd,IPPROTO_TCP,TCP_NODELAY,(char *)&true,sizeof(true));
			      hsrvFdSet(ccb->clientFd,&rsrvFds);
			      msgType = htons(SOCK_DPAREG);
			      rgMsg.dpaPrefixLen = strlen(appName) + 1;
			      send(ccb->clientFd,(char *)&msgType,sizeof(msgType),0);
			      send(ccb->clientFd,(char *)&rgMsg,sizeof(SM_DPA),0);
			      send(ccb->clientFd,appName,rgMsg.dpaPrefixLen,0);
			    }
			}
		      else
			{
			  VERBOSE(V_NOVERBOSE)(0,0, "Illegal message type %04X from %s",
					       (unsigned)msgType,addrToString(currCCB -> clientInet.s_addr));
			  continue;
			}
		    }
		}			/* fd check for loop */
	    }
/*
	adjust timer
*/
	  gettimeofday(&exitTime,0);
	  exitTime.tv_sec -= currTime.tv_sec;
	  exitTime.tv_usec -= currTime.tv_usec;
	  if(exitTime.tv_usec < 0)
	    {
	      exitTime.tv_usec += 1000000;
	      exitTime.tv_sec--;
	    }
	  maxWait.tv_sec -= exitTime.tv_sec;
	  maxWait.tv_usec -= exitTime.tv_usec;
	  if(maxWait.tv_usec < 0)
	    {
	      maxWait.tv_usec += 1000000;
	      maxWait.tv_sec--;
	    }
	}	       				/* end of while((selectStat = select... */
/*
	Timeout. Lets check for any host that we may have missed the
	last time around.
*/
      if(--sockRetry <= 0)
	{
	  sockRetry = SOCK_RETRY_INTERVAL; /* 50 = 5 seconds */
	  ccb2 = firstClient;
	  while (ccb2)
	    {
	      if (ccb2 -> clientFd == -1)
		{
		  portAddr.sin_family = AF_INET;
		  portAddr.sin_addr.s_addr = htonl(ccb2->clientInet.s_addr);
		  portAddr.sin_port = htons(SCM_PORT);
	      /*
		send BOOMERANG to host. If I get a response then connect!
		*/
		  msgType = htons(BOOMERANG);
		  VERBOSE(V_DEBUG)(0,0,"Looking for SOCKMNT at %s",
					addrToString(ccb2->clientInet.s_addr));
		  if(sendto(hsrvUdpFd, (char *) &msgType, sizeof(msgType), 0, 
			    (struct sockaddr *) &portAddr, sizeof(portAddr)) < 0)
		    VERBOSE(V_DEBUG)(0,0,"Send BOOMERANG error:%m");
		}
	      ccb2 = ccb2 -> nextClient;
	    }
	}
	/*
	Now push out anything that is left in the buffers. We don't want to wait to keep the system
	responsive.
	*/
      ccb1 = firstClient;
      while (ccb1)
	{
	  if (ccb1 -> txBuffSize)
	    {
	      int size = ccb1 -> txBuffSize;
	      if(clientSend(ccb1, (fd_set *) 0))
		VERBOSE(V_DEBUG)(0,0,"Flushing %d bytes to...%d @%s",
				size,ccb1->clientFd,addrToString(ccb1->clientInet.s_addr));
	    }
	  ccb1 = ccb1-> nextClient;
	}
    }				/* first BIG for loop */
}
/****************************************************************/
/*								*/
/*	FIND A CCB GIVEN THE CLIENT'S INTERNET ADDRESS		*/
/*								*/
/****************************************************************/

CCB *findCCB(inetAddr)
int	inetAddr;
{
  CCB		*currCCB;
  currCCB = firstClient;
  while (currCCB)
    {
      if (currCCB -> clientInet.s_addr == inetAddr)
	break;
      currCCB = currCCB -> nextClient;
    }
  return (currCCB);
}


/****************************************************************/
/*								*/
/*		FIND A CCB GIVEN THE SOCKET FD			*/
/*								*/
/****************************************************************/

CCB *rsrvFindFd(fd)
int	fd;
{
  CCB		*currCCB;
  currCCB = firstClient;
  while (currCCB)
    {
      if (currCCB -> clientFd == fd)
	break;
      currCCB = currCCB -> nextClient;
    }
  return (currCCB);
}


/****************************************************************/
/*								*/
/*		SEND BYTE STREAM TO THE CLIENT			*/
/*								*/
/*	Return a 1 if successful, zero on failure.		*/
/*								*/
/****************************************************************/

clientSend(currCCB, otherFds)
CCB	*currCCB;
fd_set	*otherFds;
{
  int		bytesSent;

  if (currCCB -> clientFd == -1)
    {
      currCCB -> txBuffSize = 0;
      return (0);
    }
  if ((bytesSent = send(currCCB -> clientFd, currCCB -> txBuff, currCCB -> txBuffSize, 0)) == ERROR)
    {
      if (errno == EWOULDBLOCK)
	{
	  currCCB -> clientTxBuff += currCCB -> txBuffSize;
	  if (setsockopt(currCCB -> clientFd, SOL_SOCKET, SO_SNDBUF, (char *)&(currCCB -> clientTxBuff),
			 sizeof(currCCB -> clientTxBuff)) == ERROR)
	    {
	      hsrvCloseFd(currCCB -> clientFd,&rsrvFds);
	      if (otherFds)
		FD_CLR(currCCB -> clientFd, otherFds);
	      currCCB -> clientFd = -1;
	      currCCB -> txBuffSize = 0;
	      return (0);
	    }
	}
      else
	{
	  hsrvCloseFd(currCCB -> clientFd,&rsrvFds);
	  if (otherFds)
	    FD_CLR(currCCB -> clientFd, otherFds);
	  currCCB -> clientFd = -1;
	  currCCB -> txBuffSize = 0;
	  return (0);
	}
      return (clientSend(currCCB, otherFds));
    }
  else if (bytesSent != currCCB -> txBuffSize)
    {
      currCCB -> txBuffSize -= bytesSent;
      bcopy(currCCB -> txBuff + bytesSent, currCCB -> txBuff, currCCB -> txBuffSize);
      return (clientSend(currCCB, otherFds));
    }
  currCCB -> txBuffSize = 0;
  return (1);
}


/****************************************************************/
/*								*/
/*		GET BYTES FROM A STREAM SOCKET			*/
/*								*/
/*	Return a 1 if successful, zero on failure.		*/
/*								*/
/****************************************************************/

clientRecv(buff, nbytes, currCCB, localFds)
char	*buff;
int	nbytes;
CCB	*currCCB;
fd_set	*localFds;
{
  int			bytesGot;
  fd_set		oneFd;
  struct timeval	timeOut;
  
  if (currCCB -> clientFd == -1)
    return (0);
  timeOut.tv_sec = 5;
  timeOut.tv_usec = 0;
  FD_ZERO(&oneFd);
  FD_SET(currCCB -> clientFd, &oneFd);
  while (nbytes)
    {
      if (select(currCCB -> clientFd + 1, &oneFd, NULL, NULL, &timeOut) <= 0)
	{
	  if (debugging)
	    printf("Timeout!\n");
	  hsrvCloseFd(currCCB -> clientFd,&rsrvFds);
	  FD_CLR(currCCB -> clientFd, localFds);
	  currCCB -> clientFd = -1;
	  currCCB -> txBuffSize = 0;
	  return (0);
	}
      if ((bytesGot = recv(currCCB -> clientFd, buff, nbytes, 0)) <= 0)
	{
	  if (debugging)
	    {
	      if (bytesGot == 0)
		printf("Zero length receive\n");
	      else
		perror("Read error");
	    }
	  hsrvCloseFd(currCCB -> clientFd,&rsrvFds);
	  FD_CLR(currCCB -> clientFd, localFds);
	  currCCB -> clientFd = -1;
	  currCCB -> txBuffSize = 0;
	  return (0);
	}
      buff += bytesGot;
      nbytes -= bytesGot;
    }
  return (1);
}


/****************************************************************/
/*								*/
/*		ENQUEUE MESSAGE TO A CLIENT			*/
/*								*/
/*	This routine builds an output message to a client.	*/
/*	A pointer to a CCB tells us which client and also	*/
/*	locates his transmit buffer. The message is made up	*/
/*	from a message type and up to two additional buffer	*/
/*	pointers. If there isn't room in the transmit buffer	*/
/*	for the complete message, then we empty the existing	*/
/*	buffer first. The entire message goes into the buffer	*/
/*	as one chunk to minimize TCP's message fragmenting.	*/
/*	This routine supplies the SOCK_READ wrapper for the	*/
/*	message.						*/
/*								*/
/*	Return a 1 if successful, a 0 on failure.		*/
/*								*/
/****************************************************************/

sendClient(pCCB, msgType, buff1, buff1Len, buff2, buff2Len, fdMask)
CCB	*pCCB;
short	msgType;
char	*buff1;
int	buff1Len;
char	*buff2;
int	buff2Len;
fd_set	*fdMask;
{
  SM_READ	readHdr;
  short	sockReadType = htons(SOCK_READ);
  int		bytesRqd;
    /*
    Calculate the total size that we need.
    */
  bytesRqd= sizeof(sockReadType) + sizeof(readHdr) + sizeof(msgType) + buff1Len + buff2Len;
  if (bytesRqd > sizeof(pCCB -> txBuff))
    return (0);				/* Won't ever fit! */
  if ((pCCB -> txBuffSize + bytesRqd) > sizeof(pCCB -> txBuff))
    if (!clientSend(pCCB, fdMask))
      return (0);
    /*
    The message we want to send consists of a SOCK_READ message type, a struct SM_READ, and
    then the user's message type followed by his buffers.
    */

  bcopy((char *) &sockReadType, pCCB -> txBuff + pCCB -> txBuffSize, sizeof(sockReadType));
  pCCB -> txBuffSize += sizeof(sockReadType);
  readHdr.smLength = sizeof(msgType) + buff1Len + buff2Len;
  bcopy((char *) &readHdr, pCCB -> txBuff + pCCB -> txBuffSize, sizeof(readHdr));
  pCCB -> txBuffSize += sizeof(readHdr);
  bcopy((char *) &msgType, pCCB -> txBuff + pCCB -> txBuffSize, sizeof(msgType));
  pCCB -> txBuffSize += sizeof(msgType);
  if (buff1 && buff1Len)
    {
      bcopy(buff1, pCCB -> txBuff + pCCB -> txBuffSize, buff1Len);
      pCCB -> txBuffSize += buff1Len;
    }
  if (buff2 && buff2Len)
    {
      bcopy(buff2, pCCB -> txBuff + pCCB -> txBuffSize, buff2Len);
      pCCB -> txBuffSize += buff2Len;
    }
  return (1);
}

/****************************************************************/
/*								*/
/*		CLOSE A CHANNEL WITH GIVEN MCB			*/
/*								*/
/****************************************************************/

closeMCB(pMCB)
MCB	*pMCB;
{
  /*
    First disconnect this MCB from the CCB. That won't affect anyone else.
    */
  if (pMCB -> nextClient)
    pMCB -> nextClient -> prevClient = pMCB -> prevClient;
  if (pMCB -> prevClient)
    pMCB -> prevClient -> nextClient = pMCB -> nextClient;
  else
    pMCB -> client -> firstChan = pMCB -> nextClient;
  deleteMCB(pMCB);
}
/****************************************************************/
/*								*/
/*	PUT AN ALARM MESSAGE INTO THE QUEUE AND SEND IT TO	*/
/*	ALL THE ALARM HANDLERS THAT WE KNOW ABOUT 		*/
/*								*/
/****************************************************************/

sendNewAlarm(alarmMsg, alarmData)
CH_ALARM	*alarmMsg;
char		*alarmData;
{
  short		msgType;
  CCB	       	*currCCB;
  ALARM_QUE    	*newAlarm;
/*
  struct dbCommon	*recHdr;
*/
/*
First thing to do is to send this message to all the registered
alarm handlers.
*/
  VERBOSE(V_DEBUG)(0,0,"New alarm for %s", alarmData);
  msgType = htons(CHAN_ALARM);
  currCCB = firstClient;
  while (currCCB)
    {
      if ((currCCB -> clientFd >= 0) && currCCB -> alarmHandler)
	sendClient(currCCB, msgType, (char *) alarmMsg, sizeof(CH_ALARM),
		   alarmData, ntohs(alarmMsg -> alarmDataLen), (fd_set *) 0);
      currCCB = currCCB -> nextClient;
    }
/*
Now insert this message into the queue.
*/
  if (freeAlarms)
    {
      newAlarm = freeAlarms;
      freeAlarms = freeAlarms -> nextAlq;
    }
  else if (alarmQueSize < maxAlarmQue)
    {
      newAlarm = (ALARM_QUE *) malloc(sizeof(ALARM_QUE));
      alarmQueSize++;
    }
  else
    {
      ALARM_QUE	*prevAlarm;
    /*
    The alarm queue has grown to its maximum size. The queue is time
    ordered oldest to newest, so the first one we find with a matching
    PV name can be deleted.
    */
      prevAlarm = 0;
      newAlarm = alarmHead;
      while (newAlarm)
	{
	  if (!strcmp(newAlarm -> alarmData, alarmData))
	    break;
	  prevAlarm = newAlarm;
	  newAlarm = newAlarm -> nextAlq;
	}
      if (newAlarm)
	{
/*
		if (rsrvPrintingWarnings)
		    printf("Oldest duplicate alarm for %s discarded\n", newAlarm -> alarmData);
*/
	  if (prevAlarm)
	    {
	      prevAlarm -> nextAlq = newAlarm -> nextAlq;
	      if (alarmTail == newAlarm)
		alarmTail = prevAlarm;
	    }
	  else
	    {
	      alarmHead = newAlarm -> nextAlq;
	      if (alarmTail == newAlarm)
		alarmTail = alarmHead;
	    }
	}
      else
	{
		/*
		No duplicates. Drop the oldest alarm in our queue.
		*/
		newAlarm = alarmHead;
		alarmHead = alarmHead -> nextAlq;
		if (alarmTail == newAlarm)
		  alarmTail = alarmHead;
	}
    }

/*
Fill out the data and put the message onto the tail of the queue.
*/

  newAlarm -> alarmMsg = *alarmMsg;
  bcopy(alarmData, newAlarm -> alarmData, ntohs(alarmMsg -> alarmDataLen));
  newAlarm -> nextAlq = 0;
  if (alarmTail)
    {
      alarmTail -> nextAlq = newAlarm;
      alarmTail = newAlarm;
    }
  else
    {
      alarmHead = newAlarm;
      alarmTail = newAlarm;
    }
/*
Also send this alarm to our redundant partner in case of our untimely death.


sendRedAlarms(((RES_TABLE *)(recHdr -> radr)) -> redAlarmFd, msgType, (char *) newAlarm,
	sizeof(CH_ALARM) + ntohs(alarmMsg -> alarmDataLen));
*/
}

/*
	The following code is used to read the application socket. The scheme is the
	same as in dbClient. The maximum number of characters (4096) are read into the
	buffer. The buffer is then examined for complete messages based on message
	type and size. If a full message is detected then it is processed. If the bytes
	read do not make up a complete message, more reads are attempted until it does.

	Read socket function. drain the fd!
*/
int	rsrvReadFd(int fd)
{
  int		stat,readSize;
  if(rxCount && rxNext)		/* slide existing bytes down */
    {
      bcopy(&rxBuff[rxNext],&rxBuff[0],rxCount);
      rxNext = 0;
    }
  readSize = RSRV_MAX_CHARS - rxCount;
  if((stat = recv(fd,&rxBuff[rxCount],readSize,0)) < 0)
    {
      VERBOSE(V_NOVERBOSE)(0,0,"Exiting due to error reading %s socket:%m",appName);
      exit(3);
    }
  if(stat == 0)
    {
      VERBOSE(V_NOVERBOSE)(0,0,"Exiting due to zero length read from %s",appName);
      exit(2);
    }
  rxCount += stat;
  rxNext = 0;
  return(stat);
}
int	rsrvCompleteMsg()
{
  short	mType,mSize;
  DPA_DRRDATA		*rMsg;
  CH_SINGMON		*smMsg;
  CH_MONDATA		*mdMsg;
  CH_TRDATA		*trMsg;
  CH_ALARM		*alMsg;
  
  mSize = RSRV_MAX_CHARS;
  if(rxCount < sizeof(mType))
    return(0);
  bcopy(&rxBuff[rxNext],&mType,sizeof(mType));
  switch(mType)
    {
    case CHAN_DRRDATA:
      if(rxCount > sizeof(DPA_DRRDATA) + sizeof(mType))
	{
	  rMsg = (DPA_DRRDATA *)&rxBuff[rxNext + sizeof(mType)];
	  mSize = sizeof(DPA_DRRDATA) + rMsg->drr.chDataLen;
	}
      break;
    case CHAN_SINGMON:
      if(rxCount > sizeof(CH_SINGMON) + sizeof(mType))
	{
	  smMsg = (CH_SINGMON *)&rxBuff[rxNext + sizeof(mType)];
	  mSize = sizeof(CH_SINGMON) + smMsg->chDataLen;
	}
      break;
    case CHAN_MONDATA:
      if(rxCount > sizeof(CH_MONDATA) + sizeof(mType))
	{
	  mdMsg = (CH_MONDATA *)&rxBuff[rxNext + sizeof(mType)];
	  mSize = sizeof(CH_MONDATA) + mdMsg->chDataLen;
	}
      break;
    case CHAN_TRDATA:
      if(rxCount > sizeof(CH_TRDATA) + sizeof(mType))
	{
	  trMsg = (CH_TRDATA *)&rxBuff[rxNext + sizeof(mType)];
	  mSize = sizeof(CH_TRDATA) + trMsg->chNameLen + trMsg->chDataLen;
	}
      break;
    case CHAN_ALARM:
      if(rxCount > sizeof(CH_ALARM) + sizeof(mType))
	{
	  alMsg = (CH_ALARM *)&rxBuff[rxNext + sizeof(mType)];
	  mSize = sizeof(CH_ALARM) + alMsg->alarmDataLen;
	}
      break;
    case CHAN_RESMODE:
      mSize = sizeof(CH_RESMODE);
      break;
    case CHAN_BLDACK:
      mSize = sizeof(CH_BLDACK);
      break;
    case CHAN_CLOSE:
      mSize = sizeof(CH_CLOSE);
      break;
    default:
      VERBOSE(V_NOVERBOSE)(0,0,"Unknown message type from %s...%04X",
				appName,(unsigned)mType);
      rxNext = 0;
      rxCount = 0;
      return(0);
    }
  return(mSize > rxCount ? 0 : 1);
}

int	hsrvReadChars(char *userBuff,int userSize)
{
  bcopy(&rxBuff[rxNext],userBuff,userSize);
  rxNext += userSize;
  rxCount -= userSize;
  return(userSize);
}
/****************************************************************/
/*								*/
/*			INITIALIZE				*/
/*								*/
/****************************************************************/
hsrvInit(char *portString)
{
  struct sockaddr_in	portAddr;
  int  	optVal = 1;
  short	msgType,portNumber;
  struct	hostent	*ent;
  unsigned int	addr;
  int	true = 1;
  char	*hostName;
  CCB		*ccb;
    /*
    Create a UDP port, bind it to our well know address, and also broadcast out a message that we are 
    alive. Do this one time and one time only. This is an optimization so that all the clients that 
    are anxiously waiting for us to arrive won't wait until their next scheduled time to invite us
    to connect to them.
    */
  hsrvUdpFd = socket(PF_INET, SOCK_DGRAM, 0);
  setsockopt(hsrvUdpFd, SOL_SOCKET, SO_BROADCAST, (caddr_t) &optVal, sizeof(optVal));
    /*
    Set up the FD mask so we can use select to serve multiple sockets from a single thread.
    */
  FD_ZERO(&rsrvFds);
  hsrvFdSet(hsrvUdpFd, &rsrvFds);
    hsrvEnvString = getenv("DPAHOST");
  if(!hsrvEnvString)
    hsrvEnvString = "localhost";
  hostName = strtok(hsrvEnvString,",");
  while(hostName)
    {
      if((ent = gethostbyname(hostName)) != 0)
	{
	  addr = *(unsigned int *)*ent->h_addr_list;
	  portAddr.sin_family = AF_INET;
	  portAddr.sin_addr.s_addr = htonl(addr);
	  portAddr.sin_port = htons(SCM_PORT);
/*
	send BOOMERANG to host. If I get a response then connect!
*/
	  msgType = htons(BOOMERANG);
	  VERBOSE(V_STATUS)(0,0,"Attempting to connect to SockMnt on host %s at address %s",hostName,addrToString(addr));
	  if(sendto(hsrvUdpFd, (char *) &msgType, sizeof(msgType), 0, (struct sockaddr *) &portAddr, sizeof(portAddr)) < 0)
	    VERBOSE(V_DEBUG)(0,0,"Send BOOMERANG error:%m");
/*
	Add to CCB list anyway so we can check later.
*/
	  ccb = (CCB *) calloc(1, sizeof(CCB));
	  ccb -> nextClient = (struct cCl *) firstClient;
	  firstClient = ccb;
	  ccb -> clientInet.s_addr = addr;
	  ccb -> clientFd = -1;
	  ccb -> clientSession = 0;
	}
      hostName = strtok(NULL,",");
    }

/*
	Attempt to connect to parent task. If this fails then exit.
*/

  if((ent = gethostbyname("localhost")) != 0)
    {
      VERBOSE(V_STATUS)(0,0,"Attempting to connect to parent task at address %s",
				addrToString(addr));
      addr = *(long *)*ent->h_addr_list;
      hsrvAppFd = socket(AF_INET,SOCK_STREAM,0);
      portAddr.sin_family = AF_INET;
      portAddr.sin_addr.s_addr = htonl(addr);
      portNumber = (short)atoi(portString);
      portAddr.sin_port = htons(portNumber);
      hsrvInet = portAddr.sin_addr.s_addr;
      if(connect(hsrvAppFd,(struct sockaddr *)&portAddr,sizeof(portAddr)) < 0)
	{
	  VERBOSE(V_DEBUG)(0,0,"Application connect error: %m");
	  exit(2);
	}
      else
	{
	  hsrvFdSet(hsrvAppFd,&rsrvFds);
	  setsockopt(hsrvAppFd,IPPROTO_TCP,TCP_NODELAY,(char *)&true,sizeof(true));
	}
    }
}
void	hsrvSignalHandler(int whichone)
{
  time_t	end_time;
  end_time = time(0);
  switch(whichone)
    {
    case SIGINT:
      VERBOSE(V_NOVERBOSE)(0,0,"Exiting due to ^C or parent exit: %s",ctime(&end_time));
      close(hsrvAppFd);
      close(hsrvUdpFd);
      exit(1);
    case SIGUSR1:
      if(++hsrvVerbose > V_DEBUG) hsrvVerbose = V_DEBUG;
      VERBOSE(V_NOVERBOSE)(0,0,"Message level now at %s",msgLevel[hsrvVerbose]);
      break;
    case SIGUSR2:
      if(--hsrvVerbose < V_NOVERBOSE) hsrvVerbose = V_NOVERBOSE;
      VERBOSE(V_NOVERBOSE)(0,0,"Message level now at %s\n",msgLevel[hsrvVerbose]);
    case SIGHUP:
      break;
    }
}
char	*addrToString(unsigned int addr)
{
static char *inetString = 0;
if(inetString == 0)
  inetString = (char *)calloc(1,32);
else
  memset(inetString,0,32);
sprintf(inetString,"%d.%d.%d.%d\0",
	((addr >> 24) & 0xff),			/* hi-order nibble first 31 - 24 */
	((addr >> 16) & 0xff),			/* 23 - 16 next */
	((addr >>  8) & 0xff),			/* 15 - 8 next */
	(addr & 0xff));	       			/*  7 - 0 at last */
return(inetString);
}
void hsrvCloseFd(int fd,fd_set *mask)
{
  char buf[80];
  fd_set oldMask;

  if((fd >=0 ) && (mask != 0))
    {
      close(fd);
      bcopy(mask,&oldMask,sizeof(fd_set));
      FD_CLR(fd,mask);
      sprintf(buf,"FD=%d maxFD=%d",fd,hsrvMaxFd);
      VERBOSE(V_STATUS)(0,0,"Socket closed %s",buf);
      sprintf(buf,"%08X %08X %08X %08X",oldMask.fds_bits[0],oldMask.fds_bits[1],oldMask.fds_bits[2],oldMask.fds_bits[3]);
      VERBOSE(V_STATUS)(0,0,"Old fds_bits %s",buf);
      sprintf(buf,"%08X %08X %08X %08X",mask->fds_bits[0],mask->fds_bits[1],mask->fds_bits[2],mask->fds_bits[3]);
      VERBOSE(V_STATUS)(0,0,"After clear  %s",buf);
    }
}
void hsrvFdSet(int fd,fd_set *mask)
{
  char buf[80];
  fd_set oldMask;

  if((fd >=0 ) && (mask != 0))
    {
      bcopy(mask,&oldMask,sizeof(fd_set));
      FD_SET(fd,mask);
      if(fd > hsrvMaxFd) hsrvMaxFd = fd;
      sprintf(buf,"FD=%d maxFD=%d",fd,hsrvMaxFd);
      VERBOSE(V_STATUS)(0,0,"Socket opened %s",buf);
      sprintf(buf,"%08X %08X %08X %08X",oldMask.fds_bits[0],oldMask.fds_bits[1],oldMask.fds_bits[2],oldMask.fds_bits[3]);
      VERBOSE(V_STATUS)(0,0,"Old fds_bits %s",buf);
      sprintf(buf,"%08X %08X %08X %08X",mask->fds_bits[0],mask->fds_bits[1],mask->fds_bits[2],mask->fds_bits[3]);
      VERBOSE(V_STATUS)(0,0,"After set  %s",buf);
    }
}
/****************************************************************/
/*								*/
/*		VALIDATE A POINTER TO AN MCB			*/
/*								*/
/*	If the pointer isn't valid, then we should close	*/
/*	the socket and then put an impossible value in the	*/
/*	session handle. That will force chanAcq at the other	*/
/*	end of this conversation to rebuild all channels	*/
/*	with us from scratch. Print a message when this		*/
/*	happens because this is a fairly severe problem.	*/
/*								*/
/****************************************************************/

validMCB(MCB *pMCB, CCB *pCCB, fd_set *pFdset)
{
  MCB		*currMCB;

  if (pMCB)
    {
      currMCB = pCCB -> firstChan;
      while (currMCB)
	{
	  if (currMCB == pMCB)
	    return (1);
	  currMCB = currMCB -> nextClient;
	}
    }
  /*
    If we get here, then the pointer is invalid. Take corrective
    action and then notify the world that this has happened.
  */
  if (pCCB -> clientFd)
    {
      close(pCCB -> clientFd);
      FD_CLR(pCCB -> clientFd, &rsrvFds);
      if (pFdset)
	FD_CLR(pCCB -> clientFd, pFdset);
      pCCB -> clientFd = 0;
      pCCB -> handlerMsgRcvd = 0;
    }
  pCCB -> clientSession = 0;
  VERBOSE(V_INFO)(0,0,"Invalid handle from %s",
			addrToString(pCCB -> clientInet.s_addr));
  return (0);
}
