/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		REWRITTEN db_post_events()			*/
/*								*/
/*	This routine replaces db_post_events() which was	*/
/*	originally encoded in dbEvent.c. This routine is for	*/
/*	the new channel access.					*/
/*								*/
/*	An important note is that a single call to this routine	*/
/*	could result in up to three different message types	*/
/*	leaving rsrvTask. A point could have a monitor on the	*/
/*	VAL field, cross MDEL and also change its alarm state.	*/
/*	For any one call to this routine, we only have to put	*/
/*	one message into the ring buffer. rsrvTask will take	*/
/*	care of the multiplexing.				*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <vxWorks.h>
#include <inetLib.h>
#include <types.h>

#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#include "serverControl.h"
#include "dbAccess.h"
#include "dbRecDes.h"
#include "fast_lock.h"
#include "dataTypes.h"
#include "rbMgt.h"
#include "rsrv.h"

int db_post_events(precord, pfield, mask)
struct dbCommon	*precord;
caddr_t		pfield;
int		mask;
{
    FCB		*pFCB;

    /*
    If there are no monitors on the record, then we are done very quickly.
    However, if the mask has the log mask bit (and trending is enabled) or 
    the alarm mask bit is on, then we should still stick the record into 
    the ring buffer.
    */

    if (!(precord -> mlis.count))
	{
	if (mask & AGNT_LOG)
	    {
	    if (precord -> tenb)
		{
	    	caRbHandlerPut(precord, pfield, mask);
		return (0);
		}
	    }
	if (mask & AGNT_ALARM)
	    {
	    caRbHandlerPut(precord, pfield, mask);
	    }
	return (0);
	}

    /*
    We'll enqueue a message to the server task if a field that we are monitoring
    is also interested in the mask presented by the caller.
    */

    FASTLOCK(&(precord -> mlok));
    pFCB = (FCB *) (precord -> mlis.node.next);
    while (pFCB)
	{
	if (pfield == (pFCB -> dbAddr.pfield))
	    if (pFCB -> currMask & mask)
		{
		caRbPut(pFCB, precord, pfield, pFCB -> dbAddr.field_size, (short) mask);
		break;
		}
	pFCB = pFCB -> nextFCB;
	}
    FASTUNLOCK(&(precord -> mlok));

    /*
    If an entry was stuck into the queue, then we are done.
    */

    if (pFCB)
	return (0);

    /*
    There may have been monitors on the record, but not one that worked this call.
    We might have to still put an entry in the queue for any trend data loggers and 
    alarm handlers that may be out there. If the alarm bits in the mask are clear,
    then check the tenb field in the record. If that is clear, then the point is
    not on trend and we should drop the message.
    */

    if ((mask & AGNT_LOG) && precord -> tenb)
	{
        caRbHandlerPut(precord, pfield, mask);
	return (0);			/* Even if alarm is on, one send is enough */
	}
    if (mask & AGNT_ALARM)
	{
	caRbHandlerPut(precord, pfield, mask);
	}
    return (0);
}
