/*serverDump.c,v 1.1.1.1 1999/08/03 14:44:32 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		DUMP CA SERVER CONTROL BLOCKS			*/
/*								*/
/*	This is a couple of shell callable routines and a set	*/
/*	of utility routines that print each of the control 	*/
/*	blocks that the remote Channel Access Server uses to 	*/
/*	keep track of its open channels and monitors.		*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <vxWorks.h>
#include <types.h>
#include <semLib.h>
#include <inetLib.h>
#include <stdio.h>

#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#include "serverControl.h"
#include "dbRecDes.h"

extern CCB	*firstClient;

/* protype definitions */

int	clientDump (void);
void	dumpCCB (CCB *pCCB);
void	dumpFCB (FCB *pFCB);
void	dumpMCB (MCB *pMCB);

/******
	Dump control blocks organized by client
******/

int clientDump (void)
{
    CCB		*pDumpCCB;
    MCB		*pDumpMCB;

    printf("DUMP BY CLIENT:\n\n");
    pDumpCCB = firstClient;
    while(pDumpCCB)
	{
	dumpCCB(pDumpCCB);
	pDumpMCB = pDumpCCB -> firstChan;
	while (pDumpMCB)
	    {
	    dumpMCB(pDumpMCB);
	    dumpFCB(pDumpMCB -> fldCntl);
	    pDumpMCB = pDumpMCB -> nextClient;
	    }
	pDumpCCB = pDumpCCB -> nextClient;
	}

    return (0);
}

void dumpCCB (CCB *pCCB)
{
    struct	in_addr iaddr;
    char	inet[20];

    iaddr = pCCB -> clientInet;

    printf("CCB at %X:\n", (int) pCCB);
    inet_ntoa_b(iaddr, inet);
    printf("\tclientInet: %s\tclientSession: %d\tclientFd: %d\n", inet, pCCB -> clientSession,
	pCCB -> clientFd);
    printf("\ttxBuffSize: %d\tnextClient: %p\tfirstChan: %p\n", pCCB -> txBuffSize, 
	(void *)(pCCB -> nextClient), (void *)(pCCB -> firstChan));
}

void dumpMCB (MCB *pMCB)
{
    printf("MCB at %X:\n", (int) pMCB);
    printf("\tchanMask: %X\tdataType: %X\tnumElem: %d\tclient: %p\n", pMCB -> chanMask,
	pMCB -> dataType, pMCB -> numElem, (void *)(pMCB -> client));
    printf("\tnextMCB: %X\tnextClient: %X\tprevClient: %X\tfldCntl: %X\thandle: %X\n",
	(int) (pMCB -> nextMCB), (int) (pMCB -> nextClient), (int) (pMCB -> prevClient),
	(int) (pMCB -> fldCntl), (int) (pMCB -> handle));
}

void dumpFCB (FCB *pFCB)
{
    printf("FCB at %X:\n", (int) pFCB);
    printf("\tdbAddr.precord: %X (%s)\tdbAddr.pfield: %X\tdbAddr.pfldDes: %X (%4s)\n",
	(int) (pFCB -> dbAddr.precord), (char *) (pFCB -> dbAddr.precord),
	(int) (pFCB -> dbAddr.pfield), (int) (pFCB -> dbAddr.pfldDes),
	((struct fldDes *)(pFCB -> dbAddr.pfldDes)) -> fldname);
    printf("\tdbAddr.no_elements: %ld\tdbAddr.record_type: %d\tdbAddr.field_type: %d\n",
	pFCB -> dbAddr.no_elements, pFCB -> dbAddr.record_type, pFCB -> dbAddr.field_type);
    printf("\tdbAddr.field_size: %d\tdbAddr.special: %d\tdbAddr.choice_set: %d\n",
	pFCB -> dbAddr.field_size, pFCB -> dbAddr.special, pFCB -> dbAddr.choice_set);
    printf("\tdbAddr.field_type: %d\tnextFCB: %X\tprevFCB: %X\tfirstChan: %X\tcurrMask: %X\n",
	pFCB -> dbAddr.field_type, (int) (pFCB -> nextFCB), (int) (pFCB -> prevFCB), 
	(int) (pFCB -> firstChan), pFCB -> currMask);
}
