/*serverMgt.c,v 1.1.1.1 1999/08/03 14:44:31 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*	BUFFER MANAGEMENT FOR SERVER CONTROL STRUCTURES		*/
/*								*/
/*	This set of functions is very similar to buffPoolMgt.c	*/
/*	and memMgt.c. The purpose is to set up our own pool	*/
/*	of memory so that we can minimize the number of trips	*/
/*	that we take through the the vxWorks memLib routines.	*/
/*								*/
/*	Author: Gordon Uchenick					*/
/*								*/
/****************************************************************/
#ifdef	vxWorks
#include <vxWorks.h>
#include <memLib.h>
#include <string.h>
#include <stdlib.h>
#include <inetLib.h>
#else
#include <sys/types.h>
#include <in.h>
#endif	vxWorks

#ifdef	vxWorks
#include "dbDefs.h"
#include "dbCommon.h"
#include "link.h"
#include "db_field_log.h"
#endif	vxWorks

#include "serverControl.h"

/****************************************************************/
/*			TUNING PARAMETERS			*/
/*								*/
/*	Whenever we allocate structures from the operating	*/
/*	system, we allocate a bunch of them. These variables	*/
/*	control the size of the chunks.				*/
/*								*/
/****************************************************************/

#define FCB_ALLOC	20
#define	MCB_ALLOC	20

/******
	Pointers to the internal memory pools
******/

static FCB	*fcbPool = 0;
static MCB	*mcbPool = 0;


/****************************************************************/
/*								*/
/*		ALLOCATE A FIELD CONTROL BLOCK			*/
/*								*/
/****************************************************************/

FCB	*newFCB(void)
{
    FCB		*currFCB;
    int		index;

    /*
    If the pool is empty, fill it up again.
    */

    if (!fcbPool)
	{
	fcbPool = (FCB *) malloc(FCB_ALLOC * sizeof(FCB));
	if (!fcbPool)
	    return (0);
	currFCB = fcbPool;
	index = FCB_ALLOC;
	while (index--)
	    {
	    if (index)
		currFCB -> nextFCB = (struct fCb *)(currFCB + 1);
	    else
		currFCB -> nextFCB = 0;
	    currFCB++;
	    }
	}

    /*
    Take the one off the front of the queue and clear it.
    */

    currFCB = fcbPool;
    fcbPool = (FCB *) currFCB -> nextFCB;
#if defined(SOLARIS)
    memset((char *) currFCB, 0,sizeof(FCB));
#else
    bzero((char *) currFCB, sizeof(FCB));
#endif
    return (currFCB);
}


/****************************************************************/
/*								*/
/*		FREE A FIELD CONTROL BLOCK (FCB)		*/
/*								*/
/****************************************************************/

void deleteFCB (oldFCB)
FCB	*oldFCB;
{
    /*
    Push this one onto the front of the pool.
    */

    oldFCB -> nextFCB = (struct fCb *) fcbPool;
    fcbPool = oldFCB;
}


/****************************************************************/
/*								*/
/*		ALLOCATE A MONITOR CONTROL BLOCK (MCB)		*/
/*								*/
/****************************************************************/

MCB	*newMCB(void)
{
    MCB		*currMCB;
    int		index;

    /*
    If the pool is empty, fill it up again.
    */

    if (!mcbPool)
	{
	mcbPool = (MCB *) malloc(MCB_ALLOC * sizeof(MCB));
	if (!mcbPool)
	    return (0);
	currMCB = mcbPool;
	index = MCB_ALLOC;
	while (index--)
	    {
	    if (index)
		currMCB -> nextMCB = (struct mCb *)(currMCB + 1);
	    else
		currMCB -> nextMCB = 0;
	    currMCB++;
	    }
	}

    /*
    Take the one off the front of the queue and clear it.
    */

    currMCB = mcbPool;
    mcbPool = (MCB *) currMCB -> nextMCB;
#if defined(SOLARIS)
    memset((char *) currMCB, 0,sizeof(MCB));
#else
    bzero((char *) currMCB, sizeof(MCB));
#endif
    return (currMCB);
}


/****************************************************************/
/*								*/
/*		FREE A MONITOR CONTROL BLOCK (MCB)		*/
/*								*/
/****************************************************************/

void deleteMCB (oldMCB)
MCB	*oldMCB;
{
    /*
    Push this one onto the front of the pool.
    */

    oldMCB -> nextMCB = (struct mCb *) mcbPool;
    mcbPool = oldMCB;
}
