/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include	<unistd.h>
#include	<stdio.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<signal.h>	/* for signal handling */
#include	<malloc.h>
#include	<memory.h>
#include	<string.h>
#include	<syslog.h>	/* for messages to system console */
#include	<time.h>	/* for time_t and ctime() */
#include	<errno.h>
#include	<d4all.h>	/* for codebase */
#include	<cadef.h>	/* channel access .../share/epicsH */
#include	<dbDefs.h>	/* .../share/epicsH */

#define	SECURITY_FILE_PATH	"/home/tis4000/tasks/"
#define	SECURITY_FILE_NAME	"security_monitor.lst"
#define	SECURITY_FILE_REPORT	"/tis4000db/security.dbf"

struct	security
{
    struct security	*NextRtu;
    chid		statuschid;;
    chid		oprchid;
    chid		logchid;
    evid		statusevid;
    char		rtuname[PVNAME_SZ];
    short		status;
    char		operator[18];
    char		logtime[18];
    short		numpts;
};

struct security	*FirstRtu = (struct security *)0;
struct security	*LastRtu;

char	*logStrs[] = 
	{
	"logged off",
	"logged on",
	"timed out"
	};

/* define stuff for codebase and the output database file */

FIELD4INFO	securityFieldInfo [] =
	{
	{ "OPERATOR",	'C',	20,	0},
	{ "STATUS",	'C',	18,	0},
	{ "LOGTIME",	'C',	18,	0},
	{ 0, 0, 0, 0},
	};

TAG4INFO	securityTagInfo [] = 
	{
	{ "OPERATOR",	"OPERATOR+LOGTIME", 	0, 	0, 	0},
	{ 0, 0, 0, 0, 0},
	};

CODE4	cb;
DATA4	*securityData;
TAG4	*operatortag;
FIELD4	*operator;
FIELD4	*opstatus;
FIELD4	*logtime;


/* function prototypes */

void	connect_handler (struct connection_handler_args);
void	free_mem (void);
void	rtuDataHandler (struct event_handler_args);
void	rtuEventHandler (struct event_handler_args);
void	secmonSignalHandler (int);

    


/*****************************************************************************/

main (argc, argv)
    int		argc;
    char	*argv[];
{
    FILE		*rtufilefd;	/* input file with rtu names */
    char		buff[128];
    chid		pCh;
    char		*p;
    int			status;
    time_t		start_time;
    struct security	*newrtu;
    struct stat		statbuf;	/* for stat () */

    start_time = time (0);

    /* set up signal handler to capture ^C */
    signal (SIGINT, secmonSignalHandler);

    /* open the system log to send messages to the condsole */
    openlog ("secmon", LOG_PID, LOG_LOCAL1);

    syslog (LOG_ALERT|LOG_LOCAL1, "started at %s", ctime (&start_time));

    sprintf(buff, "%s%s", SECURITY_FILE_PATH, SECURITY_FILE_NAME);
    if ((rtufilefd = fopen (buff, "r")) == NULL)
    {
	syslog (LOG_ERR|LOG_LOCAL1, "open error for %s (%m)\n", buff);
	exit (1);
    }

    d4init (&cb);
    cb.go_error = 0;
    cb.open_error = 0;
    cb.create_error = 0;
    cb.skip_error = 0;
    cb.field_name_error = 0;

    if (stat (SECURITY_FILE_REPORT, &statbuf) == -1)
    {
	if (errno == ENOENT)
	{
	    if ((securityData = d4create(&cb, SECURITY_FILE_REPORT, securityFieldInfo, securityTagInfo)) == 0)
	    {
		syslog (LOG_ALERT|LOG_LOCAL1, "failed to create %s (%d)\n", 
			SECURITY_FILE_REPORT, cb.error_code);
		fclose (rtufilefd);
		exit (1);
	    }
	}
	else
	{
	    syslog (LOG_ERR|LOG_LOCAL1, "stat error for %s (%m)\n", SECURITY_FILE_REPORT);
	    fclose (rtufilefd);
	    exit (1);
	}
    }
    else
    {
	if ((securityData  = d4open (&cb, SECURITY_FILE_REPORT)) == 0)
	{
	    syslog (LOG_ALERT|LOG_LOCAL1, "failed to open %s (%d)\n",
		    SECURITY_FILE_REPORT, cb.error_code);
	    fclose (rtufilefd);
	    exit (1);
	}
    }
    operatortag = d4tag (securityData, "OPERATOR");
    d4tag_select (securityData, operatortag);
    operator = d4field (securityData, "OPERATOR");
    opstatus = d4field (securityData, "STATUS");
    logtime = d4field (securityData, "LOGTIME");

    while (ca_task_initialize () != ECA_NORMAL)
	sleep (1);

    while (1)
    {
	bzero (buff, sizeof (buff));
	fgets (buff, sizeof (buff) - 1, rtufilefd);
	if (feof (rtufilefd) != 0)
	    break;
	if ((newrtu = (struct security *)calloc (1, sizeof (struct security))) == NULL)
	{
	    syslog (LOG_ERR|LOG_LOCAL1, "calloc failed (%m)\n");
	    ca_task_exit ();
	    free_mem ();
	    exit (1);
	}
	if ((p = strchr (buff, '\n')) != NULL)
	    *p = NULL;
	strcpy (newrtu -> rtuname, buff);
	strcat (buff, "-STATUS");
	if ((status = ca_build_and_connect (buff,
					    AGNT_NOCONN,
					    0,
					    &pCh,
					    0,
					    connect_handler,
					    (void *)newrtu)) == ECA_NORMAL)
	{
	    newrtu -> statuschid = pCh;
	}
	else
	{
	    syslog (LOG_ERR|LOG_LOCAL1, "ca build for %s failed (%s)\n",
				buff, ca_message());
	    free_mem();
	    ca_task_exit ();
	    exit (1);
	}
	strcpy (buff, newrtu -> rtuname);
	strcat (buff, "-OPR");
	if ((status = ca_build_and_connect (buff,
					    AGNT_NOCONN,
					    0,
					    &pCh,
					    0,
					    connect_handler,
					    (void *)newrtu)) == ECA_NORMAL)
	{
	    newrtu -> oprchid = pCh;
	}
	else
	{
	    syslog (LOG_ERR|LOG_LOCAL1, "ca build for %s failed (%s)\n",
			      buff, ca_message());
	    free_mem();
	    ca_task_exit ();
	    exit (1);
	}
	strcpy (buff, newrtu -> rtuname);
	strcat (buff, "-LOGTIME");
	if ((status = ca_build_and_connect (buff,
					    AGNT_NOCONN,
					    0,
					    &pCh,
					    0,
					    connect_handler,
					    (void *)newrtu)) == ECA_NORMAL)
	{
	    newrtu -> logchid = pCh;
	}
	else
	{
	    syslog (LOG_ERR|LOG_LOCAL1, "ca build for %s failed (%s)\n",
			      buff, ca_message());
	    free_mem();
	    ca_task_exit ();
	    exit (1);
	}
	if (FirstRtu == (struct security *)0)
	    FirstRtu = newrtu;
	else
	    LastRtu -> NextRtu = newrtu;
	LastRtu = newrtu;
    }
    fclose (rtufilefd);
    while (1)
    {
	ca_pend_io (10.0);
    }
}


/*****************************************************************************/

void connect_handler (connArgs)
    struct connection_handler_args	connArgs;
{
    chid		pCh;
    evid		pEv;
    chtype		fld_type;
    long		status;
    struct security	*pSecurity;
    char		errmsg[256];

    pCh = connArgs.chid;
    pSecurity = (struct security *)ca_puser (pCh);

    fld_type = ca_field_type (pCh);
    if (connArgs.op == CA_OP_CONN_UP)
    {
	if (!pSecurity -> statusevid)
	{
	    if ((status = ca_add_event (fld_type, 
					pCh, 
					rtuEventHandler,
					(void *)pSecurity, 
					&pEv)) != ECA_NORMAL)
	    {
		syslog (LOG_ERR | LOG_LOCAL1, "add event failed for %s-%s (%s)\n",
				  pSecurity -> rtuname, "STATUS", ca_message (status));
	    }
	    else
	    {
		pSecurity -> statusevid = (void *)pEv;
		ca_flush_io ();
	    }
	}
	/* on a reconnect */
	else
	{
	    if ((status = ca_get_callback (fld_type,
					   pCh,
					   rtuEventHandler,
					   (void *)pSecurity)) != ECA_NORMAL)
	    {
		syslog (LOG_ERR | LOG_LOCAL1, "reconnect callback failed for %s-%s (%s)\n",
				       pSecurity -> rtuname, "STATUS", ca_message(status));
	    }
	}
    }

    return;
}


/*****************************************************************************/

void free_mem ()
{
    struct security	*curr_rtu;
    struct security	*prev_rtu;

    if (FirstRtu == (struct security *) 0)
	return;

    curr_rtu = prev_rtu = FirstRtu;
    while (curr_rtu)
    {
	prev_rtu = curr_rtu;
	curr_rtu = curr_rtu -> NextRtu;
	free ((void *)prev_rtu);
    }

    return;
}


/*****************************************************************************/

void rtuDataHandler (evArgs)
    struct event_handler_args	evArgs;
{
    chid		pCh;
    struct security	*pSecurity;
    union db_access_val	*pData;
    char		*p;

    pCh = evArgs.chid;
    pData = (union db_access_val *)evArgs.dbr;
    pSecurity = (struct security *)evArgs.usr;
    if ((p = strrchr (ca_name (pCh), '-')) == NULL)
	return;
    p++;
    if (strncmp (p, "OPR", 3) == NULL)
    {
	strncpy (pSecurity -> operator, pData -> strval, 
					sizeof (pSecurity -> operator));
	if (pSecurity -> operator[0] == NULL)
	    return;
	pSecurity -> numpts++;
    }
    else if (strncmp (p, "LOGTIME", 7) == NULL)
    {
	strncpy (pSecurity -> logtime, pData -> strval,
					sizeof (pSecurity -> logtime));
	pSecurity -> numpts++;
    }
    if (pSecurity -> numpts >= 2)
    {
	if (pSecurity -> status >= (sizeof (logStrs)/sizeof (char *)))
	    return;
	d4go_eof (securityData);
	d4append_start (securityData, 0);
	d4blank (securityData);
	f4assign (operator, pSecurity -> operator);
	f4assign (opstatus, logStrs[pSecurity -> status]);
	f4assign (logtime, pSecurity -> logtime);
	d4append (securityData);
	d4flush (securityData);
	fprintf (stdout, "Operator %s %s at %s\n",
			  pSecurity -> operator,
			  logStrs[pSecurity -> status],
			  pSecurity -> logtime);
    }
	
    return;
}


/*****************************************************************************/

void rtuEventHandler (evArgs)
    struct event_handler_args	evArgs;
{
    struct security	*pSecurity;
    union db_access_val	*pData;

    pSecurity = (struct security *)(evArgs.usr);
    pSecurity -> numpts = 0;
    pData = (union db_access_val *)evArgs.dbr;
    pSecurity -> status = pData -> intval;

    if ((ca_state (pSecurity -> oprchid)) == cs_conn)
    {
	ca_get_callback (DBR_STRING, pSecurity -> oprchid, rtuDataHandler,
			 (void *)pSecurity);
    }
    if ((ca_state (pSecurity -> logchid)) == cs_conn)
    {
	ca_get_callback (DBR_STRING, pSecurity -> logchid, rtuDataHandler,
			 (void *)pSecurity);
    }

    return;
}



/*****************************************************************************/

void secmonSignalHandler (which_sig)
    int		which_sig;
{
    time_t	end_time;

    switch (which_sig)
    {
	case SIGINT:
	    ca_task_exit ();		/* close channel access */
	    d4flush (securityData);
	    d4close (securityData);	/* close the report database */
	    end_time = time (0);
	    syslog (LOG_ALERT|LOG_LOCAL1, "exiting due to ^C: %s", ctime (&end_time));
	    closelog ();
	    exit (0);
	default:
	    break;
    }
}

