/*seq_monitor.c,v 1.1.1.1 1999/08/03 14:44:37 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/************************************************************************/
/*                                                                      */
/*            State Notation Language program monitor                   */
/*                                                                      */
/*  This task is responsible for starting SNL programs, monitoring	*/
/*  their status, and program termination.				*/
/*                                                                      */
/************************************************************************/

#include <vxWorks.h>
#include <symLib.h>
#include <sysSymTbl.h>
#include <socket.h>
#include <sockLib.h>
#include <selectLib.h>
#include <stdio.h>
#include <ioLib.h>
#include <usrLib.h>
#include <unldLib.h>
#include <string.h>
#include <tickLib.h>
#include <in.h>
#include <errno.h>
#include <sys/fcntlcom.h>
#include <task_params.h>        /* For getting task priorities */

#include "sysseq.h"
#include "restable.h"
#include "resseq.h"
#include "mon_defs.h"
#include "cadef.h"

extern int ca_import (int tid);
extern int seq_sendState (SPROG *pSP);
extern int resSeqStatus (unsigned short seq_token, int *activity);

/* forward reference in this file */
int snlStart (char *ProgramName, char *Parameters, int StackSize, char *Resource, char *ActualName);

/* Tuning Parameters */

#define MAX_FDS		10	/* max number of clients we can handle */

/* Local data */

static int              rmfd;		/* fd for listen socket */
static fd_set           rmReadFds;	/* fd mask for select() */

typedef struct {
	unsigned long	Inet;		/* Connected Inet address */
	int		fd;		/* connection fd */
	long		lastmsg;	/* last message report rx time */
} CONNECTION;

static	CONNECTION	resFds [MAX_FDS];

/* Message buffers */

static	MSG_HEADER	msgHead;
static	char		msgBody [1000];

static	int	DebugOn = FALSE;
static	int	Running = TRUE;

/* Module Lists - Keeps track of all modules loaded. */

typedef	struct {
    MODULE_NAME		Name;
    int			Location;	
} MODULE_LIST_ENTRY;
static	MODULE_LIST_ENTRY	ModuleList [MAX_MODULES];
static	int	NumModules;

/* Program List - Keeps track of all invokations of each module */

typedef	struct {	/* Program list entry */
    PROGRAM_NAME	OrigName;
    PROGRAM_NAME	ActualName;
    PARAMETERS		Parameters;
    RESOURCE_NAME	Resource;
    int			TaskId;
} ProgListEntry;
static	ProgListEntry	ProgramList [MAX_PROGRAMS];
static	int	NumPrograms;	

/* Functions defined in this module */

	int	snlMonInit ();
static	void	snlMonTask ();
static	void	acceptNewClient ();
static	int	readClientMsg ();
static	void	doDisconnect ();
static	int	sendMessage ();

static	void	processRequest ();
static	void	processPing ();
static	void	processLoad ();
int snlLoad (char *ModuleName);
static	void	processStart ();
static	void	processKill ();
int snlKill (char *ProgramName, int TaskId, char *Parameters);
static	void	processUnload ();
int snlUnload (char *ModuleName);
static	void	processReportModules ();
static	void	processReportPrograms ();
static	void	processReportStateSets ();
static	void	processReportChans (int fd);
static	void	processReportVariables ();
static	void	processReportVariable (int fd);
static	void	processChangeVariable ();
static	void	processSuspend ();
static	void	processResume ();

static	BOOL	AddModule ();
static	BOOL	RemoveModule ();
static	int	FindModule ();
static	BOOL	ModTableFull ();

static	BOOL	AddProgram ();
static	BOOL	RemoveProgram ();
static	int	FindProgram ();
static	BOOL	ProgTableFull ();

/*----------*/
int snlMonInit(Deb)
    int		Deb;
{
    DebugOn = Deb;
 
    /* Spawn the snlMon task. For now, use the same priority as the */
    /* periodic scan task. */

    if (taskSpawn(SNLMON_NAME, SNLMON_PRI, VX_FP_TASK, SNLMON_STACK,
                  (FUNCPTR)snlMonTask, 0,0,0,0,0,0,0,0,0,0) == ERROR) {
        printErr("Can't spawn snlMon.\n");
        return (-1);
    }

    /* Initialize various things */

    NumModules = 0;
    NumPrograms = 0;

    /* Initialization complete */

    return(0);
}

/*----------*/
static void snlMonTask ()
{
    struct sockaddr_in serverAddr;    /* server's address   */
    short           serveport;
    register int    status, i;
    fd_set          tempfds;
    int		    addrSize;

    if (DebugOn)
	printf ("\nSNlMon started, Debug mode on!\n\n");

    if (rmfd != 0 && rmfd != ERROR)
        if ((status = close(rmfd)) == ERROR)
            printErr("snlMon: Unable to close open snlMon socket\n");

    /* Open a stream socket. */

    if ((rmfd = socket(AF_INET, SOCK_STREAM, 0)) == ERROR) {
        printErr("snlMon: Socket creation error\n");
        taskSuspend(0);
    }
    
    /* Zero the sock_addr structure */

    bzero ((char *)&serverAddr, sizeof(serverAddr)); 
    serverAddr.sin_family = AF_INET;
    serverAddr.sin_port = htons(SNLMON_PORT);

    /* get server port address */

    if (bind(rmfd, (struct sockaddr *)&serverAddr, sizeof(serverAddr)) == ERROR) {
        printErr("snlMon: Bind error\n");
        close(rmfd);
        taskSuspend(0);
    }
    addrSize = sizeof(serverAddr);
    if (getsockname (rmfd, (struct sockaddr *)&serverAddr, &addrSize)) {
        printErr("snlMon: getsockname error\n");
        close(rmfd);
        taskSuspend(0);
    }
    serveport = htons (serverAddr.sin_port);

    /* clear fd mask listen and accept new connections */

    FD_ZERO (&rmReadFds);
    FD_SET (rmfd, &rmReadFds); /* select on incoming connections */
    if (listen(rmfd, 0) == ERROR) {
        printErr("snlMon: Listen error\n");
        close(rmfd);
        taskSuspend(0);
    }

    while (Running) {
	tempfds = rmReadFds;
	status = select (sizeof(tempfds)*8, &tempfds,
                 NULL, NULL, NULL);
	if (DebugOn)
	    printf ("snlMon: select triggered, status = %d\n", status);
        if (status > 0) { /* One of the targets has activity */
            if (FD_ISSET (rmfd, &tempfds)) {
                acceptNewClient (rmfd);
                status--;
            }
            /* check the resource table fds for incoming signals */
	    for (i = 0; i < MAX_FDS && status; i++) {
                if (resFds[i].Inet && FD_ISSET (resFds[i].fd, &tempfds)) {
                    if (readClientMsg (resFds[i].fd))
			processRequest (resFds[i].fd);
		    else
			doDisconnect (resFds[i].fd);
                    status--;
                }
            }
        } else if (status < 0) { /* error processing */
            printErr("snlMon: error on select statement\n");
        }

    }

    /* Server has terminated so close all sockets */

    close (rmfd);
    for (i=0 ; i<MAX_FDS ; i++)
	if (resFds [i].fd != 0)
	    close (resFds [i].fd);

}

/*----------*/
static void acceptNewClient (rmfd)
int	rmfd;
{
    int     tempfd;
    struct  sockaddr_in newaddress;
    long    newInet;
    int	    i;
    int     blankIx, size;

    size = sizeof (newaddress);
    if ((tempfd = accept(rmfd, (struct sockaddr *)&newaddress, &size)) != ERROR) {
        /* scan the fd table for the connecting IP */
	newInet = ntohs(newaddress.sin_addr.s_addr);
        blankIx = MAX_FDS;
	for (i = 0; i < MAX_FDS; i++) {
	    if (resFds[i].Inet == newInet)
		return;
	    else if ((resFds[i].Inet == 0) && (blankIx == MAX_FDS)) {
		blankIx = i;
            }
	}
	if (i == MAX_FDS) {
            if (blankIx < MAX_FDS) {
		resFds[blankIx].Inet = newInet;
		resFds[blankIx].fd = tempfd;
		resFds[blankIx].lastmsg = tickGet ();
                FD_SET (tempfd, &rmReadFds);
		if (DebugOn)
		    printErr("snlMon: accepted new client, fd = %d\n", tempfd);
            } else {
  	        printErr("snlMon: accept out of resource fd map space\n");
		/* This condition may clear, don't do- taskSuspend(0); */
		close (tempfd);
            }
        }
    }
    else {
        printErr("snlMon: Accept error, errno = %d\n", errno);
        /* This condition may clear, don't do- taskSuspend(0); */
    }
}

/*----------*/
static int readClientMsg (fd)
int	fd;
{
    int		i;
    int		bytesGot, bytesNeeded;
    char	*p;

    /* Read the message header */

    bytesNeeded = sizeof (msgHead);
    p = (char *)&msgHead;
    while (bytesNeeded > 0) {
	bytesGot = read (fd, p, bytesNeeded);
	if (bytesGot <= 0)
	    return 0;
	p += bytesGot;
	bytesNeeded -= bytesGot;
    }

    if (DebugOn)
	printf ("Head received: Type=%ld, Length=%ld, Info=%ld\n",
			msgHead.Type, msgHead.Length, msgHead.Info);

    /* Read the message body */

    bytesNeeded = msgHead.Length;
    p = msgBody;
    while (bytesNeeded > 0) {
	bytesGot = read (fd, p, bytesNeeded);
	if (bytesGot <= 0)
	    return 0;
	p += bytesGot;
	bytesNeeded -= bytesGot;
    }

    if (DebugOn) {
	printf ("Message body: ");
	for (i=0 ; i<msgHead.Length ; i++)
	    printf ("%x ", msgBody [i] & 0xFF);
	printf ("\n");
    }

    return 1;
}

/*----------*/
static void doDisconnect (fd)
int	fd;
{
    int		i;

    if (DebugOn)
	printf ("Disconnecting, fd = %d\n", fd);

    /* find and close the fd struct */ 

    for (i = 0; i < MAX_FDS; i++) {
	if (resFds[i].fd == fd)
	    break;
    }
    resFds[i].Inet = 0;
    resFds[i].fd = 0;
    resFds[i].lastmsg = 0;

    /* close socket */

    close (fd);
    FD_CLR (fd, &rmReadFds);
}

/*----------*/
static int sendMessage (Type, Body, Length, Info, fd)

    int		Type, Length, Info, fd;
    char	*Body;

{
    MSG_HEADER	msgHeader;
    int		i;

    if (DebugOn) {
	printf ("Sending message: Type=%d, Sequence=%ld, Length=%d, Info=%d, Msg = ",
		Type, msgHead.seq, Length, Info);
	for (i=0 ; i<Length ; i++)
	    printf ("%x ", *(Body + i));
	printf ("\n");
    }

    /* Send the message header */

    msgHeader.Type = Type;
    msgHeader.seq = msgHead.seq;
    msgHeader.Length = Length;
    msgHeader.Info = Info;
    write (fd, (char *)&msgHeader, sizeof (msgHeader));

    /* Send the message body */

    if (Length > 0)
	write (fd, Body, Length);

    return 1;
}

/*----------*/
static void processRequest (fd)

    int		fd;

{
    switch (msgHead.Type) {
	case PING_MSG_TYPE:
	    processPing (fd);
	    break;
	case LOAD_MSG_TYPE:
	    processLoad (fd);
	    break;
	case START_MSG_TYPE:
	    processStart (fd);
	    break;
	case KILL_MSG_TYPE:
	    processKill (fd);
	    break;
	case UNLOAD_MSG_TYPE:
	    processUnload (fd);
	    break;
	case TERMINATE_MSG_TYPE:
	    Running = 0;
	    break;
	case REPMOD_MSG_TYPE:
	    processReportModules (fd);
	    break;
	case REPPROG_MSG_TYPE:
	    processReportPrograms (fd);
	    break;
	case REPSS_MSG_TYPE:
	    processReportStateSets (fd);
	    break;
	case REPCHANS_MSG_TYPE:
	    processReportChans (fd);
	    break;
	case REPVARS_MSG_TYPE:
	    processReportVariables (fd);
	    break;
	case REPVAR_MSG_TYPE:
	    processReportVariable (fd);
	    break;
	case CHVAR_MSG_TYPE:
	    processChangeVariable (fd);
	    break;
	case SUSPEND_MSG_TYPE:
	    processSuspend (fd);
	    break;
	case RESUME_MSG_TYPE:
	    processResume (fd);
	    break;
	default:
	    break;
    }
}

/*----------*/
static void processPing (fd)

    int		fd;

{
    if (DebugOn)
        printf ("Ping!  Info = %lx\n", msgHead.Info);

    /* Echo the Info field of message as a reply */

    sendMessage (PING_REPLY_TYPE, NULL, 0, msgHead.Info, fd);
}

/*----------*/
static void processLoad (fd)

    int		fd;

{
    LOAD_MSG	*loadMsg;
    int		Result;
		
    loadMsg = (LOAD_MSG *)msgBody;

    Result = snlLoad (loadMsg->ModuleName);

    sendMessage (LOAD_REPLY_TYPE, NULL, 0, Result, fd);
}

/*----------*/
int snlLoad (ModuleName)

    char	*ModuleName;

{
    MODULE_ID		Location;

    if (DebugOn)
	printf ("Received Load request: ModuleName = %s\n", ModuleName);

    /* Check if there is room in the module table.  If yes, then load the module */

    if (FindModule (ModuleName) == -1) {	/* Not already loaded */
	if (!ModTableFull ()) {		/* room in the table */
	    if ((Location = ld (0, TRUE, ModuleName)) != NULL) {
		AddModule (ModuleName, Location);
		if (DebugOn)
		    printf ("%s loaded, Location = %x(%d)\n", ModuleName, (unsigned int)Location, (int)Location);
		return TRUE;
	    }
	}
    } else {
	if (DebugOn)
	    printf ("Unable to load %s\n", ModuleName);
    }

    return FALSE;
}

/*----------*/
static void processStart (fd)

    int		fd;

{

    START_MSG	*startMsg;
    START_REPLY	startReply;
    PROGRAM_NAME	ActualName;
    int			Result;

    startMsg = (START_MSG *)msgBody;

    strcpy (ActualName, "");
    Result = snlStart (startMsg->ProgramName, startMsg->Parameters,
			startMsg->StackSize, startMsg->Resource, ActualName); 

    strcpy (startReply.ProgramName, ActualName);
    sendMessage (START_REPLY_TYPE, (char *)&startReply, sizeof (startReply), Result, fd);
}

/*----------*/
int snlStart (ProgramName, Parameters, StackSize, Resource, ActualName)

    char	*ProgramName, *Parameters, *Resource, *ActualName;
    int		StackSize;

{
    int		Status, tid;
    SPROG	*pSP;
    char	EntrySym [50];
    SYM_TYPE	pType;
    TASK_DESC	Desc;

    if (DebugOn) {
	printf ("Received Start request: ProgramName=%s, Parameters=%s\n",
		ProgramName, Parameters);
	printf ("Received Start request: StackSize=%d, Resource=%s\n",
		StackSize, Resource);
    }

    /* Lookup the address of the entry point and start the task */

    if (!ProgTableFull ()) {
	sprintf (EntrySym, "_%s", ProgramName);
	Status = symFindByName (sysSymTbl, EntrySym, (char **)&pSP, &pType);
	if (Status == OK) {
	    tid = seq (pSP, Parameters, StackSize, Resource);
	    if (tid > 0) {
		if (taskInfoGet (tid, &Desc) == OK) {
		    if (ActualName)
			strcpy (ActualName, Desc.td_name);
		    AddProgram (ProgramName, Desc.td_name, Parameters,
					Resource, tid);
		    return TRUE;
		}
	    }
	}
    }

    return FALSE;    
}

/*----------*/
static void processKill (fd)

    int		fd;

{

    KILL_MSG	*killMsg;
    int		Result;

    /* A program is killed by killing the first task in its task list. */
    /* The rest of the tasks will be killed automatically. */

    killMsg = (KILL_MSG *)msgBody;

    Result = snlKill (killMsg->ProgramName, killMsg->TaskId, killMsg->Parameters);

    sendMessage (KILL_REPLY_TYPE, NULL, 0, Result, fd);
 }

/*----------*/
int snlKill (ProgramName, TaskId, Parameters)

    char	*ProgramName, *Parameters;
    int		TaskId;

{
    SPROG	*pSP;

    if (DebugOn)
	printf ("Received Kill request: ProgName=/%s/, Parameters=/%s/\n",
			ProgramName, Parameters);

    /* Find the taskid for the requested */

    if (FindProgram (ProgramName, TaskId) != -1) {
	if ((pSP = seqQryFind (TaskId)) != NULL) {
	    if (taskDelete (pSP->task_id) == OK) {
		RemoveProgram (ProgramName, pSP->task_id, Parameters);	
		return TRUE;
	    }
	}
    }

    return FALSE;
}

/*----------*/
static void processUnload (fd)

    int		fd;

{
    UNLOAD_MSG	*unloadMsg;
    int		Result;

    unloadMsg = (UNLOAD_MSG *)msgBody;

    Result = snlUnload (unloadMsg->ModuleName);

    sendMessage (UNLOAD_REPLY_TYPE, NULL, 0, Result, fd);
}

/*----------*/
int snlUnload (ModuleName)

    char	*ModuleName;

{
    printf ("Received Unload request: ModuleName = %s\n", ModuleName);

    /* Remove the module from the module list. */

    if (RemoveModule (ModuleName)) {
	if (unld (ModuleName, 0) == OK) {
	    if (DebugOn)
		printf ("%s unloaded.\n", ModuleName);
	    return TRUE;
	}
    } else {
	if (DebugOn)
	    printf ("Unable to unload %s\n", ModuleName);
    }

    return FALSE;
}

/*----------*/
static void processReportModules (fd)

    int	fd;

{
    int			Length, i;
    REPMOD_REPLY	*repmodsReply, *p;

    if (DebugOn)
	printf ("Received RepMod request\n");

    Length = sizeof (REPMOD_REPLY) * NumModules;
    repmodsReply = p = (REPMOD_REPLY *) calloc (1, Length);

    for (i=0 ; i<NumModules ; i++, p++) {
	strcpy (p->ModuleName, ModuleList [i].Name);
    }

    sendMessage (REPMOD_REPLY_TYPE, (char *)repmodsReply, Length, NumModules, fd);
    free (repmodsReply);
}
	 
/*----------*/
static void processReportPrograms (fd)

    int	fd;

{
    int			Length, i, j;
    REPPROG_REPLY	*repprogReply, *p;
    SPROG		*pSP;

    if (DebugOn)
	printf ("Received RepProg request\n");

    Length = sizeof (REPPROG_REPLY) * NumPrograms;
    repprogReply = p = (REPPROG_REPLY *) calloc (1, Length);

    for (i=0 ; i<NumPrograms ; i++, p++) {

	/* Check to see if the program is waiting on channel connect */

        if ((pSP = seqQryFind (ProgramList [i].TaskId)) != NULL) {
	    /* Copy in the easy stuff */
	    strcpy (p->ProgramName, ProgramList [i].ActualName);
	    p->TaskId = ProgramList [i].TaskId;
	    strcpy (p->Parameters, ProgramList [i].Parameters);
	    strcpy (p->Resource, ProgramList [i].Resource);
	    p->Stat = REPPROG_STARTED;

	    if (pSP->nchan != pSP->conn_count)
		p->Stat = REPPROG_CHWAIT;
	}
	else {
	    j = i+1;
	    while (j < NumPrograms) {
		ProgramList [j-1] = ProgramList [j];
		j++;
	    }
	    NumPrograms--;
	}
    }
    /* recompute length in case of deleted members */
    Length = sizeof (REPPROG_REPLY) * NumPrograms;

    sendMessage (REPPROG_REPLY_TYPE, (char *)repprogReply, Length, NumPrograms, fd);
    free (repprogReply);
}

/*----------*/
static void processReportStateSets (fd)

    int	fd;

{

    REPSS_MSG	*repssMsg;
    REPSS_REPLY	*repssReply, *p;
    int		tid, nss;
    SPROG	*pSP;
    STATE	*pST;
    SSCB	*pSS;
    TASK_DESC	TaskDesc;

    repssMsg = (REPSS_MSG *)msgBody;

    if (DebugOn)
	printf ("Received RepStateSets request: Program Name = %s\n", repssMsg->ProgramName);

    /* Find all the tasks for this program and send replies */

    if (FindProgram (repssMsg->ProgramName, repssMsg->TaskId) != -1) {
	tid = repssMsg->TaskId; /* taskIdFigure (repssMsg->ProgramName); */
	if ((pSP = seqQryFind (tid)) != NULL) {
	    pSS = pSP->sscb;
	    repssReply = p = (REPSS_REPLY *) calloc (1, sizeof (REPSS_REPLY) * pSP->nss);
	    for (nss=0 ; nss < pSP->nss ; nss++, pSS++) {
		strcpy (p->StateSetName, pSS->name);
		strcpy (p->InitState, pSS->states->name);
		pST = pSS->states + pSS->current_state;
		strcpy (p->CurState, pST->name);
		pST = pSS->states + pSS->prev_state;
		strcpy (p->PrevState, pST->name);
		p->TimeEnt = pSS->time / 60.0;		
		p->TimeIn = (tickGet () - pSS->time) / 60.0;		
	        p->Priority = pSP->task_priority;
		p->TaskId = pSS->task_id;
		taskInfoGet (pSS->task_id, &TaskDesc);
		p->Stat = TaskDesc.td_status;
		p++;
	    }
	    sendMessage (REPSS_REPLY_TYPE, repssReply,
			sizeof (REPSS_REPLY) * pSP->nss, pSP->nss, fd);	
	    free (repssReply);
	    return;
	}
    }
    sendMessage (REPSS_REPLY_TYPE, NULL, 0, 0, fd);
}

/*----------*/
static void processReportChans (fd)

    int		fd;

{

    REPCHANS_MSG	*repchansMsg;
    REPCHANS_REPLY	*repchansReply, *p;
    int			tid, nch;
    SPROG		*pSP;
    CHAN		*pDB;
    char		textvalue [40];

    repchansMsg = (REPCHANS_MSG *)msgBody;

    if (DebugOn)
	printf ("Received RepChans request: Program Name = %s\n", repchansMsg->ProgramName);

    /* Find all the channels for this program */

    if (FindProgram (repchansMsg->ProgramName, repchansMsg->TaskId) != -1) {
	tid = repchansMsg->TaskId; /* taskIdFigure (repchansMsg->ProgramName); */
	if ((pSP = seqQryFind (tid)) != NULL) {
	    repchansReply = p = (REPCHANS_REPLY *) calloc (1, sizeof (REPCHANS_REPLY) * pSP->nchan);
	    pDB = pSP->channels;
	    for (nch=0 ; nch < pSP->nchan ; nch++, pDB++) {
		strcpy (p->ChanName, pDB->db_name);
		if (pDB->var == NULL)
		    sprintf(textvalue, "invalid buffer type: %12d", pDB->get_type);
		else if (dbr_type_is_STRING(pDB->get_type))
		    sprintf(textvalue, " %12.12s", (char *)pDB->var);
		else if (dbr_type_is_SHORT(pDB->get_type))
		    sprintf(textvalue, " %12d", *(short *)pDB->var);
		else if (dbr_type_is_LONG(pDB->get_type))
		    sprintf(textvalue, " %12ld", *(long *)pDB->var);
		else if (dbr_type_is_CHAR(pDB->get_type))
		    sprintf(textvalue, " %12d", *(unsigned char *)pDB->var);
		else if (dbr_type_is_ENUM(pDB->get_type)) {
 		    sprintf(textvalue, " %12d", *(short *)pDB->var);
		}
		else if (dbr_type_is_FLOAT(pDB->get_type)) {
		    sprintf(textvalue, " %12.5g", *(float *)pDB->var);
		}
		else if (dbr_type_is_DOUBLE(pDB->get_type)) {
		    sprintf(textvalue, " %12.5g", *(double *)pDB->var);
		}
		if (pDB->count > 1) {
		    strcat (textvalue, " ...");
		}
		strncpy (p->Value, textvalue, sizeof(p->Value));
		p->Monitored = pDB->monitored;
		p->Connected = pDB->connected;
		p->Stat = pDB->status;
		p->Severity = pDB->severity;
		p++;
	    }
	    sendMessage (REPCHANS_REPLY_TYPE, repchansReply,
				sizeof (REPCHANS_REPLY) * pSP->nchan, pSP->nchan, fd);	
	    free (repchansReply);
	    return;
	}
    }
    sendMessage (REPCHANS_REPLY_TYPE, NULL, 0, 0, fd);

}

/*----------*/
static void processReportVariables (fd)

    int	fd;

{
    REPVARS_MSG		*repvarsMsg;
    REPVARS_REPLY	*repvarsReply, *p;
    int			tid, nsym;
    SPROG		*pSP;
    SymTabItem		*pSYM;
    int			msgsize;
    int			Length;

    repvarsMsg = (REPVARS_MSG *)msgBody;

    if (DebugOn)
	printf ("Received RepVars request: Program Name = %s\n", repvarsMsg->ProgramName);

    /* Find all the variables for this program */

    if (FindProgram (repvarsMsg->ProgramName, repvarsMsg->TaskId) != -1) {
	tid = repvarsMsg->TaskId; /* taskIdFigure (repvarsMsg->ProgramName); */
	if ((pSP = seqQryFind (tid)) != NULL) {
	    pSYM = pSP->sym_table;
	    for (nsym=0, msgsize = 0 ; nsym < pSP->num_symbols ; nsym++, pSYM++) {
		msgsize += (strlen (pSYM->Name) + 1);
		msgsize += sizeof (pSYM->Type);
		msgsize += sizeof (pSYM->Size);
		msgsize += sizeof (pSYM->NumElements);
		msgsize += sizeof (pSYM->Offset);
		if (pSYM->Type == V_STRING)
		     msgsize += pSYM->Size;
		else if (pSYM->NumElements == 0)	/* scalar */
		    msgsize += pSYM->Size;
		else
		    msgsize += pSYM->Size * pSYM->NumElements;
	    }
	    repvarsReply = p = (REPVARS_REPLY *) calloc (1, msgsize);
	    pSYM = pSP->sym_table;
	    for (nsym=0 ; nsym < pSP->num_symbols ; nsym++, pSYM++) {
		strcpy (p->Name, pSYM->Name);
		memcpy (&p->Type, &pSYM->Type, sizeof (pSYM->Type));
		memcpy (&p->Size, &pSYM->Size, sizeof (pSYM->Size));
		memcpy (&p->NumElements, &pSYM->NumElements, sizeof (pSYM->NumElements));
		memcpy (&p->Offset, &pSYM->Offset, sizeof (pSYM->Offset));
		if (pSYM->Type == V_STRING)
		     Length = pSYM->Size;
		else if (pSYM->NumElements == 0)	/* scalar */
		    Length = pSYM->Size;
		else
		    Length = pSYM->Size * pSYM->NumElements;
		memcpy (p->Name + strlen(p->Name)+1, pSP->user_area + pSYM->Offset, Length);
		(char *)p += ((sizeof(REPVARS_REPLY)-sizeof(VARIABLE_NAME)) + (strlen (pSYM->Name) + 1) + Length);
	    }
	    sendMessage (REPVARS_REPLY_TYPE, repvarsReply,
				msgsize, pSP->num_symbols, fd);	
	    free (repvarsReply);
	    return;
	}
    }

    sendMessage (REPVARS_REPLY_TYPE, NULL, 0, 0, fd);

}

/*----------*/
static void processReportVariable (fd)

    int		fd;

{

    REPVAR_MSG		*repvarMsg;
    int			tid, nsym, Length;
    SPROG		*pSP;
    SymTabItem		*pSYM;

    repvarMsg = (REPVAR_MSG *)msgBody;

    if (DebugOn)
	printf ("Received RepVar request: Program Name = %s, Variable = %s\n",
		repvarMsg->ProgramName, repvarMsg->VariableName);

    /* Retrieve the current value for a variable */

    if (FindProgram (repvarMsg->ProgramName, repvarMsg->TaskId) != -1) {
	tid = repvarMsg->TaskId; /* taskIdFigure (repvarMsg->ProgramName); */
	if ((pSP = seqQryFind (tid)) != NULL) {
	    pSYM = pSP->sym_table;
	    for (nsym=0 ; nsym < pSP->num_symbols ; nsym++, pSYM++) {
		if (strcmp (repvarMsg->VariableName, pSYM->Name) == 0) {
		    if (pSYM->Type == V_STRING)
			Length = pSYM->Size;
		    else if (pSYM->NumElements == 0)	/* scalar */
			Length = pSYM->Size;
		    else
			Length = pSYM->Size * pSYM->NumElements;
		    sendMessage (REPVAR_REPLY_TYPE, pSP->user_area + pSYM->Offset, Length, Length, fd);
		}
	    }
	    return;
	}
    }

    sendMessage (REPVAR_REPLY_TYPE, NULL, 0, 0, fd);

}

/*----------*/
static void processChangeVariable (fd)

    int		fd;

{
    int			activity;
    CHVAR_MSG		*chvarMsg;
    int			tid, nsym, Length;
    SPROG		*pSP;
    SymTabItem		*pSYM;

    chvarMsg = (CHVAR_MSG *)msgBody;

    if (DebugOn)
	printf ("Received ChVar request: Program Name = %s, Variable = %s\n",
		chvarMsg->ProgramName, chvarMsg->VariableName);

    /* Change the current value for a variable */

    if (FindProgram (chvarMsg->ProgramName, chvarMsg->TaskId) != -1) {
	tid = chvarMsg->TaskId; /* taskIdFigure (chvarMsg->ProgramName); */
	if ((pSP = seqQryFind (tid)) != NULL) {
	    pSYM = pSP->sym_table;
	    for (nsym=0 ; nsym < pSP->num_symbols ; nsym++, pSYM++) {
		if (strcmp (chvarMsg->VariableName, pSYM->Name) == 0) {
		    if (pSYM->Type == V_STRING)
			Length = pSYM->Size;
		    else if (pSYM->NumElements == 0)	/* scalar */
			Length = pSYM->Size;
		    else
			Length = pSYM->Size * pSYM->NumElements;
		/* here is where it gets modified, find out state of this program,
		** return an error (use sendMessage) if secondary ->
		*/
		    if (pSP->seq_token
		      && (resSeqStatus (pSP->seq_token, &activity) != RES_MSTR)) {
			sendMessage (CHVAR_REPLY_TYPE, NULL, 0, 0, fd);
		    }
		    else {
			memcpy (pSP->user_area + pSYM->Offset,
				chvarMsg + 1,
				Length);
			/* If this symbol is associated with a channel, */
			/* do a pvPut to update the world. */
			if (pSYM->ChanIndex != -1) {
			    ca_import (tid);
			    seq_pvPut (pSP, 0, &pSP->channels[pSYM->ChanIndex]);
			}
			seq_sendState (pSP);
			sendMessage (CHVAR_REPLY_TYPE, NULL, 0, 1, fd);
		    }
		    break;
		}
	    }
	    return;
	}
    }

    sendMessage (CHVAR_REPLY_TYPE, NULL, 0, 0, fd);

}

/*----------*/
static void processSuspend (fd)

    int		fd;

{
    SUSPEND_MSG	*suspendMsg;
    int		tid, nss;
    SPROG	*pSP;
    SSCB	*pSS;

    suspendMsg = (SUSPEND_MSG *)msgBody;


    if (DebugOn)
	printf ("Received Suspend request: ProgName=%s, TaskName=%s\n",
			suspendMsg->ProgramName, suspendMsg->TaskName);

    /* Find the taskid for the requested */

    if (FindProgram (suspendMsg->ProgramName, suspendMsg->TaskId) != -1) {
	tid = suspendMsg->TaskId; /* taskIdFigure (suspendMsg->ProgramName); */
	if ((pSP = seqQryFind (tid)) != NULL) {
	    pSS = pSP->sscb;
	    for (nss=0 ; nss < pSP->nss ; nss++, pSS++) {
		if (pSS->task_id && strcmp (suspendMsg->TaskName, pSS->name) == 0)
		    taskSuspend (pSS->task_id);
	    }
	    sendMessage (SUSPEND_REPLY_TYPE, NULL, 0, 1, fd);	
	    return;
	}
    }
    sendMessage (SUSPEND_REPLY_TYPE, NULL, 0, 0, fd);
}

/*----------*/
static void processResume (fd)

    int		fd;

{
    RESUME_MSG	*resumeMsg;
    int		tid, nss;
    SPROG	*pSP;
    SSCB	*pSS;

    resumeMsg = (RESUME_MSG *)msgBody;


    if (DebugOn)
	printf ("Received Resume request: ProgName=%s, TaskName=%s\n",
			resumeMsg->ProgramName, resumeMsg->TaskName);

    /* Find the taskid for the requested */

    if (FindProgram (resumeMsg->ProgramName, resumeMsg->TaskId) != -1) {
	tid = resumeMsg->TaskId; /* taskIdFigure (resumeMsg->ProgramName); */
	if ((pSP = seqQryFind (tid)) != NULL) {
	    pSS = pSP->sscb;
	    for (nss=0 ; nss < pSP->nss ; nss++, pSS++) {
		if (pSS->task_id && strcmp (resumeMsg->TaskName, pSS->name) == 0)
		    taskResume (pSS->task_id);
	    }
	    sendMessage (RESUME_REPLY_TYPE, NULL, 0, 1, fd);	
	    return;
	}
    }
    sendMessage (RESUME_REPLY_TYPE, NULL, 0, 0, fd);
}

/*----------------------- Module List Utilities ----------------------*/

/*----------*/
static BOOL AddModule (Name, Location)

    char	*Name;
    int		Location;

{

    if (NumModules == MAX_MODULES)
	return FALSE;

    strcpy (ModuleList [NumModules].Name, Name);
    ModuleList [NumModules].Location = Location;
    NumModules++;

    return TRUE;
}

/*----------*/
static BOOL RemoveModule (Name)

    char	*Name;

{

    int		i;

    for (i=0 ; i<NumModules ; i++) {
	if (strcmp (Name, ModuleList [i].Name) == 0) {
	    i++;
	    while (i < NumModules) {
		ModuleList [i-1] = ModuleList [i];
		i++;
	    }
	    NumModules--;
	    return TRUE;
	}
    }
    return FALSE;
}

/*----------*/
static int FindModule (Name)

    char	*Name;

{
    int		i;

    for (i=0 ; i<NumModules ; i++)
	if (strcmp (Name, ModuleList [i].Name) == 0)
	    return i;
    return -1;
}

/*----------*/
static BOOL ModTableFull ()
{
    return NumModules == MAX_MODULES;
}

/*----------------------- Program List Utilities ----------------------*/

/*----------*/
static BOOL AddProgram (OrigName, ActualName, Parameters, Resource, TaskId)

    char	*OrigName, *ActualName, *Parameters, *Resource;
    int		TaskId;

{
    if (DebugOn)
	printf ("Adding program: ProgName=/%s/, Parameters=/%s/\n",
		OrigName, Parameters);

    if (NumPrograms == MAX_PROGRAMS)
	return FALSE;

    strcpy (ProgramList [NumPrograms].OrigName, OrigName);
    strcpy (ProgramList [NumPrograms].ActualName, ActualName);
    strcpy (ProgramList [NumPrograms].Parameters, Parameters);
    strcpy (ProgramList [NumPrograms].Resource, Resource);
    ProgramList [NumPrograms].TaskId = TaskId;
    NumPrograms++;

    return TRUE;
}

/*----------*/
static BOOL RemoveProgram (Name, TaskId, Parameters)

    char	*Name, *Parameters;
    int		TaskId;
{

    int		i;

    if (DebugOn)
	printf ("Removing program: ProgName=/%s/, Parameters=/%s/\n",
		Name, Parameters);

    for (i=0 ; i<NumPrograms ; i++) {
	if ((strcmp (Name, ProgramList [i].ActualName) == 0) &&
	    (ProgramList [i].TaskId == TaskId) &&
	    (strcmp (Parameters, ProgramList [i].Parameters) == 0)) {
	    i++;
	    while (i < NumPrograms) {
		ProgramList [i-1] = ProgramList [i];
		i++;
	    }
	    NumPrograms--;
	    return TRUE;
	}
    }
    return FALSE;
}

/*----------*/
static int FindProgram (Name, TaskId)

    char	*Name;
    int		TaskId;
{
    int		i;

    for (i=0 ; i<NumPrograms ; i++) {
	if ((strcmp (Name, ProgramList [i].ActualName) == 0) &&
	    (ProgramList [i].TaskId == TaskId)) {
	    return i;
	}
    }
    return -1;
}

/*----------*/
static BOOL ProgTableFull ()
{
    return NumPrograms == MAX_PROGRAMS;
}

