/*seq_prog.c,v 1.1.1.1 1999/08/03 14:44:37 tis4000 Exp */

/**************************************************************************
			GTA PROJECT   AT division
	Copyright, 1991, The Regents of the University of California.
		         Los Alamos National Laboratory

	@(#)seq_prog.c	1.2	1/29/92
	DESCRIPTION: Seq_prog.c: state program list management functions.
	All active state programs are inserted into the list when created
	and removed from the list when deleted.

	ENVIRONMENT: VxWorks

	HISTORY:
09dec91,ajk	original	***************************************************************************/


#include <vxWorks.h>
#include <ioLib.h>
#include <semLib.h>
#include <taskLib.h>
#include <lstLib.h>
#include <sysseq.h>

#define LOCAL static

LOCAL SIDNT	seqProgListSemId;
LOCAL int	seqProgListInited = FALSE;
LOCAL LIST	seqProgList;
LOCAL void	seqProgListInit();

typedef struct prog_node
{
	NODE		node;
	SPROG		*pSP;
} PROG_NODE;

#define	seqListFirst(pList)	(PROG_NODE *)lstFirst((LIST *)pList)

#define	seqListNext(pNode)	(PROG_NODE *)lstNext((NODE *)pNode)

/*
 * seqFindProg() - find a program in the state program list from task id.
 */
SPROG *seqFindProg(task_id)
int		task_id;
{
	PROG_NODE	*pNode;
	SPROG		*pSP;
	SSCB		*pSS;
	int		n;

	if (!seqProgListInited || task_id == 0)
		return NULL;

	for (pNode = seqListFirst(&seqProgList); pNode != NULL;
	     pNode = seqListNext(pNode) )
	{
		pSP = pNode->pSP;
		if (pSP->task_id == task_id)
			return pSP;
		pSS = pSP->sscb;
		for (n = 0; n < pSP->nss; n++, pSS++)
		{
			if (pSS->task_id == task_id)
				return pSP;
		}
	}

	return NULL; /* not in list */
}

/*
 * seqTraverseProg() - travers programs in the state program list and
 * call the specified routine or function.  Passes one parameter of
 * pointer size.
 */
STATUS seqTraverseProg(pFunc, param)
VOID		(*pFunc)();	/* function to call */
VOID		*param;		/* any parameter */
{
	PROG_NODE	*pNode;
	SPROG		*pSP;

	if (!seqProgListInited)
		return ERROR;

	for (pNode = seqListFirst(&seqProgList); pNode != NULL;
	     pNode = seqListNext(pNode) )
	{
		pSP = pNode->pSP;
		pFunc(pSP, param);
	}

	return OK;
}

/*
 * seqAddProg() - add a program to the state program list.
 * Returns ERROR if program is already in list, else TRUE.
 */
STATUS seqAddProg(pSP)
SPROG		*pSP;
{
	PROG_NODE	*pNode;

	if (!seqProgListInited)
		seqProgListInit(); /* Initialize list */

	for (pNode = seqListFirst(&seqProgList); pNode != NULL;
	     pNode = seqListNext(pNode) )
	{

		if (pSP == pNode->pSP)
			return ERROR; /* already in list */
	}

	/* Insert at head of list */
	pNode = (PROG_NODE *)malloc(sizeof(PROG_NODE) );
	if (pNode == NULL)
		return ERROR;

	pNode->pSP = pSP;
	lstAdd(&seqProgList, (NODE *)pNode);

	return OK;
}

/* 
 *seqDelProg() - delete a program from the program list.
 * Returns TRUE if deleted, else FALSE.
 */
STATUS seqDelProg(pSP)
SPROG		*pSP;
{
	PROG_NODE	*pNode;

	if (!seqProgListInited)
		return ERROR;

	for (pNode = seqListFirst(&seqProgList); pNode != NULL;
	     pNode = seqListNext(pNode) )
	{
		if (pNode->pSP == pSP)
		{
			lstDelete(&seqProgList, (NODE *)pNode);
			return OK;
		}
	}	

	return ERROR; /* not in list */
}

/*
 * seqProgListInit() - initialize the state program list.
 */
LOCAL void seqProgListInit()
{
	/* Init linked list */
	lstInit(&seqProgList);

	/* Create a semaphore for mutual exclusion */
	seqProgListSemId = semBCreate(SEM_Q_FIFO, 0);

	seqProgListInited = TRUE;

}

