/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include <fcntl.h>

#if !(defined(MIPSEL) || defined(AXP))
#include "memory.h"
#else
#include <sys/types.h>
#include <sys/stat.h>
extern int lstat ();
extern int readlink ();
#endif

#include "client.h"
#include "dbDefs.h"
#include "OpPrivData.h"
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32
#include "OperPriv.h"

static CLIENT *client;		/* the client control structure */

static char privServer [132] = { ":OpPrivData" };

static	char	CurrSystem[80];		/* Name of this system */
static	char	CurrOper[OPERATOR_SZ+1];	/* Name of operator who's modifying ChanName */
static	int	CurrErrVal;		/* Last error value for priv/track */
static char *( ErrorMsg[] ) = { 0, 0, 0, 0, 0, 0, 0 };

static ReqUnion	    rqmsg;
static int	    useRemote;	/* Forces direct use privServer instead of local passThru */

void SetRemotePrivServer (void)
{
    useRemote = TRUE;
}

void SetPrivServer (char *servername)
{
    char	privHostPath [100];
    FILE	*f;

    strcpy (privServer, servername);
    strcat (privServer, ":OpPrivData");
    sprintf (privHostPath, "/tis4000/operators/logins/privhost");
    f = fopen (privHostPath, "w");
    if (f != NULL) {
	fputs (privServer, f);
	fputs ("\n", f);
	fclose (f);
    }
}


char *GetPrivServer (char *servername, int namelen)
{
    char	privHostPath [100];
    FILE	*f;

    /* read privelege host from /tis4000/operators/logins/privhost */
    sprintf (privHostPath, "/tis4000/operators/logins/privhost");
    f = fopen (privHostPath, "r");
    if (f != NULL)
    {
	fgets (privServer, sizeof (privServer), f);
	if (strchr (privServer, '\n'))
	    *strchr (privServer, '\n') = '\0';
	fclose (f);
    }
    if (strchr (privServer, ':'))
	*strchr (privServer, ':') = 0;
    strcat (privServer, ":OpPrivData");
    if (servername)
    {
	strncpy (servername, privServer, namelen);
	if (strchr (servername, ':'))
	    *strchr (servername, ':') = 0;
    }
    return privServer;
}

void GetOperName (char *OperName)
{
    char	loginFilePath [100];
    FILE	*f;

    /* read operator name from /tis4000/operators/logins/localoperator */
    sprintf (loginFilePath, "/tis4000/operators/logins/localoperator");
    f = fopen (loginFilePath, "r");
    if (f != NULL) {
	fgets (CurrOper, sizeof (CurrOper), f);
	CurrOper [strlen (CurrOper) - 1] = '\0';
	fclose (f);
    } else {	/* No login file, report as "nooperator" */
	strcpy (CurrOper, "nooperator");
    }

    /* Return the name to the caller */
 
    if (OperName)
	strcpy (OperName, CurrOper);
}


static CLIENT *OpenPrivServer ()
{
    /* read privelege host from /tis4000/operators/logins/privhost */
    GetPrivServer (NULL, 0); /* this initializes privServer name string */
#ifndef DEBUG
    if (useRemote) /* only the tisPassword privServer should set this! */
	return OpenServer (privServer, 1, "");
    else
	return OpenServer (privServer, 1, "APP_CONTROL");
#else
    return OpenServer (privServer, 1, "DEBUG");
#endif DEBUG
}


int InitOperPriv (char *OperName)
{
    LogonMsg *logmsg;
    LogonRsp *answer;
    int	    result;

#ifdef WIN32
    int	opnamesize = OPERATOR_SZ+1;
#endif WIN32

    /* open trend data server if necessary*/
    if (!client)
	client = OpenPrivServer ();
    if (!client)
	return PRIVNOCONN;

    /* Set local logon name info, confirm logged on with server */
    GetOperName (OperName);
    logmsg = (LogonMsg *)&rqmsg;
    logmsg->type = htons (OP_LOGCHECK);
    gethostname (CurrSystem, 80);
    strncpy (logmsg->sysname, CurrSystem, sizeof(logmsg->sysname));
    strncpy (logmsg->sysname + strlen(CurrSystem) + 1,
	     CurrOper,
	     sizeof(logmsg->opname));
    strncpy (logmsg->sysname + strlen(CurrOper)
	      + strlen(CurrOper) + 2,
	     "",
	     sizeof(logmsg->pass));
    answer = (LogonRsp *)SendReq(client,
				 logmsg,
				 (short)(sizeof(logmsg->type)
				  + strlen(CurrSystem)
				  + strlen(CurrOper) + 3));
    if (answer == NULL)
    {
	if (OperName)
	    strcpy (OperName, "nooperator");
	return PRIVNOCONN;
    }
    result = (signed short)ntohs(answer->value);
    free (answer);
    if (result == PRIVOK)
    {
	if (OperName)
	    strcpy (OperName, CurrOper);
    }
    else
    {
	if (OperName)
	    strcpy (OperName, "nooperator");
    }
    return result;
}


int InitOperTracking (char *OperName)
{
    LogonMsg *logmsg;
    LogonRsp *answer;
    int	    result;

#ifdef WIN32
    int	opnamesize = OPERATOR_SZ+1;
#endif WIN32

    /* open trend data server if necessary*/
    if (!client)
	client = OpenPrivServer ();
    if (!client)
	return PRIVNOCONN;

    /* Set local logon name info, confirm logged on with server */
    GetOperName (OperName);
    logmsg = (LogonMsg *)&rqmsg;
    logmsg->type = htons (OP_LOGCHECK);
    gethostname (CurrSystem, 80);
    strncpy (logmsg->sysname, CurrSystem, sizeof(logmsg->sysname));
    strncpy (logmsg->sysname + strlen(CurrSystem) + 1,
	     CurrOper,
	     sizeof(logmsg->opname));
    strncpy (logmsg->sysname + strlen(CurrSystem)
	      + strlen(CurrOper) + 2,
	     "",
	     sizeof(logmsg->pass));
    answer = (LogonRsp *)SendReq(client,
				 logmsg,
				 (short)(sizeof(logmsg->type)
				  + strlen(CurrSystem)
				  + strlen(CurrOper) + 3));
    if (answer == NULL)
    {
	strcpy (OperName, "nooperator");
	return PRIVNOCONN;
    }
    result = (signed short)ntohs(answer->value);
    free (answer);
    if (result == PRIVOK)
    {
	if (OperName)
	    strcpy (OperName, CurrOper);
    }
    else
    {
	if (OperName)
	    strcpy (OperName, "nooperator");
    }
    return result;
}


int OperLogon (char *OperName, char *PassWord)
{
    LogonRsp *answer;
    int	    result;
    char	loginFilePath [100];
    FILE	*f;

    /* open trend data server if necessary */
    if (!client)
	client = OpenPrivServer ();
    if (!client)
	return PRIVNOCONN;

    /* Set local logon name info, confirm logged on with server */
    strncpy (CurrOper, OperName, OPERATOR_SZ+1);
    if (!strlen (OperName))
    {
	return PRIVNOCONN;
    }
    else
    {
	LogonMsg *logmsg = (LogonMsg *)&rqmsg;
	logmsg->type = htons (OP_LOGON);
	gethostname (CurrSystem, 80);
	strncpy (logmsg->sysname, CurrSystem, sizeof(logmsg->sysname));
	strncpy (logmsg->sysname + strlen(CurrSystem) + 1,
		 CurrOper,
		 sizeof(logmsg->opname));
	strncpy (logmsg->sysname + strlen(CurrSystem)
		  + strlen(CurrOper) + 2,
		 PassWord,
		 sizeof(logmsg->pass));
	answer = (LogonRsp *)SendReq(client,
				     logmsg,
				     (short)(sizeof(logmsg->type)
				      + strlen(CurrSystem)
				      + strlen(CurrOper)
				      + strlen(PassWord) + 3));
	if (answer == NULL)
	{
	    return PRIVNOCONN;
	}
	result = (signed short)ntohs(answer->value);
	free (answer);
	if (result == PRIVOK)
	{
	    /* write operator name to /tis4000/operators/logins/localoperator */
	    sprintf (loginFilePath, "/tis4000/operators/logins/localoperator");
	    f = fopen (loginFilePath, "w");
	    if (f != NULL)
	    {
		fputs (CurrOper, f);
		fputs ("\n", f);
		fclose (f);
	    }
	    if (OperName)
		strcpy (OperName, CurrOper);
	}
	else
	{
	    if (OperName)
		strcpy (OperName, "nooperator");
	}
    }
    return result;
}


int OperLogoff ()
{
    LogonMsg *logmsg;
    LogonRsp *answer;
    int	    result;
    char	loginFilePath [100];

    /* open trend data server */
    if (!client)
	return PRIVNOCONN;

    logmsg = (LogonMsg *)&rqmsg;
    logmsg->type = htons (OP_LOGOFF);
    gethostname (CurrSystem, 80);
    strncpy (logmsg->sysname, CurrSystem, sizeof(logmsg->sysname));
    strncpy (logmsg->sysname + strlen(CurrSystem) + 1,
	     CurrOper,
	     sizeof(logmsg->opname));
    strncpy (logmsg->sysname + strlen(CurrSystem)
	      + strlen(CurrOper) + 2,
	     "",
	     sizeof(logmsg->pass));
    answer = (LogonRsp *)SendReq(client,
				 logmsg,
				 (short)(sizeof(logmsg->type)
				  + strlen(CurrSystem)
				  + strlen(CurrOper)
				  + 3));
    if (answer == NULL)
    {
	return PRIVNOCONN;
    }
    result = (signed short)ntohs(answer->value);
    free (answer);
    if (result == PRIVOK)
    {
	sprintf (loginFilePath, "/tis4000/operators/logins/localoperator");
#ifdef WIN32
	_unlink (loginFilePath);
#else
	unlink (loginFilePath);
#endif WIN32
	strcpy (CurrOper, "nooperator");
    }
    return result;
}


int CheckOperPriv (char *Channel)
{
    PrivRsp *answer;
    int	    result;

    PrivMsg *priv = (PrivMsg *)&rqmsg;

    /* open trend data server if necessary*/
    if (!client)
	InitOperPriv (NULL);
    if (!client)
    {
	CurrErrVal = PRIVNOCODE;
	if (ErrorMsg[0] == NULL)
	    ErrorMsg[0] = calloc (1, 100);
	if (ErrorMsg[0])
	    strcpy (ErrorMsg[0], "No response from OpPriv server.");
	if (ErrorMsg[1])
	    free (ErrorMsg[1]);
	ErrorMsg[1] = NULL;
	return PRIVNOCONN;
    }

    priv->type = htons (OP_CHECK_PRIV);
    strncpy (priv->priv, Channel, sizeof(priv->priv));
    answer = (PrivRsp *)SendReq(client, priv, sizeof(*priv));
    if (answer == NULL)
    {
	CurrErrVal = PRIVNOCODE;
	if (ErrorMsg[0] == NULL)
	    ErrorMsg[0] = calloc (1, 100);
	if (ErrorMsg[0])
	    strcpy (ErrorMsg[0], "No response from OpPriv server.");
	if (ErrorMsg[1])
	    free (ErrorMsg[1]);
	ErrorMsg[1] = NULL;
	return PRIVNOCONN;
    }
    result = (signed short)ntohs(answer->value);
    CurrErrVal = result;
    if (result != PRIVOK)
    {
	char *errp = answer->errString;
	int i = 0;
	while (*errp)
	{
	    if (ErrorMsg[i] == NULL)
		ErrorMsg[i] = calloc (1, 100);
	    if (ErrorMsg[i])
		strcpy (ErrorMsg[i], errp);
	    i++;
	    errp += strlen(errp)+1;
	}
	if (ErrorMsg[i])
	    free (ErrorMsg[i]);
	ErrorMsg[i] = NULL;
    }
    else
    {
	if (ErrorMsg[0] == NULL)
	    ErrorMsg[0] = calloc (1, 100);
	if (ErrorMsg[0])
	    strcpy (ErrorMsg[0], "No error, last status PRIVOK");
	if (ErrorMsg[1])
	    free (ErrorMsg[1]);
	ErrorMsg[1] = NULL;
    }
    free (answer);
    return result;
}


int OperPriv (C4CODE *codebasePtr, char *Channel)
{
    PrivRsp *answer;
    int	    result;

    PrivMsg *priv = (PrivMsg *)&rqmsg;

    /* open trend data server if necessary*/
    if (!client)
	InitOperPriv (NULL);
    if (!client)
	return 0;

    priv->type = htons (OP_PRIV);
    strncpy (priv->priv, Channel, sizeof(priv->priv));
    answer = (PrivRsp *)SendReq(client, priv, sizeof(*priv));
    if (answer == NULL)
    {
	return 0;
    }
    result = (signed short)ntohs(answer->value);
    free (answer);
    return result;
}


int WriteOperTracking (char *Channel, char *oldValue, char *newValue)
{
    TrackRsp *answer;
    int	    result;
    TrackMsg *track = (TrackMsg *)&rqmsg;

    if (!client)
    {
	CurrErrVal = PRIVNOCODE;
	if (ErrorMsg[0] == NULL)
	    ErrorMsg[0] = calloc (1, 100);
	if (ErrorMsg[0])
	    strcpy (ErrorMsg[0], "OpPriv server not open");
	if (ErrorMsg[1])
	    free (ErrorMsg[1]);
	ErrorMsg[1] = NULL;
	return PRIVNOCONN;
    }
    track->type = htons (WRITE_TRACK);
    strncpy (track->channel, Channel, sizeof(track->channel));
    strncpy (track->channel + strlen(Channel) + 1,
	     oldValue,
	     sizeof(track->oldvalue));
    strncpy (track->channel + strlen(Channel)
	      + strlen(oldValue) + 2,
	     newValue,
	     sizeof(track->newvalue));
    answer = (TrackRsp *)SendReq(client,
				 track,
				 (short)(sizeof(track->type)
				  + strlen(Channel)
				  + strlen(oldValue)
				  + strlen(newValue)
				  + 3));
    if (answer == NULL)
    {
	CurrErrVal = PRIVNOCODE;
	if (ErrorMsg[0] == NULL)
	    ErrorMsg[0] = calloc (1, 100);
	if (ErrorMsg[0])
	    strcpy (ErrorMsg[0], "No response from OpPriv server.");
	if (ErrorMsg[1])
	    free (ErrorMsg[1]);
	ErrorMsg[1] = NULL;
	return PRIVNOCONN;
    }
    result = (signed short)ntohs(answer->value);
    CurrErrVal = result;
    if (result != PRIVOK)
    {
	char *errp = answer->errString;
	int i = 0;
	while (*errp)
	{
	    if (ErrorMsg[i] == NULL)
		ErrorMsg[i] = calloc (1, 100);
	    if (ErrorMsg[i])
		strcpy (ErrorMsg[i], errp);
	    i++;
	    errp += strlen(errp)+1;
	}
	if (ErrorMsg[i])
	    free (ErrorMsg[i]);
	ErrorMsg[i] = NULL;
    }
    else
    {
	if (ErrorMsg[0] == NULL)
	    ErrorMsg[0] = calloc (1, 100);
	if (ErrorMsg[0])
	    strcpy (ErrorMsg[0], "No error, last status PRIVOK");
	if (ErrorMsg[1])
	    free (ErrorMsg[1]);
	ErrorMsg[1] = NULL;
    }
    free (answer);
    return result;
}


int CloseOperTracking (char *OperName)
{
    OpenCloseTrackRsp *answer;
    int	    result;

    OpenCloseTrackMsg *track = (OpenCloseTrackMsg *)&rqmsg;
    track->type = htons (CLOSE_TRACK);
    strncpy (track->opname, OperName, sizeof(track->opname));
    answer = (OpenCloseTrackRsp *)SendReq(client,
				 track,
				 (short)(sizeof(track->type)
				  + strlen(OperName)
				  + 1));
    if (answer == NULL)
    {
	CurrErrVal = PRIVNOCODE;
	if (ErrorMsg[0] == NULL)
	    ErrorMsg[0] = calloc (1, 100);
	if (ErrorMsg[0])
	    strcpy (ErrorMsg[0], "No response from OpPriv server.");
	if (ErrorMsg[1])
	    free (ErrorMsg[1]);
	ErrorMsg[1] = NULL;
	return PRIVNOCONN;
    }
    result = (signed short)ntohs(answer->value);
    CurrErrVal = result;
    free (answer);
    return result;
}


int CloseOperPriv (void)
{
    OpenClosePrivRsp *answer;
    int	    result;

    OpenClosePrivMsg *priv = (OpenClosePrivMsg *)&rqmsg;
    priv->type = htons (CLOSE_PRIV);
    answer = (OpenClosePrivRsp *)SendReq(client,
				 priv,
				 (short)(sizeof(priv->type)));
    if (answer == NULL)
    {
	CurrErrVal = PRIVNOCODE;
	if (ErrorMsg[0] == NULL)
	    ErrorMsg[0] = calloc (1, 100);
	if (ErrorMsg[0])
	    strcpy (ErrorMsg[0], "No response from OpPriv server.");
	if (ErrorMsg[1])
	    free (ErrorMsg[1]);
	ErrorMsg[1] = NULL;
	return PRIVNOCONN;
    }
    result = (signed short)ntohs(answer->value);
    CurrErrVal = result;
    free (answer);
    return result;
}


int	OperPrivTrackErr (int Code, char ***Message)
{
    if (Code != CurrErrVal)
    {
	if (ErrorMsg[0] == NULL)
	    ErrorMsg[0] = calloc (1, 100);
	if (ErrorMsg[0])
	    strcpy (ErrorMsg[0], "Error value does not match last OpPriv call");
	if (ErrorMsg[1])
	    free (ErrorMsg[1]);
	ErrorMsg[1] = NULL;
    }
    *Message = ErrorMsg;
    return PRIVOK;
}

