
/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/************************************************************************/
/*                                                                      */
/*            IOC TREND UPDATE TASK					*/
/*                                                                      */
/*    This task exchanges status with the IOCs sharing responsibility   */
/*    for this IOC's resources.                                         */
/*                                                                      */
/*    Summary of execution:                                             */
/*    1. Initialization - Create synchronization semaphore, mutual	*/
/*	  exclusion semaphore, and queue area. Start the trend_update	*/
/*	  task. This task then waits on synchronization semaphore for	*/
/*	  indication of queued trend info.				*/
/*    2. Data transfer - Get queued information, and check connection	*/
/*	  to trenddata server at workstation (open if necessary). Send	*/
/*	  information to server, if transaction response is good, set	*/
/*	  bit specified in queued information.				*/
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/*            INCLUDE FILES                                             */
/*                                                                      */
/************************************************************************/

#include <vxWorks.h>
#include <socket.h>
#include <in.h>
#include <errno.h>
#include <task_params.h>        /* For getting task priorities */
#include <systime.h>
#include <bootLib.h>
#include <lstLib.h>
#include "ioc_trend_update.h"
#include "client.h"
#include "trendSizes.h"
#include "trendupdatemsgs.h"

/************************************************************************/
/*                                                                      */
/*            GLOBAL DATA WE DEFINE                                     */
/*                                                                      */
/************************************************************************/

int trendUpdateTask ();

/************************************************************************/
/*                                                                      */
/*            GLOBAL DATA WE REFERENCE                                  */
/*                                                                      */
/************************************************************************/

/* following routine in anyHandlers.c */
extern int anyTrendHandlers ();

unsigned char iocName [IOCNAME_SIZE]; /* ioc name */

#define MAX_TRENDSERVERS	2
unsigned char serverName [100];	/* server IP addresses(separated by ','):NAME */

#define TRUPDSERVICE	"TrendUpdate"
#define TRUPD_NAME	"plcTrend"
#define TRUPD_PRI	99
#define TRUPD_STACK	4000

CLIENT	*client;		/* server connection (NULL if closed) */

SEM_ID freeMux;			/* queue access control */
LIST freeHead;			/* list of available data chunks */

SEM_ID	qMux;			/* queue access control */
SEM_ID	syncSem;		/* stuff has shown up on the queue */
LIST qHead;			/* list of waiting data chunks */

/************************************************************************/
/*                                                                      */
/*            TUNING PARAMETERS                                         */
/*                                                                      */
/************************************************************************/
#define	TICKSPERSEC	60	/* number of ticks per second on vxWorks */
#define	TICKSPERMIN	(60*TICKSPERSEC)

#define	MAX_TIMEOUT	(1<<30) /* like 2 years */

/************************************************************************/
/*                                                                      */
/*            OUR PRIVATE DATA                                          */
/*                                                                      */
/************************************************************************/

static int		selectcalls;
static int		msgsent;
static int		msgrecv;

int TRUPDDEBUG;

UpdateMsgUnion updateMsg;
UpdateRspUnion updateRsp;

/* for local consumption only */
int trendUpdateTask();

static QENTRY *GetTrendUpdate ( );

static QENTRY *QueueFreeUpdate (QENTRY *current);

/****************************************************************/
/*                                                              */
/*            TREND INITIALIZATION ROUTINE			*/
/*                                                              */
/*    Initialize and start the update processing task.		*/
/*								*/
/****************************************************************/

int StartTrendUpdate (char *IOCName, char *ip1, char *ip2)
{
    TRUPDDEBUG = 0; /* start with debug off */

    /* record IOC name */
    strncpy (iocName, IOCName, IOCNAME_SIZE);

    /* Build service name from args */
    if (ip1)
	strcpy (serverName, ip1);
    else
	strcpy (serverName, "*");
    if (ip2)
    {
	strcat (serverName, ",");
	strcat (serverName, ip2);
    }
    strcat (serverName, ":");
    strcat (serverName, TRUPDSERVICE);
    logMsg ("Starting TrendUpdate with service host: %s\n", serverName);
    
    /*
    Initialize queue and semaphores
    */
    syncSem = semCCreate (SEM_Q_FIFO, 0);

    freeMux = semMCreate (SEM_Q_PRIORITY|SEM_INVERSION_SAFE|SEM_DELETE_SAFE);
    qMux = semMCreate (SEM_Q_PRIORITY|SEM_INVERSION_SAFE|SEM_DELETE_SAFE);

    lstInit (&freeHead);
    lstInit (&qHead);

    /*
    Spawn the trendUpdate task. For now, use the same
    priority as the periodic scan task.
    */

    if (taskSpawn(TRUPD_NAME, TRUPD_PRI, VX_FP_TASK, TRUPD_STACK,
                   trendUpdateTask) == ERROR)
    {
        logMsg("Can't spawn trendUpdateTask.\n");
        return (-1);
    }

    /*
    Initialization complete
    */
    return(0);
}

int TestTrendQueue ()
{
    static short updData [] ={ 1, 2, 3 };
    static char	stat;

	return QueueTrendUpdate (tickGet(), (short) 1, (short) 1,
		      updData, (short) 3,
		      &stat, 1);
}
/****************************************************************/
/*                                                              */
/*            TREND QUEUE ROUTINE				*/
/*                                                              */
/*    Move data to queue feeding update task.			*/
/*								*/
/****************************************************************/

int QueueTrendUpdate (long updTime, short plcline, short plcaddr,
		      void *updData, short datalen,
		      char *statPtr, char statBit)
{
    QENTRY *current;

    /* see if we have a suitable leftover element */
    semTake(freeMux, WAIT_FOREVER); /* Lock queue */
    current = (QENTRY *)lstGet (&freeHead);
    while (current && current->dataSize < datalen)
    {
	current = (QENTRY *) lstNext (current);
    }
    if (current)
    {
	lstDelete (&freeHead, current);
    }
    else
    {
	/* note - datalen is # of shorts */
	current = calloc (1, sizeof(QENTRY)+(datalen*(sizeof(short))));
	if (current)
	{
	    current->dataSize = datalen;
	}
    }
    semGive(freeMux); /* Unlock queue */
    if (current)
    {
	/* build entry */
	current->statusLoc = statPtr;
	current->statusBit = statBit;
	current->dataLen = datalen;
	current->snapTime = updTime;
	current->line = plcline;
	current->address = plcaddr;
	memcpy (&current->dataPtr, updData, datalen*sizeof(short));

	semTake(qMux, WAIT_FOREVER); /* Lock queue */
	lstAdd (&qHead, current);
	semGive(qMux); /* Unlock queue */

	semGive( syncSem, MAX_TIMEOUT );
	return TRUE;
    }
    /* could not queue data */
    return FALSE;
}

/****************************************************************/
/*                                                              */
/*            TREND ACTIVE ROUTINE				*/
/*                                                              */
/*    Determine if workstation trend registered with ca.	*/
/*								*/
/****************************************************************/

int TrendingActive ()
{
    return anyTrendHandlers () != 0;
}

/****************************************************************/
/*                                                              */
/*            THE TREND UPDATE TASK				*/
/*                                                              */
/*    This is the trend update task.				*/
/*								*/
/****************************************************************/

int trendUpdateTask()
{
    QENTRY	*current;
    int		transferred;
    UpdateMsg	*update = &updateMsg.update;
    UpdateRsp	*response = &updateRsp.updateRsp;
    int		msgLen;
    int		iocNameLen = strlen (iocName) + 1;
    short	*srcPtr;
    short	*dstPtr;
    int		i;

    while (TRUE)
    {
	/* Wait for something to do */
	semTake ( syncSem, MAX_TIMEOUT );

	if (TRUPDDEBUG > 2)
	    logMsg("Trend update task wake up.\n");

	/* get data from queue */
	current = GetTrendUpdate ( );

	/* structure server message */
	msgLen = sizeof (UpdateMsg) - IOCNAME_SIZE - 512;
	update->type = htons (UPDATE_INFO);
	update->dataLen = htons (current->dataLen);
	update->snapTime = htonl (current->snapTime);
	update->line = htons (current->line);
	update->address = htons (current->address);
	for (srcPtr = &current->dataPtr, dstPtr = update->data, i = current->dataLen;
	     i > 0;
	     srcPtr++, dstPtr++, i--)
	{
	    *dstPtr = htons (*srcPtr);
	}
	strcpy (dstPtr, iocName);
	
	msgLen = msgLen + (update->dataLen * 2) + strlen(iocName) + 1;


	/* transfer data to workstation service */
	transferred = 0;
	while (!transferred)
	{
	    while (client == NULL)
	    {
		client = OpenServer (serverName, TRUE, NULL);
		if (client)
		    break;
		if (TRUPDDEBUG > 2)
		    logMsg("Can't open server %s.\n", serverName);
#ifdef vxWorks
		taskDelay (1800);
#else
		sleep (30);
#endif vxWorks
	    }
	    /* perform transaction */
	    response = SendReq (client, update, msgLen);
	    if (response)
	    {
		transferred = response->value;
		free (response);
	    }
	    else
	    {
	        CloseSession (client);
	        client = NULL;
	    }
	}
	/* Mark as complete */
	if (current->statusLoc)
	    *current->statusLoc = *current->statusLoc | current->statusBit;

	/* save used entry on freeQ */
	QueueFreeUpdate (current);
    } /* task forever loop */
}

/****************************************************************/
/*                                                              */
/*            TREND QUEUE SUPPORT ROUTINES			*/
/*                                                              */
/*    Access queues feeding update task, and free/reuse queue.	*/
/*								*/
/****************************************************************/

static QENTRY *GetTrendUpdate ( )
{
    QENTRY *current;

    semTake(qMux, WAIT_FOREVER); /* Lock queue */
    current = (QENTRY *)lstGet (&qHead);
    semGive(qMux); /* Unlock queue */
    return current;
}

static QENTRY *QueueFreeUpdate (QENTRY *current)
{
    semTake(freeMux, WAIT_FOREVER); /* Lock queue */
    lstAdd (&freeHead, current);
    semGive(freeMux); /* Unlock queue */
    return current;
}
