/*
  Support functions for PLC trending

  GCT 23-Oct-1997

*/
#include <stdio.h>
#include <math.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#include "cb45to63.h"
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "plcTrendLib.h"
static FILE *fp = 0;
static char Token[80];
static char token;
/*
  Codebase data
*/
static CODE4 codebase;
/* database file handles */
static DATA4 *aiFile = 0;
static DATA4 *diFile = 0;
static DATA4 *arFile = 0;
/* field handles */
static FIELD4 *aslo = 0;
static FIELD4 *aoff = 0;
static FIELD4 *linr = 0;
static FIELD4 *irlo = 0;
static FIELD4 *irhi = 0;
static FIELD4 *egul = 0;
static FIELD4 *eguf = 0;
static FIELD4 *invt = 0;
static FIELD4 *hopr = 0;
static FIELD4 *lopr = 0;
static FIELD4 *egut = 0;
static FIELD4 *rate = 0;
static int firstTime = 0;

#ifdef TEST_PLCTRENDLIB
void main(int numArgs,char *argv[])
{
  char *pv,pvString[40];
  short offs,mask,i;
  double val;

  pv = pvString;
  i = 1;
  if(StartUpdateGroup(argv[1],(short)atoi(argv[2]),(short)atoi(argv[3])))
    while(NextUpdatePoint(&pv,&offs,&mask))
      {
	val = GetTrendData(pv,5*i++,PCT_ANALOG);
	printf("%s - %d - %d - %f\n",pv,offs,mask,val);
      }
  if(fp) fclose(fp);
  if(aiFile) d4close(aiFile);
  if(diFile) d4close(diFile);
  if(arFile) d4close(arFile);
}
#endif
/*
  Convert raw data from IOC to value for trending. Use the same calculations
  as in the IOC. If pointType is AI(1) then calculate value based on EGUL,EGUH,
  IRLO,IRHI. If it is DI(0) then use INVT.
*/
double GetTrendData(char *pointName,short value,short pointType,double *hoprValue,double *loprValue,char *egutValue)
{
  double asloValue;
  double aoffValue;
  double irloValue;
  double irhiValue;
  double egulValue;
  double egufValue;
  char linrValue[15],i;
  short invtValue;
  /* local */
  double esloValue;
  double roffValue;
  double returnValue;
  /*
    Find the record and get the constant values from the database file
  */
  returnValue = value;				/* assign raw value to return value incase nothing needs to be done */
  *hoprValue = 0.0;				/* default or return for digital inputs */
  *loprValue = 0.0;
  *egutValue = '\0';
  if(pointType == PCT_ANALOG)
    {
      if(d4seek(aiFile,pointName) == 0)
	{
	  asloValue = f4double(aslo);
	  aoffValue = f4double(aoff);
	  irloValue = f4double(irlo);
	  irhiValue = f4double(irhi);
	  egulValue = f4double(egul);
	  egufValue = f4double(eguf);
	  *hoprValue = f4double(hopr);
	  *loprValue = f4double(lopr);
	  strcpy(egutValue,f4str(egut));
	  strcpy(linrValue,f4str(linr));
	  for(i = strlen(linrValue)-1;i >= 0;i--)
	    {
	      if(linrValue[i] == ' ')
		linrValue[i] = 0;
	      else
		break;
	    }
	  /*
	    calculate new slope and offset
	  */
	  if(irhiValue == irloValue)
	    esloValue = 0.0;
	  else
	    esloValue = (egufValue - egulValue)/(irhiValue - irloValue);
	  roffValue = egulValue - (esloValue*irloValue);
	  /* adjust slope and offset */
	  if (asloValue != 0.0) 
	    {
	      returnValue *= asloValue;
	    }
	  if (aoffValue != 0.0) 
	    {
	      returnValue += aoffValue;
	    }
	  /*
	    check for specific conversion types:
	    "NO CONVERSION"
	    "LINEAR"
	    "Square Root"
	  */
	  if(!strcmp(linrValue,"NO CONVERSION"))
	    {
	      ;
	    }
	  else if(!strcmp(linrValue,"LINEAR"))
	    {
	      /*
		 y = mx + b, m = slope = eslo, b = offset = roff 
		 calc'd in slope_and_offset()
	      */
	      returnValue = (esloValue * returnValue) + roffValue;
	    }
	  else if(!strcmp(linrValue,"Square Root"))
	    {
	      double temp_dbl;
	      double sqrt_val;
	      /*
		val = eng units full - eng units low * square root of
		raw aval - instrument range low divided by
		instrument range high - instrument range low
		plus eng units low
	      */
	      if (irhiValue == irloValue)
		sqrt_val = 0.0;
	      else
		{
		  temp_dbl = ( (returnValue - irloValue) / (irhiValue - irloValue) );
		  sqrt_val = sqrt(temp_dbl);
		}
	      returnValue = ((egufValue - egulValue) * sqrt_val) + egulValue;
	    }
	}
    }
  else				/* must be digital */
    {
      if(d4seek(diFile,pointName) == 0)
	{
	  invtValue = (short)f4int(invt);
	  if (value == 0) 
	    returnValue = 0;
	  else 
	    returnValue = 1;
	  /* Handle inversion */
	  if (invtValue)
	    {
	      if (returnValue)
		returnValue = 0;
	      else
		returnValue = 1;
	    }
	}
    }
  return(returnValue);
}
/*
  Open PLCTREND.TXT file, if not already open,and sets file position to read the first point group.
*/
int StartUpdateGroup(char *IOCName,short line,short address)
{
  char *iocName;
  char *num;
  if(fp == 0)
#ifdef WIN32
    if((fp = fopen("\\tis4000db\\plctrend.txt","r")) == 0)
#else
    if((fp = fopen("/tis4000db/plctrend.txt","r")) == 0)
#endif WIN32
      return(0);
  /*
    Only open the files and get the field pointers the first time
  */
  if(firstTime == 0)
    {
      d4init(&codebase);
      aiFile = d4open(&codebase,AI_FILE);
      diFile = d4open(&codebase,DI_FILE);
      arFile = d4open(&codebase,AR_FILE);
      d4tag_select(arFile,d4tag(arFile,AR_METHOD_TAG));

      aslo = d4field(aiFile,ASLO_FIELD);
      aoff = d4field(aiFile,AOFF_FIELD);
      linr = d4field(aiFile,LINR_FIELD);
      irlo = d4field(aiFile,IRLO_FIELD);
      irhi = d4field(aiFile,IRHI_FIELD);
      egul = d4field(aiFile,EGUL_FIELD);
      eguf = d4field(aiFile,EGUF_FIELD);
      hopr = d4field(aiFile,HOPR_FIELD);
      lopr = d4field(aiFile,LOPR_FIELD);
      egut = d4field(aiFile,EGUT_FIELD);
      invt = d4field(diFile,INVT_FIELD);
      rate = d4field(arFile,RATE_FIELD);

      firstTime = 1;
    }
  rewind(fp);						/* start from the begining */
  while((token = fgetc(fp)) != EOF)
    {
      if(skipToIOC(fp) == 0)			/* skip characters until "@" */
	return(0);					/* nothing good until EOF */
      iocName = getToken(fp);	       		/* found "@" so next thing must be IOC name */
      if(strlen(iocName) == 0)
	return(0);					/* iocName is zero length then error */
      if(strcmp(iocName,IOCName))			/* not equal to IOC specified...keep looking */
	continue;
      num = getToken(fp);	       		/* get line value from file */
      if(line != (short)atoi(num))
	continue;					/* not equal to line specified...keep looking */
      num = getToken(fp);		       	/* get line value from file */
      if(address != (short)atoi(num))
	continue;					/* not equal to address specified...keep looking */
      num = getToken(fp);		       	/* skip last two parameters on line*/
      num = getToken(fp);
      nextLine (fp);				/* go to next line */
      return(1);					/* everything matched return OK */
    }
  return(0);						/* EOF so return 0 */
}
/*
  Get the data from the next point group and return a 1 if OK else if next
  line encountered contains a "@" or fgets returns an EOF then return 0
*/
int NextUpdatePoint(char *pointName,short *wordOffset,short *bitOffset)
{
  char *num;
  int sawComment = 0;
  char key[40];
  while(1)
    {
      if(token == '#') sawComment = 1;
      if(token == '\n') sawComment = 0;
      if((token == '@') || (token == EOF))     	/* must be new IOC spec or we're done */
	return(0);
      if((sawComment) || ((sawComment == 0) && (token == '\n')))
	{
	  token = fgetc(fp);
	  continue;
	}
      if((num = getToken(fp)) == 0) break;	/* get point name */
      strcpy(pointName,num);
      if((num = getToken(fp)) == 0) break; 	/* get the word offset */
      *wordOffset = (short)atoi(num);
      if(strlen (num = getToken(fp))) 	/* get bit offset */
        *bitOffset = (short)atoi(num);
      else
        *bitOffset = -1;
      nextLine (fp);
      /* if point is not valid and on trend at 5 min sample, keep searching */
      sprintf(key,"%-28.28sSMP",pointName);
      if(d4seek(arFile,key) != 0)
	{
	  continue;
	}
      else
	{
	  if(strcmp(f4str(rate),"300  "))		/* rate must be 5 minutes (300 seconds) */
	    {
	      continue;
	    }
	}
      return(1);
    }
return(0);
}
int skipToIOC(FILE *fp)		/* Advance file pointer to next @ char */
{
  int sawComment = 0;
  do
    {
      if(token == '#') sawComment = 1;
      if(token == '\n') sawComment = 0;
      if(token == '@')
	{
	  if(sawComment)
	    continue;
	  else
	    {
	      token = fgetc(fp);
	      return(1);
	    }
	}
    }
  while((token = fgetc(fp)) != EOF);
  return(0);
}
char *getToken(FILE *fp)	/* get token string between delimiters */
{
  int sawComment = 0;			/* no comments yet */
  int index = 0;			/* stick new char here */
  Token[0] = 0;				/* no token either */
  do
    {
      if(token == '#') sawComment = 1;
      if((token == EOF) || (token == '\n')) return(Token);
      if(index && ((token == ' ') || (token == ',') || (token == '\t')))
	{
	  token = fgetc(fp);
	  break;
	}
      if(!sawComment)
	Token[index++] = token;
    }
  while(((token = fgetc(fp)) != EOF) && (token != '\n'));
  Token[index] = 0;
  return(Token);
}
int nextLine (FILE *fp)
{
  while(token != '\n' && token != EOF) token = fgetc(fp);
  if(token == '\n') token = fgetc(fp);
}
