
/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/* svcquery.c - service query handlerfor TIS client/server programs.
**	This program is started by inetd, and expects a message on fd 0.
**	The program will read and format the tisservice.conf file into a
**	udp message, and return the data to requesting host:port.
*/

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <sys/types.h>
#ifdef WIN32
#include <winsock.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif WIN32
#include "msgMon.h"

#include "inet_ntoa.h"
#include "tisservice.h"

SvcReq	inmsg;		/* input message information */
SvcResp	msg;		/* output message information */

char	inpstring [512];

int main (argc, argv)
{
    int ix;
    char *sp;
    FILE *fp;
    char *cp;
    int client_len;
    struct sockaddr_in client_addr;
    struct hostent *hp;

    tisOpenLog ("svcquery", 6, 2, NULL, NULL);
    tisSysLog (1, 0, "Service query started");

    /* get message, exit if error or insufficient data */
    client_len = sizeof(client_addr);
    if ((ix = recvfrom(0, (char *)&inmsg, sizeof(inmsg), 0, (struct sockaddr *)&client_addr, &client_len)) < sizeof(inmsg))
    {
	tisSysLog (0, 0, "Error receiving UDP message size %d, %m", ix);
	tisCloseLog ();
	exit (1);
    }

    if ((ntohs(inmsg.type) == SVCREQ) && (ntohl(inmsg.sftlvl) == SOFTWARE_LEVEL))
    {
	/* look up requested service, exit if not found */
	fp = fopen ("/tis4000db/tisservice.conf", "rt");
	if (fp)
	{
            while (!feof (fp))
            {
        	ix = 0;
		while (!feof (fp))
		{
                    inpstring [ix] = fgetc(fp);
                    if ((inpstring [ix] == 0x0d) || (inpstring [ix] == 0x0a) 
		      || (inpstring [ix] == -1))
		    {
                	break;
 		    }
		    ix++;
 		    if (ix == 511)
		    {
			inpstring[511] = 0;
			tisSysLog (0, 0, "bad string in config file %s", inpstring);
			tisCloseLog ();
			exit (1); /* bad string in file. */
		    }
		}
		inpstring [ix] = 0;

		/* process input line */
		sp = inpstring;
		while (*sp && *sp == ' ')
		    sp++;
		if ((*sp == 0) || (*sp == '#'))
		    continue; /* blank line or comment! */
		/* process input- line is formatted as
		** <service_name> <executable> <args-as in inetd.conf>
		*/
		cp = strchr (sp, ' ');
		if (cp)
		{
		    *cp = 0;
		    if (strcmp (sp, inmsg.service) == 0)
		    {
			/* return response message */
			msg.type = htons(SVCRESP);
			msg.sftlvl = htonl(SOFTWARE_LEVEL);
			/* send response to other guy */
			sendto (0, (char *)&msg, sizeof(msg), 0,
				 (struct sockaddr *)&client_addr, sizeof(client_addr));
			close (0);
			close (fp);
			if ((hp = gethostbyaddr ((char *)&client_addr.sin_addr,
						sizeof(struct sockaddr_in), AF_INET)) == NULL)
			{
			    tisSysLog (1, 0, "Service %s sent to %s",
						 inmsg.service, INET_NTOA(client_addr.sin_addr));
			}
			else
			{
			    tisSysLog (1, 0, "Service %s sent to %s",
						 inmsg.service, hp->h_name);
			}
			tisCloseLog ();
			exit (0);
		    }
		}
	    }
	}
	close(fp);
    }
    if ((hp = gethostbyaddr ((char *)&client_addr.sin_addr,
			    sizeof(struct sockaddr_in), AF_INET)) == NULL)
    {
	tisSysLog (1, 0, "No service %s for %s",
			    inmsg.service, INET_NTOA(client_addr.sin_addr));
    }
    else
    {
	tisSysLog (1, 0, "No service %s for %s",
			   inmsg.service, hp->h_name);
    }
    tisCloseLog ();
    exit (1); /* not found */
}
