/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include "slm.h"

/* create and manage the RTU dialog box */
static Widget perList,perName,perUpdate,perCard,perKeypad;
static Widget sel,perAdmit,perIsGlobal,perIsOPI,perLogout,logoutLabel;
static char perPassword[SLM_KEYPADID_SIZE+1],cardChanged=0,firstNew=0,idChanged=0,fromInit=0;
char entryChanged=0,notSaved=0;
char slmCurrentName[SLM_NAME_SIZE+1];
Widget slmPgmButton;
static XmString previousItem=0;
typedef struct
{
  char *newPassword;
  char notUnique;
} UNIQUE_PASSWORD;

void slmCreatePersonnel(Widget parent)
{
  Widget label;
  int n=0;
  Arg args[10];
  XmString title,listLabel,selLabel,okLabel,cancelLabel,helpLabel;
#ifdef SOLARIS
  title = XmStringCreateLocalized("Personnel List");
#else
  title = XmStringCreateSimple("Personnel List");
#endif
  w_personnel = XmCreateBulletinBoardDialog(parent,"PERSONNEL",args,n);
  XtVaSetValues(w_personnel,
		XmNnoResize,True,
		XmNautoUnmanage,False,
		XmNdialogStyle,XmDIALOG_PRIMARY_APPLICATION_MODAL,
		XmNdialogTitle,title,NULL);
  n=0;
#ifdef SOLARIS
  listLabel = XmStringCreateLocalized("Names:");
  selLabel = XmStringCreateLocalized("Name:");
  cancelLabel = XmStringCreateLocalized("Dismiss");
  helpLabel = XmStringCreateLocalized("Delete");
  okLabel = XmStringCreateLocalized("Apply");
#else
  listLabel = XmStringCreateSimple("Names:");
  selLabel = XmStringCreateSimple("Name:");
  cancelLabel = XmStringCreateSimple("Dismiss");
  helpLabel = XmStringCreateSimple("Delete");
  okLabel = XmStringCreateSimple("Apply");
#endif
  /* set args */
  XtSetArg(args[n],XmNselectionLabelString,selLabel);n++;
  XtSetArg(args[n],XmNlistLabelString,listLabel);n++;
  XtSetArg(args[n],XmNcancelLabelString,cancelLabel);n++;
  XtSetArg(args[n],XmNhelpLabelString,helpLabel);n++;
  XtSetArg(args[n],XmNokLabelString,okLabel);n++;
  XtSetArg(args[n],XmNmustMatch,False);n++;
  /* create dialog */
  sel = XmCreateSelectionBox(w_personnel,"PERSEL",args,n);
  /* get children widgets */
  perList = XmSelectionBoxGetChild(sel,XmDIALOG_LIST);
  perName = XmSelectionBoxGetChild(sel,XmDIALOG_TEXT);
  XtVaSetValues(perName,XmNmaxLength,SLM_NAME_SIZE,NULL);
  XtVaSetValues(perList,XmNscrollBarDisplayPolicy,XmSTATIC,NULL);
  XtRemoveAllCallbacks(perList,XmNdefaultActionCallback);
  XtRemoveAllCallbacks(perList,XmNbrowseSelectionCallback);
  XtAddCallback(perList,XmNbrowseSelectionCallback,(XtCallbackProc)slmPerEdit,(XtPointer)0);
  XtAddCallback(perName,XmNvalueChangedCallback,(XtCallbackProc)slmPerNew,0);
  /* add callbacks */
  XtAddCallback(sel,XmNokCallback,slmPerAddReplace,(XtPointer)0);
  XtAddCallback(sel,XmNhelpCallback,slmPerDelete,(XtPointer)0);
  XtAddCallback(sel,XmNcancelCallback,slmPerDismiss,(XtPointer)0);
  /* manage selection box and free strings */
  XtManageChild(sel);
  XmStringFree(title);
  XmStringFree(listLabel);
  XmStringFree(selLabel);
  XmStringFree(cancelLabel);
  XmStringFree(helpLabel);
  /* Add text widgets */
  /* Card ID */
  n=0;
  XtSetArg(args[n],XmNx,470);n++;
  XtSetArg(args[n],XmNy,35);n++;
  XtSetArg(args[n],XmNmaxLength,SLM_CARDID_SIZE);n++;
  perCard = XmCreateText(w_personnel,"CARD",args,n);
  XtManageChild(perCard);
  XtAddCallback(perCard,XmNmodifyVerifyCallback,(XtCallbackProc)slmPerCheckCard,(XtPointer)0);
  label = XtVaCreateManagedWidget("label",xmLabelWidgetClass,w_personnel,
				  XtVaTypedArg,XmNlabelString,XmRString,"Card ID:",9,
				  XmNx,402,
				  XmNy,41,
				  NULL);
 /* Keypad ID */
  n=0;
  XtSetArg(args[n],XmNx,470);n++;
  XtSetArg(args[n],XmNy,85);n++;
  XtSetArg(args[n],XmNmaxLength,SLM_KEYPADID_SIZE);n++;
  perKeypad = XmCreateText(w_personnel,"KEYPAD",args,n);
  XtAddCallback(perKeypad,XmNmodifyVerifyCallback,slmCheckPassword,(XtPointer)perPassword);
  XtAddCallback(perKeypad,XmNactivateCallback,slmCheckPassword,(XtPointer)0);
  XtManageChild(perKeypad);
  label = XtVaCreateManagedWidget("label",xmLabelWidgetClass,w_personnel,
				  XtVaTypedArg,XmNlabelString,XmRString,"Keypad ID:",11,
				  XmNx,387,
				  XmNy,91,
				  NULL);
  /* Logout Time */
  n=0;
  XtSetArg(args[n],XmNx,470);n++;
  XtSetArg(args[n],XmNy,135);n++;
  XtSetArg(args[n],XmNsensitive,False);n++;
  perLogout = XmCreateText(w_personnel,"LOGOUT",args,n);
  XtManageChild(perLogout);
  logoutLabel = XtVaCreateManagedWidget("label",xmLabelWidgetClass,w_personnel,
					XtVaTypedArg,XmNlabelString,XmRString,"Auto Logout Time:",18,
					XmNx,330,
					XmNy,141,
					XmNsensitive,False,
					NULL);
  XtVaSetValues(perLogout,XmNverifyBell,True,0);
  XtAddCallback(perLogout,XmNmodifyVerifyCallback,(XtCallbackProc)slmPerCheckTime,0);
  XtAddCallback(perLogout,XmNvalueChangedCallback,(XtCallbackProc)slmPerCheckTime,0);
  XtAddCallback(perLogout,XmNmotionVerifyCallback,(XtCallbackProc)slmPerCheckTime,0);
  XtAddCallback(perLogout,XmNlosingFocusCallback,(XtCallbackProc)slmPerCheckTime,0);


  /* enable global access */
  perIsGlobal = XtVaCreateManagedWidget("GLOBAL",xmToggleButtonWidgetClass,w_personnel,
					XtVaTypedArg,XmNlabelString,XmRString,"Enable Global RTU Access",25,
					XmNx,330,
					XmNy,190,
					XmNindicatorSize,15,
					NULL);
  XtAddCallback(perIsGlobal,XmNvalueChangedCallback,(XtCallbackProc)slmPerAddGlobal,0);

  /* enable work station access */
  perIsOPI = XtVaCreateManagedWidget("WORKSTATION",xmToggleButtonWidgetClass,w_personnel,
					XtVaTypedArg,XmNlabelString,XmRString,"Enable Workstation Access",26,
					XmNx,330,
					XmNy,228,
					XmNindicatorSize,15,
					NULL);
  XtAddCallback(perIsOPI,XmNvalueChangedCallback,(XtCallbackProc)slmPerToggleOPI,0);

  label = XtVaCreateManagedWidget("sep",xmSeparatorWidgetClass,w_personnel,
				  XmNx,330,
				  XmNy,258,
				  XmNwidth,322,
				  NULL);
  /* Admit button */
  perAdmit = XtVaCreateManagedWidget("Admit",xmPushButtonWidgetClass,w_personnel,
				      XmNx,330,
				      XmNy,276,
				      XmNheight,29,
				      XmNwidth,75,
				      NULL);
  XtAddCallback(perAdmit,XmNactivateCallback,slmPerAdmit,(XtPointer)0);
  /* Update button */
  perUpdate = XtVaCreateManagedWidget("Save",xmPushButtonWidgetClass,w_personnel,
				      XmNx,417,
				      XmNy,276,
				      XmNheight,29,
				      XmNwidth,75,
				      NULL);
  XtAddCallback(perUpdate,XmNactivateCallback,slmPerUpdate,(XtPointer)0);
  /* Program button */
  slmPgmButton = XtVaCreateManagedWidget("Program",xmPushButtonWidgetClass,w_personnel,
				      XmNx,504,
				      XmNy,276,
				      XmNheight,29,
				      XmNwidth,75,
				      NULL);
  XtAddCallback(slmPgmButton,XmNactivateCallback,slmPerProgram,(XtPointer)0);
}
void slmPerInit()
{

  memset(perPassword,0,SLM_KEYPADID_SIZE);
  XmListDeleteAllItems(perList);
  avl_traverse(slmNameTree,INORDER,slmPerLoadList,(XtPointer)perList);
  cardChanged = 0;
  idChanged = 0;
  entryChanged = 0;
  
  XmTextSetString(perCard,"");
  XmTextSetString(perKeypad,"");
  XmTextSetString(perName,"");
  XmTextSetString(perLogout,"");
  XmToggleButtonSetState(perIsGlobal,0,True);
  XmToggleButtonSetState(perIsOPI,0,False);
  XtManageChild(w_personnel);
}
void  slmPerLoadList(void *p1,void *p2)
{
  ACCESS_ENTRY_PTR node = (ACCESS_ENTRY_PTR)p2;
  XmString  item;
  Widget list = (Widget)p1;
#ifdef SOLARIS
  item = XmStringCreateLocalized(node->slmName);
#else
  item = XmStringCreateSimple(node->slmName);
#endif
  XmListAddItem(list,item);
  XmStringFree(item);
}
void slmPerDismiss(Widget w, XtPointer client_data, XtPointer call_data)
{
  if(notSaved)
    {
      slmCreateAreYouSure("Dismiss without Saving ?","SLM");
      XtRemoveAllCallbacks(slmAreYouSure,XmNokCallback);
      XtAddCallback(slmAreYouSure,XmNokCallback,slmPerOkToDismiss,(XtPointer)0);
      return;
    }
XtUnmanageChild(w_personnel);
}
void slmPerOkToDismiss(Widget w, XtPointer client_data, XtPointer call_data)
{
fromInit = 1;
XtUnmanageChild(w_personnel);
}

void slmPerProgram(Widget w, XtPointer client_data, XtPointer call_data)
{
slmPgmInit();
XtManageChild(w_pgm);
}

/*
  Add name to list and info from text widgets to list element or
  replace edited data
*/
void slmPerAddReplace(Widget w, XtPointer client_data, XtPointer call_data)
{
  char *str,*card,keypad[SLM_KEYPADID_SIZE*2+1],*logout,*clear;
  XmString item;
  ACCESS_ENTRY  *pEntry,entry;
  UNIQUE_PASSWORD uniquePassword;
  XmToggleButtonCallbackStruct cbs;
  str = XmTextGetString(perName);
  card = XmTextGetString(perCard);			/* get the card ID */
  logout = XmTextGetString(perLogout);			/* and the Auto logout time */
  clear = XmTextGetString(perKeypad);			/* GCT 10-Apr-97 for SNWS so we can check for blank field */
  if(strlen(clear) == 0)
    {
      XmTextSetString(perKeypad,card);			/* GCT 10-Apr-97 for SNWS if keypadID is blank then copy cardID */
      clear = XmTextGetString(perKeypad);      		/* GCT 10-Apr-97 for SNWS so we can check for blank field */
    }
  if(strlen(str))
    {
#ifdef SOLARIS
      item = XmStringCreateLocalized(str);
#else
      item = XmStringCreateSimple(str);
#endif
      if(XmListItemExists(perList,item) == False)	/* False means not in list*/
	{
	  if(strlen(clear))				/* keypadID must not be EMPTY */
	    {
	      strcpy(keypad,clear);
	      tisEncrypt(keypad,SLM_KEYPADID_SIZE);
	      uniquePassword.newPassword = keypad;
	      uniquePassword.notUnique = 0;
	      avl_traverse(slmNameTree,INORDER,slmPerCheckUnique,&uniquePassword);
	      if(uniquePassword.notUnique)
		{
		  slmCreateInfo(SLM_KEYPAD_INUSE,"SLM");
		  XmStringFree(item);
		  XmTextSetString(perKeypad,"");
		  idChanged = 0;
		  XtFree(str);
		  XtFree(card);
		  XtFree(logout);
		  XtFree(clear);
		  return;
		}
	      idChanged = 0;
	    }
	  if(strlen(card))
	    {
	      uniquePassword.newPassword = card;
	      uniquePassword.notUnique = 0;
	      avl_traverse(slmNameTree,INORDER,slmPerCheckUniqueCard,&uniquePassword);
	      if(uniquePassword.notUnique)
		{
		  slmCreateInfo(SLM_CARD_INUSE,"SLM");
		  XmStringFree(item);
		  XmTextSetString(perCard,"");
		  cardChanged = 0;
		  XtFree(str);
		  XtFree(card);
		  XtFree(logout);
		  XtFree(clear);
		  return;
		}
	      cardChanged = 0;
	    }
	  pEntry = avl_newnode(ACCESS_ENTRY_SIZE);	/* obtain new node memory */
	  strcpy(pEntry->slmName,str);
	  avl_insert(slmNameTree,pEntry);	       	/* insert new node in tree */
	  strcpy(&pEntry->slmCardID[0],"0");
	  strcpy(&pEntry->slmKeypadID[0],"0");
	  strcpy(&pEntry->slmClearID[0],"0");		/* GCT 10-Apr-1997 for SNWS set this to "0" */
	  sprintf(&pEntry->slmAutoLogoutTime[0],"00:00:00\0");
	  if(strlen(logout))
	    strcpy(&pEntry->slmAutoLogoutTime[0],logout);
	  if(strlen(card))
	    strcpy(&pEntry->slmCardID[0],card);		/* replace data in tree entry with widget string */
	  strcpy(&pEntry->slmKeypadID[0],keypad);	/* save encrypted password in entry */
	  strcpy(pEntry->slmClearID,clear);	/* GCT 10-Apr-1997 for SNWS save un-encrypted password */
	  pEntry->slmRTUList = 0;			/* init RTU list */
	  pEntry->slmIsGlobal = XmToggleButtonGetState(perIsGlobal);   	/* set/clear global flag */
	  if((cbs.set = pEntry->slmIsGlobal) != 0)
	    slmPerAddGlobal(0,0,&cbs);
	  pEntry->slmIsOPI = XmToggleButtonGetState(perIsOPI); 		/* set/clear workstations flag */
	  XmListAddItem(perList,item,0);	       	/* add to list */
	}
      else						/* item in list. Search tree and replace info */
	{
	  strcpy(entry.slmName,str);		       	/* copy to entry */
	  if((pEntry = avl_findsym(slmNameTree,&entry)) != 0)
	    {
	      if(strlen(clear) && idChanged)			/* password must not be NULL */
		{
		  strcpy(keypad,clear);
		  tisEncrypt(keypad,SLM_KEYPADID_SIZE);
		  uniquePassword.newPassword = keypad;
		  uniquePassword.notUnique = 0;
		  avl_traverse(slmNameTree,INORDER,slmPerCheckUnique,&uniquePassword);
		  if(uniquePassword.notUnique)
		    {
		      slmCreateInfo(SLM_KEYPAD_INUSE,"SLM");
		      XmTextSetString(perKeypad,"");
		      idChanged = 0;
		      XmStringFree(item);
		      XtFree(str);
		      XtFree(card);
		      XtFree(logout);
		      XtFree(clear);
		      return;
		    }
		  strcpy(&pEntry->slmKeypadID[0],keypad);	/* save encrypted password in entry */
		  strcpy(pEntry->slmClearID,clear);	/* GCT 10-Apr-1997 for SNWS save un-encrypted password */
		  idChanged = 0;
		}
	      if(strlen(card) && cardChanged)
		{
		  uniquePassword.newPassword = card;
		  uniquePassword.notUnique = 0;
		  avl_traverse(slmNameTree,INORDER,slmPerCheckUniqueCard,&uniquePassword);
		  if(uniquePassword.notUnique)
		    {
		      slmCreateInfo(SLM_CARD_INUSE,"SLM");
		      XmStringFree(item);
		      XmTextSetString(perCard,"");
		      cardChanged = 0;
		      XtFree(str);
		      XtFree(card);
		      XtFree(logout);
		      XtFree(clear);
		      return;
		    }
		  strcpy(&pEntry->slmCardID[0],card);	/* replace data in tree entry with widget string */
		  cardChanged = 0;
		}
	      if(strlen(logout))
		strcpy(&pEntry->slmAutoLogoutTime[0],logout);
	      pEntry->slmIsGlobal = XmToggleButtonGetState(perIsGlobal);   	/* set/clear global flag */
	      if((cbs.set = pEntry->slmIsGlobal) != 0)
		slmPerAddGlobal(0,0,&cbs);
	      pEntry->slmIsOPI = XmToggleButtonGetState(perIsOPI); 		/* set/clear workstations flag */
	    }
	}
      XmStringFree(item);
    }
  XtFree(str);
  XtFree(card);
  XtFree(logout);
  XtFree(clear);
  entryChanged=0;
}
void slmPerNew(Widget w, XtPointer client_data, XtPointer call_data)
{
  if(!firstNew)
    {
      firstNew = 1;
      XmTextSetString(perCard,"");
      XmTextSetString(perKeypad,"");
    }
}
void slmPerCheckUnique(void *p1,void *p2)
{
ACCESS_ENTRY_PTR node = (ACCESS_ENTRY_PTR)p2;
UNIQUE_PASSWORD *unq = (UNIQUE_PASSWORD *)p1;
if(!strcmp(node->slmKeypadID,unq->newPassword))
  unq->notUnique = 1;
}
void slmPerCheckUniqueCard(void *p1,void *p2)
{
ACCESS_ENTRY_PTR node = (ACCESS_ENTRY_PTR)p2;
UNIQUE_PASSWORD *unq = (UNIQUE_PASSWORD *)p1;
if(!strcmp(node->slmCardID,unq->newPassword))
  unq->notUnique = 1;
}
void slmPerEdit(Widget w, XtPointer client_data, XmSelectionBoxCallbackStruct *cbs)
{
  char *str;
#ifdef SOLARIS
  XmStringGetLtoR(cbs->value,XmFONTLIST_DEFAULT_TAG,&str);
#else
  XmStringGetLtoR(cbs->value,XmSTRING_DEFAULT_CHARSET,&str);
#endif
  if(entryChanged)
    {
      slmCreateAreYouSure("Continue without Applying changes?","SLM");
      XtRemoveAllCallbacks(slmAreYouSure,XmNokCallback);
      XtRemoveAllCallbacks(slmAreYouSure,XmNcancelCallback);
      XtAddCallback(slmAreYouSure,XmNokCallback,slmPerOkToEdit,(XtPointer)str);
      XtAddCallback(slmAreYouSure,XmNcancelCallback,slmPerNoEdit,(XtPointer)str);
    }
  else
    slmPerOkToEdit(0,str,0);
}
void slmPerNoEdit(Widget w, XtPointer client_data, XtPointer call_data)
{
  char *str;

  entryChanged = 0;
  cardChanged = 0;
  entryChanged = 0;
  str = XmTextGetString(perName);
#ifdef SOLARIS
  previousItem = XmStringCreateLocalized(str);
#else
  previousItem = XmStringCreateSimple(str);
#endif
  XmListSelectItem(perList,previousItem,True);
}
void slmPerOkToEdit(Widget w, XtPointer client_data, XtPointer call_data)
{
ACCESS_ENTRY  *pEntry,entry;
char *name = (char *)client_data;
strcpy(entry.slmName,name);				/* copy to entry */
XtFree(name);
if((pEntry = avl_findsym(slmNameTree,&entry)) != 0)
  {
    XtRemoveAllCallbacks(perKeypad,XmNmodifyVerifyCallback);
    XtRemoveAllCallbacks(perKeypad,XmNactivateCallback);
    XmTextSetString(perName,pEntry->slmName);
    XmTextSetString(perCard,pEntry->slmCardID);
    XmTextSetString(perKeypad,pEntry->slmClearID);
    XmTextSetString(perLogout,pEntry->slmAutoLogoutTime);
    XmToggleButtonSetState(perIsGlobal,pEntry->slmIsGlobal,True);
    XmToggleButtonSetState(perIsOPI,pEntry->slmIsOPI,False);
    if(pEntry->slmIsOPI)
      {
	XtVaSetValues(perLogout,XmNsensitive,True,NULL);
	XtVaSetValues(logoutLabel,XmNsensitive,True,NULL);
      }
    else
      {
	XtVaSetValues(perLogout,XmNsensitive,False,NULL);
	XtVaSetValues(logoutLabel,XmNsensitive,False,NULL);
      }
    cardChanged = 0;
    entryChanged = 0;
    idChanged = 0;
    notSaved = 0;
    strcpy(slmCurrentName,entry.slmName);
    XtAddCallback(perKeypad,XmNmodifyVerifyCallback,slmCheckPassword,(XtPointer)perPassword);
    XtAddCallback(perKeypad,XmNactivateCallback,slmCheckPassword,(XtPointer)0);
  }
}

void slmPerDelete(Widget w, XtPointer client_data, XtPointer call_data)
{
  char msg[80];
  static ACCESS_ENTRY  entry;
  char *str;

  str = XmTextGetString(perName);		       	/* get the selected name */
  strcpy(entry.slmName,str);				/* copy to entry */
  sprintf(msg,"Delete %s\0",entry.slmName);
  slmCreateAreYouSure(msg,"SLM");
  XtRemoveAllCallbacks(slmAreYouSure,XmNokCallback);
  XtAddCallback(slmAreYouSure,XmNokCallback,slmPerOkToDelete,(XtPointer)&entry);
}
void slmPerOkToDelete(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmString item;
  ACCESS_ENTRY_PTR entry;
  ACCESS_ENTRY_PTR pEntry = (ACCESS_ENTRY_PTR)client_data;
  if((entry = avl_delete(slmNameTree,pEntry)) != NULL)
    {
#ifdef SOLARIS
      item = XmStringCreateLocalized(pEntry->slmName);
#else
      item = XmStringCreateSimple(pEntry->slmName);
#endif
      XmListDeleteItem(perList,item);
      XmStringFree(item);
      avl_freenode(entry);
      XmTextSetString(perName,"");
      XmTextSetString(perCard,"");
      XmTextSetString(perKeypad,"");
      XmTextSetString(perLogout,"");
      XmToggleButtonSetState(perIsGlobal,0,True);
      XmToggleButtonSetState(perIsOPI,0,False);
      cardChanged = 0;
      notSaved |= 1;
    }
}
void slmPerAdmit(Widget w, XtPointer client_data, XtPointer call_data)
{
  ACCESS_ENTRY_PTR pEntry;
  ACCESS_ENTRY entry;
  char *str;
  Widget list;
  str = XmTextGetString(perName);	       	/* get the selected name */
  strcpy(entry.slmName,str);		       	/* copy to entry */
  if((pEntry = avl_findsym(slmNameTree,&entry)) != 0)
    {
      list = XmSelectionBoxGetChild(w_adm,XmDIALOG_LIST);
      XmListDeselectAllItems(list);
      slmPerRTUSelect(pEntry->slmRTUList,list);
      XtRemoveAllCallbacks(XmSelectionBoxGetChild(w_adm,XmDIALOG_OK_BUTTON),XmNactivateCallback);
      XtAddCallback(XmSelectionBoxGetChild(w_adm,XmDIALOG_OK_BUTTON),XmNactivateCallback,slmAdmUpdate,(XtPointer)pEntry);
      XtManageChild(w_adm);
    }
}

/* write the main database file overwriting the old one and create & write the RTU files too */
void slmPerUpdate(Widget w,XtPointer client_data,XtPointer call_data)
{
  slmWriteRTUFiles();
  slmWriteMain();
  notSaved=0;
}

void slmPerRTUSelect(RTU_ENTRY_PTR rtu,Widget list)
{
  XmString  item;
  char rName[20];
  XtVaSetValues(list,XmNselectionPolicy,XmMULTIPLE_SELECT,NULL);
  while(rtu)
    {
      strcpy(rName,rtu->slmRTUName);
      if(rtu->slmIsRedundant != NON_REDUNDANT)
	{
	  rName[strlen(rName)-1] = 0;		/* drop a or b */
	  rtu = (RTU_ENTRY_PTR)rtu->next;	/* skip next entry to avoid de-selecting this one */
	}
#ifdef SOLARIS
      item = XmStringCreateLocalized(rName);
#else
      item = XmStringCreateSimple(rName);
#endif
      XmListSelectItem(list,item,False);
      XmStringFree(item);
      if((rtu = (RTU_ENTRY_PTR)rtu->next) == 0)
	break;
    }
  XtVaSetValues(list,XmNselectionPolicy,XmEXTENDED_SELECT,NULL);
}
int	slmPerCheckCard(Widget w,char *data,XmTextVerifyCallbackStruct *cbs)
{

  if(cbs -> reason == XmCR_ACTIVATE) return;
#ifdef SOLARIS
  if(cbs->text->length == 0) return;
#else
  if(cbs->text->ptr == NULL) return;
#endif
  if(cbs->startPos < cbs->currInsert)		/* delete any where... delete whole string */
    {
      cardChanged = 1;
      entryChanged = 1;
      notSaved |= entryChanged;
      return;
    }
/* Only 0 - 9 */
  if(!isdigit(cbs->text->ptr[0]))
    cbs->doit = False;				/* don't put char in widget */
  else
    {
      cardChanged = 1;
      entryChanged = 1;
      notSaved |= entryChanged;
      cbs->doit = True;
    }
}
void slmCheckPassword(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmTextVerifyCallbackStruct *cbs = (XmTextVerifyCallbackStruct *)call_data;

  if(cbs->reason == XmCR_ACTIVATE) return;

  if(cbs->startPos < cbs->currInsert)		/* delete any where... delete whole string */
    {
      idChanged = 1;
      entryChanged = 1;
      notSaved |= entryChanged;
      return;
    }

  if(cbs->text->ptr)
    {
      if(isdigit(cbs->text->ptr[0]))
	{
	  idChanged = 1;
	  entryChanged = 1;
	  notSaved |= entryChanged;
	  cbs->doit = True;
	}
      else if(cbs->text->ptr[0] == 0)		/* this istrue if XmTextSetString uses empty string */
	cbs->doit = True;
      else
	cbs->doit = False;
    }
}
int	slmPerCheckTime(Widget w,char *data,XmTextVerifyCallbackStruct *cbs)
{
  char c,*str,field[11],pc,*pChar;
  int len = XmTextGetLastPosition(w);
  static char *timeComplete = "00:00:00";

  XtVaGetValues(w,XmNvalue,&pChar,NULL);
  if(len) pc = pChar[len-1];
  if(cbs->event == NULL) return;
  if(cbs->reason == XmCR_VALUE_CHANGED)
    {
      XmTextSetInsertionPosition(w,len);
      return;
    }
/* no cursor movement allowed */
  if(cbs->reason == XmCR_MOVING_INSERT_CURSOR)
    {
      if((cbs->newInsert != len) && cbs->event)
	cbs->doit = False;
      return;
    }
  if(cbs->reason == XmCR_LOSING_FOCUS)
    {
      if((len > 0) && (len < 8))
	{
	  str = XmTextGetString(w);
	  strcpy(field,str);
	  XtFree(str);
	  strcat(field,&timeComplete[len]);
	  XmTextSetString(w,field);
	  entryChanged = 1;
	  notSaved |= entryChanged;
	}
      return;
    }
  /* no backspacing or typing in middle of string */
  if(cbs->currInsert < len)
    {
      cbs->doit = False;
      return;
    }
  /* delete extra chars too */
#ifdef SOLARIS
  if(cbs->text->length == 0)
    {
      if(cbs->startPos == 3 || cbs->startPos == 6)
	cbs->startPos--;
      entryChanged = 1;
      notSaved |= entryChanged;
      cbs->doit = True;
      return;
    }
#else
  if(cbs->text->ptr == NULL)
    {
      if(cbs->startPos == 2 || cbs->startPos == 5)
	cbs->startPos--;
      entryChanged = 1;
      notSaved |= entryChanged;
      cbs->doit = True;
      return;
    }
#endif
  /* no insertions from clipboard */
  if(cbs->text->length > 1)
    {
      cbs->doit = False;
      return;
    }
  /* digits only and no more than 8 */
  if(!isdigit(c = cbs->text->ptr[0]) || len >= 8)
    cbs->doit = False;
  else if(((len == 3) || (len == 6)) && c > '5')
    cbs->doit = False;		/* no minute or second > 59 */
  else if((len == 0) && (c > '2'))
    cbs->doit = False;
  else if(len == 1 && c > '3' && ((pc != '0') && (pc != '1')))
    cbs->doit = False;		/* no hour greater than 23 */
  else if(len == 1 || len == 4)
    {
      cbs->text->ptr = XtRealloc(cbs->text->ptr,2);
      cbs->text->length = 2;
      cbs->text->ptr[0] = c;
      cbs->text->ptr[1] = ':';	/* colon in these positions only */
      cbs->doit = True;
      entryChanged = 1;
      notSaved |= entryChanged;
    }
}
void slmPerAddGlobal(Widget w, XtPointer client_data, XmToggleButtonCallbackStruct *cbs)
{
  char *str;
  ACCESS_ENTRY  *pEntry,entry;
  RTU_ENTRY_PTR rtu,pRtu;

  str = XmTextGetString(perName);
  strcpy(entry.slmName,str);
  if((pEntry = avl_findsym(slmNameTree,&entry)) != 0)
    {
      slmEmptyList(pEntry->slmRTUList);
      pEntry->slmRTUList = 0;
      if(cbs->set)
	{
	  rtu = slmRTUNameTree;
	  while(rtu)
	    {
	      pRtu = (RTU_ENTRY_PTR)malloc(RTU_ENTRY_SIZE);
	      memcpy((char *)pRtu,(char *)rtu,RTU_ENTRY_SIZE);
	      pRtu->next = pEntry->slmRTUList;
	      pEntry->slmRTUList = pRtu;
	      rtu = rtu->next;
	    }
	}
      entryChanged = 1;
      notSaved |= entryChanged;
    }
  XtFree(str);
}
void slmPerToggleOPI(Widget w, XtPointer client_data, XmToggleButtonCallbackStruct *cbs)
{
  if(cbs->set)
    {
      XtVaSetValues(perLogout,XmNsensitive,True,NULL);
      XtVaSetValues(logoutLabel,XmNsensitive,True,NULL);
    }
  else
    {
      XtVaSetValues(perLogout,XmNsensitive,False,NULL);
      XtVaSetValues(logoutLabel,XmNsensitive,False,NULL);
    }
  entryChanged = 1;
  notSaved |= entryChanged;
}
