#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vdictHEADER
#include vlimitsHEADER
#include vprefHEADER
#include vrulerHEADER
#include vlistviewHEADER
#include vbuttonHEADER

extern "C"
{
#ifdef	WIN32
#include    <winsock.h>
#endif	WIN32
}

#include "snlvu.h"
#include "snlvuApp.hxx"
#include "snlvuMain.hxx"
#include "snlvuChannels.hxx"
#include "snlvuVariables.hxx"
#include "snlvuExpand.hxx"
#include "snlvuStateSets.hxx"
#include "snlvuUtils.hxx"


/*******************************************************************
*
*    Main ()
*
*******************************************************************/

int main (int argc, char **argv)
{
    const vchar	**newargv;
    int		newargc;
    char	iocname[IOCNAME_SZ + 1];

 #ifdef WIN32
    WSADATA wsaData; 
    WORD    wVersionRequested;
    int err;
#endif WIN32

    tisOpenLog ("snlvu", 6, 2, NULL, NULL);

    // Initialize Galaxy
    vstartup (argc, argv);

    // intern tag names
    snlvuInternNames ();

    // get the arg list after galaxy has taken out what it needs
    vpref::GetArgs (&newargc, &newargv);

    if (newargc != 2)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_ArgErr);
	tisSysLog (0,0, "Usage: snlvu iocname\n");
	tisSysLog (0,0, "Abnormal exit\n");
	exit (EXIT_FAILURE);
    }

#ifdef WIN32
    wVersionRequested = MAKEWORD(1, 1);
    // Tell the user that we couldn't find a useable winsock.dll.
    if ((err = WSAStartup(wVersionRequested, &wsaData)) != 0) 
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_WinsockErr);
	exit(EXIT_FAILURE);
    }
#endif WIN32

    vcharExportBounded (vcharScribeSystemExport (newargv[1]),
			iocname,
			IOCNAME_SZ);
    iocname[IOCNAME_SZ] = 0;
    if ((strlen (iocname)) > IOCNAME_SZ)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_IocNameLenErr, IOCNAME_SZ);
	tisSysLog (0,0, "Ioc name must less or equal to %d\n", IOCNAME_SZ);
	tisSysLog (0,0, "Abnormal exit\n");
	exit(EXIT_FAILURE);
    }
    snlvuApplication *snlvuApp = new snlvuApplication (iocname);

    // if any errors during iniitalization
    if ((snlvuApp -> GetInitErrors()) == SNLVU_ERROR)
    {
	delete snlvuApp;
	tisSysLog (0,0, "Abnormal exit due to app init errors\n");
	exit (EXIT_FAILURE);
    }

    /* Enter the event loop */
    veventProcess ();

    delete snlvuApp;

    tisSysLog (0,0, "Normal exit\n");

    exit (EXIT_SUCCESS);
    return EXIT_FAILURE;
}


/*******************************************************************
*
*    snlvuApplication constructor
*
*******************************************************************/

snlvuApplication::snlvuApplication (char *iocname)
{
    connected = 0;
    initErrors = SNLVU_SUCCESS;
    initedTisGalaxy = vFALSE;
    strncpy (iocName, (const char *)iocname, IOCNAME_SZ);
    iocName[IOCNAME_SZ] = 0;
    snlvuTimer = NULL;
    mainDialog = NULL;

    snlvuVApplication = vapplication::GetCurrent ();
    snlvuResource = snlvuVApplication -> GetResources ();

    messagesResource = vresourceGet (snlvuResource, snlvu_MessagesDictionary);
    titlesResource = vresourceGet (snlvuResource, snlvu_TitlesDictionary);

    // init tisGalaxy, see tisgalaxylib.<hxx,cxx>
    tisGalaxyResource = startTisGalaxy ();
    initedTisGalaxy = vTRUE;

    connected = snlMonConnect (iocname);

    // load the main dialog
    vresource mainDlgRes = vresourceGet (snlvuResource, snlvu_MainDialog);
    mainDialog = new MainDialog (mainDlgRes);
    if ((mainDialog -> GetLoadInitErrors ()) == SNLVU_ERROR)
    {
	delete mainDialog;
	initErrors = SNLVU_ERROR;
	return;
    }
    mainDialog -> OpenMain (this);

    snlvuTimer = new TimerClass;
    snlvuTimer -> StoreApp (this);
    snlvuTimer -> StartTimer ();
}


/*******************************************************************
*
*    snlvuApplication destructor
*
*******************************************************************/

snlvuApplication::~snlvuApplication (void)
{
    if (connected)
    {
	snlMonDisconnect ();
    }
    // close down tisgalaxy if used
    if (initedTisGalaxy == vTRUE)
    {
	stopTisGalaxy (tisGalaxyResource);
    }
    if (snlvuTimer != NULL)
    {
	delete snlvuTimer;
    }
    if (mainDialog != NULL)
    {
	delete mainDialog;
    }
}


/*******************************************************************
*
*    TimerClass
*
*******************************************************************/

void TimerClass::StartTimer (void)
{
    // set timer to restart after observer is finished
    SetRecurrent ();
    // set timer to expire every 5 seconds
    // args are seconds and nanoseconds
    SetPeriod (5, 0);
    Start ();
}


/*******************************************************************
*
*    TimerClass, observer
*
*******************************************************************/

void TimerClass::ObserveTimer (void)
{
    MainDialog *mainDialog = snlvuApp -> GetMainDialog ();
printf("timer fired, connected = %d\n",snlvuApp -> IsConnected ());
    if (snlvuApp -> IsConnected ())
    {
	mainDialog -> UpdateProgramList (0);
	if (mainDialog -> channelsDialog != NULL)
	{
	    mainDialog -> channelsDialog -> UpdateChannelsList ();
	}
	if (mainDialog -> variablesDialog != NULL)
	{
	    if (mainDialog -> variablesDialog -> InUse == vFALSE)
	    {
		mainDialog -> variablesDialog -> UpdateVariablesList (0);
	    }
	}
	if (mainDialog -> expandDialog != NULL)
	{
	    if (mainDialog -> expandDialog -> InUse == vFALSE)
	    {
		mainDialog -> expandDialog -> UpdateExpandList ();
	    }
	}
	if (mainDialog -> stateSetsDialog != NULL)
	{
	    if (mainDialog -> stateSetsDialog -> InUse == vFALSE)
	    {
		mainDialog -> stateSetsDialog -> UpdateStateSetsList ();
	    }
	}
    }
    else
    {
	if (mainDialog -> channelsDialog != NULL)
	{
	    delete mainDialog -> channelsDialog;
	    mainDialog -> channelsDialog = NULL;
	}
	if (mainDialog -> variablesDialog != NULL)
	{
	    delete mainDialog -> variablesDialog;
	    mainDialog -> variablesDialog = NULL;
	}
	if (mainDialog -> expandDialog != NULL)
	{
	    delete mainDialog -> expandDialog;
	    mainDialog -> expandDialog = NULL;
	}
	if (mainDialog -> stateSetsDialog != NULL)
	{
	    delete mainDialog -> stateSetsDialog;
	    mainDialog -> stateSetsDialog = NULL;
	}
    }
    snlvuApp -> Connect ();
}

