#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vbuttonHEADER
#include vstartupHEADER
#include vcontrolHEADER
#include vtimerHEADER
#include vprefHEADER
#include vlistviewHEADER
#include vrulerHEADER
#include vlistitemHEADER

extern "C"
{
#include "alarm.h"
#include "alarmString.h"
}

#include "snlvu.h"
#include "snlvuApp.hxx"
#include "snlvuMain.hxx"
#include "snlvuChannels.hxx"
#include "snlvuUtils.hxx"


vkindTYPED_FULL_DEFINITION(ChannelsDialog, vdialog, NULL);

/*******************************************************************
*
*    ChannelsDialog override for constructors init ()
*
*******************************************************************/

void ChannelsDialog::LoadInit (vresource res)
{
    // init stuff
    PrevChannels = NULL;

    // RequestSubLoad returns ptr to uninited dialog object
    // gets init'd when dialog gets init'd by LoadInit()

    // ruler header
    channelsRuler = vruler::RequestSubLoad (this, snlvu_DialogRuler);

    channelsListView = ChannelsListView::RequestSubLoad (this, snlvu_ChannelsListView);

    exitButton = ChannelsExitButton::RequestSubLoad (this, snlvu_ExitButton);

    // init inherited data
    vdialog::LoadInit (res);

    // store ptrs to main dialog
    channelsListView -> StoreChannelsDialog (this);
    exitButton -> StoreChannelsDialog (this);

    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
}


/*******************************************************************
*
*    ChannelsDialog Destroy function
*
*******************************************************************/

void ChannelsDialog::Destroy (void)
{
    if (PrevChannels != NULL)
    {
	free ((void *)PrevChannels);
    }
    if (IsOpen ())
    {
	Close ();
    }
    vdialog::Destroy ();
}

/*******************************************************************
*
*    ChannelsDialog observe window event, looking for exit
*
*******************************************************************/

void ChannelsDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
	vevent::StopProcessing ();
}


/*******************************************************************
*
*    ChannelsDialog observe dialog reconfigure, looking for a 
*	resize event. we want to resize the ruler and its components
*	also
*
*******************************************************************/

int ChannelsDialog::HandleConfigure (vevent *event)
{
    vwindow::HandleConfigure (event);
    if ( event -> GetType () == veventWINDOW_CONFIGURE && event -> IsSize () )
    {
	int		rulerWidth;
	int		itemCount;
	int		totalItemWidth = 0;
	int		itemWidth;
	int		i;
	vrulerItem	*item;
	const vrect	*rulerRect;
	vlist *list = channelsListView -> GetList ();

	rulerRect = channelsRuler -> GetRect ();
	rulerWidth = rulerRect -> GetWidth ();
	itemCount = channelsRuler -> GetItemCount ();
	for (i = 0; i < itemCount; i++)
	{
	    item = channelsRuler -> GetItemAt (i);
	    itemWidth = item -> GetWidth ();
	    totalItemWidth += itemWidth;
	}
	if (totalItemWidth < rulerWidth)
	{
	    float	itemWidth;
	    float	itemWidthPct;
	    int		newItemWidth;
	    int		remainingWidth = rulerWidth;

	    for (i = 0; i < itemCount; i++)
	    {
		item = channelsRuler -> GetItemAt (i);
		itemWidth = (float)(item -> GetWidth ());
		itemWidthPct = itemWidth / (float)totalItemWidth;
		newItemWidth = (int)(itemWidthPct * rulerWidth);
		if ((remainingWidth - newItemWidth) <= 0)
		{
		    newItemWidth = remainingWidth;
		}
		item -> SetWidth (newItemWidth);
		remainingWidth -= newItemWidth;
	    }
	    // update the listviews columns widths
	    channelsListView -> SetColumnsWidths ();
	}

    }
    return (vTRUE);
}


/*******************************************************************
*
*    ChannelsDialog 
*
*******************************************************************/

void ChannelsDialog::OpenChannels (MainDialog *dlg, char *progName, long tid)
{
    mainDialog = dlg;
    strncpy (programName, progName, sizeof (programName));
    programName[PROGRAM_NAME_SIZE] = 0;
    TaskId = tid;

    SetChannelsTitle ();

    channelsListView -> SetDefaultCellContentIndent (0);

    UpdateChannelsList ();

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open ();
    vevent::Process ();
}


/*******************************************************************
*
*    ChannelsDialog function to set the dialog title 
*
*******************************************************************/

void ChannelsDialog::SetChannelsTitle ()
{
    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource titleRes = app -> GetTitlesResource ();

    vstr *titleFormat = vresourceGetString (titleRes, snlvu_ChannelsTitle);
    vstr *progName = vstrCloneScribed (vcharScribeSystem (programName));
    vstr *channelsTitle = vstrCloneScribed (vcharScribeFormat (titleFormat, progName));
    SetTitle (channelsTitle);
    vstrDestroy (progName);
    vstrDestroy (titleFormat);
    vstrDestroy (channelsTitle);
}


/*******************************************************************
*
*    ChannelsDialog function to update the channels list in the main
*	view
*
*******************************************************************/

void ChannelsDialog::UpdateChannelsList (void)
{
    REPCHANS_REPLY *channels;
    int		  numChannels;

    snlMon_ReportChannels (programName, TaskId, &channels, &numChannels);

    // check the list against the previous list for a change
    if (PrevChannels != NULL)
    {
	// if no change, free the returned list and get out
	if ((memcmp ((const void *)channels, (const void *)PrevChannels,
		      sizeof (REPCHANS_REPLY) * numChannels)) == 0)
	{
	    free ((void *)channels);
	    return;
	}
    }
    // the list has changed, free the previous list, and set current
    if (PrevChannels != NULL)
    {
	free ((void *)PrevChannels);
	PrevChannels = NULL;
    }
    // remember the new stuff
    PrevChannels = channels;

    // build list for listview
    channelsListView -> BuildList (channels, numChannels);
}



vkindTYPED_FULL_DEFINITION(ChannelsListView, vlistview, NULL);

/*******************************************************************
*
*    ChannelsListView 
*
*******************************************************************/

void ChannelsListView::Destroy (void)
{
    vlist *list = GetList ();
    if (list != NULL)
    {
	SetList (NULL);
	delete list;
    }
    vlistview::Destroy ();
}


/*******************************************************************
*
*    ChannelsListView, build the list
*
*******************************************************************/

void ChannelsListView::BuildList (REPCHANS_REPLY *channels, int numChannels)
{
    REPCHANS_REPLY	*p = channels;
    int			i;
    int			row = 0;

    vlist *newList = new vlist;
    vlist *oldList = GetList ();

    // each cell has same value, the p ptr, what gets printed in each
    // cell is determined in the DrawCellContents override
    for (i = 0; i < numChannels; i++, p++)
    {

	newList -> SetCellValue (row, 0, (const void *)p);
	newList -> SetCellValue (row, 1, (const void *)p);
	newList -> SetCellValue (row, 2, (const void *)p);
	newList -> SetCellValue (row, 3, (const void *)p);
	newList -> SetCellValue (row, 4, (const void *)p);
	newList -> SetCellValue (row, 5, (const void *)p);
	row++;
    }
    if (oldList != NULL)
    {
	delete oldList;
    }
    int rowCount = newList -> GetRowCount ();
    if (rowCount > 0)
    {
	newList -> Sort (vlistROW, 0, NULL);
    }
    SetList (newList);
    if (rowCount > 0)
    {
	SetColumnsWidths ();
	vlistSelection *selected = GetSelection ();
	newList -> SetSelectMethod (selected, vlistNO_SELECTION);
    }
}


/*******************************************************************
*
*    ChannelsListView, override 
*
*******************************************************************/

void ChannelsListView::DrawCellContents (int row, int col, int selected, vrect *cellRect)
{
    REPCHANS_REPLY	*p;
    char		*cPtr;
    vlist		*viewList = (vlist *)GetList ();

    vdraw::GSave ();
    vdraw::RectsClip (cellRect, 1);

    p = (REPCHANS_REPLY *)(viewList -> GetCellValue (row, col));

    cPtr = " ";

    if (col == 0)
    {
	if (strlen (p -> ChanName))
	{
	    cPtr = p -> ChanName;
	}
	vdraw::ScribedRectShowCenterY (vcharScribeSystem (cPtr), cellRect);
    }
    else if (col == 1)
    {
	// the value has leading spaces
	char	*ptr = p -> Value;
	while (*ptr)
	{
	    if (*ptr == ' ')
	    {
		ptr++;
	    }
	    else
	    {
		break;
	    }
	}
	if (strlen (ptr))
	{
	    cPtr = ptr;
	}
	vdraw::ScribedRectShowCenterY (vcharScribeSystem (cPtr), cellRect);
    }
    else if (col == 2)
    {
	cPtr = p -> Monitored ? "Yes" : "No";
	vdraw::ScribedRectShowCenterY (vcharScribeSystem (cPtr), cellRect);
    }
    else if (col == 3)
    {
	cPtr = p -> Connected ? "Yes" : "No";
	vdraw::ScribedRectShowCenterY (vcharScribeSystem (cPtr), cellRect);
    }
    else if (col == 4)
    {
	// 0 = Return to normal, chg to no alarm
	if (p -> Stat == 0)
	{
	    cPtr = alarmSeverityString[p -> Stat];
	}
	else
	{
	    cPtr = alarmStatusString[p -> Stat];
	}
	vdraw::ScribedRectShowCenterY (vcharScribeSystem (cPtr), cellRect);
    }
    else
    {
	cPtr = alarmSeverityString[p-> Severity];
	vdraw::ScribedRectShowCenterY (vcharScribeSystem (cPtr), cellRect);
    }

    vdraw::GRestore ();
}


/*******************************************************************
*
*    ChannelsListView, set list column widths based on main dialog's
*	ruler item widths
*
*******************************************************************/

void ChannelsListView::SetColumnsWidths (void)
{
    vruler	*channelsRuler = channelsDialog -> GetChannelsRuler ();
    vrulerItem	*item;
    int		itemWidth;
    int		itemCount = channelsRuler -> GetItemCount ();
    int		i;

    for (i = 0; i < itemCount; i++)
    {
	item = channelsRuler -> GetItemAt (i);
	itemWidth = item -> GetWidth ();
	SetColumnWidth (i, itemWidth);
    }
}


vkindTYPED_FULL_DEFINITION(ChannelsExitButton, vbutton, NULL);

/*******************************************************************
*
*    ChannelsExitButton
*
*******************************************************************/

void ChannelsExitButton::ObserveDialogItem (vevent *event)
{
    // stop secondary event loop started in open()
    vevent::StopProcessing ();
}

