#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vbuttonHEADER
#include vstartupHEADER
#include vcontrolHEADER
#include vtimerHEADER
#include vprefHEADER
#include vlistviewHEADER
#include vrulerHEADER

extern "C"
{
#include "usrseq.h"
}

#include "snlvu.h"
#include "snlvuApp.hxx"
#include "snlvuMain.hxx"
#include "snlvuEnter.hxx"
#include "snlvuInstances.hxx"
#include "snlvuEditInstance.hxx"
#include "snlvuUtils.hxx"

static int InstancesCompare (const void *first, const void *second);


vkindTYPED_FULL_DEFINITION(InstancesDialog, vdialog, NULL);

/*******************************************************************
*
*    InstancesDialog override for constructors init ()
*
*******************************************************************/

void InstancesDialog::LoadInit (vresource res)
{
    // init some stuff
    instances = NULL;
    instancesModified = vFALSE;

    // RequestSubLoad returns ptr to uninited dialog object
    // gets init'd when dialog gets init'd by LoadInit()

    // ruler header
    instancesRuler = vruler::RequestSubLoad (this, snlvu_InstancesRuler);

    instancesListView = InstancesListView::RequestSubLoad (this, snlvu_InstancesListView);

    startButton = InstancesStartButton::RequestSubLoad (this, snlvu_StartButton);

    defineButton = InstancesDefineButton::RequestSubLoad (this, snlvu_DefineButton);

    deleteButton = InstancesDeleteButton::RequestSubLoad (this, snlvu_DeleteButton);

    editButton = InstancesEditButton::RequestSubLoad (this, snlvu_EditButton);

    exitButton = InstancesExitButton::RequestSubLoad (this, snlvu_ExitButton);

    // init inherited data
    vdialog::LoadInit (res);

    // store ptrs to main dialog
    instancesListView -> StoreInstancesDialog (this);
    startButton -> StoreInstancesDialog (this);
    defineButton -> StoreInstancesDialog (this);
    deleteButton -> StoreInstancesDialog (this);
    editButton -> StoreInstancesDialog (this);
    exitButton -> StoreInstancesDialog (this);

    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
}


/*******************************************************************
*
*    InstancesDialog Destroy function
*	free up memory of allocated Instances in the list, 
*	then delete the list
*
*******************************************************************/

void InstancesDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }
    vliststr *list = (vliststr *)(instancesListView -> GetList ());
    instancesListView -> SetList (NULL);
    if (list != NULL)
    {
	delete list;
    }
    if (instances != NULL)
    {
	vlistIterator	iterator;
	Instance	*instance;
	iterator.Start (instances);
	while (iterator.Next ())
	{
	    instance = (Instance *)(iterator.GetValue ());
	    vmemFree ((void *)instance);
	}
	iterator.Finish ();
	delete instances;
    }
    vdialog::Destroy ();
}

/*******************************************************************
*
*    InstancesDialog observe window event, looking for exit
*
*******************************************************************/

void InstancesDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    InstancesDialog observe dialog reconfigure, looking for a 
*	resize event. we want to resize the ruler and its components
*	also
*
*******************************************************************/

int InstancesDialog::HandleConfigure (vevent *event)
{
    vwindow::HandleConfigure (event);
    if ( event -> GetType () == veventWINDOW_CONFIGURE && event -> IsSize () )
    {
	int		rulerWidth;
	int		itemCount;
	int		totalItemWidth = 0;
	int		itemWidth;
	int		i;
	vrulerItem	*item;
	const vrect	*rulerRect;
	vlist *list = instancesListView -> GetList ();

	rulerRect = instancesRuler -> GetRect ();
	rulerWidth = rulerRect -> GetWidth ();
	itemCount = instancesRuler -> GetItemCount ();
	for (i = 0; i < itemCount; i++)
	{
	    item = instancesRuler -> GetItemAt (i);
	    itemWidth = item -> GetWidth ();
	    totalItemWidth += itemWidth;
	}
	if (totalItemWidth < rulerWidth)
	{
	    float	itemWidth;
	    float	itemWidthPct;
	    int		newItemWidth;
	    int		remainingWidth = rulerWidth;

	    for (i = 0; i < itemCount; i++)
	    {
		item = instancesRuler -> GetItemAt (i);
		itemWidth = (float)(item -> GetWidth ());
		itemWidthPct = itemWidth / (float)totalItemWidth;
		newItemWidth = (int)(itemWidthPct * rulerWidth);
		if ((remainingWidth - newItemWidth) <= 0)
		{
		    newItemWidth = remainingWidth;
		}
		item -> SetWidth (newItemWidth);
		remainingWidth -= newItemWidth;
	    }
	    // update the listviews columns widths
	    instancesListView -> SetColumnsWidths ();
	}
    }
    return (vTRUE);
}


/*******************************************************************
*
*    InstancesDialog 
*
*******************************************************************/

void InstancesDialog::OpenInstances (MainDialog *dlg)
{
    mainDialog = dlg;

    if (mainDialog -> startInstance == vFALSE)
    {
	startButton -> Disable ();
    }

    if (mainDialog -> defineInstance == vFALSE)
    {
	defineButton -> Disable ();
    }

    if (mainDialog -> deleteInstance == vFALSE)
    {
	deleteButton -> Disable ();
    }

    if (mainDialog -> editInstance == vFALSE)
    {
	editButton -> Disable ();
    }

    LoadInstances ();

    UpdateInstancesList ();

    instancesListView -> SetDefaultCellContentIndent (0);

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open ();
    vevent::Process ();
}


/*******************************************************************
*
*    InstancesDialog function to update the instances list
*
*******************************************************************/

void InstancesDialog::LoadInstances (void)
{
    Instance	*instance;
    int		row = 0;

    snlvuApplication *app = mainDialog -> GetApplication ();
    char *iocName = app -> GetIocName ();

    instances = new vlist;

#ifdef	WIN32
    char	pathName[_MAX_PATH];
    sprintf (pathName, "\tis4000db\\snlvudb.%s",iocName);
#else
    char	pathName[MAXPATHLEN];
    sprintf (pathName, "/tis4000db/snlvudb.%s",iocName);
#endif	WIN32

    FILE *fd;
    if ((fd = fopen (pathName, "r")) == NULL)
    {
	return;
    }

    char inputStr[100];

    for (;;)
    {
	instance = (Instance *)(vmemAllocAndClear (sizeof (Instance)));
	if (!fgets (inputStr, PROGRAM_NAME_SIZE - 1, fd))
	{
	    vmemFree ((void *)instance);
	    break;
	}
	inputStr[strlen(inputStr) - 2] = 0;	// strip \n and :
	strcpy (inputStr, &inputStr[1]);	// strip leading :
	strcpy (instance -> programName, inputStr);

	if (!fgets (inputStr, PARAMETERS_SIZE - 1, fd))
	{
	    break;
	}
	inputStr[strlen(inputStr) - 2] = 0;	// strip \n and :
	strcpy (inputStr, &inputStr[1]);	// strip leading :
	strcpy (instance -> parameters, inputStr);

	if (!fgets (inputStr, RESOURCE_NAME_SIZE - 1, fd))
	{
	    break;
	}
	inputStr[strlen(inputStr) - 2] = 0;	// strip \n and :
	strcpy (inputStr, &inputStr[1]);	// strip leading :
	strcpy (instance -> resourceName, inputStr);

	instances -> SetCellValue (row, 0, (const void *)instance);
	row++;
    }
    fclose(fd);
}


/*******************************************************************
*
*    InstancesDialog function to update the instances file
*	File has form:
*		snlvudb.<iocname>
*
*******************************************************************/

void InstancesDialog::StoreInstances (void)
{
    Instance	*instance;
    int		row = 0;

    if (instances == NULL)
    {
	return;
    }
    if (instancesModified == vFALSE)
    {
	return;
    }

    snlvuApplication *app = mainDialog -> GetApplication ();
    char *iocName = app -> GetIocName ();

#ifdef	WIN32
    char	pathName[_MAX_PATH];
    sprintf (pathName, "\tis4000db\\snlvudb.%s",iocName);
#else
    char	pathName[MAXPATHLEN];
    sprintf (pathName, "/tis4000db/snlvudb.%s",iocName);
#endif	WIN32

    FILE *fd;
    if ((fd = fopen (pathName, "w")) == NULL)
    {
	return;
    }
    vlistIterator iterator;
    iterator.Start (instances);
    while (iterator.Next ())
    {
	instance = (Instance *)(iterator.GetValue ());
	fprintf (fd, ":%s:\n:%s:\n:%s:\n",
			instance -> programName,
			instance -> parameters,
			instance -> resourceName);
    }
    iterator.Finish ();
    fclose (fd);
    instancesModified = vFALSE;
}


/*******************************************************************
*
*    InstancesDialog function to update the instances list
*
*******************************************************************/

void InstancesDialog::UpdateInstancesList (void)
{
    instancesListView -> BuildList ();
}

/*******************************************************************
*
*    InstancesDialog function to look for duplicate programNames
*	with out a differance in the task name
*	parameter list has form:
*		"name=test1, input=321.123"
*		or:
*		"name=test2"
*	we want to warn user if they try to create more than 1
*	instance of a program with the same task name
*
*******************************************************************/

int InstancesDialog::VerifyInstance (Instance inst, int instRow)
{
    vlistIterator	iterator;
    int			retVal = vTRUE;
    Instance		*instVal;
    char		*cPtr1;
    char		*cPtr2;
    char		*cPtr;
    vstr		*progStr;
    int			row;

    if (instances == NULL)
    {
	return (retVal);
    }
    // iterator through our list of program instances
    iterator.Start (instances);
    while (iterator.Next ())
    {
	row = iterator.GetRow ();
	instVal = (Instance *)(iterator.GetValue ());
	if ((strcmp (instVal ->  programName, inst.programName)) == 0)
	{
	    if ((cPtr1 = strstr (instVal -> parameters, "name")) != NULL)
	    {
		if ((cPtr2 = strstr (inst.parameters, "name")) != NULL)
		{
		    char instValTaskName [TASK_NAME_SIZE + 1];
		    char instTaskName [TASK_NAME_SIZE + 1];

		    memset ((void *)instValTaskName, 0, sizeof (instValTaskName));
		    memset ((void *)instTaskName, 0, sizeof (instTaskName));

		    // skip to end of = sign in my list program
		    while (*cPtr1 != '=')
		    {
			cPtr1++;
		    }
		    cPtr1++;

		    // skip to end of = sign in new instance
		    while (*cPtr2 != '=')
		    {
			cPtr2++;
		    }
		    cPtr2++;

		    // get task names
		    cPtr = &instValTaskName[0];
		    while (*cPtr1)
		    {
			if ((*cPtr1 == ',') || (*cPtr1 == '"'))
			{
			    break;
			}
			if (*cPtr1 == ' ')
			{
			    cPtr1++;
			    continue;
			}
			*cPtr = *cPtr1;
			cPtr++;
			cPtr1++;
		    }
		    cPtr = &instTaskName[0];
		    while (*cPtr2)
		    {
			if ((*cPtr2 == ',') || (*cPtr2 == '"'))
			{
			    break;
			}
			if (*cPtr2 == ' ')
			{
			    cPtr2++;
			    continue;
			}
			*cPtr = *cPtr2;
			cPtr++;
			cPtr2++;
		    }
		    if ((strcmp (instValTaskName, instTaskName)) == 0)
		    {
			if (row != instRow) 
			{
			    progStr = vstrCloneScribed (vcharScribeSystem (instVal -> programName));
			    retVal = snlvuMessageCenter (this, SNLVU_CAUTION_NOTICE,
							snlvu_TaskNameWarning1Msg, progStr);
			    vstrDestroy (progStr);
			    break;
			}
		    }
		}
	    }
	    else
	    {
		// my list instance does not have a task name and either does the
		// new one, bad scene like above
		if ((cPtr2 = strstr (inst.parameters, "name")) == NULL)
		{
		    if (row != instRow)
		    {
			progStr = vstrCloneScribed (vcharScribeSystem (instVal -> programName));
			retVal = snlvuMessageCenter (this, SNLVU_CAUTION_NOTICE,
						    snlvu_TaskNameWarning2Msg, progStr);
			vstrDestroy (progStr);
			break;
		    }
		}
	    }
	}
    }
    iterator.Finish ();

    return (retVal);
}


/*******************************************************************
*
*    InstancesDialog Instance helper functions for the instances
*	list
*
*******************************************************************/

void InstancesDialog::AddInstance (Instance inst)
{
    int row = instances -> GetRowCount ();
    Instance *instance = (Instance *)(vmemAllocAndClear (sizeof (Instance)));
    *instance = inst;
    instances -> SetCellValue (row, 0, (const void *)instance);
    instances -> Sort (vlistROW, 0, InstancesCompare);
    instancesModified = vTRUE;
}

void InstancesDialog::DeleteInstance (int row)
{
    Instance *instance = (Instance *)(instances -> GetCellValue (row, 0));
    vmemFree ((void *)instance);
    instances -> RemoveRow (row);
    instancesModified = vTRUE;
}

void InstancesDialog::GetInstance (int row, Instance *inst)
{
    Instance *instance = (Instance *)(instances -> GetCellValue (row, 0));
    *inst = *instance;
}

void InstancesDialog::UpdateInstance (int row, Instance *inst)
{
    Instance *instance = (Instance *)(instances -> GetCellValue (row, 0));
    *instance = *inst;
    instancesModified = vTRUE;
}


static int InstancesCompare (const void *first, const void *second)
{
    return (strcmp (((Instance *)first) -> programName, 
		    ((Instance *)second) -> programName));
}



vkindTYPED_FULL_DEFINITION(InstancesListView, vlistview, NULL);

/*******************************************************************
*
*    InstancesListView
*
*******************************************************************/

void InstancesListView::BuildList (void)
{
    int		row = 0;

    vliststr *newList = new vliststr;
    vliststr *oldList = (vliststr *)(GetList ());
    vlist *instances = instancesDialog -> GetInstancesList ();
    Instance	*instance;
    vlistIterator iterator;

    iterator.Start (instances);
    while (iterator.Next())
    {
	instance = (Instance *)(iterator.GetValue ());
	newList -> SetCellValueScribed (row, 0, 
				vcharScribeSystem (instance -> programName));
	newList -> SetCellValueScribed (row, 1, 
				vcharScribeSystem (instance -> parameters));
	newList -> SetCellValueScribed (row, 2, 
				vcharScribeSystem (instance -> resourceName));
	row++;
    }
    iterator.Finish ();

    SetList (newList);
    if (oldList != NULL)
    {
	delete oldList;
    }
    int rowCount = newList -> GetRowCount ();
    if (rowCount > 0)
    {
	newList -> Sort (vlistROW, 0, NULL);
	SetColumnsWidths ();
	vlistSelection *selected = GetSelection ();
	newList -> SetSelectMethod (selected, vlistONE_ONLY_SELECTION);
	newList -> SetExtendRowSelection (selected, vTRUE);
    }
}

/*******************************************************************
*
*    InstancesListView, set list column widths based on main dialog's
*	ruler item widths
*
*******************************************************************/

void InstancesListView::SetColumnsWidths (void)
{
    vruler	*instancesRuler = instancesDialog -> GetInstancesRuler ();
    vrulerItem	*item;
    int		itemWidth;
    int		itemCount = instancesRuler -> GetItemCount ();
    int		i;

    for (i = 0; i < itemCount; i++)
    {
	item = instancesRuler -> GetItemAt (i);
	itemWidth = item -> GetWidth ();
	SetColumnWidth (i, itemWidth);
    }
}


vkindTYPED_FULL_DEFINITION(InstancesStartButton, vbutton, NULL);

/*******************************************************************
*
*    InstancesStartButton
*
*******************************************************************/

void InstancesStartButton::ObserveDialogItem (vevent *event)
{
    InstancesListView	*instancesListView = instancesDialog -> GetInstancesListView ();
    vlistSelection	*selection = instancesListView -> GetSelection ();
    vliststr		*list = (vliststr *)(instancesListView -> GetList ());

    if (list == NULL)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	return;
    }

    int rowCount = list -> GetRowCount ();

    if (rowCount == 0)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	return;
    }
    if ((list -> IsSelectionEmpty (selection)) == vTRUE)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_NothingSelectedMsg);
	return;
    }
    int row, col;
    list -> GetSelectedCell (selection, &row, &col);

    char actualName [MODULE_NAME_SIZE + 1];
    Instance inst;

    instancesDialog -> GetInstance (row, &inst);

    if (!snlMon_Start (inst.programName, inst.parameters, 0, inst.resourceName, actualName))
    {
	int res = snlvuMessageCenter (instancesDialog, SNLVU_CAUTION_NOTICE, 
				      snlvu_LoadItMsg);
	if (res == vTRUE)
	{
	    char temptxt[100];

	    MainDialog *mainDialog = instancesDialog -> GetMainDialog ();
	    snlvuApplication *app = mainDialog -> GetApplication ();
	    vresource appRes = app -> GetSnlvuResource ();
	    vresource enterRes = vresourceGet (appRes, snlvu_EnterDialog);
	    EnterDialog *enterDialog = new EnterDialog (enterRes);
	    strcpy (temptxt, inst.programName);
	    if ((strstr (temptxt, ".o")) == NULL)
	    {
		strcat (temptxt, ".o");
	    }
	    enterDialog -> OpenEnter (mainDialog, ENTER_MODULE_NAME, temptxt);
	    vchar *enteredData = enterDialog -> GetEnteredData ();
	    vcharExportBounded (vcharScribeSystemExport (enteredData),
				temptxt,
				sizeof (temptxt));
	    temptxt[sizeof(temptxt) - 1] = 0;
	    delete (enterDialog);

	    if (!snlMon_Load (temptxt))
	    {
		snlvuMessageCenter (instancesDialog, SNLVU_INFO_NOTICE, 
				    snlvu_ModuleNotLoadedMsg, temptxt);
	    }
	    else
	    {
		if (!snlMon_Start (inst.programName, inst.parameters, 0, 
						inst.resourceName, actualName))
		{
		    snlvuMessageCenter (instancesDialog, SNLVU_INFO_NOTICE,
						snlvu_StartFailedMsg, inst.programName);
		}
	    }
	}
    }
}


vkindTYPED_FULL_DEFINITION(InstancesDefineButton, vbutton, NULL);

/*******************************************************************
*
*    InstancesDefineButton
*
*******************************************************************/

void InstancesDefineButton::ObserveDialogItem (vevent *event)
{
    MainDialog *mainDialog = instancesDialog -> GetMainDialog ();
    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource appRes = app -> GetSnlvuResource ();
    vresource edinstRes = vresourceGet (appRes, snlvu_EditInstanceDialog);
    EditInstanceDialog *editInstanceDialog = new EditInstanceDialog (edinstRes);
    editInstanceDialog -> OpenEditInstance (mainDialog, DEFINE_NEW_INSTANCE, NULL);
    Instance newInst;
    editInstanceDialog -> GetInstanceData (&newInst);
    delete editInstanceDialog;
    if (newInst.programName[0] == 0)
    {
	return;
    }
    if ((instancesDialog -> VerifyInstance (newInst, -1)) == vFALSE)
    {
	return;
    }
    instancesDialog -> AddInstance (newInst);
    instancesDialog -> UpdateInstancesList ();
    instancesDialog -> StoreInstances ();
}


vkindTYPED_FULL_DEFINITION(InstancesDeleteButton, vbutton, NULL);

/*******************************************************************
*
*    InstancesDeleteButton
*
*******************************************************************/

void InstancesDeleteButton::ObserveDialogItem (vevent *event)
{
    InstancesListView	*instancesListView = instancesDialog -> GetInstancesListView ();
    vlistSelection	*selection = instancesListView -> GetSelection ();
    vliststr		*list = (vliststr *)(instancesListView -> GetList ());

    if (list == NULL)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	return;
    }

    int rowCount = list -> GetRowCount ();

    if (rowCount == 0)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	return;
    }
    if ((list -> IsSelectionEmpty (selection)) == vTRUE)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_NothingSelectedMsg);
	return;
    }
    int row, col;
    list -> GetSelectedCell (selection, &row, &col);
    list -> RemoveRow (row);
    instancesDialog -> DeleteInstance (row);
    instancesDialog -> UpdateInstancesList ();
    instancesDialog -> StoreInstances ();
}


vkindTYPED_FULL_DEFINITION(InstancesEditButton, vbutton, NULL);

/*******************************************************************
*
*    InstancesEditButton
*
*******************************************************************/

void InstancesEditButton::ObserveDialogItem (vevent *event)
{
    InstancesListView	*instancesListView = instancesDialog -> GetInstancesListView ();
    vlistSelection	*selection = instancesListView -> GetSelection ();
    vliststr		*list = (vliststr *)(instancesListView -> GetList ());

    if (list == NULL)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	return;
    }

    int rowCount = list -> GetRowCount ();

    if (rowCount == 0)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	return;
    }
    if ((list -> IsSelectionEmpty (selection)) == vTRUE)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_NothingSelectedMsg);
	return;
    }
    int row, col;
    list -> GetSelectedCell (selection, &row, &col);

    Instance inst;
    instancesDialog -> GetInstance (row, &inst);

    MainDialog *mainDialog = instancesDialog -> GetMainDialog ();
    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource appRes = app -> GetSnlvuResource ();
    vresource edinstRes = vresourceGet (appRes, snlvu_EditInstanceDialog);
    EditInstanceDialog *editInstanceDialog = new EditInstanceDialog (edinstRes);
    editInstanceDialog -> OpenEditInstance (mainDialog, EDIT_INSTANCE, &inst);
    editInstanceDialog -> GetInstanceData (&inst);
    delete editInstanceDialog;
    if (inst.programName[0] == 0)
    {
	return;
    }
    if ((instancesDialog -> VerifyInstance (inst, row)) == vFALSE)
    {
	return;
    }
    instancesDialog -> UpdateInstance (row, &inst);
    instancesDialog -> UpdateInstancesList ();
    instancesDialog -> StoreInstances ();
}


vkindTYPED_FULL_DEFINITION(InstancesExitButton, vbutton, NULL);

/*******************************************************************
*
*    InstancesExitButton
*
*******************************************************************/

void InstancesExitButton::ObserveDialogItem (vevent *event)
{
    instancesDialog -> StoreInstances ();
    // stop secondary event loop started in open()
    vevent::StopProcessing ();
}

