#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vbuttonHEADER
#include vstartupHEADER
#include vcontrolHEADER
#include vtimerHEADER
#include vprefHEADER
#include vlistviewHEADER
#include vrulerHEADER
#include vlistitemHEADER

extern "C"
{
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "OperPriv.h"
}
#include "snlvu.h"
#include "snlvuApp.hxx"
#include "snlvuMain.hxx"
#include "snlvuChannels.hxx"
#include "snlvuModules.hxx"
#include "snlvuVariables.hxx"
#include "snlvuExpand.hxx"
#include "snlvuEnter.hxx"
#include "snlvuStateSets.hxx"
#include "snlvuInstances.hxx"
#include "snlvuUtils.hxx"


vkindTYPED_FULL_DEFINITION(MainDialog, vdialog, NULL);

/*******************************************************************
*
*    MainDialog override for constructors init ()
*
*******************************************************************/

void MainDialog::LoadInit (vresource res)
{
    int		i;

    channelsDialog = NULL;
    modulesDialog = NULL;
    variablesDialog = NULL;
    expandDialog = NULL;
    stateSetsDialog = NULL;

    // store the main dialog's resource
    mainDialogResource = res;

    // init stuff
    tisAbout	 = NULL;
    loadInitErrors = SNLVU_SUCCESS;
    PrevPrograms = NULL;
    PrevNumProg = 0;

    // init privileges
    loadModules = vTRUE;
    unloadModules = vTRUE;
    modifyVariables = vTRUE;
    pauseStateSets = vTRUE;
    continueStateSets = vTRUE;
    startInstance = vTRUE;
    defineInstance = vTRUE;
    deleteInstance = vTRUE;
    editInstance = vTRUE;

    // RequestSubLoad returns ptr to uninited dialog object
    // gets init'd when dialog gets init'd by LoadInit()
    vmenubar *menubar = vmenubar::RequestSubLoad (this,
					snlvu_MainMenubar);
    // File menu
    vmenu *fileMenu = vmenu::RequestSubLoad (menubar,
					snlvu_FileMenu);
    vmenuItem *exitItem = vmenuItem::RequestSubLoad (fileMenu,
				    	snlvu_ExitMenuitem);
    // Help menu
    helpMenu = vmenu::RequestSubLoad (menubar,
					snlvu_HelpMenu);
    // ruler header
    mainRuler = vruler::RequestSubLoad (this, snlvu_MainDialogRuler);

    mainListView = MainListView::RequestSubLoad (this, snlvu_MainListView);

    modulesButton = ModulesButton::RequestSubLoad (this, snlvu_ModulesButton);

    stopButton = StopButton::RequestSubLoad (this, snlvu_StopButton);

    channelsButton = ChannelsButton::RequestSubLoad (this, snlvu_ChannelsButton);

    variablesButton = VariablesButton::RequestSubLoad (this, snlvu_VariablesButton);

    selectButton = SelectButton::RequestSubLoad (this, snlvu_SelectButton);

    stateSetsButton = StateSetsButton::RequestSubLoad (this, snlvu_StateSetsButton);

    instancesButton = InstancesButton::RequestSubLoad (this, snlvu_InstancesButton);

    // init inherited data
    vdialog::LoadInit (res);

    // store ptrs to main dialog
    mainListView -> StoreMainDialog (this);
    modulesButton -> StoreMainDialog (this);
    stopButton -> StoreMainDialog (this);
    channelsButton -> StoreMainDialog (this);
    variablesButton -> StoreMainDialog (this);
    selectButton -> StoreMainDialog (this);
    stateSetsButton -> StoreMainDialog (this);
    instancesButton -> StoreMainDialog (this);

    // init privilege checking, see .../share/src/libtis/OperPrvtrk.c
    if ((InitOperPriv (NULL)) != PRIVOK)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_InitOperPrivErr);
	loadInitErrors = SNLVU_ERROR;
	return;
    }
    // check on user privileges to snlvu functions

    if ((OperPriv (NULL, "*MMI_SNLVU_MODULES_LOAD")) == 0)
    {
	loadModules = vFALSE;
    }

    if ((OperPriv (NULL, "*MMI_SNLVU_MODULES_UNLOAD")) == 0)
    {
	unloadModules = vFALSE;
    }

    if ((OperPriv (NULL, "*MMI_SNLVU_STOP_STOP")) == 0)
    {
	stopButton -> Disable ();
    }

    if ((OperPriv (NULL, "*MMI_SNLVU_VARIABLES_MODIFY")) == 0)
    {
	modifyVariables = vFALSE;
    }

    if ((OperPriv (NULL, "*MMI_SNLVU_SELECT_IOC")) == 0)
    {
	selectButton -> Disable ();
    }

    if ((OperPriv (NULL, "*MMI_SNLVU_STATESETS_PAUSE")) == 0)
    {
	pauseStateSets = vFALSE;
    }

    if ((OperPriv (NULL, "*MMI_SNLVU_STATESETS_CONTINUE")) == 0)
    {
	continueStateSets = vFALSE;
    }

    if ((OperPriv (NULL, "*MMI_SNLVU_INSTANCES_START")) == 0)
    {
	startInstance = vFALSE;
    }

    if ((OperPriv (NULL, "*MMI_SNLVU_INSTANCES_DEFINE")) == 0)
    {
	defineInstance = vFALSE;
    }

    if ((OperPriv (NULL, "*MMI_SNLVU_INSTANCES_DELETE")) == 0)
    {
	deleteInstance = vFALSE;
    }

    if ((OperPriv (NULL, "*MMI_SNLVU_INSTANCES_EDIT")) == 0)
    {
	editInstance = vFALSE;
    }

    // close privilege checking, see .../share/src/libtis/OperPrvtrk.c
    CloseOperPriv ();

    // store internal names for the main menu tags
    commandNames[EXIT_MENUITEM]   = snlvu_ExitMenuitem;

    // create a command space
    vcommandSpace *space = new vcommandSpace;

    // create a dictionary with a default hash size and hash function
    vdict *dict = new vdict (vnameHash);

    // populate our dictionary with menu commands
    for (i = EXIT_MENUITEM; i < MENUITEM_COUNT; i++)
    {
	// get an instance of our subclassed selector command
	MainCommand *command = new MainCommand;
	// set the commands selector value
	command -> SetSelector(i);
	// store command in the dictionary using the 
	// intern'd tag name
	dict -> SetElem (commandNames[i], command);
    }
    // add dictionary to the command space and give the command
    // space ownership of the dictionary, vname_Command defines
    // the type attribute of the things in the dictionary
    space -> AddDictOwned (vname_Command, dict);

    SetSpaceOwned (space);
    Place (GetRoot (), vrectPLACE_LEFT, vrectPLACE_TOP);
}


/*******************************************************************
*
*    MainDialog Destroy function
*
*******************************************************************/

void MainDialog::Destroy (void)
{
    if (tisAbout != NULL)
    {
	delete tisAbout;
    }
    if (PrevPrograms != NULL)
    {
	free ((void *)PrevPrograms);
    }
    if (modulesDialog != NULL)
    {
	delete modulesDialog;
    }
    if (channelsDialog != NULL)
    {
	delete channelsDialog;
    }
    if (variablesDialog != NULL)
    {
	delete variablesDialog;
    }
    if (expandDialog != NULL)
    {
	delete expandDialog;
    }
    if (stateSetsDialog != NULL)
    {
	delete stateSetsDialog;
    }
    if (IsOpen ())
    {
	Close ();
    }
    vdialog::Destroy ();
}


/*******************************************************************
*
*    MainDialog observe window event, looking for exit
*
*******************************************************************/

void MainDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
	vevent::StopProcessing ();
}


/*******************************************************************
*
*    MainDialog observe dialog reconfigure, looking for a 
*	resize event. we want to resize the ruler and its components
*	also
*
*******************************************************************/

int MainDialog::HandleConfigure (vevent *event)
{
    vwindow::HandleConfigure (event);
    if ( event -> GetType () == veventWINDOW_CONFIGURE && event -> IsSize () )
    {
	int		rulerWidth;
	int		itemCount;
	int		totalItemWidth = 0;
	int		itemWidth;
	int		i;
	vrulerItem	*item;
	const vrect	*rulerRect;
	vlist *list = mainListView -> GetList ();

	rulerRect = mainRuler -> GetRect ();
	rulerWidth = rulerRect -> GetWidth ();
	itemCount = mainRuler -> GetItemCount ();
	for (i = 0; i < itemCount; i++)
	{
	    item = mainRuler -> GetItemAt (i);
	    itemWidth = item -> GetWidth ();
	    totalItemWidth += itemWidth;
	}
	if (totalItemWidth < rulerWidth)
	{
	    float	itemWidth;
	    float	itemWidthPct;
	    int		newItemWidth;
	    int		remainingWidth = rulerWidth;

	    for (i = 0; i < itemCount; i++)
	    {
		item = mainRuler -> GetItemAt (i);
		itemWidth = (float)(item -> GetWidth ());
		itemWidthPct = itemWidth / (float)totalItemWidth;
		newItemWidth = (int)(itemWidthPct * rulerWidth);
		if ((remainingWidth - newItemWidth) <= 0)
		{
		    newItemWidth = remainingWidth;
		}
		item -> SetWidth (newItemWidth);
		remainingWidth -= newItemWidth;
	    }
	    // update the listviews columns widths
	    mainListView -> SetColumnsWidths ();
	}

    }
    return (vTRUE);
}


/*******************************************************************
*
*    MainDialog 
*
*******************************************************************/

void MainDialog::OpenMain (snlvuApplication *App)
{
    snlvuApp = App;

    BuildAboutMenu ();

    UpdateProgramList (0);

    // add so snlvu will iconize in toolbar under NT
    AddStyle (vwindowSTYLE_GLOBAL);
    vdialog::Open ();
}


/*******************************************************************
*
*    MainDialog function to build the about menu and attach
*    it to the help menu
*
*******************************************************************/

void MainDialog::BuildAboutMenu (void)
{
    vstr *verMsg = vresourceGetString (snlvuApp -> GetMessagesResource(),
				       snlvu_VersionMsg);
    tisAbout = new TisAbout (snlvuApp -> GetGalaxyRes (), verMsg);
    vstrDestroy (verMsg);
    vstr *aboutTitle = vresourceGetString (snlvuApp -> GetTitlesResource (),
				      snlvu_AboutTitle);
    tisAbout -> SetAboutTitle (aboutTitle);
    vstrDestroy (aboutTitle);
    AboutMenuItem *aboutItem = tisAbout -> GetAboutMenuItem ();
    helpMenu -> AppendItem (aboutItem);
}


/*******************************************************************
*
*    MainDialog function to set the dialog title based on whether
*	we are still connected to the IOC
*
*******************************************************************/

void MainDialog::SetMainTitle (ConnectStatus connected)
{
    char *iocName = snlvuApp -> GetIocName ();
    vresource titleRes = snlvuApp -> GetTitlesResource ();
    vstr *titleFormat;
    if (connected == SNLVU_CONNECTED)
    {
	titleFormat = vresourceGetString (titleRes, snlvu_MainTitle);
    }
    else
    {
	titleFormat = vresourceGetString (titleRes, snlvu_MainTitleDisconnected);
    }
    vstr *iocname = vstrCloneScribed (vcharScribeSystem (iocName));
    vstr *mainTitle = vstrCloneScribed (vcharScribeFormat (titleFormat, iocname));
    SetTitle (mainTitle);
    vstrDestroy (titleFormat);
    vstrDestroy (iocname);
    vstrDestroy (mainTitle);
}


/*******************************************************************
*
*    MainDialog function to update the programs list in the main
*	view
*
*******************************************************************/

void MainDialog::UpdateProgramList (int ignorePrev)
{
    if ((snlvuApp -> IsConnected ()) == 0)
    {
	SetMainTitle (SNLVU_NOTCONNECTED);
	return;
    }
    SetMainTitle (SNLVU_CONNECTED);

    REPPROG_REPLY *programs;
    int		  numPrograms;

    snlMon_ReportPrograms (&programs, &numPrograms);

    // check the list against the previous list for a change
    if (programs && PrevPrograms && !ignorePrev)
    {
	// if no change, free the returned list and get out
	if ( (PrevNumProg == numPrograms) &&
	     (memcmp ((const void *)programs, (const void *)PrevPrograms,
		      sizeof (REPPROG_REPLY) * numPrograms) == 0))
	{
	    free ((void *)programs);
	    return;
	}
    }
    // the list has changed, free the previous list, and set current
    if (PrevPrograms != NULL)
    {
	free ((void *)PrevPrograms);
	PrevPrograms = NULL;
    }
    // remember the new stuff
    PrevPrograms = programs;
    PrevNumProg = numPrograms;

    // build list for listview
    mainListView -> BuildList (programs, numPrograms);
}


vkindTYPED_FULL_DEFINITION(MainCommand, vcommandSelector, NULL);

/*******************************************************************
*
*    MainCommand menu selector commands
*
*******************************************************************/

int MainCommand::Issue (vdict *dict)
{
    switch (GetSelector ())
    {
	case EXIT_MENUITEM:
	    vevent::StopProcessing ();
	    break;

	default:
	    return (vFALSE);
    }
    return (vTRUE);
}


vkindTYPED_FULL_DEFINITION(MainListView, vlistview, NULL);

/*******************************************************************
*
*    MainListView 
*
*******************************************************************/

void MainListView::Destroy (void)
{
    vlist *list = GetList ();
    if (list != NULL)
    {
	SetList (NULL);
	delete list;
    }
    vlistview::Destroy ();
}


/*******************************************************************
*
*    MainListView, build the list
*
*******************************************************************/

void MainListView::BuildList (REPPROG_REPLY *programs, int numProgs)
{
    REPPROG_REPLY	*p = programs;
    int			i;
    int			row = 0;

    vlist *newList = new vlist;
    vlist *oldList = GetList ();

    // each cell has same value, the p ptr, what gets printed in each
    // cell is determined in the DrawCellContents override
    for (i = 0; i < numProgs; i++, p++)
    {

	newList -> SetCellValue (row, 0, (const void *)p);
	newList -> SetCellValue (row, 1, (const void *)p);
	newList -> SetCellValue (row, 2, (const void *)p);
	newList -> SetCellValue (row, 3, (const void *)p);
	row++;
    }
    if (oldList != NULL)
    {
	delete oldList;
    }
    int rowCount = newList -> GetRowCount ();
    if (rowCount > 0)
    {
	newList -> Sort (vlistROW, 0, NULL);
    }
    SetList (newList);
    if (rowCount > 0)
    {
	SetColumnsWidths ();
	vlistSelection *selected = GetSelection ();
	newList -> SetSelectMethod (selected, vlistONE_ONLY_SELECTION);
	newList -> SetExtendRowSelection (selected, vTRUE);
    }
}


/*******************************************************************
*
*    MainListView, override 
*
*******************************************************************/

void MainListView::DrawCellContents (int row, int col, int selected, vrect *cellRect)
{
    REPPROG_REPLY	*p;
    char		*cPtr;
    vlist		*viewList = (vlist *)GetList ();

    vdraw::GSave ();
    vdraw::RectsClip (cellRect, 1);

    p = (REPPROG_REPLY *)(viewList -> GetCellValue (row, col));

    cPtr = " ";

    if (col == 0)
    {
	if (strlen (p -> ProgramName))
	{
	    cPtr = p -> ProgramName;
	}
	vdraw::ScribedRectShowCenterY (vcharScribeSystem (cPtr), cellRect);
    }
    else if (col == 1)
    {
	if (strlen (p -> Parameters))
	{
	    cPtr = p -> Parameters;
	}
	vdraw::ScribedRectShowCenterY (vcharScribeSystem (cPtr), cellRect);
    }
    else if (col == 2)
    {
	if (strlen (p -> Resource))
	{
	    cPtr = p -> Resource;
	}
	vdraw::ScribedRectShowCenterY (vcharScribeSystem (cPtr), cellRect);
    }
    else
    {
	if (p -> Stat == REPPROG_PENDING)
	{
	    cPtr = "Pending";
	}
	else if (p -> Stat == REPPROG_STARTED)
	{
	    cPtr = "Started";
	}
	else
	{
	    cPtr = "ChWait";
	}
	vdraw::ScribedRectShowCenterY (vcharScribeSystem (cPtr), cellRect);
    }

    vdraw::GRestore ();
}


/*******************************************************************
*
*    MainListView, set list column widths based on main dialog's
*	ruler item widths
*
*******************************************************************/

void MainListView::SetColumnsWidths (void)
{
    vruler	*mainRuler = mainDialog -> GetMainRuler ();
    vrulerItem	*item;
    int		itemWidth;
    int		itemCount = mainRuler -> GetItemCount ();
    int		i;

    for (i = 0; i < itemCount; i++)
    {
	item = mainRuler -> GetItemAt (i);
	itemWidth = item -> GetWidth ();
	SetColumnWidth (i, itemWidth);
    }
}


vkindTYPED_FULL_DEFINITION(ModulesButton, vbutton, NULL);

/*******************************************************************
*
*    ModulesButton
*	call up the modules dialog
*
*******************************************************************/

void ModulesButton::ObserveDialogItem (vevent *event)
{
    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource appRes = app -> GetSnlvuResource ();
    vresource modRes = vresourceGet (appRes, snlvu_ModulesDialog);
    mainDialog -> modulesDialog = new ModulesDialog (modRes);
    mainDialog -> modulesDialog -> OpenModules (mainDialog);
    delete mainDialog -> modulesDialog;
    mainDialog -> modulesDialog = NULL;

    mainDialog -> UpdateProgramList (0);
}


vkindTYPED_FULL_DEFINITION(StopButton, vbutton, NULL);

/*******************************************************************
*
*    StopButton 
*	the snlMon_Kill will do a taskDelete of the selected
*	snl program in the IOC
*
*******************************************************************/

void StopButton::ObserveDialogItem (vevent *event)
{
    REPPROG_REPLY	*p;
    MainListView	*mainListView = mainDialog -> GetMainListView ();
    vlistSelection	*selection = mainListView -> GetSelection ();
    vlist		*list = mainListView -> GetList ();

    if (list == NULL)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	return;
    }

    int rowCount = list -> GetRowCount ();

    if (rowCount == 0)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	return;
    }
    if ((list -> IsSelectionEmpty (selection)) == vTRUE)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_NothingSelectedMsg);
	return;
    }
    int row, col;
    list -> GetSelectedCell (selection, &row, &col);
    p = (REPPROG_REPLY *)(list -> GetCellValue (row, 0));
    snlMon_Kill (p -> ProgramName, p -> TaskId, p -> Parameters);
    mainDialog -> UpdateProgramList (0);
}


vkindTYPED_FULL_DEFINITION(ChannelsButton, vbutton, NULL);

/*******************************************************************
*
*    ChannelsButton
*	call up the channels dialog
*
*******************************************************************/

void ChannelsButton::ObserveDialogItem (vevent *event)
{
    REPPROG_REPLY	*p;
    MainListView	*mainListView = mainDialog -> GetMainListView ();
    vlistSelection	*selection = mainListView -> GetSelection ();
    vlist		*list = mainListView -> GetList ();

    if (mainDialog -> channelsDialog != NULL)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_ChannelsShownMsg);
	return;
    }

    if (list == NULL)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	return;
    }

    int rowCount = list -> GetRowCount ();

    if (rowCount == 0)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	return;
    }
    if ((list -> IsSelectionEmpty (selection)) == vTRUE)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_NothingSelectedMsg);
	return;
    }
    int row, col;
    list -> GetSelectedCell (selection, &row, &col);
    p = (REPPROG_REPLY *)(list -> GetCellValue (row, 0));

    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource appRes = app -> GetSnlvuResource ();
    vresource chanRes = vresourceGet (appRes, snlvu_ChannelsDialog);
    mainDialog -> channelsDialog = new ChannelsDialog (chanRes);
    mainDialog -> channelsDialog -> OpenChannels (mainDialog,
						  p -> ProgramName,
						  p -> TaskId);
    delete mainDialog -> channelsDialog;
    mainDialog -> channelsDialog = NULL;

    mainDialog -> UpdateProgramList (0);
}


vkindTYPED_FULL_DEFINITION(VariablesButton, vbutton, NULL);

/*******************************************************************
*
*    VariablesButton
*	call up the variables dialog
*
*******************************************************************/

void VariablesButton::ObserveDialogItem (vevent *event)
{
    REPPROG_REPLY	*p;
    MainListView	*mainListView = mainDialog -> GetMainListView ();
    vlistSelection	*selection = mainListView -> GetSelection ();
    vlist		*list = mainListView -> GetList ();

    if (mainDialog -> variablesDialog != NULL)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_VariablesShownMsg);
	return;
    }

    if (list == NULL)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	return;
    }

    int rowCount = list -> GetRowCount ();

    if (rowCount == 0)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	return;
    }
    if ((list -> IsSelectionEmpty (selection)) == vTRUE)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_NothingSelectedMsg);
	return;
    }
    int row, col;
    list -> GetSelectedCell (selection, &row, &col);
    p = (REPPROG_REPLY *)(list -> GetCellValue (row, 0));

    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource appRes = app -> GetSnlvuResource ();
    vresource varRes = vresourceGet (appRes, snlvu_VariablesDialog);
    mainDialog -> variablesDialog = new VariablesDialog (varRes);
    mainDialog -> variablesDialog -> OpenVariables (mainDialog,
						    p -> ProgramName,
						    p -> TaskId);
    delete mainDialog -> variablesDialog;
    mainDialog -> variablesDialog = NULL;

    mainDialog -> UpdateProgramList (0);
}



vkindTYPED_FULL_DEFINITION(SelectButton, vbutton, NULL);

/*******************************************************************
*
*    SelectButton
*	select a new IOC
*
*******************************************************************/

void SelectButton::ObserveDialogItem (vevent *event)
{
    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource appRes = app -> GetSnlvuResource ();
    vresource enterRes = vresourceGet (appRes, snlvu_EnterDialog);
    EnterDialog *enterDialog = new EnterDialog (enterRes);
    enterDialog -> OpenEnter (mainDialog, NEW_IOC, NULL);
    vchar *enteredData = enterDialog -> GetEnteredData ();
    char newValue[100];
    vcharExportBounded (vcharScribeSystemExport (enteredData),
			newValue,
			sizeof (newValue));
    newValue[sizeof(newValue) - 1] = 0;
    delete (enterDialog);

    char *q = newValue;
    // remove leading blanks
    while (*q)
    {
	if (*q != ' ')
	    break;
	q++;
    }

    // remove any trailing spaces
    while (*(q + strlen (q) - 1) == ' ')
    {
	*(q + strlen (q) - 1) = 0;
    }

    if ((strlen (q)) == 0)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_NoIocErr);
	return;
    }
    snlMonDisconnect ();
    app -> StoreIocName (q);
    app -> Connect ();
    mainDialog -> UpdateProgramList (0);
}



vkindTYPED_FULL_DEFINITION(StateSetsButton, vbutton, NULL);

/*******************************************************************
*
*    StateSetsButton
*
*******************************************************************/

void StateSetsButton::ObserveDialogItem (vevent *event)
{
    REPPROG_REPLY	*p;
    MainListView	*mainListView = mainDialog -> GetMainListView ();
    vlistSelection	*selection = mainListView -> GetSelection ();
    vlist		*list = mainListView -> GetList ();

    if (mainDialog -> stateSetsDialog != NULL)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_StateSetsShownMsg);
	return;
    }

    if (list == NULL)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	return;
    }

    int rowCount = list -> GetRowCount ();

    if (rowCount == 0)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	return;
    }
    if ((list -> IsSelectionEmpty (selection)) == vTRUE)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_NothingSelectedMsg);
	return;
    }
    int row, col;
    list -> GetSelectedCell (selection, &row, &col);
    p = (REPPROG_REPLY *)(list -> GetCellValue (row, 0));

    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource appRes = app -> GetSnlvuResource ();
    vresource ssRes = vresourceGet (appRes, snlvu_StateSetsDialog);
    mainDialog -> stateSetsDialog = new StateSetsDialog (ssRes);
    mainDialog -> stateSetsDialog -> OpenStateSets (mainDialog,
						    p -> ProgramName,
						    p -> TaskId);
    delete mainDialog -> stateSetsDialog;
    mainDialog -> stateSetsDialog = NULL;

    mainDialog -> UpdateProgramList (0);
}



vkindTYPED_FULL_DEFINITION(InstancesButton, vbutton, NULL);

/*******************************************************************
*
*    InstancesButton
*
*******************************************************************/

void InstancesButton::ObserveDialogItem (vevent *event)
{
    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource appRes = app -> GetSnlvuResource ();
    vresource instRes = vresourceGet (appRes, snlvu_InstancesDialog);
    InstancesDialog *instDlg = new InstancesDialog (instRes);
    instDlg -> OpenInstances (mainDialog);
    delete instDlg;

    mainDialog -> UpdateProgramList (0);
}
