#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vbuttonHEADER
#include vstartupHEADER
#include vcontrolHEADER
#include vtimerHEADER
#include vprefHEADER
#include vlistviewHEADER
#include vrulerHEADER

extern "C"
{
#include "usrseq.h"
}

#include "snlvu.h"
#include "snlvuApp.hxx"
#include "snlvuMain.hxx"
#include "snlvuEnter.hxx"
#include "snlvuVariables.hxx"
#include "snlvuExpand.hxx"
#include "snlvuUtils.hxx"


vkindTYPED_FULL_DEFINITION(VariablesDialog, vdialog, NULL);

/*******************************************************************
*
*    VariablesDialog override for constructors init ()
*
*******************************************************************/

void VariablesDialog::LoadInit (vresource res)
{
    // init stuff
    PrevVariables = NULL;
    InUse = vFALSE;

    // RequestSubLoad returns ptr to uninited dialog object
    // gets init'd when dialog gets init'd by LoadInit()

    // ruler header
    variablesRuler = vruler::RequestSubLoad (this, snlvu_VariablesRuler);

    variablesListView = VariablesListView::RequestSubLoad (this, snlvu_VariablesListView);

    modifyButton = VariablesModifyButton::RequestSubLoad (this, snlvu_ModifyButton);

    expandButton = VariablesExpandButton::RequestSubLoad (this, snlvu_ExpandButton);

    exitButton = VariablesExitButton::RequestSubLoad (this, snlvu_ExitButton);

    // init inherited data
    vdialog::LoadInit (res);

    // store ptrs to main dialog
    variablesListView -> StoreVariablesDialog (this);
    modifyButton -> StoreVariablesDialog (this);
    expandButton -> StoreVariablesDialog (this);
    exitButton -> StoreVariablesDialog (this);

    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
}


/*******************************************************************
*
*    VariablesDialog Destroy function
*
*******************************************************************/

void VariablesDialog::Destroy (void)
{
    if (PrevVariables != NULL)
    {
	free ((void *)PrevVariables);
    }
    if (IsOpen ())
    {
	Close ();
    }
    vdialog::Destroy ();
}

/*******************************************************************
*
*    VariablesDialog observe window event, looking for exit
*
*******************************************************************/

void VariablesDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	InUse = vTRUE;
	vevent::StopProcessing ();
    }
}


/*******************************************************************
*
*    VariablesDialog observe dialog reconfigure, looking for a 
*	resize event. we want to resize the ruler and its components
*	also
*
*******************************************************************/

int VariablesDialog::HandleConfigure (vevent *event)
{
    vwindow::HandleConfigure (event);
    if ( event -> GetType () == veventWINDOW_CONFIGURE && event -> IsSize () )
    {
	int		rulerWidth;
	int		itemCount;
	int		totalItemWidth = 0;
	int		itemWidth;
	int		i;
	vrulerItem	*item;
	const vrect	*rulerRect;
	vlist *list = variablesListView -> GetList ();

	rulerRect = variablesRuler -> GetRect ();
	rulerWidth = rulerRect -> GetWidth ();
	itemCount = variablesRuler -> GetItemCount ();
	for (i = 0; i < itemCount; i++)
	{
	    item = variablesRuler -> GetItemAt (i);
	    itemWidth = item -> GetWidth ();
	    totalItemWidth += itemWidth;
	}
	if (totalItemWidth < rulerWidth)
	{
	    float	itemWidth;
	    float	itemWidthPct;
	    int		newItemWidth;
	    int		remainingWidth = rulerWidth;

	    for (i = 0; i < itemCount; i++)
	    {
		item = variablesRuler -> GetItemAt (i);
		itemWidth = (float)(item -> GetWidth ());
		itemWidthPct = itemWidth / (float)totalItemWidth;
		newItemWidth = (int)(itemWidthPct * rulerWidth);
		if ((remainingWidth - newItemWidth) <= 0)
		{
		    newItemWidth = remainingWidth;
		}
		item -> SetWidth (newItemWidth);
		remainingWidth -= newItemWidth;
	    }
	    // update the listviews columns widths
	    variablesListView -> SetColumnsWidths ();
	}
    }
    return (vTRUE);
}


/*******************************************************************
*
*    VariablesDialog 
*
*******************************************************************/

void VariablesDialog::OpenVariables (MainDialog *dlg, char *progName, long tid)
{
    mainDialog = dlg;
    strncpy (programName, progName, sizeof (programName));
    programName[PROGRAM_NAME_SIZE] = 0;
    TaskId = tid;

    if (mainDialog -> modifyVariables == vFALSE)
    {
	modifyButton -> Disable ();
    }

    SetVariablesTitle ();

    variablesListView -> SetDefaultCellContentIndent (0);

    UpdateVariablesList (1);

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open ();
    vevent::Process ();
}


/*******************************************************************
*
*    VariablesDialog function to set the dialog title 
*
*******************************************************************/

void VariablesDialog::SetVariablesTitle ()
{
    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource titleRes = app -> GetTitlesResource ();

    vstr *titleFormat = vresourceGetString (titleRes, snlvu_VariablesTitle);
    vstr *progName = vstrCloneScribed (vcharScribeSystem (programName));
    vstr *variablesTitle = vstrCloneScribed (vcharScribeFormat (titleFormat, progName));
    SetTitle (variablesTitle);
    vstrDestroy (progName);
    vstrDestroy (titleFormat);
    vstrDestroy (variablesTitle);
}


/*******************************************************************
*
*    VariablesDialog function to update the variables list in the main
*	view
*
*******************************************************************/

void VariablesDialog::UpdateVariablesList (int ignorePrev)
{
    REPVARS_REPLY *variables;
    int		  numVariables;

    snlMon_ReportVariables (programName, TaskId, &variables, &numVariables);

    if (PrevVariables != NULL)
    {
	free ((void *)PrevVariables);
	PrevVariables = NULL;
    }
    // remember the new stuff
    PrevVariables = variables;

    // build list for listview
    variablesListView -> BuildList (variables, numVariables);
}



vkindTYPED_FULL_DEFINITION(VariablesListView, vlistview, NULL);

/*******************************************************************
*
*    VariablesListView 
*
*******************************************************************/

void VariablesListView::Destroy (void)
{
    vliststr *list = (vliststr *)(GetList ());
    if (list != NULL)
    {
	SetList (NULL);
	delete list;
    }
    vlistview::Destroy ();
}


/*******************************************************************
*
*    VariablesListView, override
*
*******************************************************************/

void VariablesListView::ObserveSelect (int state)
{
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	variablesDialog -> InUse = vTRUE;
    }
}


/*******************************************************************
*
*    VariablesListView, build the list
*
*******************************************************************/

static char *VarTypes [] = {
	"none", "char", "short", "int", "long", "float", "double",
	"string", "evflag", "func", "uchar", "ushort", "uint",
	"ulong"
	};

void VariablesListView::BuildList (REPVARS_REPLY *variables, int numVariables)
{
    REPVARS_REPLY	*p = variables;
    REPVARS_REPLY	tmpreply;
    int			i;
    int			row = 0;
    char		*newp;
    char		VarType[20];
    char		Value[50];

    vliststr *newList = new vliststr;
    vliststr *oldList = (vliststr *)(GetList ());

    for (i = 0; i < numVariables; i++)
    {
	memcpy (&tmpreply, p, ((sizeof(REPVARS_REPLY)-sizeof(VARIABLE_NAME)) + 
								strlen(p->Name)+1));
	newp =  (char *)p + ((sizeof(REPVARS_REPLY)-sizeof(VARIABLE_NAME)) + 
								strlen(p->Name)+1);
	strcpy (VarType, VarTypes [tmpreply.Type]);
	if (tmpreply.NumElements)
	{
	    strcat (VarType, "[]");
	}
	GetValue (&tmpreply, newp, Value);
	newList -> SetCellValueScribed (row, 0, vcharScribeSystem (tmpreply.Name));
	newList -> SetCellValueScribed (row, 1, vcharScribeSystem (VarType));
	newList -> SetCellValueScribed (row, 2, vcharScribeSystem (Value));
	row++;

	if (tmpreply.Type == V_STRING)
	    newp += tmpreply.Size;
	else if (tmpreply.NumElements == 0)	/* scalar */
	    newp += tmpreply.Size;
	else
	    newp += (tmpreply.Size * tmpreply.NumElements);

	p = (REPVARS_REPLY *)newp;
    }
    SetList (newList);
    if (oldList != NULL)
    {
	delete oldList;
    }
    int rowCount = newList -> GetRowCount ();
    if (rowCount > 0)
    {
	SetColumnsWidths ();
	vlistSelection *selected = GetSelection ();
	newList -> SetSelectMethod (selected, vlistONE_ONLY_SELECTION);
	newList -> SetExtendRowSelection (selected, vTRUE);
    }
}


/*******************************************************************
*
*    VariablesListView, get a value for BuildList
*
*******************************************************************/

void VariablesListView::GetValue (REPVARS_REPLY *p, char *Var, char *Value)
{
    int	Count;
    char AValue [50];

    if (p->NumElements == 0)
	Count = 1;
    else
	Count = p->NumElements;
    char *q = Var;
    strcpy (Value, "");
    int NumChars = 0;
    for (int j=0 ; j<Count ; j++)
    {	
	switch (p->Type)
	{
	    case V_CHAR:
		sprintf (AValue, "%c", *q);
		break;
	    case V_INT:
		int tmpint;
		memcpy (&tmpint, q, p->Size);
		sprintf (AValue, "%d", tmpint);
		break;
	    case V_LONG:
		long tmplong;
		memcpy (&tmplong, q, p->Size);
		sprintf (AValue, "%ld", tmplong);
		break;
	    case V_SHORT:
		short tmpshort;
		memcpy (&tmpshort, q, p->Size);
		sprintf (AValue, "%d", tmpshort);
		break;
	    case V_FLOAT:
		float tmpfloat;
		memcpy (&tmpfloat, q, p->Size);
		sprintf (AValue, "%.2f", tmpfloat);
		break;
 	    case V_DOUBLE:
		double tmpdouble;
		memcpy (&tmpdouble, q, p->Size);
		sprintf (AValue, "%.2lf", tmpdouble);
		break;
	    case V_STRING:
		sprintf (AValue, "%s", q);
		break;
	}
	if (NumChars + strlen (AValue) > 25)
	{
	    strcat (Value, " ...");
	    break;
	}
	strcat (Value, AValue);
	NumChars += strlen (AValue);
	if (p->Type != V_STRING)
	{
	    strcat (Value, " ");
	    NumChars++;
	}
	q += p->Size;
    }
}


/*******************************************************************
*
*    VariablesListView, set list column widths based on main dialog's
*	ruler item widths
*
*******************************************************************/

void VariablesListView::SetColumnsWidths (void)
{
    vruler	*variablesRuler = variablesDialog -> GetVariablesRuler ();
    vrulerItem	*item;
    int		itemWidth;
    int		itemCount = variablesRuler -> GetItemCount ();
    int		i;

    for (i = 0; i < itemCount; i++)
    {
	item = variablesRuler -> GetItemAt (i);
	itemWidth = item -> GetWidth ();
	SetColumnWidth (i, itemWidth);
    }
}


vkindTYPED_FULL_DEFINITION(VariablesModifyButton, vbutton, NULL);

/*******************************************************************
*
*    VariablesModifyButton
*
*******************************************************************/

void VariablesModifyButton::ObserveDialogItem (vevent *event)
{
    REPVARS_REPLY	*p, tmpreply;
    REPVARS_REPLY	*PrevVariables = variablesDialog -> GetPrevVariables ();;
    char		*newp;
    VariablesListView	*variablesListView = variablesDialog -> GetVariablesListView ();
    vlistSelection	*selection = variablesListView -> GetSelection ();
    vliststr		*list = (vliststr *)(variablesListView -> GetList ());

    variablesDialog -> InUse = vTRUE;
    if (list == NULL)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	variablesDialog -> InUse = vFALSE;
	return;
    }

    int rowCount = list -> GetRowCount ();

    if (rowCount == 0)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	variablesDialog -> InUse = vFALSE;
	return;
    }
    if ((list -> IsSelectionEmpty (selection)) == vTRUE)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_NothingSelectedMsg);
	variablesDialog -> InUse = vFALSE;
	return;
    }
    int row, col;
    list -> GetSelectedCell (selection, &row, &col);

    int	i;
    for (p = PrevVariables, i = 0; i <= row ; i++)
    {
	memcpy (&tmpreply, p, ((sizeof(REPVARS_REPLY)-sizeof(VARIABLE_NAME)) + 
								strlen(p->Name)+1));
	newp =  (char *)p + ((sizeof(REPVARS_REPLY)-sizeof(VARIABLE_NAME)) + 
								strlen(p->Name)+1);
	if (tmpreply.Type == V_STRING)
	    newp += tmpreply.Size;
	else if (tmpreply.NumElements == 0)	/* scalar */
	    newp += tmpreply.Size;
	else
	    newp += (tmpreply.Size * tmpreply.NumElements);

	p = (REPVARS_REPLY *)newp;
    }
    if (tmpreply.NumElements != 0)
    {
	snlvuMessageCenter (variablesDialog, SNLVU_INFO_NOTICE, snlvu_ExpandArrayMsg);
	variablesDialog -> InUse = vFALSE;
	return;
    }

    MainDialog *mainDialog = variablesDialog -> GetMainDialog ();
    snlvuApplication *app = mainDialog -> GetApplication ();
    vresource appRes = app -> GetSnlvuResource ();
    vresource enterRes = vresourceGet (appRes, snlvu_EnterDialog);
    EnterDialog *enterDialog = new EnterDialog (enterRes);
    enterDialog -> OpenEnter (mainDialog, ENTER_VALUE, tmpreply.Name);
    vchar *enteredData = enterDialog -> GetEnteredData ();
    char newValue[100];
    vcharExportBounded (vcharScribeSystemExport (enteredData),
			newValue,
			sizeof (newValue));
    newValue[sizeof(newValue) - 1] = 0;
    delete (enterDialog);

    char *q = newValue;
    // remove leading blanks
    while (*q)
    {
	if (*q != ' ')
	    break;
	q++;
    }

    // Now have new value, save it in the proper format and send to server

    if (strlen (q))
    {
	char MsgBuf [100];
	switch (tmpreply.Type)
	{
	    case V_CHAR:
		MsgBuf [0] = *q;
		break;
	    case V_INT:
		*((int *)MsgBuf) = atoi (q);
		break;
	    case V_LONG:
		*((long *)MsgBuf) = atol (q);
		break;
	    case V_SHORT:
		*((short *)MsgBuf) = atoi (q);
		break;
	    case V_FLOAT:
		*((float *)MsgBuf) = (float)(atof (q));
		break;
	    case V_DOUBLE:
		*((double *)MsgBuf) = atof (q);
		break;
	    case V_STRING:
		for (int i=strlen (newValue) - 1 ; i>=0 ; i--)
		{
		    if (newValue [i] != ' ')
		    {
			newValue [i+1] = '\0';
			break;
		    }
		}
		strcpy (MsgBuf, newValue);
		break;
	}

	char *programName = variablesDialog -> GetProgramName ();
	long taskId = variablesDialog -> GetTaskId ();

	if (snlMon_ChangeVariable (programName, taskId, tmpreply.Name, MsgBuf, 
								tmpreply.Size) == 0)
	{
	    snlvuMessageCenter (variablesDialog, SNLVU_INFO_NOTICE,
							snlvu_VariableModifyError);
	    variablesDialog -> InUse = vFALSE;
	    return;
	}
	variablesDialog -> UpdateVariablesList (1);
    }
    variablesDialog -> InUse = vFALSE;
}


vkindTYPED_FULL_DEFINITION(VariablesExpandButton, vbutton, NULL);

/*******************************************************************
*
*    VariablesExpandButton
*
*******************************************************************/

void VariablesExpandButton::ObserveDialogItem (vevent *event)
{
    REPVARS_REPLY	*p, tmpreply;
    REPVARS_REPLY	*PrevVariables = variablesDialog -> GetPrevVariables ();;
    char		*newp;
    VariablesListView	*variablesListView = variablesDialog -> GetVariablesListView ();
    vlistSelection	*selection = variablesListView -> GetSelection ();
    vliststr		*list = (vliststr *)(variablesListView -> GetList ());

    variablesDialog -> InUse = vTRUE;
    if (list == NULL)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	variablesDialog -> InUse = vFALSE;
	return;
    }

    int rowCount = list -> GetRowCount ();

    if (rowCount == 0)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_EmptyListMsg);
	variablesDialog -> InUse = vFALSE;
	return;
    }
    if ((list -> IsSelectionEmpty (selection)) == vTRUE)
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_NothingSelectedMsg);
	variablesDialog -> InUse = vFALSE;
	return;
    }

    int row, col;
    list -> GetSelectedCell (selection, &row, &col);

    int i;
    for (p = PrevVariables, i = 0; i <= row ; i++)
    {
	memcpy (&tmpreply, p, ((sizeof(REPVARS_REPLY)-sizeof(VARIABLE_NAME)) + 
								strlen(p->Name)+1));
	newp =  (char *)p + ((sizeof(REPVARS_REPLY)-sizeof(VARIABLE_NAME)) + 
								strlen(p->Name)+1);
	if (tmpreply.Type == V_STRING)
	    newp += tmpreply.Size;
	else if (tmpreply.NumElements == 0)	/* scalar */
	    newp += tmpreply.Size;
	else
	    newp += (tmpreply.Size * tmpreply.NumElements);

	p = (REPVARS_REPLY *)newp;
    }
    if (tmpreply.NumElements > 0)
    {
	MainDialog *mainDialog = variablesDialog -> GetMainDialog ();
	snlvuApplication *app = mainDialog -> GetApplication ();
	vresource appRes = app -> GetSnlvuResource ();
	vresource expandRes = vresourceGet (appRes, snlvu_ExpandValuesDialog);
	mainDialog -> expandDialog = new ExpandDialog (expandRes);
	char *programName = variablesDialog -> GetProgramName ();
	long taskId = variablesDialog -> GetTaskId ();
	mainDialog -> expandDialog -> OpenExpand (mainDialog, 
						  programName,
						  taskId,
						  &tmpreply);
	delete (mainDialog -> expandDialog);
	mainDialog -> expandDialog = NULL;
	variablesDialog -> UpdateVariablesList (1);
    }
    else
    {
	snlvuMessageCenter (NULL, SNLVU_INFO_NOTICE, snlvu_NotArrayMsg, tmpreply.Name);
    }
    variablesDialog -> InUse = vFALSE;
}


vkindTYPED_FULL_DEFINITION(VariablesExitButton, vbutton, NULL);

/*******************************************************************
*
*    VariablesExitButton
*
*******************************************************************/

void VariablesExitButton::ObserveDialogItem (vevent *event)
{
    variablesDialog -> InUse = vTRUE;
    // stop secondary event loop started in open()
    vevent::StopProcessing ();
}

