/****
*
*	File name :		netdata.c
*	Author :		Steve Johnson
*	Creation date :	December 1, 1985
*
*	Copyright (c) 1985,1999 Sigma Six Corporation
*
*	Provides the following entry points
*
*		put_net_short (...)
*		put_net_long (...)
*		get_host_short (...)
*		get_host_long (....)
*
*	This procedure handles the byte swapping required for
*	passing data between machines of differing architecture
*
****/

#include <sscommon.h>
#include <ssnet.h>

static void	translate (void *, void *, size_t);
static int	is_net_type (void);

/****
*
****/

void put_net_short (void *net_short, short host_short)
{
	translate (net_short, &host_short, sizeof (short));
}

/****
*
****/

void put_net_long (void *net_long, long host_long)
{
	translate (net_long, &host_long, sizeof (long));
}

/****
*
****/

short get_host_short (void *net_short)
{
	short	host_short;

	translate (&host_short, net_short, sizeof (short));
	return host_short;
}

/****
*
****/

long get_host_long (void *net_long)
{
	long	host_long;

	translate (&host_long, net_long, sizeof (long));
	return host_long;
}

static void translate (void *target, void *source, size_t size)
{
	size_t		x;
	UCHAR		*c_target;
	UCHAR		*c_source;

	if (is_net_type ()) {
		memcpy (target, source, size);
	} else {
		c_target = (UCHAR *) target;
		c_source = (UCHAR *) source;
		for (x = 0; x < size; x++) {
			c_target [(size - 1) - x] = c_source [x];
		}
	}
}

static int is_net_type (void)
{
	long	value;

	memcpy (&value, "1234", 4);
	return (value != 0x34333231L);
}

