/****
*
*	File name :		tcpvxw.c
*	Author :		Steve Johnson
*	Creation date :	September 10, 1998
*
*	Copyright (c) 1998,1999 Sigma Six Corporation
*
*	Provides the following entry points
*
*		localtcp_init
*		localtcp_exit
*		localtcp_add
*		localtcp_remove
*		localtcp_select
*		localtcp_poll
*		localtcp_connect
*
*	TCP/IP support library
*
****/

#include <sscommon.h>
#include <sstrace.h>
#include <sstcpip.h>
#include <sstcplib.h>

/****
*
*	Return the first handle id with pending I/O
*
****/

int localtcp_init (void)
{
	TRACE ("localtcp_init, entry point");
	selectInit ();
	return TRUE;
}

void localtcp_exit (void)
{
	TRACE ("localtcp_exit, entry point");
}

int localtcp_add (int socket_descr)
{
	TRACE ("localtcp_add, entry point");
	return TRUE;
}

void localtcp_remove (int socket_descr)
{
	TRACE ("localtcp_remove, entry point");
}

int localtcp_select (FD_BIT_MASK *read_mask)
{
	int				count;
	FD_BIT_MASK		write_mask;
	FD_BIT_MASK		except_mask;
	struct timeval	time_value;
	struct timeval	*null_time;

	time_value . tv_sec = 0;
	time_value . tv_usec = 0;
	null_time = &time_value;
	FD_ZERO (&write_mask);
	FD_ZERO (&except_mask);
	count = select (FD_SETSIZE,
					read_mask,
					&write_mask,
					&except_mask,
					null_time);
	return count;
}

int localtcp_poll (int socket_descr)
{
	int				count;
	FD_BIT_MASK		write_mask;
	FD_BIT_MASK		read_mask;
	FD_BIT_MASK		except_mask;
	struct timeval	time_value;
	struct timeval	*null_time;

	time_value . tv_sec = 0;
	time_value . tv_usec = 0;
	null_time = &time_value;
	FD_ZERO (&write_mask);
	FD_ZERO (&except_mask);
	FD_ZERO (&read_mask);
	FD_SET (socket_descr, &read_mask);
	count = select (FD_SETSIZE,
					&read_mask,
					&write_mask,
					&except_mask,
					null_time);
	if (count < 0) {
		ERROR_TRACE_SYS ("localtcp_poll, select = (%ld) = %s");
	}
	return count;
}

/****
*
****/

int localtcp_connect (int 		socket_descr
					, char		*address
					, short		port
					, long		timeout)
{
	int				count;
	int				status;
	long			blocking;
	int				addr_len;
	struct timeval	time_value;
	struct timeval	*null_time;
	FD_BIT_MASK		write_mask;
	FD_BIT_MASK		except_mask;
	FD_BIT_MASK		read_mask;
	TCP_ADDR_IN		local_addr;
	TCP_ADDR_IN		remote_addr;

	TRACE ("localtcp_connect, entry point");
	TRACE_S ("localtcp_connect, address = (%s)", address);
	TRACE_I ("localtcp_connect, port = %d", port);
	TRACE_I ("localtcp_connect, socket_descr = %d", socket_descr);
  	memset (&local_addr, 0, sizeof (local_addr));
	local_addr . sin_family = AF_INET;
	local_addr . sin_port = htons (port);
	local_addr . sin_addr . s_addr = inet_addr (address);
	time_value . tv_sec = timeout / 1000;
	time_value . tv_usec = (timeout % 1000) * 1000;
	status = connectWithTimeout (socket_descr
							   , (TCP_ADDR *) &local_addr
							   , sizeof (local_addr)
							   , &time_value);
	TRACE ("localtcp_connect, after call to connect");
	if (status) {
		TRACE_I ("localtcp_connect, connect status = %d", status);
		TRACE_I ("localtcp_connect, socket_descr = %d", socket_descr);
		TRACE_I ("localtcp_connect, ConnectWithTimeout = %d", GetLastError ());
		TRACE_B ("localtcp_connect", sizeof (local_addr), &local_addr);
		return TCPIP_BAD_ADDRESS;
	}
	blocking = 1;
	NON_BLOCKING (socket_descr, blocking);
	TRACE ("localtcp_connect, successful");
	time_value . tv_sec = 0;
	time_value . tv_usec = 0;
	null_time = &time_value;
	FD_ZERO (&write_mask);
	FD_ZERO (&except_mask);
	FD_ZERO (&read_mask);
	FD_SET (socket_descr, &except_mask);
	FD_SET (socket_descr, &write_mask);
	count = select (FD_SETSIZE,
					&read_mask,
					&write_mask,
					&except_mask,
					null_time);
	if (count == -1) {
		ERROR_TRACE_I ("vxworks_connect, select = %d", GetLastError ());
		return TCPIP_ABORTED;
	} else if (count) {
		if (FD_ISSET (socket_descr, &except_mask)) {
			ERROR_TRACE_I ("vxworks_connect, select = %d", GetLastError ());
			return TCPIP_ABORTED;
		}
		if (FD_ISSET (socket_descr, &write_mask)) {
			addr_len = sizeof (remote_addr);
			status = getpeername (socket_descr,
							      (TCP_ADDR *) &remote_addr,
							      &addr_len);
			if (status) {
				ERROR_TRACE_I ("vxworks_connect, getpeername = %d",
						GetLastError ());
				return TCPIP_ABORTED;
			}
			blocking = 1;
			NON_BLOCKING (socket_descr, blocking);
			TRACE ("vxworks_connect, successful");
			return TCPIP_NORMAL;
		} else {
			ERROR_TRACE ("vxworks_connect, select lied here too");
			return TCPIP_ABORTED;
		}
	} else {
		ERROR_TRACE ("vxworks_connect, select lied");
		return TCPIP_ABORTED;
	}
}

