/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vbuttonHEADER
#include vtextitemHEADER
#include vstartupHEADER
#include vtimeHEADER
#include vtimerHEADER
#include vmemHEADER
#include vcontrolHEADER
#include vlistviewHEADER
#include vprintHEADER
#include "tisgalaxylib.hxx"
#include "tagvunames.h"
/*
	This is the source for the main dialog class
*/
// command enums for vcommand
enum
{
    EXIT_MENU = 0,
    FILTER_MENU,
    TAG_MENU,
    UNTAG_REMOVE_MENUITEM,
    PRINT_MENU,
    MENUITEM_COUNT
};
class tagvuListView;
class tagvuEntry;
class tagvuTimer;

class tagvuMainDialogClass : public vdialog
{
private:
	char *dateToOutput(char *inDate);		/* inDate is in CCYYMMDD format */
	tagvuEntry *currentSelection;
	TAGVU_ERROR_CODES errCode;
	vlist		*tagList;
	vdialogLabelItem *numFound;
	vprint *printer;
	FILTER_DATA_PTR filterData;
#ifdef TAGVU_CLIENT
	CLIENT *client;
	REQ_MSG_UNION reqMsg;
#endif TAGVU_CLIENT
	void filterIt(FILTER_DATA_PTR filter,PntTagRecord *tagRecord);
	// new functionality see source
	int firstTagRecord(FILTER_DATA_PTR filter,PntTagRecord *tagRecord);
	// new functionality see source
	int nextTagRecord(FILTER_DATA_PTR filter,PntTagRecord *tagRecord);

	// stuff for help menu etc
	vmenu		*helpMenu;
	TisAbout	*tisAbout;
	void		BuildAboutMenu ();
	vresource	tisGalaxyvresource;
	int		loadiniterror;
public:
    // macros for standard constructors and desstructors
    vkindTYPED_FULL_DECLARATION(tagvuMainDialogClass);
    vloadableINLINE_CONSTRUCTORS(tagvuMainDialogClass, vdialog);
    
    // place to store internal names for main dialog menu items
    const vname *commandNames[MENUITEM_COUNT];
#ifndef TAGVU_CLIENT
	CODE4		codeBaseDesc;
	DATA4		*pvmapDesc;
	DATA4		*opTrackDesc;
	DATA4		*tagDesc;
	INDEX4		*tagIndexDesc;
	int		caFd;
	TagValueData *pendingRequest;
#endif TAGVU_CLIENT
	int		timeout;
	char		operatorName[OPERATOR_SZ + 1];
//
    void LoadInit(vresource res);// override to load dialog
    void Destroy ();
    void Open (void);
    void ObserveWindow (vevent *event);
	tagvuListView *tagListView;
	tagvuTimer *caTimer;
	vmenuItem *removeItem;
	void updateListView();
	void updateList();
	void closeFiles();
	void setCurrentSelection(tagvuEntry *sel){currentSelection = sel;}
	tagvuEntry *getCurrentSelection(){return(currentSelection);}
	void ConnectionHandler(connection_handler_args args);
	char *stripWhiteSpace(char *);
	int SendTagValue(TagValueData *tagValueDataPtr);
	void FillFromTagDesc(TagValueData *tagValueDataPtr);
	void TagValueEvent(event_handler_args args);
	void addTagData(TagValueData *tvd);
	void errMessage(char *message);
	TAGVU_ERROR_CODES getErrorCode(){return(errCode);}
	vprint *getPrinter(){return(printer);}
	void tagvuMainDialogClass::printList();
	FILTER_DATA_PTR getFilterData(){return(filterData);}
	void setFilterData(FILTER_DATA_PTR fd){filterData = fd;}
	int GetLoadInitError (void) { return (loadiniterror); }
#ifdef TAGVU_CLIENT
	CLIENT *getClient(){return(client);}
	REQ_MSG_UNION *getReqMsg(){return(&reqMsg);}
#endif TAGVU_CLIENT
};
// ** CLASS FOR THE MAIN DIALOGS MENU COMMANDS *************************************

class  tagvuCommand : public vcommandSelector
{
private:
	tagvuMainDialogClass *mainDialog;
public:
    vkindTYPED_FULL_DECLARATION(tagvuCommand);
    vloadableINLINE_CONSTRUCTORS(tagvuCommand, vcommandSelector);
    int Issue(vdict *);				//override virtual issue function
	tagvuMainDialogClass *getMainDialog(){return(mainDialog);}
	void setMainDialog(tagvuMainDialogClass *m){mainDialog = m;}
};
//
// class for list view used to display list of tagged/untagged points
//
class tagvuListView : public vlistview
{
private:
	tagvuMainDialogClass *mainDialog;
public:
    vkindTYPED_FULL_DECLARATION(tagvuListView);
    vloadableINLINE_CONSTRUCTORS(tagvuListView, vlistview);

	void DrawCellContents(int row,int col,int sel,vrect *rect);
	void ObserveSelect(int state);
	tagvuMainDialogClass *getMainDialog(){return(mainDialog);}
	void setMainDialog(tagvuMainDialogClass *m){mainDialog = m;}
};
//
// This class holds 3 strings cooresponding to the tag status lines in the list view
//
class tagvuEntry
{
private:
	char tagLines[LINES_PER_TAG][LINE_SIZE+1];
	char tagName[PVNAME_SZ+1];
	int recNumber;
public:
	tagvuEntry(){;}
	~tagvuEntry(){;}
	char *getLineIndexed(int index);
	void putLineIndexed(int index, char *newline);
	//
	void setRecordNumber(int num){recNumber = num;}
	int getRecordNumber(){return(recNumber);}
	//
	void setTagName(char *name){strcpy(tagName,name);}
	char *getTagName(){return(tagName);}
};
//
// This is my timer class extension. I did this so I could overide
// the ObserveTimer function to check CA with ca_pend_io
//
class tagvuTimer : public vtimer
{
private:
    tagvuMainDialogClass *mainDialog;
public:
    tagvuTimer(tagvuMainDialogClass *m){mainDialog = m;}
    void ObserveTimer();
    tagvuMainDialogClass *getMainDialog(){return(mainDialog);}
};
