/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vmenubarHEADER
#include vbuttonHEADER
#include vtextitemHEADER
#include vstartupHEADER
#include vtimeHEADER
#include vmemHEADER
#include vcontrolHEADER
#include vcontainerHEADER
extern "C"
{
#undef INT_MAX
#undef UINT_MAX
#undef USHRT_MAX
#include "cadef.h"
#ifdef WIN32
#include <winsock.h>
#else
#include <sys/types.h>
#include <netinet/in.h>
#endif WIN32
}
typedef char *caddr_t;
#include "tagvu.h"
#include "tagvunames.h"
#include "textFilters.hxx"
#include "tagvuFilterDialogClass.h"
/*
	This is the source for the filter dialog class
*/
vkindTYPED_FULL_DEFINITION(tagvuFilterDialogClass, vdialog, "tagvuFilterDialogClass");
vkindTYPED_FULL_DEFINITION(tagvuFilterOKButton, vbutton, "tagvuFilterOKButton");
vkindTYPED_FULL_DEFINITION(tagvuFilterCancelButton, vbutton, "tagvuFilterCancelButton");
vkindTYPED_FULL_DEFINITION(tagvuFilterClearButton, vbutton, "tagvuFilterClearButton");

void tagvuFilterDialogClass::LoadInit(vresource res)
{
	// text fields
	nameText = filteredTextitem::RequestSubLoad(this,tagvunames_filterPointName);
	operatorText = filteredTextitem::RequestSubLoad(this,tagvunames_filterOperator);
	locationText = filteredTextitem::RequestSubLoad(this,tagvunames_filterLocation);
	dateText = filteredTextitem::RequestSubLoad(this,tagvunames_filterDate);
	// tag state filter radio buttons
	stateContainer = vcontainerExclusiveGroup::RequestSubLoad(this,
						tagvunames_filterStateContainer);
	//tagState = vcontrolToggle::RequestSubLoad(this,tagvunames_filterTagState);
	// before/after radio button container w/ buttons
	vcontainerExclusiveGroup *radio = vcontainerExclusiveGroup::RequestSubLoad(this,
						tagvunames_tagvuExclusiveGroup);
	dateDirection = vcontrolToggle::RequestSubLoad(radio,tagvunames_filterAfter);
	// finally the buttons
	okButton = tagvuFilterOKButton::RequestSubLoad(this,tagvunames_filterOK);
 	cancelButton = tagvuFilterCancelButton::RequestSubLoad(this,tagvunames_filterCancel);
 	clearButton = tagvuFilterClearButton::RequestSubLoad(this,tagvunames_filterClear);
	// set parent dialog
	vdialog::LoadInit(res);
	okButton->setParent(this);
	cancelButton->setParent(this);
	clearButton->setParent(this);
	filterData = NULL;

	// set filters
	nameText -> SetItemFilterFunc (pvnameFilter);
	operatorText -> SetItemFilterFunc (operatorNameFilter);
	locationText -> SetItemFilterFunc (locationFilter);
	dateText -> SetItemFilterFunc (dateFilter);
}
void tagvuFilterDialogClass::Destroy()
{
}
void tagvuFilterDialogClass::ObserveWindow(vevent *event)
{
    if(event -> GetType() == veventWINDOW_STATE && event -> IsClose())
    {
	Close();
	vevent::StopProcessing();
    }
}
void tagvuFilterDialogClass::Open()
{
	vdialog::Open();
    vevent::Process();
}
//
// button callbacks
//
/*
	The OK button formats a dBase/Codebase filter string,re-builds the
	index file using the new filter and displays the resulting list
*/
void tagvuFilterOKButton::ObserveDialogItem(vevent *event)
{
	char pointString[PVNAME_SZ+1];
	char operatorString[OPERATOR_SZ+1];
	char locationString[LOCATION_SZ+1];
	char dateString[EXT_DATE_SIZE+1];
	FILTER_DATA_PTR filterData;


	vbutton::ObserveDialogItem(event);
	Close();
	if((filterData = (FILTER_DATA_PTR)vmemAllocAndClear(sizeof(FILTER_DATA))))
	{
	/*
		Extract the text strings, the desired tag state and the before/after
		switch value.
	*/
	// point name
		vstr *str1 = getParent()->nameText->GetTextAsString();
		vcharExportBounded (vcharScribeSystemExport(str1),pointString,PVNAME_SZ);
		vstrDestroy (str1);
		if(strlen(pointString)) strcpy(filterData->pointName,pointString);
	// operator
		str1 = getParent()->operatorText->GetTextAsString();
		vcharExportBounded (vcharScribeSystemExport(str1),operatorString,OPERATOR_SZ);
		vstrDestroy (str1);
		if(strlen(operatorString)) strcpy(filterData->operatorApplied,operatorString);
	// location
		str1 = getParent()->locationText->GetTextAsString();
		vcharExportBounded (vcharScribeSystemExport(str1),locationString,LOCATION_SZ);
		vstrDestroy (str1);
		if(strlen(locationString)) strcpy(filterData->location,locationString);
	// date
		str1 = getParent()->dateText->GetTextAsString();
		vcharExportBounded (vcharScribeSystemExport(str1),dateString,EXT_DATE_SIZE);
		vstrDestroy (str1);
		if(strlen(dateString)) strcpy(filterData->dateApplied,inputToDate(dateString));
	// date relational (< or >)
		filterData->dateDirection = getParent()->dateDirection->GetValue();
		filterData->dateDirection = htonl(filterData->dateDirection);
	// tag state
		vdialogItem *whichOne = (vdialogItem *)getParent()->stateContainer->GetSelection();
		const vname *stateName = whichOne->GetTag();
		char *tagString = (char *)vcharExportAlloc(vnameScribeGlobal(stateName));
		if(!strcmp((const char *)tagvunames_filterAll,tagString))
			filterData->tagState = FILTER_SHOW_ALL;
		else if(!strcmp((const char *)tagvunames_filterTagged,tagString))
			filterData->tagState = FILTER_SHOW_TAGGED;
		else
			filterData->tagState = FILTER_SHOW_UNTAGGED;
		filterData->tagState = htonl(filterData->tagState);
		vcharExportFree(tagString);
	// all done! copy to parent
		getParent()->setFilterData(filterData);
		vevent::StopProcessing();
	}
}
/*
  GCT 14-May-1998 new function to convert date from "MM/DD/CCYY" to
  "CCYYMMDD"
*/
char *tagvuFilterOKButton::inputToDate(char *dstr)
{
  static char dbaseDateStr[20];
  char *m,*d,*y;
  char temp[20];
  strcpy(temp,dstr);
  m = strtok(temp,"/");
  d = strtok(NULL,"/");
  y = strtok(NULL,"/");
  sprintf(dbaseDateStr,"%4.4s%2.2s%2.2s",y,m,d);
  return(dbaseDateStr);
}
//
// Dismiss the dialog and kill local event loop
//
void tagvuFilterCancelButton::ObserveDialogItem(vevent *event)
{
	vbutton::ObserveDialogItem(event);
	getParent()->setFilterData(NULL);
	Close();
	vevent::StopProcessing();
}
//
// Clear fields and stick around
//
void tagvuFilterClearButton::ObserveDialogItem(vevent *event)
{
	vbutton::ObserveDialogItem(event);
	getParent()->nameText->SetText((vchar *)"");
	getParent()->locationText->SetText((vchar *)"");
	getParent()->operatorText->SetText((vchar *)"");
	getParent()->dateText->SetText((vchar *)"");
	getParent()->tagState->SetValue(0);
	getParent()->dateDirection->SetValue(0);
}
