/*CSLCompiler.c,v 1.1.1.1 1999/08/03 14:44:52 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <CSLScript.h>
#include "symtab.h"
#include "csldefs.h"

	FILE	*yyerfp = stdout;	/* error stream */
extern	char	yytext [];	/* Current token */
extern	int	yyleng;		/* Length of current token */

#ifdef LINUX
	int	yylineno = 0;
#else
extern	int	yylineno;
#endif

extern	int	yynerrs;	/* Total number of errors */
static	char	*source = 0;	/* Current input source file */
static	int	NumErrors;	/* Number of errors found */

extern	void	WriteSymbols ();

/*----------*/
main (argc, argv)
int	argc;
char	*argv[];

    {
    if (argc > 1)
	for (argc--, argv++; argc > 0; argc--)
		CompileProgram(*argv++);
    else
	CompileProgram(NULL);
    }


CompileProgram(name)

char	*name;		/* use stdin/stdout if NULL */
    {
        char	*getenv();
	long	Major, Minor, Magic;
	char	filename[256];
	char	*path	= (path = getenv(CSL_SCRIPT_PATH)) ? path : ".";

	if (name != NULL)
	    {
	    sprintf(filename, "%s/%s", path, name);
	    if (freopen(filename, "r", stdin) != stdin)
		{
	 	fprintf(stderr, "0:can't open script %s\n", filename);
		return;
		}
	    }

	NumErrors = 0;

	SymTabInit ();

	/* Build a parse tree */

	while (yyparse());

	/* Output version numbers and magic number */

	if (NumErrors == 0) {
	    
	    if (name != NULL )
	    	{
		strcat(filename, CSL_SCRIPT_EXTENSION);
		if (freopen(filename, "w+", stdout) != stdout)
	    	    {
	    	    fprintf(stderr, "0:can't create %s\n", filename);
	    	    return;
		    }
	    	}

	    Magic = CSL_MAGIC;

	    putchar (Magic);
	    putchar (Magic >> 8);
	    putchar (Magic >> 16);
	    putchar (Magic >> 24);

	    Major = CSL_MAJOR_VERSION;
	    putchar (Major);
	    putchar (Major >> 8);
	    putchar (Major >> 16);
	    putchar (Major >> 24);

	    Minor = CSL_MINOR_VERSION;
	    putchar (Minor);
	    putchar (Minor >> 8);
	    putchar (Minor >> 16);
	    putchar (Minor >> 24);

	    /* Write the symbols */

	    WriteSymbols ();

	    /* Generate code and write to output */

	    Gen_Program ();

	    /* Check for undefined labels */

	    CheckLabels ();

	}
    }

/* end main */

/*----------*/
void Error (Msg)	/* Send an error message to stderr */

    char	Msg [];

    {

	fprintf (stderr, "%d:%s\n", yylineno, Msg);
	NumErrors++;

    }

/* end Error */

/*----------*/
yyerror (s)

    char	*s;

    {
	yywhere ();
	fputs (s, stderr);
	putc ('\n', stderr);
	NumErrors++;
    }

/* end yyerror */
	
/*----------*/
yywhere ()

    {

	fprintf (stderr, "%d:", yylineno -
			(*yytext == '\n' || !*yytext));
	if (*yytext) {
	    register int i;
	    for (i=0 ; i<20 ; ++i)
		if (!yytext [i] || yytext[i] == '\n')
		    break;
	    if (i) {
		fprintf (stderr, "near \"%.*s\" ", i, yytext);
	    }
	}

    }

/* end yywhere */

/*----------*/
yymark ()

    {

	if (source)
	    cfree (source);
	source = (char *) calloc (yyleng, sizeof (char));
	if (source) {
#ifndef LINUX
	    sscanf (yytext, "# %d %s", &yylineno, source);
	    printf ("Line = %d, file = %s\n", yylineno, source);
#else
	    sscanf (yytext, "# %s", source);
	    printf ("File = %s\n", source);
#endif
	}

    }

/* end yymark */
