/*CSLDump.c,v 1.1.1.1 1999/08/03 14:44:52 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

#include <stdio.h>
#include "CSLScript.h"
#include "opcodes.h"
#include "csldefs.h"

#define	FALSE	0
#define	TRUE	1

static struct optableentry {
    int		OpCode;
    char	*String;
    int		NumOps;
    int		AsString [3];
} OpTable [] = {
	{NOP,		"NOP",		0},
	{JUMP,		"JUMP",		1,	FALSE},
	{JZERO,		"JZERO",	1,	FALSE},
	{JNZERO,	"JNZERO",	1,	FALSE},
	{EQ,		"EQ",		0},
	{NE,		"NE",		0},
	{GT,		"GT",		0},
	{GE,		"GE",		0},
	{LT,		"LT",		0},
	{LE,		"LE",		0},
	{LEQ,		"LEQ",		0},
	{LNE,		"LNE",		0},
	{LGT,		"LGT",		0},
	{LGE,		"LGE",		0},
	{LLT,		"LLT",		0},
	{LLE,		"LLE",		0},
	{AND,		"AND",		0},
	{OR,		"OR",		0},
	{NOT,		"NOT",		0},
	{ADD,		"ADD",		0},
	{SUB,		"SUB",		0},
	{MULT,		"MULT",		0},
	{DIV,		"DIV",		0},
	{POWER,		"POWER",	0},
	{NEG,		"NEG",		0},
	{ABS,		"ABS",		0},
	{ACOS,		"ACOS",		0},
	{ASIN,		"ASIN",		0},
	{ATAN,		"ATAN",		0},
	{CEILING,	"CEILING",	0},
	{COS,		"COS",		0},
	{EXP,		"EXP",		0},
	{FLOOR,		"FLOOR",	0},
	{LN,		"LN",		0},
	{LOG,		"LOG",		0},
	{MAX,		"MAX",		0},
	{MIN,		"MIN",		0},
	{MOD,		"MOD",		0},
	{SIN,		"SIN",		0},
	{SQRT,		"SQRT",		0},
	{TAN,		"TAN",		0},
	{PUSH,		"PUSH",		1,	FALSE},
	{POP,		"POP",		0},
	{LOAD,		"LOAD",		1,	TRUE},
	{STORE,		"STORE",	1,	TRUE},
	{TSTART,	"TSTART",	1,	TRUE},
	{TSTOP,		"TSTOP",	1,	TRUE},
	{TEXP,		"TEXP",		1,	TRUE},
	{ACCESS,	"ACCESS",	1,	TRUE},
	{COMD,		"COMD",		2,	TRUE,	TRUE},
	{WAITT,		"WAITT",	1,	TRUE},
	{WAITTOD,	"WAITTOD",	1,	TRUE},
	{WAITCSL,	"WAITCSL",	1,	TRUE},
	{WAITCSLT,	"WAITCSLT",	2,	TRUE,	TRUE},
	{PAUSE,		"PAUSE",	0},
	{SET,		"SET",		1,	TRUE},
	{SETV,		"SETV",		1,	TRUE},
	{SETT,		"SETT",		2,	TRUE,	TRUE},
	{ALARM,		"ALARM",	3,	TRUE,	TRUE,	TRUE},
	{SCSL,		"SCSL",		1,	TRUE},
	{SCSLW,		"SCSLW",	1,	TRUE},
	{SSH,		"SSH",		1,	TRUE},
	{SSHW,		"SSHW",		1,	TRUE},
	{ACT,		"ACT",		1,	TRUE},
	{STOP,		"STOP",		0},
	{STOPX,		"STOPX",	0},
	{STOPCSL,	"STOPCSL",	1,	TRUE},
	{LINENUM,	"LINENUM",	1,	FALSE},
	{WITHIN,	"WITHIN",	0},
	{NWITHIN,	"NWITHIN",	0},
	{SCSLTOD,	"SCSLTOD",	2,	TRUE,	TRUE},
	{SCSLTODW,	"SCSLTODW",	2,	TRUE,	TRUE},
	{SCSLINT,	"SCSLINT",	2,	TRUE,	TRUE},
	{SCSLINTW,	"SCSLINTW",	2,	TRUE,	TRUE},
	{SSHTOD,	"SSHTOD",	2,	TRUE,	TRUE},
	{SSHTODW,	"SSHTODW",	2,	TRUE,	TRUE},
	{SSHINT,	"SSHINT",	2,	TRUE,	TRUE},
	{SSHINTW,	"SSHINTW",	2,	TRUE,	TRUE},
	{SSHH,		"SSHH",		2,	TRUE,	TRUE},
	{SSHWH,		"SSHWH",	2,	TRUE,	TRUE},
	{SSHTODH,	"SSHTODH",	3,	TRUE,	TRUE,	TRUE},
	{SSHTODWH,	"SSHTODWH",	3,	TRUE,	TRUE,	TRUE},
	{SSHINTH,	"SSHINTH",	3,	TRUE,	TRUE,	TRUE},
	{SSHINTWH,	"SSHINTWH",	3,	TRUE,	TRUE,	TRUE},
	{RUNSTATUS,	"SCRIPTRUNSTATUS", 0},
	{COMPSTATUS,	"SCRIPTCOMPLETIONSTATUS", 0},
	{STORETINT,	"STORETINT",	1,	TRUE},
	{STORETTOD,	"STORETTOD",	1,	TRUE},
	{DECRT,		"DECRT",	0}
};
long	GetLongWord ();
void	LoadSymbols ();
void	PrintSymbols ();

char	SymBuf [50000];
char	*Symbols [5000];
int	NumSymbols;

/*----------*/
main (argc, argv)
int	argc;
char	*argv[];

    {
	int	Byte, SymNumber, Word, i, j, pc;
	char	Symbol [50];
	long	Major, Minor, Magic; 
        char	*getenv();
	char	filename[256];
	char	*path	= (path = getenv(CSL_SCRIPT_PATH)) ? path : ".";
	char	*name	= (argc == 2) ? *(++argv) : NULL;

	if (name != NULL)
	    {
	    sprintf(filename, "%s/%s%s", path, name, CSL_SCRIPT_EXTENSION);
	    if (freopen(filename, "r", stdin) != stdin)
		{
	 	printf("0:can't open %s\n", filename);
		return;
		}
	    }

	/* Print major and minor version numbers and magic number */
	
	Magic = GetLongWord ();
	printf ("Magic Number: 0x%lx\n", Magic);

	/* Check for valid magic number */

	if (Magic != CSL_MAGIC) {
	    printf ("File is not a CSL compiled code file.\n");
	    exit (1);
	}

	Major = GetLongWord ();
	printf ("Major Version: %ld\n", Major);

	Minor = GetLongWord ();
	printf ("Minor Version: %ld\n", Minor);

	/* Load the symbols */

	LoadSymbols ();

	/* Print the symbols */

	PrintSymbols ();

	printf ("Code:\n");

	pc = 0;	
	while ((Word = getchar ()) != EOF) {
	    Word |= getchar () << 8;
	    for (i=0 ; i<(sizeof(OpTable)/sizeof(struct optableentry)) ; i++) {
		if (Word == OpTable [i].OpCode)
		    break;
	    }
	    printf ("%3d: %s ", pc++, OpTable [i].String);
	    for (j=0 ; j<OpTable [i].NumOps ; j++) {
		Word = getchar ();
		Word |= getchar () << 8;
		if (OpTable [i].AsString [j])
		    printf ("\"%s\" ", Symbols [Word] + 1);
		else
		    printf ("%d ", Word);
		pc++;
	    }
	    printf ("\n");
	}

    }

/*----------*/
long GetLongWord ()

    {

	int	Byte1, Byte2, Byte3, Byte4;

	Byte1 = getchar ();
	Byte2 = getchar ();
	Byte3 = getchar ();
	Byte4 = getchar ();
	return Byte1 | (Byte2 << 8) | (Byte3 << 16) | (Byte4 << 24);

    }

/*----------*/
void LoadSymbols ()

    {
	int	ch, i, j;

	i = 0;
	j = 0;
	Symbols [j++] = SymBuf;
	while (ch = getchar ()) {
	    if (ch != '\n')
		SymBuf [i++] = ch;
	    else {
		SymBuf [i++] = '\0';
		Symbols [j++] = &SymBuf [i];
	    }
	} 	
	NumSymbols = j;
    }
  
/*----------*/
void PrintSymbols ()

    {

	int	i;

	printf ("Symbols:\n");
	for (i=0 ; i<(NumSymbols-1) ; i++) {
	    printf ("%3d: ", i);
	    switch (*Symbols [i]) {
		case 'V':	printf ("Variable ");	break;
		case 'C':	printf ("Channel  ");	break;
		case 'T':	printf ("Timer    ");	break;
		case 'I':	printf ("Integer  ");	break;
		case 'R':	printf ("Real     ");	break;
		case 'L':	printf ("Literal  ");	break;
		case 't':	printf ("Time     ");	break;
	    }
	    printf ("\"%s\"\n", Symbols [i] + 1);
	} 
   
    }

/* end PrintSymbols */
 
