/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/* Values use for MyRole */

#define	ROLE_NOREDUNDANCY	0
#define	ROLE_MASTER		1
#define	ROLE_SLAVE		2
#define	ROLE_MASTER_DP		3

#define	MAXMESSAGELEN		104
#define	MAXSCRIPTNAMELEN	104
#define	MAXENVIRONSTRINGLEN	200
#define	HOSTNAMELEN		24

/* Some types that can ensure predictable sizes of things */

typedef	unsigned int	suint32;
typedef	int		sint32;

/* Redundancy message types - All message lengths are divisable by 8 */
/* to ensure portability. */

typedef enum _msgTypes {
	EXECUTE,
	KILL,
	GET_STATUS,
	SET_STATUS,
	MESSAGE,
	RELEASE,
	STARTALL,
	DETACH
} msgTypes;

typedef struct _MsgHeader {
	suint32	msgType;
} RedMsgHeader;

typedef struct _SetStatusMsg {
	suint32	msgType;
	suint32	scriptId;
	suint32	status;
	char	completionCode [20];
} SetStatusMsg;

typedef struct _MessageMsg {
	suint32	msgType;
	suint32	scriptId;
	char	message [MAXMESSAGELEN];
} MessageMsg;

typedef struct _ExecuteMsg {
	suint32	msgType;
	suint32	scriptType;
	suint32	preStartDelay;
	suint32	scriptId;
	char	scriptName [MAXSCRIPTNAMELEN];
	char	environmentString [MAXENVIRONSTRINGLEN];
} ExecuteMsg;

typedef struct _KillMsg {
	suint32	msgType;
	suint32	scriptId;
} KillMsg;

typedef struct _GetStatusMsg {
	suint32	msgType;
	suint32	scriptId;
} GetStatusMsg;

typedef struct _ReleaseMsg {
	suint32	msgType;
	suint32	scriptId;
} ReleaseMsg;

typedef struct _StartAllMsg {
	suint32	msgType;
	char	pad [8];
} StartAllMsg;

typedef struct _DetachMsg {
	suint32	msgType;
	suint32	scriptId;
} DetachMsg;

/* Union of all redundancy message types results in a message that can be used to */
/* safely receive any message that can show up. */

typedef union _RedundMsg {
	RedMsgHeader	msgHeader;
	SetStatusMsg	setStatusMsg;
	MessageMsg	messageMsg;
	ExecuteMsg	executeMsg;
	KillMsg		killMsg;
	GetStatusMsg	getStatusMsg;
	ReleaseMsg	releaseMsg;
	StartAllMsg	startAllMsg;
	DetachMsg	detachMsg;
} RedundMsg;

typedef struct {
	NODE		node;
	int		scriptType;		/* CSLscript or SHELLscript */
	char		scriptName[MAXSCRIPTNAMELEN];	/* the scripts name (and args) */
	char		scriptEnviron [MAXENVIRONSTRINGLEN];
	ScriptId 	scriptId;		/* identifing handle */
	ScriptStatus	scriptStatus;		/* current status */
	time_t		statusChangeTime;	/* time of last status change */
	int		scriptStatusTaken;	/* 1 iff get CSLMonitorGetScriptStatus() */
						/*   called on this after it exited */
	double		completionCode;		/* only reasonable if scriptStatus == Exited */
	int		haveCompletionCode;	/* 1 iff get wait4() called on this successfully */
	int		pid;			/* pid of controlling interpreter or shell */
	char		message[MAXMESSAGELEN];	/* text of the last message */
	int		Detached;		/* Script is detached and thus is not restarted */
						/* when the slave monitor activates. */
} Script;

#ifdef MIPSEL
#define	LOG_OPTIONS	LOG_SALERT
#else
#define	LOG_OPTIONS	LOG_ALERT|LOG_LOCAL1
#endif

