/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/* This is a custom version of the main program for CSLMonitor_svc.  It is
/* identical to that which would be generated by rpcgen except the call to
/* svc_run has been replaced by a call to CSLMonitor_svc_run which is also
/* defined in this module.
/*
/* CSLMonitor_svc_run differs from svc_run in that it places a timeout on
/* the select call.  When the timeout expires, the arbiter is checked to
/* see if the master/slave status has changed and for any incoming messages
/* from our redundant partner.
*/

#include <stdio.h>
#include <errno.h>
#include <rpc/rpc.h>
#ifdef SOLARIS
#include <sys/resource.h>
#include <rpc/svc_soc.h>
#endif

#include <CSLMonitor.h>
#include <CSLMonitorDefs.h>
#include <msapp.h>

/* Functions defined in this module */

static	void	CSLMonitor_svc_run ();

/* External functions */

extern	void	cslmonitorprog_1 ();
extern	int	*csl_monitor_set_my_status_1 ();
extern	int	*csl_monitor_message_1 ();
extern	ScriptId *csl_monitor_execute_1 ();
extern	int	*csl_monitor_kill_1 ();
extern	getStatusResult *csl_monitor_get_status_1 ();
extern	int	*csl_monitor_release_1 ();
extern	int	*csl_monitor_detach_1 ();

/* External data items */

extern	int	MyHandle, MyRole, Redundant, Verbose, Debug;

/* Local data */

static	RedundMsg	msgBuffer;	/* Redund msg buffer that can hold any message. */
static	int		myPid;

static	struct	executeArg executeArgStruct;
static	struct	killArg killArgStruct;
static	struct	setStatusArg setStatusArgStruct;
static	struct	getStatusArg getStatusArgStruct;
static	struct	releaseArg releaseArgStruct;
static	struct	messageArg messageArgStruct;
static	struct	detachArg detachArgStruct;

/* Script list stuff */

extern	LIST		scriptList;	/* the list of scripts */
extern	ScriptId	MasterScriptId;
#define foreachScript(x) for(x = (Script *)lstFirst(&scriptList); x; x = (Script *)lstNext(x))

#ifdef SOLARIS
int	_rpcpmstart;
#endif

/*----------*/
main(argc, argv)

    int		argc;
    char	*argv [];

    {

	register SVCXPRT *transp;
	int	Cmd;

	myPid = getpid ();

	serverInit ("CSLMonitor_svc", "/tis4000/tasks/CSLMonitor_svc.lock", "lock");

	Redundant = No;
	Verbose = No;
	Debug = No;
	while((Cmd = getopt(argc,argv,"rvd")) != -1) {
	    switch(Cmd) {
		case 'r':
		    Redundant = Yes;
		    break;
		case 'v':
		    Verbose = Yes;
		    break;
		case 'd':
		    Debug = Yes;
		    Verbose = Yes;
		    break;
	    }
	}

	(void) pmap_unset(CSLMONITORPROG, CSLMONITORVERS);

	transp = svcudp_create(RPC_ANYSOCK);
	if (transp == NULL) {
		fprintf(stderr, "cannot create udp service.");
		exit(1);
	}
	if (!svc_register(transp, CSLMONITORPROG, CSLMONITORVERS,
			cslmonitorprog_1, IPPROTO_UDP)) {
		fprintf(stderr, "unable to register (CSLMONITORPROG, CSLMONITORVERS, udp).");
		exit(1);
	}

	transp = svctcp_create(RPC_ANYSOCK, 0, 0);
	if (transp == NULL) {
		fprintf(stderr, "cannot create tcp service.");
		exit(1);
	}
	if (!svc_register(transp, CSLMONITORPROG, CSLMONITORVERS,
			cslmonitorprog_1, IPPROTO_TCP)) {
		fprintf(stderr, "unable to register (CSLMONITORPROG, CSLMONITORVERS, tcp).");
		exit(1);
	}
	
	/* Start processing requests */

	CSLMonitor_svc_run();
	fprintf(stderr, "svc_run returned");
	exit(1);
	/* NOTREACHED */

    }

/*----------*/
static void RegisterWithArbiter ()

    {

	if (!Redundant) {
	    MyRole = ROLE_NOREDUNDANCY;
	    return;
	}

	/* Keep trying to register with arbiter until successful. */

	for (;;) {
	    int Status;
	    if (Status = appRegister ("CSLMonitor", &MyHandle)) {
		appGetStatus (MyHandle, &Status);
		switch (Status) {
		    case MASTER:
			MyRole = ROLE_MASTER;
			if (Verbose)
			    syslog (LOG_OPTIONS, "CSLMonitor_svc(%d): Running as master.", myPid);
			break;
		    case SLAVE:
			MyRole = ROLE_SLAVE;
			if (Verbose)
			    syslog (LOG_OPTIONS, "CSLMonitor_svc(%d): Running as slave.", myPid);
			break;
		    case MASTER_NOPARTNER:
			MyRole = ROLE_MASTER_DP;
			if (Verbose)
			    syslog (LOG_OPTIONS, "CSLMonitor_svc(%d): Running as master with no partner.", myPid);
			break;
		}
		return;
	    } else {
		if (Verbose)
		    syslog (LOG_OPTIONS, "CSLMonitor_svc(%d): Unable to connect with arbiter -- retrying.", myPid);
		sleep (5);
	    }
	}

    }

/*----------*/
static void CSLMonitor_svc_run ()
{
	fd_set		readfds;
	int		dtbsz;
	int		Status, OldStatus;
	int		Sock;
	static struct timeval timeout, *to;
	Script		*pScript;
	StartAllMsg	startAllMsg;
	ExecuteMsg	executeMsg;
#ifdef SOLARIS
	struct rlimit	rlp;
#endif

	if (Verbose)
	    syslog (LOG_OPTIONS, "CSLMonitor_svc(%d): Running.", myPid);

	/* Determine the maximem fd value */

#ifdef SOLARIS
	getrlimit (RLIMIT_NOFILE, &rlp);
	dtbsz = rlp.rlim_cur;
#else
	dtbsz = getdtablesize ();
#endif

	/* Register with the arbiter if necessary and get initial status */

	RegisterWithArbiter ();

	/* Set up timeout for checking the arbiter for a role change. */

	if (!Redundant) {
	    to = NULL;
	} else {	    	
	    timeout.tv_sec = 1;		/* Every second */
	    timeout.tv_usec = 0;
	    to = &timeout;
	}

	/* If I am a slave, send a message to my redundant partner to start up all scripts */
	/* that should be running. */

	if (MyRole == ROLE_SLAVE) {
	    if (Debug)
		syslog (LOG_OPTIONS, "CSLMonitor_svc(%d): Sending STARTALL message to master.", myPid);
	    startAllMsg.msgType = htonl (STARTALL);
	    if (!appSendData (MyHandle, (char *)&startAllMsg, sizeof (startAllMsg))) {
		syslog (LOG_OPTIONS, "CSLMonitor(%d): Arbiter has died -- CSLMonitor terminating.",
			getpid ());
		exit (-1);
	    }
	}
	
	/* Wait for RPC requests and incoming messages */

	for (;;) {
	    readfds = svc_fdset;
	    if (Redundant)
		FD_SET (MyHandle, &readfds);
	    Sock = select (dtbsz, &readfds, NULL, NULL, to);
	    switch (Sock) {
		case -1:

		    if (errno != EBADF)
			continue;
		    perror ("CSLMonitor: select");
		    return;
	    
		default:

		    if (Redundant) {

			/* Check arbiter for latest status. */

			OldStatus = Status;
			if (!appGetStatus (MyHandle, &Status)) {
			    syslog (LOG_OPTIONS, "CSLMonitor(%d): Arbiter has died -- CSLMonitor terminating.",
					getpid ());
			    exit (-1);
	    		}

			switch (Status) {
			    case MASTER:
				MyRole = ROLE_MASTER;
				if (Verbose && (OldStatus != Status))
				    syslog (LOG_OPTIONS, "CSLMonitor_svc(%d): Changed to master.", myPid);
				break;
			    case SLAVE:
				MyRole = ROLE_SLAVE;
				if (Verbose && (OldStatus != Status))
				    syslog (LOG_OPTIONS, "CSLMonitor_svc(%d): Running as slave.", myPid);
				break;
			    case MASTER_NOPARTNER:
				MyRole = ROLE_MASTER_DP;
				if (Verbose && (OldStatus != Status))
				    syslog (LOG_OPTIONS, "CSLMonitor_svc(%d): Running as master with no partner.", myPid);
				break;
			}

		    }

		    if (FD_ISSET (MyHandle, &readfds)) { /* Arbiter message */
			while (appRecvData (MyHandle, &msgBuffer, 1)) {
			    switch (ntohl (msgBuffer.msgHeader.msgType)) {

				case EXECUTE:

				    msgBuffer.executeMsg.scriptType = ntohl (msgBuffer.executeMsg.scriptType);
				    msgBuffer.executeMsg.preStartDelay = ntohl (msgBuffer.executeMsg.preStartDelay);
				    msgBuffer.executeMsg.scriptId = ntohl (msgBuffer.executeMsg.scriptId);

				    if (Debug) {
					syslog (LOG_OPTIONS,
						"CSLMonitor_svc(%d): EXECUTE: Type=%d, Name=%s, Delay=%ld, Env=%s\n",
						myPid,
						msgBuffer.executeMsg.scriptType,
						msgBuffer.executeMsg.scriptName,
						msgBuffer.executeMsg.preStartDelay,
						msgBuffer.executeMsg.environmentString);
				    }
				    executeArgStruct.scriptType = msgBuffer.executeMsg.scriptType;
				    executeArgStruct.scriptName = msgBuffer.executeMsg.scriptName;
				    executeArgStruct.preStartDelay = msgBuffer.executeMsg.preStartDelay;
				    executeArgStruct.environmentString =
						msgBuffer.executeMsg.environmentString;
				    executeArgStruct.scriptId = msgBuffer.executeMsg.scriptId;
				    MasterScriptId = msgBuffer.executeMsg.scriptId;
				    csl_monitor_execute_1 (&executeArgStruct, NULL); 
				    break;

				case KILL:

				    msgBuffer.killMsg.scriptId = ntohl (msgBuffer.killMsg.scriptId);

				    if (Debug)
					syslog (LOG_OPTIONS,
						"CSLMonitor_svc(%d): KILL: Id=%d\n",
						myPid,
						msgBuffer.killMsg.scriptId);
				    killArgStruct.scriptId = msgBuffer.killMsg.scriptId;
				    csl_monitor_kill_1 (&killArgStruct, NULL);
				    break;

				case SET_STATUS:

				    msgBuffer.setStatusMsg.scriptId = ntohl (msgBuffer.setStatusMsg.scriptId);
				    msgBuffer.setStatusMsg.status = ntohl (msgBuffer.setStatusMsg.status);

				    if (Debug)
					syslog (LOG_OPTIONS,
						"CSLMonitor_svc(%d): SET_STATUS: Id=%d, Status=%d, Code=%10.3lf\n",
						myPid,
						msgBuffer.setStatusMsg.scriptId,
						msgBuffer.setStatusMsg.status,
						msgBuffer.setStatusMsg.completionCode);
				    setStatusArgStruct.scriptId = msgBuffer.setStatusMsg.scriptId;
				    setStatusArgStruct.status = msgBuffer.setStatusMsg.status;
				    setStatusArgStruct.completionCode =
						atof (msgBuffer.setStatusMsg.completionCode);
				    csl_monitor_set_my_status_1 (&setStatusArgStruct, NULL);
				    break;

				case GET_STATUS:

				    msgBuffer.getStatusMsg.scriptId = ntohl (msgBuffer.getStatusMsg.scriptId);

				    if (Debug)
					syslog (LOG_OPTIONS,
						"CSLMonitor_svc(%d): GET_STATUS: Id=%d\n",
						myPid,
						msgBuffer.getStatusMsg.scriptId);
				    getStatusArgStruct.scriptId = msgBuffer.getStatusMsg.scriptId;
				    csl_monitor_get_status_1 (&getStatusArgStruct, NULL);
				    break;

				case MESSAGE:

				    msgBuffer.messageMsg.scriptId = ntohl (msgBuffer.messageMsg.scriptId);

				    if (Debug)
					syslog (LOG_OPTIONS,
						"CSLMonitor_svc(%d): MESSAGE: Id=%d, Msg=%s\n",
						myPid,
						msgBuffer.messageMsg.scriptId,
						msgBuffer.messageMsg.message);
				    messageArgStruct.scriptId = msgBuffer.messageMsg.scriptId;
				    messageArgStruct.message = msgBuffer.messageMsg.message;
				    csl_monitor_message_1 (&messageArgStruct, NULL);
				    break;

				case RELEASE:

				    msgBuffer.releaseMsg.scriptId = ntohl (msgBuffer.releaseMsg.scriptId);

				    if (Debug)
					syslog (LOG_OPTIONS,
						"CSLMonitor_svc(%d): RELEASE: Id=%d\n",
							myPid,
							msgBuffer.releaseMsg.scriptId);
				    releaseArgStruct.scriptId = msgBuffer.releaseMsg.scriptId;
				    csl_monitor_release_1 (&releaseArgStruct, NULL);
				    break;

				case STARTALL:

				    if (Debug)
					syslog (LOG_OPTIONS, "CSLMonitor_svc(%d): Received STARTALL from slave, starting:",
							myPid);
				    foreachScript (pScript) {
					if (Debug)
					    syslog (LOG_OPTIONS,
							"CSLMonitor_svc(%d): Id=%d, Detached=%d, Type=%d, Name=%s, Env=%s",
							myPid,
							pScript->scriptId, pScript->Detached,
							pScript->scriptType,
						pScript->scriptName, pScript->scriptEnviron);
					if ((pScript->Detached == No) && (pScript->scriptStatus == Running)) {
					    executeMsg.msgType = htonl (EXECUTE);
					    executeMsg.scriptType = htonl (pScript->scriptType);
					    strcpy (executeMsg.scriptName, pScript->scriptName);
					    executeMsg.preStartDelay = 0;
					    strcpy (executeMsg.environmentString,
							pScript->scriptEnviron);
					    executeMsg.scriptId = htonl (pScript->scriptId);
					    appSendData (MyHandle, (char *)&executeMsg,
								sizeof (executeMsg));
					}
				    }
				    break;

				case DETACH:

				    msgBuffer.detachMsg.scriptId = ntohl (msgBuffer.detachMsg.scriptId);

				    if (Debug)
					syslog (LOG_OPTIONS,
						"CSLMonitor(%d): DETACH: Id=%d\n",
						myPid,
						msgBuffer.detachMsg.scriptId);
				    detachArgStruct.scriptId = msgBuffer.detachMsg.scriptId;
				    csl_monitor_detach_1 (&detachArgStruct, NULL);
				    break;

				default:

				    syslog (LOG_OPTIONS,
					"CSLMonitor(%d): Received illegal message, type = %d -- CSLMonitor terminated.", myPid);
				    exit (1);
			    }
			}

		    } else {	/* RPC request */
			svc_getreqset (&readfds);
		    }
		    break;
	    }
	}
}

