/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*
 *	CSLManager.x:	CSL manager protocols
 */
#ifdef RPC_HDR
%#include	<stdio.h>
%#include	<sys/types.h>
%#include	<string.h>
%#include	<rpc/rpc.h>
#ifdef SOLARIS
#include	<dirent.h>
#endif
%#include	"lstLib.h"
%#include	<TankFarm.h>
%
%typedef struct scriptInfo
%	{
%	char		*scriptName;
%	unsigned long	modTime;
%	unsigned long	buildTime;
%	unsigned long	size;
%	unsigned short	mode;
%	int		more;
%	} ScriptInfo;
%
%ScriptInfo *CSLManagerGetScriptInfo();
#endif

const	MYMAXNAMELEN = 255;	/* max length of directory entry */
typedef string 	nametype<MYMAXNAMELEN>;	/* directory entry */
typedef struct getInfo	*GetInfoResult;		/* link in the listing */

/* A node in the directory listing */
struct getInfo
	{
	GetInfoResult	next;		/* next entry */
	nametype	scriptName;	/* name */
	unsigned long	modTime;	/* last edit time */
	unsigned long	buildTime;	/* last build time */
	unsigned long	size;		/* size */
	unsigned short	mode;		/* mode */
	};

union getInfoResult
	switch (int errno)
		{
		case 0:
		    GetInfoResult list;	/* no error: return directory listing */
		default:
		    void;		/* on error: nothing else to return */
		};

program CSLMANAGERPROG
	{
	version CSLMANAGERVERS
		{
		getInfoResult CSL_MANAGER_GET_INFO(nametype) = 1;
		} = 1;
	} = CSL_MANAGER_PROGRAM_NUMBER;

#ifdef RPC_CLNT
%	/*
%	 * This is the CSL Manager client library implementation.  These functions
%	 * are called by the CSL interperter to make the requests.
%	 */
%
%typedef struct
%	{
%	NODE	node;
%	char	*hostName;
%	CLIENT	*client;
%	} Server;
%	
%static LIST	serverList;	/* the list of servers */
%
%#define foreachServer(x) for(x=(Server *)lstFirst(&serverList);x;x=(Server *)lstNext(x))
%
%static Server *getManagerServer(hostName)
%char	*hostName;
%	{
%	static int firstTime = 1;
%	Server	*pServer;
%
%	if (firstTime)
%		{
%		firstTime = 0;
%		lstInit(&serverList);
%		}
%	/*
%	 * search for the named server and return a pointer (or a NULL)
%	 */
%	foreachServer(pServer)
%		if(strcmp(pServer->hostName, hostName) == 0)
%			break;
%	if (!pServer)
%		{
%		/*
%		 * allocate space for the new server and it's name
%		 */
%		pServer = (Server *)calloc(1, sizeof (Server));
%		if (!pServer)
%			return NULL;
%	
%		pServer->hostName = (char *)calloc(strlen(hostName) + 1, sizeof (char));
%		if (!pServer->hostName)
%			return NULL;
%	
%		/*
%		 * initialize the script and add it to the list
%		 */
%		strcpy(pServer->hostName, hostName);
%		pServer->client = clnt_create(hostName, CSLMANAGERPROG, CSLMANAGERVERS, "tcp");
%	/*clnt_pcreateerror("CSLManager");*/
%		lstAdd(&serverList, &pServer->node);
%		}
%
%	return pServer;
%	}
%
%ScriptInfo *CSLManagerGetScriptInfo(hostName, dirname)
%char	*hostName;
%char	*dirname;
%{
%	static getInfoResult *result = NULL;
%	static ScriptInfo scriptInfo;
%	
%	if (!result || !result->getInfoResult_u.list)	/* update the list */
%		{
%		Server *pServer = getManagerServer(hostName);
%
%		if (!pServer || !pServer->client)
%			return NULL;
%
%		if (result)	/* get rid of the old (null) one */	
%			xdr_free((xdrproc_t)xdr_getInfoResult, (char *)result);
%
%		result = csl_manager_get_info_1(&dirname, pServer->client);
%		if (!result || !result->getInfoResult_u.list)
%			return NULL;
%		}
%
%	scriptInfo.scriptName = result->getInfoResult_u.list->scriptName;
%	scriptInfo.modTime = result->getInfoResult_u.list->modTime;
%	scriptInfo.buildTime = result->getInfoResult_u.list->buildTime;
%	scriptInfo.size	= result->getInfoResult_u.list->size;
%	scriptInfo.mode	= result->getInfoResult_u.list->mode;
%	scriptInfo.more	= (result->getInfoResult_u.list->next != NULL);
%
%	result->getInfoResult_u.list = result->getInfoResult_u.list->next;
%
%	return &scriptInfo;
%	}
%
#endif


#ifdef RPC_SVC
%
%	/* remote getInfoResult implementation */
%
%#ifndef SOLARIS
%#include <sys/dir.h>
%#else
%#include <sys/dirent.h>
%#include <dirent.h>
%#endif
%#include <sys/stat.h>
%#include <sys/errno.h>
%
%#include <CSLScript.h>
%
%extern int	errno;
%
%getInfoResult *csl_manager_get_info_1(dirname, req)
%nametype *dirname;
%struct svc_req	*req;
%	{
%	DIR	*dirp;
%#ifdef SOLARIS
%	struct dirent *d;
%#else
%	struct direct *d;
%#endif
%	GetInfoResult nl;
%	GetInfoResult *nlp;
%	static getInfoResult res;	/* static to return the result in */
%
%	xdr_free((xdrproc_t)xdr_getInfoResult, (char *)&res);
%
%	dirp = opendir(*dirname);	/* open the directory */
%	if (!dirp)
%		{
%		res.errno = errno;
%		return &res;
%		}
%
%	nlp = &res.getInfoResult_u.list;
%	while (d = readdir(dirp))	/* make the list */
%		{
%		char		path[MYMAXNAMELEN];
%		struct stat	script_stat;
%		struct stat	parse_tree_stat;
%
%		if (!strcmp(d->d_name, ".") || !strcmp(d->d_name, ".."))
%			/* skip . and .. */
%			continue;
%		if (strlen(d->d_name) > 3 && !strcmp(&d->d_name[strlen(d->d_name)-3], CSL_SCRIPT_EXTENSION))
%			/* skip the parse trees, we only report the scripts */
%			continue;
%		strcat(strcat(strcpy(path, (char *)(*dirname)), "/"), d->d_name);
%		if (stat(path, &script_stat) == -1)
%			{
%			res.errno = errno;
%			closedir(dirp);
%			return &res;
%			}
%		strcat(path, CSL_SCRIPT_EXTENSION);
%		if (stat(path, &parse_tree_stat) == -1)
%			if (errno == ENOENT)
%				/* not parsed yet, mark it very old */
%				parse_tree_stat.st_mtime = 0;
%			else
%				{
%				res.errno = errno;
%				closedir(dirp);
%				return &res;
%				}
%		nl = *nlp = (getInfo *)malloc(sizeof (getInfo));
%		if (!nl)
%			{
%			res.errno = EAGAIN;
%			closedir(dirp);
%			return &res;
%			}
%		nl->scriptName = strcpy(malloc(strlen(d->d_name) + 1), d->d_name);
%		nl->modTime = script_stat.st_mtime;
%		nl->buildTime = parse_tree_stat.st_mtime;
%		nl->size = script_stat.st_size;
%		nl->mode = script_stat.st_mode;
%		nlp = &nl->next;
%		}
%
%	*nlp = (GetInfoResult)NULL;	/* terminate the list and return */
%	res.errno = 0;
%	closedir(dirp);
%	return &res;
%	}
#endif
