/****************************************************************/
/*								*/
/*	This software was developed for DLA by Tate Integrated	*/
/*	Systems, L.P. via a contract with Vitro Corporation.	*/
/*	It is designed to work with TIS4000 Release 4.2.	*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/*****************************************************************************
 *                                                                           *
 *  evolman.c                                                                *
 *                                                                           *
 *****************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/signal.h>

#include <d4all.h>
#undef	BOOL

#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <Xm/Xm.h>
#include <Xm/MessageB.h>
#include <Xm/SelectioB.h>
#include <Xm/Text.h>
#include <Xm/List.h>
#include <Xm/MwmUtil.h>
#include <Mrm/MrmPublic.h>

#include <TFdBASE.h>
#include <evolman.h>
#include <EVOLManager.h>
#include <msapp.h>
#include <lstLib.h>

#ifndef __STDC__
char		*calloc();
#endif
double		atof();
char		*mktemp ();

void		OpenDatabase();
void		ReOpenEvolFiles ();
void		RegisterWidgets();
void		ButtonCallback();
void		EvolDialogCallback();
void		QuantityDialogCallback();
void		EvolListCallback();
void		SegDialogCallback();
void		SegListCallback();
void		SetButtonTitles();
void		DeleteEvol();
void		DeleteSeg();
void		SourceDestinationListCallback();
void		ErrorMessage();
void		ErrorMessageCallback();
void		WarningMessage();
void		WarningMessageOkCallback();
void		WarningMessageCancelCallback();
void		PopupConfirmMessage();
char		*StripWhitespace();
char		*CheckUfloat();
int		CheckDate();
int		CheckTime();
void		formatLabels();
void		getEvol();
void		getEvolSegment();
void		RefreshCallback();
double		remainingGallons();
static void setOperPriv();

char		*hostName = "localhost";	/* currently uses only the local host */
XtAppContext	applContext;
Display		*display;
Widget		appShell;
Widget		evolmanWindow;
MrmType		evolmanWindowClass;
MrmHierarchy	mrmHierarchy;
String		uidFilenames[] =
	{
	"evolman.uid"
	};
#define	NUM_UID_FILES	(sizeof(uidFilenames) / sizeof(uidFilenames[0]))

MrmRegisterArg	namesRegister[] =
	{
	{(String) "RegisterWidgets",			(caddr_t) RegisterWidgets },
	{(String) "ButtonCallback",			(caddr_t) ButtonCallback },
	{(String) "EvolDialogCallback",			(caddr_t) EvolDialogCallback },
	{(String) "EvolListCallback",			(caddr_t) EvolListCallback },
	{(String) "SegDialogCallback",			(caddr_t) SegDialogCallback },
	{(String) "SegListCallback",			(caddr_t) SegListCallback },
	{(String) "SourceDestinationListCallback",	(caddr_t) SourceDestinationListCallback },
	{(String) "QuantityDialogCallback",		(caddr_t) QuantityDialogCallback },
	};
#define	NUM_NAME_REGISTER	(sizeof(namesRegister) / sizeof(namesRegister[0]))

int		editInProgress;
int		insertFlag;
int		selectedEvol;
int		selectedSeg;
int		selectedOrd;
int		forcedUpdate;
EvolRecord	tmpEvol;
SegRecord	tmpSeg;
XmString	insertEvolutionString;
XmString	changeEvolutionString;
XmString	insertEvolButtonString;
XmString	editEvolButtonString;
XmString	deleteEvolButtonString;
XmString	insertSegButtonString;
XmString	editSegButtonString;
XmString	deleteSegButtonString;
XmString	insertSegmentString;
XmString	changeSegmentString;
XmString	errorMessageString;
Widget		errorMessage;
void		(*warningMessageFunc)();
XmString	warningMessageString;
Widget		warningMessage;

Widget			apiEntry;
Widget			clientEntry;
Widget			clientIdEntry;
Widget			dateEntry;
Widget			deleteButton;
Widget			destinationList;
Widget			dicCodeEntry;
Widget			displayButton;
Widget			editButton;
Widget			editEvolDialog;
Widget			editSegDialog;
Widget			emrgstopButton;
Widget			evolBblsButton;
Widget			evolGalsButton;
Widget			evolList;
Widget			evolQuantityEntry;
Widget			exitButton;
Widget			insertButton;
Widget			lineupList;
Widget			manualButton;
Widget			noAutoSwingButton;
Widget			pauseButton;
Widget			pauseOnCompButton;
Widget			pierIssueButton;
Widget			pierReceiveButton;
Widget			printButton;
Widget			productList;
Widget			quantityButton;
Widget			quantityDialog;
Widget			quantityEntry;
Widget			resumeButton;
Widget			segBblsButton;
Widget			segGalsButton;
Widget			segList;
Widget			segQuantityEntry;
Widget			sourceList;
Widget			startButton;
Widget			stopButton;
Widget			timeEntry;
Widget			transferButton;
Widget			truckIssueButton;
Widget			truckReceiveButton;
Widget			yesAutoSwingButton;

C4CODE		codeBaseDesc;

D4DATA		*lineupDesc = NULL;
F4FIELD		*lineup_LINEUPNAME;
F4FIELD		*lineup_METHOD;
F4FIELD		*lineup_FLOWRATE;
F4FIELD		*lineup_RELDSPLY;

D4DATA		*evolDesc = NULL;
F4FIELD		*evol_EVOLNUMBER;
F4FIELD		*evol_PRODUCT;
F4FIELD		*evol_QUANTITY;
F4FIELD		*evol_UNITS;
F4FIELD		*evol_MOVETYPE;
F4FIELD		*evol_AUTOSWNG;
F4FIELD		*evol_CLIENT;
F4FIELD		*evol_CLIENTID;
F4FIELD		*evol_DICCODE;
F4FIELD		*evol_DATE;
F4FIELD		*evol_TIME;
F4FIELD		*evol_TIMESTAMP;
F4FIELD		*evol_STATUS;
F4FIELD		*evol_EVOLMODE;
F4FIELD		*evol_WEIGHTAVG;

D4DATA		*evolSegsDesc = NULL;
F4FIELD		*evolSegs_EVOLNUMBER;
F4FIELD		*evolSegs_SEGNUMBER;
F4FIELD		*evolSegs_ORDERNUM;
F4FIELD		*evolSegs_QUANTITY;
F4FIELD		*evolSegs_UNITS;
F4FIELD		*evolSegs_LINEUPNAME;
F4FIELD		*evolSegs_PAUSE;
F4FIELD		*evolSegs_STATUS;
F4FIELD		*evolSegs_VOLXFERED;
F4FIELD		*evolSegs_ACTFLWRT;
F4FIELD		*evolSegs_SCRIPTID;
F4FIELD		*evolSegs_REQTYPE;
F4FIELD		*evolSegs_AVGAPI;
F4FIELD		*evolSegs_ACTFLWTIME;

D4DATA		*segSrcsDesc = NULL;
F4FIELD		*segSrcs_EVOLNUMBER;
F4FIELD		*segSrcs_SEGNUMBER;
F4FIELD		*segSrcs_SOURCE;
F4FIELD		*segSrcs_API;

D4DATA		*segDestsDesc = NULL;
F4FIELD		*segDests_EVOLNUMBER;
F4FIELD		*segDests_SEGNUMBER;
F4FIELD		*segDests_DEST;

int	Verbose, Debug, Trace, Redundancy;
int	arbiterHandle;
static	void	RegisterWithArbiter ();

/*
 * since setting/getting ONE widget resource happens so often:
 */
#define	XtSetOneValue(widget, resource, value) \
	do \
	    { \
	    Arg		arg[1];	\
	    XtSetArg(arg[0],(resource),(value)); \
	    XtSetValues((widget), arg, 1); \
	    } \
	while (0)

#define	XtGetOneValue(widget, resource, value) \
	do \
	    { \
	    Arg		arg[1]; \
	    XtSetArg(arg[0],(resource),(value)); \
	    XtGetValues((widget), arg, 1); \
	    } \
	while (0)

#define	dupString(d, s, n)	StripWhitespace(strncpy(d, s, n))

/*****************************************************************************
 *                                                                           *
 *      main prgram for evolution interface                                  *
 *                                                                           *
 *****************************************************************************/

main(argc, argv)
int		argc;
char	*argv[];
	{
	Arg	args[10];
	int	n, Cmd, Stat;

	Verbose = 0;
	Debug = 0;
	Trace = 0;
	Redundancy = 0;
	while((Cmd = getopt(argc,argv,"vdt")) != -1) {
	    switch(Cmd) {
		case 'v':
		    Verbose = 1;
		    break;
		case 'd':
		    Debug = 1;
		    Verbose = 1;
		    break;
		case 't':
		    Trace = 1;
		    break;
			
	    }
	}

	/* Register with the arbiter (if necessary). */

	RegisterWithArbiter ();

	/* If we are the slave the we must hold here until we become master */

	if (Redundancy) {
	    for (;;) {
		appGetStatus (arbiterHandle, &Stat);
		if (Stat != SLAVE) break;
		if (Verbose)
		    printf ("evolman: In slave mode, waiting to become master\n");
		sleep (5);
	    }
	}

	/*
	 * initialize variables
	 */
	editInProgress = FALSE;
	insertFlag = FALSE;
	selectedEvol = 0;
	selectedSeg = 0;
	selectedOrd =0;
	errorMessage = NULL;
	errorMessageString = NULL;
	warningMessage = NULL;
	warningMessageString = NULL;
	insertEvolutionString = XmStringCreate("Insert Evolution", XmSTRING_DEFAULT_CHARSET);
	changeEvolutionString = XmStringCreate("Change Evolution", XmSTRING_DEFAULT_CHARSET);
	insertSegmentString = XmStringCreate("Insert Segment", XmSTRING_DEFAULT_CHARSET);
	changeSegmentString = XmStringCreate("Change Segment", XmSTRING_DEFAULT_CHARSET);
	insertEvolButtonString = XmStringCreate("Insert Evol", XmSTRING_DEFAULT_CHARSET);
	editEvolButtonString = XmStringCreate("Edit Evol", XmSTRING_DEFAULT_CHARSET);
	deleteEvolButtonString = XmStringCreate("Delete Evol", XmSTRING_DEFAULT_CHARSET);
	insertSegButtonString = XmStringCreate("Insert Seg", XmSTRING_DEFAULT_CHARSET);
	editSegButtonString = XmStringCreate("Edit Seg", XmSTRING_DEFAULT_CHARSET);
	deleteSegButtonString = XmStringCreate("Delete Seg", XmSTRING_DEFAULT_CHARSET);

	/*
	 * initialize and open display
	 */
	OpenDatabase ();
	MrmInitialize();
	XtToolkitInitialize();
	applContext = XtCreateApplicationContext();
	display = XtOpenDisplay(applContext, NULL, "evolman", "Evolman", NULL, 0, &argc, argv);
	if (!display)
		printf("Cannot open display\n"), exit(1);

	appShell = XtAppCreateShell("Evolution Management", "Evolman",
			applicationShellWidgetClass, display, NULL, 0);

	/*
	 * create main window widget and its children
	 */
	if (MrmOpenHierarchy(NUM_UID_FILES, uidFilenames, NULL, &mrmHierarchy) != MrmSUCCESS)
		printf("Cannot open UID file %s\n", uidFilenames[0]), exit(1);

	MrmRegisterNames(namesRegister, NUM_NAME_REGISTER);
	if (MrmFetchWidget(mrmHierarchy, "evolmanWindow", appShell, &evolmanWindow,
			&evolmanWindowClass) != MrmSUCCESS)
		printf("Cannot fetch evolution management window widget\n"), exit(1);

	/*
	 * create error msg dialog box
	 */
	n = 0;
	XtSetArg(args[n], XmNmessageAlignment, XmALIGNMENT_CENTER); n++;
	XtSetArg(args[n], XmNdialogType, XmDIALOG_ERROR); n++;
	XtSetArg(args[n], XmNdeleteResponse, XmDO_NOTHING); n++;
	XtSetArg(args[n], XmNdialogStyle, XmDIALOG_APPLICATION_MODAL); n++;
	XtSetArg(args[n], XmNtitle, "Error"); n++;
	errorMessage = XmCreateMessageDialog(evolmanWindow, "Error Message", args, n);
	XtUnmanageChild(XmMessageBoxGetChild(errorMessage, XmDIALOG_HELP_BUTTON));
	XtUnmanageChild(XmMessageBoxGetChild(errorMessage, XmDIALOG_CANCEL_BUTTON));
	XtAddCallback(XmMessageBoxGetChild(errorMessage, XmDIALOG_OK_BUTTON),
			XmNactivateCallback, ErrorMessageCallback, NULL);

	/*
	 * create warning msg dialog box
	 */
	n = 0;
	XtSetArg(args[n], XmNmessageAlignment, XmALIGNMENT_CENTER); n++;
	XtSetArg(args[n], XmNdialogType, XmDIALOG_WARNING); n++;
	XtSetArg(args[n], XmNdeleteResponse, XmDO_NOTHING); n++;
	XtSetArg(args[n], XmNdialogStyle, XmDIALOG_APPLICATION_MODAL); n++;
	XtSetArg(args[n], XmNtitle, "Warning"); n++;
	warningMessage = XmCreateQuestionDialog(evolmanWindow, "Warning Message", args, n);
	XtUnmanageChild(XmMessageBoxGetChild(warningMessage, XmDIALOG_HELP_BUTTON));
	XtAddCallback(XmMessageBoxGetChild(warningMessage, XmDIALOG_CANCEL_BUTTON),
			XmNactivateCallback, WarningMessageCancelCallback, NULL);
	XtSetOneValue(XmMessageBoxGetChild(warningMessage, XmDIALOG_CANCEL_BUTTON),
			XmNlabelString, XmStringCreate("NO", XmSTRING_DEFAULT_CHARSET));
	XtAddCallback(XmMessageBoxGetChild(warningMessage, XmDIALOG_OK_BUTTON),
			XmNactivateCallback, WarningMessageOkCallback, NULL);
	XtSetOneValue(XmMessageBoxGetChild(warningMessage, XmDIALOG_OK_BUTTON),
			XmNlabelString, XmStringCreate("YES", XmSTRING_DEFAULT_CHARSET));

	EVOLManagerAttach(hostName);		/* allow communications */		

	/*
	 * realize main window
	 */
	XtManageChild(evolmanWindow);
	XtRealizeWidget(appShell);

	/*
	 * get and dispatch events
	 */
	forcedUpdate = TRUE;
	RefreshCallback();
	XtAppMainLoop(applContext);
	}

/*----------*/
static void RegisterWithArbiter ()
{

	/* Keep trying to register with arbiter until successful. */

	int Stat;
	if (appRegister ("evolman", &arbiterHandle)) {
	    appGetStatus (arbiterHandle, &Stat);
	    switch (Stat) {
		case MASTER:
		    if (Verbose)
			printf ("evolman: Running as master.\n");
		    break;
		case SLAVE:
		    if (Verbose)
			printf ("evolman: Running as slave.\n");
		    break;
		case MASTER_NOPARTNER:
		    if (Verbose)
			printf ("evolman: Running as master with no partner.\n");
		    break;
	    }
	    Redundancy = 1;
	} else {
	    if (Verbose)
		printf ("evolman: Running without redundancy.\n");
	}
}

/*****************************************************************************
 *      OpenDatabase                                                         *
 *                                                                           *
 *      Inputs                                                               *
 *          none                                                             *
 *                                                                           *
 *      Return Value                                                         *
 *          none                                                             *
 *                                                                           *
 *      Description                                                          *
 *          open database file                                               *
 *****************************************************************************/

void
OpenAllFiles ()

    {

	/* This will close and reopen files if they are currently open */

	if (lineupDesc)
	    d4close (lineupDesc);
	if (!(lineupDesc = d4open(&codeBaseDesc, lineupName)))
		printf("Cannot open lineup database file \"%s\"\n", lineupName), exit(0);
	lineup_LINEUPNAME = d4field (lineupDesc, "LINEUPNAME");
	lineup_METHOD = d4field (lineupDesc, "METHOD");
	lineup_FLOWRATE = d4field (lineupDesc, "FLOWRATE");
	lineup_RELDSPLY = d4field (lineupDesc, "RELDSPLY");

	if (evolDesc)
	    d4close (evolDesc);
	if (!(evolDesc = d4open(&codeBaseDesc, evolName)))
		printf("Cannot open evolution database file \"%s\"\n", evolName), exit(0);
	evol_EVOLNUMBER = d4field (evolDesc, "EVOLNUMBER");
	evol_PRODUCT = d4field (evolDesc, "PRODUCT");
	evol_QUANTITY = d4field (evolDesc, "QUANTITY");
	evol_UNITS = d4field (evolDesc, "UNITS");
	evol_MOVETYPE = d4field (evolDesc, "MOVETYPE");
	evol_AUTOSWNG = d4field (evolDesc, "AUTOSWNG");
	evol_CLIENT = d4field (evolDesc, "CLIENT");
	evol_CLIENTID = d4field (evolDesc, "CLIENTID");
	evol_DICCODE = d4field (evolDesc, "DICCODE");
	evol_DATE = d4field (evolDesc, "DATE");
	evol_TIME = d4field (evolDesc, "TIME");
	evol_TIMESTAMP = d4field (evolDesc, "TIMESTAMP");
	evol_STATUS = d4field (evolDesc, "STATUS");
	evol_EVOLMODE = d4field (evolDesc, "EVOLMODE");
	evol_WEIGHTAVG = d4field (evolDesc, "WEIGHTAVG");

	if (evolSegsDesc)
	    d4close (evolSegsDesc);
	if (!(evolSegsDesc = d4open(&codeBaseDesc, evolSegsName)))
		printf("Cannot open evolution segments database file \"%s\"\n", evolSegsName), exit(0);
	evolSegs_EVOLNUMBER = d4field (evolSegsDesc, "EVOLNUMBER");
	evolSegs_SEGNUMBER = d4field (evolSegsDesc, "SEGNUMBER");
	evolSegs_ORDERNUM = d4field (evolSegsDesc, "ORDERNUM");
	evolSegs_QUANTITY = d4field (evolSegsDesc, "QUANTITY");
	evolSegs_UNITS = d4field (evolSegsDesc, "UNITS");
	evolSegs_LINEUPNAME = d4field (evolSegsDesc, "LINEUPNAME");
	evolSegs_PAUSE = d4field (evolSegsDesc, "PAUSE");
	evolSegs_STATUS = d4field (evolSegsDesc, "STATUS");
	evolSegs_VOLXFERED = d4field (evolSegsDesc, "VOLXFERED");
	evolSegs_ACTFLWRT = d4field (evolSegsDesc, "ACTFLWRT");
	evolSegs_SCRIPTID = d4field (evolSegsDesc, "SCRIPTID");
	evolSegs_REQTYPE = d4field (evolSegsDesc, "REQTYPE");
	evolSegs_AVGAPI = d4field (evolSegsDesc, "AVGAPI");
	evolSegs_ACTFLWTIME = d4field (evolSegsDesc, "ACTFLWTIME");

	if (segSrcsDesc)
	    d4close (segSrcsDesc);
	if (!(segSrcsDesc = d4open(&codeBaseDesc, segSrcsName)))
		printf("Cannot open segment sources database file \"%s\"\n", segSrcsName), exit(0);
	segSrcs_EVOLNUMBER = d4field (segSrcsDesc, "EVOLNUMBER");
	segSrcs_SEGNUMBER = d4field (segSrcsDesc, "SEGNUMBER");
	segSrcs_SOURCE = d4field (segSrcsDesc, "SOURCE");
	segSrcs_API = d4field (segSrcsDesc, "API");

	if (segDestsDesc)
	    d4close (segDestsDesc); 
	if (!(segDestsDesc = d4open(&codeBaseDesc, segDestsName)))
		printf("Cannot open segment destinations database file \"%s\"\n", segDestsName), exit(0);
	segDests_EVOLNUMBER = d4field (segDestsDesc, "EVOLNUMBER");
	segDests_SEGNUMBER = d4field (segDestsDesc, "SEGNUMBER");
	segDests_DEST = d4field (segDestsDesc, "DEST");


    }

void
OpenDatabase()
	{
	d4init(&codeBaseDesc);
	codeBaseDesc.read_lock = 0;
	codeBaseDesc.create_error = 0;
	codeBaseDesc.field_name_error = 0;
	codeBaseDesc.go_error = 0;
	codeBaseDesc.open_error = 0;
	codeBaseDesc.skip_error = 0;
	codeBaseDesc.tag_name_error = 0;

	OpenAllFiles ();

	}


/*****************************************************************************
 *      RegisterWidgets                                                      *
 *                                                                           *
 *      Inputs                                                               *
 *          widget          widget descriptor                                *
 *          widgetID        widget ID                                        *
 *          widgetData      widget data                                      *
 *                                                                           *
 *      Return Value                                                         *
 *          none                                                             *
 *                                                                           *
 *      Description                                                          *
 *          register a widget that was created from the UIL.                 *
 *****************************************************************************/

void
RegisterWidgets(widget, widgetID, widgetData)
Widget 			widget;
int			*widgetID;
XmAnyCallbackStruct	*widgetData;
	{
	char	txtBuf[BUFSIZ];
	int		n;

	switch(*widgetID)
	    {
	    case EVOL_COLUMNS1:
	    case EVOL_COLUMNS2:
	    case SEG_COLUMNS1:
	    case SEG_COLUMNS2:
		formatLabels(*widgetID, txtBuf);
		XtSetOneValue(widget, XmNlabelString, XmStringCreate(txtBuf, XmSTRING_DEFAULT_CHARSET));
		break;
	    case EVOL_LIST:
		evolList = widget;
		break;
	    case SEG_LIST:
		segList = widget;
		break;
	    case START_BUTTON:
		startButton = widget;
		break;
	    case STOP_BUTTON:
		stopButton = widget;
		break;
	    case PAUSE_BUTTON:
		pauseButton = widget;
		break;
	    case RESUME_BUTTON:
		resumeButton = widget;
		break;
	    case MAN_EVOL_BUTTON:
		manualButton = widget;
		break;
	    case QUANTITY_BUTTON:
		quantityButton = widget;
		break;
	    case EMRG_STOP_BUTTON:
		emrgstopButton = widget;
		break;
	    case INSERT_BUTTON:
		insertButton = widget;
		break;
	    case EDIT_BUTTON:
		editButton = widget;
		break;
	    case DELETE_BUTTON:
		deleteButton = widget;
		break;
	    case PRINT_BUTTON:
		printButton = widget;
		break;
	    case DISPLAY_BUTTON:
		displayButton = widget;
		break;
	    case EXIT_BUTTON:
		exitButton = widget;
		break;
	    case EDIT_EVOL_DIALOG:
		editEvolDialog = widget;
		break;
	    case EVOL_QUANTITY_ENTRY:
		XmTextSetMaxLength(evolQuantityEntry = widget, QUANTITY_ENTRY_SIZE);
		break;
	    case EVOL_BBLS_BUTTON:
		XtSetOneValue(evolBblsButton = widget, XmNset, True);
		break;
	    case EVOL_GALS_BUTTON:
		XtSetOneValue(evolGalsButton = widget, XmNset, False);
		break;
	    case PRODUCT_LIST:
		productList = widget;
		break;
	    case PIER_ISSUE_BUTTON:
		XtSetOneValue(pierIssueButton = widget, XmNset, True);
		break;
	    case PIER_RECEIVE_BUTTON:
		XtSetOneValue(pierReceiveButton = widget, XmNset, False);
		break;
	    case TRUCK_ISSUE_BUTTON:
		XtSetOneValue(truckIssueButton = widget, XmNset, False);
		break;
	    case TRUCK_RECEIVE_BUTTON:
		XtSetOneValue(truckReceiveButton = widget, XmNset, False);
		break;
	    case TRANSFER_BUTTON:
		XtSetOneValue(transferButton = widget, XmNset, False);
		break;
	    case YES_AUTO_SWING_BUTTON:
		XtSetOneValue(yesAutoSwingButton = widget, XmNset, True);
		break;
	    case NO_AUTO_SWING_BUTTON:
		XtSetOneValue(noAutoSwingButton = widget, XmNset, False);
		break;
	    case CLIENT_ENTRY:
		XmTextSetMaxLength(clientEntry = widget, CLIENT_ENTRY_SIZE);
		break;
	    case CLIENT_ID_ENTRY:
		XmTextSetMaxLength(clientIdEntry = widget, CLIENT_ID_ENTRY_SIZE);
		break;
	    case DIC_CODE_ENTRY:
		XmTextSetMaxLength(dicCodeEntry = widget, DIC_CODE_ENTRY_SIZE);
		break;
	    case DATE_ENTRY:
		XmTextSetMaxLength(dateEntry = widget, DATE_STRING_ENTRY_SIZE);
		break;
	    case TIME_ENTRY:
		XmTextSetMaxLength(timeEntry = widget, TIME_STRING_ENTRY_SIZE);
		break;
	    case EDIT_SEG_DIALOG:
		editSegDialog = widget;
		break;
	    case SEG_QUANTITY_ENTRY:
		XmTextSetMaxLength(segQuantityEntry = widget, QUANTITY_ENTRY_SIZE);
		break;
	    case SEG_BBLS_BUTTON:
		XtSetOneValue(segBblsButton = widget, XmNset, True);
		break;
	    case SEG_GALS_BUTTON:
		XtSetOneValue(segGalsButton = widget, XmNset, False);
		break;
	    case SOURCE_LIST:
		sourceList = widget;
		break;
	    case DESTINATION_LIST:
		destinationList = widget;
		break;
	    case LINEUP_LIST:
		lineupList = widget;
		break;
	    case PAUSE_ON_COMP_BUTTON:
		pauseOnCompButton = widget;
		break;
	    case API_ENTRY:
		XmTextSetMaxLength(apiEntry = widget, API_ENTRY_SIZE);
		break;
	    case QUANTITY_DIALOG:
		quantityDialog = widget;
		break;
	    case QUANTITY_ENTRY:XmTextSetMaxLength(quantityEntry = widget, QUANTITY_ENTRY_SIZE);
		break;
	}
}

/*
 * a signal handler for CHILD process completions to keep zombies from lasting forever
 */
#include <sys/wait.h>

void childHandler(sig)
int sig;
	{
#ifdef SOLARIS
	int	val;
	while (waitpid ((pid_t)0, &val, WNOHANG|WUNTRACED) > 0)
#else
	while (wait3((union wait *)NULL, WNOHANG|WUNTRACED, (struct rusage *)NULL) > 0)
#endif
		;
	}

/*****************************************************************************
 *      ButtonCallback                                                       *
 *                                                                           *
 *      Inputs                                                               *
 *          widget          widget descriptor                                *
 *          widgetID        widget ID                                        *
 *          widgetData      widget data                                      *
 *                                                                           *
 *      Return Value                                                         *
 *          none                                                             *
 *                                                                           *
 *      Description                                                          *
 *          callback for button presses                                      *
 *****************************************************************************/

void
ButtonCallback(widget, widgetID, widgetData)
Widget 			widget;
int			*widgetID;
XmAnyCallbackStruct	*widgetData;
{
	char				txtBuf[BUFSIZ];
	int				i, j;
	FILE				*printFileDesc;
#define	PRINT_FILE_TEMPLATE	"/tmp/evolmanXXXXXX"
	char				printFileName[sizeof PRINT_FILE_TEMPLATE];
	EvolutionProcessorRequestType	requestType;
	MakeRequestResult		*requestResult;	/* a place to store the request result */
	int				segment	= selectedSeg ? tmpSeg.orderNumber : NO_SEGMENT;

	switch (*widgetID)
	  {
	  case HELP_BUTTON:
		system("evolman.help");
		break;

	  case EMRG_STOP_BUTTON:
		if ((requestResult = EVOLManagerPerformESTOP(hostName))->status != TF_OK)
			{
			ErrorMessage(requestResult->statusText);
			if (requestResult->status == TF_ERROR)
				{
				EVOLManagerDetach();
				EVOLManagerAttach(hostName);
				}
			}
		break;

	  case START_BUTTON:
	  case STOP_BUTTON:
	  case PAUSE_BUTTON:
	  case RESUME_BUTTON:
	  case MAN_EVOL_BUTTON:
	    switch (*widgetID)
		  {
		  case START_BUTTON:	requestType = START;	break;
		  case STOP_BUTTON:	requestType = STOP;	break;
		  case PAUSE_BUTTON:	requestType = PAUSE;	break;
		  case RESUME_BUTTON:	requestType = RESUME;	break;
		  case MAN_EVOL_BUTTON:	requestType = MANUAL;	break;
		  }
		if ((requestResult = EVOLManagerMakeRequest(hostName, requestType,
				selectedEvol, segment))->status != TF_OK)
			{
			ErrorMessage(requestResult->statusText);
			if (requestResult->status == TF_ERROR)
				{
				EVOLManagerDetach();	/* TF_ERROR means "no server", try to connect to a new one */
				EVOLManagerAttach(hostName);
				}
			}
		break;

	  case QUANTITY_BUTTON:
		editInProgress	= TRUE;
		SetButtonTitles();
		XtSetOneValue(evolList, XmNsensitive, False);
		XtSetOneValue(segList, XmNsensitive, False);
		XmTextSetString(quantityEntry, tmpSeg.quantityEntry);
		XtManageChild(quantityDialog);
		break;

	  case INSERT_BUTTON:
		insertFlag = TRUE;
		editInProgress = TRUE;
		SetButtonTitles();
		XtSetOneValue(evolList, XmNsensitive, False);
		XtSetOneValue(segList, XmNsensitive, False);
		if (!selectedEvol)
			{
			XmListDeselectAllItems(evolList);
			XtSetOneValue(editEvolDialog, XmNdialogTitle, insertEvolutionString);
			setProductNames("");
			XtManageChild(editEvolDialog);
			}
		else
			{
			int	barrelsSpecified = (strcmp(tmpEvol.quantityUnitsEntry, BBLS_STRING) == 0);

			XtSetOneValue(editSegDialog, XmNdialogTitle, insertSegmentString);
			XtSetOneValue(segBblsButton, XmNset, barrelsSpecified);
			XtSetOneValue(segGalsButton, XmNset, !barrelsSpecified);
			sprintf(txtBuf, "%-*.3lf", QUANTITY_ENTRY_SIZE, remainingGallons() /
					 (barrelsSpecified ? 42.0 : 1.0));
			XmTextSetString(segQuantityEntry, StripWhitespace(txtBuf));
			XmTextSetString(apiEntry, "");
			XtSetOneValue(apiEntry, XmNsensitive, strcmp(tmpEvol.movementTypeEntry,
				PIER_RECEIVE_STRING) == 0 ||
			 strcmp(tmpEvol.movementTypeEntry, TRUCK_RECEIVE_STRING) == 0);
			setSourceNames();
			setDestinationNames();
			setLineupNames("");
			XtManageChild(editSegDialog);
			}
		OpenAllFiles ();
		break;

	  case EDIT_BUTTON:
		insertFlag = FALSE;
		editInProgress = TRUE;
		SetButtonTitles();
		XtSetOneValue(evolList, XmNsensitive, False);
		XtSetOneValue(segList, XmNsensitive, False);
		if (selectedSeg)
			{
			XtSetOneValue(editSegDialog, XmNdialogTitle, changeSegmentString);
			setSourceNames();
			setDestinationNames();
			setLineupNames(tmpSeg.lineupEntry);
			XmTextSetString(segQuantityEntry, tmpSeg.quantityEntry);
			XtSetOneValue(segBblsButton, XmNset, strcmp(tmpSeg.quantityUnitsEntry,
				BBLS_STRING) == 0);
			XtSetOneValue(segGalsButton, XmNset, strcmp(tmpSeg.quantityUnitsEntry,
				GALS_STRING) == 0);
			XtSetOneValue(pauseOnCompButton, XmNset, strcmp(tmpSeg.pauseEntry,
				YES_STRING) == 0);
			XmTextSetString(apiEntry, tmpSeg.avgapi);
			XtSetOneValue(apiEntry, XmNsensitive, strcmp(tmpEvol.movementTypeEntry,
				PIER_RECEIVE_STRING) == 0 ||
			  strcmp(tmpEvol.movementTypeEntry, TRUCK_RECEIVE_STRING) == 0);
			XtManageChild(editSegDialog);
			}
		else
			{
			XtSetOneValue(editEvolDialog, XmNdialogTitle, changeEvolutionString);
			setProductNames(tmpEvol.productEntry);
			XmTextSetString(evolQuantityEntry, tmpEvol.quantityEntry);
			XtSetOneValue(evolBblsButton, XmNset, strcmp(tmpEvol.quantityUnitsEntry,
				BBLS_STRING) == 0);
			XtSetOneValue(evolGalsButton, XmNset, strcmp(tmpEvol.quantityUnitsEntry,
				GALS_STRING) == 0);
			XtSetOneValue(pierIssueButton, XmNset, strcmp(tmpEvol.movementTypeEntry,
				PIER_ISSUE_STRING) == 0);
			XtSetOneValue(truckIssueButton, XmNset, strcmp(tmpEvol.movementTypeEntry,
				TRUCK_ISSUE_STRING) == 0);
			XtSetOneValue(pierReceiveButton, XmNset, strcmp(tmpEvol.movementTypeEntry,
				PIER_RECEIVE_STRING) == 0);
			XtSetOneValue(truckReceiveButton, XmNset, strcmp(tmpEvol.movementTypeEntry,
				TRUCK_RECEIVE_STRING) == 0);
			XtSetOneValue(transferButton, XmNset, strcmp(tmpEvol.movementTypeEntry,
				TRANSFER_STRING) == 0);
			XtSetOneValue(yesAutoSwingButton, XmNset, strcmp(tmpEvol.autoSwingEntry,
				YES_STRING) == 0);
			XtSetOneValue(noAutoSwingButton, XmNset, strcmp(tmpEvol.autoSwingEntry,
				NO_STRING) == 0);
			XmTextSetString(clientEntry, tmpEvol.clientEntry);
			XmTextSetString(clientIdEntry, tmpEvol.clientIdEntry);
			XmTextSetString(dicCodeEntry, tmpEvol.dicCodeEntry);
			XmTextSetString(dateEntry, tmpEvol.dateEntry);
			XmTextSetString(timeEntry, tmpEvol.timeEntry);
			XtManageChild(editEvolDialog);
			}
		OpenAllFiles ();
		break;

	  case DELETE_BUTTON:
		WarningMessage("Are you sure you want to delete?", selectedSeg ? DeleteSeg : DeleteEvol);
		break;

	  case PRINT_BUTTON:
		if (!(printFileDesc = fopen(mktemp(strcpy(printFileName, PRINT_FILE_TEMPLATE)), "w")))
			{
			sprintf(txtBuf, "Could not create temporary file: %s", printFileName);
			ErrorMessage(txtBuf);
			break;
			}
		d4tag_select(evolDesc, d4tag(evolDesc, "TIMESTAMP"));
		if (d4top(evolDesc) == 0)
			{
			do
				{
				EvolRecord	evol;
				int		i;

				formatLabels(EVOL_COLUMNS1, txtBuf), fprintf(printFileDesc, "%s\n", txtBuf);
				formatLabels(EVOL_COLUMNS2, txtBuf), fprintf(printFileDesc, "%s\n", txtBuf);
				for ( i = strlen(txtBuf); i > 0; i--)
					fprintf(printFileDesc, "-");
				fprintf(printFileDesc, "\n");
				getEvol(&evol);
				formatEvolutionString(&evol, txtBuf), fprintf(printFileDesc, "%s\n\n\n\n", txtBuf);
				formatLabels(SEG_COLUMNS1, txtBuf), fprintf(printFileDesc, "%s\n", txtBuf);
				formatLabels(SEG_COLUMNS2, txtBuf), fprintf(printFileDesc, "%s\n", txtBuf);
				for ( i = strlen(txtBuf); i > 0; i--)
					fprintf(printFileDesc, "-");
				fprintf(printFileDesc, "\n");

				d4tag_select(evolSegsDesc, d4tag(evolSegsDesc, "ORDERTAG"));
				if (d4top(evolSegsDesc) == 0)
					do
						{
						SegRecord	seg;

						getEvolSegment(&seg);
						if (seg.evolNumber == evol.evolNumber)
							formatSegmentString(&seg, txtBuf),
					 			fprintf(printFileDesc, "%s\n", txtBuf);
						d4unlock_all(evolSegsDesc);
						}
					while (d4skip(evolSegsDesc, 1) == 0);
				d4unlock_all(evolSegsDesc);

				fprintf(printFileDesc, "\f");
				d4unlock_all(evolDesc);
				}
			while (d4skip(evolDesc, 1) == 0);
			}
		d4unlock_all(evolDesc);
		fclose(printFileDesc);
		sprintf(txtBuf, "lpr -p -r -T\"Evolution Management\" %s", printFileName);
		system(txtBuf);
		break;

	  case DISPLAY_BUTTON:
		txtBuf[0] = 0;
		if (d4top(lineupDesc) == 0)
			do {
				if (strcmp(tmpSeg.lineupEntry, StripWhitespace(f4str(lineup_LINEUPNAME))) == 0)
					{
					char	*dmPath	= getenv("EPICS_DISPLAY_PATH");

					if (dmPath)
						strcat(strcpy(txtBuf, dmPath), "/");
					else
						txtBuf[0] = 0;
					sprintf(&txtBuf[strlen(txtBuf)], "dm %s", f4str(lineup_RELDSPLY));
					break;
					}
				d4unlock_all (lineupDesc);
			} while (d4skip(lineupDesc, 1) == 0);
		d4unlock_all(lineupDesc);
		if (txtBuf[0])
			{
			signal(SIGCHLD, childHandler);
			if (fork() == 0)
				{
				execlp("/bin/sh", "sh", "-c", txtBuf, 0);	/* the child becomes dm */
				exit(1);					/* in case execlp() fails */
				}
			}
	  	break;

	  case EXIT_BUTTON:
		d4close(lineupDesc);
		d4close(evolDesc);
		d4close(evolSegsDesc);
		d4close(segSrcsDesc);
		d4close(segDestsDesc);
		EVOLManagerDetach();		/* clean up communications */
		exit(0);
		break;
	  }
	}

/*****************************************************************************
 *      EvolDialogCallback                                                   *
 *                                                                           *
 *      Inputs                                                               *
 *          widget          widget descriptor                                *
 *          widgetID        widget ID                                        *
 *          widgetData      widget data                                      *
 *                                                                           *
 *      Return Value                                                         *
 *          none                                                             *
 *                                                                           *
 *      Description                                                          *
 *          callback for evolution dialog button presses                     *
 *****************************************************************************/

void
EvolDialogCallback(widget, widgetID, widgetData)
Widget 			widget;
int			*widgetID;
XmAnyCallbackStruct	*widgetData;
{
	char	*quantityPtr;
	char	*quantityUnitsPtr;
	char	*movementTypePtr;
	char	*autoSwingPtr;
	char	*clientPtr;
	char	*clientIdPtr;
	char	*dicCodePtr;
	char	*datePtr;
	char	*timePtr;
	int	count;
	char	productBuf[PRODUCT_ENTRY_SIZE+1];

	switch(*widgetID)
	  {
	  case EVOL_BBLS_BUTTON:
	  case EVOL_GALS_BUTTON:
		XtSetOneValue(evolBblsButton, XmNset, *widgetID == EVOL_BBLS_BUTTON);
		XtSetOneValue(evolGalsButton, XmNset, *widgetID == EVOL_GALS_BUTTON);
		break;

	  case PIER_ISSUE_BUTTON:
	  case PIER_RECEIVE_BUTTON:
	  case TRUCK_ISSUE_BUTTON:
	  case TRUCK_RECEIVE_BUTTON:
	  case TRANSFER_BUTTON:
		XtSetOneValue(pierIssueButton, XmNset, *widgetID == PIER_ISSUE_BUTTON);
		XtSetOneValue(pierReceiveButton, XmNset, *widgetID == PIER_RECEIVE_BUTTON);
		XtSetOneValue(truckIssueButton,	 XmNset, *widgetID == TRUCK_ISSUE_BUTTON);
		XtSetOneValue(truckReceiveButton, XmNset, *widgetID == TRUCK_RECEIVE_BUTTON);
		XtSetOneValue(transferButton, XmNset, *widgetID == TRANSFER_BUTTON);
		break;

	  case YES_AUTO_SWING_BUTTON:
	  case NO_AUTO_SWING_BUTTON:
		XtSetOneValue(yesAutoSwingButton, XmNset, *widgetID == YES_AUTO_SWING_BUTTON);
		XtSetOneValue(noAutoSwingButton, XmNset, *widgetID == NO_AUTO_SWING_BUTTON);
		break;

	  case CANCEL_EDIT_EVOL_BUTTON:
		XtUnmanageChild(editEvolDialog);
		XtSetOneValue(evolList, XmNsensitive, True);
		XtSetOneValue(segList, XmNsensitive, True);
		editInProgress = FALSE;
		forcedUpdate = TRUE;
		break;

	  case OK_EDIT_EVOL_BUTTON:
		XtGetOneValue(productList, XmNselectedItemCount, &count);
	  	if (count == 1)
			{
			XmStringTable	item;
			char		*p;

			XtGetOneValue(productList, XmNselectedItems, &item);
			XmStringGetLtoR(item[0], XmSTRING_DEFAULT_CHARSET, &p);
			dupString(productBuf, p, (sizeof productBuf)-1);
			XtFree(p);
			}
		else
			{
			ErrorMessage("Must select from \"Product\" list");
			break;
			}
		quantityPtr = StripWhitespace(XmTextGetString(evolQuantityEntry));
		quantityUnitsPtr = XmToggleButtonGetState(evolBblsButton) ? BBLS_STRING : GALS_STRING;
		movementTypePtr = XmToggleButtonGetState(pierIssueButton) ? PIER_ISSUE_STRING :
				XmToggleButtonGetState(truckIssueButton) ? TRUCK_ISSUE_STRING :
				XmToggleButtonGetState(pierReceiveButton) ? PIER_RECEIVE_STRING :
				XmToggleButtonGetState(truckReceiveButton) ? TRUCK_RECEIVE_STRING : TRANSFER_STRING;
		autoSwingPtr = XmToggleButtonGetState(yesAutoSwingButton) ? YES_STRING : NO_STRING;
		clientPtr = StripWhitespace(XmTextGetString(clientEntry));
		clientIdPtr = StripWhitespace(XmTextGetString(clientIdEntry));
		dicCodePtr = StripWhitespace(XmTextGetString(dicCodeEntry));
		datePtr = StripWhitespace(XmTextGetString(dateEntry));
		timePtr = StripWhitespace(XmTextGetString(timeEntry));

		if (!*quantityPtr)
			ErrorMessage("Must fill in \"Quantity\" field");
		else if (!*clientPtr)
			ErrorMessage("Must fill in \"Client\" field");
		else if (!*clientIdPtr)
			ErrorMessage("Must fill in \"Client ID Code\" field");
		else if (!*dicCodePtr)
			ErrorMessage("Must fill in \"DIC Code\" field");
		else if (!*datePtr)
			ErrorMessage("Must fill in \"Date\" field");
		else if (!*timePtr)
			ErrorMessage("Must fill in \"Time\" field");
		else if (!(quantityPtr = CheckUfloat(quantityPtr)))
			ErrorMessage("Invalid entry in \"Quantity\" field (a valid entry example is: 19500)");
		else if (!CheckDate(datePtr))
			ErrorMessage("Invalid entry in \"Date\" field (a valid entry example is: 06/07/93)");
		else if (!CheckTime(timePtr))
			ErrorMessage("Invalid entry in \"Time\" field (a valid entry example is: 0830)");
		else
			{
			char	evolNumber[EVOL_NUMBER_ENTRY_SIZE+1];
			char	timestamp[TIMESTAMP_ENTRY_SIZE+1];
			
			d4tag_select(evolDesc, d4tag(evolDesc, "EVOLTAG"));
			if (insertFlag)
				{
				selectedEvol = (d4bottom(evolDesc) == 0) ? f4int(evol_EVOLNUMBER) + 1 : 1;
				d4append_start(evolDesc);
				}
			else
				{
				char	searchString[EVOL_NUMBER_ENTRY_SIZE+1];

				sprintf(searchString, "%*d", EVOL_NUMBER_ENTRY_SIZE, selectedEvol);
				if (d4seek(evolDesc, searchString) != 0)
					{
					XtFree(quantityPtr);
					XtFree(clientPtr);
					XtFree(clientIdPtr);
					XtFree(dicCodePtr);
					XtFree(datePtr);
					XtFree(timePtr);
					editInProgress = FALSE;
					forcedUpdate = TRUE;
					break;
					}
				}

			sprintf(evolNumber, "%-*d", EVOL_NUMBER_ENTRY_SIZE, selectedEvol);
			f4assign(evol_EVOLNUMBER, evolNumber);
			f4assign(evol_PRODUCT, productBuf);
			f4assign(evol_QUANTITY, quantityPtr);
			f4assign(evol_UNITS, quantityUnitsPtr);
			f4assign(evol_MOVETYPE, movementTypePtr);
			f4assign(evol_AUTOSWNG, autoSwingPtr);
			f4assign(evol_CLIENT, clientPtr);
			f4assign(evol_CLIENTID, clientIdPtr);
			f4assign(evol_DICCODE, dicCodePtr);
			f4assign(evol_DATE, datePtr);
			f4assign(evol_TIME, timePtr);
			timestamp[0] = (datePtr[6] == '9') ? '1' : '2';
			timestamp[1] = (datePtr[6] == '9') ? '9' : '0';
			timestamp[2] = datePtr[6];
			timestamp[3] = datePtr[7];
			timestamp[4] = datePtr[0];
			timestamp[5] = datePtr[1];
			timestamp[6] = datePtr[3];
			timestamp[7] = datePtr[4];
			timestamp[8] = timePtr[0];
			timestamp[9] = timePtr[1];
			timestamp[10] = timePtr[2];
			timestamp[11] = timePtr[3];
			timestamp[12] = 0;
			f4assign(evol_TIMESTAMP, timestamp);
			f4assign(evol_STATUS, PENDING_STRING);
			f4assign(evol_EVOLMODE, "NOMODE   ");

			if (insertFlag) {
			    d4append (evolDesc);
			    d4unlock_all (evolDesc);
			    d4go (evolDesc, d4recno (evolDesc));
			}

			updateDBASEfile(evolDesc);

			XtUnmanageChild(editEvolDialog);
			XtSetOneValue(evolList, XmNsensitive, True);
			XtSetOneValue(segList, XmNsensitive, True);
			editInProgress = FALSE;
			forcedUpdate = TRUE;
			}
		XtFree(quantityPtr);
		XtFree(clientPtr);
		XtFree(clientIdPtr);
		XtFree(dicCodePtr);
		XtFree(datePtr);
		XtFree(timePtr);
		break;
	  }
	}

/*****************************************************************************
 *      SegDialogCallback                                                    *
 *                                                                           *
 *      Inputs                                                               *
 *          widget          widget descriptor                                *
 *          widgetID        widget ID                                        *
 *          widgetData      widget data                                      *
 *                                                                           *
 *      Return Value                                                         *
 *          none                                                             *
 *                                                                           *
 *      Description                                                          *
 *          callback for segment dialog button presses                       *
 *****************************************************************************/

void
SegDialogCallback(widget, widgetID, widgetData)
Widget 			widget;
int			*widgetID;
XmAnyCallbackStruct	*widgetData;
	{
	char	*quantityPtr;
	char	*quantityUnitsPtr;
	char	*pausePtr;
	char	*apiPtr;
	int	d4status;
	int	sourceCount;
	int	destinationCount;
	int	lineupCount;
	char	lineupBuf[LINEUP_NAME_ENTRY_SIZE+1];
	int	receive	= !strcmp(tmpEvol.movementTypeEntry, PIER_RECEIVE_STRING) ||
			!strcmp(tmpEvol.movementTypeEntry, TRUCK_RECEIVE_STRING);

	switch(*widgetID)
	  {
	  case SEG_BBLS_BUTTON:
	  case SEG_GALS_BUTTON:
		XtSetOneValue(segBblsButton, XmNset, *widgetID == SEG_BBLS_BUTTON);
		XtSetOneValue(segGalsButton, XmNset, *widgetID == SEG_GALS_BUTTON);
		break;

	  case CANCEL_EDIT_SEG_BUTTON:
		XtUnmanageChild(editSegDialog);
		XtSetOneValue(evolList, XmNsensitive, True);
		XtSetOneValue(segList, XmNsensitive, True);
		editInProgress = FALSE;
		forcedUpdate = TRUE;
		break;

	  case OK_EDIT_SEG_BUTTON:
		XtGetOneValue(sourceList, XmNselectedItemCount, &sourceCount);
	  	if (sourceCount == 0)
			{
			ErrorMessage("Must select from \"Source\" list");
			break;
			}
		XtGetOneValue(destinationList, XmNselectedItemCount, &destinationCount);
	  	if (destinationCount == 0)
			{
			ErrorMessage("Must select from \"Destination\" list");
			break;
			}
		XtGetOneValue(lineupList, XmNselectedItemCount, &lineupCount);
	  	if (lineupCount == 1)
			{
			XmStringTable	item;
			char			*p;

			XtGetOneValue(lineupList, XmNselectedItems, &item);
			XmStringGetLtoR(item[0], XmSTRING_DEFAULT_CHARSET, &p);
			dupString(lineupBuf, p, (sizeof lineupBuf)-1);
			XtFree(p);
			}
		else
			{
			ErrorMessage("Must select from \"Lineup\" list");
			break;
			}
		quantityPtr = StripWhitespace(XmTextGetString(segQuantityEntry));
		quantityUnitsPtr = XmToggleButtonGetState(segBblsButton) ? BBLS_STRING : GALS_STRING;
		pausePtr = XmToggleButtonGetState(pauseOnCompButton) ? YES_STRING : NO_STRING;
		apiPtr = StripWhitespace(XmTextGetString(apiEntry));
		if (!*quantityPtr)
			ErrorMessage("Must fill in \"Quantity\" field");
		else if (receive && !*apiPtr)
			ErrorMessage("Must fill in \"API\" field");
		if (!(quantityPtr = CheckUfloat(quantityPtr)))
			ErrorMessage("Invalid entry in \"Quantity\" field (a valid entry example is: 19500)");
		else if (atof(quantityPtr) > (remainingGallons() / (XmToggleButtonGetState(segBblsButton) ? 42.0 : 1.0)))
			ErrorMessage("Value too large in \"Quantity\" field");
		else if (receive && !(apiPtr = CheckUfloat(apiPtr)))
			ErrorMessage("Invalid entry in \"API\" field");
		else
			{
			char		evolNumber[EVOL_NUMBER_ENTRY_SIZE+1];
			char		segNumber[SEG_NUMBER_ENTRY_SIZE+1];
			char		orderNumber[ORDER_NUMBER_ENTRY_SIZE+1];
			char		searchString[BUFSIZ];
			int		ordNumber = 0;
			XmStringTable	item;
			int		i;

			if (insertFlag)
				{
				int	maxSegment = 0;

				/* get the largest segNumber in the evolution and the orderNumber for this segment */
				d4tag_select(evolSegsDesc, d4tag(evolSegsDesc, "EVOLSEGTAG"));
				memset ( searchString, 0, sizeof searchString );
				sprintf(searchString, "%*d", EVOL_NUMBER_ENTRY_SIZE, selectedEvol);
				if (d4seek(evolSegsDesc, searchString) == 0)
					do
						{
						SegRecord	seg;
						
						getEvolSegment(&seg);
						if (seg.evolNumber != selectedEvol)
							break;
						maxSegment = seg.segNumber;
						if (seg.segNumber == selectedSeg)
							selectedOrd = ordNumber = seg.orderNumber;
						d4unlock_all (evolSegsDesc);
						}
					while (d4skip(evolSegsDesc, 1) == 0);
					d4unlock_all (evolSegsDesc);	
				selectedSeg = maxSegment + 1;
				d4append_start(evolSegsDesc);
				}
			else
				{					
				sprintf(searchString, "%*d%*d", EVOL_NUMBER_ENTRY_SIZE, selectedEvol,
						 SEG_NUMBER_ENTRY_SIZE, selectedSeg);
				d4tag_select(evolSegsDesc, d4tag(evolSegsDesc, "EVOLSEGTAG"));
				if (d4seek(evolSegsDesc, searchString) != 0)
					{
					XtFree(quantityPtr);
					XtFree(apiPtr);
					editInProgress = FALSE;
					forcedUpdate = TRUE;
					break;
					}
				ordNumber = f4int (evolSegs_ORDERNUM);
				}

			sprintf(evolNumber, "%-*d", EVOL_NUMBER_ENTRY_SIZE, selectedEvol);
			sprintf(segNumber, "%-*d", SEG_NUMBER_ENTRY_SIZE, selectedSeg);
			sprintf(orderNumber, "%-*d", ORDER_NUMBER_ENTRY_SIZE, ordNumber);
			f4assign (evolSegs_EVOLNUMBER, evolNumber);
			f4assign (evolSegs_SEGNUMBER, segNumber);
			f4assign (evolSegs_ORDERNUM, orderNumber);
			f4assign (evolSegs_QUANTITY, quantityPtr);
			f4assign (evolSegs_UNITS, quantityUnitsPtr);
			f4assign (evolSegs_LINEUPNAME, lineupBuf);
			f4assign (evolSegs_PAUSE, pausePtr);
			f4assign (evolSegs_STATUS, PENDING_STRING);
			f4assign_double (evolSegs_VOLXFERED, 0.0);
			f4assign_double (evolSegs_ACTFLWRT, 0.0);
			f4assign_long (evolSegs_SCRIPTID, 0);
			f4assign (evolSegs_REQTYPE, "NONE  ");
			f4assign (evolSegs_AVGAPI, apiPtr);

			if (insertFlag) {
			    d4append (evolSegsDesc);
			    d4go (evolSegsDesc, d4recno (evolSegsDesc));
			}

			updateDBASEfile(evolSegsDesc);

			reorderSegments();

			d4tag_select(segSrcsDesc, d4tag(segSrcsDesc, "SEGSRCSTAG"));
			if (d4top(segSrcsDesc) == 0)	/* first delete all of them for the evol,seg */
				do
					{
					long	evol = f4long (segSrcs_EVOLNUMBER);
					long	seg = f4long (segSrcs_SEGNUMBER);
			
					if (evol == selectedEvol && seg == selectedSeg)  {
						d4delete(segSrcsDesc);
					}
					d4unlock_all (segSrcsDesc);
					}
				while (d4skip(segSrcsDesc, 1) == 0);
				d4unlock_all (segSrcsDesc);
	
			XtGetOneValue(sourceList, XmNselectedItems, &item);
			for ( i = 0; i < sourceCount; i++ )
				{
				char	*p;

				XmStringGetLtoR(item[i], XmSTRING_DEFAULT_CHARSET, &p);
				d4append_start(segSrcsDesc);
				f4assign (segSrcs_EVOLNUMBER, evolNumber);
				f4assign (segSrcs_SEGNUMBER, segNumber);
				f4assign (segSrcs_SOURCE, p);
				f4assign (segSrcs_API, apiPtr);
				d4append (segSrcsDesc);
				d4go (segSrcsDesc, d4recno (segSrcsDesc));
				XtFree(p);
				d4unlock_all (segSrcsDesc);
				}
			updateDBASEfile(segSrcsDesc);
			
			d4tag_select(segDestsDesc, d4tag(segDestsDesc, "SEGDESTSTAG"));
			if (d4top(segDestsDesc) == 0)	/* first delete all of them for the evol,seg */
				do
					{
					long	evol = f4long (segDests_EVOLNUMBER);
					long	seg = f4long (segDests_SEGNUMBER);
			
					if (evol == selectedEvol && seg == selectedSeg)  {	
						d4delete(segDestsDesc);
					}
					d4unlock_all (segDestsDesc);
					}
				while (d4skip(segDestsDesc, 1) == 0);
				d4unlock_all (segDestsDesc);

			XtGetOneValue(destinationList, XmNselectedItems, &item);
			for ( i = 0; i < destinationCount; i++ )
				{
				char	*p;

				XmStringGetLtoR(item[i], XmSTRING_DEFAULT_CHARSET, &p);
				d4append_start(segDestsDesc);
				f4assign (segDests_EVOLNUMBER, evolNumber);
				f4assign (segDests_SEGNUMBER, segNumber);
				f4assign (segDests_DEST, p);
				d4append (segDestsDesc);
				d4unlock_all (segDestsDesc);
				d4go (segDestsDesc, d4recno (segDestsDesc));
				XtFree(p);
				}
			updateDBASEfile(segDestsDesc);

			XtUnmanageChild(editSegDialog);
			XtSetOneValue(evolList, XmNsensitive, True);
			XtSetOneValue(segList, XmNsensitive, True);
			editInProgress = FALSE;
			forcedUpdate = TRUE;
			}
		XtFree(quantityPtr);
		XtFree(apiPtr);
		break;
	  }
	}

/*****************************************************************************
 *      QuantityDialogCallback                                               *
 *                                                                           *
 *      Inputs                                                               *
 *          widget          widget descriptor                                *
 *          widgetID        widget ID                                        *
 *          widgetData      widget data                                      *
 *                                                                           *
 *      Return Value                                                         *
 *          none                                                             *
 *                                                                           *
 *      Description                                                          *
 *          callback for quantity dialog button presses                      *
 *****************************************************************************/

void
QuantityDialogCallback(widget, widgetID, widgetData)
Widget 			widget;
int			*widgetID;
XmAnyCallbackStruct	*widgetData;
	{
	char	*quantityPtr;

	switch(*widgetID)
	  {
	  case CANCEL_QUANTITY_BUTTON:
		XtUnmanageChild(quantityDialog);
		XtSetOneValue(evolList, XmNsensitive, True);
		XtSetOneValue(segList, XmNsensitive, True);
		editInProgress = FALSE;
		forcedUpdate = TRUE;
		break;

	  case OK_QUANTITY_BUTTON:
		quantityPtr = StripWhitespace(XmTextGetString(quantityEntry));
		if (*quantityPtr)
			{
			if (!(quantityPtr = CheckUfloat(quantityPtr)))
				ErrorMessage("Invalid entry in \"Quantity\" field (a valid entry example is: 19500)");
			else
				{
				char	searchString[BUFSIZ];
				double	old, delta;
				
				sprintf(searchString, "%*d%*d", EVOL_NUMBER_ENTRY_SIZE, selectedEvol, SEG_NUMBER_ENTRY_SIZE, selectedSeg);
				d4tag_select(evolSegsDesc, d4tag(evolSegsDesc, "EVOLSEGTAG"));
				if (d4seek(evolSegsDesc, searchString) != 0)
					{
					XtFree(quantityPtr);
					editInProgress = FALSE;
					forcedUpdate = TRUE;
					break;
					}
				old = f4double( evolSegs_QUANTITY);	/* get current value */
				f4assign (evolSegs_QUANTITY, quantityPtr);
				updateDBASEfile(evolSegsDesc);
				
				delta = f4double (evolSegs_QUANTITY) - old;  /* get how much he changed it by */

				sprintf(searchString, "%*d", EVOL_NUMBER_ENTRY_SIZE, selectedEvol);
				d4tag_select(evolDesc, d4tag(evolDesc, "EVOLTAG"));
				if (d4seek(evolDesc, searchString) != 0)
					{
					XtFree(quantityPtr);
					editInProgress = FALSE;
					forcedUpdate = TRUE;
					break;
					}
				old = f4double (evol_QUANTITY);
				f4assign_double (evol_QUANTITY, old + delta);
				updateDBASEfile(evolDesc);
				}
			}
		XtUnmanageChild(quantityDialog);
		XtSetOneValue(evolList, XmNsensitive, True);
		XtSetOneValue(segList, XmNsensitive, True);
		editInProgress = FALSE;
		forcedUpdate = TRUE;
		XtFree(quantityPtr);
		break;
	  }
	}

/*****************************************************************************
 *      EvolListCallback                                                     *
 *                                                                           *
 *      Inputs                                                               *
 *          widget          widget descriptor                                *
 *          widgetID        widget ID                                        *
 *          widgetData      widget data                                      *
 *                                                                           *
 *      Return Value                                                         *
 *          none                                                             *
 *                                                                           *
 *      Description                                                          *
 *          callback for the evolution list                                  *
 *****************************************************************************/

void
EvolListCallback(widget, widgetID, widgetData)
Widget 					widget;
int						*widgetID;
XmListCallbackStruct	*widgetData;
	{
	char	*p;
	int		newSelectedEvol;

	XmStringGetLtoR(widgetData->item, XmSTRING_DEFAULT_CHARSET, &p);
	if (sscanf(p, "%d", &newSelectedEvol) != 1)
		selectedEvol = 0;
	else
		selectedEvol = (newSelectedEvol == selectedEvol) ? 0: newSelectedEvol;

	XtFree(p);

	forcedUpdate = TRUE;
	selectedSeg = 0;
	}

/*****************************************************************************
 *      SegListCallback                                                      *
 *                                                                           *
 *      Inputs                                                               *
 *          widget          widget descriptor                                *
 *          widgetID        widget ID                                        *
 *          widgetData      widget data                                      *
 *                                                                           *
 *      Return Value                                                         *
 *          none                                                             *
 *                                                                           *
 *      Description                                                          *
 *          callback for the segment list                                    *
 *****************************************************************************/
typedef struct
	{
	NODE		node;
	int			idNumber;
	} SegList;

LIST	segIdList;

void
SegListCallback(widget, widgetID, widgetData)
Widget 			widget;
int			*widgetID;
XmListCallbackStruct	*widgetData;
	{
	int		*positionList;
	int		positionCount;

	if (XmListGetSelectedPos(segList, &positionList, &positionCount))
		{
		selectedSeg = ((SegList *)lstNth(&segIdList, positionList[0]))->idNumber;
		XtFree((char *)positionList);
		}
	else
		selectedSeg = 0;

	forcedUpdate = TRUE;
	}

/*****************************************************************************
 *      SetButtonTitles                                                      *
 *                                                                           *
 *      Inputs                                                               *
 *          none                                                             *
 *                                                                           *
 *      Return Value                                                         *
 *          none                                                             *
 *                                                                           *
 *      Description                                                          *
 *          set button titles and sensitivity                                *
 *****************************************************************************/
char	pendingString[STATUS_ENTRY_SIZE+1];
char	stoppedString[STATUS_ENTRY_SIZE+1];
char	estoppedString[STATUS_ENTRY_SIZE+1];
char	errorString[STATUS_ENTRY_SIZE+1];
char	completedString[STATUS_ENTRY_SIZE+1];
char	mcompletedString[STATUS_ENTRY_SIZE+1];

isDeletable(status)
char	*status;
	{
	return (strcmp(status, pendingString) == 0) ? True : False;
	}

isEditable(status)
char	*status;
	{
/*	return	((strcmp(status, pendingString)  == 0) ||
			 (strcmp(status, stoppedString)  == 0) ||
			 (strcmp(status, estoppedString) == 0) ||
			 (strcmp(status, errorString)    == 0)) ? True : False;*/
	return	((strcmp(status, completedString)  != 0) &&
			 (strcmp(status, mcompletedString) != 0)) ? True : False;
	}

isChangeable(status)
char	*status;
	{
	return	((strcmp(status, completedString)  != 0) &&
			 (strcmp(status, mcompletedString) != 0)) ? True : False;
	}

void
SetButtonTitles()
	{
	int		firstTime	= TRUE;
	
	if (firstTime)
		{
		firstTime = FALSE;
		dupString(pendingString,	EvolutionStateText[Pending],	sizeof pendingString);
		dupString(stoppedString,	EvolutionStateText[Stopped],	sizeof stoppedString);
		dupString(estoppedString,	EvolutionStateText[EStopped],	sizeof estoppedString);
		dupString(errorString,		EvolutionStateText[Error],		sizeof errorString);
		dupString(completedString,	EvolutionStateText[Completed],	sizeof completedString);
		dupString(mcompletedString,	EvolutionStateText[MCompleted],	sizeof mcompletedString);
		}
	if (!selectedEvol)
		{
		XtSetOneValue(insertButton,	XmNlabelString, insertEvolButtonString);
		XtSetOneValue(editButton,	XmNlabelString, editEvolButtonString);
		XtSetOneValue(deleteButton,	XmNlabelString, deleteEvolButtonString);
		}
	else
		{
		XtSetOneValue(insertButton, XmNlabelString, insertSegButtonString);
		XtSetOneValue(editButton,	XmNlabelString, selectedSeg ? editSegButtonString   : editEvolButtonString);
		XtSetOneValue(deleteButton, XmNlabelString, selectedSeg ? deleteSegButtonString : deleteEvolButtonString);
		}
	XtSetOneValue(displayButton,	XmNsensitive, OperPriv(&codeBaseDesc, "EVOL_MGT_DISP") && !editInProgress && selectedSeg);
	XtSetOneValue(printButton,	XmNsensitive, OperPriv(&codeBaseDesc, "EVOL_MGT_PRNT") && !editInProgress);
	XtSetOneValue(exitButton,	XmNsensitive, !editInProgress);
	XtSetOneValue(startButton,	XmNsensitive, OperPriv(&codeBaseDesc, "EVOL_MGT_STRT") && !editInProgress && selectedEvol);
	XtSetOneValue(stopButton,	XmNsensitive, OperPriv(&codeBaseDesc, "EVOL_MGT_STOP") && !editInProgress && selectedEvol);
	XtSetOneValue(pauseButton,	XmNsensitive, OperPriv(&codeBaseDesc, "EVOL_MGT_PAUS") && !editInProgress && selectedEvol);
	XtSetOneValue(resumeButton,	XmNsensitive, OperPriv(&codeBaseDesc, "EVOL_MGT_RSUM") && !editInProgress && selectedEvol);
	XtSetOneValue(manualButton,	XmNsensitive, OperPriv(&codeBaseDesc, "EVOL_MGT_MANE") && !editInProgress && selectedEvol);
	XtSetOneValue(quantityButton,	XmNsensitive, OperPriv(&codeBaseDesc, "EVOL_MGT_SQNT") && !editInProgress && selectedSeg && isChangeable(tmpSeg.statusEntry));
	XtSetOneValue(insertButton,	XmNsensitive, OperPriv(&codeBaseDesc, "EVOL_MGT_INST") && !editInProgress && (!selectedEvol || isEditable(tmpEvol.statusEntry)));
	XtSetOneValue(editButton,	XmNsensitive, OperPriv(&codeBaseDesc, "EVOL_MGT_EDIT") && !editInProgress &&
		(selectedEvol && ((selectedSeg && isEditable(tmpSeg.statusEntry)) ||
		(!selectedSeg && isEditable(tmpEvol.statusEntry)))));
	XtSetOneValue(deleteButton,	XmNsensitive, OperPriv(&codeBaseDesc, "EVOL_MGT_DELT") && !editInProgress &&
		(selectedEvol && ((selectedSeg && isDeletable(tmpSeg.statusEntry)) ||
		(!selectedSeg && isDeletable(tmpEvol.statusEntry)))));
	setOperPriv();
	}

/*****************************************************************************
 *      DeleteEvol                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *          none                                                             *
 *                                                                           *
 *      Return Value                                                         *
 *          none                                                             *
 *                                                                           *
 *      Description                                                          *
 *          delete a evolution                                               *
 *****************************************************************************/

void
DeleteEvol()
	{
	char	searchString[BUFSIZ];
	
	/*
	 * delete the evolution and all of its segments
	 */
	sprintf(searchString, "%*d", EVOL_NUMBER_ENTRY_SIZE, selectedEvol);
	d4tag_select(evolDesc, d4tag(evolDesc, "EVOLTAG"));
	if (d4seek(evolDesc, searchString) == 0)
		{
		d4delete(evolDesc);
		
		d4tag_select(evolSegsDesc, d4tag(evolSegsDesc, "EVOLSEGTAG"));
		if (d4top(evolSegsDesc) == 0) {
			do
				{
				SegRecord	seg;

				getEvolSegment(&seg);
				if (seg.evolNumber == selectedEvol) {
					d4delete(evolSegsDesc);
				}
				d4unlock_all (evolSegsDesc);
				}
			while (d4skip(evolSegsDesc, 1) == 0);
		}
		updateDBASEfile(evolSegsDesc);
		
		d4tag_select(segSrcsDesc, d4tag(segSrcsDesc, "SEGSRCSTAG"));
		if (d4top(segSrcsDesc) == 0) {
			do
				{
				long	evol = f4long (segSrcs_EVOLNUMBER);
			
				if (evol == selectedEvol)  {
					d4delete(segSrcsDesc);
				}
				d4unlock_all (segSrcsDesc);
				}
			while (d4skip(segSrcsDesc, 1) == 0);
		}
		updateDBASEfile(segSrcsDesc); 

		d4tag_select(segDestsDesc, d4tag(segDestsDesc, "SEGDESTSTAG"));
		if (d4top(segDestsDesc) == 0) {
			do
				{
				long	evol = f4long (segDests_EVOLNUMBER);
			
				if (evol == selectedEvol)  {	
					d4delete(segDestsDesc);
				}
				d4unlock_all (segDestsDesc);
				}
			while (d4skip(segDestsDesc, 1) == 0);
		}
		updateDBASEfile(segDestsDesc);

		selectedEvol = 0;
		forcedUpdate = TRUE;
		}
	updateDBASEfile(evolDesc);
	}

/*****************************************************************************
 *      DeleteSeg                                                            *
 *                                                                           *
 *      Inputs                                                               *
 *          none                                                             *
 *                                                                           *
 *      Return Value                                                         *
 *          none                                                             *
 *                                                                           *
 *      Description                                                          *
 *          delete a segment                                                 *
 *****************************************************************************/

void
DeleteSeg()
	{
	char	searchString[BUFSIZ];
			
	sprintf(searchString, "%*d%*d", EVOL_NUMBER_ENTRY_SIZE, selectedEvol, SEG_NUMBER_ENTRY_SIZE, selectedSeg);
	d4tag_select(evolSegsDesc, d4tag(evolSegsDesc, "EVOLSEGTAG"));
	if (d4seek(evolSegsDesc, searchString) == 0) {
		d4delete(evolSegsDesc);
	}

	updateDBASEfile(evolSegsDesc);
	
	reorderSegments();

	selectedSeg = 0;
	forcedUpdate = TRUE;
	}

/*****************************************************************************
 *      SourceDestinationListCallback                                        *
 *                                                                           *
 *      Inputs                                                               *
 *          widget          widget descriptor                                *
 *          widgetID        widget ID                                        *
 *          widgetData      widget data                                      *
 *                                                                           *
 *      Return Value                                                         *
 *          none                                                             *
 *                                                                           *
 *      Description                                                          *
 *          callback for the source and destination selection list selections*
 *****************************************************************************/

void
SourceDestinationListCallback(widget, widgetID, widgetData)
Widget 			widget;
int			*widgetID;
XmAnyCallbackStruct	*widgetData;
	{
	setLineupNames(tmpSeg.lineupEntry);
	}

/*****************************************************************************
 *      ErrorMessage                                                         *
 *                                                                           *
 *      Inputs                                                               *
 *          msgPtr            pointer to error msg                           *
 *                                                                           *
 *      Description                                                          *
 *          popup error msg dialog box                                       *
 *****************************************************************************/

void
ErrorMessage(msgPtr)
char	*msgPtr;
	{
	if (!errorMessage)
		printf("%s\n", msgPtr);
	else if (!errorMessageString)
		{
		errorMessageString = XmStringCreate(msgPtr, XmSTRING_DEFAULT_CHARSET);
		XtSetOneValue(errorMessage, XmNmessageString, errorMessageString);
		XtManageChild(errorMessage);
		XBell(XtDisplay(evolmanWindow), 100);
		}
	}

/*****************************************************************************
 *      ErrorMessageCallback                                                 *
 *                                                                           *
 *      Inputs                                                               *
 *          widget          widget ID to register                            *
 *          index           index into widget array                          *
 *          widgetData      widget data                                      *
 *                                                                           *
 *      Description                                                          *
 *          callback when dismissing error msg dialog box                    *
 *****************************************************************************/

void
ErrorMessageCallback(widget, index, data)
Widget 			widget;
int			*index;
XmAnyCallbackStruct	*data;
	{
	XmStringFree(errorMessageString);
	errorMessageString = NULL;
	}

/*****************************************************************************
 *      WarningMessage                                                       *
 *                                                                           *
 *      Inputs                                                               *
 *          msgPtr            pointer to warning msg                         *
 *          funcPtr           function to call on ok button press            *
 *                                                                           *
 *      Description                                                          *
 *          popup warning msg dialog box                                     *
 *****************************************************************************/

void
WarningMessage(msgPtr, funcPtr)
char	*msgPtr;
void	(*funcPtr)();
	{
	if (!warningMessage)
		printf("%s\n", msgPtr);
	else if (!warningMessageString)
		{
		warningMessageFunc = funcPtr;
		warningMessageString = XmStringCreate(msgPtr, XmSTRING_DEFAULT_CHARSET);
		XtSetOneValue(warningMessage, XmNmessageString, warningMessageString);
		XtManageChild(warningMessage);
		XBell(XtDisplay(evolmanWindow), 100);
		}
	}

/*****************************************************************************
 *      WarningMessageOkCallback                                             *
 *                                                                           *
 *      Inputs                                                               *
 *          widget          widget ID to register                            *
 *          index           index into widget array                          *
 *          widgetData      widget data                                      *
 *                                                                           *
 *      Description                                                          *
 *          callback when oking warning msg dialog box                       *
 *****************************************************************************/

void
WarningMessageOkCallback(widget, index, data)
Widget 			widget;
int			*index;
XmAnyCallbackStruct	*data;
	{
	XmStringFree(warningMessageString);
	warningMessageString = NULL;
	(*warningMessageFunc)();
	}

/*****************************************************************************
 *      WarningMessageCancelCallback                                         *
 *                                                                           *
 *      Inputs                                                               *
 *          widget          widget ID to register                            *
 *          index           index into widget array                          *
 *          widgetData      widget data                                      *
 *                                                                           *
 *      Description                                                          *
 *          callback when cancelling warning msg dialog box                  *
 *****************************************************************************/

void
WarningMessageCancelCallback(widget, index, data)
Widget 			widget;
int			*index;
XmAnyCallbackStruct	*data;
	{
	XmStringFree(warningMessageString);
	warningMessageString = NULL;
	}

/*****************************************************************************
 *      StripWhitespace                                                      *
 *                                                                           *
 *      Inputs                                                               *
 *          string      pointer to character string                          *
 *                                                                           *
 *      Return Value                                                         *
 *          pointer to first non-space character in string                   *
 *                                                                           *
 *      Description                                                          *
 *          strip leading and trailing whitespace from character string      *
 *****************************************************************************/

char *
StripWhitespace(string)
char	*string;
	{
	int	i;

	for (i=strlen(string)-1; i >= 0 && isspace(string[i]); string[i--] = 0)
		;
	for (i=0; string[i] && isspace(string[i]); i++)
		;
	if (i)
		strcpy(string, &string[i]);
	return string;
	}

/******************************************************************************
 *     CheckUfloat(ufloatPtr)                                                 *
 *                                                                            *
 *     Inputs                                                                 *
 *          ufloatPtr        pointer to an unsigned float string              *
 *                                                                            *
 *     Return value                                                           *
 *          pointer to first non space character in string                    *
 *                                                                            *
 *     Description                                                            *
 *          checks the validity of an unsigned float string.                  *
 ******************************************************************************/

char *
CheckUfloat(ufloatPtr)
char	*ufloatPtr;
	{
	int		i, j;
	int		decFlag;

	for (i=strlen(ufloatPtr)-1; i >= 0 && ufloatPtr[i] == ' '; ufloatPtr[i--] = 0)
		;
	for (; *ufloatPtr && *ufloatPtr == ' '; ufloatPtr++)
		;
	for (j=0; *ufloatPtr && *ufloatPtr == '0'; j++, ufloatPtr++)
		;
	decFlag = FALSE;
	for (i=0; isdigit(*(ufloatPtr + i)) || *(ufloatPtr + i) == '.'; i++)
		{
		if (*(ufloatPtr + i) == '.')
			{
			if (decFlag)
				break;
			else
				decFlag = TRUE;
			}
		}
	if (!*(ufloatPtr + i))
		{
		if (j)
			return ufloatPtr - 1;
		else if (strlen(ufloatPtr) == 0)
			return NULL;
		else
			return ufloatPtr;
		}
	else
		return NULL;
	}

/******************************************************************************
 *     CheckDate                                                              *
 *                                                                            *
 *     Inputs                                                                 *
 *         datePtr          pointer to date string                            *
 *                                                                            *
 *     Return value                                                           *
 *         TRUE or FALSE                                                      *
 *                                                                            *
 *     Description                                                            *
 *         check date syntax MM/DD/YY                                         *
 ******************************************************************************/

int
CheckDate(datePtr)
char	*datePtr;
	{   
	char		dateBuf[DATE_STRING_ENTRY_SIZE+1];
	int		month;
	int		day;
	static int	monthDays [] =
		{
		00, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
		};

	if (strlen(datePtr) != DATE_STRING_ENTRY_SIZE)
		return FALSE;
	strcpy(dateBuf, datePtr);
	if (dateBuf[2] != '/' || dateBuf[5] != '/')
		return FALSE;
	dateBuf[2] = 0;
	dateBuf[5] = 0;
	month = atoi(&dateBuf[0]);
	day   = atoi(&dateBuf[3]);
	return((1 <= month && month <= 12) &&(1 <= day && day <= monthDays[month]));
	}

/******************************************************************************
 *     CheckTime                                                              *
 *                                                                            *
 *     Inputs                                                                 *
 *         timePtr          pointer to time string                            *
 *                                                                            *
 *     Return value                                                           *
 *         TRUE or FALSE                                                      *
 *                                                                            *
 *     Description                                                            *
 *         check time syntax HHMM                                             *
 ******************************************************************************/

int
CheckTime(timePtr)
char	*timePtr;
	{ 
	int		hour;
	int		minute;

	if (!isdigit(timePtr[0]) || !isdigit(timePtr[1]) || !isdigit(timePtr[2]) || !isdigit(timePtr[3]))
		return FALSE;

	hour   = ((timePtr[0] - '0') * 10) + (timePtr[1] - '0');
	minute = ((timePtr[2] - '0') * 10) + (timePtr[3] - '0');
	return (hour < 24 && minute < 60);
	}

/******************************************************************************
 *     setProductNames                                                        *
 *                                                                            *
 *     Description                                                            *
 *         generate the list of possible products                             *
 ******************************************************************************/

setProductNames(selectedProduct)
char	*selectedProduct;
	{
	D4DATA		*tankDesc;
	F4FIELD		*productField;

	red_RedundancyOff ();
	if (!(tankDesc = d4open(&codeBaseDesc, tankName)))
		printf("Cannot open tank database file \"%s\"\n", tankName), exit(0);
	red_RedundancyOn ();

	productField = d4field(tankDesc, "PRODUCT");

	XmListDeleteAllItems(productList);
	d4tag_select(tankDesc, d4tag(tankDesc, "NAME"));
	if (d4top(tankDesc) == 0)
		do
			{
			int	*positionList;
			int	positionCount;
			XmString s = XmStringCreate(StripWhitespace(f4str(productField)),
					XmSTRING_DEFAULT_CHARSET);

			if (!XmListGetMatchPos(productList, s, &positionList, &positionCount))
				XmListAddItem(productList, s, 0);
			else
				XtFree((char *)positionList);
			XmStringFree(s);
			d4unlock_all (tankDesc);
			}
		while (d4skip(tankDesc, 1) == 0);

	if (*selectedProduct)
		{
		XmString selectedString = XmStringCreate(StripWhitespace(selectedProduct), 
					XmSTRING_DEFAULT_CHARSET);
		
		XmListSelectItem(productList, selectedString, False);
		XmListSetItem(productList, selectedString);
		XmStringFree(selectedString);
		}
	else
		{
		XmListSelectPos(productList, 1, False);
		XmListSetPos(productList, 1);
		}
	d4close(tankDesc);
	}

/******************************************************************************
 *     srcDestMatch                                                           *
 *                                                                            *
 *     Description                                                            *
 *         determine whether the current list of sources or destinations      *
 *         matches those in the lineup sources or destinations database       *
 ******************************************************************************/
typedef struct
	{
	NODE	node;
	char	*text;
	} StringList;

LIST	haveList;

#define foreachHave(x)		for(x = (StringList *)lstFirst(&haveList); x; x = (StringList *)lstNext(x))
	
srcDestMatch(lineupName, type)
char	*lineupName;
char	type;
	{
	D4DATA		*desc;
	F4FIELD		*desc_LINEUPNAME, *desc_SRCS, *desc_DESTS;
	char		searchString[LINEUP_NAME_ENTRY_SIZE+1];
	int		*positionList;
	int		positionCount;
	char		*filename;
	StringList	*p;
	Widget		listWidget = (type == 's') ?  sourceList : destinationList;
	int		status = FALSE;

	red_RedundancyOff ();
	if (!(desc = d4open(&codeBaseDesc, filename = (type == 's') ? lineupSrcsName : lineupDestsName)))
	    printf("Cannot open lineup database file \"%s\"\n", filename), exit(0);
	red_RedundancyOn ();

	desc_LINEUPNAME = d4field (desc, "LINEUPNAME");
	if (type == 's')
	    desc_SRCS = d4field (desc, "SRCS");
	else
	    desc_DESTS = d4field (desc, "DESTS");

	lstInit(&haveList);
	d4tag_select(desc, d4tag(desc, "LINEUPTAG"));
	sprintf(searchString, "%-*s", LINEUP_NAME_ENTRY_SIZE, lineupName);
	if (d4seek(desc, searchString) == 0)
		while (strcmp(StripWhitespace(f4str(desc_LINEUPNAME)), lineupName) == 0)
			{
			StringList *p = (StringList *)calloc(1, sizeof (StringList));
			char *name = StripWhitespace (f4str (type == 's' ?  desc_SRCS : desc_DESTS));

			p->text = calloc(strlen(name)+1, sizeof (char));
			strcpy(p->text, name);
			lstAdd(&haveList, &p->node);
			if (d4skip(desc, 1) != 0)
				break;
			d4unlock_all (desc);
			}
	d4close(desc);

	if (XmListGetSelectedPos(listWidget, &positionList, &positionCount))
		{
		if (positionCount == lstCount(&haveList))
			{
			int i = 0;
			
			foreachHave(p)
				{
				XmStringTable	item;
				char		*s;

				XtGetOneValue(listWidget, XmNselectedItems, &item);
				XmStringGetLtoR(item[i], XmSTRING_DEFAULT_CHARSET, &s);
				if (strcmp(s, p->text) != 0)
					{
					XtFree(s);
					break;
					}
				XtFree(s);
				i++;
				}
			status = (i == positionCount);
			}
		XtFree((char *)positionList);
		}

	foreachHave(p)
		free(p->text);
	lstFree(&haveList);

	return status;
	}
	
/******************************************************************************
 *     setSourceNames                                                         *
 *                                                                            *
 *     Description                                                            *
 *         generate the list of possible sources                              *
 ******************************************************************************/

setSourceNames()
	{
	D4DATA		*desc;
	F4FIELD		*nameField;
	F4FIELD		*productField;
	char		*desiredProduct;
	char		*databaseFilename;
	char		*databaseTag;
	int		receive;
	int		issue;
	XmString	s;

	receive = strcmp(tmpEvol.movementTypeEntry, PIER_RECEIVE_STRING) == 0 ||
			  strcmp(tmpEvol.movementTypeEntry, TRUCK_RECEIVE_STRING) == 0;
	databaseFilename = receive ? potName : tankName;
	desiredProduct = receive ? NULL : tmpEvol.productEntry;

	/*
	 * open the selected database and add all of the matching (if needed) names
	 */

	red_RedundancyOff ();
	if (!(desc = d4open(&codeBaseDesc, databaseFilename)))
		printf("Cannot open database file \"%s\"\n", databaseFilename), exit(0);
	red_RedundancyOn ();

	XmListDeleteAllItems(sourceList);
	nameField = d4field (desc, "NAME");
	productField = d4field (desc, "PRODUCT");
	d4tag_select(desc, d4tag (desc, "NAME"));
	if (d4top(desc) == 0)
		do
			{
			if (desiredProduct)
				if (strcmp(desiredProduct, StripWhitespace(f4str(productField))) != 0)
					continue;

			s = XmStringCreate(StripWhitespace(f4str(nameField)), XmSTRING_DEFAULT_CHARSET);
			XmListAddItem(sourceList, s, 0);
			XmStringFree(s);
			d4unlock_all (desc);
			}
		while (d4skip(desc, 1) == 0);

	d4close(desc);

	d4tag_select(segSrcsDesc, d4tag(segSrcsDesc, "SEGSRCSTAG"));
	if (d4top(segSrcsDesc) == 0)
		do
			{
			long	evolNumber = f4long (segSrcs_EVOLNUMBER);
			long	segNumber = f4long (segSrcs_SEGNUMBER);
			
			if (evolNumber == selectedEvol && segNumber == selectedSeg)
				{
				s = XmStringCreate(StripWhitespace(f4str(segSrcs_SOURCE)),
						XmSTRING_DEFAULT_CHARSET);
				XmListSelectItem(sourceList, s, False);
				XmListSetItem(sourceList, s);
				XmStringFree(s);
				}
			d4unlock_all (segSrcsDesc);
			}
		while (d4skip(segSrcsDesc, 1) == 0);

	d4unlock_all(segSrcsDesc);
	}

/******************************************************************************
 *     setDestinationNames                                                    *
 *                                                                            *
 *     Description                                                            *
 *         generate the list of possible destinations                         *
 ******************************************************************************/

setDestinationNames()
	{
	D4DATA		*desc;
	F4FIELD		*nameField;
	F4FIELD		*productField;
	char		*desiredProduct;
	char		*databaseFilename;
	char		*databaseTag;
	int		receive;
	int		issue;
	XmString	s;

	issue = strcmp(tmpEvol.movementTypeEntry, PIER_ISSUE_STRING) == 0 ||
		    strcmp(tmpEvol.movementTypeEntry, TRUCK_ISSUE_STRING) == 0;
	databaseFilename = issue ? potName : tankName;
	desiredProduct = issue ? NULL : tmpEvol.productEntry;

	/*
	 * open the selected database and add all of the matching (if needed) names
	 */
	red_RedundancyOff ();
	if (!(desc = d4open(&codeBaseDesc, databaseFilename)))
		printf("Cannot open database file \"%s\"\n", databaseFilename), exit(0);
	red_RedundancyOn ();

	XmListDeleteAllItems(destinationList);
	nameField = d4field(desc, "NAME");
	productField = d4field(desc, "PRODUCT");
	d4tag_select(desc, d4tag(desc, "NAME"));
	if (d4top(desc) == 0)
		do
			{
			if (desiredProduct)
				if (strcmp(desiredProduct, StripWhitespace(f4str(productField))) != 0)
					continue;

			s = XmStringCreate(StripWhitespace(f4str(nameField)), XmSTRING_DEFAULT_CHARSET);
			XmListAddItem(destinationList, s, 0);
			XmStringFree(s);
			d4unlock_all (desc);
			}
		while (d4skip(desc, 1) == 0);
		
	d4close(desc);

	d4tag_select(segDestsDesc, d4tag(segDestsDesc, "SEGDESTSTAG"));
	if (d4top(segDestsDesc) == 0)
		do
			{
			long	evolNumber	= f4long (segDests_EVOLNUMBER);
			long	segNumber	= f4long (segDests_SEGNUMBER);
			
			if (evolNumber == selectedEvol && segNumber == selectedSeg)
				{
				s = XmStringCreate(StripWhitespace(f4str(segDests_DEST)),
						 XmSTRING_DEFAULT_CHARSET);
				XmListSelectItem(destinationList, s, False);
				XmListSetItem(destinationList, s);
				XmStringFree(s);
				}
			d4unlock_all (segDestsDesc);
			}
		while (d4skip(segDestsDesc, 1) == 0);
		
	d4unlock_all(segDestsDesc);
	}

/******************************************************************************
 *     setLineupNames                                                         *
 *                                                                            *
 *     Description                                                            *
 *         generate the list of possible lineup                               *
 ******************************************************************************/

setLineupNames(selectedLineup)
char	*selectedLineup;
	{

	XmListDeleteAllItems(lineupList);
	d4tag_select(lineupDesc, d4tag(lineupDesc, "LINEUPTAG"));
	if (d4top(lineupDesc) == 0)
		do
			{
			char	lineupName[LINEUP_NAME_ENTRY_SIZE+1];

			dupString(lineupName, StripWhitespace(f4str(lineup_LINEUPNAME)), sizeof lineupName);
			if (srcDestMatch(lineupName, 's') && srcDestMatch(lineupName, 'd'))
				{
				XmString s = XmStringCreate(lineupName, XmSTRING_DEFAULT_CHARSET);
				
				XmListAddItem(lineupList, s, 0);
				XmStringFree(s);
				}
			d4unlock_all (lineupDesc);
			}
		while (d4skip(lineupDesc, 1) == 0);
	d4unlock_all(lineupDesc);

	if (*selectedLineup)
		{
		XmString selectedString = XmStringCreate(StripWhitespace(selectedLineup), XmSTRING_DEFAULT_CHARSET);
		
		XmListSelectItem(lineupList, selectedString, False);
		XmListSetItem(lineupList, selectedString);
		XmStringFree(selectedString);
		}
	else
		{
		XmListSelectPos(lineupList, 1, False);
		XmListSetPos(lineupList, 1);
		}
	}

double
remainingGallons()	/* in gallons */
	{
	double	remainingQty = atof(tmpEvol.quantityEntry) * ((strcmp(tmpEvol.quantityUnitsEntry, BBLS_STRING) == 0)
					? 42.0 : 1.0);
	
	d4tag_select(evolSegsDesc, d4tag(evolSegsDesc, "EVOLSEGTAG"));
	if (d4top(evolSegsDesc) == 0)
		do
			{
			SegRecord	seg;

			getEvolSegment(&seg);
			if (seg.evolNumber == selectedEvol && (insertFlag || seg.segNumber != selectedSeg))
				remainingQty -= atof(seg.quantityEntry) * ((strcmp(seg.quantityUnitsEntry, BBLS_STRING) == 0)
							? 42.0 : 1.0);
			d4unlock_all (evolSegsDesc);
			}
		while (d4skip(evolSegsDesc, 1) == 0);
	d4unlock_all(evolSegsDesc);
	return remainingQty;
	}

#define	SRCBUFLEN		15
#define	DESTBUFLEN		15

void
formatLabels(widgetID, txtBuf)
int		widgetID;
char	*txtBuf;
	{
	switch(widgetID)
	  {
	  case EVOL_COLUMNS1:
		sprintf(txtBuf, "%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s",
			EVOL_NUMBER_ENTRY_SIZE+2,	"Evol",
			PRODUCT_ENTRY_SIZE+2,		"Product",
			QUANTITY_ENTRY_SIZE+2,		"Quantity",
			UNITS_ENTRY_SIZE+2,		"BBLS/",
			WEIGHTAVG_ENTRY_SIZE+2,		"Weighted",
			MOVE_TYPE_ENTRY_SIZE+2,		"Movement",
			AUTO_SWING_ENTRY_SIZE+5,	"Auto",
			CLIENT_ENTRY_SIZE+2,		"Client",
			CLIENT_ID_ENTRY_SIZE+2,		"Client ID Code",
			DIC_CODE_ENTRY_SIZE+2,		"DIC",
			DATE_STRING_ENTRY_SIZE+2,	"Date",
			TIME_STRING_ENTRY_SIZE+2,	"Time",
			STATUS_ENTRY_SIZE+2,		"Status");
		break;

	  case EVOL_COLUMNS2:
		sprintf(txtBuf, "%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s",
			EVOL_NUMBER_ENTRY_SIZE+2,	"#",
			PRODUCT_ENTRY_SIZE+2,		"",
			QUANTITY_ENTRY_SIZE+2,		"",
			UNITS_ENTRY_SIZE+2,		"GALS",
			WEIGHTAVG_ENTRY_SIZE+2,		"Avg API",
			MOVE_TYPE_ENTRY_SIZE+2,		"Type",
			AUTO_SWING_ENTRY_SIZE+5,	"Swing",
			CLIENT_ENTRY_SIZE+2,		"",
			CLIENT_ID_ENTRY_SIZE+2,		"",
			DIC_CODE_ENTRY_SIZE+2,		"",
			DATE_STRING_ENTRY_SIZE+2,	"",
			TIME_STRING_ENTRY_SIZE+2,	"",
			STATUS_ENTRY_SIZE+2,		"");
		break;

	  case SEG_COLUMNS1:
		sprintf(txtBuf, "%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s",
			ORDER_NUMBER_ENTRY_SIZE+2,	"Seq",
			SRCBUFLEN+2,			"Source",
			DESTBUFLEN+2,			"Destination",
			QUANTITY_ENTRY_SIZE+2,		"Quantity",
			UNITS_ENTRY_SIZE+2,		"BBLS/",
			METHOD_ENTRY_SIZE+2,		"Lineup",
			PAUSE_ON_COMPL_ENTRY_SIZE+9,	"Pause On",
			AVGAPI_ENTRY_SIZE+2,		"Avg API",
			EST_FLOW_RATE_ENTRY_SIZE+2,	"Estimated",
			ACT_FLOW_RATE_ENTRY_SIZE+2,	"Actual",
			EST_FLOW_TIME_ENTRY_SIZE+2,	"Estimated",
			ACT_FLOW_TIME_ENTRY_SIZE+2,	"Actual",
			STATUS_ENTRY_SIZE+2,		"",
			QUANTITY_ENTRY_SIZE+2,		"Quantity");
		break;

	  case SEG_COLUMNS2:
		sprintf(txtBuf, "%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s",
			ORDER_NUMBER_ENTRY_SIZE+2,	"#",
			SRCBUFLEN+2,			"",
			DESTBUFLEN+2,			"",
			QUANTITY_ENTRY_SIZE+2,		"",
			UNITS_ENTRY_SIZE+2,		"GALS",
			METHOD_ENTRY_SIZE+2,		"",
			PAUSE_ON_COMPL_ENTRY_SIZE+9,	"Completion",
			AVGAPI_ENTRY_SIZE+2,		"",
			EST_FLOW_RATE_ENTRY_SIZE+2,	"Flow Rate",
			ACT_FLOW_RATE_ENTRY_SIZE+2,	"Flow Rate",
			EST_FLOW_TIME_ENTRY_SIZE+2,	"Flow Time",
			ACT_FLOW_TIME_ENTRY_SIZE+2,	"Flow Time",
			STATUS_ENTRY_SIZE+2,		"Status",
			QUANTITY_ENTRY_SIZE+2,		"Transferred");
		break;
		}
	}
	
formatEvolutionString(pevol, txtBuf)
EvolRecord	*pevol;
char		*txtBuf;
	{
	char	*status	= (strncmp(pevol->statusEntry, "Swing", 5) == 0) ? "Swinging": pevol->statusEntry;

	sprintf(txtBuf, "%-*s%-*s%-*.3lf%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s%-*s",
		EVOL_NUMBER_ENTRY_SIZE+2,	pevol->evolNumberEntry,
		PRODUCT_ENTRY_SIZE+2,		pevol->productEntry,
		QUANTITY_ENTRY_SIZE+2,		atof(pevol->quantityEntry),
		UNITS_ENTRY_SIZE+2,		pevol->quantityUnitsEntry,
		WEIGHTAVG_ENTRY_SIZE+2,		pevol->weightavg,
		MOVE_TYPE_ENTRY_SIZE+2,		pevol->movementTypeEntry,
		AUTO_SWING_ENTRY_SIZE+5,	pevol->autoSwingEntry,
		CLIENT_ENTRY_SIZE+2,		pevol->clientEntry,
		CLIENT_ID_ENTRY_SIZE+2,		pevol->clientIdEntry,
		DIC_CODE_ENTRY_SIZE+2,		pevol->dicCodeEntry,
		DATE_STRING_ENTRY_SIZE+2,	pevol->dateEntry,
		TIME_STRING_ENTRY_SIZE+2,	pevol->timeEntry,
		STATUS_ENTRY_SIZE+2,		status);
	}

char *
formatSrcDestField(type, pseg, d, n)		/* provide tank* type formatting */
char		type;
SegRecord	*pseg;
char		*d;
int		n;
	{
	int		i, j, nstr;
	char		*p, **str;
	char		s[BUFSIZ];
	D4DATA		*desc;
	F4FIELD		*field, *evolField, *segField;
	char		*fieldname, *tagname;
	
	switch (type) {
	    case 's':
		desc = segSrcsDesc;
		field = segSrcs_SOURCE;
		evolField = segSrcs_EVOLNUMBER;
		segField = segSrcs_SEGNUMBER;	
		tagname = "SEGSRCSTAG";	
		break;
	    case 'd':
		desc = segDestsDesc;
		field = segDests_DEST;
		evolField = segDests_EVOLNUMBER;
		segField = segDests_SEGNUMBER;
		tagname = "SEGDESTSTAG";
		break;
	}

	/*
	 * get a list of the source or destination names
	 */
	s[0] = 0;
	d4tag_select(desc, d4tag(desc, tagname));
	if (d4top(desc) == 0)
		do
			{
			long	evolNumber	= f4long (evolField);
			long	segNumber	= f4long (segField);
			
			if (evolNumber == pseg->evolNumber && segNumber == pseg->segNumber)
				{
				char	*name	= f4str (field);
				
				strcat(StripWhitespace(strncat(s, name, strlen(name))), " ");
				}
			d4unlock_all (desc);
			}
		while (d4skip(desc, 1) == 0);
	d4unlock_all(desc);

	/*
	 * compare the ith character of the first string to any others
	 */
	strncpy(d, s, n);
	
	for (nstr = 0, p = StripWhitespace(s); *p; p++)
		if (*p == ' ')
			nstr++;
	if (nstr)
		{
		str = (char **)calloc(nstr, sizeof (char *));
		for (i = 0, p = s; *p; p++)
			if (*p == ' ')
				str[i++] = p+1;
		for (p = s, j = 0; *p && *p != ' '; p++, j++)
			for (i = 0; i < nstr; i++)
				if (*p != str[i][j])
					{
					d[j] = '*';
					d[++j] = 0;
					free(str);
					return d;
					}
		for (i = 0; i < nstr; i++)
			if (str[i][j] && str[i][j] != ' ')
				{
				d[j] = '*';
				d[++j] = 0;
				free(str);
				return d;
				}
		d[j] = 0;
		free(str);
		}
	d[n] = 0;
	return d;
	}
	
formatSegmentString(pseg, txtBuf)
SegRecord	*pseg;
char		*txtBuf;
	{
	char	srcBuf[SRCBUFLEN+1];
	char	destBuf[DESTBUFLEN+1];
	char	methodName[METHOD_ENTRY_SIZE+1];
	char	searchString[BUFSIZ];
	int	d4status;
	double	estFlowTime; 
	double	estFlowRate;
	char	*status, estTimeString [9], actTimeString [9];
	
	if ((d4status = d4top(lineupDesc)) == 0) {
		do {
			if (strcmp(pseg->lineupEntry, StripWhitespace(f4str(lineup_LINEUPNAME))) == 0)
				break;
		    d4unlock_all(lineupDesc);
		} while (d4skip(lineupDesc, 1) == 0);
		d4unlock_all(lineupDesc);
	}

	if (d4status || d4eof(lineupDesc))
		{
		methodName[0] = 0;
		estFlowRate = 0.0;
		estFlowTime = 0.0;
		}
	else
		{
		double	qtyInBBLS = atof(pseg->quantityEntry) / ((strcmp(pseg->quantityUnitsEntry, BBLS_STRING) == 0) ? 1.0 : 42.0);

		dupString(methodName, f4str (lineup_METHOD), METHOD_ENTRY_SIZE+1);
		estFlowRate	= f4double (lineup_FLOWRATE);	/* this is assumed to be bbls/min */
		estFlowTime	= qtyInBBLS / estFlowRate * 60.0;	/* this is then specified in seconds */
		}
	status = (strncmp(pseg->statusEntry, "Swing", 5) == 0) ? "Swinging": pseg->statusEntry;
	formatTime ((time_t)(pseg->actFlowTime), actTimeString);
	formatTime ((time_t)estFlowTime, estTimeString);

	sprintf(txtBuf, "%-*s%-*s%-*s%-*.3lf%-*s%-*s%-*s%-*s%-*.3lf%-*.3lf%-*s%-*s%-*s%-*.3lf",
		ORDER_NUMBER_ENTRY_SIZE+2,	pseg->orderNumberEntry,
		SRCBUFLEN+2,			formatSrcDestField('s', pseg, srcBuf, (sizeof srcBuf) - 1),
		DESTBUFLEN+2,			formatSrcDestField('d', pseg, destBuf, (sizeof destBuf) - 1),
		QUANTITY_ENTRY_SIZE+2,		atof(pseg->quantityEntry),
		UNITS_ENTRY_SIZE+2,		pseg->quantityUnitsEntry,
		METHOD_ENTRY_SIZE+2,		methodName,
		PAUSE_ON_COMPL_ENTRY_SIZE+9,	pseg->pauseEntry,
		AVGAPI_ENTRY_SIZE+2,		pseg->avgapi,
		EST_FLOW_RATE_ENTRY_SIZE+2,	estFlowRate,
		ACT_FLOW_RATE_ENTRY_SIZE+2,	pseg->actFlowRate,
		EST_FLOW_TIME_ENTRY_SIZE+2,	estTimeString,
		ACT_FLOW_TIME_ENTRY_SIZE+2,	actTimeString,
		STATUS_ENTRY_SIZE+2,		status,
		QUANTITY_ENTRY_SIZE+2,		pseg->quantityXfer);
	}

void
getEvol(pevol)
EvolRecord	*pevol;
	{
	memset ((char *) pevol, 0, sizeof *pevol);
	dupString(pevol->evolNumberEntry,	f4str (evol_EVOLNUMBER),EVOL_NUMBER_ENTRY_SIZE);
	dupString(pevol->productEntry,		f4str (evol_PRODUCT),	PRODUCT_ENTRY_SIZE);
	dupString(pevol->quantityEntry,		f4str (evol_QUANTITY),	QUANTITY_ENTRY_SIZE);
	dupString(pevol->quantityUnitsEntry,	f4str (evol_UNITS),	UNITS_ENTRY_SIZE);
	dupString(pevol->weightavg,		f4str (evol_WEIGHTAVG),	WEIGHTAVG_ENTRY_SIZE);
	dupString(pevol->movementTypeEntry,	f4str (evol_MOVETYPE),	MOVE_TYPE_ENTRY_SIZE);
	dupString(pevol->autoSwingEntry,	f4str (evol_AUTOSWNG),	AUTO_SWING_ENTRY_SIZE);
	dupString(pevol->clientEntry,		f4str (evol_CLIENT),	CLIENT_ENTRY_SIZE);
	dupString(pevol->clientIdEntry,		f4str (evol_CLIENTID),	CLIENT_ID_ENTRY_SIZE);
	dupString(pevol->dicCodeEntry,		f4str (evol_DICCODE),	DIC_CODE_ENTRY_SIZE);
	dupString(pevol->dateEntry,		f4str (evol_DATE),	DATE_STRING_ENTRY_SIZE);
	dupString(pevol->timeEntry,		f4str (evol_TIME),	TIME_STRING_ENTRY_SIZE);
	dupString(pevol->statusEntry,		f4str (evol_STATUS),	STATUS_ENTRY_SIZE);
	pevol->evolNumber = atol(pevol->evolNumberEntry);
	if (pevol->evolNumber == selectedEvol)
		tmpEvol = *pevol;
	}
	
void
getEvolSegment(pseg)
SegRecord	*pseg;
	{
	memset ((char *) pseg, 0, sizeof *pseg);
	dupString(pseg->evolNumberEntry,	f4str (evolSegs_EVOLNUMBER),	EVOL_NUMBER_ENTRY_SIZE);
	dupString(pseg->segNumberEntry,		f4str (evolSegs_SEGNUMBER),	SEG_NUMBER_ENTRY_SIZE);
	dupString(pseg->orderNumberEntry,	f4str (evolSegs_ORDERNUM),	ORDER_NUMBER_ENTRY_SIZE);
	dupString(pseg->quantityEntry,		f4str (evolSegs_QUANTITY),	QUANTITY_ENTRY_SIZE);
	dupString(pseg->quantityUnitsEntry,	f4str (evolSegs_UNITS),		UNITS_ENTRY_SIZE);
	dupString(pseg->lineupEntry,		f4str (evolSegs_LINEUPNAME),	LINEUP_NAME_ENTRY_SIZE);
	dupString(pseg->pauseEntry,		f4str (evolSegs_PAUSE),		PAUSE_ENTRY_SIZE);
	dupString(pseg->statusEntry,		f4str (evolSegs_STATUS),	STATUS_ENTRY_SIZE);
	dupString(pseg->avgapi,			f4str (evolSegs_AVGAPI),	STATUS_ENTRY_SIZE);
	pseg->evolNumber	= f4long (evolSegs_EVOLNUMBER);
	pseg->segNumber		= f4long (evolSegs_SEGNUMBER);
	pseg->orderNumber	= f4long (evolSegs_ORDERNUM);
	pseg->actFlowRate	= f4double (evolSegs_ACTFLWRT);
	pseg->actFlowTime	= f4double (evolSegs_ACTFLWTIME);
	pseg->quantityXfer	= f4double (evolSegs_VOLXFERED);
	if (pseg->evolNumber == selectedEvol && pseg->segNumber == selectedSeg)
		tmpSeg = *pseg;
	}

reorderSegments()
	{
	SegRecord	seg;
	int		max	= 0;
	char		txtBuf[100];		/* 4/13/94 DEK - Buffer for seeking */
	int		rv;			/* 4/13/94 DEK - CodeBase return */
	unsigned long	segNumbers[100];	/* 4/14/94 DEK - Stores segment #s */
	int		i;

	d4tag_select(evolSegsDesc, d4tag(evolSegsDesc, "ORDERTAG"));

	/* 4/13/94 DEK - Setup search string with selectedEvol */
	memset ( txtBuf, 0, sizeof txtBuf );
	sprintf (txtBuf, "%*d", EVOL_NUMBER_ENTRY_SIZE, selectedEvol );

	if (d4top(evolSegsDesc) == 0) {
	    /* 4/13/94 DEK - Perform seek to first record with selectedEvol */
		rv = d4seek( evolSegsDesc, txtBuf );

		while(1) {
			getEvolSegment(&seg);
			for ( i=0; (i<max) && (segNumbers[i] != seg.segNumber); i++);
			if (i < max) {
				/* This was already written.  Skip it.  */
				rv = d4skip(evolSegsDesc, 1);
				continue;
			}
			if (seg.evolNumber == selectedEvol) {

				/* Only process this evolution's records. */
				char	orderNumber[ORDER_NUMBER_ENTRY_SIZE+1];

				/* Remember segment # - it may come around one more. */
				segNumbers[max++] = seg.segNumber;

				/* Assign new order #. */
				sprintf(orderNumber, "%-*d", ORDER_NUMBER_ENTRY_SIZE, max);
				f4assign (evolSegs_ORDERNUM, orderNumber);
				updateDBASEfile(evolSegsDesc);	/*Get it all out to disk. */
				d4reindex(evolSegsDesc );  /* Sorry, but it's the only way... */
				d4tag_select(evolSegsDesc, d4tag(evolSegsDesc, "ORDERTAG"));   /* Just in case */
				memset (txtBuf, 0, sizeof txtBuf );
				if (seg.orderNumber == max ) {
					/* Nothing was changed.  Skip to next record.  */
					sprintf (txtBuf, "%*d%*d%*d", EVOL_NUMBER_ENTRY_SIZE, selectedEvol,
								ORDER_NUMBER_ENTRY_SIZE, seg.orderNumber,
								SEG_NUMBER_ENTRY_SIZE, seg.segNumber );
				} else {
					sprintf (txtBuf, "%*d%*d", EVOL_NUMBER_ENTRY_SIZE, selectedEvol,
								ORDER_NUMBER_ENTRY_SIZE, max );
				}
				rv = d4seek( evolSegsDesc, txtBuf );
				if (seg.orderNumber == max) {
					/* Skip to next record. */
					rv = d4skip(evolSegsDesc, 1);
				}
			} else
				break;
		}
	}
	updateDBASEfile(evolSegsDesc);
}

void
RefreshCallback()
	{
	char		txtBuf[BUFSIZ];
	int		evolFileChanged;
	int		evolSegsFileChanged;
	struct stat	evolStatbuf;
	struct stat	evolSegsStatbuf;
	static time_t	evolMtime = 0L;
	static time_t	evolSegsMtime = 0L;
	int		item;

	XtAppAddTimeOut(applContext, (long)REFRESH_RATE, RefreshCallback, NULL);	/* set up for the next time */
	if (editInProgress)
		return;

	inhibitRedundancy ();

	/*
	 * check the modification dates of the dBASE files to see if any were changed
	 * save these file times for future tests
	 */
	stat(evolName, &evolStatbuf);
	evolFileChanged = (evolStatbuf.st_mtime > evolMtime);
	evolMtime = evolStatbuf.st_mtime;

	if (evolFileChanged || forcedUpdate)
		{
		int		evolCount	= 0;

		/*
		 * read and display the data
		 */
		XmListDeleteAllItems(evolList);
		d4tag_select(evolDesc, d4tag(evolDesc, "TIMESTAMP"));
		if (d4top(evolDesc) == 0)
			{
			do
				{
				EvolRecord	evol;
				XmString	s;

				getEvol(&evol);
				formatEvolutionString(&evol, txtBuf);
				s = XmStringCreate(txtBuf, XmSTRING_DEFAULT_CHARSET);
				XmListAddItem(evolList, s, 0);
				XmStringFree(s);
				if (evol.evolNumber == selectedEvol)
					{
					XmListSelectPos(evolList, 0, False);
					XmListSetPos(evolList, 0);
					}
				evolCount++;
				d4unlock_all (evolDesc);
				}
			while (d4skip(evolDesc, 1) == 0);
			}
		d4unlock_all(evolDesc);
		XtSetOneValue(printButton, XmNsensitive, (evolCount > 0) ? OperPriv(&codeBaseDesc, "EVOL_MGT_PRNT") : False);
		}

	stat(evolSegsName, &evolSegsStatbuf);
	evolSegsFileChanged	= (evolSegsStatbuf.st_mtime > evolSegsMtime);
	evolSegsMtime		= evolSegsStatbuf.st_mtime;

	if (evolSegsFileChanged || forcedUpdate)
		{
		static int	firstTime	= TRUE;
		
		if (firstTime)
			{
			lstInit(&segIdList);
			firstTime = FALSE;
			}
		else
			lstFree(&segIdList);

		XmListDeleteAllItems(segList);
		if (selectedEvol)
			{
			d4tag_select(evolSegsDesc, d4tag(evolSegsDesc, "ORDERTAG"));
			item = 1;
			if (d4top(evolSegsDesc) == 0)
				{
				do
					{
					SegList		*psegIdList;
					SegRecord	seg;

					getEvolSegment(&seg);
					if (seg.evolNumber == selectedEvol)
						{
						XmString	s;
						
						formatSegmentString(&seg, txtBuf);
						s = XmStringCreate(txtBuf, XmSTRING_DEFAULT_CHARSET); 
						XmListAddItem(segList, s, 0);
						XmStringFree(s);
						if (seg.segNumber == selectedSeg)
							{
							XmListSelectPos(segList, 0, False);
							XmListSetPos(segList, 0);
							}
						psegIdList = (SegList *)calloc(1, sizeof (SegList));
						psegIdList->idNumber = seg.segNumber;
						lstAdd(&segIdList, &psegIdList->node);
						}
					d4unlock_all (evolSegsDesc);
					}
				while (d4skip(evolSegsDesc, 1) == 0);
				}
			d4unlock_all(evolSegsDesc);
			}
		}
	if (evolFileChanged || evolSegsFileChanged || forcedUpdate)
		SetButtonTitles();

	forcedUpdate	= FALSE;

	allowRedundancy ();

	}

struct privData
	{
	Widget	*widget;
	char	*groupName;
	}
privData[] =
	{
	&apiEntry,		"EVOL_MGT_APIE",
	&clientEntry,		"EVOL_MGT_CLNT",
	&clientIdEntry,		"EVOL_MGT_CIDC",
	&dateEntry,		"EVOL_MGT_DATE",
/*	&deleteButton,		"EVOL_MGT_DELT", */
	&dicCodeEntry,		"EVOL_MGT_DICC",
/*	&displayButton,		"EVOL_MGT_DISP", */
/*	&editButton,		"EVOL_MGT_EDIT", */
	&emrgstopButton,	"EVOL_MGT_EMGS",
	&evolBblsButton,	"EVOL_MGT_EUNT",
	&evolGalsButton,	"EVOL_MGT_EUNT",
	&evolQuantityEntry,	"EVOL_MGT_QTYE",
/*	&insertButton,		"EVOL_MGT_INST", */
/*	&manualButton,		"EVOL_MGT_MANE", */
	&noAutoSwingButton,	"EVOL_MGT_AUTO",
/*	&pauseButton,		"EVOL_MGT_PASE", */
	&pauseOnCompButton,	"EVOL_MGT_PAUS",
	&pierIssueButton,	"EVOL_MGT_MOVE",
	&pierReceiveButton,	"EVOL_MGT_MOVE",
/*	&printButton,		"EVOL_MGT_PRNT", */
/*	&quantityButton,	"EVOL_MGT_SQNT", */
	&quantityEntry,		"EVOL_MGT_EQNT",
/*	&resumeButton,		"EVOL_MGT_RSUM", */
	&segBblsButton,		"EVOL_MGT_SUNT",
	&segGalsButton,		"EVOL_MGT_SUNT",
	&segQuantityEntry,	"EVOL_MGT_SQNT",
/*	&startButton,		"EVOL_MGT_STRT", */
/*	&stopButton,		"EVOL_MGT_STOP", */
	&timeEntry,		"EVOL_MGT_TIME",
	&transferButton,	"EVOL_MGT_MOVE",
	&truckIssueButton,	"EVOL_MGT_MOVE",
	&truckReceiveButton,	"EVOL_MGT_MOVE",
	&yesAutoSwingButton,	"EVOL_MGT_AUTO",
	&productList,		"EVOL_MGT_PROD",
	&lineupList,		"EVOL_MGT_LINE",
	&sourceList,		"EVOL_MGT_SOUR",
	&destinationList,	"EVOL_MGT_DEST",
	0,			0
	};

static void setOperPriv()
	{
	struct privData	*p;
	int	ok;
	
	for (p = &privData[0]; p->widget; p++)
		{
		ok = OperPriv(&codeBaseDesc, p->groupName);
		XtSetOneValue(*p->widget, XmNsensitive, ok);
		}
	}

/*----------*/
updateDBASEfile (data)

    D4DATA	*data;

    {

	d4flush_all (data);
#ifndef CB51
	data->file_lock = TRUE;		/* to force d4update_header() to write the num_recs field */
	d4update_header (data, 1, 1);
	data->file_lock = FALSE;	/* to force d4unlock_all() to write the num_recs field */
#endif
	d4unlock_all (data);

    }

/*----------*/
formatTime (longTime, strTime)

    time_t	longTime;
    char	*strTime;

    {

	struct tm	*tim;

	tim = gmtime (&longTime);
	strftime (strTime, 9, "%T", tim);

    }
