/****************************************************************/
/*								*/
/*	This software was developed for DLA by Tate Integrated  */
/*	Systems, L.P. via a contract with Vitro Corporation.	*/
/*	It is designed to work with TIS4000 Release 4.2.	*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*****************************************************************************
 *                                                                           *
 *      name:   lineup.c                                                     *
 *      date:   april 22, 1993                                               *
 *                                                                           *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <d4all.h>

#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>

#include <Xm/Xm.h>
#include <Xm/PushB.h>
#include <Xm/MessageB.h>
#include <Xm/SelectioB.h>
#include <Xm/Text.h>
#include <Xm/List.h>
#include <Xm/MwmUtil.h>
#include <Mrm/MrmPublic.h>

#include <OperPriv.h>

#include <TFdBASE.h>
#include <lineup.h>

char	*mktemp ();


/*****************************************************************************
 *                                                                           *
 *      static functions                                                     *
 *                                                                           *
 *****************************************************************************/

static void		OpenDatabase ();
static void		LoadCurrentRecord ();
static void		SaveCurrentRecord ();
static void		DisplayCurrentRecord ();
static int		GetDisplay ();
static void		RegisterWidgets ();
static void		ButtonCallback ();
static void		ChangeCallback ();
static void		CmpntListCallback ();
static void		TankListCallback ();
static void		FocusCallback ();
static void		NextLineup ();
static void		PreviousLineup ();
static void		ChangeLineup ();
static void		SetNextPrev ();
static void		InsertLineup ();
static void		DeleteLineup ();
static void		DeleteLineupParts ();
static void		DeleteCmpnt ();
static void		ErrorMessage ();
static void		ErrorMessageCallback ();
static void		WarningMessage ();
static void		WarningMessageOkCallback ();
static void		WarningMessageCancelCallback ();
static void		PopupConfirmMessage ();
static char		*StripWhitespace ();
static void		setOperPriv();

/*****************************************************************************
 *                                                                           *
 *      static variables                                                     *
 *                                                                           *
 *****************************************************************************/

static XtAppContext	applContext;
static Display		*display;
static Widget		appShell;
static Widget		lineupWindow;
static MrmType		lineupWindowClass;
static MrmHierarchy	mrmHierarchy;
static String		uidFilenames[] =
{
	"lineup.uid"
};
#define	NUM_UID_FILES	(sizeof (uidFilenames) / sizeof (uidFilenames[0]))

static MrmRegisterArg	namesRegister[] =
{
	{ (String) "RegisterWidgets",		(caddr_t) RegisterWidgets },
	{ (String) "ButtonCallback",		(caddr_t) ButtonCallback },
	{ (String) "ChangeCallback",		(caddr_t) ChangeCallback },
	{ (String) "CmpntListCallback",		(caddr_t) CmpntListCallback },
	{ (String) "FocusCallback",		(caddr_t) FocusCallback }
};
#define	NUM_NAME_REGISTER	(sizeof (namesRegister) / sizeof (namesRegister[0]))

static int		insertFlag;
static int		changeFlag;

static XmString		nullString;

static XmString		errorMessageString;
static Widget		errorMessage;

static void		(*warningMessageFunc) ();
static XmString		warningMessageString;
static Widget		warningMessage;

static char		*whitespace = " \t\r\n\f";;

static Widget		saveButton;
static Widget		insertButton;
static Widget		deleteButton;
static Widget		printButton;
static Widget		nextButton;
static Widget		previousButton;
static Widget		getCmpntsButton;
static Widget		clearButton;
static Widget		cslMngrButton;
static Widget		lineupNameEntry;
static Widget		methodEntry;
static Widget		srcsEntry;
static Widget		destsEntry;
static Widget		setupTimeEntry;
static Widget		shtdwnTimeEntry;
static Widget		flowRateEntry;
static Widget		msrmntVolsEntry;
static Widget		xrefVolsEntry;
static Widget		relDsplyEntry;
static Widget		manCslScriptEntry;
static Widget		startScriptEntry;
static Widget		stopScriptEntry;
static Widget		pauseScriptEntry;
static Widget		resumeScriptEntry;
static Widget		monitorScriptEntry;
static Widget		cmpntsList;
static Widget		cmpntsDialog;
static Widget		tankList;
static Widget		cmpntEntry;
static Widget		stateEntry;

static unsigned int	maxCmpnts;
static unsigned int	numCmpnts;
static int		selectedCmpnt;

static int		focusedEntry;
static int		addFlag = FALSE;

static unsigned int	lineupRecordNum;
static LineupRecord	lineupRecord;

static C4CODE		codeBaseDesc;
static D4DATA		*lineupDesc;
static D4DATA		*lineupCmpntsDesc;
static D4DATA		*lineupSrcsDesc;
static D4DATA		*lineupDestsDesc;
static D4DATA		*lineupMsrmntVolsDesc;
static D4DATA		*lineupXrefVolsDesc;
static D4DATA		*tankDesc;
static D4DATA		*potDesc;

/*
 * since setting/getting ONE widget resource happens so often:
 */
#define	XtSetOneValue(widget, resource, value) \
	do \
		{ \
		Arg		arg[1]; \
		XtSetArg(arg[0],(resource),(value)); \
		XtSetValues((widget), arg, 1); \
		} \
	while (0)


/*****************************************************************************
 *                                                                           *
 *      main prgram for lineup interface                                     *
 *                                                                           *
 *****************************************************************************/

main (argc, argv)
int	argc;
char	*argv[];
{
	char	txtBuf[BUFSIZ];
	int	i;
	int	n;
	Arg	args[MAX_ARGS];
	int	Cmd, Stat;

	/*
	 * initialize variables
	 */
	focusedEntry = 0;
	addFlag = FALSE;
	errorMessage = NULL;
	errorMessageString = NULL;
	warningMessage = NULL;
	warningMessageString = NULL;
	nullString = XmStringCreate ("", XmSTRING_DEFAULT_CHARSET);

	/*
	 * initialize lineup strings
	 */
	if ( !(lineupRecord.cmpnts = (Component *) calloc (CMPNTS_BLOCK_SIZE, sizeof (Component))))
	{
	    printf ("Out of memory\n");
	    exit (1);
	}
	maxCmpnts = CMPNTS_BLOCK_SIZE;
	numCmpnts = 0;
	selectedCmpnt = 0;

	/*
	 * initialize and open display
	 */
	OpenDatabase ();
	MrmInitialize ();
	XtToolkitInitialize ();
	applContext = XtCreateApplicationContext ();
	display = XtOpenDisplay (applContext, NULL, "lineup", "Lineup", NULL, 0, &argc, argv);
	if ( !display)
	{
	    printf ("Cannot open display\n");
	    exit (1);
	}
	appShell = XtAppCreateShell ("Lineup", "Lineup", applicationShellWidgetClass, display, NULL, 0);

	/*
	 * create main window widget and its children
	 */
	if (MrmOpenHierarchy (NUM_UID_FILES, uidFilenames, NULL, &mrmHierarchy) != MrmSUCCESS)
	{
	    printf ("Cannot open UID file %s\n", uidFilenames[0]);
	    exit (1);
	}
	MrmRegisterNames (namesRegister, NUM_NAME_REGISTER);
	if (MrmFetchWidget (mrmHierarchy, "lineupWindow", appShell, &lineupWindow, &lineupWindowClass) != MrmSUCCESS)
	{
	    printf ("Cannot fetch lineup window widget\n");
	    exit (1);
	}

	/*
	 * create error msg dialog box
	 */
	n = 0;
	XtSetArg (args[n], XmNmessageAlignment, XmALIGNMENT_CENTER); n++;
	XtSetArg (args[n], XmNdialogType, XmDIALOG_ERROR); n++;
	XtSetArg (args[n], XmNdeleteResponse, XmDO_NOTHING); n++;
	XtSetArg (args[n], XmNdialogStyle, XmDIALOG_APPLICATION_MODAL); n++;
	XtSetArg (args[n], XmNtitle, "Error"); n++;
	errorMessage = XmCreateMessageDialog (lineupWindow, "Error Message", args, n);
	XtUnmanageChild (XmMessageBoxGetChild (errorMessage, XmDIALOG_HELP_BUTTON));
	XtUnmanageChild (XmMessageBoxGetChild (errorMessage, XmDIALOG_CANCEL_BUTTON));
	XtAddCallback (XmMessageBoxGetChild (errorMessage, XmDIALOG_OK_BUTTON),
			XmNactivateCallback, ErrorMessageCallback, NULL);

	/*
	 * create warning msg dialog box
	 */
	n = 0;
	XtSetArg (args[n], XmNmessageAlignment, XmALIGNMENT_CENTER); n++;
	XtSetArg (args[n], XmNdialogType, XmDIALOG_WARNING); n++;
	XtSetArg (args[n], XmNdeleteResponse, XmDO_NOTHING); n++;
	XtSetArg (args[n], XmNdialogStyle, XmDIALOG_APPLICATION_MODAL); n++;
	XtSetArg (args[n], XmNtitle, "Warning"); n++;
	warningMessage = XmCreateQuestionDialog (lineupWindow, "Warning Message", args, n);
	XtUnmanageChild (XmMessageBoxGetChild (warningMessage, XmDIALOG_HELP_BUTTON));
	XtAddCallback (XmMessageBoxGetChild (warningMessage, XmDIALOG_CANCEL_BUTTON),
			XmNactivateCallback, WarningMessageCancelCallback, NULL);
	n = 0;
	XtSetArg (args[n], XmNlabelString, XmStringCreate ("NO", XmSTRING_DEFAULT_CHARSET)); n++;
	XtSetValues (XmMessageBoxGetChild (warningMessage, XmDIALOG_CANCEL_BUTTON), args, n);
	XtAddCallback (XmMessageBoxGetChild (warningMessage, XmDIALOG_OK_BUTTON),
			XmNactivateCallback, WarningMessageOkCallback, NULL);
	n = 0;
	XtSetArg (args[n], XmNlabelString, XmStringCreate ("YES", XmSTRING_DEFAULT_CHARSET)); n++;
	XtSetValues (XmMessageBoxGetChild (warningMessage, XmDIALOG_OK_BUTTON), args, n);

	/*
	 * open database
	 */
	d4tag_select (lineupDesc, d4tag (lineupDesc, lineupTagInfo[0].name));
	if (d4top (lineupDesc) == 0)
	{
	    lineupRecordNum = d4recno (lineupDesc);
	    LoadCurrentRecord ();
	    DisplayCurrentRecord ();
	    insertFlag = FALSE;
	    n = 0;
	    XtSetArg (args[n], XmNsensitive, OperPriv(&codeBaseDesc, "LINE_DEF_DELT")); n++;
	    XtSetValues (deleteButton, args, n);
	    n = 0;
	    XtSetArg (args[n], XmNsensitive, False); n++;
	    XtSetValues (previousButton, args, n);
	    if (d4reccount (lineupDesc) == 1)
	    {
		XtSetValues (nextButton, args, n);
	    }
	}
	else
	{
	    lineupRecordNum = 0;
	    insertFlag = TRUE;
	    n = 0;
	    XtSetArg (args[n], XmNsensitive, False); n++;
	    XtSetValues (deleteButton, args, n);
	    XtSetValues (printButton, args, n);
	    XtSetValues (previousButton, args, n);
	    XtSetValues (nextButton, args, n);
	}
	changeFlag = FALSE;
	d4tag_select (lineupDesc, NULL);
	d4unlock_all (lineupDesc);

	d4tag_select (tankDesc, d4tag (tankDesc, "NAME"));
	if (d4top (tankDesc) == 0)
	{
	    i = 1;
	    do
	    {
		XmListAddItem (XmSelectionBoxGetChild (tankList, XmDIALOG_LIST),
			XmStringCreate (StripWhitespace (f4str (d4field (tankDesc, "NAME"))),
			XmSTRING_DEFAULT_CHARSET), i++);
	    }
	    while (d4skip (tankDesc, 1) == 0);
	}
	d4tag_select (tankDesc, NULL);
	d4unlock_all (tankDesc);

	d4tag_select (potDesc, d4tag (potDesc, "NAME"));
	if (d4top (potDesc) == 0)
	{
	    i = 1;
	    do
	    {
		XmListAddItem (XmSelectionBoxGetChild (tankList, XmDIALOG_LIST),
			XmStringCreate (StripWhitespace (f4str (d4field (potDesc, "NAME"))),
			XmSTRING_DEFAULT_CHARSET), i++);
	    }
	    while (d4skip (potDesc, 1) == 0);
	}
	d4tag_select (potDesc, NULL);
	d4unlock_all (potDesc);

	n = 0;
	XtSetArg (args[n], XmNsensitive, False); n++;
	XtSetValues (cslMngrButton, args, n);

	XmAddTabGroup (lineupNameEntry);
	XmAddTabGroup (methodEntry);
	XmAddTabGroup (srcsEntry);
	XmAddTabGroup (destsEntry);
	XmAddTabGroup (setupTimeEntry);
	XmAddTabGroup (shtdwnTimeEntry);
	XmAddTabGroup (flowRateEntry);
	XmAddTabGroup (msrmntVolsEntry);
	XmAddTabGroup (xrefVolsEntry);
	XmAddTabGroup (relDsplyEntry);
	XmAddTabGroup (manCslScriptEntry);
	XmAddTabGroup (startScriptEntry);
	XmAddTabGroup (stopScriptEntry);
	XmAddTabGroup (pauseScriptEntry);
	XmAddTabGroup (resumeScriptEntry);
	XmAddTabGroup (monitorScriptEntry);
	XmAddTabGroup (cmpntsList);

	/*
	 * realize main window
	 */
	XtManageChild (lineupWindow);
	XtRealizeWidget (appShell);
	setOperPriv();

	/*
	 * get and dispatch events
	 */
	XtAppMainLoop (applContext);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      OpenDatabase                                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          open lineup database file                                        *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
OpenDatabase ()
{
	struct stat	statBuf;

	d4init (&codeBaseDesc);
	codeBaseDesc.read_lock = 0;
	codeBaseDesc.create_error = 0;
	codeBaseDesc.field_name_error = 0;
	codeBaseDesc.go_error = 0;
	codeBaseDesc.open_error = 0;
	codeBaseDesc.skip_error = 0;
	codeBaseDesc.tag_name_error = 0;
	
	if ( !(lineupDesc = d4open (&codeBaseDesc, lineupName)))
	    printf ("Cannot open lineup database file \"%s\"\n", lineupName), exit (0);
	d4unlock_all (lineupDesc);

	if ( !(lineupCmpntsDesc = d4open (&codeBaseDesc, lineupCmpntsName)))
	    printf ("Cannot open lineup components database file \"%s\"\n", lineupCmpntsName), exit (0);
	d4unlock_all (lineupCmpntsDesc);

	if ( !(lineupSrcsDesc = d4open (&codeBaseDesc, lineupSrcsName)))
	    printf ("Cannot open lineup sources database file \"%s\"\n", lineupSrcsName), exit (0);
	d4unlock_all (lineupSrcsDesc);

	if ( !(lineupDestsDesc = d4open (&codeBaseDesc, lineupDestsName)))
	    printf ("Cannot open lineup destinations database file \"%s\"\n", lineupDestsName), exit (0);
	d4unlock_all (lineupDestsDesc);

	if ( !(lineupMsrmntVolsDesc = d4open (&codeBaseDesc, lineupMsrmntVolsName)))
	    printf ("Cannot open lineup measurement volumes database file \"%s\"\n", lineupMsrmntVolsName), exit (0);
	d4unlock_all (lineupMsrmntVolsDesc);

	if ( !(lineupXrefVolsDesc = d4open (&codeBaseDesc, lineupXrefVolsName)))
	    printf ("Cannot open lineup cross reference volumes database file \"%s\"\n", lineupXrefVolsName), exit (0);
	d4unlock_all (lineupXrefVolsDesc);

	if ( !(tankDesc = d4open (&codeBaseDesc, tankName)))
	    printf ("Cannot open lineup tank database file \"%s\"\n", tankName), exit (0);
	d4unlock_all (tankDesc);

	if ( !(potDesc = d4open (&codeBaseDesc, potName)))
	    printf ("Cannot open lineup pot database file \"%s\"\n", potName), exit (0);
	d4unlock_all (potDesc);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      LoadCurrentRecord                                                    *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          load current lineup database record                              *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
LoadCurrentRecord ()
{
	int		i, j;
	int		n;
	Arg		args[MAX_ARGS];
	char		lineupName[LINEUP_NAME_ENTRY_SIZE+1];
	char		cmpnt[CHANNEL_NAME_ENTRY_SIZE+1];
	char		state[STATE_ENTRY_SIZE+1];
	Component	*oldCmpnts;

	oldCmpnts = lineupRecord.cmpnts;
	memset ((char *) &lineupRecord, 0, sizeof (LineupRecord));
	lineupRecord.cmpnts = oldCmpnts;

	strncpy (lineupRecord.lineupName, f4str (d4field (lineupDesc, "LINEUPNAME")), LINEUP_NAME_ENTRY_SIZE);
	StripWhitespace (lineupRecord.lineupName);
	sprintf (lineupName, "%-*s", LINEUP_NAME_ENTRY_SIZE, f4str (d4field (lineupDesc, "LINEUPNAME")));

	strncpy (lineupRecord.method, f4str (d4field (lineupDesc, "METHOD")), METHOD_ENTRY_SIZE);
	StripWhitespace (lineupRecord.method);

	strncpy (lineupRecord.setupTime, f4str (d4field (lineupDesc, "SETUPTIME")), TIME_STRING_ENTRY_SIZE);
	StripWhitespace (lineupRecord.setupTime);

	strncpy (lineupRecord.shtdwnTime, f4str (d4field (lineupDesc, "SHTDWNTIME")), TIME_STRING_ENTRY_SIZE);
	StripWhitespace (lineupRecord.shtdwnTime);

	strncpy (lineupRecord.flowRate, f4str (d4field (lineupDesc, "FLOWRATE")), FLOW_RATE_ENTRY_SIZE);
	StripWhitespace (lineupRecord.flowRate);

	strncpy (lineupRecord.relDsply, f4str (d4field (lineupDesc, "RELDSPLY")), REL_DSPLY_ENTRY_SIZE);
	StripWhitespace (lineupRecord.relDsply);

	strncpy (lineupRecord.manCslScript, f4str (d4field (lineupDesc, "MANCSL")), CSL_SCRIPT_ENTRY_SIZE);
	StripWhitespace (lineupRecord.manCslScript);

	strncpy (lineupRecord.startScript, f4str (d4field (lineupDesc, "STARTSCRPT")), CSL_SCRIPT_ENTRY_SIZE);
	StripWhitespace (lineupRecord.startScript);

	strncpy (lineupRecord.stopScript, f4str (d4field (lineupDesc, "STOPSCRPT")), CSL_SCRIPT_ENTRY_SIZE);
	StripWhitespace (lineupRecord.stopScript);

	strncpy (lineupRecord.pauseScript, f4str (d4field (lineupDesc, "PAUSESCRPT")), CSL_SCRIPT_ENTRY_SIZE);
	StripWhitespace (lineupRecord.pauseScript);

	strncpy (lineupRecord.resumeScript, f4str (d4field (lineupDesc, "RESUMSCRPT")), CSL_SCRIPT_ENTRY_SIZE);
	StripWhitespace (lineupRecord.resumeScript);

	strncpy (lineupRecord.monitorScript, f4str (d4field (lineupDesc, "MONSCRPT")), CSL_SCRIPT_ENTRY_SIZE);
	StripWhitespace (lineupRecord.monitorScript);

	d4tag_select (lineupCmpntsDesc, d4tag (lineupCmpntsDesc, lineupTagInfo[0].name));
	i = 0;
	if (d4seek (lineupCmpntsDesc, lineupName) == 0)
	{
	    do
	    {
		if (numCmpnts >= maxCmpnts)
		{
		    oldCmpnts = lineupRecord.cmpnts;
		    maxCmpnts += CMPNTS_BLOCK_SIZE;
		    if ( !(lineupRecord.cmpnts = (Component *) calloc (CMPNTS_BLOCK_SIZE * maxCmpnts, sizeof (Component))))
		    {
			printf ("Out of memory\n");
			exit (1);
		    }
		    memcpy ((char *) lineupRecord.cmpnts, (char *) oldCmpnts, (maxCmpnts - CMPNTS_BLOCK_SIZE) *
							sizeof (Component));
		    cfree ((char *) oldCmpnts);
		}
		strncpy (cmpnt, f4str (d4field (lineupCmpntsDesc, "CMPNTS")), CHANNEL_NAME_ENTRY_SIZE);
		cmpnt[CHANNEL_NAME_ENTRY_SIZE] = 0;
		strncpy (state, f4str (d4field (lineupCmpntsDesc, "STATE")), STATE_ENTRY_SIZE);
		state[STATE_ENTRY_SIZE] = 0;
		for (i=1; i <= numCmpnts; i++)
		{
		    if (strcmp (cmpnt, lineupRecord.cmpnts[i-1].entry) < 0)
		    {
				break;
		    }
		}
		for (j=numCmpnts; j >= i; j--)
		{
		    lineupRecord.cmpnts[j] = lineupRecord.cmpnts[j - 1];
		}
		StripWhitespace (strcpy (lineupRecord.cmpnts[i - 1].entry, cmpnt));
		StripWhitespace (strcpy (lineupRecord.cmpnts[i - 1].state, state));
		{
		char	buf[BUFSIZ];
		
		sprintf(buf, "%*s %s", CHANNEL_NAME_ENTRY_SIZE, lineupRecord.cmpnts[i - 1].entry,
				lineupRecord.cmpnts[i - 1].state);
		lineupRecord.cmpnts[i - 1].string = XmStringCreate (buf, XmSTRING_DEFAULT_CHARSET);
		}
		XmListAddItem (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST), lineupRecord.cmpnts[i - 1].string, i);
		selectedCmpnt = 0;
		numCmpnts++;
	    }
	    while (d4skip (lineupCmpntsDesc, 1) == 0 &&
			strcmp (f4str (d4field (lineupCmpntsDesc, "LINEUPNAME")), lineupName) == 0);
	    n = 0;
	    XtSetArg (args[n], XmNsensitive, False); n++;
	    XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_CANCEL_BUTTON), args, n);
		XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_APPLY_BUTTON), args, n);
		XtSetOneValue(XmSelectionBoxGetChild (cmpntsList, XmDIALOG_OK_BUTTON), XmNsensitive,
				OperPriv(&codeBaseDesc, "LINE_DEF_COMP"));
	    XmListDeselectAllItems (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST));
	    XmListSetBottomPos (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST), numCmpnts);
	}
	d4tag_select (lineupCmpntsDesc, NULL);
	d4unlock_all (lineupCmpntsDesc);

	d4tag_select (lineupSrcsDesc, d4tag (lineupSrcsDesc, lineupTagInfo[0].name));
	i = 0;
	if (d4seek (lineupSrcsDesc, lineupName) == 0)
	{
	    do
	    {
		strncpy (lineupRecord.srcs[i], f4str (d4field (lineupSrcsDesc, "SRCS")), CHANNEL_NAME_ENTRY_SIZE);
		StripWhitespace (lineupRecord.srcs[i]);
	    }
	    while (d4skip (lineupSrcsDesc, 1) == 0 &&
			strcmp (f4str (d4field (lineupSrcsDesc, "LINEUPNAME")), lineupName) == 0 && ++i < MAX_SRCS);
	}
	d4tag_select (lineupSrcsDesc, NULL);
	d4unlock_all (lineupSrcsDesc);

	d4tag_select (lineupDestsDesc, d4tag (lineupDestsDesc, lineupTagInfo[0].name));
	i = 0;
	if (d4seek (lineupDestsDesc, lineupName) == 0)
	{
	    do
	    {
		strncpy (lineupRecord.dests[i], f4str (d4field (lineupDestsDesc, "DESTS")), CHANNEL_NAME_ENTRY_SIZE);
		StripWhitespace (lineupRecord.dests[i]);
	    }
	    while (d4skip (lineupDestsDesc, 1) == 0 &&
			strcmp (f4str (d4field (lineupDestsDesc, "LINEUPNAME")), lineupName) == 0 && ++i < MAX_DESTS);
	}
	d4tag_select (lineupDestsDesc, NULL);
	d4unlock_all (lineupDestsDesc);

	d4tag_select (lineupMsrmntVolsDesc, d4tag (lineupMsrmntVolsDesc, lineupTagInfo[0].name));
	i = 0;
	if (d4seek (lineupMsrmntVolsDesc, lineupName) == 0)
	{
	    do
	    {
		strncpy (lineupRecord.msrmntVols[i],
			f4str (d4field (lineupMsrmntVolsDesc, "MSRMNTVOLS")), CHANNEL_NAME_ENTRY_SIZE);
		StripWhitespace (lineupRecord.msrmntVols[i]);
	    }
	    while (d4skip (lineupMsrmntVolsDesc, 1) == 0 &&
		strcmp (f4str (d4field (lineupMsrmntVolsDesc, "LINEUPNAME")), lineupName) == 0 && ++i < MAX_MSRMNT_VOLS);
	}
	d4tag_select (lineupMsrmntVolsDesc, NULL);
	d4unlock_all (lineupMsrmntVolsDesc);

	d4tag_select (lineupXrefVolsDesc, d4tag (lineupXrefVolsDesc, lineupTagInfo[0].name));
	i = 0;
	if (d4seek (lineupXrefVolsDesc, lineupName) == 0)
	{
	    do
	    {
		strncpy (lineupRecord.xrefVols[i],
			f4str (d4field (lineupXrefVolsDesc, "XREFVOLS")), CHANNEL_NAME_ENTRY_SIZE);
		StripWhitespace (lineupRecord.xrefVols[i]);
	    }
	    while (d4skip (lineupXrefVolsDesc, 1) == 0 &&
		strcmp (f4str (d4field (lineupXrefVolsDesc, "LINEUPNAME")), lineupName) == 0 && ++i < MAX_XREF_VOLS);
	}
	d4tag_select (lineupXrefVolsDesc, NULL);
	d4unlock_all (lineupXrefVolsDesc);

}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      SaveCurrentRecord                                                    *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          save current lineup database record                              *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
SaveCurrentRecord ()
{
	char		lineupName[LINEUP_NAME_ENTRY_SIZE+1];
	int		i;

	f4assign (d4field (lineupDesc, "LINEUPNAME"), lineupRecord.lineupName);
	strcpy (lineupName, f4str (d4field (lineupDesc, "LINEUPNAME")));
	f4assign (d4field (lineupDesc, "METHOD"), lineupRecord.method);
	f4assign (d4field (lineupDesc, "SETUPTIME"), lineupRecord.setupTime);
	f4assign (d4field (lineupDesc, "SHTDWNTIME"), lineupRecord.shtdwnTime);
	f4assign (d4field (lineupDesc, "FLOWRATE"), lineupRecord.flowRate);
	f4assign (d4field (lineupDesc, "RELDSPLY"), lineupRecord.relDsply);
	f4assign (d4field (lineupDesc, "MANCSL"), lineupRecord.manCslScript);
	f4assign (d4field (lineupDesc, "STARTSCRPT"), lineupRecord.startScript);
	f4assign (d4field (lineupDesc, "STOPSCRPT"), lineupRecord.stopScript);
	f4assign (d4field (lineupDesc, "PAUSESCRPT"), lineupRecord.pauseScript);
	f4assign (d4field (lineupDesc, "RESUMSCRPT"), lineupRecord.resumeScript);
	f4assign (d4field (lineupDesc, "MONSCRPT"), lineupRecord.monitorScript);
	updateDBASEfile (lineupDesc);

	DeleteLineupParts (lineupName);

	for (i=0; i < numCmpnts; i++)
	{
	    d4append_blank (lineupCmpntsDesc);
	    f4assign (d4field (lineupCmpntsDesc, "LINEUPNAME"), lineupRecord.lineupName);
	    f4assign (d4field (lineupCmpntsDesc, "CMPNTS"), lineupRecord.cmpnts[i].entry);
	    f4assign (d4field (lineupCmpntsDesc, "STATE"), lineupRecord.cmpnts[i].state);
	}
	updateDBASEfile (lineupCmpntsDesc);

	for (i=0; i < MAX_SRCS && lineupRecord.srcs[i][0]; i++)
	{
	    d4append_blank (lineupSrcsDesc);
	    f4assign (d4field (lineupSrcsDesc, "LINEUPNAME"), lineupRecord.lineupName);
	    f4assign (d4field (lineupSrcsDesc, "SRCS"), lineupRecord.srcs[i]);
	}
	updateDBASEfile (lineupSrcsDesc);

	for (i=0; i < MAX_DESTS && lineupRecord.dests[i][0]; i++)
	{
	    d4append_blank (lineupDestsDesc);
	    f4assign (d4field (lineupDestsDesc, "LINEUPNAME"), lineupRecord.lineupName);
	    f4assign (d4field (lineupDestsDesc, "DESTS"), lineupRecord.dests[i]);
	}
	updateDBASEfile (lineupDestsDesc);

	for (i=0; i < MAX_MSRMNT_VOLS && lineupRecord.msrmntVols[i][0]; i++)
	{
	    d4append_blank (lineupMsrmntVolsDesc);
	    f4assign (d4field (lineupMsrmntVolsDesc, "LINEUPNAME"), lineupRecord.lineupName);
	    f4assign (d4field (lineupMsrmntVolsDesc, "MSRMNTVOLS"), lineupRecord.msrmntVols[i]);
	}
	updateDBASEfile (lineupMsrmntVolsDesc);

	for (i=0; i < MAX_XREF_VOLS && lineupRecord.xrefVols[i][0]; i++)
	{
	    d4append_blank (lineupXrefVolsDesc);
	    f4assign (d4field (lineupXrefVolsDesc, "LINEUPNAME"), lineupRecord.lineupName);
	    f4assign (d4field (lineupXrefVolsDesc, "XREFVOLS"), lineupRecord.xrefVols[i]);
	}
	updateDBASEfile (lineupXrefVolsDesc);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      DisplayCurrentRecord                                                 *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          display current lineup record                                    *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
DisplayCurrentRecord ()
{
	char		txtBuf[BUFSIZ];
	int		i;

	XmTextSetString (lineupNameEntry, lineupRecord.lineupName);
	XmTextSetString (methodEntry, lineupRecord.method);
	txtBuf[0] = 0;
	for (i=0; i < MAX_SRCS && lineupRecord.srcs[i][0]; i++)
	{
	    strcat (txtBuf, lineupRecord.srcs[i]);
	    strcat (txtBuf, " ");
	}
	XmTextSetString (srcsEntry, txtBuf);
	txtBuf[0] = 0;
	for (i=0; i < MAX_DESTS && lineupRecord.dests[i][0]; i++)
	{
	    strcat (txtBuf, lineupRecord.dests[i]);
	    strcat (txtBuf, " ");
	}
	XmTextSetString (destsEntry, txtBuf);
	XmTextSetString (setupTimeEntry, lineupRecord.setupTime);
	XmTextSetString (shtdwnTimeEntry, lineupRecord.shtdwnTime);
	XmTextSetString (flowRateEntry, lineupRecord.flowRate);
	txtBuf[0] = 0;
	for (i=0; i < MAX_MSRMNT_VOLS && lineupRecord.msrmntVols[i][0]; i++)
	{
	    strcat (txtBuf, lineupRecord.msrmntVols[i]);
	    strcat (txtBuf, " ");
	}
	XmTextSetString (msrmntVolsEntry, txtBuf);
	txtBuf[0] = 0;
	for (i=0; i < MAX_XREF_VOLS && lineupRecord.xrefVols[i][0]; i++)
	{
	    strcat (txtBuf, lineupRecord.xrefVols[i]);
	    strcat (txtBuf, " ");
	}
	XmTextSetString (xrefVolsEntry, txtBuf);
	XmTextSetString (relDsplyEntry, lineupRecord.relDsply);
	XmTextSetString (manCslScriptEntry, lineupRecord.manCslScript);
	XmTextSetString (startScriptEntry, lineupRecord.startScript);
	XmTextSetString (stopScriptEntry, lineupRecord.stopScript);
	XmTextSetString (pauseScriptEntry, lineupRecord.pauseScript);
	XmTextSetString (resumeScriptEntry, lineupRecord.resumeScript);
	XmTextSetString (monitorScriptEntry, lineupRecord.monitorScript);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      GetDisplay                                                           *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          get display text entries                                         *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static int
GetDisplay ()
{
	char		txtBuf[BUFSIZ];
	char		*strPtr, *strEntryPtr;
	int		i;
	int		returnStatus;

	returnStatus = TRUE;
	strcpy (lineupRecord.lineupName, strEntryPtr = StripWhitespace (XmTextGetString (lineupNameEntry)));
	strcpy (lineupRecord.method, strEntryPtr = StripWhitespace (XmTextGetString (methodEntry)));
	XtFree (strEntryPtr);
	strPtr = strtok (strEntryPtr = StripWhitespace (XmTextGetString (srcsEntry)), whitespace);
	for (i=0; i < MAX_SRCS; i++)
	{
	    strcpy (lineupRecord.srcs[i], (strPtr ? strPtr : ""));
	    if (strPtr)
	    {
		strPtr = strtok (NULL, whitespace);
	    }
	    if (i == (MAX_SRCS - 1) && strPtr)
	    {
		sprintf (txtBuf, "More than %d entries in \"Source\" field", MAX_SRCS);
		ErrorMessage (txtBuf);
		returnStatus = FALSE;
	    }
	}
	XtFree (strEntryPtr);
	strPtr = strtok (strEntryPtr = StripWhitespace (XmTextGetString (destsEntry)), whitespace);
	for (i=0; i < MAX_DESTS; i++)
	{
	    strcpy (lineupRecord.dests[i], (strPtr ? strPtr : ""));
	    if (strPtr)
	    {
		strPtr = strtok (NULL, whitespace);
	    }
	    if (i == (MAX_DESTS - 1) && strPtr)
	    {
		sprintf (txtBuf, "More than %d entries in \"Destination\" field", MAX_DESTS);
		ErrorMessage (txtBuf);
		returnStatus = FALSE;
	    }
	}
	XtFree (strEntryPtr);
	strcpy (lineupRecord.setupTime, strEntryPtr = StripWhitespace (XmTextGetString (setupTimeEntry)));
	XtFree (strEntryPtr);
	strcpy (lineupRecord.shtdwnTime, strEntryPtr = StripWhitespace (XmTextGetString (shtdwnTimeEntry)));
	XtFree (strEntryPtr);
	strcpy (lineupRecord.flowRate, strEntryPtr = StripWhitespace (XmTextGetString (flowRateEntry)));
	XtFree (strEntryPtr);
	strPtr = strtok (strEntryPtr = StripWhitespace (XmTextGetString (msrmntVolsEntry)), whitespace);
	for (i=0; i < MAX_MSRMNT_VOLS; i++)
	{
	    strcpy (lineupRecord.msrmntVols[i], (strPtr ? strPtr : ""));
	    if (strPtr)
	    {
		strPtr = strtok (NULL, whitespace);
	    }
	    if (i == (MAX_MSRMNT_VOLS - 1) && strPtr)
	    {
		sprintf (txtBuf, "More than %d entries in \"Measurement Volume\" field", MAX_MSRMNT_VOLS);
		ErrorMessage (txtBuf);
		returnStatus = FALSE;
	    }
	}
	XtFree (strEntryPtr);
	strPtr = strtok (strEntryPtr = StripWhitespace (XmTextGetString (xrefVolsEntry)), whitespace);
	for (i=0; i < MAX_XREF_VOLS; i++)
	{
	    strcpy (lineupRecord.xrefVols[i], (strPtr ? strPtr : ""));
	    if (strPtr)
	    {
		strPtr = strtok (NULL, whitespace);
	    }
	    if (i == (MAX_XREF_VOLS - 1) && strPtr)
	    {
		sprintf (txtBuf, "More than %d entries in \"Cross Reference Volume\" field", MAX_XREF_VOLS);
		ErrorMessage (txtBuf);
		returnStatus = FALSE;
	    }
	}
	XtFree (strEntryPtr);
	strcpy (lineupRecord.relDsply, strEntryPtr = StripWhitespace (XmTextGetString (relDsplyEntry)));
	XtFree (strEntryPtr);
	strcpy (lineupRecord.manCslScript, strEntryPtr = StripWhitespace (XmTextGetString (manCslScriptEntry)));
	XtFree (strEntryPtr);
	strcpy (lineupRecord.startScript, strEntryPtr = StripWhitespace (XmTextGetString (startScriptEntry)));
	XtFree (strEntryPtr);
	strcpy (lineupRecord.stopScript, strEntryPtr = StripWhitespace (XmTextGetString (stopScriptEntry)));
	XtFree (strEntryPtr);
	strcpy (lineupRecord.pauseScript, strEntryPtr = StripWhitespace (XmTextGetString (pauseScriptEntry)));
	XtFree (strEntryPtr);
	strcpy (lineupRecord.resumeScript, strEntryPtr = StripWhitespace (XmTextGetString (resumeScriptEntry)));
	XtFree (strEntryPtr);
	strcpy (lineupRecord.monitorScript, strEntryPtr = StripWhitespace (XmTextGetString (monitorScriptEntry)));
	XtFree (strEntryPtr);
	return (returnStatus);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      RegisterWidgets                                                      *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget descriptor                                *
 *                                                                           *
 *          widgetID        widget ID                                        *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          register a widget that was created from the UIL.                 *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
RegisterWidgets (widget, widgetID, widgetData)
Widget 			widget;
int			*widgetID;
XmAnyCallbackStruct	*widgetData;
{
	int		n;
	Arg		args[MAX_ARGS];

	switch (*widgetID)
	{

	    case SAVE_BUTTON:
		saveButton = widget;
		break;

	    case INSERT_BUTTON:
		insertButton = widget;
		break;

	    case DELETE_BUTTON:
		deleteButton = widget;
		break;

	    case PRINT_BUTTON:
		printButton = widget;
		break;

	    case NEXT_BUTTON:
		nextButton = widget;
		break;

	    case PREVIOUS_BUTTON:
		previousButton = widget;
		break;

	    case GET_CMPNTS_BUTTON:
		getCmpntsButton = widget;
		break;

	    case CLEAR_BUTTON:
		clearButton = widget;
		break;

	    case CSL_MNGR_BUTTON:
		cslMngrButton = widget;
		break;

	    case LINEUP_NAME_ENTRY:
		lineupNameEntry = widget;
		break;

	    case METHOD_ENTRY:
		methodEntry = widget;
		break;

	    case SRCS_ENTRY:
		srcsEntry = widget;
		break;

	    case DESTS_ENTRY:
		destsEntry = widget;
		break;

	    case SETUP_TIME_ENTRY:
		setupTimeEntry = widget;
		break;

	    case SHTDWN_TIME_ENTRY:
		shtdwnTimeEntry = widget;
		break;

	    case FLOW_RATE_ENTRY:
		flowRateEntry = widget;
		break;

	    case MSRMNT_VOLS_ENTRY:
		msrmntVolsEntry = widget;
		break;

	    case XREF_VOLS_ENTRY:
		xrefVolsEntry = widget;
		break;

	    case REL_DSPLY_ENTRY:
		relDsplyEntry = widget;
		break;

	    case MAN_CSL_SCRIPT_ENTRY:
		manCslScriptEntry = widget;
		break;

	    case START_SCRIPT_ENTRY:
		startScriptEntry = widget;
		break;

	    case STOP_SCRIPT_ENTRY:
		stopScriptEntry = widget;
		break;

	    case PAUSE_SCRIPT_ENTRY:
		pauseScriptEntry = widget;
		break;

	    case RESUME_SCRIPT_ENTRY:
		resumeScriptEntry = widget;
		break;

	    case MONITOR_SCRIPT_ENTRY:
		monitorScriptEntry = widget;
		break;

	    case CMPNTS_LIST:
		cmpntsList = widget;
		XtUnmanageChild (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST_LABEL));
		XtUnmanageChild (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_SEPARATOR));
		XtUnmanageChild (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_SELECTION_LABEL));
		XtUnmanageChild (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_TEXT));
		XtUnmanageChild (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_HELP_BUTTON));
		XtManageChild (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_APPLY_BUTTON));
		n = 0;
		XtSetArg (args[n], XmNsensitive, False); n++;
		XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_CANCEL_BUTTON), args, n);
		XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_APPLY_BUTTON), args, n);
		XtSetOneValue(XmSelectionBoxGetChild (cmpntsList, XmDIALOG_OK_BUTTON),
				XmNsensitive, OperPriv(&codeBaseDesc, "LINE_DEF_COMP"));
		n = 0;
		XtSetArg (args[n], XmNselectionPolicy, XmSINGLE_SELECT); n++;
		XtSetArg (args[n], XmNhighlightThickness, 0); n++;
		XtSetArg (args[n], XmNlistSpacing, 2); n++;
		XtSetArg (args[n], XmNlistSizePolicy, XmCONSTANT); n++;
		XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST), args, n);
		XtAddCallback (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST),
				XmNsingleSelectionCallback, CmpntListCallback, NULL);
		break;

	    case CMPNTS_DIALOG:
		cmpntsDialog = widget;
		break;

	    case CMPNTS_ENTRY:
		cmpntEntry = widget;
		break;

	    case STATE_ENTRY:
		stateEntry = widget;
		break;

	    case TANK_LIST:
		tankList = widget;
		XtUnmanageChild (XmSelectionBoxGetChild (tankList, XmDIALOG_LIST_LABEL));
		XtUnmanageChild (XmSelectionBoxGetChild (tankList, XmDIALOG_SEPARATOR));
		XtUnmanageChild (XmSelectionBoxGetChild (tankList, XmDIALOG_SELECTION_LABEL));
		XtUnmanageChild (XmSelectionBoxGetChild (tankList, XmDIALOG_TEXT));
		XtUnmanageChild (XmSelectionBoxGetChild (tankList, XmDIALOG_OK_BUTTON));
		XtUnmanageChild (XmSelectionBoxGetChild (tankList, XmDIALOG_CANCEL_BUTTON));
		XtUnmanageChild (XmSelectionBoxGetChild (tankList, XmDIALOG_APPLY_BUTTON));
		XtUnmanageChild (XmSelectionBoxGetChild (tankList, XmDIALOG_HELP_BUTTON));
		n = 0;
		XtSetArg (args[n], XmNselectionPolicy, XmSINGLE_SELECT); n++;
		XtSetArg (args[n], XmNhighlightThickness, 0); n++;
		XtSetArg (args[n], XmNlistSpacing, 2); n++;
		XtSetArg (args[n], XmNlistSizePolicy, XmCONSTANT); n++;
		XtSetValues (XmSelectionBoxGetChild (tankList, XmDIALOG_LIST), args, n);
		XtAddCallback (XmSelectionBoxGetChild (tankList, XmDIALOG_LIST),
				XmNdefaultActionCallback, TankListCallback, NULL);
		n = 0;
		XtSetArg (args[n], XmNtitle, "Tanks"); n++;
		XtSetValues (XtParent (tankList), args, n);
		break;
	}
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      ButtonCallback                                                       *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget descriptor                                *
 *                                                                           *
 *          widgetID        widget ID                                        *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback for button presses                                      *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
ButtonCallback (widget, widgetID, widgetData)
Widget 			widget;
int			*widgetID;
XmAnyCallbackStruct	*widgetData;
{
	char			txtBuf[BUFSIZ];
	char			*cmpntPtr;
	char			*statePtr;
	int			errorFlag;
	int			i, j;
	int			n;
	Component		*oldCmpnts;
	Arg			args[MAX_ARGS];
	FILE			*printFileDesc;
#define	PRINT_FILE_TEMPLATE	"/tmp/lineupXXXXXX"
	char			printFileName[sizeof PRINT_FILE_TEMPLATE];

	setOperPriv();
	focusedEntry = 0;
	XtUnmanageChild (tankList);
	switch (*widgetID)
	{

	    case INSERT_BUTTON:
		if (changeFlag)
		{
		    WarningMessage ("Abandon changes?", InsertLineup);
		}
		else
		{
		    InsertLineup ();
		}
		break;

	    case DELETE_BUTTON:
		if ( !insertFlag && !changeFlag)
		{
		    WarningMessage ("Are you sure you want to delete?", DeleteLineup);
		}
		else
		{
		    ErrorMessage ("No current record");
		}
		break;

	    case PRINT_BUTTON:
		if ( !(printFileDesc = fopen (mktemp(strcpy(printFileName, PRINT_FILE_TEMPLATE)), "w")))
		{
		    sprintf (txtBuf, "Could not create temporary file: %s", printFileName);
		    ErrorMessage (txtBuf);
		}
		else
		{
		    fprintf (printFileDesc, "\r\nLineup Name\t\t%s\r\n", lineupRecord.lineupName);
		    fprintf (printFileDesc, "Mothod\t\t%s\r\n", lineupRecord.method);
		    fprintf (printFileDesc, "Sources:\r");
		    for (i=0; i < MAX_SRCS && lineupRecord.srcs[i][0]; i++)
		    {
			fprintf (printFileDesc, "\t\t\t%s\r\n", lineupRecord.srcs[i]);
		    }
		    fprintf (printFileDesc, "Destinations:\r");
		    for (i=0; i < MAX_DESTS && lineupRecord.dests[i][0]; i++)
		    {
			fprintf (printFileDesc, "\t\t\t%s\r\n", lineupRecord.dests[i]);
		    }
		    fprintf (printFileDesc, "Setup Time\t\t%s\r\n", lineupRecord.setupTime);
		    fprintf (printFileDesc, "Shutdown Time\t\t%s\r\n", lineupRecord.shtdwnTime);
		    fprintf (printFileDesc, "Flow Rate\t\t%s\r\n", lineupRecord.flowRate);
		    fprintf (printFileDesc, "Measurement Volumes:\r");
		    for (i=0; i < MAX_MSRMNT_VOLS && lineupRecord.msrmntVols[i][0]; i++)
		    {
			fprintf (printFileDesc, "\t\t\t%s\r\n", lineupRecord.msrmntVols[i]);
		    }
		    fprintf (printFileDesc, "Cross Reference Volumes:\r");
		    for (i=0; i < MAX_XREF_VOLS && lineupRecord.xrefVols[i][0]; i++)
		    {
			fprintf (printFileDesc, "\t\t\t%s\r\n", lineupRecord.xrefVols[i]);
		    }
		    fprintf (printFileDesc, "Related Display Name\t%s\r\n", lineupRecord.relDsply);
		    fprintf (printFileDesc, "Manual CSL Script\t%s\r\n", lineupRecord.manCslScript);
		    fprintf (printFileDesc, "Start Script\t\t%s\r\n", lineupRecord.startScript);
		    fprintf (printFileDesc, "Stop Script\t\t%s\r\n", lineupRecord.stopScript);
		    fprintf (printFileDesc, "Pause Script\t\t%s\r\n", lineupRecord.pauseScript);
		    fprintf (printFileDesc, "Resume Script\t\t%s\r\n", lineupRecord.resumeScript);
		    fprintf (printFileDesc, "Monitor Script\t\t%s\r\n", lineupRecord.monitorScript);
		    fprintf (printFileDesc, "Components:\r");
		    for (i=0; i < numCmpnts; i++)
		    {
			fprintf (printFileDesc, "\t\t\t%s %s\r\n", lineupRecord.cmpnts[i].entry,
					lineupRecord.cmpnts[i].state);
		    }
		    fclose (printFileDesc);
		    sprintf (txtBuf, "lpr -p -r -T\"Lineup Swings\" %s", printFileName);
		    system (txtBuf);
		}
		break;

	    case SAVE_BUTTON:
		if (GetDisplay ())
		{
		    if (lineupRecord.lineupName[0] == 0)
		    {
			ErrorMessage ("Must fill in \"Lineup Name\" field");
		    }
		    else if (lineupRecord.method[0] == 0)
		    {
			ErrorMessage ("Must fill in \"Method\" field");
		    }
		    else if (lineupRecord.srcs[0][0] == 0)
		    {
			ErrorMessage ("Must fill in \"Sources\" field");
		    }
		    else if (lineupRecord.dests[0][0] == 0)
		    {
			ErrorMessage ("Must fill in \"Destinations\" field");
		    }
		    else if (lineupRecord.msrmntVols[0][0] == 0)
		    {
			ErrorMessage ("Must fill in \"Measurement Reference Volumes\" field");
		    }
		    else
		    {
			if (d4seek (lineupDesc, lineupRecord.lineupName) == 0 && d4recno (lineupDesc) != lineupRecordNum)
			{
			    ErrorMessage ("Duplicate \"Lineup Name\" field");
			}
			else
			{
			    errorFlag = FALSE;
			    for (i=0; i < MAX_SRCS && lineupRecord.srcs[i][0] && !errorFlag; i++)
			    {
				for (j=i+1; j <= MAX_SRCS && lineupRecord.srcs[j][0] && !errorFlag; j++)
				{
				    if (strcmp (lineupRecord.srcs[i], lineupRecord.srcs[j]) == 0)
				    {
					errorFlag = TRUE;
					ErrorMessage ("Duplicate \"Source\"");
				    }
				}
			    }
			    if ( !errorFlag)
			    {
				for (i=0; i < MAX_DESTS && lineupRecord.dests[i][0] && !errorFlag; i++)
				{
				    for (j=i+1; j < MAX_DESTS && lineupRecord.dests[j][0] && !errorFlag; j++)
				    {
					if (strcmp (lineupRecord.dests[i], lineupRecord.dests[j]) == 0)
					{
					    errorFlag = TRUE;
					    ErrorMessage ("Duplicate \"Destination\"");
					}
				    }
				}
				if ( !errorFlag)
				{
				    for (i=0; i < MAX_MSRMNT_VOLS && lineupRecord.msrmntVols[i][0] && !errorFlag; i++)
				    {
					for (j=i+1; j < MAX_MSRMNT_VOLS && lineupRecord.msrmntVols[j][0] && !errorFlag; j++)
					{
					    if (strcmp (lineupRecord.msrmntVols[i], lineupRecord.msrmntVols[j]) == 0)
					    {
						errorFlag = TRUE;
						ErrorMessage ("Duplicate \"Measurement Volume\"");
					    }
					}
				    }
				    if ( !errorFlag)
				    {
					for (i=0; i < MAX_XREF_VOLS && lineupRecord.xrefVols[i][0] && !errorFlag; i++)
					{
					    for (j=i+1; j < MAX_XREF_VOLS && lineupRecord.xrefVols[j][0] && !errorFlag; j++)
					    {
						if (strcmp (lineupRecord.xrefVols[i], lineupRecord.xrefVols[j]) == 0)
						{
						    errorFlag = TRUE;
						    ErrorMessage ("Duplicate \"Cross Reference Volume\"");
						}
					    }
					}
					if ( !errorFlag)
					{
					    if (insertFlag)
					    {
						d4append_blank (lineupDesc);
						lineupRecordNum = d4recno (lineupDesc);
					    }
					    else
					    {
						d4go (lineupDesc, lineupRecordNum);
					    }
					    SaveCurrentRecord ();
					    d4tag_select (lineupDesc, d4tag (lineupDesc, lineupTagInfo[0].name));
					    d4go (lineupDesc, lineupRecordNum);
					    SetNextPrev ();
					    d4tag_select (lineupDesc, NULL);
					    d4unlock_all (lineupDesc);
					    insertFlag = FALSE;
					    changeFlag = FALSE;
					    n = 0;
					    XtSetArg (args[n], XmNsensitive, OperPriv(&codeBaseDesc, "LINE_DEF_DELT")); n++;
					    XtSetValues (deleteButton, args, n);
					    n = 0;
					    XtSetArg (args[n], XmNsensitive, OperPriv(&codeBaseDesc, "LINE_DEF_PRNT")); n++;
					    XtSetValues (printButton, args, n);
					}
				    }
				}
			    }
			}
		    }
		}
		break;

	    case ADD_CMPNT_BUTTON:
		selectedCmpnt = 0;
		addFlag = TRUE;
		n = 0;
		XtSetArg (args[n], XmNsensitive, False); n++;
		XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_CANCEL_BUTTON), args, n);
		XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_APPLY_BUTTON), args, n);
		XtSetOneValue(XmSelectionBoxGetChild (cmpntsList, XmDIALOG_OK_BUTTON),
				XmNsensitive, OperPriv(&codeBaseDesc, "LINE_DEF_COMP"));
		XmListDeselectAllItems (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST));
		XmTextSetString (cmpntEntry, "");
		XmTextSetString (stateEntry, "");
		XtManageChild (cmpntsDialog);
		setOperPriv();
		break;

	    case MODIFY_CMPNT_BUTTON:
		addFlag = FALSE;
		n = 0;
		XtSetArg (args[n], XmNsensitive, False); n++;
		XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_CANCEL_BUTTON), args, n);
		XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_APPLY_BUTTON), args, n);
		XtSetOneValue(XmSelectionBoxGetChild (cmpntsList, XmDIALOG_OK_BUTTON),
				XmNsensitive, OperPriv(&codeBaseDesc, "LINE_DEF_COMP"));
		XmListDeselectAllItems (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST));
		XmTextSetString (cmpntEntry, lineupRecord.cmpnts[selectedCmpnt-1].entry);
		XmTextSetString (stateEntry, lineupRecord.cmpnts[selectedCmpnt-1].state);
		XtManageChild (cmpntsDialog);
		setOperPriv();
		break;

	    case DELETE_CMPNT_BUTTON:
		WarningMessage ("Are you sure you want to delete?", DeleteCmpnt);
		break;

	    case OK_CMPNT_BUTTON:
		cmpntPtr = StripWhitespace(XmTextGetString(cmpntEntry));
		statePtr = StripWhitespace(XmTextGetString(stateEntry));
		if (*cmpntPtr)
			{
		    for (i=0; i < numCmpnts; i++)
				if (strcmp (lineupRecord.cmpnts[i].entry, cmpntPtr) == 0)
				    break;
		    if ( addFlag && i < numCmpnts)
				ErrorMessage ("Duplicate \"Component\"");
		    else
				{
				char	buf[BUFSIZ];
				
				if (!addFlag)
					DeleteCmpnt ();
				
				if (numCmpnts >= maxCmpnts)
					{
					oldCmpnts = lineupRecord.cmpnts;
			    	maxCmpnts += CMPNTS_BLOCK_SIZE;
					if ( !(lineupRecord.cmpnts =
							(Component *) calloc (CMPNTS_BLOCK_SIZE * maxCmpnts,
							sizeof (Component))))
						printf ("Out of memory\n"), exit (1);

					memcpy ((char *) lineupRecord.cmpnts, (char *) oldCmpnts,
						(maxCmpnts - CMPNTS_BLOCK_SIZE) * sizeof (Component));
					cfree ((char *) oldCmpnts);
					}
				for (i=1; i <= numCmpnts; i++)
				   	if (strcmp (cmpntPtr, lineupRecord.cmpnts[i-1].entry) < 0)
						break;
				for (j=numCmpnts; j >= i; j--)
					   lineupRecord.cmpnts[j] = lineupRecord.cmpnts[j - 1];

				strcpy (lineupRecord.cmpnts[i - 1].entry, cmpntPtr);
				strcpy (lineupRecord.cmpnts[i - 1].state, *statePtr ? statePtr : "reserved");
				sprintf(buf, "%*s %s", CHANNEL_NAME_ENTRY_SIZE, lineupRecord.cmpnts[i - 1].entry,
						lineupRecord.cmpnts[i - 1].state);
				lineupRecord.cmpnts[i - 1].string = XmStringCreate (buf, XmSTRING_DEFAULT_CHARSET);
				XmListAddItem (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST),
						lineupRecord.cmpnts[i - 1].string, i);
				XmListSelectPos (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST), i, False);
				XmListSetBottomPos (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST), i);
				selectedCmpnt = 0;
				n = 0;
				XtSetArg (args[n], XmNsensitive, False); n++;
				XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_CANCEL_BUTTON), args, n);
				XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_APPLY_BUTTON), args, n);
				XtSetOneValue(XmSelectionBoxGetChild (cmpntsList, XmDIALOG_OK_BUTTON),
					XmNsensitive, OperPriv(&codeBaseDesc, "LINE_DEF_COMP"));
				XmListDeselectAllItems (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST));
				numCmpnts++;
				changeFlag = TRUE;
		    	}
			}
		XtFree (cmpntPtr);
		XtUnmanageChild (cmpntsDialog);
		break;

	    case CANCEL_CMPNT_BUTTON:
		XtUnmanageChild (cmpntsDialog);
		break;
		
	    case CLEAR_BUTTON:
		XmTextSetString (lineupNameEntry, "");
		XmTextSetString (methodEntry, "");
		XmTextSetString (srcsEntry, "");
		XmTextSetString (destsEntry, "");
		XmTextSetString (setupTimeEntry, "");
		XmTextSetString (shtdwnTimeEntry, "");
		XmTextSetString (flowRateEntry, "");
		XmTextSetString (msrmntVolsEntry, "");
		XmTextSetString (xrefVolsEntry, "");
		XmTextSetString (relDsplyEntry, "");
		XmTextSetString (manCslScriptEntry, "");
		XmTextSetString (startScriptEntry, "");
		XmTextSetString (stopScriptEntry, "");
		XmTextSetString (pauseScriptEntry, "");
		XmTextSetString (resumeScriptEntry, "");
		XmTextSetString (monitorScriptEntry, "");
		for (i=0; i < numCmpnts; i++)
		{
		    XmStringFree (lineupRecord.cmpnts[i].string);
		}
		numCmpnts = 0;
		n = 0;
		XtSetArg (args[n], XmNsensitive, False); n++;
		XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_CANCEL_BUTTON), args, n);
		XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_APPLY_BUTTON), args, n);
		XtSetOneValue(XmSelectionBoxGetChild (cmpntsList, XmDIALOG_OK_BUTTON),
				XmNsensitive, OperPriv(&codeBaseDesc, "LINE_DEF_COMP"));
		selectedCmpnt = 0;
		XmListDeselectAllItems (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST));
		XmListDeleteAllItems (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST));
		InsertLineup ();
		break;

	    case GET_CMPNTS_BUTTON:
		getComponents();
		break;

	    case CSL_MNGR_BUTTON:
		break;

	    case NEXT_BUTTON:
	    case PREVIOUS_BUTTON:
		if (changeFlag)
		{
		    WarningMessage ("Abandon changes?", (*widgetID == NEXT_BUTTON) ? NextLineup : PreviousLineup);
		}
		else
		{
		    (*widgetID == NEXT_BUTTON) ? NextLineup() : PreviousLineup();
		}
		break;

	    case EXIT_BUTTON:
		if (changeFlag)
		{
		    WarningMessage ("Exit without saving changes?", exit);
		}
		else
		{
		    exit (0);
		}
		break;
	}
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      ChangeCallback                                                       *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget descriptor                                *
 *                                                                           *
 *          widgetID        widget ID                                        *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback for any entry modification                              *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
ChangeCallback (widget, widgetID, widgetData)
Widget 			widget;
int			*widgetID;
XmAnyCallbackStruct	*widgetData;
{
	int		n;
	Arg		args[MAX_ARGS];

	changeFlag = TRUE;
	n = 0;
	XtSetArg (args[n], XmNsensitive, False); n++;
	XtSetValues (deleteButton, args, n);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      CmpntListCallback                                                    *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget descriptor                                *
 *                                                                           *
 *          widgetID        widget ID                                        *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback for the component list click                            *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
CmpntListCallback (widget, widgetID, widgetData)
Widget 			widget;
int			*widgetID;
XmAnyCallbackStruct	*widgetData;
{
	char		txtBuf[BUFSIZ];
	int		*positionList;
	int		positionCount;
	Arg		args[1];

	focusedEntry = 0;
	XtUnmanageChild (tankList);
	XtSetArg (args[0], XmNsensitive, False);
	selectedCmpnt = 0;
	if (XmListGetSelectedPos (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST),
			&positionList, &positionCount) && selectedCmpnt != positionList[0])
	{
	    XtSetArg (args[0], XmNsensitive, OperPriv(&codeBaseDesc, "LINE_DEF_COMP"));
	    selectedCmpnt = positionList[0];
	    XtFree ((char *)positionList);
	}
	XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_CANCEL_BUTTON), args, 1);
	XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_APPLY_BUTTON), args, 1);
	XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_OK_BUTTON), args, 1);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      TankListCallback                                                     *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget descriptor                                *
 *                                                                           *
 *          widgetID        widget ID                                        *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback for the tank list double click                          *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
TankListCallback (widget, widgetID, widgetData)
Widget 			widget;
int			*widgetID;
XmAnyCallbackStruct	*widgetData;
{
	char		txtBuf[BUFSIZ];
	char		*strEntryPtr;
	int		*positionList;
	int		positionCount;

	if (XmListGetSelectedPos (XmSelectionBoxGetChild (tankList, XmDIALOG_LIST), &positionList, &positionCount))
	{
	    XtFree ((char *)positionList);
	    txtBuf[0] = 0;
	    if (focusedEntry == SRCS_ENTRY)
	    {
		strcpy (txtBuf, strEntryPtr = StripWhitespace (XmTextGetString (srcsEntry)));
	    }
	    else if (focusedEntry == DESTS_ENTRY)
	    {
		strcpy (txtBuf, strEntryPtr = StripWhitespace (XmTextGetString (destsEntry)));
	    }
	    XtFree (strEntryPtr);
	    strcat (txtBuf, " ");
	    strcat (txtBuf, strEntryPtr =
			StripWhitespace (XmTextGetString (XmSelectionBoxGetChild (tankList, XmDIALOG_TEXT))));
	    XtFree (strEntryPtr);
	    if (focusedEntry == SRCS_ENTRY)
	    {
		XmTextSetString (srcsEntry, txtBuf);
		XmTextSetInsertionPosition (srcsEntry, strlen (txtBuf));
	    }
	    else if (focusedEntry == DESTS_ENTRY)
	    {
		XmTextSetString (destsEntry, txtBuf);
		XmTextSetInsertionPosition (destsEntry, strlen (txtBuf));
	    }
	    XmListDeselectAllItems (XmSelectionBoxGetChild (tankList, XmDIALOG_LIST));
	}
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      FocusCallback                                                        *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget descriptor                                *
 *                                                                           *
 *          widgetID        widget ID                                        *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback for the tank list double click                          *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
FocusCallback (widget, widgetID, widgetData)
Widget 			widget;
int			*widgetID;
XmAnyCallbackStruct	*widgetData;
{
	int			n;
	Arg			args[MAX_ARGS];
	Position		entryX, entryY;
	Position		listX, listY;
	Dimension		entryWidth;
	static int		firstTime = TRUE;

	if (firstTime && (*widgetID == SRCS_ENTRY || *widgetID == DESTS_ENTRY))
	{
	    XtManageChild (tankList);
	    firstTime = FALSE;
	}
	n = 0;
	XtSetArg (args[n], XmNx, &entryX); n++;
	XtSetArg (args[n], XmNy, &entryY); n++;
	XtSetArg (args[n], XmNwidth, &entryWidth); n++;
	switch (*widgetID)
	{

	    case SRCS_ENTRY:
		if (focusedEntry != SRCS_ENTRY)
		{
		    focusedEntry = SRCS_ENTRY;
		    XtGetValues (srcsEntry, args, n);
		    n = 0;
		    XtSetArg (args[n], XmNx, entryX + entryWidth); n++;
		    XtSetArg (args[n], XmNy, entryY); n++;
		    XtSetValues (tankList, args, n);
		    XtManageChild (tankList);
		}
		break;

	    case DESTS_ENTRY:
		if (focusedEntry != DESTS_ENTRY)
		{
		    focusedEntry = DESTS_ENTRY;
		    XtGetValues (destsEntry, args, n);
		    n = 0;
		    XtSetArg (args[n], XmNx, entryX + entryWidth); n++;
		    XtSetArg (args[n], XmNy, entryY); n++;
		    XtSetValues (tankList, args, n);
		    XtManageChild (tankList);
		}
		break;

	    default:
		if (focusedEntry)
		{
		    focusedEntry = 0;
		    XtUnmanageChild (tankList);
		}
		break;

	}
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      NextLineup                                                           *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          position to next lineup                                          *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
NextLineup ()
{
	d4tag_select (lineupDesc, d4tag (lineupDesc, lineupTagInfo[0].name));
	if (lineupRecordNum)
	{
	    d4go (lineupDesc, lineupRecordNum);
	}
	if (d4skip (lineupDesc, 1) == 0)
	{
	    ChangeLineup ();
	}
	d4tag_select (lineupDesc, NULL);
	d4unlock_all (lineupDesc);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      PreviousLineup                                                       *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          position to previous lineup                                      *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
PreviousLineup ()
{
	d4tag_select (lineupDesc, d4tag (lineupDesc, lineupTagInfo[0].name));
	if (lineupRecordNum)
	{
	    d4go (lineupDesc, lineupRecordNum);
	}
	if (d4skip (lineupDesc, -1) == 0)
	{
	    ChangeLineup ();
	}
	d4tag_select (lineupDesc, NULL);
	d4unlock_all (lineupDesc);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      ChangeLineup                                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          change lineup display                                            *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
ChangeLineup ()
{
	int			i;
	int			n;
	Arg			args[MAX_ARGS];

	for (i=0; i < numCmpnts; i++)
	{
	    XmStringFree (lineupRecord.cmpnts[i].string);
	}
	numCmpnts = 0;
	n = 0;
	XtSetArg (args[n], XmNsensitive, False); n++;
	XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_CANCEL_BUTTON), args, n);
	XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_APPLY_BUTTON), args, n);
	XtSetOneValue(XmSelectionBoxGetChild (cmpntsList, XmDIALOG_OK_BUTTON),
			XmNsensitive, OperPriv(&codeBaseDesc, "LINE_DEF_COMP"));
	selectedCmpnt = 0;
	XmListDeselectAllItems (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST));
	XmListDeleteAllItems (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST));
	lineupRecordNum = d4recno (lineupDesc);
	LoadCurrentRecord ();
	DisplayCurrentRecord ();
	insertFlag = FALSE;
	changeFlag = FALSE;
	n = 0;
	XtSetArg (args[n], XmNeditable, False); n++;
	XtSetValues (lineupNameEntry, args, n);
	n = 0;
	XtSetArg (args[n], XmNsensitive, OperPriv(&codeBaseDesc, "LINE_DEF_DELT")); n++;
	XtSetValues (deleteButton, args, n);
	SetNextPrev ();
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      SetNextPrev                                                          *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          sensitize next/previous buttons                                  *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
SetNextPrev ()
{
	int			n;
	Arg			args[MAX_ARGS];

	n = 0;
	if (d4skip (lineupDesc, 1) == 0)
	{
	    XtSetArg (args[n], XmNsensitive, OperPriv(&codeBaseDesc, "LINE_DEF_NEXT")); n++;
	    XtSetValues (nextButton, args, n);
	}
	else
	{
	    XtSetArg (args[n], XmNsensitive, False); n++;
	    XtSetValues (nextButton, args, n);
	}
	d4skip (lineupDesc, -1);
	n = 0;
	if (d4skip (lineupDesc, -1) == 0)
	{
	    XtSetArg (args[n], XmNsensitive, OperPriv(&codeBaseDesc, "LINE_DEF_PREV")); n++;
	    XtSetValues (previousButton, args, n);
	}
	else
	{
	    XtSetArg (args[n], XmNsensitive, False); n++;
	    XtSetValues (previousButton, args, n);
	}
	d4skip (lineupDesc, 1);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      InsertLineup                                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          insert a lineup                                                  *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
InsertLineup ()
{
	int			n;
	Arg			args[MAX_ARGS];

	lineupRecordNum = 0;
	insertFlag = TRUE;
	changeFlag = TRUE;
	n = 0;
	XtSetArg (args[n], XmNeditable, True); n++;
	XtSetValues (lineupNameEntry, args, n);
/*	XmTextSetString (lineupNameEntry, "");*/
	n = 0;
	XtSetArg (args[n], XmNsensitive, False); n++;
	XtSetValues (deleteButton, args, n);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      DeleteLineup                                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          delete a lineup                                                  *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
DeleteLineup ()
{
	char		lineupName[LINEUP_NAME_ENTRY_SIZE+1];
	int		i;
	int		n;
	Arg		args[MAX_ARGS];

	d4tag_select (lineupDesc, d4tag (lineupDesc, lineupTagInfo[0].name));
	d4go (lineupDesc, lineupRecordNum);
	strcpy (lineupName, f4str (d4field (lineupDesc, "LINEUPNAME")));
	d4delete (lineupDesc);

	for (i=0; i < numCmpnts; i++)
	{
	    XmStringFree (lineupRecord.cmpnts[i].string);
	}
	numCmpnts = 0;
	n = 0;
	XtSetArg (args[n], XmNsensitive, False); n++;
	XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_CANCEL_BUTTON), args, n);
	XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_APPLY_BUTTON), args, n);
	XtSetOneValue(XmSelectionBoxGetChild (cmpntsList, XmDIALOG_OK_BUTTON),
			XmNsensitive, OperPriv(&codeBaseDesc, "LINE_DEF_COMP"));
	selectedCmpnt = 0;
	XmListDeselectAllItems (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST));
	XmListDeleteAllItems (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST));
	if (d4reccount (lineupDesc) >= 1)
	{
		d4skip (lineupDesc, -1);
	    lineupRecordNum = d4recno (lineupDesc);
	    LoadCurrentRecord ();
	    DisplayCurrentRecord ();
	    insertFlag = FALSE;
	    changeFlag = FALSE;
	    n = 0;
	    XtSetArg (args[n], XmNeditable, False); n++;
	    XtSetValues (lineupNameEntry, args, n);
	    n = 0;
	    XtSetArg (args[n], XmNsensitive, OperPriv(&codeBaseDesc, "LINE_DEF_DELT")); n++;
	    XtSetValues (deleteButton, args, n);
	    d4pack (lineupDesc);
		updateDBASEfile (lineupDesc);
	    d4seek (lineupDesc, lineupRecord.lineupName);
	    lineupRecordNum = d4recno (lineupDesc);
	}
	else
	{
	    XmTextSetString (lineupNameEntry, "");
	    XmTextSetString (methodEntry, "");
	    XmTextSetString (srcsEntry, "");
	    XmTextSetString (destsEntry, "");
	    XmTextSetString (setupTimeEntry, "");
	    XmTextSetString (shtdwnTimeEntry, "");
	    XmTextSetString (flowRateEntry, "");
	    XmTextSetString (msrmntVolsEntry, "");
	    XmTextSetString (xrefVolsEntry, "");
	    XmTextSetString (relDsplyEntry, "");
	    XmTextSetString (manCslScriptEntry, "");
	    XmTextSetString (startScriptEntry, "");
	    XmTextSetString (stopScriptEntry, "");
	    XmTextSetString (pauseScriptEntry, "");
	    XmTextSetString (resumeScriptEntry, "");
	    XmTextSetString (monitorScriptEntry, "");
	    lineupRecordNum = 0;
	    insertFlag = TRUE;
	    changeFlag = FALSE;
	    n = 0;
	    XtSetArg (args[n], XmNsensitive, False); n++;
	    XtSetValues (deleteButton, args, n);
	    XtSetValues (printButton, args, n);
	    XtSetValues (nextButton, args, n);
	    XtSetValues (previousButton, args, n);
	    d4pack (lineupDesc);
		updateDBASEfile (lineupDesc);
	}
	SetNextPrev ();
	d4unlock_all (lineupDesc);
	d4tag_select (lineupDesc, NULL);

	DeleteLineupParts (lineupName);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      DeleteLineupParts                                                    *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          lineupName          lineup name                                  *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          delete a lineup parts in other databases                         *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
DeleteLineupParts (lineupName)
char	*lineupName;
{
	d4tag_select (lineupCmpntsDesc, d4tag (lineupCmpntsDesc, lineupTagInfo[0].name));
	if (d4seek (lineupCmpntsDesc, lineupName) == 0)
	{
	    do
	    {
		d4delete (lineupCmpntsDesc);
	    }
	    while (d4skip (lineupCmpntsDesc, 1) == 0 &&
			strcmp (f4str (d4field (lineupCmpntsDesc, "LINEUPNAME")), lineupName) == 0);
	    d4pack (lineupCmpntsDesc);
	}
	d4tag_select (lineupCmpntsDesc, NULL);
	d4unlock_all (lineupCmpntsDesc);

	d4tag_select (lineupSrcsDesc, d4tag (lineupSrcsDesc, lineupTagInfo[0].name));
	if (d4seek (lineupSrcsDesc, lineupName) == 0)
	{
	    do
	    {
		d4delete (lineupSrcsDesc);
	    }
	    while (d4skip (lineupSrcsDesc, 1) == 0 &&
			strcmp (f4str (d4field (lineupSrcsDesc, "LINEUPNAME")), lineupName) == 0);
	    d4pack (lineupSrcsDesc);
	}
	d4tag_select (lineupSrcsDesc, NULL);
	d4unlock_all (lineupSrcsDesc);

	d4tag_select (lineupDestsDesc, d4tag (lineupDestsDesc, lineupTagInfo[0].name));
	if (d4seek (lineupDestsDesc, lineupName) == 0)
	{
	    do
	    {
		d4delete (lineupDestsDesc);
	    }
	    while (d4skip (lineupDestsDesc, 1) == 0 &&
			strcmp (f4str (d4field (lineupDestsDesc, "LINEUPNAME")), lineupName) == 0);
	    d4pack (lineupDestsDesc);
	}
	d4tag_select (lineupDestsDesc, NULL);
	d4unlock_all (lineupDestsDesc);

	d4tag_select (lineupMsrmntVolsDesc, d4tag (lineupMsrmntVolsDesc, lineupTagInfo[0].name));
	if (d4seek (lineupMsrmntVolsDesc, lineupName) == 0)
	{
	    do
	    {
		d4delete (lineupMsrmntVolsDesc);
	    }
	    while (d4skip (lineupMsrmntVolsDesc, 1) == 0
			&& strcmp (f4str (d4field (lineupMsrmntVolsDesc, "LINEUPNAME")), lineupName) == 0);
	    d4pack (lineupMsrmntVolsDesc);
	}
	d4tag_select (lineupMsrmntVolsDesc, NULL);
	d4unlock_all (lineupMsrmntVolsDesc);

	d4tag_select (lineupXrefVolsDesc, d4tag (lineupXrefVolsDesc, lineupTagInfo[0].name));
	if (d4seek (lineupXrefVolsDesc, lineupName) == 0)
	{
	    do
	    {
		d4delete (lineupXrefVolsDesc);
	    }
	    while (d4skip (lineupXrefVolsDesc, 1) == 0 &&
			strcmp (f4str (d4field (lineupXrefVolsDesc, "LINEUPNAME")), lineupName) == 0);
	    d4pack (lineupXrefVolsDesc);
	}
	d4tag_select (lineupXrefVolsDesc, NULL);
	d4unlock_all (lineupXrefVolsDesc);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      DeleteCmpnt                                                          *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          delete a lineup swing                                            *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
DeleteCmpnt ()
{
	int		n;
	Arg		args[MAX_ARGS];

	changeFlag = TRUE;
	XmStringFree (lineupRecord.cmpnts[selectedCmpnt-1].string);
	memcpy ((char *) &lineupRecord.cmpnts[selectedCmpnt-1], (char *) &lineupRecord.cmpnts[selectedCmpnt], 
			(numCmpnts - selectedCmpnt) * sizeof (Component));
	XmListDeletePos (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST), selectedCmpnt);
	--numCmpnts;
	n = 0;
	XtSetArg (args[n], XmNsensitive, False); n++;
	XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_CANCEL_BUTTON), args, n);
	XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_APPLY_BUTTON), args, n);
	selectedCmpnt = 0;
	XmListDeselectAllItems (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_LIST));
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      ErrorMessage                                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          msgPtr            pointer to error msg                           *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          popup error msg dialog box                                       *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
ErrorMessage (msgPtr)
char	*msgPtr;
{
	int		n;
	Arg		args[MAX_ARGS];

	if ( !errorMessage)
	{
	    printf ("%s\n", msgPtr);
	}
	else if ( !errorMessageString)
	{
	    errorMessageString = XmStringCreate (msgPtr, XmSTRING_DEFAULT_CHARSET);
	    n = 0;
	    XtSetArg (args[n], XmNmessageString, errorMessageString); n++;
	    XtSetValues (errorMessage, args, n);
	    XtManageChild (errorMessage);
	    XBell (XtDisplay (lineupWindow), 100);
	}
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      ErrorMessageCallback                                                 *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget ID to register                            *
 *                                                                           *
 *          index           index into widget array                          *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback when dismissing error msg dialog box                    *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
ErrorMessageCallback (widget, index, data)
Widget 			widget;
int			*index;
XmAnyCallbackStruct	*data;
{
	XmStringFree (errorMessageString);
	errorMessageString = NULL;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      WarningMessage                                                       *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          msgPtr            pointer to warning msg                         *
 *                                                                           *
 *          funcPtr           function to call on ok button press            *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          popup warning msg dialog box                                     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
WarningMessage (msgPtr, funcPtr)
char	*msgPtr;
void	(*funcPtr) ();
{
	int		n;
	Arg		args[MAX_ARGS];

	if ( !warningMessage)
	{
	    printf ("%s\n", msgPtr);
	}
	else if ( !warningMessageString)
	{
	    warningMessageFunc = funcPtr;
	    warningMessageString = XmStringCreate (msgPtr, XmSTRING_DEFAULT_CHARSET);
	    n = 0;
	    XtSetArg (args[n], XmNmessageString, warningMessageString); n++;
	    XtSetValues (warningMessage, args, n);
	    XtManageChild (warningMessage);
	    XBell (XtDisplay (lineupWindow), 100);
	}
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      WarningMessageOkCallback                                             *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget ID to register                            *
 *                                                                           *
 *          index           index into widget array                          *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback when oking warning msg dialog box                       *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
WarningMessageOkCallback (widget, index, data)
Widget 			widget;
int			*index;
XmAnyCallbackStruct	*data;
{
	XmStringFree (warningMessageString);
	warningMessageString = NULL;
	(*warningMessageFunc) ();
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      WarningMessageCancelCallback                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget ID to register                            *
 *                                                                           *
 *          index           index into widget array                          *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback when cancelling warning msg dialog box                  *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
WarningMessageCancelCallback (widget, index, data)
Widget 			widget;
int			*index;
XmAnyCallbackStruct	*data;
{
	XmStringFree (warningMessageString);
	warningMessageString = NULL;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      StripWhitespace                                                      *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          string      pointer to character string                          *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          pointer to first non-space character in string                   *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          strip leading and trailing whitespace from character string      *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static char *
StripWhitespace (string)
char	*string;
{
	int	i;

	for (i=strlen (string)-1; i >= 0 && isspace(string[i]); string[i--] = 0)
	    ;
	for (i=0; string[i] && isspace(string[i]); i++)
	    ;
	if (i)
	{
	    strcpy (string, &string[i]);
	}
	return (string);
}

getScriptAndArgs(script, name, args)
char	*script;
char	**name, **args;
	{
	static char	buf[100];
	char		*p;

	strcpy(*name = buf, script);
	if (*args = strchr(buf, ' '))
		{
		*(*args)++ = 0;
		if (p = strchr(*args, ' '))
			*p++ = 0;
		}
	else
		*args = "";
	}

getComponents()
	{
	char	command[BUFSIZ];
	char	*getenv(), *path;
	char	*name, *args;
	FILE	*components;
	char	*s;

	strcpy(lineupRecord.manCslScript,  s = StripWhitespace(XmTextGetString(manCslScriptEntry))),  XtFree(s);
	strcpy(lineupRecord.startScript,   s = StripWhitespace(XmTextGetString(startScriptEntry))),   XtFree(s);
	strcpy(lineupRecord.stopScript,    s = StripWhitespace(XmTextGetString(stopScriptEntry))),    XtFree(s);
	strcpy(lineupRecord.pauseScript,   s = StripWhitespace(XmTextGetString(pauseScriptEntry))),   XtFree(s);
	strcpy(lineupRecord.resumeScript,  s = StripWhitespace(XmTextGetString(resumeScriptEntry))),  XtFree(s);
	strcpy(lineupRecord.monitorScript, s = StripWhitespace(XmTextGetString(monitorScriptEntry))), XtFree(s);

	strcpy(command, "CG=");
	if (path = getenv("CSL_BIN_PATH"))
		strcat(strcat(command, path), "/");
	strcat(command, "CSLComponents;export CG;(");
	getScriptAndArgs(lineupRecord.startScript,   &name, &args), sprintf(&command[strlen(command)], "$CG - %s \"%s\";", name, args);
	getScriptAndArgs(lineupRecord.stopScript,    &name, &args), sprintf(&command[strlen(command)], "$CG - %s \"%s\";", name, args);
	getScriptAndArgs(lineupRecord.pauseScript,   &name, &args), sprintf(&command[strlen(command)], "$CG - %s \"%s\";", name, args);
	getScriptAndArgs(lineupRecord.resumeScript,  &name, &args), sprintf(&command[strlen(command)], "$CG - %s \"%s\";", name, args);
	getScriptAndArgs(lineupRecord.monitorScript, &name, &args), sprintf(&command[strlen(command)], "$CG - %s \"%s\";", name, args);
	getScriptAndArgs(lineupRecord.manCslScript,  &name, &args), sprintf(&command[strlen(command)], "$CG - %s \"%s\";", name, args);
	strcat(command, ")|sort|uniq");

	components = popen(command, "r");
	while (!feof(components))
		{
		char		buf[BUFSIZ];
		int			i, j;
		Component	*oldCmpnts;

		if (fgets(buf, sizeof buf, components))
			{
			buf[strlen(buf)-1] = 0;
			if (numCmpnts >= maxCmpnts)
				{
			    oldCmpnts = lineupRecord.cmpnts;
			    maxCmpnts += CMPNTS_BLOCK_SIZE;
			    if ( !(lineupRecord.cmpnts =
					(Component *) calloc(CMPNTS_BLOCK_SIZE * maxCmpnts, sizeof (Component))))
			    	{
					printf("Out of memory\n");
					exit(1);
			    	}
			    memcpy ((char *) lineupRecord.cmpnts, (char *) oldCmpnts,
					(maxCmpnts - CMPNTS_BLOCK_SIZE) * sizeof (Component));
			    cfree((char *) oldCmpnts);
				}
			for (i=1; i <= numCmpnts; i++)		/* find position in the list */
			    if (strcmp(buf, lineupRecord.cmpnts[i-1].entry) <= 0)
					break;
					
			if (strcmp(buf, lineupRecord.cmpnts[i-1].entry) == 0)	/* don't add it if already in the list */
				continue;

			for (j=numCmpnts; j >= i; j--)
			    lineupRecord.cmpnts[j] = lineupRecord.cmpnts[j - 1];
				
			strcpy (lineupRecord.cmpnts[i - 1].entry, buf);
			strcpy (lineupRecord.cmpnts[i - 1].state, "reserved");
			sprintf(buf, "%*s %s", CHANNEL_NAME_ENTRY_SIZE, lineupRecord.cmpnts[i - 1].entry,
					lineupRecord.cmpnts[i - 1].state);
			lineupRecord.cmpnts[i - 1].string = XmStringCreate(buf, XmSTRING_DEFAULT_CHARSET);
			XmListAddItem(XmSelectionBoxGetChild(cmpntsList, XmDIALOG_LIST),
					lineupRecord.cmpnts[i - 1].string, i);
			XmListSelectPos(XmSelectionBoxGetChild(cmpntsList, XmDIALOG_LIST), i, False);
			XmListSetBottomPos(XmSelectionBoxGetChild(cmpntsList, XmDIALOG_LIST), i);
			selectedCmpnt = 0;
			{
			Arg args[1];
			int	n;
			
			n = 0;
			XtSetArg(args[n], XmNsensitive, False); n++;
			XtSetValues(XmSelectionBoxGetChild(cmpntsList, XmDIALOG_CANCEL_BUTTON), args, n);
			XtSetValues (XmSelectionBoxGetChild (cmpntsList, XmDIALOG_APPLY_BUTTON), args, n);
			XtSetOneValue(XmSelectionBoxGetChild (cmpntsList, XmDIALOG_OK_BUTTON),
					XmNsensitive, OperPriv(&codeBaseDesc, "LINE_DEF_COMP"));
			}
			XmListDeselectAllItems(XmSelectionBoxGetChild(cmpntsList, XmDIALOG_LIST));
			numCmpnts++;
			changeFlag = TRUE;
			}
		}
	pclose(components);
	}

struct privData
	{
	Widget	*widget;
	char	*groupName;
	}
privData[] =
	{
	&clearButton,		"LINE_DEF_CLER",
	&saveButton,		"LINE_DEF_SAVE",
	&insertButton,		"LINE_DEF_INST",
/*	&deleteButton,		"LINE_DEF_DELT", */
/*	&printButton,		"LINE_DEF_PRNT", */
/*	&nextButton,		"LINE_DEF_NEXT", */
/*	&previousButton,	"LINE_DEF_PREV", */
	&getCmpntsButton,	"LINE_DEF_GCMP",
	&cslMngrButton,		"LINE_DEF_CSLM",
	&lineupNameEntry,	"LINE_DEF_NAME",
	&methodEntry,		"LINE_DEF_METH",
	&srcsEntry,		"LINE_DEF_SOUR",
	&destsEntry,		"LINE_DEF_DEST",
	&setupTimeEntry,	"LINE_DEF_SETU",
	&shtdwnTimeEntry,	"LINE_DEF_SHUT",
	&flowRateEntry,		"LINE_DEF_FLOW",
	&msrmntVolsEntry,	"LINE_DEF_MEAS",
	&xrefVolsEntry,		"LINE_DEF_CREF",
	&relDsplyEntry,		"LINE_DEF_RDIS",
	&manCslScriptEntry,	"LINE_DEF_MCSL",
	&startScriptEntry,	"LINE_DEF_SCSL",
	&stopScriptEntry,	"LINE_DEF_STOP",
	&pauseScriptEntry,	"LINE_DEF_PCSL",
	&resumeScriptEntry,	"LINE_DEF_RCSL",
	&monitorScriptEntry,	"LINE_DEF_MONT",
	&cmpntEntry,		"LINE_DEF_CMPE",
	&stateEntry,		"LINE_DEF_STAT",
	0,			0
	};

static void setOperPriv()
	{
	struct privData	*p;
	
	for (p = &privData[0]; p->widget; p++)
		XtSetOneValue(*p->widget, XmNsensitive, OperPriv(&codeBaseDesc, p->groupName));
	}

/*----------*/
updateDBASEfile (data)

    D4DATA	*data;

    {

	d4flush_all (data);
#ifndef CB51
	data->file_lock = TRUE;		/* to force d4update_header() to write the num_recs field */
	d4update_header (data, 1, 1);
	data->file_lock = FALSE;	/* to force d4unlock_all() to write the num_recs field */
#endif
	d4unlock_all (data);

    }

