/*****************************************************************************
 *                                                                           *
 *      name:   lineupsw.c                                                   *
 *      date:   april 12, 1993                                               *
 *                                                                           *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>

#include <Xm/Xm.h>
#include <Xm/MessageB.h>
#include <Xm/Text.h>
#include <Xm/List.h>
#include <Xm/MwmUtil.h>

#include <Mrm/MrmPublic.h>

#include <d4all.h>

#include <TFdBASE.h>
#include <lineupsw.h>

extern	char	*mktemp ();


/*****************************************************************************
 *                                                                           *
 *      static functions                                                     *
 *                                                                           *
 *****************************************************************************/

static void		OpenDatabase ();
static void		RegisterWidgets ();
static void		ButtonCallback ();
static void		ListCallback ();
static void		DeleteLineup ();
static void		ErrorMessage ();
static void		ErrorMessageCallback ();
static void		WarningMessage ();
static void		WarningMessageOkCallback ();
static void		WarningMessageCancelCallback ();
static void		PopupConfirmMessage ();
static char		*StripWhitespace ();
static void		setOperPriv();

/*****************************************************************************
 *                                                                           *
 *      static variables                                                     *
 *                                                                           *
 *****************************************************************************/

static XtAppContext	applContext;
static Display		*display;
static Widget		appShell;
static Widget		lineupWindow;
static MrmType		lineupWindowClass;
static MrmHierarchy	mrmHierarchy;
static String		uidFilenames[] =
{
	"lineupsw.uid"
};
#define	NUM_UID_FILES	(sizeof (uidFilenames) / sizeof (uidFilenames[0]))

static MrmRegisterArg	namesRegister[] =
{
	{ (String) "RegisterWidgets",		(caddr_t) RegisterWidgets },
	{ (String) "ButtonCallback",		(caddr_t) ButtonCallback },
	{ (String) "ListCallback",		(caddr_t) ListCallback }
};
#define	NUM_NAME_REGISTER	(sizeof (namesRegister) / sizeof (namesRegister[0]))

static int		insertFlag;
static int		changeFlag;
	
static int		maxLineups;
static int		numLineups;
static int		selectedLineup;
static LineupSwingRecord	*lineups;

static XmString		nullString;
static XmString		insertLineupString;
static XmString		changeLineupString;

static XmString		errorMessageString;
static Widget		errorMessage;

static void		(*warningMessageFunc) ();
static XmString		warningMessageString;
static Widget		warningMessage;

static Widget		lineupList;
static Widget		insertButton;
static Widget		saveButton;
static Widget		deleteButton;
static Widget		printButton;
static Widget		editLineupDialog;
static Widget		fromLineupEntry;
static Widget		toLineupEntry;
static Widget		swingTimeEntry;
static Widget		cslScriptEntry;

static C4CODE		codeBaseDesc;
static D4DATA		*lineupSwingsDesc;

/*
 * since setting/getting ONE widget resource happens so often:
 */
#define	XtSetOneValue(widget, resource, value) \
	do \
		{ \
		Arg		arg[1];	\
		XtSetArg(arg[0],(resource),(value)); \
		XtSetValues((widget), arg, 1); \
		} \
	while (0)

/*****************************************************************************
 *                                                                           *
 *      main prgram for lineup interface                                     *
 *                                                                           *
 *****************************************************************************/

main (argc, argv)
int	argc;
char	*argv[];
{
	char			txtBuf[BUFSIZ];
	int			n, Stat, Cmd;
	Arg			args[MAX_ARGS];
	LineupSwingRecord	*oldLineups;

	/*
	 * initialize variables
	 */
	changeFlag = FALSE;
	insertFlag = FALSE;
	selectedLineup = 0;
	errorMessage = NULL;
	errorMessageString = NULL;
	warningMessage = NULL;
	warningMessageString = NULL;
	nullString = XmStringCreate ("", XmSTRING_DEFAULT_CHARSET);
	insertLineupString = XmStringCreate ("Insert Lineup", XmSTRING_DEFAULT_CHARSET);
	changeLineupString = XmStringCreate ("Change Lineup", XmSTRING_DEFAULT_CHARSET);

	/*
	 * initialize and open display
	 */
	OpenDatabase ();
	MrmInitialize ();
	XtToolkitInitialize ();
	applContext = XtCreateApplicationContext ();
	display = XtOpenDisplay (applContext, NULL, "lineupsw", "Lineupsw", NULL, 0, &argc, argv);
	if ( !display)
	{
	    printf ("Cannot open display\n");
	    exit (1);
	}
	appShell = XtAppCreateShell ("Lineup Swings Specification", "Lineupsw",
			applicationShellWidgetClass, display, NULL, 0);

	/*
	 * create main window widget and its children
	 */
	if (MrmOpenHierarchy (NUM_UID_FILES, uidFilenames, NULL, &mrmHierarchy) != MrmSUCCESS)
	{
	    printf ("Cannot open UID file %s\n", uidFilenames[0]);
	    exit (1);
	}
	MrmRegisterNames (namesRegister, NUM_NAME_REGISTER);
	if (MrmFetchWidget (mrmHierarchy, "lineupWindow", appShell, &lineupWindow, &lineupWindowClass) != MrmSUCCESS)
	{
	    printf ("Cannot fetch lineup window widget\n");
	    exit (1);
	}

	/*
	 * create error msg dialog box
	 */
	n = 0;
	XtSetArg (args[n], XmNmessageAlignment, XmALIGNMENT_CENTER); n++;
	XtSetArg (args[n], XmNdialogType, XmDIALOG_ERROR); n++;
	XtSetArg (args[n], XmNdeleteResponse, XmDO_NOTHING); n++;
	XtSetArg (args[n], XmNdialogStyle, XmDIALOG_APPLICATION_MODAL); n++;
	XtSetArg (args[n], XmNtitle, "Error"); n++;
	errorMessage = XmCreateMessageDialog (lineupWindow, "Error Message", args, n);
	XtUnmanageChild (XmMessageBoxGetChild (errorMessage, XmDIALOG_HELP_BUTTON));
	XtUnmanageChild (XmMessageBoxGetChild (errorMessage, XmDIALOG_CANCEL_BUTTON));
	XtAddCallback (XmMessageBoxGetChild (errorMessage, XmDIALOG_OK_BUTTON),
			XmNactivateCallback, ErrorMessageCallback, NULL);

	/*
	 * create warning msg dialog box
	 */
	n = 0;
	XtSetArg (args[n], XmNmessageAlignment, XmALIGNMENT_CENTER); n++;
	XtSetArg (args[n], XmNdialogType, XmDIALOG_WARNING); n++;
	XtSetArg (args[n], XmNdeleteResponse, XmDO_NOTHING); n++;
	XtSetArg (args[n], XmNdialogStyle, XmDIALOG_APPLICATION_MODAL); n++;
	XtSetArg (args[n], XmNtitle, "Warning"); n++;
	warningMessage = XmCreateQuestionDialog (lineupWindow, "Warning Message", args, n);
	XtUnmanageChild (XmMessageBoxGetChild (warningMessage, XmDIALOG_HELP_BUTTON));
	XtAddCallback (XmMessageBoxGetChild (warningMessage, XmDIALOG_CANCEL_BUTTON),
			XmNactivateCallback, WarningMessageCancelCallback, NULL);
	n = 0;
	XtSetArg (args[n], XmNlabelString, XmStringCreate ("NO", XmSTRING_DEFAULT_CHARSET)); n++;
	XtSetValues (XmMessageBoxGetChild (warningMessage, XmDIALOG_CANCEL_BUTTON), args, n);
	XtAddCallback (XmMessageBoxGetChild (warningMessage, XmDIALOG_OK_BUTTON),
			XmNactivateCallback, WarningMessageOkCallback, NULL);
	n = 0;
	XtSetArg (args[n], XmNlabelString, XmStringCreate ("YES", XmSTRING_DEFAULT_CHARSET)); n++;
	XtSetValues (XmMessageBoxGetChild (warningMessage, XmDIALOG_OK_BUTTON), args, n);

	/*
	 * initialize lineup strings
	 */
	if ( !(lineups = (LineupSwingRecord *) calloc (LINEUP_BLOCK_SIZE, sizeof (LineupSwingRecord))))
	{
	    printf ("Out of memory\n");
	    exit (1);
	}
	maxLineups = LINEUP_BLOCK_SIZE;
	numLineups = 0;

	/*
	 * open and read database
	 */
	d4tag_select (lineupSwingsDesc, d4tag (lineupSwingsDesc, lineupSwingsTagsInfo[0].name));
	if (d4top (lineupSwingsDesc) == 0)
	{
	    do
	    {
		if (numLineups >= maxLineups)
		{
		    oldLineups = lineups;
		    maxLineups += LINEUP_BLOCK_SIZE;
		    if ( !(lineups =
			(LineupSwingRecord *) calloc (LINEUP_BLOCK_SIZE * maxLineups, sizeof (LineupSwingRecord))))
		    {
			printf ("Out of memory\n");
			exit (1);
		    }
		    memcpy ((char *) lineups, (char *) oldLineups, (maxLineups - LINEUP_BLOCK_SIZE) * sizeof (XmString));
		    cfree ((char *) oldLineups);
		}
		memset ((char *) &lineups[numLineups], 0, sizeof (LineupSwingRecord));
		strncpy (lineups[numLineups].fromLineup, f4str (d4field (lineupSwingsDesc,
				"FROMLINEUP")), FROM_LINEUP_SIZE);
		StripWhitespace (lineups[numLineups].fromLineup);
		strncpy (lineups[numLineups].toLineup, f4str (d4field (lineupSwingsDesc, "TOLINEUP")), TO_LINEUP_SIZE);
		StripWhitespace (lineups[numLineups].toLineup);
		strncpy (lineups[numLineups].swingTime, f4str (d4field (lineupSwingsDesc,
				"SWINGTIME")), TIME_STRING_ENTRY_SIZE);
		StripWhitespace (lineups[numLineups].swingTime);
		strncpy (lineups[numLineups].cslScript, f4str (d4field (lineupSwingsDesc,
				"CSLSCRIPT")), CSL_SCRIPT_ENTRY_SIZE);
		StripWhitespace (lineups[numLineups].cslScript);
		sprintf (txtBuf, "%-*s     %-*s     %-*s     %-*s",
			 FROM_LINEUP_SIZE, lineups[numLineups].fromLineup,
			 TO_LINEUP_SIZE, lineups[numLineups].toLineup,
			 TIME_STRING_ENTRY_SIZE, lineups[numLineups].swingTime,
			 CSL_SCRIPT_ENTRY_SIZE, lineups[numLineups].cslScript);
		lineups[numLineups].lineupStr = XmStringCreate (txtBuf, XmSTRING_DEFAULT_CHARSET);
		XmListAddItem (lineupList, lineups[numLineups].lineupStr, numLineups + 1);
		numLineups++;
	    }
	    while (d4skip (lineupSwingsDesc, 1) == 0);
	}
	d4tag_select (lineupSwingsDesc, NULL);
	d4unlock_all (lineupSwingsDesc);

	n = 0;
	XtSetArg (args[n], XmNsensitive, False); n++;
	XtSetValues (deleteButton, args, n);
	if (numLineups == 0)
	{
	    XtSetValues (printButton, args, n);
	}

	/*
	 * realize main window
	 */
	XtManageChild (lineupWindow);
	XtRealizeWidget (appShell);
	setOperPriv();

	/*
	 * get and dispatch events
	 */
	XtAppMainLoop (applContext);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      OpenDatabase                                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          open lineup database file                                        *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
OpenDatabase ()
{
	struct stat	statBuf;

	d4init (&codeBaseDesc);
	codeBaseDesc.read_lock = 0;
	codeBaseDesc.create_error = 0;
	codeBaseDesc.field_name_error = 0;
	codeBaseDesc.go_error = 0;
	codeBaseDesc.open_error = 0;
	codeBaseDesc.skip_error = 0;
	codeBaseDesc.tag_name_error = 0;
	
	if ( !(lineupSwingsDesc = d4open (&codeBaseDesc, lineupSwingsName)))
	    printf ("Cannot open lineup swings database file \"%s\"", lineupSwingsName), exit (0);
	d4unlock_all (lineupSwingsDesc);
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      RegisterWidgets                                                      *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget descriptor                                *
 *                                                                           *
 *          widgetID        widget ID                                        *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          register a widget that was created from the UIL.                 *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
RegisterWidgets (widget, widgetID, widgetData)
Widget 			widget;
int			*widgetID;
XmAnyCallbackStruct	*widgetData;
{
	int		n;
	Arg		args[MAX_ARGS];

	switch (*widgetID)
	{
	    case LINEUP_COLUMNS:
		n = 0;
		XtSetArg (args[n], XmNlabelString, XmStringCreate ("From Lineup                                                      To Lineup                                                        Swing Time     CSL Swing Script", XmSTRING_DEFAULT_CHARSET)); n++;
		XtSetValues (widget, args, n);
		break;

	    case LINEUP_LIST:
		lineupList = widget;
		break;

	    case SAVE_BUTTON:
		saveButton = widget;
		break;

	    case INSERT_BUTTON:
		insertButton = widget;
		break;

	    case DELETE_BUTTON:
		deleteButton = widget;
		break;

	    case PRINT_BUTTON:
		printButton = widget;
		break;

	    case EDIT_LINEUP_DIALOG:
		editLineupDialog = widget;
		break;

	    case FROM_LINEUP_ENTRY:
		fromLineupEntry = widget;
		break;

	    case TO_LINEUP_ENTRY:
		toLineupEntry = widget;
		break;

	    case SWING_TIME_ENTRY:
		swingTimeEntry = widget;
		break;

	    case CSL_SCRIPT_ENTRY:
		cslScriptEntry = widget;
		break;
	}
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      ButtonCallback                                                       *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget descriptor                                *
 *                                                                           *
 *          widgetID        widget ID                                        *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback for button presses                                      *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
ButtonCallback (widget, widgetID, widgetData)
Widget 			widget;
int			*widgetID;
XmAnyCallbackStruct	*widgetData;
{
	char			txtBuf[BUFSIZ];
	char			*fromLineupPtr;
	char			*toLineupPtr;
	char			*swingTimePtr;
	char			*cslScriptPtr;
	int			compareStatus;
	int			i;
	int			n;
	Arg			args[MAX_ARGS];
	LineupSwingRecord	*oldLineups;
	FILE			*printFileDesc;
	static char		*printFileName = "/tmp/tagsXXXXXX";
	static char		*printFileTemp = "/tmp/tagsXXXXXX";

	setOperPriv();
	switch (*widgetID)
	{

	    case INSERT_BUTTON:
		insertFlag = TRUE;
		XmListDeselectAllItems (lineupList);
		selectedLineup = 0;
		n = 0;
		XtSetArg (args[n], XmNdialogTitle, insertLineupString); n++;
		XtSetValues (editLineupDialog, args, n);
		XmTextSetMaxLength (fromLineupEntry, FROM_LINEUP_SIZE);
		XmTextSetMaxLength (toLineupEntry, TO_LINEUP_SIZE);
		XmTextSetMaxLength (swingTimeEntry, TIME_STRING_ENTRY_SIZE);
		XmTextSetMaxLength (cslScriptEntry, CSL_SCRIPT_ENTRY_SIZE);
		XtManageChild (editLineupDialog);
		break;

	    case DELETE_BUTTON:
		WarningMessage ("Are you sure you want to delete?", DeleteLineup);
		break;

	    case PRINT_BUTTON:
		strcpy (printFileName, printFileTemp);
		if ( !mktemp (printFileName) || !(printFileDesc = fopen (printFileName, "w")))
		{
		    sprintf (txtBuf, "Could not create temporary file: %s", printFileName);
		    ErrorMessage (txtBuf);
		}
		else
		{
		    fprintf (printFileDesc, "\r\nFrom Lineup                                                      To Lineup                                                        Swing Time\r\n");
		    for (i=0; i < numLineups; i++)
		    {
			fprintf (printFileDesc, "%-*s     %-*s     %-*s\r\n",
				 FROM_LINEUP_SIZE, lineups[i].fromLineup,
				 TO_LINEUP_SIZE, lineups[i].toLineup,
				 TIME_STRING_ENTRY_SIZE, lineups[i].swingTime);
			fprintf (printFileDesc, "%-*s\r\n\r\n", CSL_SCRIPT_ENTRY_SIZE, lineups[i].cslScript);
		    }
		    fclose (printFileDesc);
		    sprintf (txtBuf, "lpr -p -r -T\"Lineup Swings\" %s", printFileName);
		    system (txtBuf);
		}
		break;

	    case SAVE_BUTTON:
		changeFlag = FALSE;
		for (i=0; i < numLineups; i++)
		{
			char	searchString[FROM_LINEUP_SIZE+TO_LINEUP_SIZE+1];
			
			sprintf(searchString, "%-*s%-*s", FROM_LINEUP_SIZE, lineups[i].fromLineup,
					TO_LINEUP_SIZE, lineups[i].toLineup);
			searchString[I4MAX_KEY_SIZE] = 0;
			d4tag_select(lineupSwingsDesc, d4tag(lineupSwingsDesc, "LINEUPTAG"));
			if (d4seek(lineupSwingsDesc, searchString) != 0)
			    d4append_blank (lineupSwingsDesc);
		    f4assign (d4field (lineupSwingsDesc, "FROMLINEUP"), lineups[i].fromLineup);
		    f4assign (d4field (lineupSwingsDesc, "TOLINEUP"), lineups[i].toLineup);
		    f4assign (d4field (lineupSwingsDesc, "SWINGTIME"), lineups[i].swingTime);
		    f4assign (d4field (lineupSwingsDesc, "CSLSCRIPT"), lineups[i].cslScript);
		}
		updateDBASEfile (lineupSwingsDesc);
		break;

	    case EXIT_BUTTON:
		if (changeFlag)
		{
		    WarningMessage ("Exit without saving changes?", exit);
		}
		else
		{
		    exit (0);
		}
		break;

	    case OK_EDIT_LINEUP_BUTTON:
		sprintf (txtBuf, "%-*s     ", FROM_LINEUP_SIZE,
				fromLineupPtr = StripWhitespace (XmTextGetString (fromLineupEntry)));
		sprintf (&txtBuf[strlen (txtBuf)], "%-*s     ", TO_LINEUP_SIZE,
				toLineupPtr = StripWhitespace (XmTextGetString (toLineupEntry)));
		sprintf (&txtBuf[strlen (txtBuf)], "%-*s     ", TIME_STRING_ENTRY_SIZE,
				swingTimePtr = StripWhitespace (XmTextGetString (swingTimeEntry)));
		sprintf (&txtBuf[strlen (txtBuf)], "%-*s", CSL_SCRIPT_ENTRY_SIZE,
				cslScriptPtr = StripWhitespace (XmTextGetString (cslScriptEntry)));
		if (*fromLineupPtr && *toLineupPtr && *cslScriptPtr)
		{
		    for (i=1; i <= numLineups; i++)
		    {
			if (selectedLineup != i)
			{
			    if (strcmp (fromLineupPtr, lineups[i-1].fromLineup) == 0 &&
					strcmp (toLineupPtr, lineups[i-1].toLineup) == 0)
			    {
				break;
			    }
			}
		    }
		    if (i <= numLineups)
		    {
			ErrorMessage ("Duplicate entries not allowed");
			XtFree (fromLineupPtr);
			XtFree (toLineupPtr);
			XtFree (swingTimePtr);
			XtFree (cslScriptPtr);
			return;
		    }
		    if ( !insertFlag)
		    {
			DeleteLineup ();
		    }
		    changeFlag = TRUE;
		    if (numLineups >= maxLineups)
		    {
			oldLineups = lineups;
			maxLineups += LINEUP_BLOCK_SIZE;
			if ( !(lineups = (LineupSwingRecord *) calloc (LINEUP_BLOCK_SIZE * maxLineups,
					sizeof (LineupSwingRecord))))
			{
			    printf ("Out of memory\n");
			    exit (1);
			}
			memcpy ((char *) lineups, (char *) oldLineups, (maxLineups - LINEUP_BLOCK_SIZE) * sizeof (XmString));
			cfree ((char *) oldLineups);
		    }
		    for (selectedLineup=1; selectedLineup <= numLineups; selectedLineup++)
		    {
			if ((compareStatus = strcmp (fromLineupPtr, lineups[selectedLineup-1].fromLineup)) < 0)
			{
			    break;
			}
			else if (compareStatus == 0)
			{
			    if (strcmp (toLineupPtr, lineups[selectedLineup-1].toLineup) < 0)
			    {
				break;
			    }
			}
		    }
		    for (i=numLineups; i >= selectedLineup; i--)
		    {
			lineups[i] = lineups[i - 1];
		    }
		    lineups[selectedLineup - 1].lineupStr = XmStringCreate (txtBuf, XmSTRING_DEFAULT_CHARSET);
		    strcpy (lineups[selectedLineup - 1].fromLineup, fromLineupPtr);
		    strcpy (lineups[selectedLineup - 1].toLineup, toLineupPtr);
		    strcpy (lineups[selectedLineup - 1].swingTime, swingTimePtr);
		    strcpy (lineups[selectedLineup - 1].cslScript, cslScriptPtr);
		    XmListAddItem (lineupList, lineups[selectedLineup - 1].lineupStr, selectedLineup);
		    XmListSelectPos (lineupList, selectedLineup, False);
		    XmListSetBottomPos (lineupList, selectedLineup);
		    n = 0;
		    XtSetArg (args[n], XmNsensitive, OperPriv(&codeBaseDesc, "LINE_SWG_PRNT")); n++;
		    XtSetValues (printButton, args, n);
		    numLineups++;
		    XtUnmanageChild (editLineupDialog);
		}
		else if ( !*fromLineupPtr)
		{
		    ErrorMessage ("Must fill in \"From Lineup\" field");
		}
		else if ( !*toLineupPtr)
		{
		    ErrorMessage ("Must fill in \"To Lineup\" field");
		}
		else if ( !*cslScriptPtr)
		{
		    ErrorMessage ("Must fill in \"CSL Script\" field");
		}
		XtFree (fromLineupPtr);
		XtFree (toLineupPtr);
		XtFree (swingTimePtr);
		XtFree (cslScriptPtr);
		break;

	    case CANCEL_EDIT_LINEUP_BUTTON:
		XtUnmanageChild (editLineupDialog);
		break;
	}
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      ListCallback                                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget descriptor                                *
 *                                                                           *
 *          widgetID        widget ID                                        *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback for the list                                            *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
ListCallback (widget, widgetID, widgetData)
Widget 			widget;
int			*widgetID;
XmAnyCallbackStruct	*widgetData;
{
	char		txtBuf[BUFSIZ];
	int		*positionList;
	int		positionCount;
	int		n;
	Arg		args[MAX_ARGS];

	n = 0;
	if (XmListGetSelectedPos (lineupList, &positionList, &positionCount))
	{
	    XtSetArg (args[n], XmNsensitive, OperPriv(&codeBaseDesc, "LINE_SWG_DELT")); n++;
	    XtSetValues (deleteButton, args, n);
	    switch (*widgetID)
	    {

		case CLICK:
		    selectedLineup = positionList[0];
		    break;

		case DOUBLE_CLICK:
		    insertFlag = FALSE;
		    selectedLineup = positionList[0];
		    XmTextSetString (fromLineupEntry, lineups[selectedLineup - 1].fromLineup);
		    XmTextSetString (toLineupEntry, lineups[selectedLineup - 1].toLineup);
		    XmTextSetString (swingTimeEntry, lineups[selectedLineup - 1].swingTime);
		    XmTextSetString (cslScriptEntry, lineups[selectedLineup - 1].cslScript);
		    n = 0;
		    XtSetArg (args[n], XmNdialogTitle, changeLineupString); n++;
		    XtSetValues (editLineupDialog, args, n);
		    XtManageChild (editLineupDialog);
		    break;
	    }
	    XtFree ((char *)positionList);
	}
	else
	{
	    XtSetArg (args[n], XmNsensitive, False); n++;
	    XtSetValues (deleteButton, args, n);
	    selectedLineup = 0;
	}
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      DeleteLineup                                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          delete a lineup swing                                            *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
DeleteLineup ()
{
	int		n;
	Arg		args[MAX_ARGS];

	changeFlag = TRUE;
	XmStringFree (lineups[selectedLineup-1].lineupStr);
	memcpy ((char *) &lineups[selectedLineup-1], (char *) &lineups[selectedLineup],
				(numLineups - selectedLineup) * sizeof (LineupSwingRecord));
	XmListDeletePos (lineupList, selectedLineup);
	numLineups--;
	n = 0;
	XtSetArg (args[n], XmNsensitive, False); n++;
	XtSetValues (deleteButton, args, n);
	selectedLineup = 0;
	if (numLineups == 0)
	{
	    n = 0;
	    XtSetArg (args[n], XmNsensitive, False); n++;
	    XtSetValues (printButton, args, n);
	}
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      ErrorMessage                                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          msgPtr            pointer to error msg                           *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          popup error msg dialog box                                       *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
ErrorMessage (msgPtr)
char	*msgPtr;
{
	int		n;
	Arg		args[MAX_ARGS];

	if ( !errorMessage)
	{
	    printf ("%s\n", msgPtr);
	}
	else if ( !errorMessageString)
	{
	    errorMessageString = XmStringCreate (msgPtr, XmSTRING_DEFAULT_CHARSET);
	    n = 0;
	    XtSetArg (args[n], XmNmessageString, errorMessageString); n++;
	    XtSetValues (errorMessage, args, n);
	    XtManageChild (errorMessage);
	    XBell (XtDisplay (lineupWindow), 100);
	}
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      ErrorMessageCallback                                                 *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget ID to register                            *
 *                                                                           *
 *          index           index into widget array                          *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback when dismissing error msg dialog box                    *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
ErrorMessageCallback (widget, index, data)
Widget 			widget;
int			*index;
XmAnyCallbackStruct	*data;
{
	XmStringFree (errorMessageString);
	errorMessageString = NULL;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      WarningMessage                                                       *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          msgPtr            pointer to warning msg                         *
 *                                                                           *
 *          funcPtr           function to call on ok button press            *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          popup warning msg dialog box                                     *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
WarningMessage (msgPtr, funcPtr)
char	*msgPtr;
void	(*funcPtr) ();
{
	int		n;
	Arg		args[MAX_ARGS];

	if ( !warningMessage)
	{
	    printf ("%s\n", msgPtr);
	}
	else if ( !warningMessageString)
	{
	    warningMessageFunc = funcPtr;
	    warningMessageString = XmStringCreate (msgPtr, XmSTRING_DEFAULT_CHARSET);
	    n = 0;
	    XtSetArg (args[n], XmNmessageString, warningMessageString); n++;
	    XtSetValues (warningMessage, args, n);
	    XtManageChild (warningMessage);
	    XBell (XtDisplay (lineupWindow), 100);
	}
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      WarningMessageOkCallback                                             *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget ID to register                            *
 *                                                                           *
 *          index           index into widget array                          *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback when oking warning msg dialog box                       *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
WarningMessageOkCallback (widget, index, data)
Widget 			widget;
int			*index;
XmAnyCallbackStruct	*data;
{
	XmStringFree (warningMessageString);
	warningMessageString = NULL;
	(*warningMessageFunc) ();
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      WarningMessageCancelCallback                                         *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          widget          widget ID to register                            *
 *                                                                           *
 *          index           index into widget array                          *
 *                                                                           *
 *          widgetData      widget data                                      *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          none                                                             *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          callback when cancelling warning msg dialog box                  *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static void
WarningMessageCancelCallback (widget, index, data)
Widget 			widget;
int			*index;
XmAnyCallbackStruct	*data;
{
	XmStringFree (warningMessageString);
	warningMessageString = NULL;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *      StripWhitespace                                                      *
 *                                                                           *
 *                                                                           *
 *      Inputs                                                               *
 *                                                                           *
 *          string      pointer to character string                          *
 *                                                                           *
 *                                                                           *
 *      Return Value                                                         *
 *                                                                           *
 *          pointer to first non-space character in string                   *
 *                                                                           *
 *                                                                           *
 *      Description                                                          *
 *                                                                           *
 *          strip leading and trailing whitespace from character string      *
 *                                                                           *
 *                                                                           *
 *****************************************************************************/

static char *
StripWhitespace (string)
char	*string;
{
	int	i;

	for (i=strlen (string)-1; i >= 0 && isspace(string[i]); string[i--] = 0)
	    ;
	for (i=0; string[i] && isspace(string[i]); i++)
	    ;
	if (i)
	{
	    strcpy (string, &string[i]);
	}
	return (string);
}

struct privData
	{
	Widget	*widget;
	char	*groupName;
	}
privData[] =
	{
	&saveButton,		"LINE_SWG_SAVE",
	&insertButton,		"LINE_SWG_INST",
	&deleteButton,		"LINE_SWG_DELT",
	&printButton,		"LINE_SWG_PRNT",
	&fromLineupEntry,	"LINE_SWG_FROM",
	&toLineupEntry,		"LINE_SWG_TOUP",
	&swingTimeEntry,	"LINE_SWG_SWGN",
	&cslScriptEntry,	"LINE_SWG_SCSL",
	0,			0
	};

static void setOperPriv()
	{
	struct privData	*p;
	
	for (p = &privData[0]; p->widget; p++)
		XtSetOneValue(*p->widget, XmNsensitive, OperPriv(&codeBaseDesc, p->groupName));
	}


/*----------*/
updateDBASEfile (data)

    D4DATA	*data;

    {

	d4flush_all (data);
#ifndef CB51
	data->file_lock = TRUE;		/* to force d4update_header() to write the num_recs field */
	d4update_header (data, 1, 1);
	data->file_lock = FALSE;	/* to force d4unlock_all() to write the num_recs field */
#endif
	d4unlock_all (data);

    }

