/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/****************************************************************/
/*								*/
/*		HANDLE THE HARDWARE TIME OF DAY CLOCK		*/
/*								*/
/*	Author:	Gordon Uchenick					*/
/*								*/
/****************************************************************/

#include <string.h>
#include <stdio.h>

#include "m48t08.h"
#include "tsDefs.h"

char			*sysModel();

extern TS_STAMP		time_stamp;

/****************************************************************/
/*								*/
/*			READ THE TOD CLOCK			*/
/*								*/
/*	This routine would normally be called just once		*/
/*	at IOC initialization time.				*/
/*								*/
/****************************************************************/

void readTodClock(void)
{
    char	*model;
    TS_STAMP	newTime;

    /*
    See if we are on a known set of hardware and handle it
    accordingly.
    */

    model = sysModel();
    if ((strcmp(model, "Motorola MVME162") == 0) || (strcmp(model, "Motorola MVM162LX") == 0))
	{
	M48T08		*todChip = (M48T08 *) 0xfffc1ff8;
	M48T08		timeOfDay;
	static int	month;
	static int	day;
	static int	year;
	static int	hours;
	static int	minutes;
	static int	seconds;
	char		timeString[100];
	char		*cptr = timeString;

	/*
	Any variant of the MVME162. The TOD clock is a SGS-THOMSON M48T08.
	The TOD clock bytes appear at 0xfffc1ff8.
	*/
	
	/*
	Check the stop bit. If this is set, then the TOD clock hasn't been programmed
	yet.
	*/

	if (todChip -> seconds & M48T08_SEC_STOP)
	    return;
	/*
	The first step is to set the READ bit in the control word. That will stop the
	clock from updating the structure. Then copy the structure out of the chip, then
	clear the READ bit to let it go.
	*/

	todChip -> control |= M48T08_CTL_READ;
	timeOfDay = *todChip;
	todChip -> control &= ~M48T08_CTL_READ;

	/*
	Now convert this into an Epics TOD stamp.
	*/

	month = ((timeOfDay.month & M48T08_MNTH_10MON_MASK) >> M48T08_MNTH_10MON_SHFT) * 10
		+ ((timeOfDay.month & M48T08_MNTH_01MON_MASK) >> M48T08_MNTH_01MON_SHFT);
	day = ((timeOfDay.date & M48T08_DATE_10DAY_MASK) >> M48T08_DATE_10DAY_SHFT) * 10
		+ ((timeOfDay.date & M48T08_DATE_01DAY_MASK) >> M48T08_DATE_01DAY_SHFT);
	year = ((timeOfDay.year & M48T08_YEAR_10YR_MASK) >> M48T08_YEAR_10YR_SHFT) * 10
		+ ((timeOfDay.year & M48T08_YEAR_01YR_MASK) >> M48T08_YEAR_01YR_SHFT);
	hours = ((timeOfDay.hour & M48T08_HOUR_10HR_MASK) >> M48T08_HOUR_10HR_SHFT) * 10
		+ ((timeOfDay.hour & M48T08_HOUR_01HR_MASK) >> M48T08_HOUR_01HR_SHFT);
	minutes = ((timeOfDay.minutes & M48T08_MIN_10MIN_MASK) >> M48T08_MIN_10MIN_SHFT) * 10
		+ ((timeOfDay.minutes & M48T08_MIN_01MIN_MASK) >> M48T08_MIN_01MIN_SHFT);
	seconds = ((timeOfDay.seconds & M48T08_SEC_10SEC_MASK) >> M48T08_SEC_10SEC_SHFT) * 10
		+ ((timeOfDay.seconds & M48T08_SEC_01SEC_MASK) >> M48T08_SEC_01SEC_SHFT);
	sprintf(timeString, "%d/%d/%d %d:%d:%d", month, day, year, hours, minutes, seconds);

	if (tsTextToStamp(&newTime, &cptr) == S_ts_OK)
	    time_stamp = newTime;
	}
}

/****************************************************************/
/*								*/
/*			SET TOD CLOCK				*/
/*								*/
/****************************************************************/


void setTodClock(void)
{
    char		*model;
    TS_STAMP		currTime;
    struct tsDetail	detail;

    /*
    See if we are on a know set of hardware and handle it accordingly.
    */

    model = sysModel();
    if ((strcmp(model, "Motorola MVME162") == 0) || (strcmp(model, "Motorola MVM162LX") == 0))
	{
	M48T08		*todChip = (M48T08 *) 0xfffc1ff8;
	int		monthReg;
	int		dateReg;
	int		dayReg;
	int		yearReg;
	int		hourReg;
	int		minuteReg;
	int		secondReg;

	/*
	Get the time of day to set. Bust it up into its components.
	*/

	currTime = time_stamp;
	tsStampToLocal(currTime, &detail);

	detail.monthNum++;	/* Make January = 1 */
	monthReg = ((detail.monthNum / 10) << M48T08_MNTH_10MON_SHFT) + ((detail.monthNum % 10) << M48T08_MNTH_01MON_SHFT);
	detail.dayMonth++;
	dateReg = ((detail.dayMonth / 10) << M48T08_DATE_10DAY_SHFT) + ((detail.dayMonth % 10) << M48T08_DATE_01DAY_SHFT);
	detail.dayOfWeek++;
	dayReg = detail.dayOfWeek << M48T08_DAY_SHFT;
	if (detail.year >= 2000)
	    detail.year -= 2000;
	else
	    detail.year -= 1900;
	yearReg = ((detail.year / 10) << M48T08_YEAR_10YR_SHFT) + ((detail.year % 10) << M48T08_YEAR_01YR_SHFT);
	hourReg = ((detail.hours / 10) << M48T08_HOUR_10HR_SHFT) + ((detail.hours % 10) << M48T08_HOUR_01HR_SHFT);
	minuteReg = ((detail.minutes / 10) << M48T08_MIN_10MIN_SHFT) + ((detail.minutes % 10) << M48T08_MIN_01MIN_SHFT);
	secondReg = ((detail.seconds / 10) << M48T08_SEC_10SEC_SHFT) + ((detail.seconds % 10) << M48T08_SEC_01SEC_SHFT);

	/*
	Set the Write bit, set the registers, and then clear the bit.
	*/

	todChip -> control |= M48T08_CTL_WRITE;
	todChip -> seconds = secondReg;
	todChip -> minutes = minuteReg;
	todChip -> hour = hourReg;
	todChip -> weekday = dayReg;
	todChip -> date = dateReg;
	todChip -> month = monthReg;
	todChip -> year = yearReg;
	todChip -> control &= ~M48T08_CTL_WRITE;
	}
}


