#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vmenuHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstrHEADER
#include vlistviewHEADER
#include vrulerHEADER
#include vfontHEADER
#include vimageHEADER
#include vcontrolHEADER
#include vnumHEADER

#include "tisperm.h"
#include "tisPermApp.hxx"
#include "tisPermDatabase.hxx"
#include "tisPermMain.hxx"
#include "tisPermOptions.hxx"
#include "tisPermAccess.hxx"
#include "tisPermUtils.hxx"


vkindTYPED_FULL_DEFINITION(AccessDialog, vdialog, NULL);

/*******************************************************************
*
*    AccessDialog override for constructors init ()
*
*******************************************************************/

void AccessDialog::LoadInit (vresource res)
{
    // load the dialogs vlistviews and rulers
    utilRuler  = vruler::RequestSubLoad (this, 
					tisperm_UtilRuler);
    utilListView  = UtilListView::RequestSubLoad (this, 
					tisperm_UtilListView);
    ptProtRuler = vruler::RequestSubLoad (this, 
					tisperm_ProtGrpRuler);
    ptProtListView = PtProtListView::RequestSubLoad (this, 
					tisperm_ProtGrpListView);
    privilegeLevel = PtProtPrivLevel::RequestSubLoad (this,
					tisperm_PrivilegeLevel);
    applyButton = PtProtApplyButton::RequestSubLoad (this,
					tisperm_ApplyButton);
    dismissButton = PtProtDismissButton::RequestSubLoad (this,
					tisperm_DismissButton);

    // init inherited data
    vdialog::LoadInit (res);

    utilListView -> StoreAccessDialog (this);
    ptProtListView -> StoreAccessDialog (this);
    privilegeLevel -> StoreAccessDialog (this);
    applyButton -> StoreAccessDialog (this);
    dismissButton -> StoreAccessDialog (this);

    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
}


/*******************************************************************
*
*    AccessDialog Destroy function
*
*******************************************************************/

void AccessDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }

    vdialog::Destroy ();
}


/*******************************************************************
*
*    AccessDialog Handle window resize function
*	need to resize the ruler items in the rulers
*
*******************************************************************/

int AccessDialog::HandleConfigure (vevent *event)
{
    vwindow::HandleConfigure (event);
    if (event -> GetType () == veventWINDOW_CONFIGURE && event -> IsSize ())
    {
	SetUtilRulerWidth ();
	SetPtProtRulerWidth ();
	SetPtProtCellWidth ();
    }
    return (vTRUE);
}

/*******************************************************************
*
*    AccessDialog Handle window lose of focus function
*	need to deselect any selected rows, for visual, 
*	without this can see background highlist on part of the 
*	selected row
*
*******************************************************************/

int AccessDialog::HandleLose (vevent *event)
{
    vlist *list = utilListView -> GetList ();
    vlistSelection *selection = utilListView -> GetSelection ();
    if (list != NULL)
    {
	if ((list -> IsSelectionEmpty (selection)) == vFALSE)
	{
	    int row, column;
	    list -> GetSelectedCell (selection, &row, &column);
	    list -> UnselectRow (selection, row);
	    ListEntry *entry = utilListView -> GetSelectedEntry ();
	    entry = NULL;
	}
    }
    return (vwindow::HandleLose (event));
}


/*******************************************************************
*
*    AccessDialog init the selected flag in the option tree
*    NOTE:  a permission has form MMI_AAA_BBB_CCC
*	    we want the tree node pointer which points to the
*	    last group (ex. CCC)
*
*******************************************************************/

void AccessDialog::InitSelectedFlag (void)
{
    TisPermApp *app = mainDialog -> GetPermApp ();

    char groupname [GROUPNAME_SZ + 1];
    vcharExportBounded (vcharScribeSystemExport (groupName),
			groupname,
			sizeof (groupname));
    groupname [sizeof (groupname) - 1] = 0;

    // get a list of utility permissions for the group
    TisPermDatabase *db = app -> GetDb ();
    Tisutilacc	    *utilAcc = db -> GetTisUtilAcc ();
    vliststr	    *list = utilAcc -> GetUtilPermissions (groupname);
    if (list == NULL)
    {
	return;
    }

    // now update the selected flag for each permission
    char	    utilPerm [UTILACCPRIV_SZ + 1];
    char	    *permission;
    TREENODE_PTR    treeNode;
    vlistIterator   iterator;
    vchar	    *listValue;
    OptionTree	    *optionTree = app -> GetOptionTree ();
    TREENODE_PTR    treeRoot = optionTree -> GetTreeRoot ();

    iterator.Start (list);
    while (iterator.Next ())
    {
	listValue = (vchar *)(iterator.GetValue ());
	vcharExportBounded (vcharScribeSystemExport (listValue),
			    utilPerm,
			    sizeof (utilPerm));
	utilPerm [sizeof (utilPerm) - 1] = 0;
	// start at the head of the class
	treeNode = treeRoot;
	permission = strtok (utilPerm, OPTION_DELIMETER_STR);
	while (permission != NULL)
	{
	    treeNode = optionTree -> FindNode (treeNode, permission);
	    permission = strtok (NULL, OPTION_DELIMETER_STR);
	}
	treeNode -> selected = vTRUE;
    }
    iterator.Finish ();
    delete list;
}


/*******************************************************************
*
*    AccessDialog Open function
*
*******************************************************************/

void AccessDialog::OpenAccess (MainDialog *main, vchar *groupname)
{
    mainDialog = main;
    vcharCopyBounded (groupname, groupName, sizeof (groupName));
    groupName [sizeof (groupName) - 1] = 0;

    // clear the selected flag in the options tree
    TisPermApp *app = mainDialog -> GetPermApp ();
    OptionTree *optionTree = app -> GetOptionTree ();
    TREENODE_PTR treeRoot = optionTree -> GetTreeRoot ();
    optionTree -> ClearSelected (treeRoot);

    SetDialogTitle ();
    SetUtilRulerWidth ();
    SetPtProtRulerWidth ();

    // init the selected flags in the option tree for this group
    // based on the utilacc database
    InitSelectedFlag ();

    // update all children of parents who are selected since all that
    // is stored in utilacc is a parent if it is selected
    optionTree -> SetParentsChildrenSelected (treeRoot);

    // set up Toggle array
    utilListView -> AddToggles ();
    // give toggles ptr to modified flag
    UpdateToggleModifyPtr ();
    utilModified = vFALSE;
    // insert main option into list
    utilListView -> InsertMain ();

    // build the point protection list
    ptProtListView -> BuildPtGroupsList ();
    ptgrpModified = vFALSE;
    // pass ptr to modified flag to radio button container
    privilegeLevel -> UpdatePtGrpModified (&this -> ptgrpModified);

    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    AccessDialog notify function
*
*******************************************************************/

void AccessDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    AccessDialog set dialog title function
*
*******************************************************************/

void AccessDialog::SetDialogTitle (void)
{
    TisPermApp	*app = mainDialog -> GetPermApp ();
    vresource	titlesRes = app -> GetTitlesRes ();
    
    vstr *titleFormat = vresourceGetString (titlesRes, 
					tisperm_AccessDialogTitle);
    vstr *dialogTitle = vstrCloneScribed (vcharScribeFormat (titleFormat,
							    groupName));
    SetTitle (dialogTitle);
    vstrDestroy (titleFormat);
    vstrDestroy (dialogTitle);
}


/*******************************************************************
*
*   AccessDialog resize the prProtListView"s list cell size because
*   we had a window resize event
*
*******************************************************************/

void AccessDialog::SetPtProtCellWidth (void)
{
    // get list views drawing area
    const vrect *crect = ptProtListView -> GetListRect ();
    int width = crect -> GetWidth ();
    // width / 2 only 2 columns
    ptProtListView -> SetColumnWidth (vlistviewALL_COLUMNS, width / 2);
}


/*******************************************************************
*
*    AccessDialog set util ruler width function
*
*******************************************************************/

void AccessDialog::SetUtilRulerWidth (void)
{
    const vrect *rulerRect = utilRuler -> GetRect ();
    int rulerwidth = rulerRect -> GetWidth ();
    vrulerItem  *item = utilRuler -> GetItemAt (0);
    item -> SetWidth (rulerwidth - 3);
    utilRuler -> Compact ();
}


/*******************************************************************
*
*    AccessDialog set point protection ruler width function
*
*******************************************************************/

void AccessDialog::SetPtProtRulerWidth (void)
{
    const vrect *rulerRect = ptProtRuler -> GetRect ();
    int rulerwidth = rulerRect -> GetWidth ();
    vrulerItem  *item = ptProtRuler -> GetItemAt (0);
    item -> SetWidth (rulerwidth - 3);
    utilRuler -> Compact ();
}


/*******************************************************************
*
*    AccessDialog update the utilModifed ptr in the toggles [] in 
*	UtilListView
*******************************************************************/

void AccessDialog::UpdateToggleModifyPtr (void)
{
    int i;
    for (i = 0; i < UTIL_MAX_TOGGLES; i++)
    {
	utilListView -> toggles [i] -> SetUtilModified (&this -> utilModified);
    }
}


vkindTYPED_FULL_DEFINITION(UtilListView, vlistview, NULL);

/*******************************************************************
*
*    UtilList listview class
*
*******************************************************************/

void UtilListView::LoadInit (vresource res)
{
    vlistview::LoadInit (res);

    // load in the images from the image dictionary

    vresource appRes = vapplication::GetCurrent () -> GetResources ();
    vresource imageRes = vresourceGet (appRes, tisperm_ImageDict);

    mainsClosedImage = new vimage (vresourceGet (imageRes, 
						tisperm_MainClsIm));
    mainsOpenImage = new vimage (vresourceGet (imageRes, 
						tisperm_MainOpIm));
    closedWithSubGroups = new vimage (vresourceGet (imageRes, 
						tisperm_GrpStClsIm));
    closedWithNoSubGroups = new vimage (vresourceGet (imageRes,
						tisperm_GrpEndClsIm));
    openWithGroups = new vimage (vresourceGet (imageRes,
						tisperm_GrpStOpIm));
    openWithNoGroups = new vimage (vresourceGet (imageRes,
						tisperm_GrpEndOpIm));
    continueLine = new vimage (vresourceGet (imageRes,
						tisperm_ContLineIm));
    continueLineWithGroups = new vimage (vresourceGet (imageRes,
						tisperm_ContGrpIm));
    endOfGroup = new vimage (vresourceGet (imageRes,
						tisperm_ContGrpEndIm));
    selectedEntry = NULL;
    buttonUpFlag = vFALSE;

    // init an item list for expanding/collapsing toggles in the list
    itemList = new UtilItemList;
    itemList -> SetOwner (this);

    // make an empty list and attach it to the utilListView
    list = new vlist;
    SetList (list);
    vlistSelection *lselection = GetSelection ();
    list -> SetSelectMethod (lselection, vlistONE_ONLY_SELECTION);
    list -> SetRowCount (0);
}


void UtilListView::Destroy (void)
{
    if (list != NULL)
    {
	SetList (NULL);
	delete list;
    }

    delete mainsClosedImage;
    delete mainsOpenImage;
    delete closedWithSubGroups;
    delete closedWithNoSubGroups;
    delete openWithGroups;
    delete openWithNoGroups;
    delete continueLine;
    delete continueLineWithGroups;
    delete endOfGroup;

    // delete the vdialogItemList
    delete itemList;

    vlistview::Destroy ( );
}


/*******************************************************************
*
*    UtilList listview fill toggles array
*
*******************************************************************/

void UtilListView::AddToggles (void)
{
    int i;

    for (i = 0; i < UTIL_MAX_TOGGLES; i++)
    {
	toggles [i] = new Toggle;
	toggles [i] -> SetUtilListView (this);
	itemList -> Append (toggles [i]);
    }
}


/*******************************************************************
*
*    UtilList listview close the item list when open listview
*
*******************************************************************/

void UtilListView::Close()
{
    itemList -> Close ();
    vlistview::Close ();
}


/*******************************************************************
*
*    UtilListView override of draw cell background function
*
*******************************************************************/

void UtilListView::DrawCellBackground (int row, int col, 
					     int selected, vrect *cellRect)
{
    vlist *viewList = GetList ();
    ListEntry *tempEntry = (ListEntry *)viewList -> GetCellValue (row, col);
    if (tempEntry == NULL)
    {
	return;
    }
    TREENODE_PTR node = tempEntry -> GetTreeNode ();
    if ( (selected == vFALSE) || (node == NULL) )
    {
	vlistview::DrawCellBackground (row, col, selected, cellRect);
    }
    else
    {
	vimage *image = tempEntry -> GetImage (this);
	vrect tempRect = *cellRect;
	int offset = node -> level * ENTRY_LEVEL_INSET + (image -> GetWidth ());
	Toggle *toggle = tempEntry -> GetToggle ();
	int toggleWidth = toggle -> GetWidth ();
	offset += toggleWidth + TOGGLE_OPTION_SPACE;
	vfont *font = vfont::FindScribed (vcharScribeLiteral (""));
	int stringWidth = font -> ScribedStringWidthX (vcharScribeSystem (node -> optionName));
	tempRect.SetX (tempRect.GetX() + offset);
	tempRect.SetWidth (0);
    }
}


/*******************************************************************
*
*    UtilListView override of Draw Cell contents. this is the guy
*	that does all the work 
*
*******************************************************************/

void UtilListView::DrawCellContents (int row, int col, 
				   int selected, vrect *cellRect)
{
    vlist *viewList = GetList();
    ListEntry *entry = (ListEntry *)(viewList -> GetCellValue(row, col));
    if (entry == NULL)
	return;
    TREENODE_PTR node = entry -> GetTreeNode ();
    if (node == NULL)
	return;

    vdrawGSave();
    vdrawRectsClip(cellRect, 1);

    // visible row is used as an index into UtilListView::toggles []
    int visibleRow = row - GetTopRow ();

    vrect temp = *cellRect;
    vdraw::RectClip(temp.GetX(), temp.GetY(), temp.GetWidth(), temp.GetHeight());

    vimage *image;
    int imageOffset;

    if (row > 0)
    {
	image = continueLine;
	ListEntry *tempEntry;
	int tempRow;
	int lastLevel = -1;
	int lastImage = -1;
	TREENODE_PTR tempNode;
	for (tempRow = row - 1; tempRow >= 1; tempRow--)
	{
	    tempEntry = (ListEntry *)(viewList -> GetCellValue (tempRow, 0));
	    tempNode = tempEntry -> GetTreeNode ();
	    if (tempNode -> level < node -> level)
	    {
		if (tempNode -> level >= 1)
		{
		    if ((lastLevel == tempNode -> level) && 
			((lastImage == TISPERM_PARENT) || (lastImage == TISPERM_LAST_PARENT)))
		    {
			continue;
		    }
		    if (tempNode -> imageType == TISPERM_PARENT)
		    {
			imageOffset = tempNode -> level * ENTRY_LEVEL_INSET;
			vdrawMoveTo( temp.GetX () + imageOffset,
			             temp.GetY() + 
				     (temp.GetHeight() - image->GetHeight())/2);
			vdrawImageCompositeIdent(image);
		    }
		    lastLevel = tempNode -> level;
		    lastImage = tempNode->imageType;
		}
		if (tempNode -> level == 1)
		{
		    break;
		}
	    }
	}
    }

    image = entry -> GetImage (this);

    // draw the box image
    imageOffset = node -> level * ENTRY_LEVEL_INSET;

    vdraw::MoveTo( temp.GetX () + imageOffset,
		temp.GetY() + (temp.GetHeight() - image->GetHeight())/2);
    vdraw::ImageCompositeIdent(image);

    // draw the check box
    vrect toggleRect = temp;
    entry -> SetToggle (toggles [visibleRow]);
    toggles [visibleRow] -> SetEntry (entry);
    TREENODE_PTR treeNode = entry -> GetTreeNode ();
    toggles [visibleRow] -> SetValue (treeNode -> selected);
    int offset = node -> level * ENTRY_LEVEL_INSET + image -> GetWidth();
    toggleRect.SetX (toggleRect.GetX() + offset);
    toggleRect.SetY (toggleRect.GetY() + 1);
    toggleRect.SetHeight (temp.GetHeight ());
    toggleRect.SetWidth(temp.GetHeight ());
    toggles [visibleRow] -> SetRect (&toggleRect);
    toggles [visibleRow] -> DrawDevice ();
    toggles [visibleRow] -> SetDraw ();
    toggles [visibleRow] -> Draw ();
    
    // draw the text option
    int toggleWidth = toggles [visibleRow] -> GetWidth ();
    offset = toggleRect.GetWidth () + TOGGLE_OPTION_SPACE;
//    vfont *font = vfont::FindScribed (vcharScribeLiteral (""));
    vfont *font = DetermineFont ();
    int stringWidth = font -> ScribedStringWidthX (vcharScribeSystem (node -> optionName));
    toggleRect.SetX(toggleRect.GetX() + offset);
    toggleRect.SetWidth (stringWidth);
    vstr *option;
    option = vstrCloneScribed (vcharScribeSystem (node -> optionName));
    vdraw::RectShowCenterY(option, &toggleRect);


    vdrawGRestore();

    vstrDestroy (option);
}


/*******************************************************************
*
*    UtilListView override of draw cell focus function
*	do nothing since we do not want to see any focus
*	highlighting
*
*******************************************************************/

void UtilListView::DrawCellFocus (int row, int col, vrect *cellRect)
{
}


/*******************************************************************
*
*    UtilList listview focus the itemList
*
*******************************************************************/

void UtilListView::Focus()
{
    vlistview::Focus ();
    itemList -> Focus ();
}


/*******************************************************************
*
*    UtilList listview handle a button down event
*
*******************************************************************/

int UtilListView::HandleButtonDown (vevent *event)
{
    int ret = vFALSE;

    if (GetListRect() -> ContainsPoint (event -> GetX (), event -> GetY ())) 
    {
	// let the item list handle the work of figuring out who gets it
	ret = itemList -> HandleButtonDown (event);
    }
    
    // the dialog item didn't handle the event.
    // Let the listview have a chance at it.
    if (!ret) 
    {
	ret = vlistview::HandleButtonDown (event);
    }
    
    return (ret);
}


/*******************************************************************
*
*    UtilList listview handle a button up event
*
*******************************************************************/

int UtilListView::HandleButtonUp (vevent *event)
{
    int ret = vFALSE;
    if ((event -> GetClickCount ()) > 1)
    {
	return (ret);
    }
    // see if one of our dialogItem want to handle this
    vdialogItem *item = itemList -> GetDragItem ();
    if (item != NULL) 
    {
	// see if the item wants to handle it
	ret = item -> HandleButtonUp (event);

	// clear out the value so we don't get confused on a subsequent
        // drag or mouse up event
	itemList -> SetDragItem (NULL);
    }

    // either there wasn't a tracking item, or it didn't want to handle
    // this.  We've had our fun, let the listview have a crack at it.
    
    if (!ret) 
    {
	buttonUpFlag = vFALSE;
	ret = vlistview::HandleButtonUp(event);
    }

    return (ret);
}


/*******************************************************************
*
*    UtilList listview handle a drag event
*
*******************************************************************/

int UtilListView::HandleDrag (vevent *event)
{
    int ret = vFALSE;
    // see if one of our dialogItemInAListview is tracking.
    vdialogItem *item = itemList -> GetDragItem ();

    if (item != NULL) 
    {
	// see if the item wants to handle it.
	ret = item -> HandleDrag (event);

    }

    // either there wasn't a tracking item, or it didn't want to handle
    // this.  We've had our fun, let the listview have a crack at it.
    
    if (!ret) 
    {
        ret = vlistview::HandleDrag (event);
    }
	
    return (ret);
}


/*******************************************************************
*
*    UtilListView handle a key down event
*
*******************************************************************/

int UtilListView::HandleKeyDown(vevent *event)
{
    int ret = vFALSE;

    // try to bind the key to our key map
    vkeyStroke key = event -> GetKeyStroke ();
    
    const vname *binding = LoadKeyMap (key);
    if (binding != NULL) 
    {
	event -> SetBinding (binding);
    }

    // try to dispatch the event to an item in the key dispatcher

    vdialogItem *item = (vdialogItem *)LoadKeyDispatch (key);
    if ((item != NULL) && item -> IsEnabled () && item -> HandleKeyDown (event)) 
    {
	ret = vTRUE;
    }

    if (!ret) 
    {
	item = itemList -> GetFocusItem ();
	if ((item != NULL) && item -> IsEnabled () &&
	    item -> HandleKeyDown (event))
	{
	    ret = vTRUE;
	} else 
	{
	    ret = vlistview::HandleKeyDown (event);
	}
    }

    return (ret);
}


/*******************************************************************
*
*    UtilList listview handle a key up event
*
*******************************************************************/

int UtilListView::HandleKeyUp (vevent *event)
{
    if (selectedEntry && selectedEntry -> GetToggle () &&
	itemList -> GetFocusItem() == selectedEntry -> GetToggle ())
	return (selectedEntry -> GetToggle() -> HandleKeyUp (event));

    return (vlistview::HandleKeyUp (event));
}


/**********************************************************************
*
*    UtilListView::IssueSelector ()
*	handle selector commands issued against the listview
*	-> selector -- the ID of the command being issued
*	-> context -- the context this command is issuing in.
*	Right now we're just sniffing the focus moving so we can move 
*	the focus in our owned itemlist.
*
***********************************************************************/
 
int UtilListView::IssueSelector (int selector, vdict *context)
{
    int ret;

    switch (selector) 
    {
	case vdialogITEM_SELECT_MOVE_FOCUS: 
	{
	    const int *direction =
	      (const int *)context -> GetElem (vname_Direction);
	    if (direction != NULL) 
	    {
	        itemList -> MoveFocus (*direction,
                                       context -> GetElem (vname_Text)
                                       != NULL);

		// since we're effectively a "flat" container, see if we
		// need to pop ourselves out
		
		if (itemList -> HasFocusWrapped ()) 
		{
		    // let superclass handle it
		    ret = vFALSE;
		} else 
		{
		    ret = vTRUE;
		}
	    }
	    break;
	}
    }

    if (!ret) 
    {
        vlistview::IssueSelector(selector, context);
    }

    return (ret);

}


/*******************************************************************
*
*    UtilList listview open the item list when open listview
*
*******************************************************************/

void UtilListView::Open()
{
    vlistview::Open ();
    itemList -> Open ();
}


/*******************************************************************
*
*    UtilList listview scroll the contents of the itemList
*
*******************************************************************/

void UtilListView::ScrollAbsolute(int x, int y)
{
    vlist *viewList = GetList ();
    vlistSelection *lselection = GetSelection ();
    if ((viewList -> IsSelectionEmpty (lselection)) == vFALSE)
    {
	int row, column;
	viewList -> GetSelectedCell (lselection, &row, &column);
	viewList -> UnselectRow (lselection, row);
	selectedEntry = NULL;
    }
    vlistview::ScrollAbsolute (x, y);
}


/*******************************************************************
*
*    UtilList listview observe function
*
*******************************************************************/

void UtilListView::ObserveSelect (int state)
{
    vlist *viewList = GetList ();
    vlistSelection *lselection = GetSelection ();
    int row = 0;
    int column = 0;
    viewList -> GetSelectedCell (lselection, &row, &column);
    if (row == vlistNO_CELLS_SELECTED)
    {
	return;
    }
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	ListEntry *entry = (ListEntry *)(viewList -> GetCellValue (row, 0));
	if (entry == NULL)
	{
	    return;
	}
	TREENODE_PTR node = entry -> GetTreeNode ();
	if (node == NULL)
	{
	    return;
	}
	if (node -> boxType == TISPERM_NO_GROUP_BOX)
	{
	    selectedEntry = entry;
	    return;
	}
	int currx, curry;
	GetLastPointerPosition (&currx, &curry);
	vrect tempRect;
	GetCellRect (row, 0, &tempRect);
	vimage *image = entry -> GetImage (this);
	tempRect.SetX (tempRect.GetX () + (node -> level * ENTRY_LEVEL_INSET));
	tempRect.SetWidth (image -> GetWidth ());
	if ( ((entry -> IsLocked ()) == vFALSE) &&
	     (buttonUpFlag == vFALSE) &&
	     (tempRect.ContainsPoint (currx, curry)) )
	{
	     entry -> SetLocked (vTRUE);
	     buttonUpFlag = vTRUE;
	     if ( (entry -> IsOpen ()) == TISPERM_BOX_OPENED)
	     {
		CloseEntry (row, entry);
	     }
	     else
	     {
		OpenEntry (row, entry);
	     }
	     entry -> SetLocked (vFALSE);
	     InvalRect ((const vrect *)&tempRect, vwindowINVAL_OPAQUE | vwindowINVAL_UPDATE);
	}
	selectedEntry = entry;
    }
}


/*******************************************************************
*
*    UtilList listview unfocus the itemList
*
*******************************************************************/

void UtilListView::Unfocus()
{
    itemList -> Unfocus ();
    vlistview::Unfocus ();
}



// START EXTENSIONS TO THE CLASS ***********************************
//
//
/*******************************************************************
*
*    UtilListView close an entry
*
*******************************************************************/

void UtilListView::CloseEntry (int row, ListEntry *entry)
{
    vlist *viewList = GetList ();
    TREENODE_PTR parent = entry -> GetTreeNode ();
    row++;
    while (viewList -> GetRowCount () && row < viewList -> GetRowCount ())
    {
	ListEntry *currEntry = (ListEntry *)(viewList -> GetCellValue (row, 0));
	TREENODE_PTR node = currEntry -> GetTreeNode ();
	if ( (currEntry == NULL) || (node -> parent != parent) )
	{
	    break;
	}
	if ( (currEntry -> IsOpen ()) == TISPERM_BOX_OPENED)
	{
	    CloseEntry (row, currEntry);
	}
	viewList -> RemoveRow (row);
	delete currEntry;
    }
    entry -> SetOpen (TISPERM_BOX_CLOSED);
}


/*******************************************************************
*
*    UtilListView insert MAIN group 
*
*******************************************************************/

void UtilListView::InsertMain (void)
{
    vlist *viewList = GetList();

    ListEntry *entry = new ListEntry ();

    MainDialog *mainDialog = accessDialog -> GetMainDialog ();
    TisPermApp *app = mainDialog -> GetPermApp ();
    OptionTree *optionTree = app -> GetOptionTree ();
    TREENODE_PTR treeRoot = optionTree -> GetTreeRoot ();
    entry -> SetTreeNode (treeRoot);
    if (treeRoot -> sublink == NULL)
    {
	entry -> SetOpen (TISPERM_BOX_OPENED);
    }
    else
    {
	entry -> SetOpen (TISPERM_BOX_CLOSED);
    }
    viewList -> SetCellValue (0, 0, (const void *)entry);
}


/*******************************************************************
*
*    UtilListView open an entry
*
*******************************************************************/

void UtilListView::OpenEntry (int row, ListEntry *entry)
{
    TREENODE_PTR node = entry -> GetTreeNode ();
    if (node -> sublink != NULL)
    {
	vlist *viewList = GetList ();
	TREENODE_PTR currNode = node -> sublink;
	while (currNode)
	{
	    // get new list item
	    ListEntry *newEntry = new ListEntry ();
	    // store ptr to treeNode item
	    newEntry -> SetTreeNode (currNode);
	    row++;
	    viewList -> InsertRow (row);
	    viewList -> SetCellValue (row, 0, (const void *)newEntry);
	    currNode = currNode -> fwdlink;
	}
	entry -> SetOpen (TISPERM_BOX_OPENED);
	InvalViewCell (row, 0);
    }
}


/*******************************************************************
*
*    PermItemList Class to define functions to handle the toggles
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(UtilItemList, vdialogItemList, NULL);

/**********************************************************************
*
*    UtilItemList::DetermineBounds
*	return the bounding box of the owning object in the 
*	coordinate system of its window.
*
***********************************************************************/
 
void UtilItemList::DetermineBounds (vrect *rect)
{
    UtilListView *view = (UtilListView *)GetOwner();   
    UtilItemList *itemList = (UtilItemList *)(view -> GetItemListIn ());

    if (itemList != NULL)
    {
	itemList -> DetermineBounds (rect);
    }
}

/**********************************************************************
*
*    UtilItemList::DetermineColorsFrom
*	given an owner, figure out a set of colors to use for drawing.
*	note: phlags is spelled to avoid hiding vdialogItemList::flags
*
***********************************************************************/
 
void UtilItemList::DetermineColorsFrom (long phlags, vcolor **known,
				        int *ids, int count, vcolor **colors)
{
    ((UtilListView *)GetOwner()) -> DetermineColorsFrom (phlags, known,
							  ids, count,
							  colors);
}


/**********************************************************************
*
*    UtilItemList::DetermineColorSchemeFrom
*	given an owner, figure out the color scheme
*	-> known   -- a set of colors that we know already
*	-> phlags  -- the elements that are set in known
*	-> scheme  -- the color scheme to modify to fill in the 
*		      unknown slots
*	NOTE: 'phlags' is spelled so to avoid hiding vdialogItemList::flags
*
***********************************************************************/
 
void UtilItemList::DetermineColorSchemeFrom (vcolorScheme *known, 
					      unsigned long phlags, 
					      vcolorScheme *scheme)
{
    ((UtilListView *)GetOwner ()) -> DetermineColorSchemeFrom (known,
							       phlags,
							       scheme);
}

/**********************************************************************
*
*    UtilItemList::DetermineContent
*	return the bounding box of the visible portion of the owning 
*	object in the coordinate system of its window.
*
***********************************************************************/
 
void UtilItemList::DetermineContent (vrect *rect)
{
    UtilListView *view = (UtilListView *)GetOwner ();

    *rect = *view -> GetListRect ();
}


/**********************************************************************
*
*    UtilItemList::DetermineCursor
*	given an owner, figure the cursor that should be used when 
*	the mouse pointer is over this area.
*	Just use the cursor that's set for the owning dialog item.
*
***********************************************************************/
 
vcursor *UtilItemList::DetermineCursor ()
{
    return ((UtilListView *)GetOwner ()) -> DetermineCursor ();
}

/**********************************************************************
*
*    UtilItemList::DetermineDialog
*	given an owner, figure out what dialog it's in.
*
***********************************************************************/
 
vdialog *UtilItemList::DetermineDialog ()
{
    return ((UtilListView *)GetOwner ()) -> DetermineDialog ();
}

/**********************************************************************
*
*    UtilItemList::DetermineEnabled
*	given an owner, figure out if it's currently enabled.
*	This allows for a container-like object to disable its contents 
*	when it gets disabled.
*
***********************************************************************/
 
int UtilItemList::DetermineEnabled ()
{
    return ((UtilListView *)GetOwner()) -> DetermineEnabled ();
}


/**********************************************************************
*
*    UtilItemList::DetermineFont
*	given an owner, figure out the font to use for text drawing.
*	Just use the font that'd be used for text drawing in owner.
*
***********************************************************************/
 
vfont *UtilItemList::DetermineFont ()
{
    return ((UtilListView *)GetOwner())->DetermineFont();
}

/**********************************************************************
*
*    UtilItemList::DetermineLineWidth
*	given an owner, figure out the line width to use for vdraw 
*	Just use the line width that'd be used for path stroking 
*	for the owner.
*
***********************************************************************/
 
int UtilItemList::DetermineLineWidth ()
{
    return ((UtilListView *)GetOwner ()) -> DetermineLineWidth ();
}

/**********************************************************************
*
*    UtilItemList::DetermineShown
*	given an owner, figure out if it's currently visible
*	This allows for a container-like object to hide its contents 
*	when it gets hidden.
*
***********************************************************************/
 
int UtilItemList::DetermineShown ()
{
    return ((UtilListView *)GetOwner ()) -> DetermineShown ();
} 


/**********************************************************************
*
*    UtilItemList::DrawDevice
*	the drawing device is being set for an item in
*	the item list.
*
***********************************************************************/
 
void UtilItemList::DrawDevice()
{
    UtilListView *view = (UtilListView *)(GetOwner());

    view -> DrawDevice ();

    vdraw::RectsClip (view -> GetListRect(), 1);
}

/**********************************************************************
*
*    UtilItemList::GetKeyDispatch
*	given an owner, figure out the key dispatch dict to use.
*	Just use the key dispatch dict	that'd be used for our owner.
*
***********************************************************************/
 
vdict *UtilItemList::GetKeyDispatch ()
{
    return ((UtilListView *)GetOwner ()) -> GetKeyDispatch ();
}

/**********************************************************************
*
*    UtilItemList::GetKeyMap ()
*	given an owner, figure out the key map to use.
*	Just use the key map that'd be used for our owner.
*
***********************************************************************/
 
vdict *UtilItemList::GetKeyMap ()
{
    return ((UtilListView *)GetOwner ()) -> GetKeyMap( );
}


/*******************************************************************
*
*    ListEntry Class
*
*******************************************************************/

ListEntry::ListEntry ()
{
    isOpen = TISPERM_BOX_CLOSED;
    isLocked = vFALSE;
    treeNode = NULL;
    toggle = NULL;
}

ListEntry::~ListEntry ()
{
}


/*******************************************************************
*
*    ListEntry Class get image for a row
*
*******************************************************************/

vimage *ListEntry::GetImage (UtilListView *mainList)
{
    vimage *image = NULL;
    switch (treeNode -> boxType)
    {
	case TISPERM_NO_GROUP_BOX:
	    if (treeNode -> imageType == TISPERM_CHILD)
	    {
		image = mainList -> continueLineWithGroups;
	    }
	    else
	    {
		image = mainList -> endOfGroup;
	    }
	    break;

	case TISPERM_GROUP_BOX:
	    if (isOpen == TISPERM_BOX_CLOSED)
	    {
		switch (treeNode -> imageType)
		{
		    case TISPERM_MAIN_PARENT:
			image = mainList -> mainsClosedImage;
			break;

		    case TISPERM_PARENT:
			image = mainList -> closedWithSubGroups;
			break;

		    case TISPERM_LAST_PARENT:
			image = mainList -> closedWithNoSubGroups;
			break;
		}
	    }
	    // need an open box
	    else
	    {
		switch (treeNode -> imageType)
		{
		    case TISPERM_MAIN_PARENT:
			image = mainList -> mainsOpenImage;
			break;

		    case TISPERM_PARENT:
			image = mainList -> openWithGroups;
			break;

		    case TISPERM_LAST_PARENT:
			image = mainList -> openWithNoGroups;
			break;
		}
	    }
	    break;
    }
    return (image);
}


vkindTYPED_FULL_DEFINITION(Toggle, vcontrolToggle, NULL);

/*******************************************************************
*
*   Toggle class, observe function
*   NOTE: treeNode -> selected, selected follows the current state
*	    of the toggle
*
*******************************************************************/

void Toggle::ObserveDialogItem (vevent *event)
{
    TREENODE_PTR    treeNode = listEntry -> GetTreeNode ();
    int		    lastSelected = treeNode -> selected;
    AccessDialog    *accessDialog = utilListView -> GetAccessDialog ();
    MainDialog	    *mainDialog = accessDialog -> GetMainDialog ();
    TisPermApp	    *app = mainDialog -> GetPermApp ();
    OptionTree	    *optionTree = app -> GetOptionTree ();
    int		    changed = vFALSE;

    // get new value
    treeNode -> selected = GetValue ();
    *utilModified = vTRUE;

    // the box has been checked
    if (lastSelected == vFALSE)
    {
	// set Parent if all siblings are selected
	optionTree -> SetParentSelected (treeNode, &changed);
	if (treeNode -> sublink != NULL)
	{
	    optionTree -> SetChildrenSelected (treeNode -> sublink, &changed);
	}
    }
    // the box has been unchecked
    else
    {
	optionTree -> ClearParentSelected (treeNode, &changed);
	if (treeNode -> sublink != NULL)
	{
	    optionTree -> ClearChildrenSelected (treeNode -> sublink, &changed);
	}
    }

    if (changed == vTRUE)
    {
	const vrect *listRect = utilListView -> GetListRect ();
	utilListView -> InvalRect (listRect,
			    vwindowINVAL_OPAQUE | vwindowINVAL_UPDATE);
    }
}



vkindTYPED_FULL_DEFINITION(PtProtListView, vlistview, NULL);

/*******************************************************************
*
*    PtProtListView listview class
*
*******************************************************************/
void PtProtListView::LoadInit (vresource res)
{
    vlistview::LoadInit (res);

    list = NULL;
}

void PtProtListView::Destroy (void)
{
    if (list != NULL)
    {
	SetList (NULL);
	delete list;
    }

    vlistview::Destroy ( );
}


/*******************************************************************
*
*   ProtGrpListView Build list function
*	create a list for the view
*	get a list of protection groups
*	get a list of protection groups that group belongs to
*	update the permissions column of list form the previous list
*
*******************************************************************/

void PtProtListView::BuildPtGroupsList (void)
{
    // make an empty list 
    list = new vliststr;
    list -> SetColumnCount (2);
    list -> SetRowCount (0);
    char groupname [GROUPNAME_SZ + 1];
    vchar *groupName = accessDialog -> GetGroupName ();
    vcharExportBounded (vcharScribeSystemExport (groupName),
			groupname,
			sizeof (groupname));
    groupname [sizeof (groupname) -1] = 0;

    MainDialog *mainDialog = accessDialog -> GetMainDialog ();
    TisPermApp *app = mainDialog -> GetPermApp ();
    TisPermDatabase *db = app -> GetDb ();

    Tisprotgroups *protGroups = db -> GetTisProtGroups ();
    // get a list of protection groups
    vliststr *ptgrpsList = protGroups -> GetProtGroupsList ();
    if ( ptgrpsList != NULL)
    {
	Tisprivtab *privTab = db -> GetPrivTab ();
	// get list of protectiongroups our group belongs to along with its
	// access or privilege level
	vliststr *ptgrpPermList = privTab -> GetPrGrpPermissions (groupname);
	int	listRow = 0;
	int	row;
	int	col;
	int	foundMatch;
	vchar	*ptgrpsListValue;
	vchar	*permListValue;
	vchar	*permLevel;
	char	permLevelStr [PRIV_STR_SZ + 1];
	int	intPermValue;
	vlistIterator iterator;
	vlistIterator permIterator;
	// start to loop through the list of protection groups
	iterator.Start (ptgrpsList);
	while (iterator.Next ())
	{
	    col = iterator.GetColumn ();
	    // col 1 = descripton, not needed here
	    if (col != 0)
	    {
		continue;
	    }
	    ptgrpsListValue = (vchar *)(iterator.GetValue ());
	    // add this protection group to our list
	    list -> SetCellValueScribed (listRow, 0, vcharScribe (ptgrpsListValue));
	    // now look in the list of protection groups that this group
	    // belongs to if any, and if we find a match then update the privilege 
	    // level for this protection group in our list
	    if (ptgrpPermList != NULL)
	    {
		foundMatch = vFALSE;
		permIterator.Start (ptgrpPermList);
		while (permIterator.Next ())
		{
		    col = permIterator.GetColumn ();
		    // all we want at this point is the protection 
		    // group name in col 0
		    if (col != 0)
		    {
			continue;
		    }
		    permListValue = (vchar *)(permIterator.GetValue ());
		    // if we have a match
		    if ((vcharCompare (ptgrpsListValue, permListValue)) == 0)
		    {
			row = permIterator.GetRow ();
			// get the privilege level
			permLevel = (vchar *)(ptgrpPermList -> GetCellValue (row, 1));
			if ((vcharLength (permLevel)) != 0)
			{
			    intPermValue = vnum::ParseInt (permLevel, NULL);
			    // funcs in tisPermUitls.cxx
#ifdef	CALISO
			    GetProtectionLevelStr (intPermValue, permLevelStr);
#else
			    GetPrivLevelStr (intPermValue, permLevelStr);
#endif	CALISO
			    // update our list with the privilege string
			    list -> SetCellValueScribed (listRow, 1, 
						      vcharScribeSystem (permLevelStr));
			}
			else
			{
#ifdef	CALISO
			    list -> SetCellValueScribed (listRow, 1,
						vcharScribeSystem (NO_ACCESS_STR));
#else
			    list -> SetCellValueScribed (listRow, 1,
						vcharScribeSystem (VIEW_STR));
#endif	CALISO
			}
			foundMatch = vTRUE;
			break;
		    }
		}
		permIterator.Finish ();
		if (foundMatch == vFALSE)
		{
#ifdef	CALISO
		    list -> SetCellValueScribed (listRow, 1,
						vcharScribeSystem (NO_ACCESS_STR));
#else
		    list -> SetCellValueScribed (listRow, 1,
						vcharScribeSystem (VIEW_STR));
#endif	CALISO
		}
	    }
	    else
	    {
#ifdef	CALISO
		list -> SetCellValueScribed (listRow, 1,
						vcharScribeSystem (NO_ACCESS_STR));
#else
		list -> SetCellValueScribed (listRow, 1,
						vcharScribeSystem (VIEW_STR));
#endif	CALISO
	    }
	    listRow++;
	}
	iterator.Finish ();
	// some cleanup of work list
	if (ptgrpPermList != NULL)
	{
	    delete ptgrpPermList;
	}
    }
    // clean up this work list also
    if (ptgrpsList != NULL)
    {
	delete ptgrpsList;
    }
    // atach our list to the PtProtListView
    SetList (list);

    // calculate column widths
    vlistsize *size = GetSizes ();
    SetColumnWidth (vlistviewALL_COLUMNS, size -> GetColumnWidth (0) / 2);

    vlistSelection *lselection = GetSelection ();
    list -> SetSelectMethod (lselection, vlistFREE_SELECTION);
    list -> SetExtendRowSelection (lselection, vTRUE);
}


/*******************************************************************
*
*   ProtGrpListView observe function
*   list is part of PtProtListView class def and is private   
*
*******************************************************************/

void PtProtListView::ObserveSelect (int state)
{
    if (state == vlistviewSELECTION_HAS_CHANGED)
    {
	vlistSelection *listSelection = GetSelection ();
	int row, column;
	list -> GetSelectedCell (listSelection, &row, &column);
	if (row == vlistNO_CELLS_SELECTED)
	{
	    return;
	}
	vchar *privLevel = (vchar *)(list -> GetCellValue (row, 1));
	if ((vcharLength (privLevel)) == 0)
	{
	    return;
	}
	char privLevelStr [PRIV_STR_SZ + 1];
	vcharExportBounded (vcharScribeSystemExport (privLevel),
			    privLevelStr,
			    sizeof (privLevelStr));
	privLevelStr [sizeof (privLevelStr) - 1] = 0;
#ifdef	CALISO
	DefaultProtectionLevels privLevelValue = GetProtectionLevelValue (privLevelStr);
	accessDialog -> privilegeLevel -> SetProtectionButton (privLevelValue);
#else
	DefaultPrivilegeLevels privLevelValue = GetPrivLevelValue (privLevelStr);
	accessDialog -> privilegeLevel -> SetPrivilegeButton (privLevelValue);
#endif	CALISO
    }
    return;
}


vkindTYPED_FULL_DEFINITION(PtProtPrivLevel, vcontainerExclusiveGroup, NULL);

/*******************************************************************
*
*    PtProtPrivLevel load function
*
*******************************************************************/

void PtProtPrivLevel::LoadInit (vresource res)
{
#ifdef	CALISO
    fullAccessToggle = vcontrolToggle::RequestSubLoad (this,
					tisperm_FullAccess);

    viewOnlyToggle = vcontrolToggle::RequestSubLoad (this,
					tisperm_ViewOnly);

    noAccessToggle = vcontrolToggle::RequestSubLoad (this,
					tisperm_NoAccess);
#else
    adminToggle = vcontrolToggle::RequestSubLoad (this,
					tisperm_Admin);

    engineerToggle = vcontrolToggle::RequestSubLoad (this,
					tisperm_Engineer);

    supervisorToggle = vcontrolToggle::RequestSubLoad (this,
					tisperm_Supervisor);

    operatorToggle = vcontrolToggle::RequestSubLoad (this,
					tisperm_Operator);

    traineeToggle = vcontrolToggle::RequestSubLoad (this,
					tisperm_Training);

    viewToggle = vcontrolToggle::RequestSubLoad (this,
					tisperm_View);
#endif	CALISO

    vcontainerExclusiveGroup::LoadInit (res);

    return;
}


/*******************************************************************
*
*    PtProtPrivLevel observe function
*
*******************************************************************/

void PtProtPrivLevel::ObserveDialogItem (vevent *event)
{
    vliststr *list = (vliststr *)(accessDialog -> ptProtListView -> GetList ());
    if (list == NULL)
    {
	return;
    }
    vlistSelection *lselection = accessDialog -> ptProtListView -> GetSelection ();
    int row, column;
    list -> GetSelectedCell (lselection, &row, &column);
    if (row == vlistNO_CELLS_SELECTED)
    {
	return;
    }
    char *privLevelStr = GetPrivStr ();
    if (privLevelStr == NULL)
    {
	return;
    }

    vlistIterator iterator;
    iterator.StartWithSelection (lselection);
    while (iterator.Next ())
    {
	row = iterator.GetRow ();
	column = iterator.GetColumn ();
	if (column == 1)
	{
	    list -> SetCellValueScribed (row, 1, vcharScribeSystem (privLevelStr));
	}
    }
    iterator.Finish ();
    *ptgrpModified = vTRUE;
    return;
}


/*******************************************************************
*
*    PtProtPrivLevel set radio button function
*
*******************************************************************/

char *PtProtPrivLevel::GetPrivStr (void)
{
    char *privLevelStr = NULL;
    vcontrolToggle *toggle = (vcontrolToggle *)(GetSelection ());
#ifdef	CALISO
    if (toggle == fullAccessToggle)
    {
	privLevelStr = FULL_ACCESS_STR;
    }
    else if (toggle == viewOnlyToggle)
    {
	privLevelStr = VIEW_ONLY_STR;
    }
    else if (toggle == noAccessToggle)
    {
	privLevelStr = NO_ACCESS_STR;
    }
#else
    if (toggle == adminToggle)
    {
	privLevelStr = ADMIN_STR;
    }
    else if (toggle == engineerToggle)
    {
	privLevelStr = ENGINEER_STR;
    }
    else if (toggle == supervisorToggle)
    {
	privLevelStr = SUPERVISOR_STR;
    }
    else if (toggle == operatorToggle)
    {
	privLevelStr = OPERATOR_STR;
    }
    else if (toggle == traineeToggle)
    {
	privLevelStr = TRAINEE_STR;
    }
    else if (toggle == viewToggle)
    {
	privLevelStr = VIEW_STR;
    }
#endif	CALISO
    return (privLevelStr);
}


/*******************************************************************
*
*    PtProtPrivLevel set radio button function
*
*******************************************************************/

void PtProtPrivLevel::SetPrivilegeButton (DefaultPrivilegeLevels privLevel)
{
    switch (privLevel)
    {
#ifndef	CALISO
	case ViewPriv:
	    SetSelection (viewToggle);
	    break;
	case TraineePriv:
	    SetSelection (traineeToggle);
	    break;
	case OperatorPriv:
	    SetSelection (operatorToggle);
	    break;
	case SupervisorPriv:
	    SetSelection (supervisorToggle);
	    break;
	case EngineerPriv:
	    SetSelection (engineerToggle);
	    break;
	case AdminPriv:
	    SetSelection (adminToggle);
	    break;
#endif	!CALISO
	default:
	    break;
    }
}


/*******************************************************************
*
*    PtProtPrivLevel set radio button function
*
*******************************************************************/

void PtProtPrivLevel::SetProtectionButton (DefaultProtectionLevels privLevel)
{
#ifdef	CALISO
    switch (privLevel)
    {
	case NoAccess:
	    SetSelection (noAccessToggle);
	    break;
	case ViewOnly:
	    SetSelection (viewOnlyToggle);
	    break;
	case FullAccess:
	    SetSelection (fullAccessToggle);
	    break;
	default:
	    break;
    }
#endif	CALISO
}


vkindTYPED_FULL_DEFINITION(PtProtApplyButton, vbutton, NULL);

/*******************************************************************
*
*    PtProtApplyButton observe function
*
*******************************************************************/

void PtProtApplyButton::ObserveDialogItem (vevent *event)
{
    MainDialog *mainDialog = accessDialog -> GetMainDialog ();
    TisPermApp *app = mainDialog -> GetPermApp ();
    TisPermDatabase *db = app -> GetDb ();
    Tisprivtab *privTab = db -> GetPrivTab ();
    OptionTree *optionTree = app -> GetOptionTree ();
    TREENODE_PTR treeRoot = optionTree -> GetTreeRoot ();
    int utilModifed = accessDialog -> GetUtilModifiedFlag ();
    int prGrpModified = accessDialog -> GetPtGrpModifiedFlag ();
    vchar *groupName = accessDialog -> GetGroupName ();

    if ((utilModifed == vFALSE) && (prGrpModified == vFALSE) )
    {
	// no mods made msg
	tisPermMsgCenter (accessDialog, TISPERM_INFO_NOTICE,
						tisperm_NothingModMsg);
	return;
    }
    // ask if want to update with chgs made
    int result = tisPermMsgCenter (accessDialog, TISPERM_CAUTION_NOTICE,
						 tisperm_UpdateAccessReq,
						 groupName);
    if (result == vFALSE)
    {
	return;
    }
    char groupname [GROUPNAME_SZ + 1];
    vcharExportBounded (vcharScribeSystemExport (groupName),
			groupname,
			sizeof (groupname));
    groupname [sizeof (groupname) - 1] = 0;
    if (utilModifed == vTRUE)
    {
	optionTree -> WriteTree (treeRoot, groupname);
    }
    if (prGrpModified == vTRUE)
    {
	vliststr *protGroupList = (vliststr *)(accessDialog -> ptProtListView -> GetList ());
	if (protGroupList != NULL)
	{
	    vliststr *newList = new vliststr;
	    int			    row, col;
	    int			    listRow = 0;
	    vchar		    *listValue;
	    char		    privLevelStr [PRIV_STR_SZ + 1];
	    vlistIterator	    iterator;

	    iterator.Start (protGroupList);
	    while (iterator.Next ())
	    {
		row = iterator.GetRow ();
		col = iterator.GetColumn ();
		// want to be in column 0, the protection group name
		if (col != 0)
		{
		    continue;
		}
		// get privilege level
		listValue = (vchar *)(protGroupList -> GetCellValue (row, 1));
		// is there a privilege level
		if ((vcharLength (listValue)) == 0)
		{
		    continue;
		}
		vcharExportBounded (vcharScribeSystemExport (listValue),
				    privLevelStr,
				    sizeof (privLevelStr));
		privLevelStr [sizeof (privLevelStr) - 1] = 0;
#ifdef	CALISO
		// dont pass on NO_ACCESS, a zero privilege
		if ((strcmp (privLevelStr, NO_ACCESS_STR)) == 0)
		{
		    continue;
		}
		// get the numeric value for privilege, see tisPermutils.cxx
		DefaultProtectionLevels privLevel = GetProtectionLevelValue (privLevelStr);
#else
		// dont pass on VIEW, a zero privilege
		if ((strcmp (privLevelStr, VIEW_STR)) == 0)
		{
		    continue;
		}
		// get the numeric value for privilege, see tisPermutils.cxx
		DefaultPrivilegeLevels privLevel = GetPrivLevelValue (privLevelStr);
#endif	CALISO
		// convert to string (vnum::ScribeInt) and store in new list
		newList -> SetCellValueScribed (listRow, 1, vnum::ScribeInt (privLevel));
		// get protection group name and store in new list
		listValue = (vchar *)(protGroupList -> GetCellValue (row, 0));
		newList -> SetCellValueScribed (listRow, 0, vcharScribe (listValue));
		listRow++;

	    }
	    iterator.Finish ();
	    privTab -> UpdatePrtGrpPrivs (newList, groupname);
	}
    }
    // inform of update
    tisPermMsgCenter (accessDialog, TISPERM_INFO_NOTICE,
				    tisperm_UpdateAccessMsg,
				    groupName);
    vevent::StopProcessing ();
    return;
}


vkindTYPED_FULL_DEFINITION(PtProtDismissButton, vbutton, NULL);

/*******************************************************************
*
*    PtProtDismissButton observe function
*
*******************************************************************/

void PtProtDismissButton::ObserveDialogItem (vevent *event)
{
    vevent::StopProcessing ();
    return;
}
