#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstrHEADER
#include vlistviewHEADER
#include vrulerHEADER
#include vcomboboxHEADER

#include "tisperm.h"
#include "tisPermApp.hxx"
#include "tisPermDatabase.hxx"
#include "tisPermMain.hxx"
#include "tisPermUtils.hxx"
#include "tisPermGroup.hxx"

static int groupNameFilter (filteredText *text, vtextSelection *sel,
			    const vchar **s, long *len);
static int descriptionFilter (filteredText *text, vtextSelection *sel,
			      const vchar **s, long *len);

vkindTYPED_FULL_DEFINITION(GroupDialog, vdialog, NULL);

/*******************************************************************
*
*    GroupDialog override for constructors init ()
*
*******************************************************************/

void GroupDialog::LoadInit (vresource res)
{
    modifiedFlag = vFALSE;
    groupName = filteredTextitem::RequestSubLoad (this,
					tisperm_Groupname);

    description = filteredTextitem::RequestSubLoad (this,
					tisperm_Description);

    membersList = vcomboboxPopdown::RequestSubLoad (this,
					tisperm_MemberOfPrGrp);
   
    okButton = GroupOkButton::RequestSubLoad (this,
					tisperm_OkButton);
   
    cancelButton = GroupCancelButton::RequestSubLoad (this,
					tisperm_CancelButton);
    vdialog::LoadInit (res);

    // give buttons ptr to group dialog
    okButton -> SetGroupDialog (this);
    cancelButton -> SetGroupDialog (this);

    // set some filters
    groupName -> SetItemFilterFunc (groupNameFilter);
    groupName -> SetTextModifiedPtr (&this -> modifiedFlag);
    description -> SetItemFilterFunc (descriptionFilter);
    description -> SetTextModifiedPtr (&this -> modifiedFlag);
    Place (GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);
}


/*******************************************************************
*
*    GroupDialog Destroy function
*
*******************************************************************/

void GroupDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }

    vdialog::Destroy ();
}



/*******************************************************************
*
*    GroupDialog notify function
*
*******************************************************************/

void GroupDialog::ObserveWindow (vevent *event)
{
    if ( event -> GetType () == veventWINDOW_STATE && event -> IsClose () )
    {
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}


/*******************************************************************
*
*    GroupDialog Open function
*
*******************************************************************/

void GroupDialog::OpenGroup (MainDialog *mnDlg, GroupFunc grpFunc)
{
    mainDialog = mnDlg;
    groupFunc = grpFunc;

    if (groupFunc == GroupModify)
    {
	// they can't modify the users name on a modify
	UpdateTextFields ();
	groupName -> Disable ();
    }
    SetDialogTitle ();
    // reset here because UpdateTextFields will set it to true
    modifiedFlag = vFALSE;
    // open the dialog and start a sub event loop, halts the caller
    // at the statement after the call to this function
    // allows caller to access the dialog data after dialog is closed
    vdialog::Open();
    MakeTopMostDialog (this, NULL);
    vevent::Process ();
}


/*******************************************************************
*
*    GroupDialog Set dialog title function, depends on what
*	function is to be done
*
*******************************************************************/

void GroupDialog::SetDialogTitle (void)
{
    vstr	*dialogTitle;
    TisPermApp	*tisPerm = mainDialog -> GetPermApp ();
    vresource	titlesRes = tisPerm -> GetTitlesRes ();

    switch (groupFunc)
    {
	case GroupAdd:
	    dialogTitle = vresourceGetString (titlesRes, tisperm_NewGroup);
	    break;

	case GroupModify:
	    dialogTitle = vresourceGetString (titlesRes, tisperm_ModGroup);
	    break;
    }

    SetTitle (dialogTitle);
    vstrDestroy (dialogTitle);
}


/*******************************************************************
*
*    GroupDialog 
*	update groupname and description fields
*
*******************************************************************/

void GroupDialog::UpdateTextFields (void)
{
    int		    row, col;
    const vchar	    *groupValue, *descValue;
    vlistSelection  *selection;
    vliststr	    *groupList;
    char	    group [GROUPNAME_SZ];
    
    selection = mainDialog -> groupListView -> GetSelection ();
    groupList = (vliststr *)mainDialog -> groupListView -> GetList ();
    groupList -> GetSelectedCell (selection, &row, &col);
    groupValue = (vchar *)(groupList -> GetCellValue (row, 0));
    groupName -> SetText (groupValue);
    descValue = (vchar *)(groupList -> GetCellValue (row, 1));
    description -> SetText (descValue);

    TisPermApp *app = mainDialog -> GetPermApp ();
    TisPermDatabase *db = app -> GetDb ();
    Tisprivtab *privTab = db -> GetPrivTab ();
    vcharExportBounded (vcharScribeSystemExport (groupValue),
			group,
			sizeof (group));
    group [sizeof (group) -1 ] = 0;
    // get a list of protection groups this group has privileges in
    vliststr *list = privTab -> GetPrGrpPermissions (group);
    if (list != NULL)
    {
	vchar		*listValue;
	int		col;
	vlistIterator	iterator;

	iterator.Start (list);
	while (iterator.Next ())
	{
	    // col 0 has the protection group name
	    col = iterator.GetColumn ();
	    if (col != 0)
	    {
		continue;
	    }
	    listValue = (vchar *)(iterator.GetValue ());
	    vcomboboxValue *value = membersList -> CreateValue ();
	    value -> SetTitle (listValue);
 	    membersList -> AppendValue (value);
	}
	iterator.Finish ();
	membersList -> SelectValueAt (0);
    }
    else
    {
	membersList -> Disable ();
    }
    delete list;
}


/*******************************************************************
*
*    Some filter functions,for the GroupDialog
*	NCharFilter (see tisgalaxyLib/textFilters.cxx
*
*******************************************************************/

static int groupNameFilter (filteredText *text, vtextSelection *sel,
			    const vchar **s, long *len)
{
    return (NCharFilter (text, sel, s, len, GROUPNAME_SZ));
}

static int descriptionFilter (filteredText *text, vtextSelection *sel,
			      const vchar **s, long *len)
{
    return (NCharFilter (text, sel, s, len, DESCRIPTION_SZ));
}


vkindTYPED_FULL_DEFINITION(GroupOkButton, vbutton, NULL);

/*******************************************************************
*
*    GroupOkButton observe function
*
*******************************************************************/

void GroupOkButton::ObserveDialogItem (vevent *event)
{
    vstr    *nameStr = groupDialog -> groupName -> GetTextAsString ();
    vstr    *descStr = groupDialog -> description -> GetTextAsString ();
    GroupFunc	func = groupDialog -> GetGroupFunction ();
    MainDialog	*mainDialog = groupDialog -> GetMainDialog ();
    int	    groupExists =  mainDialog -> 
			    groupListView -> GroupExists (nameStr);
    int	    todo = TISPERM_ERROR;

    if ((groupDialog -> GetModifiedFlag ()) == vFALSE)
    {
	tisPermMsgCenter (groupDialog, TISPERM_INFO_NOTICE, tisperm_NothingModMsg);
	vstrDestroy (nameStr);
	vstrDestroy (descStr);
	return;
    }

    if (vcharLength (nameStr) == 0)
    {
	tisPermMsgCenter (groupDialog, TISPERM_INFO_NOTICE, tisperm_NoGroupMsg);
	vstrDestroy (nameStr);
	vstrDestroy (descStr);
	return;
    }
    if (func == GroupAdd)
    {
	if (groupExists == vTRUE)
	{
	    tisPermMsgCenter (groupDialog,TISPERM_INFO_NOTICE,
			       tisperm_GroupAlreadyExistsMsg, nameStr);
	    vstrDestroy (nameStr);
	    vstrDestroy (descStr);
	    return;
	}
	else
	{
	    mainDialog -> groupListView -> ModList (AddToList, nameStr, descStr);
	    todo = AddToList;
	}
    }
    // modify existing group
    else if (func == GroupModify)
    {
	if (groupExists == vFALSE)
	{
	    tisPermMsgCenter (groupDialog, TISPERM_INFO_NOTICE,
			       tisperm_GroupDoesNotExistMsg, nameStr);
	    vstrDestroy (nameStr);
	    vstrDestroy (descStr);
	    return;
	}
	else
	{
	    mainDialog -> groupListView -> ModList (ModTheList, nameStr, descStr);
	    todo = ModTheList;
	}
    }

    vstr	*todoStr;
    TisPermApp	*app = mainDialog -> GetPermApp ();
    vresource	msgsRes = app -> GetMessagesRes ();
    switch (todo)
    {
	case TISPERM_ERROR:
	    vstrDestroy (nameStr);
	    vstrDestroy (descStr);
	    return;

	case AddToList:
	    todoStr = vresourceGetString (msgsRes, tisperm_AddedStr);
	    break;

	case ModTheList:
	    todoStr = vresourceGetString (msgsRes, tisperm_ModStr);
	    break;
    }
    tisPermMsgCenter (groupDialog, TISPERM_INFO_NOTICE, 
		       tisperm_GroupDbUpdateMsg, nameStr, todoStr);
    vstrDestroy (nameStr);
    vstrDestroy (descStr);
    vstrDestroy (todoStr);
    vevent::StopProcessing ();
    return;
}


vkindTYPED_FULL_DEFINITION(GroupCancelButton, vbutton, NULL);

/*******************************************************************
*
*    GroupCancelButton observe function
*
*******************************************************************/

void GroupCancelButton::ObserveDialogItem (vevent *event)
{
    vevent::StopProcessing ();
    return;
}

