#include <vport.h>
/* ANSI headers */
#include vstdlibHEADER	/* for exit(), EXIT_ codes */

/* Galaxy headers */
#include vkindHEADER
#include vloadableHEADER
#include vobjectHEADER
#include vscribeHEADER
#include vrectHEADER
#include vcharHEADER
#include vnameHEADER
#include veventHEADER
#include vresourceHEADER
#include vwindowHEADER
#include vapplicationHEADER
#include vdialogHEADER
#include vstartupHEADER
#include vtextitemHEADER
#include vbuttonHEADER
#include vcontainerHEADER
#include vnoticeHEADER
#include vstrHEADER
#include vspinnerHEADER
#include vnumHEADER
#include vfsHEADER

#include "tisgalaxy.h"
#include "tisgalaxylib.hxx"

#ifdef	WIN32
// Galaxy call to get the windows handle for a vwindow
#include    <windows.h>
extern vportLINK HWND vwindowGetMSWWindow (vwindow *window);
#endif	WIN32


/*******************************************************************
*
*    Notice Class
*	needed so ObserveWondow can be overrriden to capture 
*	stack events
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(noticeClass, vnotice, NULL);

void noticeClass::ObserveWindow (vevent *event)
{
#ifndef WIN32
    // this only works on Solaris (motif)
    if (event -> IsStack ())
    {
	// raise the window if it is covered
	Raise ();
    }
#endif	WIN32
}

/*******************************************************************
*
*    cautionNotice
*
*******************************************************************/

int cautionNotice (vdialog *dialog, vstr *title, vstr *msg)
{
    // create a notice dialog
//    noticeClass *cautionNotice = new noticeClass;
    vnotice *cautionNotice = new vnotice;
    int		result = vFALSE;

    cautionNotice -> SetCaution (vTRUE);
    if (title != NULL)
	cautionNotice -> SetTitle (title);
    if (msg != NULL)
    {
	// place message in notice
	cautionNotice -> SetText(msg);
	// center the notice in the dialog that called it
	if (dialog != NULL)
	    cautionNotice -> Place (dialog, vrectPLACE_CENTER, vrectPLACE_CENTER);
	else
	    cautionNotice -> Place (vwindow::GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);

#ifdef	WIN32
	// make sure this dialog is always on top
	HWND windowHandle = vwindowGetMSWWindow (cautionNotice);
	SetWindowPos (windowHandle, HWND_TOPMOST, 0,0,0,0, SWP_NOMOVE | SWP_NOSIZE);
#else
	// raise this dialog to the top, the ObserveWindow will keep it there
	cautionNotice -> Raise ();
#endif	WIN32

	const vname *tag = cautionNotice -> Process ();
	if ( (vcharCompare (tag, vname_No)) == 0)
	    result = vFALSE;
	else
	    result = vTRUE;
    }
    else
    {
	result = vFALSE;
    }
    delete cautionNotice;
    return (result);
}


/*******************************************************************
*
*    infoNotice
*
*******************************************************************/

void infoNotice (vdialog *dialog, vstr *title, vstr *msg)
{
    // create a notice dialog
//    noticeClass *infoNotice = new noticeClass;
    vnotice *infoNotice = new vnotice;

    // overwrite the notice's title
    if (title != NULL)
	infoNotice -> SetTitle (title);
    if (msg != NULL)
    {
	// place our message in the notice
	infoNotice -> SetText (msg);
	// center the notice in the dialog that called it
	if (dialog != NULL)
	    infoNotice -> Place (dialog, vrectPLACE_CENTER, vrectPLACE_CENTER);
	else
	    infoNotice -> Place ( vwindow::GetRoot (), vrectPLACE_CENTER, vrectPLACE_CENTER);

#ifdef	WIN32
	// make sure this dialog is always on top
	HWND windowHandle = vwindowGetMSWWindow (infoNotice);
	SetWindowPos (windowHandle, HWND_TOPMOST, 0,0,0,0, SWP_NOMOVE | SWP_NOSIZE);
#else
	// raise this dialog to the top, the ObserveWindow will keep it there
	infoNotice -> Raise ();
#endif	WIN32

	const vname *tag = infoNotice -> Process ();
    }
    delete infoNotice;
}


/**********************************************************************
*
*   A function to make sure a dialog (windows/solaris) is always
*   on top
*
**********************************************************************/
void MakeTopMostDialog (vwindow *window, vevent *event)
{
/*
#ifdef	WIN32
    // for windows environment
    HWND windowHandle = vwindowGetMSWWindow (window);
    SetWindowPos (windowHandle, HWND_TOPMOST, 0,0,0,0, SWP_NOMOVE | SWP_NOSIZE);
#else
    if (event != NULL)
    {
	// this only works on Solaris (motif)
	if (event -> IsStack ())
	{
	    window -> Raise ();
	}
    }
#endif	WIN32
*/
}


/**********************************************************************
*
*	Some usefull class definitions
*
**********************************************************************/

/**********************************************************************
*
*    class that defines a check box observer
*
*    if you override the ObserveDialogItem, first call
*	optionMenuClass::ObserveDialogItem (vevent *event) so that
*	mod flag can be set
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION(checkBoxClass, vcontrolToggle, NULL);

void checkBoxClass::ObserveDialogItem (vevent *event)
{
    int *mod = GetModifiedPtr ();
    if (mod != NULL)
	SetModifiedFlag ();
}

void checkBoxClass::LoadInit (vresource resource)
{
    vcontrolToggle::LoadInit (resource);
    modified = NULL;
}



/**********************************************************************
*
*    class that defines exclusive container class
*
*    if you override the ObserveDialogItem, first call
*	optionMenuClass::ObserveDialogItem (vevent *event) so that
*	mod flag can be set
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION(exclusiveContainerClass, vcontainerExclusiveGroup, NULL);

void exclusiveContainerClass::ObserveDialogItem (vevent *event)
{
    int *mod = GetModifiedPtr ();
    if (mod != NULL)
	SetModifiedFlag ();
}

void exclusiveContainerClass::LoadInit (vresource resource)
{
    vcontainerGroup::LoadInit (resource);
    modified = NULL;
}


/**********************************************************************
*
*    class that defines group container class
*
*    if you override the ObserveDialogItem, first call
*	optionMenuClass::ObserveDialogItem (vevent *event) so that
*	mod flag can be set
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION(groupContainerClass, vcontainerGroup, NULL);

void groupContainerClass::ObserveDialogItem (vevent *event)
{
    int *mod = GetModifiedPtr ();
    if (mod != NULL)
	SetModifiedFlag ();
}

void groupContainerClass::LoadInit (vresource resource)
{
    vcontainerGroup::LoadInit (resource);
    modified = NULL;
}



/**********************************************************************
*
*    class that defines option menu class
*
*    if you override the ObserveDialogItem, first call
*	optionMenuClass::ObserveDialogItem (vevent *event) so that
*	mod flag can be set
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION(optionMenuClass, vcontrolOptionMenu, NULL);

void optionMenuClass::ObserveDialogItem (vevent *event)
{
    // get current value of control
    int controlValue = this -> GetValue ();

    if ( (controlValue >= this -> GetMin ()) &&
	 (controlValue <= this -> GetMax ()) )
    {
	int *mod = GetModifiedPtr ();
	if (mod != NULL)
	    SetModifiedFlag ();
    }
}

void optionMenuClass::LoadInit (vresource resource)
{
    vcontrolOptionMenu::LoadInit (resource);
    modified = NULL;
}

void optionMenuClass::buildMenu (void)
{
    // no-op, user overrides this function to supply own code
    return;
}

/*******************************************************************
*
*    GetOptionMenuValue ()
*	return the current menu value 
*
*******************************************************************/

const vchar *optionMenuClass::GetOptionMenuValue (void)
{
    const vchar	*menuValue;
    int		controlValue;
    vmenu	*menu;
    vmenuItem	*menuItem;

    controlValue = GetValue ();
    menu = GetMenu ();
    menuItem = menu -> GetItemAt (controlValue);
    menuValue = menuItem -> GetTitle ();

    return (menuValue);
}


/*******************************************************************
*
*    SetOptionMenuValue ()
*	given a string, find the string
*	in the attached menu at set the controls value
*
*******************************************************************/

void optionMenuClass::SetOptionMenuValue (vstr *ptr)
{
    int   i;
    vmenu *menu = GetMenu ();
    int   maxVal   = GetMax ();
    int   minVal   = GetMin ();

    for (i = minVal; i <= maxVal; i++)
    {
	vmenuItem *menuItem = menu -> GetItemAt (i);
	if ( (vcharCompare ( (const)ptr, menuItem -> GetTitle ())) == 0)
	{
	    SetValue (i);
	    return;
	}
    }
}


/**********************************************************************
*
*    class that defines a vspinner observer
*
*    if you override the ObserveDialogItem, first call
*	optionMenuClass::ObserveDialogItem (vevent *event) so that
*	mod flag can be set
*
***********************************************************************/

vkindTYPED_FULL_DEFINITION(spinnerClass, vspinner, "spinnerClass");

void spinnerClass::ObserveDialogItem (vevent *event)
{
    int *mod = GetModifiedPtr ();
    if (mod != NULL)
	SetModifiedFlag ();
}

void spinnerClass::LoadInit (vresource resource)
{
    vspinner::LoadInit (resource);
    modified = NULL;
}

// overirde due to galaxy bug on NT where it show value 
// outside of its range, workaround from Visix

void spinnerClass::SetValue (int val)
{
    vspinner::SetValue (val);
    vtextitem *textItem = GetTextitem ();
    textItem -> SetTextScribed (vnumScribeInt (GetValue ()));
}


/*******************************************************************
*
*    tisGalaxy classes and functions that rely on tisgalaxy.vr
*
*******************************************************************/
/*******************************************************************
*
*    Call startTisGalaxy() at the start of your program, after the
*    call to galaxy's vstartup, and save the vresource returned
*    it is used at the end of your program when you call 
*    stopTisGalaxy (). The file that is opened is:
*	 /tis4000/tasks/tisgalaxy.vr
*    Suggestion:
*	If you have an application wrapper class, call start in the
*	constructor and stop in the destructor
*
*******************************************************************/

vresource startTisGalaxy (void)
{
    vfsPath	*path;
    vresource	tisRes;

    path = vfsPath::BuildScribed (vfsPATH_TYPE_FULL,
				  vcharScribeLiteral ("tis4000"),
				  vcharScribeLiteral ("tasks"),
				  vcharScribeLiteral ("tisgalaxy.vr"), NULL);
    tisRes = vresourceOpenFile ((const vfsPath *)path, vfsOPEN_READ_ONLY);

    // internal globals, found in tisgalaxy.c
    tisgalaxyInternNames ();
    delete path;

    return (tisRes);
}

void stopTisGalaxy (vresource res)
{
    vresourceCloseFile (res);
}


/*******************************************************************
*
*    tisGalaxy About MenuItem Class,
*    provides an about menu item for your dialog
*    Instantiate the class with the vresource from startTisGalaxy ()
*    and a vchar ptr to a string that will become the version label
*    on the about menu
*
*******************************************************************/

TisAbout::TisAbout (vresource res, vchar *verText)
{
    aboutDialog = NULL;
    tisGalaxyRes = res;
    versionText = NULL;
    aboutRes = vresourceGet (tisGalaxyRes, tisgalaxy_aboutDialog);
    aboutDialog = new AboutDialog (aboutRes);
    if (verText != NULL)
    {
	versionText = vstrClone ((const vchar *)verText);
	// change verion label to what user has specified
	SetVersionLabel (versionText);
    }
}

TisAbout::~TisAbout ()
{
    if (versionText != NULL)
    {
	vstrDestroy (versionText);
    }
    if (aboutDialog != NULL)
    {
	delete aboutDialog;
    }
}

// provide if user has chosen to use an observe function
void TisAbout::OpenAboutDialog (void)
{
    aboutDialog -> Place (aboutDialog -> GetRoot(), vrectPLACE_CENTER, vrectPLACE_CENTER);
    aboutDialog -> Open ();
}

void TisAbout::SetAboutTitle (vchar *title)
{
    if (aboutDialog != NULL)
    {
	aboutDialog -> SetTitle (title);
    }
}

void TisAbout::SetResizable (void)
{
    if (aboutDialog != NULL)
    {
	int style = aboutDialog -> GetStyle ();
	aboutDialog -> SetStyle (style | vwindowSTYLE_RESIZABLE);
    }
}

// provide if user needs to chg version label before About dialog is opened
void TisAbout::SetVersionLabel (vchar *text)
{ 
    vtextitem *versionLabel = aboutDialog -> GetVersionLabel ();
    versionLabel -> SetJustification (vtextCENTER_JUSTIFY);
    versionLabel -> SetTextScribed (vcharScribeSystem (""));
    versionLabel -> SetText ((const vchar *)text);
}


/*******************************************************************
*
*   tisGalaxy About Class
*	return a vemnuItem, titled About
*   AboutMenuItem - override class has ObserveMenuItem override
*
*******************************************************************/

AboutMenuItem *TisAbout::GetAboutMenuItem (void)
{
    vresource titlesRes = vresourceGet (tisGalaxyRes, tisgalaxy_titlesDictionary);
    vstr *aboutTitle = vresourceGetString (titlesRes, tisgalaxy_aboutMenuTitle);
    AboutMenuItem *item = new AboutMenuItem ();
    item -> SetTitle (aboutTitle);
    item -> SetAboutDialog (aboutDialog);
    vstrDestroy (aboutTitle);

    return (item);
}
/*******************************************************************
*
*   tisGalaxy About Class
*	return a vemnuItem, titled About
*   OtherAboutMenuItem - does not have override because user has own
*	observe function
*
*******************************************************************/

OtherAboutMenuItem *TisAbout::GetOtherAboutMenuItem (void)
{
    vresource titlesRes = vresourceGet (tisGalaxyRes, tisgalaxy_titlesDictionary);
    vstr *aboutTitle = vresourceGetString (titlesRes, tisgalaxy_aboutMenuTitle);
    OtherAboutMenuItem *item = new OtherAboutMenuItem ();
    item -> SetTitle (aboutTitle);
    item -> SetAboutDialog (aboutDialog);
    vstrDestroy (aboutTitle);

    return (item);
}


/*******************************************************************
*
*    tisGalaxy About Dialog Class
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(AboutDialog, vdialog, NULL);

void AboutDialog::LoadInit (vresource res)
{
    aboutRes = res;
    versionLabel = vtextitem::RequestSubLoad (this, tisgalaxy_aboutVersionLabel);
    okButton = AboutOkButton::RequestSubLoad (this, tisgalaxy_aboutOkButton);
    vdialog::LoadInit (res);
    okButton -> SetAboutDialog (this);
}

void AboutDialog::Destroy (void)
{
    if (IsOpen ())
    {
	Close ();
    }
    vdialog::Destroy ();
}

void AboutDialog::ObserveWindow (vevent *event)
{
    if (event -> GetType () == veventWINDOW_STATE && event -> IsClose ())
    {
	Close ();
    }
}


/*******************************************************************
*
*    tisGalaxy AboutOkButton Classes
*
*******************************************************************/
vkindTYPED_FULL_DEFINITION(AboutOkButton, vbutton, NULL);

void AboutOkButton::ObserveDialogItem (vevent *event)
{
    aboutDialog -> Close ();
}

/*******************************************************************
*
*    tisGalaxy AboutMenuItem Classes, needed to override the
*    ObserveMenuItem class function
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(AboutMenuItem, vmenuItem, NULL);

void AboutMenuItem::ObserveMenuItem (vevent *event)
{
    aboutDialog -> Place (aboutDialog -> GetRoot(), vrectPLACE_CENTER, vrectPLACE_CENTER);
    aboutDialog -> Open ();
}

/*******************************************************************
*
*    tisGalaxy OtherAboutMenuItem Classes,
*
*******************************************************************/

vkindTYPED_FULL_DEFINITION(OtherAboutMenuItem, vmenuItem, NULL);

