/* trendcontrols.cxx,v 1.1.1.1 1999/08/03 14:45:09 tis4000 Exp
 *
 */

/*
 * TrendControls - trend set controls dialog class
 */

#include <vport.h>		/* for v*HEADER */
#include vstdlibHEADER		/* for random, exit, EXIT_SUCCESS */
#include vstartupHEADER
#include vcontainerHEADER
#include vmemHEADER
#include vprefHEADER
#include vtextitemHEADER
#include vbuttonHEADER
#include vcharHEADER
#include vnumHEADER

#ifdef WIN32
#include <windows.h>
#include <winbase.h>
#include <stdio.h>
#endif WIN32

extern "C"
{
    #include "trendutils.h"
}
#include "trenddocs.h"
#include "trendcontrols.h"
#include "tisgalaxylib.hxx"

// Help functions
class tcControlButton : public vbutton {
public:
  // macro for overrides from vkindTyped
  vkindTYPED_FULL_DECLARATION(tcControlButton);
  
  // macro for the standard versions of the 
  // necessary constructors and destructors
  vloadableINLINE_CONSTRUCTORS(tcControlButton, vbutton);
  
  // virtual member function overrides
  void ObserveDialogItem(vevent *event);

  void SetParent (TrendControls *p) { parent = p; }
  void SetType (int t) { type = t; }
private:
    TrendControls *parent;
    int		type;
};

vkindTYPED_FULL_DEFINITION(tcControlButton, vbutton, NULL);
void tcControlButton::ObserveDialogItem (vevent *event)
{
    parent->ControlClick (type);
}

class tcControlToggle : public vcontrolToggle {
public:
  // macro for overrides from vkindTyped
  vkindTYPED_FULL_DECLARATION(tcControlToggle);
  
  // macro for the standard versions of the 
  // necessary constructors and destructors
  vloadableINLINE_CONSTRUCTORS(tcControlToggle, vcontrolToggle);
  
  // virtual member function overrides
  void ObserveDialogItem(vevent *event);

  void SetParent (TrendControls *p) { parent = p; }
  void SetType (int t) { type = t; }
private:
    TrendControls *parent;
    int		type;
};

vkindTYPED_FULL_DEFINITION(tcControlToggle, vcontrolToggle, NULL);
void tcControlToggle::ObserveDialogItem (vevent *event)
{
    parent->ToggleClick (type, GetValue());
}

// macro definitions for vkindtyped overrides
vkindTYPED_FULL_DEFINITION(TrendControls, vcontainer, NULL);

// vloadable
void TrendControls::LoadInit(vresource resource)
{
    // get handles to interesting elements
    timeDown = tcControlButton::RequestSubLoad (this, vnameInternGlobalLiteral("TimeBackTag"));
    timeSet = tcControlButton::RequestSubLoad (this, vnameInternGlobalLiteral("TimeTag"));
    timeUp = tcControlButton::RequestSubLoad (this, vnameInternGlobalLiteral("TimeForwardTag"));
    cursorDown = tcControlButton::RequestSubLoad (this, vnameInternGlobalLiteral("CursorBackTag"));
    cursorSet = tcControlButton::RequestSubLoad (this, vnameInternGlobalLiteral("CursorTag"));
    cursorUp = tcControlButton::RequestSubLoad (this, vnameInternGlobalLiteral("CursorForwardTag"));
    zoomDown = tcControlButton::RequestSubLoad (this, vnameInternGlobalLiteral("ZoomInTag"));
    zoomSet = tcControlButton::RequestSubLoad (this, vnameInternGlobalLiteral("ZoomTag"));
    zoomUp = tcControlButton::RequestSubLoad (this, vnameInternGlobalLiteral("ZoomOutTag"));
    textToggle = tcControlToggle::RequestSubLoad (this, vnameInternGlobalLiteral("TextCheckTag"));
    autoToggle = tcControlToggle::RequestSubLoad (this, vnameInternGlobalLiteral("AutoCheckTag"));
    spanText = vtextitem::RequestSubLoad (this, vnameInternGlobalLiteral("SpanTextTag"));
    rateText = vtextitem::RequestSubLoad (this, vnameInternGlobalLiteral("RateTextTag"));
    locationText = vtextitem::RequestSubLoad (this, vnameInternGlobalLiteral("LocationTextTag"));

    vcontainer::LoadInit(resource);

    timeDown->SetParent (this);
    timeDown->SetType (tcTIMEDOWN);
    timeSet->SetParent (this);
    timeSet->SetType (tcTIME);
    timeUp->SetParent (this);
    timeUp->SetType (tcTIMEUP);
    cursorDown->SetParent (this);
    cursorDown->SetType (tcCURSORDOWN);
    cursorSet->SetParent (this);
    cursorSet->SetType (tcCURSOR);
    cursorUp->SetParent (this);
    cursorUp->SetType (tcCURSORUP);
    zoomDown->SetParent (this);
    zoomDown->SetType (tcZOOMDOWN);
    zoomSet->SetParent (this);
    zoomSet->SetType (tcZOOM);
    zoomUp->SetParent (this);
    zoomUp->SetType (tcZOOMUP);

    textToggle->SetParent (this);
    textToggle->SetType (tcTEXTTOGGLE);
    autoToggle->SetParent (this);
    autoToggle->SetType (tcAUTOTOGGLE);
}

/*********
 *  TrendControls::ControlClick (int type)
 *
 *  The procedure gets called whenever one of the control
 *  buttons is clicked 
 *
 **********/
void TrendControls::ControlClick (int type)
{
}

/*********
 *  TrendControls::ToggleClick (int type, int value)
 *
 *  The procedure gets called whenever one of the toggle
 *  buttons is clicked 
 *
 **********/
void TrendControls::ToggleClick (int type, int value)
{
}

/*********
 *  TrendControls::SetTextIndicator (int value)
 *
 *  The procedure gets called whenever the text indicator
 *  needs a new value
 *
 **********/
void TrendControls::SetTextIndicator (int value)
{
    textToggle->SetValue (value);
}

/*********
 *  TrendControls::SetTheCursor (int cursor)
 *
 *  The procedure gets called whenever the cursor should be
 *   redisplayed
 *
 **********/
void TrendControls::SetTheCursor (int cursor)
{
    char    tempstring [40];

    spanText->SetText ((unsigned char *)StrSpan (trendSet->zoom*512L, tempstring));
    rateText->SetTextScribed (vnum::ScribeFloat(trendSet->zoom));
    locationText->SetTextScribed (vnum::ScribeFloat(cursor));
}

/*********
 *  TrendControls::SetTrendSet (TrendSet *set))
 *
 *  The procedure gets called whenever the cursor should be
 *   redisplayed
 *
 **********/
void TrendControls::SetTrendSet (TrendSet *set)
{
    trendSet = set;

    textToggle->SetValue (0);
    autoToggle->SetValue (trendSet->autoupdate);
}
