/* $Id: Trenddocs.cxx
 */

/*
 * Trenddocs.cxx - trend document classes
 */

#include <vport.h>		/* for v*HEADER */
#include vstdlibHEADER		/* for random, exit, EXIT_SUCCESS */
#include vstartupHEADER
#include vdialogHEADER
#include vbuttonHEADER
#ifdef WIN32
#include <windows.h>
#include <winbase.h>
#include <stdio.h>
#endif WIN32
#include <time.h>

extern "C"
{
    #include "trendutils.h"
    #include "oracleSnap.h"
    #include "archiveLib.h"
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
    #include "cb45to63.h"
    #include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32
    #include "avmDefs.h"
    #include "xbaseSnapLib.h"
extern char	*dbreducs [];
extern char	*qual;
extern char	*sev;
}
#include "trenddocs.h"
#include "tisgalaxylib.hxx"


TrendData::TrendData ()
{
    strcpy (pointname,"");
    strcpy (egutag,"");
    dbreduc = 0;
    reduc = 0;
    starttime = 0;
    zoom = 60;
    dbrate = 60;
    dispLo = dispHi = 0;
    ctlLo = ctlHi = 0;
    ctlEnable = 1;
    enable = 1;
}

void TrendData::UpdateData (CHARSPEC trspec, int vuzoom)
{
    ctlEnable = 1;
    enable = 1;
    GetTSpec (trspec, pointname, &dbreduc, &reduc, date, off,
		    &starttime, &dispLo, &dispHi, &ctlLo, &ctlHi, vuzoom);
    UpdateData (long(0), vuzoom);
}

void TrendData::UpdateData (long vutimechange, int vuzoom)
{
    int retcnt;

    starttime += vutimechange;
    zoom = vuzoom;

    /* Calculated sets return to caller */
    if (pointname[0] == '#')
	return;

    /* collect a set of ar_snap value/status pairs for the given time units/span,
    ** reducing data if necessary. Returns number of values entered. */
    else if (pointname[0] == '>')
	retcnt = GetSnapValues (&pointname[1], dbreducs[dbreduc], dbreducs[reduc],
				vuzoom, starttime, &dbrate,
				egutag, &dbdispLo, &dbdispHi,
				C_TRENDSAMPLES, value1, status, value2);
    else
	retcnt = GetSnapValues (pointname, dbreducs[dbreduc], dbreducs[reduc],
				vuzoom, starttime, &dbrate,
				egutag, &dbdispLo, &dbdispHi,
				C_TRENDSAMPLES, value1, status, value2);

/* turn this on later
    if (GetOfflineTables ())
	archTables = GetOfflineTables ();
*/
}

void TrendData::UpdateData (TrendData *ds)
{
    for (int i = 0; i < C_TRENDSAMPLES; i++)
    {
        status[i][0] = ds->status[i][0];
        status[i][1] = ds->status[i][1];
        status[i][2] = ds->status[i][2];
        value1[i] = ds->value1[i];
    }
}

void TrendData::AddData (TrendData *ds)
{
    for (int i = 0; i < C_TRENDSAMPLES; i++)
    {
        if ((status[i][0] == 0xff) || (ds->status[i][0] == 0xff))
        {
            status[i][0] = (unsigned char)0xff;
        }
        else
        {
            value1[i] += ds->value1[i];
            if (reduc == 4)
                value2[i] += ds->value2[i];
            else if (value2[i] < ds->value2[i])
                value2[i] = ds->value2[i];
            // Take the most severe status
            if (strchr(qual, ds->status[i][0]) < strchr(qual, status[i][0]))
	    {
                strncpy ((char *)status[i], (char *)ds->status[i], sizeof(status[i]));
	    }
            else if (strchr (sev, ds->status[i][1]) < strchr (sev, status[i][1]))
	    {
                status[i][1] = ds->status[i][1];
                status[i][2] = ds->status[i][2];
	    }
        }
    }
}

void TrendData::SubtractData (TrendData *ds)
{
    for (int i = 0; i < C_TRENDSAMPLES; i++)
    {
        if ((status[i][0] == 0xff) || (ds->status[i][0] == 0xff))
        {
            status[i][0] = (unsigned char)0xff;
        }
        else
        {
            value1[i] -= ds->value1[i];
            if (reduc == 4)
                value2[i] -= ds->value2[i];
            else if (value2[i] < ds->value2[i])
                value2[i] = ds->value2[i];
            // Take the most severe status
            if (strchr(qual, ds->status[i][0]) < strchr(qual, status[i][0]))
	    {
                strncpy ((char *)status[i], (char *)ds->status[i], sizeof(status[i]));
	    }
            else if (strchr (sev, ds->status[i][1]) < strchr (sev, status[i][1]))
	    {
                status[i][1] = ds->status[i][1];
                status[i][2] = ds->status[i][2];
	    }
        }
    }
}

int TrendData::Search (SEARCH *srch, int htcursor)
{
    int    ix;
    int    retix;
    int    dir;

    if (srch->type [1] == 'S')
    {
        ix = 0;
        dir = 1;
    }
    else if (srch->type [1] == 'C')
    {
        if (htcursor < (C_TRENDSAMPLES - 1))
        {
            ix = htcursor + 1;
            dir = 1;
        }
        else
            return (-1);
    }
    else if (srch->type [1] == 'P')
    {
        if (htcursor != 0)
        {
            ix = htcursor - 1;
            dir = -1;
        }
        else
            return (-1);
    }
    else
        return (-1);

    if (pointname [0])
    {
        if (srch->type [2] == '+')
        {
            for (retix = ix; (0 <= ix) && (ix < C_TRENDSAMPLES); ix += dir)
                if (value1[ix] > value1[retix])
                    retix = ix;
        }
        else if (srch->type [2] == '-')
        {
            for (retix = ix; (0 <= ix) && (ix < C_TRENDSAMPLES); ix += dir)
                if (value1[ix] < value1[retix])
                    retix = ix;
        }
        else
        {
            for (; (0 <= ix) && (ix < C_TRENDSAMPLES); ix += dir)
            {
                if (strchr((char *)srch->severity, status[ix][1])
		  || strchr((char *)srch->status, status[ix][2]))
                    return (ix);
            }
            return (-1);
        }
        return (retix);
    }
    return (-1);
}

void TrendData::GetTrendSpec (CHARSPEC trspec, int convertAbs)
{
    if (convertAbs)
    {
        GmtDateString (date, starttime);
	strcpy (off, "");
    }
    MakeTSpec ( trspec, PointName (), dbreduc, reduc,
		date, off, dispLo, dispHi, ctlLo, ctlHi );
}

time_t TrendData::GetNoteTime (time_t when)
{
    // must fill in later
    return 0;
}

void TrendData::GetDetailString (char *text, int textlen, int cursloc)
{

    if (*(pointname) == '#')
    {
        sprintf (text, "%-28.28s %3.3s %-10.10s %-8.8s %-9.9s %-8.8s %-9.9s %-3.3s",
			        PointName (), "",
			        "", "",
			        dValue (cursloc), "", dValue2 (cursloc),
			        dStatus (cursloc));
    }
    else
    {
        if (reduc != dbreduc)
            sprintf (text, "%-23.23s(%3.3s) %3.3s %-10.10s %-8.8s %-9.9s %-8.8s %-9.9s %-3.3s",
                    			PointName (), PointReduc (), PointDBReduc (),
                    			PointDate (cursloc), PointTime (cursloc),
                    			dValue (cursloc), PointTag (), dValue2 (cursloc),
                    			dStatus (cursloc));
        else
            sprintf (text, "%-28.28s %3.3s %-10.10s %-8.8s %-9.9s %-8.8s %-9.9s %-3.3s",
                    			PointName (), PointReduc (),
                    			PointDate (cursloc), PointTime (cursloc),
                    			dValue (cursloc), PointTag (), dValue2 (cursloc),
                    			dStatus (cursloc));
    }
}

void TrendData::GetValueString (char *text, int textlen, int cursloc)
{
    sprintf (text, "%-9.9s %9.9s %-3.3s", dValue (cursloc), dValue2 (cursloc), dStatus (cursloc));
}

char * TrendData::PointName ()
{
    static char name [C_PTNMLEN];

    if (pointname [0])
    {
	return (ResolveName (name, pointname, sizeof(pointname)));
    }
    else
    {
        return (NULL);
    }
}

char * TrendData::PointReduc ()
{
    return (dbreducs[reduc]);
}

char * TrendData::PointDBReduc ()
{
    return (dbreducs[dbreduc]);
}

char * TrendData::PointTag ()
{
    return (egutag);
}

char    cnvdate[10];
char    cnvtime[10];

char * TrendData::PointDate (int cursloc)
{
    return (LocalStrDate (starttime + cursloc * zoom, cnvdate));
}
char * TrendData::PointTime (int cursloc)
{
    return (LocalStrTime (starttime + cursloc * zoom, cnvtime));
}

long TrendData::PointTick (int cursloc)
{
    return (starttime + cursloc * zoom);
}

char cnvvalue[20];

char * TrendData::dValue (int cursloc)
{
    if ((status[cursloc][0] == 0xff) || (status[cursloc][0] == 'C'))
    {
        sprintf (cnvvalue, "         ");
    }
    else
    {
        sprintf (cnvvalue, "%9g", value1[cursloc]);
    }
    return (cnvvalue);
}

char cnvvalue2[20];

char * TrendData::dValue2 (int cursloc)
{
    if ((status[cursloc][0] == 0xff) || (status[cursloc][0] == 'C'))
    {
        sprintf (cnvvalue2, "         ");
    }
    else
    {
        sprintf (cnvvalue2, "%9g", value2[cursloc]);
    }
    return (cnvvalue2);
}

char * TrendData::dStatus (int    cursloc)
{
    if (status[cursloc][0] == 0xff)
    {
        return "   ";
    }
    else
    {
	return (char *)status[cursloc];
    }
}

char hitext[20];

char * TrendData::PointHiLim (void)
{
    sprintf (hitext, "%9g", dispHi);
    return (hitext);
}

char lotext[20];

char * TrendData::PointLoLim (void)
{
    sprintf (lotext, "%9g", dispLo);
    return (lotext);
}

TrendSet::TrendSet ()
{
    strcpy (setname, "");
    zoom = 1;
    autoupdate = 0;
    numTrends = 0;
    for (int i = 0; i < C_TRENDPTS; i++)
    {
	specs[i][0] = 0;
    }
}

void TrendSet::ChangeSet (char *setName)
{
    if (GetTrendInfo (setName, &zoom, &autoupdate, specs))
    {
	strncpy (setname, setName, sizeof(setname));
	for (numTrends = 0; numTrends < C_TRENDPTS; numTrends++)
	{
	    if (strlen (specs[numTrends]) == 0)
		break;
	    // Unpack the specification
	    trends[numTrends].UpdateData (specs[numTrends], zoom);
	}
    }
}

void TrendSet::SaveSet (char *setName, int convertAbs)
{
    strncpy (setname, setName, sizeof(setname));
    for (int i = 0; i < numTrends; i++)
    {
	trends[i].GetTrendSpec (specs[i], convertAbs);
    }
    PutTrendInfo (setname, zoom, autoupdate, specs, TRUE);
}

void TrendSet::UpdateSet (long vutimechange, int vuzoom)
{
    char *cp;
    int i;

    for (i = 0; i < numTrends; i++)
    {
	trends[i].UpdateData (vutimechange, vuzoom);
    }
    /* Now execute calculated sets */
    for (i = 0; i < numTrends; i++)
    {
        cp = trends[i].PointName();
	if (*cp == '#')
	{
	    cp++;
	    if ((0 <= (*cp-'1')) && ((*cp-'1') < i))
	    {
		trends[i].UpdateData (&trends[*cp-'1']);
		cp++;
		while (*cp && (*cp != '|'))
		{
		    while (*cp == ' ') cp++;
		    if (*cp == '+')
		    {
			cp++;
			while (*cp == ' ') cp++;
			if (*cp == '#')
			{
			    cp++;
			    if ((0 <= (*cp-'1')) && ((*cp-'1') < i))
				trends[i].AddData (&trends[*cp-'1']);
			    else
			    {
				infoNotice (NULL, (unsigned char *)"Specification error",
					    (unsigned char *)"~Bad composite specification");
				break;
			    }
			    cp++;
			}
			else
			{
			    infoNotice (NULL, (unsigned char *)"Specification error",
					(unsigned char *)"~Bad composite specification");
			    break;
			}
		    }
		    else if (*cp == '-')
		    {
			cp++;
			while (*cp == ' ') cp++;
			if (*cp == '#')
			{
			    cp++;
			    if ((0 <= (*cp-'1')) && ((*cp-'1') < i))
				trends[i].SubtractData (&trends[*cp-'1']);
			    else
			    {
				infoNotice (NULL, (unsigned char *)"Specification error",
					    (unsigned char *)"~Bad composite specification");
				break;
			    }
			    cp++;
			}
			else
			{
			    infoNotice (NULL, (unsigned char *)"Specification error",
					(unsigned char *)"~Bad composite specification");
			    break;
			}
		    }
		    else	// invalid character in spec
		    {
			infoNotice (NULL, (unsigned char *)"Specification error",
				    (unsigned char *)"~Bad composite specification");
			break;
		    }
		}
	    }
	    else
	    {
		infoNotice (NULL, (unsigned char *)"Specification error",
			    (unsigned char *)"~Bad composite specification");
	    }
	}
    }
}

time_t TrendSet::GetStartTime (void)
{
    for (int i = 0; i < numTrends; i++)
    {
	if (trends[i].pointname [0] == '>')
	    return trends[i].starttime;
    }
    return 0;
}

time_t TrendSet::GetEndTime (void)
{
    for (int i = 0; i < numTrends; i++)
    {
	if (trends[i].pointname [0] == '>')
	    return trends[i].starttime + (zoom * 511);
    }
    return 0;
}

