/* trendgraph.cxx,v 1.1.1.1 1999/08/03 14:45:10 tis4000 Exp
 *
 */

/*
 * TrendGraph - trend graph container class
 */

#include <vport.h>		/* for v*HEADER */
#include vstdlibHEADER		/* for random, exit, EXIT_SUCCESS */
#include vcontainerHEADER
#include vstartupHEADER
#include vmemHEADER
#include vprefHEADER
#include vtextitemHEADER
#include vcharHEADER

#ifdef WIN32
#include <windows.h>
#include <winbase.h>
#include <stdio.h>
#endif WIN32

extern "C"
{
    #include "trendutils.h"
}
#include "trenddocs.h"
#include "trendgraph.h"

// Private classes
class TrendFabric : public vdialogItem
{
public:
    // constructor/destructors
    vloadableINLINE_CONSTRUCTORS (TrendFabric, vdialogItem);

    // overrides from vloadable
    vkindTYPED_FULL_DECLARATION (TrendFabric);

    virtual int HandleButtonDown (vevent *event);
    virtual int HandleButtonUp (vevent *event);
    virtual int HandleDrag (vevent *event);
    
    // members
    void Init ();
    void Draw (void);
    void DrawCursor (int loc);
    void DrawControls (TrendData *trendData, vcolor *color, vrect *inset);
    void DrawPlots (TrendData *trendData, vcolor *color, vrect *inset);
    void SetParent (TrendGraph *p) {parent = p; };
private:
    void DrawFeedback (int x1, int x2);
    TrendGraph *parent;
    int	lastCursor;
    vcolorSpec vred;
    vcolor *redColor;
    int inDrag;
    int startX;
    int stopX;
};
vkindTYPED_FULL_DEFINITION (TrendFabric, vdialogItem, "TrendFabric");

int TrendFabric::HandleButtonDown (vevent *event)
{
    vrect   tfrect;
    vrect   inset;

    if (event->GetButton () == 0) // left (cursor) button
    {
	tfrect = *GetRect();
	inset.Inset(&tfrect, 2, 2);
	parent->NewCursor ((event->GetX () - inset.x) * 512 / inset.w);
	return vTRUE;
    }
    else if (event->GetButton () == 2) // right (range) button
    {
	inDrag = TRUE;
	tfrect = *GetRect();
	inset.Inset(&tfrect, 2, 2);
	startX = stopX = (event->GetX () - inset.x) * 512 / inset.w;
	DrawFeedback (startX, stopX + 1);
	return vTRUE;
    }
    return vdialogItem::HandleButtonDown (event);
}

int TrendFabric::HandleButtonUp (vevent *event)
{
    vrect   tfrect;
    vrect   inset;

    if (inDrag && event->GetButton () == 2) // right (range) button
    {
	tfrect = *GetRect();
	inset.Inset(&tfrect, 2, 2);
	if (startX == stopX)
	    DrawFeedback (startX, stopX+1);
	else
	    DrawFeedback (startX, stopX);
	stopX = (event->GetX () - inset.x) * 512 / inset.w;
	inDrag = FALSE;
	parent->NewRange (startX, stopX);
	return vTRUE;
    }
    return vdialogItem::HandleButtonUp (event);
}

int TrendFabric::HandleDrag (vevent *event)
{
    vrect   tfrect;
    vrect   inset;

    if (inDrag) // right (range) button
    {
	tfrect = *GetRect();
	inset.Inset(&tfrect, 2, 2);
	if (startX == stopX)
	    DrawFeedback (startX, stopX+1);
	else
	    DrawFeedback (startX, stopX);
	stopX = (event->GetX () - inset.x) * 512 / inset.w;
	DrawFeedback (startX, stopX);
	return vTRUE;
    }
    return vdialogItem::HandleDrag (event);
}

//	While tracking our selection, draw an outline indicating the
//	the active position of the selection. We assume the gstate
//	has been properly saved.
//

void TrendFabric::DrawFeedback (int x1, int x2)
{
    vrect   tfrect;
    vrect   inset;
    int	saveW;

    if (!IsOpen())
	return;

    vdraw::GSave ();
    DrawDevice ();
    tfrect = *GetRect();
    vdraw::Translate (tfrect.x, tfrect.y);
    tfrect.SetX(0);
    tfrect.SetY(0);
    inset.Inset(&tfrect, 2, 2);

    inset.SetX (inset.x + (inset.w * x1 / 512));
    inset.SetWidth (inset.w * (x2 - x1) / 512);

    saveW = vdraw::GetLineWidth();
    vdraw::SetLineWidth(0);
    vdraw::SetFunction(vdrawFUNC_INVERT);
    vdraw::RectsStroke(&inset, 1);
    vdraw::SetLineWidth(saveW);
    vdraw::SetFunction(vdrawFUNC_COPY);
    vdraw::GRestore ();
}

void TrendFabric::Init (void)
{
    vred.SetRGB (vcolorCOMPONENT_MAX, 0, 0);
    redColor = vcolor::Intern (&vred);
    inDrag = FALSE;
}

void TrendFabric::Draw ()
{
    vrect   tfrect;
    vrect   inset;
    float   dotpattern [2] = {(float)2.0, (float)4.0};
    int	    i;
    TrendSet *trendSet = parent->GetTrendSet ();

    if (!IsOpen())
	return;

    lastCursor = -1;
    vdraw::GSave ();
    DrawDevice ();
    // Clear the mainDraw area, draw the 3 horizontal lines
    tfrect = *GetRect();
    vdraw::Translate (tfrect.x, tfrect.y);
    tfrect.SetX(0);
    tfrect.SetY(0);
    vdraw::RectsHiliteFillInside
      (&tfrect, 1, parent->bgColor,
       DetermineColor(vdrawCOLOR_SHADOW_BOTTOM),
       DetermineColor(vdrawCOLOR_SHADOW_TOP));
    inset.Inset(&tfrect, 2, 2);
    vdraw::RectsClip(&inset, 1);
    // set the color to the right one for the register lines
    vdraw::SetColor(vcolor::GetBlack());
    vdraw::SetLineWidth(1);
    vdraw::FSetDash(2, dotpattern, 0.0);
    vdraw::NewPath();
    vdraw::MoveTo(inset.x, inset.y+(inset.h/4));
    vdraw::LineTo(inset.x+inset.w, inset.y+(inset.h/4));
    vdraw::Stroke ();
    vdraw::MoveTo(inset.x, inset.y+(inset.h/2));
    vdraw::LineTo(inset.x+inset.w, inset.y+(inset.h/2));
    vdraw::Stroke ();
    vdraw::MoveTo(inset.x, inset.y+((3*inset.h)/4));
    vdraw::LineTo(inset.x+inset.w, inset.y+((3*inset.h)/4));
    vdraw::Stroke ();

    // bail here if no document to draw from
    if (trendSet == NULL)
    {
        vdraw::GRestore ();
	return;
    }

    for (i = 0; i < trendSet->numTrends; i++)
    {
	// Draw each enabled trend in the trend set
	if (trendSet->trends[i].enable && trendSet->trends[i].ctlEnable)
	{
	    DrawControls (&trendSet->trends[i], parent->fgColors[i], &inset);
	}
    }
    for (i = 0; i < trendSet->numTrends; i++)
    {
	// Draw each enabled trend in the trend set
	if (trendSet->trends[i].enable)
	{
	    DrawPlots (&trendSet->trends[i], parent->fgColors[i], &inset);
	}
    }
    vdraw::GRestore ();
    DrawCursor (parent->GetCursor());
}

void TrendFabric::DrawCursor (int loc)
{
    vrect   tfrect;
    vrect   inset;

    if (!IsOpen())
	return;

    vdraw::GSave ();
    DrawDevice ();
    // Clear the mainDraw area, draw the 3 horizontal lines
    tfrect = *GetRect();
    vdraw::Translate (tfrect.x, tfrect.y);
    tfrect.SetX(0);
    tfrect.SetY(0);
    inset.Inset(&tfrect, 2, 2);
    // set the color to red
    vdraw::SetColor(redColor);
    vdraw::SetLineWidth(1);
    vdraw::FSetDash(0, 0, 0.0);
    vdraw::NewPath();
    vdraw::MoveTo(inset.x + (inset.w * loc / 512), inset.y);
    vdraw::LineTo(inset.x + (inset.w * loc / 512), inset.y + inset.h);
    vdraw::Stroke ();

    vdraw::GRestore ();
}

void TrendFabric::DrawControls (TrendData *trendData, vcolor *color, vrect *inset)
{
    vdraw::SetLineWidth(1);
    vdraw::FSetDash(0, 0, 0.0);
    vdraw::SetColor(color);
    double vrange = trendData->dispHi
	- trendData->dispLo;
    if (vrange == 0) vrange = 1000000.;
    double vstep = (inset->h-1)/vrange;
    vdraw::MoveTo(inset->x, inset->y
	    + (int)((trendData->ctlLo - trendData->dispLo)
		* vstep));
    vdraw::LineTo(inset->x + inset->w, inset->y
	    + (int)((trendData->ctlLo - trendData->dispLo)
		*vstep));
    vdraw::Stroke ();
    vdraw::MoveTo(inset->x, inset->y
	    + (int)((trendData->ctlHi - trendData->dispLo)
		* vstep));
    vdraw::LineTo(inset->x + inset->w, inset->y
	    + (int)((trendData->ctlHi - trendData->dispLo)
		*vstep));
    vdraw::Stroke ();
}

void TrendFabric::DrawPlots (TrendData *trendData, vcolor *color, vrect *inset)
{
    int	    px;
    int	    py;
    int	    py2;

    vdraw::SetLineWidth(1);
    vdraw::FSetDash(0, 0, 0.0);
    vdraw::SetColor(color);
    double vrange = trendData->dispHi
	- trendData->dispLo;
    if (vrange == 0) vrange = 1000000.;
    int vsize = inset->h-1;
    int hsize = inset->w-1;
    if (vrange == 0) vrange = 1000000.;
    double vstep = vsize/vrange;
    double xstep = hsize/(double)C_TRENDSAMPLES;

    int xix = 0;
    for (int i = 0; i < C_TRENDSAMPLES; i++)
    {
        if ((trendData->status[i][0] != 'C') && (trendData->status[i][0] != 0xff)) // stuff to plot?
        {
            if (trendData->reduc == 4)
            {
                px = (int)(i*xstep + .5);
                py = vsize - (int) ((trendData->value1[i] - trendData->dispLo) * vstep + .5);
                py2 = vsize - (int) ((trendData->value2[i] - trendData->dispLo) * vstep + .5);
		vdraw::MoveTo(inset->x + px, inset->y + py);
		vdraw::LineTo(inset->x+px, inset->y + py2);
		vdraw::Stroke ();
            }
            else
            {
                if (xix == 0)
                {
                    px = (int)(i*xstep + .5);
                    py = (int) ((trendData->value1[i] - trendData->dispLo) * vstep + .5);
		    vdraw::MoveTo(inset->x + px, inset->y + py);
                    xix++;
                }
                else
                {
                    int dx = (int)(i*xstep + .5) - px;
                    int dy = (int) ((trendData->value1[i] - trendData->dispLo) * vstep + .5);
                    if ( dx || dy)
                    {
                        px += dx;
                        py = dy;
			vdraw::LineTo(inset->x + px, inset->y + py);
                        xix++;
                    }
                }
            }
        }
        else if (xix)
        {
	    vdraw::Stroke ();
            xix = 0;
        }
    }
    if (xix)
    {
	vdraw::Stroke ();
        xix = 0;
    }
}

class IconFabric : public vdialogItem
{
public:
    // constructor/destructors
    vloadableINLINE_CONSTRUCTORS (IconFabric, vdialogItem);

    // overrides from vloadable
    vkindTYPED_FULL_DECLARATION (IconFabric);

    // virtual members
    void Init ();
    void Draw (void);
    void DrawCursor (int loc);
    void DrawIcons (TrendData *trendData, vcolor *color, vrect *inset);
    void SetParent (TrendGraph *p) {parent = p; }
    int HvaeNotes () { return haveNotes; }
private:
    TrendGraph *parent;
    int haveNotes;
    vcolorSpec vred;
    vcolor *redColor;
};

vkindTYPED_FULL_DEFINITION (IconFabric, vdialogItem, "IconFabric");

void IconFabric::Init (void)
{
    vred.SetRGB (vcolorCOMPONENT_MAX, 0, 0);
    redColor = vcolor::Intern (&vred);
}

void IconFabric::Draw ()
{
    vrect   ifrect;
    vrect   inset;

    if (!IsOpen())
	return;

    TrendSet *trendSet = parent->GetTrendSet ();

    vdraw::GSave ();
    DrawDevice ();
    // Clear the iconDraw area, display notes symbols for enabled trends
    ifrect = *GetRect();
    vdraw::Translate (ifrect.x, ifrect.y);
    vdraw::SetLineWidth(2);
    ifrect.SetX(0);
    ifrect.SetY(0);
    vdraw::RectsHiliteFillInside
      (&ifrect, 1, parent->bgColor,
       DetermineColor(vdrawCOLOR_SHADOW_BOTTOM),
       DetermineColor(vdrawCOLOR_SHADOW_TOP));
    inset.Inset(&ifrect, 2, 2);

    haveNotes = FALSE;
    // bail here if no document to draw from
    if (trendSet == NULL)
    {
        vdraw::GRestore ();
	return;
    }

    for (int i = 0; i < trendSet->numTrends; i++)
    {
	// Draw each enabled trend in the trend set
	if (trendSet->trends[i].enable)
	{
	    DrawIcons (&trendSet->trends[i], parent->fgColors[i], &inset);
	}
    }
    vdraw::GRestore ();
    DrawCursor (parent->GetCursor());
}

void IconFabric::DrawCursor (int loc)
{
    vrect   ifrect;
    vrect   inset;

    if (!IsOpen())
	return;

    vdraw::GSave ();
    DrawDevice ();
    // Clear the mainDraw area, draw the 3 horizontal lines
    ifrect = *GetRect();
    vdraw::Translate (ifrect.x, ifrect.y);
    ifrect.SetX(0);
    ifrect.SetY(0);
    inset.Inset(&ifrect, 2, 2);
    // set the color to red
    vdraw::SetColor(redColor);
    vdraw::SetLineWidth(1);
    vdraw::FSetDash(0, 0, 0.0);
    vdraw::NewPath();
    vdraw::MoveTo(inset.x + (inset.w * loc / 512), inset.y);
    vdraw::LineTo(inset.x + (inset.w * loc / 512), inset.y + inset.h);
    vdraw::Stroke ();

    vdraw::GRestore ();
}

void IconFabric::DrawIcons (TrendData *trendData, vcolor *color, vrect *inset)
{
    char *notetext;

    // Draw indicators on icon area (if any are found, set notes flag!
    if (trendData->pointname [0] && (trendData->pointname [0] != '#'))
    {
        int hsize = inset->GetWidth ();
        double xscale = hsize/(double)C_TRENDSAMPLES;
	vdraw::SetColor(color);
	vrect temp = *inset;
        long notetime = trendData->starttime;
	if (GetTrendNote (trendData->pointname, &notetime, &notetext))
	{
	    unsigned long noteX = ((notetime - trendData->starttime)/trendData->zoom);
	    while (noteX < C_TRENDSAMPLES)
	    {
		haveNotes = TRUE;
		temp.SetX (inset->GetX () + (int)(noteX * xscale));
		vdrawRectShowCenterY ((unsigned char *)"N", &temp);// this shows the text
		if (NextTrendNote (&notetime, &notetext))
		{
		    noteX = ((notetime - trendData->starttime)/trendData->zoom);
		}
		else
		    break; // no more 
	    }
        }
    }
}

// macro definitions for vkindtyped overrides

vkindTYPED_FULL_DEFINITION(TrendGraph, vcontainer, NULL);

// vloadable
void TrendGraph::LoadInit(vresource resource)
{
    trendSet = NULL;
    
    leftTime = vtextitem::RequestSubLoad (this, vnameInternGlobalLiteral("LeftTimeTag"));
    rightTime = vtextitem::RequestSubLoad (this, vnameInternGlobalLiteral("RightTimeTag"));
    iconDraw = IconFabric::RequestSubLoad (this, vnameInternGlobalLiteral("IconDrawTag"));
    mainDraw = TrendFabric::RequestSubLoad (this, vnameInternGlobalLiteral("MainDrawTag"));

    vcontainer::LoadInit(resource);

    cursor = 0;

    iconDraw->SetParent (this);
    iconDraw->Init ();
    mainDraw->SetParent (this);
    mainDraw->Init ();
}

void TrendGraph::SetPrefs (vcolorSpec fgColorSpec[C_TRENDPTS], vcolorSpec bgColorSpec)
{
    int i;

    // convert to vcolor and store
    for (i = 0; i < 8; i++)
    {
	fgColors[i] = vcolor::Intern (&fgColorSpec [i]);
    }
    bgColor = vcolor::Intern (&bgColorSpec);
    Inval (0);
}

void TrendGraph::SetTheCursor (int pos)
{
    cursor = pos;
    Inval (0);
}

void TrendGraph::NewCursor (int pos) // override with position callback!
{
    // User code should adjust document for new position
    cursor = pos;
    Inval (0);
}

void TrendGraph::NewRange (int lpos, int rpos) // overide with expand event callback!
{
    // User code should adjust document for new width and position
}

void TrendGraph::DrawTimes (void)
{
    char    ptstring [40];
    char    cnvdate [20];
    char    cnvtime [20];

    // Set startTime to 0, change to timePoint if found
    startTime = trendSet->GetStartTime ();
    if (startTime)
    {
	//   (draw left and right times if timepoint was found)
        sprintf (ptstring, "%s %s",
		    LocalStrDate (startTime, cnvdate),
		    LocalStrTime (startTime, cnvtime));
	leftTime->SetTextScribed (vcharScribeSystem (&ptstring));
	time_t endTime = trendSet->GetEndTime ();
        sprintf (ptstring, "%s %s",
		    LocalStrDate (endTime, cnvdate),
		    LocalStrTime (endTime, cnvtime));
	rightTime->SetTextScribed (vcharScribeSystem (&ptstring));
    }
    else
    {
	// draw 0, 511
        StrSpan (-(long)cursor * trendSet->zoom, ptstring);
	leftTime->SetTextScribed (vcharScribeSystem (&ptstring));
        StrSpan ((511L-(long)cursor) * trendSet->zoom, ptstring);
	rightTime->SetTextScribed (vcharScribeSystem (&ptstring));
    }
}
