/* trendsaveasdlg.cxx,v 1.1.1.1 1999/08/03 14:45:10 tis4000 Exp
 *
 */

/*
 * TrendSaveAsDialog - trend graph container class
 */

#include <vport.h>		/* for v*HEADER */
#include vstdlibHEADER		/* for random, exit, EXIT_SUCCESS */
#include vlistHEADER
#include vstartupHEADER
#include vmemHEADER
#include vprefHEADER
#include vtextitemHEADER
#include vbuttonHEADER
#include vcharHEADER
#include vcomboboxHEADER
#include vdialogHEADER

#ifdef WIN32
#include <windows.h>
#include <winbase.h>
#include <stdio.h>
#endif WIN32

extern "C"
{
    #include "trendutils.h"
}
#include "trenddocs.h"
#include "trendsaveasdlg.h"
#include "tisgalaxylib.hxx"

// Help functions
class tsaAcceptButton : public vbutton {
public:
  // SaveAs for overrides from vkindTyped
  vkindTYPED_FULL_DECLARATION(tsaAcceptButton);
  
  // SaveAs for the standard versions of the 
  // necessary constructors and destructors
  vloadableINLINE_CONSTRUCTORS(tsaAcceptButton, vbutton);
  
  // virtual member function overrides
  void ObserveDialogItem(vevent *event);

  void SetParent (TrendSaveAsDialog *p) { parent = p; }
private:
    TrendSaveAsDialog *parent;
};

vkindTYPED_FULL_DEFINITION(tsaAcceptButton, vbutton, NULL);
void tsaAcceptButton::ObserveDialogItem (vevent *event)
{
    parent->AcceptClick();
}

class tsaCancelButton : public vbutton {
public:
  // SaveAs for overrides from vkindTyped
  vkindTYPED_FULL_DECLARATION(tsaCancelButton);
  
  // SaveAs for the standard versions of the 
  // necessary constructors and destructors
  vloadableINLINE_CONSTRUCTORS(tsaCancelButton, vbutton);
  
  // virtual member function overrides
  void ObserveDialogItem(vevent *event);

  void SetParent (TrendSaveAsDialog *p) { parent = p; }
private:
    TrendSaveAsDialog *parent;
};

vkindTYPED_FULL_DEFINITION(tsaCancelButton, vbutton, NULL);
void tsaCancelButton::ObserveDialogItem (vevent *event)
{
    parent->CancelClick();
}

// macro definitions for vkindtyped overrides
vkindTYPED_FULL_DEFINITION(TrendSaveAsDialog, vdialog, NULL);

// vloadable
void TrendSaveAsDialog::LoadInit(vresource resource)
{
    // get handles to interesting elements
    descText = vtextitem::RequestSubLoad (this, vnameInternGlobalLiteral("SubstTextTag"));
    convertToggle = vcontrolToggle::RequestSubLoad (this, vnameInternGlobalLiteral("ConvertTimesToggleTag"));
    accept = tsaAcceptButton::RequestSubLoad (this, vnameInternGlobalLiteral("OKButtonTag"));
    cancel = tsaCancelButton::RequestSubLoad (this, vnameInternGlobalLiteral("CancelButtonTag"));

    vdialog::LoadInit(resource);

    accept->SetParent (this);
    cancel->SetParent (this);
}

/*********
 *  TrendSaveAsDialog::ObserveWindow()
 *
 *  The dialog notification procedure gets called whenever any 
 *  major event applies to the dialog as a whole.  We need to
 *  look for the event which tells us that the dialog has
 *  closed.
 *
 **********/
void TrendSaveAsDialog::ObserveWindow (vevent *event)
{
    result = RESULT_CANCEL;
    if (event->GetType() == veventWINDOW_STATE && event->IsClose ())
    {
	vevent::StopProcessing ();
    }
    else
    {
	MakeTopMostDialog (this, event);
    }
}

/*********
 *  TrendSaveAsDialog::AcceptClick()
 *
 *  The procedure gets called whenever the accept button is clicked 
 *
 **********/
void TrendSaveAsDialog::AcceptClick ()
{
    descText->GetText ((unsigned char *)descriptor, sizeof(descriptor));
    convertAbs = convertToggle->GetValue();
    result = RESULT_OK;
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }
}

/*********
 *  TrendSaveAsDialog::CancelClick()
 *
 *  The procedure gets called whenever the cancel button is clicked 
 *
 **********/
void TrendSaveAsDialog::CancelClick ()
{
    result = RESULT_CANCEL;
    if (IsOpen())
    {
	Close ();
	vevent::StopProcessing ();
    }
}

int TrendSaveAsDialog::Alert (char *descript)
{
    // initialize the descText item
    descText->SetText ((unsigned char *)descript);
    convertToggle->SetValue (0);

    // Pop it up and handle events
    Open ();
    MakeTopMostDialog (this, NULL);
    // process events
    vevent::Process();

    // return answer
    return result;
}
