/*cmdClient.c,v 1.1.1.1 1999/08/03 14:45:12 tis4000 Exp */

/*	@(#)cmdClient.c	1.5 1/7/92
 *	Author:	Roger A. Cole
 *	Date:	11-26-90
 *
 *	Experimental Physics and Industrial Control System (EPICS)
 *
 *	Copyright 1991, the Regents of the University of California,
 *	and the University of Chicago Board of Governors.
 *
 *	This software was produced under  U.S. Government contracts:
 *	(W-7405-ENG-36) at the Los Alamos National Laboratory,
 *	and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *	Initial development by:
 *		The Controls and Automation Group (AT-8)
 *		Ground Test Accelerator
 *		Accelerator Technology Division
 *		Los Alamos National Laboratory
 *
 *	Co-developed with
 *		The Controls and Computing Group
 *		Accelerator Systems Division
 *		Advanced Photon Source
 *		Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .01	11-26-90	rac	initial version
 * .02	07-30-91	rac	installed in SCCS
 * .03  09-11-91	joh	updated for v5 vxWorks
 * .03	11-07-91	rac	be more forgiving of unexpected interactions
 * .04	12-06-91	rac	total rewrite, Unix-only, no LWP
 *
 * make options
 *	-DNDEBUG	don't compile assert() checking
 *      -DDEBUG         compile various debug code, including checks on
 *                      malloc'd memory
 */
/*+/mod***********************************************************************
* TITLE	cmdClient.c - general purpose client for command-based servers
*
* DESCRIPTION
*	Connects to a text-command-based server.
*
*	Usage on SunOS:
*		% cmdClient hostName portNum
*	    or
*		execl("cmdClient", "cmdClient", "hostName", "portNum",
*								(char *)0);
*
*	Usage on VxWorks:
*		> cmdClient "hostName",portNum
*
* BUGS
* o	the stdout stream from this program contains, intermixed, the
*	server's stdout and stderr streams
* o	not all signals are caught
*-***************************************************************************/

#include <stdio.h>
#include <ctype.h>
#ifdef SOLARIS
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#else
#include <strings.h>
#endif
#include <signal.h>
#include <setjmp.h>
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>

#include "genDefs.h"
#include "tsDefs.h"
#include "ezsSockSubr.h"
#include "msgMon.h"



/*-----------------------------------------------------------------------------
* prototypes
*----------------------------------------------------------------------------*/
int cmdcl();
void cmdclCmdProcess();
char *cmdclInTask();
long cmdclTask();
void cmdclTaskSigHandler();

/*-----------------------------------------------------------------------------
* global definitions
*----------------------------------------------------------------------------*/

jmp_buf	sigEnv;		/* environment for longjmp at signal time */

int main(int argc, char *argv[])
{
    char	*hostName;	/* host name for server */
    int		portNum;	/* port number for server */

    tisOpenLog ("cmdClient", 6, 2, NULL, NULL);
    if (argc != 3)		/* must be command and 2 args */
	goto mainUsage;
    hostName = argv[1];
    if (sscanf(argv[2], "%d", &portNum) != 1) {
	tisSysLog(0,0,"error scanning port number");
	goto mainUsage;
    }

    return cmdClient(hostName, portNum);

mainUsage:
    printf ("Usage: %s serverHostName serverPortNumber", argv[0]);
    return -1;
}

/*+/subr**********************************************************************
* NAME	cmdClient - shell callable interface for cmdClient
*
*-*/
int cmdClient(char *hostName, int portNum)
{
    int		serverSock = -1;	/* socket connected to server */
    char	serverBuf[80];
    char	*pBuf;
    FILE	*myIn=NULL;
    FILE	*myOut=NULL;
    char	*findNl;
    char	message[80];
    int		i;
    char	keyboard[80];	/* line from keyboard */
    int		eofFlag=0;
    int		discardNL=0;

    genSigInit(cmdclTaskSigHandler);

    if (setjmp(sigEnv) != 0)
	goto cmdclTaskWrapup;

/*-----------------------------------------------------------------------------
*    attempt to connect to the server.  If the connection is successful,
*    print the message returned by the server.  Then open two streams to
*    the socket, one for characters from keyboard to server, the other
*    for characters from server to stdout.
*----------------------------------------------------------------------------*/
    if ((i=ezsConnectToServer(&serverSock, portNum, hostName, serverBuf)) < 0) {
	if (i == -1) {
	    tisSysLog(0,0, "connect to %s port:%d (%d)", hostName, portNum, errno);
	    perror(message);
	}
	else
	    printf("%s\n", serverBuf);
	serverSock = -1;
	goto cmdclTaskWrapup;
    }
    tisSysLog(0,0,"%s", serverBuf);
    if (ezsFopenToFd(&myIn, &serverSock) < 0) {
	tisSysLog(0,0,"myIn %d",errno);
	myIn = NULL;
	goto cmdclTaskWrapup;
    }
    setbuf(myIn, NULL);
    if (ezsFopenToFd(&myOut, &serverSock) < 0) {
	tisSysLog(0,0,"myOut %d",errno);
	myOut = NULL;
	goto cmdclTaskWrapup;
    }

/*----------------------------------------------------------------------------
*    "processing loop"
*	do the interactions with the server, attempting as much as possible
*	to show the messages as they come in from the server.
*---------------------------------------------------------------------------*/
    while (!eofFlag) {
        if (cmdclInTask(keyboard) != NULL) {
	    fputs(keyboard, myOut);
	    fflush(myOut);
	}

	if (ezsCheckFpRead(myIn)) {
	    int		i, c;

	    i = 0;
	    do {
		if ((c = fgetc(myIn)) == EOF) {
		    eofFlag++;
		    break;
		}
		serverBuf[i++] = c;
		if (i >= 79 || c == '\n')
		    break;
	    } while (ezsCheckFpRead(myIn));
	    pBuf = (char *)&serverBuf[0];
	    serverBuf[i] = '\0';
	    if (i > 0) {
		char	*findChr;
#ifdef SOLARIS
		if ((findChr = strchr(pBuf, '#')) != NULL) {
#else
		if ((findChr = index(pBuf, '#')) != NULL) {
#endif
		    while (pBuf != findChr) {
			putchar(*pBuf);
			pBuf++;
		    }
		    if (strncmp(pBuf, "#p#r#", 5) == 0) {
			pBuf += 5;
			discardNL = 1;
		    }
		}
		else
		    pBuf = (char *)&serverBuf[0];
		if (discardNL) {
#ifdef SOLARIS
		    if ((findNl = strchr(pBuf, '\n')) != NULL) {
#else
		    if ((findNl = index(pBuf, '\n')) != NULL) {
#endif
			*findNl = '\0';
			discardNL = 0;
		    }
		}
		fputs(pBuf, stdout);
		fflush(stdout);
	    }
	    else {
		if (strncmp(keyboard, "quit", 4) == 0)
		    break;
		else if (strncmp(keyboard, "close", 5) == 0)
		    break;
		else
		    tisSysLog(0,0, "null message from server");
	    }
	}
#ifdef SOLARIS
	sleep (1); /* round out to a second in Solaris */
#else
	usleep(100000);
#endif
    }

cmdclTaskWrapup:
    if (myIn != NULL)
	fclose(myIn);
    if (myOut != NULL)
	fclose(myOut);
    if (serverSock >= 0)
	close(serverSock);

    return 0;
}

/*+/subr**********************************************************************
* NAME	cmdclTaskSig - signal handling and initialization
*
* DESCRIPTION
*	These routines set up for the signals to be caught for cmdclTask
*	and handle the signals when they occur.
*
* RETURNS
*	void
*
* BUGS
* o	not all signals are caught
* o	it's not clear how useful it is to catch the signals which originate
*	from the keyboard
*
*-*/
void cmdclTaskSigHandler(int signo)
{
    tisSysLog(0,0,"entered cmdclTaskSigHandler for signal:%d", signo);
    signal(signo, SIG_DFL);
    longjmp(sigEnv, 1);
}

/*+/subr**********************************************************************
* NAME	cmdclInTask - handle keyboard input
*
* DESCRIPTION
*	Checks to see if input is available and reads it, if so.  If
*	there is EOF, then question is asked whether to shut down the
*	server.
*
* RETURNS
*	char *, or
*	NULL
*
*-*/
char *cmdclInTask(char *buf)
{
    char	message[80];
    fd_set      fdSet;          /* set of fd's to watch with select */
    int         fdSetWidth;     /* width of select bit mask */
    struct timeval fdSetTimeout;/* timeout interval for select */

#ifdef SOLARIS
    { struct rlimit rl;
	  if (getrlimit (RLIMIT_NOFILE, &rl)== 0)
	      fdSetWidth = (int)rl.rlim_cur;
	  else
	      fdSetWidth = 0;
    }
#else
    fdSetWidth = getdtablesize();
#endif
    fdSetTimeout.tv_sec = 0;
    fdSetTimeout.tv_usec = 0;
    FD_ZERO(&fdSet);
    FD_SET(fileno(stdin), &fdSet);

    if (select(fdSetWidth, &fdSet, NULL, NULL, &fdSetTimeout) == 0)
	return NULL;
    if (fgets(buf, 80, stdin) == NULL) {
	strcpy(buf, "quit\n");
	clearerr(stdin);
    }
    if (strncmp(buf, "quit", 4) == 0) {
	(void)printf("stop the server (y/n) ? ");
	if (fgets(message, 80, stdin) == NULL) {
	    strcpy(message, "y");
	}
	if (message[0] == 'y' || message[0] == 'Y')
	    strcpy(buf, "quit\n");
	else
	    strcpy(buf, "\n");
    }

    return buf;
}
