/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/******************************************************************************
  ackgroup

  input :  the groupname to be acknowledged.
  output  :

********************************************************************************/   
#include <stdio.h>
#include <stdlib.h>  
#include <string.h>
#include <errno.h>
#include <time.h>
#if !(defined(AXP) || defined(WIN32))
#    include <sys/filio.h>
#endif
#ifndef	WIN32
#    include <sys/types.h>    /* Socket setup */
#    include <sys/socket.h>
#    include <netinet/in.h>
#    include <netdb.h>
#    include <sys/ioctl.h>
#    define SOCKET		int
#    define SOCKET_ERROR	-1
#    define INVALID_SOCKET	-1
#else
#    include <winsock.h>
#    include <fcntl.h>
#endif	!WIN32
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "dbDefs.h"
#include "alarmDefs.h"
#include "shareMem.h"    /* alarmMgr communication */
#include "alarmLib.h"
#include "ack.h"
#include "OperPriv.h"
#include "alarm.h"
#include "alarmString.h"
#include "messages.h"
#include "alarmvu.h"
#include "msgMon.h"

#include "SocketIO_alarmmgr.h"

/***********************************************************************************
  for each pvname in the group send an acknowledge message to alarmmgr, it
  the operator has ack privledges on the pvname.
************************************************************************************/
void main(int argc, char *argv[])
{
    SOCKET	alarmmgrSocket;
    char	groupname[GROUPNAME_SZ +1];
    int		operatorName[OPERATOR_SZ + 1];
    char	field[PVNAME_SZ + FLDNAME_SZ + 2];
    char	**PvNameList, **p;
    int		debug = 0;
    int		priv = PRIVOK;

    /* set debug to 1 if print messages are to be printed */
    if (argc > 2) debug = 1;

    /* get the group name to be acknowledged */
    /* if a parameter is passed in when ackgroup is invoked that is assumed to 
     * be to group name.  if no parameters are present check for the group 
     * invironment variable.
     */

    memset(groupname, 0, sizeof groupname);

    if (argc > 1)
	strncpy(groupname, argv[1], sizeof (groupname));
    else if (getenv("group"))  
	strncpy(groupname, getenv("group"), sizeof (groupname));
    else
	groupname[0] = 0;

    if (debug)
    {
	tisSysLog (0,0, "ACKGROUP for %s", groupname);
    }

    if (groupname[0] != 0)
    {
	/* open dbase files need to determine priv and to track operator.
	 * create and connect to the socket required for communication with
	 * alarmmgr. And initialize shared memory
	 */

	if ( ((priv = InitOperPriv((char *)(&operatorName[0]))) == PRIVOK) &&
	     ((priv = InitOperTracking((char *)(&operatorName[0]))) == PRIVOK) && 
	     (InitAlarmAccess()) && (init_ca(debug)) &&
	     (create_and_connect_socket(&alarmmgrSocket)) )
	{
	    /* let alarmmgr know that a group is being acknowledged */

	    if (ack_group(alarmmgrSocket, groupname, debug))
	    {
		/* get an array of all the pvnames in the group */

		if (PvNameList = p = GetPVNames(groupname))
		{
		    while (*p)
		    {
			/* for each point in the group ack it */

			sprintf( field, "%s.ACK", *p);

			/* if the operator has priv on the point, 
			 * the pvname must include the field,
			 * and the point is currently unacked, acknowledge it
			 */

			if ( (CheckOperPriv(field) == PRIVOK)  &&
			     (needs_acknowledgement(*p)) )
			{
			    ack_point(alarmmgrSocket, *p, 1, groupname, debug);
			}
			else if (debug)
			{
			    tisSysLog (0,0, "%s NOT acked", *p);
			}
			p++;

			if (!SharedMemorySegmentAvailable(alarmmgrSocket))
			{
			    tisSysLog (0,0, "Can not ackgroup %s- alarmMgr not accepting communcations", groupname);
			    break;
			}
		    }
		}

		free(PvNameList);
	    }
	} /* end big if */
	else
	{
	    if (priv != PRIVOK)
	    {
		tisSysLog (0,0, "Can not ack group %s, can not access privileges", groupname);
	    }
	    else
	    {
		tisSysLog (0,0, "Can not ack group %s, can not access resources", groupname);
	    }
	}
	close_ca ();
#ifdef	WIN32
	closesocket (alarmmgrSocket);
#else
	close (alarmmgrSocket);
#endif	WIN32
	CloseOperTracking((char *)(&operatorName[0]));
    } /* end if groupname[0] */
    else
    {
	if (debug)
	{
	    tisSysLog (0,0, "No group to ACK");
	}
    }
}

