/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/*********************************************************************************
*
*    alm
*
*    This utility gets the names of the points associated with this application
*    The groups associated with each point are derived from the pointnames.
*    If the groups are defined, alarms for the "group" are then tracked. 
*    If the alarm status is changed the point on the host is updated.
*
*    Meaning of the groupData.groupStatus (see main() )
*	status 0 - no alarm condition
*	status 1 - not returned to normal and Ack'd
*	status 2 - returned to normal but not Ack'd
*	status 3 - not returned to normal and not Ack'd
*
*    NOTE:
*	the form of the point name is:
*		alm:alarmgrpname.VAL
*	the entire name can not be > 27 characters, 1 less than PVNAME_SZ
*
*	the form for the point severity is:
*		alm:alarmgrpname.GRSV
*	the entire name can not be > 27 characters, 1 less than PVNAME_SZ
*
*    This utility also tracks the group severity
*	status = 0	severity = 0
*	status = 1	severity = 1 = INFO alarm
*				   2 = MINOR alarm
*				   3 = MAJOR alarm
*	status = 2	severity = 4 = INFO alarm
*				   5 = MINOR alarm
*				   6 = MAJOR alarm
*	status = 3	severity = 7 = INFO alarm
*				   8 = MINOR alarm
*				   9 = MAJOR alarm
*
*********************************************************************************/

/* Modification history
*
*  .01	mm-dd-yy	xxx	comment
*/

#include <stdio.h>
#include <stdlib.h>  
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#include <syslog.h>
#ifndef AXP
#include <sys/filio.h>
#endif
#ifndef	WIN32
#    include <sys/types.h>    /* Socket setup */
#    include <sys/socket.h>
#    include <netinet/in.h>
#    include <netdb.h>
#    include <sys/ioctl.h>
#    include <signal.h>
#    include <memory.h>
#    define SOCKET		int
#    define SOCKET_ERROR	-1
#    define INVALID_SOCKET	-1
#else
#    include <winsock.h>
#    include <fcntl.h>
#endif	!WIN32

#include "alarmDefs.h"
#include "shareMem.h"    /* alarmMgr communication */
#include "alarmLib.h"
#include "messages.h"
#include "alarmvu.h"
#include "cadef.h"
#include "agentMsgs.h"
#include "dpa.h"
#include "msgMon.h"

struct groupdata
{
    struct groupdata	*nextgroup;
    char		pvname[PVNAME_SZ+1];
    char		group[PVNAME_SZ+1];
    short		status;
    short		val_or_sevr;	/* val = 1, sevr = 0 */
    short		lastsevr;
    short		firstpass;
    void		*handle;
};

static SOCKET		alarmMgrFd;
static int		debug = 0;
static int		dpaFd;
static int		okToWrite = 0;
static int		signals[] = {SIGINT,SIGHUP,SIGQUIT,SIGTERM,SIGUSR1,SIGUSR2,0};
static struct groupdata	*FirstAlmGroup = (struct groupdata *)0;

static void	AlmClose (void *);
static void	AlmOpen (char *, void *),AlmRead(void *);
static void	AlmSignalHandler (int);
static void	Almfree_data (void);
static int	AlmConnectToAlarmMgr (void);
static void	AlmUpdatePoints (void);


/***************************************************************************/
void main(int argc, char *argv[])
{
    char		appName[64];
    char		*p;
    TypeOnlyMessage     alarmMgrMsg;
    fd_set		chanmask;
    int			status;
    int			numbytes;
#ifdef	WIN32
    WSADATA	wsaData
    WORD	wVersionRequested = MAKEWORD (1, 1);
    int		err;
#endif	WIN32

    tisOpenLog ("alm", 6, 2, NULL, NULL);

#ifdef	WIN32
    err = WSAStartup (wVersionRequested, &wsaData);
    if (err != 0)
    {
	tisSysLog (0,0, "could not start windows socket\n");
	exit (1);
    }
#endif

    if ( (p = (char *)strrchr (argv[0], '/')) == NULL)
	strcpy (appName, argv[0]);
    else
    {
	p++;
	strcpy (appName, p);
    }

    /* call the dpa init routine */
    if ( (dpaFd = dpaInit (appName, AlmOpen, AlmRead, 0, AlmClose, AlmSignalHandler, signals)) < 0)
    {
	tisSysLog (0,0, "dpaInit failed, will exit\n");
	exit (1);
    }

    /* set myself as MASTER (for redundancy) */
    dpaSetResMode (1);

    /* connect with the alarmmgr socket, keep trying until successful */
    if (AlmConnectToAlarmMgr() == 0)
    {
	tisSysLog (0,0,"unable to connect to alarmMgr, will exit\n");
	exit (1);
    }

    /* access the shared memory segment, keep trying until successful */
    if (InitAlarmAccess() == 0)
    {
	tisSysLog (0,0,"unable to access alarmMgr shared memory, will exit\n");
	exit (1);
    }

    while(1)
    {
	FD_ZERO(&chanmask);
	FD_SET(alarmMgrFd, &chanmask);
	FD_SET(dpaFd, &chanmask);

	/* wait here until we get either a message from dpa or the
	   alarmMgr
	*/
	status = select(FD_SETSIZE, &chanmask, NULL, NULL, NULL);
	if (status == -1)
	{
	    tisSysLog (0,0,"select error, will exit\n");
#ifdef	WIN32
	    closesocket (alarmMgrFd);
#else
	    close (alarmMgrFd);
#endif	WIN32
	    Almfree_data ();
	    exit (1);
	}
	
	/* dpa message, flush any thing in the buffer */
	if (FD_ISSET(dpaFd, &chanmask) )
	{
	    dpaEvent (0.1);
	}
	/* alarm mgr msg waiting */
	else
	{
	    numbytes = recv (alarmMgrFd, (char *)&alarmMgrMsg, sizeof (alarmMgrMsg), 0);

	    if (numbytes <= 0 )
	    {
		/* alarmMgr does not exist */
	        tisSysLog (0,0,"alarm mgr has disconnected, will exit\n");
#ifdef	WIN32
		closesocket (alarmMgrFd);
#else
		close (alarmMgrFd);
#endif	WIN32
		Almfree_data ();
	        exit (1);
	    }
	    switch (alarmMgrMsg.msgType)
	    {
		case UPDATE:
		case UPDATE_NEW_ALARM:
		    AlmUpdatePoints ();
		    break;

		case UNAVAILABLE: 
		/* alarmMgr is reconfiguring the shared memory segment.
		   Send UNAVAILABLE message back to alarmMgr and 
		*/
		    send(alarmMgrFd, (char *)&alarmMgrMsg, sizeof(alarmMgrMsg), 0);
		    break;
	    }
	}
    }
}


/********************************************
*
*    update dpa points if a chg has occured
*
********************************************/

static void AlmUpdatePoints ()
{
    struct groupdata	*currgroup;
    TreeNode		*node;

	if (okToWrite == 0)
	    return;
	currgroup = FirstAlmGroup;
	while (currgroup != (struct groupdata *)0)
	{
	    /* if the group is in the configuration */

	    if (node = (TreeNode *)FindGroup(NULL, currgroup -> group)) 
	    {
		if (currgroup -> val_or_sevr)
		{
		    /* determine if the points status has changed */

		    if ( (currgroup -> status != node -> groupData.groupStatus) ||
			 (currgroup -> firstpass) )
		    {
			currgroup -> status = node -> groupData.groupStatus; 
			if ( (currgroup -> status >= 0) && (currgroup -> status <= 3) )
			{
			    if (dpaWrite (currgroup -> handle, 
					  &currgroup -> status, 
					  sizeof (short), AGNT_VALUE, 
					  0, 0) == 0)
			    {
				currgroup -> firstpass = 0;
			    }
			}
		    }
		}
		else
		{
		    /* determine if group severity has changed */

		    if ( (currgroup -> lastsevr != node -> groupData.groupSeverity) ||
			 (currgroup -> firstpass) )
		    {
			currgroup -> lastsevr = node -> groupData.groupSeverity;
			if ( (currgroup -> lastsevr >= 0) && 
			     (currgroup -> lastsevr <= 9) )
			{
			    if (dpaWrite (currgroup -> handle,
					  &currgroup -> lastsevr,
					  sizeof (short), AGNT_VALUE, 
					  0, 0) == 0)
			    {
				currgroup -> firstpass = 0;
			    }
			}
		    }
		}
	    }
	    currgroup = currgroup -> nextgroup;
	}
}


/***********************************************
*
*   called when a points connection is broken
*
***********************************************/

static void AlmClose (void *handle)
{
    struct groupdata	*currgroup;
    int			found_handle = 0;

    okToWrite = 0;
    if (FirstAlmGroup == (struct groupdata *)0)
    {
	tisSysLog (0,0, "no handles to close (handle %0x)\n",handle);
	return;
    }
    else
    {
	currgroup = FirstAlmGroup;
	while (currgroup != (struct groupdata *)0)
	{
	    if (currgroup -> handle == handle)
	    {
		currgroup -> handle = 0;
		currgroup -> status = 0;
		currgroup -> lastsevr = 0;
		found_handle = 1;
		break;
	    }
	    currgroup = currgroup -> nextgroup;
	}
    }
    if (found_handle == 0)
	tisSysLog (0,0, "close on non existant handle %0x\n",handle);
    return;
}


/**********************************************************************
*
*    create a socket and connect to alarmMgr existing "message" socket.
*
***********************************************************************/

static int AlmConnectToAlarmMgr(void)
{
    struct hostent	*hostval;
    struct sockaddr_in	setPort;
    short		msgPort = MSGPORT;
    int			err;
 
    /* create the socket */
    if ((alarmMgrFd = socket(AF_INET, SOCK_STREAM, 0)) == SOCKET_ERROR) 
    {
#ifdef	WIN32
	err = WSAGetLastError ();
#else
	err = errno;
#endif	WIN32
	tisSysLog (0,0, "can not create socket to alarm mgr (%d), will exit\n",err);
	return (0);
    }
    /* make a connection with the alarmmgr existing socket */
    memset ((char*) &setPort, 0, sizeof(setPort));
    setPort.sin_family = AF_INET;
    hostval = gethostbyname ("localhost");
    if (hostval == NULL) 
    {
#ifdef	WIN32
	closesocket (alarmMgrFd);
	err = WSAGetLastError ();
#else
	close (alarmMgrFd);
	err = errno;
#endif	WIN32
	tisSysLog (0,0, "can not get hostname (%d), will exit\n", err);
	return (0);
    }
    setPort.sin_addr.s_addr =**((long **)(hostval -> h_addr_list));
    setPort.sin_port = htons(msgPort);
    if (connect(alarmMgrFd, (struct sockaddr *) &setPort, sizeof (setPort)) == SOCKET_ERROR) 
    {
#ifdef	WIN32
	closesocket (alarmMgrFd);
	err = WSAGetLastError ();
#else
	close (alarmMgrFd);
	err = errno;
#endif	WIN32
	tisSysLog (0,0, "can not connect to TCP port (%d)\n", err);
	return (0);
    }
    return (1);
}


/*********************
*
*    free up memory
*
*********************/

static void Almfree_data ()
{
    struct groupdata	*currgroup, *lastgroup;

    if (FirstAlmGroup != NULL)
    {
	currgroup = FirstAlmGroup;
	while (currgroup != (struct groupdata *)0)
	{
	    lastgroup = currgroup;
	    currgroup = currgroup -> nextgroup;
	    free ( (void*)lastgroup);
	}
    }
    return;
}


/******************************************
*
*    called when a point is to be opened
*
******************************************/

static void AlmOpen (char *pvname, void *handle)
{
    int			newname;
    char		tempPvname[PVNAME_SZ + 1];
    char		groupName[PVNAME_SZ + 1];
    char		*fieldName;
    struct groupdata	*newgroup, *currgroup, *lastgroup;

    okToWrite = 1;
    strncpy (tempPvname, pvname, PVNAME_SZ);
    strtok (tempPvname, ":");
    strncpy (groupName, strtok (NULL, "."), PVNAME_SZ);
    if ( (strlen (groupName) == 0))
    {
	tisSysLog (0,0, "pvname error, %s\n", pvname);
	return;
    }
    fieldName = strrchr (pvname, '.');
    if (fieldName)
	fieldName++;
    else
    {
	tisSysLog (0,0, "pvname error, %s\n", pvname);
	return;
    }
    if ( (newgroup = (struct groupdata *)calloc(1, sizeof (struct groupdata))) == NULL)
    {
	tisSysLog (0,0, "new group calloc error, program will exit\n");
	Almfree_data ();
	exit (1);
    }
    newgroup -> nextgroup = (struct groupdata *)0;
    strncpy (newgroup -> pvname, pvname, PVNAME_SZ);
    strncpy (newgroup -> group, groupName, PVNAME_SZ);
    if ( (strcmp (fieldName, "VAL")) == 0)
	newgroup -> val_or_sevr = 1;
    else if ( (strcmp (fieldName, "GRSV")) == 0)
	newgroup -> val_or_sevr = 0;
    else
    {
	tisSysLog (0,0, "pvname error, %s\n", pvname);
	free ( (void *)newgroup);
	return;
    }
    newgroup -> status = 0;
    newgroup -> lastsevr = NO_ALARM;
    newgroup -> firstpass = 1;
    newgroup -> handle = dpaSetUserField (handle, fieldName, sizeof(short),AGNT_SHORT, 1,0,0,0,0);
    if(debug)
    {
    	tisSysLog (0,0, "pv -%s-, group -%s-, fld = %d\n", 
	     newgroup->pvname,newgroup->group,newgroup->val_or_sevr);
    }
    if (FirstAlmGroup == (struct groupdata *)0)
	FirstAlmGroup = newgroup;
    else
    {
	currgroup = FirstAlmGroup;
	newname = 1;
	while (currgroup != (struct groupdata *)0)
	{
	    lastgroup = currgroup;
	    /* if alreaday in list do not add, but update handle, user may have
	       exited the graphic and then gone back in without getting out of
	       alm
	    */
	    if (strncmp (currgroup -> pvname, pvname, PVNAME_SZ) == 0)
	    {
		currgroup -> handle = newgroup -> handle;
		currgroup -> firstpass = 1;
		free ((void *)newgroup);
		newname = 0;
		break;
	    }
	    currgroup = currgroup -> nextgroup;
	}
	if (newname)
	    lastgroup -> nextgroup = newgroup;
    }

    return;
}


/**********************
*
*    process signals
*
***********************/

static void AlmSignalHandler (int which_one)
{
    time_t	end_time;
    int		do_exit = 0;

    end_time = time (0);

    switch (which_one)
    {
	case SIGINT:
	    tisSysLog (0,0, "exit due to ^C at %s\n", ctime (&end_time));
	    do_exit = 1;
	    break;

	case SIGHUP:
	    tisSysLog (0,0, "exit due to hangup at %s\n", ctime (&end_time));
	    do_exit = 1;
	    break;

	case SIGQUIT:
	    tisSysLog (0,0, "exit due to sigquit at %s\n", ctime (&end_time));
	    do_exit = 1;
	    break;

	case SIGTERM:
	    tisSysLog (0,0, "exit due to term at %s\n", ctime (&end_time));
	    do_exit = 1;
	    break;

	case SIGUSR1:
	    tisSysLog (0,0, "alm debug on\n");
	    debug = 1;
	    break;

	case SIGUSR2:
	    tisSysLog (0,0, "alm debug going silent\n");
	    debug = 0;
	    break;

	default:
	    break;
    }
    if (do_exit)
    {
	Almfree_data ();
	exit (0);
    }
}


/**************************************
*
*    this is needed to make dpa happy
*
**************************************/

static void AlmRead(void *handle)
{
    /*dpaFieldResponse(handle);*/
    AlmUpdatePoints ();
}

