/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <stdio.h>
#include <string.h>
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32
#ifdef	WIN32
#    include "NTUnixSubs.h"
     extern char *optarg;
#endif WIN32

/* Functions defined in this module */

static	void	dumpStructure (char *);

/* Data types use in this module */

#define	PATH_NAME_SIZE	100
typedef	char	PathName [PATH_NAME_SIZE+1];

/* Database directory */

static	PathName	dbDir;

/* Field specification file */

static	FILE		*specFile;

/* Output file */

static	FILE		*outFile;

/* Codebase related things */

CODE4	codebase;

/*----------*/
void main (int argc, char *argv [])

    {

	int	c;
	extern	char	*optarg;
	PathName	specFileName;
	char		line [500];

	/* Set up the defaults */

	strcpy (specFileName, "Filelist");
#ifdef WIN32
	strcpy (dbDir, "\\tis4000db\\");
#else
	strcpy (dbDir, "/tis4000db/");
#endif WIN32
	outFile = stdout;

	/* Process command line options */

	while ((c = getopt (argc, argv, "f:p:o:h")) != -1)
	    switch (c) {
		case 'f':	/* Specification file */
		    strcpy (specFileName, optarg);
		    break;
		case 'p':	/* Database directory */
		    strcpy (dbDir, optarg);
		    break;
		case 'o':	/* Output file */
		    outFile = fopen (optarg, "w");
		    if (outFile == NULL) {
			fprintf (stderr, "Unable to create %s\n", optarg);
			exit (-1);
		    }
		    break;
		case 'h':
		case '?':
		    fprintf (stderr, "Usage: dbstruct [-f <filelist>] [-o <outputfile>] [-p <directory>]\n");
		    fprintf (stderr, "  Default <filelist> is \"Filelist\"\n");
		    exit (0);
	    }

	/* Open the specification file */

	specFile = fopen (specFileName, "r");
	if (specFile == NULL) {
	    fprintf (stderr, "Unable to open file list \"%s\"\n", specFileName);
	    exit (-1);
	}

	/* Initialize the Codebase system */

	d4init (&codebase);
#if defined(USECB63) || defined(WIN32)
	codebase.errOpen = 0;
	codebase.errOff = 1;		/* disable codebase error msgs */
	codebase.autoOpen = 0;
#else
	codebase.open_error = 0;
	codebase.auto_open = 0;		/* Don't automatically open the index */
#endif	USECB63 or WIN32

	/* Read each line from the spec file and dump the information */

	while (fgets (line, sizeof (line), specFile)) {
	    line [strlen (line) - 1] = '\0';
	    dumpStructure (line);
	}

	/* Close the files and we're done */

	fclose (specFile);
	fclose (outFile);

	exit (0);


    }

/*----------*/
static void dumpStructure (char *line)

    {

	PathName	pathName, fileName;
	DATA4		*data;
	INDEX4		*indx;
	FIELD4INFO	*fieldInfo, *f;
	TAG4INFO	*tagInfo, *t;
	char		filter [200];

	/* Get the name of the database and open it and it's index */

	strcpy (fileName, strtok (line, ". "));
	sprintf (pathName, "%s%s" , dbDir, fileName);

	data = d4open (&codebase, pathName);
	if (data == NULL) {
	    fprintf (stderr, "Unable to open data file for %s\n", pathName);
	    return;
	}
	
	indx = i4open (data, pathName);
	
	/* Get the structure of the database */

	f = fieldInfo = d4field_info (data);
	if (fieldInfo == NULL) {
	    fprintf (stderr, "Unable to allocate memory for field information\n");
	    d4close (data);
	    exit (-1);
	}

	/* Output the structure */

	while (f->name) {
	    fprintf (outFile, "field: %s %s %c %d %d\n", fileName, f->name, f->type,
			f->len, f->dec);
	    f++;
	}

	/* Get the tags for the database (if any) and output them */

	if (indx) {
	    t = tagInfo = i4tag_info (indx);
	    while (t->name) {
		if (strlen (t->filter))
		    sprintf (filter, "\"%s\"", t->filter);
		else
		    strcpy (filter, "nofilter");
		fprintf (outFile, "tag: %s %s \"%s\" %s %s %s\n", fileName, t->name, t->expression,
			filter,
			t->unique ? "unique" : "notunique",
			t->descending ? "descending" : "ascending");
		t++;
	    }
	}

	/* Close the file and release the memory used by d4field_info */

	if (indx) {
	    u4free (tagInfo);
	    i4close (indx);
	}
	u4free (fieldInfo);
	d4close (data);

    }

