/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/*******************************************************************************
  lamptest <group> <pointname>

  This utility determines the lamp address for the group in the mapbddb dBase file.
  The lamp address is then "put" to the pointname using channel access.  The points
  should be set up as follows:

	<LAMP_TEST_ON>
	<LAMP_TEST_OFF>
	<LAMP_TEST_SLOW>
	<LAMP_TEST_FAST>
	<LAMP_TEST_RESTORE>

  where "putting" the lamp address to the point causes the indicated function.
*******************************************************************************/
#include <stdio.h>

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "cadef.h"
#include "db_access.h"
#include "msgMon.h"

#define  GROUP_SZ	16
short       lamp;
chid       pchid;

/*******************************************************************************
  get_lamp
********************************************************************************/
int    get_lamp(char *name)
{
int       success = 0;
C4CODE    cb;
D4DATA    *data;
char      dbname[300];

    d4init(&cb);

#if defined(USECB63) || defined(WIN32)
    cb.errOpen = 0;
    cb.errCreate = 0;
#else
    cb.open_error = 0;
    cb.create_error = 0;
#endif	USECB63

    memset(dbname, 0, sizeof dbname);
#ifdef	WIN32
    resolveLinks("\\tis4000db\\ioaddr.dbf", dbname);
#else
    resolveLinks("/tis4000db/ioaddr.dbf", dbname);
#endif	WIN32

    if (! (data = d4open(&cb, dbname)) )
    {
	tisSysLog(0,0, "ioaddr.dbf can not be opened");
	exit (1);
    }

    d4tag_select(data, d4tag(data,"NAME"));
    d4top(data);

    if (d4seek(data, name)==0) {

          lamp  = f4int(d4field(data, "SIGNAL"));
	  success = 1;
    }
    d4unlock_all(data);
    if (data) d4close(data);

    return success;
}


/*******************************************************************************
   testlamp
********************************************************************************/

void testlamp( struct connection_handler_args handlerargs)
{
     ca_put(OLD_DBR_SHORT, pchid, &lamp);
     ca_pend_io(3.0);
}
/********************************************************************************
  map_test
********************************************************************************/
main(int argc, char *argv[])

{
char    group[GROUP_SZ];
char    pointname[PVNAME_SZ];
int     success;
char    pvname[PVNAME_SZ];


    tisOpenLog ("lamp_test", 6, 2, NULL, NULL);

  if (argc == 3) { 
    strncpy(group , argv[1], GROUP_SZ);
    strncpy(pointname , argv[2], PVNAME_SZ);
    success = 1;
  }
  else if( getenv("G") && getenv("L") ) {

    strncpy(group , getenv("G"), GROUP_SZ);
    strncpy(pointname , getenv("L"), PVNAME_SZ);
    success = 1;
  }
  else 
    success = 0;

  /* build the pvname to be searched for in io_addr.dbf */
  sprintf(pvname,"MAP_%s_CTRL", group);

  if ( success && get_lamp(pvname))  {

     /*  setup channel access for point name */
     
     ca_task_initialize();

     if(ca_build_and_connect(pointname, TYPENOTCONN, 1, &pchid, 0, testlamp, 0)==ECA_NORMAL)
         ca_pend_io(3.0);
 }
}
