/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/***********************************************************************
  function :  launch

  parameters : key, this is used to identify the graphic name in the dbase file.
               [macro_replacement_sting] passed to dm.

  This utility accepts as a parameter a key, using this key the name of the
  corresponding graphic is retrieved from the dbase file.  The graphic is     
  launched using the standard dm command.  The second optional parameter is the 
  macro replacement string passed as dm second argument.

  K.GRIMES	7/94

*************************************************************************/
#include <stdio.h>
#include <errno.h>
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32
#ifdef	WIN32
#include <process.h>
#include <direct.h>
#endif
#include "launch.h"
#include "resolveLinks.h"
#include "msgMon.h"

/*********************************************************************************
 remove_trailing_blanks : given a string and its maximum remove trailing blanks 
			  returned from the database. 
*********************************************************************************/
void remove_trailing_blanks( char *from_string, int size)
{
    int   i;

    for (i=size-1; i>=0 && from_string[i]==' '; i--);
    from_string[++i] = '\0';

}
/**************************************************************************
  get the graphics name associated with the key
**************************************************************************/

int  get_graphic_name(char *key, char *position, char *graphic, char *envstr)
{
C4CODE       cb;
D4DATA       *data;
T4TAG        *tag;
int          success = 0;
char         dbname[300];
char	     paddedKey [KEY_SZ+1];

	/* pad key on right with blanks so that exact match will be made */

	sprintf (paddedKey, "%-*s", KEY_SZ, key);

	/* initialize codebase */

	d4init(&cb);

#if defined(USECB63) || defined(WIN32)
	cb.errOpen = 0;
#else
	cb.open_error = 0;
#endif	USECB63

	/* open file, get and select tag, positon to top of file */

	memset(dbname, 0, sizeof dbname);
#ifdef WIN32
	resolveLinks("\\tis4000db\\grphicdb.dbf", dbname);
#else
	resolveLinks("/tis4000db/grphicdb.dbf", dbname);
#endif WIN32
	if ( !(data = d4open(&cb, dbname)) ) {

	      /* the file does not exist create it and exit */
	      if (data = d4create(&cb, dbname, graphic_launch_fields, launch_tag) ) 
	          d4close(data);
	      return success;
        }

	tag = d4tag(data, "TAG");

	d4tag_select(data, tag);

	d4top(data);

	switch(position[0])
	{  
	   case 'P':

	          if ( d4seek(data, paddedKey) == 0)
		  {
		       strcpy(key, f4str(d4field(data, "PREV")));
		       sprintf (paddedKey, "%-*s", KEY_SZ, key);
		       d4top(data);

	               if ( d4seek(data, paddedKey) == 0)
		       {
                          strcpy(graphic, f4str(d4field(data, "FILE")));
			  strcpy(envstr, f4str(d4field(data, "PARAM")));
			  remove_trailing_blanks(graphic, FILE_SZ);
			  remove_trailing_blanks(envstr, 130);
			  remove_trailing_blanks(key, KEY_SZ);
	                  success = 1;
                       }
                  }
                  break;

           case 'M' :

	          if ( d4seek(data, paddedKey) == 0)
		  {
                     strcpy(graphic, f4str(d4field(data, "FILE")));
		     strcpy(envstr, f4str(d4field(data, "PARAM")));
		     remove_trailing_blanks(graphic, FILE_SZ);
		     remove_trailing_blanks(envstr, 130);
		     remove_trailing_blanks(key, KEY_SZ);
	             success = 1;
                  }
		  break;

           case 'N' :

		  if (d4seek(data, paddedKey) == 0)
		  {
		       strcpy(key, f4str(d4field(data, "NEXT")));
		       sprintf (paddedKey, "%-*s", KEY_SZ, key);
		       d4top(data);
	               if ( d4seek(data, paddedKey) == 0)
		       {
                          strcpy(graphic, f4str(d4field(data, "FILE")));
		          strcpy(envstr, f4str(d4field(data, "PARAM")));
			  remove_trailing_blanks(graphic, FILE_SZ);
			  remove_trailing_blanks(envstr, 130);
			  remove_trailing_blanks(key, KEY_SZ);
			  
	                  success = 1;
                       }
                  }
		  break;
        }

	d4close(data);
	return success;
}	    
/******************************************************************************/
void usage()
{
   printf ("usage :  launch  key  order   [macro_replacement_string] where");
   printf ("\tkey  is the index into the graphic table");
   printf ("\torder is a P, M, or N");
   printf ("\t[macro_replacement_string] is the string passed to dm, 130 chars max.");
}

/******************************************************************************/
void main(int argc, char *argv[])
{
char   filename[FILE_SZ+1];
char   name[FILE_SZ+1];
char   envstr[PARAM_SZ+1];
char   param[PARAM_SZ+1];
char   key[KEY_SZ+1];
#ifndef	WIN32
char   txtBuf[300];
#endif	!WIN32
int    numparams = 0;
int    exec = 0, i;
#ifdef	WIN32
char	savedir [200];
#endif

    tisOpenLog ("launch", 2, 6, NULL, NULL);

    /* build the parameters to invoke the script */

    memset(filename, 0, sizeof filename);
    memset(envstr, 0, sizeof envstr);
    memset(param, 0, sizeof param);
    memset(key, 0, sizeof key);

    if (argc > 2)
    {
       strncpy(key, argv[1], KEY_SZ);

       if ( (*key != '\0') &&
	    (get_graphic_name(key, argv[2], filename, envstr)) )
       {
	 /* the filename may be a command sequence such as alarmvu etc.
	  * and not a graphic
	  */
	 if (filename[0] == '!')
	 {
		exec = 1;
		strcpy(name, &filename[1]);  /* skip past the ! */
		strcpy(filename, name); 
         }
         if (argc > 3)
	 {
	     /*if the macro replacement string was passed in as a parameter use it*/
	     if (exec)
	     {
		for (i=3; i <= argc-1; i++)
		{
		    strncat(param, argv[i], PARAM_SZ - strlen(param));
		    strncat(param, " ", PARAM_SZ - strlen(param));
 	    	}
             }	
	     else
	     {
	     	sprintf(param, "KEY=%s,", key);
	     	strncat(param, argv[3], PARAM_SZ - strlen(param));
	     }
         } 
	 else if (envstr[0] != '\0')
	 {
	    /* otherwise if the macro string was in the database use it */

	    if (exec) 
	    	sprintf(param,"%s", envstr);
	    else
	    	sprintf(param,"KEY=%s,%s", key, envstr);
	 }
         else 
	 {
	    /* otherwise only the key will be passed */
	    if (!exec)
	    	sprintf(param,"KEY=%s", key);
	 }
	 numparams = 2;

       }
       else
       {
	  /* there is a graphic called ERROR */
	  strcpy(filename, "ERROR");
	  numparams = 1;
	  exec = 0;
       }
#ifdef	WIN32
       if (numparams == 2)
       {
	   if (exec)
	   {
	       _getcwd (savedir, sizeof (savedir));
	       _chdir ("\\tis4000\\tasks");
	       _spawnlp (_P_DETACH, filename, filename, param, NULL);
	       _chdir (savedir);
	   }
	   else
	   {
	       _getcwd (savedir, sizeof (savedir));
	       _chdir ("\\tis4000\\graphics");
	       _spawnlp (_P_DETACH, "dm", "dm", filename, param, NULL);
	       _chdir (savedir);
	   }
       }
       else if (numparams == 1)
       {
	   if (exec)
	   {
	       _getcwd (savedir, sizeof (savedir));
	       _chdir ("\\tis4000\\tasks");
	       _spawnlp (_P_DETACH, filename, filename, NULL);
	       _chdir (savedir);
	   }
	   else
	   {
	       _getcwd (savedir, sizeof (savedir));
	       _chdir ("\\tis4000\\graphics");
	       _spawnlp (_P_DETACH, "dm", "dm", filename, NULL);
	       _chdir (savedir);
	   }
       }
#else
       if (numparams == 2)
       {
	  if (exec)
       	  	sprintf(txtBuf, "%s %s &", filename, param);
          else
       	  	sprintf(txtBuf, "csh -c \"dm %s %s &\"", filename, param);
       }
       else if (numparams == 1)	
       {
	  if (exec)
       	   	sprintf(txtBuf, "%s &", filename);
   	  else
       	  	sprintf(txtBuf, "csh -c \"dm %s &\"", filename);
       }
       system(txtBuf);

#endif	WIN32
   }
   else
     usage();
}
