/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

#include  <stdio.h>
#include  <time.h>
#include  <errno.h>
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32
#include "script.h"
#include "msgMon.h"

/* Functions defined in this module */

static	int	openDatabase ();
static	int	lastDayOfMonth ();
static	int	firstSelectedDow ();
static	int	buildCommand ();
static	void	executeScript ();
static	void	removeTrailingBlanks ();


/* Misc definitions */

#define	NUM_PARMS	10

/* CodeBase related data */

C4CODE		cb;
D4DATA		*data;
F4FIELD		*scriptIdFld, *processFld, *shellFld, *descFld, *hourFld, *minuteFld;
F4FIELD		*domFld, *dowFld [7], *monthFld [12], *useFld, *parmFld [NUM_PARMS];

/* Local data */

static	struct tm	*timeNow;
 
/*----------*/
void main ()
    {
	char	command [500], key [5], domValue [20];
	time_t	clock;

	/* Open the database */

	tisOpenLog ("launchscript", 2, 6, NULL, NULL);

	if (!openDatabase ())
	    exit (1);

	/* Get the current date/time */

	time (&clock);
	timeNow = localtime (&clock);

	/* Build a key with the current hour and minute and seek to the record */

	sprintf (key, "%2d%2d", timeNow->tm_hour, timeNow->tm_min);
	if (d4seek (data, key) != 0)
	{
	    d4close (data);
	    exit (0);
	}
	d4unlock_all (data);

	/* Seek to each record with the current hour:minute and check to see if it is time */
	/* to fire off the script. */

	while ((timeNow->tm_hour == f4int (hourFld)) && 
	       (timeNow->tm_min == f4int (minuteFld)) && !d4eof (data))
	{
	    /* If the month field for this month is not selected then skip this entry. */

	    if (!f4true (monthFld [timeNow->tm_mon]))
	    {
		d4skip (data, 1L);
		d4unlock_all (data);
		continue;
	    }

	    /* if the day of month field is non blank the script is scheduled to 
	     * be run on a monthly basis.  Otherwise the script is run on a weekly to daily 
	     * basis.
	     */ 
 
	    strcpy (domValue, f4str (domFld));

	    if (strcmp (domValue, "LASTDOM  ") == 0)
	    {
		/* if timeNow is the last day of the month */
		if (timeNow->tm_mday ==  lastDayOfMonth (timeNow->tm_mon, timeNow->tm_year))
		{
		    if (buildCommand (command))
		    {
			executeScript (command);
		    }
		}
	    }
	    else if (strcmp (domValue, "FIRSTWDOM") == 0)
	    {
		/* if timeNow is the first selected day of the week this month */

		if ((timeNow->tm_mday <= 7) && (timeNow->tm_wday == firstSelectedDow ()) )
		{
		    if (buildCommand (command))
		    {
			executeScript (command);
		    }
		}
	    }
	    else if (strcmp (domValue, "LASTWDOM ") == 0)
	    {
		/* if timeNow is the last selected day of the week this month */

		if ((timeNow->tm_wday == firstSelectedDow ()) &&
			(timeNow->tm_mday > 
			(lastDayOfMonth (timeNow->tm_mon, timeNow->tm_year) - 7)))
		{
		    if (buildCommand (command))
		    {
			executeScript (command);
		    }
		}
	    }
	    else if (strcmp(domValue, "         ") == 0)
	    {
		/* if timeNow is a selected day of the week. */

		if (f4true (dowFld [timeNow->tm_wday]))
		{
		    if (buildCommand(command))
		    {
			executeScript (command);
		    }
		}
	    }
	    else if (timeNow->tm_mday == atoi (domValue) )
	    {
		/* if the DOM is timeNow */

		if (buildCommand (command))
		{
		    executeScript (command);
		}
	    }

	    d4skip (data, 1L);
	    d4unlock_all (data);

	}

	d4close (data);

	exit (0);

    }

/*----------*/
static int openDatabase ()

    {
	char	dbName [100], buf [10];
	int	i;
	static	char	*days [] = {"SUN", "MON", "TU", "WED", "TH", "FRI", "SAT"};
	static	char	*months [] = {"JAN", "FEB", "MAR", "APR", "MAY", "JUN",
				      "JUL", "AUG", "SEP", "OCT", "NOV", "DEC"};

	d4init (&cb);
#if defined(USECB63) || defined(WIN32)
	cb.errOpen = 0;
#else
	cb.open_error = 0;
#endif	USECB63

#ifdef WIN32
	resolveLinks("\\tis4000db\\scriptdb.dbf", dbName);
#else
	resolveLinks("/tis4000db/scriptdb.dbf", dbName);
#endif WIN32
	data = d4open (&cb, dbName);
	if (data == NULL)
	{
	    tisSysLog (0,0, "Unable to open %s (%d)", dbName, getpid ());
	    return 0;
	}

	d4tag_select (data, d4tag (data, "TAG"));
	d4top (data);

	/* Open all the fields */

	scriptIdFld = d4field (data, "SCRIPTID");
	processFld = d4field (data, "PROCESS");
	shellFld = d4field (data, "SHELL");
	descFld = d4field (data, "DESC");
	hourFld = d4field (data, "HOUR");
	minuteFld = d4field (data, "MINUTE");
	domFld = d4field (data, "DOM");
	for (i=0 ; i<7 ; i++)
	    dowFld [i] = d4field (data, days [i]);
	for (i=0 ; i<12 ; i++)
	    monthFld [i] = d4field (data, months [i]);
	useFld = d4field (data, "USE");
	for (i=0 ; i<NUM_PARMS ; i++)
	{
	    sprintf (buf, "P%d", i);
	    parmFld [i] = d4field (data, buf);
	}

	return 1;

    }

/*----------*/		
static int lastDayOfMonth (int month, int year)

    {
	static	int	daysInMonth [] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

	if (month == 2)
	{
	    if (((year % 4) == 0) && ((year % 100) != 0) || ((year % 400) == 0))
		return 28;
	    else
		return 29;
	}

	return daysInMonth [month];
    }

/*----------*/
static int firstSelectedDow ()

	/* Returns the first selected day of the week or a -1 is no days are selected */

    {

	int	i;

	for (i=0 ; i<7 ; i++)
	    if (f4true (dowFld [i]))
		return i;

	return -1;

    }

/*----------*/
static int buildCommand (char *command)

    {
	int   i;
	char  use [NUM_OPTIONS + 1];
	char  shell [SHELL_SIZE +1];
	char  process [PROCESS_SIZE+1];
	char  param [PARAM_SIZE+1];

	strcpy (command, "");
	strcpy (shell, f4str(shellFld));
	removeTrailingBlanks (shell);
	if (shell [0])
	{
	    if ((strcmp (shell,  "sh") == 0) || (strcmp (shell, "csh") == 0) || 
		(strcmp (shell, "ksh") == 0) )
	    {
		strcpy (command, shell);
		strcat (command, " ");
	    }
	}

	strcpy (process, f4str(processFld)); 
	if (process [0] != ' ')
	{
	    removeTrailingBlanks (process);
	    strcat (command, process);
            strcat (command, " ");
	    strcpy (use, f4str(useFld));
	    removeTrailingBlanks (use);
	    i = 0;
	    while (use [i])
	    {
		if ((use [i] != ' ') && (use [i] >= '0') && (use [i] <= '9') )
		{
		    strcpy (param, f4str(parmFld [use [i] - '0']));
		    removeTrailingBlanks (param);
		    if (param [0] != '\0')
		    {
			strcat (command, param);
			strcat (command, " ");
		    }
		}
		i++;
	    }
	    strcat (command, "&");
	    return 1;
	}
	else
	{
	    return 0;
	}

	return (0);

    }

/*----------*/
static void executeScript (char *command)

    {
	system (command);
    }

/*----------*/
static void removeTrailingBlanks (char *str) 	/* Strip trailing blanks */

    {
	int	i, size;

	size = strlen (str);
	for (i=size-1 ; (i >= 0) && (str [i] == ' ') ; i--);
        str [i + 1] = '\0';

    }

