/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include <Xm/Label.h>
#include <sys/types.h>
#include <sys/unistd.h>
#include <sys/stat.h>
#include <sys/procfs.h>
#include <fcntl.h>
#include <time.h>

#include "msgMon.h"

String fallbacks[] =
{
  "*background: gray",
  "*foreground: black",
  "*fontList: 8x13bold",
  "*geometry: 260x50",
  "*.title: Hathaway Industrial Automation",
  NULL
};

typedef struct ProcStruct {
    char procname [80];
    time_t  starttime;
    pid_t pid;
    char procentry [80];
} ProcData;

XtAppContext ctx;
int suspendCount;
int Debug;

XtTimerCallbackProc CheckProcess ( XtPointer data, XtIntervalId *id);
XtTimerCallbackProc CheckSuspend ( XtPointer data, XtIntervalId *id);


void main (int argc, char *argv[])
{
    Widget toplevel;
    Widget msg_widget;
    Arg wargs[5];
    int n;
    XmString message;
    ProcData procdata;
    Pixmap pixmap;

    tisOpenLog ("startMessage", 6, 2, NULL, NULL);

    if (argc < 3)
    {
	printf ("Usage: StartProcess <message> <program_image_name>\n");
	printf ("       Displays a message while starting program.\n");
	printf ("       If message starts with '.' - message is pixmap name\n");
	printf ("       If message starts with '#' - run in diagnostic mode\n");
	exit (1);
    }

    XtSetLanguageProc(NULL,NULL,NULL);
    toplevel = XtVaAppInitialize (&ctx, "message", NULL, 0, &argc, argv, fallbacks, NULL);
 
    strncpy (procdata.procname, argv[1], 80);
    procdata.starttime = time(0);
    procdata.pid = fork();
    if (procdata.pid == 0)
    {
	execvp (argv[2], &argv[2]);
    }
    sprintf (procdata.procentry, "/proc/%d", procdata.pid);

    n = 0;

    if ((argv[1])[0] == '#') /* debug switch */
    {
	Debug = 1;
	message = XmStringCreate (argv[1]+1, XmSTRING_DEFAULT_CHARSET);
	XtSetArg (wargs[n], XmNlabelString, message);
	n++;
    }
    else if ((argv[1])[0] == '.')
    {
	pixmap = XmGetPixmap (XtScreen(toplevel), argv[1]+1,
			 BlackPixelOfScreen (XtScreen(toplevel)),
			 WhitePixelOfScreen (XtScreen(toplevel)));
	if (pixmap == XmUNSPECIFIED_PIXMAP)
	{
	    tisSysLog(0,0,"can't create pixmap from %s\n", argv[1]+1);
	    message = XmStringCreate (argv[1]+1, XmSTRING_DEFAULT_CHARSET);
	    XtSetArg (wargs[n], XmNlabelString, message);
	    n++;
	}
	else
	{
	    XtSetArg (wargs[n], XmNlabelType, XmPIXMAP);
	    n++;
	    XtSetArg (wargs[n], XmNlabelPixmap, pixmap);
	    n++;
	}
    }
    else
    {
	message = XmStringCreate (argv[1], XmSTRING_DEFAULT_CHARSET);
	XtSetArg (wargs[n], XmNlabelString, message);
	n++;
    }

    msg_widget = XtCreateManagedWidget ("msg", xmLabelWidgetClass, toplevel, wargs, n);
    XtRealizeWidget (toplevel);

    /* first check for process status at 1/4 seconds */
    XtAppAddTimeOut (ctx, 250, CheckProcess, &procdata);

    XtAppMainLoop(ctx);
}

XtTimerCallbackProc CheckProcess ( XtPointer proc_data, XtIntervalId *id)
{
    int pfd;
    prpsinfo_t p;
    ProcData *procdata = (ProcData *)proc_data;

    if ((time(0) - procdata->starttime) > 60)
	exit (0);

    /* open process informatione pseudofile, bail on error */
    if ((pfd = open (procdata->procentry, O_RDONLY)) == -1)
    {
	tisSysLog (0,0,"CheckProcess can't open process info. (%m)");
	exit (0);
    }

    /* get data about spawned process, bail on error */
    if (ioctl (pfd, PIOCPSINFO, &p) == -1)
    {
	tisSysLog (0,0,"CheckProcess can't get process info. (%m)");
	exit (0);
    }

    if (Debug)
	tisSysLog (0,0,"Spawn proc status: %c\n", p.pr_sname);

    /* quit immediately on zombie! */
    if (p.pr_sname == 'Z')
    {
	/* Delay until suspended */    
	exit (1);
    }

    /* wait for a number of suspends, 1/4 sec apart */
    if (p.pr_sname == 'S')
    {
	/* Delay for repeated suspend */
	suspendCount = 0;    
	XtAppAddTimeOut (ctx, 250, CheckSuspend, procdata);
	return;
    }

    close (pfd);

    /* check back in a second */    
    XtAppAddTimeOut (ctx, 250, CheckProcess, procdata);
}

XtTimerCallbackProc CheckSuspend ( XtPointer proc_data, XtIntervalId *id)
{
    int pfd;
    prpsinfo_t p;
    ProcData *procdata = (ProcData *)proc_data;

    if ((time(0) - procdata->starttime) > 60)
	exit (0);

    /* open process informatione pseudofile, bail on error */
    if ((pfd = open (procdata->procentry, O_RDONLY)) == -1)
    {
	tisSysLog (0,0,"CheckSuspend can't open process info. (%m)");
	exit (0);
    }

    /* get data about spawned process, bail on error */
    if (ioctl (pfd, PIOCPSINFO, &p) == -1)
    {
	perror ("CheckSuspend can't get process info. (%m)");
	exit (0);
    }

    if (Debug)
	tisSysLog (0,0,"Spawn proc status: %c %d\n", p.pr_sname, suspendCount);

    /* quit immediately on zombie! */
    if (p.pr_sname == 'Z')
    {
	/* Delay until suspended */    
	exit (1);
    }

    /* exit if process is ready to run */
    if (p.pr_sname == 'S')
    {
	if (suspendCount >= 16)
	    exit (0);
	suspendCount++;
    }
    else
	suspendCount = 0;

    close (pfd);

    /* check back in 1 second */    
    XtAppAddTimeOut (ctx, 250, CheckSuspend, procdata);
}
