/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/

/***************************************************************** 
*
*    pidcnvrt
*	convert an existing pid.dbf file
*
*    NOTES:
*	We move the record from old to new by pieces because
*	the OUTCLAMPLO field was changed fro 6 to 7 and the
*	AUTOCLOSE field was changed from Logical to Character
*	of 10 in size
*
*****************************************************************/


#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32
#include <pid.h>

main ()
{
    char	pathbuff[128];
    char	systembuff[256];
    struct stat	stabuff;
    CODE4	cb;
    DATA4	*old_pid, *new_pid;
    FIELD4	*aclo_old, *aclo_new, *name_old, *name_new; 
    FIELD4	*clamplo_old, *clamplo_new, *outrev_old, *outrev_new; 
    FIELD4	*initact_old, *initact_new;
    char	*aclo_old_ptr, *aclo_new_ptr, *name_old_ptr, *name_new_ptr;
    char	*clamplo_old_ptr, *outrev_old_ptr, *outrev_new_ptr; 
    char	*initact_old_ptr, *initact_new_ptr;
    long	old_record_width;
    double	clamplo_val;

    d4init (&cb);
    cb.go_error = 0;
    cb.open_error = 0;
    cb.create_error = 0;
    cb.skip_error = 0;
    cb.field_name_error = 0;

    strcpy (pathbuff, "pid");
    if (!(old_pid = d4open (&cb, pathbuff)))
    {
	printf ("pid conversion error, failed to open %s, will exit\n", pathbuff);
	exit (1);
    }
    d4alias_set (old_pid, "X");
    aclo_old = d4field (old_pid, "AUTOCLOSE");
    if (f4type (aclo_old) == 'C')
    {
	printf ("pid conversion error, file has already been converted, will exit\n");
	exit (1);
    }
    strcpy (systembuff, "mkdir pidcnv");
    system (systembuff);
    strcpy (systembuff, "cp -p pid.dbf pidcnv/pid.dbf.sav");
    system (systembuff);
    strcpy (systembuff, "cp -p pid.mdx pidcnv/pid.mdx.sav");
    system (systembuff);
    strcpy (pathbuff, "pidcnv/pid");
    if (!(new_pid = d4create (&cb, pathbuff, pidFieldInfo, pidTagInfo)))
    {
	printf ("pid conversion error, failed to create %s, will exit\n", pathbuff);
	d4close (old_pid);
	exit (1);
    }

    name_old = d4field (old_pid, "NAME");
    name_new = d4field (new_pid, "NAME");
    clamplo_old = d4field (old_pid, "OUTCLAMPLO");
    clamplo_new = d4field (new_pid, "OUTCLAMPLO");
    outrev_old = d4field (old_pid, "OUTREVERSE");
    outrev_new = d4field (new_pid, "OUTREVERSE");
    aclo_new = d4field (new_pid, "AUTOCLOSE");
    initact_old = d4field (old_pid, "INITACT");
    initact_new = d4field (new_pid, "INITACT");
    old_record_width = d4record_width (old_pid);

    name_old_ptr = f4ptr (name_old);
    name_new_ptr = f4ptr (name_new);
    clamplo_old_ptr = f4ptr (clamplo_old);
    outrev_old_ptr = f4ptr (outrev_old);
    outrev_new_ptr = f4ptr (outrev_new);
    aclo_old_ptr = f4ptr (aclo_old);
    aclo_new_ptr = f4ptr (aclo_new);
    initact_old_ptr = f4ptr (initact_old);
    initact_new_ptr = f4ptr (initact_new);

    for (d4top (old_pid); !d4eof (old_pid); d4skip (old_pid, 1L))
    {
	d4append_start (new_pid, 0);
	d4blank (new_pid);
	bcopy (name_old_ptr, name_new_ptr, clamplo_old_ptr - name_old_ptr);
	bcopy (outrev_old_ptr, outrev_new_ptr, aclo_old_ptr - outrev_old_ptr);
	bcopy (initact_old_ptr, initact_new_ptr, old_record_width - (initact_old_ptr - name_old_ptr));
	clamplo_val = f4double (clamplo_old);
	f4assign_double (clamplo_new, clamplo_val);
	if ( (f4char (aclo_old) == 'F') || (f4char (aclo_old) == ' ') )
	    f4assign (aclo_new, "Manual");
	else
	    f4assign (aclo_new, "Automatic");
	d4append (new_pid);
	d4flush (new_pid);
    }

    d4close (old_pid);
    d4close (new_pid);
    printf ("pid conversion finished\n");

    exit (0);
}

