/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


#include <fcntl.h>
#include <unistd.h>
#include <sys/file.h>
#include <termios.h>
#include <signal.h>
#include <syslog.h>
#include <sys/ioctl.h>
#include <string.h>

#include "msgMon.h"

void serverInit(char *appName, char *dirPath, char *lockFile)
{
    int		fd, stat;
    char	buff[10];

/******
	Change our controlling TTY so that we don't pick up any
	stray signals.
******/

    fd = open("/dev/tty", O_RDWR);
    ioctl(fd, TIOCNOTTY, 0);
    close(fd);

/******
	Move to the defined directory so that the core file
	can be found in case we die an unnatural death.
******/

    if (chdir(dirPath) != 0) {
	tisSysLog (0,0, "Unable to find trend lock directory (%s)", dirPath);
	if (mkdir (dirPath, 0775) == 0)
	{
	    tisSysLog (0,0, " Lock directory %s created.", dirPath);
	}
	else
	{
	    tisSysLog (0,0, "Lock directory %s could not be created.", dirPath);
	    exit (-1);
	}
    }

/******
	Make a process group just for ourselves
******/

#if defined(SOLARIS) || defined(AXP)
    setpgrp ( );
#else
    setpgrp(0, getpid());
#endif

/******
	Insure that we are the only copy of the server running
******/

    fd = open(lockFile, O_RDWR | O_CREAT, 0640);
    if (fd < 0) {
	tisSysLog (0,0, "Unable to access lock file in %s\n", dirPath); 
	exit(-1);
    }
#ifdef SOLARIS
    if (stat = (lockf (fd, F_TLOCK, 0)))
#else
    if (stat = flock(fd, LOCK_EX | LOCK_NB))
#endif
    {
	tisSysLog (0,0, " Already running.\n");
	exit(-1);
    }

/******
	Record our PID in the lock file to make it easy to find
******/

    sprintf(buff, "%d\n", getpid());
    write(fd, buff, strlen(buff));
    close(fd);		/* I think that we keep the lock even if
			   the file is closed. */

}

