/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/


/**************************************************************************
 update_mapboard

 This utility gets the names of the points associated with the mapboard.
 The groups associated with each point are derived from the pointnames.
 If the groups are defined, alarms for the "group" are then tracked.  Every
 N seconds the status of each defined point is checked.  If the alarm status is
 changed the mapboard is updated.  The mapboard is updated as follows :
	if NOT RTN and UN ACK - fast blink
	if RTN and UN ACK     - slow blink
	if NOT RTN and ACK    - on
	if not in alarm       - off
 Update_mapboard handles the interrupt kill - ,  when the interrupt is 
 received update_mapboard terminates.
***************************************************************************/
#include <stdio.h>
#include <stdlib.h>  
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#ifndef AXP
#include <sys/filio.h>
#endif

#ifndef	WIN32
#    include <sys/types.h>    /* Socket setup */
#    include <sys/socket.h>
#    include <netinet/in.h>
#    include <netdb.h>
#    include <sys/ioctl.h>
#    include <signal.h>
#    define SOCKET		int
#    define SOCKET_ERROR	-1
#    define INVALID_SOCKET	-1
#else
#    include <winsock.h>
#    include <fcntl.h>
#endif	!WIN32

#ifdef WIN32
#pragma pack (push, 1)
#endif WIN32
#if defined(WIN32) || defined(USECB63)
#include "cb45to63.h"
#endif WIN32
#include "d4all.h"
#ifdef WIN32
#pragma pack (pop)
#endif WIN32

#include "alarmDefs.h"
#include "shareMem.h"    /* alarmMgr communication */
#include "alarmLib.h"
#include "ack.h"
#include "OperPriv.h"
#include "alarm.h"
#include "alarmString.h"
#include "messages.h"
#include "alarmvu.h"
#include "msgMon.h"
#include "SocketIO_alarmmgr.h"

#define	N		5         /*every N seconds the points status are checked*/
#define	WAITTIME	30        /* retry to get shared mem segment */

struct groupdata{
    char      group[PVNAME_SZ+1];
    short     status;
    chid      pchid;
    int       connected;
};

int    running=1;

/*********************************************************************************
 remove_trailing_blanks : given a string and its maximum remove trailing blanks 
			  returned from the database. 
*********************************************************************************/
void remove_trailing_blanks(char *from_string, int size)
{
int   i;

     for (i=size-1; i>=0 && from_string[i]==' '; i--);
     from_string[++i] = '\0';

}

/**************************************************************************
  set running to 0 to indicate interrupt has been received
***************************************************************************/
#ifndef	WIN32
void terminateProcessor( int sig)
{
  running = 0;
}
#endif	!WIN32
/***************************************************************************
 "put" the value determined by the status to the point.
***************************************************************************/
void update_mapbd( int status, chid pchid)
{
    switch(status) {
    case 0:          /* serv = 0  No Alarms*/
	ca_put(OLD_DBR_STRING, pchid, "OFF");
	break;
    case 1:          /* sevr > 0 and sevr < 4  Acked and Not RTN */
	ca_put(OLD_DBR_STRING, pchid, "ON");
        break;
    case 2:          /* sevr > 3 and sevr < 7  UnAcked and TRN*/
	ca_put(OLD_DBR_STRING, pchid, "SLOW FLASH");
	break;
    case 3:          /* sevr > 6  Unacked and Not RTN*/
	ca_put(OLD_DBR_STRING, pchid, "FAST FLASH");
	break;
    default:
	break;
    }
    ca_flush_io();
}
/**************************************************************************
  get_mapboard_pts
  determine the names of mapboard points.  For each point that is currently
  defined get the current status of the point, and set up for channel access.
***************************************************************************/
struct groupdata *get_groups(int *total)
{
    C4CODE	cb;
    D4DATA	*pvmap;
    int		i;
    chid	pchid;
    struct groupdata	*mapboard;
    char	name[PVNAME_SZ+1], tempname[PVNAME_SZ+1];
    char	dbname[300];

    d4init(&cb);
#if defined(USECB63) || defined(WIN32)
    cb.errOpen = 0;
#else
    cb.open_error=0;
#endif	USECB63

    memset(dbname, 0, sizeof dbname);
#ifdef WIN32
    resolveLinks("\\tis4000db\\pvmap.dbf", dbname);
#else
    resolveLinks("/tis4000db/pvmap.dbf", dbname);
#endif WIN32
    if (! (pvmap = d4open(&cb, dbname))) {
	/* pvmap does not exist */
	tisSysLog(0,0, "can not open pvmap.dbf\n");
	exit(1);
    }
    d4top(pvmap);
    d4unlock_all(pvmap);

    /* create an array big enough to hold group data for all mapboard points */
    i = d4reccount(pvmap);
    mapboard = (struct groupdata *)calloc(i, sizeof(struct groupdata));

    i = 0;
    while (!d4eof(pvmap)) {

	/* get the alarmgroup name */

	if (strncmp("MAPCTRL ", f4str(d4field(pvmap, "ALARMGROUP")), 8) == 0) {

       		/*Get the group associated with point name */
		/* The point name must have the following footprint
		   MAP_<groupname>_CTRL	*/

       		strcpy(name, f4str(d4field(pvmap, "NAME")));
		strcpy(tempname, name);
		d4unlock_all(pvmap);
       		remove_trailing_blanks(name, PVNAME_SZ);
		
		strtok(tempname, "_");
		strcpy(mapboard[i].group, strtok(NULL, "_"));
		if (mapboard[i].group == NULL) {
			d4skip(pvmap, 1);
			continue;
		}
	 	ca_search(name, &pchid);
	 	if (ca_pend_io(.5) == ECA_NORMAL) 
	     		mapboard[i].connected = 1;
         	else
	     		mapboard[i].connected = 0;

          	mapboard[i].status = -1;
	  	mapboard[i].pchid = pchid;

	  	i++;
    	}
       	d4skip(pvmap, 1);
	d4unlock_all(pvmap);
    }
    d4unlock_all(pvmap);
    d4close(pvmap);
    *total = i;
    return mapboard;
}
/***************************************************************************/
void main()
{
    struct groupdata *mapboard;

    int		i, total;
    TreeNode	*node;
    int		alarmMgrSocket;

    /* set up signal handler for kill signal 
       send signal as kill -[QUIT/TERM/INT] processid */
#ifndef	WIN32
    signal (SIGQUIT, terminateProcessor);    /*  3  -QUIT */
    signal (SIGTERM, terminateProcessor);    /* 15  -TERM */
    signal (SIGINT, terminateProcessor);     /*  2  -INT  */
#endif	!WIN32
       
    ca_task_initialize();

    /* connect with the alarmmgr socket, keep trying until successful */
    while (!create_and_connect_socket(&alarmMgrSocket))
	   sleep(WAITTIME);

    /* access the shared memory segment, keep trying until successful */
    while (!InitAlarmAccess())
	   sleep(WAITTIME);
    

    /* get the array of group names associated with mapboard */
    mapboard = get_groups(&total);
    
    if (total == 0)
 	exit(1);

    /* while the kill signal has not been trapped */
    while(running) {

       for ( i = 0; i < total; i++) {

	   /* if the group is in the configuration */

	   if (node = (TreeNode *)FindGroup(NULL, mapboard[i].group)) {

	       /* determine if the points status has changed since the last time the
		  mapboard was updated. */

	       if (mapboard[i].status != node->groupData.groupStatus) {

		   /* if the point has connected with channel access update the mapbd.
		      otherwise check for connection */
		   if (mapboard[i].connected) {
		       mapboard[i].status = node->groupData.groupStatus; 
                       update_mapbd(mapboard[i].status, mapboard[i].pchid);
                   }
                   else {
		       ca_flush_io();
		       if (ca_state(mapboard[i].pchid) == currOpen) 
		          mapboard[i].connected = 1;
                   }
               }
           }
         }

         if (!GetAlarmMgrMessage(alarmMgrSocket, N))
		exit(1);
  }
}
