/*drvSv85.c,v 1.1.1.1 1999/08/03 14:44:49 tis4000 Exp */

/****************************************************************/
/*								*/
/*	This software is the proprietary property of Tate	*/
/*	Integrated Systems, L.P.  Patents and copyrights are	*/
/*	pending.						*/
/*								*/
/*	This software is furnished under a license and may	*/
/*	be used and copied only in accordance with the terms	*/
/*	of such license and with the inclusion of the above	*/
/*	notice. This software or any other copies thereof	*/
/*	may not be provided or otherwise made available to	*/
/*	any other person. No title to and ownership of the	*/
/*	software is hereby transferred.				*/
/*								*/
/*	The information in this software is subject to		*/
/*	change without notice and should not be construed 	*/
/*	as a commitment by Tate Integrated Systems L.P. or	*/
/*	its third party suppliers.				*/
/*								*/
/*	Tate Integrated Systems, L.P. (TIS) and its third	*/
/*	party suppliers assume no responsibility for the	*/
/*	use or inability to use any of its software. TIS	*/
/*	software is provided "as is" without warranty of	*/
/*	any kind and TIS expressly disclaims all implied	*/
/*	warranties, including but not limited to the implied	*/
/*	warranties of merchantability and fitness for a		*/
/*	particular purpose.					*/
/*								*/
/*	Notice: Notwithstanding any other lease or license	*/
/*	that may pertain to or accompany the delivery of this	*/
/*	computer software, the rights of the Government		*/
/*	regarding its use, reproduction, and disclosure are	*/
/*	as set forth in Section 52.227-19 of the FARS		*/
/*	Computer Software-Restricted Rights clause.		*/
/*								*/
/****************************************************************/
/*
*  File: drvSv85.c
*  Edit:  2/25/93 12:33 JLD
*      : 12/08/93 10:15 WEP: comments only
*	: 12/16/93      WEP: half duplex mode with rts/cts delay and data tail squelch avoidance
*	: 12/17/93      WEP: dynamically allocate data communications buffers on a per line basis
*	: 12/21/93      WEP: dynamically allocate MODPLC structures on a per PLC basis
*	: 12/23/93      WEP: dynamically allocate output queue structures on a per line basis
*	:  1/03/94      WEP: mutual exclusion logic used when a diagnostic driver shares data path.
*	:  1/06/94      WEP: add pointname pointer to all driver/read functions. for error messages.
*	:  1/10/94      WEP: '#define' public routine names to make the driver cloning effort easier.
*	:  1/20/94	WEP: Duluth fix: added microsol support to sv85_aoread
*      :  2/21/94      WEP: take semaphore immediately after its creation.
*	:  3/30/94	ems: startsv85, STARTDRIVER2, remove test ln >= 0, it is unsigned
*			     msgxfer(), remove cast (char *), not needed
*      :  4/18/94      WEP: tasklocking added to improve IOC terminal messgae viewing.
*                         : implemented Polled Report By Exception Function.
*                         : added half duplex RTS control with timing set in call.
*                         : modified msgxfer RX logic to timeout on inactivity.
*                         : modified sv85_aidriver to mask out MSBit of statistics counts.
*                         : added string support - sv85_sodriver & sv85_sidriver (see devArcom.c).
*                         : modified sv85_aidriver to support non-block-aligned FLOAT,
*                           DOUBLE INTEGER, and DOUBLE FLOAT data types.
*                         : modified sv85_aodriver and sv85_aoread to support non-block-aligned
*                           ASCII FLOAT
*     :  05/05/94      WEP: moved while(TRUE) taskDelay logic to while(master) loop.
*                         : added taskDelay to while(pstate=slavewant).
*                         : change all failed msgxfer holdflag=FALSE to holdflag=TRUE.
*     :  05/06/94      WEP: modified loopback logic to correct performance statistics
*     :  06/06/94      WEP: modified msgxfer RX timing logic to handle long responses
*     :  06/08/94      WEP: modified rx_int_proc to reject pre-response false input.
*     :  06/14/94      WEP: modified register_resource to print pointname causing error.
*     :  06/22/94      WEP: corrected performance and use of the cmd_force_offline flag.
*     :  06/29/94      WEP: added printing of pstate & activity on initial master state.
*     :  07/07/94      WEP: added wait for pstate to become valid before deciding
*                         : whether to become master.
*     :  07/13/94      WEP: modified A0DRIVER, sv85_bodriver, & sv85_sodriver to increment the
*                         : 'badoutcnt' pseudo point counter when the output request is
*                         : rejected due to the output queue being full.
*     :  07/14/94      WEP: modified master state logic in sv85_driver to detect and recover
*                         : from the local and partner drivers being in lockstep.
*     :  07/15/94         : added test of is_responding flag in checkoutputs and, if
*                         : FALSE, discard the output queue entry.
*     :  07/18/94	WEP: modified scan_plc, readcoils, readdis, readais, readregs,
*                         : setcoil, setreg, & setstring to disable points and continue
*                         : processing after NACK received in response to Modbus cmd.
*     :  08/31/94	WEP: modified sv85_aodriver, sv85_bodriver, & sv85_sodriver to generate a
*                         : sv85traceon enabled diagnostic message whenever the output
*                         : buffer overflows on an output driver call.
*                         : removed mltrace details from the three output drivers while
*                         : adding diagnostic messages for driver call failures.
*     :  09/12/94	WEP: modified sv85_aodriver, sv85_bodriver, sv85_sodriver, & checkoutputs to
*                         : generate, at trace level 3, output queue and dequeue msgs.
*                         : added line level communication statistics to accumulate RX
*                         : interrupt counts, TX interrupt counts, and msgxfer counts.
*                         : added routines to display and clear the same counters.
*                         : added logic to measure time to scan the entire set of PLCs,
*                         : and the elapsed time since scanning the entire set of PLCs.
*                         : modified sv85_driver to generate a debug message, when sv85traceon,
*                         : displaying the elapsed time.
*     :  09/16/94	WEP: added calculated statistics to line level statistics above.
*     :  09/15/94	WEP: modified traffic trace logic to trace either all or a
*                         : single selected plc address, and either all or a single
*                         : selected modbus command type.
*     :  09/20/94	WEP: added modhelp routine to display this driver's utility info.
*     :  09/21/94	WEP: modified drivers to simulate responsive plcs. while in
*                         : simulation mode, no communications take place with the plcs,
*                         : and if in slave mode, the driver is forced to master mode.
*                         : added modsimon and modsimoff to enable & disable simulation.
*     :  12/05/94      WEP: modified dumpplcstats to mask out MSBit of stat counts.
*     :  01/16/95      WEP: implemented calls to watchdog timer logic in platform.
*     :  01/18/95      WEP: added explict disableable error messages to driver calls.
*     :  01/27/95      WEP: parameterized and renamed statistics control functions.
*     :  02/14/95      WEP: clarified and made conditional the illegal lstate message.
*     :  02/14/95      WEP: corrected pstatewait logic.
*                         :
*
*/                                                                                                                      

#include <vxWorks.h>
#include <netinet/tcp.h>
#include <inetLib.h>
#include <selectLib.h>
#include <ioLib.h>
#include <logLib.h>
#include <remLib.h>
#include <tickLib.h>
#include <string.h>
#include <sockLib.h>
#include <task_params.h>
#include <stdioLib.h>
#include <intLib.h>
#include <sysLib.h>
#include <pppLib.h>

#include <drvSup.h>
#include <restable.h>
#include <iocsiodvr_bridge.h>
#include <serialTask.h>
#include <resGroupTask.h>
#include <drvSv85.h>				/* NOTE: change here */
#include <drvSv85_proto.h>			/* NOTE: change here */

static	long init ();

	void	sv85_driver (unsigned short line, unsigned short address, unsigned short delay);

	int	sv85initsequence (unsigned short line, unsigned char *mbaddr, unsigned short doconfig);

static	void	sim_plc (MODPLC *pmod, unsigned short line, unsigned short mod);

static  int	checkresponses (unsigned short line);

static	unsigned short get_plc_stats (unsigned short line, unsigned short mod,
			 MODPLC *pmod);

static	unsigned short readcoilsfd(unsigned short line, unsigned short mod,
			 MODPLC *pmod, unsigned short bitaddr,
			 unsigned short readlen,unsigned short updateflag);

static	unsigned short readdisfd (unsigned short line, unsigned short mod,
			 MODPLC *pmod, unsigned short bitaddr,
			 unsigned short readlen, unsigned short updateflag);

static	unsigned short readaisfd (unsigned short line, unsigned short mod,
			 MODPLC *pmod, unsigned short regaddr,
			 unsigned short reglen, unsigned short updateflag);
static	unsigned short readregsfd (unsigned short line, unsigned short mod,
			 MODPLC *pmod, unsigned short regaddr,
			 unsigned short reglen, unsigned short updateflag);

static	unsigned short setstring (unsigned short line, unsigned short mod,
			          MODPLC *pmod, unsigned short signal,
			          unsigned char *buf, unsigned short olen);

	int	sv85putmstcmd (unsigned short line, unsigned char *buf);

	int	sv85getsvcreq (unsigned short line, unsigned char *buf);

	int	sv85abort (unsigned short line, unsigned char path);

	int	sv85getmstrsp (unsigned short line, unsigned char path, unsigned char *buf);

static  int	clearputpath (unsigned short line);

static	int	process_paths (unsigned short line, unsigned short mask);

	unsigned char sv85msgxfer (unsigned short line, unsigned char cmd , char *buf,
		unsigned short tics);
	int	sv85reset (unsigned short line);

	void	sv85initduarts (unsigned short line);

	int	sv85init (unsigned short line);

	int	sv85infcdiag (unsigned short line);

static 	void on_line (unsigned short line,unsigned short mod);

static 	void mark_not_valid (MODPLC  *pmod);

static	void scan_plc (MODPLC *pmod, unsigned short line,unsigned short mod);

static	void checkoutputs (unsigned short line);
static void  register_resource (unsigned short link, MODPLC *pmod,
                                unsigned char *resname, char *pointname);

static	void outqueinit (unsigned short line);

static	OUTQENT *ogetemptyp (unsigned short line);

static	void omarkfull (unsigned short line);

static	OUTQENT *ogetnextp (unsigned short line);

static	void oreleasefull (unsigned short line);

static 	int allocate_plc (unsigned short line, unsigned short plc);

static	int check_line_activity (unsigned short line, unsigned short predelay);

static	unsigned short msgxfer (unsigned short fline,unsigned short resplen,
        MODPLC *pmod, unsigned short rxlength);

static	void init_plc_tab (unsigned short line, unsigned short mod);

static 	void initplc (MODPLC  *pmod);

static	void initstats (MODPLC *pmod);

static	unsigned short modloopback (unsigned short line, unsigned short mod, MODPLC *pmod);

static	void prepresponses (unsigned short line);

static	int any_responding (unsigned short line);

static	unsigned short setcoil (unsigned short line, unsigned short mod,
		       MODPLC *pmod, unsigned short bitaddr, unsigned long mask,
		       unsigned long val);

static	unsigned short setreg (unsigned short line, unsigned short mod,
		       MODPLC *pmod, unsigned short regaddr, unsigned short val,
		       double dblval, unsigned short conv);

static 	int loadinit (unsigned short line,MODPLC *pmod,unsigned short mod);
static 	int loadnextblock (unsigned short line);

static 	int loadlogic (LOADINFO *plinfo);

static 	int loadmemory (LOADINFO *plinfo, unsigned short type);

static 	terminateload (LOADINFO *plinfo,unsigned short line, unsigned short success,
        unsigned short code);

static 	int plc_logout (LOADINFO *plinfo);

static 	int plc_login (LOADINFO *plinfo);

static 	int plc_start (LOADINFO *plinfo);

static 	int plc_stop (LOADINFO *plinfo);

static 	int pollfordone (LOADINFO *plinfo, int subfunction, unsigned short geterrcode);

static 	int plc_init (LOADINFO *plinfo);

static 	int plc_getdim (LOADINFO *plinfo);

static 	int plc_setdim (LOADINFO *plinfo);

static 	int setrunmode (LOADINFO *plinfo, short value);

static 	int getrunmode (LOADINFO *plinfo, short *pvalue);

static 	int get_plc_state (LOADINFO *plinfo, short *pvalue);

static 	int readmem (LOADINFO *plinfo, unsigned short page, unsigned short address,
	    unsigned short datalen, char *databuf);

static  int writemem (LOADINFO *plinfo, unsigned short page, unsigned short address,
	    unsigned short datalen, char *databuf);

static 	int writelogic (LOADINFO *plinfo, unsigned short page, unsigned short address,
	    unsigned short datalen, char *databuf);

static 	int writeascii (LOADINFO *plinfo, unsigned short page, unsigned short address,
	    unsigned short datalen, char *databuf);

static	void makecrc (unsigned char *chkbuf, int len,char unsigned *retbuf);

static 	void printcoilhdr (unsigned char *buffer, unsigned short cnt);

static 	void printdihdr (unsigned char *buffer, unsigned short cnt);

static	void printaihdr (unsigned char *buffer, unsigned short cnt);

static 	void printreghdr (unsigned char *buffer, unsigned short cnt);

	void sv85_int_proc (unsigned short line);

	int sv85SockCreate (unsigned short line);

	void sv85UDPCheck (unsigned short line);

	int haveCompleteSv85Msg (unsigned int line);

	int getSv85Bytes (char *buff, int nbytes, unsigned int line);

	int putSv85Bytes (short msgType, char *buff, int nbytes, unsigned int line);

	int sv85Flush (unsigned int line);

	void redSv85Task (unsigned short line);



struct
{
	long		number;
	DRVSUPFUN	report;
	DRVSUPFUN	init;
} drvSv85 =
	{
		2,
		NULL,
		init
	};


static	char		lastsvcreq [4];

static	unsigned short	sv85xfertrace [1024];

static	unsigned short	sv85tix;

/* Delay between commands until loaded */

static short		needdelay [MODLINES];

static short		sv85_needs_init [MODLINES] = {TRUE, TRUE, TRUE, TRUE};

static LPATHS		*lpath [MODLINES] = {NULL, NULL, NULL, NULL};

static unsigned char	*bases [MODLINES] =
{
	(unsigned char *) SV85L1BASE,
	(unsigned char *) SV85L2BASE,
	(unsigned char *) SV85L3BASE,
	(unsigned char *) SV85L4BASE
};

static unsigned		vectors [MODLINES] =
{
	SV85L1VEC,
	SV85L2VEC,
	SV85L3VEC,
	SV85L4VEC
};

static long		masks [] =
{
	0x00000001,
	0x00000002,
	0x00000004,
	0x00000008,
	0x00000010,
	0x00000020,
	0x00000040,
	0x00000080,
	0x00000100,
	0x00000200,
	0x00000400,
	0x00000800,
	0x00001000,
	0x00002000,
	0x00004000,
	0x00008000,
	0x00010000,
	0x00020000,
	0x00040000,
	0x00080000,
	0x00100000,
	0x00200000,
	0x00400000,
	0x00800000,
	0x01000000,
	0x02000000,
	0x04000000,
	0x08000000,
	0x10000000,
	0x20000000,
	0x40000000,
	0x80000000
};

static unsigned char	mbplusaddr [MODLINES];
static short		resp [MODLINES];
static long		SVCDelay;
static unsigned short	putpath [MODLINES];		/*Path in use flags*/
static short		intrs [MODLINES];

static SEM_ID		redundSem [MODLINES];		/* Wakes up redundancy task */
static int		redFd [MODLINES];		/* fd for redundant update */
static char		*sockBuff [MODLINES];		/* For tx/rx on UDP, tx only on TCP */
static int		sockBytes [MODLINES];		/* # bytes in sockBuff */
static long		partnerInet [MODLINES];		/* 0 = unknown */
static int		redRxBuffmax [MODLINES];	/* Has Rx buffer topped out ? */
static int		tcpRxBuffBytes [MODLINES];	/* # bytes in tcpRxBuff */
static char		*tcpRxBuff [MODLINES];		/* For rx only on TCP */
static long		lastSeqTx [MODLINES];		/* Last sequence number sent 118 AI*/
static int		redTxBuffsize [MODLINES];	/* Current Tx buffer size for redFd */
static int		redRxBuffsize [MODLINES];	/* Current Rx buffer size for redFd */
static	int		udpFd [MODLINES];		/* fd for UDP socket */
static int		tcpFd [MODLINES];		/* fd where slave connects */
static char		lastStrobeRx [MODLINES];	/* null, 'M' or 'S' */
static int		tcpRxBuffIdx [MODLINES];	/* Next byte to be output */
static long		lastSeqRx [MODLINES];		/* Last sequence number ack'd 119 AI*/
static long		lastSeqRxTime [MODLINES];	/* From TickGet () */
static long		lastStrobeRxTime [MODLINES];
static int		partnerAlive [MODLINES];

/* array of resource name strings. */

static	unsigned char	resnames [MODLINES] [RESNAMELEN];

/* array of pointers to resource tables */

static	RES_TABLE	*prestable [MODLINES];

/* array of resource response flags. */
/*	False = not active resource (slave), */
/*	True = active resource (master) */

static	unsigned short	resresponse [MODLINES];

/* true / false line active flag. calculated */
/* using activity_time below. */

static	unsigned short	line_activity [MODLINES];

/* assigned value of line_activity prior to */
/* it's recalculation. */

static	unsigned short	lastactivity [MODLINES];

/* controls output of communication error log */
/* messages (true / false) */

static	unsigned short	modtrace [MODLINES];

/* controls output of certain resource related */
/* log messages (level) */

static	unsigned short	mltrace [MODLINES];

/* controls output of MASTER/SLAVE arbitration */
/* log messages (binary) */

static	unsigned short	modrestrace [MODLINES];

/* 'new' semaphore logic. the old SEMAPHORE is */
/* still used. */

static	SEM_ID		semids [MODLINES];

/* holds value passed in task spawn. used for */
/* each msgxfer attempt */

static	unsigned short	intermsgdelay [MODLINES];

/* holds post RTS delay before TX. */

static	unsigned short	pretxdelay [MODLINES];

/* holds post TX delay before dropping RTS. */

static	unsigned short	posttxdelay [MODLINES];

/* holds timetick of last activity. used to */
/* set line_activity flag */

static	long		activity_time [MODLINES];

/* store half/full duplex line status */

static  unsigned short  halfduplex [MODLINES];

/* holds max time in seconds that driver will */
/* wait to obtain partner state on first */
/* attempting to decide if it should be master. */

static  unsigned short  pstatewait [MODLINES];

/*
*      Storage for MODBUS Comm Line Structures
*
*/

/* array of pointers to dynamically */
/* allocated comm controls/buffers */

static RXLINE          *rxlinetab [] = 
{
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL
};

/* array of pointers to dynamically */
/* allocated comm controls/buffers */

static TXLINE          *txlinetab [] = 
{
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL
};


/*
*      Storage for MODBUS PLC Structures
*
*/

/* 0 = normal operation, 1 = plc simulation.  */

static unsigned int     modsim [MODLINES];

static unsigned int	simdelay [MODLINES];

static 		int	simlow [MODLINES];

static 		int	simhigh [MODLINES];

static		int	simdigon [MODLINES];

static unsigned int     tick [MODLINES][5];

/* holds timestamp of beginning of major loop. */

static unsigned int     lastscan [MODLINES];

/* holds actual time taken to scan plcs. */

static unsigned int     scantime [MODLINES];

static unsigned int     loopoutmsgs [MODLINES];

static unsigned int     looptotmsgs [MODLINES];

static unsigned int     loopmsgs [MODLINES];

/* array of line level statistics. */

static LINESTATS	linestats [MODLINES];

/* array of control structures for loading PLC headers */

static LOADINFO		linfo [MODLINES];

/* two dimensional array of pointers to the */
/* fixed length dynamically allocated PLC */
/* control and data structures. extra pointer */
/* is for Query structure used independently */
/* of the others. it also allows for the array */
/* index to match the mod value. */

static MODPLC		*plcs [MODLINES] [NUMMODPLCS + 1];

/* controls output of system error messages */
/* bit mapped controls */

static	unsigned short	trace [MODLINES];
static  unsigned char   traceadr [MODLINES];
static  unsigned char   tracetype [MODLINES];

/* array of ptrs to the line output queues and */
/* their control structures */

static OUTPUTQUE	*oq [MODLINES];

static unsigned long	nextouttime [MODLINES];

static unsigned long	outagain [MODLINES];

/* holds id and password of the user that */
/* spawned the driver task required by host to */
/* check file read privileges on PLC load files */

static char		user [40], password [40];

/* store half/full duplex line status */

static int		taskId [MODLINES];
static int		redtaskId [MODLINES];


static char *tasknames [] =
{
	"SV85 Ch0", "SV85 Ch1", "SV85 Ch2", "SV85 Ch3",
	"SV85 Ch4", "SV85 Ch5", "SV85 Ch6", "SV85 Ch7",
	"SV85 Ch8", "SV85 Ch9", "SV85 Ch10", "SV85 Ch11",
	"SV85 Ch12", "SV85 Ch13", "SV85 Ch14", "SV85 Ch15",
	"SV85 Ch16", "SV85 Ch17", "SV85 Ch18", "SV85 Ch19",
	"SV85 Ch20"
};

static char *redtasknames [] =
{
	"SV85 R0","SV85 R1","SV85 R2","SV85 R3",
	"SV85 R4","SV85 R5","SV85 R6","SV85 R7",
	"SV85 R8","SV85 R9","SV85 R10","SV85 R11",
	"SV85 R12","SV85 R13","SV85 R14",
	"SV85 R15","SV85 R16","SV85 R17",
	"SV85 R18","SV85 R19","SV85 R20"
};

static	unsigned long		stat_refresh_rate [MODLINES];
static	unsigned long		path_timeout [MODLINES];
static	unsigned long		slv_want_tmout [MODLINES];
static	unsigned short		sv85nodetimeout = 3; /* Seconds before offline */

/**********
*
*	This routine is required by iocinit in the EPICS system.
*
**********/
static long init ()
{
	return (0);
}

/********
*
*	This routine outputs information to help with debugging.
*
********/
void sv85help (void)
{
    printf (" Modbus Driver Control & Display Functions (* = default)\n");
    printf ("\n");
    printf (" modtraceon(N)         Enable the following debug messages for line N:             \n");
    printf ("                         1) Major Scan loop Info:                                  \n");
    printf ("                           scan time, total output messages, and total messages.   \n"); 
    printf ("                         2) Output Queue Overflow.                                 \n");
    printf ("                         3) All communication errors encountered in msgxfer.       \n");
    printf ("*modtraceoff(N)        Disable all of the modtraceon debug messages for line N.    \n");
    printf ("\n");
    printf (" modrestrace(N,X)      Resource Trace Level: 0 = off. 1 = state change; 2 = delay. \n");
    printf ("*modrestraceon(N)      Enable display of master/slave/etc state changes in sv85_driver.   \n");
    printf (" modrestraceoff(N)     Disable display of master/slave/etc state changes in sv85_driver.  \n");
    printf ("\n");

    taskDelay (1 * 60); /* Time for operator to see it and pause the output */

    printf (" modcontraceon(N)      Enable display of configuration errors in driver calls.     \n");
    printf ("*modcontraceoff(N)     Disable display of configuration errors in driver calls.    \n");
    printf (" modruntraceon(N)      Enable display of runtime errors (queue overflow).          \n");
    printf ("*modruntraceoff(N)     Disable display of runtime errors (queue overflow).         \n");
    printf (" modrejecttraceon(N)   Enable display of driver call rejections in driver calls.   \n");
    printf ("*modrejecttraceoff(N)  Disable display of driver call rejections in driver calls.  \n");
    printf (" modnacktraceon(N)     Enable display of point deselection on MODBUS command NACK. \n");
    printf ("*modnacktraceoff(N)    Disable display of point deselection on MODBUS command NACK.\n");
    printf ("\n");

    taskDelay (1 * 60); /* Time for operator to see it and pause the output */

    printf (" modtrlv(N,X)          Enable the following message for line N:                  \n");
    printf ("                         X=0,  display none of the messages below.               \n");
    printf ("                         X=1,  beginning of plc scan loop, bringing plcs online. \n");
    printf ("                         X=2,  beginning of plc scan.                            \n");
    printf ("                         X=3,  AO queue actions (normal: sv85_aodriver & checkoutputs)\n");
    printf ("                         X=4,  BO queue actions (normal: sv85_bodriver & checkoutputs)\n");
    printf ("                         X=5,  SO queue actions (normal: sv85_sodriver & checkoutputs)\n");
    printf ("                         X=6,  calls to checkoutputs.                            \n");
    printf ("                         X=7,  checkoutputs: trace each queue entry read.        \n");
    printf ("                         X=8,  sv85_aidriver, sv85_aoread: normal call return.             \n");
    printf ("                         X=9,  sv85_bidriver, sv85_boread: normal call return.             \n");
    printf ("                         X=10, sv85_sidriver: normal call return.                     \n");
    printf ("                         X=11 to 14,  reserved.                                  \n");
    printf ("                         X=15, 'Microsol Busy', waiting in checkoutputs.         \n");
    printf ("                         X=16 to 254,  reserved.                                 \n");
    printf ("\n");

    taskDelay (1 * 60); /* Time for operator to see it and pause the output */

    printf (" modtrafficon(N)       Enable traffic tracing for all plc addresses                \n");
    printf ("                         and all modbus functions on line N.                       \n");
    printf ("*modtrafficoff(N)      Disable all traffic tracing on line N.                      \n");
    printf (" modtraffic(N,X,Y)     Modifies line N traffic tracing as follows:                 \n");
    printf ("                         X=0   disables tracing of all addresses.                  \n");
    printf ("                         X=255 enables tracing of all addresses.                   \n");
    printf ("                         0 < X < 255 enables tracing of address X traffic.         \n");
    printf ("                         Y=0   disables tracing of all modbus functions.           \n");
    printf ("                         Y=255 enables tracing of all modbus functions.            \n");
    printf ("                         0 < Y < 255 enables tracing of function Y traffic.        \n");
    printf ("\n");
    printf (" moddefaulttrace(N)    Restores default setting for all traces.                    \n");
    printf ("\n");

    taskDelay (1 * 60); /* Time for operator to see it and pause the output */

    printf (" modsimon(N,D)         Enables the simulation of responding plcs, reading all      \n");
    printf ("                        input values as zero.                                      \n");
    printf ("*modsimoff(N)          Disables simulation, thus enabling normal driver function.  \n");
    printf ("\n");

    taskDelay (1 * 60); /* Time for operator to see it and pause the output */

    printf (" modcomm(N,t1,f,t2,t3) Set communication timing parameters for line N as follows:  \n");
    printf ("                         t1: intermessage delay (in ticks).                        \n");
    printf ("                          f: 0 = full duplex, 1 = half duplex.                     \n");
    printf ("                         t2: post RTS assert, pre TX delay (in ticks).             \n");
    printf ("                         t3: post TX, pre RTS drop delay (in ticks).               \n");
    printf ("\n");
    printf (" moddumpmsgcnts(N)     Display line N communication statistics.                    \n");
    printf (" modclearmsgcnts(N)    Display and clear line N communication statistics.          \n");
    printf (" moddumpplcstats(N,X)  Display line N, plc X, communication statistics.            \n");
    printf (" modclearplcstats(N,X) Display and clear line N, plc X, xfers, status & errors.    \n");
    printf (" dump_mod_table(N,X)   Display line N, plc X, xfers, status, errors and input data.\n");
}


/********
*
*	This routine sets the timeout value for the nodes on the MB+
*
********/
void sv85to (unsigned short secs)
{
	sv85nodetimeout = secs;
}

/********
*
*	This routine enables logging of comm trace messages
*
********/
void sv85traceon (unsigned short line)
{
	modtrace [line] = TRUE;
	looptotmsgs [line] = 0;	    /* init the total message count	*/
}

/********
*
*	This routine disables logging of comm trace messages
*
********/
void sv85traceoff (unsigned short line)
{
	modtrace [line] = FALSE;
}

/********
*
*	This routine enables detailed display of normal communications traffic
*
********/
void sv85traffic (unsigned short line, unsigned char add, unsigned type)
{
	trace [line] = trace [line] | ALLTRAFFICTRACE;
        traceadr [line] = add;
        tracetype [line] = type;
        return;
}

/********
*
*	This routine enables detailed display of normal communications traffic
*
********/
void sv85trafficon (unsigned short line)
{
	trace [line] = trace [line] | ALLTRAFFICTRACE;
        traceadr [line] = 255;  /* every address */
        tracetype [line] = 255; /* every message type */
}

/********
*
*	This routine disables detailed display of normal communications traffic
*
********/
void sv85trafficoff (unsigned short line)
{
	trace [line] = trace [line] & ~ALLTRAFFICTRACE;
}

/********
*
*	This routine enables logging of resource trace messages
*
********/
void sv85restrlv (unsigned short line, unsigned short level)
{
	modrestrace [line] = level;
}

/********
*
*	This routine enables logging of resource trace messages
*
********/
void sv85restraceon (unsigned short line)
{
	modrestrace [line] = 1;
}

/********
*
*	This routine disables logging of resource trace messages
*
********/
void sv85restraceoff (unsigned short line)
{
	modrestrace [line] = 0;
}

/********
*
*	This routine controls logging of trace messages
*
********/
void sv85trlv (unsigned short line, unsigned short level)
{
	mltrace [line] = level;
}

/********
*
*	This routine enables the simulation of responding plcs, reading all input values as zero.
*
********/
void sv85simon (unsigned short line, unsigned short delay, int low, int high, int digenable)
{
	if ((delay < 1) || (delay > 60))
	{
	    simdelay [line] = 60;
	}
	else
	{
	    simdelay [line] = delay;
	}
	if (low > high)
	{
	    simhigh [line] = low;
	    simlow [line] = high;
	}
	else
	{
	    simlow [line] = low;
	    simhigh [line] = high;
	}
	modsim [line] = TRUE;
	simdigon [line] = digenable;
}

/********
*
*	This routine disables simulation, thus enabling normal driver function.
*
********/
void sv85simoff (unsigned short line)
{
	modsim [line] = FALSE;
}

/********
*
*	These routines enable and disable tracing of database configuration
*	errors in driver calls
*
********/
void sv85contraceon (unsigned short line)
{
	trace [line] = trace [line] | CONFIGURATIONTRACE;
}

void sv85contraceoff (unsigned short line)
{
	trace [line] = trace [line] & ~CONFIGURATIONTRACE;
}

/********
*
*	These routines enable and disable tracing of runtime errors in driver calls
*
********/
void sv85runtraceon (unsigned short line)
{
	trace [line] = trace [line] | RUNTIMEERRORTRACE;
}

void sv85runtraceoff (unsigned short line)
{
	trace [line] = trace [line] & ~RUNTIMEERRORTRACE;
}

/********
*
*	These routines enable and disable tracing of call rejection
*	errors in driver calls
*
********/
void sv85rejecttraceon (unsigned short line)
{
	trace[line] = trace [line] | CALLREJECTTRACE;
}

void sv85rejecttraceoff (unsigned short line)
{
	trace[line] = trace [line] & ~CALLREJECTTRACE;
}


/********
*
*	These routines enable and disable tracing of call rejection
*	errors in driver calls
*
********/
void sv85nacktraceon (unsigned short line)
{
	trace [line] = trace [line] | MODBUSNACKTRACE;
}

void sv85nacktraceoff (unsigned short line)
{
	trace [line] = trace [line] & ~MODBUSNACKTRACE;
}

/********
*
*	This routine restores default setting for all traces.
*
********/
void sv85defaulttrace (unsigned short line)
{
        /* debug structures */

	modtrace [line] = FALSE;
	mltrace [line] = 0;
	modrestrace [line] = 1;
        trace [line] = 0;
        traceadr [line] = 255;  /* every address      */
        tracetype [line] = 255; /* every message type */
}

/********
*
*	This routine turns on all error tracing.
*
********/
void moderrortrace (unsigned short line)
{
        /* debug structures */

	modtrace [line] = TRUE;
	modrestrace [line] = 2;
	trace[line] = MODBUSNACKTRACE ||
                      CONFIGURATIONTRACE ||
                      RUNTIMEERRORTRACE ||
                      CALLREJECTTRACE ||
                      0;
        traceadr [line] = 255;  /* every address      */
        tracetype [line] = 255; /* every message type */
}

/**********
*
*	This routine is used to modify the receive timeout used in
*	path processing.
*
**********/
void sv85pathtimeout (unsigned short ln, unsigned long tmout)
{
	if (tmout > 25)
	{
	    path_timeout [ln] = tmout;
	    logMsg ("Path timeout set to %d\n", tmout, 0, 0, 0, 0, 0);
	}
	else
	{
	    logMsg ("Path timeout must be greater than 25\n", 0, 0, 0, 0, 0, 0);
	}
}
/**********
*
*	This routine is used to modify the slv_want_tmout which controls how
*	long the driver will wait before changing to SLVWANT state.
*
**********/
void sv85slvwant (unsigned short ln, unsigned long tmout)
{
	slv_want_tmout [ln] = tmout;
	logMsg ("SLVWANT timeout changed to %d ticks\n", tmout, 0, 0, 0, 0, 0);
}

/**********
*
*	This routine attempts to stop the driver by detaching from the watchdog
*	timer and deleting the task.
*
**********/
int stopsv85 (short ln)
{
	if (comm_disconnect (ln) != -1)
	{
            /* if watchdog timer has been attached, detach it before deleting the task */
	    taskDelete (taskId [ln]);
	    return (1);
	}
	else
	{
	    return (-1);
	}
}
/**********
*
*	This routines starts the driver. It is usually called from startup.cmd
*
**********/
int startsv85 (unsigned short ln, unsigned short address, unsigned short delay,
			unsigned long stat_freq, unsigned long pathtimeout)
{

	if (ln < MODLINES)
	{
	    stat_refresh_rate [ln] = stat_freq;
	    slv_want_tmout [ln] = SLWWINDOW;
	    if (pathtimeout > 25)
	    {
		path_timeout [ln] = pathtimeout;
	    }
	    else
	    {
		path_timeout [ln] = 30;
	    }
	    taskId [ln] = taskSpawn (tasknames [ln],
					MOD_PRIORITY,
	    				MOD_TASK_OPTIONS,
					MOD_STACK,
					(FUNCPTR) sv85_driver,
					ln,
					address,
					delay, 0, 0, 0, 0, 0, 0, 0);

	    logMsg ("Started the SV85 driver on line %d with MD`BPlus Address = %d\n",
			ln, address, 0, 0, 0, 0);

	    remCurIdGet (user, password);
	    return (1);
	}
	else
	{
	    logMsg ("Invalid line number %d\n", ln, 0, 0, 0, 0, 0);
	    return (-1);
	}
}

/********
*
* Set communication logic time delays to handle half/full duplex, radio delays, etc
*
********/
void sv85comm (unsigned short line,
              unsigned short tickcount1,
              unsigned short hdflag,
              unsigned short tickcount2,
              unsigned short tickcount3)
{
	intermsgdelay [line] = tickcount1;  /* at beginning new message/response sequence.               */
	halfduplex [line]    = hdflag;
	pretxdelay [line]    = tickcount2;  /* after asserting RTS but before beginning transmission.    */
	posttxdelay [line]   = tickcount3;  /* after sending last character but before dropping RTS.     */
}


/**********
*
*	SV85_driver
*
*	The purpose of this task is to scan the modbus plc
*	lines. It will maintain a configuration list supporting online additions
*	and replacement of modules, and maintain a table of most recently
*	scanned values.
*
**********/                                                                                     	

/* not called from within this driver 	*/

void sv85_driver (unsigned short line, unsigned short address, unsigned short delay)
{

	unsigned short  mod, cfgchk, cnt, modspolled, onlinefreq;
	unsigned short  activity, pstate, lstate, execute, command, i;
	long	        this_time, last_time;
	MODPLC          *pmod;
        unsigned int    temptick, tempdiff;
        int             temp;
	unsigned char	*BASE;

/* Initialize the data structures */

	intermsgdelay [line] = delay;

	mbplusaddr [line] = address;

	linfo [line].loading = FALSE;

	looptotmsgs [line] = 0;	    /* init the total message count	*/
	resresponse [line] = FALSE;
	resnames [line] [0] = 0;
	prestable [line] = NULL;

	activity_time [line] = tickGet ();
	lastactivity [line] = line_activity [line] = FALSE;
	nextouttime [line] = 0;
	outagain [line] = 0;
	sv85tix = 0;

	/* will wait up to 30 seconds for partner to become valid */

        pstatewait [line] = 30;

        /* debug structures */

/*	moddebugtrace (line);	*/

/*	moddefaulttrace (line); */

        tick [line] [0] = tick [line] [1] = tick [line] [2] = tick [line] [3] = 0;

	linestats [line].TXintcount = 0;
	linestats [line].RXintcount = 0;
	linestats [line].msgxfercount = 0;
        linestats [line].startingtick = tickGet ();

	/* start with normal operation. */

	modsim [line] = FALSE;

	lpath [line] = (LPATHS *) calloc (1, (sizeof (LPATHS)));
	if (!lpath [line]) {
	    logMsg ("Not enough memory for path table line %d\n", line, 0, 0, 0, 0, 0);
	    return;
	}

	sockBuff [line] = calloc (1, 4096);
	if (sockBuff [line] == 0)
	{
	    logMsg ("Cannot allocate buffer for redundant status messages\n", 0, 0, 0, 0, 0, 0);
	    return;
	}

	tcpRxBuff [line] = calloc (1, 4096);
	if (tcpRxBuff [line] == 0)
	{
	    logMsg ("Cannot allocate buffer for redundant status messages\n", 0, 0, 0, 0, 0, 0);
	    return;
	}

	redundSem [line] = semBCreate (SEM_Q_PRIORITY, SEM_EMPTY);

        /* output queue initialization */

	oq [line] = (OUTPUTQUE *) calloc (1, (sizeof (OUTPUTQUE)));
	if (!oq [line])
	{
	    logMsg ("Not enough memory for line %d data structure (data output queue)\n",
			line, 0, 0, 0, 0, 0);
	    return;
	}
        outqueinit (line);

        /* communication initialization */

	rxlinetab [line] = (RXLINE *) calloc (1, (sizeof (RXLINE)));
	if (!rxlinetab [line])
	{
	    logMsg ("Not enough memory for line %d data structure (comm: rx)\n",
			line, 0, 0, 0, 0, 0);
	    return;
	}
	txlinetab [line] = (TXLINE *) calloc (1, (sizeof (TXLINE)));
	if (!txlinetab [line])
	{
	    logMsg ("Not enough memory for line %d data structure (comm: tx)\n",
			line, 0, 0, 0, 0, 0);
	    return;
	}
	/* make display easier to read - avoids interleaving of other task data */


	/* causes on_line calls to begin on first pass if master */

	onlinefreq = 1;

	/* Wait for resource to be defined via an initial call to
	   'register_resource' via driver call.  */

	while ((resnames [line] [0] == 0) || (prestable [line] == NULL))
	{
            if (modrestrace [line] > 1)
	    {
	    	logMsg ( "No Resource Defined for line - %d \n", line, 0, 0, 0, 0, 0);
	    }
	    taskDelay (600);
	}

	/* !!! FOR DEBUGGING !!! */
/*
*	while (mltrace [line] == 0) {
*	    taskDelay (60);
*	}
*/
	/* Initialize this tasks semaphore*/

	semids [line] = semBCreate (SEM_Q_PRIORITY, SEM_EMPTY);

	/* Initialize the Interrupt Vector */

	BASE = bases [line];

	BASE [VECTORREG] = vectors [line];

	intConnect ((VOIDFUNCPTR *)(vectors [line] * 4), (VOIDFUNCPTR) sv85_int_proc, line);
	sysIntEnable (SV85INTLEV); 

	while (!sv85initsequence (line, &mbplusaddr [line], TRUE))
	{
	    logMsg ( "Board will not init on line - %d \n", line, 0, 0, 0, 0, 0);
	    taskDelay (180);
	}	

	/* Start out as slave */

	resClearCaAnswer (prestable [line]);	/* start quiet */
	resPutLocalState (prestable [line], RES_SLV);
	if (modrestrace [line])
	{
	    logMsg ("1] Initialized as RES_SLV\n", 0, 0, 0, 0, 0, 0);
	}

        /* wait finite time for pstate to be resolved */

        logMsg ("Begin wait for pstate [%d] to become valid.\n",
			pstatewait [line], 0, 0, 0, 0, 0);
        for (i = 0; i <= pstatewait [line]; i++)
        {
            resGetStates (prestable [line], &lstate, &pstate);
	    if (pstate != 0)
	    {
		break;
	    }
            taskDelay (60);
        }
        if (pstate == 0)
	{
            logMsg ("Timeout of pstate wait.\n", 0, 0, 0, 0, 0, 0);
        }
	else
	{
            logMsg ("Completion of pstate wait.\n", 0, 0, 0, 0, 0, 0);
	}
	sv85SockCreate (line);

	redtaskId [line] = taskSpawn (redtasknames [line],
					MOD_PRIORITY,
					MOD_TASK_OPTIONS,
					MOD_STACK,
					(FUNCPTR) redSv85Task,
					line, 0, 0, 0, 0, 0, 0, 0, 0, 0
                                     );

	logMsg ("Started the SV85 redundant task on line %d\n",
			line, 0, 0, 0, 0, 0);

	lastStrobeRxTime [line] = tickGet ();

	semGive (redundSem [line]);

	taskDelay (120);

	if (lastStrobeRx [line] == 'M')
	{
	    line_activity [line] = TRUE;
	    partnerAlive [line] = TRUE;
	}
	else if (lastStrobeRx [line] == 'S')
	{
	    line_activity [line] = FALSE;
	    partnerAlive [line] = TRUE;
	}
	else
	{
	    line_activity [line] = FALSE;
	    partnerAlive [line] = FALSE;
	}

        for (i = 0; i < ACTIVITYTIME / 10; i++) ;  /* ten seconds */
        {
            taskDelay (10);
        }
	activity = check_line_activity (line, 0);
	resGetStates (prestable [line], &lstate, &pstate);

	/* if the partner is not there, be master */

	if ((pstate != RES_MSTR) && (pstate != RES_SLVWANT) && (!activity))
	{
	    resSetCaAnswer (prestable [line]);
	    resPutLocalState (prestable [line], RES_MSTR);

            lastscan [line] = tickGet ();
            scantime [line] = 0;

	    lstate = RES_MSTR;
	    if (modrestrace [line])
	    {
	    	logMsg ("2] Became RES_MSTR   P= %x A= %d \n",
				pstate, activity, 0, 0, 0, 0);
	    }
	}

	cfgchk = 0;			/* Init the new config pointer */
	last_time = tickGet ();		/* Initialize the time  */

	while (TRUE)
	{

	    activity_time [line] = tickGet ();
	    lastactivity [line] = line_activity [line] = FALSE;

            /**********
	    * we arrive here because we are a slave.
	    * we stay until:
	    * asked to be master by partner if we can reach PLCs, become master.
            * the partner is not master and there is no activity on the line.
            *********/

	    while (lstate == RES_SLV)
	    {
		sv85UDPCheck (line);
		if (lastStrobeRx [line] == 'M')
		{
		    line_activity [line] = TRUE;
		    partnerAlive [line] = TRUE;
		}
		else if (lastStrobeRx [line] == 'S')
		{
		    line_activity [line] = FALSE;
		    partnerAlive [line] = TRUE;
		}
		else
		{
		    line_activity [line] = FALSE;
		    partnerAlive [line] = FALSE;
		}

                /* check for plc simulated mode. */

                if (modsim [line])
                {
		    resSetCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_MSTR);
		    lstate = RES_MSTR;

                    lastscan [line] = tickGet ();
                    scantime [line] = 0;

		    if (modrestrace [line])
		    {
                        logMsg ("Simulation Mode] Became RES_MSTR  P= %x A= %d \n",
					pstate, activity, 0, 0, 0, 0);
		    }
		    break;
                }

 		/* make sure driver is not responding */

		resresponse [line] = FALSE;

 		/* clear out commands */

		resPutCommand (prestable [line], 0);

 		/* wait 10 seconds before checking */

		activity = check_line_activity (line, 10);
		resGetStates (prestable [line], &lstate, &pstate);

		/* see if needed by partner to be master. */
		/* if yes, do it and break out of slave state. */

		if (pstate == RES_SLVWANTCA)
		{
		    resSetCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_MSTR);
		    lstate = RES_MSTR;

                    lastscan [line] = tickGet ();
                    scantime [line] = 0;

		    if (modrestrace [line])
		    {
			logMsg ("3] Became RES_MSTR  P= %x A= %d \n",
					pstate, activity, 0, 0, 0, 0);
		    }
		    break;
		}

		/* If partner is there and wants to give it up, try to take it */

		if (pstate == RES_SLVWANT)
		{
		    execute = FALSE;
		    taskDelay (2);		/* Time for switch to occur */

		    /* See if anyone is out there and able to respond */
		    /* to a loopback command */

		    for (mod = 1; (mod <= NUMMODPLCS); mod++)
		    {
			pmod = plcs [line] [mod];

			/* verify pointer validity prior to use */
			/* data points have been requested      */
			/* user has NOT forced plc offline      */

			if ((pmod) &&
			    (pmod->was_requested) &&
                            (!pmod->cmd_force_offline))
			{
			    on_line (line, mod);

			    if (sv85_needs_init[line])
			    {
				sv85initsequence (line,  &mbplusaddr [line], TRUE);
			    }

                            /* check cmd_force_offline flag before AND after */
			    /* due to asynchronous tasks */

                            if ((pmod->is_responding) && (!pmod->cmd_force_offline))
			    {
				execute = TRUE;
				break;
			    }
			}
		    }
		    /* If someone answered, become master and make */
		    /* sure partner really gave it up */

		    if (execute)
		    {
			resPutLocalState (prestable [line], RES_MSTR);
			lstate = RES_MSTR;

                        lastscan [line] = tickGet ();
                        scantime [line] = 0;

			if (modrestrace [line])
			{
			    logMsg ("4] Became RES_MSTR  P= %x A= %d \n",
					pstate, activity, 0, 0, 0, 0);
			}
			execute = FALSE;
			activity_time [line] = tickGet ();
			lastactivity [line] = line_activity [line] = FALSE;
			for (i = 0; (i < 60); i++)
			{
			    /* pause before testing */

			    activity = check_line_activity (line, 2);
			    resGetStates (prestable [line], &lstate, &pstate);
			    if ((pstate != RES_MSTR) && (pstate != RES_SLVWANT) && (!activity) )
			    {
				execute = TRUE; /* Partner Gave It Up so take CA */
				break;
			    }
			}
			if (!execute)
			{
			    /* Partner did not give it up go back to slave */

			    resPutLocalState (prestable [line], RES_SLV);
			    lstate = RES_SLV;
			    if (modrestrace [line])
			    {
			    	logMsg ("5] Became RES_SLV  P= %x A= %d \n",
						pstate, activity, 0, 0, 0, 0);
			    }
			}
			else
			{
			    /* Assume control by taking the CA responsibilities */ 

			    resSetCaAnswer (prestable [line]);
			    prepresponses (line);
			    break;
			}
		    }
		}

		/* if the partner is not talking, be master */

		if ((pstate != RES_MSTR) &&  (pstate != RES_SLVWANT) && (!activity))
		{
		    resSetCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_MSTR);
		    lstate = RES_MSTR;

                    lastscan [line] = tickGet ();
                    scantime [line] = 0;

		    if (modrestrace [line])
		    {
		    	logMsg ("6] Became RES_MSTR P= %x A= %d \n",
					pstate, activity, 0, 0, 0, 0);
		    }
		    prepresponses (line);
		}
	    }

	    /* to arrive here, we used to be master, and had been told */
	    /* to relinquish master status. */
	    /* stay in this loop until the partner takes over, */
	    /* or the command is canceled */

	    while (lstate == RES_SLVWANTCA)
	    {
		sv85UDPCheck (line);
		if (lastStrobeRx [line] == 'M')
		{
		    line_activity [line] = TRUE;
		    partnerAlive [line] = TRUE;
		}
		else if (lastStrobeRx [line] == 'S')
		{
		    line_activity [line] = FALSE;
		    partnerAlive [line] = TRUE;
		}
		else
		{
		    line_activity [line] = FALSE;
		    partnerAlive [line] = FALSE;
		}
	        taskDelay (2);		/* Give OS a chance...  */

		/* Make Sure driver is not responding   */

		resresponse [line] = FALSE;

		resGetStates (prestable [line], &lstate, &pstate);
		if ((pstate == RES_MSTR) || (pstate == RES_SLVWANT))
		{
		    resClearCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_SLV);
		    lstate = RES_SLV;
		    if (modrestrace [line])
		    {
		    	logMsg ("7] Became RES_SLV  P= %x A= %d \n",
					pstate, activity, 0, 0, 0, 0);
		    }
		    break;
		}
		resGetCommand (prestable [line], &command);
		resPutCommand (prestable [line], 0); /* clear out commands */
		if (command == RES_CANCELREL)
		{
		    resSetCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_MSTR);
		    lstate = RES_MSTR;
		    if (modrestrace [line])
		    {
		    	logMsg ("8] Became RES_MSTR  P= %x A= %d \n",
					pstate, activity, 0, 0, 0, 0);
		    }
		    prepresponses (line);
		}
	    }

	    /* it doesn't appear that we ever come here in a RES_SLVWANT */
	    /* lstate mode. it happens within this loop, but is never seen */
	    /* by the while loop conditional evaluation. */

	    while (lstate == RES_SLVWANT)
	    {
		if ((modrestrace [line]) || (trace[line] & CONFIGURATIONTRACE))
		{
		    /* this condition should NEVER occur. print error message every 5 seconds */
		    logMsg ("Illegal Local Resource State.\n", 0, 0, 0, 0, 0, 0);
		    if (trace[line] & CONFIGURATIONTRACE)
		    {
			logMsg ("  Possible Resource Configuration Error:\n", 0, 0, 0, 0, 0, 0);
			logMsg ("   Check for multiple drivers sharing single Resource.\n",
					0, 0, 0, 0, 0, 0);
		    }
                    taskDelay (600);
		}
		taskDelay (2);
            }

	    /* we come here because we have gone to, or are looping in, */
	    /* the master state. we stay in this loop until: */
	    /* we have been told to give master state to the partner.*/
	    /* we break out to RES_SLVWANTCA.                        */
	    /* we have successfully handed master state to partner.  */
	    /* While master, this task will process all modules sending */
	    /* outputs and getting inputs. */
	    /* It will then try to configure requested non-responding modules. */
            
	    last_time = tickGet ();		/* initialize gap */
	    while (lstate == RES_MSTR)
	    {
		sv85UDPCheck (line);
		if (lastStrobeRx [line] == 'M')
		{
		    line_activity [line] = TRUE;
		    partnerAlive [line] = TRUE;
		}
		else if (lastStrobeRx [line] == 'S')
		{
		    line_activity [line] = FALSE;
		    partnerAlive [line] = TRUE;
		} 
		else
		{
		    line_activity [line] = FALSE;
		    partnerAlive [line] = FALSE;
		}
		if (mltrace [line] == 1)
		{
		    logMsg ( "Top of Master Loop... \n", 0, 0, 0, 0, 0, 0);
	        }
                /* check for plc simulated mode. */

                while (modsim [line])
                {
		    resresponse [line] = TRUE; /* Make sure driver is responding. driver won't         */
		    for (mod = 1; (mod <= NUMMODPLCS); mod++)
		    {
		        pmod = plcs [line] [mod];	/* shorthand pointer */

			/* verify pointer validity prior to use */

		        if (pmod)
		        {
			    if (pmod->cmd_force_offline)
			    { 
				/* if user has turned this plc offline */
				/* via pseudo commands, */
                                /* mark it as not responding. */

			 	if (pmod->is_responding)
				{
				    pmod->no_offlines++;
				}
 				pmod->is_responding = FALSE;
	    		    }
			    else
			    {
				sim_plc (pmod, line, mod);	/* I/O done here! */
		   		taskDelay (simdelay [line]);
			    }
			}
		    }
                }
		/* initial stuff */
		/* Make sure driver is responding. driver won't           */
                /* issue commands if not set.                             */

		resresponse [line] = TRUE;

		resGetCommand (prestable [line], &command);
		resPutCommand (prestable [line], 0);	/* clear out commands */

		/* if relinquish command sent, change state and */
		/* break out to outer loop */

		if (command == RES_RELINQUISH)
		{
		    resClearCaAnswer (prestable [line]);
		    resPutLocalState (prestable [line], RES_SLVWANTCA);
		    lstate = RES_SLVWANTCA;
		    if (modrestrace [line])
		    {
                       logMsg ("9] Became RES_SLVWANTCA  P= %x A= %d \n",
					pstate, activity, 0, 0, 0, 0);
		    }
		    break;
		}

                /* measure time between scans of the entire set of plcs */
		/* display if requested. */ 

                temptick = tickGet ();
                tempdiff = temptick - lastscan [line];

                if (modtrace [line])
		{
                    logMsg ("ScanTime: =%4d, # OutputMsgs =%5d, # TotalMsgs =%3d, accum = %d\n",
                            scantime [line], loopoutmsgs [line], loopmsgs [line], looptotmsgs[line], 0, 0);
		}

                loopoutmsgs [line] = 0;
                loopmsgs [line] = 0;
                lastscan [line] = temptick;

		/* scan plcs for data i/o, if allowed */

		modspolled = 0;

		/* call 'scan_plc' for each PLC device, unless user has */
                /* taken it off line or it is not responding.           */

		for (mod = 1; (mod <= NUMMODPLCS); mod++)
		{
		    pmod = plcs [line] [mod];	/* shorthand pointer */

		    /* verify pointer validity prior to use */

		    if (pmod)
		    {
			if (pmod->cmd_force_offline)
			{ 
			    /* if user has turned this plc offline via pseudo */
			    /* commands, mark it as not responding. */

			    if (pmod->is_responding)
			    {
				pmod->no_offlines++;
			    }
 			    pmod->is_responding = FALSE;
	    		}
			else if (pmod->is_responding)
			{
                            /* local/partner lockstep recovery */

                            resGetStates (prestable [line], &lstate, &pstate);
                            if (pstate == RES_MSTR)
                            {
                                resClearCaAnswer (prestable [line]);
                                resPutLocalState (prestable [line], RES_SLV);
                                lstate = RES_SLV;


                                if (modrestrace [line])
				{
                                    logMsg ("13] Became RES_SLV  P= %x A= %d \n",
						pstate, activity, 0, 0, 0, 0);
				}
                                temp = ((tickGet () % 8) + 5); /* random 5 to 12 second delay */
                                for ( i = 0; i < temp; i++)
                                {
                                   taskDelay(60);
                                }
                                break;
                            }
			    modspolled++;

			    last_time = tickGet ();

			    scan_plc (pmod, line, mod);	/* I/O done here! */

			    if (sv85_needs_init [line])
			    {
				sv85initsequence (line,  &mbplusaddr [line], TRUE);
			    }
			    if (!pmod->is_responding)
			    {
				mark_not_valid (pmod);
                            }
			}
		    }
		    /* end of single module action */

		    /* if told to relinquish our master state, stop plc */
		    /* scanning immediately. */
		    /* don't trash the command, so that we can still read it */
		    /* on the next loop through this while statement. */

		    resGetCommand (prestable [line], &command);
		    if (command == RES_RELINQUISH)
		    {
			break;
		    }
		}
		/* end of scan loop */
                /* measure how long the scan took. */
		/* to be displayed at beginning of next loop. */

                temptick = tickGet ();
                scantime [line] = temptick - lastscan [line];

		/* try and bring all nonresponding plcs on line, but */
                /* don't bring in more than a specified number.      */

		cnt = 0;
		if (--onlinefreq == 0) 		/* do we try to bring any in on this pass? */
		{ 
		    onlinefreq = BRINGBACKCNT;
		    for (mod = 1; (mod <= NUMMODPLCS); mod++)
		    {
			/* manage plc index */
			cfgchk++;
			if (cfgchk == NUMMODPLCS + 1)
			{
			    cfgchk = 1;	/* wrap plc index */
                        }
			pmod = plcs [line] [cfgchk];

			/* verify pointer validity prior to use*/

			if ((pmod) &&
			    (!pmod->is_responding) &&
			    (pmod->was_requested) &&
			    (!pmod->cmd_force_offline))
		    	{
 			    on_line (line, cfgchk);

			    if (sv85_needs_init [line])
			    {
				sv85initsequence (line,  &mbplusaddr [line], TRUE);
			    }

			    modspolled++;
			    cnt++;
                            if (cnt == MAXCFGCK)
			    {
				break;
			    }
			}
		    }
		}

		/* see if anyone on the line has been responding. */
                /* if not, try to get partner to become slave.    */
		/* Check for no one responding */

		if (!any_responding (line))
		{
 		    /* See if it is time to let partner try */

		    this_time = tickGet ();
		    if ((this_time - last_time) > (slv_want_tmout [line]))
		    {
			/* #ticks for slave to try */
			/* Switch States to let partner know it's its time to try */

			resPutLocalState (prestable [line], RES_SLVWANT);
			lstate = RES_SLVWANT;
			if (modrestrace [line])
			{
			    logMsg ("10] Became RES_SLVWANT  P= %x A= %d \n",
					pstate, activity, 0, 0, 0, 0);
			}

			/* Remember the time of this try */

			last_time = tickGet ();
			lastactivity [line] = line_activity [line] = FALSE;
			for (i = 0; (i < SLWWINDOW); i += 10)
			{
			    /* SLWINDOW is the # of ticks for slave to try */

			    taskDelay (10);

		    	    /* If partner picked up let it have control */
			    /* check immediately */

			    activity = check_line_activity (line, 0);
			    resGetStates (prestable [line], &lstate, &pstate);
			    if ((pstate == RES_MSTR) || (pstate == RES_SLVWANT) || (activity))
			    {
				resClearCaAnswer (prestable [line]);
				resPutLocalState (prestable [line], RES_SLV);
				lstate = RES_SLV;
				if (modrestrace [line])
				{
				    logMsg ("11] Became RES_SLV  P= %x A= %d \n",
						pstate, activity, 0, 0, 0, 0);
				}
                                taskDelay (60);
                                taskDelay (60);
                                taskDelay (60);
				break;
			    }
			}
			/* Go Back to being the Master and take control */
			/* of the hardware */
			if (lstate != RES_SLV)
			{
			    resPutLocalState (prestable [line], RES_MSTR);
			    lstate = RES_MSTR;
			    if (modrestrace [line])
			    {
			    	logMsg ("12] Became RES_MSTR  P= %x A= %d \n",
						pstate, activity, 0, 0, 0, 0);
			    }
			}
		    }
		}

		/* take plcs off hold status. the plc's data are not to */
		/* be held off since now master. */

		for (mod = 1; (mod <= NUMMODPLCS); mod++ )
		{
		    pmod = plcs [line] [mod];

		    /* verify pointer validity prior to use */

		    if (pmod)
		    {        
			pmod->holdflag = FALSE;
                    }
		}
		if (modspolled == 0)
		{
		    taskDelay (6);
		}
                else
		{
		    taskDelay (2);
		}
	    }
	}
	/* end infinite task loop */
}

/**********
*
* Try to bring a PLC on line
*
**********/

static void on_line (unsigned short line, unsigned short mod)
{
	MODPLC		*pmod;

	pmod = plcs [line] [mod];	

	if (pmod)			/* verify pointer prior to use */
	{
	    if (mltrace [line] == 1)
	    {
		logMsg ( "Bring Online PLC %d\n", mod, 0, 0, 0, 0, 0 );
	    }
	    /* Use the loop back function to see if the unit */
	    /* is there and responding */

	    if (!modloopback (ONERETRY + line, mod, pmod))
	    {
		/* if no response, get out */
		if (modtrace [line])
		{
		    logMsg ("Bring On-line  Failed for Mod %d \n", mod, 0, 0, 0, 0, 0);
		}
		pmod->no_rxnoresp++;
		return;
	    }

	    /* a PLC that has just been brought online must have    */
	    /* it's data marked as being not valid.                 */

	    mark_not_valid (pmod);

	    pmod->is_responding = TRUE;
        }
}

/**********
*							
* Mark all plc data elements as not valid     		
*                                             		
**********/                                             	

/* called from on_line and prepresponses      	*/    	

static void mark_not_valid (MODPLC  *pmod)
{
	unsigned	short i;

/*
*      COILS ARE INITIALIZED AS READ WRITE VALUES      
*              - READ BACK NOT VALID 			
*/
	for (i = 0; (i < NUMMODCOILS); i++)
	{
	    pmod->coils [i].data_valid = FALSE;
	}

/*
*      DIGITALS ARE INITIALIZED AS NOT VALID		
*/
	for (i = 0; (i < NUMMODDIS); i++)
	{
	    pmod->dis [i].data_valid = FALSE;
	}

/*
*      ANALOG INS ARE INITIALIZED AS NOT VALID		
*/
	for (i = 0; (i < NUMMODAIS); i++)
	{
	    pmod->ais [i].data_valid = FALSE;
	}

/*
*      REGISTERS ARE INITIALIZED AS READ WRITE VALUES		
*      WRITES ARE TREATED BY INDIVIDUAL REGISTERS NOT GROUPS 	
*              - READ BACK NOT VALID      	               	
*/
	for (i = 0; (i < NUMREGISTERS); i++)
	{
	    pmod->regs [i].data_valid = FALSE;
	}
}

/**********
*                                                           	
* Scan a MODBUS PLC -  get inputs and set outputs as required	
*                                                            	
**********/

static	void scan_plc (MODPLC *pmod, unsigned short line, unsigned short mod)
{
	unsigned short	i, j, bitaddr, regaddr, readlen;
	LOADINFO	sslinfo;

	if (mltrace [line] == 2)
	{
	    logMsg ( "Scanning PLC - %d.\n", mod, 0, 0, 0, 0, 0 );
	}
	checkoutputs (line);

	/* Test to see if load requested or in progress */

	if (pmod->prog_load_stat == PLC_LOAD_REQUESTED)
	{
	    loadinit (line, pmod, mod);
	    return;
	}
	if (pmod->prog_load_stat == PLC_LOAD_IN_PROGRESS)
	{
	    loadnextblock (line);
	    return;
	}

	/* Test to see if a start or stop is requested */

	if (pmod->plc_stop)
	{
	    sslinfo.line = line;
	    sslinfo.mod = mod;
	    sslinfo.pmod = pmod;
	    if (plc_login (&sslinfo))
	    {
		plc_stop (&sslinfo);
	    }
	    plc_logout (&sslinfo);
	    pmod->plc_stop = FALSE;
	}
	if (pmod->plc_start)
	{
	    sslinfo.line = line;
	    sslinfo.mod = mod;
	    sslinfo.pmod = pmod;
	    if (plc_login (&sslinfo))
	    {
		plc_start (&sslinfo);
	    }
	    plc_logout (&sslinfo);
	    pmod->plc_start = FALSE;
	}

	checkresponses (line);

	/* if enabled, get plc cpu status words */

	if (stat_refresh_rate [line])
	{
	    if (pmod->stat_fetch_time < tickGet ())
	    {
		get_plc_stats (line, mod, pmod);
		pmod->stat_fetch_time = (tickGet () + (stat_refresh_rate [line] * 60));
	    }
	}

	/* read the coil values to be scanned */

	for (i = 0; (i < NUMMODCOILS); i++)
	{
	    if (pmod->coils [i].was_requested)
	    {
		for (j = 1; ((j < MAXDIGBLKSIZE) && ((i + j) < NUMMODCOILS )); j++)
		{
		    if (!pmod->coils [i + j].was_requested)
		    {
		    	break;
		    }
		}
		bitaddr = (i * BITBLOCK);
		readlen = j * BITBLOCK;
		if (!readcoilsfd (line, mod, pmod, bitaddr, readlen, 0))
		{
		    return;
		}
/*
                if (temp == CMDFAIL)
		{
		    return;
		}


                if ((temp == CMDNACK) || (temp == CMDREJECT))
                {
*/
                    /* disable the entire range of NACKed points */
/*
                    cmd = MCOILS;
                    if (trace [line] & MODBUSNACKTRACE)
		    {
                        logMsg ("Modbus Command NACK: Disabling %d Points starting at L=%d,A=%d,C=%d,S=%d\n",
                           readlen, line, mod, cmd, bitaddr + 1);
		    }
                    for (k = 0 ; k < j ; k++)
                    {
                        pmod->coils [i + k].was_requested = FALSE;
                        pmod->coils[i + k].data_valid = FALSE;
                    }
                }
*/
		checkresponses (line);
		checkoutputs (line);
		i = (i + j) - 1;
	    }
	}

	/* read the di values to be scanned */

	for (i = 0; (i < NUMMODDIS); i++)
	{
	    if (pmod->dis [i].was_requested)
	    {
		for (j = 1; ((j < MAXDIGBLKSIZE) && ((i + j) < NUMMODDIS )); j++)
		{
		    if (!pmod->dis [i + j].was_requested)
		    {
		    	break;
		    }
		}
		bitaddr = (i * BITBLOCK);
		readlen = j * BITBLOCK;
		if (!readdisfd (line, mod, pmod, bitaddr, readlen, 0))
		{
		    return;
		}
/*
                if (temp == CMDFAIL)
		{
		    return;
		}

*/
/*
                if ((temp == CMDNACK) || (temp == CMDREJECT))
                {
*/
                    /* disable the entire range of NACKed points */
/*
                    cmd = MDINS;
                    if (trace [line] & MODBUSNACKTRACE)
		    {
                        logMsg ("Modbus Command NACK: Disabling %d Points starting at L=%d,A=%d,C=%d,S=%d\n",
                           readlen, line, mod, cmd, bitaddr+1);
		    }
                    for (k = 0; k < j; k++)
                    {
                        pmod->dis [i + k].was_requested = FALSE;
                        pmod->dis [i + k].data_valid = FALSE;
                    }
                }
*/
		checkresponses (line);
		checkoutputs (line);
		i = (i + j) - 1;
	    }
	}

	/* read the ai values to be scanned */

	for (i = 0; (i < NUMMODAIS); i++)
	{
	    if (pmod->ais [i].was_requested)
	    {
		for (j = 1; ((j < MAXREGBLKSIZE) && ((i + j) < NUMMODAIS)); j++)
		{
		    if (!pmod->ais [i + j].was_requested)
		    	break;
		}
		regaddr = i * REGBLOCK;
		readlen = j * REGBLOCK;
		if (!readaisfd (line, mod, pmod, regaddr, readlen, 0))
		{
		    return;
		}
/*
                if (temp == CMDFAIL)
		{
		    return;
		}
*/
/*
                if ((temp == CMDNACK) || (temp == CMDREJECT))
                {
*/
                    /* disable the entire range of NACKed points */
/*
                    cmd = MAINS;
                    if (trace [line] & MODBUSNACKTRACE)
		    {
                        logMsg ("Modbus Command NACK: Disabling %d Points starting at L=%d,A=%d,C=%d,S=%d\n",
                           readlen, line, mod, cmd, bitaddr + 1);
		    }
                    for (k = 0; k < j; k++)
                    {
                        pmod->ais [i + k].was_requested = FALSE;
                        pmod->ais [i + k].data_valid = FALSE;
                    }
                }
*/
		checkresponses (line);
		checkoutputs (line);
		i = (i + j) - 1;
	    }
	}

	/* read the reg values to be scanned */

	for (i = 0; (i < NUMREGISTERS); i++)
	{
	    if (pmod->regs [i].was_requested)
	    {
		for (j = 1; ((j < MAXREGBLKSIZE) && ((i + j) < NUMREGISTERS)); j++)
		{
		    if (!pmod->regs [i + j].was_requested)
		    {
		    	break;
		    }
		}
		regaddr = i * REGBLOCK;
		readlen = j * REGBLOCK;
		if (!readregsfd (line, mod, pmod, regaddr, readlen, 0))
		{
		    return;
		}
/*
                if (temp == CMDFAIL)
		{
		    return;
		}
*/
/*                if ((temp == CMDNACK) || (temp == CMDREJECT))
                {
*/
                    /* disable the entire range of NACKed points */
/*
                    cmd = MREGS;
                    if (trace [line] & MODBUSNACKTRACE)
		    {
                        logMsg ("Modbus Command NACK: Disabling %d Points starting at L=%d,A=%d,C=%d,S=%d\n",
                           readlen, line, mod, cmd, bitaddr+1);
		    }
                    for (k = 0; k < j; k++)
                    {
                        pmod->regs [i + k].was_requested = FALSE;
                        pmod->regs [i + k].data_valid = FALSE;
                    }
                }
*/
		checkresponses (line);
		checkoutputs (line);
		i = (i + j) - 1;
	    }
	}
}

static	void sim_plc (MODPLC *pmod, unsigned short line, unsigned short mod)
{
	unsigned short	i, j;

	pmod->is_responding = pmod->was_requested;

	/* read the coil values to be scanned */

	for (i = 0; (i < NUMMODCOILS); i++)
	{
	    pmod->coils [i].data_valid = pmod->coils [i].was_requested;
	    if (simdigon [line])
	    {
		pmod->coils [i].last_read = ~pmod->coils [i].last_read;
	    }
	}

	/* read the di values to be scanned */

	for (i = 0; (i < NUMMODDIS); i++)
	{
	    pmod->dis [i].data_valid = pmod->coils [i].was_requested;
	    if (simdigon [line])
	    {
	        pmod->dis [i].last_read = ~pmod->dis [i].last_read;
	    }
	}

	/* read the ai values to be scanned */

	for (i = 0; (i < NUMMODAIS); i++)
	{
	    pmod->ais [i].data_valid = pmod->ais [i].was_requested;
	    for (j = 0; j < REGBLOCK; j++)
	    {
		if ((pmod->ais [i].aivals [j] < simlow [line]) ||
		    (pmod->ais [i].aivals [j] >= simhigh [line]))
		{
		    pmod->ais [i].aivals [j] = simlow [line];
		}
		else
		{
		    pmod->ais [i].aivals [j] += 1;
		}
	    }
	}

	/* read the reg values to be scanned */

	for (i = 0; (i < NUMREGISTERS); i++)
	{
	    pmod->regs [i].data_valid = pmod->regs [i].was_requested;
	    for (j = 0; j < REGBLOCK; j++)
	    {
		if ((pmod->regs [i].ioregs [j] < simlow [line]) ||
		    (pmod->regs [i].ioregs [j] >= simhigh [line]))
		{
		    pmod->regs [i].ioregs [j] = simlow [line];
		}
		else
		{
		    pmod->regs [i].ioregs [j] += 1;
		}
	    }
	}
}

/**********
*
* checkoutputs
*      checks output que and returns if empty.
*      if there is something on the queue it attempts to process all
*      pending outputs, doing immediate retries. If attempts fail the request is dropped.      
*
**********/                                                                                              

/* this routine only called from scan_plc */

static	void checkoutputs (unsigned short line)
{
	OUTQENT         *optr;
	unsigned short  regaddr, regval, link, adapter, coiladdr, conv;
	unsigned long   coilmask, coilval;
	MODPLC          *pmod;
	double 		dblaovalue;
        unsigned short  stringaddr, stringlen;
	unsigned char	*pstringval;
	unsigned short	write_status;

        if (mltrace [line] == 6)
	{
            logMsg ("Checking Output Queue ...\n", 0, 0, 0, 0, 0, 0); 
	}
	while ((optr = ogetnextp (line)) != NULLP)
	{
	    if ((optr->otype) == AOUTTYPE)
	    {
		link = optr->out.aonode.link;
		adapter = optr->out.aonode.adapter;
		pmod = plcs [link] [adapter];
		if (pmod == NULL)
		{
		    if (!allocate_plc (link, adapter))
		    {
			/* if not allocated, try. if can't allocate, get out. */
			return;
		    }
        	    pmod = plcs [link] [adapter];
		}
		regaddr = optr->out.aonode.aosignal;
		regval = optr->out.aonode.aovalue;
		dblaovalue = optr->out.aonode.dblaovalue;
		conv = optr->out.aonode.conversion;
                if (mltrace [link] == 7)
                {
                    logMsg ("AO IN QUEUE: L=%d, A=%d,       S=%d, Value=%d\n",
                           link, adapter, regaddr, regval, 0, 0);
                } 

		if ((!pmod->cmd_force_offline) &&
                    (resresponse [link]) &&
                    (pmod->is_responding))
		{
                    loopoutmsgs [line]++;
		    write_status = setreg (link, adapter, pmod, regaddr, regval, dblaovalue, conv);
		    if (write_status != CMDDONE)
		    {
		        pmod->badoutcnt++;		    
                        if (mltrace [link] == 3)
			{
                            logMsg("AO DEQUEUE: Output failed: L=%d, A=%d,       S=%d, Value=%d\n",
                                   link, adapter, regaddr, regval, 0, 0);
			} 
		    }
                    else
		    {
                        if (mltrace [link] == 3)
			{
                            logMsg ("AO DEQUEUE: Output succeeded: L=%d, A=%d,       S=%d, Value=%d\n",
                                   link, adapter, regaddr, regval, 0, 0);
			} 
		    }

		}
	    }
	    else if ((optr->otype) == BOUTTYPE)
	    {
		link = optr->out.bonode.link;
		adapter = optr->out.bonode.adapter;
		pmod = plcs [link] [adapter];
		if (pmod == NULL)
		{
		    if (!allocate_plc (link, adapter))
		    {
			/* if not allocated, try. if can't allocate, get out. */
			return;
		    }
        	    pmod = plcs [link] [adapter];
		}
		coiladdr = optr->out.bonode.bosignal;
		coilval = optr->out.bonode.bovalue;
		coilmask = optr->out.bonode.bomask;
                if (mltrace [link] == 7)
                {
                    logMsg("BO IN QUEUE: L=%d, A=%d,       S=%d, Value=%d\n",
                           link, adapter, coiladdr, coilval, 0, 0);
                }

		if ((!pmod->cmd_force_offline) &&
                    (resresponse [link]) &&
                    (pmod->is_responding))
		{
                    loopoutmsgs [line]++;
		    write_status = setcoil (link, adapter, pmod, coiladdr, coilmask, coilval);
/*
		    if (write_status != CMDDONE)
		    {
			pmod->badoutcnt++;
                        if (mltrace [link] == 4)
			{
                            logMsg("BO DEQUEUE: Output failed: L=%d, A=%d,       S=%d, Value=%d\n",
                                   link, adapter, coiladdr, coilval);
			}
		    }
                    else
		    {
                        if (mltrace [link] == 4)
			{
                            logMsg("BO DEQUEUE: Output succeeded: L=%d, A=%d,       S=%d, Value=%d\n",
                                   link, adapter, coiladdr, coilval);
			}
		    }
*/
		}
	    }
	    else if ((optr->otype) == SOUTTYPE)
	    {
		link = optr->out.sonode.link;
		adapter = optr->out.sonode.adapter;
/*		card = optr->out.sonode.card; */
		pmod = plcs [link] [adapter];
		if (pmod == NULL)
		{
		    if (!allocate_plc (link, adapter))
		    {
			return; /* if not allocated, try. if can't allocate, get out. */
		    }
        	    pmod = plcs [link] [adapter];
		}
		stringaddr = optr->out.sonode.sosignal;
		stringlen  = optr->out.sonode.solen;
		pstringval = optr->out.sonode.sovalue;
		if (mltrace [line] == 7)
		{
                    logMsg("SO IN QUEUE: L=%d, A=%d,       S=%d\n",
                           link, adapter, stringaddr, 0, 0, 0);
		}

		if ((!pmod->cmd_force_offline) &&
                    (resresponse [link]) &&
                    (pmod->is_responding))
		{
                    loopoutmsgs [line]++;
		    if ((setstring (link, adapter, pmod, stringaddr, pstringval, stringlen)) != CMDDONE)
		    {
			pmod->badoutcnt++;
                        if (mltrace [link] == 5)
			{
                            logMsg ("SO DEQUEUE: Output failed: L=%d, A=%d,       S=%d\n",
                                   link,adapter,stringaddr, 0, 0, 0);
			}
		    }
                    else
		    {
                        if (mltrace [link] == 5)
			{
                            logMsg ("SO DEQUEUE: Output succeeded: L=%d, A=%d,       S=%d\n",
                                   link, adapter, stringaddr, 0, 0, 0);
			}
		    }
		}
	    }
	    oreleasefull (line);     /* release the output que entry */
	}
}

/**********
*
*      sv85_aidriver
*
*      This routines main purpose is to return the analog input value for the
*      specified module. For a module value a (-1) will be returned if there
*       is no valid data ready.
*
**********/

/* this routine not called from within this driver */

int sv85_aidriver (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal, short *pvalue,
		unsigned short conversion, unsigned char *resname)
{
	MODPLC		*pmod;
	unsigned short  blkix;
	unsigned short	*pdata;

	if (link > MODLINES)
        {
            if (trace [link] & CONFIGURATIONTRACE)
	    {
                logMsg("sv85_aidriver: Illegal line:[%s] L=%d, A=%d, C=%d, S=%d\n",
                        (int) pointname, link, adapter, card, signal, 0);
	    }
            return (-1);
        }
	if (adapter > NUMMODPLCS)
	{
            if (trace [link] & CONFIGURATIONTRACE)
	    {
                logMsg("sv85_aidriver: Illegal adapter:[%s] L=%d, A=%d, C=%d, S=%d\n",
                        (int) pointname, link, adapter, card, signal, 0);
	    }
            return (-1);
        }
	if ((card != MAINS) && (card != MREGS) && (card != MPSEUDO))
        {
            if (trace [link] & CONFIGURATIONTRACE)
	    {
                logMsg ("sv85_aidriver: Illegal card:[%s] L=%d, A=%d, C=%d, S=%d\n",
                        (int) pointname, link, adapter, card, signal, 0);
	    }
            return (-1);
        }

	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (!allocate_plc (link, adapter))
	    {
		/* if not allocated, try. if can't allocate, get out. */
		return (-1);
	    }
	    pmod = plcs [link] [adapter];
	}

	if (card == MPSEUDO)
	{
	    switch (signal)
	    {
	    case 0: /* Were the PLC values requested */
		*pvalue = pmod->was_requested;
		break;

	    case 1: /* Is the PLC responding */
		*pvalue = pmod->is_responding;
		break;

	    case 2: /* The program Load request */
		*pvalue = pmod->prog_load_stat;
		break;

	    case 3: /* PLC Program Load Block Number */
		*pvalue = pmod->prog_load_page;
		break;

	    case 4: /* Set the program number to Load */
		*pvalue = pmod->prog_load_number;
		break;

	    case 5: /* Output Messages */
		*pvalue = pmod->no_dataout_msgs & 0x7fff;
		break;

	    case 6: /* Input Messages */
		*pvalue = pmod->no_datain_msgs & 0x7fff;
		break;

	    case 7: /* No Receiver Responses */
		*pvalue = pmod->no_rxnoresp & 0x7fff;
		break;

	    case 8: /* No Transmitter Response */
		*pvalue = pmod->no_txnoresp & 0x7fff;
		break;

	    case 9: /* Data Echo Errors */
		*pvalue = pmod->no_dataerr & 0x7fff;
		break;

	    case 10: /* Message Retries */
		*pvalue = pmod->no_retries & 0x7fff;
		break;

	    case 11: /* RX Hardware errors */
		*pvalue = pmod->no_rxhwerrs & 0x7fff;
		break;

	    case 12: /* TX Hardware errors */
		*pvalue = pmod->no_txhwerrs & 0x7fff;
		break;

	    case 13: /* CRC errors */
		*pvalue = pmod->no_crcerrs & 0x7fff;
		break;

	    case 14: /* Offline transitions */
		*pvalue = pmod->no_offlines & 0x7fff;
		break;

	    case 15: /* reset statistics */
		*pvalue = pmod->cmd_reset_stat;
		break;

	    case 16: /* Force the PLC offline */
		*pvalue = pmod->cmd_force_offline;
		break;

	    case 17: /* Failed output commands */
		*pvalue = pmod->badoutcnt & 0x7fff;
		break;

	    case 20: /* What is the program load comp code for the plc? */
		*pvalue = pmod->prog_load_code;
		break;

	    case 21: /* Is the redundant arbitration link active ? */
		*pvalue = partnerAlive [link];
		break;

	    default:
                if (trace[link] & CONFIGURATIONTRACE)
		{
                    logMsg("sv85_aidriver: Illegal Pseudo Point:[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal, 0);
		}
		return (-1);
	    }
	    return (0);
	}

	if (card == MAINS)
	{
            if ((signal > MAXAINO) || (signal < MINAINO))
            {
                if (trace [link] & CONFIGURATIONTRACE)
		{
                    logMsg ("sv85_aidriver: Illegal signal(0):[%s] L=%d, A=%d, C=%d, S=%d,: max=%d\n",
                            (int) pointname, link, adapter, card, signal, MAXAINO);
		}
                return (-1);
            }

            if (resname != 0)
            { 
                register_resource(link,pmod,resname,pointname);
            }

            pmod->was_requested = TRUE;

            signal--;		/* Signals are defined as 1 relative */

            blkix = (signal / REGBLOCK);

            /* check for trans-REGBLOCK boundary access */

            if (((conversion == AICNVDOUBLEINT)  ||
                 (conversion == AICNVFLOAT    )) &&
                ((signal % REGBLOCK) == 3      ))
            {
		/* if at top of register space, give up */

		if (blkix + 1 >= NUMMODAIS)
		{
		    if (trace [link] & CONFIGURATIONTRACE)
		    {
			logMsg ("sv85_aidriver: Point extends beyond legal signal range(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		/* as soon as the address is confirmed as valid, */
		/* set requested bits. */

		pmod->ais [blkix].was_requested = TRUE;
		pmod->ais [blkix + 1].was_requested = TRUE;
		if (!resresponse [link])
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg ("sv85_aidriver: Driver Offline(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		if (pmod->holdflag)
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg ("sv85_aidriver: On Hold(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
                                    (int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-2);
		}
		if (!pmod->is_responding)
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg ("sv85_aidriver: Not Responding(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
                                   (int)  pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		if ((!pmod->ais [blkix].data_valid) || (!pmod->ais [blkix + 1].data_valid))
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg("sv85_aidriver: Data Not Valid(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
                                    (int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		pvalue [1] = pmod->ais [blkix].aivals [3];
		pvalue [0] = pmod->ais [blkix + 1].aivals [0];
		if (mltrace [link] == 8)
		{
		    logMsg("sv85_aidriver: Data Returned(1):[%s] A=%d, C=%d, S=%d, V=0x%x 0x%x\n",
                                (int) pointname, adapter, card, signal + 1, pvalue [0], pvalue [1]);
		}
		return (0);
            }

            /* check for trans-REGBLOCK boundary access */

            if ((conversion == AICNVDOUBLEFLT) &&
                ((signal % REGBLOCK) != 0      ))
            {
		/* if at top of register space, give up */
		if (blkix + 1 >= NUMMODAIS)
		{
		    if (trace [link] & CONFIGURATIONTRACE)
		    {
			logMsg ("sv85_aidriver: Point extends beyond legal signal range(2):[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}

		/* as soon as the address is confirmed as valid, */
		/* set requested bits. */

		pmod->ais [blkix].was_requested = TRUE;
		pmod->ais [blkix + 1].was_requested = TRUE;
		if (!resresponse [link])
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg ("sv85_aidriver: Driver Offline(2):[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		if (pmod->holdflag)
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg("sv85_aidriver: On Hold(2):[%s] L=%d, A=%d, C=%d, S=%d\n",
                                    (int)pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-2);
		}
		if (!pmod->is_responding)
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg("sv85_aidriver: Not Responding(2):[%s] L=%d, A=%d, C=%d, S=%d\n",
                                    (int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		if ((!pmod->ais [blkix].data_valid) || (!pmod->ais [blkix + 1].data_valid))
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg ("sv85_aidriver: Data Not Valid(2):[%s] L=%d, A=%d, C=%d, S=%d\n",
                                    (int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		switch (signal % REGBLOCK)
		{
		case (1):
		    pvalue [3] = pmod->ais [blkix].aivals [1];
		    pvalue [2] = pmod->ais [blkix].aivals [2];
		    pvalue [1] = pmod->ais [blkix].aivals [3];
		    pvalue [0] = pmod->ais [blkix+1].aivals [0];
		    break;

		case (2):
		    pvalue [3] = pmod->ais [blkix].aivals [2];
		    pvalue [2] = pmod->ais [blkix].aivals [3];
		    pvalue [1] = pmod->ais [blkix + 1].aivals [0];
		    pvalue [0] = pmod->ais [blkix + 1].aivals [1];
		    break;

		case (3):
		    pvalue [3] = pmod->ais [blkix].aivals [3];
		    pvalue [2] = pmod->ais [blkix+1].aivals [0];
		    pvalue [1] = pmod->ais [blkix + 1].aivals [1];
		    pvalue [0] = pmod->ais [blkix + 1].aivals [2];
		    break;

		default:
		    return (-1); /* condition can never happen */
		}
		if (mltrace [link] == 8)
		{
		    logMsg("sv85_aidriver: Data Returned(2):[%s] A=%d, C=%d, S=%d, V=0x%x 0x%x\n",
                                (int) pointname, adapter, card, signal + 1, pvalue [0], pvalue [1]);
		}
	  	return (0);
            }

            /* as soon as the address is confirmed as valid, set requested bits. */

	    pmod->ais [blkix].was_requested = TRUE;
            if (!resresponse [link])
            {
                if (trace [link] & CALLREJECTTRACE)
		{
                    logMsg ("sv85_aidriver: Driver Offline(3):[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal + 1, 0);
		}
                return (-1);
            }
            if (pmod->holdflag)
            {
                if (trace [link] & CALLREJECTTRACE)
		{
                    logMsg ("sv85_aidriver: On Hold(3):[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal + 1, 0);
		}
                return (-2);
            }
            if (!pmod->is_responding)
            {
                if (trace [link] & CALLREJECTTRACE)
		{
                    logMsg("sv85_aidriver: Not Responding(3):[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal + 1, 0);
		}
                return (-1);
            }
	    if (!pmod->ais [blkix].data_valid)
            {
               if (trace[link] & CALLREJECTTRACE)
		{
                    logMsg("sv85_aidriver: Data Not Valid(3):[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal + 1, 0);
		}
		return (-1);
	    }
	    pdata = (unsigned short *) &(pmod->ais [blkix].aivals [signal % REGBLOCK]);
	    if (mltrace [link] == 8)
	    {
		logMsg ("sv85_aidriver: Data Returned(3):[%s] A=%d, C=%d, S=%d, V=0x%x 0x%x\n",
                            (int) pointname, adapter, card, signal + 1, pvalue [0], pvalue [1]);
	    }
	    /* Select the Size and Type of Response based on Conversion */
	    switch (conversion)
	    {
	        case (AICNVDOUBLEINT):
		     pvalue [0] = pdata [1];
		     pvalue [1] = pdata [0];
		     break;

	        case (AICNVFLOAT):
		     pvalue [0] = pdata [1];
		     pvalue [1] = pdata [0];
		     break;

	        case (AICNVDOUBLEFLT):
		     pvalue [0] = pdata [3];
		     pvalue [1] = pdata [2];
		     pvalue [2] = pdata [1];
		     pvalue [3] = pdata [0];
		     break;

	        case (AICNVINTEGER):
	        default:
		     *pvalue = *pdata;
		     break;
            }
            if (mltrace [link] == 8)
	    {
                logMsg ("sv85_aidriver: Data Returned(3):[%s] A=%d, C=%d, S=%d, V=0x%x 0x%x\n",
                        (int) pointname, adapter, card, signal + 1, pvalue [0], pvalue [1]);
	    }
            return (0);
	}

	if (card == MREGS)
	{
            if ((signal > MAXREGNO) || (signal < MINREGNO))
            {
                if (trace [link] & CONFIGURATIONTRACE)
		{
                    logMsg ("sv85_aidriver: Illegal signal(0):[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal, 0);
		}
                return (-1);
            }

            if (resname != 0)
            { 
                register_resource (link, pmod, resname, pointname);
            }

            pmod->was_requested = TRUE;

            signal--;			/* Signals are defined as 1 relative */
            blkix = (signal / REGBLOCK);

            /* check for trans-REGBLOCK boundary access */

            if (((conversion == AICNVDOUBLEINT)  ||
                 (conversion == AICNVFLOAT)) &&
                ((signal % REGBLOCK) == 3))
            {
		if (blkix + 1 >= NUMREGISTERS)   /* if at top of register space, give up */
		{
		    if (trace [link] & CONFIGURATIONTRACE)
		    {
			logMsg ("sv85_aidriver: Point extends beyond legal signal range(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
                                    (int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		/* as soon as the address is confirmed as valid, set requested bits. */
		pmod->regs [blkix].was_requested = TRUE;
		pmod->regs [blkix + 1].was_requested = TRUE;

		if (!resresponse [link])
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
		    
			logMsg ("sv85_aidriver: Driver Offline(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
                                    (int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		if (pmod->holdflag)
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg("sv85_aidriver: On Hold(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
                                    (int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-2);
		}
		if (!pmod->is_responding)
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg ("sv85_aidriver: Not Responding(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
                                    (int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		if ((!pmod->regs [blkix].data_valid) || (!pmod->regs [blkix+1].data_valid))
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg ("sv85_aidriver: Data Not Valid(1):[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		pvalue [1] = pmod->regs [blkix].ioregs [3];
		pvalue [0] = pmod->regs [blkix + 1].ioregs [0];
		if (mltrace [link] == 8)
		{
		    logMsg("sv85_aidriver: Data Returned(2):[%s] A=%d, C=%d, S=%d, V=0x%x 0x%x\n",
			(int) pointname, adapter, card, signal + 1, pvalue [0], pvalue [1]);
		}
		return (0);
            }
            /* check for trans-REGBLOCK boundary access */
            if ((conversion == AICNVDOUBLEFLT) &&
                ((signal % REGBLOCK) != 0))
            {
		if (blkix + 1 >= NUMREGISTERS)   /* if at top of register space, give up */
		{
		    if (trace [link] & CONFIGURATIONTRACE)
		    {
			logMsg ("sv85_aidriver: Point extends beyond legal signal range(2):[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}

		/* as soon as the address is confirmed as valid, set requested bits. */

		pmod->regs [blkix].was_requested = TRUE;
		pmod->regs [blkix + 1].was_requested = TRUE;
		if (!resresponse [link])
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg ("sv85_aidriver: Driver Offline(2):[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		if (pmod->holdflag)
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg ("sv85_aidriver: On Hold(2):[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-2);
		}
		if (!pmod->is_responding)
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg ("sv85_aidriver: Not Responding(2):[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		if ((!pmod->regs [blkix].data_valid) || (!pmod->regs [blkix + 1].data_valid))
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg ("sv85_aidriver: Data Not Valid(2):[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		switch (signal % REGBLOCK)
		{
		case (1):
		    pvalue [3] = pmod->regs [blkix].ioregs [1];
		    pvalue [2] = pmod->regs [blkix].ioregs [2];
		    pvalue [1] = pmod->regs [blkix].ioregs [3];
		    pvalue [0] = pmod->regs [blkix + 1].ioregs [0];
		    break;

		case (2):
		    pvalue [3] = pmod->regs [blkix].ioregs [2];
		    pvalue [2] = pmod->regs [blkix].ioregs [3];
		    pvalue [1] = pmod->regs [blkix + 1].ioregs [0];
		    pvalue [0] = pmod->regs [blkix + 1].ioregs [1];
		    break;

		case (3):
		    pvalue [3] = pmod->regs [blkix].ioregs [3];
		    pvalue [2] = pmod->regs [blkix + 1].ioregs [0];
		    pvalue [1] = pmod->regs [blkix + 1].ioregs [1];
		    pvalue [0] = pmod->regs [blkix + 1].ioregs [2];
		    break;

		default:
		    return (-1); /* condition can never happen */
		}
		if (mltrace [link] == 8)
		{
		    logMsg ("sv85_aidriver: Data Returned(2):[%s] A=%d, C=%d, S=%d, V=0x%x 0x%x\n",
			(int) pointname, adapter, card, signal + 1, pvalue [0], pvalue [1]);
		}
		return (0);
	    }

	    /* as soon as the address is confirmed as valid, set requested bits. */

	    pmod->regs [blkix].was_requested = TRUE;
	    if (!resresponse [link])
	    {
		if (trace [link] & CALLREJECTTRACE)
		{
		    logMsg ("sv85_aidriver: Driver Offline(3):[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal + 1, 0);
		}
		return (-1);
	    }
	    if (pmod->holdflag)
	    {
		if (trace [link] & CALLREJECTTRACE)
		{
		    logMsg ("sv85_aidriver: On Hold(3):[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal + 1, 0);
		}
		return (-2);
	    }
	    if (!pmod->is_responding)
	    {
		if (trace [link] & CALLREJECTTRACE)
		{
		    logMsg ("sv85_aidriver: Not Responding(3):[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal + 1, 0);
		}
		return (-1);
	    }
	    if (!pmod->regs [blkix].data_valid)
	    {
		if (trace [link] & CALLREJECTTRACE)
		{
		    logMsg("sv85_aidriver: Data Not Valid(3):[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal + 1, 0);
		}
		return (-1);
	    }
	    pdata = &(pmod->regs [blkix].ioregs [signal % REGBLOCK]);

	    /* Select the Size and Type of Response based on Conversion */

	    switch (conversion)
	    {
	    case (AICNVDOUBLEINT):
		pvalue [0] = pdata [1];
		pvalue [1] = pdata [0];
		break;

	    case (AICNVFLOAT):
		pvalue [0] = pdata [1];
		pvalue [1] = pdata [0];
		break;

	    case (AICNVDOUBLEFLT):
		pvalue [0] = pdata [3];
		pvalue [1] = pdata [2];
		pvalue [2] = pdata [1];
		pvalue [3] = pdata [0];
		break;

	    case (AICNVINTEGER):
	    default:
		*pvalue = *pdata;
		break;
            }
	    if (mltrace [link] == 8)
	    {
                logMsg ("sv85_aidriver: Data Returned(3):[%s] A=%d, C=%d, S=%d, V=0x%x 0x%x\n",
                        (int) pointname, adapter, card, signal + 1, pvalue [0], pvalue [1]);
	    }
            return (0);
	}
	return (-1);
}


/**********
*  										
*      sv85_aodriver                                                                
*                                                                            	
*      This routines main purpose is to set requested analog outputs to the  	
*      specified module.                                                     	
*
**********/                                                                            	

/* this routine not called from within this driver */

int sv85_aodriver (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal, double dvalue,
		unsigned short value, unsigned short conversion,
		unsigned char *resname)

{
	MODPLC		*pmod;
	unsigned short  blkix;
	OUTQENT		*oqp;

	if (link > MODLINES)
        {
            if (trace [link] & CONFIGURATIONTRACE)
	    {
                logMsg ("sv85_aodriver: Illegal line:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                        (int) pointname, link, adapter, card, signal, value);
	    }
            return (-1);
        }
	if (adapter > NUMMODPLCS)
	{
            if (trace [link] & CONFIGURATIONTRACE)
	    {
                logMsg ("sv85_aodriver: Illegal adapter:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                        (int) pointname, link, adapter, card, signal, value);
	    }
            return (-1);
        }
	if ((card != MREGS) && (card != MPSEUDO))
        {
            if (trace [link] & CONFIGURATIONTRACE)
	    {
                logMsg ("sv85_aodriver: Illegal card:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                        (int) pointname, link, adapter, card, signal, value);
	    }
            return (-1);
        }

	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (!allocate_plc (link, adapter))
	    {
		return (-1); /* if not allocated, try. if can't allocate, get out. */
	    }
	    pmod = plcs [link] [adapter];
	}
	if (card == MPSEUDO)
	{
	    switch (signal)
	    {
	    case 2: /* Set the program Load request */
		if (value != 0)
		{
		    pmod->prog_load_stat = PLC_LOAD_REQUESTED;
		}
		break;

	    case 4: /* Set the program number to Load */
		pmod->prog_load_number = value;
		break;

	    case 15: /* Reset the accumulated statistics */
		if (value != 0)
		{
		    initstats (pmod);
		}
		break;

	    case 16: /* Force the PLC online/offline */
		pmod->cmd_force_offline = (value != 0);
                if (value != 0)
		{
                    pmod->is_responding = FALSE;/* only required when asserting cmd_force_offline */
		}
		break;

	    default:
                if (trace [link] & CONFIGURATIONTRACE)
		{
                    logMsg ("sv85_aodriver: Illegal Pseudo Point:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                            (int) pointname, link, adapter, card, signal, value);
		}
		return (-1);
	    }
	    return (0);
	}

        if ((signal > MAXREGNO) || (signal < MINREGNO))
	{
            if (trace [link] & CONFIGURATIONTRACE)
	    {
                logMsg ("sv85_aodriver: Illegal signal:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                        (int) pointname, link, adapter, card, signal, value);
	    }
	    return (-1);
	}
	if (resname != 0)
	{ 
	    register_resource (link, pmod, resname, pointname);
	}

	pmod->was_requested = TRUE;
	pmod->red_update = TRUE;

	signal--; /* Signals are defined as 1 relative */
	blkix = (signal / REGBLOCK);

	pmod->regs [blkix].was_requested = TRUE;
	pmod->regs [blkix].red_update = TRUE;

        /* check for trans-REGBLOCK boundary access */

        if ((conversion == ASCFLT) && ((signal % REGBLOCK) == 3))
        {
	    /* if at top of register space, give up */

            if (blkix + 1 >= NUMREGISTERS)
            {
                if (trace [link] & CONFIGURATIONTRACE)
		{
                    logMsg ("sv85_aodriver: Point extends beyond legal signal range:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                            (int) pointname, link, adapter, card, signal + 1, value);
		}
                return (-1);
            }

            /* as soon as the address is confirmed as valid,
		set next block's requested bits. */

            pmod->regs [blkix + 1].was_requested = TRUE;
            pmod->regs [blkix + 1].red_update = TRUE;
        }

	if (!resresponse [link])
        {
            if (trace [link] & CALLREJECTTRACE)
	    {
                logMsg ("sv85_aodriver: Driver Offline:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                        (int) pointname, link, adapter, card, signal + 1, value);
	    }
            return (-1);
        }

        /* check for plc simulated mode. */

        if (modsim [link])
        {
            if (mltrace [link] == 3)
	    {
                logMsg ("sv85_aodriver: SimReturn:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                        (int) pointname, link, adapter, card, signal + 1, value);
	    }
            return (0);
        }

	if ((!pmod->is_responding) || (pmod->cmd_force_offline))
	{
            if (trace [link] & CALLREJECTTRACE)
	    {
                logMsg ("sv85_aodriver: Not Responding or Forced offline:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                        (int) pointname, link, adapter, card, signal + 1, value);
	    }
	    return (-1);
	}

	oqp = ogetemptyp (link);
	if (oqp == NULLP)
	{
            if (trace [link] & RUNTIMEERRORTRACE)
            {
                logMsg("sv85_aodriver: Queue Overflow:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                        (int) pointname, link, adapter, card, signal + 1, value);
	    }
	    pmod->badoutcnt++;
	    return (-1);		/* output queue is full */
	}
	oqp->otype = AOUTTYPE;
	oqp->out.aonode.link = link;
	oqp->out.aonode.adapter = adapter;
	oqp->out.aonode.card = card;
	oqp->out.aonode.aosignal = signal;
	oqp->out.aonode.aovalue = value;
	oqp->out.aonode.dblaovalue = dvalue;
	oqp->out.aonode.conversion = conversion;

	omarkfull (link);		/* Mark this entry as ready for output */

        if (mltrace [link] == 3)
	{
            logMsg("sv85_aodriver: Queued:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                    (int) pointname, link, adapter, card, signal + 1, value);
	}
	return (0);
}

/**********
*                                                                             
*      sv85_aoread                                                                  
*                                                                          	
*      This routines main purpose is to return the value last read back     	
*      from the analog output hardware. This is used to initialize analog   	
*      outputs.                                                             	
*
**********/

/* this routine not called from within this driver */

int sv85_aoread (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal,
		unsigned short *pvalue, unsigned short conversion,
		unsigned char *resname)

{
	MODPLC		*pmod;
	unsigned short  blkix, *pdata;
	
	if (link > MODLINES)
        {
            if (trace [link] & CONFIGURATIONTRACE)
	    {
                logMsg ("sv85_aoread: Illegal line:[%s] L=%d, A=%d, C=%d, S=%d\n",
                       (int) pointname, link, adapter, card, signal, 0);
	    }
            return (-1);
        }
	if (adapter > NUMMODPLCS)
	{
            if (trace [link] & CONFIGURATIONTRACE)
	    {
                logMsg("sv85_aoread: Illegal adapter:[%s] L=%d, A=%d, C=%d, S=%d\n",
                        (int) pointname, link, adapter, card, signal, 0);
	    }
            return (-1);
        }
	if ((card != MREGS) && (card != MPSEUDO))
        {
            if (trace [link] & CONFIGURATIONTRACE)
	    {
                logMsg("sv85_aoread: Illegal card:[%s] L=%d, A=%d, C=%d, S=%d\n",
                        (int) pointname, link, adapter, card, signal, 0);
	    }
            return (-1);
        }

	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (!allocate_plc (link, adapter))
	    {
		/* if not allocated, try. if can't allocate, get out. */
                return (-1);
	    }
	    pmod = plcs [link] [adapter];
	}

	if (card == MPSEUDO)
	{
	    switch (signal)
	    {
	    case 2: /* read the most recent program Load request */
		*pvalue = pmod->prog_load_stat;
		break;

	    case 4: /* read the most recent program number to Load */
		*pvalue = pmod->prog_load_number;
		break;

	    case 15: /* read statistics command value */
		*pvalue = pmod->cmd_reset_stat;
		break;

	    case 16: /* Read the current PLC offline command status */
		*pvalue = pmod->cmd_force_offline;
		break;

	    default:
                if (trace [link] & CONFIGURATIONTRACE)
		{
                    logMsg ("sv85_aoread: Illegal Pseudo Point:[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal, 0);
		}
		return(-1);
	    }
	    return (0);
	}

	if (card == MREGS)
	{
            if ((signal > MAXREGNO) || (signal < MINREGNO))
	    {
                if (trace [link] & CONFIGURATIONTRACE)
		{
                    logMsg("sv85_aoread: Illegal signal:[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal, 0);
		}
		return (-1);
	    }

            if (resname != 0)
            { 
                register_resource (link, pmod, resname, pointname);
            }

            pmod->was_requested = TRUE;

            signal--;			/* Signals are defined as 1 relative */
            blkix = (signal / REGBLOCK);

            /* check for trans-REGBLOCK boundary access */

            if (  (conversion == ASCFLT)  &&
                  ((signal % REGBLOCK) == 3))
            {
		/* if at top of register space, give up */

		if (blkix + 1 >= NUMREGISTERS)
		{
		    if (trace [link] & CONFIGURATIONTRACE)
		    {
			logMsg ("sv85_aoread: Point extends beyond legal signal range:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		/* as soon as the address is confirmed as valid,
			set requested bits. */

		pmod->regs [blkix].was_requested = TRUE;
		pmod->regs [blkix + 1].was_requested = TRUE;
		if (!resresponse [link])
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg ("sv85_aoread: Driver Offline:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		if (pmod->holdflag)
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg ("sv85_aoread: On Hold:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-2);
		}
		if (!pmod->is_responding)
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg ("sv85_aoread: Not Responding:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		if ((!pmod->regs [blkix].data_valid) || (!pmod->regs [blkix + 1].data_valid))
		{
		    if (trace [link] & CALLREJECTTRACE)
		    {
			logMsg("sv85_aoread: Data Not Valid:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		    }
		    return (-1);
		}
		pvalue [1] = pmod->regs [blkix].ioregs [3];
		pvalue [0] = pmod->regs [blkix + 1].ioregs [0];
		if (mltrace [link] == 8)
		{
		    logMsg ("sv85_aoread: Data Returned:[%s] A=%d, C=%d, S=%d, V=0x%x 0x%x\n",
			(int) pointname, adapter, card, signal + 1, pvalue [0], pvalue [1]);
		}
		return(0);
            }

            /* as soon as the address is confirmed as valid, set requested bits. */

	    pmod->regs [blkix].was_requested = TRUE;
            if (!resresponse [link])
            {
                if (trace [link] & CALLREJECTTRACE)
		{
                    logMsg ("sv85_aoread: Driver Offline:[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal + 1, 0);
		}
                return (-1);
            }
            if (pmod->holdflag)
            {
                if (trace [link] & CALLREJECTTRACE)
		{
                    logMsg ("sv85_aoread: On Hold:[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal + 1, 0);
		}
                return (-2);
            }
            if (!pmod->is_responding)
            {
                if (trace [link] & CALLREJECTTRACE)
		{
                    logMsg ("sv85_aoread: Not Responding:[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal + 1, 0);
		}
                return (-1);
            }
	    if (!pmod->regs [blkix].data_valid)
	    {
		if (trace [link] & CALLREJECTTRACE)
		{
		    logMsg ("sv85_aoread: Data Not Valid:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal + 1, 0);
		}
		return (-1);
	    }
	    pdata = &(pmod->regs [blkix].ioregs [signal % REGBLOCK]);

	    /* Select the Size and Type of Response based on Conversion */

	    switch (conversion)
	    {
	    case (ASCFLT):
		pvalue [0] = pdata [1];
		pvalue [1] = pdata [0];
		break;

	    default:
		*pvalue = *pdata;
		break;
            }
            if (mltrace [link] == 8)
	    {
                logMsg ("sv85_aoread: Data Returned:[%s] A=%d, C=%d, S=%d, V=0x%x 0x%x\n",
                        (int) pointname, adapter, card, signal + 1, pvalue [0], pvalue [1]);
	    }
	    return (0);
	}
	return (-1);
}


/**********
*
*      sv85_bidriver
*
*      The main purpose of this routine is to return the request digital input
*      value(s) from the requested card. This routine will return a (-1) if the
*      requested module is the wrong type or the data is not yet valid.
*
**********/

/* this routine not called from within this driver */

int sv85_bidriver (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal,
		unsigned long mask, unsigned long *pvalue,
		unsigned char *resname)
{
	MODPLC 		*pmod;
	unsigned short	blkix;

	if (link > MODLINES)
        {
            if (trace [link] & CONFIGURATIONTRACE)
	    {
                logMsg ("sv85_bidriver: Illegal line:[%s] L=%d, A=%d, C=%d, S=%d\n",
                        (int) pointname, link, adapter, card, signal, 0);
	    }
            return (-1);
        }
	if (adapter > NUMMODPLCS)
        {
            if (trace [link] & CONFIGURATIONTRACE)
	    {
                logMsg("sv85_bidriver: Illegal adapter:[%s] L=%d, A=%d, C=%d, S=%d\n",
                        (int) pointname, link, adapter, card, signal, 0);
	    }
            return (-1);
        }
	if ((card != MDINS) && (card != MCOILS) && (card != MPSEUDO) && (card != MPSEUDO1))
        {
            if (trace [link] & CONFIGURATIONTRACE)
	    {
                logMsg ("sv85_bidriver: Illegal card:[%s] L=%d, A=%d, C=%d, S=%d\n",
                        (int) pointname, link, adapter, card, signal, 0);
	    }
            return (-1);
        }

	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (!allocate_plc (link, adapter))
	    {
		/* if not allocated, try. if can't allocate, get out. */
		return (-1);
	    }
	    pmod = plcs [link] [adapter];
	}

	pmod->was_requested = TRUE;

	if (card == MPSEUDO)
	{
	    switch (signal)
	    {
	    case 0: /* Were the PLC values requested */
		*pvalue = pmod->was_requested;
		break;

	    case 1: /* Is the PLC responding */
		*pvalue = pmod->is_responding;
		break;

	    case 15: /* reset statistics */
		*pvalue = pmod->cmd_reset_stat;
		break;

	    case 16: /* Force the PLC offline */
		*pvalue = pmod->cmd_force_offline;
		break;

	    case 20: /* What is the selected channel for the line */
		*pvalue = pmod->prog_load_code;
		break;

	    case 21: /* Is the redundant arbitration link active ? */
		*pvalue = partnerAlive [link] ? TRUE : FALSE;
		break;

	    default:
		if (trace [link] & CONFIGURATIONTRACE)
		{
		    logMsg("sv85_bidriver: Illegal Pseudo Point:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal, 0);
		}
		return (-1);
            }
            return (0);
        }
	if (card == MPSEUDO1)
	{
	    if ((signal < 4) || (signal > 7))
	    {
                if (trace [link] & CONFIGURATIONTRACE)
		{
                    logMsg("sv85_bidriver: Illegal signal:[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal, 0);
		}
                return (-1);
            }
            if (resname != 0)
            { 
                register_resource (link, pmod, resname, pointname);
            }

            pmod->was_requested = TRUE;

            if (!resresponse [link])
            {
                if (trace [link] & CALLREJECTTRACE)
		{
		    logMsg ("sv85_bidriver: Driver Offline:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		}
                return (-1);
            }
            if (pmod->holdflag)
            {
                if (trace [link] & CALLREJECTTRACE)
		{
		    logMsg ("sv85_bidriver: On Hold:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		}
                return (-2);
            }
            if (!pmod->is_responding)
            {
		if (trace [link] & CALLREJECTTRACE)
		{
		    logMsg ("sv85_bidriver: Not Responding:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		}
                return (-1);
            }
	    switch (signal)
	    {
	    case 4:
		if (pmod->battery_okay & 0x02)
		{
		    return (-1);
		}
		else
		{
		    *pvalue = (pmod->battery_okay & 0x01);
		}
		break;

	    case 5:
		if (pmod->memory_protect & 0x02)
		{
		    return (-1);
		}
		else
		{
		    *pvalue = (pmod->memory_protect & 0x01);
		}
		break;

	    case 6:
		if (pmod->run_light & 0x02)
		{
		    return (-1);
		}
		else
		{
		    *pvalue = (pmod->run_light & 0x01);
		}
		break;

	    case 7:
		if (pmod->power_okay & 0x02)
		{
		    return (-1);
		}
		else
		{
		    *pvalue = (pmod->power_okay & 0x01);
		}
		break;
	    }
	    return (0);
	}

	if (card == MDINS)
	{
            if ((signal > MAXDINO) || (signal < MINDINO))
            {
                if (trace [link] & CONFIGURATIONTRACE)
		{
                    logMsg("sv85_bidriver: Illegal signal:[%s] L=%d, A=%d, C=%d, S=%d\n",
                            (int) pointname, link, adapter, card, signal, 0);
		}
                return (-1);
            }
            if (resname != 0)
            { 
                register_resource (link, pmod, resname, pointname);
            }

            pmod->was_requested = TRUE;

            signal--;			/* Signals are defined as 1 relative */
            blkix = (signal / BITBLOCK);	/* BITBLOCK bits per digital BLOCK */

	    pmod->dis [blkix].was_requested = TRUE;
            if (!resresponse [link])
            {
                if (trace [link] & CALLREJECTTRACE)
		{
		    logMsg ("sv85_bidriver: Driver Offline:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		}
                return (-1);
            }
            if (pmod->holdflag)
            {
                if (trace [link] & CALLREJECTTRACE)
		{
		    logMsg ("sv85_bidriver: On Hold:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		}
                return (-2);
            }
            if (!pmod->is_responding)
            {
		if (trace [link] & CALLREJECTTRACE)
		{
		    logMsg ("sv85_bidriver: Not Responding:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		}
                return (-1);
            }
	    if (!pmod->dis [blkix].data_valid)
	    {
		if (trace [link] & CALLREJECTTRACE)
		{
		    logMsg ("sv85_bidriver: Data Not Valid:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		}
		return (-1);
	    }
	    *pvalue = pmod->dis [blkix].last_read & mask;
            if (mltrace [link] == 8)
	    {
                logMsg ("sv85_bidriver: Data Returned:[%s] L=%d, A=%d, C=%d, S=%d, V=0x%x\n",
                        (int) pointname, link, adapter, card, signal + 1, (int) pvalue);
	    }
	    return (0);
	}
	if (card == MCOILS)
	{
	    if ((signal > MAXCOILNO) || (signal < MINCOILNO))
	    {
		if (trace [link] & CONFIGURATIONTRACE)
		{
		    logMsg ("sv85_bidriver: Illegal signal:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal, 0);
		}
		return (-1);
	    }
	    if (resname != 0)
	    { 
		register_resource (link, pmod, resname, pointname);
	    }

	    pmod->was_requested = TRUE;

	    signal--;				/* Signals are defined as 1 relative */
	    blkix = (signal / BITBLOCK);	/* BITBLOCK bits per digital BLOCK */

	    pmod->coils [blkix].was_requested = TRUE;
	    if (!resresponse [link])
            {
		if (trace [link] & CALLREJECTTRACE)
		{
		    logMsg ("sv85_bidriver: Driver Offline:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		}
		return (-1);
	    }
	    if (pmod->holdflag)
	    {
		if (trace [link] & CALLREJECTTRACE)
		{
		    logMsg ("sv85_bidriver: On Hold:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		}
		return (-2);
	    }
	    if (!pmod->is_responding)
	    {
		if (trace [link] & CALLREJECTTRACE)
		{
		    logMsg ("sv85_bidriver: Not Responding:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		}
		return (-1);
	    }
	    if (!pmod->coils [blkix].data_valid)
	    {
		if (trace [link] & CALLREJECTTRACE)
		{
		    logMsg ("sv85_bidriver: Data Not Valid:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
		}
		return (-1);
	    }
	    *pvalue = pmod->coils [blkix].last_read & mask;
	    if (mltrace [link] == 9)
	    {
		logMsg ("sv85_bidriver: Data Returned:[%s] L=%d, A=%d, C=%d, S=%d, V=0x%x\n",
			(int) pointname, link, adapter, card, signal + 1, (int) pvalue);
	    }
	    return (0);
	}
	return (-1);
}

/**********
*
*      sv85_bodriver
*
*      The main purpose of this routine is to set the request digital output
*      value(s) on the requested card. This routine will return a (-1) if a
*      requested module is the wrong type or not configured yet.
*
**********/

/* this routine not called from within this driver */

int sv85_bodriver (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal,
		unsigned long value, unsigned long mask,
		unsigned char *resname)
{
	MODPLC 		*pmod;
	OUTQENT		*oqp;
	unsigned short  blkix;

	if (link > MODLINES)
	{
	    if (trace [link] & CONFIGURATIONTRACE)
	    {
		logMsg ("sv85_bodriver: Illegal line:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
			(int) pointname, link, adapter, card, signal, value);
	    }
	    return (-1);
	}
	if (adapter > NUMMODPLCS)
	{
	    if (trace [link] & CONFIGURATIONTRACE)
	    {
		logMsg ("sv85_bodriver: Illegal adapter:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
			(int) pointname, link, adapter, card, signal, value);
	    }
	    return (-1);
	}
	if ((card != MCOILS) && (card != MPSEUDO))
	{
	    if (trace [link] & CONFIGURATIONTRACE)
	    {
		logMsg ("sv85_bodriver: Illegal card:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
			(int) pointname, link, adapter, card, signal, value);
	    }
	    return (-1);
	}

	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (!allocate_plc (link, adapter))
	    {
		/* if not allocated, try. if can't allocate, get out. */
		return(-1);
	    }
	    pmod = plcs [link] [adapter];
	}

	if (card == MPSEUDO)
	{
	    switch (signal & 0x1f)
	    {

	    case 2: /* Set the program Load request */
		if (value != 0)
		{
		    pmod->prog_load_stat = PLC_LOAD_REQUESTED;
		}
		break;

	    case 15: /* Reset the accumulated statistics */
		if (value != 0)
		{
		    initstats (pmod);
		}
		break;

	    case 16: /* Force the PLC offline */
		pmod->cmd_force_offline = (value != 0);
                if (value != 0)
		{
                    pmod->is_responding = FALSE;/* only required when asserting cmd_force_offline */
		}
		break;

	    case 21: /* Start the PLC */
		pmod->plc_start = (value != 0);
		break;

	    case 22: /* What is the selected channel for the line */
		pmod->plc_stop = (value != 0);
		break;

	    default:
		if (trace [link] & CONFIGURATIONTRACE)
		{
		    logMsg ("sv85_bodriver: Illegal Pseudo Point:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
			(int) pointname, link, adapter, card, signal, value);
		}
		return (-1);
	    }
	    return (0);
	}

	if ((signal > MAXCOILNO) || (signal < MINCOILNO))
	{
	    if (trace [link] & CONFIGURATIONTRACE)
	    {
		logMsg ("sv85_bodriver: Illegal signal:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
			(int) pointname, link, adapter, card, signal, value);
	    }
	    return (-1);
	}

	if (resname != 0)
	{ 
	    register_resource (link, pmod, resname, pointname);
	}

	pmod->was_requested = TRUE;
	pmod->red_update = TRUE;

	signal--;				/* Signals are defined as 1 relative */
	blkix = (signal / BITBLOCK);		/* BITBLOCK bits per digital BLOCK */
						/* masks aligned on 32 bits */

	pmod->coils [blkix].was_requested = TRUE;
	pmod->coils [blkix].red_update = TRUE;

	if (!resresponse[link])
	{
	    if (trace [link] & CALLREJECTTRACE)
	    {
		logMsg ("sv85_bodriver: Driver Offline:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
			(int) pointname, link, adapter, card, signal + 1, value);
	    }
            return (-1);
        }

	if (modsim [link])
        {
	    if (mltrace [link] == 4)
	    {
		logMsg("sv85_bodriver: SimReturn:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
			(int) pointname, link, adapter, card, signal + 1, value);
	    }
	    return (0);
	}
	if ((!pmod->is_responding) || (pmod->cmd_force_offline))
	{
	    if (trace [link] & CALLREJECTTRACE)
	    {
                logMsg ("sv85_bodriver: Not Responding or Forced offline:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
			(int) pointname, link, adapter, card, signal+1, value);
	    }
	    return (-1);
	}

	oqp = ogetemptyp (link);
	if (oqp == NULLP)
	{
	    if (trace [link] & RUNTIMEERRORTRACE)
	    {
		logMsg ("sv85_bodriver: Queue Overflow:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
			(int) pointname, link, adapter, card, signal + 1, value);
	    }
	    pmod->badoutcnt++;
	    return (-1);		/* output queue is full */
	}
	oqp->otype = BOUTTYPE;
	oqp->out.bonode.link = link;
	oqp->out.bonode.adapter = adapter;
	oqp->out.bonode.card = card;
	oqp->out.bonode.bosignal = signal;
	oqp->out.bonode.bomask = mask;
	oqp->out.bonode.bovalue = value;

	omarkfull (link);		/* Mark this entry as ready for output */

	if (mltrace [link] == 4)
	{
            logMsg ("sv85_bodriver: Queued:[%s] L=%d, A=%d, C=%d, S=%d, V=%d\n",
                    (int) pointname, link, adapter, card, signal + 1, value);
	}
	return (0);
}

/**********
*
*      sv85_boread
*
*      This routines main purpose is to return the value last read back
*      from the digital output hardware. This is used to initialize digital
*      outputs. This will return a (-1) if the module is the wrong type or
*      the data is not yet valid
*
**********/

/* this routine not called from within this driver */

int sv85_boread (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal,
		unsigned long *pvalue, unsigned long mask,
		unsigned char *resname)
{
	MODPLC		*pmod;
	unsigned short  blkix;

	if (link > MODLINES)
	{
	    if (trace [link] & CONFIGURATIONTRACE)
	    {
		logMsg ("sv85_boread: Illegal line:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal, 0);
	    }
	    return (-1);
        }
	if (adapter > NUMMODPLCS)
	{
	    if (trace [link] & CONFIGURATIONTRACE)
	    {
		logMsg ("sv85_boread: Illegal adapter:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal, 0);
	    }
	return (-1);
	}
	if ((card != MCOILS) && (card != MPSEUDO))
	{
	    if (trace [link] & CONFIGURATIONTRACE)
	    {
		logMsg ("sv85_boread: Illegal card:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal, 0);
	    }
	    return (-1);
	}

	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (!allocate_plc (link, adapter))
	    {
		/* if not allocated, try. if can't allocate, get out. */
		return (-1);
	    }
	    pmod = plcs [link] [adapter];
	}

	if (card == MPSEUDO)
	{
	    switch (signal)
	    {
	    case 2: /* Set the program Load request */
		*pvalue = 0;
		break;

	    case 15: /* reset statistics */
		*pvalue = pmod->cmd_reset_stat;
		break;

	    case 16: /* Force the PLC offline */
		*pvalue = pmod->cmd_force_offline;
		break;

	    case 21: /* Start the PLC */
		*pvalue = pmod->plc_start;
		break;

	    case 22: /* What is the selected channel for the line */
		*pvalue = pmod->plc_stop;
		break;

	    default:
		if (trace [link] & CONFIGURATIONTRACE)
		{
		    logMsg ("sv85_boread: Illegal Pseudo Point:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal, 0);
		}
		return (-1);
	    }
	    return (0);
	}

	if ((signal > MAXCOILNO) || (signal < MINCOILNO))
	{
	    if (trace [link] & CONFIGURATIONTRACE)
	    {
		logMsg ("sv85_boread: Illegal signal:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal + 1, 0);
	    }
	    return (-1);
	}

	if (resname != 0)
	{
	    register_resource (link, pmod, resname, pointname);
	}

	pmod->was_requested = TRUE;

	signal--;				/* Signals are defined as 1 relative */
	blkix = (signal / BITBLOCK);		/* BITBLOCK bits per digital BLOCK */
						/* masks aligned on 32 bits */
	pmod->coils [blkix].was_requested = TRUE;
	if (!resresponse [link])
	{
	    if (trace [link] & CALLREJECTTRACE)
	    {
		logMsg ("BIREAD: Driver Offline:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal + 1, 0);
	    }
	    return (-1);
	}
	if (pmod->holdflag)
	{
	    if (trace [link] & CALLREJECTTRACE)
	    {
		logMsg ("BIREAD: On Hold:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal + 1, 0);
	     }
	    return (-2);
	}
	if (!pmod->is_responding)
	{
	    if (trace [link] & CALLREJECTTRACE)
	    {
		logMsg ("BIREAD: Not Responding:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal + 1, 0);
	    }
	    return (-1);
	}
	if (!pmod->coils [blkix].data_valid)
	{
	    if (trace [link] & CALLREJECTTRACE)
	    {
		logMsg ("sv85_boread: Data Not Valid:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal + 1, 0);
	    }
	    return (-1);
	}
	*pvalue = pmod->coils [blkix].last_read & mask;
	if (mltrace [link] == 9)
	{
	    logMsg ("sv85_boread: Data Returned:[%s] L=%d, A=%d, C=%d, S=%d, V=0x%x\n",
			(int) pointname, link, adapter, card, signal + 1, (int) pvalue);
	}
	return (0);
}

/**********
*
*      sv85_sidriver
*
*      This routines main purpose is to return the analog input value for the
*      specified module. For a module value
*      a (-1) will be returned if there is no valid data ready.
*
**********/

/* this routine not called from within this driver */

int sv85_sidriver (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal,
		short *pvalue, unsigned short len,
		unsigned char *resname)
{
	MODPLC          *pmod;
	short           oldblkix, blkix, regix;
        unsigned short  reglen;
        int             i;

	if (link > MODLINES)
	{
	    if (trace [link] & CONFIGURATIONTRACE)
	    {
		logMsg ("sv85_sidriver: Illegal line:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal, 0);
	    }
	    return (-1);
	}
	if (adapter > NUMMODPLCS)
	{
	    if (trace [link] & CONFIGURATIONTRACE)
	    {
		logMsg ("sv85_sidriver: Illegal adapter:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal, 0);
	    }
	    return (-1);
	}
	if ((card != MAINS) && (card != MREGS) && (card != MPSEUDO))
	{
	    if (trace [link] & CONFIGURATIONTRACE)
	    {
		logMsg ("sv85_sidriver: Illegal card:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal, 0);
	    }
	    return (-1);
	}

	pmod = plcs[link][adapter];
	if (pmod == NULL)
	{
	    if (!allocate_plc (link, adapter))
	    {
		/* if not allocated, try. if can't allocate, get out. */
		return(-1);
	    }
	    pmod = plcs [link] [adapter];
	}

	if (card == MAINS)
	{
	    /* ensure string contained entirely within the address
		space of the allocated region */

	    if ((signal > MAXAINO) || (signal < MINAINO))
	    {
		if (trace [link] & CONFIGURATIONTRACE)
		{
		    logMsg ("sv85_sidriver: Illegal signal:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal, 0);
		}
		return (-1);
	    }
	    if ((signal + len) > MAXAINO)
	    {
		if (trace [link] & CONFIGURATIONTRACE)
		{
		    logMsg ("sv85_sidriver: String extends beyond legal signal range:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal, 0);
		}
		return (-1);
	    }
	    if (resname != 0)
	    { 
		register_resource (link, pmod, resname, pointname);
	    }

	    pmod->was_requested = TRUE;

	    signal--;			/* Signals are defined as 1 relative */
	    reglen = (len + 1) / 2;	/* round up odd length character count string lengths */
	    blkix = -1;

	    for (i = 0; i < reglen ; i++)
	    {
		oldblkix = blkix;
		blkix = ((signal + i) / REGBLOCK);
		regix = ((signal + i) % REGBLOCK);
		if (oldblkix != blkix)
		{
		    pmod->ais [blkix].was_requested = TRUE;
		    if (!resresponse [link])
		    {
			if (trace [link] & CALLREJECTTRACE)
			{
			    logMsg ("sv85_sidriver: Driver Offline:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
			}
			return (-1);
		    }
		    if (pmod->holdflag)
		    {
			if (trace [link] & CALLREJECTTRACE)
		 	{
			    logMsg ("sv85_sidriver: On Hold:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
			}
			return (-2);
		    }

		    /* check for plc simulated mode. */

		    if (modsim [link])
		    {
			*pvalue = 0;
			if (mltrace [link] == 10)
			{
			    logMsg ("sv85_bodriver: SimReturn:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
			}
			return (0);
		    }
		    if (!pmod->is_responding)
		    {
			if (trace [link] & CALLREJECTTRACE)
			{
			    logMsg ("sv85_sidriver: Not Responding:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
			}
			return (-1);
		    }
		    if (!pmod->ais [blkix].data_valid)
		    {
			if (trace [link] & CALLREJECTTRACE)
			{
			    logMsg ("sv85_sidriver: Data Not Valid:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
			}
			return (-1);
		    }
		}
		pvalue [i] = pmod->ais [blkix].aivals [regix];
	    }
	    pvalue [i] = 0; /* null terminate the string */
	    return(0);
	}

	if (card == MREGS)
	{
	    /* ensure string contained entirely within the address
		space of the allocated region */

	    if ((signal > MAXREGNO) || (signal < MINREGNO))
	    {
		if (trace [link] & CONFIGURATIONTRACE)
		{
		    logMsg ("sv85_sidriver: Illegal signal:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal + 1, 0);
		}
		return (-1);
	    }
	    if ((signal + len) > MAXREGNO)
	    {
		if (trace [link] & CONFIGURATIONTRACE)
		{
		    logMsg ("sv85_sidriver: String extends beyond legal signal range:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal + 1, 0);
		}
		return (-1);
	    }
	    if (resname != 0)
	    { 
		register_resource (link, pmod, resname, pointname);
	    }

	    pmod->was_requested = TRUE;

	    signal--;			/* Signals are defined as 1 relative */
	    reglen = (len + 1) / 2;	/* round up odd length character count string lengths */
	    blkix = -1;

	    for (i = 0; i < reglen ; i++)
	    {
		oldblkix = blkix;
		blkix = ((signal + i) / REGBLOCK);
		regix = ((signal + i) % REGBLOCK);
		if (oldblkix != blkix)
		{
		    if (!resresponse [link])
		    {
			if (trace [link] & CALLREJECTTRACE)
			{
			    logMsg ("sv85_sidriver: Driver Offline:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
			}
			return (-1);
		    }
		    if (pmod->holdflag)
		    {
			if (trace [link] & CALLREJECTTRACE)
			{
			    logMsg ("sv85_sidriver: On Hold:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
			}
			return (-2);
		    }

		    /* check for plc simulated mode. */

		    if (modsim [link])
		    {
			*pvalue = 0;
			if (mltrace [link] == 10)
			{
			    logMsg ("sv85_sidriver: SimReturn:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
			}
			return (0);
		    }
		    if (!pmod->is_responding)
		    {
			if (trace [link] & CALLREJECTTRACE)
			{
			    logMsg("sv85_sidriver: Not Responding:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
			}
			return (-1);
		    }
		    if (!pmod->regs [blkix].data_valid)
		    {
			if (trace [link] & CALLREJECTTRACE)
			{
			    logMsg ("sv85_sidriver: Data Not Valid:[%s] L=%d, A=%d, C=%d, S=%d\n",
				(int) pointname, link, adapter, card, signal + 1, 0);
			}
			return (-1);
		    }
		    return (-1);
		}
		pvalue [i] = pmod->regs [blkix].ioregs [regix];
	    }
	    pvalue [i] = 0; /* null terminate the string */
	    return (0);
	}
	return(-1);
}


/**********
*
*      sv85_sodriver
*
*      This routines main purpose is to set requested string outputs to the
*      specified module via the output buffer.
*
**********/

/* this routine not called from within this driver */

int sv85_sodriver (char *pointname, unsigned short link, unsigned short adapter,
		unsigned short card, unsigned short signal,
		unsigned char *value, unsigned short len,
		unsigned char *resname)
{
	MODPLC		*pmod;
	unsigned short	blkix;
	OUTQENT		*oqp;
        unsigned short	reglen;
        int		i;

	if (link > MODLINES)
	{
	    if (trace [link] & CONFIGURATIONTRACE)
	    {
		logMsg ("sv85_sodriver: Illegal line:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal, 0);
	    }
	    return (-1);
	}
	if (adapter > NUMMODPLCS)
	{
	    if (trace [link] & CONFIGURATIONTRACE)
	    {
		logMsg ("sv85_sodriver: Illegal adapter:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal, 0);
	    }
	    return (-1);
	}
	if (card != MREGS)
	{
	    if (trace [link] & CONFIGURATIONTRACE)
	    {
		logMsg ("sv85_sodriver: Illegal card:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal, 0);
	    }
	    return (-1);
	}
	pmod = plcs [link] [adapter];
	if (pmod == NULL)
	{
	    if (!allocate_plc (link, adapter))
	    {
		/* if not allocated, try. if can't allocate, get out. */
		return (-1);
	    }
	    pmod = plcs [link] [adapter];
	}

        /* ensure string contained entirely within the address
		space of the allocated region */

	if ((signal > MAXREGNO) || (signal < MINREGNO))
	{
	    if (trace [link] & CONFIGURATIONTRACE)
	    {
		logMsg ("sv85_sodriver: Illegal signal:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal, 0);
	    }
	    return (-1);
	}
	if ((signal + len) > MAXREGNO)
	{
	    if (trace [link] & CONFIGURATIONTRACE)
	    {
		logMsg ("sv85_sodriver: String extends beyond legal signal range:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal, 0);
	    }
	    return (-1);
	}
	if (resname != 0)
	{ 
	    register_resource (link, pmod, resname, pointname);
	}
	pmod->was_requested = TRUE;

	signal--;			/* Signals are defined as 1 relative */
	reglen = (len + 1) / 2;		/* round up odd length character count string lengths */
	blkix = (signal / REGBLOCK);

	pmod->regs [blkix].was_requested = TRUE; /* notice we only set one here */

	if (!resresponse [link])
	{
	    if (trace [link] & CALLREJECTTRACE)
	    {
		logMsg ("sv85_sodriver: Driver Offline:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal + 1, 0);
	    }
	    return (-1);
	}

	/* check for plc simulated mode. */

	if (modsim [link])
	{
	    if (mltrace [link] == 5)
	    {
		logMsg("sv85_sodriver: SimReturn:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal + 1, 0);
	    }
	    return (0);
	}
	if ((!pmod->is_responding) || (pmod->cmd_force_offline))
	{
	    if (trace [link] & CALLREJECTTRACE)
	    {
		logMsg("sv85_sodriver: Not Responding or Forced offline:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal + 1, 0);
	    }
	    return (-1);
	}

	oqp = ogetemptyp (link);
	if (oqp == NULLP)
	{
	    if (trace [link] & RUNTIMEERRORTRACE)
	    {
		logMsg ("sv85_sodriver: Queue Overflow:[%s] L=%d, A=%d, C=%d, S=%d\n",
			(int) pointname, link, adapter, card, signal + 1, 0);
	    }
	    pmod->badoutcnt++;
	    return (-1);		/* output queue is full */
	}
	oqp->otype = SOUTTYPE;
	oqp->out.sonode.link = link;
	oqp->out.sonode.adapter = adapter;
	oqp->out.sonode.card = card;
	oqp->out.sonode.sosignal = signal;
	oqp->out.sonode.solen = len;

	/* Move in the output string (strcpy not used because */
        /*  string content could contain terminators)         */

	for (i = 0 ;(i < STRINGMSGLEN) ; i++)
	{
	    oqp->out.sonode.sovalue [i] = value [i];
	}

	omarkfull (link);		/* Mark this entry as ready for output */

	if (mltrace [link] == 5)
	{
	    logMsg ("sv85_sodriver: Queued:[%s] L=%d, A=%d, C=%d, S=%d, V=%s\n",
		(int) pointname, link, adapter, card, signal + 1, (int) value);
	}
	return (0);
}

/**********
*
* register_resource
*
*	This routine is used to initialize the resource pointers
*  If this is the first time the resource for this line is being
*  referenced the pointer to the resource table entry will be set
*  and the resource name will be saved. If the resource name has
*  already been set, this routine will verify that all other points
*  for this driver are using the same resource name.
*
**********/                                                                                    	

static void  register_resource (unsigned short link, MODPLC *pmod,
                                unsigned char *resname, char *pointname)
{
	if (resnames [link] [0] == 0) /* This is the first reference */
	{
	    if (resGetByName (resname, &prestable [link]) != RES_OK)
	    {
		logMsg ("Resource not found in table. (Point Name = %s, Resource Name = %s)\n",
                                (int) pointname,
				(int) resname, 0, 0, 0, 0 );
	    }
	    else
	    {
		strcpy (resnames [link], resname);
	    }
	}
	else
	{
	    if (strcmp (resnames [link], resname) != 0)
	    {
		logMsg ("Resource name [%s] conflict. (Point Name = %s, Resource Name = %s)\n",
				(int) resnames[link],
                                (int) pointname,
				(int) resname, 0, 0, 0 );
	    }
	}
}

/**********
*
*      Initialize the Queue                                        		
*
**********/                                                                        	

/* not called from within this driver */

static	void outqueinit (unsigned short line)
{
	oq [line]->ooutix = oq [line]->oinix = 0;
}


/**********
*                                                                     		
*  ogetempty - get the next available slot on the output list         		
*
**********/

static	OUTQENT *ogetemptyp (unsigned short line)
{
	if (((oq [line]->oinix + 1) % OUTQSIZE) != oq [line]->ooutix)
	{
	    return (&oq [line]->outque[oq [line]->oinix]);
	}
        if (mltrace [line] == 11)
	{
	    logMsg ( "Output Queue Full\n", 0, 0, 0, 0, 0, 0);
	}
	return (NULLP);
}


/**********
*                                                                    		
* omarkfull                                                          		
*      mark the next block as full                                     	
*
**********/

static	void omarkfull (unsigned short line)
{
	oq [line]->oinix = ((oq [line]->oinix + 1) % OUTQSIZE);
}

/**********
*                                                                      	
*  ogetnextp - get the next full slot on the output list               	
*
**********/                                                                      	

/* only called from checkoutputs */

static	OUTQENT *ogetnextp (unsigned short line)
{
	if (oq [line]->ooutix != oq [line]->oinix)
	{
	    return (&oq [line]->outque[oq [line]->ooutix]);
	}
	else
	{
	    return (NULLP);
	}
}

/**********
*                                                       	               	
* oreleasefull - release the current full output block    	              	
*
**********/                                                                            	

/* only called from checkoutputs */

static	void oreleasefull (unsigned short line)
{
	if (oq [line]->ooutix == oq [line]->oinix)
	{
	    return;
	}
	oq [line]->ooutix = ((oq [line]->ooutix + 1) % OUTQSIZE);
}

/**********                                                                             	
*                                                                              
*	allocate_plc                                                            
*                                                                              
*	The purpose of this routine is to allocate the memory for a single      
*	plc data structure. If not successful, a console error message is  	
*	printed and an error flag is returned to the caller.                    
*	If successful, the newly allocated data structures are initialized.     
*                                                                             	
**********/

static int allocate_plc (unsigned short line, unsigned short plc )
{
	plcs [line] [plc] = (MODPLC *) calloc (1, (sizeof (MODPLC)));
	if (!plcs [line] [plc])
	{
	    logMsg ("Unable to allocate memory", 0, 0, 0, 0, 0, 0);
            printf (" for MODPLC data structure on line %d, plc %d\n", line, plc);
	    return (FALSE);
	}
	else
	{
	    logMsg ("Allocated memory for MODPLC data structure on line %d, plc %d\n",
			line, plc, 0, 0, 0, 0);
	    init_plc_tab (line, plc);	     /*	initialize the data structures. */
	}
	return (TRUE);
}

/**********
*                                                                             
*      check_line_activity                                                     
*                                                                             	
**********/


static	int check_line_activity (unsigned short line, unsigned short predelay )

    {
	RXLINE          *rxp;
	long		thistime;
	short		status;

	if (predelay != 0)
	{
	    taskDelay (predelay);
	}
	thistime = tickGet ();
	if ((thistime - activity_time [line]) < ACTIVITYTIME)
	{
	    return (lastactivity [line] || (line_activity [line]));
	}
	lastactivity [line] = status = line_activity [line];
	line_activity [line] = FALSE;
	activity_time [line] = thistime;
	rxp = rxlinetab [line];
	rxp->rxstate = RXIGNORE;

	return (status);
}

/**********
*
*      msgxfer
*
*      The purpose of this routine is to transmit a buffer to the
*      analog devices I/O and to await a response. The transmit and receive
*      buffers are in their respective line tables. This routine will retry
*      the message up to a maximum number of times or until a response
*      is received witha proper checksum.
*              returns         XFERGOOD if response received
*                              XFERBAD  if no good repsonse is received
*
**********/
static	unsigned short  msgxfer (unsigned short fline, unsigned short resplen,
				MODPLC *pmod, unsigned short rxlength)
{

	unsigned short  retry, line, ix;
	unsigned short	timeout;
	RXLINE          *rxp;
	TXLINE          *txp;
	unsigned char	buf [512], opcode;
	unsigned char	svcreq [4];
	unsigned char	*BASE;

	line = (fline & (~ONERETRY));          /* strip off the no retry flag*/
	BASE = bases [line];

	looptotmsgs[line]++;	    /* init the total message count	*/
	loopmsgs[line]++;	    /* init the total message count	*/

	rxp = rxlinetab [line];
	txp = txlinetab [line];

        /* set the path number */

	opcode = txp->txbuf [1];		      /* extract the opcode */
	if ((opcode == 9) || (opcode == 10) ||
	     (opcode == 13) || (opcode == 14) ||
	     (opcode == 18) || (opcode == 126))
	{
	   buf [0] = 0x81;
	}
	else
	{
	   buf [0] = 1;
	}
	/* set the broadcast address */

	buf [1] = 0xff;

	/* set the destination node address */

	buf [2] = txp->txbuf [0];	/* get the address from the modbus msg */

	/* set the source nodes address */

        buf [3] = mbplusaddr [line];	/* read back from interface */

	/* Set the MAC Command code */

	buf [4] = 0x11;

	/* Set The byte count */

	buf [5] = (txp->txmsglen - 3 + 8) & 0xff; /* take out dest addr and crc add in mb+ oh */

	buf [6] = (txp->txmsglen - 3 + 8) / 256;

	/* copy path number */

	buf [7] = buf [0];

	/* reserved */

	buf [8] = 0;

	/* skip sequence number */
	/* Routing Path */

	buf [0x0a] = buf [2];
	buf [0x0b] = buf [0x0c] = buf [0x0d] = buf [0x0e] = 0;

	/* copy in the message without the destination address or the crc */

	for (ix = 0;(ix < (txp->txmsglen - 3)); ix++)
	{
	    buf [ix + 0x0f] = txp->txbuf [ix + 1];
	} 

	for (retry = ((fline == line) ? 0 : MAXRETRYS - 1); (retry < MAXRETRYS); retry++ )
	{
	    if (intermsgdelay [line] != 0)
	    {
		taskDelay (intermsgdelay [line]);
	    }

	    /* Send out the message */

	    if (sv85putmstcmd (line, buf))
	    {

		/* wait in a loop for time out or response on my path */

		timeout = FALSE;
                ix = 0;

		while (!timeout)
		{
		    /* Enable Unsollicited interrupts */

		    BASE [CONTROLREG] = SV85SRRESET + SV85INTENB;

		    if (((semTake (semids [line], 2)) == ERROR) ||
			(resp [line] != 0x80) )
		    {
		         SVCDelay++;
			 if (modtrace [line])
			{ 
			     logMsg ("SVCDelay -  peermbx = %x,\n", resp [line], 0, 0, 0, 0, 0);
			}
	    	    }

		    BASE [CONTROLREG] = SV85SRLOCK + SV85SRRESET + SV85INTENB;

		    if (!sv85getsvcreq (line, svcreq))
		    {
			if (modtrace[line])
			{
			    logMsg ("Get Service Request in MSGXFER\n", 0, 0, 0, 0, 0, 0);
			}
			return (XFERBAD);
		    }

		    ix++;
		    if (ix == RXTIMEOUT)
		    {
			timeout = TRUE;
			break;
		    }

                    if ((buf [0] == 1) && ((svcreq [0] & 1) != 0))
		    {
			break;
		    }
		    if ((buf [0] == 0x81) && ((svcreq [2] & 1) != 0))
		    {
			break;
		    }
		    SVCDelay++;
		}

                if (timeout)
		{
		    if (modtrace [line])
		    {
			 logMsg ("Receive Timeout... Address: %d\n", buf [2], 0, 0, 0, 0, 0);
		    }
		    if (!sv85abort (line, buf [0]))
		    {
			if (modtrace [line])
			{
			    logMsg ("Abort Failed in MSGXFER\n", 0, 0, 0, 0, 0, 0);
			}
			return (XFERBAD);
		    }
		}
		else
		{
                    if (!sv85getmstrsp (line, buf [0], &rxp->rxbuf [1]))
		    {
			if (modtrace [line])
			{
			    logMsg ("Get Response Failed in MSGXFER\n", 0, 0, 0, 0, 0, 0);
			}
			return (XFERBAD);
		    }
		    else
		    {
			rxp->rxbuf [0] = buf [2]; /* put the node address in buffer */
			pmod->last_good_msg_time = tickGet (); 
			return (XFERGOOD);
		    }
		}
            }
	    else
	    {
	        if (modtrace [line])
		{
		    logMsg ("Put Command Failed in MSGXFER\n", 0, 0, 0, 0, 0, 0);
		}
		return (XFERBAD);
            }

	    if (retry != MAXRETRYS)
	    {
		pmod->no_retries++;
	    }

	}
	if (modtrace [line])
	{
	    logMsg ("Xfer Bad\n", 0, 0, 0, 0, 0, 0);
	}
	return (XFERBAD);
}

/**********
*
*      msgxmit
*
*      The purpose of this routine is to transmit a buffer to the
*      I/O. The buffers are in their respective line tables. 
*              returns         XFERGOOD if response received
*                              XFERBAD  if no good repsonse is received
*
**********/
static	unsigned short  msgxmit (unsigned short fline, unsigned short resplen,
				MODPLC *pmod, unsigned short rxlength, PATHINFO **path)
{

	unsigned short  line, ix, mask, i;
	RXLINE          *rxp;
	TXLINE          *txp;
	unsigned char	buf [512];
	unsigned char	*BASE;

	line = (fline & (~ONERETRY));          /* strip off the no retry flag*/
	BASE = bases [line];


	looptotmsgs[line]++;	    /* init the total message count	*/
	loopmsgs[line]++;	    /* init the total message count	*/

	rxp = rxlinetab [line];
	txp = txlinetab [line];

	/* See if there is a put path available */

	if ((putpath [line] & 0xFE) == 0xFE)
	{
	    if (!clearputpath (line))
	    {
	        if (modtrace [line])
		{
		    logMsg ("Put paths full in msgxmit\n", 0, 0, 0, 0, 0, 0);
		}
		return (XFERBAD);
	    }
	}
	for (mask = 2, i = 2; (i <= 8); i++)
	{
	    if (((putpath [line] & mask) == 0) && (i != 9))
	    {
		break;
	    }
	    mask = mask << 1;
	}

	*path = &(lpath [line]->paths [i]);

	if (i > 8)		/* adjust higher path numbers to use program paths */
	{
	    i = 0x80 + (i - 8);
	}
	buf [0] = i;

	/* set the broadcast address */

	buf [1] = 0xff;

	/* set the destination node address */

	buf [2] = txp->txbuf [0];	/* get the address from the modbus msg */

	/* set the source nodes address */

        buf [3] = mbplusaddr [line];	/* read back from interface */

	/* Set the MAC Command code */

	buf [4] = 0x11;

	/* Set The byte count */

	buf [5] = (txp->txmsglen - 3 + 8) & 0xff; /* take out dest addr and crc add in mb+ oh */

	buf [6] = (txp->txmsglen - 3 + 8) / 256;

	/* copy path number */

	buf [7] = buf [0];

	/* reserved */

	buf [8] = 0;

	/* skip sequence number */
	/* Routing Path */

	buf [0x0a] = buf [2];
	buf [0x0b] = buf [0x0c] = buf [0x0d] = buf [0x0e] = 0;

	/* copy in the message without the destination address or the crc */

	for (ix = 0;(ix < (txp->txmsglen - 3)); ix++)
	{
	    buf [ix + 0x0f] = txp->txbuf [ix + 1];
	} 

	/* Send out the message */

	if (sv85putmstcmd (line, buf))
	{
	    putpath[line] |= mask;
	}
	else
	{

 	    if (modtrace [line])
	    {
		logMsg ("put failed on path: %d in MSGXMIT\n", i, 0, 0, 0, 0, 0);
	    }
	    if (!sv85abort (line, buf [0]))
	    {
	        if (modtrace [line])
		{
		    logMsg ("Abort Failed in MSGXMIT\n", 0, 0, 0, 0, 0, 0);
		}
	    }
	    return (XFERBAD);
	}
        return (XFERGOOD);
}

/**********
*
* Check Responses (line)
*
**********/
static  int checkresponses (unsigned short line)

    {
	unsigned long	starttime, thistime;
	unsigned short	timeout, mask, i, p, svcmask;
	unsigned char	svcreq [4];
	PATHINFO	*path;
	unsigned char	*BASE;

	BASE = bases [line];

	starttime = tickGet ();
	timeout = FALSE;

	/* see if there are any puts which have timed out */

	thistime = tickGet();
	mask = 2;
	for (i = 2; (i < 9); i++)
	{
	    if ((putpath [line] & mask) != 0)
	    {
		path = &(lpath [line]->paths [i]);
		if (i > 8)
		{
		    p = 0x80 + (i - 8);
		}
		else
		{
		    p = i;
		}
		if ((thistime - path->timeout) > path_timeout [line])
		{
		    if (!sv85abort (line, p))
		    {
			if (modtrace [line])
			{
			    logMsg ("Abort Failed in check responses\n", 0, 0, 0, 0, 0, 0);
			}
		    }
		    else
		    {
			putpath [line] = putpath [line] & (~mask);
			if (modtrace [line])
			{
			     logMsg ("Abort path = %x in check responses because path timeout \n",
					p, 0, 0, 0, 0, 0);
			}
		    }

		    if (  (thistime - path->pmod->last_good_msg_time)
		        > (sv85nodetimeout*60)) {
			if (path->pmod->is_responding)
			{
			    path->pmod->no_offlines++;
			}
			/* >>> Path has timed out so mark unit as off line */
			path->pmod->is_responding = FALSE;
		        if (modtrace [line])
		        {
			    logMsg ("Unit going offline - mod = %d \n",
					path->mod, 0, 0, 0, 0, 0);
		        }
		    }

		    if (modtrace [line])
		    {
			logMsg ("Path timed out in check responses mod = %d \n",
					path->mod, 0, 0, 0, 0, 0);
		    }
		    path->pmod->no_rxnoresp++;
		}
	    }
	    mask = mask << 1;
	}

	if (!sv85getsvcreq (line, svcreq))
	{
	    if (modtrace [line])
	    {
		logMsg ("Get Service Request in check responses failed\n", 0, 0, 0, 0, 0, 0);
	    }
	    return (FALSE);
	}

	svcmask = svcreq [2];
	svcmask = ((svcmask << 8) | svcreq [0]);

	if (svcmask != 0)
	{
	    if (!process_paths (line, svcmask))
	    {
		return (FALSE);
	    }
	}
	return (TRUE);
}

/**********
*
* Clear the put paths because they are all used
*
**********/
static  int clearputpath (unsigned short line)

    {
	unsigned long	starttime, thistime;
	unsigned short	timeout, mask, i, p, svcmask;
	int		status;
	unsigned char	svcreq [4];
	PATHINFO	*path;
	unsigned char	*BASE;

	BASE = bases [line];

	starttime = tickGet ();
	timeout = FALSE;

	/* see if there are any puts which have timed out */

	thistime = tickGet ();
	mask = 2;
	for (i = 2; (i < 9); i++)
	{
	    if ((putpath [line] & mask) != 0)
	    {
		path = &(lpath [line]->paths [i]);
		if (i > 8)
		{
		    p = 0x80 + (i - 8);
		}
		else
		{
		    p = i;
		}
		if ((thistime - path->timeout) > path_timeout [line])
		{
		    if (!sv85abort (line, p))
		    {
			if (modtrace [line])
			{
			    logMsg ("Abort Failed in clear paths\n", 0, 0, 0, 0, 0, 0);
			}
			return (FALSE);
		    }
		    else
		    {
			if (modtrace [line])
			{
			     logMsg ("Abort path = %x in clear put path because path timeout \n",
					p, 0, 0, 0, 0, 0);
			}
			putpath [line] = putpath [line] & (~mask);
		    }
		    if (  (thistime - path->pmod->last_good_msg_time)
		        > (sv85nodetimeout*60)) {
			if (path->pmod->is_responding)
			{
			    path->pmod->no_offlines++;
			}
		        if (modtrace [line])
		        {
			    logMsg ("Unit going offline - mod = %d \n", path->mod, 0, 0, 0, 0, 0);
		        }
			/* >>> Path has timed out so mark unit as off line */
			path->pmod->is_responding = FALSE;
		    }
		    if (modtrace [line])
		    {
			logMsg ("Path timed out in clear_put_path  mod = %d \n", path->mod, 0, 0, 0, 0, 0);
		    }
		    path->pmod->no_rxnoresp++;
		}
	    }

	    mask = mask << 1;
	}

	while ((!timeout) && (putpath [line] == 0xFE))
	{

	    /* Enable Unsollicited interrupts */

	    BASE [CONTROLREG] = SV85SRRESET + SV85INTENB;
	    if (((status = semTake (semids [line], 10)) == ERROR) ||
		    (resp [line] != 0x80) )
	    {
		if ((status == ERROR) && (modtrace [line]))
		{
		    logMsg ("Timeout waiting for interrupt... semTake\n", 0, 0, 0, 0, 0, 0);
		}
		SVCDelay++;
		if (modtrace [line])
		{
		    logMsg ("SVCDelay -  peermbx = %x,\n", resp [line], 0, 0, 0, 0, 0);
		}			
	    }

	    BASE [CONTROLREG] = SV85SRLOCK + SV85SRRESET + SV85INTENB;

	    if (!sv85getsvcreq (line, svcreq))
	    {
		if (modtrace [line])
		{
		    logMsg ("Get Service Request in clearputpath failed\n", 0, 0, 0, 0, 0, 0);
		}
		return (FALSE);
	    }

	    svcmask = svcreq [2];
	    svcmask = ((svcmask << 8) | svcreq [0]) & 0xFE;

	    if (svcmask == 0)
	    {
		thistime = tickGet ();
		timeout = ((thistime - starttime) > 60);

	    	SVCDelay++;
	    }
	    else
	    {
		if (!process_paths (line, svcmask))
		{
		    return (FALSE);
		}
	    }
	}
	if (putpath [line] == 0xFE)
	{
	    return (FALSE);
	}
	return (TRUE);
}

/**********
*			
* process_paths (line, mask)
*
*	process the response to outstanding write commands to the SV85
*
**********/
static int process_paths (unsigned short line, unsigned short mask)

{

	unsigned short  m, i, tmask, blkix, j, p, bitaddr, regaddr, buildreg;
	PATHINFO	*path;
	RXLINE          *rxp;
	TXLINE          *txp;
	unsigned long	cval,thistime;

	thistime = tickGet();
	rxp = rxlinetab [line];
	txp = txlinetab [line];

	tmask = 1;

	for (m = 1; (m < 9); m++)
	{
	    if (((tmask & mask) != 0) && (m != 1) && (m != 9))
	    {

		/* have a response for path*/

		path = &(lpath [line]->paths [m]);

		if (m > 8)
		{
		    p = 0x80 + (m - 8);
		}
		else
		{
		    p = m;
		}
                if (!sv85getmstrsp (line, p, &rxp->rxbuf [1]))
		{
		    putpath [line] = (putpath [line] & (~tmask));
		    if (  (thistime - path->pmod->last_good_msg_time)
		        > (sv85nodetimeout*60)) {
			if (path->pmod->is_responding)
			{
			    path->pmod->no_offlines++;
			}
		        if (modtrace [line])
		        {
			    logMsg ("Unit going offline - mod = %d \n", path->mod, 0, 0, 0, 0, 0);
		        }
			/* >>> Get Master Response failed so mark unit as off line */
			path->pmod->is_responding = FALSE;
		    }

		    path->function = 0xffff; /* skip the switch */
		    if (modtrace [line])
		    {
			logMsg ("Get Response Failed in process put path #: %x  mod = %d\n",
				 p, path->mod, 0, 0, 0, 0);
		    }
		}
		else
		{
	            /* Verify the response - the message is repeated back*/
		    path->pmod->last_good_msg_time = tickGet (); 

	            for (j = 1; (j < path->resplen); j++)
		    {
			if (path->respbuf [j] != rxp->rxbuf [j])
			{
		            path->pmod->no_dataerr++;
			    putpath [line] = (putpath [line] & (~tmask));
			    if (path->function == SCOIL || path->function == SMCOIL)
			    {
				path->pmod->badoutcnt++;
			    }
			    path->function = 0xffff;		/* skip the switch */
			    /* >>> There is a data format error so mark unit as off line */
			    path->pmod->is_responding = FALSE;
		            if (modtrace [line])
		            {
			        logMsg ("Unit going offline - mod = %d \n", path->mod, 0, 0, 0, 0, 0);
		            }
			    if (modtrace [line])
			    {
				logMsg (" Data Error from local in process put path# %d, mod = %d\n"
						, p, path->mod, 0, 0, 0, 0);
			    }
			}
	            }
		}

		switch (path->function)
		{

		case SCOIL:
	    	    /* update the internal copy so partner will get it on update */
	    	    if (path->lval != 0)
		    {
	    		path->pmod->coils [(path->lsignal / BITBLOCK)].last_read |= masks [path->lsignal % BITBLOCK];
		    }
	    	    else
		    {
	    		path->pmod->coils [(path->lsignal / BITBLOCK)].last_read &= ~(masks [path->lsignal % BITBLOCK]);
		    }
		    break;

		case SMCOIL:

	    	    /* update the internal copy so partner will get it on update */
		
		    cval = path->pmod->coils [(path->lsignal / BITBLOCK)].last_read;
		    cval &= (~(path->mask));
		    cval |= path->val;
		    path->pmod->coils [(path->lsignal / BITBLOCK)].last_read = cval;
		    break;

		case SREG:

		    /* update the internal copy so partner will get it on update */

		    blkix = (path->regaddr / REGBLOCK);
		    path->pmod->regs [blkix].ioregs [path->regaddr % REGBLOCK] = path->sval;
		    break;

		case RCOIL:

		    if (path->reqlen != rxp->rxbuf [2])
		    {
			path->pmod->no_dataerr++;
			putpath [line] = (putpath [line] & (~tmask));
			if (modtrace [line])
			{
			    logMsg ("RCOIL Data Error from local in process put path# %d\n",
					p, 0, 0, 0, 0, 0);
			}
		    }
		    else
		    {

			/* Store the returned data		*/
			/* data is aligned by 8 pseudo register	*/
			/* and a multiple of 8 bits long	*/

			bitaddr = path->lsignal;
			for (i = j = 0; (i < rxp->rxbuf [2]); i += (BITBLOCK / 8), j++)
			{
	    		    path->pmod->coils [(bitaddr / BITBLOCK) + j].last_read = rxp->rxbuf [3 + i];
	    		    path->pmod->coils [(bitaddr / BITBLOCK) + j].last_read += (rxp->rxbuf [4 + i] << 8);
	    		    path->pmod->coils [(bitaddr / BITBLOCK) + j].last_read += (rxp->rxbuf [5 + i] << 16);
	    		    path->pmod->coils [(bitaddr / BITBLOCK) + j].last_read += (rxp->rxbuf [6 + i] << 24);
	   		    path->pmod->coils [(bitaddr / BITBLOCK) + j].data_valid = TRUE;
			}
		    }
		    break;

		case RDI:

		    if (path->reqlen != rxp->rxbuf [2])
		    {
			path->pmod->no_dataerr++;
			putpath [line] = (putpath [line] & (~tmask));
			if (modtrace [line])
			{
			    logMsg ("RDI Data Error from local in process put path# %d\n",
					p, 0, 0, 0, 0, 0);
			}
		    }
		    else
		    {
			/* Store the returned data			*/
			/*  data is aligned by 8 pseudo register	*/
			/*  and a multiple of 8 bits long		*/

			bitaddr = path->lsignal;
			for (i = j = 0; (i < rxp->rxbuf [2]); i += (BITBLOCK / 8), j++)
			{
	    		    path->pmod->dis [(bitaddr / BITBLOCK) + j].last_read = rxp->rxbuf [3 + i];
	    		    path->pmod->dis [(bitaddr / BITBLOCK) + j].last_read += (rxp->rxbuf [4 + i] << 8);
	    		    path->pmod->dis [(bitaddr / BITBLOCK) + j].last_read += (rxp->rxbuf [5 + i] << 16);
	    		    path->pmod->dis [(bitaddr / BITBLOCK) + j].last_read += (rxp->rxbuf [6 + i] << 24);
	   		    path->pmod->dis [(bitaddr / BITBLOCK) + j].data_valid = TRUE;
			}
		    }
		    break;

		case RREG:

		    if (path->reqlen != rxp->rxbuf [2])
		    {
		    	path->pmod->no_dataerr++;
			putpath [line] = (putpath [line] & (~tmask));
			if (modtrace [line])
			{
			    logMsg ("RREG Data Error from local in process put path# %d\n",
					p, 0, 0, 0, 0, 0);
			}
		    }
		    else
		    {

			/* Store the returned data                      */
			/*  data is aligned by 4 register sets          */
			/*  and a multiple of 4 registers long          */

			regaddr = path->regaddr;

			for (i = 0; (i < rxp->rxbuf [2]); i += (REGBLOCK * 2))
			{
	    		    for (j = 0; (j < REGBLOCK); j++)
			    {
				buildreg = rxp->rxbuf [3 + i + (j * 2)];
				buildreg = (buildreg << 8) + rxp->rxbuf [3 + i + (j * 2) + 1];
				path->pmod->regs [(regaddr / REGBLOCK) + (i / (REGBLOCK * 2))].ioregs [j] = buildreg;
			    }
			    path->pmod->regs [(regaddr / REGBLOCK) + (i / (REGBLOCK * 2))].data_valid = TRUE;
			}
		    }
		    break;

		case RAI:

		    if (path->reqlen != rxp->rxbuf [2])
		    {
		    	path->pmod->no_dataerr++;
			putpath [line] = (putpath [line] & (~tmask));
			if (modtrace [line])
			{
			    logMsg ("RAI Data Error from local in process put path# %d\n",
					p, 0, 0, 0, 0, 0);
			}
		    }
		    else
		    {

			/* Store the returned data			*/
			/*  data is aligned by REGBLOCK register sets	*/
			/*  and a multiple of REGBLOCK registers long	*/

			regaddr = path->regaddr;

			for (i = 0; (i < rxp->rxbuf [2]); i += (REGBLOCK * 2))
			{
	    		    for (j = 0; (j < REGBLOCK); j++)
			    {
				buildreg = rxp->rxbuf [3 + i + (j * 2)];
				buildreg = (buildreg << 8) + rxp->rxbuf [3 + i + (j * 2) + 1];
				path->pmod->ais [(regaddr / REGBLOCK) + (i / (REGBLOCK * 2))].aivals [j] = buildreg;
			    }
			    path->pmod->ais [(regaddr / REGBLOCK) + (i / (REGBLOCK * 2))].data_valid = TRUE;
			}
		    }
		    break;

		case WMCOIL:
		case WMREG:
		default:
		    /* If response was ok we are done */
		    break;

		}

		/* clear the flag for the processed message */

	        putpath [line] = (putpath [line] & (~tmask));
	    }
	    if (((tmask & mask) != 0) && ((m == 1) || (m == 9)))
	    {
		if (!sv85abort (line, m))
		{
		    if (modtrace [line])
		    {
			logMsg ("Abort Failed in process put path# %d\n", m, 0, 0, 0, 0, 0);
		    }
		    return (FALSE);
		}
		else
		{
		    putpath [line] = putpath [line] & (~tmask);
		    if (modtrace [line])
		    {
			logMsg ("Aborted message in process put path# %d\n", m, 0, 0, 0, 0, 0);
		    }
		}
	    }
	    tmask = tmask << 1;
	}
	if (modtrace [line] == 2)
	{
	    logMsg ("PP mask = %x,lsvc = %x, pp[line] = %x\n",
			mask, lastsvcreq [0], putpath [line], 0, 0, 0);
	}
	return (TRUE);
}

/**********
*
*	Put master command to output path
*
**********/
int sv85putmstcmd (unsigned short line, unsigned char *buf)
{
	unsigned char	stat, path;
	unsigned short	buflen, i;
	unsigned char	*BASE;

	BASE = bases [line];

	if (modtrace [line] == 6)
	{
	    logMsg ("Put path = %d ,pp = %x\n",buf [0], (putpath [line]), 0, 0, 0, 0);
	}
	buflen = (buf [6] * 256) + buf [5] + 7;
	for (i = 0; i < buflen; i++)
	{
	    BASE [(RAMBASE + i)] = buf [i];
	}
	path = BASE [RAMBASE];		/* save the path number for echo test */
	stat = sv85msgxfer (line, 0x01, NULLP, 60);
	if (stat == TRUE)
	{
	    if (BASE [RAMBASE] == path)	/* requested internal path # echoed back */
	    {
		return (TRUE);
	    }
	}
	sv85reset (line);
	return (FALSE);
}

/**********
*
*	Put host command to output path
*
**********/
int sv85puthostcmd (unsigned short line, unsigned char *buf,
		unsigned short len, unsigned char * respbuf)
{
	unsigned char	stat, path;
	unsigned short  i, mlen;
	unsigned char	*BASE;

	BASE = bases [line];

	if (modtrace [line] == 6)
	{
	    logMsg ("Put path = %d ,pp = %x\n", buf [0], (putpath [line]), 0, 0, 0, 0);
	}
	for (i = 0; i < len; i++)
	{
	    BASE [(RAMBASE + i)] = buf [i];
	}
	path = BASE [RAMBASE];		/* save the path number for echo test */
	stat = sv85msgxfer (line, 0x0b, NULLP, 60);
	if (stat == TRUE)
	{
	    mlen = (unsigned short) BASE [RAMBASE + 2];
	    mlen <<= 8;
	    mlen += (unsigned short) BASE [RAMBASE + 1];
	    for (i = 0; i < mlen ; i++)
	    {
		respbuf [i] = BASE [(RAMBASE + 3 + i)];
	    }

	    return (TRUE);
	}
	sv85reset (line);
	return (FALSE);
}

/**********
*
*	Get master response from output path
*
**********/
int sv85getmstrsp (unsigned short line, unsigned char path, unsigned char *buf)
{
	unsigned char	stat, maccode, errcode;
	unsigned short	i, len;
	unsigned char	*BASE;

	BASE = bases [line];

	if (modtrace [line] == 6)
	{
	    logMsg ("Get path = %d ,pp = %x\n", path, (putpath [line]), 0, 0, 0, 0);
	}

	BASE [RAMBASE] = path;		/* save the path number for echo test */
	stat = sv85msgxfer (line, 0x04, NULLP, 60);
	if (stat == TRUE)
	{
	    len = (unsigned short) BASE [RAMBASE + 6];
	    len <<= 8;
	    len += (unsigned short) BASE [RAMBASE + 5];
	    len -= 8; 				/* add frame overhead */
	    for (i = 0; i < len ; i++)
	    {
		buf [i] = BASE [(RAMBASE + MODBUSBASE + i)];
	    }
	    maccode = BASE [(RAMBASE + 4)];
	    errcode = BASE [(RAMBASE + 16)];

	    if (((maccode & 0x1f) != 0x16))
	    {
		if (modtrace [line])
		{
		    logMsg ("Get MST Resp path = %d ,pp = %x, maccode = %x, err = %x, len = %d\n",
				path, (putpath [line]), maccode, errcode, len, 0);
		}
		return (FALSE);
	    }
	    return (TRUE);
	}
	sv85reset (line);
	return (FALSE);
}

/**********
* 
*     sv85_int_proc
*
*      The purpose of this routine is to process a receive character interrupt
*      for a given line.
*
**********/

unsigned long	Ireqs = 0, Irsps = 0, Iwait = 0, Iduart1 = 0, Iduart2 = 0;
unsigned long	Iatx1 = 0, Iarx1 = 0, Iabrk1 = 0, Itmr1 = 0;
unsigned long	Ibtx1 = 0, Ibrx1 = 0, Ibbrk1 = 0, Iprtch1 = 0;

void sv85_int_proc (unsigned short line)
{

	unsigned char	*BASE;
	unsigned char	intstat;

	++Irsps;
	BASE = bases [line];
	intrs [line] = BASE [STATUSREG];

	if ((intrs [line] & 0x04) == 0)
	{
	    Iduart1++;
	    intstat = BASE [D1INTSTA];
	    if (intstat & ATXRDY)
	    {
		Iatx1++;
	    }
	    if (intstat & ARXRDY)
	    {
		Iarx1++;
	    }
	    if (intstat & ABRK)
	    {
		Iabrk1++;
	    }
	    if (intstat & TMRRDY)
	    {
		Itmr1++;
	    }
	    if (intstat & BTXRDY)
	    {
		Ibtx1++;
	    }
	    if (intstat & BRXRDY)
	    {
		Ibrx1++;
	    }
	    if (intstat & BBRK)
	    {
		Ibbrk1++;
	    }
	    if (intstat & PORTCH)
	    {
		Iprtch1++;
	    }
	    sv85initduarts (line);
	}
	if ((intrs [line] & 0x08) == 0)
	{
	    Iduart2++;
	    sv85initduarts (line);
	}
	resp [line] = BASE [PEERMBX];
	sv85initduarts (line);
	semGive (semids [line]);
	return;
}
/**********
*
*	This routine prints out the interrupt statistics
*
**********/
void chkints (unsigned short duart)
{
	printf ("ATX: %ld ARX: %ld ABRK: %ld TMR: %ld BTX: %ld BRX: %ld BBRK: %ld PCH: %ld\n",
		Iatx1, Iarx1, Iabrk1, Itmr1, Ibtx1, Ibrx1, Ibbrk1, Iprtch1);
}

/**********
*
*	This routine contains the sv85 initialization sequence
*
**********/
int sv85initsequence (unsigned short line, unsigned char *mbaddr, unsigned short doconfig)
{
	unsigned short	status;
	unsigned short	i;
	unsigned char	*BASE;

	BASE = bases [line];

	status = sv85init (line);
	if (status == FALSE)
	{
	    return(FALSE);
	}
	status = sv85reset (line);
	if (status == FALSE)
	{
	    return (FALSE);
	}

	/* Program Loading does not want to start watchdog */

	if (!doconfig)
	{
	    return (TRUE);
	}
	status = sv85infcdiag (line);
	if (status == FALSE)
	{
	    return (FALSE);
	}

/*
	status = sv85configstat (line, 1, 0, mbaddr);
	if (status == FALSE)
	{
	    return (FALSE);
	}
*/
	for (i = 0; i < 8; i++)
	{
	    status = sv85abort (line, (i + 1));
	    if (status == FALSE)
	    {
		return(FALSE);
	    }
	}
	for (i = 0; i < 8; i++)
	{
	    status = sv85abort (line, ((i +1)  | 0x40));
	    if (status == FALSE)
	    {
		return (FALSE);
	    }
	}
	for (i = 0; i < 8; i++)
	{
	    status = sv85abort (line, ((i + 1) | 0x80));
	    if (status == FALSE)
	    {
		return (FALSE);
	    }
	}
	for (i = 0; i < 8; i++)
	{
	    status = sv85abort (line, ((i + 1)  | 0xc0));
	    if (status == FALSE)
	    {
		return (FALSE);
	    }
	}
	sv85_needs_init[line] = FALSE;
	putpath [line] = 0;
	return (TRUE);
}

/**********
*
*      init_plc_tab
*
*      The purpose of this routine is to initialize the line tables for the
*      analog devices I/O
*
**********/

/* only called from allocate_plc  */

static	void init_plc_tab (unsigned short line, unsigned short mod)
{
	MODPLC		*pmod;

	pmod = plcs [line] [mod];
	if (!pmod)
	{
	    return;
	}
	pmod->is_responding = FALSE;
	pmod->holdflag = FALSE;
	pmod->was_requested = FALSE;
	pmod->cmd_force_offline = FALSE;
	pmod->prog_load_stat = PLC_NOT_LOADING;
	pmod->prog_load_number = 0;
	pmod->prog_load_page = 0;
	pmod->plc_start = FALSE;
	pmod->plc_stop = FALSE;
	pmod->stat_fetch_time = 0;
	pmod->last_good_msg_time = 0;
	pmod->battery_okay |= 0x02;
	pmod->memory_protect |= 0x02;
	pmod->run_light |= 0x02;
	pmod->power_okay |= 0x02;
	initplc (pmod);
	initstats (pmod);

}

/**********
*
*	Get Service Request
*
**********/
int sv85getsvcreq (unsigned short line, unsigned char *buf)
{
	unsigned char	stat;
	unsigned short	i;
	unsigned char	*BASE;

	BASE = bases [line];

	stat = sv85msgxfer (line, 0x00, NULLP, 60);
	if (stat == TRUE)
	{
	    for (i = 0; i < 4; i++)
	    {
		lastsvcreq [i] = buf [i] = BASE [(RAMBASE + i)];	/* just get the first four byte for now */
	    }
	    return (TRUE);
	}
	sv85reset (line);
	return (FALSE);
}

int sv85init (unsigned short line)
{

	unsigned char	*BASE;

	BASE = bases [line];

	BASE [CONTROLREG] = SV85SRLOCK;
	taskDelay (15);
	BASE [CONTROLREG] = SV85SRLOCK + SV85SRRESET;

	taskDelay (600);

	sv85initduarts (line);

	BASE [CONTROLREG] = SV85SRLOCK + SV85INTENB + SV85SRRESET;

	taskDelay (60);
	return (TRUE);
}
void sv85initduarts (unsigned short line)
{
	unsigned char	*BASE;

	BASE = bases [line];
	BASE [D1CMDA] = RXDISABLE + TXDISABLE;
	BASE [D1CMDA] = RXRESET + TXRESET + ERSRESET + BRDRESET;
	BASE [D1CMDB] = RXDISABLE + TXDISABLE;
	BASE [D1CMDB] = RXRESET + TXRESET + ERSRESET + BRDRESET;
	BASE [D1AUX] = DI_DSRINTB + DI_DSRINTA + DI_CTSINTB + DI_CTSINTA;
	BASE [D1INTMSK] = 0;

	BASE [D2CMDA] = RXDISABLE + TXDISABLE;
	BASE [D2CMDA] = RXRESET + TXRESET + ERSRESET + BRDRESET;
	BASE [D2CMDB] = RXDISABLE + TXDISABLE;
	BASE [D2CMDB] = RXRESET + TXRESET + ERSRESET + BRDRESET;
	BASE [D2AUX] = DI_DSRINTB + DI_DSRINTA + DI_CTSINTB + DI_CTSINTA;
	BASE [D2INTMSK] = 0;
}
/**********
*
*	Software reset
*
**********/
int sv85reset (unsigned short line)
{
	unsigned char	stat;
	unsigned char	*BASE;

	BASE = bases [line];
	stat = sv85msgxfer (line, 0x08, NULLP, 120);

	BASE [CONTROLREG] = SV85SRLOCK + SV85SRRESET;

	taskDelay (180);

	BASE [CONTROLREG] = SV85SRLOCK + SV85INTENB + SV85SRRESET;


	if (mltrace [line])
	{
	    logMsg("SV85reset crash code = %d \n", BASE [RAMBASE], 0, 0, 0, 0, 0);
	}

	/* 0x00 idicates healthy peer processor interface 	*/
	/* 0x01-0x7f indicates crashed peer procesor		*/
	/*   caused  by invalid interface command		*/
	/* 0x80-0xff indicates crashed peer processor		*/
	/*   caused by internal fatal error			*/

	return (stat);
}
/**********
*
*	Abort Transaction
*
**********/
int sv85abort (unsigned short line, unsigned char path)
{
	unsigned char	stat;
	unsigned char	*BASE;

	BASE = bases [line];

	if (modtrace [line] == 5)
	{
	    logMsg ("Abort Processed for Path %d, pp = %x \n", path, putpath [line], 0, 0, 0, 0);
	}

	BASE [RAMBASE] = path;
	stat = sv85msgxfer (line, 0x05, NULLP, 60);
	if (stat == TRUE)
	{
	    if (BASE [RAMBASE] == path)	/* requested internal path # echoed back */
	    {
		return (TRUE);
	    }
	}
	sv85reset (line);
	return (FALSE);
}
/**********
*
*	Message transfer routine
*
**********/
unsigned char sv85msgxfer (unsigned short line, unsigned char cmd , char *buf,
		unsigned short tics)
{
	unsigned long	i;
	unsigned char	hmbx, pmbx;
	unsigned char	*BASE;


	BASE = bases [line];

        /* Clear the semaphore for sync */

        semTake (semids [line], NO_WAIT);
        semTake (semids [line], NO_WAIT);

	resp [line] = cmd;
	intrs [line] = 0xff;

	BASE [HOSTMBX] = cmd;

	if (modtrace [line])
	{
	    sv85xfertrace [sv85tix] = cmd;
	    sv85xfertrace [sv85tix] = (sv85xfertrace [sv85tix] << 8) + BASE [RAMBASE];
	    sv85tix = ((sv85tix + 1) % 1024);
	}

	if (needdelay [line])
	{
	    taskDelay (1);
	}
	++Ireqs;
	for (i = 0; i < tics; i += 10)
	{
	    if ((semTake (semids [line], 1)) != ERROR)
	    {

		if ((intrs [line] & SV85SRPEERRAM) == 0x00)	/*  see  if peer relinqished dual port ram */
		{
		    hmbx = BASE [HOSTMBX];
		    pmbx = ~resp [line];
		    if (hmbx == pmbx)	/* check for cmd echo */
		    {
			return(TRUE);
		    }
		    logMsg ("hostmbx = %x, peermbx = %x,\n", BASE [HOSTMBX], 
				BASE [PEERMBX], 0, 0, 0, 0);			
		    sv85_needs_init [line] = TRUE;
		    return (FALSE);
		}
	    }
	    ++Iwait;
	}
	logMsg ("command = %x,	peermbx = %x,	status register = %x\n",
				 cmd,
				 resp [line],
				 intrs [line], 0, 0, 0);	
		
	sv85_needs_init [line] = TRUE;

	return (FALSE);	/* TIMEOUT! did not get shared RAM back in time */
}

/**********
*
*	Interface Diagnostic
*
**********/
int sv85infcdiag (unsigned short line)
{
	unsigned char	stat, i;
	unsigned short	bufix;
	unsigned char	*BASE;

	BASE = bases [line];

	bufix = 0;
	
	while (bufix < 0x3fe)
	{
	    for (i = 0; (i < 0xf0); i++)
	    {
		BASE [(RAMBASE + bufix++)] = (0xfe - i);
		if (bufix >= 0x3fe)
		{
		    break;
		}
	    }
	}
	stat = sv85msgxfer (line, 0x07, NULLP, 120);

	if (stat == TRUE)
	{
	    bufix = 0;
	    while (bufix < 0x3fe)
	    {
		for (i = 0; (i < 0xf0); i++)
		{
		    if (BASE [(RAMBASE + bufix++)] !=  ((~(0xfe - i)) & 0xff))
		    {
			return(FALSE);
		    }
		    if (bufix >= 0x3fe)
		    {
			break;
		    }
		}
	    }
	}
	else
	{
	    sv85reset (line);
	    return (FALSE);
	}
	return( TRUE);
}


/**********
*
* Cold Init a PLC
*
**********/

/* only called from init_line_tabs  */

static void initplc (MODPLC  *pmod)
{

	unsigned short	i, j;

                                                                	
/*                                                              	
*      COILS ARE INITIALIZED AS READ WRITE VALUES              	
*      WRITES ARE TREATED BY INDIVIDUAL BITS NOT GROUPS       		
*              - NOT REQUESTED                                		
*              - READ BACK NOT VALID                             	
*              - NO NEW VALUES TO GO                                	
*/

	for (i = 0; (i < NUMMODCOILS); i++)
	{
	    pmod->coils [i].was_requested = FALSE;
	    pmod->coils [i].data_valid = FALSE;
	    pmod->coils [i].last_read = 0;
	}

/*                                                                 	
*      DIGITALS ARE INITIALIZED AS NOT REQUESTED AND NOT VALID		
*/
	for (i = 0; (i < NUMMODDIS); i++)
	{
	    pmod->dis [i].was_requested = FALSE;
	    pmod->dis [i].data_valid = FALSE;
	}

/*                                                                 	
*      ANALOG INS ARE INITIALIZED AS NOT REQUESTED AND NOT VALID    	
*/
	for (i = 0; (i < NUMMODAIS); i++)
	{
	    pmod->ais [i].was_requested = FALSE;
	    pmod->ais [i].data_valid = FALSE;
	}

/*                                                             		
*      REGISTERS ARE INITIALIZED AS READ WRITE VALUES             	
*      WRITES ARE TREATED BY INDIVIDUAL REGISTERS NOT GROUPS      	
*              - NOT REQUESTED                                    	
*              - READ BACK NOT VALID                              	
*              - NO NEW VALUES TO GO                               	
*/                                                                  	
	for (i = 0; (i < NUMREGISTERS); i++)
	{
	    pmod->regs [i].was_requested = FALSE;
	    pmod->regs [i].data_valid = FALSE;
	    for (j = 0; j < REGBLOCK; j++)
	    {
		pmod->regs [i].data_togo [j] = FALSE;
	    }
	}

}


/**********                                                           		
*                                                              	
* initialize the statistics for a modicon PLC                  	
*                                                              	
**********/                                                               	

static	void initstats (MODPLC *pmod)
{
	pmod->no_dataout_msgs = 0;  /* 5  Count Output Msgs 		*/
	pmod->no_datain_msgs = 0;   /* 6  Count Data Requests 		*/
	pmod->no_rxnoresp = 0;      /* 7  Count No Responses 		*/
	pmod->no_txnoresp = 0;      /* 8  Count No Responses 		*/
	pmod->no_dataerr = 0;       /* 9  Count Message Echo Errors 	*/
	pmod->no_retries = 0;       /* 10 Count Retries 		*/          	
	pmod->no_rxhwerrs = 0;      /* 11 Count Comm Hardware Errors	*/
	pmod->no_txhwerrs = 0;      /* 12 Count Comm Hardware Errors 	*/
	pmod->no_crcerrs = 0;       /* 13 Count CRC Errors 		*/
	pmod->no_offlines = 0;      /* 14 Count Off Line Transitions 	*/
	pmod->badoutcnt = 0;	    /* 17 Count failed output commands 	*/
}

/**********
*
* modloopback                                                            
*                                                                        
**********/                                                                          	

/* only called from on_line */

static	unsigned short modloopback (unsigned short line, unsigned short mod,
			 MODPLC *pmod)
{
	TXLINE 		*txp;
	RXLINE		*rxp;

	/* Set up pointers for comm buffers */

	txp = txlinetab [line & (~ONERETRY)];
	rxp = rxlinetab [line & (~ONERETRY)];

	/* Format the output message */

	txp->txbuf [0] =  mod;
	txp->txbuf [1] =  MODLOOPBACK;
	txp->txbuf [2] =  0;
	txp->txbuf [3] =  0;
	txp->txbuf [4] =  0xA5;
	txp->txbuf [5] =  0x37;

	makecrc (&(txp->txbuf [0]), MODLOOPBACKLEN, &(txp->txbuf [MODLOOPBACKLEN]));

	txp->txmsglen = MODLOOPBACKLEN + MODCRCLEN;

	pmod->no_datain_msgs++;

	/* Send out the message */

	if (msgxfer (line, FIXEDRESP, pmod, MODLOOPBACKRESPLEN) == XFERBAD)
	{
	    /* >>> Did not get turn around message so mark unit as off line */
	    if (modtrace [line])
	    {
		logMsg ("Failed to bring unit on line  mod = %d \n", mod, 0, 0, 0, 0, 0);
	    }
	    pmod->is_responding = FALSE;
	    return (FALSE);
	}
	return (TRUE);
}

/*********
*
*     prepresponses
*                                                                             	
*      The purpose of this routine is to mark the PLCs as on hold and the    	
*	data as not valid until read again from the PLC                       	
*                                                                           	
**********/

static	void prepresponses (unsigned short line)
{
	unsigned short	mod;
	MODPLC		*pmod;

	for (mod = 1; (mod <= NUMMODPLCS); mod++)
	{
            pmod = plcs [line] [mod];
	    if (pmod)			/* verify pointer validity prior to use */
	    {
	    	pmod->holdflag = TRUE;
		mark_not_valid (pmod);
            }
	}
}

/**********
*
*      any_responding                                               	
*                                                                   	
*      The purpose of this routine is to see if any units are       	
*	responding on this line                                      	
*                                                                   	
**********/

static	int any_responding (unsigned short line)
{
	unsigned short  mod;
	MODPLC		*pmod;

	for (mod = 1; (mod <= NUMMODPLCS); mod++)
	{
	    pmod = plcs [line] [mod];
	    if (pmod)			/* verify pointer validity prior to use */
	    {
                /* check cmd_force_offline flag because it is set in
			an asynchronous task */

                if ((pmod->is_responding) && (!pmod->cmd_force_offline))
		{
		    return (TRUE);
		}
	    }
	}
	return (FALSE);
}

/**********
*
* Readcoilsfd
*  - full duplex version
*
**********/
static	unsigned short readcoilsfd(unsigned short line, unsigned short mod,
			 MODPLC *pmod, unsigned short bitaddr,
			 unsigned short readlen,unsigned short updateflag)
{

	TXLINE  	*txp;
	RXLINE  	*rxp;
	PATHINFO	*path;
	unsigned short  i;
	unsigned long	thistime;


	/* Set up pointers for comm buffers */

	txp = txlinetab [line & (~ONERETRY)];
	rxp = rxlinetab [line & (~ONERETRY)];

	/* Count number of data request messages */

	pmod->no_datain_msgs++;


	/* Format the output message */

	txp->txbuf [0] =  mod;
	txp->txbuf [1] =  MODCOILREAD;
	txp->txbuf [2] =  (bitaddr >> 8);
	txp->txbuf [3] =  (bitaddr & 0xFF);
	txp->txbuf [4] =  (readlen >> 8);
	txp->txbuf [5] =  (readlen & 0xFF);

	makecrc (&(txp->txbuf [0]), MODREADMSGLEN, &(txp->txbuf [MODREADMSGLEN]));

	txp->txmsglen = MODREADMSGLEN + MODCRCLEN;

	/* Send out the message */

	if (msgxmit (line, VARRESP, pmod, 0, &path) == XFERBAD)
	{
	    thistime = tickGet();
	    if (  (thistime - pmod->last_good_msg_time)
	        > (sv85nodetimeout*60)) {
		if (pmod->is_responding)
		{
		    pmod->no_offlines++;
		}
		/* >>> Could not send message so mark unit as off line */
		pmod->holdflag = TRUE;
		pmod->is_responding = FALSE;
	        if (modtrace [line])
	        {
		    logMsg ("Unit going offline - mod = %d \n", mod, 0, 0, 0, 0, 0);
	        }
	    }
	    if (modtrace [line])
	    {
		logMsg ("MsgXmit Read Coils Failed for mod = %d\n",
			 mod, 0, 0, 0, 0, 0);
	    }

	    return (FALSE);
	}
	path->mod = mod;
	path->timeout = tickGet ();
	path->function = RCOIL;
	path->pmod = pmod;
	path->line = line;
	path->lsignal = bitaddr;
	path->resplen = 1;
	path->reqlen = (readlen + 7) / 8;
	path->updateflag = updateflag;
	for (i = 1; (i < 20); i++)
	{
	    path->respbuf [i] = txp->txbuf [i];
	}
	return (TRUE);
}

/**********
*
* get_plc_stats
*
**********/

/* only called from scan_plc */

static	unsigned short get_plc_stats (unsigned short line, unsigned short mod,
			 MODPLC *pmod)
{
	TXLINE		*txp;
	RXLINE		*rxp;
	unsigned long	thistime;

	/* Set up pointers for comm buffers */

	txp = txlinetab [line & (~ONERETRY)];
	rxp = rxlinetab [line & (~ONERETRY)];

	/* Format the output message */

	txp->txbuf [0] = mod;
	txp->txbuf [1] = MODREPSLVID;

	makecrc (&(txp->txbuf [0]), MODREPSLVIDMSGLEN, &(txp->txbuf [MODREPSLVIDMSGLEN]));

	txp->txmsglen = MODREPSLVIDMSGLEN + MODCRCLEN;

	/* Send out the message */

	if (msgxfer (line, FIXEDRESP, pmod, MODREPSLVIDRESPLEN) == XFERBAD)
	{
	    thistime = tickGet();
	    if (  (thistime - pmod->last_good_msg_time)
	        > (sv85nodetimeout*60)) {
		if (pmod->is_responding)
		{
		    pmod->no_offlines++;
		}
		/* >>> Could not get statistics so mark unit as off line */
		pmod->holdflag = TRUE;
		pmod->is_responding = FALSE;
		pmod->battery_okay |= 0x02;
		pmod->memory_protect |= 0x02;
		pmod->run_light |= 0x02;
		pmod->power_okay |= 0x02;
	        if (modtrace [line])
	        {
		    logMsg ("Unit going offline - mod = %d \n", mod, 0, 0, 0, 0, 0);
	        }
	    }
	    if (modtrace [line])
	    {
	 	logMsg ("MsgXfer Get Plc Stats Failed for mod = %d\n",
			  mod, 0, 0, 0, 0, 0);
	    }
	    return (-1);
	}

	/* Verify the response */

	if (rxp->rxbuf [0] != mod)
	{
	    return (-1);
	}
	if (rxp->rxbuf [1] != MODREPSLVID)
	{
	    return (-1);
	}
	if (rxp->rxbuf [2] != 0x09)
	{
	    return (-1);
	}
	if (rxp->rxbuf [3] != 0x09)
	{
	    return (-1);
	}
	pmod->battery_okay = (rxp->rxbuf [9] & BATTERY_OK ? 1 : 0);
	pmod->memory_protect = (rxp->rxbuf [9] & MEM_PROTECT ? 1 : 0);
	pmod->run_light = (rxp->rxbuf [9] & RUN_LIGHT ? 1 : 0);
	pmod->power_okay = (rxp->rxbuf [9] & POWER_OK ? 1 : 0);
	return (0);
}



/**********
*
* Readdisfd
*
**********/

/* only called from scan_plc */

static	unsigned short readdisfd (unsigned short line, unsigned short mod,
			 MODPLC *pmod, unsigned short bitaddr,
			 unsigned short readlen, unsigned short updateflag)
{
	TXLINE		*txp;
	RXLINE		*rxp;
	PATHINFO	*path;
	unsigned short  i;
	unsigned long	thistime;

	/* Set up pointers for comm buffers */

	txp = txlinetab [line & (~ONERETRY)];
	rxp = rxlinetab [line & (~ONERETRY)];

	/* Count number of data request messages */

	pmod->no_datain_msgs++;

	/* Format the output message */

	txp->txbuf [0] =  mod;
	txp->txbuf [1] =  MODDIREAD;
	txp->txbuf [2] =  (bitaddr >> 8);
	txp->txbuf [3] =  (bitaddr & 0xFF);
	txp->txbuf [4] =  (readlen >> 8);
	txp->txbuf [5] =  (readlen & 0xFF);

	makecrc (&(txp->txbuf [0]), MODREADMSGLEN, &(txp->txbuf [MODREADMSGLEN]));

	txp->txmsglen = MODREADMSGLEN + MODCRCLEN;

	/* Send out the message */

	if (msgxmit (line, VARRESP, pmod, 0, &path) == XFERBAD)
	{
	    thistime = tickGet();
	    if (  (thistime - pmod->last_good_msg_time)
	        > (sv85nodetimeout*60)) {
		if (pmod->is_responding)
		{
		    pmod->no_offlines++;
		}
		/* >>> Could not send message so mark unit as off line */
		pmod->holdflag = TRUE;
		pmod->is_responding = FALSE;
	        if (modtrace [line])
	        {
		    logMsg ("Unit going offline - mod = %d \n", mod, 0, 0, 0, 0, 0);
	        }
	    }
	    if (modtrace [line])
	    {
	 	logMsg ("MsgXmit read disfd Failed for mod = %d\n",
			  mod, 0, 0, 0, 0, 0);
	    }
	    return (FALSE);
	}
	path->mod = mod;
	path->timeout = tickGet ();
	path->function = RDI;
	path->pmod = pmod;
	path->line = line;
	path->lsignal = bitaddr;
	path->resplen = 1;
	path->reqlen = (readlen + 7) / 8;
	path->updateflag = updateflag;
	for (i = 1; (i < 20); i++)
	{
	    path->respbuf [i] = txp->txbuf [i];
	}
	return (TRUE);
}

/**********
*                                                               	
* Readaisfd                                                        	
*                                                                	
**********/

/* only called from scan_plc */

static	unsigned short readaisfd (unsigned short line, unsigned short mod,
			 MODPLC *pmod, unsigned short regaddr,
			 unsigned short reglen, unsigned short updateflag)
{
	TXLINE		*txp;
	RXLINE		*rxp;
	PATHINFO	*path;
	unsigned short  i;
	unsigned long	thistime;

	/* Set up pointers for comm buffers */

	txp = txlinetab [line & (~ONERETRY)];
	rxp = rxlinetab [line & (~ONERETRY)];

	/* Count number of data request messages */

	pmod->no_datain_msgs++;

	/* Format the output message */

	txp->txbuf [0] =  mod;
	txp->txbuf [1] =  MODAIREAD;
	txp->txbuf [2] =  (regaddr >> 8);
	txp->txbuf [3] =  (regaddr & 0xFF);
	txp->txbuf [4] =  (reglen >> 8);
	txp->txbuf [5] =  (reglen & 0xFF);

	makecrc (&(txp->txbuf [0]), MODREADMSGLEN, &(txp->txbuf [MODREADMSGLEN]));

	txp->txmsglen = MODREADMSGLEN + MODCRCLEN;

	/* Send out the message */

	if (msgxmit (line, VARRESP, pmod, 0, &path) == XFERBAD)
	{
	    thistime = tickGet();
	    if (  (thistime - pmod->last_good_msg_time)
	        > (sv85nodetimeout*60)) {
		if (pmod->is_responding)
		{
		    pmod->no_offlines++;
		}
		/* >>> Could not send message so mark unit as off line */
		pmod->holdflag = TRUE;
		pmod->is_responding = FALSE;
	        if (modtrace [line])
	        {
		    logMsg ("Unit going offline - mod = %d \n", mod, 0, 0, 0, 0, 0);
	        }
	    }
	    if (modtrace [line])
	    {
	 	logMsg ("MsgXmit read aisfd Failed for mod = %d\n",
			  mod, 0, 0, 0, 0, 0);
	    }
	    return (FALSE);
	}

	/* Verify the response data */
	path->mod = mod;
	path->timeout = tickGet ();
	path->function = RAI;
	path->regaddr = regaddr;
	path->pmod = pmod;
	path->line = line;
	path->resplen = 1;
	path->reqlen = (reglen * 2);
	path->updateflag = updateflag;
	for (i = 1; (i < 20); i++)
	{
	    path->respbuf [i] = txp->txbuf [i];
	}
	return (TRUE);
}

/**********
*
* Readregsfd
*	read registers full duplex
*
***********/
static	unsigned short readregsfd (unsigned short line, unsigned short mod,
			 MODPLC *pmod, unsigned short regaddr,
			 unsigned short reglen, unsigned short updateflag)
{

	TXLINE  	*txp;
	RXLINE  	*rxp;
	PATHINFO	*path;
	unsigned short  i;
	unsigned long	thistime;

	/* Set up pointers for comm buffers */

	txp = txlinetab [line & (~ONERETRY)];
	rxp = rxlinetab [line & (~ONERETRY)];

	/* Count number of data request messages */

	pmod->no_datain_msgs++;

	/* Format the output message */

	txp->txbuf [0] =  mod;
	txp->txbuf [1] =  MODREGREAD;
	txp->txbuf [2] =  (regaddr >> 8);
	txp->txbuf [3] =  (regaddr & 0xFF);
	txp->txbuf [4] =  (reglen >> 8);
	txp->txbuf [5] =  (reglen & 0xFF);

	makecrc (&(txp->txbuf [0]), MODREADMSGLEN, &(txp->txbuf [MODREADMSGLEN]));

	txp->txmsglen = MODREADMSGLEN+MODCRCLEN;

	/* Send out the message */

	if (msgxmit (line, VARRESP, pmod, 0, &path) == XFERBAD)
	{
	    thistime = tickGet();
	    if (  (thistime - pmod->last_good_msg_time)
	        > (sv85nodetimeout*60)) {
		if (pmod->is_responding)
		{
		    pmod->no_offlines++;
		}
		/* >>> Could not send message so mark unit as off line */
		pmod->no_offlines++;
		pmod->holdflag = TRUE;
		pmod->is_responding = FALSE;
	        if (modtrace [line])
	        {
		    logMsg ("Unit going offline - mod = %d \n", mod, 0, 0, 0, 0, 0);
	        }
	    }
	    if (modtrace [line])
	    {
	 	logMsg ("MsgXmit read regsfd Failed for mod = %d\n",
			  mod, 0, 0, 0, 0, 0);
	    }
	    return (FALSE);
	}

	path->mod = mod;
	path->timeout = tickGet ();
	path->function = RREG;
	path->regaddr = regaddr;
	path->pmod = pmod;
	path->line = line;
	path->resplen = 1;
	path->reqlen = (reglen * 2);
	path->updateflag = updateflag;
	for (i = 1; (i < 20); i++)
	{
	    path->respbuf [i] = txp->txbuf [i];
	}
	return (TRUE);
}

/**********
*
* Setcoil                                                        	
*                                                                 	
**********/

/* only called from checkoutputs */

static	unsigned short setcoil (unsigned short line, unsigned short mod,
			 MODPLC *pmod, unsigned short bitaddr,
			 unsigned long mask, unsigned long val)
{
	TXLINE		*txp;
	RXLINE		*rxp;
	unsigned short  i, ix, lsignal;
	unsigned long	lmask, lval, tmask ,tval, thistime;
	PATHINFO	*path;

	if (mask == 0)
	{
	    return (CMDREJECT);
	}
	tmask = lmask = mask;
	tval = lval = val;
	lsignal = bitaddr;

	/* Set up pointers for comm buffers */

	txp = txlinetab [line & (~ONERETRY)];
	rxp = rxlinetab [line & (~ONERETRY)];

	/* Count number of data request messages */

	pmod->no_dataout_msgs++;

	 /* adjust the bits to the low order bit */

	while ((tmask & 1) == 0)
	{
	    tval = tval >> 1;          /* adjust the value */
	    tmask = tmask >> 1;         /* adjust the mask */
	}

	/* Is it a single coil output */

	if ((tmask & 2) == 0)
	{
	    /* Format the output message */

	    txp->txbuf [0] =  mod;
	    txp->txbuf [1] =  MODCOILWRITE;
	    txp->txbuf [2] =  (lsignal >> 8);
	    txp->txbuf [3] =  (lsignal & 0xFF);
	    txp->txbuf [4] =  (lval == 0) ? 0 : 0xFF;  /* 0 => off  0xFF => on */
	    txp->txbuf [5] =  0;

	    makecrc (&(txp->txbuf [0]), MODSETCOILLEN, &(txp->txbuf [MODSETCOILLEN]));

	    txp->txmsglen = MODSETCOILLEN + MODCRCLEN;

	    /* Send out the message */

	    if (msgxmit (line, FIXEDRESP, pmod, MODWRITERESPLEN, &path) == XFERBAD)
	    {
		thistime = tickGet();
		if (  (thistime - pmod->last_good_msg_time)
	            > (sv85nodetimeout*60)) {
		    if (pmod->is_responding)
		    {
			pmod->no_offlines++;
		    }
		    /* >>> Could not send message so mark unit as off line */
		    pmod->holdflag = TRUE;
		    pmod->is_responding = FALSE;
	            if (modtrace [line])
	            {
		        logMsg ("Unit going offline - mod = %d \n", mod, 0, 0, 0, 0, 0);
	            }
	        }
	        if (modtrace [line])
	        {
	 	    logMsg ("MsgXmit setcoil Failed for mod = %d\n",
			      mod, 0, 0, 0, 0, 0);
	        }
		return (FALSE);
	    }

	    path->mod = mod;
	    path->timeout = tickGet ();
	    path->function = SCOIL;
	    path->pmod = pmod;
	    path->line = line;
	    path->lsignal = lsignal;
	    path->lval = lval;
	    path->resplen = MODSETCOILLEN - 1;
	    for (i = 1; (i < 20); i++)
	    {
		path->respbuf [i] = txp->txbuf [i];
	    }

	    return (TRUE);

	}
	else
	{
	    /* Must be a multiple coil output */

	    lmask = tmask;
	    lval = tval;

	    /* count the number of coils to output */

	    for (i = 0, tmask = 1; ((tmask & lmask) != 0); i++)
	    {
	    	tmask = tmask << 1;
	    }

	    /* Format the output message */

	    txp->txbuf [0] =  mod;
	    txp->txbuf [1] =  MODMCOILWRITE;
	    txp->txbuf [2] =  (lsignal >> 8);
	    txp->txbuf [3] =  (lsignal & 0xFF);
	    txp->txbuf [4] =  (i >> 8);          /* number of coils */
	    txp->txbuf [5] =  (i & 0xFF );
	    txp->txbuf [6] =  (((i + 7) / 8) & 0xFF); /* number of data bytes */
	    txp->txbuf [7] =  (lval & 0xFF );
	    ix = 8;
	    if (i > 8)
	    {
		txp->txbuf [ix++] =  ((lval >> 8) & 0xFF );
		if (i > 16)
		{
		    txp->txbuf [ix++] =  ((lval >> 16) & 0xFF );
		    if (i > 24)
		    {
			txp->txbuf [ix++] =  ((lval >> 24) & 0xFF );
		    }
		}
	    }

	    makecrc (&(txp->txbuf [0]), ix, &(txp->txbuf [ix]));

	    txp->txmsglen = ix+MODCRCLEN;

	    /* Send out the message */

	    if (msgxmit (line, FIXEDRESP, pmod, MODWRITERESPLEN, &path) == XFERBAD)
	    {
		thistime = tickGet();
		if (  (thistime - pmod->last_good_msg_time)
	            > (sv85nodetimeout*60)) {
		    if (pmod->is_responding)
		    {
			pmod->no_offlines++;
		    }
		    /* >>> Could not send message so mark unit as off line */
		    pmod->no_offlines++;
		    pmod->holdflag = TRUE;
		    pmod->is_responding = FALSE;
	            if (modtrace [line])
	            {
		        logMsg ("Unit going offline - mod = %d \n", mod, 0, 0, 0, 0, 0);
	            }
	        }
	        if (modtrace [line])
	        {
	 	    logMsg ("MsgXmit set multiple coils Failed for mod = %d\n",
			      mod, 0, 0, 0, 0, 0);
	        }
		return (FALSE);
	    }
	    path->mod = mod;
	    path->timeout = tickGet ();
	    path->function = SMCOIL;
	    path->pmod = pmod;
	    path->line = line;
	    path->lsignal = lsignal;
	    path->val = val;
	    path->mask = mask;
	    path->resplen = MODMSETCOILRESPLEN - 1;
	    for (i = 1; (i < 20); i++)
	    {
		path->respbuf [i] = txp->txbuf [i];
	    }
	    return (TRUE);
	}
}

/**********
*
* Setreg                                                       	
*                                                                  	
**********/

/* only called from checkoutputs */

static	unsigned short setreg (unsigned short line, unsigned short mod,
			 MODPLC *pmod, unsigned short regaddr,
			 unsigned short val,double dblval, unsigned short conv)
{
	unsigned long	thistime;
	TXLINE 		*txp;
	RXLINE 		*rxp;
	unsigned short	i;
   	union 	u_value
		{
		    short		ivalue;
		    long		lvalue;
		    float		fvalue;
		    double		dvalue;
		    char		afvalue [100];
		}  uval;


	/* Set up pointers for comm buffers */

	txp = txlinetab [line & (~ONERETRY)];
	rxp = rxlinetab [line & (~ONERETRY)];

	/* Count number of data request messages */

	pmod->no_dataout_msgs++;

	uval.fvalue = dblval;

	/* Format the output message */

	txp->txbuf [0] =  mod;
	txp->txbuf [2] =  (regaddr >> 8);
	txp->txbuf [3] =  (regaddr & 0xFF);

	if (conv != ASCFLT)
	{
	    txp->txbuf [1] =  MODREGWRITE;
	    txp->txbuf [4] =  (val >> 8);
	    txp->txbuf [5] =  (val & 0xFF);

	    makecrc (&(txp->txbuf [0]), MODSETREGLEN, &(txp->txbuf [MODSETREGLEN]));

	    txp->txmsglen = MODSETREGLEN + MODCRCLEN;

	    /* Send out the message */

	    if (msgxfer (line, FIXEDRESP, pmod, MODWRITERESPLEN) == XFERBAD)
	    {
		thistime = tickGet();
		if (  (thistime - pmod->last_good_msg_time)
	            > (sv85nodetimeout*60)) {
		    if (pmod->is_responding)
		    {
			pmod->no_offlines++;
		    }
		    /* >>> Could not set registers so mark unit as off line */
		    pmod->holdflag = TRUE;
		    pmod->is_responding = FALSE;
	            if (modtrace [line])
	            {
		        logMsg ("Unit going offline - mod = %d \n", mod, 0, 0, 0, 0, 0);
	            }
	        }
	        if (modtrace [line])
	        {
	 	    logMsg ("MsgXfer set regs Failed for mod = %d\n",
			      mod, 0, 0, 0, 0, 0);
	        }
	        return (CMDFAIL);
	    }

	    /* Verify the response - the message is repeated back*/

	    for (i = 0; (i < MODSETREGLEN); i++)
	    {
		if (txp->txbuf [i] != rxp->rxbuf [i])
		{
		    pmod->no_dataerr++;
		    return (CMDNACK);
	        }
	    }
	    return (CMDDONE);
	}
	else
	{ /* Send out float as multiple register write */

	    txp->txbuf [1] =  MODMREGWRITE;
	    txp->txbuf [4] = 0x00;
	    txp->txbuf [5] = 0x02;	/* Number of registers */
	    txp->txbuf [6] = 0x04;	/* Number of bytes */

	    txp->txbuf [7] = (uval.lvalue >> 8) & 0xFF;
	    txp->txbuf [8] =  uval.lvalue & 0xFF;
	    txp->txbuf [9] = (uval.lvalue >> 24) & 0xFF;
	    txp->txbuf [10] = (uval.lvalue >> 16) & 0xFF;


	    makecrc (&(txp->txbuf [0]), MODSET2REGLEN, &(txp->txbuf [MODSET2REGLEN]));

	    txp->txmsglen = MODSET2REGLEN + MODCRCLEN;

	    /* Send out the message */

	    if (msgxfer (line, FIXEDRESP, pmod, MODWRITERESPLEN) == XFERBAD)
	    {
		thistime = tickGet();
		if (  (thistime - pmod->last_good_msg_time)
	            > (sv85nodetimeout*60)) {
		    if (pmod->is_responding)
		    {
			pmod->no_offlines++;
		    }
		    /* >>> Could not set registers so mark unit as off line */
		    pmod->holdflag = TRUE;
		    pmod->is_responding = FALSE;
	            if (modtrace [line])
	            {
		        logMsg ("Unit going offline - mod = %d \n", mod, 0, 0, 0, 0, 0);
	            }
	        }
	        if (modtrace [line])
	        {
	 	    logMsg ("MsgXfer set multiple regs Failed for mod = %d\n",
			      mod, 0, 0, 0, 0, 0);
	        }
	        return (CMDFAIL);
	    }

	    /* Verify the response - the message is repeated back*/

	    for (i = 0; (i < MODSETMREGRESPLEN); i++)
	    {
		if (txp->txbuf [i] != rxp->rxbuf [i])
		{
		    pmod->no_dataerr++;
		    return (CMDNACK);
	        }
	    }
	    return (CMDDONE);
	}
}

/**********
*
* Setstring                                                       	
*                                                                  	
**********/
static	unsigned short setstring (unsigned short line, unsigned short mod,
			          MODPLC *pmod, unsigned short signal,
			          unsigned char *buf, unsigned short olen)
/*  if (!(setstring (link, adapter, pmod, stringaddr, pstringval, stringlen))) */
/* only called from checkoutputs */
{
	TXLINE 		*txp;
	RXLINE 		*rxp;
	unsigned short  i, slen, tlen, base;
	unsigned long	thistime;


	/* Set up pointers for comm buffers */

	txp = txlinetab [line & (~ONERETRY)];
	rxp = rxlinetab [line & (~ONERETRY)];

	/* Count number of data request messages */

	pmod->no_dataout_msgs++;

	/* Format the output message */

	txp->txbuf [0] =  mod;
        txp->txbuf [1] =  MODMREGWRITE;
        txp->txbuf [2] =  (signal >> 8);
        txp->txbuf [3] =  (signal & 0xFF);

	/* Store the transmit data                      */

	base = 7;
        i = 0;
	slen = olen;
	while  (i < STRINGMSGLEN)
        {
	    if ((slen != 0) && (i >= slen))
	    {
               break;
	    }
	    txp->txbuf [base + i] = buf [i];

	    if ((slen == 0) && ((buf [i] == CR) || (buf [i] == 0)))
            {
                i++;
                break;
            }
	    i++;
	}
        if ((i % 2) == 1) /* convert all odd length strings to even length strings */
        {
            txp->txbuf [base + i] = buf [i];
            i++;
        }
 	slen = i;
        tlen = base + slen;

	txp->txbuf [4] = ((slen / 2) / 256);
	txp->txbuf [5] = ((slen / 2) & 0xFF);
        txp->txbuf [6] = slen;

	makecrc (&(txp->txbuf [0]), tlen, &(txp->txbuf [tlen]));
        txp->txmsglen = tlen + MODCRCLEN;

        /* Send out the message */

        if (msgxfer (line, FIXEDRESP, pmod, MODWRITERESPLEN) == XFERBAD)
        {
	    thistime = tickGet();
	    if (  (thistime - pmod->last_good_msg_time)
	        > (sv85nodetimeout*60)) {
		if (pmod->is_responding)
		{
		    pmod->no_offlines++;
		}
		/* >>> Could not write a string so mark unit as off line */
		pmod->holdflag = TRUE;
		pmod->is_responding = FALSE;
	        if (modtrace [line])
	        {
		    logMsg ("Unit going offline - mod = %d \n", mod, 0, 0, 0, 0, 0);
	        }
	    }
	    if (modtrace [line])
	    {
	 	logMsg ("MsgXfer set string Failed for mod = %d\n",
			  mod, 0, 0, 0, 0, 0);
	    }
	    return (CMDFAIL);
        }

        /* Verify the response - the message is repeated back*/

        for (i = 0; (i < MODSETMREGRESPLEN); i++)
        {
	    if (txp->txbuf [i] != rxp->rxbuf [i])
            {
                pmod->no_dataerr++;
                return (CMDNACK);
            }
        }
        return (CMDDONE);
}

/**********
*
*	Read Area Descriptors                                		
*
**********/                                                                 	

/* only called from scan_plc */

static int loadinit (unsigned short line,MODPLC *pmod,unsigned short mod)
{
	LOADINFO	*plinfo;
	char		addr [8];
        int 		status;
	long		pos;
	unsigned short	prognum;

	plinfo = &linfo [line];		/* establish the pointer */

	/* If already one being loaded wait else claim it */

	if (plinfo->loading)
	{
	    return(0);
	}
	plinfo->line = line;	/* easy way to find it */
	plinfo->loading = TRUE;
	plinfo->pmod = pmod;
	plinfo->mod = mod;
	prognum = pmod->prog_load_number;

	/* Make File Name	*/

	strcpy (plinfo->filename, "P");
	addr [0] = 'L';			/* Line Number */
	addr [1] = '0' + (line / 10);
	addr [2] = '0' + (line % 10);
	addr [3] = 'A';			/* Address */
	addr [4] = '0' + (mod / 10);
	addr [5] = '0' + (mod % 10);
	addr [6] = '0' + (prognum % 10);
	addr [7] = 0;
	strcat (plinfo->filename, addr);
	strcat (plinfo->filename, ".prg");

	/* Open File	*/

	iam (user, password);
	if ((plinfo->loadfile = fopen (&plinfo->filename [0], "r")) == NULL)
	{
	    terminateload (plinfo, line, FALSE, CANNOTOPENFILE);
	    return (-1);
	}

	/* Seek to First Area */

	plinfo->fileaddr = (PAGE0 * WORDLEN);
	if ((pos = fseek (plinfo->loadfile, plinfo->fileaddr, 0)) != OK)
	{
	    terminateload (plinfo, line, FALSE, CANNOTSEEKTO);
	    return (-1);
	}

	/* Read First Dump Block */

	if (fread (&plinfo->dadesc [0], 1, DUMPAREASIZE, plinfo->loadfile) != DUMPAREASIZE)
	{
	    terminateload (plinfo, line, FALSE, CANNOTREAD);
	    return (-1);
	}

	/* Seek to Next Area */

	plinfo->fileaddr = (PAGEF * WORDLEN);
	if ((pos = fseek (plinfo->loadfile, plinfo->fileaddr, 0)) != OK)
	{
	    terminateload (plinfo, line, FALSE, CANNOTSEEKTO);
	    return (-1);
	}

	/* Read Next Dump Block */

	if (fread (&plinfo->dadesc [1], 1, DUMPAREASIZE, plinfo->loadfile) != DUMPAREASIZE)
	{
	    terminateload (plinfo, line, FALSE, CANNOTREAD);
	    return (-1);
	}

	/* Seek to Next Area */

	plinfo->fileaddr = (ASCII * WORDLEN);
	if ((pos = fseek (plinfo->loadfile, plinfo->fileaddr, 0)) != OK)
	{
	    terminateload (plinfo, line, FALSE, CANNOTSEEKTO);
	    return (-1);
	}

	/* Read Next Dump Block */

	if (fread (&plinfo->dadesc [2], 1, DUMPAREASIZE, plinfo->loadfile) != DUMPAREASIZE)
	{
	    terminateload (plinfo, line, FALSE, CANNOTREAD);
	    return (-1);
	}

	/* Seek to Last Area */

	plinfo->fileaddr = (FUTURE * WORDLEN);
	if ((pos = fseek (plinfo->loadfile, plinfo->fileaddr, 0)) != OK)
	{
	    terminateload (plinfo, line, FALSE, CANNOTSEEKTO);
	    return (-1);
	}

	/* Read Last Dump Block */

	if (fread (&plinfo->dadesc [3], 1, DUMPAREASIZE, plinfo->loadfile) != DUMPAREASIZE)
	{
	    terminateload (plinfo, line, FALSE, CANNOTREAD);
	    return (-1);
	}

	/* Seek to The First Area in the File */

	plinfo->fileaddr = ((plinfo->dadesc [0].areas [0].strtaddr) * WORDLEN);
	if ((pos = fseek (plinfo->loadfile, plinfo->fileaddr, 0)) != OK)
	{
	    terminateload (plinfo, line, FALSE, CANNOTSEEKTO);
	    return (-1);
	}

	/* LogOut on PLC	*/

	if (!plc_logout (plinfo))
	{
	    terminateload (plinfo, line, FALSE, CANNOTLOGOUT);
	    return (-1);
	}

	/* LogIn on PLC	*/

	if (!plc_login (plinfo))
	{
	    terminateload (plinfo, line, FALSE, CANNOTLOGIN);
	    return (-1);
	}

	/* Read The Machine State	*/

	if (!get_plc_state (plinfo, &plinfo->machstate))
	{
	    terminateload (plinfo, line, FALSE, CANNOTREAD);
	    return (-1);
	}

	/* If Memory Protect Abort */

	if ((plinfo->machstate & MEMPROTECT) == 0)
	{
	    terminateload (plinfo, line, FALSE, MEMORYPROTECT);
	    return (-1);
	}

	/* If Running Stop The Controller */

	if ((plinfo->machstate & RUNNINGSTATE) == 0)
	{
	    if(!plc_stop (plinfo))
	    {
		terminateload (plinfo, line, FALSE, PLCWONTSTOP);
		return (-1);
	    }
	}

	/* Check for dim Awareness */

	status = plc_getdim (plinfo);
	if (status < 0)
	{
	    terminateload (plinfo, line, FALSE, PLCWONTGETDIM);
	    return (-1);
	}

	/* If not in dim awareness */

	if (status == 0)
	{
	    /* Send to dim awareness */

	    if (!plc_setdim (plinfo))
	    {
		terminateload (plinfo, line, FALSE, PLCWONTSETDIM);
		return (-1);
	    }
	    /* Verify dim awareness */

	    status = plc_getdim (plinfo);
	    if (status != 1)
	    {
		terminateload (plinfo, line, FALSE, PLCWONTBEDIM);
		return (-1);
	    }
	}

	/* Initialize the Load Pointer to the first area */

	plinfo->ldblock = 0;
	plinfo->ldarea = 0;

	/* Set The Load State Variables */

	plinfo->pmod->prog_load_stat = PLC_LOAD_IN_PROGRESS;
	plinfo->pmod->prog_load_page = 0;

	return (0);
}


/**********
*                                                              	
*	Load The Next Dump Block                               		
*
**********/

/* only called from scan_plc */

static int loadnextblock (unsigned short line)
{
	LOADINFO	*plinfo;
	unsigned short	type_mask, currarea;
	int		status;

	plinfo = &linfo [line];	/* establish the pointer */

	type_mask = ((plinfo->dadesc [plinfo->ldblock].areaspec << plinfo->ldarea) & AREATYPEMASK);
	switch (type_mask)
	{
	case LOGICAREATYPE:
	    status = loadlogic (plinfo);
	    break;

	case ASCIIAREATYPE:
	    status = loadmemory( plinfo, ASCIIAREATYPE);
	    break;

	case MEMORYAREATYPE:
	default:
	    status = loadmemory (plinfo, MEMORYAREATYPE);
	    break;
	}

	if (status != 0)
	{
	    return(-1);
	}
	currarea = ((plinfo->ldblock * AREASPERBLK) + plinfo->ldarea);
	if (currarea == CONFIGAREA)
	{
	    /* Exit Dim Arareness with a start command*/

	    if (!plc_start (plinfo))
	    {
		terminateload (plinfo, line, FALSE, PLCWONTSTART);
		return (-1);
	    }

	    /* !!Three messages were issued here: 	*/
	    /*	 sf=0x31				*/
	    /*	 sf=0x34				*/
	    /*	confirm dim awareness			*/

	    /* Init PLC (Special Message Issued Here) initialize msg*/

	    status = plc_init (plinfo);
	    if (!status)
	    {
		terminateload (plinfo, line, FALSE, PLCWONTINIT);
		return (-1);
	    }

	    /* Read and Save the Run Load Switch from the PLC */

	    if (!getrunmode (plinfo, &plinfo->rlmode))
	    {
		terminateload (plinfo, line, FALSE, CANTGETMODE);
		return (-1);
	    }

	    /* Set the Run Load Switch to Load */

	    if (!setrunmode (plinfo, LOADSWITCH))
	    {
		terminateload (plinfo, line, FALSE, CANTSETLOADMODE);
		return (-1);
	    }

	}

	/* Advance the block and area pointers for the next load operation */

	plinfo->ldarea = ((plinfo->ldarea + 1) % AREASPERBLK);
	if (plinfo->ldarea == 0)
	{
	    plinfo->ldblock++;
	}
	plinfo->pmod->prog_load_page = currarea + 1;

	/* When we reach the end clean up and enable the next load */

	if (plinfo->ldblock >= NUMDUMPAREAS)
	{
	    terminateload (plinfo, line, TRUE, LOADOK);
	}

	return (0);

}

/**********
*                                                                      	
*	Load a Logic Area                                                	
*
**********/                                                                        	

/* only called from loadnextblock */

static int loadlogic (LOADINFO *plinfo)
{
	long		pos;
	short		length, page, startaddr, startnode, endnode, dbix, status;
	short		dowrite, endwritten, thisnodecount;
	unsigned char	databuf [MAXTXMSG];

	page = plinfo->dadesc [plinfo->ldblock].page;
	startaddr = plinfo->dadesc [plinfo->ldblock].areas [plinfo->ldarea].strtaddr;
	length = plinfo->dadesc [plinfo->ldblock].areas [plinfo->ldarea].length * NODELEN;
	startnode = FALSE;
	endnode = FALSE;
	endwritten = FALSE;
	dowrite = FALSE;
	dbix = 0;
	thisnodecount = 0;

	while (length)
	{
	    /* While there are nodes to read */
	    /* Read a node (3 Bytes) if error abort */

	    status = fread (&databuf [dbix], 1, NODELEN, plinfo->loadfile);
	    if (status == 0)
	    {
		terminateload (plinfo, plinfo->line, FALSE, UNEXPECTDEOF);
		return (-1);
	    }
	    thisnodecount++;
	    length -= NODELEN;

	    /* If 5X0 set Endnode found */

	    if ((databuf [dbix] == EOLNODE) && (databuf [dbix + 2] == 0))
	    {
		endnode = TRUE;
	    }
	    /* If (<5)XX (this is a start) and I already have one */
	    /*   backup and write what We have so far and resume */

	    if ( (databuf [dbix] <= BOCNODE) && (startnode) )
	    {
		/* Back up over start node from current position */

		pos = fseek (plinfo->loadfile, (long) -NODELEN, 1);
		thisnodecount--;
		length += NODELEN;
		dowrite = TRUE;
		dbix -= NODELEN;
	    }

	    /* If this is the first start node set the flag */

	    if ((databuf [dbix] < BOCNODE) && (!startnode))
	    {
	        startnode = TRUE;
	    }

	    dbix += NODELEN;

	    if ((length == 0) ||
		 (dowrite) || 
		 ((!startnode) && (thisnodecount > 4) ) ||
		 (thisnodecount >= MAXMSGNODES))
	    {
		if (!endwritten)
		{   /* Continue to read but don't write */
		    /* Write logic load message out */
		    status = writelogic (
		    plinfo,
		    page,
		    startaddr,
		    dbix,
		    (char*) databuf);  /* databuf); ###*/
                                      /* cast because the local array is unsigned char */

		    /* Abort on error */
		    if (!status)
		    {
			terminateload (plinfo, plinfo->line, FALSE, LOGICWRITEERROR);
			return (-1);
		    }
		}
		dowrite = FALSE;
		startaddr += thisnodecount; /* Set Start for next block */
		thisnodecount = 0;
		startnode = FALSE;
		endwritten = endnode; 
		dbix = 0;		   /* Set index for next block */
 	    }
	}
	return (0);
}

/**********
*                                                       		
*	Load A Memory or ASCII Area                                  	
*
**********/

/* only called from loadnextblock */

static int loadmemory (LOADINFO *plinfo, unsigned short type)
{
	short		length, page, startaddr, status, thislength;
	char		databuf [MAXTXMSG];

	page = plinfo->dadesc [plinfo->ldblock].page;
	startaddr = plinfo->dadesc [plinfo->ldblock].areas [plinfo->ldarea].strtaddr;
	length = plinfo->dadesc [plinfo->ldblock].areas [plinfo->ldarea].length * WORDLEN;

	while (length)
	{
	    /* While there are nodes to read */
	    /* Read a block of data if error abort */

	    thislength = (length > MAXWORDMSGLEN) ? MAXWORDMSGLEN : length;
	    status = fread (databuf, 1, thislength, plinfo->loadfile);
	    if (status == 0)
	    {
		terminateload (plinfo, plinfo->line, FALSE, UNEXPECTDEOF);
		return (-1);
	    }
	    length -= thislength;
	    if (type == MEMORYAREATYPE)
	    {
		status = writemem (plinfo, page, startaddr, thislength, databuf);
	    }
	    else
	    {
		status = writeascii (plinfo, page, startaddr, thislength, databuf);
	    }
	    /* Abort on error */
	    if (!status)
	    {
		terminateload (plinfo, plinfo->line, FALSE, MEMORYWRITEERROR);
		return (-1);
	    }
	    startaddr += thislength / 2;
	}
	return (0);
}


/**********
*                                                             	
*	End The Load Sequence                                      	
*
**********/

/* called from loadinit, loadnextblock, loadlogic, loadmemory */

static int terminateload (LOADINFO *plinfo,unsigned short line,
	       unsigned short success,unsigned short code)
{
	/* Close The File		*/

	fclose (plinfo->loadfile);

	/* Release for another PLC */

	plinfo->loading = FALSE;

	if (!success)
	{
	    plinfo->pmod->prog_load_stat = PLC_LOAD_FAILED;
	    plinfo->pmod->prog_load_code = code;
	    return (0);
	}

	/* Switch The Controller from Load to Run */

	if (!setrunmode (plinfo, RUNSWITCH))
	{
	    terminateload (plinfo, line, FALSE, CANTSETRUNMODE);
	    return (-1);
	}

	/* Set The Controller to the mode stored in the file */
	/* Which was read back from the controller after config area loaded */

	if (!setrunmode (plinfo, plinfo->rlmode))
	{
	    terminateload (plinfo, line, FALSE, CANTSETLOADEDMODE);
	    return (-1);
	}

	/* !!Two messages were issued here: 		*/
	/*	 sf=0x31				*/
	/*	 sf=0x34				*/

	if (!plc_logout (plinfo))
	{
	    terminateload (plinfo, line, FALSE, CANNOTLOGOUT);
	    return (-1);
	}

	/* Set Pmod status back to no loading */

	plinfo->pmod->prog_load_stat = PLC_NOT_LOADING;
	plinfo->pmod->prog_load_code = code;
	return (0);
}

/**********
*                                                            		
*	PLC Logout                                                	
*
**********/

/* called from scan_plc, loadinit, and terminateload  */

static int plc_logout (LOADINFO *plinfo)
{
	TXLINE		*txp;
	RXLINE		*rxp;

	/* Set up pointers for comm buffers */

	txp = txlinetab [plinfo->line];
	rxp = rxlinetab [plinfo->line];

	/* Format the output message */

	txp->txbuf [0] =  plinfo->mod;	/* Address		*/
	txp->txbuf [1] =  MODLOAD;	/* Function Code	*/
	txp->txbuf [2] =  1;		/* Length		*/
	txp->txbuf [3] =  MODSFLOGOUT;	/* Subfunction Code 	*/

	makecrc (&(txp->txbuf [0]), 4, &(txp->txbuf [4]));

	txp->txmsglen = 6;

	/* Send out the message */

	if (msgxfer (plinfo->line, VARRESP, plinfo->pmod, 0) == XFERBAD)
	{
/* >>> Could not send logout message so mark unit as off line */
	    plinfo->pmod->no_offlines++;
	    plinfo->pmod->holdflag = TRUE;
	    plinfo->pmod->is_responding = FALSE;
	    return (FALSE);
	}

	/* Verify the response */

	if ((rxp->rxbuf [0] != plinfo->mod) ||
	    (rxp->rxbuf [1] != MODLOAD) ||
	    (rxp->rxbuf [2] != 2 ) ||
	    (rxp->rxbuf [3] != MODSFLOGOUT) ||
	    (rxp->rxbuf [4] != MODCMDDONE) )
	{
	    plinfo->pmod->no_dataerr++;
	    return (FALSE);
	}
	return (TRUE);
}

/**********
*                                                               	
*	PLC Login                                                 	
*
**********/

/* called from scan_plc and loadinit  */

static int plc_login (LOADINFO *plinfo)
{
	TXLINE		*txp;
	RXLINE		*rxp;

	/* Set up pointers for comm buffers */

	txp = txlinetab [plinfo->line];
	rxp = rxlinetab [plinfo->line];

	/* Format the output message */

	txp->txbuf [0] =  plinfo->mod;	/* Address	        */
	txp->txbuf [1] =  MODLOAD;	/* Function Code	*/
	txp->txbuf [2] =  1;		/* Length	        */
	txp->txbuf [3] =  MODSFLOGIN;	/* Subfunction Code     */

	makecrc (&(txp->txbuf [0]), 4, &(txp->txbuf [4]));

	txp->txmsglen = 6;

	/* Send out the message */

	if (msgxfer (plinfo->line, VARRESP, plinfo->pmod, 0) == XFERBAD)
	{
/* >>> Could not send log in message so mark unit as off line */
	    plinfo->pmod->no_offlines++;
	    plinfo->pmod->holdflag = TRUE;
	    plinfo->pmod->is_responding = FALSE;
	    return (FALSE);
	}

	/* Verify the response */

	if ((rxp->rxbuf [0] != plinfo->mod) ||
	    (rxp->rxbuf [1] != MODLOAD) ||
	    (rxp->rxbuf [2] != 2 ) ||
	    (rxp->rxbuf [3] != MODSFLOGIN) ||
	    ((rxp->rxbuf [4] & 0x80) != 0) )
	{
	    plinfo->pmod->no_dataerr++;
	    return (FALSE);
	}

	return (TRUE);
}

/*                                                          		
*	PLC Start                                                 	
*/
/* called from scan_plc and loadnextblock  */


static int plc_start (LOADINFO *plinfo)
    {
  	TXLINE		*txp;
	RXLINE		*rxp;

	/* Set up pointers for comm buffers */

	txp = txlinetab [plinfo->line];
	rxp = rxlinetab [plinfo->line];

	/* Format the output message */

	txp->txbuf [0] =  plinfo->mod;	/* Address	        */
	txp->txbuf [1] =  MODLOAD;	/* Function Code	*/
	txp->txbuf [2] =  1;		/* Length	        */
	txp->txbuf [3] =  MODSFSTART;	/* Subfunction Code     */

	makecrc (&(txp->txbuf [0]), 4, &(txp->txbuf [4]));

	txp->txmsglen = 6;

	/* Send out the message */

	if (msgxfer (plinfo->line, VARRESP, plinfo->pmod, 0) == XFERBAD)
	{
/* >>> Could not send plc start message so mark unit as off line */
	    plinfo->pmod->no_offlines++;
	    plinfo->pmod->holdflag = TRUE;
	    plinfo->pmod->is_responding = FALSE;
	    return (FALSE);
	}

	/* Verify the response */

	if ((rxp->rxbuf [0] != plinfo->mod) ||
	    ((rxp->rxbuf [1] & 0x7F) != MODLOAD) )
	{
	    plinfo->pmod->no_dataerr++;
	    return (FALSE);
	}

	/* Is it a command done */

	if ((rxp->rxbuf [1] == MODLOAD) &&
	    (rxp->rxbuf [3] == MODSFSTART) &&
	    (rxp->rxbuf [4] == MODCMDDONE) )
	{
	    return (TRUE);
	}

	/* Is it an ACK done */

	if ((rxp->rxbuf [1] == (0x80 + MODLOAD) )&&
	    (rxp->rxbuf [2] == MODCMDSTARTED) )
	{
	    return (pollfordone (plinfo, MODSFSTART, FALSE));
	}
	return (FALSE);
}

/**********
*                                                                	
*	PLC Stop                                         		
*
**********/

/* called from loadinit and scan_plc */

static int plc_stop (LOADINFO *plinfo)
{
	TXLINE		*txp;
	RXLINE		*rxp;

	/* Set up pointers for comm buffers */

	txp = txlinetab [plinfo->line];
	rxp = rxlinetab [plinfo->line];

	/* Format the output message */

	txp->txbuf [0] =  plinfo->mod;	/* Address	        */
	txp->txbuf [1] =  MODLOAD;	/* Function Code	*/
	txp->txbuf [2] =  1;		/* Length	        */
	txp->txbuf [3] =  MODSFSTOP;	/* Subfunction Code     */

	makecrc (&(txp->txbuf [0]), 4, &(txp->txbuf [4]));

	txp->txmsglen = 6;

	/* Send out the message */

	if (msgxfer (plinfo->line, VARRESP, plinfo->pmod, 0) == XFERBAD)
	{
/* >>> Could not send plc stop message so mark unit as off line */
	    plinfo->pmod->no_offlines++;
	    plinfo->pmod->holdflag = TRUE;
	    plinfo->pmod->is_responding = FALSE;
	    return (FALSE);
	}

	/* Verify the response */

	if ((rxp->rxbuf [0] != plinfo->mod) ||
	    ((rxp->rxbuf [1] & 0x7F) != MODLOAD) )
	{
	    plinfo->pmod->no_dataerr++;
	    return (FALSE);
	}

	/* Is it a command done */

	if ((rxp->rxbuf [1] == MODLOAD) &&
	    (rxp->rxbuf [3] == MODSFSTOP) &&
	    (rxp->rxbuf [4] == MODCMDDONE) )
	{
	    return (TRUE);
	}

	/* Is it an ACK done */

	if ( (rxp->rxbuf [1] == (0x80 + MODLOAD)) /*&&
	     (rxp->rxbuf [2] == MODCMDSTARTED) */ )
	{
	    return (pollfordone (plinfo, MODSFSTOP, FALSE));
	}
	return (FALSE);
}

/**********
*    									
* pollfordone                                                     	
*	subfunction                                                 	
*                                                                  	
**********/

/* called from plc_start, plc_stop, and writelogic  */

static int pollfordone (LOADINFO *plinfo, int subfunction, unsigned short geterrcode)
{
	unsigned short	i;
	TXLINE		*txp;
	RXLINE		*rxp;

	/* Set up pointers for comm buffers */

	txp = txlinetab [plinfo->line];
	rxp = rxlinetab [plinfo->line];

	for (i = 0;i < MAXPOLLSFORDONE; i++)
	{

	    /* Format the output message */

	    txp->txbuf [0] =  plinfo->mod;	/* Address	        */
	    txp->txbuf [1] =  MODPOLL;		/* Function Code	*/

	    makecrc (&(txp->txbuf [0]), 2, &(txp->txbuf [2]));

	    txp->txmsglen = 4;

	    /* Send out the message */

	    if (msgxfer (plinfo->line, VARRESP, plinfo->pmod, 0) == XFERBAD)
	    {
/* >>> Could not send poll message so mark unit as off line */
		plinfo->pmod->no_offlines++;
	    	plinfo->pmod->holdflag = TRUE;
		plinfo->pmod->is_responding = FALSE;
		return (FALSE);
	    }

	    /* May be used to fetch error code only */

	    if (geterrcode)
	    {
	    	return (TRUE);
	    }
	    /* Verify the response */

	    if ((rxp->rxbuf [0] == plinfo->mod) &&
		(rxp->rxbuf [1] == MODPOLL) &&
		(rxp->rxbuf [2] == 2 ) &&
		(rxp->rxbuf [3] == subfunction) &&
		(rxp->rxbuf [4] == MODCMDDONE) )
	    {
		return (TRUE);
	    }
	}
	return (FALSE);
}

/**********
* 										
*	PLC Init                                                     		
*
**********/                                                                      	

/* only called from loadnextblock  */

static int plc_init (LOADINFO *plinfo)
{
	TXLINE		*txp;
	RXLINE		*rxp;

	/* Set up pointers for comm buffers */

	txp = txlinetab [plinfo->line];
	rxp = rxlinetab [plinfo->line];

	/* Format the output message */

	txp->txbuf [0] =  plinfo->mod;		/* Address		*/
	txp->txbuf [1] =  MODLOAD;		/* Function Code	*/
	txp->txbuf [2] =  1;			/* Length		*/
	txp->txbuf [3] =  MODSFINITIALIZE;	/* Subfunction Code 	*/

	makecrc (&(txp->txbuf [0]), 4, &(txp->txbuf [4]));

	txp->txmsglen = 6;

	/* Send out the message */

	if (msgxfer (plinfo->line, VARRESP, plinfo->pmod, 0) == XFERBAD)
	{
/* >>> Could not send plc init message so mark unit as off line */
	    plinfo->pmod->no_offlines++;
	    plinfo->pmod->holdflag = TRUE;
	    plinfo->pmod->is_responding = FALSE;
	    return (FALSE);
	}

	/* Verify the response */

	if ((rxp->rxbuf [0] != plinfo->mod) ||
	    (rxp->rxbuf [1] != MODLOAD) ||
	    (rxp->rxbuf [2] != 2 ) ||
	    (rxp->rxbuf [3] != MODSFINITIALIZE) ||
	    (rxp->rxbuf [4] != MODCMDDONE) )
	{
	    plinfo->pmod->no_dataerr++;
	    return (FALSE);
	}
	return (TRUE);
}

/**********
*                    								
*	Get Dim Awareness
*		-1 => error    0 => not dim   1=> dim                       	
*
**********/

/* only called from loadinit  */

static int plc_getdim (LOADINFO *plinfo)
{
	TXLINE		*txp;
	RXLINE		*rxp;

	/* Set up pointers for comm buffers */

	txp = txlinetab [plinfo->line];
	rxp = rxlinetab [plinfo->line];

	/* Format the output message */

	txp->txbuf [0] =  plinfo->mod;	/* Address	        */
	txp->txbuf [1] =  MODLOAD;	/* Function Code	*/
	txp->txbuf [2] =  1;		/* Length	        */
	txp->txbuf [3] =  MODSFCHKDIM;	/* Subfunction Code     */

	makecrc (&(txp->txbuf [0]), 4, &(txp->txbuf [4]));

	txp->txmsglen = 6;

	/* Send out the message */

	if (msgxfer (plinfo->line, VARRESP, plinfo->pmod, 0) == XFERBAD)
	{
/* >>> Could not send get dim message so mark unit as off line */
	    plinfo->pmod->no_offlines++;
	    plinfo->pmod->holdflag = TRUE;
	    plinfo->pmod->is_responding = FALSE;
	    return (-1);
	}

	/* Verify the response */

	if ((rxp->rxbuf [0] != plinfo->mod) ||
	    (rxp->rxbuf [1] != MODLOAD) ||
	    (rxp->rxbuf [3] != MODSFCHKDIM) )
	{
	    plinfo->pmod->no_dataerr++;
	    return (-1);
	}

	if  (rxp->rxbuf [4] == 0)
	{
	    return(0); /* Not in Dim Awareness */
	}
	return (1); /* Am in Dim Awareness */
}

/**********
*          									
*	Set Dim Awareness                                                   	
*
**********/

/* only called from loadinit  */

static int plc_setdim (LOADINFO *plinfo)
{
	TXLINE		*txp;
	RXLINE		*rxp;

	/* Set up pointers for comm buffers */

	txp = txlinetab [plinfo->line];
	rxp = rxlinetab [plinfo->line];

	/* Format the output message */

	txp->txbuf [0] =  plinfo->mod;	/* Address	        */
	txp->txbuf [1] =  MODLOAD;	/* Function Code	*/
	txp->txbuf [2] =  1;		/* Length	        */
	txp->txbuf [3] =  MODSFENTERDIM;	/* Subfunction Code     */

	makecrc (&(txp->txbuf [0]), 4, &(txp->txbuf [4]));

	txp->txmsglen = 6;

	/* Send out the message */

	if (msgxfer (plinfo->line, VARRESP, plinfo->pmod, 0) == XFERBAD)
	{
/* >>> Could not send set dim message so mark unit as off line */
	    plinfo->pmod->no_offlines++;
	    plinfo->pmod->holdflag = TRUE;
	    plinfo->pmod->is_responding = FALSE;
	    return (FALSE);
	}

	/* Verify the response */

	if ((rxp->rxbuf [0] != plinfo->mod) ||
	    (rxp->rxbuf [1] != MODLOAD) ||
	    (rxp->rxbuf [3] != MODSFENTERDIM) ||
	    (rxp->rxbuf [4] != MODCMDDONE))
	{
	    plinfo->pmod->no_dataerr++;
	    return (FALSE);
	}
	return (TRUE);
}

/**********
*                                                			
*	Set Run Mode                                                	
*
**********/

/* called from loadnextblock and terminateload  */

static int setrunmode (LOADINFO *plinfo, short value)
{
	return (writemem (
		plinfo,
		0,
	 	RUNLOADSWITCHADDR,
	 	2,
		(char*) &value));  /* &value)); ### */
			   /* (wep: cast required when prototyping was performed. */
}

/**********
*                                            				
*	Get Run Mode                                               	
*
**********/

/* only called from loadnextblock  */

static int getrunmode (LOADINFO *plinfo, short *pvalue)
{
	return (readmem(
		plinfo,
		0,
		RUNLOADSWITCHADDR,
		2,
		(char*) pvalue)); /* pvalue)); ###*/
}

/**********
*                                                                 	
*	Get Machine State                                          	
*
**********/

/* only called from loadinit  */

static int get_plc_state (LOADINFO *plinfo, short *pvalue)
{
	return (readmem (
		plinfo,
		0,
		MACHINESTATEADDR,
		2,
		(char*) pvalue));  /* pvalue)); ###*/
}

/**********
*                                                                	
*	Read Memory (Page,Address,datalength,buffer)               	
*
**********/

/* called from getrunmode and get_plc_state */

static int readmem (LOADINFO *plinfo, unsigned short page, unsigned short address,
	     unsigned short datalen, char *databuf)
{
	TXLINE		*txp;
	RXLINE		*rxp;
	unsigned short	i;

	/* Set up pointers for comm buffers */

	txp = txlinetab [plinfo->line];
	rxp = rxlinetab [plinfo->line];

	/* Format the output message */

	txp->txbuf [0] =  plinfo->mod;		/* Address		*/
	txp->txbuf [1] =  MODLOAD;		/* Function Code	*/
	txp->txbuf [2] =  5;			/* Length		*/
	txp->txbuf [3] =  MODSFREADMEM;		/* Subfunction Code 	*/
	txp->txbuf [4] =  (datalen / 2);	/* Count of elements 	*/
	txp->txbuf [5] =  page;			/* Memory Page 		*/
	txp->txbuf [6] =  (address >> 8);	/* Address High 	*/
	txp->txbuf [7] =  (address & 0xFF);	/* Address Low 		*/

	makecrc (&(txp->txbuf [0]), 8, &(txp->txbuf [8]));

	txp->txmsglen = 10;

	/* Send out the message */

	if (msgxfer (plinfo->line, VARRESP, plinfo->pmod, 0) == XFERBAD)
	{
/* >>> Could not send read mem message so mark unit as off line */
	    plinfo->pmod->no_offlines++;
	    plinfo->pmod->holdflag = TRUE;
	    plinfo->pmod->is_responding = FALSE;
	    return (FALSE);
	}

	/* Verify the response */

	if ((rxp->rxbuf [0] != plinfo->mod) ||
	    (rxp->rxbuf [1] != MODLOAD) ||
	    (rxp->rxbuf [2] != datalen + 5) ||
	    (rxp->rxbuf [3] != MODSFREADMEM) ||
	    (rxp->rxbuf [4] != (datalen / 2)) ||
	    (rxp->rxbuf [5] != page) ||
	    (rxp->rxbuf [6] != (unsigned char) (address >> 8)) ||
	    (rxp->rxbuf [7] != (unsigned char) (address & 0xFF))  )
	{
	    plinfo->pmod->no_dataerr++;
	    return (FALSE);
	}
	/* move in the data */
	for (i = 0;(i < datalen); i++)
	{
	    databuf [i] = rxp->rxbuf [8 + i];
	}
	return (TRUE);
}

/**********
*                                                            		
*	Write Memory (Page,Address,datalength,buffer)              	
*
**********/

/* called from loadmemory and setrunmode */

static int writemem (LOADINFO *plinfo, unsigned short page, unsigned short address,
	     unsigned short datalen, char *databuf)
{
	unsigned short	i;
	TXLINE		*txp;
	RXLINE		*rxp;

	/* Set up pointers for comm buffers */

	txp = txlinetab [plinfo->line];
	rxp = rxlinetab [plinfo->line];

	/* Format the output message */

	txp->txbuf [0] =  plinfo->mod;	/* Address	        */
	txp->txbuf [1] =  MODLOAD;	/* Function Code	*/
	txp->txbuf [2] =  5 + datalen;	/* Length	        */
	txp->txbuf [3] =  MODSFWRITEMEM;	/* Subfunction Code     */
	txp->txbuf [4] =  (datalen  /2);	/* Count of elements    */
	txp->txbuf [5] =  page;		/* Memory Page */
	txp->txbuf [6] =  (address >> 8);	/* Address High */
	txp->txbuf [7] =  (address & 0xFF);	/* Address Low  */

	/* move in the data */
	for (i = 0;(i < datalen); i++)
	{
	    txp->txbuf [8 + i] = databuf [i];
	}
	makecrc (&(txp->txbuf [0]), 8 + datalen, &(txp->txbuf [8 + datalen]));

	txp->txmsglen = 10 + datalen;

	/* Send out the message */

	if (msgxfer (plinfo->line, VARRESP, plinfo->pmod, 0) == XFERBAD)
	{
/* >>> Could not send write mem message so mark unit as off line */
	    plinfo->pmod->no_offlines++;
	    plinfo->pmod->holdflag = TRUE;
	    plinfo->pmod->is_responding = FALSE;
	    return (FALSE);
	}

	/* Verify the response is command done*/

	if ((rxp->rxbuf [0] != plinfo->mod) ||
	    (rxp->rxbuf [1] != MODLOAD) ||
	    (rxp->rxbuf [2] != 2) ||
	    (rxp->rxbuf [3] != MODSFWRITEMEM) ||
	    (rxp->rxbuf [4] != MODCMDDONE)  )
	{
	    plinfo->pmod->no_dataerr++;
	    return (FALSE);
	}
	return (TRUE);
}

/**********
*                                                                	
*	Write Logic (Page,Address,datalength,buffer)               	
*
**********/

/* only called from loadlogic */

static int writelogic (LOADINFO *plinfo, unsigned short page, unsigned short address,
	     unsigned short datalen, char *databuf)
{
	unsigned short	i;
	TXLINE		*txp;
	RXLINE		*rxp;

	/* Set up pointers for comm buffers */

	txp = txlinetab [plinfo->line];
	rxp = rxlinetab [plinfo->line];

	/* Format the output message */

	txp->txbuf [0] =  plinfo->mod;		/* Address		*/
	txp->txbuf [1] =  MODLOAD;		/* Function Code	*/
	txp->txbuf [2] =  5 + datalen;		/* Length		*/
	txp->txbuf [3] =  MODSFWRITENODE;	/* Subfunction Code 	*/
	txp->txbuf [4] =  (datalen / 3);		/* Count of elements 	*/
	txp->txbuf [5] =  page;			/* Memory Page 		*/
	txp->txbuf [6] =  (address >> 8);	/* Address High 	*/
	txp->txbuf [7] =  (address & 0xFF);	/* Address Low 		*/

	/* move in the data */
	for (i = 0; (i < datalen); i++)
	{
	    txp->txbuf [8 + i] = databuf [i];
	}
	makecrc (&(txp->txbuf [0]), 8 + datalen, &(txp->txbuf [8 + datalen]));

	txp->txmsglen = 10 + datalen;

	/* Send out the message */

	if (msgxfer (plinfo->line, VARRESP, plinfo->pmod, 0) == XFERBAD)
	{
/* >>> Could not send write logic message so mark unit as off line */
	    plinfo->pmod->no_offlines++;
	    plinfo->pmod->holdflag = TRUE;
	    plinfo->pmod->is_responding = FALSE;
	    return (FALSE);
	}

	/* Verify the response is command done*/

	if ((rxp->rxbuf [0] == plinfo->mod) &&
	    (rxp->rxbuf [1] == MODLOAD) &&
	    (rxp->rxbuf [2] == 2) &&
	    (rxp->rxbuf [3] == MODSFWRITENODE) &&
	    (rxp->rxbuf [4] == MODCMDDONE)  )
	{
	    return (TRUE);
	}

	/* Is it an ACK done */

	if ((rxp->rxbuf [1] == (0x80 + MODLOAD)) &&
	    (rxp->rxbuf [2] == MODCMDSTARTED))
	{
	    return (pollfordone (plinfo, MODSFWRITENODE, FALSE));
	}
	pollfordone (plinfo, MODSFWRITENODE, TRUE);
	return (FALSE);
}

/**********
*                                                            		
*	Write Ascii (Page,Address,datalength,buffer)              	
*
**********/

/* only called from loadmemory */

static int writeascii (LOADINFO *plinfo, unsigned short page, unsigned short address,
	     unsigned short datalen, char *databuf)
{
	unsigned short	i;
	TXLINE		*txp;
	RXLINE		*rxp;

	/* Set up pointers for comm buffers */

	txp = txlinetab [plinfo->line];
	rxp = rxlinetab [plinfo->line];

	/* Format the output message */

	txp->txbuf [0] =  plinfo->mod;		/* Address		*/
	txp->txbuf [1] =  MODLOAD;		/* Function Code	*/
	txp->txbuf [2] =  5 + datalen;		/* Length		*/
	txp->txbuf [3] =  MODSFWRITEASCII;	/* Subfunction Code 	*/
	txp->txbuf [4] =  (datalen / 2);		/* Count of elements 	*/
	txp->txbuf [5] =  page;			/* Memory Page 		*/
	txp->txbuf [6] =  (address >> 8);	/* Address High 	*/
	txp->txbuf [7] =  (address & 0xFF);	/* Address Low 		*/

	/* move in the data */
	for (i = 0; (i < datalen); i++)
	{
	    txp->txbuf [8 + i] = databuf [i];
	}
	makecrc (&(txp->txbuf [0]), 8 + datalen, &(txp->txbuf [8 + datalen]));

	txp->txmsglen = 10 + datalen;

	/* Send out the message */

	if (msgxfer (plinfo->line, VARRESP, plinfo->pmod, 0) == XFERBAD)
	{
/* >>> Could not send write ascii message so mark unit as off line */
	    plinfo->pmod->no_offlines++;
	    plinfo->pmod->holdflag = TRUE;
	    plinfo->pmod->is_responding = FALSE;
	    return (FALSE);
	}

	/* Verify the response is command done*/

	if ((rxp->rxbuf [0] != plinfo->mod) ||
	    (rxp->rxbuf [1] != MODLOAD) ||
	    (rxp->rxbuf [2] != 2) ||
	    (rxp->rxbuf [3] != MODSFWRITEASCII) ||
	    (rxp->rxbuf [4] != MODCMDDONE)  )
	{
	    plinfo->pmod->no_dataerr++;
	    return (FALSE);
	}
	return (TRUE);
}

/**********
*                                                                           	
* makecrc 									
*                                                                        	
**********/

/* called from msgxfer, modloopback, readcoils, readdis, readais, readregs, setcoil, setreg,  */
/*             plc_logout, plc_login, plc_start, plc_stop, pollfordone, plc_init, plc_getdim, */
/*             plc_setdim, readmem, writemem, writelogic, and writeascii.                     */

static	void makecrc (unsigned char *chkbuf, int len,char unsigned *retbuf)

{
	unsigned short	checksum;
	unsigned int    i, j;
	unsigned short  flag;
	unsigned char   tempch;

	checksum = 0x0ffff;
	for (j = 0; j < len; j++)
	{
	    tempch = chkbuf [j];
	    checksum ^= tempch;
	    for (i = 0; i < 8; i++)
	    {
		flag = checksum & 1;
		checksum >>= 1;
		if (flag)
		{
		    checksum ^= (unsigned short) 0x0a001;
		}
	    }
	}
	retbuf [1] = ((checksum >> 8) & 0xff);
	retbuf [0] = (checksum & 0xff);
}

/**********
*
*      sv85dumpmsgcnts
*
*      The purpose of this routine is to dump the line level statistics for this driver. 
*
**********/

void sv85dumpmsgcnts (unsigned short line)
{
        unsigned long	temp, t, x, y, z;

        temp = tickGet ();
        t = (temp - linestats [line].startingtick);

 	x = linestats [line].TXintcount;
	y = linestats [line].RXintcount;
	z = linestats [line].msgxfercount;

        if (z == 0)
        {
            printf (" ******  No Messages Transferred ****** \n");
            return;
        }

 	printf ("\n");
	printf ("Measurement Interval: clock ticks = %ld, seconds = %ld.%ld \n", t, t/60, (t%60)/6);
 	printf ("\n");
 	printf ("TX Interrupt Count    = %ld   Avg TX cnts/sec        = %ld \n", x, (60*x)/t);
 	printf ("RX Interrupt Count    = %ld   Avg RX cnts/sec        = %ld \n", y, (60*y)/t);
 	printf ("                                     Avg RX cnts/Msgxfer    = %ld.\n", x/z);
 	printf ("                                     Avg TX cnts/Msgxfer    = %ld.\n", y/z);
 	printf ("\n");
 	printf ("RX+TX Interrupt Count = %ld   Avg RX+TX cnts/sec     = %ld \n",(x+y),(60*(x+y))/t);
 	printf ("                                     Avg RX+TX cnts/Msgxfer = %ld \n",(x+y)/z);
 	printf ("\n");
	printf ("Msgxfer Count         = %ld   Avg Msgxfers/sec       = %ld \n", z,(60*z)/t);
	printf ("                                     Avg msec/Msgxfer       = %ld \n",(t*1000)/(z*60));
 	printf ("\n");
        return;
}

/**********
*
*      sv85clearmsgcnts
*
*      The purpose of this routine is to dump and clear the line level statistics
*      for this driver.
*
**********/

void sv85clearmsgcnts (unsigned short line)
{
        sv85dumpmsgcnts (line);

	linestats [line].TXintcount = 0;
	linestats [line].RXintcount = 0;
	linestats [line].msgxfercount = 0;

        linestats [line].startingtick = tickGet ();
	printf (" ******  All Counts Cleared  ****** \n");
 	printf ("\n");
        return;
}

/**********
*
*      sv85dumpplcstats
*
*      The purpose of this routine is to dump the plc level statistics for this driver.  
*
**********/

void sv85dumpplcstats (unsigned short line, unsigned short maddr)
{
	MODPLC		*pmod;

	pmod = plcs [line] [maddr];
	if (!pmod)
	{
	    logMsg ("Line/ModbusAddress Not Defined \n", 0, 0, 0, 0, 0, 0);
	    return;
	}

	printf (" Was requested = %d \n", pmod->was_requested);
	printf (" Is responding = %d \n", pmod->is_responding);
	printf (" Prog Load State = %d \n", pmod->prog_load_stat);
	printf (" Prog Load Page = %d \n",  pmod->prog_load_page);
	printf (" Prog Number = %d \n", pmod->prog_load_number);
	printf (" # Output messages = %d \n", pmod->no_dataout_msgs & 0x7fff);
	printf (" # Data in messages = %d \n", pmod->no_datain_msgs & 0x7fff);
	printf (" #No Rx Response = %d \n", pmod->no_rxnoresp & 0x7fff);
	printf (" #No Tx Responses = %d \n", pmod->no_txnoresp & 0x7fff);
	printf (" #Data Errors = %d \n", pmod->no_dataerr & 0x7fff);
	printf (" #Retries = %d \n", pmod->no_retries & 0x7fff);
	printf (" #Rx Hw Errors = %d \n", pmod->no_rxhwerrs & 0x7fff);
	printf (" #Tx Hw Errors = %d \n", pmod->no_txhwerrs & 0x7fff);
	printf (" #Crc Errors = %d \n", pmod->no_crcerrs & 0x7fff);
	printf (" #Off Lines = %d \n", pmod->no_offlines & 0x7fff);
	printf (" Reset Statistics = %d \n", pmod->cmd_reset_stat);
	printf (" Force Offline = %d \n", pmod->cmd_force_offline);
	printf (" #Failed Outputs = %d \n", pmod->badoutcnt & 0x7fff);


	printf (" Resource Responding = %d \n", resresponse[line]);
	printf (" Hold flag  = %d \n", pmod->holdflag );
        return;
}

/**********
*
*      sv85clearplcstats
*
*      The purpose of this routine is to clear the plc level statistics for this driver. 
*
**********/

void sv85clearplcstats (unsigned short line, unsigned short maddr)
{
	MODPLC		*pmod;

	pmod = plcs [line] [maddr];
	if (!pmod)
	{
	    logMsg ("Line/ModbusAddress Not Defined \n", 0, 0, 0, 0, 0, 0);
	    return;
	}

	initstats (pmod); /* use existing procedure */

        return;
    }

/**********
*
*      dump_mod_table                                                       	
*                                                                           	
*      The purpose of this routine is twofold:                                 
*         1) dump the plc level statistics for the modbus driver.              
*         2) dump the data for the specified modbus plc.                       
*                                                                              
**********/

/* not called from within this driver */

void dump_sv85_table (unsigned short line, unsigned short maddr)
{
	MODPLC		*pmod;
	unsigned	short	i, n, j; 

	MODCOILS	*coilptr;
	MODDIS		*diptr;
	REGISTERS	*regptr;
	MODAIS		*aiptr;
	unsigned	short addr, linecnt;
	unsigned	char	buffer [200];


	pmod = plcs [line] [maddr];
	if (!pmod)
	{
	    logMsg("Line/ModbusAddress Not Defined \n", 0, 0, 0, 0, 0, 0);
	    return;
	}

        /* previously integral code moved to seperate procedure for debugging convenience */

        sv85dumpplcstats (line, maddr);

	linecnt = 1;
	printcoilhdr (buffer, linecnt);
	addr = 1;
	for (i = 0; i < (MAXCOILNO / 8); i += 4)
	{
	    printf("  %4d    ", addr);
	    for (n = 0; n < 4; n++)
	    {		
		coilptr = &pmod->coils[i+n];
		printf("%1x,  %1x,         %lx,\t",
                          coilptr->was_requested,
                          coilptr->data_valid,
                          coilptr->last_read);
	    }
	    printf ("\n");
	    linecnt += 1;
	    if ( (linecnt >= 17) && ((i + 4)  < (MAXCOILNO /8)) )
	    {
		linecnt = 1;
		printcoilhdr(buffer, linecnt);
	    }
	    addr += 32;		
	}

	linecnt = 1;
	printdihdr (buffer, linecnt);
	addr = 1;
	for (i = 0; i < (MAXDINO / 8); i += 4)
	{
	    printf ("  %4d    ", addr);
	    for (n = 0; n < 4; n++)
	    {		
 		diptr = &pmod->dis[i+n];
		printf("%1x,  %1x,         %lx,\t",
                           diptr->was_requested,
                           diptr->data_valid,
                           diptr->last_read);
	    }
	    printf ("\n");
	    linecnt += 1;
	    if ((linecnt >= 17) && ((i + 4)  < (MAXDINO / 8)) )
	    {
		linecnt = 1;
		printdihdr (buffer, linecnt);
	    }
	    addr += 32;		
	}


	linecnt = 1;
	printaihdr(buffer, linecnt);
	addr = 1;
	for (i = 0; i < MAXAINO; i += 8)
	{

	    printf ("  %4d\t", addr);
				   
	    for (n = 0; n < 2; n++ )
	    {		
		aiptr = &pmod->ais [(i / 4) + n];
		printf ("\t%1x, %1x,\t", aiptr->was_requested, aiptr->data_valid);
		for (j = 0; j < 4; j++)
		{
		    printf("%6d,\t",  aiptr->aivals[j]);
		}
	    }
	    printf("\n");
	    linecnt += 1;
	    if ((linecnt >= 17) && ((i + 8)  < NUMMODAIS) )
	    {
		linecnt = 1;
		printaihdr (buffer, linecnt);
	    }
	    addr += 8;		
	}

	linecnt = 1;
	printreghdr (buffer, linecnt);
	addr = 1;
	for (i = 0; i < MAXREGNO; i += 8)
	{
	    printf("  %4d\t", addr);

	    for (n = 0; n < 2; n++)
	    {		
		regptr = &pmod->regs[(i/4)+n];
		printf("\t%1x, %1x,\t", regptr->was_requested, regptr->data_valid);
		for (j = 0; j < 4; j++)
		{
		    printf("%6d,\t",regptr->ioregs[j]);
		}
	    }
	    printf ("\n");
	    linecnt += 1;
	    if ((linecnt >= 17) && ((i + 8)  < NUMREGISTERS) )
	    {
		linecnt = 1;
		printreghdr (buffer, linecnt);
	    }
	    addr += 8;		
	}
}

/**********
*
* only called from dump_mod_table 
*
**********/

static void printcoilhdr (unsigned char *buffer, unsigned short cnt)
{	
	taskDelay (120);
	printf ("\n");
	sprintf (buffer,"                                                    PLC REPORT         ");
	printf ( "%s\n", buffer);
	cnt += 1;
	printf ("\n");
	cnt += 1;
        sprintf (buffer,"                                                   MODBUS COILS        ");
	printf ( "%s\n", buffer);
	cnt += 1;
        sprintf (buffer,
               "          R   V      Value\tR   V      Value\tR   V      Value\tR   V      Value");
 	printf ( "%s\n", buffer);
 	cnt += 1;
}
/*
* only called from dump_mod_table
*/


static void printdihdr (unsigned char *buffer, unsigned short cnt)
{
	taskDelay (120);
	printf ("\n");
	sprintf (buffer,"                                                    PLC REPORT         ");
	printf ( "%s\n", buffer);
	cnt += 1;
	printf ("\n");
	cnt += 1;
        sprintf (buffer,"                                              MODBUS DIGITAL INPUTS        ");
	printf ( "%s\n", buffer);
	cnt += 1;
        sprintf (buffer,
                "          R   V      Value\tR   V      Value\tR   V      Value\tR   V      Value");
	printf ( "%s\n", buffer);
 	cnt += 1;
}
/*
* only called from dump_mod_table
*/


static void printaihdr(unsigned char *buffer, unsigned short cnt)
{
	taskDelay (120);
	printf ("\n");
	sprintf (buffer,"                                               PLC REPORT         ");
	printf ( "%s\n", buffer);
	cnt += 1;
	printf ("\n");
	cnt += 1;
       sprintf (buffer,"                                            MODBUS ANALOG INPUTS        ");
	printf ( "%s\n", buffer);
	cnt += 1;
       sprintf (buffer,"\t\tR  V\tValue\tValue\tValue\tValue\t\tR  V\tValue\tValue\tValue\tValue");
 	printf ( "%s\n", buffer);
 	cnt += 1;
}

/*
* only called from dump_mod_table
*/


static void printreghdr (unsigned char *buffer, unsigned short cnt)
{
	taskDelay (120);
	printf ("\n");
	sprintf (buffer,"                                               PLC REPORT         ");
	printf ( "%s\n", buffer);
	cnt += 1;
	printf ("\n");
	cnt += 1;
       sprintf (buffer,"                                             MODBUS REGISTERS        ");
	printf ( "%s\n", buffer);
	cnt += 1;
       sprintf (buffer,"\t\tR  V\tValue\tValue\tValue\tValue\t\tR  V\tValue\tValue\tValue\tValue");
 	printf ( "%s\n", buffer);
 	cnt += 1;

}

/**********                                                               		
*                                                                	 	
*	The purpose of this routine is to display the contents            	
*	of the static variables contained within this driver.             	
*                                                                       	
**********/

/* this routine not called from within this driver */


void sv85showlineinfo ()
{
	short	i;
	char	buffer [132];
	char	*resrsp;
	char	*lineact;


        printf ("                                 Modbus Line Information Table          \n");
        printf ("\n");
        printf (" Line        Pointer to        Resource      Resource");
        printf ("       Line        Delay Times:       Activity\n");
        printf ("Number      MODPLC Table         Name        Response");
        printf ("     Activity      IM  .TX  TX.         Time\n");
	printf ("\n");

	for (i = 0; i < MODLINES; i++)
	{
	    if(plcs [i] == NULL)
	    {
		continue;
	    }
	    if (resresponse [i] == 0)
	    {
		resrsp = "FALSE";
	    }
	    else
	    {
		resrsp = "TRUE";
	    }
	    if (line_activity [i] == 0)
	    {
		lineact = "FALSE";
	    }
	    else
	    {
		lineact = "TRUE";
	    }
	    sprintf (buffer,
                       "  %2d         %4x   %16s        %5s        %5s        %2d   %2d   %2d        %4ld\n",
			i, (int) plcs[i], (char *) &resnames[i],
			&resrsp[0],
			&lineact[0],
			intermsgdelay[i],
			pretxdelay[i],
			posttxdelay[i],
			activity_time[i]);
	    printf ( "%s\n", buffer);
	}
	return;
}

/****************************************************************/
/*								*/
/*		SV85 I/O REDUNDANT UPDATE TASK			*/
/*								*/
/****************************************************************/

extern	char	consoleName [];
int		redtaskhold = 0;

void redSv85Task (unsigned short line)
{
	int			tycoFd;
	char			heartBeat;
	char			tyconame [16];
	int			tycoport;
	struct sockaddr_in	addr;
	long			ourInet;
	int			intSize;
	int			optVal = 1;
	fd_set			redFds;
	struct timeval		pollTimer = {0, 0};
	int			lastPeriodic = 0;

	/*
	Start off by grabbing the tyCo port for redundant heartbeats.
	*/
	while (redtaskhold)
	{
	    taskDelay (10);
	}

	strcpy (tyconame, consoleName);
	tyconame [strlen (tyconame) - 1] += 1;
	if (tyconame [strlen (tyconame) - 2] == '/')
	{
	    tycoport = atoi (& (tyconame [strlen (tyconame) - 1]));
	}
	else
	{
	    tycoport = atoi (& (tyconame [strlen (tyconame) - 2]));
	}
/*
*	if (tyCoDevCreate (tyconame, tycoport, 512, 512) == ERROR)
*	{
*	    logMsg ("Can't create device for serial line redundancy %s %d\n",
*			tyconame, tycoport);
*	    exit (1);
*	}
*/
	if ((tycoFd = open (tyconame, O_RDWR, 0777)) == ERROR)
	{
	    perror ("Can't open serial line for redundancy");
	    exit (1);
	}

	/*
	Find our local Inet address.
	*/

	local_addr (0, &addr);
	ourInet = ntohl (addr.sin_addr.s_addr);

	/*
	This is our main loop.
	*/

	for (;;)
	{
	    /*
	    Wait for a "kick" to get going.
	    */

/*	    semTake (redundSem [line], WAIT_FOREVER);	*/

	    taskDelay (15);

	    /*
	    See if we don't have a connection to our redundant partner.
	    */

	    if (redFd [line] == 0)
	    {
		/*
		Is the partner Inet address known?
		*/

		if (partnerInet [line])
		{
		    /*
		    It is known. Do we make the connection or does our partner?
		    */

		    if (ourInet > partnerInet [line])
		    {
			/*
			We do. Create a socket and try to connect it.
			*/

			redFd [line] = socket (PF_INET, SOCK_STREAM, 0);
			addr.sin_family = AF_INET;
			addr.sin_addr.s_addr = htonl (partnerInet [line]);
			addr.sin_port = htons ((MOD_PORT + line));
			if (connect (redFd [line], (struct sockaddr *) &addr, sizeof (addr)) == ERROR)
			{
			    /*
			    Can't connect. We'll try again later.
			    */

			    close (redFd [line]);
			    redFd [line] = 0;
			}
			else
			{
			    /*
			    Got it. Get current buffer sizes and set up for non-blocking I/O.
			    */
			    redRxBuffmax [line] = 0;
			    tcpRxBuffBytes [line] = 0;
			    lastSeqTx [line] = 0;
			    lastPeriodic = 0;
			    intSize = sizeof (redTxBuffsize [line]);
			    getsockopt (redFd [line], SOL_SOCKET, SO_SNDBUF,
					(char *)&(redTxBuffsize [line]), &intSize);
			    intSize = sizeof (redRxBuffsize [line]);
			    getsockopt (redFd [line], SOL_SOCKET, SO_SNDBUF,
					(char *) &(redRxBuffsize [line]),
					&intSize);
			    ioctl (redFd [line], FIONBIO, (int) &optVal);
			}
		    }
		    else
		    {
			/*
			Partner connects. See if he's trying to connect.
			*/

			FD_ZERO (&redFds);
			FD_SET (tcpFd [line], &redFds);
			if (select (tcpFd [line] + 1, &redFds, NULL, NULL,
				&pollTimer) > 0)
			{
			    /*
			    He is. Set him up and condition the socket for non-blocking I/O.
			    */

			    intSize = sizeof (addr);
			    redFd [line] = accept (tcpFd [line], (struct sockaddr *) &addr, &intSize);
			    redRxBuffmax [line] = 0;
			    tcpRxBuffBytes [line] = 0;
			    lastSeqTx [line] = 0;
			    lastPeriodic = 0;
			    intSize = sizeof (redTxBuffsize [line]);
			    getsockopt (redFd [line], SOL_SOCKET, SO_SNDBUF,
					(char *) &(redTxBuffsize [line]), &intSize);
			    intSize = sizeof (redRxBuffsize [line]);
			    getsockopt (redFd [line], SOL_SOCKET, SO_SNDBUF,
					(char *) &(redRxBuffsize [line]), &intSize);
			    ioctl (redFd [line], FIONBIO, (int) &optVal);
			}
		    }
		}
	    }

	    /*
	    Determine which byte to send out the async port as a heartbeat.
	    */

	    if (prestable [line]->caAnswer)
	    {
		heartBeat = 'M';
	    }
	    else
	    {
		heartBeat = 'S';
	    }
	    write (tycoFd, &heartBeat, sizeof (heartBeat));

	    /*
	    See if there's a byte to read on the serial port.
	    */

	    heartBeat = 0;
	    FD_ZERO (&redFds);
	    FD_SET (tycoFd, &redFds);
	    while (select (tycoFd + 1, &redFds, NULL, NULL, &pollTimer) > 0)
	    {
		read (tycoFd, &heartBeat, sizeof (heartBeat));
	        lastStrobeRxTime [line] = tickGet ();
	    }
	    lastStrobeRx [line] = heartBeat;
	
	    /*
	    Remaining processing requires a connection with our redundant partner.
	    If there is no such connection, then there is nothing to do!
	    */
	    if (redFd [line])
	    {
		short		msgType;
		MOD_UPDT_MSG	updtMsg;

		FD_ZERO (&redFds);
		FD_SET (redFd [line], &redFds);
		sockBytes [line] = 0;

		/*
		Check the socket for any inputs from the other side and process all of them
		*/

		while (select (redFd [line] + 1, &redFds, NULL, NULL, &pollTimer) > 0)
		{
		    /*
		    Check to see if we need to expand the receive buffer size
		    */

		    if (!(redRxBuffmax [line]))
		    {
			ioctl (redFd [line], FIONREAD, (int) &intSize);
			if (intSize == redRxBuffsize [line])
			{
			    redRxBuffsize [line] += 1024;
			    if (setsockopt (redFd [line], SOL_SOCKET, SO_RCVBUF,
				(char *) &(redRxBuffsize [line]),
				sizeof (redRxBuffsize [line])) == ERROR)
			    {
				redRxBuffmax [line] = 1;
			    }
			}
		    }

		    /*
		    Get a big chunk into our buffer. Start by moving things down if there
		    is already stuff in the buffer and it's not at the beginning.
		    */

		    if (tcpRxBuffBytes [line] && tcpRxBuffIdx [line])
		    {
			memmove (tcpRxBuff [line],
					tcpRxBuff [line] + tcpRxBuffIdx [line],
			    		tcpRxBuffBytes [line]);
		    }
		    intSize = recv (redFd [line],
				tcpRxBuff [line] + tcpRxBuffBytes [line],
				sizeof (tcpRxBuff [line]) - tcpRxBuffBytes [line],
				0);
		    if (intSize <= 0)
		    {
			/*
			Error. Close the socket and bail out of here.
			*/

			close (redFd [line]);
			redFd [line] = 0;
			partnerInet [line] = 0;
			break;
		    }
		    tcpRxBuffBytes [line] += intSize;
		    tcpRxBuffIdx [line] = 0;

		    /*
		    Process all of the messages that we can.
		    */

		    while (haveCompleteSv85Msg (line))
		    {
			MOD_ACK_MSG		ackMsg;

			/*
			Get the message type and handle accordingly.
			*/

			getSv85Bytes ((char *) &msgType, sizeof (msgType), line);
			if (msgType == MOD_ACK)
			{
			    /*
			    Acknowledgment for a redundant update. Read the rest of
			    the message, post the sequence number and the current
			    vxWorks tick count so that the driver can make redundancy
			    role decisions. However, if we're slave, ignore the message.
			    */

			    getSv85Bytes ((char *) &ackMsg, sizeof (ackMsg), line);
			    if (!(prestable [line]->caAnswer))
			    {
				continue;
			    }
			    lastSeqRx [line] = ackMsg.ackSeqNo;
			    lastSeqRxTime [line] = tickGet();
			}
			else if (msgType == MOD_REDUPDT)
			{
			    /*
			    A redundant update. Read the rest of the message. If we are
			    slave, ack the message and then post the data in the shared
			    area. Ignore the message if we're master.
			    */

			    getSv85Bytes ((char *) &updtMsg, sizeof (updtMsg),
					line);
			    if (prestable [line]->caAnswer)
			    {
				continue;
			    }
			    ackMsg.ackSeqNo = updtMsg.ackSeqNo;
			    msgType = MOD_ACK;
			    putSv85Bytes (msgType, (char *) &ackMsg,
					sizeof (ackMsg), line);
			}
		    }
		}

		/*
		If it was more than one minute since our last periodic report, create one now
		*/

		if ((prestable [line]->caAnswer) && ((tickGet() - lastPeriodic) > 3600))
		{
/*		    int	index;

		    msgType = MOD_REDUPDT;
		    for (index = 0; index < NUM_ICM_PTS; index++)
		    {
		    }
*/		    lastPeriodic = tickGet();
		}
		else if (prestable [line] -> caAnswer)
		{
/*		    int		index;	*/

	            /*
	            We are master. Traverse the point image looking for the "needs redundant
	            update" bit set. Create an update message for each one of them.
	            */

/*		    msgType = MOD_REDUPDT;
		    for (index = 0; index < NUM_ICM_PTS; index++)
		    {
		    }
*/		}

		/*
		Flush the buffer if there's anything left in it
		*/

		sv85Flush (line);
	    }
	}
}

/****************************************************************/
/*								*/
/*	FLUSH THE TCP/IP BUFFER TO OUR REDUNDANT PARTNER	*/
/*								*/
/****************************************************************/

int sv85Flush (unsigned int line)
{
	int		bytesSent;

	if (redFd [line] == 0)
	{
	    return (0);
	}
	bytesSent = send (redFd [line], sockBuff [line], sockBytes [line], 0);
	if (bytesSent == ERROR)
	{
	    if (errno == EWOULDBLOCK)
	    {
		redTxBuffsize [line] += sockBytes [line];
		if (setsockopt(redFd [line], SOL_SOCKET, SO_SNDBUF,
			(char *) &(redTxBuffsize [line]),
			sizeof (redTxBuffsize [line])) == ERROR)
		{
		    close (redFd [line]);
		    redFd [line] = 0;
		    partnerInet [line] = 0;
		    return (0);
		}
		return (sv85Flush (line));
	    }
	    else
	    {
		close (redFd [line]);
		redFd [line] = 0;
		partnerInet [line] = 0;
		return (0);
	    }
	}
	else if (bytesSent != sockBytes [line])
	{
	    sockBytes [line] -= bytesSent;
	    memmove (sockBuff [line], sockBuff [line] + bytesSent,
			sockBytes [line]);
	    return (sv85Flush (line));
	}
	sockBytes [line] = 0;
	return (1);
}

/****************************************************************/
/*								*/
/*		PUT A MESSAGE INTO THE TCP/IP BUFFER		*/
/*								*/
/****************************************************************/

int putSv85Bytes (short msgType, char *buff, int nbytes, unsigned int line)
{
	/*
	See if there's room in the buffer. If not, make some.
	*/

	if ((sockBytes [line] + sizeof (msgType) + nbytes) > sizeof (sockBuff [line]))
	{
	    sv85Flush (line);
	}
	memmove (sockBuff [line] + sockBytes [line], (char *) &msgType,
			sizeof (msgType));
	sockBytes [line] += sizeof(msgType);
	memmove (sockBuff [line] + sockBytes [line], buff, nbytes);
	sockBytes  [line] += nbytes;
	return (0);
}

/****************************************************************/
/*								*/
/*		GET BYTE STREAM FROM THE RX BUFFER		*/
/*								*/
/****************************************************************/

int getSv85Bytes (char *buff, int nbytes, unsigned int line)
{
	memmove (buff, tcpRxBuff [line] + tcpRxBuffIdx [line], nbytes);
	tcpRxBuffIdx [line] += nbytes;
	tcpRxBuffBytes [line] -= nbytes;
	return (0);
}

/****************************************************************/
/*								*/
/*	DETERMINE IF THE BUFFER HAS A COMPLETE MESSAGE		*/
/*								*/
/****************************************************************/

int haveCompleteSv85Msg (unsigned int line)
{
	short	msgType;

	if (tcpRxBuffBytes [line] < sizeof (short))
	{
	    return (0);
	}
	memmove ((char *) &msgType, tcpRxBuff [line] + tcpRxBuffIdx [line],
			sizeof (msgType));
	if (msgType == MOD_ACK)
	{
	    if (tcpRxBuffBytes [line] >= (sizeof(short) + sizeof (MOD_ACK_MSG)))
	    {
		return(1);
	    }
	    else
	    {
		return (0);
	    }
	}
	else if (msgType == MOD_REDUPDT)
	{
	    if (tcpRxBuffBytes [line] >= (sizeof(short) + sizeof (MOD_UPDT_MSG)))
	    {
		return (1);
	    }
	    else
	    {
		return (0);
	    }
	}
	printf ("Sv85 redundant socket got out of sync!\n");
	return (0);
}

/****************************************************************/
/*								*/
/*		CREATE THE TCP & UDP SOCKETS			*/
/*								*/
/****************************************************************/

int sv85SockCreate (unsigned short line)
{
	struct sockaddr_in	addr;

	/*
	Create and bind the UDP socket.
	*/

	udpFd [line] = socket (PF_INET, SOCK_DGRAM, 0);
	if (udpFd [line] == ERROR)
	{
	    perror ("Can't create SV85 UDP socket");
	    return (ERROR);
	}
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl (INADDR_ANY);
	addr.sin_port = htons (MOD_PORT);
	if (bind (udpFd [line], (struct sockaddr *) &addr, sizeof (addr)) == ERROR)
	{
	    perror ("Can't bind SV85 UDP socket");
	    return (ERROR);
	}
	/*
	Create and bind the TCP socket.
	*/

	tcpFd [line] = socket (PF_INET, SOCK_STREAM, 0);
	if (tcpFd [line] == ERROR)
	{
	    perror ("Can't create SV85 TCP socket");
	    return (ERROR);
	}
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl (INADDR_ANY);
	addr.sin_port = htons (MOD_PORT);

	if (bind (tcpFd [line], (struct sockaddr *) &addr, sizeof (addr)) == ERROR)
	{
	    perror ("Can't bind SV85 TCP socket");
	    return (ERROR);
	}
	listen (tcpFd [line], 2);

	return (1);
}

/****************************************************************/
/*								*/
/*		CHECK THE STATE OF THE UDP SOCKET		*/
/*								*/
/*	This routine is to be called every 1/4 second. It	*/
/*	will check the UDP socket to see if there is any	*/
/*	available input datagram. If the data is a boot		*/
/*	request from the ICM, then we will send the shared	*/
/*	memory address, exec address, block data address, and	*/
/*	then invoke the download. If not from the ICM, then we	*/
/*	post the address for action by the redundancy task.	*/
/*								*/
/****************************************************************/

void sv85UDPCheck (unsigned short line)
{
	fd_set			sv85Fds;
	struct sockaddr_in	addr;
	struct timeval		pollTimer;
	int			addrSize;

	/*
	See if we haven't heard from the redundant partner.
	*/
	if (partnerInet [line] == 0)
	{
	    short		message;

	    /*
	    We haven't. If we know his Inet address, boink him.
	    When he responds, we'll post his address which will
	    make the redundant update task connect.
	    */

/*	    if (prestable [line]->partnerInet)
	    {
		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr = htonl (prestable [line] -> partnerInet);
		addr.sin_port = htons ((MOD_PORT + line));
		sendto (udpFd [line], (char *) &message, sizeof (message),
			0, (struct sockaddr *) &addr, sizeof (addr));
	    }
*/
	}

	/*
	Check the UDP port to see if there's anything there to read.
	If so, drain the socket completely.
	*/

	FD_ZERO (&sv85Fds);
	FD_SET (udpFd [line], &sv85Fds);
	pollTimer.tv_sec = 0;
	pollTimer.tv_usec = 0;
	while (select (udpFd [line] + 1, &sv85Fds, NULL, NULL, &pollTimer) > 0)
	{

	    /*
	    There was. Get it and then see who it was from
	    */

	    addrSize = sizeof (addr);
	    recvfrom (udpFd [line], sockBuff [line], sizeof (sockBuff [line]), 
		0, (struct sockaddr *) &addr, &addrSize);

	    partnerInet [line] = ntohl(addr.sin_addr.s_addr);
	}
}

/*	End of File:	drvSv85.c */
